#include <stdint.h>
#include <stddef.h>

typedef enum JISCharacterClass {
    JISCharacterClass_RESERVED = 0,
    JISCharacterClass_KANJI_LEVEL_1 = 1,
    JISCharacterClass_KANJI_LEVEL_2 = 2,
    JISCharacterClass_KANJI_LEVEL_3 = 3,
    JISCharacterClass_KANJI_LEVEL_4 = 4,
    JISCharacterClass_JISX0208_NON_KANJI = 9,
    JISCharacterClass_JISX0213_NON_KANJI = 11
} JISCharacterClass;

typedef struct ShrinkingTransliterationMapping {
    uint16_t jis;
    uint32_t us[2];
    uint32_t sus[2];
    JISCharacterClass class_;
    size_t tx_len:8;
    uint16_t tx_jis[4];
    uint32_t tx_us[4];
} ShrinkingTransliterationMapping;

typedef struct URangeToJISMapping {
    uint32_t start:24, end:24;
    const uint16_t *jis;
} URangeToJISMapping;

static const ShrinkingTransliterationMapping tx_mappings[2 * 94 * 94] = {
    {
        0,
        {12288, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {0},
        {12288}
    },
    {
        1,
        {12289, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {1},
        {12289}
    },
    {
        2,
        {12290, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {2},
        {12290}
    },
    {
        3,
        {65292, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {3},
        {65292}
    },
    {
        4,
        {65294, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {4},
        {65294}
    },
    {
        5,
        {12539, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {5},
        {12539}
    },
    {
        6,
        {65306, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {6},
        {65306}
    },
    {
        7,
        {65307, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {7},
        {65307}
    },
    {
        8,
        {65311, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {8},
        {65311}
    },
    {
        9,
        {65281, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {9},
        {65281}
    },
    {
        10,
        {12443, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {10},
        {12443}
    },
    {
        11,
        {12444, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {11},
        {12444}
    },
    {
        12,
        {180, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {12},
        {180}
    },
    {
        13,
        {65344, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {13},
        {65344}
    },
    {
        14,
        {168, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {14},
        {168}
    },
    {
        15,
        {65342, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {15},
        {65342}
    },
    {
        16,
        {65507, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {16},
        {65507}
    },
    {
        17,
        {65343, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {17},
        {65343}
    },
    {
        18,
        {12541, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {18},
        {12541}
    },
    {
        19,
        {12542, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {19},
        {12542}
    },
    {
        20,
        {12445, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {20},
        {12445}
    },
    {
        21,
        {12446, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {21},
        {12446}
    },
    {
        22,
        {12291, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {22},
        {12291}
    },
    {
        23,
        {20189, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {23},
        {20189}
    },
    {
        24,
        {12293, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {24},
        {12293}
    },
    {
        25,
        {12294, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {25},
        {12294}
    },
    {
        26,
        {12295, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {26},
        {12295}
    },
    {
        27,
        {12540, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {27},
        {12540}
    },
    {
        28,
        {8213, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {28},
        {8213}
    },
    {
        29,
        {8208, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {29},
        {8208}
    },
    {
        30,
        {65295, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {30},
        {65295}
    },
    {
        31,
        {65340, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {31},
        {65340}
    },
    {
        32,
        {65374, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {32},
        {65374}
    },
    {
        33,
        {8741, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {33},
        {8741}
    },
    {
        34,
        {65372, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {34},
        {65372}
    },
    {
        35,
        {8230, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {35},
        {8230}
    },
    {
        36,
        {8229, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {36},
        {8229}
    },
    {
        37,
        {8216, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {37},
        {8216}
    },
    {
        38,
        {8217, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {38},
        {8217}
    },
    {
        39,
        {8220, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {39},
        {8220}
    },
    {
        40,
        {8221, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {40},
        {8221}
    },
    {
        41,
        {65288, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {41},
        {65288}
    },
    {
        42,
        {65289, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {42},
        {65289}
    },
    {
        43,
        {12308, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {43},
        {12308}
    },
    {
        44,
        {12309, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {44},
        {12309}
    },
    {
        45,
        {65339, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {45},
        {65339}
    },
    {
        46,
        {65341, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {46},
        {65341}
    },
    {
        47,
        {65371, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {47},
        {65371}
    },
    {
        48,
        {65373, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {48},
        {65373}
    },
    {
        49,
        {12296, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {49},
        {12296}
    },
    {
        50,
        {12297, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {50},
        {12297}
    },
    {
        51,
        {12298, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {51},
        {12298}
    },
    {
        52,
        {12299, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {52},
        {12299}
    },
    {
        53,
        {12300, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {53},
        {12300}
    },
    {
        54,
        {12301, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {54},
        {12301}
    },
    {
        55,
        {12302, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {55},
        {12302}
    },
    {
        56,
        {12303, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {56},
        {12303}
    },
    {
        57,
        {12304, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {57},
        {12304}
    },
    {
        58,
        {12305, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {58},
        {12305}
    },
    {
        59,
        {65291, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {59},
        {65291}
    },
    {
        60,
        {65293, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {60},
        {65293}
    },
    {
        61,
        {177, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {61},
        {177}
    },
    {
        62,
        {215, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {62},
        {215}
    },
    {
        63,
        {247, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {63},
        {247}
    },
    {
        64,
        {65309, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {64},
        {65309}
    },
    {
        65,
        {8800, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {65},
        {8800}
    },
    {
        66,
        {65308, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {66},
        {65308}
    },
    {
        67,
        {65310, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {67},
        {65310}
    },
    {
        68,
        {8806, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {68},
        {8806}
    },
    {
        69,
        {8807, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {69},
        {8807}
    },
    {
        70,
        {8734, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {70},
        {8734}
    },
    {
        71,
        {8756, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {71},
        {8756}
    },
    {
        72,
        {9794, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {72},
        {9794}
    },
    {
        73,
        {9792, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {73},
        {9792}
    },
    {
        74,
        {176, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {74},
        {176}
    },
    {
        75,
        {8242, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {75},
        {8242}
    },
    {
        76,
        {8243, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {76},
        {8243}
    },
    {
        77,
        {8451, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {77},
        {8451}
    },
    {
        78,
        {65509, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {78},
        {65509}
    },
    {
        79,
        {65284, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {79},
        {65284}
    },
    {
        80,
        {65504, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {80},
        {65504}
    },
    {
        81,
        {65505, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {81},
        {65505}
    },
    {
        82,
        {65285, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {82},
        {65285}
    },
    {
        83,
        {65283, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {83},
        {65283}
    },
    {
        84,
        {65286, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {84},
        {65286}
    },
    {
        85,
        {65290, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {85},
        {65290}
    },
    {
        86,
        {65312, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {86},
        {65312}
    },
    {
        87,
        {167, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {87},
        {167}
    },
    {
        88,
        {9734, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {88},
        {9734}
    },
    {
        89,
        {9733, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {89},
        {9733}
    },
    {
        90,
        {9675, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {90},
        {9675}
    },
    {
        91,
        {9679, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {91},
        {9679}
    },
    {
        92,
        {9678, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {92},
        {9678}
    },
    {
        93,
        {9671, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {93},
        {9671}
    },
    {
        94,
        {9670, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {94},
        {9670}
    },
    {
        95,
        {9633, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {95},
        {9633}
    },
    {
        96,
        {9632, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {96},
        {9632}
    },
    {
        97,
        {9651, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {97},
        {9651}
    },
    {
        98,
        {9650, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {98},
        {9650}
    },
    {
        99,
        {9661, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {99},
        {9661}
    },
    {
        100,
        {9660, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {100},
        {9660}
    },
    {
        101,
        {8251, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {101},
        {8251}
    },
    {
        102,
        {12306, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {102},
        {12306}
    },
    {
        103,
        {8594, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {103},
        {8594}
    },
    {
        104,
        {8592, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {104},
        {8592}
    },
    {
        105,
        {8593, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {105},
        {8593}
    },
    {
        106,
        {8595, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {106},
        {8595}
    },
    {
        107,
        {12307, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {107},
        {12307}
    },
    {
        108,
        {65287, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        109,
        {65282, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        110,
        {12339, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        111,
        {12339, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        112,
        {12339, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        113,
        {12340, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        114,
        {12341, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        115,
        {12347, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        116,
        {12348, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        117,
        {12543, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        118,
        {12447, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        119,
        {8712, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {119},
        {8712}
    },
    {
        120,
        {8715, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {120},
        {8715}
    },
    {
        121,
        {8838, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {121},
        {8838}
    },
    {
        122,
        {8839, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {122},
        {8839}
    },
    {
        123,
        {8834, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {123},
        {8834}
    },
    {
        124,
        {8835, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {124},
        {8835}
    },
    {
        125,
        {8746, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {125},
        {8746}
    },
    {
        126,
        {8745, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {126},
        {8745}
    },
    {
        127,
        {8836, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        128,
        {8837, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        129,
        {8842, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        130,
        {8843, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        131,
        {8713, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        132,
        {8709, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        133,
        {8965, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        134,
        {8966, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        135,
        {8743, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {135},
        {8743}
    },
    {
        136,
        {8744, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {136},
        {8744}
    },
    {
        137,
        {65506, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {137},
        {65506}
    },
    {
        138,
        {8658, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {138},
        {8658}
    },
    {
        139,
        {8660, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {139},
        {8660}
    },
    {
        140,
        {8704, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {140},
        {8704}
    },
    {
        141,
        {8707, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {141},
        {8707}
    },
    {
        142,
        {8853, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        143,
        {8854, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        144,
        {8855, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        145,
        {8742, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        146,
        {8742, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        147,
        {65375, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {41},
        {65288}
    },
    {
        148,
        {65376, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {42},
        {65289}
    },
    {
        149,
        {12312, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {45},
        {65339}
    },
    {
        150,
        {12313, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {46},
        {65341}
    },
    {
        151,
        {12310, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {57},
        {12304}
    },
    {
        152,
        {12311, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {58},
        {12305}
    },
    {
        153,
        {8736, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {153},
        {8736}
    },
    {
        154,
        {8869, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {154},
        {8869}
    },
    {
        155,
        {8978, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {155},
        {8978}
    },
    {
        156,
        {8706, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {156},
        {8706}
    },
    {
        157,
        {8711, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {157},
        {8711}
    },
    {
        158,
        {8801, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {158},
        {8801}
    },
    {
        159,
        {8786, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {159},
        {8786}
    },
    {
        160,
        {8810, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {160},
        {8810}
    },
    {
        161,
        {8811, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {161},
        {8811}
    },
    {
        162,
        {8730, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {162},
        {8730}
    },
    {
        163,
        {8765, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {163},
        {8765}
    },
    {
        164,
        {8733, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {164},
        {8733}
    },
    {
        165,
        {8757, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {165},
        {8757}
    },
    {
        166,
        {8747, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {166},
        {8747}
    },
    {
        167,
        {8748, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {167},
        {8748}
    },
    {
        168,
        {8802, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {65},
        {8800}
    },
    {
        169,
        {8771, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {64},
        {65309}
    },
    {
        170,
        {8773, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        171,
        {8776, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        172,
        {8822, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        173,
        {8823, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        174,
        {8596, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        175,
        {8491, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {175},
        {8491}
    },
    {
        176,
        {8240, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {176},
        {8240}
    },
    {
        177,
        {9839, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {177},
        {9839}
    },
    {
        178,
        {9837, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {178},
        {9837}
    },
    {
        179,
        {9834, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {179},
        {9834}
    },
    {
        180,
        {8224, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {180},
        {8224}
    },
    {
        181,
        {8225, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {181},
        {8225}
    },
    {
        182,
        {182, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {182},
        {182}
    },
    {
        183,
        {9838, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        184,
        {9835, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        185,
        {9836, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        186,
        {9833, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        187,
        {9711, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {187},
        {9711}
    },
    {
        188,
        {9655, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        189,
        {9654, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        190,
        {9665, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        191,
        {9664, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        192,
        {8599, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        193,
        {8600, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        194,
        {8598, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        195,
        {8601, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        196,
        {8644, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        197,
        {8680, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        198,
        {8678, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        199,
        {8679, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        200,
        {8681, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        201,
        {10548, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        202,
        {10549, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        203,
        {65296, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {203},
        {65296}
    },
    {
        204,
        {65297, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {204},
        {65297}
    },
    {
        205,
        {65298, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {205},
        {65298}
    },
    {
        206,
        {65299, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {206},
        {65299}
    },
    {
        207,
        {65300, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {207},
        {65300}
    },
    {
        208,
        {65301, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {208},
        {65301}
    },
    {
        209,
        {65302, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {209},
        {65302}
    },
    {
        210,
        {65303, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {210},
        {65303}
    },
    {
        211,
        {65304, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {211},
        {65304}
    },
    {
        212,
        {65305, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {212},
        {65305}
    },
    {
        213,
        {10687, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        214,
        {9673, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        215,
        {12349, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        216,
        {65094, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        217,
        {65093, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        218,
        {9702, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        219,
        {8226, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        220,
        {65313, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {220},
        {65313}
    },
    {
        221,
        {65314, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {221},
        {65314}
    },
    {
        222,
        {65315, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {222},
        {65315}
    },
    {
        223,
        {65316, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {223},
        {65316}
    },
    {
        224,
        {65317, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {224},
        {65317}
    },
    {
        225,
        {65318, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {225},
        {65318}
    },
    {
        226,
        {65319, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {226},
        {65319}
    },
    {
        227,
        {65320, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {227},
        {65320}
    },
    {
        228,
        {65321, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {228},
        {65321}
    },
    {
        229,
        {65322, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {229},
        {65322}
    },
    {
        230,
        {65323, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {230},
        {65323}
    },
    {
        231,
        {65324, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {231},
        {65324}
    },
    {
        232,
        {65325, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {232},
        {65325}
    },
    {
        233,
        {65326, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {233},
        {65326}
    },
    {
        234,
        {65327, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {234},
        {65327}
    },
    {
        235,
        {65328, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {235},
        {65328}
    },
    {
        236,
        {65329, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {236},
        {65329}
    },
    {
        237,
        {65330, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {237},
        {65330}
    },
    {
        238,
        {65331, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {238},
        {65331}
    },
    {
        239,
        {65332, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {239},
        {65332}
    },
    {
        240,
        {65333, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {240},
        {65333}
    },
    {
        241,
        {65334, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {241},
        {65334}
    },
    {
        242,
        {65335, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {242},
        {65335}
    },
    {
        243,
        {65336, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {243},
        {65336}
    },
    {
        244,
        {65337, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {244},
        {65337}
    },
    {
        245,
        {65338, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {245},
        {65338}
    },
    {
        246,
        {8723, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        247,
        {8501, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        248,
        {8463, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        249,
        {13259, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        250,
        {8467, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        251,
        {8487, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        252,
        {65345, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {252},
        {65345}
    },
    {
        253,
        {65346, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {253},
        {65346}
    },
    {
        254,
        {65347, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {254},
        {65347}
    },
    {
        255,
        {65348, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {255},
        {65348}
    },
    {
        256,
        {65349, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {256},
        {65349}
    },
    {
        257,
        {65350, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {257},
        {65350}
    },
    {
        258,
        {65351, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {258},
        {65351}
    },
    {
        259,
        {65352, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {259},
        {65352}
    },
    {
        260,
        {65353, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {260},
        {65353}
    },
    {
        261,
        {65354, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {261},
        {65354}
    },
    {
        262,
        {65355, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {262},
        {65355}
    },
    {
        263,
        {65356, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {263},
        {65356}
    },
    {
        264,
        {65357, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {264},
        {65357}
    },
    {
        265,
        {65358, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {265},
        {65358}
    },
    {
        266,
        {65359, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {266},
        {65359}
    },
    {
        267,
        {65360, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {267},
        {65360}
    },
    {
        268,
        {65361, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {268},
        {65361}
    },
    {
        269,
        {65362, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {269},
        {65362}
    },
    {
        270,
        {65363, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {270},
        {65363}
    },
    {
        271,
        {65364, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {271},
        {65364}
    },
    {
        272,
        {65365, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {272},
        {65365}
    },
    {
        273,
        {65366, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {273},
        {65366}
    },
    {
        274,
        {65367, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {274},
        {65367}
    },
    {
        275,
        {65368, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {275},
        {65368}
    },
    {
        276,
        {65369, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {276},
        {65369}
    },
    {
        277,
        {65370, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {277},
        {65370}
    },
    {
        278,
        {12448, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        279,
        {8211, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {29},
        {8208}
    },
    {
        280,
        {10746, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        281,
        {10747, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        282,
        {12353, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {282},
        {12353}
    },
    {
        283,
        {12354, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {283},
        {12354}
    },
    {
        284,
        {12355, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {284},
        {12355}
    },
    {
        285,
        {12356, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {285},
        {12356}
    },
    {
        286,
        {12357, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {286},
        {12357}
    },
    {
        287,
        {12358, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {287},
        {12358}
    },
    {
        288,
        {12359, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {288},
        {12359}
    },
    {
        289,
        {12360, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {289},
        {12360}
    },
    {
        290,
        {12361, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {290},
        {12361}
    },
    {
        291,
        {12362, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {291},
        {12362}
    },
    {
        292,
        {12363, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {292},
        {12363}
    },
    {
        293,
        {12364, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {293},
        {12364}
    },
    {
        294,
        {12365, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {294},
        {12365}
    },
    {
        295,
        {12366, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {295},
        {12366}
    },
    {
        296,
        {12367, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {296},
        {12367}
    },
    {
        297,
        {12368, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {297},
        {12368}
    },
    {
        298,
        {12369, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {298},
        {12369}
    },
    {
        299,
        {12370, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {299},
        {12370}
    },
    {
        300,
        {12371, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {300},
        {12371}
    },
    {
        301,
        {12372, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {301},
        {12372}
    },
    {
        302,
        {12373, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {302},
        {12373}
    },
    {
        303,
        {12374, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {303},
        {12374}
    },
    {
        304,
        {12375, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {304},
        {12375}
    },
    {
        305,
        {12376, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {305},
        {12376}
    },
    {
        306,
        {12377, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {306},
        {12377}
    },
    {
        307,
        {12378, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {307},
        {12378}
    },
    {
        308,
        {12379, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {308},
        {12379}
    },
    {
        309,
        {12380, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {309},
        {12380}
    },
    {
        310,
        {12381, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {310},
        {12381}
    },
    {
        311,
        {12382, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {311},
        {12382}
    },
    {
        312,
        {12383, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {312},
        {12383}
    },
    {
        313,
        {12384, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {313},
        {12384}
    },
    {
        314,
        {12385, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {314},
        {12385}
    },
    {
        315,
        {12386, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {315},
        {12386}
    },
    {
        316,
        {12387, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {316},
        {12387}
    },
    {
        317,
        {12388, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {317},
        {12388}
    },
    {
        318,
        {12389, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {318},
        {12389}
    },
    {
        319,
        {12390, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {319},
        {12390}
    },
    {
        320,
        {12391, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {320},
        {12391}
    },
    {
        321,
        {12392, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {321},
        {12392}
    },
    {
        322,
        {12393, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {322},
        {12393}
    },
    {
        323,
        {12394, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {323},
        {12394}
    },
    {
        324,
        {12395, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {324},
        {12395}
    },
    {
        325,
        {12396, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {325},
        {12396}
    },
    {
        326,
        {12397, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {326},
        {12397}
    },
    {
        327,
        {12398, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {327},
        {12398}
    },
    {
        328,
        {12399, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {328},
        {12399}
    },
    {
        329,
        {12400, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {329},
        {12400}
    },
    {
        330,
        {12401, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {330},
        {12401}
    },
    {
        331,
        {12402, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {331},
        {12402}
    },
    {
        332,
        {12403, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {332},
        {12403}
    },
    {
        333,
        {12404, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {333},
        {12404}
    },
    {
        334,
        {12405, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {334},
        {12405}
    },
    {
        335,
        {12406, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {335},
        {12406}
    },
    {
        336,
        {12407, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {336},
        {12407}
    },
    {
        337,
        {12408, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {337},
        {12408}
    },
    {
        338,
        {12409, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {338},
        {12409}
    },
    {
        339,
        {12410, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {339},
        {12410}
    },
    {
        340,
        {12411, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {340},
        {12411}
    },
    {
        341,
        {12412, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {341},
        {12412}
    },
    {
        342,
        {12413, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {342},
        {12413}
    },
    {
        343,
        {12414, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {343},
        {12414}
    },
    {
        344,
        {12415, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {344},
        {12415}
    },
    {
        345,
        {12416, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {345},
        {12416}
    },
    {
        346,
        {12417, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {346},
        {12417}
    },
    {
        347,
        {12418, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {347},
        {12418}
    },
    {
        348,
        {12419, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {348},
        {12419}
    },
    {
        349,
        {12420, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {349},
        {12420}
    },
    {
        350,
        {12421, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {350},
        {12421}
    },
    {
        351,
        {12422, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {351},
        {12422}
    },
    {
        352,
        {12423, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {352},
        {12423}
    },
    {
        353,
        {12424, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {353},
        {12424}
    },
    {
        354,
        {12425, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {354},
        {12425}
    },
    {
        355,
        {12426, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {355},
        {12426}
    },
    {
        356,
        {12427, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {356},
        {12427}
    },
    {
        357,
        {12428, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {357},
        {12428}
    },
    {
        358,
        {12429, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {358},
        {12429}
    },
    {
        359,
        {12430, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {359},
        {12430}
    },
    {
        360,
        {12431, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {360},
        {12431}
    },
    {
        361,
        {12432, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {361},
        {12432}
    },
    {
        362,
        {12433, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {362},
        {12433}
    },
    {
        363,
        {12434, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {363},
        {12434}
    },
    {
        364,
        {12435, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {364},
        {12435}
    },
    {
        365,
        {12436, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {459},
        {12532}
    },
    {
        366,
        {12437, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {460},
        {12533}
    },
    {
        367,
        {12438, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {461},
        {12534}
    },
    {
        368,
        {12363, 12442},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        369,
        {12365, 12442},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        370,
        {12367, 12442},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        371,
        {12369, 12442},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        372,
        {12371, 12442},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        373,
        {12449, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        374,
        {12449, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        375,
        {12449, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        376,
        {12449, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {376},
        {12449}
    },
    {
        377,
        {12450, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {377},
        {12450}
    },
    {
        378,
        {12451, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {378},
        {12451}
    },
    {
        379,
        {12452, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {379},
        {12452}
    },
    {
        380,
        {12453, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {380},
        {12453}
    },
    {
        381,
        {12454, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {381},
        {12454}
    },
    {
        382,
        {12455, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {382},
        {12455}
    },
    {
        383,
        {12456, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {383},
        {12456}
    },
    {
        384,
        {12457, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {384},
        {12457}
    },
    {
        385,
        {12458, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {385},
        {12458}
    },
    {
        386,
        {12459, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {386},
        {12459}
    },
    {
        387,
        {12460, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {387},
        {12460}
    },
    {
        388,
        {12461, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {388},
        {12461}
    },
    {
        389,
        {12462, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {389},
        {12462}
    },
    {
        390,
        {12463, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {390},
        {12463}
    },
    {
        391,
        {12464, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {391},
        {12464}
    },
    {
        392,
        {12465, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {392},
        {12465}
    },
    {
        393,
        {12466, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {393},
        {12466}
    },
    {
        394,
        {12467, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {394},
        {12467}
    },
    {
        395,
        {12468, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {395},
        {12468}
    },
    {
        396,
        {12469, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {396},
        {12469}
    },
    {
        397,
        {12470, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {397},
        {12470}
    },
    {
        398,
        {12471, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {398},
        {12471}
    },
    {
        399,
        {12472, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {399},
        {12472}
    },
    {
        400,
        {12473, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {400},
        {12473}
    },
    {
        401,
        {12474, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {401},
        {12474}
    },
    {
        402,
        {12475, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {402},
        {12475}
    },
    {
        403,
        {12476, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {403},
        {12476}
    },
    {
        404,
        {12477, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {404},
        {12477}
    },
    {
        405,
        {12478, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {405},
        {12478}
    },
    {
        406,
        {12479, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {406},
        {12479}
    },
    {
        407,
        {12480, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {407},
        {12480}
    },
    {
        408,
        {12481, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {408},
        {12481}
    },
    {
        409,
        {12482, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {409},
        {12482}
    },
    {
        410,
        {12483, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {410},
        {12483}
    },
    {
        411,
        {12484, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {411},
        {12484}
    },
    {
        412,
        {12485, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {412},
        {12485}
    },
    {
        413,
        {12486, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {413},
        {12486}
    },
    {
        414,
        {12487, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {414},
        {12487}
    },
    {
        415,
        {12488, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {415},
        {12488}
    },
    {
        416,
        {12489, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {416},
        {12489}
    },
    {
        417,
        {12490, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {417},
        {12490}
    },
    {
        418,
        {12491, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {418},
        {12491}
    },
    {
        419,
        {12492, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {419},
        {12492}
    },
    {
        420,
        {12493, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {420},
        {12493}
    },
    {
        421,
        {12494, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {421},
        {12494}
    },
    {
        422,
        {12495, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {422},
        {12495}
    },
    {
        423,
        {12496, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {423},
        {12496}
    },
    {
        424,
        {12497, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {424},
        {12497}
    },
    {
        425,
        {12498, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {425},
        {12498}
    },
    {
        426,
        {12499, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {426},
        {12499}
    },
    {
        427,
        {12500, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {427},
        {12500}
    },
    {
        428,
        {12501, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {428},
        {12501}
    },
    {
        429,
        {12502, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {429},
        {12502}
    },
    {
        430,
        {12503, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {430},
        {12503}
    },
    {
        431,
        {12504, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {431},
        {12504}
    },
    {
        432,
        {12505, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {432},
        {12505}
    },
    {
        433,
        {12506, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {433},
        {12506}
    },
    {
        434,
        {12507, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {434},
        {12507}
    },
    {
        435,
        {12508, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {435},
        {12508}
    },
    {
        436,
        {12509, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {436},
        {12509}
    },
    {
        437,
        {12510, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {437},
        {12510}
    },
    {
        438,
        {12511, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {438},
        {12511}
    },
    {
        439,
        {12512, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {439},
        {12512}
    },
    {
        440,
        {12513, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {440},
        {12513}
    },
    {
        441,
        {12514, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {441},
        {12514}
    },
    {
        442,
        {12515, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {442},
        {12515}
    },
    {
        443,
        {12516, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {443},
        {12516}
    },
    {
        444,
        {12517, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {444},
        {12517}
    },
    {
        445,
        {12518, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {445},
        {12518}
    },
    {
        446,
        {12519, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {446},
        {12519}
    },
    {
        447,
        {12520, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {447},
        {12520}
    },
    {
        448,
        {12521, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {448},
        {12521}
    },
    {
        449,
        {12522, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {449},
        {12522}
    },
    {
        450,
        {12523, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {450},
        {12523}
    },
    {
        451,
        {12524, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {451},
        {12524}
    },
    {
        452,
        {12525, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {452},
        {12525}
    },
    {
        453,
        {12526, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {453},
        {12526}
    },
    {
        454,
        {12527, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {454},
        {12527}
    },
    {
        455,
        {12528, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {455},
        {12528}
    },
    {
        456,
        {12529, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {456},
        {12529}
    },
    {
        457,
        {12530, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {457},
        {12530}
    },
    {
        458,
        {12531, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {458},
        {12531}
    },
    {
        459,
        {12532, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {459},
        {12532}
    },
    {
        460,
        {12533, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {460},
        {12533}
    },
    {
        461,
        {12534, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {461},
        {12534}
    },
    {
        462,
        {12459, 12442},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        463,
        {12461, 12442},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        464,
        {12463, 12442},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        465,
        {12465, 12442},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        466,
        {12467, 12442},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        467,
        {12475, 12442},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        468,
        {12484, 12442},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        469,
        {12488, 12442},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        470,
        {913, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {470},
        {913}
    },
    {
        471,
        {914, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {471},
        {914}
    },
    {
        472,
        {915, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {472},
        {915}
    },
    {
        473,
        {916, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {473},
        {916}
    },
    {
        474,
        {917, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {474},
        {917}
    },
    {
        475,
        {918, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {475},
        {918}
    },
    {
        476,
        {919, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {476},
        {919}
    },
    {
        477,
        {920, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {477},
        {920}
    },
    {
        478,
        {921, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {478},
        {921}
    },
    {
        479,
        {922, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {479},
        {922}
    },
    {
        480,
        {923, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {480},
        {923}
    },
    {
        481,
        {924, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {481},
        {924}
    },
    {
        482,
        {925, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {482},
        {925}
    },
    {
        483,
        {926, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {483},
        {926}
    },
    {
        484,
        {927, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {484},
        {927}
    },
    {
        485,
        {928, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {485},
        {928}
    },
    {
        486,
        {929, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {486},
        {929}
    },
    {
        487,
        {931, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {487},
        {931}
    },
    {
        488,
        {932, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {488},
        {932}
    },
    {
        489,
        {933, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {489},
        {933}
    },
    {
        490,
        {934, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {490},
        {934}
    },
    {
        491,
        {935, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {491},
        {935}
    },
    {
        492,
        {936, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {492},
        {936}
    },
    {
        493,
        {937, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {493},
        {937}
    },
    {
        494,
        {9828, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        495,
        {9824, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        496,
        {9826, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        497,
        {9830, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        498,
        {9825, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        499,
        {9829, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        500,
        {9831, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        501,
        {9827, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        502,
        {945, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {502},
        {945}
    },
    {
        503,
        {946, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {503},
        {946}
    },
    {
        504,
        {947, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {504},
        {947}
    },
    {
        505,
        {948, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {505},
        {948}
    },
    {
        506,
        {949, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {506},
        {949}
    },
    {
        507,
        {950, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {507},
        {950}
    },
    {
        508,
        {951, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {508},
        {951}
    },
    {
        509,
        {952, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {509},
        {952}
    },
    {
        510,
        {953, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {510},
        {953}
    },
    {
        511,
        {954, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {511},
        {954}
    },
    {
        512,
        {955, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {512},
        {955}
    },
    {
        513,
        {956, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {513},
        {956}
    },
    {
        514,
        {957, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {514},
        {957}
    },
    {
        515,
        {958, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {515},
        {958}
    },
    {
        516,
        {959, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {516},
        {959}
    },
    {
        517,
        {960, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {517},
        {960}
    },
    {
        518,
        {961, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {518},
        {961}
    },
    {
        519,
        {963, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {519},
        {963}
    },
    {
        520,
        {964, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {520},
        {964}
    },
    {
        521,
        {965, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {521},
        {965}
    },
    {
        522,
        {966, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {522},
        {966}
    },
    {
        523,
        {967, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {523},
        {967}
    },
    {
        524,
        {968, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {524},
        {968}
    },
    {
        525,
        {969, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {525},
        {969}
    },
    {
        526,
        {962, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        527,
        {9461, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {204},
        {65297}
    },
    {
        528,
        {9462, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {205},
        {65298}
    },
    {
        529,
        {9463, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {206},
        {65299}
    },
    {
        530,
        {9464, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {207},
        {65300}
    },
    {
        531,
        {9465, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {208},
        {65301}
    },
    {
        532,
        {9466, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {209},
        {65302}
    },
    {
        533,
        {9467, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {210},
        {65303}
    },
    {
        534,
        {9468, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {211},
        {65304}
    },
    {
        535,
        {9469, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {212},
        {65305}
    },
    {
        536,
        {9470, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {204, 203},
        {65297, 65296}
    },
    {
        537,
        {9750, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        538,
        {9751, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        539,
        {12320, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        540,
        {9742, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        541,
        {9728, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        542,
        {9729, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        543,
        {9730, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        544,
        {9731, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        545,
        {9832, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        546,
        {9649, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        547,
        {12784, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {390},
        {12463}
    },
    {
        548,
        {12785, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {398},
        {12471}
    },
    {
        549,
        {12786, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {400},
        {12473}
    },
    {
        550,
        {12787, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {415},
        {12488}
    },
    {
        551,
        {12788, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {419},
        {12492}
    },
    {
        552,
        {12789, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {422},
        {12495}
    },
    {
        553,
        {12790, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {425},
        {12498}
    },
    {
        554,
        {12791, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {428},
        {12501}
    },
    {
        555,
        {12792, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {431},
        {12504}
    },
    {
        556,
        {12793, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {434},
        {12507}
    },
    {
        557,
        {12791, 12442},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        558,
        {12794, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {439},
        {12512}
    },
    {
        559,
        {12795, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {448},
        {12521}
    },
    {
        560,
        {12796, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {449},
        {12522}
    },
    {
        561,
        {12797, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {450},
        {12523}
    },
    {
        562,
        {12798, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {451},
        {12524}
    },
    {
        563,
        {12799, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {452},
        {12525}
    },
    {
        564,
        {1040, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {564},
        {1040}
    },
    {
        565,
        {1041, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {565},
        {1041}
    },
    {
        566,
        {1042, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {566},
        {1042}
    },
    {
        567,
        {1043, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {567},
        {1043}
    },
    {
        568,
        {1044, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {568},
        {1044}
    },
    {
        569,
        {1045, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {569},
        {1045}
    },
    {
        570,
        {1025, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {570},
        {1025}
    },
    {
        571,
        {1046, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {571},
        {1046}
    },
    {
        572,
        {1047, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {572},
        {1047}
    },
    {
        573,
        {1048, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {573},
        {1048}
    },
    {
        574,
        {1049, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {574},
        {1049}
    },
    {
        575,
        {1050, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {575},
        {1050}
    },
    {
        576,
        {1051, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {576},
        {1051}
    },
    {
        577,
        {1052, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {577},
        {1052}
    },
    {
        578,
        {1053, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {578},
        {1053}
    },
    {
        579,
        {1054, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {579},
        {1054}
    },
    {
        580,
        {1055, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {580},
        {1055}
    },
    {
        581,
        {1056, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {581},
        {1056}
    },
    {
        582,
        {1057, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {582},
        {1057}
    },
    {
        583,
        {1058, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {583},
        {1058}
    },
    {
        584,
        {1059, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {584},
        {1059}
    },
    {
        585,
        {1060, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {585},
        {1060}
    },
    {
        586,
        {1061, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {586},
        {1061}
    },
    {
        587,
        {1062, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {587},
        {1062}
    },
    {
        588,
        {1063, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {588},
        {1063}
    },
    {
        589,
        {1064, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {589},
        {1064}
    },
    {
        590,
        {1065, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {590},
        {1065}
    },
    {
        591,
        {1066, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {591},
        {1066}
    },
    {
        592,
        {1067, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {592},
        {1067}
    },
    {
        593,
        {1068, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {593},
        {1068}
    },
    {
        594,
        {1069, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {594},
        {1069}
    },
    {
        595,
        {1070, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {595},
        {1070}
    },
    {
        596,
        {1071, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {596},
        {1071}
    },
    {
        597,
        {9150, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        598,
        {9151, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        599,
        {9152, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        600,
        {9153, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        601,
        {9154, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        602,
        {9155, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        603,
        {9156, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        604,
        {9157, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        605,
        {9158, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        606,
        {9159, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        607,
        {9160, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        608,
        {9161, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        609,
        {9162, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        610,
        {9163, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        611,
        {9164, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        612,
        {1072, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {612},
        {1072}
    },
    {
        613,
        {1073, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {613},
        {1073}
    },
    {
        614,
        {1074, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {614},
        {1074}
    },
    {
        615,
        {1075, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {615},
        {1075}
    },
    {
        616,
        {1076, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {616},
        {1076}
    },
    {
        617,
        {1077, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {617},
        {1077}
    },
    {
        618,
        {1105, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {618},
        {1105}
    },
    {
        619,
        {1078, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {619},
        {1078}
    },
    {
        620,
        {1079, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {620},
        {1079}
    },
    {
        621,
        {1080, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {621},
        {1080}
    },
    {
        622,
        {1081, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {622},
        {1081}
    },
    {
        623,
        {1082, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {623},
        {1082}
    },
    {
        624,
        {1083, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {624},
        {1083}
    },
    {
        625,
        {1084, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {625},
        {1084}
    },
    {
        626,
        {1085, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {626},
        {1085}
    },
    {
        627,
        {1086, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {627},
        {1086}
    },
    {
        628,
        {1087, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {628},
        {1087}
    },
    {
        629,
        {1088, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {629},
        {1088}
    },
    {
        630,
        {1089, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {630},
        {1089}
    },
    {
        631,
        {1090, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {631},
        {1090}
    },
    {
        632,
        {1091, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {632},
        {1091}
    },
    {
        633,
        {1092, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {633},
        {1092}
    },
    {
        634,
        {1093, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {634},
        {1093}
    },
    {
        635,
        {1094, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {635},
        {1094}
    },
    {
        636,
        {1095, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {636},
        {1095}
    },
    {
        637,
        {1096, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {637},
        {1096}
    },
    {
        638,
        {1097, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {638},
        {1097}
    },
    {
        639,
        {1098, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {639},
        {1098}
    },
    {
        640,
        {1099, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {640},
        {1099}
    },
    {
        641,
        {1100, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {641},
        {1100}
    },
    {
        642,
        {1101, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {642},
        {1101}
    },
    {
        643,
        {1102, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {643},
        {1102}
    },
    {
        644,
        {1103, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {644},
        {1103}
    },
    {
        645,
        {12535, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {454},
        {12527}
    },
    {
        646,
        {12536, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {455},
        {12528}
    },
    {
        647,
        {12537, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {456},
        {12529}
    },
    {
        648,
        {12538, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {457},
        {12530}
    },
    {
        649,
        {8922, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        650,
        {8923, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        651,
        {8531, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        652,
        {8532, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        653,
        {8533, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        654,
        {10003, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        655,
        {8984, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        656,
        {9251, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        657,
        {9166, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        658,
        {9472, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {658},
        {9472}
    },
    {
        659,
        {9474, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {659},
        {9474}
    },
    {
        660,
        {9484, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {660},
        {9484}
    },
    {
        661,
        {9488, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {661},
        {9488}
    },
    {
        662,
        {9496, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {662},
        {9496}
    },
    {
        663,
        {9492, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {663},
        {9492}
    },
    {
        664,
        {9500, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {664},
        {9500}
    },
    {
        665,
        {9516, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {665},
        {9516}
    },
    {
        666,
        {9508, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {666},
        {9508}
    },
    {
        667,
        {9524, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {667},
        {9524}
    },
    {
        668,
        {9532, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {668},
        {9532}
    },
    {
        669,
        {9473, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {669},
        {9473}
    },
    {
        670,
        {9475, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {670},
        {9475}
    },
    {
        671,
        {9487, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {671},
        {9487}
    },
    {
        672,
        {9491, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {672},
        {9491}
    },
    {
        673,
        {9499, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {673},
        {9499}
    },
    {
        674,
        {9495, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {674},
        {9495}
    },
    {
        675,
        {9507, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {675},
        {9507}
    },
    {
        676,
        {9523, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {676},
        {9523}
    },
    {
        677,
        {9515, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {677},
        {9515}
    },
    {
        678,
        {9531, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {678},
        {9531}
    },
    {
        679,
        {9547, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {679},
        {9547}
    },
    {
        680,
        {9504, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {680},
        {9504}
    },
    {
        681,
        {9519, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {681},
        {9519}
    },
    {
        682,
        {9512, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {682},
        {9512}
    },
    {
        683,
        {9527, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {683},
        {9527}
    },
    {
        684,
        {9535, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {684},
        {9535}
    },
    {
        685,
        {9501, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {685},
        {9501}
    },
    {
        686,
        {9520, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {686},
        {9520}
    },
    {
        687,
        {9509, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {687},
        {9509}
    },
    {
        688,
        {9528, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {688},
        {9528}
    },
    {
        689,
        {9538, -1},
        {-1, -1},
        JISCharacterClass_JISX0208_NON_KANJI,
        1,
        {689},
        {9538}
    },
    {
        690,
        {12881, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {205, 204},
        {65298, 65297}
    },
    {
        691,
        {12882, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {205, 205},
        {65298, 65298}
    },
    {
        692,
        {12883, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {205, 206},
        {65298, 65299}
    },
    {
        693,
        {12884, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {205, 207},
        {65298, 65300}
    },
    {
        694,
        {12885, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {205, 208},
        {65298, 65301}
    },
    {
        695,
        {12886, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {205, 209},
        {65298, 65302}
    },
    {
        696,
        {12887, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {205, 210},
        {65298, 65303}
    },
    {
        697,
        {12888, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {205, 211},
        {65298, 65304}
    },
    {
        698,
        {12889, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {205, 212},
        {65298, 65305}
    },
    {
        699,
        {12890, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {206, 203},
        {65299, 65296}
    },
    {
        700,
        {12891, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {206, 204},
        {65299, 65297}
    },
    {
        701,
        {12892, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {206, 205},
        {65299, 65298}
    },
    {
        702,
        {12893, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {206, 206},
        {65299, 65299}
    },
    {
        703,
        {12894, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {206, 207},
        {65299, 65300}
    },
    {
        704,
        {12895, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {206, 208},
        {65299, 65301}
    },
    {
        705,
        {12977, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {206, 209},
        {65299, 65302}
    },
    {
        706,
        {12978, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {206, 210},
        {65299, 65303}
    },
    {
        707,
        {12979, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {206, 211},
        {65299, 65304}
    },
    {
        708,
        {12980, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {206, 212},
        {65299, 65305}
    },
    {
        709,
        {12981, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {207, 203},
        {65300, 65296}
    },
    {
        710,
        {12982, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {207, 204},
        {65300, 65297}
    },
    {
        711,
        {12983, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {207, 205},
        {65300, 65298}
    },
    {
        712,
        {12984, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {207, 206},
        {65300, 65299}
    },
    {
        713,
        {12985, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {207, 207},
        {65300, 65300}
    },
    {
        714,
        {12986, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {207, 208},
        {65300, 65301}
    },
    {
        715,
        {12987, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {207, 209},
        {65300, 65302}
    },
    {
        716,
        {12988, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {207, 210},
        {65300, 65303}
    },
    {
        717,
        {12989, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {207, 211},
        {65300, 65304}
    },
    {
        718,
        {12990, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {207, 212},
        {65300, 65305}
    },
    {
        719,
        {12991, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {208, 203},
        {65301, 65296}
    },
    {
        720,
        {9680, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        721,
        {9680, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        722,
        {9680, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        723,
        {9680, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        724,
        {9680, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        725,
        {9680, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        726,
        {9680, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        727,
        {9680, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        728,
        {9680, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        729,
        {9681, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        730,
        {9682, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        731,
        {9683, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        732,
        {8252, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {9, 9},
        {65281, 65281}
    },
    {
        733,
        {8263, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {8, 8},
        {65311, 65311}
    },
    {
        734,
        {8264, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {8, 9},
        {65311, 65281}
    },
    {
        735,
        {8265, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {9, 8},
        {65281, 65311}
    },
    {
        736,
        {461, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        737,
        {462, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        738,
        {464, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        739,
        {7742, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        740,
        {7743, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        741,
        {504, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        742,
        {505, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        743,
        {465, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        744,
        {466, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        745,
        {468, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        746,
        {470, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        747,
        {472, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        748,
        {474, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        749,
        {476, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        750,
        {8364, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        751,
        {8364, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        752,
        {8364, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        753,
        {160, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {0},
        {12288}
    },
    {
        754,
        {161, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {9},
        {65281}
    },
    {
        755,
        {164, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        756,
        {166, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        757,
        {169, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        758,
        {170, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        759,
        {171, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        760,
        {173, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {29},
        {8208}
    },
    {
        761,
        {174, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {17},
        {65343}
    },
    {
        762,
        {175, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {16},
        {65507}
    },
    {
        763,
        {178, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        764,
        {179, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        765,
        {183, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {5},
        {12539}
    },
    {
        766,
        {184, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {3},
        {65292}
    },
    {
        767,
        {185, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        768,
        {186, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        769,
        {187, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        770,
        {188, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        771,
        {189, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        772,
        {190, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        773,
        {191, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {8},
        {65311}
    },
    {
        774,
        {192, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        775,
        {193, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        776,
        {194, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        777,
        {195, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        778,
        {196, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        779,
        {197, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        780,
        {198, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        781,
        {199, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        782,
        {200, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        783,
        {201, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        784,
        {202, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        785,
        {203, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        786,
        {204, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        787,
        {205, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        788,
        {206, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        789,
        {207, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        790,
        {208, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        791,
        {209, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        792,
        {210, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        793,
        {211, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        794,
        {212, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        795,
        {213, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        796,
        {214, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        797,
        {216, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        798,
        {217, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        799,
        {218, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        800,
        {219, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        801,
        {220, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        802,
        {221, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        803,
        {222, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        804,
        {223, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        805,
        {224, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        806,
        {225, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        807,
        {226, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        808,
        {227, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        809,
        {228, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        810,
        {229, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        811,
        {230, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        812,
        {231, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        813,
        {232, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        814,
        {233, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {256},
        {65349}
    },
    {
        815,
        {234, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        816,
        {235, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        817,
        {236, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        818,
        {237, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        819,
        {238, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        820,
        {239, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        821,
        {240, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        822,
        {241, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        823,
        {242, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        824,
        {243, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        825,
        {244, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        826,
        {245, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {266},
        {65359}
    },
    {
        827,
        {246, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        828,
        {248, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        829,
        {249, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        830,
        {250, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        831,
        {251, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        832,
        {252, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        833,
        {253, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        834,
        {254, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        835,
        {255, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        836,
        {256, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        837,
        {298, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        838,
        {362, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        839,
        {274, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        840,
        {332, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        841,
        {257, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        842,
        {299, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        843,
        {363, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        844,
        {275, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        845,
        {333, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        846,
        {260, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        847,
        {728, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        848,
        {321, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        849,
        {317, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        850,
        {346, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        851,
        {352, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        852,
        {350, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        853,
        {356, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        854,
        {377, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        855,
        {381, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        856,
        {379, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        857,
        {261, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        858,
        {731, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        859,
        {322, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        860,
        {318, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        861,
        {347, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        862,
        {711, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        863,
        {353, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        864,
        {351, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        865,
        {357, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        866,
        {378, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        867,
        {733, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        868,
        {382, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        869,
        {380, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        870,
        {340, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        871,
        {258, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        872,
        {313, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        873,
        {262, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        874,
        {268, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        875,
        {280, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        876,
        {282, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        877,
        {270, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        878,
        {323, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        879,
        {327, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        880,
        {336, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        881,
        {344, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        882,
        {366, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        883,
        {368, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        884,
        {354, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        885,
        {341, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        886,
        {259, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        887,
        {314, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        888,
        {263, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        889,
        {269, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        890,
        {281, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        891,
        {283, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        892,
        {271, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        893,
        {273, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        894,
        {324, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        895,
        {328, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        896,
        {337, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        897,
        {345, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        898,
        {367, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        899,
        {369, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        900,
        {355, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        901,
        {729, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        902,
        {264, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        903,
        {284, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        904,
        {292, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        905,
        {308, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        906,
        {348, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        907,
        {364, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        908,
        {265, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        909,
        {285, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        910,
        {293, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        911,
        {309, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        912,
        {349, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        913,
        {365, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        914,
        {625, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        915,
        {651, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        916,
        {638, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        917,
        {643, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        918,
        {658, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        919,
        {620, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        920,
        {622, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        921,
        {633, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        922,
        {648, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        923,
        {598, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        924,
        {627, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        925,
        {637, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        926,
        {642, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        927,
        {656, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        928,
        {635, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        929,
        {621, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        930,
        {607, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        931,
        {626, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        932,
        {669, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        933,
        {654, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        934,
        {609, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        935,
        {331, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        936,
        {624, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        937,
        {641, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        938,
        {295, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        939,
        {661, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        940,
        {660, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        941,
        {614, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        942,
        {664, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        943,
        {450, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        944,
        {595, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        945,
        {599, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        946,
        {644, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        947,
        {608, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        948,
        {403, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        949,
        {339, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        950,
        {338, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        951,
        {616, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        952,
        {649, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        953,
        {600, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        954,
        {629, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        955,
        {601, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        956,
        {604, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        957,
        {606, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        958,
        {592, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        959,
        {623, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        960,
        {650, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        961,
        {612, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        962,
        {652, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        963,
        {596, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        964,
        {593, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        965,
        {594, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        966,
        {653, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        967,
        {613, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        968,
        {674, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        969,
        {673, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        970,
        {597, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        971,
        {657, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        972,
        {634, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        973,
        {615, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        974,
        {602, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        975,
        {230, 768},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        976,
        {509, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        977,
        {8048, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        978,
        {8049, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        979,
        {596, 768},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        980,
        {596, 769},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        981,
        {652, 768},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        982,
        {652, 769},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        983,
        {601, 768},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        984,
        {601, 769},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        985,
        {602, 768},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        986,
        {602, 769},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        987,
        {8050, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        988,
        {8051, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        989,
        {865, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        990,
        {712, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        991,
        {716, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        992,
        {720, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        993,
        {721, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        994,
        {774, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        995,
        {8255, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        996,
        {779, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        997,
        {769, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        998,
        {772, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        999,
        {768, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1000,
        {783, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1001,
        {780, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1002,
        {770, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1003,
        {741, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1004,
        {742, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1005,
        {743, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1006,
        {744, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1007,
        {745, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1008,
        {745, 741},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1009,
        {741, 745},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1010,
        {805, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1011,
        {812, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1012,
        {825, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1013,
        {796, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1014,
        {799, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1015,
        {800, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1016,
        {776, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1017,
        {829, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1018,
        {809, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1019,
        {815, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1020,
        {734, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1021,
        {804, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1022,
        {816, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {32},
        {65374}
    },
    {
        1023,
        {828, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1024,
        {820, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {32},
        {65374}
    },
    {
        1025,
        {797, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1026,
        {798, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1027,
        {792, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1028,
        {793, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1029,
        {810, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1030,
        {826, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1031,
        {827, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1032,
        {771, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {32},
        {65374}
    },
    {
        1033,
        {794, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1034,
        {10102, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {204},
        {65297}
    },
    {
        1035,
        {10103, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {205},
        {65298}
    },
    {
        1036,
        {10104, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {206},
        {65299}
    },
    {
        1037,
        {10105, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {207},
        {65300}
    },
    {
        1038,
        {10106, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {208},
        {65301}
    },
    {
        1039,
        {10107, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {209},
        {65302}
    },
    {
        1040,
        {10108, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {210},
        {65303}
    },
    {
        1041,
        {10109, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {211},
        {65304}
    },
    {
        1042,
        {10110, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {212},
        {65305}
    },
    {
        1043,
        {10111, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {204, 203},
        {65297, 65296}
    },
    {
        1044,
        {9451, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {204, 204},
        {65297, 65297}
    },
    {
        1045,
        {9452, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {204, 205},
        {65297, 65298}
    },
    {
        1046,
        {9453, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {204, 206},
        {65297, 65299}
    },
    {
        1047,
        {9454, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {204, 207},
        {65297, 65300}
    },
    {
        1048,
        {9455, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {204, 208},
        {65297, 65301}
    },
    {
        1049,
        {9456, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {204, 209},
        {65297, 65302}
    },
    {
        1050,
        {9457, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {204, 210},
        {65297, 65303}
    },
    {
        1051,
        {9458, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {204, 211},
        {65297, 65304}
    },
    {
        1052,
        {9459, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {204, 212},
        {65297, 65305}
    },
    {
        1053,
        {9460, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {205, 203},
        {65298, 65296}
    },
    {
        1054,
        {8560, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {260},
        {65353}
    },
    {
        1055,
        {8561, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {260, 260},
        {65353, 65353}
    },
    {
        1056,
        {8562, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        3,
        {260, 260, 260},
        {65353, 65353, 65353}
    },
    {
        1057,
        {8563, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {260, 273},
        {65353, 65366}
    },
    {
        1058,
        {8564, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {273},
        {65366}
    },
    {
        1059,
        {8565, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {273, 260},
        {65366, 65353}
    },
    {
        1060,
        {8566, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        3,
        {273, 260, 260},
        {65366, 65353, 65353}
    },
    {
        1061,
        {8567, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        4,
        {273, 260, 260, 260},
        {65366, 65353, 65353, 65353}
    },
    {
        1062,
        {8568, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {260, 275},
        {65353, 65368}
    },
    {
        1063,
        {8569, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {275},
        {65368}
    },
    {
        1064,
        {8570, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {275, 260},
        {65368, 65353}
    },
    {
        1065,
        {8571, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        3,
        {275, 260, 260},
        {65368, 65353, 65353}
    },
    {
        1066,
        {9424, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {252},
        {65345}
    },
    {
        1067,
        {9425, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {253},
        {65346}
    },
    {
        1068,
        {9426, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {254},
        {65347}
    },
    {
        1069,
        {9427, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {255},
        {65348}
    },
    {
        1070,
        {9428, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {256},
        {65349}
    },
    {
        1071,
        {9429, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {257},
        {65350}
    },
    {
        1072,
        {9430, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {258},
        {65351}
    },
    {
        1073,
        {9431, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {259},
        {65352}
    },
    {
        1074,
        {9432, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {260},
        {65353}
    },
    {
        1075,
        {9433, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {261},
        {65354}
    },
    {
        1076,
        {9434, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {262},
        {65355}
    },
    {
        1077,
        {9435, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {263},
        {65356}
    },
    {
        1078,
        {9436, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {264},
        {65357}
    },
    {
        1079,
        {9437, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {265},
        {65358}
    },
    {
        1080,
        {9438, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {266},
        {65359}
    },
    {
        1081,
        {9439, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {267},
        {65360}
    },
    {
        1082,
        {9440, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {268},
        {65361}
    },
    {
        1083,
        {9441, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {269},
        {65362}
    },
    {
        1084,
        {9442, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {270},
        {65363}
    },
    {
        1085,
        {9443, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {271},
        {65364}
    },
    {
        1086,
        {9444, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {272},
        {65365}
    },
    {
        1087,
        {9445, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {273},
        {65366}
    },
    {
        1088,
        {9446, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {274},
        {65367}
    },
    {
        1089,
        {9447, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {275},
        {65368}
    },
    {
        1090,
        {9448, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {276},
        {65369}
    },
    {
        1091,
        {9449, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {277},
        {65370}
    },
    {
        1092,
        {13008, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {377},
        {12450}
    },
    {
        1093,
        {13009, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {379},
        {12452}
    },
    {
        1094,
        {13010, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {381},
        {12454}
    },
    {
        1095,
        {13011, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {383},
        {12456}
    },
    {
        1096,
        {13012, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {385},
        {12458}
    },
    {
        1097,
        {13013, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {386},
        {12459}
    },
    {
        1098,
        {13014, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {388},
        {12461}
    },
    {
        1099,
        {13015, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {390},
        {12463}
    },
    {
        1100,
        {13016, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {392},
        {12465}
    },
    {
        1101,
        {13017, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {394},
        {12467}
    },
    {
        1102,
        {13018, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {396},
        {12469}
    },
    {
        1103,
        {13019, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {398},
        {12471}
    },
    {
        1104,
        {13020, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {400},
        {12473}
    },
    {
        1105,
        {13021, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {402},
        {12475}
    },
    {
        1106,
        {13022, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {404},
        {12477}
    },
    {
        1107,
        {13023, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {406},
        {12479}
    },
    {
        1108,
        {13024, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {408},
        {12481}
    },
    {
        1109,
        {13025, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {411},
        {12484}
    },
    {
        1110,
        {13026, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {413},
        {12486}
    },
    {
        1111,
        {13027, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {415},
        {12488}
    },
    {
        1112,
        {13050, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {452},
        {12525}
    },
    {
        1113,
        {13033, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {422},
        {12495}
    },
    {
        1114,
        {13029, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {418},
        {12491}
    },
    {
        1115,
        {13037, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {434},
        {12507}
    },
    {
        1116,
        {13036, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {431},
        {12504}
    },
    {
        1117,
        {8273, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        1118,
        {8273, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        1119,
        {8273, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        1120,
        {8273, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        1121,
        {8273, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        1122,
        {8273, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        1123,
        {8273, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        1124,
        {8273, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        1125,
        {8273, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        1126,
        {8273, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1127,
        {8258, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1128,
        {9312, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {204},
        {65297}
    },
    {
        1129,
        {9313, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {205},
        {65298}
    },
    {
        1130,
        {9314, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {206},
        {65299}
    },
    {
        1131,
        {9315, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {207},
        {65300}
    },
    {
        1132,
        {9316, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {208},
        {65301}
    },
    {
        1133,
        {9317, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {209},
        {65302}
    },
    {
        1134,
        {9318, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {210},
        {65303}
    },
    {
        1135,
        {9319, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {211},
        {65304}
    },
    {
        1136,
        {9320, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {212},
        {65305}
    },
    {
        1137,
        {9321, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {204, 203},
        {65297, 65296}
    },
    {
        1138,
        {9322, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {204, 204},
        {65297, 65297}
    },
    {
        1139,
        {9323, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {204, 205},
        {65297, 65298}
    },
    {
        1140,
        {9324, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {204, 206},
        {65297, 65299}
    },
    {
        1141,
        {9325, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {204, 207},
        {65297, 65300}
    },
    {
        1142,
        {9326, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {204, 208},
        {65297, 65301}
    },
    {
        1143,
        {9327, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {204, 209},
        {65297, 65302}
    },
    {
        1144,
        {9328, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {204, 210},
        {65297, 65303}
    },
    {
        1145,
        {9329, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {204, 211},
        {65297, 65304}
    },
    {
        1146,
        {9330, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {204, 212},
        {65297, 65305}
    },
    {
        1147,
        {9331, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {205, 203},
        {65298, 65296}
    },
    {
        1148,
        {8544, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {228},
        {65321}
    },
    {
        1149,
        {8545, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {228, 228},
        {65321, 65321}
    },
    {
        1150,
        {8546, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        3,
        {228, 228, 228},
        {65321, 65321, 65321}
    },
    {
        1151,
        {8547, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {228, 241},
        {65321, 65334}
    },
    {
        1152,
        {8548, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {241},
        {65334}
    },
    {
        1153,
        {8549, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {241, 228},
        {65334, 65321}
    },
    {
        1154,
        {8550, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        3,
        {241, 228, 228},
        {65334, 65321, 65321}
    },
    {
        1155,
        {8551, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        4,
        {241, 228, 228, 228},
        {65334, 65321, 65321, 65321}
    },
    {
        1156,
        {8552, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {228, 243},
        {65321, 65336}
    },
    {
        1157,
        {8553, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {243},
        {65336}
    },
    {
        1158,
        {8554, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        2,
        {243, 228},
        {65336, 65321}
    },
    {
        1159,
        {13129, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1160,
        {13076, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1161,
        {13090, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1162,
        {13133, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1163,
        {13080, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1164,
        {13095, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1165,
        {13059, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1166,
        {13110, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1167,
        {13137, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1168,
        {13143, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1169,
        {13069, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1170,
        {13094, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1171,
        {13091, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1172,
        {13099, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1173,
        {13130, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1174,
        {13115, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1175,
        {13212, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1176,
        {13213, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1177,
        {13214, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1178,
        {13198, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1179,
        {13199, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1180,
        {13252, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1181,
        {13217, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1182,
        {8555, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        3,
        {243, 228, 228},
        {65336, 65321, 65321}
    },
    {
        1183,
        {13179, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        1184,
        {13179, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        1185,
        {13179, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        1186,
        {13179, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        1187,
        {13179, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        1188,
        {13179, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        1189,
        {13179, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        1190,
        {13179, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1191,
        {12317, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {39},
        {8220}
    },
    {
        1192,
        {12319, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        1,
        {40},
        {8221}
    },
    {
        1193,
        {8470, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        3,
        {233, 266, 4},
        {65326, 65359, 65294}
    },
    {
        1194,
        {13261, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1195,
        {8481, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1196,
        {12964, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1197,
        {12965, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1198,
        {12966, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1199,
        {12967, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1200,
        {12968, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1201,
        {12849, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        3,
        {41, 1775, 42},
        {65288, 26666, 65289}
    },
    {
        1202,
        {12850, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        3,
        {41, 4148, 42},
        {65288, 26377, 65289}
    },
    {
        1203,
        {12857, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        3,
        {41, 3170, 42},
        {65288, 20195, 65289}
    },
    {
        1204,
        {13182, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1205,
        {13181, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1206,
        {13180, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1207,
        {8750, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        1208,
        {8750, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        1209,
        {8750, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        1210,
        {8750, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1211,
        {8735, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        1212,
        {8735, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        1213,
        {8735, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        1214,
        {8735, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        1215,
        {8735, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1216,
        {8895, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1217,
        {10070, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        1218,
        {10070, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        1219,
        {10070, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        1220,
        {10070, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1221,
        {9758, -1},
        {-1, -1},
        JISCharacterClass_JISX0213_NON_KANJI,
        0,
        {0},
        {0}
    },
    {
        1222,
        {20465, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2043},
        {20534}
    },
    {
        1223,
        {131083, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2795},
        {19976}
    },
    {
        1224,
        {13314, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1863},
        {21916}
    },
    {
        1225,
        {20008, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1226,
        {20015, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1227,
        {20016, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1228,
        {20109, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1229,
        {20193, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1230,
        {20221, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3803},
        {25996}
    },
    {
        1231,
        {20223, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1232,
        {20227, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1233,
        {20235, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1234,
        {20320, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4972},
        {23571}
    },
    {
        1235,
        {20296, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1236,
        {20297, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1237,
        {20310, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1238,
        {20319, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1239,
        {20330, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1240,
        {20332, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1241,
        {20350, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1242,
        {20362, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1243,
        {20372, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1244,
        {20375, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1245,
        {64048, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3837},
        {20398}
    },
    {
        1246,
        {20425, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1247,
        {20448, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1975},
        {20384}
    },
    {
        1248,
        {20481, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1249,
        {20482, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3880},
        {20341}
    },
    {
        1250,
        {20494, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1251,
        {20504, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1252,
        {20519, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1253,
        {20526, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1254,
        {20544, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1552},
        {33521}
    },
    {
        1255,
        {20539, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1256,
        {20545, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {6245},
        {31281}
    },
    {
        1257,
        {20628, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1258,
        {20684, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1259,
        {20722, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4513},
        {20682}
    },
    {
        1260,
        {20688, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1261,
        {20710, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1262,
        {64049, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3053},
        {20711}
    },
    {
        1263,
        {20742, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1264,
        {20739, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1265,
        {20747, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1266,
        {20766, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2541},
        {29246}
    },
    {
        1267,
        {20789, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1268,
        {20810, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4539},
        {20812}
    },
    {
        1269,
        {64050, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4081},
        {20813}
    },
    {
        1270,
        {20821, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1271,
        {20823, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1272,
        {13493, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {6115},
        {30339}
    },
    {
        1273,
        {20893, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1904},
        {23452}
    },
    {
        1274,
        {20931, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1275,
        {20938, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1276,
        {20958, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {5898},
        {29064}
    },
    {
        1277,
        {20962, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4009},
        {20961}
    },
    {
        1278,
        {20974, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3846},
        {39080}
    },
    {
        1279,
        {20993, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1280,
        {13531, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2250},
        {21151}
    },
    {
        1281,
        {21011, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1282,
        {21013, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1283,
        {21065, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1284,
        {21079, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1285,
        {21089, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1286,
        {21139, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1287,
        {21192, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4141},
        {21191}
    },
    {
        1288,
        {64051, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3910},
        {21193}
    },
    {
        1289,
        {21196, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1290,
        {21200, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4093},
        {29467}
    },
    {
        1291,
        {21206, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4613},
        {21207}
    },
    {
        1292,
        {21211, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4619},
        {21235}
    },
    {
        1293,
        {64052, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2021},
        {21220}
    },
    {
        1294,
        {21232, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1295,
        {21243, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1296,
        {21248, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1297,
        {21255, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4623},
        {21254}
    },
    {
        1298,
        {21276, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1299,
        {64053, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3725},
        {21329}
    },
    {
        1300,
        {21345, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1301,
        {21347, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1302,
        {21373, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3109},
        {21363}
    },
    {
        1303,
        {21395, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1304,
        {21405, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1305,
        {21426, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1306,
        {21522, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4702},
        {21668}
    },
    {
        1307,
        {21543, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1308,
        {21581, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4668},
        {21565}
    },
    {
        1309,
        {21660, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4702},
        {21668}
    },
    {
        1310,
        {21611, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1311,
        {21620, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1312,
        {21631, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1313,
        {21640, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1314,
        {21654, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1315,
        {21665, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1316,
        {21673, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1317,
        {21702, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1318,
        {21759, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1319,
        {21774, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1320,
        {21803, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1321,
        {21813, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1322,
        {21840, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1323,
        {21854, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1411},
        {21782}
    },
    {
        1324,
        {21889, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1325,
        {21894, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3398},
        {21746}
    },
    {
        1326,
        {21902, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4704},
        {21692}
    },
    {
        1327,
        {64054, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1760},
        {21917}
    },
    {
        1328,
        {21933, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1329,
        {21966, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3616},
        {32629}
    },
    {
        1330,
        {64055, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3213},
        {22022}
    },
    {
        1331,
        {22024, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1332,
        {22030, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1333,
        {22075, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1334,
        {22089, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4722},
        {21846}
    },
    {
        1335,
        {22134, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1336,
        {22118, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1337,
        {64056, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1864},
        {22120}
    },
    {
        1338,
        {22127, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1339,
        {22129, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1340,
        {22130, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1341,
        {22169, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1781},
        {22107}
    },
    {
        1342,
        {22174, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3398},
        {21746}
    },
    {
        1343,
        {22185, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1344,
        {22188, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1345,
        {22195, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1346,
        {22217, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1347,
        {22218, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3596},
        {22178}
    },
    {
        1348,
        {22282, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1349,
        {135741, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3441},
        {22303}
    },
    {
        1350,
        {22305, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3441},
        {22303}
    },
    {
        1351,
        {22319, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1352,
        {22323, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1353,
        {22324, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1354,
        {22384, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1355,
        {22391, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1356,
        {22396, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1357,
        {22428, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1358,
        {64015, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1359,
        {135963, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4119},
        {37326}
    },
    {
        1360,
        {22456, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1361,
        {22471, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1362,
        {22472, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1363,
        {22479, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1364,
        {22500, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1365,
        {22509, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1366,
        {22517, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1367,
        {22518, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1368,
        {22527, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3388},
        {27877}
    },
    {
        1369,
        {22537, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1370,
        {64016, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3334},
        {22618}
    },
    {
        1371,
        {22625, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3405},
        {22635}
    },
    {
        1372,
        {22628, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1373,
        {64057, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3882},
        {22592}
    },
    {
        1374,
        {22652, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1375,
        {22665, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1376,
        {22686, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3100},
        {22679}
    },
    {
        1377,
        {64058, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3994},
        {22696}
    },
    {
        1378,
        {22697, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1379,
        {136302, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1380,
        {22738, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1381,
        {22734, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1382,
        {22740, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1383,
        {22746, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1384,
        {22752, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4859},
        {22751}
    },
    {
        1385,
        {22761, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1386,
        {22796, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1387,
        {34369, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1388,
        {22877, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1389,
        {22893, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1390,
        {22923, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1391,
        {22930, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3425},
        {22956}
    },
    {
        1392,
        {22948, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1393,
        {22979, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1394,
        {22994, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1395,
        {23005, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1396,
        {23059, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1397,
        {23075, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1398,
        {23143, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1399,
        {23149, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1400,
        {23159, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1401,
        {23166, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1402,
        {23172, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1403,
        {23198, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1404,
        {23207, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1405,
        {23236, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1406,
        {137405, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1407,
        {23321, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1408,
        {23333, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        1409,
        {21085, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3648},
        {21093}
    },
    {
        1410,
        {20124, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1410},
        {20124}
    },
    {
        1411,
        {21782, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1411},
        {21782}
    },
    {
        1412,
        {23043, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1412},
        {23043}
    },
    {
        1413,
        {38463, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1413},
        {38463}
    },
    {
        1414,
        {21696, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1414},
        {21696}
    },
    {
        1415,
        {24859, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1415},
        {24859}
    },
    {
        1416,
        {25384, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1416},
        {25384}
    },
    {
        1417,
        {23030, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1417},
        {23030}
    },
    {
        1418,
        {36898, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1418},
        {36898}
    },
    {
        1419,
        {33909, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1419},
        {33909}
    },
    {
        1420,
        {33564, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1420},
        {33564}
    },
    {
        1421,
        {31312, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1421},
        {31312}
    },
    {
        1422,
        {24746, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1422},
        {24746}
    },
    {
        1423,
        {25569, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1423},
        {25569}
    },
    {
        1424,
        {28197, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1424},
        {28197}
    },
    {
        1425,
        {26093, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1425},
        {26093}
    },
    {
        1426,
        {33894, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1426},
        {33894}
    },
    {
        1427,
        {33446, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1427},
        {33446}
    },
    {
        1428,
        {39925, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1428},
        {39925}
    },
    {
        1429,
        {26771, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1429},
        {26771}
    },
    {
        1430,
        {22311, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1430},
        {22311}
    },
    {
        1431,
        {26017, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1431},
        {26017}
    },
    {
        1432,
        {25201, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1432},
        {25201}
    },
    {
        1433,
        {23451, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1433},
        {23451}
    },
    {
        1434,
        {22992, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1434},
        {22992}
    },
    {
        1435,
        {34427, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1435},
        {34427}
    },
    {
        1436,
        {39156, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1436},
        {39156}
    },
    {
        1437,
        {32098, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1437},
        {32098}
    },
    {
        1438,
        {32190, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1438},
        {32190}
    },
    {
        1439,
        {39822, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1439},
        {39822}
    },
    {
        1440,
        {25110, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1440},
        {25110}
    },
    {
        1441,
        {31903, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1441},
        {31903}
    },
    {
        1442,
        {34999, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1442},
        {34999}
    },
    {
        1443,
        {23433, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1443},
        {23433}
    },
    {
        1444,
        {24245, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1444},
        {24245}
    },
    {
        1445,
        {25353, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1445},
        {25353}
    },
    {
        1446,
        {26263, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1446},
        {26263}
    },
    {
        1447,
        {26696, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1447},
        {26696}
    },
    {
        1448,
        {38343, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1448},
        {38343}
    },
    {
        1449,
        {38797, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1449},
        {38797}
    },
    {
        1450,
        {26447, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1450},
        {26447}
    },
    {
        1451,
        {20197, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1451},
        {20197}
    },
    {
        1452,
        {20234, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1452},
        {20234}
    },
    {
        1453,
        {20301, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1453},
        {20301}
    },
    {
        1454,
        {20381, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1454},
        {20381}
    },
    {
        1455,
        {20553, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1455},
        {20553}
    },
    {
        1456,
        {22258, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1456},
        {22258}
    },
    {
        1457,
        {22839, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1457},
        {22839}
    },
    {
        1458,
        {22996, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1458},
        {22996}
    },
    {
        1459,
        {23041, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1459},
        {23041}
    },
    {
        1460,
        {23561, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1460},
        {23561}
    },
    {
        1461,
        {24799, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1461},
        {24799}
    },
    {
        1462,
        {24847, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1462},
        {24847}
    },
    {
        1463,
        {24944, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1463},
        {24944}
    },
    {
        1464,
        {26131, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1464},
        {26131}
    },
    {
        1465,
        {26885, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1465},
        {26885}
    },
    {
        1466,
        {28858, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1466},
        {28858}
    },
    {
        1467,
        {30031, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1467},
        {30031}
    },
    {
        1468,
        {30064, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1468},
        {30064}
    },
    {
        1469,
        {31227, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1469},
        {31227}
    },
    {
        1470,
        {32173, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1470},
        {32173}
    },
    {
        1471,
        {32239, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1471},
        {32239}
    },
    {
        1472,
        {32963, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1472},
        {32963}
    },
    {
        1473,
        {33806, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1473},
        {33806}
    },
    {
        1474,
        {34915, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1474},
        {34915}
    },
    {
        1475,
        {35586, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1475},
        {35586}
    },
    {
        1476,
        {36949, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1476},
        {36949}
    },
    {
        1477,
        {36986, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1477},
        {36986}
    },
    {
        1478,
        {21307, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1478},
        {21307}
    },
    {
        1479,
        {20117, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1479},
        {20117}
    },
    {
        1480,
        {20133, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1480},
        {20133}
    },
    {
        1481,
        {22495, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1481},
        {22495}
    },
    {
        1482,
        {32946, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1482},
        {32946}
    },
    {
        1483,
        {37057, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1483},
        {37057}
    },
    {
        1484,
        {30959, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1484},
        {30959}
    },
    {
        1485,
        {19968, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1485},
        {19968}
    },
    {
        1486,
        {22769, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1486},
        {22769}
    },
    {
        1487,
        {28322, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1487},
        {28322}
    },
    {
        1488,
        {36920, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1488},
        {36920}
    },
    {
        1489,
        {31282, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1489},
        {31282}
    },
    {
        1490,
        {33576, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1490},
        {33576}
    },
    {
        1491,
        {33419, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1491},
        {33419}
    },
    {
        1492,
        {39983, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1492},
        {39983}
    },
    {
        1493,
        {20801, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1493},
        {20801}
    },
    {
        1494,
        {21360, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1494},
        {21360}
    },
    {
        1495,
        {21693, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1495},
        {21693}
    },
    {
        1496,
        {21729, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1496},
        {21729}
    },
    {
        1497,
        {22240, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1497},
        {22240}
    },
    {
        1498,
        {23035, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1498},
        {23035}
    },
    {
        1499,
        {24341, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1499},
        {24341}
    },
    {
        1500,
        {39154, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1500},
        {39154}
    },
    {
        1501,
        {28139, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1501},
        {28139}
    },
    {
        1502,
        {32996, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1502},
        {32996}
    },
    {
        1503,
        {34093, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1503},
        {34093}
    },
    {
        1504,
        {38498, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1504},
        {38498}
    },
    {
        1505,
        {38512, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1505},
        {38512}
    },
    {
        1506,
        {38560, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1506},
        {38560}
    },
    {
        1507,
        {38907, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1507},
        {38907}
    },
    {
        1508,
        {21515, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1508},
        {21515}
    },
    {
        1509,
        {21491, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1509},
        {21491}
    },
    {
        1510,
        {23431, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1510},
        {23431}
    },
    {
        1511,
        {28879, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1511},
        {28879}
    },
    {
        1512,
        {32701, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1512},
        {32701}
    },
    {
        1513,
        {36802, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1513},
        {36802}
    },
    {
        1514,
        {38632, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1514},
        {38632}
    },
    {
        1515,
        {21359, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1515},
        {21359}
    },
    {
        1516,
        {40284, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1516},
        {40284}
    },
    {
        1517,
        {31418, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1517},
        {31418}
    },
    {
        1518,
        {19985, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1518},
        {19985}
    },
    {
        1519,
        {30867, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1519},
        {30867}
    },
    {
        1520,
        {33276, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1520},
        {33276}
    },
    {
        1521,
        {28198, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1521},
        {28198}
    },
    {
        1522,
        {22040, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1522},
        {22040}
    },
    {
        1523,
        {21764, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1523},
        {21764}
    },
    {
        1524,
        {27421, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1524},
        {27421}
    },
    {
        1525,
        {34074, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1525},
        {34074}
    },
    {
        1526,
        {39995, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1526},
        {39995}
    },
    {
        1527,
        {23013, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1527},
        {23013}
    },
    {
        1528,
        {21417, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1528},
        {21417}
    },
    {
        1529,
        {28006, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1529},
        {28006}
    },
    {
        1530,
        {29916, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1530},
        {29916}
    },
    {
        1531,
        {38287, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1531},
        {38287}
    },
    {
        1532,
        {22082, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1532},
        {22082}
    },
    {
        1533,
        {20113, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1533},
        {20113}
    },
    {
        1534,
        {36939, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1534},
        {36939}
    },
    {
        1535,
        {38642, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1535},
        {38642}
    },
    {
        1536,
        {33615, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1536},
        {33615}
    },
    {
        1537,
        {39180, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1537},
        {39180}
    },
    {
        1538,
        {21473, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1538},
        {21473}
    },
    {
        1539,
        {21942, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1539},
        {21942}
    },
    {
        1540,
        {23344, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1540},
        {23344}
    },
    {
        1541,
        {24433, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1541},
        {24433}
    },
    {
        1542,
        {26144, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1542},
        {26144}
    },
    {
        1543,
        {26355, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1543},
        {26355}
    },
    {
        1544,
        {26628, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1544},
        {26628}
    },
    {
        1545,
        {27704, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1545},
        {27704}
    },
    {
        1546,
        {27891, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1546},
        {27891}
    },
    {
        1547,
        {27945, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1547},
        {27945}
    },
    {
        1548,
        {29787, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1548},
        {29787}
    },
    {
        1549,
        {30408, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1549},
        {30408}
    },
    {
        1550,
        {31310, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1550},
        {31310}
    },
    {
        1551,
        {38964, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1551},
        {38964}
    },
    {
        1552,
        {33521, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1552},
        {33521}
    },
    {
        1553,
        {34907, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1553},
        {34907}
    },
    {
        1554,
        {35424, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1554},
        {35424}
    },
    {
        1555,
        {37613, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1555},
        {37613}
    },
    {
        1556,
        {28082, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1556},
        {28082}
    },
    {
        1557,
        {30123, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1557},
        {30123}
    },
    {
        1558,
        {30410, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1558},
        {30410}
    },
    {
        1559,
        {39365, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1559},
        {39365}
    },
    {
        1560,
        {24742, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1560},
        {24742}
    },
    {
        1561,
        {35585, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1561},
        {35585}
    },
    {
        1562,
        {36234, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1562},
        {36234}
    },
    {
        1563,
        {38322, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1563},
        {38322}
    },
    {
        1564,
        {27022, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1564},
        {27022}
    },
    {
        1565,
        {21421, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1565},
        {21421}
    },
    {
        1566,
        {20870, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1566},
        {20870}
    },
    {
        1567,
        {22290, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1567},
        {22290}
    },
    {
        1568,
        {22576, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1568},
        {22576}
    },
    {
        1569,
        {22852, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1569},
        {22852}
    },
    {
        1570,
        {23476, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1570},
        {23476}
    },
    {
        1571,
        {24310, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1571},
        {24310}
    },
    {
        1572,
        {24616, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1572},
        {24616}
    },
    {
        1573,
        {25513, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1573},
        {25513}
    },
    {
        1574,
        {25588, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1574},
        {25588}
    },
    {
        1575,
        {27839, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1575},
        {27839}
    },
    {
        1576,
        {28436, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1576},
        {28436}
    },
    {
        1577,
        {28814, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1577},
        {28814}
    },
    {
        1578,
        {28948, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1578},
        {28948}
    },
    {
        1579,
        {29017, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1579},
        {29017}
    },
    {
        1580,
        {29141, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1580},
        {29141}
    },
    {
        1581,
        {29503, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1581},
        {29503}
    },
    {
        1582,
        {32257, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1582},
        {32257}
    },
    {
        1583,
        {33398, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1583},
        {33398}
    },
    {
        1584,
        {33489, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1584},
        {33489}
    },
    {
        1585,
        {34199, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1585},
        {34199}
    },
    {
        1586,
        {36960, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1586},
        {36960}
    },
    {
        1587,
        {37467, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1587},
        {37467}
    },
    {
        1588,
        {40219, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1588},
        {40219}
    },
    {
        1589,
        {22633, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1589},
        {22633}
    },
    {
        1590,
        {26044, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1590},
        {26044}
    },
    {
        1591,
        {27738, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1591},
        {27738}
    },
    {
        1592,
        {29989, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1592},
        {29989}
    },
    {
        1593,
        {20985, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1593},
        {20985}
    },
    {
        1594,
        {22830, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1594},
        {22830}
    },
    {
        1595,
        {22885, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1595},
        {22885}
    },
    {
        1596,
        {24448, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1596},
        {24448}
    },
    {
        1597,
        {24540, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1597},
        {24540}
    },
    {
        1598,
        {25276, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1598},
        {25276}
    },
    {
        1599,
        {26106, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1599},
        {26106}
    },
    {
        1600,
        {27178, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1600},
        {27178}
    },
    {
        1601,
        {27431, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1601},
        {27431}
    },
    {
        1602,
        {27572, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1602},
        {27572}
    },
    {
        1603,
        {29579, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1603},
        {29579}
    },
    {
        1604,
        {32705, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1604},
        {32705}
    },
    {
        1605,
        {35158, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1605},
        {35158}
    },
    {
        1606,
        {40236, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1606},
        {40236}
    },
    {
        1607,
        {40206, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1607},
        {40206}
    },
    {
        1608,
        {40644, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1608},
        {40644}
    },
    {
        1609,
        {23713, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1609},
        {23713}
    },
    {
        1610,
        {27798, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1610},
        {27798}
    },
    {
        1611,
        {33659, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1611},
        {33659}
    },
    {
        1612,
        {20740, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1612},
        {20740}
    },
    {
        1613,
        {23627, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1613},
        {23627}
    },
    {
        1614,
        {25014, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1614},
        {25014}
    },
    {
        1615,
        {33222, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1615},
        {33222}
    },
    {
        1616,
        {26742, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1616},
        {26742}
    },
    {
        1617,
        {29281, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1617},
        {29281}
    },
    {
        1618,
        {20057, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1618},
        {20057}
    },
    {
        1619,
        {20474, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1619},
        {20474}
    },
    {
        1620,
        {21368, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1620},
        {21368}
    },
    {
        1621,
        {24681, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1621},
        {24681}
    },
    {
        1622,
        {28201, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1622},
        {28201}
    },
    {
        1623,
        {31311, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1623},
        {31311}
    },
    {
        1624,
        {38899, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1624},
        {38899}
    },
    {
        1625,
        {19979, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1625},
        {19979}
    },
    {
        1626,
        {21270, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1626},
        {21270}
    },
    {
        1627,
        {20206, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1627},
        {20206}
    },
    {
        1628,
        {20309, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1628},
        {20309}
    },
    {
        1629,
        {20285, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1629},
        {20285}
    },
    {
        1630,
        {20385, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1630},
        {20385}
    },
    {
        1631,
        {20339, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1631},
        {20339}
    },
    {
        1632,
        {21152, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1632},
        {21152}
    },
    {
        1633,
        {21487, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1633},
        {21487}
    },
    {
        1634,
        {22025, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1634},
        {22025}
    },
    {
        1635,
        {22799, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1635},
        {22799}
    },
    {
        1636,
        {23233, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1636},
        {23233}
    },
    {
        1637,
        {23478, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1637},
        {23478}
    },
    {
        1638,
        {23521, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1638},
        {23521}
    },
    {
        1639,
        {31185, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1639},
        {31185}
    },
    {
        1640,
        {26247, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1640},
        {26247}
    },
    {
        1641,
        {26524, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1641},
        {26524}
    },
    {
        1642,
        {26550, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1642},
        {26550}
    },
    {
        1643,
        {27468, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1643},
        {27468}
    },
    {
        1644,
        {27827, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1644},
        {27827}
    },
    {
        1645,
        {28779, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1645},
        {28779}
    },
    {
        1646,
        {29634, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1646},
        {29634}
    },
    {
        1647,
        {31117, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1647},
        {31117}
    },
    {
        1648,
        {31166, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1648},
        {31166}
    },
    {
        1649,
        {31292, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1649},
        {31292}
    },
    {
        1650,
        {31623, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1650},
        {31623}
    },
    {
        1651,
        {33457, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1651},
        {33457}
    },
    {
        1652,
        {33499, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1652},
        {33499}
    },
    {
        1653,
        {33540, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1653},
        {33540}
    },
    {
        1654,
        {33655, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1654},
        {33655}
    },
    {
        1655,
        {33775, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1655},
        {33775}
    },
    {
        1656,
        {33747, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1656},
        {33747}
    },
    {
        1657,
        {34662, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1657},
        {34662}
    },
    {
        1658,
        {35506, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1658},
        {35506}
    },
    {
        1659,
        {22057, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1659},
        {22057}
    },
    {
        1660,
        {36008, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1660},
        {36008}
    },
    {
        1661,
        {36838, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1661},
        {36838}
    },
    {
        1662,
        {36942, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1662},
        {36942}
    },
    {
        1663,
        {38686, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1663},
        {38686}
    },
    {
        1664,
        {34442, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1664},
        {34442}
    },
    {
        1665,
        {20420, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1665},
        {20420}
    },
    {
        1666,
        {23784, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1666},
        {23784}
    },
    {
        1667,
        {25105, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1667},
        {25105}
    },
    {
        1668,
        {29273, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1668},
        {29273}
    },
    {
        1669,
        {30011, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1669},
        {30011}
    },
    {
        1670,
        {33253, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1670},
        {33253}
    },
    {
        1671,
        {33469, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1671},
        {33469}
    },
    {
        1672,
        {34558, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1672},
        {34558}
    },
    {
        1673,
        {36032, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1673},
        {36032}
    },
    {
        1674,
        {38597, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1674},
        {38597}
    },
    {
        1675,
        {39187, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1675},
        {39187}
    },
    {
        1676,
        {39381, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1676},
        {39381}
    },
    {
        1677,
        {20171, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1677},
        {20171}
    },
    {
        1678,
        {20250, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1678},
        {20250}
    },
    {
        1679,
        {35299, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1679},
        {35299}
    },
    {
        1680,
        {22238, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1680},
        {22238}
    },
    {
        1681,
        {22602, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1681},
        {22602}
    },
    {
        1682,
        {22730, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1682},
        {22730}
    },
    {
        1683,
        {24315, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1683},
        {24315}
    },
    {
        1684,
        {24555, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1684},
        {24555}
    },
    {
        1685,
        {24618, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1685},
        {24618}
    },
    {
        1686,
        {24724, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1686},
        {24724}
    },
    {
        1687,
        {24674, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1687},
        {24674}
    },
    {
        1688,
        {25040, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1688},
        {25040}
    },
    {
        1689,
        {25106, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1689},
        {25106}
    },
    {
        1690,
        {25296, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1690},
        {25296}
    },
    {
        1691,
        {25913, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1691},
        {25913}
    },
    {
        1692,
        {39745, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1692},
        {39745}
    },
    {
        1693,
        {26214, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1693},
        {26214}
    },
    {
        1694,
        {26800, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1694},
        {26800}
    },
    {
        1695,
        {28023, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1695},
        {28023}
    },
    {
        1696,
        {28784, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1696},
        {28784}
    },
    {
        1697,
        {30028, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1697},
        {30028}
    },
    {
        1698,
        {30342, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1698},
        {30342}
    },
    {
        1699,
        {32117, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1699},
        {32117}
    },
    {
        1700,
        {33445, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1700},
        {33445}
    },
    {
        1701,
        {34809, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1701},
        {34809}
    },
    {
        1702,
        {38283, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1702},
        {38283}
    },
    {
        1703,
        {38542, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1703},
        {38542}
    },
    {
        1704,
        {35997, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1704},
        {35997}
    },
    {
        1705,
        {20977, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1705},
        {20977}
    },
    {
        1706,
        {21182, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1706},
        {21182}
    },
    {
        1707,
        {22806, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1707},
        {22806}
    },
    {
        1708,
        {21683, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1708},
        {21683}
    },
    {
        1709,
        {23475, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1709},
        {23475}
    },
    {
        1710,
        {23830, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1710},
        {23830}
    },
    {
        1711,
        {24936, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1711},
        {24936}
    },
    {
        1712,
        {27010, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1712},
        {27010}
    },
    {
        1713,
        {28079, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1713},
        {28079}
    },
    {
        1714,
        {30861, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1714},
        {30861}
    },
    {
        1715,
        {33995, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1715},
        {33995}
    },
    {
        1716,
        {34903, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1716},
        {34903}
    },
    {
        1717,
        {35442, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1717},
        {35442}
    },
    {
        1718,
        {37799, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1718},
        {37799}
    },
    {
        1719,
        {39608, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1719},
        {39608}
    },
    {
        1720,
        {28012, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1720},
        {28012}
    },
    {
        1721,
        {39336, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1721},
        {39336}
    },
    {
        1722,
        {34521, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1722},
        {34521}
    },
    {
        1723,
        {22435, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1723},
        {22435}
    },
    {
        1724,
        {26623, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1724},
        {26623}
    },
    {
        1725,
        {34510, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1725},
        {34510}
    },
    {
        1726,
        {37390, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1726},
        {37390}
    },
    {
        1727,
        {21123, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1727},
        {21123}
    },
    {
        1728,
        {22151, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1728},
        {22151}
    },
    {
        1729,
        {21508, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1729},
        {21508}
    },
    {
        1730,
        {24275, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1730},
        {24275}
    },
    {
        1731,
        {25313, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1731},
        {25313}
    },
    {
        1732,
        {25785, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1732},
        {25785}
    },
    {
        1733,
        {26684, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1733},
        {26684}
    },
    {
        1734,
        {26680, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1734},
        {26680}
    },
    {
        1735,
        {27579, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1735},
        {27579}
    },
    {
        1736,
        {29554, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1736},
        {29554}
    },
    {
        1737,
        {30906, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1737},
        {30906}
    },
    {
        1738,
        {31339, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1738},
        {31339}
    },
    {
        1739,
        {35226, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1739},
        {35226}
    },
    {
        1740,
        {35282, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1740},
        {35282}
    },
    {
        1741,
        {36203, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1741},
        {36203}
    },
    {
        1742,
        {36611, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1742},
        {36611}
    },
    {
        1743,
        {37101, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1743},
        {37101}
    },
    {
        1744,
        {38307, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1744},
        {38307}
    },
    {
        1745,
        {38548, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1745},
        {38548}
    },
    {
        1746,
        {38761, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1746},
        {38761}
    },
    {
        1747,
        {23398, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1747},
        {23398}
    },
    {
        1748,
        {23731, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1748},
        {23731}
    },
    {
        1749,
        {27005, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1749},
        {27005}
    },
    {
        1750,
        {38989, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1750},
        {38989}
    },
    {
        1751,
        {38990, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1751},
        {38990}
    },
    {
        1752,
        {25499, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1752},
        {25499}
    },
    {
        1753,
        {31520, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1753},
        {31520}
    },
    {
        1754,
        {27179, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1754},
        {27179}
    },
    {
        1755,
        {27263, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1755},
        {27263}
    },
    {
        1756,
        {26806, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1756},
        {26806}
    },
    {
        1757,
        {39949, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1757},
        {39949}
    },
    {
        1758,
        {28511, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1758},
        {28511}
    },
    {
        1759,
        {21106, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1759},
        {21106}
    },
    {
        1760,
        {21917, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1760},
        {21917}
    },
    {
        1761,
        {24688, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1761},
        {24688}
    },
    {
        1762,
        {25324, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1762},
        {25324}
    },
    {
        1763,
        {27963, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1763},
        {27963}
    },
    {
        1764,
        {28167, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1764},
        {28167}
    },
    {
        1765,
        {28369, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1765},
        {28369}
    },
    {
        1766,
        {33883, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1766},
        {33883}
    },
    {
        1767,
        {35088, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1767},
        {35088}
    },
    {
        1768,
        {36676, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1768},
        {36676}
    },
    {
        1769,
        {19988, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1769},
        {19988}
    },
    {
        1770,
        {39993, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1770},
        {39993}
    },
    {
        1771,
        {21494, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1771},
        {21494}
    },
    {
        1772,
        {26907, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1772},
        {26907}
    },
    {
        1773,
        {27194, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1773},
        {27194}
    },
    {
        1774,
        {38788, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1774},
        {38788}
    },
    {
        1775,
        {26666, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1775},
        {26666}
    },
    {
        1776,
        {20828, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1776},
        {20828}
    },
    {
        1777,
        {31427, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1777},
        {31427}
    },
    {
        1778,
        {33970, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1778},
        {33970}
    },
    {
        1779,
        {37340, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1779},
        {37340}
    },
    {
        1780,
        {37772, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1780},
        {37772}
    },
    {
        1781,
        {22107, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1781},
        {22107}
    },
    {
        1782,
        {40232, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1782},
        {40232}
    },
    {
        1783,
        {26658, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1783},
        {26658}
    },
    {
        1784,
        {33541, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1784},
        {33541}
    },
    {
        1785,
        {33841, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1785},
        {33841}
    },
    {
        1786,
        {31909, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1786},
        {31909}
    },
    {
        1787,
        {21000, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1787},
        {21000}
    },
    {
        1788,
        {33477, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1788},
        {33477}
    },
    {
        1789,
        {29926, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1789},
        {29926}
    },
    {
        1790,
        {20094, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1790},
        {20094}
    },
    {
        1791,
        {20355, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1791},
        {20355}
    },
    {
        1792,
        {20896, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1792},
        {20896}
    },
    {
        1793,
        {23506, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1793},
        {23506}
    },
    {
        1794,
        {21002, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1794},
        {21002}
    },
    {
        1795,
        {21208, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1795},
        {21208}
    },
    {
        1796,
        {21223, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1796},
        {21223}
    },
    {
        1797,
        {24059, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1797},
        {24059}
    },
    {
        1798,
        {21914, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1798},
        {21914}
    },
    {
        1799,
        {22570, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1799},
        {22570}
    },
    {
        1800,
        {23014, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1800},
        {23014}
    },
    {
        1801,
        {23436, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1801},
        {23436}
    },
    {
        1802,
        {23448, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1802},
        {23448}
    },
    {
        1803,
        {23515, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1803},
        {23515}
    },
    {
        1804,
        {24178, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1804},
        {24178}
    },
    {
        1805,
        {24185, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1805},
        {24185}
    },
    {
        1806,
        {24739, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1806},
        {24739}
    },
    {
        1807,
        {24863, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1807},
        {24863}
    },
    {
        1808,
        {24931, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1808},
        {24931}
    },
    {
        1809,
        {25022, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1809},
        {25022}
    },
    {
        1810,
        {25563, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1810},
        {25563}
    },
    {
        1811,
        {25954, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1811},
        {25954}
    },
    {
        1812,
        {26577, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1812},
        {26577}
    },
    {
        1813,
        {26707, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1813},
        {26707}
    },
    {
        1814,
        {26874, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1814},
        {26874}
    },
    {
        1815,
        {27454, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1815},
        {27454}
    },
    {
        1816,
        {27475, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1816},
        {27475}
    },
    {
        1817,
        {27735, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1817},
        {27735}
    },
    {
        1818,
        {28450, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1818},
        {28450}
    },
    {
        1819,
        {28567, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1819},
        {28567}
    },
    {
        1820,
        {28485, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1820},
        {28485}
    },
    {
        1821,
        {29872, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1821},
        {29872}
    },
    {
        1822,
        {29976, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1822},
        {29976}
    },
    {
        1823,
        {30435, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1823},
        {30435}
    },
    {
        1824,
        {30475, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1824},
        {30475}
    },
    {
        1825,
        {31487, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1825},
        {31487}
    },
    {
        1826,
        {31649, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1826},
        {31649}
    },
    {
        1827,
        {31777, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1827},
        {31777}
    },
    {
        1828,
        {32233, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1828},
        {32233}
    },
    {
        1829,
        {32566, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1829},
        {32566}
    },
    {
        1830,
        {32752, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1830},
        {32752}
    },
    {
        1831,
        {32925, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1831},
        {32925}
    },
    {
        1832,
        {33382, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1832},
        {33382}
    },
    {
        1833,
        {33694, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1833},
        {33694}
    },
    {
        1834,
        {35251, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1834},
        {35251}
    },
    {
        1835,
        {35532, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1835},
        {35532}
    },
    {
        1836,
        {36011, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1836},
        {36011}
    },
    {
        1837,
        {36996, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1837},
        {36996}
    },
    {
        1838,
        {37969, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1838},
        {37969}
    },
    {
        1839,
        {38291, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1839},
        {38291}
    },
    {
        1840,
        {38289, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1840},
        {38289}
    },
    {
        1841,
        {38306, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1841},
        {38306}
    },
    {
        1842,
        {38501, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1842},
        {38501}
    },
    {
        1843,
        {38867, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1843},
        {38867}
    },
    {
        1844,
        {39208, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1844},
        {39208}
    },
    {
        1845,
        {33304, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1845},
        {33304}
    },
    {
        1846,
        {20024, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1846},
        {20024}
    },
    {
        1847,
        {21547, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1847},
        {21547}
    },
    {
        1848,
        {23736, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1848},
        {23736}
    },
    {
        1849,
        {24012, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1849},
        {24012}
    },
    {
        1850,
        {29609, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1850},
        {29609}
    },
    {
        1851,
        {30284, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1851},
        {30284}
    },
    {
        1852,
        {30524, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1852},
        {30524}
    },
    {
        1853,
        {23721, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1853},
        {23721}
    },
    {
        1854,
        {32747, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1854},
        {32747}
    },
    {
        1855,
        {36107, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1855},
        {36107}
    },
    {
        1856,
        {38593, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1856},
        {38593}
    },
    {
        1857,
        {38929, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1857},
        {38929}
    },
    {
        1858,
        {38996, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1858},
        {38996}
    },
    {
        1859,
        {39000, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1859},
        {39000}
    },
    {
        1860,
        {20225, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1860},
        {20225}
    },
    {
        1861,
        {20238, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1861},
        {20238}
    },
    {
        1862,
        {21361, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1862},
        {21361}
    },
    {
        1863,
        {21916, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1863},
        {21916}
    },
    {
        1864,
        {22120, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1864},
        {22120}
    },
    {
        1865,
        {22522, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1865},
        {22522}
    },
    {
        1866,
        {22855, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1866},
        {22855}
    },
    {
        1867,
        {23305, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1867},
        {23305}
    },
    {
        1868,
        {23492, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1868},
        {23492}
    },
    {
        1869,
        {23696, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1869},
        {23696}
    },
    {
        1870,
        {24076, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1870},
        {24076}
    },
    {
        1871,
        {24190, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1871},
        {24190}
    },
    {
        1872,
        {24524, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1872},
        {24524}
    },
    {
        1873,
        {25582, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1873},
        {25582}
    },
    {
        1874,
        {26426, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1874},
        {26426}
    },
    {
        1875,
        {26071, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1875},
        {26071}
    },
    {
        1876,
        {26082, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1876},
        {26082}
    },
    {
        1877,
        {26399, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1877},
        {26399}
    },
    {
        1878,
        {26827, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1878},
        {26827}
    },
    {
        1879,
        {26820, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1879},
        {26820}
    },
    {
        1880,
        {27231, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1880},
        {27231}
    },
    {
        1881,
        {24112, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1881},
        {24112}
    },
    {
        1882,
        {27589, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1882},
        {27589}
    },
    {
        1883,
        {27671, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1883},
        {27671}
    },
    {
        1884,
        {27773, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1884},
        {27773}
    },
    {
        1885,
        {30079, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1885},
        {30079}
    },
    {
        1886,
        {31048, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1886},
        {31048}
    },
    {
        1887,
        {23395, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1887},
        {23395}
    },
    {
        1888,
        {31232, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1888},
        {31232}
    },
    {
        1889,
        {32000, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1889},
        {32000}
    },
    {
        1890,
        {24509, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1890},
        {24509}
    },
    {
        1891,
        {35215, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1891},
        {35215}
    },
    {
        1892,
        {35352, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1892},
        {35352}
    },
    {
        1893,
        {36020, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1893},
        {36020}
    },
    {
        1894,
        {36215, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1894},
        {36215}
    },
    {
        1895,
        {36556, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1895},
        {36556}
    },
    {
        1896,
        {36637, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1896},
        {36637}
    },
    {
        1897,
        {39138, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1897},
        {39138}
    },
    {
        1898,
        {39438, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1898},
        {39438}
    },
    {
        1899,
        {39740, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1899},
        {39740}
    },
    {
        1900,
        {20096, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1900},
        {20096}
    },
    {
        1901,
        {20605, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1901},
        {20605}
    },
    {
        1902,
        {20736, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1902},
        {20736}
    },
    {
        1903,
        {22931, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1903},
        {22931}
    },
    {
        1904,
        {23452, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1904},
        {23452}
    },
    {
        1905,
        {25135, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1905},
        {25135}
    },
    {
        1906,
        {25216, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1906},
        {25216}
    },
    {
        1907,
        {25836, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1907},
        {25836}
    },
    {
        1908,
        {27450, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1908},
        {27450}
    },
    {
        1909,
        {29344, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1909},
        {29344}
    },
    {
        1910,
        {30097, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1910},
        {30097}
    },
    {
        1911,
        {31047, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1911},
        {31047}
    },
    {
        1912,
        {32681, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1912},
        {32681}
    },
    {
        1913,
        {34811, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1913},
        {34811}
    },
    {
        1914,
        {35516, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1914},
        {35516}
    },
    {
        1915,
        {35696, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1915},
        {35696}
    },
    {
        1916,
        {25516, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1916},
        {25516}
    },
    {
        1917,
        {33738, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1917},
        {33738}
    },
    {
        1918,
        {38816, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1918},
        {38816}
    },
    {
        1919,
        {21513, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1919},
        {21513}
    },
    {
        1920,
        {21507, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1920},
        {21507}
    },
    {
        1921,
        {21931, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1921},
        {21931}
    },
    {
        1922,
        {26708, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1922},
        {26708}
    },
    {
        1923,
        {27224, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1923},
        {27224}
    },
    {
        1924,
        {35440, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1924},
        {35440}
    },
    {
        1925,
        {30759, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1925},
        {30759}
    },
    {
        1926,
        {26485, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1926},
        {26485}
    },
    {
        1927,
        {40653, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1927},
        {40653}
    },
    {
        1928,
        {21364, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1928},
        {21364}
    },
    {
        1929,
        {23458, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1929},
        {23458}
    },
    {
        1930,
        {33050, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1930},
        {33050}
    },
    {
        1931,
        {34384, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1931},
        {34384}
    },
    {
        1932,
        {36870, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1932},
        {36870}
    },
    {
        1933,
        {19992, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1933},
        {19992}
    },
    {
        1934,
        {20037, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1934},
        {20037}
    },
    {
        1935,
        {20167, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1935},
        {20167}
    },
    {
        1936,
        {20241, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1936},
        {20241}
    },
    {
        1937,
        {21450, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1937},
        {21450}
    },
    {
        1938,
        {21560, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1938},
        {21560}
    },
    {
        1939,
        {23470, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1939},
        {23470}
    },
    {
        1940,
        {24339, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1940},
        {24339}
    },
    {
        1941,
        {24613, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1941},
        {24613}
    },
    {
        1942,
        {25937, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1942},
        {25937}
    },
    {
        1943,
        {26429, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1943},
        {26429}
    },
    {
        1944,
        {27714, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1944},
        {27714}
    },
    {
        1945,
        {27762, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1945},
        {27762}
    },
    {
        1946,
        {27875, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1946},
        {27875}
    },
    {
        1947,
        {28792, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1947},
        {28792}
    },
    {
        1948,
        {29699, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1948},
        {29699}
    },
    {
        1949,
        {31350, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1949},
        {31350}
    },
    {
        1950,
        {31406, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1950},
        {31406}
    },
    {
        1951,
        {31496, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1951},
        {31496}
    },
    {
        1952,
        {32026, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1952},
        {32026}
    },
    {
        1953,
        {31998, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1953},
        {31998}
    },
    {
        1954,
        {32102, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1954},
        {32102}
    },
    {
        1955,
        {26087, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1955},
        {26087}
    },
    {
        1956,
        {29275, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1956},
        {29275}
    },
    {
        1957,
        {21435, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1957},
        {21435}
    },
    {
        1958,
        {23621, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1958},
        {23621}
    },
    {
        1959,
        {24040, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1959},
        {24040}
    },
    {
        1960,
        {25298, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1960},
        {25298}
    },
    {
        1961,
        {25312, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1961},
        {25312}
    },
    {
        1962,
        {25369, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1962},
        {25369}
    },
    {
        1963,
        {28192, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1963},
        {28192}
    },
    {
        1964,
        {34394, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1964},
        {34394}
    },
    {
        1965,
        {35377, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1965},
        {35377}
    },
    {
        1966,
        {36317, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1966},
        {36317}
    },
    {
        1967,
        {37624, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1967},
        {37624}
    },
    {
        1968,
        {28417, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1968},
        {28417}
    },
    {
        1969,
        {31142, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1969},
        {31142}
    },
    {
        1970,
        {39770, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1970},
        {39770}
    },
    {
        1971,
        {20136, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1971},
        {20136}
    },
    {
        1972,
        {20139, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1972},
        {20139}
    },
    {
        1973,
        {20140, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1973},
        {20140}
    },
    {
        1974,
        {20379, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1974},
        {20379}
    },
    {
        1975,
        {20384, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1975},
        {20384}
    },
    {
        1976,
        {20689, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1976},
        {20689}
    },
    {
        1977,
        {20807, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1977},
        {20807}
    },
    {
        1978,
        {31478, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1978},
        {31478}
    },
    {
        1979,
        {20849, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1979},
        {20849}
    },
    {
        1980,
        {20982, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1980},
        {20982}
    },
    {
        1981,
        {21332, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1981},
        {21332}
    },
    {
        1982,
        {21281, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1982},
        {21281}
    },
    {
        1983,
        {21375, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1983},
        {21375}
    },
    {
        1984,
        {21483, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1984},
        {21483}
    },
    {
        1985,
        {21932, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1985},
        {21932}
    },
    {
        1986,
        {22659, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1986},
        {22659}
    },
    {
        1987,
        {23777, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1987},
        {23777}
    },
    {
        1988,
        {24375, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1988},
        {24375}
    },
    {
        1989,
        {24394, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1989},
        {24394}
    },
    {
        1990,
        {24623, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1990},
        {24623}
    },
    {
        1991,
        {24656, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1991},
        {24656}
    },
    {
        1992,
        {24685, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1992},
        {24685}
    },
    {
        1993,
        {25375, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1993},
        {25375}
    },
    {
        1994,
        {25945, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1994},
        {25945}
    },
    {
        1995,
        {27211, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1995},
        {27211}
    },
    {
        1996,
        {27841, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1996},
        {27841}
    },
    {
        1997,
        {29378, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1997},
        {29378}
    },
    {
        1998,
        {29421, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1998},
        {29421}
    },
    {
        1999,
        {30703, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {1999},
        {30703}
    },
    {
        2000,
        {33016, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2000},
        {33016}
    },
    {
        2001,
        {33029, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2001},
        {33029}
    },
    {
        2002,
        {33288, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2002},
        {33288}
    },
    {
        2003,
        {34126, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2003},
        {34126}
    },
    {
        2004,
        {37111, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2004},
        {37111}
    },
    {
        2005,
        {37857, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2005},
        {37857}
    },
    {
        2006,
        {38911, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2006},
        {38911}
    },
    {
        2007,
        {39255, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2007},
        {39255}
    },
    {
        2008,
        {39514, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2008},
        {39514}
    },
    {
        2009,
        {20208, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2009},
        {20208}
    },
    {
        2010,
        {20957, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2010},
        {20957}
    },
    {
        2011,
        {23597, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2011},
        {23597}
    },
    {
        2012,
        {26241, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2012},
        {26241}
    },
    {
        2013,
        {26989, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2013},
        {26989}
    },
    {
        2014,
        {23616, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2014},
        {23616}
    },
    {
        2015,
        {26354, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2015},
        {26354}
    },
    {
        2016,
        {26997, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2016},
        {26997}
    },
    {
        2017,
        {29577, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2017},
        {29577}
    },
    {
        2018,
        {26704, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2018},
        {26704}
    },
    {
        2019,
        {31873, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2019},
        {31873}
    },
    {
        2020,
        {20677, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2020},
        {20677}
    },
    {
        2021,
        {21220, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2021},
        {21220}
    },
    {
        2022,
        {22343, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2022},
        {22343}
    },
    {
        2023,
        {24062, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2023},
        {24062}
    },
    {
        2024,
        {37670, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2024},
        {37670}
    },
    {
        2025,
        {26020, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2025},
        {26020}
    },
    {
        2026,
        {27427, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2026},
        {27427}
    },
    {
        2027,
        {27453, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2027},
        {27453}
    },
    {
        2028,
        {29748, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2028},
        {29748}
    },
    {
        2029,
        {31105, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2029},
        {31105}
    },
    {
        2030,
        {31165, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2030},
        {31165}
    },
    {
        2031,
        {31563, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2031},
        {31563}
    },
    {
        2032,
        {32202, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2032},
        {32202}
    },
    {
        2033,
        {33465, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2033},
        {33465}
    },
    {
        2034,
        {33740, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2034},
        {33740}
    },
    {
        2035,
        {34943, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2035},
        {34943}
    },
    {
        2036,
        {35167, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2036},
        {35167}
    },
    {
        2037,
        {35641, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2037},
        {35641}
    },
    {
        2038,
        {36817, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2038},
        {36817}
    },
    {
        2039,
        {37329, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2039},
        {37329}
    },
    {
        2040,
        {21535, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2040},
        {21535}
    },
    {
        2041,
        {37504, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2041},
        {37504}
    },
    {
        2042,
        {20061, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2042},
        {20061}
    },
    {
        2043,
        {20534, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2043},
        {20534}
    },
    {
        2044,
        {21477, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2044},
        {21477}
    },
    {
        2045,
        {21306, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2045},
        {21306}
    },
    {
        2046,
        {29399, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2046},
        {29399}
    },
    {
        2047,
        {29590, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2047},
        {29590}
    },
    {
        2048,
        {30697, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2048},
        {30697}
    },
    {
        2049,
        {33510, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2049},
        {33510}
    },
    {
        2050,
        {36527, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2050},
        {36527}
    },
    {
        2051,
        {39366, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2051},
        {39366}
    },
    {
        2052,
        {39368, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2052},
        {39368}
    },
    {
        2053,
        {39378, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2053},
        {39378}
    },
    {
        2054,
        {20855, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2054},
        {20855}
    },
    {
        2055,
        {24858, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2055},
        {24858}
    },
    {
        2056,
        {34398, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2056},
        {34398}
    },
    {
        2057,
        {21936, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2057},
        {21936}
    },
    {
        2058,
        {31354, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2058},
        {31354}
    },
    {
        2059,
        {20598, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2059},
        {20598}
    },
    {
        2060,
        {23507, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2060},
        {23507}
    },
    {
        2061,
        {36935, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2061},
        {36935}
    },
    {
        2062,
        {38533, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2062},
        {38533}
    },
    {
        2063,
        {20018, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2063},
        {20018}
    },
    {
        2064,
        {27355, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2064},
        {27355}
    },
    {
        2065,
        {37351, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2065},
        {37351}
    },
    {
        2066,
        {23633, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2066},
        {23633}
    },
    {
        2067,
        {23624, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2067},
        {23624}
    },
    {
        2068,
        {25496, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2068},
        {25496}
    },
    {
        2069,
        {31391, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2069},
        {31391}
    },
    {
        2070,
        {27795, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2070},
        {27795}
    },
    {
        2071,
        {38772, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2071},
        {38772}
    },
    {
        2072,
        {36705, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2072},
        {36705}
    },
    {
        2073,
        {31402, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2073},
        {31402}
    },
    {
        2074,
        {29066, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2074},
        {29066}
    },
    {
        2075,
        {38536, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2075},
        {38536}
    },
    {
        2076,
        {31874, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2076},
        {31874}
    },
    {
        2077,
        {26647, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2077},
        {26647}
    },
    {
        2078,
        {32368, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2078},
        {32368}
    },
    {
        2079,
        {26705, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2079},
        {26705}
    },
    {
        2080,
        {37740, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2080},
        {37740}
    },
    {
        2081,
        {21234, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2081},
        {21234}
    },
    {
        2082,
        {21531, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2082},
        {21531}
    },
    {
        2083,
        {34219, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2083},
        {34219}
    },
    {
        2084,
        {35347, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2084},
        {35347}
    },
    {
        2085,
        {32676, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2085},
        {32676}
    },
    {
        2086,
        {36557, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2086},
        {36557}
    },
    {
        2087,
        {37089, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2087},
        {37089}
    },
    {
        2088,
        {21350, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2088},
        {21350}
    },
    {
        2089,
        {34952, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2089},
        {34952}
    },
    {
        2090,
        {31041, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2090},
        {31041}
    },
    {
        2091,
        {20418, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2091},
        {20418}
    },
    {
        2092,
        {20670, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2092},
        {20670}
    },
    {
        2093,
        {21009, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2093},
        {21009}
    },
    {
        2094,
        {20804, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2094},
        {20804}
    },
    {
        2095,
        {21843, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2095},
        {21843}
    },
    {
        2096,
        {22317, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2096},
        {22317}
    },
    {
        2097,
        {29674, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2097},
        {29674}
    },
    {
        2098,
        {22411, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2098},
        {22411}
    },
    {
        2099,
        {22865, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2099},
        {22865}
    },
    {
        2100,
        {24418, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2100},
        {24418}
    },
    {
        2101,
        {24452, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2101},
        {24452}
    },
    {
        2102,
        {24693, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2102},
        {24693}
    },
    {
        2103,
        {24950, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2103},
        {24950}
    },
    {
        2104,
        {24935, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2104},
        {24935}
    },
    {
        2105,
        {25001, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2105},
        {25001}
    },
    {
        2106,
        {25522, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2106},
        {25522}
    },
    {
        2107,
        {25658, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2107},
        {25658}
    },
    {
        2108,
        {25964, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2108},
        {25964}
    },
    {
        2109,
        {26223, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2109},
        {26223}
    },
    {
        2110,
        {26690, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2110},
        {26690}
    },
    {
        2111,
        {28179, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2111},
        {28179}
    },
    {
        2112,
        {30054, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2112},
        {30054}
    },
    {
        2113,
        {31293, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2113},
        {31293}
    },
    {
        2114,
        {31995, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2114},
        {31995}
    },
    {
        2115,
        {32076, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2115},
        {32076}
    },
    {
        2116,
        {32153, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2116},
        {32153}
    },
    {
        2117,
        {32331, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2117},
        {32331}
    },
    {
        2118,
        {32619, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2118},
        {32619}
    },
    {
        2119,
        {33550, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2119},
        {33550}
    },
    {
        2120,
        {33610, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2120},
        {33610}
    },
    {
        2121,
        {34509, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2121},
        {34509}
    },
    {
        2122,
        {35336, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2122},
        {35336}
    },
    {
        2123,
        {35427, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2123},
        {35427}
    },
    {
        2124,
        {35686, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2124},
        {35686}
    },
    {
        2125,
        {36605, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2125},
        {36605}
    },
    {
        2126,
        {38938, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2126},
        {38938}
    },
    {
        2127,
        {40335, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2127},
        {40335}
    },
    {
        2128,
        {33464, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2128},
        {33464}
    },
    {
        2129,
        {36814, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2129},
        {36814}
    },
    {
        2130,
        {39912, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2130},
        {39912}
    },
    {
        2131,
        {21127, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2131},
        {21127}
    },
    {
        2132,
        {25119, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2132},
        {25119}
    },
    {
        2133,
        {25731, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2133},
        {25731}
    },
    {
        2134,
        {28608, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2134},
        {28608}
    },
    {
        2135,
        {38553, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2135},
        {38553}
    },
    {
        2136,
        {26689, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2136},
        {26689}
    },
    {
        2137,
        {20625, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2137},
        {20625}
    },
    {
        2138,
        {27424, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2138},
        {27424}
    },
    {
        2139,
        {27770, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2139},
        {27770}
    },
    {
        2140,
        {28500, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2140},
        {28500}
    },
    {
        2141,
        {31348, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2141},
        {31348}
    },
    {
        2142,
        {32080, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2142},
        {32080}
    },
    {
        2143,
        {34880, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2143},
        {34880}
    },
    {
        2144,
        {35363, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2144},
        {35363}
    },
    {
        2145,
        {26376, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2145},
        {26376}
    },
    {
        2146,
        {20214, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2146},
        {20214}
    },
    {
        2147,
        {20537, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2147},
        {20537}
    },
    {
        2148,
        {20518, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2148},
        {20518}
    },
    {
        2149,
        {20581, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2149},
        {20581}
    },
    {
        2150,
        {20860, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2150},
        {20860}
    },
    {
        2151,
        {21048, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2151},
        {21048}
    },
    {
        2152,
        {21091, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2152},
        {21091}
    },
    {
        2153,
        {21927, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2153},
        {21927}
    },
    {
        2154,
        {22287, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2154},
        {22287}
    },
    {
        2155,
        {22533, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2155},
        {22533}
    },
    {
        2156,
        {23244, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2156},
        {23244}
    },
    {
        2157,
        {24314, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2157},
        {24314}
    },
    {
        2158,
        {25010, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2158},
        {25010}
    },
    {
        2159,
        {25080, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2159},
        {25080}
    },
    {
        2160,
        {25331, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2160},
        {25331}
    },
    {
        2161,
        {25458, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2161},
        {25458}
    },
    {
        2162,
        {26908, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2162},
        {26908}
    },
    {
        2163,
        {27177, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2163},
        {27177}
    },
    {
        2164,
        {29309, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2164},
        {29309}
    },
    {
        2165,
        {29356, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2165},
        {29356}
    },
    {
        2166,
        {29486, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2166},
        {29486}
    },
    {
        2167,
        {30740, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2167},
        {30740}
    },
    {
        2168,
        {30831, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2168},
        {30831}
    },
    {
        2169,
        {32121, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2169},
        {32121}
    },
    {
        2170,
        {30476, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2170},
        {30476}
    },
    {
        2171,
        {32937, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2171},
        {32937}
    },
    {
        2172,
        {35211, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2172},
        {35211}
    },
    {
        2173,
        {35609, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2173},
        {35609}
    },
    {
        2174,
        {36066, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2174},
        {36066}
    },
    {
        2175,
        {36562, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2175},
        {36562}
    },
    {
        2176,
        {36963, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2176},
        {36963}
    },
    {
        2177,
        {37749, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2177},
        {37749}
    },
    {
        2178,
        {38522, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2178},
        {38522}
    },
    {
        2179,
        {38997, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2179},
        {38997}
    },
    {
        2180,
        {39443, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2180},
        {39443}
    },
    {
        2181,
        {40568, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2181},
        {40568}
    },
    {
        2182,
        {20803, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2182},
        {20803}
    },
    {
        2183,
        {21407, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2183},
        {21407}
    },
    {
        2184,
        {21427, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2184},
        {21427}
    },
    {
        2185,
        {24187, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2185},
        {24187}
    },
    {
        2186,
        {24358, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2186},
        {24358}
    },
    {
        2187,
        {28187, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2187},
        {28187}
    },
    {
        2188,
        {28304, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2188},
        {28304}
    },
    {
        2189,
        {29572, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2189},
        {29572}
    },
    {
        2190,
        {29694, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2190},
        {29694}
    },
    {
        2191,
        {32067, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2191},
        {32067}
    },
    {
        2192,
        {33335, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2192},
        {33335}
    },
    {
        2193,
        {35328, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2193},
        {35328}
    },
    {
        2194,
        {35578, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2194},
        {35578}
    },
    {
        2195,
        {38480, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2195},
        {38480}
    },
    {
        2196,
        {20046, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2196},
        {20046}
    },
    {
        2197,
        {20491, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2197},
        {20491}
    },
    {
        2198,
        {21476, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2198},
        {21476}
    },
    {
        2199,
        {21628, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2199},
        {21628}
    },
    {
        2200,
        {22266, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2200},
        {22266}
    },
    {
        2201,
        {22993, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2201},
        {22993}
    },
    {
        2202,
        {23396, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2202},
        {23396}
    },
    {
        2203,
        {24049, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2203},
        {24049}
    },
    {
        2204,
        {24235, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2204},
        {24235}
    },
    {
        2205,
        {24359, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2205},
        {24359}
    },
    {
        2206,
        {25144, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2206},
        {25144}
    },
    {
        2207,
        {25925, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2207},
        {25925}
    },
    {
        2208,
        {26543, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2208},
        {26543}
    },
    {
        2209,
        {28246, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2209},
        {28246}
    },
    {
        2210,
        {29392, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2210},
        {29392}
    },
    {
        2211,
        {31946, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2211},
        {31946}
    },
    {
        2212,
        {34996, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2212},
        {34996}
    },
    {
        2213,
        {32929, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2213},
        {32929}
    },
    {
        2214,
        {32993, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2214},
        {32993}
    },
    {
        2215,
        {33776, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2215},
        {33776}
    },
    {
        2216,
        {34382, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2216},
        {34382}
    },
    {
        2217,
        {35463, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2217},
        {35463}
    },
    {
        2218,
        {36328, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2218},
        {36328}
    },
    {
        2219,
        {37431, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2219},
        {37431}
    },
    {
        2220,
        {38599, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2220},
        {38599}
    },
    {
        2221,
        {39015, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2221},
        {39015}
    },
    {
        2222,
        {40723, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2222},
        {40723}
    },
    {
        2223,
        {20116, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2223},
        {20116}
    },
    {
        2224,
        {20114, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2224},
        {20114}
    },
    {
        2225,
        {20237, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2225},
        {20237}
    },
    {
        2226,
        {21320, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2226},
        {21320}
    },
    {
        2227,
        {21577, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2227},
        {21577}
    },
    {
        2228,
        {21566, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2228},
        {21566}
    },
    {
        2229,
        {23087, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2229},
        {23087}
    },
    {
        2230,
        {24460, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2230},
        {24460}
    },
    {
        2231,
        {24481, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2231},
        {24481}
    },
    {
        2232,
        {24735, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2232},
        {24735}
    },
    {
        2233,
        {26791, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2233},
        {26791}
    },
    {
        2234,
        {27278, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2234},
        {27278}
    },
    {
        2235,
        {29786, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2235},
        {29786}
    },
    {
        2236,
        {30849, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2236},
        {30849}
    },
    {
        2237,
        {35486, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2237},
        {35486}
    },
    {
        2238,
        {35492, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2238},
        {35492}
    },
    {
        2239,
        {35703, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2239},
        {35703}
    },
    {
        2240,
        {37264, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2240},
        {37264}
    },
    {
        2241,
        {20062, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2241},
        {20062}
    },
    {
        2242,
        {39881, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2242},
        {39881}
    },
    {
        2243,
        {20132, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2243},
        {20132}
    },
    {
        2244,
        {20348, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2244},
        {20348}
    },
    {
        2245,
        {20399, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2245},
        {20399}
    },
    {
        2246,
        {20505, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2246},
        {20505}
    },
    {
        2247,
        {20502, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2247},
        {20502}
    },
    {
        2248,
        {20809, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2248},
        {20809}
    },
    {
        2249,
        {20844, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2249},
        {20844}
    },
    {
        2250,
        {21151, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2250},
        {21151}
    },
    {
        2251,
        {21177, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2251},
        {21177}
    },
    {
        2252,
        {21246, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2252},
        {21246}
    },
    {
        2253,
        {21402, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2253},
        {21402}
    },
    {
        2254,
        {21475, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2254},
        {21475}
    },
    {
        2255,
        {21521, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2255},
        {21521}
    },
    {
        2256,
        {21518, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2256},
        {21518}
    },
    {
        2257,
        {21897, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2257},
        {21897}
    },
    {
        2258,
        {22353, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2258},
        {22353}
    },
    {
        2259,
        {22434, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2259},
        {22434}
    },
    {
        2260,
        {22909, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2260},
        {22909}
    },
    {
        2261,
        {23380, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2261},
        {23380}
    },
    {
        2262,
        {23389, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2262},
        {23389}
    },
    {
        2263,
        {23439, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2263},
        {23439}
    },
    {
        2264,
        {24037, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2264},
        {24037}
    },
    {
        2265,
        {24039, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2265},
        {24039}
    },
    {
        2266,
        {24055, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2266},
        {24055}
    },
    {
        2267,
        {24184, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2267},
        {24184}
    },
    {
        2268,
        {24195, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2268},
        {24195}
    },
    {
        2269,
        {24218, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2269},
        {24218}
    },
    {
        2270,
        {24247, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2270},
        {24247}
    },
    {
        2271,
        {24344, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2271},
        {24344}
    },
    {
        2272,
        {24658, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2272},
        {24658}
    },
    {
        2273,
        {24908, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2273},
        {24908}
    },
    {
        2274,
        {25239, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2274},
        {25239}
    },
    {
        2275,
        {25304, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2275},
        {25304}
    },
    {
        2276,
        {25511, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2276},
        {25511}
    },
    {
        2277,
        {25915, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2277},
        {25915}
    },
    {
        2278,
        {26114, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2278},
        {26114}
    },
    {
        2279,
        {26179, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2279},
        {26179}
    },
    {
        2280,
        {26356, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2280},
        {26356}
    },
    {
        2281,
        {26477, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2281},
        {26477}
    },
    {
        2282,
        {26657, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2282},
        {26657}
    },
    {
        2283,
        {26775, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2283},
        {26775}
    },
    {
        2284,
        {27083, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2284},
        {27083}
    },
    {
        2285,
        {27743, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2285},
        {27743}
    },
    {
        2286,
        {27946, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2286},
        {27946}
    },
    {
        2287,
        {28009, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2287},
        {28009}
    },
    {
        2288,
        {28207, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2288},
        {28207}
    },
    {
        2289,
        {28317, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2289},
        {28317}
    },
    {
        2290,
        {30002, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2290},
        {30002}
    },
    {
        2291,
        {30343, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2291},
        {30343}
    },
    {
        2292,
        {30828, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2292},
        {30828}
    },
    {
        2293,
        {31295, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2293},
        {31295}
    },
    {
        2294,
        {31968, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2294},
        {31968}
    },
    {
        2295,
        {32005, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2295},
        {32005}
    },
    {
        2296,
        {32024, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2296},
        {32024}
    },
    {
        2297,
        {32094, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2297},
        {32094}
    },
    {
        2298,
        {32177, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2298},
        {32177}
    },
    {
        2299,
        {32789, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2299},
        {32789}
    },
    {
        2300,
        {32771, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2300},
        {32771}
    },
    {
        2301,
        {32943, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2301},
        {32943}
    },
    {
        2302,
        {32945, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2302},
        {32945}
    },
    {
        2303,
        {33108, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2303},
        {33108}
    },
    {
        2304,
        {33167, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2304},
        {33167}
    },
    {
        2305,
        {33322, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2305},
        {33322}
    },
    {
        2306,
        {33618, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2306},
        {33618}
    },
    {
        2307,
        {34892, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2307},
        {34892}
    },
    {
        2308,
        {34913, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2308},
        {34913}
    },
    {
        2309,
        {35611, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2309},
        {35611}
    },
    {
        2310,
        {36002, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2310},
        {36002}
    },
    {
        2311,
        {36092, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2311},
        {36092}
    },
    {
        2312,
        {37066, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2312},
        {37066}
    },
    {
        2313,
        {37237, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2313},
        {37237}
    },
    {
        2314,
        {37489, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2314},
        {37489}
    },
    {
        2315,
        {30783, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2315},
        {30783}
    },
    {
        2316,
        {37628, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2316},
        {37628}
    },
    {
        2317,
        {38308, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2317},
        {38308}
    },
    {
        2318,
        {38477, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2318},
        {38477}
    },
    {
        2319,
        {38917, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2319},
        {38917}
    },
    {
        2320,
        {39321, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2320},
        {39321}
    },
    {
        2321,
        {39640, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2321},
        {39640}
    },
    {
        2322,
        {40251, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2322},
        {40251}
    },
    {
        2323,
        {21083, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2323},
        {21083}
    },
    {
        2324,
        {21163, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2324},
        {21163}
    },
    {
        2325,
        {21495, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2325},
        {21495}
    },
    {
        2326,
        {21512, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2326},
        {21512}
    },
    {
        2327,
        {22741, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2327},
        {22741}
    },
    {
        2328,
        {25335, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2328},
        {25335}
    },
    {
        2329,
        {28640, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2329},
        {28640}
    },
    {
        2330,
        {35946, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2330},
        {35946}
    },
    {
        2331,
        {36703, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2331},
        {36703}
    },
    {
        2332,
        {40633, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2332},
        {40633}
    },
    {
        2333,
        {20811, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2333},
        {20811}
    },
    {
        2334,
        {21051, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2334},
        {21051}
    },
    {
        2335,
        {21578, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2335},
        {21578}
    },
    {
        2336,
        {22269, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2336},
        {22269}
    },
    {
        2337,
        {31296, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2337},
        {31296}
    },
    {
        2338,
        {37239, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2338},
        {37239}
    },
    {
        2339,
        {40288, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2339},
        {40288}
    },
    {
        2340,
        {40658, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2340},
        {40658}
    },
    {
        2341,
        {29508, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2341},
        {29508}
    },
    {
        2342,
        {28425, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2342},
        {28425}
    },
    {
        2343,
        {33136, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2343},
        {33136}
    },
    {
        2344,
        {29969, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2344},
        {29969}
    },
    {
        2345,
        {24573, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2345},
        {24573}
    },
    {
        2346,
        {24794, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2346},
        {24794}
    },
    {
        2347,
        {39592, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2347},
        {39592}
    },
    {
        2348,
        {29403, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2348},
        {29403}
    },
    {
        2349,
        {36796, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2349},
        {36796}
    },
    {
        2350,
        {27492, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2350},
        {27492}
    },
    {
        2351,
        {38915, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2351},
        {38915}
    },
    {
        2352,
        {20170, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2352},
        {20170}
    },
    {
        2353,
        {22256, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2353},
        {22256}
    },
    {
        2354,
        {22372, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2354},
        {22372}
    },
    {
        2355,
        {22718, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2355},
        {22718}
    },
    {
        2356,
        {23130, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2356},
        {23130}
    },
    {
        2357,
        {24680, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2357},
        {24680}
    },
    {
        2358,
        {25031, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2358},
        {25031}
    },
    {
        2359,
        {26127, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2359},
        {26127}
    },
    {
        2360,
        {26118, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2360},
        {26118}
    },
    {
        2361,
        {26681, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2361},
        {26681}
    },
    {
        2362,
        {26801, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2362},
        {26801}
    },
    {
        2363,
        {28151, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2363},
        {28151}
    },
    {
        2364,
        {30165, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2364},
        {30165}
    },
    {
        2365,
        {32058, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2365},
        {32058}
    },
    {
        2366,
        {33390, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2366},
        {33390}
    },
    {
        2367,
        {39746, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2367},
        {39746}
    },
    {
        2368,
        {20123, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2368},
        {20123}
    },
    {
        2369,
        {20304, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2369},
        {20304}
    },
    {
        2370,
        {21449, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2370},
        {21449}
    },
    {
        2371,
        {21766, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2371},
        {21766}
    },
    {
        2372,
        {23919, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2372},
        {23919}
    },
    {
        2373,
        {24038, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2373},
        {24038}
    },
    {
        2374,
        {24046, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2374},
        {24046}
    },
    {
        2375,
        {26619, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2375},
        {26619}
    },
    {
        2376,
        {27801, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2376},
        {27801}
    },
    {
        2377,
        {29811, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2377},
        {29811}
    },
    {
        2378,
        {30722, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2378},
        {30722}
    },
    {
        2379,
        {35408, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2379},
        {35408}
    },
    {
        2380,
        {37782, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2380},
        {37782}
    },
    {
        2381,
        {35039, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2381},
        {35039}
    },
    {
        2382,
        {22352, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2382},
        {22352}
    },
    {
        2383,
        {24231, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2383},
        {24231}
    },
    {
        2384,
        {25387, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2384},
        {25387}
    },
    {
        2385,
        {20661, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2385},
        {20661}
    },
    {
        2386,
        {20652, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2386},
        {20652}
    },
    {
        2387,
        {20877, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2387},
        {20877}
    },
    {
        2388,
        {26368, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2388},
        {26368}
    },
    {
        2389,
        {21705, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2389},
        {21705}
    },
    {
        2390,
        {22622, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2390},
        {22622}
    },
    {
        2391,
        {22971, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2391},
        {22971}
    },
    {
        2392,
        {23472, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2392},
        {23472}
    },
    {
        2393,
        {24425, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2393},
        {24425}
    },
    {
        2394,
        {25165, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2394},
        {25165}
    },
    {
        2395,
        {25505, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2395},
        {25505}
    },
    {
        2396,
        {26685, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2396},
        {26685}
    },
    {
        2397,
        {27507, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2397},
        {27507}
    },
    {
        2398,
        {28168, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2398},
        {28168}
    },
    {
        2399,
        {28797, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2399},
        {28797}
    },
    {
        2400,
        {37319, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2400},
        {37319}
    },
    {
        2401,
        {29312, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2401},
        {29312}
    },
    {
        2402,
        {30741, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2402},
        {30741}
    },
    {
        2403,
        {30758, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2403},
        {30758}
    },
    {
        2404,
        {31085, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2404},
        {31085}
    },
    {
        2405,
        {25998, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2405},
        {25998}
    },
    {
        2406,
        {32048, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2406},
        {32048}
    },
    {
        2407,
        {33756, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2407},
        {33756}
    },
    {
        2408,
        {35009, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2408},
        {35009}
    },
    {
        2409,
        {36617, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2409},
        {36617}
    },
    {
        2410,
        {38555, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2410},
        {38555}
    },
    {
        2411,
        {21092, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2411},
        {21092}
    },
    {
        2412,
        {22312, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2412},
        {22312}
    },
    {
        2413,
        {26448, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2413},
        {26448}
    },
    {
        2414,
        {32618, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2414},
        {32618}
    },
    {
        2415,
        {36001, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2415},
        {36001}
    },
    {
        2416,
        {20916, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2416},
        {20916}
    },
    {
        2417,
        {22338, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2417},
        {22338}
    },
    {
        2418,
        {38442, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2418},
        {38442}
    },
    {
        2419,
        {22586, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2419},
        {22586}
    },
    {
        2420,
        {27018, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2420},
        {27018}
    },
    {
        2421,
        {32948, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2421},
        {32948}
    },
    {
        2422,
        {21682, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2422},
        {21682}
    },
    {
        2423,
        {23822, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2423},
        {23822}
    },
    {
        2424,
        {22524, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2424},
        {22524}
    },
    {
        2425,
        {30869, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2425},
        {30869}
    },
    {
        2426,
        {40442, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2426},
        {40442}
    },
    {
        2427,
        {20316, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2427},
        {20316}
    },
    {
        2428,
        {21066, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2428},
        {21066}
    },
    {
        2429,
        {21643, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2429},
        {21643}
    },
    {
        2430,
        {25662, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2430},
        {25662}
    },
    {
        2431,
        {26152, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2431},
        {26152}
    },
    {
        2432,
        {26388, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2432},
        {26388}
    },
    {
        2433,
        {26613, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2433},
        {26613}
    },
    {
        2434,
        {31364, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2434},
        {31364}
    },
    {
        2435,
        {31574, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2435},
        {31574}
    },
    {
        2436,
        {32034, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2436},
        {32034}
    },
    {
        2437,
        {37679, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2437},
        {37679}
    },
    {
        2438,
        {26716, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2438},
        {26716}
    },
    {
        2439,
        {39853, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2439},
        {39853}
    },
    {
        2440,
        {31545, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2440},
        {31545}
    },
    {
        2441,
        {21273, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2441},
        {21273}
    },
    {
        2442,
        {20874, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2442},
        {20874}
    },
    {
        2443,
        {21047, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2443},
        {21047}
    },
    {
        2444,
        {23519, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2444},
        {23519}
    },
    {
        2445,
        {25334, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2445},
        {25334}
    },
    {
        2446,
        {25774, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2446},
        {25774}
    },
    {
        2447,
        {25830, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2447},
        {25830}
    },
    {
        2448,
        {26413, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2448},
        {26413}
    },
    {
        2449,
        {27578, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2449},
        {27578}
    },
    {
        2450,
        {34217, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2450},
        {34217}
    },
    {
        2451,
        {38609, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2451},
        {38609}
    },
    {
        2452,
        {30352, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2452},
        {30352}
    },
    {
        2453,
        {39894, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2453},
        {39894}
    },
    {
        2454,
        {25420, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2454},
        {25420}
    },
    {
        2455,
        {37638, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2455},
        {37638}
    },
    {
        2456,
        {39851, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2456},
        {39851}
    },
    {
        2457,
        {30399, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2457},
        {30399}
    },
    {
        2458,
        {26194, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2458},
        {26194}
    },
    {
        2459,
        {19977, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2459},
        {19977}
    },
    {
        2460,
        {20632, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2460},
        {20632}
    },
    {
        2461,
        {21442, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2461},
        {21442}
    },
    {
        2462,
        {23665, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2462},
        {23665}
    },
    {
        2463,
        {24808, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2463},
        {24808}
    },
    {
        2464,
        {25746, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2464},
        {25746}
    },
    {
        2465,
        {25955, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2465},
        {25955}
    },
    {
        2466,
        {26719, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2466},
        {26719}
    },
    {
        2467,
        {29158, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2467},
        {29158}
    },
    {
        2468,
        {29642, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2468},
        {29642}
    },
    {
        2469,
        {29987, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2469},
        {29987}
    },
    {
        2470,
        {31639, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2470},
        {31639}
    },
    {
        2471,
        {32386, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2471},
        {32386}
    },
    {
        2472,
        {34453, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2472},
        {34453}
    },
    {
        2473,
        {35715, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2473},
        {35715}
    },
    {
        2474,
        {36059, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2474},
        {36059}
    },
    {
        2475,
        {37240, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2475},
        {37240}
    },
    {
        2476,
        {39184, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2476},
        {39184}
    },
    {
        2477,
        {26028, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2477},
        {26028}
    },
    {
        2478,
        {26283, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2478},
        {26283}
    },
    {
        2479,
        {27531, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2479},
        {27531}
    },
    {
        2480,
        {20181, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2480},
        {20181}
    },
    {
        2481,
        {20180, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2481},
        {20180}
    },
    {
        2482,
        {20282, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2482},
        {20282}
    },
    {
        2483,
        {20351, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2483},
        {20351}
    },
    {
        2484,
        {21050, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2484},
        {21050}
    },
    {
        2485,
        {21496, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2485},
        {21496}
    },
    {
        2486,
        {21490, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2486},
        {21490}
    },
    {
        2487,
        {21987, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2487},
        {21987}
    },
    {
        2488,
        {22235, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2488},
        {22235}
    },
    {
        2489,
        {22763, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2489},
        {22763}
    },
    {
        2490,
        {22987, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2490},
        {22987}
    },
    {
        2491,
        {22985, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2491},
        {22985}
    },
    {
        2492,
        {23039, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2492},
        {23039}
    },
    {
        2493,
        {23376, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2493},
        {23376}
    },
    {
        2494,
        {23629, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2494},
        {23629}
    },
    {
        2495,
        {24066, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2495},
        {24066}
    },
    {
        2496,
        {24107, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2496},
        {24107}
    },
    {
        2497,
        {24535, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2497},
        {24535}
    },
    {
        2498,
        {24605, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2498},
        {24605}
    },
    {
        2499,
        {25351, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2499},
        {25351}
    },
    {
        2500,
        {25903, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2500},
        {25903}
    },
    {
        2501,
        {23388, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2501},
        {23388}
    },
    {
        2502,
        {26031, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2502},
        {26031}
    },
    {
        2503,
        {26045, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2503},
        {26045}
    },
    {
        2504,
        {26088, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2504},
        {26088}
    },
    {
        2505,
        {26525, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2505},
        {26525}
    },
    {
        2506,
        {27490, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2506},
        {27490}
    },
    {
        2507,
        {27515, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2507},
        {27515}
    },
    {
        2508,
        {27663, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2508},
        {27663}
    },
    {
        2509,
        {29509, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2509},
        {29509}
    },
    {
        2510,
        {31049, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2510},
        {31049}
    },
    {
        2511,
        {31169, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2511},
        {31169}
    },
    {
        2512,
        {31992, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2512},
        {31992}
    },
    {
        2513,
        {32025, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2513},
        {32025}
    },
    {
        2514,
        {32043, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2514},
        {32043}
    },
    {
        2515,
        {32930, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2515},
        {32930}
    },
    {
        2516,
        {33026, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2516},
        {33026}
    },
    {
        2517,
        {33267, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2517},
        {33267}
    },
    {
        2518,
        {35222, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2518},
        {35222}
    },
    {
        2519,
        {35422, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2519},
        {35422}
    },
    {
        2520,
        {35433, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2520},
        {35433}
    },
    {
        2521,
        {35430, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2521},
        {35430}
    },
    {
        2522,
        {35468, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2522},
        {35468}
    },
    {
        2523,
        {35566, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2523},
        {35566}
    },
    {
        2524,
        {36039, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2524},
        {36039}
    },
    {
        2525,
        {36060, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2525},
        {36060}
    },
    {
        2526,
        {38604, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2526},
        {38604}
    },
    {
        2527,
        {39164, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2527},
        {39164}
    },
    {
        2528,
        {27503, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2528},
        {27503}
    },
    {
        2529,
        {20107, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2529},
        {20107}
    },
    {
        2530,
        {20284, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2530},
        {20284}
    },
    {
        2531,
        {20365, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2531},
        {20365}
    },
    {
        2532,
        {20816, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2532},
        {20816}
    },
    {
        2533,
        {23383, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2533},
        {23383}
    },
    {
        2534,
        {23546, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2534},
        {23546}
    },
    {
        2535,
        {24904, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2535},
        {24904}
    },
    {
        2536,
        {25345, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2536},
        {25345}
    },
    {
        2537,
        {26178, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2537},
        {26178}
    },
    {
        2538,
        {27425, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2538},
        {27425}
    },
    {
        2539,
        {28363, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2539},
        {28363}
    },
    {
        2540,
        {27835, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2540},
        {27835}
    },
    {
        2541,
        {29246, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2541},
        {29246}
    },
    {
        2542,
        {29885, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2542},
        {29885}
    },
    {
        2543,
        {30164, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2543},
        {30164}
    },
    {
        2544,
        {30913, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2544},
        {30913}
    },
    {
        2545,
        {31034, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2545},
        {31034}
    },
    {
        2546,
        {32780, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2546},
        {32780}
    },
    {
        2547,
        {32819, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2547},
        {32819}
    },
    {
        2548,
        {33258, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2548},
        {33258}
    },
    {
        2549,
        {33940, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2549},
        {33940}
    },
    {
        2550,
        {36766, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2550},
        {36766}
    },
    {
        2551,
        {27728, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2551},
        {27728}
    },
    {
        2552,
        {40575, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2552},
        {40575}
    },
    {
        2553,
        {24335, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2553},
        {24335}
    },
    {
        2554,
        {35672, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2554},
        {35672}
    },
    {
        2555,
        {40235, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2555},
        {40235}
    },
    {
        2556,
        {31482, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2556},
        {31482}
    },
    {
        2557,
        {36600, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2557},
        {36600}
    },
    {
        2558,
        {23437, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2558},
        {23437}
    },
    {
        2559,
        {38635, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2559},
        {38635}
    },
    {
        2560,
        {19971, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2560},
        {19971}
    },
    {
        2561,
        {21489, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2561},
        {21489}
    },
    {
        2562,
        {22519, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2562},
        {22519}
    },
    {
        2563,
        {22833, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2563},
        {22833}
    },
    {
        2564,
        {23241, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2564},
        {23241}
    },
    {
        2565,
        {23460, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2565},
        {23460}
    },
    {
        2566,
        {24713, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2566},
        {24713}
    },
    {
        2567,
        {28287, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2567},
        {28287}
    },
    {
        2568,
        {28422, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2568},
        {28422}
    },
    {
        2569,
        {30142, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2569},
        {30142}
    },
    {
        2570,
        {36074, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2570},
        {36074}
    },
    {
        2571,
        {23455, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2571},
        {23455}
    },
    {
        2572,
        {34048, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2572},
        {34048}
    },
    {
        2573,
        {31712, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2573},
        {31712}
    },
    {
        2574,
        {20594, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2574},
        {20594}
    },
    {
        2575,
        {26612, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2575},
        {26612}
    },
    {
        2576,
        {33437, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2576},
        {33437}
    },
    {
        2577,
        {23649, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2577},
        {23649}
    },
    {
        2578,
        {34122, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2578},
        {34122}
    },
    {
        2579,
        {32286, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2579},
        {32286}
    },
    {
        2580,
        {33294, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2580},
        {33294}
    },
    {
        2581,
        {20889, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2581},
        {20889}
    },
    {
        2582,
        {23556, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2582},
        {23556}
    },
    {
        2583,
        {25448, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2583},
        {25448}
    },
    {
        2584,
        {36198, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2584},
        {36198}
    },
    {
        2585,
        {26012, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2585},
        {26012}
    },
    {
        2586,
        {29038, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2586},
        {29038}
    },
    {
        2587,
        {31038, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2587},
        {31038}
    },
    {
        2588,
        {32023, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2588},
        {32023}
    },
    {
        2589,
        {32773, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2589},
        {32773}
    },
    {
        2590,
        {35613, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2590},
        {35613}
    },
    {
        2591,
        {36554, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2591},
        {36554}
    },
    {
        2592,
        {36974, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2592},
        {36974}
    },
    {
        2593,
        {34503, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2593},
        {34503}
    },
    {
        2594,
        {37034, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2594},
        {37034}
    },
    {
        2595,
        {20511, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2595},
        {20511}
    },
    {
        2596,
        {21242, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2596},
        {21242}
    },
    {
        2597,
        {23610, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2597},
        {23610}
    },
    {
        2598,
        {26451, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2598},
        {26451}
    },
    {
        2599,
        {28796, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2599},
        {28796}
    },
    {
        2600,
        {29237, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2600},
        {29237}
    },
    {
        2601,
        {37196, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2601},
        {37196}
    },
    {
        2602,
        {37320, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2602},
        {37320}
    },
    {
        2603,
        {37675, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2603},
        {37675}
    },
    {
        2604,
        {33509, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2604},
        {33509}
    },
    {
        2605,
        {23490, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2605},
        {23490}
    },
    {
        2606,
        {24369, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2606},
        {24369}
    },
    {
        2607,
        {24825, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2607},
        {24825}
    },
    {
        2608,
        {20027, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2608},
        {20027}
    },
    {
        2609,
        {21462, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2609},
        {21462}
    },
    {
        2610,
        {23432, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2610},
        {23432}
    },
    {
        2611,
        {25163, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2611},
        {25163}
    },
    {
        2612,
        {26417, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2612},
        {26417}
    },
    {
        2613,
        {27530, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2613},
        {27530}
    },
    {
        2614,
        {29417, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2614},
        {29417}
    },
    {
        2615,
        {29664, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2615},
        {29664}
    },
    {
        2616,
        {31278, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2616},
        {31278}
    },
    {
        2617,
        {33131, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2617},
        {33131}
    },
    {
        2618,
        {36259, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2618},
        {36259}
    },
    {
        2619,
        {37202, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2619},
        {37202}
    },
    {
        2620,
        {39318, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2620},
        {39318}
    },
    {
        2621,
        {20754, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2621},
        {20754}
    },
    {
        2622,
        {21463, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2622},
        {21463}
    },
    {
        2623,
        {21610, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2623},
        {21610}
    },
    {
        2624,
        {23551, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2624},
        {23551}
    },
    {
        2625,
        {25480, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2625},
        {25480}
    },
    {
        2626,
        {27193, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2626},
        {27193}
    },
    {
        2627,
        {32172, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2627},
        {32172}
    },
    {
        2628,
        {38656, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2628},
        {38656}
    },
    {
        2629,
        {22234, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2629},
        {22234}
    },
    {
        2630,
        {21454, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2630},
        {21454}
    },
    {
        2631,
        {21608, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2631},
        {21608}
    },
    {
        2632,
        {23447, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2632},
        {23447}
    },
    {
        2633,
        {23601, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2633},
        {23601}
    },
    {
        2634,
        {24030, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2634},
        {24030}
    },
    {
        2635,
        {20462, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2635},
        {20462}
    },
    {
        2636,
        {24833, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2636},
        {24833}
    },
    {
        2637,
        {25342, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2637},
        {25342}
    },
    {
        2638,
        {27954, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2638},
        {27954}
    },
    {
        2639,
        {31168, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2639},
        {31168}
    },
    {
        2640,
        {31179, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2640},
        {31179}
    },
    {
        2641,
        {32066, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2641},
        {32066}
    },
    {
        2642,
        {32333, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2642},
        {32333}
    },
    {
        2643,
        {32722, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2643},
        {32722}
    },
    {
        2644,
        {33261, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2644},
        {33261}
    },
    {
        2645,
        {33311, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2645},
        {33311}
    },
    {
        2646,
        {33936, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2646},
        {33936}
    },
    {
        2647,
        {34886, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2647},
        {34886}
    },
    {
        2648,
        {35186, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2648},
        {35186}
    },
    {
        2649,
        {35728, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2649},
        {35728}
    },
    {
        2650,
        {36468, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2650},
        {36468}
    },
    {
        2651,
        {36655, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2651},
        {36655}
    },
    {
        2652,
        {36913, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2652},
        {36913}
    },
    {
        2653,
        {37195, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2653},
        {37195}
    },
    {
        2654,
        {37228, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2654},
        {37228}
    },
    {
        2655,
        {38598, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2655},
        {38598}
    },
    {
        2656,
        {37276, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2656},
        {37276}
    },
    {
        2657,
        {20160, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2657},
        {20160}
    },
    {
        2658,
        {20303, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2658},
        {20303}
    },
    {
        2659,
        {20805, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2659},
        {20805}
    },
    {
        2660,
        {21313, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2660},
        {21313}
    },
    {
        2661,
        {24467, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2661},
        {24467}
    },
    {
        2662,
        {25102, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2662},
        {25102}
    },
    {
        2663,
        {26580, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2663},
        {26580}
    },
    {
        2664,
        {27713, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2664},
        {27713}
    },
    {
        2665,
        {28171, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2665},
        {28171}
    },
    {
        2666,
        {29539, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2666},
        {29539}
    },
    {
        2667,
        {32294, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2667},
        {32294}
    },
    {
        2668,
        {37325, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2668},
        {37325}
    },
    {
        2669,
        {37507, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2669},
        {37507}
    },
    {
        2670,
        {21460, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2670},
        {21460}
    },
    {
        2671,
        {22809, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2671},
        {22809}
    },
    {
        2672,
        {23487, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2672},
        {23487}
    },
    {
        2673,
        {28113, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2673},
        {28113}
    },
    {
        2674,
        {31069, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2674},
        {31069}
    },
    {
        2675,
        {32302, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2675},
        {32302}
    },
    {
        2676,
        {31899, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2676},
        {31899}
    },
    {
        2677,
        {22654, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2677},
        {22654}
    },
    {
        2678,
        {29087, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2678},
        {29087}
    },
    {
        2679,
        {20986, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2679},
        {20986}
    },
    {
        2680,
        {34899, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2680},
        {34899}
    },
    {
        2681,
        {36848, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2681},
        {36848}
    },
    {
        2682,
        {20426, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2682},
        {20426}
    },
    {
        2683,
        {23803, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2683},
        {23803}
    },
    {
        2684,
        {26149, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2684},
        {26149}
    },
    {
        2685,
        {30636, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2685},
        {30636}
    },
    {
        2686,
        {31459, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2686},
        {31459}
    },
    {
        2687,
        {33308, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2687},
        {33308}
    },
    {
        2688,
        {39423, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2688},
        {39423}
    },
    {
        2689,
        {20934, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2689},
        {20934}
    },
    {
        2690,
        {24490, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2690},
        {24490}
    },
    {
        2691,
        {26092, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2691},
        {26092}
    },
    {
        2692,
        {26991, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2692},
        {26991}
    },
    {
        2693,
        {27529, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2693},
        {27529}
    },
    {
        2694,
        {28147, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2694},
        {28147}
    },
    {
        2695,
        {28310, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2695},
        {28310}
    },
    {
        2696,
        {28516, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2696},
        {28516}
    },
    {
        2697,
        {30462, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2697},
        {30462}
    },
    {
        2698,
        {32020, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2698},
        {32020}
    },
    {
        2699,
        {24033, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2699},
        {24033}
    },
    {
        2700,
        {36981, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2700},
        {36981}
    },
    {
        2701,
        {37255, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2701},
        {37255}
    },
    {
        2702,
        {38918, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2702},
        {38918}
    },
    {
        2703,
        {20966, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2703},
        {20966}
    },
    {
        2704,
        {21021, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2704},
        {21021}
    },
    {
        2705,
        {25152, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2705},
        {25152}
    },
    {
        2706,
        {26257, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2706},
        {26257}
    },
    {
        2707,
        {26329, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2707},
        {26329}
    },
    {
        2708,
        {28186, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2708},
        {28186}
    },
    {
        2709,
        {24246, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2709},
        {24246}
    },
    {
        2710,
        {32210, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2710},
        {32210}
    },
    {
        2711,
        {32626, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2711},
        {32626}
    },
    {
        2712,
        {26360, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2712},
        {26360}
    },
    {
        2713,
        {34223, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2713},
        {34223}
    },
    {
        2714,
        {34295, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2714},
        {34295}
    },
    {
        2715,
        {35576, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2715},
        {35576}
    },
    {
        2716,
        {21161, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2716},
        {21161}
    },
    {
        2717,
        {21465, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2717},
        {21465}
    },
    {
        2718,
        {22899, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2718},
        {22899}
    },
    {
        2719,
        {24207, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2719},
        {24207}
    },
    {
        2720,
        {24464, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2720},
        {24464}
    },
    {
        2721,
        {24661, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2721},
        {24661}
    },
    {
        2722,
        {37604, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2722},
        {37604}
    },
    {
        2723,
        {38500, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2723},
        {38500}
    },
    {
        2724,
        {20663, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2724},
        {20663}
    },
    {
        2725,
        {20767, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2725},
        {20767}
    },
    {
        2726,
        {21213, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2726},
        {21213}
    },
    {
        2727,
        {21280, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2727},
        {21280}
    },
    {
        2728,
        {21319, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2728},
        {21319}
    },
    {
        2729,
        {21484, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2729},
        {21484}
    },
    {
        2730,
        {21736, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2730},
        {21736}
    },
    {
        2731,
        {21830, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2731},
        {21830}
    },
    {
        2732,
        {21809, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2732},
        {21809}
    },
    {
        2733,
        {22039, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2733},
        {22039}
    },
    {
        2734,
        {22888, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2734},
        {22888}
    },
    {
        2735,
        {22974, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2735},
        {22974}
    },
    {
        2736,
        {23100, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2736},
        {23100}
    },
    {
        2737,
        {23477, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2737},
        {23477}
    },
    {
        2738,
        {23558, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2738},
        {23558}
    },
    {
        2739,
        {23567, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2739},
        {23567}
    },
    {
        2740,
        {23569, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2740},
        {23569}
    },
    {
        2741,
        {23578, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2741},
        {23578}
    },
    {
        2742,
        {24196, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2742},
        {24196}
    },
    {
        2743,
        {24202, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2743},
        {24202}
    },
    {
        2744,
        {24288, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2744},
        {24288}
    },
    {
        2745,
        {24432, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2745},
        {24432}
    },
    {
        2746,
        {25215, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2746},
        {25215}
    },
    {
        2747,
        {25220, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2747},
        {25220}
    },
    {
        2748,
        {25307, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2748},
        {25307}
    },
    {
        2749,
        {25484, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2749},
        {25484}
    },
    {
        2750,
        {25463, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2750},
        {25463}
    },
    {
        2751,
        {26119, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2751},
        {26119}
    },
    {
        2752,
        {26124, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2752},
        {26124}
    },
    {
        2753,
        {26157, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2753},
        {26157}
    },
    {
        2754,
        {26230, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2754},
        {26230}
    },
    {
        2755,
        {26494, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2755},
        {26494}
    },
    {
        2756,
        {26786, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2756},
        {26786}
    },
    {
        2757,
        {27167, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2757},
        {27167}
    },
    {
        2758,
        {27189, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2758},
        {27189}
    },
    {
        2759,
        {27836, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2759},
        {27836}
    },
    {
        2760,
        {28040, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2760},
        {28040}
    },
    {
        2761,
        {28169, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2761},
        {28169}
    },
    {
        2762,
        {28248, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2762},
        {28248}
    },
    {
        2763,
        {28988, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2763},
        {28988}
    },
    {
        2764,
        {28966, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2764},
        {28966}
    },
    {
        2765,
        {29031, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2765},
        {29031}
    },
    {
        2766,
        {30151, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2766},
        {30151}
    },
    {
        2767,
        {30465, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2767},
        {30465}
    },
    {
        2768,
        {30813, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2768},
        {30813}
    },
    {
        2769,
        {30977, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2769},
        {30977}
    },
    {
        2770,
        {31077, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2770},
        {31077}
    },
    {
        2771,
        {31216, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2771},
        {31216}
    },
    {
        2772,
        {31456, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2772},
        {31456}
    },
    {
        2773,
        {31505, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2773},
        {31505}
    },
    {
        2774,
        {31911, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2774},
        {31911}
    },
    {
        2775,
        {32057, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2775},
        {32057}
    },
    {
        2776,
        {32918, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2776},
        {32918}
    },
    {
        2777,
        {33750, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2777},
        {33750}
    },
    {
        2778,
        {33931, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2778},
        {33931}
    },
    {
        2779,
        {34121, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2779},
        {34121}
    },
    {
        2780,
        {34909, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2780},
        {34909}
    },
    {
        2781,
        {35059, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2781},
        {35059}
    },
    {
        2782,
        {35359, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2782},
        {35359}
    },
    {
        2783,
        {35388, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2783},
        {35388}
    },
    {
        2784,
        {35412, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2784},
        {35412}
    },
    {
        2785,
        {35443, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2785},
        {35443}
    },
    {
        2786,
        {35937, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2786},
        {35937}
    },
    {
        2787,
        {36062, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2787},
        {36062}
    },
    {
        2788,
        {37284, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2788},
        {37284}
    },
    {
        2789,
        {37478, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2789},
        {37478}
    },
    {
        2790,
        {37758, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2790},
        {37758}
    },
    {
        2791,
        {37912, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2791},
        {37912}
    },
    {
        2792,
        {38556, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2792},
        {38556}
    },
    {
        2793,
        {38808, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2793},
        {38808}
    },
    {
        2794,
        {19978, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2794},
        {19978}
    },
    {
        2795,
        {19976, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2795},
        {19976}
    },
    {
        2796,
        {19998, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2796},
        {19998}
    },
    {
        2797,
        {20055, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2797},
        {20055}
    },
    {
        2798,
        {20887, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2798},
        {20887}
    },
    {
        2799,
        {21104, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2799},
        {21104}
    },
    {
        2800,
        {22478, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2800},
        {22478}
    },
    {
        2801,
        {22580, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2801},
        {22580}
    },
    {
        2802,
        {22732, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2802},
        {22732}
    },
    {
        2803,
        {23330, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2803},
        {23330}
    },
    {
        2804,
        {24120, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2804},
        {24120}
    },
    {
        2805,
        {24773, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2805},
        {24773}
    },
    {
        2806,
        {25854, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2806},
        {25854}
    },
    {
        2807,
        {26465, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2807},
        {26465}
    },
    {
        2808,
        {26454, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2808},
        {26454}
    },
    {
        2809,
        {27972, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2809},
        {27972}
    },
    {
        2810,
        {29366, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2810},
        {29366}
    },
    {
        2811,
        {30067, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2811},
        {30067}
    },
    {
        2812,
        {31331, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2812},
        {31331}
    },
    {
        2813,
        {33976, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2813},
        {33976}
    },
    {
        2814,
        {35698, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2814},
        {35698}
    },
    {
        2815,
        {37304, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2815},
        {37304}
    },
    {
        2816,
        {37664, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2816},
        {37664}
    },
    {
        2817,
        {22065, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2817},
        {22065}
    },
    {
        2818,
        {22516, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2818},
        {22516}
    },
    {
        2819,
        {39166, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2819},
        {39166}
    },
    {
        2820,
        {25325, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2820},
        {25325}
    },
    {
        2821,
        {26893, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2821},
        {26893}
    },
    {
        2822,
        {27542, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2822},
        {27542}
    },
    {
        2823,
        {29165, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2823},
        {29165}
    },
    {
        2824,
        {32340, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2824},
        {32340}
    },
    {
        2825,
        {32887, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2825},
        {32887}
    },
    {
        2826,
        {33394, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2826},
        {33394}
    },
    {
        2827,
        {35302, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2827},
        {35302}
    },
    {
        2828,
        {39135, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2828},
        {39135}
    },
    {
        2829,
        {34645, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2829},
        {34645}
    },
    {
        2830,
        {36785, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2830},
        {36785}
    },
    {
        2831,
        {23611, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2831},
        {23611}
    },
    {
        2832,
        {20280, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2832},
        {20280}
    },
    {
        2833,
        {20449, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2833},
        {20449}
    },
    {
        2834,
        {20405, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2834},
        {20405}
    },
    {
        2835,
        {21767, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2835},
        {21767}
    },
    {
        2836,
        {23072, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2836},
        {23072}
    },
    {
        2837,
        {23517, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2837},
        {23517}
    },
    {
        2838,
        {23529, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2838},
        {23529}
    },
    {
        2839,
        {24515, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2839},
        {24515}
    },
    {
        2840,
        {24910, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2840},
        {24910}
    },
    {
        2841,
        {25391, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2841},
        {25391}
    },
    {
        2842,
        {26032, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2842},
        {26032}
    },
    {
        2843,
        {26187, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2843},
        {26187}
    },
    {
        2844,
        {26862, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2844},
        {26862}
    },
    {
        2845,
        {27035, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2845},
        {27035}
    },
    {
        2846,
        {28024, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2846},
        {28024}
    },
    {
        2847,
        {28145, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2847},
        {28145}
    },
    {
        2848,
        {30003, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2848},
        {30003}
    },
    {
        2849,
        {30137, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2849},
        {30137}
    },
    {
        2850,
        {30495, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2850},
        {30495}
    },
    {
        2851,
        {31070, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2851},
        {31070}
    },
    {
        2852,
        {31206, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2852},
        {31206}
    },
    {
        2853,
        {32051, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2853},
        {32051}
    },
    {
        2854,
        {33251, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2854},
        {33251}
    },
    {
        2855,
        {33455, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2855},
        {33455}
    },
    {
        2856,
        {34218, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2856},
        {34218}
    },
    {
        2857,
        {35242, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2857},
        {35242}
    },
    {
        2858,
        {35386, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2858},
        {35386}
    },
    {
        2859,
        {36523, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2859},
        {36523}
    },
    {
        2860,
        {36763, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2860},
        {36763}
    },
    {
        2861,
        {36914, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2861},
        {36914}
    },
    {
        2862,
        {37341, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2862},
        {37341}
    },
    {
        2863,
        {38663, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2863},
        {38663}
    },
    {
        2864,
        {20154, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2864},
        {20154}
    },
    {
        2865,
        {20161, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2865},
        {20161}
    },
    {
        2866,
        {20995, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2866},
        {20995}
    },
    {
        2867,
        {22645, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2867},
        {22645}
    },
    {
        2868,
        {22764, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2868},
        {22764}
    },
    {
        2869,
        {23563, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2869},
        {23563}
    },
    {
        2870,
        {29978, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2870},
        {29978}
    },
    {
        2871,
        {23613, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2871},
        {23613}
    },
    {
        2872,
        {33102, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2872},
        {33102}
    },
    {
        2873,
        {35338, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2873},
        {35338}
    },
    {
        2874,
        {36805, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2874},
        {36805}
    },
    {
        2875,
        {38499, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2875},
        {38499}
    },
    {
        2876,
        {38765, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2876},
        {38765}
    },
    {
        2877,
        {31525, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2877},
        {31525}
    },
    {
        2878,
        {35535, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2878},
        {35535}
    },
    {
        2879,
        {38920, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2879},
        {38920}
    },
    {
        2880,
        {37218, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2880},
        {37218}
    },
    {
        2881,
        {22259, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2881},
        {22259}
    },
    {
        2882,
        {21416, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2882},
        {21416}
    },
    {
        2883,
        {36887, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2883},
        {36887}
    },
    {
        2884,
        {21561, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2884},
        {21561}
    },
    {
        2885,
        {22402, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2885},
        {22402}
    },
    {
        2886,
        {24101, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2886},
        {24101}
    },
    {
        2887,
        {25512, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2887},
        {25512}
    },
    {
        2888,
        {27700, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2888},
        {27700}
    },
    {
        2889,
        {28810, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2889},
        {28810}
    },
    {
        2890,
        {30561, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2890},
        {30561}
    },
    {
        2891,
        {31883, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2891},
        {31883}
    },
    {
        2892,
        {32736, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2892},
        {32736}
    },
    {
        2893,
        {34928, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2893},
        {34928}
    },
    {
        2894,
        {36930, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2894},
        {36930}
    },
    {
        2895,
        {37204, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2895},
        {37204}
    },
    {
        2896,
        {37648, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2896},
        {37648}
    },
    {
        2897,
        {37656, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2897},
        {37656}
    },
    {
        2898,
        {38543, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2898},
        {38543}
    },
    {
        2899,
        {29790, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2899},
        {29790}
    },
    {
        2900,
        {39620, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2900},
        {39620}
    },
    {
        2901,
        {23815, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2901},
        {23815}
    },
    {
        2902,
        {23913, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2902},
        {23913}
    },
    {
        2903,
        {25968, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2903},
        {25968}
    },
    {
        2904,
        {26530, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2904},
        {26530}
    },
    {
        2905,
        {36264, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2905},
        {36264}
    },
    {
        2906,
        {38619, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2906},
        {38619}
    },
    {
        2907,
        {25454, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2907},
        {25454}
    },
    {
        2908,
        {26441, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2908},
        {26441}
    },
    {
        2909,
        {26905, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2909},
        {26905}
    },
    {
        2910,
        {33733, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2910},
        {33733}
    },
    {
        2911,
        {38935, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2911},
        {38935}
    },
    {
        2912,
        {38592, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2912},
        {38592}
    },
    {
        2913,
        {35070, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2913},
        {35070}
    },
    {
        2914,
        {28548, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2914},
        {28548}
    },
    {
        2915,
        {25722, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2915},
        {25722}
    },
    {
        2916,
        {23544, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2916},
        {23544}
    },
    {
        2917,
        {19990, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2917},
        {19990}
    },
    {
        2918,
        {28716, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2918},
        {28716}
    },
    {
        2919,
        {30045, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2919},
        {30045}
    },
    {
        2920,
        {26159, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2920},
        {26159}
    },
    {
        2921,
        {20932, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2921},
        {20932}
    },
    {
        2922,
        {21046, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2922},
        {21046}
    },
    {
        2923,
        {21218, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2923},
        {21218}
    },
    {
        2924,
        {22995, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2924},
        {22995}
    },
    {
        2925,
        {24449, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2925},
        {24449}
    },
    {
        2926,
        {24615, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2926},
        {24615}
    },
    {
        2927,
        {25104, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2927},
        {25104}
    },
    {
        2928,
        {25919, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2928},
        {25919}
    },
    {
        2929,
        {25972, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2929},
        {25972}
    },
    {
        2930,
        {26143, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2930},
        {26143}
    },
    {
        2931,
        {26228, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2931},
        {26228}
    },
    {
        2932,
        {26866, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2932},
        {26866}
    },
    {
        2933,
        {26646, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2933},
        {26646}
    },
    {
        2934,
        {27491, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2934},
        {27491}
    },
    {
        2935,
        {28165, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2935},
        {28165}
    },
    {
        2936,
        {29298, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2936},
        {29298}
    },
    {
        2937,
        {29983, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2937},
        {29983}
    },
    {
        2938,
        {30427, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2938},
        {30427}
    },
    {
        2939,
        {31934, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2939},
        {31934}
    },
    {
        2940,
        {32854, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2940},
        {32854}
    },
    {
        2941,
        {22768, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2941},
        {22768}
    },
    {
        2942,
        {35069, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2942},
        {35069}
    },
    {
        2943,
        {35199, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2943},
        {35199}
    },
    {
        2944,
        {35488, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2944},
        {35488}
    },
    {
        2945,
        {35475, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2945},
        {35475}
    },
    {
        2946,
        {35531, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2946},
        {35531}
    },
    {
        2947,
        {36893, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2947},
        {36893}
    },
    {
        2948,
        {37266, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2948},
        {37266}
    },
    {
        2949,
        {38738, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2949},
        {38738}
    },
    {
        2950,
        {38745, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2950},
        {38745}
    },
    {
        2951,
        {25993, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2951},
        {25993}
    },
    {
        2952,
        {31246, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2952},
        {31246}
    },
    {
        2953,
        {33030, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2953},
        {33030}
    },
    {
        2954,
        {38587, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2954},
        {38587}
    },
    {
        2955,
        {24109, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2955},
        {24109}
    },
    {
        2956,
        {24796, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2956},
        {24796}
    },
    {
        2957,
        {25114, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2957},
        {25114}
    },
    {
        2958,
        {26021, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2958},
        {26021}
    },
    {
        2959,
        {26132, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2959},
        {26132}
    },
    {
        2960,
        {26512, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2960},
        {26512}
    },
    {
        2961,
        {30707, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2961},
        {30707}
    },
    {
        2962,
        {31309, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2962},
        {31309}
    },
    {
        2963,
        {31821, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2963},
        {31821}
    },
    {
        2964,
        {32318, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2964},
        {32318}
    },
    {
        2965,
        {33034, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2965},
        {33034}
    },
    {
        2966,
        {36012, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2966},
        {36012}
    },
    {
        2967,
        {36196, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2967},
        {36196}
    },
    {
        2968,
        {36321, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2968},
        {36321}
    },
    {
        2969,
        {36447, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2969},
        {36447}
    },
    {
        2970,
        {30889, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2970},
        {30889}
    },
    {
        2971,
        {20999, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2971},
        {20999}
    },
    {
        2972,
        {25305, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2972},
        {25305}
    },
    {
        2973,
        {25509, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2973},
        {25509}
    },
    {
        2974,
        {25666, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2974},
        {25666}
    },
    {
        2975,
        {25240, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2975},
        {25240}
    },
    {
        2976,
        {35373, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2976},
        {35373}
    },
    {
        2977,
        {31363, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2977},
        {31363}
    },
    {
        2978,
        {31680, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2978},
        {31680}
    },
    {
        2979,
        {35500, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2979},
        {35500}
    },
    {
        2980,
        {38634, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2980},
        {38634}
    },
    {
        2981,
        {32118, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2981},
        {32118}
    },
    {
        2982,
        {33292, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2982},
        {33292}
    },
    {
        2983,
        {34633, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2983},
        {34633}
    },
    {
        2984,
        {20185, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2984},
        {20185}
    },
    {
        2985,
        {20808, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2985},
        {20808}
    },
    {
        2986,
        {21315, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2986},
        {21315}
    },
    {
        2987,
        {21344, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2987},
        {21344}
    },
    {
        2988,
        {23459, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2988},
        {23459}
    },
    {
        2989,
        {23554, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2989},
        {23554}
    },
    {
        2990,
        {23574, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2990},
        {23574}
    },
    {
        2991,
        {24029, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2991},
        {24029}
    },
    {
        2992,
        {25126, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2992},
        {25126}
    },
    {
        2993,
        {25159, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2993},
        {25159}
    },
    {
        2994,
        {25776, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2994},
        {25776}
    },
    {
        2995,
        {26643, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2995},
        {26643}
    },
    {
        2996,
        {26676, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2996},
        {26676}
    },
    {
        2997,
        {27849, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2997},
        {27849}
    },
    {
        2998,
        {27973, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2998},
        {27973}
    },
    {
        2999,
        {27927, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {2999},
        {27927}
    },
    {
        3000,
        {26579, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3000},
        {26579}
    },
    {
        3001,
        {28508, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3001},
        {28508}
    },
    {
        3002,
        {29006, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3002},
        {29006}
    },
    {
        3003,
        {29053, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3003},
        {29053}
    },
    {
        3004,
        {26059, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3004},
        {26059}
    },
    {
        3005,
        {31359, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3005},
        {31359}
    },
    {
        3006,
        {31661, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3006},
        {31661}
    },
    {
        3007,
        {32218, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3007},
        {32218}
    },
    {
        3008,
        {32330, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3008},
        {32330}
    },
    {
        3009,
        {32680, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3009},
        {32680}
    },
    {
        3010,
        {33146, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3010},
        {33146}
    },
    {
        3011,
        {33307, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3011},
        {33307}
    },
    {
        3012,
        {33337, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3012},
        {33337}
    },
    {
        3013,
        {34214, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3013},
        {34214}
    },
    {
        3014,
        {35438, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3014},
        {35438}
    },
    {
        3015,
        {36046, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3015},
        {36046}
    },
    {
        3016,
        {36341, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3016},
        {36341}
    },
    {
        3017,
        {36984, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3017},
        {36984}
    },
    {
        3018,
        {36983, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3018},
        {36983}
    },
    {
        3019,
        {37549, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3019},
        {37549}
    },
    {
        3020,
        {37521, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3020},
        {37521}
    },
    {
        3021,
        {38275, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3021},
        {38275}
    },
    {
        3022,
        {39854, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3022},
        {39854}
    },
    {
        3023,
        {21069, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3023},
        {21069}
    },
    {
        3024,
        {21892, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3024},
        {21892}
    },
    {
        3025,
        {28472, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3025},
        {28472}
    },
    {
        3026,
        {28982, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3026},
        {28982}
    },
    {
        3027,
        {20840, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3027},
        {20840}
    },
    {
        3028,
        {31109, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3028},
        {31109}
    },
    {
        3029,
        {32341, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3029},
        {32341}
    },
    {
        3030,
        {33203, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3030},
        {33203}
    },
    {
        3031,
        {31950, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3031},
        {31950}
    },
    {
        3032,
        {22092, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3032},
        {22092}
    },
    {
        3033,
        {22609, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3033},
        {22609}
    },
    {
        3034,
        {23720, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3034},
        {23720}
    },
    {
        3035,
        {25514, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3035},
        {25514}
    },
    {
        3036,
        {26366, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3036},
        {26366}
    },
    {
        3037,
        {26365, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3037},
        {26365}
    },
    {
        3038,
        {26970, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3038},
        {26970}
    },
    {
        3039,
        {29401, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3039},
        {29401}
    },
    {
        3040,
        {30095, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3040},
        {30095}
    },
    {
        3041,
        {30094, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3041},
        {30094}
    },
    {
        3042,
        {30990, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3042},
        {30990}
    },
    {
        3043,
        {31062, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3043},
        {31062}
    },
    {
        3044,
        {31199, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3044},
        {31199}
    },
    {
        3045,
        {31895, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3045},
        {31895}
    },
    {
        3046,
        {32032, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3046},
        {32032}
    },
    {
        3047,
        {32068, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3047},
        {32068}
    },
    {
        3048,
        {34311, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3048},
        {34311}
    },
    {
        3049,
        {35380, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3049},
        {35380}
    },
    {
        3050,
        {38459, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3050},
        {38459}
    },
    {
        3051,
        {36961, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3051},
        {36961}
    },
    {
        3052,
        {40736, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3052},
        {40736}
    },
    {
        3053,
        {20711, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3053},
        {20711}
    },
    {
        3054,
        {21109, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3054},
        {21109}
    },
    {
        3055,
        {21452, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3055},
        {21452}
    },
    {
        3056,
        {21474, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3056},
        {21474}
    },
    {
        3057,
        {20489, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3057},
        {20489}
    },
    {
        3058,
        {21930, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3058},
        {21930}
    },
    {
        3059,
        {22766, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3059},
        {22766}
    },
    {
        3060,
        {22863, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3060},
        {22863}
    },
    {
        3061,
        {29245, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3061},
        {29245}
    },
    {
        3062,
        {23435, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3062},
        {23435}
    },
    {
        3063,
        {23652, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3063},
        {23652}
    },
    {
        3064,
        {21277, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3064},
        {21277}
    },
    {
        3065,
        {24803, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3065},
        {24803}
    },
    {
        3066,
        {24819, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3066},
        {24819}
    },
    {
        3067,
        {25436, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3067},
        {25436}
    },
    {
        3068,
        {25475, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3068},
        {25475}
    },
    {
        3069,
        {25407, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3069},
        {25407}
    },
    {
        3070,
        {25531, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3070},
        {25531}
    },
    {
        3071,
        {25805, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3071},
        {25805}
    },
    {
        3072,
        {26089, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3072},
        {26089}
    },
    {
        3073,
        {26361, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3073},
        {26361}
    },
    {
        3074,
        {24035, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3074},
        {24035}
    },
    {
        3075,
        {27085, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3075},
        {27085}
    },
    {
        3076,
        {27133, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3076},
        {27133}
    },
    {
        3077,
        {28437, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3077},
        {28437}
    },
    {
        3078,
        {29157, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3078},
        {29157}
    },
    {
        3079,
        {20105, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3079},
        {20105}
    },
    {
        3080,
        {30185, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3080},
        {30185}
    },
    {
        3081,
        {30456, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3081},
        {30456}
    },
    {
        3082,
        {31379, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3082},
        {31379}
    },
    {
        3083,
        {31967, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3083},
        {31967}
    },
    {
        3084,
        {32207, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3084},
        {32207}
    },
    {
        3085,
        {32156, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3085},
        {32156}
    },
    {
        3086,
        {32865, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3086},
        {32865}
    },
    {
        3087,
        {33609, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3087},
        {33609}
    },
    {
        3088,
        {33624, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3088},
        {33624}
    },
    {
        3089,
        {33900, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3089},
        {33900}
    },
    {
        3090,
        {33980, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3090},
        {33980}
    },
    {
        3091,
        {34299, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3091},
        {34299}
    },
    {
        3092,
        {35013, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3092},
        {35013}
    },
    {
        3093,
        {36208, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3093},
        {36208}
    },
    {
        3094,
        {36865, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3094},
        {36865}
    },
    {
        3095,
        {36973, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3095},
        {36973}
    },
    {
        3096,
        {37783, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3096},
        {37783}
    },
    {
        3097,
        {38684, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3097},
        {38684}
    },
    {
        3098,
        {39442, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3098},
        {39442}
    },
    {
        3099,
        {20687, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3099},
        {20687}
    },
    {
        3100,
        {22679, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3100},
        {22679}
    },
    {
        3101,
        {24974, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3101},
        {24974}
    },
    {
        3102,
        {33235, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3102},
        {33235}
    },
    {
        3103,
        {34101, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3103},
        {34101}
    },
    {
        3104,
        {36104, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3104},
        {36104}
    },
    {
        3105,
        {36896, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3105},
        {36896}
    },
    {
        3106,
        {20419, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3106},
        {20419}
    },
    {
        3107,
        {20596, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3107},
        {20596}
    },
    {
        3108,
        {21063, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3108},
        {21063}
    },
    {
        3109,
        {21363, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3109},
        {21363}
    },
    {
        3110,
        {24687, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3110},
        {24687}
    },
    {
        3111,
        {25417, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3111},
        {25417}
    },
    {
        3112,
        {26463, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3112},
        {26463}
    },
    {
        3113,
        {28204, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3113},
        {28204}
    },
    {
        3114,
        {36275, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3114},
        {36275}
    },
    {
        3115,
        {36895, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3115},
        {36895}
    },
    {
        3116,
        {20439, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3116},
        {20439}
    },
    {
        3117,
        {23646, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3117},
        {23646}
    },
    {
        3118,
        {36042, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3118},
        {36042}
    },
    {
        3119,
        {26063, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3119},
        {26063}
    },
    {
        3120,
        {32154, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3120},
        {32154}
    },
    {
        3121,
        {21330, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3121},
        {21330}
    },
    {
        3122,
        {34966, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3122},
        {34966}
    },
    {
        3123,
        {20854, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3123},
        {20854}
    },
    {
        3124,
        {25539, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3124},
        {25539}
    },
    {
        3125,
        {23384, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3125},
        {23384}
    },
    {
        3126,
        {23403, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3126},
        {23403}
    },
    {
        3127,
        {23562, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3127},
        {23562}
    },
    {
        3128,
        {25613, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3128},
        {25613}
    },
    {
        3129,
        {26449, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3129},
        {26449}
    },
    {
        3130,
        {36956, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3130},
        {36956}
    },
    {
        3131,
        {20182, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3131},
        {20182}
    },
    {
        3132,
        {22810, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3132},
        {22810}
    },
    {
        3133,
        {22826, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3133},
        {22826}
    },
    {
        3134,
        {27760, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3134},
        {27760}
    },
    {
        3135,
        {35409, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3135},
        {35409}
    },
    {
        3136,
        {21822, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3136},
        {21822}
    },
    {
        3137,
        {22549, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3137},
        {22549}
    },
    {
        3138,
        {22949, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3138},
        {22949}
    },
    {
        3139,
        {24816, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3139},
        {24816}
    },
    {
        3140,
        {25171, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3140},
        {25171}
    },
    {
        3141,
        {26561, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3141},
        {26561}
    },
    {
        3142,
        {33333, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3142},
        {33333}
    },
    {
        3143,
        {26965, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3143},
        {26965}
    },
    {
        3144,
        {38464, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3144},
        {38464}
    },
    {
        3145,
        {39364, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3145},
        {39364}
    },
    {
        3146,
        {39464, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3146},
        {39464}
    },
    {
        3147,
        {20307, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3147},
        {20307}
    },
    {
        3148,
        {22534, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3148},
        {22534}
    },
    {
        3149,
        {23550, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3149},
        {23550}
    },
    {
        3150,
        {32784, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3150},
        {32784}
    },
    {
        3151,
        {23729, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3151},
        {23729}
    },
    {
        3152,
        {24111, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3152},
        {24111}
    },
    {
        3153,
        {24453, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3153},
        {24453}
    },
    {
        3154,
        {24608, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3154},
        {24608}
    },
    {
        3155,
        {24907, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3155},
        {24907}
    },
    {
        3156,
        {25140, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3156},
        {25140}
    },
    {
        3157,
        {26367, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3157},
        {26367}
    },
    {
        3158,
        {27888, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3158},
        {27888}
    },
    {
        3159,
        {28382, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3159},
        {28382}
    },
    {
        3160,
        {32974, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3160},
        {32974}
    },
    {
        3161,
        {33151, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3161},
        {33151}
    },
    {
        3162,
        {33492, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3162},
        {33492}
    },
    {
        3163,
        {34955, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3163},
        {34955}
    },
    {
        3164,
        {36024, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3164},
        {36024}
    },
    {
        3165,
        {36864, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3165},
        {36864}
    },
    {
        3166,
        {36910, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3166},
        {36910}
    },
    {
        3167,
        {38538, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3167},
        {38538}
    },
    {
        3168,
        {40667, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3168},
        {40667}
    },
    {
        3169,
        {39899, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3169},
        {39899}
    },
    {
        3170,
        {20195, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3170},
        {20195}
    },
    {
        3171,
        {21488, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3171},
        {21488}
    },
    {
        3172,
        {22823, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3172},
        {22823}
    },
    {
        3173,
        {31532, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3173},
        {31532}
    },
    {
        3174,
        {37261, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3174},
        {37261}
    },
    {
        3175,
        {38988, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3175},
        {38988}
    },
    {
        3176,
        {40441, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3176},
        {40441}
    },
    {
        3177,
        {28381, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3177},
        {28381}
    },
    {
        3178,
        {28711, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3178},
        {28711}
    },
    {
        3179,
        {21331, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3179},
        {21331}
    },
    {
        3180,
        {21828, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3180},
        {21828}
    },
    {
        3181,
        {23429, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3181},
        {23429}
    },
    {
        3182,
        {25176, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3182},
        {25176}
    },
    {
        3183,
        {25246, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3183},
        {25246}
    },
    {
        3184,
        {25299, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3184},
        {25299}
    },
    {
        3185,
        {27810, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3185},
        {27810}
    },
    {
        3186,
        {28655, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3186},
        {28655}
    },
    {
        3187,
        {29730, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3187},
        {29730}
    },
    {
        3188,
        {35351, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3188},
        {35351}
    },
    {
        3189,
        {37944, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3189},
        {37944}
    },
    {
        3190,
        {28609, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3190},
        {28609}
    },
    {
        3191,
        {35582, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3191},
        {35582}
    },
    {
        3192,
        {33592, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3192},
        {33592}
    },
    {
        3193,
        {20967, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3193},
        {20967}
    },
    {
        3194,
        {34552, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3194},
        {34552}
    },
    {
        3195,
        {21482, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3195},
        {21482}
    },
    {
        3196,
        {21481, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3196},
        {21481}
    },
    {
        3197,
        {20294, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3197},
        {20294}
    },
    {
        3198,
        {36948, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3198},
        {36948}
    },
    {
        3199,
        {36784, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3199},
        {36784}
    },
    {
        3200,
        {22890, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3200},
        {22890}
    },
    {
        3201,
        {33073, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3201},
        {33073}
    },
    {
        3202,
        {24061, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3202},
        {24061}
    },
    {
        3203,
        {31466, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3203},
        {31466}
    },
    {
        3204,
        {36799, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3204},
        {36799}
    },
    {
        3205,
        {26842, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3205},
        {26842}
    },
    {
        3206,
        {35895, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3206},
        {35895}
    },
    {
        3207,
        {29432, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3207},
        {29432}
    },
    {
        3208,
        {40008, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3208},
        {40008}
    },
    {
        3209,
        {27197, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3209},
        {27197}
    },
    {
        3210,
        {35504, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3210},
        {35504}
    },
    {
        3211,
        {20025, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3211},
        {20025}
    },
    {
        3212,
        {21336, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3212},
        {21336}
    },
    {
        3213,
        {22022, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3213},
        {22022}
    },
    {
        3214,
        {22374, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3214},
        {22374}
    },
    {
        3215,
        {25285, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3215},
        {25285}
    },
    {
        3216,
        {25506, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3216},
        {25506}
    },
    {
        3217,
        {26086, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3217},
        {26086}
    },
    {
        3218,
        {27470, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3218},
        {27470}
    },
    {
        3219,
        {28129, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3219},
        {28129}
    },
    {
        3220,
        {28251, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3220},
        {28251}
    },
    {
        3221,
        {28845, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3221},
        {28845}
    },
    {
        3222,
        {30701, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3222},
        {30701}
    },
    {
        3223,
        {31471, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3223},
        {31471}
    },
    {
        3224,
        {31658, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3224},
        {31658}
    },
    {
        3225,
        {32187, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3225},
        {32187}
    },
    {
        3226,
        {32829, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3226},
        {32829}
    },
    {
        3227,
        {32966, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3227},
        {32966}
    },
    {
        3228,
        {34507, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3228},
        {34507}
    },
    {
        3229,
        {35477, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3229},
        {35477}
    },
    {
        3230,
        {37723, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3230},
        {37723}
    },
    {
        3231,
        {22243, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3231},
        {22243}
    },
    {
        3232,
        {22727, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3232},
        {22727}
    },
    {
        3233,
        {24382, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3233},
        {24382}
    },
    {
        3234,
        {26029, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3234},
        {26029}
    },
    {
        3235,
        {26262, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3235},
        {26262}
    },
    {
        3236,
        {27264, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3236},
        {27264}
    },
    {
        3237,
        {27573, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3237},
        {27573}
    },
    {
        3238,
        {30007, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3238},
        {30007}
    },
    {
        3239,
        {35527, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3239},
        {35527}
    },
    {
        3240,
        {20516, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3240},
        {20516}
    },
    {
        3241,
        {30693, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3241},
        {30693}
    },
    {
        3242,
        {22320, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3242},
        {22320}
    },
    {
        3243,
        {24347, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3243},
        {24347}
    },
    {
        3244,
        {24677, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3244},
        {24677}
    },
    {
        3245,
        {26234, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3245},
        {26234}
    },
    {
        3246,
        {27744, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3246},
        {27744}
    },
    {
        3247,
        {30196, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3247},
        {30196}
    },
    {
        3248,
        {31258, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3248},
        {31258}
    },
    {
        3249,
        {32622, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3249},
        {32622}
    },
    {
        3250,
        {33268, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3250},
        {33268}
    },
    {
        3251,
        {34584, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3251},
        {34584}
    },
    {
        3252,
        {36933, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3252},
        {36933}
    },
    {
        3253,
        {39347, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3253},
        {39347}
    },
    {
        3254,
        {31689, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3254},
        {31689}
    },
    {
        3255,
        {30044, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3255},
        {30044}
    },
    {
        3256,
        {31481, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3256},
        {31481}
    },
    {
        3257,
        {31569, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3257},
        {31569}
    },
    {
        3258,
        {33988, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3258},
        {33988}
    },
    {
        3259,
        {36880, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3259},
        {36880}
    },
    {
        3260,
        {31209, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3260},
        {31209}
    },
    {
        3261,
        {31378, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3261},
        {31378}
    },
    {
        3262,
        {33590, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3262},
        {33590}
    },
    {
        3263,
        {23265, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3263},
        {23265}
    },
    {
        3264,
        {30528, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3264},
        {30528}
    },
    {
        3265,
        {20013, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3265},
        {20013}
    },
    {
        3266,
        {20210, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3266},
        {20210}
    },
    {
        3267,
        {23449, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3267},
        {23449}
    },
    {
        3268,
        {24544, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3268},
        {24544}
    },
    {
        3269,
        {25277, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3269},
        {25277}
    },
    {
        3270,
        {26172, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3270},
        {26172}
    },
    {
        3271,
        {26609, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3271},
        {26609}
    },
    {
        3272,
        {27880, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3272},
        {27880}
    },
    {
        3273,
        {34411, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3273},
        {34411}
    },
    {
        3274,
        {34935, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3274},
        {34935}
    },
    {
        3275,
        {35387, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3275},
        {35387}
    },
    {
        3276,
        {37198, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3276},
        {37198}
    },
    {
        3277,
        {37619, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3277},
        {37619}
    },
    {
        3278,
        {39376, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3278},
        {39376}
    },
    {
        3279,
        {27159, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3279},
        {27159}
    },
    {
        3280,
        {28710, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3280},
        {28710}
    },
    {
        3281,
        {29482, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3281},
        {29482}
    },
    {
        3282,
        {33511, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3282},
        {33511}
    },
    {
        3283,
        {33879, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3283},
        {33879}
    },
    {
        3284,
        {36015, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3284},
        {36015}
    },
    {
        3285,
        {19969, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3285},
        {19969}
    },
    {
        3286,
        {20806, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3286},
        {20806}
    },
    {
        3287,
        {20939, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3287},
        {20939}
    },
    {
        3288,
        {21899, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3288},
        {21899}
    },
    {
        3289,
        {23541, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3289},
        {23541}
    },
    {
        3290,
        {24086, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3290},
        {24086}
    },
    {
        3291,
        {24115, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3291},
        {24115}
    },
    {
        3292,
        {24193, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3292},
        {24193}
    },
    {
        3293,
        {24340, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3293},
        {24340}
    },
    {
        3294,
        {24373, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3294},
        {24373}
    },
    {
        3295,
        {24427, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3295},
        {24427}
    },
    {
        3296,
        {24500, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3296},
        {24500}
    },
    {
        3297,
        {25074, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3297},
        {25074}
    },
    {
        3298,
        {25361, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3298},
        {25361}
    },
    {
        3299,
        {26274, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3299},
        {26274}
    },
    {
        3300,
        {26397, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3300},
        {26397}
    },
    {
        3301,
        {28526, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3301},
        {28526}
    },
    {
        3302,
        {29266, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3302},
        {29266}
    },
    {
        3303,
        {30010, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3303},
        {30010}
    },
    {
        3304,
        {30522, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3304},
        {30522}
    },
    {
        3305,
        {32884, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3305},
        {32884}
    },
    {
        3306,
        {33081, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3306},
        {33081}
    },
    {
        3307,
        {33144, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3307},
        {33144}
    },
    {
        3308,
        {34678, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3308},
        {34678}
    },
    {
        3309,
        {35519, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3309},
        {35519}
    },
    {
        3310,
        {35548, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3310},
        {35548}
    },
    {
        3311,
        {36229, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3311},
        {36229}
    },
    {
        3312,
        {36339, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3312},
        {36339}
    },
    {
        3313,
        {37530, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3313},
        {37530}
    },
    {
        3314,
        {38263, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3314},
        {38263}
    },
    {
        3315,
        {38914, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3315},
        {38914}
    },
    {
        3316,
        {40165, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3316},
        {40165}
    },
    {
        3317,
        {21189, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3317},
        {21189}
    },
    {
        3318,
        {25431, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3318},
        {25431}
    },
    {
        3319,
        {30452, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3319},
        {30452}
    },
    {
        3320,
        {26389, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3320},
        {26389}
    },
    {
        3321,
        {27784, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3321},
        {27784}
    },
    {
        3322,
        {29645, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3322},
        {29645}
    },
    {
        3323,
        {36035, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3323},
        {36035}
    },
    {
        3324,
        {37806, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3324},
        {37806}
    },
    {
        3325,
        {38515, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3325},
        {38515}
    },
    {
        3326,
        {27941, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3326},
        {27941}
    },
    {
        3327,
        {22684, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3327},
        {22684}
    },
    {
        3328,
        {26894, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3328},
        {26894}
    },
    {
        3329,
        {27084, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3329},
        {27084}
    },
    {
        3330,
        {36861, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3330},
        {36861}
    },
    {
        3331,
        {37786, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3331},
        {37786}
    },
    {
        3332,
        {30171, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3332},
        {30171}
    },
    {
        3333,
        {36890, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3333},
        {36890}
    },
    {
        3334,
        {22618, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3334},
        {22618}
    },
    {
        3335,
        {26626, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3335},
        {26626}
    },
    {
        3336,
        {25524, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3336},
        {25524}
    },
    {
        3337,
        {27131, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3337},
        {27131}
    },
    {
        3338,
        {20291, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3338},
        {20291}
    },
    {
        3339,
        {28460, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3339},
        {28460}
    },
    {
        3340,
        {26584, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3340},
        {26584}
    },
    {
        3341,
        {36795, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3341},
        {36795}
    },
    {
        3342,
        {34086, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3342},
        {34086}
    },
    {
        3343,
        {32180, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3343},
        {32180}
    },
    {
        3344,
        {37716, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3344},
        {37716}
    },
    {
        3345,
        {26943, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3345},
        {26943}
    },
    {
        3346,
        {28528, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3346},
        {28528}
    },
    {
        3347,
        {22378, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3347},
        {22378}
    },
    {
        3348,
        {22775, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3348},
        {22775}
    },
    {
        3349,
        {23340, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3349},
        {23340}
    },
    {
        3350,
        {32044, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3350},
        {32044}
    },
    {
        3351,
        {29226, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3351},
        {29226}
    },
    {
        3352,
        {21514, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3352},
        {21514}
    },
    {
        3353,
        {37347, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3353},
        {37347}
    },
    {
        3354,
        {40372, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3354},
        {40372}
    },
    {
        3355,
        {20141, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3355},
        {20141}
    },
    {
        3356,
        {20302, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3356},
        {20302}
    },
    {
        3357,
        {20572, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3357},
        {20572}
    },
    {
        3358,
        {20597, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3358},
        {20597}
    },
    {
        3359,
        {21059, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3359},
        {21059}
    },
    {
        3360,
        {35998, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3360},
        {35998}
    },
    {
        3361,
        {21576, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3361},
        {21576}
    },
    {
        3362,
        {22564, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3362},
        {22564}
    },
    {
        3363,
        {23450, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3363},
        {23450}
    },
    {
        3364,
        {24093, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3364},
        {24093}
    },
    {
        3365,
        {24213, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3365},
        {24213}
    },
    {
        3366,
        {24237, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3366},
        {24237}
    },
    {
        3367,
        {24311, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3367},
        {24311}
    },
    {
        3368,
        {24351, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3368},
        {24351}
    },
    {
        3369,
        {24716, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3369},
        {24716}
    },
    {
        3370,
        {25269, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3370},
        {25269}
    },
    {
        3371,
        {25402, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3371},
        {25402}
    },
    {
        3372,
        {25552, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3372},
        {25552}
    },
    {
        3373,
        {26799, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3373},
        {26799}
    },
    {
        3374,
        {27712, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3374},
        {27712}
    },
    {
        3375,
        {30855, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3375},
        {30855}
    },
    {
        3376,
        {31118, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3376},
        {31118}
    },
    {
        3377,
        {31243, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3377},
        {31243}
    },
    {
        3378,
        {32224, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3378},
        {32224}
    },
    {
        3379,
        {33351, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3379},
        {33351}
    },
    {
        3380,
        {35330, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3380},
        {35330}
    },
    {
        3381,
        {35558, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3381},
        {35558}
    },
    {
        3382,
        {36420, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3382},
        {36420}
    },
    {
        3383,
        {36883, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3383},
        {36883}
    },
    {
        3384,
        {37048, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3384},
        {37048}
    },
    {
        3385,
        {37165, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3385},
        {37165}
    },
    {
        3386,
        {37336, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3386},
        {37336}
    },
    {
        3387,
        {40718, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3387},
        {40718}
    },
    {
        3388,
        {27877, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3388},
        {27877}
    },
    {
        3389,
        {25688, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3389},
        {25688}
    },
    {
        3390,
        {25826, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3390},
        {25826}
    },
    {
        3391,
        {25973, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3391},
        {25973}
    },
    {
        3392,
        {28404, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3392},
        {28404}
    },
    {
        3393,
        {30340, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3393},
        {30340}
    },
    {
        3394,
        {31515, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3394},
        {31515}
    },
    {
        3395,
        {36969, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3395},
        {36969}
    },
    {
        3396,
        {37841, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3396},
        {37841}
    },
    {
        3397,
        {28346, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3397},
        {28346}
    },
    {
        3398,
        {21746, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3398},
        {21746}
    },
    {
        3399,
        {24505, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3399},
        {24505}
    },
    {
        3400,
        {25764, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3400},
        {25764}
    },
    {
        3401,
        {36685, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3401},
        {36685}
    },
    {
        3402,
        {36845, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3402},
        {36845}
    },
    {
        3403,
        {37444, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3403},
        {37444}
    },
    {
        3404,
        {20856, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3404},
        {20856}
    },
    {
        3405,
        {22635, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3405},
        {22635}
    },
    {
        3406,
        {22825, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3406},
        {22825}
    },
    {
        3407,
        {23637, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3407},
        {23637}
    },
    {
        3408,
        {24215, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3408},
        {24215}
    },
    {
        3409,
        {28155, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3409},
        {28155}
    },
    {
        3410,
        {32399, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3410},
        {32399}
    },
    {
        3411,
        {29980, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3411},
        {29980}
    },
    {
        3412,
        {36028, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3412},
        {36028}
    },
    {
        3413,
        {36578, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3413},
        {36578}
    },
    {
        3414,
        {39003, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3414},
        {39003}
    },
    {
        3415,
        {28857, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3415},
        {28857}
    },
    {
        3416,
        {20253, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3416},
        {20253}
    },
    {
        3417,
        {27583, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3417},
        {27583}
    },
    {
        3418,
        {28593, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3418},
        {28593}
    },
    {
        3419,
        {30000, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3419},
        {30000}
    },
    {
        3420,
        {38651, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3420},
        {38651}
    },
    {
        3421,
        {20814, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3421},
        {20814}
    },
    {
        3422,
        {21520, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3422},
        {21520}
    },
    {
        3423,
        {22581, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3423},
        {22581}
    },
    {
        3424,
        {22615, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3424},
        {22615}
    },
    {
        3425,
        {22956, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3425},
        {22956}
    },
    {
        3426,
        {23648, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3426},
        {23648}
    },
    {
        3427,
        {24466, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3427},
        {24466}
    },
    {
        3428,
        {26007, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3428},
        {26007}
    },
    {
        3429,
        {26460, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3429},
        {26460}
    },
    {
        3430,
        {28193, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3430},
        {28193}
    },
    {
        3431,
        {30331, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3431},
        {30331}
    },
    {
        3432,
        {33759, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3432},
        {33759}
    },
    {
        3433,
        {36077, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3433},
        {36077}
    },
    {
        3434,
        {36884, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3434},
        {36884}
    },
    {
        3435,
        {37117, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3435},
        {37117}
    },
    {
        3436,
        {37709, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3436},
        {37709}
    },
    {
        3437,
        {30757, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3437},
        {30757}
    },
    {
        3438,
        {30778, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3438},
        {30778}
    },
    {
        3439,
        {21162, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3439},
        {21162}
    },
    {
        3440,
        {24230, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3440},
        {24230}
    },
    {
        3441,
        {22303, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3441},
        {22303}
    },
    {
        3442,
        {22900, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3442},
        {22900}
    },
    {
        3443,
        {24594, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3443},
        {24594}
    },
    {
        3444,
        {20498, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3444},
        {20498}
    },
    {
        3445,
        {20826, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3445},
        {20826}
    },
    {
        3446,
        {20908, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3446},
        {20908}
    },
    {
        3447,
        {20941, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3447},
        {20941}
    },
    {
        3448,
        {20992, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3448},
        {20992}
    },
    {
        3449,
        {21776, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3449},
        {21776}
    },
    {
        3450,
        {22612, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3450},
        {22612}
    },
    {
        3451,
        {22616, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3451},
        {22616}
    },
    {
        3452,
        {22871, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3452},
        {22871}
    },
    {
        3453,
        {23445, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3453},
        {23445}
    },
    {
        3454,
        {23798, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3454},
        {23798}
    },
    {
        3455,
        {23947, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3455},
        {23947}
    },
    {
        3456,
        {24764, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3456},
        {24764}
    },
    {
        3457,
        {25237, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3457},
        {25237}
    },
    {
        3458,
        {25645, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3458},
        {25645}
    },
    {
        3459,
        {26481, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3459},
        {26481}
    },
    {
        3460,
        {26691, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3460},
        {26691}
    },
    {
        3461,
        {26812, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3461},
        {26812}
    },
    {
        3462,
        {26847, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3462},
        {26847}
    },
    {
        3463,
        {30423, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3463},
        {30423}
    },
    {
        3464,
        {28120, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3464},
        {28120}
    },
    {
        3465,
        {28271, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3465},
        {28271}
    },
    {
        3466,
        {28059, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3466},
        {28059}
    },
    {
        3467,
        {28783, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3467},
        {28783}
    },
    {
        3468,
        {29128, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3468},
        {29128}
    },
    {
        3469,
        {24403, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3469},
        {24403}
    },
    {
        3470,
        {30168, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3470},
        {30168}
    },
    {
        3471,
        {31095, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3471},
        {31095}
    },
    {
        3472,
        {31561, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3472},
        {31561}
    },
    {
        3473,
        {31572, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3473},
        {31572}
    },
    {
        3474,
        {31570, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3474},
        {31570}
    },
    {
        3475,
        {31958, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3475},
        {31958}
    },
    {
        3476,
        {32113, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3476},
        {32113}
    },
    {
        3477,
        {21040, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3477},
        {21040}
    },
    {
        3478,
        {33891, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3478},
        {33891}
    },
    {
        3479,
        {34153, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3479},
        {34153}
    },
    {
        3480,
        {34276, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3480},
        {34276}
    },
    {
        3481,
        {35342, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3481},
        {35342}
    },
    {
        3482,
        {35588, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3482},
        {35588}
    },
    {
        3483,
        {35910, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3483},
        {35910}
    },
    {
        3484,
        {36367, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3484},
        {36367}
    },
    {
        3485,
        {36867, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3485},
        {36867}
    },
    {
        3486,
        {36879, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3486},
        {36879}
    },
    {
        3487,
        {37913, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3487},
        {37913}
    },
    {
        3488,
        {38518, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3488},
        {38518}
    },
    {
        3489,
        {38957, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3489},
        {38957}
    },
    {
        3490,
        {39472, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3490},
        {39472}
    },
    {
        3491,
        {38360, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3491},
        {38360}
    },
    {
        3492,
        {20685, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3492},
        {20685}
    },
    {
        3493,
        {21205, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3493},
        {21205}
    },
    {
        3494,
        {21516, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3494},
        {21516}
    },
    {
        3495,
        {22530, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3495},
        {22530}
    },
    {
        3496,
        {23566, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3496},
        {23566}
    },
    {
        3497,
        {24999, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3497},
        {24999}
    },
    {
        3498,
        {25758, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3498},
        {25758}
    },
    {
        3499,
        {27934, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3499},
        {27934}
    },
    {
        3500,
        {30643, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3500},
        {30643}
    },
    {
        3501,
        {31461, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3501},
        {31461}
    },
    {
        3502,
        {33012, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3502},
        {33012}
    },
    {
        3503,
        {33796, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3503},
        {33796}
    },
    {
        3504,
        {36947, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3504},
        {36947}
    },
    {
        3505,
        {37509, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3505},
        {37509}
    },
    {
        3506,
        {23776, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3506},
        {23776}
    },
    {
        3507,
        {40199, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3507},
        {40199}
    },
    {
        3508,
        {21311, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3508},
        {21311}
    },
    {
        3509,
        {24471, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3509},
        {24471}
    },
    {
        3510,
        {24499, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3510},
        {24499}
    },
    {
        3511,
        {28060, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3511},
        {28060}
    },
    {
        3512,
        {29305, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3512},
        {29305}
    },
    {
        3513,
        {30563, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3513},
        {30563}
    },
    {
        3514,
        {31167, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3514},
        {31167}
    },
    {
        3515,
        {31716, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3515},
        {31716}
    },
    {
        3516,
        {27602, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3516},
        {27602}
    },
    {
        3517,
        {29420, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3517},
        {29420}
    },
    {
        3518,
        {35501, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3518},
        {35501}
    },
    {
        3519,
        {26627, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3519},
        {26627}
    },
    {
        3520,
        {27233, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3520},
        {27233}
    },
    {
        3521,
        {20984, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3521},
        {20984}
    },
    {
        3522,
        {31361, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3522},
        {31361}
    },
    {
        3523,
        {26932, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3523},
        {26932}
    },
    {
        3524,
        {23626, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3524},
        {23626}
    },
    {
        3525,
        {40182, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3525},
        {40182}
    },
    {
        3526,
        {33515, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3526},
        {33515}
    },
    {
        3527,
        {23493, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3527},
        {23493}
    },
    {
        3528,
        {37193, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3528},
        {37193}
    },
    {
        3529,
        {28702, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3529},
        {28702}
    },
    {
        3530,
        {22136, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3530},
        {22136}
    },
    {
        3531,
        {23663, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3531},
        {23663}
    },
    {
        3532,
        {24775, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3532},
        {24775}
    },
    {
        3533,
        {25958, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3533},
        {25958}
    },
    {
        3534,
        {27788, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3534},
        {27788}
    },
    {
        3535,
        {35930, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3535},
        {35930}
    },
    {
        3536,
        {36929, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3536},
        {36929}
    },
    {
        3537,
        {38931, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3537},
        {38931}
    },
    {
        3538,
        {21585, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3538},
        {21585}
    },
    {
        3539,
        {26311, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3539},
        {26311}
    },
    {
        3540,
        {37389, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3540},
        {37389}
    },
    {
        3541,
        {22856, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3541},
        {22856}
    },
    {
        3542,
        {37027, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3542},
        {37027}
    },
    {
        3543,
        {20869, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3543},
        {20869}
    },
    {
        3544,
        {20045, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3544},
        {20045}
    },
    {
        3545,
        {20970, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3545},
        {20970}
    },
    {
        3546,
        {34201, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3546},
        {34201}
    },
    {
        3547,
        {35598, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3547},
        {35598}
    },
    {
        3548,
        {28760, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3548},
        {28760}
    },
    {
        3549,
        {25466, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3549},
        {25466}
    },
    {
        3550,
        {37707, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3550},
        {37707}
    },
    {
        3551,
        {26978, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3551},
        {26978}
    },
    {
        3552,
        {39348, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3552},
        {39348}
    },
    {
        3553,
        {32260, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3553},
        {32260}
    },
    {
        3554,
        {30071, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3554},
        {30071}
    },
    {
        3555,
        {21335, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3555},
        {21335}
    },
    {
        3556,
        {26976, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3556},
        {26976}
    },
    {
        3557,
        {36575, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3557},
        {36575}
    },
    {
        3558,
        {38627, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3558},
        {38627}
    },
    {
        3559,
        {27741, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3559},
        {27741}
    },
    {
        3560,
        {20108, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3560},
        {20108}
    },
    {
        3561,
        {23612, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3561},
        {23612}
    },
    {
        3562,
        {24336, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3562},
        {24336}
    },
    {
        3563,
        {36841, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3563},
        {36841}
    },
    {
        3564,
        {21250, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3564},
        {21250}
    },
    {
        3565,
        {36049, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3565},
        {36049}
    },
    {
        3566,
        {32905, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3566},
        {32905}
    },
    {
        3567,
        {34425, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3567},
        {34425}
    },
    {
        3568,
        {24319, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3568},
        {24319}
    },
    {
        3569,
        {26085, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3569},
        {26085}
    },
    {
        3570,
        {20083, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3570},
        {20083}
    },
    {
        3571,
        {20837, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3571},
        {20837}
    },
    {
        3572,
        {22914, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3572},
        {22914}
    },
    {
        3573,
        {23615, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3573},
        {23615}
    },
    {
        3574,
        {38894, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3574},
        {38894}
    },
    {
        3575,
        {20219, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3575},
        {20219}
    },
    {
        3576,
        {22922, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3576},
        {22922}
    },
    {
        3577,
        {24525, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3577},
        {24525}
    },
    {
        3578,
        {35469, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3578},
        {35469}
    },
    {
        3579,
        {28641, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3579},
        {28641}
    },
    {
        3580,
        {31152, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3580},
        {31152}
    },
    {
        3581,
        {31074, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3581},
        {31074}
    },
    {
        3582,
        {23527, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3582},
        {23527}
    },
    {
        3583,
        {33905, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3583},
        {33905}
    },
    {
        3584,
        {29483, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3584},
        {29483}
    },
    {
        3585,
        {29105, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3585},
        {29105}
    },
    {
        3586,
        {24180, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3586},
        {24180}
    },
    {
        3587,
        {24565, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3587},
        {24565}
    },
    {
        3588,
        {25467, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3588},
        {25467}
    },
    {
        3589,
        {25754, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3589},
        {25754}
    },
    {
        3590,
        {29123, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3590},
        {29123}
    },
    {
        3591,
        {31896, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3591},
        {31896}
    },
    {
        3592,
        {20035, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3592},
        {20035}
    },
    {
        3593,
        {24316, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3593},
        {24316}
    },
    {
        3594,
        {20043, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3594},
        {20043}
    },
    {
        3595,
        {22492, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3595},
        {22492}
    },
    {
        3596,
        {22178, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3596},
        {22178}
    },
    {
        3597,
        {24745, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3597},
        {24745}
    },
    {
        3598,
        {28611, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3598},
        {28611}
    },
    {
        3599,
        {32013, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3599},
        {32013}
    },
    {
        3600,
        {33021, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3600},
        {33021}
    },
    {
        3601,
        {33075, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3601},
        {33075}
    },
    {
        3602,
        {33215, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3602},
        {33215}
    },
    {
        3603,
        {36786, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3603},
        {36786}
    },
    {
        3604,
        {35223, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3604},
        {35223}
    },
    {
        3605,
        {34468, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3605},
        {34468}
    },
    {
        3606,
        {24052, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3606},
        {24052}
    },
    {
        3607,
        {25226, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3607},
        {25226}
    },
    {
        3608,
        {25773, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3608},
        {25773}
    },
    {
        3609,
        {35207, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3609},
        {35207}
    },
    {
        3610,
        {26487, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3610},
        {26487}
    },
    {
        3611,
        {27874, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3611},
        {27874}
    },
    {
        3612,
        {27966, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3612},
        {27966}
    },
    {
        3613,
        {29750, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3613},
        {29750}
    },
    {
        3614,
        {30772, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3614},
        {30772}
    },
    {
        3615,
        {23110, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3615},
        {23110}
    },
    {
        3616,
        {32629, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3616},
        {32629}
    },
    {
        3617,
        {33453, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3617},
        {33453}
    },
    {
        3618,
        {39340, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3618},
        {39340}
    },
    {
        3619,
        {20467, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3619},
        {20467}
    },
    {
        3620,
        {24259, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3620},
        {24259}
    },
    {
        3621,
        {25309, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3621},
        {25309}
    },
    {
        3622,
        {25490, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3622},
        {25490}
    },
    {
        3623,
        {25943, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3623},
        {25943}
    },
    {
        3624,
        {26479, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3624},
        {26479}
    },
    {
        3625,
        {30403, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3625},
        {30403}
    },
    {
        3626,
        {29260, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3626},
        {29260}
    },
    {
        3627,
        {32972, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3627},
        {32972}
    },
    {
        3628,
        {32954, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3628},
        {32954}
    },
    {
        3629,
        {36649, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3629},
        {36649}
    },
    {
        3630,
        {37197, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3630},
        {37197}
    },
    {
        3631,
        {20493, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3631},
        {20493}
    },
    {
        3632,
        {22521, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3632},
        {22521}
    },
    {
        3633,
        {23186, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3633},
        {23186}
    },
    {
        3634,
        {26757, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3634},
        {26757}
    },
    {
        3635,
        {26995, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3635},
        {26995}
    },
    {
        3636,
        {29028, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3636},
        {29028}
    },
    {
        3637,
        {29437, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3637},
        {29437}
    },
    {
        3638,
        {36023, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3638},
        {36023}
    },
    {
        3639,
        {22770, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3639},
        {22770}
    },
    {
        3640,
        {36064, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3640},
        {36064}
    },
    {
        3641,
        {38506, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3641},
        {38506}
    },
    {
        3642,
        {36889, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3642},
        {36889}
    },
    {
        3643,
        {34687, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3643},
        {34687}
    },
    {
        3644,
        {31204, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3644},
        {31204}
    },
    {
        3645,
        {30695, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3645},
        {30695}
    },
    {
        3646,
        {33833, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3646},
        {33833}
    },
    {
        3647,
        {20271, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3647},
        {20271}
    },
    {
        3648,
        {21093, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3648},
        {21093}
    },
    {
        3649,
        {21338, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3649},
        {21338}
    },
    {
        3650,
        {25293, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3650},
        {25293}
    },
    {
        3651,
        {26575, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3651},
        {26575}
    },
    {
        3652,
        {27850, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3652},
        {27850}
    },
    {
        3653,
        {30333, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3653},
        {30333}
    },
    {
        3654,
        {31636, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3654},
        {31636}
    },
    {
        3655,
        {31893, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3655},
        {31893}
    },
    {
        3656,
        {33334, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3656},
        {33334}
    },
    {
        3657,
        {34180, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3657},
        {34180}
    },
    {
        3658,
        {36843, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3658},
        {36843}
    },
    {
        3659,
        {26333, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3659},
        {26333}
    },
    {
        3660,
        {28448, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3660},
        {28448}
    },
    {
        3661,
        {29190, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3661},
        {29190}
    },
    {
        3662,
        {32283, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3662},
        {32283}
    },
    {
        3663,
        {33707, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3663},
        {33707}
    },
    {
        3664,
        {39361, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3664},
        {39361}
    },
    {
        3665,
        {40614, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3665},
        {40614}
    },
    {
        3666,
        {20989, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3666},
        {20989}
    },
    {
        3667,
        {31665, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3667},
        {31665}
    },
    {
        3668,
        {30834, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3668},
        {30834}
    },
    {
        3669,
        {31672, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3669},
        {31672}
    },
    {
        3670,
        {32903, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3670},
        {32903}
    },
    {
        3671,
        {31560, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3671},
        {31560}
    },
    {
        3672,
        {27368, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3672},
        {27368}
    },
    {
        3673,
        {24161, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3673},
        {24161}
    },
    {
        3674,
        {32908, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3674},
        {32908}
    },
    {
        3675,
        {30033, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3675},
        {30033}
    },
    {
        3676,
        {30048, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3676},
        {30048}
    },
    {
        3677,
        {20843, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3677},
        {20843}
    },
    {
        3678,
        {37474, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3678},
        {37474}
    },
    {
        3679,
        {28300, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3679},
        {28300}
    },
    {
        3680,
        {30330, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3680},
        {30330}
    },
    {
        3681,
        {37271, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3681},
        {37271}
    },
    {
        3682,
        {39658, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3682},
        {39658}
    },
    {
        3683,
        {20240, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3683},
        {20240}
    },
    {
        3684,
        {32624, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3684},
        {32624}
    },
    {
        3685,
        {25244, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3685},
        {25244}
    },
    {
        3686,
        {31567, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3686},
        {31567}
    },
    {
        3687,
        {38309, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3687},
        {38309}
    },
    {
        3688,
        {40169, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3688},
        {40169}
    },
    {
        3689,
        {22138, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3689},
        {22138}
    },
    {
        3690,
        {22617, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3690},
        {22617}
    },
    {
        3691,
        {34532, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3691},
        {34532}
    },
    {
        3692,
        {38588, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3692},
        {38588}
    },
    {
        3693,
        {20276, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3693},
        {20276}
    },
    {
        3694,
        {21028, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3694},
        {21028}
    },
    {
        3695,
        {21322, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3695},
        {21322}
    },
    {
        3696,
        {21453, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3696},
        {21453}
    },
    {
        3697,
        {21467, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3697},
        {21467}
    },
    {
        3698,
        {24070, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3698},
        {24070}
    },
    {
        3699,
        {25644, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3699},
        {25644}
    },
    {
        3700,
        {26001, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3700},
        {26001}
    },
    {
        3701,
        {26495, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3701},
        {26495}
    },
    {
        3702,
        {27710, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3702},
        {27710}
    },
    {
        3703,
        {27726, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3703},
        {27726}
    },
    {
        3704,
        {29256, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3704},
        {29256}
    },
    {
        3705,
        {29359, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3705},
        {29359}
    },
    {
        3706,
        {29677, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3706},
        {29677}
    },
    {
        3707,
        {30036, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3707},
        {30036}
    },
    {
        3708,
        {32321, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3708},
        {32321}
    },
    {
        3709,
        {33324, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3709},
        {33324}
    },
    {
        3710,
        {34281, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3710},
        {34281}
    },
    {
        3711,
        {36009, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3711},
        {36009}
    },
    {
        3712,
        {31684, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3712},
        {31684}
    },
    {
        3713,
        {37318, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3713},
        {37318}
    },
    {
        3714,
        {29033, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3714},
        {29033}
    },
    {
        3715,
        {38930, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3715},
        {38930}
    },
    {
        3716,
        {39151, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3716},
        {39151}
    },
    {
        3717,
        {25405, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3717},
        {25405}
    },
    {
        3718,
        {26217, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3718},
        {26217}
    },
    {
        3719,
        {30058, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3719},
        {30058}
    },
    {
        3720,
        {30436, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3720},
        {30436}
    },
    {
        3721,
        {30928, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3721},
        {30928}
    },
    {
        3722,
        {34115, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3722},
        {34115}
    },
    {
        3723,
        {34542, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3723},
        {34542}
    },
    {
        3724,
        {21290, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3724},
        {21290}
    },
    {
        3725,
        {21329, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3725},
        {21329}
    },
    {
        3726,
        {21542, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3726},
        {21542}
    },
    {
        3727,
        {22915, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3727},
        {22915}
    },
    {
        3728,
        {24199, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3728},
        {24199}
    },
    {
        3729,
        {24444, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3729},
        {24444}
    },
    {
        3730,
        {24754, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3730},
        {24754}
    },
    {
        3731,
        {25161, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3731},
        {25161}
    },
    {
        3732,
        {25209, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3732},
        {25209}
    },
    {
        3733,
        {25259, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3733},
        {25259}
    },
    {
        3734,
        {26000, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3734},
        {26000}
    },
    {
        3735,
        {27604, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3735},
        {27604}
    },
    {
        3736,
        {27852, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3736},
        {27852}
    },
    {
        3737,
        {30130, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3737},
        {30130}
    },
    {
        3738,
        {30382, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3738},
        {30382}
    },
    {
        3739,
        {30865, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3739},
        {30865}
    },
    {
        3740,
        {31192, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3740},
        {31192}
    },
    {
        3741,
        {32203, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3741},
        {32203}
    },
    {
        3742,
        {32631, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3742},
        {32631}
    },
    {
        3743,
        {32933, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3743},
        {32933}
    },
    {
        3744,
        {34987, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3744},
        {34987}
    },
    {
        3745,
        {35513, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3745},
        {35513}
    },
    {
        3746,
        {36027, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3746},
        {36027}
    },
    {
        3747,
        {36991, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3747},
        {36991}
    },
    {
        3748,
        {38750, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3748},
        {38750}
    },
    {
        3749,
        {39131, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3749},
        {39131}
    },
    {
        3750,
        {27147, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3750},
        {27147}
    },
    {
        3751,
        {31800, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3751},
        {31800}
    },
    {
        3752,
        {20633, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3752},
        {20633}
    },
    {
        3753,
        {23614, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3753},
        {23614}
    },
    {
        3754,
        {24494, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3754},
        {24494}
    },
    {
        3755,
        {26503, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3755},
        {26503}
    },
    {
        3756,
        {27608, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3756},
        {27608}
    },
    {
        3757,
        {29749, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3757},
        {29749}
    },
    {
        3758,
        {30473, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3758},
        {30473}
    },
    {
        3759,
        {32654, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3759},
        {32654}
    },
    {
        3760,
        {40763, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3760},
        {40763}
    },
    {
        3761,
        {26570, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3761},
        {26570}
    },
    {
        3762,
        {31255, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3762},
        {31255}
    },
    {
        3763,
        {21305, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3763},
        {21305}
    },
    {
        3764,
        {30091, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3764},
        {30091}
    },
    {
        3765,
        {39661, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3765},
        {39661}
    },
    {
        3766,
        {24422, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3766},
        {24422}
    },
    {
        3767,
        {33181, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3767},
        {33181}
    },
    {
        3768,
        {33777, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3768},
        {33777}
    },
    {
        3769,
        {32920, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3769},
        {32920}
    },
    {
        3770,
        {24380, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3770},
        {24380}
    },
    {
        3771,
        {24517, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3771},
        {24517}
    },
    {
        3772,
        {30050, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3772},
        {30050}
    },
    {
        3773,
        {31558, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3773},
        {31558}
    },
    {
        3774,
        {36924, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3774},
        {36924}
    },
    {
        3775,
        {26727, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3775},
        {26727}
    },
    {
        3776,
        {23019, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3776},
        {23019}
    },
    {
        3777,
        {23195, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3777},
        {23195}
    },
    {
        3778,
        {32016, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3778},
        {32016}
    },
    {
        3779,
        {30334, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3779},
        {30334}
    },
    {
        3780,
        {35628, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3780},
        {35628}
    },
    {
        3781,
        {20469, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3781},
        {20469}
    },
    {
        3782,
        {24426, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3782},
        {24426}
    },
    {
        3783,
        {27161, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3783},
        {27161}
    },
    {
        3784,
        {27703, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3784},
        {27703}
    },
    {
        3785,
        {28418, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3785},
        {28418}
    },
    {
        3786,
        {29922, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3786},
        {29922}
    },
    {
        3787,
        {31080, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3787},
        {31080}
    },
    {
        3788,
        {34920, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3788},
        {34920}
    },
    {
        3789,
        {35413, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3789},
        {35413}
    },
    {
        3790,
        {35961, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3790},
        {35961}
    },
    {
        3791,
        {24287, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3791},
        {24287}
    },
    {
        3792,
        {25551, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3792},
        {25551}
    },
    {
        3793,
        {30149, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3793},
        {30149}
    },
    {
        3794,
        {31186, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3794},
        {31186}
    },
    {
        3795,
        {33495, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3795},
        {33495}
    },
    {
        3796,
        {37672, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3796},
        {37672}
    },
    {
        3797,
        {37618, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3797},
        {37618}
    },
    {
        3798,
        {33948, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3798},
        {33948}
    },
    {
        3799,
        {34541, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3799},
        {34541}
    },
    {
        3800,
        {39981, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3800},
        {39981}
    },
    {
        3801,
        {21697, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3801},
        {21697}
    },
    {
        3802,
        {24428, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3802},
        {24428}
    },
    {
        3803,
        {25996, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3803},
        {25996}
    },
    {
        3804,
        {27996, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3804},
        {27996}
    },
    {
        3805,
        {28693, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3805},
        {28693}
    },
    {
        3806,
        {36007, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3806},
        {36007}
    },
    {
        3807,
        {36051, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3807},
        {36051}
    },
    {
        3808,
        {38971, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3808},
        {38971}
    },
    {
        3809,
        {25935, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3809},
        {25935}
    },
    {
        3810,
        {29942, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3810},
        {29942}
    },
    {
        3811,
        {19981, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3811},
        {19981}
    },
    {
        3812,
        {20184, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3812},
        {20184}
    },
    {
        3813,
        {22496, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3813},
        {22496}
    },
    {
        3814,
        {22827, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3814},
        {22827}
    },
    {
        3815,
        {23142, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3815},
        {23142}
    },
    {
        3816,
        {23500, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3816},
        {23500}
    },
    {
        3817,
        {20904, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3817},
        {20904}
    },
    {
        3818,
        {24067, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3818},
        {24067}
    },
    {
        3819,
        {24220, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3819},
        {24220}
    },
    {
        3820,
        {24598, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3820},
        {24598}
    },
    {
        3821,
        {25206, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3821},
        {25206}
    },
    {
        3822,
        {25975, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3822},
        {25975}
    },
    {
        3823,
        {26023, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3823},
        {26023}
    },
    {
        3824,
        {26222, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3824},
        {26222}
    },
    {
        3825,
        {28014, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3825},
        {28014}
    },
    {
        3826,
        {29238, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3826},
        {29238}
    },
    {
        3827,
        {31526, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3827},
        {31526}
    },
    {
        3828,
        {33104, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3828},
        {33104}
    },
    {
        3829,
        {33178, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3829},
        {33178}
    },
    {
        3830,
        {33433, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3830},
        {33433}
    },
    {
        3831,
        {35676, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3831},
        {35676}
    },
    {
        3832,
        {36000, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3832},
        {36000}
    },
    {
        3833,
        {36070, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3833},
        {36070}
    },
    {
        3834,
        {36212, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3834},
        {36212}
    },
    {
        3835,
        {38428, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3835},
        {38428}
    },
    {
        3836,
        {38468, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3836},
        {38468}
    },
    {
        3837,
        {20398, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3837},
        {20398}
    },
    {
        3838,
        {25771, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3838},
        {25771}
    },
    {
        3839,
        {27494, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3839},
        {27494}
    },
    {
        3840,
        {33310, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3840},
        {33310}
    },
    {
        3841,
        {33889, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3841},
        {33889}
    },
    {
        3842,
        {34154, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3842},
        {34154}
    },
    {
        3843,
        {37096, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3843},
        {37096}
    },
    {
        3844,
        {23553, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3844},
        {23553}
    },
    {
        3845,
        {26963, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3845},
        {26963}
    },
    {
        3846,
        {39080, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3846},
        {39080}
    },
    {
        3847,
        {33914, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3847},
        {33914}
    },
    {
        3848,
        {34135, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3848},
        {34135}
    },
    {
        3849,
        {20239, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3849},
        {20239}
    },
    {
        3850,
        {21103, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3850},
        {21103}
    },
    {
        3851,
        {24489, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3851},
        {24489}
    },
    {
        3852,
        {24133, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3852},
        {24133}
    },
    {
        3853,
        {26381, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3853},
        {26381}
    },
    {
        3854,
        {31119, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3854},
        {31119}
    },
    {
        3855,
        {33145, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3855},
        {33145}
    },
    {
        3856,
        {35079, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3856},
        {35079}
    },
    {
        3857,
        {35206, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3857},
        {35206}
    },
    {
        3858,
        {28149, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3858},
        {28149}
    },
    {
        3859,
        {24343, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3859},
        {24343}
    },
    {
        3860,
        {25173, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3860},
        {25173}
    },
    {
        3861,
        {27832, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3861},
        {27832}
    },
    {
        3862,
        {20175, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3862},
        {20175}
    },
    {
        3863,
        {29289, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3863},
        {29289}
    },
    {
        3864,
        {39826, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3864},
        {39826}
    },
    {
        3865,
        {20998, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3865},
        {20998}
    },
    {
        3866,
        {21563, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3866},
        {21563}
    },
    {
        3867,
        {22132, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3867},
        {22132}
    },
    {
        3868,
        {22707, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3868},
        {22707}
    },
    {
        3869,
        {24996, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3869},
        {24996}
    },
    {
        3870,
        {25198, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3870},
        {25198}
    },
    {
        3871,
        {28954, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3871},
        {28954}
    },
    {
        3872,
        {22894, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3872},
        {22894}
    },
    {
        3873,
        {31881, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3873},
        {31881}
    },
    {
        3874,
        {31966, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3874},
        {31966}
    },
    {
        3875,
        {32027, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3875},
        {32027}
    },
    {
        3876,
        {38640, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3876},
        {38640}
    },
    {
        3877,
        {25991, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3877},
        {25991}
    },
    {
        3878,
        {32862, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3878},
        {32862}
    },
    {
        3879,
        {19993, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3879},
        {19993}
    },
    {
        3880,
        {20341, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3880},
        {20341}
    },
    {
        3881,
        {20853, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3881},
        {20853}
    },
    {
        3882,
        {22592, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3882},
        {22592}
    },
    {
        3883,
        {24163, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3883},
        {24163}
    },
    {
        3884,
        {24179, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3884},
        {24179}
    },
    {
        3885,
        {24330, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3885},
        {24330}
    },
    {
        3886,
        {26564, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3886},
        {26564}
    },
    {
        3887,
        {20006, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3887},
        {20006}
    },
    {
        3888,
        {34109, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3888},
        {34109}
    },
    {
        3889,
        {38281, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3889},
        {38281}
    },
    {
        3890,
        {38491, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3890},
        {38491}
    },
    {
        3891,
        {31859, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3891},
        {31859}
    },
    {
        3892,
        {38913, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3892},
        {38913}
    },
    {
        3893,
        {20731, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3893},
        {20731}
    },
    {
        3894,
        {22721, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3894},
        {22721}
    },
    {
        3895,
        {30294, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3895},
        {30294}
    },
    {
        3896,
        {30887, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3896},
        {30887}
    },
    {
        3897,
        {21029, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3897},
        {21029}
    },
    {
        3898,
        {30629, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3898},
        {30629}
    },
    {
        3899,
        {34065, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3899},
        {34065}
    },
    {
        3900,
        {31622, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3900},
        {31622}
    },
    {
        3901,
        {20559, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3901},
        {20559}
    },
    {
        3902,
        {22793, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3902},
        {22793}
    },
    {
        3903,
        {29255, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3903},
        {29255}
    },
    {
        3904,
        {31687, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3904},
        {31687}
    },
    {
        3905,
        {32232, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3905},
        {32232}
    },
    {
        3906,
        {36794, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3906},
        {36794}
    },
    {
        3907,
        {36820, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3907},
        {36820}
    },
    {
        3908,
        {36941, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3908},
        {36941}
    },
    {
        3909,
        {20415, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3909},
        {20415}
    },
    {
        3910,
        {21193, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3910},
        {21193}
    },
    {
        3911,
        {23081, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3911},
        {23081}
    },
    {
        3912,
        {24321, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3912},
        {24321}
    },
    {
        3913,
        {38829, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3913},
        {38829}
    },
    {
        3914,
        {20445, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3914},
        {20445}
    },
    {
        3915,
        {33303, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3915},
        {33303}
    },
    {
        3916,
        {37610, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3916},
        {37610}
    },
    {
        3917,
        {22275, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3917},
        {22275}
    },
    {
        3918,
        {25429, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3918},
        {25429}
    },
    {
        3919,
        {27497, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3919},
        {27497}
    },
    {
        3920,
        {29995, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3920},
        {29995}
    },
    {
        3921,
        {35036, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3921},
        {35036}
    },
    {
        3922,
        {36628, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3922},
        {36628}
    },
    {
        3923,
        {31298, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3923},
        {31298}
    },
    {
        3924,
        {21215, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3924},
        {21215}
    },
    {
        3925,
        {22675, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3925},
        {22675}
    },
    {
        3926,
        {24917, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3926},
        {24917}
    },
    {
        3927,
        {25098, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3927},
        {25098}
    },
    {
        3928,
        {26286, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3928},
        {26286}
    },
    {
        3929,
        {27597, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3929},
        {27597}
    },
    {
        3930,
        {31807, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3930},
        {31807}
    },
    {
        3931,
        {33769, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3931},
        {33769}
    },
    {
        3932,
        {20515, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3932},
        {20515}
    },
    {
        3933,
        {20472, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3933},
        {20472}
    },
    {
        3934,
        {21253, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3934},
        {21253}
    },
    {
        3935,
        {21574, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3935},
        {21574}
    },
    {
        3936,
        {22577, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3936},
        {22577}
    },
    {
        3937,
        {22857, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3937},
        {22857}
    },
    {
        3938,
        {23453, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3938},
        {23453}
    },
    {
        3939,
        {23792, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3939},
        {23792}
    },
    {
        3940,
        {23791, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3940},
        {23791}
    },
    {
        3941,
        {23849, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3941},
        {23849}
    },
    {
        3942,
        {24214, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3942},
        {24214}
    },
    {
        3943,
        {25265, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3943},
        {25265}
    },
    {
        3944,
        {25447, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3944},
        {25447}
    },
    {
        3945,
        {25918, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3945},
        {25918}
    },
    {
        3946,
        {26041, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3946},
        {26041}
    },
    {
        3947,
        {26379, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3947},
        {26379}
    },
    {
        3948,
        {27861, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3948},
        {27861}
    },
    {
        3949,
        {27873, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3949},
        {27873}
    },
    {
        3950,
        {28921, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3950},
        {28921}
    },
    {
        3951,
        {30770, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3951},
        {30770}
    },
    {
        3952,
        {32299, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3952},
        {32299}
    },
    {
        3953,
        {32990, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3953},
        {32990}
    },
    {
        3954,
        {33459, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3954},
        {33459}
    },
    {
        3955,
        {33804, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3955},
        {33804}
    },
    {
        3956,
        {34028, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3956},
        {34028}
    },
    {
        3957,
        {34562, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3957},
        {34562}
    },
    {
        3958,
        {35090, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3958},
        {35090}
    },
    {
        3959,
        {35370, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3959},
        {35370}
    },
    {
        3960,
        {35914, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3960},
        {35914}
    },
    {
        3961,
        {37030, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3961},
        {37030}
    },
    {
        3962,
        {37586, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3962},
        {37586}
    },
    {
        3963,
        {39165, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3963},
        {39165}
    },
    {
        3964,
        {40179, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3964},
        {40179}
    },
    {
        3965,
        {40300, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3965},
        {40300}
    },
    {
        3966,
        {20047, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3966},
        {20047}
    },
    {
        3967,
        {20129, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3967},
        {20129}
    },
    {
        3968,
        {20621, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3968},
        {20621}
    },
    {
        3969,
        {21078, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3969},
        {21078}
    },
    {
        3970,
        {22346, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3970},
        {22346}
    },
    {
        3971,
        {22952, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3971},
        {22952}
    },
    {
        3972,
        {24125, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3972},
        {24125}
    },
    {
        3973,
        {24536, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3973},
        {24536}
    },
    {
        3974,
        {24537, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3974},
        {24537}
    },
    {
        3975,
        {25151, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3975},
        {25151}
    },
    {
        3976,
        {26292, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3976},
        {26292}
    },
    {
        3977,
        {26395, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3977},
        {26395}
    },
    {
        3978,
        {26576, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3978},
        {26576}
    },
    {
        3979,
        {26834, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3979},
        {26834}
    },
    {
        3980,
        {20882, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3980},
        {20882}
    },
    {
        3981,
        {32033, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3981},
        {32033}
    },
    {
        3982,
        {32938, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3982},
        {32938}
    },
    {
        3983,
        {33192, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3983},
        {33192}
    },
    {
        3984,
        {35584, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3984},
        {35584}
    },
    {
        3985,
        {35980, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3985},
        {35980}
    },
    {
        3986,
        {36031, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3986},
        {36031}
    },
    {
        3987,
        {37502, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3987},
        {37502}
    },
    {
        3988,
        {38450, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3988},
        {38450}
    },
    {
        3989,
        {21536, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3989},
        {21536}
    },
    {
        3990,
        {38956, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3990},
        {38956}
    },
    {
        3991,
        {21271, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3991},
        {21271}
    },
    {
        3992,
        {20693, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3992},
        {20693}
    },
    {
        3993,
        {21340, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3993},
        {21340}
    },
    {
        3994,
        {22696, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3994},
        {22696}
    },
    {
        3995,
        {25778, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3995},
        {25778}
    },
    {
        3996,
        {26420, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3996},
        {26420}
    },
    {
        3997,
        {29287, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3997},
        {29287}
    },
    {
        3998,
        {30566, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3998},
        {30566}
    },
    {
        3999,
        {31302, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {3999},
        {31302}
    },
    {
        4000,
        {37350, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4000},
        {37350}
    },
    {
        4001,
        {21187, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4001},
        {21187}
    },
    {
        4002,
        {27809, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4002},
        {27809}
    },
    {
        4003,
        {27526, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4003},
        {27526}
    },
    {
        4004,
        {22528, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4004},
        {22528}
    },
    {
        4005,
        {24140, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4005},
        {24140}
    },
    {
        4006,
        {22868, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4006},
        {22868}
    },
    {
        4007,
        {26412, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4007},
        {26412}
    },
    {
        4008,
        {32763, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4008},
        {32763}
    },
    {
        4009,
        {20961, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4009},
        {20961}
    },
    {
        4010,
        {30406, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4010},
        {30406}
    },
    {
        4011,
        {25705, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4011},
        {25705}
    },
    {
        4012,
        {30952, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4012},
        {30952}
    },
    {
        4013,
        {39764, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4013},
        {39764}
    },
    {
        4014,
        {40635, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4014},
        {40635}
    },
    {
        4015,
        {22475, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4015},
        {22475}
    },
    {
        4016,
        {22969, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4016},
        {22969}
    },
    {
        4017,
        {26151, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4017},
        {26151}
    },
    {
        4018,
        {26522, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4018},
        {26522}
    },
    {
        4019,
        {27598, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4019},
        {27598}
    },
    {
        4020,
        {21737, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4020},
        {21737}
    },
    {
        4021,
        {27097, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4021},
        {27097}
    },
    {
        4022,
        {24149, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4022},
        {24149}
    },
    {
        4023,
        {33180, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4023},
        {33180}
    },
    {
        4024,
        {26517, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4024},
        {26517}
    },
    {
        4025,
        {39850, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4025},
        {39850}
    },
    {
        4026,
        {26622, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4026},
        {26622}
    },
    {
        4027,
        {40018, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4027},
        {40018}
    },
    {
        4028,
        {26717, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4028},
        {26717}
    },
    {
        4029,
        {20134, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4029},
        {20134}
    },
    {
        4030,
        {20451, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4030},
        {20451}
    },
    {
        4031,
        {21448, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4031},
        {21448}
    },
    {
        4032,
        {25273, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4032},
        {25273}
    },
    {
        4033,
        {26411, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4033},
        {26411}
    },
    {
        4034,
        {27819, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4034},
        {27819}
    },
    {
        4035,
        {36804, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4035},
        {36804}
    },
    {
        4036,
        {20397, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4036},
        {20397}
    },
    {
        4037,
        {32365, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4037},
        {32365}
    },
    {
        4038,
        {40639, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4038},
        {40639}
    },
    {
        4039,
        {19975, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4039},
        {19975}
    },
    {
        4040,
        {24930, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4040},
        {24930}
    },
    {
        4041,
        {28288, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4041},
        {28288}
    },
    {
        4042,
        {28459, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4042},
        {28459}
    },
    {
        4043,
        {34067, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4043},
        {34067}
    },
    {
        4044,
        {21619, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4044},
        {21619}
    },
    {
        4045,
        {26410, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4045},
        {26410}
    },
    {
        4046,
        {39749, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4046},
        {39749}
    },
    {
        4047,
        {24051, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4047},
        {24051}
    },
    {
        4048,
        {31637, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4048},
        {31637}
    },
    {
        4049,
        {23724, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4049},
        {23724}
    },
    {
        4050,
        {23494, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4050},
        {23494}
    },
    {
        4051,
        {34588, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4051},
        {34588}
    },
    {
        4052,
        {28234, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4052},
        {28234}
    },
    {
        4053,
        {34001, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4053},
        {34001}
    },
    {
        4054,
        {31252, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4054},
        {31252}
    },
    {
        4055,
        {33032, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4055},
        {33032}
    },
    {
        4056,
        {22937, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4056},
        {22937}
    },
    {
        4057,
        {31885, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4057},
        {31885}
    },
    {
        4058,
        {27665, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4058},
        {27665}
    },
    {
        4059,
        {30496, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4059},
        {30496}
    },
    {
        4060,
        {21209, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4060},
        {21209}
    },
    {
        4061,
        {22818, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4061},
        {22818}
    },
    {
        4062,
        {28961, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4062},
        {28961}
    },
    {
        4063,
        {29279, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4063},
        {29279}
    },
    {
        4064,
        {30683, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4064},
        {30683}
    },
    {
        4065,
        {38695, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4065},
        {38695}
    },
    {
        4066,
        {40289, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4066},
        {40289}
    },
    {
        4067,
        {26891, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4067},
        {26891}
    },
    {
        4068,
        {23167, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4068},
        {23167}
    },
    {
        4069,
        {23064, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4069},
        {23064}
    },
    {
        4070,
        {20901, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4070},
        {20901}
    },
    {
        4071,
        {21517, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4071},
        {21517}
    },
    {
        4072,
        {21629, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4072},
        {21629}
    },
    {
        4073,
        {26126, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4073},
        {26126}
    },
    {
        4074,
        {30431, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4074},
        {30431}
    },
    {
        4075,
        {36855, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4075},
        {36855}
    },
    {
        4076,
        {37528, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4076},
        {37528}
    },
    {
        4077,
        {40180, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4077},
        {40180}
    },
    {
        4078,
        {23018, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4078},
        {23018}
    },
    {
        4079,
        {29277, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4079},
        {29277}
    },
    {
        4080,
        {28357, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4080},
        {28357}
    },
    {
        4081,
        {20813, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4081},
        {20813}
    },
    {
        4082,
        {26825, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4082},
        {26825}
    },
    {
        4083,
        {32191, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4083},
        {32191}
    },
    {
        4084,
        {32236, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4084},
        {32236}
    },
    {
        4085,
        {38754, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4085},
        {38754}
    },
    {
        4086,
        {40634, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4086},
        {40634}
    },
    {
        4087,
        {25720, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4087},
        {25720}
    },
    {
        4088,
        {27169, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4088},
        {27169}
    },
    {
        4089,
        {33538, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4089},
        {33538}
    },
    {
        4090,
        {22916, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4090},
        {22916}
    },
    {
        4091,
        {23391, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4091},
        {23391}
    },
    {
        4092,
        {27611, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4092},
        {27611}
    },
    {
        4093,
        {29467, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4093},
        {29467}
    },
    {
        4094,
        {30450, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4094},
        {30450}
    },
    {
        4095,
        {32178, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4095},
        {32178}
    },
    {
        4096,
        {32791, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4096},
        {32791}
    },
    {
        4097,
        {33945, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4097},
        {33945}
    },
    {
        4098,
        {20786, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4098},
        {20786}
    },
    {
        4099,
        {26408, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4099},
        {26408}
    },
    {
        4100,
        {40665, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4100},
        {40665}
    },
    {
        4101,
        {30446, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4101},
        {30446}
    },
    {
        4102,
        {26466, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4102},
        {26466}
    },
    {
        4103,
        {21247, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4103},
        {21247}
    },
    {
        4104,
        {39173, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4104},
        {39173}
    },
    {
        4105,
        {23588, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4105},
        {23588}
    },
    {
        4106,
        {25147, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4106},
        {25147}
    },
    {
        4107,
        {31870, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4107},
        {31870}
    },
    {
        4108,
        {36016, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4108},
        {36016}
    },
    {
        4109,
        {21839, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4109},
        {21839}
    },
    {
        4110,
        {24758, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4110},
        {24758}
    },
    {
        4111,
        {32011, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4111},
        {32011}
    },
    {
        4112,
        {38272, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4112},
        {38272}
    },
    {
        4113,
        {21249, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4113},
        {21249}
    },
    {
        4114,
        {20063, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4114},
        {20063}
    },
    {
        4115,
        {20918, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4115},
        {20918}
    },
    {
        4116,
        {22812, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4116},
        {22812}
    },
    {
        4117,
        {29242, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4117},
        {29242}
    },
    {
        4118,
        {32822, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4118},
        {32822}
    },
    {
        4119,
        {37326, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4119},
        {37326}
    },
    {
        4120,
        {24357, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4120},
        {24357}
    },
    {
        4121,
        {30690, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4121},
        {30690}
    },
    {
        4122,
        {21380, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4122},
        {21380}
    },
    {
        4123,
        {24441, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4123},
        {24441}
    },
    {
        4124,
        {32004, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4124},
        {32004}
    },
    {
        4125,
        {34220, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4125},
        {34220}
    },
    {
        4126,
        {35379, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4126},
        {35379}
    },
    {
        4127,
        {36493, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4127},
        {36493}
    },
    {
        4128,
        {38742, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4128},
        {38742}
    },
    {
        4129,
        {26611, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4129},
        {26611}
    },
    {
        4130,
        {34222, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4130},
        {34222}
    },
    {
        4131,
        {37971, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4131},
        {37971}
    },
    {
        4132,
        {24841, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4132},
        {24841}
    },
    {
        4133,
        {24840, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4133},
        {24840}
    },
    {
        4134,
        {27833, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4134},
        {27833}
    },
    {
        4135,
        {30290, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4135},
        {30290}
    },
    {
        4136,
        {35565, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4136},
        {35565}
    },
    {
        4137,
        {36664, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4137},
        {36664}
    },
    {
        4138,
        {21807, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4138},
        {21807}
    },
    {
        4139,
        {20305, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4139},
        {20305}
    },
    {
        4140,
        {20778, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4140},
        {20778}
    },
    {
        4141,
        {21191, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4141},
        {21191}
    },
    {
        4142,
        {21451, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4142},
        {21451}
    },
    {
        4143,
        {23461, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4143},
        {23461}
    },
    {
        4144,
        {24189, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4144},
        {24189}
    },
    {
        4145,
        {24736, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4145},
        {24736}
    },
    {
        4146,
        {24962, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4146},
        {24962}
    },
    {
        4147,
        {25558, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4147},
        {25558}
    },
    {
        4148,
        {26377, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4148},
        {26377}
    },
    {
        4149,
        {26586, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4149},
        {26586}
    },
    {
        4150,
        {28263, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4150},
        {28263}
    },
    {
        4151,
        {28044, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4151},
        {28044}
    },
    {
        4152,
        {29494, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4152},
        {29494}
    },
    {
        4153,
        {29495, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4153},
        {29495}
    },
    {
        4154,
        {30001, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4154},
        {30001}
    },
    {
        4155,
        {31056, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4155},
        {31056}
    },
    {
        4156,
        {35029, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4156},
        {35029}
    },
    {
        4157,
        {35480, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4157},
        {35480}
    },
    {
        4158,
        {36938, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4158},
        {36938}
    },
    {
        4159,
        {37009, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4159},
        {37009}
    },
    {
        4160,
        {37109, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4160},
        {37109}
    },
    {
        4161,
        {38596, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4161},
        {38596}
    },
    {
        4162,
        {34701, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4162},
        {34701}
    },
    {
        4163,
        {22805, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4163},
        {22805}
    },
    {
        4164,
        {20104, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4164},
        {20104}
    },
    {
        4165,
        {20313, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4165},
        {20313}
    },
    {
        4166,
        {19982, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4166},
        {19982}
    },
    {
        4167,
        {35465, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4167},
        {35465}
    },
    {
        4168,
        {36671, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4168},
        {36671}
    },
    {
        4169,
        {38928, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4169},
        {38928}
    },
    {
        4170,
        {20653, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4170},
        {20653}
    },
    {
        4171,
        {24188, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4171},
        {24188}
    },
    {
        4172,
        {22934, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4172},
        {22934}
    },
    {
        4173,
        {23481, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4173},
        {23481}
    },
    {
        4174,
        {24248, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4174},
        {24248}
    },
    {
        4175,
        {25562, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4175},
        {25562}
    },
    {
        4176,
        {25594, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4176},
        {25594}
    },
    {
        4177,
        {25793, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4177},
        {25793}
    },
    {
        4178,
        {26332, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4178},
        {26332}
    },
    {
        4179,
        {26954, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4179},
        {26954}
    },
    {
        4180,
        {27096, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4180},
        {27096}
    },
    {
        4181,
        {27915, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4181},
        {27915}
    },
    {
        4182,
        {28342, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4182},
        {28342}
    },
    {
        4183,
        {29076, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4183},
        {29076}
    },
    {
        4184,
        {29992, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4184},
        {29992}
    },
    {
        4185,
        {31407, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4185},
        {31407}
    },
    {
        4186,
        {32650, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4186},
        {32650}
    },
    {
        4187,
        {32768, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4187},
        {32768}
    },
    {
        4188,
        {33865, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4188},
        {33865}
    },
    {
        4189,
        {33993, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4189},
        {33993}
    },
    {
        4190,
        {35201, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4190},
        {35201}
    },
    {
        4191,
        {35617, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4191},
        {35617}
    },
    {
        4192,
        {36362, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4192},
        {36362}
    },
    {
        4193,
        {36965, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4193},
        {36965}
    },
    {
        4194,
        {38525, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4194},
        {38525}
    },
    {
        4195,
        {39178, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4195},
        {39178}
    },
    {
        4196,
        {24958, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4196},
        {24958}
    },
    {
        4197,
        {25233, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4197},
        {25233}
    },
    {
        4198,
        {27442, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4198},
        {27442}
    },
    {
        4199,
        {27779, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4199},
        {27779}
    },
    {
        4200,
        {28020, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4200},
        {28020}
    },
    {
        4201,
        {32716, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4201},
        {32716}
    },
    {
        4202,
        {32764, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4202},
        {32764}
    },
    {
        4203,
        {28096, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4203},
        {28096}
    },
    {
        4204,
        {32645, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4204},
        {32645}
    },
    {
        4205,
        {34746, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4205},
        {34746}
    },
    {
        4206,
        {35064, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4206},
        {35064}
    },
    {
        4207,
        {26469, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4207},
        {26469}
    },
    {
        4208,
        {33713, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4208},
        {33713}
    },
    {
        4209,
        {38972, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4209},
        {38972}
    },
    {
        4210,
        {38647, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4210},
        {38647}
    },
    {
        4211,
        {27931, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4211},
        {27931}
    },
    {
        4212,
        {32097, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4212},
        {32097}
    },
    {
        4213,
        {33853, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4213},
        {33853}
    },
    {
        4214,
        {37226, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4214},
        {37226}
    },
    {
        4215,
        {20081, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4215},
        {20081}
    },
    {
        4216,
        {21365, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4216},
        {21365}
    },
    {
        4217,
        {23888, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4217},
        {23888}
    },
    {
        4218,
        {27396, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4218},
        {27396}
    },
    {
        4219,
        {28651, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4219},
        {28651}
    },
    {
        4220,
        {34253, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4220},
        {34253}
    },
    {
        4221,
        {34349, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4221},
        {34349}
    },
    {
        4222,
        {35239, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4222},
        {35239}
    },
    {
        4223,
        {21033, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4223},
        {21033}
    },
    {
        4224,
        {21519, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4224},
        {21519}
    },
    {
        4225,
        {23653, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4225},
        {23653}
    },
    {
        4226,
        {26446, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4226},
        {26446}
    },
    {
        4227,
        {26792, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4227},
        {26792}
    },
    {
        4228,
        {29702, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4228},
        {29702}
    },
    {
        4229,
        {29827, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4229},
        {29827}
    },
    {
        4230,
        {30178, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4230},
        {30178}
    },
    {
        4231,
        {35023, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4231},
        {35023}
    },
    {
        4232,
        {35041, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4232},
        {35041}
    },
    {
        4233,
        {37324, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4233},
        {37324}
    },
    {
        4234,
        {38626, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4234},
        {38626}
    },
    {
        4235,
        {38520, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4235},
        {38520}
    },
    {
        4236,
        {24459, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4236},
        {24459}
    },
    {
        4237,
        {29575, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4237},
        {29575}
    },
    {
        4238,
        {31435, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4238},
        {31435}
    },
    {
        4239,
        {33870, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4239},
        {33870}
    },
    {
        4240,
        {25504, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4240},
        {25504}
    },
    {
        4241,
        {30053, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4241},
        {30053}
    },
    {
        4242,
        {21129, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4242},
        {21129}
    },
    {
        4243,
        {27969, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4243},
        {27969}
    },
    {
        4244,
        {28316, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4244},
        {28316}
    },
    {
        4245,
        {29705, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4245},
        {29705}
    },
    {
        4246,
        {30041, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4246},
        {30041}
    },
    {
        4247,
        {30827, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4247},
        {30827}
    },
    {
        4248,
        {31890, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4248},
        {31890}
    },
    {
        4249,
        {38534, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4249},
        {38534}
    },
    {
        4250,
        {31452, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4250},
        {31452}
    },
    {
        4251,
        {40845, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4251},
        {40845}
    },
    {
        4252,
        {20406, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4252},
        {20406}
    },
    {
        4253,
        {24942, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4253},
        {24942}
    },
    {
        4254,
        {26053, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4254},
        {26053}
    },
    {
        4255,
        {34396, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4255},
        {34396}
    },
    {
        4256,
        {20102, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4256},
        {20102}
    },
    {
        4257,
        {20142, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4257},
        {20142}
    },
    {
        4258,
        {20698, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4258},
        {20698}
    },
    {
        4259,
        {20001, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4259},
        {20001}
    },
    {
        4260,
        {20940, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4260},
        {20940}
    },
    {
        4261,
        {23534, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4261},
        {23534}
    },
    {
        4262,
        {26009, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4262},
        {26009}
    },
    {
        4263,
        {26753, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4263},
        {26753}
    },
    {
        4264,
        {28092, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4264},
        {28092}
    },
    {
        4265,
        {29471, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4265},
        {29471}
    },
    {
        4266,
        {30274, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4266},
        {30274}
    },
    {
        4267,
        {30637, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4267},
        {30637}
    },
    {
        4268,
        {31260, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4268},
        {31260}
    },
    {
        4269,
        {31975, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4269},
        {31975}
    },
    {
        4270,
        {33391, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4270},
        {33391}
    },
    {
        4271,
        {35538, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4271},
        {35538}
    },
    {
        4272,
        {36988, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4272},
        {36988}
    },
    {
        4273,
        {37327, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4273},
        {37327}
    },
    {
        4274,
        {38517, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4274},
        {38517}
    },
    {
        4275,
        {38936, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4275},
        {38936}
    },
    {
        4276,
        {21147, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4276},
        {21147}
    },
    {
        4277,
        {32209, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4277},
        {32209}
    },
    {
        4278,
        {20523, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4278},
        {20523}
    },
    {
        4279,
        {21400, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4279},
        {21400}
    },
    {
        4280,
        {26519, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4280},
        {26519}
    },
    {
        4281,
        {28107, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4281},
        {28107}
    },
    {
        4282,
        {29136, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4282},
        {29136}
    },
    {
        4283,
        {29747, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4283},
        {29747}
    },
    {
        4284,
        {33256, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4284},
        {33256}
    },
    {
        4285,
        {36650, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4285},
        {36650}
    },
    {
        4286,
        {38563, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4286},
        {38563}
    },
    {
        4287,
        {40023, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4287},
        {40023}
    },
    {
        4288,
        {40607, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4288},
        {40607}
    },
    {
        4289,
        {29792, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4289},
        {29792}
    },
    {
        4290,
        {22593, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4290},
        {22593}
    },
    {
        4291,
        {28057, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4291},
        {28057}
    },
    {
        4292,
        {32047, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4292},
        {32047}
    },
    {
        4293,
        {39006, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4293},
        {39006}
    },
    {
        4294,
        {20196, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4294},
        {20196}
    },
    {
        4295,
        {20278, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4295},
        {20278}
    },
    {
        4296,
        {20363, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4296},
        {20363}
    },
    {
        4297,
        {20919, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4297},
        {20919}
    },
    {
        4298,
        {21169, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4298},
        {21169}
    },
    {
        4299,
        {23994, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4299},
        {23994}
    },
    {
        4300,
        {24604, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4300},
        {24604}
    },
    {
        4301,
        {29618, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4301},
        {29618}
    },
    {
        4302,
        {31036, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4302},
        {31036}
    },
    {
        4303,
        {33491, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4303},
        {33491}
    },
    {
        4304,
        {37428, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4304},
        {37428}
    },
    {
        4305,
        {38583, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4305},
        {38583}
    },
    {
        4306,
        {38646, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4306},
        {38646}
    },
    {
        4307,
        {38666, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4307},
        {38666}
    },
    {
        4308,
        {40599, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4308},
        {40599}
    },
    {
        4309,
        {40802, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4309},
        {40802}
    },
    {
        4310,
        {26278, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4310},
        {26278}
    },
    {
        4311,
        {27508, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4311},
        {27508}
    },
    {
        4312,
        {21015, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4312},
        {21015}
    },
    {
        4313,
        {21155, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4313},
        {21155}
    },
    {
        4314,
        {28872, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4314},
        {28872}
    },
    {
        4315,
        {35010, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4315},
        {35010}
    },
    {
        4316,
        {24265, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4316},
        {24265}
    },
    {
        4317,
        {24651, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4317},
        {24651}
    },
    {
        4318,
        {24976, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4318},
        {24976}
    },
    {
        4319,
        {28451, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4319},
        {28451}
    },
    {
        4320,
        {29001, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4320},
        {29001}
    },
    {
        4321,
        {31806, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4321},
        {31806}
    },
    {
        4322,
        {32244, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4322},
        {32244}
    },
    {
        4323,
        {32879, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4323},
        {32879}
    },
    {
        4324,
        {34030, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4324},
        {34030}
    },
    {
        4325,
        {36899, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4325},
        {36899}
    },
    {
        4326,
        {37676, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4326},
        {37676}
    },
    {
        4327,
        {21570, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4327},
        {21570}
    },
    {
        4328,
        {39791, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4328},
        {39791}
    },
    {
        4329,
        {27347, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4329},
        {27347}
    },
    {
        4330,
        {28809, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4330},
        {28809}
    },
    {
        4331,
        {36034, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4331},
        {36034}
    },
    {
        4332,
        {36335, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4332},
        {36335}
    },
    {
        4333,
        {38706, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4333},
        {38706}
    },
    {
        4334,
        {21172, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4334},
        {21172}
    },
    {
        4335,
        {23105, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4335},
        {23105}
    },
    {
        4336,
        {24266, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4336},
        {24266}
    },
    {
        4337,
        {24324, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4337},
        {24324}
    },
    {
        4338,
        {26391, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4338},
        {26391}
    },
    {
        4339,
        {27004, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4339},
        {27004}
    },
    {
        4340,
        {27028, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4340},
        {27028}
    },
    {
        4341,
        {28010, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4341},
        {28010}
    },
    {
        4342,
        {28431, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4342},
        {28431}
    },
    {
        4343,
        {29282, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4343},
        {29282}
    },
    {
        4344,
        {29436, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4344},
        {29436}
    },
    {
        4345,
        {31725, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4345},
        {31725}
    },
    {
        4346,
        {32769, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4346},
        {32769}
    },
    {
        4347,
        {32894, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4347},
        {32894}
    },
    {
        4348,
        {34635, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4348},
        {34635}
    },
    {
        4349,
        {37070, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4349},
        {37070}
    },
    {
        4350,
        {20845, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4350},
        {20845}
    },
    {
        4351,
        {40595, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4351},
        {40595}
    },
    {
        4352,
        {31108, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4352},
        {31108}
    },
    {
        4353,
        {32907, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4353},
        {32907}
    },
    {
        4354,
        {37682, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4354},
        {37682}
    },
    {
        4355,
        {35542, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4355},
        {35542}
    },
    {
        4356,
        {20525, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4356},
        {20525}
    },
    {
        4357,
        {21644, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4357},
        {21644}
    },
    {
        4358,
        {35441, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4358},
        {35441}
    },
    {
        4359,
        {27498, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4359},
        {27498}
    },
    {
        4360,
        {36036, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4360},
        {36036}
    },
    {
        4361,
        {33031, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4361},
        {33031}
    },
    {
        4362,
        {24785, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4362},
        {24785}
    },
    {
        4363,
        {26528, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4363},
        {26528}
    },
    {
        4364,
        {40434, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4364},
        {40434}
    },
    {
        4365,
        {20121, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4365},
        {20121}
    },
    {
        4366,
        {20120, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4366},
        {20120}
    },
    {
        4367,
        {39952, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4367},
        {39952}
    },
    {
        4368,
        {35435, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4368},
        {35435}
    },
    {
        4369,
        {34241, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4369},
        {34241}
    },
    {
        4370,
        {34152, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4370},
        {34152}
    },
    {
        4371,
        {26880, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4371},
        {26880}
    },
    {
        4372,
        {28286, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4372},
        {28286}
    },
    {
        4373,
        {30871, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4373},
        {30871}
    },
    {
        4374,
        {33109, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_1,
        1,
        {4374},
        {33109}
    },
    {
        4375,
        {134047, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2561},
        {21489}
    },
    {
        4376,
        {23361, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        4377,
        {23382, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        4378,
        {23421, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        4379,
        {23443, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        4380,
        {23512, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        4381,
        {23532, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1803},
        {23515}
    },
    {
        4382,
        {23570, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4972},
        {23571}
    },
    {
        4383,
        {23582, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {5916},
        {29134}
    },
    {
        4384,
        {23587, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4974},
        {23586}
    },
    {
        4385,
        {23595, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        4386,
        {14221, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2831},
        {23611}
    },
    {
        4387,
        {23650, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2577},
        {23649}
    },
    {
        4388,
        {64059, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3063},
        {23652}
    },
    {
        4389,
        {64060, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4986},
        {23662}
    },
    {
        4390,
        {136884, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4993},
        {22939}
    },
    {
        4391,
        {23674, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        4392,
        {23695, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        4393,
        {23711, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        4394,
        {23715, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        4395,
        {23722, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        4396,
        {23738, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        4397,
        {23755, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        4398,
        {23760, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        4399,
        {23762, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        4400,
        {23796, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        4401,
        {138804, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {5034},
        {23982}
    },
    {
        4402,
        {14306, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {5012},
        {23900}
    },
    {
        4403,
        {23821, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        4404,
        {23847, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2902},
        {23913}
    },
    {
        4405,
        {64017, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2423},
        {23822}
    },
    {
        4406,
        {23878, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        4407,
        {23879, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        4408,
        {23891, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {5022},
        {23890}
    },
    {
        4409,
        {23882, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        4410,
        {23917, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        4411,
        {23937, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        4412,
        {23968, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        4413,
        {23972, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {5031},
        {23970}
    },
    {
        4414,
        {23975, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        4415,
        {23992, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        4416,
        {24011, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        4417,
        {21534, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3538},
        {21585}
    },
    {
        4418,
        {24332, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4418},
        {24332}
    },
    {
        4419,
        {19984, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4419},
        {19984}
    },
    {
        4420,
        {19989, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4420},
        {19989}
    },
    {
        4421,
        {20010, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4421},
        {20010}
    },
    {
        4422,
        {20017, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4422},
        {20017}
    },
    {
        4423,
        {20022, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4423},
        {20022}
    },
    {
        4424,
        {20028, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4424},
        {20028}
    },
    {
        4425,
        {20031, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4425},
        {20031}
    },
    {
        4426,
        {20034, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4426},
        {20034}
    },
    {
        4427,
        {20054, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4427},
        {20054}
    },
    {
        4428,
        {20056, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4428},
        {20056}
    },
    {
        4429,
        {20098, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4429},
        {20098}
    },
    {
        4430,
        {20101, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4430},
        {20101}
    },
    {
        4431,
        {35947, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4431},
        {35947}
    },
    {
        4432,
        {20106, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4432},
        {20106}
    },
    {
        4433,
        {33298, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4433},
        {33298}
    },
    {
        4434,
        {24333, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4434},
        {24333}
    },
    {
        4435,
        {20110, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4435},
        {20110}
    },
    {
        4436,
        {20126, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4436},
        {20126}
    },
    {
        4437,
        {20127, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4437},
        {20127}
    },
    {
        4438,
        {20128, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4438},
        {20128}
    },
    {
        4439,
        {20130, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4439},
        {20130}
    },
    {
        4440,
        {20144, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4440},
        {20144}
    },
    {
        4441,
        {20147, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4441},
        {20147}
    },
    {
        4442,
        {20150, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4442},
        {20150}
    },
    {
        4443,
        {20174, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4443},
        {20174}
    },
    {
        4444,
        {20173, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4444},
        {20173}
    },
    {
        4445,
        {20164, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4445},
        {20164}
    },
    {
        4446,
        {20166, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4446},
        {20166}
    },
    {
        4447,
        {20162, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4447},
        {20162}
    },
    {
        4448,
        {20183, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4448},
        {20183}
    },
    {
        4449,
        {20190, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4449},
        {20190}
    },
    {
        4450,
        {20205, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4450},
        {20205}
    },
    {
        4451,
        {20191, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4451},
        {20191}
    },
    {
        4452,
        {20215, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4452},
        {20215}
    },
    {
        4453,
        {20233, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4453},
        {20233}
    },
    {
        4454,
        {20314, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4454},
        {20314}
    },
    {
        4455,
        {20272, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4455},
        {20272}
    },
    {
        4456,
        {20315, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4456},
        {20315}
    },
    {
        4457,
        {20317, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4457},
        {20317}
    },
    {
        4458,
        {20311, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4458},
        {20311}
    },
    {
        4459,
        {20295, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4459},
        {20295}
    },
    {
        4460,
        {20342, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4460},
        {20342}
    },
    {
        4461,
        {20360, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4461},
        {20360}
    },
    {
        4462,
        {20367, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4462},
        {20367}
    },
    {
        4463,
        {20376, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4463},
        {20376}
    },
    {
        4464,
        {20347, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4464},
        {20347}
    },
    {
        4465,
        {20329, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4465},
        {20329}
    },
    {
        4466,
        {20336, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4466},
        {20336}
    },
    {
        4467,
        {20369, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4467},
        {20369}
    },
    {
        4468,
        {20335, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4468},
        {20335}
    },
    {
        4469,
        {20358, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4469},
        {20358}
    },
    {
        4470,
        {20374, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4470},
        {20374}
    },
    {
        4471,
        {20760, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4471},
        {20760}
    },
    {
        4472,
        {20436, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4472},
        {20436}
    },
    {
        4473,
        {20447, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4473},
        {20447}
    },
    {
        4474,
        {20430, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4474},
        {20430}
    },
    {
        4475,
        {20440, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4475},
        {20440}
    },
    {
        4476,
        {20443, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4476},
        {20443}
    },
    {
        4477,
        {20433, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4477},
        {20433}
    },
    {
        4478,
        {20442, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4478},
        {20442}
    },
    {
        4479,
        {20432, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4479},
        {20432}
    },
    {
        4480,
        {20452, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4480},
        {20452}
    },
    {
        4481,
        {20453, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4481},
        {20453}
    },
    {
        4482,
        {20506, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4482},
        {20506}
    },
    {
        4483,
        {20520, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4483},
        {20520}
    },
    {
        4484,
        {20500, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4484},
        {20500}
    },
    {
        4485,
        {20522, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4485},
        {20522}
    },
    {
        4486,
        {20517, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4486},
        {20517}
    },
    {
        4487,
        {20485, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4487},
        {20485}
    },
    {
        4488,
        {20252, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4488},
        {20252}
    },
    {
        4489,
        {20470, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4489},
        {20470}
    },
    {
        4490,
        {20513, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4490},
        {20513}
    },
    {
        4491,
        {20521, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4491},
        {20521}
    },
    {
        4492,
        {20524, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4492},
        {20524}
    },
    {
        4493,
        {20478, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4493},
        {20478}
    },
    {
        4494,
        {20463, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4494},
        {20463}
    },
    {
        4495,
        {20497, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4495},
        {20497}
    },
    {
        4496,
        {20486, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4496},
        {20486}
    },
    {
        4497,
        {20547, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4497},
        {20547}
    },
    {
        4498,
        {20551, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4498},
        {20551}
    },
    {
        4499,
        {26371, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4499},
        {26371}
    },
    {
        4500,
        {20565, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4500},
        {20565}
    },
    {
        4501,
        {20560, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4501},
        {20560}
    },
    {
        4502,
        {20552, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4502},
        {20552}
    },
    {
        4503,
        {20570, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4503},
        {20570}
    },
    {
        4504,
        {20566, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4504},
        {20566}
    },
    {
        4505,
        {20588, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4505},
        {20588}
    },
    {
        4506,
        {20600, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4506},
        {20600}
    },
    {
        4507,
        {20608, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4507},
        {20608}
    },
    {
        4508,
        {20634, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4508},
        {20634}
    },
    {
        4509,
        {20613, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4509},
        {20613}
    },
    {
        4510,
        {20660, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4510},
        {20660}
    },
    {
        4511,
        {20658, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4511},
        {20658}
    },
    {
        4512,
        {20681, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4512},
        {20681}
    },
    {
        4513,
        {20682, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4513},
        {20682}
    },
    {
        4514,
        {20659, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4514},
        {20659}
    },
    {
        4515,
        {20674, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4515},
        {20674}
    },
    {
        4516,
        {20694, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4516},
        {20694}
    },
    {
        4517,
        {20702, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4517},
        {20702}
    },
    {
        4518,
        {20709, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4518},
        {20709}
    },
    {
        4519,
        {20717, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4519},
        {20717}
    },
    {
        4520,
        {20707, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4520},
        {20707}
    },
    {
        4521,
        {20718, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4521},
        {20718}
    },
    {
        4522,
        {20729, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4522},
        {20729}
    },
    {
        4523,
        {20725, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4523},
        {20725}
    },
    {
        4524,
        {20745, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4524},
        {20745}
    },
    {
        4525,
        {20737, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4525},
        {20737}
    },
    {
        4526,
        {20738, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4526},
        {20738}
    },
    {
        4527,
        {20758, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4527},
        {20758}
    },
    {
        4528,
        {20757, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4528},
        {20757}
    },
    {
        4529,
        {20756, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4529},
        {20756}
    },
    {
        4530,
        {20762, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4530},
        {20762}
    },
    {
        4531,
        {20769, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4531},
        {20769}
    },
    {
        4532,
        {20794, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4532},
        {20794}
    },
    {
        4533,
        {20791, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4533},
        {20791}
    },
    {
        4534,
        {20796, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4534},
        {20796}
    },
    {
        4535,
        {20795, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4535},
        {20795}
    },
    {
        4536,
        {20799, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4536},
        {20799}
    },
    {
        4537,
        {20800, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4537},
        {20800}
    },
    {
        4538,
        {20818, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4538},
        {20818}
    },
    {
        4539,
        {20812, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4539},
        {20812}
    },
    {
        4540,
        {20820, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4540},
        {20820}
    },
    {
        4541,
        {20834, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4541},
        {20834}
    },
    {
        4542,
        {31480, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4542},
        {31480}
    },
    {
        4543,
        {20841, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4543},
        {20841}
    },
    {
        4544,
        {20842, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4544},
        {20842}
    },
    {
        4545,
        {20846, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4545},
        {20846}
    },
    {
        4546,
        {20864, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4546},
        {20864}
    },
    {
        4547,
        {20866, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4547},
        {20866}
    },
    {
        4548,
        {22232, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4548},
        {22232}
    },
    {
        4549,
        {20876, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4549},
        {20876}
    },
    {
        4550,
        {20873, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4550},
        {20873}
    },
    {
        4551,
        {20879, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4551},
        {20879}
    },
    {
        4552,
        {20881, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4552},
        {20881}
    },
    {
        4553,
        {20883, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4553},
        {20883}
    },
    {
        4554,
        {20885, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4554},
        {20885}
    },
    {
        4555,
        {20886, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4555},
        {20886}
    },
    {
        4556,
        {20900, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4556},
        {20900}
    },
    {
        4557,
        {20902, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4557},
        {20902}
    },
    {
        4558,
        {20898, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4558},
        {20898}
    },
    {
        4559,
        {20905, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4559},
        {20905}
    },
    {
        4560,
        {20906, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4560},
        {20906}
    },
    {
        4561,
        {20907, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4561},
        {20907}
    },
    {
        4562,
        {20915, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4562},
        {20915}
    },
    {
        4563,
        {20913, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4563},
        {20913}
    },
    {
        4564,
        {20914, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4564},
        {20914}
    },
    {
        4565,
        {20912, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4565},
        {20912}
    },
    {
        4566,
        {20917, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4566},
        {20917}
    },
    {
        4567,
        {20925, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4567},
        {20925}
    },
    {
        4568,
        {20933, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4568},
        {20933}
    },
    {
        4569,
        {20937, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4569},
        {20937}
    },
    {
        4570,
        {20955, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4570},
        {20955}
    },
    {
        4571,
        {20960, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4571},
        {20960}
    },
    {
        4572,
        {34389, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4572},
        {34389}
    },
    {
        4573,
        {20969, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4573},
        {20969}
    },
    {
        4574,
        {20973, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4574},
        {20973}
    },
    {
        4575,
        {20976, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4575},
        {20976}
    },
    {
        4576,
        {20981, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4576},
        {20981}
    },
    {
        4577,
        {20990, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4577},
        {20990}
    },
    {
        4578,
        {20996, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4578},
        {20996}
    },
    {
        4579,
        {21003, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4579},
        {21003}
    },
    {
        4580,
        {21012, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4580},
        {21012}
    },
    {
        4581,
        {21006, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4581},
        {21006}
    },
    {
        4582,
        {21031, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4582},
        {21031}
    },
    {
        4583,
        {21034, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4583},
        {21034}
    },
    {
        4584,
        {21038, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4584},
        {21038}
    },
    {
        4585,
        {21043, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4585},
        {21043}
    },
    {
        4586,
        {21049, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4586},
        {21049}
    },
    {
        4587,
        {21071, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4587},
        {21071}
    },
    {
        4588,
        {21060, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4588},
        {21060}
    },
    {
        4589,
        {21067, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4589},
        {21067}
    },
    {
        4590,
        {21068, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4590},
        {21068}
    },
    {
        4591,
        {21086, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4591},
        {21086}
    },
    {
        4592,
        {21076, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4592},
        {21076}
    },
    {
        4593,
        {21098, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4593},
        {21098}
    },
    {
        4594,
        {21108, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4594},
        {21108}
    },
    {
        4595,
        {21097, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4595},
        {21097}
    },
    {
        4596,
        {21107, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4596},
        {21107}
    },
    {
        4597,
        {21119, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4597},
        {21119}
    },
    {
        4598,
        {21117, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4598},
        {21117}
    },
    {
        4599,
        {21133, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4599},
        {21133}
    },
    {
        4600,
        {21140, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4600},
        {21140}
    },
    {
        4601,
        {21138, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4601},
        {21138}
    },
    {
        4602,
        {21105, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4602},
        {21105}
    },
    {
        4603,
        {21128, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4603},
        {21128}
    },
    {
        4604,
        {21137, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4604},
        {21137}
    },
    {
        4605,
        {36776, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4605},
        {36776}
    },
    {
        4606,
        {36775, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4606},
        {36775}
    },
    {
        4607,
        {21164, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4607},
        {21164}
    },
    {
        4608,
        {21165, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4608},
        {21165}
    },
    {
        4609,
        {21180, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4609},
        {21180}
    },
    {
        4610,
        {21173, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4610},
        {21173}
    },
    {
        4611,
        {21185, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4611},
        {21185}
    },
    {
        4612,
        {21197, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4612},
        {21197}
    },
    {
        4613,
        {21207, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4613},
        {21207}
    },
    {
        4614,
        {21214, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4614},
        {21214}
    },
    {
        4615,
        {21219, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4615},
        {21219}
    },
    {
        4616,
        {21222, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4616},
        {21222}
    },
    {
        4617,
        {39149, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4617},
        {39149}
    },
    {
        4618,
        {21216, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4618},
        {21216}
    },
    {
        4619,
        {21235, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4619},
        {21235}
    },
    {
        4620,
        {21237, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4620},
        {21237}
    },
    {
        4621,
        {21240, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4621},
        {21240}
    },
    {
        4622,
        {21241, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4622},
        {21241}
    },
    {
        4623,
        {21254, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4623},
        {21254}
    },
    {
        4624,
        {21256, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4624},
        {21256}
    },
    {
        4625,
        {30008, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4625},
        {30008}
    },
    {
        4626,
        {21261, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4626},
        {21261}
    },
    {
        4627,
        {21264, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4627},
        {21264}
    },
    {
        4628,
        {21263, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4628},
        {21263}
    },
    {
        4629,
        {21269, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4629},
        {21269}
    },
    {
        4630,
        {21274, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4630},
        {21274}
    },
    {
        4631,
        {21283, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4631},
        {21283}
    },
    {
        4632,
        {21295, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4632},
        {21295}
    },
    {
        4633,
        {21297, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4633},
        {21297}
    },
    {
        4634,
        {21299, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4634},
        {21299}
    },
    {
        4635,
        {21304, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4635},
        {21304}
    },
    {
        4636,
        {21312, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4636},
        {21312}
    },
    {
        4637,
        {21318, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4637},
        {21318}
    },
    {
        4638,
        {21317, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4638},
        {21317}
    },
    {
        4639,
        {19991, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4639},
        {19991}
    },
    {
        4640,
        {21321, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4640},
        {21321}
    },
    {
        4641,
        {21325, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4641},
        {21325}
    },
    {
        4642,
        {20950, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4642},
        {20950}
    },
    {
        4643,
        {21342, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4643},
        {21342}
    },
    {
        4644,
        {21353, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4644},
        {21353}
    },
    {
        4645,
        {21358, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4645},
        {21358}
    },
    {
        4646,
        {22808, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4646},
        {22808}
    },
    {
        4647,
        {21371, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4647},
        {21371}
    },
    {
        4648,
        {21367, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4648},
        {21367}
    },
    {
        4649,
        {21378, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4649},
        {21378}
    },
    {
        4650,
        {21398, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4650},
        {21398}
    },
    {
        4651,
        {21408, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4651},
        {21408}
    },
    {
        4652,
        {21414, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4652},
        {21414}
    },
    {
        4653,
        {21413, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4653},
        {21413}
    },
    {
        4654,
        {21422, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4654},
        {21422}
    },
    {
        4655,
        {21424, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4655},
        {21424}
    },
    {
        4656,
        {21430, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4656},
        {21430}
    },
    {
        4657,
        {21443, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4657},
        {21443}
    },
    {
        4658,
        {31762, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4658},
        {31762}
    },
    {
        4659,
        {38617, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4659},
        {38617}
    },
    {
        4660,
        {21471, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4660},
        {21471}
    },
    {
        4661,
        {26364, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4661},
        {26364}
    },
    {
        4662,
        {29166, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4662},
        {29166}
    },
    {
        4663,
        {21486, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4663},
        {21486}
    },
    {
        4664,
        {21480, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4664},
        {21480}
    },
    {
        4665,
        {21485, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4665},
        {21485}
    },
    {
        4666,
        {21498, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4666},
        {21498}
    },
    {
        4667,
        {21505, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4667},
        {21505}
    },
    {
        4668,
        {21565, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4668},
        {21565}
    },
    {
        4669,
        {21568, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4669},
        {21568}
    },
    {
        4670,
        {21548, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4670},
        {21548}
    },
    {
        4671,
        {21549, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4671},
        {21549}
    },
    {
        4672,
        {21564, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4672},
        {21564}
    },
    {
        4673,
        {21550, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4673},
        {21550}
    },
    {
        4674,
        {21558, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4674},
        {21558}
    },
    {
        4675,
        {21545, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4675},
        {21545}
    },
    {
        4676,
        {21533, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4676},
        {21533}
    },
    {
        4677,
        {21582, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4677},
        {21582}
    },
    {
        4678,
        {21647, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4678},
        {21647}
    },
    {
        4679,
        {21621, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4679},
        {21621}
    },
    {
        4680,
        {21646, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4680},
        {21646}
    },
    {
        4681,
        {21599, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4681},
        {21599}
    },
    {
        4682,
        {21617, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4682},
        {21617}
    },
    {
        4683,
        {21623, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4683},
        {21623}
    },
    {
        4684,
        {21616, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4684},
        {21616}
    },
    {
        4685,
        {21650, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4685},
        {21650}
    },
    {
        4686,
        {21627, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4686},
        {21627}
    },
    {
        4687,
        {21632, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4687},
        {21632}
    },
    {
        4688,
        {21622, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4688},
        {21622}
    },
    {
        4689,
        {21636, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4689},
        {21636}
    },
    {
        4690,
        {21648, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4690},
        {21648}
    },
    {
        4691,
        {21638, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4691},
        {21638}
    },
    {
        4692,
        {21703, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4692},
        {21703}
    },
    {
        4693,
        {21666, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4693},
        {21666}
    },
    {
        4694,
        {21688, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4694},
        {21688}
    },
    {
        4695,
        {21669, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4695},
        {21669}
    },
    {
        4696,
        {21676, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4696},
        {21676}
    },
    {
        4697,
        {21700, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4697},
        {21700}
    },
    {
        4698,
        {21704, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4698},
        {21704}
    },
    {
        4699,
        {21672, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4699},
        {21672}
    },
    {
        4700,
        {21675, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4700},
        {21675}
    },
    {
        4701,
        {21698, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4701},
        {21698}
    },
    {
        4702,
        {21668, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4702},
        {21668}
    },
    {
        4703,
        {21694, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4703},
        {21694}
    },
    {
        4704,
        {21692, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4704},
        {21692}
    },
    {
        4705,
        {21720, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4705},
        {21720}
    },
    {
        4706,
        {21733, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4706},
        {21733}
    },
    {
        4707,
        {21734, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4707},
        {21734}
    },
    {
        4708,
        {21775, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4708},
        {21775}
    },
    {
        4709,
        {21780, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4709},
        {21780}
    },
    {
        4710,
        {21757, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4710},
        {21757}
    },
    {
        4711,
        {21742, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4711},
        {21742}
    },
    {
        4712,
        {21741, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4712},
        {21741}
    },
    {
        4713,
        {21754, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4713},
        {21754}
    },
    {
        4714,
        {21730, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4714},
        {21730}
    },
    {
        4715,
        {21817, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4715},
        {21817}
    },
    {
        4716,
        {21824, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4716},
        {21824}
    },
    {
        4717,
        {21859, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4717},
        {21859}
    },
    {
        4718,
        {21836, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4718},
        {21836}
    },
    {
        4719,
        {21806, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4719},
        {21806}
    },
    {
        4720,
        {21852, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4720},
        {21852}
    },
    {
        4721,
        {21829, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4721},
        {21829}
    },
    {
        4722,
        {21846, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4722},
        {21846}
    },
    {
        4723,
        {21847, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4723},
        {21847}
    },
    {
        4724,
        {21816, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4724},
        {21816}
    },
    {
        4725,
        {21811, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4725},
        {21811}
    },
    {
        4726,
        {21853, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4726},
        {21853}
    },
    {
        4727,
        {21913, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4727},
        {21913}
    },
    {
        4728,
        {21888, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4728},
        {21888}
    },
    {
        4729,
        {21679, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4729},
        {21679}
    },
    {
        4730,
        {21898, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4730},
        {21898}
    },
    {
        4731,
        {21919, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4731},
        {21919}
    },
    {
        4732,
        {21883, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4732},
        {21883}
    },
    {
        4733,
        {21886, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4733},
        {21886}
    },
    {
        4734,
        {21912, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4734},
        {21912}
    },
    {
        4735,
        {21918, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4735},
        {21918}
    },
    {
        4736,
        {21934, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4736},
        {21934}
    },
    {
        4737,
        {21884, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4737},
        {21884}
    },
    {
        4738,
        {21891, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4738},
        {21891}
    },
    {
        4739,
        {21929, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4739},
        {21929}
    },
    {
        4740,
        {21895, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4740},
        {21895}
    },
    {
        4741,
        {21928, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4741},
        {21928}
    },
    {
        4742,
        {21978, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4742},
        {21978}
    },
    {
        4743,
        {21957, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4743},
        {21957}
    },
    {
        4744,
        {21983, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4744},
        {21983}
    },
    {
        4745,
        {21956, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4745},
        {21956}
    },
    {
        4746,
        {21980, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4746},
        {21980}
    },
    {
        4747,
        {21988, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4747},
        {21988}
    },
    {
        4748,
        {21972, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4748},
        {21972}
    },
    {
        4749,
        {22036, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4749},
        {22036}
    },
    {
        4750,
        {22007, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4750},
        {22007}
    },
    {
        4751,
        {22038, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4751},
        {22038}
    },
    {
        4752,
        {22014, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4752},
        {22014}
    },
    {
        4753,
        {22013, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4753},
        {22013}
    },
    {
        4754,
        {22043, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4754},
        {22043}
    },
    {
        4755,
        {22009, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4755},
        {22009}
    },
    {
        4756,
        {22094, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4756},
        {22094}
    },
    {
        4757,
        {22096, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4757},
        {22096}
    },
    {
        4758,
        {29151, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4758},
        {29151}
    },
    {
        4759,
        {22068, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4759},
        {22068}
    },
    {
        4760,
        {22070, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4760},
        {22070}
    },
    {
        4761,
        {22066, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4761},
        {22066}
    },
    {
        4762,
        {22072, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4762},
        {22072}
    },
    {
        4763,
        {22123, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4763},
        {22123}
    },
    {
        4764,
        {22116, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4764},
        {22116}
    },
    {
        4765,
        {22063, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4765},
        {22063}
    },
    {
        4766,
        {22124, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4766},
        {22124}
    },
    {
        4767,
        {22122, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4767},
        {22122}
    },
    {
        4768,
        {22150, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4768},
        {22150}
    },
    {
        4769,
        {22144, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4769},
        {22144}
    },
    {
        4770,
        {22154, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4770},
        {22154}
    },
    {
        4771,
        {22176, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4771},
        {22176}
    },
    {
        4772,
        {22164, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4772},
        {22164}
    },
    {
        4773,
        {22159, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4773},
        {22159}
    },
    {
        4774,
        {22181, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4774},
        {22181}
    },
    {
        4775,
        {22190, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4775},
        {22190}
    },
    {
        4776,
        {22198, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4776},
        {22198}
    },
    {
        4777,
        {22196, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4777},
        {22196}
    },
    {
        4778,
        {22210, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4778},
        {22210}
    },
    {
        4779,
        {22204, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4779},
        {22204}
    },
    {
        4780,
        {22209, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4780},
        {22209}
    },
    {
        4781,
        {22211, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4781},
        {22211}
    },
    {
        4782,
        {22208, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4782},
        {22208}
    },
    {
        4783,
        {22216, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4783},
        {22216}
    },
    {
        4784,
        {22222, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4784},
        {22222}
    },
    {
        4785,
        {22225, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4785},
        {22225}
    },
    {
        4786,
        {22227, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4786},
        {22227}
    },
    {
        4787,
        {22231, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4787},
        {22231}
    },
    {
        4788,
        {22254, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4788},
        {22254}
    },
    {
        4789,
        {22265, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4789},
        {22265}
    },
    {
        4790,
        {22272, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4790},
        {22272}
    },
    {
        4791,
        {22271, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4791},
        {22271}
    },
    {
        4792,
        {22276, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4792},
        {22276}
    },
    {
        4793,
        {22281, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4793},
        {22281}
    },
    {
        4794,
        {22280, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4794},
        {22280}
    },
    {
        4795,
        {22283, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4795},
        {22283}
    },
    {
        4796,
        {22285, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4796},
        {22285}
    },
    {
        4797,
        {22291, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4797},
        {22291}
    },
    {
        4798,
        {22296, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4798},
        {22296}
    },
    {
        4799,
        {22294, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4799},
        {22294}
    },
    {
        4800,
        {21959, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4800},
        {21959}
    },
    {
        4801,
        {22300, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4801},
        {22300}
    },
    {
        4802,
        {22310, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4802},
        {22310}
    },
    {
        4803,
        {22327, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4803},
        {22327}
    },
    {
        4804,
        {22328, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4804},
        {22328}
    },
    {
        4805,
        {22350, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4805},
        {22350}
    },
    {
        4806,
        {22331, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4806},
        {22331}
    },
    {
        4807,
        {22336, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4807},
        {22336}
    },
    {
        4808,
        {22351, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4808},
        {22351}
    },
    {
        4809,
        {22377, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4809},
        {22377}
    },
    {
        4810,
        {22464, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4810},
        {22464}
    },
    {
        4811,
        {22408, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4811},
        {22408}
    },
    {
        4812,
        {22369, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4812},
        {22369}
    },
    {
        4813,
        {22399, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4813},
        {22399}
    },
    {
        4814,
        {22409, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4814},
        {22409}
    },
    {
        4815,
        {22419, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4815},
        {22419}
    },
    {
        4816,
        {22432, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4816},
        {22432}
    },
    {
        4817,
        {22451, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4817},
        {22451}
    },
    {
        4818,
        {22436, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4818},
        {22436}
    },
    {
        4819,
        {22442, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4819},
        {22442}
    },
    {
        4820,
        {22448, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4820},
        {22448}
    },
    {
        4821,
        {22467, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4821},
        {22467}
    },
    {
        4822,
        {22470, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4822},
        {22470}
    },
    {
        4823,
        {22484, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4823},
        {22484}
    },
    {
        4824,
        {22482, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4824},
        {22482}
    },
    {
        4825,
        {22483, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4825},
        {22483}
    },
    {
        4826,
        {22538, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4826},
        {22538}
    },
    {
        4827,
        {22486, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4827},
        {22486}
    },
    {
        4828,
        {22499, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4828},
        {22499}
    },
    {
        4829,
        {22539, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4829},
        {22539}
    },
    {
        4830,
        {22553, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4830},
        {22553}
    },
    {
        4831,
        {22557, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4831},
        {22557}
    },
    {
        4832,
        {22642, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4832},
        {22642}
    },
    {
        4833,
        {22561, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4833},
        {22561}
    },
    {
        4834,
        {22626, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4834},
        {22626}
    },
    {
        4835,
        {22603, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4835},
        {22603}
    },
    {
        4836,
        {22640, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4836},
        {22640}
    },
    {
        4837,
        {27584, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4837},
        {27584}
    },
    {
        4838,
        {22610, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4838},
        {22610}
    },
    {
        4839,
        {22589, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4839},
        {22589}
    },
    {
        4840,
        {22649, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4840},
        {22649}
    },
    {
        4841,
        {22661, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4841},
        {22661}
    },
    {
        4842,
        {22713, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4842},
        {22713}
    },
    {
        4843,
        {22687, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4843},
        {22687}
    },
    {
        4844,
        {22699, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4844},
        {22699}
    },
    {
        4845,
        {22714, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4845},
        {22714}
    },
    {
        4846,
        {22750, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4846},
        {22750}
    },
    {
        4847,
        {22715, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4847},
        {22715}
    },
    {
        4848,
        {22712, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4848},
        {22712}
    },
    {
        4849,
        {22702, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4849},
        {22702}
    },
    {
        4850,
        {22725, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4850},
        {22725}
    },
    {
        4851,
        {22739, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4851},
        {22739}
    },
    {
        4852,
        {22737, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4852},
        {22737}
    },
    {
        4853,
        {22743, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4853},
        {22743}
    },
    {
        4854,
        {22745, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4854},
        {22745}
    },
    {
        4855,
        {22744, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4855},
        {22744}
    },
    {
        4856,
        {22757, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4856},
        {22757}
    },
    {
        4857,
        {22748, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4857},
        {22748}
    },
    {
        4858,
        {22756, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4858},
        {22756}
    },
    {
        4859,
        {22751, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4859},
        {22751}
    },
    {
        4860,
        {22767, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4860},
        {22767}
    },
    {
        4861,
        {22778, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4861},
        {22778}
    },
    {
        4862,
        {22777, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4862},
        {22777}
    },
    {
        4863,
        {22779, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4863},
        {22779}
    },
    {
        4864,
        {22780, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4864},
        {22780}
    },
    {
        4865,
        {22781, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4865},
        {22781}
    },
    {
        4866,
        {22786, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4866},
        {22786}
    },
    {
        4867,
        {22794, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4867},
        {22794}
    },
    {
        4868,
        {22800, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4868},
        {22800}
    },
    {
        4869,
        {22811, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4869},
        {22811}
    },
    {
        4870,
        {26790, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4870},
        {26790}
    },
    {
        4871,
        {22821, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4871},
        {22821}
    },
    {
        4872,
        {22828, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4872},
        {22828}
    },
    {
        4873,
        {22829, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4873},
        {22829}
    },
    {
        4874,
        {22834, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4874},
        {22834}
    },
    {
        4875,
        {22840, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4875},
        {22840}
    },
    {
        4876,
        {22846, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4876},
        {22846}
    },
    {
        4877,
        {31442, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4877},
        {31442}
    },
    {
        4878,
        {22869, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4878},
        {22869}
    },
    {
        4879,
        {22864, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4879},
        {22864}
    },
    {
        4880,
        {22862, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4880},
        {22862}
    },
    {
        4881,
        {22874, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4881},
        {22874}
    },
    {
        4882,
        {22872, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4882},
        {22872}
    },
    {
        4883,
        {22882, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4883},
        {22882}
    },
    {
        4884,
        {22880, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4884},
        {22880}
    },
    {
        4885,
        {22887, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4885},
        {22887}
    },
    {
        4886,
        {22892, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4886},
        {22892}
    },
    {
        4887,
        {22889, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4887},
        {22889}
    },
    {
        4888,
        {22904, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4888},
        {22904}
    },
    {
        4889,
        {22913, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4889},
        {22913}
    },
    {
        4890,
        {22941, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4890},
        {22941}
    },
    {
        4891,
        {20318, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4891},
        {20318}
    },
    {
        4892,
        {20395, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4892},
        {20395}
    },
    {
        4893,
        {22947, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4893},
        {22947}
    },
    {
        4894,
        {22962, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4894},
        {22962}
    },
    {
        4895,
        {22982, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4895},
        {22982}
    },
    {
        4896,
        {23016, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4896},
        {23016}
    },
    {
        4897,
        {23004, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4897},
        {23004}
    },
    {
        4898,
        {22925, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4898},
        {22925}
    },
    {
        4899,
        {23001, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4899},
        {23001}
    },
    {
        4900,
        {23002, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4900},
        {23002}
    },
    {
        4901,
        {23077, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4901},
        {23077}
    },
    {
        4902,
        {23071, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4902},
        {23071}
    },
    {
        4903,
        {23057, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4903},
        {23057}
    },
    {
        4904,
        {23068, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4904},
        {23068}
    },
    {
        4905,
        {23049, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4905},
        {23049}
    },
    {
        4906,
        {23066, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4906},
        {23066}
    },
    {
        4907,
        {23104, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4907},
        {23104}
    },
    {
        4908,
        {23148, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4908},
        {23148}
    },
    {
        4909,
        {23113, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4909},
        {23113}
    },
    {
        4910,
        {23093, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4910},
        {23093}
    },
    {
        4911,
        {23094, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4911},
        {23094}
    },
    {
        4912,
        {23138, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4912},
        {23138}
    },
    {
        4913,
        {23146, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4913},
        {23146}
    },
    {
        4914,
        {23194, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4914},
        {23194}
    },
    {
        4915,
        {23228, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4915},
        {23228}
    },
    {
        4916,
        {23230, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4916},
        {23230}
    },
    {
        4917,
        {23243, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4917},
        {23243}
    },
    {
        4918,
        {23234, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4918},
        {23234}
    },
    {
        4919,
        {23229, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4919},
        {23229}
    },
    {
        4920,
        {23267, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4920},
        {23267}
    },
    {
        4921,
        {23255, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4921},
        {23255}
    },
    {
        4922,
        {23270, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4922},
        {23270}
    },
    {
        4923,
        {23273, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4923},
        {23273}
    },
    {
        4924,
        {23254, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4924},
        {23254}
    },
    {
        4925,
        {23290, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4925},
        {23290}
    },
    {
        4926,
        {23291, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4926},
        {23291}
    },
    {
        4927,
        {23308, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4927},
        {23308}
    },
    {
        4928,
        {23307, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4928},
        {23307}
    },
    {
        4929,
        {23318, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4929},
        {23318}
    },
    {
        4930,
        {23346, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4930},
        {23346}
    },
    {
        4931,
        {23248, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4931},
        {23248}
    },
    {
        4932,
        {23338, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4932},
        {23338}
    },
    {
        4933,
        {23350, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4933},
        {23350}
    },
    {
        4934,
        {23358, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4934},
        {23358}
    },
    {
        4935,
        {23363, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4935},
        {23363}
    },
    {
        4936,
        {23365, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4936},
        {23365}
    },
    {
        4937,
        {23360, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4937},
        {23360}
    },
    {
        4938,
        {23377, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4938},
        {23377}
    },
    {
        4939,
        {23381, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4939},
        {23381}
    },
    {
        4940,
        {23386, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4940},
        {23386}
    },
    {
        4941,
        {23387, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4941},
        {23387}
    },
    {
        4942,
        {23397, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4942},
        {23397}
    },
    {
        4943,
        {23401, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4943},
        {23401}
    },
    {
        4944,
        {23408, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4944},
        {23408}
    },
    {
        4945,
        {23411, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4945},
        {23411}
    },
    {
        4946,
        {23413, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4946},
        {23413}
    },
    {
        4947,
        {23416, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4947},
        {23416}
    },
    {
        4948,
        {25992, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4948},
        {25992}
    },
    {
        4949,
        {23418, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4949},
        {23418}
    },
    {
        4950,
        {23424, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4950},
        {23424}
    },
    {
        4951,
        {23427, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4951},
        {23427}
    },
    {
        4952,
        {23462, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4952},
        {23462}
    },
    {
        4953,
        {23480, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4953},
        {23480}
    },
    {
        4954,
        {23491, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4954},
        {23491}
    },
    {
        4955,
        {23495, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4955},
        {23495}
    },
    {
        4956,
        {23497, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4956},
        {23497}
    },
    {
        4957,
        {23508, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4957},
        {23508}
    },
    {
        4958,
        {23504, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4958},
        {23504}
    },
    {
        4959,
        {23524, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4959},
        {23524}
    },
    {
        4960,
        {23526, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4960},
        {23526}
    },
    {
        4961,
        {23522, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4961},
        {23522}
    },
    {
        4962,
        {23518, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4962},
        {23518}
    },
    {
        4963,
        {23525, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4963},
        {23525}
    },
    {
        4964,
        {23531, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4964},
        {23531}
    },
    {
        4965,
        {23536, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4965},
        {23536}
    },
    {
        4966,
        {23542, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4966},
        {23542}
    },
    {
        4967,
        {23539, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4967},
        {23539}
    },
    {
        4968,
        {23557, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4968},
        {23557}
    },
    {
        4969,
        {23559, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4969},
        {23559}
    },
    {
        4970,
        {23560, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4970},
        {23560}
    },
    {
        4971,
        {23565, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4971},
        {23565}
    },
    {
        4972,
        {23571, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4972},
        {23571}
    },
    {
        4973,
        {23584, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4973},
        {23584}
    },
    {
        4974,
        {23586, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4974},
        {23586}
    },
    {
        4975,
        {23592, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4975},
        {23592}
    },
    {
        4976,
        {23608, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4976},
        {23608}
    },
    {
        4977,
        {23609, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4977},
        {23609}
    },
    {
        4978,
        {23617, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4978},
        {23617}
    },
    {
        4979,
        {23622, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4979},
        {23622}
    },
    {
        4980,
        {23630, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4980},
        {23630}
    },
    {
        4981,
        {23635, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4981},
        {23635}
    },
    {
        4982,
        {23632, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4982},
        {23632}
    },
    {
        4983,
        {23631, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4983},
        {23631}
    },
    {
        4984,
        {23409, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4984},
        {23409}
    },
    {
        4985,
        {23660, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4985},
        {23660}
    },
    {
        4986,
        {23662, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4986},
        {23662}
    },
    {
        4987,
        {20066, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4987},
        {20066}
    },
    {
        4988,
        {23670, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4988},
        {23670}
    },
    {
        4989,
        {23673, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4989},
        {23673}
    },
    {
        4990,
        {23692, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4990},
        {23692}
    },
    {
        4991,
        {23697, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4991},
        {23697}
    },
    {
        4992,
        {23700, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4992},
        {23700}
    },
    {
        4993,
        {22939, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4993},
        {22939}
    },
    {
        4994,
        {23723, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4994},
        {23723}
    },
    {
        4995,
        {23739, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4995},
        {23739}
    },
    {
        4996,
        {23734, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4996},
        {23734}
    },
    {
        4997,
        {23740, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4997},
        {23740}
    },
    {
        4998,
        {23735, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4998},
        {23735}
    },
    {
        4999,
        {23749, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {4999},
        {23749}
    },
    {
        5000,
        {23742, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5000},
        {23742}
    },
    {
        5001,
        {23751, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5001},
        {23751}
    },
    {
        5002,
        {23769, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5002},
        {23769}
    },
    {
        5003,
        {23785, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5003},
        {23785}
    },
    {
        5004,
        {23805, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5004},
        {23805}
    },
    {
        5005,
        {23802, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5005},
        {23802}
    },
    {
        5006,
        {23789, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5006},
        {23789}
    },
    {
        5007,
        {23948, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5007},
        {23948}
    },
    {
        5008,
        {23786, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5008},
        {23786}
    },
    {
        5009,
        {23819, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5009},
        {23819}
    },
    {
        5010,
        {23829, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5010},
        {23829}
    },
    {
        5011,
        {23831, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5011},
        {23831}
    },
    {
        5012,
        {23900, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5012},
        {23900}
    },
    {
        5013,
        {23839, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5013},
        {23839}
    },
    {
        5014,
        {23835, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5014},
        {23835}
    },
    {
        5015,
        {23825, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5015},
        {23825}
    },
    {
        5016,
        {23828, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5016},
        {23828}
    },
    {
        5017,
        {23842, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5017},
        {23842}
    },
    {
        5018,
        {23834, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5018},
        {23834}
    },
    {
        5019,
        {23833, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5019},
        {23833}
    },
    {
        5020,
        {23832, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5020},
        {23832}
    },
    {
        5021,
        {23884, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5021},
        {23884}
    },
    {
        5022,
        {23890, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5022},
        {23890}
    },
    {
        5023,
        {23886, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5023},
        {23886}
    },
    {
        5024,
        {23883, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5024},
        {23883}
    },
    {
        5025,
        {23916, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5025},
        {23916}
    },
    {
        5026,
        {23923, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5026},
        {23923}
    },
    {
        5027,
        {23926, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5027},
        {23926}
    },
    {
        5028,
        {23943, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5028},
        {23943}
    },
    {
        5029,
        {23940, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5029},
        {23940}
    },
    {
        5030,
        {23938, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5030},
        {23938}
    },
    {
        5031,
        {23970, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5031},
        {23970}
    },
    {
        5032,
        {23965, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5032},
        {23965}
    },
    {
        5033,
        {23980, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5033},
        {23980}
    },
    {
        5034,
        {23982, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5034},
        {23982}
    },
    {
        5035,
        {23997, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5035},
        {23997}
    },
    {
        5036,
        {23952, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5036},
        {23952}
    },
    {
        5037,
        {23991, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5037},
        {23991}
    },
    {
        5038,
        {23996, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5038},
        {23996}
    },
    {
        5039,
        {24009, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5039},
        {24009}
    },
    {
        5040,
        {24013, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5040},
        {24013}
    },
    {
        5041,
        {24019, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5041},
        {24019}
    },
    {
        5042,
        {24018, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5042},
        {24018}
    },
    {
        5043,
        {24022, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5043},
        {24022}
    },
    {
        5044,
        {24027, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5044},
        {24027}
    },
    {
        5045,
        {24043, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5045},
        {24043}
    },
    {
        5046,
        {24050, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5046},
        {24050}
    },
    {
        5047,
        {24053, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5047},
        {24053}
    },
    {
        5048,
        {24075, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5048},
        {24075}
    },
    {
        5049,
        {24090, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5049},
        {24090}
    },
    {
        5050,
        {24089, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5050},
        {24089}
    },
    {
        5051,
        {24081, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5051},
        {24081}
    },
    {
        5052,
        {24091, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5052},
        {24091}
    },
    {
        5053,
        {24118, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5053},
        {24118}
    },
    {
        5054,
        {24119, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5054},
        {24119}
    },
    {
        5055,
        {24132, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5055},
        {24132}
    },
    {
        5056,
        {24131, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5056},
        {24131}
    },
    {
        5057,
        {24128, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5057},
        {24128}
    },
    {
        5058,
        {24142, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5058},
        {24142}
    },
    {
        5059,
        {24151, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5059},
        {24151}
    },
    {
        5060,
        {24148, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5060},
        {24148}
    },
    {
        5061,
        {24159, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5061},
        {24159}
    },
    {
        5062,
        {24162, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5062},
        {24162}
    },
    {
        5063,
        {24164, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5063},
        {24164}
    },
    {
        5064,
        {24135, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5064},
        {24135}
    },
    {
        5065,
        {24181, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5065},
        {24181}
    },
    {
        5066,
        {24182, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5066},
        {24182}
    },
    {
        5067,
        {24186, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5067},
        {24186}
    },
    {
        5068,
        {40636, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5068},
        {40636}
    },
    {
        5069,
        {24191, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5069},
        {24191}
    },
    {
        5070,
        {24224, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5070},
        {24224}
    },
    {
        5071,
        {24257, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5071},
        {24257}
    },
    {
        5072,
        {24258, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5072},
        {24258}
    },
    {
        5073,
        {24264, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5073},
        {24264}
    },
    {
        5074,
        {24272, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5074},
        {24272}
    },
    {
        5075,
        {24271, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5075},
        {24271}
    },
    {
        5076,
        {24278, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5076},
        {24278}
    },
    {
        5077,
        {24291, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5077},
        {24291}
    },
    {
        5078,
        {24285, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5078},
        {24285}
    },
    {
        5079,
        {24282, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5079},
        {24282}
    },
    {
        5080,
        {24283, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5080},
        {24283}
    },
    {
        5081,
        {24290, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5081},
        {24290}
    },
    {
        5082,
        {24289, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5082},
        {24289}
    },
    {
        5083,
        {24296, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5083},
        {24296}
    },
    {
        5084,
        {24297, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5084},
        {24297}
    },
    {
        5085,
        {24300, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5085},
        {24300}
    },
    {
        5086,
        {24305, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5086},
        {24305}
    },
    {
        5087,
        {24307, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5087},
        {24307}
    },
    {
        5088,
        {24304, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5088},
        {24304}
    },
    {
        5089,
        {24308, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5089},
        {24308}
    },
    {
        5090,
        {24312, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5090},
        {24312}
    },
    {
        5091,
        {24318, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5091},
        {24318}
    },
    {
        5092,
        {24323, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5092},
        {24323}
    },
    {
        5093,
        {24329, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5093},
        {24329}
    },
    {
        5094,
        {24413, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5094},
        {24413}
    },
    {
        5095,
        {24412, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5095},
        {24412}
    },
    {
        5096,
        {24331, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5096},
        {24331}
    },
    {
        5097,
        {24337, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5097},
        {24337}
    },
    {
        5098,
        {24342, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5098},
        {24342}
    },
    {
        5099,
        {24361, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5099},
        {24361}
    },
    {
        5100,
        {24365, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5100},
        {24365}
    },
    {
        5101,
        {24376, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5101},
        {24376}
    },
    {
        5102,
        {24385, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5102},
        {24385}
    },
    {
        5103,
        {24392, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5103},
        {24392}
    },
    {
        5104,
        {24396, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5104},
        {24396}
    },
    {
        5105,
        {24398, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5105},
        {24398}
    },
    {
        5106,
        {24367, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5106},
        {24367}
    },
    {
        5107,
        {24401, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5107},
        {24401}
    },
    {
        5108,
        {24406, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5108},
        {24406}
    },
    {
        5109,
        {24407, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5109},
        {24407}
    },
    {
        5110,
        {24409, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5110},
        {24409}
    },
    {
        5111,
        {24417, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5111},
        {24417}
    },
    {
        5112,
        {24429, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5112},
        {24429}
    },
    {
        5113,
        {24435, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5113},
        {24435}
    },
    {
        5114,
        {24439, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5114},
        {24439}
    },
    {
        5115,
        {24451, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5115},
        {24451}
    },
    {
        5116,
        {24450, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5116},
        {24450}
    },
    {
        5117,
        {24447, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5117},
        {24447}
    },
    {
        5118,
        {24458, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5118},
        {24458}
    },
    {
        5119,
        {24456, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5119},
        {24456}
    },
    {
        5120,
        {24465, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5120},
        {24465}
    },
    {
        5121,
        {24455, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5121},
        {24455}
    },
    {
        5122,
        {24478, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5122},
        {24478}
    },
    {
        5123,
        {24473, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5123},
        {24473}
    },
    {
        5124,
        {24472, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5124},
        {24472}
    },
    {
        5125,
        {24480, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5125},
        {24480}
    },
    {
        5126,
        {24488, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5126},
        {24488}
    },
    {
        5127,
        {24493, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5127},
        {24493}
    },
    {
        5128,
        {24508, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5128},
        {24508}
    },
    {
        5129,
        {24534, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5129},
        {24534}
    },
    {
        5130,
        {24571, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5130},
        {24571}
    },
    {
        5131,
        {24548, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5131},
        {24548}
    },
    {
        5132,
        {24568, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5132},
        {24568}
    },
    {
        5133,
        {24561, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5133},
        {24561}
    },
    {
        5134,
        {24541, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5134},
        {24541}
    },
    {
        5135,
        {24755, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5135},
        {24755}
    },
    {
        5136,
        {24575, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5136},
        {24575}
    },
    {
        5137,
        {24609, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5137},
        {24609}
    },
    {
        5138,
        {24672, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5138},
        {24672}
    },
    {
        5139,
        {24601, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5139},
        {24601}
    },
    {
        5140,
        {24592, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5140},
        {24592}
    },
    {
        5141,
        {24617, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5141},
        {24617}
    },
    {
        5142,
        {24590, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5142},
        {24590}
    },
    {
        5143,
        {24625, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5143},
        {24625}
    },
    {
        5144,
        {24603, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5144},
        {24603}
    },
    {
        5145,
        {24597, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5145},
        {24597}
    },
    {
        5146,
        {24619, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5146},
        {24619}
    },
    {
        5147,
        {24614, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5147},
        {24614}
    },
    {
        5148,
        {24591, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5148},
        {24591}
    },
    {
        5149,
        {24634, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5149},
        {24634}
    },
    {
        5150,
        {24666, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5150},
        {24666}
    },
    {
        5151,
        {24641, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5151},
        {24641}
    },
    {
        5152,
        {24682, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5152},
        {24682}
    },
    {
        5153,
        {24695, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5153},
        {24695}
    },
    {
        5154,
        {24671, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5154},
        {24671}
    },
    {
        5155,
        {24650, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5155},
        {24650}
    },
    {
        5156,
        {24646, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5156},
        {24646}
    },
    {
        5157,
        {24653, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5157},
        {24653}
    },
    {
        5158,
        {24675, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5158},
        {24675}
    },
    {
        5159,
        {24643, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5159},
        {24643}
    },
    {
        5160,
        {24676, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5160},
        {24676}
    },
    {
        5161,
        {24642, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5161},
        {24642}
    },
    {
        5162,
        {24684, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5162},
        {24684}
    },
    {
        5163,
        {24683, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5163},
        {24683}
    },
    {
        5164,
        {24665, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5164},
        {24665}
    },
    {
        5165,
        {24705, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5165},
        {24705}
    },
    {
        5166,
        {24717, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5166},
        {24717}
    },
    {
        5167,
        {24807, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5167},
        {24807}
    },
    {
        5168,
        {24707, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5168},
        {24707}
    },
    {
        5169,
        {24730, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5169},
        {24730}
    },
    {
        5170,
        {24708, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5170},
        {24708}
    },
    {
        5171,
        {24731, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5171},
        {24731}
    },
    {
        5172,
        {24726, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5172},
        {24726}
    },
    {
        5173,
        {24727, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5173},
        {24727}
    },
    {
        5174,
        {24722, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5174},
        {24722}
    },
    {
        5175,
        {24743, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5175},
        {24743}
    },
    {
        5176,
        {24715, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5176},
        {24715}
    },
    {
        5177,
        {24801, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5177},
        {24801}
    },
    {
        5178,
        {24760, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5178},
        {24760}
    },
    {
        5179,
        {24800, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5179},
        {24800}
    },
    {
        5180,
        {24787, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5180},
        {24787}
    },
    {
        5181,
        {24756, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5181},
        {24756}
    },
    {
        5182,
        {24560, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5182},
        {24560}
    },
    {
        5183,
        {24765, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5183},
        {24765}
    },
    {
        5184,
        {24774, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5184},
        {24774}
    },
    {
        5185,
        {24757, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5185},
        {24757}
    },
    {
        5186,
        {24792, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5186},
        {24792}
    },
    {
        5187,
        {24909, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5187},
        {24909}
    },
    {
        5188,
        {24853, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5188},
        {24853}
    },
    {
        5189,
        {24838, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5189},
        {24838}
    },
    {
        5190,
        {24822, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5190},
        {24822}
    },
    {
        5191,
        {24823, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5191},
        {24823}
    },
    {
        5192,
        {24832, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5192},
        {24832}
    },
    {
        5193,
        {24820, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5193},
        {24820}
    },
    {
        5194,
        {24826, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5194},
        {24826}
    },
    {
        5195,
        {24835, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5195},
        {24835}
    },
    {
        5196,
        {24865, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5196},
        {24865}
    },
    {
        5197,
        {24827, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5197},
        {24827}
    },
    {
        5198,
        {24817, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5198},
        {24817}
    },
    {
        5199,
        {24845, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5199},
        {24845}
    },
    {
        5200,
        {24846, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5200},
        {24846}
    },
    {
        5201,
        {24903, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5201},
        {24903}
    },
    {
        5202,
        {24894, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5202},
        {24894}
    },
    {
        5203,
        {24872, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5203},
        {24872}
    },
    {
        5204,
        {24871, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5204},
        {24871}
    },
    {
        5205,
        {24906, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5205},
        {24906}
    },
    {
        5206,
        {24895, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5206},
        {24895}
    },
    {
        5207,
        {24892, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5207},
        {24892}
    },
    {
        5208,
        {24876, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5208},
        {24876}
    },
    {
        5209,
        {24884, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5209},
        {24884}
    },
    {
        5210,
        {24893, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5210},
        {24893}
    },
    {
        5211,
        {24898, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5211},
        {24898}
    },
    {
        5212,
        {24900, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5212},
        {24900}
    },
    {
        5213,
        {24947, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5213},
        {24947}
    },
    {
        5214,
        {24951, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5214},
        {24951}
    },
    {
        5215,
        {24920, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5215},
        {24920}
    },
    {
        5216,
        {24921, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5216},
        {24921}
    },
    {
        5217,
        {24922, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5217},
        {24922}
    },
    {
        5218,
        {24939, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5218},
        {24939}
    },
    {
        5219,
        {24948, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5219},
        {24948}
    },
    {
        5220,
        {24943, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5220},
        {24943}
    },
    {
        5221,
        {24933, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5221},
        {24933}
    },
    {
        5222,
        {24945, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5222},
        {24945}
    },
    {
        5223,
        {24927, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5223},
        {24927}
    },
    {
        5224,
        {24925, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5224},
        {24925}
    },
    {
        5225,
        {24915, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5225},
        {24915}
    },
    {
        5226,
        {24949, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5226},
        {24949}
    },
    {
        5227,
        {24985, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5227},
        {24985}
    },
    {
        5228,
        {24982, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5228},
        {24982}
    },
    {
        5229,
        {24967, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5229},
        {24967}
    },
    {
        5230,
        {25004, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5230},
        {25004}
    },
    {
        5231,
        {24980, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5231},
        {24980}
    },
    {
        5232,
        {24986, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5232},
        {24986}
    },
    {
        5233,
        {24970, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5233},
        {24970}
    },
    {
        5234,
        {24977, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5234},
        {24977}
    },
    {
        5235,
        {25003, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5235},
        {25003}
    },
    {
        5236,
        {25006, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5236},
        {25006}
    },
    {
        5237,
        {25036, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5237},
        {25036}
    },
    {
        5238,
        {25034, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5238},
        {25034}
    },
    {
        5239,
        {25033, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5239},
        {25033}
    },
    {
        5240,
        {25079, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5240},
        {25079}
    },
    {
        5241,
        {25032, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5241},
        {25032}
    },
    {
        5242,
        {25027, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5242},
        {25027}
    },
    {
        5243,
        {25030, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5243},
        {25030}
    },
    {
        5244,
        {25018, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5244},
        {25018}
    },
    {
        5245,
        {25035, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5245},
        {25035}
    },
    {
        5246,
        {32633, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5246},
        {32633}
    },
    {
        5247,
        {25037, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5247},
        {25037}
    },
    {
        5248,
        {25062, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5248},
        {25062}
    },
    {
        5249,
        {25059, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5249},
        {25059}
    },
    {
        5250,
        {25078, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5250},
        {25078}
    },
    {
        5251,
        {25082, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5251},
        {25082}
    },
    {
        5252,
        {25076, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5252},
        {25076}
    },
    {
        5253,
        {25087, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5253},
        {25087}
    },
    {
        5254,
        {25085, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5254},
        {25085}
    },
    {
        5255,
        {25084, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5255},
        {25084}
    },
    {
        5256,
        {25086, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5256},
        {25086}
    },
    {
        5257,
        {25088, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5257},
        {25088}
    },
    {
        5258,
        {25096, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5258},
        {25096}
    },
    {
        5259,
        {25097, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5259},
        {25097}
    },
    {
        5260,
        {25101, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5260},
        {25101}
    },
    {
        5261,
        {25100, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5261},
        {25100}
    },
    {
        5262,
        {25108, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5262},
        {25108}
    },
    {
        5263,
        {25115, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5263},
        {25115}
    },
    {
        5264,
        {25118, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5264},
        {25118}
    },
    {
        5265,
        {25121, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5265},
        {25121}
    },
    {
        5266,
        {25130, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5266},
        {25130}
    },
    {
        5267,
        {25134, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5267},
        {25134}
    },
    {
        5268,
        {25136, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5268},
        {25136}
    },
    {
        5269,
        {25138, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5269},
        {25138}
    },
    {
        5270,
        {25139, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5270},
        {25139}
    },
    {
        5271,
        {25153, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5271},
        {25153}
    },
    {
        5272,
        {25166, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5272},
        {25166}
    },
    {
        5273,
        {25182, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5273},
        {25182}
    },
    {
        5274,
        {25187, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5274},
        {25187}
    },
    {
        5275,
        {25179, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5275},
        {25179}
    },
    {
        5276,
        {25184, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5276},
        {25184}
    },
    {
        5277,
        {25192, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5277},
        {25192}
    },
    {
        5278,
        {25212, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5278},
        {25212}
    },
    {
        5279,
        {25218, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5279},
        {25218}
    },
    {
        5280,
        {25225, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5280},
        {25225}
    },
    {
        5281,
        {25214, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5281},
        {25214}
    },
    {
        5282,
        {25234, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5282},
        {25234}
    },
    {
        5283,
        {25235, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5283},
        {25235}
    },
    {
        5284,
        {25238, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5284},
        {25238}
    },
    {
        5285,
        {25300, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5285},
        {25300}
    },
    {
        5286,
        {25219, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5286},
        {25219}
    },
    {
        5287,
        {25236, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5287},
        {25236}
    },
    {
        5288,
        {25303, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5288},
        {25303}
    },
    {
        5289,
        {25297, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5289},
        {25297}
    },
    {
        5290,
        {25275, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5290},
        {25275}
    },
    {
        5291,
        {25295, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5291},
        {25295}
    },
    {
        5292,
        {25343, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5292},
        {25343}
    },
    {
        5293,
        {25286, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5293},
        {25286}
    },
    {
        5294,
        {25812, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5294},
        {25812}
    },
    {
        5295,
        {25288, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5295},
        {25288}
    },
    {
        5296,
        {25308, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5296},
        {25308}
    },
    {
        5297,
        {25292, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5297},
        {25292}
    },
    {
        5298,
        {25290, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5298},
        {25290}
    },
    {
        5299,
        {25282, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5299},
        {25282}
    },
    {
        5300,
        {25287, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5300},
        {25287}
    },
    {
        5301,
        {25243, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5301},
        {25243}
    },
    {
        5302,
        {25289, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5302},
        {25289}
    },
    {
        5303,
        {25356, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5303},
        {25356}
    },
    {
        5304,
        {25326, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5304},
        {25326}
    },
    {
        5305,
        {25329, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5305},
        {25329}
    },
    {
        5306,
        {25383, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5306},
        {25383}
    },
    {
        5307,
        {25346, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5307},
        {25346}
    },
    {
        5308,
        {25352, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5308},
        {25352}
    },
    {
        5309,
        {25327, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5309},
        {25327}
    },
    {
        5310,
        {25333, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5310},
        {25333}
    },
    {
        5311,
        {25424, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5311},
        {25424}
    },
    {
        5312,
        {25406, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5312},
        {25406}
    },
    {
        5313,
        {25421, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5313},
        {25421}
    },
    {
        5314,
        {25628, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5314},
        {25628}
    },
    {
        5315,
        {25423, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5315},
        {25423}
    },
    {
        5316,
        {25494, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5316},
        {25494}
    },
    {
        5317,
        {25486, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5317},
        {25486}
    },
    {
        5318,
        {25472, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5318},
        {25472}
    },
    {
        5319,
        {25515, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5319},
        {25515}
    },
    {
        5320,
        {25462, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5320},
        {25462}
    },
    {
        5321,
        {25507, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5321},
        {25507}
    },
    {
        5322,
        {25487, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5322},
        {25487}
    },
    {
        5323,
        {25481, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5323},
        {25481}
    },
    {
        5324,
        {25503, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5324},
        {25503}
    },
    {
        5325,
        {25525, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5325},
        {25525}
    },
    {
        5326,
        {25451, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5326},
        {25451}
    },
    {
        5327,
        {25449, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5327},
        {25449}
    },
    {
        5328,
        {25534, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5328},
        {25534}
    },
    {
        5329,
        {25577, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5329},
        {25577}
    },
    {
        5330,
        {25536, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5330},
        {25536}
    },
    {
        5331,
        {25542, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5331},
        {25542}
    },
    {
        5332,
        {25571, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5332},
        {25571}
    },
    {
        5333,
        {25545, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5333},
        {25545}
    },
    {
        5334,
        {25554, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5334},
        {25554}
    },
    {
        5335,
        {25590, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5335},
        {25590}
    },
    {
        5336,
        {25540, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5336},
        {25540}
    },
    {
        5337,
        {25622, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5337},
        {25622}
    },
    {
        5338,
        {25652, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5338},
        {25652}
    },
    {
        5339,
        {25606, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5339},
        {25606}
    },
    {
        5340,
        {25619, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5340},
        {25619}
    },
    {
        5341,
        {25638, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5341},
        {25638}
    },
    {
        5342,
        {25654, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5342},
        {25654}
    },
    {
        5343,
        {25885, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5343},
        {25885}
    },
    {
        5344,
        {25623, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5344},
        {25623}
    },
    {
        5345,
        {25640, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5345},
        {25640}
    },
    {
        5346,
        {25615, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5346},
        {25615}
    },
    {
        5347,
        {25703, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5347},
        {25703}
    },
    {
        5348,
        {25711, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5348},
        {25711}
    },
    {
        5349,
        {25718, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5349},
        {25718}
    },
    {
        5350,
        {25678, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5350},
        {25678}
    },
    {
        5351,
        {25898, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5351},
        {25898}
    },
    {
        5352,
        {25749, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5352},
        {25749}
    },
    {
        5353,
        {25747, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5353},
        {25747}
    },
    {
        5354,
        {25765, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5354},
        {25765}
    },
    {
        5355,
        {25769, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5355},
        {25769}
    },
    {
        5356,
        {25736, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5356},
        {25736}
    },
    {
        5357,
        {25788, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5357},
        {25788}
    },
    {
        5358,
        {25818, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5358},
        {25818}
    },
    {
        5359,
        {25810, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5359},
        {25810}
    },
    {
        5360,
        {25797, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5360},
        {25797}
    },
    {
        5361,
        {25799, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5361},
        {25799}
    },
    {
        5362,
        {25787, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5362},
        {25787}
    },
    {
        5363,
        {25816, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5363},
        {25816}
    },
    {
        5364,
        {25794, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5364},
        {25794}
    },
    {
        5365,
        {25841, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5365},
        {25841}
    },
    {
        5366,
        {25831, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5366},
        {25831}
    },
    {
        5367,
        {33289, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5367},
        {33289}
    },
    {
        5368,
        {25824, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5368},
        {25824}
    },
    {
        5369,
        {25825, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5369},
        {25825}
    },
    {
        5370,
        {25260, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5370},
        {25260}
    },
    {
        5371,
        {25827, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5371},
        {25827}
    },
    {
        5372,
        {25839, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5372},
        {25839}
    },
    {
        5373,
        {25900, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5373},
        {25900}
    },
    {
        5374,
        {25846, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5374},
        {25846}
    },
    {
        5375,
        {25844, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5375},
        {25844}
    },
    {
        5376,
        {25842, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5376},
        {25842}
    },
    {
        5377,
        {25850, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5377},
        {25850}
    },
    {
        5378,
        {25856, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5378},
        {25856}
    },
    {
        5379,
        {25853, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5379},
        {25853}
    },
    {
        5380,
        {25880, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5380},
        {25880}
    },
    {
        5381,
        {25884, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5381},
        {25884}
    },
    {
        5382,
        {25861, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5382},
        {25861}
    },
    {
        5383,
        {25892, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5383},
        {25892}
    },
    {
        5384,
        {25891, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5384},
        {25891}
    },
    {
        5385,
        {25899, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5385},
        {25899}
    },
    {
        5386,
        {25908, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5386},
        {25908}
    },
    {
        5387,
        {25909, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5387},
        {25909}
    },
    {
        5388,
        {25911, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5388},
        {25911}
    },
    {
        5389,
        {25910, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5389},
        {25910}
    },
    {
        5390,
        {25912, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5390},
        {25912}
    },
    {
        5391,
        {30027, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5391},
        {30027}
    },
    {
        5392,
        {25928, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5392},
        {25928}
    },
    {
        5393,
        {25942, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5393},
        {25942}
    },
    {
        5394,
        {25941, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5394},
        {25941}
    },
    {
        5395,
        {25933, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5395},
        {25933}
    },
    {
        5396,
        {25944, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5396},
        {25944}
    },
    {
        5397,
        {25950, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5397},
        {25950}
    },
    {
        5398,
        {25949, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5398},
        {25949}
    },
    {
        5399,
        {25970, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5399},
        {25970}
    },
    {
        5400,
        {25976, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5400},
        {25976}
    },
    {
        5401,
        {25986, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5401},
        {25986}
    },
    {
        5402,
        {25987, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5402},
        {25987}
    },
    {
        5403,
        {35722, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5403},
        {35722}
    },
    {
        5404,
        {26011, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5404},
        {26011}
    },
    {
        5405,
        {26015, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5405},
        {26015}
    },
    {
        5406,
        {26027, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5406},
        {26027}
    },
    {
        5407,
        {26039, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5407},
        {26039}
    },
    {
        5408,
        {26051, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5408},
        {26051}
    },
    {
        5409,
        {26054, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5409},
        {26054}
    },
    {
        5410,
        {26049, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5410},
        {26049}
    },
    {
        5411,
        {26052, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5411},
        {26052}
    },
    {
        5412,
        {26060, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5412},
        {26060}
    },
    {
        5413,
        {26066, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5413},
        {26066}
    },
    {
        5414,
        {26075, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5414},
        {26075}
    },
    {
        5415,
        {26073, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5415},
        {26073}
    },
    {
        5416,
        {26080, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5416},
        {26080}
    },
    {
        5417,
        {26081, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5417},
        {26081}
    },
    {
        5418,
        {26097, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5418},
        {26097}
    },
    {
        5419,
        {26482, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5419},
        {26482}
    },
    {
        5420,
        {26122, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5420},
        {26122}
    },
    {
        5421,
        {26115, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5421},
        {26115}
    },
    {
        5422,
        {26107, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5422},
        {26107}
    },
    {
        5423,
        {26483, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5423},
        {26483}
    },
    {
        5424,
        {26165, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5424},
        {26165}
    },
    {
        5425,
        {26166, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5425},
        {26166}
    },
    {
        5426,
        {26164, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5426},
        {26164}
    },
    {
        5427,
        {26140, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5427},
        {26140}
    },
    {
        5428,
        {26191, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5428},
        {26191}
    },
    {
        5429,
        {26180, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5429},
        {26180}
    },
    {
        5430,
        {26185, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5430},
        {26185}
    },
    {
        5431,
        {26177, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5431},
        {26177}
    },
    {
        5432,
        {26206, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5432},
        {26206}
    },
    {
        5433,
        {26205, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5433},
        {26205}
    },
    {
        5434,
        {26212, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5434},
        {26212}
    },
    {
        5435,
        {26215, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5435},
        {26215}
    },
    {
        5436,
        {26216, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5436},
        {26216}
    },
    {
        5437,
        {26207, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5437},
        {26207}
    },
    {
        5438,
        {26210, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5438},
        {26210}
    },
    {
        5439,
        {26224, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5439},
        {26224}
    },
    {
        5440,
        {26243, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5440},
        {26243}
    },
    {
        5441,
        {26248, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5441},
        {26248}
    },
    {
        5442,
        {26254, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5442},
        {26254}
    },
    {
        5443,
        {26249, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5443},
        {26249}
    },
    {
        5444,
        {26244, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5444},
        {26244}
    },
    {
        5445,
        {26264, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5445},
        {26264}
    },
    {
        5446,
        {26269, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5446},
        {26269}
    },
    {
        5447,
        {26305, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5447},
        {26305}
    },
    {
        5448,
        {26297, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5448},
        {26297}
    },
    {
        5449,
        {26313, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5449},
        {26313}
    },
    {
        5450,
        {26302, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5450},
        {26302}
    },
    {
        5451,
        {26300, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5451},
        {26300}
    },
    {
        5452,
        {26308, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5452},
        {26308}
    },
    {
        5453,
        {26296, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5453},
        {26296}
    },
    {
        5454,
        {26326, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5454},
        {26326}
    },
    {
        5455,
        {26330, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5455},
        {26330}
    },
    {
        5456,
        {26336, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5456},
        {26336}
    },
    {
        5457,
        {26175, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5457},
        {26175}
    },
    {
        5458,
        {26342, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5458},
        {26342}
    },
    {
        5459,
        {26345, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5459},
        {26345}
    },
    {
        5460,
        {26352, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5460},
        {26352}
    },
    {
        5461,
        {26357, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5461},
        {26357}
    },
    {
        5462,
        {26359, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5462},
        {26359}
    },
    {
        5463,
        {26383, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5463},
        {26383}
    },
    {
        5464,
        {26390, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5464},
        {26390}
    },
    {
        5465,
        {26398, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5465},
        {26398}
    },
    {
        5466,
        {26406, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5466},
        {26406}
    },
    {
        5467,
        {26407, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5467},
        {26407}
    },
    {
        5468,
        {38712, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5468},
        {38712}
    },
    {
        5469,
        {26414, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5469},
        {26414}
    },
    {
        5470,
        {26431, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5470},
        {26431}
    },
    {
        5471,
        {26422, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5471},
        {26422}
    },
    {
        5472,
        {26433, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5472},
        {26433}
    },
    {
        5473,
        {26424, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5473},
        {26424}
    },
    {
        5474,
        {26423, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5474},
        {26423}
    },
    {
        5475,
        {26438, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5475},
        {26438}
    },
    {
        5476,
        {26462, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5476},
        {26462}
    },
    {
        5477,
        {26464, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5477},
        {26464}
    },
    {
        5478,
        {26457, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5478},
        {26457}
    },
    {
        5479,
        {26467, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5479},
        {26467}
    },
    {
        5480,
        {26468, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5480},
        {26468}
    },
    {
        5481,
        {26505, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5481},
        {26505}
    },
    {
        5482,
        {26480, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5482},
        {26480}
    },
    {
        5483,
        {26537, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5483},
        {26537}
    },
    {
        5484,
        {26492, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5484},
        {26492}
    },
    {
        5485,
        {26474, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5485},
        {26474}
    },
    {
        5486,
        {26508, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5486},
        {26508}
    },
    {
        5487,
        {26507, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5487},
        {26507}
    },
    {
        5488,
        {26534, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5488},
        {26534}
    },
    {
        5489,
        {26529, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5489},
        {26529}
    },
    {
        5490,
        {26501, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5490},
        {26501}
    },
    {
        5491,
        {26551, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5491},
        {26551}
    },
    {
        5492,
        {26607, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5492},
        {26607}
    },
    {
        5493,
        {26548, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5493},
        {26548}
    },
    {
        5494,
        {26604, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5494},
        {26604}
    },
    {
        5495,
        {26547, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5495},
        {26547}
    },
    {
        5496,
        {26601, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5496},
        {26601}
    },
    {
        5497,
        {26552, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5497},
        {26552}
    },
    {
        5498,
        {26596, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5498},
        {26596}
    },
    {
        5499,
        {26590, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5499},
        {26590}
    },
    {
        5500,
        {26589, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5500},
        {26589}
    },
    {
        5501,
        {26594, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5501},
        {26594}
    },
    {
        5502,
        {26606, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5502},
        {26606}
    },
    {
        5503,
        {26553, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5503},
        {26553}
    },
    {
        5504,
        {26574, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5504},
        {26574}
    },
    {
        5505,
        {26566, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5505},
        {26566}
    },
    {
        5506,
        {26599, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5506},
        {26599}
    },
    {
        5507,
        {27292, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5507},
        {27292}
    },
    {
        5508,
        {26654, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5508},
        {26654}
    },
    {
        5509,
        {26694, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5509},
        {26694}
    },
    {
        5510,
        {26665, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5510},
        {26665}
    },
    {
        5511,
        {26688, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5511},
        {26688}
    },
    {
        5512,
        {26701, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5512},
        {26701}
    },
    {
        5513,
        {26674, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5513},
        {26674}
    },
    {
        5514,
        {26702, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5514},
        {26702}
    },
    {
        5515,
        {26803, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5515},
        {26803}
    },
    {
        5516,
        {26667, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5516},
        {26667}
    },
    {
        5517,
        {26713, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5517},
        {26713}
    },
    {
        5518,
        {26723, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5518},
        {26723}
    },
    {
        5519,
        {26743, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5519},
        {26743}
    },
    {
        5520,
        {26751, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5520},
        {26751}
    },
    {
        5521,
        {26783, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5521},
        {26783}
    },
    {
        5522,
        {26767, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5522},
        {26767}
    },
    {
        5523,
        {26797, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5523},
        {26797}
    },
    {
        5524,
        {26772, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5524},
        {26772}
    },
    {
        5525,
        {26781, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5525},
        {26781}
    },
    {
        5526,
        {26779, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5526},
        {26779}
    },
    {
        5527,
        {26755, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5527},
        {26755}
    },
    {
        5528,
        {27310, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5528},
        {27310}
    },
    {
        5529,
        {26809, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5529},
        {26809}
    },
    {
        5530,
        {26740, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5530},
        {26740}
    },
    {
        5531,
        {26805, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5531},
        {26805}
    },
    {
        5532,
        {26784, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5532},
        {26784}
    },
    {
        5533,
        {26810, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5533},
        {26810}
    },
    {
        5534,
        {26895, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5534},
        {26895}
    },
    {
        5535,
        {26765, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5535},
        {26765}
    },
    {
        5536,
        {26750, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5536},
        {26750}
    },
    {
        5537,
        {26881, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5537},
        {26881}
    },
    {
        5538,
        {26826, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5538},
        {26826}
    },
    {
        5539,
        {26888, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5539},
        {26888}
    },
    {
        5540,
        {26840, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5540},
        {26840}
    },
    {
        5541,
        {26914, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5541},
        {26914}
    },
    {
        5542,
        {26918, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5542},
        {26918}
    },
    {
        5543,
        {26849, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5543},
        {26849}
    },
    {
        5544,
        {26892, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5544},
        {26892}
    },
    {
        5545,
        {26829, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5545},
        {26829}
    },
    {
        5546,
        {26836, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5546},
        {26836}
    },
    {
        5547,
        {26855, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5547},
        {26855}
    },
    {
        5548,
        {26837, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5548},
        {26837}
    },
    {
        5549,
        {26934, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5549},
        {26934}
    },
    {
        5550,
        {26898, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5550},
        {26898}
    },
    {
        5551,
        {26884, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5551},
        {26884}
    },
    {
        5552,
        {26839, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5552},
        {26839}
    },
    {
        5553,
        {26851, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5553},
        {26851}
    },
    {
        5554,
        {26917, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5554},
        {26917}
    },
    {
        5555,
        {26873, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5555},
        {26873}
    },
    {
        5556,
        {26848, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5556},
        {26848}
    },
    {
        5557,
        {26863, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5557},
        {26863}
    },
    {
        5558,
        {26920, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5558},
        {26920}
    },
    {
        5559,
        {26922, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5559},
        {26922}
    },
    {
        5560,
        {26906, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5560},
        {26906}
    },
    {
        5561,
        {26915, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5561},
        {26915}
    },
    {
        5562,
        {26913, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5562},
        {26913}
    },
    {
        5563,
        {26822, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5563},
        {26822}
    },
    {
        5564,
        {27001, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5564},
        {27001}
    },
    {
        5565,
        {26999, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5565},
        {26999}
    },
    {
        5566,
        {26972, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5566},
        {26972}
    },
    {
        5567,
        {27000, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5567},
        {27000}
    },
    {
        5568,
        {26987, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5568},
        {26987}
    },
    {
        5569,
        {26964, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5569},
        {26964}
    },
    {
        5570,
        {27006, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5570},
        {27006}
    },
    {
        5571,
        {26990, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5571},
        {26990}
    },
    {
        5572,
        {26937, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5572},
        {26937}
    },
    {
        5573,
        {26996, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5573},
        {26996}
    },
    {
        5574,
        {26941, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5574},
        {26941}
    },
    {
        5575,
        {26969, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5575},
        {26969}
    },
    {
        5576,
        {26928, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5576},
        {26928}
    },
    {
        5577,
        {26977, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5577},
        {26977}
    },
    {
        5578,
        {26974, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5578},
        {26974}
    },
    {
        5579,
        {26973, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5579},
        {26973}
    },
    {
        5580,
        {27009, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5580},
        {27009}
    },
    {
        5581,
        {26986, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5581},
        {26986}
    },
    {
        5582,
        {27058, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5582},
        {27058}
    },
    {
        5583,
        {27054, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5583},
        {27054}
    },
    {
        5584,
        {27088, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5584},
        {27088}
    },
    {
        5585,
        {27071, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5585},
        {27071}
    },
    {
        5586,
        {27073, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5586},
        {27073}
    },
    {
        5587,
        {27091, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5587},
        {27091}
    },
    {
        5588,
        {27070, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5588},
        {27070}
    },
    {
        5589,
        {27086, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5589},
        {27086}
    },
    {
        5590,
        {23528, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5590},
        {23528}
    },
    {
        5591,
        {27082, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5591},
        {27082}
    },
    {
        5592,
        {27101, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5592},
        {27101}
    },
    {
        5593,
        {27067, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5593},
        {27067}
    },
    {
        5594,
        {27075, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5594},
        {27075}
    },
    {
        5595,
        {27047, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5595},
        {27047}
    },
    {
        5596,
        {27182, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5596},
        {27182}
    },
    {
        5597,
        {27025, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5597},
        {27025}
    },
    {
        5598,
        {27040, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5598},
        {27040}
    },
    {
        5599,
        {27036, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5599},
        {27036}
    },
    {
        5600,
        {27029, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5600},
        {27029}
    },
    {
        5601,
        {27060, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5601},
        {27060}
    },
    {
        5602,
        {27102, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5602},
        {27102}
    },
    {
        5603,
        {27112, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5603},
        {27112}
    },
    {
        5604,
        {27138, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5604},
        {27138}
    },
    {
        5605,
        {27163, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5605},
        {27163}
    },
    {
        5606,
        {27135, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5606},
        {27135}
    },
    {
        5607,
        {27402, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5607},
        {27402}
    },
    {
        5608,
        {27129, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5608},
        {27129}
    },
    {
        5609,
        {27122, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5609},
        {27122}
    },
    {
        5610,
        {27111, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5610},
        {27111}
    },
    {
        5611,
        {27141, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5611},
        {27141}
    },
    {
        5612,
        {27057, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5612},
        {27057}
    },
    {
        5613,
        {27166, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5613},
        {27166}
    },
    {
        5614,
        {27117, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5614},
        {27117}
    },
    {
        5615,
        {27156, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5615},
        {27156}
    },
    {
        5616,
        {27115, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5616},
        {27115}
    },
    {
        5617,
        {27146, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5617},
        {27146}
    },
    {
        5618,
        {27154, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5618},
        {27154}
    },
    {
        5619,
        {27329, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5619},
        {27329}
    },
    {
        5620,
        {27171, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5620},
        {27171}
    },
    {
        5621,
        {27155, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5621},
        {27155}
    },
    {
        5622,
        {27204, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5622},
        {27204}
    },
    {
        5623,
        {27148, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5623},
        {27148}
    },
    {
        5624,
        {27250, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5624},
        {27250}
    },
    {
        5625,
        {27190, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5625},
        {27190}
    },
    {
        5626,
        {27256, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5626},
        {27256}
    },
    {
        5627,
        {27207, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5627},
        {27207}
    },
    {
        5628,
        {27234, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5628},
        {27234}
    },
    {
        5629,
        {27225, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5629},
        {27225}
    },
    {
        5630,
        {27238, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5630},
        {27238}
    },
    {
        5631,
        {27208, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5631},
        {27208}
    },
    {
        5632,
        {27192, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5632},
        {27192}
    },
    {
        5633,
        {27170, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5633},
        {27170}
    },
    {
        5634,
        {27280, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5634},
        {27280}
    },
    {
        5635,
        {27277, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5635},
        {27277}
    },
    {
        5636,
        {27296, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5636},
        {27296}
    },
    {
        5637,
        {27268, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5637},
        {27268}
    },
    {
        5638,
        {27298, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5638},
        {27298}
    },
    {
        5639,
        {27299, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5639},
        {27299}
    },
    {
        5640,
        {27287, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5640},
        {27287}
    },
    {
        5641,
        {34327, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5641},
        {34327}
    },
    {
        5642,
        {27323, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5642},
        {27323}
    },
    {
        5643,
        {27331, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5643},
        {27331}
    },
    {
        5644,
        {27330, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5644},
        {27330}
    },
    {
        5645,
        {27320, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5645},
        {27320}
    },
    {
        5646,
        {27315, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5646},
        {27315}
    },
    {
        5647,
        {27308, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5647},
        {27308}
    },
    {
        5648,
        {27358, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5648},
        {27358}
    },
    {
        5649,
        {27345, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5649},
        {27345}
    },
    {
        5650,
        {27359, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5650},
        {27359}
    },
    {
        5651,
        {27306, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5651},
        {27306}
    },
    {
        5652,
        {27354, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5652},
        {27354}
    },
    {
        5653,
        {27370, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5653},
        {27370}
    },
    {
        5654,
        {27387, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5654},
        {27387}
    },
    {
        5655,
        {27397, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5655},
        {27397}
    },
    {
        5656,
        {34326, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5656},
        {34326}
    },
    {
        5657,
        {27386, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5657},
        {27386}
    },
    {
        5658,
        {27410, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5658},
        {27410}
    },
    {
        5659,
        {27414, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5659},
        {27414}
    },
    {
        5660,
        {39729, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5660},
        {39729}
    },
    {
        5661,
        {27423, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5661},
        {27423}
    },
    {
        5662,
        {27448, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5662},
        {27448}
    },
    {
        5663,
        {27447, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5663},
        {27447}
    },
    {
        5664,
        {30428, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5664},
        {30428}
    },
    {
        5665,
        {27449, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5665},
        {27449}
    },
    {
        5666,
        {39150, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5666},
        {39150}
    },
    {
        5667,
        {27463, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5667},
        {27463}
    },
    {
        5668,
        {27459, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5668},
        {27459}
    },
    {
        5669,
        {27465, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5669},
        {27465}
    },
    {
        5670,
        {27472, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5670},
        {27472}
    },
    {
        5671,
        {27481, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5671},
        {27481}
    },
    {
        5672,
        {27476, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5672},
        {27476}
    },
    {
        5673,
        {27483, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5673},
        {27483}
    },
    {
        5674,
        {27487, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5674},
        {27487}
    },
    {
        5675,
        {27489, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5675},
        {27489}
    },
    {
        5676,
        {27512, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5676},
        {27512}
    },
    {
        5677,
        {27513, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5677},
        {27513}
    },
    {
        5678,
        {27519, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5678},
        {27519}
    },
    {
        5679,
        {27520, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5679},
        {27520}
    },
    {
        5680,
        {27524, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5680},
        {27524}
    },
    {
        5681,
        {27523, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5681},
        {27523}
    },
    {
        5682,
        {27533, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5682},
        {27533}
    },
    {
        5683,
        {27544, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5683},
        {27544}
    },
    {
        5684,
        {27541, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5684},
        {27541}
    },
    {
        5685,
        {27550, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5685},
        {27550}
    },
    {
        5686,
        {27556, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5686},
        {27556}
    },
    {
        5687,
        {27562, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5687},
        {27562}
    },
    {
        5688,
        {27563, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5688},
        {27563}
    },
    {
        5689,
        {27567, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5689},
        {27567}
    },
    {
        5690,
        {27570, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5690},
        {27570}
    },
    {
        5691,
        {27569, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5691},
        {27569}
    },
    {
        5692,
        {27571, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5692},
        {27571}
    },
    {
        5693,
        {27575, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5693},
        {27575}
    },
    {
        5694,
        {27580, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5694},
        {27580}
    },
    {
        5695,
        {27590, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5695},
        {27590}
    },
    {
        5696,
        {27595, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5696},
        {27595}
    },
    {
        5697,
        {27603, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5697},
        {27603}
    },
    {
        5698,
        {27615, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5698},
        {27615}
    },
    {
        5699,
        {27628, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5699},
        {27628}
    },
    {
        5700,
        {27627, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5700},
        {27627}
    },
    {
        5701,
        {27635, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5701},
        {27635}
    },
    {
        5702,
        {27631, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5702},
        {27631}
    },
    {
        5703,
        {40638, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5703},
        {40638}
    },
    {
        5704,
        {27656, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5704},
        {27656}
    },
    {
        5705,
        {27667, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5705},
        {27667}
    },
    {
        5706,
        {27668, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5706},
        {27668}
    },
    {
        5707,
        {27675, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5707},
        {27675}
    },
    {
        5708,
        {27684, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5708},
        {27684}
    },
    {
        5709,
        {27683, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5709},
        {27683}
    },
    {
        5710,
        {27742, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5710},
        {27742}
    },
    {
        5711,
        {27733, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5711},
        {27733}
    },
    {
        5712,
        {27746, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5712},
        {27746}
    },
    {
        5713,
        {27754, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5713},
        {27754}
    },
    {
        5714,
        {27778, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5714},
        {27778}
    },
    {
        5715,
        {27789, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5715},
        {27789}
    },
    {
        5716,
        {27802, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5716},
        {27802}
    },
    {
        5717,
        {27777, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5717},
        {27777}
    },
    {
        5718,
        {27803, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5718},
        {27803}
    },
    {
        5719,
        {27774, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5719},
        {27774}
    },
    {
        5720,
        {27752, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5720},
        {27752}
    },
    {
        5721,
        {27763, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5721},
        {27763}
    },
    {
        5722,
        {27794, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5722},
        {27794}
    },
    {
        5723,
        {27792, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5723},
        {27792}
    },
    {
        5724,
        {27844, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5724},
        {27844}
    },
    {
        5725,
        {27889, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5725},
        {27889}
    },
    {
        5726,
        {27859, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5726},
        {27859}
    },
    {
        5727,
        {27837, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5727},
        {27837}
    },
    {
        5728,
        {27863, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5728},
        {27863}
    },
    {
        5729,
        {27845, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5729},
        {27845}
    },
    {
        5730,
        {27869, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5730},
        {27869}
    },
    {
        5731,
        {27822, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5731},
        {27822}
    },
    {
        5732,
        {27825, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5732},
        {27825}
    },
    {
        5733,
        {27838, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5733},
        {27838}
    },
    {
        5734,
        {27834, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5734},
        {27834}
    },
    {
        5735,
        {27867, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5735},
        {27867}
    },
    {
        5736,
        {27887, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5736},
        {27887}
    },
    {
        5737,
        {27865, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5737},
        {27865}
    },
    {
        5738,
        {27882, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5738},
        {27882}
    },
    {
        5739,
        {27935, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5739},
        {27935}
    },
    {
        5740,
        {34893, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5740},
        {34893}
    },
    {
        5741,
        {27958, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5741},
        {27958}
    },
    {
        5742,
        {27947, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5742},
        {27947}
    },
    {
        5743,
        {27965, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5743},
        {27965}
    },
    {
        5744,
        {27960, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5744},
        {27960}
    },
    {
        5745,
        {27929, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5745},
        {27929}
    },
    {
        5746,
        {27957, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5746},
        {27957}
    },
    {
        5747,
        {27955, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5747},
        {27955}
    },
    {
        5748,
        {27922, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5748},
        {27922}
    },
    {
        5749,
        {27916, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5749},
        {27916}
    },
    {
        5750,
        {28003, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5750},
        {28003}
    },
    {
        5751,
        {28051, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5751},
        {28051}
    },
    {
        5752,
        {28004, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5752},
        {28004}
    },
    {
        5753,
        {27994, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5753},
        {27994}
    },
    {
        5754,
        {28025, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5754},
        {28025}
    },
    {
        5755,
        {27993, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5755},
        {27993}
    },
    {
        5756,
        {28046, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5756},
        {28046}
    },
    {
        5757,
        {28053, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5757},
        {28053}
    },
    {
        5758,
        {28644, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5758},
        {28644}
    },
    {
        5759,
        {28037, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5759},
        {28037}
    },
    {
        5760,
        {28153, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5760},
        {28153}
    },
    {
        5761,
        {28181, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5761},
        {28181}
    },
    {
        5762,
        {28170, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5762},
        {28170}
    },
    {
        5763,
        {28085, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5763},
        {28085}
    },
    {
        5764,
        {28103, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5764},
        {28103}
    },
    {
        5765,
        {28134, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5765},
        {28134}
    },
    {
        5766,
        {28088, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5766},
        {28088}
    },
    {
        5767,
        {28102, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5767},
        {28102}
    },
    {
        5768,
        {28140, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5768},
        {28140}
    },
    {
        5769,
        {28126, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5769},
        {28126}
    },
    {
        5770,
        {28108, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5770},
        {28108}
    },
    {
        5771,
        {28136, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5771},
        {28136}
    },
    {
        5772,
        {28114, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5772},
        {28114}
    },
    {
        5773,
        {28101, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5773},
        {28101}
    },
    {
        5774,
        {28154, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5774},
        {28154}
    },
    {
        5775,
        {28121, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5775},
        {28121}
    },
    {
        5776,
        {28132, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5776},
        {28132}
    },
    {
        5777,
        {28117, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5777},
        {28117}
    },
    {
        5778,
        {28138, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5778},
        {28138}
    },
    {
        5779,
        {28142, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5779},
        {28142}
    },
    {
        5780,
        {28205, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5780},
        {28205}
    },
    {
        5781,
        {28270, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5781},
        {28270}
    },
    {
        5782,
        {28206, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5782},
        {28206}
    },
    {
        5783,
        {28185, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5783},
        {28185}
    },
    {
        5784,
        {28274, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5784},
        {28274}
    },
    {
        5785,
        {28255, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5785},
        {28255}
    },
    {
        5786,
        {28222, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5786},
        {28222}
    },
    {
        5787,
        {28195, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5787},
        {28195}
    },
    {
        5788,
        {28267, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5788},
        {28267}
    },
    {
        5789,
        {28203, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5789},
        {28203}
    },
    {
        5790,
        {28278, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5790},
        {28278}
    },
    {
        5791,
        {28237, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5791},
        {28237}
    },
    {
        5792,
        {28191, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5792},
        {28191}
    },
    {
        5793,
        {28227, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5793},
        {28227}
    },
    {
        5794,
        {28218, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5794},
        {28218}
    },
    {
        5795,
        {28238, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5795},
        {28238}
    },
    {
        5796,
        {28196, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5796},
        {28196}
    },
    {
        5797,
        {28415, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5797},
        {28415}
    },
    {
        5798,
        {28189, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5798},
        {28189}
    },
    {
        5799,
        {28216, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5799},
        {28216}
    },
    {
        5800,
        {28290, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5800},
        {28290}
    },
    {
        5801,
        {28330, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5801},
        {28330}
    },
    {
        5802,
        {28312, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5802},
        {28312}
    },
    {
        5803,
        {28361, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5803},
        {28361}
    },
    {
        5804,
        {28343, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5804},
        {28343}
    },
    {
        5805,
        {28371, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5805},
        {28371}
    },
    {
        5806,
        {28349, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5806},
        {28349}
    },
    {
        5807,
        {28335, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5807},
        {28335}
    },
    {
        5808,
        {28356, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5808},
        {28356}
    },
    {
        5809,
        {28338, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5809},
        {28338}
    },
    {
        5810,
        {28372, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5810},
        {28372}
    },
    {
        5811,
        {28373, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5811},
        {28373}
    },
    {
        5812,
        {28303, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5812},
        {28303}
    },
    {
        5813,
        {28325, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5813},
        {28325}
    },
    {
        5814,
        {28354, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5814},
        {28354}
    },
    {
        5815,
        {28319, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5815},
        {28319}
    },
    {
        5816,
        {28481, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5816},
        {28481}
    },
    {
        5817,
        {28433, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5817},
        {28433}
    },
    {
        5818,
        {28748, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5818},
        {28748}
    },
    {
        5819,
        {28396, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5819},
        {28396}
    },
    {
        5820,
        {28408, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5820},
        {28408}
    },
    {
        5821,
        {28414, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5821},
        {28414}
    },
    {
        5822,
        {28479, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5822},
        {28479}
    },
    {
        5823,
        {28402, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5823},
        {28402}
    },
    {
        5824,
        {28465, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5824},
        {28465}
    },
    {
        5825,
        {28399, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5825},
        {28399}
    },
    {
        5826,
        {28466, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5826},
        {28466}
    },
    {
        5827,
        {28364, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5827},
        {28364}
    },
    {
        5828,
        {28478, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5828},
        {28478}
    },
    {
        5829,
        {28435, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5829},
        {28435}
    },
    {
        5830,
        {28407, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5830},
        {28407}
    },
    {
        5831,
        {28550, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5831},
        {28550}
    },
    {
        5832,
        {28538, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5832},
        {28538}
    },
    {
        5833,
        {28536, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5833},
        {28536}
    },
    {
        5834,
        {28545, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5834},
        {28545}
    },
    {
        5835,
        {28544, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5835},
        {28544}
    },
    {
        5836,
        {28527, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5836},
        {28527}
    },
    {
        5837,
        {28507, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5837},
        {28507}
    },
    {
        5838,
        {28659, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5838},
        {28659}
    },
    {
        5839,
        {28525, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5839},
        {28525}
    },
    {
        5840,
        {28546, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5840},
        {28546}
    },
    {
        5841,
        {28540, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5841},
        {28540}
    },
    {
        5842,
        {28504, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5842},
        {28504}
    },
    {
        5843,
        {28558, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5843},
        {28558}
    },
    {
        5844,
        {28561, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5844},
        {28561}
    },
    {
        5845,
        {28610, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5845},
        {28610}
    },
    {
        5846,
        {28518, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5846},
        {28518}
    },
    {
        5847,
        {28595, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5847},
        {28595}
    },
    {
        5848,
        {28579, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5848},
        {28579}
    },
    {
        5849,
        {28577, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5849},
        {28577}
    },
    {
        5850,
        {28580, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5850},
        {28580}
    },
    {
        5851,
        {28601, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5851},
        {28601}
    },
    {
        5852,
        {28614, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5852},
        {28614}
    },
    {
        5853,
        {28586, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5853},
        {28586}
    },
    {
        5854,
        {28639, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5854},
        {28639}
    },
    {
        5855,
        {28629, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5855},
        {28629}
    },
    {
        5856,
        {28652, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5856},
        {28652}
    },
    {
        5857,
        {28628, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5857},
        {28628}
    },
    {
        5858,
        {28632, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5858},
        {28632}
    },
    {
        5859,
        {28657, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5859},
        {28657}
    },
    {
        5860,
        {28654, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5860},
        {28654}
    },
    {
        5861,
        {28635, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5861},
        {28635}
    },
    {
        5862,
        {28681, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5862},
        {28681}
    },
    {
        5863,
        {28683, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5863},
        {28683}
    },
    {
        5864,
        {28666, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5864},
        {28666}
    },
    {
        5865,
        {28689, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5865},
        {28689}
    },
    {
        5866,
        {28673, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5866},
        {28673}
    },
    {
        5867,
        {28687, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5867},
        {28687}
    },
    {
        5868,
        {28670, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5868},
        {28670}
    },
    {
        5869,
        {28699, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5869},
        {28699}
    },
    {
        5870,
        {28698, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5870},
        {28698}
    },
    {
        5871,
        {28532, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5871},
        {28532}
    },
    {
        5872,
        {28701, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5872},
        {28701}
    },
    {
        5873,
        {28696, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5873},
        {28696}
    },
    {
        5874,
        {28703, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5874},
        {28703}
    },
    {
        5875,
        {28720, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5875},
        {28720}
    },
    {
        5876,
        {28734, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5876},
        {28734}
    },
    {
        5877,
        {28722, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5877},
        {28722}
    },
    {
        5878,
        {28753, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5878},
        {28753}
    },
    {
        5879,
        {28771, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5879},
        {28771}
    },
    {
        5880,
        {28825, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5880},
        {28825}
    },
    {
        5881,
        {28818, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5881},
        {28818}
    },
    {
        5882,
        {28847, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5882},
        {28847}
    },
    {
        5883,
        {28913, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5883},
        {28913}
    },
    {
        5884,
        {28844, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5884},
        {28844}
    },
    {
        5885,
        {28856, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5885},
        {28856}
    },
    {
        5886,
        {28851, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5886},
        {28851}
    },
    {
        5887,
        {28846, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5887},
        {28846}
    },
    {
        5888,
        {28895, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5888},
        {28895}
    },
    {
        5889,
        {28875, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5889},
        {28875}
    },
    {
        5890,
        {28893, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5890},
        {28893}
    },
    {
        5891,
        {28889, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5891},
        {28889}
    },
    {
        5892,
        {28937, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5892},
        {28937}
    },
    {
        5893,
        {28925, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5893},
        {28925}
    },
    {
        5894,
        {28956, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5894},
        {28956}
    },
    {
        5895,
        {28953, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5895},
        {28953}
    },
    {
        5896,
        {29029, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5896},
        {29029}
    },
    {
        5897,
        {29013, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5897},
        {29013}
    },
    {
        5898,
        {29064, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5898},
        {29064}
    },
    {
        5899,
        {29030, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5899},
        {29030}
    },
    {
        5900,
        {29026, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5900},
        {29026}
    },
    {
        5901,
        {29004, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5901},
        {29004}
    },
    {
        5902,
        {29014, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5902},
        {29014}
    },
    {
        5903,
        {29036, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5903},
        {29036}
    },
    {
        5904,
        {29071, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5904},
        {29071}
    },
    {
        5905,
        {29179, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5905},
        {29179}
    },
    {
        5906,
        {29060, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5906},
        {29060}
    },
    {
        5907,
        {29077, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5907},
        {29077}
    },
    {
        5908,
        {29096, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5908},
        {29096}
    },
    {
        5909,
        {29100, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5909},
        {29100}
    },
    {
        5910,
        {29143, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5910},
        {29143}
    },
    {
        5911,
        {29113, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5911},
        {29113}
    },
    {
        5912,
        {29118, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5912},
        {29118}
    },
    {
        5913,
        {29138, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5913},
        {29138}
    },
    {
        5914,
        {29129, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5914},
        {29129}
    },
    {
        5915,
        {29140, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5915},
        {29140}
    },
    {
        5916,
        {29134, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5916},
        {29134}
    },
    {
        5917,
        {29152, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5917},
        {29152}
    },
    {
        5918,
        {29164, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5918},
        {29164}
    },
    {
        5919,
        {29159, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5919},
        {29159}
    },
    {
        5920,
        {29173, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5920},
        {29173}
    },
    {
        5921,
        {29180, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5921},
        {29180}
    },
    {
        5922,
        {29177, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5922},
        {29177}
    },
    {
        5923,
        {29183, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5923},
        {29183}
    },
    {
        5924,
        {29197, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5924},
        {29197}
    },
    {
        5925,
        {29200, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5925},
        {29200}
    },
    {
        5926,
        {29211, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5926},
        {29211}
    },
    {
        5927,
        {29224, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5927},
        {29224}
    },
    {
        5928,
        {29229, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5928},
        {29229}
    },
    {
        5929,
        {29228, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5929},
        {29228}
    },
    {
        5930,
        {29232, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5930},
        {29232}
    },
    {
        5931,
        {29234, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5931},
        {29234}
    },
    {
        5932,
        {29243, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5932},
        {29243}
    },
    {
        5933,
        {29244, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5933},
        {29244}
    },
    {
        5934,
        {29247, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5934},
        {29247}
    },
    {
        5935,
        {29248, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5935},
        {29248}
    },
    {
        5936,
        {29254, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5936},
        {29254}
    },
    {
        5937,
        {29259, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5937},
        {29259}
    },
    {
        5938,
        {29272, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5938},
        {29272}
    },
    {
        5939,
        {29300, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5939},
        {29300}
    },
    {
        5940,
        {29310, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5940},
        {29310}
    },
    {
        5941,
        {29314, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5941},
        {29314}
    },
    {
        5942,
        {29313, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5942},
        {29313}
    },
    {
        5943,
        {29319, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5943},
        {29319}
    },
    {
        5944,
        {29330, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5944},
        {29330}
    },
    {
        5945,
        {29334, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5945},
        {29334}
    },
    {
        5946,
        {29346, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5946},
        {29346}
    },
    {
        5947,
        {29351, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5947},
        {29351}
    },
    {
        5948,
        {29369, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5948},
        {29369}
    },
    {
        5949,
        {29362, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5949},
        {29362}
    },
    {
        5950,
        {29379, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5950},
        {29379}
    },
    {
        5951,
        {29382, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5951},
        {29382}
    },
    {
        5952,
        {29380, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5952},
        {29380}
    },
    {
        5953,
        {29390, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5953},
        {29390}
    },
    {
        5954,
        {29394, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5954},
        {29394}
    },
    {
        5955,
        {29410, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5955},
        {29410}
    },
    {
        5956,
        {29408, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5956},
        {29408}
    },
    {
        5957,
        {29409, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5957},
        {29409}
    },
    {
        5958,
        {29433, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5958},
        {29433}
    },
    {
        5959,
        {29431, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5959},
        {29431}
    },
    {
        5960,
        {20495, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5960},
        {20495}
    },
    {
        5961,
        {29463, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5961},
        {29463}
    },
    {
        5962,
        {29450, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5962},
        {29450}
    },
    {
        5963,
        {29468, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5963},
        {29468}
    },
    {
        5964,
        {29462, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5964},
        {29462}
    },
    {
        5965,
        {29469, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5965},
        {29469}
    },
    {
        5966,
        {29492, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5966},
        {29492}
    },
    {
        5967,
        {29487, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5967},
        {29487}
    },
    {
        5968,
        {29481, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5968},
        {29481}
    },
    {
        5969,
        {29477, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5969},
        {29477}
    },
    {
        5970,
        {29502, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5970},
        {29502}
    },
    {
        5971,
        {29518, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5971},
        {29518}
    },
    {
        5972,
        {29519, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5972},
        {29519}
    },
    {
        5973,
        {40664, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5973},
        {40664}
    },
    {
        5974,
        {29527, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5974},
        {29527}
    },
    {
        5975,
        {29546, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5975},
        {29546}
    },
    {
        5976,
        {29544, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5976},
        {29544}
    },
    {
        5977,
        {29552, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5977},
        {29552}
    },
    {
        5978,
        {29560, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5978},
        {29560}
    },
    {
        5979,
        {29557, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5979},
        {29557}
    },
    {
        5980,
        {29563, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5980},
        {29563}
    },
    {
        5981,
        {29562, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5981},
        {29562}
    },
    {
        5982,
        {29640, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5982},
        {29640}
    },
    {
        5983,
        {29619, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5983},
        {29619}
    },
    {
        5984,
        {29646, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5984},
        {29646}
    },
    {
        5985,
        {29627, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5985},
        {29627}
    },
    {
        5986,
        {29632, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5986},
        {29632}
    },
    {
        5987,
        {29669, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5987},
        {29669}
    },
    {
        5988,
        {29678, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5988},
        {29678}
    },
    {
        5989,
        {29662, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5989},
        {29662}
    },
    {
        5990,
        {29858, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5990},
        {29858}
    },
    {
        5991,
        {29701, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5991},
        {29701}
    },
    {
        5992,
        {29807, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5992},
        {29807}
    },
    {
        5993,
        {29733, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5993},
        {29733}
    },
    {
        5994,
        {29688, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5994},
        {29688}
    },
    {
        5995,
        {29746, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5995},
        {29746}
    },
    {
        5996,
        {29754, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5996},
        {29754}
    },
    {
        5997,
        {29781, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5997},
        {29781}
    },
    {
        5998,
        {29759, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5998},
        {29759}
    },
    {
        5999,
        {29791, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {5999},
        {29791}
    },
    {
        6000,
        {29785, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6000},
        {29785}
    },
    {
        6001,
        {29761, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6001},
        {29761}
    },
    {
        6002,
        {29788, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6002},
        {29788}
    },
    {
        6003,
        {29801, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6003},
        {29801}
    },
    {
        6004,
        {29808, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6004},
        {29808}
    },
    {
        6005,
        {29795, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6005},
        {29795}
    },
    {
        6006,
        {29802, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6006},
        {29802}
    },
    {
        6007,
        {29814, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6007},
        {29814}
    },
    {
        6008,
        {29822, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6008},
        {29822}
    },
    {
        6009,
        {29835, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6009},
        {29835}
    },
    {
        6010,
        {29854, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6010},
        {29854}
    },
    {
        6011,
        {29863, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6011},
        {29863}
    },
    {
        6012,
        {29898, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6012},
        {29898}
    },
    {
        6013,
        {29903, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6013},
        {29903}
    },
    {
        6014,
        {29908, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6014},
        {29908}
    },
    {
        6015,
        {29681, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6015},
        {29681}
    },
    {
        6016,
        {29920, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6016},
        {29920}
    },
    {
        6017,
        {29923, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6017},
        {29923}
    },
    {
        6018,
        {29927, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6018},
        {29927}
    },
    {
        6019,
        {29929, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6019},
        {29929}
    },
    {
        6020,
        {29934, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6020},
        {29934}
    },
    {
        6021,
        {29938, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6021},
        {29938}
    },
    {
        6022,
        {29936, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6022},
        {29936}
    },
    {
        6023,
        {29937, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6023},
        {29937}
    },
    {
        6024,
        {29944, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6024},
        {29944}
    },
    {
        6025,
        {29943, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6025},
        {29943}
    },
    {
        6026,
        {29956, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6026},
        {29956}
    },
    {
        6027,
        {29955, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6027},
        {29955}
    },
    {
        6028,
        {29957, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6028},
        {29957}
    },
    {
        6029,
        {29964, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6029},
        {29964}
    },
    {
        6030,
        {29966, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6030},
        {29966}
    },
    {
        6031,
        {29965, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6031},
        {29965}
    },
    {
        6032,
        {29973, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6032},
        {29973}
    },
    {
        6033,
        {29971, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6033},
        {29971}
    },
    {
        6034,
        {29982, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6034},
        {29982}
    },
    {
        6035,
        {29990, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6035},
        {29990}
    },
    {
        6036,
        {29996, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6036},
        {29996}
    },
    {
        6037,
        {30012, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6037},
        {30012}
    },
    {
        6038,
        {30020, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6038},
        {30020}
    },
    {
        6039,
        {30029, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6039},
        {30029}
    },
    {
        6040,
        {30026, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6040},
        {30026}
    },
    {
        6041,
        {30025, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6041},
        {30025}
    },
    {
        6042,
        {30043, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6042},
        {30043}
    },
    {
        6043,
        {30022, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6043},
        {30022}
    },
    {
        6044,
        {30042, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6044},
        {30042}
    },
    {
        6045,
        {30057, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6045},
        {30057}
    },
    {
        6046,
        {30052, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6046},
        {30052}
    },
    {
        6047,
        {30055, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6047},
        {30055}
    },
    {
        6048,
        {30059, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6048},
        {30059}
    },
    {
        6049,
        {30061, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6049},
        {30061}
    },
    {
        6050,
        {30072, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6050},
        {30072}
    },
    {
        6051,
        {30070, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6051},
        {30070}
    },
    {
        6052,
        {30086, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6052},
        {30086}
    },
    {
        6053,
        {30087, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6053},
        {30087}
    },
    {
        6054,
        {30068, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6054},
        {30068}
    },
    {
        6055,
        {30090, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6055},
        {30090}
    },
    {
        6056,
        {30089, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6056},
        {30089}
    },
    {
        6057,
        {30082, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6057},
        {30082}
    },
    {
        6058,
        {30100, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6058},
        {30100}
    },
    {
        6059,
        {30106, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6059},
        {30106}
    },
    {
        6060,
        {30109, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6060},
        {30109}
    },
    {
        6061,
        {30117, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6061},
        {30117}
    },
    {
        6062,
        {30115, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6062},
        {30115}
    },
    {
        6063,
        {30146, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6063},
        {30146}
    },
    {
        6064,
        {30131, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6064},
        {30131}
    },
    {
        6065,
        {30147, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6065},
        {30147}
    },
    {
        6066,
        {30133, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6066},
        {30133}
    },
    {
        6067,
        {30141, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6067},
        {30141}
    },
    {
        6068,
        {30136, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6068},
        {30136}
    },
    {
        6069,
        {30140, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6069},
        {30140}
    },
    {
        6070,
        {30129, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6070},
        {30129}
    },
    {
        6071,
        {30157, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6071},
        {30157}
    },
    {
        6072,
        {30154, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6072},
        {30154}
    },
    {
        6073,
        {30162, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6073},
        {30162}
    },
    {
        6074,
        {30169, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6074},
        {30169}
    },
    {
        6075,
        {30179, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6075},
        {30179}
    },
    {
        6076,
        {30174, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6076},
        {30174}
    },
    {
        6077,
        {30206, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6077},
        {30206}
    },
    {
        6078,
        {30207, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6078},
        {30207}
    },
    {
        6079,
        {30204, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6079},
        {30204}
    },
    {
        6080,
        {30209, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6080},
        {30209}
    },
    {
        6081,
        {30192, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6081},
        {30192}
    },
    {
        6082,
        {30202, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6082},
        {30202}
    },
    {
        6083,
        {30194, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6083},
        {30194}
    },
    {
        6084,
        {30195, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6084},
        {30195}
    },
    {
        6085,
        {30219, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6085},
        {30219}
    },
    {
        6086,
        {30221, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6086},
        {30221}
    },
    {
        6087,
        {30217, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6087},
        {30217}
    },
    {
        6088,
        {30239, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6088},
        {30239}
    },
    {
        6089,
        {30247, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6089},
        {30247}
    },
    {
        6090,
        {30240, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6090},
        {30240}
    },
    {
        6091,
        {30241, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6091},
        {30241}
    },
    {
        6092,
        {30242, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6092},
        {30242}
    },
    {
        6093,
        {30244, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6093},
        {30244}
    },
    {
        6094,
        {30260, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6094},
        {30260}
    },
    {
        6095,
        {30256, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6095},
        {30256}
    },
    {
        6096,
        {30267, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6096},
        {30267}
    },
    {
        6097,
        {30279, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6097},
        {30279}
    },
    {
        6098,
        {30280, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6098},
        {30280}
    },
    {
        6099,
        {30278, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6099},
        {30278}
    },
    {
        6100,
        {30300, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6100},
        {30300}
    },
    {
        6101,
        {30296, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6101},
        {30296}
    },
    {
        6102,
        {30305, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6102},
        {30305}
    },
    {
        6103,
        {30306, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6103},
        {30306}
    },
    {
        6104,
        {30312, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6104},
        {30312}
    },
    {
        6105,
        {30313, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6105},
        {30313}
    },
    {
        6106,
        {30314, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6106},
        {30314}
    },
    {
        6107,
        {30311, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6107},
        {30311}
    },
    {
        6108,
        {30316, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6108},
        {30316}
    },
    {
        6109,
        {30320, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6109},
        {30320}
    },
    {
        6110,
        {30322, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6110},
        {30322}
    },
    {
        6111,
        {30326, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6111},
        {30326}
    },
    {
        6112,
        {30328, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6112},
        {30328}
    },
    {
        6113,
        {30332, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6113},
        {30332}
    },
    {
        6114,
        {30336, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6114},
        {30336}
    },
    {
        6115,
        {30339, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6115},
        {30339}
    },
    {
        6116,
        {30344, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6116},
        {30344}
    },
    {
        6117,
        {30347, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6117},
        {30347}
    },
    {
        6118,
        {30350, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6118},
        {30350}
    },
    {
        6119,
        {30358, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6119},
        {30358}
    },
    {
        6120,
        {30355, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6120},
        {30355}
    },
    {
        6121,
        {30361, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6121},
        {30361}
    },
    {
        6122,
        {30362, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6122},
        {30362}
    },
    {
        6123,
        {30384, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6123},
        {30384}
    },
    {
        6124,
        {30388, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6124},
        {30388}
    },
    {
        6125,
        {30392, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6125},
        {30392}
    },
    {
        6126,
        {30393, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6126},
        {30393}
    },
    {
        6127,
        {30394, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6127},
        {30394}
    },
    {
        6128,
        {30402, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6128},
        {30402}
    },
    {
        6129,
        {30413, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6129},
        {30413}
    },
    {
        6130,
        {30422, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6130},
        {30422}
    },
    {
        6131,
        {30418, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6131},
        {30418}
    },
    {
        6132,
        {30430, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6132},
        {30430}
    },
    {
        6133,
        {30433, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6133},
        {30433}
    },
    {
        6134,
        {30437, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6134},
        {30437}
    },
    {
        6135,
        {30439, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6135},
        {30439}
    },
    {
        6136,
        {30442, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6136},
        {30442}
    },
    {
        6137,
        {34351, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6137},
        {34351}
    },
    {
        6138,
        {30459, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6138},
        {30459}
    },
    {
        6139,
        {30472, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6139},
        {30472}
    },
    {
        6140,
        {30471, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6140},
        {30471}
    },
    {
        6141,
        {30468, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6141},
        {30468}
    },
    {
        6142,
        {30505, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6142},
        {30505}
    },
    {
        6143,
        {30500, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6143},
        {30500}
    },
    {
        6144,
        {30494, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6144},
        {30494}
    },
    {
        6145,
        {30501, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6145},
        {30501}
    },
    {
        6146,
        {30502, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6146},
        {30502}
    },
    {
        6147,
        {30491, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6147},
        {30491}
    },
    {
        6148,
        {30519, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6148},
        {30519}
    },
    {
        6149,
        {30520, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6149},
        {30520}
    },
    {
        6150,
        {30535, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6150},
        {30535}
    },
    {
        6151,
        {30554, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6151},
        {30554}
    },
    {
        6152,
        {30568, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6152},
        {30568}
    },
    {
        6153,
        {30571, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6153},
        {30571}
    },
    {
        6154,
        {30555, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6154},
        {30555}
    },
    {
        6155,
        {30565, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6155},
        {30565}
    },
    {
        6156,
        {30591, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6156},
        {30591}
    },
    {
        6157,
        {30590, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6157},
        {30590}
    },
    {
        6158,
        {30585, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6158},
        {30585}
    },
    {
        6159,
        {30606, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6159},
        {30606}
    },
    {
        6160,
        {30603, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6160},
        {30603}
    },
    {
        6161,
        {30609, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6161},
        {30609}
    },
    {
        6162,
        {30624, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6162},
        {30624}
    },
    {
        6163,
        {30622, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6163},
        {30622}
    },
    {
        6164,
        {30640, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6164},
        {30640}
    },
    {
        6165,
        {30646, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6165},
        {30646}
    },
    {
        6166,
        {30649, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6166},
        {30649}
    },
    {
        6167,
        {30655, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6167},
        {30655}
    },
    {
        6168,
        {30652, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6168},
        {30652}
    },
    {
        6169,
        {30653, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6169},
        {30653}
    },
    {
        6170,
        {30651, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6170},
        {30651}
    },
    {
        6171,
        {30663, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6171},
        {30663}
    },
    {
        6172,
        {30669, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6172},
        {30669}
    },
    {
        6173,
        {30679, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6173},
        {30679}
    },
    {
        6174,
        {30682, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6174},
        {30682}
    },
    {
        6175,
        {30684, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6175},
        {30684}
    },
    {
        6176,
        {30691, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6176},
        {30691}
    },
    {
        6177,
        {30702, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6177},
        {30702}
    },
    {
        6178,
        {30716, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6178},
        {30716}
    },
    {
        6179,
        {30732, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6179},
        {30732}
    },
    {
        6180,
        {30738, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6180},
        {30738}
    },
    {
        6181,
        {31014, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6181},
        {31014}
    },
    {
        6182,
        {30752, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6182},
        {30752}
    },
    {
        6183,
        {31018, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6183},
        {31018}
    },
    {
        6184,
        {30789, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6184},
        {30789}
    },
    {
        6185,
        {30862, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6185},
        {30862}
    },
    {
        6186,
        {30836, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6186},
        {30836}
    },
    {
        6187,
        {30854, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6187},
        {30854}
    },
    {
        6188,
        {30844, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6188},
        {30844}
    },
    {
        6189,
        {30874, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6189},
        {30874}
    },
    {
        6190,
        {30860, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6190},
        {30860}
    },
    {
        6191,
        {30883, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6191},
        {30883}
    },
    {
        6192,
        {30901, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6192},
        {30901}
    },
    {
        6193,
        {30890, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6193},
        {30890}
    },
    {
        6194,
        {30895, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6194},
        {30895}
    },
    {
        6195,
        {30929, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6195},
        {30929}
    },
    {
        6196,
        {30918, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6196},
        {30918}
    },
    {
        6197,
        {30923, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6197},
        {30923}
    },
    {
        6198,
        {30932, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6198},
        {30932}
    },
    {
        6199,
        {30910, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6199},
        {30910}
    },
    {
        6200,
        {30908, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6200},
        {30908}
    },
    {
        6201,
        {30917, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6201},
        {30917}
    },
    {
        6202,
        {30922, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6202},
        {30922}
    },
    {
        6203,
        {30956, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6203},
        {30956}
    },
    {
        6204,
        {30951, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6204},
        {30951}
    },
    {
        6205,
        {30938, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6205},
        {30938}
    },
    {
        6206,
        {30973, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6206},
        {30973}
    },
    {
        6207,
        {30964, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6207},
        {30964}
    },
    {
        6208,
        {30983, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6208},
        {30983}
    },
    {
        6209,
        {30994, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6209},
        {30994}
    },
    {
        6210,
        {30993, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6210},
        {30993}
    },
    {
        6211,
        {31001, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6211},
        {31001}
    },
    {
        6212,
        {31020, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6212},
        {31020}
    },
    {
        6213,
        {31019, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6213},
        {31019}
    },
    {
        6214,
        {31040, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6214},
        {31040}
    },
    {
        6215,
        {31072, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6215},
        {31072}
    },
    {
        6216,
        {31063, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6216},
        {31063}
    },
    {
        6217,
        {31071, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6217},
        {31071}
    },
    {
        6218,
        {31066, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6218},
        {31066}
    },
    {
        6219,
        {31061, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6219},
        {31061}
    },
    {
        6220,
        {31059, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6220},
        {31059}
    },
    {
        6221,
        {31098, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6221},
        {31098}
    },
    {
        6222,
        {31103, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6222},
        {31103}
    },
    {
        6223,
        {31114, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6223},
        {31114}
    },
    {
        6224,
        {31133, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6224},
        {31133}
    },
    {
        6225,
        {31143, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6225},
        {31143}
    },
    {
        6226,
        {40779, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6226},
        {40779}
    },
    {
        6227,
        {31146, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6227},
        {31146}
    },
    {
        6228,
        {31150, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6228},
        {31150}
    },
    {
        6229,
        {31155, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6229},
        {31155}
    },
    {
        6230,
        {31161, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6230},
        {31161}
    },
    {
        6231,
        {31162, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6231},
        {31162}
    },
    {
        6232,
        {31177, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6232},
        {31177}
    },
    {
        6233,
        {31189, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6233},
        {31189}
    },
    {
        6234,
        {31207, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6234},
        {31207}
    },
    {
        6235,
        {31212, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6235},
        {31212}
    },
    {
        6236,
        {31201, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6236},
        {31201}
    },
    {
        6237,
        {31203, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6237},
        {31203}
    },
    {
        6238,
        {31240, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6238},
        {31240}
    },
    {
        6239,
        {31245, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6239},
        {31245}
    },
    {
        6240,
        {31256, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6240},
        {31256}
    },
    {
        6241,
        {31257, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6241},
        {31257}
    },
    {
        6242,
        {31264, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6242},
        {31264}
    },
    {
        6243,
        {31263, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6243},
        {31263}
    },
    {
        6244,
        {31104, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6244},
        {31104}
    },
    {
        6245,
        {31281, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6245},
        {31281}
    },
    {
        6246,
        {31291, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6246},
        {31291}
    },
    {
        6247,
        {31294, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6247},
        {31294}
    },
    {
        6248,
        {31287, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6248},
        {31287}
    },
    {
        6249,
        {31299, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6249},
        {31299}
    },
    {
        6250,
        {31319, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6250},
        {31319}
    },
    {
        6251,
        {31305, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6251},
        {31305}
    },
    {
        6252,
        {31329, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6252},
        {31329}
    },
    {
        6253,
        {31330, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6253},
        {31330}
    },
    {
        6254,
        {31337, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6254},
        {31337}
    },
    {
        6255,
        {40861, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6255},
        {40861}
    },
    {
        6256,
        {31344, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6256},
        {31344}
    },
    {
        6257,
        {31353, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6257},
        {31353}
    },
    {
        6258,
        {31357, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6258},
        {31357}
    },
    {
        6259,
        {31368, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6259},
        {31368}
    },
    {
        6260,
        {31383, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6260},
        {31383}
    },
    {
        6261,
        {31381, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6261},
        {31381}
    },
    {
        6262,
        {31384, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6262},
        {31384}
    },
    {
        6263,
        {31382, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6263},
        {31382}
    },
    {
        6264,
        {31401, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6264},
        {31401}
    },
    {
        6265,
        {31432, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6265},
        {31432}
    },
    {
        6266,
        {31408, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6266},
        {31408}
    },
    {
        6267,
        {31414, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6267},
        {31414}
    },
    {
        6268,
        {31429, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6268},
        {31429}
    },
    {
        6269,
        {31428, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6269},
        {31428}
    },
    {
        6270,
        {31423, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6270},
        {31423}
    },
    {
        6271,
        {36995, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6271},
        {36995}
    },
    {
        6272,
        {31431, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6272},
        {31431}
    },
    {
        6273,
        {31434, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6273},
        {31434}
    },
    {
        6274,
        {31437, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6274},
        {31437}
    },
    {
        6275,
        {31439, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6275},
        {31439}
    },
    {
        6276,
        {31445, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6276},
        {31445}
    },
    {
        6277,
        {31443, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6277},
        {31443}
    },
    {
        6278,
        {31449, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6278},
        {31449}
    },
    {
        6279,
        {31450, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6279},
        {31450}
    },
    {
        6280,
        {31453, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6280},
        {31453}
    },
    {
        6281,
        {31457, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6281},
        {31457}
    },
    {
        6282,
        {31458, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6282},
        {31458}
    },
    {
        6283,
        {31462, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6283},
        {31462}
    },
    {
        6284,
        {31469, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6284},
        {31469}
    },
    {
        6285,
        {31472, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6285},
        {31472}
    },
    {
        6286,
        {31490, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6286},
        {31490}
    },
    {
        6287,
        {31503, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6287},
        {31503}
    },
    {
        6288,
        {31498, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6288},
        {31498}
    },
    {
        6289,
        {31494, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6289},
        {31494}
    },
    {
        6290,
        {31539, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6290},
        {31539}
    },
    {
        6291,
        {31512, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6291},
        {31512}
    },
    {
        6292,
        {31513, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6292},
        {31513}
    },
    {
        6293,
        {31518, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6293},
        {31518}
    },
    {
        6294,
        {31541, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6294},
        {31541}
    },
    {
        6295,
        {31528, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6295},
        {31528}
    },
    {
        6296,
        {31542, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6296},
        {31542}
    },
    {
        6297,
        {31568, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6297},
        {31568}
    },
    {
        6298,
        {31610, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6298},
        {31610}
    },
    {
        6299,
        {31492, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6299},
        {31492}
    },
    {
        6300,
        {31565, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6300},
        {31565}
    },
    {
        6301,
        {31499, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6301},
        {31499}
    },
    {
        6302,
        {31564, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6302},
        {31564}
    },
    {
        6303,
        {31557, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6303},
        {31557}
    },
    {
        6304,
        {31605, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6304},
        {31605}
    },
    {
        6305,
        {31589, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6305},
        {31589}
    },
    {
        6306,
        {31604, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6306},
        {31604}
    },
    {
        6307,
        {31591, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6307},
        {31591}
    },
    {
        6308,
        {31600, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6308},
        {31600}
    },
    {
        6309,
        {31601, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6309},
        {31601}
    },
    {
        6310,
        {31596, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6310},
        {31596}
    },
    {
        6311,
        {31598, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6311},
        {31598}
    },
    {
        6312,
        {31645, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6312},
        {31645}
    },
    {
        6313,
        {31640, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6313},
        {31640}
    },
    {
        6314,
        {31647, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6314},
        {31647}
    },
    {
        6315,
        {31629, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6315},
        {31629}
    },
    {
        6316,
        {31644, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6316},
        {31644}
    },
    {
        6317,
        {31642, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6317},
        {31642}
    },
    {
        6318,
        {31627, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6318},
        {31627}
    },
    {
        6319,
        {31634, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6319},
        {31634}
    },
    {
        6320,
        {31631, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6320},
        {31631}
    },
    {
        6321,
        {31581, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6321},
        {31581}
    },
    {
        6322,
        {31641, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6322},
        {31641}
    },
    {
        6323,
        {31691, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6323},
        {31691}
    },
    {
        6324,
        {31681, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6324},
        {31681}
    },
    {
        6325,
        {31692, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6325},
        {31692}
    },
    {
        6326,
        {31695, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6326},
        {31695}
    },
    {
        6327,
        {31668, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6327},
        {31668}
    },
    {
        6328,
        {31686, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6328},
        {31686}
    },
    {
        6329,
        {31709, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6329},
        {31709}
    },
    {
        6330,
        {31721, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6330},
        {31721}
    },
    {
        6331,
        {31761, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6331},
        {31761}
    },
    {
        6332,
        {31764, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6332},
        {31764}
    },
    {
        6333,
        {31718, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6333},
        {31718}
    },
    {
        6334,
        {31717, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6334},
        {31717}
    },
    {
        6335,
        {31840, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6335},
        {31840}
    },
    {
        6336,
        {31744, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6336},
        {31744}
    },
    {
        6337,
        {31751, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6337},
        {31751}
    },
    {
        6338,
        {31763, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6338},
        {31763}
    },
    {
        6339,
        {31731, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6339},
        {31731}
    },
    {
        6340,
        {31735, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6340},
        {31735}
    },
    {
        6341,
        {31767, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6341},
        {31767}
    },
    {
        6342,
        {31757, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6342},
        {31757}
    },
    {
        6343,
        {31734, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6343},
        {31734}
    },
    {
        6344,
        {31779, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6344},
        {31779}
    },
    {
        6345,
        {31783, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6345},
        {31783}
    },
    {
        6346,
        {31786, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6346},
        {31786}
    },
    {
        6347,
        {31775, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6347},
        {31775}
    },
    {
        6348,
        {31799, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6348},
        {31799}
    },
    {
        6349,
        {31787, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6349},
        {31787}
    },
    {
        6350,
        {31805, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6350},
        {31805}
    },
    {
        6351,
        {31820, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6351},
        {31820}
    },
    {
        6352,
        {31811, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6352},
        {31811}
    },
    {
        6353,
        {31828, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6353},
        {31828}
    },
    {
        6354,
        {31823, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6354},
        {31823}
    },
    {
        6355,
        {31808, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6355},
        {31808}
    },
    {
        6356,
        {31824, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6356},
        {31824}
    },
    {
        6357,
        {31832, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6357},
        {31832}
    },
    {
        6358,
        {31839, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6358},
        {31839}
    },
    {
        6359,
        {31844, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6359},
        {31844}
    },
    {
        6360,
        {31830, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6360},
        {31830}
    },
    {
        6361,
        {31845, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6361},
        {31845}
    },
    {
        6362,
        {31852, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6362},
        {31852}
    },
    {
        6363,
        {31861, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6363},
        {31861}
    },
    {
        6364,
        {31875, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6364},
        {31875}
    },
    {
        6365,
        {31888, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6365},
        {31888}
    },
    {
        6366,
        {31908, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6366},
        {31908}
    },
    {
        6367,
        {31917, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6367},
        {31917}
    },
    {
        6368,
        {31906, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6368},
        {31906}
    },
    {
        6369,
        {31915, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6369},
        {31915}
    },
    {
        6370,
        {31905, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6370},
        {31905}
    },
    {
        6371,
        {31912, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6371},
        {31912}
    },
    {
        6372,
        {31923, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6372},
        {31923}
    },
    {
        6373,
        {31922, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6373},
        {31922}
    },
    {
        6374,
        {31921, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6374},
        {31921}
    },
    {
        6375,
        {31918, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6375},
        {31918}
    },
    {
        6376,
        {31929, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6376},
        {31929}
    },
    {
        6377,
        {31933, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6377},
        {31933}
    },
    {
        6378,
        {31936, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6378},
        {31936}
    },
    {
        6379,
        {31941, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6379},
        {31941}
    },
    {
        6380,
        {31938, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6380},
        {31938}
    },
    {
        6381,
        {31960, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6381},
        {31960}
    },
    {
        6382,
        {31954, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6382},
        {31954}
    },
    {
        6383,
        {31964, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6383},
        {31964}
    },
    {
        6384,
        {31970, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6384},
        {31970}
    },
    {
        6385,
        {39739, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6385},
        {39739}
    },
    {
        6386,
        {31983, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6386},
        {31983}
    },
    {
        6387,
        {31986, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6387},
        {31986}
    },
    {
        6388,
        {31988, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6388},
        {31988}
    },
    {
        6389,
        {31990, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6389},
        {31990}
    },
    {
        6390,
        {31994, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6390},
        {31994}
    },
    {
        6391,
        {32006, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6391},
        {32006}
    },
    {
        6392,
        {32002, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6392},
        {32002}
    },
    {
        6393,
        {32028, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6393},
        {32028}
    },
    {
        6394,
        {32021, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6394},
        {32021}
    },
    {
        6395,
        {32010, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6395},
        {32010}
    },
    {
        6396,
        {32069, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6396},
        {32069}
    },
    {
        6397,
        {32075, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6397},
        {32075}
    },
    {
        6398,
        {32046, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6398},
        {32046}
    },
    {
        6399,
        {32050, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6399},
        {32050}
    },
    {
        6400,
        {32063, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6400},
        {32063}
    },
    {
        6401,
        {32053, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6401},
        {32053}
    },
    {
        6402,
        {32070, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6402},
        {32070}
    },
    {
        6403,
        {32115, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6403},
        {32115}
    },
    {
        6404,
        {32086, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6404},
        {32086}
    },
    {
        6405,
        {32078, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6405},
        {32078}
    },
    {
        6406,
        {32114, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6406},
        {32114}
    },
    {
        6407,
        {32104, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6407},
        {32104}
    },
    {
        6408,
        {32110, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6408},
        {32110}
    },
    {
        6409,
        {32079, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6409},
        {32079}
    },
    {
        6410,
        {32099, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6410},
        {32099}
    },
    {
        6411,
        {32147, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6411},
        {32147}
    },
    {
        6412,
        {32137, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6412},
        {32137}
    },
    {
        6413,
        {32091, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6413},
        {32091}
    },
    {
        6414,
        {32143, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6414},
        {32143}
    },
    {
        6415,
        {32125, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6415},
        {32125}
    },
    {
        6416,
        {32155, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6416},
        {32155}
    },
    {
        6417,
        {32186, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6417},
        {32186}
    },
    {
        6418,
        {32174, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6418},
        {32174}
    },
    {
        6419,
        {32163, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6419},
        {32163}
    },
    {
        6420,
        {32181, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6420},
        {32181}
    },
    {
        6421,
        {32199, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6421},
        {32199}
    },
    {
        6422,
        {32189, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6422},
        {32189}
    },
    {
        6423,
        {32171, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6423},
        {32171}
    },
    {
        6424,
        {32317, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6424},
        {32317}
    },
    {
        6425,
        {32162, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6425},
        {32162}
    },
    {
        6426,
        {32175, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6426},
        {32175}
    },
    {
        6427,
        {32220, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6427},
        {32220}
    },
    {
        6428,
        {32184, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6428},
        {32184}
    },
    {
        6429,
        {32159, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6429},
        {32159}
    },
    {
        6430,
        {32176, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6430},
        {32176}
    },
    {
        6431,
        {32216, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6431},
        {32216}
    },
    {
        6432,
        {32221, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6432},
        {32221}
    },
    {
        6433,
        {32228, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6433},
        {32228}
    },
    {
        6434,
        {32222, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6434},
        {32222}
    },
    {
        6435,
        {32251, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6435},
        {32251}
    },
    {
        6436,
        {32242, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6436},
        {32242}
    },
    {
        6437,
        {32225, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6437},
        {32225}
    },
    {
        6438,
        {32261, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6438},
        {32261}
    },
    {
        6439,
        {32266, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6439},
        {32266}
    },
    {
        6440,
        {32291, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6440},
        {32291}
    },
    {
        6441,
        {32289, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6441},
        {32289}
    },
    {
        6442,
        {32274, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6442},
        {32274}
    },
    {
        6443,
        {32305, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6443},
        {32305}
    },
    {
        6444,
        {32287, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6444},
        {32287}
    },
    {
        6445,
        {32265, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6445},
        {32265}
    },
    {
        6446,
        {32267, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6446},
        {32267}
    },
    {
        6447,
        {32290, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6447},
        {32290}
    },
    {
        6448,
        {32326, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6448},
        {32326}
    },
    {
        6449,
        {32358, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6449},
        {32358}
    },
    {
        6450,
        {32315, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6450},
        {32315}
    },
    {
        6451,
        {32309, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6451},
        {32309}
    },
    {
        6452,
        {32313, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6452},
        {32313}
    },
    {
        6453,
        {32323, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6453},
        {32323}
    },
    {
        6454,
        {32311, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6454},
        {32311}
    },
    {
        6455,
        {32306, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6455},
        {32306}
    },
    {
        6456,
        {32314, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6456},
        {32314}
    },
    {
        6457,
        {32359, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6457},
        {32359}
    },
    {
        6458,
        {32349, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6458},
        {32349}
    },
    {
        6459,
        {32342, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6459},
        {32342}
    },
    {
        6460,
        {32350, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6460},
        {32350}
    },
    {
        6461,
        {32345, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6461},
        {32345}
    },
    {
        6462,
        {32346, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6462},
        {32346}
    },
    {
        6463,
        {32377, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6463},
        {32377}
    },
    {
        6464,
        {32362, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6464},
        {32362}
    },
    {
        6465,
        {32361, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6465},
        {32361}
    },
    {
        6466,
        {32380, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6466},
        {32380}
    },
    {
        6467,
        {32379, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6467},
        {32379}
    },
    {
        6468,
        {32387, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6468},
        {32387}
    },
    {
        6469,
        {32213, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6469},
        {32213}
    },
    {
        6470,
        {32381, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6470},
        {32381}
    },
    {
        6471,
        {36782, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6471},
        {36782}
    },
    {
        6472,
        {32383, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6472},
        {32383}
    },
    {
        6473,
        {32392, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6473},
        {32392}
    },
    {
        6474,
        {32393, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6474},
        {32393}
    },
    {
        6475,
        {32396, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6475},
        {32396}
    },
    {
        6476,
        {32402, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6476},
        {32402}
    },
    {
        6477,
        {32400, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6477},
        {32400}
    },
    {
        6478,
        {32403, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6478},
        {32403}
    },
    {
        6479,
        {32404, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6479},
        {32404}
    },
    {
        6480,
        {32406, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6480},
        {32406}
    },
    {
        6481,
        {32398, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6481},
        {32398}
    },
    {
        6482,
        {32411, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6482},
        {32411}
    },
    {
        6483,
        {32412, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6483},
        {32412}
    },
    {
        6484,
        {32568, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6484},
        {32568}
    },
    {
        6485,
        {32570, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6485},
        {32570}
    },
    {
        6486,
        {32581, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6486},
        {32581}
    },
    {
        6487,
        {32588, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6487},
        {32588}
    },
    {
        6488,
        {32589, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6488},
        {32589}
    },
    {
        6489,
        {32590, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6489},
        {32590}
    },
    {
        6490,
        {32592, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6490},
        {32592}
    },
    {
        6491,
        {32593, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6491},
        {32593}
    },
    {
        6492,
        {32597, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6492},
        {32597}
    },
    {
        6493,
        {32596, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6493},
        {32596}
    },
    {
        6494,
        {32600, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6494},
        {32600}
    },
    {
        6495,
        {32607, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6495},
        {32607}
    },
    {
        6496,
        {32608, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6496},
        {32608}
    },
    {
        6497,
        {32616, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6497},
        {32616}
    },
    {
        6498,
        {32617, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6498},
        {32617}
    },
    {
        6499,
        {32615, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6499},
        {32615}
    },
    {
        6500,
        {32632, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6500},
        {32632}
    },
    {
        6501,
        {32642, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6501},
        {32642}
    },
    {
        6502,
        {32646, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6502},
        {32646}
    },
    {
        6503,
        {32643, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6503},
        {32643}
    },
    {
        6504,
        {32648, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6504},
        {32648}
    },
    {
        6505,
        {32647, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6505},
        {32647}
    },
    {
        6506,
        {32652, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6506},
        {32652}
    },
    {
        6507,
        {32660, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6507},
        {32660}
    },
    {
        6508,
        {32670, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6508},
        {32670}
    },
    {
        6509,
        {32669, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6509},
        {32669}
    },
    {
        6510,
        {32666, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6510},
        {32666}
    },
    {
        6511,
        {32675, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6511},
        {32675}
    },
    {
        6512,
        {32687, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6512},
        {32687}
    },
    {
        6513,
        {32690, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6513},
        {32690}
    },
    {
        6514,
        {32697, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6514},
        {32697}
    },
    {
        6515,
        {32686, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6515},
        {32686}
    },
    {
        6516,
        {32694, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6516},
        {32694}
    },
    {
        6517,
        {32696, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6517},
        {32696}
    },
    {
        6518,
        {35697, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6518},
        {35697}
    },
    {
        6519,
        {32709, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6519},
        {32709}
    },
    {
        6520,
        {32710, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6520},
        {32710}
    },
    {
        6521,
        {32714, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6521},
        {32714}
    },
    {
        6522,
        {32725, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6522},
        {32725}
    },
    {
        6523,
        {32724, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6523},
        {32724}
    },
    {
        6524,
        {32737, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6524},
        {32737}
    },
    {
        6525,
        {32742, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6525},
        {32742}
    },
    {
        6526,
        {32745, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6526},
        {32745}
    },
    {
        6527,
        {32755, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6527},
        {32755}
    },
    {
        6528,
        {32761, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6528},
        {32761}
    },
    {
        6529,
        {39132, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6529},
        {39132}
    },
    {
        6530,
        {32774, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6530},
        {32774}
    },
    {
        6531,
        {32772, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6531},
        {32772}
    },
    {
        6532,
        {32779, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6532},
        {32779}
    },
    {
        6533,
        {32786, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6533},
        {32786}
    },
    {
        6534,
        {32792, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6534},
        {32792}
    },
    {
        6535,
        {32793, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6535},
        {32793}
    },
    {
        6536,
        {32796, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6536},
        {32796}
    },
    {
        6537,
        {32801, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6537},
        {32801}
    },
    {
        6538,
        {32808, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6538},
        {32808}
    },
    {
        6539,
        {32831, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6539},
        {32831}
    },
    {
        6540,
        {32827, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6540},
        {32827}
    },
    {
        6541,
        {32842, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6541},
        {32842}
    },
    {
        6542,
        {32838, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6542},
        {32838}
    },
    {
        6543,
        {32850, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6543},
        {32850}
    },
    {
        6544,
        {32856, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6544},
        {32856}
    },
    {
        6545,
        {32858, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6545},
        {32858}
    },
    {
        6546,
        {32863, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6546},
        {32863}
    },
    {
        6547,
        {32866, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6547},
        {32866}
    },
    {
        6548,
        {32872, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6548},
        {32872}
    },
    {
        6549,
        {32883, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6549},
        {32883}
    },
    {
        6550,
        {32882, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6550},
        {32882}
    },
    {
        6551,
        {32880, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6551},
        {32880}
    },
    {
        6552,
        {32886, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6552},
        {32886}
    },
    {
        6553,
        {32889, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6553},
        {32889}
    },
    {
        6554,
        {32893, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6554},
        {32893}
    },
    {
        6555,
        {32895, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6555},
        {32895}
    },
    {
        6556,
        {32900, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6556},
        {32900}
    },
    {
        6557,
        {32902, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6557},
        {32902}
    },
    {
        6558,
        {32901, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6558},
        {32901}
    },
    {
        6559,
        {32923, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6559},
        {32923}
    },
    {
        6560,
        {32915, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6560},
        {32915}
    },
    {
        6561,
        {32922, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6561},
        {32922}
    },
    {
        6562,
        {32941, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6562},
        {32941}
    },
    {
        6563,
        {20880, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6563},
        {20880}
    },
    {
        6564,
        {32940, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6564},
        {32940}
    },
    {
        6565,
        {32987, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6565},
        {32987}
    },
    {
        6566,
        {32997, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6566},
        {32997}
    },
    {
        6567,
        {32985, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6567},
        {32985}
    },
    {
        6568,
        {32989, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6568},
        {32989}
    },
    {
        6569,
        {32964, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6569},
        {32964}
    },
    {
        6570,
        {32986, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6570},
        {32986}
    },
    {
        6571,
        {32982, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6571},
        {32982}
    },
    {
        6572,
        {33033, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6572},
        {33033}
    },
    {
        6573,
        {33007, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6573},
        {33007}
    },
    {
        6574,
        {33009, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6574},
        {33009}
    },
    {
        6575,
        {33051, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6575},
        {33051}
    },
    {
        6576,
        {33065, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6576},
        {33065}
    },
    {
        6577,
        {33059, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6577},
        {33059}
    },
    {
        6578,
        {33071, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6578},
        {33071}
    },
    {
        6579,
        {33099, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6579},
        {33099}
    },
    {
        6580,
        {38539, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6580},
        {38539}
    },
    {
        6581,
        {33094, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6581},
        {33094}
    },
    {
        6582,
        {33086, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6582},
        {33086}
    },
    {
        6583,
        {33107, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6583},
        {33107}
    },
    {
        6584,
        {33105, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6584},
        {33105}
    },
    {
        6585,
        {33020, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6585},
        {33020}
    },
    {
        6586,
        {33137, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6586},
        {33137}
    },
    {
        6587,
        {33134, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6587},
        {33134}
    },
    {
        6588,
        {33125, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6588},
        {33125}
    },
    {
        6589,
        {33126, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6589},
        {33126}
    },
    {
        6590,
        {33140, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6590},
        {33140}
    },
    {
        6591,
        {33155, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6591},
        {33155}
    },
    {
        6592,
        {33160, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6592},
        {33160}
    },
    {
        6593,
        {33162, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6593},
        {33162}
    },
    {
        6594,
        {33152, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6594},
        {33152}
    },
    {
        6595,
        {33154, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6595},
        {33154}
    },
    {
        6596,
        {33184, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6596},
        {33184}
    },
    {
        6597,
        {33173, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6597},
        {33173}
    },
    {
        6598,
        {33188, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6598},
        {33188}
    },
    {
        6599,
        {33187, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6599},
        {33187}
    },
    {
        6600,
        {33119, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6600},
        {33119}
    },
    {
        6601,
        {33171, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6601},
        {33171}
    },
    {
        6602,
        {33193, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6602},
        {33193}
    },
    {
        6603,
        {33200, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6603},
        {33200}
    },
    {
        6604,
        {33205, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6604},
        {33205}
    },
    {
        6605,
        {33214, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6605},
        {33214}
    },
    {
        6606,
        {33208, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6606},
        {33208}
    },
    {
        6607,
        {33213, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6607},
        {33213}
    },
    {
        6608,
        {33216, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6608},
        {33216}
    },
    {
        6609,
        {33218, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6609},
        {33218}
    },
    {
        6610,
        {33210, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6610},
        {33210}
    },
    {
        6611,
        {33225, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6611},
        {33225}
    },
    {
        6612,
        {33229, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6612},
        {33229}
    },
    {
        6613,
        {33233, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6613},
        {33233}
    },
    {
        6614,
        {33241, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6614},
        {33241}
    },
    {
        6615,
        {33240, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6615},
        {33240}
    },
    {
        6616,
        {33224, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6616},
        {33224}
    },
    {
        6617,
        {33242, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6617},
        {33242}
    },
    {
        6618,
        {33247, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6618},
        {33247}
    },
    {
        6619,
        {33248, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6619},
        {33248}
    },
    {
        6620,
        {33255, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6620},
        {33255}
    },
    {
        6621,
        {33274, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6621},
        {33274}
    },
    {
        6622,
        {33275, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6622},
        {33275}
    },
    {
        6623,
        {33278, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6623},
        {33278}
    },
    {
        6624,
        {33281, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6624},
        {33281}
    },
    {
        6625,
        {33282, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6625},
        {33282}
    },
    {
        6626,
        {33285, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6626},
        {33285}
    },
    {
        6627,
        {33287, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6627},
        {33287}
    },
    {
        6628,
        {33290, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6628},
        {33290}
    },
    {
        6629,
        {33293, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6629},
        {33293}
    },
    {
        6630,
        {33296, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6630},
        {33296}
    },
    {
        6631,
        {33302, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6631},
        {33302}
    },
    {
        6632,
        {33321, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6632},
        {33321}
    },
    {
        6633,
        {33323, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6633},
        {33323}
    },
    {
        6634,
        {33336, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6634},
        {33336}
    },
    {
        6635,
        {33331, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6635},
        {33331}
    },
    {
        6636,
        {33344, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6636},
        {33344}
    },
    {
        6637,
        {33369, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6637},
        {33369}
    },
    {
        6638,
        {33368, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6638},
        {33368}
    },
    {
        6639,
        {33373, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6639},
        {33373}
    },
    {
        6640,
        {33370, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6640},
        {33370}
    },
    {
        6641,
        {33375, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6641},
        {33375}
    },
    {
        6642,
        {33380, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6642},
        {33380}
    },
    {
        6643,
        {33378, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6643},
        {33378}
    },
    {
        6644,
        {33384, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6644},
        {33384}
    },
    {
        6645,
        {33386, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6645},
        {33386}
    },
    {
        6646,
        {33387, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6646},
        {33387}
    },
    {
        6647,
        {33326, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6647},
        {33326}
    },
    {
        6648,
        {33393, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6648},
        {33393}
    },
    {
        6649,
        {33399, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6649},
        {33399}
    },
    {
        6650,
        {33400, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6650},
        {33400}
    },
    {
        6651,
        {33406, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6651},
        {33406}
    },
    {
        6652,
        {33421, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6652},
        {33421}
    },
    {
        6653,
        {33426, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6653},
        {33426}
    },
    {
        6654,
        {33451, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6654},
        {33451}
    },
    {
        6655,
        {33439, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6655},
        {33439}
    },
    {
        6656,
        {33467, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6656},
        {33467}
    },
    {
        6657,
        {33452, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6657},
        {33452}
    },
    {
        6658,
        {33505, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6658},
        {33505}
    },
    {
        6659,
        {33507, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6659},
        {33507}
    },
    {
        6660,
        {33503, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6660},
        {33503}
    },
    {
        6661,
        {33490, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6661},
        {33490}
    },
    {
        6662,
        {33524, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6662},
        {33524}
    },
    {
        6663,
        {33523, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6663},
        {33523}
    },
    {
        6664,
        {33530, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6664},
        {33530}
    },
    {
        6665,
        {33683, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6665},
        {33683}
    },
    {
        6666,
        {33539, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6666},
        {33539}
    },
    {
        6667,
        {33531, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6667},
        {33531}
    },
    {
        6668,
        {33529, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6668},
        {33529}
    },
    {
        6669,
        {33502, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6669},
        {33502}
    },
    {
        6670,
        {33542, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6670},
        {33542}
    },
    {
        6671,
        {33500, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6671},
        {33500}
    },
    {
        6672,
        {33545, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6672},
        {33545}
    },
    {
        6673,
        {33497, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6673},
        {33497}
    },
    {
        6674,
        {33589, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6674},
        {33589}
    },
    {
        6675,
        {33588, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6675},
        {33588}
    },
    {
        6676,
        {33558, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6676},
        {33558}
    },
    {
        6677,
        {33586, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6677},
        {33586}
    },
    {
        6678,
        {33585, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6678},
        {33585}
    },
    {
        6679,
        {33600, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6679},
        {33600}
    },
    {
        6680,
        {33593, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6680},
        {33593}
    },
    {
        6681,
        {33616, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6681},
        {33616}
    },
    {
        6682,
        {33605, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6682},
        {33605}
    },
    {
        6683,
        {33583, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6683},
        {33583}
    },
    {
        6684,
        {33579, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6684},
        {33579}
    },
    {
        6685,
        {33559, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6685},
        {33559}
    },
    {
        6686,
        {33560, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6686},
        {33560}
    },
    {
        6687,
        {33669, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6687},
        {33669}
    },
    {
        6688,
        {33690, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6688},
        {33690}
    },
    {
        6689,
        {33706, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6689},
        {33706}
    },
    {
        6690,
        {33695, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6690},
        {33695}
    },
    {
        6691,
        {33698, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6691},
        {33698}
    },
    {
        6692,
        {33686, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6692},
        {33686}
    },
    {
        6693,
        {33571, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6693},
        {33571}
    },
    {
        6694,
        {33678, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6694},
        {33678}
    },
    {
        6695,
        {33671, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6695},
        {33671}
    },
    {
        6696,
        {33674, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6696},
        {33674}
    },
    {
        6697,
        {33660, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6697},
        {33660}
    },
    {
        6698,
        {33717, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6698},
        {33717}
    },
    {
        6699,
        {33651, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6699},
        {33651}
    },
    {
        6700,
        {33653, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6700},
        {33653}
    },
    {
        6701,
        {33696, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6701},
        {33696}
    },
    {
        6702,
        {33673, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6702},
        {33673}
    },
    {
        6703,
        {33704, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6703},
        {33704}
    },
    {
        6704,
        {33780, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6704},
        {33780}
    },
    {
        6705,
        {33811, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6705},
        {33811}
    },
    {
        6706,
        {33771, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6706},
        {33771}
    },
    {
        6707,
        {33742, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6707},
        {33742}
    },
    {
        6708,
        {33789, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6708},
        {33789}
    },
    {
        6709,
        {33795, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6709},
        {33795}
    },
    {
        6710,
        {33752, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6710},
        {33752}
    },
    {
        6711,
        {33803, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6711},
        {33803}
    },
    {
        6712,
        {33729, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6712},
        {33729}
    },
    {
        6713,
        {33783, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6713},
        {33783}
    },
    {
        6714,
        {33799, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6714},
        {33799}
    },
    {
        6715,
        {33760, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6715},
        {33760}
    },
    {
        6716,
        {33778, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6716},
        {33778}
    },
    {
        6717,
        {33805, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6717},
        {33805}
    },
    {
        6718,
        {33826, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6718},
        {33826}
    },
    {
        6719,
        {33824, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6719},
        {33824}
    },
    {
        6720,
        {33725, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6720},
        {33725}
    },
    {
        6721,
        {33848, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6721},
        {33848}
    },
    {
        6722,
        {34054, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6722},
        {34054}
    },
    {
        6723,
        {33787, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6723},
        {33787}
    },
    {
        6724,
        {33901, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6724},
        {33901}
    },
    {
        6725,
        {33834, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6725},
        {33834}
    },
    {
        6726,
        {33852, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6726},
        {33852}
    },
    {
        6727,
        {34138, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6727},
        {34138}
    },
    {
        6728,
        {33924, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6728},
        {33924}
    },
    {
        6729,
        {33911, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6729},
        {33911}
    },
    {
        6730,
        {33899, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6730},
        {33899}
    },
    {
        6731,
        {33965, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6731},
        {33965}
    },
    {
        6732,
        {33902, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6732},
        {33902}
    },
    {
        6733,
        {33922, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6733},
        {33922}
    },
    {
        6734,
        {33897, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6734},
        {33897}
    },
    {
        6735,
        {33862, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6735},
        {33862}
    },
    {
        6736,
        {33836, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6736},
        {33836}
    },
    {
        6737,
        {33903, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6737},
        {33903}
    },
    {
        6738,
        {33913, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6738},
        {33913}
    },
    {
        6739,
        {33845, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6739},
        {33845}
    },
    {
        6740,
        {33994, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6740},
        {33994}
    },
    {
        6741,
        {33890, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6741},
        {33890}
    },
    {
        6742,
        {33977, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6742},
        {33977}
    },
    {
        6743,
        {33983, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6743},
        {33983}
    },
    {
        6744,
        {33951, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6744},
        {33951}
    },
    {
        6745,
        {34009, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6745},
        {34009}
    },
    {
        6746,
        {33997, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6746},
        {33997}
    },
    {
        6747,
        {33979, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6747},
        {33979}
    },
    {
        6748,
        {34010, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6748},
        {34010}
    },
    {
        6749,
        {34000, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6749},
        {34000}
    },
    {
        6750,
        {33985, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6750},
        {33985}
    },
    {
        6751,
        {33990, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6751},
        {33990}
    },
    {
        6752,
        {34006, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6752},
        {34006}
    },
    {
        6753,
        {33953, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6753},
        {33953}
    },
    {
        6754,
        {34081, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6754},
        {34081}
    },
    {
        6755,
        {34047, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6755},
        {34047}
    },
    {
        6756,
        {34036, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6756},
        {34036}
    },
    {
        6757,
        {34071, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6757},
        {34071}
    },
    {
        6758,
        {34072, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6758},
        {34072}
    },
    {
        6759,
        {34092, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6759},
        {34092}
    },
    {
        6760,
        {34079, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6760},
        {34079}
    },
    {
        6761,
        {34069, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6761},
        {34069}
    },
    {
        6762,
        {34068, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6762},
        {34068}
    },
    {
        6763,
        {34044, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6763},
        {34044}
    },
    {
        6764,
        {34112, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6764},
        {34112}
    },
    {
        6765,
        {34147, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6765},
        {34147}
    },
    {
        6766,
        {34136, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6766},
        {34136}
    },
    {
        6767,
        {34120, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6767},
        {34120}
    },
    {
        6768,
        {34113, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6768},
        {34113}
    },
    {
        6769,
        {34306, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6769},
        {34306}
    },
    {
        6770,
        {34123, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6770},
        {34123}
    },
    {
        6771,
        {34133, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6771},
        {34133}
    },
    {
        6772,
        {34176, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6772},
        {34176}
    },
    {
        6773,
        {34212, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6773},
        {34212}
    },
    {
        6774,
        {34184, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6774},
        {34184}
    },
    {
        6775,
        {34193, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6775},
        {34193}
    },
    {
        6776,
        {34186, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6776},
        {34186}
    },
    {
        6777,
        {34216, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6777},
        {34216}
    },
    {
        6778,
        {34157, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6778},
        {34157}
    },
    {
        6779,
        {34196, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6779},
        {34196}
    },
    {
        6780,
        {34203, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6780},
        {34203}
    },
    {
        6781,
        {34282, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6781},
        {34282}
    },
    {
        6782,
        {34183, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6782},
        {34183}
    },
    {
        6783,
        {34204, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6783},
        {34204}
    },
    {
        6784,
        {34167, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6784},
        {34167}
    },
    {
        6785,
        {34174, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6785},
        {34174}
    },
    {
        6786,
        {34192, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6786},
        {34192}
    },
    {
        6787,
        {34249, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6787},
        {34249}
    },
    {
        6788,
        {34234, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6788},
        {34234}
    },
    {
        6789,
        {34255, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6789},
        {34255}
    },
    {
        6790,
        {34233, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6790},
        {34233}
    },
    {
        6791,
        {34256, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6791},
        {34256}
    },
    {
        6792,
        {34261, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6792},
        {34261}
    },
    {
        6793,
        {34269, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6793},
        {34269}
    },
    {
        6794,
        {34277, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6794},
        {34277}
    },
    {
        6795,
        {34268, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6795},
        {34268}
    },
    {
        6796,
        {34297, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6796},
        {34297}
    },
    {
        6797,
        {34314, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6797},
        {34314}
    },
    {
        6798,
        {34323, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6798},
        {34323}
    },
    {
        6799,
        {34315, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6799},
        {34315}
    },
    {
        6800,
        {34302, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6800},
        {34302}
    },
    {
        6801,
        {34298, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6801},
        {34298}
    },
    {
        6802,
        {34310, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6802},
        {34310}
    },
    {
        6803,
        {34338, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6803},
        {34338}
    },
    {
        6804,
        {34330, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6804},
        {34330}
    },
    {
        6805,
        {34352, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6805},
        {34352}
    },
    {
        6806,
        {34367, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6806},
        {34367}
    },
    {
        6807,
        {34381, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6807},
        {34381}
    },
    {
        6808,
        {20053, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6808},
        {20053}
    },
    {
        6809,
        {34388, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6809},
        {34388}
    },
    {
        6810,
        {34399, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6810},
        {34399}
    },
    {
        6811,
        {34407, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6811},
        {34407}
    },
    {
        6812,
        {34417, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6812},
        {34417}
    },
    {
        6813,
        {34451, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6813},
        {34451}
    },
    {
        6814,
        {34467, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6814},
        {34467}
    },
    {
        6815,
        {34473, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6815},
        {34473}
    },
    {
        6816,
        {34474, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6816},
        {34474}
    },
    {
        6817,
        {34443, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6817},
        {34443}
    },
    {
        6818,
        {34444, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6818},
        {34444}
    },
    {
        6819,
        {34486, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6819},
        {34486}
    },
    {
        6820,
        {34479, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6820},
        {34479}
    },
    {
        6821,
        {34500, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6821},
        {34500}
    },
    {
        6822,
        {34502, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6822},
        {34502}
    },
    {
        6823,
        {34480, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6823},
        {34480}
    },
    {
        6824,
        {34505, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6824},
        {34505}
    },
    {
        6825,
        {34851, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6825},
        {34851}
    },
    {
        6826,
        {34475, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6826},
        {34475}
    },
    {
        6827,
        {34516, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6827},
        {34516}
    },
    {
        6828,
        {34526, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6828},
        {34526}
    },
    {
        6829,
        {34537, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6829},
        {34537}
    },
    {
        6830,
        {34540, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6830},
        {34540}
    },
    {
        6831,
        {34527, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6831},
        {34527}
    },
    {
        6832,
        {34523, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6832},
        {34523}
    },
    {
        6833,
        {34543, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6833},
        {34543}
    },
    {
        6834,
        {34578, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6834},
        {34578}
    },
    {
        6835,
        {34566, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6835},
        {34566}
    },
    {
        6836,
        {34568, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6836},
        {34568}
    },
    {
        6837,
        {34560, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6837},
        {34560}
    },
    {
        6838,
        {34563, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6838},
        {34563}
    },
    {
        6839,
        {34555, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6839},
        {34555}
    },
    {
        6840,
        {34577, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6840},
        {34577}
    },
    {
        6841,
        {34569, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6841},
        {34569}
    },
    {
        6842,
        {34573, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6842},
        {34573}
    },
    {
        6843,
        {34553, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6843},
        {34553}
    },
    {
        6844,
        {34570, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6844},
        {34570}
    },
    {
        6845,
        {34612, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6845},
        {34612}
    },
    {
        6846,
        {34623, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6846},
        {34623}
    },
    {
        6847,
        {34615, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6847},
        {34615}
    },
    {
        6848,
        {34619, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6848},
        {34619}
    },
    {
        6849,
        {34597, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6849},
        {34597}
    },
    {
        6850,
        {34601, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6850},
        {34601}
    },
    {
        6851,
        {34586, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6851},
        {34586}
    },
    {
        6852,
        {34656, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6852},
        {34656}
    },
    {
        6853,
        {34655, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6853},
        {34655}
    },
    {
        6854,
        {34680, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6854},
        {34680}
    },
    {
        6855,
        {34636, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6855},
        {34636}
    },
    {
        6856,
        {34638, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6856},
        {34638}
    },
    {
        6857,
        {34676, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6857},
        {34676}
    },
    {
        6858,
        {34647, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6858},
        {34647}
    },
    {
        6859,
        {34664, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6859},
        {34664}
    },
    {
        6860,
        {34670, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6860},
        {34670}
    },
    {
        6861,
        {34649, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6861},
        {34649}
    },
    {
        6862,
        {34643, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6862},
        {34643}
    },
    {
        6863,
        {34659, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6863},
        {34659}
    },
    {
        6864,
        {34666, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6864},
        {34666}
    },
    {
        6865,
        {34821, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6865},
        {34821}
    },
    {
        6866,
        {34722, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6866},
        {34722}
    },
    {
        6867,
        {34719, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6867},
        {34719}
    },
    {
        6868,
        {34690, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6868},
        {34690}
    },
    {
        6869,
        {34735, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6869},
        {34735}
    },
    {
        6870,
        {34763, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6870},
        {34763}
    },
    {
        6871,
        {34749, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6871},
        {34749}
    },
    {
        6872,
        {34752, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6872},
        {34752}
    },
    {
        6873,
        {34768, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6873},
        {34768}
    },
    {
        6874,
        {38614, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6874},
        {38614}
    },
    {
        6875,
        {34731, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6875},
        {34731}
    },
    {
        6876,
        {34756, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6876},
        {34756}
    },
    {
        6877,
        {34739, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6877},
        {34739}
    },
    {
        6878,
        {34759, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6878},
        {34759}
    },
    {
        6879,
        {34758, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6879},
        {34758}
    },
    {
        6880,
        {34747, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6880},
        {34747}
    },
    {
        6881,
        {34799, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6881},
        {34799}
    },
    {
        6882,
        {34802, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6882},
        {34802}
    },
    {
        6883,
        {34784, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6883},
        {34784}
    },
    {
        6884,
        {34831, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6884},
        {34831}
    },
    {
        6885,
        {34829, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6885},
        {34829}
    },
    {
        6886,
        {34814, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6886},
        {34814}
    },
    {
        6887,
        {34806, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6887},
        {34806}
    },
    {
        6888,
        {34807, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6888},
        {34807}
    },
    {
        6889,
        {34830, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6889},
        {34830}
    },
    {
        6890,
        {34770, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6890},
        {34770}
    },
    {
        6891,
        {34833, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6891},
        {34833}
    },
    {
        6892,
        {34838, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6892},
        {34838}
    },
    {
        6893,
        {34837, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6893},
        {34837}
    },
    {
        6894,
        {34850, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6894},
        {34850}
    },
    {
        6895,
        {34849, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6895},
        {34849}
    },
    {
        6896,
        {34865, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6896},
        {34865}
    },
    {
        6897,
        {34870, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6897},
        {34870}
    },
    {
        6898,
        {34873, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6898},
        {34873}
    },
    {
        6899,
        {34855, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6899},
        {34855}
    },
    {
        6900,
        {34875, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6900},
        {34875}
    },
    {
        6901,
        {34884, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6901},
        {34884}
    },
    {
        6902,
        {34882, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6902},
        {34882}
    },
    {
        6903,
        {34898, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6903},
        {34898}
    },
    {
        6904,
        {34905, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6904},
        {34905}
    },
    {
        6905,
        {34910, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6905},
        {34910}
    },
    {
        6906,
        {34914, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6906},
        {34914}
    },
    {
        6907,
        {34923, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6907},
        {34923}
    },
    {
        6908,
        {34945, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6908},
        {34945}
    },
    {
        6909,
        {34942, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6909},
        {34942}
    },
    {
        6910,
        {34974, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6910},
        {34974}
    },
    {
        6911,
        {34933, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6911},
        {34933}
    },
    {
        6912,
        {34941, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6912},
        {34941}
    },
    {
        6913,
        {34997, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6913},
        {34997}
    },
    {
        6914,
        {34930, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6914},
        {34930}
    },
    {
        6915,
        {34946, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6915},
        {34946}
    },
    {
        6916,
        {34967, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6916},
        {34967}
    },
    {
        6917,
        {34962, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6917},
        {34962}
    },
    {
        6918,
        {34990, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6918},
        {34990}
    },
    {
        6919,
        {34969, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6919},
        {34969}
    },
    {
        6920,
        {34978, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6920},
        {34978}
    },
    {
        6921,
        {34957, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6921},
        {34957}
    },
    {
        6922,
        {34980, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6922},
        {34980}
    },
    {
        6923,
        {34992, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6923},
        {34992}
    },
    {
        6924,
        {35007, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6924},
        {35007}
    },
    {
        6925,
        {34993, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6925},
        {34993}
    },
    {
        6926,
        {35011, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6926},
        {35011}
    },
    {
        6927,
        {35012, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6927},
        {35012}
    },
    {
        6928,
        {35028, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6928},
        {35028}
    },
    {
        6929,
        {35032, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6929},
        {35032}
    },
    {
        6930,
        {35033, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6930},
        {35033}
    },
    {
        6931,
        {35037, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6931},
        {35037}
    },
    {
        6932,
        {35065, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6932},
        {35065}
    },
    {
        6933,
        {35074, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6933},
        {35074}
    },
    {
        6934,
        {35068, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6934},
        {35068}
    },
    {
        6935,
        {35060, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6935},
        {35060}
    },
    {
        6936,
        {35048, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6936},
        {35048}
    },
    {
        6937,
        {35058, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6937},
        {35058}
    },
    {
        6938,
        {35076, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6938},
        {35076}
    },
    {
        6939,
        {35084, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6939},
        {35084}
    },
    {
        6940,
        {35082, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6940},
        {35082}
    },
    {
        6941,
        {35091, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6941},
        {35091}
    },
    {
        6942,
        {35139, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6942},
        {35139}
    },
    {
        6943,
        {35102, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6943},
        {35102}
    },
    {
        6944,
        {35109, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6944},
        {35109}
    },
    {
        6945,
        {35114, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6945},
        {35114}
    },
    {
        6946,
        {35115, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6946},
        {35115}
    },
    {
        6947,
        {35137, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6947},
        {35137}
    },
    {
        6948,
        {35140, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6948},
        {35140}
    },
    {
        6949,
        {35131, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6949},
        {35131}
    },
    {
        6950,
        {35126, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6950},
        {35126}
    },
    {
        6951,
        {35128, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6951},
        {35128}
    },
    {
        6952,
        {35148, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6952},
        {35148}
    },
    {
        6953,
        {35101, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6953},
        {35101}
    },
    {
        6954,
        {35168, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6954},
        {35168}
    },
    {
        6955,
        {35166, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6955},
        {35166}
    },
    {
        6956,
        {35174, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6956},
        {35174}
    },
    {
        6957,
        {35172, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6957},
        {35172}
    },
    {
        6958,
        {35181, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6958},
        {35181}
    },
    {
        6959,
        {35178, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6959},
        {35178}
    },
    {
        6960,
        {35183, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6960},
        {35183}
    },
    {
        6961,
        {35188, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6961},
        {35188}
    },
    {
        6962,
        {35191, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6962},
        {35191}
    },
    {
        6963,
        {35198, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6963},
        {35198}
    },
    {
        6964,
        {35203, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6964},
        {35203}
    },
    {
        6965,
        {35208, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6965},
        {35208}
    },
    {
        6966,
        {35210, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6966},
        {35210}
    },
    {
        6967,
        {35219, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6967},
        {35219}
    },
    {
        6968,
        {35224, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6968},
        {35224}
    },
    {
        6969,
        {35233, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6969},
        {35233}
    },
    {
        6970,
        {35241, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6970},
        {35241}
    },
    {
        6971,
        {35238, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6971},
        {35238}
    },
    {
        6972,
        {35244, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6972},
        {35244}
    },
    {
        6973,
        {35247, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6973},
        {35247}
    },
    {
        6974,
        {35250, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6974},
        {35250}
    },
    {
        6975,
        {35258, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6975},
        {35258}
    },
    {
        6976,
        {35261, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6976},
        {35261}
    },
    {
        6977,
        {35263, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6977},
        {35263}
    },
    {
        6978,
        {35264, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6978},
        {35264}
    },
    {
        6979,
        {35290, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6979},
        {35290}
    },
    {
        6980,
        {35292, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6980},
        {35292}
    },
    {
        6981,
        {35293, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6981},
        {35293}
    },
    {
        6982,
        {35303, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6982},
        {35303}
    },
    {
        6983,
        {35316, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6983},
        {35316}
    },
    {
        6984,
        {35320, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6984},
        {35320}
    },
    {
        6985,
        {35331, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6985},
        {35331}
    },
    {
        6986,
        {35350, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6986},
        {35350}
    },
    {
        6987,
        {35344, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6987},
        {35344}
    },
    {
        6988,
        {35340, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6988},
        {35340}
    },
    {
        6989,
        {35355, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6989},
        {35355}
    },
    {
        6990,
        {35357, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6990},
        {35357}
    },
    {
        6991,
        {35365, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6991},
        {35365}
    },
    {
        6992,
        {35382, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6992},
        {35382}
    },
    {
        6993,
        {35393, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6993},
        {35393}
    },
    {
        6994,
        {35419, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6994},
        {35419}
    },
    {
        6995,
        {35410, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6995},
        {35410}
    },
    {
        6996,
        {35398, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6996},
        {35398}
    },
    {
        6997,
        {35400, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6997},
        {35400}
    },
    {
        6998,
        {35452, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6998},
        {35452}
    },
    {
        6999,
        {35437, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {6999},
        {35437}
    },
    {
        7000,
        {35436, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7000},
        {35436}
    },
    {
        7001,
        {35426, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7001},
        {35426}
    },
    {
        7002,
        {35461, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7002},
        {35461}
    },
    {
        7003,
        {35458, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7003},
        {35458}
    },
    {
        7004,
        {35460, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7004},
        {35460}
    },
    {
        7005,
        {35496, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7005},
        {35496}
    },
    {
        7006,
        {35489, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7006},
        {35489}
    },
    {
        7007,
        {35473, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7007},
        {35473}
    },
    {
        7008,
        {35493, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7008},
        {35493}
    },
    {
        7009,
        {35494, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7009},
        {35494}
    },
    {
        7010,
        {35482, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7010},
        {35482}
    },
    {
        7011,
        {35491, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7011},
        {35491}
    },
    {
        7012,
        {35524, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7012},
        {35524}
    },
    {
        7013,
        {35533, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7013},
        {35533}
    },
    {
        7014,
        {35522, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7014},
        {35522}
    },
    {
        7015,
        {35546, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7015},
        {35546}
    },
    {
        7016,
        {35563, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7016},
        {35563}
    },
    {
        7017,
        {35571, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7017},
        {35571}
    },
    {
        7018,
        {35559, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7018},
        {35559}
    },
    {
        7019,
        {35556, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7019},
        {35556}
    },
    {
        7020,
        {35569, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7020},
        {35569}
    },
    {
        7021,
        {35604, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7021},
        {35604}
    },
    {
        7022,
        {35552, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7022},
        {35552}
    },
    {
        7023,
        {35554, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7023},
        {35554}
    },
    {
        7024,
        {35575, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7024},
        {35575}
    },
    {
        7025,
        {35550, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7025},
        {35550}
    },
    {
        7026,
        {35547, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7026},
        {35547}
    },
    {
        7027,
        {35596, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7027},
        {35596}
    },
    {
        7028,
        {35591, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7028},
        {35591}
    },
    {
        7029,
        {35610, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7029},
        {35610}
    },
    {
        7030,
        {35553, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7030},
        {35553}
    },
    {
        7031,
        {35606, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7031},
        {35606}
    },
    {
        7032,
        {35600, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7032},
        {35600}
    },
    {
        7033,
        {35607, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7033},
        {35607}
    },
    {
        7034,
        {35616, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7034},
        {35616}
    },
    {
        7035,
        {35635, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7035},
        {35635}
    },
    {
        7036,
        {38827, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7036},
        {38827}
    },
    {
        7037,
        {35622, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7037},
        {35622}
    },
    {
        7038,
        {35627, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7038},
        {35627}
    },
    {
        7039,
        {35646, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7039},
        {35646}
    },
    {
        7040,
        {35624, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7040},
        {35624}
    },
    {
        7041,
        {35649, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7041},
        {35649}
    },
    {
        7042,
        {35660, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7042},
        {35660}
    },
    {
        7043,
        {35663, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7043},
        {35663}
    },
    {
        7044,
        {35662, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7044},
        {35662}
    },
    {
        7045,
        {35657, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7045},
        {35657}
    },
    {
        7046,
        {35670, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7046},
        {35670}
    },
    {
        7047,
        {35675, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7047},
        {35675}
    },
    {
        7048,
        {35674, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7048},
        {35674}
    },
    {
        7049,
        {35691, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7049},
        {35691}
    },
    {
        7050,
        {35679, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7050},
        {35679}
    },
    {
        7051,
        {35692, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7051},
        {35692}
    },
    {
        7052,
        {35695, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7052},
        {35695}
    },
    {
        7053,
        {35700, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7053},
        {35700}
    },
    {
        7054,
        {35709, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7054},
        {35709}
    },
    {
        7055,
        {35712, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7055},
        {35712}
    },
    {
        7056,
        {35724, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7056},
        {35724}
    },
    {
        7057,
        {35726, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7057},
        {35726}
    },
    {
        7058,
        {35730, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7058},
        {35730}
    },
    {
        7059,
        {35731, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7059},
        {35731}
    },
    {
        7060,
        {35734, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7060},
        {35734}
    },
    {
        7061,
        {35737, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7061},
        {35737}
    },
    {
        7062,
        {35738, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7062},
        {35738}
    },
    {
        7063,
        {35898, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7063},
        {35898}
    },
    {
        7064,
        {35905, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7064},
        {35905}
    },
    {
        7065,
        {35903, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7065},
        {35903}
    },
    {
        7066,
        {35912, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7066},
        {35912}
    },
    {
        7067,
        {35916, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7067},
        {35916}
    },
    {
        7068,
        {35918, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7068},
        {35918}
    },
    {
        7069,
        {35920, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7069},
        {35920}
    },
    {
        7070,
        {35925, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7070},
        {35925}
    },
    {
        7071,
        {35938, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7071},
        {35938}
    },
    {
        7072,
        {35948, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7072},
        {35948}
    },
    {
        7073,
        {35960, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7073},
        {35960}
    },
    {
        7074,
        {35962, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7074},
        {35962}
    },
    {
        7075,
        {35970, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7075},
        {35970}
    },
    {
        7076,
        {35977, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7076},
        {35977}
    },
    {
        7077,
        {35973, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7077},
        {35973}
    },
    {
        7078,
        {35978, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7078},
        {35978}
    },
    {
        7079,
        {35981, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7079},
        {35981}
    },
    {
        7080,
        {35982, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7080},
        {35982}
    },
    {
        7081,
        {35988, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7081},
        {35988}
    },
    {
        7082,
        {35964, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7082},
        {35964}
    },
    {
        7083,
        {35992, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7083},
        {35992}
    },
    {
        7084,
        {25117, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7084},
        {25117}
    },
    {
        7085,
        {36013, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7085},
        {36013}
    },
    {
        7086,
        {36010, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7086},
        {36010}
    },
    {
        7087,
        {36029, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7087},
        {36029}
    },
    {
        7088,
        {36018, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7088},
        {36018}
    },
    {
        7089,
        {36019, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7089},
        {36019}
    },
    {
        7090,
        {36014, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7090},
        {36014}
    },
    {
        7091,
        {36022, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7091},
        {36022}
    },
    {
        7092,
        {36040, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7092},
        {36040}
    },
    {
        7093,
        {36033, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7093},
        {36033}
    },
    {
        7094,
        {36068, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7094},
        {36068}
    },
    {
        7095,
        {36067, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7095},
        {36067}
    },
    {
        7096,
        {36058, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7096},
        {36058}
    },
    {
        7097,
        {36093, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7097},
        {36093}
    },
    {
        7098,
        {36090, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7098},
        {36090}
    },
    {
        7099,
        {36091, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7099},
        {36091}
    },
    {
        7100,
        {36100, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7100},
        {36100}
    },
    {
        7101,
        {36101, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7101},
        {36101}
    },
    {
        7102,
        {36106, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7102},
        {36106}
    },
    {
        7103,
        {36103, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7103},
        {36103}
    },
    {
        7104,
        {36111, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7104},
        {36111}
    },
    {
        7105,
        {36109, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7105},
        {36109}
    },
    {
        7106,
        {36112, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7106},
        {36112}
    },
    {
        7107,
        {40782, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7107},
        {40782}
    },
    {
        7108,
        {36115, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7108},
        {36115}
    },
    {
        7109,
        {36045, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7109},
        {36045}
    },
    {
        7110,
        {36116, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7110},
        {36116}
    },
    {
        7111,
        {36118, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7111},
        {36118}
    },
    {
        7112,
        {36199, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7112},
        {36199}
    },
    {
        7113,
        {36205, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7113},
        {36205}
    },
    {
        7114,
        {36209, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7114},
        {36209}
    },
    {
        7115,
        {36211, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7115},
        {36211}
    },
    {
        7116,
        {36225, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7116},
        {36225}
    },
    {
        7117,
        {36249, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7117},
        {36249}
    },
    {
        7118,
        {36290, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7118},
        {36290}
    },
    {
        7119,
        {36286, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7119},
        {36286}
    },
    {
        7120,
        {36282, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7120},
        {36282}
    },
    {
        7121,
        {36303, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7121},
        {36303}
    },
    {
        7122,
        {36314, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7122},
        {36314}
    },
    {
        7123,
        {36310, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7123},
        {36310}
    },
    {
        7124,
        {36300, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7124},
        {36300}
    },
    {
        7125,
        {36315, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7125},
        {36315}
    },
    {
        7126,
        {36299, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7126},
        {36299}
    },
    {
        7127,
        {36330, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7127},
        {36330}
    },
    {
        7128,
        {36331, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7128},
        {36331}
    },
    {
        7129,
        {36319, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7129},
        {36319}
    },
    {
        7130,
        {36323, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7130},
        {36323}
    },
    {
        7131,
        {36348, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7131},
        {36348}
    },
    {
        7132,
        {36360, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7132},
        {36360}
    },
    {
        7133,
        {36361, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7133},
        {36361}
    },
    {
        7134,
        {36351, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7134},
        {36351}
    },
    {
        7135,
        {36381, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7135},
        {36381}
    },
    {
        7136,
        {36382, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7136},
        {36382}
    },
    {
        7137,
        {36368, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7137},
        {36368}
    },
    {
        7138,
        {36383, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7138},
        {36383}
    },
    {
        7139,
        {36418, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7139},
        {36418}
    },
    {
        7140,
        {36405, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7140},
        {36405}
    },
    {
        7141,
        {36400, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7141},
        {36400}
    },
    {
        7142,
        {36404, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7142},
        {36404}
    },
    {
        7143,
        {36426, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7143},
        {36426}
    },
    {
        7144,
        {36423, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7144},
        {36423}
    },
    {
        7145,
        {36425, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7145},
        {36425}
    },
    {
        7146,
        {36428, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7146},
        {36428}
    },
    {
        7147,
        {36432, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7147},
        {36432}
    },
    {
        7148,
        {36424, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7148},
        {36424}
    },
    {
        7149,
        {36441, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7149},
        {36441}
    },
    {
        7150,
        {36452, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7150},
        {36452}
    },
    {
        7151,
        {36448, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7151},
        {36448}
    },
    {
        7152,
        {36394, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7152},
        {36394}
    },
    {
        7153,
        {36451, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7153},
        {36451}
    },
    {
        7154,
        {36437, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7154},
        {36437}
    },
    {
        7155,
        {36470, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7155},
        {36470}
    },
    {
        7156,
        {36466, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7156},
        {36466}
    },
    {
        7157,
        {36476, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7157},
        {36476}
    },
    {
        7158,
        {36481, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7158},
        {36481}
    },
    {
        7159,
        {36487, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7159},
        {36487}
    },
    {
        7160,
        {36485, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7160},
        {36485}
    },
    {
        7161,
        {36484, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7161},
        {36484}
    },
    {
        7162,
        {36491, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7162},
        {36491}
    },
    {
        7163,
        {36490, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7163},
        {36490}
    },
    {
        7164,
        {36499, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7164},
        {36499}
    },
    {
        7165,
        {36497, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7165},
        {36497}
    },
    {
        7166,
        {36500, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7166},
        {36500}
    },
    {
        7167,
        {36505, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7167},
        {36505}
    },
    {
        7168,
        {36522, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7168},
        {36522}
    },
    {
        7169,
        {36513, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7169},
        {36513}
    },
    {
        7170,
        {36524, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7170},
        {36524}
    },
    {
        7171,
        {36528, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7171},
        {36528}
    },
    {
        7172,
        {36550, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7172},
        {36550}
    },
    {
        7173,
        {36529, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7173},
        {36529}
    },
    {
        7174,
        {36542, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7174},
        {36542}
    },
    {
        7175,
        {36549, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7175},
        {36549}
    },
    {
        7176,
        {36552, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7176},
        {36552}
    },
    {
        7177,
        {36555, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7177},
        {36555}
    },
    {
        7178,
        {36571, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7178},
        {36571}
    },
    {
        7179,
        {36579, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7179},
        {36579}
    },
    {
        7180,
        {36604, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7180},
        {36604}
    },
    {
        7181,
        {36603, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7181},
        {36603}
    },
    {
        7182,
        {36587, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7182},
        {36587}
    },
    {
        7183,
        {36606, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7183},
        {36606}
    },
    {
        7184,
        {36618, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7184},
        {36618}
    },
    {
        7185,
        {36613, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7185},
        {36613}
    },
    {
        7186,
        {36629, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7186},
        {36629}
    },
    {
        7187,
        {36626, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7187},
        {36626}
    },
    {
        7188,
        {36633, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7188},
        {36633}
    },
    {
        7189,
        {36627, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7189},
        {36627}
    },
    {
        7190,
        {36636, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7190},
        {36636}
    },
    {
        7191,
        {36639, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7191},
        {36639}
    },
    {
        7192,
        {36635, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7192},
        {36635}
    },
    {
        7193,
        {36620, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7193},
        {36620}
    },
    {
        7194,
        {36646, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7194},
        {36646}
    },
    {
        7195,
        {36659, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7195},
        {36659}
    },
    {
        7196,
        {36667, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7196},
        {36667}
    },
    {
        7197,
        {36665, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7197},
        {36665}
    },
    {
        7198,
        {36677, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7198},
        {36677}
    },
    {
        7199,
        {36674, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7199},
        {36674}
    },
    {
        7200,
        {36670, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7200},
        {36670}
    },
    {
        7201,
        {36684, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7201},
        {36684}
    },
    {
        7202,
        {36681, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7202},
        {36681}
    },
    {
        7203,
        {36678, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7203},
        {36678}
    },
    {
        7204,
        {36686, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7204},
        {36686}
    },
    {
        7205,
        {36695, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7205},
        {36695}
    },
    {
        7206,
        {36700, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7206},
        {36700}
    },
    {
        7207,
        {36706, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7207},
        {36706}
    },
    {
        7208,
        {36707, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7208},
        {36707}
    },
    {
        7209,
        {36708, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7209},
        {36708}
    },
    {
        7210,
        {36764, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7210},
        {36764}
    },
    {
        7211,
        {36767, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7211},
        {36767}
    },
    {
        7212,
        {36771, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7212},
        {36771}
    },
    {
        7213,
        {36781, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7213},
        {36781}
    },
    {
        7214,
        {36783, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7214},
        {36783}
    },
    {
        7215,
        {36791, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7215},
        {36791}
    },
    {
        7216,
        {36826, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7216},
        {36826}
    },
    {
        7217,
        {36837, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7217},
        {36837}
    },
    {
        7218,
        {36834, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7218},
        {36834}
    },
    {
        7219,
        {36842, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7219},
        {36842}
    },
    {
        7220,
        {36847, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7220},
        {36847}
    },
    {
        7221,
        {36999, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7221},
        {36999}
    },
    {
        7222,
        {36852, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7222},
        {36852}
    },
    {
        7223,
        {36869, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7223},
        {36869}
    },
    {
        7224,
        {36857, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7224},
        {36857}
    },
    {
        7225,
        {36858, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7225},
        {36858}
    },
    {
        7226,
        {36881, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7226},
        {36881}
    },
    {
        7227,
        {36885, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7227},
        {36885}
    },
    {
        7228,
        {36897, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7228},
        {36897}
    },
    {
        7229,
        {36877, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7229},
        {36877}
    },
    {
        7230,
        {36894, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7230},
        {36894}
    },
    {
        7231,
        {36886, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7231},
        {36886}
    },
    {
        7232,
        {36875, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7232},
        {36875}
    },
    {
        7233,
        {36903, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7233},
        {36903}
    },
    {
        7234,
        {36918, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7234},
        {36918}
    },
    {
        7235,
        {36917, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7235},
        {36917}
    },
    {
        7236,
        {36921, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7236},
        {36921}
    },
    {
        7237,
        {36856, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7237},
        {36856}
    },
    {
        7238,
        {36943, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7238},
        {36943}
    },
    {
        7239,
        {36944, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7239},
        {36944}
    },
    {
        7240,
        {36945, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7240},
        {36945}
    },
    {
        7241,
        {36946, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7241},
        {36946}
    },
    {
        7242,
        {36878, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7242},
        {36878}
    },
    {
        7243,
        {36937, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7243},
        {36937}
    },
    {
        7244,
        {36926, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7244},
        {36926}
    },
    {
        7245,
        {36950, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7245},
        {36950}
    },
    {
        7246,
        {36952, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7246},
        {36952}
    },
    {
        7247,
        {36958, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7247},
        {36958}
    },
    {
        7248,
        {36968, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7248},
        {36968}
    },
    {
        7249,
        {36975, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7249},
        {36975}
    },
    {
        7250,
        {36982, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7250},
        {36982}
    },
    {
        7251,
        {38568, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7251},
        {38568}
    },
    {
        7252,
        {36978, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7252},
        {36978}
    },
    {
        7253,
        {36994, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7253},
        {36994}
    },
    {
        7254,
        {36989, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7254},
        {36989}
    },
    {
        7255,
        {36993, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7255},
        {36993}
    },
    {
        7256,
        {36992, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7256},
        {36992}
    },
    {
        7257,
        {37002, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7257},
        {37002}
    },
    {
        7258,
        {37001, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7258},
        {37001}
    },
    {
        7259,
        {37007, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7259},
        {37007}
    },
    {
        7260,
        {37032, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7260},
        {37032}
    },
    {
        7261,
        {37039, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7261},
        {37039}
    },
    {
        7262,
        {37041, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7262},
        {37041}
    },
    {
        7263,
        {37045, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7263},
        {37045}
    },
    {
        7264,
        {37090, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7264},
        {37090}
    },
    {
        7265,
        {37092, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7265},
        {37092}
    },
    {
        7266,
        {25160, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7266},
        {25160}
    },
    {
        7267,
        {37083, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7267},
        {37083}
    },
    {
        7268,
        {37122, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7268},
        {37122}
    },
    {
        7269,
        {37138, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7269},
        {37138}
    },
    {
        7270,
        {37145, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7270},
        {37145}
    },
    {
        7271,
        {37170, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7271},
        {37170}
    },
    {
        7272,
        {37168, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7272},
        {37168}
    },
    {
        7273,
        {37194, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7273},
        {37194}
    },
    {
        7274,
        {37206, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7274},
        {37206}
    },
    {
        7275,
        {37208, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7275},
        {37208}
    },
    {
        7276,
        {37219, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7276},
        {37219}
    },
    {
        7277,
        {37221, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7277},
        {37221}
    },
    {
        7278,
        {37225, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7278},
        {37225}
    },
    {
        7279,
        {37235, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7279},
        {37235}
    },
    {
        7280,
        {37234, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7280},
        {37234}
    },
    {
        7281,
        {37259, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7281},
        {37259}
    },
    {
        7282,
        {37257, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7282},
        {37257}
    },
    {
        7283,
        {37250, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7283},
        {37250}
    },
    {
        7284,
        {37282, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7284},
        {37282}
    },
    {
        7285,
        {37291, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7285},
        {37291}
    },
    {
        7286,
        {37295, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7286},
        {37295}
    },
    {
        7287,
        {37290, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7287},
        {37290}
    },
    {
        7288,
        {37301, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7288},
        {37301}
    },
    {
        7289,
        {37300, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7289},
        {37300}
    },
    {
        7290,
        {37306, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7290},
        {37306}
    },
    {
        7291,
        {37312, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7291},
        {37312}
    },
    {
        7292,
        {37313, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7292},
        {37313}
    },
    {
        7293,
        {37321, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7293},
        {37321}
    },
    {
        7294,
        {37323, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7294},
        {37323}
    },
    {
        7295,
        {37328, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7295},
        {37328}
    },
    {
        7296,
        {37334, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7296},
        {37334}
    },
    {
        7297,
        {37343, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7297},
        {37343}
    },
    {
        7298,
        {37345, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7298},
        {37345}
    },
    {
        7299,
        {37339, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7299},
        {37339}
    },
    {
        7300,
        {37372, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7300},
        {37372}
    },
    {
        7301,
        {37365, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7301},
        {37365}
    },
    {
        7302,
        {37366, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7302},
        {37366}
    },
    {
        7303,
        {37406, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7303},
        {37406}
    },
    {
        7304,
        {37375, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7304},
        {37375}
    },
    {
        7305,
        {37396, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7305},
        {37396}
    },
    {
        7306,
        {37420, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7306},
        {37420}
    },
    {
        7307,
        {37397, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7307},
        {37397}
    },
    {
        7308,
        {37393, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7308},
        {37393}
    },
    {
        7309,
        {37470, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7309},
        {37470}
    },
    {
        7310,
        {37463, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7310},
        {37463}
    },
    {
        7311,
        {37445, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7311},
        {37445}
    },
    {
        7312,
        {37449, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7312},
        {37449}
    },
    {
        7313,
        {37476, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7313},
        {37476}
    },
    {
        7314,
        {37448, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7314},
        {37448}
    },
    {
        7315,
        {37525, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7315},
        {37525}
    },
    {
        7316,
        {37439, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7316},
        {37439}
    },
    {
        7317,
        {37451, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7317},
        {37451}
    },
    {
        7318,
        {37456, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7318},
        {37456}
    },
    {
        7319,
        {37532, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7319},
        {37532}
    },
    {
        7320,
        {37526, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7320},
        {37526}
    },
    {
        7321,
        {37523, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7321},
        {37523}
    },
    {
        7322,
        {37531, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7322},
        {37531}
    },
    {
        7323,
        {37466, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7323},
        {37466}
    },
    {
        7324,
        {37583, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7324},
        {37583}
    },
    {
        7325,
        {37561, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7325},
        {37561}
    },
    {
        7326,
        {37559, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7326},
        {37559}
    },
    {
        7327,
        {37609, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7327},
        {37609}
    },
    {
        7328,
        {37647, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7328},
        {37647}
    },
    {
        7329,
        {37626, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7329},
        {37626}
    },
    {
        7330,
        {37700, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7330},
        {37700}
    },
    {
        7331,
        {37678, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7331},
        {37678}
    },
    {
        7332,
        {37657, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7332},
        {37657}
    },
    {
        7333,
        {37666, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7333},
        {37666}
    },
    {
        7334,
        {37658, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7334},
        {37658}
    },
    {
        7335,
        {37667, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7335},
        {37667}
    },
    {
        7336,
        {37690, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7336},
        {37690}
    },
    {
        7337,
        {37685, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7337},
        {37685}
    },
    {
        7338,
        {37691, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7338},
        {37691}
    },
    {
        7339,
        {37724, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7339},
        {37724}
    },
    {
        7340,
        {37728, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7340},
        {37728}
    },
    {
        7341,
        {37756, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7341},
        {37756}
    },
    {
        7342,
        {37742, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7342},
        {37742}
    },
    {
        7343,
        {37718, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7343},
        {37718}
    },
    {
        7344,
        {37808, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7344},
        {37808}
    },
    {
        7345,
        {37804, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7345},
        {37804}
    },
    {
        7346,
        {37805, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7346},
        {37805}
    },
    {
        7347,
        {37780, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7347},
        {37780}
    },
    {
        7348,
        {37817, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7348},
        {37817}
    },
    {
        7349,
        {37846, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7349},
        {37846}
    },
    {
        7350,
        {37847, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7350},
        {37847}
    },
    {
        7351,
        {37864, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7351},
        {37864}
    },
    {
        7352,
        {37861, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7352},
        {37861}
    },
    {
        7353,
        {37848, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7353},
        {37848}
    },
    {
        7354,
        {37827, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7354},
        {37827}
    },
    {
        7355,
        {37853, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7355},
        {37853}
    },
    {
        7356,
        {37840, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7356},
        {37840}
    },
    {
        7357,
        {37832, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7357},
        {37832}
    },
    {
        7358,
        {37860, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7358},
        {37860}
    },
    {
        7359,
        {37914, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7359},
        {37914}
    },
    {
        7360,
        {37908, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7360},
        {37908}
    },
    {
        7361,
        {37907, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7361},
        {37907}
    },
    {
        7362,
        {37891, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7362},
        {37891}
    },
    {
        7363,
        {37895, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7363},
        {37895}
    },
    {
        7364,
        {37904, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7364},
        {37904}
    },
    {
        7365,
        {37942, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7365},
        {37942}
    },
    {
        7366,
        {37931, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7366},
        {37931}
    },
    {
        7367,
        {37941, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7367},
        {37941}
    },
    {
        7368,
        {37921, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7368},
        {37921}
    },
    {
        7369,
        {37946, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7369},
        {37946}
    },
    {
        7370,
        {37953, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7370},
        {37953}
    },
    {
        7371,
        {37970, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7371},
        {37970}
    },
    {
        7372,
        {37956, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7372},
        {37956}
    },
    {
        7373,
        {37979, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7373},
        {37979}
    },
    {
        7374,
        {37984, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7374},
        {37984}
    },
    {
        7375,
        {37986, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7375},
        {37986}
    },
    {
        7376,
        {37982, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7376},
        {37982}
    },
    {
        7377,
        {37994, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7377},
        {37994}
    },
    {
        7378,
        {37417, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7378},
        {37417}
    },
    {
        7379,
        {38000, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7379},
        {38000}
    },
    {
        7380,
        {38005, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7380},
        {38005}
    },
    {
        7381,
        {38007, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7381},
        {38007}
    },
    {
        7382,
        {38013, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7382},
        {38013}
    },
    {
        7383,
        {37978, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7383},
        {37978}
    },
    {
        7384,
        {38012, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7384},
        {38012}
    },
    {
        7385,
        {38014, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7385},
        {38014}
    },
    {
        7386,
        {38017, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7386},
        {38017}
    },
    {
        7387,
        {38015, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7387},
        {38015}
    },
    {
        7388,
        {38274, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7388},
        {38274}
    },
    {
        7389,
        {38279, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7389},
        {38279}
    },
    {
        7390,
        {38282, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7390},
        {38282}
    },
    {
        7391,
        {38292, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7391},
        {38292}
    },
    {
        7392,
        {38294, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7392},
        {38294}
    },
    {
        7393,
        {38296, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7393},
        {38296}
    },
    {
        7394,
        {38297, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7394},
        {38297}
    },
    {
        7395,
        {38304, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7395},
        {38304}
    },
    {
        7396,
        {38312, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7396},
        {38312}
    },
    {
        7397,
        {38311, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7397},
        {38311}
    },
    {
        7398,
        {38317, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7398},
        {38317}
    },
    {
        7399,
        {38332, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7399},
        {38332}
    },
    {
        7400,
        {38331, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7400},
        {38331}
    },
    {
        7401,
        {38329, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7401},
        {38329}
    },
    {
        7402,
        {38334, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7402},
        {38334}
    },
    {
        7403,
        {38346, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7403},
        {38346}
    },
    {
        7404,
        {28662, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7404},
        {28662}
    },
    {
        7405,
        {38339, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7405},
        {38339}
    },
    {
        7406,
        {38349, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7406},
        {38349}
    },
    {
        7407,
        {38348, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7407},
        {38348}
    },
    {
        7408,
        {38357, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7408},
        {38357}
    },
    {
        7409,
        {38356, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7409},
        {38356}
    },
    {
        7410,
        {38358, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7410},
        {38358}
    },
    {
        7411,
        {38364, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7411},
        {38364}
    },
    {
        7412,
        {38369, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7412},
        {38369}
    },
    {
        7413,
        {38373, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7413},
        {38373}
    },
    {
        7414,
        {38370, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7414},
        {38370}
    },
    {
        7415,
        {38433, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7415},
        {38433}
    },
    {
        7416,
        {38440, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7416},
        {38440}
    },
    {
        7417,
        {38446, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7417},
        {38446}
    },
    {
        7418,
        {38447, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7418},
        {38447}
    },
    {
        7419,
        {38466, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7419},
        {38466}
    },
    {
        7420,
        {38476, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7420},
        {38476}
    },
    {
        7421,
        {38479, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7421},
        {38479}
    },
    {
        7422,
        {38475, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7422},
        {38475}
    },
    {
        7423,
        {38519, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7423},
        {38519}
    },
    {
        7424,
        {38492, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7424},
        {38492}
    },
    {
        7425,
        {38494, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7425},
        {38494}
    },
    {
        7426,
        {38493, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7426},
        {38493}
    },
    {
        7427,
        {38495, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7427},
        {38495}
    },
    {
        7428,
        {38502, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7428},
        {38502}
    },
    {
        7429,
        {38514, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7429},
        {38514}
    },
    {
        7430,
        {38508, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7430},
        {38508}
    },
    {
        7431,
        {38541, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7431},
        {38541}
    },
    {
        7432,
        {38552, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7432},
        {38552}
    },
    {
        7433,
        {38549, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7433},
        {38549}
    },
    {
        7434,
        {38551, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7434},
        {38551}
    },
    {
        7435,
        {38570, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7435},
        {38570}
    },
    {
        7436,
        {38567, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7436},
        {38567}
    },
    {
        7437,
        {38577, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7437},
        {38577}
    },
    {
        7438,
        {38578, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7438},
        {38578}
    },
    {
        7439,
        {38576, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7439},
        {38576}
    },
    {
        7440,
        {38580, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7440},
        {38580}
    },
    {
        7441,
        {38582, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7441},
        {38582}
    },
    {
        7442,
        {38584, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7442},
        {38584}
    },
    {
        7443,
        {38585, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7443},
        {38585}
    },
    {
        7444,
        {38606, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7444},
        {38606}
    },
    {
        7445,
        {38603, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7445},
        {38603}
    },
    {
        7446,
        {38601, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7446},
        {38601}
    },
    {
        7447,
        {38605, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7447},
        {38605}
    },
    {
        7448,
        {35149, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7448},
        {35149}
    },
    {
        7449,
        {38620, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7449},
        {38620}
    },
    {
        7450,
        {38669, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7450},
        {38669}
    },
    {
        7451,
        {38613, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7451},
        {38613}
    },
    {
        7452,
        {38649, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7452},
        {38649}
    },
    {
        7453,
        {38660, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7453},
        {38660}
    },
    {
        7454,
        {38662, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7454},
        {38662}
    },
    {
        7455,
        {38664, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7455},
        {38664}
    },
    {
        7456,
        {38675, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7456},
        {38675}
    },
    {
        7457,
        {38670, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7457},
        {38670}
    },
    {
        7458,
        {38673, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7458},
        {38673}
    },
    {
        7459,
        {38671, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7459},
        {38671}
    },
    {
        7460,
        {38678, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7460},
        {38678}
    },
    {
        7461,
        {38681, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7461},
        {38681}
    },
    {
        7462,
        {38692, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7462},
        {38692}
    },
    {
        7463,
        {38698, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7463},
        {38698}
    },
    {
        7464,
        {38704, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7464},
        {38704}
    },
    {
        7465,
        {38713, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7465},
        {38713}
    },
    {
        7466,
        {38717, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7466},
        {38717}
    },
    {
        7467,
        {38718, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7467},
        {38718}
    },
    {
        7468,
        {38724, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7468},
        {38724}
    },
    {
        7469,
        {38726, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7469},
        {38726}
    },
    {
        7470,
        {38728, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7470},
        {38728}
    },
    {
        7471,
        {38722, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7471},
        {38722}
    },
    {
        7472,
        {38729, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7472},
        {38729}
    },
    {
        7473,
        {38748, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7473},
        {38748}
    },
    {
        7474,
        {38752, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7474},
        {38752}
    },
    {
        7475,
        {38756, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7475},
        {38756}
    },
    {
        7476,
        {38758, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7476},
        {38758}
    },
    {
        7477,
        {38760, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7477},
        {38760}
    },
    {
        7478,
        {21202, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7478},
        {21202}
    },
    {
        7479,
        {38763, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7479},
        {38763}
    },
    {
        7480,
        {38769, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7480},
        {38769}
    },
    {
        7481,
        {38777, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7481},
        {38777}
    },
    {
        7482,
        {38789, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7482},
        {38789}
    },
    {
        7483,
        {38780, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7483},
        {38780}
    },
    {
        7484,
        {38785, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7484},
        {38785}
    },
    {
        7485,
        {38778, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7485},
        {38778}
    },
    {
        7486,
        {38790, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7486},
        {38790}
    },
    {
        7487,
        {38795, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7487},
        {38795}
    },
    {
        7488,
        {38799, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7488},
        {38799}
    },
    {
        7489,
        {38800, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7489},
        {38800}
    },
    {
        7490,
        {38812, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7490},
        {38812}
    },
    {
        7491,
        {38824, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7491},
        {38824}
    },
    {
        7492,
        {38822, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7492},
        {38822}
    },
    {
        7493,
        {38819, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7493},
        {38819}
    },
    {
        7494,
        {38835, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7494},
        {38835}
    },
    {
        7495,
        {38836, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7495},
        {38836}
    },
    {
        7496,
        {38851, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7496},
        {38851}
    },
    {
        7497,
        {38854, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7497},
        {38854}
    },
    {
        7498,
        {38856, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7498},
        {38856}
    },
    {
        7499,
        {38859, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7499},
        {38859}
    },
    {
        7500,
        {38876, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7500},
        {38876}
    },
    {
        7501,
        {38893, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7501},
        {38893}
    },
    {
        7502,
        {40783, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7502},
        {40783}
    },
    {
        7503,
        {38898, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7503},
        {38898}
    },
    {
        7504,
        {31455, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7504},
        {31455}
    },
    {
        7505,
        {38902, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7505},
        {38902}
    },
    {
        7506,
        {38901, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7506},
        {38901}
    },
    {
        7507,
        {38927, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7507},
        {38927}
    },
    {
        7508,
        {38924, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7508},
        {38924}
    },
    {
        7509,
        {38968, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7509},
        {38968}
    },
    {
        7510,
        {38948, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7510},
        {38948}
    },
    {
        7511,
        {38945, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7511},
        {38945}
    },
    {
        7512,
        {38967, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7512},
        {38967}
    },
    {
        7513,
        {38973, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7513},
        {38973}
    },
    {
        7514,
        {38982, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7514},
        {38982}
    },
    {
        7515,
        {38991, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7515},
        {38991}
    },
    {
        7516,
        {38987, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7516},
        {38987}
    },
    {
        7517,
        {39019, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7517},
        {39019}
    },
    {
        7518,
        {39023, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7518},
        {39023}
    },
    {
        7519,
        {39024, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7519},
        {39024}
    },
    {
        7520,
        {39025, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7520},
        {39025}
    },
    {
        7521,
        {39028, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7521},
        {39028}
    },
    {
        7522,
        {39027, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7522},
        {39027}
    },
    {
        7523,
        {39082, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7523},
        {39082}
    },
    {
        7524,
        {39087, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7524},
        {39087}
    },
    {
        7525,
        {39089, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7525},
        {39089}
    },
    {
        7526,
        {39094, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7526},
        {39094}
    },
    {
        7527,
        {39108, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7527},
        {39108}
    },
    {
        7528,
        {39107, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7528},
        {39107}
    },
    {
        7529,
        {39110, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7529},
        {39110}
    },
    {
        7530,
        {39145, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7530},
        {39145}
    },
    {
        7531,
        {39147, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7531},
        {39147}
    },
    {
        7532,
        {39171, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7532},
        {39171}
    },
    {
        7533,
        {39177, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7533},
        {39177}
    },
    {
        7534,
        {39186, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7534},
        {39186}
    },
    {
        7535,
        {39188, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7535},
        {39188}
    },
    {
        7536,
        {39192, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7536},
        {39192}
    },
    {
        7537,
        {39201, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7537},
        {39201}
    },
    {
        7538,
        {39197, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7538},
        {39197}
    },
    {
        7539,
        {39198, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7539},
        {39198}
    },
    {
        7540,
        {39204, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7540},
        {39204}
    },
    {
        7541,
        {39200, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7541},
        {39200}
    },
    {
        7542,
        {39212, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7542},
        {39212}
    },
    {
        7543,
        {39214, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7543},
        {39214}
    },
    {
        7544,
        {39229, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7544},
        {39229}
    },
    {
        7545,
        {39230, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7545},
        {39230}
    },
    {
        7546,
        {39234, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7546},
        {39234}
    },
    {
        7547,
        {39241, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7547},
        {39241}
    },
    {
        7548,
        {39237, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7548},
        {39237}
    },
    {
        7549,
        {39248, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7549},
        {39248}
    },
    {
        7550,
        {39243, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7550},
        {39243}
    },
    {
        7551,
        {39249, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7551},
        {39249}
    },
    {
        7552,
        {39250, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7552},
        {39250}
    },
    {
        7553,
        {39244, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7553},
        {39244}
    },
    {
        7554,
        {39253, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7554},
        {39253}
    },
    {
        7555,
        {39319, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7555},
        {39319}
    },
    {
        7556,
        {39320, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7556},
        {39320}
    },
    {
        7557,
        {39333, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7557},
        {39333}
    },
    {
        7558,
        {39341, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7558},
        {39341}
    },
    {
        7559,
        {39342, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7559},
        {39342}
    },
    {
        7560,
        {39356, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7560},
        {39356}
    },
    {
        7561,
        {39391, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7561},
        {39391}
    },
    {
        7562,
        {39387, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7562},
        {39387}
    },
    {
        7563,
        {39389, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7563},
        {39389}
    },
    {
        7564,
        {39384, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7564},
        {39384}
    },
    {
        7565,
        {39377, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7565},
        {39377}
    },
    {
        7566,
        {39405, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7566},
        {39405}
    },
    {
        7567,
        {39406, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7567},
        {39406}
    },
    {
        7568,
        {39409, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7568},
        {39409}
    },
    {
        7569,
        {39410, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7569},
        {39410}
    },
    {
        7570,
        {39419, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7570},
        {39419}
    },
    {
        7571,
        {39416, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7571},
        {39416}
    },
    {
        7572,
        {39425, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7572},
        {39425}
    },
    {
        7573,
        {39439, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7573},
        {39439}
    },
    {
        7574,
        {39429, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7574},
        {39429}
    },
    {
        7575,
        {39394, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7575},
        {39394}
    },
    {
        7576,
        {39449, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7576},
        {39449}
    },
    {
        7577,
        {39467, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7577},
        {39467}
    },
    {
        7578,
        {39479, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7578},
        {39479}
    },
    {
        7579,
        {39493, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7579},
        {39493}
    },
    {
        7580,
        {39490, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7580},
        {39490}
    },
    {
        7581,
        {39488, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7581},
        {39488}
    },
    {
        7582,
        {39491, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7582},
        {39491}
    },
    {
        7583,
        {39486, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7583},
        {39486}
    },
    {
        7584,
        {39509, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7584},
        {39509}
    },
    {
        7585,
        {39501, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7585},
        {39501}
    },
    {
        7586,
        {39515, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7586},
        {39515}
    },
    {
        7587,
        {39511, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7587},
        {39511}
    },
    {
        7588,
        {39519, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7588},
        {39519}
    },
    {
        7589,
        {39522, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7589},
        {39522}
    },
    {
        7590,
        {39525, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7590},
        {39525}
    },
    {
        7591,
        {39524, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7591},
        {39524}
    },
    {
        7592,
        {39529, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7592},
        {39529}
    },
    {
        7593,
        {39531, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7593},
        {39531}
    },
    {
        7594,
        {39530, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7594},
        {39530}
    },
    {
        7595,
        {39597, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7595},
        {39597}
    },
    {
        7596,
        {39600, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7596},
        {39600}
    },
    {
        7597,
        {39612, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7597},
        {39612}
    },
    {
        7598,
        {39616, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7598},
        {39616}
    },
    {
        7599,
        {39631, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7599},
        {39631}
    },
    {
        7600,
        {39633, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7600},
        {39633}
    },
    {
        7601,
        {39635, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7601},
        {39635}
    },
    {
        7602,
        {39636, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7602},
        {39636}
    },
    {
        7603,
        {39646, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7603},
        {39646}
    },
    {
        7604,
        {39647, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7604},
        {39647}
    },
    {
        7605,
        {39650, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7605},
        {39650}
    },
    {
        7606,
        {39651, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7606},
        {39651}
    },
    {
        7607,
        {39654, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7607},
        {39654}
    },
    {
        7608,
        {39663, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7608},
        {39663}
    },
    {
        7609,
        {39659, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7609},
        {39659}
    },
    {
        7610,
        {39662, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7610},
        {39662}
    },
    {
        7611,
        {39668, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7611},
        {39668}
    },
    {
        7612,
        {39665, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7612},
        {39665}
    },
    {
        7613,
        {39671, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7613},
        {39671}
    },
    {
        7614,
        {39675, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7614},
        {39675}
    },
    {
        7615,
        {39686, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7615},
        {39686}
    },
    {
        7616,
        {39704, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7616},
        {39704}
    },
    {
        7617,
        {39706, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7617},
        {39706}
    },
    {
        7618,
        {39711, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7618},
        {39711}
    },
    {
        7619,
        {39714, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7619},
        {39714}
    },
    {
        7620,
        {39715, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7620},
        {39715}
    },
    {
        7621,
        {39717, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7621},
        {39717}
    },
    {
        7622,
        {39719, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7622},
        {39719}
    },
    {
        7623,
        {39720, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7623},
        {39720}
    },
    {
        7624,
        {39721, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7624},
        {39721}
    },
    {
        7625,
        {39722, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7625},
        {39722}
    },
    {
        7626,
        {39726, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7626},
        {39726}
    },
    {
        7627,
        {39727, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7627},
        {39727}
    },
    {
        7628,
        {39730, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7628},
        {39730}
    },
    {
        7629,
        {39748, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7629},
        {39748}
    },
    {
        7630,
        {39747, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7630},
        {39747}
    },
    {
        7631,
        {39759, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7631},
        {39759}
    },
    {
        7632,
        {39757, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7632},
        {39757}
    },
    {
        7633,
        {39758, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7633},
        {39758}
    },
    {
        7634,
        {39761, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7634},
        {39761}
    },
    {
        7635,
        {39768, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7635},
        {39768}
    },
    {
        7636,
        {39796, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7636},
        {39796}
    },
    {
        7637,
        {39827, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7637},
        {39827}
    },
    {
        7638,
        {39811, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7638},
        {39811}
    },
    {
        7639,
        {39825, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7639},
        {39825}
    },
    {
        7640,
        {39830, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7640},
        {39830}
    },
    {
        7641,
        {39831, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7641},
        {39831}
    },
    {
        7642,
        {39839, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7642},
        {39839}
    },
    {
        7643,
        {39840, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7643},
        {39840}
    },
    {
        7644,
        {39848, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7644},
        {39848}
    },
    {
        7645,
        {39860, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7645},
        {39860}
    },
    {
        7646,
        {39872, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7646},
        {39872}
    },
    {
        7647,
        {39882, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7647},
        {39882}
    },
    {
        7648,
        {39865, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7648},
        {39865}
    },
    {
        7649,
        {39878, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7649},
        {39878}
    },
    {
        7650,
        {39887, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7650},
        {39887}
    },
    {
        7651,
        {39889, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7651},
        {39889}
    },
    {
        7652,
        {39890, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7652},
        {39890}
    },
    {
        7653,
        {39907, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7653},
        {39907}
    },
    {
        7654,
        {39906, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7654},
        {39906}
    },
    {
        7655,
        {39908, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7655},
        {39908}
    },
    {
        7656,
        {39892, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7656},
        {39892}
    },
    {
        7657,
        {39905, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7657},
        {39905}
    },
    {
        7658,
        {39994, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7658},
        {39994}
    },
    {
        7659,
        {39922, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7659},
        {39922}
    },
    {
        7660,
        {39921, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7660},
        {39921}
    },
    {
        7661,
        {39920, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7661},
        {39920}
    },
    {
        7662,
        {39957, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7662},
        {39957}
    },
    {
        7663,
        {39956, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7663},
        {39956}
    },
    {
        7664,
        {39945, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7664},
        {39945}
    },
    {
        7665,
        {39955, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7665},
        {39955}
    },
    {
        7666,
        {39948, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7666},
        {39948}
    },
    {
        7667,
        {39942, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7667},
        {39942}
    },
    {
        7668,
        {39944, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7668},
        {39944}
    },
    {
        7669,
        {39954, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7669},
        {39954}
    },
    {
        7670,
        {39946, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7670},
        {39946}
    },
    {
        7671,
        {39940, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7671},
        {39940}
    },
    {
        7672,
        {39982, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7672},
        {39982}
    },
    {
        7673,
        {39963, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7673},
        {39963}
    },
    {
        7674,
        {39973, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7674},
        {39973}
    },
    {
        7675,
        {39972, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7675},
        {39972}
    },
    {
        7676,
        {39969, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7676},
        {39969}
    },
    {
        7677,
        {39984, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7677},
        {39984}
    },
    {
        7678,
        {40007, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7678},
        {40007}
    },
    {
        7679,
        {39986, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7679},
        {39986}
    },
    {
        7680,
        {40006, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7680},
        {40006}
    },
    {
        7681,
        {39998, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7681},
        {39998}
    },
    {
        7682,
        {40026, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7682},
        {40026}
    },
    {
        7683,
        {40032, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7683},
        {40032}
    },
    {
        7684,
        {40039, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7684},
        {40039}
    },
    {
        7685,
        {40054, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7685},
        {40054}
    },
    {
        7686,
        {40056, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7686},
        {40056}
    },
    {
        7687,
        {40167, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7687},
        {40167}
    },
    {
        7688,
        {40172, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7688},
        {40172}
    },
    {
        7689,
        {40176, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7689},
        {40176}
    },
    {
        7690,
        {40201, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7690},
        {40201}
    },
    {
        7691,
        {40200, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7691},
        {40200}
    },
    {
        7692,
        {40171, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7692},
        {40171}
    },
    {
        7693,
        {40195, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7693},
        {40195}
    },
    {
        7694,
        {40198, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7694},
        {40198}
    },
    {
        7695,
        {40234, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7695},
        {40234}
    },
    {
        7696,
        {40230, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7696},
        {40230}
    },
    {
        7697,
        {40367, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7697},
        {40367}
    },
    {
        7698,
        {40227, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7698},
        {40227}
    },
    {
        7699,
        {40223, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7699},
        {40223}
    },
    {
        7700,
        {40260, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7700},
        {40260}
    },
    {
        7701,
        {40213, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7701},
        {40213}
    },
    {
        7702,
        {40210, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7702},
        {40210}
    },
    {
        7703,
        {40257, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7703},
        {40257}
    },
    {
        7704,
        {40255, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7704},
        {40255}
    },
    {
        7705,
        {40254, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7705},
        {40254}
    },
    {
        7706,
        {40262, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7706},
        {40262}
    },
    {
        7707,
        {40264, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7707},
        {40264}
    },
    {
        7708,
        {40285, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7708},
        {40285}
    },
    {
        7709,
        {40286, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7709},
        {40286}
    },
    {
        7710,
        {40292, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7710},
        {40292}
    },
    {
        7711,
        {40273, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7711},
        {40273}
    },
    {
        7712,
        {40272, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7712},
        {40272}
    },
    {
        7713,
        {40281, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7713},
        {40281}
    },
    {
        7714,
        {40306, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7714},
        {40306}
    },
    {
        7715,
        {40329, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7715},
        {40329}
    },
    {
        7716,
        {40327, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7716},
        {40327}
    },
    {
        7717,
        {40363, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7717},
        {40363}
    },
    {
        7718,
        {40303, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7718},
        {40303}
    },
    {
        7719,
        {40314, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7719},
        {40314}
    },
    {
        7720,
        {40346, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7720},
        {40346}
    },
    {
        7721,
        {40356, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7721},
        {40356}
    },
    {
        7722,
        {40361, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7722},
        {40361}
    },
    {
        7723,
        {40370, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7723},
        {40370}
    },
    {
        7724,
        {40388, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7724},
        {40388}
    },
    {
        7725,
        {40385, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7725},
        {40385}
    },
    {
        7726,
        {40379, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7726},
        {40379}
    },
    {
        7727,
        {40376, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7727},
        {40376}
    },
    {
        7728,
        {40378, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7728},
        {40378}
    },
    {
        7729,
        {40390, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7729},
        {40390}
    },
    {
        7730,
        {40399, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7730},
        {40399}
    },
    {
        7731,
        {40386, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7731},
        {40386}
    },
    {
        7732,
        {40409, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7732},
        {40409}
    },
    {
        7733,
        {40403, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7733},
        {40403}
    },
    {
        7734,
        {40440, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7734},
        {40440}
    },
    {
        7735,
        {40422, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7735},
        {40422}
    },
    {
        7736,
        {40429, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7736},
        {40429}
    },
    {
        7737,
        {40431, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7737},
        {40431}
    },
    {
        7738,
        {40445, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7738},
        {40445}
    },
    {
        7739,
        {40474, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7739},
        {40474}
    },
    {
        7740,
        {40475, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7740},
        {40475}
    },
    {
        7741,
        {40478, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7741},
        {40478}
    },
    {
        7742,
        {40565, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7742},
        {40565}
    },
    {
        7743,
        {40569, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7743},
        {40569}
    },
    {
        7744,
        {40573, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7744},
        {40573}
    },
    {
        7745,
        {40577, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7745},
        {40577}
    },
    {
        7746,
        {40584, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7746},
        {40584}
    },
    {
        7747,
        {40587, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7747},
        {40587}
    },
    {
        7748,
        {40588, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7748},
        {40588}
    },
    {
        7749,
        {40594, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7749},
        {40594}
    },
    {
        7750,
        {40597, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7750},
        {40597}
    },
    {
        7751,
        {40593, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7751},
        {40593}
    },
    {
        7752,
        {40605, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7752},
        {40605}
    },
    {
        7753,
        {40613, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7753},
        {40613}
    },
    {
        7754,
        {40617, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7754},
        {40617}
    },
    {
        7755,
        {40632, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7755},
        {40632}
    },
    {
        7756,
        {40618, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7756},
        {40618}
    },
    {
        7757,
        {40621, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7757},
        {40621}
    },
    {
        7758,
        {38753, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7758},
        {38753}
    },
    {
        7759,
        {40652, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7759},
        {40652}
    },
    {
        7760,
        {40654, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7760},
        {40654}
    },
    {
        7761,
        {40655, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7761},
        {40655}
    },
    {
        7762,
        {40656, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7762},
        {40656}
    },
    {
        7763,
        {40660, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7763},
        {40660}
    },
    {
        7764,
        {40668, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7764},
        {40668}
    },
    {
        7765,
        {40670, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7765},
        {40670}
    },
    {
        7766,
        {40669, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7766},
        {40669}
    },
    {
        7767,
        {40672, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7767},
        {40672}
    },
    {
        7768,
        {40677, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7768},
        {40677}
    },
    {
        7769,
        {40680, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7769},
        {40680}
    },
    {
        7770,
        {40687, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7770},
        {40687}
    },
    {
        7771,
        {40692, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7771},
        {40692}
    },
    {
        7772,
        {40694, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7772},
        {40694}
    },
    {
        7773,
        {40695, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7773},
        {40695}
    },
    {
        7774,
        {40697, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7774},
        {40697}
    },
    {
        7775,
        {40699, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7775},
        {40699}
    },
    {
        7776,
        {40700, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7776},
        {40700}
    },
    {
        7777,
        {40701, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7777},
        {40701}
    },
    {
        7778,
        {40711, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7778},
        {40711}
    },
    {
        7779,
        {40712, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7779},
        {40712}
    },
    {
        7780,
        {30391, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7780},
        {30391}
    },
    {
        7781,
        {40725, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7781},
        {40725}
    },
    {
        7782,
        {40737, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7782},
        {40737}
    },
    {
        7783,
        {40748, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7783},
        {40748}
    },
    {
        7784,
        {40766, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7784},
        {40766}
    },
    {
        7785,
        {40778, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7785},
        {40778}
    },
    {
        7786,
        {40786, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7786},
        {40786}
    },
    {
        7787,
        {40788, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7787},
        {40788}
    },
    {
        7788,
        {40803, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7788},
        {40803}
    },
    {
        7789,
        {40799, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7789},
        {40799}
    },
    {
        7790,
        {40800, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7790},
        {40800}
    },
    {
        7791,
        {40801, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7791},
        {40801}
    },
    {
        7792,
        {40806, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7792},
        {40806}
    },
    {
        7793,
        {40807, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7793},
        {40807}
    },
    {
        7794,
        {40812, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7794},
        {40812}
    },
    {
        7795,
        {40810, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7795},
        {40810}
    },
    {
        7796,
        {40823, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7796},
        {40823}
    },
    {
        7797,
        {40818, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7797},
        {40818}
    },
    {
        7798,
        {40822, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7798},
        {40822}
    },
    {
        7799,
        {40853, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7799},
        {40853}
    },
    {
        7800,
        {40860, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7800},
        {40860}
    },
    {
        7801,
        {40864, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7801},
        {40864}
    },
    {
        7802,
        {22575, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7802},
        {22575}
    },
    {
        7803,
        {27079, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7803},
        {27079}
    },
    {
        7804,
        {36953, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7804},
        {36953}
    },
    {
        7805,
        {29796, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7805},
        {29796}
    },
    {
        7806,
        {20956, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7806},
        {20956}
    },
    {
        7807,
        {29081, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_2,
        1,
        {7807},
        {29081}
    },
    {
        7808,
        {22099, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1522},
        {22040}
    },
    {
        7809,
        {24034, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3074},
        {24035}
    },
    {
        7810,
        {24084, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7811,
        {24088, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7812,
        {24152, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7813,
        {24158, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7814,
        {24254, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7815,
        {63784, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4336},
        {24266}
    },
    {
        7816,
        {24267, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7817,
        {24313, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3658},
        {36843}
    },
    {
        7818,
        {24320, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1702},
        {38283}
    },
    {
        7819,
        {24322, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7820,
        {24327, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7821,
        {24349, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7822,
        {24355, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7823,
        {24372, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7824,
        {24374, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7825,
        {24381, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7826,
        {24384, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7827,
        {24389, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7828,
        {24404, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7829,
        {24408, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7830,
        {24420, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7831,
        {24423, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7832,
        {24445, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7833,
        {24457, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7834,
        {24476, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7835,
        {24487, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3908},
        {36941}
    },
    {
        7836,
        {24495, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {7143},
        {36426}
    },
    {
        7837,
        {24501, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3296},
        {24500}
    },
    {
        7838,
        {24503, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3510},
        {24499}
    },
    {
        7839,
        {24521, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7840,
        {24542, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7841,
        {24545, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7842,
        {24553, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {5143},
        {24625}
    },
    {
        7843,
        {24589, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7844,
        {24596, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7845,
        {24600, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2200},
        {22266}
    },
    {
        7846,
        {24627, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7847,
        {24629, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7848,
        {24647, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7849,
        {64061, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1686},
        {24724}
    },
    {
        7850,
        {24733, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7851,
        {24734, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2238},
        {35492}
    },
    {
        7852,
        {24779, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7853,
        {24788, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7854,
        {24789, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7855,
        {24797, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7856,
        {24824, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7857,
        {24860, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7858,
        {24875, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7859,
        {24880, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7860,
        {24887, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7861,
        {64062, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1711},
        {24936}
    },
    {
        7862,
        {24973, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7863,
        {64063, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3101},
        {24974}
    },
    {
        7864,
        {25020, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7865,
        {25017, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7866,
        {64064, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3297},
        {25074}
    },
    {
        7867,
        {25122, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7868,
        {25150, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4106},
        {25147}
    },
    {
        7869,
        {25155, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7870,
        {25174, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7871,
        {25178, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7872,
        {25199, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4504},
        {20566}
    },
    {
        7873,
        {25221, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2275},
        {25304}
    },
    {
        7874,
        {25284, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7875,
        {25302, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7876,
        {25340, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7877,
        {25354, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4337},
        {24324}
    },
    {
        7878,
        {25368, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {5698},
        {27615}
    },
    {
        7879,
        {25401, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7880,
        {25411, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7881,
        {25445, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4374},
        {33109}
    },
    {
        7882,
        {25468, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7883,
        {25573, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7884,
        {25581, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2106},
        {25522}
    },
    {
        7885,
        {25589, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7886,
        {25616, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7887,
        {25620, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3070},
        {25531}
    },
    {
        7888,
        {25634, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7889,
        {25721, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7890,
        {25681, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3336},
        {25524}
    },
    {
        7891,
        {25696, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {6424},
        {32317}
    },
    {
        7892,
        {25709, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7893,
        {25806, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7894,
        {25790, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7895,
        {25791, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7896,
        {25796, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7897,
        {25802, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2133},
        {25731}
    },
    {
        7898,
        {25808, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7899,
        {25847, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7900,
        {25851, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7901,
        {25890, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {5382},
        {25861}
    },
    {
        7902,
        {25897, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7903,
        {64065, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3809},
        {25935}
    },
    {
        7904,
        {25959, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7905,
        {26013, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7906,
        {64066, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1876},
        {26082}
    },
    {
        7907,
        {26112, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7908,
        {26121, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7909,
        {26133, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7910,
        {26142, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {5886},
        {28851}
    },
    {
        7911,
        {26170, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {5886},
        {28851}
    },
    {
        7912,
        {26146, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7913,
        {26148, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7914,
        {26155, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7915,
        {26160, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2920},
        {26159}
    },
    {
        7916,
        {26161, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7917,
        {26163, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7918,
        {26363, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2751},
        {26119}
    },
    {
        7919,
        {26184, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {6118},
        {30350}
    },
    {
        7920,
        {26188, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7921,
        {143812, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {5438},
        {26210}
    },
    {
        7922,
        {26201, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7923,
        {26202, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3718},
        {26217}
    },
    {
        7924,
        {26209, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7925,
        {26213, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {6119},
        {30358}
    },
    {
        7926,
        {26227, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {5439},
        {26224}
    },
    {
        7927,
        {26231, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7928,
        {26232, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7929,
        {26253, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7930,
        {64067, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2706},
        {26257}
    },
    {
        7931,
        {26272, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {6120},
        {30355}
    },
    {
        7932,
        {26290, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7933,
        {26299, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2109},
        {26223}
    },
    {
        7934,
        {26310, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4310},
        {26278}
    },
    {
        7935,
        {26312, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7936,
        {15138, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {5458},
        {26342}
    },
    {
        7937,
        {26331, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7938,
        {26344, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7939,
        {26362, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3073},
        {26361}
    },
    {
        7940,
        {26387, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7941,
        {63785, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4338},
        {26391}
    },
    {
        7942,
        {26419, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7943,
        {26470, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2908},
        {26441}
    },
    {
        7944,
        {26439, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7945,
        {26440, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7946,
        {26491, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7947,
        {26497, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7948,
        {26515, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7949,
        {26520, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7950,
        {26523, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7951,
        {26555, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7952,
        {26617, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1724},
        {26623}
    },
    {
        7953,
        {26560, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7954,
        {26583, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2755},
        {26494}
    },
    {
        7955,
        {26620, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7956,
        {26625, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4129},
        {26611}
    },
    {
        7957,
        {26706, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2079},
        {26705}
    },
    {
        7958,
        {26653, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7959,
        {26668, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7960,
        {26673, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7961,
        {26715, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7962,
        {26738, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7963,
        {26741, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7964,
        {64068, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3634},
        {26757}
    },
    {
        7965,
        {26787, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7966,
        {26789, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2755},
        {26494}
    },
    {
        7967,
        {26802, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7968,
        {26824, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7969,
        {26832, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7970,
        {26856, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7971,
        {26861, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7972,
        {26864, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7973,
        {26865, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {5578},
        {26974}
    },
    {
        7974,
        {26876, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7975,
        {26890, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7976,
        {26953, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7977,
        {144836, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7978,
        {26933, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7979,
        {26946, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7980,
        {26967, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7981,
        {26979, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7982,
        {26980, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7983,
        {26984, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7984,
        {27008, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7985,
        {64020, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {5655},
        {27397}
    },
    {
        7986,
        {27045, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7987,
        {27053, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7988,
        {27087, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7989,
        {15286, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {5591},
        {27082}
    },
    {
        7990,
        {15299, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2568},
        {28422}
    },
    {
        7991,
        {27106, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7992,
        {27113, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1712},
        {27010}
    },
    {
        7993,
        {27114, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1712},
        {27010}
    },
    {
        7994,
        {27125, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7995,
        {27126, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {5541},
        {26914}
    },
    {
        7996,
        {27151, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7997,
        {27157, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7998,
        {145215, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        7999,
        {27195, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8000,
        {27198, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8001,
        {27205, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4088},
        {27169}
    },
    {
        8002,
        {27216, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8003,
        {27222, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8004,
        {27227, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8005,
        {27243, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1600},
        {27178}
    },
    {
        8006,
        {27251, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8007,
        {145251, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2600},
        {29237}
    },
    {
        8008,
        {27273, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8009,
        {27284, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {5518},
        {26723}
    },
    {
        8010,
        {27293, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {5568},
        {26987}
    },
    {
        8011,
        {27294, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8012,
        {27301, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {6642},
        {33380}
    },
    {
        8013,
        {27364, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8014,
        {27367, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8015,
        {15375, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8016,
        {63773, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4218},
        {27396}
    },
    {
        8017,
        {27419, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3610},
        {26487}
    },
    {
        8018,
        {27422, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {5657},
        {27386}
    },
    {
        8019,
        {27436, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8020,
        {27445, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1815},
        {27454}
    },
    {
        8021,
        {27462, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8022,
        {27478, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1863},
        {21916}
    },
    {
        8023,
        {27488, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8024,
        {27493, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3919},
        {27497}
    },
    {
        8025,
        {27495, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8026,
        {27511, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4311},
        {27508}
    },
    {
        8027,
        {27522, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8028,
        {27561, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8029,
        {27565, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8030,
        {63856, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2449},
        {27578}
    },
    {
        8031,
        {27599, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4019},
        {27598}
    },
    {
        8032,
        {27606, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8033,
        {27607, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3756},
        {27608}
    },
    {
        8034,
        {27647, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8035,
        {27653, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8036,
        {27664, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {5098},
        {24342}
    },
    {
        8037,
        {27699, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8038,
        {27737, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1591},
        {27738}
    },
    {
        8039,
        {27740, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8040,
        {27818, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {5873},
        {28696}
    },
    {
        8041,
        {27764, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {5721},
        {27763}
    },
    {
        8042,
        {27766, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8043,
        {27781, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8044,
        {27782, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8045,
        {27800, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8046,
        {27804, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8047,
        {27899, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8048,
        {27846, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8049,
        {27860, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8050,
        {27872, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8051,
        {27883, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8052,
        {27886, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8053,
        {146686, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3158},
        {27888}
    },
    {
        8054,
        {27908, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8055,
        {27918, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8056,
        {27950, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8057,
        {27953, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8058,
        {27961, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8059,
        {27967, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1591},
        {27738}
    },
    {
        8060,
        {27992, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8061,
        {28005, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8062,
        {64069, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1695},
        {28023}
    },
    {
        8063,
        {28034, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8064,
        {28039, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8065,
        {28041, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2761},
        {28169}
    },
    {
        8066,
        {28052, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8067,
        {28074, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8068,
        {28076, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8069,
        {28095, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8070,
        {28100, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8071,
        {28118, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8072,
        {28122, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4291},
        {28057}
    },
    {
        8073,
        {28123, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8074,
        {28125, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8075,
        {28156, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8076,
        {64070, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2708},
        {28186}
    },
    {
        8077,
        {28212, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1764},
        {28167}
    },
    {
        8078,
        {28228, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8079,
        {28252, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8080,
        {28254, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8081,
        {28331, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1622},
        {28201}
    },
    {
        8082,
        {28337, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8083,
        {28353, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8084,
        {28359, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8085,
        {28366, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8086,
        {28432, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8087,
        {28442, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8088,
        {64071, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1818},
        {28450}
    },
    {
        8089,
        {28458, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8090,
        {28463, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8091,
        {28467, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8092,
        {28497, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3679},
        {28300}
    },
    {
        8093,
        {28505, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8094,
        {28510, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8095,
        {28513, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8096,
        {28514, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8097,
        {28542, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8098,
        {28552, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8099,
        {28556, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8100,
        {28557, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8101,
        {28564, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2287},
        {28009}
    },
    {
        8102,
        {28576, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8103,
        {28583, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8104,
        {28598, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8105,
        {28604, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8106,
        {28615, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8107,
        {28618, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8108,
        {28665, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8109,
        {28656, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8110,
        {28661, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {5859},
        {28657}
    },
    {
        8111,
        {28677, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8112,
        {28678, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3511},
        {28060}
    },
    {
        8113,
        {28712, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2918},
        {28716}
    },
    {
        8114,
        {28746, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8115,
        {28765, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8116,
        {28766, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8117,
        {28750, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8118,
        {28772, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8119,
        {28789, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4307},
        {38666}
    },
    {
        8120,
        {28805, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8121,
        {28836, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2753},
        {26157}
    },
    {
        8122,
        {28843, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8123,
        {28855, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8124,
        {28884, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {5883},
        {28913}
    },
    {
        8125,
        {28888, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8126,
        {28900, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8127,
        {28943, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4437},
        {20127}
    },
    {
        8128,
        {28971, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8129,
        {28958, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8130,
        {28960, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8131,
        {28974, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8132,
        {28976, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1578},
        {28948}
    },
    {
        8133,
        {28998, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8134,
        {28999, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {5443},
        {26249}
    },
    {
        8135,
        {29009, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2586},
        {29038}
    },
    {
        8136,
        {64072, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2586},
        {29038}
    },
    {
        8137,
        {29010, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8138,
        {29020, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8139,
        {29024, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8140,
        {29032, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8141,
        {64021, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {5898},
        {29064}
    },
    {
        8142,
        {29061, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8143,
        {29063, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8144,
        {29074, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8145,
        {29121, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8146,
        {29114, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {5911},
        {29113}
    },
    {
        8147,
        {29124, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1578},
        {28948}
    },
    {
        8148,
        {29182, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8149,
        {29184, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1741},
        {36203}
    },
    {
        8150,
        {29205, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4662},
        {29166}
    },
    {
        8151,
        {29269, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3082},
        {31379}
    },
    {
        8152,
        {29270, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8153,
        {15935, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {5946},
        {29346}
    },
    {
        8154,
        {29325, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8155,
        {29339, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8156,
        {29374, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8157,
        {29376, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2810},
        {29366}
    },
    {
        8158,
        {29435, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8159,
        {149489, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8160,
        {29479, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8161,
        {29480, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1581},
        {29503}
    },
    {
        8162,
        {64022, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3281},
        {29482}
    },
    {
        8163,
        {29520, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8164,
        {29542, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8165,
        {29564, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8166,
        {29589, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8167,
        {29599, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8168,
        {29600, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8169,
        {29602, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8170,
        {29606, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8171,
        {29611, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8172,
        {29641, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8173,
        {29647, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8174,
        {29654, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8175,
        {29657, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8176,
        {29667, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8177,
        {29673, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8178,
        {29703, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8179,
        {29706, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8180,
        {29722, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8181,
        {29723, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8182,
        {64074, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3187},
        {29730}
    },
    {
        8183,
        {29734, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8184,
        {29736, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8185,
        {29738, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8186,
        {29739, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8187,
        {29740, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8188,
        {29742, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8189,
        {29743, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8190,
        {29744, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8191,
        {29764, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8192,
        {29766, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8193,
        {29767, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {5983},
        {29619}
    },
    {
        8194,
        {29771, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8195,
        {29783, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8196,
        {29794, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8197,
        {29803, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8198,
        {29805, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8199,
        {29830, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1948},
        {29699}
    },
    {
        8200,
        {29831, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8201,
        {29833, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8202,
        {29848, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8203,
        {29852, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8204,
        {29855, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8205,
        {29859, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8206,
        {29840, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8207,
        {29862, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8208,
        {29864, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8209,
        {29865, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8210,
        {29877, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8211,
        {29887, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8212,
        {29896, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8213,
        {29897, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8214,
        {29914, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8215,
        {29951, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8216,
        {29953, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3810},
        {29942}
    },
    {
        8217,
        {29975, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8218,
        {29999, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3582},
        {23527}
    },
    {
        8219,
        {30063, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8220,
        {30073, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8221,
        {30098, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8222,
        {16242, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8223,
        {30158, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8224,
        {30180, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8225,
        {30208, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8226,
        {30210, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1411},
        {21782}
    },
    {
        8227,
        {30216, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8228,
        {30229, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8229,
        {30230, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8230,
        {30233, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8231,
        {30238, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8232,
        {30253, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8233,
        {30261, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8234,
        {30275, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8235,
        {30283, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8236,
        {30308, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8237,
        {30309, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8238,
        {30317, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8239,
        {30319, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8240,
        {30321, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8241,
        {30337, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {5535},
        {26765}
    },
    {
        8242,
        {30363, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8243,
        {30365, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8244,
        {30366, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8245,
        {30374, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {6118},
        {30350}
    },
    {
        8246,
        {30378, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8247,
        {30390, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8248,
        {30405, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8249,
        {30412, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4373},
        {30871}
    },
    {
        8250,
        {30414, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8251,
        {30420, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8252,
        {30438, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8253,
        {30449, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8254,
        {30460, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8255,
        {30474, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8256,
        {30489, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8257,
        {30516, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8258,
        {30518, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8259,
        {30534, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8260,
        {30541, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8261,
        {30542, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8262,
        {30556, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8263,
        {30559, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8264,
        {30562, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8265,
        {30586, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8266,
        {30592, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8267,
        {30612, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8268,
        {30634, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {6162},
        {30624}
    },
    {
        8269,
        {30688, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8270,
        {30765, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8271,
        {152718, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8272,
        {30787, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2612},
        {26417}
    },
    {
        8273,
        {30798, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8274,
        {30799, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2167},
        {30740}
    },
    {
        8275,
        {30801, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8276,
        {30824, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8277,
        {30830, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4822},
        {22470}
    },
    {
        8278,
        {64075, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3739},
        {30865}
    },
    {
        8279,
        {30896, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8280,
        {152846, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2425},
        {30869}
    },
    {
        8281,
        {30893, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8282,
        {30948, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8283,
        {30962, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8284,
        {30976, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1819},
        {28567}
    },
    {
        8285,
        {30967, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8286,
        {31004, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8287,
        {31022, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3951},
        {30770}
    },
    {
        8288,
        {31025, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8289,
        {31028, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8290,
        {64076, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2587},
        {31038}
    },
    {
        8291,
        {64077, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2510},
        {31049}
    },
    {
        8292,
        {31045, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4172},
        {22934}
    },
    {
        8293,
        {31046, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8294,
        {64078, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1886},
        {31048}
    },
    {
        8295,
        {64079, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4155},
        {31056}
    },
    {
        8296,
        {64080, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3043},
        {31062}
    },
    {
        8297,
        {31068, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8298,
        {64081, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2674},
        {31069}
    },
    {
        8299,
        {64025, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2851},
        {31070}
    },
    {
        8300,
        {64026, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2770},
        {31077}
    },
    {
        8301,
        {31097, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8302,
        {64082, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1647},
        {31117}
    },
    {
        8303,
        {64083, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3376},
        {31118}
    },
    {
        8304,
        {64027, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3854},
        {31119}
    },
    {
        8305,
        {31128, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8306,
        {31153, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3471},
        {31095}
    },
    {
        8307,
        {31160, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8308,
        {31176, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8309,
        {31178, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3586},
        {24180}
    },
    {
        8310,
        {153457, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8311,
        {31188, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {6372},
        {31923}
    },
    {
        8312,
        {31198, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8313,
        {31211, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8314,
        {31213, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8315,
        {31235, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8316,
        {64084, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2337},
        {31296}
    },
    {
        8317,
        {31289, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8318,
        {31325, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8319,
        {31341, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8320,
        {64085, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3522},
        {31361}
    },
    {
        8321,
        {31365, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8322,
        {31392, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8323,
        {154052, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1777},
        {31427}
    },
    {
        8324,
        {31411, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8325,
        {31419, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {6260},
        {31383}
    },
    {
        8326,
        {31438, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8327,
        {31467, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4128},
        {38742}
    },
    {
        8328,
        {31485, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8329,
        {31506, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8330,
        {31533, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8331,
        {31547, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8332,
        {31559, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8333,
        {31566, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8334,
        {31584, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8335,
        {31597, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2470},
        {31639}
    },
    {
        8336,
        {31599, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3669},
        {31672}
    },
    {
        8337,
        {31602, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8338,
        {31646, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8339,
        {64086, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2978},
        {31680}
    },
    {
        8340,
        {31703, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8341,
        {31705, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8342,
        {31745, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {6330},
        {31721}
    },
    {
        8343,
        {31793, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {6354},
        {31823}
    },
    {
        8344,
        {31774, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3224},
        {31658}
    },
    {
        8345,
        {31776, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8346,
        {31795, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8347,
        {31798, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8348,
        {16996, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {6353},
        {31828}
    },
    {
        8349,
        {155041, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8350,
        {31833, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8351,
        {31853, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {6330},
        {31721}
    },
    {
        8352,
        {31865, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8353,
        {31887, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8354,
        {31892, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8355,
        {31904, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8356,
        {31932, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8357,
        {31957, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8358,
        {31961, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8359,
        {31965, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {6380},
        {31938}
    },
    {
        8360,
        {32007, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8361,
        {32008, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8362,
        {32019, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8363,
        {32029, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8364,
        {32035, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8365,
        {32049, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8366,
        {32065, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8367,
        {32072, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8368,
        {32083, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8369,
        {32092, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8370,
        {32122, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8371,
        {32131, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8372,
        {32139, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2296},
        {32024}
    },
    {
        8373,
        {32160, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4277},
        {32209}
    },
    {
        8374,
        {32166, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8375,
        {32194, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8376,
        {32204, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8377,
        {32214, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2710},
        {32210}
    },
    {
        8378,
        {32227, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1582},
        {32257}
    },
    {
        8379,
        {64087, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4322},
        {32244}
    },
    {
        8380,
        {32296, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4005},
        {24140}
    },
    {
        8381,
        {32264, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8382,
        {32273, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8383,
        {32277, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8384,
        {64089, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3708},
        {32321}
    },
    {
        8385,
        {32327, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8386,
        {32338, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8387,
        {32353, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2642},
        {32333}
    },
    {
        8388,
        {32394, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {6397},
        {32075}
    },
    {
        8389,
        {32397, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4292},
        {32047}
    },
    {
        8390,
        {32583, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3209},
        {27197}
    },
    {
        8391,
        {64090, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2711},
        {32626}
    },
    {
        8392,
        {32657, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8393,
        {32663, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {6506},
        {32652}
    },
    {
        8394,
        {32703, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8395,
        {32718, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8396,
        {32731, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8397,
        {32735, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8398,
        {32748, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8399,
        {32750, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8400,
        {32762, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8401,
        {64091, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2589},
        {32773}
    },
    {
        8402,
        {32788, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8403,
        {32806, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8404,
        {32821, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8405,
        {32823, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8406,
        {32828, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8407,
        {32970, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8408,
        {32983, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8409,
        {32992, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8410,
        {33011, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8411,
        {33048, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8412,
        {33098, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8413,
        {33120, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8414,
        {33127, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8415,
        {33128, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8416,
        {33133, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {7798},
        {40822}
    },
    {
        8417,
        {33211, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8418,
        {33226, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8419,
        {33231, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8420,
        {33239, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8421,
        {64092, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2644},
        {33261}
    },
    {
        8422,
        {17491, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {6621},
        {33274}
    },
    {
        8423,
        {17499, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {6630},
        {33296}
    },
    {
        8424,
        {33376, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8425,
        {33396, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8426,
        {158463, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {6651},
        {33406}
    },
    {
        8427,
        {33422, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8428,
        {33441, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8429,
        {33443, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8430,
        {33444, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8431,
        {33449, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8432,
        {33454, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8433,
        {33463, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8434,
        {33470, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8435,
        {33471, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8436,
        {33478, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8437,
        {33493, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8438,
        {33533, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2215},
        {33776}
    },
    {
        8439,
        {33534, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8440,
        {33536, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8441,
        {33537, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8442,
        {33634, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8443,
        {33570, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8444,
        {33581, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8445,
        {33594, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8446,
        {33603, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8447,
        {33607, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8448,
        {33617, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8449,
        {33621, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8450,
        {33661, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8451,
        {33670, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8452,
        {33682, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8453,
        {33688, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8454,
        {33703, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8455,
        {33705, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8456,
        {33727, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8457,
        {33728, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8458,
        {33735, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2215},
        {33776}
    },
    {
        8459,
        {33743, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8460,
        {33745, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8461,
        {33761, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8462,
        {33770, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8463,
        {33793, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8464,
        {33798, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8465,
        {33802, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4208},
        {33713}
    },
    {
        8466,
        {64095, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3283},
        {33879}
    },
    {
        8467,
        {33864, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8468,
        {33887, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8469,
        {33904, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8470,
        {33907, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8471,
        {33925, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8472,
        {33950, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {6687},
        {33669}
    },
    {
        8473,
        {33967, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8474,
        {33972, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8475,
        {33978, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8476,
        {33984, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8477,
        {33986, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8478,
        {159296, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8479,
        {34098, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8480,
        {34078, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8481,
        {34083, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2778},
        {33931}
    },
    {
        8482,
        {34095, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8483,
        {34137, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8484,
        {34148, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8485,
        {64031, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {6616},
        {33224}
    },
    {
        8486,
        {34221, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8487,
        {34170, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8488,
        {34188, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8489,
        {34191, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8490,
        {34210, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8491,
        {34224, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2083},
        {34219}
    },
    {
        8492,
        {34251, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8493,
        {34254, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8494,
        {34285, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8495,
        {34322, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8496,
        {34303, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8497,
        {34308, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8498,
        {34309, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8499,
        {34320, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1785},
        {33841}
    },
    {
        8500,
        {159988, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8501,
        {34328, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8502,
        {34345, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8503,
        {34360, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8504,
        {34391, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1964},
        {34394}
    },
    {
        8505,
        {34395, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1964},
        {34394}
    },
    {
        8506,
        {63798, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4255},
        {34396}
    },
    {
        8507,
        {34402, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8508,
        {17821, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3273},
        {34411}
    },
    {
        8509,
        {34412, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8510,
        {34421, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2593},
        {34503}
    },
    {
        8511,
        {34456, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {6827},
        {34516}
    },
    {
        8512,
        {34488, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8513,
        {34554, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8514,
        {34556, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8515,
        {34557, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1704},
        {35997}
    },
    {
        8516,
        {34571, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {6868},
        {34690}
    },
    {
        8517,
        {34673, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1435},
        {34427}
    },
    {
        8518,
        {34695, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8519,
        {34696, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8520,
        {34732, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8521,
        {34733, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8522,
        {34741, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8523,
        {17898, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8524,
        {34774, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8525,
        {34796, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2983},
        {34633}
    },
    {
        8526,
        {34822, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8527,
        {34826, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8528,
        {34832, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8529,
        {34836, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8530,
        {34847, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4348},
        {34635}
    },
    {
        8531,
        {34968, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8532,
        {34986, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8533,
        {35018, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8534,
        {35022, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8535,
        {161412, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8536,
        {35061, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {6935},
        {35060}
    },
    {
        8537,
        {35100, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8538,
        {64096, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1767},
        {35088}
    },
    {
        8539,
        {35096, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8540,
        {35097, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8541,
        {35098, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8542,
        {35111, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8543,
        {35120, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8544,
        {35122, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2212},
        {34996}
    },
    {
        8545,
        {35129, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8546,
        {35136, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8547,
        {35220, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {6967},
        {35219}
    },
    {
        8548,
        {64097, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2518},
        {35222}
    },
    {
        8549,
        {35284, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2031},
        {31563}
    },
    {
        8550,
        {35301, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8551,
        {35318, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8552,
        {35346, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8553,
        {35349, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8554,
        {35362, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8555,
        {35383, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2832},
        {20280}
    },
    {
        8556,
        {35399, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8557,
        {35406, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8558,
        {35421, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8559,
        {35425, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8560,
        {35445, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8561,
        {35449, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8562,
        {35495, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8563,
        {35536, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {5189},
        {24838}
    },
    {
        8564,
        {35551, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8565,
        {35572, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8566,
        {35574, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8567,
        {64034, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2715},
        {35576}
    },
    {
        8568,
        {64098, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1561},
        {35585}
    },
    {
        8569,
        {64099, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2037},
        {35641}
    },
    {
        8570,
        {35654, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8571,
        {35668, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8572,
        {35673, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {7010},
        {35482}
    },
    {
        8573,
        {35689, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8574,
        {35741, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8575,
        {35913, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8576,
        {35944, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8577,
        {64100, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3807},
        {36051}
    },
    {
        8578,
        {36065, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3120},
        {32154}
    },
    {
        8579,
        {36084, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4209},
        {38972}
    },
    {
        8580,
        {36088, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8581,
        {36094, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8582,
        {64101, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3104},
        {36104}
    },
    {
        8583,
        {36114, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2174},
        {36066}
    },
    {
        8584,
        {36123, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8585,
        {36271, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8586,
        {36302, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8587,
        {36305, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8588,
        {36311, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8589,
        {36384, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8590,
        {36387, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8591,
        {36413, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8592,
        {36464, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8593,
        {36475, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8594,
        {164471, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8595,
        {36544, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2050},
        {36527}
    },
    {
        8596,
        {18500, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1895},
        {36556}
    },
    {
        8597,
        {36602, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8598,
        {36638, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8599,
        {36653, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3557},
        {36575}
    },
    {
        8600,
        {36662, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8601,
        {36692, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8602,
        {164813, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8603,
        {36774, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8604,
        {36789, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8605,
        {36836, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8606,
        {36840, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8607,
        {36846, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8608,
        {36872, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {7217},
        {36837}
    },
    {
        8609,
        {36909, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8610,
        {64103, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1488},
        {36920}
    },
    {
        8611,
        {37000, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8612,
        {37013, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8613,
        {37015, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8614,
        {37017, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8615,
        {37019, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8616,
        {37026, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8617,
        {37043, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8618,
        {37054, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8619,
        {37060, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {7265},
        {37092}
    },
    {
        8620,
        {37061, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8621,
        {37063, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8622,
        {37079, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8623,
        {37085, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8624,
        {37086, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4349},
        {37070}
    },
    {
        8625,
        {37103, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8626,
        {37108, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8627,
        {64038, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3435},
        {37117}
    },
    {
        8628,
        {37140, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8629,
        {37141, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2004},
        {37111}
    },
    {
        8630,
        {37142, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8631,
        {37154, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8632,
        {37155, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8633,
        {37159, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8634,
        {37167, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8635,
        {37169, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8636,
        {37172, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8637,
        {37181, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {5080},
        {24283}
    },
    {
        8638,
        {37192, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8639,
        {37211, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8640,
        {37251, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8641,
        {37278, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8642,
        {37292, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2788},
        {37284}
    },
    {
        8643,
        {37297, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3681},
        {37271}
    },
    {
        8644,
        {37308, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {7056},
        {35724}
    },
    {
        8645,
        {37335, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8646,
        {37371, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8647,
        {37348, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8648,
        {37349, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8649,
        {37357, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8650,
        {37361, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8651,
        {37383, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8652,
        {37392, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8653,
        {37432, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8654,
        {37433, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8655,
        {37434, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8656,
        {37436, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8657,
        {37440, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8658,
        {37443, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8659,
        {37455, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2722},
        {37604}
    },
    {
        8660,
        {37496, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8661,
        {37512, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8662,
        {37570, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8663,
        {37579, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8664,
        {37580, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8665,
        {37587, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8666,
        {37600, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8667,
        {37631, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8668,
        {37636, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4354},
        {37682}
    },
    {
        8669,
        {37663, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {7322},
        {37531}
    },
    {
        8670,
        {37665, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8671,
        {37669, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8672,
        {37704, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8673,
        {37705, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8674,
        {37706, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4326},
        {37676}
    },
    {
        8675,
        {37732, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8676,
        {37733, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8677,
        {37738, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8678,
        {37744, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8679,
        {37787, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8680,
        {37795, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8681,
        {37818, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8682,
        {37830, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8683,
        {37854, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8684,
        {37855, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8685,
        {37892, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8686,
        {37885, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {7325},
        {37561}
    },
    {
        8687,
        {37939, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8688,
        {37962, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8689,
        {37987, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8690,
        {37995, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8691,
        {38001, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8692,
        {38002, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8693,
        {38286, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8694,
        {38303, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8695,
        {38310, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8696,
        {38313, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8697,
        {38316, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8698,
        {38326, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8699,
        {38333, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8700,
        {38347, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8701,
        {38352, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8702,
        {38355, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8703,
        {18864, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {7626},
        {39726}
    },
    {
        8704,
        {38362, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1517},
        {31418}
    },
    {
        8705,
        {38366, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8706,
        {38488, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8707,
        {38532, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3362},
        {22564}
    },
    {
        8708,
        {63964, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4249},
        {38534}
    },
    {
        8709,
        {38557, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8710,
        {38564, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8711,
        {38565, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8712,
        {38610, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8713,
        {38622, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {7724},
        {40388}
    },
    {
        8714,
        {64104, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3558},
        {38627}
    },
    {
        8715,
        {38633, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8716,
        {38639, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8717,
        {38707, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8718,
        {38715, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8719,
        {38733, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3354},
        {40372}
    },
    {
        8720,
        {38734, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3354},
        {40372}
    },
    {
        8721,
        {38735, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3354},
        {40372}
    },
    {
        8722,
        {38746, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8723,
        {38766, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8724,
        {38771, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8725,
        {38805, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8726,
        {38830, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8727,
        {38842, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8728,
        {38849, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8729,
        {38857, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8730,
        {38878, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8731,
        {38875, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8732,
        {38900, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8733,
        {64105, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2006},
        {38911}
    },
    {
        8734,
        {38922, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8735,
        {38942, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8736,
        {38955, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8737,
        {38960, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3990},
        {38956}
    },
    {
        8738,
        {64106, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3808},
        {38971}
    },
    {
        8739,
        {38994, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8740,
        {38995, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8741,
        {38998, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8742,
        {38999, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8743,
        {39001, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8744,
        {39002, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3414},
        {39003}
    },
    {
        8745,
        {63952, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4293},
        {39006}
    },
    {
        8746,
        {39013, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8747,
        {39020, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8748,
        {39098, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8749,
        {39112, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {7529},
        {39110}
    },
    {
        8750,
        {39143, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8751,
        {39256, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8752,
        {39326, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8753,
        {39426, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8754,
        {39427, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8755,
        {39460, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8756,
        {39469, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {7438},
        {38578}
    },
    {
        8757,
        {39470, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8758,
        {39480, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8759,
        {39498, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8760,
        {39502, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8761,
        {39506, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3146},
        {39464}
    },
    {
        8762,
        {39606, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8763,
        {39617, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8764,
        {39619, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8765,
        {39630, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8766,
        {39638, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8767,
        {39673, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8768,
        {39682, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {7619},
        {39714}
    },
    {
        8769,
        {39688, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8770,
        {39712, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8771,
        {19479, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {7626},
        {39726}
    },
    {
        8772,
        {39725, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {7625},
        {39722}
    },
    {
        8773,
        {39774, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8774,
        {39801, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8775,
        {39782, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8776,
        {39794, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {7686},
        {40056}
    },
    {
        8777,
        {39797, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8778,
        {39812, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8779,
        {39818, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8780,
        {39823, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8781,
        {39838, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8782,
        {39847, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8783,
        {39873, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8784,
        {39886, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8785,
        {39909, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8786,
        {39928, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8787,
        {39933, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8788,
        {39936, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8789,
        {39971, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8790,
        {40001, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8791,
        {40015, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8792,
        {40016, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8793,
        {40019, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8794,
        {40035, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8795,
        {40037, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8796,
        {40055, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4367},
        {39952}
    },
    {
        8797,
        {40221, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8798,
        {40222, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8799,
        {40259, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8800,
        {40263, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8801,
        {40274, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8802,
        {40291, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8803,
        {40304, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {7451},
        {38613}
    },
    {
        8804,
        {40316, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8805,
        {40330, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8806,
        {40342, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8807,
        {40384, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8808,
        {40364, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8809,
        {40380, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8810,
        {40407, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1607},
        {40206}
    },
    {
        8811,
        {172432, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8812,
        {40423, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8813,
        {40455, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8814,
        {40469, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8815,
        {40572, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2181},
        {40568}
    },
    {
        8816,
        {40606, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8817,
        {40612, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {7745},
        {40577}
    },
    {
        8818,
        {40620, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {7754},
        {40617}
    },
    {
        8819,
        {40623, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2332},
        {40633}
    },
    {
        8820,
        {40628, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2332},
        {40633}
    },
    {
        8821,
        {40629, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4086},
        {40634}
    },
    {
        8822,
        {40643, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {1608},
        {40644}
    },
    {
        8823,
        {40657, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2340},
        {40658}
    },
    {
        8824,
        {40720, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8825,
        {40761, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8826,
        {40791, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8827,
        {40848, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8828,
        {40852, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8829,
        {40855, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        0,
        {0},
        {0}
    },
    {
        8830,
        {40866, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4357},
        {21644}
    },
    {
        8831,
        {23032, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4898},
        {22925}
    },
    {
        8832,
        {23643, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {4983},
        {23631}
    },
    {
        8833,
        {24183, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {5066},
        {24182}
    },
    {
        8834,
        {30246, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {3080},
        {30185}
    },
    {
        8835,
        {32363, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_3,
        1,
        {2117},
        {32331}
    },
    {
        8836,
        {131209, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8837,
        {19970, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8838,
        {19983, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8839,
        {19986, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1518},
        {19985}
    },
    {
        8840,
        {20009, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8841,
        {20011, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8842,
        {20014, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8843,
        {20032, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8844,
        {20039, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8845,
        {20040, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5067},
        {24186}
    },
    {
        8846,
        {131234, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3612},
        {27966}
    },
    {
        8847,
        {20049, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8848,
        {13318, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8849,
        {131236, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3148},
        {22534}
    },
    {
        8850,
        {20058, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1506},
        {38560}
    },
    {
        8851,
        {20073, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8852,
        {20125, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2951},
        {25993}
    },
    {
        8853,
        {13356, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5413},
        {26066}
    },
    {
        8854,
        {13358, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6948},
        {35140}
    },
    {
        8855,
        {20153, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8856,
        {20155, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8857,
        {131490, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8858,
        {20156, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8859,
        {20163, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7273},
        {37194}
    },
    {
        8860,
        {20168, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8861,
        {20176, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2460},
        {20632}
    },
    {
        8862,
        {20203, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8863,
        {20186, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8864,
        {20209, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8865,
        {20213, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8866,
        {20224, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8867,
        {20246, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5051},
        {24081}
    },
    {
        8868,
        {20324, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8869,
        {20279, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6569},
        {32964}
    },
    {
        8870,
        {20286, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8871,
        {20308, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8872,
        {20312, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8873,
        {131603, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4496},
        {20486}
    },
    {
        8874,
        {20343, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5119},
        {24456}
    },
    {
        8875,
        {20344, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8876,
        {20346, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8877,
        {20349, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8878,
        {20354, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8879,
        {20357, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8880,
        {20370, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8881,
        {20378, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5121},
        {24455}
    },
    {
        8882,
        {20454, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4529},
        {20756}
    },
    {
        8883,
        {20402, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8884,
        {20414, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8885,
        {20421, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8886,
        {20427, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8887,
        {20431, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8888,
        {20434, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8889,
        {13418, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4528},
        {20757}
    },
    {
        8890,
        {20466, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4508},
        {20634}
    },
    {
        8891,
        {20480, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8892,
        {20496, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5960},
        {20495}
    },
    {
        8893,
        {20499, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8894,
        {20508, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8895,
        {20510, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8896,
        {20514, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8897,
        {13416, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4030},
        {20451}
    },
    {
        8898,
        {20546, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3023},
        {21069}
    },
    {
        8899,
        {20550, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8900,
        {20558, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8901,
        {20563, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8902,
        {20567, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8903,
        {20579, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8904,
        {20582, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6566},
        {32997}
    },
    {
        8905,
        {20586, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8906,
        {20592, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8907,
        {20643, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8908,
        {20616, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8909,
        {20626, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8910,
        {20627, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8911,
        {20629, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8912,
        {20630, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8913,
        {20636, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5127},
        {24493}
    },
    {
        8914,
        {20650, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8915,
        {131883, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7154},
        {36437}
    },
    {
        8916,
        {20657, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8917,
        {20666, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8918,
        {20667, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8919,
        {20676, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8920,
        {20679, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5267},
        {25134}
    },
    {
        8921,
        {20723, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8922,
        {131969, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8923,
        {20686, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8924,
        {131953, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2655},
        {38598}
    },
    {
        8925,
        {20692, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8926,
        {20697, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8927,
        {20705, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5179},
        {24800}
    },
    {
        8928,
        {20713, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        8929,
        {13458, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3992},
        {20693}
    },
    {
        8930,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8931,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8932,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8933,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8934,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8935,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8936,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8937,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8938,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8939,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8940,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8941,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8942,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8943,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8944,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8945,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8946,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8947,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8948,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8949,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8950,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8951,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8952,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8953,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8954,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8955,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8956,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8957,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8958,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8959,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8960,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8961,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8962,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8963,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8964,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8965,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8966,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8967,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8968,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8969,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8970,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8971,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8972,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8973,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8974,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8975,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8976,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8977,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8978,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8979,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8980,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8981,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8982,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8983,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8984,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8985,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8986,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8987,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8988,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8989,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8990,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8991,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8992,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8993,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8994,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8995,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8996,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8997,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8998,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        8999,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9000,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9001,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9002,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9003,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9004,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9005,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9006,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9007,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9008,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9009,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9010,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9011,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9012,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9013,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9014,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9015,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9016,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9017,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9018,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9019,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9020,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9021,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9022,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9023,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9024,
        {20744, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9025,
        {132089, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9026,
        {20759, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9027,
        {20763, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3840},
        {33310}
    },
    {
        9028,
        {132170, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4534},
        {20796}
    },
    {
        9029,
        {20832, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1776},
        {20828}
    },
    {
        9030,
        {132361, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9031,
        {20851, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2773},
        {31505}
    },
    {
        9032,
        {20867, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9033,
        {20875, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4547},
        {20866}
    },
    {
        9034,
        {13500, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9035,
        {20888, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9036,
        {20899, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2388},
        {26368}
    },
    {
        9037,
        {20909, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3133},
        {22826}
    },
    {
        9038,
        {13511, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9039,
        {20924, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9040,
        {132566, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9041,
        {132648, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9042,
        {20979, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9043,
        {20980, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4574},
        {20973}
    },
    {
        9044,
        {20994, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9045,
        {21010, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9046,
        {21014, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9047,
        {132943, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2152},
        {21091}
    },
    {
        9048,
        {21077, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9049,
        {21084, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9050,
        {21100, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2922},
        {21046}
    },
    {
        9051,
        {21111, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9052,
        {21124, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6317},
        {31642}
    },
    {
        9053,
        {21122, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9054,
        {133127, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9055,
        {21144, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9056,
        {133178, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9057,
        {21156, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9058,
        {21158, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9059,
        {21167, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9060,
        {21178, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9061,
        {21179, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9062,
        {21194, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9063,
        {13599, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2923},
        {21218}
    },
    {
        9064,
        {21201, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3317},
        {21189}
    },
    {
        9065,
        {133305, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9066,
        {21239, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9067,
        {21258, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9068,
        {21259, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9069,
        {21284, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9070,
        {21301, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9071,
        {21310, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9072,
        {21314, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9073,
        {133500, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9074,
        {133533, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9075,
        {21351, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1670},
        {33253}
    },
    {
        9076,
        {21356, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9077,
        {21370, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9078,
        {21412, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4310},
        {26278}
    },
    {
        9079,
        {21428, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9080,
        {133843, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9081,
        {21431, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2302},
        {32945}
    },
    {
        9082,
        {21440, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9083,
        {133917, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9084,
        {13661, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2529},
        {20107}
    },
    {
        9085,
        {13662, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9086,
        {21461, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9087,
        {21466, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9088,
        {13667, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9089,
        {21492, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9090,
        {21493, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9091,
        {21589, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4749},
        {22036}
    },
    {
        9092,
        {21540, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9093,
        {21544, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9094,
        {13678, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4713},
        {21754}
    },
    {
        9095,
        {21571, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9096,
        {21602, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9097,
        {21606, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9098,
        {21612, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9099,
        {21642, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4357},
        {21644}
    },
    {
        9100,
        {21645, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9101,
        {21653, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9102,
        {21664, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9103,
        {21670, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9104,
        {21677, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9105,
        {21678, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9106,
        {21687, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9107,
        {21690, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9108,
        {21695, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9109,
        {21699, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9110,
        {134469, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1659},
        {22057}
    },
    {
        9111,
        {21740, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6992},
        {35382}
    },
    {
        9112,
        {21743, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9113,
        {21745, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9114,
        {21747, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9115,
        {21760, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4157},
        {35480}
    },
    {
        9116,
        {21761, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9117,
        {21769, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9118,
        {21820, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9119,
        {21825, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9120,
        {13734, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9121,
        {21831, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9122,
        {21834, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9123,
        {13736, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9124,
        {21856, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3398},
        {21746}
    },
    {
        9125,
        {21857, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9126,
        {21860, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9127,
        {134625, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9128,
        {21885, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9129,
        {21890, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9130,
        {21896, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9131,
        {21905, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9132,
        {13765, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9133,
        {21970, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9134,
        {134805, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9135,
        {134765, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9136,
        {21951, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9137,
        {21961, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9138,
        {21964, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9139,
        {21969, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9140,
        {21981, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9141,
        {13786, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9142,
        {21986, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9143,
        {134756, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9144,
        {21993, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9145,
        {22056, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4765},
        {22063}
    },
    {
        9146,
        {135007, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9147,
        {22023, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9148,
        {22032, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9149,
        {22064, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9150,
        {22071, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9151,
        {13812, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9152,
        {22077, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9153,
        {22079, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9154,
        {22080, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9155,
        {22087, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9156,
        {22110, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9157,
        {22112, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9158,
        {22125, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9159,
        {13829, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7319},
        {37532}
    },
    {
        9160,
        {22152, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9161,
        {22156, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9162,
        {22165, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9163,
        {22170, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9164,
        {22173, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9165,
        {22184, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9166,
        {22189, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9167,
        {22194, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9168,
        {22213, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9169,
        {22221, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9170,
        {22239, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9171,
        {22248, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9172,
        {22262, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4790},
        {22272}
    },
    {
        9173,
        {22263, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9174,
        {135681, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9175,
        {22293, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9176,
        {22307, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9177,
        {135765, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9178,
        {22313, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9179,
        {135803, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4828},
        {22499}
    },
    {
        9180,
        {22341, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9181,
        {22342, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9182,
        {22348, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9183,
        {22349, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9184,
        {135796, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9185,
        {22376, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9186,
        {22383, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4808},
        {22351}
    },
    {
        9187,
        {22387, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9188,
        {22388, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9189,
        {22389, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1933},
        {19992}
    },
    {
        9190,
        {22395, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9191,
        {135908, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4853},
        {22743}
    },
    {
        9192,
        {135895, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9193,
        {22444, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9194,
        {22426, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7802},
        {22575}
    },
    {
        9195,
        {22429, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9196,
        {22430, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9197,
        {22440, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2610},
        {23432}
    },
    {
        9198,
        {22487, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9199,
        {135933, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9200,
        {22476, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9201,
        {135990, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9202,
        {136004, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9203,
        {22494, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3362},
        {22564}
    },
    {
        9204,
        {22502, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4371},
        {26880}
    },
    {
        9205,
        {22512, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9206,
        {13898, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9207,
        {22520, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9208,
        {22523, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9209,
        {22525, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3068},
        {25475}
    },
    {
        9210,
        {22532, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9211,
        {22558, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9212,
        {22560, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9213,
        {22567, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9214,
        {22578, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9215,
        {22585, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9216,
        {136132, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9217,
        {22601, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9218,
        {22604, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9219,
        {22631, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9220,
        {22666, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9221,
        {22667, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9222,
        {22669, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9223,
        {22671, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9224,
        {22672, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9225,
        {22676, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9226,
        {22685, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6206},
        {30973}
    },
    {
        9227,
        {22698, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9228,
        {22705, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9229,
        {136301, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9230,
        {22723, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6052},
        {30086}
    },
    {
        9231,
        {22733, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4840},
        {22649}
    },
    {
        9232,
        {22754, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9233,
        {22771, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5694},
        {27580}
    },
    {
        9234,
        {22772, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9235,
        {22789, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2318},
        {38477}
    },
    {
        9236,
        {22790, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9237,
        {22795, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9238,
        {22797, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3851},
        {24489}
    },
    {
        9239,
        {22804, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9240,
        {22820, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9241,
        {136663, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9242,
        {13969, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4307},
        {38666}
    },
    {
        9243,
        {22845, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9244,
        {13977, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9245,
        {22854, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9246,
        {13974, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3185},
        {27810}
    },
    {
        9247,
        {158761, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9248,
        {22875, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9249,
        {22879, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9250,
        {136775, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6485},
        {32570}
    },
    {
        9251,
        {22901, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9252,
        {22902, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9253,
        {22908, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9254,
        {22943, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9255,
        {22958, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9256,
        {22972, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9257,
        {22984, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9258,
        {22989, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9259,
        {23006, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9260,
        {23011, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9261,
        {23012, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9262,
        {23015, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1800},
        {23014}
    },
    {
        9263,
        {23022, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4922},
        {23270}
    },
    {
        9264,
        {136966, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9265,
        {137026, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4890},
        {22941}
    },
    {
        9266,
        {14031, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9267,
        {23052, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9268,
        {23053, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9269,
        {23063, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9270,
        {23079, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9271,
        {23085, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1867},
        {23305}
    },
    {
        9272,
        {23125, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9273,
        {23141, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9274,
        {23162, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9275,
        {23179, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9276,
        {23196, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9277,
        {23199, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9278,
        {23200, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9279,
        {23202, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9280,
        {23217, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9281,
        {23219, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9282,
        {23221, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9283,
        {23226, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9284,
        {23231, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5204},
        {24871}
    },
    {
        9285,
        {23258, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9286,
        {23260, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9287,
        {23264, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9288,
        {23269, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9289,
        {23280, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4923},
        {23273}
    },
    {
        9290,
        {23278, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9291,
        {23285, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9292,
        {23296, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9293,
        {23304, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9294,
        {23319, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9295,
        {23348, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9296,
        {23341, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9297,
        {23372, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9298,
        {23378, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9299,
        {23400, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9300,
        {23407, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9301,
        {23420, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9302,
        {23423, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9303,
        {23425, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9304,
        {23428, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9305,
        {137667, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1510},
        {23431}
    },
    {
        9306,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9307,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9308,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9309,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9310,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9311,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9312,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9313,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9314,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9315,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9316,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9317,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9318,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9319,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9320,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9321,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9322,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9323,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9324,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9325,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9326,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9327,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9328,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9329,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9330,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9331,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9332,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9333,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9334,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9335,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9336,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9337,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9338,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9339,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9340,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9341,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9342,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9343,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9344,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9345,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9346,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9347,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9348,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9349,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9350,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9351,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9352,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9353,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9354,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9355,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9356,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9357,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9358,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9359,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9360,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9361,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9362,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9363,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9364,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9365,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9366,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9367,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9368,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9369,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9370,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9371,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9372,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9373,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9374,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9375,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9376,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9377,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9378,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9379,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9380,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9381,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9382,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9383,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9384,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9385,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9386,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9387,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9388,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9389,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9390,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9391,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9392,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9393,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9394,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9395,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9396,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9397,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9398,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9399,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9400,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9401,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9402,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9403,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9404,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9405,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9406,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9407,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9408,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9409,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9410,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9411,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9412,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9413,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9414,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9415,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9416,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9417,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9418,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9419,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9420,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9421,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9422,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9423,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9424,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9425,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9426,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9427,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9428,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9429,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9430,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9431,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9432,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9433,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9434,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9435,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9436,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9437,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9438,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9439,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9440,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9441,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9442,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9443,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9444,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9445,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9446,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9447,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9448,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9449,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9450,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9451,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9452,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9453,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9454,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9455,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9456,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9457,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9458,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9459,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9460,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9461,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9462,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9463,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9464,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9465,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9466,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9467,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9468,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9469,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9470,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9471,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9472,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9473,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9474,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9475,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9476,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9477,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9478,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9479,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9480,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9481,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9482,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9483,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9484,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9485,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9486,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9487,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9488,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9489,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9490,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9491,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9492,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9493,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9494,
        {23446, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9495,
        {23468, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9496,
        {14177, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2388},
        {26368}
    },
    {
        9497,
        {23488, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9498,
        {14178, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1510},
        {23431}
    },
    {
        9499,
        {23502, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9500,
        {23510, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2846},
        {28024}
    },
    {
        9501,
        {14188, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9502,
        {14187, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6270},
        {31423}
    },
    {
        9503,
        {23537, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9504,
        {23549, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9505,
        {14197, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1714},
        {30861}
    },
    {
        9506,
        {23555, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9507,
        {23593, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9508,
        {23600, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9509,
        {138326, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9510,
        {23647, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9511,
        {23651, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9512,
        {23655, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9513,
        {23656, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9514,
        {23657, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9515,
        {23664, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9516,
        {138541, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9517,
        {138565, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9518,
        {23676, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9519,
        {138616, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9520,
        {138594, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9521,
        {23688, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7063},
        {35898}
    },
    {
        9522,
        {23690, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9523,
        {14273, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1848},
        {23736}
    },
    {
        9524,
        {138657, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9525,
        {138652, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1848},
        {23736}
    },
    {
        9526,
        {23712, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9527,
        {23714, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9528,
        {23718, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9529,
        {23719, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9530,
        {138642, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9531,
        {23725, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9532,
        {23733, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9533,
        {138679, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9534,
        {23753, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9535,
        {138720, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9536,
        {138803, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5031},
        {23970}
    },
    {
        9537,
        {23814, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9538,
        {23824, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5015},
        {23825}
    },
    {
        9539,
        {23851, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9540,
        {23837, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9541,
        {23840, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9542,
        {23844, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9543,
        {23846, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9544,
        {23857, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9545,
        {23865, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9546,
        {23874, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9547,
        {14312, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9548,
        {23905, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9549,
        {23914, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9550,
        {14324, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5025},
        {23916}
    },
    {
        9551,
        {23920, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9552,
        {139038, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9553,
        {14333, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9554,
        {23944, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9555,
        {14336, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5007},
        {23948}
    },
    {
        9556,
        {23954, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9557,
        {23956, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9558,
        {23959, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9559,
        {23961, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9560,
        {23984, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9561,
        {23986, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9562,
        {23988, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4885},
        {22887}
    },
    {
        9563,
        {139126, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9564,
        {23993, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7428},
        {38502}
    },
    {
        9565,
        {24017, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9566,
        {24023, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5043},
        {24022}
    },
    {
        9567,
        {24024, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9568,
        {24032, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9569,
        {139258, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6589},
        {33126}
    },
    {
        9570,
        {24036, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9571,
        {24041, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9572,
        {14383, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1451},
        {20197}
    },
    {
        9573,
        {24064, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3064},
        {21277}
    },
    {
        9574,
        {14390, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3698},
        {24070}
    },
    {
        9575,
        {24082, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3163},
        {34955}
    },
    {
        9576,
        {24085, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9577,
        {14400, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9578,
        {24095, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9579,
        {24110, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5064},
        {24135}
    },
    {
        9580,
        {24126, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9581,
        {24137, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9582,
        {14428, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9583,
        {24150, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9584,
        {14433, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9585,
        {24171, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5064},
        {24135}
    },
    {
        9586,
        {24172, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9587,
        {24173, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9588,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9589,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9590,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9591,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9592,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9593,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9594,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9595,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9596,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9597,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9598,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9599,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9600,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9601,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9602,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9603,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9604,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9605,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9606,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9607,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9608,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9609,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9610,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9611,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9612,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9613,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9614,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9615,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9616,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9617,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9618,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9619,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9620,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9621,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9622,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9623,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9624,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9625,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9626,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9627,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9628,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9629,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9630,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9631,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9632,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9633,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9634,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9635,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9636,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9637,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9638,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9639,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9640,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9641,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9642,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9643,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9644,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9645,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9646,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9647,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9648,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9649,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9650,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9651,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9652,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9653,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9654,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9655,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9656,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9657,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9658,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9659,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9660,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9661,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9662,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9663,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9664,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9665,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9666,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9667,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9668,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9669,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9670,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9671,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9672,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9673,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9674,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9675,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9676,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9677,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9678,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9679,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9680,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9681,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9682,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9683,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9684,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9685,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9686,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9687,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9688,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9689,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9690,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9691,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9692,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9693,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9694,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9695,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9696,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9697,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9698,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9699,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9700,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9701,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9702,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9703,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9704,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9705,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9706,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9707,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9708,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9709,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9710,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9711,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9712,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9713,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9714,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9715,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9716,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9717,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9718,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9719,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9720,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9721,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9722,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9723,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9724,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9725,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9726,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9727,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9728,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9729,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9730,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9731,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9732,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9733,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9734,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9735,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9736,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9737,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9738,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9739,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9740,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9741,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9742,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9743,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9744,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9745,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9746,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9747,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9748,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9749,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9750,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9751,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9752,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9753,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9754,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9755,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9756,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9757,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9758,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9759,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9760,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9761,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9762,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9763,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9764,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9765,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9766,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9767,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9768,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9769,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9770,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9771,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9772,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9773,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9774,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9775,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9776,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9777,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9778,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9779,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9780,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9781,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9782,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9783,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9784,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9785,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9786,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9787,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9788,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9789,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9790,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9791,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9792,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9793,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9794,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9795,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9796,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9797,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9798,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9799,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9800,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9801,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9802,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9803,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9804,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9805,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9806,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9807,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9808,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9809,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9810,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9811,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9812,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9813,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9814,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9815,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9816,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9817,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9818,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9819,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9820,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9821,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9822,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9823,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9824,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9825,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9826,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9827,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9828,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9829,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9830,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9831,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9832,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9833,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9834,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9835,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9836,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9837,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9838,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9839,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9840,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9841,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9842,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9843,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9844,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9845,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9846,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9847,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9848,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9849,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9850,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9851,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9852,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9853,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9854,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9855,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9856,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9857,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9858,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9859,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9860,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9861,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9862,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9863,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9864,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9865,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9866,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9867,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9868,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9869,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        9870,
        {24174, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9871,
        {139643, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9872,
        {24229, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9873,
        {24234, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9874,
        {24236, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4650},
        {21398}
    },
    {
        9875,
        {24249, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9876,
        {24255, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3791},
        {24287}
    },
    {
        9877,
        {24262, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9878,
        {24274, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9879,
        {24281, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9880,
        {140062, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1683},
        {24315}
    },
    {
        9881,
        {24317, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1683},
        {24315}
    },
    {
        9882,
        {24328, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9883,
        {24334, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2459},
        {19977}
    },
    {
        9884,
        {24348, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9885,
        {140205, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9886,
        {24350, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9887,
        {24391, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9888,
        {24419, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9889,
        {24434, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9890,
        {24446, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9891,
        {24463, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7427},
        {38495}
    },
    {
        9892,
        {24482, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9893,
        {24484, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2149},
        {20581}
    },
    {
        9894,
        {24504, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9895,
        {24516, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2839},
        {24515}
    },
    {
        9896,
        {14586, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2839},
        {24515}
    },
    {
        9897,
        {24519, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9898,
        {24523, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9899,
        {24530, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9900,
        {24531, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9901,
        {24532, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9902,
        {24546, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2232},
        {24735}
    },
    {
        9903,
        {24558, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9904,
        {24559, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9905,
        {24563, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9906,
        {24572, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5214},
        {24951}
    },
    {
        9907,
        {14615, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3158},
        {27888}
    },
    {
        9908,
        {24599, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9909,
        {24610, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9910,
        {24612, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9911,
        {14618, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9912,
        {24652, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9913,
        {24703, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4141},
        {21191}
    },
    {
        9914,
        {24714, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3398},
        {21746}
    },
    {
        9915,
        {24725, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9916,
        {24744, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9917,
        {141043, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3510},
        {24499}
    },
    {
        9918,
        {24752, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9919,
        {24753, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9920,
        {24766, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9921,
        {24776, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9922,
        {24793, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9923,
        {24795, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9924,
        {24814, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5232},
        {24986}
    },
    {
        9925,
        {24818, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9926,
        {24821, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9927,
        {24848, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9928,
        {24850, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2105},
        {25001}
    },
    {
        9929,
        {24851, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9930,
        {24857, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5152},
        {24682}
    },
    {
        9931,
        {24862, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5248},
        {25062}
    },
    {
        9932,
        {24890, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9933,
        {14703, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9934,
        {24897, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9935,
        {24902, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9936,
        {24928, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4511},
        {20658}
    },
    {
        9937,
        {24956, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9938,
        {141403, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5249},
        {25059}
    },
    {
        9939,
        {24978, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9940,
        {24979, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9941,
        {24983, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5228},
        {24982}
    },
    {
        9942,
        {24984, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1863},
        {21916}
    },
    {
        9943,
        {24997, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4614},
        {21214}
    },
    {
        9944,
        {25000, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9945,
        {25005, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9946,
        {141483, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9947,
        {25045, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9948,
        {25053, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9949,
        {25055, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9950,
        {25077, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9951,
        {141711, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9952,
        {25109, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9953,
        {25123, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9954,
        {25129, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9955,
        {25158, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9956,
        {25164, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2611},
        {25163}
    },
    {
        9957,
        {25169, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9958,
        {25170, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9959,
        {25185, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9960,
        {25188, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9961,
        {25211, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9962,
        {25197, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9963,
        {25203, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9964,
        {25241, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9965,
        {25254, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3886},
        {26564}
    },
    {
        9966,
        {25301, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9967,
        {142008, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9968,
        {25341, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1543},
        {26355}
    },
    {
        9969,
        {25347, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9970,
        {25357, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2282},
        {26657}
    },
    {
        9971,
        {25360, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5291},
        {25295}
    },
    {
        9972,
        {142159, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5371},
        {25827}
    },
    {
        9973,
        {142160, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5356},
        {25736}
    },
    {
        9974,
        {25394, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9975,
        {25397, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4337},
        {24324}
    },
    {
        9976,
        {25403, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9977,
        {25404, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9978,
        {25409, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9979,
        {25412, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9980,
        {25422, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9981,
        {142150, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9982,
        {25433, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9983,
        {142365, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9984,
        {142246, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9985,
        {25452, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9986,
        {25476, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9987,
        {25497, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9988,
        {142372, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5317},
        {25486}
    },
    {
        9989,
        {25492, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9990,
        {25533, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9991,
        {25591, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5334},
        {25554}
    },
    {
        9992,
        {25556, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3084},
        {32207}
    },
    {
        9993,
        {25557, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9994,
        {25564, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9995,
        {25568, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9996,
        {25579, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9997,
        {25580, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9998,
        {25586, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        9999,
        {25609, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10000,
        {25630, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10001,
        {25637, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10002,
        {25641, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10003,
        {25647, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10004,
        {25690, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10005,
        {25691, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10006,
        {25693, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10007,
        {25715, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10008,
        {25725, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10009,
        {25735, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10010,
        {25745, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10011,
        {25757, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10012,
        {25759, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10013,
        {25803, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10014,
        {25804, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10015,
        {25813, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2107},
        {25658}
    },
    {
        10016,
        {25815, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10017,
        {142817, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10018,
        {25828, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10019,
        {25829, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5373},
        {25900}
    },
    {
        10020,
        {25855, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5376},
        {25842}
    },
    {
        10021,
        {25860, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10022,
        {14958, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10023,
        {25871, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10024,
        {25876, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10025,
        {25878, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10026,
        {14963, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10027,
        {25886, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10028,
        {25906, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10029,
        {25924, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10030,
        {25940, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10031,
        {25963, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10032,
        {25978, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7579},
        {39493}
    },
    {
        10033,
        {25985, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10034,
        {25988, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10035,
        {25989, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4947},
        {23416}
    },
    {
        10036,
        {25994, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2951},
        {25993}
    },
    {
        10037,
        {26034, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10038,
        {26037, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10039,
        {26040, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10040,
        {26047, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5799},
        {28216}
    },
    {
        10041,
        {26050, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10042,
        {26057, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3822},
        {25975}
    },
    {
        10043,
        {26068, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10044,
        {15062, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2504},
        {26088}
    },
    {
        10045,
        {26098, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10046,
        {26105, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2537},
        {26178}
    },
    {
        10047,
        {26108, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10048,
        {26116, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10049,
        {26120, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10050,
        {26145, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10051,
        {26154, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10052,
        {26181, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10053,
        {26193, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10054,
        {26190, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10055,
        {15082, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6625},
        {33282}
    },
    {
        10056,
        {143811, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6625},
        {33282}
    },
    {
        10057,
        {26199, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10058,
        {26203, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10059,
        {26211, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5438},
        {26210}
    },
    {
        10060,
        {143861, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10061,
        {143798, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3718},
        {26217}
    },
    {
        10062,
        {26218, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10063,
        {26219, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10064,
        {26220, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10065,
        {26221, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10066,
        {26235, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10067,
        {26240, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1599},
        {26106}
    },
    {
        10068,
        {26256, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10069,
        {26258, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2931},
        {26228}
    },
    {
        10070,
        {26265, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10071,
        {15118, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7518},
        {39023}
    },
    {
        10072,
        {26285, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10073,
        {26289, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5424},
        {26165}
    },
    {
        10074,
        {26293, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10075,
        {15130, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10076,
        {26303, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5911},
        {29113}
    },
    {
        10077,
        {15132, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2843},
        {26187}
    },
    {
        10078,
        {26348, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2458},
        {26194}
    },
    {
        10079,
        {15063, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2253},
        {21402}
    },
    {
        10080,
        {26369, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10081,
        {26373, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10082,
        {26386, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10083,
        {144242, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10084,
        {26393, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4073},
        {26126}
    },
    {
        10085,
        {144339, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1943},
        {26429}
    },
    {
        10086,
        {144338, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10087,
        {26444, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10088,
        {26445, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4226},
        {26446}
    },
    {
        10089,
        {26452, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10090,
        {26461, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10091,
        {144336, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2598},
        {26451}
    },
    {
        10092,
        {144356, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10093,
        {144341, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10094,
        {26484, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10095,
        {26486, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10096,
        {144346, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10097,
        {26514, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10098,
        {144351, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10099,
        {33635, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1544},
        {26628}
    },
    {
        10100,
        {26640, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10101,
        {26544, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10102,
        {26546, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10103,
        {26563, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10104,
        {26568, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10105,
        {26578, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2568},
        {28422}
    },
    {
        10106,
        {26585, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10107,
        {26587, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10108,
        {26608, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3541},
        {22856}
    },
    {
        10109,
        {26615, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10110,
        {144458, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10111,
        {144465, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10112,
        {144459, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10113,
        {26648, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10114,
        {26655, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10115,
        {26669, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10116,
        {144485, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5513},
        {26674}
    },
    {
        10117,
        {26675, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10118,
        {26683, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10119,
        {26686, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5658},
        {27410}
    },
    {
        10120,
        {26692, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10121,
        {26693, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5524},
        {26772}
    },
    {
        10122,
        {26697, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1447},
        {26696}
    },
    {
        10123,
        {26700, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3179},
        {21331}
    },
    {
        10124,
        {26709, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10125,
        {26711, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10126,
        {15223, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10127,
        {26731, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10128,
        {26734, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3624},
        {26479}
    },
    {
        10129,
        {26746, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4129},
        {26611}
    },
    {
        10130,
        {26748, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2568},
        {28422}
    },
    {
        10131,
        {26754, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10132,
        {26768, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10133,
        {26774, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10134,
        {15213, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10135,
        {26776, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10136,
        {26777, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10137,
        {26778, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10138,
        {26780, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10139,
        {26794, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10140,
        {26795, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10141,
        {26804, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10142,
        {26811, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10143,
        {26875, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10144,
        {144612, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5631},
        {27208}
    },
    {
        10145,
        {144730, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10146,
        {64019, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5558},
        {26920}
    },
    {
        10147,
        {26819, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4227},
        {26792}
    },
    {
        10148,
        {26821, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3886},
        {26564}
    },
    {
        10149,
        {26828, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10150,
        {26831, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10151,
        {26838, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10152,
        {26841, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10153,
        {26852, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10154,
        {26853, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5617},
        {27146}
    },
    {
        10155,
        {26860, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10156,
        {26871, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10157,
        {26883, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10158,
        {26887, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10159,
        {15239, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10160,
        {15240, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3541},
        {22856}
    },
    {
        10161,
        {144788, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10162,
        {26939, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1568},
        {22576}
    },
    {
        10163,
        {15245, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6318},
        {31627}
    },
    {
        10164,
        {26950, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10165,
        {26985, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10166,
        {26988, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10167,
        {26994, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10168,
        {27002, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5333},
        {25545}
    },
    {
        10169,
        {27007, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10170,
        {27026, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10171,
        {15268, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2755},
        {26494}
    },
    {
        10172,
        {27030, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10173,
        {27032, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2048},
        {30697}
    },
    {
        10174,
        {27046, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1805},
        {24185}
    },
    {
        10175,
        {27056, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10176,
        {27063, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10177,
        {27066, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10178,
        {27068, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10179,
        {27072, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5586},
        {27073}
    },
    {
        10180,
        {27089, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3635},
        {26995}
    },
    {
        10181,
        {27094, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10182,
        {144953, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10183,
        {144967, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10184,
        {27184, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10185,
        {144952, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10186,
        {144954, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1995},
        {27211}
    },
    {
        10187,
        {27107, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10188,
        {27118, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10189,
        {27119, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10190,
        {27123, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10191,
        {15309, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10192,
        {27124, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10193,
        {27134, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7355},
        {37853}
    },
    {
        10194,
        {27153, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4263},
        {26753}
    },
    {
        10195,
        {27162, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10196,
        {27165, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10197,
        {145180, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5480},
        {26468}
    },
    {
        10198,
        {27186, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10199,
        {27187, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10200,
        {27188, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10201,
        {27199, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10202,
        {27206, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10203,
        {27209, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10204,
        {27258, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10205,
        {27214, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10206,
        {27218, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10207,
        {27236, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10208,
        {145164, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10209,
        {27262, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10210,
        {27267, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10211,
        {27275, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10212,
        {15344, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2821},
        {26893}
    },
    {
        10213,
        {27281, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10214,
        {27295, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1564},
        {27022}
    },
    {
        10215,
        {27297, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10216,
        {145252, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10217,
        {27307, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10218,
        {27325, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10219,
        {27334, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10220,
        {27348, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3519},
        {26627}
    },
    {
        10221,
        {27344, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10222,
        {27356, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10223,
        {27357, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10224,
        {145407, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10225,
        {145383, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10226,
        {27372, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10227,
        {27377, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5656},
        {34326}
    },
    {
        10228,
        {27378, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10229,
        {27379, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5602},
        {27102}
    },
    {
        10230,
        {27389, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10231,
        {145444, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5661},
        {27423}
    },
    {
        10232,
        {27403, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10233,
        {27407, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10234,
        {27408, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10235,
        {27409, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10236,
        {145469, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10237,
        {27415, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4218},
        {27396}
    },
    {
        10238,
        {15398, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10239,
        {27439, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10240,
        {27466, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10241,
        {27480, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10242,
        {27500, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3023},
        {21069}
    },
    {
        10243,
        {27509, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        10244,
        {27514, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5677},
        {27513}
    },
    {
        10245,
        {27521, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5678},
        {27519}
    },
    {
        10246,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10247,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10248,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10249,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10250,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10251,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10252,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10253,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10254,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10255,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10256,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10257,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10258,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10259,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10260,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10261,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10262,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10263,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10264,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10265,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10266,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10267,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10268,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10269,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10270,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10271,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10272,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10273,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10274,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10275,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10276,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10277,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10278,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10279,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10280,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10281,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10282,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10283,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10284,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10285,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10286,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10287,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10288,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10289,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10290,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10291,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10292,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10293,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10294,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10295,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10296,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10297,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10298,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10299,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10300,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10301,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10302,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10303,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10304,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10305,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10306,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10307,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10308,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10309,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10310,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10311,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10312,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10313,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10314,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10315,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10316,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10317,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10318,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10319,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10320,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10321,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10322,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10323,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10324,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10325,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10326,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10327,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10328,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10329,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10330,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10331,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10332,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10333,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10334,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10335,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10336,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10337,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10338,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10339,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10340,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10341,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10342,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10343,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10344,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10345,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10346,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10347,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10348,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10349,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10350,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10351,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10352,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10353,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10354,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10355,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10356,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10357,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10358,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10359,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10360,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10361,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10362,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10363,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10364,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10365,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10366,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10367,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10368,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10369,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10370,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10371,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10372,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10373,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10374,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10375,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10376,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10377,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10378,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10379,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10380,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10381,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10382,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10383,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10384,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10385,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10386,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10387,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10388,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10389,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10390,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10391,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10392,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10393,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10394,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10395,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10396,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10397,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10398,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10399,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10400,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10401,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10402,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10403,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10404,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10405,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10406,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10407,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10408,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10409,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10410,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10411,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10412,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10413,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10414,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10415,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10416,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10417,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10418,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10419,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10420,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10421,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10422,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10423,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10424,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10425,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10426,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10427,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10428,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10429,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10430,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10431,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10432,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10433,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10434,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10435,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10436,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10437,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10438,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10439,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10440,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10441,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10442,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10443,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10444,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10445,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10446,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10447,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10448,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10449,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10450,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10451,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10452,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10453,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10454,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10455,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10456,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10457,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10458,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10459,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10460,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10461,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10462,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10463,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10464,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10465,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10466,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10467,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10468,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10469,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10470,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10471,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10472,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10473,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10474,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10475,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10476,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10477,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10478,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10479,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10480,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10481,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10482,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10483,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10484,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10485,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10486,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10487,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10488,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10489,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10490,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10491,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10492,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10493,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10494,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10495,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10496,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10497,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10498,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10499,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10500,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10501,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10502,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10503,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10504,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10505,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10506,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10507,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10508,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10509,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10510,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10511,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10512,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10513,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10514,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10515,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10516,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10517,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10518,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10519,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10520,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10521,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10522,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10523,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10524,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10525,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10526,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10527,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10528,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10529,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10530,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10531,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10532,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10533,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10534,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10535,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10536,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10537,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10538,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10539,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10540,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10541,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10542,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10543,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10544,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10545,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10546,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10547,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10548,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10549,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10550,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10551,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10552,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10553,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10554,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10555,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10556,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10557,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10558,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10559,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10560,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10561,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10562,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10563,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10564,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10565,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10566,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10567,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10568,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10569,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10570,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10571,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10572,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10573,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10574,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10575,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10576,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10577,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10578,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10579,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10580,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10581,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10582,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10583,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10584,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10585,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10586,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10587,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10588,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10589,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10590,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10591,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10592,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10593,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10594,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10595,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10596,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10597,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10598,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10599,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10600,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10601,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10602,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10603,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10604,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10605,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10606,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10607,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10608,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10609,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10610,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10611,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10612,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10613,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10614,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10615,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10616,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10617,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10618,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10619,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10620,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10621,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10622,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10623,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10624,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10625,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10626,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10627,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10628,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10629,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10630,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10631,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10632,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10633,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10634,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10635,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10636,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10637,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10638,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10639,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10640,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10641,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10642,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10643,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10644,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10645,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10646,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10647,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10648,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10649,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10650,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10651,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10652,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10653,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10654,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10655,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10656,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10657,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10658,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10659,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10660,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10661,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10662,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10663,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10664,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10665,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10666,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10667,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10668,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10669,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10670,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10671,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10672,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10673,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10674,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10675,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10676,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10677,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10678,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10679,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10680,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10681,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10682,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10683,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10684,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10685,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10686,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10687,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10688,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10689,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10690,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10691,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10692,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10693,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10694,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10695,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10696,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10697,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10698,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10699,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10700,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10701,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10702,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10703,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10704,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10705,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10706,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10707,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10708,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10709,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10710,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10711,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10712,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10713,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10714,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10715,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10716,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10717,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10718,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10719,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10720,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10721,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10722,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10723,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10724,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10725,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10726,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10727,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10728,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10729,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10730,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10731,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10732,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10733,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10734,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10735,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10736,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10737,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10738,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10739,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10740,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10741,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10742,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10743,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10744,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10745,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10746,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10747,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10748,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10749,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10750,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10751,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10752,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10753,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10754,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10755,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10756,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10757,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10758,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10759,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10760,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10761,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10762,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10763,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10764,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10765,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10766,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10767,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10768,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10769,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10770,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10771,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10772,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10773,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10774,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10775,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10776,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10777,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10778,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10779,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10780,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10781,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10782,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10783,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10784,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10785,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10786,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10787,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10788,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10789,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10790,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10791,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10792,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10793,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10794,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10795,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10796,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10797,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10798,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10799,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10800,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10801,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10802,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10803,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10804,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10805,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10806,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10807,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10808,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10809,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10810,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10811,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10812,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10813,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10814,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10815,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10816,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10817,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10818,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10819,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10820,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10821,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10822,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10823,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10824,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10825,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10826,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10827,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10828,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10829,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10830,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10831,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10832,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10833,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10834,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10835,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10836,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10837,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10838,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10839,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10840,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10841,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10842,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10843,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10844,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10845,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10846,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10847,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10848,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10849,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10850,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10851,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10852,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10853,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10854,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10855,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10856,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10857,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10858,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10859,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10860,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10861,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10862,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10863,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10864,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10865,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10866,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10867,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10868,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10869,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10870,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10871,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10872,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10873,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10874,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10875,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10876,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10877,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10878,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10879,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10880,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10881,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10882,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10883,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10884,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10885,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10886,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10887,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10888,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10889,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10890,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10891,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10892,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10893,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10894,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10895,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10896,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10897,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10898,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10899,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10900,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10901,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10902,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10903,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10904,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10905,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10906,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10907,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10908,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10909,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10910,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10911,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10912,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10913,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10914,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10915,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10916,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10917,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10918,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10919,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10920,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10921,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10922,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10923,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10924,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10925,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10926,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10927,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10928,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10929,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10930,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10931,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10932,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10933,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10934,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10935,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10936,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10937,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10938,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10939,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10940,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10941,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10942,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10943,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10944,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10945,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10946,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10947,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10948,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10949,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10950,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10951,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10952,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10953,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10954,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10955,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10956,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10957,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10958,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10959,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10960,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10961,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10962,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10963,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10964,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10965,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10966,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10967,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10968,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10969,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10970,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10971,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10972,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10973,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10974,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10975,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10976,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10977,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10978,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10979,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10980,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10981,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10982,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10983,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10984,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10985,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10986,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10987,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10988,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10989,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10990,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10991,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10992,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10993,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10994,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10995,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10996,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10997,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10998,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        10999,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11000,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11001,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11002,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11003,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11004,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11005,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11006,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11007,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11008,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11009,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11010,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11011,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11012,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11013,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11014,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11015,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11016,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11017,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11018,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11019,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11020,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11021,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11022,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11023,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11024,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11025,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11026,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11027,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11028,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11029,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11030,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11031,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11032,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11033,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11034,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11035,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11036,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11037,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11038,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11039,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11040,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11041,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11042,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11043,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11044,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11045,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11046,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11047,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11048,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11049,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11050,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11051,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11052,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11053,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11054,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11055,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11056,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11057,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11058,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11059,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11060,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11061,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11062,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11063,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11064,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11065,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11066,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11067,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11068,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11069,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11070,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11071,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11072,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11073,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11074,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11075,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11076,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11077,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11078,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11079,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11080,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11081,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11082,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11083,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11084,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11085,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11086,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11087,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11088,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11089,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11090,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11091,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11092,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11093,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11094,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11095,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11096,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11097,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11098,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11099,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11100,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11101,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11102,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11103,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11104,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11105,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11106,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11107,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11108,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11109,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11110,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11111,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11112,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11113,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11114,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11115,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11116,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11117,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11118,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11119,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11120,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11121,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11122,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11123,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11124,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11125,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11126,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11127,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11128,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11129,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11130,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11131,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11132,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11133,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11134,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11135,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11136,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11137,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11138,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11139,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11140,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11141,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11142,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11143,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11144,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11145,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11146,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11147,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11148,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11149,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11150,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11151,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11152,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11153,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11154,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11155,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11156,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11157,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11158,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11159,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11160,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11161,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11162,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11163,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11164,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11165,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11166,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11167,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11168,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11169,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11170,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11171,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11172,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11173,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11174,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11175,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11176,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11177,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11178,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11179,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11180,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11181,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11182,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11183,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11184,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11185,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11186,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11187,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11188,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11189,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11190,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11191,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11192,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11193,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11194,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11195,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11196,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11197,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11198,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11199,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11200,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11201,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11202,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11203,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11204,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11205,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11206,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11207,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11208,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11209,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11210,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11211,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11212,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11213,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11214,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11215,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11216,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11217,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11218,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11219,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11220,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11221,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11222,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11223,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11224,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11225,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11226,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11227,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11228,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11229,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11230,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11231,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11232,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11233,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11234,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11235,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11236,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11237,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11238,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11239,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11240,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11241,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11242,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11243,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11244,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11245,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11246,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11247,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11248,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11249,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11250,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11251,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11252,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11253,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11254,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11255,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11256,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11257,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11258,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11259,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11260,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11261,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11262,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11263,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11264,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11265,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11266,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11267,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11268,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11269,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11270,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11271,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11272,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11273,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11274,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11275,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11276,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11277,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11278,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11279,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11280,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11281,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11282,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11283,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11284,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11285,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11286,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11287,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11288,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11289,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11290,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11291,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11292,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11293,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11294,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11295,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11296,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11297,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11298,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11299,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11300,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11301,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11302,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11303,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11304,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11305,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11306,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11307,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11308,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11309,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11310,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11311,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11312,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11313,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11314,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11315,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11316,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11317,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11318,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11319,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11320,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11321,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11322,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11323,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11324,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11325,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11326,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11327,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11328,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11329,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11330,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11331,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11332,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11333,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11334,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11335,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11336,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11337,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11338,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11339,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11340,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11341,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11342,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11343,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11344,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11345,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11346,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11347,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11348,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11349,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11350,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11351,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11352,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11353,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11354,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11355,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11356,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11357,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11358,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11359,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11360,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11361,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11362,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11363,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11364,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11365,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11366,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11367,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11368,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11369,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11370,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11371,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11372,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11373,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11374,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11375,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11376,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11377,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11378,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11379,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11380,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11381,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11382,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11383,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11384,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11385,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11386,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11387,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11388,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11389,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11390,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11391,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11392,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11393,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11394,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11395,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11396,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11397,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11398,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11399,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11400,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11401,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11402,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11403,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11404,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11405,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11406,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11407,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11408,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11409,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11410,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11411,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11412,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11413,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11414,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11415,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11416,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11417,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11418,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11419,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11420,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11421,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11422,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11423,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11424,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11425,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11426,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11427,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11428,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11429,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11430,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11431,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11432,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11433,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11434,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11435,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11436,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11437,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11438,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11439,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11440,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11441,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11442,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11443,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11444,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11445,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11446,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11447,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11448,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11449,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11450,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11451,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11452,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11453,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11454,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11455,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11456,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11457,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11458,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11459,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11460,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11461,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11462,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11463,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11464,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11465,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11466,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11467,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11468,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11469,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11470,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11471,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11472,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11473,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11474,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11475,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11476,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11477,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11478,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11479,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11480,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11481,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11482,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11483,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11484,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11485,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11486,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11487,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11488,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11489,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11490,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11491,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11492,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11493,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11494,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11495,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11496,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11497,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11498,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11499,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11500,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11501,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11502,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11503,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11504,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11505,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11506,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11507,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11508,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11509,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11510,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11511,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11512,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11513,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11514,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11515,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11516,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11517,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11518,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11519,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11520,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11521,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11522,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11523,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11524,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11525,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11526,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11527,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11528,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11529,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11530,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11531,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11532,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11533,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11534,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11535,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11536,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11537,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11538,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11539,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11540,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11541,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11542,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11543,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11544,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11545,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11546,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11547,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11548,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11549,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11550,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11551,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11552,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11553,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11554,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11555,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11556,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11557,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11558,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11559,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11560,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11561,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11562,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11563,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11564,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11565,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11566,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11567,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11568,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11569,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11570,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11571,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11572,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11573,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11574,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11575,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11576,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11577,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11578,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11579,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11580,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11581,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11582,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11583,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11584,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11585,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11586,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11587,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11588,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11589,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11590,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11591,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11592,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11593,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11594,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11595,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11596,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11597,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11598,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11599,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11600,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11601,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11602,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11603,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11604,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11605,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11606,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11607,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11608,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11609,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11610,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11611,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11612,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11613,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11614,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11615,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11616,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11617,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11618,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11619,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11620,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11621,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11622,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11623,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11624,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11625,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11626,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11627,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11628,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11629,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11630,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11631,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11632,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11633,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11634,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11635,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11636,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11637,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11638,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11639,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11640,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11641,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11642,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11643,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11644,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11645,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11646,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11647,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11648,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11649,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11650,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11651,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11652,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11653,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11654,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11655,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11656,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11657,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11658,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11659,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11660,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11661,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11662,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11663,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11664,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11665,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11666,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11667,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11668,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11669,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11670,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11671,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11672,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11673,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11674,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11675,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11676,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11677,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11678,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11679,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11680,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11681,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11682,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11683,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11684,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11685,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11686,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11687,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11688,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11689,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11690,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11691,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11692,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11693,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11694,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11695,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11696,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11697,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11698,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11699,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11700,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11701,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11702,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11703,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11704,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11705,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11706,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11707,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11708,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11709,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11710,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11711,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11712,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11713,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11714,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11715,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11716,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11717,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11718,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11719,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11720,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11721,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11722,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11723,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11724,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11725,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11726,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11727,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11728,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11729,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11730,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11731,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11732,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11733,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11734,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11735,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11736,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11737,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11738,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11739,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11740,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11741,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11742,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11743,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11744,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11745,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11746,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11747,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11748,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11749,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11750,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11751,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11752,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11753,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11754,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11755,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11756,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11757,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11758,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11759,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11760,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11761,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11762,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11763,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11764,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11765,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11766,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11767,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11768,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11769,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11770,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11771,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11772,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11773,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11774,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11775,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11776,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11777,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11778,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11779,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11780,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11781,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11782,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11783,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11784,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11785,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11786,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11787,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11788,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11789,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11790,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11791,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11792,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11793,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11794,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11795,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11796,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11797,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11798,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11799,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11800,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11801,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11802,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11803,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11804,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11805,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11806,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11807,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11808,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11809,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11810,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11811,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11812,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11813,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11814,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11815,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11816,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11817,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11818,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11819,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11820,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11821,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11822,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11823,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11824,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11825,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11826,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11827,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11828,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11829,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11830,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11831,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11832,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11833,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11834,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11835,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11836,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11837,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11838,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11839,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11840,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11841,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11842,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11843,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11844,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11845,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11846,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11847,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11848,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11849,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11850,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11851,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11852,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11853,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11854,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11855,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11856,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11857,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11858,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11859,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11860,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11861,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11862,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11863,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11864,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11865,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11866,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11867,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11868,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11869,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11870,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11871,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11872,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11873,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11874,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11875,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11876,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11877,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11878,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11879,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11880,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11881,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11882,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11883,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11884,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11885,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11886,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11887,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11888,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11889,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11890,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11891,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11892,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11893,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11894,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11895,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11896,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11897,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11898,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11899,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11900,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11901,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11902,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11903,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11904,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11905,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11906,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11907,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11908,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11909,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11910,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11911,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11912,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11913,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11914,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11915,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11916,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11917,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11918,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11919,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11920,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11921,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11922,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11923,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11924,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11925,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11926,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11927,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11928,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11929,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11930,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11931,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11932,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11933,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11934,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11935,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11936,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11937,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11938,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11939,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11940,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11941,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11942,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11943,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11944,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11945,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11946,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11947,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11948,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11949,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11950,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11951,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11952,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11953,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11954,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11955,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11956,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11957,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11958,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11959,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11960,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11961,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11962,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11963,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11964,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11965,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11966,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11967,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11968,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11969,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11970,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11971,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11972,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11973,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11974,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11975,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11976,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11977,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11978,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11979,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11980,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11981,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11982,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11983,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11984,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11985,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11986,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11987,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11988,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11989,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11990,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11991,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11992,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11993,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11994,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11995,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11996,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11997,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11998,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        11999,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12000,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12001,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12002,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12003,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12004,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12005,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12006,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12007,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12008,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12009,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12010,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12011,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12012,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12013,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12014,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12015,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12016,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12017,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12018,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12019,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12020,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12021,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12022,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12023,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12024,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12025,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12026,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12027,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12028,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12029,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12030,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12031,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12032,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12033,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12034,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12035,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12036,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12037,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12038,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12039,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12040,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12041,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12042,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12043,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12044,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12045,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12046,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12047,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12048,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12049,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12050,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12051,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12052,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12053,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12054,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12055,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12056,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12057,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12058,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12059,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12060,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12061,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12062,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12063,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12064,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12065,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12066,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12067,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12068,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12069,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12070,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12071,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12072,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12073,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12074,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12075,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12076,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12077,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12078,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12079,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12080,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12081,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12082,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12083,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12084,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12085,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12086,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12087,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12088,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12089,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12090,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12091,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12092,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12093,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12094,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12095,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12096,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12097,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12098,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12099,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12100,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12101,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12102,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12103,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12104,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12105,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12106,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12107,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12108,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12109,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12110,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12111,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12112,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12113,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12114,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12115,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12116,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12117,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12118,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12119,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12120,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12121,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12122,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12123,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12124,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12125,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12126,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12127,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12128,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12129,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12130,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12131,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12132,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12133,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12134,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12135,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12136,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12137,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12138,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12139,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12140,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12141,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12142,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12143,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12144,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12145,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12146,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12147,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12148,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12149,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12150,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12151,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12152,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12153,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12154,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12155,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12156,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12157,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12158,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12159,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12160,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12161,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12162,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12163,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12164,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12165,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12166,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12167,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12168,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12169,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12170,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12171,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12172,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12173,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12174,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12175,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12176,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12177,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12178,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12179,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12180,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12181,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12182,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12183,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12184,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12185,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12186,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12187,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12188,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12189,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12190,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12191,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12192,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12193,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12194,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12195,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12196,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12197,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12198,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12199,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12200,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12201,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12202,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12203,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12204,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12205,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12206,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12207,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12208,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12209,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12210,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12211,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12212,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12213,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12214,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12215,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12216,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12217,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12218,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12219,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12220,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12221,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12222,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12223,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12224,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12225,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12226,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12227,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12228,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12229,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12230,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12231,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12232,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12233,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12234,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12235,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12236,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12237,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12238,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12239,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12240,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12241,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12242,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12243,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12244,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12245,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12246,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12247,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12248,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12249,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12250,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12251,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12252,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12253,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12254,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12255,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12256,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12257,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12258,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12259,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12260,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12261,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12262,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12263,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12264,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12265,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12266,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12267,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12268,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12269,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12270,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12271,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12272,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12273,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12274,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12275,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12276,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12277,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12278,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12279,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12280,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12281,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12282,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12283,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12284,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12285,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12286,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12287,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12288,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12289,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12290,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12291,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12292,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12293,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12294,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12295,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12296,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12297,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12298,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12299,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12300,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12301,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12302,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12303,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12304,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12305,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12306,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12307,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12308,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12309,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12310,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12311,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12312,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12313,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12314,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12315,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12316,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12317,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12318,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12319,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12320,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12321,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12322,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12323,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12324,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12325,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12326,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12327,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12328,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12329,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12330,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12331,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12332,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12333,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12334,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12335,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12336,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12337,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12338,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12339,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12340,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12341,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12342,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12343,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12344,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12345,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12346,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12347,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12348,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12349,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12350,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12351,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12352,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12353,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12354,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12355,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12356,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12357,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12358,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12359,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12360,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12361,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12362,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12363,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12364,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12365,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12366,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12367,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12368,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12369,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12370,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12371,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12372,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12373,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12374,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12375,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12376,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12377,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12378,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12379,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12380,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12381,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12382,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12383,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12384,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12385,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12386,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12387,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12388,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12389,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12390,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12391,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12392,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12393,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12394,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12395,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12396,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12397,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12398,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12399,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12400,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12401,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12402,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12403,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12404,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12405,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12406,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12407,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12408,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12409,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12410,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12411,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12412,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12413,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12414,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12415,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12416,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12417,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12418,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12419,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12420,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12421,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12422,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12423,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12424,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12425,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12426,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12427,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12428,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12429,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12430,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12431,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12432,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12433,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12434,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12435,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12436,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12437,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12438,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12439,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12440,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12441,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12442,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12443,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12444,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12445,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12446,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12447,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12448,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12449,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12450,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12451,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12452,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12453,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12454,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12455,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12456,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12457,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12458,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12459,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12460,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12461,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12462,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12463,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12464,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12465,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12466,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12467,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12468,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12469,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12470,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12471,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12472,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12473,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12474,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12475,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12476,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12477,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12478,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12479,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12480,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12481,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12482,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12483,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12484,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12485,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12486,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12487,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12488,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12489,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12490,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12491,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12492,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12493,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12494,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12495,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12496,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12497,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12498,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12499,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12500,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12501,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12502,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12503,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12504,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12505,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12506,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12507,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12508,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12509,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12510,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12511,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12512,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12513,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12514,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12515,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12516,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12517,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12518,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12519,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12520,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12521,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12522,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12523,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12524,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12525,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12526,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12527,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12528,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12529,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12530,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12531,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12532,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12533,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12534,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12535,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12536,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12537,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12538,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12539,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12540,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12541,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12542,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12543,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12544,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12545,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12546,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12547,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12548,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12549,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12550,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12551,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12552,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12553,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12554,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12555,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12556,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12557,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12558,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12559,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12560,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12561,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12562,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12563,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12564,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12565,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12566,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12567,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12568,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12569,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12570,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12571,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12572,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12573,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12574,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12575,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12576,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12577,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12578,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12579,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12580,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12581,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12582,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12583,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12584,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12585,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12586,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12587,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12588,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12589,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12590,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12591,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12592,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12593,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12594,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12595,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12596,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12597,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12598,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12599,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12600,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12601,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12602,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12603,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12604,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12605,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12606,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12607,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12608,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12609,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12610,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12611,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12612,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12613,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12614,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12615,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12616,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12617,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12618,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12619,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12620,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12621,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12622,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12623,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12624,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12625,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12626,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12627,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12628,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12629,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12630,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12631,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12632,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12633,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12634,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12635,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12636,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12637,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12638,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12639,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12640,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12641,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12642,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12643,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12644,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12645,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12646,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12647,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12648,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12649,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12650,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12651,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12652,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12653,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12654,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12655,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12656,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12657,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12658,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12659,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12660,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12661,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12662,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12663,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12664,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12665,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12666,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12667,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12668,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12669,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12670,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12671,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12672,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12673,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12674,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12675,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12676,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12677,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12678,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12679,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12680,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12681,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12682,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12683,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12684,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12685,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12686,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12687,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12688,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12689,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12690,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12691,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12692,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12693,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12694,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12695,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12696,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12697,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12698,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12699,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12700,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12701,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12702,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12703,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12704,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12705,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12706,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12707,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12708,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12709,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12710,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12711,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12712,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12713,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12714,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12715,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12716,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12717,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12718,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12719,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12720,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12721,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12722,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12723,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12724,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12725,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12726,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12727,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12728,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12729,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12730,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12731,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12732,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12733,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12734,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12735,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12736,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12737,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12738,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12739,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12740,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12741,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12742,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12743,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12744,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12745,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12746,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12747,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12748,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12749,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12750,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12751,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12752,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12753,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12754,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12755,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12756,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12757,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12758,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12759,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12760,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12761,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12762,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12763,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12764,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12765,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12766,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12767,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12768,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12769,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12770,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12771,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12772,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12773,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12774,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12775,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12776,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12777,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12778,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12779,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12780,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12781,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12782,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12783,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12784,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12785,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12786,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12787,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12788,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12789,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12790,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12791,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12792,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12793,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12794,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12795,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12796,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12797,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12798,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12799,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12800,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12801,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12802,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12803,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12804,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12805,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12806,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12807,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12808,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12809,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12810,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12811,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12812,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12813,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12814,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12815,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12816,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12817,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12818,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12819,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12820,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12821,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12822,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12823,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12824,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12825,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12826,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12827,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12828,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12829,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12830,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12831,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12832,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12833,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12834,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12835,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12836,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12837,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12838,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12839,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12840,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12841,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12842,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12843,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12844,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12845,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12846,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12847,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12848,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12849,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12850,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12851,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12852,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12853,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12854,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12855,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12856,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12857,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12858,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12859,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12860,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12861,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12862,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12863,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12864,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12865,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12866,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12867,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12868,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12869,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12870,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12871,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12872,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12873,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12874,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12875,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12876,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12877,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12878,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12879,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12880,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12881,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12882,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12883,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12884,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12885,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12886,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12887,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12888,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12889,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12890,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12891,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12892,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12893,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12894,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12895,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12896,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12897,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12898,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12899,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12900,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12901,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12902,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12903,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12904,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12905,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12906,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12907,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12908,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12909,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12910,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12911,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12912,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12913,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12914,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12915,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12916,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12917,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12918,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12919,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12920,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12921,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12922,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12923,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12924,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12925,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12926,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12927,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12928,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12929,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12930,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12931,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12932,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12933,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12934,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12935,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12936,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12937,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12938,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12939,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12940,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12941,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12942,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12943,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12944,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12945,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12946,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12947,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12948,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12949,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12950,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12951,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12952,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12953,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12954,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12955,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12956,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12957,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12958,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12959,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12960,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12961,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12962,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12963,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12964,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12965,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12966,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12967,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12968,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12969,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12970,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12971,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12972,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12973,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12974,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12975,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12976,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12977,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12978,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12979,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12980,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12981,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12982,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12983,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12984,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12985,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12986,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12987,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12988,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12989,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12990,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12991,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12992,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12993,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12994,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12995,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12996,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12997,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12998,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        12999,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13000,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13001,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13002,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13003,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13004,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13005,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13006,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13007,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13008,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13009,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13010,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13011,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13012,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13013,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13014,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13015,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13016,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13017,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13018,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13019,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13020,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13021,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13022,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13023,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13024,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13025,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13026,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13027,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13028,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13029,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13030,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13031,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13032,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13033,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13034,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13035,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13036,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13037,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13038,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13039,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13040,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13041,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13042,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13043,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13044,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13045,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13046,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13047,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13048,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13049,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13050,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13051,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13052,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13053,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13054,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13055,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13056,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13057,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13058,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13059,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13060,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13061,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13062,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13063,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13064,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13065,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13066,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13067,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13068,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13069,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13070,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13071,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13072,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13073,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13074,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13075,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13076,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13077,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13078,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13079,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13080,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13081,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13082,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13083,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13084,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13085,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13086,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13087,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13088,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13089,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13090,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13091,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13092,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13093,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13094,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13095,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13096,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13097,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13098,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13099,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13100,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13101,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13102,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13103,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13104,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13105,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13106,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13107,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13108,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13109,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13110,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13111,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13112,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13113,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13114,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13115,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13116,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13117,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13118,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13119,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13120,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13121,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13122,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13123,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13124,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13125,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13126,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13127,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13128,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13129,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13130,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13131,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13132,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13133,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13134,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13135,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13136,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13137,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13138,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13139,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13140,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13141,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13142,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13143,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13144,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13145,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13146,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13147,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13148,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13149,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13150,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13151,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13152,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13153,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13154,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13155,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13156,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13157,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13158,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13159,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13160,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13161,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13162,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13163,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13164,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13165,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13166,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13167,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13168,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13169,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13170,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13171,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13172,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13173,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13174,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13175,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13176,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13177,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13178,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13179,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13180,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13181,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13182,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13183,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13184,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13185,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13186,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13187,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13188,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13189,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13190,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13191,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13192,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13193,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13194,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13195,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13196,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13197,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13198,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13199,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13200,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13201,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13202,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13203,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13204,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13205,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13206,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13207,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13208,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13209,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13210,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13211,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13212,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13213,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13214,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13215,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13216,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13217,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13218,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13219,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13220,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13221,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13222,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13223,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13224,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13225,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13226,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13227,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13228,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13229,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13230,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13231,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13232,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13233,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13234,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13235,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13236,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13237,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13238,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13239,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13240,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13241,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13242,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13243,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13244,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13245,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13246,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13247,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13248,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13249,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13250,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13251,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13252,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13253,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13254,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13255,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13256,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13257,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13258,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13259,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13260,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13261,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13262,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13263,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13264,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13265,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13266,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13267,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13268,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13269,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13270,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13271,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13272,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13273,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13274,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13275,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13276,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13277,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13278,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13279,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13280,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13281,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13282,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13283,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13284,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13285,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13286,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13287,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13288,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13289,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13290,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13291,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13292,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13293,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13294,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13295,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13296,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13297,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13298,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13299,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13300,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13301,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13302,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13303,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13304,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13305,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13306,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13307,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13308,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13309,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13310,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13311,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13312,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13313,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13314,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13315,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13316,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13317,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13318,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13319,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13320,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13321,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13322,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13323,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13324,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13325,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13326,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13327,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13328,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13329,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13330,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13331,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13332,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13333,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13334,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13335,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13336,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13337,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13338,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13339,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13340,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13341,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13342,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13343,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13344,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13345,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13346,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13347,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13348,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13349,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13350,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13351,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13352,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13353,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13354,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13355,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13356,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13357,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13358,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13359,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13360,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13361,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13362,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13363,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13364,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13365,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13366,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13367,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13368,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13369,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13370,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13371,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13372,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13373,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13374,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13375,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13376,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13377,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13378,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13379,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13380,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13381,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13382,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13383,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13384,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13385,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13386,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13387,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13388,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13389,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13390,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13391,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13392,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13393,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13394,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13395,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13396,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13397,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13398,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13399,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13400,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13401,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13402,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13403,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13404,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13405,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13406,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13407,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13408,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13409,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13410,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13411,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13412,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13413,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13414,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13415,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13416,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13417,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13418,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13419,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13420,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13421,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13422,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13423,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13424,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13425,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13426,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13427,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13428,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13429,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13430,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13431,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13432,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13433,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13434,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13435,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13436,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13437,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13438,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13439,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13440,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13441,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13442,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13443,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13444,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13445,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13446,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13447,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13448,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13449,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13450,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13451,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13452,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13453,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13454,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13455,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13456,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13457,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13458,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13459,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13460,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13461,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13462,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13463,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13464,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13465,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13466,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13467,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13468,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13469,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13470,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13471,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13472,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13473,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13474,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13475,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13476,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13477,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13478,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13479,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13480,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13481,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13482,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13483,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13484,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13485,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13486,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13487,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13488,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13489,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13490,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13491,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13492,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13493,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13494,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13495,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13496,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13497,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13498,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13499,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13500,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13501,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13502,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13503,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13504,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13505,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13506,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13507,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13508,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13509,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13510,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13511,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13512,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13513,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13514,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13515,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13516,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13517,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13518,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13519,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13520,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13521,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13522,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13523,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13524,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13525,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13526,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13527,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13528,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13529,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13530,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13531,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13532,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13533,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13534,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13535,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13536,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13537,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13538,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13539,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13540,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13541,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13542,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13543,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13544,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13545,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13546,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13547,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13548,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13549,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13550,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13551,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13552,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13553,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13554,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13555,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13556,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13557,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13558,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13559,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13560,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13561,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13562,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13563,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13564,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13565,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13566,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13567,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13568,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13569,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13570,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13571,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13572,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13573,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13574,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13575,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13576,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13577,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13578,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13579,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13580,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13581,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13582,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13583,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13584,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13585,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13586,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13587,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13588,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13589,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13590,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13591,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13592,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13593,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13594,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13595,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13596,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13597,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13598,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13599,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13600,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13601,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13602,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13603,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13604,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13605,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13606,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13607,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13608,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13609,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13610,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13611,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13612,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13613,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13614,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13615,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13616,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13617,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13618,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13619,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13620,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13621,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13622,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13623,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13624,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13625,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13626,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13627,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13628,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13629,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13630,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13631,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13632,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13633,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13634,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13635,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13636,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13637,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13638,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13639,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13640,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13641,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13642,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13643,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13644,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13645,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13646,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13647,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13648,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13649,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13650,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13651,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13652,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13653,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13654,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13655,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13656,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13657,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13658,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13659,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13660,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13661,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13662,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13663,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13664,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13665,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13666,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13667,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13668,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13669,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13670,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13671,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13672,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13673,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13674,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13675,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13676,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13677,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13678,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13679,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13680,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13681,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13682,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13683,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13684,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13685,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13686,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13687,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13688,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13689,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13690,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13691,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13692,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13693,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13694,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13695,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13696,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13697,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13698,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13699,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13700,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13701,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13702,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13703,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13704,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13705,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13706,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13707,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13708,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13709,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13710,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13711,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13712,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13713,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13714,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13715,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13716,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13717,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13718,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13719,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13720,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13721,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13722,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13723,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13724,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13725,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13726,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13727,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13728,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13729,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13730,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13731,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13732,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13733,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13734,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13735,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13736,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13737,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13738,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13739,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13740,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13741,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13742,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13743,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13744,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13745,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13746,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13747,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13748,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13749,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13750,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13751,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13752,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13753,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13754,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13755,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13756,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13757,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13758,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13759,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13760,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13761,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13762,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13763,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13764,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13765,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13766,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13767,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13768,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13769,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13770,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13771,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13772,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13773,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13774,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13775,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13776,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13777,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13778,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13779,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13780,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13781,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13782,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13783,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13784,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13785,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13786,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13787,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13788,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13789,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13790,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13791,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13792,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13793,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13794,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13795,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13796,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13797,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13798,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13799,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13800,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13801,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13802,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13803,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13804,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13805,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13806,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13807,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13808,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13809,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13810,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13811,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13812,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13813,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13814,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13815,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13816,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13817,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13818,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13819,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13820,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13821,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13822,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13823,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13824,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13825,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13826,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13827,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13828,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13829,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13830,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13831,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13832,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13833,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13834,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13835,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13836,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13837,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13838,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13839,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13840,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13841,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13842,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13843,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13844,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13845,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13846,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13847,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13848,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13849,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13850,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13851,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13852,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13853,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13854,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13855,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13856,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13857,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13858,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13859,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13860,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13861,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13862,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13863,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13864,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13865,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13866,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13867,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13868,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13869,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13870,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13871,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13872,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13873,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13874,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13875,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13876,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13877,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13878,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13879,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13880,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13881,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13882,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13883,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13884,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13885,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13886,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13887,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13888,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13889,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13890,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13891,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13892,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13893,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13894,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13895,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13896,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13897,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13898,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13899,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13900,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13901,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13902,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13903,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13904,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13905,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13906,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13907,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13908,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13909,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13910,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13911,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13912,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13913,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13914,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13915,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13916,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13917,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13918,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13919,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13920,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13921,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13922,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13923,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13924,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13925,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13926,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13927,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13928,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13929,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13930,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13931,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13932,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13933,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13934,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13935,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13936,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13937,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13938,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13939,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13940,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13941,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13942,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13943,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13944,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13945,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13946,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13947,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13948,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13949,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13950,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13951,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13952,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13953,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13954,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13955,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13956,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13957,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13958,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13959,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13960,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13961,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13962,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13963,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13964,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13965,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13966,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13967,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13968,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13969,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13970,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13971,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13972,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13973,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13974,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13975,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13976,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13977,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13978,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13979,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13980,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13981,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13982,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13983,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13984,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13985,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13986,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13987,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13988,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13989,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13990,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13991,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13992,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13993,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13994,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13995,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13996,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13997,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13998,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        13999,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14000,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14001,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14002,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14003,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14004,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14005,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14006,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14007,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14008,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14009,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14010,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14011,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14012,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14013,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14014,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14015,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14016,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14017,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14018,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14019,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14020,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14021,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14022,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14023,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14024,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14025,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14026,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14027,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14028,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14029,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14030,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14031,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14032,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14033,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14034,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14035,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14036,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14037,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14038,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14039,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14040,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14041,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14042,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14043,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14044,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14045,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14046,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14047,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14048,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14049,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14050,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14051,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14052,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14053,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14054,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14055,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14056,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14057,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14058,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14059,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14060,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14061,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14062,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14063,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14064,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14065,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14066,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14067,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14068,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14069,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14070,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14071,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14072,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14073,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14074,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14075,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14076,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14077,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14078,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14079,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14080,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14081,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14082,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14083,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14084,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14085,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14086,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14087,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14088,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14089,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14090,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14091,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14092,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14093,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14094,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14095,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14096,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14097,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14098,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14099,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14100,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14101,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14102,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14103,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14104,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14105,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14106,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14107,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14108,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14109,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14110,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14111,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14112,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14113,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14114,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14115,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14116,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14117,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14118,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14119,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14120,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14121,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14122,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14123,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14124,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14125,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14126,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14127,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14128,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14129,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14130,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14131,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14132,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14133,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14134,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14135,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14136,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14137,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14138,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14139,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14140,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14141,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14142,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14143,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14144,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14145,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14146,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14147,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14148,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14149,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14150,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14151,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14152,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14153,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14154,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14155,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14156,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14157,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14158,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14159,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14160,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14161,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14162,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14163,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14164,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14165,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14166,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14167,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14168,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14169,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14170,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14171,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14172,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14173,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14174,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14175,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14176,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14177,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14178,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14179,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14180,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14181,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14182,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14183,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14184,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14185,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14186,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14187,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14188,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14189,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14190,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14191,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14192,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14193,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14194,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14195,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14196,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14197,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14198,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14199,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14200,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14201,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14202,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14203,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14204,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14205,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14206,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14207,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14208,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14209,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14210,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14211,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14212,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14213,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14214,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14215,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14216,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14217,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14218,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14219,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14220,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14221,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14222,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14223,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14224,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14225,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14226,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14227,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14228,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14229,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14230,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14231,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14232,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14233,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14234,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14235,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14236,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14237,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14238,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14239,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14240,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14241,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14242,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14243,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14244,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14245,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14246,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14247,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14248,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14249,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14250,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14251,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14252,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14253,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14254,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14255,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14256,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14257,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14258,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14259,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14260,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14261,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14262,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14263,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14264,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14265,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14266,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14267,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14268,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14269,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14270,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14271,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14272,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14273,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14274,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14275,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14276,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14277,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14278,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14279,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14280,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14281,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14282,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14283,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14284,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14285,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14286,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14287,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14288,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14289,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14290,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14291,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14292,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14293,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14294,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14295,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14296,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14297,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14298,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14299,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14300,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14301,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14302,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14303,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14304,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14305,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14306,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14307,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14308,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14309,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14310,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14311,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14312,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14313,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14314,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14315,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14316,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14317,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14318,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14319,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14320,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14321,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14322,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14323,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14324,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14325,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14326,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14327,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14328,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14329,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14330,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14331,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14332,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14333,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14334,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14335,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14336,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14337,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14338,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14339,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14340,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14341,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14342,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14343,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14344,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14345,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14346,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14347,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14348,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14349,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14350,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14351,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14352,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14353,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14354,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14355,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14356,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14357,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14358,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14359,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14360,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14361,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14362,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14363,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14364,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14365,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14366,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14367,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14368,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14369,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14370,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14371,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14372,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14373,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14374,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14375,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14376,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14377,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14378,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14379,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14380,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14381,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14382,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14383,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14384,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14385,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14386,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14387,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14388,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14389,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14390,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14391,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14392,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14393,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14394,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14395,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14396,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14397,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14398,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14399,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14400,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14401,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14402,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14403,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14404,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14405,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14406,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14407,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14408,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14409,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14410,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14411,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14412,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14413,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14414,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14415,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14416,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14417,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14418,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14419,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14420,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14421,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14422,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14423,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14424,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14425,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14426,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14427,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14428,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14429,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14430,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14431,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14432,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14433,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14434,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14435,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14436,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14437,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14438,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14439,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14440,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14441,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14442,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14443,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14444,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14445,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14446,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14447,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14448,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14449,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14450,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14451,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14452,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14453,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14454,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14455,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14456,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14457,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14458,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14459,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14460,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14461,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14462,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14463,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14464,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14465,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14466,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14467,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14468,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14469,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14470,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14471,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14472,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14473,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14474,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14475,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14476,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14477,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14478,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14479,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14480,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14481,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14482,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14483,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14484,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14485,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14486,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14487,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14488,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14489,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14490,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14491,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14492,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14493,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14494,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14495,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14496,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14497,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14498,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14499,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14500,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14501,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14502,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14503,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14504,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14505,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14506,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14507,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14508,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14509,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14510,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14511,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14512,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14513,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14514,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14515,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14516,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14517,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14518,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14519,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14520,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14521,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14522,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14523,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14524,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14525,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14526,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14527,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14528,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14529,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14530,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14531,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14532,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14533,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14534,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14535,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14536,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14537,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14538,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14539,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14540,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14541,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14542,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14543,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14544,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14545,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14546,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14547,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14548,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14549,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14550,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14551,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14552,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14553,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14554,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14555,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14556,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14557,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14558,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14559,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14560,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14561,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14562,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14563,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14564,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14565,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14566,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14567,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14568,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14569,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14570,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14571,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14572,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14573,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14574,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14575,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14576,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14577,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14578,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14579,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14580,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14581,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14582,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14583,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14584,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14585,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14586,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14587,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14588,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14589,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14590,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14591,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14592,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14593,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14594,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14595,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14596,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14597,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14598,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14599,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14600,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14601,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14602,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14603,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14604,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14605,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14606,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14607,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14608,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14609,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14610,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14611,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14612,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14613,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14614,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14615,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14616,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14617,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14618,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14619,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14620,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14621,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14622,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14623,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14624,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14625,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14626,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14627,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14628,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14629,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14630,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14631,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14632,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14633,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14634,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14635,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14636,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14637,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14638,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14639,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14640,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14641,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14642,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14643,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14644,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14645,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14646,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14647,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14648,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14649,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14650,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14651,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14652,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14653,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14654,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14655,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14656,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14657,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14658,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14659,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14660,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14661,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14662,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14663,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14664,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14665,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14666,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14667,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14668,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14669,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14670,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14671,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14672,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14673,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14674,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14675,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14676,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14677,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14678,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14679,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14680,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14681,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14682,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14683,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14684,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14685,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14686,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14687,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14688,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14689,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14690,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14691,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14692,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14693,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14694,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14695,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14696,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14697,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14698,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14699,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14700,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14701,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14702,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14703,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14704,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14705,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14706,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14707,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14708,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14709,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14710,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14711,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14712,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14713,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14714,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14715,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14716,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14717,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14718,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14719,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14720,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14721,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14722,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14723,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14724,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14725,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14726,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14727,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14728,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14729,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14730,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14731,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14732,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14733,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14734,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14735,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14736,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14737,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14738,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14739,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14740,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14741,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14742,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14743,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14744,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14745,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14746,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14747,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14748,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14749,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14750,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14751,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14752,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14753,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14754,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14755,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14756,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14757,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14758,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14759,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14760,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14761,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14762,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14763,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14764,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14765,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14766,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14767,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14768,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14769,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14770,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14771,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14772,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14773,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14774,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14775,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14776,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14777,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14778,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14779,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14780,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14781,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14782,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14783,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14784,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14785,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14786,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14787,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14788,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14789,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14790,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14791,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14792,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14793,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14794,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14795,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14796,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14797,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14798,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14799,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14800,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14801,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14802,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14803,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14804,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14805,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14806,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14807,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14808,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14809,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14810,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14811,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14812,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14813,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14814,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14815,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14816,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14817,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14818,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14819,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14820,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14821,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14822,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14823,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14824,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14825,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14826,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14827,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14828,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14829,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14830,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14831,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14832,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14833,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14834,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14835,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14836,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14837,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14838,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14839,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14840,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14841,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14842,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14843,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14844,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14845,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14846,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14847,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14848,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14849,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14850,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14851,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14852,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14853,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14854,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14855,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14856,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14857,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14858,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14859,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14860,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14861,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14862,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14863,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14864,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14865,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14866,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14867,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14868,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14869,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14870,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14871,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14872,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14873,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14874,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14875,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14876,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14877,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14878,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14879,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14880,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14881,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14882,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14883,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14884,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14885,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14886,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14887,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14888,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14889,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14890,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14891,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14892,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14893,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14894,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14895,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14896,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14897,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14898,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14899,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14900,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14901,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14902,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14903,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14904,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14905,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14906,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14907,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14908,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14909,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14910,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14911,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14912,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14913,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14914,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14915,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14916,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14917,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14918,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14919,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14920,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14921,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14922,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14923,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14924,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14925,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14926,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14927,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14928,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14929,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14930,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14931,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14932,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14933,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14934,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14935,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14936,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14937,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14938,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14939,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14940,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14941,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14942,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14943,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14944,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14945,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14946,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14947,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14948,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14949,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14950,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14951,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14952,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14953,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14954,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14955,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14956,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14957,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14958,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14959,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14960,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14961,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14962,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14963,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14964,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14965,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14966,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14967,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14968,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14969,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14970,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14971,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14972,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14973,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14974,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14975,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14976,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14977,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14978,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14979,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14980,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14981,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14982,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14983,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14984,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14985,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14986,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14987,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14988,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14989,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14990,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14991,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14992,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14993,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14994,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14995,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14996,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14997,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14998,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        14999,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15000,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15001,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15002,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15003,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15004,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15005,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15006,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15007,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15008,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15009,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15010,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15011,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15012,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15013,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15014,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15015,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15016,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15017,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15018,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15019,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15020,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15021,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15022,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15023,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15024,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15025,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15026,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15027,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15028,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15029,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15030,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15031,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15032,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15033,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15034,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15035,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15036,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15037,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15038,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15039,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15040,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15041,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15042,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15043,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15044,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15045,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15046,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15047,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15048,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15049,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15050,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15051,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15052,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15053,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15054,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15055,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15056,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15057,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15058,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15059,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15060,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15061,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15062,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15063,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15064,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15065,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15066,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15067,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15068,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15069,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15070,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15071,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15072,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15073,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15074,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15075,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15076,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15077,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15078,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15079,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15080,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15081,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15082,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15083,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15084,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15085,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15086,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15087,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15088,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15089,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15090,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15091,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15092,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15093,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15094,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15095,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15096,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15097,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15098,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15099,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15100,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15101,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15102,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15103,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15104,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15105,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15106,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15107,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15108,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15109,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15110,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15111,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15112,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15113,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15114,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15115,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15116,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15117,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15118,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15119,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15120,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15121,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15122,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15123,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15124,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15125,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15126,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15127,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15128,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15129,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15130,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15131,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15132,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15133,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15134,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15135,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15136,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15137,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15138,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15139,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15140,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15141,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15142,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15143,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15144,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15145,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15146,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15147,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15148,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15149,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15150,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15151,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15152,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15153,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15154,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15155,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15156,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15157,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15158,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15159,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15160,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15161,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15162,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15163,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15164,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15165,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15166,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15167,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15168,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15169,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15170,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15171,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15172,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15173,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15174,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15175,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15176,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15177,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15178,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15179,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15180,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15181,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15182,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15183,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15184,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15185,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15186,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15187,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15188,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15189,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15190,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15191,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15192,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15193,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15194,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15195,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15196,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15197,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15198,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15199,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15200,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15201,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15202,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15203,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15204,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15205,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15206,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15207,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15208,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15209,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15210,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15211,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15212,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15213,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15214,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15215,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15216,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15217,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15218,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15219,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15220,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15221,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15222,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15223,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15224,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15225,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15226,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15227,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15228,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15229,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15230,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15231,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15232,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15233,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15234,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15235,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15236,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15237,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15238,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15239,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15240,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15241,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15242,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15243,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15244,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15245,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15246,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15247,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15248,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15249,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15250,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15251,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15252,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15253,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15254,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15255,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15256,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15257,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15258,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15259,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15260,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15261,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15262,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15263,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15264,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15265,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15266,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15267,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15268,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15269,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15270,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15271,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15272,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15273,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15274,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15275,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15276,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15277,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15278,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15279,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15280,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15281,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15282,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15283,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15284,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15285,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15286,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15287,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15288,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15289,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15290,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15291,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15292,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15293,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15294,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15295,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15296,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15297,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15298,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15299,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15300,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15301,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15302,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15303,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15304,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15305,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15306,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15307,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15308,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15309,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15310,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15311,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15312,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15313,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15314,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15315,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15316,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15317,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15318,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15319,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15320,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15321,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15322,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15323,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15324,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15325,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15326,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15327,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15328,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15329,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15330,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15331,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15332,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15333,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15334,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15335,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15336,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15337,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15338,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15339,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15340,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15341,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15342,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15343,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15344,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15345,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15346,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15347,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15348,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15349,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15350,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15351,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15352,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15353,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15354,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15355,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15356,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15357,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15358,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15359,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15360,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15361,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15362,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15363,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15364,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15365,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15366,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15367,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15368,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15369,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15370,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15371,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15372,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15373,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15374,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15375,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15376,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15377,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15378,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15379,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15380,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15381,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15382,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15383,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15384,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15385,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15386,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15387,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15388,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15389,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15390,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15391,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15392,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15393,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15394,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15395,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15396,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15397,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15398,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15399,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15400,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15401,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15402,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15403,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15404,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15405,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15406,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15407,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15408,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15409,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15410,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15411,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15412,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15413,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15414,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15415,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15416,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15417,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15418,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15419,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15420,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15421,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15422,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15423,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15424,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15425,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15426,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15427,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15428,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15429,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15430,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15431,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15432,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15433,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15434,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15435,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15436,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15437,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15438,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15439,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15440,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15441,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15442,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15443,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15444,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15445,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15446,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15447,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15448,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15449,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15450,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15451,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15452,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15453,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15454,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15455,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15456,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15457,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15458,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15459,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15460,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15461,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15462,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15463,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15464,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15465,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15466,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15467,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15468,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15469,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15470,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15471,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15472,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15473,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15474,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15475,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15476,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15477,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15478,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15479,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15480,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15481,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15482,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15483,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15484,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15485,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15486,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15487,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15488,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15489,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15490,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15491,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15492,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15493,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15494,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15495,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15496,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15497,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15498,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15499,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15500,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15501,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15502,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15503,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15504,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15505,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15506,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15507,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15508,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15509,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15510,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15511,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15512,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15513,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15514,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15515,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15516,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15517,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15518,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15519,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15520,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15521,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15522,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15523,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15524,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15525,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15526,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15527,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15528,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15529,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15530,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15531,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15532,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15533,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15534,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15535,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15536,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15537,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15538,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15539,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15540,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15541,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15542,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15543,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15544,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15545,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15546,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15547,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15548,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15549,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15550,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15551,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15552,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15553,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15554,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15555,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15556,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15557,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15558,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15559,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15560,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15561,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15562,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15563,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15564,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15565,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15566,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15567,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15568,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15569,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15570,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15571,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15572,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15573,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15574,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15575,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15576,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15577,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15578,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15579,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15580,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15581,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15582,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15583,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15584,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15585,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15586,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15587,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15588,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15589,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15590,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15591,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15592,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15593,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15594,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15595,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15596,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15597,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15598,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15599,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15600,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15601,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15602,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15603,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15604,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15605,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15606,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15607,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15608,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15609,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15610,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15611,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15612,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15613,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15614,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15615,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15616,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15617,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15618,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15619,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15620,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15621,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15622,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15623,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15624,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15625,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15626,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15627,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15628,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15629,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15630,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15631,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15632,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15633,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15634,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15635,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15636,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15637,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15638,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15639,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15640,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15641,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15642,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15643,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15644,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15645,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15646,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15647,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15648,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15649,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15650,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15651,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15652,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15653,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15654,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15655,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15656,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15657,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15658,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15659,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15660,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15661,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15662,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15663,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15664,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15665,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15666,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15667,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15668,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15669,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15670,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15671,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15672,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15673,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15674,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15675,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15676,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15677,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15678,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15679,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15680,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15681,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15682,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15683,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15684,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15685,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15686,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15687,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15688,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15689,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15690,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15691,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15692,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15693,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15694,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15695,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15696,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15697,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15698,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15699,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15700,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15701,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15702,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15703,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15704,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15705,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15706,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15707,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15708,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15709,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15710,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15711,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15712,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15713,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15714,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15715,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15716,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15717,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15718,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15719,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15720,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15721,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15722,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15723,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15724,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15725,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15726,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15727,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15728,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15729,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15730,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15731,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15732,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15733,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15734,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15735,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15736,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15737,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15738,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15739,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15740,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15741,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15742,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15743,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15744,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15745,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15746,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15747,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15748,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15749,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15750,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15751,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15752,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15753,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15754,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15755,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15756,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15757,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15758,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15759,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15760,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15761,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15762,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15763,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15764,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15765,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15766,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15767,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15768,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15769,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15770,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15771,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15772,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15773,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15774,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15775,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15776,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15777,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15778,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15779,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15780,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15781,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15782,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15783,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15784,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15785,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15786,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15787,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15788,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15789,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15790,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15791,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15792,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15793,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15794,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15795,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15796,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15797,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15798,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15799,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15800,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15801,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15802,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15803,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15804,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15805,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15806,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15807,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15808,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15809,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15810,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15811,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15812,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15813,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15814,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15815,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15816,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15817,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15818,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15819,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15820,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15821,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15822,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15823,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15824,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15825,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15826,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15827,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15828,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15829,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15830,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15831,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15832,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15833,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15834,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15835,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15836,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15837,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15838,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15839,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15840,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15841,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15842,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15843,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15844,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15845,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15846,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15847,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15848,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15849,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15850,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15851,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15852,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15853,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15854,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15855,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15856,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15857,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15858,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15859,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15860,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15861,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15862,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15863,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15864,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15865,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15866,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15867,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15868,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15869,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15870,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15871,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15872,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15873,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15874,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15875,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15876,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15877,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15878,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15879,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15880,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15881,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15882,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15883,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15884,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15885,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15886,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15887,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15888,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15889,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15890,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15891,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15892,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15893,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15894,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15895,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15896,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15897,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15898,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15899,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15900,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15901,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15902,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15903,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15904,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15905,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15906,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15907,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15908,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15909,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15910,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15911,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15912,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15913,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15914,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15915,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15916,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15917,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15918,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15919,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15920,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15921,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15922,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15923,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15924,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15925,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15926,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15927,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15928,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15929,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15930,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15931,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15932,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15933,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15934,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15935,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15936,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15937,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15938,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15939,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15940,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15941,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15942,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15943,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15944,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15945,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15946,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15947,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15948,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15949,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15950,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15951,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15952,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15953,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15954,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15955,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15956,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15957,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15958,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15959,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15960,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15961,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15962,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15963,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15964,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15965,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15966,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15967,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15968,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15969,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15970,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15971,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15972,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15973,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15974,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15975,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15976,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15977,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15978,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15979,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15980,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15981,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15982,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15983,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15984,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15985,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15986,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15987,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15988,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15989,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15990,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15991,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15992,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15993,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15994,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15995,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15996,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15997,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15998,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        15999,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16000,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16001,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16002,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16003,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16004,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16005,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16006,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16007,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16008,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16009,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16010,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16011,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16012,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16013,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16014,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16015,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16016,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16017,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16018,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16019,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16020,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16021,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16022,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16023,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16024,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16025,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16026,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16027,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16028,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16029,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16030,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16031,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16032,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16033,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16034,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16035,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16036,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16037,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16038,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16039,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16040,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16041,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16042,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16043,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16044,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16045,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16046,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16047,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16048,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16049,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16050,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16051,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16052,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16053,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16054,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16055,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16056,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16057,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16058,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16059,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16060,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16061,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16062,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16063,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16064,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16065,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16066,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16067,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16068,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16069,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16070,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16071,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16072,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16073,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_RESERVED,
        0,
        {0},
        {0}
    },
    {
        16074,
        {27547, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16075,
        {27566, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16076,
        {146072, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16077,
        {27581, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16078,
        {27582, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16079,
        {27591, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16080,
        {27592, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16081,
        {27593, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7285},
        {37291}
    },
    {
        16082,
        {27610, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16083,
        {27622, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16084,
        {27623, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16085,
        {27630, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5698},
        {27615}
    },
    {
        16086,
        {27633, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1918},
        {38816}
    },
    {
        16087,
        {27650, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16088,
        {27658, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5704},
        {27656}
    },
    {
        16089,
        {27662, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16090,
        {27701, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16091,
        {27702, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2796},
        {19998}
    },
    {
        16092,
        {27706, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2888},
        {27700}
    },
    {
        16093,
        {146559, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16094,
        {27711, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16095,
        {27725, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16096,
        {27739, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16097,
        {27757, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16098,
        {27780, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16099,
        {27785, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3321},
        {27784}
    },
    {
        16100,
        {15555, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5768},
        {28140}
    },
    {
        16101,
        {27796, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16102,
        {27797, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16103,
        {27799, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3409},
        {28155}
    },
    {
        16104,
        {27821, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16105,
        {27842, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16106,
        {27856, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16107,
        {15570, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3948},
        {27861}
    },
    {
        16108,
        {27862, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16109,
        {27866, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16110,
        {27868, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16111,
        {27881, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16112,
        {27884, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16113,
        {27885, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16114,
        {146688, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16115,
        {27904, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16116,
        {27914, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16117,
        {27940, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5790},
        {28278}
    },
    {
        16118,
        {27942, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3652},
        {27850}
    },
    {
        16119,
        {27943, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16120,
        {27751, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16121,
        {27951, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2140},
        {28500}
    },
    {
        16122,
        {27964, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16123,
        {27995, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16124,
        {27998, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16125,
        {28000, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16126,
        {28016, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16127,
        {28032, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16128,
        {28033, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5823},
        {28402}
    },
    {
        16129,
        {28042, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16130,
        {28045, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16131,
        {28049, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16132,
        {28056, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16133,
        {146752, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5759},
        {28037}
    },
    {
        16134,
        {28183, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5823},
        {28402}
    },
    {
        16135,
        {146938, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16136,
        {146937, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3409},
        {28155}
    },
    {
        16137,
        {146899, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16138,
        {28075, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16139,
        {28078, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16140,
        {28084, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16141,
        {28098, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3509},
        {24471}
    },
    {
        16142,
        {27956, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16143,
        {28104, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16144,
        {28110, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16145,
        {28111, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16146,
        {28112, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16147,
        {28127, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16148,
        {28137, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16149,
        {28150, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16150,
        {28214, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16151,
        {28190, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16152,
        {28194, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16153,
        {28199, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16154,
        {15633, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5807},
        {28335}
    },
    {
        16155,
        {28210, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16156,
        {28220, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16157,
        {28232, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16158,
        {28233, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16159,
        {28235, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16160,
        {28236, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2476},
        {39184}
    },
    {
        16161,
        {28239, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16162,
        {28241, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16163,
        {28243, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16164,
        {28244, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16165,
        {28247, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16166,
        {28259, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16167,
        {15646, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5810},
        {28372}
    },
    {
        16168,
        {28307, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5845},
        {28610}
    },
    {
        16169,
        {28327, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16170,
        {28340, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16171,
        {28351, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16172,
        {28355, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16173,
        {28362, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16174,
        {28377, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4632},
        {21295}
    },
    {
        16175,
        {28469, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16176,
        {28395, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16177,
        {28409, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16178,
        {28411, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16179,
        {28426, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16180,
        {28428, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16181,
        {28440, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16182,
        {28453, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2073},
        {31402}
    },
    {
        16183,
        {28470, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16184,
        {28476, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16185,
        {147326, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16186,
        {28498, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16187,
        {28503, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16188,
        {28506, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16189,
        {28512, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16190,
        {28520, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16191,
        {28568, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5833},
        {28536}
    },
    {
        16192,
        {28541, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16193,
        {28560, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16194,
        {28566, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16195,
        {28606, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16196,
        {28575, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16197,
        {28581, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16198,
        {28591, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16199,
        {15716, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16200,
        {28597, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16201,
        {28616, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16202,
        {28617, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16203,
        {28634, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16204,
        {28638, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16205,
        {28649, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16206,
        {147606, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16207,
        {28668, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16208,
        {28672, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16209,
        {28679, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16210,
        {28682, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16211,
        {28707, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16212,
        {147715, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7404},
        {28662}
    },
    {
        16213,
        {28729, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16214,
        {28730, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16215,
        {28732, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16216,
        {28739, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7069},
        {35920}
    },
    {
        16217,
        {28743, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16218,
        {28747, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3948},
        {27861}
    },
    {
        16219,
        {15770, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16220,
        {28756, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16221,
        {28773, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16222,
        {28777, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16223,
        {28780, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16224,
        {28782, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2248},
        {20809}
    },
    {
        16225,
        {28790, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6265},
        {31432}
    },
    {
        16226,
        {28798, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2399},
        {28797}
    },
    {
        16227,
        {28801, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1883},
        {27671}
    },
    {
        16228,
        {28806, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16229,
        {28821, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16230,
        {28823, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2248},
        {20809}
    },
    {
        16231,
        {28859, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16232,
        {147910, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5888},
        {28895}
    },
    {
        16233,
        {28831, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16234,
        {28849, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16235,
        {147966, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16236,
        {28908, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5921},
        {29180}
    },
    {
        16237,
        {28874, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5903},
        {29036}
    },
    {
        16238,
        {28881, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16239,
        {28883, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16240,
        {28892, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16241,
        {28931, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1741},
        {36203}
    },
    {
        16242,
        {28932, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16243,
        {28934, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16244,
        {28935, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16245,
        {28936, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7807},
        {29081}
    },
    {
        16246,
        {28940, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16247,
        {15808, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5900},
        {29026}
    },
    {
        16248,
        {28975, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16249,
        {28977, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16250,
        {29008, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16251,
        {29002, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5444},
        {26244}
    },
    {
        16252,
        {29011, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16253,
        {29022, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2449},
        {27578}
    },
    {
        16254,
        {15828, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1578},
        {28948}
    },
    {
        16255,
        {29078, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1578},
        {28948}
    },
    {
        16256,
        {29056, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16257,
        {29083, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16258,
        {29088, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16259,
        {29090, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5893},
        {28925}
    },
    {
        16260,
        {29102, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16261,
        {29103, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16262,
        {29107, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4042},
        {28459}
    },
    {
        16263,
        {148412, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2678},
        {29087}
    },
    {
        16264,
        {29131, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16265,
        {29139, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16266,
        {29145, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16267,
        {29148, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16268,
        {29191, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16269,
        {15877, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16270,
        {64073, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16271,
        {29227, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16272,
        {29236, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5385},
        {25899}
    },
    {
        16273,
        {29240, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16274,
        {29241, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16275,
        {20012, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5934},
        {29247}
    },
    {
        16276,
        {29250, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16277,
        {29267, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16278,
        {29271, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16279,
        {29283, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16280,
        {149033, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16281,
        {29294, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16282,
        {29295, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16283,
        {29304, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16284,
        {29311, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16285,
        {29326, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16286,
        {149157, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16287,
        {29357, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16288,
        {29358, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16289,
        {29360, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16290,
        {29361, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16291,
        {29377, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16292,
        {15968, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3535},
        {35930}
    },
    {
        16293,
        {29388, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5968},
        {29481}
    },
    {
        16294,
        {15974, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7075},
        {35970}
    },
    {
        16295,
        {15976, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16296,
        {29427, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16297,
        {29434, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16298,
        {29447, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16299,
        {29458, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16300,
        {29464, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16301,
        {29465, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16302,
        {16003, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2046},
        {29399}
    },
    {
        16303,
        {29497, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16304,
        {29484, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6853},
        {34655}
    },
    {
        16305,
        {29489, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16306,
        {29491, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16307,
        {29501, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16308,
        {29522, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16309,
        {16020, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16310,
        {29547, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16311,
        {29548, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16312,
        {149654, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16313,
        {29550, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16314,
        {29551, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16315,
        {29553, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16316,
        {29559, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16317,
        {29569, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16318,
        {29573, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4056},
        {22937}
    },
    {
        16319,
        {29578, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16320,
        {29588, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16321,
        {29592, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16322,
        {29596, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16323,
        {29598, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16324,
        {29605, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16325,
        {29608, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16326,
        {29621, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16327,
        {29623, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16328,
        {29625, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16329,
        {29628, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16330,
        {29631, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16331,
        {29637, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16332,
        {29643, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4245},
        {29705}
    },
    {
        16333,
        {29665, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2028},
        {29748}
    },
    {
        16334,
        {29671, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16335,
        {29689, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16336,
        {29715, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16337,
        {29690, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16338,
        {29697, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16339,
        {29732, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16340,
        {29745, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3295},
        {24427}
    },
    {
        16341,
        {29753, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2028},
        {29748}
    },
    {
        16342,
        {29779, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16343,
        {29760, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16344,
        {29763, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16345,
        {29773, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16346,
        {29778, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16347,
        {29789, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16348,
        {29809, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16349,
        {29825, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16350,
        {29829, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16351,
        {29832, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16352,
        {150093, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16353,
        {29842, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16354,
        {29847, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16355,
        {29849, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16356,
        {29856, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16357,
        {29857, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16358,
        {29861, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16359,
        {29866, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16360,
        {29867, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16361,
        {29881, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16362,
        {29883, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16363,
        {29882, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16364,
        {29910, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16365,
        {29912, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16366,
        {29918, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16367,
        {29935, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6029},
        {29964}
    },
    {
        16368,
        {29931, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4010},
        {30406}
    },
    {
        16369,
        {150358, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16370,
        {29946, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16371,
        {150383, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5570},
        {27006}
    },
    {
        16372,
        {29984, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16373,
        {29988, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16374,
        {29994, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1740},
        {35282}
    },
    {
        16375,
        {16215, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2919},
        {30045}
    },
    {
        16376,
        {150550, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2919},
        {30045}
    },
    {
        16377,
        {30013, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16378,
        {30014, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16379,
        {30016, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16380,
        {30024, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16381,
        {30030, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16382,
        {30032, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16383,
        {30034, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2919},
        {30045}
    },
    {
        16384,
        {30060, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6049},
        {30061}
    },
    {
        16385,
        {30066, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6049},
        {30061}
    },
    {
        16386,
        {30065, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4246},
        {30041}
    },
    {
        16387,
        {30074, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6052},
        {30086}
    },
    {
        16388,
        {30077, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16389,
        {30078, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16390,
        {30081, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16391,
        {150804, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16392,
        {30092, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16393,
        {16245, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16394,
        {30114, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16395,
        {16247, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16396,
        {30128, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16397,
        {30135, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16398,
        {30143, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16399,
        {30144, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16400,
        {30150, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16401,
        {30159, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16402,
        {30163, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16403,
        {30173, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16404,
        {30175, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16405,
        {30176, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16406,
        {30183, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16407,
        {30188, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16408,
        {30190, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3306},
        {33081}
    },
    {
        16409,
        {30193, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16410,
        {30201, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6082},
        {30202}
    },
    {
        16411,
        {30211, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16412,
        {30232, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6096},
        {30267}
    },
    {
        16413,
        {30215, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16414,
        {30223, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16415,
        {16302, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16416,
        {151054, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16417,
        {30227, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16418,
        {30235, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16419,
        {30236, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16420,
        {151095, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16421,
        {30245, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16422,
        {30248, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16423,
        {30268, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16424,
        {30259, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16425,
        {151146, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16426,
        {16329, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16427,
        {30273, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16428,
        {151179, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2849},
        {30137}
    },
    {
        16429,
        {30281, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16430,
        {30293, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6109},
        {30320}
    },
    {
        16431,
        {16343, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16432,
        {30318, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16433,
        {30357, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16434,
        {30364, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16435,
        {30369, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16436,
        {30368, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16437,
        {30375, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16438,
        {30376, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2930},
        {26143}
    },
    {
        16439,
        {30383, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16440,
        {151626, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3678},
        {37474}
    },
    {
        16441,
        {30409, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16442,
        {151637, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16443,
        {30440, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16444,
        {30444, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16445,
        {151842, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16446,
        {30487, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16447,
        {30490, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16448,
        {30509, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16449,
        {30517, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16450,
        {151977, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6152},
        {30568}
    },
    {
        16451,
        {16441, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6153},
        {30571}
    },
    {
        16452,
        {152037, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6168},
        {30652}
    },
    {
        16453,
        {152013, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16454,
        {30552, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16455,
        {30560, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6148},
        {30519}
    },
    {
        16456,
        {30570, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16457,
        {152094, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6163},
        {30622}
    },
    {
        16458,
        {30578, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16459,
        {30588, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16460,
        {30589, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16461,
        {152140, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16462,
        {16472, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16463,
        {30618, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2685},
        {30636}
    },
    {
        16464,
        {30623, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16465,
        {30626, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16466,
        {30628, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16467,
        {30633, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6174},
        {30682}
    },
    {
        16468,
        {30686, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16469,
        {30687, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5591},
        {27082}
    },
    {
        16470,
        {30692, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3645},
        {30695}
    },
    {
        16471,
        {30694, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2245},
        {20399}
    },
    {
        16472,
        {30698, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16473,
        {30700, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16474,
        {16531, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1891},
        {35215}
    },
    {
        16475,
        {30704, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16476,
        {30708, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3375},
        {30855}
    },
    {
        16477,
        {30715, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16478,
        {152622, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16479,
        {30725, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16480,
        {30726, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16481,
        {30729, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16482,
        {30733, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16483,
        {30745, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16484,
        {30753, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16485,
        {30764, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16486,
        {30791, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16487,
        {30820, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16488,
        {30826, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16489,
        {152793, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16490,
        {30858, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16491,
        {30868, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16492,
        {30884, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16493,
        {30877, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16494,
        {30878, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5022},
        {23890}
    },
    {
        16495,
        {30879, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16496,
        {30907, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1737},
        {30906}
    },
    {
        16497,
        {30920, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16498,
        {30924, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16499,
        {30926, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16500,
        {30933, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16501,
        {30944, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16502,
        {30945, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16503,
        {30950, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16504,
        {30969, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16505,
        {30970, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6181},
        {31014}
    },
    {
        16506,
        {30971, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16507,
        {30974, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16508,
        {152999, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16509,
        {30992, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4822},
        {22470}
    },
    {
        16510,
        {31003, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16511,
        {31024, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16512,
        {31013, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16513,
        {31035, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16514,
        {31050, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16515,
        {31064, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16516,
        {31067, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16517,
        {16645, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16518,
        {31079, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16519,
        {31090, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16520,
        {31124, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16521,
        {31125, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16522,
        {31126, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16523,
        {31131, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16524,
        {31137, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16525,
        {31145, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6214},
        {31040}
    },
    {
        16526,
        {31156, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16527,
        {31163, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16528,
        {31170, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16529,
        {31175, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2128},
        {33464}
    },
    {
        16530,
        {31180, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2640},
        {31179}
    },
    {
        16531,
        {31181, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16532,
        {31190, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16533,
        {16712, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16534,
        {153513, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16535,
        {153524, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16536,
        {16719, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16537,
        {31242, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16538,
        {31249, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16539,
        {31253, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16540,
        {31259, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16541,
        {31262, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16542,
        {16739, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16543,
        {31277, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16544,
        {31288, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3258},
        {33988}
    },
    {
        16545,
        {31303, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16546,
        {31308, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16547,
        {31318, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16548,
        {31321, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16549,
        {31324, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16550,
        {31327, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3923},
        {31298}
    },
    {
        16551,
        {31328, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16552,
        {31335, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16553,
        {31338, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6245},
        {31281}
    },
    {
        16554,
        {31349, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16555,
        {31352, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16556,
        {31362, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4343},
        {29282}
    },
    {
        16557,
        {31370, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16558,
        {31376, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16559,
        {31395, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16560,
        {31404, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16561,
        {154068, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4978},
        {23617}
    },
    {
        16562,
        {16820, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1777},
        {31427}
    },
    {
        16563,
        {31417, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16564,
        {31420, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16565,
        {31422, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16566,
        {16831, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16567,
        {31436, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16568,
        {31441, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16569,
        {31463, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16570,
        {31464, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16571,
        {31476, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16572,
        {154340, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16573,
        {154339, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16574,
        {31495, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2470},
        {31639}
    },
    {
        16575,
        {154353, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16576,
        {31549, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16577,
        {31527, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2442},
        {20874}
    },
    {
        16578,
        {31530, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16579,
        {31534, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16580,
        {31535, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16581,
        {31537, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16582,
        {16870, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1552},
        {33521}
    },
    {
        16583,
        {16883, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6318},
        {31627}
    },
    {
        16584,
        {31615, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2573},
        {31712}
    },
    {
        16585,
        {31553, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16586,
        {16878, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16587,
        {31573, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16588,
        {31609, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6351},
        {31820}
    },
    {
        16589,
        {31588, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16590,
        {31590, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16591,
        {31593, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16592,
        {31603, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16593,
        {154546, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16594,
        {16903, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6498},
        {32617}
    },
    {
        16595,
        {31632, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16596,
        {31633, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16597,
        {31643, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16598,
        {16910, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16599,
        {31663, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16600,
        {31669, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16601,
        {31676, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16602,
        {31685, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16603,
        {31690, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16604,
        {154699, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6341},
        {31767}
    },
    {
        16605,
        {154724, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16606,
        {31700, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16607,
        {31702, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16608,
        {31706, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16609,
        {31722, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16610,
        {31728, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16611,
        {31747, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16612,
        {31755, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16613,
        {31758, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16614,
        {31759, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16615,
        {31782, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16616,
        {31813, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16617,
        {31818, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16618,
        {31825, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2994},
        {25776}
    },
    {
        16619,
        {31831, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16620,
        {31838, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16621,
        {31841, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16622,
        {31849, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16623,
        {31854, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16624,
        {31855, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16625,
        {31856, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16626,
        {155182, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2079},
        {26705}
    },
    {
        16627,
        {155222, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16628,
        {155237, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3475},
        {31958}
    },
    {
        16629,
        {31910, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4282},
        {29136}
    },
    {
        16630,
        {155234, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3878},
        {32862}
    },
    {
        16631,
        {31926, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16632,
        {31927, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16633,
        {31935, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16634,
        {155352, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16635,
        {31940, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16636,
        {155330, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16637,
        {31944, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16638,
        {31949, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16639,
        {155368, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16640,
        {31959, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16641,
        {155427, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16642,
        {31974, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16643,
        {31979, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16644,
        {155484, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3672},
        {27368}
    },
    {
        16645,
        {31989, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16646,
        {32003, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16647,
        {32009, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16648,
        {17094, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16649,
        {32018, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16650,
        {32030, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16651,
        {155616, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16652,
        {155604, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16653,
        {32061, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16654,
        {32062, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16655,
        {32064, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16656,
        {32071, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16657,
        {155660, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16658,
        {155643, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16659,
        {17110, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16660,
        {32089, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16661,
        {32090, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16662,
        {32106, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16663,
        {32112, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16664,
        {17117, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6478},
        {32403}
    },
    {
        16665,
        {32127, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16666,
        {155671, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2824},
        {32340}
    },
    {
        16667,
        {32134, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16668,
        {32136, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16669,
        {32140, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16670,
        {32151, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6396},
        {32069}
    },
    {
        16671,
        {155744, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16672,
        {32157, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16673,
        {32167, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16674,
        {32170, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16675,
        {32182, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16676,
        {32183, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16677,
        {32192, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16678,
        {32215, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16679,
        {32217, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16680,
        {32230, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16681,
        {32241, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16682,
        {32249, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16683,
        {17154, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1423},
        {25569}
    },
    {
        16684,
        {155885, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16685,
        {64088, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6445},
        {32265}
    },
    {
        16686,
        {32272, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16687,
        {32279, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16688,
        {32285, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16689,
        {32288, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16690,
        {32295, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6413},
        {32091}
    },
    {
        16691,
        {32300, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16692,
        {32325, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2078},
        {32368}
    },
    {
        16693,
        {32371, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16694,
        {32373, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16695,
        {32382, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16696,
        {32390, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16697,
        {32391, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16698,
        {17195, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6483},
        {32412}
    },
    {
        16699,
        {32401, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16700,
        {32408, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6474},
        {32393}
    },
    {
        16701,
        {32410, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16702,
        {17219, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16703,
        {32572, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6485},
        {32570}
    },
    {
        16704,
        {32571, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1829},
        {32566}
    },
    {
        16705,
        {32574, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3810},
        {29942}
    },
    {
        16706,
        {32579, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16707,
        {32580, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16708,
        {32591, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16709,
        {13505, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6491},
        {32593}
    },
    {
        16710,
        {156272, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6491},
        {32593}
    },
    {
        16711,
        {32594, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6491},
        {32593}
    },
    {
        16712,
        {156294, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16713,
        {32609, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16714,
        {32611, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16715,
        {32612, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16716,
        {32621, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16717,
        {32637, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16718,
        {32638, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16719,
        {156492, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16720,
        {32656, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16721,
        {20859, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3759},
        {32654}
    },
    {
        16722,
        {146702, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2746},
        {25215}
    },
    {
        16723,
        {32662, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16724,
        {32668, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16725,
        {32685, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16726,
        {156674, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4202},
        {32764}
    },
    {
        16727,
        {32707, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16728,
        {32719, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16729,
        {32739, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16730,
        {32741, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16731,
        {32751, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16732,
        {32754, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16733,
        {32770, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16734,
        {32778, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6532},
        {32779}
    },
    {
        16735,
        {32776, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16736,
        {32782, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16737,
        {32785, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16738,
        {32790, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16739,
        {32804, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16740,
        {32812, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16741,
        {32816, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16742,
        {32835, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16743,
        {32870, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6551},
        {32880}
    },
    {
        16744,
        {32881, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16745,
        {32885, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16746,
        {32891, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16747,
        {32921, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16748,
        {32924, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16749,
        {32932, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3829},
        {33178}
    },
    {
        16750,
        {32935, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6570},
        {32986}
    },
    {
        16751,
        {32952, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16752,
        {157310, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6575},
        {33051}
    },
    {
        16753,
        {32965, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16754,
        {32981, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16755,
        {32984, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16756,
        {32998, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16757,
        {157360, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4361},
        {33031}
    },
    {
        16758,
        {33037, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6605},
        {33214}
    },
    {
        16759,
        {33013, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16760,
        {33019, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16761,
        {17390, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4361},
        {33031}
    },
    {
        16762,
        {33077, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2213},
        {32929}
    },
    {
        16763,
        {33046, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16764,
        {33054, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16765,
        {17392, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16766,
        {33060, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16767,
        {33063, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16768,
        {33068, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16769,
        {157469, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6612},
        {33229}
    },
    {
        16770,
        {33085, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16771,
        {17416, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16772,
        {33129, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16773,
        {17431, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3281},
        {29482}
    },
    {
        16774,
        {33153, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16775,
        {17436, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16776,
        {33156, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3080},
        {30185}
    },
    {
        16777,
        {33157, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16778,
        {17442, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16779,
        {33176, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16780,
        {33202, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16781,
        {33217, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16782,
        {33219, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6109},
        {30320}
    },
    {
        16783,
        {33238, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16784,
        {33243, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16785,
        {157917, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7510},
        {38948}
    },
    {
        16786,
        {33252, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2174},
        {36066}
    },
    {
        16787,
        {157930, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1589},
        {22633}
    },
    {
        16788,
        {33260, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16789,
        {158033, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1520},
        {33276}
    },
    {
        16790,
        {33277, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16791,
        {33279, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16792,
        {158063, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16793,
        {33284, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16794,
        {158173, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16795,
        {33305, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16796,
        {33313, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16797,
        {33314, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16798,
        {158238, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16799,
        {33330, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16800,
        {33332, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16801,
        {33340, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16802,
        {33350, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16803,
        {33353, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16804,
        {33349, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16805,
        {158296, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16806,
        {33355, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16807,
        {17526, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16808,
        {33359, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16809,
        {17530, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3698},
        {24070}
    },
    {
        16810,
        {33367, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16811,
        {158348, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16812,
        {33372, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16813,
        {33379, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6645},
        {33386}
    },
    {
        16814,
        {158391, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16815,
        {64093, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6650},
        {33400}
    },
    {
        16816,
        {64094, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6650},
        {33400}
    },
    {
        16817,
        {33401, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6650},
        {33400}
    },
    {
        16818,
        {17553, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6650},
        {33400}
    },
    {
        16819,
        {33405, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16820,
        {33407, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16821,
        {33411, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16822,
        {33418, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16823,
        {33427, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16824,
        {33447, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16825,
        {33448, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16826,
        {33458, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1651},
        {33457}
    },
    {
        16827,
        {33460, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16828,
        {33466, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16829,
        {33468, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16830,
        {33506, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6658},
        {33505}
    },
    {
        16831,
        {33512, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16832,
        {33527, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16833,
        {33543, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16834,
        {33544, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16835,
        {33548, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16836,
        {33620, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6686},
        {33560}
    },
    {
        16837,
        {33563, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16838,
        {33565, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16839,
        {33584, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6721},
        {33848}
    },
    {
        16840,
        {33596, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16841,
        {33604, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16842,
        {33623, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16843,
        {17598, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16844,
        {33663, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16845,
        {17620, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16846,
        {17587, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5471},
        {26422}
    },
    {
        16847,
        {33677, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16848,
        {33684, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16849,
        {33685, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16850,
        {33691, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16851,
        {33693, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16852,
        {33737, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16853,
        {33744, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16854,
        {33748, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6762},
        {34068}
    },
    {
        16855,
        {33757, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16856,
        {33765, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16857,
        {33785, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16858,
        {33807, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16859,
        {33809, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16860,
        {33813, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6788},
        {34234}
    },
    {
        16861,
        {158835, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6677},
        {33586}
    },
    {
        16862,
        {33815, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16863,
        {33849, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16864,
        {33866, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6704},
        {33780}
    },
    {
        16865,
        {33871, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16866,
        {33873, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16867,
        {33874, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16868,
        {33881, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16869,
        {33882, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16870,
        {33884, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16871,
        {158941, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16872,
        {33893, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16873,
        {33910, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16874,
        {33912, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16875,
        {33916, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16876,
        {33921, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16877,
        {17677, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3515},
        {31716}
    },
    {
        16878,
        {34012, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16879,
        {33943, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16880,
        {33958, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16881,
        {33982, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16882,
        {17672, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16883,
        {33998, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16884,
        {33999, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16885,
        {34003, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16886,
        {159333, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1503},
        {34093}
    },
    {
        16887,
        {34023, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16888,
        {34026, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16889,
        {34031, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16890,
        {34032, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16891,
        {34033, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6717},
        {33805}
    },
    {
        16892,
        {34042, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6793},
        {34269}
    },
    {
        16893,
        {34045, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16894,
        {34060, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16895,
        {34075, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16896,
        {34084, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16897,
        {34085, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3583},
        {33905}
    },
    {
        16898,
        {34091, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16899,
        {34100, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4014},
        {40635}
    },
    {
        16900,
        {34127, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16901,
        {34159, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16902,
        {17701, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6778},
        {34157}
    },
    {
        16903,
        {17731, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1503},
        {34093}
    },
    {
        16904,
        {34110, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16905,
        {34129, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16906,
        {34131, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16907,
        {34142, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16908,
        {34145, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16909,
        {34146, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16910,
        {159636, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6628},
        {33290}
    },
    {
        16911,
        {34171, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16912,
        {34173, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16913,
        {34175, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1785},
        {33841}
    },
    {
        16914,
        {34177, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16915,
        {34182, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16916,
        {34195, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6758},
        {34072}
    },
    {
        16917,
        {34205, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16918,
        {34207, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16919,
        {159736, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16920,
        {159734, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16921,
        {159735, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1611},
        {33659}
    },
    {
        16922,
        {34231, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16923,
        {34236, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16924,
        {34247, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16925,
        {34250, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16926,
        {34264, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16927,
        {34265, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16928,
        {34271, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16929,
        {34273, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1611},
        {33659}
    },
    {
        16930,
        {34278, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16931,
        {34294, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16932,
        {34304, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16933,
        {34321, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16934,
        {34334, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16935,
        {34337, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16936,
        {34340, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1651},
        {33457}
    },
    {
        16937,
        {34343, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16938,
        {160013, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16939,
        {34361, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16940,
        {34364, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16941,
        {160057, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16942,
        {34368, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7503},
        {38898}
    },
    {
        16943,
        {64032, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16944,
        {34387, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16945,
        {34390, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16946,
        {34415, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16947,
        {34423, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16948,
        {34426, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16949,
        {34439, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16950,
        {34441, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1664},
        {34442}
    },
    {
        16951,
        {34445, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16952,
        {34449, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16953,
        {34460, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16954,
        {34461, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16955,
        {34472, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16956,
        {64033, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16957,
        {34481, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16958,
        {34483, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16959,
        {34497, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16960,
        {34499, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16961,
        {34513, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16962,
        {34517, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6827},
        {34516}
    },
    {
        16963,
        {34519, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16964,
        {34531, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16965,
        {34534, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16966,
        {17848, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1672},
        {34558}
    },
    {
        16967,
        {34565, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16968,
        {34567, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16969,
        {34574, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16970,
        {34576, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16971,
        {34579, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16972,
        {34585, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16973,
        {34591, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16974,
        {34593, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16975,
        {34595, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16976,
        {34609, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16977,
        {34618, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16978,
        {34622, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16979,
        {34624, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16980,
        {34627, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16981,
        {34641, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16982,
        {34648, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16983,
        {34660, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16984,
        {34661, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16985,
        {34674, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16986,
        {34684, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6880},
        {34747}
    },
    {
        16987,
        {160731, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16988,
        {160730, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16989,
        {34727, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6833},
        {34543}
    },
    {
        16990,
        {34697, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16991,
        {34699, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16992,
        {34707, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16993,
        {34720, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16994,
        {160766, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16995,
        {17893, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16996,
        {34750, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6813},
        {34451}
    },
    {
        16997,
        {160784, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        16998,
        {34753, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1664},
        {34442}
    },
    {
        16999,
        {34766, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17000,
        {34805, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17001,
        {34783, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17002,
        {160841, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17003,
        {34787, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17004,
        {34789, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17005,
        {34790, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17006,
        {34794, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17007,
        {34795, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17008,
        {34797, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17009,
        {34817, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4775},
        {22190}
    },
    {
        17010,
        {34819, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4205},
        {34746}
    },
    {
        17011,
        {34827, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6837},
        {34560}
    },
    {
        17012,
        {34835, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17013,
        {34856, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17014,
        {34862, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17015,
        {34866, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17016,
        {34876, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17017,
        {17935, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5160},
        {24676}
    },
    {
        17018,
        {34890, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17019,
        {34904, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7319},
        {37532}
    },
    {
        17020,
        {34911, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3504},
        {36947}
    },
    {
        17021,
        {34916, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17022,
        {161301, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17023,
        {161300, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17024,
        {34921, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17025,
        {161329, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17026,
        {34927, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17027,
        {34976, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5050},
        {24089}
    },
    {
        17028,
        {35004, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17029,
        {35005, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17030,
        {35006, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17031,
        {35008, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17032,
        {35026, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17033,
        {161427, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17034,
        {35025, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17035,
        {35027, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17036,
        {35035, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17037,
        {35056, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17038,
        {35057, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17039,
        {17985, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17040,
        {35073, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17041,
        {161550, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6962},
        {35191}
    },
    {
        17042,
        {35127, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17043,
        {161571, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17044,
        {35138, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17045,
        {35141, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17046,
        {35145, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6458},
        {32349}
    },
    {
        17047,
        {161618, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17048,
        {18021, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17049,
        {35170, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17050,
        {35200, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6963},
        {35198}
    },
    {
        17051,
        {35209, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6505},
        {32647}
    },
    {
        17052,
        {35216, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1739},
        {35226}
    },
    {
        17053,
        {35231, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17054,
        {35248, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17055,
        {35255, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17056,
        {35286, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17057,
        {35288, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17058,
        {35307, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17059,
        {18081, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17060,
        {35313, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17061,
        {35315, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17062,
        {35325, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17063,
        {35327, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17064,
        {18095, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17065,
        {35345, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3135},
        {35409}
    },
    {
        17066,
        {35348, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17067,
        {162181, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7035},
        {35635}
    },
    {
        17068,
        {35361, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2040},
        {21535}
    },
    {
        17069,
        {35381, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17070,
        {35390, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17071,
        {35397, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17072,
        {35405, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17073,
        {35416, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17074,
        {35502, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17075,
        {35472, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17076,
        {35511, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17077,
        {35518, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17078,
        {35543, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17079,
        {35580, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7022},
        {35552}
    },
    {
        17080,
        {162436, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1892},
        {35352}
    },
    {
        17081,
        {35594, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17082,
        {35589, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17083,
        {35597, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17084,
        {35612, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17085,
        {35615, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17086,
        {35629, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17087,
        {35651, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17088,
        {18188, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17089,
        {35665, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17090,
        {35678, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17091,
        {35702, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17092,
        {35711, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17093,
        {35713, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7038},
        {35627}
    },
    {
        17094,
        {35723, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17095,
        {35732, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17096,
        {35733, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17097,
        {35740, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17098,
        {35742, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17099,
        {35897, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17100,
        {162739, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17101,
        {35901, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17102,
        {162750, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7065},
        {35903}
    },
    {
        17103,
        {162759, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17104,
        {35909, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17105,
        {35911, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17106,
        {35919, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17107,
        {35924, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1583},
        {33398}
    },
    {
        17108,
        {35927, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17109,
        {35945, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17110,
        {35949, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17111,
        {35955, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17112,
        {163000, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7075},
        {35970}
    },
    {
        17113,
        {35987, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3584},
        {29483}
    },
    {
        17114,
        {35986, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5967},
        {29487}
    },
    {
        17115,
        {35993, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17116,
        {18276, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17117,
        {35995, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17118,
        {36004, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17119,
        {36054, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17120,
        {36053, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17121,
        {36057, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17122,
        {163232, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17123,
        {36080, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17124,
        {36081, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17125,
        {163344, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6602},
        {33193}
    },
    {
        17126,
        {36105, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17127,
        {36110, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17128,
        {36204, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17129,
        {36228, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17130,
        {36245, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17131,
        {36262, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17132,
        {163767, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17133,
        {36294, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17134,
        {36296, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17135,
        {36313, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17136,
        {36332, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17137,
        {36364, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7163},
        {36490}
    },
    {
        17138,
        {18429, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3040},
        {30095}
    },
    {
        17139,
        {36349, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17140,
        {36358, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17141,
        {163978, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17142,
        {36372, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17143,
        {36374, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17144,
        {36385, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17145,
        {36386, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17146,
        {36391, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17147,
        {164027, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7148},
        {36424}
    },
    {
        17148,
        {18454, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17149,
        {36406, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17150,
        {36409, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17151,
        {36427, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3484},
        {36367}
    },
    {
        17152,
        {36436, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17153,
        {36450, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17154,
        {36460, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17155,
        {36461, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17156,
        {36463, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6363},
        {31861}
    },
    {
        17157,
        {36504, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17158,
        {36510, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17159,
        {36526, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17160,
        {36531, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7170},
        {36524}
    },
    {
        17161,
        {36533, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17162,
        {36534, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4206},
        {35064}
    },
    {
        17163,
        {36539, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17164,
        {164482, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17165,
        {36561, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17166,
        {36564, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17167,
        {18510, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7209},
        {36708}
    },
    {
        17168,
        {36601, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17169,
        {164595, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17170,
        {36608, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7206},
        {36700}
    },
    {
        17171,
        {36616, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17172,
        {36631, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17173,
        {36651, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17174,
        {36672, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17175,
        {36682, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17176,
        {36696, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17177,
        {164876, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2860},
        {36763}
    },
    {
        17178,
        {36772, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2550},
        {36766}
    },
    {
        17179,
        {36788, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17180,
        {64102, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17181,
        {36790, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17182,
        {164949, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17183,
        {36801, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3018},
        {36983}
    },
    {
        17184,
        {36806, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17185,
        {64036, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3907},
        {36820}
    },
    {
        17186,
        {36810, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3064},
        {21277}
    },
    {
        17187,
        {36813, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17188,
        {36819, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17189,
        {36821, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17190,
        {36832, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17191,
        {36849, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17192,
        {36853, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17193,
        {36859, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17194,
        {36866, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17195,
        {36876, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17196,
        {36919, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7231},
        {36886}
    },
    {
        17197,
        {165227, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17198,
        {36931, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17199,
        {36932, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17200,
        {36957, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17201,
        {165320, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17202,
        {165321, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6271},
        {36995}
    },
    {
        17203,
        {36997, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17204,
        {37004, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17205,
        {37008, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17206,
        {38429, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3835},
        {38428}
    },
    {
        17207,
        {37025, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17208,
        {18613, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17209,
        {37040, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17210,
        {37046, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17211,
        {37059, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17212,
        {37064, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17213,
        {165591, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17214,
        {37084, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17215,
        {37087, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17216,
        {165626, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17217,
        {37110, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3843},
        {37096}
    },
    {
        17218,
        {37106, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17219,
        {37120, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17220,
        {37099, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17221,
        {37118, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17222,
        {37119, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17223,
        {37124, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17224,
        {37126, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17225,
        {37144, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17226,
        {37148, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17227,
        {37150, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17228,
        {37175, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17229,
        {37177, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17230,
        {37178, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17231,
        {37190, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17232,
        {37191, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17233,
        {37207, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17234,
        {37209, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5405},
        {26015}
    },
    {
        17235,
        {37217, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17236,
        {37220, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17237,
        {37236, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17238,
        {37241, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17239,
        {37253, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17240,
        {37262, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7743},
        {40569}
    },
    {
        17241,
        {37288, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17242,
        {37294, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17243,
        {37299, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17244,
        {37302, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17245,
        {37315, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17246,
        {37316, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17247,
        {37338, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17248,
        {166217, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17249,
        {166214, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17250,
        {37356, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17251,
        {37358, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17252,
        {37377, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17253,
        {37386, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17254,
        {37398, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17255,
        {37399, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17256,
        {166251, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7323},
        {37466}
    },
    {
        17257,
        {37427, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17258,
        {37442, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17259,
        {37447, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7302},
        {37366}
    },
    {
        17260,
        {37450, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17261,
        {37454, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17262,
        {37457, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17263,
        {37462, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17264,
        {37465, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17265,
        {37472, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17266,
        {37473, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17267,
        {37477, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17268,
        {37479, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17269,
        {37480, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17270,
        {166279, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7324},
        {37583}
    },
    {
        17271,
        {166280, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17272,
        {37500, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17273,
        {37501, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17274,
        {37503, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17275,
        {37513, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17276,
        {37517, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17277,
        {37527, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17278,
        {37529, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17279,
        {37535, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17280,
        {37543, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17281,
        {37547, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17282,
        {166330, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17283,
        {166331, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17284,
        {37554, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17285,
        {37567, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2791},
        {37912}
    },
    {
        17286,
        {37568, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17287,
        {37574, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17288,
        {37582, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17289,
        {37584, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17290,
        {37591, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17291,
        {37593, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17292,
        {37605, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17293,
        {37607, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17294,
        {37649, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17295,
        {166430, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3823},
        {26023}
    },
    {
        17296,
        {166441, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17297,
        {37623, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17298,
        {37625, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17299,
        {37627, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17300,
        {37634, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17301,
        {37645, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17302,
        {37653, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17303,
        {37661, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17304,
        {37662, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17305,
        {37671, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17306,
        {37673, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17307,
        {166513, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17308,
        {166467, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17309,
        {37703, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17310,
        {37713, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17311,
        {37719, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17312,
        {37722, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17313,
        {37739, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2080},
        {37740}
    },
    {
        17314,
        {37745, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17315,
        {37747, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7371},
        {37970}
    },
    {
        17316,
        {37793, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17317,
        {166553, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7353},
        {37848}
    },
    {
        17318,
        {166605, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17319,
        {37768, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17320,
        {37771, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1768},
        {36676}
    },
    {
        17321,
        {37775, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17322,
        {37790, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17323,
        {37877, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17324,
        {166628, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17325,
        {166621, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7380},
        {38005}
    },
    {
        17326,
        {37873, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17327,
        {37825, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2380},
        {37782}
    },
    {
        17328,
        {37831, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17329,
        {37852, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7369},
        {37946}
    },
    {
        17330,
        {37858, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17331,
        {37863, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17332,
        {37897, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17333,
        {37903, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17334,
        {37910, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17335,
        {37911, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17336,
        {37883, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17337,
        {37938, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17338,
        {37940, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17339,
        {37947, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17340,
        {37957, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17341,
        {166849, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17342,
        {166895, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4131},
        {37971}
    },
    {
        17343,
        {37997, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17344,
        {37999, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17345,
        {38264, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3314},
        {38263}
    },
    {
        17346,
        {38265, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7604},
        {39647}
    },
    {
        17347,
        {38278, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7400},
        {38331}
    },
    {
        17348,
        {38284, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17349,
        {38285, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17350,
        {167184, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17351,
        {38315, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17352,
        {38324, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7405},
        {38339}
    },
    {
        17353,
        {167281, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17354,
        {38344, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17355,
        {167419, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17356,
        {167455, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17357,
        {38444, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2258},
        {22353}
    },
    {
        17358,
        {38451, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4194},
        {38525}
    },
    {
        17359,
        {38452, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1505},
        {38512}
    },
    {
        17360,
        {167478, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17361,
        {38460, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17362,
        {38465, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3144},
        {38464}
    },
    {
        17363,
        {38497, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17364,
        {167561, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17365,
        {38530, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1505},
        {38512}
    },
    {
        17366,
        {167659, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7440},
        {38580}
    },
    {
        17367,
        {38554, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17368,
        {167730, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1503},
        {34093}
    },
    {
        17369,
        {18919, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5033},
        {23980}
    },
    {
        17370,
        {38569, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17371,
        {38575, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7428},
        {38502}
    },
    {
        17372,
        {38579, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17373,
        {38586, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4956},
        {23497}
    },
    {
        17374,
        {38589, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7445},
        {38603}
    },
    {
        17375,
        {18938, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4161},
        {38596}
    },
    {
        17376,
        {167928, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2127},
        {40335}
    },
    {
        17377,
        {38616, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17378,
        {38618, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17379,
        {38621, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7447},
        {38605}
    },
    {
        17380,
        {18948, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17381,
        {38676, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17382,
        {38691, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17383,
        {18985, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4307},
        {38666}
    },
    {
        17384,
        {38710, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5814},
        {28354}
    },
    {
        17385,
        {38721, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4210},
        {38647}
    },
    {
        17386,
        {38727, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17387,
        {38741, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17388,
        {38743, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17389,
        {38747, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17390,
        {38762, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17391,
        {168608, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17392,
        {168625, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17393,
        {38806, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17394,
        {38810, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17395,
        {38814, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17396,
        {38818, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17397,
        {38833, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7500},
        {38876}
    },
    {
        17398,
        {38834, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17399,
        {38846, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2071},
        {38772}
    },
    {
        17400,
        {38860, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2876},
        {38765}
    },
    {
        17401,
        {38865, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17402,
        {38868, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17403,
        {38872, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17404,
        {38873, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17405,
        {38881, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17406,
        {38897, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17407,
        {38916, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17408,
        {38925, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17409,
        {38926, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17410,
        {38932, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17411,
        {38934, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17412,
        {19132, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17413,
        {169104, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17414,
        {38947, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17415,
        {38962, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17416,
        {38963, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17417,
        {38949, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7510},
        {38948}
    },
    {
        17418,
        {38983, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17419,
        {39014, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17420,
        {39083, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17421,
        {39085, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17422,
        {39088, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17423,
        {169423, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17424,
        {39095, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7529},
        {39110}
    },
    {
        17425,
        {39096, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17426,
        {39099, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17427,
        {39100, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17428,
        {39103, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3698},
        {24070}
    },
    {
        17429,
        {39106, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17430,
        {39111, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7529},
        {39110}
    },
    {
        17431,
        {39115, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17432,
        {39136, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17433,
        {169599, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17434,
        {39137, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2476},
        {39184}
    },
    {
        17435,
        {39139, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17436,
        {39141, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17437,
        {39146, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17438,
        {39152, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3716},
        {39151}
    },
    {
        17439,
        {39153, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17440,
        {39155, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17441,
        {39176, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {6368},
        {31906}
    },
    {
        17442,
        {19259, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7539},
        {39198}
    },
    {
        17443,
        {169712, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17444,
        {39190, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17445,
        {39191, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17446,
        {169753, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7552},
        {39250}
    },
    {
        17447,
        {39194, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2421},
        {32948}
    },
    {
        17448,
        {39195, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17449,
        {39196, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17450,
        {169808, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17451,
        {39217, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17452,
        {39218, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17453,
        {39219, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3475},
        {31958}
    },
    {
        17454,
        {39226, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17455,
        {39227, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17456,
        {39228, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17457,
        {39232, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17458,
        {39233, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17459,
        {39238, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17460,
        {39245, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {3030},
        {33203}
    },
    {
        17461,
        {39246, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17462,
        {39260, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17463,
        {39263, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7533},
        {39177}
    },
    {
        17464,
        {39264, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17465,
        {39331, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17466,
        {39334, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17467,
        {39353, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17468,
        {39357, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17469,
        {39359, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7589},
        {39522}
    },
    {
        17470,
        {39363, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17471,
        {39369, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17472,
        {39380, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17473,
        {39385, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17474,
        {39390, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7563},
        {39389}
    },
    {
        17475,
        {170182, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17476,
        {39408, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17477,
        {39417, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17478,
        {39420, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17479,
        {39434, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17480,
        {39441, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17481,
        {39446, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17482,
        {39450, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17483,
        {39456, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17484,
        {39473, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17485,
        {39478, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17486,
        {39492, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17487,
        {39500, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17488,
        {39512, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7583},
        {39486}
    },
    {
        17489,
        {19394, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17490,
        {39599, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17491,
        {19402, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17492,
        {39607, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17493,
        {19410, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17494,
        {39609, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17495,
        {170610, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17496,
        {39622, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17497,
        {39632, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17498,
        {39634, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17499,
        {39637, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17500,
        {19432, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5399},
        {25970}
    },
    {
        17501,
        {39644, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17502,
        {39648, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17503,
        {39653, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7608},
        {39663}
    },
    {
        17504,
        {39657, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7619},
        {39714}
    },
    {
        17505,
        {39683, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17506,
        {39692, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17507,
        {39696, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17508,
        {39698, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17509,
        {39702, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17510,
        {39708, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17511,
        {39723, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17512,
        {39731, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17513,
        {39741, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4046},
        {39749}
    },
    {
        17514,
        {19488, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17515,
        {39755, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17516,
        {39779, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17517,
        {39781, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17518,
        {39787, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17519,
        {39788, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17520,
        {39795, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17521,
        {39798, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17522,
        {39799, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17523,
        {39846, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17524,
        {39852, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17525,
        {39857, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17526,
        {171483, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7652},
        {39890}
    },
    {
        17527,
        {171581, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17528,
        {39858, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17529,
        {39864, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17530,
        {39870, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17531,
        {39879, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17532,
        {39923, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17533,
        {39896, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17534,
        {39901, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17535,
        {39911, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17536,
        {39914, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17537,
        {39915, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17538,
        {39919, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17539,
        {39918, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17540,
        {171541, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17541,
        {39930, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17542,
        {171658, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17543,
        {39927, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17544,
        {171593, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17545,
        {39958, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17546,
        {39960, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17547,
        {39961, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17548,
        {39962, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17549,
        {39965, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17550,
        {39970, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17551,
        {39975, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17552,
        {39977, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17553,
        {39978, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17554,
        {171716, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17555,
        {39985, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17556,
        {39990, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17557,
        {39991, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17558,
        {40005, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17559,
        {40028, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17560,
        {171753, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17561,
        {40009, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7779},
        {40712}
    },
    {
        17562,
        {40010, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17563,
        {171739, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17564,
        {40020, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17565,
        {40024, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17566,
        {40027, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17567,
        {40029, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17568,
        {40031, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17569,
        {40041, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17570,
        {40042, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17571,
        {40043, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17572,
        {40045, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17573,
        {40046, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17574,
        {40048, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17575,
        {40050, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17576,
        {40053, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17577,
        {40058, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17578,
        {40166, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17579,
        {40178, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17580,
        {40203, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17581,
        {40194, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7693},
        {40195}
    },
    {
        17582,
        {171982, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7690},
        {40201}
    },
    {
        17583,
        {40209, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17584,
        {40215, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17585,
        {40216, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17586,
        {172079, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17587,
        {19652, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17588,
        {172058, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17589,
        {40242, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17590,
        {19665, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17591,
        {40258, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17592,
        {40266, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17593,
        {40287, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17594,
        {40290, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17595,
        {172281, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17596,
        {40297, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17597,
        {40299, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17598,
        {172162, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17599,
        {40307, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17600,
        {40310, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7690},
        {40201}
    },
    {
        17601,
        {40311, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17602,
        {40318, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7721},
        {40356}
    },
    {
        17603,
        {40324, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17604,
        {40333, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17605,
        {40345, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17606,
        {40353, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17607,
        {40383, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17608,
        {40373, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2906},
        {38619}
    },
    {
        17609,
        {40377, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17610,
        {40381, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17611,
        {40387, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17612,
        {40391, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17613,
        {40393, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17614,
        {40406, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17615,
        {40410, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17616,
        {40415, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17617,
        {40416, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17618,
        {40419, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17619,
        {40436, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17620,
        {19719, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7702},
        {40210}
    },
    {
        17621,
        {40458, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17622,
        {40450, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17623,
        {40461, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17624,
        {40473, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17625,
        {40476, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17626,
        {40477, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17627,
        {40571, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17628,
        {139800, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {2552},
        {40575}
    },
    {
        17629,
        {40576, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17630,
        {40581, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17631,
        {40603, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17632,
        {40616, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17633,
        {172940, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17634,
        {40637, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {5068},
        {40636}
    },
    {
        17635,
        {173111, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17636,
        {40671, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17637,
        {40679, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17638,
        {40686, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17639,
        {40703, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17640,
        {40706, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17641,
        {19831, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1722},
        {34521}
    },
    {
        17642,
        {40707, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1722},
        {34521}
    },
    {
        17643,
        {40727, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17644,
        {40729, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17645,
        {40751, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17646,
        {40759, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17647,
        {40762, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17648,
        {40765, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17649,
        {40769, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17650,
        {40773, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {4743},
        {21957}
    },
    {
        17651,
        {40774, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17652,
        {40787, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7787},
        {40788}
    },
    {
        17653,
        {40789, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17654,
        {40792, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17655,
        {173553, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7789},
        {40799}
    },
    {
        17656,
        {40797, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17657,
        {173570, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7793},
        {40807}
    },
    {
        17658,
        {40809, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {1781},
        {22107}
    },
    {
        17659,
        {173594, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        1,
        {7794},
        {40812}
    },
    {
        17660,
        {40813, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17661,
        {40816, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17662,
        {40821, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    },
    {
        17663,
        {173746, -1},
        {-1, -1},
        JISCharacterClass_KANJI_LEVEL_4,
        0,
        {0},
        {0}
    }
};


static const uint16_t jis_urange_0000a0_000451[946] = {
    753,
    754,
    -1,
    -1,
    755,
    -1,
    756,
    87,
    14,
    757,
    758,
    759,
    -1,
    760,
    761,
    762,
    74,
    61,
    763,
    764,
    12,
    -1,
    182,
    765,
    766,
    767,
    768,
    769,
    770,
    771,
    772,
    773,
    774,
    775,
    776,
    777,
    778,
    779,
    780,
    781,
    782,
    783,
    784,
    785,
    786,
    787,
    788,
    789,
    790,
    791,
    792,
    793,
    794,
    795,
    796,
    62,
    797,
    798,
    799,
    800,
    801,
    802,
    803,
    804,
    805,
    806,
    807,
    808,
    809,
    810,
    811,
    812,
    813,
    814,
    815,
    816,
    817,
    818,
    819,
    820,
    821,
    822,
    823,
    824,
    825,
    826,
    827,
    63,
    828,
    829,
    830,
    831,
    832,
    833,
    834,
    835,
    836,
    841,
    871,
    886,
    846,
    857,
    873,
    888,
    902,
    908,
    -1,
    -1,
    874,
    889,
    877,
    892,
    -1,
    893,
    839,
    844,
    -1,
    -1,
    -1,
    -1,
    875,
    890,
    876,
    891,
    903,
    909,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    904,
    910,
    -1,
    938,
    -1,
    -1,
    837,
    842,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    905,
    911,
    -1,
    -1,
    -1,
    872,
    887,
    -1,
    -1,
    849,
    860,
    -1,
    -1,
    848,
    859,
    878,
    894,
    -1,
    -1,
    879,
    895,
    -1,
    -1,
    935,
    840,
    845,
    -1,
    -1,
    880,
    896,
    950,
    949,
    870,
    885,
    -1,
    -1,
    881,
    897,
    850,
    861,
    906,
    912,
    852,
    864,
    851,
    863,
    884,
    900,
    853,
    865,
    -1,
    -1,
    -1,
    -1,
    838,
    843,
    907,
    913,
    882,
    898,
    883,
    899,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    854,
    866,
    856,
    869,
    855,
    868,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    948,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    943,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    736,
    737,
    -1,
    738,
    743,
    744,
    -1,
    745,
    -1,
    746,
    -1,
    747,
    -1,
    748,
    -1,
    749,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    741,
    742,
    -1,
    -1,
    -1,
    976,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    958,
    964,
    965,
    944,
    963,
    970,
    923,
    945,
    953,
    955,
    974,
    -1,
    956,
    -1,
    957,
    930,
    947,
    934,
    -1,
    -1,
    961,
    967,
    941,
    973,
    951,
    -1,
    -1,
    -1,
    919,
    929,
    920,
    959,
    936,
    914,
    931,
    924,
    -1,
    954,
    -1,
    -1,
    -1,
    921,
    972,
    928,
    -1,
    925,
    916,
    -1,
    -1,
    937,
    926,
    917,
    946,
    -1,
    -1,
    -1,
    922,
    952,
    960,
    915,
    962,
    966,
    933,
    -1,
    927,
    971,
    918,
    -1,
    940,
    939,
    -1,
    -1,
    942,
    -1,
    -1,
    -1,
    -1,
    932,
    -1,
    -1,
    -1,
    969,
    968,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    862,
    990,
    -1,
    -1,
    -1,
    991,
    -1,
    -1,
    -1,
    992,
    993,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    847,
    901,
    -1,
    858,
    -1,
    867,
    1020,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1003,
    1004,
    1005,
    1006,
    1007,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    999,
    997,
    1002,
    1032,
    998,
    -1,
    994,
    -1,
    1016,
    -1,
    -1,
    996,
    1001,
    -1,
    -1,
    1000,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1027,
    1028,
    1033,
    -1,
    1013,
    1025,
    1026,
    1014,
    1015,
    -1,
    -1,
    -1,
    1021,
    1010,
    -1,
    -1,
    -1,
    1018,
    1029,
    -1,
    1011,
    -1,
    -1,
    1019,
    1022,
    -1,
    -1,
    -1,
    1024,
    -1,
    -1,
    -1,
    -1,
    1012,
    1030,
    1031,
    1023,
    1017,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    989,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    470,
    471,
    472,
    473,
    474,
    475,
    476,
    477,
    478,
    479,
    480,
    481,
    482,
    483,
    484,
    485,
    486,
    -1,
    487,
    488,
    489,
    490,
    491,
    492,
    493,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    502,
    503,
    504,
    505,
    506,
    507,
    508,
    509,
    510,
    511,
    512,
    513,
    514,
    515,
    516,
    517,
    518,
    526,
    519,
    520,
    521,
    522,
    523,
    524,
    525,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    570,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    564,
    565,
    566,
    567,
    568,
    569,
    571,
    572,
    573,
    574,
    575,
    576,
    577,
    578,
    579,
    580,
    581,
    582,
    583,
    584,
    585,
    586,
    587,
    588,
    589,
    590,
    591,
    592,
    593,
    594,
    595,
    596,
    612,
    613,
    614,
    615,
    616,
    617,
    619,
    620,
    621,
    622,
    623,
    624,
    625,
    626,
    627,
    628,
    629,
    630,
    631,
    632,
    633,
    634,
    635,
    636,
    637,
    638,
    639,
    640,
    641,
    642,
    643,
    644,
    -1,
    618
};
static const uint16_t jis_urange_001e3e_001e3f[2] = {
    739,
    740
};
static const uint16_t jis_urange_001f70_00277f[2064] = {
    977,
    978,
    987,
    988,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    29,
    -1,
    -1,
    279,
    -1,
    28,
    -1,
    -1,
    37,
    38,
    -1,
    -1,
    39,
    40,
    -1,
    -1,
    180,
    181,
    219,
    -1,
    -1,
    36,
    35,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    176,
    -1,
    75,
    76,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    101,
    732,
    -1,
    -1,
    995,
    -1,
    -1,
    1127,
    -1,
    -1,
    -1,
    -1,
    733,
    734,
    735,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1126,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    752,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    77,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    248,
    -1,
    -1,
    -1,
    250,
    -1,
    -1,
    1193,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1195,
    -1,
    -1,
    -1,
    -1,
    -1,
    251,
    -1,
    -1,
    -1,
    175,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    247,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    651,
    652,
    653,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1148,
    1149,
    1150,
    1151,
    1152,
    1153,
    1154,
    1155,
    1156,
    1157,
    1158,
    1182,
    -1,
    -1,
    -1,
    -1,
    1054,
    1055,
    1056,
    1057,
    1058,
    1059,
    1060,
    1061,
    1062,
    1063,
    1064,
    1065,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    104,
    105,
    103,
    106,
    174,
    -1,
    194,
    192,
    193,
    195,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    196,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    138,
    -1,
    139,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    198,
    199,
    197,
    200,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    140,
    -1,
    156,
    141,
    -1,
    132,
    -1,
    157,
    119,
    131,
    -1,
    120,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    246,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    162,
    -1,
    -1,
    164,
    70,
    1215,
    153,
    -1,
    -1,
    -1,
    -1,
    33,
    146,
    135,
    136,
    126,
    125,
    166,
    167,
    -1,
    1210,
    -1,
    -1,
    -1,
    -1,
    -1,
    71,
    165,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    163,
    -1,
    -1,
    -1,
    -1,
    -1,
    169,
    -1,
    170,
    -1,
    -1,
    171,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    159,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    65,
    158,
    168,
    -1,
    -1,
    -1,
    68,
    69,
    -1,
    -1,
    160,
    161,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    172,
    173,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    123,
    124,
    127,
    128,
    121,
    122,
    -1,
    -1,
    129,
    130,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    142,
    143,
    144,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    154,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1216,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    649,
    650,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    133,
    134,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    155,
    -1,
    -1,
    -1,
    -1,
    -1,
    655,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    597,
    598,
    599,
    600,
    601,
    602,
    603,
    604,
    605,
    606,
    607,
    608,
    609,
    610,
    611,
    -1,
    657,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    656,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1128,
    1129,
    1130,
    1131,
    1132,
    1133,
    1134,
    1135,
    1136,
    1137,
    1138,
    1139,
    1140,
    1141,
    1142,
    1143,
    1144,
    1145,
    1146,
    1147,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1066,
    1067,
    1068,
    1069,
    1070,
    1071,
    1072,
    1073,
    1074,
    1075,
    1076,
    1077,
    1078,
    1079,
    1080,
    1081,
    1082,
    1083,
    1084,
    1085,
    1086,
    1087,
    1088,
    1089,
    1090,
    1091,
    -1,
    1044,
    1045,
    1046,
    1047,
    1048,
    1049,
    1050,
    1051,
    1052,
    1053,
    527,
    528,
    529,
    530,
    531,
    532,
    533,
    534,
    535,
    536,
    -1,
    658,
    669,
    659,
    670,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    660,
    -1,
    -1,
    671,
    661,
    -1,
    -1,
    672,
    663,
    -1,
    -1,
    674,
    662,
    -1,
    -1,
    673,
    664,
    685,
    -1,
    -1,
    680,
    -1,
    -1,
    675,
    666,
    687,
    -1,
    -1,
    682,
    -1,
    -1,
    677,
    665,
    -1,
    -1,
    681,
    686,
    -1,
    -1,
    676,
    667,
    -1,
    -1,
    683,
    688,
    -1,
    -1,
    678,
    668,
    -1,
    -1,
    684,
    -1,
    -1,
    689,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    679,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    96,
    95,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    546,
    98,
    97,
    -1,
    -1,
    189,
    188,
    -1,
    -1,
    -1,
    -1,
    100,
    99,
    -1,
    -1,
    191,
    190,
    -1,
    -1,
    -1,
    -1,
    94,
    93,
    -1,
    214,
    -1,
    90,
    -1,
    -1,
    92,
    91,
    728,
    729,
    730,
    731,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    218,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    187,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    541,
    542,
    543,
    544,
    -1,
    89,
    88,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    540,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    537,
    538,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1221,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    73,
    -1,
    72,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    495,
    498,
    496,
    501,
    494,
    499,
    497,
    500,
    545,
    186,
    179,
    184,
    185,
    178,
    183,
    177,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    654,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1220,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1034,
    1035,
    1036,
    1037,
    1038,
    1039,
    1040,
    1041,
    1042,
    1043
};
static const uint16_t jis_urange_002934_0029fb[200] = {
    201,
    202,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    213,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    280,
    281
};
static const uint16_t jis_urange_003000_009fa2[28579] = {
    0,
    1,
    2,
    22,
    -1,
    24,
    25,
    26,
    49,
    50,
    51,
    52,
    53,
    54,
    55,
    56,
    57,
    58,
    102,
    107,
    43,
    44,
    151,
    152,
    149,
    150,
    -1,
    -1,
    -1,
    1191,
    -1,
    1192,
    539,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    112,
    113,
    114,
    -1,
    -1,
    -1,
    -1,
    -1,
    115,
    116,
    215,
    -1,
    -1,
    -1,
    282,
    283,
    284,
    285,
    286,
    287,
    288,
    289,
    290,
    291,
    292,
    293,
    294,
    295,
    296,
    297,
    298,
    299,
    300,
    301,
    302,
    303,
    304,
    305,
    306,
    307,
    308,
    309,
    310,
    311,
    312,
    313,
    314,
    315,
    316,
    317,
    318,
    319,
    320,
    321,
    322,
    323,
    324,
    325,
    326,
    327,
    328,
    329,
    330,
    331,
    332,
    333,
    334,
    335,
    336,
    337,
    338,
    339,
    340,
    341,
    342,
    343,
    344,
    345,
    346,
    347,
    348,
    349,
    350,
    351,
    352,
    353,
    354,
    355,
    356,
    357,
    358,
    359,
    360,
    361,
    362,
    363,
    364,
    365,
    366,
    367,
    -1,
    -1,
    -1,
    -1,
    10,
    11,
    20,
    21,
    118,
    278,
    376,
    377,
    378,
    379,
    380,
    381,
    382,
    383,
    384,
    385,
    386,
    387,
    388,
    389,
    390,
    391,
    392,
    393,
    394,
    395,
    396,
    397,
    398,
    399,
    400,
    401,
    402,
    403,
    404,
    405,
    406,
    407,
    408,
    409,
    410,
    411,
    412,
    413,
    414,
    415,
    416,
    417,
    418,
    419,
    420,
    421,
    422,
    423,
    424,
    425,
    426,
    427,
    428,
    429,
    430,
    431,
    432,
    433,
    434,
    435,
    436,
    437,
    438,
    439,
    440,
    441,
    442,
    443,
    444,
    445,
    446,
    447,
    448,
    449,
    450,
    451,
    452,
    453,
    454,
    455,
    456,
    457,
    458,
    459,
    460,
    461,
    645,
    646,
    647,
    648,
    5,
    27,
    18,
    19,
    117,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    547,
    548,
    549,
    550,
    551,
    552,
    553,
    554,
    555,
    556,
    558,
    559,
    560,
    561,
    562,
    563,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1201,
    1202,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1203,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    690,
    691,
    692,
    693,
    694,
    695,
    696,
    697,
    698,
    699,
    700,
    701,
    702,
    703,
    704,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1196,
    1197,
    1198,
    1199,
    1200,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    705,
    706,
    707,
    708,
    709,
    710,
    711,
    712,
    713,
    714,
    715,
    716,
    717,
    718,
    719,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1092,
    1093,
    1094,
    1095,
    1096,
    1097,
    1098,
    1099,
    1100,
    1101,
    1102,
    1103,
    1104,
    1105,
    1106,
    1107,
    1108,
    1109,
    1110,
    1111,
    -1,
    1114,
    -1,
    -1,
    -1,
    1113,
    -1,
    -1,
    1116,
    1115,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1112,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1165,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1169,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1160,
    -1,
    -1,
    -1,
    1163,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1161,
    1171,
    -1,
    -1,
    1170,
    1164,
    -1,
    -1,
    -1,
    1172,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1166,
    -1,
    -1,
    -1,
    -1,
    1174,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1159,
    1173,
    -1,
    -1,
    1162,
    -1,
    -1,
    -1,
    1167,
    -1,
    -1,
    -1,
    -1,
    -1,
    1168,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1190,
    1206,
    1205,
    1204,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1178,
    1179,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1175,
    1176,
    1177,
    -1,
    -1,
    1181,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1180,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    249,
    -1,
    1194,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1224,
    -1,
    -1,
    -1,
    8848,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8853,
    -1,
    8854,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8897,
    -1,
    8889,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8929,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1272,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9034,
    -1,
    -1,
    -1,
    -1,
    16709,
    -1,
    -1,
    -1,
    -1,
    -1,
    9038,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1280,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9063,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9084,
    9085,
    -1,
    -1,
    -1,
    -1,
    9088,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9094,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9120,
    -1,
    9123,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9132,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9141,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9151,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9159,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9206,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9242,
    -1,
    -1,
    -1,
    -1,
    9246,
    -1,
    -1,
    9244,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9266,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9496,
    9498,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9502,
    9501,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9505,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    4386,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9523,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    4402,
    -1,
    -1,
    -1,
    -1,
    -1,
    9547,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9550,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9553,
    -1,
    -1,
    9555,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9572,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9574,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9577,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9582,
    -1,
    -1,
    -1,
    -1,
    9584,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9896,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9907,
    -1,
    -1,
    9911,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9933,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    10022,
    -1,
    -1,
    -1,
    -1,
    10026,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    10044,
    10079,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    10055,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    10071,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    10075,
    -1,
    10077,
    -1,
    -1,
    -1,
    -1,
    -1,
    7936,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    10134,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    10126,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    10159,
    10160,
    -1,
    -1,
    -1,
    -1,
    10163,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    10171,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    7989,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    7990,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    10191,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    10212,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8015,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    10238,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16100,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16107,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16154,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16167,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16199,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16219,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16247,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16254,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16269,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8153,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16292,
    -1,
    -1,
    -1,
    -1,
    -1,
    16294,
    -1,
    16295,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16302,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16309,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16375,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8222,
    -1,
    -1,
    16393,
    -1,
    16395,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16415,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16426,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16431,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16451,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16462,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16474,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16517,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16533,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16536,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16542,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16562,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16566,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16582,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16586,
    -1,
    -1,
    -1,
    -1,
    16583,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16594,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16598,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8348,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16648,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16659,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16664,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16683,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16698,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16702,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16761,
    -1,
    16765,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16771,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16773,
    -1,
    -1,
    -1,
    -1,
    16775,
    -1,
    -1,
    -1,
    -1,
    -1,
    16778,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8422,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8423,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16807,
    -1,
    -1,
    -1,
    16809,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16818,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16846,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16843,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16845,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16882,
    -1,
    -1,
    -1,
    -1,
    16877,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16902,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16903,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8508,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16966,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16995,
    -1,
    -1,
    -1,
    -1,
    8523,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17017,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17039,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17048,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17059,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17064,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17088,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17116,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17138,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17148,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8596,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17167,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17208,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8703,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17369,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17375,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17380,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17383,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17412,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17442,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17489,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17491,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17493,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17500,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8771,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17514,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17587,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17590,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17620,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17641,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1485,
    3285,
    8837,
    2560,
    -1,
    -1,
    -1,
    4039,
    2795,
    2459,
    2794,
    1625,
    -1,
    3811,
    4166,
    8838,
    4419,
    1518,
    8839,
    -1,
    1769,
    4420,
    2917,
    4639,
    1933,
    3879,
    -1,
    -1,
    -1,
    -1,
    2796,
    -1,
    -1,
    4259,
    -1,
    -1,
    -1,
    -1,
    3887,
    -1,
    1225,
    8840,
    4421,
    8841,
    16275,
    3265,
    8842,
    1226,
    1227,
    4422,
    2063,
    -1,
    -1,
    -1,
    4423,
    -1,
    1846,
    3211,
    -1,
    2608,
    4424,
    -1,
    -1,
    4425,
    8843,
    -1,
    4426,
    3592,
    -1,
    1934,
    -1,
    8844,
    8845,
    -1,
    -1,
    3594,
    -1,
    3544,
    2196,
    3966,
    -1,
    8847,
    -1,
    -1,
    -1,
    6808,
    4427,
    2797,
    4428,
    1618,
    8850,
    -1,
    -1,
    2042,
    2241,
    4114,
    -1,
    -1,
    4987,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8851,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    4215,
    -1,
    3570,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1790,
    -1,
    1900,
    -1,
    4429,
    -1,
    -1,
    4430,
    4256,
    -1,
    4164,
    3079,
    4432,
    2529,
    3560,
    1228,
    4435,
    -1,
    -1,
    1533,
    2224,
    -1,
    2223,
    1479,
    -1,
    -1,
    4366,
    4365,
    -1,
    2368,
    1410,
    8852,
    4436,
    4437,
    4438,
    3967,
    4439,
    -1,
    2243,
    1480,
    4029,
    -1,
    1971,
    -1,
    -1,
    1972,
    1973,
    3355,
    4257,
    -1,
    4440,
    -1,
    -1,
    4441,
    -1,
    -1,
    4442,
    -1,
    -1,
    8855,
    2864,
    8856,
    8858,
    -1,
    -1,
    -1,
    2657,
    2865,
    4447,
    8859,
    4445,
    -1,
    4446,
    1935,
    8860,
    -1,
    2352,
    1677,
    -1,
    4444,
    4443,
    3862,
    8861,
    -1,
    -1,
    -1,
    2481,
    2480,
    3131,
    4448,
    3812,
    2984,
    8863,
    -1,
    -1,
    23,
    4449,
    4451,
    -1,
    1229,
    -1,
    3170,
    4294,
    1451,
    -1,
    -1,
    -1,
    -1,
    -1,
    8862,
    -1,
    4450,
    1627,
    -1,
    2009,
    8864,
    3266,
    -1,
    -1,
    8865,
    2146,
    4452,
    -1,
    -1,
    -1,
    3575,
    -1,
    1230,
    -1,
    1231,
    8866,
    1860,
    -1,
    1232,
    -1,
    -1,
    -1,
    -1,
    -1,
    4453,
    1452,
    1233,
    -1,
    2225,
    1861,
    3849,
    3683,
    1936,
    -1,
    -1,
    -1,
    -1,
    8867,
    -1,
    -1,
    -1,
    1678,
    -1,
    4488,
    3416,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    3647,
    4455,
    -1,
    -1,
    -1,
    3693,
    -1,
    4295,
    8869,
    2832,
    -1,
    2482,
    -1,
    2530,
    1629,
    8870,
    -1,
    -1,
    -1,
    -1,
    3338,
    -1,
    -1,
    3197,
    4459,
    1235,
    1236,
    -1,
    -1,
    -1,
    1453,
    3356,
    2658,
    2369,
    4139,
    -1,
    3147,
    8871,
    1628,
    1237,
    4458,
    8872,
    4165,
    4454,
    4456,
    2427,
    4457,
    4891,
    1238,
    1234,
    -1,
    -1,
    -1,
    8868,
    -1,
    -1,
    -1,
    -1,
    4465,
    1239,
    -1,
    1240,
    -1,
    -1,
    4468,
    4466,
    -1,
    -1,
    1631,
    -1,
    3880,
    4460,
    8874,
    8875,
    -1,
    8876,
    4464,
    2244,
    8877,
    1241,
    2483,
    -1,
    -1,
    8878,
    1791,
    -1,
    8879,
    4469,
    -1,
    4461,
    -1,
    1242,
    4296,
    -1,
    2531,
    -1,
    4462,
    -1,
    4467,
    8880,
    -1,
    1243,
    -1,
    4470,
    1244,
    4463,
    -1,
    8881,
    1974,
    -1,
    1454,
    -1,
    -1,
    1975,
    1630,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    4892,
    -1,
    4036,
    3837,
    2245,
    -1,
    -1,
    8883,
    -1,
    -1,
    2834,
    4252,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8884,
    3909,
    -1,
    -1,
    2091,
    3106,
    1665,
    8885,
    -1,
    -1,
    -1,
    1246,
    2682,
    8886,
    -1,
    -1,
    4474,
    8887,
    4479,
    4477,
    8888,
    -1,
    4472,
    -1,
    -1,
    3116,
    4475,
    -1,
    4478,
    4476,
    -1,
    3914,
    -1,
    4473,
    1247,
    2833,
    -1,
    4030,
    4480,
    4481,
    8882,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    2635,
    4494,
    -1,
    1222,
    8890,
    3619,
    -1,
    3781,
    4489,
    -1,
    3933,
    -1,
    1619,
    -1,
    -1,
    -1,
    4493,
    -1,
    8891,
    1248,
    1249,
    -1,
    -1,
    4487,
    4496,
    -1,
    -1,
    3057,
    -1,
    2197,
    -1,
    3631,
    1250,
    5960,
    8892,
    4495,
    3444,
    8893,
    4484,
    -1,
    2247,
    -1,
    1251,
    2246,
    4482,
    -1,
    8894,
    -1,
    8895,
    2595,
    -1,
    4490,
    8896,
    3932,
    3240,
    4486,
    2148,
    1252,
    4483,
    4491,
    4485,
    4278,
    4492,
    4356,
    1253,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    2043,
    -1,
    -1,
    2147,
    -1,
    1255,
    -1,
    -1,
    -1,
    -1,
    1254,
    1256,
    8898,
    4497,
    -1,
    -1,
    8899,
    4498,
    4502,
    1455,
    -1,
    -1,
    -1,
    -1,
    8900,
    3901,
    4501,
    -1,
    -1,
    8901,
    -1,
    4500,
    4504,
    8902,
    -1,
    -1,
    4503,
    -1,
    3357,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8903,
    -1,
    2149,
    8904,
    -1,
    -1,
    -1,
    8905,
    -1,
    4505,
    -1,
    -1,
    -1,
    8906,
    -1,
    2574,
    -1,
    3107,
    3358,
    2059,
    -1,
    4506,
    -1,
    -1,
    -1,
    -1,
    1901,
    -1,
    -1,
    4507,
    -1,
    -1,
    -1,
    -1,
    4509,
    -1,
    -1,
    8908,
    -1,
    -1,
    -1,
    -1,
    3968,
    -1,
    -1,
    -1,
    2137,
    8909,
    8910,
    1257,
    8911,
    8912,
    -1,
    2460,
    3752,
    4508,
    -1,
    8913,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8907,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8914,
    -1,
    2386,
    4170,
    -1,
    -1,
    -1,
    8916,
    4511,
    4514,
    4510,
    2385,
    -1,
    2724,
    -1,
    -1,
    8917,
    8918,
    -1,
    -1,
    2092,
    -1,
    -1,
    -1,
    4515,
    -1,
    8919,
    2020,
    -1,
    8920,
    -1,
    4512,
    4513,
    -1,
    1258,
    3492,
    8923,
    3099,
    1260,
    1976,
    -1,
    -1,
    8925,
    3992,
    4516,
    -1,
    -1,
    8926,
    4258,
    -1,
    -1,
    -1,
    4517,
    -1,
    -1,
    8927,
    -1,
    4520,
    -1,
    4518,
    1261,
    3053,
    -1,
    8928,
    -1,
    -1,
    -1,
    4519,
    4521,
    -1,
    -1,
    -1,
    1259,
    8921,
    -1,
    4523,
    -1,
    -1,
    -1,
    4522,
    -1,
    3893,
    -1,
    -1,
    -1,
    -1,
    1902,
    4525,
    4526,
    1264,
    1612,
    -1,
    1263,
    -1,
    9024,
    4524,
    -1,
    1265,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    2621,
    -1,
    4529,
    4528,
    4527,
    9026,
    4471,
    -1,
    4530,
    9027,
    -1,
    -1,
    1266,
    2725,
    -1,
    4531,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    4140,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    4098,
    -1,
    -1,
    1267,
    -1,
    4533,
    -1,
    -1,
    4532,
    4535,
    4534,
    -1,
    -1,
    4536,
    4537,
    1493,
    -1,
    2182,
    2094,
    2659,
    3286,
    1977,
    2985,
    2248,
    1268,
    2333,
    4539,
    4081,
    3421,
    -1,
    2532,
    -1,
    4538,
    -1,
    4540,
    1270,
    -1,
    1271,
    -1,
    -1,
    3445,
    -1,
    1776,
    -1,
    -1,
    -1,
    9029,
    -1,
    4541,
    -1,
    -1,
    3571,
    -1,
    -1,
    3027,
    4543,
    4544,
    3677,
    2249,
    4350,
    4545,
    -1,
    -1,
    1979,
    -1,
    9031,
    -1,
    3881,
    3123,
    2054,
    3404,
    -1,
    -1,
    16721,
    2150,
    -1,
    -1,
    -1,
    4546,
    -1,
    4547,
    9032,
    -1,
    3543,
    1566,
    -1,
    -1,
    4550,
    2442,
    9033,
    4549,
    2387,
    -1,
    4551,
    6563,
    4552,
    3980,
    4553,
    -1,
    4554,
    4555,
    2798,
    9035,
    2581,
    -1,
    -1,
    -1,
    1273,
    -1,
    -1,
    1792,
    -1,
    4558,
    9036,
    4556,
    4070,
    4557,
    -1,
    3817,
    4559,
    4560,
    4561,
    3446,
    9037,
    -1,
    -1,
    4565,
    4563,
    4564,
    4562,
    2416,
    4566,
    4115,
    4297,
    -1,
    -1,
    -1,
    -1,
    9039,
    4567,
    -1,
    -1,
    -1,
    -1,
    -1,
    1274,
    2921,
    4568,
    2689,
    -1,
    -1,
    4569,
    1275,
    3287,
    4260,
    3447,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    4642,
    -1,
    -1,
    -1,
    -1,
    4570,
    7806,
    2010,
    1276,
    -1,
    4571,
    4009,
    1277,
    -1,
    -1,
    -1,
    2703,
    3193,
    -1,
    4573,
    3545,
    -1,
    -1,
    4574,
    1278,
    -1,
    4575,
    1705,
    -1,
    9042,
    9043,
    4576,
    1980,
    -1,
    3521,
    1593,
    2679,
    -1,
    -1,
    3666,
    4577,
    -1,
    3448,
    1279,
    9044,
    2866,
    4578,
    -1,
    3865,
    2971,
    1787,
    -1,
    1794,
    4579,
    -1,
    -1,
    4581,
    -1,
    -1,
    2093,
    9045,
    1281,
    4580,
    1282,
    9046,
    4312,
    -1,
    -1,
    -1,
    -1,
    -1,
    2704,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    3694,
    3897,
    -1,
    4582,
    -1,
    4223,
    4583,
    -1,
    -1,
    -1,
    4584,
    -1,
    3477,
    -1,
    -1,
    4585,
    -1,
    -1,
    2922,
    2443,
    2151,
    4586,
    2484,
    2334,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    3359,
    4588,
    -1,
    -1,
    3108,
    -1,
    1283,
    2428,
    4589,
    4590,
    3023,
    -1,
    4587,
    -1,
    -1,
    -1,
    -1,
    4592,
    9048,
    3969,
    1284,
    -1,
    -1,
    -1,
    2323,
    9049,
    1409,
    4591,
    -1,
    -1,
    1285,
    -1,
    2152,
    2411,
    3648,
    -1,
    -1,
    -1,
    4595,
    4593,
    -1,
    9050,
    -1,
    -1,
    3850,
    2799,
    4602,
    1759,
    4596,
    4594,
    3054,
    -1,
    9051,
    -1,
    -1,
    -1,
    -1,
    -1,
    4598,
    -1,
    4597,
    -1,
    -1,
    9053,
    1727,
    9052,
    -1,
    -1,
    2131,
    4603,
    4242,
    -1,
    -1,
    -1,
    4599,
    -1,
    -1,
    -1,
    4604,
    4601,
    1286,
    4600,
    -1,
    -1,
    -1,
    9055,
    -1,
    -1,
    4276,
    -1,
    -1,
    -1,
    2250,
    1632,
    -1,
    -1,
    4313,
    9057,
    -1,
    9058,
    -1,
    -1,
    2716,
    3439,
    2324,
    4607,
    4608,
    -1,
    9059,
    -1,
    4298,
    -1,
    -1,
    4334,
    4610,
    -1,
    -1,
    -1,
    2251,
    9060,
    9061,
    4609,
    -1,
    1706,
    -1,
    -1,
    4611,
    -1,
    4001,
    -1,
    3317,
    -1,
    4141,
    1287,
    3910,
    9062,
    -1,
    1289,
    4612,
    -1,
    -1,
    1290,
    9064,
    7478,
    -1,
    -1,
    3493,
    1291,
    4613,
    1795,
    4060,
    -1,
    1292,
    -1,
    2726,
    4614,
    3924,
    4618,
    -1,
    2923,
    4615,
    2021,
    -1,
    4616,
    1796,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1294,
    -1,
    2081,
    4619,
    -1,
    4620,
    -1,
    9066,
    4621,
    4622,
    2596,
    1295,
    -1,
    -1,
    2252,
    4103,
    1296,
    4113,
    3564,
    -1,
    -1,
    3934,
    4623,
    1297,
    4624,
    -1,
    9067,
    9068,
    -1,
    4626,
    -1,
    4628,
    4627,
    -1,
    -1,
    -1,
    -1,
    4629,
    1626,
    3991,
    -1,
    2441,
    4630,
    -1,
    1298,
    3064,
    -1,
    -1,
    2727,
    1982,
    -1,
    4631,
    9069,
    -1,
    -1,
    -1,
    -1,
    -1,
    3724,
    -1,
    -1,
    -1,
    -1,
    4632,
    -1,
    4633,
    -1,
    4634,
    -1,
    9070,
    -1,
    -1,
    4635,
    3763,
    2045,
    1478,
    -1,
    -1,
    9071,
    3508,
    4636,
    2660,
    9072,
    2986,
    -1,
    4638,
    4637,
    2728,
    2226,
    4640,
    3695,
    -1,
    -1,
    4641,
    -1,
    -1,
    -1,
    3725,
    3121,
    3179,
    1981,
    -1,
    -1,
    3555,
    3212,
    -1,
    3649,
    -1,
    3993,
    -1,
    4643,
    -1,
    2987,
    1300,
    -1,
    1301,
    -1,
    -1,
    2088,
    9075,
    -1,
    4644,
    -1,
    -1,
    9076,
    -1,
    4645,
    1515,
    1494,
    1862,
    -1,
    3109,
    1928,
    4216,
    -1,
    4648,
    1620,
    -1,
    9077,
    4647,
    -1,
    1302,
    -1,
    1983,
    -1,
    -1,
    4649,
    -1,
    4122,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1303,
    -1,
    -1,
    4650,
    -1,
    4279,
    -1,
    2253,
    -1,
    -1,
    1304,
    -1,
    2183,
    4651,
    -1,
    -1,
    -1,
    9078,
    4653,
    4652,
    -1,
    2882,
    1528,
    -1,
    -1,
    -1,
    1565,
    4654,
    -1,
    4655,
    -1,
    1305,
    2184,
    9079,
    -1,
    4656,
    9081,
    -1,
    -1,
    -1,
    1957,
    -1,
    -1,
    -1,
    -1,
    9082,
    -1,
    2461,
    4657,
    -1,
    -1,
    -1,
    -1,
    4031,
    2370,
    1937,
    4142,
    3055,
    3696,
    2630,
    -1,
    -1,
    -1,
    -1,
    -1,
    2670,
    9086,
    2609,
    2622,
    -1,
    2717,
    9087,
    3697,
    -1,
    -1,
    -1,
    4660,
    -1,
    1538,
    3056,
    2254,
    2198,
    2044,
    -1,
    -1,
    4664,
    3196,
    3195,
    1984,
    2729,
    4665,
    4663,
    1633,
    3171,
    2561,
    2486,
    1509,
    9089,
    9090,
    1771,
    2325,
    2485,
    -1,
    4666,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    4667,
    -1,
    1920,
    1729,
    -1,
    -1,
    -1,
    2326,
    1919,
    3352,
    1508,
    3494,
    4071,
    2256,
    4224,
    3422,
    2255,
    1306,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    2082,
    -1,
    4676,
    4417,
    2040,
    3989,
    -1,
    -1,
    -1,
    9092,
    -1,
    3726,
    1307,
    9093,
    4675,
    -1,
    1847,
    4670,
    4671,
    4673,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    4674,
    -1,
    1938,
    2884,
    -1,
    3866,
    4672,
    4668,
    2228,
    -1,
    4669,
    -1,
    4327,
    9095,
    -1,
    -1,
    3935,
    -1,
    3361,
    2227,
    2335,
    -1,
    -1,
    1308,
    4677,
    -1,
    -1,
    3538,
    -1,
    -1,
    -1,
    9091,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    4681,
    -1,
    -1,
    9096,
    -1,
    -1,
    -1,
    9097,
    -1,
    2631,
    -1,
    2623,
    1310,
    9098,
    -1,
    -1,
    -1,
    4684,
    4682,
    -1,
    4044,
    1311,
    4679,
    4688,
    4683,
    -1,
    -1,
    -1,
    4686,
    2199,
    4072,
    -1,
    1312,
    4687,
    -1,
    -1,
    -1,
    4689,
    -1,
    4691,
    -1,
    1313,
    -1,
    9099,
    2429,
    4357,
    9100,
    4680,
    4678,
    4690,
    -1,
    4685,
    -1,
    -1,
    9101,
    1314,
    -1,
    -1,
    -1,
    -1,
    -1,
    1309,
    -1,
    -1,
    -1,
    9102,
    1315,
    4693,
    -1,
    4702,
    4695,
    9103,
    -1,
    4699,
    1316,
    -1,
    4700,
    4696,
    9104,
    9105,
    4729,
    -1,
    -1,
    2422,
    1708,
    -1,
    -1,
    -1,
    9106,
    4694,
    -1,
    9107,
    -1,
    4704,
    1495,
    4703,
    9108,
    1414,
    3801,
    4701,
    9109,
    4697,
    -1,
    1317,
    4692,
    4698,
    2389,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    4705,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1496,
    4714,
    -1,
    -1,
    4706,
    4707,
    -1,
    2730,
    4020,
    -1,
    -1,
    9111,
    4712,
    4711,
    9112,
    -1,
    9113,
    3398,
    9114,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    4713,
    -1,
    -1,
    4710,
    -1,
    1318,
    9115,
    9116,
    -1,
    -1,
    1523,
    -1,
    2371,
    2835,
    -1,
    9117,
    -1,
    -1,
    -1,
    -1,
    1319,
    4708,
    3449,
    -1,
    -1,
    -1,
    4709,
    -1,
    1411,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1320,
    -1,
    -1,
    4719,
    4138,
    -1,
    2732,
    -1,
    4725,
    -1,
    1321,
    -1,
    -1,
    4724,
    4715,
    -1,
    -1,
    9118,
    -1,
    3136,
    -1,
    4716,
    9119,
    -1,
    -1,
    3180,
    4721,
    2731,
    9121,
    -1,
    -1,
    9122,
    -1,
    4718,
    -1,
    -1,
    4109,
    1322,
    -1,
    -1,
    2095,
    -1,
    -1,
    4722,
    4723,
    -1,
    -1,
    -1,
    -1,
    4720,
    4726,
    1323,
    -1,
    9124,
    9125,
    -1,
    4717,
    9126,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    4732,
    4737,
    9128,
    4733,
    -1,
    4728,
    1324,
    9129,
    4738,
    3024,
    -1,
    1325,
    4740,
    9130,
    2257,
    4730,
    3288,
    -1,
    -1,
    1326,
    -1,
    -1,
    9131,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    4734,
    4727,
    1798,
    -1,
    1863,
    1760,
    4735,
    4731,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    2153,
    4741,
    4739,
    3058,
    1921,
    1985,
    1328,
    4736,
    -1,
    2057,
    -1,
    -1,
    -1,
    -1,
    -1,
    1539,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9136,
    -1,
    -1,
    -1,
    -1,
    4745,
    4743,
    -1,
    4800,
    -1,
    9137,
    -1,
    -1,
    9138,
    -1,
    1329,
    -1,
    -1,
    9139,
    9133,
    -1,
    4748,
    -1,
    -1,
    -1,
    -1,
    -1,
    4742,
    -1,
    4746,
    9140,
    -1,
    4744,
    -1,
    -1,
    9142,
    2487,
    4747,
    -1,
    -1,
    -1,
    -1,
    9144,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    4750,
    -1,
    4755,
    -1,
    -1,
    -1,
    4753,
    4752,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    3213,
    9147,
    1331,
    1634,
    -1,
    -1,
    -1,
    -1,
    1332,
    -1,
    9148,
    -1,
    -1,
    -1,
    4749,
    -1,
    4751,
    2733,
    1522,
    -1,
    -1,
    4754,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9145,
    1659,
    -1,
    -1,
    -1,
    -1,
    -1,
    4765,
    9149,
    2817,
    4761,
    -1,
    4759,
    -1,
    4760,
    9150,
    4762,
    -1,
    -1,
    1333,
    -1,
    9152,
    -1,
    9153,
    9154,
    -1,
    1532,
    -1,
    -1,
    -1,
    -1,
    9155,
    -1,
    1334,
    -1,
    -1,
    3032,
    -1,
    4756,
    -1,
    4757,
    -1,
    -1,
    7808,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1781,
    -1,
    -1,
    9156,
    -1,
    9157,
    -1,
    -1,
    -1,
    4764,
    -1,
    1336,
    -1,
    1864,
    -1,
    4767,
    4763,
    4766,
    9158,
    -1,
    1338,
    -1,
    1339,
    1340,
    -1,
    3867,
    -1,
    1335,
    -1,
    3530,
    -1,
    3689,
    -1,
    -1,
    -1,
    -1,
    -1,
    4769,
    -1,
    -1,
    -1,
    -1,
    -1,
    4768,
    1728,
    9160,
    -1,
    4770,
    -1,
    9161,
    -1,
    -1,
    4773,
    -1,
    -1,
    -1,
    -1,
    4772,
    9162,
    -1,
    -1,
    -1,
    1341,
    9163,
    -1,
    -1,
    9164,
    1342,
    -1,
    4771,
    -1,
    3596,
    -1,
    -1,
    4774,
    -1,
    -1,
    9165,
    1343,
    -1,
    -1,
    1344,
    9166,
    4775,
    -1,
    -1,
    -1,
    9167,
    1345,
    4777,
    -1,
    4776,
    -1,
    -1,
    -1,
    -1,
    -1,
    4779,
    -1,
    -1,
    -1,
    4782,
    4780,
    4778,
    4781,
    -1,
    9168,
    -1,
    -1,
    4783,
    1346,
    1347,
    -1,
    -1,
    9169,
    4784,
    -1,
    -1,
    4785,
    -1,
    4786,
    -1,
    -1,
    -1,
    4787,
    4548,
    -1,
    2629,
    2488,
    -1,
    -1,
    1680,
    9170,
    1497,
    -1,
    -1,
    3231,
    -1,
    -1,
    -1,
    -1,
    9171,
    -1,
    -1,
    -1,
    -1,
    -1,
    4788,
    -1,
    2353,
    -1,
    1456,
    2881,
    -1,
    -1,
    9172,
    9173,
    -1,
    4789,
    2200,
    -1,
    -1,
    2336,
    -1,
    4791,
    4790,
    -1,
    -1,
    3917,
    4792,
    -1,
    -1,
    -1,
    4794,
    4793,
    1348,
    4795,
    -1,
    4796,
    -1,
    2154,
    -1,
    -1,
    1567,
    4797,
    -1,
    9175,
    4799,
    -1,
    4798,
    -1,
    -1,
    -1,
    4801,
    -1,
    -1,
    3441,
    -1,
    1350,
    -1,
    9176,
    -1,
    -1,
    4802,
    1430,
    2412,
    9178,
    -1,
    -1,
    -1,
    2096,
    -1,
    1351,
    3242,
    -1,
    -1,
    1352,
    1353,
    -1,
    -1,
    4803,
    4804,
    -1,
    -1,
    4806,
    -1,
    -1,
    -1,
    -1,
    4807,
    -1,
    2417,
    -1,
    -1,
    9180,
    9181,
    2022,
    -1,
    -1,
    3970,
    -1,
    9182,
    9183,
    4805,
    4808,
    2382,
    2258,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    4812,
    -1,
    -1,
    2354,
    -1,
    3214,
    -1,
    9185,
    4809,
    3347,
    -1,
    -1,
    -1,
    -1,
    9186,
    1354,
    -1,
    -1,
    9187,
    9188,
    9189,
    -1,
    1355,
    -1,
    -1,
    -1,
    9190,
    1356,
    -1,
    -1,
    4813,
    -1,
    -1,
    2885,
    -1,
    -1,
    -1,
    -1,
    -1,
    4811,
    4814,
    -1,
    2098,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    4815,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9194,
    -1,
    1357,
    9195,
    9196,
    -1,
    4816,
    -1,
    2259,
    1723,
    4818,
    -1,
    -1,
    -1,
    9197,
    -1,
    4819,
    -1,
    9193,
    -1,
    -1,
    -1,
    4820,
    -1,
    -1,
    4817,
    -1,
    -1,
    -1,
    -1,
    1360,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    4810,
    -1,
    -1,
    4821,
    -1,
    -1,
    4822,
    1361,
    1362,
    -1,
    -1,
    4015,
    9200,
    -1,
    2800,
    1363,
    -1,
    -1,
    4824,
    4825,
    4823,
    -1,
    4827,
    9198,
    -1,
    -1,
    -1,
    -1,
    3595,
    -1,
    9203,
    1481,
    3813,
    -1,
    -1,
    4828,
    1364,
    -1,
    9204,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1365,
    -1,
    -1,
    9205,
    -1,
    -1,
    -1,
    2818,
    1366,
    1367,
    2562,
    9207,
    3632,
    1865,
    9208,
    2424,
    9209,
    -1,
    1368,
    4004,
    -1,
    3495,
    -1,
    9210,
    2155,
    3148,
    -1,
    -1,
    1369,
    4826,
    4829,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    3137,
    -1,
    -1,
    -1,
    4830,
    -1,
    -1,
    -1,
    4831,
    9211,
    -1,
    9212,
    4833,
    -1,
    -1,
    3362,
    -1,
    -1,
    9213,
    -1,
    -1,
    1799,
    -1,
    -1,
    -1,
    -1,
    7802,
    1568,
    3936,
    9214,
    -1,
    2801,
    3423,
    -1,
    -1,
    -1,
    9215,
    2419,
    -1,
    -1,
    4839,
    -1,
    -1,
    3882,
    4290,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9217,
    1681,
    4835,
    9218,
    -1,
    -1,
    -1,
    -1,
    3033,
    4838,
    -1,
    3450,
    -1,
    -1,
    3424,
    3451,
    3690,
    3334,
    -1,
    -1,
    -1,
    2390,
    -1,
    -1,
    1371,
    4834,
    -1,
    1372,
    -1,
    -1,
    9219,
    -1,
    1589,
    -1,
    3405,
    -1,
    -1,
    -1,
    -1,
    4836,
    -1,
    4832,
    -1,
    -1,
    2867,
    -1,
    -1,
    -1,
    4840,
    -1,
    -1,
    1374,
    -1,
    2677,
    -1,
    -1,
    -1,
    -1,
    1986,
    -1,
    4841,
    -1,
    -1,
    -1,
    1375,
    9220,
    9221,
    -1,
    9222,
    -1,
    9223,
    9224,
    -1,
    -1,
    3925,
    9225,
    -1,
    -1,
    3100,
    -1,
    -1,
    -1,
    -1,
    3327,
    9226,
    1376,
    4843,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    3994,
    1378,
    9227,
    4844,
    -1,
    -1,
    4849,
    -1,
    -1,
    9228,
    -1,
    3868,
    -1,
    -1,
    -1,
    -1,
    4848,
    4842,
    4845,
    4847,
    -1,
    -1,
    2355,
    -1,
    -1,
    3894,
    -1,
    9230,
    -1,
    4850,
    -1,
    3232,
    -1,
    -1,
    1682,
    -1,
    2802,
    9231,
    1381,
    -1,
    -1,
    4852,
    1380,
    4851,
    1382,
    2327,
    -1,
    4853,
    4855,
    4854,
    1383,
    -1,
    4857,
    -1,
    4846,
    4859,
    1384,
    -1,
    9232,
    -1,
    4858,
    4856,
    -1,
    -1,
    -1,
    1385,
    -1,
    2489,
    2868,
    -1,
    3059,
    4860,
    2941,
    1486,
    3639,
    9233,
    9234,
    -1,
    -1,
    3348,
    -1,
    4862,
    4861,
    4863,
    4864,
    4865,
    -1,
    -1,
    -1,
    -1,
    4866,
    -1,
    -1,
    9235,
    9236,
    -1,
    -1,
    3902,
    4867,
    9237,
    1386,
    9238,
    -1,
    1635,
    4868,
    -1,
    -1,
    -1,
    9239,
    4163,
    1707,
    -1,
    4646,
    2671,
    3132,
    4869,
    4116,
    -1,
    -1,
    -1,
    -1,
    -1,
    4061,
    -1,
    9240,
    4871,
    -1,
    3172,
    -1,
    3406,
    3133,
    3814,
    4872,
    4873,
    1594,
    -1,
    -1,
    2563,
    4874,
    -1,
    -1,
    -1,
    -1,
    1457,
    4875,
    -1,
    -1,
    -1,
    -1,
    9243,
    4876,
    -1,
    -1,
    -1,
    -1,
    -1,
    1569,
    -1,
    9245,
    1866,
    3541,
    3937,
    -1,
    -1,
    -1,
    -1,
    4880,
    3060,
    4879,
    2099,
    -1,
    -1,
    4006,
    4878,
    -1,
    3452,
    4882,
    -1,
    4881,
    9248,
    -1,
    1388,
    -1,
    9249,
    4884,
    -1,
    4883,
    -1,
    -1,
    1595,
    -1,
    4885,
    2734,
    4887,
    3200,
    -1,
    4886,
    1389,
    3872,
    -1,
    -1,
    -1,
    -1,
    2718,
    3442,
    9251,
    9252,
    -1,
    4888,
    -1,
    -1,
    -1,
    9253,
    2260,
    -1,
    -1,
    -1,
    4889,
    3572,
    3727,
    4090,
    -1,
    -1,
    -1,
    -1,
    -1,
    3576,
    1390,
    -1,
    4898,
    -1,
    -1,
    -1,
    -1,
    1391,
    1903,
    -1,
    -1,
    4172,
    -1,
    -1,
    4056,
    -1,
    4993,
    -1,
    4890,
    -1,
    9254,
    -1,
    -1,
    -1,
    4893,
    1392,
    3138,
    -1,
    -1,
    3971,
    -1,
    -1,
    -1,
    3425,
    -1,
    9255,
    -1,
    -1,
    -1,
    4894,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    4016,
    -1,
    2391,
    9256,
    -1,
    2735,
    -1,
    -1,
    -1,
    -1,
    1393,
    -1,
    -1,
    4895,
    -1,
    9257,
    2491,
    -1,
    2490,
    -1,
    9258,
    -1,
    -1,
    1434,
    2201,
    1394,
    2924,
    1458,
    -1,
    -1,
    -1,
    -1,
    4899,
    4900,
    -1,
    4897,
    1395,
    9259,
    -1,
    -1,
    -1,
    -1,
    9260,
    9261,
    1527,
    1800,
    9262,
    4896,
    -1,
    4078,
    3776,
    -1,
    -1,
    9263,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1417,
    -1,
    8831,
    -1,
    -1,
    1498,
    -1,
    -1,
    -1,
    2492,
    -1,
    1459,
    -1,
    1412,
    -1,
    -1,
    -1,
    -1,
    -1,
    4905,
    -1,
    -1,
    9267,
    9268,
    -1,
    -1,
    -1,
    4903,
    -1,
    1396,
    -1,
    -1,
    -1,
    9269,
    4069,
    -1,
    4906,
    -1,
    4904,
    -1,
    -1,
    4902,
    2836,
    -1,
    -1,
    1397,
    -1,
    4901,
    -1,
    9270,
    -1,
    3911,
    -1,
    -1,
    -1,
    9271,
    -1,
    2229,
    -1,
    -1,
    -1,
    -1,
    -1,
    4910,
    4911,
    -1,
    -1,
    -1,
    -1,
    -1,
    2736,
    -1,
    -1,
    -1,
    4907,
    4335,
    -1,
    -1,
    -1,
    -1,
    3615,
    -1,
    -1,
    4909,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9272,
    -1,
    -1,
    -1,
    -1,
    2356,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    4912,
    -1,
    -1,
    9273,
    3815,
    1398,
    -1,
    -1,
    4913,
    -1,
    4908,
    1399,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1400,
    -1,
    -1,
    9274,
    -1,
    -1,
    -1,
    1401,
    4068,
    -1,
    -1,
    -1,
    -1,
    1402,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9275,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    3633,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    4914,
    3777,
    9276,
    -1,
    1403,
    9277,
    9278,
    -1,
    9279,
    -1,
    -1,
    -1,
    -1,
    1404,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9280,
    -1,
    9281,
    -1,
    9282,
    -1,
    -1,
    -1,
    -1,
    9283,
    -1,
    4915,
    4919,
    4916,
    9284,
    -1,
    1636,
    4918,
    -1,
    1405,
    -1,
    -1,
    -1,
    -1,
    2564,
    -1,
    4917,
    2156,
    -1,
    -1,
    -1,
    4931,
    -1,
    -1,
    -1,
    -1,
    -1,
    4924,
    4921,
    -1,
    -1,
    9285,
    -1,
    9286,
    -1,
    -1,
    -1,
    9287,
    3263,
    -1,
    4920,
    -1,
    9288,
    4922,
    -1,
    -1,
    4923,
    -1,
    -1,
    -1,
    -1,
    9290,
    -1,
    9289,
    -1,
    -1,
    -1,
    -1,
    9291,
    -1,
    -1,
    -1,
    -1,
    4925,
    4926,
    -1,
    -1,
    -1,
    -1,
    9292,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9293,
    1867,
    -1,
    4928,
    4927,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    4929,
    9294,
    -1,
    1407,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    2803,
    -1,
    -1,
    1408,
    -1,
    -1,
    -1,
    -1,
    4932,
    -1,
    3349,
    9296,
    -1,
    -1,
    1540,
    -1,
    4930,
    -1,
    9295,
    -1,
    4933,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    4934,
    -1,
    4937,
    4376,
    -1,
    4935,
    -1,
    4936,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9297,
    -1,
    -1,
    -1,
    2493,
    4938,
    9298,
    -1,
    2261,
    4939,
    4377,
    2533,
    3125,
    -1,
    4940,
    4941,
    2501,
    2262,
    -1,
    4091,
    -1,
    -1,
    -1,
    1887,
    2202,
    4942,
    1747,
    -1,
    9299,
    4943,
    -1,
    3126,
    -1,
    -1,
    -1,
    9300,
    4944,
    4984,
    -1,
    4945,
    -1,
    4946,
    -1,
    -1,
    4947,
    -1,
    4949,
    -1,
    9301,
    4378,
    -1,
    9302,
    4950,
    9303,
    -1,
    4951,
    9304,
    3181,
    -1,
    1510,
    2610,
    1443,
    -1,
    3062,
    1801,
    2558,
    -1,
    2263,
    -1,
    -1,
    -1,
    4379,
    -1,
    3453,
    9494,
    2632,
    1802,
    3267,
    3363,
    1433,
    1904,
    3938,
    -1,
    2571,
    -1,
    -1,
    1929,
    2988,
    2565,
    4143,
    4952,
    -1,
    -1,
    -1,
    -1,
    -1,
    9495,
    -1,
    1939,
    -1,
    2392,
    -1,
    -1,
    1709,
    1570,
    2737,
    1637,
    -1,
    4953,
    4173,
    -1,
    -1,
    -1,
    -1,
    -1,
    2672,
    9497,
    -1,
    2605,
    4954,
    1868,
    3527,
    4050,
    4955,
    -1,
    4956,
    -1,
    -1,
    3816,
    -1,
    9499,
    -1,
    4958,
    -1,
    1793,
    2060,
    4957,
    -1,
    9500,
    -1,
    4380,
    -1,
    -1,
    1803,
    -1,
    2837,
    4962,
    2444,
    -1,
    1638,
    4961,
    -1,
    4959,
    4963,
    4960,
    3582,
    5590,
    2838,
    -1,
    4964,
    4381,
    -1,
    4261,
    -1,
    4965,
    9503,
    -1,
    4967,
    -1,
    3289,
    4966,
    -1,
    2916,
    -1,
    2534,
    -1,
    -1,
    9504,
    3149,
    2624,
    -1,
    3844,
    2989,
    9506,
    2582,
    4968,
    2738,
    4969,
    4970,
    1460,
    3127,
    2869,
    -1,
    4971,
    3496,
    2739,
    -1,
    2740,
    4382,
    4972,
    -1,
    -1,
    2990,
    -1,
    -1,
    -1,
    2741,
    -1,
    -1,
    -1,
    4383,
    -1,
    4973,
    -1,
    4974,
    4384,
    4105,
    -1,
    -1,
    -1,
    4975,
    9507,
    -1,
    4385,
    -1,
    2011,
    -1,
    -1,
    9508,
    2633,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    4976,
    4977,
    2597,
    2831,
    3561,
    2871,
    3753,
    3573,
    2014,
    4978,
    -1,
    -1,
    -1,
    1958,
    4979,
    -1,
    2067,
    -1,
    3524,
    1613,
    -1,
    2494,
    4980,
    4983,
    4982,
    2066,
    -1,
    4981,
    -1,
    3407,
    -1,
    -1,
    -1,
    -1,
    -1,
    8832,
    -1,
    -1,
    3117,
    9510,
    3426,
    2577,
    4387,
    9511,
    3063,
    4225,
    -1,
    9512,
    9513,
    9514,
    -1,
    -1,
    4985,
    -1,
    4986,
    3531,
    9515,
    2462,
    -1,
    -1,
    -1,
    -1,
    4988,
    -1,
    -1,
    4989,
    4391,
    -1,
    9518,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9521,
    -1,
    9522,
    -1,
    4990,
    -1,
    -1,
    4392,
    1869,
    4991,
    -1,
    -1,
    4992,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    4393,
    9526,
    1609,
    9527,
    4394,
    -1,
    -1,
    9528,
    9529,
    3034,
    1853,
    4395,
    4994,
    4049,
    9531,
    -1,
    -1,
    -1,
    3151,
    -1,
    1748,
    -1,
    9532,
    4996,
    4998,
    1848,
    -1,
    4396,
    4995,
    4997,
    -1,
    5000,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    4999,
    -1,
    5001,
    -1,
    9534,
    -1,
    4397,
    -1,
    -1,
    -1,
    -1,
    4398,
    -1,
    4399,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    5002,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    3506,
    1987,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1666,
    5003,
    5008,
    -1,
    -1,
    5006,
    -1,
    3940,
    3939,
    -1,
    -1,
    -1,
    4400,
    -1,
    3454,
    -1,
    -1,
    -1,
    5005,
    2683,
    -1,
    5004,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9537,
    2901,
    -1,
    -1,
    -1,
    5009,
    -1,
    4403,
    2423,
    -1,
    9538,
    5015,
    -1,
    -1,
    5016,
    5010,
    1710,
    5011,
    5020,
    5019,
    5018,
    5014,
    -1,
    9540,
    -1,
    5013,
    9541,
    -1,
    5017,
    -1,
    9542,
    -1,
    9543,
    4404,
    -1,
    3941,
    -1,
    9539,
    -1,
    -1,
    -1,
    -1,
    -1,
    9544,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9545,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9546,
    -1,
    -1,
    -1,
    4406,
    4407,
    -1,
    -1,
    4409,
    5024,
    5021,
    -1,
    5023,
    -1,
    4217,
    -1,
    5022,
    4408,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    5012,
    -1,
    -1,
    -1,
    -1,
    9548,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    2902,
    9549,
    -1,
    5025,
    4410,
    -1,
    2372,
    9551,
    -1,
    -1,
    5026,
    -1,
    -1,
    5027,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    4411,
    5030,
    -1,
    5029,
    -1,
    -1,
    5028,
    9554,
    -1,
    -1,
    3455,
    5007,
    -1,
    -1,
    -1,
    5036,
    -1,
    9556,
    -1,
    9557,
    -1,
    -1,
    9558,
    -1,
    9559,
    -1,
    -1,
    -1,
    5032,
    -1,
    -1,
    4412,
    -1,
    5031,
    -1,
    4413,
    -1,
    -1,
    4414,
    -1,
    -1,
    -1,
    -1,
    5033,
    -1,
    5034,
    -1,
    9560,
    -1,
    9561,
    -1,
    9562,
    -1,
    -1,
    5037,
    4415,
    9564,
    4299,
    -1,
    5038,
    5035,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    5039,
    -1,
    4416,
    1849,
    5040,
    -1,
    -1,
    -1,
    9565,
    5042,
    5041,
    -1,
    -1,
    5043,
    9566,
    9567,
    -1,
    -1,
    5044,
    -1,
    2991,
    2634,
    -1,
    9568,
    2699,
    7809,
    3074,
    9570,
    2264,
    2373,
    2265,
    1959,
    9571,
    -1,
    5045,
    -1,
    -1,
    2374,
    -1,
    -1,
    2203,
    5046,
    4047,
    3606,
    5047,
    -1,
    2266,
    -1,
    -1,
    -1,
    1797,
    -1,
    3202,
    2023,
    -1,
    9573,
    -1,
    2495,
    3818,
    -1,
    -1,
    3698,
    -1,
    -1,
    -1,
    -1,
    5048,
    1870,
    -1,
    -1,
    -1,
    -1,
    5051,
    9575,
    -1,
    7810,
    9576,
    3290,
    -1,
    7811,
    5050,
    5049,
    5052,
    -1,
    3364,
    -1,
    9578,
    -1,
    -1,
    -1,
    -1,
    -1,
    2886,
    -1,
    -1,
    -1,
    -1,
    -1,
    2496,
    -1,
    2955,
    9579,
    3152,
    1881,
    -1,
    -1,
    3291,
    -1,
    -1,
    5053,
    5054,
    2804,
    -1,
    -1,
    -1,
    -1,
    3972,
    9580,
    -1,
    5057,
    -1,
    -1,
    5056,
    5055,
    3852,
    -1,
    5064,
    -1,
    9581,
    -1,
    -1,
    4005,
    -1,
    5058,
    -1,
    -1,
    -1,
    -1,
    -1,
    5060,
    4022,
    9583,
    5059,
    7812,
    -1,
    -1,
    -1,
    -1,
    -1,
    7813,
    5061,
    -1,
    3673,
    5062,
    3883,
    5063,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9585,
    9586,
    9587,
    9870,
    -1,
    -1,
    -1,
    1804,
    3884,
    3586,
    5065,
    5066,
    8833,
    2267,
    1805,
    5067,
    2185,
    4171,
    4144,
    1871,
    5069,
    -1,
    3292,
    -1,
    2268,
    2742,
    -1,
    -1,
    3728,
    -1,
    -1,
    2743,
    -1,
    -1,
    -1,
    -1,
    2719,
    -1,
    -1,
    -1,
    -1,
    -1,
    3365,
    3942,
    3408,
    -1,
    -1,
    2269,
    -1,
    3819,
    -1,
    -1,
    -1,
    5070,
    -1,
    -1,
    -1,
    -1,
    9872,
    3440,
    2383,
    -1,
    -1,
    9873,
    2204,
    9874,
    3366,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1444,
    2709,
    2270,
    4174,
    9875,
    -1,
    -1,
    -1,
    -1,
    7814,
    9876,
    -1,
    5071,
    5072,
    3620,
    -1,
    -1,
    9877,
    -1,
    5073,
    4316,
    4336,
    7816,
    -1,
    -1,
    -1,
    5075,
    5074,
    -1,
    9878,
    1730,
    -1,
    -1,
    5076,
    -1,
    -1,
    9879,
    5079,
    5080,
    -1,
    5078,
    -1,
    3791,
    2744,
    5082,
    5081,
    5077,
    -1,
    -1,
    -1,
    -1,
    5083,
    5084,
    -1,
    -1,
    5085,
    -1,
    -1,
    -1,
    5088,
    5086,
    -1,
    5087,
    5089,
    -1,
    1571,
    3367,
    5090,
    7817,
    2157,
    1683,
    3593,
    9881,
    5091,
    3568,
    7818,
    3912,
    7819,
    5092,
    4337,
    -1,
    -1,
    7820,
    9882,
    5093,
    3885,
    5096,
    4418,
    4434,
    9883,
    2553,
    3562,
    5097,
    -1,
    1940,
    3293,
    1499,
    5098,
    3859,
    2271,
    -1,
    -1,
    3243,
    9884,
    7821,
    9886,
    3368,
    -1,
    -1,
    -1,
    7822,
    -1,
    4120,
    2186,
    2205,
    -1,
    5099,
    -1,
    -1,
    -1,
    5100,
    -1,
    5106,
    -1,
    2606,
    -1,
    -1,
    7823,
    3294,
    7824,
    1988,
    5101,
    -1,
    -1,
    -1,
    3770,
    7825,
    3233,
    -1,
    7826,
    5102,
    -1,
    -1,
    -1,
    7827,
    -1,
    9887,
    5103,
    -1,
    1989,
    -1,
    5104,
    -1,
    5105,
    -1,
    -1,
    5107,
    -1,
    3469,
    7828,
    -1,
    5108,
    5109,
    7829,
    5110,
    -1,
    -1,
    5095,
    5094,
    -1,
    -1,
    -1,
    5111,
    2100,
    9888,
    7830,
    -1,
    3766,
    7831,
    -1,
    2393,
    3782,
    3295,
    3802,
    5112,
    -1,
    -1,
    2745,
    1541,
    9889,
    5113,
    -1,
    -1,
    -1,
    5114,
    -1,
    4123,
    -1,
    -1,
    3729,
    7832,
    9890,
    5117,
    1596,
    2925,
    5116,
    5115,
    2101,
    3153,
    -1,
    5121,
    5119,
    7833,
    5118,
    4236,
    2230,
    -1,
    -1,
    9891,
    2720,
    5120,
    3427,
    2661,
    -1,
    -1,
    -1,
    3509,
    5124,
    5123,
    -1,
    -1,
    7834,
    -1,
    5122,
    -1,
    5125,
    2231,
    9892,
    -1,
    9893,
    -1,
    -1,
    7835,
    5126,
    3851,
    2690,
    -1,
    -1,
    5127,
    3754,
    7836,
    -1,
    -1,
    -1,
    3510,
    3296,
    7837,
    -1,
    7838,
    9894,
    3399,
    -1,
    -1,
    5128,
    1890,
    -1,
    -1,
    -1,
    -1,
    -1,
    2839,
    9895,
    3771,
    -1,
    9897,
    -1,
    7839,
    -1,
    9898,
    1872,
    3577,
    -1,
    -1,
    -1,
    -1,
    9899,
    9900,
    9901,
    -1,
    5129,
    2497,
    3973,
    3974,
    -1,
    -1,
    1597,
    5134,
    7840,
    -1,
    3268,
    7841,
    9902,
    -1,
    5131,
    -1,
    -1,
    -1,
    -1,
    7842,
    -1,
    1684,
    -1,
    -1,
    9903,
    9904,
    5182,
    5133,
    -1,
    9905,
    -1,
    3587,
    -1,
    -1,
    5132,
    -1,
    -1,
    5130,
    9906,
    2345,
    -1,
    5136,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    7843,
    5142,
    5148,
    5140,
    -1,
    3443,
    -1,
    7844,
    5145,
    3820,
    9908,
    7845,
    5139,
    -1,
    5144,
    4300,
    2498,
    -1,
    -1,
    3154,
    5137,
    9909,
    -1,
    9910,
    1941,
    5147,
    2926,
    1572,
    5141,
    1685,
    5146,
    -1,
    -1,
    -1,
    1990,
    -1,
    5143,
    -1,
    7846,
    -1,
    7847,
    -1,
    -1,
    -1,
    -1,
    5149,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    5151,
    5161,
    5159,
    -1,
    -1,
    5156,
    7848,
    -1,
    -1,
    5155,
    4317,
    9912,
    5157,
    -1,
    -1,
    1991,
    -1,
    2272,
    -1,
    -1,
    2721,
    -1,
    -1,
    -1,
    5164,
    5150,
    -1,
    -1,
    -1,
    -1,
    5154,
    5138,
    -1,
    1687,
    5158,
    5160,
    3244,
    -1,
    -1,
    2357,
    1621,
    5152,
    5163,
    5162,
    1992,
    -1,
    3110,
    1761,
    -1,
    -1,
    -1,
    -1,
    2102,
    -1,
    5153,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9913,
    -1,
    5165,
    -1,
    5168,
    5170,
    -1,
    -1,
    -1,
    -1,
    2566,
    9914,
    5176,
    3369,
    5166,
    -1,
    -1,
    -1,
    -1,
    5174,
    -1,
    1686,
    9915,
    5172,
    5173,
    -1,
    -1,
    5169,
    5171,
    -1,
    7850,
    7851,
    2232,
    4145,
    -1,
    -1,
    1806,
    -1,
    -1,
    1560,
    5175,
    9916,
    3597,
    1422,
    -1,
    -1,
    -1,
    -1,
    -1,
    9918,
    9919,
    3730,
    5135,
    5181,
    5185,
    4110,
    -1,
    5178,
    -1,
    -1,
    -1,
    3456,
    5183,
    9920,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    2805,
    5184,
    3532,
    9921,
    -1,
    -1,
    7852,
    -1,
    -1,
    -1,
    -1,
    -1,
    4362,
    -1,
    5180,
    7853,
    7854,
    -1,
    -1,
    5186,
    9922,
    2346,
    9923,
    2956,
    7855,
    -1,
    1461,
    5179,
    5177,
    -1,
    3065,
    -1,
    -1,
    -1,
    5167,
    2463,
    -1,
    -1,
    -1,
    -1,
    -1,
    9924,
    -1,
    3139,
    5198,
    9925,
    3066,
    5193,
    9926,
    5190,
    5191,
    7856,
    2607,
    5194,
    5197,
    -1,
    -1,
    -1,
    -1,
    5192,
    2636,
    -1,
    5195,
    -1,
    -1,
    5189,
    -1,
    4133,
    4132,
    -1,
    -1,
    -1,
    5199,
    5200,
    1462,
    9927,
    -1,
    9928,
    9929,
    -1,
    5188,
    -1,
    -1,
    -1,
    9930,
    2055,
    1415,
    7857,
    -1,
    9931,
    1807,
    -1,
    5196,
    -1,
    -1,
    -1,
    -1,
    -1,
    5204,
    5203,
    -1,
    -1,
    7858,
    5208,
    -1,
    -1,
    -1,
    7859,
    -1,
    -1,
    -1,
    5209,
    -1,
    -1,
    7860,
    -1,
    -1,
    9932,
    -1,
    5207,
    5210,
    5202,
    5206,
    -1,
    9934,
    5211,
    -1,
    5212,
    -1,
    9935,
    5201,
    2535,
    -1,
    5205,
    3155,
    2273,
    5187,
    2840,
    -1,
    -1,
    -1,
    -1,
    5225,
    -1,
    3926,
    -1,
    -1,
    5215,
    5216,
    5217,
    -1,
    -1,
    5224,
    -1,
    5223,
    9936,
    -1,
    4040,
    1808,
    -1,
    5221,
    -1,
    2104,
    1711,
    -1,
    -1,
    5218,
    -1,
    -1,
    4253,
    5220,
    1463,
    5222,
    -1,
    5213,
    5219,
    5226,
    2103,
    5214,
    -1,
    -1,
    -1,
    -1,
    9937,
    -1,
    4196,
    -1,
    -1,
    -1,
    4146,
    -1,
    -1,
    -1,
    -1,
    5229,
    -1,
    -1,
    5233,
    -1,
    -1,
    7862,
    3101,
    -1,
    4318,
    5234,
    9939,
    9940,
    5231,
    -1,
    5228,
    9941,
    9942,
    5227,
    5232,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    3869,
    9943,
    -1,
    3497,
    9944,
    2105,
    -1,
    5235,
    5230,
    9945,
    5236,
    -1,
    -1,
    -1,
    2158,
    -1,
    -1,
    -1,
    1614,
    -1,
    -1,
    7865,
    5244,
    -1,
    7864,
    -1,
    1809,
    -1,
    -1,
    -1,
    -1,
    5242,
    -1,
    -1,
    5243,
    2358,
    5241,
    5239,
    5238,
    5245,
    5237,
    5247,
    -1,
    -1,
    1688,
    -1,
    -1,
    -1,
    -1,
    9947,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9948,
    -1,
    9949,
    -1,
    -1,
    -1,
    5249,
    -1,
    -1,
    5248,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    3297,
    -1,
    5252,
    9950,
    5250,
    5240,
    2159,
    -1,
    5251,
    -1,
    5255,
    5254,
    5256,
    5253,
    5257,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    5258,
    5259,
    3927,
    -1,
    5261,
    5260,
    2662,
    -1,
    2927,
    1667,
    1689,
    -1,
    5262,
    9952,
    1440,
    -1,
    -1,
    -1,
    2957,
    5263,
    -1,
    7084,
    5264,
    2132,
    -1,
    5265,
    7867,
    9953,
    -1,
    -1,
    2992,
    -1,
    -1,
    9954,
    5266,
    -1,
    -1,
    -1,
    5267,
    1905,
    5268,
    -1,
    5269,
    5270,
    3156,
    -1,
    -1,
    -1,
    2206,
    -1,
    -1,
    4106,
    -1,
    -1,
    7868,
    3975,
    2705,
    5271,
    -1,
    7869,
    -1,
    -1,
    9955,
    2993,
    7266,
    3731,
    -1,
    2611,
    9956,
    2394,
    5272,
    -1,
    -1,
    9957,
    9958,
    3140,
    -1,
    3860,
    7870,
    -1,
    3182,
    -1,
    7871,
    5275,
    -1,
    -1,
    5273,
    -1,
    5276,
    9959,
    -1,
    5274,
    9960,
    -1,
    -1,
    -1,
    5277,
    -1,
    -1,
    -1,
    -1,
    9962,
    3870,
    7872,
    -1,
    1432,
    -1,
    9963,
    -1,
    -1,
    3821,
    -1,
    -1,
    3732,
    -1,
    9961,
    5278,
    -1,
    5281,
    2746,
    1906,
    -1,
    5279,
    5286,
    2747,
    7873,
    -1,
    -1,
    -1,
    5280,
    3607,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    4197,
    5282,
    5283,
    5287,
    3457,
    5284,
    2274,
    2975,
    9964,
    -1,
    5301,
    3685,
    -1,
    3183,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9965,
    -1,
    -1,
    -1,
    -1,
    3733,
    5370,
    -1,
    -1,
    -1,
    -1,
    3943,
    -1,
    -1,
    -1,
    3370,
    -1,
    -1,
    -1,
    4032,
    -1,
    5290,
    1598,
    3269,
    -1,
    -1,
    -1,
    -1,
    5299,
    -1,
    7874,
    3215,
    5293,
    5300,
    5295,
    5302,
    5298,
    -1,
    5297,
    3650,
    -1,
    5291,
    1690,
    5289,
    1960,
    3184,
    5285,
    9966,
    7875,
    5288,
    2275,
    2972,
    -1,
    2748,
    5296,
    3621,
    -1,
    -1,
    1961,
    1731,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1762,
    2820,
    5304,
    5309,
    -1,
    5305,
    -1,
    2160,
    -1,
    5310,
    2445,
    2328,
    -1,
    -1,
    -1,
    -1,
    7876,
    9968,
    2637,
    5292,
    -1,
    2536,
    5307,
    9969,
    -1,
    -1,
    -1,
    2499,
    5308,
    1445,
    7877,
    -1,
    5303,
    9970,
    -1,
    -1,
    9971,
    3298,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    7878,
    1962,
    -1,
    -1,
    -1,
    -1,
    -1,
    1993,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    5306,
    1416,
    -1,
    -1,
    2384,
    -1,
    -1,
    -1,
    2841,
    -1,
    -1,
    9974,
    -1,
    -1,
    9975,
    -1,
    -1,
    -1,
    7879,
    3371,
    9976,
    9977,
    3717,
    5312,
    3069,
    -1,
    9978,
    -1,
    7880,
    9979,
    -1,
    -1,
    -1,
    -1,
    3111,
    -1,
    -1,
    2454,
    5313,
    9980,
    5315,
    5311,
    -1,
    -1,
    -1,
    -1,
    3918,
    -1,
    3318,
    -1,
    9982,
    -1,
    -1,
    3067,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    7881,
    -1,
    3944,
    2583,
    5327,
    -1,
    5326,
    9985,
    -1,
    2907,
    -1,
    -1,
    -1,
    2161,
    -1,
    -1,
    -1,
    5320,
    2750,
    -1,
    -1,
    3549,
    3588,
    7882,
    -1,
    -1,
    -1,
    5318,
    -1,
    -1,
    3068,
    9986,
    -1,
    -1,
    -1,
    2625,
    5323,
    -1,
    -1,
    2749,
    -1,
    5317,
    5322,
    -1,
    -1,
    3622,
    -1,
    9989,
    -1,
    5316,
    -1,
    2068,
    9987,
    -1,
    1752,
    -1,
    -1,
    -1,
    5324,
    4240,
    2395,
    3216,
    5321,
    -1,
    2973,
    -1,
    2276,
    2887,
    1573,
    3035,
    5319,
    1916,
    -1,
    -1,
    -1,
    -1,
    -1,
    2106,
    -1,
    3336,
    5325,
    -1,
    -1,
    -1,
    -1,
    -1,
    3070,
    -1,
    9990,
    5328,
    -1,
    5330,
    -1,
    -1,
    3124,
    5336,
    -1,
    5331,
    -1,
    -1,
    5333,
    -1,
    -1,
    -1,
    -1,
    -1,
    3792,
    3372,
    -1,
    5334,
    -1,
    9992,
    9993,
    4147,
    -1,
    -1,
    -1,
    4175,
    1810,
    9994,
    -1,
    -1,
    -1,
    9995,
    1423,
    -1,
    5332,
    -1,
    7883,
    -1,
    -1,
    -1,
    5329,
    -1,
    9996,
    9997,
    7884,
    1873,
    -1,
    -1,
    -1,
    9998,
    -1,
    1574,
    7885,
    5335,
    9991,
    -1,
    -1,
    4176,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    5339,
    -1,
    -1,
    9999,
    -1,
    -1,
    -1,
    3128,
    -1,
    5346,
    7886,
    -1,
    -1,
    5340,
    7887,
    -1,
    5337,
    5344,
    -1,
    -1,
    -1,
    -1,
    5314,
    -1,
    10000,
    -1,
    -1,
    -1,
    7888,
    -1,
    -1,
    10001,
    5341,
    -1,
    5345,
    10002,
    -1,
    -1,
    3699,
    3458,
    -1,
    10003,
    -1,
    -1,
    -1,
    -1,
    5338,
    -1,
    5342,
    -1,
    -1,
    -1,
    2107,
    -1,
    -1,
    -1,
    2430,
    -1,
    -1,
    -1,
    2974,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    5350,
    -1,
    -1,
    7890,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    3389,
    -1,
    10004,
    10005,
    -1,
    10006,
    -1,
    -1,
    7891,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    5347,
    -1,
    4011,
    -1,
    -1,
    -1,
    7892,
    -1,
    5348,
    -1,
    -1,
    -1,
    10007,
    -1,
    -1,
    5349,
    -1,
    4087,
    7889,
    2915,
    -1,
    -1,
    10008,
    -1,
    -1,
    -1,
    -1,
    -1,
    2133,
    -1,
    -1,
    -1,
    10009,
    5356,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    10010,
    2464,
    5353,
    -1,
    5352,
    -1,
    -1,
    -1,
    -1,
    3589,
    -1,
    -1,
    10011,
    3498,
    10012,
    -1,
    -1,
    -1,
    -1,
    3400,
    5354,
    -1,
    -1,
    -1,
    5355,
    -1,
    3838,
    -1,
    3608,
    2446,
    -1,
    2994,
    -1,
    3995,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1732,
    -1,
    5362,
    5357,
    -1,
    7894,
    7895,
    -1,
    4177,
    5364,
    -1,
    7896,
    5360,
    -1,
    5361,
    -1,
    -1,
    7897,
    10013,
    10014,
    3071,
    7893,
    -1,
    7898,
    -1,
    5359,
    -1,
    5294,
    10015,
    -1,
    10016,
    5363,
    -1,
    5358,
    -1,
    -1,
    -1,
    -1,
    -1,
    5368,
    5369,
    3390,
    5371,
    10018,
    10019,
    2447,
    5366,
    -1,
    -1,
    -1,
    -1,
    1907,
    -1,
    -1,
    5372,
    -1,
    5365,
    5376,
    -1,
    5375,
    -1,
    5374,
    7899,
    -1,
    -1,
    5377,
    7900,
    -1,
    5379,
    2806,
    10020,
    5378,
    -1,
    -1,
    -1,
    10021,
    5382,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    10023,
    -1,
    -1,
    -1,
    -1,
    10024,
    -1,
    10025,
    -1,
    5380,
    -1,
    -1,
    -1,
    5381,
    5343,
    10027,
    -1,
    -1,
    -1,
    7901,
    5384,
    5383,
    -1,
    -1,
    -1,
    -1,
    7902,
    5351,
    5385,
    5373,
    -1,
    -1,
    2500,
    -1,
    -1,
    10028,
    -1,
    5386,
    5387,
    5389,
    5388,
    5390,
    1691,
    -1,
    2277,
    -1,
    -1,
    3945,
    2928,
    -1,
    -1,
    -1,
    -1,
    10029,
    2207,
    -1,
    -1,
    5392,
    -1,
    -1,
    -1,
    -1,
    5395,
    -1,
    3809,
    -1,
    1942,
    -1,
    -1,
    10030,
    5394,
    5393,
    3623,
    5396,
    1994,
    -1,
    -1,
    -1,
    5398,
    5397,
    -1,
    -1,
    -1,
    1811,
    2465,
    -1,
    -1,
    3533,
    7904,
    -1,
    -1,
    -1,
    10031,
    2108,
    -1,
    -1,
    -1,
    2903,
    -1,
    5399,
    -1,
    2929,
    3391,
    -1,
    3822,
    5400,
    -1,
    10032,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    10033,
    5401,
    5402,
    10034,
    10035,
    -1,
    3877,
    4948,
    2951,
    10036,
    -1,
    3803,
    -1,
    2405,
    -1,
    3734,
    3700,
    -1,
    -1,
    -1,
    -1,
    -1,
    3428,
    -1,
    4262,
    -1,
    5404,
    2585,
    7905,
    -1,
    5405,
    -1,
    1431,
    -1,
    -1,
    2025,
    2958,
    -1,
    3823,
    -1,
    -1,
    -1,
    5406,
    2477,
    3234,
    -1,
    2502,
    2842,
    -1,
    10037,
    -1,
    -1,
    10038,
    -1,
    5407,
    10039,
    3946,
    -1,
    -1,
    1590,
    2503,
    -1,
    10040,
    -1,
    5410,
    10041,
    5408,
    5411,
    4254,
    5409,
    -1,
    -1,
    10042,
    -1,
    3004,
    5412,
    -1,
    -1,
    3119,
    -1,
    -1,
    5413,
    -1,
    10043,
    -1,
    -1,
    1875,
    -1,
    5415,
    -1,
    5414,
    -1,
    -1,
    -1,
    -1,
    5416,
    5417,
    1876,
    -1,
    -1,
    3569,
    3217,
    1955,
    2504,
    3072,
    -1,
    -1,
    2691,
    1425,
    -1,
    -1,
    -1,
    5418,
    10045,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    10046,
    1599,
    5422,
    10047,
    -1,
    -1,
    -1,
    7907,
    -1,
    2278,
    5421,
    10048,
    -1,
    2360,
    2751,
    10049,
    7908,
    5420,
    -1,
    2752,
    -1,
    4073,
    2359,
    -1,
    -1,
    -1,
    1464,
    2959,
    7909,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    5427,
    -1,
    7910,
    2930,
    1542,
    10050,
    7912,
    -1,
    7913,
    2684,
    -1,
    4017,
    2431,
    -1,
    10051,
    7914,
    -1,
    2753,
    -1,
    2920,
    7915,
    7916,
    -1,
    7917,
    5426,
    5424,
    5425,
    -1,
    -1,
    -1,
    7911,
    -1,
    3270,
    -1,
    -1,
    5457,
    -1,
    5431,
    2537,
    2279,
    5429,
    10052,
    -1,
    -1,
    7919,
    5430,
    -1,
    2843,
    7920,
    -1,
    10054,
    5428,
    -1,
    10053,
    2458,
    -1,
    -1,
    -1,
    -1,
    10057,
    -1,
    7922,
    7923,
    10058,
    -1,
    5433,
    5432,
    5437,
    -1,
    7924,
    5438,
    10059,
    5434,
    7925,
    1693,
    5435,
    5436,
    3718,
    10062,
    10063,
    10064,
    10065,
    3824,
    2109,
    5439,
    -1,
    -1,
    7926,
    2931,
    -1,
    2754,
    7927,
    7928,
    -1,
    3245,
    10066,
    -1,
    -1,
    -1,
    -1,
    10067,
    2012,
    -1,
    5440,
    5444,
    -1,
    -1,
    1640,
    5441,
    5443,
    -1,
    -1,
    -1,
    7929,
    5442,
    -1,
    10068,
    2706,
    10069,
    -1,
    -1,
    -1,
    3235,
    1446,
    5445,
    10070,
    -1,
    -1,
    -1,
    5446,
    -1,
    -1,
    7931,
    -1,
    3299,
    -1,
    -1,
    -1,
    4310,
    -1,
    -1,
    -1,
    -1,
    2478,
    -1,
    10072,
    3928,
    -1,
    -1,
    10073,
    7932,
    -1,
    3976,
    10074,
    -1,
    -1,
    5453,
    5448,
    -1,
    7933,
    5451,
    -1,
    5450,
    10076,
    -1,
    5447,
    -1,
    -1,
    5452,
    -1,
    7934,
    3539,
    7935,
    5449,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    5454,
    -1,
    -1,
    2707,
    5455,
    7937,
    4178,
    3659,
    -1,
    -1,
    5456,
    -1,
    -1,
    -1,
    -1,
    -1,
    5458,
    -1,
    7938,
    5459,
    -1,
    -1,
    10078,
    -1,
    -1,
    -1,
    5460,
    -1,
    2015,
    1543,
    2280,
    5461,
    -1,
    5462,
    2712,
    3073,
    7939,
    7918,
    4661,
    3037,
    3036,
    3157,
    2388,
    10080,
    -1,
    4499,
    -1,
    10081,
    -1,
    -1,
    2145,
    4148,
    -1,
    3947,
    -1,
    3853,
    -1,
    5463,
    -1,
    -1,
    10082,
    7940,
    2432,
    3320,
    5464,
    4338,
    -1,
    10084,
    -1,
    3977,
    -1,
    3300,
    5465,
    1877,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    5466,
    5467,
    4099,
    -1,
    4045,
    4033,
    4007,
    2448,
    5469,
    -1,
    -1,
    2612,
    -1,
    7942,
    3996,
    -1,
    5471,
    5474,
    5473,
    -1,
    1874,
    -1,
    -1,
    1943,
    -1,
    5470,
    -1,
    5472,
    -1,
    -1,
    -1,
    -1,
    5475,
    7944,
    7945,
    2908,
    -1,
    -1,
    10087,
    10088,
    4226,
    1450,
    2413,
    3129,
    -1,
    2598,
    10089,
    -1,
    2808,
    -1,
    -1,
    5478,
    -1,
    -1,
    3429,
    10090,
    5476,
    3112,
    5477,
    2807,
    4102,
    5479,
    5480,
    4207,
    7943,
    -1,
    -1,
    -1,
    5485,
    -1,
    -1,
    2281,
    -1,
    3624,
    5482,
    3459,
    5419,
    5423,
    10094,
    1926,
    10095,
    3610,
    -1,
    -1,
    -1,
    7946,
    5484,
    -1,
    2755,
    3701,
    -1,
    7947,
    -1,
    -1,
    -1,
    5490,
    -1,
    3755,
    -1,
    5481,
    -1,
    5487,
    5486,
    -1,
    -1,
    -1,
    2960,
    -1,
    10097,
    7948,
    -1,
    4024,
    -1,
    4280,
    7949,
    -1,
    4018,
    7950,
    1641,
    2505,
    -1,
    -1,
    4363,
    5489,
    2904,
    -1,
    -1,
    -1,
    5488,
    -1,
    -1,
    5483,
    -1,
    -1,
    -1,
    -1,
    -1,
    2208,
    10101,
    -1,
    10102,
    5495,
    5493,
    -1,
    1642,
    5491,
    5497,
    5503,
    -1,
    7951,
    -1,
    -1,
    -1,
    -1,
    7953,
    3141,
    -1,
    10103,
    3886,
    -1,
    5505,
    -1,
    10104,
    -1,
    3761,
    -1,
    -1,
    -1,
    5504,
    3651,
    3978,
    1812,
    10105,
    3000,
    2663,
    -1,
    -1,
    7954,
    3340,
    10106,
    4149,
    10107,
    -1,
    5500,
    5499,
    -1,
    -1,
    -1,
    5501,
    -1,
    5498,
    -1,
    -1,
    5506,
    -1,
    5496,
    -1,
    -1,
    5494,
    -1,
    5502,
    5492,
    10108,
    3271,
    -1,
    4129,
    2575,
    2433,
    -1,
    10109,
    -1,
    7952,
    -1,
    2375,
    7955,
    -1,
    4026,
    1724,
    -1,
    7956,
    3335,
    3519,
    1544,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    10100,
    -1,
    -1,
    2995,
    -1,
    -1,
    2933,
    2077,
    10113,
    -1,
    -1,
    -1,
    -1,
    7958,
    5508,
    10114,
    -1,
    2282,
    1783,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    5510,
    1775,
    5516,
    7959,
    10115,
    -1,
    -1,
    -1,
    7960,
    5513,
    10117,
    2996,
    -1,
    -1,
    -1,
    1734,
    2361,
    -1,
    10118,
    1733,
    2396,
    10119,
    -1,
    5511,
    2136,
    2110,
    3460,
    10120,
    10121,
    5509,
    -1,
    1447,
    10122,
    -1,
    -1,
    10123,
    5512,
    5514,
    -1,
    2018,
    2079,
    7957,
    1813,
    1922,
    10124,
    -1,
    10125,
    -1,
    5517,
    -1,
    7961,
    2438,
    4028,
    -1,
    2466,
    -1,
    -1,
    -1,
    5518,
    -1,
    -1,
    -1,
    3775,
    -1,
    -1,
    -1,
    10127,
    -1,
    -1,
    10128,
    -1,
    -1,
    -1,
    7962,
    -1,
    5530,
    7963,
    1616,
    5519,
    -1,
    -1,
    10129,
    -1,
    10130,
    -1,
    5536,
    5520,
    -1,
    4263,
    10131,
    5527,
    -1,
    3634,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    5535,
    -1,
    5522,
    10132,
    -1,
    -1,
    1429,
    5524,
    -1,
    10133,
    2283,
    10135,
    10136,
    10137,
    5526,
    10138,
    5525,
    -1,
    5521,
    5532,
    -1,
    2756,
    7965,
    -1,
    7966,
    4870,
    2233,
    4227,
    -1,
    10139,
    10140,
    -1,
    5523,
    -1,
    3373,
    1694,
    2362,
    7967,
    5515,
    10141,
    5531,
    1756,
    -1,
    -1,
    5529,
    5533,
    10142,
    3461,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    10147,
    1879,
    10148,
    5563,
    -1,
    7968,
    4082,
    5538,
    1878,
    10149,
    5545,
    -1,
    10150,
    7969,
    -1,
    3979,
    -1,
    5546,
    5548,
    10151,
    5552,
    5540,
    10152,
    3205,
    -1,
    -1,
    -1,
    -1,
    3462,
    5556,
    5543,
    -1,
    5553,
    10153,
    10154,
    -1,
    5547,
    7970,
    -1,
    -1,
    -1,
    10155,
    7971,
    2844,
    5557,
    7972,
    7973,
    2932,
    -1,
    -1,
    -1,
    -1,
    10156,
    -1,
    5555,
    1814,
    10143,
    7974,
    -1,
    -1,
    -1,
    4371,
    5537,
    -1,
    10157,
    5551,
    1465,
    -1,
    10158,
    5539,
    -1,
    7975,
    4067,
    5544,
    2821,
    3328,
    5534,
    -1,
    -1,
    5550,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    2909,
    5560,
    1772,
    2162,
    -1,
    -1,
    -1,
    -1,
    5562,
    5541,
    5561,
    -1,
    5554,
    5542,
    -1,
    5558,
    -1,
    5559,
    -1,
    -1,
    -1,
    -1,
    -1,
    5576,
    -1,
    -1,
    -1,
    3523,
    7978,
    5549,
    -1,
    -1,
    5572,
    -1,
    10162,
    -1,
    5574,
    -1,
    3345,
    -1,
    -1,
    7979,
    -1,
    -1,
    -1,
    10164,
    -1,
    -1,
    7976,
    4179,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    3845,
    5569,
    3143,
    -1,
    7980,
    -1,
    5575,
    3038,
    -1,
    5566,
    5579,
    5578,
    -1,
    3556,
    5577,
    3551,
    7981,
    7982,
    -1,
    -1,
    -1,
    7983,
    10165,
    5581,
    5568,
    10166,
    2013,
    5571,
    2692,
    -1,
    -1,
    10167,
    3635,
    5573,
    2016,
    -1,
    5565,
    5567,
    5564,
    10168,
    -1,
    4339,
    1749,
    5570,
    10169,
    7984,
    5580,
    1712,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    2420,
    -1,
    -1,
    -1,
    1564,
    -1,
    -1,
    5597,
    10170,
    -1,
    4340,
    5600,
    10172,
    -1,
    10173,
    -1,
    -1,
    2845,
    5599,
    -1,
    -1,
    -1,
    5598,
    -1,
    -1,
    -1,
    -1,
    7986,
    10174,
    5595,
    -1,
    -1,
    -1,
    -1,
    -1,
    7987,
    5583,
    -1,
    10175,
    5612,
    5582,
    -1,
    5601,
    -1,
    -1,
    10176,
    -1,
    -1,
    10177,
    5593,
    10178,
    -1,
    5588,
    5585,
    10179,
    5586,
    -1,
    5594,
    -1,
    -1,
    -1,
    7803,
    -1,
    -1,
    5591,
    2284,
    3329,
    3075,
    5589,
    7988,
    5584,
    10180,
    -1,
    5587,
    -1,
    -1,
    10181,
    -1,
    4180,
    4021,
    -1,
    -1,
    -1,
    5592,
    5602,
    -1,
    -1,
    -1,
    7991,
    10187,
    -1,
    -1,
    -1,
    5610,
    5603,
    7992,
    7993,
    5616,
    -1,
    5614,
    10188,
    10189,
    -1,
    -1,
    5609,
    10190,
    10192,
    7994,
    7995,
    -1,
    -1,
    5608,
    -1,
    3337,
    -1,
    3076,
    10193,
    5606,
    -1,
    -1,
    5604,
    -1,
    -1,
    5611,
    -1,
    -1,
    -1,
    -1,
    5617,
    3750,
    5623,
    -1,
    -1,
    7996,
    -1,
    10194,
    5618,
    5621,
    5615,
    7997,
    -1,
    3279,
    -1,
    3783,
    10195,
    5605,
    -1,
    10196,
    5613,
    2757,
    -1,
    4088,
    5633,
    5620,
    -1,
    -1,
    -1,
    -1,
    -1,
    2163,
    1600,
    1754,
    -1,
    -1,
    5596,
    -1,
    10184,
    -1,
    10198,
    10199,
    10200,
    2758,
    5625,
    -1,
    5632,
    2626,
    1773,
    7999,
    -1,
    3209,
    8000,
    10201,
    -1,
    -1,
    -1,
    -1,
    5622,
    8001,
    10202,
    5627,
    5631,
    10203,
    -1,
    1995,
    -1,
    -1,
    10205,
    -1,
    8002,
    -1,
    10206,
    -1,
    -1,
    -1,
    8003,
    -1,
    1923,
    5629,
    -1,
    8004,
    -1,
    -1,
    -1,
    1880,
    -1,
    3520,
    5628,
    -1,
    10207,
    -1,
    5630,
    -1,
    -1,
    -1,
    -1,
    8005,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    5624,
    8006,
    -1,
    -1,
    -1,
    -1,
    5626,
    -1,
    10204,
    -1,
    -1,
    -1,
    10209,
    1755,
    3236,
    -1,
    -1,
    10210,
    5637,
    -1,
    -1,
    -1,
    -1,
    8008,
    -1,
    10211,
    -1,
    5635,
    2234,
    -1,
    5634,
    10213,
    -1,
    -1,
    8009,
    -1,
    -1,
    5640,
    -1,
    -1,
    -1,
    -1,
    5507,
    8010,
    8011,
    10214,
    5636,
    10215,
    5638,
    5639,
    -1,
    8012,
    -1,
    -1,
    -1,
    -1,
    5651,
    10217,
    5647,
    -1,
    5528,
    -1,
    -1,
    -1,
    -1,
    5646,
    -1,
    -1,
    -1,
    -1,
    5645,
    -1,
    -1,
    5642,
    -1,
    10218,
    -1,
    -1,
    -1,
    5619,
    5644,
    5643,
    -1,
    -1,
    10219,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    10221,
    5649,
    -1,
    4329,
    10220,
    -1,
    -1,
    -1,
    -1,
    -1,
    5652,
    2064,
    10222,
    10223,
    5648,
    5650,
    -1,
    -1,
    -1,
    -1,
    8013,
    -1,
    -1,
    8014,
    3672,
    -1,
    5653,
    -1,
    10226,
    -1,
    -1,
    -1,
    -1,
    10227,
    10228,
    10229,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    5657,
    5654,
    -1,
    10230,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    4218,
    5655,
    -1,
    -1,
    -1,
    -1,
    5607,
    10232,
    -1,
    -1,
    -1,
    10233,
    10234,
    10235,
    5658,
    -1,
    -1,
    -1,
    5659,
    10237,
    -1,
    -1,
    -1,
    8017,
    -1,
    1524,
    8018,
    5661,
    2138,
    2538,
    -1,
    2026,
    -1,
    -1,
    -1,
    1601,
    -1,
    -1,
    -1,
    -1,
    8019,
    -1,
    -1,
    10239,
    -1,
    -1,
    4198,
    -1,
    -1,
    8020,
    -1,
    5663,
    5662,
    5665,
    1908,
    -1,
    -1,
    2027,
    1815,
    -1,
    -1,
    -1,
    -1,
    5668,
    -1,
    -1,
    8021,
    5667,
    -1,
    5669,
    10240,
    -1,
    1643,
    -1,
    3218,
    -1,
    5670,
    -1,
    -1,
    1816,
    5672,
    -1,
    8022,
    -1,
    10241,
    5671,
    -1,
    5673,
    -1,
    -1,
    -1,
    5674,
    8023,
    5675,
    2506,
    2934,
    2350,
    8024,
    3839,
    8025,
    -1,
    3919,
    4359,
    -1,
    10242,
    -1,
    -1,
    2528,
    -1,
    -1,
    -1,
    2397,
    4311,
    10243,
    -1,
    8026,
    5676,
    5677,
    10244,
    2507,
    -1,
    -1,
    -1,
    5678,
    5679,
    10245,
    8027,
    5681,
    5680,
    -1,
    4003,
    -1,
    -1,
    2693,
    2613,
    2479,
    -1,
    5682,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    5684,
    2822,
    -1,
    5683,
    -1,
    -1,
    16074,
    -1,
    -1,
    5685,
    -1,
    -1,
    -1,
    -1,
    -1,
    5686,
    -1,
    -1,
    -1,
    -1,
    8028,
    5687,
    5688,
    -1,
    8029,
    16075,
    5689,
    -1,
    5691,
    5690,
    5692,
    1602,
    3237,
    -1,
    5693,
    -1,
    -1,
    2449,
    1735,
    5694,
    16077,
    16078,
    3417,
    4837,
    -1,
    -1,
    -1,
    -1,
    1882,
    5695,
    16079,
    16080,
    16081,
    -1,
    5696,
    -1,
    3929,
    4019,
    8031,
    -1,
    -1,
    3516,
    5697,
    3735,
    -1,
    8032,
    8033,
    3756,
    -1,
    16082,
    4092,
    -1,
    -1,
    -1,
    5698,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16083,
    16084,
    -1,
    -1,
    -1,
    5700,
    5699,
    -1,
    16085,
    5702,
    -1,
    16086,
    -1,
    5701,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8034,
    -1,
    -1,
    16087,
    -1,
    -1,
    8035,
    -1,
    -1,
    5704,
    -1,
    16088,
    -1,
    -1,
    -1,
    16089,
    2508,
    8036,
    4058,
    -1,
    5705,
    5706,
    -1,
    -1,
    1883,
    -1,
    -1,
    -1,
    5707,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    5709,
    5708,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8037,
    2888,
    16090,
    16091,
    3784,
    1545,
    -1,
    16092,
    -1,
    -1,
    -1,
    3702,
    16094,
    3374,
    2664,
    1944,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16095,
    3703,
    -1,
    2551,
    -1,
    -1,
    -1,
    -1,
    5711,
    -1,
    1817,
    -1,
    8038,
    1591,
    16096,
    8039,
    3559,
    5710,
    2285,
    3246,
    -1,
    5712,
    -1,
    -1,
    -1,
    -1,
    16120,
    5720,
    -1,
    5713,
    -1,
    -1,
    16097,
    -1,
    -1,
    3134,
    -1,
    1945,
    5721,
    8041,
    -1,
    8042,
    -1,
    -1,
    -1,
    2139,
    -1,
    -1,
    1884,
    5719,
    -1,
    -1,
    5717,
    5714,
    4199,
    16098,
    8043,
    8044,
    -1,
    3321,
    16099,
    -1,
    -1,
    3534,
    5715,
    -1,
    -1,
    5723,
    -1,
    5722,
    2070,
    16101,
    16102,
    1610,
    16103,
    8045,
    2376,
    5716,
    5718,
    8046,
    -1,
    -1,
    -1,
    -1,
    4002,
    3185,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8040,
    4034,
    -1,
    16104,
    5731,
    -1,
    -1,
    5732,
    -1,
    1644,
    -1,
    -1,
    -1,
    -1,
    3861,
    4134,
    5734,
    2540,
    2759,
    5727,
    5733,
    1575,
    -1,
    1996,
    16105,
    -1,
    5724,
    5729,
    8048,
    -1,
    -1,
    2997,
    3652,
    -1,
    3736,
    -1,
    -1,
    -1,
    16106,
    -1,
    -1,
    5726,
    8049,
    3948,
    16108,
    5728,
    -1,
    5737,
    16109,
    5735,
    16110,
    5730,
    -1,
    -1,
    8050,
    3949,
    3611,
    1946,
    -1,
    3388,
    -1,
    -1,
    3272,
    16111,
    5738,
    8051,
    16112,
    16113,
    8052,
    5736,
    3158,
    5725,
    -1,
    1546,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8047,
    -1,
    -1,
    -1,
    -1,
    16115,
    -1,
    -1,
    -1,
    8054,
    -1,
    -1,
    -1,
    -1,
    -1,
    16116,
    4181,
    5749,
    -1,
    8055,
    -1,
    -1,
    -1,
    5748,
    -1,
    -1,
    -1,
    -1,
    2999,
    -1,
    5745,
    -1,
    4211,
    -1,
    -1,
    3499,
    5739,
    -1,
    -1,
    -1,
    -1,
    16117,
    3326,
    16118,
    16119,
    -1,
    1547,
    2286,
    5742,
    -1,
    -1,
    8056,
    16121,
    -1,
    8057,
    2638,
    5747,
    16142,
    5746,
    5741,
    -1,
    5744,
    8058,
    -1,
    1763,
    16122,
    5743,
    3612,
    8059,
    -1,
    4243,
    -1,
    -1,
    2809,
    2998,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8060,
    5755,
    5753,
    16123,
    3804,
    -1,
    16124,
    -1,
    16125,
    -1,
    -1,
    5750,
    5752,
    8061,
    1529,
    -1,
    -1,
    2287,
    4341,
    -1,
    1720,
    -1,
    3825,
    -1,
    16126,
    -1,
    -1,
    -1,
    4200,
    -1,
    -1,
    1695,
    2846,
    5754,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16127,
    16128,
    8063,
    -1,
    -1,
    5759,
    -1,
    8064,
    2760,
    8065,
    16129,
    -1,
    4151,
    16130,
    5756,
    -1,
    -1,
    16131,
    -1,
    5751,
    8066,
    5757,
    -1,
    -1,
    16132,
    4291,
    -1,
    3466,
    3511,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8067,
    16138,
    8068,
    -1,
    16139,
    1713,
    -1,
    -1,
    1556,
    -1,
    16140,
    5763,
    -1,
    -1,
    5766,
    -1,
    -1,
    -1,
    4264,
    -1,
    -1,
    8069,
    4203,
    -1,
    16141,
    -1,
    8070,
    5773,
    5767,
    5764,
    16143,
    -1,
    -1,
    4281,
    5770,
    -1,
    16144,
    16145,
    16146,
    2673,
    5772,
    -1,
    -1,
    5777,
    8071,
    -1,
    3464,
    5775,
    8072,
    8073,
    -1,
    8074,
    5769,
    16147,
    -1,
    3219,
    -1,
    -1,
    5776,
    -1,
    5765,
    -1,
    5771,
    16148,
    5778,
    1501,
    5768,
    -1,
    5779,
    -1,
    -1,
    2847,
    -1,
    2694,
    -1,
    3858,
    16149,
    2363,
    -1,
    5760,
    5774,
    3409,
    8075,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    2935,
    -1,
    1764,
    2398,
    2761,
    5762,
    2665,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    2111,
    -1,
    5761,
    -1,
    16134,
    -1,
    5783,
    2708,
    2187,
    -1,
    5798,
    16151,
    5792,
    1963,
    3430,
    16152,
    5787,
    5796,
    1424,
    1521,
    16153,
    -1,
    1622,
    -1,
    5789,
    3113,
    5780,
    5782,
    2288,
    -1,
    -1,
    16155,
    -1,
    8077,
    -1,
    16150,
    -1,
    5799,
    -1,
    5794,
    -1,
    16156,
    -1,
    5786,
    -1,
    -1,
    -1,
    -1,
    5793,
    8078,
    -1,
    -1,
    -1,
    16157,
    16158,
    4052,
    16159,
    16160,
    5791,
    5795,
    16161,
    -1,
    16162,
    -1,
    16163,
    16164,
    -1,
    2209,
    16165,
    2762,
    -1,
    -1,
    3220,
    8079,
    -1,
    8080,
    5785,
    -1,
    -1,
    -1,
    16166,
    -1,
    -1,
    -1,
    4150,
    -1,
    -1,
    -1,
    5788,
    -1,
    -1,
    5781,
    3465,
    -1,
    -1,
    5784,
    -1,
    -1,
    -1,
    5790,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    4372,
    2567,
    4041,
    -1,
    5800,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    3679,
    -1,
    -1,
    5812,
    2188,
    -1,
    -1,
    16168,
    -1,
    -1,
    2695,
    -1,
    5802,
    -1,
    -1,
    -1,
    4244,
    2289,
    -1,
    5815,
    -1,
    -1,
    1487,
    -1,
    -1,
    5813,
    -1,
    16169,
    -1,
    -1,
    5801,
    8081,
    -1,
    -1,
    -1,
    5807,
    -1,
    8082,
    5809,
    -1,
    16170,
    -1,
    4182,
    5804,
    -1,
    -1,
    3397,
    -1,
    -1,
    5806,
    -1,
    16171,
    -1,
    8083,
    5814,
    16172,
    5808,
    4080,
    -1,
    8084,
    -1,
    5803,
    16173,
    2539,
    5827,
    -1,
    8085,
    -1,
    -1,
    1765,
    -1,
    5805,
    5810,
    5811,
    -1,
    -1,
    -1,
    16174,
    -1,
    -1,
    -1,
    3177,
    3159,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16176,
    5819,
    -1,
    -1,
    5825,
    -1,
    -1,
    5823,
    -1,
    3392,
    -1,
    -1,
    5830,
    5820,
    16177,
    -1,
    16178,
    -1,
    -1,
    5821,
    5797,
    -1,
    1968,
    3785,
    -1,
    -1,
    -1,
    2568,
    -1,
    -1,
    2342,
    16179,
    -1,
    16180,
    -1,
    -1,
    4342,
    8086,
    5817,
    -1,
    5829,
    1576,
    3077,
    -1,
    -1,
    16181,
    -1,
    8087,
    -1,
    -1,
    -1,
    -1,
    -1,
    3660,
    -1,
    1818,
    4319,
    -1,
    16182,
    -1,
    -1,
    -1,
    -1,
    8089,
    4042,
    3339,
    -1,
    -1,
    8090,
    -1,
    5824,
    5826,
    8091,
    -1,
    16175,
    16183,
    -1,
    3025,
    -1,
    -1,
    -1,
    16184,
    -1,
    5828,
    5822,
    -1,
    5816,
    -1,
    -1,
    -1,
    1820,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8092,
    16186,
    -1,
    2140,
    -1,
    -1,
    16187,
    5842,
    8093,
    16188,
    5837,
    3001,
    -1,
    8094,
    1758,
    16189,
    8095,
    8096,
    -1,
    2696,
    -1,
    5846,
    -1,
    16190,
    -1,
    -1,
    -1,
    -1,
    5839,
    3301,
    5836,
    3346,
    -1,
    -1,
    -1,
    5871,
    -1,
    -1,
    -1,
    5833,
    -1,
    5832,
    -1,
    5841,
    16192,
    8097,
    -1,
    5835,
    5834,
    5840,
    -1,
    2914,
    -1,
    5831,
    -1,
    8098,
    -1,
    -1,
    -1,
    8099,
    8100,
    5843,
    -1,
    16193,
    5844,
    -1,
    -1,
    8101,
    -1,
    16194,
    1819,
    16191,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16196,
    8102,
    5849,
    -1,
    5848,
    5850,
    16197,
    -1,
    8103,
    -1,
    -1,
    5853,
    -1,
    -1,
    -1,
    -1,
    16198,
    -1,
    3418,
    -1,
    5847,
    -1,
    16200,
    8104,
    -1,
    -1,
    5851,
    -1,
    -1,
    8105,
    -1,
    16195,
    -1,
    2134,
    3190,
    5845,
    3598,
    -1,
    -1,
    5852,
    8106,
    16201,
    16202,
    8107,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    5857,
    5855,
    -1,
    -1,
    5858,
    -1,
    16203,
    5861,
    -1,
    -1,
    16204,
    5854,
    2329,
    3579,
    -1,
    -1,
    5758,
    -1,
    -1,
    -1,
    -1,
    16205,
    -1,
    4219,
    5856,
    -1,
    5860,
    3186,
    8109,
    5859,
    -1,
    5838,
    -1,
    8110,
    7404,
    -1,
    -1,
    8108,
    5864,
    -1,
    16207,
    -1,
    5868,
    -1,
    16208,
    5866,
    -1,
    -1,
    -1,
    8111,
    8112,
    16209,
    -1,
    5862,
    16210,
    5863,
    -1,
    -1,
    -1,
    5867,
    -1,
    5865,
    -1,
    -1,
    -1,
    3805,
    -1,
    -1,
    5873,
    -1,
    5870,
    5869,
    -1,
    5872,
    3529,
    5874,
    -1,
    -1,
    -1,
    16211,
    -1,
    -1,
    3280,
    3178,
    8113,
    -1,
    -1,
    -1,
    2918,
    -1,
    -1,
    -1,
    5875,
    -1,
    5877,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16213,
    16214,
    -1,
    16215,
    -1,
    5876,
    -1,
    -1,
    -1,
    -1,
    16216,
    -1,
    -1,
    -1,
    16217,
    -1,
    -1,
    8114,
    16218,
    5818,
    -1,
    8117,
    -1,
    -1,
    5878,
    -1,
    -1,
    16220,
    -1,
    -1,
    -1,
    3548,
    -1,
    -1,
    -1,
    -1,
    8115,
    8116,
    -1,
    -1,
    -1,
    -1,
    5879,
    8118,
    16221,
    -1,
    -1,
    -1,
    16222,
    -1,
    1645,
    16223,
    -1,
    16224,
    3467,
    1696,
    -1,
    -1,
    -1,
    -1,
    8119,
    16225,
    -1,
    1947,
    -1,
    -1,
    -1,
    2599,
    2399,
    16226,
    -1,
    -1,
    16227,
    -1,
    -1,
    -1,
    8120,
    16228,
    -1,
    -1,
    4330,
    2889,
    -1,
    -1,
    -1,
    1577,
    -1,
    -1,
    -1,
    5881,
    -1,
    -1,
    16229,
    -1,
    16230,
    -1,
    5880,
    -1,
    -1,
    -1,
    -1,
    -1,
    16233,
    -1,
    -1,
    -1,
    -1,
    8121,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8122,
    5884,
    3221,
    5887,
    5882,
    -1,
    16234,
    -1,
    5886,
    -1,
    -1,
    -1,
    8123,
    5885,
    3415,
    1466,
    16231,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    4314,
    -1,
    16237,
    5889,
    -1,
    -1,
    -1,
    1511,
    -1,
    16238,
    -1,
    16239,
    8124,
    -1,
    -1,
    -1,
    8125,
    5891,
    -1,
    -1,
    16240,
    5890,
    -1,
    5888,
    -1,
    -1,
    -1,
    -1,
    8126,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16236,
    -1,
    -1,
    -1,
    -1,
    5883,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    3950,
    -1,
    -1,
    -1,
    5893,
    -1,
    -1,
    -1,
    -1,
    -1,
    16241,
    16242,
    -1,
    16243,
    16244,
    16245,
    5892,
    -1,
    -1,
    16246,
    -1,
    -1,
    8127,
    -1,
    -1,
    -1,
    -1,
    1578,
    -1,
    -1,
    -1,
    -1,
    5895,
    3871,
    -1,
    5894,
    -1,
    8129,
    -1,
    8130,
    4062,
    -1,
    -1,
    -1,
    -1,
    2764,
    -1,
    -1,
    -1,
    -1,
    8128,
    -1,
    -1,
    8131,
    16248,
    8132,
    16249,
    -1,
    -1,
    -1,
    -1,
    3026,
    -1,
    -1,
    -1,
    -1,
    -1,
    2763,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8133,
    8134,
    -1,
    4320,
    16251,
    -1,
    5901,
    -1,
    3002,
    -1,
    16250,
    8135,
    8137,
    16252,
    -1,
    5897,
    5902,
    -1,
    -1,
    1579,
    -1,
    -1,
    8138,
    -1,
    16253,
    -1,
    8139,
    -1,
    5900,
    -1,
    3636,
    5896,
    5899,
    2765,
    8140,
    3714,
    -1,
    -1,
    5903,
    -1,
    2586,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    3003,
    -1,
    -1,
    16256,
    -1,
    -1,
    -1,
    5906,
    8142,
    -1,
    8143,
    5898,
    -1,
    2074,
    -1,
    -1,
    -1,
    -1,
    5904,
    -1,
    -1,
    8144,
    -1,
    4183,
    5907,
    16255,
    -1,
    -1,
    7807,
    -1,
    16257,
    -1,
    -1,
    -1,
    2678,
    16258,
    -1,
    16259,
    -1,
    -1,
    -1,
    -1,
    -1,
    5908,
    -1,
    -1,
    -1,
    5909,
    -1,
    16260,
    16261,
    -1,
    3585,
    -1,
    16262,
    -1,
    -1,
    -1,
    -1,
    -1,
    5911,
    8146,
    -1,
    -1,
    -1,
    5912,
    -1,
    -1,
    8145,
    -1,
    3590,
    8147,
    -1,
    -1,
    -1,
    3468,
    5914,
    -1,
    16264,
    -1,
    -1,
    5916,
    -1,
    4282,
    -1,
    5913,
    16265,
    5915,
    1580,
    -1,
    5910,
    -1,
    16266,
    -1,
    -1,
    16267,
    -1,
    -1,
    4758,
    5917,
    -1,
    -1,
    -1,
    -1,
    3078,
    2467,
    5919,
    -1,
    -1,
    -1,
    -1,
    5918,
    2823,
    4662,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    5920,
    -1,
    -1,
    -1,
    5922,
    -1,
    5905,
    5921,
    -1,
    8148,
    5923,
    8149,
    -1,
    -1,
    -1,
    -1,
    -1,
    3661,
    16268,
    -1,
    -1,
    -1,
    -1,
    -1,
    5924,
    -1,
    -1,
    5925,
    -1,
    -1,
    -1,
    -1,
    8150,
    -1,
    -1,
    -1,
    -1,
    -1,
    5926,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    5927,
    -1,
    3351,
    16271,
    5929,
    5928,
    -1,
    -1,
    5930,
    -1,
    5931,
    -1,
    16272,
    2600,
    3826,
    -1,
    16273,
    16274,
    4117,
    5932,
    5933,
    3061,
    2541,
    5934,
    5935,
    -1,
    16276,
    -1,
    -1,
    -1,
    5936,
    3903,
    3704,
    -1,
    -1,
    5937,
    3626,
    -1,
    -1,
    -1,
    -1,
    -1,
    3302,
    16277,
    -1,
    8151,
    8152,
    16278,
    5938,
    1668,
    -1,
    1956,
    -1,
    4079,
    -1,
    4063,
    -1,
    1617,
    4343,
    16279,
    -1,
    -1,
    -1,
    3997,
    -1,
    3863,
    -1,
    -1,
    -1,
    -1,
    16281,
    16282,
    -1,
    -1,
    2936,
    -1,
    5939,
    -1,
    -1,
    -1,
    16283,
    3512,
    -1,
    -1,
    -1,
    2164,
    5940,
    16284,
    2401,
    5942,
    5941,
    -1,
    -1,
    -1,
    -1,
    5943,
    -1,
    -1,
    -1,
    -1,
    -1,
    8154,
    16285,
    -1,
    -1,
    -1,
    5944,
    -1,
    -1,
    -1,
    5945,
    -1,
    -1,
    -1,
    -1,
    8155,
    -1,
    -1,
    -1,
    -1,
    1909,
    -1,
    5946,
    -1,
    -1,
    -1,
    -1,
    5947,
    -1,
    -1,
    -1,
    -1,
    2165,
    16287,
    16288,
    3705,
    16289,
    16290,
    5949,
    -1,
    -1,
    -1,
    2810,
    -1,
    -1,
    5948,
    -1,
    -1,
    -1,
    -1,
    8156,
    -1,
    8157,
    16291,
    1997,
    5950,
    5952,
    -1,
    5951,
    -1,
    -1,
    -1,
    -1,
    -1,
    16293,
    -1,
    5953,
    -1,
    2210,
    -1,
    5954,
    -1,
    -1,
    -1,
    -1,
    2046,
    -1,
    3039,
    -1,
    2348,
    -1,
    -1,
    -1,
    -1,
    5956,
    5957,
    5955,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    2614,
    -1,
    -1,
    3517,
    1998,
    -1,
    -1,
    -1,
    -1,
    -1,
    16296,
    -1,
    -1,
    -1,
    5959,
    3207,
    5958,
    16297,
    8158,
    4344,
    3637,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16298,
    -1,
    -1,
    5962,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16299,
    -1,
    -1,
    -1,
    5964,
    5961,
    16300,
    16301,
    -1,
    4093,
    5963,
    5965,
    -1,
    4265,
    -1,
    -1,
    -1,
    -1,
    -1,
    5969,
    -1,
    8160,
    8161,
    5968,
    3281,
    3584,
    16304,
    -1,
    2166,
    5967,
    -1,
    16305,
    -1,
    16306,
    5966,
    -1,
    4152,
    4153,
    -1,
    16303,
    -1,
    -1,
    -1,
    16307,
    5970,
    1581,
    -1,
    -1,
    -1,
    -1,
    2341,
    2509,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    5971,
    5972,
    8163,
    -1,
    16308,
    -1,
    -1,
    -1,
    -1,
    5974,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    2666,
    -1,
    -1,
    8164,
    -1,
    5976,
    -1,
    5975,
    16310,
    16311,
    -1,
    16313,
    16314,
    5977,
    16315,
    1736,
    -1,
    -1,
    5979,
    -1,
    16316,
    5978,
    -1,
    5981,
    5980,
    8165,
    -1,
    -1,
    -1,
    -1,
    16317,
    -1,
    -1,
    2189,
    16318,
    -1,
    4237,
    -1,
    2017,
    16319,
    1603,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16320,
    8166,
    2047,
    -1,
    16321,
    -1,
    -1,
    -1,
    16322,
    -1,
    16323,
    8167,
    8168,
    -1,
    8169,
    -1,
    -1,
    16324,
    8170,
    -1,
    16325,
    1850,
    -1,
    8171,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    4301,
    5983,
    -1,
    16326,
    -1,
    16327,
    -1,
    16328,
    -1,
    5985,
    16329,
    -1,
    -1,
    16330,
    5986,
    -1,
    1646,
    -1,
    -1,
    16331,
    -1,
    -1,
    5982,
    8172,
    2468,
    16332,
    -1,
    3322,
    5984,
    8173,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8174,
    -1,
    -1,
    8175,
    -1,
    -1,
    -1,
    -1,
    5989,
    -1,
    2615,
    16333,
    -1,
    8176,
    -1,
    5987,
    -1,
    16334,
    -1,
    8177,
    2097,
    -1,
    -1,
    3706,
    5988,
    -1,
    -1,
    6015,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    5994,
    16335,
    16337,
    -1,
    -1,
    -1,
    2190,
    -1,
    -1,
    16338,
    -1,
    1948,
    -1,
    5991,
    4228,
    8178,
    -1,
    4245,
    8179,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16336,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8180,
    8181,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    3187,
    -1,
    16339,
    5993,
    8183,
    -1,
    8184,
    -1,
    8185,
    8186,
    8187,
    -1,
    8188,
    8189,
    8190,
    16340,
    5995,
    4283,
    2028,
    3757,
    3613,
    -1,
    -1,
    16341,
    5996,
    -1,
    -1,
    -1,
    -1,
    5998,
    16343,
    6001,
    -1,
    16344,
    8191,
    -1,
    8192,
    8193,
    -1,
    -1,
    -1,
    8194,
    -1,
    16345,
    -1,
    -1,
    -1,
    -1,
    16346,
    16342,
    -1,
    5997,
    -1,
    8195,
    -1,
    6000,
    2235,
    1548,
    6002,
    16347,
    2899,
    5999,
    4289,
    -1,
    8196,
    6005,
    7805,
    -1,
    -1,
    -1,
    -1,
    6003,
    6006,
    8197,
    -1,
    8198,
    -1,
    5992,
    6004,
    16348,
    -1,
    2377,
    -1,
    -1,
    6007,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    6008,
    -1,
    -1,
    16349,
    -1,
    4229,
    -1,
    16350,
    8199,
    8200,
    16351,
    8201,
    -1,
    6009,
    -1,
    -1,
    -1,
    -1,
    8206,
    -1,
    16353,
    -1,
    -1,
    -1,
    -1,
    16354,
    8202,
    16355,
    -1,
    -1,
    8203,
    -1,
    6010,
    8204,
    16356,
    16357,
    5990,
    8205,
    -1,
    16358,
    8207,
    6011,
    8208,
    8209,
    16359,
    16360,
    -1,
    -1,
    -1,
    -1,
    1821,
    -1,
    -1,
    -1,
    -1,
    8210,
    -1,
    -1,
    -1,
    16361,
    16363,
    16362,
    -1,
    2542,
    -1,
    8211,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8212,
    8213,
    6012,
    -1,
    -1,
    -1,
    -1,
    6013,
    -1,
    -1,
    -1,
    -1,
    6014,
    -1,
    16364,
    -1,
    16365,
    -1,
    8214,
    -1,
    1530,
    -1,
    16366,
    -1,
    6016,
    -1,
    3786,
    6017,
    -1,
    -1,
    1789,
    6018,
    -1,
    6019,
    -1,
    16368,
    -1,
    -1,
    6020,
    16367,
    6022,
    6023,
    6021,
    -1,
    -1,
    -1,
    3810,
    6025,
    6024,
    -1,
    16370,
    -1,
    -1,
    -1,
    -1,
    8215,
    -1,
    8216,
    -1,
    6027,
    6026,
    6028,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    6029,
    6031,
    6030,
    -1,
    -1,
    2344,
    -1,
    6033,
    -1,
    6032,
    -1,
    8217,
    1822,
    -1,
    2870,
    -1,
    3411,
    -1,
    6034,
    2937,
    16372,
    -1,
    -1,
    2469,
    16373,
    1592,
    6035,
    -1,
    4184,
    -1,
    16374,
    3920,
    6036,
    -1,
    -1,
    8218,
    3419,
    4154,
    2290,
    2848,
    -1,
    -1,
    -1,
    3238,
    4625,
    -1,
    3303,
    1669,
    6037,
    16377,
    16378,
    -1,
    16379,
    -1,
    -1,
    -1,
    6038,
    -1,
    6043,
    -1,
    16380,
    6041,
    6040,
    5391,
    1697,
    6039,
    16381,
    1467,
    16382,
    3675,
    16383,
    -1,
    3707,
    -1,
    -1,
    -1,
    -1,
    4246,
    6044,
    6042,
    3255,
    2919,
    -1,
    -1,
    3676,
    -1,
    3772,
    -1,
    6046,
    4241,
    2112,
    6047,
    -1,
    6045,
    3719,
    6048,
    16384,
    6049,
    -1,
    8219,
    1468,
    16386,
    16385,
    2811,
    6054,
    -1,
    6051,
    3554,
    6050,
    8220,
    16387,
    -1,
    -1,
    16388,
    16389,
    1885,
    -1,
    16390,
    6057,
    -1,
    -1,
    -1,
    6052,
    6053,
    -1,
    6056,
    6055,
    3764,
    16392,
    -1,
    3041,
    3040,
    -1,
    1910,
    8221,
    -1,
    6058,
    -1,
    -1,
    -1,
    -1,
    -1,
    6059,
    -1,
    -1,
    6060,
    -1,
    -1,
    -1,
    -1,
    16394,
    6062,
    -1,
    6061,
    -1,
    -1,
    -1,
    -1,
    -1,
    1557,
    -1,
    -1,
    -1,
    -1,
    16396,
    6070,
    3737,
    6064,
    -1,
    6066,
    -1,
    16397,
    6068,
    2849,
    -1,
    -1,
    6069,
    6067,
    2569,
    16398,
    16399,
    -1,
    6063,
    6065,
    -1,
    3793,
    16400,
    2766,
    -1,
    -1,
    6072,
    -1,
    -1,
    6071,
    8223,
    16401,
    -1,
    -1,
    6073,
    16402,
    2543,
    2364,
    -1,
    -1,
    3470,
    6074,
    -1,
    3332,
    -1,
    16403,
    6076,
    16404,
    16405,
    -1,
    4230,
    6075,
    8224,
    -1,
    -1,
    16406,
    -1,
    3080,
    -1,
    -1,
    16407,
    -1,
    16408,
    -1,
    6081,
    16409,
    6083,
    6084,
    3247,
    -1,
    -1,
    -1,
    -1,
    16410,
    6082,
    -1,
    6079,
    -1,
    6077,
    6078,
    8225,
    6080,
    8226,
    16411,
    -1,
    -1,
    -1,
    16413,
    8227,
    6087,
    -1,
    6085,
    -1,
    6086,
    -1,
    16414,
    -1,
    -1,
    -1,
    16417,
    -1,
    8228,
    8229,
    -1,
    16412,
    8230,
    -1,
    16418,
    16419,
    -1,
    8231,
    6088,
    6090,
    6091,
    6092,
    -1,
    6093,
    16421,
    8834,
    6089,
    16422,
    -1,
    -1,
    -1,
    -1,
    8232,
    -1,
    -1,
    6095,
    -1,
    -1,
    16424,
    6094,
    8233,
    -1,
    -1,
    -1,
    -1,
    -1,
    6096,
    16423,
    -1,
    -1,
    -1,
    -1,
    16427,
    4266,
    8234,
    -1,
    -1,
    6099,
    6097,
    6098,
    16429,
    -1,
    8235,
    1851,
    -1,
    -1,
    -1,
    -1,
    -1,
    4135,
    -1,
    -1,
    16430,
    3895,
    -1,
    6101,
    -1,
    -1,
    -1,
    6100,
    -1,
    -1,
    -1,
    -1,
    6102,
    6103,
    -1,
    8236,
    8237,
    -1,
    6107,
    6104,
    6105,
    6106,
    -1,
    6108,
    8238,
    16432,
    8239,
    6109,
    8240,
    6110,
    -1,
    -1,
    -1,
    6111,
    -1,
    6112,
    -1,
    3680,
    3431,
    6113,
    3653,
    3779,
    -1,
    6114,
    8241,
    -1,
    6115,
    3393,
    -1,
    1698,
    2291,
    6116,
    -1,
    -1,
    6117,
    -1,
    -1,
    6118,
    -1,
    2452,
    -1,
    -1,
    6120,
    -1,
    16433,
    6119,
    -1,
    -1,
    6121,
    6122,
    8242,
    16434,
    8243,
    8244,
    -1,
    16436,
    16435,
    -1,
    -1,
    -1,
    -1,
    8245,
    16437,
    16438,
    -1,
    8246,
    -1,
    -1,
    -1,
    3738,
    16439,
    6123,
    -1,
    -1,
    -1,
    6124,
    -1,
    8247,
    7780,
    6125,
    6126,
    6127,
    -1,
    -1,
    -1,
    -1,
    2457,
    -1,
    -1,
    6128,
    3625,
    -1,
    8248,
    4010,
    -1,
    1549,
    16441,
    1558,
    -1,
    8249,
    6129,
    8250,
    -1,
    -1,
    -1,
    6131,
    -1,
    8251,
    -1,
    6130,
    3463,
    -1,
    -1,
    -1,
    2938,
    5664,
    -1,
    6132,
    4074,
    -1,
    6133,
    -1,
    1823,
    3720,
    6134,
    8252,
    6135,
    16443,
    -1,
    6136,
    -1,
    16444,
    -1,
    4101,
    -1,
    -1,
    8253,
    4094,
    -1,
    3319,
    -1,
    -1,
    -1,
    3081,
    -1,
    -1,
    6138,
    8254,
    -1,
    2697,
    -1,
    -1,
    2767,
    -1,
    -1,
    6141,
    -1,
    -1,
    6140,
    6139,
    3758,
    8255,
    1824,
    2170,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16446,
    -1,
    8256,
    16447,
    6147,
    -1,
    -1,
    6144,
    2850,
    4059,
    -1,
    -1,
    -1,
    6143,
    6145,
    6146,
    -1,
    -1,
    6142,
    -1,
    -1,
    -1,
    16448,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8257,
    16449,
    8258,
    6148,
    6149,
    -1,
    3304,
    -1,
    1852,
    -1,
    -1,
    -1,
    3264,
    -1,
    -1,
    -1,
    -1,
    -1,
    8259,
    6150,
    -1,
    -1,
    -1,
    -1,
    -1,
    8260,
    8261,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16454,
    -1,
    6151,
    6154,
    8262,
    -1,
    -1,
    8263,
    16455,
    2890,
    8264,
    3513,
    -1,
    6155,
    3998,
    -1,
    6152,
    -1,
    16456,
    6153,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16458,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    6158,
    8265,
    -1,
    16459,
    16460,
    6157,
    6156,
    8266,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    6160,
    -1,
    -1,
    6159,
    -1,
    -1,
    6161,
    -1,
    -1,
    8267,
    -1,
    -1,
    -1,
    -1,
    -1,
    16463,
    -1,
    -1,
    -1,
    6163,
    16464,
    6162,
    -1,
    16465,
    -1,
    16466,
    3898,
    -1,
    -1,
    -1,
    16467,
    8268,
    -1,
    2685,
    4267,
    -1,
    -1,
    6164,
    -1,
    -1,
    3500,
    -1,
    -1,
    6165,
    -1,
    -1,
    6166,
    -1,
    6170,
    6168,
    6169,
    -1,
    6167,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    6171,
    -1,
    -1,
    -1,
    -1,
    -1,
    6172,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    6173,
    -1,
    -1,
    6174,
    4064,
    6175,
    -1,
    16468,
    16469,
    8269,
    -1,
    4121,
    6176,
    16470,
    3241,
    16471,
    3645,
    -1,
    2048,
    16472,
    -1,
    16473,
    3222,
    6177,
    1999,
    16475,
    -1,
    -1,
    2961,
    16476,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16477,
    6178,
    -1,
    -1,
    -1,
    -1,
    -1,
    2378,
    -1,
    -1,
    16479,
    16480,
    -1,
    -1,
    16481,
    -1,
    -1,
    6179,
    16482,
    -1,
    -1,
    -1,
    -1,
    6180,
    -1,
    2167,
    2402,
    -1,
    -1,
    -1,
    16483,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    6182,
    16484,
    -1,
    -1,
    -1,
    3437,
    2403,
    1925,
    -1,
    -1,
    -1,
    -1,
    16485,
    8270,
    -1,
    -1,
    -1,
    -1,
    3951,
    -1,
    3614,
    -1,
    -1,
    -1,
    -1,
    -1,
    3438,
    -1,
    -1,
    -1,
    -1,
    2315,
    -1,
    -1,
    -1,
    8272,
    -1,
    6184,
    -1,
    16486,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8273,
    8274,
    -1,
    8275,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    2768,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16487,
    -1,
    -1,
    -1,
    8276,
    -1,
    16488,
    4247,
    2292,
    -1,
    8277,
    2168,
    -1,
    -1,
    3668,
    -1,
    6186,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    6188,
    -1,
    -1,
    -1,
    -1,
    2236,
    -1,
    -1,
    -1,
    -1,
    6187,
    3375,
    -1,
    -1,
    16490,
    -1,
    6190,
    1714,
    6185,
    -1,
    -1,
    3739,
    -1,
    1519,
    16491,
    2425,
    -1,
    4373,
    -1,
    -1,
    6189,
    -1,
    -1,
    16493,
    16494,
    16495,
    -1,
    -1,
    -1,
    6191,
    16492,
    -1,
    -1,
    3896,
    -1,
    2970,
    6193,
    -1,
    -1,
    8281,
    -1,
    6194,
    8279,
    -1,
    -1,
    -1,
    -1,
    6192,
    -1,
    -1,
    -1,
    -1,
    1737,
    16496,
    6200,
    -1,
    6199,
    -1,
    -1,
    2544,
    -1,
    -1,
    -1,
    6201,
    6196,
    -1,
    16497,
    -1,
    6202,
    6197,
    16498,
    -1,
    16499,
    -1,
    3721,
    6195,
    -1,
    -1,
    6198,
    16500,
    -1,
    -1,
    -1,
    -1,
    6205,
    -1,
    -1,
    -1,
    -1,
    -1,
    16501,
    16502,
    -1,
    -1,
    8282,
    -1,
    16503,
    6204,
    4012,
    -1,
    -1,
    -1,
    6203,
    -1,
    -1,
    1484,
    -1,
    -1,
    8283,
    -1,
    6207,
    -1,
    -1,
    8285,
    -1,
    16504,
    16505,
    16506,
    -1,
    6206,
    16507,
    -1,
    8284,
    2769,
    -1,
    -1,
    -1,
    -1,
    -1,
    6208,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    3042,
    -1,
    16509,
    6210,
    6209,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    6211,
    -1,
    16510,
    8286,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16512,
    6181,
    -1,
    -1,
    -1,
    6183,
    6213,
    6212,
    -1,
    8287,
    -1,
    16511,
    8288,
    -1,
    -1,
    8289,
    -1,
    -1,
    -1,
    -1,
    -1,
    2545,
    16513,
    4302,
    -1,
    2587,
    -1,
    6214,
    2090,
    -1,
    -1,
    -1,
    8292,
    8293,
    1911,
    1886,
    2510,
    16514,
    -1,
    -1,
    -1,
    -1,
    -1,
    4155,
    -1,
    -1,
    6220,
    -1,
    6219,
    3043,
    6216,
    16515,
    -1,
    6218,
    16516,
    8297,
    2674,
    2851,
    6217,
    6215,
    -1,
    3581,
    -1,
    -1,
    2770,
    -1,
    16518,
    3787,
    -1,
    -1,
    -1,
    -1,
    2404,
    -1,
    -1,
    -1,
    -1,
    16519,
    -1,
    -1,
    -1,
    -1,
    3471,
    -1,
    8301,
    6221,
    -1,
    -1,
    -1,
    -1,
    6222,
    6244,
    2029,
    -1,
    -1,
    4352,
    3028,
    -1,
    -1,
    -1,
    -1,
    6223,
    -1,
    -1,
    1647,
    3376,
    3854,
    -1,
    -1,
    -1,
    -1,
    16520,
    16521,
    16522,
    -1,
    8305,
    -1,
    -1,
    16523,
    -1,
    6224,
    -1,
    -1,
    -1,
    16524,
    -1,
    -1,
    -1,
    -1,
    1969,
    6225,
    -1,
    16525,
    6227,
    -1,
    -1,
    -1,
    6228,
    -1,
    3580,
    8306,
    -1,
    6229,
    16526,
    -1,
    -1,
    -1,
    8307,
    6230,
    6231,
    16527,
    -1,
    2030,
    1648,
    3514,
    2639,
    2511,
    16528,
    -1,
    -1,
    -1,
    -1,
    16529,
    8308,
    6232,
    8309,
    2640,
    16530,
    16531,
    -1,
    -1,
    -1,
    1639,
    3794,
    -1,
    8311,
    6233,
    16532,
    -1,
    3740,
    -1,
    -1,
    -1,
    -1,
    -1,
    8312,
    3044,
    -1,
    6236,
    -1,
    6237,
    3644,
    -1,
    2852,
    6234,
    -1,
    3260,
    -1,
    8313,
    6235,
    8314,
    -1,
    -1,
    2771,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1469,
    -1,
    -1,
    -1,
    -1,
    1888,
    -1,
    -1,
    8315,
    -1,
    -1,
    -1,
    -1,
    6238,
    -1,
    16537,
    3377,
    -1,
    6239,
    2952,
    -1,
    -1,
    16538,
    -1,
    -1,
    4054,
    16539,
    -1,
    3762,
    6240,
    6241,
    3248,
    16540,
    4268,
    -1,
    16541,
    6243,
    6242,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16543,
    2616,
    -1,
    -1,
    6245,
    1489,
    -1,
    -1,
    -1,
    -1,
    6248,
    16544,
    8317,
    -1,
    6246,
    1649,
    2113,
    6247,
    2293,
    2337,
    -1,
    3923,
    6249,
    -1,
    -1,
    3999,
    16545,
    -1,
    6251,
    -1,
    -1,
    16546,
    2962,
    1550,
    1623,
    1421,
    -1,
    -1,
    -1,
    -1,
    -1,
    16547,
    6250,
    -1,
    16548,
    -1,
    -1,
    16549,
    8318,
    -1,
    16550,
    16551,
    6252,
    6253,
    2812,
    -1,
    -1,
    -1,
    16552,
    -1,
    6254,
    16553,
    1738,
    -1,
    8319,
    -1,
    -1,
    6256,
    -1,
    -1,
    -1,
    2141,
    16554,
    1949,
    -1,
    16555,
    6257,
    2058,
    -1,
    -1,
    6258,
    -1,
    3005,
    -1,
    3522,
    16556,
    2977,
    2434,
    8321,
    -1,
    -1,
    6259,
    -1,
    16557,
    -1,
    -1,
    -1,
    -1,
    -1,
    16558,
    -1,
    3261,
    3082,
    -1,
    6261,
    6263,
    6260,
    6262,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    2069,
    8322,
    -1,
    -1,
    16559,
    -1,
    -1,
    -1,
    -1,
    -1,
    6264,
    2073,
    -1,
    16560,
    -1,
    1950,
    4185,
    6266,
    -1,
    -1,
    8324,
    -1,
    -1,
    6267,
    -1,
    -1,
    16563,
    1517,
    8325,
    16564,
    -1,
    16565,
    6270,
    -1,
    -1,
    -1,
    1777,
    6269,
    6268,
    -1,
    6272,
    6265,
    -1,
    6273,
    4238,
    16567,
    6274,
    8326,
    6275,
    -1,
    16568,
    4877,
    6277,
    -1,
    6276,
    -1,
    -1,
    -1,
    6278,
    6279,
    -1,
    4250,
    6280,
    -1,
    7504,
    2772,
    6281,
    6282,
    2686,
    -1,
    3501,
    6283,
    16569,
    16570,
    -1,
    3203,
    8327,
    -1,
    6284,
    -1,
    3223,
    6285,
    -1,
    -1,
    -1,
    16571,
    -1,
    1978,
    -1,
    4542,
    3256,
    2556,
    -1,
    -1,
    8328,
    -1,
    1825,
    -1,
    -1,
    6286,
    -1,
    6299,
    -1,
    6289,
    16574,
    1951,
    -1,
    6288,
    6301,
    -1,
    -1,
    -1,
    6287,
    -1,
    2773,
    8329,
    -1,
    -1,
    -1,
    -1,
    -1,
    6291,
    6292,
    -1,
    3394,
    -1,
    -1,
    6293,
    -1,
    1753,
    -1,
    -1,
    -1,
    -1,
    2877,
    3827,
    16577,
    6295,
    -1,
    16578,
    -1,
    3173,
    8330,
    16579,
    16580,
    -1,
    16581,
    -1,
    6290,
    -1,
    6294,
    6296,
    -1,
    -1,
    2440,
    -1,
    8331,
    -1,
    16576,
    -1,
    -1,
    -1,
    16585,
    -1,
    -1,
    -1,
    6303,
    3773,
    8332,
    3671,
    3472,
    -1,
    2031,
    6302,
    6300,
    8333,
    3686,
    6297,
    3257,
    3474,
    -1,
    3473,
    16587,
    2435,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    6321,
    -1,
    -1,
    8334,
    -1,
    -1,
    -1,
    16589,
    6305,
    16590,
    6307,
    -1,
    16591,
    -1,
    -1,
    6310,
    8335,
    6311,
    8336,
    6308,
    6309,
    8337,
    16592,
    6306,
    6304,
    -1,
    -1,
    -1,
    16588,
    6298,
    -1,
    -1,
    -1,
    -1,
    16584,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    3900,
    1650,
    -1,
    -1,
    -1,
    6318,
    -1,
    6315,
    -1,
    6320,
    16595,
    16596,
    6319,
    -1,
    3654,
    4048,
    -1,
    2470,
    6313,
    6322,
    6317,
    16597,
    6316,
    6312,
    8338,
    6314,
    -1,
    1826,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    3224,
    -1,
    -1,
    3006,
    -1,
    16599,
    -1,
    3667,
    -1,
    -1,
    6327,
    16600,
    -1,
    -1,
    3669,
    -1,
    -1,
    -1,
    16601,
    -1,
    -1,
    -1,
    2978,
    6324,
    -1,
    -1,
    3712,
    16602,
    6328,
    3904,
    -1,
    3254,
    16603,
    6323,
    6325,
    -1,
    -1,
    6326,
    -1,
    -1,
    -1,
    -1,
    16606,
    -1,
    16607,
    8340,
    -1,
    8341,
    16608,
    -1,
    -1,
    6329,
    -1,
    -1,
    2573,
    -1,
    -1,
    -1,
    3515,
    6334,
    6333,
    -1,
    -1,
    6330,
    16609,
    -1,
    -1,
    4345,
    -1,
    -1,
    16610,
    -1,
    -1,
    6339,
    -1,
    -1,
    6343,
    6340,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    6336,
    8342,
    -1,
    16611,
    -1,
    -1,
    -1,
    6337,
    -1,
    -1,
    -1,
    16612,
    -1,
    6342,
    16613,
    16614,
    -1,
    6331,
    4658,
    6338,
    6332,
    -1,
    -1,
    6341,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8344,
    6347,
    8345,
    1827,
    -1,
    6344,
    -1,
    -1,
    16615,
    6345,
    -1,
    -1,
    6346,
    6349,
    -1,
    -1,
    -1,
    -1,
    -1,
    8343,
    -1,
    8346,
    -1,
    -1,
    8347,
    6348,
    3751,
    -1,
    -1,
    -1,
    -1,
    6350,
    4321,
    3930,
    6355,
    -1,
    -1,
    6352,
    -1,
    16616,
    -1,
    -1,
    -1,
    -1,
    16617,
    -1,
    6351,
    2963,
    -1,
    6354,
    6356,
    16618,
    -1,
    -1,
    6353,
    -1,
    6360,
    16619,
    6357,
    8350,
    -1,
    -1,
    -1,
    -1,
    16620,
    6358,
    6335,
    16621,
    -1,
    -1,
    6359,
    6361,
    -1,
    -1,
    -1,
    16622,
    -1,
    -1,
    6362,
    8351,
    16623,
    16624,
    16625,
    -1,
    -1,
    3891,
    -1,
    6363,
    -1,
    -1,
    -1,
    8352,
    -1,
    -1,
    -1,
    -1,
    4107,
    -1,
    -1,
    2019,
    2076,
    6364,
    -1,
    -1,
    -1,
    -1,
    -1,
    3873,
    -1,
    2891,
    -1,
    4057,
    -1,
    8353,
    6365,
    -1,
    4248,
    -1,
    8354,
    3655,
    -1,
    3045,
    3591,
    -1,
    -1,
    2676,
    -1,
    -1,
    -1,
    1441,
    8355,
    6370,
    6368,
    -1,
    6366,
    1786,
    16629,
    2774,
    6371,
    -1,
    -1,
    6369,
    -1,
    6367,
    6375,
    -1,
    -1,
    6374,
    6373,
    6372,
    -1,
    -1,
    16631,
    16632,
    -1,
    6376,
    -1,
    -1,
    8356,
    6377,
    2939,
    16633,
    6378,
    -1,
    6380,
    -1,
    16635,
    6379,
    -1,
    -1,
    16637,
    -1,
    2211,
    -1,
    -1,
    16638,
    3031,
    -1,
    -1,
    -1,
    6382,
    -1,
    -1,
    8357,
    3475,
    16640,
    6381,
    8358,
    -1,
    -1,
    6383,
    8359,
    3874,
    3083,
    2294,
    -1,
    6384,
    -1,
    -1,
    -1,
    16642,
    4269,
    -1,
    -1,
    -1,
    16643,
    -1,
    -1,
    -1,
    6386,
    -1,
    -1,
    6387,
    -1,
    6388,
    16645,
    6389,
    -1,
    2512,
    -1,
    6390,
    2114,
    -1,
    -1,
    1953,
    -1,
    1889,
    -1,
    6392,
    16646,
    4124,
    2295,
    6391,
    8360,
    8361,
    16647,
    6395,
    4111,
    -1,
    3599,
    -1,
    -1,
    3778,
    -1,
    16649,
    8362,
    2698,
    6394,
    -1,
    2588,
    2296,
    2513,
    1952,
    3875,
    6393,
    8363,
    16650,
    -1,
    3046,
    3981,
    2436,
    8364,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    2514,
    3350,
    -1,
    6398,
    4292,
    2406,
    8365,
    6399,
    2853,
    -1,
    6401,
    -1,
    -1,
    -1,
    2775,
    2365,
    -1,
    -1,
    16653,
    16654,
    6400,
    16655,
    8366,
    2641,
    2191,
    3047,
    6396,
    6402,
    16656,
    8367,
    -1,
    -1,
    6397,
    2115,
    -1,
    6405,
    6409,
    2142,
    -1,
    -1,
    8368,
    -1,
    -1,
    6404,
    -1,
    -1,
    16660,
    16661,
    6413,
    8369,
    -1,
    2297,
    -1,
    -1,
    4212,
    1437,
    6410,
    -1,
    -1,
    1954,
    -1,
    6407,
    -1,
    16662,
    -1,
    -1,
    -1,
    6408,
    -1,
    16663,
    3476,
    6406,
    6403,
    -1,
    1699,
    2981,
    -1,
    -1,
    2169,
    8370,
    -1,
    -1,
    6415,
    -1,
    16665,
    -1,
    -1,
    -1,
    8371,
    -1,
    -1,
    16667,
    -1,
    16668,
    6412,
    -1,
    8372,
    16669,
    -1,
    -1,
    6414,
    -1,
    -1,
    -1,
    6411,
    -1,
    -1,
    -1,
    16670,
    -1,
    2116,
    3120,
    6416,
    3085,
    16672,
    -1,
    6429,
    8373,
    -1,
    6425,
    6419,
    -1,
    -1,
    8374,
    16673,
    -1,
    -1,
    16674,
    6423,
    2627,
    1470,
    6418,
    6426,
    6430,
    2298,
    4095,
    -1,
    3343,
    6420,
    16675,
    16676,
    6428,
    -1,
    6417,
    3225,
    -1,
    6422,
    1438,
    4083,
    16677,
    -1,
    8375,
    -1,
    -1,
    -1,
    -1,
    6421,
    -1,
    -1,
    2032,
    3741,
    8376,
    -1,
    -1,
    3084,
    -1,
    4277,
    2710,
    -1,
    -1,
    6469,
    8377,
    16678,
    6431,
    16679,
    3007,
    -1,
    6427,
    6432,
    6434,
    -1,
    3378,
    6437,
    -1,
    8378,
    6433,
    -1,
    16680,
    -1,
    3905,
    1828,
    -1,
    -1,
    4084,
    -1,
    -1,
    1471,
    -1,
    16681,
    6436,
    -1,
    4322,
    -1,
    -1,
    -1,
    -1,
    16682,
    -1,
    6435,
    -1,
    -1,
    -1,
    -1,
    -1,
    1582,
    -1,
    -1,
    3553,
    6438,
    -1,
    -1,
    8381,
    6445,
    6439,
    6446,
    -1,
    -1,
    -1,
    -1,
    16686,
    8382,
    6442,
    -1,
    -1,
    8383,
    -1,
    16687,
    -1,
    -1,
    -1,
    3662,
    -1,
    16688,
    2579,
    6444,
    16689,
    6441,
    6447,
    6440,
    -1,
    -1,
    2667,
    16690,
    8380,
    -1,
    -1,
    3952,
    16691,
    -1,
    2675,
    -1,
    -1,
    6443,
    6455,
    -1,
    -1,
    6451,
    -1,
    6454,
    -1,
    6452,
    6456,
    6450,
    -1,
    6424,
    2964,
    -1,
    -1,
    3708,
    -1,
    6453,
    -1,
    16692,
    6448,
    8385,
    -1,
    -1,
    3008,
    2117,
    -1,
    2642,
    -1,
    -1,
    -1,
    -1,
    8386,
    -1,
    2824,
    3029,
    6459,
    -1,
    -1,
    6461,
    6462,
    -1,
    -1,
    6458,
    6460,
    -1,
    -1,
    8387,
    -1,
    -1,
    -1,
    -1,
    6449,
    6457,
    -1,
    6465,
    6464,
    8835,
    -1,
    4037,
    -1,
    -1,
    2078,
    -1,
    -1,
    16693,
    -1,
    16694,
    -1,
    -1,
    -1,
    6463,
    -1,
    6467,
    6466,
    6470,
    16695,
    6472,
    -1,
    -1,
    2471,
    6468,
    -1,
    -1,
    16696,
    16697,
    6473,
    6474,
    8388,
    -1,
    6475,
    8389,
    6481,
    3410,
    6477,
    16699,
    6476,
    6478,
    6479,
    -1,
    6480,
    -1,
    16700,
    -1,
    16701,
    6482,
    6483,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1829,
    -1,
    6484,
    -1,
    6485,
    16704,
    16703,
    -1,
    16705,
    -1,
    -1,
    -1,
    -1,
    16706,
    16707,
    6486,
    -1,
    8390,
    -1,
    -1,
    -1,
    -1,
    6487,
    6488,
    6489,
    16708,
    6490,
    6491,
    16711,
    -1,
    6493,
    6492,
    -1,
    -1,
    6494,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    6495,
    6496,
    16713,
    -1,
    16714,
    16715,
    -1,
    -1,
    6499,
    6497,
    6498,
    2414,
    2118,
    -1,
    16716,
    3249,
    -1,
    3684,
    -1,
    2711,
    -1,
    -1,
    3616,
    -1,
    3742,
    6500,
    5246,
    -1,
    -1,
    -1,
    16717,
    16718,
    -1,
    -1,
    -1,
    6501,
    6503,
    -1,
    4204,
    6502,
    6505,
    6504,
    -1,
    4186,
    -1,
    6506,
    -1,
    3759,
    -1,
    16720,
    8392,
    -1,
    -1,
    6507,
    -1,
    16723,
    8393,
    -1,
    -1,
    6510,
    -1,
    16724,
    6509,
    6508,
    -1,
    -1,
    -1,
    -1,
    6511,
    2085,
    -1,
    -1,
    -1,
    3009,
    1912,
    -1,
    -1,
    -1,
    16725,
    6515,
    6512,
    -1,
    -1,
    6513,
    -1,
    -1,
    -1,
    6516,
    -1,
    6517,
    6514,
    -1,
    -1,
    -1,
    1512,
    -1,
    8394,
    -1,
    1604,
    -1,
    16727,
    -1,
    6519,
    6520,
    -1,
    -1,
    -1,
    6521,
    -1,
    4201,
    -1,
    8395,
    16728,
    -1,
    -1,
    2643,
    -1,
    6523,
    6522,
    -1,
    -1,
    -1,
    -1,
    -1,
    8396,
    -1,
    -1,
    -1,
    8397,
    2892,
    6524,
    -1,
    16729,
    -1,
    16730,
    6525,
    -1,
    -1,
    6526,
    -1,
    1854,
    8398,
    -1,
    8399,
    16731,
    1830,
    -1,
    16732,
    6527,
    -1,
    -1,
    -1,
    -1,
    -1,
    6528,
    8400,
    4008,
    4202,
    -1,
    -1,
    -1,
    4187,
    4346,
    16733,
    2300,
    6531,
    2589,
    6530,
    -1,
    16735,
    -1,
    16734,
    6532,
    2546,
    -1,
    16736,
    -1,
    3150,
    16737,
    6533,
    -1,
    8402,
    2299,
    16738,
    4096,
    6534,
    6535,
    -1,
    -1,
    6536,
    -1,
    -1,
    -1,
    -1,
    6537,
    -1,
    -1,
    16739,
    -1,
    8403,
    -1,
    6538,
    -1,
    -1,
    -1,
    16740,
    -1,
    -1,
    -1,
    16741,
    -1,
    -1,
    2547,
    -1,
    8404,
    4118,
    8405,
    -1,
    -1,
    -1,
    6540,
    8406,
    3226,
    -1,
    6539,
    -1,
    -1,
    -1,
    16742,
    -1,
    -1,
    6542,
    -1,
    -1,
    -1,
    6541,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    6543,
    -1,
    -1,
    -1,
    2940,
    -1,
    6544,
    -1,
    6545,
    -1,
    -1,
    -1,
    3878,
    6546,
    -1,
    3086,
    6547,
    -1,
    -1,
    -1,
    16743,
    -1,
    6548,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    4323,
    6551,
    16744,
    6550,
    6549,
    3305,
    16745,
    6552,
    2825,
    -1,
    6553,
    -1,
    16746,
    -1,
    6554,
    4347,
    6555,
    -1,
    -1,
    -1,
    -1,
    6556,
    6558,
    6557,
    3670,
    -1,
    3566,
    -1,
    4353,
    3674,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    6560,
    -1,
    -1,
    2776,
    -1,
    3769,
    16747,
    6561,
    6559,
    16748,
    1831,
    -1,
    -1,
    -1,
    2213,
    2515,
    -1,
    16749,
    3743,
    -1,
    16750,
    -1,
    2171,
    3982,
    -1,
    6564,
    6562,
    -1,
    2301,
    -1,
    2302,
    1482,
    -1,
    2421,
    -1,
    -1,
    -1,
    16751,
    -1,
    3628,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1472,
    6569,
    16753,
    3227,
    -1,
    -1,
    -1,
    8407,
    -1,
    3627,
    -1,
    3160,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16754,
    6571,
    8408,
    16755,
    6567,
    6570,
    6565,
    -1,
    6568,
    3953,
    -1,
    8409,
    2214,
    -1,
    -1,
    1502,
    6566,
    16756,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    6573,
    -1,
    6574,
    -1,
    8410,
    3502,
    16759,
    -1,
    -1,
    2000,
    -1,
    -1,
    16760,
    6585,
    3600,
    -1,
    -1,
    -1,
    -1,
    2516,
    -1,
    -1,
    2001,
    2953,
    4361,
    4055,
    6572,
    2965,
    -1,
    -1,
    16758,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16763,
    -1,
    8411,
    -1,
    1930,
    6575,
    -1,
    -1,
    16764,
    -1,
    -1,
    -1,
    -1,
    6577,
    16766,
    -1,
    -1,
    16767,
    -1,
    6576,
    -1,
    -1,
    16768,
    -1,
    -1,
    6578,
    -1,
    3201,
    -1,
    3601,
    -1,
    16762,
    -1,
    -1,
    -1,
    3306,
    -1,
    -1,
    -1,
    16770,
    6582,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    6581,
    -1,
    -1,
    -1,
    8412,
    6579,
    -1,
    -1,
    2872,
    -1,
    3828,
    6584,
    -1,
    6583,
    2303,
    4374,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    6600,
    8413,
    -1,
    -1,
    -1,
    -1,
    6588,
    6589,
    8414,
    8415,
    16772,
    -1,
    2617,
    -1,
    8416,
    6587,
    -1,
    2343,
    6586,
    -1,
    -1,
    6590,
    -1,
    -1,
    -1,
    3307,
    3855,
    3010,
    -1,
    -1,
    -1,
    -1,
    3161,
    6594,
    16774,
    6595,
    6591,
    16776,
    16777,
    -1,
    -1,
    6592,
    -1,
    6593,
    -1,
    -1,
    -1,
    -1,
    2304,
    -1,
    -1,
    -1,
    6601,
    -1,
    6597,
    -1,
    -1,
    16779,
    -1,
    3829,
    -1,
    4023,
    3767,
    -1,
    -1,
    6596,
    -1,
    -1,
    6599,
    6598,
    -1,
    -1,
    -1,
    3983,
    6602,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    6603,
    -1,
    16780,
    3030,
    -1,
    6604,
    -1,
    -1,
    6606,
    -1,
    6610,
    8417,
    -1,
    6607,
    6605,
    3602,
    6608,
    16781,
    6609,
    16782,
    -1,
    -1,
    1615,
    -1,
    6616,
    6611,
    8418,
    -1,
    -1,
    6612,
    -1,
    8419,
    -1,
    6613,
    -1,
    3102,
    -1,
    -1,
    16783,
    8420,
    6615,
    6614,
    6617,
    16784,
    -1,
    -1,
    -1,
    6618,
    6619,
    -1,
    -1,
    2854,
    16786,
    1670,
    -1,
    6620,
    4284,
    -1,
    2548,
    -1,
    16788,
    2644,
    -1,
    -1,
    -1,
    -1,
    -1,
    2517,
    3250,
    -1,
    -1,
    -1,
    -1,
    -1,
    6621,
    6622,
    1520,
    16790,
    6623,
    16791,
    -1,
    6624,
    6625,
    -1,
    16793,
    6626,
    -1,
    6627,
    2002,
    5367,
    6628,
    -1,
    2982,
    6629,
    2580,
    -1,
    6630,
    -1,
    4433,
    -1,
    -1,
    -1,
    6631,
    3915,
    1845,
    16795,
    -1,
    3011,
    2687,
    -1,
    3840,
    2645,
    -1,
    16796,
    16797,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    6632,
    2305,
    6633,
    3709,
    -1,
    6647,
    -1,
    -1,
    -1,
    16799,
    6635,
    16800,
    3142,
    3656,
    2192,
    6634,
    3012,
    -1,
    -1,
    16801,
    -1,
    -1,
    -1,
    6636,
    -1,
    -1,
    -1,
    -1,
    16804,
    16802,
    3379,
    -1,
    16803,
    -1,
    16806,
    -1,
    -1,
    -1,
    16808,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16810,
    6638,
    6637,
    6640,
    -1,
    16812,
    6639,
    -1,
    6641,
    8424,
    -1,
    6643,
    16813,
    6642,
    -1,
    1832,
    -1,
    6644,
    -1,
    6645,
    6646,
    -1,
    -1,
    2366,
    4270,
    -1,
    6648,
    2826,
    -1,
    8425,
    -1,
    1583,
    6649,
    6650,
    16817,
    -1,
    -1,
    -1,
    16819,
    6651,
    16820,
    -1,
    -1,
    -1,
    16821,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16822,
    1491,
    -1,
    6652,
    8427,
    -1,
    -1,
    -1,
    6653,
    16823,
    -1,
    -1,
    -1,
    -1,
    -1,
    3830,
    -1,
    -1,
    -1,
    2576,
    -1,
    6655,
    -1,
    8428,
    -1,
    8429,
    8430,
    1700,
    1427,
    16824,
    16825,
    8431,
    -1,
    6654,
    6657,
    3617,
    8432,
    2855,
    -1,
    1651,
    16826,
    3954,
    16827,
    -1,
    -1,
    8433,
    2128,
    2033,
    16828,
    6656,
    16829,
    1671,
    8434,
    8435,
    -1,
    -1,
    -1,
    -1,
    -1,
    1788,
    8436,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1584,
    6661,
    4303,
    3162,
    8437,
    -1,
    3795,
    -1,
    6673,
    -1,
    1652,
    6671,
    -1,
    6669,
    6660,
    -1,
    6658,
    16830,
    6659,
    -1,
    2604,
    2049,
    3282,
    16831,
    -1,
    -1,
    3526,
    -1,
    -1,
    -1,
    -1,
    -1,
    1552,
    -1,
    6663,
    6662,
    -1,
    -1,
    16832,
    -1,
    6668,
    6664,
    6667,
    -1,
    8438,
    8439,
    -1,
    8440,
    8441,
    4089,
    6666,
    1653,
    1784,
    6670,
    16833,
    16834,
    6672,
    -1,
    -1,
    16835,
    -1,
    2119,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    6676,
    6685,
    6686,
    -1,
    -1,
    16837,
    1420,
    16838,
    -1,
    -1,
    -1,
    -1,
    8443,
    6693,
    -1,
    -1,
    -1,
    -1,
    1490,
    -1,
    -1,
    6684,
    -1,
    8444,
    -1,
    6683,
    16839,
    6678,
    6677,
    -1,
    6675,
    6674,
    3262,
    -1,
    3192,
    6680,
    8445,
    -1,
    16840,
    -1,
    -1,
    -1,
    6679,
    -1,
    -1,
    8446,
    16841,
    6682,
    -1,
    8447,
    -1,
    3087,
    2120,
    -1,
    -1,
    -1,
    -1,
    1536,
    6681,
    8448,
    2306,
    -1,
    16836,
    8449,
    -1,
    16842,
    3088,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8442,
    10099,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    6699,
    -1,
    6700,
    -1,
    1654,
    -1,
    -1,
    -1,
    1611,
    6697,
    8450,
    -1,
    16844,
    -1,
    -1,
    -1,
    -1,
    -1,
    6687,
    8451,
    6695,
    -1,
    6702,
    6696,
    -1,
    -1,
    16847,
    6694,
    -1,
    -1,
    -1,
    8452,
    6665,
    16848,
    16849,
    6692,
    -1,
    8453,
    -1,
    6688,
    16850,
    -1,
    16851,
    1833,
    6690,
    6701,
    -1,
    6691,
    -1,
    -1,
    -1,
    -1,
    8454,
    6703,
    8455,
    6689,
    3663,
    -1,
    -1,
    -1,
    -1,
    -1,
    4208,
    -1,
    -1,
    -1,
    6698,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    6720,
    -1,
    8456,
    8457,
    6712,
    -1,
    -1,
    -1,
    2910,
    -1,
    8458,
    -1,
    16852,
    1917,
    -1,
    2034,
    -1,
    6707,
    8459,
    16853,
    8460,
    -1,
    1656,
    16854,
    -1,
    2777,
    -1,
    6710,
    -1,
    -1,
    -1,
    2407,
    16855,
    -1,
    3432,
    6715,
    8461,
    -1,
    -1,
    -1,
    16856,
    -1,
    -1,
    -1,
    3931,
    8462,
    6706,
    -1,
    -1,
    -1,
    1655,
    2215,
    3768,
    6716,
    -1,
    6704,
    -1,
    -1,
    6713,
    -1,
    16857,
    -1,
    6723,
    -1,
    6708,
    -1,
    -1,
    -1,
    8463,
    -1,
    6709,
    3503,
    -1,
    8464,
    6714,
    -1,
    -1,
    8465,
    6711,
    3955,
    6717,
    1473,
    16858,
    -1,
    16859,
    -1,
    6705,
    -1,
    16860,
    -1,
    16862,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    6719,
    -1,
    6718,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    3646,
    6725,
    -1,
    6736,
    -1,
    -1,
    -1,
    -1,
    1785,
    -1,
    -1,
    -1,
    6739,
    -1,
    -1,
    6721,
    16863,
    -1,
    -1,
    6726,
    4213,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    6735,
    -1,
    8467,
    4188,
    16864,
    -1,
    -1,
    -1,
    4239,
    16865,
    -1,
    16866,
    16867,
    -1,
    -1,
    -1,
    -1,
    3283,
    -1,
    16868,
    16869,
    1766,
    16870,
    -1,
    -1,
    8468,
    -1,
    3841,
    6741,
    3478,
    -1,
    16872,
    1426,
    -1,
    -1,
    6734,
    -1,
    6730,
    3089,
    6724,
    6732,
    6737,
    8469,
    3583,
    -1,
    8470,
    -1,
    1419,
    16873,
    6729,
    16874,
    6738,
    3847,
    -1,
    16875,
    -1,
    -1,
    -1,
    -1,
    16876,
    6733,
    -1,
    6728,
    8471,
    -1,
    -1,
    -1,
    -1,
    -1,
    2778,
    -1,
    -1,
    -1,
    -1,
    2646,
    -1,
    -1,
    -1,
    2549,
    -1,
    -1,
    16879,
    -1,
    4097,
    -1,
    -1,
    3798,
    -1,
    8472,
    6744,
    -1,
    6753,
    -1,
    -1,
    -1,
    -1,
    16880,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    6731,
    -1,
    8473,
    -1,
    -1,
    1778,
    -1,
    8474,
    -1,
    -1,
    -1,
    2813,
    6742,
    8475,
    6747,
    3090,
    -1,
    16881,
    6743,
    8476,
    6750,
    8477,
    -1,
    3258,
    -1,
    6751,
    -1,
    -1,
    4189,
    6740,
    1715,
    -1,
    6746,
    16883,
    16884,
    6749,
    4053,
    -1,
    16885,
    -1,
    -1,
    6752,
    -1,
    -1,
    6745,
    6748,
    -1,
    16878,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16887,
    -1,
    -1,
    16888,
    -1,
    3956,
    -1,
    4324,
    16889,
    16890,
    16891,
    -1,
    -1,
    6756,
    -1,
    -1,
    -1,
    -1,
    -1,
    16892,
    -1,
    6763,
    16893,
    -1,
    6755,
    2572,
    -1,
    -1,
    -1,
    -1,
    -1,
    6722,
    -1,
    -1,
    -1,
    -1,
    -1,
    16894,
    -1,
    -1,
    -1,
    -1,
    3899,
    -1,
    4043,
    6762,
    6761,
    -1,
    6757,
    6758,
    -1,
    1525,
    16895,
    -1,
    -1,
    8480,
    6760,
    -1,
    6754,
    -1,
    8481,
    16896,
    16897,
    3342,
    -1,
    -1,
    -1,
    -1,
    16898,
    6759,
    1503,
    -1,
    8482,
    -1,
    -1,
    8479,
    -1,
    16899,
    3103,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    3888,
    16904,
    -1,
    6764,
    6768,
    -1,
    3722,
    -1,
    -1,
    -1,
    -1,
    6767,
    2779,
    2578,
    6770,
    -1,
    -1,
    2003,
    16900,
    -1,
    16905,
    -1,
    16906,
    -1,
    6771,
    -1,
    3848,
    6766,
    8483,
    6727,
    -1,
    -1,
    -1,
    16907,
    -1,
    -1,
    16908,
    16909,
    6765,
    8484,
    -1,
    -1,
    -1,
    4370,
    3479,
    3842,
    -1,
    -1,
    6778,
    -1,
    16901,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    6784,
    -1,
    -1,
    8487,
    16911,
    -1,
    16912,
    6785,
    16913,
    6772,
    16914,
    -1,
    -1,
    3657,
    -1,
    16915,
    6782,
    6774,
    -1,
    6776,
    -1,
    8488,
    -1,
    -1,
    8489,
    6786,
    6775,
    -1,
    16916,
    6779,
    -1,
    -1,
    1585,
    -1,
    3546,
    -1,
    6780,
    6783,
    16917,
    -1,
    16918,
    -1,
    -1,
    8490,
    -1,
    6773,
    -1,
    3013,
    -1,
    6777,
    2450,
    2856,
    2083,
    4125,
    8486,
    4130,
    2713,
    8491,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16922,
    -1,
    6790,
    6788,
    -1,
    16923,
    -1,
    -1,
    -1,
    -1,
    4369,
    -1,
    -1,
    -1,
    -1,
    -1,
    16924,
    -1,
    6787,
    16925,
    8492,
    -1,
    4220,
    8493,
    6789,
    6791,
    -1,
    -1,
    -1,
    -1,
    6792,
    -1,
    -1,
    16926,
    16927,
    -1,
    -1,
    6795,
    6793,
    -1,
    16928,
    -1,
    16929,
    -1,
    -1,
    3480,
    6794,
    16930,
    -1,
    -1,
    3710,
    6781,
    -1,
    -1,
    8494,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16931,
    2714,
    -1,
    6796,
    6801,
    3091,
    -1,
    -1,
    6800,
    8496,
    16932,
    -1,
    6769,
    -1,
    8497,
    8498,
    6802,
    3048,
    -1,
    -1,
    6797,
    6799,
    -1,
    -1,
    -1,
    -1,
    8499,
    16933,
    8495,
    6798,
    -1,
    -1,
    5656,
    5641,
    8501,
    -1,
    6804,
    -1,
    -1,
    -1,
    16934,
    -1,
    -1,
    16935,
    6803,
    -1,
    16936,
    -1,
    -1,
    16937,
    -1,
    8502,
    -1,
    -1,
    -1,
    4221,
    -1,
    6137,
    6805,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8503,
    16939,
    -1,
    -1,
    16940,
    -1,
    -1,
    6806,
    16942,
    1387,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    6807,
    2216,
    -1,
    1931,
    -1,
    -1,
    16944,
    6809,
    4572,
    16945,
    8504,
    -1,
    -1,
    1964,
    8505,
    4255,
    -1,
    2056,
    6810,
    -1,
    -1,
    8507,
    -1,
    -1,
    -1,
    -1,
    6811,
    -1,
    -1,
    -1,
    3273,
    8509,
    -1,
    -1,
    16946,
    -1,
    6812,
    -1,
    -1,
    -1,
    8510,
    -1,
    16947,
    -1,
    3567,
    16948,
    1435,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16949,
    -1,
    16950,
    1664,
    6817,
    6818,
    16951,
    -1,
    -1,
    -1,
    16952,
    -1,
    6813,
    -1,
    2472,
    -1,
    -1,
    8511,
    -1,
    -1,
    -1,
    16953,
    16954,
    -1,
    -1,
    -1,
    -1,
    -1,
    6814,
    3605,
    -1,
    -1,
    -1,
    16955,
    6815,
    6816,
    6826,
    -1,
    -1,
    -1,
    6820,
    6823,
    16957,
    -1,
    16958,
    -1,
    -1,
    6819,
    -1,
    8512,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16959,
    -1,
    16960,
    6821,
    -1,
    6822,
    2593,
    -1,
    6824,
    -1,
    3228,
    -1,
    2121,
    1725,
    -1,
    -1,
    16961,
    -1,
    -1,
    6827,
    16962,
    -1,
    16963,
    -1,
    1722,
    -1,
    6832,
    -1,
    -1,
    6828,
    6831,
    -1,
    -1,
    -1,
    16964,
    3691,
    -1,
    16965,
    -1,
    -1,
    6829,
    -1,
    -1,
    6830,
    3799,
    3723,
    6833,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    3194,
    6843,
    8513,
    6839,
    8514,
    8515,
    1672,
    -1,
    6837,
    -1,
    3957,
    6838,
    -1,
    16967,
    6835,
    16968,
    6836,
    6841,
    6844,
    8516,
    -1,
    6842,
    16969,
    -1,
    16970,
    6840,
    6834,
    16971,
    -1,
    -1,
    -1,
    -1,
    3251,
    16972,
    6851,
    -1,
    4051,
    -1,
    -1,
    16973,
    -1,
    16974,
    -1,
    16975,
    -1,
    6849,
    -1,
    -1,
    -1,
    6850,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16976,
    -1,
    -1,
    6845,
    -1,
    -1,
    6847,
    -1,
    -1,
    16977,
    6848,
    -1,
    -1,
    16978,
    6846,
    16979,
    -1,
    -1,
    16980,
    -1,
    -1,
    -1,
    -1,
    -1,
    2983,
    -1,
    4348,
    6855,
    -1,
    6856,
    -1,
    -1,
    16981,
    -1,
    6862,
    -1,
    2829,
    -1,
    6858,
    16982,
    6861,
    -1,
    -1,
    -1,
    -1,
    -1,
    6853,
    6852,
    -1,
    -1,
    6863,
    16983,
    16984,
    1657,
    -1,
    6859,
    -1,
    6864,
    -1,
    -1,
    -1,
    6860,
    -1,
    -1,
    8517,
    16985,
    -1,
    6857,
    -1,
    3308,
    -1,
    6854,
    -1,
    -1,
    -1,
    16986,
    -1,
    -1,
    3643,
    -1,
    -1,
    6868,
    -1,
    -1,
    -1,
    -1,
    8518,
    8519,
    16990,
    -1,
    16991,
    -1,
    4162,
    -1,
    -1,
    -1,
    -1,
    -1,
    16992,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    6867,
    16993,
    -1,
    6866,
    -1,
    -1,
    -1,
    -1,
    16989,
    -1,
    -1,
    -1,
    6875,
    8520,
    8521,
    -1,
    6869,
    -1,
    -1,
    -1,
    6877,
    -1,
    8522,
    -1,
    -1,
    -1,
    -1,
    4205,
    6880,
    -1,
    6871,
    16996,
    -1,
    6872,
    16998,
    -1,
    -1,
    6876,
    -1,
    6879,
    6878,
    -1,
    -1,
    -1,
    6870,
    -1,
    -1,
    16999,
    -1,
    6873,
    -1,
    6890,
    -1,
    -1,
    -1,
    8524,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17001,
    6883,
    -1,
    -1,
    17003,
    -1,
    17004,
    17005,
    -1,
    -1,
    -1,
    17006,
    17007,
    8525,
    17008,
    -1,
    6881,
    -1,
    -1,
    6882,
    -1,
    -1,
    17000,
    6887,
    6888,
    -1,
    1701,
    -1,
    1913,
    -1,
    -1,
    6886,
    -1,
    -1,
    17009,
    -1,
    17010,
    -1,
    6865,
    8526,
    -1,
    -1,
    -1,
    8527,
    17011,
    -1,
    6885,
    6889,
    6884,
    8528,
    6891,
    -1,
    17012,
    8529,
    6893,
    6892,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8530,
    -1,
    6895,
    6894,
    6825,
    -1,
    -1,
    -1,
    6899,
    17013,
    -1,
    -1,
    -1,
    -1,
    -1,
    17014,
    -1,
    -1,
    6896,
    17015,
    -1,
    -1,
    -1,
    6897,
    -1,
    -1,
    6898,
    -1,
    6900,
    17016,
    -1,
    -1,
    -1,
    2143,
    -1,
    6902,
    -1,
    6901,
    -1,
    2647,
    -1,
    -1,
    -1,
    17018,
    -1,
    2307,
    5740,
    -1,
    -1,
    -1,
    -1,
    6903,
    2680,
    -1,
    -1,
    -1,
    1716,
    17019,
    6904,
    -1,
    1553,
    -1,
    2780,
    6905,
    17020,
    -1,
    2308,
    6906,
    1474,
    17021,
    -1,
    -1,
    -1,
    3788,
    17024,
    -1,
    6907,
    -1,
    -1,
    -1,
    17026,
    2893,
    -1,
    6914,
    -1,
    -1,
    6911,
    -1,
    3274,
    -1,
    -1,
    -1,
    -1,
    -1,
    6912,
    6909,
    2035,
    -1,
    6908,
    6915,
    -1,
    -1,
    -1,
    -1,
    -1,
    2089,
    -1,
    -1,
    3163,
    -1,
    6921,
    -1,
    -1,
    -1,
    -1,
    6917,
    -1,
    -1,
    -1,
    3122,
    6916,
    8531,
    6919,
    -1,
    -1,
    -1,
    -1,
    6910,
    -1,
    17027,
    -1,
    6920,
    -1,
    6922,
    -1,
    -1,
    -1,
    -1,
    -1,
    8532,
    3744,
    -1,
    -1,
    6918,
    -1,
    6923,
    6925,
    -1,
    -1,
    2212,
    6913,
    -1,
    1442,
    -1,
    -1,
    -1,
    -1,
    17028,
    17029,
    17030,
    6924,
    17031,
    2408,
    4315,
    6926,
    6927,
    3092,
    -1,
    -1,
    -1,
    -1,
    8533,
    -1,
    -1,
    -1,
    8534,
    4231,
    -1,
    17034,
    17032,
    17035,
    6928,
    4156,
    -1,
    -1,
    6929,
    6930,
    -1,
    17036,
    3921,
    6931,
    -1,
    2381,
    -1,
    4232,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    6936,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17037,
    17038,
    6937,
    2781,
    6935,
    8536,
    -1,
    -1,
    4206,
    6932,
    -1,
    -1,
    6934,
    2942,
    2913,
    -1,
    -1,
    17040,
    6933,
    -1,
    6938,
    -1,
    -1,
    3856,
    -1,
    -1,
    6940,
    -1,
    6939,
    -1,
    -1,
    -1,
    1767,
    -1,
    3958,
    6941,
    -1,
    -1,
    -1,
    -1,
    8539,
    8540,
    8541,
    -1,
    8537,
    6953,
    6943,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    6944,
    -1,
    8542,
    -1,
    -1,
    6945,
    6946,
    -1,
    -1,
    -1,
    -1,
    8543,
    -1,
    8544,
    -1,
    -1,
    -1,
    6950,
    17042,
    6951,
    8545,
    -1,
    6949,
    -1,
    -1,
    -1,
    -1,
    8546,
    6947,
    17044,
    6942,
    6948,
    17045,
    -1,
    -1,
    -1,
    17046,
    -1,
    -1,
    6952,
    7448,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1605,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    6955,
    2036,
    6954,
    -1,
    17049,
    -1,
    6957,
    -1,
    6956,
    -1,
    -1,
    -1,
    6959,
    -1,
    -1,
    6958,
    -1,
    6960,
    -1,
    -1,
    2648,
    -1,
    6961,
    -1,
    -1,
    6962,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    6963,
    2943,
    17050,
    4190,
    -1,
    6964,
    -1,
    -1,
    3857,
    3609,
    6965,
    17051,
    6966,
    2172,
    -1,
    -1,
    -1,
    1891,
    17052,
    -1,
    -1,
    6967,
    8547,
    -1,
    2518,
    3604,
    6968,
    -1,
    1739,
    -1,
    -1,
    -1,
    -1,
    17053,
    -1,
    6969,
    -1,
    -1,
    -1,
    -1,
    6971,
    4222,
    -1,
    6970,
    2857,
    -1,
    6972,
    -1,
    -1,
    6973,
    17054,
    -1,
    6974,
    1834,
    -1,
    -1,
    -1,
    17055,
    -1,
    -1,
    6975,
    -1,
    -1,
    6976,
    -1,
    6977,
    6978,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1740,
    -1,
    8549,
    -1,
    17056,
    -1,
    17057,
    -1,
    6979,
    -1,
    6980,
    6981,
    -1,
    -1,
    -1,
    -1,
    -1,
    1679,
    -1,
    8550,
    2827,
    6982,
    -1,
    -1,
    -1,
    17058,
    -1,
    -1,
    -1,
    -1,
    -1,
    17060,
    -1,
    17061,
    6983,
    -1,
    8551,
    -1,
    6984,
    -1,
    -1,
    -1,
    -1,
    17062,
    -1,
    17063,
    2193,
    -1,
    3380,
    6985,
    -1,
    -1,
    -1,
    -1,
    2122,
    -1,
    2873,
    -1,
    6988,
    -1,
    3481,
    -1,
    6987,
    17065,
    8552,
    2084,
    17066,
    8553,
    6986,
    3188,
    1892,
    -1,
    -1,
    6989,
    -1,
    6990,
    -1,
    2782,
    -1,
    17068,
    8554,
    2144,
    -1,
    6991,
    -1,
    -1,
    -1,
    -1,
    3959,
    -1,
    -1,
    2976,
    -1,
    -1,
    -1,
    1965,
    -1,
    4126,
    3049,
    17069,
    6992,
    8555,
    -1,
    -1,
    2858,
    3275,
    2783,
    -1,
    17070,
    -1,
    -1,
    6993,
    -1,
    -1,
    -1,
    17071,
    6996,
    8556,
    6997,
    -1,
    -1,
    -1,
    -1,
    17072,
    8557,
    -1,
    2379,
    3135,
    6995,
    -1,
    2784,
    3789,
    -1,
    -1,
    17073,
    -1,
    -1,
    6994,
    -1,
    8558,
    2519,
    -1,
    1554,
    8559,
    7001,
    2123,
    -1,
    -1,
    2521,
    -1,
    -1,
    2520,
    -1,
    4368,
    7000,
    6999,
    3014,
    -1,
    1924,
    4358,
    1717,
    2785,
    -1,
    8560,
    -1,
    -1,
    -1,
    8561,
    -1,
    -1,
    6998,
    -1,
    -1,
    -1,
    -1,
    -1,
    7003,
    -1,
    7004,
    7002,
    -1,
    2217,
    -1,
    4167,
    -1,
    -1,
    2522,
    3578,
    -1,
    -1,
    17075,
    7007,
    -1,
    2945,
    -1,
    3229,
    -1,
    -1,
    4157,
    -1,
    7010,
    -1,
    -1,
    -1,
    2237,
    -1,
    2944,
    7006,
    -1,
    7011,
    2238,
    7008,
    7009,
    8562,
    7005,
    -1,
    -1,
    -1,
    2979,
    3518,
    17074,
    -1,
    3210,
    -1,
    1658,
    -1,
    -1,
    -1,
    -1,
    17076,
    -1,
    3745,
    -1,
    -1,
    1914,
    -1,
    17077,
    3309,
    -1,
    -1,
    7014,
    -1,
    7012,
    -1,
    -1,
    3239,
    -1,
    -1,
    -1,
    2946,
    1835,
    7013,
    -1,
    2878,
    8563,
    -1,
    4271,
    -1,
    -1,
    -1,
    4355,
    17078,
    -1,
    -1,
    7015,
    7026,
    3310,
    -1,
    7025,
    8564,
    7022,
    7030,
    7023,
    -1,
    7019,
    -1,
    3381,
    7018,
    -1,
    -1,
    -1,
    7016,
    -1,
    4136,
    2523,
    -1,
    -1,
    7020,
    -1,
    7017,
    8565,
    -1,
    8566,
    7024,
    2715,
    -1,
    2194,
    -1,
    17079,
    -1,
    3191,
    -1,
    3984,
    1561,
    1475,
    -1,
    3482,
    17082,
    -1,
    7028,
    -1,
    -1,
    17081,
    -1,
    7027,
    17083,
    3547,
    -1,
    7032,
    -1,
    -1,
    -1,
    7021,
    -1,
    7031,
    7033,
    -1,
    2173,
    7029,
    2309,
    17084,
    2590,
    -1,
    17085,
    7034,
    4191,
    -1,
    -1,
    -1,
    -1,
    7037,
    -1,
    7040,
    -1,
    -1,
    7038,
    3780,
    17086,
    -1,
    -1,
    -1,
    -1,
    -1,
    7035,
    -1,
    -1,
    -1,
    -1,
    -1,
    2037,
    -1,
    -1,
    -1,
    -1,
    7039,
    -1,
    -1,
    7041,
    -1,
    17087,
    -1,
    -1,
    8570,
    -1,
    -1,
    7045,
    -1,
    -1,
    7042,
    -1,
    7044,
    7043,
    -1,
    17089,
    -1,
    -1,
    8571,
    -1,
    7046,
    -1,
    2554,
    8572,
    7048,
    7047,
    3831,
    -1,
    17090,
    7050,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    2124,
    -1,
    -1,
    8573,
    -1,
    7049,
    7051,
    -1,
    -1,
    7052,
    1915,
    6518,
    2814,
    -1,
    7053,
    -1,
    17091,
    2239,
    -1,
    -1,
    -1,
    -1,
    -1,
    7054,
    -1,
    17092,
    7055,
    17093,
    -1,
    2473,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    5403,
    17094,
    7056,
    -1,
    7057,
    -1,
    2649,
    -1,
    7058,
    7059,
    17095,
    17096,
    7060,
    -1,
    -1,
    7061,
    7062,
    -1,
    17097,
    8574,
    17098,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    3206,
    -1,
    17099,
    7063,
    -1,
    -1,
    17101,
    -1,
    7065,
    -1,
    7064,
    -1,
    -1,
    -1,
    17104,
    3483,
    17105,
    7066,
    8575,
    3960,
    -1,
    7067,
    -1,
    7068,
    17106,
    7069,
    -1,
    -1,
    -1,
    17107,
    7070,
    -1,
    17108,
    -1,
    -1,
    3535,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    2786,
    7071,
    -1,
    -1,
    -1,
    -1,
    -1,
    8576,
    17109,
    2330,
    4431,
    7072,
    17110,
    -1,
    -1,
    -1,
    -1,
    -1,
    17111,
    -1,
    -1,
    -1,
    -1,
    7073,
    3790,
    7074,
    -1,
    7082,
    -1,
    -1,
    -1,
    -1,
    -1,
    7075,
    -1,
    -1,
    7077,
    -1,
    -1,
    -1,
    7076,
    7078,
    -1,
    3985,
    7079,
    7080,
    -1,
    -1,
    -1,
    17114,
    17113,
    7081,
    -1,
    -1,
    -1,
    7083,
    17115,
    -1,
    17117,
    -1,
    1704,
    3360,
    -1,
    3832,
    2415,
    2310,
    -1,
    17118,
    -1,
    -1,
    3806,
    1660,
    3711,
    7086,
    1836,
    2966,
    7085,
    7090,
    3284,
    4108,
    -1,
    7088,
    7089,
    1893,
    -1,
    7091,
    3638,
    3164,
    -1,
    -1,
    3746,
    3412,
    7087,
    -1,
    3986,
    1673,
    7093,
    4331,
    3323,
    4360,
    -1,
    -1,
    2524,
    7092,
    -1,
    3118,
    -1,
    -1,
    7109,
    3015,
    -1,
    -1,
    3565,
    -1,
    3807,
    -1,
    17120,
    17119,
    -1,
    -1,
    17121,
    7096,
    2474,
    2525,
    -1,
    2787,
    -1,
    3640,
    8578,
    2174,
    7095,
    7094,
    -1,
    3833,
    -1,
    -1,
    -1,
    2570,
    -1,
    -1,
    3433,
    -1,
    -1,
    17123,
    17124,
    -1,
    -1,
    8579,
    -1,
    -1,
    -1,
    8580,
    -1,
    7098,
    7099,
    2311,
    7097,
    8581,
    -1,
    -1,
    -1,
    -1,
    -1,
    7100,
    7101,
    -1,
    7103,
    3104,
    17126,
    7102,
    1855,
    -1,
    7105,
    17127,
    7104,
    7106,
    -1,
    8583,
    7108,
    7110,
    -1,
    7111,
    -1,
    -1,
    -1,
    -1,
    8584,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    2967,
    -1,
    2584,
    7112,
    -1,
    -1,
    -1,
    1741,
    17128,
    7113,
    -1,
    -1,
    3093,
    7114,
    -1,
    7115,
    3834,
    -1,
    -1,
    1894,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    7116,
    -1,
    -1,
    17129,
    3311,
    -1,
    -1,
    -1,
    -1,
    1562,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17130,
    -1,
    -1,
    -1,
    7117,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    2618,
    -1,
    -1,
    17131,
    -1,
    2905,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8585,
    -1,
    -1,
    -1,
    3114,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    7120,
    -1,
    -1,
    -1,
    7119,
    -1,
    -1,
    -1,
    7118,
    -1,
    -1,
    -1,
    17133,
    -1,
    17134,
    -1,
    -1,
    7126,
    7124,
    -1,
    8586,
    7121,
    -1,
    8587,
    -1,
    -1,
    -1,
    -1,
    7123,
    8588,
    -1,
    17135,
    7122,
    7125,
    -1,
    1966,
    -1,
    7129,
    -1,
    2968,
    -1,
    7130,
    -1,
    -1,
    -1,
    -1,
    2218,
    -1,
    7127,
    7128,
    17136,
    -1,
    -1,
    4332,
    -1,
    -1,
    -1,
    3312,
    -1,
    3016,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    7131,
    17139,
    -1,
    7134,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17140,
    -1,
    7132,
    7133,
    4192,
    -1,
    17137,
    -1,
    -1,
    3484,
    7137,
    -1,
    -1,
    -1,
    17142,
    -1,
    17143,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    7135,
    7136,
    7138,
    8589,
    17144,
    17145,
    8590,
    -1,
    -1,
    -1,
    17146,
    -1,
    -1,
    7152,
    -1,
    -1,
    -1,
    -1,
    -1,
    7141,
    -1,
    -1,
    -1,
    7142,
    7140,
    17149,
    -1,
    -1,
    17150,
    -1,
    -1,
    -1,
    8591,
    -1,
    -1,
    -1,
    -1,
    7139,
    -1,
    3382,
    -1,
    -1,
    7144,
    7148,
    7145,
    7143,
    17151,
    7146,
    -1,
    -1,
    -1,
    7147,
    -1,
    -1,
    -1,
    17152,
    7154,
    -1,
    -1,
    -1,
    7149,
    -1,
    -1,
    -1,
    -1,
    -1,
    2969,
    7151,
    -1,
    17153,
    7153,
    7150,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17154,
    17155,
    -1,
    17156,
    8592,
    -1,
    7156,
    -1,
    2650,
    -1,
    7155,
    -1,
    -1,
    -1,
    -1,
    8593,
    7157,
    -1,
    -1,
    -1,
    -1,
    7158,
    -1,
    -1,
    7161,
    7160,
    -1,
    7159,
    -1,
    -1,
    7163,
    7162,
    -1,
    4127,
    -1,
    -1,
    -1,
    7165,
    -1,
    7164,
    7166,
    -1,
    -1,
    -1,
    17157,
    7167,
    -1,
    -1,
    -1,
    -1,
    17158,
    -1,
    -1,
    7169,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    7168,
    2859,
    7170,
    -1,
    17159,
    2050,
    7171,
    7173,
    -1,
    17160,
    -1,
    17161,
    17162,
    -1,
    -1,
    -1,
    -1,
    17163,
    -1,
    -1,
    7174,
    -1,
    8595,
    -1,
    -1,
    -1,
    -1,
    7175,
    7172,
    -1,
    7176,
    -1,
    2591,
    7177,
    1895,
    2086,
    -1,
    -1,
    -1,
    17165,
    2175,
    -1,
    17166,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    7178,
    -1,
    -1,
    -1,
    3557,
    -1,
    -1,
    3413,
    7179,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    7182,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    2557,
    17168,
    8597,
    7181,
    7180,
    2125,
    7183,
    -1,
    17170,
    -1,
    -1,
    1742,
    -1,
    7185,
    -1,
    -1,
    17171,
    2409,
    7184,
    -1,
    7193,
    -1,
    -1,
    -1,
    -1,
    -1,
    7187,
    7189,
    3922,
    7186,
    -1,
    17172,
    -1,
    7188,
    -1,
    7192,
    7190,
    1896,
    8598,
    7191,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    7194,
    -1,
    -1,
    3629,
    4285,
    17173,
    -1,
    8599,
    -1,
    2651,
    -1,
    -1,
    -1,
    7195,
    -1,
    -1,
    8600,
    -1,
    4137,
    7197,
    -1,
    7196,
    -1,
    -1,
    7200,
    4168,
    17174,
    -1,
    7199,
    -1,
    1768,
    7198,
    7203,
    -1,
    -1,
    7202,
    17175,
    -1,
    7201,
    3401,
    7204,
    -1,
    -1,
    -1,
    -1,
    -1,
    8601,
    -1,
    -1,
    7205,
    17176,
    -1,
    -1,
    -1,
    7206,
    -1,
    -1,
    2331,
    -1,
    2072,
    7207,
    7208,
    7209,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    2860,
    7210,
    -1,
    2550,
    7211,
    -1,
    -1,
    -1,
    7212,
    17178,
    -1,
    8603,
    4606,
    4605,
    -1,
    -1,
    -1,
    -1,
    7213,
    6471,
    7214,
    3199,
    2830,
    3603,
    -1,
    17179,
    8604,
    17181,
    7215,
    -1,
    -1,
    3906,
    3341,
    2349,
    -1,
    -1,
    3204,
    -1,
    17183,
    1513,
    -1,
    4035,
    2874,
    17184,
    -1,
    -1,
    -1,
    17186,
    -1,
    -1,
    17187,
    2129,
    -1,
    -1,
    2038,
    -1,
    17188,
    3907,
    17189,
    -1,
    -1,
    -1,
    -1,
    7216,
    -1,
    -1,
    -1,
    -1,
    -1,
    17190,
    -1,
    7218,
    -1,
    8605,
    7217,
    1661,
    -1,
    8606,
    3563,
    7219,
    3658,
    -1,
    3402,
    8607,
    7220,
    2681,
    17191,
    -1,
    -1,
    7222,
    17192,
    -1,
    4075,
    7237,
    7224,
    7225,
    17193,
    -1,
    3330,
    -1,
    -1,
    3165,
    3094,
    17194,
    3485,
    -1,
    7223,
    1932,
    -1,
    8608,
    -1,
    -1,
    7232,
    17195,
    7229,
    7242,
    3486,
    3259,
    7226,
    -1,
    3383,
    3434,
    7227,
    7231,
    2883,
    -1,
    3642,
    3333,
    -1,
    -1,
    2947,
    7230,
    3115,
    3105,
    7228,
    1418,
    4325,
    -1,
    -1,
    -1,
    7233,
    -1,
    -1,
    -1,
    -1,
    -1,
    8609,
    3166,
    -1,
    -1,
    2652,
    2861,
    -1,
    -1,
    7235,
    7234,
    17196,
    1488,
    7236,
    -1,
    -1,
    3774,
    -1,
    7244,
    -1,
    -1,
    3536,
    2894,
    17198,
    17199,
    3252,
    -1,
    2061,
    -1,
    7243,
    4158,
    1534,
    -1,
    3908,
    1662,
    7238,
    7239,
    7240,
    7241,
    3504,
    3198,
    1476,
    7245,
    -1,
    7246,
    7804,
    -1,
    -1,
    3130,
    17200,
    7247,
    -1,
    1586,
    3051,
    -1,
    2176,
    -1,
    4193,
    -1,
    -1,
    7248,
    3395,
    -1,
    -1,
    -1,
    3095,
    2592,
    7249,
    -1,
    -1,
    7252,
    -1,
    -1,
    2700,
    7250,
    3018,
    3017,
    -1,
    1477,
    -1,
    4272,
    7254,
    -1,
    3747,
    7256,
    7255,
    7253,
    6271,
    1837,
    17203,
    -1,
    7221,
    8611,
    7258,
    7257,
    -1,
    17204,
    -1,
    -1,
    7259,
    17205,
    4159,
    -1,
    -1,
    -1,
    8612,
    -1,
    8613,
    -1,
    8614,
    -1,
    8615,
    -1,
    -1,
    -1,
    -1,
    -1,
    17207,
    8616,
    3542,
    -1,
    -1,
    3961,
    -1,
    7260,
    -1,
    2594,
    -1,
    -1,
    -1,
    -1,
    7261,
    17209,
    7262,
    -1,
    8617,
    -1,
    7263,
    17210,
    -1,
    3384,
    -1,
    -1,
    -1,
    -1,
    -1,
    8618,
    -1,
    -1,
    1483,
    -1,
    17211,
    8619,
    8620,
    -1,
    8621,
    17212,
    -1,
    2312,
    -1,
    -1,
    -1,
    4349,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8622,
    -1,
    -1,
    -1,
    7267,
    17214,
    8623,
    8624,
    17215,
    -1,
    2087,
    7264,
    -1,
    7265,
    -1,
    -1,
    -1,
    3843,
    -1,
    -1,
    17220,
    -1,
    1743,
    -1,
    8625,
    -1,
    -1,
    17218,
    -1,
    8626,
    4160,
    17217,
    2004,
    -1,
    -1,
    -1,
    -1,
    -1,
    3435,
    17221,
    17222,
    17219,
    -1,
    7268,
    -1,
    17223,
    -1,
    17224,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    7269,
    -1,
    8628,
    8629,
    8630,
    -1,
    17225,
    7270,
    -1,
    -1,
    17226,
    -1,
    17227,
    -1,
    -1,
    -1,
    8631,
    8632,
    -1,
    -1,
    -1,
    8633,
    -1,
    -1,
    -1,
    -1,
    -1,
    3385,
    -1,
    8634,
    7272,
    8635,
    7271,
    -1,
    8636,
    -1,
    -1,
    17228,
    -1,
    17229,
    17230,
    -1,
    -1,
    8637,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17231,
    17232,
    8638,
    3528,
    7273,
    2653,
    2601,
    3630,
    3276,
    -1,
    -1,
    -1,
    2619,
    -1,
    2895,
    -1,
    7274,
    17233,
    7275,
    17234,
    -1,
    8639,
    -1,
    -1,
    -1,
    -1,
    -1,
    17235,
    2880,
    7276,
    17236,
    7277,
    -1,
    -1,
    -1,
    7278,
    4214,
    -1,
    2654,
    -1,
    -1,
    -1,
    -1,
    -1,
    7280,
    7279,
    17237,
    2313,
    -1,
    2338,
    2475,
    17238,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    7283,
    8640,
    -1,
    17239,
    -1,
    2701,
    -1,
    7282,
    -1,
    7281,
    -1,
    3174,
    17240,
    -1,
    2240,
    -1,
    2948,
    -1,
    -1,
    -1,
    -1,
    3681,
    -1,
    -1,
    -1,
    -1,
    2656,
    -1,
    8641,
    -1,
    -1,
    -1,
    7284,
    -1,
    2788,
    -1,
    -1,
    -1,
    17241,
    -1,
    7287,
    7285,
    8642,
    -1,
    17242,
    7286,
    -1,
    8643,
    -1,
    17243,
    7289,
    7288,
    17244,
    -1,
    2815,
    -1,
    7290,
    -1,
    8644,
    -1,
    -1,
    -1,
    7291,
    7292,
    -1,
    17245,
    17246,
    -1,
    3713,
    2400,
    2602,
    7293,
    -1,
    7294,
    4233,
    2668,
    4119,
    4273,
    7295,
    2039,
    -1,
    -1,
    -1,
    -1,
    7296,
    8645,
    3386,
    -1,
    17247,
    7299,
    1779,
    2862,
    -1,
    7297,
    -1,
    7298,
    -1,
    3353,
    8647,
    8648,
    4000,
    2065,
    -1,
    -1,
    -1,
    -1,
    17250,
    8649,
    17251,
    -1,
    -1,
    8650,
    -1,
    -1,
    -1,
    7301,
    7302,
    -1,
    -1,
    -1,
    -1,
    8646,
    7300,
    -1,
    -1,
    7304,
    -1,
    17252,
    -1,
    -1,
    -1,
    -1,
    -1,
    8651,
    -1,
    -1,
    17253,
    -1,
    -1,
    3540,
    1726,
    -1,
    8652,
    7308,
    -1,
    -1,
    7305,
    7307,
    17254,
    17255,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    7303,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    7378,
    -1,
    -1,
    7306,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17257,
    4304,
    -1,
    -1,
    2219,
    8653,
    8654,
    8655,
    -1,
    8656,
    -1,
    -1,
    7316,
    8657,
    -1,
    17258,
    8658,
    3403,
    7311,
    -1,
    17259,
    7314,
    7312,
    17260,
    7317,
    -1,
    -1,
    17261,
    8659,
    7318,
    17262,
    -1,
    -1,
    -1,
    -1,
    17263,
    7310,
    -1,
    17264,
    7323,
    1587,
    -1,
    -1,
    7309,
    -1,
    17265,
    17266,
    3678,
    -1,
    7313,
    17267,
    2789,
    17268,
    17269,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    2314,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8660,
    -1,
    -1,
    -1,
    17272,
    17273,
    3987,
    17274,
    2041,
    -1,
    -1,
    2669,
    -1,
    3505,
    -1,
    -1,
    8661,
    17275,
    -1,
    -1,
    -1,
    17276,
    -1,
    -1,
    -1,
    3020,
    -1,
    7321,
    -1,
    7315,
    7320,
    17277,
    4076,
    17278,
    3313,
    7322,
    7319,
    -1,
    -1,
    17279,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17280,
    -1,
    -1,
    -1,
    17281,
    -1,
    3019,
    -1,
    -1,
    -1,
    -1,
    17284,
    -1,
    -1,
    -1,
    -1,
    7326,
    -1,
    7325,
    -1,
    -1,
    -1,
    -1,
    -1,
    17285,
    17286,
    -1,
    8662,
    -1,
    -1,
    -1,
    17287,
    -1,
    -1,
    -1,
    -1,
    8663,
    8664,
    -1,
    17288,
    7324,
    17289,
    -1,
    3962,
    8665,
    -1,
    -1,
    -1,
    17290,
    -1,
    17291,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8666,
    -1,
    -1,
    -1,
    2722,
    17292,
    -1,
    17293,
    -1,
    7327,
    3916,
    -1,
    -1,
    1555,
    -1,
    -1,
    -1,
    -1,
    3797,
    3277,
    -1,
    -1,
    -1,
    17297,
    1967,
    17298,
    7329,
    17299,
    2316,
    -1,
    -1,
    8667,
    -1,
    -1,
    17300,
    -1,
    8668,
    -1,
    2455,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17301,
    -1,
    7328,
    2896,
    17294,
    -1,
    -1,
    -1,
    17302,
    -1,
    -1,
    2897,
    7332,
    7334,
    -1,
    -1,
    17303,
    17304,
    8669,
    2816,
    8670,
    7333,
    7335,
    -1,
    8671,
    2024,
    17305,
    3796,
    17306,
    -1,
    2603,
    4326,
    -1,
    7331,
    2437,
    -1,
    -1,
    4354,
    -1,
    -1,
    7337,
    -1,
    -1,
    -1,
    -1,
    7336,
    7338,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    7330,
    -1,
    -1,
    17309,
    8672,
    8673,
    8674,
    3550,
    -1,
    3436,
    -1,
    -1,
    -1,
    17310,
    -1,
    -1,
    3344,
    -1,
    7343,
    17311,
    -1,
    -1,
    17312,
    3230,
    7339,
    -1,
    -1,
    -1,
    7340,
    -1,
    -1,
    -1,
    8675,
    8676,
    -1,
    -1,
    -1,
    -1,
    8677,
    17313,
    2080,
    -1,
    7342,
    -1,
    8678,
    17314,
    -1,
    17315,
    -1,
    2177,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    7341,
    -1,
    2790,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17319,
    -1,
    -1,
    17320,
    1780,
    -1,
    -1,
    17321,
    -1,
    -1,
    -1,
    -1,
    7347,
    -1,
    2380,
    3096,
    -1,
    -1,
    3331,
    8679,
    -1,
    -1,
    17322,
    -1,
    -1,
    17316,
    -1,
    8680,
    -1,
    -1,
    -1,
    1718,
    -1,
    -1,
    -1,
    -1,
    7345,
    7346,
    3324,
    -1,
    7344,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    7348,
    8681,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17327,
    -1,
    7354,
    -1,
    -1,
    8682,
    17328,
    7357,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    7356,
    3396,
    -1,
    -1,
    -1,
    -1,
    7349,
    7350,
    7353,
    -1,
    -1,
    -1,
    17329,
    7355,
    8683,
    8684,
    -1,
    2005,
    17330,
    -1,
    7358,
    7352,
    -1,
    17331,
    7351,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17326,
    -1,
    -1,
    -1,
    17323,
    -1,
    -1,
    -1,
    -1,
    -1,
    17336,
    -1,
    8686,
    -1,
    -1,
    -1,
    -1,
    -1,
    7362,
    8685,
    -1,
    -1,
    7363,
    -1,
    17332,
    -1,
    -1,
    -1,
    -1,
    -1,
    17333,
    7364,
    -1,
    -1,
    7361,
    7360,
    -1,
    17334,
    17335,
    2791,
    3487,
    7359,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    7368,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    7366,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17337,
    8687,
    17338,
    7367,
    7365,
    -1,
    3189,
    -1,
    7369,
    17339,
    -1,
    -1,
    -1,
    -1,
    -1,
    7370,
    -1,
    -1,
    7372,
    17340,
    -1,
    -1,
    -1,
    -1,
    8688,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1838,
    7371,
    4131,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    7383,
    7373,
    -1,
    -1,
    7376,
    -1,
    7374,
    -1,
    7375,
    8689,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    7377,
    8690,
    -1,
    17343,
    -1,
    17344,
    7379,
    8691,
    8692,
    -1,
    -1,
    7380,
    -1,
    7381,
    -1,
    -1,
    -1,
    -1,
    7384,
    7382,
    7385,
    7387,
    -1,
    7386,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    3314,
    17345,
    17346,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    4112,
    -1,
    7388,
    3021,
    -1,
    -1,
    17347,
    7389,
    -1,
    3889,
    7390,
    1702,
    17348,
    17349,
    8693,
    1531,
    -1,
    1840,
    -1,
    1839,
    7391,
    -1,
    7392,
    -1,
    7393,
    7394,
    -1,
    -1,
    -1,
    -1,
    -1,
    8694,
    7395,
    -1,
    1841,
    1744,
    2317,
    3687,
    8695,
    7397,
    7396,
    8696,
    -1,
    17351,
    8697,
    7398,
    -1,
    -1,
    -1,
    -1,
    1563,
    -1,
    17352,
    -1,
    8698,
    -1,
    -1,
    7401,
    -1,
    7400,
    7399,
    8699,
    7402,
    -1,
    -1,
    -1,
    -1,
    7405,
    -1,
    -1,
    -1,
    1448,
    17354,
    -1,
    7403,
    8700,
    7407,
    7406,
    -1,
    -1,
    8701,
    -1,
    -1,
    8702,
    7409,
    7408,
    7410,
    -1,
    3491,
    -1,
    8704,
    -1,
    7411,
    -1,
    8705,
    -1,
    -1,
    7412,
    7414,
    -1,
    -1,
    7413,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    3835,
    17206,
    -1,
    -1,
    -1,
    7415,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    7416,
    -1,
    2418,
    -1,
    17357,
    -1,
    7417,
    7418,
    -1,
    -1,
    3988,
    17358,
    17359,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    3050,
    17361,
    -1,
    -1,
    1413,
    3144,
    17362,
    7419,
    -1,
    3836,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    7422,
    7420,
    2318,
    -1,
    7421,
    2195,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8706,
    -1,
    -1,
    3890,
    7424,
    7426,
    7425,
    7427,
    -1,
    17363,
    1504,
    2875,
    2723,
    1842,
    7428,
    -1,
    -1,
    -1,
    3641,
    -1,
    7430,
    -1,
    -1,
    -1,
    1505,
    -1,
    7429,
    3325,
    -1,
    4274,
    3488,
    7423,
    4235,
    -1,
    2178,
    -1,
    -1,
    4194,
    -1,
    -1,
    -1,
    -1,
    17365,
    -1,
    8707,
    2062,
    4249,
    -1,
    2075,
    -1,
    3167,
    6580,
    -1,
    7431,
    1703,
    2898,
    -1,
    -1,
    -1,
    -1,
    1745,
    7433,
    -1,
    7434,
    7432,
    2135,
    17367,
    2410,
    2792,
    8709,
    -1,
    -1,
    1506,
    -1,
    -1,
    4286,
    8710,
    8711,
    -1,
    7436,
    7251,
    17370,
    7435,
    -1,
    -1,
    -1,
    -1,
    17371,
    7439,
    7437,
    7438,
    17372,
    7440,
    -1,
    7441,
    4305,
    7442,
    7443,
    17373,
    2954,
    3692,
    17374,
    -1,
    -1,
    2912,
    1856,
    -1,
    -1,
    4161,
    1674,
    2655,
    2220,
    -1,
    7446,
    -1,
    7445,
    2526,
    7447,
    7444,
    -1,
    -1,
    2451,
    8712,
    -1,
    -1,
    7451,
    6874,
    -1,
    17377,
    4659,
    17378,
    2906,
    7449,
    17379,
    8713,
    -1,
    -1,
    -1,
    4234,
    3558,
    -1,
    -1,
    -1,
    -1,
    1514,
    8715,
    2980,
    2559,
    -1,
    -1,
    -1,
    8716,
    3876,
    -1,
    1535,
    -1,
    -1,
    -1,
    4306,
    4210,
    -1,
    7452,
    -1,
    3420,
    -1,
    -1,
    -1,
    -1,
    2628,
    -1,
    -1,
    -1,
    7453,
    -1,
    7454,
    2863,
    7455,
    -1,
    4307,
    -1,
    -1,
    7450,
    7457,
    7459,
    -1,
    7458,
    -1,
    7456,
    17381,
    -1,
    7460,
    -1,
    -1,
    7461,
    -1,
    -1,
    3097,
    -1,
    1663,
    -1,
    -1,
    -1,
    -1,
    17382,
    7462,
    -1,
    -1,
    4065,
    -1,
    -1,
    7463,
    -1,
    -1,
    -1,
    -1,
    -1,
    7464,
    -1,
    4333,
    8717,
    -1,
    -1,
    17384,
    -1,
    5468,
    7465,
    -1,
    8718,
    -1,
    7466,
    7467,
    -1,
    -1,
    17385,
    7471,
    -1,
    7468,
    -1,
    7469,
    17386,
    7470,
    7472,
    -1,
    -1,
    -1,
    8719,
    8720,
    8721,
    -1,
    -1,
    2949,
    -1,
    -1,
    17387,
    4128,
    17388,
    -1,
    2950,
    8722,
    17389,
    7473,
    -1,
    3748,
    -1,
    7474,
    7758,
    4085,
    -1,
    7475,
    -1,
    7476,
    -1,
    7477,
    1746,
    17390,
    7479,
    -1,
    2876,
    8723,
    -1,
    -1,
    7480,
    -1,
    8724,
    2071,
    -1,
    -1,
    -1,
    -1,
    7481,
    7485,
    -1,
    7483,
    -1,
    -1,
    -1,
    -1,
    7484,
    -1,
    -1,
    1774,
    7482,
    7486,
    -1,
    -1,
    -1,
    -1,
    7487,
    -1,
    1449,
    -1,
    7488,
    7489,
    -1,
    -1,
    -1,
    -1,
    8725,
    17393,
    -1,
    2793,
    -1,
    17394,
    -1,
    7490,
    -1,
    17395,
    -1,
    1918,
    -1,
    17396,
    7493,
    -1,
    -1,
    7492,
    -1,
    7491,
    -1,
    -1,
    7036,
    -1,
    3913,
    8726,
    -1,
    -1,
    17397,
    17398,
    7494,
    7495,
    -1,
    -1,
    -1,
    -1,
    -1,
    8727,
    -1,
    -1,
    -1,
    17399,
    -1,
    -1,
    8728,
    -1,
    7496,
    -1,
    -1,
    7497,
    -1,
    7498,
    8729,
    -1,
    7499,
    17400,
    -1,
    -1,
    -1,
    -1,
    17401,
    -1,
    1843,
    17402,
    -1,
    -1,
    -1,
    17403,
    17404,
    -1,
    8731,
    7500,
    -1,
    8730,
    -1,
    -1,
    17405,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    7501,
    3574,
    -1,
    -1,
    17406,
    7503,
    1624,
    8732,
    7506,
    7505,
    -1,
    -1,
    -1,
    -1,
    1507,
    -1,
    -1,
    -1,
    2006,
    -1,
    3892,
    3315,
    2351,
    17407,
    2319,
    2702,
    -1,
    2879,
    -1,
    8734,
    -1,
    7508,
    17408,
    17409,
    7507,
    4169,
    1857,
    3715,
    3537,
    17410,
    -1,
    17411,
    2911,
    4275,
    -1,
    2126,
    -1,
    -1,
    -1,
    8735,
    -1,
    -1,
    7511,
    -1,
    17414,
    7510,
    17417,
    -1,
    -1,
    -1,
    -1,
    -1,
    8736,
    3990,
    3489,
    -1,
    -1,
    8737,
    -1,
    17415,
    17416,
    1551,
    -1,
    -1,
    7512,
    7509,
    -1,
    -1,
    3808,
    4209,
    7513,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    7514,
    17418,
    -1,
    -1,
    -1,
    7516,
    3175,
    1750,
    1751,
    7515,
    -1,
    -1,
    8739,
    8740,
    1858,
    2179,
    8741,
    8742,
    1859,
    8743,
    8744,
    3414,
    -1,
    -1,
    4293,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8746,
    17419,
    2221,
    -1,
    -1,
    -1,
    7517,
    8747,
    -1,
    -1,
    7518,
    7519,
    7520,
    -1,
    7522,
    7521,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    3846,
    -1,
    7523,
    17420,
    -1,
    17421,
    -1,
    7524,
    17422,
    7525,
    -1,
    -1,
    -1,
    -1,
    7526,
    17424,
    17425,
    -1,
    8748,
    17426,
    17427,
    -1,
    -1,
    17428,
    -1,
    -1,
    17429,
    7528,
    7527,
    -1,
    7529,
    17430,
    8749,
    -1,
    -1,
    17431,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    3749,
    6529,
    -1,
    -1,
    2828,
    17432,
    17434,
    1897,
    17435,
    -1,
    17436,
    -1,
    8750,
    -1,
    7530,
    17437,
    7531,
    -1,
    4617,
    5666,
    3716,
    17438,
    17439,
    1500,
    17440,
    1436,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    2527,
    3963,
    2819,
    -1,
    -1,
    -1,
    -1,
    7532,
    -1,
    4104,
    -1,
    -1,
    17441,
    7533,
    4195,
    -1,
    1537,
    -1,
    -1,
    -1,
    2476,
    -1,
    7534,
    1675,
    7535,
    -1,
    17444,
    17445,
    7536,
    -1,
    17447,
    17448,
    17449,
    7538,
    7539,
    -1,
    7541,
    7537,
    -1,
    -1,
    7540,
    -1,
    -1,
    -1,
    1844,
    -1,
    -1,
    -1,
    7542,
    -1,
    7543,
    -1,
    -1,
    17451,
    17452,
    17453,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17454,
    17455,
    17456,
    7544,
    7545,
    -1,
    17457,
    17458,
    7546,
    -1,
    -1,
    7548,
    17459,
    -1,
    -1,
    7547,
    -1,
    7550,
    7553,
    17460,
    17461,
    -1,
    7549,
    7551,
    7552,
    -1,
    -1,
    7554,
    -1,
    2007,
    8751,
    -1,
    -1,
    -1,
    17462,
    -1,
    -1,
    17463,
    17464,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    2620,
    7555,
    7556,
    2320,
    -1,
    -1,
    -1,
    -1,
    8752,
    -1,
    -1,
    -1,
    -1,
    17465,
    -1,
    7557,
    17466,
    -1,
    1721,
    -1,
    -1,
    -1,
    3618,
    7558,
    7559,
    -1,
    -1,
    -1,
    -1,
    3253,
    3552,
    -1,
    -1,
    -1,
    -1,
    17467,
    -1,
    -1,
    7560,
    17468,
    -1,
    17469,
    -1,
    3664,
    -1,
    17470,
    3145,
    1559,
    2051,
    -1,
    2052,
    17471,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    3278,
    7565,
    2053,
    -1,
    17472,
    1676,
    -1,
    -1,
    7564,
    17473,
    -1,
    7562,
    -1,
    7563,
    17474,
    7561,
    -1,
    -1,
    7575,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    7566,
    7567,
    -1,
    17476,
    7568,
    7569,
    -1,
    -1,
    -1,
    -1,
    -1,
    7571,
    17477,
    -1,
    7570,
    17478,
    -1,
    -1,
    2688,
    -1,
    7572,
    8753,
    8754,
    -1,
    7574,
    -1,
    -1,
    -1,
    -1,
    17479,
    -1,
    -1,
    -1,
    1898,
    7573,
    -1,
    17480,
    3098,
    2180,
    -1,
    -1,
    17481,
    -1,
    -1,
    7576,
    17482,
    -1,
    -1,
    -1,
    -1,
    -1,
    17483,
    -1,
    -1,
    -1,
    8755,
    -1,
    -1,
    -1,
    3146,
    -1,
    -1,
    7577,
    -1,
    8756,
    8757,
    -1,
    3490,
    17484,
    -1,
    -1,
    -1,
    -1,
    17485,
    7578,
    8758,
    -1,
    -1,
    -1,
    -1,
    -1,
    7583,
    -1,
    7581,
    -1,
    7580,
    7582,
    17486,
    7579,
    -1,
    -1,
    -1,
    -1,
    8759,
    -1,
    17487,
    7585,
    8760,
    -1,
    -1,
    -1,
    8761,
    -1,
    -1,
    7584,
    -1,
    7587,
    17488,
    -1,
    2008,
    7586,
    -1,
    -1,
    -1,
    7588,
    -1,
    -1,
    7589,
    -1,
    7591,
    7590,
    -1,
    -1,
    -1,
    7592,
    7594,
    7593,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    2347,
    -1,
    -1,
    -1,
    -1,
    7595,
    -1,
    17490,
    7596,
    -1,
    -1,
    -1,
    -1,
    -1,
    8762,
    17492,
    1719,
    17494,
    -1,
    -1,
    7597,
    -1,
    -1,
    -1,
    7598,
    8763,
    -1,
    8764,
    2900,
    -1,
    17496,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8765,
    7599,
    17497,
    7600,
    17498,
    7601,
    7602,
    17499,
    8766,
    -1,
    2321,
    -1,
    -1,
    -1,
    17501,
    -1,
    7603,
    7604,
    17502,
    -1,
    7605,
    7606,
    -1,
    17503,
    7607,
    -1,
    -1,
    17504,
    3682,
    7609,
    -1,
    3765,
    7610,
    7608,
    -1,
    7612,
    -1,
    -1,
    7611,
    -1,
    -1,
    7613,
    -1,
    8767,
    -1,
    7614,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8768,
    17505,
    -1,
    -1,
    7615,
    -1,
    8769,
    -1,
    -1,
    -1,
    17506,
    -1,
    -1,
    -1,
    17507,
    -1,
    17508,
    -1,
    -1,
    -1,
    17509,
    -1,
    7616,
    -1,
    7617,
    -1,
    17510,
    -1,
    -1,
    7618,
    8770,
    -1,
    7619,
    7620,
    -1,
    7621,
    -1,
    7622,
    7623,
    7624,
    7625,
    17511,
    -1,
    8772,
    7626,
    7627,
    -1,
    5660,
    7628,
    17512,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    6385,
    1899,
    17513,
    -1,
    -1,
    -1,
    1692,
    2367,
    7630,
    7629,
    4046,
    -1,
    -1,
    -1,
    -1,
    -1,
    17515,
    -1,
    7632,
    7633,
    7631,
    -1,
    7634,
    -1,
    -1,
    4013,
    -1,
    -1,
    -1,
    7635,
    -1,
    1970,
    -1,
    -1,
    -1,
    8773,
    -1,
    -1,
    -1,
    -1,
    17516,
    -1,
    17517,
    8775,
    -1,
    -1,
    -1,
    -1,
    17518,
    17519,
    -1,
    -1,
    4328,
    -1,
    -1,
    8776,
    17520,
    7636,
    8777,
    17521,
    17522,
    -1,
    8774,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    7638,
    8778,
    -1,
    -1,
    -1,
    -1,
    -1,
    8779,
    -1,
    -1,
    -1,
    1439,
    8780,
    -1,
    7639,
    3864,
    7637,
    -1,
    -1,
    7640,
    7641,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8781,
    7642,
    7643,
    -1,
    -1,
    -1,
    -1,
    -1,
    17523,
    8782,
    7644,
    -1,
    4025,
    2456,
    17524,
    2439,
    3022,
    -1,
    -1,
    17525,
    17528,
    -1,
    7645,
    -1,
    -1,
    -1,
    17529,
    7648,
    -1,
    -1,
    -1,
    -1,
    17530,
    -1,
    7646,
    8783,
    -1,
    -1,
    -1,
    -1,
    7649,
    17531,
    -1,
    2242,
    7647,
    -1,
    -1,
    -1,
    8784,
    7650,
    -1,
    7651,
    7652,
    -1,
    7656,
    -1,
    2453,
    -1,
    17533,
    -1,
    -1,
    3169,
    -1,
    17534,
    -1,
    -1,
    -1,
    7657,
    7654,
    7653,
    7655,
    8785,
    -1,
    17535,
    2130,
    -1,
    17536,
    17537,
    -1,
    -1,
    17539,
    17538,
    7661,
    7660,
    7659,
    17532,
    -1,
    1428,
    -1,
    17543,
    8786,
    -1,
    17541,
    -1,
    -1,
    8787,
    -1,
    -1,
    8788,
    -1,
    -1,
    -1,
    7671,
    -1,
    7667,
    -1,
    7668,
    7664,
    7670,
    -1,
    7666,
    1757,
    -1,
    -1,
    4367,
    -1,
    7669,
    7665,
    7663,
    7662,
    17545,
    -1,
    17546,
    17547,
    17548,
    7673,
    -1,
    17549,
    -1,
    -1,
    -1,
    7676,
    17550,
    8789,
    7675,
    7674,
    -1,
    17551,
    -1,
    17552,
    17553,
    -1,
    -1,
    3800,
    7672,
    1492,
    7677,
    17555,
    7679,
    -1,
    -1,
    -1,
    17556,
    17557,
    -1,
    1770,
    7658,
    1526,
    -1,
    -1,
    7681,
    -1,
    -1,
    8790,
    -1,
    -1,
    -1,
    17558,
    7680,
    7678,
    3208,
    17561,
    17562,
    -1,
    -1,
    -1,
    -1,
    8791,
    8792,
    -1,
    4027,
    8793,
    17564,
    -1,
    -1,
    4287,
    17565,
    -1,
    7682,
    17566,
    17559,
    17567,
    -1,
    17568,
    7683,
    -1,
    -1,
    8794,
    -1,
    8795,
    -1,
    7684,
    -1,
    17569,
    17570,
    17571,
    -1,
    17572,
    17573,
    -1,
    17574,
    -1,
    17575,
    -1,
    -1,
    17576,
    7685,
    8796,
    7686,
    -1,
    17577,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    3316,
    17578,
    7687,
    -1,
    3688,
    -1,
    7692,
    7688,
    -1,
    -1,
    -1,
    7689,
    -1,
    17579,
    3964,
    4077,
    -1,
    3525,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17581,
    7693,
    -1,
    -1,
    7694,
    3507,
    7691,
    7690,
    -1,
    17580,
    -1,
    -1,
    1607,
    -1,
    -1,
    17583,
    7702,
    -1,
    -1,
    7701,
    -1,
    17584,
    17585,
    -1,
    -1,
    1588,
    -1,
    8797,
    8798,
    7699,
    -1,
    -1,
    -1,
    7698,
    -1,
    -1,
    7696,
    -1,
    1782,
    -1,
    7695,
    2555,
    1606,
    -1,
    -1,
    -1,
    -1,
    -1,
    17589,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    2322,
    -1,
    -1,
    7705,
    7704,
    -1,
    7703,
    17591,
    8799,
    7700,
    -1,
    7706,
    8800,
    7707,
    -1,
    17592,
    -1,
    -1,
    -1,
    -1,
    -1,
    7712,
    7711,
    8801,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    7713,
    -1,
    -1,
    1516,
    7708,
    7709,
    17593,
    2339,
    4066,
    17594,
    8802,
    7710,
    -1,
    -1,
    -1,
    -1,
    17596,
    -1,
    17597,
    3965,
    -1,
    -1,
    7718,
    8803,
    -1,
    7714,
    17599,
    -1,
    -1,
    17600,
    17601,
    -1,
    -1,
    7719,
    -1,
    8804,
    -1,
    17602,
    -1,
    -1,
    -1,
    -1,
    -1,
    17603,
    -1,
    -1,
    7716,
    -1,
    7715,
    8805,
    -1,
    -1,
    17604,
    -1,
    2127,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8806,
    -1,
    -1,
    17605,
    7720,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17606,
    -1,
    -1,
    7721,
    -1,
    -1,
    -1,
    -1,
    7722,
    -1,
    7717,
    8808,
    -1,
    -1,
    7697,
    -1,
    -1,
    7723,
    -1,
    3354,
    17608,
    -1,
    -1,
    7727,
    17609,
    7728,
    7726,
    8809,
    17610,
    -1,
    17607,
    8807,
    7725,
    7731,
    17611,
    7724,
    -1,
    7729,
    17612,
    -1,
    17613,
    -1,
    -1,
    -1,
    -1,
    -1,
    7730,
    -1,
    -1,
    -1,
    7733,
    -1,
    -1,
    17614,
    8810,
    -1,
    7732,
    17615,
    -1,
    -1,
    -1,
    -1,
    17616,
    17617,
    -1,
    -1,
    17618,
    -1,
    -1,
    7735,
    8812,
    -1,
    -1,
    -1,
    -1,
    -1,
    7736,
    -1,
    7737,
    -1,
    -1,
    4364,
    -1,
    17619,
    -1,
    -1,
    -1,
    7734,
    3176,
    2426,
    -1,
    -1,
    7738,
    -1,
    -1,
    -1,
    -1,
    17622,
    -1,
    -1,
    -1,
    -1,
    8813,
    -1,
    -1,
    17621,
    -1,
    -1,
    17623,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8814,
    -1,
    -1,
    -1,
    17624,
    7739,
    7740,
    17625,
    17626,
    7741,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    7742,
    -1,
    -1,
    2181,
    7743,
    -1,
    17627,
    8815,
    7744,
    -1,
    2552,
    17629,
    7745,
    -1,
    -1,
    -1,
    17630,
    -1,
    -1,
    7746,
    -1,
    -1,
    7747,
    7748,
    -1,
    -1,
    -1,
    -1,
    7751,
    7749,
    4351,
    -1,
    7750,
    -1,
    4308,
    -1,
    -1,
    -1,
    17631,
    -1,
    7752,
    8816,
    4288,
    -1,
    -1,
    -1,
    -1,
    8817,
    7753,
    3665,
    -1,
    17632,
    7754,
    7756,
    -1,
    8818,
    7757,
    -1,
    8819,
    -1,
    -1,
    -1,
    -1,
    8820,
    8821,
    -1,
    -1,
    7755,
    2332,
    4086,
    4014,
    5068,
    17634,
    5703,
    4038,
    -1,
    -1,
    -1,
    8822,
    1608,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    7759,
    1927,
    7760,
    7761,
    7762,
    8823,
    2340,
    -1,
    7763,
    -1,
    -1,
    -1,
    5973,
    4100,
    -1,
    3168,
    7764,
    7766,
    7765,
    17636,
    7767,
    -1,
    -1,
    -1,
    -1,
    7768,
    -1,
    17637,
    7769,
    -1,
    -1,
    -1,
    -1,
    -1,
    17638,
    7770,
    -1,
    -1,
    -1,
    -1,
    7771,
    -1,
    7772,
    7773,
    -1,
    7774,
    -1,
    7775,
    7776,
    7777,
    -1,
    17639,
    -1,
    -1,
    17640,
    17642,
    -1,
    -1,
    -1,
    7778,
    7779,
    -1,
    -1,
    -1,
    -1,
    -1,
    3387,
    -1,
    8824,
    -1,
    -1,
    2222,
    -1,
    7781,
    -1,
    17643,
    -1,
    17644,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    3052,
    7782,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    7783,
    -1,
    -1,
    17645,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17646,
    -1,
    8825,
    17647,
    3760,
    -1,
    17648,
    7784,
    -1,
    -1,
    17649,
    -1,
    -1,
    -1,
    17650,
    17651,
    -1,
    -1,
    -1,
    7785,
    6226,
    -1,
    -1,
    7107,
    7502,
    -1,
    -1,
    7786,
    17652,
    7787,
    17653,
    -1,
    8826,
    17654,
    -1,
    -1,
    -1,
    -1,
    17656,
    -1,
    7789,
    7790,
    7791,
    4309,
    7788,
    -1,
    -1,
    7792,
    7793,
    -1,
    17658,
    7795,
    -1,
    7794,
    17660,
    -1,
    -1,
    17661,
    -1,
    7797,
    -1,
    -1,
    17662,
    7798,
    7796,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    4251,
    -1,
    -1,
    8827,
    -1,
    -1,
    -1,
    8828,
    7799,
    -1,
    8829,
    -1,
    -1,
    -1,
    -1,
    7800,
    6255,
    -1,
    -1,
    7801,
    -1,
    8830
};
static const uint16_t jis_urange_00f91d_00fa6a[334] = {
    8016,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    7815,
    7941,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8506,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8030,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8745,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8708,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1358,
    1370,
    4405,
    -1,
    10146,
    7985,
    8141,
    8162,
    -1,
    -1,
    8299,
    8300,
    8304,
    -1,
    -1,
    -1,
    8485,
    16943,
    16956,
    8567,
    -1,
    17185,
    -1,
    8627,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1245,
    1262,
    1269,
    1288,
    1293,
    1299,
    1327,
    1330,
    1337,
    1373,
    1377,
    4388,
    4389,
    7849,
    7861,
    7863,
    7866,
    7903,
    7906,
    7930,
    7964,
    8062,
    8076,
    8088,
    8136,
    16270,
    8182,
    8278,
    8290,
    8291,
    8294,
    8295,
    8296,
    8298,
    8302,
    8303,
    8316,
    8320,
    8339,
    8379,
    16685,
    8384,
    8391,
    8401,
    8421,
    16815,
    16816,
    8466,
    8538,
    8548,
    8568,
    8569,
    8577,
    8582,
    17180,
    8610,
    8714,
    8733,
    8738
};
static const uint16_t jis_urange_00fe45_00ffe5[417] = {
    217,
    216,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9,
    109,
    83,
    79,
    82,
    84,
    108,
    41,
    42,
    85,
    59,
    3,
    60,
    4,
    30,
    203,
    204,
    205,
    206,
    207,
    208,
    209,
    210,
    211,
    212,
    6,
    7,
    66,
    64,
    67,
    8,
    86,
    220,
    221,
    222,
    223,
    224,
    225,
    226,
    227,
    228,
    229,
    230,
    231,
    232,
    233,
    234,
    235,
    236,
    237,
    238,
    239,
    240,
    241,
    242,
    243,
    244,
    245,
    45,
    31,
    46,
    15,
    17,
    13,
    252,
    253,
    254,
    255,
    256,
    257,
    258,
    259,
    260,
    261,
    262,
    263,
    264,
    265,
    266,
    267,
    268,
    269,
    270,
    271,
    272,
    273,
    274,
    275,
    276,
    277,
    47,
    34,
    48,
    32,
    147,
    148,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    80,
    81,
    137,
    16,
    -1,
    78
};
static const uint16_t jis_urange_02000b_020213[521] = {
    1223,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8836,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8846,
    -1,
    8849,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8857,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8873
};
static const uint16_t jis_urange_02032b_020628[766] = {
    8915,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8924,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8922,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9025,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9028,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9030,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9040,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9041
};
static const uint16_t jis_urange_02074f_02099d[591] = {
    9047,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9054,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9056,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9065,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9073,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9074
};
static const uint16_t jis_urange_020ad3_020b9f[205] = {
    9080,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9083,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    4375
};
static const uint16_t jis_urange_020d45_020f5f[539] = {
    9110,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9127,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9143,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9135,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9134,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9146
};
static const uint16_t jis_urange_021201_02146e[622] = {
    9174,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1349,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9177,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9184,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9179,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9192,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9191,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9199,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    1359,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9201,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9202,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9216,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9229,
    1379
};
static const uint16_t jis_urange_0215d7_021742[364] = {
    9241,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9250,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    4390,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9264,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9265
};
static const uint16_t jis_urange_0218bd_0218bd[1] = {
    1406
};
static const uint16_t jis_urange_0219c3_0219c3[1] = {
    9305
};
static const uint16_t jis_urange_021c56_021ffa[933] = {
    9509,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9516,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9517,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9520,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9519,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9530,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9525,
    -1,
    -1,
    -1,
    -1,
    9524,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9533,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9535,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9536,
    4401,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9552,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9563,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9569
};
static const uint16_t jis_urange_02217b_022218[158] = {
    9871,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17628
};
static const uint16_t jis_urange_02231e_0223ad[144] = {
    9880,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9885
};
static const uint16_t jis_urange_0226f3_0226f3[1] = {
    9917
};
static const uint16_t jis_urange_02285b_02298f[309] = {
    9938,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9946,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9951
};
static const uint16_t jis_urange_022ab8_022c24[365] = {
    9967,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9981,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9972,
    9973,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9984,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9983,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    9988
};
static const uint16_t jis_urange_022de1_022de1[1] = {
    10017
};
static const uint16_t jis_urange_0231b6_0231f5[64] = {
    10061,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    10056,
    7921,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    10060
};
static const uint16_t jis_urange_023372_02383d[1228] = {
    10083,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    10091,
    -1,
    10086,
    10085,
    -1,
    10093,
    -1,
    -1,
    -1,
    -1,
    10096,
    -1,
    -1,
    -1,
    -1,
    10098,
    -1,
    -1,
    -1,
    -1,
    10092,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    10110,
    10112,
    -1,
    -1,
    -1,
    -1,
    -1,
    10111,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    10116,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    10144,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    10145,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    10161,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    7977,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    10185,
    10182,
    10186,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    10183,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    10208,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    10197,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    7998,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8007,
    10216,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    10225,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    10224,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    10231,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    10236
};
static const uint16_t jis_urange_023a98_023a98[1] = {
    16076
};
static const uint16_t jis_urange_023c7f_023dfa[380] = {
    16093,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8053,
    -1,
    16114,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16722,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16133,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16137,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16136,
    16135
};
static const uint16_t jis_urange_023f7e_023f7e[1] = {
    16185
};
static const uint16_t jis_urange_024096_0241fe[361] = {
    16206,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16212,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16232,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16235
};
static const uint16_t jis_urange_0243bc_0243bc[1] = {
    16263
};
static const uint16_t jis_urange_024629_0246a5[125] = {
    16280,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16286
};
static const uint16_t jis_urange_0247f1_024896[166] = {
    8159,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16312
};
static const uint16_t jis_urange_024a4d_024a4d[1] = {
    16352
};
static const uint16_t jis_urange_024b56_024e8b[822] = {
    16369,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16371,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16376,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16391,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16416,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16420,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16425,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16428
};
static const uint16_t jis_urange_02504a_02524c[515] = {
    16440,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16442,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16445,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16450,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16453,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16452,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16457,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16461
};
static const uint16_t jis_urange_02542e_0255a7[378] = {
    16478,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8271,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16489,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8280,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16508
};
static const uint16_t jis_urange_025771_0257b4[68] = {
    8310,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16534,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16535
};
static const uint16_t jis_urange_0259c4_0259d4[17] = {
    8323,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16561
};
static const uint16_t jis_urange_025ae3_025c64[386] = {
    16573,
    16572,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16575,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16593,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16604,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16605
};
static const uint16_t jis_urange_025da1_0260ed[845] = {
    8349,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16626,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16627,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16630,
    -1,
    -1,
    16628,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16636,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16634,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16639,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16641,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16644,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16652,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16651,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16658,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16657,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16666,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16671,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16684
};
static const uint16_t jis_urange_026270_026402[403] = {
    16710,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16712,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16719,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16726
};
static const uint16_t jis_urange_02667e_02671d[160] = {
    16752,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16757,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16769
};
static const uint16_t jis_urange_0268dd_026aff[547] = {
    16785,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16787,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16789,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16792,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16794,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16798,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16805,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16811,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16814,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8426
};
static const uint16_t jis_urange_026c29_026cdd[181] = {
    9247,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16861,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16871
};
static const uint16_t jis_urange_026e40_026e65[38] = {
    8478,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16886
};
static const uint16_t jis_urange_026f94_027139[422] = {
    16910,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16920,
    16921,
    16919,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8500,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16938,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16941
};
static const uint16_t jis_urange_0273da_027449[112] = {
    16988,
    16987,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16994,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    16997,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17002
};
static const uint16_t jis_urange_027614_027752[319] = {
    17023,
    17022,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17025,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8535,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17033,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17041,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17043,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17047
};
static const uint16_t jis_urange_027985_027a84[256] = {
    17067,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17080
};
static const uint16_t jis_urange_027bb3_027e10[606] = {
    17100,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17102,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17103,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17112,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17122,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17125
};
static const uint16_t jis_urange_027fb7_0280bb[261] = {
    17132,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17141,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17147
};
static const uint16_t jis_urange_028277_028455[479] = {
    8594,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17164,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17169,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8602,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17177,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17182
};
static const uint16_t jis_urange_02856b_0285c9[95] = {
    17197,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17201,
    17202
};
static const uint16_t jis_urange_0286d7_0286fa[36] = {
    17213,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17216
};
static const uint16_t jis_urange_028946_028bef[682] = {
    17249,
    -1,
    -1,
    17248,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17256,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17270,
    17271,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17282,
    17283,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17295,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17296,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17308,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17307,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17317,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17318,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17325,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17324,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17341,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17342
};
static const uint16_t jis_urange_028d10_028ff8[745] = {
    17350,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17353,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17355,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17356,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17360,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17364,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17366,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17368,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17376
};
static const uint16_t jis_urange_0292a0_0292b1[18] = {
    17391,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17392
};
static const uint16_t jis_urange_029490_029490[1] = {
    17413
};
static const uint16_t jis_urange_0295cf_029750[386] = {
    17423,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17433,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17443,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17446,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17450
};
static const uint16_t jis_urange_0298c6_0298c6[1] = {
    17475
};
static const uint16_t jis_urange_029a72_029a72[1] = {
    17495
};
static const uint16_t jis_urange_029ddb_02a190[950] = {
    17526,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17540,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17527,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17544,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17542,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17554,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17563,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17560,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17582,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17588,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17586,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17598,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17595,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    8811
};
static const uint16_t jis_urange_02a38c_02a437[172] = {
    17633,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17635
};
static const uint16_t jis_urange_02a5f1_02a6b2[194] = {
    17655,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17657,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17659,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    17663
};

static const URangeToJISMapping urange_to_jis_mappings[] = {
    {160, 1105, jis_urange_0000a0_000451},
    {7742, 7743, jis_urange_001e3e_001e3f},
    {8048, 10111, jis_urange_001f70_00277f},
    {10548, 10747, jis_urange_002934_0029fb},
    {12288, 40866, jis_urange_003000_009fa2},
    {63773, 64106, jis_urange_00f91d_00fa6a},
    {65093, 65509, jis_urange_00fe45_00ffe5},
    {131083, 131603, jis_urange_02000b_020213},
    {131883, 132648, jis_urange_02032b_020628},
    {132943, 133533, jis_urange_02074f_02099d},
    {133843, 134047, jis_urange_020ad3_020b9f},
    {134469, 135007, jis_urange_020d45_020f5f},
    {135681, 136302, jis_urange_021201_02146e},
    {136663, 137026, jis_urange_0215d7_021742},
    {137405, 137405, jis_urange_0218bd_0218bd},
    {137667, 137667, jis_urange_0219c3_0219c3},
    {138326, 139258, jis_urange_021c56_021ffa},
    {139643, 139800, jis_urange_02217b_022218},
    {140062, 140205, jis_urange_02231e_0223ad},
    {141043, 141043, jis_urange_0226f3_0226f3},
    {141403, 141711, jis_urange_02285b_02298f},
    {142008, 142372, jis_urange_022ab8_022c24},
    {142817, 142817, jis_urange_022de1_022de1},
    {143798, 143861, jis_urange_0231b6_0231f5},
    {144242, 145469, jis_urange_023372_02383d},
    {146072, 146072, jis_urange_023a98_023a98},
    {146559, 146938, jis_urange_023c7f_023dfa},
    {147326, 147326, jis_urange_023f7e_023f7e},
    {147606, 147966, jis_urange_024096_0241fe},
    {148412, 148412, jis_urange_0243bc_0243bc},
    {149033, 149157, jis_urange_024629_0246a5},
    {149489, 149654, jis_urange_0247f1_024896},
    {150093, 150093, jis_urange_024a4d_024a4d},
    {150358, 151179, jis_urange_024b56_024e8b},
    {151626, 152140, jis_urange_02504a_02524c},
    {152622, 152999, jis_urange_02542e_0255a7},
    {153457, 153524, jis_urange_025771_0257b4},
    {154052, 154068, jis_urange_0259c4_0259d4},
    {154339, 154724, jis_urange_025ae3_025c64},
    {155041, 155885, jis_urange_025da1_0260ed},
    {156272, 156674, jis_urange_026270_026402},
    {157310, 157469, jis_urange_02667e_02671d},
    {157917, 158463, jis_urange_0268dd_026aff},
    {158761, 158941, jis_urange_026c29_026cdd},
    {159296, 159333, jis_urange_026e40_026e65},
    {159636, 160057, jis_urange_026f94_027139},
    {160730, 160841, jis_urange_0273da_027449},
    {161300, 161618, jis_urange_027614_027752},
    {162181, 162436, jis_urange_027985_027a84},
    {162739, 163344, jis_urange_027bb3_027e10},
    {163767, 164027, jis_urange_027fb7_0280bb},
    {164471, 164949, jis_urange_028277_028455},
    {165227, 165321, jis_urange_02856b_0285c9},
    {165591, 165626, jis_urange_0286d7_0286fa},
    {166214, 166895, jis_urange_028946_028bef},
    {167184, 167928, jis_urange_028d10_028ff8},
    {168608, 168625, jis_urange_0292a0_0292b1},
    {169104, 169104, jis_urange_029490_029490},
    {169423, 169808, jis_urange_0295cf_029750},
    {170182, 170182, jis_urange_0298c6_0298c6},
    {170610, 170610, jis_urange_029a72_029a72},
    {171483, 172432, jis_urange_029ddb_02a190},
    {172940, 173111, jis_urange_02a38c_02a437},
    {173553, 173746, jis_urange_02a5f1_02a6b2}
};

typedef struct SMUniToJISTuple {
    int state;
    uint32_t u;
} SMUniToJISTuple;

static uint16_t sm_uni_to_jis_mapping(int *state, uint32_t u)
{
    uint16_t j = 0;
    int _state = *state;
reenter:
    switch (_state) {
    case 0:
        if (u < 230 || u > 12791) {
            break;
        }
        switch (u) {
        case 230:
            _state = 1;
            break;
        case 596:
            _state = 2;
            break;
        case 601:
            _state = 3;
            break;
        case 602:
            _state = 4;
            break;
        case 652:
            _state = 5;
            break;
        case 741:
            _state = 6;
            break;
        case 745:
            _state = 7;
            break;
        case 12363:
            _state = 8;
            break;
        case 12365:
            _state = 9;
            break;
        case 12367:
            _state = 10;
            break;
        case 12369:
            _state = 11;
            break;
        case 12371:
            _state = 12;
            break;
        case 12459:
            _state = 13;
            break;
        case 12461:
            _state = 14;
            break;
        case 12463:
            _state = 15;
            break;
        case 12465:
            _state = 16;
            break;
        case 12467:
            _state = 17;
            break;
        case 12475:
            _state = 18;
            break;
        case 12484:
            _state = 19;
            break;
        case 12488:
            _state = 20;
            break;
        case 12791:
            _state = 21;
            break;
        }
        break;
    case 1:
        switch (u) {
        case 768:
            j = 975;
            _state = -1;
            break;
        default:
            _state = 0;
            goto reenter;
        }
        break;
    case 2:
        switch (u) {
        case 768:
            j = 979;
            _state = -1;
            break;
        case 769:
            j = 980;
            _state = -1;
            break;
        default:
            _state = 0;
            goto reenter;
        }
        break;
    case 3:
        switch (u) {
        case 768:
            j = 983;
            _state = -1;
            break;
        case 769:
            j = 984;
            _state = -1;
            break;
        default:
            _state = 0;
            goto reenter;
        }
        break;
    case 4:
        switch (u) {
        case 768:
            j = 985;
            _state = -1;
            break;
        case 769:
            j = 986;
            _state = -1;
            break;
        default:
            _state = 0;
            goto reenter;
        }
        break;
    case 5:
        switch (u) {
        case 768:
            j = 981;
            _state = -1;
            break;
        case 769:
            j = 982;
            _state = -1;
            break;
        default:
            _state = 0;
            goto reenter;
        }
        break;
    case 6:
        switch (u) {
        case 745:
            j = 1009;
            _state = -1;
            break;
        default:
            _state = 0;
            goto reenter;
        }
        break;
    case 7:
        switch (u) {
        case 741:
            j = 1008;
            _state = -1;
            break;
        default:
            _state = 0;
            goto reenter;
        }
        break;
    case 8:
        switch (u) {
        case 12442:
            j = 368;
            _state = -1;
            break;
        default:
            _state = 0;
            goto reenter;
        }
        break;
    case 9:
        switch (u) {
        case 12442:
            j = 369;
            _state = -1;
            break;
        default:
            _state = 0;
            goto reenter;
        }
        break;
    case 10:
        switch (u) {
        case 12442:
            j = 370;
            _state = -1;
            break;
        default:
            _state = 0;
            goto reenter;
        }
        break;
    case 11:
        switch (u) {
        case 12442:
            j = 371;
            _state = -1;
            break;
        default:
            _state = 0;
            goto reenter;
        }
        break;
    case 12:
        switch (u) {
        case 12442:
            j = 372;
            _state = -1;
            break;
        default:
            _state = 0;
            goto reenter;
        }
        break;
    case 13:
        switch (u) {
        case 12442:
            j = 462;
            _state = -1;
            break;
        default:
            _state = 0;
            goto reenter;
        }
        break;
    case 14:
        switch (u) {
        case 12442:
            j = 463;
            _state = -1;
            break;
        default:
            _state = 0;
            goto reenter;
        }
        break;
    case 15:
        switch (u) {
        case 12442:
            j = 464;
            _state = -1;
            break;
        default:
            _state = 0;
            goto reenter;
        }
        break;
    case 16:
        switch (u) {
        case 12442:
            j = 465;
            _state = -1;
            break;
        default:
            _state = 0;
            goto reenter;
        }
        break;
    case 17:
        switch (u) {
        case 12442:
            j = 466;
            _state = -1;
            break;
        default:
            _state = 0;
            goto reenter;
        }
        break;
    case 18:
        switch (u) {
        case 12442:
            j = 467;
            _state = -1;
            break;
        default:
            _state = 0;
            goto reenter;
        }
        break;
    case 19:
        switch (u) {
        case 12442:
            j = 468;
            _state = -1;
            break;
        default:
            _state = 0;
            goto reenter;
        }
        break;
    case 20:
        switch (u) {
        case 12442:
            j = 469;
            _state = -1;
            break;
        default:
            _state = 0;
            goto reenter;
        }
        break;
    case 21:
        switch (u) {
        case 12442:
            j = 557;
            _state = -1;
            break;
        default:
            _state = 0;
            goto reenter;
        }
        break;
    }
    *state = _state;
    return j;
}

typedef enum MJShrinkScheme {
    JISCharacterClass_JIS_INCORPORATION_UCS_UNIFICATION_RULE = 0,
    JISCharacterClass_INFERENCE_BY_READING_AND_GLYPH = 1,
    JISCharacterClass_MOJ_NOTICE_582 = 2,
    JISCharacterClass_MOJ_FAMILY_REGISTER_ACT_RELATED_NOTICE = 3
} MJShrinkScheme;

typedef struct UIVSPair {
    uint32_t u:22;
    unsigned int v:1;
    unsigned int sv:1;
    uint8_t s;
} UIVSPair;

typedef struct MJMapping {
    uint32_t mj:24;
    UIVSPair v[5];
} MJMapping;

typedef struct MJMappingSet {
    size_t l:8;
    MJMapping ms[32];
} MJMappingSet;

typedef struct URangeToMJMappings {
    uint32_t start:24, end:24;
    const MJMappingSet* mss;
} URangeToMJMappings;

typedef struct MJShrinkMappingUnicodeSet {
    uint32_t _0[2];
    uint32_t _1[1];
    uint32_t _2[2];
    uint32_t _3[3];
} MJShrinkMappingUnicodeSet;


static const MJMappingSet chunked_mj_mappings_003005_00303b[55] = {
    {1, {{1, {{12293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2, {{12294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3, {{12347, 1, 0, 0}, {0, 0, 0, 0}}}}}
};

static const MJMappingSet chunked_mj_mappings_003400_004db5[6582] = {
    {1, {{4, {{13312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5, {{13313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6, {{13314, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{7, {{13316, 1, 0, 0},{13316, 1, 1, 17}, {0, 0, 0, 0}}},{7, {{13316, 1, 0, 0},{13316, 1, 1, 17}, {0, 0, 0, 0}}},{8, {{13316, 1, 0, 0},{13316, 1, 1, 16}, {0, 0, 0, 0}}},{8, {{13316, 1, 0, 0},{13316, 1, 1, 16}, {0, 0, 0, 0}}},{68055, {{13316, 1, 0, 0},{13316, 1, 1, 18}, {0, 0, 0, 0}}},{68055, {{13316, 1, 0, 0},{13316, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9, {{13317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10, {{13318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67953, {{13319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67954, {{13320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67955, {{13321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67956, {{13322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67957, {{13323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11, {{13324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67958, {{13325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67959, {{13326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67960, {{13327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67961, {{13328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67962, {{13329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67963, {{13330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67964, {{13331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67965, {{13332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67966, {{13333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13, {{13334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67967, {{13335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67968, {{13336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67969, {{13337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67970, {{13338, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67971, {{13339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15, {{13340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67972, {{13341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67973, {{13342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67974, {{13343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67975, {{13344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16, {{13345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67976, {{13346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67977, {{13347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17, {{13348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67978, {{13349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67979, {{13350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18, {{13351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20, {{13352, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21, {{13353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22, {{13354, 1, 0, 0},{13354, 1, 1, 19}, {0, 0, 0, 0}}},{22, {{13354, 1, 0, 0},{13354, 1, 1, 19}, {0, 0, 0, 0}}},{23, {{13354, 1, 0, 0},{13354, 1, 1, 17}, {0, 0, 0, 0}}},{23, {{13354, 1, 0, 0},{13354, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{57038, {{13355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25, {{13356, 1, 0, 0},{13356, 1, 1, 17}, {0, 0, 0, 0}}},{25, {{13356, 1, 0, 0},{13356, 1, 1, 17}, {0, 0, 0, 0}}},{57397, {{13356, 1, 0, 0},{13356, 1, 1, 18}, {0, 0, 0, 0}}},{57397, {{13356, 1, 0, 0},{13356, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26, {{13357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{27, {{13358, 1, 0, 0},{13358, 1, 1, 17}, {0, 0, 0, 0}}},{27, {{13358, 1, 0, 0},{13358, 1, 1, 17}, {0, 0, 0, 0}}},{28, {{13358, 1, 0, 0},{13358, 1, 1, 18}, {0, 0, 0, 0}}},{28, {{13358, 1, 0, 0},{13358, 1, 1, 18}, {0, 0, 0, 0}}},{29, {{13358, 1, 0, 0},{13358, 1, 1, 19}, {0, 0, 0, 0}}},{29, {{13358, 1, 0, 0},{13358, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{30, {{13359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31, {{13360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32, {{13361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33, {{13362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34, {{13363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35, {{13364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{36, {{13366, 1, 0, 0},{13366, 1, 1, 16}, {0, 0, 0, 0}}},{36, {{13366, 1, 0, 0},{13366, 1, 1, 16}, {0, 0, 0, 0}}},{37, {{13366, 1, 0, 0},{13366, 1, 1, 17}, {0, 0, 0, 0}}},{37, {{13366, 1, 0, 0},{13366, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38, {{13368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39, {{13369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40, {{13370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41, {{13371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42, {{13372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{45, {{13377, 1, 0, 0},{13377, 1, 1, 16}, {0, 0, 0, 0}}},{45, {{13377, 1, 0, 0},{13377, 1, 1, 16}, {0, 0, 0, 0}}},{46, {{13377, 1, 0, 0},{13377, 1, 1, 17}, {0, 0, 0, 0}}},{46, {{13377, 1, 0, 0},{13377, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{47, {{13378, 1, 0, 0},{13378, 1, 1, 16}, {0, 0, 0, 0}}},{47, {{13378, 1, 0, 0},{13378, 1, 1, 16}, {0, 0, 0, 0}}},{48, {{13378, 1, 0, 0},{13378, 1, 1, 17}, {0, 0, 0, 0}}},{48, {{13378, 1, 0, 0},{13378, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{49, {{13379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50, {{13380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51, {{13381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52, {{13382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53, {{13385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54, {{13386, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55, {{13387, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56, {{13388, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57, {{13389, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59, {{13390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60, {{13392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{61, {{13393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{62, {{13394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{63, {{13395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{64, {{13399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{65, {{13400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{66, {{13401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67, {{13402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{68, {{13403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{69, {{13404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{70, {{13406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{71, {{13407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{72, {{13411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{73, {{13412, 1, 0, 0},{13412, 1, 1, 17}, {0, 0, 0, 0}}},{73, {{13412, 1, 0, 0},{13412, 1, 1, 17}, {0, 0, 0, 0}}},{74, {{13412, 1, 0, 0},{13412, 1, 1, 16}, {0, 0, 0, 0}}},{74, {{13412, 1, 0, 0},{13412, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{75, {{13413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{76, {{13414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{77, {{13415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{78, {{13416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{79, {{13418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{80, {{13422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{81, {{13423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{82, {{13425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{83, {{13427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{84, {{13428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{85, {{13429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{86, {{13430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{87, {{13431, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{88, {{13432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{89, {{13433, 1, 0, 0},{13433, 1, 1, 16}, {0, 0, 0, 0}}},{89, {{13433, 1, 0, 0},{13433, 1, 1, 16}, {0, 0, 0, 0}}},{13511, {{13433, 1, 0, 0},{13433, 1, 1, 17}, {0, 0, 0, 0}}},{13511, {{13433, 1, 0, 0},{13433, 1, 1, 17}, {0, 0, 0, 0}}},{57674, {{13433, 1, 0, 0},{13433, 1, 1, 18}, {0, 0, 0, 0}}},{57674, {{13433, 1, 0, 0},{13433, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{90, {{13435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{91, {{13436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{92, {{13437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{93, {{13438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{94, {{13439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{95, {{13440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{96, {{13441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{97, {{13442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{98, {{13443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{99, {{13444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{100, {{13445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{101, {{13446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{102, {{13447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{103, {{13448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{104, {{13449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{105, {{13450, 1, 0, 0},{13450, 1, 1, 17}, {0, 0, 0, 0}}},{105, {{13450, 1, 0, 0},{13450, 1, 1, 17}, {0, 0, 0, 0}}},{106, {{13450, 1, 0, 0},{13450, 1, 1, 16}, {0, 0, 0, 0}}},{106, {{13450, 1, 0, 0},{13450, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{107, {{13451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{108, {{13452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{109, {{13453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{110, {{13454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{111, {{13457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{112, {{13458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{113, {{13459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{114, {{13460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{115, {{13461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{116, {{13462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{117, {{13465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{118, {{13466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{119, {{13467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{120, {{13468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{121, {{13469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{122, {{13470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{123, {{13471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{124, {{13472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{125, {{13473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{126, {{13476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{127, {{13477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{128, {{13478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{129, {{13479, 1, 0, 0},{13479, 1, 1, 16}, {0, 0, 0, 0}}},{129, {{13479, 1, 0, 0},{13479, 1, 1, 16}, {0, 0, 0, 0}}},{130, {{13479, 1, 0, 0},{13479, 1, 1, 17}, {0, 0, 0, 0}}},{130, {{13479, 1, 0, 0},{13479, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{131, {{13480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{132, {{13481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{133, {{13482, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{134, {{13483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{135, {{13485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{136, {{13487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{137, {{13488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{138, {{13490, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{139, {{13491, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{140, {{13492, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{141, {{13493, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{142, {{13494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{143, {{13496, 1, 0, 0},{13496, 1, 1, 16}, {0, 0, 0, 0}}},{143, {{13496, 1, 0, 0},{13496, 1, 1, 16}, {0, 0, 0, 0}}},{144, {{13496, 1, 0, 0},{13496, 1, 1, 17}, {0, 0, 0, 0}}},{144, {{13496, 1, 0, 0},{13496, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{145, {{13497, 1, 0, 0},{13497, 1, 1, 17}, {0, 0, 0, 0}}},{145, {{13497, 1, 0, 0},{13497, 1, 1, 17}, {0, 0, 0, 0}}},{146, {{13497, 1, 0, 0},{13497, 1, 1, 16}, {0, 0, 0, 0}}},{146, {{13497, 1, 0, 0},{13497, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{147, {{13498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{148, {{13499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{150, {{13500, 1, 0, 0},{13500, 1, 1, 17}, {0, 0, 0, 0}}},{150, {{13500, 1, 0, 0},{13500, 1, 1, 17}, {0, 0, 0, 0}}},{57021, {{13500, 1, 0, 0},{13500, 1, 1, 18}, {0, 0, 0, 0}}},{57021, {{13500, 1, 0, 0},{13500, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{151, {{13501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{152, {{13502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{153, {{13503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{154, {{13505, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{155, {{13506, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{156, {{13507, 1, 0, 0},{13507, 1, 1, 16}, {0, 0, 0, 0}}},{156, {{13507, 1, 0, 0},{13507, 1, 1, 16}, {0, 0, 0, 0}}},{157, {{13507, 1, 0, 0},{13507, 1, 1, 17}, {0, 0, 0, 0}}},{157, {{13507, 1, 0, 0},{13507, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{158, {{13508, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{159, {{13509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{160, {{13511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{162, {{13512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{163, {{13513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{164, {{13514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{165, {{13515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{166, {{13516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{167, {{13518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{168, {{13519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{169, {{13520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{170, {{13521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{171, {{13523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{172, {{13524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{173, {{13525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{174, {{13526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{175, {{13527, 1, 0, 0},{13527, 1, 1, 16}, {0, 0, 0, 0}}},{175, {{13527, 1, 0, 0},{13527, 1, 1, 16}, {0, 0, 0, 0}}},{176, {{13527, 1, 0, 0},{13527, 1, 1, 17}, {0, 0, 0, 0}}},{176, {{13527, 1, 0, 0},{13527, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{177, {{13528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{178, {{13529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{179, {{13530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{180, {{13531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{181, {{13532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{182, {{13533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{183, {{13534, 1, 0, 0},{13534, 1, 1, 16}, {0, 0, 0, 0}}},{183, {{13534, 1, 0, 0},{13534, 1, 1, 16}, {0, 0, 0, 0}}},{184, {{13534, 1, 0, 0},{13534, 1, 1, 18}, {0, 0, 0, 0}}},{184, {{13534, 1, 0, 0},{13534, 1, 1, 18}, {0, 0, 0, 0}}},{185, {{13534, 1, 0, 0},{13534, 1, 1, 19}, {0, 0, 0, 0}}},{185, {{13534, 1, 0, 0},{13534, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{186, {{13535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{187, {{13536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{188, {{13537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{189, {{13538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{190, {{13539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{191, {{13540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{192, {{13543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{193, {{13544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{194, {{13545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{195, {{13546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{196, {{13548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{197, {{13549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{198, {{13550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{200, {{13551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{201, {{13553, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{202, {{13554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{204, {{13555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{205, {{13556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{206, {{13557, 1, 0, 0},{13557, 1, 1, 16}, {0, 0, 0, 0}}},{206, {{13557, 1, 0, 0},{13557, 1, 1, 16}, {0, 0, 0, 0}}},{207, {{13557, 1, 0, 0},{13557, 1, 1, 17}, {0, 0, 0, 0}}},{207, {{13557, 1, 0, 0},{13557, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{208, {{13558, 1, 0, 0},{13558, 1, 1, 16}, {0, 0, 0, 0}}},{208, {{13558, 1, 0, 0},{13558, 1, 1, 16}, {0, 0, 0, 0}}},{209, {{13558, 1, 0, 0},{13558, 1, 1, 17}, {0, 0, 0, 0}}},{209, {{13558, 1, 0, 0},{13558, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{210, {{13559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{211, {{13560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{212, {{13561, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{213, {{13562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{214, {{13563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{215, {{13564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{216, {{13565, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{217, {{13566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{218, {{13568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{219, {{13569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{220, {{13570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{221, {{13571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{222, {{13572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{223, {{13573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{224, {{13574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{225, {{13575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{226, {{13578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{227, {{13579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{228, {{13580, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{229, {{13581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{230, {{13582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{231, {{13583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{232, {{13585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{233, {{13586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{234, {{13587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67980, {{13588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{235, {{13589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67981, {{13590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{237, {{13591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{238, {{13592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{239, {{13593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{240, {{13594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{241, {{13596, 1, 0, 0},{13596, 1, 1, 17}, {0, 0, 0, 0}}},{241, {{13596, 1, 0, 0},{13596, 1, 1, 17}, {0, 0, 0, 0}}},{242, {{13596, 1, 0, 0},{13596, 1, 1, 16}, {0, 0, 0, 0}}},{242, {{13596, 1, 0, 0},{13596, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{243, {{13597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{244, {{13598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{245, {{13599, 1, 0, 0},{13599, 1, 1, 17}, {0, 0, 0, 0}}},{245, {{13599, 1, 0, 0},{13599, 1, 1, 17}, {0, 0, 0, 0}}},{59383, {{13599, 1, 0, 0},{13599, 1, 1, 18}, {0, 0, 0, 0}}},{59383, {{13599, 1, 0, 0},{13599, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{246, {{13600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{249, {{13601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{250, {{13603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{251, {{13604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{252, {{13605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{253, {{13606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{255, {{13607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{256, {{13608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{257, {{13609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{258, {{13610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{259, {{13611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{260, {{13612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{261, {{13613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{262, {{13614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{263, {{13615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{264, {{13616, 1, 0, 0},{13616, 1, 1, 17}, {0, 0, 0, 0}}},{264, {{13616, 1, 0, 0},{13616, 1, 1, 17}, {0, 0, 0, 0}}},{265, {{13616, 1, 0, 0},{13616, 1, 1, 16}, {0, 0, 0, 0}}},{265, {{13616, 1, 0, 0},{13616, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {4, {{266, {{13617, 1, 0, 0},{13617, 1, 1, 16}, {0, 0, 0, 0}}},{266, {{13617, 1, 0, 0},{13617, 1, 1, 16}, {0, 0, 0, 0}}},{267, {{13617, 1, 0, 0},{13617, 1, 1, 17}, {0, 0, 0, 0}}},{267, {{13617, 1, 0, 0},{13617, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{268, {{13619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{269, {{13620, 1, 0, 0},{13620, 1, 1, 16}, {0, 0, 0, 0}}},{269, {{13620, 1, 0, 0},{13620, 1, 1, 16}, {0, 0, 0, 0}}},{270, {{13620, 1, 0, 0},{13620, 1, 1, 17}, {0, 0, 0, 0}}},{270, {{13620, 1, 0, 0},{13620, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{271, {{13621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{272, {{13622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{273, {{13623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{274, {{13624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{275, {{13625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{276, {{13626, 1, 0, 0},{13626, 1, 1, 16}, {0, 0, 0, 0}}},{276, {{13626, 1, 0, 0},{13626, 1, 1, 16}, {0, 0, 0, 0}}},{277, {{13626, 1, 0, 0},{13626, 1, 1, 17}, {0, 0, 0, 0}}},{277, {{13626, 1, 0, 0},{13626, 1, 1, 17}, {0, 0, 0, 0}}},{278, {{13626, 1, 0, 0},{13626, 1, 1, 18}, {0, 0, 0, 0}}},{278, {{13626, 1, 0, 0},{13626, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{279, {{13627, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{280, {{13628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{281, {{13629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{282, {{13630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{284, {{13631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{285, {{13633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{286, {{13634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{287, {{13635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{288, {{13636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{289, {{13637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{290, {{13638, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{291, {{13639, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{292, {{13640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{293, {{13641, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{294, {{13642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{295, {{13643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{296, {{13644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{297, {{13645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{298, {{13646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{299, {{13648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{300, {{13649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{301, {{13650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{302, {{13651, 1, 0, 0},{13651, 1, 1, 17}, {0, 0, 0, 0}}},{302, {{13651, 1, 0, 0},{13651, 1, 1, 17}, {0, 0, 0, 0}}},{303, {{13651, 1, 0, 0},{13651, 1, 1, 16}, {0, 0, 0, 0}}},{303, {{13651, 1, 0, 0},{13651, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{304, {{13653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{305, {{13654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{306, {{13655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{307, {{13657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{308, {{13658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{309, {{13659, 1, 0, 0},{13659, 1, 1, 16}, {0, 0, 0, 0}}},{309, {{13659, 1, 0, 0},{13659, 1, 1, 16}, {0, 0, 0, 0}}},{310, {{13659, 1, 0, 0},{13659, 1, 1, 17}, {0, 0, 0, 0}}},{310, {{13659, 1, 0, 0},{13659, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{311, {{13660, 1, 0, 0},{13660, 1, 1, 16}, {0, 0, 0, 0}}},{311, {{13660, 1, 0, 0},{13660, 1, 1, 16}, {0, 0, 0, 0}}},{312, {{13660, 1, 0, 0},{13660, 1, 1, 17}, {0, 0, 0, 0}}},{312, {{13660, 1, 0, 0},{13660, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{313, {{13661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{314, {{13662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{315, {{13663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{316, {{13664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{317, {{13665, 1, 0, 0},{13665, 1, 1, 16}, {0, 0, 0, 0}}},{317, {{13665, 1, 0, 0},{13665, 1, 1, 16}, {0, 0, 0, 0}}},{318, {{13665, 1, 0, 0},{13665, 1, 1, 17}, {0, 0, 0, 0}}},{318, {{13665, 1, 0, 0},{13665, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{319, {{13666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{320, {{13667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{321, {{13668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{322, {{13669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{323, {{13670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{324, {{13671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{325, {{13672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{326, {{13673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{327, {{13674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{328, {{13675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{329, {{13676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{330, {{13677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{331, {{13678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{332, {{13679, 1, 0, 0},{13679, 1, 1, 16}, {0, 0, 0, 0}}},{332, {{13679, 1, 0, 0},{13679, 1, 1, 16}, {0, 0, 0, 0}}},{333, {{13679, 1, 0, 0},{13679, 1, 1, 17}, {0, 0, 0, 0}}},{333, {{13679, 1, 0, 0},{13679, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{334, {{13680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{335, {{13681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{336, {{13682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{337, {{13683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{338, {{13684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{339, {{13685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{340, {{13686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{341, {{13688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{342, {{13689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{343, {{13690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{344, {{13691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{345, {{13692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{346, {{13693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67982, {{13694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{348, {{13696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{349, {{13697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{350, {{13698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{351, {{13699, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{352, {{13700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{353, {{13701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{354, {{13702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{355, {{13703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{67983, {{13707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67984, {{13708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67985, {{13709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67986, {{13710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{356, {{13711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{357, {{13712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{358, {{13713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{359, {{13714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{360, {{13715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{361, {{13716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{362, {{13717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{363, {{13718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{364, {{13719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{365, {{13720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67987, {{13721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67988, {{13722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67989, {{13723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67990, {{13724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67991, {{13725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{366, {{13727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{367, {{13728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{368, {{13729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{369, {{13730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{370, {{13731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{371, {{13732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{372, {{13733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{373, {{13734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{374, {{13735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{375, {{13736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{376, {{13737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{377, {{13738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{378, {{13739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{67992, {{13743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67993, {{13744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{67994, {{13746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67995, {{13747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{380, {{13750, 1, 0, 0},{13750, 1, 1, 17}, {0, 0, 0, 0}}},{380, {{13750, 1, 0, 0},{13750, 1, 1, 17}, {0, 0, 0, 0}}},{381, {{13750, 1, 0, 0},{13750, 1, 1, 16}, {0, 0, 0, 0}}},{381, {{13750, 1, 0, 0},{13750, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{382, {{13751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{383, {{13752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{384, {{13753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{385, {{13754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{386, {{13755, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{387, {{13756, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{388, {{13758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{389, {{13759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{390, {{13760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{391, {{13761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{392, {{13762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{393, {{13763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{394, {{13764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{395, {{13765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{396, {{13766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{397, {{13767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{398, {{13769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{399, {{13770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{400, {{13771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{401, {{13772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{402, {{13773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{403, {{13776, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{404, {{13779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{405, {{13780, 1, 0, 0},{13780, 1, 1, 16}, {0, 0, 0, 0}}},{405, {{13780, 1, 0, 0},{13780, 1, 1, 16}, {0, 0, 0, 0}}},{406, {{13780, 1, 0, 0},{13780, 1, 1, 17}, {0, 0, 0, 0}}},{406, {{13780, 1, 0, 0},{13780, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{407, {{13781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{408, {{13782, 1, 0, 0},{13782, 1, 1, 16}, {0, 0, 0, 0}}},{408, {{13782, 1, 0, 0},{13782, 1, 1, 16}, {0, 0, 0, 0}}},{409, {{13782, 1, 0, 0},{13782, 1, 1, 17}, {0, 0, 0, 0}}},{409, {{13782, 1, 0, 0},{13782, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{410, {{13783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{411, {{13784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{412, {{13785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{413, {{13786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{414, {{13787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{415, {{13788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{416, {{13790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67996, {{13791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67997, {{13792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{67998, {{13793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{417, {{13794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{418, {{13795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{419, {{13796, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{420, {{13797, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{421, {{13798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{422, {{13799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{423, {{13800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{424, {{13801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{425, {{13802, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{426, {{13803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{427, {{13804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{428, {{13805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{67999, {{13807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{429, {{13809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{430, {{13810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{431, {{13811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{432, {{13812, 1, 0, 0},{13812, 1, 1, 18}, {0, 0, 0, 0}}},{432, {{13812, 1, 0, 0},{13812, 1, 1, 18}, {0, 0, 0, 0}}},{433, {{13812, 1, 0, 0},{13812, 1, 1, 17}, {0, 0, 0, 0}}},{433, {{13812, 1, 0, 0},{13812, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{434, {{13813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{435, {{13814, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{436, {{13819, 1, 0, 0},{13819, 1, 1, 17}, {0, 0, 0, 0}}},{436, {{13819, 1, 0, 0},{13819, 1, 1, 17}, {0, 0, 0, 0}}},{437, {{13819, 1, 0, 0},{13819, 1, 1, 16}, {0, 0, 0, 0}}},{437, {{13819, 1, 0, 0},{13819, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{438, {{13820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{439, {{13821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{440, {{13822, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{441, {{13823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{442, {{13824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{443, {{13825, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{444, {{13826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{445, {{13829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{446, {{13830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{447, {{13831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{448, {{13832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{449, {{13833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{450, {{13834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{451, {{13835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{452, {{13836, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{453, {{13837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{68000, {{13839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{454, {{13840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{455, {{13841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{68001, {{13842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{456, {{13843, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{457, {{13845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{458, {{13846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{459, {{13849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{460, {{13851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{461, {{13852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{462, {{13853, 1, 0, 0},{13853, 1, 1, 16}, {0, 0, 0, 0}}},{462, {{13853, 1, 0, 0},{13853, 1, 1, 16}, {0, 0, 0, 0}}},{463, {{13853, 1, 0, 0},{13853, 1, 1, 17}, {0, 0, 0, 0}}},{463, {{13853, 1, 0, 0},{13853, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{464, {{13854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{466, {{13856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{467, {{13857, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{468, {{13858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{469, {{13859, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{470, {{13860, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{471, {{13861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{472, {{13862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{473, {{13863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{474, {{13864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{476, {{13865, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{477, {{13866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{478, {{13867, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{479, {{13868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{480, {{13869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{481, {{13872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{482, {{13873, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{483, {{13874, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{484, {{13875, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{485, {{13876, 1, 0, 0},{13876, 1, 1, 16}, {0, 0, 0, 0}}},{485, {{13876, 1, 0, 0},{13876, 1, 1, 16}, {0, 0, 0, 0}}},{486, {{13876, 1, 0, 0},{13876, 1, 1, 17}, {0, 0, 0, 0}}},{486, {{13876, 1, 0, 0},{13876, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{487, {{13878, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{488, {{13879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{489, {{13880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{490, {{13881, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{491, {{13882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{492, {{13883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{493, {{13887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{494, {{13888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{495, {{13889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{496, {{13890, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{498, {{13891, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{499, {{13892, 1, 0, 0},{13892, 1, 1, 16}, {0, 0, 0, 0}}},{499, {{13892, 1, 0, 0},{13892, 1, 1, 16}, {0, 0, 0, 0}}},{500, {{13892, 1, 0, 0},{13892, 1, 1, 17}, {0, 0, 0, 0}}},{500, {{13892, 1, 0, 0},{13892, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{501, {{13893, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{502, {{13895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{503, {{13896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{504, {{13897, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{505, {{13898, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{506, {{13902, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{507, {{13903, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{508, {{13904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{509, {{13905, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{511, {{13906, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{513, {{13907, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{514, {{13908, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{515, {{13909, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{517, {{13913, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{518, {{13914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{519, {{13915, 1, 0, 0},{13915, 1, 1, 16}, {0, 0, 0, 0}}},{519, {{13915, 1, 0, 0},{13915, 1, 1, 16}, {0, 0, 0, 0}}},{520, {{13915, 1, 0, 0},{13915, 1, 1, 17}, {0, 0, 0, 0}}},{520, {{13915, 1, 0, 0},{13915, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{522, {{13917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{523, {{13918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{524, {{13920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{525, {{13921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{526, {{13922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{527, {{13923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{528, {{13924, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{529, {{13925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{530, {{13927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{531, {{13928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{532, {{13929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{533, {{13930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{534, {{13931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{535, {{13932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{536, {{13935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{537, {{13936, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{538, {{13937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{539, {{13938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{540, {{13939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{541, {{13940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{542, {{13941, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{543, {{13942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{544, {{13943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{545, {{13944, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{546, {{13945, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{547, {{13946, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{548, {{13947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{549, {{13948, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{550, {{13950, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{551, {{13952, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{552, {{13953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{553, {{13954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{554, {{13955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{555, {{13957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{556, {{13958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{557, {{13959, 1, 0, 0},{13959, 1, 1, 17}, {0, 0, 0, 0}}},{557, {{13959, 1, 0, 0},{13959, 1, 1, 17}, {0, 0, 0, 0}}},{57255, {{13959, 1, 0, 0},{13959, 1, 1, 16}, {0, 0, 0, 0}}},{57255, {{13959, 1, 0, 0},{13959, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {4, {{558, {{13960, 1, 0, 0},{13960, 1, 1, 16}, {0, 0, 0, 0}}},{558, {{13960, 1, 0, 0},{13960, 1, 1, 16}, {0, 0, 0, 0}}},{57261, {{13960, 1, 0, 0},{13960, 1, 1, 17}, {0, 0, 0, 0}}},{57261, {{13960, 1, 0, 0},{13960, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{559, {{13961, 1, 0, 0},{13961, 1, 1, 17}, {0, 0, 0, 0}}},{559, {{13961, 1, 0, 0},{13961, 1, 1, 17}, {0, 0, 0, 0}}},{560, {{13961, 1, 0, 0},{13961, 1, 1, 16}, {0, 0, 0, 0}}},{560, {{13961, 1, 0, 0},{13961, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{562, {{13962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{563, {{13963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{564, {{13964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{565, {{13965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{566, {{13968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{567, {{13969, 1, 0, 0},{13969, 1, 1, 18}, {0, 0, 0, 0}}},{567, {{13969, 1, 0, 0},{13969, 1, 1, 18}, {0, 0, 0, 0}}},{568, {{13969, 1, 0, 0},{13969, 1, 1, 17}, {0, 0, 0, 0}}},{568, {{13969, 1, 0, 0},{13969, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{569, {{13970, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{570, {{13971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{571, {{13972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{572, {{13973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{573, {{13974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{574, {{13975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{575, {{13976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{576, {{13977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{577, {{13978, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{578, {{13980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{579, {{13981, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{580, {{13982, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{581, {{13984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{582, {{13985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{583, {{13986, 1, 0, 0},{13986, 1, 1, 16}, {0, 0, 0, 0}}},{583, {{13986, 1, 0, 0},{13986, 1, 1, 16}, {0, 0, 0, 0}}},{584, {{13986, 1, 0, 0},{13986, 1, 1, 17}, {0, 0, 0, 0}}},{584, {{13986, 1, 0, 0},{13986, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{585, {{13987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{586, {{13988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{587, {{13989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{588, {{13990, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{589, {{13991, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{590, {{13993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{591, {{13994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{592, {{13995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{594, {{13996, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{595, {{13997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{596, {{13998, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{597, {{14000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{598, {{14001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{599, {{14002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{600, {{14003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{601, {{14004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{602, {{14005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{603, {{14006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{604, {{14007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{605, {{14008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{606, {{14009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{607, {{14010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{608, {{14011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{609, {{14012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{610, {{14013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{611, {{14014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{612, {{14015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{613, {{14017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{614, {{14018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{615, {{14019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{618, {{14020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{619, {{14021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{620, {{14022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{621, {{14029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{622, {{14030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{623, {{14031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{624, {{14032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{625, {{14033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{626, {{14034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{627, {{14035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{628, {{14036, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{629, {{14037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{630, {{14038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{631, {{14039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{632, {{14040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{633, {{14041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{634, {{14042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{635, {{14043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{636, {{14044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{637, {{14045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{638, {{14046, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{639, {{14053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{640, {{14054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{641, {{14055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{642, {{14056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{643, {{14057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{644, {{14058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{645, {{14059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{646, {{14060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{647, {{14061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{648, {{14062, 1, 0, 0},{14062, 1, 1, 16}, {0, 0, 0, 0}}},{648, {{14062, 1, 0, 0},{14062, 1, 1, 16}, {0, 0, 0, 0}}},{649, {{14062, 1, 0, 0},{14062, 1, 1, 17}, {0, 0, 0, 0}}},{649, {{14062, 1, 0, 0},{14062, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{650, {{14063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{651, {{14064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{652, {{14065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{653, {{14067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{654, {{14068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{655, {{14069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{656, {{14070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{657, {{14071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{658, {{14072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{659, {{14073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{660, {{14074, 1, 0, 0},{14074, 1, 1, 16}, {0, 0, 0, 0}}},{660, {{14074, 1, 0, 0},{14074, 1, 1, 16}, {0, 0, 0, 0}}},{661, {{14074, 1, 0, 0},{14074, 1, 1, 17}, {0, 0, 0, 0}}},{661, {{14074, 1, 0, 0},{14074, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{662, {{14075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{663, {{14076, 1, 0, 0},{14076, 1, 1, 16}, {0, 0, 0, 0}}},{663, {{14076, 1, 0, 0},{14076, 1, 1, 16}, {0, 0, 0, 0}}},{664, {{14076, 1, 0, 0},{14076, 1, 1, 17}, {0, 0, 0, 0}}},{664, {{14076, 1, 0, 0},{14076, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{665, {{14077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{666, {{14078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{667, {{14082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{668, {{14083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{669, {{14084, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{670, {{14085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{671, {{14086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{672, {{14087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{673, {{14088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{674, {{14089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{675, {{14090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{676, {{14091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{677, {{14092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{678, {{14093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{679, {{14094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{680, {{14095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{681, {{14096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{682, {{14097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{683, {{14098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{684, {{14099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{685, {{14102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{686, {{14103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{687, {{14104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{688, {{14105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{689, {{14106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{690, {{14107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{691, {{14108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{692, {{14109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{693, {{14110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{694, {{14112, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{695, {{14113, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{696, {{14114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{697, {{14115, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{698, {{14116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{699, {{14117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{700, {{14118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{701, {{14119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{702, {{14121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{703, {{14122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{704, {{14123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{705, {{14126, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{706, {{14127, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{709, {{14128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{710, {{14129, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{711, {{14130, 1, 0, 0},{14130, 1, 1, 16}, {0, 0, 0, 0}}},{711, {{14130, 1, 0, 0},{14130, 1, 1, 16}, {0, 0, 0, 0}}},{712, {{14130, 1, 0, 0},{14130, 1, 1, 17}, {0, 0, 0, 0}}},{712, {{14130, 1, 0, 0},{14130, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{713, {{14131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{714, {{14132, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{715, {{14133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{716, {{14134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{717, {{14135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{718, {{14136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{719, {{14137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{720, {{14138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{721, {{14139, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{722, {{14140, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{723, {{14141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{724, {{14142, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{725, {{14144, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{726, {{14145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{727, {{14146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{728, {{14147, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{729, {{14148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{730, {{14149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{731, {{14151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{732, {{14152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{733, {{14154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{734, {{14155, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{735, {{14156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{736, {{14157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{737, {{14158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{738, {{14159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{739, {{14160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{740, {{14161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{741, {{14162, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{742, {{14163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{743, {{14164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{744, {{14165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{745, {{14166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{746, {{14167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{747, {{14168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{748, {{14169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{749, {{14170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{750, {{14171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{751, {{14172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{752, {{14173, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{753, {{14174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{754, {{14175, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{755, {{14176, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{756, {{14177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{757, {{14178, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{758, {{14179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{760, {{14180, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{761, {{14181, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{762, {{14182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{763, {{14183, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{764, {{14185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{765, {{14186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{766, {{14187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{768, {{14188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{769, {{14189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{770, {{14190, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{771, {{14191, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{772, {{14192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{773, {{14193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{774, {{14194, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{775, {{14196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{776, {{14197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{777, {{14198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{778, {{14200, 1, 0, 0},{14200, 1, 1, 17}, {0, 0, 0, 0}}},{778, {{14200, 1, 0, 0},{14200, 1, 1, 17}, {0, 0, 0, 0}}},{34205, {{14200, 1, 0, 0},{14200, 1, 1, 16}, {0, 0, 0, 0}}},{34205, {{14200, 1, 0, 0},{14200, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{779, {{14202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{780, {{14203, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{781, {{14204, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{783, {{14206, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{784, {{14207, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{785, {{14208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{786, {{14209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{787, {{14210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{788, {{14211, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{789, {{14212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{790, {{14213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{791, {{14214, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{792, {{14215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{793, {{14216, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{794, {{14217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{795, {{14218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{796, {{14219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{797, {{14220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{798, {{14221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{799, {{14223, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{800, {{14224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{801, {{14225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{802, {{14226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{803, {{14227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{804, {{14229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{805, {{14230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{806, {{14231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{807, {{14232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{808, {{14233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{809, {{14234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{810, {{14235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{811, {{14236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{812, {{14239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{813, {{14241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{814, {{14242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{815, {{14243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{816, {{14244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{818, {{14245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{819, {{14246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{820, {{14247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{821, {{14248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{822, {{14249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{823, {{14250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{824, {{14251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{825, {{14252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{826, {{14253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{827, {{14254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{828, {{14255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{829, {{14256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{830, {{14257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{831, {{14258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{832, {{14259, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{833, {{14260, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{834, {{14262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{835, {{14263, 1, 0, 0},{14263, 1, 1, 17}, {0, 0, 0, 0}}},{835, {{14263, 1, 0, 0},{14263, 1, 1, 17}, {0, 0, 0, 0}}},{836, {{14263, 1, 0, 0},{14263, 1, 1, 16}, {0, 0, 0, 0}}},{836, {{14263, 1, 0, 0},{14263, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{837, {{14265, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{838, {{14266, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{839, {{14267, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{840, {{14268, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{841, {{14269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{842, {{14270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{843, {{14271, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{844, {{14272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{845, {{14273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{847, {{14275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{848, {{14276, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{849, {{14277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{850, {{14279, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{851, {{14280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{852, {{14281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{853, {{14282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{854, {{14283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{855, {{14284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{856, {{14285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{857, {{14286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{858, {{14287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{859, {{14288, 1, 0, 0},{14288, 1, 1, 16}, {0, 0, 0, 0}}},{859, {{14288, 1, 0, 0},{14288, 1, 1, 16}, {0, 0, 0, 0}}},{860, {{14288, 1, 0, 0},{14288, 1, 1, 17}, {0, 0, 0, 0}}},{860, {{14288, 1, 0, 0},{14288, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{861, {{14289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{862, {{14290, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{863, {{14291, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{864, {{14292, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{865, {{14295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{866, {{14296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{867, {{14297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{868, {{14298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{869, {{14299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{870, {{14300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{871, {{14301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{872, {{14302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{873, {{14303, 1, 0, 0},{14303, 1, 1, 16}, {0, 0, 0, 0}}},{873, {{14303, 1, 0, 0},{14303, 1, 1, 16}, {0, 0, 0, 0}}},{874, {{14303, 1, 0, 0},{14303, 1, 1, 17}, {0, 0, 0, 0}}},{874, {{14303, 1, 0, 0},{14303, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{875, {{14304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{876, {{14305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{877, {{14306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{878, {{14310, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{879, {{14311, 1, 0, 0},{14311, 1, 1, 16}, {0, 0, 0, 0}}},{879, {{14311, 1, 0, 0},{14311, 1, 1, 16}, {0, 0, 0, 0}}},{880, {{14311, 1, 0, 0},{14311, 1, 1, 17}, {0, 0, 0, 0}}},{880, {{14311, 1, 0, 0},{14311, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{881, {{14312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{882, {{14313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{883, {{14314, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{884, {{14315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{885, {{14316, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{886, {{14317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{887, {{14320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{888, {{14321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{890, {{14322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{891, {{14323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{892, {{14324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{893, {{14325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{894, {{14326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{895, {{14328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{896, {{14329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{897, {{14330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{898, {{14331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{899, {{14332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{900, {{14333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{901, {{14334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{902, {{14335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{903, {{14336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{904, {{14337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{905, {{14338, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{906, {{14341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{907, {{14342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{908, {{14343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{909, {{14344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{910, {{14345, 1, 0, 0},{14345, 1, 1, 16}, {0, 0, 0, 0}}},{910, {{14345, 1, 0, 0},{14345, 1, 1, 16}, {0, 0, 0, 0}}},{911, {{14345, 1, 0, 0},{14345, 1, 1, 17}, {0, 0, 0, 0}}},{911, {{14345, 1, 0, 0},{14345, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{912, {{14346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{913, {{14347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{914, {{14348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{915, {{14349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{916, {{14350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{917, {{14351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{918, {{14352, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{919, {{14353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{920, {{14354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{921, {{14355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{922, {{14356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{923, {{14357, 1, 0, 0},{14357, 1, 1, 16}, {0, 0, 0, 0}}},{923, {{14357, 1, 0, 0},{14357, 1, 1, 16}, {0, 0, 0, 0}}},{924, {{14357, 1, 0, 0},{14357, 1, 1, 17}, {0, 0, 0, 0}}},{924, {{14357, 1, 0, 0},{14357, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{925, {{14358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{926, {{14359, 1, 0, 0},{14359, 1, 1, 16}, {0, 0, 0, 0}}},{926, {{14359, 1, 0, 0},{14359, 1, 1, 16}, {0, 0, 0, 0}}},{927, {{14359, 1, 0, 0},{14359, 1, 1, 17}, {0, 0, 0, 0}}},{927, {{14359, 1, 0, 0},{14359, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{928, {{14360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{929, {{14361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{930, {{14362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{931, {{14363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{932, {{14365, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{933, {{14368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{934, {{14369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{935, {{14370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{936, {{14373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{937, {{14374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{938, {{14375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{939, {{14376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{940, {{14377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{941, {{14378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{942, {{14381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{943, {{14382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{944, {{14383, 1, 0, 0},{14383, 1, 1, 18}, {0, 0, 0, 0}}},{944, {{14383, 1, 0, 0},{14383, 1, 1, 18}, {0, 0, 0, 0}}},{945, {{14383, 1, 0, 0},{14383, 1, 1, 19}, {0, 0, 0, 0}}},{945, {{14383, 1, 0, 0},{14383, 1, 1, 19}, {0, 0, 0, 0}}},{946, {{14383, 1, 0, 0},{14383, 1, 1, 17}, {0, 0, 0, 0}}},{946, {{14383, 1, 0, 0},{14383, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{68002, {{14384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{947, {{14385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{948, {{14386, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{949, {{14388, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{950, {{14389, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{951, {{14390, 1, 0, 0},{14390, 1, 1, 17}, {0, 0, 0, 0}}},{951, {{14390, 1, 0, 0},{14390, 1, 1, 17}, {0, 0, 0, 0}}},{59541, {{14390, 1, 0, 0},{14390, 1, 1, 18}, {0, 0, 0, 0}}},{59541, {{14390, 1, 0, 0},{14390, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{952, {{14391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{953, {{14392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{954, {{14393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{955, {{14394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{956, {{14395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{957, {{14396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{958, {{14397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{959, {{14398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{960, {{14399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{961, {{14400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{962, {{14401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{963, {{14402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{965, {{14403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{966, {{14404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{967, {{14405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{964, {{14406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{968, {{14407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{969, {{14408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{970, {{14409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{971, {{14410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{972, {{14411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{973, {{14412, 1, 0, 0},{14412, 1, 1, 16}, {0, 0, 0, 0}}},{973, {{14412, 1, 0, 0},{14412, 1, 1, 16}, {0, 0, 0, 0}}},{974, {{14412, 1, 0, 0},{14412, 1, 1, 17}, {0, 0, 0, 0}}},{974, {{14412, 1, 0, 0},{14412, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{975, {{14415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{976, {{14416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{977, {{14417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{978, {{14418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{979, {{14419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{980, {{14420, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{981, {{14421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{983, {{14422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{984, {{14423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{985, {{14424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{986, {{14425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{987, {{14426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{988, {{14427, 1, 0, 0},{14427, 1, 1, 16}, {0, 0, 0, 0}}},{988, {{14427, 1, 0, 0},{14427, 1, 1, 16}, {0, 0, 0, 0}}},{989, {{14427, 1, 0, 0},{14427, 1, 1, 17}, {0, 0, 0, 0}}},{989, {{14427, 1, 0, 0},{14427, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{990, {{14428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{991, {{14429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{992, {{14430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{993, {{14431, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{994, {{14432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{995, {{14433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{996, {{14434, 1, 0, 0},{14434, 1, 1, 16}, {0, 0, 0, 0}}},{996, {{14434, 1, 0, 0},{14434, 1, 1, 16}, {0, 0, 0, 0}}},{997, {{14434, 1, 0, 0},{14434, 1, 1, 17}, {0, 0, 0, 0}}},{997, {{14434, 1, 0, 0},{14434, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{998, {{14435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{999, {{14437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1000, {{14438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1001, {{14439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1002, {{14440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1003, {{14441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1004, {{14442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1005, {{14443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1006, {{14444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{1007, {{14445, 1, 0, 0},{14445, 1, 1, 16}, {0, 0, 0, 0}}},{1007, {{14445, 1, 0, 0},{14445, 1, 1, 16}, {0, 0, 0, 0}}},{1008, {{14445, 1, 0, 0},{14445, 1, 1, 17}, {0, 0, 0, 0}}},{1008, {{14445, 1, 0, 0},{14445, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{1009, {{14446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1010, {{14447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1011, {{14448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1013, {{14449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1014, {{14450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1015, {{14451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1016, {{14452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1017, {{14453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1018, {{14454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1019, {{14455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1020, {{14456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1021, {{14457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1022, {{14458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1023, {{14459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1024, {{14460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1025, {{14461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1026, {{14462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1027, {{14463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1028, {{14464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1029, {{14465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1030, {{14466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1031, {{14467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1032, {{14468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1033, {{14469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1034, {{14470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1035, {{14472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1036, {{14473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1037, {{14474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1038, {{14475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1039, {{14477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1040, {{14478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1041, {{14479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1042, {{14481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1043, {{14482, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1044, {{14483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1045, {{14484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1046, {{14485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1047, {{14486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1048, {{14487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1050, {{14489, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1051, {{14490, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1052, {{14491, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1053, {{14492, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1054, {{14493, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1055, {{14494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1056, {{14495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1057, {{14497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{1058, {{14498, 1, 0, 0},{14498, 1, 1, 16}, {0, 0, 0, 0}}},{1058, {{14498, 1, 0, 0},{14498, 1, 1, 16}, {0, 0, 0, 0}}},{1059, {{14498, 1, 0, 0},{14498, 1, 1, 17}, {0, 0, 0, 0}}},{1059, {{14498, 1, 0, 0},{14498, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{1060, {{14499, 1, 0, 0},{14499, 1, 1, 16}, {0, 0, 0, 0}}},{1060, {{14499, 1, 0, 0},{14499, 1, 1, 16}, {0, 0, 0, 0}}},{1061, {{14499, 1, 0, 0},{14499, 1, 1, 17}, {0, 0, 0, 0}}},{1061, {{14499, 1, 0, 0},{14499, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1062, {{14501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1063, {{14502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1064, {{14503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1065, {{14504, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1066, {{14505, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1067, {{14506, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1068, {{14508, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1069, {{14510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1070, {{14511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1071, {{14512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1073, {{14514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1074, {{14515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{1075, {{14516, 1, 0, 0},{14516, 1, 1, 16}, {0, 0, 0, 0}}},{1075, {{14516, 1, 0, 0},{14516, 1, 1, 16}, {0, 0, 0, 0}}},{1076, {{14516, 1, 0, 0},{14516, 1, 1, 17}, {0, 0, 0, 0}}},{1076, {{14516, 1, 0, 0},{14516, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{1077, {{14517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1078, {{14518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1079, {{14520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1080, {{14522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1081, {{14523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1082, {{14524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1083, {{14525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1084, {{14526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1085, {{14528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1086, {{14529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1087, {{14530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1088, {{14531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1089, {{14532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1090, {{14533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1091, {{14534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{1092, {{14535, 1, 0, 0},{14535, 1, 1, 16}, {0, 0, 0, 0}}},{1092, {{14535, 1, 0, 0},{14535, 1, 1, 16}, {0, 0, 0, 0}}},{1093, {{14535, 1, 0, 0},{14535, 1, 1, 17}, {0, 0, 0, 0}}},{1093, {{14535, 1, 0, 0},{14535, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{1094, {{14536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1095, {{14537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1096, {{14539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1097, {{14540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1098, {{14541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1099, {{14542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1100, {{14543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1101, {{14544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1102, {{14545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1103, {{14546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1104, {{14547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1105, {{14548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1106, {{14552, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1107, {{14553, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1108, {{14554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1109, {{14555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1110, {{14556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1111, {{14557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1112, {{14558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1113, {{14559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1114, {{14560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1115, {{14562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1116, {{14563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1117, {{14564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1118, {{14565, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1119, {{14566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1120, {{14570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1121, {{14571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1122, {{14572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1123, {{14573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1124, {{14575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1125, {{14576, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1126, {{14577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1127, {{14578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1128, {{14581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1129, {{14582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1130, {{14583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{1132, {{14586, 1, 0, 0},{14586, 1, 1, 17}, {0, 0, 0, 0}}},{1132, {{14586, 1, 0, 0},{14586, 1, 1, 17}, {0, 0, 0, 0}}},{1133, {{14586, 1, 0, 0},{14586, 1, 1, 18}, {0, 0, 0, 0}}},{1133, {{14586, 1, 0, 0},{14586, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{1134, {{14587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{1135, {{14588, 1, 0, 0},{14588, 1, 1, 16}, {0, 0, 0, 0}}},{1135, {{14588, 1, 0, 0},{14588, 1, 1, 16}, {0, 0, 0, 0}}},{1136, {{14588, 1, 0, 0},{14588, 1, 1, 17}, {0, 0, 0, 0}}},{1136, {{14588, 1, 0, 0},{14588, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{1137, {{14589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1138, {{14590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1139, {{14591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{57490, {{14592, 1, 0, 0},{14592, 1, 1, 16}, {0, 0, 0, 0}}},{57490, {{14592, 1, 0, 0},{14592, 1, 1, 16}, {0, 0, 0, 0}}},{57493, {{14592, 1, 0, 0},{14592, 1, 1, 17}, {0, 0, 0, 0}}},{57493, {{14592, 1, 0, 0},{14592, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{1140, {{14593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1141, {{14594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1142, {{14595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1143, {{14596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{1144, {{14597, 1, 0, 0},{14597, 1, 1, 16}, {0, 0, 0, 0}}},{1144, {{14597, 1, 0, 0},{14597, 1, 1, 16}, {0, 0, 0, 0}}},{1145, {{14597, 1, 0, 0},{14597, 1, 1, 17}, {0, 0, 0, 0}}},{1145, {{14597, 1, 0, 0},{14597, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{1146, {{14598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1148, {{14599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1149, {{14600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1150, {{14601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1151, {{14602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1152, {{14603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1153, {{14604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1154, {{14605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1155, {{14606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1156, {{14607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1157, {{14608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1158, {{14609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1159, {{14610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1160, {{14611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1161, {{14612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1162, {{14613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1163, {{14614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{1164, {{14615, 1, 0, 0},{14615, 1, 1, 17}, {0, 0, 0, 0}}},{1164, {{14615, 1, 0, 0},{14615, 1, 1, 17}, {0, 0, 0, 0}}},{1165, {{14615, 1, 0, 0},{14615, 1, 1, 18}, {0, 0, 0, 0}}},{1165, {{14615, 1, 0, 0},{14615, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1166, {{14617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1167, {{14618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1168, {{14619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1169, {{14620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1170, {{14621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1171, {{14622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1172, {{14623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1173, {{14624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1174, {{14625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1175, {{14626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1176, {{14627, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1177, {{14628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1178, {{14629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1179, {{14630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1180, {{14631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1181, {{14632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1182, {{14633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1183, {{14634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1184, {{14635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1185, {{14636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1186, {{14638, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1187, {{14639, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1188, {{14640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1189, {{14641, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1190, {{14642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1191, {{14643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1192, {{14644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1193, {{14645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1194, {{14646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1195, {{14647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1196, {{14648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1197, {{14649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{1198, {{14650, 1, 0, 0},{14650, 1, 1, 17}, {0, 0, 0, 0}}},{1198, {{14650, 1, 0, 0},{14650, 1, 1, 17}, {0, 0, 0, 0}}},{1199, {{14650, 1, 0, 0},{14650, 1, 1, 16}, {0, 0, 0, 0}}},{1199, {{14650, 1, 0, 0},{14650, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{1200, {{14651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1202, {{14654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1203, {{14655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1204, {{14656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1205, {{14657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1206, {{14658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1207, {{14659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1208, {{14660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1209, {{14661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1210, {{14662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1211, {{14663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1212, {{14664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1213, {{14665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1214, {{14666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1215, {{14667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1216, {{14668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1217, {{14669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1218, {{14670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1219, {{14671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1220, {{14672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1221, {{14673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1222, {{14674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1223, {{14675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1224, {{14676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1225, {{14677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1226, {{14678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1228, {{14682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1229, {{14683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1230, {{14684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1231, {{14685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1232, {{14686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1233, {{14687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1234, {{14688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1235, {{14689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1236, {{14690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1237, {{14691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1238, {{14692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1239, {{14693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1240, {{14694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1241, {{14695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1242, {{14696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1243, {{14697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1244, {{14700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1245, {{14701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1246, {{14702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{1247, {{14703, 1, 0, 0},{14703, 1, 1, 17}, {0, 0, 0, 0}}},{1247, {{14703, 1, 0, 0},{14703, 1, 1, 17}, {0, 0, 0, 0}}},{1248, {{14703, 1, 0, 0},{14703, 1, 1, 18}, {0, 0, 0, 0}}},{1248, {{14703, 1, 0, 0},{14703, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{1249, {{14704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{1250, {{14705, 1, 0, 0},{14705, 1, 1, 16}, {0, 0, 0, 0}}},{1250, {{14705, 1, 0, 0},{14705, 1, 1, 16}, {0, 0, 0, 0}}},{1251, {{14705, 1, 0, 0},{14705, 1, 1, 17}, {0, 0, 0, 0}}},{1251, {{14705, 1, 0, 0},{14705, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{1252, {{14706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1253, {{14707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1254, {{14708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1255, {{14709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1256, {{14710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1257, {{14711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1258, {{14712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1259, {{14713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1260, {{14714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1261, {{14716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1262, {{14717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1263, {{14718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1264, {{14719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1265, {{14720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1266, {{14721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1267, {{14722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1268, {{14723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1270, {{14724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1271, {{14725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1272, {{14726, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1273, {{14727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1274, {{14729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1275, {{14730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1276, {{14731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1277, {{14732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1278, {{14733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1279, {{14734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1280, {{14735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1281, {{14736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1282, {{14737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1283, {{14738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1284, {{14739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1285, {{14740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1286, {{14741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1287, {{14742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1288, {{14743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1289, {{14744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1290, {{14746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1291, {{14747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1292, {{14748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1293, {{14749, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1294, {{14750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1295, {{14751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1296, {{14753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{1297, {{14754, 1, 0, 0},{14754, 1, 1, 16}, {0, 0, 0, 0}}},{1297, {{14754, 1, 0, 0},{14754, 1, 1, 16}, {0, 0, 0, 0}}},{1298, {{14754, 1, 0, 0},{14754, 1, 1, 17}, {0, 0, 0, 0}}},{1298, {{14754, 1, 0, 0},{14754, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{1301, {{14755, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1302, {{14756, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1303, {{14757, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1305, {{14758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1306, {{14759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1307, {{14760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1308, {{14761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1309, {{14762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1310, {{14763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1311, {{14764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1312, {{14765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{1313, {{14766, 1, 0, 0},{14766, 1, 1, 16}, {0, 0, 0, 0}}},{1313, {{14766, 1, 0, 0},{14766, 1, 1, 16}, {0, 0, 0, 0}}},{1314, {{14766, 1, 0, 0},{14766, 1, 1, 18}, {0, 0, 0, 0}}},{1314, {{14766, 1, 0, 0},{14766, 1, 1, 18}, {0, 0, 0, 0}}},{57537, {{14766, 1, 0, 0},{14766, 1, 1, 17}, {0, 0, 0, 0}}},{57537, {{14766, 1, 0, 0},{14766, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{1315, {{14767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1316, {{14768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1317, {{14769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1318, {{14770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1320, {{14772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1321, {{14773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{1322, {{14774, 1, 0, 0},{14774, 1, 1, 17}, {0, 0, 0, 0}}},{1322, {{14774, 1, 0, 0},{14774, 1, 1, 17}, {0, 0, 0, 0}}},{1323, {{14774, 1, 0, 0},{14774, 1, 1, 16}, {0, 0, 0, 0}}},{1323, {{14774, 1, 0, 0},{14774, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{1324, {{14775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{1325, {{14776, 1, 0, 0},{14776, 1, 1, 17}, {0, 0, 0, 0}}},{1325, {{14776, 1, 0, 0},{14776, 1, 1, 17}, {0, 0, 0, 0}}},{57539, {{14776, 1, 0, 0},{14776, 1, 1, 18}, {0, 0, 0, 0}}},{57539, {{14776, 1, 0, 0},{14776, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{1326, {{14777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1327, {{14778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1328, {{14779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1329, {{14780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1330, {{14781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1331, {{14782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1332, {{14783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1333, {{14784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1334, {{14785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1335, {{14786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1336, {{14787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1337, {{14788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1338, {{14789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1339, {{14790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1340, {{14791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1341, {{14792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1342, {{14793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1343, {{14794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1344, {{14795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1345, {{14796, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1346, {{14797, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1347, {{14798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1348, {{14802, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1349, {{14803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1350, {{14804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1351, {{14805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1352, {{14806, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1353, {{14807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1354, {{14808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1355, {{14809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{1356, {{14814, 1, 0, 0},{14814, 1, 1, 16}, {0, 0, 0, 0}}},{1356, {{14814, 1, 0, 0},{14814, 1, 1, 16}, {0, 0, 0, 0}}},{1357, {{14814, 1, 0, 0},{14814, 1, 1, 17}, {0, 0, 0, 0}}},{1357, {{14814, 1, 0, 0},{14814, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1359, {{14817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1360, {{14818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1361, {{14819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1362, {{14820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1363, {{14821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1364, {{14822, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1365, {{14823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1366, {{14824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1367, {{14825, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1368, {{14826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1369, {{14827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1370, {{14828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1371, {{14829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1372, {{14830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1373, {{14831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1374, {{14834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1375, {{14835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1376, {{14836, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1377, {{14838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1378, {{14839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1379, {{14840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1380, {{14841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1381, {{14842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1382, {{14843, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1383, {{14844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1384, {{14845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1385, {{14846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1386, {{14847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1387, {{14848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1388, {{14849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1389, {{14850, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1390, {{14851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1391, {{14852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1392, {{14853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1393, {{14854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1394, {{14855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1395, {{14857, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1396, {{14858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1397, {{14859, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1398, {{14861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1399, {{14862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1400, {{14863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1401, {{14864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1402, {{14866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1403, {{14867, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1404, {{14868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1405, {{14869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1406, {{14870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{1407, {{14871, 1, 0, 0},{14871, 1, 1, 16}, {0, 0, 0, 0}}},{1407, {{14871, 1, 0, 0},{14871, 1, 1, 16}, {0, 0, 0, 0}}},{1408, {{14871, 1, 0, 0},{14871, 1, 1, 17}, {0, 0, 0, 0}}},{1408, {{14871, 1, 0, 0},{14871, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1409, {{14873, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1410, {{14874, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1411, {{14875, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1412, {{14876, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1413, {{14877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1414, {{14878, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1415, {{14879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1416, {{14880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1417, {{14881, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1418, {{14882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1419, {{14883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1420, {{14884, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1421, {{14885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1422, {{14886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1423, {{14887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1424, {{14888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1426, {{14890, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1427, {{14893, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1428, {{14894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{1429, {{14895, 1, 0, 0},{14895, 1, 1, 16}, {0, 0, 0, 0}}},{1429, {{14895, 1, 0, 0},{14895, 1, 1, 16}, {0, 0, 0, 0}}},{1430, {{14895, 1, 0, 0},{14895, 1, 1, 17}, {0, 0, 0, 0}}},{1430, {{14895, 1, 0, 0},{14895, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{1431, {{14896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1432, {{14897, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1433, {{14898, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1434, {{14899, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1435, {{14900, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1436, {{14901, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1437, {{14902, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1438, {{14903, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1439, {{14904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1440, {{14905, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1441, {{14906, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1442, {{14907, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1443, {{14908, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1444, {{14909, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{1445, {{14910, 1, 0, 0},{14910, 1, 1, 16}, {0, 0, 0, 0}}},{1445, {{14910, 1, 0, 0},{14910, 1, 1, 16}, {0, 0, 0, 0}}},{57568, {{14910, 1, 0, 0},{14910, 1, 1, 17}, {0, 0, 0, 0}}},{57568, {{14910, 1, 0, 0},{14910, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{1446, {{14911, 1, 0, 0},{14911, 1, 1, 16}, {0, 0, 0, 0}}},{1446, {{14911, 1, 0, 0},{14911, 1, 1, 16}, {0, 0, 0, 0}}},{1447, {{14911, 1, 0, 0},{14911, 1, 1, 17}, {0, 0, 0, 0}}},{1447, {{14911, 1, 0, 0},{14911, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{1448, {{14912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1449, {{14915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1450, {{14916, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1451, {{14917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1452, {{14918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1453, {{14919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1454, {{14920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1455, {{14921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1456, {{14922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1457, {{14923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1458, {{14924, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1459, {{14925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1460, {{14926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1461, {{14930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1462, {{14931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1463, {{14932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1464, {{14933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1465, {{14934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1466, {{14935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1467, {{14937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1468, {{14938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1469, {{14939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1470, {{14940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1471, {{14941, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1472, {{14942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1473, {{14944, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1474, {{14945, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1475, {{14946, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1476, {{14947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1477, {{14948, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1478, {{14949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1479, {{14950, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1480, {{14951, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1481, {{14952, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1482, {{14953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1483, {{14954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1484, {{14956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1485, {{14957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{1486, {{14958, 1, 0, 0},{14958, 1, 1, 18}, {0, 0, 0, 0}}},{1486, {{14958, 1, 0, 0},{14958, 1, 1, 18}, {0, 0, 0, 0}}},{1487, {{14958, 1, 0, 0},{14958, 1, 1, 17}, {0, 0, 0, 0}}},{1487, {{14958, 1, 0, 0},{14958, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{1488, {{14959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1489, {{14960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1490, {{14961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1491, {{14962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1492, {{14963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1493, {{14964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1494, {{14965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1495, {{14966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1496, {{14968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1497, {{14970, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1498, {{14971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1499, {{14972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1500, {{14973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1501, {{14974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1502, {{14975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1503, {{14976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1504, {{14977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1505, {{14978, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1506, {{14979, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1507, {{14980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1508, {{14981, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1509, {{14982, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1510, {{14983, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1511, {{14984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1512, {{14985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1513, {{14986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1514, {{14987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1515, {{14988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1516, {{14989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57592, {{14990, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1517, {{14991, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1518, {{14992, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1519, {{14993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1520, {{14994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1521, {{14995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1522, {{14996, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1523, {{14997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1524, {{14998, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1525, {{14999, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1526, {{15000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57599, {{15002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1527, {{15003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{1528, {{15004, 1, 0, 0},{15004, 1, 1, 16}, {0, 0, 0, 0}}},{1528, {{15004, 1, 0, 0},{15004, 1, 1, 16}, {0, 0, 0, 0}}},{57606, {{15004, 1, 0, 0},{15004, 1, 1, 17}, {0, 0, 0, 0}}},{57606, {{15004, 1, 0, 0},{15004, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{1529, {{15005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1530, {{15006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1531, {{15007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1532, {{15008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1533, {{15009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1534, {{15010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1535, {{15011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{1536, {{15012, 1, 0, 0},{15012, 1, 1, 17}, {0, 0, 0, 0}}},{1536, {{15012, 1, 0, 0},{15012, 1, 1, 17}, {0, 0, 0, 0}}},{1537, {{15012, 1, 0, 0},{15012, 1, 1, 16}, {0, 0, 0, 0}}},{1537, {{15012, 1, 0, 0},{15012, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{1538, {{15013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1539, {{15014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1540, {{15015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1541, {{15016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1542, {{15017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1543, {{15020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1544, {{15021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1545, {{15022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1546, {{15023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1547, {{15024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1548, {{15025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{68003, {{15026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{68004, {{15027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1549, {{15028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1551, {{15029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1552, {{15030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1553, {{15031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1554, {{15032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1555, {{15033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1556, {{15034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1557, {{15035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1558, {{15036, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1559, {{15037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1560, {{15038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1561, {{15039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1562, {{15040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1563, {{15041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1564, {{15042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1565, {{15043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{1566, {{15044, 1, 0, 0},{15044, 1, 1, 17}, {0, 0, 0, 0}}},{1566, {{15044, 1, 0, 0},{15044, 1, 1, 17}, {0, 0, 0, 0}}},{59671, {{15044, 1, 0, 0},{15044, 1, 1, 18}, {0, 0, 0, 0}}},{59671, {{15044, 1, 0, 0},{15044, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{1567, {{15045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{68005, {{15047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{68006, {{15048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1568, {{15050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1569, {{15051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1570, {{15053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1571, {{15054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1572, {{15055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1573, {{15056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1574, {{15057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1575, {{15058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1576, {{15059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1577, {{15060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1578, {{15061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1579, {{15062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1580, {{15063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1581, {{15064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1582, {{15065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1583, {{15066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1584, {{15067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1585, {{15069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1586, {{15070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1587, {{15071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{1588, {{15072, 1, 0, 0},{15072, 1, 1, 16}, {0, 0, 0, 0}}},{1588, {{15072, 1, 0, 0},{15072, 1, 1, 16}, {0, 0, 0, 0}}},{1589, {{15072, 1, 0, 0},{15072, 1, 1, 18}, {0, 0, 0, 0}}},{1589, {{15072, 1, 0, 0},{15072, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{1590, {{15073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1591, {{15074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1592, {{15076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1593, {{15077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{1594, {{15078, 1, 0, 0},{15078, 1, 1, 17}, {0, 0, 0, 0}}},{1594, {{15078, 1, 0, 0},{15078, 1, 1, 17}, {0, 0, 0, 0}}},{1595, {{15078, 1, 0, 0},{15078, 1, 1, 16}, {0, 0, 0, 0}}},{1595, {{15078, 1, 0, 0},{15078, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{1596, {{15079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1597, {{15080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1598, {{15081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1599, {{15082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1600, {{15083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1601, {{15084, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1602, {{15087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1604, {{15088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1605, {{15089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1606, {{15090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1607, {{15091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{1608, {{15092, 1, 0, 0},{15092, 1, 1, 16}, {0, 0, 0, 0}}},{1608, {{15092, 1, 0, 0},{15092, 1, 1, 16}, {0, 0, 0, 0}}},{1609, {{15092, 1, 0, 0},{15092, 1, 1, 17}, {0, 0, 0, 0}}},{1609, {{15092, 1, 0, 0},{15092, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{1610, {{15093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1611, {{15094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{1612, {{15095, 1, 0, 0},{15095, 1, 1, 17}, {0, 0, 0, 0}}},{1612, {{15095, 1, 0, 0},{15095, 1, 1, 17}, {0, 0, 0, 0}}},{1613, {{15095, 1, 0, 0},{15095, 1, 1, 16}, {0, 0, 0, 0}}},{1613, {{15095, 1, 0, 0},{15095, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{1614, {{15096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1615, {{15097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1616, {{15098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1617, {{15099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1618, {{15100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1619, {{15101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1620, {{15105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1621, {{15106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1622, {{15107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1623, {{15109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1625, {{15110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1626, {{15111, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{1627, {{15112, 1, 0, 0},{15112, 1, 1, 16}, {0, 0, 0, 0}}},{1627, {{15112, 1, 0, 0},{15112, 1, 1, 16}, {0, 0, 0, 0}}},{1628, {{15112, 1, 0, 0},{15112, 1, 1, 17}, {0, 0, 0, 0}}},{1628, {{15112, 1, 0, 0},{15112, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{1629, {{15113, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{1630, {{15114, 1, 0, 0},{15114, 1, 1, 16}, {0, 0, 0, 0}}},{1630, {{15114, 1, 0, 0},{15114, 1, 1, 16}, {0, 0, 0, 0}}},{1631, {{15114, 1, 0, 0},{15114, 1, 1, 17}, {0, 0, 0, 0}}},{1631, {{15114, 1, 0, 0},{15114, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{1632, {{15115, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1633, {{15116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1634, {{15117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1635, {{15118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1636, {{15119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1637, {{15120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1638, {{15122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1639, {{15123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1640, {{15124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1642, {{15127, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1643, {{15128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1644, {{15129, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1646, {{15130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1647, {{15131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1648, {{15132, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{1649, {{15133, 1, 0, 0},{15133, 1, 1, 16}, {0, 0, 0, 0}}},{1649, {{15133, 1, 0, 0},{15133, 1, 1, 16}, {0, 0, 0, 0}}},{1650, {{15133, 1, 0, 0},{15133, 1, 1, 17}, {0, 0, 0, 0}}},{1650, {{15133, 1, 0, 0},{15133, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{1651, {{15134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1652, {{15136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1653, {{15137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1654, {{15138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1655, {{15139, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1656, {{15140, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1657, {{15141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{1658, {{15142, 1, 0, 0},{15142, 1, 1, 17}, {0, 0, 0, 0}}},{1658, {{15142, 1, 0, 0},{15142, 1, 1, 17}, {0, 0, 0, 0}}},{1659, {{15142, 1, 0, 0},{15142, 1, 1, 16}, {0, 0, 0, 0}}},{1659, {{15142, 1, 0, 0},{15142, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {6, {{1660, {{15143, 1, 0, 0},{15143, 1, 1, 16}, {0, 0, 0, 0}}},{1660, {{15143, 1, 0, 0},{15143, 1, 1, 16}, {0, 0, 0, 0}}},{1661, {{15143, 1, 0, 0},{15143, 1, 1, 17}, {0, 0, 0, 0}}},{1661, {{15143, 1, 0, 0},{15143, 1, 1, 17}, {0, 0, 0, 0}}},{1662, {{15143, 1, 0, 0},{15143, 1, 1, 18}, {0, 0, 0, 0}}},{1662, {{15143, 1, 0, 0},{15143, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1663, {{15145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1664, {{15146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1665, {{15147, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1666, {{15148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1667, {{15149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1669, {{15150, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1670, {{15151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1671, {{15152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1672, {{15153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1673, {{15154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1674, {{15155, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1675, {{15156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1676, {{15157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1677, {{15158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1679, {{15160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1680, {{15161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1681, {{15163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{1682, {{15164, 1, 0, 0},{15164, 1, 1, 17}, {0, 0, 0, 0}}},{1682, {{15164, 1, 0, 0},{15164, 1, 1, 17}, {0, 0, 0, 0}}},{1683, {{15164, 1, 0, 0},{15164, 1, 1, 16}, {0, 0, 0, 0}}},{1683, {{15164, 1, 0, 0},{15164, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1684, {{15167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1685, {{15169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1686, {{15170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1687, {{15171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1688, {{15172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1689, {{15175, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1690, {{15176, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1691, {{15177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1692, {{15178, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1693, {{15179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1694, {{15180, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1695, {{15185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{1696, {{15186, 1, 0, 0},{15186, 1, 1, 16}, {0, 0, 0, 0}}},{1696, {{15186, 1, 0, 0},{15186, 1, 1, 16}, {0, 0, 0, 0}}},{57714, {{15186, 1, 0, 0},{15186, 1, 1, 17}, {0, 0, 0, 0}}},{57714, {{15186, 1, 0, 0},{15186, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{1697, {{15187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1698, {{15188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1699, {{15189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1700, {{15192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1701, {{15193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1702, {{15194, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1703, {{15195, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1705, {{15196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1706, {{15197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1707, {{15198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1708, {{15199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1709, {{15200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1710, {{15201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1711, {{15202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1712, {{15208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1713, {{15209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1716, {{15210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1717, {{15211, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1718, {{15212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1719, {{15213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1720, {{15214, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1721, {{15215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1722, {{15216, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1723, {{15217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1724, {{15218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{1726, {{15220, 1, 0, 0},{15220, 1, 1, 16}, {0, 0, 0, 0}}},{1726, {{15220, 1, 0, 0},{15220, 1, 1, 16}, {0, 0, 0, 0}}},{1727, {{15220, 1, 0, 0},{15220, 1, 1, 17}, {0, 0, 0, 0}}},{1727, {{15220, 1, 0, 0},{15220, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1728, {{15223, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{1729, {{15224, 1, 0, 0},{15224, 1, 1, 17}, {0, 0, 0, 0}}},{1729, {{15224, 1, 0, 0},{15224, 1, 1, 17}, {0, 0, 0, 0}}},{1730, {{15224, 1, 0, 0},{15224, 1, 1, 16}, {0, 0, 0, 0}}},{1730, {{15224, 1, 0, 0},{15224, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{1731, {{15225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1732, {{15226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1733, {{15227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1734, {{15228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1735, {{15229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1736, {{15230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1737, {{15231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1738, {{15232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1739, {{15233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1740, {{15234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1741, {{15235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1742, {{15236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1743, {{15237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1744, {{15238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1745, {{15239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1746, {{15240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1747, {{15242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1748, {{15243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1749, {{15244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1750, {{15245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1752, {{15246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1753, {{15247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1754, {{15248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1755, {{15249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1756, {{15250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1757, {{15251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1758, {{15252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1759, {{15253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1760, {{15254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1761, {{15255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1762, {{15256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1763, {{15257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1764, {{15258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1765, {{15259, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1766, {{15260, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1767, {{15261, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1768, {{15262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1769, {{15263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1770, {{15265, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1772, {{15268, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1774, {{15269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1775, {{15270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1776, {{15271, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1777, {{15272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1778, {{15273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1779, {{15274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1780, {{15275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1782, {{15276, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1783, {{15277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{1784, {{15278, 1, 0, 0},{15278, 1, 1, 16}, {0, 0, 0, 0}}},{1784, {{15278, 1, 0, 0},{15278, 1, 1, 16}, {0, 0, 0, 0}}},{57785, {{15278, 1, 0, 0},{15278, 1, 1, 17}, {0, 0, 0, 0}}},{57785, {{15278, 1, 0, 0},{15278, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{1785, {{15279, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1786, {{15280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1787, {{15281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1788, {{15282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1789, {{15283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1790, {{15284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{1791, {{15285, 1, 0, 0},{15285, 1, 1, 17}, {0, 0, 0, 0}}},{1791, {{15285, 1, 0, 0},{15285, 1, 1, 17}, {0, 0, 0, 0}}},{1792, {{15285, 1, 0, 0},{15285, 1, 1, 16}, {0, 0, 0, 0}}},{1792, {{15285, 1, 0, 0},{15285, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{1793, {{15286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1794, {{15287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{1795, {{15288, 1, 0, 0},{15288, 1, 1, 16}, {0, 0, 0, 0}}},{1795, {{15288, 1, 0, 0},{15288, 1, 1, 16}, {0, 0, 0, 0}}},{1796, {{15288, 1, 0, 0},{15288, 1, 1, 17}, {0, 0, 0, 0}}},{1796, {{15288, 1, 0, 0},{15288, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{1797, {{15289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1798, {{15290, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1799, {{15291, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1800, {{15295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1801, {{15296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1802, {{15297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1803, {{15298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1804, {{15299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1805, {{15300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1806, {{15301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1807, {{15302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1808, {{15303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1809, {{15304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1810, {{15305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1811, {{15306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1812, {{15307, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1813, {{15308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{1814, {{15309, 1, 0, 0},{15309, 1, 1, 17}, {0, 0, 0, 0}}},{1814, {{15309, 1, 0, 0},{15309, 1, 1, 17}, {0, 0, 0, 0}}},{57784, {{15309, 1, 0, 0},{15309, 1, 1, 18}, {0, 0, 0, 0}}},{57784, {{15309, 1, 0, 0},{15309, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{1815, {{15310, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1816, {{15311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1817, {{15312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1818, {{15315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1819, {{15316, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1820, {{15317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1821, {{15318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1822, {{15319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1823, {{15320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1824, {{15321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1825, {{15322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1826, {{15323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1827, {{15324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1828, {{15325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1830, {{15327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1831, {{15328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1832, {{15329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1833, {{15330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1834, {{15331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1835, {{15332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1836, {{15333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1837, {{15334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1838, {{15338, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1839, {{15339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{1840, {{15340, 1, 0, 0},{15340, 1, 1, 16}, {0, 0, 0, 0}}},{1840, {{15340, 1, 0, 0},{15340, 1, 1, 16}, {0, 0, 0, 0}}},{1841, {{15340, 1, 0, 0},{15340, 1, 1, 17}, {0, 0, 0, 0}}},{1841, {{15340, 1, 0, 0},{15340, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{1842, {{15341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1843, {{15342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1844, {{15343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1845, {{15344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1846, {{15345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1847, {{15346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{1848, {{15347, 1, 0, 0},{15347, 1, 1, 17}, {0, 0, 0, 0}}},{1848, {{15347, 1, 0, 0},{15347, 1, 1, 17}, {0, 0, 0, 0}}},{1849, {{15347, 1, 0, 0},{15347, 1, 1, 18}, {0, 0, 0, 0}}},{1849, {{15347, 1, 0, 0},{15347, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57841, {{15350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1850, {{15351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1851, {{15352, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1852, {{15354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1853, {{15355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57853, {{15357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{1854, {{15358, 1, 0, 0},{15358, 1, 1, 17}, {0, 0, 0, 0}}},{1854, {{15358, 1, 0, 0},{15358, 1, 1, 17}, {0, 0, 0, 0}}},{1855, {{15358, 1, 0, 0},{15358, 1, 1, 16}, {0, 0, 0, 0}}},{1855, {{15358, 1, 0, 0},{15358, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{1856, {{15359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1857, {{15360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1858, {{15361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1859, {{15362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1860, {{15363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1861, {{15364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{1862, {{15365, 1, 0, 0},{15365, 1, 1, 17}, {0, 0, 0, 0}}},{1862, {{15365, 1, 0, 0},{15365, 1, 1, 17}, {0, 0, 0, 0}}},{1863, {{15365, 1, 0, 0},{15365, 1, 1, 16}, {0, 0, 0, 0}}},{1863, {{15365, 1, 0, 0},{15365, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{1864, {{15366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1866, {{15369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1867, {{15370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1868, {{15371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1869, {{15372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{1870, {{15373, 1, 0, 0},{15373, 1, 1, 16}, {0, 0, 0, 0}}},{1870, {{15373, 1, 0, 0},{15373, 1, 1, 16}, {0, 0, 0, 0}}},{1871, {{15373, 1, 0, 0},{15373, 1, 1, 17}, {0, 0, 0, 0}}},{1871, {{15373, 1, 0, 0},{15373, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{1872, {{15374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1873, {{15375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1874, {{15376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1875, {{15377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1876, {{15378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1877, {{15379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1878, {{15380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1879, {{15381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1880, {{15382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1881, {{15383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1882, {{15384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1883, {{15385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1884, {{15386, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1885, {{15387, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1887, {{15389, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1888, {{15390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1889, {{15391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1890, {{15392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1891, {{15393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1892, {{15394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1893, {{15395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1894, {{15396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1895, {{15397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1896, {{15398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1897, {{15399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1898, {{15400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1899, {{15401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1900, {{15402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1901, {{15403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1902, {{15404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1903, {{15405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1904, {{15406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1905, {{15407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1906, {{15408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1907, {{15409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1908, {{15410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1909, {{15411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1910, {{15412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1911, {{15413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1912, {{15414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1913, {{15415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1914, {{15416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1915, {{15417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1916, {{15418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1917, {{15419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1918, {{15420, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1919, {{15421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1920, {{15422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1921, {{15423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1922, {{15424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1923, {{15425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1924, {{15426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1925, {{15427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1926, {{15428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1927, {{15429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1928, {{15430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1929, {{15431, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1930, {{15432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1931, {{15433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1932, {{15434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1933, {{15435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1934, {{15436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1935, {{15437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{1936, {{15438, 1, 0, 0},{15438, 1, 1, 17}, {0, 0, 0, 0}}},{1936, {{15438, 1, 0, 0},{15438, 1, 1, 17}, {0, 0, 0, 0}}},{1937, {{15438, 1, 0, 0},{15438, 1, 1, 16}, {0, 0, 0, 0}}},{1937, {{15438, 1, 0, 0},{15438, 1, 1, 16}, {0, 0, 0, 0}}},{68056, {{15438, 1, 0, 0},{15438, 1, 1, 18}, {0, 0, 0, 0}}},{68056, {{15438, 1, 0, 0},{15438, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{1938, {{15439, 1, 0, 0},{15439, 1, 1, 16}, {0, 0, 0, 0}}},{1938, {{15439, 1, 0, 0},{15439, 1, 1, 16}, {0, 0, 0, 0}}},{57890, {{15439, 1, 0, 0},{15439, 1, 1, 17}, {0, 0, 0, 0}}},{57890, {{15439, 1, 0, 0},{15439, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1939, {{15442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1940, {{15445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1941, {{15446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1942, {{15447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1943, {{15448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1944, {{15449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1945, {{15450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1946, {{15451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1947, {{15452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1948, {{15453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1949, {{15454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1950, {{15455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1951, {{15456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1952, {{15457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1953, {{15458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1954, {{15459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1955, {{15460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1956, {{15461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1957, {{15462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1958, {{15463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1959, {{15464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1960, {{15466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1961, {{15467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1962, {{15468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1963, {{15469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1964, {{15471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1965, {{15472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1966, {{15473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1967, {{15474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1968, {{15475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1970, {{15476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1971, {{15477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1972, {{15478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1973, {{15480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1974, {{15481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1975, {{15482, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1976, {{15483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1977, {{15484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1978, {{15485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1979, {{15486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1980, {{15487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1981, {{15488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1982, {{15489, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1983, {{15490, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1984, {{15491, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1985, {{15492, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1986, {{15493, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1987, {{15494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1988, {{15495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1989, {{15496, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1990, {{15497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{1991, {{15498, 1, 0, 0},{15498, 1, 1, 17}, {0, 0, 0, 0}}},{1991, {{15498, 1, 0, 0},{15498, 1, 1, 17}, {0, 0, 0, 0}}},{1992, {{15498, 1, 0, 0},{15498, 1, 1, 16}, {0, 0, 0, 0}}},{1992, {{15498, 1, 0, 0},{15498, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{1993, {{15499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1994, {{15500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1995, {{15501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1996, {{15502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1997, {{15503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1998, {{15506, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1999, {{15507, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2000, {{15508, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2001, {{15509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2002, {{15510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2003, {{15511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2004, {{15513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2005, {{15514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2006, {{15515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2007, {{15516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2008, {{15517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2009, {{15518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2010, {{15519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2011, {{15520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2012, {{15521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2013, {{15522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2014, {{15523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2015, {{15524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2016, {{15525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2017, {{15526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2018, {{15527, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2019, {{15528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2020, {{15530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2021, {{15531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2022, {{15532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2023, {{15533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2024, {{15535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2025, {{15536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2026, {{15537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2027, {{15539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2028, {{15540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2029, {{15541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2030, {{15542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2031, {{15543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2032, {{15544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2033, {{15545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2034, {{15546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2035, {{15547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2036, {{15548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2037, {{15549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2038, {{15550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2039, {{15552, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2040, {{15553, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2041, {{15554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2042, {{15555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2043, {{15556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2044, {{15557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2045, {{15558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2046, {{15560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2047, {{15562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2048, {{15563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2049, {{15564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2050, {{15565, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2051, {{15566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2052, {{15567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2053, {{15568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2054, {{15569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2055, {{15570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2056, {{15574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2057, {{15575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2058, {{15576, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2059, {{15577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2060, {{15578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2061, {{15579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2062, {{15580, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2063, {{15582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{2064, {{15583, 1, 0, 0},{15583, 1, 1, 16}, {0, 0, 0, 0}}},{2064, {{15583, 1, 0, 0},{15583, 1, 1, 16}, {0, 0, 0, 0}}},{2065, {{15583, 1, 0, 0},{15583, 1, 1, 18}, {0, 0, 0, 0}}},{2065, {{15583, 1, 0, 0},{15583, 1, 1, 18}, {0, 0, 0, 0}}},{2066, {{15583, 1, 0, 0},{15583, 1, 1, 17}, {0, 0, 0, 0}}},{2066, {{15583, 1, 0, 0},{15583, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{2067, {{15588, 1, 0, 0},{15588, 1, 1, 16}, {0, 0, 0, 0}}},{2067, {{15588, 1, 0, 0},{15588, 1, 1, 16}, {0, 0, 0, 0}}},{2068, {{15588, 1, 0, 0},{15588, 1, 1, 17}, {0, 0, 0, 0}}},{2068, {{15588, 1, 0, 0},{15588, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{2069, {{15589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2070, {{15590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2071, {{15591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2072, {{15593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2073, {{15594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2074, {{15596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2075, {{15597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2076, {{15598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2077, {{15599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2079, {{15604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2080, {{15605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2081, {{15606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2082, {{15607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2083, {{15608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2085, {{15609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2086, {{15610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2087, {{15611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2088, {{15612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2089, {{15613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2091, {{15614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2092, {{15615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{2093, {{15616, 1, 0, 0},{15616, 1, 1, 16}, {0, 0, 0, 0}}},{2093, {{15616, 1, 0, 0},{15616, 1, 1, 16}, {0, 0, 0, 0}}},{2094, {{15616, 1, 0, 0},{15616, 1, 1, 17}, {0, 0, 0, 0}}},{2094, {{15616, 1, 0, 0},{15616, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{2095, {{15617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2096, {{15618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2097, {{15619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{2098, {{15620, 1, 0, 0},{15620, 1, 1, 17}, {0, 0, 0, 0}}},{2098, {{15620, 1, 0, 0},{15620, 1, 1, 17}, {0, 0, 0, 0}}},{2099, {{15620, 1, 0, 0},{15620, 1, 1, 16}, {0, 0, 0, 0}}},{2099, {{15620, 1, 0, 0},{15620, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{2100, {{15621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2101, {{15622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2102, {{15623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2103, {{15624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2104, {{15625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2105, {{15626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2106, {{15631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2107, {{15632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2108, {{15633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2109, {{15634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2110, {{15635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2111, {{15636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2112, {{15637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2113, {{15638, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2114, {{15639, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2115, {{15640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2116, {{15641, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2117, {{15642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{2118, {{15643, 1, 0, 0},{15643, 1, 1, 16}, {0, 0, 0, 0}}},{2118, {{15643, 1, 0, 0},{15643, 1, 1, 16}, {0, 0, 0, 0}}},{2119, {{15643, 1, 0, 0},{15643, 1, 1, 17}, {0, 0, 0, 0}}},{2119, {{15643, 1, 0, 0},{15643, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{2120, {{15644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2121, {{15645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{2122, {{15646, 1, 0, 0},{15646, 1, 1, 18}, {0, 0, 0, 0}}},{2122, {{15646, 1, 0, 0},{15646, 1, 1, 18}, {0, 0, 0, 0}}},{2123, {{15646, 1, 0, 0},{15646, 1, 1, 17}, {0, 0, 0, 0}}},{2123, {{15646, 1, 0, 0},{15646, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2125, {{15648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2127, {{15653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2128, {{15654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2129, {{15655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2130, {{15656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2131, {{15657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2132, {{15658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2133, {{15659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2134, {{15660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2135, {{15661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2136, {{15662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2137, {{15663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2138, {{15664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{2139, {{15665, 1, 0, 0},{15665, 1, 1, 17}, {0, 0, 0, 0}}},{2139, {{15665, 1, 0, 0},{15665, 1, 1, 17}, {0, 0, 0, 0}}},{2140, {{15665, 1, 0, 0},{15665, 1, 1, 18}, {0, 0, 0, 0}}},{2140, {{15665, 1, 0, 0},{15665, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{2141, {{15666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2142, {{15667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2143, {{15668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2144, {{15669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2145, {{15670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2146, {{15671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2147, {{15672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2148, {{15676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2149, {{15677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2150, {{15678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2151, {{15679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2152, {{15680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2153, {{15681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2155, {{15682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2156, {{15683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2157, {{15684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2158, {{15685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2159, {{15686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2160, {{15690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2161, {{15691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2162, {{15692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2163, {{15693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2164, {{15694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2165, {{15695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2166, {{15696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2167, {{15697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2168, {{15698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2169, {{15699, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2170, {{15700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2171, {{15701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2172, {{15702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2173, {{15703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2174, {{15704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{2176, {{15709, 1, 0, 0},{15709, 1, 1, 16}, {0, 0, 0, 0}}},{2176, {{15709, 1, 0, 0},{15709, 1, 1, 16}, {0, 0, 0, 0}}},{2177, {{15709, 1, 0, 0},{15709, 1, 1, 17}, {0, 0, 0, 0}}},{2177, {{15709, 1, 0, 0},{15709, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{2178, {{15710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2179, {{15711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2180, {{15712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2181, {{15713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2182, {{15714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2183, {{15715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2184, {{15716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2185, {{15717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2186, {{15718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{2187, {{15719, 1, 0, 0},{15719, 1, 1, 16}, {0, 0, 0, 0}}},{2187, {{15719, 1, 0, 0},{15719, 1, 1, 16}, {0, 0, 0, 0}}},{2188, {{15719, 1, 0, 0},{15719, 1, 1, 17}, {0, 0, 0, 0}}},{2188, {{15719, 1, 0, 0},{15719, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{2189, {{15720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2190, {{15721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2192, {{15722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2194, {{15723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2195, {{15724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2196, {{15725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2197, {{15726, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2198, {{15727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2199, {{15728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2200, {{15729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2201, {{15730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2202, {{15731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2203, {{15732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2204, {{15733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2205, {{15734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2206, {{15735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2207, {{15736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2208, {{15737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2209, {{15738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2210, {{15739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2211, {{15740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2212, {{15741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{2213, {{15742, 1, 0, 0},{15742, 1, 1, 16}, {0, 0, 0, 0}}},{2213, {{15742, 1, 0, 0},{15742, 1, 1, 16}, {0, 0, 0, 0}}},{2214, {{15742, 1, 0, 0},{15742, 1, 1, 17}, {0, 0, 0, 0}}},{2214, {{15742, 1, 0, 0},{15742, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{2215, {{15743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{2216, {{15744, 1, 0, 0},{15744, 1, 1, 16}, {0, 0, 0, 0}}},{2216, {{15744, 1, 0, 0},{15744, 1, 1, 16}, {0, 0, 0, 0}}},{2217, {{15744, 1, 0, 0},{15744, 1, 1, 17}, {0, 0, 0, 0}}},{2217, {{15744, 1, 0, 0},{15744, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{2218, {{15745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2219, {{15748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2220, {{15749, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2221, {{15750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2222, {{15751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2224, {{15752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2225, {{15756, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2226, {{15757, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2227, {{15758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2228, {{15759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2229, {{15762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{2230, {{15763, 1, 0, 0},{15763, 1, 1, 17}, {0, 0, 0, 0}}},{2230, {{15763, 1, 0, 0},{15763, 1, 1, 17}, {0, 0, 0, 0}}},{2231, {{15763, 1, 0, 0},{15763, 1, 1, 16}, {0, 0, 0, 0}}},{2231, {{15763, 1, 0, 0},{15763, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{2232, {{15764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2234, {{15766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2235, {{15767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2236, {{15768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2238, {{15770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2239, {{15771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2240, {{15772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2241, {{15774, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2242, {{15775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2243, {{15776, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2244, {{15777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2245, {{15779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2246, {{15780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2247, {{15781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2248, {{15782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2249, {{15783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2251, {{15784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2252, {{15788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2253, {{15789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2254, {{15790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2255, {{15791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2256, {{15792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2257, {{15793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{2258, {{15795, 1, 0, 0},{15795, 1, 1, 16}, {0, 0, 0, 0}}},{2258, {{15795, 1, 0, 0},{15795, 1, 1, 16}, {0, 0, 0, 0}}},{57991, {{15795, 1, 0, 0},{15795, 1, 1, 17}, {0, 0, 0, 0}}},{57991, {{15795, 1, 0, 0},{15795, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{2259, {{15796, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2260, {{15797, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2262, {{15801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2263, {{15802, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2264, {{15803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2265, {{15804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2266, {{15806, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2267, {{15807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{2268, {{15808, 1, 0, 0},{15808, 1, 1, 18}, {0, 0, 0, 0}}},{2268, {{15808, 1, 0, 0},{15808, 1, 1, 18}, {0, 0, 0, 0}}},{2269, {{15808, 1, 0, 0},{15808, 1, 1, 17}, {0, 0, 0, 0}}},{2269, {{15808, 1, 0, 0},{15808, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2270, {{15810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2271, {{15811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2272, {{15812, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2273, {{15813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2274, {{15814, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2275, {{15815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2276, {{15816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2277, {{15817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2278, {{15818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2279, {{15819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2280, {{15822, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2281, {{15823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2282, {{15824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2283, {{15825, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2284, {{15826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2285, {{15827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{2286, {{15828, 1, 0, 0},{15828, 1, 1, 18}, {0, 0, 0, 0}}},{2286, {{15828, 1, 0, 0},{15828, 1, 1, 18}, {0, 0, 0, 0}}},{2287, {{15828, 1, 0, 0},{15828, 1, 1, 17}, {0, 0, 0, 0}}},{2287, {{15828, 1, 0, 0},{15828, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{2288, {{15829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2289, {{15830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2291, {{15832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2292, {{15833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{2293, {{15839, 1, 0, 0},{15839, 1, 1, 17}, {0, 0, 0, 0}}},{2293, {{15839, 1, 0, 0},{15839, 1, 1, 17}, {0, 0, 0, 0}}},{2294, {{15839, 1, 0, 0},{15839, 1, 1, 16}, {0, 0, 0, 0}}},{2294, {{15839, 1, 0, 0},{15839, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{2295, {{15840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2296, {{15841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2297, {{15842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2298, {{15843, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2299, {{15844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2300, {{15845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2301, {{15846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2302, {{15851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2303, {{15852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{2305, {{15853, 1, 0, 0},{15853, 1, 1, 17}, {0, 0, 0, 0}}},{2305, {{15853, 1, 0, 0},{15853, 1, 1, 17}, {0, 0, 0, 0}}},{2306, {{15853, 1, 0, 0},{15853, 1, 1, 16}, {0, 0, 0, 0}}},{2306, {{15853, 1, 0, 0},{15853, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{2307, {{15854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2308, {{15855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2309, {{15856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{2310, {{15857, 1, 0, 0},{15857, 1, 1, 16}, {0, 0, 0, 0}}},{2310, {{15857, 1, 0, 0},{15857, 1, 1, 16}, {0, 0, 0, 0}}},{58003, {{15857, 1, 0, 0},{15857, 1, 1, 17}, {0, 0, 0, 0}}},{58003, {{15857, 1, 0, 0},{15857, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2311, {{15859, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2312, {{15860, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2313, {{15861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2314, {{15862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2315, {{15863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2316, {{15864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2317, {{15865, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2318, {{15867, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2319, {{15868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2320, {{15870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2321, {{15871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2322, {{15872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2323, {{15873, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{2324, {{15874, 1, 0, 0},{15874, 1, 1, 17}, {0, 0, 0, 0}}},{2324, {{15874, 1, 0, 0},{15874, 1, 1, 17}, {0, 0, 0, 0}}},{2325, {{15874, 1, 0, 0},{15874, 1, 1, 16}, {0, 0, 0, 0}}},{2325, {{15874, 1, 0, 0},{15874, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{2326, {{15875, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2327, {{15876, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2328, {{15877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2329, {{15880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2330, {{15881, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2331, {{15882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2332, {{15883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2333, {{15885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2334, {{15886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{2335, {{15887, 1, 0, 0},{15887, 1, 1, 16}, {0, 0, 0, 0}}},{2335, {{15887, 1, 0, 0},{15887, 1, 1, 16}, {0, 0, 0, 0}}},{2336, {{15887, 1, 0, 0},{15887, 1, 1, 17}, {0, 0, 0, 0}}},{2336, {{15887, 1, 0, 0},{15887, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{2337, {{15888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2338, {{15889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2339, {{15890, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{58161, {{15892, 1, 0, 0},{15892, 1, 1, 16}, {0, 0, 0, 0}}},{58161, {{15892, 1, 0, 0},{15892, 1, 1, 16}, {0, 0, 0, 0}}},{58168, {{15892, 1, 0, 0},{15892, 1, 1, 17}, {0, 0, 0, 0}}},{58168, {{15892, 1, 0, 0},{15892, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{2340, {{15893, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2341, {{15894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2342, {{15895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2343, {{15896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2344, {{15897, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2345, {{15898, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2346, {{15899, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2347, {{15900, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2348, {{15901, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2349, {{15902, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2350, {{15903, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2351, {{15904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2352, {{15905, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2353, {{15906, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2354, {{15907, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2355, {{15908, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2356, {{15909, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2357, {{15910, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2358, {{15911, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2359, {{15912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2360, {{15913, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2361, {{15914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2362, {{15915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2363, {{15916, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2364, {{15917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2365, {{15918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2366, {{15919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2367, {{15920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2368, {{15921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2369, {{15922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2370, {{15923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2371, {{15925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2373, {{15926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{2374, {{15927, 1, 0, 0},{15927, 1, 1, 16}, {0, 0, 0, 0}}},{2374, {{15927, 1, 0, 0},{15927, 1, 1, 16}, {0, 0, 0, 0}}},{2375, {{15927, 1, 0, 0},{15927, 1, 1, 17}, {0, 0, 0, 0}}},{2375, {{15927, 1, 0, 0},{15927, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{2376, {{15928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2377, {{15929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2378, {{15930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2379, {{15931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2380, {{15932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2381, {{15933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2382, {{15934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2383, {{15935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2384, {{15936, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2385, {{15937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2386, {{15938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2387, {{15939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2388, {{15940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2389, {{15941, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2390, {{15942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2391, {{15943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2392, {{15945, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2393, {{15946, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2394, {{15947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2395, {{15948, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2396, {{15949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2397, {{15950, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2398, {{15951, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2399, {{15952, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2400, {{15953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2401, {{15954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2402, {{15955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2403, {{15956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2404, {{15957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2405, {{15958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2406, {{15959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2407, {{15960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2408, {{15961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2409, {{15962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2410, {{15963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2411, {{15964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2412, {{15965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2413, {{15966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2414, {{15967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2415, {{15968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2416, {{15969, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2417, {{15970, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2418, {{15971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2419, {{15972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2420, {{15973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2421, {{15974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2422, {{15975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2423, {{15976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2424, {{15979, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2425, {{15980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2426, {{15981, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2427, {{15982, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2428, {{15983, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2429, {{15985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2430, {{15986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2431, {{15987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2432, {{15988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2433, {{15989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2434, {{15990, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2435, {{15991, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2436, {{15992, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2437, {{15993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2438, {{15994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2439, {{15995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2440, {{15996, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2441, {{15997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2442, {{15998, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2443, {{15999, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2444, {{16000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2445, {{16001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2446, {{16002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{2447, {{16003, 1, 0, 0},{16003, 1, 1, 17}, {0, 0, 0, 0}}},{2447, {{16003, 1, 0, 0},{16003, 1, 1, 17}, {0, 0, 0, 0}}},{2448, {{16003, 1, 0, 0},{16003, 1, 1, 18}, {0, 0, 0, 0}}},{2448, {{16003, 1, 0, 0},{16003, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{2449, {{16004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2450, {{16005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2451, {{16007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2452, {{16008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2453, {{16009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2454, {{16010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2455, {{16011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2456, {{16012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2457, {{16014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2458, {{16015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2459, {{16016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2460, {{16017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2461, {{16018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2462, {{16019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2463, {{16020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2464, {{16021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2465, {{16022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2466, {{16023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2467, {{16024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{2468, {{16026, 1, 0, 0},{16026, 1, 1, 17}, {0, 0, 0, 0}}},{2468, {{16026, 1, 0, 0},{16026, 1, 1, 17}, {0, 0, 0, 0}}},{2469, {{16026, 1, 0, 0},{16026, 1, 1, 16}, {0, 0, 0, 0}}},{2469, {{16026, 1, 0, 0},{16026, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{2470, {{16027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2471, {{16028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2472, {{16029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2473, {{16030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2474, {{16031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2475, {{16032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2476, {{16033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2477, {{16035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2478, {{16036, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2479, {{16037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2480, {{16040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2481, {{16041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2482, {{16042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2483, {{16043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2484, {{16044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2485, {{16045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2486, {{16046, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2487, {{16048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2488, {{16049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2489, {{16050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2490, {{16051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2491, {{16052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2492, {{16055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2493, {{16056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2494, {{16057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2495, {{16058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2496, {{16063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2498, {{16064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2499, {{16065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2500, {{16066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2501, {{16067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2502, {{16068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2503, {{16070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2505, {{16073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2506, {{16074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2507, {{16075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2508, {{16076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2509, {{16077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2510, {{16078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2511, {{16081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2512, {{16082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2513, {{16083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2514, {{16084, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2515, {{16085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2516, {{16086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2517, {{16087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2518, {{16090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2519, {{16093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2520, {{16094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2521, {{16095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2522, {{16096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2523, {{16097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2525, {{16103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{2526, {{16104, 1, 0, 0},{16104, 1, 1, 16}, {0, 0, 0, 0}}},{2526, {{16104, 1, 0, 0},{16104, 1, 1, 16}, {0, 0, 0, 0}}},{59908, {{16104, 1, 0, 0},{16104, 1, 1, 17}, {0, 0, 0, 0}}},{59908, {{16104, 1, 0, 0},{16104, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2527, {{16107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2528, {{16108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2529, {{16109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2530, {{16110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2531, {{16111, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2532, {{16112, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2533, {{16113, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2534, {{16114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2535, {{16117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2536, {{16118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2537, {{16119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2538, {{16120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2539, {{16121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2540, {{16122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2541, {{16125, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2542, {{16126, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2543, {{16127, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2546, {{16132, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2547, {{16135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2548, {{16137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2549, {{16138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{2550, {{16139, 1, 0, 0},{16139, 1, 1, 17}, {0, 0, 0, 0}}},{2550, {{16139, 1, 0, 0},{16139, 1, 1, 17}, {0, 0, 0, 0}}},{2551, {{16139, 1, 0, 0},{16139, 1, 1, 16}, {0, 0, 0, 0}}},{2551, {{16139, 1, 0, 0},{16139, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{2552, {{16140, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2553, {{16141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2554, {{16142, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2555, {{16143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2556, {{16144, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2557, {{16145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2558, {{16146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2559, {{16147, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2560, {{16148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2561, {{16149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2562, {{16150, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2563, {{16151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2564, {{16152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2565, {{16153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2566, {{16154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{2567, {{16155, 1, 0, 0},{16155, 1, 1, 16}, {0, 0, 0, 0}}},{2567, {{16155, 1, 0, 0},{16155, 1, 1, 16}, {0, 0, 0, 0}}},{2568, {{16155, 1, 0, 0},{16155, 1, 1, 17}, {0, 0, 0, 0}}},{2568, {{16155, 1, 0, 0},{16155, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{2569, {{16156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2570, {{16157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2571, {{16158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2572, {{16160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2573, {{16161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2574, {{16162, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2575, {{16163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2576, {{16164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2577, {{16165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2578, {{16166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2579, {{16167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2580, {{16168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2581, {{16169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2582, {{16170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2583, {{16171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2584, {{16172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2585, {{16173, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2586, {{16174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2587, {{16175, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2588, {{16176, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2589, {{16177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2590, {{16178, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2591, {{16179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2592, {{16180, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2593, {{16181, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2594, {{16182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2595, {{16183, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2596, {{16184, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2597, {{16185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2598, {{16186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2599, {{16187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2600, {{16188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2601, {{16189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2602, {{16190, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2603, {{16191, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2604, {{16192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2605, {{16193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2606, {{16194, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2607, {{16195, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2608, {{16196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2609, {{16198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2610, {{16199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2611, {{16200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2612, {{16201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2613, {{16202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2614, {{16203, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2615, {{16205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2616, {{16206, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2617, {{16208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2618, {{16209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2619, {{16210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2620, {{16213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2621, {{16214, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2622, {{16215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2623, {{16216, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2624, {{16217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2625, {{16218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2626, {{16219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2627, {{16220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2628, {{16221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2629, {{16222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2630, {{16223, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2631, {{16224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2632, {{16225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2633, {{16226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2634, {{16227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2635, {{16228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2636, {{16229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2637, {{16230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2638, {{16231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2639, {{16232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2640, {{16233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2641, {{16234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2642, {{16235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2643, {{16236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2644, {{16237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2645, {{16238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2646, {{16239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2647, {{16240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2648, {{16241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2649, {{16242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2650, {{16243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2651, {{16244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2652, {{16245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2653, {{16247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2654, {{16248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2655, {{16249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2656, {{16250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2657, {{16251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2658, {{16253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2659, {{16254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2660, {{16255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2661, {{16256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2662, {{16257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2663, {{16258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2664, {{16259, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2665, {{16260, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2666, {{16261, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2667, {{16262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2668, {{16263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2669, {{16264, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2670, {{16265, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2671, {{16266, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2672, {{16267, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2673, {{16268, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2674, {{16269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2675, {{16270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2676, {{16271, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2677, {{16272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2678, {{16273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2679, {{16274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2680, {{16275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2681, {{16276, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2682, {{16277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2683, {{16278, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2684, {{16279, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2685, {{16280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2686, {{16281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2687, {{16282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2688, {{16283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2689, {{16284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2690, {{16285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2691, {{16286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2692, {{16287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2693, {{16288, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2694, {{16289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2695, {{16290, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2696, {{16291, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2697, {{16292, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2698, {{16293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2699, {{16294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2700, {{16295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2701, {{16296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2702, {{16297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2703, {{16298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2704, {{16299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2705, {{16300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2706, {{16301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2707, {{16302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2708, {{16303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2709, {{16304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2710, {{16305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2711, {{16310, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2712, {{16311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2713, {{16312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2714, {{16313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2715, {{16314, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2716, {{16315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2717, {{16316, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2718, {{16317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2719, {{16318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2720, {{16319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{2721, {{16322, 1, 0, 0},{16322, 1, 1, 16}, {0, 0, 0, 0}}},{2721, {{16322, 1, 0, 0},{16322, 1, 1, 16}, {0, 0, 0, 0}}},{2722, {{16322, 1, 0, 0},{16322, 1, 1, 17}, {0, 0, 0, 0}}},{2722, {{16322, 1, 0, 0},{16322, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{2723, {{16323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2724, {{16324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2725, {{16325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2726, {{16326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2727, {{16327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2728, {{16328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2729, {{16329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{2730, {{16330, 1, 0, 0},{16330, 1, 1, 16}, {0, 0, 0, 0}}},{2730, {{16330, 1, 0, 0},{16330, 1, 1, 16}, {0, 0, 0, 0}}},{2731, {{16330, 1, 0, 0},{16330, 1, 1, 17}, {0, 0, 0, 0}}},{2731, {{16330, 1, 0, 0},{16330, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{2732, {{16331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2733, {{16332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2734, {{16333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2735, {{16334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2736, {{16335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2737, {{16337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2738, {{16338, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2739, {{16339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2740, {{16341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2741, {{16342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2742, {{16343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2743, {{16344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2744, {{16345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2745, {{16346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2746, {{16347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2747, {{16349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2748, {{16350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{2749, {{16351, 1, 0, 0},{16351, 1, 1, 17}, {0, 0, 0, 0}}},{2749, {{16351, 1, 0, 0},{16351, 1, 1, 17}, {0, 0, 0, 0}}},{2750, {{16351, 1, 0, 0},{16351, 1, 1, 16}, {0, 0, 0, 0}}},{2750, {{16351, 1, 0, 0},{16351, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2752, {{16353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2753, {{16354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2754, {{16355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2755, {{16356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2756, {{16357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2757, {{16358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2758, {{16359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2759, {{16360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2760, {{16361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2762, {{16362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2763, {{16363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2764, {{16364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2765, {{16365, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2766, {{16366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2767, {{16367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2768, {{16368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2769, {{16369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2770, {{16370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2771, {{16371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2772, {{16372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2773, {{16373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2774, {{16374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2775, {{16375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2776, {{16376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2777, {{16377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2778, {{16378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2779, {{16379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{2780, {{16380, 1, 0, 0},{16380, 1, 1, 16}, {0, 0, 0, 0}}},{2780, {{16380, 1, 0, 0},{16380, 1, 1, 16}, {0, 0, 0, 0}}},{2781, {{16380, 1, 0, 0},{16380, 1, 1, 17}, {0, 0, 0, 0}}},{2781, {{16380, 1, 0, 0},{16380, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{2782, {{16381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2783, {{16382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2784, {{16383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2785, {{16384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2786, {{16385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2787, {{16386, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2788, {{16387, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2789, {{16388, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2790, {{16389, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2791, {{16390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2792, {{16391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2793, {{16392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2794, {{16393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2795, {{16394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2796, {{16395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2797, {{16397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2798, {{16398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2799, {{16399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2800, {{16400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2801, {{16402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2802, {{16403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2803, {{16404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2804, {{16405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2805, {{16406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2806, {{16407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{2807, {{16408, 1, 0, 0},{16408, 1, 1, 16}, {0, 0, 0, 0}}},{2807, {{16408, 1, 0, 0},{16408, 1, 1, 16}, {0, 0, 0, 0}}},{2808, {{16408, 1, 0, 0},{16408, 1, 1, 17}, {0, 0, 0, 0}}},{2808, {{16408, 1, 0, 0},{16408, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{2809, {{16409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2810, {{16410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2811, {{16411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2812, {{16412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2813, {{16414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2814, {{16415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2815, {{16416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2816, {{16417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2817, {{16418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2818, {{16419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2819, {{16420, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2820, {{16423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2821, {{16424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2822, {{16425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2823, {{16426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2824, {{16427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2825, {{16428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2826, {{16429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2827, {{16430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2828, {{16431, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2829, {{16432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2830, {{16433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2831, {{16434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2832, {{16435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2833, {{16436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2834, {{16437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2835, {{16438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2836, {{16439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2837, {{16440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2839, {{16441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2840, {{16442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2841, {{16443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2842, {{16444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2843, {{16445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2844, {{16446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2845, {{16447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2846, {{16448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2847, {{16449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2848, {{16450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2849, {{16451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2850, {{16452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2851, {{16453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2852, {{16454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2854, {{16455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{2855, {{16456, 1, 0, 0},{16456, 1, 1, 17}, {0, 0, 0, 0}}},{2855, {{16456, 1, 0, 0},{16456, 1, 1, 17}, {0, 0, 0, 0}}},{2856, {{16456, 1, 0, 0},{16456, 1, 1, 16}, {0, 0, 0, 0}}},{2856, {{16456, 1, 0, 0},{16456, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{2857, {{16457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2858, {{16458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2859, {{16459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2860, {{16460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2861, {{16461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2862, {{16462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{2863, {{16463, 1, 0, 0},{16463, 1, 1, 16}, {0, 0, 0, 0}}},{2863, {{16463, 1, 0, 0},{16463, 1, 1, 16}, {0, 0, 0, 0}}},{2864, {{16463, 1, 0, 0},{16463, 1, 1, 17}, {0, 0, 0, 0}}},{2864, {{16463, 1, 0, 0},{16463, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{2865, {{16464, 1, 0, 0},{16464, 1, 1, 16}, {0, 0, 0, 0}}},{2865, {{16464, 1, 0, 0},{16464, 1, 1, 16}, {0, 0, 0, 0}}},{2866, {{16464, 1, 0, 0},{16464, 1, 1, 17}, {0, 0, 0, 0}}},{2866, {{16464, 1, 0, 0},{16464, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{2867, {{16465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2868, {{16466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2869, {{16467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2870, {{16468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2872, {{16469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2873, {{16471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2874, {{16472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2875, {{16473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2876, {{16474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2877, {{16475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2878, {{16476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2879, {{16477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2880, {{16478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2881, {{16479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2882, {{16480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2883, {{16482, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2884, {{16483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2885, {{16484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2886, {{16485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2887, {{16486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2888, {{16487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2889, {{16488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2890, {{16489, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2891, {{16491, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2892, {{16492, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2893, {{16493, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2894, {{16494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2895, {{16495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2896, {{16496, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{2897, {{16497, 1, 0, 0},{16497, 1, 1, 16}, {0, 0, 0, 0}}},{2897, {{16497, 1, 0, 0},{16497, 1, 1, 16}, {0, 0, 0, 0}}},{59940, {{16497, 1, 0, 0},{16497, 1, 1, 17}, {0, 0, 0, 0}}},{59940, {{16497, 1, 0, 0},{16497, 1, 1, 17}, {0, 0, 0, 0}}},{59941, {{16497, 1, 0, 0},{16497, 1, 1, 18}, {0, 0, 0, 0}}},{59941, {{16497, 1, 0, 0},{16497, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{2898, {{16498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2899, {{16499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2900, {{16500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2901, {{16501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2902, {{16502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2903, {{16503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2904, {{16504, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2905, {{16505, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2906, {{16506, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2907, {{16507, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2908, {{16508, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2909, {{16509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{2910, {{16510, 1, 0, 0},{16510, 1, 1, 18}, {0, 0, 0, 0}}},{2910, {{16510, 1, 0, 0},{16510, 1, 1, 18}, {0, 0, 0, 0}}},{2911, {{16510, 1, 0, 0},{16510, 1, 1, 17}, {0, 0, 0, 0}}},{2911, {{16510, 1, 0, 0},{16510, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{2912, {{16511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2913, {{16512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2914, {{16513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2915, {{16514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2916, {{16515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2917, {{16516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2918, {{16517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2919, {{16518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2920, {{16519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2921, {{16520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2922, {{16521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2923, {{16522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2924, {{16524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2925, {{16525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2926, {{16526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2927, {{16527, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2928, {{16528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2929, {{16529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2930, {{16530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2931, {{16531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2932, {{16532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2933, {{16533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{2934, {{16534, 1, 0, 0},{16534, 1, 1, 17}, {0, 0, 0, 0}}},{2934, {{16534, 1, 0, 0},{16534, 1, 1, 17}, {0, 0, 0, 0}}},{2935, {{16534, 1, 0, 0},{16534, 1, 1, 18}, {0, 0, 0, 0}}},{2935, {{16534, 1, 0, 0},{16534, 1, 1, 18}, {0, 0, 0, 0}}},{2936, {{16534, 1, 0, 0},{16534, 1, 1, 16}, {0, 0, 0, 0}}},{2936, {{16534, 1, 0, 0},{16534, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{2937, {{16535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2938, {{16536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2939, {{16537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2940, {{16538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2941, {{16539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2942, {{16540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2943, {{16541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2944, {{16542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2945, {{16543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2946, {{16544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2947, {{16545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2948, {{16546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2949, {{16547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2950, {{16548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2951, {{16549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2952, {{16550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2953, {{16551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2954, {{16552, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2955, {{16553, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2956, {{16554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2957, {{16555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2958, {{16556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2959, {{16557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{2960, {{16558, 1, 0, 0},{16558, 1, 1, 16}, {0, 0, 0, 0}}},{2960, {{16558, 1, 0, 0},{16558, 1, 1, 16}, {0, 0, 0, 0}}},{2961, {{16558, 1, 0, 0},{16558, 1, 1, 17}, {0, 0, 0, 0}}},{2961, {{16558, 1, 0, 0},{16558, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{2962, {{16559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2963, {{16560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2964, {{16561, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2965, {{16562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2966, {{16563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2967, {{16564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2968, {{16567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2969, {{16568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2970, {{16569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2971, {{16570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2972, {{16571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2973, {{16572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2974, {{16573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2975, {{16574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2976, {{16575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2977, {{16576, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2978, {{16577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2979, {{16578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2980, {{16584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2981, {{16585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2982, {{16586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2983, {{16587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2984, {{16588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{2985, {{16589, 1, 0, 0},{16589, 1, 1, 17}, {0, 0, 0, 0}}},{2985, {{16589, 1, 0, 0},{16589, 1, 1, 17}, {0, 0, 0, 0}}},{2986, {{16589, 1, 0, 0},{16589, 1, 1, 16}, {0, 0, 0, 0}}},{2986, {{16589, 1, 0, 0},{16589, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{2987, {{16590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2988, {{16591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2989, {{16593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2990, {{16594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2991, {{16595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2992, {{16596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2993, {{16597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2994, {{16598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2995, {{16599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2996, {{16600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2997, {{16601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2998, {{16602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2999, {{16603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3000, {{16604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3001, {{16605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3003, {{16606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3004, {{16609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3005, {{16610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3006, {{16611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3007, {{16612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3008, {{16613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3009, {{16614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3010, {{16615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3011, {{16617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3012, {{16618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3013, {{16619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3014, {{16620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3015, {{16621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3016, {{16622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3018, {{16623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3019, {{16624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3020, {{16625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3021, {{16626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3022, {{16627, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3023, {{16628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3024, {{16629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3025, {{16630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3026, {{16631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3027, {{16632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3028, {{16633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3029, {{16634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3030, {{16635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{3031, {{16637, 1, 0, 0},{16637, 1, 1, 17}, {0, 0, 0, 0}}},{3031, {{16637, 1, 0, 0},{16637, 1, 1, 17}, {0, 0, 0, 0}}},{3032, {{16637, 1, 0, 0},{16637, 1, 1, 16}, {0, 0, 0, 0}}},{3032, {{16637, 1, 0, 0},{16637, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {4, {{3033, {{16638, 1, 0, 0},{16638, 1, 1, 17}, {0, 0, 0, 0}}},{3033, {{16638, 1, 0, 0},{16638, 1, 1, 17}, {0, 0, 0, 0}}},{3034, {{16638, 1, 0, 0},{16638, 1, 1, 16}, {0, 0, 0, 0}}},{3034, {{16638, 1, 0, 0},{16638, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{3035, {{16639, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3036, {{16640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3037, {{16641, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{3038, {{16642, 1, 0, 0},{16642, 1, 1, 16}, {0, 0, 0, 0}}},{3038, {{16642, 1, 0, 0},{16642, 1, 1, 16}, {0, 0, 0, 0}}},{3039, {{16642, 1, 0, 0},{16642, 1, 1, 17}, {0, 0, 0, 0}}},{3039, {{16642, 1, 0, 0},{16642, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{3040, {{16643, 1, 0, 0},{16643, 1, 1, 17}, {0, 0, 0, 0}}},{3040, {{16643, 1, 0, 0},{16643, 1, 1, 17}, {0, 0, 0, 0}}},{3041, {{16643, 1, 0, 0},{16643, 1, 1, 19}, {0, 0, 0, 0}}},{3041, {{16643, 1, 0, 0},{16643, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{3042, {{16644, 1, 0, 0},{16644, 1, 1, 17}, {0, 0, 0, 0}}},{3042, {{16644, 1, 0, 0},{16644, 1, 1, 17}, {0, 0, 0, 0}}},{3043, {{16644, 1, 0, 0},{16644, 1, 1, 16}, {0, 0, 0, 0}}},{3043, {{16644, 1, 0, 0},{16644, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {4, {{3044, {{16645, 1, 0, 0},{16645, 1, 1, 17}, {0, 0, 0, 0}}},{3044, {{16645, 1, 0, 0},{16645, 1, 1, 17}, {0, 0, 0, 0}}},{3045, {{16645, 1, 0, 0},{16645, 1, 1, 18}, {0, 0, 0, 0}}},{3045, {{16645, 1, 0, 0},{16645, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{3046, {{16646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{3048, {{16647, 1, 0, 0},{16647, 1, 1, 18}, {0, 0, 0, 0}}},{3048, {{16647, 1, 0, 0},{16647, 1, 1, 18}, {0, 0, 0, 0}}},{3049, {{16647, 1, 0, 0},{16647, 1, 1, 16}, {0, 0, 0, 0}}},{3049, {{16647, 1, 0, 0},{16647, 1, 1, 16}, {0, 0, 0, 0}}},{3050, {{16647, 1, 0, 0},{16647, 1, 1, 17}, {0, 0, 0, 0}}},{3050, {{16647, 1, 0, 0},{16647, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{3051, {{16648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3052, {{16649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3053, {{16651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3054, {{16652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{3056, {{16653, 1, 0, 0},{16653, 1, 1, 17}, {0, 0, 0, 0}}},{3056, {{16653, 1, 0, 0},{16653, 1, 1, 17}, {0, 0, 0, 0}}},{3057, {{16653, 1, 0, 0},{16653, 1, 1, 16}, {0, 0, 0, 0}}},{3057, {{16653, 1, 0, 0},{16653, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{3058, {{16654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{3059, {{16655, 1, 0, 0},{16655, 1, 1, 17}, {0, 0, 0, 0}}},{3059, {{16655, 1, 0, 0},{16655, 1, 1, 17}, {0, 0, 0, 0}}},{3060, {{16655, 1, 0, 0},{16655, 1, 1, 16}, {0, 0, 0, 0}}},{3060, {{16655, 1, 0, 0},{16655, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{3061, {{16656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3062, {{16657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{3063, {{16658, 1, 0, 0},{16658, 1, 1, 17}, {0, 0, 0, 0}}},{3063, {{16658, 1, 0, 0},{16658, 1, 1, 17}, {0, 0, 0, 0}}},{3064, {{16658, 1, 0, 0},{16658, 1, 1, 16}, {0, 0, 0, 0}}},{3064, {{16658, 1, 0, 0},{16658, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{3065, {{16659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3067, {{16660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3068, {{16661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3070, {{16664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3071, {{16665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3072, {{16666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3073, {{16667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3074, {{16668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3075, {{16669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3076, {{16671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{3077, {{16672, 1, 0, 0},{16672, 1, 1, 17}, {0, 0, 0, 0}}},{3077, {{16672, 1, 0, 0},{16672, 1, 1, 17}, {0, 0, 0, 0}}},{3078, {{16672, 1, 0, 0},{16672, 1, 1, 16}, {0, 0, 0, 0}}},{3078, {{16672, 1, 0, 0},{16672, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{3079, {{16673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3080, {{16674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3081, {{16676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3082, {{16677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3083, {{16678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3084, {{16679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3085, {{16680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3086, {{16681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{3087, {{16682, 1, 0, 0},{16682, 1, 1, 17}, {0, 0, 0, 0}}},{3087, {{16682, 1, 0, 0},{16682, 1, 1, 17}, {0, 0, 0, 0}}},{3088, {{16682, 1, 0, 0},{16682, 1, 1, 16}, {0, 0, 0, 0}}},{3088, {{16682, 1, 0, 0},{16682, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{3089, {{16683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3090, {{16684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3091, {{16685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3092, {{16686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{3093, {{16687, 1, 0, 0},{16687, 1, 1, 16}, {0, 0, 0, 0}}},{3093, {{16687, 1, 0, 0},{16687, 1, 1, 16}, {0, 0, 0, 0}}},{3094, {{16687, 1, 0, 0},{16687, 1, 1, 17}, {0, 0, 0, 0}}},{3094, {{16687, 1, 0, 0},{16687, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{3095, {{16688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3096, {{16689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3097, {{16690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3098, {{16691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3099, {{16694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3100, {{16695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3101, {{16696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3102, {{16698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3103, {{16699, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3104, {{16700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3105, {{16702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3106, {{16703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3107, {{16704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3108, {{16705, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3109, {{16706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3110, {{16707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3111, {{16708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3112, {{16709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{3113, {{16710, 1, 0, 0},{16710, 1, 1, 17}, {0, 0, 0, 0}}},{3113, {{16710, 1, 0, 0},{16710, 1, 1, 17}, {0, 0, 0, 0}}},{3114, {{16710, 1, 0, 0},{16710, 1, 1, 16}, {0, 0, 0, 0}}},{3114, {{16710, 1, 0, 0},{16710, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{3115, {{16711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{3116, {{16712, 1, 0, 0},{16712, 1, 1, 18}, {0, 0, 0, 0}}},{3116, {{16712, 1, 0, 0},{16712, 1, 1, 18}, {0, 0, 0, 0}}},{3117, {{16712, 1, 0, 0},{16712, 1, 1, 17}, {0, 0, 0, 0}}},{3117, {{16712, 1, 0, 0},{16712, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3120, {{16714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3121, {{16715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3122, {{16716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3123, {{16717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3124, {{16718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3125, {{16719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3126, {{16720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3127, {{16721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3128, {{16722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3129, {{16723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3130, {{16724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3131, {{16725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3132, {{16726, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3133, {{16727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3134, {{16728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3135, {{16729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3136, {{16730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3137, {{16731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3138, {{16732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3139, {{16733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3140, {{16734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3141, {{16736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3142, {{16737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3143, {{16738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3144, {{16739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3145, {{16740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{3146, {{16741, 1, 0, 0},{16741, 1, 1, 16}, {0, 0, 0, 0}}},{3146, {{16741, 1, 0, 0},{16741, 1, 1, 16}, {0, 0, 0, 0}}},{3147, {{16741, 1, 0, 0},{16741, 1, 1, 17}, {0, 0, 0, 0}}},{3147, {{16741, 1, 0, 0},{16741, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{3148, {{16742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3149, {{16743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3150, {{16744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3151, {{16745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3152, {{16748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3153, {{16749, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3154, {{16750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3155, {{16751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3156, {{16752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3157, {{16753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3158, {{16754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3159, {{16755, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3160, {{16756, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3161, {{16757, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3162, {{16758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3163, {{16759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3164, {{16760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3165, {{16761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3166, {{16762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3167, {{16763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3168, {{16764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3169, {{16765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3170, {{16766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3172, {{16767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3173, {{16768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3174, {{16769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3176, {{16770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{3177, {{16771, 1, 0, 0},{16771, 1, 1, 17}, {0, 0, 0, 0}}},{3177, {{16771, 1, 0, 0},{16771, 1, 1, 17}, {0, 0, 0, 0}}},{3178, {{16771, 1, 0, 0},{16771, 1, 1, 16}, {0, 0, 0, 0}}},{3178, {{16771, 1, 0, 0},{16771, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{3180, {{16772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3181, {{16773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3182, {{16776, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3183, {{16777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3184, {{16778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30162, {{16779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3185, {{16781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3186, {{16782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3187, {{16783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3188, {{16784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3189, {{16785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{3190, {{16786, 1, 0, 0},{16786, 1, 1, 16}, {0, 0, 0, 0}}},{3190, {{16786, 1, 0, 0},{16786, 1, 1, 16}, {0, 0, 0, 0}}},{3191, {{16786, 1, 0, 0},{16786, 1, 1, 17}, {0, 0, 0, 0}}},{3191, {{16786, 1, 0, 0},{16786, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{3192, {{16787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3193, {{16788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3194, {{16789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3195, {{16790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3197, {{16791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3198, {{16792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3199, {{16793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3200, {{16794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3201, {{16795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3202, {{16796, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3203, {{16797, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3204, {{16798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3205, {{16799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3206, {{16800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3207, {{16801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3208, {{16802, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3209, {{16803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3210, {{16804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3211, {{16805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3212, {{16806, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3213, {{16807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3214, {{16808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3215, {{16809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3216, {{16810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3217, {{16811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3218, {{16812, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3219, {{16813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3220, {{16815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3221, {{16816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3222, {{16817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3223, {{16818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{3224, {{16819, 1, 0, 0},{16819, 1, 1, 16}, {0, 0, 0, 0}}},{3224, {{16819, 1, 0, 0},{16819, 1, 1, 16}, {0, 0, 0, 0}}},{3225, {{16819, 1, 0, 0},{16819, 1, 1, 17}, {0, 0, 0, 0}}},{3225, {{16819, 1, 0, 0},{16819, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{3226, {{16820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3228, {{16821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3229, {{16822, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3230, {{16823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3231, {{16824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3232, {{16825, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3233, {{16826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3234, {{16827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3235, {{16828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3236, {{16829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3237, {{16830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3238, {{16831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3239, {{16832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3240, {{16833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3241, {{16834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3242, {{16835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3243, {{16837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3244, {{16838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3245, {{16839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3246, {{16840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3247, {{16841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3248, {{16842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3249, {{16843, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3250, {{16844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3252, {{16845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3253, {{16846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3254, {{16847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3255, {{16848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3256, {{16849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{3257, {{16850, 1, 0, 0},{16850, 1, 1, 17}, {0, 0, 0, 0}}},{3257, {{16850, 1, 0, 0},{16850, 1, 1, 17}, {0, 0, 0, 0}}},{3258, {{16850, 1, 0, 0},{16850, 1, 1, 16}, {0, 0, 0, 0}}},{3258, {{16850, 1, 0, 0},{16850, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{3259, {{16851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3260, {{16852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3261, {{16853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3262, {{16854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3263, {{16855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3264, {{16856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3265, {{16857, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3266, {{16858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3267, {{16859, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3268, {{16860, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3269, {{16861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3270, {{16862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3271, {{16863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3272, {{16864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3273, {{16865, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3274, {{16866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3275, {{16867, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3276, {{16868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3277, {{16869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3278, {{16870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3279, {{16871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3280, {{16872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3281, {{16873, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3282, {{16874, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3283, {{16875, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3284, {{16876, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3285, {{16877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3286, {{16878, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3287, {{16879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3288, {{16880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{3289, {{16881, 1, 0, 0},{16881, 1, 1, 17}, {0, 0, 0, 0}}},{3289, {{16881, 1, 0, 0},{16881, 1, 1, 17}, {0, 0, 0, 0}}},{3290, {{16881, 1, 0, 0},{16881, 1, 1, 16}, {0, 0, 0, 0}}},{3290, {{16881, 1, 0, 0},{16881, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{3291, {{16882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3292, {{16883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3293, {{16885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3294, {{16886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3295, {{16887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3296, {{16888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3298, {{16889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3299, {{16890, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3300, {{16891, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3301, {{16892, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3302, {{16893, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3303, {{16894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3304, {{16895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3305, {{16896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3306, {{16897, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3307, {{16898, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3308, {{16899, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3309, {{16900, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3310, {{16901, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3311, {{16902, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3312, {{16903, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3313, {{16904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3314, {{16905, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3315, {{16906, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3317, {{16907, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3318, {{16908, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3319, {{16909, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3320, {{16910, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3321, {{16911, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3322, {{16912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3323, {{16913, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3324, {{16914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3325, {{16915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3326, {{16916, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3327, {{16917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3328, {{16918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3329, {{16919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3330, {{16920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3331, {{16921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3332, {{16922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3333, {{16923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3334, {{16924, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3335, {{16925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3336, {{16926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3337, {{16927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3338, {{16928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3339, {{16929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3340, {{16930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3341, {{16931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3342, {{16932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3343, {{16933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3344, {{16934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{3345, {{16935, 1, 0, 0},{16935, 1, 1, 17}, {0, 0, 0, 0}}},{3345, {{16935, 1, 0, 0},{16935, 1, 1, 17}, {0, 0, 0, 0}}},{3346, {{16935, 1, 0, 0},{16935, 1, 1, 16}, {0, 0, 0, 0}}},{3346, {{16935, 1, 0, 0},{16935, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{3347, {{16938, 1, 0, 0},{16938, 1, 1, 16}, {0, 0, 0, 0}}},{3347, {{16938, 1, 0, 0},{16938, 1, 1, 16}, {0, 0, 0, 0}}},{3348, {{16938, 1, 0, 0},{16938, 1, 1, 17}, {0, 0, 0, 0}}},{3348, {{16938, 1, 0, 0},{16938, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{3349, {{16939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3350, {{16940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3351, {{16941, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3352, {{16942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3353, {{16943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3354, {{16944, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3355, {{16945, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3356, {{16946, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3357, {{16947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3358, {{16948, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3360, {{16949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3361, {{16950, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3362, {{16951, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3363, {{16952, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3364, {{16953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3365, {{16954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3366, {{16955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3367, {{16956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3368, {{16957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3369, {{16958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3370, {{16959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3371, {{16960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3373, {{16961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3374, {{16962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3375, {{16963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3376, {{16964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3377, {{16965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3378, {{16966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3379, {{16968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3380, {{16969, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3381, {{16970, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3382, {{16971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3383, {{16972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3384, {{16973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3385, {{16974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3386, {{16975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3387, {{16976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3388, {{16977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3389, {{16978, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3390, {{16979, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3391, {{16980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3392, {{16981, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3393, {{16982, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3394, {{16983, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3395, {{16984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3396, {{16985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3397, {{16986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3398, {{16987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3399, {{16988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3400, {{16989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3401, {{16990, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3402, {{16991, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3403, {{16992, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3404, {{16993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3405, {{16994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3406, {{16995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3407, {{16996, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3408, {{16997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3409, {{16998, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3410, {{16999, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3411, {{17000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3412, {{17001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3413, {{17002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3414, {{17003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3415, {{17004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3416, {{17006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3417, {{17007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3418, {{17008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3419, {{17009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3420, {{17010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3421, {{17011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3422, {{17012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{3424, {{17013, 1, 0, 0},{17013, 1, 1, 16}, {0, 0, 0, 0}}},{3424, {{17013, 1, 0, 0},{17013, 1, 1, 16}, {0, 0, 0, 0}}},{3425, {{17013, 1, 0, 0},{17013, 1, 1, 17}, {0, 0, 0, 0}}},{3425, {{17013, 1, 0, 0},{17013, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{3426, {{17014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3427, {{17015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3428, {{17016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3429, {{17017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3430, {{17018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3431, {{17019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3432, {{17020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3433, {{17021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3435, {{17022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3436, {{17023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3437, {{17024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3438, {{17025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3439, {{17026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3440, {{17027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3441, {{17028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3442, {{17029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3443, {{17030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3444, {{17031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3445, {{17032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3446, {{17033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3447, {{17034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3448, {{17035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3449, {{17037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3450, {{17038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3451, {{17039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3452, {{17040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3453, {{17041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3455, {{17042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3456, {{17043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3457, {{17044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3459, {{17045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3460, {{17046, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3461, {{17047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3462, {{17048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3463, {{17049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3464, {{17050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3465, {{17051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3466, {{17052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3467, {{17053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3468, {{17054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3469, {{17055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3470, {{17056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3471, {{17057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3472, {{17058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3473, {{17059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3475, {{17060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3476, {{17061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3477, {{17062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3478, {{17063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3479, {{17064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3480, {{17065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3481, {{17066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3482, {{17067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3483, {{17068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3484, {{17069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3485, {{17070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3487, {{17071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3488, {{17072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3489, {{17073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3490, {{17074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3491, {{17075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3492, {{17076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3493, {{17077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3494, {{17078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3495, {{17079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{3496, {{17080, 1, 0, 0},{17080, 1, 1, 16}, {0, 0, 0, 0}}},{3496, {{17080, 1, 0, 0},{17080, 1, 1, 16}, {0, 0, 0, 0}}},{3497, {{17080, 1, 0, 0},{17080, 1, 1, 17}, {0, 0, 0, 0}}},{3497, {{17080, 1, 0, 0},{17080, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3499, {{17082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3500, {{17083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3501, {{17084, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3502, {{17085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3503, {{17086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3504, {{17087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3505, {{17088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3506, {{17089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3507, {{17090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3508, {{17091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3509, {{17092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{3510, {{17094, 1, 0, 0},{17094, 1, 1, 17}, {0, 0, 0, 0}}},{3510, {{17094, 1, 0, 0},{17094, 1, 1, 17}, {0, 0, 0, 0}}},{3511, {{17094, 1, 0, 0},{17094, 1, 1, 18}, {0, 0, 0, 0}}},{3511, {{17094, 1, 0, 0},{17094, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{3513, {{17095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3515, {{17096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3516, {{17097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3517, {{17098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3518, {{17099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3519, {{17100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3520, {{17101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3521, {{17102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3522, {{17103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3523, {{17104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3524, {{17105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3525, {{17107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3526, {{17108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3527, {{17109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3528, {{17110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3529, {{17111, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3530, {{17112, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3531, {{17113, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3532, {{17114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3534, {{17115, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3535, {{17116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3536, {{17117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3537, {{17119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3538, {{17120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3539, {{17121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3540, {{17122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{3541, {{17123, 1, 0, 0},{17123, 1, 1, 16}, {0, 0, 0, 0}}},{3541, {{17123, 1, 0, 0},{17123, 1, 1, 16}, {0, 0, 0, 0}}},{3542, {{17123, 1, 0, 0},{17123, 1, 1, 17}, {0, 0, 0, 0}}},{3542, {{17123, 1, 0, 0},{17123, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{3544, {{17124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3545, {{17125, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58305, {{17126, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3546, {{17127, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3547, {{17128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3548, {{17129, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3549, {{17130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3550, {{17131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3551, {{17132, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3552, {{17133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3553, {{17134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3554, {{17135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3555, {{17136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3556, {{17137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58316, {{17138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3557, {{17139, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3558, {{17140, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3559, {{17141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3560, {{17142, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3561, {{17143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3562, {{17144, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3563, {{17145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3564, {{17146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3565, {{17147, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3566, {{17148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3567, {{17149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3568, {{17150, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3569, {{17151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3570, {{17152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{3571, {{17153, 1, 0, 0},{17153, 1, 1, 16}, {0, 0, 0, 0}}},{3571, {{17153, 1, 0, 0},{17153, 1, 1, 16}, {0, 0, 0, 0}}},{3572, {{17153, 1, 0, 0},{17153, 1, 1, 17}, {0, 0, 0, 0}}},{3572, {{17153, 1, 0, 0},{17153, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{3573, {{17154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3574, {{17155, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3575, {{17156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3576, {{17157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3577, {{17158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3578, {{17159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3579, {{17160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3580, {{17161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3581, {{17164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3582, {{17165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3583, {{17166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3584, {{17167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3585, {{17168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3586, {{17169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3587, {{17170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{3588, {{17171, 1, 0, 0},{17171, 1, 1, 18}, {0, 0, 0, 0}}},{3588, {{17171, 1, 0, 0},{17171, 1, 1, 18}, {0, 0, 0, 0}}},{3589, {{17171, 1, 0, 0},{17171, 1, 1, 17}, {0, 0, 0, 0}}},{3589, {{17171, 1, 0, 0},{17171, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{3590, {{17172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3591, {{17173, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3592, {{17174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3593, {{17175, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3594, {{17176, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3595, {{17177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3596, {{17178, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3597, {{17179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3598, {{17180, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3599, {{17181, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3600, {{17182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3601, {{17183, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3602, {{17184, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3603, {{17185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3604, {{17186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3605, {{17187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3606, {{17188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3607, {{17189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3608, {{17190, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3609, {{17191, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3610, {{17192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3611, {{17193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3612, {{17194, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3613, {{17195, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3614, {{17196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3615, {{17197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3616, {{17198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3617, {{17199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3619, {{17200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3620, {{17201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3621, {{17202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3622, {{17203, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{3623, {{17204, 1, 0, 0},{17204, 1, 1, 16}, {0, 0, 0, 0}}},{3623, {{17204, 1, 0, 0},{17204, 1, 1, 16}, {0, 0, 0, 0}}},{3624, {{17204, 1, 0, 0},{17204, 1, 1, 17}, {0, 0, 0, 0}}},{3624, {{17204, 1, 0, 0},{17204, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{3625, {{17205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3626, {{17218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{3627, {{17219, 1, 0, 0},{17219, 1, 1, 17}, {0, 0, 0, 0}}},{3627, {{17219, 1, 0, 0},{17219, 1, 1, 17}, {0, 0, 0, 0}}},{3628, {{17219, 1, 0, 0},{17219, 1, 1, 18}, {0, 0, 0, 0}}},{3628, {{17219, 1, 0, 0},{17219, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{3629, {{17220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3630, {{17221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3631, {{17222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3632, {{17223, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3633, {{17224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3634, {{17225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3635, {{17226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3636, {{17227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3637, {{17228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3638, {{17229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3639, {{17230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3640, {{17231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3641, {{17232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3642, {{17233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3643, {{17235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3644, {{17236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3645, {{17237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3646, {{17238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3647, {{17239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3648, {{17240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{3649, {{17241, 1, 0, 0},{17241, 1, 1, 16}, {0, 0, 0, 0}}},{3649, {{17241, 1, 0, 0},{17241, 1, 1, 16}, {0, 0, 0, 0}}},{3650, {{17241, 1, 0, 0},{17241, 1, 1, 17}, {0, 0, 0, 0}}},{3650, {{17241, 1, 0, 0},{17241, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{3651, {{17242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3652, {{17243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3653, {{17244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3654, {{17245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3655, {{17246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3656, {{17247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3657, {{17249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3658, {{17250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3659, {{17251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3660, {{17252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3661, {{17253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3662, {{17254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3663, {{17255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3664, {{17256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3665, {{17257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3666, {{17258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3667, {{17259, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3668, {{17260, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3669, {{17261, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3670, {{17262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3671, {{17263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3672, {{17264, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3673, {{17265, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3674, {{17266, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3675, {{17267, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3676, {{17268, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3677, {{17269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3678, {{17270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3679, {{17271, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3680, {{17272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3681, {{17273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3682, {{17274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3683, {{17275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3684, {{17276, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3685, {{17277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3686, {{17278, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3687, {{17279, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3688, {{17280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3689, {{17281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3690, {{17282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3691, {{17283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3692, {{17284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3693, {{17285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3694, {{17286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3695, {{17287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3696, {{17288, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3697, {{17289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3698, {{17290, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3699, {{17291, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3700, {{17292, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3701, {{17293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3702, {{17294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3703, {{17295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3704, {{17296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3705, {{17297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3706, {{17298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3707, {{17299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3708, {{17300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3709, {{17301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3711, {{17302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3712, {{17303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3713, {{17304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3714, {{17305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3715, {{17306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3716, {{17308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3717, {{17309, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{68007, {{17310, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {2, {{3718, {{17311, 1, 0, 0}, {0, 0, 0, 0}}},{3719, {{17311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3720, {{17312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3721, {{17313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3722, {{17314, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3723, {{17315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3724, {{17316, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3725, {{17318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3726, {{17319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3727, {{17320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{3729, {{17321, 1, 0, 0},{17321, 1, 1, 17}, {0, 0, 0, 0}}},{3729, {{17321, 1, 0, 0},{17321, 1, 1, 17}, {0, 0, 0, 0}}},{3730, {{17321, 1, 0, 0},{17321, 1, 1, 16}, {0, 0, 0, 0}}},{3730, {{17321, 1, 0, 0},{17321, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{3731, {{17322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3733, {{17323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3734, {{17326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3735, {{17327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3737, {{17328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3738, {{17329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3739, {{17330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3740, {{17331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3741, {{17332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3742, {{17333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3743, {{17334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3744, {{17335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3745, {{17336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3746, {{17337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3747, {{17338, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3748, {{17339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3749, {{17340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3750, {{17341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3751, {{17343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3752, {{17344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3753, {{17345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3754, {{17346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3755, {{17347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3756, {{17348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3757, {{17349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3758, {{17350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3759, {{17351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3760, {{17352, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3761, {{17353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{3762, {{17354, 1, 0, 0},{17354, 1, 1, 16}, {0, 0, 0, 0}}},{3762, {{17354, 1, 0, 0},{17354, 1, 1, 16}, {0, 0, 0, 0}}},{3763, {{17354, 1, 0, 0},{17354, 1, 1, 17}, {0, 0, 0, 0}}},{3763, {{17354, 1, 0, 0},{17354, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{3764, {{17355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3767, {{17356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3768, {{17357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3769, {{17358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3770, {{17359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3771, {{17360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3772, {{17361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3773, {{17362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3774, {{17363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3775, {{17364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{3776, {{17365, 1, 0, 0},{17365, 1, 1, 17}, {0, 0, 0, 0}}},{3776, {{17365, 1, 0, 0},{17365, 1, 1, 17}, {0, 0, 0, 0}}},{3777, {{17365, 1, 0, 0},{17365, 1, 1, 16}, {0, 0, 0, 0}}},{3777, {{17365, 1, 0, 0},{17365, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{3779, {{17366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3780, {{17367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3781, {{17368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{3782, {{17369, 1, 0, 0},{17369, 1, 1, 16}, {0, 0, 0, 0}}},{3782, {{17369, 1, 0, 0},{17369, 1, 1, 16}, {0, 0, 0, 0}}},{3783, {{17369, 1, 0, 0},{17369, 1, 1, 17}, {0, 0, 0, 0}}},{3783, {{17369, 1, 0, 0},{17369, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{3784, {{17370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3785, {{17371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3786, {{17372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3787, {{17374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3788, {{17375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3789, {{17376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3790, {{17377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{3791, {{17378, 1, 0, 0},{17378, 1, 1, 16}, {0, 0, 0, 0}}},{3791, {{17378, 1, 0, 0},{17378, 1, 1, 16}, {0, 0, 0, 0}}},{3792, {{17378, 1, 0, 0},{17378, 1, 1, 17}, {0, 0, 0, 0}}},{3792, {{17378, 1, 0, 0},{17378, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{3793, {{17379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3794, {{17380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3795, {{17381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3796, {{17382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3797, {{17383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3798, {{17384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3799, {{17385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3800, {{17386, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3801, {{17387, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3802, {{17388, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3803, {{17389, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3804, {{17390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3805, {{17391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{3806, {{17392, 1, 0, 0},{17392, 1, 1, 18}, {0, 0, 0, 0}}},{3806, {{17392, 1, 0, 0},{17392, 1, 1, 18}, {0, 0, 0, 0}}},{3807, {{17392, 1, 0, 0},{17392, 1, 1, 17}, {0, 0, 0, 0}}},{3807, {{17392, 1, 0, 0},{17392, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{3808, {{17393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3809, {{17394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3810, {{17395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3811, {{17396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3812, {{17397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3813, {{17398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3814, {{17399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3815, {{17400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3816, {{17401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3817, {{17403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3818, {{17404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3819, {{17405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3820, {{17406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3821, {{17407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3822, {{17408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3823, {{17409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3824, {{17410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3825, {{17411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3826, {{17412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3827, {{17413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3828, {{17414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3829, {{17415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3830, {{17416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3831, {{17417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3832, {{17418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{3833, {{17419, 1, 0, 0},{17419, 1, 1, 16}, {0, 0, 0, 0}}},{3833, {{17419, 1, 0, 0},{17419, 1, 1, 16}, {0, 0, 0, 0}}},{3834, {{17419, 1, 0, 0},{17419, 1, 1, 17}, {0, 0, 0, 0}}},{3834, {{17419, 1, 0, 0},{17419, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{3835, {{17420, 1, 0, 0},{17420, 1, 1, 18}, {0, 0, 0, 0}}},{3835, {{17420, 1, 0, 0},{17420, 1, 1, 18}, {0, 0, 0, 0}}},{3836, {{17420, 1, 0, 0},{17420, 1, 1, 17}, {0, 0, 0, 0}}},{3836, {{17420, 1, 0, 0},{17420, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{3837, {{17421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3838, {{17422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3839, {{17423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3840, {{17424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3841, {{17425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3842, {{17426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3843, {{17427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3844, {{17428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3845, {{17429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3846, {{17430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3847, {{17431, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3848, {{17432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3849, {{17433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3850, {{17434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3851, {{17435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3852, {{17436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3853, {{17437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3854, {{17438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3855, {{17439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3856, {{17440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3857, {{17441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3858, {{17442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3859, {{17443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3860, {{17444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3861, {{17445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3862, {{17446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3863, {{17447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3864, {{17448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3865, {{17449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3866, {{17451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3867, {{17452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3868, {{17453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3869, {{17454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3870, {{17455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3871, {{17456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3872, {{17457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3873, {{17458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3874, {{17459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3875, {{17460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3876, {{17461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3877, {{17462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3878, {{17463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3879, {{17464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3880, {{17465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3881, {{17466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3882, {{17467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3883, {{17468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3884, {{17469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3885, {{17470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3886, {{17471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3887, {{17472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3888, {{17473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3889, {{17474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3890, {{17475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3891, {{17476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3892, {{17477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3893, {{17478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3894, {{17479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3895, {{17480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3896, {{17481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3897, {{17482, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3898, {{17483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3899, {{17484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3900, {{17485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3901, {{17486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3902, {{17487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3903, {{17488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3904, {{17489, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3905, {{17491, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3906, {{17492, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3907, {{17493, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3908, {{17495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3909, {{17496, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3910, {{17497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3911, {{17498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3912, {{17499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3913, {{17500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3914, {{17501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3915, {{17502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3916, {{17503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3917, {{17504, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3918, {{17505, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3919, {{17506, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3920, {{17507, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3921, {{17508, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3922, {{17509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3923, {{17510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3924, {{17511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3925, {{17512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3926, {{17514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{3927, {{17515, 1, 0, 0},{17515, 1, 1, 16}, {0, 0, 0, 0}}},{3927, {{17515, 1, 0, 0},{17515, 1, 1, 16}, {0, 0, 0, 0}}},{3928, {{17515, 1, 0, 0},{17515, 1, 1, 17}, {0, 0, 0, 0}}},{3928, {{17515, 1, 0, 0},{17515, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{3929, {{17516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3930, {{17517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3931, {{17519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3932, {{17520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3933, {{17521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3934, {{17522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3935, {{17523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3936, {{17524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3937, {{17526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3939, {{17529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3940, {{17530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3941, {{17531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3942, {{17532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3943, {{17533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3944, {{17535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3945, {{17536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3946, {{17537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3947, {{17538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3948, {{17539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3949, {{17540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3950, {{17541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3951, {{17542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3952, {{17544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3953, {{17545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3954, {{17546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3955, {{17547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3956, {{17548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3957, {{17549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3958, {{17550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3959, {{17551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3960, {{17552, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3961, {{17553, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3962, {{17554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3963, {{17555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{3965, {{17556, 1, 0, 0},{17556, 1, 1, 16}, {0, 0, 0, 0}}},{3965, {{17556, 1, 0, 0},{17556, 1, 1, 16}, {0, 0, 0, 0}}},{3966, {{17556, 1, 0, 0},{17556, 1, 1, 17}, {0, 0, 0, 0}}},{3966, {{17556, 1, 0, 0},{17556, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{3967, {{17557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3968, {{17558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3969, {{17559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3970, {{17560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3972, {{17562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3973, {{17563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3974, {{17564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3975, {{17565, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3976, {{17566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3977, {{17567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3978, {{17568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3979, {{17569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{3980, {{17570, 1, 0, 0},{17570, 1, 1, 16}, {0, 0, 0, 0}}},{3980, {{17570, 1, 0, 0},{17570, 1, 1, 16}, {0, 0, 0, 0}}},{3981, {{17570, 1, 0, 0},{17570, 1, 1, 17}, {0, 0, 0, 0}}},{3981, {{17570, 1, 0, 0},{17570, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{3982, {{17571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3983, {{17572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3984, {{17573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3985, {{17574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3986, {{17575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3988, {{17576, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{3989, {{17577, 1, 0, 0},{17577, 1, 1, 16}, {0, 0, 0, 0}}},{3989, {{17577, 1, 0, 0},{17577, 1, 1, 16}, {0, 0, 0, 0}}},{3990, {{17577, 1, 0, 0},{17577, 1, 1, 17}, {0, 0, 0, 0}}},{3990, {{17577, 1, 0, 0},{17577, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{3991, {{17578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{3992, {{17579, 1, 0, 0},{17579, 1, 1, 16}, {0, 0, 0, 0}}},{3992, {{17579, 1, 0, 0},{17579, 1, 1, 16}, {0, 0, 0, 0}}},{3993, {{17579, 1, 0, 0},{17579, 1, 1, 17}, {0, 0, 0, 0}}},{3993, {{17579, 1, 0, 0},{17579, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{3994, {{17580, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{3995, {{17581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{3996, {{17584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{3997, {{17585, 1, 0, 0},{17585, 1, 1, 16}, {0, 0, 0, 0}}},{3997, {{17585, 1, 0, 0},{17585, 1, 1, 16}, {0, 0, 0, 0}}},{3998, {{17585, 1, 0, 0},{17585, 1, 1, 17}, {0, 0, 0, 0}}},{3998, {{17585, 1, 0, 0},{17585, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{3999, {{17586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4000, {{17587, 1, 0, 0},{17587, 1, 1, 17}, {0, 0, 0, 0}}},{4000, {{17587, 1, 0, 0},{17587, 1, 1, 17}, {0, 0, 0, 0}}},{4001, {{17587, 1, 0, 0},{17587, 1, 1, 18}, {0, 0, 0, 0}}},{4001, {{17587, 1, 0, 0},{17587, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{4002, {{17588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4003, {{17589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4004, {{17590, 1, 0, 0},{17590, 1, 1, 16}, {0, 0, 0, 0}}},{4004, {{17590, 1, 0, 0},{17590, 1, 1, 16}, {0, 0, 0, 0}}},{4005, {{17590, 1, 0, 0},{17590, 1, 1, 17}, {0, 0, 0, 0}}},{4005, {{17590, 1, 0, 0},{17590, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{4006, {{17591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4007, {{17592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{4008, {{17593, 1, 0, 0},{17593, 1, 1, 17}, {0, 0, 0, 0}}},{4008, {{17593, 1, 0, 0},{17593, 1, 1, 17}, {0, 0, 0, 0}}},{4009, {{17593, 1, 0, 0},{17593, 1, 1, 19}, {0, 0, 0, 0}}},{4009, {{17593, 1, 0, 0},{17593, 1, 1, 19}, {0, 0, 0, 0}}},{4010, {{17593, 1, 0, 0},{17593, 1, 1, 16}, {0, 0, 0, 0}}},{4010, {{17593, 1, 0, 0},{17593, 1, 1, 16}, {0, 0, 0, 0}}},{4011, {{17593, 1, 0, 0},{17593, 1, 1, 18}, {0, 0, 0, 0}}},{4011, {{17593, 1, 0, 0},{17593, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{4013, {{17594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4015, {{17595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4016, {{17596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4017, {{17597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4018, {{17598, 1, 0, 0},{17598, 1, 1, 17}, {0, 0, 0, 0}}},{4018, {{17598, 1, 0, 0},{17598, 1, 1, 17}, {0, 0, 0, 0}}},{4019, {{17598, 1, 0, 0},{17598, 1, 1, 18}, {0, 0, 0, 0}}},{4019, {{17598, 1, 0, 0},{17598, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{4020, {{17601, 1, 0, 0},{17601, 1, 1, 17}, {0, 0, 0, 0}}},{4020, {{17601, 1, 0, 0},{17601, 1, 1, 17}, {0, 0, 0, 0}}},{4021, {{17601, 1, 0, 0},{17601, 1, 1, 16}, {0, 0, 0, 0}}},{4021, {{17601, 1, 0, 0},{17601, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{4022, {{17602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4023, {{17603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4024, {{17604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4025, {{17605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4026, {{17606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4028, {{17607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4030, {{17608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4031, {{17609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4032, {{17610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4033, {{17611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4034, {{17612, 1, 0, 0},{17612, 1, 1, 16}, {0, 0, 0, 0}}},{4034, {{17612, 1, 0, 0},{17612, 1, 1, 16}, {0, 0, 0, 0}}},{4035, {{17612, 1, 0, 0},{17612, 1, 1, 17}, {0, 0, 0, 0}}},{4035, {{17612, 1, 0, 0},{17612, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{4036, {{17613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4037, {{17614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4038, {{17615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4039, {{17616, 1, 0, 0},{17616, 1, 1, 16}, {0, 0, 0, 0}}},{4039, {{17616, 1, 0, 0},{17616, 1, 1, 16}, {0, 0, 0, 0}}},{4040, {{17616, 1, 0, 0},{17616, 1, 1, 17}, {0, 0, 0, 0}}},{4040, {{17616, 1, 0, 0},{17616, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{4041, {{17617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4042, {{17618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4043, {{17619, 1, 0, 0},{17619, 1, 1, 16}, {0, 0, 0, 0}}},{4043, {{17619, 1, 0, 0},{17619, 1, 1, 16}, {0, 0, 0, 0}}},{4044, {{17619, 1, 0, 0},{17619, 1, 1, 17}, {0, 0, 0, 0}}},{4044, {{17619, 1, 0, 0},{17619, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{4045, {{17620, 1, 0, 0},{17620, 1, 1, 17}, {0, 0, 0, 0}}},{4045, {{17620, 1, 0, 0},{17620, 1, 1, 17}, {0, 0, 0, 0}}},{4046, {{17620, 1, 0, 0},{17620, 1, 1, 18}, {0, 0, 0, 0}}},{4046, {{17620, 1, 0, 0},{17620, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4047, {{17623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4048, {{17624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4049, {{17625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4050, {{17626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4051, {{17627, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4052, {{17628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4053, {{17629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4054, {{17630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4055, {{17631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4056, {{17632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4057, {{17633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4059, {{17634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4060, {{17635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4061, {{17636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4063, {{17637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4064, {{17638, 1, 0, 0},{17638, 1, 1, 16}, {0, 0, 0, 0}}},{4064, {{17638, 1, 0, 0},{17638, 1, 1, 16}, {0, 0, 0, 0}}},{4065, {{17638, 1, 0, 0},{17638, 1, 1, 17}, {0, 0, 0, 0}}},{4065, {{17638, 1, 0, 0},{17638, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{4066, {{17639, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58486, {{17642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4067, {{17645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4068, {{17646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4069, {{17647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4070, {{17648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4071, {{17649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4072, {{17650, 1, 0, 0},{17650, 1, 1, 16}, {0, 0, 0, 0}}},{4072, {{17650, 1, 0, 0},{17650, 1, 1, 16}, {0, 0, 0, 0}}},{4073, {{17650, 1, 0, 0},{17650, 1, 1, 17}, {0, 0, 0, 0}}},{4073, {{17650, 1, 0, 0},{17650, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{4075, {{17651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4076, {{17652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4077, {{17653, 1, 0, 0},{17653, 1, 1, 16}, {0, 0, 0, 0}}},{4077, {{17653, 1, 0, 0},{17653, 1, 1, 16}, {0, 0, 0, 0}}},{4078, {{17653, 1, 0, 0},{17653, 1, 1, 17}, {0, 0, 0, 0}}},{4078, {{17653, 1, 0, 0},{17653, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{4079, {{17654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4081, {{17655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4082, {{17656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4083, {{17657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4084, {{17658, 1, 0, 0},{17658, 1, 1, 16}, {0, 0, 0, 0}}},{4084, {{17658, 1, 0, 0},{17658, 1, 1, 16}, {0, 0, 0, 0}}},{4085, {{17658, 1, 0, 0},{17658, 1, 1, 17}, {0, 0, 0, 0}}},{4085, {{17658, 1, 0, 0},{17658, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{4086, {{17659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4087, {{17660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4088, {{17661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4089, {{17662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4090, {{17663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4091, {{17664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4092, {{17665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4093, {{17666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4094, {{17667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4095, {{17668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4096, {{17669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4097, {{17670, 1, 0, 0},{17670, 1, 1, 16}, {0, 0, 0, 0}}},{4097, {{17670, 1, 0, 0},{17670, 1, 1, 16}, {0, 0, 0, 0}}},{4098, {{17670, 1, 0, 0},{17670, 1, 1, 17}, {0, 0, 0, 0}}},{4098, {{17670, 1, 0, 0},{17670, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{4099, {{17671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4100, {{17672, 1, 0, 0},{17672, 1, 1, 17}, {0, 0, 0, 0}}},{4100, {{17672, 1, 0, 0},{17672, 1, 1, 17}, {0, 0, 0, 0}}},{4101, {{17672, 1, 0, 0},{17672, 1, 1, 18}, {0, 0, 0, 0}}},{4101, {{17672, 1, 0, 0},{17672, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{4102, {{17673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4103, {{17674, 1, 0, 0},{17674, 1, 1, 16}, {0, 0, 0, 0}}},{4103, {{17674, 1, 0, 0},{17674, 1, 1, 16}, {0, 0, 0, 0}}},{4104, {{17674, 1, 0, 0},{17674, 1, 1, 17}, {0, 0, 0, 0}}},{4104, {{17674, 1, 0, 0},{17674, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4105, {{17676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4106, {{17677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4108, {{17678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4109, {{17679, 1, 0, 0},{17679, 1, 1, 17}, {0, 0, 0, 0}}},{4109, {{17679, 1, 0, 0},{17679, 1, 1, 17}, {0, 0, 0, 0}}},{4110, {{17679, 1, 0, 0},{17679, 1, 1, 16}, {0, 0, 0, 0}}},{4110, {{17679, 1, 0, 0},{17679, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{4111, {{17680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4112, {{17681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4113, {{17682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4114, {{17683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4115, {{17684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4116, {{17685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4117, {{17686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4118, {{17687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4119, {{17688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4120, {{17689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4121, {{17690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4122, {{17691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4123, {{17692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4124, {{17693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4125, {{17694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4126, {{17695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4127, {{17696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4128, {{17697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4129, {{17698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4130, {{17699, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4131, {{17700, 1, 0, 0},{17700, 1, 1, 16}, {0, 0, 0, 0}}},{4131, {{17700, 1, 0, 0},{17700, 1, 1, 16}, {0, 0, 0, 0}}},{4132, {{17700, 1, 0, 0},{17700, 1, 1, 17}, {0, 0, 0, 0}}},{4132, {{17700, 1, 0, 0},{17700, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{4133, {{17701, 1, 0, 0},{17701, 1, 1, 17}, {0, 0, 0, 0}}},{4133, {{17701, 1, 0, 0},{17701, 1, 1, 17}, {0, 0, 0, 0}}},{4134, {{17701, 1, 0, 0},{17701, 1, 1, 18}, {0, 0, 0, 0}}},{4134, {{17701, 1, 0, 0},{17701, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4135, {{17703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4136, {{17706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4137, {{17708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4138, {{17709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4139, {{17710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4140, {{17711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4141, {{17712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4142, {{17713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4144, {{17714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4146, {{17715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4147, {{17716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4148, {{17717, 1, 0, 0},{17717, 1, 1, 16}, {0, 0, 0, 0}}},{4148, {{17717, 1, 0, 0},{17717, 1, 1, 16}, {0, 0, 0, 0}}},{4149, {{17717, 1, 0, 0},{17717, 1, 1, 17}, {0, 0, 0, 0}}},{4149, {{17717, 1, 0, 0},{17717, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{4150, {{17718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4151, {{17719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4152, {{17720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4153, {{17721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4154, {{17722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4155, {{17723, 1, 0, 0},{17723, 1, 1, 16}, {0, 0, 0, 0}}},{4155, {{17723, 1, 0, 0},{17723, 1, 1, 16}, {0, 0, 0, 0}}},{4156, {{17723, 1, 0, 0},{17723, 1, 1, 17}, {0, 0, 0, 0}}},{4156, {{17723, 1, 0, 0},{17723, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{4157, {{17724, 1, 0, 0},{17724, 1, 1, 16}, {0, 0, 0, 0}}},{4157, {{17724, 1, 0, 0},{17724, 1, 1, 16}, {0, 0, 0, 0}}},{4158, {{17724, 1, 0, 0},{17724, 1, 1, 17}, {0, 0, 0, 0}}},{4158, {{17724, 1, 0, 0},{17724, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{4159, {{17725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4161, {{17726, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4162, {{17727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4163, {{17728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4165, {{17729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{4166, {{17731, 1, 0, 0},{17731, 1, 1, 17}, {0, 0, 0, 0}}},{4166, {{17731, 1, 0, 0},{17731, 1, 1, 17}, {0, 0, 0, 0}}},{4167, {{17731, 1, 0, 0},{17731, 1, 1, 19}, {0, 0, 0, 0}}},{4167, {{17731, 1, 0, 0},{17731, 1, 1, 19}, {0, 0, 0, 0}}},{58538, {{17731, 1, 0, 0},{17731, 1, 1, 18}, {0, 0, 0, 0}}},{58538, {{17731, 1, 0, 0},{17731, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{4168, {{17732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4169, {{17733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4170, {{17734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4171, {{17735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4172, {{17736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4173, {{17737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4174, {{17738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4175, {{17739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{4176, {{17740, 1, 0, 0},{17740, 1, 1, 17}, {0, 0, 0, 0}}},{4176, {{17740, 1, 0, 0},{17740, 1, 1, 17}, {0, 0, 0, 0}}},{4177, {{17740, 1, 0, 0},{17740, 1, 1, 16}, {0, 0, 0, 0}}},{4177, {{17740, 1, 0, 0},{17740, 1, 1, 16}, {0, 0, 0, 0}}},{4178, {{17740, 1, 0, 0},{17740, 1, 1, 18}, {0, 0, 0, 0}}},{4178, {{17740, 1, 0, 0},{17740, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{4179, {{17741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4180, {{17742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4181, {{17743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4183, {{17745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4184, {{17746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4185, {{17747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4186, {{17748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4187, {{17750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4188, {{17751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4189, {{17752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4190, {{17753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4191, {{17754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4192, {{17755, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4195, {{17757, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{4196, {{17758, 1, 0, 0},{17758, 1, 1, 18}, {0, 0, 0, 0}}},{4196, {{17758, 1, 0, 0},{17758, 1, 1, 18}, {0, 0, 0, 0}}},{4197, {{17758, 1, 0, 0},{17758, 1, 1, 16}, {0, 0, 0, 0}}},{4197, {{17758, 1, 0, 0},{17758, 1, 1, 16}, {0, 0, 0, 0}}},{4198, {{17758, 1, 0, 0},{17758, 1, 1, 17}, {0, 0, 0, 0}}},{4198, {{17758, 1, 0, 0},{17758, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{4199, {{17759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4200, {{17760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4201, {{17761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4202, {{17762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4203, {{17763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4204, {{17764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4205, {{17765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4206, {{17766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4207, {{17767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4208, {{17768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4210, {{17769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4211, {{17770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4213, {{17771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4214, {{17772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4215, {{17773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4216, {{17774, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4217, {{17775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4218, {{17776, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4219, {{17777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4220, {{17778, 1, 0, 0},{17778, 1, 1, 16}, {0, 0, 0, 0}}},{4220, {{17778, 1, 0, 0},{17778, 1, 1, 16}, {0, 0, 0, 0}}},{4221, {{17778, 1, 0, 0},{17778, 1, 1, 17}, {0, 0, 0, 0}}},{4221, {{17778, 1, 0, 0},{17778, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{4222, {{17779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4223, {{17780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4224, {{17781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4225, {{17782, 1, 0, 0},{17782, 1, 1, 17}, {0, 0, 0, 0}}},{4225, {{17782, 1, 0, 0},{17782, 1, 1, 17}, {0, 0, 0, 0}}},{4226, {{17782, 1, 0, 0},{17782, 1, 1, 16}, {0, 0, 0, 0}}},{4226, {{17782, 1, 0, 0},{17782, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{4227, {{17783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4228, {{17784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4229, {{17785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4230, {{17786, 1, 0, 0},{17786, 1, 1, 17}, {0, 0, 0, 0}}},{4230, {{17786, 1, 0, 0},{17786, 1, 1, 17}, {0, 0, 0, 0}}},{4231, {{17786, 1, 0, 0},{17786, 1, 1, 18}, {0, 0, 0, 0}}},{4231, {{17786, 1, 0, 0},{17786, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{4232, {{17787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4233, {{17788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4234, {{17789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4235, {{17790, 1, 0, 0},{17790, 1, 1, 16}, {0, 0, 0, 0}}},{4235, {{17790, 1, 0, 0},{17790, 1, 1, 16}, {0, 0, 0, 0}}},{4236, {{17790, 1, 0, 0},{17790, 1, 1, 17}, {0, 0, 0, 0}}},{4236, {{17790, 1, 0, 0},{17790, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{4237, {{17791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4238, {{17792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4239, {{17793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4240, {{17794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4241, {{17795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4242, {{17797, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4243, {{17798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4244, {{17799, 1, 0, 0},{17799, 1, 1, 16}, {0, 0, 0, 0}}},{4244, {{17799, 1, 0, 0},{17799, 1, 1, 16}, {0, 0, 0, 0}}},{4245, {{17799, 1, 0, 0},{17799, 1, 1, 17}, {0, 0, 0, 0}}},{4245, {{17799, 1, 0, 0},{17799, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{4246, {{17800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4247, {{17801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4248, {{17802, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4249, {{17803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4250, {{17804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4251, {{17805, 1, 0, 0},{17805, 1, 1, 16}, {0, 0, 0, 0}}},{4251, {{17805, 1, 0, 0},{17805, 1, 1, 16}, {0, 0, 0, 0}}},{4252, {{17805, 1, 0, 0},{17805, 1, 1, 17}, {0, 0, 0, 0}}},{4252, {{17805, 1, 0, 0},{17805, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{4255, {{17806, 1, 0, 0},{17806, 1, 1, 16}, {0, 0, 0, 0}}},{4255, {{17806, 1, 0, 0},{17806, 1, 1, 16}, {0, 0, 0, 0}}},{4256, {{17806, 1, 0, 0},{17806, 1, 1, 17}, {0, 0, 0, 0}}},{4256, {{17806, 1, 0, 0},{17806, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{4257, {{17807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4258, {{17808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4259, {{17809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4260, {{17810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4261, {{17811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4262, {{17812, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4264, {{17813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4265, {{17814, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4268, {{17815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4269, {{17816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4272, {{17817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4273, {{17818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4274, {{17819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4275, {{17820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4276, {{17821, 1, 0, 0},{17821, 1, 1, 18}, {0, 0, 0, 0}}},{4276, {{17821, 1, 0, 0},{17821, 1, 1, 18}, {0, 0, 0, 0}}},{4277, {{17821, 1, 0, 0},{17821, 1, 1, 17}, {0, 0, 0, 0}}},{4277, {{17821, 1, 0, 0},{17821, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{4278, {{17822, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4279, {{17823, 1, 0, 0},{17823, 1, 1, 16}, {0, 0, 0, 0}}},{4279, {{17823, 1, 0, 0},{17823, 1, 1, 16}, {0, 0, 0, 0}}},{4280, {{17823, 1, 0, 0},{17823, 1, 1, 17}, {0, 0, 0, 0}}},{4280, {{17823, 1, 0, 0},{17823, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{4281, {{17824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4282, {{17825, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4283, {{17826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4284, {{17827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4285, {{17828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4286, {{17829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4287, {{17830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4288, {{17831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4289, {{17832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4290, {{17833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4291, {{17834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4292, {{17835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4293, {{17837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4294, {{17838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4295, {{17839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4296, {{17840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4297, {{17841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4298, {{17842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4299, {{17843, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4300, {{17844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4301, {{17845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4302, {{17846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4303, {{17847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4304, {{17848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4305, {{17849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4306, {{17850, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4307, {{17851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4308, {{17852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4309, {{17853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4310, {{17855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4311, {{17856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4312, {{17857, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4313, {{17858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4314, {{17859, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4315, {{17860, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4316, {{17861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4317, {{17862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4318, {{17863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4319, {{17864, 1, 0, 0},{17864, 1, 1, 16}, {0, 0, 0, 0}}},{4319, {{17864, 1, 0, 0},{17864, 1, 1, 16}, {0, 0, 0, 0}}},{4320, {{17864, 1, 0, 0},{17864, 1, 1, 17}, {0, 0, 0, 0}}},{4320, {{17864, 1, 0, 0},{17864, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{4321, {{17865, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4322, {{17866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4323, {{17867, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4324, {{17868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4325, {{17869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4326, {{17870, 1, 0, 0},{17870, 1, 1, 16}, {0, 0, 0, 0}}},{4326, {{17870, 1, 0, 0},{17870, 1, 1, 16}, {0, 0, 0, 0}}},{4327, {{17870, 1, 0, 0},{17870, 1, 1, 17}, {0, 0, 0, 0}}},{4327, {{17870, 1, 0, 0},{17870, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{4328, {{17871, 1, 0, 0},{17871, 1, 1, 16}, {0, 0, 0, 0}}},{4328, {{17871, 1, 0, 0},{17871, 1, 1, 16}, {0, 0, 0, 0}}},{58641, {{17871, 1, 0, 0},{17871, 1, 1, 17}, {0, 0, 0, 0}}},{58641, {{17871, 1, 0, 0},{17871, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{4329, {{17872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4330, {{17873, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4331, {{17874, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4332, {{17875, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4333, {{17876, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4334, {{17877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{4335, {{17879, 1, 0, 0},{17879, 1, 1, 16}, {0, 0, 0, 0}}},{4335, {{17879, 1, 0, 0},{17879, 1, 1, 16}, {0, 0, 0, 0}}},{4336, {{17879, 1, 0, 0},{17879, 1, 1, 17}, {0, 0, 0, 0}}},{4336, {{17879, 1, 0, 0},{17879, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{4337, {{17880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4338, {{17881, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4339, {{17882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4340, {{17883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4341, {{17884, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4342, {{17885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4343, {{17886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4344, {{17887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4345, {{17888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4346, {{17889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4347, {{17890, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4348, {{17891, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4349, {{17892, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4350, {{17893, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4351, {{17894, 1, 0, 0},{17894, 1, 1, 17}, {0, 0, 0, 0}}},{4351, {{17894, 1, 0, 0},{17894, 1, 1, 17}, {0, 0, 0, 0}}},{4352, {{17894, 1, 0, 0},{17894, 1, 1, 16}, {0, 0, 0, 0}}},{4352, {{17894, 1, 0, 0},{17894, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{4353, {{17895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4354, {{17896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4355, {{17897, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4356, {{17898, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4357, {{17899, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4358, {{17900, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4359, {{17902, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4360, {{17903, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4361, {{17904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4362, {{17905, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4363, {{17906, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4364, {{17907, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4365, {{17909, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4366, {{17910, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4367, {{17911, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4368, {{17912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4369, {{17913, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4370, {{17914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4371, {{17916, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4372, {{17917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4373, {{17918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4374, {{17919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4375, {{17920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4376, {{17921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4377, {{17922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4378, {{17923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4379, {{17924, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4380, {{17925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4381, {{17926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4382, {{17927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4383, {{17929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4384, {{17930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4385, {{17931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4386, {{17932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4387, {{17933, 1, 0, 0},{17933, 1, 1, 16}, {0, 0, 0, 0}}},{4387, {{17933, 1, 0, 0},{17933, 1, 1, 16}, {0, 0, 0, 0}}},{4388, {{17933, 1, 0, 0},{17933, 1, 1, 17}, {0, 0, 0, 0}}},{4388, {{17933, 1, 0, 0},{17933, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{4389, {{17934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4390, {{17935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4391, {{17936, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4392, {{17937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4393, {{17938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4394, {{17939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4395, {{17940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4396, {{17941, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4397, {{17942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4398, {{17944, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4399, {{17945, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4400, {{17946, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4401, {{17948, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4402, {{17949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4403, {{17951, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4404, {{17953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4405, {{17954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4406, {{17955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4407, {{17956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4408, {{17957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4409, {{17958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4410, {{17960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4411, {{17961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4412, {{17962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4413, {{17963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4414, {{17964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4415, {{17965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4416, {{17966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4417, {{17967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4419, {{17968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4420, {{17969, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4421, {{17971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4422, {{17972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4423, {{17973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4424, {{17974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4425, {{17975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4426, {{17976, 1, 0, 0},{17976, 1, 1, 16}, {0, 0, 0, 0}}},{4426, {{17976, 1, 0, 0},{17976, 1, 1, 16}, {0, 0, 0, 0}}},{4427, {{17976, 1, 0, 0},{17976, 1, 1, 17}, {0, 0, 0, 0}}},{4427, {{17976, 1, 0, 0},{17976, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{4428, {{17977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4429, {{17978, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4430, {{17979, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4431, {{17980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4432, {{17981, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4433, {{17982, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4434, {{17983, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4435, {{17984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4436, {{17985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4437, {{17986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4438, {{17987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4439, {{17988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4440, {{17989, 1, 0, 0},{17989, 1, 1, 16}, {0, 0, 0, 0}}},{4440, {{17989, 1, 0, 0},{17989, 1, 1, 16}, {0, 0, 0, 0}}},{58700, {{17989, 1, 0, 0},{17989, 1, 1, 17}, {0, 0, 0, 0}}},{58700, {{17989, 1, 0, 0},{17989, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{4441, {{17990, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4442, {{17991, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4443, {{17992, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4444, {{17993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4445, {{17998, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4446, {{17999, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4447, {{18000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4448, {{18001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4449, {{18002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4450, {{18004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4451, {{18005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4452, {{18006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4453, {{18007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4454, {{18008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4455, {{18009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4456, {{18010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4457, {{18011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4458, {{18012, 1, 0, 0},{18012, 1, 1, 17}, {0, 0, 0, 0}}},{4458, {{18012, 1, 0, 0},{18012, 1, 1, 17}, {0, 0, 0, 0}}},{4459, {{18012, 1, 0, 0},{18012, 1, 1, 16}, {0, 0, 0, 0}}},{4459, {{18012, 1, 0, 0},{18012, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {4, {{4460, {{18013, 1, 0, 0},{18013, 1, 1, 16}, {0, 0, 0, 0}}},{4460, {{18013, 1, 0, 0},{18013, 1, 1, 16}, {0, 0, 0, 0}}},{4461, {{18013, 1, 0, 0},{18013, 1, 1, 17}, {0, 0, 0, 0}}},{4461, {{18013, 1, 0, 0},{18013, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{4462, {{18014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4463, {{18015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4464, {{18016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4465, {{18017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4466, {{18018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4467, {{18019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4468, {{18020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4469, {{18021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4470, {{18022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4471, {{18023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4472, {{18024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4474, {{18026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4475, {{18028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4476, {{18029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4477, {{18030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4478, {{18031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4479, {{18032, 1, 0, 0},{18032, 1, 1, 17}, {0, 0, 0, 0}}},{4479, {{18032, 1, 0, 0},{18032, 1, 1, 17}, {0, 0, 0, 0}}},{4480, {{18032, 1, 0, 0},{18032, 1, 1, 16}, {0, 0, 0, 0}}},{4480, {{18032, 1, 0, 0},{18032, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{4481, {{18033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4482, {{18034, 1, 0, 0},{18034, 1, 1, 16}, {0, 0, 0, 0}}},{4482, {{18034, 1, 0, 0},{18034, 1, 1, 16}, {0, 0, 0, 0}}},{4483, {{18034, 1, 0, 0},{18034, 1, 1, 17}, {0, 0, 0, 0}}},{4483, {{18034, 1, 0, 0},{18034, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{4484, {{18035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{4485, {{18036, 1, 0, 0},{18036, 1, 1, 16}, {0, 0, 0, 0}}},{4485, {{18036, 1, 0, 0},{18036, 1, 1, 16}, {0, 0, 0, 0}}},{4486, {{18036, 1, 0, 0},{18036, 1, 1, 18}, {0, 0, 0, 0}}},{4486, {{18036, 1, 0, 0},{18036, 1, 1, 18}, {0, 0, 0, 0}}},{4487, {{18036, 1, 0, 0},{18036, 1, 1, 17}, {0, 0, 0, 0}}},{4487, {{18036, 1, 0, 0},{18036, 1, 1, 17}, {0, 0, 0, 0}}},{58717, {{18036, 1, 0, 0},{18036, 1, 1, 19}, {0, 0, 0, 0}}},{58717, {{18036, 1, 0, 0},{18036, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{4488, {{18037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4489, {{18039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4490, {{18040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4491, {{18041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4492, {{18042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4493, {{18044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4494, {{18045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4495, {{18046, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4496, {{18047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4497, {{18048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4498, {{18049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4499, {{18050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4500, {{18051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4501, {{18052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4502, {{18053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4503, {{18054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4504, {{18055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4505, {{18056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4506, {{18057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4507, {{18058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4508, {{18059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4509, {{18061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4510, {{18062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4511, {{18063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4512, {{18064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4513, {{18065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4514, {{18066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4515, {{18067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4516, {{18068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4517, {{18069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4518, {{18070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4519, {{18071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4520, {{18072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4521, {{18073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4522, {{18074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4523, {{18075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4524, {{18076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4525, {{18077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4526, {{18078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4527, {{18079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4528, {{18080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4529, {{18081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4530, {{18082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4531, {{18083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4532, {{18085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4533, {{18086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4534, {{18087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4535, {{18088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4536, {{18089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4537, {{18090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4538, {{18091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4539, {{18092, 1, 0, 0},{18092, 1, 1, 16}, {0, 0, 0, 0}}},{4539, {{18092, 1, 0, 0},{18092, 1, 1, 16}, {0, 0, 0, 0}}},{4540, {{18092, 1, 0, 0},{18092, 1, 1, 17}, {0, 0, 0, 0}}},{4540, {{18092, 1, 0, 0},{18092, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{4541, {{18093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4542, {{18094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4543, {{18095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4544, {{18096, 1, 0, 0},{18096, 1, 1, 16}, {0, 0, 0, 0}}},{4544, {{18096, 1, 0, 0},{18096, 1, 1, 16}, {0, 0, 0, 0}}},{4545, {{18096, 1, 0, 0},{18096, 1, 1, 17}, {0, 0, 0, 0}}},{4545, {{18096, 1, 0, 0},{18096, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{4546, {{18097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4547, {{18098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4548, {{18099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4549, {{18100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4550, {{18101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4551, {{18102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4552, {{18103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4553, {{18104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4554, {{18105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4555, {{18106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4556, {{18107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4557, {{18108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4558, {{18109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4560, {{18110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4561, {{18111, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4562, {{18112, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4563, {{18113, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4564, {{18114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4565, {{18115, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4566, {{18116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4567, {{18117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4568, {{18118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4569, {{18119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4570, {{18120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4571, {{18121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4572, {{18122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4573, {{18123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4574, {{18124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4575, {{18125, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4576, {{18126, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4577, {{18127, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4578, {{18128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4579, {{18129, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4580, {{18130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4581, {{18132, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4582, {{18133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4583, {{18134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4584, {{18135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4585, {{18136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4586, {{18137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4587, {{18138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4588, {{18139, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4589, {{18140, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4590, {{18141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4591, {{18142, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4592, {{18143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4593, {{18144, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4594, {{18145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4595, {{18146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4596, {{18147, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4598, {{18148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4599, {{18149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4600, {{18150, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4601, {{18152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4602, {{18153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4603, {{18154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4604, {{18155, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4605, {{18156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4606, {{18157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4607, {{18158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4608, {{18159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4609, {{18160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4610, {{18161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4611, {{18162, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4612, {{18163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4613, {{18164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4614, {{18165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4615, {{18166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4616, {{18167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4617, {{18169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4618, {{18170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4619, {{18171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4620, {{18172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4621, {{18173, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4622, {{18174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4623, {{18175, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4624, {{18176, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4625, {{18177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4626, {{18179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4627, {{18180, 1, 0, 0},{18180, 1, 1, 16}, {0, 0, 0, 0}}},{4627, {{18180, 1, 0, 0},{18180, 1, 1, 16}, {0, 0, 0, 0}}},{4628, {{18180, 1, 0, 0},{18180, 1, 1, 17}, {0, 0, 0, 0}}},{4628, {{18180, 1, 0, 0},{18180, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{4629, {{18181, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4630, {{18182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4631, {{18183, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4632, {{18184, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4633, {{18185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4634, {{18186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4635, {{18187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4636, {{18188, 1, 0, 0},{18188, 1, 1, 17}, {0, 0, 0, 0}}},{4636, {{18188, 1, 0, 0},{18188, 1, 1, 17}, {0, 0, 0, 0}}},{58764, {{18188, 1, 0, 0},{18188, 1, 1, 18}, {0, 0, 0, 0}}},{58764, {{18188, 1, 0, 0},{18188, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{4637, {{18189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4638, {{18190, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4639, {{18191, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4640, {{18192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4641, {{18193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4642, {{18194, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4643, {{18195, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4644, {{18197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4645, {{18199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4646, {{18200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4647, {{18201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4648, {{18202, 1, 0, 0},{18202, 1, 1, 17}, {0, 0, 0, 0}}},{4648, {{18202, 1, 0, 0},{18202, 1, 1, 17}, {0, 0, 0, 0}}},{4649, {{18202, 1, 0, 0},{18202, 1, 1, 16}, {0, 0, 0, 0}}},{4649, {{18202, 1, 0, 0},{18202, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{4650, {{18203, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4651, {{18204, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4652, {{18205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4653, {{18206, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4654, {{18207, 1, 0, 0},{18207, 1, 1, 17}, {0, 0, 0, 0}}},{4654, {{18207, 1, 0, 0},{18207, 1, 1, 17}, {0, 0, 0, 0}}},{4655, {{18207, 1, 0, 0},{18207, 1, 1, 18}, {0, 0, 0, 0}}},{4655, {{18207, 1, 0, 0},{18207, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{4656, {{18208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4657, {{18209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4658, {{18210, 1, 0, 0},{18210, 1, 1, 16}, {0, 0, 0, 0}}},{4658, {{18210, 1, 0, 0},{18210, 1, 1, 16}, {0, 0, 0, 0}}},{4659, {{18210, 1, 0, 0},{18210, 1, 1, 17}, {0, 0, 0, 0}}},{4659, {{18210, 1, 0, 0},{18210, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4660, {{18218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4661, {{18219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4662, {{18220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4663, {{18221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4664, {{18222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4665, {{18223, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4666, {{18224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4667, {{18225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4668, {{18226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{68008, {{18227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4669, {{18228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4670, {{18229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4671, {{18230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4672, {{18231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4673, {{18232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4674, {{18233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4675, {{18234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4676, {{18235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4677, {{18236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4678, {{18237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4679, {{18238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4680, {{18240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4681, {{18241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4682, {{18242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4683, {{18243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4684, {{18244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4685, {{18245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4686, {{18246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4687, {{18247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4688, {{18248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4689, {{18249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4690, {{18250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4691, {{18251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4692, {{18252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4693, {{18253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4694, {{18254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4695, {{18255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4696, {{18256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4697, {{18257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4698, {{18258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4699, {{18259, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4700, {{18260, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4701, {{18261, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4702, {{18262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4703, {{18263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4704, {{18264, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4705, {{18266, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4706, {{18267, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4707, {{18268, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4709, {{18269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4710, {{18270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4711, {{18271, 1, 0, 0},{18271, 1, 1, 16}, {0, 0, 0, 0}}},{4711, {{18271, 1, 0, 0},{18271, 1, 1, 16}, {0, 0, 0, 0}}},{4712, {{18271, 1, 0, 0},{18271, 1, 1, 17}, {0, 0, 0, 0}}},{4712, {{18271, 1, 0, 0},{18271, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{4713, {{18272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4714, {{18273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4715, {{18274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4716, {{18275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4717, {{18276, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4718, {{18277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4719, {{18278, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4720, {{18279, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4721, {{18280, 1, 0, 0},{18280, 1, 1, 16}, {0, 0, 0, 0}}},{4721, {{18280, 1, 0, 0},{18280, 1, 1, 16}, {0, 0, 0, 0}}},{4722, {{18280, 1, 0, 0},{18280, 1, 1, 17}, {0, 0, 0, 0}}},{4722, {{18280, 1, 0, 0},{18280, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{4723, {{18281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4724, {{18282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4725, {{18283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4726, {{18284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4727, {{18285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4728, {{18286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4729, {{18287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4730, {{18288, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4731, {{18289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4732, {{18291, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4733, {{18292, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4734, {{18293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4735, {{18294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4736, {{18295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4737, {{18296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4738, {{18297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4739, {{18298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4740, {{18299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4741, {{18300, 1, 0, 0},{18300, 1, 1, 16}, {0, 0, 0, 0}}},{4741, {{18300, 1, 0, 0},{18300, 1, 1, 16}, {0, 0, 0, 0}}},{4742, {{18300, 1, 0, 0},{18300, 1, 1, 17}, {0, 0, 0, 0}}},{4742, {{18300, 1, 0, 0},{18300, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{4743, {{18301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4744, {{18302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4745, {{18303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4746, {{18304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4747, {{18305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4748, {{18306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4749, {{18307, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4750, {{18308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4751, {{18309, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4752, {{18310, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4753, {{18311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4754, {{18312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4755, {{18315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4756, {{18321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4757, {{18322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4758, {{18323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4759, {{18324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4760, {{18325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4761, {{18326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4762, {{18327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4764, {{18328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4765, {{18329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4766, {{18330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4767, {{18331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4768, {{18332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4769, {{18333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4770, {{18334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4771, {{18335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4772, {{18336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4773, {{18337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4774, {{18338, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4775, {{18339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4776, {{18340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4777, {{18341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4778, {{18342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4779, {{18343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4780, {{18344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4781, {{18345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4782, {{18346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4783, {{18347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4784, {{18348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4785, {{18349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4786, {{18350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4787, {{18351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4788, {{18352, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4789, {{18353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4790, {{18355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4791, {{18356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4792, {{18357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4793, {{18358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4794, {{18359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4795, {{18361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4796, {{18362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4797, {{18363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4798, {{18364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4799, {{18365, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4800, {{18366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4801, {{18367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4802, {{18368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4803, {{18369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4804, {{18370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4805, {{18371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4806, {{18372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4807, {{18373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4808, {{18374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4809, {{18375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4810, {{18376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4811, {{18377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4812, {{18378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4813, {{18379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4814, {{18380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4815, {{18381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4816, {{18382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4817, {{18383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4818, {{18384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4819, {{18385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4820, {{18386, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4821, {{18387, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4822, {{18388, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4823, {{18389, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4824, {{18390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4825, {{18391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4826, {{18392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4827, {{18393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4828, {{18394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4829, {{18395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4830, {{18396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4831, {{18397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4832, {{18398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4833, {{18399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4834, {{18400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4835, {{18401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4836, {{18402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4837, {{18403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4838, {{18404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4839, {{18405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4840, {{18406, 1, 0, 0},{18406, 1, 1, 17}, {0, 0, 0, 0}}},{4840, {{18406, 1, 0, 0},{18406, 1, 1, 17}, {0, 0, 0, 0}}},{58806, {{18406, 1, 0, 0},{18406, 1, 1, 18}, {0, 0, 0, 0}}},{58806, {{18406, 1, 0, 0},{18406, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{4841, {{18407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4842, {{18408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4843, {{18409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4844, {{18410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4845, {{18411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4846, {{18412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4847, {{18413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4848, {{18414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4849, {{18415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4850, {{18416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4851, {{18417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4852, {{18420, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4853, {{18421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4854, {{18422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4855, {{18423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4856, {{18424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4857, {{18425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4858, {{18426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4859, {{18427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4860, {{18428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4861, {{18429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4862, {{18430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4863, {{18431, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4864, {{18432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4865, {{18433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4866, {{18434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4867, {{18435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4868, {{18436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4869, {{18437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4870, {{18438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4871, {{18439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4872, {{18440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4873, {{18441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4874, {{18442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4875, {{18443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4876, {{18445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4877, {{18446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4878, {{18447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4879, {{18448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4880, {{18449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4881, {{18450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4882, {{18451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4883, {{18452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4884, {{18453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4885, {{18454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4886, {{18455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4887, {{18456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4888, {{18457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4889, {{18458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4890, {{18459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4891, {{18460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4892, {{18461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4893, {{18462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4894, {{18463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4895, {{18464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4896, {{18465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4897, {{18466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4898, {{18467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4899, {{18468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4900, {{18469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4901, {{18470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4902, {{18471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4903, {{18472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4904, {{18473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4905, {{18474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4906, {{18475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4907, {{18476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4908, {{18477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4909, {{18478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4910, {{18479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4911, {{18480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4912, {{18481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4913, {{18483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4914, {{18484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4915, {{18485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4916, {{18486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4917, {{18487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4918, {{18488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4919, {{18489, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4920, {{18491, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4921, {{18492, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4922, {{18493, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4923, {{18494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4924, {{18495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4925, {{18496, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4926, {{18497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4927, {{18498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4928, {{18499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4929, {{18500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4930, {{18501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4931, {{18502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4932, {{18503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4933, {{18504, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4934, {{18505, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4935, {{18506, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4936, {{18507, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4937, {{18508, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4938, {{18509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4939, {{18510, 1, 0, 0},{18510, 1, 1, 17}, {0, 0, 0, 0}}},{4939, {{18510, 1, 0, 0},{18510, 1, 1, 17}, {0, 0, 0, 0}}},{4940, {{18510, 1, 0, 0},{18510, 1, 1, 18}, {0, 0, 0, 0}}},{4940, {{18510, 1, 0, 0},{18510, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{4941, {{18511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4942, {{18512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4943, {{18513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4944, {{18514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4945, {{18515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4946, {{18516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4947, {{18517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4948, {{18518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4949, {{18519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4950, {{18520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4951, {{18521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4952, {{18522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4953, {{18523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4954, {{18524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4955, {{18525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4956, {{18526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4957, {{18527, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4958, {{18528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4959, {{18529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4960, {{18530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4961, {{18531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4962, {{18532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4963, {{18533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4964, {{18534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4965, {{18535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4966, {{18536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4967, {{18537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4968, {{18538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4969, {{18539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4970, {{18540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4971, {{18541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4972, {{18542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4973, {{18543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4974, {{18544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4975, {{18545, 1, 0, 0},{18545, 1, 1, 17}, {0, 0, 0, 0}}},{4975, {{18545, 1, 0, 0},{18545, 1, 1, 17}, {0, 0, 0, 0}}},{4976, {{18545, 1, 0, 0},{18545, 1, 1, 16}, {0, 0, 0, 0}}},{4976, {{18545, 1, 0, 0},{18545, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{4977, {{18546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4978, {{18547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4979, {{18548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4980, {{18549, 1, 0, 0},{18549, 1, 1, 17}, {0, 0, 0, 0}}},{4980, {{18549, 1, 0, 0},{18549, 1, 1, 17}, {0, 0, 0, 0}}},{4981, {{18549, 1, 0, 0},{18549, 1, 1, 16}, {0, 0, 0, 0}}},{4981, {{18549, 1, 0, 0},{18549, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{4982, {{18550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4983, {{18551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4984, {{18554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4985, {{18556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4986, {{18557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4987, {{18558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4988, {{18559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4989, {{18563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4990, {{18564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4991, {{18565, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{4992, {{18567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4993, {{18568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4994, {{18569, 1, 0, 0},{18569, 1, 1, 17}, {0, 0, 0, 0}}},{4994, {{18569, 1, 0, 0},{18569, 1, 1, 17}, {0, 0, 0, 0}}},{4995, {{18569, 1, 0, 0},{18569, 1, 1, 16}, {0, 0, 0, 0}}},{4995, {{18569, 1, 0, 0},{18569, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{4996, {{18570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{4997, {{18571, 1, 0, 0},{18571, 1, 1, 16}, {0, 0, 0, 0}}},{4997, {{18571, 1, 0, 0},{18571, 1, 1, 16}, {0, 0, 0, 0}}},{58841, {{18571, 1, 0, 0},{18571, 1, 1, 17}, {0, 0, 0, 0}}},{58841, {{18571, 1, 0, 0},{18571, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{4998, {{18572, 1, 0, 0},{18572, 1, 1, 17}, {0, 0, 0, 0}}},{4998, {{18572, 1, 0, 0},{18572, 1, 1, 17}, {0, 0, 0, 0}}},{4999, {{18572, 1, 0, 0},{18572, 1, 1, 16}, {0, 0, 0, 0}}},{4999, {{18572, 1, 0, 0},{18572, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{5000, {{18573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5001, {{18574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{5002, {{18576, 1, 0, 0},{18576, 1, 1, 17}, {0, 0, 0, 0}}},{5002, {{18576, 1, 0, 0},{18576, 1, 1, 17}, {0, 0, 0, 0}}},{5003, {{18576, 1, 0, 0},{18576, 1, 1, 16}, {0, 0, 0, 0}}},{5003, {{18576, 1, 0, 0},{18576, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{5004, {{18577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5005, {{18578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5006, {{18579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{5008, {{18580, 1, 0, 0},{18580, 1, 1, 17}, {0, 0, 0, 0}}},{5008, {{18580, 1, 0, 0},{18580, 1, 1, 17}, {0, 0, 0, 0}}},{5009, {{18580, 1, 0, 0},{18580, 1, 1, 16}, {0, 0, 0, 0}}},{5009, {{18580, 1, 0, 0},{18580, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{5010, {{18581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5011, {{18582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5013, {{18583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5015, {{18585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5016, {{18586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5017, {{18587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5018, {{18588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5019, {{18589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5020, {{18590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5021, {{18591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5022, {{18592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5023, {{18593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5024, {{18594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5025, {{18595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5026, {{18596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{5029, {{18598, 1, 0, 0},{18598, 1, 1, 17}, {0, 0, 0, 0}}},{5029, {{18598, 1, 0, 0},{18598, 1, 1, 17}, {0, 0, 0, 0}}},{5030, {{18598, 1, 0, 0},{18598, 1, 1, 16}, {0, 0, 0, 0}}},{5030, {{18598, 1, 0, 0},{18598, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{5031, {{18599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5032, {{18600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5033, {{18601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5034, {{18602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{5035, {{18603, 1, 0, 0},{18603, 1, 1, 17}, {0, 0, 0, 0}}},{5035, {{18603, 1, 0, 0},{18603, 1, 1, 17}, {0, 0, 0, 0}}},{5036, {{18603, 1, 0, 0},{18603, 1, 1, 16}, {0, 0, 0, 0}}},{5036, {{18603, 1, 0, 0},{18603, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{5037, {{18604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5038, {{18605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5039, {{18606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5040, {{18607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{5041, {{18608, 1, 0, 0},{18608, 1, 1, 17}, {0, 0, 0, 0}}},{5041, {{18608, 1, 0, 0},{18608, 1, 1, 17}, {0, 0, 0, 0}}},{5042, {{18608, 1, 0, 0},{18608, 1, 1, 16}, {0, 0, 0, 0}}},{5042, {{18608, 1, 0, 0},{18608, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{5043, {{18609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5044, {{18610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5045, {{18611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5046, {{18612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5047, {{18613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5048, {{18614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5049, {{18616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5050, {{18617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5051, {{18619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5052, {{18620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5053, {{18621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5054, {{18622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5055, {{18623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5056, {{18624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5057, {{18625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5058, {{18626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5059, {{18627, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5060, {{18628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5061, {{18629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5062, {{18630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5063, {{18631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5064, {{18632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5065, {{18634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5067, {{18635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5068, {{18636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5069, {{18637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5070, {{18638, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5071, {{18639, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{5072, {{18640, 1, 0, 0},{18640, 1, 1, 16}, {0, 0, 0, 0}}},{5072, {{18640, 1, 0, 0},{18640, 1, 1, 16}, {0, 0, 0, 0}}},{5073, {{18640, 1, 0, 0},{18640, 1, 1, 17}, {0, 0, 0, 0}}},{5073, {{18640, 1, 0, 0},{18640, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{5074, {{18641, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5075, {{18643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5076, {{18644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5077, {{18645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5078, {{18646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5079, {{18648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5080, {{18649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5081, {{18650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5082, {{18651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5083, {{18652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5084, {{18653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5085, {{18654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5086, {{18655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5087, {{18657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5088, {{18658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5089, {{18659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5090, {{18660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5091, {{18661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5092, {{18662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5093, {{18663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5094, {{18664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5095, {{18665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5096, {{18666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5097, {{18667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5098, {{18668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5099, {{18670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5100, {{18671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5101, {{18672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5102, {{18673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5103, {{18674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5104, {{18675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5105, {{18676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5106, {{18677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5107, {{18678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5108, {{18679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5109, {{18680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5110, {{18681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5111, {{18682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5112, {{18683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5113, {{18684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5114, {{18685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5115, {{18686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5116, {{18687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5117, {{18689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5118, {{18690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5119, {{18691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5120, {{18692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5121, {{18693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5122, {{18694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5123, {{18695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5124, {{18696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5125, {{18697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5126, {{18698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5127, {{18699, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5128, {{18701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5129, {{18702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5130, {{18703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5131, {{18704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5132, {{18705, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5133, {{18706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5134, {{18707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5135, {{18708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5136, {{18709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5137, {{18710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5138, {{18711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5139, {{18712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5140, {{18713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5141, {{18714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5142, {{18715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5143, {{18716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5144, {{18717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5145, {{18718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5146, {{18719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5147, {{18720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5148, {{18721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5150, {{18722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5151, {{18723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5152, {{18724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5153, {{18725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5154, {{18727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5155, {{18728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5156, {{18729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5157, {{18731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5158, {{18732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5159, {{18733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5160, {{18734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5161, {{18735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5162, {{18736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5163, {{18737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5164, {{18738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5165, {{18739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5167, {{18740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5168, {{18741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5169, {{18742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5170, {{18743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5171, {{18744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5173, {{18745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5174, {{18746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5175, {{18747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5177, {{18749, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5178, {{18750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{5179, {{18751, 1, 0, 0},{18751, 1, 1, 16}, {0, 0, 0, 0}}},{5179, {{18751, 1, 0, 0},{18751, 1, 1, 16}, {0, 0, 0, 0}}},{5180, {{18751, 1, 0, 0},{18751, 1, 1, 17}, {0, 0, 0, 0}}},{5180, {{18751, 1, 0, 0},{18751, 1, 1, 17}, {0, 0, 0, 0}}},{5181, {{18751, 1, 0, 0},{18751, 1, 1, 18}, {0, 0, 0, 0}}},{5181, {{18751, 1, 0, 0},{18751, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{5182, {{18752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5184, {{18753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5185, {{18754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5186, {{18755, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5188, {{18756, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5189, {{18757, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5190, {{18758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5191, {{18760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5192, {{18761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5193, {{18762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5194, {{18763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5195, {{18764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5196, {{18765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{5197, {{18766, 1, 0, 0},{18766, 1, 1, 17}, {0, 0, 0, 0}}},{5197, {{18766, 1, 0, 0},{18766, 1, 1, 17}, {0, 0, 0, 0}}},{5198, {{18766, 1, 0, 0},{18766, 1, 1, 16}, {0, 0, 0, 0}}},{5198, {{18766, 1, 0, 0},{18766, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{5199, {{18767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5200, {{18768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5201, {{18770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5202, {{18771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5203, {{18772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5204, {{18773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5205, {{18774, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5206, {{18775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5207, {{18776, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5208, {{18777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5209, {{18778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5210, {{18779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5211, {{18781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5212, {{18782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5213, {{18783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5214, {{18784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5215, {{18785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5216, {{18786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5217, {{18787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5218, {{18788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5220, {{18791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5221, {{18792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5222, {{18793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5223, {{18794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5224, {{18795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{5225, {{18796, 1, 0, 0},{18796, 1, 1, 16}, {0, 0, 0, 0}}},{5225, {{18796, 1, 0, 0},{18796, 1, 1, 16}, {0, 0, 0, 0}}},{5226, {{18796, 1, 0, 0},{18796, 1, 1, 17}, {0, 0, 0, 0}}},{5226, {{18796, 1, 0, 0},{18796, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{5227, {{18797, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5228, {{18798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5229, {{18799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5230, {{18800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5231, {{18801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5232, {{18802, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5233, {{18803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5234, {{18805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5235, {{18806, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5237, {{18807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5238, {{18808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5239, {{18823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5240, {{18824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5241, {{18826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5242, {{18827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5243, {{18828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5244, {{18829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5245, {{18830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5246, {{18831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5247, {{18832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5248, {{18833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5249, {{18834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5250, {{18835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5251, {{18836, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{5252, {{18837, 1, 0, 0},{18837, 1, 1, 16}, {0, 0, 0, 0}}},{5252, {{18837, 1, 0, 0},{18837, 1, 1, 16}, {0, 0, 0, 0}}},{5253, {{18837, 1, 0, 0},{18837, 1, 1, 17}, {0, 0, 0, 0}}},{5253, {{18837, 1, 0, 0},{18837, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{5255, {{18838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5256, {{18839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5257, {{18840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5258, {{18841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5259, {{18842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5260, {{18844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5261, {{18845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5262, {{18846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5263, {{18847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5264, {{18848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5265, {{18849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5267, {{18851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5268, {{18852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5269, {{18853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5270, {{18854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5271, {{18855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{5272, {{18856, 1, 0, 0},{18856, 1, 1, 16}, {0, 0, 0, 0}}},{5272, {{18856, 1, 0, 0},{18856, 1, 1, 16}, {0, 0, 0, 0}}},{5273, {{18856, 1, 0, 0},{18856, 1, 1, 17}, {0, 0, 0, 0}}},{5273, {{18856, 1, 0, 0},{18856, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5274, {{18858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5275, {{18859, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5276, {{18860, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5277, {{18861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5278, {{18862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5279, {{18863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5280, {{18864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5281, {{18865, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5282, {{18866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5283, {{18867, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5284, {{18868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5285, {{18873, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5286, {{18874, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5287, {{18875, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5288, {{18876, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5289, {{18877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5290, {{18878, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5291, {{18881, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5292, {{18882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5293, {{18883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5294, {{18884, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5295, {{18885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5296, {{18886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5297, {{18887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5298, {{18889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5299, {{18890, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5300, {{18891, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5301, {{18892, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5302, {{18893, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5303, {{18895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5304, {{18896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5305, {{18897, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5306, {{18898, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5307, {{18899, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5309, {{18900, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5310, {{18901, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5311, {{18902, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5312, {{18903, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5313, {{18904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5314, {{18905, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5315, {{18906, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5316, {{18907, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5317, {{18908, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5318, {{18909, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5319, {{18910, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{58975, {{18911, 1, 0, 0},{18911, 1, 1, 16}, {0, 0, 0, 0}}},{58975, {{18911, 1, 0, 0},{18911, 1, 1, 16}, {0, 0, 0, 0}}},{68096, {{18911, 1, 0, 0},{18911, 1, 1, 17}, {0, 0, 0, 0}}},{68096, {{18911, 1, 0, 0},{18911, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{5320, {{18912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5321, {{18913, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{5322, {{18914, 1, 0, 0},{18914, 1, 1, 16}, {0, 0, 0, 0}}},{5322, {{18914, 1, 0, 0},{18914, 1, 1, 16}, {0, 0, 0, 0}}},{5323, {{18914, 1, 0, 0},{18914, 1, 1, 17}, {0, 0, 0, 0}}},{5323, {{18914, 1, 0, 0},{18914, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{5324, {{18915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5325, {{18916, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5326, {{18917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5327, {{18918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5328, {{18919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5330, {{18920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5331, {{18922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5332, {{18923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5333, {{18924, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5334, {{18926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5335, {{18927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5336, {{18928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5337, {{18929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5338, {{18930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5339, {{18931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5340, {{18932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5341, {{18933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5342, {{18934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5343, {{18935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5344, {{18936, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5345, {{18937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5346, {{18938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5347, {{18939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5348, {{18940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5349, {{18941, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5350, {{18942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5351, {{18943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5352, {{18944, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5353, {{18945, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5354, {{18946, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5355, {{18947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5356, {{18948, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5357, {{18949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5358, {{18950, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5359, {{18951, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5360, {{18952, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5361, {{18953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5362, {{18954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5363, {{18955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5364, {{18956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5365, {{18957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5366, {{18958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5367, {{18959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5368, {{18960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5369, {{18961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5370, {{18962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5371, {{18963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5372, {{18965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5373, {{18966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5374, {{18967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5375, {{18968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5376, {{18970, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5377, {{18971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5378, {{18972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5379, {{18973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5380, {{18974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5381, {{18975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5382, {{18976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5383, {{18977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{5384, {{18978, 1, 0, 0},{18978, 1, 1, 16}, {0, 0, 0, 0}}},{5384, {{18978, 1, 0, 0},{18978, 1, 1, 16}, {0, 0, 0, 0}}},{5385, {{18978, 1, 0, 0},{18978, 1, 1, 17}, {0, 0, 0, 0}}},{5385, {{18978, 1, 0, 0},{18978, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{5386, {{18979, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{5387, {{18980, 1, 0, 0},{18980, 1, 1, 17}, {0, 0, 0, 0}}},{5387, {{18980, 1, 0, 0},{18980, 1, 1, 17}, {0, 0, 0, 0}}},{5388, {{18980, 1, 0, 0},{18980, 1, 1, 16}, {0, 0, 0, 0}}},{5388, {{18980, 1, 0, 0},{18980, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{5389, {{18981, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5390, {{18982, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{5391, {{18984, 1, 0, 0},{18984, 1, 1, 17}, {0, 0, 0, 0}}},{5391, {{18984, 1, 0, 0},{18984, 1, 1, 17}, {0, 0, 0, 0}}},{5392, {{18984, 1, 0, 0},{18984, 1, 1, 16}, {0, 0, 0, 0}}},{5392, {{18984, 1, 0, 0},{18984, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{5393, {{18985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5394, {{18986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5395, {{18987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5396, {{18988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5397, {{18989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5398, {{18990, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5399, {{18991, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5400, {{18992, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5401, {{18993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5402, {{18994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5403, {{18995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5404, {{18996, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5405, {{18997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5406, {{18999, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5407, {{19000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5408, {{19001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5409, {{19002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5410, {{19003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{5411, {{19004, 1, 0, 0},{19004, 1, 1, 16}, {0, 0, 0, 0}}},{5411, {{19004, 1, 0, 0},{19004, 1, 1, 16}, {0, 0, 0, 0}}},{5412, {{19004, 1, 0, 0},{19004, 1, 1, 17}, {0, 0, 0, 0}}},{5412, {{19004, 1, 0, 0},{19004, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{5413, {{19005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5414, {{19006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5415, {{19007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5416, {{19008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5417, {{19009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5418, {{19010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5419, {{19011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5420, {{19013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5421, {{19014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5422, {{19015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5423, {{19016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5424, {{19017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5425, {{19018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5426, {{19019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5427, {{19020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5428, {{19021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5429, {{19023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5430, {{19024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5431, {{19025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5432, {{19026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5433, {{19027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5434, {{19028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5435, {{19029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5436, {{19030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5437, {{19031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5438, {{19032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5439, {{19033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5440, {{19034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5441, {{19035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5442, {{19036, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5443, {{19037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5444, {{19038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5445, {{19039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5446, {{19040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5447, {{19041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5448, {{19042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5449, {{19043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5450, {{19044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5451, {{19045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5452, {{19046, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5453, {{19047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5454, {{19048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5455, {{19049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5456, {{19050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5457, {{19051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5458, {{19052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5459, {{19053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5460, {{19054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5461, {{19055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5462, {{19056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5463, {{19057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5464, {{19058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5465, {{19059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5466, {{19060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5467, {{19061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5468, {{19062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5469, {{19063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5470, {{19064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5471, {{19065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5472, {{19066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5473, {{19067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5474, {{19068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5475, {{19069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5476, {{19070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{5477, {{19071, 1, 0, 0},{19071, 1, 1, 16}, {0, 0, 0, 0}}},{5477, {{19071, 1, 0, 0},{19071, 1, 1, 16}, {0, 0, 0, 0}}},{59013, {{19071, 1, 0, 0},{19071, 1, 1, 17}, {0, 0, 0, 0}}},{59013, {{19071, 1, 0, 0},{19071, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{5478, {{19072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5479, {{19073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5480, {{19074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5481, {{19075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5482, {{19076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5483, {{19077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5484, {{19078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5485, {{19079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5486, {{19080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5487, {{19081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5488, {{19082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5489, {{19083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5490, {{19084, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5491, {{19085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5492, {{19086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5493, {{19087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5494, {{19088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5495, {{19089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5496, {{19090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5497, {{19091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5498, {{19092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5499, {{19093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5500, {{19094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5501, {{19095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5502, {{19096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5503, {{19097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5504, {{19098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5505, {{19099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5506, {{19100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{5507, {{19101, 1, 0, 0},{19101, 1, 1, 16}, {0, 0, 0, 0}}},{5507, {{19101, 1, 0, 0},{19101, 1, 1, 16}, {0, 0, 0, 0}}},{68057, {{19101, 1, 0, 0},{19101, 1, 1, 17}, {0, 0, 0, 0}}},{68057, {{19101, 1, 0, 0},{19101, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{5508, {{19102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5509, {{19104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5510, {{19105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5511, {{19107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5512, {{19108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5513, {{19109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5514, {{19110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{5515, {{19111, 1, 0, 0},{19111, 1, 1, 16}, {0, 0, 0, 0}}},{5515, {{19111, 1, 0, 0},{19111, 1, 1, 16}, {0, 0, 0, 0}}},{5516, {{19111, 1, 0, 0},{19111, 1, 1, 17}, {0, 0, 0, 0}}},{5516, {{19111, 1, 0, 0},{19111, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{5517, {{19112, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5518, {{19113, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{68009, {{19114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{5520, {{19115, 1, 0, 0},{19115, 1, 1, 16}, {0, 0, 0, 0}}},{5520, {{19115, 1, 0, 0},{19115, 1, 1, 16}, {0, 0, 0, 0}}},{5521, {{19115, 1, 0, 0},{19115, 1, 1, 17}, {0, 0, 0, 0}}},{5521, {{19115, 1, 0, 0},{19115, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{5522, {{19116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5523, {{19117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5524, {{19118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5525, {{19119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5526, {{19120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5527, {{19121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5528, {{19122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5529, {{19123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{5530, {{19124, 1, 0, 0},{19124, 1, 1, 16}, {0, 0, 0, 0}}},{5530, {{19124, 1, 0, 0},{19124, 1, 1, 16}, {0, 0, 0, 0}}},{5531, {{19124, 1, 0, 0},{19124, 1, 1, 17}, {0, 0, 0, 0}}},{5531, {{19124, 1, 0, 0},{19124, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{5532, {{19125, 1, 0, 0},{19125, 1, 1, 17}, {0, 0, 0, 0}}},{5532, {{19125, 1, 0, 0},{19125, 1, 1, 17}, {0, 0, 0, 0}}},{5533, {{19125, 1, 0, 0},{19125, 1, 1, 16}, {0, 0, 0, 0}}},{5533, {{19125, 1, 0, 0},{19125, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{5534, {{19126, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5535, {{19127, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5536, {{19128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5537, {{19129, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5538, {{19130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5539, {{19131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5540, {{19132, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5541, {{19133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5542, {{19134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5543, {{19135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5544, {{19136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5545, {{19137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5546, {{19138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5547, {{19139, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5548, {{19140, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5549, {{19141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5550, {{19142, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5551, {{19143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5552, {{19144, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5553, {{19145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5554, {{19146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5555, {{19147, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5556, {{19148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5557, {{19150, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5558, {{19151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5559, {{19152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5560, {{19153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5561, {{19154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5562, {{19155, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5563, {{19156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5564, {{19157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5565, {{19158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5566, {{19159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5567, {{19160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5568, {{19161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5569, {{19162, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5570, {{19163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5571, {{19164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{5572, {{19165, 1, 0, 0},{19165, 1, 1, 16}, {0, 0, 0, 0}}},{5572, {{19165, 1, 0, 0},{19165, 1, 1, 16}, {0, 0, 0, 0}}},{5573, {{19165, 1, 0, 0},{19165, 1, 1, 17}, {0, 0, 0, 0}}},{5573, {{19165, 1, 0, 0},{19165, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{5574, {{19166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5575, {{19167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5576, {{19168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5577, {{19169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5578, {{19170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5579, {{19171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5580, {{19172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5581, {{19173, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5582, {{19174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5583, {{19176, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5584, {{19177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5585, {{19178, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5586, {{19179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5587, {{19180, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5588, {{19181, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5589, {{19182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5590, {{19184, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5591, {{19185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5592, {{19186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5593, {{19187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5594, {{19188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5595, {{19189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5596, {{19190, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5597, {{19191, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5598, {{19192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5599, {{19193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5600, {{19194, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5601, {{19195, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5602, {{19196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5603, {{19197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5604, {{19198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5605, {{19199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5606, {{19200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5607, {{19201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5608, {{19202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5609, {{19203, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5610, {{19204, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5611, {{19205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5612, {{19206, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5613, {{19207, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5614, {{19208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5615, {{19209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5616, {{19210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5617, {{19211, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5618, {{19212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5619, {{19213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5620, {{19214, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5621, {{19215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5622, {{19216, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5623, {{19217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{5624, {{19218, 1, 0, 0},{19218, 1, 1, 16}, {0, 0, 0, 0}}},{5624, {{19218, 1, 0, 0},{19218, 1, 1, 16}, {0, 0, 0, 0}}},{59037, {{19218, 1, 0, 0},{19218, 1, 1, 17}, {0, 0, 0, 0}}},{59037, {{19218, 1, 0, 0},{19218, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{5625, {{19219, 1, 0, 0},{19219, 1, 1, 16}, {0, 0, 0, 0}}},{5625, {{19219, 1, 0, 0},{19219, 1, 1, 16}, {0, 0, 0, 0}}},{5626, {{19219, 1, 0, 0},{19219, 1, 1, 17}, {0, 0, 0, 0}}},{5626, {{19219, 1, 0, 0},{19219, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{5627, {{19220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5628, {{19221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5629, {{19222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5630, {{19223, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5631, {{19224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{5632, {{19225, 1, 0, 0},{19225, 1, 1, 16}, {0, 0, 0, 0}}},{5632, {{19225, 1, 0, 0},{19225, 1, 1, 16}, {0, 0, 0, 0}}},{5633, {{19225, 1, 0, 0},{19225, 1, 1, 17}, {0, 0, 0, 0}}},{5633, {{19225, 1, 0, 0},{19225, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{5634, {{19226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5635, {{19227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5636, {{19228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5637, {{19229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5638, {{19230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{5639, {{19231, 1, 0, 0},{19231, 1, 1, 16}, {0, 0, 0, 0}}},{5639, {{19231, 1, 0, 0},{19231, 1, 1, 16}, {0, 0, 0, 0}}},{5640, {{19231, 1, 0, 0},{19231, 1, 1, 17}, {0, 0, 0, 0}}},{5640, {{19231, 1, 0, 0},{19231, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{5641, {{19232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5642, {{19233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{5643, {{19234, 1, 0, 0},{19234, 1, 1, 17}, {0, 0, 0, 0}}},{5643, {{19234, 1, 0, 0},{19234, 1, 1, 17}, {0, 0, 0, 0}}},{5644, {{19234, 1, 0, 0},{19234, 1, 1, 16}, {0, 0, 0, 0}}},{5644, {{19234, 1, 0, 0},{19234, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{5645, {{19235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5646, {{19236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5647, {{19237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5648, {{19238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5649, {{19239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5650, {{19240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5651, {{19241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{5652, {{19242, 1, 0, 0},{19242, 1, 1, 17}, {0, 0, 0, 0}}},{5652, {{19242, 1, 0, 0},{19242, 1, 1, 17}, {0, 0, 0, 0}}},{5653, {{19242, 1, 0, 0},{19242, 1, 1, 16}, {0, 0, 0, 0}}},{5653, {{19242, 1, 0, 0},{19242, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{5654, {{19243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5655, {{19244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5656, {{19245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{5657, {{19246, 1, 0, 0},{19246, 1, 1, 17}, {0, 0, 0, 0}}},{5657, {{19246, 1, 0, 0},{19246, 1, 1, 17}, {0, 0, 0, 0}}},{5658, {{19246, 1, 0, 0},{19246, 1, 1, 16}, {0, 0, 0, 0}}},{5658, {{19246, 1, 0, 0},{19246, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{5659, {{19247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5660, {{19248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5662, {{19249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5663, {{19250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{5664, {{19251, 1, 0, 0},{19251, 1, 1, 17}, {0, 0, 0, 0}}},{5664, {{19251, 1, 0, 0},{19251, 1, 1, 17}, {0, 0, 0, 0}}},{5665, {{19251, 1, 0, 0},{19251, 1, 1, 16}, {0, 0, 0, 0}}},{5665, {{19251, 1, 0, 0},{19251, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {4, {{5666, {{19252, 1, 0, 0},{19252, 1, 1, 17}, {0, 0, 0, 0}}},{5666, {{19252, 1, 0, 0},{19252, 1, 1, 17}, {0, 0, 0, 0}}},{5667, {{19252, 1, 0, 0},{19252, 1, 1, 16}, {0, 0, 0, 0}}},{5667, {{19252, 1, 0, 0},{19252, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{5668, {{19253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5669, {{19254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5670, {{19255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5671, {{19256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{5672, {{19257, 1, 0, 0},{19257, 1, 1, 17}, {0, 0, 0, 0}}},{5672, {{19257, 1, 0, 0},{19257, 1, 1, 17}, {0, 0, 0, 0}}},{5673, {{19257, 1, 0, 0},{19257, 1, 1, 16}, {0, 0, 0, 0}}},{5673, {{19257, 1, 0, 0},{19257, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{5674, {{19258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{5675, {{19259, 1, 0, 0},{19259, 1, 1, 17}, {0, 0, 0, 0}}},{5675, {{19259, 1, 0, 0},{19259, 1, 1, 17}, {0, 0, 0, 0}}},{5676, {{19259, 1, 0, 0},{19259, 1, 1, 18}, {0, 0, 0, 0}}},{5676, {{19259, 1, 0, 0},{19259, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{5677, {{19260, 1, 0, 0},{19260, 1, 1, 17}, {0, 0, 0, 0}}},{5677, {{19260, 1, 0, 0},{19260, 1, 1, 17}, {0, 0, 0, 0}}},{5678, {{19260, 1, 0, 0},{19260, 1, 1, 16}, {0, 0, 0, 0}}},{5678, {{19260, 1, 0, 0},{19260, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{5679, {{19261, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5680, {{19262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5681, {{19263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{5682, {{19264, 1, 0, 0},{19264, 1, 1, 17}, {0, 0, 0, 0}}},{5682, {{19264, 1, 0, 0},{19264, 1, 1, 17}, {0, 0, 0, 0}}},{5683, {{19264, 1, 0, 0},{19264, 1, 1, 16}, {0, 0, 0, 0}}},{5683, {{19264, 1, 0, 0},{19264, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{5684, {{19265, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5685, {{19266, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{5687, {{19267, 1, 0, 0},{19267, 1, 1, 17}, {0, 0, 0, 0}}},{5687, {{19267, 1, 0, 0},{19267, 1, 1, 17}, {0, 0, 0, 0}}},{5688, {{19267, 1, 0, 0},{19267, 1, 1, 16}, {0, 0, 0, 0}}},{5688, {{19267, 1, 0, 0},{19267, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{5689, {{19268, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{5690, {{19269, 1, 0, 0},{19269, 1, 1, 17}, {0, 0, 0, 0}}},{5690, {{19269, 1, 0, 0},{19269, 1, 1, 17}, {0, 0, 0, 0}}},{5691, {{19269, 1, 0, 0},{19269, 1, 1, 16}, {0, 0, 0, 0}}},{5691, {{19269, 1, 0, 0},{19269, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{5692, {{19270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{5693, {{19271, 1, 0, 0},{19271, 1, 1, 17}, {0, 0, 0, 0}}},{5693, {{19271, 1, 0, 0},{19271, 1, 1, 17}, {0, 0, 0, 0}}},{5694, {{19271, 1, 0, 0},{19271, 1, 1, 16}, {0, 0, 0, 0}}},{5694, {{19271, 1, 0, 0},{19271, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{5695, {{19272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{5696, {{19273, 1, 0, 0},{19273, 1, 1, 17}, {0, 0, 0, 0}}},{5696, {{19273, 1, 0, 0},{19273, 1, 1, 17}, {0, 0, 0, 0}}},{5697, {{19273, 1, 0, 0},{19273, 1, 1, 16}, {0, 0, 0, 0}}},{5697, {{19273, 1, 0, 0},{19273, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{5698, {{19274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{5699, {{19275, 1, 0, 0},{19275, 1, 1, 17}, {0, 0, 0, 0}}},{5699, {{19275, 1, 0, 0},{19275, 1, 1, 17}, {0, 0, 0, 0}}},{5700, {{19275, 1, 0, 0},{19275, 1, 1, 16}, {0, 0, 0, 0}}},{5700, {{19275, 1, 0, 0},{19275, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{5701, {{19276, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5702, {{19277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5703, {{19278, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{5704, {{19280, 1, 0, 0},{19280, 1, 1, 17}, {0, 0, 0, 0}}},{5704, {{19280, 1, 0, 0},{19280, 1, 1, 17}, {0, 0, 0, 0}}},{5705, {{19280, 1, 0, 0},{19280, 1, 1, 16}, {0, 0, 0, 0}}},{5705, {{19280, 1, 0, 0},{19280, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {4, {{5706, {{19281, 1, 0, 0},{19281, 1, 1, 17}, {0, 0, 0, 0}}},{5706, {{19281, 1, 0, 0},{19281, 1, 1, 17}, {0, 0, 0, 0}}},{5707, {{19281, 1, 0, 0},{19281, 1, 1, 16}, {0, 0, 0, 0}}},{5707, {{19281, 1, 0, 0},{19281, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {4, {{5708, {{19282, 1, 0, 0},{19282, 1, 1, 17}, {0, 0, 0, 0}}},{5708, {{19282, 1, 0, 0},{19282, 1, 1, 17}, {0, 0, 0, 0}}},{5709, {{19282, 1, 0, 0},{19282, 1, 1, 16}, {0, 0, 0, 0}}},{5709, {{19282, 1, 0, 0},{19282, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{5710, {{19283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{5711, {{19284, 1, 0, 0},{19284, 1, 1, 17}, {0, 0, 0, 0}}},{5711, {{19284, 1, 0, 0},{19284, 1, 1, 17}, {0, 0, 0, 0}}},{5712, {{19284, 1, 0, 0},{19284, 1, 1, 16}, {0, 0, 0, 0}}},{5712, {{19284, 1, 0, 0},{19284, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{5713, {{19285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5714, {{19286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5715, {{19287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5716, {{19288, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5717, {{19289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5718, {{19290, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5719, {{19291, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5720, {{19293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5721, {{19294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5722, {{19295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5723, {{19296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{5724, {{19297, 1, 0, 0},{19297, 1, 1, 17}, {0, 0, 0, 0}}},{5724, {{19297, 1, 0, 0},{19297, 1, 1, 17}, {0, 0, 0, 0}}},{5725, {{19297, 1, 0, 0},{19297, 1, 1, 16}, {0, 0, 0, 0}}},{5725, {{19297, 1, 0, 0},{19297, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{5726, {{19298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{5727, {{19299, 1, 0, 0},{19299, 1, 1, 17}, {0, 0, 0, 0}}},{5727, {{19299, 1, 0, 0},{19299, 1, 1, 17}, {0, 0, 0, 0}}},{5728, {{19299, 1, 0, 0},{19299, 1, 1, 16}, {0, 0, 0, 0}}},{5728, {{19299, 1, 0, 0},{19299, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{5729, {{19300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5730, {{19301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5731, {{19302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5732, {{19303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{5733, {{19304, 1, 0, 0},{19304, 1, 1, 17}, {0, 0, 0, 0}}},{5733, {{19304, 1, 0, 0},{19304, 1, 1, 17}, {0, 0, 0, 0}}},{5734, {{19304, 1, 0, 0},{19304, 1, 1, 16}, {0, 0, 0, 0}}},{5734, {{19304, 1, 0, 0},{19304, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {4, {{5735, {{19305, 1, 0, 0},{19305, 1, 1, 17}, {0, 0, 0, 0}}},{5735, {{19305, 1, 0, 0},{19305, 1, 1, 17}, {0, 0, 0, 0}}},{5736, {{19305, 1, 0, 0},{19305, 1, 1, 16}, {0, 0, 0, 0}}},{5736, {{19305, 1, 0, 0},{19305, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5737, {{19307, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5738, {{19308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5739, {{19309, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5740, {{19310, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5741, {{19311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5742, {{19312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5743, {{19313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5744, {{19314, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5745, {{19315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5746, {{19316, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5747, {{19317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5748, {{19318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5749, {{19319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5750, {{19320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5751, {{19321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5752, {{19322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5753, {{19323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5754, {{19324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5755, {{19325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5756, {{19327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5757, {{19328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5758, {{19329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5759, {{19330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5760, {{19331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5761, {{19332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5762, {{19334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5763, {{19335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5764, {{19336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5765, {{19337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5766, {{19338, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5767, {{19339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5768, {{19340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5769, {{19341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5770, {{19342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5771, {{19343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5772, {{19344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5773, {{19345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5774, {{19346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5775, {{19347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5776, {{19348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5777, {{19349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5778, {{19350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5779, {{19351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5780, {{19352, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5781, {{19353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5782, {{19354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5784, {{19355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5785, {{19356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5786, {{19357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5787, {{19358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5788, {{19359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5789, {{19360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5790, {{19361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5791, {{19362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5792, {{19363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5793, {{19364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5794, {{19365, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5795, {{19366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5796, {{19367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5797, {{19368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5798, {{19369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5799, {{19370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5800, {{19371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5801, {{19372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5802, {{19373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5803, {{19374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5804, {{19375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5805, {{19376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5806, {{19377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5807, {{19378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5809, {{19379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5810, {{19380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5811, {{19381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5812, {{19382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5813, {{19383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5814, {{19384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5815, {{19385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5816, {{19386, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5817, {{19387, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5818, {{19388, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5819, {{19389, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5820, {{19390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5821, {{19391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5822, {{19392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5823, {{19394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5824, {{19398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5825, {{19399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5826, {{19400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5827, {{19401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5828, {{19402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5829, {{19403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5830, {{19404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5831, {{19405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5832, {{19406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5833, {{19407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5834, {{19408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5835, {{19409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5836, {{19410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5837, {{19411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5838, {{19412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5839, {{19413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5840, {{19414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5841, {{19415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5842, {{19416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5843, {{19417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5844, {{19418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5845, {{19419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5846, {{19420, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5847, {{19421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5848, {{19422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5849, {{19423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5850, {{19424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5851, {{19425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5852, {{19426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5853, {{19427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5854, {{19428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5855, {{19429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5856, {{19430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5857, {{19431, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5858, {{19432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5859, {{19434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5860, {{19435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{5861, {{19436, 1, 0, 0},{19436, 1, 1, 16}, {0, 0, 0, 0}}},{5861, {{19436, 1, 0, 0},{19436, 1, 1, 16}, {0, 0, 0, 0}}},{5862, {{19436, 1, 0, 0},{19436, 1, 1, 17}, {0, 0, 0, 0}}},{5862, {{19436, 1, 0, 0},{19436, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{5863, {{19437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5864, {{19438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5865, {{19439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5866, {{19440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5867, {{19441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5868, {{19442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5869, {{19443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5870, {{19444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5871, {{19445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5872, {{19447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5873, {{19448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5874, {{19449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5875, {{19450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5876, {{19453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5877, {{19454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5878, {{19455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5879, {{19456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5880, {{19457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5881, {{19458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59070, {{19459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5882, {{19460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5883, {{19461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5884, {{19462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5885, {{19463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5886, {{19464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5887, {{19465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5888, {{19466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5889, {{19467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5890, {{19468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5891, {{19469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5892, {{19470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5893, {{19471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5894, {{19472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5895, {{19473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5896, {{19474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5897, {{19475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5898, {{19476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5899, {{19477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5900, {{19478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{5901, {{19479, 1, 0, 0},{19479, 1, 1, 17}, {0, 0, 0, 0}}},{5901, {{19479, 1, 0, 0},{19479, 1, 1, 17}, {0, 0, 0, 0}}},{5902, {{19479, 1, 0, 0},{19479, 1, 1, 18}, {0, 0, 0, 0}}},{5902, {{19479, 1, 0, 0},{19479, 1, 1, 18}, {0, 0, 0, 0}}},{5903, {{19479, 1, 0, 0},{19479, 1, 1, 19}, {0, 0, 0, 0}}},{5903, {{19479, 1, 0, 0},{19479, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{5904, {{19480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5905, {{19481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5906, {{19482, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5907, {{19483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5908, {{19484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5909, {{19485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5910, {{19486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5911, {{19487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5912, {{19488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5913, {{19489, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5914, {{19490, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5915, {{19491, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5916, {{19492, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5917, {{19493, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5918, {{19494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5919, {{19495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5920, {{19496, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5921, {{19497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5922, {{19499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5923, {{19500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5924, {{19501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5925, {{19503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5926, {{19504, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5927, {{19505, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5928, {{19506, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5929, {{19507, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5930, {{19508, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5931, {{19509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5932, {{19511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5933, {{19512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5934, {{19513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5935, {{19514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5936, {{19515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5937, {{19516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5938, {{19517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5939, {{19519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5940, {{19520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5941, {{19521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5942, {{19522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5943, {{19523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5944, {{19525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5945, {{19526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5946, {{19527, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5947, {{19528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5948, {{19529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5949, {{19530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5950, {{19531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5951, {{19532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5952, {{19533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5953, {{19534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5954, {{19535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{5955, {{19536, 1, 0, 0},{19536, 1, 1, 16}, {0, 0, 0, 0}}},{5955, {{19536, 1, 0, 0},{19536, 1, 1, 16}, {0, 0, 0, 0}}},{5956, {{19536, 1, 0, 0},{19536, 1, 1, 17}, {0, 0, 0, 0}}},{5956, {{19536, 1, 0, 0},{19536, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{5957, {{19537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5958, {{19538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5959, {{19539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5960, {{19540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5961, {{19541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5962, {{19542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5963, {{19544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5964, {{19545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5965, {{19546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5966, {{19547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5967, {{19548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5968, {{19549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5969, {{19550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5970, {{19551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5971, {{19552, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5972, {{19553, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5973, {{19554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5974, {{19555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5975, {{19556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5976, {{19557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5977, {{19558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5978, {{19559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5979, {{19560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5980, {{19561, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5981, {{19563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5982, {{19564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5983, {{19565, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5984, {{19566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5985, {{19567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5986, {{19568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5987, {{19569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5988, {{19570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5989, {{19571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5990, {{19573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5991, {{19574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5992, {{19575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5993, {{19576, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5994, {{19577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5995, {{19578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5996, {{19579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5997, {{19580, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5998, {{19581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{5999, {{19582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6000, {{19583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6001, {{19584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6002, {{19585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6003, {{19586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6004, {{19587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6005, {{19588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6006, {{19589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6007, {{19590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{6008, {{19592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6009, {{19593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6010, {{19594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6011, {{19595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6012, {{19596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6013, {{19597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{6014, {{19599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6015, {{19600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6016, {{19601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6017, {{19602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6018, {{19603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6019, {{19604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6020, {{19605, 1, 0, 0},{19605, 1, 1, 17}, {0, 0, 0, 0}}},{6020, {{19605, 1, 0, 0},{19605, 1, 1, 17}, {0, 0, 0, 0}}},{6021, {{19605, 1, 0, 0},{19605, 1, 1, 16}, {0, 0, 0, 0}}},{6021, {{19605, 1, 0, 0},{19605, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{6022, {{19606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6023, {{19607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{6024, {{19609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6025, {{19610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6026, {{19611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6027, {{19612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{6028, {{19621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6029, {{19622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6030, {{19623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6031, {{19624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6032, {{19625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6033, {{19626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{6034, {{19628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6035, {{19629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6036, {{19630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6037, {{19631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6038, {{19632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6039, {{19633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6040, {{19634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6041, {{19635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6042, {{19636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6043, {{19637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{6044, {{19639, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6045, {{19640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6046, {{19641, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6047, {{19642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6048, {{19643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6049, {{19644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6050, {{19645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6051, {{19646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6052, {{19647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6053, {{19648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6054, {{19649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6055, {{19650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6056, {{19651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6057, {{19652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6058, {{19653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6059, {{19654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6060, {{19655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6061, {{19656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6062, {{19657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6063, {{19658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {2, {{6064, {{19659, 1, 0, 0}, {0, 0, 0, 0}}},{6065, {{19659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6066, {{19660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6067, {{19661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6068, {{19662, 1, 0, 0},{19662, 1, 1, 16}, {0, 0, 0, 0}}},{6068, {{19662, 1, 0, 0},{19662, 1, 1, 16}, {0, 0, 0, 0}}},{6069, {{19662, 1, 0, 0},{19662, 1, 1, 17}, {0, 0, 0, 0}}},{6069, {{19662, 1, 0, 0},{19662, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{6070, {{19663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6071, {{19664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6072, {{19665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6073, {{19666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6074, {{19667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6075, {{19668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6076, {{19669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6077, {{19670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6078, {{19671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6079, {{19672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6080, {{19673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6081, {{19674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6082, {{19675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6083, {{19676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6084, {{19677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6085, {{19678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6086, {{19679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6087, {{19680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6088, {{19681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6089, {{19682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6090, {{19683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6091, {{19684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{6092, {{19686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6093, {{19687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6094, {{19688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6095, {{19689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6097, {{19690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6098, {{19691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6099, {{19692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6100, {{19693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6101, {{19694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6102, {{19695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6103, {{19696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6104, {{19697, 1, 0, 0},{19697, 1, 1, 16}, {0, 0, 0, 0}}},{6104, {{19697, 1, 0, 0},{19697, 1, 1, 16}, {0, 0, 0, 0}}},{6105, {{19697, 1, 0, 0},{19697, 1, 1, 17}, {0, 0, 0, 0}}},{6105, {{19697, 1, 0, 0},{19697, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{6106, {{19698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6107, {{19699, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6108, {{19700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6109, {{19701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6110, {{19702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6111, {{19703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6112, {{19704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{6113, {{19706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6114, {{19707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6115, {{19708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6116, {{19709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6117, {{19710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6118, {{19711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6119, {{19712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6120, {{19713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6121, {{19714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6122, {{19715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6123, {{19716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6124, {{19717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6125, {{19718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6126, {{19719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6127, {{19720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6128, {{19721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6129, {{19722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6130, {{19723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6131, {{19724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6132, {{19725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6133, {{19726, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6134, {{19727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6135, {{19728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6136, {{19729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6137, {{19730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{6138, {{19738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6139, {{19739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6140, {{19740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6141, {{19741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6142, {{19742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6143, {{19743, 1, 0, 0},{19743, 1, 1, 17}, {0, 0, 0, 0}}},{6143, {{19743, 1, 0, 0},{19743, 1, 1, 17}, {0, 0, 0, 0}}},{6144, {{19743, 1, 0, 0},{19743, 1, 1, 16}, {0, 0, 0, 0}}},{6144, {{19743, 1, 0, 0},{19743, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{6145, {{19744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6146, {{19745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{6147, {{19747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6148, {{19748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6149, {{19749, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6150, {{19750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6151, {{19751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{6152, {{19753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6153, {{19754, 1, 0, 0},{19754, 1, 1, 17}, {0, 0, 0, 0}}},{6153, {{19754, 1, 0, 0},{19754, 1, 1, 17}, {0, 0, 0, 0}}},{6154, {{19754, 1, 0, 0},{19754, 1, 1, 16}, {0, 0, 0, 0}}},{6154, {{19754, 1, 0, 0},{19754, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{6155, {{19755, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6156, {{19756, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6157, {{19757, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6158, {{19758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6159, {{19759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6160, {{19760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6161, {{19761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6162, {{19762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6163, {{19763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6164, {{19764, 1, 0, 0},{19764, 1, 1, 16}, {0, 0, 0, 0}}},{6164, {{19764, 1, 0, 0},{19764, 1, 1, 16}, {0, 0, 0, 0}}},{6165, {{19764, 1, 0, 0},{19764, 1, 1, 17}, {0, 0, 0, 0}}},{6165, {{19764, 1, 0, 0},{19764, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{6166, {{19765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6167, {{19766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6168, {{19767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6169, {{19768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6170, {{19769, 1, 0, 0},{19769, 1, 1, 16}, {0, 0, 0, 0}}},{6170, {{19769, 1, 0, 0},{19769, 1, 1, 16}, {0, 0, 0, 0}}},{6171, {{19769, 1, 0, 0},{19769, 1, 1, 17}, {0, 0, 0, 0}}},{6171, {{19769, 1, 0, 0},{19769, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{6172, {{19770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6173, {{19771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6174, {{19772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6175, {{19773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6176, {{19774, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6177, {{19775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6178, {{19776, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6179, {{19777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6180, {{19778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6181, {{19779, 1, 0, 0},{19779, 1, 1, 16}, {0, 0, 0, 0}}},{6181, {{19779, 1, 0, 0},{19779, 1, 1, 16}, {0, 0, 0, 0}}},{6182, {{19779, 1, 0, 0},{19779, 1, 1, 17}, {0, 0, 0, 0}}},{6182, {{19779, 1, 0, 0},{19779, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{6183, {{19780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6184, {{19781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6185, {{19782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6186, {{19783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6187, {{19784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6188, {{19785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6189, {{19786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6190, {{19787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6191, {{19788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6192, {{19789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6193, {{19790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6194, {{19791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6195, {{19792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{6196, {{19794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6197, {{19795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6198, {{19796, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6199, {{19797, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6200, {{19798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6201, {{19799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6202, {{19800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6203, {{19801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6204, {{19802, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6205, {{19803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6206, {{19804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6207, {{19805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6208, {{19806, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6209, {{19807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6210, {{19808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6211, {{19809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6212, {{19810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6213, {{19811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6214, {{19812, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6215, {{19813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6216, {{19814, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6217, {{19815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6218, {{19816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6219, {{19817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6220, {{19818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6221, {{19819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6222, {{19820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6223, {{19821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6224, {{19822, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6225, {{19823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6226, {{19824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6227, {{19825, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6228, {{19826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6229, {{19827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6230, {{19828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6231, {{19829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6232, {{19830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6233, {{19831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6234, {{19832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6235, {{19833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6236, {{19834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6237, {{19835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6238, {{19836, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6239, {{19837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6240, {{19838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6241, {{19839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6242, {{19840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6243, {{19841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6244, {{19842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6246, {{19843, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6247, {{19844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6248, {{19845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6249, {{19846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{6250, {{19848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6251, {{19849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6252, {{19850, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6253, {{19851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6254, {{19852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6255, {{19853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6256, {{19854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6257, {{19855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6258, {{19856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6259, {{19857, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6260, {{19858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6261, {{19859, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6262, {{19860, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6263, {{19861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6264, {{19862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6265, {{19863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6266, {{19864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6267, {{19865, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6268, {{19866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6269, {{19867, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6270, {{19868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6271, {{19869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6272, {{19870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6273, {{19871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6274, {{19872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6275, {{19873, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6276, {{19874, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6277, {{19875, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6278, {{19876, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6279, {{19877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6280, {{19878, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6281, {{19879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6282, {{19880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6283, {{19881, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6284, {{19882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6285, {{19883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6286, {{19884, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{6287, {{19887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6288, {{19888, 1, 0, 0},{19888, 1, 1, 18}, {0, 0, 0, 0}}},{6288, {{19888, 1, 0, 0},{19888, 1, 1, 18}, {0, 0, 0, 0}}},{6289, {{19888, 1, 0, 0},{19888, 1, 1, 17}, {0, 0, 0, 0}}},{6289, {{19888, 1, 0, 0},{19888, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{6290, {{19889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{6291, {{19891, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6292, {{19892, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6293, {{19893, 1, 0, 0}, {0, 0, 0, 0}}}}}
};

static const MJMappingSet chunked_mj_mappings_004e00_007e9e[12447] = {
    {1, {{6294, {{19968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6295, {{19969, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6296, {{19970, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6297, {{19971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6298, {{19972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6299, {{19973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{6300, {{19975, 1, 0, 0},{19975, 1, 1, 17}, {0, 0, 0, 0}}},{6300, {{19975, 1, 0, 0},{19975, 1, 1, 17}, {0, 0, 0, 0}}},{6301, {{19975, 1, 0, 0},{19975, 1, 1, 18}, {0, 0, 0, 0}}},{6301, {{19975, 1, 0, 0},{19975, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{6303, {{19976, 1, 0, 0},{19976, 1, 1, 18}, {0, 0, 0, 0}}},{6303, {{19976, 1, 0, 0},{19976, 1, 1, 18}, {0, 0, 0, 0}}},{6304, {{19976, 1, 0, 0},{19976, 1, 1, 19}, {0, 0, 0, 0}}},{6304, {{19976, 1, 0, 0},{19976, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{6305, {{19977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6306, {{19978, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6307, {{19979, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6308, {{19980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6309, {{19981, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{6310, {{19982, 1, 0, 0},{19982, 1, 1, 18}, {0, 0, 0, 0}}},{6310, {{19982, 1, 0, 0},{19982, 1, 1, 18}, {0, 0, 0, 0}}},{6311, {{19982, 1, 0, 0},{19982, 1, 1, 19}, {0, 0, 0, 0}}},{6311, {{19982, 1, 0, 0},{19982, 1, 1, 19}, {0, 0, 0, 0}}},{6312, {{19982, 1, 0, 0},{19982, 1, 1, 20}, {0, 0, 0, 0}}},{6312, {{19982, 1, 0, 0},{19982, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{6313, {{19983, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6314, {{19984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{6315, {{19985, 1, 0, 0},{19985, 1, 1, 17}, {0, 0, 0, 0}}},{6315, {{19985, 1, 0, 0},{19985, 1, 1, 17}, {0, 0, 0, 0}}},{6318, {{19985, 1, 0, 0},{19985, 1, 1, 18}, {0, 0, 0, 0}}},{6318, {{19985, 1, 0, 0},{19985, 1, 1, 18}, {0, 0, 0, 0}}},{6320, {{19985, 1, 0, 0},{19985, 1, 1, 20}, {0, 0, 0, 0}}},{6320, {{19985, 1, 0, 0},{19985, 1, 1, 20}, {0, 0, 0, 0}}},{56824, {{19985, 1, 0, 0},{19985, 1, 1, 19}, {0, 0, 0, 0}}},{56824, {{19985, 1, 0, 0},{19985, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {6, {{6316, {{19986, 1, 0, 0},{19986, 1, 1, 17}, {0, 0, 0, 0}}},{6316, {{19986, 1, 0, 0},{19986, 1, 1, 17}, {0, 0, 0, 0}}},{6317, {{19986, 1, 0, 0},{19986, 1, 1, 19}, {0, 0, 0, 0}}},{6317, {{19986, 1, 0, 0},{19986, 1, 1, 19}, {0, 0, 0, 0}}},{6319, {{19986, 1, 0, 0},{19986, 1, 1, 18}, {0, 0, 0, 0}}},{6319, {{19986, 1, 0, 0},{19986, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{6321, {{19988, 1, 0, 0},{19988, 1, 1, 17}, {0, 0, 0, 0}}},{6321, {{19988, 1, 0, 0},{19988, 1, 1, 17}, {0, 0, 0, 0}}},{56831, {{19988, 1, 0, 0},{19988, 1, 1, 18}, {0, 0, 0, 0}}},{56831, {{19988, 1, 0, 0},{19988, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{6322, {{19989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6323, {{19990, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6324, {{19991, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6325, {{19992, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{6327, {{19993, 1, 0, 0},{19993, 1, 1, 18}, {0, 0, 0, 0}}},{6327, {{19993, 1, 0, 0},{19993, 1, 1, 18}, {0, 0, 0, 0}}},{6328, {{19993, 1, 0, 0},{19993, 1, 1, 19}, {0, 0, 0, 0}}},{6328, {{19993, 1, 0, 0},{19993, 1, 1, 19}, {0, 0, 0, 0}}},{6329, {{19993, 1, 0, 0},{19993, 1, 1, 20}, {0, 0, 0, 0}}},{6329, {{19993, 1, 0, 0},{19993, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{6330, {{19994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{6332, {{19998, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6334, {{19999, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6335, {{20000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6336, {{20001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6337, {{20002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6338, {{20003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6339, {{20004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{6340, {{20006, 1, 0, 0},{20006, 1, 1, 19}, {0, 0, 0, 0}}},{6340, {{20006, 1, 0, 0},{20006, 1, 1, 19}, {0, 0, 0, 0}}},{6341, {{20006, 1, 0, 0},{20006, 1, 1, 18}, {0, 0, 0, 0}}},{6341, {{20006, 1, 0, 0},{20006, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{6342, {{20008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6343, {{20009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6344, {{20010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6345, {{20011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6346, {{20012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6347, {{20013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6348, {{20014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6349, {{20015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6350, {{20016, 1, 0, 0},{20016, 1, 1, 18}, {0, 0, 0, 0}}},{6350, {{20016, 1, 0, 0},{20016, 1, 1, 18}, {0, 0, 0, 0}}},{6351, {{20016, 1, 0, 0},{20016, 1, 1, 19}, {0, 0, 0, 0}}},{6351, {{20016, 1, 0, 0},{20016, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{6352, {{20017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6353, {{20018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6354, {{20019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{6355, {{20021, 1, 0, 0},{20021, 1, 1, 17}, {0, 0, 0, 0}}},{6355, {{20021, 1, 0, 0},{20021, 1, 1, 17}, {0, 0, 0, 0}}},{6356, {{20021, 1, 0, 0},{20021, 1, 1, 18}, {0, 0, 0, 0}}},{6356, {{20021, 1, 0, 0},{20021, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{6357, {{20022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{6358, {{20024, 1, 0, 0},{20024, 1, 1, 17}, {0, 0, 0, 0}}},{6358, {{20024, 1, 0, 0},{20024, 1, 1, 17}, {0, 0, 0, 0}}},{6359, {{20024, 1, 0, 0},{20024, 1, 1, 18}, {0, 0, 0, 0}}},{6359, {{20024, 1, 0, 0},{20024, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{6360, {{20025, 1, 0, 0},{20025, 1, 1, 18}, {0, 0, 0, 0}}},{6360, {{20025, 1, 0, 0},{20025, 1, 1, 18}, {0, 0, 0, 0}}},{56826, {{20025, 1, 0, 0},{20025, 1, 1, 19}, {0, 0, 0, 0}}},{56826, {{20025, 1, 0, 0},{20025, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{56827, {{20026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6362, {{20027, 1, 0, 0},{20027, 1, 1, 18}, {0, 0, 0, 0}}},{6362, {{20027, 1, 0, 0},{20027, 1, 1, 18}, {0, 0, 0, 0}}},{6363, {{20027, 1, 0, 0},{20027, 1, 1, 19}, {0, 0, 0, 0}}},{6363, {{20027, 1, 0, 0},{20027, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{6364, {{20028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6365, {{20029, 1, 0, 0},{20029, 1, 1, 16}, {0, 0, 0, 0}}},{6365, {{20029, 1, 0, 0},{20029, 1, 1, 16}, {0, 0, 0, 0}}},{6366, {{20029, 1, 0, 0},{20029, 1, 1, 17}, {0, 0, 0, 0}}},{6366, {{20029, 1, 0, 0},{20029, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{6368, {{20031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6369, {{20032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6370, {{20033, 1, 0, 0},{20033, 1, 1, 17}, {0, 0, 0, 0}}},{6370, {{20033, 1, 0, 0},{20033, 1, 1, 17}, {0, 0, 0, 0}}},{6371, {{20033, 1, 0, 0},{20033, 1, 1, 18}, {0, 0, 0, 0}}},{6371, {{20033, 1, 0, 0},{20033, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{6372, {{20034, 1, 0, 0},{20034, 1, 1, 17}, {0, 0, 0, 0}}},{6372, {{20034, 1, 0, 0},{20034, 1, 1, 17}, {0, 0, 0, 0}}},{6373, {{20034, 1, 0, 0},{20034, 1, 1, 18}, {0, 0, 0, 0}}},{6373, {{20034, 1, 0, 0},{20034, 1, 1, 18}, {0, 0, 0, 0}}},{6374, {{20034, 1, 0, 0},{20034, 1, 1, 19}, {0, 0, 0, 0}}},{6374, {{20034, 1, 0, 0},{20034, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{6375, {{20035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6376, {{20036, 1, 0, 0},{20036, 1, 1, 17}, {0, 0, 0, 0}}},{6376, {{20036, 1, 0, 0},{20036, 1, 1, 17}, {0, 0, 0, 0}}},{6377, {{20036, 1, 0, 0},{20036, 1, 1, 18}, {0, 0, 0, 0}}},{6377, {{20036, 1, 0, 0},{20036, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{6378, {{20037, 1, 0, 0},{20037, 1, 1, 17}, {0, 0, 0, 0}}},{6378, {{20037, 1, 0, 0},{20037, 1, 1, 17}, {0, 0, 0, 0}}},{6379, {{20037, 1, 0, 0},{20037, 1, 1, 18}, {0, 0, 0, 0}}},{6379, {{20037, 1, 0, 0},{20037, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{6380, {{20038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6381, {{20039, 1, 0, 0},{20039, 1, 1, 17}, {0, 0, 0, 0}}},{6381, {{20039, 1, 0, 0},{20039, 1, 1, 17}, {0, 0, 0, 0}}},{56858, {{20039, 1, 0, 0},{20039, 1, 1, 18}, {0, 0, 0, 0}}},{56858, {{20039, 1, 0, 0},{20039, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{6382, {{20040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6383, {{20041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{6384, {{20043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{6385, {{20045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6386, {{20046, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6387, {{20047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{6389, {{20049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6390, {{20050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6391, {{20051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6392, {{20052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{6394, {{20053, 1, 0, 0},{20053, 1, 1, 18}, {0, 0, 0, 0}}},{6394, {{20053, 1, 0, 0},{20053, 1, 1, 18}, {0, 0, 0, 0}}},{6395, {{20053, 1, 0, 0},{20053, 1, 1, 17}, {0, 0, 0, 0}}},{6395, {{20053, 1, 0, 0},{20053, 1, 1, 17}, {0, 0, 0, 0}}},{6396, {{20053, 1, 0, 0},{20053, 1, 1, 22}, {0, 0, 0, 0}}},{6396, {{20053, 1, 0, 0},{20053, 1, 1, 22}, {0, 0, 0, 0}}},{59544, {{20053, 1, 0, 0},{20053, 1, 1, 19}, {0, 0, 0, 0}}},{59544, {{20053, 1, 0, 0},{20053, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{6397, {{20054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6398, {{20055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6399, {{20056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6401, {{20057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6402, {{20058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{6403, {{20060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6404, {{20061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6405, {{20062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6406, {{20063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{6407, {{20066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6408, {{20067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{68010, {{20068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{68011, {{20071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6410, {{20072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6411, {{20073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6412, {{20074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{68012, {{20075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{68013, {{20076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{68014, {{20077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{68015, {{20078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{68016, {{20079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{6415, {{20081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{68017, {{20082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6416, {{20083, 1, 0, 0},{20083, 1, 1, 19}, {0, 0, 0, 0}}},{6416, {{20083, 1, 0, 0},{20083, 1, 1, 19}, {0, 0, 0, 0}}},{6417, {{20083, 1, 0, 0},{20083, 1, 1, 18}, {0, 0, 0, 0}}},{6417, {{20083, 1, 0, 0},{20083, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{6418, {{20084, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6419, {{20085, 1, 0, 0},{20085, 1, 1, 17}, {0, 0, 0, 0}}},{6419, {{20085, 1, 0, 0},{20085, 1, 1, 17}, {0, 0, 0, 0}}},{6420, {{20085, 1, 0, 0},{20085, 1, 1, 18}, {0, 0, 0, 0}}},{6420, {{20085, 1, 0, 0},{20085, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{68018, {{20086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{68019, {{20087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{6421, {{20089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{68020, {{20090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{68021, {{20091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{68022, {{20092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{68023, {{20093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6422, {{20094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6423, {{20095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6424, {{20096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6425, {{20097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6426, {{20098, 1, 0, 0},{20098, 1, 1, 17}, {0, 0, 0, 0}}},{6426, {{20098, 1, 0, 0},{20098, 1, 1, 17}, {0, 0, 0, 0}}},{6427, {{20098, 1, 0, 0},{20098, 1, 1, 18}, {0, 0, 0, 0}}},{6427, {{20098, 1, 0, 0},{20098, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{6428, {{20099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6429, {{20100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6430, {{20101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6431, {{20102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{6433, {{20104, 1, 0, 0},{20104, 1, 1, 17}, {0, 0, 0, 0}}},{6433, {{20104, 1, 0, 0},{20104, 1, 1, 17}, {0, 0, 0, 0}}},{56873, {{20104, 1, 0, 0},{20104, 1, 1, 18}, {0, 0, 0, 0}}},{56873, {{20104, 1, 0, 0},{20104, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{6434, {{20105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6435, {{20106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6436, {{20107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6437, {{20108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6438, {{20109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6439, {{20110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6440, {{20111, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56879, {{20112, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6441, {{20113, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6442, {{20114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6443, {{20115, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6444, {{20116, 1, 0, 0},{20116, 1, 1, 17}, {0, 0, 0, 0}}},{6444, {{20116, 1, 0, 0},{20116, 1, 1, 17}, {0, 0, 0, 0}}},{56822, {{20116, 1, 0, 0},{20116, 1, 1, 18}, {0, 0, 0, 0}}},{56822, {{20116, 1, 0, 0},{20116, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{6445, {{20117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6446, {{20118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6447, {{20119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6448, {{20120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6450, {{20121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{6453, {{20123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6454, {{20124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6455, {{20125, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6456, {{20126, 1, 0, 0},{20126, 1, 1, 17}, {0, 0, 0, 0}}},{6456, {{20126, 1, 0, 0},{20126, 1, 1, 17}, {0, 0, 0, 0}}},{56837, {{20126, 1, 0, 0},{20126, 1, 1, 18}, {0, 0, 0, 0}}},{56837, {{20126, 1, 0, 0},{20126, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{6457, {{20127, 1, 0, 0},{20127, 1, 1, 18}, {0, 0, 0, 0}}},{6457, {{20127, 1, 0, 0},{20127, 1, 1, 18}, {0, 0, 0, 0}}},{6458, {{20127, 1, 0, 0},{20127, 1, 1, 17}, {0, 0, 0, 0}}},{6458, {{20127, 1, 0, 0},{20127, 1, 1, 17}, {0, 0, 0, 0}}},{6459, {{20127, 1, 0, 0},{20127, 1, 1, 19}, {0, 0, 0, 0}}},{6459, {{20127, 1, 0, 0},{20127, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{6460, {{20128, 1, 0, 0},{20128, 1, 1, 17}, {0, 0, 0, 0}}},{6460, {{20128, 1, 0, 0},{20128, 1, 1, 17}, {0, 0, 0, 0}}},{6461, {{20128, 1, 0, 0},{20128, 1, 1, 18}, {0, 0, 0, 0}}},{6461, {{20128, 1, 0, 0},{20128, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{6462, {{20129, 1, 0, 0},{20129, 1, 1, 19}, {0, 0, 0, 0}}},{6462, {{20129, 1, 0, 0},{20129, 1, 1, 19}, {0, 0, 0, 0}}},{6463, {{20129, 1, 0, 0},{20129, 1, 1, 18}, {0, 0, 0, 0}}},{6463, {{20129, 1, 0, 0},{20129, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {8, {{6464, {{20130, 1, 0, 0},{20130, 1, 1, 17}, {0, 0, 0, 0}}},{6464, {{20130, 1, 0, 0},{20130, 1, 1, 17}, {0, 0, 0, 0}}},{6465, {{20130, 1, 0, 0},{20130, 1, 1, 20}, {0, 0, 0, 0}}},{6465, {{20130, 1, 0, 0},{20130, 1, 1, 20}, {0, 0, 0, 0}}},{56886, {{20130, 1, 0, 0},{20130, 1, 1, 18}, {0, 0, 0, 0}}},{56886, {{20130, 1, 0, 0},{20130, 1, 1, 18}, {0, 0, 0, 0}}},{56979, {{20130, 1, 0, 0},{20130, 1, 1, 21}, {0, 0, 0, 0}}},{56979, {{20130, 1, 0, 0},{20130, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{6466, {{20131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{6467, {{20132, 1, 0, 0},{20132, 1, 1, 18}, {0, 0, 0, 0}}},{6467, {{20132, 1, 0, 0},{20132, 1, 1, 18}, {0, 0, 0, 0}}},{6468, {{20132, 1, 0, 0},{20132, 1, 1, 19}, {0, 0, 0, 0}}},{6468, {{20132, 1, 0, 0},{20132, 1, 1, 19}, {0, 0, 0, 0}}},{6469, {{20132, 1, 0, 0},{20132, 1, 1, 20}, {0, 0, 0, 0}}},{6469, {{20132, 1, 0, 0},{20132, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{6470, {{20133, 1, 0, 0},{20133, 1, 1, 17}, {0, 0, 0, 0}}},{6470, {{20133, 1, 0, 0},{20133, 1, 1, 17}, {0, 0, 0, 0}}},{6471, {{20133, 1, 0, 0},{20133, 1, 1, 18}, {0, 0, 0, 0}}},{6471, {{20133, 1, 0, 0},{20133, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{6472, {{20134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{6473, {{20136, 1, 0, 0},{20136, 1, 1, 17}, {0, 0, 0, 0}}},{6473, {{20136, 1, 0, 0},{20136, 1, 1, 17}, {0, 0, 0, 0}}},{56890, {{20136, 1, 0, 0},{20136, 1, 1, 18}, {0, 0, 0, 0}}},{56890, {{20136, 1, 0, 0},{20136, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{6474, {{20139, 1, 0, 0},{20139, 1, 1, 17}, {0, 0, 0, 0}}},{6474, {{20139, 1, 0, 0},{20139, 1, 1, 17}, {0, 0, 0, 0}}},{56895, {{20139, 1, 0, 0},{20139, 1, 1, 18}, {0, 0, 0, 0}}},{56895, {{20139, 1, 0, 0},{20139, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{6475, {{20140, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6476, {{20141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6477, {{20142, 1, 0, 0},{20142, 1, 1, 17}, {0, 0, 0, 0}}},{6477, {{20142, 1, 0, 0},{20142, 1, 1, 17}, {0, 0, 0, 0}}},{6478, {{20142, 1, 0, 0},{20142, 1, 1, 18}, {0, 0, 0, 0}}},{6478, {{20142, 1, 0, 0},{20142, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{6479, {{20143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6480, {{20144, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6481, {{20145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6482, {{20146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6483, {{20147, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6484, {{20148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{6485, {{20150, 1, 0, 0},{20150, 1, 1, 17}, {0, 0, 0, 0}}},{6485, {{20150, 1, 0, 0},{20150, 1, 1, 17}, {0, 0, 0, 0}}},{6486, {{20150, 1, 0, 0},{20150, 1, 1, 18}, {0, 0, 0, 0}}},{6486, {{20150, 1, 0, 0},{20150, 1, 1, 18}, {0, 0, 0, 0}}},{6487, {{20150, 1, 0, 0},{20150, 1, 1, 19}, {0, 0, 0, 0}}},{6487, {{20150, 1, 0, 0},{20150, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{6488, {{20151, 1, 0, 0},{20151, 1, 1, 17}, {0, 0, 0, 0}}},{6488, {{20151, 1, 0, 0},{20151, 1, 1, 17}, {0, 0, 0, 0}}},{6489, {{20151, 1, 0, 0},{20151, 1, 1, 16}, {0, 0, 0, 0}}},{6489, {{20151, 1, 0, 0},{20151, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{6490, {{20153, 1, 0, 0},{20153, 1, 1, 18}, {0, 0, 0, 0}}},{6490, {{20153, 1, 0, 0},{20153, 1, 1, 18}, {0, 0, 0, 0}}},{6491, {{20153, 1, 0, 0},{20153, 1, 1, 17}, {0, 0, 0, 0}}},{6491, {{20153, 1, 0, 0},{20153, 1, 1, 17}, {0, 0, 0, 0}}},{6492, {{20153, 1, 0, 0},{20153, 1, 1, 19}, {0, 0, 0, 0}}},{6492, {{20153, 1, 0, 0},{20153, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{6493, {{20154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6494, {{20155, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6495, {{20156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56901, {{20158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{6496, {{20160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6497, {{20161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6498, {{20162, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6499, {{20163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6500, {{20164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6502, {{20165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6503, {{20166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6504, {{20167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6506, {{20168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6507, {{20169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6508, {{20170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6509, {{20171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6510, {{20172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6511, {{20173, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6512, {{20174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6513, {{20175, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6514, {{20176, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{6515, {{20180, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6516, {{20181, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6517, {{20182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{6518, {{20183, 1, 0, 0},{20183, 1, 1, 19}, {0, 0, 0, 0}}},{6518, {{20183, 1, 0, 0},{20183, 1, 1, 19}, {0, 0, 0, 0}}},{6519, {{20183, 1, 0, 0},{20183, 1, 1, 20}, {0, 0, 0, 0}}},{6519, {{20183, 1, 0, 0},{20183, 1, 1, 20}, {0, 0, 0, 0}}},{56911, {{20183, 1, 0, 0},{20183, 1, 1, 18}, {0, 0, 0, 0}}},{56911, {{20183, 1, 0, 0},{20183, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{6520, {{20184, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6521, {{20185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6522, {{20186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6523, {{20187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6524, {{20188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6525, {{20189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6526, {{20190, 1, 0, 0},{20190, 1, 1, 17}, {0, 0, 0, 0}}},{6526, {{20190, 1, 0, 0},{20190, 1, 1, 17}, {0, 0, 0, 0}}},{6527, {{20190, 1, 0, 0},{20190, 1, 1, 18}, {0, 0, 0, 0}}},{6527, {{20190, 1, 0, 0},{20190, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{6528, {{20191, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6529, {{20192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6530, {{20193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6531, {{20194, 1, 0, 0},{20194, 1, 1, 17}, {0, 0, 0, 0}}},{6531, {{20194, 1, 0, 0},{20194, 1, 1, 17}, {0, 0, 0, 0}}},{56903, {{20194, 1, 0, 0},{20194, 1, 1, 18}, {0, 0, 0, 0}}},{56903, {{20194, 1, 0, 0},{20194, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{6532, {{20195, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6533, {{20196, 1, 0, 0},{20196, 1, 1, 17}, {0, 0, 0, 0}}},{6533, {{20196, 1, 0, 0},{20196, 1, 1, 17}, {0, 0, 0, 0}}},{56905, {{20196, 1, 0, 0},{20196, 1, 1, 18}, {0, 0, 0, 0}}},{56905, {{20196, 1, 0, 0},{20196, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{6534, {{20197, 1, 0, 0},{20197, 1, 1, 17}, {0, 0, 0, 0}}},{6534, {{20197, 1, 0, 0},{20197, 1, 1, 17}, {0, 0, 0, 0}}},{57123, {{20197, 1, 0, 0},{20197, 1, 1, 18}, {0, 0, 0, 0}}},{57123, {{20197, 1, 0, 0},{20197, 1, 1, 18}, {0, 0, 0, 0}}},{57124, {{20197, 1, 0, 0},{20197, 1, 1, 19}, {0, 0, 0, 0}}},{57124, {{20197, 1, 0, 0},{20197, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{6535, {{20199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6536, {{20200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{6537, {{20203, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{6538, {{20205, 1, 0, 0},{20205, 1, 1, 17}, {0, 0, 0, 0}}},{6538, {{20205, 1, 0, 0},{20205, 1, 1, 17}, {0, 0, 0, 0}}},{6539, {{20205, 1, 0, 0},{20205, 1, 1, 18}, {0, 0, 0, 0}}},{6539, {{20205, 1, 0, 0},{20205, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{6540, {{20206, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6541, {{20207, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6542, {{20208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6543, {{20209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6544, {{20210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6545, {{20211, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6546, {{20212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6547, {{20213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6548, {{20214, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6549, {{20215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6550, {{20216, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6551, {{20217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6552, {{20218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6553, {{20219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6554, {{20220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6555, {{20221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6556, {{20222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6557, {{20223, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6558, {{20224, 1, 0, 0},{20224, 1, 1, 17}, {0, 0, 0, 0}}},{6558, {{20224, 1, 0, 0},{20224, 1, 1, 17}, {0, 0, 0, 0}}},{6559, {{20224, 1, 0, 0},{20224, 1, 1, 18}, {0, 0, 0, 0}}},{6559, {{20224, 1, 0, 0},{20224, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{6560, {{20225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6561, {{20226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6562, {{20227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6563, {{20228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6564, {{20229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6565, {{20230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6566, {{20231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6567, {{20232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6568, {{20233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6569, {{20234, 1, 0, 0},{20234, 1, 1, 17}, {0, 0, 0, 0}}},{6569, {{20234, 1, 0, 0},{20234, 1, 1, 17}, {0, 0, 0, 0}}},{59314, {{20234, 1, 0, 0},{20234, 1, 1, 18}, {0, 0, 0, 0}}},{59314, {{20234, 1, 0, 0},{20234, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{6570, {{20235, 1, 0, 0},{20235, 1, 1, 18}, {0, 0, 0, 0}}},{6570, {{20235, 1, 0, 0},{20235, 1, 1, 18}, {0, 0, 0, 0}}},{6571, {{20235, 1, 0, 0},{20235, 1, 1, 17}, {0, 0, 0, 0}}},{6571, {{20235, 1, 0, 0},{20235, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{6572, {{20236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6573, {{20237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6574, {{20238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6575, {{20239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6576, {{20240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6577, {{20241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6578, {{20242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6579, {{20243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6580, {{20244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6581, {{20245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6582, {{20246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6583, {{20247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6584, {{20248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6585, {{20249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6586, {{20250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6587, {{20251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6588, {{20252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6589, {{20253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{6592, {{20264, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{6593, {{20269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6594, {{20270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6595, {{20271, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6596, {{20272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6597, {{20273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6598, {{20274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6599, {{20275, 1, 0, 0},{20275, 1, 1, 17}, {0, 0, 0, 0}}},{6599, {{20275, 1, 0, 0},{20275, 1, 1, 17}, {0, 0, 0, 0}}},{6600, {{20275, 1, 0, 0},{20275, 1, 1, 18}, {0, 0, 0, 0}}},{6600, {{20275, 1, 0, 0},{20275, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{6601, {{20276, 1, 0, 0},{20276, 1, 1, 19}, {0, 0, 0, 0}}},{6601, {{20276, 1, 0, 0},{20276, 1, 1, 19}, {0, 0, 0, 0}}},{6602, {{20276, 1, 0, 0},{20276, 1, 1, 18}, {0, 0, 0, 0}}},{6602, {{20276, 1, 0, 0},{20276, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{6603, {{20277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6605, {{20278, 1, 0, 0},{20278, 1, 1, 17}, {0, 0, 0, 0}}},{6605, {{20278, 1, 0, 0},{20278, 1, 1, 17}, {0, 0, 0, 0}}},{6606, {{20278, 1, 0, 0},{20278, 1, 1, 18}, {0, 0, 0, 0}}},{6606, {{20278, 1, 0, 0},{20278, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{6607, {{20279, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6608, {{20280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6609, {{20281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6610, {{20282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6611, {{20283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6612, {{20284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6613, {{20285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6614, {{20286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6615, {{20287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6616, {{20288, 1, 0, 0},{20288, 1, 1, 17}, {0, 0, 0, 0}}},{6616, {{20288, 1, 0, 0},{20288, 1, 1, 17}, {0, 0, 0, 0}}},{6617, {{20288, 1, 0, 0},{20288, 1, 1, 18}, {0, 0, 0, 0}}},{6617, {{20288, 1, 0, 0},{20288, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{6618, {{20289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6619, {{20290, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6620, {{20291, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6621, {{20292, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6622, {{20293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6623, {{20294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6624, {{20295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6625, {{20296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6626, {{20297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6627, {{20298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6628, {{20299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6629, {{20300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6630, {{20301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6631, {{20302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6632, {{20303, 1, 0, 0},{20303, 1, 1, 18}, {0, 0, 0, 0}}},{6632, {{20303, 1, 0, 0},{20303, 1, 1, 18}, {0, 0, 0, 0}}},{6633, {{20303, 1, 0, 0},{20303, 1, 1, 19}, {0, 0, 0, 0}}},{6633, {{20303, 1, 0, 0},{20303, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{6634, {{20304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6635, {{20305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6636, {{20306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6637, {{20307, 1, 0, 0},{20307, 1, 1, 17}, {0, 0, 0, 0}}},{6637, {{20307, 1, 0, 0},{20307, 1, 1, 17}, {0, 0, 0, 0}}},{6638, {{20307, 1, 0, 0},{20307, 1, 1, 18}, {0, 0, 0, 0}}},{6638, {{20307, 1, 0, 0},{20307, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{6639, {{20308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6640, {{20309, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6641, {{20310, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6642, {{20311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6643, {{20312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6644, {{20313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6645, {{20314, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6646, {{20315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6647, {{20316, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6648, {{20317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6649, {{20318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6650, {{20319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{6651, {{20320, 1, 0, 0},{20320, 1, 1, 18}, {0, 0, 0, 0}}},{6651, {{20320, 1, 0, 0},{20320, 1, 1, 18}, {0, 0, 0, 0}}},{6652, {{20320, 1, 0, 0},{20320, 1, 1, 19}, {0, 0, 0, 0}}},{6652, {{20320, 1, 0, 0},{20320, 1, 1, 19}, {0, 0, 0, 0}}},{59316, {{20320, 1, 0, 0},{20320, 1, 1, 20}, {0, 0, 0, 0}}},{59316, {{20320, 1, 0, 0},{20320, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{6653, {{20321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6654, {{20322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6655, {{20323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6656, {{20324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{6657, {{20328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6658, {{20329, 1, 0, 0},{20329, 1, 1, 17}, {0, 0, 0, 0}}},{6658, {{20329, 1, 0, 0},{20329, 1, 1, 17}, {0, 0, 0, 0}}},{56914, {{20329, 1, 0, 0},{20329, 1, 1, 18}, {0, 0, 0, 0}}},{56914, {{20329, 1, 0, 0},{20329, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{6659, {{20330, 1, 0, 0},{20330, 1, 1, 17}, {0, 0, 0, 0}}},{6659, {{20330, 1, 0, 0},{20330, 1, 1, 17}, {0, 0, 0, 0}}},{6660, {{20330, 1, 0, 0},{20330, 1, 1, 18}, {0, 0, 0, 0}}},{6660, {{20330, 1, 0, 0},{20330, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{6661, {{20331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6662, {{20332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6663, {{20333, 1, 0, 0},{20333, 1, 1, 17}, {0, 0, 0, 0}}},{6663, {{20333, 1, 0, 0},{20333, 1, 1, 17}, {0, 0, 0, 0}}},{6664, {{20333, 1, 0, 0},{20333, 1, 1, 16}, {0, 0, 0, 0}}},{6664, {{20333, 1, 0, 0},{20333, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{6665, {{20334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6666, {{20335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6667, {{20336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6668, {{20337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6669, {{20338, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6670, {{20339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6671, {{20340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6672, {{20341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6673, {{20342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6675, {{20343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6676, {{20344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6677, {{20345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6678, {{20346, 1, 0, 0},{20346, 1, 1, 17}, {0, 0, 0, 0}}},{6678, {{20346, 1, 0, 0},{20346, 1, 1, 17}, {0, 0, 0, 0}}},{6679, {{20346, 1, 0, 0},{20346, 1, 1, 18}, {0, 0, 0, 0}}},{6679, {{20346, 1, 0, 0},{20346, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{6680, {{20347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6681, {{20348, 1, 0, 0},{20348, 1, 1, 17}, {0, 0, 0, 0}}},{6681, {{20348, 1, 0, 0},{20348, 1, 1, 17}, {0, 0, 0, 0}}},{6682, {{20348, 1, 0, 0},{20348, 1, 1, 18}, {0, 0, 0, 0}}},{6682, {{20348, 1, 0, 0},{20348, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{6683, {{20349, 1, 0, 0},{20349, 1, 1, 17}, {0, 0, 0, 0}}},{6683, {{20349, 1, 0, 0},{20349, 1, 1, 17}, {0, 0, 0, 0}}},{6684, {{20349, 1, 0, 0},{20349, 1, 1, 18}, {0, 0, 0, 0}}},{6684, {{20349, 1, 0, 0},{20349, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{6685, {{20350, 1, 0, 0},{20350, 1, 1, 18}, {0, 0, 0, 0}}},{6685, {{20350, 1, 0, 0},{20350, 1, 1, 18}, {0, 0, 0, 0}}},{6686, {{20350, 1, 0, 0},{20350, 1, 1, 17}, {0, 0, 0, 0}}},{6686, {{20350, 1, 0, 0},{20350, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{6687, {{20351, 1, 0, 0},{20351, 1, 1, 18}, {0, 0, 0, 0}}},{6687, {{20351, 1, 0, 0},{20351, 1, 1, 18}, {0, 0, 0, 0}}},{6688, {{20351, 1, 0, 0},{20351, 1, 1, 19}, {0, 0, 0, 0}}},{6688, {{20351, 1, 0, 0},{20351, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{6689, {{20352, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6690, {{20353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6691, {{20354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6692, {{20355, 1, 0, 0},{20355, 1, 1, 17}, {0, 0, 0, 0}}},{6692, {{20355, 1, 0, 0},{20355, 1, 1, 17}, {0, 0, 0, 0}}},{59324, {{20355, 1, 0, 0},{20355, 1, 1, 18}, {0, 0, 0, 0}}},{59324, {{20355, 1, 0, 0},{20355, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{6693, {{20356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6694, {{20357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6695, {{20358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6696, {{20359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6697, {{20360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6698, {{20361, 1, 0, 0},{20361, 1, 1, 17}, {0, 0, 0, 0}}},{6698, {{20361, 1, 0, 0},{20361, 1, 1, 17}, {0, 0, 0, 0}}},{56918, {{20361, 1, 0, 0},{20361, 1, 1, 18}, {0, 0, 0, 0}}},{56918, {{20361, 1, 0, 0},{20361, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{6699, {{20362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6700, {{20363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6701, {{20364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6702, {{20365, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6703, {{20366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6704, {{20367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6705, {{20368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6706, {{20369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6707, {{20370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6708, {{20371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6709, {{20372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6710, {{20373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{6711, {{20374, 1, 0, 0},{20374, 1, 1, 17}, {0, 0, 0, 0}}},{6711, {{20374, 1, 0, 0},{20374, 1, 1, 17}, {0, 0, 0, 0}}},{56919, {{20374, 1, 0, 0},{20374, 1, 1, 18}, {0, 0, 0, 0}}},{56919, {{20374, 1, 0, 0},{20374, 1, 1, 18}, {0, 0, 0, 0}}},{56926, {{20374, 1, 0, 0},{20374, 1, 1, 19}, {0, 0, 0, 0}}},{56926, {{20374, 1, 0, 0},{20374, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{6712, {{20375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6713, {{20376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6714, {{20377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6715, {{20378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6716, {{20379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6717, {{20380, 1, 0, 0},{20380, 1, 1, 17}, {0, 0, 0, 0}}},{6717, {{20380, 1, 0, 0},{20380, 1, 1, 17}, {0, 0, 0, 0}}},{6718, {{20380, 1, 0, 0},{20380, 1, 1, 16}, {0, 0, 0, 0}}},{6718, {{20380, 1, 0, 0},{20380, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{6719, {{20381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6720, {{20382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6721, {{20383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6722, {{20384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6723, {{20385, 1, 0, 0},{20385, 1, 1, 17}, {0, 0, 0, 0}}},{6723, {{20385, 1, 0, 0},{20385, 1, 1, 17}, {0, 0, 0, 0}}},{56917, {{20385, 1, 0, 0},{20385, 1, 1, 18}, {0, 0, 0, 0}}},{56917, {{20385, 1, 0, 0},{20385, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{6724, {{20388, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{6725, {{20395, 1, 0, 0},{20395, 1, 1, 17}, {0, 0, 0, 0}}},{6725, {{20395, 1, 0, 0},{20395, 1, 1, 17}, {0, 0, 0, 0}}},{6726, {{20395, 1, 0, 0},{20395, 1, 1, 18}, {0, 0, 0, 0}}},{6726, {{20395, 1, 0, 0},{20395, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{6727, {{20397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{6728, {{20398, 1, 0, 0},{20398, 1, 1, 18}, {0, 0, 0, 0}}},{6728, {{20398, 1, 0, 0},{20398, 1, 1, 18}, {0, 0, 0, 0}}},{30244, {{20398, 1, 0, 0},{20398, 1, 1, 0},{20398, 1, 1, 19},{64048, 1, 0, 0}, {0, 0, 0, 0}}},{30244, {{20398, 1, 0, 0},{20398, 1, 1, 0},{20398, 1, 1, 19},{64048, 1, 0, 0}, {0, 0, 0, 0}}},{30244, {{20398, 1, 0, 0},{20398, 1, 1, 0},{20398, 1, 1, 19},{64048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6729, {{20399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6730, {{20400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6731, {{20401, 1, 0, 0},{20401, 1, 1, 17}, {0, 0, 0, 0}}},{6731, {{20401, 1, 0, 0},{20401, 1, 1, 17}, {0, 0, 0, 0}}},{6732, {{20401, 1, 0, 0},{20401, 1, 1, 16}, {0, 0, 0, 0}}},{6732, {{20401, 1, 0, 0},{20401, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{6733, {{20402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6734, {{20403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6735, {{20404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6736, {{20405, 1, 0, 0},{20405, 1, 1, 19}, {0, 0, 0, 0}}},{6736, {{20405, 1, 0, 0},{20405, 1, 1, 19}, {0, 0, 0, 0}}},{6737, {{20405, 1, 0, 0},{20405, 1, 1, 18}, {0, 0, 0, 0}}},{6737, {{20405, 1, 0, 0},{20405, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{6738, {{20406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6739, {{20407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6740, {{20408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6741, {{20409, 1, 0, 0},{20409, 1, 1, 18}, {0, 0, 0, 0}}},{6741, {{20409, 1, 0, 0},{20409, 1, 1, 18}, {0, 0, 0, 0}}},{6742, {{20409, 1, 0, 0},{20409, 1, 1, 17}, {0, 0, 0, 0}}},{6742, {{20409, 1, 0, 0},{20409, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{6743, {{20410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6744, {{20411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6746, {{20412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6747, {{20413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6748, {{20414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6749, {{20415, 1, 0, 0},{20415, 1, 1, 18}, {0, 0, 0, 0}}},{6749, {{20415, 1, 0, 0},{20415, 1, 1, 18}, {0, 0, 0, 0}}},{6750, {{20415, 1, 0, 0},{20415, 1, 1, 19}, {0, 0, 0, 0}}},{6750, {{20415, 1, 0, 0},{20415, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{6751, {{20416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6752, {{20417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6753, {{20418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6754, {{20419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6755, {{20420, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6756, {{20421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6757, {{20422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6758, {{20423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6759, {{20424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6761, {{20425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6762, {{20426, 1, 0, 0},{20426, 1, 1, 18}, {0, 0, 0, 0}}},{6762, {{20426, 1, 0, 0},{20426, 1, 1, 18}, {0, 0, 0, 0}}},{6763, {{20426, 1, 0, 0},{20426, 1, 1, 17}, {0, 0, 0, 0}}},{6763, {{20426, 1, 0, 0},{20426, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{6764, {{20427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6765, {{20428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6766, {{20429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6767, {{20430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6768, {{20431, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6769, {{20432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6770, {{20433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6771, {{20434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6772, {{20435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6773, {{20436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6774, {{20437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{6775, {{20439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6776, {{20440, 1, 0, 0},{20440, 1, 1, 17}, {0, 0, 0, 0}}},{6776, {{20440, 1, 0, 0},{20440, 1, 1, 17}, {0, 0, 0, 0}}},{6777, {{20440, 1, 0, 0},{20440, 1, 1, 18}, {0, 0, 0, 0}}},{6777, {{20440, 1, 0, 0},{20440, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{6778, {{20441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6779, {{20442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6780, {{20443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6781, {{20444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6782, {{20445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{6783, {{20446, 1, 0, 0},{20446, 1, 1, 17}, {0, 0, 0, 0}}},{6783, {{20446, 1, 0, 0},{20446, 1, 1, 17}, {0, 0, 0, 0}}},{6784, {{20446, 1, 0, 0},{20446, 1, 1, 16}, {0, 0, 0, 0}}},{6784, {{20446, 1, 0, 0},{20446, 1, 1, 16}, {0, 0, 0, 0}}},{6785, {{20446, 1, 0, 0},{20446, 1, 1, 18}, {0, 0, 0, 0}}},{6785, {{20446, 1, 0, 0},{20446, 1, 1, 18}, {0, 0, 0, 0}}},{56932, {{20446, 1, 0, 0},{20446, 1, 1, 19}, {0, 0, 0, 0}}},{56932, {{20446, 1, 0, 0},{20446, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{6786, {{20447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6787, {{20448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6788, {{20449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6789, {{20450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6790, {{20451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6791, {{20452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6792, {{20453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6793, {{20454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{6795, {{20462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6796, {{20463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6797, {{20464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6798, {{20465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6799, {{20466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6800, {{20467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6801, {{20468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6802, {{20469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6803, {{20470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6804, {{20471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6805, {{20472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6806, {{20473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6807, {{20474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6808, {{20475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6809, {{20476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6810, {{20477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6811, {{20478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6812, {{20479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6813, {{20480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6814, {{20481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6815, {{20482, 1, 0, 0},{20482, 1, 1, 18}, {0, 0, 0, 0}}},{6815, {{20482, 1, 0, 0},{20482, 1, 1, 18}, {0, 0, 0, 0}}},{6816, {{20482, 1, 0, 0},{20482, 1, 1, 17}, {0, 0, 0, 0}}},{6816, {{20482, 1, 0, 0},{20482, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{6817, {{20483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6818, {{20484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6819, {{20485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6820, {{20486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6821, {{20487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6822, {{20488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6823, {{20489, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6824, {{20490, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6825, {{20491, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6826, {{20492, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6827, {{20493, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6828, {{20494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6829, {{20495, 1, 0, 0},{20495, 1, 1, 17}, {0, 0, 0, 0}}},{6829, {{20495, 1, 0, 0},{20495, 1, 1, 17}, {0, 0, 0, 0}}},{6830, {{20495, 1, 0, 0},{20495, 1, 1, 18}, {0, 0, 0, 0}}},{6830, {{20495, 1, 0, 0},{20495, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{6831, {{20496, 1, 0, 0},{20496, 1, 1, 17}, {0, 0, 0, 0}}},{6831, {{20496, 1, 0, 0},{20496, 1, 1, 17}, {0, 0, 0, 0}}},{6832, {{20496, 1, 0, 0},{20496, 1, 1, 18}, {0, 0, 0, 0}}},{6832, {{20496, 1, 0, 0},{20496, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{6833, {{20497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6834, {{20498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6835, {{20499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6836, {{20500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6837, {{20501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6838, {{20502, 1, 0, 0},{20502, 1, 1, 17}, {0, 0, 0, 0}}},{6838, {{20502, 1, 0, 0},{20502, 1, 1, 17}, {0, 0, 0, 0}}},{6839, {{20502, 1, 0, 0},{20502, 1, 1, 18}, {0, 0, 0, 0}}},{6839, {{20502, 1, 0, 0},{20502, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{6840, {{20503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6842, {{20504, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6843, {{20505, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6844, {{20506, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6845, {{20507, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6846, {{20508, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6848, {{20509, 1, 0, 0},{20509, 1, 1, 17}, {0, 0, 0, 0}}},{6848, {{20509, 1, 0, 0},{20509, 1, 1, 17}, {0, 0, 0, 0}}},{56931, {{20509, 1, 0, 0},{20509, 1, 1, 18}, {0, 0, 0, 0}}},{56931, {{20509, 1, 0, 0},{20509, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{6849, {{20510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6850, {{20511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6851, {{20512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6852, {{20513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6853, {{20514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6854, {{20515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6855, {{20516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6856, {{20517, 1, 0, 0},{20517, 1, 1, 17}, {0, 0, 0, 0}}},{6856, {{20517, 1, 0, 0},{20517, 1, 1, 17}, {0, 0, 0, 0}}},{6857, {{20517, 1, 0, 0},{20517, 1, 1, 18}, {0, 0, 0, 0}}},{6857, {{20517, 1, 0, 0},{20517, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{6858, {{20518, 1, 0, 0},{20518, 1, 1, 19}, {0, 0, 0, 0}}},{6858, {{20518, 1, 0, 0},{20518, 1, 1, 19}, {0, 0, 0, 0}}},{6859, {{20518, 1, 0, 0},{20518, 1, 1, 18}, {0, 0, 0, 0}}},{6859, {{20518, 1, 0, 0},{20518, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{6860, {{20519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6861, {{20520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6862, {{20521, 1, 0, 0},{20521, 1, 1, 17}, {0, 0, 0, 0}}},{6862, {{20521, 1, 0, 0},{20521, 1, 1, 17}, {0, 0, 0, 0}}},{6863, {{20521, 1, 0, 0},{20521, 1, 1, 18}, {0, 0, 0, 0}}},{6863, {{20521, 1, 0, 0},{20521, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{6864, {{20522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6865, {{20523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6867, {{20524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6868, {{20525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6869, {{20526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6870, {{20527, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6871, {{20528, 1, 0, 0},{20528, 1, 1, 17}, {0, 0, 0, 0}}},{6871, {{20528, 1, 0, 0},{20528, 1, 1, 17}, {0, 0, 0, 0}}},{6872, {{20528, 1, 0, 0},{20528, 1, 1, 18}, {0, 0, 0, 0}}},{6872, {{20528, 1, 0, 0},{20528, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{6873, {{20529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6874, {{20530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6875, {{20531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6876, {{20532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6877, {{20533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6878, {{20534, 1, 0, 0},{20534, 1, 1, 17}, {0, 0, 0, 0}}},{6878, {{20534, 1, 0, 0},{20534, 1, 1, 17}, {0, 0, 0, 0}}},{6879, {{20534, 1, 0, 0},{20534, 1, 1, 18}, {0, 0, 0, 0}}},{6879, {{20534, 1, 0, 0},{20534, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{6880, {{20535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6881, {{20536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6882, {{20537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{6883, {{20539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{6884, {{20543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6885, {{20544, 1, 0, 0},{20544, 1, 1, 17}, {0, 0, 0, 0}}},{6885, {{20544, 1, 0, 0},{20544, 1, 1, 17}, {0, 0, 0, 0}}},{6886, {{20544, 1, 0, 0},{20544, 1, 1, 18}, {0, 0, 0, 0}}},{6886, {{20544, 1, 0, 0},{20544, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{6887, {{20545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6888, {{20546, 1, 0, 0},{20546, 1, 1, 17}, {0, 0, 0, 0}}},{6888, {{20546, 1, 0, 0},{20546, 1, 1, 17}, {0, 0, 0, 0}}},{6889, {{20546, 1, 0, 0},{20546, 1, 1, 18}, {0, 0, 0, 0}}},{6889, {{20546, 1, 0, 0},{20546, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{6890, {{20547, 1, 0, 0},{20547, 1, 1, 17}, {0, 0, 0, 0}}},{6890, {{20547, 1, 0, 0},{20547, 1, 1, 17}, {0, 0, 0, 0}}},{6891, {{20547, 1, 0, 0},{20547, 1, 1, 18}, {0, 0, 0, 0}}},{6891, {{20547, 1, 0, 0},{20547, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{6892, {{20548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6893, {{20549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6894, {{20550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6895, {{20551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6896, {{20552, 1, 0, 0},{20552, 1, 1, 17}, {0, 0, 0, 0}}},{6896, {{20552, 1, 0, 0},{20552, 1, 1, 17}, {0, 0, 0, 0}}},{6897, {{20552, 1, 0, 0},{20552, 1, 1, 18}, {0, 0, 0, 0}}},{6897, {{20552, 1, 0, 0},{20552, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{6898, {{20553, 1, 0, 0},{20553, 1, 1, 19}, {0, 0, 0, 0}}},{6898, {{20553, 1, 0, 0},{20553, 1, 1, 19}, {0, 0, 0, 0}}},{6899, {{20553, 1, 0, 0},{20553, 1, 1, 18}, {0, 0, 0, 0}}},{6899, {{20553, 1, 0, 0},{20553, 1, 1, 18}, {0, 0, 0, 0}}},{6900, {{20553, 1, 0, 0},{20553, 1, 1, 21}, {0, 0, 0, 0}}},{6900, {{20553, 1, 0, 0},{20553, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{6901, {{20554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6902, {{20555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6903, {{20556, 1, 0, 0},{20556, 1, 1, 17}, {0, 0, 0, 0}}},{6903, {{20556, 1, 0, 0},{20556, 1, 1, 17}, {0, 0, 0, 0}}},{6904, {{20556, 1, 0, 0},{20556, 1, 1, 18}, {0, 0, 0, 0}}},{6904, {{20556, 1, 0, 0},{20556, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{6905, {{20557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6906, {{20558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6907, {{20559, 1, 0, 0},{20559, 1, 1, 19}, {0, 0, 0, 0}}},{6907, {{20559, 1, 0, 0},{20559, 1, 1, 19}, {0, 0, 0, 0}}},{6908, {{20559, 1, 0, 0},{20559, 1, 1, 18}, {0, 0, 0, 0}}},{6908, {{20559, 1, 0, 0},{20559, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{6909, {{20560, 1, 0, 0},{20560, 1, 1, 17}, {0, 0, 0, 0}}},{6909, {{20560, 1, 0, 0},{20560, 1, 1, 17}, {0, 0, 0, 0}}},{6910, {{20560, 1, 0, 0},{20560, 1, 1, 18}, {0, 0, 0, 0}}},{6910, {{20560, 1, 0, 0},{20560, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{6911, {{20561, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6912, {{20562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6913, {{20563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{6914, {{20565, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6915, {{20566, 1, 0, 0},{20566, 1, 1, 18}, {0, 0, 0, 0}}},{6915, {{20566, 1, 0, 0},{20566, 1, 1, 18}, {0, 0, 0, 0}}},{6916, {{20566, 1, 0, 0},{20566, 1, 1, 19}, {0, 0, 0, 0}}},{6916, {{20566, 1, 0, 0},{20566, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{6917, {{20567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6918, {{20568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6919, {{20569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6920, {{20570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6921, {{20571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6922, {{20572, 1, 0, 0},{20572, 1, 1, 17}, {0, 0, 0, 0}}},{6922, {{20572, 1, 0, 0},{20572, 1, 1, 17}, {0, 0, 0, 0}}},{56942, {{20572, 1, 0, 0},{20572, 1, 1, 18}, {0, 0, 0, 0}}},{56942, {{20572, 1, 0, 0},{20572, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{6923, {{20573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6924, {{20574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6925, {{20575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6926, {{20576, 1, 0, 0},{20576, 1, 1, 17}, {0, 0, 0, 0}}},{6926, {{20576, 1, 0, 0},{20576, 1, 1, 17}, {0, 0, 0, 0}}},{6927, {{20576, 1, 0, 0},{20576, 1, 1, 18}, {0, 0, 0, 0}}},{6927, {{20576, 1, 0, 0},{20576, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{6928, {{20577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6929, {{20578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6930, {{20579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6931, {{20580, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{6933, {{20581, 1, 0, 0},{20581, 1, 1, 18}, {0, 0, 0, 0}}},{6933, {{20581, 1, 0, 0},{20581, 1, 1, 18}, {0, 0, 0, 0}}},{6934, {{20581, 1, 0, 0},{20581, 1, 1, 20}, {0, 0, 0, 0}}},{6934, {{20581, 1, 0, 0},{20581, 1, 1, 20}, {0, 0, 0, 0}}},{56945, {{20581, 1, 0, 0},{20581, 1, 1, 19}, {0, 0, 0, 0}}},{56945, {{20581, 1, 0, 0},{20581, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{6935, {{20582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6936, {{20583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6937, {{20584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6938, {{20585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6939, {{20586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6940, {{20587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6941, {{20588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6942, {{20589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6943, {{20590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6944, {{20591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{6945, {{20592, 1, 0, 0},{20592, 1, 1, 17}, {0, 0, 0, 0}}},{6945, {{20592, 1, 0, 0},{20592, 1, 1, 17}, {0, 0, 0, 0}}},{6946, {{20592, 1, 0, 0},{20592, 1, 1, 19}, {0, 0, 0, 0}}},{6946, {{20592, 1, 0, 0},{20592, 1, 1, 19}, {0, 0, 0, 0}}},{6947, {{20592, 1, 0, 0},{20592, 1, 1, 18}, {0, 0, 0, 0}}},{6947, {{20592, 1, 0, 0},{20592, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{6948, {{20593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6949, {{20594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6950, {{20595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6951, {{20596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6952, {{20597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6953, {{20598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6954, {{20599, 1, 0, 0},{20599, 1, 1, 18}, {0, 0, 0, 0}}},{6954, {{20599, 1, 0, 0},{20599, 1, 1, 18}, {0, 0, 0, 0}}},{6955, {{20599, 1, 0, 0},{20599, 1, 1, 17}, {0, 0, 0, 0}}},{6955, {{20599, 1, 0, 0},{20599, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{6956, {{20600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6957, {{20601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6958, {{20602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6959, {{20603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6960, {{20604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6961, {{20605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{6962, {{20608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6963, {{20609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6964, {{20610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6965, {{20611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6966, {{20612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6967, {{20613, 1, 0, 0},{20613, 1, 1, 18}, {0, 0, 0, 0}}},{6967, {{20613, 1, 0, 0},{20613, 1, 1, 18}, {0, 0, 0, 0}}},{6968, {{20613, 1, 0, 0},{20613, 1, 1, 19}, {0, 0, 0, 0}}},{6968, {{20613, 1, 0, 0},{20613, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{6971, {{20614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6972, {{20615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6973, {{20616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6974, {{20617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6975, {{20618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6976, {{20619, 1, 0, 0},{20619, 1, 1, 16}, {0, 0, 0, 0}}},{6976, {{20619, 1, 0, 0},{20619, 1, 1, 16}, {0, 0, 0, 0}}},{6977, {{20619, 1, 0, 0},{20619, 1, 1, 17}, {0, 0, 0, 0}}},{6977, {{20619, 1, 0, 0},{20619, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{6978, {{20620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6979, {{20621, 1, 0, 0},{20621, 1, 1, 17}, {0, 0, 0, 0}}},{6979, {{20621, 1, 0, 0},{20621, 1, 1, 17}, {0, 0, 0, 0}}},{6980, {{20621, 1, 0, 0},{20621, 1, 1, 18}, {0, 0, 0, 0}}},{6980, {{20621, 1, 0, 0},{20621, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{6981, {{20622, 1, 0, 0},{20622, 1, 1, 17}, {0, 0, 0, 0}}},{6981, {{20622, 1, 0, 0},{20622, 1, 1, 17}, {0, 0, 0, 0}}},{6982, {{20622, 1, 0, 0},{20622, 1, 1, 18}, {0, 0, 0, 0}}},{6982, {{20622, 1, 0, 0},{20622, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{6983, {{20623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6984, {{20624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{6986, {{20625, 1, 0, 0},{20625, 1, 1, 20}, {0, 0, 0, 0}}},{6986, {{20625, 1, 0, 0},{20625, 1, 1, 20}, {0, 0, 0, 0}}},{6987, {{20625, 1, 0, 0},{20625, 1, 1, 19}, {0, 0, 0, 0}}},{6987, {{20625, 1, 0, 0},{20625, 1, 1, 19}, {0, 0, 0, 0}}},{6988, {{20625, 1, 0, 0},{20625, 1, 1, 21}, {0, 0, 0, 0}}},{6988, {{20625, 1, 0, 0},{20625, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{6989, {{20626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6990, {{20627, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6991, {{20628, 1, 0, 0},{20628, 1, 1, 17}, {0, 0, 0, 0}}},{6991, {{20628, 1, 0, 0},{20628, 1, 1, 17}, {0, 0, 0, 0}}},{6992, {{20628, 1, 0, 0},{20628, 1, 1, 18}, {0, 0, 0, 0}}},{6992, {{20628, 1, 0, 0},{20628, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{6993, {{20629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6994, {{20630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6995, {{20631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{6996, {{20632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{6997, {{20633, 1, 0, 0},{20633, 1, 1, 17}, {0, 0, 0, 0}}},{6997, {{20633, 1, 0, 0},{20633, 1, 1, 17}, {0, 0, 0, 0}}},{6998, {{20633, 1, 0, 0},{20633, 1, 1, 18}, {0, 0, 0, 0}}},{6998, {{20633, 1, 0, 0},{20633, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{6999, {{20634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7000, {{20635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7001, {{20636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7002, {{20637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7003, {{20638, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7004, {{20639, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7005, {{20640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7006, {{20641, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7007, {{20642, 1, 0, 0},{20642, 1, 1, 17}, {0, 0, 0, 0}}},{7007, {{20642, 1, 0, 0},{20642, 1, 1, 17}, {0, 0, 0, 0}}},{7008, {{20642, 1, 0, 0},{20642, 1, 1, 18}, {0, 0, 0, 0}}},{7008, {{20642, 1, 0, 0},{20642, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7009, {{20643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{7010, {{20645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{7011, {{20650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7012, {{20651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7013, {{20652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7014, {{20653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7015, {{20654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7016, {{20655, 1, 0, 0},{20655, 1, 1, 17}, {0, 0, 0, 0}}},{7016, {{20655, 1, 0, 0},{20655, 1, 1, 17}, {0, 0, 0, 0}}},{7017, {{20655, 1, 0, 0},{20655, 1, 1, 18}, {0, 0, 0, 0}}},{7017, {{20655, 1, 0, 0},{20655, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7018, {{20656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7020, {{20657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7021, {{20658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7023, {{20659, 1, 0, 0},{20659, 1, 1, 17}, {0, 0, 0, 0}}},{7023, {{20659, 1, 0, 0},{20659, 1, 1, 17}, {0, 0, 0, 0}}},{7024, {{20659, 1, 0, 0},{20659, 1, 1, 18}, {0, 0, 0, 0}}},{7024, {{20659, 1, 0, 0},{20659, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{7026, {{20660, 1, 0, 0},{20660, 1, 1, 17}, {0, 0, 0, 0}}},{7026, {{20660, 1, 0, 0},{20660, 1, 1, 17}, {0, 0, 0, 0}}},{7027, {{20660, 1, 0, 0},{20660, 1, 1, 18}, {0, 0, 0, 0}}},{7027, {{20660, 1, 0, 0},{20660, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7028, {{20661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7029, {{20662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7030, {{20663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7031, {{20664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7032, {{20665, 1, 0, 0},{20665, 1, 1, 17}, {0, 0, 0, 0}}},{7032, {{20665, 1, 0, 0},{20665, 1, 1, 17}, {0, 0, 0, 0}}},{7033, {{20665, 1, 0, 0},{20665, 1, 1, 18}, {0, 0, 0, 0}}},{7033, {{20665, 1, 0, 0},{20665, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7034, {{20666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7035, {{20667, 1, 0, 0},{20667, 1, 1, 18}, {0, 0, 0, 0}}},{7035, {{20667, 1, 0, 0},{20667, 1, 1, 18}, {0, 0, 0, 0}}},{7036, {{20667, 1, 0, 0},{20667, 1, 1, 17}, {0, 0, 0, 0}}},{7036, {{20667, 1, 0, 0},{20667, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{7037, {{20668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7038, {{20669, 1, 0, 0},{20669, 1, 1, 17}, {0, 0, 0, 0}}},{7038, {{20669, 1, 0, 0},{20669, 1, 1, 17}, {0, 0, 0, 0}}},{7039, {{20669, 1, 0, 0},{20669, 1, 1, 18}, {0, 0, 0, 0}}},{7039, {{20669, 1, 0, 0},{20669, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7040, {{20670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7041, {{20671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7042, {{20672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7043, {{20673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7044, {{20674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7045, {{20675, 1, 0, 0},{20675, 1, 1, 17}, {0, 0, 0, 0}}},{7045, {{20675, 1, 0, 0},{20675, 1, 1, 17}, {0, 0, 0, 0}}},{7046, {{20675, 1, 0, 0},{20675, 1, 1, 18}, {0, 0, 0, 0}}},{7046, {{20675, 1, 0, 0},{20675, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7047, {{20676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7048, {{20677, 1, 0, 0},{20677, 1, 1, 18}, {0, 0, 0, 0}}},{7048, {{20677, 1, 0, 0},{20677, 1, 1, 18}, {0, 0, 0, 0}}},{7049, {{20677, 1, 0, 0},{20677, 1, 1, 19}, {0, 0, 0, 0}}},{7049, {{20677, 1, 0, 0},{20677, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{7050, {{20678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7051, {{20679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7052, {{20680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{7053, {{20681, 1, 0, 0},{20681, 1, 1, 17}, {0, 0, 0, 0}}},{7053, {{20681, 1, 0, 0},{20681, 1, 1, 17}, {0, 0, 0, 0}}},{56936, {{20681, 1, 0, 0},{20681, 1, 1, 18}, {0, 0, 0, 0}}},{56936, {{20681, 1, 0, 0},{20681, 1, 1, 18}, {0, 0, 0, 0}}},{56948, {{20681, 1, 0, 0},{20681, 1, 1, 19}, {0, 0, 0, 0}}},{56948, {{20681, 1, 0, 0},{20681, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {8, {{7054, {{20682, 1, 0, 0},{20682, 1, 1, 20}, {0, 0, 0, 0}}},{7054, {{20682, 1, 0, 0},{20682, 1, 1, 20}, {0, 0, 0, 0}}},{7055, {{20682, 1, 0, 0},{20682, 1, 1, 21}, {0, 0, 0, 0}}},{7055, {{20682, 1, 0, 0},{20682, 1, 1, 21}, {0, 0, 0, 0}}},{7056, {{20682, 1, 0, 0},{20682, 1, 1, 19}, {0, 0, 0, 0}}},{7056, {{20682, 1, 0, 0},{20682, 1, 1, 19}, {0, 0, 0, 0}}},{7057, {{20682, 1, 0, 0},{20682, 1, 1, 22}, {0, 0, 0, 0}}},{7057, {{20682, 1, 0, 0},{20682, 1, 1, 22}, {0, 0, 0, 0}}}}},
    {1, {{7058, {{20683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7059, {{20684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7060, {{20685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7061, {{20686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{7062, {{20687, 1, 0, 0},{20687, 1, 1, 18}, {0, 0, 0, 0}}},{7062, {{20687, 1, 0, 0},{20687, 1, 1, 18}, {0, 0, 0, 0}}},{7063, {{20687, 1, 0, 0},{20687, 1, 1, 19}, {0, 0, 0, 0}}},{7063, {{20687, 1, 0, 0},{20687, 1, 1, 19}, {0, 0, 0, 0}}},{7064, {{20687, 1, 0, 0},{20687, 1, 1, 20}, {0, 0, 0, 0}}},{7064, {{20687, 1, 0, 0},{20687, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{7065, {{20688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7066, {{20689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7067, {{20690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7068, {{20691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7069, {{20692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7070, {{20693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7071, {{20694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7072, {{20695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7073, {{20696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7074, {{20697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7076, {{20698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7077, {{20699, 1, 0, 0},{20699, 1, 1, 16}, {0, 0, 0, 0}}},{7077, {{20699, 1, 0, 0},{20699, 1, 1, 16}, {0, 0, 0, 0}}},{7078, {{20699, 1, 0, 0},{20699, 1, 1, 17}, {0, 0, 0, 0}}},{7078, {{20699, 1, 0, 0},{20699, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{7079, {{20700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7080, {{20701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7081, {{20702, 1, 0, 0},{20702, 1, 1, 17}, {0, 0, 0, 0}}},{7081, {{20702, 1, 0, 0},{20702, 1, 1, 17}, {0, 0, 0, 0}}},{7082, {{20702, 1, 0, 0},{20702, 1, 1, 18}, {0, 0, 0, 0}}},{7082, {{20702, 1, 0, 0},{20702, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7083, {{20703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7086, {{20704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7087, {{20705, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7088, {{20706, 1, 0, 0},{20706, 1, 1, 18}, {0, 0, 0, 0}}},{7088, {{20706, 1, 0, 0},{20706, 1, 1, 18}, {0, 0, 0, 0}}},{7089, {{20706, 1, 0, 0},{20706, 1, 1, 17}, {0, 0, 0, 0}}},{7089, {{20706, 1, 0, 0},{20706, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{7090, {{20707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7091, {{20708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7092, {{20709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7093, {{20710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {9, {{7094, {{20711, 1, 0, 0},{20711, 1, 1, 19}, {0, 0, 0, 0}}},{7094, {{20711, 1, 0, 0},{20711, 1, 1, 19}, {0, 0, 0, 0}}},{7095, {{20711, 1, 0, 0},{20711, 1, 1, 18}, {0, 0, 0, 0}}},{7095, {{20711, 1, 0, 0},{20711, 1, 1, 18}, {0, 0, 0, 0}}},{7096, {{20711, 1, 0, 0},{20711, 1, 1, 20}, {0, 0, 0, 0}}},{7096, {{20711, 1, 0, 0},{20711, 1, 1, 20}, {0, 0, 0, 0}}},{30245, {{20711, 1, 0, 0},{20711, 1, 1, 0},{20711, 1, 1, 21},{64049, 1, 0, 0}, {0, 0, 0, 0}}},{30245, {{20711, 1, 0, 0},{20711, 1, 1, 0},{20711, 1, 1, 21},{64049, 1, 0, 0}, {0, 0, 0, 0}}},{30245, {{20711, 1, 0, 0},{20711, 1, 1, 0},{20711, 1, 1, 21},{64049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7097, {{20712, 1, 0, 0},{20712, 1, 1, 17}, {0, 0, 0, 0}}},{7097, {{20712, 1, 0, 0},{20712, 1, 1, 17}, {0, 0, 0, 0}}},{56951, {{20712, 1, 0, 0},{20712, 1, 1, 18}, {0, 0, 0, 0}}},{56951, {{20712, 1, 0, 0},{20712, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7098, {{20713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7099, {{20714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7100, {{20715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7101, {{20716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7102, {{20717, 1, 0, 0},{20717, 1, 1, 18}, {0, 0, 0, 0}}},{7102, {{20717, 1, 0, 0},{20717, 1, 1, 18}, {0, 0, 0, 0}}},{7103, {{20717, 1, 0, 0},{20717, 1, 1, 19}, {0, 0, 0, 0}}},{7103, {{20717, 1, 0, 0},{20717, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{7104, {{20718, 1, 0, 0},{20718, 1, 1, 17}, {0, 0, 0, 0}}},{7104, {{20718, 1, 0, 0},{20718, 1, 1, 17}, {0, 0, 0, 0}}},{7105, {{20718, 1, 0, 0},{20718, 1, 1, 18}, {0, 0, 0, 0}}},{7105, {{20718, 1, 0, 0},{20718, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{7106, {{20719, 1, 0, 0},{20719, 1, 1, 18}, {0, 0, 0, 0}}},{7106, {{20719, 1, 0, 0},{20719, 1, 1, 18}, {0, 0, 0, 0}}},{7107, {{20719, 1, 0, 0},{20719, 1, 1, 17}, {0, 0, 0, 0}}},{7107, {{20719, 1, 0, 0},{20719, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{7108, {{20720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7109, {{20721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{7110, {{20722, 1, 0, 0},{20722, 1, 1, 18}, {0, 0, 0, 0}}},{7110, {{20722, 1, 0, 0},{20722, 1, 1, 18}, {0, 0, 0, 0}}},{7111, {{20722, 1, 0, 0},{20722, 1, 1, 19}, {0, 0, 0, 0}}},{7111, {{20722, 1, 0, 0},{20722, 1, 1, 19}, {0, 0, 0, 0}}},{7112, {{20722, 1, 0, 0},{20722, 1, 1, 20}, {0, 0, 0, 0}}},{7112, {{20722, 1, 0, 0},{20722, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{7113, {{20723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7114, {{20724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7115, {{20725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{7116, {{20726, 1, 0, 0},{20726, 1, 1, 19}, {0, 0, 0, 0}}},{7116, {{20726, 1, 0, 0},{20726, 1, 1, 19}, {0, 0, 0, 0}}},{7117, {{20726, 1, 0, 0},{20726, 1, 1, 17}, {0, 0, 0, 0}}},{7117, {{20726, 1, 0, 0},{20726, 1, 1, 17}, {0, 0, 0, 0}}},{56953, {{20726, 1, 0, 0},{20726, 1, 1, 18}, {0, 0, 0, 0}}},{56953, {{20726, 1, 0, 0},{20726, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7118, {{20727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7120, {{20728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7121, {{20729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7122, {{20730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7123, {{20731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7124, {{20732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7125, {{20733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7126, {{20734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7127, {{20735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7128, {{20736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7129, {{20737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7130, {{20738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7131, {{20739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7133, {{20740, 1, 0, 0},{20740, 1, 1, 17}, {0, 0, 0, 0}}},{7133, {{20740, 1, 0, 0},{20740, 1, 1, 17}, {0, 0, 0, 0}}},{7134, {{20740, 1, 0, 0},{20740, 1, 1, 18}, {0, 0, 0, 0}}},{7134, {{20740, 1, 0, 0},{20740, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7135, {{20741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7136, {{20742, 1, 0, 0},{20742, 1, 1, 18}, {0, 0, 0, 0}}},{7136, {{20742, 1, 0, 0},{20742, 1, 1, 18}, {0, 0, 0, 0}}},{7137, {{20742, 1, 0, 0},{20742, 1, 1, 17}, {0, 0, 0, 0}}},{7137, {{20742, 1, 0, 0},{20742, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{7138, {{20743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{7139, {{20744, 1, 0, 0},{20744, 1, 1, 17}, {0, 0, 0, 0}}},{7139, {{20744, 1, 0, 0},{20744, 1, 1, 17}, {0, 0, 0, 0}}},{56949, {{20744, 1, 0, 0},{20744, 1, 1, 19}, {0, 0, 0, 0}}},{56949, {{20744, 1, 0, 0},{20744, 1, 1, 19}, {0, 0, 0, 0}}},{59332, {{20744, 1, 0, 0},{20744, 1, 1, 18}, {0, 0, 0, 0}}},{59332, {{20744, 1, 0, 0},{20744, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7140, {{20745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{7141, {{20747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7142, {{20748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7143, {{20749, 1, 0, 0},{20749, 1, 1, 18}, {0, 0, 0, 0}}},{7143, {{20749, 1, 0, 0},{20749, 1, 1, 18}, {0, 0, 0, 0}}},{7144, {{20749, 1, 0, 0},{20749, 1, 1, 17}, {0, 0, 0, 0}}},{7144, {{20749, 1, 0, 0},{20749, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{7145, {{20750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{7146, {{20752, 1, 0, 0},{20752, 1, 1, 17}, {0, 0, 0, 0}}},{7146, {{20752, 1, 0, 0},{20752, 1, 1, 17}, {0, 0, 0, 0}}},{7147, {{20752, 1, 0, 0},{20752, 1, 1, 18}, {0, 0, 0, 0}}},{7147, {{20752, 1, 0, 0},{20752, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7148, {{20753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7149, {{20754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7150, {{20755, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7151, {{20756, 1, 0, 0},{20756, 1, 1, 17}, {0, 0, 0, 0}}},{7151, {{20756, 1, 0, 0},{20756, 1, 1, 17}, {0, 0, 0, 0}}},{7152, {{20756, 1, 0, 0},{20756, 1, 1, 18}, {0, 0, 0, 0}}},{7152, {{20756, 1, 0, 0},{20756, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7153, {{20757, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7154, {{20758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7155, {{20759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7156, {{20760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7157, {{20761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7158, {{20762, 1, 0, 0},{20762, 1, 1, 18}, {0, 0, 0, 0}}},{7158, {{20762, 1, 0, 0},{20762, 1, 1, 18}, {0, 0, 0, 0}}},{7159, {{20762, 1, 0, 0},{20762, 1, 1, 19}, {0, 0, 0, 0}}},{7159, {{20762, 1, 0, 0},{20762, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{7160, {{20763, 1, 0, 0},{20763, 1, 1, 18}, {0, 0, 0, 0}}},{7160, {{20763, 1, 0, 0},{20763, 1, 1, 18}, {0, 0, 0, 0}}},{7161, {{20763, 1, 0, 0},{20763, 1, 1, 17}, {0, 0, 0, 0}}},{7161, {{20763, 1, 0, 0},{20763, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{7162, {{20764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7163, {{20765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7164, {{20766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7165, {{20767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7166, {{20768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7167, {{20769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7168, {{20770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7169, {{20771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7170, {{20772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7171, {{20773, 1, 0, 0},{20773, 1, 1, 16}, {0, 0, 0, 0}}},{7171, {{20773, 1, 0, 0},{20773, 1, 1, 16}, {0, 0, 0, 0}}},{7172, {{20773, 1, 0, 0},{20773, 1, 1, 17}, {0, 0, 0, 0}}},{7172, {{20773, 1, 0, 0},{20773, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{7174, {{20774, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7175, {{20775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7176, {{20776, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7177, {{20777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{7178, {{20778, 1, 0, 0},{20778, 1, 1, 19}, {0, 0, 0, 0}}},{7178, {{20778, 1, 0, 0},{20778, 1, 1, 19}, {0, 0, 0, 0}}},{7179, {{20778, 1, 0, 0},{20778, 1, 1, 17}, {0, 0, 0, 0}}},{7179, {{20778, 1, 0, 0},{20778, 1, 1, 17}, {0, 0, 0, 0}}},{7180, {{20778, 1, 0, 0},{20778, 1, 1, 18}, {0, 0, 0, 0}}},{7180, {{20778, 1, 0, 0},{20778, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7181, {{20779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7182, {{20780, 1, 0, 0},{20780, 1, 1, 17}, {0, 0, 0, 0}}},{7182, {{20780, 1, 0, 0},{20780, 1, 1, 17}, {0, 0, 0, 0}}},{7183, {{20780, 1, 0, 0},{20780, 1, 1, 18}, {0, 0, 0, 0}}},{7183, {{20780, 1, 0, 0},{20780, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7184, {{20781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7185, {{20782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7186, {{20783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7187, {{20784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7188, {{20785, 1, 0, 0},{20785, 1, 1, 17}, {0, 0, 0, 0}}},{7188, {{20785, 1, 0, 0},{20785, 1, 1, 17}, {0, 0, 0, 0}}},{7189, {{20785, 1, 0, 0},{20785, 1, 1, 18}, {0, 0, 0, 0}}},{7189, {{20785, 1, 0, 0},{20785, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{7190, {{20786, 1, 0, 0},{20786, 1, 1, 18}, {0, 0, 0, 0}}},{7190, {{20786, 1, 0, 0},{20786, 1, 1, 18}, {0, 0, 0, 0}}},{7191, {{20786, 1, 0, 0},{20786, 1, 1, 19}, {0, 0, 0, 0}}},{7191, {{20786, 1, 0, 0},{20786, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{7192, {{20787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7193, {{20788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7194, {{20789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7195, {{20790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7196, {{20791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7197, {{20792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7198, {{20793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7199, {{20794, 1, 0, 0},{20794, 1, 1, 17}, {0, 0, 0, 0}}},{7199, {{20794, 1, 0, 0},{20794, 1, 1, 17}, {0, 0, 0, 0}}},{7200, {{20794, 1, 0, 0},{20794, 1, 1, 18}, {0, 0, 0, 0}}},{7200, {{20794, 1, 0, 0},{20794, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{7201, {{20795, 1, 0, 0},{20795, 1, 1, 17}, {0, 0, 0, 0}}},{7201, {{20795, 1, 0, 0},{20795, 1, 1, 17}, {0, 0, 0, 0}}},{7202, {{20795, 1, 0, 0},{20795, 1, 1, 18}, {0, 0, 0, 0}}},{7202, {{20795, 1, 0, 0},{20795, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{7203, {{20796, 1, 0, 0},{20796, 1, 1, 17}, {0, 0, 0, 0}}},{7203, {{20796, 1, 0, 0},{20796, 1, 1, 17}, {0, 0, 0, 0}}},{56958, {{20796, 1, 0, 0},{20796, 1, 1, 18}, {0, 0, 0, 0}}},{56958, {{20796, 1, 0, 0},{20796, 1, 1, 18}, {0, 0, 0, 0}}},{56960, {{20796, 1, 0, 0},{20796, 1, 1, 20}, {0, 0, 0, 0}}},{56960, {{20796, 1, 0, 0},{20796, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{7204, {{20797, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7205, {{20798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7206, {{20799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{7207, {{20800, 1, 0, 0},{20800, 1, 1, 17}, {0, 0, 0, 0}}},{7207, {{20800, 1, 0, 0},{20800, 1, 1, 17}, {0, 0, 0, 0}}},{7208, {{20800, 1, 0, 0},{20800, 1, 1, 18}, {0, 0, 0, 0}}},{7208, {{20800, 1, 0, 0},{20800, 1, 1, 18}, {0, 0, 0, 0}}},{59294, {{20800, 1, 0, 0},{20800, 1, 1, 19}, {0, 0, 0, 0}}},{59294, {{20800, 1, 0, 0},{20800, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{7209, {{20801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{7210, {{20802, 1, 0, 0},{20802, 1, 1, 18}, {0, 0, 0, 0}}},{7210, {{20802, 1, 0, 0},{20802, 1, 1, 18}, {0, 0, 0, 0}}},{7211, {{20802, 1, 0, 0},{20802, 1, 1, 19}, {0, 0, 0, 0}}},{7211, {{20802, 1, 0, 0},{20802, 1, 1, 19}, {0, 0, 0, 0}}},{7212, {{20802, 1, 0, 0},{20802, 1, 1, 17}, {0, 0, 0, 0}}},{7212, {{20802, 1, 0, 0},{20802, 1, 1, 17}, {0, 0, 0, 0}}},{56966, {{20802, 1, 0, 0},{20802, 1, 1, 20}, {0, 0, 0, 0}}},{56966, {{20802, 1, 0, 0},{20802, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{7213, {{20803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7214, {{20804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7215, {{20805, 1, 0, 0},{20805, 1, 1, 18}, {0, 0, 0, 0}}},{7215, {{20805, 1, 0, 0},{20805, 1, 1, 18}, {0, 0, 0, 0}}},{7216, {{20805, 1, 0, 0},{20805, 1, 1, 17}, {0, 0, 0, 0}}},{7216, {{20805, 1, 0, 0},{20805, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{7217, {{20806, 1, 0, 0},{20806, 1, 1, 18}, {0, 0, 0, 0}}},{7217, {{20806, 1, 0, 0},{20806, 1, 1, 18}, {0, 0, 0, 0}}},{7218, {{20806, 1, 0, 0},{20806, 1, 1, 19}, {0, 0, 0, 0}}},{7218, {{20806, 1, 0, 0},{20806, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{7219, {{20807, 1, 0, 0},{20807, 1, 1, 17}, {0, 0, 0, 0}}},{7219, {{20807, 1, 0, 0},{20807, 1, 1, 17}, {0, 0, 0, 0}}},{56967, {{20807, 1, 0, 0},{20807, 1, 1, 18}, {0, 0, 0, 0}}},{56967, {{20807, 1, 0, 0},{20807, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7220, {{20808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7221, {{20809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{7222, {{20810, 1, 0, 0},{20810, 1, 1, 17}, {0, 0, 0, 0}}},{7222, {{20810, 1, 0, 0},{20810, 1, 1, 17}, {0, 0, 0, 0}}},{7223, {{20810, 1, 0, 0},{20810, 1, 1, 20}, {0, 0, 0, 0}}},{7223, {{20810, 1, 0, 0},{20810, 1, 1, 20}, {0, 0, 0, 0}}},{7224, {{20810, 1, 0, 0},{20810, 1, 1, 18}, {0, 0, 0, 0}}},{7224, {{20810, 1, 0, 0},{20810, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7226, {{20811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7227, {{20812, 1, 0, 0},{20812, 1, 1, 17}, {0, 0, 0, 0}}},{7227, {{20812, 1, 0, 0},{20812, 1, 1, 17}, {0, 0, 0, 0}}},{7228, {{20812, 1, 0, 0},{20812, 1, 1, 18}, {0, 0, 0, 0}}},{7228, {{20812, 1, 0, 0},{20812, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {7, {{7229, {{20813, 1, 0, 0},{20813, 1, 1, 19}, {0, 0, 0, 0}}},{7229, {{20813, 1, 0, 0},{20813, 1, 1, 19}, {0, 0, 0, 0}}},{7230, {{20813, 1, 0, 0},{20813, 1, 1, 18}, {0, 0, 0, 0}}},{7230, {{20813, 1, 0, 0},{20813, 1, 1, 18}, {0, 0, 0, 0}}},{30246, {{20813, 1, 0, 0},{20813, 1, 1, 0},{20813, 1, 1, 20},{64050, 1, 0, 0}, {0, 0, 0, 0}}},{30246, {{20813, 1, 0, 0},{20813, 1, 1, 0},{20813, 1, 1, 20},{64050, 1, 0, 0}, {0, 0, 0, 0}}},{30246, {{20813, 1, 0, 0},{20813, 1, 1, 0},{20813, 1, 1, 20},{64050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7231, {{20814, 1, 0, 0},{20814, 1, 1, 19}, {0, 0, 0, 0}}},{7231, {{20814, 1, 0, 0},{20814, 1, 1, 19}, {0, 0, 0, 0}}},{7232, {{20814, 1, 0, 0},{20814, 1, 1, 18}, {0, 0, 0, 0}}},{7232, {{20814, 1, 0, 0},{20814, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7233, {{20815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7234, {{20816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7235, {{20817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7236, {{20818, 1, 0, 0},{20818, 1, 1, 17}, {0, 0, 0, 0}}},{7236, {{20818, 1, 0, 0},{20818, 1, 1, 17}, {0, 0, 0, 0}}},{7237, {{20818, 1, 0, 0},{20818, 1, 1, 18}, {0, 0, 0, 0}}},{7237, {{20818, 1, 0, 0},{20818, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{7239, {{20819, 1, 0, 0},{20819, 1, 1, 18}, {0, 0, 0, 0}}},{7239, {{20819, 1, 0, 0},{20819, 1, 1, 18}, {0, 0, 0, 0}}},{7240, {{20819, 1, 0, 0},{20819, 1, 1, 17}, {0, 0, 0, 0}}},{7240, {{20819, 1, 0, 0},{20819, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {6, {{7241, {{20820, 1, 0, 0},{20820, 1, 1, 21}, {0, 0, 0, 0}}},{7241, {{20820, 1, 0, 0},{20820, 1, 1, 21}, {0, 0, 0, 0}}},{7242, {{20820, 1, 0, 0},{20820, 1, 1, 19}, {0, 0, 0, 0}}},{7242, {{20820, 1, 0, 0},{20820, 1, 1, 19}, {0, 0, 0, 0}}},{7243, {{20820, 1, 0, 0},{20820, 1, 1, 20}, {0, 0, 0, 0}}},{7243, {{20820, 1, 0, 0},{20820, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{7244, {{20821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7245, {{20822, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7246, {{20823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7247, {{20824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7248, {{20825, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7249, {{20826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7250, {{20827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7251, {{20828, 1, 0, 0},{20828, 1, 1, 17}, {0, 0, 0, 0}}},{7251, {{20828, 1, 0, 0},{20828, 1, 1, 17}, {0, 0, 0, 0}}},{57036, {{20828, 1, 0, 0},{20828, 1, 1, 18}, {0, 0, 0, 0}}},{57036, {{20828, 1, 0, 0},{20828, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7252, {{20829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7253, {{20830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7254, {{20831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7255, {{20832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7256, {{20833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7257, {{20834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7258, {{20835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7259, {{20836, 1, 0, 0},{20836, 1, 1, 17}, {0, 0, 0, 0}}},{7259, {{20836, 1, 0, 0},{20836, 1, 1, 17}, {0, 0, 0, 0}}},{7260, {{20836, 1, 0, 0},{20836, 1, 1, 19}, {0, 0, 0, 0}}},{7260, {{20836, 1, 0, 0},{20836, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{7264, {{20837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7265, {{20838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7266, {{20839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7267, {{20840, 1, 0, 0},{20840, 1, 1, 19}, {0, 0, 0, 0}}},{7267, {{20840, 1, 0, 0},{20840, 1, 1, 19}, {0, 0, 0, 0}}},{7268, {{20840, 1, 0, 0},{20840, 1, 1, 18}, {0, 0, 0, 0}}},{7268, {{20840, 1, 0, 0},{20840, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{7269, {{20841, 1, 0, 0},{20841, 1, 1, 18}, {0, 0, 0, 0}}},{7269, {{20841, 1, 0, 0},{20841, 1, 1, 18}, {0, 0, 0, 0}}},{7270, {{20841, 1, 0, 0},{20841, 1, 1, 17}, {0, 0, 0, 0}}},{7270, {{20841, 1, 0, 0},{20841, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{7271, {{20842, 1, 0, 0},{20842, 1, 1, 17}, {0, 0, 0, 0}}},{7271, {{20842, 1, 0, 0},{20842, 1, 1, 17}, {0, 0, 0, 0}}},{7272, {{20842, 1, 0, 0},{20842, 1, 1, 18}, {0, 0, 0, 0}}},{7272, {{20842, 1, 0, 0},{20842, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{7273, {{20843, 1, 0, 0},{20843, 1, 1, 18}, {0, 0, 0, 0}}},{7273, {{20843, 1, 0, 0},{20843, 1, 1, 18}, {0, 0, 0, 0}}},{7274, {{20843, 1, 0, 0},{20843, 1, 1, 19}, {0, 0, 0, 0}}},{7274, {{20843, 1, 0, 0},{20843, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{7275, {{20844, 1, 0, 0},{20844, 1, 1, 18}, {0, 0, 0, 0}}},{7275, {{20844, 1, 0, 0},{20844, 1, 1, 18}, {0, 0, 0, 0}}},{7276, {{20844, 1, 0, 0},{20844, 1, 1, 19}, {0, 0, 0, 0}}},{7276, {{20844, 1, 0, 0},{20844, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{7277, {{20845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7278, {{20846, 1, 0, 0},{20846, 1, 1, 17}, {0, 0, 0, 0}}},{7278, {{20846, 1, 0, 0},{20846, 1, 1, 17}, {0, 0, 0, 0}}},{7279, {{20846, 1, 0, 0},{20846, 1, 1, 18}, {0, 0, 0, 0}}},{7279, {{20846, 1, 0, 0},{20846, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{7281, {{20849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7282, {{20850, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7283, {{20851, 1, 0, 0},{20851, 1, 1, 18}, {0, 0, 0, 0}}},{7283, {{20851, 1, 0, 0},{20851, 1, 1, 18}, {0, 0, 0, 0}}},{7284, {{20851, 1, 0, 0},{20851, 1, 1, 17}, {0, 0, 0, 0}}},{7284, {{20851, 1, 0, 0},{20851, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{7285, {{20852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7286, {{20853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7287, {{20854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{7288, {{20855, 1, 0, 0},{20855, 1, 1, 18}, {0, 0, 0, 0}}},{7288, {{20855, 1, 0, 0},{20855, 1, 1, 18}, {0, 0, 0, 0}}},{7289, {{20855, 1, 0, 0},{20855, 1, 1, 20}, {0, 0, 0, 0}}},{7289, {{20855, 1, 0, 0},{20855, 1, 1, 20}, {0, 0, 0, 0}}},{7290, {{20855, 1, 0, 0},{20855, 1, 1, 19}, {0, 0, 0, 0}}},{7290, {{20855, 1, 0, 0},{20855, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{7291, {{20856, 1, 0, 0},{20856, 1, 1, 17}, {0, 0, 0, 0}}},{7291, {{20856, 1, 0, 0},{20856, 1, 1, 17}, {0, 0, 0, 0}}},{56983, {{20856, 1, 0, 0},{20856, 1, 1, 18}, {0, 0, 0, 0}}},{56983, {{20856, 1, 0, 0},{20856, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7292, {{20857, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{68024, {{20858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7293, {{20859, 1, 0, 0},{20859, 1, 1, 17}, {0, 0, 0, 0}}},{7293, {{20859, 1, 0, 0},{20859, 1, 1, 17}, {0, 0, 0, 0}}},{7294, {{20859, 1, 0, 0},{20859, 1, 1, 18}, {0, 0, 0, 0}}},{7294, {{20859, 1, 0, 0},{20859, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {10, {{7296, {{20860, 1, 0, 0},{20860, 1, 1, 20}, {0, 0, 0, 0}}},{7296, {{20860, 1, 0, 0},{20860, 1, 1, 20}, {0, 0, 0, 0}}},{7297, {{20860, 1, 0, 0},{20860, 1, 1, 18}, {0, 0, 0, 0}}},{7297, {{20860, 1, 0, 0},{20860, 1, 1, 18}, {0, 0, 0, 0}}},{7298, {{20860, 1, 0, 0},{20860, 1, 1, 19}, {0, 0, 0, 0}}},{7298, {{20860, 1, 0, 0},{20860, 1, 1, 19}, {0, 0, 0, 0}}},{56985, {{20860, 1, 0, 0},{20860, 1, 1, 21}, {0, 0, 0, 0}}},{56985, {{20860, 1, 0, 0},{20860, 1, 1, 21}, {0, 0, 0, 0}}},{56989, {{20860, 1, 0, 0},{20860, 1, 1, 22}, {0, 0, 0, 0}}},{56989, {{20860, 1, 0, 0},{20860, 1, 1, 22}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{7299, {{20862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{7300, {{20864, 1, 0, 0},{20864, 1, 1, 17}, {0, 0, 0, 0}}},{7300, {{20864, 1, 0, 0},{20864, 1, 1, 17}, {0, 0, 0, 0}}},{56999, {{20864, 1, 0, 0},{20864, 1, 1, 18}, {0, 0, 0, 0}}},{56999, {{20864, 1, 0, 0},{20864, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{7301, {{20866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7302, {{20867, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7303, {{20868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7304, {{20869, 1, 0, 0},{20869, 1, 1, 17}, {0, 0, 0, 0}}},{7304, {{20869, 1, 0, 0},{20869, 1, 1, 17}, {0, 0, 0, 0}}},{57000, {{20869, 1, 0, 0},{20869, 1, 1, 18}, {0, 0, 0, 0}}},{57000, {{20869, 1, 0, 0},{20869, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7305, {{20870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{7306, {{20872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7307, {{20873, 1, 0, 0},{20873, 1, 1, 19}, {0, 0, 0, 0}}},{7307, {{20873, 1, 0, 0},{20873, 1, 1, 19}, {0, 0, 0, 0}}},{7308, {{20873, 1, 0, 0},{20873, 1, 1, 18}, {0, 0, 0, 0}}},{7308, {{20873, 1, 0, 0},{20873, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7309, {{20874, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7310, {{20875, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7311, {{20876, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{7312, {{20877, 1, 0, 0},{20877, 1, 1, 19}, {0, 0, 0, 0}}},{7312, {{20877, 1, 0, 0},{20877, 1, 1, 19}, {0, 0, 0, 0}}},{7313, {{20877, 1, 0, 0},{20877, 1, 1, 18}, {0, 0, 0, 0}}},{7313, {{20877, 1, 0, 0},{20877, 1, 1, 18}, {0, 0, 0, 0}}},{57004, {{20877, 1, 0, 0},{20877, 1, 1, 20}, {0, 0, 0, 0}}},{57004, {{20877, 1, 0, 0},{20877, 1, 1, 20}, {0, 0, 0, 0}}},{57010, {{20877, 1, 0, 0},{20877, 1, 1, 21}, {0, 0, 0, 0}}},{57010, {{20877, 1, 0, 0},{20877, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {6, {{7314, {{20878, 1, 0, 0},{20878, 1, 1, 17}, {0, 0, 0, 0}}},{7314, {{20878, 1, 0, 0},{20878, 1, 1, 17}, {0, 0, 0, 0}}},{57006, {{20878, 1, 0, 0},{20878, 1, 1, 18}, {0, 0, 0, 0}}},{57006, {{20878, 1, 0, 0},{20878, 1, 1, 18}, {0, 0, 0, 0}}},{57007, {{20878, 1, 0, 0},{20878, 1, 1, 19}, {0, 0, 0, 0}}},{57007, {{20878, 1, 0, 0},{20878, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{7315, {{20879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7316, {{20880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7317, {{20881, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7318, {{20882, 1, 0, 0},{20882, 1, 1, 19}, {0, 0, 0, 0}}},{7318, {{20882, 1, 0, 0},{20882, 1, 1, 19}, {0, 0, 0, 0}}},{7319, {{20882, 1, 0, 0},{20882, 1, 1, 18}, {0, 0, 0, 0}}},{7319, {{20882, 1, 0, 0},{20882, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {8, {{7320, {{20883, 1, 0, 0},{20883, 1, 1, 19}, {0, 0, 0, 0}}},{7320, {{20883, 1, 0, 0},{20883, 1, 1, 19}, {0, 0, 0, 0}}},{7321, {{20883, 1, 0, 0},{20883, 1, 1, 18}, {0, 0, 0, 0}}},{7321, {{20883, 1, 0, 0},{20883, 1, 1, 18}, {0, 0, 0, 0}}},{7322, {{20883, 1, 0, 0},{20883, 1, 1, 20}, {0, 0, 0, 0}}},{7322, {{20883, 1, 0, 0},{20883, 1, 1, 20}, {0, 0, 0, 0}}},{57016, {{20883, 1, 0, 0},{20883, 1, 1, 21}, {0, 0, 0, 0}}},{57016, {{20883, 1, 0, 0},{20883, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{7323, {{20884, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{7324, {{20885, 1, 0, 0},{20885, 1, 1, 19}, {0, 0, 0, 0}}},{7324, {{20885, 1, 0, 0},{20885, 1, 1, 19}, {0, 0, 0, 0}}},{7325, {{20885, 1, 0, 0},{20885, 1, 1, 21}, {0, 0, 0, 0}}},{7325, {{20885, 1, 0, 0},{20885, 1, 1, 21}, {0, 0, 0, 0}}},{7326, {{20885, 1, 0, 0},{20885, 1, 1, 20}, {0, 0, 0, 0}}},{7326, {{20885, 1, 0, 0},{20885, 1, 1, 20}, {0, 0, 0, 0}}},{7327, {{20885, 1, 0, 0},{20885, 1, 1, 22}, {0, 0, 0, 0}}},{7327, {{20885, 1, 0, 0},{20885, 1, 1, 22}, {0, 0, 0, 0}}}}},
    {1, {{7328, {{20886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7329, {{20887, 1, 0, 0},{20887, 1, 1, 17}, {0, 0, 0, 0}}},{7329, {{20887, 1, 0, 0},{20887, 1, 1, 17}, {0, 0, 0, 0}}},{7330, {{20887, 1, 0, 0},{20887, 1, 1, 18}, {0, 0, 0, 0}}},{7330, {{20887, 1, 0, 0},{20887, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{7331, {{20888, 1, 0, 0},{20888, 1, 1, 17}, {0, 0, 0, 0}}},{7331, {{20888, 1, 0, 0},{20888, 1, 1, 17}, {0, 0, 0, 0}}},{57022, {{20888, 1, 0, 0},{20888, 1, 1, 18}, {0, 0, 0, 0}}},{57022, {{20888, 1, 0, 0},{20888, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7332, {{20889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{7334, {{20893, 1, 0, 0},{20893, 1, 1, 17}, {0, 0, 0, 0}}},{7334, {{20893, 1, 0, 0},{20893, 1, 1, 17}, {0, 0, 0, 0}}},{7335, {{20893, 1, 0, 0},{20893, 1, 1, 18}, {0, 0, 0, 0}}},{7335, {{20893, 1, 0, 0},{20893, 1, 1, 18}, {0, 0, 0, 0}}},{57024, {{20893, 1, 0, 0},{20893, 1, 1, 19}, {0, 0, 0, 0}}},{57024, {{20893, 1, 0, 0},{20893, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{7336, {{20894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7337, {{20895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7338, {{20896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7339, {{20897, 1, 0, 0},{20897, 1, 1, 17}, {0, 0, 0, 0}}},{7339, {{20897, 1, 0, 0},{20897, 1, 1, 17}, {0, 0, 0, 0}}},{7340, {{20897, 1, 0, 0},{20897, 1, 1, 18}, {0, 0, 0, 0}}},{7340, {{20897, 1, 0, 0},{20897, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {10, {{7341, {{20898, 1, 0, 0},{20898, 1, 1, 18}, {0, 0, 0, 0}}},{7341, {{20898, 1, 0, 0},{20898, 1, 1, 18}, {0, 0, 0, 0}}},{7342, {{20898, 1, 0, 0},{20898, 1, 1, 17}, {0, 0, 0, 0}}},{7342, {{20898, 1, 0, 0},{20898, 1, 1, 17}, {0, 0, 0, 0}}},{7343, {{20898, 1, 0, 0},{20898, 1, 1, 19}, {0, 0, 0, 0}}},{7343, {{20898, 1, 0, 0},{20898, 1, 1, 19}, {0, 0, 0, 0}}},{7344, {{20898, 1, 0, 0},{20898, 1, 1, 20}, {0, 0, 0, 0}}},{7344, {{20898, 1, 0, 0},{20898, 1, 1, 20}, {0, 0, 0, 0}}},{68058, {{20898, 1, 0, 0},{20898, 1, 1, 21}, {0, 0, 0, 0}}},{68058, {{20898, 1, 0, 0},{20898, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{7345, {{20899, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {10, {{7346, {{20900, 1, 0, 0},{20900, 1, 1, 19}, {0, 0, 0, 0}}},{7346, {{20900, 1, 0, 0},{20900, 1, 1, 19}, {0, 0, 0, 0}}},{7347, {{20900, 1, 0, 0},{20900, 1, 1, 18}, {0, 0, 0, 0}}},{7347, {{20900, 1, 0, 0},{20900, 1, 1, 18}, {0, 0, 0, 0}}},{7348, {{20900, 1, 0, 0},{20900, 1, 1, 21}, {0, 0, 0, 0}}},{7348, {{20900, 1, 0, 0},{20900, 1, 1, 21}, {0, 0, 0, 0}}},{7349, {{20900, 1, 0, 0},{20900, 1, 1, 20}, {0, 0, 0, 0}}},{7349, {{20900, 1, 0, 0},{20900, 1, 1, 20}, {0, 0, 0, 0}}},{68059, {{20900, 1, 0, 0},{20900, 1, 1, 22}, {0, 0, 0, 0}}},{68059, {{20900, 1, 0, 0},{20900, 1, 1, 22}, {0, 0, 0, 0}}}}},
    {4, {{7350, {{20901, 1, 0, 0},{20901, 1, 1, 17}, {0, 0, 0, 0}}},{7350, {{20901, 1, 0, 0},{20901, 1, 1, 17}, {0, 0, 0, 0}}},{57027, {{20901, 1, 0, 0},{20901, 1, 1, 18}, {0, 0, 0, 0}}},{57027, {{20901, 1, 0, 0},{20901, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{7351, {{20902, 1, 0, 0},{20902, 1, 1, 17}, {0, 0, 0, 0}}},{7351, {{20902, 1, 0, 0},{20902, 1, 1, 17}, {0, 0, 0, 0}}},{7352, {{20902, 1, 0, 0},{20902, 1, 1, 18}, {0, 0, 0, 0}}},{7352, {{20902, 1, 0, 0},{20902, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{7353, {{20904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7354, {{20905, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7355, {{20906, 1, 0, 0},{20906, 1, 1, 18}, {0, 0, 0, 0}}},{7355, {{20906, 1, 0, 0},{20906, 1, 1, 18}, {0, 0, 0, 0}}},{7356, {{20906, 1, 0, 0},{20906, 1, 1, 17}, {0, 0, 0, 0}}},{7356, {{20906, 1, 0, 0},{20906, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{7357, {{20907, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7358, {{20908, 1, 0, 0},{20908, 1, 1, 19}, {0, 0, 0, 0}}},{7358, {{20908, 1, 0, 0},{20908, 1, 1, 19}, {0, 0, 0, 0}}},{7359, {{20908, 1, 0, 0},{20908, 1, 1, 18}, {0, 0, 0, 0}}},{7359, {{20908, 1, 0, 0},{20908, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7360, {{20909, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7361, {{20910, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{7363, {{20912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7364, {{20913, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7365, {{20914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7366, {{20915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{7367, {{20916, 1, 0, 0},{20916, 1, 1, 20}, {0, 0, 0, 0}}},{7367, {{20916, 1, 0, 0},{20916, 1, 1, 20}, {0, 0, 0, 0}}},{7368, {{20916, 1, 0, 0},{20916, 1, 1, 21}, {0, 0, 0, 0}}},{7368, {{20916, 1, 0, 0},{20916, 1, 1, 21}, {0, 0, 0, 0}}},{7369, {{20916, 1, 0, 0},{20916, 1, 1, 19}, {0, 0, 0, 0}}},{7369, {{20916, 1, 0, 0},{20916, 1, 1, 19}, {0, 0, 0, 0}}},{7370, {{20916, 1, 0, 0},{20916, 1, 1, 22}, {0, 0, 0, 0}}},{7370, {{20916, 1, 0, 0},{20916, 1, 1, 22}, {0, 0, 0, 0}}}}},
    {4, {{7371, {{20917, 1, 0, 0},{20917, 1, 1, 18}, {0, 0, 0, 0}}},{7371, {{20917, 1, 0, 0},{20917, 1, 1, 18}, {0, 0, 0, 0}}},{7372, {{20917, 1, 0, 0},{20917, 1, 1, 17}, {0, 0, 0, 0}}},{7372, {{20917, 1, 0, 0},{20917, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{7373, {{20918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7374, {{20919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7375, {{20920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7376, {{20921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7377, {{20922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{7378, {{20924, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7379, {{20925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7380, {{20926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7381, {{20927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{7382, {{20929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7383, {{20930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7384, {{20931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7385, {{20932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7386, {{20933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7387, {{20934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7388, {{20935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7389, {{20936, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7390, {{20937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7391, {{20938, 1, 0, 0},{20938, 1, 1, 17}, {0, 0, 0, 0}}},{7391, {{20938, 1, 0, 0},{20938, 1, 1, 17}, {0, 0, 0, 0}}},{7392, {{20938, 1, 0, 0},{20938, 1, 1, 18}, {0, 0, 0, 0}}},{7392, {{20938, 1, 0, 0},{20938, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{7393, {{20939, 1, 0, 0},{20939, 1, 1, 19}, {0, 0, 0, 0}}},{7393, {{20939, 1, 0, 0},{20939, 1, 1, 19}, {0, 0, 0, 0}}},{7394, {{20939, 1, 0, 0},{20939, 1, 1, 18}, {0, 0, 0, 0}}},{7394, {{20939, 1, 0, 0},{20939, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{7395, {{20940, 1, 0, 0},{20940, 1, 1, 18}, {0, 0, 0, 0}}},{7395, {{20940, 1, 0, 0},{20940, 1, 1, 18}, {0, 0, 0, 0}}},{7396, {{20940, 1, 0, 0},{20940, 1, 1, 17}, {0, 0, 0, 0}}},{7396, {{20940, 1, 0, 0},{20940, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{7397, {{20941, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{7398, {{20943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7399, {{20944, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7400, {{20945, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7401, {{20946, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7402, {{20947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7403, {{20948, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7404, {{20949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7405, {{20950, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7406, {{20951, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7407, {{20952, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7408, {{20953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7409, {{20954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7410, {{20955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7411, {{20956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7412, {{20957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {13, {{7413, {{20958, 1, 0, 0},{20958, 1, 1, 22}, {0, 0, 0, 0}}},{7413, {{20958, 1, 0, 0},{20958, 1, 1, 22}, {0, 0, 0, 0}}},{7414, {{20958, 1, 0, 0},{20958, 1, 1, 23}, {0, 0, 0, 0}}},{7414, {{20958, 1, 0, 0},{20958, 1, 1, 23}, {0, 0, 0, 0}}},{30203, {{20958, 1, 0, 0},{20958, 1, 1, 24}, {0, 0, 0, 0}}},{30203, {{20958, 1, 0, 0},{20958, 1, 1, 24}, {0, 0, 0, 0}}},{30204, {{20958, 1, 0, 0},{20958, 1, 1, 25}, {0, 0, 0, 0}}},{30204, {{20958, 1, 0, 0},{20958, 1, 1, 25}, {0, 0, 0, 0}}},{30205, {{20958, 1, 0, 0},{20958, 1, 1, 0},{20958, 1, 1, 26},{64021, 1, 0, 0}, {0, 0, 0, 0}}},{30205, {{20958, 1, 0, 0},{20958, 1, 1, 0},{20958, 1, 1, 26},{64021, 1, 0, 0}, {0, 0, 0, 0}}},{30205, {{20958, 1, 0, 0},{20958, 1, 1, 0},{20958, 1, 1, 26},{64021, 1, 0, 0}, {0, 0, 0, 0}}},{59370, {{20958, 1, 0, 0},{20958, 1, 1, 27}, {0, 0, 0, 0}}},{59370, {{20958, 1, 0, 0},{20958, 1, 1, 27}, {0, 0, 0, 0}}}}},
    {1, {{7415, {{20959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7416, {{20960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7417, {{20961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7418, {{20962, 1, 0, 0},{20962, 1, 1, 17}, {0, 0, 0, 0}}},{7418, {{20962, 1, 0, 0},{20962, 1, 1, 17}, {0, 0, 0, 0}}},{7419, {{20962, 1, 0, 0},{20962, 1, 1, 18}, {0, 0, 0, 0}}},{7419, {{20962, 1, 0, 0},{20962, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{7420, {{20964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7421, {{20965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7422, {{20966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7423, {{20967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{7424, {{20969, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7425, {{20970, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{7426, {{20972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7427, {{20973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7428, {{20974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{7430, {{20976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7431, {{20977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7432, {{20978, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7434, {{20979, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7435, {{20980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7436, {{20981, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7438, {{20982, 1, 0, 0},{20982, 1, 1, 17}, {0, 0, 0, 0}}},{7438, {{20982, 1, 0, 0},{20982, 1, 1, 17}, {0, 0, 0, 0}}},{57037, {{20982, 1, 0, 0},{20982, 1, 1, 18}, {0, 0, 0, 0}}},{57037, {{20982, 1, 0, 0},{20982, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7439, {{20983, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7440, {{20984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7441, {{20985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7442, {{20986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{7443, {{20989, 1, 0, 0},{20989, 1, 1, 18}, {0, 0, 0, 0}}},{7443, {{20989, 1, 0, 0},{20989, 1, 1, 18}, {0, 0, 0, 0}}},{7444, {{20989, 1, 0, 0},{20989, 1, 1, 19}, {0, 0, 0, 0}}},{7444, {{20989, 1, 0, 0},{20989, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{7445, {{20990, 1, 0, 0},{20990, 1, 1, 17}, {0, 0, 0, 0}}},{7445, {{20990, 1, 0, 0},{20990, 1, 1, 17}, {0, 0, 0, 0}}},{7446, {{20990, 1, 0, 0},{20990, 1, 1, 18}, {0, 0, 0, 0}}},{7446, {{20990, 1, 0, 0},{20990, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{7447, {{20992, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7448, {{20993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7449, {{20994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7450, {{20995, 1, 0, 0},{20995, 1, 1, 19}, {0, 0, 0, 0}}},{7450, {{20995, 1, 0, 0},{20995, 1, 1, 19}, {0, 0, 0, 0}}},{7451, {{20995, 1, 0, 0},{20995, 1, 1, 18}, {0, 0, 0, 0}}},{7451, {{20995, 1, 0, 0},{20995, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{7452, {{20996, 1, 0, 0},{20996, 1, 1, 17}, {0, 0, 0, 0}}},{7452, {{20996, 1, 0, 0},{20996, 1, 1, 17}, {0, 0, 0, 0}}},{7453, {{20996, 1, 0, 0},{20996, 1, 1, 18}, {0, 0, 0, 0}}},{7453, {{20996, 1, 0, 0},{20996, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7454, {{20997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7455, {{20998, 1, 0, 0},{20998, 1, 1, 18}, {0, 0, 0, 0}}},{7455, {{20998, 1, 0, 0},{20998, 1, 1, 18}, {0, 0, 0, 0}}},{7456, {{20998, 1, 0, 0},{20998, 1, 1, 19}, {0, 0, 0, 0}}},{7456, {{20998, 1, 0, 0},{20998, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {8, {{7457, {{20999, 1, 0, 0},{20999, 1, 1, 17}, {0, 0, 0, 0}}},{7457, {{20999, 1, 0, 0},{20999, 1, 1, 17}, {0, 0, 0, 0}}},{7459, {{20999, 1, 0, 0},{20999, 1, 1, 18}, {0, 0, 0, 0}}},{7459, {{20999, 1, 0, 0},{20999, 1, 1, 18}, {0, 0, 0, 0}}},{7460, {{20999, 1, 0, 0},{20999, 1, 1, 19}, {0, 0, 0, 0}}},{7460, {{20999, 1, 0, 0},{20999, 1, 1, 19}, {0, 0, 0, 0}}},{57061, {{20999, 1, 0, 0},{20999, 1, 1, 20}, {0, 0, 0, 0}}},{57061, {{20999, 1, 0, 0},{20999, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{7461, {{21000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7462, {{21001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7463, {{21002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7464, {{21003, 1, 0, 0},{21003, 1, 1, 17}, {0, 0, 0, 0}}},{7464, {{21003, 1, 0, 0},{21003, 1, 1, 17}, {0, 0, 0, 0}}},{7465, {{21003, 1, 0, 0},{21003, 1, 1, 18}, {0, 0, 0, 0}}},{7465, {{21003, 1, 0, 0},{21003, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7466, {{21004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7467, {{21005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7468, {{21006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7469, {{21007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7470, {{21008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7471, {{21009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7472, {{21010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7473, {{21011, 1, 0, 0},{21011, 1, 1, 17}, {0, 0, 0, 0}}},{7473, {{21011, 1, 0, 0},{21011, 1, 1, 17}, {0, 0, 0, 0}}},{7474, {{21011, 1, 0, 0},{21011, 1, 1, 18}, {0, 0, 0, 0}}},{7474, {{21011, 1, 0, 0},{21011, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7475, {{21012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7476, {{21013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7477, {{21014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7478, {{21015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7479, {{21016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{7481, {{21020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7483, {{21021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7484, {{21022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7485, {{21023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7486, {{21024, 1, 0, 0},{21024, 1, 1, 17}, {0, 0, 0, 0}}},{7486, {{21024, 1, 0, 0},{21024, 1, 1, 17}, {0, 0, 0, 0}}},{7487, {{21024, 1, 0, 0},{21024, 1, 1, 16}, {0, 0, 0, 0}}},{7487, {{21024, 1, 0, 0},{21024, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{7488, {{21025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7489, {{21026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7490, {{21027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{7491, {{21028, 1, 0, 0},{21028, 1, 1, 18}, {0, 0, 0, 0}}},{7491, {{21028, 1, 0, 0},{21028, 1, 1, 18}, {0, 0, 0, 0}}},{7492, {{21028, 1, 0, 0},{21028, 1, 1, 20}, {0, 0, 0, 0}}},{7492, {{21028, 1, 0, 0},{21028, 1, 1, 20}, {0, 0, 0, 0}}},{7493, {{21028, 1, 0, 0},{21028, 1, 1, 19}, {0, 0, 0, 0}}},{7493, {{21028, 1, 0, 0},{21028, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{7494, {{21029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7496, {{21030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7497, {{21031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7498, {{21032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7499, {{21033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7500, {{21034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7501, {{21035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57045, {{21037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7502, {{21038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7503, {{21039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7504, {{21040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7505, {{21041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7506, {{21042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7507, {{21043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7508, {{21044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7509, {{21045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7510, {{21046, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7511, {{21047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7512, {{21048, 1, 0, 0},{21048, 1, 1, 18}, {0, 0, 0, 0}}},{7512, {{21048, 1, 0, 0},{21048, 1, 1, 18}, {0, 0, 0, 0}}},{7513, {{21048, 1, 0, 0},{21048, 1, 1, 19}, {0, 0, 0, 0}}},{7513, {{21048, 1, 0, 0},{21048, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{7514, {{21049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7515, {{21050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7516, {{21051, 1, 0, 0},{21051, 1, 1, 17}, {0, 0, 0, 0}}},{7516, {{21051, 1, 0, 0},{21051, 1, 1, 17}, {0, 0, 0, 0}}},{7517, {{21051, 1, 0, 0},{21051, 1, 1, 18}, {0, 0, 0, 0}}},{7517, {{21051, 1, 0, 0},{21051, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7518, {{21052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{7519, {{21054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{7520, {{21059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7521, {{21060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7522, {{21061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7523, {{21062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7524, {{21063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7525, {{21064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7526, {{21065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7527, {{21066, 1, 0, 0},{21066, 1, 1, 19}, {0, 0, 0, 0}}},{7527, {{21066, 1, 0, 0},{21066, 1, 1, 19}, {0, 0, 0, 0}}},{7528, {{21066, 1, 0, 0},{21066, 1, 1, 18}, {0, 0, 0, 0}}},{7528, {{21066, 1, 0, 0},{21066, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7529, {{21067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7530, {{21068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7531, {{21069, 1, 0, 0},{21069, 1, 1, 19}, {0, 0, 0, 0}}},{7531, {{21069, 1, 0, 0},{21069, 1, 1, 19}, {0, 0, 0, 0}}},{7532, {{21069, 1, 0, 0},{21069, 1, 1, 18}, {0, 0, 0, 0}}},{7532, {{21069, 1, 0, 0},{21069, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7533, {{21070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7534, {{21071, 1, 0, 0},{21071, 1, 1, 17}, {0, 0, 0, 0}}},{7534, {{21071, 1, 0, 0},{21071, 1, 1, 17}, {0, 0, 0, 0}}},{7535, {{21071, 1, 0, 0},{21071, 1, 1, 18}, {0, 0, 0, 0}}},{7535, {{21071, 1, 0, 0},{21071, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{7536, {{21074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7537, {{21075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7538, {{21076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7539, {{21077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7540, {{21078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7541, {{21079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7542, {{21080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7543, {{21081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7544, {{21082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7545, {{21083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7546, {{21084, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7547, {{21085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7548, {{21086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7549, {{21087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7550, {{21088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7551, {{21089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7552, {{21090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7553, {{21091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7554, {{21092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7555, {{21093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7556, {{21094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{7557, {{21096, 1, 0, 0},{21096, 1, 1, 16}, {0, 0, 0, 0}}},{7557, {{21096, 1, 0, 0},{21096, 1, 1, 16}, {0, 0, 0, 0}}},{7558, {{21096, 1, 0, 0},{21096, 1, 1, 17}, {0, 0, 0, 0}}},{7558, {{21096, 1, 0, 0},{21096, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{7559, {{21097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7560, {{21098, 1, 0, 0},{21098, 1, 1, 17}, {0, 0, 0, 0}}},{7560, {{21098, 1, 0, 0},{21098, 1, 1, 17}, {0, 0, 0, 0}}},{7561, {{21098, 1, 0, 0},{21098, 1, 1, 18}, {0, 0, 0, 0}}},{7561, {{21098, 1, 0, 0},{21098, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7562, {{21099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7563, {{21100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7564, {{21101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7565, {{21102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7566, {{21103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7567, {{21104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{7568, {{21105, 1, 0, 0},{21105, 1, 1, 20}, {0, 0, 0, 0}}},{7568, {{21105, 1, 0, 0},{21105, 1, 1, 20}, {0, 0, 0, 0}}},{7569, {{21105, 1, 0, 0},{21105, 1, 1, 19}, {0, 0, 0, 0}}},{7569, {{21105, 1, 0, 0},{21105, 1, 1, 19}, {0, 0, 0, 0}}},{7570, {{21105, 1, 0, 0},{21105, 1, 1, 21}, {0, 0, 0, 0}}},{7570, {{21105, 1, 0, 0},{21105, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {6, {{7571, {{21106, 1, 0, 0},{21106, 1, 1, 19}, {0, 0, 0, 0}}},{7571, {{21106, 1, 0, 0},{21106, 1, 1, 19}, {0, 0, 0, 0}}},{7572, {{21106, 1, 0, 0},{21106, 1, 1, 20}, {0, 0, 0, 0}}},{7572, {{21106, 1, 0, 0},{21106, 1, 1, 20}, {0, 0, 0, 0}}},{7573, {{21106, 1, 0, 0},{21106, 1, 1, 21}, {0, 0, 0, 0}}},{7573, {{21106, 1, 0, 0},{21106, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {4, {{7574, {{21107, 1, 0, 0},{21107, 1, 1, 17}, {0, 0, 0, 0}}},{7574, {{21107, 1, 0, 0},{21107, 1, 1, 17}, {0, 0, 0, 0}}},{7575, {{21107, 1, 0, 0},{21107, 1, 1, 18}, {0, 0, 0, 0}}},{7575, {{21107, 1, 0, 0},{21107, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7576, {{21108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7577, {{21109, 1, 0, 0},{21109, 1, 1, 18}, {0, 0, 0, 0}}},{7577, {{21109, 1, 0, 0},{21109, 1, 1, 18}, {0, 0, 0, 0}}},{7578, {{21109, 1, 0, 0},{21109, 1, 1, 19}, {0, 0, 0, 0}}},{7578, {{21109, 1, 0, 0},{21109, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{7579, {{21110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7580, {{21111, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7582, {{21112, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7583, {{21113, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7584, {{21114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7585, {{21115, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7586, {{21116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7587, {{21117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7588, {{21118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7589, {{21119, 1, 0, 0},{21119, 1, 1, 17}, {0, 0, 0, 0}}},{7589, {{21119, 1, 0, 0},{21119, 1, 1, 17}, {0, 0, 0, 0}}},{7590, {{21119, 1, 0, 0},{21119, 1, 1, 18}, {0, 0, 0, 0}}},{7590, {{21119, 1, 0, 0},{21119, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7591, {{21120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7592, {{21121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7593, {{21122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7594, {{21123, 1, 0, 0},{21123, 1, 1, 17}, {0, 0, 0, 0}}},{7594, {{21123, 1, 0, 0},{21123, 1, 1, 17}, {0, 0, 0, 0}}},{7595, {{21123, 1, 0, 0},{21123, 1, 1, 18}, {0, 0, 0, 0}}},{7595, {{21123, 1, 0, 0},{21123, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7596, {{21124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7597, {{21125, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7598, {{21126, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7599, {{21127, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7600, {{21128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{7601, {{21129, 1, 0, 0},{21129, 1, 1, 18}, {0, 0, 0, 0}}},{7601, {{21129, 1, 0, 0},{21129, 1, 1, 18}, {0, 0, 0, 0}}},{7602, {{21129, 1, 0, 0},{21129, 1, 1, 17}, {0, 0, 0, 0}}},{7602, {{21129, 1, 0, 0},{21129, 1, 1, 17}, {0, 0, 0, 0}}},{57057, {{21129, 1, 0, 0},{21129, 1, 1, 19}, {0, 0, 0, 0}}},{57057, {{21129, 1, 0, 0},{21129, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{7603, {{21130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7604, {{21131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7605, {{21132, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7606, {{21133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7607, {{21134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{7608, {{21136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7609, {{21137, 1, 0, 0},{21137, 1, 1, 17}, {0, 0, 0, 0}}},{7609, {{21137, 1, 0, 0},{21137, 1, 1, 17}, {0, 0, 0, 0}}},{7610, {{21137, 1, 0, 0},{21137, 1, 1, 18}, {0, 0, 0, 0}}},{7610, {{21137, 1, 0, 0},{21137, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{7611, {{21138, 1, 0, 0},{21138, 1, 1, 17}, {0, 0, 0, 0}}},{7611, {{21138, 1, 0, 0},{21138, 1, 1, 17}, {0, 0, 0, 0}}},{7612, {{21138, 1, 0, 0},{21138, 1, 1, 18}, {0, 0, 0, 0}}},{7612, {{21138, 1, 0, 0},{21138, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7613, {{21139, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7614, {{21140, 1, 0, 0},{21140, 1, 1, 19}, {0, 0, 0, 0}}},{7614, {{21140, 1, 0, 0},{21140, 1, 1, 19}, {0, 0, 0, 0}}},{7615, {{21140, 1, 0, 0},{21140, 1, 1, 20}, {0, 0, 0, 0}}},{7615, {{21140, 1, 0, 0},{21140, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{7616, {{21141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7617, {{21142, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7618, {{21143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7619, {{21144, 1, 0, 0},{21144, 1, 1, 17}, {0, 0, 0, 0}}},{7619, {{21144, 1, 0, 0},{21144, 1, 1, 17}, {0, 0, 0, 0}}},{7620, {{21144, 1, 0, 0},{21144, 1, 1, 18}, {0, 0, 0, 0}}},{7620, {{21144, 1, 0, 0},{21144, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7621, {{21145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7622, {{21146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7623, {{21147, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7624, {{21148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7625, {{21149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7626, {{21150, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7627, {{21151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7628, {{21152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{7629, {{21155, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7630, {{21156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7631, {{21157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7632, {{21158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7633, {{21159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{7634, {{21161, 1, 0, 0},{21161, 1, 1, 17}, {0, 0, 0, 0}}},{7634, {{21161, 1, 0, 0},{21161, 1, 1, 17}, {0, 0, 0, 0}}},{57064, {{21161, 1, 0, 0},{21161, 1, 1, 18}, {0, 0, 0, 0}}},{57064, {{21161, 1, 0, 0},{21161, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7635, {{21162, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7636, {{21163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7637, {{21164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7638, {{21165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7639, {{21166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7640, {{21167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7641, {{21168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7642, {{21169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{7643, {{21171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7644, {{21172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7645, {{21173, 1, 0, 0},{21173, 1, 1, 17}, {0, 0, 0, 0}}},{7645, {{21173, 1, 0, 0},{21173, 1, 1, 17}, {0, 0, 0, 0}}},{7646, {{21173, 1, 0, 0},{21173, 1, 1, 18}, {0, 0, 0, 0}}},{7646, {{21173, 1, 0, 0},{21173, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7647, {{21174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7648, {{21175, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7649, {{21176, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7650, {{21177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7651, {{21178, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7652, {{21179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7653, {{21180, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7654, {{21181, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7655, {{21182, 1, 0, 0},{21182, 1, 1, 17}, {0, 0, 0, 0}}},{7655, {{21182, 1, 0, 0},{21182, 1, 1, 17}, {0, 0, 0, 0}}},{7656, {{21182, 1, 0, 0},{21182, 1, 1, 18}, {0, 0, 0, 0}}},{7656, {{21182, 1, 0, 0},{21182, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7657, {{21183, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7658, {{21184, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7659, {{21185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7660, {{21186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7662, {{21187, 1, 0, 0},{21187, 1, 1, 17}, {0, 0, 0, 0}}},{7662, {{21187, 1, 0, 0},{21187, 1, 1, 17}, {0, 0, 0, 0}}},{57070, {{21187, 1, 0, 0},{21187, 1, 1, 18}, {0, 0, 0, 0}}},{57070, {{21187, 1, 0, 0},{21187, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7663, {{21188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7664, {{21189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7665, {{21190, 1, 0, 0},{21190, 1, 1, 17}, {0, 0, 0, 0}}},{7665, {{21190, 1, 0, 0},{21190, 1, 1, 17}, {0, 0, 0, 0}}},{7666, {{21190, 1, 0, 0},{21190, 1, 1, 18}, {0, 0, 0, 0}}},{7666, {{21190, 1, 0, 0},{21190, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{7667, {{21191, 1, 0, 0},{21191, 1, 1, 19}, {0, 0, 0, 0}}},{7667, {{21191, 1, 0, 0},{21191, 1, 1, 19}, {0, 0, 0, 0}}},{7668, {{21191, 1, 0, 0},{21191, 1, 1, 18}, {0, 0, 0, 0}}},{7668, {{21191, 1, 0, 0},{21191, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7669, {{21192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{7670, {{21193, 1, 0, 0},{21193, 1, 1, 18}, {0, 0, 0, 0}}},{7670, {{21193, 1, 0, 0},{21193, 1, 1, 18}, {0, 0, 0, 0}}},{30247, {{21193, 1, 0, 0},{21193, 1, 1, 0},{21193, 1, 1, 19},{64051, 1, 0, 0}, {0, 0, 0, 0}}},{30247, {{21193, 1, 0, 0},{21193, 1, 1, 0},{21193, 1, 1, 19},{64051, 1, 0, 0}, {0, 0, 0, 0}}},{30247, {{21193, 1, 0, 0},{21193, 1, 1, 0},{21193, 1, 1, 19},{64051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7671, {{21194, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{7672, {{21196, 1, 0, 0},{21196, 1, 1, 17}, {0, 0, 0, 0}}},{7672, {{21196, 1, 0, 0},{21196, 1, 1, 17}, {0, 0, 0, 0}}},{7673, {{21196, 1, 0, 0},{21196, 1, 1, 18}, {0, 0, 0, 0}}},{7673, {{21196, 1, 0, 0},{21196, 1, 1, 18}, {0, 0, 0, 0}}},{7674, {{21196, 1, 0, 0},{21196, 1, 1, 20}, {0, 0, 0, 0}}},{7674, {{21196, 1, 0, 0},{21196, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{7681, {{21197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7682, {{21198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7684, {{21199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7685, {{21200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7686, {{21201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7687, {{21202, 1, 0, 0},{21202, 1, 1, 17}, {0, 0, 0, 0}}},{7687, {{21202, 1, 0, 0},{21202, 1, 1, 17}, {0, 0, 0, 0}}},{7688, {{21202, 1, 0, 0},{21202, 1, 1, 18}, {0, 0, 0, 0}}},{7688, {{21202, 1, 0, 0},{21202, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7689, {{21203, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7690, {{21204, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7691, {{21205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7692, {{21206, 1, 0, 0},{21206, 1, 1, 17}, {0, 0, 0, 0}}},{7692, {{21206, 1, 0, 0},{21206, 1, 1, 17}, {0, 0, 0, 0}}},{7693, {{21206, 1, 0, 0},{21206, 1, 1, 18}, {0, 0, 0, 0}}},{7693, {{21206, 1, 0, 0},{21206, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{7694, {{21207, 1, 0, 0},{21207, 1, 1, 19}, {0, 0, 0, 0}}},{7694, {{21207, 1, 0, 0},{21207, 1, 1, 19}, {0, 0, 0, 0}}},{7695, {{21207, 1, 0, 0},{21207, 1, 1, 18}, {0, 0, 0, 0}}},{7695, {{21207, 1, 0, 0},{21207, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{7696, {{21208, 1, 0, 0},{21208, 1, 1, 17}, {0, 0, 0, 0}}},{7696, {{21208, 1, 0, 0},{21208, 1, 1, 17}, {0, 0, 0, 0}}},{59386, {{21208, 1, 0, 0},{21208, 1, 1, 18}, {0, 0, 0, 0}}},{59386, {{21208, 1, 0, 0},{21208, 1, 1, 18}, {0, 0, 0, 0}}},{59387, {{21208, 1, 0, 0},{21208, 1, 1, 19}, {0, 0, 0, 0}}},{59387, {{21208, 1, 0, 0},{21208, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{7697, {{21209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{7698, {{21211, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7699, {{21212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7700, {{21213, 1, 0, 0},{21213, 1, 1, 19}, {0, 0, 0, 0}}},{7700, {{21213, 1, 0, 0},{21213, 1, 1, 19}, {0, 0, 0, 0}}},{7701, {{21213, 1, 0, 0},{21213, 1, 1, 18}, {0, 0, 0, 0}}},{7701, {{21213, 1, 0, 0},{21213, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7702, {{21214, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7703, {{21215, 1, 0, 0},{21215, 1, 1, 17}, {0, 0, 0, 0}}},{7703, {{21215, 1, 0, 0},{21215, 1, 1, 17}, {0, 0, 0, 0}}},{7704, {{21215, 1, 0, 0},{21215, 1, 1, 18}, {0, 0, 0, 0}}},{7704, {{21215, 1, 0, 0},{21215, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{7705, {{21216, 1, 0, 0},{21216, 1, 1, 17}, {0, 0, 0, 0}}},{7705, {{21216, 1, 0, 0},{21216, 1, 1, 17}, {0, 0, 0, 0}}},{7706, {{21216, 1, 0, 0},{21216, 1, 1, 18}, {0, 0, 0, 0}}},{7706, {{21216, 1, 0, 0},{21216, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7707, {{21217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7708, {{21218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7709, {{21219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {9, {{7710, {{21220, 1, 0, 0},{21220, 1, 1, 18}, {0, 0, 0, 0}}},{7710, {{21220, 1, 0, 0},{21220, 1, 1, 18}, {0, 0, 0, 0}}},{30248, {{21220, 1, 0, 0},{21220, 1, 1, 0},{21220, 1, 1, 19},{64052, 1, 0, 0}, {0, 0, 0, 0}}},{30248, {{21220, 1, 0, 0},{21220, 1, 1, 0},{21220, 1, 1, 19},{64052, 1, 0, 0}, {0, 0, 0, 0}}},{30248, {{21220, 1, 0, 0},{21220, 1, 1, 0},{21220, 1, 1, 19},{64052, 1, 0, 0}, {0, 0, 0, 0}}},{59388, {{21220, 1, 0, 0},{21220, 1, 1, 20}, {0, 0, 0, 0}}},{59388, {{21220, 1, 0, 0},{21220, 1, 1, 20}, {0, 0, 0, 0}}},{59389, {{21220, 1, 0, 0},{21220, 1, 1, 21}, {0, 0, 0, 0}}},{59389, {{21220, 1, 0, 0},{21220, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{7712, {{21221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7714, {{21222, 1, 0, 0},{21222, 1, 1, 17}, {0, 0, 0, 0}}},{7714, {{21222, 1, 0, 0},{21222, 1, 1, 17}, {0, 0, 0, 0}}},{7715, {{21222, 1, 0, 0},{21222, 1, 1, 18}, {0, 0, 0, 0}}},{7715, {{21222, 1, 0, 0},{21222, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7716, {{21223, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7717, {{21224, 1, 0, 0},{21224, 1, 1, 17}, {0, 0, 0, 0}}},{7717, {{21224, 1, 0, 0},{21224, 1, 1, 17}, {0, 0, 0, 0}}},{7718, {{21224, 1, 0, 0},{21224, 1, 1, 18}, {0, 0, 0, 0}}},{7718, {{21224, 1, 0, 0},{21224, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7719, {{21225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7720, {{21226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7721, {{21227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7722, {{21228, 1, 0, 0},{21228, 1, 1, 17}, {0, 0, 0, 0}}},{7722, {{21228, 1, 0, 0},{21228, 1, 1, 17}, {0, 0, 0, 0}}},{7723, {{21228, 1, 0, 0},{21228, 1, 1, 18}, {0, 0, 0, 0}}},{7723, {{21228, 1, 0, 0},{21228, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7724, {{21229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7725, {{21230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7726, {{21231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7728, {{21232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7729, {{21233, 1, 0, 0},{21233, 1, 1, 17}, {0, 0, 0, 0}}},{7729, {{21233, 1, 0, 0},{21233, 1, 1, 17}, {0, 0, 0, 0}}},{7730, {{21233, 1, 0, 0},{21233, 1, 1, 18}, {0, 0, 0, 0}}},{7730, {{21233, 1, 0, 0},{21233, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{7731, {{21234, 1, 0, 0},{21234, 1, 1, 17}, {0, 0, 0, 0}}},{7731, {{21234, 1, 0, 0},{21234, 1, 1, 17}, {0, 0, 0, 0}}},{7732, {{21234, 1, 0, 0},{21234, 1, 1, 18}, {0, 0, 0, 0}}},{7732, {{21234, 1, 0, 0},{21234, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{7733, {{21235, 1, 0, 0},{21235, 1, 1, 17}, {0, 0, 0, 0}}},{7733, {{21235, 1, 0, 0},{21235, 1, 1, 17}, {0, 0, 0, 0}}},{7734, {{21235, 1, 0, 0},{21235, 1, 1, 18}, {0, 0, 0, 0}}},{7734, {{21235, 1, 0, 0},{21235, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7735, {{21236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7736, {{21237, 1, 0, 0},{21237, 1, 1, 17}, {0, 0, 0, 0}}},{7736, {{21237, 1, 0, 0},{21237, 1, 1, 17}, {0, 0, 0, 0}}},{7737, {{21237, 1, 0, 0},{21237, 1, 1, 18}, {0, 0, 0, 0}}},{7737, {{21237, 1, 0, 0},{21237, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7738, {{21238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7739, {{21239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7740, {{21240, 1, 0, 0},{21240, 1, 1, 17}, {0, 0, 0, 0}}},{7740, {{21240, 1, 0, 0},{21240, 1, 1, 17}, {0, 0, 0, 0}}},{7741, {{21240, 1, 0, 0},{21240, 1, 1, 18}, {0, 0, 0, 0}}},{7741, {{21240, 1, 0, 0},{21240, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{7742, {{21241, 1, 0, 0},{21241, 1, 1, 17}, {0, 0, 0, 0}}},{7742, {{21241, 1, 0, 0},{21241, 1, 1, 17}, {0, 0, 0, 0}}},{7743, {{21241, 1, 0, 0},{21241, 1, 1, 18}, {0, 0, 0, 0}}},{7743, {{21241, 1, 0, 0},{21241, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{7744, {{21242, 1, 0, 0},{21242, 1, 1, 19}, {0, 0, 0, 0}}},{7744, {{21242, 1, 0, 0},{21242, 1, 1, 19}, {0, 0, 0, 0}}},{7745, {{21242, 1, 0, 0},{21242, 1, 1, 18}, {0, 0, 0, 0}}},{7745, {{21242, 1, 0, 0},{21242, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7746, {{21243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7747, {{21244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7748, {{21245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7749, {{21246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7750, {{21247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7751, {{21248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7752, {{21249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7753, {{21250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7754, {{21251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7755, {{21252, 1, 0, 0},{21252, 1, 1, 16}, {0, 0, 0, 0}}},{7755, {{21252, 1, 0, 0},{21252, 1, 1, 16}, {0, 0, 0, 0}}},{7756, {{21252, 1, 0, 0},{21252, 1, 1, 17}, {0, 0, 0, 0}}},{7756, {{21252, 1, 0, 0},{21252, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{7757, {{21253, 1, 0, 0},{21253, 1, 1, 19}, {0, 0, 0, 0}}},{7757, {{21253, 1, 0, 0},{21253, 1, 1, 19}, {0, 0, 0, 0}}},{7758, {{21253, 1, 0, 0},{21253, 1, 1, 18}, {0, 0, 0, 0}}},{7758, {{21253, 1, 0, 0},{21253, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{7759, {{21254, 1, 0, 0},{21254, 1, 1, 18}, {0, 0, 0, 0}}},{7759, {{21254, 1, 0, 0},{21254, 1, 1, 18}, {0, 0, 0, 0}}},{7760, {{21254, 1, 0, 0},{21254, 1, 1, 17}, {0, 0, 0, 0}}},{7760, {{21254, 1, 0, 0},{21254, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{7761, {{21255, 1, 0, 0},{21255, 1, 1, 18}, {0, 0, 0, 0}}},{7761, {{21255, 1, 0, 0},{21255, 1, 1, 18}, {0, 0, 0, 0}}},{7762, {{21255, 1, 0, 0},{21255, 1, 1, 19}, {0, 0, 0, 0}}},{7762, {{21255, 1, 0, 0},{21255, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {6, {{7763, {{21256, 1, 0, 0},{21256, 1, 1, 17}, {0, 0, 0, 0}}},{7763, {{21256, 1, 0, 0},{21256, 1, 1, 17}, {0, 0, 0, 0}}},{57080, {{21256, 1, 0, 0},{21256, 1, 1, 19}, {0, 0, 0, 0}}},{57080, {{21256, 1, 0, 0},{21256, 1, 1, 19}, {0, 0, 0, 0}}},{57081, {{21256, 1, 0, 0},{21256, 1, 1, 20}, {0, 0, 0, 0}}},{57081, {{21256, 1, 0, 0},{21256, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{7764, {{21257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7765, {{21258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{7766, {{21259, 1, 0, 0},{21259, 1, 1, 17}, {0, 0, 0, 0}}},{7766, {{21259, 1, 0, 0},{21259, 1, 1, 17}, {0, 0, 0, 0}}},{57082, {{21259, 1, 0, 0},{21259, 1, 1, 18}, {0, 0, 0, 0}}},{57082, {{21259, 1, 0, 0},{21259, 1, 1, 18}, {0, 0, 0, 0}}},{57083, {{21259, 1, 0, 0},{21259, 1, 1, 19}, {0, 0, 0, 0}}},{57083, {{21259, 1, 0, 0},{21259, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{7767, {{21260, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7768, {{21261, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7769, {{21262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7770, {{21263, 1, 0, 0},{21263, 1, 1, 17}, {0, 0, 0, 0}}},{7770, {{21263, 1, 0, 0},{21263, 1, 1, 17}, {0, 0, 0, 0}}},{7771, {{21263, 1, 0, 0},{21263, 1, 1, 18}, {0, 0, 0, 0}}},{7771, {{21263, 1, 0, 0},{21263, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7772, {{21264, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7773, {{21265, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7774, {{21266, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7775, {{21267, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7776, {{21268, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7777, {{21269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7778, {{21270, 1, 0, 0},{21270, 1, 1, 19}, {0, 0, 0, 0}}},{7778, {{21270, 1, 0, 0},{21270, 1, 1, 19}, {0, 0, 0, 0}}},{7779, {{21270, 1, 0, 0},{21270, 1, 1, 18}, {0, 0, 0, 0}}},{7779, {{21270, 1, 0, 0},{21270, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{7780, {{21271, 1, 0, 0},{21271, 1, 1, 17}, {0, 0, 0, 0}}},{7780, {{21271, 1, 0, 0},{21271, 1, 1, 17}, {0, 0, 0, 0}}},{7781, {{21271, 1, 0, 0},{21271, 1, 1, 18}, {0, 0, 0, 0}}},{7781, {{21271, 1, 0, 0},{21271, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7782, {{21272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7783, {{21273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7784, {{21274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7785, {{21275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7786, {{21276, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7787, {{21277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7788, {{21278, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7789, {{21279, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7790, {{21280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7791, {{21281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7792, {{21282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7793, {{21283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7794, {{21284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7796, {{21285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{7797, {{21287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7798, {{21288, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7799, {{21289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7800, {{21290, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7801, {{21291, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7802, {{21292, 1, 0, 0},{21292, 1, 1, 17}, {0, 0, 0, 0}}},{7802, {{21292, 1, 0, 0},{21292, 1, 1, 17}, {0, 0, 0, 0}}},{7803, {{21292, 1, 0, 0},{21292, 1, 1, 18}, {0, 0, 0, 0}}},{7803, {{21292, 1, 0, 0},{21292, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7804, {{21293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{7805, {{21295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7806, {{21296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7807, {{21297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7808, {{21298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7809, {{21299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7810, {{21300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7811, {{21301, 1, 0, 0},{21301, 1, 1, 17}, {0, 0, 0, 0}}},{7811, {{21301, 1, 0, 0},{21301, 1, 1, 17}, {0, 0, 0, 0}}},{7812, {{21301, 1, 0, 0},{21301, 1, 1, 18}, {0, 0, 0, 0}}},{7812, {{21301, 1, 0, 0},{21301, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{7813, {{21302, 1, 0, 0},{21302, 1, 1, 16}, {0, 0, 0, 0}}},{7813, {{21302, 1, 0, 0},{21302, 1, 1, 16}, {0, 0, 0, 0}}},{7814, {{21302, 1, 0, 0},{21302, 1, 1, 17}, {0, 0, 0, 0}}},{7814, {{21302, 1, 0, 0},{21302, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{7815, {{21303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7816, {{21304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7817, {{21305, 1, 0, 0},{21305, 1, 1, 19}, {0, 0, 0, 0}}},{7817, {{21305, 1, 0, 0},{21305, 1, 1, 19}, {0, 0, 0, 0}}},{7818, {{21305, 1, 0, 0},{21305, 1, 1, 18}, {0, 0, 0, 0}}},{7818, {{21305, 1, 0, 0},{21305, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7819, {{21306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7820, {{21307, 1, 0, 0},{21307, 1, 1, 17}, {0, 0, 0, 0}}},{7820, {{21307, 1, 0, 0},{21307, 1, 1, 17}, {0, 0, 0, 0}}},{7821, {{21307, 1, 0, 0},{21307, 1, 1, 18}, {0, 0, 0, 0}}},{7821, {{21307, 1, 0, 0},{21307, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{7822, {{21308, 1, 0, 0},{21308, 1, 1, 17}, {0, 0, 0, 0}}},{7822, {{21308, 1, 0, 0},{21308, 1, 1, 17}, {0, 0, 0, 0}}},{7823, {{21308, 1, 0, 0},{21308, 1, 1, 18}, {0, 0, 0, 0}}},{7823, {{21308, 1, 0, 0},{21308, 1, 1, 18}, {0, 0, 0, 0}}},{7824, {{21308, 1, 0, 0},{21308, 1, 1, 19}, {0, 0, 0, 0}}},{7824, {{21308, 1, 0, 0},{21308, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{7825, {{21309, 1, 0, 0},{21309, 1, 1, 17}, {0, 0, 0, 0}}},{7825, {{21309, 1, 0, 0},{21309, 1, 1, 17}, {0, 0, 0, 0}}},{7826, {{21309, 1, 0, 0},{21309, 1, 1, 18}, {0, 0, 0, 0}}},{7826, {{21309, 1, 0, 0},{21309, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{7827, {{21310, 1, 0, 0},{21310, 1, 1, 17}, {0, 0, 0, 0}}},{7827, {{21310, 1, 0, 0},{21310, 1, 1, 17}, {0, 0, 0, 0}}},{7828, {{21310, 1, 0, 0},{21310, 1, 1, 18}, {0, 0, 0, 0}}},{7828, {{21310, 1, 0, 0},{21310, 1, 1, 18}, {0, 0, 0, 0}}},{7829, {{21310, 1, 0, 0},{21310, 1, 1, 19}, {0, 0, 0, 0}}},{7829, {{21310, 1, 0, 0},{21310, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{7830, {{21311, 1, 0, 0},{21311, 1, 1, 18}, {0, 0, 0, 0}}},{7830, {{21311, 1, 0, 0},{21311, 1, 1, 18}, {0, 0, 0, 0}}},{7831, {{21311, 1, 0, 0},{21311, 1, 1, 19}, {0, 0, 0, 0}}},{7831, {{21311, 1, 0, 0},{21311, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{7832, {{21312, 1, 0, 0},{21312, 1, 1, 18}, {0, 0, 0, 0}}},{7832, {{21312, 1, 0, 0},{21312, 1, 1, 18}, {0, 0, 0, 0}}},{7833, {{21312, 1, 0, 0},{21312, 1, 1, 19}, {0, 0, 0, 0}}},{7833, {{21312, 1, 0, 0},{21312, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{7834, {{21313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7835, {{21314, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7836, {{21315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7837, {{21316, 1, 0, 0},{21316, 1, 1, 16}, {0, 0, 0, 0}}},{7837, {{21316, 1, 0, 0},{21316, 1, 1, 16}, {0, 0, 0, 0}}},{57090, {{21316, 1, 0, 0},{21316, 1, 1, 17}, {0, 0, 0, 0}}},{57090, {{21316, 1, 0, 0},{21316, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{7838, {{21317, 1, 0, 0},{21317, 1, 1, 17}, {0, 0, 0, 0}}},{7838, {{21317, 1, 0, 0},{21317, 1, 1, 17}, {0, 0, 0, 0}}},{57092, {{21317, 1, 0, 0},{21317, 1, 1, 18}, {0, 0, 0, 0}}},{57092, {{21317, 1, 0, 0},{21317, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7839, {{21318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7840, {{21319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7841, {{21320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7842, {{21321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7843, {{21322, 1, 0, 0},{21322, 1, 1, 19}, {0, 0, 0, 0}}},{7843, {{21322, 1, 0, 0},{21322, 1, 1, 19}, {0, 0, 0, 0}}},{7844, {{21322, 1, 0, 0},{21322, 1, 1, 18}, {0, 0, 0, 0}}},{7844, {{21322, 1, 0, 0},{21322, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7845, {{21323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7847, {{21324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7848, {{21325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{7850, {{21328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{7851, {{21329, 1, 0, 0},{21329, 1, 1, 18}, {0, 0, 0, 0}}},{7851, {{21329, 1, 0, 0},{21329, 1, 1, 18}, {0, 0, 0, 0}}},{30249, {{21329, 1, 0, 0},{21329, 1, 1, 0},{21329, 1, 1, 19},{64053, 1, 0, 0}, {0, 0, 0, 0}}},{30249, {{21329, 1, 0, 0},{21329, 1, 1, 0},{21329, 1, 1, 19},{64053, 1, 0, 0}, {0, 0, 0, 0}}},{30249, {{21329, 1, 0, 0},{21329, 1, 1, 0},{21329, 1, 1, 19},{64053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7852, {{21330, 1, 0, 0},{21330, 1, 1, 17}, {0, 0, 0, 0}}},{7852, {{21330, 1, 0, 0},{21330, 1, 1, 17}, {0, 0, 0, 0}}},{7853, {{21330, 1, 0, 0},{21330, 1, 1, 18}, {0, 0, 0, 0}}},{7853, {{21330, 1, 0, 0},{21330, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7854, {{21331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7855, {{21332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{7856, {{21335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7857, {{21336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7858, {{21337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {12, {{7859, {{21338, 1, 0, 0},{21338, 1, 1, 20}, {0, 0, 0, 0}}},{7859, {{21338, 1, 0, 0},{21338, 1, 1, 20}, {0, 0, 0, 0}}},{7860, {{21338, 1, 0, 0},{21338, 1, 1, 18}, {0, 0, 0, 0}}},{7860, {{21338, 1, 0, 0},{21338, 1, 1, 18}, {0, 0, 0, 0}}},{7861, {{21338, 1, 0, 0},{21338, 1, 1, 22}, {0, 0, 0, 0}}},{7861, {{21338, 1, 0, 0},{21338, 1, 1, 22}, {0, 0, 0, 0}}},{7862, {{21338, 1, 0, 0},{21338, 1, 1, 19}, {0, 0, 0, 0}}},{7862, {{21338, 1, 0, 0},{21338, 1, 1, 19}, {0, 0, 0, 0}}},{59398, {{21338, 1, 0, 0},{21338, 1, 1, 21}, {0, 0, 0, 0}}},{59398, {{21338, 1, 0, 0},{21338, 1, 1, 21}, {0, 0, 0, 0}}},{59399, {{21338, 1, 1, 26},{178660, 1, 0, 0},{178660, 1, 1, 16}, {0, 0, 0, 0}}},{59400, {{21338, 1, 1, 27},{178660, 1, 0, 0},{178660, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{7865, {{21339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7866, {{21340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7867, {{21341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7868, {{21342, 1, 0, 0},{21342, 1, 1, 17}, {0, 0, 0, 0}}},{7868, {{21342, 1, 0, 0},{21342, 1, 1, 17}, {0, 0, 0, 0}}},{7869, {{21342, 1, 0, 0},{21342, 1, 1, 18}, {0, 0, 0, 0}}},{7869, {{21342, 1, 0, 0},{21342, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7870, {{21343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7871, {{21344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7872, {{21345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{7874, {{21347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7875, {{21348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7876, {{21349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7877, {{21350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7878, {{21351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7879, {{21352, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7880, {{21353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7881, {{21354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{7882, {{21356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7883, {{21357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7884, {{21358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7885, {{21359, 1, 0, 0},{21359, 1, 1, 17}, {0, 0, 0, 0}}},{7885, {{21359, 1, 0, 0},{21359, 1, 1, 17}, {0, 0, 0, 0}}},{57108, {{21359, 1, 0, 0},{21359, 1, 1, 18}, {0, 0, 0, 0}}},{57108, {{21359, 1, 0, 0},{21359, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7886, {{21360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{7887, {{21361, 1, 0, 0},{21361, 1, 1, 20}, {0, 0, 0, 0}}},{7887, {{21361, 1, 0, 0},{21361, 1, 1, 20}, {0, 0, 0, 0}}},{7888, {{21361, 1, 0, 0},{21361, 1, 1, 18}, {0, 0, 0, 0}}},{7888, {{21361, 1, 0, 0},{21361, 1, 1, 18}, {0, 0, 0, 0}}},{7889, {{21361, 1, 0, 0},{21361, 1, 1, 19}, {0, 0, 0, 0}}},{7889, {{21361, 1, 0, 0},{21361, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{7890, {{21362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7891, {{21363, 1, 0, 0},{21363, 1, 1, 17}, {0, 0, 0, 0}}},{7891, {{21363, 1, 0, 0},{21363, 1, 1, 17}, {0, 0, 0, 0}}},{7892, {{21363, 1, 0, 0},{21363, 1, 1, 18}, {0, 0, 0, 0}}},{7892, {{21363, 1, 0, 0},{21363, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7894, {{21364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7895, {{21365, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7896, {{21366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {10, {{7897, {{21367, 1, 0, 0},{21367, 1, 1, 17}, {0, 0, 0, 0}}},{7897, {{21367, 1, 0, 0},{21367, 1, 1, 17}, {0, 0, 0, 0}}},{7898, {{21367, 1, 0, 0},{21367, 1, 1, 18}, {0, 0, 0, 0}}},{7898, {{21367, 1, 0, 0},{21367, 1, 1, 18}, {0, 0, 0, 0}}},{7899, {{21367, 1, 0, 0},{21367, 1, 1, 22}, {0, 0, 0, 0}}},{7899, {{21367, 1, 0, 0},{21367, 1, 1, 22}, {0, 0, 0, 0}}},{7900, {{21367, 1, 0, 0},{21367, 1, 1, 19}, {0, 0, 0, 0}}},{7900, {{21367, 1, 0, 0},{21367, 1, 1, 19}, {0, 0, 0, 0}}},{57415, {{21367, 1, 0, 0},{21367, 1, 1, 20}, {0, 0, 0, 0}}},{57415, {{21367, 1, 0, 0},{21367, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{7902, {{21368, 1, 0, 0},{21368, 1, 1, 18}, {0, 0, 0, 0}}},{7902, {{21368, 1, 0, 0},{21368, 1, 1, 18}, {0, 0, 0, 0}}},{7903, {{21368, 1, 0, 0},{21368, 1, 1, 19}, {0, 0, 0, 0}}},{7903, {{21368, 1, 0, 0},{21368, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{7904, {{21369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7905, {{21370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7906, {{21371, 1, 0, 0},{21371, 1, 1, 17}, {0, 0, 0, 0}}},{7906, {{21371, 1, 0, 0},{21371, 1, 1, 17}, {0, 0, 0, 0}}},{7907, {{21371, 1, 0, 0},{21371, 1, 1, 18}, {0, 0, 0, 0}}},{7907, {{21371, 1, 0, 0},{21371, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7908, {{21372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7893, {{21373, 1, 0, 0},{21373, 1, 1, 18}, {0, 0, 0, 0}}},{7893, {{21373, 1, 0, 0},{21373, 1, 1, 18}, {0, 0, 0, 0}}},{7909, {{21373, 1, 0, 0},{21373, 1, 1, 17}, {0, 0, 0, 0}}},{7909, {{21373, 1, 0, 0},{21373, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{7910, {{21374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {10, {{7911, {{21375, 1, 0, 0},{21375, 1, 1, 19}, {0, 0, 0, 0}}},{7911, {{21375, 1, 0, 0},{21375, 1, 1, 19}, {0, 0, 0, 0}}},{7912, {{21375, 1, 0, 0},{21375, 1, 1, 23}, {0, 0, 0, 0}}},{7912, {{21375, 1, 0, 0},{21375, 1, 1, 23}, {0, 0, 0, 0}}},{7913, {{21375, 1, 0, 0},{21375, 1, 1, 21}, {0, 0, 0, 0}}},{7913, {{21375, 1, 0, 0},{21375, 1, 1, 21}, {0, 0, 0, 0}}},{7914, {{21375, 1, 0, 0},{21375, 1, 1, 22}, {0, 0, 0, 0}}},{7914, {{21375, 1, 0, 0},{21375, 1, 1, 22}, {0, 0, 0, 0}}},{7915, {{21375, 1, 0, 0},{21375, 1, 1, 20}, {0, 0, 0, 0}}},{7915, {{21375, 1, 0, 0},{21375, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{7916, {{21376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{7917, {{21378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7918, {{21379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7919, {{21380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{7922, {{21383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7923, {{21384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7924, {{21385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7925, {{21386, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{7926, {{21390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7927, {{21391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{7928, {{21394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7929, {{21395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7930, {{21396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{7931, {{21398, 1, 0, 0},{21398, 1, 1, 17}, {0, 0, 0, 0}}},{7931, {{21398, 1, 0, 0},{21398, 1, 1, 17}, {0, 0, 0, 0}}},{7932, {{21398, 1, 0, 0},{21398, 1, 1, 18}, {0, 0, 0, 0}}},{7932, {{21398, 1, 0, 0},{21398, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7933, {{21399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7934, {{21400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7935, {{21401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7936, {{21402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{7937, {{21404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7938, {{21405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7939, {{21406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7940, {{21407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7941, {{21408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7942, {{21409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{7943, {{21412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7944, {{21413, 1, 0, 0},{21413, 1, 1, 17}, {0, 0, 0, 0}}},{7944, {{21413, 1, 0, 0},{21413, 1, 1, 17}, {0, 0, 0, 0}}},{57119, {{21413, 1, 0, 0},{21413, 1, 1, 18}, {0, 0, 0, 0}}},{57119, {{21413, 1, 0, 0},{21413, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7945, {{21414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7946, {{21415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7947, {{21416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {16, {{7948, {{21417, 1, 0, 0},{21417, 1, 1, 22}, {0, 0, 0, 0}}},{7948, {{21417, 1, 0, 0},{21417, 1, 1, 22}, {0, 0, 0, 0}}},{7949, {{21417, 1, 0, 0},{21417, 1, 1, 28}, {0, 0, 0, 0}}},{7949, {{21417, 1, 0, 0},{21417, 1, 1, 28}, {0, 0, 0, 0}}},{7950, {{21417, 1, 0, 0},{21417, 1, 1, 25}, {0, 0, 0, 0}}},{7950, {{21417, 1, 0, 0},{21417, 1, 1, 25}, {0, 0, 0, 0}}},{7951, {{21417, 1, 0, 0},{21417, 1, 1, 24}, {0, 0, 0, 0}}},{7951, {{21417, 1, 0, 0},{21417, 1, 1, 24}, {0, 0, 0, 0}}},{7952, {{21417, 1, 0, 0},{21417, 1, 1, 26}, {0, 0, 0, 0}}},{7952, {{21417, 1, 0, 0},{21417, 1, 1, 26}, {0, 0, 0, 0}}},{7953, {{21417, 1, 0, 0},{21417, 1, 1, 23}, {0, 0, 0, 0}}},{7953, {{21417, 1, 0, 0},{21417, 1, 1, 23}, {0, 0, 0, 0}}},{7954, {{21417, 1, 0, 0},{21417, 1, 1, 29}, {0, 0, 0, 0}}},{7954, {{21417, 1, 0, 0},{21417, 1, 1, 29}, {0, 0, 0, 0}}},{7955, {{21417, 1, 0, 0},{21417, 1, 1, 27}, {0, 0, 0, 0}}},{7955, {{21417, 1, 0, 0},{21417, 1, 1, 27}, {0, 0, 0, 0}}}}},
    {1, {{7956, {{21418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7957, {{21419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7958, {{21420, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7959, {{21421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7960, {{21422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7961, {{21423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7962, {{21424, 1, 0, 0},{21424, 1, 1, 17}, {0, 0, 0, 0}}},{7962, {{21424, 1, 0, 0},{21424, 1, 1, 17}, {0, 0, 0, 0}}},{7963, {{21424, 1, 0, 0},{21424, 1, 1, 18}, {0, 0, 0, 0}}},{7963, {{21424, 1, 0, 0},{21424, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7964, {{21425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7965, {{21426, 1, 0, 0},{21426, 1, 1, 18}, {0, 0, 0, 0}}},{7965, {{21426, 1, 0, 0},{21426, 1, 1, 18}, {0, 0, 0, 0}}},{7966, {{21426, 1, 0, 0},{21426, 1, 1, 17}, {0, 0, 0, 0}}},{7966, {{21426, 1, 0, 0},{21426, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{7967, {{21427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7968, {{21428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7969, {{21429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7970, {{21430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7971, {{21431, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7972, {{21432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7973, {{21433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7974, {{21434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7975, {{21435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{7976, {{21437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7977, {{21438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{7979, {{21440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7980, {{21441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7981, {{21442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7982, {{21443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7983, {{21444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7984, {{21445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{7985, {{21448, 1, 0, 0},{21448, 1, 1, 17}, {0, 0, 0, 0}}},{7985, {{21448, 1, 0, 0},{21448, 1, 1, 17}, {0, 0, 0, 0}}},{7986, {{21448, 1, 0, 0},{21448, 1, 1, 18}, {0, 0, 0, 0}}},{7986, {{21448, 1, 0, 0},{21448, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{7987, {{21449, 1, 0, 0},{21449, 1, 1, 18}, {0, 0, 0, 0}}},{7987, {{21449, 1, 0, 0},{21449, 1, 1, 18}, {0, 0, 0, 0}}},{7988, {{21449, 1, 0, 0},{21449, 1, 1, 19}, {0, 0, 0, 0}}},{7988, {{21449, 1, 0, 0},{21449, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{7989, {{21450, 1, 0, 0},{21450, 1, 1, 19}, {0, 0, 0, 0}}},{7989, {{21450, 1, 0, 0},{21450, 1, 1, 19}, {0, 0, 0, 0}}},{7990, {{21450, 1, 0, 0},{21450, 1, 1, 18}, {0, 0, 0, 0}}},{7990, {{21450, 1, 0, 0},{21450, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{7992, {{21451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7993, {{21452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7994, {{21453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{7995, {{21454, 1, 0, 0},{21454, 1, 1, 18}, {0, 0, 0, 0}}},{7995, {{21454, 1, 0, 0},{21454, 1, 1, 18}, {0, 0, 0, 0}}},{7996, {{21454, 1, 0, 0},{21454, 1, 1, 19}, {0, 0, 0, 0}}},{7996, {{21454, 1, 0, 0},{21454, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{7997, {{21455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{7998, {{21456, 1, 0, 0},{21456, 1, 1, 16}, {0, 0, 0, 0}}},{7998, {{21456, 1, 0, 0},{21456, 1, 1, 16}, {0, 0, 0, 0}}},{56859, {{21456, 1, 0, 0},{21456, 1, 1, 17}, {0, 0, 0, 0}}},{56859, {{21456, 1, 0, 0},{21456, 1, 1, 17}, {0, 0, 0, 0}}},{56861, {{21456, 1, 0, 0},{21456, 1, 1, 18}, {0, 0, 0, 0}}},{56861, {{21456, 1, 0, 0},{21456, 1, 1, 18}, {0, 0, 0, 0}}},{57134, {{21456, 1, 0, 0},{21456, 1, 1, 19}, {0, 0, 0, 0}}},{57134, {{21456, 1, 0, 0},{21456, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{7999, {{21458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8000, {{21459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8002, {{21460, 1, 0, 0},{21460, 1, 1, 17}, {0, 0, 0, 0}}},{8002, {{21460, 1, 0, 0},{21460, 1, 1, 17}, {0, 0, 0, 0}}},{57135, {{21460, 1, 0, 0},{21460, 1, 1, 18}, {0, 0, 0, 0}}},{57135, {{21460, 1, 0, 0},{21460, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8003, {{21461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8004, {{21462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8005, {{21463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{8007, {{21465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8008, {{21466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {10, {{8009, {{21467, 1, 0, 0},{21467, 1, 1, 19}, {0, 0, 0, 0}}},{8009, {{21467, 1, 0, 0},{21467, 1, 1, 19}, {0, 0, 0, 0}}},{8010, {{21467, 1, 0, 0},{21467, 1, 1, 18}, {0, 0, 0, 0}}},{8010, {{21467, 1, 0, 0},{21467, 1, 1, 18}, {0, 0, 0, 0}}},{8011, {{21467, 1, 0, 0},{21467, 1, 1, 20}, {0, 0, 0, 0}}},{8011, {{21467, 1, 0, 0},{21467, 1, 1, 20}, {0, 0, 0, 0}}},{8012, {{21467, 1, 0, 0},{21467, 1, 1, 21}, {0, 0, 0, 0}}},{8012, {{21467, 1, 0, 0},{21467, 1, 1, 21}, {0, 0, 0, 0}}},{57140, {{21467, 1, 0, 0},{21467, 1, 1, 22}, {0, 0, 0, 0}}},{57140, {{21467, 1, 0, 0},{21467, 1, 1, 22}, {0, 0, 0, 0}}}}},
    {4, {{8013, {{21468, 1, 0, 0},{21468, 1, 1, 16}, {0, 0, 0, 0}}},{8013, {{21468, 1, 0, 0},{21468, 1, 1, 16}, {0, 0, 0, 0}}},{57139, {{21468, 1, 0, 0},{21468, 1, 1, 17}, {0, 0, 0, 0}}},{57139, {{21468, 1, 0, 0},{21468, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{8014, {{21469, 1, 0, 0},{21469, 1, 1, 17}, {0, 0, 0, 0}}},{8014, {{21469, 1, 0, 0},{21469, 1, 1, 17}, {0, 0, 0, 0}}},{8015, {{21469, 1, 0, 0},{21469, 1, 1, 18}, {0, 0, 0, 0}}},{8015, {{21469, 1, 0, 0},{21469, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8016, {{21470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{8017, {{21471, 1, 0, 0},{21471, 1, 1, 20}, {0, 0, 0, 0}}},{8017, {{21471, 1, 0, 0},{21471, 1, 1, 20}, {0, 0, 0, 0}}},{8018, {{21471, 1, 0, 0},{21471, 1, 1, 18}, {0, 0, 0, 0}}},{8018, {{21471, 1, 0, 0},{21471, 1, 1, 18}, {0, 0, 0, 0}}},{8019, {{21471, 1, 0, 0},{21471, 1, 1, 19}, {0, 0, 0, 0}}},{8019, {{21471, 1, 0, 0},{21471, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{8020, {{21472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8021, {{21473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8022, {{21474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8023, {{21475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8024, {{21476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8025, {{21477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8026, {{21478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8027, {{21479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8028, {{21480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8029, {{21481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8030, {{21482, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8031, {{21483, 1, 0, 0},{21483, 1, 1, 18}, {0, 0, 0, 0}}},{8031, {{21483, 1, 0, 0},{21483, 1, 1, 18}, {0, 0, 0, 0}}},{8032, {{21483, 1, 0, 0},{21483, 1, 1, 17}, {0, 0, 0, 0}}},{8032, {{21483, 1, 0, 0},{21483, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{8033, {{21484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8034, {{21485, 1, 0, 0},{21485, 1, 1, 17}, {0, 0, 0, 0}}},{8034, {{21485, 1, 0, 0},{21485, 1, 1, 17}, {0, 0, 0, 0}}},{8035, {{21485, 1, 0, 0},{21485, 1, 1, 18}, {0, 0, 0, 0}}},{8035, {{21485, 1, 0, 0},{21485, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8036, {{21486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8037, {{21487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8038, {{21488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8039, {{21489, 1, 0, 0},{21489, 1, 1, 18}, {0, 0, 0, 0}}},{8039, {{21489, 1, 0, 0},{21489, 1, 1, 18}, {0, 0, 0, 0}}},{8040, {{21489, 1, 0, 0},{21489, 1, 1, 19}, {0, 0, 0, 0}}},{8040, {{21489, 1, 0, 0},{21489, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{8041, {{21490, 1, 0, 0},{21490, 1, 1, 18}, {0, 0, 0, 0}}},{8041, {{21490, 1, 0, 0},{21490, 1, 1, 18}, {0, 0, 0, 0}}},{8042, {{21490, 1, 0, 0},{21490, 1, 1, 19}, {0, 0, 0, 0}}},{8042, {{21490, 1, 0, 0},{21490, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{8043, {{21491, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8044, {{21492, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8045, {{21493, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8046, {{21494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8047, {{21495, 1, 0, 0},{21495, 1, 1, 17}, {0, 0, 0, 0}}},{8047, {{21495, 1, 0, 0},{21495, 1, 1, 17}, {0, 0, 0, 0}}},{57144, {{21495, 1, 0, 0},{21495, 1, 1, 18}, {0, 0, 0, 0}}},{57144, {{21495, 1, 0, 0},{21495, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8048, {{21496, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8049, {{21497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8050, {{21498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8051, {{21499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8052, {{21500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{8053, {{21503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8054, {{21504, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8055, {{21505, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8056, {{21506, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8057, {{21507, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8058, {{21508, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8059, {{21509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8060, {{21510, 1, 0, 0},{21510, 1, 1, 16}, {0, 0, 0, 0}}},{8060, {{21510, 1, 0, 0},{21510, 1, 1, 16}, {0, 0, 0, 0}}},{8061, {{21510, 1, 0, 0},{21510, 1, 1, 17}, {0, 0, 0, 0}}},{8061, {{21510, 1, 0, 0},{21510, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{8062, {{21511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8063, {{21512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8064, {{21513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8065, {{21514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8066, {{21515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8067, {{21516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8068, {{21517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8069, {{21518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8070, {{21519, 1, 0, 0},{21519, 1, 1, 18}, {0, 0, 0, 0}}},{8070, {{21519, 1, 0, 0},{21519, 1, 1, 18}, {0, 0, 0, 0}}},{8071, {{21519, 1, 0, 0},{21519, 1, 1, 20}, {0, 0, 0, 0}}},{8071, {{21519, 1, 0, 0},{21519, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{8073, {{21520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8074, {{21521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8075, {{21522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8076, {{21523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8077, {{21524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{8078, {{21528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8079, {{21529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8080, {{21530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8081, {{21531, 1, 0, 0},{21531, 1, 1, 17}, {0, 0, 0, 0}}},{8081, {{21531, 1, 0, 0},{21531, 1, 1, 17}, {0, 0, 0, 0}}},{8082, {{21531, 1, 0, 0},{21531, 1, 1, 18}, {0, 0, 0, 0}}},{8082, {{21531, 1, 0, 0},{21531, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8083, {{21532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8084, {{21533, 1, 0, 0},{21533, 1, 1, 17}, {0, 0, 0, 0}}},{8084, {{21533, 1, 0, 0},{21533, 1, 1, 17}, {0, 0, 0, 0}}},{8085, {{21533, 1, 0, 0},{21533, 1, 1, 18}, {0, 0, 0, 0}}},{8085, {{21533, 1, 0, 0},{21533, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8086, {{21534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8087, {{21535, 1, 0, 0},{21535, 1, 1, 17}, {0, 0, 0, 0}}},{8087, {{21535, 1, 0, 0},{21535, 1, 1, 17}, {0, 0, 0, 0}}},{59417, {{21535, 1, 0, 0},{21535, 1, 1, 18}, {0, 0, 0, 0}}},{59417, {{21535, 1, 0, 0},{21535, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8088, {{21536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8089, {{21537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8090, {{21538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8091, {{21539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8092, {{21540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8093, {{21541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8094, {{21542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8095, {{21543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8096, {{21544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8097, {{21545, 1, 0, 0},{21545, 1, 1, 17}, {0, 0, 0, 0}}},{8097, {{21545, 1, 0, 0},{21545, 1, 1, 17}, {0, 0, 0, 0}}},{8098, {{21545, 1, 0, 0},{21545, 1, 1, 18}, {0, 0, 0, 0}}},{8098, {{21545, 1, 0, 0},{21545, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8099, {{21546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8100, {{21547, 1, 0, 0},{21547, 1, 1, 17}, {0, 0, 0, 0}}},{8100, {{21547, 1, 0, 0},{21547, 1, 1, 17}, {0, 0, 0, 0}}},{57151, {{21547, 1, 0, 0},{21547, 1, 1, 18}, {0, 0, 0, 0}}},{57151, {{21547, 1, 0, 0},{21547, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8101, {{21548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8102, {{21549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8103, {{21550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8104, {{21551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8105, {{21552, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8106, {{21553, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8107, {{21554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8108, {{21555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8109, {{21556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8110, {{21557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8111, {{21558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8112, {{21559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8113, {{21560, 1, 0, 0},{21560, 1, 1, 19}, {0, 0, 0, 0}}},{8113, {{21560, 1, 0, 0},{21560, 1, 1, 19}, {0, 0, 0, 0}}},{8114, {{21560, 1, 0, 0},{21560, 1, 1, 18}, {0, 0, 0, 0}}},{8114, {{21560, 1, 0, 0},{21560, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8115, {{21561, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8116, {{21562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8117, {{21563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8118, {{21564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8119, {{21565, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8120, {{21566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8122, {{21567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8123, {{21568, 1, 0, 0},{21568, 1, 1, 18}, {0, 0, 0, 0}}},{8123, {{21568, 1, 0, 0},{21568, 1, 1, 18}, {0, 0, 0, 0}}},{8124, {{21568, 1, 0, 0},{21568, 1, 1, 19}, {0, 0, 0, 0}}},{8124, {{21568, 1, 0, 0},{21568, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{8125, {{21569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8126, {{21570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8127, {{21571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8128, {{21572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8129, {{21573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8130, {{21574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8131, {{21575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{8132, {{21576, 1, 0, 0},{21576, 1, 1, 20}, {0, 0, 0, 0}}},{8132, {{21576, 1, 0, 0},{21576, 1, 1, 20}, {0, 0, 0, 0}}},{8133, {{21576, 1, 0, 0},{21576, 1, 1, 18}, {0, 0, 0, 0}}},{8133, {{21576, 1, 0, 0},{21576, 1, 1, 18}, {0, 0, 0, 0}}},{8134, {{21576, 1, 0, 0},{21576, 1, 1, 19}, {0, 0, 0, 0}}},{8134, {{21576, 1, 0, 0},{21576, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{8135, {{21577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8136, {{21578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8137, {{21579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8138, {{21580, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8139, {{21581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8140, {{21582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8141, {{21583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8142, {{21584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8143, {{21585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{8144, {{21589, 1, 0, 0},{21589, 1, 1, 17}, {0, 0, 0, 0}}},{8144, {{21589, 1, 0, 0},{21589, 1, 1, 17}, {0, 0, 0, 0}}},{8145, {{21589, 1, 0, 0},{21589, 1, 1, 18}, {0, 0, 0, 0}}},{8145, {{21589, 1, 0, 0},{21589, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{8146, {{21594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{8147, {{21597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8148, {{21598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8149, {{21599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8150, {{21600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8151, {{21601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8152, {{21602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8153, {{21603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8154, {{21604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8155, {{21605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8156, {{21606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8157, {{21607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8158, {{21608, 1, 0, 0},{21608, 1, 1, 19}, {0, 0, 0, 0}}},{8158, {{21608, 1, 0, 0},{21608, 1, 1, 19}, {0, 0, 0, 0}}},{8159, {{21608, 1, 0, 0},{21608, 1, 1, 18}, {0, 0, 0, 0}}},{8159, {{21608, 1, 0, 0},{21608, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8160, {{21609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8161, {{21610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8162, {{21611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8163, {{21612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8164, {{21613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8165, {{21614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8166, {{21615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8167, {{21616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8168, {{21617, 1, 0, 0},{21617, 1, 1, 18}, {0, 0, 0, 0}}},{8168, {{21617, 1, 0, 0},{21617, 1, 1, 18}, {0, 0, 0, 0}}},{8169, {{21617, 1, 0, 0},{21617, 1, 1, 17}, {0, 0, 0, 0}}},{8169, {{21617, 1, 0, 0},{21617, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{8170, {{21618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8171, {{21619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8172, {{21620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8173, {{21621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8174, {{21622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8175, {{21623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8176, {{21624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8177, {{21625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8178, {{21626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8179, {{21627, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8180, {{21628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8181, {{21629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8182, {{21630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8183, {{21631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8184, {{21632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8185, {{21633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8186, {{21634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8187, {{21635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8188, {{21636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8189, {{21637, 1, 0, 0},{21637, 1, 1, 17}, {0, 0, 0, 0}}},{8189, {{21637, 1, 0, 0},{21637, 1, 1, 17}, {0, 0, 0, 0}}},{8190, {{21637, 1, 0, 0},{21637, 1, 1, 18}, {0, 0, 0, 0}}},{8190, {{21637, 1, 0, 0},{21637, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{8191, {{21638, 1, 0, 0},{21638, 1, 1, 17}, {0, 0, 0, 0}}},{8191, {{21638, 1, 0, 0},{21638, 1, 1, 17}, {0, 0, 0, 0}}},{8192, {{21638, 1, 0, 0},{21638, 1, 1, 18}, {0, 0, 0, 0}}},{8192, {{21638, 1, 0, 0},{21638, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8193, {{21639, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8194, {{21640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8195, {{21641, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8196, {{21642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8198, {{21643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8199, {{21644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8200, {{21645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{8201, {{21646, 1, 0, 0},{21646, 1, 1, 17}, {0, 0, 0, 0}}},{8201, {{21646, 1, 0, 0},{21646, 1, 1, 17}, {0, 0, 0, 0}}},{57147, {{21646, 1, 0, 0},{21646, 1, 1, 18}, {0, 0, 0, 0}}},{57147, {{21646, 1, 0, 0},{21646, 1, 1, 18}, {0, 0, 0, 0}}},{57149, {{21646, 1, 0, 0},{21646, 1, 1, 19}, {0, 0, 0, 0}}},{57149, {{21646, 1, 0, 0},{21646, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{8202, {{21647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8203, {{21648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8204, {{21649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8205, {{21650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8207, {{21651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8208, {{21652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8209, {{21653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8210, {{21654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{8211, {{21660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{8212, {{21662, 1, 0, 0},{21662, 1, 1, 17}, {0, 0, 0, 0}}},{8212, {{21662, 1, 0, 0},{21662, 1, 1, 17}, {0, 0, 0, 0}}},{8213, {{21662, 1, 0, 0},{21662, 1, 1, 16}, {0, 0, 0, 0}}},{8213, {{21662, 1, 0, 0},{21662, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{8214, {{21663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8215, {{21664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8216, {{21665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8217, {{21666, 1, 0, 0},{21666, 1, 1, 17}, {0, 0, 0, 0}}},{8217, {{21666, 1, 0, 0},{21666, 1, 1, 17}, {0, 0, 0, 0}}},{8218, {{21666, 1, 0, 0},{21666, 1, 1, 18}, {0, 0, 0, 0}}},{8218, {{21666, 1, 0, 0},{21666, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8219, {{21667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8220, {{21668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8221, {{21669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8222, {{21670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8223, {{21671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8224, {{21672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8225, {{21673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8226, {{21674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8227, {{21675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8228, {{21676, 1, 0, 0},{21676, 1, 1, 18}, {0, 0, 0, 0}}},{8228, {{21676, 1, 0, 0},{21676, 1, 1, 18}, {0, 0, 0, 0}}},{8229, {{21676, 1, 0, 0},{21676, 1, 1, 19}, {0, 0, 0, 0}}},{8229, {{21676, 1, 0, 0},{21676, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{8230, {{21677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8231, {{21678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8232, {{21679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8233, {{21680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8234, {{21681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8235, {{21682, 1, 0, 0},{21682, 1, 1, 19}, {0, 0, 0, 0}}},{8235, {{21682, 1, 0, 0},{21682, 1, 1, 19}, {0, 0, 0, 0}}},{8236, {{21682, 1, 0, 0},{21682, 1, 1, 18}, {0, 0, 0, 0}}},{8236, {{21682, 1, 0, 0},{21682, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8237, {{21683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8238, {{21684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8239, {{21685, 1, 0, 0},{21685, 1, 1, 16}, {0, 0, 0, 0}}},{8239, {{21685, 1, 0, 0},{21685, 1, 1, 16}, {0, 0, 0, 0}}},{8240, {{21685, 1, 0, 0},{21685, 1, 1, 17}, {0, 0, 0, 0}}},{8240, {{21685, 1, 0, 0},{21685, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{8241, {{21686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8242, {{21687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8243, {{21688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8244, {{21689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8245, {{21690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8246, {{21691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{8247, {{21692, 1, 0, 0},{21692, 1, 1, 17}, {0, 0, 0, 0}}},{8247, {{21692, 1, 0, 0},{21692, 1, 1, 17}, {0, 0, 0, 0}}},{57153, {{21692, 1, 0, 0},{21692, 1, 1, 18}, {0, 0, 0, 0}}},{57153, {{21692, 1, 0, 0},{21692, 1, 1, 18}, {0, 0, 0, 0}}},{57154, {{21692, 1, 0, 0},{21692, 1, 1, 19}, {0, 0, 0, 0}}},{57154, {{21692, 1, 0, 0},{21692, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{8248, {{21693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8249, {{21694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8250, {{21695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8251, {{21696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8252, {{21697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8253, {{21698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8254, {{21699, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8255, {{21700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8256, {{21701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8257, {{21702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8258, {{21703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8259, {{21704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8260, {{21705, 1, 0, 0},{21705, 1, 1, 17}, {0, 0, 0, 0}}},{8260, {{21705, 1, 0, 0},{21705, 1, 1, 17}, {0, 0, 0, 0}}},{8261, {{21705, 1, 0, 0},{21705, 1, 1, 18}, {0, 0, 0, 0}}},{8261, {{21705, 1, 0, 0},{21705, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8262, {{21706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8263, {{21707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{8264, {{21709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8265, {{21710, 1, 0, 0},{21710, 1, 1, 17}, {0, 0, 0, 0}}},{8265, {{21710, 1, 0, 0},{21710, 1, 1, 17}, {0, 0, 0, 0}}},{8266, {{21710, 1, 0, 0},{21710, 1, 1, 18}, {0, 0, 0, 0}}},{8266, {{21710, 1, 0, 0},{21710, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8267, {{21711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{8268, {{21720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8269, {{21721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{68025, {{21723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{8270, {{21728, 1, 0, 0},{21728, 1, 1, 17}, {0, 0, 0, 0}}},{8270, {{21728, 1, 0, 0},{21728, 1, 1, 17}, {0, 0, 0, 0}}},{8271, {{21728, 1, 0, 0},{21728, 1, 1, 18}, {0, 0, 0, 0}}},{8271, {{21728, 1, 0, 0},{21728, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8272, {{21729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8273, {{21730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8274, {{21731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8275, {{21732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8276, {{21733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8277, {{21734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8278, {{21735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8279, {{21736, 1, 0, 0},{21736, 1, 1, 19}, {0, 0, 0, 0}}},{8279, {{21736, 1, 0, 0},{21736, 1, 1, 19}, {0, 0, 0, 0}}},{8280, {{21736, 1, 0, 0},{21736, 1, 1, 18}, {0, 0, 0, 0}}},{8280, {{21736, 1, 0, 0},{21736, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8281, {{21737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8282, {{21738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8283, {{21739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8284, {{21740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8285, {{21741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8286, {{21742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8287, {{21743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8288, {{21744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8289, {{21745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8290, {{21746, 1, 0, 0},{21746, 1, 1, 17}, {0, 0, 0, 0}}},{8290, {{21746, 1, 0, 0},{21746, 1, 1, 17}, {0, 0, 0, 0}}},{8291, {{21746, 1, 0, 0},{21746, 1, 1, 18}, {0, 0, 0, 0}}},{8291, {{21746, 1, 0, 0},{21746, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8292, {{21747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8293, {{21748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8294, {{21749, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8295, {{21750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8296, {{21751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8297, {{21752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8298, {{21753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8299, {{21754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8300, {{21755, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8301, {{21756, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8302, {{21757, 1, 0, 0},{21757, 1, 1, 17}, {0, 0, 0, 0}}},{8302, {{21757, 1, 0, 0},{21757, 1, 1, 17}, {0, 0, 0, 0}}},{8303, {{21757, 1, 0, 0},{21757, 1, 1, 18}, {0, 0, 0, 0}}},{8303, {{21757, 1, 0, 0},{21757, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8304, {{21758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8306, {{21759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8307, {{21760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8308, {{21761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8309, {{21762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8310, {{21763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8311, {{21764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8312, {{21765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8313, {{21766, 1, 0, 0},{21766, 1, 1, 18}, {0, 0, 0, 0}}},{8313, {{21766, 1, 0, 0},{21766, 1, 1, 18}, {0, 0, 0, 0}}},{8314, {{21766, 1, 0, 0},{21766, 1, 1, 17}, {0, 0, 0, 0}}},{8314, {{21766, 1, 0, 0},{21766, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{8315, {{21767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8316, {{21768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8317, {{21769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8318, {{21770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8319, {{21771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8320, {{21772, 1, 0, 0},{21772, 1, 1, 17}, {0, 0, 0, 0}}},{8320, {{21772, 1, 0, 0},{21772, 1, 1, 17}, {0, 0, 0, 0}}},{8321, {{21772, 1, 0, 0},{21772, 1, 1, 18}, {0, 0, 0, 0}}},{8321, {{21772, 1, 0, 0},{21772, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8322, {{21773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8323, {{21774, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8324, {{21775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8325, {{21776, 1, 0, 0},{21776, 1, 1, 19}, {0, 0, 0, 0}}},{8325, {{21776, 1, 0, 0},{21776, 1, 1, 19}, {0, 0, 0, 0}}},{8326, {{21776, 1, 0, 0},{21776, 1, 1, 18}, {0, 0, 0, 0}}},{8326, {{21776, 1, 0, 0},{21776, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8327, {{21777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8328, {{21778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8329, {{21779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8330, {{21780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8331, {{21781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8332, {{21782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8333, {{21783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8334, {{21784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8335, {{21785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8336, {{21786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{68026, {{21788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{68027, {{21791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{8337, {{21799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{8338, {{21801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8339, {{21802, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8340, {{21803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8341, {{21804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8343, {{21805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8344, {{21806, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8345, {{21807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8346, {{21808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8347, {{21809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8348, {{21810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{8349, {{21811, 1, 0, 0},{21811, 1, 1, 19}, {0, 0, 0, 0}}},{8349, {{21811, 1, 0, 0},{21811, 1, 1, 19}, {0, 0, 0, 0}}},{8350, {{21811, 1, 0, 0},{21811, 1, 1, 21}, {0, 0, 0, 0}}},{8350, {{21811, 1, 0, 0},{21811, 1, 1, 21}, {0, 0, 0, 0}}},{8351, {{21811, 1, 0, 0},{21811, 1, 1, 20}, {0, 0, 0, 0}}},{8351, {{21811, 1, 0, 0},{21811, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{8352, {{21812, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8353, {{21813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8354, {{21814, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8355, {{21815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8356, {{21816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8357, {{21817, 1, 0, 0},{21817, 1, 1, 18}, {0, 0, 0, 0}}},{8357, {{21817, 1, 0, 0},{21817, 1, 1, 18}, {0, 0, 0, 0}}},{8358, {{21817, 1, 0, 0},{21817, 1, 1, 19}, {0, 0, 0, 0}}},{8358, {{21817, 1, 0, 0},{21817, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{8359, {{21818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8360, {{21819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8361, {{21820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8362, {{21821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8363, {{21822, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8364, {{21823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8365, {{21824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8366, {{21825, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8368, {{21826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8369, {{21827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {10, {{8370, {{21828, 1, 0, 0},{21828, 1, 1, 18}, {0, 0, 0, 0}}},{8370, {{21828, 1, 0, 0},{21828, 1, 1, 18}, {0, 0, 0, 0}}},{8371, {{21828, 1, 0, 0},{21828, 1, 1, 21}, {0, 0, 0, 0}}},{8371, {{21828, 1, 0, 0},{21828, 1, 1, 21}, {0, 0, 0, 0}}},{8372, {{21828, 1, 0, 0},{21828, 1, 1, 20}, {0, 0, 0, 0}}},{8372, {{21828, 1, 0, 0},{21828, 1, 1, 20}, {0, 0, 0, 0}}},{8373, {{21828, 1, 0, 0},{21828, 1, 1, 22}, {0, 0, 0, 0}}},{8373, {{21828, 1, 0, 0},{21828, 1, 1, 22}, {0, 0, 0, 0}}},{8374, {{21828, 1, 0, 0},{21828, 1, 1, 19}, {0, 0, 0, 0}}},{8374, {{21828, 1, 0, 0},{21828, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{8375, {{21829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{8376, {{21830, 1, 0, 0},{21830, 1, 1, 18}, {0, 0, 0, 0}}},{8376, {{21830, 1, 0, 0},{21830, 1, 1, 18}, {0, 0, 0, 0}}},{8377, {{21830, 1, 0, 0},{21830, 1, 1, 19}, {0, 0, 0, 0}}},{8377, {{21830, 1, 0, 0},{21830, 1, 1, 19}, {0, 0, 0, 0}}},{57166, {{21830, 1, 0, 0},{21830, 1, 1, 20}, {0, 0, 0, 0}}},{57166, {{21830, 1, 0, 0},{21830, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{8378, {{21831, 1, 0, 0},{21831, 1, 1, 17}, {0, 0, 0, 0}}},{8378, {{21831, 1, 0, 0},{21831, 1, 1, 17}, {0, 0, 0, 0}}},{57165, {{21831, 1, 0, 0},{21831, 1, 1, 18}, {0, 0, 0, 0}}},{57165, {{21831, 1, 0, 0},{21831, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8379, {{21832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8380, {{21833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8381, {{21834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8382, {{21835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8383, {{21836, 1, 0, 0},{21836, 1, 1, 17}, {0, 0, 0, 0}}},{8383, {{21836, 1, 0, 0},{21836, 1, 1, 17}, {0, 0, 0, 0}}},{8384, {{21836, 1, 0, 0},{21836, 1, 1, 18}, {0, 0, 0, 0}}},{8384, {{21836, 1, 0, 0},{21836, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8385, {{21837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8386, {{21838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8387, {{21839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8388, {{21840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8389, {{21841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8390, {{21842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8391, {{21843, 1, 0, 0},{21843, 1, 1, 19}, {0, 0, 0, 0}}},{8391, {{21843, 1, 0, 0},{21843, 1, 1, 19}, {0, 0, 0, 0}}},{8392, {{21843, 1, 0, 0},{21843, 1, 1, 18}, {0, 0, 0, 0}}},{8392, {{21843, 1, 0, 0},{21843, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8393, {{21844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8395, {{21845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8396, {{21846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8397, {{21847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8398, {{21848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8399, {{21849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{8400, {{21850, 1, 0, 0},{21850, 1, 1, 17}, {0, 0, 0, 0}}},{8400, {{21850, 1, 0, 0},{21850, 1, 1, 17}, {0, 0, 0, 0}}},{8401, {{21850, 1, 0, 0},{21850, 1, 1, 18}, {0, 0, 0, 0}}},{8401, {{21850, 1, 0, 0},{21850, 1, 1, 18}, {0, 0, 0, 0}}},{57180, {{21850, 1, 0, 0},{21850, 1, 1, 19}, {0, 0, 0, 0}}},{57180, {{21850, 1, 0, 0},{21850, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{8402, {{21851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8403, {{21852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8404, {{21853, 1, 0, 0},{21853, 1, 1, 17}, {0, 0, 0, 0}}},{8404, {{21853, 1, 0, 0},{21853, 1, 1, 17}, {0, 0, 0, 0}}},{8405, {{21853, 1, 0, 0},{21853, 1, 1, 18}, {0, 0, 0, 0}}},{8405, {{21853, 1, 0, 0},{21853, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8406, {{21854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8407, {{21855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8408, {{21856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8409, {{21857, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8410, {{21858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8411, {{21859, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8412, {{21860, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8413, {{21861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8414, {{21862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{8415, {{21864, 1, 0, 0},{21864, 1, 1, 16}, {0, 0, 0, 0}}},{8415, {{21864, 1, 0, 0},{21864, 1, 1, 16}, {0, 0, 0, 0}}},{8416, {{21864, 1, 0, 0},{21864, 1, 1, 17}, {0, 0, 0, 0}}},{8416, {{21864, 1, 0, 0},{21864, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{8417, {{21877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{8418, {{21882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8419, {{21883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8420, {{21884, 1, 0, 0},{21884, 1, 1, 17}, {0, 0, 0, 0}}},{8420, {{21884, 1, 0, 0},{21884, 1, 1, 17}, {0, 0, 0, 0}}},{8421, {{21884, 1, 0, 0},{21884, 1, 1, 18}, {0, 0, 0, 0}}},{8421, {{21884, 1, 0, 0},{21884, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8422, {{21885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8423, {{21886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8424, {{21887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8425, {{21888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8426, {{21889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8427, {{21890, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8428, {{21891, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {10, {{8429, {{21892, 1, 0, 0},{21892, 1, 1, 18}, {0, 0, 0, 0}}},{8429, {{21892, 1, 0, 0},{21892, 1, 1, 18}, {0, 0, 0, 0}}},{8430, {{21892, 1, 0, 0},{21892, 1, 1, 22}, {0, 0, 0, 0}}},{8430, {{21892, 1, 0, 0},{21892, 1, 1, 22}, {0, 0, 0, 0}}},{8431, {{21892, 1, 0, 0},{21892, 1, 1, 19}, {0, 0, 0, 0}}},{8431, {{21892, 1, 0, 0},{21892, 1, 1, 19}, {0, 0, 0, 0}}},{8432, {{21892, 1, 0, 0},{21892, 1, 1, 20}, {0, 0, 0, 0}}},{8432, {{21892, 1, 0, 0},{21892, 1, 1, 20}, {0, 0, 0, 0}}},{8433, {{21892, 1, 0, 0},{21892, 1, 1, 21}, {0, 0, 0, 0}}},{8433, {{21892, 1, 0, 0},{21892, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{8434, {{21893, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8435, {{21894, 1, 0, 0},{21894, 1, 1, 17}, {0, 0, 0, 0}}},{8435, {{21894, 1, 0, 0},{21894, 1, 1, 17}, {0, 0, 0, 0}}},{8436, {{21894, 1, 0, 0},{21894, 1, 1, 18}, {0, 0, 0, 0}}},{8436, {{21894, 1, 0, 0},{21894, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8437, {{21895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8438, {{21896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8439, {{21897, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8440, {{21898, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8441, {{21899, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8442, {{21900, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8443, {{21901, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8444, {{21902, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8445, {{21903, 1, 0, 0},{21903, 1, 1, 17}, {0, 0, 0, 0}}},{8445, {{21903, 1, 0, 0},{21903, 1, 1, 17}, {0, 0, 0, 0}}},{8446, {{21903, 1, 0, 0},{21903, 1, 1, 18}, {0, 0, 0, 0}}},{8446, {{21903, 1, 0, 0},{21903, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8447, {{21904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8448, {{21905, 1, 0, 0},{21905, 1, 1, 17}, {0, 0, 0, 0}}},{8448, {{21905, 1, 0, 0},{21905, 1, 1, 17}, {0, 0, 0, 0}}},{8449, {{21905, 1, 0, 0},{21905, 1, 1, 18}, {0, 0, 0, 0}}},{8449, {{21905, 1, 0, 0},{21905, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8450, {{21906, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8451, {{21907, 1, 0, 0},{21907, 1, 1, 17}, {0, 0, 0, 0}}},{8451, {{21907, 1, 0, 0},{21907, 1, 1, 17}, {0, 0, 0, 0}}},{8452, {{21907, 1, 0, 0},{21907, 1, 1, 18}, {0, 0, 0, 0}}},{8452, {{21907, 1, 0, 0},{21907, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8453, {{21908, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8454, {{21909, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8455, {{21910, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8456, {{21911, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8457, {{21912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{8458, {{21913, 1, 0, 0},{21913, 1, 1, 17}, {0, 0, 0, 0}}},{8458, {{21913, 1, 0, 0},{21913, 1, 1, 17}, {0, 0, 0, 0}}},{8459, {{21913, 1, 0, 0},{21913, 1, 1, 20}, {0, 0, 0, 0}}},{8459, {{21913, 1, 0, 0},{21913, 1, 1, 20}, {0, 0, 0, 0}}},{8460, {{21913, 1, 0, 0},{21913, 1, 1, 18}, {0, 0, 0, 0}}},{8460, {{21913, 1, 0, 0},{21913, 1, 1, 18}, {0, 0, 0, 0}}},{8461, {{21913, 1, 0, 0},{21913, 1, 1, 19}, {0, 0, 0, 0}}},{8461, {{21913, 1, 0, 0},{21913, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{8462, {{21914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8463, {{21915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{8464, {{21916, 1, 0, 0},{21916, 1, 1, 18}, {0, 0, 0, 0}}},{8464, {{21916, 1, 0, 0},{21916, 1, 1, 18}, {0, 0, 0, 0}}},{8465, {{21916, 1, 0, 0},{21916, 1, 1, 19}, {0, 0, 0, 0}}},{8465, {{21916, 1, 0, 0},{21916, 1, 1, 19}, {0, 0, 0, 0}}},{57175, {{21916, 1, 0, 0},{21916, 1, 1, 20}, {0, 0, 0, 0}}},{57175, {{21916, 1, 0, 0},{21916, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {5, {{8468, {{21917, 1, 0, 0},{21917, 1, 1, 18}, {0, 0, 0, 0}}},{8468, {{21917, 1, 0, 0},{21917, 1, 1, 18}, {0, 0, 0, 0}}},{30250, {{21917, 1, 0, 0},{21917, 1, 1, 0},{21917, 1, 1, 19},{64054, 1, 0, 0}, {0, 0, 0, 0}}},{30250, {{21917, 1, 0, 0},{21917, 1, 1, 0},{21917, 1, 1, 19},{64054, 1, 0, 0}, {0, 0, 0, 0}}},{30250, {{21917, 1, 0, 0},{21917, 1, 1, 0},{21917, 1, 1, 19},{64054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8469, {{21918, 1, 0, 0},{21918, 1, 1, 17}, {0, 0, 0, 0}}},{8469, {{21918, 1, 0, 0},{21918, 1, 1, 17}, {0, 0, 0, 0}}},{8470, {{21918, 1, 0, 0},{21918, 1, 1, 18}, {0, 0, 0, 0}}},{8470, {{21918, 1, 0, 0},{21918, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8471, {{21919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8472, {{21920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8473, {{21921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8474, {{21922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8475, {{21923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8476, {{21924, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8477, {{21925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8478, {{21926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8479, {{21927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8480, {{21928, 1, 0, 0},{21928, 1, 1, 17}, {0, 0, 0, 0}}},{8480, {{21928, 1, 0, 0},{21928, 1, 1, 17}, {0, 0, 0, 0}}},{8481, {{21928, 1, 0, 0},{21928, 1, 1, 18}, {0, 0, 0, 0}}},{8481, {{21928, 1, 0, 0},{21928, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{8482, {{21929, 1, 0, 0},{21929, 1, 1, 19}, {0, 0, 0, 0}}},{8482, {{21929, 1, 0, 0},{21929, 1, 1, 19}, {0, 0, 0, 0}}},{8483, {{21929, 1, 0, 0},{21929, 1, 1, 20}, {0, 0, 0, 0}}},{8483, {{21929, 1, 0, 0},{21929, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{8484, {{21930, 1, 0, 0},{21930, 1, 1, 17}, {0, 0, 0, 0}}},{8484, {{21930, 1, 0, 0},{21930, 1, 1, 17}, {0, 0, 0, 0}}},{57176, {{21930, 1, 0, 0},{21930, 1, 1, 18}, {0, 0, 0, 0}}},{57176, {{21930, 1, 0, 0},{21930, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {8, {{8485, {{21931, 1, 0, 0},{21931, 1, 1, 20}, {0, 0, 0, 0}}},{8485, {{21931, 1, 0, 0},{21931, 1, 1, 20}, {0, 0, 0, 0}}},{8486, {{21931, 1, 0, 0},{21931, 1, 1, 19}, {0, 0, 0, 0}}},{8486, {{21931, 1, 0, 0},{21931, 1, 1, 19}, {0, 0, 0, 0}}},{8487, {{21931, 1, 0, 0},{21931, 1, 1, 21}, {0, 0, 0, 0}}},{8487, {{21931, 1, 0, 0},{21931, 1, 1, 21}, {0, 0, 0, 0}}},{8488, {{21931, 1, 0, 0},{21931, 1, 1, 22}, {0, 0, 0, 0}}},{8488, {{21931, 1, 0, 0},{21931, 1, 1, 22}, {0, 0, 0, 0}}}}},
    {1, {{8489, {{21932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8490, {{21933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{8491, {{21934, 1, 0, 0},{21934, 1, 1, 17}, {0, 0, 0, 0}}},{8491, {{21934, 1, 0, 0},{21934, 1, 1, 17}, {0, 0, 0, 0}}},{57167, {{21934, 1, 0, 0},{21934, 1, 1, 19}, {0, 0, 0, 0}}},{57167, {{21934, 1, 0, 0},{21934, 1, 1, 19}, {0, 0, 0, 0}}},{57178, {{21934, 1, 0, 0},{21934, 1, 1, 18}, {0, 0, 0, 0}}},{57178, {{21934, 1, 0, 0},{21934, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8492, {{21935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8493, {{21936, 1, 0, 0},{21936, 1, 1, 18}, {0, 0, 0, 0}}},{8493, {{21936, 1, 0, 0},{21936, 1, 1, 18}, {0, 0, 0, 0}}},{8494, {{21936, 1, 0, 0},{21936, 1, 1, 19}, {0, 0, 0, 0}}},{8494, {{21936, 1, 0, 0},{21936, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{8495, {{21937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8496, {{21938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8497, {{21939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8498, {{21940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8499, {{21941, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8500, {{21942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{68028, {{21944, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{8501, {{21947, 1, 0, 0},{21947, 1, 1, 16}, {0, 0, 0, 0}}},{8501, {{21947, 1, 0, 0},{21947, 1, 1, 16}, {0, 0, 0, 0}}},{8502, {{21947, 1, 0, 0},{21947, 1, 1, 17}, {0, 0, 0, 0}}},{8502, {{21947, 1, 0, 0},{21947, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{57169, {{21948, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{8504, {{21951, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8505, {{21952, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8506, {{21953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8507, {{21954, 1, 0, 0},{21954, 1, 1, 16}, {0, 0, 0, 0}}},{8507, {{21954, 1, 0, 0},{21954, 1, 1, 16}, {0, 0, 0, 0}}},{8508, {{21954, 1, 0, 0},{21954, 1, 1, 17}, {0, 0, 0, 0}}},{8508, {{21954, 1, 0, 0},{21954, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{8509, {{21955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8510, {{21956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8511, {{21957, 1, 0, 0},{21957, 1, 1, 17}, {0, 0, 0, 0}}},{8511, {{21957, 1, 0, 0},{21957, 1, 1, 17}, {0, 0, 0, 0}}},{8512, {{21957, 1, 0, 0},{21957, 1, 1, 18}, {0, 0, 0, 0}}},{8512, {{21957, 1, 0, 0},{21957, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8513, {{21958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8514, {{21959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8515, {{21960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8516, {{21961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8517, {{21962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8518, {{21963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8519, {{21964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8520, {{21965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8521, {{21966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8522, {{21967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8523, {{21968, 1, 0, 0},{21968, 1, 1, 16}, {0, 0, 0, 0}}},{8523, {{21968, 1, 0, 0},{21968, 1, 1, 16}, {0, 0, 0, 0}}},{8524, {{21968, 1, 0, 0},{21968, 1, 1, 17}, {0, 0, 0, 0}}},{8524, {{21968, 1, 0, 0},{21968, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{8525, {{21969, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8526, {{21970, 1, 0, 0},{21970, 1, 1, 17}, {0, 0, 0, 0}}},{8526, {{21970, 1, 0, 0},{21970, 1, 1, 17}, {0, 0, 0, 0}}},{8527, {{21970, 1, 0, 0},{21970, 1, 1, 18}, {0, 0, 0, 0}}},{8527, {{21970, 1, 0, 0},{21970, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8528, {{21971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8529, {{21972, 1, 0, 0},{21972, 1, 1, 17}, {0, 0, 0, 0}}},{8529, {{21972, 1, 0, 0},{21972, 1, 1, 17}, {0, 0, 0, 0}}},{59423, {{21972, 1, 0, 0},{21972, 1, 1, 18}, {0, 0, 0, 0}}},{59423, {{21972, 1, 0, 0},{21972, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8530, {{21973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8531, {{21974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8532, {{21975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8533, {{21976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8534, {{21977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8535, {{21978, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8536, {{21979, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8537, {{21980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8538, {{21981, 1, 0, 0},{21981, 1, 1, 18}, {0, 0, 0, 0}}},{8538, {{21981, 1, 0, 0},{21981, 1, 1, 18}, {0, 0, 0, 0}}},{8539, {{21981, 1, 0, 0},{21981, 1, 1, 17}, {0, 0, 0, 0}}},{8539, {{21981, 1, 0, 0},{21981, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{8540, {{21982, 1, 0, 0},{21982, 1, 1, 17}, {0, 0, 0, 0}}},{8540, {{21982, 1, 0, 0},{21982, 1, 1, 17}, {0, 0, 0, 0}}},{8541, {{21982, 1, 0, 0},{21982, 1, 1, 18}, {0, 0, 0, 0}}},{8541, {{21982, 1, 0, 0},{21982, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8542, {{21983, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8543, {{21984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8544, {{21985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8545, {{21986, 1, 0, 0},{21986, 1, 1, 18}, {0, 0, 0, 0}}},{8545, {{21986, 1, 0, 0},{21986, 1, 1, 18}, {0, 0, 0, 0}}},{8546, {{21986, 1, 0, 0},{21986, 1, 1, 17}, {0, 0, 0, 0}}},{8546, {{21986, 1, 0, 0},{21986, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{8547, {{21987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{8549, {{21988, 1, 0, 0},{21988, 1, 1, 19}, {0, 0, 0, 0}}},{8549, {{21988, 1, 0, 0},{21988, 1, 1, 19}, {0, 0, 0, 0}}},{8550, {{21988, 1, 0, 0},{21988, 1, 1, 20}, {0, 0, 0, 0}}},{8550, {{21988, 1, 0, 0},{21988, 1, 1, 20}, {0, 0, 0, 0}}},{8551, {{21988, 1, 0, 0},{21988, 1, 1, 21}, {0, 0, 0, 0}}},{8551, {{21988, 1, 0, 0},{21988, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{8552, {{21989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8553, {{21990, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8554, {{21991, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8555, {{21992, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8556, {{21993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{68029, {{21997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{8557, {{22001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{8558, {{22006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8559, {{22007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8560, {{22008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8561, {{22009, 1, 0, 0},{22009, 1, 1, 18}, {0, 0, 0, 0}}},{8561, {{22009, 1, 0, 0},{22009, 1, 1, 18}, {0, 0, 0, 0}}},{8562, {{22009, 1, 0, 0},{22009, 1, 1, 17}, {0, 0, 0, 0}}},{8562, {{22009, 1, 0, 0},{22009, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{8563, {{22010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8564, {{22011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8565, {{22012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8566, {{22013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8567, {{22014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8568, {{22015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8569, {{22016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8570, {{22017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8571, {{22018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8572, {{22019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8573, {{22020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{8574, {{22021, 1, 0, 0},{22021, 1, 1, 18}, {0, 0, 0, 0}}},{8574, {{22021, 1, 0, 0},{22021, 1, 1, 18}, {0, 0, 0, 0}}},{8575, {{22021, 1, 0, 0},{22021, 1, 1, 20}, {0, 0, 0, 0}}},{8575, {{22021, 1, 0, 0},{22021, 1, 1, 20}, {0, 0, 0, 0}}},{8576, {{22021, 1, 0, 0},{22021, 1, 1, 19}, {0, 0, 0, 0}}},{8576, {{22021, 1, 0, 0},{22021, 1, 1, 19}, {0, 0, 0, 0}}},{8577, {{22021, 1, 0, 0},{22021, 1, 1, 21}, {0, 0, 0, 0}}},{8577, {{22021, 1, 0, 0},{22021, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {5, {{8578, {{22022, 1, 0, 0},{22022, 1, 1, 18}, {0, 0, 0, 0}}},{8578, {{22022, 1, 0, 0},{22022, 1, 1, 18}, {0, 0, 0, 0}}},{30251, {{22022, 1, 0, 0},{22022, 1, 1, 0},{22022, 1, 1, 19},{64055, 1, 0, 0}, {0, 0, 0, 0}}},{30251, {{22022, 1, 0, 0},{22022, 1, 1, 0},{22022, 1, 1, 19},{64055, 1, 0, 0}, {0, 0, 0, 0}}},{30251, {{22022, 1, 0, 0},{22022, 1, 1, 0},{22022, 1, 1, 19},{64055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8579, {{22023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8580, {{22024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{8581, {{22025, 1, 0, 0},{22025, 1, 1, 18}, {0, 0, 0, 0}}},{8581, {{22025, 1, 0, 0},{22025, 1, 1, 18}, {0, 0, 0, 0}}},{8582, {{22025, 1, 0, 0},{22025, 1, 1, 19}, {0, 0, 0, 0}}},{8582, {{22025, 1, 0, 0},{22025, 1, 1, 19}, {0, 0, 0, 0}}},{8583, {{22025, 1, 0, 0},{22025, 1, 1, 20}, {0, 0, 0, 0}}},{8583, {{22025, 1, 0, 0},{22025, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{8585, {{22026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8586, {{22027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8587, {{22028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8588, {{22029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8589, {{22030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8590, {{22031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8591, {{22032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8592, {{22033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8593, {{22034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8594, {{22035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8595, {{22036, 1, 0, 0},{22036, 1, 1, 17}, {0, 0, 0, 0}}},{8595, {{22036, 1, 0, 0},{22036, 1, 1, 17}, {0, 0, 0, 0}}},{8596, {{22036, 1, 0, 0},{22036, 1, 1, 18}, {0, 0, 0, 0}}},{8596, {{22036, 1, 0, 0},{22036, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8597, {{22037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8598, {{22038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8599, {{22039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8600, {{22040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8601, {{22041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8602, {{22042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8603, {{22043, 1, 0, 0},{22043, 1, 1, 17}, {0, 0, 0, 0}}},{8603, {{22043, 1, 0, 0},{22043, 1, 1, 17}, {0, 0, 0, 0}}},{8604, {{22043, 1, 0, 0},{22043, 1, 1, 18}, {0, 0, 0, 0}}},{8604, {{22043, 1, 0, 0},{22043, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8605, {{22044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8606, {{22045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8607, {{22046, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8608, {{22047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59437, {{22048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{8609, {{22056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8610, {{22057, 1, 0, 0},{22057, 1, 1, 17}, {0, 0, 0, 0}}},{8610, {{22057, 1, 0, 0},{22057, 1, 1, 17}, {0, 0, 0, 0}}},{8611, {{22057, 1, 0, 0},{22057, 1, 1, 18}, {0, 0, 0, 0}}},{8611, {{22057, 1, 0, 0},{22057, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8612, {{22058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8613, {{22059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8614, {{22060, 1, 0, 0},{22060, 1, 1, 18}, {0, 0, 0, 0}}},{8614, {{22060, 1, 0, 0},{22060, 1, 1, 18}, {0, 0, 0, 0}}},{8615, {{22060, 1, 0, 0},{22060, 1, 1, 17}, {0, 0, 0, 0}}},{8615, {{22060, 1, 0, 0},{22060, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{8616, {{22061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8617, {{22062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8618, {{22063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8619, {{22064, 1, 0, 0},{22064, 1, 1, 17}, {0, 0, 0, 0}}},{8619, {{22064, 1, 0, 0},{22064, 1, 1, 17}, {0, 0, 0, 0}}},{8620, {{22064, 1, 0, 0},{22064, 1, 1, 18}, {0, 0, 0, 0}}},{8620, {{22064, 1, 0, 0},{22064, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8621, {{22065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8622, {{22066, 1, 0, 0},{22066, 1, 1, 19}, {0, 0, 0, 0}}},{8622, {{22066, 1, 0, 0},{22066, 1, 1, 19}, {0, 0, 0, 0}}},{8623, {{22066, 1, 0, 0},{22066, 1, 1, 18}, {0, 0, 0, 0}}},{8623, {{22066, 1, 0, 0},{22066, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8624, {{22067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8625, {{22068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8626, {{22069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8627, {{22070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8629, {{22071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8630, {{22072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8631, {{22073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8632, {{22074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8633, {{22075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8634, {{22076, 1, 0, 0},{22076, 1, 1, 17}, {0, 0, 0, 0}}},{8634, {{22076, 1, 0, 0},{22076, 1, 1, 17}, {0, 0, 0, 0}}},{59434, {{22076, 1, 0, 0},{22076, 1, 1, 18}, {0, 0, 0, 0}}},{59434, {{22076, 1, 0, 0},{22076, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8635, {{22077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8636, {{22078, 1, 0, 0},{22078, 1, 1, 16}, {0, 0, 0, 0}}},{8636, {{22078, 1, 0, 0},{22078, 1, 1, 16}, {0, 0, 0, 0}}},{8637, {{22078, 1, 0, 0},{22078, 1, 1, 17}, {0, 0, 0, 0}}},{8637, {{22078, 1, 0, 0},{22078, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{8638, {{22079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8639, {{22080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8640, {{22081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{8641, {{22082, 1, 0, 0},{22082, 1, 1, 19}, {0, 0, 0, 0}}},{8641, {{22082, 1, 0, 0},{22082, 1, 1, 19}, {0, 0, 0, 0}}},{8642, {{22082, 1, 0, 0},{22082, 1, 1, 18}, {0, 0, 0, 0}}},{8642, {{22082, 1, 0, 0},{22082, 1, 1, 18}, {0, 0, 0, 0}}},{8643, {{22082, 1, 0, 0},{22082, 1, 1, 20}, {0, 0, 0, 0}}},{8643, {{22082, 1, 0, 0},{22082, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{8644, {{22083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8645, {{22084, 1, 0, 0},{22084, 1, 1, 17}, {0, 0, 0, 0}}},{8645, {{22084, 1, 0, 0},{22084, 1, 1, 17}, {0, 0, 0, 0}}},{8646, {{22084, 1, 0, 0},{22084, 1, 1, 18}, {0, 0, 0, 0}}},{8646, {{22084, 1, 0, 0},{22084, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8647, {{22085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8648, {{22086, 1, 0, 0},{22086, 1, 1, 18}, {0, 0, 0, 0}}},{8648, {{22086, 1, 0, 0},{22086, 1, 1, 18}, {0, 0, 0, 0}}},{8649, {{22086, 1, 0, 0},{22086, 1, 1, 17}, {0, 0, 0, 0}}},{8649, {{22086, 1, 0, 0},{22086, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{8650, {{22087, 1, 0, 0},{22087, 1, 1, 17}, {0, 0, 0, 0}}},{8650, {{22087, 1, 0, 0},{22087, 1, 1, 17}, {0, 0, 0, 0}}},{8651, {{22087, 1, 0, 0},{22087, 1, 1, 18}, {0, 0, 0, 0}}},{8651, {{22087, 1, 0, 0},{22087, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8652, {{22088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8653, {{22089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8654, {{22090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8655, {{22091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8656, {{22092, 1, 0, 0},{22092, 1, 1, 18}, {0, 0, 0, 0}}},{8656, {{22092, 1, 0, 0},{22092, 1, 1, 18}, {0, 0, 0, 0}}},{8657, {{22092, 1, 0, 0},{22092, 1, 1, 19}, {0, 0, 0, 0}}},{8657, {{22092, 1, 0, 0},{22092, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{8658, {{22093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8659, {{22094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8660, {{22095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8661, {{22096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8628, {{22097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8662, {{22098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8663, {{22099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8664, {{22100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8665, {{22101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8666, {{22102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8667, {{22103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8668, {{22104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8669, {{22105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8670, {{22106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8671, {{22107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{8672, {{22110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8673, {{22111, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8674, {{22112, 1, 0, 0},{22112, 1, 1, 18}, {0, 0, 0, 0}}},{8674, {{22112, 1, 0, 0},{22112, 1, 1, 18}, {0, 0, 0, 0}}},{8675, {{22112, 1, 0, 0},{22112, 1, 1, 17}, {0, 0, 0, 0}}},{8675, {{22112, 1, 0, 0},{22112, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{8676, {{22113, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8677, {{22114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8678, {{22115, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8679, {{22116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8680, {{22117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8681, {{22118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8682, {{22119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {9, {{8683, {{22120, 1, 0, 0},{22120, 1, 1, 18}, {0, 0, 0, 0}}},{8683, {{22120, 1, 0, 0},{22120, 1, 1, 18}, {0, 0, 0, 0}}},{30252, {{22120, 1, 0, 0},{22120, 1, 1, 0},{22120, 1, 1, 19},{64056, 1, 0, 0}, {0, 0, 0, 0}}},{30252, {{22120, 1, 0, 0},{22120, 1, 1, 0},{22120, 1, 1, 19},{64056, 1, 0, 0}, {0, 0, 0, 0}}},{30252, {{22120, 1, 0, 0},{22120, 1, 1, 0},{22120, 1, 1, 19},{64056, 1, 0, 0}, {0, 0, 0, 0}}},{57164, {{22120, 1, 0, 0},{22120, 1, 1, 20}, {0, 0, 0, 0}}},{57164, {{22120, 1, 0, 0},{22120, 1, 1, 20}, {0, 0, 0, 0}}},{58033, {{22120, 1, 0, 0},{22120, 1, 1, 21}, {0, 0, 0, 0}}},{58033, {{22120, 1, 0, 0},{22120, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{8684, {{22121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8685, {{22122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8686, {{22123, 1, 0, 0},{22123, 1, 1, 17}, {0, 0, 0, 0}}},{8686, {{22123, 1, 0, 0},{22123, 1, 1, 17}, {0, 0, 0, 0}}},{8687, {{22123, 1, 0, 0},{22123, 1, 1, 18}, {0, 0, 0, 0}}},{8687, {{22123, 1, 0, 0},{22123, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8688, {{22124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8689, {{22125, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8690, {{22126, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8691, {{22127, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8692, {{22128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8693, {{22129, 1, 0, 0},{22129, 1, 1, 17}, {0, 0, 0, 0}}},{8693, {{22129, 1, 0, 0},{22129, 1, 1, 17}, {0, 0, 0, 0}}},{8694, {{22129, 1, 0, 0},{22129, 1, 1, 18}, {0, 0, 0, 0}}},{8694, {{22129, 1, 0, 0},{22129, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8695, {{22130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8696, {{22131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{8697, {{22132, 1, 0, 0},{22132, 1, 1, 18}, {0, 0, 0, 0}}},{8697, {{22132, 1, 0, 0},{22132, 1, 1, 18}, {0, 0, 0, 0}}},{8698, {{22132, 1, 0, 0},{22132, 1, 1, 19}, {0, 0, 0, 0}}},{8698, {{22132, 1, 0, 0},{22132, 1, 1, 19}, {0, 0, 0, 0}}},{8699, {{22132, 1, 0, 0},{22132, 1, 1, 20}, {0, 0, 0, 0}}},{8699, {{22132, 1, 0, 0},{22132, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{8700, {{22133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8701, {{22134, 1, 0, 0},{22134, 1, 1, 18}, {0, 0, 0, 0}}},{8701, {{22134, 1, 0, 0},{22134, 1, 1, 18}, {0, 0, 0, 0}}},{8702, {{22134, 1, 0, 0},{22134, 1, 1, 17}, {0, 0, 0, 0}}},{8702, {{22134, 1, 0, 0},{22134, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{8703, {{22135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8704, {{22136, 1, 0, 0},{22136, 1, 1, 18}, {0, 0, 0, 0}}},{8704, {{22136, 1, 0, 0},{22136, 1, 1, 18}, {0, 0, 0, 0}}},{8705, {{22136, 1, 0, 0},{22136, 1, 1, 19}, {0, 0, 0, 0}}},{8705, {{22136, 1, 0, 0},{22136, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{8706, {{22137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8707, {{22138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{8708, {{22141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8709, {{22142, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8710, {{22143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8711, {{22144, 1, 0, 0},{22144, 1, 1, 17}, {0, 0, 0, 0}}},{8711, {{22144, 1, 0, 0},{22144, 1, 1, 17}, {0, 0, 0, 0}}},{8712, {{22144, 1, 0, 0},{22144, 1, 1, 18}, {0, 0, 0, 0}}},{8712, {{22144, 1, 0, 0},{22144, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8713, {{22145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8714, {{22146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8715, {{22147, 1, 0, 0},{22147, 1, 1, 16}, {0, 0, 0, 0}}},{8715, {{22147, 1, 0, 0},{22147, 1, 1, 16}, {0, 0, 0, 0}}},{8716, {{22147, 1, 0, 0},{22147, 1, 1, 17}, {0, 0, 0, 0}}},{8716, {{22147, 1, 0, 0},{22147, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{8717, {{22148, 1, 0, 0},{22148, 1, 1, 17}, {0, 0, 0, 0}}},{8717, {{22148, 1, 0, 0},{22148, 1, 1, 17}, {0, 0, 0, 0}}},{8718, {{22148, 1, 0, 0},{22148, 1, 1, 18}, {0, 0, 0, 0}}},{8718, {{22148, 1, 0, 0},{22148, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8719, {{22149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8720, {{22150, 1, 0, 0},{22150, 1, 1, 17}, {0, 0, 0, 0}}},{8720, {{22150, 1, 0, 0},{22150, 1, 1, 17}, {0, 0, 0, 0}}},{8721, {{22150, 1, 0, 0},{22150, 1, 1, 18}, {0, 0, 0, 0}}},{8721, {{22150, 1, 0, 0},{22150, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8722, {{22151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8723, {{22152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8724, {{22153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8725, {{22154, 1, 0, 0},{22154, 1, 1, 17}, {0, 0, 0, 0}}},{8725, {{22154, 1, 0, 0},{22154, 1, 1, 17}, {0, 0, 0, 0}}},{8726, {{22154, 1, 0, 0},{22154, 1, 1, 18}, {0, 0, 0, 0}}},{8726, {{22154, 1, 0, 0},{22154, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8727, {{22155, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8728, {{22156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8729, {{22157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8730, {{22158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8731, {{22159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8732, {{22160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{8733, {{22164, 1, 0, 0},{22164, 1, 1, 17}, {0, 0, 0, 0}}},{8733, {{22164, 1, 0, 0},{22164, 1, 1, 17}, {0, 0, 0, 0}}},{8734, {{22164, 1, 0, 0},{22164, 1, 1, 18}, {0, 0, 0, 0}}},{8734, {{22164, 1, 0, 0},{22164, 1, 1, 18}, {0, 0, 0, 0}}},{8735, {{22164, 1, 0, 0},{22164, 1, 1, 19}, {0, 0, 0, 0}}},{8735, {{22164, 1, 0, 0},{22164, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{8736, {{22165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8737, {{22166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8738, {{22167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8739, {{22168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8740, {{22169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8741, {{22170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8742, {{22171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8743, {{22172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8744, {{22173, 1, 0, 0},{22173, 1, 1, 17}, {0, 0, 0, 0}}},{8744, {{22173, 1, 0, 0},{22173, 1, 1, 17}, {0, 0, 0, 0}}},{8745, {{22173, 1, 0, 0},{22173, 1, 1, 18}, {0, 0, 0, 0}}},{8745, {{22173, 1, 0, 0},{22173, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8746, {{22174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8748, {{22175, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8749, {{22176, 1, 0, 0},{22176, 1, 1, 17}, {0, 0, 0, 0}}},{8749, {{22176, 1, 0, 0},{22176, 1, 1, 17}, {0, 0, 0, 0}}},{8750, {{22176, 1, 0, 0},{22176, 1, 1, 18}, {0, 0, 0, 0}}},{8750, {{22176, 1, 0, 0},{22176, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{8751, {{22178, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{8752, {{22181, 1, 0, 0},{22181, 1, 1, 18}, {0, 0, 0, 0}}},{8752, {{22181, 1, 0, 0},{22181, 1, 1, 18}, {0, 0, 0, 0}}},{8753, {{22181, 1, 0, 0},{22181, 1, 1, 19}, {0, 0, 0, 0}}},{8753, {{22181, 1, 0, 0},{22181, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{8754, {{22182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8756, {{22183, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8757, {{22184, 1, 0, 0},{22184, 1, 1, 17}, {0, 0, 0, 0}}},{8757, {{22184, 1, 0, 0},{22184, 1, 1, 17}, {0, 0, 0, 0}}},{8758, {{22184, 1, 0, 0},{22184, 1, 1, 18}, {0, 0, 0, 0}}},{8758, {{22184, 1, 0, 0},{22184, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8759, {{22185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8760, {{22186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8761, {{22187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8762, {{22188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8763, {{22189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{8764, {{22190, 1, 0, 0},{22190, 1, 1, 20}, {0, 0, 0, 0}}},{8764, {{22190, 1, 0, 0},{22190, 1, 1, 20}, {0, 0, 0, 0}}},{8765, {{22190, 1, 0, 0},{22190, 1, 1, 21}, {0, 0, 0, 0}}},{8765, {{22190, 1, 0, 0},{22190, 1, 1, 21}, {0, 0, 0, 0}}},{8766, {{22190, 1, 0, 0},{22190, 1, 1, 22}, {0, 0, 0, 0}}},{8766, {{22190, 1, 0, 0},{22190, 1, 1, 22}, {0, 0, 0, 0}}},{8767, {{22190, 1, 0, 0},{22190, 1, 1, 23}, {0, 0, 0, 0}}},{8767, {{22190, 1, 0, 0},{22190, 1, 1, 23}, {0, 0, 0, 0}}}}},
    {1, {{8768, {{22191, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{8769, {{22193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8770, {{22194, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8771, {{22195, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8773, {{22196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8774, {{22197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8775, {{22198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8776, {{22199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{8777, {{22203, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8778, {{22204, 1, 0, 0},{22204, 1, 1, 17}, {0, 0, 0, 0}}},{8778, {{22204, 1, 0, 0},{22204, 1, 1, 17}, {0, 0, 0, 0}}},{8779, {{22204, 1, 0, 0},{22204, 1, 1, 18}, {0, 0, 0, 0}}},{8779, {{22204, 1, 0, 0},{22204, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8780, {{22205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8781, {{22206, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8782, {{22207, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8783, {{22208, 1, 0, 0},{22208, 1, 1, 19}, {0, 0, 0, 0}}},{8783, {{22208, 1, 0, 0},{22208, 1, 1, 19}, {0, 0, 0, 0}}},{8784, {{22208, 1, 0, 0},{22208, 1, 1, 18}, {0, 0, 0, 0}}},{8784, {{22208, 1, 0, 0},{22208, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{8785, {{22209, 1, 0, 0},{22209, 1, 1, 18}, {0, 0, 0, 0}}},{8785, {{22209, 1, 0, 0},{22209, 1, 1, 18}, {0, 0, 0, 0}}},{8786, {{22209, 1, 0, 0},{22209, 1, 1, 19}, {0, 0, 0, 0}}},{8786, {{22209, 1, 0, 0},{22209, 1, 1, 19}, {0, 0, 0, 0}}},{8787, {{22209, 1, 0, 0},{22209, 1, 1, 20}, {0, 0, 0, 0}}},{8787, {{22209, 1, 0, 0},{22209, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{8788, {{22210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8789, {{22211, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{8790, {{22213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8791, {{22214, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8792, {{22215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8793, {{22216, 1, 0, 0},{22216, 1, 1, 17}, {0, 0, 0, 0}}},{8793, {{22216, 1, 0, 0},{22216, 1, 1, 17}, {0, 0, 0, 0}}},{8794, {{22216, 1, 0, 0},{22216, 1, 1, 18}, {0, 0, 0, 0}}},{8794, {{22216, 1, 0, 0},{22216, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8795, {{22217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8796, {{22218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8797, {{22219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8798, {{22220, 1, 0, 0},{22220, 1, 1, 17}, {0, 0, 0, 0}}},{8798, {{22220, 1, 0, 0},{22220, 1, 1, 17}, {0, 0, 0, 0}}},{8799, {{22220, 1, 0, 0},{22220, 1, 1, 18}, {0, 0, 0, 0}}},{8799, {{22220, 1, 0, 0},{22220, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8800, {{22221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{8801, {{22222, 1, 0, 0},{22222, 1, 1, 18}, {0, 0, 0, 0}}},{8801, {{22222, 1, 0, 0},{22222, 1, 1, 18}, {0, 0, 0, 0}}},{8802, {{22222, 1, 0, 0},{22222, 1, 1, 19}, {0, 0, 0, 0}}},{8802, {{22222, 1, 0, 0},{22222, 1, 1, 19}, {0, 0, 0, 0}}},{8803, {{22222, 1, 0, 0},{22222, 1, 1, 20}, {0, 0, 0, 0}}},{8803, {{22222, 1, 0, 0},{22222, 1, 1, 20}, {0, 0, 0, 0}}},{59441, {{22222, 1, 0, 0},{22222, 1, 1, 21}, {0, 0, 0, 0}}},{59441, {{22222, 1, 0, 0},{22222, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{8804, {{22223, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8805, {{22224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8806, {{22225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8807, {{22226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{8808, {{22227, 1, 0, 0},{22227, 1, 1, 17}, {0, 0, 0, 0}}},{8808, {{22227, 1, 0, 0},{22227, 1, 1, 17}, {0, 0, 0, 0}}},{8809, {{22227, 1, 0, 0},{22227, 1, 1, 19}, {0, 0, 0, 0}}},{8809, {{22227, 1, 0, 0},{22227, 1, 1, 19}, {0, 0, 0, 0}}},{8810, {{22227, 1, 0, 0},{22227, 1, 1, 18}, {0, 0, 0, 0}}},{8810, {{22227, 1, 0, 0},{22227, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8811, {{22228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{8812, {{22231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8813, {{22232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8814, {{22233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8815, {{22234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8816, {{22235, 1, 0, 0},{22235, 1, 1, 17}, {0, 0, 0, 0}}},{8816, {{22235, 1, 0, 0},{22235, 1, 1, 17}, {0, 0, 0, 0}}},{8817, {{22235, 1, 0, 0},{22235, 1, 1, 18}, {0, 0, 0, 0}}},{8817, {{22235, 1, 0, 0},{22235, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8818, {{22236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8819, {{22237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8820, {{22238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8821, {{22239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8822, {{22240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8823, {{22241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8824, {{22242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8826, {{22243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8827, {{22244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8828, {{22245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8829, {{22246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8830, {{22247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8831, {{22248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8832, {{22249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57204, {{22250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8833, {{22251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8834, {{22252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8835, {{22253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8836, {{22254, 1, 0, 0},{22254, 1, 1, 18}, {0, 0, 0, 0}}},{8836, {{22254, 1, 0, 0},{22254, 1, 1, 18}, {0, 0, 0, 0}}},{8837, {{22254, 1, 0, 0},{22254, 1, 1, 19}, {0, 0, 0, 0}}},{8837, {{22254, 1, 0, 0},{22254, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{8838, {{22255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8839, {{22256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8840, {{22257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8841, {{22258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8842, {{22259, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{8843, {{22262, 1, 0, 0},{22262, 1, 1, 17}, {0, 0, 0, 0}}},{8843, {{22262, 1, 0, 0},{22262, 1, 1, 17}, {0, 0, 0, 0}}},{8844, {{22262, 1, 0, 0},{22262, 1, 1, 18}, {0, 0, 0, 0}}},{8844, {{22262, 1, 0, 0},{22262, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8845, {{22263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8846, {{22264, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8847, {{22265, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8848, {{22266, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8849, {{22267, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8850, {{22268, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8851, {{22269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{8852, {{22271, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8853, {{22272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8854, {{22273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8855, {{22274, 1, 0, 0},{22274, 1, 1, 17}, {0, 0, 0, 0}}},{8855, {{22274, 1, 0, 0},{22274, 1, 1, 17}, {0, 0, 0, 0}}},{8856, {{22274, 1, 0, 0},{22274, 1, 1, 18}, {0, 0, 0, 0}}},{8856, {{22274, 1, 0, 0},{22274, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8857, {{22275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8858, {{22276, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8859, {{22277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{8860, {{22279, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8861, {{22280, 1, 0, 0},{22280, 1, 1, 17}, {0, 0, 0, 0}}},{8861, {{22280, 1, 0, 0},{22280, 1, 1, 17}, {0, 0, 0, 0}}},{8862, {{22280, 1, 0, 0},{22280, 1, 1, 18}, {0, 0, 0, 0}}},{8862, {{22280, 1, 0, 0},{22280, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8863, {{22281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8864, {{22282, 1, 0, 0},{22282, 1, 1, 17}, {0, 0, 0, 0}}},{8864, {{22282, 1, 0, 0},{22282, 1, 1, 17}, {0, 0, 0, 0}}},{8865, {{22282, 1, 0, 0},{22282, 1, 1, 18}, {0, 0, 0, 0}}},{8865, {{22282, 1, 0, 0},{22282, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8866, {{22283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8867, {{22284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{8868, {{22285, 1, 0, 0},{22285, 1, 1, 18}, {0, 0, 0, 0}}},{8868, {{22285, 1, 0, 0},{22285, 1, 1, 18}, {0, 0, 0, 0}}},{8869, {{22285, 1, 0, 0},{22285, 1, 1, 19}, {0, 0, 0, 0}}},{8869, {{22285, 1, 0, 0},{22285, 1, 1, 19}, {0, 0, 0, 0}}},{8870, {{22285, 1, 0, 0},{22285, 1, 1, 20}, {0, 0, 0, 0}}},{8870, {{22285, 1, 0, 0},{22285, 1, 1, 20}, {0, 0, 0, 0}}},{8871, {{22285, 1, 0, 0},{22285, 1, 1, 21}, {0, 0, 0, 0}}},{8871, {{22285, 1, 0, 0},{22285, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{8872, {{22286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8873, {{22287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{8874, {{22289, 1, 0, 0},{22289, 1, 1, 17}, {0, 0, 0, 0}}},{8874, {{22289, 1, 0, 0},{22289, 1, 1, 17}, {0, 0, 0, 0}}},{8875, {{22289, 1, 0, 0},{22289, 1, 1, 18}, {0, 0, 0, 0}}},{8875, {{22289, 1, 0, 0},{22289, 1, 1, 18}, {0, 0, 0, 0}}},{59444, {{22289, 1, 0, 0},{22289, 1, 1, 19}, {0, 0, 0, 0}}},{59444, {{22289, 1, 0, 0},{22289, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{8876, {{22290, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8877, {{22291, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8878, {{22292, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8879, {{22293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{8880, {{22294, 1, 0, 0},{22294, 1, 1, 17}, {0, 0, 0, 0}}},{8880, {{22294, 1, 0, 0},{22294, 1, 1, 17}, {0, 0, 0, 0}}},{8881, {{22294, 1, 0, 0},{22294, 1, 1, 19}, {0, 0, 0, 0}}},{8881, {{22294, 1, 0, 0},{22294, 1, 1, 19}, {0, 0, 0, 0}}},{57207, {{22294, 1, 0, 0},{22294, 1, 1, 18}, {0, 0, 0, 0}}},{57207, {{22294, 1, 0, 0},{22294, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{8882, {{22295, 1, 0, 0},{22295, 1, 1, 16}, {0, 0, 0, 0}}},{8882, {{22295, 1, 0, 0},{22295, 1, 1, 16}, {0, 0, 0, 0}}},{8883, {{22295, 1, 0, 0},{22295, 1, 1, 17}, {0, 0, 0, 0}}},{8883, {{22295, 1, 0, 0},{22295, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{8885, {{22296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{8886, {{22298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8887, {{22299, 1, 0, 0},{22299, 1, 1, 17}, {0, 0, 0, 0}}},{8887, {{22299, 1, 0, 0},{22299, 1, 1, 17}, {0, 0, 0, 0}}},{8888, {{22299, 1, 0, 0},{22299, 1, 1, 18}, {0, 0, 0, 0}}},{8888, {{22299, 1, 0, 0},{22299, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8889, {{22300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8890, {{22301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8891, {{22302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8892, {{22303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8893, {{22304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8894, {{22305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8895, {{22306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8896, {{22307, 1, 0, 0},{22307, 1, 1, 17}, {0, 0, 0, 0}}},{8896, {{22307, 1, 0, 0},{22307, 1, 1, 17}, {0, 0, 0, 0}}},{8897, {{22307, 1, 0, 0},{22307, 1, 1, 18}, {0, 0, 0, 0}}},{8897, {{22307, 1, 0, 0},{22307, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8898, {{22308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8899, {{22309, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8900, {{22310, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8901, {{22311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8902, {{22312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8903, {{22313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8904, {{22314, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8905, {{22315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8906, {{22316, 1, 0, 0},{22316, 1, 1, 17}, {0, 0, 0, 0}}},{8906, {{22316, 1, 0, 0},{22316, 1, 1, 17}, {0, 0, 0, 0}}},{8907, {{22316, 1, 0, 0},{22316, 1, 1, 18}, {0, 0, 0, 0}}},{8907, {{22316, 1, 0, 0},{22316, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{8908, {{22317, 1, 0, 0},{22317, 1, 1, 17}, {0, 0, 0, 0}}},{8908, {{22317, 1, 0, 0},{22317, 1, 1, 17}, {0, 0, 0, 0}}},{57213, {{22317, 1, 0, 0},{22317, 1, 1, 18}, {0, 0, 0, 0}}},{57213, {{22317, 1, 0, 0},{22317, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8909, {{22318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8911, {{22319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8912, {{22320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8913, {{22321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8914, {{22322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8915, {{22323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8916, {{22324, 1, 0, 0},{22324, 1, 1, 17}, {0, 0, 0, 0}}},{8916, {{22324, 1, 0, 0},{22324, 1, 1, 17}, {0, 0, 0, 0}}},{8917, {{22324, 1, 0, 0},{22324, 1, 1, 18}, {0, 0, 0, 0}}},{8917, {{22324, 1, 0, 0},{22324, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8918, {{22325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8919, {{22326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8920, {{22327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8921, {{22328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{8922, {{22331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8923, {{22332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8924, {{22333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8925, {{22334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8926, {{22335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8927, {{22336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8928, {{22337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8929, {{22338, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8930, {{22339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8931, {{22340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8932, {{22341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8933, {{22342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8934, {{22343, 1, 0, 0},{22343, 1, 1, 18}, {0, 0, 0, 0}}},{8934, {{22343, 1, 0, 0},{22343, 1, 1, 18}, {0, 0, 0, 0}}},{8935, {{22343, 1, 0, 0},{22343, 1, 1, 19}, {0, 0, 0, 0}}},{8935, {{22343, 1, 0, 0},{22343, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{8936, {{22344, 1, 0, 0},{22344, 1, 1, 16}, {0, 0, 0, 0}}},{8936, {{22344, 1, 0, 0},{22344, 1, 1, 16}, {0, 0, 0, 0}}},{8937, {{22344, 1, 0, 0},{22344, 1, 1, 17}, {0, 0, 0, 0}}},{8937, {{22344, 1, 0, 0},{22344, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{8938, {{22345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8939, {{22346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8940, {{22347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8941, {{22348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8942, {{22349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8943, {{22350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8944, {{22351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8945, {{22352, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8946, {{22353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8947, {{22354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8948, {{22355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8949, {{22356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8950, {{22357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8951, {{22358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8952, {{22359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8953, {{22360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8954, {{22361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8955, {{22362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{8957, {{22369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8958, {{22370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8959, {{22371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8960, {{22372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8961, {{22373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8962, {{22374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8963, {{22375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8964, {{22376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8965, {{22377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8966, {{22378, 1, 0, 0},{22378, 1, 1, 19}, {0, 0, 0, 0}}},{8966, {{22378, 1, 0, 0},{22378, 1, 1, 19}, {0, 0, 0, 0}}},{8967, {{22378, 1, 0, 0},{22378, 1, 1, 18}, {0, 0, 0, 0}}},{8967, {{22378, 1, 0, 0},{22378, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{8968, {{22379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8969, {{22380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8970, {{22381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8971, {{22382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8972, {{22383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8973, {{22384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8974, {{22385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8975, {{22386, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8976, {{22387, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8977, {{22388, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8979, {{22389, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8980, {{22390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8981, {{22391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8982, {{22392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8983, {{22393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8984, {{22394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8985, {{22395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8986, {{22396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8987, {{22397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8989, {{22398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8990, {{22399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8991, {{22400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8992, {{22401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8993, {{22402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8994, {{22403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{8995, {{22408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{8996, {{22409, 1, 0, 0},{22409, 1, 1, 17}, {0, 0, 0, 0}}},{8996, {{22409, 1, 0, 0},{22409, 1, 1, 17}, {0, 0, 0, 0}}},{8997, {{22409, 1, 0, 0},{22409, 1, 1, 18}, {0, 0, 0, 0}}},{8997, {{22409, 1, 0, 0},{22409, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{8998, {{22411, 1, 0, 0},{22411, 1, 1, 17}, {0, 0, 0, 0}}},{8998, {{22411, 1, 0, 0},{22411, 1, 1, 17}, {0, 0, 0, 0}}},{8999, {{22411, 1, 0, 0},{22411, 1, 1, 18}, {0, 0, 0, 0}}},{8999, {{22411, 1, 0, 0},{22411, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9000, {{22412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9001, {{22413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9002, {{22414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9003, {{22415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9004, {{22416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9005, {{22417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9006, {{22418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9007, {{22419, 1, 0, 0},{22419, 1, 1, 17}, {0, 0, 0, 0}}},{9007, {{22419, 1, 0, 0},{22419, 1, 1, 17}, {0, 0, 0, 0}}},{9008, {{22419, 1, 0, 0},{22419, 1, 1, 18}, {0, 0, 0, 0}}},{9008, {{22419, 1, 0, 0},{22419, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{9009, {{22420, 1, 0, 0},{22420, 1, 1, 17}, {0, 0, 0, 0}}},{9009, {{22420, 1, 0, 0},{22420, 1, 1, 17}, {0, 0, 0, 0}}},{9010, {{22420, 1, 0, 0},{22420, 1, 1, 18}, {0, 0, 0, 0}}},{9010, {{22420, 1, 0, 0},{22420, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9011, {{22421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9012, {{22422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9013, {{22423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9014, {{22424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9015, {{22425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9016, {{22426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9017, {{22427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9018, {{22428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9019, {{22429, 1, 0, 0},{22429, 1, 1, 17}, {0, 0, 0, 0}}},{9019, {{22429, 1, 0, 0},{22429, 1, 1, 17}, {0, 0, 0, 0}}},{9020, {{22429, 1, 0, 0},{22429, 1, 1, 18}, {0, 0, 0, 0}}},{9020, {{22429, 1, 0, 0},{22429, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9021, {{22430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9022, {{22431, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9023, {{22432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9024, {{22433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9025, {{22434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9026, {{22435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9027, {{22436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9028, {{22437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9029, {{22438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9030, {{22439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9031, {{22440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9032, {{22441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9033, {{22442, 1, 0, 0},{22442, 1, 1, 17}, {0, 0, 0, 0}}},{9033, {{22442, 1, 0, 0},{22442, 1, 1, 17}, {0, 0, 0, 0}}},{57228, {{22442, 1, 0, 0},{22442, 1, 1, 18}, {0, 0, 0, 0}}},{57228, {{22442, 1, 0, 0},{22442, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{9034, {{22444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{9035, {{22448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{9036, {{22451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{9037, {{22454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9038, {{22455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9039, {{22456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9040, {{22457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9041, {{22458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9042, {{22459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9043, {{22460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9044, {{22461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9045, {{22462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9046, {{22463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9047, {{22464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9048, {{22465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9049, {{22466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9050, {{22467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9051, {{22468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9052, {{22469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9053, {{22470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9054, {{22471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9055, {{22472, 1, 0, 0},{22472, 1, 1, 18}, {0, 0, 0, 0}}},{9055, {{22472, 1, 0, 0},{22472, 1, 1, 18}, {0, 0, 0, 0}}},{9056, {{22472, 1, 0, 0},{22472, 1, 1, 17}, {0, 0, 0, 0}}},{9056, {{22472, 1, 0, 0},{22472, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{9057, {{22473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9058, {{22474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9059, {{22475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9060, {{22476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9061, {{22477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9062, {{22478, 1, 0, 0},{22478, 1, 1, 18}, {0, 0, 0, 0}}},{9062, {{22478, 1, 0, 0},{22478, 1, 1, 18}, {0, 0, 0, 0}}},{9063, {{22478, 1, 0, 0},{22478, 1, 1, 19}, {0, 0, 0, 0}}},{9063, {{22478, 1, 0, 0},{22478, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{9064, {{22479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9065, {{22480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9066, {{22481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9067, {{22482, 1, 0, 0},{22482, 1, 1, 17}, {0, 0, 0, 0}}},{9067, {{22482, 1, 0, 0},{22482, 1, 1, 17}, {0, 0, 0, 0}}},{9068, {{22482, 1, 0, 0},{22482, 1, 1, 18}, {0, 0, 0, 0}}},{9068, {{22482, 1, 0, 0},{22482, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{9069, {{22483, 1, 0, 0},{22483, 1, 1, 18}, {0, 0, 0, 0}}},{9069, {{22483, 1, 0, 0},{22483, 1, 1, 18}, {0, 0, 0, 0}}},{9070, {{22483, 1, 0, 0},{22483, 1, 1, 17}, {0, 0, 0, 0}}},{9070, {{22483, 1, 0, 0},{22483, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{9072, {{22484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9073, {{22485, 1, 0, 0},{22485, 1, 1, 17}, {0, 0, 0, 0}}},{9073, {{22485, 1, 0, 0},{22485, 1, 1, 17}, {0, 0, 0, 0}}},{9074, {{22485, 1, 0, 0},{22485, 1, 1, 18}, {0, 0, 0, 0}}},{9074, {{22485, 1, 0, 0},{22485, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{9075, {{22486, 1, 0, 0},{22486, 1, 1, 18}, {0, 0, 0, 0}}},{9075, {{22486, 1, 0, 0},{22486, 1, 1, 18}, {0, 0, 0, 0}}},{57226, {{22486, 1, 0, 0},{22486, 1, 1, 19}, {0, 0, 0, 0}}},{57226, {{22486, 1, 0, 0},{22486, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{9076, {{22487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{9077, {{22492, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9078, {{22493, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9079, {{22494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9080, {{22495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9081, {{22496, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9082, {{22497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9083, {{22498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9085, {{22499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9086, {{22500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9087, {{22501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9089, {{22502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9090, {{22503, 1, 0, 0},{22503, 1, 1, 17}, {0, 0, 0, 0}}},{9090, {{22503, 1, 0, 0},{22503, 1, 1, 17}, {0, 0, 0, 0}}},{9091, {{22503, 1, 0, 0},{22503, 1, 1, 18}, {0, 0, 0, 0}}},{9091, {{22503, 1, 0, 0},{22503, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9092, {{22504, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9093, {{22505, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9094, {{22506, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9097, {{22507, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9098, {{22508, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9099, {{22509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9100, {{22510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9101, {{22511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9102, {{22512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9103, {{22513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9104, {{22514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9105, {{22515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9106, {{22516, 1, 0, 0},{22516, 1, 1, 20}, {0, 0, 0, 0}}},{9106, {{22516, 1, 0, 0},{22516, 1, 1, 20}, {0, 0, 0, 0}}},{9107, {{22516, 1, 0, 0},{22516, 1, 1, 19}, {0, 0, 0, 0}}},{9107, {{22516, 1, 0, 0},{22516, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{9108, {{22517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9109, {{22518, 1, 0, 0},{22518, 1, 1, 17}, {0, 0, 0, 0}}},{9109, {{22518, 1, 0, 0},{22518, 1, 1, 17}, {0, 0, 0, 0}}},{9110, {{22518, 1, 0, 0},{22518, 1, 1, 18}, {0, 0, 0, 0}}},{9110, {{22518, 1, 0, 0},{22518, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9111, {{22519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9112, {{22520, 1, 0, 0},{22520, 1, 1, 17}, {0, 0, 0, 0}}},{9112, {{22520, 1, 0, 0},{22520, 1, 1, 17}, {0, 0, 0, 0}}},{9113, {{22520, 1, 0, 0},{22520, 1, 1, 18}, {0, 0, 0, 0}}},{9113, {{22520, 1, 0, 0},{22520, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9114, {{22521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9115, {{22522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9116, {{22523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9117, {{22524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9118, {{22525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9119, {{22526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9120, {{22527, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9121, {{22528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9122, {{22529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9123, {{22530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9124, {{22531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9125, {{22532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9126, {{22533, 1, 0, 0},{22533, 1, 1, 17}, {0, 0, 0, 0}}},{9126, {{22533, 1, 0, 0},{22533, 1, 1, 17}, {0, 0, 0, 0}}},{57223, {{22533, 1, 0, 0},{22533, 1, 1, 18}, {0, 0, 0, 0}}},{57223, {{22533, 1, 0, 0},{22533, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9127, {{22534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9128, {{22535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9131, {{22536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9132, {{22537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9133, {{22538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9134, {{22539, 1, 0, 0},{22539, 1, 1, 19}, {0, 0, 0, 0}}},{9134, {{22539, 1, 0, 0},{22539, 1, 1, 19}, {0, 0, 0, 0}}},{9135, {{22539, 1, 0, 0},{22539, 1, 1, 18}, {0, 0, 0, 0}}},{9135, {{22539, 1, 0, 0},{22539, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9136, {{22540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9137, {{22541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{9139, {{22549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{9140, {{22551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9141, {{22552, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9142, {{22553, 1, 0, 0},{22553, 1, 1, 18}, {0, 0, 0, 0}}},{9142, {{22553, 1, 0, 0},{22553, 1, 1, 18}, {0, 0, 0, 0}}},{9143, {{22553, 1, 0, 0},{22553, 1, 1, 19}, {0, 0, 0, 0}}},{9143, {{22553, 1, 0, 0},{22553, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{9144, {{22554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9145, {{22555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9146, {{22556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9147, {{22557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9148, {{22558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9149, {{22559, 1, 0, 0},{22559, 1, 1, 17}, {0, 0, 0, 0}}},{9149, {{22559, 1, 0, 0},{22559, 1, 1, 17}, {0, 0, 0, 0}}},{9150, {{22559, 1, 0, 0},{22559, 1, 1, 18}, {0, 0, 0, 0}}},{9150, {{22559, 1, 0, 0},{22559, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9151, {{22560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9152, {{22561, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9153, {{22562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9154, {{22563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9155, {{22564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9156, {{22565, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9157, {{22566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9158, {{22567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9159, {{22568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9160, {{22569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9161, {{22570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9162, {{22571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9163, {{22572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9164, {{22573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9165, {{22574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9166, {{22575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9167, {{22576, 1, 0, 0},{22576, 1, 1, 17}, {0, 0, 0, 0}}},{9167, {{22576, 1, 0, 0},{22576, 1, 1, 17}, {0, 0, 0, 0}}},{9168, {{22576, 1, 0, 0},{22576, 1, 1, 19}, {0, 0, 0, 0}}},{9168, {{22576, 1, 0, 0},{22576, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{9169, {{22577, 1, 0, 0},{22577, 1, 1, 17}, {0, 0, 0, 0}}},{9169, {{22577, 1, 0, 0},{22577, 1, 1, 17}, {0, 0, 0, 0}}},{9170, {{22577, 1, 0, 0},{22577, 1, 1, 18}, {0, 0, 0, 0}}},{9170, {{22577, 1, 0, 0},{22577, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{9171, {{22578, 1, 0, 0},{22578, 1, 1, 17}, {0, 0, 0, 0}}},{9171, {{22578, 1, 0, 0},{22578, 1, 1, 17}, {0, 0, 0, 0}}},{9172, {{22578, 1, 0, 0},{22578, 1, 1, 18}, {0, 0, 0, 0}}},{9172, {{22578, 1, 0, 0},{22578, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9173, {{22579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9174, {{22580, 1, 0, 0},{22580, 1, 1, 17}, {0, 0, 0, 0}}},{9174, {{22580, 1, 0, 0},{22580, 1, 1, 17}, {0, 0, 0, 0}}},{9175, {{22580, 1, 0, 0},{22580, 1, 1, 18}, {0, 0, 0, 0}}},{9175, {{22580, 1, 0, 0},{22580, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{9176, {{22581, 1, 0, 0},{22581, 1, 1, 18}, {0, 0, 0, 0}}},{9176, {{22581, 1, 0, 0},{22581, 1, 1, 18}, {0, 0, 0, 0}}},{9177, {{22581, 1, 0, 0},{22581, 1, 1, 19}, {0, 0, 0, 0}}},{9177, {{22581, 1, 0, 0},{22581, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{9178, {{22582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9179, {{22583, 1, 0, 0},{22583, 1, 1, 16}, {0, 0, 0, 0}}},{9179, {{22583, 1, 0, 0},{22583, 1, 1, 16}, {0, 0, 0, 0}}},{9180, {{22583, 1, 0, 0},{22583, 1, 1, 17}, {0, 0, 0, 0}}},{9180, {{22583, 1, 0, 0},{22583, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{9181, {{22584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9182, {{22585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9183, {{22586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9184, {{22587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9185, {{22588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9186, {{22589, 1, 0, 0},{22589, 1, 1, 18}, {0, 0, 0, 0}}},{9186, {{22589, 1, 0, 0},{22589, 1, 1, 18}, {0, 0, 0, 0}}},{9187, {{22589, 1, 0, 0},{22589, 1, 1, 19}, {0, 0, 0, 0}}},{9187, {{22589, 1, 0, 0},{22589, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{9188, {{22590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9189, {{22591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{9190, {{22592, 1, 0, 0},{22592, 1, 1, 18}, {0, 0, 0, 0}}},{9190, {{22592, 1, 0, 0},{22592, 1, 1, 18}, {0, 0, 0, 0}}},{30253, {{22592, 1, 0, 0},{22592, 1, 1, 0},{22592, 1, 1, 19},{64057, 1, 0, 0}, {0, 0, 0, 0}}},{30253, {{22592, 1, 0, 0},{22592, 1, 1, 0},{22592, 1, 1, 19},{64057, 1, 0, 0}, {0, 0, 0, 0}}},{30253, {{22592, 1, 0, 0},{22592, 1, 1, 0},{22592, 1, 1, 19},{64057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9191, {{22593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{9192, {{22601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9193, {{22602, 1, 0, 0},{22602, 1, 1, 17}, {0, 0, 0, 0}}},{9193, {{22602, 1, 0, 0},{22602, 1, 1, 17}, {0, 0, 0, 0}}},{57234, {{22602, 1, 0, 0},{22602, 1, 1, 18}, {0, 0, 0, 0}}},{57234, {{22602, 1, 0, 0},{22602, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9194, {{22603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9195, {{22604, 1, 0, 0},{22604, 1, 1, 18}, {0, 0, 0, 0}}},{9195, {{22604, 1, 0, 0},{22604, 1, 1, 18}, {0, 0, 0, 0}}},{9196, {{22604, 1, 0, 0},{22604, 1, 1, 17}, {0, 0, 0, 0}}},{9196, {{22604, 1, 0, 0},{22604, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{9197, {{22605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9198, {{22606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9199, {{22607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9200, {{22608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9201, {{22609, 1, 0, 0},{22609, 1, 1, 18}, {0, 0, 0, 0}}},{9201, {{22609, 1, 0, 0},{22609, 1, 1, 18}, {0, 0, 0, 0}}},{9202, {{22609, 1, 0, 0},{22609, 1, 1, 17}, {0, 0, 0, 0}}},{9202, {{22609, 1, 0, 0},{22609, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{9203, {{22610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9204, {{22611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9205, {{22612, 1, 0, 0},{22612, 1, 1, 17}, {0, 0, 0, 0}}},{9205, {{22612, 1, 0, 0},{22612, 1, 1, 17}, {0, 0, 0, 0}}},{9206, {{22612, 1, 0, 0},{22612, 1, 1, 18}, {0, 0, 0, 0}}},{9206, {{22612, 1, 0, 0},{22612, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9207, {{22613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9208, {{22614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9209, {{22615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9210, {{22616, 1, 0, 0},{22616, 1, 1, 19}, {0, 0, 0, 0}}},{9210, {{22616, 1, 0, 0},{22616, 1, 1, 19}, {0, 0, 0, 0}}},{9211, {{22616, 1, 0, 0},{22616, 1, 1, 18}, {0, 0, 0, 0}}},{9211, {{22616, 1, 0, 0},{22616, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{9212, {{22617, 1, 0, 0},{22617, 1, 1, 18}, {0, 0, 0, 0}}},{9212, {{22617, 1, 0, 0},{22617, 1, 1, 18}, {0, 0, 0, 0}}},{9213, {{22617, 1, 0, 0},{22617, 1, 1, 20}, {0, 0, 0, 0}}},{9213, {{22617, 1, 0, 0},{22617, 1, 1, 20}, {0, 0, 0, 0}}},{9214, {{22617, 1, 0, 0},{22617, 1, 1, 19}, {0, 0, 0, 0}}},{9214, {{22617, 1, 0, 0},{22617, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {9, {{9215, {{22618, 1, 0, 0},{22618, 1, 1, 19}, {0, 0, 0, 0}}},{9215, {{22618, 1, 0, 0},{22618, 1, 1, 19}, {0, 0, 0, 0}}},{9216, {{22618, 1, 0, 0},{22618, 1, 1, 20}, {0, 0, 0, 0}}},{9216, {{22618, 1, 0, 0},{22618, 1, 1, 20}, {0, 0, 0, 0}}},{30194, {{22618, 1, 0, 0},{22618, 1, 1, 0},{22618, 1, 1, 21},{64016, 1, 0, 0}, {0, 0, 0, 0}}},{30194, {{22618, 1, 0, 0},{22618, 1, 1, 0},{22618, 1, 1, 21},{64016, 1, 0, 0}, {0, 0, 0, 0}}},{30194, {{22618, 1, 0, 0},{22618, 1, 1, 0},{22618, 1, 1, 21},{64016, 1, 0, 0}, {0, 0, 0, 0}}},{30195, {{22618, 1, 0, 0},{22618, 1, 1, 22}, {0, 0, 0, 0}}},{30195, {{22618, 1, 0, 0},{22618, 1, 1, 22}, {0, 0, 0, 0}}}}},
    {1, {{9217, {{22619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9218, {{22620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9219, {{22621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9220, {{22622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9221, {{22623, 1, 0, 0},{22623, 1, 1, 17}, {0, 0, 0, 0}}},{9221, {{22623, 1, 0, 0},{22623, 1, 1, 17}, {0, 0, 0, 0}}},{9222, {{22623, 1, 0, 0},{22623, 1, 1, 18}, {0, 0, 0, 0}}},{9222, {{22623, 1, 0, 0},{22623, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9223, {{22624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9224, {{22625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9225, {{22626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9226, {{22627, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9228, {{22628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9229, {{22629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9230, {{22630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9231, {{22631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9233, {{22632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9234, {{22633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{9235, {{22635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{9236, {{22640, 1, 0, 0},{22640, 1, 1, 17}, {0, 0, 0, 0}}},{9236, {{22640, 1, 0, 0},{22640, 1, 1, 17}, {0, 0, 0, 0}}},{9237, {{22640, 1, 0, 0},{22640, 1, 1, 18}, {0, 0, 0, 0}}},{9237, {{22640, 1, 0, 0},{22640, 1, 1, 18}, {0, 0, 0, 0}}},{57249, {{22640, 1, 0, 0},{22640, 1, 1, 19}, {0, 0, 0, 0}}},{57249, {{22640, 1, 0, 0},{22640, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{9238, {{22642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9239, {{22643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9240, {{22644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9242, {{22645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9243, {{22646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9244, {{22647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9245, {{22648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9246, {{22649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9247, {{22650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9248, {{22651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9249, {{22652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9250, {{22653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9252, {{22654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9253, {{22655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9254, {{22656, 1, 0, 0},{22656, 1, 1, 18}, {0, 0, 0, 0}}},{9254, {{22656, 1, 0, 0},{22656, 1, 1, 18}, {0, 0, 0, 0}}},{9255, {{22656, 1, 0, 0},{22656, 1, 1, 17}, {0, 0, 0, 0}}},{9255, {{22656, 1, 0, 0},{22656, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{9256, {{22657, 1, 0, 0},{22657, 1, 1, 17}, {0, 0, 0, 0}}},{9256, {{22657, 1, 0, 0},{22657, 1, 1, 17}, {0, 0, 0, 0}}},{9257, {{22657, 1, 0, 0},{22657, 1, 1, 18}, {0, 0, 0, 0}}},{9257, {{22657, 1, 0, 0},{22657, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9258, {{22658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9259, {{22659, 1, 0, 0},{22659, 1, 1, 17}, {0, 0, 0, 0}}},{9259, {{22659, 1, 0, 0},{22659, 1, 1, 17}, {0, 0, 0, 0}}},{9260, {{22659, 1, 0, 0},{22659, 1, 1, 18}, {0, 0, 0, 0}}},{9260, {{22659, 1, 0, 0},{22659, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9261, {{22660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9262, {{22661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9263, {{22662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9264, {{22663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9265, {{22664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9266, {{22665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9267, {{22666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9268, {{22667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9269, {{22668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{9270, {{22669, 1, 0, 0},{22669, 1, 1, 18}, {0, 0, 0, 0}}},{9270, {{22669, 1, 0, 0},{22669, 1, 1, 18}, {0, 0, 0, 0}}},{9271, {{22669, 1, 0, 0},{22669, 1, 1, 17}, {0, 0, 0, 0}}},{9271, {{22669, 1, 0, 0},{22669, 1, 1, 17}, {0, 0, 0, 0}}},{9272, {{22669, 1, 0, 0},{22669, 1, 1, 19}, {0, 0, 0, 0}}},{9272, {{22669, 1, 0, 0},{22669, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{9273, {{22670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9274, {{22671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9275, {{22672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9277, {{22673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9278, {{22674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9279, {{22675, 1, 0, 0},{22675, 1, 1, 17}, {0, 0, 0, 0}}},{9279, {{22675, 1, 0, 0},{22675, 1, 1, 17}, {0, 0, 0, 0}}},{9280, {{22675, 1, 0, 0},{22675, 1, 1, 18}, {0, 0, 0, 0}}},{9280, {{22675, 1, 0, 0},{22675, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9281, {{22676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9282, {{22677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9283, {{22678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9284, {{22679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{9285, {{22684, 1, 0, 0},{22684, 1, 1, 20}, {0, 0, 0, 0}}},{9285, {{22684, 1, 0, 0},{22684, 1, 1, 20}, {0, 0, 0, 0}}},{9286, {{22684, 1, 0, 0},{22684, 1, 1, 18}, {0, 0, 0, 0}}},{9286, {{22684, 1, 0, 0},{22684, 1, 1, 18}, {0, 0, 0, 0}}},{9287, {{22684, 1, 0, 0},{22684, 1, 1, 19}, {0, 0, 0, 0}}},{9287, {{22684, 1, 0, 0},{22684, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{9288, {{22685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9289, {{22686, 1, 0, 0},{22686, 1, 1, 17}, {0, 0, 0, 0}}},{9289, {{22686, 1, 0, 0},{22686, 1, 1, 17}, {0, 0, 0, 0}}},{9290, {{22686, 1, 0, 0},{22686, 1, 1, 18}, {0, 0, 0, 0}}},{9290, {{22686, 1, 0, 0},{22686, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{9291, {{22687, 1, 0, 0},{22687, 1, 1, 17}, {0, 0, 0, 0}}},{9291, {{22687, 1, 0, 0},{22687, 1, 1, 17}, {0, 0, 0, 0}}},{9292, {{22687, 1, 0, 0},{22687, 1, 1, 18}, {0, 0, 0, 0}}},{9292, {{22687, 1, 0, 0},{22687, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9293, {{22688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9294, {{22689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9295, {{22690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9296, {{22691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9297, {{22692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9298, {{22693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9299, {{22694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9300, {{22695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{9301, {{22696, 1, 0, 0},{22696, 1, 1, 18}, {0, 0, 0, 0}}},{9301, {{22696, 1, 0, 0},{22696, 1, 1, 18}, {0, 0, 0, 0}}},{30254, {{22696, 1, 0, 0},{22696, 1, 1, 0},{22696, 1, 1, 19},{64058, 1, 0, 0}, {0, 0, 0, 0}}},{30254, {{22696, 1, 0, 0},{22696, 1, 1, 0},{22696, 1, 1, 19},{64058, 1, 0, 0}, {0, 0, 0, 0}}},{30254, {{22696, 1, 0, 0},{22696, 1, 1, 0},{22696, 1, 1, 19},{64058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9302, {{22697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9303, {{22698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{9304, {{22699, 1, 0, 0},{22699, 1, 1, 18}, {0, 0, 0, 0}}},{9304, {{22699, 1, 0, 0},{22699, 1, 1, 18}, {0, 0, 0, 0}}},{9305, {{22699, 1, 0, 0},{22699, 1, 1, 19}, {0, 0, 0, 0}}},{9305, {{22699, 1, 0, 0},{22699, 1, 1, 19}, {0, 0, 0, 0}}},{9306, {{22699, 1, 0, 0},{22699, 1, 1, 20}, {0, 0, 0, 0}}},{9306, {{22699, 1, 0, 0},{22699, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{9307, {{22700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9308, {{22701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9309, {{22702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9311, {{22703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9312, {{22704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9313, {{22705, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9314, {{22706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9315, {{22707, 1, 0, 0},{22707, 1, 1, 17}, {0, 0, 0, 0}}},{9315, {{22707, 1, 0, 0},{22707, 1, 1, 17}, {0, 0, 0, 0}}},{57244, {{22707, 1, 0, 0},{22707, 1, 1, 18}, {0, 0, 0, 0}}},{57244, {{22707, 1, 0, 0},{22707, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9316, {{22708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9317, {{22709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{9318, {{22711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{9319, {{22712, 1, 0, 0},{22712, 1, 1, 17}, {0, 0, 0, 0}}},{9319, {{22712, 1, 0, 0},{22712, 1, 1, 17}, {0, 0, 0, 0}}},{9320, {{22712, 1, 0, 0},{22712, 1, 1, 18}, {0, 0, 0, 0}}},{9320, {{22712, 1, 0, 0},{22712, 1, 1, 18}, {0, 0, 0, 0}}},{57245, {{22712, 1, 0, 0},{22712, 1, 1, 19}, {0, 0, 0, 0}}},{57245, {{22712, 1, 0, 0},{22712, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{9321, {{22713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9322, {{22714, 1, 0, 0},{22714, 1, 1, 17}, {0, 0, 0, 0}}},{9322, {{22714, 1, 0, 0},{22714, 1, 1, 17}, {0, 0, 0, 0}}},{9323, {{22714, 1, 0, 0},{22714, 1, 1, 18}, {0, 0, 0, 0}}},{9323, {{22714, 1, 0, 0},{22714, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9324, {{22715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9325, {{22716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9326, {{22717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9327, {{22718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9328, {{22719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9329, {{22720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9330, {{22721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9331, {{22722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9332, {{22723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9333, {{22724, 1, 0, 0},{22724, 1, 1, 17}, {0, 0, 0, 0}}},{9333, {{22724, 1, 0, 0},{22724, 1, 1, 17}, {0, 0, 0, 0}}},{9334, {{22724, 1, 0, 0},{22724, 1, 1, 18}, {0, 0, 0, 0}}},{9334, {{22724, 1, 0, 0},{22724, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9335, {{22725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9336, {{22726, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9337, {{22727, 1, 0, 0},{22727, 1, 1, 17}, {0, 0, 0, 0}}},{9337, {{22727, 1, 0, 0},{22727, 1, 1, 17}, {0, 0, 0, 0}}},{9338, {{22727, 1, 0, 0},{22727, 1, 1, 18}, {0, 0, 0, 0}}},{9338, {{22727, 1, 0, 0},{22727, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9339, {{22728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9340, {{22729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9341, {{22730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{9342, {{22732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9343, {{22733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9344, {{22734, 1, 0, 0},{22734, 1, 1, 17}, {0, 0, 0, 0}}},{9344, {{22734, 1, 0, 0},{22734, 1, 1, 17}, {0, 0, 0, 0}}},{9345, {{22734, 1, 0, 0},{22734, 1, 1, 18}, {0, 0, 0, 0}}},{9345, {{22734, 1, 0, 0},{22734, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9346, {{22735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9347, {{22736, 1, 0, 0},{22736, 1, 1, 17}, {0, 0, 0, 0}}},{9347, {{22736, 1, 0, 0},{22736, 1, 1, 17}, {0, 0, 0, 0}}},{9348, {{22736, 1, 0, 0},{22736, 1, 1, 18}, {0, 0, 0, 0}}},{9348, {{22736, 1, 0, 0},{22736, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9349, {{22737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9350, {{22738, 1, 0, 0},{22738, 1, 1, 17}, {0, 0, 0, 0}}},{9350, {{22738, 1, 0, 0},{22738, 1, 1, 17}, {0, 0, 0, 0}}},{9351, {{22738, 1, 0, 0},{22738, 1, 1, 18}, {0, 0, 0, 0}}},{9351, {{22738, 1, 0, 0},{22738, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{9352, {{22739, 1, 0, 0},{22739, 1, 1, 17}, {0, 0, 0, 0}}},{9352, {{22739, 1, 0, 0},{22739, 1, 1, 17}, {0, 0, 0, 0}}},{9353, {{22739, 1, 0, 0},{22739, 1, 1, 18}, {0, 0, 0, 0}}},{9353, {{22739, 1, 0, 0},{22739, 1, 1, 18}, {0, 0, 0, 0}}},{59460, {{22739, 1, 0, 0},{22739, 1, 1, 19}, {0, 0, 0, 0}}},{59460, {{22739, 1, 0, 0},{22739, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{9354, {{22740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9355, {{22741, 1, 0, 0},{22741, 1, 1, 17}, {0, 0, 0, 0}}},{9355, {{22741, 1, 0, 0},{22741, 1, 1, 17}, {0, 0, 0, 0}}},{9356, {{22741, 1, 0, 0},{22741, 1, 1, 18}, {0, 0, 0, 0}}},{9356, {{22741, 1, 0, 0},{22741, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9357, {{22742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9358, {{22743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9359, {{22744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9360, {{22745, 1, 0, 0},{22745, 1, 1, 17}, {0, 0, 0, 0}}},{9360, {{22745, 1, 0, 0},{22745, 1, 1, 17}, {0, 0, 0, 0}}},{9361, {{22745, 1, 0, 0},{22745, 1, 1, 18}, {0, 0, 0, 0}}},{9361, {{22745, 1, 0, 0},{22745, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9362, {{22746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9363, {{22747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9364, {{22748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9365, {{22749, 1, 0, 0},{22749, 1, 1, 18}, {0, 0, 0, 0}}},{9365, {{22749, 1, 0, 0},{22749, 1, 1, 18}, {0, 0, 0, 0}}},{9366, {{22749, 1, 0, 0},{22749, 1, 1, 17}, {0, 0, 0, 0}}},{9366, {{22749, 1, 0, 0},{22749, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{9367, {{22750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9368, {{22751, 1, 0, 0},{22751, 1, 1, 17}, {0, 0, 0, 0}}},{9368, {{22751, 1, 0, 0},{22751, 1, 1, 17}, {0, 0, 0, 0}}},{9369, {{22751, 1, 0, 0},{22751, 1, 1, 18}, {0, 0, 0, 0}}},{9369, {{22751, 1, 0, 0},{22751, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{9370, {{22752, 1, 0, 0},{22752, 1, 1, 17}, {0, 0, 0, 0}}},{9370, {{22752, 1, 0, 0},{22752, 1, 1, 17}, {0, 0, 0, 0}}},{9371, {{22752, 1, 0, 0},{22752, 1, 1, 18}, {0, 0, 0, 0}}},{9371, {{22752, 1, 0, 0},{22752, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{9372, {{22753, 1, 0, 0},{22753, 1, 1, 17}, {0, 0, 0, 0}}},{9372, {{22753, 1, 0, 0},{22753, 1, 1, 17}, {0, 0, 0, 0}}},{9373, {{22753, 1, 0, 0},{22753, 1, 1, 18}, {0, 0, 0, 0}}},{9373, {{22753, 1, 0, 0},{22753, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9374, {{22754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9375, {{22755, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9376, {{22756, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9377, {{22757, 1, 0, 0},{22757, 1, 1, 17}, {0, 0, 0, 0}}},{9377, {{22757, 1, 0, 0},{22757, 1, 1, 17}, {0, 0, 0, 0}}},{9378, {{22757, 1, 0, 0},{22757, 1, 1, 18}, {0, 0, 0, 0}}},{9378, {{22757, 1, 0, 0},{22757, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9379, {{22758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9380, {{22759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9381, {{22760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9382, {{22761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{9383, {{22763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9384, {{22764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{9385, {{22766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9386, {{22767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9387, {{22768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9388, {{22769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9389, {{22770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9390, {{22771, 1, 0, 0},{22771, 1, 1, 18}, {0, 0, 0, 0}}},{9390, {{22771, 1, 0, 0},{22771, 1, 1, 18}, {0, 0, 0, 0}}},{9391, {{22771, 1, 0, 0},{22771, 1, 1, 17}, {0, 0, 0, 0}}},{9391, {{22771, 1, 0, 0},{22771, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{9392, {{22772, 1, 0, 0},{22772, 1, 1, 18}, {0, 0, 0, 0}}},{9392, {{22772, 1, 0, 0},{22772, 1, 1, 18}, {0, 0, 0, 0}}},{9393, {{22772, 1, 0, 0},{22772, 1, 1, 17}, {0, 0, 0, 0}}},{9393, {{22772, 1, 0, 0},{22772, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{9394, {{22773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{9395, {{22775, 1, 0, 0},{22775, 1, 1, 17}, {0, 0, 0, 0}}},{9395, {{22775, 1, 0, 0},{22775, 1, 1, 17}, {0, 0, 0, 0}}},{59463, {{22775, 1, 0, 0},{22775, 1, 1, 18}, {0, 0, 0, 0}}},{59463, {{22775, 1, 0, 0},{22775, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{9397, {{22777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9398, {{22778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9399, {{22779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9400, {{22780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9401, {{22781, 1, 0, 0},{22781, 1, 1, 17}, {0, 0, 0, 0}}},{9401, {{22781, 1, 0, 0},{22781, 1, 1, 17}, {0, 0, 0, 0}}},{57250, {{22781, 1, 0, 0},{22781, 1, 1, 18}, {0, 0, 0, 0}}},{57250, {{22781, 1, 0, 0},{22781, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9402, {{22782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9403, {{22783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {10, {{9404, {{22784, 1, 0, 0},{22784, 1, 1, 18}, {0, 0, 0, 0}}},{9404, {{22784, 1, 0, 0},{22784, 1, 1, 18}, {0, 0, 0, 0}}},{57355, {{22784, 1, 0, 0},{22784, 1, 1, 16}, {0, 0, 0, 0}}},{57355, {{22784, 1, 0, 0},{22784, 1, 1, 16}, {0, 0, 0, 0}}},{57356, {{22784, 1, 0, 0},{22784, 1, 1, 17}, {0, 0, 0, 0}}},{57356, {{22784, 1, 0, 0},{22784, 1, 1, 17}, {0, 0, 0, 0}}},{57357, {{22784, 1, 0, 0},{22784, 1, 1, 20}, {0, 0, 0, 0}}},{57357, {{22784, 1, 0, 0},{22784, 1, 1, 20}, {0, 0, 0, 0}}},{57358, {{22784, 1, 0, 0},{22784, 1, 1, 19}, {0, 0, 0, 0}}},{57358, {{22784, 1, 0, 0},{22784, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{9405, {{22785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9406, {{22786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9407, {{22787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{9408, {{22789, 1, 0, 0},{22789, 1, 1, 18}, {0, 0, 0, 0}}},{9408, {{22789, 1, 0, 0},{22789, 1, 1, 18}, {0, 0, 0, 0}}},{9409, {{22789, 1, 0, 0},{22789, 1, 1, 17}, {0, 0, 0, 0}}},{9409, {{22789, 1, 0, 0},{22789, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{9410, {{22790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58084, {{22791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9411, {{22792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9412, {{22793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9413, {{22794, 1, 0, 0},{22794, 1, 1, 17}, {0, 0, 0, 0}}},{9413, {{22794, 1, 0, 0},{22794, 1, 1, 17}, {0, 0, 0, 0}}},{9414, {{22794, 1, 0, 0},{22794, 1, 1, 18}, {0, 0, 0, 0}}},{9414, {{22794, 1, 0, 0},{22794, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9415, {{22795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9416, {{22796, 1, 0, 0},{22796, 1, 1, 17}, {0, 0, 0, 0}}},{9416, {{22796, 1, 0, 0},{22796, 1, 1, 17}, {0, 0, 0, 0}}},{57253, {{22796, 1, 0, 0},{22796, 1, 1, 18}, {0, 0, 0, 0}}},{57253, {{22796, 1, 0, 0},{22796, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9417, {{22797, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9418, {{22798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{9419, {{22799, 1, 0, 0},{22799, 1, 1, 18}, {0, 0, 0, 0}}},{9419, {{22799, 1, 0, 0},{22799, 1, 1, 18}, {0, 0, 0, 0}}},{9420, {{22799, 1, 0, 0},{22799, 1, 1, 17}, {0, 0, 0, 0}}},{9420, {{22799, 1, 0, 0},{22799, 1, 1, 17}, {0, 0, 0, 0}}},{57256, {{22799, 1, 0, 0},{22799, 1, 1, 19}, {0, 0, 0, 0}}},{57256, {{22799, 1, 0, 0},{22799, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{9421, {{22800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{9423, {{22802, 1, 0, 0},{22802, 1, 1, 18}, {0, 0, 0, 0}}},{9423, {{22802, 1, 0, 0},{22802, 1, 1, 18}, {0, 0, 0, 0}}},{9424, {{22802, 1, 0, 0},{22802, 1, 1, 17}, {0, 0, 0, 0}}},{9424, {{22802, 1, 0, 0},{22802, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{9425, {{22803, 1, 0, 0},{22803, 1, 1, 18}, {0, 0, 0, 0}}},{9425, {{22803, 1, 0, 0},{22803, 1, 1, 18}, {0, 0, 0, 0}}},{9426, {{22803, 1, 0, 0},{22803, 1, 1, 17}, {0, 0, 0, 0}}},{9426, {{22803, 1, 0, 0},{22803, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {6, {{9427, {{22804, 1, 0, 0},{22804, 1, 1, 18}, {0, 0, 0, 0}}},{9427, {{22804, 1, 0, 0},{22804, 1, 1, 18}, {0, 0, 0, 0}}},{9428, {{22804, 1, 0, 0},{22804, 1, 1, 19}, {0, 0, 0, 0}}},{9428, {{22804, 1, 0, 0},{22804, 1, 1, 19}, {0, 0, 0, 0}}},{9429, {{22804, 1, 0, 0},{22804, 1, 1, 17}, {0, 0, 0, 0}}},{9429, {{22804, 1, 0, 0},{22804, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{9430, {{22805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9431, {{22806, 1, 0, 0},{22806, 1, 1, 17}, {0, 0, 0, 0}}},{9431, {{22806, 1, 0, 0},{22806, 1, 1, 17}, {0, 0, 0, 0}}},{9432, {{22806, 1, 0, 0},{22806, 1, 1, 18}, {0, 0, 0, 0}}},{9432, {{22806, 1, 0, 0},{22806, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{9433, {{22807, 1, 0, 0},{22807, 1, 1, 16}, {0, 0, 0, 0}}},{9433, {{22807, 1, 0, 0},{22807, 1, 1, 16}, {0, 0, 0, 0}}},{9434, {{22807, 1, 0, 0},{22807, 1, 1, 17}, {0, 0, 0, 0}}},{9434, {{22807, 1, 0, 0},{22807, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{9435, {{22808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9436, {{22809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9437, {{22810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9438, {{22811, 1, 0, 0},{22811, 1, 1, 18}, {0, 0, 0, 0}}},{9438, {{22811, 1, 0, 0},{22811, 1, 1, 18}, {0, 0, 0, 0}}},{9439, {{22811, 1, 0, 0},{22811, 1, 1, 19}, {0, 0, 0, 0}}},{9439, {{22811, 1, 0, 0},{22811, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{9441, {{22812, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9442, {{22813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{9443, {{22815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9444, {{22816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9445, {{22817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{9446, {{22818, 1, 0, 0},{22818, 1, 1, 17}, {0, 0, 0, 0}}},{9446, {{22818, 1, 0, 0},{22818, 1, 1, 17}, {0, 0, 0, 0}}},{9447, {{22818, 1, 0, 0},{22818, 1, 1, 18}, {0, 0, 0, 0}}},{9447, {{22818, 1, 0, 0},{22818, 1, 1, 18}, {0, 0, 0, 0}}},{57269, {{22818, 1, 0, 0},{22818, 1, 1, 19}, {0, 0, 0, 0}}},{57269, {{22818, 1, 0, 0},{22818, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{9448, {{22819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9449, {{22820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9450, {{22821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9451, {{22822, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9452, {{22823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9453, {{22824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9454, {{22825, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9455, {{22826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9456, {{22827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9457, {{22828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9458, {{22829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9459, {{22830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9460, {{22831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9461, {{22832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9462, {{22833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9463, {{22834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9464, {{22835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{9465, {{22837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9466, {{22838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9467, {{22839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9468, {{22840, 1, 0, 0},{22840, 1, 1, 17}, {0, 0, 0, 0}}},{9468, {{22840, 1, 0, 0},{22840, 1, 1, 17}, {0, 0, 0, 0}}},{9469, {{22840, 1, 0, 0},{22840, 1, 1, 18}, {0, 0, 0, 0}}},{9469, {{22840, 1, 0, 0},{22840, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9470, {{22841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{9471, {{22845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9472, {{22846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9473, {{22847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9474, {{22848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{9475, {{22851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9476, {{22852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9478, {{22853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9479, {{22854, 1, 0, 0},{22854, 1, 1, 17}, {0, 0, 0, 0}}},{9479, {{22854, 1, 0, 0},{22854, 1, 1, 17}, {0, 0, 0, 0}}},{9480, {{22854, 1, 0, 0},{22854, 1, 1, 18}, {0, 0, 0, 0}}},{9480, {{22854, 1, 0, 0},{22854, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9481, {{22855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9482, {{22856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9483, {{22857, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9484, {{22858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9485, {{22859, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9486, {{22860, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{9487, {{22862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9488, {{22863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9489, {{22864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{9490, {{22865, 1, 0, 0},{22865, 1, 1, 20}, {0, 0, 0, 0}}},{9490, {{22865, 1, 0, 0},{22865, 1, 1, 20}, {0, 0, 0, 0}}},{9491, {{22865, 1, 0, 0},{22865, 1, 1, 19}, {0, 0, 0, 0}}},{9491, {{22865, 1, 0, 0},{22865, 1, 1, 19}, {0, 0, 0, 0}}},{9492, {{22865, 1, 0, 0},{22865, 1, 1, 22}, {0, 0, 0, 0}}},{9492, {{22865, 1, 0, 0},{22865, 1, 1, 22}, {0, 0, 0, 0}}},{9493, {{22865, 1, 0, 0},{22865, 1, 1, 21}, {0, 0, 0, 0}}},{9493, {{22865, 1, 0, 0},{22865, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{9494, {{22866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9495, {{22867, 1, 0, 0},{22867, 1, 1, 19}, {0, 0, 0, 0}}},{9495, {{22867, 1, 0, 0},{22867, 1, 1, 19}, {0, 0, 0, 0}}},{9496, {{22867, 1, 0, 0},{22867, 1, 1, 18}, {0, 0, 0, 0}}},{9496, {{22867, 1, 0, 0},{22867, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{9497, {{22868, 1, 0, 0},{22868, 1, 1, 17}, {0, 0, 0, 0}}},{9497, {{22868, 1, 0, 0},{22868, 1, 1, 17}, {0, 0, 0, 0}}},{9498, {{22868, 1, 0, 0},{22868, 1, 1, 18}, {0, 0, 0, 0}}},{9498, {{22868, 1, 0, 0},{22868, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9499, {{22869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{9501, {{22871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9502, {{22872, 1, 0, 0},{22872, 1, 1, 17}, {0, 0, 0, 0}}},{9502, {{22872, 1, 0, 0},{22872, 1, 1, 17}, {0, 0, 0, 0}}},{9503, {{22872, 1, 0, 0},{22872, 1, 1, 18}, {0, 0, 0, 0}}},{9503, {{22872, 1, 0, 0},{22872, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9505, {{22873, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9506, {{22874, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9507, {{22875, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9508, {{22876, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9509, {{22877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9511, {{22878, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9512, {{22879, 1, 0, 0},{22879, 1, 1, 17}, {0, 0, 0, 0}}},{9512, {{22879, 1, 0, 0},{22879, 1, 1, 17}, {0, 0, 0, 0}}},{9513, {{22879, 1, 0, 0},{22879, 1, 1, 18}, {0, 0, 0, 0}}},{9513, {{22879, 1, 0, 0},{22879, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {8, {{9514, {{22880, 1, 0, 0},{22880, 1, 1, 18}, {0, 0, 0, 0}}},{9514, {{22880, 1, 0, 0},{22880, 1, 1, 18}, {0, 0, 0, 0}}},{9515, {{22880, 1, 0, 0},{22880, 1, 1, 21}, {0, 0, 0, 0}}},{9515, {{22880, 1, 0, 0},{22880, 1, 1, 21}, {0, 0, 0, 0}}},{9516, {{22880, 1, 0, 0},{22880, 1, 1, 19}, {0, 0, 0, 0}}},{9516, {{22880, 1, 0, 0},{22880, 1, 1, 19}, {0, 0, 0, 0}}},{9517, {{22880, 1, 0, 0},{22880, 1, 1, 20}, {0, 0, 0, 0}}},{9517, {{22880, 1, 0, 0},{22880, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{9518, {{22881, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9519, {{22882, 1, 0, 0},{22882, 1, 1, 18}, {0, 0, 0, 0}}},{9519, {{22882, 1, 0, 0},{22882, 1, 1, 18}, {0, 0, 0, 0}}},{9520, {{22882, 1, 0, 0},{22882, 1, 1, 19}, {0, 0, 0, 0}}},{9520, {{22882, 1, 0, 0},{22882, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{9521, {{22883, 1, 0, 0},{22883, 1, 1, 17}, {0, 0, 0, 0}}},{9521, {{22883, 1, 0, 0},{22883, 1, 1, 17}, {0, 0, 0, 0}}},{9522, {{22883, 1, 0, 0},{22883, 1, 1, 18}, {0, 0, 0, 0}}},{9522, {{22883, 1, 0, 0},{22883, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9523, {{22884, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9524, {{22885, 1, 0, 0},{22885, 1, 1, 17}, {0, 0, 0, 0}}},{9524, {{22885, 1, 0, 0},{22885, 1, 1, 17}, {0, 0, 0, 0}}},{9525, {{22885, 1, 0, 0},{22885, 1, 1, 18}, {0, 0, 0, 0}}},{9525, {{22885, 1, 0, 0},{22885, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9526, {{22886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9527, {{22887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9528, {{22888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9530, {{22889, 1, 0, 0},{22889, 1, 1, 17}, {0, 0, 0, 0}}},{9530, {{22889, 1, 0, 0},{22889, 1, 1, 17}, {0, 0, 0, 0}}},{9531, {{22889, 1, 0, 0},{22889, 1, 1, 18}, {0, 0, 0, 0}}},{9531, {{22889, 1, 0, 0},{22889, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9532, {{22890, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9533, {{22891, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9534, {{22892, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9536, {{22893, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9537, {{22894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9538, {{22895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9539, {{22896, 1, 0, 0},{22896, 1, 1, 16}, {0, 0, 0, 0}}},{9539, {{22896, 1, 0, 0},{22896, 1, 1, 16}, {0, 0, 0, 0}}},{59485, {{22896, 1, 0, 0},{22896, 1, 1, 17}, {0, 0, 0, 0}}},{59485, {{22896, 1, 0, 0},{22896, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{9540, {{22897, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9541, {{22898, 1, 0, 0},{22898, 1, 1, 17}, {0, 0, 0, 0}}},{9541, {{22898, 1, 0, 0},{22898, 1, 1, 17}, {0, 0, 0, 0}}},{9542, {{22898, 1, 0, 0},{22898, 1, 1, 18}, {0, 0, 0, 0}}},{9542, {{22898, 1, 0, 0},{22898, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9543, {{22899, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9544, {{22900, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9545, {{22901, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9546, {{22902, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9547, {{22903, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9548, {{22904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9549, {{22905, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9550, {{22906, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9551, {{22907, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9552, {{22908, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9553, {{22909, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9554, {{22910, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9555, {{22911, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9556, {{22912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9557, {{22913, 1, 0, 0},{22913, 1, 1, 18}, {0, 0, 0, 0}}},{9557, {{22913, 1, 0, 0},{22913, 1, 1, 18}, {0, 0, 0, 0}}},{9558, {{22913, 1, 0, 0},{22913, 1, 1, 17}, {0, 0, 0, 0}}},{9558, {{22913, 1, 0, 0},{22913, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{9559, {{22914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9560, {{22915, 1, 0, 0},{22915, 1, 1, 17}, {0, 0, 0, 0}}},{9560, {{22915, 1, 0, 0},{22915, 1, 1, 17}, {0, 0, 0, 0}}},{9561, {{22915, 1, 0, 0},{22915, 1, 1, 18}, {0, 0, 0, 0}}},{9561, {{22915, 1, 0, 0},{22915, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{9563, {{22916, 1, 0, 0},{22916, 1, 1, 19}, {0, 0, 0, 0}}},{9563, {{22916, 1, 0, 0},{22916, 1, 1, 19}, {0, 0, 0, 0}}},{9564, {{22916, 1, 0, 0},{22916, 1, 1, 18}, {0, 0, 0, 0}}},{9564, {{22916, 1, 0, 0},{22916, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9565, {{22917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9566, {{22918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57291, {{22919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{9567, {{22921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9568, {{22922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9569, {{22923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9570, {{22924, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9571, {{22925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9573, {{22926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9574, {{22927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9575, {{22928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9576, {{22929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9577, {{22930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9578, {{22931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9579, {{22932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9580, {{22933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9581, {{22934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9582, {{22935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9583, {{22936, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9584, {{22937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9585, {{22938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9586, {{22939, 1, 0, 0},{22939, 1, 1, 18}, {0, 0, 0, 0}}},{9586, {{22939, 1, 0, 0},{22939, 1, 1, 18}, {0, 0, 0, 0}}},{9587, {{22939, 1, 0, 0},{22939, 1, 1, 17}, {0, 0, 0, 0}}},{9587, {{22939, 1, 0, 0},{22939, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{9588, {{22940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9589, {{22941, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9590, {{22942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9591, {{22943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9592, {{22944, 1, 0, 0},{22944, 1, 1, 16}, {0, 0, 0, 0}}},{9592, {{22944, 1, 0, 0},{22944, 1, 1, 16}, {0, 0, 0, 0}}},{9593, {{22944, 1, 0, 0},{22944, 1, 1, 17}, {0, 0, 0, 0}}},{9593, {{22944, 1, 0, 0},{22944, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{9594, {{22945, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9595, {{22946, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9596, {{22947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9597, {{22948, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9598, {{22949, 1, 0, 0},{22949, 1, 1, 19}, {0, 0, 0, 0}}},{9598, {{22949, 1, 0, 0},{22949, 1, 1, 19}, {0, 0, 0, 0}}},{9599, {{22949, 1, 0, 0},{22949, 1, 1, 18}, {0, 0, 0, 0}}},{9599, {{22949, 1, 0, 0},{22949, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9600, {{22950, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9601, {{22951, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9602, {{22952, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{9603, {{22956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9604, {{22957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9605, {{22958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9606, {{22959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9607, {{22960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9608, {{22961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9609, {{22962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9610, {{22963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9611, {{22964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9612, {{22965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9613, {{22966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9614, {{22967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9615, {{22968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9616, {{22969, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9617, {{22970, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9618, {{22971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9619, {{22972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9620, {{22973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9621, {{22974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9622, {{22975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9623, {{22976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9624, {{22977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9625, {{22978, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9626, {{22979, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9627, {{22980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9628, {{22981, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9629, {{22982, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9630, {{22983, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9631, {{22984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9632, {{22985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9633, {{22986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9634, {{22987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9635, {{22988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9636, {{22989, 1, 0, 0},{22989, 1, 1, 17}, {0, 0, 0, 0}}},{9636, {{22989, 1, 0, 0},{22989, 1, 1, 17}, {0, 0, 0, 0}}},{9637, {{22989, 1, 0, 0},{22989, 1, 1, 18}, {0, 0, 0, 0}}},{9637, {{22989, 1, 0, 0},{22989, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9638, {{22990, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9639, {{22991, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9640, {{22992, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9641, {{22993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9642, {{22994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9643, {{22995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9644, {{22996, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9645, {{22997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9646, {{22998, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{9647, {{23000, 1, 0, 0},{23000, 1, 1, 16}, {0, 0, 0, 0}}},{9647, {{23000, 1, 0, 0},{23000, 1, 1, 16}, {0, 0, 0, 0}}},{9648, {{23000, 1, 0, 0},{23000, 1, 1, 17}, {0, 0, 0, 0}}},{9648, {{23000, 1, 0, 0},{23000, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{9649, {{23001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9650, {{23002, 1, 0, 0},{23002, 1, 1, 18}, {0, 0, 0, 0}}},{9650, {{23002, 1, 0, 0},{23002, 1, 1, 18}, {0, 0, 0, 0}}},{9651, {{23002, 1, 0, 0},{23002, 1, 1, 19}, {0, 0, 0, 0}}},{9651, {{23002, 1, 0, 0},{23002, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{9652, {{23003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9653, {{23004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9654, {{23005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9655, {{23006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9656, {{23007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9657, {{23008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9658, {{23009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9659, {{23010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9660, {{23011, 1, 0, 0},{23011, 1, 1, 17}, {0, 0, 0, 0}}},{9660, {{23011, 1, 0, 0},{23011, 1, 1, 17}, {0, 0, 0, 0}}},{9661, {{23011, 1, 0, 0},{23011, 1, 1, 18}, {0, 0, 0, 0}}},{9661, {{23011, 1, 0, 0},{23011, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9662, {{23012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9663, {{23013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9664, {{23014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9665, {{23015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9666, {{23016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9667, {{23017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9668, {{23018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9669, {{23019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9670, {{23020, 1, 0, 0},{23020, 1, 1, 18}, {0, 0, 0, 0}}},{9670, {{23020, 1, 0, 0},{23020, 1, 1, 18}, {0, 0, 0, 0}}},{9671, {{23020, 1, 0, 0},{23020, 1, 1, 19}, {0, 0, 0, 0}}},{9671, {{23020, 1, 0, 0},{23020, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{9672, {{23021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9673, {{23022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9674, {{23023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9675, {{23024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9676, {{23025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9677, {{23026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9678, {{23027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9679, {{23028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9680, {{23029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9681, {{23030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9682, {{23031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9683, {{23032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9684, {{23033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9685, {{23034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9686, {{23035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9687, {{23036, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9688, {{23037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9689, {{23038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9690, {{23039, 1, 0, 0},{23039, 1, 1, 20}, {0, 0, 0, 0}}},{9690, {{23039, 1, 0, 0},{23039, 1, 1, 20}, {0, 0, 0, 0}}},{9691, {{23039, 1, 0, 0},{23039, 1, 1, 19}, {0, 0, 0, 0}}},{9691, {{23039, 1, 0, 0},{23039, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{9692, {{23040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9693, {{23041, 1, 0, 0},{23041, 1, 1, 17}, {0, 0, 0, 0}}},{9693, {{23041, 1, 0, 0},{23041, 1, 1, 17}, {0, 0, 0, 0}}},{9694, {{23041, 1, 0, 0},{23041, 1, 1, 18}, {0, 0, 0, 0}}},{9694, {{23041, 1, 0, 0},{23041, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9695, {{23042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9696, {{23043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9697, {{23044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{9698, {{23049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9699, {{23050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9700, {{23051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9701, {{23052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9702, {{23053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9703, {{23054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9704, {{23055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9705, {{23056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9706, {{23057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9707, {{23058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9708, {{23059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9709, {{23060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9710, {{23061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9711, {{23062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9712, {{23063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9713, {{23064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9714, {{23065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9716, {{23066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9717, {{23067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9719, {{23068, 1, 0, 0},{23068, 1, 1, 18}, {0, 0, 0, 0}}},{9719, {{23068, 1, 0, 0},{23068, 1, 1, 18}, {0, 0, 0, 0}}},{9720, {{23068, 1, 0, 0},{23068, 1, 1, 19}, {0, 0, 0, 0}}},{9720, {{23068, 1, 0, 0},{23068, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{9721, {{23069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9722, {{23070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9723, {{23071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9724, {{23072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9725, {{23073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9726, {{23074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9727, {{23075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9728, {{23076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9729, {{23077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9730, {{23078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9731, {{23079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9733, {{23080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9734, {{23081, 1, 0, 0},{23081, 1, 1, 19}, {0, 0, 0, 0}}},{9734, {{23081, 1, 0, 0},{23081, 1, 1, 19}, {0, 0, 0, 0}}},{9735, {{23081, 1, 0, 0},{23081, 1, 1, 18}, {0, 0, 0, 0}}},{9735, {{23081, 1, 0, 0},{23081, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9736, {{23082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9737, {{23083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9738, {{23084, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9739, {{23085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9740, {{23086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9741, {{23087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9742, {{23088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9718, {{23089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{9743, {{23093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9744, {{23094, 1, 0, 0},{23094, 1, 1, 18}, {0, 0, 0, 0}}},{9744, {{23094, 1, 0, 0},{23094, 1, 1, 18}, {0, 0, 0, 0}}},{9745, {{23094, 1, 0, 0},{23094, 1, 1, 19}, {0, 0, 0, 0}}},{9745, {{23094, 1, 0, 0},{23094, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{9746, {{23095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9747, {{23096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9748, {{23097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9749, {{23098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9750, {{23099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9751, {{23100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9752, {{23101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9753, {{23102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9754, {{23103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9755, {{23104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9756, {{23105, 1, 0, 0},{23105, 1, 1, 17}, {0, 0, 0, 0}}},{9756, {{23105, 1, 0, 0},{23105, 1, 1, 17}, {0, 0, 0, 0}}},{57300, {{23105, 1, 0, 0},{23105, 1, 1, 18}, {0, 0, 0, 0}}},{57300, {{23105, 1, 0, 0},{23105, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9757, {{23106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9758, {{23107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9759, {{23108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9760, {{23109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9761, {{23110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9762, {{23111, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9763, {{23112, 1, 0, 0},{23112, 1, 1, 18}, {0, 0, 0, 0}}},{9763, {{23112, 1, 0, 0},{23112, 1, 1, 18}, {0, 0, 0, 0}}},{9764, {{23112, 1, 0, 0},{23112, 1, 1, 17}, {0, 0, 0, 0}}},{9764, {{23112, 1, 0, 0},{23112, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{9765, {{23113, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9766, {{23114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9767, {{23115, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9768, {{23116, 1, 0, 0},{23116, 1, 1, 17}, {0, 0, 0, 0}}},{9768, {{23116, 1, 0, 0},{23116, 1, 1, 17}, {0, 0, 0, 0}}},{9769, {{23116, 1, 0, 0},{23116, 1, 1, 18}, {0, 0, 0, 0}}},{9769, {{23116, 1, 0, 0},{23116, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9770, {{23117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9771, {{23118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9772, {{23119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9773, {{23120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9774, {{23121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9775, {{23122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9776, {{23123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9777, {{23124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9778, {{23125, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9779, {{23126, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9780, {{23127, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9781, {{23128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9782, {{23129, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9783, {{23130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9784, {{23131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9785, {{23132, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9786, {{23133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9787, {{23134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9788, {{23135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9789, {{23136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9790, {{23137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9791, {{23138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9792, {{23139, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9793, {{23140, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9795, {{23141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9796, {{23142, 1, 0, 0},{23142, 1, 1, 19}, {0, 0, 0, 0}}},{9796, {{23142, 1, 0, 0},{23142, 1, 1, 19}, {0, 0, 0, 0}}},{9797, {{23142, 1, 0, 0},{23142, 1, 1, 18}, {0, 0, 0, 0}}},{9797, {{23142, 1, 0, 0},{23142, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{9798, {{23143, 1, 0, 0},{23143, 1, 1, 17}, {0, 0, 0, 0}}},{9798, {{23143, 1, 0, 0},{23143, 1, 1, 17}, {0, 0, 0, 0}}},{9799, {{23143, 1, 0, 0},{23143, 1, 1, 18}, {0, 0, 0, 0}}},{9799, {{23143, 1, 0, 0},{23143, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9800, {{23144, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9801, {{23145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9802, {{23146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9803, {{23147, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9804, {{23148, 1, 0, 0},{23148, 1, 1, 17}, {0, 0, 0, 0}}},{9804, {{23148, 1, 0, 0},{23148, 1, 1, 17}, {0, 0, 0, 0}}},{9805, {{23148, 1, 0, 0},{23148, 1, 1, 18}, {0, 0, 0, 0}}},{9805, {{23148, 1, 0, 0},{23148, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9806, {{23149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9807, {{23150, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9808, {{23151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9809, {{23152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9810, {{23153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9811, {{23154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{9813, {{23159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9814, {{23160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9815, {{23161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9816, {{23162, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9817, {{23163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9818, {{23164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9819, {{23165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9820, {{23166, 1, 0, 0},{23166, 1, 1, 17}, {0, 0, 0, 0}}},{9820, {{23166, 1, 0, 0},{23166, 1, 1, 17}, {0, 0, 0, 0}}},{9821, {{23166, 1, 0, 0},{23166, 1, 1, 18}, {0, 0, 0, 0}}},{9821, {{23166, 1, 0, 0},{23166, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9822, {{23167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9823, {{23168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9824, {{23169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9825, {{23170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9826, {{23171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9827, {{23172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9828, {{23173, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9829, {{23174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9830, {{23175, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9831, {{23176, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9832, {{23177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9833, {{23178, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9834, {{23179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9835, {{23180, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9836, {{23181, 1, 0, 0},{23181, 1, 1, 16}, {0, 0, 0, 0}}},{9836, {{23181, 1, 0, 0},{23181, 1, 1, 16}, {0, 0, 0, 0}}},{57309, {{23181, 1, 0, 0},{23181, 1, 1, 17}, {0, 0, 0, 0}}},{57309, {{23181, 1, 0, 0},{23181, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{9837, {{23182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9838, {{23183, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9839, {{23184, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9840, {{23185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9841, {{23186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9842, {{23187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9843, {{23188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9844, {{23189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9845, {{23190, 1, 0, 0},{23190, 1, 1, 17}, {0, 0, 0, 0}}},{9845, {{23190, 1, 0, 0},{23190, 1, 1, 17}, {0, 0, 0, 0}}},{9846, {{23190, 1, 0, 0},{23190, 1, 1, 18}, {0, 0, 0, 0}}},{9846, {{23190, 1, 0, 0},{23190, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9847, {{23191, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9848, {{23192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9849, {{23193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9850, {{23194, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9851, {{23195, 1, 0, 0},{23195, 1, 1, 19}, {0, 0, 0, 0}}},{9851, {{23195, 1, 0, 0},{23195, 1, 1, 19}, {0, 0, 0, 0}}},{9852, {{23195, 1, 0, 0},{23195, 1, 1, 18}, {0, 0, 0, 0}}},{9852, {{23195, 1, 0, 0},{23195, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9853, {{23196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9854, {{23197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9855, {{23198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9856, {{23199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9857, {{23200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9858, {{23201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9859, {{23202, 1, 0, 0},{23202, 1, 1, 17}, {0, 0, 0, 0}}},{9859, {{23202, 1, 0, 0},{23202, 1, 1, 17}, {0, 0, 0, 0}}},{9860, {{23202, 1, 0, 0},{23202, 1, 1, 18}, {0, 0, 0, 0}}},{9860, {{23202, 1, 0, 0},{23202, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9861, {{23203, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9862, {{23204, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9863, {{23205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9864, {{23206, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9865, {{23207, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9866, {{23208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9867, {{23209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9868, {{23210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{9869, {{23212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{9870, {{23216, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9871, {{23217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9872, {{23218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9873, {{23219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9874, {{23220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9875, {{23221, 1, 0, 0},{23221, 1, 1, 17}, {0, 0, 0, 0}}},{9875, {{23221, 1, 0, 0},{23221, 1, 1, 17}, {0, 0, 0, 0}}},{9876, {{23221, 1, 0, 0},{23221, 1, 1, 18}, {0, 0, 0, 0}}},{9876, {{23221, 1, 0, 0},{23221, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9877, {{23222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9878, {{23223, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9879, {{23224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9880, {{23225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{9881, {{23226, 1, 0, 0},{23226, 1, 1, 17}, {0, 0, 0, 0}}},{9881, {{23226, 1, 0, 0},{23226, 1, 1, 17}, {0, 0, 0, 0}}},{9882, {{23226, 1, 0, 0},{23226, 1, 1, 18}, {0, 0, 0, 0}}},{9882, {{23226, 1, 0, 0},{23226, 1, 1, 18}, {0, 0, 0, 0}}},{9883, {{23226, 1, 0, 0},{23226, 1, 1, 19}, {0, 0, 0, 0}}},{9883, {{23226, 1, 0, 0},{23226, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{9884, {{23227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9885, {{23228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9886, {{23229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9887, {{23230, 1, 0, 0},{23230, 1, 1, 20}, {0, 0, 0, 0}}},{9887, {{23230, 1, 0, 0},{23230, 1, 1, 20}, {0, 0, 0, 0}}},{9888, {{23230, 1, 0, 0},{23230, 1, 1, 19}, {0, 0, 0, 0}}},{9888, {{23230, 1, 0, 0},{23230, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{9889, {{23231, 1, 0, 0},{23231, 1, 1, 17}, {0, 0, 0, 0}}},{9889, {{23231, 1, 0, 0},{23231, 1, 1, 17}, {0, 0, 0, 0}}},{57310, {{23231, 1, 0, 0},{23231, 1, 1, 18}, {0, 0, 0, 0}}},{57310, {{23231, 1, 0, 0},{23231, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9890, {{23232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9891, {{23233, 1, 0, 0},{23233, 1, 1, 17}, {0, 0, 0, 0}}},{9891, {{23233, 1, 0, 0},{23233, 1, 1, 17}, {0, 0, 0, 0}}},{9892, {{23233, 1, 0, 0},{23233, 1, 1, 18}, {0, 0, 0, 0}}},{9892, {{23233, 1, 0, 0},{23233, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{9893, {{23234, 1, 0, 0},{23234, 1, 1, 18}, {0, 0, 0, 0}}},{9893, {{23234, 1, 0, 0},{23234, 1, 1, 18}, {0, 0, 0, 0}}},{9894, {{23234, 1, 0, 0},{23234, 1, 1, 19}, {0, 0, 0, 0}}},{9894, {{23234, 1, 0, 0},{23234, 1, 1, 19}, {0, 0, 0, 0}}},{9895, {{23234, 1, 0, 0},{23234, 1, 1, 20}, {0, 0, 0, 0}}},{9895, {{23234, 1, 0, 0},{23234, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{9896, {{23235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9899, {{23236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9900, {{23237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9901, {{23238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9903, {{23239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9904, {{23240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9905, {{23241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9906, {{23242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9907, {{23243, 1, 0, 0},{23243, 1, 1, 17}, {0, 0, 0, 0}}},{9907, {{23243, 1, 0, 0},{23243, 1, 1, 17}, {0, 0, 0, 0}}},{9908, {{23243, 1, 0, 0},{23243, 1, 1, 18}, {0, 0, 0, 0}}},{9908, {{23243, 1, 0, 0},{23243, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{9909, {{23244, 1, 0, 0},{23244, 1, 1, 19}, {0, 0, 0, 0}}},{9909, {{23244, 1, 0, 0},{23244, 1, 1, 19}, {0, 0, 0, 0}}},{9910, {{23244, 1, 0, 0},{23244, 1, 1, 18}, {0, 0, 0, 0}}},{9910, {{23244, 1, 0, 0},{23244, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9911, {{23245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9912, {{23246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9913, {{23247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9914, {{23248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{9915, {{23253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9916, {{23254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9917, {{23255, 1, 0, 0},{23255, 1, 1, 17}, {0, 0, 0, 0}}},{9917, {{23255, 1, 0, 0},{23255, 1, 1, 17}, {0, 0, 0, 0}}},{9918, {{23255, 1, 0, 0},{23255, 1, 1, 18}, {0, 0, 0, 0}}},{9918, {{23255, 1, 0, 0},{23255, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9919, {{23256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9920, {{23257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9921, {{23258, 1, 0, 0},{23258, 1, 1, 18}, {0, 0, 0, 0}}},{9921, {{23258, 1, 0, 0},{23258, 1, 1, 18}, {0, 0, 0, 0}}},{9922, {{23258, 1, 0, 0},{23258, 1, 1, 19}, {0, 0, 0, 0}}},{9922, {{23258, 1, 0, 0},{23258, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{9923, {{23259, 1, 0, 0},{23259, 1, 1, 16}, {0, 0, 0, 0}}},{9923, {{23259, 1, 0, 0},{23259, 1, 1, 16}, {0, 0, 0, 0}}},{9924, {{23259, 1, 0, 0},{23259, 1, 1, 17}, {0, 0, 0, 0}}},{9924, {{23259, 1, 0, 0},{23259, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{9925, {{23260, 1, 0, 0},{23260, 1, 1, 17}, {0, 0, 0, 0}}},{9925, {{23260, 1, 0, 0},{23260, 1, 1, 17}, {0, 0, 0, 0}}},{9926, {{23260, 1, 0, 0},{23260, 1, 1, 18}, {0, 0, 0, 0}}},{9926, {{23260, 1, 0, 0},{23260, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9927, {{23261, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9928, {{23262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9929, {{23263, 1, 0, 0},{23263, 1, 1, 16}, {0, 0, 0, 0}}},{9929, {{23263, 1, 0, 0},{23263, 1, 1, 16}, {0, 0, 0, 0}}},{9930, {{23263, 1, 0, 0},{23263, 1, 1, 17}, {0, 0, 0, 0}}},{9930, {{23263, 1, 0, 0},{23263, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{9931, {{23264, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9932, {{23265, 1, 0, 0},{23265, 1, 1, 17}, {0, 0, 0, 0}}},{9932, {{23265, 1, 0, 0},{23265, 1, 1, 17}, {0, 0, 0, 0}}},{9933, {{23265, 1, 0, 0},{23265, 1, 1, 18}, {0, 0, 0, 0}}},{9933, {{23265, 1, 0, 0},{23265, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{9934, {{23266, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9935, {{23267, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9936, {{23268, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9937, {{23269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9938, {{23270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9939, {{23271, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9940, {{23272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9941, {{23273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9942, {{23274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9943, {{23275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9945, {{23276, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9946, {{23277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9947, {{23278, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9948, {{23279, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9949, {{23280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{9950, {{23283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9951, {{23284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9952, {{23285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9953, {{23286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9954, {{23287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9955, {{23288, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9957, {{23289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9958, {{23290, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9959, {{23291, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9960, {{23292, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9961, {{23293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9962, {{23294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9963, {{23295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9964, {{23296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9965, {{23297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9966, {{23298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9967, {{23299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9968, {{23300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{9969, {{23301, 1, 0, 0},{23301, 1, 1, 17}, {0, 0, 0, 0}}},{9969, {{23301, 1, 0, 0},{23301, 1, 1, 17}, {0, 0, 0, 0}}},{9970, {{23301, 1, 0, 0},{23301, 1, 1, 16}, {0, 0, 0, 0}}},{9970, {{23301, 1, 0, 0},{23301, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{9971, {{23302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9972, {{23303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9973, {{23304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9974, {{23305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9975, {{23306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9976, {{23307, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9977, {{23308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9978, {{23309, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9979, {{23310, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9980, {{23311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9981, {{23312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9982, {{23313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9983, {{23314, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9984, {{23315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9985, {{23316, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9986, {{23317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9987, {{23318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9988, {{23319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9989, {{23320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9990, {{23321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9991, {{23322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9992, {{23323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9993, {{23324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9994, {{23325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9995, {{23326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9996, {{23327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9997, {{23328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9998, {{23329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{9999, {{23330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10000, {{23331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10001, {{23332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10002, {{23333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10003, {{23334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10004, {{23335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10005, {{23336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10007, {{23337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10008, {{23338, 1, 0, 0},{23338, 1, 1, 17}, {0, 0, 0, 0}}},{10008, {{23338, 1, 0, 0},{23338, 1, 1, 17}, {0, 0, 0, 0}}},{10009, {{23338, 1, 0, 0},{23338, 1, 1, 18}, {0, 0, 0, 0}}},{10009, {{23338, 1, 0, 0},{23338, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10010, {{23339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10011, {{23340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10012, {{23341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10013, {{23342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10014, {{23343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10015, {{23344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10016, {{23345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10017, {{23346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10018, {{23347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{10019, {{23348, 1, 0, 0},{23348, 1, 1, 17}, {0, 0, 0, 0}}},{10019, {{23348, 1, 0, 0},{23348, 1, 1, 17}, {0, 0, 0, 0}}},{10020, {{23348, 1, 0, 0},{23348, 1, 1, 19}, {0, 0, 0, 0}}},{10020, {{23348, 1, 0, 0},{23348, 1, 1, 19}, {0, 0, 0, 0}}},{10021, {{23348, 1, 0, 0},{23348, 1, 1, 18}, {0, 0, 0, 0}}},{10021, {{23348, 1, 0, 0},{23348, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10023, {{23349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10024, {{23350, 1, 0, 0},{23350, 1, 1, 17}, {0, 0, 0, 0}}},{10024, {{23350, 1, 0, 0},{23350, 1, 1, 17}, {0, 0, 0, 0}}},{10025, {{23350, 1, 0, 0},{23350, 1, 1, 18}, {0, 0, 0, 0}}},{10025, {{23350, 1, 0, 0},{23350, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10026, {{23352, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10027, {{23353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10028, {{23354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10029, {{23355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10030, {{23356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10031, {{23357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10032, {{23358, 1, 0, 0},{23358, 1, 1, 17}, {0, 0, 0, 0}}},{10032, {{23358, 1, 0, 0},{23358, 1, 1, 17}, {0, 0, 0, 0}}},{10033, {{23358, 1, 0, 0},{23358, 1, 1, 18}, {0, 0, 0, 0}}},{10033, {{23358, 1, 0, 0},{23358, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10034, {{23359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10035, {{23360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10036, {{23361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10037, {{23362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10038, {{23363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10039, {{23364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10040, {{23365, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10041, {{23366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10042, {{23367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10043, {{23368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10044, {{23369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10045, {{23370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10046, {{23371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10047, {{23372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10048, {{23373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10049, {{23374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10050, {{23375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10051, {{23376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10052, {{23377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10053, {{23378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10054, {{23380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10055, {{23381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10056, {{23382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10057, {{23383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10058, {{23384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{10060, {{23386, 1, 0, 0},{23386, 1, 1, 18}, {0, 0, 0, 0}}},{10060, {{23386, 1, 0, 0},{23386, 1, 1, 18}, {0, 0, 0, 0}}},{10061, {{23386, 1, 0, 0},{23386, 1, 1, 19}, {0, 0, 0, 0}}},{10061, {{23386, 1, 0, 0},{23386, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{10062, {{23387, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10063, {{23388, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10064, {{23389, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10065, {{23390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10066, {{23391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10067, {{23392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10068, {{23393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10069, {{23394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10070, {{23395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{10071, {{23396, 1, 0, 0},{23396, 1, 1, 19}, {0, 0, 0, 0}}},{10071, {{23396, 1, 0, 0},{23396, 1, 1, 19}, {0, 0, 0, 0}}},{10072, {{23396, 1, 0, 0},{23396, 1, 1, 18}, {0, 0, 0, 0}}},{10072, {{23396, 1, 0, 0},{23396, 1, 1, 18}, {0, 0, 0, 0}}},{10073, {{23396, 1, 0, 0},{23396, 1, 1, 17}, {0, 0, 0, 0}}},{10073, {{23396, 1, 0, 0},{23396, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{10074, {{23397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10075, {{23398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10076, {{23400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10077, {{23401, 1, 0, 0},{23401, 1, 1, 17}, {0, 0, 0, 0}}},{10077, {{23401, 1, 0, 0},{23401, 1, 1, 17}, {0, 0, 0, 0}}},{10078, {{23401, 1, 0, 0},{23401, 1, 1, 18}, {0, 0, 0, 0}}},{10078, {{23401, 1, 0, 0},{23401, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10079, {{23403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10081, {{23404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10082, {{23406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10083, {{23407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10084, {{23408, 1, 0, 0},{23408, 1, 1, 17}, {0, 0, 0, 0}}},{10084, {{23408, 1, 0, 0},{23408, 1, 1, 17}, {0, 0, 0, 0}}},{57322, {{23408, 1, 0, 0},{23408, 1, 1, 18}, {0, 0, 0, 0}}},{57322, {{23408, 1, 0, 0},{23408, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10085, {{23409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10086, {{23410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10087, {{23411, 1, 0, 0},{23411, 1, 1, 18}, {0, 0, 0, 0}}},{10087, {{23411, 1, 0, 0},{23411, 1, 1, 18}, {0, 0, 0, 0}}},{10088, {{23411, 1, 0, 0},{23411, 1, 1, 19}, {0, 0, 0, 0}}},{10088, {{23411, 1, 0, 0},{23411, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{10089, {{23412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10090, {{23413, 1, 0, 0},{23413, 1, 1, 17}, {0, 0, 0, 0}}},{10090, {{23413, 1, 0, 0},{23413, 1, 1, 17}, {0, 0, 0, 0}}},{10091, {{23413, 1, 0, 0},{23413, 1, 1, 18}, {0, 0, 0, 0}}},{10091, {{23413, 1, 0, 0},{23413, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{10092, {{23414, 1, 0, 0},{23414, 1, 1, 18}, {0, 0, 0, 0}}},{10092, {{23414, 1, 0, 0},{23414, 1, 1, 18}, {0, 0, 0, 0}}},{10093, {{23414, 1, 0, 0},{23414, 1, 1, 17}, {0, 0, 0, 0}}},{10093, {{23414, 1, 0, 0},{23414, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{10094, {{23415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10095, {{23416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10096, {{23417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10097, {{23418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10098, {{23419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10099, {{23420, 1, 0, 0},{23420, 1, 1, 19}, {0, 0, 0, 0}}},{10099, {{23420, 1, 0, 0},{23420, 1, 1, 19}, {0, 0, 0, 0}}},{10100, {{23420, 1, 0, 0},{23420, 1, 1, 18}, {0, 0, 0, 0}}},{10100, {{23420, 1, 0, 0},{23420, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{10101, {{23421, 1, 0, 0},{23421, 1, 1, 18}, {0, 0, 0, 0}}},{10101, {{23421, 1, 0, 0},{23421, 1, 1, 18}, {0, 0, 0, 0}}},{10102, {{23421, 1, 0, 0},{23421, 1, 1, 17}, {0, 0, 0, 0}}},{10102, {{23421, 1, 0, 0},{23421, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{10103, {{23422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10104, {{23423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10105, {{23424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10106, {{23425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10107, {{23426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10108, {{23427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10109, {{23428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10110, {{23429, 1, 0, 0},{23429, 1, 1, 17}, {0, 0, 0, 0}}},{10110, {{23429, 1, 0, 0},{23429, 1, 1, 17}, {0, 0, 0, 0}}},{57328, {{23429, 1, 0, 0},{23429, 1, 1, 18}, {0, 0, 0, 0}}},{57328, {{23429, 1, 0, 0},{23429, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10111, {{23430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10112, {{23431, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10113, {{23432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10114, {{23433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10115, {{23434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10116, {{23435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10117, {{23436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10118, {{23437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10119, {{23438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10120, {{23439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10121, {{23440, 1, 0, 0},{23440, 1, 1, 17}, {0, 0, 0, 0}}},{10121, {{23440, 1, 0, 0},{23440, 1, 1, 17}, {0, 0, 0, 0}}},{10122, {{23440, 1, 0, 0},{23440, 1, 1, 18}, {0, 0, 0, 0}}},{10122, {{23440, 1, 0, 0},{23440, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10123, {{23441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10124, {{23442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10125, {{23443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10126, {{23444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10127, {{23445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10128, {{23446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10129, {{23447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10130, {{23448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10131, {{23449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10132, {{23450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10133, {{23451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10134, {{23452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10135, {{23453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10136, {{23454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10137, {{23455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10138, {{23458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10139, {{23459, 1, 0, 0},{23459, 1, 1, 17}, {0, 0, 0, 0}}},{10139, {{23459, 1, 0, 0},{23459, 1, 1, 17}, {0, 0, 0, 0}}},{10140, {{23459, 1, 0, 0},{23459, 1, 1, 18}, {0, 0, 0, 0}}},{10140, {{23459, 1, 0, 0},{23459, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10141, {{23460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10142, {{23461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10143, {{23462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10144, {{23463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10145, {{23464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10146, {{23465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10147, {{23466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10148, {{23467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10149, {{23468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10151, {{23469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10153, {{23470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10154, {{23471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10155, {{23472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10156, {{23473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10157, {{23474, 1, 0, 0},{23474, 1, 1, 17}, {0, 0, 0, 0}}},{10157, {{23474, 1, 0, 0},{23474, 1, 1, 17}, {0, 0, 0, 0}}},{10158, {{23474, 1, 0, 0},{23474, 1, 1, 18}, {0, 0, 0, 0}}},{10158, {{23474, 1, 0, 0},{23474, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{10159, {{23475, 1, 0, 0},{23475, 1, 1, 21}, {0, 0, 0, 0}}},{10159, {{23475, 1, 0, 0},{23475, 1, 1, 21}, {0, 0, 0, 0}}},{10160, {{23475, 1, 0, 0},{23475, 1, 1, 20}, {0, 0, 0, 0}}},{10160, {{23475, 1, 0, 0},{23475, 1, 1, 20}, {0, 0, 0, 0}}},{10161, {{23475, 1, 0, 0},{23475, 1, 1, 22}, {0, 0, 0, 0}}},{10161, {{23475, 1, 0, 0},{23475, 1, 1, 22}, {0, 0, 0, 0}}}}},
    {1, {{10162, {{23476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10163, {{23477, 1, 0, 0},{23477, 1, 1, 19}, {0, 0, 0, 0}}},{10163, {{23477, 1, 0, 0},{23477, 1, 1, 19}, {0, 0, 0, 0}}},{10164, {{23477, 1, 0, 0},{23477, 1, 1, 18}, {0, 0, 0, 0}}},{10164, {{23477, 1, 0, 0},{23477, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{10165, {{23478, 1, 0, 0},{23478, 1, 1, 17}, {0, 0, 0, 0}}},{10165, {{23478, 1, 0, 0},{23478, 1, 1, 17}, {0, 0, 0, 0}}},{10166, {{23478, 1, 0, 0},{23478, 1, 1, 18}, {0, 0, 0, 0}}},{10166, {{23478, 1, 0, 0},{23478, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10167, {{23479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10168, {{23480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10169, {{23481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10170, {{23482, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10171, {{23483, 1, 0, 0},{23483, 1, 1, 16}, {0, 0, 0, 0}}},{10171, {{23483, 1, 0, 0},{23483, 1, 1, 16}, {0, 0, 0, 0}}},{57334, {{23483, 1, 0, 0},{23483, 1, 1, 17}, {0, 0, 0, 0}}},{57334, {{23483, 1, 0, 0},{23483, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{10172, {{23484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10174, {{23487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10175, {{23488, 1, 0, 0},{23488, 1, 1, 17}, {0, 0, 0, 0}}},{10175, {{23488, 1, 0, 0},{23488, 1, 1, 17}, {0, 0, 0, 0}}},{10176, {{23488, 1, 0, 0},{23488, 1, 1, 18}, {0, 0, 0, 0}}},{10176, {{23488, 1, 0, 0},{23488, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10177, {{23489, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10178, {{23490, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{10179, {{23491, 1, 0, 0},{23491, 1, 1, 18}, {0, 0, 0, 0}}},{10179, {{23491, 1, 0, 0},{23491, 1, 1, 18}, {0, 0, 0, 0}}},{10180, {{23491, 1, 0, 0},{23491, 1, 1, 20}, {0, 0, 0, 0}}},{10180, {{23491, 1, 0, 0},{23491, 1, 1, 20}, {0, 0, 0, 0}}},{10181, {{23491, 1, 0, 0},{23491, 1, 1, 19}, {0, 0, 0, 0}}},{10181, {{23491, 1, 0, 0},{23491, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{10182, {{23492, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10183, {{23493, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10184, {{23494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10185, {{23495, 1, 0, 0},{23495, 1, 1, 17}, {0, 0, 0, 0}}},{10185, {{23495, 1, 0, 0},{23495, 1, 1, 17}, {0, 0, 0, 0}}},{10186, {{23495, 1, 0, 0},{23495, 1, 1, 18}, {0, 0, 0, 0}}},{10186, {{23495, 1, 0, 0},{23495, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{10187, {{23496, 1, 0, 0},{23496, 1, 1, 16}, {0, 0, 0, 0}}},{10187, {{23496, 1, 0, 0},{23496, 1, 1, 16}, {0, 0, 0, 0}}},{10188, {{23496, 1, 0, 0},{23496, 1, 1, 17}, {0, 0, 0, 0}}},{10188, {{23496, 1, 0, 0},{23496, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{10189, {{23497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10190, {{23498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10191, {{23499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10192, {{23500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10193, {{23501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10194, {{23502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10195, {{23503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10196, {{23504, 1, 0, 0},{23504, 1, 1, 17}, {0, 0, 0, 0}}},{10196, {{23504, 1, 0, 0},{23504, 1, 1, 17}, {0, 0, 0, 0}}},{10197, {{23504, 1, 0, 0},{23504, 1, 1, 18}, {0, 0, 0, 0}}},{10197, {{23504, 1, 0, 0},{23504, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10198, {{23505, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10199, {{23506, 1, 0, 0},{23506, 1, 1, 18}, {0, 0, 0, 0}}},{10199, {{23506, 1, 0, 0},{23506, 1, 1, 18}, {0, 0, 0, 0}}},{10200, {{23506, 1, 0, 0},{23506, 1, 1, 19}, {0, 0, 0, 0}}},{10200, {{23506, 1, 0, 0},{23506, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{10201, {{23507, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10202, {{23508, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34014, {{23509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10204, {{23510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{10205, {{23511, 1, 0, 0},{23511, 1, 1, 17}, {0, 0, 0, 0}}},{10205, {{23511, 1, 0, 0},{23511, 1, 1, 17}, {0, 0, 0, 0}}},{10206, {{23511, 1, 0, 0},{23511, 1, 1, 18}, {0, 0, 0, 0}}},{10206, {{23511, 1, 0, 0},{23511, 1, 1, 18}, {0, 0, 0, 0}}},{10207, {{23511, 1, 0, 0},{23511, 1, 1, 19}, {0, 0, 0, 0}}},{10207, {{23511, 1, 0, 0},{23511, 1, 1, 19}, {0, 0, 0, 0}}},{10208, {{23511, 1, 0, 0},{23511, 1, 1, 20}, {0, 0, 0, 0}}},{10208, {{23511, 1, 0, 0},{23511, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{10209, {{23512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10210, {{23513, 1, 0, 0},{23513, 1, 1, 17}, {0, 0, 0, 0}}},{10210, {{23513, 1, 0, 0},{23513, 1, 1, 17}, {0, 0, 0, 0}}},{10211, {{23513, 1, 0, 0},{23513, 1, 1, 18}, {0, 0, 0, 0}}},{10211, {{23513, 1, 0, 0},{23513, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10212, {{23514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{10213, {{23515, 1, 0, 0},{23515, 1, 1, 18}, {0, 0, 0, 0}}},{10213, {{23515, 1, 0, 0},{23515, 1, 1, 18}, {0, 0, 0, 0}}},{10214, {{23515, 1, 0, 0},{23515, 1, 1, 19}, {0, 0, 0, 0}}},{10214, {{23515, 1, 0, 0},{23515, 1, 1, 19}, {0, 0, 0, 0}}},{59508, {{23515, 1, 0, 0},{23515, 1, 1, 20}, {0, 0, 0, 0}}},{59508, {{23515, 1, 0, 0},{23515, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{10215, {{23516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10216, {{23517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10217, {{23518, 1, 0, 0},{23518, 1, 1, 17}, {0, 0, 0, 0}}},{10217, {{23518, 1, 0, 0},{23518, 1, 1, 17}, {0, 0, 0, 0}}},{10218, {{23518, 1, 0, 0},{23518, 1, 1, 18}, {0, 0, 0, 0}}},{10218, {{23518, 1, 0, 0},{23518, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10219, {{23519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10220, {{23520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10221, {{23521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10222, {{23522, 1, 0, 0},{23522, 1, 1, 17}, {0, 0, 0, 0}}},{10222, {{23522, 1, 0, 0},{23522, 1, 1, 17}, {0, 0, 0, 0}}},{10223, {{23522, 1, 0, 0},{23522, 1, 1, 18}, {0, 0, 0, 0}}},{10223, {{23522, 1, 0, 0},{23522, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10224, {{23523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10225, {{23524, 1, 0, 0},{23524, 1, 1, 17}, {0, 0, 0, 0}}},{10225, {{23524, 1, 0, 0},{23524, 1, 1, 17}, {0, 0, 0, 0}}},{10226, {{23524, 1, 0, 0},{23524, 1, 1, 18}, {0, 0, 0, 0}}},{10226, {{23524, 1, 0, 0},{23524, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{10227, {{23525, 1, 0, 0},{23525, 1, 1, 17}, {0, 0, 0, 0}}},{10227, {{23525, 1, 0, 0},{23525, 1, 1, 17}, {0, 0, 0, 0}}},{10228, {{23525, 1, 0, 0},{23525, 1, 1, 18}, {0, 0, 0, 0}}},{10228, {{23525, 1, 0, 0},{23525, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10229, {{23526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10230, {{23527, 1, 0, 0},{23527, 1, 1, 19}, {0, 0, 0, 0}}},{10230, {{23527, 1, 0, 0},{23527, 1, 1, 19}, {0, 0, 0, 0}}},{10231, {{23527, 1, 0, 0},{23527, 1, 1, 18}, {0, 0, 0, 0}}},{10231, {{23527, 1, 0, 0},{23527, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10232, {{23528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10233, {{23529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10234, {{23530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10235, {{23531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{10236, {{23532, 1, 0, 0},{23532, 1, 1, 17}, {0, 0, 0, 0}}},{10236, {{23532, 1, 0, 0},{23532, 1, 1, 17}, {0, 0, 0, 0}}},{10237, {{23532, 1, 0, 0},{23532, 1, 1, 18}, {0, 0, 0, 0}}},{10237, {{23532, 1, 0, 0},{23532, 1, 1, 18}, {0, 0, 0, 0}}},{10238, {{23532, 1, 0, 0},{23532, 1, 1, 19}, {0, 0, 0, 0}}},{10238, {{23532, 1, 0, 0},{23532, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{10239, {{23533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10241, {{23534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10242, {{23535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10243, {{23536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10244, {{23537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10245, {{23538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10246, {{23539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10247, {{23540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{10248, {{23541, 1, 0, 0},{23541, 1, 1, 17}, {0, 0, 0, 0}}},{10248, {{23541, 1, 0, 0},{23541, 1, 1, 17}, {0, 0, 0, 0}}},{10249, {{23541, 1, 0, 0},{23541, 1, 1, 19}, {0, 0, 0, 0}}},{10249, {{23541, 1, 0, 0},{23541, 1, 1, 19}, {0, 0, 0, 0}}},{10250, {{23541, 1, 0, 0},{23541, 1, 1, 18}, {0, 0, 0, 0}}},{10250, {{23541, 1, 0, 0},{23541, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{10251, {{23542, 1, 0, 0},{23542, 1, 1, 17}, {0, 0, 0, 0}}},{10251, {{23542, 1, 0, 0},{23542, 1, 1, 17}, {0, 0, 0, 0}}},{59512, {{23542, 1, 0, 0},{23542, 1, 1, 18}, {0, 0, 0, 0}}},{59512, {{23542, 1, 0, 0},{23542, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10252, {{23543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10253, {{23544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10254, {{23545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10255, {{23546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{10256, {{23549, 1, 0, 0},{23549, 1, 1, 17}, {0, 0, 0, 0}}},{10256, {{23549, 1, 0, 0},{23549, 1, 1, 17}, {0, 0, 0, 0}}},{10257, {{23549, 1, 0, 0},{23549, 1, 1, 18}, {0, 0, 0, 0}}},{10257, {{23549, 1, 0, 0},{23549, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10258, {{23550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10259, {{23551, 1, 0, 0},{23551, 1, 1, 18}, {0, 0, 0, 0}}},{10259, {{23551, 1, 0, 0},{23551, 1, 1, 18}, {0, 0, 0, 0}}},{10260, {{23551, 1, 0, 0},{23551, 1, 1, 17}, {0, 0, 0, 0}}},{10260, {{23551, 1, 0, 0},{23551, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{10261, {{23552, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10262, {{23553, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10263, {{23554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10264, {{23555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10265, {{23556, 1, 0, 0},{23556, 1, 1, 17}, {0, 0, 0, 0}}},{10265, {{23556, 1, 0, 0},{23556, 1, 1, 17}, {0, 0, 0, 0}}},{10266, {{23556, 1, 0, 0},{23556, 1, 1, 18}, {0, 0, 0, 0}}},{10266, {{23556, 1, 0, 0},{23556, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10267, {{23557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{10268, {{23558, 1, 0, 0},{23558, 1, 1, 20}, {0, 0, 0, 0}}},{10268, {{23558, 1, 0, 0},{23558, 1, 1, 20}, {0, 0, 0, 0}}},{10269, {{23558, 1, 0, 0},{23558, 1, 1, 18}, {0, 0, 0, 0}}},{10269, {{23558, 1, 0, 0},{23558, 1, 1, 18}, {0, 0, 0, 0}}},{10270, {{23558, 1, 0, 0},{23558, 1, 1, 19}, {0, 0, 0, 0}}},{10270, {{23558, 1, 0, 0},{23558, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {8, {{10271, {{23559, 1, 0, 0},{23559, 1, 1, 17}, {0, 0, 0, 0}}},{10271, {{23559, 1, 0, 0},{23559, 1, 1, 17}, {0, 0, 0, 0}}},{58022, {{23559, 1, 0, 0},{23559, 1, 1, 20}, {0, 0, 0, 0}}},{58022, {{23559, 1, 0, 0},{23559, 1, 1, 20}, {0, 0, 0, 0}}},{59513, {{23559, 1, 0, 0},{23559, 1, 1, 19}, {0, 0, 0, 0}}},{59513, {{23559, 1, 0, 0},{23559, 1, 1, 19}, {0, 0, 0, 0}}},{59879, {{23559, 1, 0, 0},{23559, 1, 1, 18}, {0, 0, 0, 0}}},{59879, {{23559, 1, 0, 0},{23559, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{10272, {{23560, 1, 0, 0},{23560, 1, 1, 17}, {0, 0, 0, 0}}},{10272, {{23560, 1, 0, 0},{23560, 1, 1, 17}, {0, 0, 0, 0}}},{10273, {{23560, 1, 0, 0},{23560, 1, 1, 18}, {0, 0, 0, 0}}},{10273, {{23560, 1, 0, 0},{23560, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{10274, {{23561, 1, 0, 0},{23561, 1, 1, 17}, {0, 0, 0, 0}}},{10274, {{23561, 1, 0, 0},{23561, 1, 1, 17}, {0, 0, 0, 0}}},{57350, {{23561, 1, 0, 0},{23561, 1, 1, 18}, {0, 0, 0, 0}}},{57350, {{23561, 1, 0, 0},{23561, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{10275, {{23562, 1, 0, 0},{23562, 1, 1, 20}, {0, 0, 0, 0}}},{10275, {{23562, 1, 0, 0},{23562, 1, 1, 20}, {0, 0, 0, 0}}},{10276, {{23562, 1, 0, 0},{23562, 1, 1, 19}, {0, 0, 0, 0}}},{10276, {{23562, 1, 0, 0},{23562, 1, 1, 19}, {0, 0, 0, 0}}},{10277, {{23562, 1, 0, 0},{23562, 1, 1, 21}, {0, 0, 0, 0}}},{10277, {{23562, 1, 0, 0},{23562, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {4, {{10278, {{23563, 1, 0, 0},{23563, 1, 1, 19}, {0, 0, 0, 0}}},{10278, {{23563, 1, 0, 0},{23563, 1, 1, 19}, {0, 0, 0, 0}}},{10279, {{23563, 1, 0, 0},{23563, 1, 1, 18}, {0, 0, 0, 0}}},{10279, {{23563, 1, 0, 0},{23563, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10280, {{23564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10281, {{23565, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10282, {{23566, 1, 0, 0},{23566, 1, 1, 18}, {0, 0, 0, 0}}},{10282, {{23566, 1, 0, 0},{23566, 1, 1, 18}, {0, 0, 0, 0}}},{10283, {{23566, 1, 0, 0},{23566, 1, 1, 19}, {0, 0, 0, 0}}},{10283, {{23566, 1, 0, 0},{23566, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{10284, {{23567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10285, {{23568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10286, {{23569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10287, {{23570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10289, {{23571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10290, {{23572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10291, {{23574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10292, {{23575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10293, {{23576, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10294, {{23577, 1, 0, 0},{23577, 1, 1, 17}, {0, 0, 0, 0}}},{10294, {{23577, 1, 0, 0},{23577, 1, 1, 17}, {0, 0, 0, 0}}},{10295, {{23577, 1, 0, 0},{23577, 1, 1, 18}, {0, 0, 0, 0}}},{10295, {{23577, 1, 0, 0},{23577, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10296, {{23578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10297, {{23579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10298, {{23580, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10299, {{23582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10300, {{23583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10301, {{23584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10303, {{23586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10304, {{23587, 1, 0, 0},{23587, 1, 1, 17}, {0, 0, 0, 0}}},{10304, {{23587, 1, 0, 0},{23587, 1, 1, 17}, {0, 0, 0, 0}}},{10305, {{23587, 1, 0, 0},{23587, 1, 1, 18}, {0, 0, 0, 0}}},{10305, {{23587, 1, 0, 0},{23587, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10306, {{23588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10307, {{23589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10308, {{23590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{10309, {{23592, 1, 0, 0},{23592, 1, 1, 20}, {0, 0, 0, 0}}},{10309, {{23592, 1, 0, 0},{23592, 1, 1, 20}, {0, 0, 0, 0}}},{10310, {{23592, 1, 0, 0},{23592, 1, 1, 18}, {0, 0, 0, 0}}},{10310, {{23592, 1, 0, 0},{23592, 1, 1, 18}, {0, 0, 0, 0}}},{10311, {{23592, 1, 0, 0},{23592, 1, 1, 19}, {0, 0, 0, 0}}},{10311, {{23592, 1, 0, 0},{23592, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{10312, {{23593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10313, {{23594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10314, {{23595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10315, {{23596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{10316, {{23597, 1, 0, 0},{23597, 1, 1, 17}, {0, 0, 0, 0}}},{10316, {{23597, 1, 0, 0},{23597, 1, 1, 17}, {0, 0, 0, 0}}},{10317, {{23597, 1, 0, 0},{23597, 1, 1, 18}, {0, 0, 0, 0}}},{10317, {{23597, 1, 0, 0},{23597, 1, 1, 18}, {0, 0, 0, 0}}},{10318, {{23597, 1, 0, 0},{23597, 1, 1, 19}, {0, 0, 0, 0}}},{10318, {{23597, 1, 0, 0},{23597, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{10319, {{23598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10320, {{23599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10321, {{23600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10322, {{23601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10323, {{23602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10324, {{23603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10325, {{23605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10326, {{23606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10327, {{23607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10328, {{23608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10329, {{23609, 1, 0, 0},{23609, 1, 1, 17}, {0, 0, 0, 0}}},{10329, {{23609, 1, 0, 0},{23609, 1, 1, 17}, {0, 0, 0, 0}}},{10330, {{23609, 1, 0, 0},{23609, 1, 1, 18}, {0, 0, 0, 0}}},{10330, {{23609, 1, 0, 0},{23609, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10331, {{23610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10332, {{23611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10333, {{23612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10334, {{23613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10335, {{23614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10336, {{23615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10337, {{23616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10338, {{23617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10339, {{23620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10341, {{23621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10342, {{23622, 1, 0, 0},{23622, 1, 1, 17}, {0, 0, 0, 0}}},{10342, {{23622, 1, 0, 0},{23622, 1, 1, 17}, {0, 0, 0, 0}}},{10343, {{23622, 1, 0, 0},{23622, 1, 1, 18}, {0, 0, 0, 0}}},{10343, {{23622, 1, 0, 0},{23622, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10344, {{23623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10345, {{23624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10346, {{23625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10347, {{23626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10348, {{23627, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10349, {{23628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10350, {{23629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10351, {{23630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10352, {{23631, 1, 0, 0},{23631, 1, 1, 17}, {0, 0, 0, 0}}},{10352, {{23631, 1, 0, 0},{23631, 1, 1, 17}, {0, 0, 0, 0}}},{10353, {{23631, 1, 0, 0},{23631, 1, 1, 18}, {0, 0, 0, 0}}},{10353, {{23631, 1, 0, 0},{23631, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10354, {{23632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10355, {{23633, 1, 0, 0},{23633, 1, 1, 19}, {0, 0, 0, 0}}},{10355, {{23633, 1, 0, 0},{23633, 1, 1, 19}, {0, 0, 0, 0}}},{10356, {{23633, 1, 0, 0},{23633, 1, 1, 18}, {0, 0, 0, 0}}},{10356, {{23633, 1, 0, 0},{23633, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10357, {{23634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10358, {{23635, 1, 0, 0},{23635, 1, 1, 17}, {0, 0, 0, 0}}},{10358, {{23635, 1, 0, 0},{23635, 1, 1, 17}, {0, 0, 0, 0}}},{10359, {{23635, 1, 0, 0},{23635, 1, 1, 18}, {0, 0, 0, 0}}},{10359, {{23635, 1, 0, 0},{23635, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10360, {{23636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10361, {{23637, 1, 0, 0},{23637, 1, 1, 17}, {0, 0, 0, 0}}},{10361, {{23637, 1, 0, 0},{23637, 1, 1, 17}, {0, 0, 0, 0}}},{10362, {{23637, 1, 0, 0},{23637, 1, 1, 18}, {0, 0, 0, 0}}},{10362, {{23637, 1, 0, 0},{23637, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10363, {{23638, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10366, {{23641, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10367, {{23642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10368, {{23643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10369, {{23644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10370, {{23645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10371, {{23646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10372, {{23647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10373, {{23648, 1, 0, 0},{23648, 1, 1, 18}, {0, 0, 0, 0}}},{10373, {{23648, 1, 0, 0},{23648, 1, 1, 18}, {0, 0, 0, 0}}},{10374, {{23648, 1, 0, 0},{23648, 1, 1, 19}, {0, 0, 0, 0}}},{10374, {{23648, 1, 0, 0},{23648, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{10375, {{23649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10376, {{23650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10377, {{23651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{10378, {{23652, 1, 0, 0},{23652, 1, 1, 18}, {0, 0, 0, 0}}},{10378, {{23652, 1, 0, 0},{23652, 1, 1, 18}, {0, 0, 0, 0}}},{30255, {{23652, 1, 0, 0},{23652, 1, 1, 0},{23652, 1, 1, 19},{64059, 1, 0, 0}, {0, 0, 0, 0}}},{30255, {{23652, 1, 0, 0},{23652, 1, 1, 0},{23652, 1, 1, 19},{64059, 1, 0, 0}, {0, 0, 0, 0}}},{30255, {{23652, 1, 0, 0},{23652, 1, 1, 0},{23652, 1, 1, 19},{64059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10379, {{23653, 1, 0, 0},{23653, 1, 1, 18}, {0, 0, 0, 0}}},{10379, {{23653, 1, 0, 0},{23653, 1, 1, 18}, {0, 0, 0, 0}}},{10380, {{23653, 1, 0, 0},{23653, 1, 1, 17}, {0, 0, 0, 0}}},{10380, {{23653, 1, 0, 0},{23653, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10381, {{23655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10382, {{23656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10383, {{23657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10384, {{23658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10385, {{23659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10386, {{23660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10387, {{23661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{10388, {{23662, 1, 0, 0},{23662, 1, 1, 18}, {0, 0, 0, 0}}},{10388, {{23662, 1, 0, 0},{23662, 1, 1, 18}, {0, 0, 0, 0}}},{30256, {{23662, 1, 0, 0},{23662, 1, 1, 0},{23662, 1, 1, 19},{64060, 1, 0, 0}, {0, 0, 0, 0}}},{30256, {{23662, 1, 0, 0},{23662, 1, 1, 0},{23662, 1, 1, 19},{64060, 1, 0, 0}, {0, 0, 0, 0}}},{30256, {{23662, 1, 0, 0},{23662, 1, 1, 0},{23662, 1, 1, 19},{64060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10389, {{23663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10391, {{23664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10392, {{23665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10393, {{23667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10394, {{23668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10395, {{23669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10397, {{23670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10398, {{23672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10399, {{23673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10400, {{23674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10401, {{23675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10402, {{23676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10403, {{23677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10404, {{23678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10405, {{23680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10406, {{23684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10407, {{23685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10408, {{23686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10409, {{23687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10410, {{23688, 1, 0, 0},{23688, 1, 1, 18}, {0, 0, 0, 0}}},{10410, {{23688, 1, 0, 0},{23688, 1, 1, 18}, {0, 0, 0, 0}}},{10411, {{23688, 1, 0, 0},{23688, 1, 1, 17}, {0, 0, 0, 0}}},{10411, {{23688, 1, 0, 0},{23688, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{10412, {{23689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10413, {{23690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10414, {{23691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10416, {{23692, 1, 0, 0},{23692, 1, 1, 17}, {0, 0, 0, 0}}},{10416, {{23692, 1, 0, 0},{23692, 1, 1, 17}, {0, 0, 0, 0}}},{10417, {{23692, 1, 0, 0},{23692, 1, 1, 18}, {0, 0, 0, 0}}},{10417, {{23692, 1, 0, 0},{23692, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{10418, {{23693, 1, 0, 0},{23693, 1, 1, 16}, {0, 0, 0, 0}}},{10418, {{23693, 1, 0, 0},{23693, 1, 1, 16}, {0, 0, 0, 0}}},{10419, {{23693, 1, 0, 0},{23693, 1, 1, 17}, {0, 0, 0, 0}}},{10419, {{23693, 1, 0, 0},{23693, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{10420, {{23694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10421, {{23695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10422, {{23696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10423, {{23697, 1, 0, 0},{23697, 1, 1, 17}, {0, 0, 0, 0}}},{10423, {{23697, 1, 0, 0},{23697, 1, 1, 17}, {0, 0, 0, 0}}},{59523, {{23697, 1, 0, 0},{23697, 1, 1, 18}, {0, 0, 0, 0}}},{59523, {{23697, 1, 0, 0},{23697, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10424, {{23698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10425, {{23699, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10426, {{23700, 1, 0, 0},{23700, 1, 1, 17}, {0, 0, 0, 0}}},{10426, {{23700, 1, 0, 0},{23700, 1, 1, 17}, {0, 0, 0, 0}}},{10427, {{23700, 1, 0, 0},{23700, 1, 1, 18}, {0, 0, 0, 0}}},{10427, {{23700, 1, 0, 0},{23700, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10428, {{23701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10432, {{23709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10433, {{23710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10434, {{23711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10435, {{23712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10436, {{23713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10437, {{23714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10438, {{23715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10439, {{23716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10440, {{23717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10441, {{23718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10442, {{23719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10443, {{23720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10444, {{23721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10445, {{23722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10446, {{23723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10447, {{23724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10448, {{23725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10449, {{23726, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10450, {{23727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10451, {{23728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10452, {{23729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10453, {{23730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10454, {{23731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10455, {{23732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10456, {{23733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10457, {{23734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10458, {{23735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10459, {{23736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10461, {{23737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10462, {{23738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10464, {{23739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10465, {{23740, 1, 0, 0},{23740, 1, 1, 17}, {0, 0, 0, 0}}},{10465, {{23740, 1, 0, 0},{23740, 1, 1, 17}, {0, 0, 0, 0}}},{10466, {{23740, 1, 0, 0},{23740, 1, 1, 18}, {0, 0, 0, 0}}},{10466, {{23740, 1, 0, 0},{23740, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10467, {{23742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{10468, {{23744, 1, 0, 0},{23744, 1, 1, 16}, {0, 0, 0, 0}}},{10468, {{23744, 1, 0, 0},{23744, 1, 1, 16}, {0, 0, 0, 0}}},{68030, {{23744, 1, 0, 0},{23744, 1, 1, 17}, {0, 0, 0, 0}}},{68030, {{23744, 1, 0, 0},{23744, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10469, {{23749, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10470, {{23750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10471, {{23751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10472, {{23752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10473, {{23753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10474, {{23754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10475, {{23755, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10476, {{23756, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10477, {{23757, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10478, {{23758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10479, {{23759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10480, {{23760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10481, {{23761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10482, {{23762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10483, {{23763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10484, {{23764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10487, {{23766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10488, {{23767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10489, {{23768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10490, {{23769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10491, {{23770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10492, {{23771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10493, {{23772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10494, {{23773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10495, {{23774, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10496, {{23775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10497, {{23776, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10498, {{23777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10500, {{23782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10501, {{23784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10502, {{23785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10503, {{23786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10504, {{23787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10505, {{23788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10506, {{23789, 1, 0, 0},{23789, 1, 1, 17}, {0, 0, 0, 0}}},{10506, {{23789, 1, 0, 0},{23789, 1, 1, 17}, {0, 0, 0, 0}}},{10507, {{23789, 1, 0, 0},{23789, 1, 1, 18}, {0, 0, 0, 0}}},{10507, {{23789, 1, 0, 0},{23789, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10508, {{23790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10509, {{23791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10510, {{23792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10511, {{23793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10512, {{23794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10513, {{23795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10514, {{23796, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10515, {{23797, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10516, {{23798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10518, {{23799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10519, {{23800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10520, {{23801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10521, {{23802, 1, 0, 0},{23802, 1, 1, 17}, {0, 0, 0, 0}}},{10521, {{23802, 1, 0, 0},{23802, 1, 1, 17}, {0, 0, 0, 0}}},{10522, {{23802, 1, 0, 0},{23802, 1, 1, 18}, {0, 0, 0, 0}}},{10522, {{23802, 1, 0, 0},{23802, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{10523, {{23803, 1, 0, 0},{23803, 1, 1, 18}, {0, 0, 0, 0}}},{10523, {{23803, 1, 0, 0},{23803, 1, 1, 18}, {0, 0, 0, 0}}},{10524, {{23803, 1, 0, 0},{23803, 1, 1, 17}, {0, 0, 0, 0}}},{10524, {{23803, 1, 0, 0},{23803, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{10525, {{23804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10526, {{23805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10527, {{23806, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10528, {{23807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10529, {{23808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10530, {{23809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{10531, {{23814, 1, 0, 0},{23814, 1, 1, 17}, {0, 0, 0, 0}}},{10531, {{23814, 1, 0, 0},{23814, 1, 1, 17}, {0, 0, 0, 0}}},{10532, {{23814, 1, 0, 0},{23814, 1, 1, 18}, {0, 0, 0, 0}}},{10532, {{23814, 1, 0, 0},{23814, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10533, {{23815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10534, {{23816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10535, {{23817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10536, {{23818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10537, {{23819, 1, 0, 0},{23819, 1, 1, 17}, {0, 0, 0, 0}}},{10537, {{23819, 1, 0, 0},{23819, 1, 1, 17}, {0, 0, 0, 0}}},{10538, {{23819, 1, 0, 0},{23819, 1, 1, 18}, {0, 0, 0, 0}}},{10538, {{23819, 1, 0, 0},{23819, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10539, {{23820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10540, {{23821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10541, {{23822, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10542, {{23823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10543, {{23824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10544, {{23825, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10545, {{23826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10546, {{23827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10547, {{23828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10548, {{23829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10549, {{23830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10550, {{23831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10551, {{23832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10552, {{23833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10553, {{23834, 1, 0, 0},{23834, 1, 1, 18}, {0, 0, 0, 0}}},{10553, {{23834, 1, 0, 0},{23834, 1, 1, 18}, {0, 0, 0, 0}}},{10554, {{23834, 1, 0, 0},{23834, 1, 1, 17}, {0, 0, 0, 0}}},{10554, {{23834, 1, 0, 0},{23834, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{10555, {{23835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10556, {{23836, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10557, {{23837, 1, 0, 0},{23837, 1, 1, 17}, {0, 0, 0, 0}}},{10557, {{23837, 1, 0, 0},{23837, 1, 1, 17}, {0, 0, 0, 0}}},{10558, {{23837, 1, 0, 0},{23837, 1, 1, 18}, {0, 0, 0, 0}}},{10558, {{23837, 1, 0, 0},{23837, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10559, {{23838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10560, {{23839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10561, {{23840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10562, {{23841, 1, 0, 0},{23841, 1, 1, 16}, {0, 0, 0, 0}}},{10562, {{23841, 1, 0, 0},{23841, 1, 1, 16}, {0, 0, 0, 0}}},{10563, {{23841, 1, 0, 0},{23841, 1, 1, 17}, {0, 0, 0, 0}}},{10563, {{23841, 1, 0, 0},{23841, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{10564, {{23842, 1, 0, 0},{23842, 1, 1, 17}, {0, 0, 0, 0}}},{10564, {{23842, 1, 0, 0},{23842, 1, 1, 17}, {0, 0, 0, 0}}},{10565, {{23842, 1, 0, 0},{23842, 1, 1, 18}, {0, 0, 0, 0}}},{10565, {{23842, 1, 0, 0},{23842, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10566, {{23843, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10567, {{23844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10568, {{23845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10569, {{23846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10570, {{23847, 1, 0, 0},{23847, 1, 1, 17}, {0, 0, 0, 0}}},{10570, {{23847, 1, 0, 0},{23847, 1, 1, 17}, {0, 0, 0, 0}}},{10571, {{23847, 1, 0, 0},{23847, 1, 1, 18}, {0, 0, 0, 0}}},{10571, {{23847, 1, 0, 0},{23847, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10572, {{23848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10573, {{23849, 1, 0, 0},{23849, 1, 1, 19}, {0, 0, 0, 0}}},{10573, {{23849, 1, 0, 0},{23849, 1, 1, 19}, {0, 0, 0, 0}}},{10574, {{23849, 1, 0, 0},{23849, 1, 1, 18}, {0, 0, 0, 0}}},{10574, {{23849, 1, 0, 0},{23849, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{10575, {{23850, 1, 0, 0},{23850, 1, 1, 16}, {0, 0, 0, 0}}},{10575, {{23850, 1, 0, 0},{23850, 1, 1, 16}, {0, 0, 0, 0}}},{10576, {{23850, 1, 0, 0},{23850, 1, 1, 17}, {0, 0, 0, 0}}},{10576, {{23850, 1, 0, 0},{23850, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{10577, {{23851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10578, {{23857, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10579, {{23858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10580, {{23859, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10581, {{23860, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10582, {{23861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10583, {{23862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10584, {{23863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10585, {{23864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10586, {{23865, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10587, {{23866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10588, {{23867, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10589, {{23868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10590, {{23869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10591, {{23870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10592, {{23871, 1, 0, 0},{23871, 1, 1, 17}, {0, 0, 0, 0}}},{10592, {{23871, 1, 0, 0},{23871, 1, 1, 17}, {0, 0, 0, 0}}},{10593, {{23871, 1, 0, 0},{23871, 1, 1, 18}, {0, 0, 0, 0}}},{10593, {{23871, 1, 0, 0},{23871, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10594, {{23872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10595, {{23873, 1, 0, 0},{23873, 1, 1, 16}, {0, 0, 0, 0}}},{10595, {{23873, 1, 0, 0},{23873, 1, 1, 16}, {0, 0, 0, 0}}},{10596, {{23873, 1, 0, 0},{23873, 1, 1, 17}, {0, 0, 0, 0}}},{10596, {{23873, 1, 0, 0},{23873, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{10597, {{23874, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10598, {{23875, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10600, {{23876, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10601, {{23877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10602, {{23878, 1, 0, 0},{23878, 1, 1, 17}, {0, 0, 0, 0}}},{10602, {{23878, 1, 0, 0},{23878, 1, 1, 17}, {0, 0, 0, 0}}},{57388, {{23878, 1, 0, 0},{23878, 1, 1, 18}, {0, 0, 0, 0}}},{57388, {{23878, 1, 0, 0},{23878, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10603, {{23879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10604, {{23880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10605, {{23881, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10606, {{23882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10607, {{23883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10608, {{23884, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10609, {{23885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10610, {{23886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10611, {{23887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10612, {{23888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10613, {{23889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10614, {{23890, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10615, {{23891, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10616, {{23892, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10617, {{23893, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10618, {{23894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10619, {{23895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10620, {{23897, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10621, {{23900, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10622, {{23902, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10623, {{23903, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10624, {{23904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10625, {{23905, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10626, {{23906, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10627, {{23907, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10628, {{23908, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10629, {{23909, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10630, {{23910, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10631, {{23911, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10632, {{23912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10633, {{23913, 1, 0, 0},{23913, 1, 1, 17}, {0, 0, 0, 0}}},{10633, {{23913, 1, 0, 0},{23913, 1, 1, 17}, {0, 0, 0, 0}}},{10634, {{23913, 1, 0, 0},{23913, 1, 1, 18}, {0, 0, 0, 0}}},{10634, {{23913, 1, 0, 0},{23913, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10635, {{23914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10636, {{23915, 1, 0, 0},{23915, 1, 1, 16}, {0, 0, 0, 0}}},{10636, {{23915, 1, 0, 0},{23915, 1, 1, 16}, {0, 0, 0, 0}}},{10637, {{23915, 1, 0, 0},{23915, 1, 1, 17}, {0, 0, 0, 0}}},{10637, {{23915, 1, 0, 0},{23915, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{10638, {{23916, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10639, {{23917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10640, {{23918, 1, 0, 0},{23918, 1, 1, 17}, {0, 0, 0, 0}}},{10640, {{23918, 1, 0, 0},{23918, 1, 1, 17}, {0, 0, 0, 0}}},{10641, {{23918, 1, 0, 0},{23918, 1, 1, 16}, {0, 0, 0, 0}}},{10641, {{23918, 1, 0, 0},{23918, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{10642, {{23919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10643, {{23920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10645, {{23921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10646, {{23922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10647, {{23923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10648, {{23924, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{10649, {{23926, 1, 0, 0},{23926, 1, 1, 17}, {0, 0, 0, 0}}},{10649, {{23926, 1, 0, 0},{23926, 1, 1, 17}, {0, 0, 0, 0}}},{10650, {{23926, 1, 0, 0},{23926, 1, 1, 18}, {0, 0, 0, 0}}},{10650, {{23926, 1, 0, 0},{23926, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10651, {{23927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10652, {{23928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10653, {{23929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10654, {{23930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10655, {{23931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10656, {{23932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10657, {{23933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10658, {{23934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10659, {{23935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10660, {{23936, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10661, {{23937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10662, {{23938, 1, 0, 0},{23938, 1, 1, 17}, {0, 0, 0, 0}}},{10662, {{23938, 1, 0, 0},{23938, 1, 1, 17}, {0, 0, 0, 0}}},{10663, {{23938, 1, 0, 0},{23938, 1, 1, 18}, {0, 0, 0, 0}}},{10663, {{23938, 1, 0, 0},{23938, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10664, {{23939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10665, {{23940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10666, {{23941, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10667, {{23942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10668, {{23943, 1, 0, 0},{23943, 1, 1, 18}, {0, 0, 0, 0}}},{10668, {{23943, 1, 0, 0},{23943, 1, 1, 18}, {0, 0, 0, 0}}},{10669, {{23943, 1, 0, 0},{23943, 1, 1, 19}, {0, 0, 0, 0}}},{10669, {{23943, 1, 0, 0},{23943, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{10670, {{23944, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10671, {{23945, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10672, {{23946, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10673, {{23947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10674, {{23948, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10675, {{23949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10676, {{23951, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10677, {{23952, 1, 0, 0},{23952, 1, 1, 17}, {0, 0, 0, 0}}},{10677, {{23952, 1, 0, 0},{23952, 1, 1, 17}, {0, 0, 0, 0}}},{10678, {{23952, 1, 0, 0},{23952, 1, 1, 18}, {0, 0, 0, 0}}},{10678, {{23952, 1, 0, 0},{23952, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10679, {{23953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10680, {{23954, 1, 0, 0},{23954, 1, 1, 17}, {0, 0, 0, 0}}},{10680, {{23954, 1, 0, 0},{23954, 1, 1, 17}, {0, 0, 0, 0}}},{10681, {{23954, 1, 0, 0},{23954, 1, 1, 18}, {0, 0, 0, 0}}},{10681, {{23954, 1, 0, 0},{23954, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10682, {{23955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10683, {{23956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10684, {{23957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10685, {{23958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10686, {{23959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10687, {{23960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10688, {{23961, 1, 0, 0},{23961, 1, 1, 18}, {0, 0, 0, 0}}},{10688, {{23961, 1, 0, 0},{23961, 1, 1, 18}, {0, 0, 0, 0}}},{10689, {{23961, 1, 0, 0},{23961, 1, 1, 17}, {0, 0, 0, 0}}},{10689, {{23961, 1, 0, 0},{23961, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{10690, {{23962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10691, {{23963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10692, {{23964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10693, {{23965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10694, {{23966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10695, {{23967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10696, {{23968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10697, {{23969, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10698, {{23970, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10699, {{23971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10700, {{23972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10701, {{23974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10702, {{23975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10703, {{23976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10704, {{23977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10705, {{23978, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10706, {{23979, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10707, {{23980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10708, {{23981, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10709, {{23982, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10710, {{23983, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10711, {{23984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10712, {{23985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10713, {{23986, 1, 0, 0},{23986, 1, 1, 19}, {0, 0, 0, 0}}},{10713, {{23986, 1, 0, 0},{23986, 1, 1, 19}, {0, 0, 0, 0}}},{10714, {{23986, 1, 0, 0},{23986, 1, 1, 18}, {0, 0, 0, 0}}},{10714, {{23986, 1, 0, 0},{23986, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10715, {{23987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10716, {{23988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10717, {{23989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10718, {{23991, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10719, {{23992, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10720, {{23993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10721, {{23994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10722, {{23995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10723, {{23996, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10724, {{23997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10725, {{23998, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10726, {{24000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10727, {{24001, 1, 0, 0},{24001, 1, 1, 17}, {0, 0, 0, 0}}},{10727, {{24001, 1, 0, 0},{24001, 1, 1, 17}, {0, 0, 0, 0}}},{10728, {{24001, 1, 0, 0},{24001, 1, 1, 16}, {0, 0, 0, 0}}},{10728, {{24001, 1, 0, 0},{24001, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{10729, {{24002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{10730, {{24003, 1, 0, 0},{24003, 1, 1, 17}, {0, 0, 0, 0}}},{10730, {{24003, 1, 0, 0},{24003, 1, 1, 17}, {0, 0, 0, 0}}},{10731, {{24003, 1, 0, 0},{24003, 1, 1, 19}, {0, 0, 0, 0}}},{10731, {{24003, 1, 0, 0},{24003, 1, 1, 19}, {0, 0, 0, 0}}},{10732, {{24003, 1, 0, 0},{24003, 1, 1, 18}, {0, 0, 0, 0}}},{10732, {{24003, 1, 0, 0},{24003, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10733, {{24004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10734, {{24006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10735, {{24007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10736, {{24008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10737, {{24009, 1, 0, 0},{24009, 1, 1, 18}, {0, 0, 0, 0}}},{10737, {{24009, 1, 0, 0},{24009, 1, 1, 18}, {0, 0, 0, 0}}},{10738, {{24009, 1, 0, 0},{24009, 1, 1, 19}, {0, 0, 0, 0}}},{10738, {{24009, 1, 0, 0},{24009, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{10739, {{24010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10740, {{24011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10741, {{24012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{10742, {{24013, 1, 0, 0},{24013, 1, 1, 17}, {0, 0, 0, 0}}},{10742, {{24013, 1, 0, 0},{24013, 1, 1, 17}, {0, 0, 0, 0}}},{10743, {{24013, 1, 0, 0},{24013, 1, 1, 18}, {0, 0, 0, 0}}},{10743, {{24013, 1, 0, 0},{24013, 1, 1, 18}, {0, 0, 0, 0}}},{10744, {{24013, 1, 0, 0},{24013, 1, 1, 19}, {0, 0, 0, 0}}},{10744, {{24013, 1, 0, 0},{24013, 1, 1, 19}, {0, 0, 0, 0}}},{68060, {{24013, 1, 0, 0},{24013, 1, 1, 20}, {0, 0, 0, 0}}},{68060, {{24013, 1, 0, 0},{24013, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{10745, {{24014, 1, 0, 0},{24014, 1, 1, 18}, {0, 0, 0, 0}}},{10745, {{24014, 1, 0, 0},{24014, 1, 1, 18}, {0, 0, 0, 0}}},{10746, {{24014, 1, 0, 0},{24014, 1, 1, 17}, {0, 0, 0, 0}}},{10746, {{24014, 1, 0, 0},{24014, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{10747, {{24015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10748, {{24016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10749, {{24017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10750, {{24018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{10751, {{24019, 1, 0, 0},{24019, 1, 1, 18}, {0, 0, 0, 0}}},{10751, {{24019, 1, 0, 0},{24019, 1, 1, 18}, {0, 0, 0, 0}}},{10752, {{24019, 1, 0, 0},{24019, 1, 1, 20}, {0, 0, 0, 0}}},{10752, {{24019, 1, 0, 0},{24019, 1, 1, 20}, {0, 0, 0, 0}}},{10753, {{24019, 1, 0, 0},{24019, 1, 1, 19}, {0, 0, 0, 0}}},{10753, {{24019, 1, 0, 0},{24019, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{10754, {{24022, 1, 0, 0},{24022, 1, 1, 17}, {0, 0, 0, 0}}},{10754, {{24022, 1, 0, 0},{24022, 1, 1, 17}, {0, 0, 0, 0}}},{59537, {{24022, 1, 0, 0},{24022, 1, 1, 18}, {0, 0, 0, 0}}},{59537, {{24022, 1, 0, 0},{24022, 1, 1, 18}, {0, 0, 0, 0}}},{59539, {{24022, 1, 0, 0},{24022, 1, 1, 19}, {0, 0, 0, 0}}},{59539, {{24022, 1, 0, 0},{24022, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{10755, {{24023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10756, {{24024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{10757, {{24025, 1, 0, 0},{24025, 1, 1, 18}, {0, 0, 0, 0}}},{10757, {{24025, 1, 0, 0},{24025, 1, 1, 18}, {0, 0, 0, 0}}},{10758, {{24025, 1, 0, 0},{24025, 1, 1, 19}, {0, 0, 0, 0}}},{10758, {{24025, 1, 0, 0},{24025, 1, 1, 19}, {0, 0, 0, 0}}},{10759, {{24025, 1, 0, 0},{24025, 1, 1, 17}, {0, 0, 0, 0}}},{10759, {{24025, 1, 0, 0},{24025, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{10760, {{24026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10761, {{24027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10762, {{24028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10763, {{24029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10764, {{24030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{34906, {{24031, 1, 0, 0},{24031, 1, 1, 16}, {0, 0, 0, 0}}},{34906, {{24031, 1, 0, 0},{24031, 1, 1, 16}, {0, 0, 0, 0}}},{34907, {{24031, 1, 0, 0},{24031, 1, 1, 17}, {0, 0, 0, 0}}},{34907, {{24031, 1, 0, 0},{24031, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{10766, {{24032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10767, {{24033, 1, 0, 0},{24033, 1, 1, 18}, {0, 0, 0, 0}}},{10767, {{24033, 1, 0, 0},{24033, 1, 1, 18}, {0, 0, 0, 0}}},{10768, {{24033, 1, 0, 0},{24033, 1, 1, 19}, {0, 0, 0, 0}}},{10768, {{24033, 1, 0, 0},{24033, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {6, {{10769, {{24034, 1, 0, 0},{24034, 1, 1, 17}, {0, 0, 0, 0}}},{10769, {{24034, 1, 0, 0},{24034, 1, 1, 17}, {0, 0, 0, 0}}},{10770, {{24034, 1, 0, 0},{24034, 1, 1, 19}, {0, 0, 0, 0}}},{10770, {{24034, 1, 0, 0},{24034, 1, 1, 19}, {0, 0, 0, 0}}},{57404, {{24034, 1, 0, 0},{24034, 1, 1, 18}, {0, 0, 0, 0}}},{57404, {{24034, 1, 0, 0},{24034, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10771, {{24035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10772, {{24036, 1, 0, 0},{24036, 1, 1, 17}, {0, 0, 0, 0}}},{10772, {{24036, 1, 0, 0},{24036, 1, 1, 17}, {0, 0, 0, 0}}},{57405, {{24036, 1, 0, 0},{24036, 1, 1, 18}, {0, 0, 0, 0}}},{57405, {{24036, 1, 0, 0},{24036, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10773, {{24037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10774, {{24038, 1, 0, 0},{24038, 1, 1, 17}, {0, 0, 0, 0}}},{10774, {{24038, 1, 0, 0},{24038, 1, 1, 17}, {0, 0, 0, 0}}},{57406, {{24038, 1, 0, 0},{24038, 1, 1, 18}, {0, 0, 0, 0}}},{57406, {{24038, 1, 0, 0},{24038, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10775, {{24039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10776, {{24040, 1, 0, 0},{24040, 1, 1, 19}, {0, 0, 0, 0}}},{10776, {{24040, 1, 0, 0},{24040, 1, 1, 19}, {0, 0, 0, 0}}},{10777, {{24040, 1, 0, 0},{24040, 1, 1, 18}, {0, 0, 0, 0}}},{10777, {{24040, 1, 0, 0},{24040, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{10778, {{24041, 1, 0, 0},{24041, 1, 1, 17}, {0, 0, 0, 0}}},{10778, {{24041, 1, 0, 0},{24041, 1, 1, 17}, {0, 0, 0, 0}}},{10779, {{24041, 1, 0, 0},{24041, 1, 1, 18}, {0, 0, 0, 0}}},{10779, {{24041, 1, 0, 0},{24041, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{68031, {{24042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10780, {{24043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{10781, {{24046, 1, 0, 0},{24046, 1, 1, 17}, {0, 0, 0, 0}}},{10781, {{24046, 1, 0, 0},{24046, 1, 1, 17}, {0, 0, 0, 0}}},{57408, {{24046, 1, 0, 0},{24046, 1, 1, 18}, {0, 0, 0, 0}}},{57408, {{24046, 1, 0, 0},{24046, 1, 1, 18}, {0, 0, 0, 0}}},{57409, {{24046, 1, 0, 0},{24046, 1, 1, 19}, {0, 0, 0, 0}}},{57409, {{24046, 1, 0, 0},{24046, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10782, {{24049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10784, {{24050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10785, {{24051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10786, {{24052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10787, {{24053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10788, {{24054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10789, {{24055, 1, 0, 0},{24055, 1, 1, 19}, {0, 0, 0, 0}}},{10789, {{24055, 1, 0, 0},{24055, 1, 1, 19}, {0, 0, 0, 0}}},{10790, {{24055, 1, 0, 0},{24055, 1, 1, 18}, {0, 0, 0, 0}}},{10790, {{24055, 1, 0, 0},{24055, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{10791, {{24056, 1, 0, 0},{24056, 1, 1, 17}, {0, 0, 0, 0}}},{10791, {{24056, 1, 0, 0},{24056, 1, 1, 17}, {0, 0, 0, 0}}},{10792, {{24056, 1, 0, 0},{24056, 1, 1, 18}, {0, 0, 0, 0}}},{10792, {{24056, 1, 0, 0},{24056, 1, 1, 18}, {0, 0, 0, 0}}},{10794, {{24056, 1, 0, 0},{24056, 1, 1, 19}, {0, 0, 0, 0}}},{10794, {{24056, 1, 0, 0},{24056, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{10796, {{24057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10797, {{24058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{10798, {{24059, 1, 0, 0},{24059, 1, 1, 17}, {0, 0, 0, 0}}},{10798, {{24059, 1, 0, 0},{24059, 1, 1, 17}, {0, 0, 0, 0}}},{10799, {{24059, 1, 0, 0},{24059, 1, 1, 19}, {0, 0, 0, 0}}},{10799, {{24059, 1, 0, 0},{24059, 1, 1, 19}, {0, 0, 0, 0}}},{59540, {{24059, 1, 0, 0},{24059, 1, 1, 18}, {0, 0, 0, 0}}},{59540, {{24059, 1, 0, 0},{24059, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{68032, {{24060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{10800, {{24061, 1, 0, 0},{24061, 1, 1, 19}, {0, 0, 0, 0}}},{10800, {{24061, 1, 0, 0},{24061, 1, 1, 19}, {0, 0, 0, 0}}},{10801, {{24061, 1, 0, 0},{24061, 1, 1, 18}, {0, 0, 0, 0}}},{10801, {{24061, 1, 0, 0},{24061, 1, 1, 18}, {0, 0, 0, 0}}},{10802, {{24061, 1, 0, 0},{24061, 1, 1, 20}, {0, 0, 0, 0}}},{10802, {{24061, 1, 0, 0},{24061, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{10803, {{24062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10804, {{24063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10805, {{24064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10806, {{24065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10807, {{24066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10808, {{24067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10809, {{24068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57417, {{24069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10810, {{24070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10811, {{24071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10812, {{24072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10813, {{24073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10815, {{24074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10816, {{24075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10817, {{24076, 1, 0, 0},{24076, 1, 1, 17}, {0, 0, 0, 0}}},{10817, {{24076, 1, 0, 0},{24076, 1, 1, 17}, {0, 0, 0, 0}}},{57418, {{24076, 1, 0, 0},{24076, 1, 1, 18}, {0, 0, 0, 0}}},{57418, {{24076, 1, 0, 0},{24076, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10818, {{24077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10819, {{24078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10820, {{24081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10821, {{24082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10822, {{24083, 1, 0, 0},{24083, 1, 1, 16}, {0, 0, 0, 0}}},{10822, {{24083, 1, 0, 0},{24083, 1, 1, 16}, {0, 0, 0, 0}}},{10823, {{24083, 1, 0, 0},{24083, 1, 1, 17}, {0, 0, 0, 0}}},{10823, {{24083, 1, 0, 0},{24083, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{10824, {{24084, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10825, {{24085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10826, {{24086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10827, {{24087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10828, {{24088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10829, {{24089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10830, {{24090, 1, 0, 0},{24090, 1, 1, 17}, {0, 0, 0, 0}}},{10830, {{24090, 1, 0, 0},{24090, 1, 1, 17}, {0, 0, 0, 0}}},{10831, {{24090, 1, 0, 0},{24090, 1, 1, 18}, {0, 0, 0, 0}}},{10831, {{24090, 1, 0, 0},{24090, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10832, {{24091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{10833, {{24093, 1, 0, 0},{24093, 1, 1, 18}, {0, 0, 0, 0}}},{10833, {{24093, 1, 0, 0},{24093, 1, 1, 18}, {0, 0, 0, 0}}},{10834, {{24093, 1, 0, 0},{24093, 1, 1, 19}, {0, 0, 0, 0}}},{10834, {{24093, 1, 0, 0},{24093, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{10835, {{24094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10836, {{24095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10837, {{24096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10838, {{24097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10839, {{24098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10840, {{24099, 1, 0, 0},{24099, 1, 1, 16}, {0, 0, 0, 0}}},{10840, {{24099, 1, 0, 0},{24099, 1, 1, 16}, {0, 0, 0, 0}}},{10841, {{24099, 1, 0, 0},{24099, 1, 1, 17}, {0, 0, 0, 0}}},{10841, {{24099, 1, 0, 0},{24099, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{10842, {{24100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10843, {{24101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57423, {{24102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{10844, {{24104, 1, 0, 0},{24104, 1, 1, 17}, {0, 0, 0, 0}}},{10844, {{24104, 1, 0, 0},{24104, 1, 1, 17}, {0, 0, 0, 0}}},{10845, {{24104, 1, 0, 0},{24104, 1, 1, 18}, {0, 0, 0, 0}}},{10845, {{24104, 1, 0, 0},{24104, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10846, {{24105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10847, {{24106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{10848, {{24107, 1, 0, 0},{24107, 1, 1, 17}, {0, 0, 0, 0}}},{10848, {{24107, 1, 0, 0},{24107, 1, 1, 17}, {0, 0, 0, 0}}},{57420, {{24107, 1, 0, 0},{24107, 1, 1, 19}, {0, 0, 0, 0}}},{57420, {{24107, 1, 0, 0},{24107, 1, 1, 19}, {0, 0, 0, 0}}},{57421, {{24107, 1, 0, 0},{24107, 1, 1, 18}, {0, 0, 0, 0}}},{57421, {{24107, 1, 0, 0},{24107, 1, 1, 18}, {0, 0, 0, 0}}},{57424, {{24107, 1, 0, 0},{24107, 1, 1, 20}, {0, 0, 0, 0}}},{57424, {{24107, 1, 0, 0},{24107, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{10849, {{24108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10850, {{24109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10852, {{24110, 1, 0, 0},{24110, 1, 1, 17}, {0, 0, 0, 0}}},{10852, {{24110, 1, 0, 0},{24110, 1, 1, 17}, {0, 0, 0, 0}}},{10853, {{24110, 1, 0, 0},{24110, 1, 1, 18}, {0, 0, 0, 0}}},{10853, {{24110, 1, 0, 0},{24110, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{10854, {{24111, 1, 0, 0},{24111, 1, 1, 17}, {0, 0, 0, 0}}},{10854, {{24111, 1, 0, 0},{24111, 1, 1, 17}, {0, 0, 0, 0}}},{57426, {{24111, 1, 0, 0},{24111, 1, 1, 18}, {0, 0, 0, 0}}},{57426, {{24111, 1, 0, 0},{24111, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{10855, {{24112, 1, 0, 0},{24112, 1, 1, 18}, {0, 0, 0, 0}}},{10855, {{24112, 1, 0, 0},{24112, 1, 1, 18}, {0, 0, 0, 0}}},{10856, {{24112, 1, 0, 0},{24112, 1, 1, 19}, {0, 0, 0, 0}}},{10856, {{24112, 1, 0, 0},{24112, 1, 1, 19}, {0, 0, 0, 0}}},{10857, {{24112, 1, 0, 0},{24112, 1, 1, 20}, {0, 0, 0, 0}}},{10857, {{24112, 1, 0, 0},{24112, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10858, {{24114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10859, {{24115, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10860, {{24116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10861, {{24117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10862, {{24118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10863, {{24119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10864, {{24120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10865, {{24121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10866, {{24122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{10867, {{24125, 1, 0, 0},{24125, 1, 1, 19}, {0, 0, 0, 0}}},{10867, {{24125, 1, 0, 0},{24125, 1, 1, 19}, {0, 0, 0, 0}}},{10868, {{24125, 1, 0, 0},{24125, 1, 1, 18}, {0, 0, 0, 0}}},{10868, {{24125, 1, 0, 0},{24125, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{10869, {{24126, 1, 0, 0},{24126, 1, 1, 17}, {0, 0, 0, 0}}},{10869, {{24126, 1, 0, 0},{24126, 1, 1, 17}, {0, 0, 0, 0}}},{10870, {{24126, 1, 0, 0},{24126, 1, 1, 18}, {0, 0, 0, 0}}},{10870, {{24126, 1, 0, 0},{24126, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10871, {{24127, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10872, {{24128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10873, {{24129, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10874, {{24130, 1, 0, 0},{24130, 1, 1, 16}, {0, 0, 0, 0}}},{10874, {{24130, 1, 0, 0},{24130, 1, 1, 16}, {0, 0, 0, 0}}},{10875, {{24130, 1, 0, 0},{24130, 1, 1, 17}, {0, 0, 0, 0}}},{10875, {{24130, 1, 0, 0},{24130, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {6, {{10876, {{24131, 1, 0, 0},{24131, 1, 1, 18}, {0, 0, 0, 0}}},{10876, {{24131, 1, 0, 0},{24131, 1, 1, 18}, {0, 0, 0, 0}}},{10877, {{24131, 1, 0, 0},{24131, 1, 1, 19}, {0, 0, 0, 0}}},{10877, {{24131, 1, 0, 0},{24131, 1, 1, 19}, {0, 0, 0, 0}}},{10878, {{24131, 1, 0, 0},{24131, 1, 1, 20}, {0, 0, 0, 0}}},{10878, {{24131, 1, 0, 0},{24131, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{10879, {{24132, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10880, {{24133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10881, {{24134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10882, {{24135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10883, {{24136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10884, {{24137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10885, {{24138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10886, {{24139, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10887, {{24140, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10888, {{24141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10889, {{24142, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10890, {{24143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10891, {{24144, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10892, {{24145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10893, {{24146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10894, {{24147, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10895, {{24148, 1, 0, 0},{24148, 1, 1, 19}, {0, 0, 0, 0}}},{10895, {{24148, 1, 0, 0},{24148, 1, 1, 19}, {0, 0, 0, 0}}},{10896, {{24148, 1, 0, 0},{24148, 1, 1, 18}, {0, 0, 0, 0}}},{10896, {{24148, 1, 0, 0},{24148, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{10897, {{24149, 1, 0, 0},{24149, 1, 1, 17}, {0, 0, 0, 0}}},{10897, {{24149, 1, 0, 0},{24149, 1, 1, 17}, {0, 0, 0, 0}}},{10898, {{24149, 1, 0, 0},{24149, 1, 1, 18}, {0, 0, 0, 0}}},{10898, {{24149, 1, 0, 0},{24149, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10899, {{24150, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10900, {{24151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10901, {{24152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10902, {{24153, 1, 0, 0},{24153, 1, 1, 17}, {0, 0, 0, 0}}},{10902, {{24153, 1, 0, 0},{24153, 1, 1, 17}, {0, 0, 0, 0}}},{10903, {{24153, 1, 0, 0},{24153, 1, 1, 16}, {0, 0, 0, 0}}},{10903, {{24153, 1, 0, 0},{24153, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{10904, {{24154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10905, {{24155, 1, 0, 0},{24155, 1, 1, 17}, {0, 0, 0, 0}}},{10905, {{24155, 1, 0, 0},{24155, 1, 1, 17}, {0, 0, 0, 0}}},{10906, {{24155, 1, 0, 0},{24155, 1, 1, 18}, {0, 0, 0, 0}}},{10906, {{24155, 1, 0, 0},{24155, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10907, {{24156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10908, {{24157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10909, {{24158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10910, {{24159, 1, 0, 0},{24159, 1, 1, 17}, {0, 0, 0, 0}}},{10910, {{24159, 1, 0, 0},{24159, 1, 1, 17}, {0, 0, 0, 0}}},{10911, {{24159, 1, 0, 0},{24159, 1, 1, 18}, {0, 0, 0, 0}}},{10911, {{24159, 1, 0, 0},{24159, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10912, {{24160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10913, {{24161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10914, {{24162, 1, 0, 0},{24162, 1, 1, 17}, {0, 0, 0, 0}}},{10914, {{24162, 1, 0, 0},{24162, 1, 1, 17}, {0, 0, 0, 0}}},{10915, {{24162, 1, 0, 0},{24162, 1, 1, 18}, {0, 0, 0, 0}}},{10915, {{24162, 1, 0, 0},{24162, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{10916, {{24163, 1, 0, 0},{24163, 1, 1, 19}, {0, 0, 0, 0}}},{10916, {{24163, 1, 0, 0},{24163, 1, 1, 19}, {0, 0, 0, 0}}},{10917, {{24163, 1, 0, 0},{24163, 1, 1, 18}, {0, 0, 0, 0}}},{10917, {{24163, 1, 0, 0},{24163, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{10918, {{24164, 1, 0, 0},{24164, 1, 1, 18}, {0, 0, 0, 0}}},{10918, {{24164, 1, 0, 0},{24164, 1, 1, 18}, {0, 0, 0, 0}}},{10919, {{24164, 1, 0, 0},{24164, 1, 1, 19}, {0, 0, 0, 0}}},{10919, {{24164, 1, 0, 0},{24164, 1, 1, 19}, {0, 0, 0, 0}}},{59550, {{24164, 1, 0, 0},{24164, 1, 1, 20}, {0, 0, 0, 0}}},{59550, {{24164, 1, 0, 0},{24164, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10920, {{24166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10921, {{24167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10922, {{24168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10923, {{24169, 1, 0, 0},{24169, 1, 1, 17}, {0, 0, 0, 0}}},{10923, {{24169, 1, 0, 0},{24169, 1, 1, 17}, {0, 0, 0, 0}}},{10924, {{24169, 1, 0, 0},{24169, 1, 1, 16}, {0, 0, 0, 0}}},{10924, {{24169, 1, 0, 0},{24169, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {6, {{10925, {{24170, 1, 0, 0},{24170, 1, 1, 19}, {0, 0, 0, 0}}},{10925, {{24170, 1, 0, 0},{24170, 1, 1, 19}, {0, 0, 0, 0}}},{10926, {{24170, 1, 0, 0},{24170, 1, 1, 17}, {0, 0, 0, 0}}},{10926, {{24170, 1, 0, 0},{24170, 1, 1, 17}, {0, 0, 0, 0}}},{10927, {{24170, 1, 0, 0},{24170, 1, 1, 18}, {0, 0, 0, 0}}},{10927, {{24170, 1, 0, 0},{24170, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10928, {{24171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10929, {{24172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{10930, {{24173, 1, 0, 0},{24173, 1, 1, 19}, {0, 0, 0, 0}}},{10930, {{24173, 1, 0, 0},{24173, 1, 1, 19}, {0, 0, 0, 0}}},{10931, {{24173, 1, 0, 0},{24173, 1, 1, 18}, {0, 0, 0, 0}}},{10931, {{24173, 1, 0, 0},{24173, 1, 1, 18}, {0, 0, 0, 0}}},{10932, {{24173, 1, 0, 0},{24173, 1, 1, 17}, {0, 0, 0, 0}}},{10932, {{24173, 1, 0, 0},{24173, 1, 1, 17}, {0, 0, 0, 0}}},{10933, {{24173, 1, 0, 0},{24173, 1, 1, 20}, {0, 0, 0, 0}}},{10933, {{24173, 1, 0, 0},{24173, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{10934, {{24174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10935, {{24175, 1, 0, 0},{24175, 1, 1, 16}, {0, 0, 0, 0}}},{10935, {{24175, 1, 0, 0},{24175, 1, 1, 16}, {0, 0, 0, 0}}},{10936, {{24175, 1, 0, 0},{24175, 1, 1, 17}, {0, 0, 0, 0}}},{10936, {{24175, 1, 0, 0},{24175, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{10938, {{24176, 1, 0, 0},{24176, 1, 1, 17}, {0, 0, 0, 0}}},{10938, {{24176, 1, 0, 0},{24176, 1, 1, 17}, {0, 0, 0, 0}}},{10939, {{24176, 1, 0, 0},{24176, 1, 1, 18}, {0, 0, 0, 0}}},{10939, {{24176, 1, 0, 0},{24176, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10940, {{24177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10941, {{24178, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10942, {{24179, 1, 0, 0},{24179, 1, 1, 19}, {0, 0, 0, 0}}},{10942, {{24179, 1, 0, 0},{24179, 1, 1, 19}, {0, 0, 0, 0}}},{10943, {{24179, 1, 0, 0},{24179, 1, 1, 18}, {0, 0, 0, 0}}},{10943, {{24179, 1, 0, 0},{24179, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{10944, {{24180, 1, 0, 0},{24180, 1, 1, 17}, {0, 0, 0, 0}}},{10944, {{24180, 1, 0, 0},{24180, 1, 1, 17}, {0, 0, 0, 0}}},{10945, {{24180, 1, 0, 0},{24180, 1, 1, 18}, {0, 0, 0, 0}}},{10945, {{24180, 1, 0, 0},{24180, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{10946, {{24181, 1, 0, 0},{24181, 1, 1, 17}, {0, 0, 0, 0}}},{10946, {{24181, 1, 0, 0},{24181, 1, 1, 17}, {0, 0, 0, 0}}},{10947, {{24181, 1, 0, 0},{24181, 1, 1, 18}, {0, 0, 0, 0}}},{10947, {{24181, 1, 0, 0},{24181, 1, 1, 18}, {0, 0, 0, 0}}},{10948, {{24181, 1, 0, 0},{24181, 1, 1, 19}, {0, 0, 0, 0}}},{10948, {{24181, 1, 0, 0},{24181, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{10949, {{24182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10950, {{24183, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10951, {{24184, 1, 0, 0},{24184, 1, 1, 17}, {0, 0, 0, 0}}},{10951, {{24184, 1, 0, 0},{24184, 1, 1, 17}, {0, 0, 0, 0}}},{57432, {{24184, 1, 0, 0},{24184, 1, 1, 18}, {0, 0, 0, 0}}},{57432, {{24184, 1, 0, 0},{24184, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10953, {{24185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10954, {{24186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10955, {{24187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10956, {{24188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10957, {{24189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{10958, {{24190, 1, 0, 0},{24190, 1, 1, 18}, {0, 0, 0, 0}}},{10958, {{24190, 1, 0, 0},{24190, 1, 1, 18}, {0, 0, 0, 0}}},{57436, {{24190, 1, 0, 0},{24190, 1, 1, 19}, {0, 0, 0, 0}}},{57436, {{24190, 1, 0, 0},{24190, 1, 1, 19}, {0, 0, 0, 0}}},{57437, {{24190, 1, 0, 0},{24190, 1, 1, 20}, {0, 0, 0, 0}}},{57437, {{24190, 1, 0, 0},{24190, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{10959, {{24191, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10960, {{24192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10961, {{24193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10962, {{24194, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10963, {{24195, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10964, {{24196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10966, {{24199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10967, {{24200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10968, {{24201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10969, {{24202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10970, {{24203, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10971, {{24204, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10972, {{24205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10973, {{24206, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10974, {{24207, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10975, {{24208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10976, {{24210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{10979, {{24213, 1, 0, 0},{24213, 1, 1, 17}, {0, 0, 0, 0}}},{10979, {{24213, 1, 0, 0},{24213, 1, 1, 17}, {0, 0, 0, 0}}},{57440, {{24213, 1, 0, 0},{24213, 1, 1, 18}, {0, 0, 0, 0}}},{57440, {{24213, 1, 0, 0},{24213, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{10980, {{24214, 1, 0, 0},{24214, 1, 1, 19}, {0, 0, 0, 0}}},{10980, {{24214, 1, 0, 0},{24214, 1, 1, 19}, {0, 0, 0, 0}}},{10981, {{24214, 1, 0, 0},{24214, 1, 1, 18}, {0, 0, 0, 0}}},{10981, {{24214, 1, 0, 0},{24214, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10982, {{24215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10983, {{24216, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10984, {{24217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10985, {{24218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10986, {{24219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10987, {{24220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10988, {{24221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10990, {{24224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10991, {{24225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10992, {{24226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10993, {{24227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10994, {{24228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10995, {{24229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{10996, {{24230, 1, 0, 0},{24230, 1, 1, 17}, {0, 0, 0, 0}}},{10996, {{24230, 1, 0, 0},{24230, 1, 1, 17}, {0, 0, 0, 0}}},{57439, {{24230, 1, 0, 0},{24230, 1, 1, 18}, {0, 0, 0, 0}}},{57439, {{24230, 1, 0, 0},{24230, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{10997, {{24231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10998, {{24232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10999, {{24233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11000, {{24234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11001, {{24235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{11002, {{24236, 1, 0, 0},{24236, 1, 1, 17}, {0, 0, 0, 0}}},{11002, {{24236, 1, 0, 0},{24236, 1, 1, 17}, {0, 0, 0, 0}}},{11003, {{24236, 1, 0, 0},{24236, 1, 1, 18}, {0, 0, 0, 0}}},{11003, {{24236, 1, 0, 0},{24236, 1, 1, 18}, {0, 0, 0, 0}}},{11004, {{24236, 1, 0, 0},{24236, 1, 1, 19}, {0, 0, 0, 0}}},{11004, {{24236, 1, 0, 0},{24236, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {6, {{11005, {{24237, 1, 0, 0},{24237, 1, 1, 20}, {0, 0, 0, 0}}},{11005, {{24237, 1, 0, 0},{24237, 1, 1, 20}, {0, 0, 0, 0}}},{11006, {{24237, 1, 0, 0},{24237, 1, 1, 18}, {0, 0, 0, 0}}},{11006, {{24237, 1, 0, 0},{24237, 1, 1, 18}, {0, 0, 0, 0}}},{57441, {{24237, 1, 0, 0},{24237, 1, 1, 19}, {0, 0, 0, 0}}},{57441, {{24237, 1, 0, 0},{24237, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{11007, {{24238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11008, {{24239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11009, {{24240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11010, {{24241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11011, {{24242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11012, {{24243, 1, 0, 0},{24243, 1, 1, 17}, {0, 0, 0, 0}}},{11012, {{24243, 1, 0, 0},{24243, 1, 1, 17}, {0, 0, 0, 0}}},{57442, {{24243, 1, 0, 0},{24243, 1, 1, 18}, {0, 0, 0, 0}}},{57442, {{24243, 1, 0, 0},{24243, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11013, {{24244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11014, {{24245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11015, {{24246, 1, 0, 0},{24246, 1, 1, 17}, {0, 0, 0, 0}}},{11015, {{24246, 1, 0, 0},{24246, 1, 1, 17}, {0, 0, 0, 0}}},{57448, {{24246, 1, 0, 0},{24246, 1, 1, 18}, {0, 0, 0, 0}}},{57448, {{24246, 1, 0, 0},{24246, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{11016, {{24247, 1, 0, 0},{24247, 1, 1, 18}, {0, 0, 0, 0}}},{11016, {{24247, 1, 0, 0},{24247, 1, 1, 18}, {0, 0, 0, 0}}},{11017, {{24247, 1, 0, 0},{24247, 1, 1, 17}, {0, 0, 0, 0}}},{11017, {{24247, 1, 0, 0},{24247, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{11018, {{24248, 1, 0, 0},{24248, 1, 1, 17}, {0, 0, 0, 0}}},{11018, {{24248, 1, 0, 0},{24248, 1, 1, 17}, {0, 0, 0, 0}}},{59558, {{24248, 1, 0, 0},{24248, 1, 1, 18}, {0, 0, 0, 0}}},{59558, {{24248, 1, 0, 0},{24248, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{11019, {{24249, 1, 0, 0},{24249, 1, 1, 18}, {0, 0, 0, 0}}},{11019, {{24249, 1, 0, 0},{24249, 1, 1, 18}, {0, 0, 0, 0}}},{11020, {{24249, 1, 0, 0},{24249, 1, 1, 17}, {0, 0, 0, 0}}},{11020, {{24249, 1, 0, 0},{24249, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{11021, {{24250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11022, {{24251, 1, 0, 0},{24251, 1, 1, 17}, {0, 0, 0, 0}}},{11022, {{24251, 1, 0, 0},{24251, 1, 1, 17}, {0, 0, 0, 0}}},{11023, {{24251, 1, 0, 0},{24251, 1, 1, 16}, {0, 0, 0, 0}}},{11023, {{24251, 1, 0, 0},{24251, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{11024, {{24253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11025, {{24254, 1, 0, 0},{24254, 1, 1, 17}, {0, 0, 0, 0}}},{11025, {{24254, 1, 0, 0},{24254, 1, 1, 17}, {0, 0, 0, 0}}},{11026, {{24254, 1, 0, 0},{24254, 1, 1, 18}, {0, 0, 0, 0}}},{11026, {{24254, 1, 0, 0},{24254, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{11027, {{24255, 1, 0, 0},{24255, 1, 1, 17}, {0, 0, 0, 0}}},{11027, {{24255, 1, 0, 0},{24255, 1, 1, 17}, {0, 0, 0, 0}}},{11028, {{24255, 1, 0, 0},{24255, 1, 1, 18}, {0, 0, 0, 0}}},{11028, {{24255, 1, 0, 0},{24255, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11029, {{24256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11030, {{24257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11031, {{24258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11032, {{24259, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11033, {{24260, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11034, {{24261, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11035, {{24262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11036, {{24263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11037, {{24264, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11038, {{24265, 1, 0, 0},{24265, 1, 1, 19}, {0, 0, 0, 0}}},{11038, {{24265, 1, 0, 0},{24265, 1, 1, 19}, {0, 0, 0, 0}}},{11039, {{24265, 1, 0, 0},{24265, 1, 1, 18}, {0, 0, 0, 0}}},{11039, {{24265, 1, 0, 0},{24265, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {5, {{11040, {{24266, 1, 0, 0},{24266, 1, 1, 19}, {0, 0, 0, 0}}},{11040, {{24266, 1, 0, 0},{24266, 1, 1, 19}, {0, 0, 0, 0}}},{30184, {{24266, 1, 0, 0},{24266, 1, 1, 0},{24266, 1, 1, 20},{63784, 1, 0, 0}, {0, 0, 0, 0}}},{30184, {{24266, 1, 0, 0},{24266, 1, 1, 0},{24266, 1, 1, 20},{63784, 1, 0, 0}, {0, 0, 0, 0}}},{30184, {{24266, 1, 0, 0},{24266, 1, 1, 0},{24266, 1, 1, 20},{63784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11041, {{24267, 1, 0, 0},{24267, 1, 1, 18}, {0, 0, 0, 0}}},{11041, {{24267, 1, 0, 0},{24267, 1, 1, 18}, {0, 0, 0, 0}}},{11042, {{24267, 1, 0, 0},{24267, 1, 1, 19}, {0, 0, 0, 0}}},{11042, {{24267, 1, 0, 0},{24267, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{11043, {{24268, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{11044, {{24270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11045, {{24271, 1, 0, 0},{24271, 1, 1, 18}, {0, 0, 0, 0}}},{11045, {{24271, 1, 0, 0},{24271, 1, 1, 18}, {0, 0, 0, 0}}},{11046, {{24271, 1, 0, 0},{24271, 1, 1, 19}, {0, 0, 0, 0}}},{11046, {{24271, 1, 0, 0},{24271, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {12, {{11047, {{24272, 1, 0, 0},{24272, 1, 1, 18}, {0, 0, 0, 0}}},{11047, {{24272, 1, 0, 0},{24272, 1, 1, 18}, {0, 0, 0, 0}}},{11048, {{24272, 1, 0, 0},{24272, 1, 1, 19}, {0, 0, 0, 0}}},{11048, {{24272, 1, 0, 0},{24272, 1, 1, 19}, {0, 0, 0, 0}}},{11049, {{24272, 1, 0, 0},{24272, 1, 1, 21}, {0, 0, 0, 0}}},{11049, {{24272, 1, 0, 0},{24272, 1, 1, 21}, {0, 0, 0, 0}}},{11050, {{24272, 1, 0, 0},{24272, 1, 1, 22}, {0, 0, 0, 0}}},{11050, {{24272, 1, 0, 0},{24272, 1, 1, 22}, {0, 0, 0, 0}}},{11051, {{24272, 1, 0, 0},{24272, 1, 1, 20}, {0, 0, 0, 0}}},{11051, {{24272, 1, 0, 0},{24272, 1, 1, 20}, {0, 0, 0, 0}}},{11052, {{24272, 1, 0, 0},{24272, 1, 1, 23}, {0, 0, 0, 0}}},{11052, {{24272, 1, 0, 0},{24272, 1, 1, 23}, {0, 0, 0, 0}}}}},
    {1, {{11053, {{24273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11054, {{24274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11055, {{24275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11056, {{24276, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11057, {{24277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11058, {{24278, 1, 0, 0},{24278, 1, 1, 17}, {0, 0, 0, 0}}},{11058, {{24278, 1, 0, 0},{24278, 1, 1, 17}, {0, 0, 0, 0}}},{11059, {{24278, 1, 0, 0},{24278, 1, 1, 18}, {0, 0, 0, 0}}},{11059, {{24278, 1, 0, 0},{24278, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11060, {{24279, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11061, {{24280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11062, {{24281, 1, 0, 0},{24281, 1, 1, 17}, {0, 0, 0, 0}}},{11062, {{24281, 1, 0, 0},{24281, 1, 1, 17}, {0, 0, 0, 0}}},{11063, {{24281, 1, 0, 0},{24281, 1, 1, 18}, {0, 0, 0, 0}}},{11063, {{24281, 1, 0, 0},{24281, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11064, {{24282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11065, {{24283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11066, {{24284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11067, {{24285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11068, {{24286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11069, {{24287, 1, 0, 0},{24287, 1, 1, 19}, {0, 0, 0, 0}}},{11069, {{24287, 1, 0, 0},{24287, 1, 1, 19}, {0, 0, 0, 0}}},{11070, {{24287, 1, 0, 0},{24287, 1, 1, 18}, {0, 0, 0, 0}}},{11070, {{24287, 1, 0, 0},{24287, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{11071, {{24288, 1, 0, 0},{24288, 1, 1, 19}, {0, 0, 0, 0}}},{11071, {{24288, 1, 0, 0},{24288, 1, 1, 19}, {0, 0, 0, 0}}},{11072, {{24288, 1, 0, 0},{24288, 1, 1, 18}, {0, 0, 0, 0}}},{11072, {{24288, 1, 0, 0},{24288, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11073, {{24289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11074, {{24290, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{11075, {{24291, 1, 0, 0},{24291, 1, 1, 20}, {0, 0, 0, 0}}},{11075, {{24291, 1, 0, 0},{24291, 1, 1, 20}, {0, 0, 0, 0}}},{11076, {{24291, 1, 0, 0},{24291, 1, 1, 21}, {0, 0, 0, 0}}},{11076, {{24291, 1, 0, 0},{24291, 1, 1, 21}, {0, 0, 0, 0}}},{11077, {{24291, 1, 0, 0},{24291, 1, 1, 19}, {0, 0, 0, 0}}},{11077, {{24291, 1, 0, 0},{24291, 1, 1, 19}, {0, 0, 0, 0}}},{11078, {{24291, 1, 0, 0},{24291, 1, 1, 28}, {0, 0, 0, 0}}},{11078, {{24291, 1, 0, 0},{24291, 1, 1, 28}, {0, 0, 0, 0}}}}},
    {1, {{68033, {{24292, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11084, {{24293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11085, {{24294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11086, {{24295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11087, {{24296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11088, {{24297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11089, {{24298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11090, {{24299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11091, {{24300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{11092, {{24302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11093, {{24303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11094, {{24304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11095, {{24305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11096, {{24306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11097, {{24307, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11098, {{24308, 1, 0, 0},{24308, 1, 1, 17}, {0, 0, 0, 0}}},{11098, {{24308, 1, 0, 0},{24308, 1, 1, 17}, {0, 0, 0, 0}}},{11099, {{24308, 1, 0, 0},{24308, 1, 1, 18}, {0, 0, 0, 0}}},{11099, {{24308, 1, 0, 0},{24308, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{11100, {{24310, 1, 0, 0},{24310, 1, 1, 21}, {0, 0, 0, 0}}},{11100, {{24310, 1, 0, 0},{24310, 1, 1, 21}, {0, 0, 0, 0}}},{11101, {{24310, 1, 0, 0},{24310, 1, 1, 19}, {0, 0, 0, 0}}},{11101, {{24310, 1, 0, 0},{24310, 1, 1, 19}, {0, 0, 0, 0}}},{11102, {{24310, 1, 0, 0},{24310, 1, 1, 20}, {0, 0, 0, 0}}},{11102, {{24310, 1, 0, 0},{24310, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {6, {{11103, {{24311, 1, 0, 0},{24311, 1, 1, 19}, {0, 0, 0, 0}}},{11103, {{24311, 1, 0, 0},{24311, 1, 1, 19}, {0, 0, 0, 0}}},{11104, {{24311, 1, 0, 0},{24311, 1, 1, 18}, {0, 0, 0, 0}}},{11104, {{24311, 1, 0, 0},{24311, 1, 1, 18}, {0, 0, 0, 0}}},{11105, {{24311, 1, 0, 0},{24311, 1, 1, 20}, {0, 0, 0, 0}}},{11105, {{24311, 1, 0, 0},{24311, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{11106, {{24312, 1, 0, 0},{24312, 1, 1, 17}, {0, 0, 0, 0}}},{11106, {{24312, 1, 0, 0},{24312, 1, 1, 17}, {0, 0, 0, 0}}},{11107, {{24312, 1, 0, 0},{24312, 1, 1, 18}, {0, 0, 0, 0}}},{11107, {{24312, 1, 0, 0},{24312, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{11108, {{24313, 1, 0, 0},{24313, 1, 1, 17}, {0, 0, 0, 0}}},{11108, {{24313, 1, 0, 0},{24313, 1, 1, 17}, {0, 0, 0, 0}}},{11109, {{24313, 1, 0, 0},{24313, 1, 1, 18}, {0, 0, 0, 0}}},{11109, {{24313, 1, 0, 0},{24313, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{11110, {{24314, 1, 0, 0},{24314, 1, 1, 20}, {0, 0, 0, 0}}},{11110, {{24314, 1, 0, 0},{24314, 1, 1, 20}, {0, 0, 0, 0}}},{11111, {{24314, 1, 0, 0},{24314, 1, 1, 18}, {0, 0, 0, 0}}},{11111, {{24314, 1, 0, 0},{24314, 1, 1, 18}, {0, 0, 0, 0}}},{11112, {{24314, 1, 0, 0},{24314, 1, 1, 19}, {0, 0, 0, 0}}},{11112, {{24314, 1, 0, 0},{24314, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{11113, {{24315, 1, 0, 0},{24315, 1, 1, 18}, {0, 0, 0, 0}}},{11113, {{24315, 1, 0, 0},{24315, 1, 1, 18}, {0, 0, 0, 0}}},{11114, {{24315, 1, 0, 0},{24315, 1, 1, 19}, {0, 0, 0, 0}}},{11114, {{24315, 1, 0, 0},{24315, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {6, {{11115, {{24316, 1, 0, 0},{24316, 1, 1, 17}, {0, 0, 0, 0}}},{11115, {{24316, 1, 0, 0},{24316, 1, 1, 17}, {0, 0, 0, 0}}},{11116, {{24316, 1, 0, 0},{24316, 1, 1, 18}, {0, 0, 0, 0}}},{11116, {{24316, 1, 0, 0},{24316, 1, 1, 18}, {0, 0, 0, 0}}},{11117, {{24316, 1, 0, 0},{24316, 1, 1, 19}, {0, 0, 0, 0}}},{11117, {{24316, 1, 0, 0},{24316, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{11118, {{24317, 1, 0, 0},{24317, 1, 1, 17}, {0, 0, 0, 0}}},{11118, {{24317, 1, 0, 0},{24317, 1, 1, 17}, {0, 0, 0, 0}}},{11119, {{24317, 1, 0, 0},{24317, 1, 1, 18}, {0, 0, 0, 0}}},{11119, {{24317, 1, 0, 0},{24317, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{11120, {{24318, 1, 0, 0},{24318, 1, 1, 17}, {0, 0, 0, 0}}},{11120, {{24318, 1, 0, 0},{24318, 1, 1, 17}, {0, 0, 0, 0}}},{11121, {{24318, 1, 0, 0},{24318, 1, 1, 18}, {0, 0, 0, 0}}},{11121, {{24318, 1, 0, 0},{24318, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11122, {{24319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11123, {{24320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11124, {{24321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11125, {{24322, 1, 0, 0},{24322, 1, 1, 17}, {0, 0, 0, 0}}},{11125, {{24322, 1, 0, 0},{24322, 1, 1, 17}, {0, 0, 0, 0}}},{11126, {{24322, 1, 0, 0},{24322, 1, 1, 18}, {0, 0, 0, 0}}},{11126, {{24322, 1, 0, 0},{24322, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11127, {{24323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11128, {{24324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11129, {{24325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11130, {{24326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11131, {{24327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11132, {{24328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11133, {{24329, 1, 0, 0},{24329, 1, 1, 17}, {0, 0, 0, 0}}},{11133, {{24329, 1, 0, 0},{24329, 1, 1, 17}, {0, 0, 0, 0}}},{11134, {{24329, 1, 0, 0},{24329, 1, 1, 18}, {0, 0, 0, 0}}},{11134, {{24329, 1, 0, 0},{24329, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{11136, {{24330, 1, 0, 0},{24330, 1, 1, 19}, {0, 0, 0, 0}}},{11136, {{24330, 1, 0, 0},{24330, 1, 1, 19}, {0, 0, 0, 0}}},{11137, {{24330, 1, 0, 0},{24330, 1, 1, 18}, {0, 0, 0, 0}}},{11137, {{24330, 1, 0, 0},{24330, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11138, {{24331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11139, {{24332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11140, {{24333, 1, 0, 0},{24333, 1, 1, 17}, {0, 0, 0, 0}}},{11140, {{24333, 1, 0, 0},{24333, 1, 1, 17}, {0, 0, 0, 0}}},{11141, {{24333, 1, 0, 0},{24333, 1, 1, 18}, {0, 0, 0, 0}}},{11141, {{24333, 1, 0, 0},{24333, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{11143, {{24334, 1, 0, 0},{24334, 1, 1, 17}, {0, 0, 0, 0}}},{11143, {{24334, 1, 0, 0},{24334, 1, 1, 17}, {0, 0, 0, 0}}},{11144, {{24334, 1, 0, 0},{24334, 1, 1, 18}, {0, 0, 0, 0}}},{11144, {{24334, 1, 0, 0},{24334, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11145, {{24335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11146, {{24336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11147, {{24337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11148, {{24338, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{11149, {{24339, 1, 0, 0},{24339, 1, 1, 17}, {0, 0, 0, 0}}},{11149, {{24339, 1, 0, 0},{24339, 1, 1, 17}, {0, 0, 0, 0}}},{57459, {{24339, 1, 0, 0},{24339, 1, 1, 18}, {0, 0, 0, 0}}},{57459, {{24339, 1, 0, 0},{24339, 1, 1, 18}, {0, 0, 0, 0}}},{59569, {{24339, 1, 0, 0},{24339, 1, 1, 19}, {0, 0, 0, 0}}},{59569, {{24339, 1, 0, 0},{24339, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{11150, {{24340, 1, 0, 0},{24340, 1, 1, 17}, {0, 0, 0, 0}}},{11150, {{24340, 1, 0, 0},{24340, 1, 1, 17}, {0, 0, 0, 0}}},{57631, {{24340, 1, 0, 0},{24340, 1, 1, 18}, {0, 0, 0, 0}}},{57631, {{24340, 1, 0, 0},{24340, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11151, {{24341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11153, {{24342, 1, 0, 0},{24342, 1, 1, 17}, {0, 0, 0, 0}}},{11153, {{24342, 1, 0, 0},{24342, 1, 1, 17}, {0, 0, 0, 0}}},{11154, {{24342, 1, 0, 0},{24342, 1, 1, 18}, {0, 0, 0, 0}}},{11154, {{24342, 1, 0, 0},{24342, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{11155, {{24343, 1, 0, 0},{24343, 1, 1, 17}, {0, 0, 0, 0}}},{11155, {{24343, 1, 0, 0},{24343, 1, 1, 17}, {0, 0, 0, 0}}},{57633, {{24343, 1, 0, 0},{24343, 1, 1, 18}, {0, 0, 0, 0}}},{57633, {{24343, 1, 0, 0},{24343, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11156, {{24344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11157, {{24345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11158, {{24346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11159, {{24347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11160, {{24348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11161, {{24349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11162, {{24350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11163, {{24351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{11167, {{24353, 1, 0, 0},{24353, 1, 1, 17}, {0, 0, 0, 0}}},{11167, {{24353, 1, 0, 0},{24353, 1, 1, 17}, {0, 0, 0, 0}}},{11168, {{24353, 1, 0, 0},{24353, 1, 1, 18}, {0, 0, 0, 0}}},{11168, {{24353, 1, 0, 0},{24353, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{11169, {{24354, 1, 0, 0},{24354, 1, 1, 17}, {0, 0, 0, 0}}},{11169, {{24354, 1, 0, 0},{24354, 1, 1, 17}, {0, 0, 0, 0}}},{11170, {{24354, 1, 0, 0},{24354, 1, 1, 18}, {0, 0, 0, 0}}},{11170, {{24354, 1, 0, 0},{24354, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11171, {{24355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11172, {{24356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11173, {{24357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11176, {{24358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11177, {{24359, 1, 0, 0},{24359, 1, 1, 18}, {0, 0, 0, 0}}},{11177, {{24359, 1, 0, 0},{24359, 1, 1, 18}, {0, 0, 0, 0}}},{11178, {{24359, 1, 0, 0},{24359, 1, 1, 17}, {0, 0, 0, 0}}},{11178, {{24359, 1, 0, 0},{24359, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{11179, {{24360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11180, {{24361, 1, 0, 0},{24361, 1, 1, 17}, {0, 0, 0, 0}}},{11180, {{24361, 1, 0, 0},{24361, 1, 1, 17}, {0, 0, 0, 0}}},{11181, {{24361, 1, 0, 0},{24361, 1, 1, 18}, {0, 0, 0, 0}}},{11181, {{24361, 1, 0, 0},{24361, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{11182, {{24363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11183, {{24364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11184, {{24365, 1, 0, 0},{24365, 1, 1, 18}, {0, 0, 0, 0}}},{11184, {{24365, 1, 0, 0},{24365, 1, 1, 18}, {0, 0, 0, 0}}},{11185, {{24365, 1, 0, 0},{24365, 1, 1, 19}, {0, 0, 0, 0}}},{11185, {{24365, 1, 0, 0},{24365, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{11186, {{24366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11188, {{24367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11189, {{24368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11190, {{24369, 1, 0, 0},{24369, 1, 1, 19}, {0, 0, 0, 0}}},{11190, {{24369, 1, 0, 0},{24369, 1, 1, 19}, {0, 0, 0, 0}}},{11191, {{24369, 1, 0, 0},{24369, 1, 1, 18}, {0, 0, 0, 0}}},{11191, {{24369, 1, 0, 0},{24369, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11192, {{24370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{11193, {{24372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11194, {{24373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11195, {{24374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11196, {{24375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11197, {{24376, 1, 0, 0},{24376, 1, 1, 19}, {0, 0, 0, 0}}},{11197, {{24376, 1, 0, 0},{24376, 1, 1, 19}, {0, 0, 0, 0}}},{11198, {{24376, 1, 0, 0},{24376, 1, 1, 18}, {0, 0, 0, 0}}},{11198, {{24376, 1, 0, 0},{24376, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{11199, {{24378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11200, {{24379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11201, {{24380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11202, {{24381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11203, {{24382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11204, {{24383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11205, {{24384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11206, {{24385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{11207, {{24387, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11208, {{24388, 1, 0, 0},{24388, 1, 1, 17}, {0, 0, 0, 0}}},{11208, {{24388, 1, 0, 0},{24388, 1, 1, 17}, {0, 0, 0, 0}}},{11209, {{24388, 1, 0, 0},{24388, 1, 1, 18}, {0, 0, 0, 0}}},{11209, {{24388, 1, 0, 0},{24388, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{11210, {{24389, 1, 0, 0},{24389, 1, 1, 18}, {0, 0, 0, 0}}},{11210, {{24389, 1, 0, 0},{24389, 1, 1, 18}, {0, 0, 0, 0}}},{11211, {{24389, 1, 0, 0},{24389, 1, 1, 19}, {0, 0, 0, 0}}},{11211, {{24389, 1, 0, 0},{24389, 1, 1, 19}, {0, 0, 0, 0}}},{59583, {{24389, 1, 0, 0},{24389, 1, 1, 20}, {0, 0, 0, 0}}},{59583, {{24389, 1, 0, 0},{24389, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{11213, {{24390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11214, {{24391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11215, {{24392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11216, {{24393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11217, {{24394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11218, {{24395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{11219, {{24396, 1, 0, 0},{24396, 1, 1, 17}, {0, 0, 0, 0}}},{11219, {{24396, 1, 0, 0},{24396, 1, 1, 17}, {0, 0, 0, 0}}},{11220, {{24396, 1, 0, 0},{24396, 1, 1, 18}, {0, 0, 0, 0}}},{11220, {{24396, 1, 0, 0},{24396, 1, 1, 18}, {0, 0, 0, 0}}},{11221, {{24396, 1, 0, 0},{24396, 1, 1, 19}, {0, 0, 0, 0}}},{11221, {{24396, 1, 0, 0},{24396, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{11222, {{24397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11223, {{24398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11224, {{24399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{11225, {{24400, 1, 0, 0},{24400, 1, 1, 18}, {0, 0, 0, 0}}},{11225, {{24400, 1, 0, 0},{24400, 1, 1, 18}, {0, 0, 0, 0}}},{11226, {{24400, 1, 0, 0},{24400, 1, 1, 19}, {0, 0, 0, 0}}},{11226, {{24400, 1, 0, 0},{24400, 1, 1, 19}, {0, 0, 0, 0}}},{11227, {{24400, 1, 0, 0},{24400, 1, 1, 20}, {0, 0, 0, 0}}},{11227, {{24400, 1, 0, 0},{24400, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{11228, {{24401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11229, {{24402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11230, {{24403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11231, {{24404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11232, {{24405, 1, 0, 0},{24405, 1, 1, 16}, {0, 0, 0, 0}}},{11232, {{24405, 1, 0, 0},{24405, 1, 1, 16}, {0, 0, 0, 0}}},{57475, {{24405, 1, 0, 0},{24405, 1, 1, 17}, {0, 0, 0, 0}}},{57475, {{24405, 1, 0, 0},{24405, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {6, {{11233, {{24406, 1, 0, 0},{24406, 1, 1, 17}, {0, 0, 0, 0}}},{11233, {{24406, 1, 0, 0},{24406, 1, 1, 17}, {0, 0, 0, 0}}},{11234, {{24406, 1, 0, 0},{24406, 1, 1, 18}, {0, 0, 0, 0}}},{11234, {{24406, 1, 0, 0},{24406, 1, 1, 18}, {0, 0, 0, 0}}},{11235, {{24406, 1, 0, 0},{24406, 1, 1, 19}, {0, 0, 0, 0}}},{11235, {{24406, 1, 0, 0},{24406, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {6, {{11236, {{24407, 1, 0, 0},{24407, 1, 1, 17}, {0, 0, 0, 0}}},{11236, {{24407, 1, 0, 0},{24407, 1, 1, 17}, {0, 0, 0, 0}}},{11237, {{24407, 1, 0, 0},{24407, 1, 1, 18}, {0, 0, 0, 0}}},{11237, {{24407, 1, 0, 0},{24407, 1, 1, 18}, {0, 0, 0, 0}}},{57478, {{24407, 1, 0, 0},{24407, 1, 1, 19}, {0, 0, 0, 0}}},{57478, {{24407, 1, 0, 0},{24407, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{11238, {{24408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11239, {{24409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{11240, {{24410, 1, 0, 0},{24410, 1, 1, 16}, {0, 0, 0, 0}}},{11240, {{24410, 1, 0, 0},{24410, 1, 1, 16}, {0, 0, 0, 0}}},{57480, {{24410, 1, 0, 0},{24410, 1, 1, 18}, {0, 0, 0, 0}}},{57480, {{24410, 1, 0, 0},{24410, 1, 1, 18}, {0, 0, 0, 0}}},{59586, {{24410, 1, 0, 0},{24410, 1, 1, 17}, {0, 0, 0, 0}}},{59586, {{24410, 1, 0, 0},{24410, 1, 1, 17}, {0, 0, 0, 0}}},{68054, {{24410, 1, 0, 0},{24410, 1, 1, 20}, {0, 0, 0, 0}}},{68054, {{24410, 1, 0, 0},{24410, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{11242, {{24411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11243, {{24412, 1, 0, 0},{24412, 1, 1, 17}, {0, 0, 0, 0}}},{11243, {{24412, 1, 0, 0},{24412, 1, 1, 17}, {0, 0, 0, 0}}},{11244, {{24412, 1, 0, 0},{24412, 1, 1, 18}, {0, 0, 0, 0}}},{11244, {{24412, 1, 0, 0},{24412, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11245, {{24413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11246, {{24414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{11248, {{24416, 1, 0, 0},{24416, 1, 1, 18}, {0, 0, 0, 0}}},{11248, {{24416, 1, 0, 0},{24416, 1, 1, 18}, {0, 0, 0, 0}}},{11249, {{24416, 1, 0, 0},{24416, 1, 1, 17}, {0, 0, 0, 0}}},{11249, {{24416, 1, 0, 0},{24416, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{11252, {{24417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11253, {{24418, 1, 0, 0},{24418, 1, 1, 18}, {0, 0, 0, 0}}},{11253, {{24418, 1, 0, 0},{24418, 1, 1, 18}, {0, 0, 0, 0}}},{11254, {{24418, 1, 0, 0},{24418, 1, 1, 19}, {0, 0, 0, 0}}},{11254, {{24418, 1, 0, 0},{24418, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{11255, {{24419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11256, {{24420, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11257, {{24421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11258, {{24422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11259, {{24423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{11260, {{24425, 1, 0, 0},{24425, 1, 1, 19}, {0, 0, 0, 0}}},{11260, {{24425, 1, 0, 0},{24425, 1, 1, 19}, {0, 0, 0, 0}}},{11261, {{24425, 1, 0, 0},{24425, 1, 1, 18}, {0, 0, 0, 0}}},{11261, {{24425, 1, 0, 0},{24425, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{11262, {{24426, 1, 0, 0},{24426, 1, 1, 17}, {0, 0, 0, 0}}},{11262, {{24426, 1, 0, 0},{24426, 1, 1, 17}, {0, 0, 0, 0}}},{11263, {{24426, 1, 0, 0},{24426, 1, 1, 18}, {0, 0, 0, 0}}},{11263, {{24426, 1, 0, 0},{24426, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{11266, {{24427, 1, 0, 0},{24427, 1, 1, 19}, {0, 0, 0, 0}}},{11266, {{24427, 1, 0, 0},{24427, 1, 1, 19}, {0, 0, 0, 0}}},{11267, {{24427, 1, 0, 0},{24427, 1, 1, 18}, {0, 0, 0, 0}}},{11267, {{24427, 1, 0, 0},{24427, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11268, {{24428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11269, {{24429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11270, {{24430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11271, {{24431, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11272, {{24432, 1, 0, 0},{24432, 1, 1, 17}, {0, 0, 0, 0}}},{11272, {{24432, 1, 0, 0},{24432, 1, 1, 17}, {0, 0, 0, 0}}},{11273, {{24432, 1, 0, 0},{24432, 1, 1, 18}, {0, 0, 0, 0}}},{11273, {{24432, 1, 0, 0},{24432, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11274, {{24433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11275, {{24434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11276, {{24435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11277, {{24436, 1, 0, 0},{24436, 1, 1, 17}, {0, 0, 0, 0}}},{11277, {{24436, 1, 0, 0},{24436, 1, 1, 17}, {0, 0, 0, 0}}},{11278, {{24436, 1, 0, 0},{24436, 1, 1, 18}, {0, 0, 0, 0}}},{11278, {{24436, 1, 0, 0},{24436, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11279, {{24437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11280, {{24438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11281, {{24439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11282, {{24440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11283, {{24441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11284, {{24442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{11285, {{24444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11286, {{24445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11287, {{24446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11288, {{24447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11289, {{24448, 1, 0, 0},{24448, 1, 1, 18}, {0, 0, 0, 0}}},{11289, {{24448, 1, 0, 0},{24448, 1, 1, 18}, {0, 0, 0, 0}}},{11290, {{24448, 1, 0, 0},{24448, 1, 1, 19}, {0, 0, 0, 0}}},{11290, {{24448, 1, 0, 0},{24448, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{11291, {{24449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11292, {{24450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11293, {{24451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11294, {{24452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11295, {{24453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11296, {{24454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11297, {{24455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11298, {{24456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11299, {{24457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11300, {{24458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11301, {{24459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11302, {{24460, 1, 0, 0},{24460, 1, 1, 18}, {0, 0, 0, 0}}},{11302, {{24460, 1, 0, 0},{24460, 1, 1, 18}, {0, 0, 0, 0}}},{11303, {{24460, 1, 0, 0},{24460, 1, 1, 17}, {0, 0, 0, 0}}},{11303, {{24460, 1, 0, 0},{24460, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{11304, {{24461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11305, {{24462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11306, {{24463, 1, 0, 0},{24463, 1, 1, 17}, {0, 0, 0, 0}}},{11306, {{24463, 1, 0, 0},{24463, 1, 1, 17}, {0, 0, 0, 0}}},{11307, {{24463, 1, 0, 0},{24463, 1, 1, 18}, {0, 0, 0, 0}}},{11307, {{24463, 1, 0, 0},{24463, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11308, {{24464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11309, {{24465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11311, {{24466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11312, {{24467, 1, 0, 0},{24467, 1, 1, 17}, {0, 0, 0, 0}}},{11312, {{24467, 1, 0, 0},{24467, 1, 1, 17}, {0, 0, 0, 0}}},{57488, {{24467, 1, 0, 0},{24467, 1, 1, 18}, {0, 0, 0, 0}}},{57488, {{24467, 1, 0, 0},{24467, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{11313, {{24470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11314, {{24471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11315, {{24472, 1, 0, 0},{24472, 1, 1, 18}, {0, 0, 0, 0}}},{11315, {{24472, 1, 0, 0},{24472, 1, 1, 18}, {0, 0, 0, 0}}},{11316, {{24472, 1, 0, 0},{24472, 1, 1, 19}, {0, 0, 0, 0}}},{11316, {{24472, 1, 0, 0},{24472, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{11317, {{24473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{11318, {{24475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11319, {{24476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11320, {{24477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11321, {{24478, 1, 0, 0},{24478, 1, 1, 17}, {0, 0, 0, 0}}},{11321, {{24478, 1, 0, 0},{24478, 1, 1, 17}, {0, 0, 0, 0}}},{59595, {{24478, 1, 0, 0},{24478, 1, 1, 18}, {0, 0, 0, 0}}},{59595, {{24478, 1, 0, 0},{24478, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11322, {{24479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11323, {{24480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{11324, {{24481, 1, 0, 0},{24481, 1, 1, 21}, {0, 0, 0, 0}}},{11324, {{24481, 1, 0, 0},{24481, 1, 1, 21}, {0, 0, 0, 0}}},{11325, {{24481, 1, 0, 0},{24481, 1, 1, 19}, {0, 0, 0, 0}}},{11325, {{24481, 1, 0, 0},{24481, 1, 1, 19}, {0, 0, 0, 0}}},{11326, {{24481, 1, 0, 0},{24481, 1, 1, 22}, {0, 0, 0, 0}}},{11326, {{24481, 1, 0, 0},{24481, 1, 1, 22}, {0, 0, 0, 0}}},{57489, {{24481, 1, 0, 0},{24481, 1, 1, 20}, {0, 0, 0, 0}}},{57489, {{24481, 1, 0, 0},{24481, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{11327, {{24482, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11328, {{24483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11329, {{24484, 1, 0, 0},{24484, 1, 1, 17}, {0, 0, 0, 0}}},{11329, {{24484, 1, 0, 0},{24484, 1, 1, 17}, {0, 0, 0, 0}}},{11330, {{24484, 1, 0, 0},{24484, 1, 1, 18}, {0, 0, 0, 0}}},{11330, {{24484, 1, 0, 0},{24484, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11331, {{24485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11332, {{24486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11333, {{24487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11334, {{24488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11335, {{24489, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11336, {{24490, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11337, {{24491, 1, 0, 0},{24491, 1, 1, 18}, {0, 0, 0, 0}}},{11337, {{24491, 1, 0, 0},{24491, 1, 1, 18}, {0, 0, 0, 0}}},{11338, {{24491, 1, 0, 0},{24491, 1, 1, 17}, {0, 0, 0, 0}}},{11338, {{24491, 1, 0, 0},{24491, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{11339, {{24492, 1, 0, 0},{24492, 1, 1, 17}, {0, 0, 0, 0}}},{11339, {{24492, 1, 0, 0},{24492, 1, 1, 17}, {0, 0, 0, 0}}},{11340, {{24492, 1, 0, 0},{24492, 1, 1, 18}, {0, 0, 0, 0}}},{11340, {{24492, 1, 0, 0},{24492, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11341, {{24493, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{11342, {{24494, 1, 0, 0},{24494, 1, 1, 21}, {0, 0, 0, 0}}},{11342, {{24494, 1, 0, 0},{24494, 1, 1, 21}, {0, 0, 0, 0}}},{11343, {{24494, 1, 0, 0},{24494, 1, 1, 18}, {0, 0, 0, 0}}},{11343, {{24494, 1, 0, 0},{24494, 1, 1, 18}, {0, 0, 0, 0}}},{11344, {{24494, 1, 0, 0},{24494, 1, 1, 20}, {0, 0, 0, 0}}},{11344, {{24494, 1, 0, 0},{24494, 1, 1, 20}, {0, 0, 0, 0}}},{11345, {{24494, 1, 0, 0},{24494, 1, 1, 19}, {0, 0, 0, 0}}},{11345, {{24494, 1, 0, 0},{24494, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{11346, {{24495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11347, {{24496, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11348, {{24497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11349, {{24498, 1, 0, 0},{24498, 1, 1, 17}, {0, 0, 0, 0}}},{11349, {{24498, 1, 0, 0},{24498, 1, 1, 17}, {0, 0, 0, 0}}},{11350, {{24498, 1, 0, 0},{24498, 1, 1, 16}, {0, 0, 0, 0}}},{11350, {{24498, 1, 0, 0},{24498, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {4, {{11351, {{24499, 1, 0, 0},{24499, 1, 1, 17}, {0, 0, 0, 0}}},{11351, {{24499, 1, 0, 0},{24499, 1, 1, 17}, {0, 0, 0, 0}}},{59596, {{24499, 1, 0, 0},{24499, 1, 1, 18}, {0, 0, 0, 0}}},{59596, {{24499, 1, 0, 0},{24499, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11352, {{24500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11353, {{24501, 1, 0, 0},{24501, 1, 1, 19}, {0, 0, 0, 0}}},{11353, {{24501, 1, 0, 0},{24501, 1, 1, 19}, {0, 0, 0, 0}}},{11354, {{24501, 1, 0, 0},{24501, 1, 1, 18}, {0, 0, 0, 0}}},{11354, {{24501, 1, 0, 0},{24501, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11355, {{24502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11356, {{24503, 1, 0, 0},{24503, 1, 1, 17}, {0, 0, 0, 0}}},{11356, {{24503, 1, 0, 0},{24503, 1, 1, 17}, {0, 0, 0, 0}}},{11357, {{24503, 1, 0, 0},{24503, 1, 1, 18}, {0, 0, 0, 0}}},{11357, {{24503, 1, 0, 0},{24503, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11358, {{24504, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11359, {{24505, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{11360, {{24507, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11361, {{24508, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11362, {{24509, 1, 0, 0},{24509, 1, 1, 18}, {0, 0, 0, 0}}},{11362, {{24509, 1, 0, 0},{24509, 1, 1, 18}, {0, 0, 0, 0}}},{11363, {{24509, 1, 0, 0},{24509, 1, 1, 19}, {0, 0, 0, 0}}},{11363, {{24509, 1, 0, 0},{24509, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{11364, {{24510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11365, {{24511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11366, {{24512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11367, {{24513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11368, {{24514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11369, {{24515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11370, {{24516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11371, {{24517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{11372, {{24519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11373, {{24520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11374, {{24521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11375, {{24522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11376, {{24523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11377, {{24524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{11378, {{24525, 1, 0, 0},{24525, 1, 1, 20}, {0, 0, 0, 0}}},{11378, {{24525, 1, 0, 0},{24525, 1, 1, 20}, {0, 0, 0, 0}}},{11379, {{24525, 1, 0, 0},{24525, 1, 1, 18}, {0, 0, 0, 0}}},{11379, {{24525, 1, 0, 0},{24525, 1, 1, 18}, {0, 0, 0, 0}}},{11380, {{24525, 1, 0, 0},{24525, 1, 1, 19}, {0, 0, 0, 0}}},{11380, {{24525, 1, 0, 0},{24525, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{11381, {{24526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11382, {{24527, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11383, {{24528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11384, {{24529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11385, {{24530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11386, {{24531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11387, {{24532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11388, {{24533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11389, {{24534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11390, {{24535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11392, {{24536, 1, 0, 0},{24536, 1, 1, 19}, {0, 0, 0, 0}}},{11392, {{24536, 1, 0, 0},{24536, 1, 1, 19}, {0, 0, 0, 0}}},{11393, {{24536, 1, 0, 0},{24536, 1, 1, 18}, {0, 0, 0, 0}}},{11393, {{24536, 1, 0, 0},{24536, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{11394, {{24537, 1, 0, 0},{24537, 1, 1, 19}, {0, 0, 0, 0}}},{11394, {{24537, 1, 0, 0},{24537, 1, 1, 19}, {0, 0, 0, 0}}},{11395, {{24537, 1, 0, 0},{24537, 1, 1, 18}, {0, 0, 0, 0}}},{11395, {{24537, 1, 0, 0},{24537, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11396, {{24538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{11398, {{24540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11399, {{24541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11400, {{24542, 1, 0, 0},{24542, 1, 1, 17}, {0, 0, 0, 0}}},{11400, {{24542, 1, 0, 0},{24542, 1, 1, 17}, {0, 0, 0, 0}}},{11401, {{24542, 1, 0, 0},{24542, 1, 1, 18}, {0, 0, 0, 0}}},{11401, {{24542, 1, 0, 0},{24542, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{11402, {{24543, 1, 0, 0},{24543, 1, 1, 16}, {0, 0, 0, 0}}},{11402, {{24543, 1, 0, 0},{24543, 1, 1, 16}, {0, 0, 0, 0}}},{11403, {{24543, 1, 0, 0},{24543, 1, 1, 17}, {0, 0, 0, 0}}},{11403, {{24543, 1, 0, 0},{24543, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{11404, {{24544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11405, {{24545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11406, {{24546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11407, {{24547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11409, {{24548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11410, {{24549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11411, {{24550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11412, {{24551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11413, {{24552, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11414, {{24553, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11415, {{24554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11416, {{24555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11417, {{24556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11418, {{24557, 1, 0, 0},{24557, 1, 1, 17}, {0, 0, 0, 0}}},{11418, {{24557, 1, 0, 0},{24557, 1, 1, 17}, {0, 0, 0, 0}}},{11419, {{24557, 1, 0, 0},{24557, 1, 1, 18}, {0, 0, 0, 0}}},{11419, {{24557, 1, 0, 0},{24557, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11420, {{24558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11421, {{24559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11422, {{24560, 1, 0, 0},{24560, 1, 1, 17}, {0, 0, 0, 0}}},{11422, {{24560, 1, 0, 0},{24560, 1, 1, 17}, {0, 0, 0, 0}}},{11423, {{24560, 1, 0, 0},{24560, 1, 1, 18}, {0, 0, 0, 0}}},{11423, {{24560, 1, 0, 0},{24560, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11424, {{24561, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11425, {{24562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11426, {{24563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11427, {{24564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11428, {{24565, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11429, {{24566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11430, {{24567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11431, {{24568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11432, {{24569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11433, {{24570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11434, {{24571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11435, {{24572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11436, {{24573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11437, {{24574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11438, {{24575, 1, 0, 0},{24575, 1, 1, 17}, {0, 0, 0, 0}}},{11438, {{24575, 1, 0, 0},{24575, 1, 1, 17}, {0, 0, 0, 0}}},{11439, {{24575, 1, 0, 0},{24575, 1, 1, 18}, {0, 0, 0, 0}}},{11439, {{24575, 1, 0, 0},{24575, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11440, {{24576, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{11441, {{24583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11442, {{24584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11443, {{24585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11444, {{24586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11445, {{24587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11446, {{24588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11447, {{24589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11448, {{24590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11449, {{24591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11450, {{24592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11451, {{24593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11452, {{24594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11453, {{24595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11454, {{24596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11455, {{24597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11456, {{24598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11457, {{24599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11458, {{24600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11459, {{24601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11460, {{24602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11461, {{24603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11462, {{24604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11464, {{24605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11465, {{24606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11466, {{24607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11467, {{24608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11468, {{24609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11469, {{24610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11470, {{24611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11471, {{24612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11472, {{24613, 1, 0, 0},{24613, 1, 1, 19}, {0, 0, 0, 0}}},{11472, {{24613, 1, 0, 0},{24613, 1, 1, 19}, {0, 0, 0, 0}}},{11473, {{24613, 1, 0, 0},{24613, 1, 1, 18}, {0, 0, 0, 0}}},{11473, {{24613, 1, 0, 0},{24613, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{11475, {{24614, 1, 0, 0},{24614, 1, 1, 17}, {0, 0, 0, 0}}},{11475, {{24614, 1, 0, 0},{24614, 1, 1, 17}, {0, 0, 0, 0}}},{11476, {{24614, 1, 0, 0},{24614, 1, 1, 18}, {0, 0, 0, 0}}},{11476, {{24614, 1, 0, 0},{24614, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11477, {{24615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11478, {{24616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11479, {{24617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11480, {{24618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11481, {{24619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11482, {{24620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11483, {{24621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11484, {{24622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11485, {{24623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11486, {{24624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11487, {{24625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11488, {{24626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11489, {{24627, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11490, {{24628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11491, {{24629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11492, {{24630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11493, {{24631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11494, {{24632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11496, {{24633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11497, {{24634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{11498, {{24640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11499, {{24641, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11500, {{24642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11501, {{24643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11502, {{24644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11504, {{24645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11505, {{24646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11506, {{24647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11507, {{24648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11508, {{24649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11509, {{24650, 1, 0, 0},{24650, 1, 1, 17}, {0, 0, 0, 0}}},{11509, {{24650, 1, 0, 0},{24650, 1, 1, 17}, {0, 0, 0, 0}}},{57097, {{24650, 1, 0, 0},{24650, 1, 1, 18}, {0, 0, 0, 0}}},{57097, {{24650, 1, 0, 0},{24650, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11510, {{24651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11511, {{24652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11512, {{24653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11513, {{24654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11514, {{24655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11515, {{24656, 1, 0, 0},{24656, 1, 1, 18}, {0, 0, 0, 0}}},{11515, {{24656, 1, 0, 0},{24656, 1, 1, 18}, {0, 0, 0, 0}}},{57504, {{24656, 1, 0, 0},{24656, 1, 1, 19}, {0, 0, 0, 0}}},{57504, {{24656, 1, 0, 0},{24656, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{11516, {{24657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11517, {{24658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11518, {{24659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11519, {{24660, 1, 0, 0},{24660, 1, 1, 17}, {0, 0, 0, 0}}},{11519, {{24660, 1, 0, 0},{24660, 1, 1, 17}, {0, 0, 0, 0}}},{11520, {{24660, 1, 0, 0},{24660, 1, 1, 18}, {0, 0, 0, 0}}},{11520, {{24660, 1, 0, 0},{24660, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11521, {{24661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11522, {{24662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11524, {{24663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11525, {{24664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11526, {{24665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11527, {{24666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11528, {{24667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11529, {{24668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{11530, {{24669, 1, 0, 0},{24669, 1, 1, 17}, {0, 0, 0, 0}}},{11530, {{24669, 1, 0, 0},{24669, 1, 1, 17}, {0, 0, 0, 0}}},{11531, {{24669, 1, 0, 0},{24669, 1, 1, 19}, {0, 0, 0, 0}}},{11531, {{24669, 1, 0, 0},{24669, 1, 1, 19}, {0, 0, 0, 0}}},{11532, {{24669, 1, 0, 0},{24669, 1, 1, 18}, {0, 0, 0, 0}}},{11532, {{24669, 1, 0, 0},{24669, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11533, {{24670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11534, {{24671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11535, {{24672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11536, {{24673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{11537, {{24674, 1, 0, 0},{24674, 1, 1, 20}, {0, 0, 0, 0}}},{11537, {{24674, 1, 0, 0},{24674, 1, 1, 20}, {0, 0, 0, 0}}},{11538, {{24674, 1, 0, 0},{24674, 1, 1, 19}, {0, 0, 0, 0}}},{11538, {{24674, 1, 0, 0},{24674, 1, 1, 19}, {0, 0, 0, 0}}},{11539, {{24674, 1, 0, 0},{24674, 1, 1, 21}, {0, 0, 0, 0}}},{11539, {{24674, 1, 0, 0},{24674, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{11540, {{24675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11541, {{24676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11542, {{24677, 1, 0, 0},{24677, 1, 1, 18}, {0, 0, 0, 0}}},{11542, {{24677, 1, 0, 0},{24677, 1, 1, 18}, {0, 0, 0, 0}}},{11543, {{24677, 1, 0, 0},{24677, 1, 1, 19}, {0, 0, 0, 0}}},{11543, {{24677, 1, 0, 0},{24677, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{11544, {{24678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11545, {{24679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11546, {{24680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11547, {{24681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11548, {{24682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11549, {{24683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11550, {{24684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11551, {{24685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11552, {{24686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11553, {{24687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11554, {{24688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11555, {{24689, 1, 0, 0},{24689, 1, 1, 17}, {0, 0, 0, 0}}},{11555, {{24689, 1, 0, 0},{24689, 1, 1, 17}, {0, 0, 0, 0}}},{11556, {{24689, 1, 0, 0},{24689, 1, 1, 18}, {0, 0, 0, 0}}},{11556, {{24689, 1, 0, 0},{24689, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11558, {{24690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11559, {{24691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11560, {{24692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{11561, {{24693, 1, 0, 0},{24693, 1, 1, 18}, {0, 0, 0, 0}}},{11561, {{24693, 1, 0, 0},{24693, 1, 1, 18}, {0, 0, 0, 0}}},{30308, {{24693, 1, 0, 0},{24693, 1, 1, 0},{24693, 1, 1, 19},{64107, 1, 0, 0}, {0, 0, 0, 0}}},{30308, {{24693, 1, 0, 0},{24693, 1, 1, 0},{24693, 1, 1, 19},{64107, 1, 0, 0}, {0, 0, 0, 0}}},{30308, {{24693, 1, 0, 0},{24693, 1, 1, 0},{24693, 1, 1, 19},{64107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{11562, {{24695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{11563, {{24702, 1, 0, 0},{24702, 1, 1, 17}, {0, 0, 0, 0}}},{11563, {{24702, 1, 0, 0},{24702, 1, 1, 17}, {0, 0, 0, 0}}},{11564, {{24702, 1, 0, 0},{24702, 1, 1, 18}, {0, 0, 0, 0}}},{11564, {{24702, 1, 0, 0},{24702, 1, 1, 18}, {0, 0, 0, 0}}},{11565, {{24702, 1, 0, 0},{24702, 1, 1, 19}, {0, 0, 0, 0}}},{11565, {{24702, 1, 0, 0},{24702, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{11566, {{24703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11567, {{24704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11568, {{24705, 1, 0, 0},{24705, 1, 1, 18}, {0, 0, 0, 0}}},{11568, {{24705, 1, 0, 0},{24705, 1, 1, 18}, {0, 0, 0, 0}}},{11569, {{24705, 1, 0, 0},{24705, 1, 1, 17}, {0, 0, 0, 0}}},{11569, {{24705, 1, 0, 0},{24705, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{11570, {{24706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11571, {{24707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11572, {{24708, 1, 0, 0},{24708, 1, 1, 17}, {0, 0, 0, 0}}},{11572, {{24708, 1, 0, 0},{24708, 1, 1, 17}, {0, 0, 0, 0}}},{11573, {{24708, 1, 0, 0},{24708, 1, 1, 18}, {0, 0, 0, 0}}},{11573, {{24708, 1, 0, 0},{24708, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11574, {{24709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11575, {{24710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11576, {{24711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11577, {{24712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11578, {{24713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11579, {{24714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11580, {{24715, 1, 0, 0},{24715, 1, 1, 17}, {0, 0, 0, 0}}},{11580, {{24715, 1, 0, 0},{24715, 1, 1, 17}, {0, 0, 0, 0}}},{11581, {{24715, 1, 0, 0},{24715, 1, 1, 18}, {0, 0, 0, 0}}},{11581, {{24715, 1, 0, 0},{24715, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11582, {{24716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11583, {{24717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11584, {{24718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11586, {{24719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11587, {{24720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11588, {{24721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11589, {{24722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11590, {{24723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{11591, {{24724, 1, 0, 0},{24724, 1, 1, 18}, {0, 0, 0, 0}}},{11591, {{24724, 1, 0, 0},{24724, 1, 1, 18}, {0, 0, 0, 0}}},{30257, {{24724, 1, 0, 0},{24724, 1, 1, 0},{24724, 1, 1, 19},{64061, 1, 0, 0}, {0, 0, 0, 0}}},{30257, {{24724, 1, 0, 0},{24724, 1, 1, 0},{24724, 1, 1, 19},{64061, 1, 0, 0}, {0, 0, 0, 0}}},{30257, {{24724, 1, 0, 0},{24724, 1, 1, 0},{24724, 1, 1, 19},{64061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11592, {{24725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11593, {{24726, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{11594, {{24727, 1, 0, 0},{24727, 1, 1, 19}, {0, 0, 0, 0}}},{11594, {{24727, 1, 0, 0},{24727, 1, 1, 19}, {0, 0, 0, 0}}},{11595, {{24727, 1, 0, 0},{24727, 1, 1, 20}, {0, 0, 0, 0}}},{11595, {{24727, 1, 0, 0},{24727, 1, 1, 20}, {0, 0, 0, 0}}},{11596, {{24727, 1, 0, 0},{24727, 1, 1, 21}, {0, 0, 0, 0}}},{11596, {{24727, 1, 0, 0},{24727, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{11597, {{24728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11598, {{24729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11601, {{24730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11602, {{24731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11603, {{24732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11605, {{24733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11606, {{24734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11608, {{24735, 1, 0, 0},{24735, 1, 1, 17}, {0, 0, 0, 0}}},{11608, {{24735, 1, 0, 0},{24735, 1, 1, 17}, {0, 0, 0, 0}}},{59607, {{24735, 1, 0, 0},{24735, 1, 1, 18}, {0, 0, 0, 0}}},{59607, {{24735, 1, 0, 0},{24735, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11609, {{24736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11610, {{24737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11611, {{24738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11612, {{24739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{11613, {{24740, 1, 0, 0},{24740, 1, 1, 18}, {0, 0, 0, 0}}},{11613, {{24740, 1, 0, 0},{24740, 1, 1, 18}, {0, 0, 0, 0}}},{11614, {{24740, 1, 0, 0},{24740, 1, 1, 19}, {0, 0, 0, 0}}},{11614, {{24740, 1, 0, 0},{24740, 1, 1, 19}, {0, 0, 0, 0}}},{57507, {{24740, 1, 0, 0},{24740, 1, 1, 20}, {0, 0, 0, 0}}},{57507, {{24740, 1, 0, 0},{24740, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{11615, {{24741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11616, {{24742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11617, {{24743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11618, {{24744, 1, 0, 0},{24744, 1, 1, 17}, {0, 0, 0, 0}}},{11618, {{24744, 1, 0, 0},{24744, 1, 1, 17}, {0, 0, 0, 0}}},{11619, {{24744, 1, 0, 0},{24744, 1, 1, 18}, {0, 0, 0, 0}}},{11619, {{24744, 1, 0, 0},{24744, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11620, {{24745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11621, {{24746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{11607, {{24750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{11622, {{24752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11623, {{24753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11624, {{24754, 1, 0, 0},{24754, 1, 1, 18}, {0, 0, 0, 0}}},{11624, {{24754, 1, 0, 0},{24754, 1, 1, 18}, {0, 0, 0, 0}}},{11625, {{24754, 1, 0, 0},{24754, 1, 1, 19}, {0, 0, 0, 0}}},{11625, {{24754, 1, 0, 0},{24754, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{11626, {{24755, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11629, {{24756, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11630, {{24757, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11631, {{24758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11632, {{24759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11633, {{24760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11634, {{24761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11635, {{24762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11636, {{24763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11637, {{24764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11638, {{24765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11639, {{24766, 1, 0, 0},{24766, 1, 1, 17}, {0, 0, 0, 0}}},{11639, {{24766, 1, 0, 0},{24766, 1, 1, 17}, {0, 0, 0, 0}}},{11640, {{24766, 1, 0, 0},{24766, 1, 1, 18}, {0, 0, 0, 0}}},{11640, {{24766, 1, 0, 0},{24766, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11641, {{24767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11642, {{24768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11643, {{24769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11644, {{24770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11645, {{24771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11646, {{24772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11647, {{24773, 1, 0, 0},{24773, 1, 1, 19}, {0, 0, 0, 0}}},{11647, {{24773, 1, 0, 0},{24773, 1, 1, 19}, {0, 0, 0, 0}}},{11648, {{24773, 1, 0, 0},{24773, 1, 1, 18}, {0, 0, 0, 0}}},{11648, {{24773, 1, 0, 0},{24773, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{11649, {{24774, 1, 0, 0},{24774, 1, 1, 17}, {0, 0, 0, 0}}},{11649, {{24774, 1, 0, 0},{24774, 1, 1, 17}, {0, 0, 0, 0}}},{11650, {{24774, 1, 0, 0},{24774, 1, 1, 18}, {0, 0, 0, 0}}},{11650, {{24774, 1, 0, 0},{24774, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{11651, {{24775, 1, 0, 0},{24775, 1, 1, 17}, {0, 0, 0, 0}}},{11651, {{24775, 1, 0, 0},{24775, 1, 1, 17}, {0, 0, 0, 0}}},{11652, {{24775, 1, 0, 0},{24775, 1, 1, 18}, {0, 0, 0, 0}}},{11652, {{24775, 1, 0, 0},{24775, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11653, {{24776, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11654, {{24777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11655, {{24778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11656, {{24779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11657, {{24780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11658, {{24781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11659, {{24782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11660, {{24783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11661, {{24784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11662, {{24785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11663, {{24786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11664, {{24787, 1, 0, 0},{24787, 1, 1, 17}, {0, 0, 0, 0}}},{11664, {{24787, 1, 0, 0},{24787, 1, 1, 17}, {0, 0, 0, 0}}},{11665, {{24787, 1, 0, 0},{24787, 1, 1, 18}, {0, 0, 0, 0}}},{11665, {{24787, 1, 0, 0},{24787, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11666, {{24788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11667, {{24789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11668, {{24790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11669, {{24791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{11670, {{24792, 1, 0, 0},{24792, 1, 1, 18}, {0, 0, 0, 0}}},{11670, {{24792, 1, 0, 0},{24792, 1, 1, 18}, {0, 0, 0, 0}}},{11671, {{24792, 1, 0, 0},{24792, 1, 1, 20}, {0, 0, 0, 0}}},{11671, {{24792, 1, 0, 0},{24792, 1, 1, 20}, {0, 0, 0, 0}}},{11672, {{24792, 1, 0, 0},{24792, 1, 1, 19}, {0, 0, 0, 0}}},{11672, {{24792, 1, 0, 0},{24792, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{11673, {{24793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11674, {{24794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11675, {{24795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11676, {{24796, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11677, {{24797, 1, 0, 0},{24797, 1, 1, 17}, {0, 0, 0, 0}}},{11677, {{24797, 1, 0, 0},{24797, 1, 1, 17}, {0, 0, 0, 0}}},{11678, {{24797, 1, 0, 0},{24797, 1, 1, 18}, {0, 0, 0, 0}}},{11678, {{24797, 1, 0, 0},{24797, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11679, {{24798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11680, {{24799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{11681, {{24800, 1, 0, 0},{24800, 1, 1, 17}, {0, 0, 0, 0}}},{11681, {{24800, 1, 0, 0},{24800, 1, 1, 17}, {0, 0, 0, 0}}},{11682, {{24800, 1, 0, 0},{24800, 1, 1, 18}, {0, 0, 0, 0}}},{11682, {{24800, 1, 0, 0},{24800, 1, 1, 18}, {0, 0, 0, 0}}},{59623, {{24800, 1, 0, 0},{24800, 1, 1, 19}, {0, 0, 0, 0}}},{59623, {{24800, 1, 0, 0},{24800, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{11683, {{24801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11684, {{24802, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11685, {{24803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11686, {{24804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11687, {{24805, 1, 0, 0},{24805, 1, 1, 17}, {0, 0, 0, 0}}},{11687, {{24805, 1, 0, 0},{24805, 1, 1, 17}, {0, 0, 0, 0}}},{11688, {{24805, 1, 0, 0},{24805, 1, 1, 18}, {0, 0, 0, 0}}},{11688, {{24805, 1, 0, 0},{24805, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11689, {{24806, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11690, {{24807, 1, 0, 0},{24807, 1, 1, 17}, {0, 0, 0, 0}}},{11690, {{24807, 1, 0, 0},{24807, 1, 1, 17}, {0, 0, 0, 0}}},{11691, {{24807, 1, 0, 0},{24807, 1, 1, 18}, {0, 0, 0, 0}}},{11691, {{24807, 1, 0, 0},{24807, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11692, {{24808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11693, {{24809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57511, {{24810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{11694, {{24814, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{11695, {{24816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11696, {{24817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11697, {{24818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11698, {{24819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11699, {{24820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11700, {{24821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11701, {{24822, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11702, {{24823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11703, {{24824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11704, {{24825, 1, 0, 0},{24825, 1, 1, 17}, {0, 0, 0, 0}}},{11704, {{24825, 1, 0, 0},{24825, 1, 1, 17}, {0, 0, 0, 0}}},{11705, {{24825, 1, 0, 0},{24825, 1, 1, 18}, {0, 0, 0, 0}}},{11705, {{24825, 1, 0, 0},{24825, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11706, {{24826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11707, {{24827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11708, {{24828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11709, {{24829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11710, {{24830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11711, {{24831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11712, {{24832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11713, {{24833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11714, {{24834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11715, {{24835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11716, {{24836, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11717, {{24837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11718, {{24838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11719, {{24839, 1, 0, 0},{24839, 1, 1, 18}, {0, 0, 0, 0}}},{11719, {{24839, 1, 0, 0},{24839, 1, 1, 18}, {0, 0, 0, 0}}},{11720, {{24839, 1, 0, 0},{24839, 1, 1, 17}, {0, 0, 0, 0}}},{11720, {{24839, 1, 0, 0},{24839, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {6, {{11721, {{24840, 1, 0, 0},{24840, 1, 1, 19}, {0, 0, 0, 0}}},{11721, {{24840, 1, 0, 0},{24840, 1, 1, 19}, {0, 0, 0, 0}}},{11722, {{24840, 1, 0, 0},{24840, 1, 1, 18}, {0, 0, 0, 0}}},{11722, {{24840, 1, 0, 0},{24840, 1, 1, 18}, {0, 0, 0, 0}}},{11723, {{24840, 1, 0, 0},{24840, 1, 1, 20}, {0, 0, 0, 0}}},{11723, {{24840, 1, 0, 0},{24840, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{11725, {{24841, 1, 0, 0},{24841, 1, 1, 19}, {0, 0, 0, 0}}},{11725, {{24841, 1, 0, 0},{24841, 1, 1, 19}, {0, 0, 0, 0}}},{11726, {{24841, 1, 0, 0},{24841, 1, 1, 18}, {0, 0, 0, 0}}},{11726, {{24841, 1, 0, 0},{24841, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11727, {{24842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11729, {{24843, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11730, {{24844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11731, {{24845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11732, {{24846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11733, {{24847, 1, 0, 0},{24847, 1, 1, 18}, {0, 0, 0, 0}}},{11733, {{24847, 1, 0, 0},{24847, 1, 1, 18}, {0, 0, 0, 0}}},{11734, {{24847, 1, 0, 0},{24847, 1, 1, 19}, {0, 0, 0, 0}}},{11734, {{24847, 1, 0, 0},{24847, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{11735, {{24848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11736, {{24849, 1, 0, 0},{24849, 1, 1, 17}, {0, 0, 0, 0}}},{11736, {{24849, 1, 0, 0},{24849, 1, 1, 17}, {0, 0, 0, 0}}},{11737, {{24849, 1, 0, 0},{24849, 1, 1, 18}, {0, 0, 0, 0}}},{11737, {{24849, 1, 0, 0},{24849, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{11738, {{24850, 1, 0, 0},{24850, 1, 1, 17}, {0, 0, 0, 0}}},{11738, {{24850, 1, 0, 0},{24850, 1, 1, 17}, {0, 0, 0, 0}}},{11739, {{24850, 1, 0, 0},{24850, 1, 1, 18}, {0, 0, 0, 0}}},{11739, {{24850, 1, 0, 0},{24850, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11740, {{24851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11741, {{24852, 1, 0, 0},{24852, 1, 1, 17}, {0, 0, 0, 0}}},{11741, {{24852, 1, 0, 0},{24852, 1, 1, 17}, {0, 0, 0, 0}}},{11742, {{24852, 1, 0, 0},{24852, 1, 1, 18}, {0, 0, 0, 0}}},{11742, {{24852, 1, 0, 0},{24852, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{11743, {{24853, 1, 0, 0},{24853, 1, 1, 17}, {0, 0, 0, 0}}},{11743, {{24853, 1, 0, 0},{24853, 1, 1, 17}, {0, 0, 0, 0}}},{11744, {{24853, 1, 0, 0},{24853, 1, 1, 18}, {0, 0, 0, 0}}},{11744, {{24853, 1, 0, 0},{24853, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{11745, {{24854, 1, 0, 0},{24854, 1, 1, 17}, {0, 0, 0, 0}}},{11745, {{24854, 1, 0, 0},{24854, 1, 1, 17}, {0, 0, 0, 0}}},{11746, {{24854, 1, 0, 0},{24854, 1, 1, 18}, {0, 0, 0, 0}}},{11746, {{24854, 1, 0, 0},{24854, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11747, {{24855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11748, {{24856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11749, {{24857, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11750, {{24858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{11751, {{24859, 1, 0, 0},{24859, 1, 1, 18}, {0, 0, 0, 0}}},{11751, {{24859, 1, 0, 0},{24859, 1, 1, 18}, {0, 0, 0, 0}}},{11752, {{24859, 1, 0, 0},{24859, 1, 1, 17}, {0, 0, 0, 0}}},{11752, {{24859, 1, 0, 0},{24859, 1, 1, 17}, {0, 0, 0, 0}}},{57515, {{24859, 1, 0, 0},{24859, 1, 1, 19}, {0, 0, 0, 0}}},{57515, {{24859, 1, 0, 0},{24859, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{11753, {{24860, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11754, {{24861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11755, {{24862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11756, {{24863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11757, {{24864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11758, {{24865, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11759, {{24866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11760, {{24867, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{11761, {{24869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{11762, {{24871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11763, {{24872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11764, {{24873, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11765, {{24874, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11766, {{24875, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11767, {{24876, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11768, {{24877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11769, {{24878, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11770, {{24879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11771, {{24880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11772, {{24881, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11773, {{24882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11774, {{24883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11775, {{24884, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11776, {{24885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11777, {{24886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11778, {{24887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11779, {{24888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11780, {{24889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11781, {{24890, 1, 0, 0},{24890, 1, 1, 17}, {0, 0, 0, 0}}},{11781, {{24890, 1, 0, 0},{24890, 1, 1, 17}, {0, 0, 0, 0}}},{11782, {{24890, 1, 0, 0},{24890, 1, 1, 18}, {0, 0, 0, 0}}},{11782, {{24890, 1, 0, 0},{24890, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11783, {{24891, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{11784, {{24892, 1, 0, 0},{24892, 1, 1, 17}, {0, 0, 0, 0}}},{11784, {{24892, 1, 0, 0},{24892, 1, 1, 17}, {0, 0, 0, 0}}},{11785, {{24892, 1, 0, 0},{24892, 1, 1, 20}, {0, 0, 0, 0}}},{11785, {{24892, 1, 0, 0},{24892, 1, 1, 20}, {0, 0, 0, 0}}},{11786, {{24892, 1, 0, 0},{24892, 1, 1, 18}, {0, 0, 0, 0}}},{11786, {{24892, 1, 0, 0},{24892, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{11788, {{24893, 1, 0, 0},{24893, 1, 1, 18}, {0, 0, 0, 0}}},{11788, {{24893, 1, 0, 0},{24893, 1, 1, 18}, {0, 0, 0, 0}}},{11789, {{24893, 1, 0, 0},{24893, 1, 1, 20}, {0, 0, 0, 0}}},{11789, {{24893, 1, 0, 0},{24893, 1, 1, 20}, {0, 0, 0, 0}}},{11790, {{24893, 1, 0, 0},{24893, 1, 1, 19}, {0, 0, 0, 0}}},{11790, {{24893, 1, 0, 0},{24893, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{11791, {{24894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11792, {{24895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11794, {{24896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11795, {{24897, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11796, {{24898, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11797, {{24899, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11798, {{24900, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11799, {{24901, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11800, {{24902, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11801, {{24903, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{11802, {{24904, 1, 0, 0},{24904, 1, 1, 18}, {0, 0, 0, 0}}},{11802, {{24904, 1, 0, 0},{24904, 1, 1, 18}, {0, 0, 0, 0}}},{11803, {{24904, 1, 0, 0},{24904, 1, 1, 20}, {0, 0, 0, 0}}},{11803, {{24904, 1, 0, 0},{24904, 1, 1, 20}, {0, 0, 0, 0}}},{11804, {{24904, 1, 0, 0},{24904, 1, 1, 19}, {0, 0, 0, 0}}},{11804, {{24904, 1, 0, 0},{24904, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{11806, {{24905, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11807, {{24906, 1, 0, 0},{24906, 1, 1, 17}, {0, 0, 0, 0}}},{11807, {{24906, 1, 0, 0},{24906, 1, 1, 17}, {0, 0, 0, 0}}},{11808, {{24906, 1, 0, 0},{24906, 1, 1, 18}, {0, 0, 0, 0}}},{11808, {{24906, 1, 0, 0},{24906, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11809, {{24907, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{11810, {{24908, 1, 0, 0},{24908, 1, 1, 18}, {0, 0, 0, 0}}},{11810, {{24908, 1, 0, 0},{24908, 1, 1, 18}, {0, 0, 0, 0}}},{11811, {{24908, 1, 0, 0},{24908, 1, 1, 20}, {0, 0, 0, 0}}},{11811, {{24908, 1, 0, 0},{24908, 1, 1, 20}, {0, 0, 0, 0}}},{11812, {{24908, 1, 0, 0},{24908, 1, 1, 19}, {0, 0, 0, 0}}},{11812, {{24908, 1, 0, 0},{24908, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{11813, {{24909, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11814, {{24910, 1, 0, 0},{24910, 1, 1, 18}, {0, 0, 0, 0}}},{11814, {{24910, 1, 0, 0},{24910, 1, 1, 18}, {0, 0, 0, 0}}},{11815, {{24910, 1, 0, 0},{24910, 1, 1, 19}, {0, 0, 0, 0}}},{11815, {{24910, 1, 0, 0},{24910, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{11816, {{24911, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{11817, {{24914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11818, {{24915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11819, {{24916, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11821, {{24917, 1, 0, 0},{24917, 1, 1, 17}, {0, 0, 0, 0}}},{11821, {{24917, 1, 0, 0},{24917, 1, 1, 17}, {0, 0, 0, 0}}},{11822, {{24917, 1, 0, 0},{24917, 1, 1, 18}, {0, 0, 0, 0}}},{11822, {{24917, 1, 0, 0},{24917, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11823, {{24918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11824, {{24919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11825, {{24920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11826, {{24921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11827, {{24922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11828, {{24923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11829, {{24924, 1, 0, 0},{24924, 1, 1, 17}, {0, 0, 0, 0}}},{11829, {{24924, 1, 0, 0},{24924, 1, 1, 17}, {0, 0, 0, 0}}},{11830, {{24924, 1, 0, 0},{24924, 1, 1, 16}, {0, 0, 0, 0}}},{11830, {{24924, 1, 0, 0},{24924, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {6, {{11831, {{24925, 1, 0, 0},{24925, 1, 1, 17}, {0, 0, 0, 0}}},{11831, {{24925, 1, 0, 0},{24925, 1, 1, 17}, {0, 0, 0, 0}}},{11832, {{24925, 1, 0, 0},{24925, 1, 1, 19}, {0, 0, 0, 0}}},{11832, {{24925, 1, 0, 0},{24925, 1, 1, 19}, {0, 0, 0, 0}}},{11833, {{24925, 1, 0, 0},{24925, 1, 1, 18}, {0, 0, 0, 0}}},{11833, {{24925, 1, 0, 0},{24925, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{11834, {{24926, 1, 0, 0},{24926, 1, 1, 17}, {0, 0, 0, 0}}},{11834, {{24926, 1, 0, 0},{24926, 1, 1, 17}, {0, 0, 0, 0}}},{11835, {{24926, 1, 0, 0},{24926, 1, 1, 18}, {0, 0, 0, 0}}},{11835, {{24926, 1, 0, 0},{24926, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11836, {{24927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11837, {{24928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11838, {{24929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11839, {{24930, 1, 0, 0},{24930, 1, 1, 19}, {0, 0, 0, 0}}},{11839, {{24930, 1, 0, 0},{24930, 1, 1, 19}, {0, 0, 0, 0}}},{11840, {{24930, 1, 0, 0},{24930, 1, 1, 18}, {0, 0, 0, 0}}},{11840, {{24930, 1, 0, 0},{24930, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11841, {{24931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11842, {{24932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{11843, {{24933, 1, 0, 0},{24933, 1, 1, 18}, {0, 0, 0, 0}}},{11843, {{24933, 1, 0, 0},{24933, 1, 1, 18}, {0, 0, 0, 0}}},{11844, {{24933, 1, 0, 0},{24933, 1, 1, 17}, {0, 0, 0, 0}}},{11844, {{24933, 1, 0, 0},{24933, 1, 1, 17}, {0, 0, 0, 0}}},{11845, {{24933, 1, 0, 0},{24933, 1, 1, 19}, {0, 0, 0, 0}}},{11845, {{24933, 1, 0, 0},{24933, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{11846, {{24934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{11847, {{24935, 1, 0, 0},{24935, 1, 1, 20}, {0, 0, 0, 0}}},{11847, {{24935, 1, 0, 0},{24935, 1, 1, 20}, {0, 0, 0, 0}}},{11848, {{24935, 1, 0, 0},{24935, 1, 1, 19}, {0, 0, 0, 0}}},{11848, {{24935, 1, 0, 0},{24935, 1, 1, 19}, {0, 0, 0, 0}}},{59630, {{24935, 1, 0, 0},{24935, 1, 1, 21}, {0, 0, 0, 0}}},{59630, {{24935, 1, 0, 0},{24935, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {19, {{11849, {{24936, 1, 0, 0},{24936, 1, 1, 22}, {0, 0, 0, 0}}},{11849, {{24936, 1, 0, 0},{24936, 1, 1, 22}, {0, 0, 0, 0}}},{11850, {{24936, 1, 0, 0},{24936, 1, 1, 29}, {0, 0, 0, 0}}},{11850, {{24936, 1, 0, 0},{24936, 1, 1, 29}, {0, 0, 0, 0}}},{11851, {{24936, 1, 0, 0},{24936, 1, 1, 26}, {0, 0, 0, 0}}},{11851, {{24936, 1, 0, 0},{24936, 1, 1, 26}, {0, 0, 0, 0}}},{11852, {{24936, 1, 0, 0},{24936, 1, 1, 25}, {0, 0, 0, 0}}},{11852, {{24936, 1, 0, 0},{24936, 1, 1, 25}, {0, 0, 0, 0}}},{11853, {{24936, 1, 0, 0},{24936, 1, 1, 23}, {0, 0, 0, 0}}},{11853, {{24936, 1, 0, 0},{24936, 1, 1, 23}, {0, 0, 0, 0}}},{11854, {{24936, 1, 0, 0},{24936, 1, 1, 30}, {0, 0, 0, 0}}},{11854, {{24936, 1, 0, 0},{24936, 1, 1, 30}, {0, 0, 0, 0}}},{11855, {{24936, 1, 0, 0},{24936, 1, 1, 31}, {0, 0, 0, 0}}},{11855, {{24936, 1, 0, 0},{24936, 1, 1, 31}, {0, 0, 0, 0}}},{11856, {{24936, 1, 0, 0},{24936, 1, 1, 27}, {0, 0, 0, 0}}},{11856, {{24936, 1, 0, 0},{24936, 1, 1, 27}, {0, 0, 0, 0}}},{30258, {{24936, 1, 0, 0},{24936, 1, 1, 0},{24936, 1, 1, 24},{64062, 1, 0, 0}, {0, 0, 0, 0}}},{30258, {{24936, 1, 0, 0},{24936, 1, 1, 0},{24936, 1, 1, 24},{64062, 1, 0, 0}, {0, 0, 0, 0}}},{30258, {{24936, 1, 0, 0},{24936, 1, 1, 0},{24936, 1, 1, 24},{64062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11858, {{24937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11859, {{24938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11860, {{24939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11861, {{24940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11862, {{24941, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11863, {{24942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11864, {{24943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11865, {{24944, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11866, {{24945, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11867, {{24946, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11868, {{24947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11869, {{24948, 1, 0, 0},{24948, 1, 1, 17}, {0, 0, 0, 0}}},{11869, {{24948, 1, 0, 0},{24948, 1, 1, 17}, {0, 0, 0, 0}}},{11870, {{24948, 1, 0, 0},{24948, 1, 1, 18}, {0, 0, 0, 0}}},{11870, {{24948, 1, 0, 0},{24948, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11871, {{24949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11872, {{24950, 1, 0, 0},{24950, 1, 1, 18}, {0, 0, 0, 0}}},{11872, {{24950, 1, 0, 0},{24950, 1, 1, 18}, {0, 0, 0, 0}}},{11873, {{24950, 1, 0, 0},{24950, 1, 1, 17}, {0, 0, 0, 0}}},{11873, {{24950, 1, 0, 0},{24950, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{11874, {{24951, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11875, {{24952, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11876, {{24953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{11877, {{24954, 1, 0, 0},{24954, 1, 1, 16}, {0, 0, 0, 0}}},{11877, {{24954, 1, 0, 0},{24954, 1, 1, 16}, {0, 0, 0, 0}}},{11878, {{24954, 1, 0, 0},{24954, 1, 1, 19}, {0, 0, 0, 0}}},{11878, {{24954, 1, 0, 0},{24954, 1, 1, 19}, {0, 0, 0, 0}}},{57518, {{24954, 1, 0, 0},{24954, 1, 1, 17}, {0, 0, 0, 0}}},{57518, {{24954, 1, 0, 0},{24954, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{11880, {{24955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11881, {{24956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11882, {{24957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11883, {{24958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11884, {{24959, 1, 0, 0},{24959, 1, 1, 17}, {0, 0, 0, 0}}},{11884, {{24959, 1, 0, 0},{24959, 1, 1, 17}, {0, 0, 0, 0}}},{11885, {{24959, 1, 0, 0},{24959, 1, 1, 18}, {0, 0, 0, 0}}},{11885, {{24959, 1, 0, 0},{24959, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11886, {{24960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11887, {{24961, 1, 0, 0},{24961, 1, 1, 17}, {0, 0, 0, 0}}},{11887, {{24961, 1, 0, 0},{24961, 1, 1, 17}, {0, 0, 0, 0}}},{11888, {{24961, 1, 0, 0},{24961, 1, 1, 18}, {0, 0, 0, 0}}},{11888, {{24961, 1, 0, 0},{24961, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{11890, {{24962, 1, 0, 0},{24962, 1, 1, 18}, {0, 0, 0, 0}}},{11890, {{24962, 1, 0, 0},{24962, 1, 1, 18}, {0, 0, 0, 0}}},{11891, {{24962, 1, 0, 0},{24962, 1, 1, 17}, {0, 0, 0, 0}}},{11891, {{24962, 1, 0, 0},{24962, 1, 1, 17}, {0, 0, 0, 0}}},{57524, {{24962, 1, 0, 0},{24962, 1, 1, 19}, {0, 0, 0, 0}}},{57524, {{24962, 1, 0, 0},{24962, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{11892, {{24963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11893, {{24964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11894, {{24965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11895, {{24966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11896, {{24967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11897, {{24968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11898, {{24969, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11899, {{24970, 1, 0, 0},{24970, 1, 1, 17}, {0, 0, 0, 0}}},{11899, {{24970, 1, 0, 0},{24970, 1, 1, 17}, {0, 0, 0, 0}}},{11900, {{24970, 1, 0, 0},{24970, 1, 1, 18}, {0, 0, 0, 0}}},{11900, {{24970, 1, 0, 0},{24970, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11901, {{24971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11902, {{24972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11903, {{24973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{11904, {{24974, 1, 0, 0},{24974, 1, 1, 18}, {0, 0, 0, 0}}},{11904, {{24974, 1, 0, 0},{24974, 1, 1, 18}, {0, 0, 0, 0}}},{30259, {{24974, 1, 0, 0},{24974, 1, 1, 0},{24974, 1, 1, 19},{64063, 1, 0, 0}, {0, 0, 0, 0}}},{30259, {{24974, 1, 0, 0},{24974, 1, 1, 0},{24974, 1, 1, 19},{64063, 1, 0, 0}, {0, 0, 0, 0}}},{30259, {{24974, 1, 0, 0},{24974, 1, 1, 0},{24974, 1, 1, 19},{64063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11905, {{24975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{11906, {{24976, 1, 0, 0},{24976, 1, 1, 19}, {0, 0, 0, 0}}},{11906, {{24976, 1, 0, 0},{24976, 1, 1, 19}, {0, 0, 0, 0}}},{11907, {{24976, 1, 0, 0},{24976, 1, 1, 18}, {0, 0, 0, 0}}},{11907, {{24976, 1, 0, 0},{24976, 1, 1, 18}, {0, 0, 0, 0}}},{11908, {{24976, 1, 0, 0},{24976, 1, 1, 20}, {0, 0, 0, 0}}},{11908, {{24976, 1, 0, 0},{24976, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{11909, {{24977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11910, {{24978, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11911, {{24979, 1, 0, 0},{24979, 1, 1, 17}, {0, 0, 0, 0}}},{11911, {{24979, 1, 0, 0},{24979, 1, 1, 17}, {0, 0, 0, 0}}},{11912, {{24979, 1, 0, 0},{24979, 1, 1, 18}, {0, 0, 0, 0}}},{11912, {{24979, 1, 0, 0},{24979, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11913, {{24980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11914, {{24981, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11915, {{24982, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11916, {{24983, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11917, {{24984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11918, {{24985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11919, {{24986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11920, {{24987, 1, 0, 0},{24987, 1, 1, 16}, {0, 0, 0, 0}}},{11920, {{24987, 1, 0, 0},{24987, 1, 1, 16}, {0, 0, 0, 0}}},{11921, {{24987, 1, 0, 0},{24987, 1, 1, 17}, {0, 0, 0, 0}}},{11921, {{24987, 1, 0, 0},{24987, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{11922, {{24988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11923, {{24989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11924, {{24990, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11925, {{24991, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11926, {{24992, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11927, {{24993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11928, {{24994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11929, {{24995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{11930, {{24996, 1, 0, 0},{24996, 1, 1, 18}, {0, 0, 0, 0}}},{11930, {{24996, 1, 0, 0},{24996, 1, 1, 18}, {0, 0, 0, 0}}},{11931, {{24996, 1, 0, 0},{24996, 1, 1, 19}, {0, 0, 0, 0}}},{11931, {{24996, 1, 0, 0},{24996, 1, 1, 19}, {0, 0, 0, 0}}},{11932, {{24996, 1, 0, 0},{24996, 1, 1, 20}, {0, 0, 0, 0}}},{11932, {{24996, 1, 0, 0},{24996, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{11933, {{24997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11934, {{24998, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11935, {{24999, 1, 0, 0},{24999, 1, 1, 17}, {0, 0, 0, 0}}},{11935, {{24999, 1, 0, 0},{24999, 1, 1, 17}, {0, 0, 0, 0}}},{11936, {{24999, 1, 0, 0},{24999, 1, 1, 18}, {0, 0, 0, 0}}},{11936, {{24999, 1, 0, 0},{24999, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11937, {{25000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11938, {{25001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11939, {{25002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11940, {{25003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11941, {{25004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11942, {{25005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11943, {{25006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11944, {{25007, 1, 0, 0},{25007, 1, 1, 17}, {0, 0, 0, 0}}},{11944, {{25007, 1, 0, 0},{25007, 1, 1, 17}, {0, 0, 0, 0}}},{11945, {{25007, 1, 0, 0},{25007, 1, 1, 16}, {0, 0, 0, 0}}},{11945, {{25007, 1, 0, 0},{25007, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{11946, {{25008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11947, {{25009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{11948, {{25010, 1, 0, 0},{25010, 1, 1, 21}, {0, 0, 0, 0}}},{11948, {{25010, 1, 0, 0},{25010, 1, 1, 21}, {0, 0, 0, 0}}},{11949, {{25010, 1, 0, 0},{25010, 1, 1, 19}, {0, 0, 0, 0}}},{11949, {{25010, 1, 0, 0},{25010, 1, 1, 19}, {0, 0, 0, 0}}},{11950, {{25010, 1, 0, 0},{25010, 1, 1, 22}, {0, 0, 0, 0}}},{11950, {{25010, 1, 0, 0},{25010, 1, 1, 22}, {0, 0, 0, 0}}},{11951, {{25010, 1, 0, 0},{25010, 1, 1, 20}, {0, 0, 0, 0}}},{11951, {{25010, 1, 0, 0},{25010, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{11952, {{25011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11953, {{25012, 1, 0, 0},{25012, 1, 1, 16}, {0, 0, 0, 0}}},{11953, {{25012, 1, 0, 0},{25012, 1, 1, 16}, {0, 0, 0, 0}}},{11954, {{25012, 1, 0, 0},{25012, 1, 1, 17}, {0, 0, 0, 0}}},{11954, {{25012, 1, 0, 0},{25012, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{11955, {{25013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11956, {{25014, 1, 0, 0},{25014, 1, 1, 17}, {0, 0, 0, 0}}},{11956, {{25014, 1, 0, 0},{25014, 1, 1, 17}, {0, 0, 0, 0}}},{11957, {{25014, 1, 0, 0},{25014, 1, 1, 18}, {0, 0, 0, 0}}},{11957, {{25014, 1, 0, 0},{25014, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11958, {{25015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11959, {{25016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11960, {{25017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11961, {{25018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11962, {{25019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11963, {{25020, 1, 0, 0},{25020, 1, 1, 17}, {0, 0, 0, 0}}},{11963, {{25020, 1, 0, 0},{25020, 1, 1, 17}, {0, 0, 0, 0}}},{11964, {{25020, 1, 0, 0},{25020, 1, 1, 18}, {0, 0, 0, 0}}},{11964, {{25020, 1, 0, 0},{25020, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11965, {{25021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11967, {{25022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11968, {{25023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11969, {{25024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11970, {{25025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11971, {{25026, 1, 0, 0},{25026, 1, 1, 17}, {0, 0, 0, 0}}},{11971, {{25026, 1, 0, 0},{25026, 1, 1, 17}, {0, 0, 0, 0}}},{11972, {{25026, 1, 0, 0},{25026, 1, 1, 18}, {0, 0, 0, 0}}},{11972, {{25026, 1, 0, 0},{25026, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11973, {{25027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11974, {{25028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11975, {{25029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11976, {{25030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11977, {{25031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11978, {{25032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11979, {{25033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11980, {{25034, 1, 0, 0},{25034, 1, 1, 17}, {0, 0, 0, 0}}},{11980, {{25034, 1, 0, 0},{25034, 1, 1, 17}, {0, 0, 0, 0}}},{11981, {{25034, 1, 0, 0},{25034, 1, 1, 18}, {0, 0, 0, 0}}},{11981, {{25034, 1, 0, 0},{25034, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{11982, {{25035, 1, 0, 0},{25035, 1, 1, 17}, {0, 0, 0, 0}}},{11982, {{25035, 1, 0, 0},{25035, 1, 1, 17}, {0, 0, 0, 0}}},{59634, {{25035, 1, 0, 0},{25035, 1, 1, 18}, {0, 0, 0, 0}}},{59634, {{25035, 1, 0, 0},{25035, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{11983, {{25036, 1, 0, 0},{25036, 1, 1, 17}, {0, 0, 0, 0}}},{11983, {{25036, 1, 0, 0},{25036, 1, 1, 17}, {0, 0, 0, 0}}},{11984, {{25036, 1, 0, 0},{25036, 1, 1, 18}, {0, 0, 0, 0}}},{11984, {{25036, 1, 0, 0},{25036, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11985, {{25037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11986, {{25038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11987, {{25039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11988, {{25040, 1, 0, 0},{25040, 1, 1, 17}, {0, 0, 0, 0}}},{11988, {{25040, 1, 0, 0},{25040, 1, 1, 17}, {0, 0, 0, 0}}},{57525, {{25040, 1, 0, 0},{25040, 1, 1, 18}, {0, 0, 0, 0}}},{57525, {{25040, 1, 0, 0},{25040, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{11989, {{25043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11990, {{25044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{11991, {{25045, 1, 0, 0},{25045, 1, 1, 17}, {0, 0, 0, 0}}},{11991, {{25045, 1, 0, 0},{25045, 1, 1, 17}, {0, 0, 0, 0}}},{11992, {{25045, 1, 0, 0},{25045, 1, 1, 18}, {0, 0, 0, 0}}},{11992, {{25045, 1, 0, 0},{25045, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{11993, {{25046, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11994, {{25047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11995, {{25048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11996, {{25049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11997, {{25050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11998, {{25051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{11999, {{25052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12000, {{25053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{12001, {{25054, 1, 0, 0},{25054, 1, 1, 17}, {0, 0, 0, 0}}},{12001, {{25054, 1, 0, 0},{25054, 1, 1, 17}, {0, 0, 0, 0}}},{12002, {{25054, 1, 0, 0},{25054, 1, 1, 18}, {0, 0, 0, 0}}},{12002, {{25054, 1, 0, 0},{25054, 1, 1, 18}, {0, 0, 0, 0}}},{12003, {{25054, 1, 0, 0},{25054, 1, 1, 19}, {0, 0, 0, 0}}},{12003, {{25054, 1, 0, 0},{25054, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{12004, {{25055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12005, {{25056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12006, {{25057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12007, {{25058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12008, {{25059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12009, {{25060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12010, {{25061, 1, 0, 0},{25061, 1, 1, 18}, {0, 0, 0, 0}}},{12010, {{25061, 1, 0, 0},{25061, 1, 1, 18}, {0, 0, 0, 0}}},{12011, {{25061, 1, 0, 0},{25061, 1, 1, 17}, {0, 0, 0, 0}}},{12011, {{25061, 1, 0, 0},{25061, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{12012, {{25062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12013, {{25063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12014, {{25064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12015, {{25065, 1, 0, 0},{25065, 1, 1, 17}, {0, 0, 0, 0}}},{12015, {{25065, 1, 0, 0},{25065, 1, 1, 17}, {0, 0, 0, 0}}},{12016, {{25065, 1, 0, 0},{25065, 1, 1, 18}, {0, 0, 0, 0}}},{12016, {{25065, 1, 0, 0},{25065, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12017, {{25066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12018, {{25067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12019, {{25068, 1, 0, 0},{25068, 1, 1, 17}, {0, 0, 0, 0}}},{12019, {{25068, 1, 0, 0},{25068, 1, 1, 17}, {0, 0, 0, 0}}},{12020, {{25068, 1, 0, 0},{25068, 1, 1, 18}, {0, 0, 0, 0}}},{12020, {{25068, 1, 0, 0},{25068, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12021, {{25069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12023, {{25070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12024, {{25071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12025, {{25072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12026, {{25073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {9, {{12027, {{25074, 1, 0, 0},{25074, 1, 1, 20}, {0, 0, 0, 0}}},{12027, {{25074, 1, 0, 0},{25074, 1, 1, 20}, {0, 0, 0, 0}}},{12028, {{25074, 1, 0, 0},{25074, 1, 1, 23}, {0, 0, 0, 0}}},{12028, {{25074, 1, 0, 0},{25074, 1, 1, 23}, {0, 0, 0, 0}}},{30260, {{25074, 1, 0, 0},{25074, 1, 1, 0},{25074, 1, 1, 21},{64064, 1, 0, 0}, {0, 0, 0, 0}}},{30260, {{25074, 1, 0, 0},{25074, 1, 1, 0},{25074, 1, 1, 21},{64064, 1, 0, 0}, {0, 0, 0, 0}}},{30260, {{25074, 1, 0, 0},{25074, 1, 1, 0},{25074, 1, 1, 21},{64064, 1, 0, 0}, {0, 0, 0, 0}}},{30261, {{25074, 1, 0, 0},{25074, 1, 1, 22}, {0, 0, 0, 0}}},{30261, {{25074, 1, 0, 0},{25074, 1, 1, 22}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{12029, {{25076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12030, {{25077, 1, 0, 0},{25077, 1, 1, 17}, {0, 0, 0, 0}}},{12030, {{25077, 1, 0, 0},{25077, 1, 1, 17}, {0, 0, 0, 0}}},{12031, {{25077, 1, 0, 0},{25077, 1, 1, 18}, {0, 0, 0, 0}}},{12031, {{25077, 1, 0, 0},{25077, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{12032, {{25078, 1, 0, 0},{25078, 1, 1, 18}, {0, 0, 0, 0}}},{12032, {{25078, 1, 0, 0},{25078, 1, 1, 18}, {0, 0, 0, 0}}},{12033, {{25078, 1, 0, 0},{25078, 1, 1, 17}, {0, 0, 0, 0}}},{12033, {{25078, 1, 0, 0},{25078, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{12034, {{25079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12035, {{25080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12036, {{25081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12037, {{25082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12038, {{25083, 1, 0, 0},{25083, 1, 1, 16}, {0, 0, 0, 0}}},{12038, {{25083, 1, 0, 0},{25083, 1, 1, 16}, {0, 0, 0, 0}}},{12039, {{25083, 1, 0, 0},{25083, 1, 1, 17}, {0, 0, 0, 0}}},{12039, {{25083, 1, 0, 0},{25083, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{12040, {{25084, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12041, {{25085, 1, 0, 0},{25085, 1, 1, 17}, {0, 0, 0, 0}}},{12041, {{25085, 1, 0, 0},{25085, 1, 1, 17}, {0, 0, 0, 0}}},{12042, {{25085, 1, 0, 0},{25085, 1, 1, 18}, {0, 0, 0, 0}}},{12042, {{25085, 1, 0, 0},{25085, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{12043, {{25086, 1, 0, 0},{25086, 1, 1, 18}, {0, 0, 0, 0}}},{12043, {{25086, 1, 0, 0},{25086, 1, 1, 18}, {0, 0, 0, 0}}},{12044, {{25086, 1, 0, 0},{25086, 1, 1, 19}, {0, 0, 0, 0}}},{12044, {{25086, 1, 0, 0},{25086, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{12045, {{25087, 1, 0, 0},{25087, 1, 1, 17}, {0, 0, 0, 0}}},{12045, {{25087, 1, 0, 0},{25087, 1, 1, 17}, {0, 0, 0, 0}}},{12046, {{25087, 1, 0, 0},{25087, 1, 1, 18}, {0, 0, 0, 0}}},{12046, {{25087, 1, 0, 0},{25087, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12047, {{25088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12048, {{25089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12049, {{25090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12050, {{25091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12051, {{25092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12052, {{25093, 1, 0, 0},{25093, 1, 1, 16}, {0, 0, 0, 0}}},{12052, {{25093, 1, 0, 0},{25093, 1, 1, 16}, {0, 0, 0, 0}}},{12053, {{25093, 1, 0, 0},{25093, 1, 1, 17}, {0, 0, 0, 0}}},{12053, {{25093, 1, 0, 0},{25093, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{12054, {{25095, 1, 0, 0},{25095, 1, 1, 17}, {0, 0, 0, 0}}},{12054, {{25095, 1, 0, 0},{25095, 1, 1, 17}, {0, 0, 0, 0}}},{12055, {{25095, 1, 0, 0},{25095, 1, 1, 18}, {0, 0, 0, 0}}},{12055, {{25095, 1, 0, 0},{25095, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12056, {{25096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12057, {{25097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12058, {{25098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{12059, {{25100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12060, {{25101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12061, {{25102, 1, 0, 0},{25102, 1, 1, 17}, {0, 0, 0, 0}}},{12061, {{25102, 1, 0, 0},{25102, 1, 1, 17}, {0, 0, 0, 0}}},{12062, {{25102, 1, 0, 0},{25102, 1, 1, 18}, {0, 0, 0, 0}}},{12062, {{25102, 1, 0, 0},{25102, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{12063, {{25104, 1, 0, 0},{25104, 1, 1, 18}, {0, 0, 0, 0}}},{12063, {{25104, 1, 0, 0},{25104, 1, 1, 18}, {0, 0, 0, 0}}},{12064, {{25104, 1, 0, 0},{25104, 1, 1, 19}, {0, 0, 0, 0}}},{12064, {{25104, 1, 0, 0},{25104, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{12065, {{25105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12066, {{25106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12067, {{25107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12068, {{25108, 1, 0, 0},{25108, 1, 1, 17}, {0, 0, 0, 0}}},{12068, {{25108, 1, 0, 0},{25108, 1, 1, 17}, {0, 0, 0, 0}}},{57535, {{25108, 1, 0, 0},{25108, 1, 1, 18}, {0, 0, 0, 0}}},{57535, {{25108, 1, 0, 0},{25108, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12069, {{25109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12070, {{25110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{12071, {{25112, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{12072, {{25114, 1, 0, 0},{25114, 1, 1, 17}, {0, 0, 0, 0}}},{12072, {{25114, 1, 0, 0},{25114, 1, 1, 17}, {0, 0, 0, 0}}},{57540, {{25114, 1, 0, 0},{25114, 1, 1, 18}, {0, 0, 0, 0}}},{57540, {{25114, 1, 0, 0},{25114, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{12073, {{25115, 1, 0, 0},{25115, 1, 1, 17}, {0, 0, 0, 0}}},{12073, {{25115, 1, 0, 0},{25115, 1, 1, 17}, {0, 0, 0, 0}}},{12074, {{25115, 1, 0, 0},{25115, 1, 1, 18}, {0, 0, 0, 0}}},{12074, {{25115, 1, 0, 0},{25115, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12075, {{25116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12076, {{25117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12077, {{25118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12078, {{25119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{12079, {{25120, 1, 0, 0},{25120, 1, 1, 18}, {0, 0, 0, 0}}},{12079, {{25120, 1, 0, 0},{25120, 1, 1, 18}, {0, 0, 0, 0}}},{12080, {{25120, 1, 0, 0},{25120, 1, 1, 17}, {0, 0, 0, 0}}},{12080, {{25120, 1, 0, 0},{25120, 1, 1, 17}, {0, 0, 0, 0}}},{12081, {{25120, 1, 0, 0},{25120, 1, 1, 19}, {0, 0, 0, 0}}},{12081, {{25120, 1, 0, 0},{25120, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{12082, {{25121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12083, {{25122, 1, 0, 0},{25122, 1, 1, 18}, {0, 0, 0, 0}}},{12083, {{25122, 1, 0, 0},{25122, 1, 1, 18}, {0, 0, 0, 0}}},{12084, {{25122, 1, 0, 0},{25122, 1, 1, 17}, {0, 0, 0, 0}}},{12084, {{25122, 1, 0, 0},{25122, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{12085, {{25123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12086, {{25124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12087, {{25125, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12088, {{25126, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12089, {{25127, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12090, {{25128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12091, {{25129, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12092, {{25130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12093, {{25131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12094, {{25132, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12095, {{25133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12096, {{25134, 1, 0, 0},{25134, 1, 1, 17}, {0, 0, 0, 0}}},{12096, {{25134, 1, 0, 0},{25134, 1, 1, 17}, {0, 0, 0, 0}}},{12097, {{25134, 1, 0, 0},{25134, 1, 1, 18}, {0, 0, 0, 0}}},{12097, {{25134, 1, 0, 0},{25134, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12098, {{25135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12099, {{25136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12100, {{25137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12101, {{25138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12102, {{25139, 1, 0, 0},{25139, 1, 1, 17}, {0, 0, 0, 0}}},{12102, {{25139, 1, 0, 0},{25139, 1, 1, 17}, {0, 0, 0, 0}}},{12103, {{25139, 1, 0, 0},{25139, 1, 1, 18}, {0, 0, 0, 0}}},{12103, {{25139, 1, 0, 0},{25139, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{12104, {{25140, 1, 0, 0},{25140, 1, 1, 17}, {0, 0, 0, 0}}},{12104, {{25140, 1, 0, 0},{25140, 1, 1, 17}, {0, 0, 0, 0}}},{12105, {{25140, 1, 0, 0},{25140, 1, 1, 18}, {0, 0, 0, 0}}},{12105, {{25140, 1, 0, 0},{25140, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12106, {{25141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12107, {{25142, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57544, {{25143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12108, {{25144, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12109, {{25145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12110, {{25146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12111, {{25147, 1, 0, 0},{25147, 1, 1, 19}, {0, 0, 0, 0}}},{12111, {{25147, 1, 0, 0},{25147, 1, 1, 19}, {0, 0, 0, 0}}},{12112, {{25147, 1, 0, 0},{25147, 1, 1, 18}, {0, 0, 0, 0}}},{12112, {{25147, 1, 0, 0},{25147, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12113, {{25148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12114, {{25149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12115, {{25150, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{12116, {{25151, 1, 0, 0},{25151, 1, 1, 19}, {0, 0, 0, 0}}},{12116, {{25151, 1, 0, 0},{25151, 1, 1, 19}, {0, 0, 0, 0}}},{12117, {{25151, 1, 0, 0},{25151, 1, 1, 18}, {0, 0, 0, 0}}},{12117, {{25151, 1, 0, 0},{25151, 1, 1, 18}, {0, 0, 0, 0}}},{12118, {{25151, 1, 0, 0},{25151, 1, 1, 20}, {0, 0, 0, 0}}},{12118, {{25151, 1, 0, 0},{25151, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{12119, {{25152, 1, 0, 0},{25152, 1, 1, 19}, {0, 0, 0, 0}}},{12119, {{25152, 1, 0, 0},{25152, 1, 1, 19}, {0, 0, 0, 0}}},{12120, {{25152, 1, 0, 0},{25152, 1, 1, 18}, {0, 0, 0, 0}}},{12120, {{25152, 1, 0, 0},{25152, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{12121, {{25153, 1, 0, 0},{25153, 1, 1, 19}, {0, 0, 0, 0}}},{12121, {{25153, 1, 0, 0},{25153, 1, 1, 19}, {0, 0, 0, 0}}},{12122, {{25153, 1, 0, 0},{25153, 1, 1, 18}, {0, 0, 0, 0}}},{12122, {{25153, 1, 0, 0},{25153, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12123, {{25154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12124, {{25155, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12125, {{25156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12126, {{25157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12127, {{25158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12129, {{25159, 1, 0, 0},{25159, 1, 1, 19}, {0, 0, 0, 0}}},{12129, {{25159, 1, 0, 0},{25159, 1, 1, 19}, {0, 0, 0, 0}}},{12130, {{25159, 1, 0, 0},{25159, 1, 1, 18}, {0, 0, 0, 0}}},{12130, {{25159, 1, 0, 0},{25159, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{12131, {{25160, 1, 0, 0},{25160, 1, 1, 19}, {0, 0, 0, 0}}},{12131, {{25160, 1, 0, 0},{25160, 1, 1, 19}, {0, 0, 0, 0}}},{12132, {{25160, 1, 0, 0},{25160, 1, 1, 18}, {0, 0, 0, 0}}},{12132, {{25160, 1, 0, 0},{25160, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {8, {{12133, {{25161, 1, 0, 0},{25161, 1, 1, 21}, {0, 0, 0, 0}}},{12133, {{25161, 1, 0, 0},{25161, 1, 1, 21}, {0, 0, 0, 0}}},{12134, {{25161, 1, 0, 0},{25161, 1, 1, 20}, {0, 0, 0, 0}}},{12134, {{25161, 1, 0, 0},{25161, 1, 1, 20}, {0, 0, 0, 0}}},{12135, {{25161, 1, 0, 0},{25161, 1, 1, 22}, {0, 0, 0, 0}}},{12135, {{25161, 1, 0, 0},{25161, 1, 1, 22}, {0, 0, 0, 0}}},{12136, {{25161, 1, 0, 0},{25161, 1, 1, 23}, {0, 0, 0, 0}}},{12136, {{25161, 1, 0, 0},{25161, 1, 1, 23}, {0, 0, 0, 0}}}}},
    {1, {{12137, {{25162, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12138, {{25163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12139, {{25164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12140, {{25165, 1, 0, 0},{25165, 1, 1, 17}, {0, 0, 0, 0}}},{12140, {{25165, 1, 0, 0},{25165, 1, 1, 17}, {0, 0, 0, 0}}},{59295, {{25165, 1, 0, 0},{25165, 1, 1, 18}, {0, 0, 0, 0}}},{59295, {{25165, 1, 0, 0},{25165, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12141, {{25166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12142, {{25167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12143, {{25168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12144, {{25169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12145, {{25170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12146, {{25171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12147, {{25172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12148, {{25173, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12149, {{25174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12150, {{25175, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12151, {{25176, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12152, {{25177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12153, {{25178, 1, 0, 0},{25178, 1, 1, 17}, {0, 0, 0, 0}}},{12153, {{25178, 1, 0, 0},{25178, 1, 1, 17}, {0, 0, 0, 0}}},{12154, {{25178, 1, 0, 0},{25178, 1, 1, 18}, {0, 0, 0, 0}}},{12154, {{25178, 1, 0, 0},{25178, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12155, {{25179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12156, {{25180, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12157, {{25181, 1, 0, 0},{25181, 1, 1, 16}, {0, 0, 0, 0}}},{12157, {{25181, 1, 0, 0},{25181, 1, 1, 16}, {0, 0, 0, 0}}},{12158, {{25181, 1, 0, 0},{25181, 1, 1, 17}, {0, 0, 0, 0}}},{12158, {{25181, 1, 0, 0},{25181, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{12159, {{25182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12160, {{25183, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12161, {{25184, 1, 0, 0},{25184, 1, 1, 17}, {0, 0, 0, 0}}},{12161, {{25184, 1, 0, 0},{25184, 1, 1, 17}, {0, 0, 0, 0}}},{12162, {{25184, 1, 0, 0},{25184, 1, 1, 18}, {0, 0, 0, 0}}},{12162, {{25184, 1, 0, 0},{25184, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12163, {{25185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12164, {{25186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12165, {{25187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12166, {{25188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12167, {{25189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12168, {{25190, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12169, {{25191, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{12170, {{25192, 1, 0, 0},{25192, 1, 1, 19}, {0, 0, 0, 0}}},{12170, {{25192, 1, 0, 0},{25192, 1, 1, 19}, {0, 0, 0, 0}}},{12171, {{25192, 1, 0, 0},{25192, 1, 1, 20}, {0, 0, 0, 0}}},{12171, {{25192, 1, 0, 0},{25192, 1, 1, 20}, {0, 0, 0, 0}}},{12172, {{25192, 1, 0, 0},{25192, 1, 1, 21}, {0, 0, 0, 0}}},{12172, {{25192, 1, 0, 0},{25192, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{12173, {{25197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12174, {{25198, 1, 0, 0},{25198, 1, 1, 17}, {0, 0, 0, 0}}},{12174, {{25198, 1, 0, 0},{25198, 1, 1, 17}, {0, 0, 0, 0}}},{12175, {{25198, 1, 0, 0},{25198, 1, 1, 18}, {0, 0, 0, 0}}},{12175, {{25198, 1, 0, 0},{25198, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12176, {{25199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12177, {{25200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12178, {{25201, 1, 0, 0},{25201, 1, 1, 18}, {0, 0, 0, 0}}},{12178, {{25201, 1, 0, 0},{25201, 1, 1, 18}, {0, 0, 0, 0}}},{12179, {{25201, 1, 0, 0},{25201, 1, 1, 19}, {0, 0, 0, 0}}},{12179, {{25201, 1, 0, 0},{25201, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{12180, {{25202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12181, {{25203, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12182, {{25204, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12183, {{25205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12184, {{25206, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12185, {{25207, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12186, {{25208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12187, {{25209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12188, {{25210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12189, {{25211, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12190, {{25212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12191, {{25213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12192, {{25214, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12193, {{25215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12194, {{25216, 1, 0, 0},{25216, 1, 1, 17}, {0, 0, 0, 0}}},{12194, {{25216, 1, 0, 0},{25216, 1, 1, 17}, {0, 0, 0, 0}}},{12195, {{25216, 1, 0, 0},{25216, 1, 1, 18}, {0, 0, 0, 0}}},{12195, {{25216, 1, 0, 0},{25216, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12196, {{25217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12197, {{25218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12198, {{25219, 1, 0, 0},{25219, 1, 1, 17}, {0, 0, 0, 0}}},{12198, {{25219, 1, 0, 0},{25219, 1, 1, 17}, {0, 0, 0, 0}}},{12199, {{25219, 1, 0, 0},{25219, 1, 1, 18}, {0, 0, 0, 0}}},{12199, {{25219, 1, 0, 0},{25219, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12200, {{25220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12201, {{25221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12202, {{25222, 1, 0, 0},{25222, 1, 1, 16}, {0, 0, 0, 0}}},{12202, {{25222, 1, 0, 0},{25222, 1, 1, 16}, {0, 0, 0, 0}}},{12203, {{25222, 1, 0, 0},{25222, 1, 1, 17}, {0, 0, 0, 0}}},{12203, {{25222, 1, 0, 0},{25222, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{12204, {{25224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12205, {{25225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12206, {{25226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12207, {{25227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12208, {{25228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12209, {{25229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12210, {{25230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12211, {{25231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12212, {{25232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12213, {{25233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12214, {{25234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12215, {{25235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12216, {{25236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12217, {{25237, 1, 0, 0},{25237, 1, 1, 17}, {0, 0, 0, 0}}},{12217, {{25237, 1, 0, 0},{25237, 1, 1, 17}, {0, 0, 0, 0}}},{57553, {{25237, 1, 0, 0},{25237, 1, 1, 18}, {0, 0, 0, 0}}},{57553, {{25237, 1, 0, 0},{25237, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12218, {{25238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12219, {{25239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12220, {{25240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12221, {{25241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12222, {{25242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12223, {{25243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12224, {{25244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12225, {{25245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12226, {{25246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{12228, {{25253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12229, {{25254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12230, {{25255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12231, {{25256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12233, {{25257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12234, {{25258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12235, {{25259, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12236, {{25260, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12237, {{25261, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12238, {{25262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12239, {{25263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12240, {{25264, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12241, {{25265, 1, 0, 0},{25265, 1, 1, 19}, {0, 0, 0, 0}}},{12241, {{25265, 1, 0, 0},{25265, 1, 1, 19}, {0, 0, 0, 0}}},{12242, {{25265, 1, 0, 0},{25265, 1, 1, 18}, {0, 0, 0, 0}}},{12242, {{25265, 1, 0, 0},{25265, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12243, {{25266, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12244, {{25267, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12245, {{25268, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12246, {{25269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12247, {{25270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12248, {{25271, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12249, {{25272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12250, {{25273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12251, {{25274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12252, {{25275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12253, {{25276, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12254, {{25277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12255, {{25278, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12256, {{25279, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{12257, {{25282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12258, {{25283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12259, {{25284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12260, {{25285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12261, {{25286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12262, {{25287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12263, {{25288, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12264, {{25289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12265, {{25290, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12266, {{25291, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12267, {{25292, 1, 0, 0},{25292, 1, 1, 18}, {0, 0, 0, 0}}},{12267, {{25292, 1, 0, 0},{25292, 1, 1, 18}, {0, 0, 0, 0}}},{12268, {{25292, 1, 0, 0},{25292, 1, 1, 19}, {0, 0, 0, 0}}},{12268, {{25292, 1, 0, 0},{25292, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{12269, {{25293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12270, {{25294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12271, {{25295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{12272, {{25296, 1, 0, 0},{25296, 1, 1, 18}, {0, 0, 0, 0}}},{12272, {{25296, 1, 0, 0},{25296, 1, 1, 18}, {0, 0, 0, 0}}},{12273, {{25296, 1, 0, 0},{25296, 1, 1, 19}, {0, 0, 0, 0}}},{12273, {{25296, 1, 0, 0},{25296, 1, 1, 19}, {0, 0, 0, 0}}},{12274, {{25296, 1, 0, 0},{25296, 1, 1, 20}, {0, 0, 0, 0}}},{12274, {{25296, 1, 0, 0},{25296, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{12275, {{25297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12276, {{25298, 1, 0, 0},{25298, 1, 1, 19}, {0, 0, 0, 0}}},{12276, {{25298, 1, 0, 0},{25298, 1, 1, 19}, {0, 0, 0, 0}}},{12277, {{25298, 1, 0, 0},{25298, 1, 1, 18}, {0, 0, 0, 0}}},{12277, {{25298, 1, 0, 0},{25298, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12278, {{25299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12279, {{25300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12281, {{25301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12282, {{25302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12283, {{25303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12284, {{25304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12285, {{25305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12286, {{25306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12287, {{25307, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12288, {{25308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12289, {{25309, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12290, {{25310, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12291, {{25311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12292, {{25312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12293, {{25313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{12294, {{25322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12295, {{25323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12296, {{25324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12297, {{25325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12298, {{25326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12299, {{25327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12300, {{25328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12301, {{25329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12302, {{25330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12303, {{25331, 1, 0, 0},{25331, 1, 1, 19}, {0, 0, 0, 0}}},{12303, {{25331, 1, 0, 0},{25331, 1, 1, 19}, {0, 0, 0, 0}}},{12304, {{25331, 1, 0, 0},{25331, 1, 1, 18}, {0, 0, 0, 0}}},{12304, {{25331, 1, 0, 0},{25331, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{12305, {{25332, 1, 0, 0},{25332, 1, 1, 17}, {0, 0, 0, 0}}},{12305, {{25332, 1, 0, 0},{25332, 1, 1, 17}, {0, 0, 0, 0}}},{12306, {{25332, 1, 0, 0},{25332, 1, 1, 18}, {0, 0, 0, 0}}},{12306, {{25332, 1, 0, 0},{25332, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12307, {{25333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12308, {{25334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12309, {{25335, 1, 0, 0},{25335, 1, 1, 18}, {0, 0, 0, 0}}},{12309, {{25335, 1, 0, 0},{25335, 1, 1, 18}, {0, 0, 0, 0}}},{12310, {{25335, 1, 0, 0},{25335, 1, 1, 19}, {0, 0, 0, 0}}},{12310, {{25335, 1, 0, 0},{25335, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{12311, {{25336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12312, {{25337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12313, {{25338, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12314, {{25339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12315, {{25340, 1, 0, 0},{25340, 1, 1, 17}, {0, 0, 0, 0}}},{12315, {{25340, 1, 0, 0},{25340, 1, 1, 17}, {0, 0, 0, 0}}},{12316, {{25340, 1, 0, 0},{25340, 1, 1, 18}, {0, 0, 0, 0}}},{12316, {{25340, 1, 0, 0},{25340, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12317, {{25341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12318, {{25342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12319, {{25343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12320, {{25344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12321, {{25345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12322, {{25346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12323, {{25347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12324, {{25348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12325, {{25349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12326, {{25350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12327, {{25351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{12328, {{25352, 1, 0, 0},{25352, 1, 1, 17}, {0, 0, 0, 0}}},{12328, {{25352, 1, 0, 0},{25352, 1, 1, 17}, {0, 0, 0, 0}}},{12329, {{25352, 1, 0, 0},{25352, 1, 1, 18}, {0, 0, 0, 0}}},{12329, {{25352, 1, 0, 0},{25352, 1, 1, 18}, {0, 0, 0, 0}}},{12330, {{25352, 1, 0, 0},{25352, 1, 1, 20}, {0, 0, 0, 0}}},{12330, {{25352, 1, 0, 0},{25352, 1, 1, 20}, {0, 0, 0, 0}}},{12331, {{25352, 1, 0, 0},{25352, 1, 1, 19}, {0, 0, 0, 0}}},{12331, {{25352, 1, 0, 0},{25352, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{12332, {{25353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12333, {{25354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12334, {{25355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12335, {{25356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12336, {{25357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12337, {{25358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12338, {{25359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12339, {{25360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12340, {{25361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12341, {{25362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12342, {{25363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12343, {{25364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12344, {{25365, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12345, {{25366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12346, {{25367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12347, {{25368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12348, {{25369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{12349, {{25375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{12350, {{25383, 1, 0, 0},{25383, 1, 1, 17}, {0, 0, 0, 0}}},{12350, {{25383, 1, 0, 0},{25383, 1, 1, 17}, {0, 0, 0, 0}}},{12351, {{25383, 1, 0, 0},{25383, 1, 1, 18}, {0, 0, 0, 0}}},{12351, {{25383, 1, 0, 0},{25383, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12352, {{25384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12353, {{25385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12355, {{25386, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12356, {{25387, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12357, {{25388, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12358, {{25389, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12359, {{25390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12360, {{25391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12361, {{25392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12362, {{25393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12363, {{25394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12364, {{25395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12365, {{25396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12366, {{25397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12367, {{25398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12368, {{25399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12369, {{25400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12370, {{25401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12371, {{25402, 1, 0, 0},{25402, 1, 1, 18}, {0, 0, 0, 0}}},{12371, {{25402, 1, 0, 0},{25402, 1, 1, 18}, {0, 0, 0, 0}}},{12372, {{25402, 1, 0, 0},{25402, 1, 1, 19}, {0, 0, 0, 0}}},{12372, {{25402, 1, 0, 0},{25402, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{12373, {{25403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12374, {{25404, 1, 0, 0},{25404, 1, 1, 17}, {0, 0, 0, 0}}},{12374, {{25404, 1, 0, 0},{25404, 1, 1, 17}, {0, 0, 0, 0}}},{12375, {{25404, 1, 0, 0},{25404, 1, 1, 18}, {0, 0, 0, 0}}},{12375, {{25404, 1, 0, 0},{25404, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{12376, {{25405, 1, 0, 0},{25405, 1, 1, 19}, {0, 0, 0, 0}}},{12376, {{25405, 1, 0, 0},{25405, 1, 1, 19}, {0, 0, 0, 0}}},{12377, {{25405, 1, 0, 0},{25405, 1, 1, 18}, {0, 0, 0, 0}}},{12377, {{25405, 1, 0, 0},{25405, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12378, {{25406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12379, {{25407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12380, {{25408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12381, {{25409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12382, {{25410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12383, {{25411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12384, {{25412, 1, 0, 0},{25412, 1, 1, 17}, {0, 0, 0, 0}}},{12384, {{25412, 1, 0, 0},{25412, 1, 1, 17}, {0, 0, 0, 0}}},{12385, {{25412, 1, 0, 0},{25412, 1, 1, 18}, {0, 0, 0, 0}}},{12385, {{25412, 1, 0, 0},{25412, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12386, {{25413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12387, {{25414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12388, {{25415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12389, {{25416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12390, {{25417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12391, {{25418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12393, {{25419, 1, 0, 0},{25419, 1, 1, 17}, {0, 0, 0, 0}}},{12393, {{25419, 1, 0, 0},{25419, 1, 1, 17}, {0, 0, 0, 0}}},{12394, {{25419, 1, 0, 0},{25419, 1, 1, 18}, {0, 0, 0, 0}}},{12394, {{25419, 1, 0, 0},{25419, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12396, {{25420, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12397, {{25421, 1, 0, 0},{25421, 1, 1, 17}, {0, 0, 0, 0}}},{12397, {{25421, 1, 0, 0},{25421, 1, 1, 17}, {0, 0, 0, 0}}},{12398, {{25421, 1, 0, 0},{25421, 1, 1, 18}, {0, 0, 0, 0}}},{12398, {{25421, 1, 0, 0},{25421, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12399, {{25422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12401, {{25423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12403, {{25424, 1, 0, 0},{25424, 1, 1, 18}, {0, 0, 0, 0}}},{12403, {{25424, 1, 0, 0},{25424, 1, 1, 18}, {0, 0, 0, 0}}},{12404, {{25424, 1, 0, 0},{25424, 1, 1, 17}, {0, 0, 0, 0}}},{12404, {{25424, 1, 0, 0},{25424, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{12405, {{25425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12406, {{25426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12407, {{25427, 1, 0, 0},{25427, 1, 1, 18}, {0, 0, 0, 0}}},{12407, {{25427, 1, 0, 0},{25427, 1, 1, 18}, {0, 0, 0, 0}}},{12408, {{25427, 1, 0, 0},{25427, 1, 1, 17}, {0, 0, 0, 0}}},{12408, {{25427, 1, 0, 0},{25427, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{12409, {{25428, 1, 0, 0},{25428, 1, 1, 17}, {0, 0, 0, 0}}},{12409, {{25428, 1, 0, 0},{25428, 1, 1, 17}, {0, 0, 0, 0}}},{12410, {{25428, 1, 0, 0},{25428, 1, 1, 18}, {0, 0, 0, 0}}},{12410, {{25428, 1, 0, 0},{25428, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12411, {{25429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12412, {{25430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12413, {{25431, 1, 0, 0},{25431, 1, 1, 19}, {0, 0, 0, 0}}},{12413, {{25431, 1, 0, 0},{25431, 1, 1, 19}, {0, 0, 0, 0}}},{12414, {{25431, 1, 0, 0},{25431, 1, 1, 18}, {0, 0, 0, 0}}},{12414, {{25431, 1, 0, 0},{25431, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12415, {{25432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12416, {{25433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12417, {{25434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12418, {{25435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12419, {{25436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{12420, {{25444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12421, {{25445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12422, {{25446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12423, {{25447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12424, {{25448, 1, 0, 0},{25448, 1, 1, 19}, {0, 0, 0, 0}}},{12424, {{25448, 1, 0, 0},{25448, 1, 1, 19}, {0, 0, 0, 0}}},{12425, {{25448, 1, 0, 0},{25448, 1, 1, 18}, {0, 0, 0, 0}}},{12425, {{25448, 1, 0, 0},{25448, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{12426, {{25449, 1, 0, 0},{25449, 1, 1, 18}, {0, 0, 0, 0}}},{12426, {{25449, 1, 0, 0},{25449, 1, 1, 18}, {0, 0, 0, 0}}},{12427, {{25449, 1, 0, 0},{25449, 1, 1, 19}, {0, 0, 0, 0}}},{12427, {{25449, 1, 0, 0},{25449, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{12428, {{25450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12429, {{25451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12430, {{25452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12431, {{25453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12432, {{25454, 1, 0, 0},{25454, 1, 1, 18}, {0, 0, 0, 0}}},{12432, {{25454, 1, 0, 0},{25454, 1, 1, 18}, {0, 0, 0, 0}}},{12433, {{25454, 1, 0, 0},{25454, 1, 1, 19}, {0, 0, 0, 0}}},{12433, {{25454, 1, 0, 0},{25454, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{12434, {{25455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12435, {{25456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12436, {{25457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12437, {{25458, 1, 0, 0},{25458, 1, 1, 19}, {0, 0, 0, 0}}},{12437, {{25458, 1, 0, 0},{25458, 1, 1, 19}, {0, 0, 0, 0}}},{12438, {{25458, 1, 0, 0},{25458, 1, 1, 18}, {0, 0, 0, 0}}},{12438, {{25458, 1, 0, 0},{25458, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12439, {{25459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12440, {{25460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12441, {{25461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12442, {{25462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12443, {{25463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12444, {{25464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12445, {{25465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12446, {{25466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12447, {{25467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12448, {{25468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12449, {{25469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12450, {{25470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12451, {{25471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12452, {{25472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12453, {{25473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12454, {{25474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12455, {{25475, 1, 0, 0},{25475, 1, 1, 19}, {0, 0, 0, 0}}},{12455, {{25475, 1, 0, 0},{25475, 1, 1, 19}, {0, 0, 0, 0}}},{12456, {{25475, 1, 0, 0},{25475, 1, 1, 18}, {0, 0, 0, 0}}},{12456, {{25475, 1, 0, 0},{25475, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12457, {{25476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12458, {{25477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12459, {{25478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12460, {{25479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12461, {{25480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12462, {{25481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12463, {{25482, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12464, {{25483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12465, {{25484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12466, {{25485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12467, {{25486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12468, {{25487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12469, {{25488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12470, {{25489, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12471, {{25490, 1, 0, 0},{25490, 1, 1, 18}, {0, 0, 0, 0}}},{12471, {{25490, 1, 0, 0},{25490, 1, 1, 18}, {0, 0, 0, 0}}},{12472, {{25490, 1, 0, 0},{25490, 1, 1, 19}, {0, 0, 0, 0}}},{12472, {{25490, 1, 0, 0},{25490, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{12473, {{25491, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12474, {{25492, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12475, {{25493, 1, 0, 0},{25493, 1, 1, 18}, {0, 0, 0, 0}}},{12475, {{25493, 1, 0, 0},{25493, 1, 1, 18}, {0, 0, 0, 0}}},{12476, {{25493, 1, 0, 0},{25493, 1, 1, 17}, {0, 0, 0, 0}}},{12476, {{25493, 1, 0, 0},{25493, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{12477, {{25494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12478, {{25495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12479, {{25496, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12480, {{25497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12481, {{25498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12482, {{25499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12483, {{25500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12484, {{25501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12485, {{25502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12486, {{25503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12487, {{25504, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12488, {{25505, 1, 0, 0},{25505, 1, 1, 19}, {0, 0, 0, 0}}},{12488, {{25505, 1, 0, 0},{25505, 1, 1, 19}, {0, 0, 0, 0}}},{12489, {{25505, 1, 0, 0},{25505, 1, 1, 18}, {0, 0, 0, 0}}},{12489, {{25505, 1, 0, 0},{25505, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12490, {{25506, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12491, {{25507, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12492, {{25508, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12493, {{25509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12494, {{25510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12495, {{25511, 1, 0, 0},{25511, 1, 1, 18}, {0, 0, 0, 0}}},{12495, {{25511, 1, 0, 0},{25511, 1, 1, 18}, {0, 0, 0, 0}}},{12496, {{25511, 1, 0, 0},{25511, 1, 1, 19}, {0, 0, 0, 0}}},{12496, {{25511, 1, 0, 0},{25511, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{12497, {{25512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12498, {{25513, 1, 0, 0},{25513, 1, 1, 17}, {0, 0, 0, 0}}},{12498, {{25513, 1, 0, 0},{25513, 1, 1, 17}, {0, 0, 0, 0}}},{12499, {{25513, 1, 0, 0},{25513, 1, 1, 18}, {0, 0, 0, 0}}},{12499, {{25513, 1, 0, 0},{25513, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12500, {{25514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12501, {{25515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12502, {{25516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12503, {{25517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12504, {{25518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12505, {{25519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12506, {{25520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12507, {{25521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12508, {{25522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{12509, {{25524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12510, {{25525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{12511, {{25531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{12512, {{25533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{12513, {{25534, 1, 0, 0},{25534, 1, 1, 17}, {0, 0, 0, 0}}},{12513, {{25534, 1, 0, 0},{25534, 1, 1, 17}, {0, 0, 0, 0}}},{12514, {{25534, 1, 0, 0},{25534, 1, 1, 18}, {0, 0, 0, 0}}},{12514, {{25534, 1, 0, 0},{25534, 1, 1, 18}, {0, 0, 0, 0}}},{12515, {{25534, 1, 0, 0},{25534, 1, 1, 19}, {0, 0, 0, 0}}},{12515, {{25534, 1, 0, 0},{25534, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{12516, {{25535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12517, {{25536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12518, {{25537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12519, {{25538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12521, {{25539, 1, 0, 0},{25539, 1, 1, 19}, {0, 0, 0, 0}}},{12521, {{25539, 1, 0, 0},{25539, 1, 1, 19}, {0, 0, 0, 0}}},{12522, {{25539, 1, 0, 0},{25539, 1, 1, 18}, {0, 0, 0, 0}}},{12522, {{25539, 1, 0, 0},{25539, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{12523, {{25540, 1, 0, 0},{25540, 1, 1, 17}, {0, 0, 0, 0}}},{12523, {{25540, 1, 0, 0},{25540, 1, 1, 17}, {0, 0, 0, 0}}},{12524, {{25540, 1, 0, 0},{25540, 1, 1, 18}, {0, 0, 0, 0}}},{12524, {{25540, 1, 0, 0},{25540, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{12525, {{25541, 1, 0, 0},{25541, 1, 1, 17}, {0, 0, 0, 0}}},{12525, {{25541, 1, 0, 0},{25541, 1, 1, 17}, {0, 0, 0, 0}}},{12526, {{25541, 1, 0, 0},{25541, 1, 1, 19}, {0, 0, 0, 0}}},{12526, {{25541, 1, 0, 0},{25541, 1, 1, 19}, {0, 0, 0, 0}}},{12527, {{25541, 1, 0, 0},{25541, 1, 1, 18}, {0, 0, 0, 0}}},{12527, {{25541, 1, 0, 0},{25541, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12528, {{25542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12529, {{25543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12530, {{25544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12531, {{25545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12532, {{25546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12533, {{25547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12534, {{25548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12535, {{25549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12536, {{25550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12537, {{25551, 1, 0, 0},{25551, 1, 1, 17}, {0, 0, 0, 0}}},{12537, {{25551, 1, 0, 0},{25551, 1, 1, 17}, {0, 0, 0, 0}}},{12538, {{25551, 1, 0, 0},{25551, 1, 1, 18}, {0, 0, 0, 0}}},{12538, {{25551, 1, 0, 0},{25551, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12539, {{25552, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12402, {{25553, 1, 0, 0},{25553, 1, 1, 18}, {0, 0, 0, 0}}},{12402, {{25553, 1, 0, 0},{25553, 1, 1, 18}, {0, 0, 0, 0}}},{12540, {{25553, 1, 0, 0},{25553, 1, 1, 17}, {0, 0, 0, 0}}},{12540, {{25553, 1, 0, 0},{25553, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{12541, {{25554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12542, {{25555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12543, {{25556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12544, {{25557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12545, {{25558, 1, 0, 0},{25558, 1, 1, 17}, {0, 0, 0, 0}}},{12545, {{25558, 1, 0, 0},{25558, 1, 1, 17}, {0, 0, 0, 0}}},{12546, {{25558, 1, 0, 0},{25558, 1, 1, 18}, {0, 0, 0, 0}}},{12546, {{25558, 1, 0, 0},{25558, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12547, {{25559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12548, {{25560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12549, {{25561, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12550, {{25562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12551, {{25563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12552, {{25564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12553, {{25565, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12554, {{25566, 1, 0, 0},{25566, 1, 1, 16}, {0, 0, 0, 0}}},{12554, {{25566, 1, 0, 0},{25566, 1, 1, 16}, {0, 0, 0, 0}}},{12555, {{25566, 1, 0, 0},{25566, 1, 1, 17}, {0, 0, 0, 0}}},{12555, {{25566, 1, 0, 0},{25566, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{12556, {{25567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12557, {{25568, 1, 0, 0},{25568, 1, 1, 17}, {0, 0, 0, 0}}},{12557, {{25568, 1, 0, 0},{25568, 1, 1, 17}, {0, 0, 0, 0}}},{12558, {{25568, 1, 0, 0},{25568, 1, 1, 18}, {0, 0, 0, 0}}},{12558, {{25568, 1, 0, 0},{25568, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12559, {{25569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12560, {{25570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12561, {{25571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12562, {{25572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12563, {{25573, 1, 0, 0},{25573, 1, 1, 17}, {0, 0, 0, 0}}},{12563, {{25573, 1, 0, 0},{25573, 1, 1, 17}, {0, 0, 0, 0}}},{12564, {{25573, 1, 0, 0},{25573, 1, 1, 18}, {0, 0, 0, 0}}},{12564, {{25573, 1, 0, 0},{25573, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12565, {{25574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12566, {{25575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12567, {{25576, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12568, {{25577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12569, {{25578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12570, {{25579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12571, {{25580, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12573, {{25581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12574, {{25582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12575, {{25583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12576, {{25584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12577, {{25585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12578, {{25586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{12579, {{25587, 1, 0, 0},{25587, 1, 1, 17}, {0, 0, 0, 0}}},{12579, {{25587, 1, 0, 0},{25587, 1, 1, 17}, {0, 0, 0, 0}}},{12580, {{25587, 1, 0, 0},{25587, 1, 1, 18}, {0, 0, 0, 0}}},{12580, {{25587, 1, 0, 0},{25587, 1, 1, 18}, {0, 0, 0, 0}}},{57561, {{25587, 1, 0, 0},{25587, 1, 1, 19}, {0, 0, 0, 0}}},{57561, {{25587, 1, 0, 0},{25587, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{12581, {{25588, 1, 0, 0},{25588, 1, 1, 19}, {0, 0, 0, 0}}},{12581, {{25588, 1, 0, 0},{25588, 1, 1, 19}, {0, 0, 0, 0}}},{12582, {{25588, 1, 0, 0},{25588, 1, 1, 18}, {0, 0, 0, 0}}},{12582, {{25588, 1, 0, 0},{25588, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{12583, {{25589, 1, 0, 0},{25589, 1, 1, 17}, {0, 0, 0, 0}}},{12583, {{25589, 1, 0, 0},{25589, 1, 1, 17}, {0, 0, 0, 0}}},{12584, {{25589, 1, 0, 0},{25589, 1, 1, 18}, {0, 0, 0, 0}}},{12584, {{25589, 1, 0, 0},{25589, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12585, {{25590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12586, {{25591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12587, {{25592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12588, {{25593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12589, {{25594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{12590, {{25598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{12591, {{25606, 1, 0, 0},{25606, 1, 1, 18}, {0, 0, 0, 0}}},{12591, {{25606, 1, 0, 0},{25606, 1, 1, 18}, {0, 0, 0, 0}}},{12592, {{25606, 1, 0, 0},{25606, 1, 1, 19}, {0, 0, 0, 0}}},{12592, {{25606, 1, 0, 0},{25606, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{12593, {{25607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12594, {{25608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12595, {{25609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12596, {{25610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12597, {{25611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12598, {{25612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12599, {{25613, 1, 0, 0},{25613, 1, 1, 18}, {0, 0, 0, 0}}},{12599, {{25613, 1, 0, 0},{25613, 1, 1, 18}, {0, 0, 0, 0}}},{12600, {{25613, 1, 0, 0},{25613, 1, 1, 17}, {0, 0, 0, 0}}},{12600, {{25613, 1, 0, 0},{25613, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{12601, {{25614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12602, {{25615, 1, 0, 0},{25615, 1, 1, 18}, {0, 0, 0, 0}}},{12602, {{25615, 1, 0, 0},{25615, 1, 1, 18}, {0, 0, 0, 0}}},{12603, {{25615, 1, 0, 0},{25615, 1, 1, 19}, {0, 0, 0, 0}}},{12603, {{25615, 1, 0, 0},{25615, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{12604, {{25616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12605, {{25617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12606, {{25618, 1, 0, 0},{25618, 1, 1, 17}, {0, 0, 0, 0}}},{12606, {{25618, 1, 0, 0},{25618, 1, 1, 17}, {0, 0, 0, 0}}},{12607, {{25618, 1, 0, 0},{25618, 1, 1, 18}, {0, 0, 0, 0}}},{12607, {{25618, 1, 0, 0},{25618, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12608, {{25619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12609, {{25620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12610, {{25621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12611, {{25622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12612, {{25623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12613, {{25624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12614, {{25625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12615, {{25626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12616, {{25627, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12617, {{25628, 1, 0, 0},{25628, 1, 1, 18}, {0, 0, 0, 0}}},{12617, {{25628, 1, 0, 0},{25628, 1, 1, 18}, {0, 0, 0, 0}}},{12618, {{25628, 1, 0, 0},{25628, 1, 1, 19}, {0, 0, 0, 0}}},{12618, {{25628, 1, 0, 0},{25628, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{12619, {{25629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12620, {{25630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12621, {{25631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12622, {{25632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12623, {{25633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12624, {{25634, 1, 0, 0},{25634, 1, 1, 17}, {0, 0, 0, 0}}},{12624, {{25634, 1, 0, 0},{25634, 1, 1, 17}, {0, 0, 0, 0}}},{12625, {{25634, 1, 0, 0},{25634, 1, 1, 18}, {0, 0, 0, 0}}},{12625, {{25634, 1, 0, 0},{25634, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12626, {{25635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12627, {{25636, 1, 0, 0},{25636, 1, 1, 18}, {0, 0, 0, 0}}},{12627, {{25636, 1, 0, 0},{25636, 1, 1, 18}, {0, 0, 0, 0}}},{12628, {{25636, 1, 0, 0},{25636, 1, 1, 17}, {0, 0, 0, 0}}},{12628, {{25636, 1, 0, 0},{25636, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{12629, {{25637, 1, 0, 0},{25637, 1, 1, 18}, {0, 0, 0, 0}}},{12629, {{25637, 1, 0, 0},{25637, 1, 1, 18}, {0, 0, 0, 0}}},{12630, {{25637, 1, 0, 0},{25637, 1, 1, 17}, {0, 0, 0, 0}}},{12630, {{25637, 1, 0, 0},{25637, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{12631, {{25638, 1, 0, 0},{25638, 1, 1, 17}, {0, 0, 0, 0}}},{12631, {{25638, 1, 0, 0},{25638, 1, 1, 17}, {0, 0, 0, 0}}},{12632, {{25638, 1, 0, 0},{25638, 1, 1, 18}, {0, 0, 0, 0}}},{12632, {{25638, 1, 0, 0},{25638, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12633, {{25639, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{12634, {{25640, 1, 0, 0},{25640, 1, 1, 20}, {0, 0, 0, 0}}},{12634, {{25640, 1, 0, 0},{25640, 1, 1, 20}, {0, 0, 0, 0}}},{12635, {{25640, 1, 0, 0},{25640, 1, 1, 18}, {0, 0, 0, 0}}},{12635, {{25640, 1, 0, 0},{25640, 1, 1, 18}, {0, 0, 0, 0}}},{12636, {{25640, 1, 0, 0},{25640, 1, 1, 19}, {0, 0, 0, 0}}},{12636, {{25640, 1, 0, 0},{25640, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{12637, {{25641, 1, 0, 0},{25641, 1, 1, 18}, {0, 0, 0, 0}}},{12637, {{25641, 1, 0, 0},{25641, 1, 1, 18}, {0, 0, 0, 0}}},{12638, {{25641, 1, 0, 0},{25641, 1, 1, 17}, {0, 0, 0, 0}}},{12638, {{25641, 1, 0, 0},{25641, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{12639, {{25642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12640, {{25643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12641, {{25644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12642, {{25645, 1, 0, 0},{25645, 1, 1, 17}, {0, 0, 0, 0}}},{12642, {{25645, 1, 0, 0},{25645, 1, 1, 17}, {0, 0, 0, 0}}},{12643, {{25645, 1, 0, 0},{25645, 1, 1, 18}, {0, 0, 0, 0}}},{12643, {{25645, 1, 0, 0},{25645, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12644, {{25646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12645, {{25647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12647, {{25648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12648, {{25649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12649, {{25650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12650, {{25651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12651, {{25652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12652, {{25653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12653, {{25654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12654, {{25655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12655, {{25656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12656, {{25657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12657, {{25658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12658, {{25659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12659, {{25660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12660, {{25661, 1, 0, 0},{25661, 1, 1, 17}, {0, 0, 0, 0}}},{12660, {{25661, 1, 0, 0},{25661, 1, 1, 17}, {0, 0, 0, 0}}},{12661, {{25661, 1, 0, 0},{25661, 1, 1, 18}, {0, 0, 0, 0}}},{12661, {{25661, 1, 0, 0},{25661, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12662, {{25662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12663, {{25663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12664, {{25664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12665, {{25665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12666, {{25666, 1, 0, 0},{25666, 1, 1, 18}, {0, 0, 0, 0}}},{12666, {{25666, 1, 0, 0},{25666, 1, 1, 18}, {0, 0, 0, 0}}},{12667, {{25666, 1, 0, 0},{25666, 1, 1, 19}, {0, 0, 0, 0}}},{12667, {{25666, 1, 0, 0},{25666, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{12668, {{25675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12669, {{25676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12670, {{25677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12671, {{25678, 1, 0, 0},{25678, 1, 1, 17}, {0, 0, 0, 0}}},{12671, {{25678, 1, 0, 0},{25678, 1, 1, 17}, {0, 0, 0, 0}}},{12672, {{25678, 1, 0, 0},{25678, 1, 1, 18}, {0, 0, 0, 0}}},{12672, {{25678, 1, 0, 0},{25678, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12673, {{25679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12674, {{25680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12675, {{25681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12676, {{25682, 1, 0, 0},{25682, 1, 1, 18}, {0, 0, 0, 0}}},{12676, {{25682, 1, 0, 0},{25682, 1, 1, 18}, {0, 0, 0, 0}}},{12677, {{25682, 1, 0, 0},{25682, 1, 1, 17}, {0, 0, 0, 0}}},{12677, {{25682, 1, 0, 0},{25682, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{12678, {{25683, 1, 0, 0},{25683, 1, 1, 18}, {0, 0, 0, 0}}},{12678, {{25683, 1, 0, 0},{25683, 1, 1, 18}, {0, 0, 0, 0}}},{12679, {{25683, 1, 0, 0},{25683, 1, 1, 17}, {0, 0, 0, 0}}},{12679, {{25683, 1, 0, 0},{25683, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{12680, {{25684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12681, {{25685, 1, 0, 0},{25685, 1, 1, 16}, {0, 0, 0, 0}}},{12681, {{25685, 1, 0, 0},{25685, 1, 1, 16}, {0, 0, 0, 0}}},{12682, {{25685, 1, 0, 0},{25685, 1, 1, 17}, {0, 0, 0, 0}}},{12682, {{25685, 1, 0, 0},{25685, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{12683, {{25686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12684, {{25687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12685, {{25688, 1, 0, 0},{25688, 1, 1, 17}, {0, 0, 0, 0}}},{12685, {{25688, 1, 0, 0},{25688, 1, 1, 17}, {0, 0, 0, 0}}},{12686, {{25688, 1, 0, 0},{25688, 1, 1, 18}, {0, 0, 0, 0}}},{12686, {{25688, 1, 0, 0},{25688, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{12687, {{25689, 1, 0, 0},{25689, 1, 1, 16}, {0, 0, 0, 0}}},{12687, {{25689, 1, 0, 0},{25689, 1, 1, 16}, {0, 0, 0, 0}}},{12688, {{25689, 1, 0, 0},{25689, 1, 1, 17}, {0, 0, 0, 0}}},{12688, {{25689, 1, 0, 0},{25689, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{12689, {{25690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12690, {{25691, 1, 0, 0},{25691, 1, 1, 17}, {0, 0, 0, 0}}},{12690, {{25691, 1, 0, 0},{25691, 1, 1, 17}, {0, 0, 0, 0}}},{57570, {{25691, 1, 0, 0},{25691, 1, 1, 18}, {0, 0, 0, 0}}},{57570, {{25691, 1, 0, 0},{25691, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12691, {{25692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12692, {{25693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12693, {{25694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12694, {{25695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{12695, {{25696, 1, 0, 0},{25696, 1, 1, 17}, {0, 0, 0, 0}}},{12695, {{25696, 1, 0, 0},{25696, 1, 1, 17}, {0, 0, 0, 0}}},{12696, {{25696, 1, 0, 0},{25696, 1, 1, 18}, {0, 0, 0, 0}}},{12696, {{25696, 1, 0, 0},{25696, 1, 1, 18}, {0, 0, 0, 0}}},{12697, {{25696, 1, 0, 0},{25696, 1, 1, 19}, {0, 0, 0, 0}}},{12697, {{25696, 1, 0, 0},{25696, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{12698, {{25697, 1, 0, 0},{25697, 1, 1, 18}, {0, 0, 0, 0}}},{12698, {{25697, 1, 0, 0},{25697, 1, 1, 18}, {0, 0, 0, 0}}},{12699, {{25697, 1, 0, 0},{25697, 1, 1, 17}, {0, 0, 0, 0}}},{12699, {{25697, 1, 0, 0},{25697, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{12700, {{25698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12701, {{25699, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12702, {{25700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12703, {{25701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12704, {{25702, 1, 0, 0},{25702, 1, 1, 17}, {0, 0, 0, 0}}},{12704, {{25702, 1, 0, 0},{25702, 1, 1, 17}, {0, 0, 0, 0}}},{12705, {{25702, 1, 0, 0},{25702, 1, 1, 16}, {0, 0, 0, 0}}},{12705, {{25702, 1, 0, 0},{25702, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{12706, {{25703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12707, {{25704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12708, {{25705, 1, 0, 0},{25705, 1, 1, 19}, {0, 0, 0, 0}}},{12708, {{25705, 1, 0, 0},{25705, 1, 1, 19}, {0, 0, 0, 0}}},{12709, {{25705, 1, 0, 0},{25705, 1, 1, 18}, {0, 0, 0, 0}}},{12709, {{25705, 1, 0, 0},{25705, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12710, {{25706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12711, {{25707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12712, {{25708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12713, {{25709, 1, 0, 0},{25709, 1, 1, 17}, {0, 0, 0, 0}}},{12713, {{25709, 1, 0, 0},{25709, 1, 1, 17}, {0, 0, 0, 0}}},{12714, {{25709, 1, 0, 0},{25709, 1, 1, 18}, {0, 0, 0, 0}}},{12714, {{25709, 1, 0, 0},{25709, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12715, {{25710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12716, {{25711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12717, {{25712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12718, {{25713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12719, {{25714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12720, {{25715, 1, 0, 0},{25715, 1, 1, 17}, {0, 0, 0, 0}}},{12720, {{25715, 1, 0, 0},{25715, 1, 1, 17}, {0, 0, 0, 0}}},{12721, {{25715, 1, 0, 0},{25715, 1, 1, 18}, {0, 0, 0, 0}}},{12721, {{25715, 1, 0, 0},{25715, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12722, {{25716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12723, {{25717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12724, {{25718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12725, {{25719, 1, 0, 0},{25719, 1, 1, 16}, {0, 0, 0, 0}}},{12725, {{25719, 1, 0, 0},{25719, 1, 1, 16}, {0, 0, 0, 0}}},{12726, {{25719, 1, 0, 0},{25719, 1, 1, 17}, {0, 0, 0, 0}}},{12726, {{25719, 1, 0, 0},{25719, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{12727, {{25720, 1, 0, 0},{25720, 1, 1, 17}, {0, 0, 0, 0}}},{12727, {{25720, 1, 0, 0},{25720, 1, 1, 17}, {0, 0, 0, 0}}},{12728, {{25720, 1, 0, 0},{25720, 1, 1, 18}, {0, 0, 0, 0}}},{12728, {{25720, 1, 0, 0},{25720, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{12729, {{25721, 1, 0, 0},{25721, 1, 1, 18}, {0, 0, 0, 0}}},{12729, {{25721, 1, 0, 0},{25721, 1, 1, 18}, {0, 0, 0, 0}}},{12730, {{25721, 1, 0, 0},{25721, 1, 1, 17}, {0, 0, 0, 0}}},{12730, {{25721, 1, 0, 0},{25721, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{12732, {{25722, 1, 0, 0},{25722, 1, 1, 19}, {0, 0, 0, 0}}},{12732, {{25722, 1, 0, 0},{25722, 1, 1, 19}, {0, 0, 0, 0}}},{12733, {{25722, 1, 0, 0},{25722, 1, 1, 18}, {0, 0, 0, 0}}},{12733, {{25722, 1, 0, 0},{25722, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12734, {{25723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12735, {{25724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12736, {{25725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12737, {{25726, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12738, {{25727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12739, {{25728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12740, {{25729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12741, {{25730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12742, {{25731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{12743, {{25733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12744, {{25734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12745, {{25735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12746, {{25736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12747, {{25737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12748, {{25738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12749, {{25739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12750, {{25740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12751, {{25741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12752, {{25742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12753, {{25743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{12754, {{25744, 1, 0, 0},{25744, 1, 1, 18}, {0, 0, 0, 0}}},{12754, {{25744, 1, 0, 0},{25744, 1, 1, 18}, {0, 0, 0, 0}}},{12755, {{25744, 1, 0, 0},{25744, 1, 1, 17}, {0, 0, 0, 0}}},{12755, {{25744, 1, 0, 0},{25744, 1, 1, 17}, {0, 0, 0, 0}}},{12756, {{25744, 1, 0, 0},{25744, 1, 1, 19}, {0, 0, 0, 0}}},{12756, {{25744, 1, 0, 0},{25744, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{12757, {{25745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12758, {{25746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12759, {{25747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12760, {{25748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12761, {{25749, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12762, {{25750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12763, {{25751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12764, {{25752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12765, {{25753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12766, {{25754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12767, {{25755, 1, 0, 0},{25755, 1, 1, 18}, {0, 0, 0, 0}}},{12767, {{25755, 1, 0, 0},{25755, 1, 1, 18}, {0, 0, 0, 0}}},{12768, {{25755, 1, 0, 0},{25755, 1, 1, 17}, {0, 0, 0, 0}}},{12768, {{25755, 1, 0, 0},{25755, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{12769, {{25756, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12770, {{25757, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12771, {{25758, 1, 0, 0},{25758, 1, 1, 17}, {0, 0, 0, 0}}},{12771, {{25758, 1, 0, 0},{25758, 1, 1, 17}, {0, 0, 0, 0}}},{12772, {{25758, 1, 0, 0},{25758, 1, 1, 18}, {0, 0, 0, 0}}},{12772, {{25758, 1, 0, 0},{25758, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12773, {{25759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12774, {{25760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12775, {{25761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12776, {{25762, 1, 0, 0},{25762, 1, 1, 16}, {0, 0, 0, 0}}},{12776, {{25762, 1, 0, 0},{25762, 1, 1, 16}, {0, 0, 0, 0}}},{12777, {{25762, 1, 0, 0},{25762, 1, 1, 17}, {0, 0, 0, 0}}},{12777, {{25762, 1, 0, 0},{25762, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{12778, {{25763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12779, {{25764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12780, {{25765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12781, {{25766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12782, {{25767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12783, {{25768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12784, {{25769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12785, {{25770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12786, {{25771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12787, {{25772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12788, {{25773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12789, {{25774, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12790, {{25775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12791, {{25776, 1, 0, 0},{25776, 1, 1, 19}, {0, 0, 0, 0}}},{12791, {{25776, 1, 0, 0},{25776, 1, 1, 19}, {0, 0, 0, 0}}},{12792, {{25776, 1, 0, 0},{25776, 1, 1, 18}, {0, 0, 0, 0}}},{12792, {{25776, 1, 0, 0},{25776, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12793, {{25777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12794, {{25778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12795, {{25779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12796, {{25780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{12797, {{25785, 1, 0, 0},{25785, 1, 1, 17}, {0, 0, 0, 0}}},{12797, {{25785, 1, 0, 0},{25785, 1, 1, 17}, {0, 0, 0, 0}}},{12798, {{25785, 1, 0, 0},{25785, 1, 1, 18}, {0, 0, 0, 0}}},{12798, {{25785, 1, 0, 0},{25785, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{12799, {{25787, 1, 0, 0},{25787, 1, 1, 18}, {0, 0, 0, 0}}},{12799, {{25787, 1, 0, 0},{25787, 1, 1, 18}, {0, 0, 0, 0}}},{12800, {{25787, 1, 0, 0},{25787, 1, 1, 17}, {0, 0, 0, 0}}},{12800, {{25787, 1, 0, 0},{25787, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{12801, {{25788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12802, {{25789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12803, {{25790, 1, 0, 0},{25790, 1, 1, 18}, {0, 0, 0, 0}}},{12803, {{25790, 1, 0, 0},{25790, 1, 1, 18}, {0, 0, 0, 0}}},{12804, {{25790, 1, 0, 0},{25790, 1, 1, 17}, {0, 0, 0, 0}}},{12804, {{25790, 1, 0, 0},{25790, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{12805, {{25791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12806, {{25792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12807, {{25793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12808, {{25794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12809, {{25795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12810, {{25796, 1, 0, 0},{25796, 1, 1, 17}, {0, 0, 0, 0}}},{12810, {{25796, 1, 0, 0},{25796, 1, 1, 17}, {0, 0, 0, 0}}},{12811, {{25796, 1, 0, 0},{25796, 1, 1, 18}, {0, 0, 0, 0}}},{12811, {{25796, 1, 0, 0},{25796, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12812, {{25797, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12813, {{25798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12814, {{25799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12815, {{25800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12816, {{25801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12817, {{25802, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12818, {{25803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12819, {{25804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12820, {{25805, 1, 0, 0},{25805, 1, 1, 17}, {0, 0, 0, 0}}},{12820, {{25805, 1, 0, 0},{25805, 1, 1, 17}, {0, 0, 0, 0}}},{59648, {{25805, 1, 0, 0},{25805, 1, 1, 18}, {0, 0, 0, 0}}},{59648, {{25805, 1, 0, 0},{25805, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{12821, {{25806, 1, 0, 0},{25806, 1, 1, 17}, {0, 0, 0, 0}}},{12821, {{25806, 1, 0, 0},{25806, 1, 1, 17}, {0, 0, 0, 0}}},{12822, {{25806, 1, 0, 0},{25806, 1, 1, 18}, {0, 0, 0, 0}}},{12822, {{25806, 1, 0, 0},{25806, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12823, {{25807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12824, {{25808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12825, {{25809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12826, {{25810, 1, 0, 0},{25810, 1, 1, 17}, {0, 0, 0, 0}}},{12826, {{25810, 1, 0, 0},{25810, 1, 1, 17}, {0, 0, 0, 0}}},{12827, {{25810, 1, 0, 0},{25810, 1, 1, 18}, {0, 0, 0, 0}}},{12827, {{25810, 1, 0, 0},{25810, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12828, {{25811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12829, {{25812, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12830, {{25813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12831, {{25814, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12832, {{25815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12833, {{25816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12834, {{25817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{12835, {{25818, 1, 0, 0},{25818, 1, 1, 17}, {0, 0, 0, 0}}},{12835, {{25818, 1, 0, 0},{25818, 1, 1, 17}, {0, 0, 0, 0}}},{12836, {{25818, 1, 0, 0},{25818, 1, 1, 18}, {0, 0, 0, 0}}},{12836, {{25818, 1, 0, 0},{25818, 1, 1, 18}, {0, 0, 0, 0}}},{12837, {{25818, 1, 0, 0},{25818, 1, 1, 19}, {0, 0, 0, 0}}},{12837, {{25818, 1, 0, 0},{25818, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{12838, {{25819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{12839, {{25823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12840, {{25824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12841, {{25825, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12842, {{25826, 1, 0, 0},{25826, 1, 1, 19}, {0, 0, 0, 0}}},{12842, {{25826, 1, 0, 0},{25826, 1, 1, 19}, {0, 0, 0, 0}}},{12843, {{25826, 1, 0, 0},{25826, 1, 1, 18}, {0, 0, 0, 0}}},{12843, {{25826, 1, 0, 0},{25826, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12844, {{25827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12845, {{25828, 1, 0, 0},{25828, 1, 1, 17}, {0, 0, 0, 0}}},{12845, {{25828, 1, 0, 0},{25828, 1, 1, 17}, {0, 0, 0, 0}}},{12846, {{25828, 1, 0, 0},{25828, 1, 1, 18}, {0, 0, 0, 0}}},{12846, {{25828, 1, 0, 0},{25828, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12847, {{25829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12848, {{25830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12849, {{25831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12850, {{25832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12851, {{25833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{12852, {{25834, 1, 0, 0},{25834, 1, 1, 17}, {0, 0, 0, 0}}},{12852, {{25834, 1, 0, 0},{25834, 1, 1, 17}, {0, 0, 0, 0}}},{12853, {{25834, 1, 0, 0},{25834, 1, 1, 19}, {0, 0, 0, 0}}},{12853, {{25834, 1, 0, 0},{25834, 1, 1, 19}, {0, 0, 0, 0}}},{12854, {{25834, 1, 0, 0},{25834, 1, 1, 18}, {0, 0, 0, 0}}},{12854, {{25834, 1, 0, 0},{25834, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12855, {{25835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12856, {{25836, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12857, {{25837, 1, 0, 0},{25837, 1, 1, 17}, {0, 0, 0, 0}}},{12857, {{25837, 1, 0, 0},{25837, 1, 1, 17}, {0, 0, 0, 0}}},{12858, {{25837, 1, 0, 0},{25837, 1, 1, 18}, {0, 0, 0, 0}}},{12858, {{25837, 1, 0, 0},{25837, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12859, {{25838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12860, {{25839, 1, 0, 0},{25839, 1, 1, 17}, {0, 0, 0, 0}}},{12860, {{25839, 1, 0, 0},{25839, 1, 1, 17}, {0, 0, 0, 0}}},{12861, {{25839, 1, 0, 0},{25839, 1, 1, 18}, {0, 0, 0, 0}}},{12861, {{25839, 1, 0, 0},{25839, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12862, {{25840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12863, {{25841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{12864, {{25842, 1, 0, 0},{25842, 1, 1, 18}, {0, 0, 0, 0}}},{12864, {{25842, 1, 0, 0},{25842, 1, 1, 18}, {0, 0, 0, 0}}},{12865, {{25842, 1, 0, 0},{25842, 1, 1, 21}, {0, 0, 0, 0}}},{12865, {{25842, 1, 0, 0},{25842, 1, 1, 21}, {0, 0, 0, 0}}},{12866, {{25842, 1, 0, 0},{25842, 1, 1, 19}, {0, 0, 0, 0}}},{12866, {{25842, 1, 0, 0},{25842, 1, 1, 19}, {0, 0, 0, 0}}},{12867, {{25842, 1, 0, 0},{25842, 1, 1, 20}, {0, 0, 0, 0}}},{12867, {{25842, 1, 0, 0},{25842, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{12868, {{25843, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12869, {{25844, 1, 0, 0},{25844, 1, 1, 17}, {0, 0, 0, 0}}},{12869, {{25844, 1, 0, 0},{25844, 1, 1, 17}, {0, 0, 0, 0}}},{12870, {{25844, 1, 0, 0},{25844, 1, 1, 18}, {0, 0, 0, 0}}},{12870, {{25844, 1, 0, 0},{25844, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12874, {{25845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12875, {{25846, 1, 0, 0},{25846, 1, 1, 19}, {0, 0, 0, 0}}},{12875, {{25846, 1, 0, 0},{25846, 1, 1, 19}, {0, 0, 0, 0}}},{12876, {{25846, 1, 0, 0},{25846, 1, 1, 18}, {0, 0, 0, 0}}},{12876, {{25846, 1, 0, 0},{25846, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12877, {{25847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12878, {{25848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12879, {{25849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12880, {{25850, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12881, {{25851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12882, {{25852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12883, {{25853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12885, {{25854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{12886, {{25855, 1, 0, 0},{25855, 1, 1, 18}, {0, 0, 0, 0}}},{12886, {{25855, 1, 0, 0},{25855, 1, 1, 18}, {0, 0, 0, 0}}},{12887, {{25855, 1, 0, 0},{25855, 1, 1, 17}, {0, 0, 0, 0}}},{12887, {{25855, 1, 0, 0},{25855, 1, 1, 17}, {0, 0, 0, 0}}},{12888, {{25855, 1, 0, 0},{25855, 1, 1, 19}, {0, 0, 0, 0}}},{12888, {{25855, 1, 0, 0},{25855, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{12889, {{25856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12890, {{25857, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12891, {{25858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12892, {{25859, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12893, {{25860, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12894, {{25861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12895, {{25862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12896, {{25863, 1, 0, 0},{25863, 1, 1, 16}, {0, 0, 0, 0}}},{12896, {{25863, 1, 0, 0},{25863, 1, 1, 16}, {0, 0, 0, 0}}},{12897, {{25863, 1, 0, 0},{25863, 1, 1, 17}, {0, 0, 0, 0}}},{12897, {{25863, 1, 0, 0},{25863, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{12898, {{25864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12899, {{25865, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12900, {{25866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12901, {{25867, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12902, {{25868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12903, {{25869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12904, {{25870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12905, {{25871, 1, 0, 0},{25871, 1, 1, 17}, {0, 0, 0, 0}}},{12905, {{25871, 1, 0, 0},{25871, 1, 1, 17}, {0, 0, 0, 0}}},{12906, {{25871, 1, 0, 0},{25871, 1, 1, 18}, {0, 0, 0, 0}}},{12906, {{25871, 1, 0, 0},{25871, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12907, {{25872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{12908, {{25875, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12909, {{25876, 1, 0, 0},{25876, 1, 1, 17}, {0, 0, 0, 0}}},{12909, {{25876, 1, 0, 0},{25876, 1, 1, 17}, {0, 0, 0, 0}}},{12910, {{25876, 1, 0, 0},{25876, 1, 1, 18}, {0, 0, 0, 0}}},{12910, {{25876, 1, 0, 0},{25876, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12911, {{25877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12912, {{25878, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12913, {{25879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12914, {{25880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12915, {{25881, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{12916, {{25883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12917, {{25884, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{12918, {{25885, 1, 0, 0},{25885, 1, 1, 18}, {0, 0, 0, 0}}},{12918, {{25885, 1, 0, 0},{25885, 1, 1, 18}, {0, 0, 0, 0}}},{12919, {{25885, 1, 0, 0},{25885, 1, 1, 20}, {0, 0, 0, 0}}},{12919, {{25885, 1, 0, 0},{25885, 1, 1, 20}, {0, 0, 0, 0}}},{12920, {{25885, 1, 0, 0},{25885, 1, 1, 21}, {0, 0, 0, 0}}},{12920, {{25885, 1, 0, 0},{25885, 1, 1, 21}, {0, 0, 0, 0}}},{12921, {{25885, 1, 0, 0},{25885, 1, 1, 19}, {0, 0, 0, 0}}},{12921, {{25885, 1, 0, 0},{25885, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{12922, {{25886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12923, {{25887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12924, {{25888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12925, {{25889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12926, {{25890, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12927, {{25891, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12928, {{25892, 1, 0, 0},{25892, 1, 1, 17}, {0, 0, 0, 0}}},{12928, {{25892, 1, 0, 0},{25892, 1, 1, 17}, {0, 0, 0, 0}}},{12929, {{25892, 1, 0, 0},{25892, 1, 1, 18}, {0, 0, 0, 0}}},{12929, {{25892, 1, 0, 0},{25892, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12930, {{25893, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12931, {{25894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{12932, {{25896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12933, {{25897, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12934, {{25898, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12935, {{25899, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12936, {{25900, 1, 0, 0},{25900, 1, 1, 17}, {0, 0, 0, 0}}},{12936, {{25900, 1, 0, 0},{25900, 1, 1, 17}, {0, 0, 0, 0}}},{12937, {{25900, 1, 0, 0},{25900, 1, 1, 18}, {0, 0, 0, 0}}},{12937, {{25900, 1, 0, 0},{25900, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12938, {{25901, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12939, {{25902, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12940, {{25903, 1, 0, 0},{25903, 1, 1, 17}, {0, 0, 0, 0}}},{12940, {{25903, 1, 0, 0},{25903, 1, 1, 17}, {0, 0, 0, 0}}},{12941, {{25903, 1, 0, 0},{25903, 1, 1, 18}, {0, 0, 0, 0}}},{12941, {{25903, 1, 0, 0},{25903, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12942, {{25904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12943, {{25905, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12944, {{25906, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12945, {{25907, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12946, {{25908, 1, 0, 0},{25908, 1, 1, 17}, {0, 0, 0, 0}}},{12946, {{25908, 1, 0, 0},{25908, 1, 1, 17}, {0, 0, 0, 0}}},{57581, {{25908, 1, 0, 0},{25908, 1, 1, 18}, {0, 0, 0, 0}}},{57581, {{25908, 1, 0, 0},{25908, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{12947, {{25909, 1, 0, 0},{25909, 1, 1, 17}, {0, 0, 0, 0}}},{12947, {{25909, 1, 0, 0},{25909, 1, 1, 17}, {0, 0, 0, 0}}},{12948, {{25909, 1, 0, 0},{25909, 1, 1, 19}, {0, 0, 0, 0}}},{12948, {{25909, 1, 0, 0},{25909, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{12950, {{25910, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12951, {{25911, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12952, {{25912, 1, 0, 0},{25912, 1, 1, 18}, {0, 0, 0, 0}}},{12952, {{25912, 1, 0, 0},{25912, 1, 1, 18}, {0, 0, 0, 0}}},{12953, {{25912, 1, 0, 0},{25912, 1, 1, 17}, {0, 0, 0, 0}}},{12953, {{25912, 1, 0, 0},{25912, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{12954, {{25913, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12955, {{25914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12956, {{25915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12957, {{25916, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12958, {{25917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12959, {{25918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12960, {{25919, 1, 0, 0},{25919, 1, 1, 17}, {0, 0, 0, 0}}},{12960, {{25919, 1, 0, 0},{25919, 1, 1, 17}, {0, 0, 0, 0}}},{12961, {{25919, 1, 0, 0},{25919, 1, 1, 18}, {0, 0, 0, 0}}},{12961, {{25919, 1, 0, 0},{25919, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12962, {{25920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12963, {{25921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12964, {{25922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12965, {{25923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12966, {{25924, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12967, {{25925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12968, {{25926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12969, {{25927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12970, {{25928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12971, {{25929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12972, {{25930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{12973, {{25932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12974, {{25933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12975, {{25934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{12976, {{25935, 1, 0, 0},{25935, 1, 1, 18}, {0, 0, 0, 0}}},{12976, {{25935, 1, 0, 0},{25935, 1, 1, 18}, {0, 0, 0, 0}}},{30262, {{25935, 1, 0, 0},{25935, 1, 1, 0},{25935, 1, 1, 19},{64065, 1, 0, 0}, {0, 0, 0, 0}}},{30262, {{25935, 1, 0, 0},{25935, 1, 1, 0},{25935, 1, 1, 19},{64065, 1, 0, 0}, {0, 0, 0, 0}}},{30262, {{25935, 1, 0, 0},{25935, 1, 1, 0},{25935, 1, 1, 19},{64065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12977, {{25936, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12978, {{25937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12979, {{25938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12980, {{25939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12982, {{25940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12983, {{25941, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12984, {{25942, 1, 0, 0},{25942, 1, 1, 17}, {0, 0, 0, 0}}},{12984, {{25942, 1, 0, 0},{25942, 1, 1, 17}, {0, 0, 0, 0}}},{57586, {{25942, 1, 0, 0},{25942, 1, 1, 18}, {0, 0, 0, 0}}},{57586, {{25942, 1, 0, 0},{25942, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{12985, {{25943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12986, {{25944, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12987, {{25945, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{12988, {{25948, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{12989, {{25949, 1, 0, 0},{25949, 1, 1, 18}, {0, 0, 0, 0}}},{12989, {{25949, 1, 0, 0},{25949, 1, 1, 18}, {0, 0, 0, 0}}},{12990, {{25949, 1, 0, 0},{25949, 1, 1, 19}, {0, 0, 0, 0}}},{12990, {{25949, 1, 0, 0},{25949, 1, 1, 19}, {0, 0, 0, 0}}},{12991, {{25949, 1, 0, 0},{25949, 1, 1, 20}, {0, 0, 0, 0}}},{12991, {{25949, 1, 0, 0},{25949, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{12992, {{25950, 1, 0, 0},{25950, 1, 1, 18}, {0, 0, 0, 0}}},{12992, {{25950, 1, 0, 0},{25950, 1, 1, 18}, {0, 0, 0, 0}}},{12993, {{25950, 1, 0, 0},{25950, 1, 1, 19}, {0, 0, 0, 0}}},{12993, {{25950, 1, 0, 0},{25950, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{12994, {{25951, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12995, {{25952, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12996, {{25953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{12997, {{25954, 1, 0, 0},{25954, 1, 1, 18}, {0, 0, 0, 0}}},{12997, {{25954, 1, 0, 0},{25954, 1, 1, 18}, {0, 0, 0, 0}}},{57594, {{25954, 1, 0, 0},{25954, 1, 1, 19}, {0, 0, 0, 0}}},{57594, {{25954, 1, 0, 0},{25954, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{12998, {{25955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{12999, {{25956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13000, {{25957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13001, {{25958, 1, 0, 0},{25958, 1, 1, 17}, {0, 0, 0, 0}}},{13001, {{25958, 1, 0, 0},{25958, 1, 1, 17}, {0, 0, 0, 0}}},{57608, {{25958, 1, 0, 0},{25958, 1, 1, 18}, {0, 0, 0, 0}}},{57608, {{25958, 1, 0, 0},{25958, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13002, {{25959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{13003, {{25961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{13005, {{25963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13006, {{25964, 1, 0, 0},{25964, 1, 1, 17}, {0, 0, 0, 0}}},{13006, {{25964, 1, 0, 0},{25964, 1, 1, 17}, {0, 0, 0, 0}}},{13007, {{25964, 1, 0, 0},{25964, 1, 1, 18}, {0, 0, 0, 0}}},{13007, {{25964, 1, 0, 0},{25964, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13009, {{25965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13011, {{25966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13012, {{25967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13013, {{25968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13014, {{25969, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13015, {{25970, 1, 0, 0},{25970, 1, 1, 17}, {0, 0, 0, 0}}},{13015, {{25970, 1, 0, 0},{25970, 1, 1, 17}, {0, 0, 0, 0}}},{13016, {{25970, 1, 0, 0},{25970, 1, 1, 18}, {0, 0, 0, 0}}},{13016, {{25970, 1, 0, 0},{25970, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{57610, {{25971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13017, {{25972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13018, {{25973, 1, 0, 0},{25973, 1, 1, 17}, {0, 0, 0, 0}}},{13018, {{25973, 1, 0, 0},{25973, 1, 1, 17}, {0, 0, 0, 0}}},{13019, {{25973, 1, 0, 0},{25973, 1, 1, 18}, {0, 0, 0, 0}}},{13019, {{25973, 1, 0, 0},{25973, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13020, {{25974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13021, {{25975, 1, 0, 0},{25975, 1, 1, 19}, {0, 0, 0, 0}}},{13021, {{25975, 1, 0, 0},{25975, 1, 1, 19}, {0, 0, 0, 0}}},{13022, {{25975, 1, 0, 0},{25975, 1, 1, 18}, {0, 0, 0, 0}}},{13022, {{25975, 1, 0, 0},{25975, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{13023, {{25976, 1, 0, 0},{25976, 1, 1, 17}, {0, 0, 0, 0}}},{13023, {{25976, 1, 0, 0},{25976, 1, 1, 17}, {0, 0, 0, 0}}},{13024, {{25976, 1, 0, 0},{25976, 1, 1, 20}, {0, 0, 0, 0}}},{13024, {{25976, 1, 0, 0},{25976, 1, 1, 20}, {0, 0, 0, 0}}},{59656, {{25976, 1, 0, 0},{25976, 1, 1, 18}, {0, 0, 0, 0}}},{59656, {{25976, 1, 0, 0},{25976, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13025, {{25977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13026, {{25978, 1, 0, 0},{25978, 1, 1, 17}, {0, 0, 0, 0}}},{13026, {{25978, 1, 0, 0},{25978, 1, 1, 17}, {0, 0, 0, 0}}},{13027, {{25978, 1, 0, 0},{25978, 1, 1, 18}, {0, 0, 0, 0}}},{13027, {{25978, 1, 0, 0},{25978, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{57614, {{25979, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13028, {{25980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13029, {{25981, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{68034, {{25982, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{13030, {{25984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13031, {{25985, 1, 0, 0},{25985, 1, 1, 17}, {0, 0, 0, 0}}},{13031, {{25985, 1, 0, 0},{25985, 1, 1, 17}, {0, 0, 0, 0}}},{13032, {{25985, 1, 0, 0},{25985, 1, 1, 18}, {0, 0, 0, 0}}},{13032, {{25985, 1, 0, 0},{25985, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13033, {{25986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{13034, {{25987, 1, 0, 0},{25987, 1, 1, 18}, {0, 0, 0, 0}}},{13034, {{25987, 1, 0, 0},{25987, 1, 1, 18}, {0, 0, 0, 0}}},{13035, {{25987, 1, 0, 0},{25987, 1, 1, 20}, {0, 0, 0, 0}}},{13035, {{25987, 1, 0, 0},{25987, 1, 1, 20}, {0, 0, 0, 0}}},{13036, {{25987, 1, 0, 0},{25987, 1, 1, 19}, {0, 0, 0, 0}}},{13036, {{25987, 1, 0, 0},{25987, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{13037, {{25988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13038, {{25989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13039, {{25990, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13040, {{25991, 1, 0, 0},{25991, 1, 1, 18}, {0, 0, 0, 0}}},{13040, {{25991, 1, 0, 0},{25991, 1, 1, 18}, {0, 0, 0, 0}}},{13041, {{25991, 1, 0, 0},{25991, 1, 1, 19}, {0, 0, 0, 0}}},{13041, {{25991, 1, 0, 0},{25991, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{13042, {{25992, 1, 0, 0},{25992, 1, 1, 17}, {0, 0, 0, 0}}},{13042, {{25992, 1, 0, 0},{25992, 1, 1, 17}, {0, 0, 0, 0}}},{13043, {{25992, 1, 0, 0},{25992, 1, 1, 18}, {0, 0, 0, 0}}},{13043, {{25992, 1, 0, 0},{25992, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{13044, {{25993, 1, 0, 0},{25993, 1, 1, 18}, {0, 0, 0, 0}}},{13044, {{25993, 1, 0, 0},{25993, 1, 1, 18}, {0, 0, 0, 0}}},{59659, {{25993, 1, 0, 0},{25993, 1, 1, 19}, {0, 0, 0, 0}}},{59659, {{25993, 1, 0, 0},{25993, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{13045, {{25994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13046, {{25995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13047, {{25996, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13048, {{25997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13049, {{25998, 1, 0, 0},{25998, 1, 1, 18}, {0, 0, 0, 0}}},{13049, {{25998, 1, 0, 0},{25998, 1, 1, 18}, {0, 0, 0, 0}}},{13050, {{25998, 1, 0, 0},{25998, 1, 1, 19}, {0, 0, 0, 0}}},{13050, {{25998, 1, 0, 0},{25998, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{13051, {{26000, 1, 0, 0},{26000, 1, 1, 18}, {0, 0, 0, 0}}},{13051, {{26000, 1, 0, 0},{26000, 1, 1, 18}, {0, 0, 0, 0}}},{13052, {{26000, 1, 0, 0},{26000, 1, 1, 19}, {0, 0, 0, 0}}},{13052, {{26000, 1, 0, 0},{26000, 1, 1, 19}, {0, 0, 0, 0}}},{13053, {{26000, 1, 0, 0},{26000, 1, 1, 20}, {0, 0, 0, 0}}},{13053, {{26000, 1, 0, 0},{26000, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{13054, {{26001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13055, {{26002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{13056, {{26004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13057, {{26005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13058, {{26006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13059, {{26007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13060, {{26008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13061, {{26009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13062, {{26010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13063, {{26011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13064, {{26012, 1, 0, 0},{26012, 1, 1, 19}, {0, 0, 0, 0}}},{13064, {{26012, 1, 0, 0},{26012, 1, 1, 19}, {0, 0, 0, 0}}},{13065, {{26012, 1, 0, 0},{26012, 1, 1, 18}, {0, 0, 0, 0}}},{13065, {{26012, 1, 0, 0},{26012, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13066, {{26013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13067, {{26014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13068, {{26015, 1, 0, 0},{26015, 1, 1, 17}, {0, 0, 0, 0}}},{13068, {{26015, 1, 0, 0},{26015, 1, 1, 17}, {0, 0, 0, 0}}},{13069, {{26015, 1, 0, 0},{26015, 1, 1, 18}, {0, 0, 0, 0}}},{13069, {{26015, 1, 0, 0},{26015, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13070, {{26016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13071, {{26017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13072, {{26018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13073, {{26019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13074, {{26020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13075, {{26021, 1, 0, 0},{26021, 1, 1, 17}, {0, 0, 0, 0}}},{13075, {{26021, 1, 0, 0},{26021, 1, 1, 17}, {0, 0, 0, 0}}},{57624, {{26021, 1, 0, 0},{26021, 1, 1, 18}, {0, 0, 0, 0}}},{57624, {{26021, 1, 0, 0},{26021, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13076, {{26022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13077, {{26023, 1, 0, 0},{26023, 1, 1, 18}, {0, 0, 0, 0}}},{13077, {{26023, 1, 0, 0},{26023, 1, 1, 18}, {0, 0, 0, 0}}},{13078, {{26023, 1, 0, 0},{26023, 1, 1, 19}, {0, 0, 0, 0}}},{13078, {{26023, 1, 0, 0},{26023, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{13079, {{26024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{13080, {{26026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13081, {{26027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13082, {{26028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13083, {{26029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13084, {{26030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13085, {{26031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13086, {{26032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13087, {{26033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13088, {{26034, 1, 0, 0},{26034, 1, 1, 17}, {0, 0, 0, 0}}},{13088, {{26034, 1, 0, 0},{26034, 1, 1, 17}, {0, 0, 0, 0}}},{13089, {{26034, 1, 0, 0},{26034, 1, 1, 18}, {0, 0, 0, 0}}},{13089, {{26034, 1, 0, 0},{26034, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13090, {{26035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13091, {{26036, 1, 0, 0},{26036, 1, 1, 18}, {0, 0, 0, 0}}},{13091, {{26036, 1, 0, 0},{26036, 1, 1, 18}, {0, 0, 0, 0}}},{13092, {{26036, 1, 0, 0},{26036, 1, 1, 17}, {0, 0, 0, 0}}},{13092, {{26036, 1, 0, 0},{26036, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{13093, {{26037, 1, 0, 0},{26037, 1, 1, 18}, {0, 0, 0, 0}}},{13093, {{26037, 1, 0, 0},{26037, 1, 1, 18}, {0, 0, 0, 0}}},{13094, {{26037, 1, 0, 0},{26037, 1, 1, 17}, {0, 0, 0, 0}}},{13094, {{26037, 1, 0, 0},{26037, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{13095, {{26038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13096, {{26039, 1, 0, 0},{26039, 1, 1, 17}, {0, 0, 0, 0}}},{13096, {{26039, 1, 0, 0},{26039, 1, 1, 17}, {0, 0, 0, 0}}},{13097, {{26039, 1, 0, 0},{26039, 1, 1, 18}, {0, 0, 0, 0}}},{13097, {{26039, 1, 0, 0},{26039, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13098, {{26040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13099, {{26041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13100, {{26042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57635, {{26043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13102, {{26044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13103, {{26045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13104, {{26046, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13105, {{26047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{13106, {{26049, 1, 0, 0},{26049, 1, 1, 17}, {0, 0, 0, 0}}},{13106, {{26049, 1, 0, 0},{26049, 1, 1, 17}, {0, 0, 0, 0}}},{57637, {{26049, 1, 0, 0},{26049, 1, 1, 18}, {0, 0, 0, 0}}},{57637, {{26049, 1, 0, 0},{26049, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13107, {{26050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13108, {{26051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13109, {{26052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13110, {{26053, 1, 0, 0},{26053, 1, 1, 19}, {0, 0, 0, 0}}},{13110, {{26053, 1, 0, 0},{26053, 1, 1, 19}, {0, 0, 0, 0}}},{13111, {{26053, 1, 0, 0},{26053, 1, 1, 18}, {0, 0, 0, 0}}},{13111, {{26053, 1, 0, 0},{26053, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13112, {{26054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13113, {{26055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13114, {{26056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13115, {{26057, 1, 0, 0},{26057, 1, 1, 17}, {0, 0, 0, 0}}},{13115, {{26057, 1, 0, 0},{26057, 1, 1, 17}, {0, 0, 0, 0}}},{13116, {{26057, 1, 0, 0},{26057, 1, 1, 18}, {0, 0, 0, 0}}},{13116, {{26057, 1, 0, 0},{26057, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13117, {{26058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13118, {{26059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13119, {{26060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13120, {{26061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13121, {{26062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13122, {{26063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13123, {{26064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13124, {{26065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13125, {{26066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13126, {{26067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13127, {{26068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{68035, {{26069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13128, {{26070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13129, {{26071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13130, {{26072, 1, 0, 0},{26072, 1, 1, 17}, {0, 0, 0, 0}}},{13130, {{26072, 1, 0, 0},{26072, 1, 1, 17}, {0, 0, 0, 0}}},{13131, {{26072, 1, 0, 0},{26072, 1, 1, 18}, {0, 0, 0, 0}}},{13131, {{26072, 1, 0, 0},{26072, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13132, {{26073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13133, {{26074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13134, {{26075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13135, {{26076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13136, {{26077, 1, 0, 0},{26077, 1, 1, 16}, {0, 0, 0, 0}}},{13136, {{26077, 1, 0, 0},{26077, 1, 1, 16}, {0, 0, 0, 0}}},{13137, {{26077, 1, 0, 0},{26077, 1, 1, 17}, {0, 0, 0, 0}}},{13137, {{26077, 1, 0, 0},{26077, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {6, {{13138, {{26078, 1, 0, 0},{26078, 1, 1, 16}, {0, 0, 0, 0}}},{13138, {{26078, 1, 0, 0},{26078, 1, 1, 16}, {0, 0, 0, 0}}},{13139, {{26078, 1, 0, 0},{26078, 1, 1, 17}, {0, 0, 0, 0}}},{13139, {{26078, 1, 0, 0},{26078, 1, 1, 17}, {0, 0, 0, 0}}},{13140, {{26078, 1, 0, 0},{26078, 1, 1, 18}, {0, 0, 0, 0}}},{13140, {{26078, 1, 0, 0},{26078, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13141, {{26079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13142, {{26080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{13143, {{26081, 1, 0, 0},{26081, 1, 1, 18}, {0, 0, 0, 0}}},{13143, {{26081, 1, 0, 0},{26081, 1, 1, 18}, {0, 0, 0, 0}}},{13144, {{26081, 1, 0, 0},{26081, 1, 1, 21}, {0, 0, 0, 0}}},{13144, {{26081, 1, 0, 0},{26081, 1, 1, 21}, {0, 0, 0, 0}}},{13145, {{26081, 1, 0, 0},{26081, 1, 1, 19}, {0, 0, 0, 0}}},{13145, {{26081, 1, 0, 0},{26081, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {9, {{13146, {{26082, 1, 0, 0},{26082, 1, 1, 18}, {0, 0, 0, 0}}},{13146, {{26082, 1, 0, 0},{26082, 1, 1, 18}, {0, 0, 0, 0}}},{13147, {{26082, 1, 0, 0},{26082, 1, 1, 19}, {0, 0, 0, 0}}},{13147, {{26082, 1, 0, 0},{26082, 1, 1, 19}, {0, 0, 0, 0}}},{30263, {{26082, 1, 0, 0},{26082, 1, 1, 0},{26082, 1, 1, 20},{64066, 1, 0, 0}, {0, 0, 0, 0}}},{30263, {{26082, 1, 0, 0},{26082, 1, 1, 0},{26082, 1, 1, 20},{64066, 1, 0, 0}, {0, 0, 0, 0}}},{30263, {{26082, 1, 0, 0},{26082, 1, 1, 0},{26082, 1, 1, 20},{64066, 1, 0, 0}, {0, 0, 0, 0}}},{57647, {{26082, 1, 0, 0},{26082, 1, 1, 21}, {0, 0, 0, 0}}},{57647, {{26082, 1, 0, 0},{26082, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {4, {{13148, {{26083, 1, 0, 0},{26083, 1, 1, 18}, {0, 0, 0, 0}}},{13148, {{26083, 1, 0, 0},{26083, 1, 1, 18}, {0, 0, 0, 0}}},{13149, {{26083, 1, 0, 0},{26083, 1, 1, 17}, {0, 0, 0, 0}}},{13149, {{26083, 1, 0, 0},{26083, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{13150, {{26084, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13151, {{26085, 1, 0, 0},{26085, 1, 1, 17}, {0, 0, 0, 0}}},{13151, {{26085, 1, 0, 0},{26085, 1, 1, 17}, {0, 0, 0, 0}}},{13152, {{26085, 1, 0, 0},{26085, 1, 1, 18}, {0, 0, 0, 0}}},{13152, {{26085, 1, 0, 0},{26085, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13153, {{26086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13154, {{26087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13155, {{26088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13156, {{26089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13157, {{26090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13158, {{26091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13159, {{26092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13160, {{26093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13161, {{26094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{13162, {{26096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13163, {{26097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13164, {{26098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13166, {{26099, 1, 0, 0},{26099, 1, 1, 16}, {0, 0, 0, 0}}},{13166, {{26099, 1, 0, 0},{26099, 1, 1, 16}, {0, 0, 0, 0}}},{13167, {{26099, 1, 0, 0},{26099, 1, 1, 17}, {0, 0, 0, 0}}},{13167, {{26099, 1, 0, 0},{26099, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{13168, {{26100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13169, {{26101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13170, {{26102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{13171, {{26105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13172, {{26106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13173, {{26107, 1, 0, 0},{26107, 1, 1, 17}, {0, 0, 0, 0}}},{13173, {{26107, 1, 0, 0},{26107, 1, 1, 17}, {0, 0, 0, 0}}},{13174, {{26107, 1, 0, 0},{26107, 1, 1, 18}, {0, 0, 0, 0}}},{13174, {{26107, 1, 0, 0},{26107, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13175, {{26108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13176, {{26109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13178, {{26110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13180, {{26111, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13181, {{26112, 1, 0, 0},{26112, 1, 1, 17}, {0, 0, 0, 0}}},{13181, {{26112, 1, 0, 0},{26112, 1, 1, 17}, {0, 0, 0, 0}}},{13182, {{26112, 1, 0, 0},{26112, 1, 1, 18}, {0, 0, 0, 0}}},{13182, {{26112, 1, 0, 0},{26112, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13183, {{26113, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13184, {{26114, 1, 0, 0},{26114, 1, 1, 17}, {0, 0, 0, 0}}},{13184, {{26114, 1, 0, 0},{26114, 1, 1, 17}, {0, 0, 0, 0}}},{13185, {{26114, 1, 0, 0},{26114, 1, 1, 18}, {0, 0, 0, 0}}},{13185, {{26114, 1, 0, 0},{26114, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13186, {{26115, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13187, {{26116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13188, {{26117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13189, {{26118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13190, {{26119, 1, 0, 0},{26119, 1, 1, 17}, {0, 0, 0, 0}}},{13190, {{26119, 1, 0, 0},{26119, 1, 1, 17}, {0, 0, 0, 0}}},{59678, {{26119, 1, 0, 0},{26119, 1, 1, 18}, {0, 0, 0, 0}}},{59678, {{26119, 1, 0, 0},{26119, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{13191, {{26120, 1, 0, 0},{26120, 1, 1, 17}, {0, 0, 0, 0}}},{13191, {{26120, 1, 0, 0},{26120, 1, 1, 17}, {0, 0, 0, 0}}},{13192, {{26120, 1, 0, 0},{26120, 1, 1, 18}, {0, 0, 0, 0}}},{13192, {{26120, 1, 0, 0},{26120, 1, 1, 18}, {0, 0, 0, 0}}},{13193, {{26120, 1, 0, 0},{26120, 1, 1, 19}, {0, 0, 0, 0}}},{13193, {{26120, 1, 0, 0},{26120, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{13194, {{26121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13195, {{26122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13196, {{26123, 1, 0, 0},{26123, 1, 1, 16}, {0, 0, 0, 0}}},{13196, {{26123, 1, 0, 0},{26123, 1, 1, 16}, {0, 0, 0, 0}}},{13197, {{26123, 1, 0, 0},{26123, 1, 1, 17}, {0, 0, 0, 0}}},{13197, {{26123, 1, 0, 0},{26123, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{13198, {{26124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13199, {{26125, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13201, {{26126, 1, 0, 0},{26126, 1, 1, 18}, {0, 0, 0, 0}}},{13201, {{26126, 1, 0, 0},{26126, 1, 1, 18}, {0, 0, 0, 0}}},{13202, {{26126, 1, 0, 0},{26126, 1, 1, 19}, {0, 0, 0, 0}}},{13202, {{26126, 1, 0, 0},{26126, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{13203, {{26127, 1, 0, 0},{26127, 1, 1, 17}, {0, 0, 0, 0}}},{13203, {{26127, 1, 0, 0},{26127, 1, 1, 17}, {0, 0, 0, 0}}},{57659, {{26127, 1, 0, 0},{26127, 1, 1, 18}, {0, 0, 0, 0}}},{57659, {{26127, 1, 0, 0},{26127, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13204, {{26128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13205, {{26129, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13206, {{26130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13207, {{26131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13208, {{26132, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13209, {{26133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13210, {{26134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13211, {{26135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13212, {{26136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13213, {{26137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13214, {{26138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13215, {{26139, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13216, {{26140, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13217, {{26141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{13219, {{26142, 1, 0, 0},{26142, 1, 1, 19}, {0, 0, 0, 0}}},{13219, {{26142, 1, 0, 0},{26142, 1, 1, 19}, {0, 0, 0, 0}}},{13220, {{26142, 1, 0, 0},{26142, 1, 1, 18}, {0, 0, 0, 0}}},{13220, {{26142, 1, 0, 0},{26142, 1, 1, 18}, {0, 0, 0, 0}}},{13221, {{26142, 1, 0, 0},{26142, 1, 1, 20}, {0, 0, 0, 0}}},{13221, {{26142, 1, 0, 0},{26142, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{13222, {{26143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13223, {{26144, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13224, {{26145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13226, {{26146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13227, {{26147, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13228, {{26148, 1, 0, 0},{26148, 1, 1, 17}, {0, 0, 0, 0}}},{13228, {{26148, 1, 0, 0},{26148, 1, 1, 17}, {0, 0, 0, 0}}},{13229, {{26148, 1, 0, 0},{26148, 1, 1, 18}, {0, 0, 0, 0}}},{13229, {{26148, 1, 0, 0},{26148, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13230, {{26149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13231, {{26150, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13232, {{26151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13233, {{26152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13234, {{26153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13235, {{26154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13236, {{26155, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13237, {{26156, 1, 0, 0},{26156, 1, 1, 17}, {0, 0, 0, 0}}},{13237, {{26156, 1, 0, 0},{26156, 1, 1, 17}, {0, 0, 0, 0}}},{57666, {{26156, 1, 0, 0},{26156, 1, 1, 18}, {0, 0, 0, 0}}},{57666, {{26156, 1, 0, 0},{26156, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13238, {{26157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13239, {{26158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13240, {{26159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13241, {{26160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13242, {{26161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13243, {{26162, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13244, {{26163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13245, {{26164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13246, {{26165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13247, {{26166, 1, 0, 0},{26166, 1, 1, 17}, {0, 0, 0, 0}}},{13247, {{26166, 1, 0, 0},{26166, 1, 1, 17}, {0, 0, 0, 0}}},{13248, {{26166, 1, 0, 0},{26166, 1, 1, 18}, {0, 0, 0, 0}}},{13248, {{26166, 1, 0, 0},{26166, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13249, {{26167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13250, {{26168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13251, {{26169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13252, {{26170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13253, {{26171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13254, {{26172, 1, 0, 0},{26172, 1, 1, 18}, {0, 0, 0, 0}}},{13254, {{26172, 1, 0, 0},{26172, 1, 1, 18}, {0, 0, 0, 0}}},{13255, {{26172, 1, 0, 0},{26172, 1, 1, 17}, {0, 0, 0, 0}}},{13255, {{26172, 1, 0, 0},{26172, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{13256, {{26175, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13257, {{26176, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13258, {{26177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13259, {{26178, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13260, {{26179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13261, {{26180, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13262, {{26181, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13263, {{26182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13264, {{26183, 1, 0, 0},{26183, 1, 1, 16}, {0, 0, 0, 0}}},{13264, {{26183, 1, 0, 0},{26183, 1, 1, 16}, {0, 0, 0, 0}}},{57665, {{26183, 1, 0, 0},{26183, 1, 1, 17}, {0, 0, 0, 0}}},{57665, {{26183, 1, 0, 0},{26183, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {6, {{13265, {{26184, 1, 0, 0},{26184, 1, 1, 17}, {0, 0, 0, 0}}},{13265, {{26184, 1, 0, 0},{26184, 1, 1, 17}, {0, 0, 0, 0}}},{13266, {{26184, 1, 0, 0},{26184, 1, 1, 19}, {0, 0, 0, 0}}},{13266, {{26184, 1, 0, 0},{26184, 1, 1, 19}, {0, 0, 0, 0}}},{13267, {{26184, 1, 0, 0},{26184, 1, 1, 18}, {0, 0, 0, 0}}},{13267, {{26184, 1, 0, 0},{26184, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13268, {{26185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13269, {{26186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13270, {{26187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13271, {{26188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13272, {{26189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13273, {{26190, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13274, {{26191, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13275, {{26192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13276, {{26193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13277, {{26194, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{13279, {{26199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13280, {{26200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13281, {{26201, 1, 0, 0},{26201, 1, 1, 18}, {0, 0, 0, 0}}},{13281, {{26201, 1, 0, 0},{26201, 1, 1, 18}, {0, 0, 0, 0}}},{13282, {{26201, 1, 0, 0},{26201, 1, 1, 17}, {0, 0, 0, 0}}},{13282, {{26201, 1, 0, 0},{26201, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{13283, {{26202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13284, {{26203, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13285, {{26204, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13286, {{26205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13287, {{26206, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13288, {{26207, 1, 0, 0},{26207, 1, 1, 20}, {0, 0, 0, 0}}},{13288, {{26207, 1, 0, 0},{26207, 1, 1, 20}, {0, 0, 0, 0}}},{13289, {{26207, 1, 0, 0},{26207, 1, 1, 21}, {0, 0, 0, 0}}},{13289, {{26207, 1, 0, 0},{26207, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{13290, {{26208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13291, {{26209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13292, {{26210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13293, {{26211, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13294, {{26212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13295, {{26213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13296, {{26214, 1, 0, 0},{26214, 1, 1, 18}, {0, 0, 0, 0}}},{13296, {{26214, 1, 0, 0},{26214, 1, 1, 18}, {0, 0, 0, 0}}},{13297, {{26214, 1, 0, 0},{26214, 1, 1, 19}, {0, 0, 0, 0}}},{13297, {{26214, 1, 0, 0},{26214, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{13298, {{26215, 1, 0, 0},{26215, 1, 1, 18}, {0, 0, 0, 0}}},{13298, {{26215, 1, 0, 0},{26215, 1, 1, 18}, {0, 0, 0, 0}}},{13299, {{26215, 1, 0, 0},{26215, 1, 1, 19}, {0, 0, 0, 0}}},{13299, {{26215, 1, 0, 0},{26215, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{13300, {{26216, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13301, {{26217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13302, {{26218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13303, {{26219, 1, 0, 0},{26219, 1, 1, 17}, {0, 0, 0, 0}}},{13303, {{26219, 1, 0, 0},{26219, 1, 1, 17}, {0, 0, 0, 0}}},{59691, {{26219, 1, 0, 0},{26219, 1, 1, 18}, {0, 0, 0, 0}}},{59691, {{26219, 1, 0, 0},{26219, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13304, {{26220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13305, {{26221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13307, {{26222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13308, {{26223, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13309, {{26224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13310, {{26225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13311, {{26226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13312, {{26227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{13313, {{26228, 1, 0, 0},{26228, 1, 1, 18}, {0, 0, 0, 0}}},{13313, {{26228, 1, 0, 0},{26228, 1, 1, 18}, {0, 0, 0, 0}}},{30199, {{26228, 1, 0, 0},{26228, 1, 1, 0},{26228, 1, 1, 19},{64018, 1, 0, 0}, {0, 0, 0, 0}}},{30199, {{26228, 1, 0, 0},{26228, 1, 1, 0},{26228, 1, 1, 19},{64018, 1, 0, 0}, {0, 0, 0, 0}}},{30199, {{26228, 1, 0, 0},{26228, 1, 1, 0},{26228, 1, 1, 19},{64018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13314, {{26229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13315, {{26230, 1, 0, 0},{26230, 1, 1, 17}, {0, 0, 0, 0}}},{13315, {{26230, 1, 0, 0},{26230, 1, 1, 17}, {0, 0, 0, 0}}},{59704, {{26230, 1, 0, 0},{26230, 1, 1, 18}, {0, 0, 0, 0}}},{59704, {{26230, 1, 0, 0},{26230, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{13316, {{26231, 1, 0, 0},{26231, 1, 1, 18}, {0, 0, 0, 0}}},{13316, {{26231, 1, 0, 0},{26231, 1, 1, 18}, {0, 0, 0, 0}}},{13317, {{26231, 1, 0, 0},{26231, 1, 1, 19}, {0, 0, 0, 0}}},{13317, {{26231, 1, 0, 0},{26231, 1, 1, 19}, {0, 0, 0, 0}}},{13318, {{26231, 1, 0, 0},{26231, 1, 1, 20}, {0, 0, 0, 0}}},{13318, {{26231, 1, 0, 0},{26231, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{13319, {{26232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13320, {{26233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13321, {{26234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13322, {{26235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13323, {{26236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13324, {{26237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13325, {{26238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13326, {{26239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13327, {{26240, 1, 0, 0},{26240, 1, 1, 17}, {0, 0, 0, 0}}},{13327, {{26240, 1, 0, 0},{26240, 1, 1, 17}, {0, 0, 0, 0}}},{13328, {{26240, 1, 0, 0},{26240, 1, 1, 18}, {0, 0, 0, 0}}},{13328, {{26240, 1, 0, 0},{26240, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{13329, {{26241, 1, 0, 0},{26241, 1, 1, 18}, {0, 0, 0, 0}}},{13329, {{26241, 1, 0, 0},{26241, 1, 1, 18}, {0, 0, 0, 0}}},{13330, {{26241, 1, 0, 0},{26241, 1, 1, 19}, {0, 0, 0, 0}}},{13330, {{26241, 1, 0, 0},{26241, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{13331, {{26243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13332, {{26244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{13333, {{26246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13334, {{26247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13335, {{26248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13336, {{26249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13337, {{26250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13338, {{26251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13339, {{26252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13340, {{26253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{13341, {{26254, 1, 0, 0},{26254, 1, 1, 17}, {0, 0, 0, 0}}},{13341, {{26254, 1, 0, 0},{26254, 1, 1, 17}, {0, 0, 0, 0}}},{13342, {{26254, 1, 0, 0},{26254, 1, 1, 18}, {0, 0, 0, 0}}},{13342, {{26254, 1, 0, 0},{26254, 1, 1, 18}, {0, 0, 0, 0}}},{13343, {{26254, 1, 0, 0},{26254, 1, 1, 19}, {0, 0, 0, 0}}},{13343, {{26254, 1, 0, 0},{26254, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{13344, {{26255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13345, {{26256, 1, 0, 0},{26256, 1, 1, 18}, {0, 0, 0, 0}}},{13345, {{26256, 1, 0, 0},{26256, 1, 1, 18}, {0, 0, 0, 0}}},{13346, {{26256, 1, 0, 0},{26256, 1, 1, 17}, {0, 0, 0, 0}}},{13346, {{26256, 1, 0, 0},{26256, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {5, {{13348, {{26257, 1, 0, 0},{26257, 1, 1, 18}, {0, 0, 0, 0}}},{13348, {{26257, 1, 0, 0},{26257, 1, 1, 18}, {0, 0, 0, 0}}},{30264, {{26257, 1, 0, 0},{26257, 1, 1, 0},{26257, 1, 1, 19},{64067, 1, 0, 0}, {0, 0, 0, 0}}},{30264, {{26257, 1, 0, 0},{26257, 1, 1, 0},{26257, 1, 1, 19},{64067, 1, 0, 0}, {0, 0, 0, 0}}},{30264, {{26257, 1, 0, 0},{26257, 1, 1, 0},{26257, 1, 1, 19},{64067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13349, {{26258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13350, {{26259, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13351, {{26260, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13352, {{26261, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13354, {{26262, 1, 0, 0},{26262, 1, 1, 19}, {0, 0, 0, 0}}},{13354, {{26262, 1, 0, 0},{26262, 1, 1, 19}, {0, 0, 0, 0}}},{13355, {{26262, 1, 0, 0},{26262, 1, 1, 18}, {0, 0, 0, 0}}},{13355, {{26262, 1, 0, 0},{26262, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{13356, {{26263, 1, 0, 0},{26263, 1, 1, 18}, {0, 0, 0, 0}}},{13356, {{26263, 1, 0, 0},{26263, 1, 1, 18}, {0, 0, 0, 0}}},{13357, {{26263, 1, 0, 0},{26263, 1, 1, 19}, {0, 0, 0, 0}}},{13357, {{26263, 1, 0, 0},{26263, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{13358, {{26264, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13359, {{26265, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13360, {{26266, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13362, {{26267, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13363, {{26268, 1, 0, 0},{26268, 1, 1, 17}, {0, 0, 0, 0}}},{13363, {{26268, 1, 0, 0},{26268, 1, 1, 17}, {0, 0, 0, 0}}},{13364, {{26268, 1, 0, 0},{26268, 1, 1, 18}, {0, 0, 0, 0}}},{13364, {{26268, 1, 0, 0},{26268, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13365, {{26269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13366, {{26270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13367, {{26271, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13368, {{26272, 1, 0, 0},{26272, 1, 1, 17}, {0, 0, 0, 0}}},{13368, {{26272, 1, 0, 0},{26272, 1, 1, 17}, {0, 0, 0, 0}}},{13369, {{26272, 1, 0, 0},{26272, 1, 1, 18}, {0, 0, 0, 0}}},{13369, {{26272, 1, 0, 0},{26272, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13370, {{26273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13371, {{26274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13372, {{26275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13373, {{26276, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13375, {{26277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13376, {{26278, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{13377, {{26280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13378, {{26281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13379, {{26282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13380, {{26283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13381, {{26284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13382, {{26285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13383, {{26286, 1, 0, 0},{26286, 1, 1, 17}, {0, 0, 0, 0}}},{13383, {{26286, 1, 0, 0},{26286, 1, 1, 17}, {0, 0, 0, 0}}},{13384, {{26286, 1, 0, 0},{26286, 1, 1, 18}, {0, 0, 0, 0}}},{13384, {{26286, 1, 0, 0},{26286, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13385, {{26287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13386, {{26288, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{13387, {{26289, 1, 0, 0},{26289, 1, 1, 17}, {0, 0, 0, 0}}},{13387, {{26289, 1, 0, 0},{26289, 1, 1, 17}, {0, 0, 0, 0}}},{13388, {{26289, 1, 0, 0},{26289, 1, 1, 19}, {0, 0, 0, 0}}},{13388, {{26289, 1, 0, 0},{26289, 1, 1, 19}, {0, 0, 0, 0}}},{13389, {{26289, 1, 0, 0},{26289, 1, 1, 18}, {0, 0, 0, 0}}},{13389, {{26289, 1, 0, 0},{26289, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13390, {{26290, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13391, {{26291, 1, 0, 0},{26291, 1, 1, 17}, {0, 0, 0, 0}}},{13391, {{26291, 1, 0, 0},{26291, 1, 1, 17}, {0, 0, 0, 0}}},{68052, {{26291, 1, 0, 0},{26291, 1, 1, 20}, {0, 0, 0, 0}}},{68052, {{26291, 1, 0, 0},{26291, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{13394, {{26292, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13395, {{26293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13396, {{26294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13397, {{26295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13398, {{26296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13400, {{26297, 1, 0, 0},{26297, 1, 1, 18}, {0, 0, 0, 0}}},{13400, {{26297, 1, 0, 0},{26297, 1, 1, 18}, {0, 0, 0, 0}}},{13401, {{26297, 1, 0, 0},{26297, 1, 1, 17}, {0, 0, 0, 0}}},{13401, {{26297, 1, 0, 0},{26297, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{13403, {{26298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13404, {{26299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13405, {{26300, 1, 0, 0},{26300, 1, 1, 17}, {0, 0, 0, 0}}},{13405, {{26300, 1, 0, 0},{26300, 1, 1, 17}, {0, 0, 0, 0}}},{13406, {{26300, 1, 0, 0},{26300, 1, 1, 18}, {0, 0, 0, 0}}},{13406, {{26300, 1, 0, 0},{26300, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13407, {{26301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13408, {{26302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13409, {{26303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13410, {{26304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{13411, {{26305, 1, 0, 0},{26305, 1, 1, 19}, {0, 0, 0, 0}}},{13411, {{26305, 1, 0, 0},{26305, 1, 1, 19}, {0, 0, 0, 0}}},{13412, {{26305, 1, 0, 0},{26305, 1, 1, 20}, {0, 0, 0, 0}}},{13412, {{26305, 1, 0, 0},{26305, 1, 1, 20}, {0, 0, 0, 0}}},{13413, {{26305, 1, 0, 0},{26305, 1, 1, 21}, {0, 0, 0, 0}}},{13413, {{26305, 1, 0, 0},{26305, 1, 1, 21}, {0, 0, 0, 0}}},{13414, {{26305, 1, 0, 0},{26305, 1, 1, 22}, {0, 0, 0, 0}}},{13414, {{26305, 1, 0, 0},{26305, 1, 1, 22}, {0, 0, 0, 0}}}}},
    {1, {{13416, {{26306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13417, {{26307, 1, 0, 0},{26307, 1, 1, 18}, {0, 0, 0, 0}}},{13417, {{26307, 1, 0, 0},{26307, 1, 1, 18}, {0, 0, 0, 0}}},{13418, {{26307, 1, 0, 0},{26307, 1, 1, 17}, {0, 0, 0, 0}}},{13418, {{26307, 1, 0, 0},{26307, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{13419, {{26308, 1, 0, 0},{26308, 1, 1, 17}, {0, 0, 0, 0}}},{13419, {{26308, 1, 0, 0},{26308, 1, 1, 17}, {0, 0, 0, 0}}},{13420, {{26308, 1, 0, 0},{26308, 1, 1, 18}, {0, 0, 0, 0}}},{13420, {{26308, 1, 0, 0},{26308, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{13421, {{26309, 1, 0, 0},{26309, 1, 1, 17}, {0, 0, 0, 0}}},{13421, {{26309, 1, 0, 0},{26309, 1, 1, 17}, {0, 0, 0, 0}}},{13422, {{26309, 1, 0, 0},{26309, 1, 1, 16}, {0, 0, 0, 0}}},{13422, {{26309, 1, 0, 0},{26309, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{13423, {{26310, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13424, {{26311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13425, {{26312, 1, 0, 0},{26312, 1, 1, 17}, {0, 0, 0, 0}}},{13425, {{26312, 1, 0, 0},{26312, 1, 1, 17}, {0, 0, 0, 0}}},{57680, {{26312, 1, 0, 0},{26312, 1, 1, 18}, {0, 0, 0, 0}}},{57680, {{26312, 1, 0, 0},{26312, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{13426, {{26313, 1, 0, 0},{26313, 1, 1, 17}, {0, 0, 0, 0}}},{13426, {{26313, 1, 0, 0},{26313, 1, 1, 17}, {0, 0, 0, 0}}},{59694, {{26313, 1, 0, 0},{26313, 1, 1, 18}, {0, 0, 0, 0}}},{59694, {{26313, 1, 0, 0},{26313, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13427, {{26314, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13428, {{26315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13429, {{26316, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57679, {{26317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13431, {{26318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13432, {{26319, 1, 0, 0},{26319, 1, 1, 17}, {0, 0, 0, 0}}},{13432, {{26319, 1, 0, 0},{26319, 1, 1, 17}, {0, 0, 0, 0}}},{13433, {{26319, 1, 0, 0},{26319, 1, 1, 18}, {0, 0, 0, 0}}},{13433, {{26319, 1, 0, 0},{26319, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13437, {{26320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13438, {{26321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13439, {{26322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13440, {{26323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13441, {{26324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13442, {{26325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13443, {{26326, 1, 0, 0},{26326, 1, 1, 18}, {0, 0, 0, 0}}},{13443, {{26326, 1, 0, 0},{26326, 1, 1, 18}, {0, 0, 0, 0}}},{13444, {{26326, 1, 0, 0},{26326, 1, 1, 17}, {0, 0, 0, 0}}},{13444, {{26326, 1, 0, 0},{26326, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{13445, {{26327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13446, {{26328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13447, {{26329, 1, 0, 0},{26329, 1, 1, 18}, {0, 0, 0, 0}}},{13447, {{26329, 1, 0, 0},{26329, 1, 1, 18}, {0, 0, 0, 0}}},{13448, {{26329, 1, 0, 0},{26329, 1, 1, 19}, {0, 0, 0, 0}}},{13448, {{26329, 1, 0, 0},{26329, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {6, {{13449, {{26330, 1, 0, 0},{26330, 1, 1, 17}, {0, 0, 0, 0}}},{13449, {{26330, 1, 0, 0},{26330, 1, 1, 17}, {0, 0, 0, 0}}},{13450, {{26330, 1, 0, 0},{26330, 1, 1, 18}, {0, 0, 0, 0}}},{13450, {{26330, 1, 0, 0},{26330, 1, 1, 18}, {0, 0, 0, 0}}},{13451, {{26330, 1, 0, 0},{26330, 1, 1, 19}, {0, 0, 0, 0}}},{13451, {{26330, 1, 0, 0},{26330, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{13452, {{26331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13453, {{26332, 1, 0, 0},{26332, 1, 1, 19}, {0, 0, 0, 0}}},{13453, {{26332, 1, 0, 0},{26332, 1, 1, 19}, {0, 0, 0, 0}}},{13454, {{26332, 1, 0, 0},{26332, 1, 1, 18}, {0, 0, 0, 0}}},{13454, {{26332, 1, 0, 0},{26332, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13455, {{26333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13456, {{26334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13458, {{26335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13459, {{26336, 1, 0, 0},{26336, 1, 1, 17}, {0, 0, 0, 0}}},{13459, {{26336, 1, 0, 0},{26336, 1, 1, 17}, {0, 0, 0, 0}}},{13460, {{26336, 1, 0, 0},{26336, 1, 1, 18}, {0, 0, 0, 0}}},{13460, {{26336, 1, 0, 0},{26336, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{57682, {{26337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{13463, {{26339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13464, {{26340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13465, {{26341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13466, {{26342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13467, {{26343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13469, {{26344, 1, 0, 0},{26344, 1, 1, 17}, {0, 0, 0, 0}}},{13469, {{26344, 1, 0, 0},{26344, 1, 1, 17}, {0, 0, 0, 0}}},{13470, {{26344, 1, 0, 0},{26344, 1, 1, 18}, {0, 0, 0, 0}}},{13470, {{26344, 1, 0, 0},{26344, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13471, {{26345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13472, {{26346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13473, {{26347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13474, {{26348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13475, {{26349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13476, {{26350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13477, {{26351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13478, {{26352, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13479, {{26353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13480, {{26354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13481, {{26355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13482, {{26356, 1, 0, 0},{26356, 1, 1, 18}, {0, 0, 0, 0}}},{13482, {{26356, 1, 0, 0},{26356, 1, 1, 18}, {0, 0, 0, 0}}},{13483, {{26356, 1, 0, 0},{26356, 1, 1, 19}, {0, 0, 0, 0}}},{13483, {{26356, 1, 0, 0},{26356, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {6, {{13484, {{26357, 1, 0, 0},{26357, 1, 1, 18}, {0, 0, 0, 0}}},{13484, {{26357, 1, 0, 0},{26357, 1, 1, 18}, {0, 0, 0, 0}}},{13485, {{26357, 1, 0, 0},{26357, 1, 1, 19}, {0, 0, 0, 0}}},{13485, {{26357, 1, 0, 0},{26357, 1, 1, 19}, {0, 0, 0, 0}}},{13486, {{26357, 1, 0, 0},{26357, 1, 1, 20}, {0, 0, 0, 0}}},{13486, {{26357, 1, 0, 0},{26357, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{13487, {{26358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13488, {{26359, 1, 0, 0},{26359, 1, 1, 17}, {0, 0, 0, 0}}},{13488, {{26359, 1, 0, 0},{26359, 1, 1, 17}, {0, 0, 0, 0}}},{13489, {{26359, 1, 0, 0},{26359, 1, 1, 18}, {0, 0, 0, 0}}},{13489, {{26359, 1, 0, 0},{26359, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{13491, {{26360, 1, 0, 0},{26360, 1, 1, 18}, {0, 0, 0, 0}}},{13491, {{26360, 1, 0, 0},{26360, 1, 1, 18}, {0, 0, 0, 0}}},{13492, {{26360, 1, 0, 0},{26360, 1, 1, 19}, {0, 0, 0, 0}}},{13492, {{26360, 1, 0, 0},{26360, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{13493, {{26361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13494, {{26362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{13495, {{26363, 1, 0, 0},{26363, 1, 1, 19}, {0, 0, 0, 0}}},{13495, {{26363, 1, 0, 0},{26363, 1, 1, 19}, {0, 0, 0, 0}}},{13496, {{26363, 1, 0, 0},{26363, 1, 1, 18}, {0, 0, 0, 0}}},{13496, {{26363, 1, 0, 0},{26363, 1, 1, 18}, {0, 0, 0, 0}}},{13497, {{26363, 1, 0, 0},{26363, 1, 1, 20}, {0, 0, 0, 0}}},{13497, {{26363, 1, 0, 0},{26363, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{13498, {{26364, 1, 0, 0},{26364, 1, 1, 18}, {0, 0, 0, 0}}},{13498, {{26364, 1, 0, 0},{26364, 1, 1, 18}, {0, 0, 0, 0}}},{13499, {{26364, 1, 0, 0},{26364, 1, 1, 19}, {0, 0, 0, 0}}},{13499, {{26364, 1, 0, 0},{26364, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{13500, {{26365, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{13501, {{26366, 1, 0, 0},{26366, 1, 1, 17}, {0, 0, 0, 0}}},{13501, {{26366, 1, 0, 0},{26366, 1, 1, 17}, {0, 0, 0, 0}}},{13502, {{26366, 1, 0, 0},{26366, 1, 1, 18}, {0, 0, 0, 0}}},{13502, {{26366, 1, 0, 0},{26366, 1, 1, 18}, {0, 0, 0, 0}}},{13503, {{26366, 1, 0, 0},{26366, 1, 1, 19}, {0, 0, 0, 0}}},{13503, {{26366, 1, 0, 0},{26366, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{13504, {{26367, 1, 0, 0},{26367, 1, 1, 17}, {0, 0, 0, 0}}},{13504, {{26367, 1, 0, 0},{26367, 1, 1, 17}, {0, 0, 0, 0}}},{57672, {{26367, 1, 0, 0},{26367, 1, 1, 18}, {0, 0, 0, 0}}},{57672, {{26367, 1, 0, 0},{26367, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{13505, {{26368, 1, 0, 0},{26368, 1, 1, 19}, {0, 0, 0, 0}}},{13505, {{26368, 1, 0, 0},{26368, 1, 1, 19}, {0, 0, 0, 0}}},{13506, {{26368, 1, 0, 0},{26368, 1, 1, 18}, {0, 0, 0, 0}}},{13506, {{26368, 1, 0, 0},{26368, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{13507, {{26369, 1, 0, 0},{26369, 1, 1, 18}, {0, 0, 0, 0}}},{13507, {{26369, 1, 0, 0},{26369, 1, 1, 18}, {0, 0, 0, 0}}},{13508, {{26369, 1, 0, 0},{26369, 1, 1, 17}, {0, 0, 0, 0}}},{13508, {{26369, 1, 0, 0},{26369, 1, 1, 17}, {0, 0, 0, 0}}},{57673, {{26369, 1, 0, 0},{26369, 1, 1, 19}, {0, 0, 0, 0}}},{57673, {{26369, 1, 0, 0},{26369, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{13509, {{26370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{13510, {{26371, 1, 0, 0},{26371, 1, 1, 17}, {0, 0, 0, 0}}},{13510, {{26371, 1, 0, 0},{26371, 1, 1, 17}, {0, 0, 0, 0}}},{13512, {{26371, 1, 0, 0},{26371, 1, 1, 19}, {0, 0, 0, 0}}},{13512, {{26371, 1, 0, 0},{26371, 1, 1, 19}, {0, 0, 0, 0}}},{59330, {{26371, 1, 0, 0},{26371, 1, 1, 18}, {0, 0, 0, 0}}},{59330, {{26371, 1, 0, 0},{26371, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13513, {{26372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13514, {{26373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13515, {{26374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13516, {{26375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13519, {{26376, 1, 0, 0},{26376, 1, 1, 19}, {0, 0, 0, 0}}},{13519, {{26376, 1, 0, 0},{26376, 1, 1, 19}, {0, 0, 0, 0}}},{13520, {{26376, 1, 0, 0},{26376, 1, 1, 18}, {0, 0, 0, 0}}},{13520, {{26376, 1, 0, 0},{26376, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{13521, {{26377, 1, 0, 0},{26377, 1, 1, 18}, {0, 0, 0, 0}}},{13521, {{26377, 1, 0, 0},{26377, 1, 1, 18}, {0, 0, 0, 0}}},{13522, {{26377, 1, 0, 0},{26377, 1, 1, 19}, {0, 0, 0, 0}}},{13522, {{26377, 1, 0, 0},{26377, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{13523, {{26378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13524, {{26379, 1, 0, 0},{26379, 1, 1, 19}, {0, 0, 0, 0}}},{13524, {{26379, 1, 0, 0},{26379, 1, 1, 19}, {0, 0, 0, 0}}},{13525, {{26379, 1, 0, 0},{26379, 1, 1, 18}, {0, 0, 0, 0}}},{13525, {{26379, 1, 0, 0},{26379, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13526, {{26380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13527, {{26381, 1, 0, 0},{26381, 1, 1, 19}, {0, 0, 0, 0}}},{13527, {{26381, 1, 0, 0},{26381, 1, 1, 19}, {0, 0, 0, 0}}},{13528, {{26381, 1, 0, 0},{26381, 1, 1, 18}, {0, 0, 0, 0}}},{13528, {{26381, 1, 0, 0},{26381, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{13529, {{26382, 1, 0, 0},{26382, 1, 1, 17}, {0, 0, 0, 0}}},{13529, {{26382, 1, 0, 0},{26382, 1, 1, 17}, {0, 0, 0, 0}}},{13530, {{26382, 1, 0, 0},{26382, 1, 1, 18}, {0, 0, 0, 0}}},{13530, {{26382, 1, 0, 0},{26382, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13531, {{26383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13532, {{26384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{13533, {{26386, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13534, {{26387, 1, 0, 0},{26387, 1, 1, 17}, {0, 0, 0, 0}}},{13534, {{26387, 1, 0, 0},{26387, 1, 1, 17}, {0, 0, 0, 0}}},{13535, {{26387, 1, 0, 0},{26387, 1, 1, 18}, {0, 0, 0, 0}}},{13535, {{26387, 1, 0, 0},{26387, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{13536, {{26388, 1, 0, 0},{26388, 1, 1, 17}, {0, 0, 0, 0}}},{13536, {{26388, 1, 0, 0},{26388, 1, 1, 17}, {0, 0, 0, 0}}},{13537, {{26388, 1, 0, 0},{26388, 1, 1, 18}, {0, 0, 0, 0}}},{13537, {{26388, 1, 0, 0},{26388, 1, 1, 18}, {0, 0, 0, 0}}},{57685, {{26388, 1, 0, 0},{26388, 1, 1, 19}, {0, 0, 0, 0}}},{57685, {{26388, 1, 0, 0},{26388, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{13539, {{26389, 1, 0, 0},{26389, 1, 1, 20}, {0, 0, 0, 0}}},{13539, {{26389, 1, 0, 0},{26389, 1, 1, 20}, {0, 0, 0, 0}}},{13540, {{26389, 1, 0, 0},{26389, 1, 1, 19}, {0, 0, 0, 0}}},{13540, {{26389, 1, 0, 0},{26389, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{13541, {{26390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {7, {{13542, {{26391, 1, 0, 0},{26391, 1, 1, 20}, {0, 0, 0, 0}}},{13542, {{26391, 1, 0, 0},{26391, 1, 1, 20}, {0, 0, 0, 0}}},{13543, {{26391, 1, 0, 0},{26391, 1, 1, 21}, {0, 0, 0, 0}}},{13543, {{26391, 1, 0, 0},{26391, 1, 1, 21}, {0, 0, 0, 0}}},{30185, {{26391, 1, 0, 0},{26391, 1, 1, 0},{26391, 1, 1, 22},{63785, 1, 0, 0}, {0, 0, 0, 0}}},{30185, {{26391, 1, 0, 0},{26391, 1, 1, 0},{26391, 1, 1, 22},{63785, 1, 0, 0}, {0, 0, 0, 0}}},{30185, {{26391, 1, 0, 0},{26391, 1, 1, 0},{26391, 1, 1, 22},{63785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21070, {{26392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13545, {{26393, 1, 0, 0},{26393, 1, 1, 17}, {0, 0, 0, 0}}},{13545, {{26393, 1, 0, 0},{26393, 1, 1, 17}, {0, 0, 0, 0}}},{13546, {{26393, 1, 0, 0},{26393, 1, 1, 18}, {0, 0, 0, 0}}},{13546, {{26393, 1, 0, 0},{26393, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{13547, {{26394, 1, 0, 0},{26394, 1, 1, 16}, {0, 0, 0, 0}}},{13547, {{26394, 1, 0, 0},{26394, 1, 1, 16}, {0, 0, 0, 0}}},{13548, {{26394, 1, 0, 0},{26394, 1, 1, 17}, {0, 0, 0, 0}}},{13548, {{26394, 1, 0, 0},{26394, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {12, {{13549, {{26395, 1, 0, 0},{26395, 1, 1, 23}, {0, 0, 0, 0}}},{13549, {{26395, 1, 0, 0},{26395, 1, 1, 23}, {0, 0, 0, 0}}},{13550, {{26395, 1, 0, 0},{26395, 1, 1, 19}, {0, 0, 0, 0}}},{13550, {{26395, 1, 0, 0},{26395, 1, 1, 19}, {0, 0, 0, 0}}},{13551, {{26395, 1, 0, 0},{26395, 1, 1, 26}, {0, 0, 0, 0}}},{13551, {{26395, 1, 0, 0},{26395, 1, 1, 26}, {0, 0, 0, 0}}},{13552, {{26395, 1, 0, 0},{26395, 1, 1, 22}, {0, 0, 0, 0}}},{13552, {{26395, 1, 0, 0},{26395, 1, 1, 22}, {0, 0, 0, 0}}},{13553, {{26395, 1, 0, 0},{26395, 1, 1, 21}, {0, 0, 0, 0}}},{13553, {{26395, 1, 0, 0},{26395, 1, 1, 21}, {0, 0, 0, 0}}},{13554, {{26395, 1, 0, 0},{26395, 1, 1, 20}, {0, 0, 0, 0}}},{13554, {{26395, 1, 0, 0},{26395, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{13556, {{26396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{13557, {{26397, 1, 0, 0},{26397, 1, 1, 19}, {0, 0, 0, 0}}},{13557, {{26397, 1, 0, 0},{26397, 1, 1, 19}, {0, 0, 0, 0}}},{13558, {{26397, 1, 0, 0},{26397, 1, 1, 18}, {0, 0, 0, 0}}},{13558, {{26397, 1, 0, 0},{26397, 1, 1, 18}, {0, 0, 0, 0}}},{57690, {{26397, 1, 0, 0},{26397, 1, 1, 20}, {0, 0, 0, 0}}},{57690, {{26397, 1, 0, 0},{26397, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {6, {{13559, {{26398, 1, 0, 0},{26398, 1, 1, 17}, {0, 0, 0, 0}}},{13559, {{26398, 1, 0, 0},{26398, 1, 1, 17}, {0, 0, 0, 0}}},{13560, {{26398, 1, 0, 0},{26398, 1, 1, 18}, {0, 0, 0, 0}}},{13560, {{26398, 1, 0, 0},{26398, 1, 1, 18}, {0, 0, 0, 0}}},{13561, {{26398, 1, 0, 0},{26398, 1, 1, 19}, {0, 0, 0, 0}}},{13561, {{26398, 1, 0, 0},{26398, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{13562, {{26399, 1, 0, 0},{26399, 1, 1, 18}, {0, 0, 0, 0}}},{13562, {{26399, 1, 0, 0},{26399, 1, 1, 18}, {0, 0, 0, 0}}},{13563, {{26399, 1, 0, 0},{26399, 1, 1, 19}, {0, 0, 0, 0}}},{13563, {{26399, 1, 0, 0},{26399, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{13564, {{26400, 1, 0, 0},{26400, 1, 1, 17}, {0, 0, 0, 0}}},{13564, {{26400, 1, 0, 0},{26400, 1, 1, 17}, {0, 0, 0, 0}}},{13565, {{26400, 1, 0, 0},{26400, 1, 1, 18}, {0, 0, 0, 0}}},{13565, {{26400, 1, 0, 0},{26400, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{13566, {{26401, 1, 0, 0},{26401, 1, 1, 17}, {0, 0, 0, 0}}},{13566, {{26401, 1, 0, 0},{26401, 1, 1, 17}, {0, 0, 0, 0}}},{13567, {{26401, 1, 0, 0},{26401, 1, 1, 16}, {0, 0, 0, 0}}},{13567, {{26401, 1, 0, 0},{26401, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {4, {{13568, {{26402, 1, 0, 0},{26402, 1, 1, 17}, {0, 0, 0, 0}}},{13568, {{26402, 1, 0, 0},{26402, 1, 1, 17}, {0, 0, 0, 0}}},{13569, {{26402, 1, 0, 0},{26402, 1, 1, 18}, {0, 0, 0, 0}}},{13569, {{26402, 1, 0, 0},{26402, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{13570, {{26403, 1, 0, 0},{26403, 1, 1, 16}, {0, 0, 0, 0}}},{13570, {{26403, 1, 0, 0},{26403, 1, 1, 16}, {0, 0, 0, 0}}},{57693, {{26403, 1, 0, 0},{26403, 1, 1, 17}, {0, 0, 0, 0}}},{57693, {{26403, 1, 0, 0},{26403, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{13571, {{26404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{13572, {{26406, 1, 0, 0},{26406, 1, 1, 17}, {0, 0, 0, 0}}},{13572, {{26406, 1, 0, 0},{26406, 1, 1, 17}, {0, 0, 0, 0}}},{13573, {{26406, 1, 0, 0},{26406, 1, 1, 18}, {0, 0, 0, 0}}},{13573, {{26406, 1, 0, 0},{26406, 1, 1, 18}, {0, 0, 0, 0}}},{13574, {{26406, 1, 0, 0},{26406, 1, 1, 19}, {0, 0, 0, 0}}},{13574, {{26406, 1, 0, 0},{26406, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{13575, {{26407, 1, 0, 0},{26407, 1, 1, 17}, {0, 0, 0, 0}}},{13575, {{26407, 1, 0, 0},{26407, 1, 1, 17}, {0, 0, 0, 0}}},{45852, {{26407, 1, 0, 0},{26407, 1, 1, 18}, {0, 0, 0, 0}}},{45852, {{26407, 1, 0, 0},{26407, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13576, {{26408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57695, {{26409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13577, {{26410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13578, {{26411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13579, {{26412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13580, {{26413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13581, {{26414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57697, {{26415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{13583, {{26417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13584, {{26418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13585, {{26419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13586, {{26420, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13587, {{26421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13588, {{26422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13589, {{26423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13590, {{26424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13591, {{26425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13592, {{26426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13593, {{26427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13594, {{26428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13595, {{26429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13596, {{26430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13597, {{26431, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57698, {{26432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13598, {{26433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{13599, {{26435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13600, {{26436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13601, {{26437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13602, {{26438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13603, {{26439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13604, {{26440, 1, 0, 0},{26440, 1, 1, 17}, {0, 0, 0, 0}}},{13604, {{26440, 1, 0, 0},{26440, 1, 1, 17}, {0, 0, 0, 0}}},{13605, {{26440, 1, 0, 0},{26440, 1, 1, 18}, {0, 0, 0, 0}}},{13605, {{26440, 1, 0, 0},{26440, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13606, {{26441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13607, {{26442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13608, {{26443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13609, {{26444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13610, {{26445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13611, {{26446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13612, {{26447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13613, {{26448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13614, {{26449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13615, {{26450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13616, {{26451, 1, 0, 0},{26451, 1, 1, 19}, {0, 0, 0, 0}}},{13616, {{26451, 1, 0, 0},{26451, 1, 1, 19}, {0, 0, 0, 0}}},{13617, {{26451, 1, 0, 0},{26451, 1, 1, 18}, {0, 0, 0, 0}}},{13617, {{26451, 1, 0, 0},{26451, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13618, {{26452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13619, {{26453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{13620, {{26454, 1, 0, 0},{26454, 1, 1, 21}, {0, 0, 0, 0}}},{13620, {{26454, 1, 0, 0},{26454, 1, 1, 21}, {0, 0, 0, 0}}},{13621, {{26454, 1, 0, 0},{26454, 1, 1, 19}, {0, 0, 0, 0}}},{13621, {{26454, 1, 0, 0},{26454, 1, 1, 19}, {0, 0, 0, 0}}},{13622, {{26454, 1, 0, 0},{26454, 1, 1, 22}, {0, 0, 0, 0}}},{13622, {{26454, 1, 0, 0},{26454, 1, 1, 22}, {0, 0, 0, 0}}}}},
    {4, {{13623, {{26455, 1, 0, 0},{26455, 1, 1, 16}, {0, 0, 0, 0}}},{13623, {{26455, 1, 0, 0},{26455, 1, 1, 16}, {0, 0, 0, 0}}},{13624, {{26455, 1, 0, 0},{26455, 1, 1, 17}, {0, 0, 0, 0}}},{13624, {{26455, 1, 0, 0},{26455, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{13625, {{26456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13626, {{26457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13627, {{26458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13628, {{26459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13629, {{26460, 1, 0, 0},{26460, 1, 1, 17}, {0, 0, 0, 0}}},{13629, {{26460, 1, 0, 0},{26460, 1, 1, 17}, {0, 0, 0, 0}}},{57711, {{26460, 1, 0, 0},{26460, 1, 1, 18}, {0, 0, 0, 0}}},{57711, {{26460, 1, 0, 0},{26460, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13630, {{26461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13631, {{26462, 1, 0, 0},{26462, 1, 1, 18}, {0, 0, 0, 0}}},{13631, {{26462, 1, 0, 0},{26462, 1, 1, 18}, {0, 0, 0, 0}}},{13632, {{26462, 1, 0, 0},{26462, 1, 1, 19}, {0, 0, 0, 0}}},{13632, {{26462, 1, 0, 0},{26462, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{13634, {{26463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13635, {{26464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13636, {{26465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13638, {{26466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13639, {{26467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13640, {{26468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13641, {{26469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13642, {{26470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{13646, {{26474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13647, {{26475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13648, {{26476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13649, {{26477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13651, {{26478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13652, {{26479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13653, {{26480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13654, {{26481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13655, {{26482, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13656, {{26483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13657, {{26484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13658, {{26485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13659, {{26486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13660, {{26487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13661, {{26488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13662, {{26489, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13663, {{26490, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13664, {{26491, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13665, {{26492, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13666, {{26493, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13667, {{26494, 1, 0, 0},{26494, 1, 1, 18}, {0, 0, 0, 0}}},{13667, {{26494, 1, 0, 0},{26494, 1, 1, 18}, {0, 0, 0, 0}}},{13668, {{26494, 1, 0, 0},{26494, 1, 1, 19}, {0, 0, 0, 0}}},{13668, {{26494, 1, 0, 0},{26494, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{13669, {{26495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13670, {{26496, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13671, {{26497, 1, 0, 0},{26497, 1, 1, 18}, {0, 0, 0, 0}}},{13671, {{26497, 1, 0, 0},{26497, 1, 1, 18}, {0, 0, 0, 0}}},{13672, {{26497, 1, 0, 0},{26497, 1, 1, 17}, {0, 0, 0, 0}}},{13672, {{26497, 1, 0, 0},{26497, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{13673, {{26498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13674, {{26499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13676, {{26500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{13677, {{26501, 1, 0, 0},{26501, 1, 1, 17}, {0, 0, 0, 0}}},{13677, {{26501, 1, 0, 0},{26501, 1, 1, 17}, {0, 0, 0, 0}}},{13678, {{26501, 1, 0, 0},{26501, 1, 1, 18}, {0, 0, 0, 0}}},{13678, {{26501, 1, 0, 0},{26501, 1, 1, 18}, {0, 0, 0, 0}}},{13679, {{26501, 1, 0, 0},{26501, 1, 1, 19}, {0, 0, 0, 0}}},{13679, {{26501, 1, 0, 0},{26501, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{13681, {{26502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13682, {{26503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13683, {{26504, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13684, {{26505, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13685, {{26506, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13686, {{26507, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13687, {{26508, 1, 0, 0},{26508, 1, 1, 17}, {0, 0, 0, 0}}},{13687, {{26508, 1, 0, 0},{26508, 1, 1, 17}, {0, 0, 0, 0}}},{13688, {{26508, 1, 0, 0},{26508, 1, 1, 18}, {0, 0, 0, 0}}},{13688, {{26508, 1, 0, 0},{26508, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13689, {{26509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13690, {{26510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13691, {{26511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13692, {{26512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13693, {{26513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13694, {{26514, 1, 0, 0},{26514, 1, 1, 18}, {0, 0, 0, 0}}},{13694, {{26514, 1, 0, 0},{26514, 1, 1, 18}, {0, 0, 0, 0}}},{13695, {{26514, 1, 0, 0},{26514, 1, 1, 17}, {0, 0, 0, 0}}},{13695, {{26514, 1, 0, 0},{26514, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{13696, {{26515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13697, {{26516, 1, 0, 0},{26516, 1, 1, 16}, {0, 0, 0, 0}}},{13697, {{26516, 1, 0, 0},{26516, 1, 1, 16}, {0, 0, 0, 0}}},{59710, {{26516, 1, 0, 0},{26516, 1, 1, 17}, {0, 0, 0, 0}}},{59710, {{26516, 1, 0, 0},{26516, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{13698, {{26517, 1, 0, 0},{26517, 1, 1, 17}, {0, 0, 0, 0}}},{13698, {{26517, 1, 0, 0},{26517, 1, 1, 17}, {0, 0, 0, 0}}},{57718, {{26517, 1, 0, 0},{26517, 1, 1, 18}, {0, 0, 0, 0}}},{57718, {{26517, 1, 0, 0},{26517, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13699, {{26518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13700, {{26519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13701, {{26520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13702, {{26521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13703, {{26522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13704, {{26523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13705, {{26524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13706, {{26525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13707, {{26526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13708, {{26527, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13709, {{26528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13710, {{26529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13711, {{26530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30472, {{26531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{13712, {{26534, 1, 0, 0},{26534, 1, 1, 18}, {0, 0, 0, 0}}},{13712, {{26534, 1, 0, 0},{26534, 1, 1, 18}, {0, 0, 0, 0}}},{13713, {{26534, 1, 0, 0},{26534, 1, 1, 19}, {0, 0, 0, 0}}},{13713, {{26534, 1, 0, 0},{26534, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{13582, {{26537, 1, 0, 0},{26537, 1, 1, 20}, {0, 0, 0, 0}}},{13582, {{26537, 1, 0, 0},{26537, 1, 1, 20}, {0, 0, 0, 0}}},{13714, {{26537, 1, 0, 0},{26537, 1, 1, 18}, {0, 0, 0, 0}}},{13714, {{26537, 1, 0, 0},{26537, 1, 1, 18}, {0, 0, 0, 0}}},{13715, {{26537, 1, 0, 0},{26537, 1, 1, 19}, {0, 0, 0, 0}}},{13715, {{26537, 1, 0, 0},{26537, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{13717, {{26540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{13718, {{26542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13719, {{26543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13720, {{26544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13722, {{26545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13723, {{26546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13724, {{26547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13725, {{26548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13726, {{26549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13727, {{26550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13728, {{26551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13729, {{26552, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13730, {{26553, 1, 0, 0},{26553, 1, 1, 17}, {0, 0, 0, 0}}},{13730, {{26553, 1, 0, 0},{26553, 1, 1, 17}, {0, 0, 0, 0}}},{13731, {{26553, 1, 0, 0},{26553, 1, 1, 18}, {0, 0, 0, 0}}},{13731, {{26553, 1, 0, 0},{26553, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13732, {{26554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13733, {{26555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13734, {{26556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13735, {{26557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13736, {{26558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13737, {{26559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13738, {{26560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13739, {{26561, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13740, {{26562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13741, {{26563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13742, {{26564, 1, 0, 0},{26564, 1, 1, 18}, {0, 0, 0, 0}}},{13742, {{26564, 1, 0, 0},{26564, 1, 1, 18}, {0, 0, 0, 0}}},{13743, {{26564, 1, 0, 0},{26564, 1, 1, 19}, {0, 0, 0, 0}}},{13743, {{26564, 1, 0, 0},{26564, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{13744, {{26565, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13745, {{26566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13746, {{26567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13747, {{26568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13749, {{26569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13750, {{26570, 1, 0, 0},{26570, 1, 1, 19}, {0, 0, 0, 0}}},{13750, {{26570, 1, 0, 0},{26570, 1, 1, 19}, {0, 0, 0, 0}}},{13751, {{26570, 1, 0, 0},{26570, 1, 1, 18}, {0, 0, 0, 0}}},{13751, {{26570, 1, 0, 0},{26570, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13752, {{26571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13753, {{26572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13754, {{26573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13755, {{26574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13756, {{26575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13757, {{26576, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13758, {{26577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13759, {{26578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13760, {{26579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13761, {{26580, 1, 0, 0},{26580, 1, 1, 18}, {0, 0, 0, 0}}},{13761, {{26580, 1, 0, 0},{26580, 1, 1, 18}, {0, 0, 0, 0}}},{13762, {{26580, 1, 0, 0},{26580, 1, 1, 19}, {0, 0, 0, 0}}},{13762, {{26580, 1, 0, 0},{26580, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{13763, {{26581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13764, {{26582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13765, {{26583, 1, 0, 0},{26583, 1, 1, 17}, {0, 0, 0, 0}}},{13765, {{26583, 1, 0, 0},{26583, 1, 1, 17}, {0, 0, 0, 0}}},{59711, {{26583, 1, 0, 0},{26583, 1, 1, 18}, {0, 0, 0, 0}}},{59711, {{26583, 1, 0, 0},{26583, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13766, {{26584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13767, {{26585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13768, {{26586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13769, {{26587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13770, {{26588, 1, 0, 0},{26588, 1, 1, 17}, {0, 0, 0, 0}}},{13770, {{26588, 1, 0, 0},{26588, 1, 1, 17}, {0, 0, 0, 0}}},{13771, {{26588, 1, 0, 0},{26588, 1, 1, 18}, {0, 0, 0, 0}}},{13771, {{26588, 1, 0, 0},{26588, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13772, {{26589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13773, {{26590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13774, {{26591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13776, {{26592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13777, {{26593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13778, {{26594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13779, {{26595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13780, {{26596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13781, {{26597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13782, {{26598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13783, {{26599, 1, 0, 0},{26599, 1, 1, 19}, {0, 0, 0, 0}}},{13783, {{26599, 1, 0, 0},{26599, 1, 1, 19}, {0, 0, 0, 0}}},{13784, {{26599, 1, 0, 0},{26599, 1, 1, 18}, {0, 0, 0, 0}}},{13784, {{26599, 1, 0, 0},{26599, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13785, {{26600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13786, {{26601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13787, {{26602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13788, {{26603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13789, {{26604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13790, {{26605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13791, {{26606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13792, {{26607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13793, {{26608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13794, {{26609, 1, 0, 0},{26609, 1, 1, 18}, {0, 0, 0, 0}}},{13794, {{26609, 1, 0, 0},{26609, 1, 1, 18}, {0, 0, 0, 0}}},{13795, {{26609, 1, 0, 0},{26609, 1, 1, 19}, {0, 0, 0, 0}}},{13795, {{26609, 1, 0, 0},{26609, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{13796, {{26610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13797, {{26611, 1, 0, 0},{26611, 1, 1, 17}, {0, 0, 0, 0}}},{13797, {{26611, 1, 0, 0},{26611, 1, 1, 17}, {0, 0, 0, 0}}},{13798, {{26611, 1, 0, 0},{26611, 1, 1, 18}, {0, 0, 0, 0}}},{13798, {{26611, 1, 0, 0},{26611, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13799, {{26612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13800, {{26613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13801, {{26614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13802, {{26615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13803, {{26616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13804, {{26617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13806, {{26618, 1, 0, 0},{26618, 1, 1, 18}, {0, 0, 0, 0}}},{13806, {{26618, 1, 0, 0},{26618, 1, 1, 18}, {0, 0, 0, 0}}},{13807, {{26618, 1, 0, 0},{26618, 1, 1, 17}, {0, 0, 0, 0}}},{13807, {{26618, 1, 0, 0},{26618, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{13808, {{26619, 1, 0, 0},{26619, 1, 1, 17}, {0, 0, 0, 0}}},{13808, {{26619, 1, 0, 0},{26619, 1, 1, 17}, {0, 0, 0, 0}}},{57728, {{26619, 1, 0, 0},{26619, 1, 1, 18}, {0, 0, 0, 0}}},{57728, {{26619, 1, 0, 0},{26619, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13809, {{26620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13811, {{26621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13812, {{26622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13813, {{26623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13814, {{26624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13815, {{26625, 1, 0, 0},{26625, 1, 1, 18}, {0, 0, 0, 0}}},{13815, {{26625, 1, 0, 0},{26625, 1, 1, 18}, {0, 0, 0, 0}}},{13816, {{26625, 1, 0, 0},{26625, 1, 1, 19}, {0, 0, 0, 0}}},{13816, {{26625, 1, 0, 0},{26625, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{13817, {{26626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13818, {{26627, 1, 0, 0},{26627, 1, 1, 17}, {0, 0, 0, 0}}},{13818, {{26627, 1, 0, 0},{26627, 1, 1, 17}, {0, 0, 0, 0}}},{13819, {{26627, 1, 0, 0},{26627, 1, 1, 18}, {0, 0, 0, 0}}},{13819, {{26627, 1, 0, 0},{26627, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13820, {{26628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13821, {{26629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{13823, {{26640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{13825, {{26642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13826, {{26643, 1, 0, 0},{26643, 1, 1, 19}, {0, 0, 0, 0}}},{13826, {{26643, 1, 0, 0},{26643, 1, 1, 19}, {0, 0, 0, 0}}},{13827, {{26643, 1, 0, 0},{26643, 1, 1, 18}, {0, 0, 0, 0}}},{13827, {{26643, 1, 0, 0},{26643, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{13828, {{26644, 1, 0, 0},{26644, 1, 1, 17}, {0, 0, 0, 0}}},{13828, {{26644, 1, 0, 0},{26644, 1, 1, 17}, {0, 0, 0, 0}}},{13829, {{26644, 1, 0, 0},{26644, 1, 1, 18}, {0, 0, 0, 0}}},{13829, {{26644, 1, 0, 0},{26644, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13831, {{26645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13832, {{26646, 1, 0, 0},{26646, 1, 1, 17}, {0, 0, 0, 0}}},{13832, {{26646, 1, 0, 0},{26646, 1, 1, 17}, {0, 0, 0, 0}}},{13833, {{26646, 1, 0, 0},{26646, 1, 1, 18}, {0, 0, 0, 0}}},{13833, {{26646, 1, 0, 0},{26646, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13834, {{26647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13835, {{26648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13836, {{26649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13837, {{26650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13838, {{26651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13839, {{26652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13840, {{26653, 1, 0, 0},{26653, 1, 1, 17}, {0, 0, 0, 0}}},{13840, {{26653, 1, 0, 0},{26653, 1, 1, 17}, {0, 0, 0, 0}}},{13841, {{26653, 1, 0, 0},{26653, 1, 1, 18}, {0, 0, 0, 0}}},{13841, {{26653, 1, 0, 0},{26653, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13842, {{26654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13843, {{26655, 1, 0, 0},{26655, 1, 1, 18}, {0, 0, 0, 0}}},{13843, {{26655, 1, 0, 0},{26655, 1, 1, 18}, {0, 0, 0, 0}}},{13844, {{26655, 1, 0, 0},{26655, 1, 1, 19}, {0, 0, 0, 0}}},{13844, {{26655, 1, 0, 0},{26655, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{13845, {{26656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13846, {{26657, 1, 0, 0},{26657, 1, 1, 18}, {0, 0, 0, 0}}},{13846, {{26657, 1, 0, 0},{26657, 1, 1, 18}, {0, 0, 0, 0}}},{13847, {{26657, 1, 0, 0},{26657, 1, 1, 19}, {0, 0, 0, 0}}},{13847, {{26657, 1, 0, 0},{26657, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{13848, {{26658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13849, {{26659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13850, {{26660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13851, {{26661, 1, 0, 0},{26661, 1, 1, 16}, {0, 0, 0, 0}}},{13851, {{26661, 1, 0, 0},{26661, 1, 1, 16}, {0, 0, 0, 0}}},{13852, {{26661, 1, 0, 0},{26661, 1, 1, 17}, {0, 0, 0, 0}}},{13852, {{26661, 1, 0, 0},{26661, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{13853, {{26662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13854, {{26663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13855, {{26664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13856, {{26665, 1, 0, 0},{26665, 1, 1, 17}, {0, 0, 0, 0}}},{13856, {{26665, 1, 0, 0},{26665, 1, 1, 17}, {0, 0, 0, 0}}},{13857, {{26665, 1, 0, 0},{26665, 1, 1, 18}, {0, 0, 0, 0}}},{13857, {{26665, 1, 0, 0},{26665, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13858, {{26666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13859, {{26667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13860, {{26668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13861, {{26669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13862, {{26670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13863, {{26671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13864, {{26672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13865, {{26673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13866, {{26674, 1, 0, 0},{26674, 1, 1, 17}, {0, 0, 0, 0}}},{13866, {{26674, 1, 0, 0},{26674, 1, 1, 17}, {0, 0, 0, 0}}},{13867, {{26674, 1, 0, 0},{26674, 1, 1, 18}, {0, 0, 0, 0}}},{13867, {{26674, 1, 0, 0},{26674, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13868, {{26675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13869, {{26676, 1, 0, 0},{26676, 1, 1, 17}, {0, 0, 0, 0}}},{13869, {{26676, 1, 0, 0},{26676, 1, 1, 17}, {0, 0, 0, 0}}},{13870, {{26676, 1, 0, 0},{26676, 1, 1, 18}, {0, 0, 0, 0}}},{13870, {{26676, 1, 0, 0},{26676, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13871, {{26677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13872, {{26678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13873, {{26679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13874, {{26680, 1, 0, 0},{26680, 1, 1, 17}, {0, 0, 0, 0}}},{13874, {{26680, 1, 0, 0},{26680, 1, 1, 17}, {0, 0, 0, 0}}},{13875, {{26680, 1, 0, 0},{26680, 1, 1, 18}, {0, 0, 0, 0}}},{13875, {{26680, 1, 0, 0},{26680, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13876, {{26681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13877, {{26682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13878, {{26683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13879, {{26684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13880, {{26685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13881, {{26686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13882, {{26687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{13883, {{26688, 1, 0, 0},{26688, 1, 1, 18}, {0, 0, 0, 0}}},{13883, {{26688, 1, 0, 0},{26688, 1, 1, 18}, {0, 0, 0, 0}}},{13884, {{26688, 1, 0, 0},{26688, 1, 1, 17}, {0, 0, 0, 0}}},{13884, {{26688, 1, 0, 0},{26688, 1, 1, 17}, {0, 0, 0, 0}}},{13885, {{26688, 1, 0, 0},{26688, 1, 1, 19}, {0, 0, 0, 0}}},{13885, {{26688, 1, 0, 0},{26688, 1, 1, 19}, {0, 0, 0, 0}}},{57740, {{26688, 1, 0, 0},{26688, 1, 1, 20}, {0, 0, 0, 0}}},{57740, {{26688, 1, 0, 0},{26688, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{13886, {{26689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13887, {{26690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13888, {{26691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13889, {{26692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13890, {{26693, 1, 0, 0},{26693, 1, 1, 17}, {0, 0, 0, 0}}},{13890, {{26693, 1, 0, 0},{26693, 1, 1, 17}, {0, 0, 0, 0}}},{13891, {{26693, 1, 0, 0},{26693, 1, 1, 18}, {0, 0, 0, 0}}},{13891, {{26693, 1, 0, 0},{26693, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13892, {{26694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13893, {{26695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13894, {{26696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13895, {{26697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13896, {{26698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13897, {{26699, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13898, {{26700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13899, {{26701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13900, {{26702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13901, {{26703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13902, {{26704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13903, {{26705, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{13904, {{26706, 1, 0, 0},{26706, 1, 1, 18}, {0, 0, 0, 0}}},{13904, {{26706, 1, 0, 0},{26706, 1, 1, 18}, {0, 0, 0, 0}}},{13905, {{26706, 1, 0, 0},{26706, 1, 1, 19}, {0, 0, 0, 0}}},{13905, {{26706, 1, 0, 0},{26706, 1, 1, 19}, {0, 0, 0, 0}}},{13906, {{26706, 1, 0, 0},{26706, 1, 1, 21}, {0, 0, 0, 0}}},{13906, {{26706, 1, 0, 0},{26706, 1, 1, 21}, {0, 0, 0, 0}}},{13907, {{26706, 1, 0, 0},{26706, 1, 1, 20}, {0, 0, 0, 0}}},{13907, {{26706, 1, 0, 0},{26706, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{13908, {{26707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13909, {{26708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13910, {{26709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13912, {{26710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13913, {{26711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13914, {{26712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13915, {{26713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{13916, {{26715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13917, {{26716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{13918, {{26717, 1, 0, 0},{26717, 1, 1, 20}, {0, 0, 0, 0}}},{13918, {{26717, 1, 0, 0},{26717, 1, 1, 20}, {0, 0, 0, 0}}},{13919, {{26717, 1, 0, 0},{26717, 1, 1, 19}, {0, 0, 0, 0}}},{13919, {{26717, 1, 0, 0},{26717, 1, 1, 19}, {0, 0, 0, 0}}},{13920, {{26717, 1, 0, 0},{26717, 1, 1, 21}, {0, 0, 0, 0}}},{13920, {{26717, 1, 0, 0},{26717, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{13921, {{26719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{13923, {{26723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{13926, {{26727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{13927, {{26731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13928, {{26732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13929, {{26733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13930, {{26734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13931, {{26735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13933, {{26736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13934, {{26737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13935, {{26738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13936, {{26739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13937, {{26740, 1, 0, 0},{26740, 1, 1, 17}, {0, 0, 0, 0}}},{13937, {{26740, 1, 0, 0},{26740, 1, 1, 17}, {0, 0, 0, 0}}},{13938, {{26740, 1, 0, 0},{26740, 1, 1, 18}, {0, 0, 0, 0}}},{13938, {{26740, 1, 0, 0},{26740, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{13939, {{26741, 1, 0, 0},{26741, 1, 1, 17}, {0, 0, 0, 0}}},{13939, {{26741, 1, 0, 0},{26741, 1, 1, 17}, {0, 0, 0, 0}}},{13940, {{26741, 1, 0, 0},{26741, 1, 1, 18}, {0, 0, 0, 0}}},{13940, {{26741, 1, 0, 0},{26741, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13941, {{26742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13942, {{26743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13943, {{26744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13944, {{26745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13945, {{26746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13948, {{26747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13949, {{26748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13950, {{26749, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13951, {{26750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13952, {{26751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13953, {{26752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13954, {{26753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13956, {{26754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13957, {{26755, 1, 0, 0},{26755, 1, 1, 17}, {0, 0, 0, 0}}},{13957, {{26755, 1, 0, 0},{26755, 1, 1, 17}, {0, 0, 0, 0}}},{13958, {{26755, 1, 0, 0},{26755, 1, 1, 18}, {0, 0, 0, 0}}},{13958, {{26755, 1, 0, 0},{26755, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13959, {{26756, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{13961, {{26757, 1, 0, 0},{26757, 1, 1, 18}, {0, 0, 0, 0}}},{13961, {{26757, 1, 0, 0},{26757, 1, 1, 18}, {0, 0, 0, 0}}},{30265, {{26757, 1, 0, 0},{26757, 1, 1, 0},{26757, 1, 1, 19},{64068, 1, 0, 0}, {0, 0, 0, 0}}},{30265, {{26757, 1, 0, 0},{26757, 1, 1, 0},{26757, 1, 1, 19},{64068, 1, 0, 0}, {0, 0, 0, 0}}},{30265, {{26757, 1, 0, 0},{26757, 1, 1, 0},{26757, 1, 1, 19},{64068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13962, {{26758, 1, 0, 0},{26758, 1, 1, 17}, {0, 0, 0, 0}}},{13962, {{26758, 1, 0, 0},{26758, 1, 1, 17}, {0, 0, 0, 0}}},{13963, {{26758, 1, 0, 0},{26758, 1, 1, 18}, {0, 0, 0, 0}}},{13963, {{26758, 1, 0, 0},{26758, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13964, {{26759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13965, {{26760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13966, {{26761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13968, {{26762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13969, {{26763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13970, {{26764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13971, {{26765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13972, {{26766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13973, {{26767, 1, 0, 0},{26767, 1, 1, 17}, {0, 0, 0, 0}}},{13973, {{26767, 1, 0, 0},{26767, 1, 1, 17}, {0, 0, 0, 0}}},{13974, {{26767, 1, 0, 0},{26767, 1, 1, 18}, {0, 0, 0, 0}}},{13974, {{26767, 1, 0, 0},{26767, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13975, {{26768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13976, {{26769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13977, {{26770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13978, {{26771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13979, {{26772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13980, {{26773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13981, {{26774, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13982, {{26775, 1, 0, 0},{26775, 1, 1, 18}, {0, 0, 0, 0}}},{13982, {{26775, 1, 0, 0},{26775, 1, 1, 18}, {0, 0, 0, 0}}},{13983, {{26775, 1, 0, 0},{26775, 1, 1, 19}, {0, 0, 0, 0}}},{13983, {{26775, 1, 0, 0},{26775, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{13984, {{26776, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13985, {{26777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13986, {{26778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13988, {{26779, 1, 0, 0},{26779, 1, 1, 19}, {0, 0, 0, 0}}},{13988, {{26779, 1, 0, 0},{26779, 1, 1, 19}, {0, 0, 0, 0}}},{13989, {{26779, 1, 0, 0},{26779, 1, 1, 18}, {0, 0, 0, 0}}},{13989, {{26779, 1, 0, 0},{26779, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13990, {{26780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13991, {{26781, 1, 0, 0},{26781, 1, 1, 18}, {0, 0, 0, 0}}},{13991, {{26781, 1, 0, 0},{26781, 1, 1, 18}, {0, 0, 0, 0}}},{56933, {{26781, 1, 0, 0},{26781, 1, 1, 19}, {0, 0, 0, 0}}},{56933, {{26781, 1, 0, 0},{26781, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{13992, {{26782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13993, {{26783, 1, 0, 0},{26783, 1, 1, 17}, {0, 0, 0, 0}}},{13993, {{26783, 1, 0, 0},{26783, 1, 1, 17}, {0, 0, 0, 0}}},{13994, {{26783, 1, 0, 0},{26783, 1, 1, 18}, {0, 0, 0, 0}}},{13994, {{26783, 1, 0, 0},{26783, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13995, {{26784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13996, {{26785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{13997, {{26786, 1, 0, 0},{26786, 1, 1, 19}, {0, 0, 0, 0}}},{13997, {{26786, 1, 0, 0},{26786, 1, 1, 19}, {0, 0, 0, 0}}},{13998, {{26786, 1, 0, 0},{26786, 1, 1, 18}, {0, 0, 0, 0}}},{13998, {{26786, 1, 0, 0},{26786, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{13999, {{26787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14000, {{26788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14001, {{26789, 1, 0, 0},{26789, 1, 1, 17}, {0, 0, 0, 0}}},{14001, {{26789, 1, 0, 0},{26789, 1, 1, 17}, {0, 0, 0, 0}}},{14002, {{26789, 1, 0, 0},{26789, 1, 1, 18}, {0, 0, 0, 0}}},{14002, {{26789, 1, 0, 0},{26789, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {2, {{14005, {{26790, 1, 0, 0}, {0, 0, 0, 0}}},{14004, {{26790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14006, {{26791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14007, {{26792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14008, {{26793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14009, {{26794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14010, {{26795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14011, {{26796, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14012, {{26797, 1, 0, 0},{26797, 1, 1, 18}, {0, 0, 0, 0}}},{14012, {{26797, 1, 0, 0},{26797, 1, 1, 18}, {0, 0, 0, 0}}},{14013, {{26797, 1, 0, 0},{26797, 1, 1, 17}, {0, 0, 0, 0}}},{14013, {{26797, 1, 0, 0},{26797, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{14014, {{26798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14015, {{26799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14016, {{26800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14017, {{26801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14018, {{26802, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14019, {{26803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14020, {{26804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14021, {{26805, 1, 0, 0},{26805, 1, 1, 17}, {0, 0, 0, 0}}},{14021, {{26805, 1, 0, 0},{26805, 1, 1, 17}, {0, 0, 0, 0}}},{14022, {{26805, 1, 0, 0},{26805, 1, 1, 18}, {0, 0, 0, 0}}},{14022, {{26805, 1, 0, 0},{26805, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14023, {{26806, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14024, {{26807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14025, {{26808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14026, {{26809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14027, {{26810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14028, {{26811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14029, {{26812, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{14030, {{26817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{14031, {{26819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14032, {{26820, 1, 0, 0},{26820, 1, 1, 17}, {0, 0, 0, 0}}},{14032, {{26820, 1, 0, 0},{26820, 1, 1, 17}, {0, 0, 0, 0}}},{14033, {{26820, 1, 0, 0},{26820, 1, 1, 18}, {0, 0, 0, 0}}},{14033, {{26820, 1, 0, 0},{26820, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14034, {{26821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14035, {{26822, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14036, {{26823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14037, {{26824, 1, 0, 0},{26824, 1, 1, 18}, {0, 0, 0, 0}}},{14037, {{26824, 1, 0, 0},{26824, 1, 1, 18}, {0, 0, 0, 0}}},{14038, {{26824, 1, 0, 0},{26824, 1, 1, 19}, {0, 0, 0, 0}}},{14038, {{26824, 1, 0, 0},{26824, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{14039, {{26825, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14040, {{26826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14041, {{26827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14042, {{26828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14044, {{26829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14045, {{26830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14046, {{26831, 1, 0, 0},{26831, 1, 1, 17}, {0, 0, 0, 0}}},{14046, {{26831, 1, 0, 0},{26831, 1, 1, 17}, {0, 0, 0, 0}}},{14047, {{26831, 1, 0, 0},{26831, 1, 1, 18}, {0, 0, 0, 0}}},{14047, {{26831, 1, 0, 0},{26831, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14048, {{26832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14050, {{26833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14051, {{26834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14052, {{26835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14053, {{26836, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14054, {{26837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14055, {{26838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14056, {{26839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14057, {{26840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14058, {{26841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14059, {{26842, 1, 0, 0},{26842, 1, 1, 19}, {0, 0, 0, 0}}},{14059, {{26842, 1, 0, 0},{26842, 1, 1, 19}, {0, 0, 0, 0}}},{14060, {{26842, 1, 0, 0},{26842, 1, 1, 18}, {0, 0, 0, 0}}},{14060, {{26842, 1, 0, 0},{26842, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14061, {{26843, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14062, {{26844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14063, {{26845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14064, {{26846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14065, {{26847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14066, {{26848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14067, {{26849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14068, {{26850, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14069, {{26851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14070, {{26852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14071, {{26853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14072, {{26854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14074, {{26855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14075, {{26856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14076, {{26857, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14077, {{26858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14078, {{26859, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14079, {{26860, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14080, {{26861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14081, {{26862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14082, {{26863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14083, {{26864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14084, {{26865, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14085, {{26866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14086, {{26867, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14087, {{26868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14088, {{26869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14089, {{26870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14090, {{26871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14091, {{26872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14092, {{26873, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14093, {{26874, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14094, {{26875, 1, 0, 0},{26875, 1, 1, 17}, {0, 0, 0, 0}}},{14094, {{26875, 1, 0, 0},{26875, 1, 1, 17}, {0, 0, 0, 0}}},{14095, {{26875, 1, 0, 0},{26875, 1, 1, 18}, {0, 0, 0, 0}}},{14095, {{26875, 1, 0, 0},{26875, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14096, {{26876, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14097, {{26877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14098, {{26878, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14099, {{26879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14100, {{26880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14101, {{26881, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14102, {{26882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14104, {{26883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14106, {{26884, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14107, {{26885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14108, {{26886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14110, {{26887, 1, 0, 0},{26887, 1, 1, 18}, {0, 0, 0, 0}}},{14110, {{26887, 1, 0, 0},{26887, 1, 1, 18}, {0, 0, 0, 0}}},{14111, {{26887, 1, 0, 0},{26887, 1, 1, 17}, {0, 0, 0, 0}}},{14111, {{26887, 1, 0, 0},{26887, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{14112, {{26888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14113, {{26889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14114, {{26890, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14115, {{26891, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14116, {{26892, 1, 0, 0},{26892, 1, 1, 17}, {0, 0, 0, 0}}},{14116, {{26892, 1, 0, 0},{26892, 1, 1, 17}, {0, 0, 0, 0}}},{14117, {{26892, 1, 0, 0},{26892, 1, 1, 18}, {0, 0, 0, 0}}},{14117, {{26892, 1, 0, 0},{26892, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{14118, {{26893, 1, 0, 0},{26893, 1, 1, 19}, {0, 0, 0, 0}}},{14118, {{26893, 1, 0, 0},{26893, 1, 1, 19}, {0, 0, 0, 0}}},{57753, {{26893, 1, 0, 0},{26893, 1, 1, 20}, {0, 0, 0, 0}}},{57753, {{26893, 1, 0, 0},{26893, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{14120, {{26894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14121, {{26895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14122, {{26896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14123, {{26897, 1, 0, 0},{26897, 1, 1, 17}, {0, 0, 0, 0}}},{14123, {{26897, 1, 0, 0},{26897, 1, 1, 17}, {0, 0, 0, 0}}},{57741, {{26897, 1, 0, 0},{26897, 1, 1, 18}, {0, 0, 0, 0}}},{57741, {{26897, 1, 0, 0},{26897, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14124, {{26898, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14125, {{26899, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14126, {{26900, 1, 0, 0},{26900, 1, 1, 16}, {0, 0, 0, 0}}},{14126, {{26900, 1, 0, 0},{26900, 1, 1, 16}, {0, 0, 0, 0}}},{14127, {{26900, 1, 0, 0},{26900, 1, 1, 17}, {0, 0, 0, 0}}},{14127, {{26900, 1, 0, 0},{26900, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{14128, {{26901, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14129, {{26902, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14131, {{26903, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14132, {{26904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14133, {{26905, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14134, {{26906, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{14135, {{26907, 1, 0, 0},{26907, 1, 1, 17}, {0, 0, 0, 0}}},{14135, {{26907, 1, 0, 0},{26907, 1, 1, 17}, {0, 0, 0, 0}}},{14136, {{26907, 1, 0, 0},{26907, 1, 1, 18}, {0, 0, 0, 0}}},{14136, {{26907, 1, 0, 0},{26907, 1, 1, 18}, {0, 0, 0, 0}}},{14137, {{26907, 1, 0, 0},{26907, 1, 1, 19}, {0, 0, 0, 0}}},{14137, {{26907, 1, 0, 0},{26907, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{14138, {{26908, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{14139, {{26913, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14140, {{26914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14141, {{26915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{14142, {{26917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14143, {{26918, 1, 0, 0},{26918, 1, 1, 17}, {0, 0, 0, 0}}},{14143, {{26918, 1, 0, 0},{26918, 1, 1, 17}, {0, 0, 0, 0}}},{14144, {{26918, 1, 0, 0},{26918, 1, 1, 18}, {0, 0, 0, 0}}},{14144, {{26918, 1, 0, 0},{26918, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14145, {{26919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14146, {{26920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{14147, {{26922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{14148, {{26927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14149, {{26928, 1, 0, 0},{26928, 1, 1, 18}, {0, 0, 0, 0}}},{14149, {{26928, 1, 0, 0},{26928, 1, 1, 18}, {0, 0, 0, 0}}},{14150, {{26928, 1, 0, 0},{26928, 1, 1, 19}, {0, 0, 0, 0}}},{14150, {{26928, 1, 0, 0},{26928, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{14151, {{26929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14152, {{26930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14154, {{26931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14155, {{26932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14156, {{26933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14157, {{26934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14158, {{26935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14159, {{26936, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14160, {{26937, 1, 0, 0},{26937, 1, 1, 17}, {0, 0, 0, 0}}},{14160, {{26937, 1, 0, 0},{26937, 1, 1, 17}, {0, 0, 0, 0}}},{14161, {{26937, 1, 0, 0},{26937, 1, 1, 18}, {0, 0, 0, 0}}},{14161, {{26937, 1, 0, 0},{26937, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14162, {{26938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14163, {{26939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14164, {{26940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{14165, {{26941, 1, 0, 0},{26941, 1, 1, 18}, {0, 0, 0, 0}}},{14165, {{26941, 1, 0, 0},{26941, 1, 1, 18}, {0, 0, 0, 0}}},{14166, {{26941, 1, 0, 0},{26941, 1, 1, 19}, {0, 0, 0, 0}}},{14166, {{26941, 1, 0, 0},{26941, 1, 1, 19}, {0, 0, 0, 0}}},{14167, {{26941, 1, 0, 0},{26941, 1, 1, 20}, {0, 0, 0, 0}}},{14167, {{26941, 1, 0, 0},{26941, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{14168, {{26942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14169, {{26943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14170, {{26944, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14171, {{26945, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14172, {{26946, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14173, {{26947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14174, {{26948, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14175, {{26949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14176, {{26950, 1, 0, 0},{26950, 1, 1, 18}, {0, 0, 0, 0}}},{14176, {{26950, 1, 0, 0},{26950, 1, 1, 18}, {0, 0, 0, 0}}},{14177, {{26950, 1, 0, 0},{26950, 1, 1, 17}, {0, 0, 0, 0}}},{14177, {{26950, 1, 0, 0},{26950, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{14178, {{26951, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14179, {{26952, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14180, {{26953, 1, 0, 0},{26953, 1, 1, 17}, {0, 0, 0, 0}}},{14180, {{26953, 1, 0, 0},{26953, 1, 1, 17}, {0, 0, 0, 0}}},{14181, {{26953, 1, 0, 0},{26953, 1, 1, 18}, {0, 0, 0, 0}}},{14181, {{26953, 1, 0, 0},{26953, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14182, {{26954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14183, {{26955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14184, {{26956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14185, {{26957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14186, {{26958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14187, {{26959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14188, {{26960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14189, {{26961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14190, {{26962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14191, {{26963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{14192, {{26964, 1, 0, 0},{26964, 1, 1, 18}, {0, 0, 0, 0}}},{14192, {{26964, 1, 0, 0},{26964, 1, 1, 18}, {0, 0, 0, 0}}},{14193, {{26964, 1, 0, 0},{26964, 1, 1, 17}, {0, 0, 0, 0}}},{14193, {{26964, 1, 0, 0},{26964, 1, 1, 17}, {0, 0, 0, 0}}},{14194, {{26964, 1, 0, 0},{26964, 1, 1, 19}, {0, 0, 0, 0}}},{14194, {{26964, 1, 0, 0},{26964, 1, 1, 19}, {0, 0, 0, 0}}},{57766, {{26964, 1, 0, 0},{26964, 1, 1, 20}, {0, 0, 0, 0}}},{57766, {{26964, 1, 0, 0},{26964, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{14195, {{26965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14196, {{26966, 1, 0, 0},{26966, 1, 1, 16}, {0, 0, 0, 0}}},{14196, {{26966, 1, 0, 0},{26966, 1, 1, 16}, {0, 0, 0, 0}}},{14197, {{26966, 1, 0, 0},{26966, 1, 1, 17}, {0, 0, 0, 0}}},{14197, {{26966, 1, 0, 0},{26966, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{14198, {{26967, 1, 0, 0},{26967, 1, 1, 17}, {0, 0, 0, 0}}},{14198, {{26967, 1, 0, 0},{26967, 1, 1, 17}, {0, 0, 0, 0}}},{14199, {{26967, 1, 0, 0},{26967, 1, 1, 18}, {0, 0, 0, 0}}},{14199, {{26967, 1, 0, 0},{26967, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14200, {{26968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14201, {{26969, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14202, {{26970, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14203, {{26971, 1, 0, 0},{26971, 1, 1, 17}, {0, 0, 0, 0}}},{14203, {{26971, 1, 0, 0},{26971, 1, 1, 17}, {0, 0, 0, 0}}},{14204, {{26971, 1, 0, 0},{26971, 1, 1, 18}, {0, 0, 0, 0}}},{14204, {{26971, 1, 0, 0},{26971, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14205, {{26972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14206, {{26973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14207, {{26974, 1, 0, 0},{26974, 1, 1, 19}, {0, 0, 0, 0}}},{14207, {{26974, 1, 0, 0},{26974, 1, 1, 19}, {0, 0, 0, 0}}},{14208, {{26974, 1, 0, 0},{26974, 1, 1, 18}, {0, 0, 0, 0}}},{14208, {{26974, 1, 0, 0},{26974, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14209, {{26975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14210, {{26976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14211, {{26977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{14213, {{26978, 1, 0, 0},{26978, 1, 1, 18}, {0, 0, 0, 0}}},{14213, {{26978, 1, 0, 0},{26978, 1, 1, 18}, {0, 0, 0, 0}}},{14214, {{26978, 1, 0, 0},{26978, 1, 1, 19}, {0, 0, 0, 0}}},{14214, {{26978, 1, 0, 0},{26978, 1, 1, 19}, {0, 0, 0, 0}}},{14215, {{26978, 1, 0, 0},{26978, 1, 1, 21}, {0, 0, 0, 0}}},{14215, {{26978, 1, 0, 0},{26978, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{14216, {{26979, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14217, {{26980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14218, {{26981, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14219, {{26982, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14220, {{26983, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14221, {{26984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14222, {{26985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14223, {{26986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14224, {{26987, 1, 0, 0},{26987, 1, 1, 18}, {0, 0, 0, 0}}},{14224, {{26987, 1, 0, 0},{26987, 1, 1, 18}, {0, 0, 0, 0}}},{14225, {{26987, 1, 0, 0},{26987, 1, 1, 19}, {0, 0, 0, 0}}},{14225, {{26987, 1, 0, 0},{26987, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{14226, {{26988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14227, {{26989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14228, {{26990, 1, 0, 0},{26990, 1, 1, 18}, {0, 0, 0, 0}}},{14228, {{26990, 1, 0, 0},{26990, 1, 1, 18}, {0, 0, 0, 0}}},{14229, {{26990, 1, 0, 0},{26990, 1, 1, 17}, {0, 0, 0, 0}}},{14229, {{26990, 1, 0, 0},{26990, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{14230, {{26991, 1, 0, 0},{26991, 1, 1, 20}, {0, 0, 0, 0}}},{14230, {{26991, 1, 0, 0},{26991, 1, 1, 20}, {0, 0, 0, 0}}},{14231, {{26991, 1, 0, 0},{26991, 1, 1, 21}, {0, 0, 0, 0}}},{14231, {{26991, 1, 0, 0},{26991, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{14232, {{26992, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14233, {{26993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14234, {{26994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14235, {{26995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14236, {{26996, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14237, {{26997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14238, {{26998, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14239, {{26999, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14240, {{27000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14241, {{27001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14242, {{27002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14243, {{27003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14244, {{27004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14245, {{27005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14246, {{27006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14247, {{27007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14248, {{27008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14249, {{27009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14250, {{27010, 1, 0, 0},{27010, 1, 1, 20}, {0, 0, 0, 0}}},{14250, {{27010, 1, 0, 0},{27010, 1, 1, 20}, {0, 0, 0, 0}}},{14251, {{27010, 1, 0, 0},{27010, 1, 1, 21}, {0, 0, 0, 0}}},{14251, {{27010, 1, 0, 0},{27010, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{14252, {{27013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14253, {{27014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{14255, {{27018, 1, 0, 0},{27018, 1, 1, 18}, {0, 0, 0, 0}}},{14255, {{27018, 1, 0, 0},{27018, 1, 1, 18}, {0, 0, 0, 0}}},{14256, {{27018, 1, 0, 0},{27018, 1, 1, 19}, {0, 0, 0, 0}}},{14256, {{27018, 1, 0, 0},{27018, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{14257, {{27021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14258, {{27022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14259, {{27023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14261, {{27024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14262, {{27025, 1, 0, 0},{27025, 1, 1, 17}, {0, 0, 0, 0}}},{14262, {{27025, 1, 0, 0},{27025, 1, 1, 17}, {0, 0, 0, 0}}},{14263, {{27025, 1, 0, 0},{27025, 1, 1, 18}, {0, 0, 0, 0}}},{14263, {{27025, 1, 0, 0},{27025, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{14264, {{27026, 1, 0, 0},{27026, 1, 1, 17}, {0, 0, 0, 0}}},{14264, {{27026, 1, 0, 0},{27026, 1, 1, 17}, {0, 0, 0, 0}}},{14265, {{27026, 1, 0, 0},{27026, 1, 1, 18}, {0, 0, 0, 0}}},{14265, {{27026, 1, 0, 0},{27026, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14266, {{27027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14267, {{27028, 1, 0, 0},{27028, 1, 1, 18}, {0, 0, 0, 0}}},{14267, {{27028, 1, 0, 0},{27028, 1, 1, 18}, {0, 0, 0, 0}}},{14268, {{27028, 1, 0, 0},{27028, 1, 1, 19}, {0, 0, 0, 0}}},{14268, {{27028, 1, 0, 0},{27028, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{14270, {{27029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14271, {{27030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14272, {{27031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14273, {{27032, 1, 0, 0},{27032, 1, 1, 17}, {0, 0, 0, 0}}},{14273, {{27032, 1, 0, 0},{27032, 1, 1, 17}, {0, 0, 0, 0}}},{14274, {{27032, 1, 0, 0},{27032, 1, 1, 18}, {0, 0, 0, 0}}},{14274, {{27032, 1, 0, 0},{27032, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14275, {{27033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{14276, {{27035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14277, {{27036, 1, 0, 0},{27036, 1, 1, 17}, {0, 0, 0, 0}}},{14277, {{27036, 1, 0, 0},{27036, 1, 1, 17}, {0, 0, 0, 0}}},{14278, {{27036, 1, 0, 0},{27036, 1, 1, 18}, {0, 0, 0, 0}}},{14278, {{27036, 1, 0, 0},{27036, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14452, {{27037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14279, {{27038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14280, {{27039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14281, {{27040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14282, {{27041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14283, {{27042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14284, {{27043, 1, 0, 0},{27043, 1, 1, 16}, {0, 0, 0, 0}}},{14284, {{27043, 1, 0, 0},{27043, 1, 1, 16}, {0, 0, 0, 0}}},{14285, {{27043, 1, 0, 0},{27043, 1, 1, 17}, {0, 0, 0, 0}}},{14285, {{27043, 1, 0, 0},{27043, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{14286, {{27044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14287, {{27045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14288, {{27046, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14290, {{27047, 1, 0, 0},{27047, 1, 1, 18}, {0, 0, 0, 0}}},{14290, {{27047, 1, 0, 0},{27047, 1, 1, 18}, {0, 0, 0, 0}}},{14291, {{27047, 1, 0, 0},{27047, 1, 1, 19}, {0, 0, 0, 0}}},{14291, {{27047, 1, 0, 0},{27047, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{14292, {{27048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14293, {{27049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14294, {{27050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14295, {{27051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14296, {{27052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14297, {{27053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14298, {{27054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14299, {{27055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14300, {{27056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14301, {{27057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14302, {{27058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14303, {{27059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14304, {{27060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14305, {{27061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14306, {{27062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14308, {{27063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14309, {{27064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14310, {{27065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{14311, {{27066, 1, 0, 0},{27066, 1, 1, 17}, {0, 0, 0, 0}}},{14311, {{27066, 1, 0, 0},{27066, 1, 1, 17}, {0, 0, 0, 0}}},{14312, {{27066, 1, 0, 0},{27066, 1, 1, 18}, {0, 0, 0, 0}}},{14312, {{27066, 1, 0, 0},{27066, 1, 1, 18}, {0, 0, 0, 0}}},{14313, {{27066, 1, 0, 0},{27066, 1, 1, 20}, {0, 0, 0, 0}}},{14313, {{27066, 1, 0, 0},{27066, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {6, {{14316, {{27067, 1, 0, 0},{27067, 1, 1, 18}, {0, 0, 0, 0}}},{14316, {{27067, 1, 0, 0},{27067, 1, 1, 18}, {0, 0, 0, 0}}},{14317, {{27067, 1, 0, 0},{27067, 1, 1, 19}, {0, 0, 0, 0}}},{14317, {{27067, 1, 0, 0},{27067, 1, 1, 19}, {0, 0, 0, 0}}},{14318, {{27067, 1, 0, 0},{27067, 1, 1, 20}, {0, 0, 0, 0}}},{14318, {{27067, 1, 0, 0},{27067, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{14319, {{27068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14320, {{27069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14322, {{27070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14323, {{27071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14324, {{27072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14325, {{27073, 1, 0, 0},{27073, 1, 1, 18}, {0, 0, 0, 0}}},{14325, {{27073, 1, 0, 0},{27073, 1, 1, 18}, {0, 0, 0, 0}}},{14326, {{27073, 1, 0, 0},{27073, 1, 1, 19}, {0, 0, 0, 0}}},{14326, {{27073, 1, 0, 0},{27073, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{14327, {{27074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14328, {{27075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14329, {{27076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14331, {{27077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14332, {{27078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{14333, {{27079, 1, 0, 0},{27079, 1, 1, 17}, {0, 0, 0, 0}}},{14333, {{27079, 1, 0, 0},{27079, 1, 1, 17}, {0, 0, 0, 0}}},{14334, {{27079, 1, 0, 0},{27079, 1, 1, 18}, {0, 0, 0, 0}}},{14334, {{27079, 1, 0, 0},{27079, 1, 1, 18}, {0, 0, 0, 0}}},{14357, {{27079, 1, 0, 0},{27079, 1, 1, 19}, {0, 0, 0, 0}}},{14357, {{27079, 1, 0, 0},{27079, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{14335, {{27080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14336, {{27081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14337, {{27082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14338, {{27083, 1, 0, 0},{27083, 1, 1, 19}, {0, 0, 0, 0}}},{14338, {{27083, 1, 0, 0},{27083, 1, 1, 19}, {0, 0, 0, 0}}},{14339, {{27083, 1, 0, 0},{27083, 1, 1, 18}, {0, 0, 0, 0}}},{14339, {{27083, 1, 0, 0},{27083, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{14340, {{27084, 1, 0, 0},{27084, 1, 1, 18}, {0, 0, 0, 0}}},{14340, {{27084, 1, 0, 0},{27084, 1, 1, 18}, {0, 0, 0, 0}}},{14341, {{27084, 1, 0, 0},{27084, 1, 1, 19}, {0, 0, 0, 0}}},{14341, {{27084, 1, 0, 0},{27084, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{14342, {{27085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14343, {{27086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{14344, {{27087, 1, 0, 0},{27087, 1, 1, 19}, {0, 0, 0, 0}}},{14344, {{27087, 1, 0, 0},{27087, 1, 1, 19}, {0, 0, 0, 0}}},{14345, {{27087, 1, 0, 0},{27087, 1, 1, 18}, {0, 0, 0, 0}}},{14345, {{27087, 1, 0, 0},{27087, 1, 1, 18}, {0, 0, 0, 0}}},{14346, {{27087, 1, 0, 0},{27087, 1, 1, 17}, {0, 0, 0, 0}}},{14346, {{27087, 1, 0, 0},{27087, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{14347, {{27088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14348, {{27089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14349, {{27090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14350, {{27091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14351, {{27092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14352, {{27093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14353, {{27094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14354, {{27095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14355, {{27096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14356, {{27097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{14359, {{27101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14360, {{27102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{14361, {{27106, 1, 0, 0},{27106, 1, 1, 17}, {0, 0, 0, 0}}},{14361, {{27106, 1, 0, 0},{27106, 1, 1, 17}, {0, 0, 0, 0}}},{14362, {{27106, 1, 0, 0},{27106, 1, 1, 19}, {0, 0, 0, 0}}},{14362, {{27106, 1, 0, 0},{27106, 1, 1, 19}, {0, 0, 0, 0}}},{59729, {{27106, 1, 0, 0},{27106, 1, 1, 18}, {0, 0, 0, 0}}},{59729, {{27106, 1, 0, 0},{27106, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14363, {{27107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14364, {{27108, 1, 0, 0},{27108, 1, 1, 16}, {0, 0, 0, 0}}},{14364, {{27108, 1, 0, 0},{27108, 1, 1, 16}, {0, 0, 0, 0}}},{14365, {{27108, 1, 0, 0},{27108, 1, 1, 17}, {0, 0, 0, 0}}},{14365, {{27108, 1, 0, 0},{27108, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{14366, {{27109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14367, {{27110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14368, {{27111, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14369, {{27112, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{14370, {{27113, 1, 0, 0},{27113, 1, 1, 18}, {0, 0, 0, 0}}},{14370, {{27113, 1, 0, 0},{27113, 1, 1, 18}, {0, 0, 0, 0}}},{14371, {{27113, 1, 0, 0},{27113, 1, 1, 19}, {0, 0, 0, 0}}},{14371, {{27113, 1, 0, 0},{27113, 1, 1, 19}, {0, 0, 0, 0}}},{14372, {{27113, 1, 0, 0},{27113, 1, 1, 17}, {0, 0, 0, 0}}},{14372, {{27113, 1, 0, 0},{27113, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {6, {{14373, {{27114, 1, 0, 0},{27114, 1, 1, 20}, {0, 0, 0, 0}}},{14373, {{27114, 1, 0, 0},{27114, 1, 1, 20}, {0, 0, 0, 0}}},{14374, {{27114, 1, 0, 0},{27114, 1, 1, 19}, {0, 0, 0, 0}}},{14374, {{27114, 1, 0, 0},{27114, 1, 1, 19}, {0, 0, 0, 0}}},{14375, {{27114, 1, 0, 0},{27114, 1, 1, 18}, {0, 0, 0, 0}}},{14375, {{27114, 1, 0, 0},{27114, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{14376, {{27115, 1, 0, 0},{27115, 1, 1, 17}, {0, 0, 0, 0}}},{14376, {{27115, 1, 0, 0},{27115, 1, 1, 17}, {0, 0, 0, 0}}},{14377, {{27115, 1, 0, 0},{27115, 1, 1, 18}, {0, 0, 0, 0}}},{14377, {{27115, 1, 0, 0},{27115, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14379, {{27116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14380, {{27117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14381, {{27118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14382, {{27119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14383, {{27120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14384, {{27121, 1, 0, 0},{27121, 1, 1, 17}, {0, 0, 0, 0}}},{14384, {{27121, 1, 0, 0},{27121, 1, 1, 17}, {0, 0, 0, 0}}},{14385, {{27121, 1, 0, 0},{27121, 1, 1, 18}, {0, 0, 0, 0}}},{14385, {{27121, 1, 0, 0},{27121, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14386, {{27122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14387, {{27123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14388, {{27124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14389, {{27125, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14390, {{27126, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14391, {{27127, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14392, {{27128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14393, {{27129, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14394, {{27130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14395, {{27131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14396, {{27132, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14397, {{27133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14398, {{27134, 1, 0, 0},{27134, 1, 1, 18}, {0, 0, 0, 0}}},{14398, {{27134, 1, 0, 0},{27134, 1, 1, 18}, {0, 0, 0, 0}}},{14399, {{27134, 1, 0, 0},{27134, 1, 1, 19}, {0, 0, 0, 0}}},{14399, {{27134, 1, 0, 0},{27134, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {6, {{14400, {{27135, 1, 0, 0},{27135, 1, 1, 17}, {0, 0, 0, 0}}},{14400, {{27135, 1, 0, 0},{27135, 1, 1, 17}, {0, 0, 0, 0}}},{14401, {{27135, 1, 0, 0},{27135, 1, 1, 18}, {0, 0, 0, 0}}},{14401, {{27135, 1, 0, 0},{27135, 1, 1, 18}, {0, 0, 0, 0}}},{14402, {{27135, 1, 0, 0},{27135, 1, 1, 19}, {0, 0, 0, 0}}},{14402, {{27135, 1, 0, 0},{27135, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{14403, {{27136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14404, {{27137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14406, {{27138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14407, {{27139, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14408, {{27140, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14409, {{27141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14411, {{27142, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14412, {{27143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14413, {{27144, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14414, {{27145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14415, {{27146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14416, {{27147, 1, 0, 0},{27147, 1, 1, 18}, {0, 0, 0, 0}}},{14416, {{27147, 1, 0, 0},{27147, 1, 1, 18}, {0, 0, 0, 0}}},{14417, {{27147, 1, 0, 0},{27147, 1, 1, 19}, {0, 0, 0, 0}}},{14417, {{27147, 1, 0, 0},{27147, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{14418, {{27148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14419, {{27149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14420, {{27150, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14421, {{27151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14422, {{27152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14423, {{27153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14424, {{27154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14425, {{27155, 1, 0, 0},{27155, 1, 1, 17}, {0, 0, 0, 0}}},{14425, {{27155, 1, 0, 0},{27155, 1, 1, 17}, {0, 0, 0, 0}}},{57816, {{27155, 1, 0, 0},{27155, 1, 1, 18}, {0, 0, 0, 0}}},{57816, {{27155, 1, 0, 0},{27155, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{14426, {{27156, 1, 0, 0},{27156, 1, 1, 17}, {0, 0, 0, 0}}},{14426, {{27156, 1, 0, 0},{27156, 1, 1, 17}, {0, 0, 0, 0}}},{14427, {{27156, 1, 0, 0},{27156, 1, 1, 18}, {0, 0, 0, 0}}},{14427, {{27156, 1, 0, 0},{27156, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14428, {{27157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14429, {{27158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14430, {{27159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14431, {{27160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14432, {{27161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14433, {{27162, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14434, {{27163, 1, 0, 0},{27163, 1, 1, 17}, {0, 0, 0, 0}}},{14434, {{27163, 1, 0, 0},{27163, 1, 1, 17}, {0, 0, 0, 0}}},{14435, {{27163, 1, 0, 0},{27163, 1, 1, 18}, {0, 0, 0, 0}}},{14435, {{27163, 1, 0, 0},{27163, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14436, {{27164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14437, {{27165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14438, {{27166, 1, 0, 0},{27166, 1, 1, 17}, {0, 0, 0, 0}}},{14438, {{27166, 1, 0, 0},{27166, 1, 1, 17}, {0, 0, 0, 0}}},{14439, {{27166, 1, 0, 0},{27166, 1, 1, 18}, {0, 0, 0, 0}}},{14439, {{27166, 1, 0, 0},{27166, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{14440, {{27167, 1, 0, 0},{27167, 1, 1, 17}, {0, 0, 0, 0}}},{14440, {{27167, 1, 0, 0},{27167, 1, 1, 17}, {0, 0, 0, 0}}},{14441, {{27167, 1, 0, 0},{27167, 1, 1, 18}, {0, 0, 0, 0}}},{14441, {{27167, 1, 0, 0},{27167, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14442, {{27168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14443, {{27169, 1, 0, 0},{27169, 1, 1, 17}, {0, 0, 0, 0}}},{14443, {{27169, 1, 0, 0},{27169, 1, 1, 17}, {0, 0, 0, 0}}},{14444, {{27169, 1, 0, 0},{27169, 1, 1, 18}, {0, 0, 0, 0}}},{14444, {{27169, 1, 0, 0},{27169, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14445, {{27170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14446, {{27171, 1, 0, 0},{27171, 1, 1, 17}, {0, 0, 0, 0}}},{14446, {{27171, 1, 0, 0},{27171, 1, 1, 17}, {0, 0, 0, 0}}},{14447, {{27171, 1, 0, 0},{27171, 1, 1, 18}, {0, 0, 0, 0}}},{14447, {{27171, 1, 0, 0},{27171, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14448, {{27172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14449, {{27173, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14450, {{27174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14451, {{27175, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14453, {{27176, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14454, {{27177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14455, {{27178, 1, 0, 0},{27178, 1, 1, 17}, {0, 0, 0, 0}}},{14455, {{27178, 1, 0, 0},{27178, 1, 1, 17}, {0, 0, 0, 0}}},{14456, {{27178, 1, 0, 0},{27178, 1, 1, 18}, {0, 0, 0, 0}}},{14456, {{27178, 1, 0, 0},{27178, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14458, {{27179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14459, {{27180, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{14461, {{27182, 1, 0, 0},{27182, 1, 1, 17}, {0, 0, 0, 0}}},{14461, {{27182, 1, 0, 0},{27182, 1, 1, 17}, {0, 0, 0, 0}}},{59737, {{27182, 1, 0, 0},{27182, 1, 1, 18}, {0, 0, 0, 0}}},{59737, {{27182, 1, 0, 0},{27182, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{57809, {{27183, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14462, {{27184, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{14463, {{27186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14464, {{27187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14466, {{27188, 1, 0, 0},{27188, 1, 1, 17}, {0, 0, 0, 0}}},{14466, {{27188, 1, 0, 0},{27188, 1, 1, 17}, {0, 0, 0, 0}}},{14467, {{27188, 1, 0, 0},{27188, 1, 1, 18}, {0, 0, 0, 0}}},{14467, {{27188, 1, 0, 0},{27188, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14468, {{27189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14470, {{27190, 1, 0, 0},{27190, 1, 1, 18}, {0, 0, 0, 0}}},{14470, {{27190, 1, 0, 0},{27190, 1, 1, 18}, {0, 0, 0, 0}}},{14471, {{27190, 1, 0, 0},{27190, 1, 1, 17}, {0, 0, 0, 0}}},{14471, {{27190, 1, 0, 0},{27190, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{14472, {{27191, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14473, {{27192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14474, {{27193, 1, 0, 0},{27193, 1, 1, 17}, {0, 0, 0, 0}}},{14474, {{27193, 1, 0, 0},{27193, 1, 1, 17}, {0, 0, 0, 0}}},{14475, {{27193, 1, 0, 0},{27193, 1, 1, 18}, {0, 0, 0, 0}}},{14475, {{27193, 1, 0, 0},{27193, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{14476, {{27194, 1, 0, 0},{27194, 1, 1, 17}, {0, 0, 0, 0}}},{14476, {{27194, 1, 0, 0},{27194, 1, 1, 17}, {0, 0, 0, 0}}},{14477, {{27194, 1, 0, 0},{27194, 1, 1, 18}, {0, 0, 0, 0}}},{14477, {{27194, 1, 0, 0},{27194, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14478, {{27195, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14479, {{27196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14480, {{27197, 1, 0, 0},{27197, 1, 1, 19}, {0, 0, 0, 0}}},{14480, {{27197, 1, 0, 0},{27197, 1, 1, 19}, {0, 0, 0, 0}}},{14481, {{27197, 1, 0, 0},{27197, 1, 1, 18}, {0, 0, 0, 0}}},{14481, {{27197, 1, 0, 0},{27197, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14482, {{27198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14483, {{27199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14484, {{27200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14485, {{27201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14486, {{27202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14487, {{27203, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14488, {{27204, 1, 0, 0},{27204, 1, 1, 18}, {0, 0, 0, 0}}},{14488, {{27204, 1, 0, 0},{27204, 1, 1, 18}, {0, 0, 0, 0}}},{14489, {{27204, 1, 0, 0},{27204, 1, 1, 19}, {0, 0, 0, 0}}},{14489, {{27204, 1, 0, 0},{27204, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{14490, {{27205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14491, {{27206, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14492, {{27207, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14493, {{27208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14494, {{27209, 1, 0, 0},{27209, 1, 1, 17}, {0, 0, 0, 0}}},{14494, {{27209, 1, 0, 0},{27209, 1, 1, 17}, {0, 0, 0, 0}}},{14495, {{27209, 1, 0, 0},{27209, 1, 1, 18}, {0, 0, 0, 0}}},{14495, {{27209, 1, 0, 0},{27209, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14496, {{27210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14497, {{27211, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14498, {{27212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14499, {{27213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14500, {{27214, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14501, {{27215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14502, {{27216, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14503, {{27217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14504, {{27218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14505, {{27219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14506, {{27220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14507, {{27221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14508, {{27222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14509, {{27223, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14510, {{27224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14511, {{27225, 1, 0, 0},{27225, 1, 1, 17}, {0, 0, 0, 0}}},{14511, {{27225, 1, 0, 0},{27225, 1, 1, 17}, {0, 0, 0, 0}}},{14512, {{27225, 1, 0, 0},{27225, 1, 1, 18}, {0, 0, 0, 0}}},{14512, {{27225, 1, 0, 0},{27225, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14513, {{27226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14514, {{27227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14515, {{27228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14516, {{27229, 1, 0, 0},{27229, 1, 1, 17}, {0, 0, 0, 0}}},{14516, {{27229, 1, 0, 0},{27229, 1, 1, 17}, {0, 0, 0, 0}}},{14517, {{27229, 1, 0, 0},{27229, 1, 1, 16}, {0, 0, 0, 0}}},{14517, {{27229, 1, 0, 0},{27229, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{14518, {{27230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14520, {{27231, 1, 0, 0},{27231, 1, 1, 18}, {0, 0, 0, 0}}},{14520, {{27231, 1, 0, 0},{27231, 1, 1, 18}, {0, 0, 0, 0}}},{14521, {{27231, 1, 0, 0},{27231, 1, 1, 19}, {0, 0, 0, 0}}},{14521, {{27231, 1, 0, 0},{27231, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{14522, {{27232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14523, {{27233, 1, 0, 0},{27233, 1, 1, 17}, {0, 0, 0, 0}}},{14523, {{27233, 1, 0, 0},{27233, 1, 1, 17}, {0, 0, 0, 0}}},{14524, {{27233, 1, 0, 0},{27233, 1, 1, 18}, {0, 0, 0, 0}}},{14524, {{27233, 1, 0, 0},{27233, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14525, {{27234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14526, {{27235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14527, {{27236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14528, {{27237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14529, {{27238, 1, 0, 0},{27238, 1, 1, 17}, {0, 0, 0, 0}}},{14529, {{27238, 1, 0, 0},{27238, 1, 1, 17}, {0, 0, 0, 0}}},{14530, {{27238, 1, 0, 0},{27238, 1, 1, 18}, {0, 0, 0, 0}}},{14530, {{27238, 1, 0, 0},{27238, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{14531, {{27239, 1, 0, 0},{27239, 1, 1, 17}, {0, 0, 0, 0}}},{14531, {{27239, 1, 0, 0},{27239, 1, 1, 17}, {0, 0, 0, 0}}},{14532, {{27239, 1, 0, 0},{27239, 1, 1, 18}, {0, 0, 0, 0}}},{14532, {{27239, 1, 0, 0},{27239, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14534, {{27240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14535, {{27241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14536, {{27242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14537, {{27243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14538, {{27244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14539, {{27245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14540, {{27246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14541, {{27247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14542, {{27248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14543, {{27249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14544, {{27250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{14545, {{27251, 1, 0, 0},{27251, 1, 1, 18}, {0, 0, 0, 0}}},{14545, {{27251, 1, 0, 0},{27251, 1, 1, 18}, {0, 0, 0, 0}}},{14546, {{27251, 1, 0, 0},{27251, 1, 1, 20}, {0, 0, 0, 0}}},{14546, {{27251, 1, 0, 0},{27251, 1, 1, 20}, {0, 0, 0, 0}}},{14547, {{27251, 1, 0, 0},{27251, 1, 1, 19}, {0, 0, 0, 0}}},{14547, {{27251, 1, 0, 0},{27251, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{14548, {{27256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{14549, {{27258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{14550, {{27261, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14551, {{27262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14552, {{27263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14554, {{27264, 1, 0, 0},{27264, 1, 1, 17}, {0, 0, 0, 0}}},{14554, {{27264, 1, 0, 0},{27264, 1, 1, 17}, {0, 0, 0, 0}}},{14555, {{27264, 1, 0, 0},{27264, 1, 1, 18}, {0, 0, 0, 0}}},{14555, {{27264, 1, 0, 0},{27264, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14556, {{27265, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14557, {{27266, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14558, {{27267, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14559, {{27268, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14560, {{27269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14561, {{27270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14562, {{27271, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14563, {{27272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14564, {{27273, 1, 0, 0},{27273, 1, 1, 17}, {0, 0, 0, 0}}},{14564, {{27273, 1, 0, 0},{27273, 1, 1, 17}, {0, 0, 0, 0}}},{14565, {{27273, 1, 0, 0},{27273, 1, 1, 18}, {0, 0, 0, 0}}},{14565, {{27273, 1, 0, 0},{27273, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14566, {{27274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14567, {{27275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14568, {{27276, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14569, {{27277, 1, 0, 0},{27277, 1, 1, 17}, {0, 0, 0, 0}}},{14569, {{27277, 1, 0, 0},{27277, 1, 1, 17}, {0, 0, 0, 0}}},{14570, {{27277, 1, 0, 0},{27277, 1, 1, 18}, {0, 0, 0, 0}}},{14570, {{27277, 1, 0, 0},{27277, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{14571, {{27278, 1, 0, 0},{27278, 1, 1, 18}, {0, 0, 0, 0}}},{14571, {{27278, 1, 0, 0},{27278, 1, 1, 18}, {0, 0, 0, 0}}},{57842, {{27278, 1, 0, 0},{27278, 1, 1, 19}, {0, 0, 0, 0}}},{57842, {{27278, 1, 0, 0},{27278, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{14572, {{27279, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14573, {{27280, 1, 0, 0},{27280, 1, 1, 18}, {0, 0, 0, 0}}},{14573, {{27280, 1, 0, 0},{27280, 1, 1, 18}, {0, 0, 0, 0}}},{14574, {{27280, 1, 0, 0},{27280, 1, 1, 19}, {0, 0, 0, 0}}},{14574, {{27280, 1, 0, 0},{27280, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{14575, {{27281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14576, {{27282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14577, {{27283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14578, {{27284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14580, {{27285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14581, {{27286, 1, 0, 0},{27286, 1, 1, 16}, {0, 0, 0, 0}}},{14581, {{27286, 1, 0, 0},{27286, 1, 1, 16}, {0, 0, 0, 0}}},{14582, {{27286, 1, 0, 0},{27286, 1, 1, 17}, {0, 0, 0, 0}}},{14582, {{27286, 1, 0, 0},{27286, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{14583, {{27287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14584, {{27288, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14585, {{27289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14586, {{27290, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14587, {{27291, 1, 0, 0},{27291, 1, 1, 18}, {0, 0, 0, 0}}},{14587, {{27291, 1, 0, 0},{27291, 1, 1, 18}, {0, 0, 0, 0}}},{14588, {{27291, 1, 0, 0},{27291, 1, 1, 17}, {0, 0, 0, 0}}},{14588, {{27291, 1, 0, 0},{27291, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {8, {{14589, {{27292, 1, 0, 0},{27292, 1, 1, 20}, {0, 0, 0, 0}}},{14589, {{27292, 1, 0, 0},{27292, 1, 1, 20}, {0, 0, 0, 0}}},{14590, {{27292, 1, 0, 0},{27292, 1, 1, 18}, {0, 0, 0, 0}}},{14590, {{27292, 1, 0, 0},{27292, 1, 1, 18}, {0, 0, 0, 0}}},{14591, {{27292, 1, 0, 0},{27292, 1, 1, 21}, {0, 0, 0, 0}}},{14591, {{27292, 1, 0, 0},{27292, 1, 1, 21}, {0, 0, 0, 0}}},{14592, {{27292, 1, 0, 0},{27292, 1, 1, 19}, {0, 0, 0, 0}}},{14592, {{27292, 1, 0, 0},{27292, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{14593, {{27293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14594, {{27294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14595, {{27295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14596, {{27296, 1, 0, 0},{27296, 1, 1, 17}, {0, 0, 0, 0}}},{14596, {{27296, 1, 0, 0},{27296, 1, 1, 17}, {0, 0, 0, 0}}},{14597, {{27296, 1, 0, 0},{27296, 1, 1, 18}, {0, 0, 0, 0}}},{14597, {{27296, 1, 0, 0},{27296, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14598, {{27297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14599, {{27298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14600, {{27299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14601, {{27300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14602, {{27301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14603, {{27302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14604, {{27303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14605, {{27304, 1, 0, 0},{27304, 1, 1, 16}, {0, 0, 0, 0}}},{14605, {{27304, 1, 0, 0},{27304, 1, 1, 16}, {0, 0, 0, 0}}},{57832, {{27304, 1, 0, 0},{27304, 1, 1, 17}, {0, 0, 0, 0}}},{57832, {{27304, 1, 0, 0},{27304, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{14606, {{27306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14607, {{27307, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{14608, {{27308, 1, 0, 0},{27308, 1, 1, 17}, {0, 0, 0, 0}}},{14608, {{27308, 1, 0, 0},{27308, 1, 1, 17}, {0, 0, 0, 0}}},{14609, {{27308, 1, 0, 0},{27308, 1, 1, 18}, {0, 0, 0, 0}}},{14609, {{27308, 1, 0, 0},{27308, 1, 1, 18}, {0, 0, 0, 0}}},{14610, {{27308, 1, 0, 0},{27308, 1, 1, 19}, {0, 0, 0, 0}}},{14610, {{27308, 1, 0, 0},{27308, 1, 1, 19}, {0, 0, 0, 0}}},{14611, {{27308, 1, 0, 0},{27308, 1, 1, 20}, {0, 0, 0, 0}}},{14611, {{27308, 1, 0, 0},{27308, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{14612, {{27309, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14613, {{27310, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14614, {{27311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14615, {{27312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14616, {{27313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14617, {{27314, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{14618, {{27315, 1, 0, 0},{27315, 1, 1, 17}, {0, 0, 0, 0}}},{14618, {{27315, 1, 0, 0},{27315, 1, 1, 17}, {0, 0, 0, 0}}},{14619, {{27315, 1, 0, 0},{27315, 1, 1, 19}, {0, 0, 0, 0}}},{14619, {{27315, 1, 0, 0},{27315, 1, 1, 19}, {0, 0, 0, 0}}},{14620, {{27315, 1, 0, 0},{27315, 1, 1, 18}, {0, 0, 0, 0}}},{14620, {{27315, 1, 0, 0},{27315, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{14621, {{27316, 1, 0, 0},{27316, 1, 1, 17}, {0, 0, 0, 0}}},{14621, {{27316, 1, 0, 0},{27316, 1, 1, 17}, {0, 0, 0, 0}}},{14622, {{27316, 1, 0, 0},{27316, 1, 1, 18}, {0, 0, 0, 0}}},{14622, {{27316, 1, 0, 0},{27316, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14623, {{27317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14624, {{27318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14625, {{27319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14626, {{27320, 1, 0, 0},{27320, 1, 1, 17}, {0, 0, 0, 0}}},{14626, {{27320, 1, 0, 0},{27320, 1, 1, 17}, {0, 0, 0, 0}}},{14627, {{27320, 1, 0, 0},{27320, 1, 1, 18}, {0, 0, 0, 0}}},{14627, {{27320, 1, 0, 0},{27320, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14628, {{27321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14629, {{27322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14630, {{27323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14631, {{27324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14633, {{27325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14634, {{27326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14635, {{27327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14636, {{27328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14637, {{27329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14638, {{27330, 1, 0, 0},{27330, 1, 1, 17}, {0, 0, 0, 0}}},{14638, {{27330, 1, 0, 0},{27330, 1, 1, 17}, {0, 0, 0, 0}}},{14639, {{27330, 1, 0, 0},{27330, 1, 1, 18}, {0, 0, 0, 0}}},{14639, {{27330, 1, 0, 0},{27330, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14640, {{27331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14641, {{27332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14642, {{27333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14643, {{27334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14644, {{27335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14645, {{27336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14646, {{27337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{14648, {{27339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14649, {{27340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14650, {{27341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14651, {{27342, 1, 0, 0},{27342, 1, 1, 16}, {0, 0, 0, 0}}},{14651, {{27342, 1, 0, 0},{27342, 1, 1, 16}, {0, 0, 0, 0}}},{14652, {{27342, 1, 0, 0},{27342, 1, 1, 17}, {0, 0, 0, 0}}},{14652, {{27342, 1, 0, 0},{27342, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{14653, {{27343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14654, {{27344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14655, {{27345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14656, {{27346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14657, {{27347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{14658, {{27348, 1, 0, 0},{27348, 1, 1, 17}, {0, 0, 0, 0}}},{14658, {{27348, 1, 0, 0},{27348, 1, 1, 17}, {0, 0, 0, 0}}},{14659, {{27348, 1, 0, 0},{27348, 1, 1, 18}, {0, 0, 0, 0}}},{14659, {{27348, 1, 0, 0},{27348, 1, 1, 18}, {0, 0, 0, 0}}},{14660, {{27348, 1, 0, 0},{27348, 1, 1, 19}, {0, 0, 0, 0}}},{14660, {{27348, 1, 0, 0},{27348, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{14662, {{27349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14663, {{27350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14665, {{27351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14666, {{27352, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14667, {{27353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14668, {{27354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{14669, {{27355, 1, 0, 0},{27355, 1, 1, 19}, {0, 0, 0, 0}}},{14669, {{27355, 1, 0, 0},{27355, 1, 1, 19}, {0, 0, 0, 0}}},{14670, {{27355, 1, 0, 0},{27355, 1, 1, 21}, {0, 0, 0, 0}}},{14670, {{27355, 1, 0, 0},{27355, 1, 1, 21}, {0, 0, 0, 0}}},{14671, {{27355, 1, 0, 0},{27355, 1, 1, 20}, {0, 0, 0, 0}}},{14671, {{27355, 1, 0, 0},{27355, 1, 1, 20}, {0, 0, 0, 0}}},{14672, {{27355, 1, 0, 0},{27355, 1, 1, 22}, {0, 0, 0, 0}}},{14672, {{27355, 1, 0, 0},{27355, 1, 1, 22}, {0, 0, 0, 0}}}}},
    {4, {{14673, {{27356, 1, 0, 0},{27356, 1, 1, 17}, {0, 0, 0, 0}}},{14673, {{27356, 1, 0, 0},{27356, 1, 1, 17}, {0, 0, 0, 0}}},{14674, {{27356, 1, 0, 0},{27356, 1, 1, 18}, {0, 0, 0, 0}}},{14674, {{27356, 1, 0, 0},{27356, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{14675, {{27357, 1, 0, 0},{27357, 1, 1, 17}, {0, 0, 0, 0}}},{14675, {{27357, 1, 0, 0},{27357, 1, 1, 17}, {0, 0, 0, 0}}},{14676, {{27357, 1, 0, 0},{27357, 1, 1, 18}, {0, 0, 0, 0}}},{14676, {{27357, 1, 0, 0},{27357, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {8, {{14677, {{27358, 1, 0, 0},{27358, 1, 1, 17}, {0, 0, 0, 0}}},{14677, {{27358, 1, 0, 0},{27358, 1, 1, 17}, {0, 0, 0, 0}}},{14678, {{27358, 1, 0, 0},{27358, 1, 1, 19}, {0, 0, 0, 0}}},{14678, {{27358, 1, 0, 0},{27358, 1, 1, 19}, {0, 0, 0, 0}}},{14679, {{27358, 1, 0, 0},{27358, 1, 1, 20}, {0, 0, 0, 0}}},{14679, {{27358, 1, 0, 0},{27358, 1, 1, 20}, {0, 0, 0, 0}}},{14680, {{27358, 1, 0, 0},{27358, 1, 1, 18}, {0, 0, 0, 0}}},{14680, {{27358, 1, 0, 0},{27358, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14681, {{27359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14682, {{27360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14683, {{27361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14684, {{27362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14685, {{27363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14686, {{27364, 1, 0, 0},{27364, 1, 1, 17}, {0, 0, 0, 0}}},{14686, {{27364, 1, 0, 0},{27364, 1, 1, 17}, {0, 0, 0, 0}}},{14687, {{27364, 1, 0, 0},{27364, 1, 1, 18}, {0, 0, 0, 0}}},{14687, {{27364, 1, 0, 0},{27364, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{57861, {{27365, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{14688, {{27367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14689, {{27368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14690, {{27369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14691, {{27370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14693, {{27371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14694, {{27372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14695, {{27373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14696, {{27374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14697, {{27375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14698, {{27376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14699, {{27377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14700, {{27378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14701, {{27379, 1, 0, 0},{27379, 1, 1, 17}, {0, 0, 0, 0}}},{14701, {{27379, 1, 0, 0},{27379, 1, 1, 17}, {0, 0, 0, 0}}},{14702, {{27379, 1, 0, 0},{27379, 1, 1, 18}, {0, 0, 0, 0}}},{14702, {{27379, 1, 0, 0},{27379, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14703, {{27380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{68036, {{27382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{14704, {{27385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14705, {{27386, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14706, {{27387, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14707, {{27388, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14708, {{27389, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14709, {{27390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14710, {{27391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14711, {{27392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14712, {{27393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14713, {{27394, 1, 0, 0},{27394, 1, 1, 17}, {0, 0, 0, 0}}},{14713, {{27394, 1, 0, 0},{27394, 1, 1, 17}, {0, 0, 0, 0}}},{14714, {{27394, 1, 0, 0},{27394, 1, 1, 18}, {0, 0, 0, 0}}},{14714, {{27394, 1, 0, 0},{27394, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14715, {{27395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{14716, {{27396, 1, 0, 0},{27396, 1, 1, 18}, {0, 0, 0, 0}}},{14716, {{27396, 1, 0, 0},{27396, 1, 1, 18}, {0, 0, 0, 0}}},{30183, {{27396, 1, 0, 0},{27396, 1, 1, 0},{27396, 1, 1, 19},{63773, 1, 0, 0}, {0, 0, 0, 0}}},{30183, {{27396, 1, 0, 0},{27396, 1, 1, 0},{27396, 1, 1, 19},{63773, 1, 0, 0}, {0, 0, 0, 0}}},{30183, {{27396, 1, 0, 0},{27396, 1, 1, 0},{27396, 1, 1, 19},{63773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14717, {{27397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14718, {{27398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14719, {{27399, 1, 0, 0},{27399, 1, 1, 17}, {0, 0, 0, 0}}},{14719, {{27399, 1, 0, 0},{27399, 1, 1, 17}, {0, 0, 0, 0}}},{14720, {{27399, 1, 0, 0},{27399, 1, 1, 18}, {0, 0, 0, 0}}},{14720, {{27399, 1, 0, 0},{27399, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14721, {{27400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14722, {{27401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{14723, {{27402, 1, 0, 0},{27402, 1, 1, 17}, {0, 0, 0, 0}}},{14723, {{27402, 1, 0, 0},{27402, 1, 1, 17}, {0, 0, 0, 0}}},{14724, {{27402, 1, 0, 0},{27402, 1, 1, 18}, {0, 0, 0, 0}}},{14724, {{27402, 1, 0, 0},{27402, 1, 1, 18}, {0, 0, 0, 0}}},{59750, {{27402, 1, 0, 0},{27402, 1, 1, 19}, {0, 0, 0, 0}}},{59750, {{27402, 1, 0, 0},{27402, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{14725, {{27403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{68037, {{27404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{14726, {{27407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14727, {{27408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14728, {{27409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14729, {{27410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14730, {{27411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14732, {{27412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{14733, {{27414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14734, {{27415, 1, 0, 0},{27415, 1, 1, 17}, {0, 0, 0, 0}}},{14734, {{27415, 1, 0, 0},{27415, 1, 1, 17}, {0, 0, 0, 0}}},{14735, {{27415, 1, 0, 0},{27415, 1, 1, 18}, {0, 0, 0, 0}}},{14735, {{27415, 1, 0, 0},{27415, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14737, {{27416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14738, {{27417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14739, {{27418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14740, {{27419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14741, {{27420, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14742, {{27421, 1, 0, 0},{27421, 1, 1, 18}, {0, 0, 0, 0}}},{14742, {{27421, 1, 0, 0},{27421, 1, 1, 18}, {0, 0, 0, 0}}},{14743, {{27421, 1, 0, 0},{27421, 1, 1, 19}, {0, 0, 0, 0}}},{14743, {{27421, 1, 0, 0},{27421, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{14744, {{27422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14745, {{27423, 1, 0, 0},{27423, 1, 1, 17}, {0, 0, 0, 0}}},{14745, {{27423, 1, 0, 0},{27423, 1, 1, 17}, {0, 0, 0, 0}}},{14746, {{27423, 1, 0, 0},{27423, 1, 1, 18}, {0, 0, 0, 0}}},{14746, {{27423, 1, 0, 0},{27423, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14747, {{27424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14748, {{27425, 1, 0, 0},{27425, 1, 1, 20}, {0, 0, 0, 0}}},{14748, {{27425, 1, 0, 0},{27425, 1, 1, 20}, {0, 0, 0, 0}}},{14749, {{27425, 1, 0, 0},{27425, 1, 1, 19}, {0, 0, 0, 0}}},{14749, {{27425, 1, 0, 0},{27425, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{14750, {{27427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14751, {{27428, 1, 0, 0},{27428, 1, 1, 18}, {0, 0, 0, 0}}},{14751, {{27428, 1, 0, 0},{27428, 1, 1, 18}, {0, 0, 0, 0}}},{57876, {{27428, 1, 0, 0},{27428, 1, 1, 19}, {0, 0, 0, 0}}},{57876, {{27428, 1, 0, 0},{27428, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{14753, {{27429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14754, {{27430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14755, {{27431, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14756, {{27432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14757, {{27433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14758, {{27434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14759, {{27435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14760, {{27436, 1, 0, 0},{27436, 1, 1, 17}, {0, 0, 0, 0}}},{14760, {{27436, 1, 0, 0},{27436, 1, 1, 17}, {0, 0, 0, 0}}},{14761, {{27436, 1, 0, 0},{27436, 1, 1, 18}, {0, 0, 0, 0}}},{14761, {{27436, 1, 0, 0},{27436, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14762, {{27437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14763, {{27438, 1, 0, 0},{27438, 1, 1, 16}, {0, 0, 0, 0}}},{14763, {{27438, 1, 0, 0},{27438, 1, 1, 16}, {0, 0, 0, 0}}},{57877, {{27438, 1, 0, 0},{27438, 1, 1, 17}, {0, 0, 0, 0}}},{57877, {{27438, 1, 0, 0},{27438, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{14764, {{27439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14765, {{27440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14766, {{27441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14767, {{27442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14768, {{27443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14769, {{27444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14770, {{27445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14771, {{27446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14772, {{27447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14773, {{27448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14774, {{27449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14775, {{27450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14776, {{27451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14777, {{27452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14778, {{27453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14779, {{27454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14780, {{27455, 1, 0, 0},{27455, 1, 1, 17}, {0, 0, 0, 0}}},{14780, {{27455, 1, 0, 0},{27455, 1, 1, 17}, {0, 0, 0, 0}}},{59754, {{27455, 1, 0, 0},{27455, 1, 1, 18}, {0, 0, 0, 0}}},{59754, {{27455, 1, 0, 0},{27455, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14781, {{27456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14782, {{27457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14783, {{27458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14784, {{27459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14785, {{27460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14786, {{27461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14787, {{27462, 1, 0, 0},{27462, 1, 1, 17}, {0, 0, 0, 0}}},{14787, {{27462, 1, 0, 0},{27462, 1, 1, 17}, {0, 0, 0, 0}}},{14788, {{27462, 1, 0, 0},{27462, 1, 1, 18}, {0, 0, 0, 0}}},{14788, {{27462, 1, 0, 0},{27462, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{14789, {{27463, 1, 0, 0},{27463, 1, 1, 18}, {0, 0, 0, 0}}},{14789, {{27463, 1, 0, 0},{27463, 1, 1, 18}, {0, 0, 0, 0}}},{14790, {{27463, 1, 0, 0},{27463, 1, 1, 17}, {0, 0, 0, 0}}},{14790, {{27463, 1, 0, 0},{27463, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{14791, {{27464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14792, {{27465, 1, 0, 0},{27465, 1, 1, 17}, {0, 0, 0, 0}}},{14792, {{27465, 1, 0, 0},{27465, 1, 1, 17}, {0, 0, 0, 0}}},{14793, {{27465, 1, 0, 0},{27465, 1, 1, 18}, {0, 0, 0, 0}}},{14793, {{27465, 1, 0, 0},{27465, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14794, {{27466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14795, {{27467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14796, {{27468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14797, {{27469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{14798, {{27470, 1, 0, 0},{27470, 1, 1, 18}, {0, 0, 0, 0}}},{14798, {{27470, 1, 0, 0},{27470, 1, 1, 18}, {0, 0, 0, 0}}},{14799, {{27470, 1, 0, 0},{27470, 1, 1, 19}, {0, 0, 0, 0}}},{14799, {{27470, 1, 0, 0},{27470, 1, 1, 19}, {0, 0, 0, 0}}},{57884, {{27470, 1, 0, 0},{27470, 1, 1, 20}, {0, 0, 0, 0}}},{57884, {{27470, 1, 0, 0},{27470, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{14800, {{27471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14801, {{27472, 1, 0, 0},{27472, 1, 1, 17}, {0, 0, 0, 0}}},{14801, {{27472, 1, 0, 0},{27472, 1, 1, 17}, {0, 0, 0, 0}}},{14802, {{27472, 1, 0, 0},{27472, 1, 1, 18}, {0, 0, 0, 0}}},{14802, {{27472, 1, 0, 0},{27472, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14803, {{27473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14804, {{27474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14805, {{27475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14806, {{27476, 1, 0, 0},{27476, 1, 1, 18}, {0, 0, 0, 0}}},{14806, {{27476, 1, 0, 0},{27476, 1, 1, 18}, {0, 0, 0, 0}}},{14807, {{27476, 1, 0, 0},{27476, 1, 1, 17}, {0, 0, 0, 0}}},{14807, {{27476, 1, 0, 0},{27476, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{14808, {{27477, 1, 0, 0},{27477, 1, 1, 16}, {0, 0, 0, 0}}},{14808, {{27477, 1, 0, 0},{27477, 1, 1, 16}, {0, 0, 0, 0}}},{14809, {{27477, 1, 0, 0},{27477, 1, 1, 17}, {0, 0, 0, 0}}},{14809, {{27477, 1, 0, 0},{27477, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{14810, {{27478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14811, {{27479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14812, {{27480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14813, {{27481, 1, 0, 0},{27481, 1, 1, 17}, {0, 0, 0, 0}}},{14813, {{27481, 1, 0, 0},{27481, 1, 1, 17}, {0, 0, 0, 0}}},{14814, {{27481, 1, 0, 0},{27481, 1, 1, 18}, {0, 0, 0, 0}}},{14814, {{27481, 1, 0, 0},{27481, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{14815, {{27483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14816, {{27484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14817, {{27485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14818, {{27486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14819, {{27487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14820, {{27488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14821, {{27489, 1, 0, 0},{27489, 1, 1, 17}, {0, 0, 0, 0}}},{14821, {{27489, 1, 0, 0},{27489, 1, 1, 17}, {0, 0, 0, 0}}},{14822, {{27489, 1, 0, 0},{27489, 1, 1, 18}, {0, 0, 0, 0}}},{14822, {{27489, 1, 0, 0},{27489, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{14823, {{27490, 1, 0, 0},{27490, 1, 1, 17}, {0, 0, 0, 0}}},{14823, {{27490, 1, 0, 0},{27490, 1, 1, 17}, {0, 0, 0, 0}}},{14824, {{27490, 1, 0, 0},{27490, 1, 1, 18}, {0, 0, 0, 0}}},{14824, {{27490, 1, 0, 0},{27490, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14825, {{27491, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14826, {{27492, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14827, {{27493, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14828, {{27494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14829, {{27495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14830, {{27496, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14831, {{27497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14832, {{27498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14833, {{27499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14834, {{27500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14835, {{27501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14836, {{27502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14837, {{27503, 1, 0, 0},{27503, 1, 1, 17}, {0, 0, 0, 0}}},{14837, {{27503, 1, 0, 0},{27503, 1, 1, 17}, {0, 0, 0, 0}}},{59281, {{27503, 1, 0, 0},{27503, 1, 1, 18}, {0, 0, 0, 0}}},{59281, {{27503, 1, 0, 0},{27503, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{14838, {{27504, 1, 0, 0},{27504, 1, 1, 17}, {0, 0, 0, 0}}},{14838, {{27504, 1, 0, 0},{27504, 1, 1, 17}, {0, 0, 0, 0}}},{14839, {{27504, 1, 0, 0},{27504, 1, 1, 18}, {0, 0, 0, 0}}},{14839, {{27504, 1, 0, 0},{27504, 1, 1, 18}, {0, 0, 0, 0}}},{57898, {{27504, 1, 0, 0},{27504, 1, 1, 19}, {0, 0, 0, 0}}},{57898, {{27504, 1, 0, 0},{27504, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{14840, {{27505, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {10, {{14841, {{27506, 1, 0, 0},{27506, 1, 1, 20}, {0, 0, 0, 0}}},{14841, {{27506, 1, 0, 0},{27506, 1, 1, 20}, {0, 0, 0, 0}}},{14842, {{27506, 1, 0, 0},{27506, 1, 1, 17}, {0, 0, 0, 0}}},{14842, {{27506, 1, 0, 0},{27506, 1, 1, 17}, {0, 0, 0, 0}}},{14843, {{27506, 1, 0, 0},{27506, 1, 1, 18}, {0, 0, 0, 0}}},{14843, {{27506, 1, 0, 0},{27506, 1, 1, 18}, {0, 0, 0, 0}}},{57897, {{27506, 1, 0, 0},{27506, 1, 1, 21}, {0, 0, 0, 0}}},{57897, {{27506, 1, 0, 0},{27506, 1, 1, 21}, {0, 0, 0, 0}}},{57899, {{27506, 1, 0, 0},{27506, 1, 1, 19}, {0, 0, 0, 0}}},{57899, {{27506, 1, 0, 0},{27506, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{14844, {{27507, 1, 0, 0},{27507, 1, 1, 17}, {0, 0, 0, 0}}},{14844, {{27507, 1, 0, 0},{27507, 1, 1, 17}, {0, 0, 0, 0}}},{59760, {{27507, 1, 0, 0},{27507, 1, 1, 19}, {0, 0, 0, 0}}},{59760, {{27507, 1, 0, 0},{27507, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{14845, {{27508, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14846, {{27509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14847, {{27510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14848, {{27511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14849, {{27512, 1, 0, 0},{27512, 1, 1, 17}, {0, 0, 0, 0}}},{14849, {{27512, 1, 0, 0},{27512, 1, 1, 17}, {0, 0, 0, 0}}},{14850, {{27512, 1, 0, 0},{27512, 1, 1, 18}, {0, 0, 0, 0}}},{14850, {{27512, 1, 0, 0},{27512, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14851, {{27513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14852, {{27514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14853, {{27515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{14854, {{27517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14855, {{27518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14856, {{27519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14857, {{27520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14858, {{27521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14859, {{27522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14860, {{27523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14861, {{27524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14862, {{27525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14863, {{27526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{14864, {{27528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14865, {{27529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14866, {{27530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14867, {{27531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14868, {{27532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14869, {{27533, 1, 0, 0},{27533, 1, 1, 17}, {0, 0, 0, 0}}},{14869, {{27533, 1, 0, 0},{27533, 1, 1, 17}, {0, 0, 0, 0}}},{14870, {{27533, 1, 0, 0},{27533, 1, 1, 18}, {0, 0, 0, 0}}},{14870, {{27533, 1, 0, 0},{27533, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14871, {{27534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14872, {{27535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14873, {{27536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14874, {{27537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{14875, {{27540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14876, {{27541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14877, {{27542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14878, {{27543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14879, {{27544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14880, {{27545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{14881, {{27547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14882, {{27548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14883, {{27549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14884, {{27550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14885, {{27551, 1, 0, 0},{27551, 1, 1, 17}, {0, 0, 0, 0}}},{14885, {{27551, 1, 0, 0},{27551, 1, 1, 17}, {0, 0, 0, 0}}},{57909, {{27551, 1, 0, 0},{27551, 1, 1, 18}, {0, 0, 0, 0}}},{57909, {{27551, 1, 0, 0},{27551, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14886, {{27552, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{14887, {{27554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14888, {{27555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14889, {{27556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14890, {{27557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14891, {{27558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14892, {{27559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14893, {{27560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14894, {{27561, 1, 0, 0},{27561, 1, 1, 17}, {0, 0, 0, 0}}},{14894, {{27561, 1, 0, 0},{27561, 1, 1, 17}, {0, 0, 0, 0}}},{14895, {{27561, 1, 0, 0},{27561, 1, 1, 18}, {0, 0, 0, 0}}},{14895, {{27561, 1, 0, 0},{27561, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14896, {{27562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14897, {{27563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14898, {{27564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14899, {{27565, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14900, {{27566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{14901, {{27567, 1, 0, 0},{27567, 1, 1, 17}, {0, 0, 0, 0}}},{14901, {{27567, 1, 0, 0},{27567, 1, 1, 17}, {0, 0, 0, 0}}},{14902, {{27567, 1, 0, 0},{27567, 1, 1, 19}, {0, 0, 0, 0}}},{14902, {{27567, 1, 0, 0},{27567, 1, 1, 19}, {0, 0, 0, 0}}},{14903, {{27567, 1, 0, 0},{27567, 1, 1, 18}, {0, 0, 0, 0}}},{14903, {{27567, 1, 0, 0},{27567, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{14904, {{27568, 1, 0, 0},{27568, 1, 1, 17}, {0, 0, 0, 0}}},{14904, {{27568, 1, 0, 0},{27568, 1, 1, 17}, {0, 0, 0, 0}}},{14905, {{27568, 1, 0, 0},{27568, 1, 1, 18}, {0, 0, 0, 0}}},{14905, {{27568, 1, 0, 0},{27568, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14906, {{27569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14907, {{27570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14908, {{27571, 1, 0, 0},{27571, 1, 1, 17}, {0, 0, 0, 0}}},{14908, {{27571, 1, 0, 0},{27571, 1, 1, 17}, {0, 0, 0, 0}}},{14909, {{27571, 1, 0, 0},{27571, 1, 1, 18}, {0, 0, 0, 0}}},{14909, {{27571, 1, 0, 0},{27571, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14910, {{27572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14911, {{27573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14912, {{27574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14913, {{27575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14914, {{27576, 1, 0, 0},{27576, 1, 1, 17}, {0, 0, 0, 0}}},{14914, {{27576, 1, 0, 0},{27576, 1, 1, 17}, {0, 0, 0, 0}}},{57918, {{27576, 1, 0, 0},{27576, 1, 1, 18}, {0, 0, 0, 0}}},{57918, {{27576, 1, 0, 0},{27576, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14915, {{27577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{14916, {{27578, 1, 0, 0},{27578, 1, 1, 18}, {0, 0, 0, 0}}},{14916, {{27578, 1, 0, 0},{27578, 1, 1, 18}, {0, 0, 0, 0}}},{30187, {{27578, 1, 0, 0},{27578, 1, 1, 0},{27578, 1, 1, 19},{63856, 1, 0, 0}, {0, 0, 0, 0}}},{30187, {{27578, 1, 0, 0},{27578, 1, 1, 0},{27578, 1, 1, 19},{63856, 1, 0, 0}, {0, 0, 0, 0}}},{30187, {{27578, 1, 0, 0},{27578, 1, 1, 0},{27578, 1, 1, 19},{63856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14917, {{27579, 1, 0, 0},{27579, 1, 1, 18}, {0, 0, 0, 0}}},{14917, {{27579, 1, 0, 0},{27579, 1, 1, 18}, {0, 0, 0, 0}}},{14918, {{27579, 1, 0, 0},{27579, 1, 1, 19}, {0, 0, 0, 0}}},{14918, {{27579, 1, 0, 0},{27579, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{14919, {{27580, 1, 0, 0},{27580, 1, 1, 17}, {0, 0, 0, 0}}},{14919, {{27580, 1, 0, 0},{27580, 1, 1, 17}, {0, 0, 0, 0}}},{14920, {{27580, 1, 0, 0},{27580, 1, 1, 18}, {0, 0, 0, 0}}},{14920, {{27580, 1, 0, 0},{27580, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14921, {{27581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14922, {{27582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14923, {{27583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14924, {{27584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14925, {{27585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{14926, {{27587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14927, {{27588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14928, {{27589, 1, 0, 0},{27589, 1, 1, 17}, {0, 0, 0, 0}}},{14928, {{27589, 1, 0, 0},{27589, 1, 1, 17}, {0, 0, 0, 0}}},{60199, {{27589, 1, 0, 0},{27589, 1, 1, 18}, {0, 0, 0, 0}}},{60199, {{27589, 1, 0, 0},{27589, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{14929, {{27590, 1, 0, 0},{27590, 1, 1, 17}, {0, 0, 0, 0}}},{14929, {{27590, 1, 0, 0},{27590, 1, 1, 17}, {0, 0, 0, 0}}},{14930, {{27590, 1, 0, 0},{27590, 1, 1, 18}, {0, 0, 0, 0}}},{14930, {{27590, 1, 0, 0},{27590, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14931, {{27591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14932, {{27592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14933, {{27593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14934, {{27594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14935, {{27595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14937, {{27596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14938, {{27597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14939, {{27598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14940, {{27599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14941, {{27600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14942, {{27601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14943, {{27602, 1, 0, 0},{27602, 1, 1, 17}, {0, 0, 0, 0}}},{14943, {{27602, 1, 0, 0},{27602, 1, 1, 17}, {0, 0, 0, 0}}},{57930, {{27602, 1, 0, 0},{27602, 1, 1, 18}, {0, 0, 0, 0}}},{57930, {{27602, 1, 0, 0},{27602, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{14944, {{27603, 1, 0, 0},{27603, 1, 1, 17}, {0, 0, 0, 0}}},{14944, {{27603, 1, 0, 0},{27603, 1, 1, 17}, {0, 0, 0, 0}}},{14945, {{27603, 1, 0, 0},{27603, 1, 1, 18}, {0, 0, 0, 0}}},{14945, {{27603, 1, 0, 0},{27603, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{14947, {{27604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{14948, {{27606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14949, {{27607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14950, {{27608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{14951, {{27610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14952, {{27611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{14953, {{27614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14954, {{27615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14955, {{27616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14956, {{27617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14957, {{27618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14958, {{27619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14959, {{27620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14960, {{27621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14961, {{27622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14962, {{27623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14963, {{27624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14964, {{27625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{14965, {{27627, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14966, {{27628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14967, {{27629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14968, {{27630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14969, {{27631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14970, {{27632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14971, {{27633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14972, {{27634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14973, {{27635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14974, {{27636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{14975, {{27639, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14976, {{27640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14977, {{27641, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14978, {{27642, 1, 0, 0},{27642, 1, 1, 16}, {0, 0, 0, 0}}},{14978, {{27642, 1, 0, 0},{27642, 1, 1, 16}, {0, 0, 0, 0}}},{14979, {{27642, 1, 0, 0},{27642, 1, 1, 17}, {0, 0, 0, 0}}},{14979, {{27642, 1, 0, 0},{27642, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{14980, {{27643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14981, {{27644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14982, {{27645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14983, {{27646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14984, {{27647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14985, {{27648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14986, {{27649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14987, {{27650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{14988, {{27651, 1, 0, 0},{27651, 1, 1, 16}, {0, 0, 0, 0}}},{14988, {{27651, 1, 0, 0},{27651, 1, 1, 16}, {0, 0, 0, 0}}},{14989, {{27651, 1, 0, 0},{27651, 1, 1, 17}, {0, 0, 0, 0}}},{14989, {{27651, 1, 0, 0},{27651, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{14990, {{27652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14991, {{27653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14992, {{27654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{14993, {{27656, 1, 0, 0},{27656, 1, 1, 19}, {0, 0, 0, 0}}},{14993, {{27656, 1, 0, 0},{27656, 1, 1, 19}, {0, 0, 0, 0}}},{14994, {{27656, 1, 0, 0},{27656, 1, 1, 20}, {0, 0, 0, 0}}},{14994, {{27656, 1, 0, 0},{27656, 1, 1, 20}, {0, 0, 0, 0}}},{14995, {{27656, 1, 0, 0},{27656, 1, 1, 22}, {0, 0, 0, 0}}},{14995, {{27656, 1, 0, 0},{27656, 1, 1, 22}, {0, 0, 0, 0}}}}},
    {1, {{14997, {{27657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14998, {{27658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14999, {{27659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15000, {{27660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15001, {{27661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15002, {{27662, 1, 0, 0},{27662, 1, 1, 17}, {0, 0, 0, 0}}},{15002, {{27662, 1, 0, 0},{27662, 1, 1, 17}, {0, 0, 0, 0}}},{15003, {{27662, 1, 0, 0},{27662, 1, 1, 18}, {0, 0, 0, 0}}},{15003, {{27662, 1, 0, 0},{27662, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15004, {{27663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15005, {{27664, 1, 0, 0},{27664, 1, 1, 17}, {0, 0, 0, 0}}},{15005, {{27664, 1, 0, 0},{27664, 1, 1, 17}, {0, 0, 0, 0}}},{57935, {{27664, 1, 0, 0},{27664, 1, 1, 18}, {0, 0, 0, 0}}},{57935, {{27664, 1, 0, 0},{27664, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15006, {{27665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15007, {{27666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{15008, {{27667, 1, 0, 0},{27667, 1, 1, 18}, {0, 0, 0, 0}}},{15008, {{27667, 1, 0, 0},{27667, 1, 1, 18}, {0, 0, 0, 0}}},{15009, {{27667, 1, 0, 0},{27667, 1, 1, 19}, {0, 0, 0, 0}}},{15009, {{27667, 1, 0, 0},{27667, 1, 1, 19}, {0, 0, 0, 0}}},{15010, {{27667, 1, 0, 0},{27667, 1, 1, 20}, {0, 0, 0, 0}}},{15010, {{27667, 1, 0, 0},{27667, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{15011, {{27668, 1, 0, 0},{27668, 1, 1, 17}, {0, 0, 0, 0}}},{15011, {{27668, 1, 0, 0},{27668, 1, 1, 17}, {0, 0, 0, 0}}},{56866, {{27668, 1, 0, 0},{27668, 1, 1, 18}, {0, 0, 0, 0}}},{56866, {{27668, 1, 0, 0},{27668, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{15012, {{27670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15013, {{27671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{15014, {{27673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{15015, {{27675, 1, 0, 0},{27675, 1, 1, 17}, {0, 0, 0, 0}}},{15015, {{27675, 1, 0, 0},{27675, 1, 1, 17}, {0, 0, 0, 0}}},{15016, {{27675, 1, 0, 0},{27675, 1, 1, 18}, {0, 0, 0, 0}}},{15016, {{27675, 1, 0, 0},{27675, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15017, {{27676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15018, {{27677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{15019, {{27679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{15020, {{27683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15021, {{27684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15022, {{27685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15023, {{27686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15024, {{27687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15025, {{27688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{15026, {{27690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15027, {{27691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15028, {{27692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15029, {{27693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15030, {{27694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15031, {{27695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15032, {{27696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15033, {{27697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57937, {{27698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15034, {{27699, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15035, {{27700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15036, {{27701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15037, {{27702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15038, {{27703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15039, {{27704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15040, {{27705, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15041, {{27706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15042, {{27707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15043, {{27708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15044, {{27709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15045, {{27710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15046, {{27711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15047, {{27712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15048, {{27713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{15049, {{27714, 1, 0, 0},{27714, 1, 1, 17}, {0, 0, 0, 0}}},{15049, {{27714, 1, 0, 0},{27714, 1, 1, 17}, {0, 0, 0, 0}}},{57940, {{27714, 1, 0, 0},{27714, 1, 1, 19}, {0, 0, 0, 0}}},{57940, {{27714, 1, 0, 0},{27714, 1, 1, 19}, {0, 0, 0, 0}}},{59770, {{27714, 1, 0, 0},{27714, 1, 1, 18}, {0, 0, 0, 0}}},{59770, {{27714, 1, 0, 0},{27714, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15050, {{27715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15051, {{27716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15052, {{27717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{15056, {{27722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15057, {{27723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15058, {{27724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15059, {{27725, 1, 0, 0},{27725, 1, 1, 17}, {0, 0, 0, 0}}},{15059, {{27725, 1, 0, 0},{27725, 1, 1, 17}, {0, 0, 0, 0}}},{15060, {{27725, 1, 0, 0},{27725, 1, 1, 18}, {0, 0, 0, 0}}},{15060, {{27725, 1, 0, 0},{27725, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{15061, {{27726, 1, 0, 0},{27726, 1, 1, 17}, {0, 0, 0, 0}}},{15061, {{27726, 1, 0, 0},{27726, 1, 1, 17}, {0, 0, 0, 0}}},{15062, {{27726, 1, 0, 0},{27726, 1, 1, 18}, {0, 0, 0, 0}}},{15062, {{27726, 1, 0, 0},{27726, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15063, {{27727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15064, {{27728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15065, {{27729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15066, {{27730, 1, 0, 0},{27730, 1, 1, 17}, {0, 0, 0, 0}}},{15066, {{27730, 1, 0, 0},{27730, 1, 1, 17}, {0, 0, 0, 0}}},{15067, {{27730, 1, 0, 0},{27730, 1, 1, 18}, {0, 0, 0, 0}}},{15067, {{27730, 1, 0, 0},{27730, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15068, {{27731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15069, {{27732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15070, {{27733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{15072, {{27735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15073, {{27736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15074, {{27737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15075, {{27738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15076, {{27739, 1, 0, 0},{27739, 1, 1, 17}, {0, 0, 0, 0}}},{15076, {{27739, 1, 0, 0},{27739, 1, 1, 17}, {0, 0, 0, 0}}},{15077, {{27739, 1, 0, 0},{27739, 1, 1, 18}, {0, 0, 0, 0}}},{15077, {{27739, 1, 0, 0},{27739, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15078, {{27740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15080, {{27741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15081, {{27742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15082, {{27743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15084, {{27744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15085, {{27745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15086, {{27746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{15088, {{27749, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15089, {{27750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15090, {{27751, 1, 0, 0},{27751, 1, 1, 18}, {0, 0, 0, 0}}},{15090, {{27751, 1, 0, 0},{27751, 1, 1, 18}, {0, 0, 0, 0}}},{15091, {{27751, 1, 0, 0},{27751, 1, 1, 17}, {0, 0, 0, 0}}},{15091, {{27751, 1, 0, 0},{27751, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{15093, {{27752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15094, {{27753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15095, {{27754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15096, {{27755, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15097, {{27756, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{15098, {{27757, 1, 0, 0},{27757, 1, 1, 17}, {0, 0, 0, 0}}},{15098, {{27757, 1, 0, 0},{27757, 1, 1, 17}, {0, 0, 0, 0}}},{15099, {{27757, 1, 0, 0},{27757, 1, 1, 18}, {0, 0, 0, 0}}},{15099, {{27757, 1, 0, 0},{27757, 1, 1, 18}, {0, 0, 0, 0}}},{15100, {{27757, 1, 0, 0},{27757, 1, 1, 19}, {0, 0, 0, 0}}},{15100, {{27757, 1, 0, 0},{27757, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{15101, {{27758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15102, {{27759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15103, {{27760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15104, {{27761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15105, {{27762, 1, 0, 0},{27762, 1, 1, 18}, {0, 0, 0, 0}}},{15105, {{27762, 1, 0, 0},{27762, 1, 1, 18}, {0, 0, 0, 0}}},{15106, {{27762, 1, 0, 0},{27762, 1, 1, 19}, {0, 0, 0, 0}}},{15106, {{27762, 1, 0, 0},{27762, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{15107, {{27763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15108, {{27764, 1, 0, 0},{27764, 1, 1, 17}, {0, 0, 0, 0}}},{15108, {{27764, 1, 0, 0},{27764, 1, 1, 17}, {0, 0, 0, 0}}},{15109, {{27764, 1, 0, 0},{27764, 1, 1, 18}, {0, 0, 0, 0}}},{15109, {{27764, 1, 0, 0},{27764, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15110, {{27765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15111, {{27766, 1, 0, 0},{27766, 1, 1, 17}, {0, 0, 0, 0}}},{15111, {{27766, 1, 0, 0},{27766, 1, 1, 17}, {0, 0, 0, 0}}},{15112, {{27766, 1, 0, 0},{27766, 1, 1, 18}, {0, 0, 0, 0}}},{15112, {{27766, 1, 0, 0},{27766, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15113, {{27767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15114, {{27768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15115, {{27769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15116, {{27770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15117, {{27771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15118, {{27772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15119, {{27773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15120, {{27774, 1, 0, 0},{27774, 1, 1, 17}, {0, 0, 0, 0}}},{15120, {{27774, 1, 0, 0},{27774, 1, 1, 17}, {0, 0, 0, 0}}},{15121, {{27774, 1, 0, 0},{27774, 1, 1, 18}, {0, 0, 0, 0}}},{15121, {{27774, 1, 0, 0},{27774, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15122, {{27775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15123, {{27776, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15124, {{27777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15125, {{27778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15126, {{27779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15127, {{27780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15129, {{27781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15130, {{27782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15131, {{27783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15132, {{27784, 1, 0, 0},{27784, 1, 1, 17}, {0, 0, 0, 0}}},{15132, {{27784, 1, 0, 0},{27784, 1, 1, 17}, {0, 0, 0, 0}}},{57943, {{27784, 1, 0, 0},{27784, 1, 1, 18}, {0, 0, 0, 0}}},{57943, {{27784, 1, 0, 0},{27784, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15133, {{27785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15134, {{27786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15135, {{27787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15136, {{27788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15137, {{27789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15138, {{27790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15139, {{27791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15140, {{27792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15141, {{27793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15142, {{27794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15143, {{27795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15144, {{27796, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15145, {{27797, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15146, {{27798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{15147, {{27799, 1, 0, 0},{27799, 1, 1, 19}, {0, 0, 0, 0}}},{15147, {{27799, 1, 0, 0},{27799, 1, 1, 19}, {0, 0, 0, 0}}},{15148, {{27799, 1, 0, 0},{27799, 1, 1, 17}, {0, 0, 0, 0}}},{15148, {{27799, 1, 0, 0},{27799, 1, 1, 17}, {0, 0, 0, 0}}},{15149, {{27799, 1, 0, 0},{27799, 1, 1, 18}, {0, 0, 0, 0}}},{15149, {{27799, 1, 0, 0},{27799, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15150, {{27800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15151, {{27801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15152, {{27802, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15153, {{27803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15154, {{27804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15155, {{27805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15156, {{27806, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15157, {{27807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15158, {{27808, 1, 0, 0},{27808, 1, 1, 16}, {0, 0, 0, 0}}},{15158, {{27808, 1, 0, 0},{27808, 1, 1, 16}, {0, 0, 0, 0}}},{15159, {{27808, 1, 0, 0},{27808, 1, 1, 17}, {0, 0, 0, 0}}},{15159, {{27808, 1, 0, 0},{27808, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{15160, {{27809, 1, 0, 0},{27809, 1, 1, 17}, {0, 0, 0, 0}}},{15160, {{27809, 1, 0, 0},{27809, 1, 1, 17}, {0, 0, 0, 0}}},{15161, {{27809, 1, 0, 0},{27809, 1, 1, 18}, {0, 0, 0, 0}}},{15161, {{27809, 1, 0, 0},{27809, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15162, {{27810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{15163, {{27818, 1, 0, 0},{27818, 1, 1, 19}, {0, 0, 0, 0}}},{15163, {{27818, 1, 0, 0},{27818, 1, 1, 19}, {0, 0, 0, 0}}},{15164, {{27818, 1, 0, 0},{27818, 1, 1, 20}, {0, 0, 0, 0}}},{15164, {{27818, 1, 0, 0},{27818, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{15166, {{27819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15167, {{27820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15168, {{27821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15169, {{27822, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{15170, {{27824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15171, {{27825, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15172, {{27826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15173, {{27827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15174, {{27828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15175, {{27829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15176, {{27830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15177, {{27831, 1, 0, 0},{27831, 1, 1, 17}, {0, 0, 0, 0}}},{15177, {{27831, 1, 0, 0},{27831, 1, 1, 17}, {0, 0, 0, 0}}},{15178, {{27831, 1, 0, 0},{27831, 1, 1, 16}, {0, 0, 0, 0}}},{15178, {{27831, 1, 0, 0},{27831, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{15179, {{27832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15180, {{27833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15181, {{27834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15182, {{27835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15183, {{27836, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15184, {{27837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15185, {{27838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{15186, {{27839, 1, 0, 0},{27839, 1, 1, 19}, {0, 0, 0, 0}}},{15186, {{27839, 1, 0, 0},{27839, 1, 1, 19}, {0, 0, 0, 0}}},{15187, {{27839, 1, 0, 0},{27839, 1, 1, 20}, {0, 0, 0, 0}}},{15187, {{27839, 1, 0, 0},{27839, 1, 1, 20}, {0, 0, 0, 0}}},{15188, {{27839, 1, 0, 0},{27839, 1, 1, 21}, {0, 0, 0, 0}}},{15188, {{27839, 1, 0, 0},{27839, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{15189, {{27840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15190, {{27841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15191, {{27842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15192, {{27843, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15193, {{27844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15194, {{27845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15195, {{27846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15196, {{27847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15197, {{27848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15198, {{27849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15199, {{27850, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15200, {{27851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15201, {{27852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15202, {{27853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15204, {{27854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15205, {{27855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15206, {{27856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15207, {{27857, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{15208, {{27858, 1, 0, 0},{27858, 1, 1, 19}, {0, 0, 0, 0}}},{15208, {{27858, 1, 0, 0},{27858, 1, 1, 19}, {0, 0, 0, 0}}},{15209, {{27858, 1, 0, 0},{27858, 1, 1, 17}, {0, 0, 0, 0}}},{15209, {{27858, 1, 0, 0},{27858, 1, 1, 17}, {0, 0, 0, 0}}},{15210, {{27858, 1, 0, 0},{27858, 1, 1, 18}, {0, 0, 0, 0}}},{15210, {{27858, 1, 0, 0},{27858, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15211, {{27859, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15212, {{27860, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15213, {{27861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15214, {{27862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15215, {{27863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15216, {{27864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15217, {{27865, 1, 0, 0},{27865, 1, 1, 17}, {0, 0, 0, 0}}},{15217, {{27865, 1, 0, 0},{27865, 1, 1, 17}, {0, 0, 0, 0}}},{15218, {{27865, 1, 0, 0},{27865, 1, 1, 18}, {0, 0, 0, 0}}},{15218, {{27865, 1, 0, 0},{27865, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15219, {{27866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15220, {{27867, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15221, {{27868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15222, {{27869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15223, {{27870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15224, {{27871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15225, {{27872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15226, {{27873, 1, 0, 0},{27873, 1, 1, 19}, {0, 0, 0, 0}}},{15226, {{27873, 1, 0, 0},{27873, 1, 1, 19}, {0, 0, 0, 0}}},{15227, {{27873, 1, 0, 0},{27873, 1, 1, 18}, {0, 0, 0, 0}}},{15227, {{27873, 1, 0, 0},{27873, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15228, {{27874, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15229, {{27875, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15230, {{27876, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15231, {{27877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15232, {{27878, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15233, {{27879, 1, 0, 0},{27879, 1, 1, 17}, {0, 0, 0, 0}}},{15233, {{27879, 1, 0, 0},{27879, 1, 1, 17}, {0, 0, 0, 0}}},{15234, {{27879, 1, 0, 0},{27879, 1, 1, 18}, {0, 0, 0, 0}}},{15234, {{27879, 1, 0, 0},{27879, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{15235, {{27880, 1, 0, 0},{27880, 1, 1, 19}, {0, 0, 0, 0}}},{15235, {{27880, 1, 0, 0},{27880, 1, 1, 19}, {0, 0, 0, 0}}},{15236, {{27880, 1, 0, 0},{27880, 1, 1, 20}, {0, 0, 0, 0}}},{15236, {{27880, 1, 0, 0},{27880, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{15237, {{27881, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15238, {{27882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15239, {{27883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15240, {{27884, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15241, {{27885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15242, {{27886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15243, {{27887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{15244, {{27888, 1, 0, 0},{27888, 1, 1, 18}, {0, 0, 0, 0}}},{15244, {{27888, 1, 0, 0},{27888, 1, 1, 18}, {0, 0, 0, 0}}},{15245, {{27888, 1, 0, 0},{27888, 1, 1, 17}, {0, 0, 0, 0}}},{15245, {{27888, 1, 0, 0},{27888, 1, 1, 17}, {0, 0, 0, 0}}},{57277, {{27888, 1, 0, 0},{27888, 1, 1, 19}, {0, 0, 0, 0}}},{57277, {{27888, 1, 0, 0},{27888, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{15246, {{27889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15247, {{27890, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15248, {{27891, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15249, {{27892, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15250, {{27893, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{15251, {{27899, 1, 0, 0},{27899, 1, 1, 18}, {0, 0, 0, 0}}},{15251, {{27899, 1, 0, 0},{27899, 1, 1, 18}, {0, 0, 0, 0}}},{15252, {{27899, 1, 0, 0},{27899, 1, 1, 17}, {0, 0, 0, 0}}},{15252, {{27899, 1, 0, 0},{27899, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{15255, {{27903, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15256, {{27904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15257, {{27905, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15259, {{27906, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15260, {{27907, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15261, {{27908, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15262, {{27909, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15263, {{27910, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15264, {{27911, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15265, {{27912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15266, {{27913, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15267, {{27914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15268, {{27915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15269, {{27916, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15270, {{27917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15272, {{27918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15273, {{27919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15274, {{27920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15275, {{27921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15276, {{27922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15277, {{27923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15278, {{27924, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15279, {{27925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15280, {{27926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15281, {{27927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15282, {{27928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15283, {{27929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15284, {{27930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15286, {{27931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15287, {{27932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15288, {{27933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15289, {{27934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15290, {{27935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15291, {{27936, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15292, {{27937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15293, {{27938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15294, {{27939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15295, {{27940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15296, {{27941, 1, 0, 0},{27941, 1, 1, 17}, {0, 0, 0, 0}}},{15296, {{27941, 1, 0, 0},{27941, 1, 1, 17}, {0, 0, 0, 0}}},{59778, {{27941, 1, 0, 0},{27941, 1, 1, 18}, {0, 0, 0, 0}}},{59778, {{27941, 1, 0, 0},{27941, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15297, {{27942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15298, {{27943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15299, {{27944, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15300, {{27945, 1, 0, 0},{27945, 1, 1, 17}, {0, 0, 0, 0}}},{15300, {{27945, 1, 0, 0},{27945, 1, 1, 17}, {0, 0, 0, 0}}},{57950, {{27945, 1, 0, 0},{27945, 1, 1, 18}, {0, 0, 0, 0}}},{57950, {{27945, 1, 0, 0},{27945, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15301, {{27946, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15302, {{27947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15303, {{27948, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15304, {{27949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15305, {{27950, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15306, {{27951, 1, 0, 0},{27951, 1, 1, 17}, {0, 0, 0, 0}}},{15306, {{27951, 1, 0, 0},{27951, 1, 1, 17}, {0, 0, 0, 0}}},{15307, {{27951, 1, 0, 0},{27951, 1, 1, 18}, {0, 0, 0, 0}}},{15307, {{27951, 1, 0, 0},{27951, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15308, {{27952, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15309, {{27953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15310, {{27954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15311, {{27955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15312, {{27956, 1, 0, 0},{27956, 1, 1, 18}, {0, 0, 0, 0}}},{15312, {{27956, 1, 0, 0},{27956, 1, 1, 18}, {0, 0, 0, 0}}},{15313, {{27956, 1, 0, 0},{27956, 1, 1, 17}, {0, 0, 0, 0}}},{15313, {{27956, 1, 0, 0},{27956, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{15314, {{27957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15315, {{27958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15316, {{27959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15317, {{27960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15318, {{27961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15319, {{27962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15321, {{27963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15322, {{27964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15323, {{27965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15324, {{27966, 1, 0, 0},{27966, 1, 1, 20}, {0, 0, 0, 0}}},{15324, {{27966, 1, 0, 0},{27966, 1, 1, 20}, {0, 0, 0, 0}}},{15325, {{27966, 1, 0, 0},{27966, 1, 1, 19}, {0, 0, 0, 0}}},{15325, {{27966, 1, 0, 0},{27966, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{15326, {{27967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15328, {{27968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15329, {{27969, 1, 0, 0},{27969, 1, 1, 18}, {0, 0, 0, 0}}},{15329, {{27969, 1, 0, 0},{27969, 1, 1, 18}, {0, 0, 0, 0}}},{15330, {{27969, 1, 0, 0},{27969, 1, 1, 17}, {0, 0, 0, 0}}},{15330, {{27969, 1, 0, 0},{27969, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{15331, {{27970, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15332, {{27971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15333, {{27972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15334, {{27973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{15335, {{27978, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{15336, {{27990, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15337, {{27991, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15338, {{27992, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15339, {{27993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15340, {{27994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15341, {{27995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15342, {{27996, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15343, {{27997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15344, {{27998, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15345, {{27999, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15346, {{28000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15347, {{28001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15348, {{28002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15349, {{28003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15350, {{28004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15351, {{28005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15352, {{28006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15353, {{28007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15354, {{28008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15355, {{28009, 1, 0, 0},{28009, 1, 1, 19}, {0, 0, 0, 0}}},{15355, {{28009, 1, 0, 0},{28009, 1, 1, 19}, {0, 0, 0, 0}}},{15356, {{28009, 1, 0, 0},{28009, 1, 1, 18}, {0, 0, 0, 0}}},{15356, {{28009, 1, 0, 0},{28009, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15357, {{28010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15358, {{28011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15359, {{28012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15360, {{28013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15361, {{28014, 1, 0, 0},{28014, 1, 1, 19}, {0, 0, 0, 0}}},{15361, {{28014, 1, 0, 0},{28014, 1, 1, 19}, {0, 0, 0, 0}}},{15362, {{28014, 1, 0, 0},{28014, 1, 1, 18}, {0, 0, 0, 0}}},{15362, {{28014, 1, 0, 0},{28014, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15363, {{28015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15364, {{28016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15365, {{28017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15366, {{28018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15367, {{28019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15368, {{28020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15369, {{28021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15370, {{28022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{15371, {{28023, 1, 0, 0},{28023, 1, 1, 18}, {0, 0, 0, 0}}},{15371, {{28023, 1, 0, 0},{28023, 1, 1, 18}, {0, 0, 0, 0}}},{30266, {{28023, 1, 0, 0},{28023, 1, 1, 0},{28023, 1, 1, 19},{64069, 1, 0, 0}, {0, 0, 0, 0}}},{30266, {{28023, 1, 0, 0},{28023, 1, 1, 0},{28023, 1, 1, 19},{64069, 1, 0, 0}, {0, 0, 0, 0}}},{30266, {{28023, 1, 0, 0},{28023, 1, 1, 0},{28023, 1, 1, 19},{64069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15372, {{28024, 1, 0, 0},{28024, 1, 1, 19}, {0, 0, 0, 0}}},{15372, {{28024, 1, 0, 0},{28024, 1, 1, 19}, {0, 0, 0, 0}}},{15373, {{28024, 1, 0, 0},{28024, 1, 1, 18}, {0, 0, 0, 0}}},{15373, {{28024, 1, 0, 0},{28024, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15374, {{28025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15375, {{28026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15376, {{28027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15377, {{28028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15378, {{28029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15379, {{28030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15380, {{28031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15381, {{28032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15382, {{28033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15383, {{28034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15385, {{28035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15386, {{28036, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15387, {{28037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15388, {{28038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15389, {{28039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15390, {{28040, 1, 0, 0},{28040, 1, 1, 19}, {0, 0, 0, 0}}},{15390, {{28040, 1, 0, 0},{28040, 1, 1, 19}, {0, 0, 0, 0}}},{15391, {{28040, 1, 0, 0},{28040, 1, 1, 18}, {0, 0, 0, 0}}},{15391, {{28040, 1, 0, 0},{28040, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15392, {{28041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15393, {{28042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15394, {{28043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15395, {{28044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15396, {{28045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15397, {{28046, 1, 0, 0},{28046, 1, 1, 17}, {0, 0, 0, 0}}},{15397, {{28046, 1, 0, 0},{28046, 1, 1, 17}, {0, 0, 0, 0}}},{15398, {{28046, 1, 0, 0},{28046, 1, 1, 18}, {0, 0, 0, 0}}},{15398, {{28046, 1, 0, 0},{28046, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15399, {{28047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{15400, {{28049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15401, {{28050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15402, {{28051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15403, {{28052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15404, {{28053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15405, {{28054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15406, {{28055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15408, {{28056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15409, {{28057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{15410, {{28059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15411, {{28060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{15412, {{28070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{15413, {{28074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15414, {{28075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15415, {{28076, 1, 0, 0},{28076, 1, 1, 17}, {0, 0, 0, 0}}},{15415, {{28076, 1, 0, 0},{28076, 1, 1, 17}, {0, 0, 0, 0}}},{15416, {{28076, 1, 0, 0},{28076, 1, 1, 18}, {0, 0, 0, 0}}},{15416, {{28076, 1, 0, 0},{28076, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15417, {{28077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15418, {{28078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15419, {{28079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15420, {{28080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15421, {{28081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15422, {{28082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15423, {{28083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15424, {{28084, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15425, {{28085, 1, 0, 0},{28085, 1, 1, 17}, {0, 0, 0, 0}}},{15425, {{28085, 1, 0, 0},{28085, 1, 1, 17}, {0, 0, 0, 0}}},{15426, {{28085, 1, 0, 0},{28085, 1, 1, 18}, {0, 0, 0, 0}}},{15426, {{28085, 1, 0, 0},{28085, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15427, {{28086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15428, {{28087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15429, {{28088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15430, {{28089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15431, {{28090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15432, {{28091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15433, {{28092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15434, {{28093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15435, {{28094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{15436, {{28095, 1, 0, 0},{28095, 1, 1, 17}, {0, 0, 0, 0}}},{15436, {{28095, 1, 0, 0},{28095, 1, 1, 17}, {0, 0, 0, 0}}},{15438, {{28095, 1, 0, 0},{28095, 1, 1, 18}, {0, 0, 0, 0}}},{15438, {{28095, 1, 0, 0},{28095, 1, 1, 18}, {0, 0, 0, 0}}},{68062, {{28095, 1, 0, 0},{28095, 1, 1, 19}, {0, 0, 0, 0}}},{68062, {{28095, 1, 0, 0},{28095, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{15439, {{28096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15440, {{28097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15441, {{28098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{15442, {{28099, 1, 0, 0},{28099, 1, 1, 16}, {0, 0, 0, 0}}},{15442, {{28099, 1, 0, 0},{28099, 1, 1, 16}, {0, 0, 0, 0}}},{15443, {{28099, 1, 0, 0},{28099, 1, 1, 17}, {0, 0, 0, 0}}},{15443, {{28099, 1, 0, 0},{28099, 1, 1, 17}, {0, 0, 0, 0}}},{15444, {{28099, 1, 0, 0},{28099, 1, 1, 18}, {0, 0, 0, 0}}},{15444, {{28099, 1, 0, 0},{28099, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15445, {{28100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15446, {{28101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15447, {{28102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15448, {{28103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15449, {{28104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15450, {{28105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15451, {{28106, 1, 0, 0},{28106, 1, 1, 17}, {0, 0, 0, 0}}},{15451, {{28106, 1, 0, 0},{28106, 1, 1, 17}, {0, 0, 0, 0}}},{15452, {{28106, 1, 0, 0},{28106, 1, 1, 18}, {0, 0, 0, 0}}},{15452, {{28106, 1, 0, 0},{28106, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15453, {{28107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15454, {{28108, 1, 0, 0},{28108, 1, 1, 17}, {0, 0, 0, 0}}},{15454, {{28108, 1, 0, 0},{28108, 1, 1, 17}, {0, 0, 0, 0}}},{15455, {{28108, 1, 0, 0},{28108, 1, 1, 18}, {0, 0, 0, 0}}},{15455, {{28108, 1, 0, 0},{28108, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15456, {{28109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15458, {{28110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15459, {{28111, 1, 0, 0},{28111, 1, 1, 17}, {0, 0, 0, 0}}},{15459, {{28111, 1, 0, 0},{28111, 1, 1, 17}, {0, 0, 0, 0}}},{15460, {{28111, 1, 0, 0},{28111, 1, 1, 18}, {0, 0, 0, 0}}},{15460, {{28111, 1, 0, 0},{28111, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15461, {{28112, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15462, {{28113, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15464, {{28114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15465, {{28115, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15466, {{28116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15467, {{28117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15468, {{28118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15469, {{28119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15470, {{28120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15471, {{28121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15472, {{28122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15473, {{28123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15474, {{28124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15476, {{28125, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15477, {{28126, 1, 0, 0},{28126, 1, 1, 17}, {0, 0, 0, 0}}},{15477, {{28126, 1, 0, 0},{28126, 1, 1, 17}, {0, 0, 0, 0}}},{15478, {{28126, 1, 0, 0},{28126, 1, 1, 18}, {0, 0, 0, 0}}},{15478, {{28126, 1, 0, 0},{28126, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15479, {{28127, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15480, {{28128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15481, {{28129, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15482, {{28130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15483, {{28131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15484, {{28132, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15485, {{28133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15486, {{28134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15487, {{28135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15488, {{28136, 1, 0, 0},{28136, 1, 1, 17}, {0, 0, 0, 0}}},{15488, {{28136, 1, 0, 0},{28136, 1, 1, 17}, {0, 0, 0, 0}}},{15489, {{28136, 1, 0, 0},{28136, 1, 1, 18}, {0, 0, 0, 0}}},{15489, {{28136, 1, 0, 0},{28136, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15490, {{28137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15491, {{28138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15492, {{28139, 1, 0, 0},{28139, 1, 1, 19}, {0, 0, 0, 0}}},{15492, {{28139, 1, 0, 0},{28139, 1, 1, 19}, {0, 0, 0, 0}}},{15493, {{28139, 1, 0, 0},{28139, 1, 1, 18}, {0, 0, 0, 0}}},{15493, {{28139, 1, 0, 0},{28139, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15494, {{28140, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15495, {{28141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15496, {{28142, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15497, {{28143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15498, {{28144, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15499, {{28145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15500, {{28146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15501, {{28147, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15503, {{28148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{15504, {{28149, 1, 0, 0},{28149, 1, 1, 17}, {0, 0, 0, 0}}},{15504, {{28149, 1, 0, 0},{28149, 1, 1, 17}, {0, 0, 0, 0}}},{59782, {{28149, 1, 0, 0},{28149, 1, 1, 19}, {0, 0, 0, 0}}},{59782, {{28149, 1, 0, 0},{28149, 1, 1, 19}, {0, 0, 0, 0}}},{59783, {{28149, 1, 0, 0},{28149, 1, 1, 20}, {0, 0, 0, 0}}},{59783, {{28149, 1, 0, 0},{28149, 1, 1, 20}, {0, 0, 0, 0}}},{59791, {{28149, 1, 0, 0},{28149, 1, 1, 18}, {0, 0, 0, 0}}},{59791, {{28149, 1, 0, 0},{28149, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15505, {{28150, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15506, {{28151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15507, {{28152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15508, {{28153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15510, {{28154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15511, {{28155, 1, 0, 0},{28155, 1, 1, 18}, {0, 0, 0, 0}}},{15511, {{28155, 1, 0, 0},{28155, 1, 1, 18}, {0, 0, 0, 0}}},{15512, {{28155, 1, 0, 0},{28155, 1, 1, 19}, {0, 0, 0, 0}}},{15512, {{28155, 1, 0, 0},{28155, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{15513, {{28156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15514, {{28157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15515, {{28158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15516, {{28159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15517, {{28160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15518, {{28161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15519, {{28162, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15520, {{28163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15522, {{28164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15523, {{28165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15524, {{28166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15525, {{28167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15526, {{28168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15527, {{28169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15528, {{28170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15529, {{28171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{15531, {{28178, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15532, {{28179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{15534, {{28181, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{15535, {{28183, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15536, {{28184, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15537, {{28185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{15538, {{28186, 1, 0, 0},{28186, 1, 1, 18}, {0, 0, 0, 0}}},{15538, {{28186, 1, 0, 0},{28186, 1, 1, 18}, {0, 0, 0, 0}}},{30267, {{28186, 1, 0, 0},{28186, 1, 1, 0},{28186, 1, 1, 19},{64070, 1, 0, 0}, {0, 0, 0, 0}}},{30267, {{28186, 1, 0, 0},{28186, 1, 1, 0},{28186, 1, 1, 19},{64070, 1, 0, 0}, {0, 0, 0, 0}}},{30267, {{28186, 1, 0, 0},{28186, 1, 1, 0},{28186, 1, 1, 19},{64070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15539, {{28187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15540, {{28188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15541, {{28189, 1, 0, 0},{28189, 1, 1, 17}, {0, 0, 0, 0}}},{15541, {{28189, 1, 0, 0},{28189, 1, 1, 17}, {0, 0, 0, 0}}},{15542, {{28189, 1, 0, 0},{28189, 1, 1, 18}, {0, 0, 0, 0}}},{15542, {{28189, 1, 0, 0},{28189, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15543, {{28190, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15544, {{28191, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15545, {{28192, 1, 0, 0},{28192, 1, 1, 17}, {0, 0, 0, 0}}},{15545, {{28192, 1, 0, 0},{28192, 1, 1, 17}, {0, 0, 0, 0}}},{15546, {{28192, 1, 0, 0},{28192, 1, 1, 18}, {0, 0, 0, 0}}},{15546, {{28192, 1, 0, 0},{28192, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15547, {{28193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15548, {{28194, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15549, {{28195, 1, 0, 0},{28195, 1, 1, 19}, {0, 0, 0, 0}}},{15549, {{28195, 1, 0, 0},{28195, 1, 1, 19}, {0, 0, 0, 0}}},{15550, {{28195, 1, 0, 0},{28195, 1, 1, 20}, {0, 0, 0, 0}}},{15550, {{28195, 1, 0, 0},{28195, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{15551, {{28196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15552, {{28197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15553, {{28198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15554, {{28199, 1, 0, 0},{28199, 1, 1, 17}, {0, 0, 0, 0}}},{15554, {{28199, 1, 0, 0},{28199, 1, 1, 17}, {0, 0, 0, 0}}},{15555, {{28199, 1, 0, 0},{28199, 1, 1, 18}, {0, 0, 0, 0}}},{15555, {{28199, 1, 0, 0},{28199, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15556, {{28200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15557, {{28201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15558, {{28202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15559, {{28203, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15560, {{28204, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15561, {{28205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15562, {{28206, 1, 0, 0},{28206, 1, 1, 17}, {0, 0, 0, 0}}},{15562, {{28206, 1, 0, 0},{28206, 1, 1, 17}, {0, 0, 0, 0}}},{15563, {{28206, 1, 0, 0},{28206, 1, 1, 18}, {0, 0, 0, 0}}},{15563, {{28206, 1, 0, 0},{28206, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{15564, {{28207, 1, 0, 0},{28207, 1, 1, 19}, {0, 0, 0, 0}}},{15564, {{28207, 1, 0, 0},{28207, 1, 1, 19}, {0, 0, 0, 0}}},{15565, {{28207, 1, 0, 0},{28207, 1, 1, 18}, {0, 0, 0, 0}}},{15565, {{28207, 1, 0, 0},{28207, 1, 1, 18}, {0, 0, 0, 0}}},{15566, {{28207, 1, 0, 0},{28207, 1, 1, 20}, {0, 0, 0, 0}}},{15566, {{28207, 1, 0, 0},{28207, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{15567, {{28208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15568, {{28209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15569, {{28210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15571, {{28211, 1, 0, 0},{28211, 1, 1, 16}, {0, 0, 0, 0}}},{15571, {{28211, 1, 0, 0},{28211, 1, 1, 16}, {0, 0, 0, 0}}},{15572, {{28211, 1, 0, 0},{28211, 1, 1, 17}, {0, 0, 0, 0}}},{15572, {{28211, 1, 0, 0},{28211, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{15573, {{28212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15574, {{28213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15575, {{28214, 1, 0, 0},{28214, 1, 1, 17}, {0, 0, 0, 0}}},{15575, {{28214, 1, 0, 0},{28214, 1, 1, 17}, {0, 0, 0, 0}}},{15576, {{28214, 1, 0, 0},{28214, 1, 1, 18}, {0, 0, 0, 0}}},{15576, {{28214, 1, 0, 0},{28214, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15577, {{28215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15578, {{28216, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15579, {{28217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15580, {{28218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15581, {{28219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15582, {{28220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15583, {{28221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15584, {{28222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15585, {{28223, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15586, {{28224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15587, {{28225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15588, {{28226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15589, {{28227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15590, {{28228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15591, {{28229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15592, {{28230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15593, {{28231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15594, {{28232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15595, {{28233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15596, {{28234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15597, {{28235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15599, {{28236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15600, {{28237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15601, {{28238, 1, 0, 0},{28238, 1, 1, 17}, {0, 0, 0, 0}}},{15601, {{28238, 1, 0, 0},{28238, 1, 1, 17}, {0, 0, 0, 0}}},{57956, {{28238, 1, 0, 0},{28238, 1, 1, 18}, {0, 0, 0, 0}}},{57956, {{28238, 1, 0, 0},{28238, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15602, {{28239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15603, {{28240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15604, {{28241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15605, {{28242, 1, 0, 0},{28242, 1, 1, 17}, {0, 0, 0, 0}}},{15605, {{28242, 1, 0, 0},{28242, 1, 1, 17}, {0, 0, 0, 0}}},{15606, {{28242, 1, 0, 0},{28242, 1, 1, 18}, {0, 0, 0, 0}}},{15606, {{28242, 1, 0, 0},{28242, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15607, {{28243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15608, {{28244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15609, {{28245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15610, {{28246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15611, {{28247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15612, {{28248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15613, {{28249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15614, {{28250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{15615, {{28251, 1, 0, 0},{28251, 1, 1, 17}, {0, 0, 0, 0}}},{15615, {{28251, 1, 0, 0},{28251, 1, 1, 17}, {0, 0, 0, 0}}},{15616, {{28251, 1, 0, 0},{28251, 1, 1, 18}, {0, 0, 0, 0}}},{15616, {{28251, 1, 0, 0},{28251, 1, 1, 18}, {0, 0, 0, 0}}},{59792, {{28251, 1, 0, 0},{28251, 1, 1, 19}, {0, 0, 0, 0}}},{59792, {{28251, 1, 0, 0},{28251, 1, 1, 19}, {0, 0, 0, 0}}},{60382, {{28251, 1, 0, 0},{28251, 1, 1, 20}, {0, 0, 0, 0}}},{60382, {{28251, 1, 0, 0},{28251, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{15617, {{28252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15618, {{28253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15619, {{28254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15620, {{28255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15621, {{28256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15623, {{28257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15625, {{28258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15626, {{28259, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15627, {{28260, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15628, {{28261, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15629, {{28262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15630, {{28263, 1, 0, 0},{28263, 1, 1, 17}, {0, 0, 0, 0}}},{15630, {{28263, 1, 0, 0},{28263, 1, 1, 17}, {0, 0, 0, 0}}},{15631, {{28263, 1, 0, 0},{28263, 1, 1, 18}, {0, 0, 0, 0}}},{15631, {{28263, 1, 0, 0},{28263, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15632, {{28264, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15633, {{28265, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15634, {{28266, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15635, {{28267, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15636, {{28268, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15637, {{28269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{15638, {{28270, 1, 0, 0},{28270, 1, 1, 19}, {0, 0, 0, 0}}},{15638, {{28270, 1, 0, 0},{28270, 1, 1, 19}, {0, 0, 0, 0}}},{15639, {{28270, 1, 0, 0},{28270, 1, 1, 18}, {0, 0, 0, 0}}},{15639, {{28270, 1, 0, 0},{28270, 1, 1, 18}, {0, 0, 0, 0}}},{15640, {{28270, 1, 0, 0},{28270, 1, 1, 20}, {0, 0, 0, 0}}},{15640, {{28270, 1, 0, 0},{28270, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{15641, {{28271, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15643, {{28272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15644, {{28273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15646, {{28274, 1, 0, 0},{28274, 1, 1, 18}, {0, 0, 0, 0}}},{15646, {{28274, 1, 0, 0},{28274, 1, 1, 18}, {0, 0, 0, 0}}},{15647, {{28274, 1, 0, 0},{28274, 1, 1, 19}, {0, 0, 0, 0}}},{15647, {{28274, 1, 0, 0},{28274, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{15648, {{28275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15649, {{28276, 1, 0, 0},{28276, 1, 1, 16}, {0, 0, 0, 0}}},{15649, {{28276, 1, 0, 0},{28276, 1, 1, 16}, {0, 0, 0, 0}}},{15650, {{28276, 1, 0, 0},{28276, 1, 1, 17}, {0, 0, 0, 0}}},{15650, {{28276, 1, 0, 0},{28276, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{15651, {{28277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15652, {{28278, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15653, {{28279, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15654, {{28280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{15655, {{28282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15656, {{28283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15657, {{28284, 1, 0, 0},{28284, 1, 1, 16}, {0, 0, 0, 0}}},{15657, {{28284, 1, 0, 0},{28284, 1, 1, 16}, {0, 0, 0, 0}}},{15658, {{28284, 1, 0, 0},{28284, 1, 1, 17}, {0, 0, 0, 0}}},{15658, {{28284, 1, 0, 0},{28284, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{15659, {{28285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15660, {{28286, 1, 0, 0},{28286, 1, 1, 18}, {0, 0, 0, 0}}},{15660, {{28286, 1, 0, 0},{28286, 1, 1, 18}, {0, 0, 0, 0}}},{15661, {{28286, 1, 0, 0},{28286, 1, 1, 19}, {0, 0, 0, 0}}},{15661, {{28286, 1, 0, 0},{28286, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{15662, {{28287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15663, {{28288, 1, 0, 0},{28288, 1, 1, 17}, {0, 0, 0, 0}}},{15663, {{28288, 1, 0, 0},{28288, 1, 1, 17}, {0, 0, 0, 0}}},{15664, {{28288, 1, 0, 0},{28288, 1, 1, 18}, {0, 0, 0, 0}}},{15664, {{28288, 1, 0, 0},{28288, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{15665, {{28290, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{15666, {{28297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{15667, {{28300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15668, {{28301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15669, {{28302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15670, {{28303, 1, 0, 0},{28303, 1, 1, 17}, {0, 0, 0, 0}}},{15670, {{28303, 1, 0, 0},{28303, 1, 1, 17}, {0, 0, 0, 0}}},{15671, {{28303, 1, 0, 0},{28303, 1, 1, 18}, {0, 0, 0, 0}}},{15671, {{28303, 1, 0, 0},{28303, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{15672, {{28304, 1, 0, 0},{28304, 1, 1, 17}, {0, 0, 0, 0}}},{15672, {{28304, 1, 0, 0},{28304, 1, 1, 17}, {0, 0, 0, 0}}},{59789, {{28304, 1, 0, 0},{28304, 1, 1, 18}, {0, 0, 0, 0}}},{59789, {{28304, 1, 0, 0},{28304, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15673, {{28305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15674, {{28306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15675, {{28307, 1, 0, 0},{28307, 1, 1, 17}, {0, 0, 0, 0}}},{15675, {{28307, 1, 0, 0},{28307, 1, 1, 17}, {0, 0, 0, 0}}},{15676, {{28307, 1, 0, 0},{28307, 1, 1, 18}, {0, 0, 0, 0}}},{15676, {{28307, 1, 0, 0},{28307, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15677, {{28308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15678, {{28309, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15679, {{28310, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15680, {{28311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15681, {{28312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15682, {{28313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15683, {{28314, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15684, {{28315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15685, {{28316, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15686, {{28317, 1, 0, 0},{28317, 1, 1, 19}, {0, 0, 0, 0}}},{15686, {{28317, 1, 0, 0},{28317, 1, 1, 19}, {0, 0, 0, 0}}},{15687, {{28317, 1, 0, 0},{28317, 1, 1, 18}, {0, 0, 0, 0}}},{15687, {{28317, 1, 0, 0},{28317, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15688, {{28318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15689, {{28319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15690, {{28320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15691, {{28321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{15692, {{28322, 1, 0, 0},{28322, 1, 1, 20}, {0, 0, 0, 0}}},{15692, {{28322, 1, 0, 0},{28322, 1, 1, 20}, {0, 0, 0, 0}}},{15693, {{28322, 1, 0, 0},{28322, 1, 1, 19}, {0, 0, 0, 0}}},{15693, {{28322, 1, 0, 0},{28322, 1, 1, 19}, {0, 0, 0, 0}}},{15694, {{28322, 1, 0, 0},{28322, 1, 1, 21}, {0, 0, 0, 0}}},{15694, {{28322, 1, 0, 0},{28322, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{15695, {{28323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15696, {{28324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15697, {{28325, 1, 0, 0},{28325, 1, 1, 17}, {0, 0, 0, 0}}},{15697, {{28325, 1, 0, 0},{28325, 1, 1, 17}, {0, 0, 0, 0}}},{15698, {{28325, 1, 0, 0},{28325, 1, 1, 18}, {0, 0, 0, 0}}},{15698, {{28325, 1, 0, 0},{28325, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15699, {{28326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15700, {{28327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15701, {{28328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15702, {{28329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15703, {{28330, 1, 0, 0},{28330, 1, 1, 17}, {0, 0, 0, 0}}},{15703, {{28330, 1, 0, 0},{28330, 1, 1, 17}, {0, 0, 0, 0}}},{15704, {{28330, 1, 0, 0},{28330, 1, 1, 18}, {0, 0, 0, 0}}},{15704, {{28330, 1, 0, 0},{28330, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15705, {{28331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15706, {{28332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15707, {{28333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15708, {{28334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15709, {{28335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15710, {{28336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15711, {{28337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{15712, {{28338, 1, 0, 0},{28338, 1, 1, 18}, {0, 0, 0, 0}}},{15712, {{28338, 1, 0, 0},{28338, 1, 1, 18}, {0, 0, 0, 0}}},{15713, {{28338, 1, 0, 0},{28338, 1, 1, 17}, {0, 0, 0, 0}}},{15713, {{28338, 1, 0, 0},{28338, 1, 1, 17}, {0, 0, 0, 0}}},{15714, {{28338, 1, 0, 0},{28338, 1, 1, 19}, {0, 0, 0, 0}}},{15714, {{28338, 1, 0, 0},{28338, 1, 1, 19}, {0, 0, 0, 0}}},{15715, {{28338, 1, 0, 0},{28338, 1, 1, 20}, {0, 0, 0, 0}}},{15715, {{28338, 1, 0, 0},{28338, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{15716, {{28339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15717, {{28340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15718, {{28341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15719, {{28342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15720, {{28343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15721, {{28344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15722, {{28345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15723, {{28346, 1, 0, 0},{28346, 1, 1, 19}, {0, 0, 0, 0}}},{15723, {{28346, 1, 0, 0},{28346, 1, 1, 19}, {0, 0, 0, 0}}},{15724, {{28346, 1, 0, 0},{28346, 1, 1, 18}, {0, 0, 0, 0}}},{15724, {{28346, 1, 0, 0},{28346, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15725, {{28347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15726, {{28348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15727, {{28349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15728, {{28350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15729, {{28351, 1, 0, 0},{28351, 1, 1, 17}, {0, 0, 0, 0}}},{15729, {{28351, 1, 0, 0},{28351, 1, 1, 17}, {0, 0, 0, 0}}},{15730, {{28351, 1, 0, 0},{28351, 1, 1, 18}, {0, 0, 0, 0}}},{15730, {{28351, 1, 0, 0},{28351, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15731, {{28352, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15732, {{28353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15733, {{28354, 1, 0, 0},{28354, 1, 1, 17}, {0, 0, 0, 0}}},{15733, {{28354, 1, 0, 0},{28354, 1, 1, 17}, {0, 0, 0, 0}}},{57962, {{28354, 1, 0, 0},{28354, 1, 1, 18}, {0, 0, 0, 0}}},{57962, {{28354, 1, 0, 0},{28354, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15734, {{28355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15735, {{28356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15736, {{28357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15737, {{28358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15738, {{28359, 1, 0, 0},{28359, 1, 1, 17}, {0, 0, 0, 0}}},{15738, {{28359, 1, 0, 0},{28359, 1, 1, 17}, {0, 0, 0, 0}}},{15739, {{28359, 1, 0, 0},{28359, 1, 1, 18}, {0, 0, 0, 0}}},{15739, {{28359, 1, 0, 0},{28359, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15740, {{28360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15741, {{28361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15742, {{28362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{15743, {{28363, 1, 0, 0},{28363, 1, 1, 19}, {0, 0, 0, 0}}},{15743, {{28363, 1, 0, 0},{28363, 1, 1, 19}, {0, 0, 0, 0}}},{15744, {{28363, 1, 0, 0},{28363, 1, 1, 21}, {0, 0, 0, 0}}},{15744, {{28363, 1, 0, 0},{28363, 1, 1, 21}, {0, 0, 0, 0}}},{15745, {{28363, 1, 0, 0},{28363, 1, 1, 22}, {0, 0, 0, 0}}},{15745, {{28363, 1, 0, 0},{28363, 1, 1, 22}, {0, 0, 0, 0}}},{15746, {{28363, 1, 0, 0},{28363, 1, 1, 20}, {0, 0, 0, 0}}},{15746, {{28363, 1, 0, 0},{28363, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{15747, {{28364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15748, {{28365, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15749, {{28366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15750, {{28367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15751, {{28368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15752, {{28369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15753, {{28370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15754, {{28371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15755, {{28372, 1, 0, 0},{28372, 1, 1, 17}, {0, 0, 0, 0}}},{15755, {{28372, 1, 0, 0},{28372, 1, 1, 17}, {0, 0, 0, 0}}},{15756, {{28372, 1, 0, 0},{28372, 1, 1, 18}, {0, 0, 0, 0}}},{15756, {{28372, 1, 0, 0},{28372, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{15757, {{28373, 1, 0, 0},{28373, 1, 1, 19}, {0, 0, 0, 0}}},{15757, {{28373, 1, 0, 0},{28373, 1, 1, 19}, {0, 0, 0, 0}}},{15758, {{28373, 1, 0, 0},{28373, 1, 1, 18}, {0, 0, 0, 0}}},{15758, {{28373, 1, 0, 0},{28373, 1, 1, 18}, {0, 0, 0, 0}}},{15759, {{28373, 1, 0, 0},{28373, 1, 1, 20}, {0, 0, 0, 0}}},{15759, {{28373, 1, 0, 0},{28373, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{15761, {{28374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15762, {{28375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15763, {{28376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15764, {{28377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15765, {{28378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15766, {{28379, 1, 0, 0},{28379, 1, 1, 18}, {0, 0, 0, 0}}},{15766, {{28379, 1, 0, 0},{28379, 1, 1, 18}, {0, 0, 0, 0}}},{57957, {{28379, 1, 0, 0},{28379, 1, 1, 19}, {0, 0, 0, 0}}},{57957, {{28379, 1, 0, 0},{28379, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{15767, {{28380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15768, {{28381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15769, {{28382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{15771, {{28387, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{15773, {{28395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15774, {{28396, 1, 0, 0},{28396, 1, 1, 19}, {0, 0, 0, 0}}},{15774, {{28396, 1, 0, 0},{28396, 1, 1, 19}, {0, 0, 0, 0}}},{15775, {{28396, 1, 0, 0},{28396, 1, 1, 18}, {0, 0, 0, 0}}},{15775, {{28396, 1, 0, 0},{28396, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15776, {{28397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15777, {{28398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15778, {{28399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15779, {{28400, 1, 0, 0},{28400, 1, 1, 16}, {0, 0, 0, 0}}},{15779, {{28400, 1, 0, 0},{28400, 1, 1, 16}, {0, 0, 0, 0}}},{15780, {{28400, 1, 0, 0},{28400, 1, 1, 17}, {0, 0, 0, 0}}},{15780, {{28400, 1, 0, 0},{28400, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{15781, {{28402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15782, {{28403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15783, {{28404, 1, 0, 0},{28404, 1, 1, 17}, {0, 0, 0, 0}}},{15783, {{28404, 1, 0, 0},{28404, 1, 1, 17}, {0, 0, 0, 0}}},{15784, {{28404, 1, 0, 0},{28404, 1, 1, 18}, {0, 0, 0, 0}}},{15784, {{28404, 1, 0, 0},{28404, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15785, {{28405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15786, {{28406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15787, {{28407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15788, {{28408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15789, {{28409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15790, {{28410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15791, {{28411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15792, {{28412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15793, {{28413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15794, {{28414, 1, 0, 0},{28414, 1, 1, 18}, {0, 0, 0, 0}}},{15794, {{28414, 1, 0, 0},{28414, 1, 1, 18}, {0, 0, 0, 0}}},{15795, {{28414, 1, 0, 0},{28414, 1, 1, 19}, {0, 0, 0, 0}}},{15795, {{28414, 1, 0, 0},{28414, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {8, {{15796, {{28415, 1, 0, 0},{28415, 1, 1, 17}, {0, 0, 0, 0}}},{15796, {{28415, 1, 0, 0},{28415, 1, 1, 17}, {0, 0, 0, 0}}},{15797, {{28415, 1, 0, 0},{28415, 1, 1, 19}, {0, 0, 0, 0}}},{15797, {{28415, 1, 0, 0},{28415, 1, 1, 19}, {0, 0, 0, 0}}},{59813, {{28415, 1, 0, 0},{28415, 1, 1, 18}, {0, 0, 0, 0}}},{59813, {{28415, 1, 0, 0},{28415, 1, 1, 18}, {0, 0, 0, 0}}},{59814, {{28415, 1, 0, 0},{28415, 1, 1, 20}, {0, 0, 0, 0}}},{59814, {{28415, 1, 0, 0},{28415, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{15798, {{28416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15799, {{28417, 1, 0, 0},{28417, 1, 1, 18}, {0, 0, 0, 0}}},{15799, {{28417, 1, 0, 0},{28417, 1, 1, 18}, {0, 0, 0, 0}}},{15800, {{28417, 1, 0, 0},{28417, 1, 1, 17}, {0, 0, 0, 0}}},{15800, {{28417, 1, 0, 0},{28417, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{15801, {{28418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15802, {{28419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15803, {{28420, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15804, {{28421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15805, {{28422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15806, {{28423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15807, {{28424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15808, {{28425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15809, {{28426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15810, {{28427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15811, {{28428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15812, {{28429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15813, {{28430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15814, {{28431, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15815, {{28432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {14, {{15816, {{28433, 1, 0, 0},{28433, 1, 1, 21}, {0, 0, 0, 0}}},{15816, {{28433, 1, 0, 0},{28433, 1, 1, 21}, {0, 0, 0, 0}}},{15817, {{28433, 1, 0, 0},{28433, 1, 1, 22}, {0, 0, 0, 0}}},{15817, {{28433, 1, 0, 0},{28433, 1, 1, 22}, {0, 0, 0, 0}}},{15818, {{28433, 1, 0, 0},{28433, 1, 1, 19}, {0, 0, 0, 0}}},{15818, {{28433, 1, 0, 0},{28433, 1, 1, 19}, {0, 0, 0, 0}}},{15819, {{28433, 1, 0, 0},{28433, 1, 1, 23}, {0, 0, 0, 0}}},{15819, {{28433, 1, 0, 0},{28433, 1, 1, 23}, {0, 0, 0, 0}}},{15820, {{28433, 1, 0, 0},{28433, 1, 1, 20}, {0, 0, 0, 0}}},{15820, {{28433, 1, 0, 0},{28433, 1, 1, 20}, {0, 0, 0, 0}}},{15821, {{28433, 1, 0, 0},{28433, 1, 1, 24}, {0, 0, 0, 0}}},{15821, {{28433, 1, 0, 0},{28433, 1, 1, 24}, {0, 0, 0, 0}}},{15822, {{28433, 1, 0, 0},{28433, 1, 1, 25}, {0, 0, 0, 0}}},{15822, {{28433, 1, 0, 0},{28433, 1, 1, 25}, {0, 0, 0, 0}}}}},
    {1, {{15823, {{28434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15824, {{28435, 1, 0, 0},{28435, 1, 1, 17}, {0, 0, 0, 0}}},{15824, {{28435, 1, 0, 0},{28435, 1, 1, 17}, {0, 0, 0, 0}}},{57964, {{28435, 1, 0, 0},{28435, 1, 1, 18}, {0, 0, 0, 0}}},{57964, {{28435, 1, 0, 0},{28435, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15825, {{28436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15826, {{28437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15827, {{28438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15828, {{28439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15829, {{28440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15830, {{28441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15831, {{28442, 1, 0, 0},{28442, 1, 1, 17}, {0, 0, 0, 0}}},{15831, {{28442, 1, 0, 0},{28442, 1, 1, 17}, {0, 0, 0, 0}}},{15832, {{28442, 1, 0, 0},{28442, 1, 1, 18}, {0, 0, 0, 0}}},{15832, {{28442, 1, 0, 0},{28442, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15833, {{28443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15834, {{28444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15835, {{28445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15836, {{28446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15837, {{28447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15838, {{28448, 1, 0, 0},{28448, 1, 1, 17}, {0, 0, 0, 0}}},{15838, {{28448, 1, 0, 0},{28448, 1, 1, 17}, {0, 0, 0, 0}}},{15839, {{28448, 1, 0, 0},{28448, 1, 1, 18}, {0, 0, 0, 0}}},{15839, {{28448, 1, 0, 0},{28448, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15840, {{28449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {7, {{15841, {{28450, 1, 0, 0},{28450, 1, 1, 18}, {0, 0, 0, 0}}},{15841, {{28450, 1, 0, 0},{28450, 1, 1, 18}, {0, 0, 0, 0}}},{15842, {{28450, 1, 0, 0},{28450, 1, 1, 23}, {0, 0, 0, 0}}},{15842, {{28450, 1, 0, 0},{28450, 1, 1, 23}, {0, 0, 0, 0}}},{30268, {{28450, 1, 0, 0},{28450, 1, 1, 0},{28450, 1, 1, 19},{64071, 1, 0, 0}, {0, 0, 0, 0}}},{30268, {{28450, 1, 0, 0},{28450, 1, 1, 0},{28450, 1, 1, 19},{64071, 1, 0, 0}, {0, 0, 0, 0}}},{30268, {{28450, 1, 0, 0},{28450, 1, 1, 0},{28450, 1, 1, 19},{64071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15844, {{28451, 1, 0, 0},{28451, 1, 1, 18}, {0, 0, 0, 0}}},{15844, {{28451, 1, 0, 0},{28451, 1, 1, 18}, {0, 0, 0, 0}}},{15845, {{28451, 1, 0, 0},{28451, 1, 1, 19}, {0, 0, 0, 0}}},{15845, {{28451, 1, 0, 0},{28451, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{15846, {{28452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15847, {{28453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15848, {{28454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15849, {{28455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15850, {{28456, 1, 0, 0},{28456, 1, 1, 16}, {0, 0, 0, 0}}},{15850, {{28456, 1, 0, 0},{28456, 1, 1, 16}, {0, 0, 0, 0}}},{15851, {{28456, 1, 0, 0},{28456, 1, 1, 17}, {0, 0, 0, 0}}},{15851, {{28456, 1, 0, 0},{28456, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{15852, {{28457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15853, {{28458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15854, {{28459, 1, 0, 0},{28459, 1, 1, 18}, {0, 0, 0, 0}}},{15854, {{28459, 1, 0, 0},{28459, 1, 1, 18}, {0, 0, 0, 0}}},{15855, {{28459, 1, 0, 0},{28459, 1, 1, 17}, {0, 0, 0, 0}}},{15855, {{28459, 1, 0, 0},{28459, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{15856, {{28460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15857, {{28461, 1, 0, 0},{28461, 1, 1, 17}, {0, 0, 0, 0}}},{15857, {{28461, 1, 0, 0},{28461, 1, 1, 17}, {0, 0, 0, 0}}},{15858, {{28461, 1, 0, 0},{28461, 1, 1, 18}, {0, 0, 0, 0}}},{15858, {{28461, 1, 0, 0},{28461, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15859, {{28462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15860, {{28463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15861, {{28464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15862, {{28465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15863, {{28466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15864, {{28467, 1, 0, 0},{28467, 1, 1, 17}, {0, 0, 0, 0}}},{15864, {{28467, 1, 0, 0},{28467, 1, 1, 17}, {0, 0, 0, 0}}},{15865, {{28467, 1, 0, 0},{28467, 1, 1, 18}, {0, 0, 0, 0}}},{15865, {{28467, 1, 0, 0},{28467, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15866, {{28468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15867, {{28469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15868, {{28470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15869, {{28471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15870, {{28472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15871, {{28473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15872, {{28474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{15873, {{28475, 1, 0, 0},{28475, 1, 1, 17}, {0, 0, 0, 0}}},{15873, {{28475, 1, 0, 0},{28475, 1, 1, 17}, {0, 0, 0, 0}}},{15874, {{28475, 1, 0, 0},{28475, 1, 1, 18}, {0, 0, 0, 0}}},{15874, {{28475, 1, 0, 0},{28475, 1, 1, 18}, {0, 0, 0, 0}}},{15875, {{28475, 1, 0, 0},{28475, 1, 1, 19}, {0, 0, 0, 0}}},{15875, {{28475, 1, 0, 0},{28475, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{15876, {{28476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15877, {{28477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15879, {{28478, 1, 0, 0},{28478, 1, 1, 18}, {0, 0, 0, 0}}},{15879, {{28478, 1, 0, 0},{28478, 1, 1, 18}, {0, 0, 0, 0}}},{15880, {{28478, 1, 0, 0},{28478, 1, 1, 19}, {0, 0, 0, 0}}},{15880, {{28478, 1, 0, 0},{28478, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{15881, {{28479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15882, {{28480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15883, {{28481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15884, {{28482, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15885, {{28483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15886, {{28484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15887, {{28485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{15888, {{28494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15889, {{28495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15890, {{28496, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15891, {{28497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15892, {{28498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15893, {{28499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{15896, {{28500, 1, 0, 0},{28500, 1, 1, 20}, {0, 0, 0, 0}}},{15896, {{28500, 1, 0, 0},{28500, 1, 1, 20}, {0, 0, 0, 0}}},{15897, {{28500, 1, 0, 0},{28500, 1, 1, 18}, {0, 0, 0, 0}}},{15897, {{28500, 1, 0, 0},{28500, 1, 1, 18}, {0, 0, 0, 0}}},{15898, {{28500, 1, 0, 0},{28500, 1, 1, 19}, {0, 0, 0, 0}}},{15898, {{28500, 1, 0, 0},{28500, 1, 1, 19}, {0, 0, 0, 0}}},{15899, {{28500, 1, 0, 0},{28500, 1, 1, 21}, {0, 0, 0, 0}}},{15899, {{28500, 1, 0, 0},{28500, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{15900, {{28501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15901, {{28502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15902, {{28503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15904, {{28504, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15905, {{28505, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15906, {{28506, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {10, {{15907, {{28507, 1, 0, 0},{28507, 1, 1, 19}, {0, 0, 0, 0}}},{15907, {{28507, 1, 0, 0},{28507, 1, 1, 19}, {0, 0, 0, 0}}},{15908, {{28507, 1, 0, 0},{28507, 1, 1, 22}, {0, 0, 0, 0}}},{15908, {{28507, 1, 0, 0},{28507, 1, 1, 22}, {0, 0, 0, 0}}},{15909, {{28507, 1, 0, 0},{28507, 1, 1, 21}, {0, 0, 0, 0}}},{15909, {{28507, 1, 0, 0},{28507, 1, 1, 21}, {0, 0, 0, 0}}},{15910, {{28507, 1, 0, 0},{28507, 1, 1, 20}, {0, 0, 0, 0}}},{15910, {{28507, 1, 0, 0},{28507, 1, 1, 20}, {0, 0, 0, 0}}},{15911, {{28507, 1, 0, 0},{28507, 1, 1, 23}, {0, 0, 0, 0}}},{15911, {{28507, 1, 0, 0},{28507, 1, 1, 23}, {0, 0, 0, 0}}}}},
    {1, {{15912, {{28508, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15913, {{28509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15914, {{28510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15915, {{28511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15916, {{28512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15917, {{28513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15918, {{28514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15921, {{28515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15922, {{28516, 1, 0, 0},{28516, 1, 1, 18}, {0, 0, 0, 0}}},{15922, {{28516, 1, 0, 0},{28516, 1, 1, 18}, {0, 0, 0, 0}}},{15923, {{28516, 1, 0, 0},{28516, 1, 1, 19}, {0, 0, 0, 0}}},{15923, {{28516, 1, 0, 0},{28516, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{15925, {{28517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15926, {{28518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15927, {{28519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15928, {{28520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15929, {{28521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15930, {{28522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15931, {{28523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15933, {{28524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{15934, {{28525, 1, 0, 0},{28525, 1, 1, 18}, {0, 0, 0, 0}}},{15934, {{28525, 1, 0, 0},{28525, 1, 1, 18}, {0, 0, 0, 0}}},{15935, {{28525, 1, 0, 0},{28525, 1, 1, 17}, {0, 0, 0, 0}}},{15935, {{28525, 1, 0, 0},{28525, 1, 1, 17}, {0, 0, 0, 0}}},{59825, {{28525, 1, 0, 0},{28525, 1, 1, 19}, {0, 0, 0, 0}}},{59825, {{28525, 1, 0, 0},{28525, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{15936, {{28526, 1, 0, 0},{28526, 1, 1, 19}, {0, 0, 0, 0}}},{15936, {{28526, 1, 0, 0},{28526, 1, 1, 19}, {0, 0, 0, 0}}},{15937, {{28526, 1, 0, 0},{28526, 1, 1, 18}, {0, 0, 0, 0}}},{15937, {{28526, 1, 0, 0},{28526, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{15938, {{28527, 1, 0, 0},{28527, 1, 1, 17}, {0, 0, 0, 0}}},{15938, {{28527, 1, 0, 0},{28527, 1, 1, 17}, {0, 0, 0, 0}}},{15939, {{28527, 1, 0, 0},{28527, 1, 1, 18}, {0, 0, 0, 0}}},{15939, {{28527, 1, 0, 0},{28527, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15940, {{28528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15941, {{28529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15942, {{28530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15943, {{28531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15944, {{28532, 1, 0, 0},{28532, 1, 1, 18}, {0, 0, 0, 0}}},{15944, {{28532, 1, 0, 0},{28532, 1, 1, 18}, {0, 0, 0, 0}}},{15945, {{28532, 1, 0, 0},{28532, 1, 1, 19}, {0, 0, 0, 0}}},{15945, {{28532, 1, 0, 0},{28532, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{15946, {{28533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15947, {{28534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15948, {{28535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15949, {{28536, 1, 0, 0},{28536, 1, 1, 17}, {0, 0, 0, 0}}},{15949, {{28536, 1, 0, 0},{28536, 1, 1, 17}, {0, 0, 0, 0}}},{15950, {{28536, 1, 0, 0},{28536, 1, 1, 18}, {0, 0, 0, 0}}},{15950, {{28536, 1, 0, 0},{28536, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15951, {{28537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15952, {{28538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15953, {{28539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15954, {{28540, 1, 0, 0},{28540, 1, 1, 17}, {0, 0, 0, 0}}},{15954, {{28540, 1, 0, 0},{28540, 1, 1, 17}, {0, 0, 0, 0}}},{15955, {{28540, 1, 0, 0},{28540, 1, 1, 18}, {0, 0, 0, 0}}},{15955, {{28540, 1, 0, 0},{28540, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15956, {{28541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{15957, {{28542, 1, 0, 0},{28542, 1, 1, 19}, {0, 0, 0, 0}}},{15957, {{28542, 1, 0, 0},{28542, 1, 1, 19}, {0, 0, 0, 0}}},{15958, {{28542, 1, 0, 0},{28542, 1, 1, 17}, {0, 0, 0, 0}}},{15958, {{28542, 1, 0, 0},{28542, 1, 1, 17}, {0, 0, 0, 0}}},{15959, {{28542, 1, 0, 0},{28542, 1, 1, 18}, {0, 0, 0, 0}}},{15959, {{28542, 1, 0, 0},{28542, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{15960, {{28543, 1, 0, 0},{28543, 1, 1, 16}, {0, 0, 0, 0}}},{15960, {{28543, 1, 0, 0},{28543, 1, 1, 16}, {0, 0, 0, 0}}},{15961, {{28543, 1, 0, 0},{28543, 1, 1, 17}, {0, 0, 0, 0}}},{15961, {{28543, 1, 0, 0},{28543, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {6, {{15962, {{28544, 1, 0, 0},{28544, 1, 1, 17}, {0, 0, 0, 0}}},{15962, {{28544, 1, 0, 0},{28544, 1, 1, 17}, {0, 0, 0, 0}}},{15963, {{28544, 1, 0, 0},{28544, 1, 1, 19}, {0, 0, 0, 0}}},{15963, {{28544, 1, 0, 0},{28544, 1, 1, 19}, {0, 0, 0, 0}}},{15964, {{28544, 1, 0, 0},{28544, 1, 1, 18}, {0, 0, 0, 0}}},{15964, {{28544, 1, 0, 0},{28544, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15965, {{28545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{15966, {{28546, 1, 0, 0},{28546, 1, 1, 17}, {0, 0, 0, 0}}},{15966, {{28546, 1, 0, 0},{28546, 1, 1, 17}, {0, 0, 0, 0}}},{15967, {{28546, 1, 0, 0},{28546, 1, 1, 19}, {0, 0, 0, 0}}},{15967, {{28546, 1, 0, 0},{28546, 1, 1, 19}, {0, 0, 0, 0}}},{15968, {{28546, 1, 0, 0},{28546, 1, 1, 18}, {0, 0, 0, 0}}},{15968, {{28546, 1, 0, 0},{28546, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15969, {{28547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15970, {{28548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15971, {{28549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15972, {{28550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15973, {{28551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15974, {{28552, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15975, {{28553, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15976, {{28554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15977, {{28555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15978, {{28556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15979, {{28557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15980, {{28558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15981, {{28559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15982, {{28560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15983, {{28561, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15984, {{28562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15985, {{28563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15986, {{28564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15988, {{28565, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15989, {{28566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15990, {{28567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15991, {{28568, 1, 0, 0},{28568, 1, 1, 19}, {0, 0, 0, 0}}},{15991, {{28568, 1, 0, 0},{28568, 1, 1, 19}, {0, 0, 0, 0}}},{15992, {{28568, 1, 0, 0},{28568, 1, 1, 18}, {0, 0, 0, 0}}},{15992, {{28568, 1, 0, 0},{28568, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{15993, {{28570, 1, 0, 0},{28570, 1, 1, 17}, {0, 0, 0, 0}}},{15993, {{28570, 1, 0, 0},{28570, 1, 1, 17}, {0, 0, 0, 0}}},{15994, {{28570, 1, 0, 0},{28570, 1, 1, 19}, {0, 0, 0, 0}}},{15994, {{28570, 1, 0, 0},{28570, 1, 1, 19}, {0, 0, 0, 0}}},{60383, {{28570, 1, 0, 0},{28570, 1, 1, 18}, {0, 0, 0, 0}}},{60383, {{28570, 1, 0, 0},{28570, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{15995, {{28574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{15996, {{28575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15997, {{28576, 1, 0, 0},{28576, 1, 1, 17}, {0, 0, 0, 0}}},{15997, {{28576, 1, 0, 0},{28576, 1, 1, 17}, {0, 0, 0, 0}}},{15998, {{28576, 1, 0, 0},{28576, 1, 1, 18}, {0, 0, 0, 0}}},{15998, {{28576, 1, 0, 0},{28576, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{15999, {{28577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16000, {{28578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16001, {{28579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{16002, {{28580, 1, 0, 0},{28580, 1, 1, 17}, {0, 0, 0, 0}}},{16002, {{28580, 1, 0, 0},{28580, 1, 1, 17}, {0, 0, 0, 0}}},{16003, {{28580, 1, 0, 0},{28580, 1, 1, 19}, {0, 0, 0, 0}}},{16003, {{28580, 1, 0, 0},{28580, 1, 1, 19}, {0, 0, 0, 0}}},{16004, {{28580, 1, 0, 0},{28580, 1, 1, 18}, {0, 0, 0, 0}}},{16004, {{28580, 1, 0, 0},{28580, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16005, {{28581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16006, {{28582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16007, {{28583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16008, {{28584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16009, {{28585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16010, {{28586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{16011, {{28587, 1, 0, 0},{28587, 1, 1, 16}, {0, 0, 0, 0}}},{16011, {{28587, 1, 0, 0},{28587, 1, 1, 16}, {0, 0, 0, 0}}},{16012, {{28587, 1, 0, 0},{28587, 1, 1, 17}, {0, 0, 0, 0}}},{16012, {{28587, 1, 0, 0},{28587, 1, 1, 17}, {0, 0, 0, 0}}},{16013, {{28587, 1, 0, 0},{28587, 1, 1, 19}, {0, 0, 0, 0}}},{16013, {{28587, 1, 0, 0},{28587, 1, 1, 19}, {0, 0, 0, 0}}},{16014, {{28587, 1, 0, 0},{28587, 1, 1, 18}, {0, 0, 0, 0}}},{16014, {{28587, 1, 0, 0},{28587, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16015, {{28588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16016, {{28589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16017, {{28590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16018, {{28591, 1, 0, 0},{28591, 1, 1, 17}, {0, 0, 0, 0}}},{16018, {{28591, 1, 0, 0},{28591, 1, 1, 17}, {0, 0, 0, 0}}},{16019, {{28591, 1, 0, 0},{28591, 1, 1, 18}, {0, 0, 0, 0}}},{16019, {{28591, 1, 0, 0},{28591, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16020, {{28592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16021, {{28593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16022, {{28594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16023, {{28595, 1, 0, 0},{28595, 1, 1, 17}, {0, 0, 0, 0}}},{16023, {{28595, 1, 0, 0},{28595, 1, 1, 17}, {0, 0, 0, 0}}},{16024, {{28595, 1, 0, 0},{28595, 1, 1, 18}, {0, 0, 0, 0}}},{16024, {{28595, 1, 0, 0},{28595, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16025, {{28596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16026, {{28597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16027, {{28598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16028, {{28599, 1, 0, 0},{28599, 1, 1, 17}, {0, 0, 0, 0}}},{16028, {{28599, 1, 0, 0},{28599, 1, 1, 17}, {0, 0, 0, 0}}},{16029, {{28599, 1, 0, 0},{28599, 1, 1, 18}, {0, 0, 0, 0}}},{16029, {{28599, 1, 0, 0},{28599, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16030, {{28600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16031, {{28601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16032, {{28602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16033, {{28603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16034, {{28604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16035, {{28605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16036, {{28606, 1, 0, 0},{28606, 1, 1, 18}, {0, 0, 0, 0}}},{16036, {{28606, 1, 0, 0},{28606, 1, 1, 18}, {0, 0, 0, 0}}},{16037, {{28606, 1, 0, 0},{28606, 1, 1, 17}, {0, 0, 0, 0}}},{16037, {{28606, 1, 0, 0},{28606, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{16038, {{28607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16039, {{28608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16040, {{28609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16041, {{28610, 1, 0, 0},{28610, 1, 1, 17}, {0, 0, 0, 0}}},{16041, {{28610, 1, 0, 0},{28610, 1, 1, 17}, {0, 0, 0, 0}}},{16042, {{28610, 1, 0, 0},{28610, 1, 1, 18}, {0, 0, 0, 0}}},{16042, {{28610, 1, 0, 0},{28610, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16043, {{28611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16044, {{28612, 1, 0, 0},{28612, 1, 1, 16}, {0, 0, 0, 0}}},{16044, {{28612, 1, 0, 0},{28612, 1, 1, 16}, {0, 0, 0, 0}}},{16045, {{28612, 1, 0, 0},{28612, 1, 1, 17}, {0, 0, 0, 0}}},{16045, {{28612, 1, 0, 0},{28612, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{16046, {{28613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16047, {{28614, 1, 0, 0},{28614, 1, 1, 17}, {0, 0, 0, 0}}},{16047, {{28614, 1, 0, 0},{28614, 1, 1, 17}, {0, 0, 0, 0}}},{16048, {{28614, 1, 0, 0},{28614, 1, 1, 18}, {0, 0, 0, 0}}},{16048, {{28614, 1, 0, 0},{28614, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16049, {{28615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16050, {{28616, 1, 0, 0},{28616, 1, 1, 17}, {0, 0, 0, 0}}},{16050, {{28616, 1, 0, 0},{28616, 1, 1, 17}, {0, 0, 0, 0}}},{16051, {{28616, 1, 0, 0},{28616, 1, 1, 18}, {0, 0, 0, 0}}},{16051, {{28616, 1, 0, 0},{28616, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16052, {{28617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16053, {{28618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16054, {{28619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16055, {{28620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16056, {{28621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16057, {{28622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16058, {{28623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16059, {{28624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{16060, {{28628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16061, {{28629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16062, {{28630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16063, {{28631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16064, {{28632, 1, 0, 0},{28632, 1, 1, 17}, {0, 0, 0, 0}}},{16064, {{28632, 1, 0, 0},{28632, 1, 1, 17}, {0, 0, 0, 0}}},{16065, {{28632, 1, 0, 0},{28632, 1, 1, 18}, {0, 0, 0, 0}}},{16065, {{28632, 1, 0, 0},{28632, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16066, {{28633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16067, {{28634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{16068, {{28635, 1, 0, 0},{28635, 1, 1, 17}, {0, 0, 0, 0}}},{16068, {{28635, 1, 0, 0},{28635, 1, 1, 17}, {0, 0, 0, 0}}},{16069, {{28635, 1, 0, 0},{28635, 1, 1, 18}, {0, 0, 0, 0}}},{16069, {{28635, 1, 0, 0},{28635, 1, 1, 18}, {0, 0, 0, 0}}},{16070, {{28635, 1, 0, 0},{28635, 1, 1, 19}, {0, 0, 0, 0}}},{16070, {{28635, 1, 0, 0},{28635, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{16071, {{28636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16072, {{28637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16073, {{28638, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16074, {{28639, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16075, {{28640, 1, 0, 0},{28640, 1, 1, 17}, {0, 0, 0, 0}}},{16075, {{28640, 1, 0, 0},{28640, 1, 1, 17}, {0, 0, 0, 0}}},{16076, {{28640, 1, 0, 0},{28640, 1, 1, 18}, {0, 0, 0, 0}}},{16076, {{28640, 1, 0, 0},{28640, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16077, {{28641, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16078, {{28642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16079, {{28643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16080, {{28644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16081, {{28645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16082, {{28646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16083, {{28647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16084, {{28648, 1, 0, 0},{28648, 1, 1, 17}, {0, 0, 0, 0}}},{16084, {{28648, 1, 0, 0},{28648, 1, 1, 17}, {0, 0, 0, 0}}},{16085, {{28648, 1, 0, 0},{28648, 1, 1, 18}, {0, 0, 0, 0}}},{16085, {{28648, 1, 0, 0},{28648, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{16087, {{28649, 1, 0, 0},{28649, 1, 1, 17}, {0, 0, 0, 0}}},{16087, {{28649, 1, 0, 0},{28649, 1, 1, 17}, {0, 0, 0, 0}}},{16088, {{28649, 1, 0, 0},{28649, 1, 1, 18}, {0, 0, 0, 0}}},{16088, {{28649, 1, 0, 0},{28649, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{16092, {{28651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16093, {{28652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16095, {{28653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16096, {{28654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{16097, {{28655, 1, 0, 0},{28655, 1, 1, 19}, {0, 0, 0, 0}}},{16097, {{28655, 1, 0, 0},{28655, 1, 1, 19}, {0, 0, 0, 0}}},{16098, {{28655, 1, 0, 0},{28655, 1, 1, 18}, {0, 0, 0, 0}}},{16098, {{28655, 1, 0, 0},{28655, 1, 1, 18}, {0, 0, 0, 0}}},{16099, {{28655, 1, 0, 0},{28655, 1, 1, 20}, {0, 0, 0, 0}}},{16099, {{28655, 1, 0, 0},{28655, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{16100, {{28656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{16101, {{28657, 1, 0, 0},{28657, 1, 1, 17}, {0, 0, 0, 0}}},{16101, {{28657, 1, 0, 0},{28657, 1, 1, 17}, {0, 0, 0, 0}}},{16102, {{28657, 1, 0, 0},{28657, 1, 1, 19}, {0, 0, 0, 0}}},{16102, {{28657, 1, 0, 0},{28657, 1, 1, 19}, {0, 0, 0, 0}}},{16103, {{28657, 1, 0, 0},{28657, 1, 1, 18}, {0, 0, 0, 0}}},{16103, {{28657, 1, 0, 0},{28657, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{16104, {{28659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16105, {{28660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16106, {{28661, 1, 0, 0},{28661, 1, 1, 17}, {0, 0, 0, 0}}},{16106, {{28661, 1, 0, 0},{28661, 1, 1, 17}, {0, 0, 0, 0}}},{59833, {{28661, 1, 0, 0},{28661, 1, 1, 18}, {0, 0, 0, 0}}},{59833, {{28661, 1, 0, 0},{28661, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16107, {{28662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16108, {{28663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{16109, {{28665, 1, 0, 0},{28665, 1, 1, 18}, {0, 0, 0, 0}}},{16109, {{28665, 1, 0, 0},{28665, 1, 1, 18}, {0, 0, 0, 0}}},{16110, {{28665, 1, 0, 0},{28665, 1, 1, 19}, {0, 0, 0, 0}}},{16110, {{28665, 1, 0, 0},{28665, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{16111, {{28666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16112, {{28667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16113, {{28668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16114, {{28669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16115, {{28670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16116, {{28671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16118, {{28672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16119, {{28673, 1, 0, 0},{28673, 1, 1, 17}, {0, 0, 0, 0}}},{16119, {{28673, 1, 0, 0},{28673, 1, 1, 17}, {0, 0, 0, 0}}},{16120, {{28673, 1, 0, 0},{28673, 1, 1, 18}, {0, 0, 0, 0}}},{16120, {{28673, 1, 0, 0},{28673, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16121, {{28674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16122, {{28675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16123, {{28676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16124, {{28677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16125, {{28678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16127, {{28679, 1, 0, 0},{28679, 1, 1, 17}, {0, 0, 0, 0}}},{16127, {{28679, 1, 0, 0},{28679, 1, 1, 17}, {0, 0, 0, 0}}},{16128, {{28679, 1, 0, 0},{28679, 1, 1, 18}, {0, 0, 0, 0}}},{16128, {{28679, 1, 0, 0},{28679, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16129, {{28680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16130, {{28681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16131, {{28682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16132, {{28683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16133, {{28684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16134, {{28685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16135, {{28686, 1, 0, 0},{28686, 1, 1, 17}, {0, 0, 0, 0}}},{16135, {{28686, 1, 0, 0},{28686, 1, 1, 17}, {0, 0, 0, 0}}},{57975, {{28686, 1, 0, 0},{28686, 1, 1, 16}, {0, 0, 0, 0}}},{57975, {{28686, 1, 0, 0},{28686, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{16136, {{28687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16137, {{28688, 1, 0, 0},{28688, 1, 1, 17}, {0, 0, 0, 0}}},{16137, {{28688, 1, 0, 0},{28688, 1, 1, 17}, {0, 0, 0, 0}}},{16138, {{28688, 1, 0, 0},{28688, 1, 1, 16}, {0, 0, 0, 0}}},{16138, {{28688, 1, 0, 0},{28688, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{16139, {{28689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16140, {{28690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16141, {{28691, 1, 0, 0},{28691, 1, 1, 16}, {0, 0, 0, 0}}},{16141, {{28691, 1, 0, 0},{28691, 1, 1, 16}, {0, 0, 0, 0}}},{57976, {{28691, 1, 0, 0},{28691, 1, 1, 17}, {0, 0, 0, 0}}},{57976, {{28691, 1, 0, 0},{28691, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{16142, {{28692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16143, {{28693, 1, 0, 0},{28693, 1, 1, 19}, {0, 0, 0, 0}}},{16143, {{28693, 1, 0, 0},{28693, 1, 1, 19}, {0, 0, 0, 0}}},{16144, {{28693, 1, 0, 0},{28693, 1, 1, 18}, {0, 0, 0, 0}}},{16144, {{28693, 1, 0, 0},{28693, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16145, {{28694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16146, {{28695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16148, {{28696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16149, {{28697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16150, {{28698, 1, 0, 0},{28698, 1, 1, 17}, {0, 0, 0, 0}}},{16150, {{28698, 1, 0, 0},{28698, 1, 1, 17}, {0, 0, 0, 0}}},{16151, {{28698, 1, 0, 0},{28698, 1, 1, 18}, {0, 0, 0, 0}}},{16151, {{28698, 1, 0, 0},{28698, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {10, {{16152, {{28699, 1, 0, 0},{28699, 1, 1, 18}, {0, 0, 0, 0}}},{16152, {{28699, 1, 0, 0},{28699, 1, 1, 18}, {0, 0, 0, 0}}},{16153, {{28699, 1, 0, 0},{28699, 1, 1, 20}, {0, 0, 0, 0}}},{16153, {{28699, 1, 0, 0},{28699, 1, 1, 20}, {0, 0, 0, 0}}},{16154, {{28699, 1, 0, 0},{28699, 1, 1, 21}, {0, 0, 0, 0}}},{16154, {{28699, 1, 0, 0},{28699, 1, 1, 21}, {0, 0, 0, 0}}},{16155, {{28699, 1, 0, 0},{28699, 1, 1, 19}, {0, 0, 0, 0}}},{16155, {{28699, 1, 0, 0},{28699, 1, 1, 19}, {0, 0, 0, 0}}},{16156, {{28699, 1, 0, 0},{28699, 1, 1, 22}, {0, 0, 0, 0}}},{16156, {{28699, 1, 0, 0},{28699, 1, 1, 22}, {0, 0, 0, 0}}}}},
    {1, {{16157, {{28700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16158, {{28701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {10, {{16159, {{28702, 1, 0, 0},{28702, 1, 1, 22}, {0, 0, 0, 0}}},{16159, {{28702, 1, 0, 0},{28702, 1, 1, 22}, {0, 0, 0, 0}}},{16160, {{28702, 1, 0, 0},{28702, 1, 1, 18}, {0, 0, 0, 0}}},{16160, {{28702, 1, 0, 0},{28702, 1, 1, 18}, {0, 0, 0, 0}}},{16161, {{28702, 1, 0, 0},{28702, 1, 1, 19}, {0, 0, 0, 0}}},{16161, {{28702, 1, 0, 0},{28702, 1, 1, 19}, {0, 0, 0, 0}}},{16162, {{28702, 1, 0, 0},{28702, 1, 1, 20}, {0, 0, 0, 0}}},{16162, {{28702, 1, 0, 0},{28702, 1, 1, 20}, {0, 0, 0, 0}}},{16163, {{28702, 1, 0, 0},{28702, 1, 1, 21}, {0, 0, 0, 0}}},{16163, {{28702, 1, 0, 0},{28702, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {4, {{16165, {{28703, 1, 0, 0},{28703, 1, 1, 18}, {0, 0, 0, 0}}},{16165, {{28703, 1, 0, 0},{28703, 1, 1, 18}, {0, 0, 0, 0}}},{16166, {{28703, 1, 0, 0},{28703, 1, 1, 17}, {0, 0, 0, 0}}},{16166, {{28703, 1, 0, 0},{28703, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{16167, {{28704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16168, {{28705, 1, 0, 0},{28705, 1, 1, 17}, {0, 0, 0, 0}}},{16168, {{28705, 1, 0, 0},{28705, 1, 1, 17}, {0, 0, 0, 0}}},{16169, {{28705, 1, 0, 0},{28705, 1, 1, 16}, {0, 0, 0, 0}}},{16169, {{28705, 1, 0, 0},{28705, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {4, {{16170, {{28706, 1, 0, 0},{28706, 1, 1, 16}, {0, 0, 0, 0}}},{16170, {{28706, 1, 0, 0},{28706, 1, 1, 16}, {0, 0, 0, 0}}},{16171, {{28706, 1, 0, 0},{28706, 1, 1, 17}, {0, 0, 0, 0}}},{16171, {{28706, 1, 0, 0},{28706, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{16172, {{28707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16173, {{28708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16174, {{28709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16175, {{28710, 1, 0, 0},{28710, 1, 1, 18}, {0, 0, 0, 0}}},{16175, {{28710, 1, 0, 0},{28710, 1, 1, 18}, {0, 0, 0, 0}}},{16176, {{28710, 1, 0, 0},{28710, 1, 1, 19}, {0, 0, 0, 0}}},{16176, {{28710, 1, 0, 0},{28710, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {12, {{16177, {{28711, 1, 0, 0},{28711, 1, 1, 18}, {0, 0, 0, 0}}},{16177, {{28711, 1, 0, 0},{28711, 1, 1, 18}, {0, 0, 0, 0}}},{16178, {{28711, 1, 0, 0},{28711, 1, 1, 22}, {0, 0, 0, 0}}},{16178, {{28711, 1, 0, 0},{28711, 1, 1, 22}, {0, 0, 0, 0}}},{16179, {{28711, 1, 0, 0},{28711, 1, 1, 20}, {0, 0, 0, 0}}},{16179, {{28711, 1, 0, 0},{28711, 1, 1, 20}, {0, 0, 0, 0}}},{16180, {{28711, 1, 0, 0},{28711, 1, 1, 19}, {0, 0, 0, 0}}},{16180, {{28711, 1, 0, 0},{28711, 1, 1, 19}, {0, 0, 0, 0}}},{16181, {{28711, 1, 0, 0},{28711, 1, 1, 21}, {0, 0, 0, 0}}},{16181, {{28711, 1, 0, 0},{28711, 1, 1, 21}, {0, 0, 0, 0}}},{16182, {{28711, 1, 0, 0},{28711, 1, 1, 23}, {0, 0, 0, 0}}},{16182, {{28711, 1, 0, 0},{28711, 1, 1, 23}, {0, 0, 0, 0}}}}},
    {1, {{16185, {{28712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16186, {{28713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16187, {{28714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16189, {{28715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16190, {{28716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{16192, {{28719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16193, {{28720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16194, {{28721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16195, {{28722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16196, {{28723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16197, {{28724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16198, {{28725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16199, {{28726, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16200, {{28727, 1, 0, 0},{28727, 1, 1, 17}, {0, 0, 0, 0}}},{16200, {{28727, 1, 0, 0},{28727, 1, 1, 17}, {0, 0, 0, 0}}},{57978, {{28727, 1, 0, 0},{28727, 1, 1, 18}, {0, 0, 0, 0}}},{57978, {{28727, 1, 0, 0},{28727, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16201, {{28728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16202, {{28729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16203, {{28730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16204, {{28731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16205, {{28732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16206, {{28733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16207, {{28734, 1, 0, 0},{28734, 1, 1, 17}, {0, 0, 0, 0}}},{16207, {{28734, 1, 0, 0},{28734, 1, 1, 17}, {0, 0, 0, 0}}},{16208, {{28734, 1, 0, 0},{28734, 1, 1, 18}, {0, 0, 0, 0}}},{16208, {{28734, 1, 0, 0},{28734, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16209, {{28735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16210, {{28736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16211, {{28737, 1, 0, 0},{28737, 1, 1, 16}, {0, 0, 0, 0}}},{16211, {{28737, 1, 0, 0},{28737, 1, 1, 16}, {0, 0, 0, 0}}},{16212, {{28737, 1, 0, 0},{28737, 1, 1, 17}, {0, 0, 0, 0}}},{16212, {{28737, 1, 0, 0},{28737, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{16213, {{28738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16214, {{28739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16215, {{28740, 1, 0, 0},{28740, 1, 1, 17}, {0, 0, 0, 0}}},{16215, {{28740, 1, 0, 0},{28740, 1, 1, 17}, {0, 0, 0, 0}}},{16216, {{28740, 1, 0, 0},{28740, 1, 1, 18}, {0, 0, 0, 0}}},{16216, {{28740, 1, 0, 0},{28740, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16217, {{28741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16218, {{28742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16219, {{28743, 1, 0, 0},{28743, 1, 1, 17}, {0, 0, 0, 0}}},{16219, {{28743, 1, 0, 0},{28743, 1, 1, 17}, {0, 0, 0, 0}}},{16220, {{28743, 1, 0, 0},{28743, 1, 1, 18}, {0, 0, 0, 0}}},{16220, {{28743, 1, 0, 0},{28743, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16221, {{28744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16222, {{28745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16223, {{28746, 1, 0, 0},{28746, 1, 1, 18}, {0, 0, 0, 0}}},{16223, {{28746, 1, 0, 0},{28746, 1, 1, 18}, {0, 0, 0, 0}}},{16224, {{28746, 1, 0, 0},{28746, 1, 1, 19}, {0, 0, 0, 0}}},{16224, {{28746, 1, 0, 0},{28746, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{16225, {{28747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16226, {{28748, 1, 0, 0},{28748, 1, 1, 17}, {0, 0, 0, 0}}},{16226, {{28748, 1, 0, 0},{28748, 1, 1, 17}, {0, 0, 0, 0}}},{16227, {{28748, 1, 0, 0},{28748, 1, 1, 18}, {0, 0, 0, 0}}},{16227, {{28748, 1, 0, 0},{28748, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16228, {{28749, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16229, {{28750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{68038, {{28752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16230, {{28753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16231, {{28754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16232, {{28755, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16233, {{28756, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16234, {{28757, 1, 0, 0},{28757, 1, 1, 17}, {0, 0, 0, 0}}},{16234, {{28757, 1, 0, 0},{28757, 1, 1, 17}, {0, 0, 0, 0}}},{57980, {{28757, 1, 0, 0},{28757, 1, 1, 18}, {0, 0, 0, 0}}},{57980, {{28757, 1, 0, 0},{28757, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16235, {{28758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16236, {{28759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16237, {{28760, 1, 0, 0},{28760, 1, 1, 18}, {0, 0, 0, 0}}},{16237, {{28760, 1, 0, 0},{28760, 1, 1, 18}, {0, 0, 0, 0}}},{16238, {{28760, 1, 0, 0},{28760, 1, 1, 19}, {0, 0, 0, 0}}},{16238, {{28760, 1, 0, 0},{28760, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{16240, {{28761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16241, {{28762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16242, {{28763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{16243, {{28765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16244, {{28766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16245, {{28767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16246, {{28768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16247, {{28769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16250, {{28770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16251, {{28771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16252, {{28772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16253, {{28773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16254, {{28774, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16255, {{28775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16256, {{28776, 1, 0, 0},{28776, 1, 1, 16}, {0, 0, 0, 0}}},{16256, {{28776, 1, 0, 0},{28776, 1, 1, 16}, {0, 0, 0, 0}}},{16257, {{28776, 1, 0, 0},{28776, 1, 1, 17}, {0, 0, 0, 0}}},{16257, {{28776, 1, 0, 0},{28776, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{16258, {{28777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16259, {{28778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16260, {{28779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16261, {{28780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{16262, {{28782, 1, 0, 0},{28782, 1, 1, 17}, {0, 0, 0, 0}}},{16262, {{28782, 1, 0, 0},{28782, 1, 1, 17}, {0, 0, 0, 0}}},{16263, {{28782, 1, 0, 0},{28782, 1, 1, 19}, {0, 0, 0, 0}}},{16263, {{28782, 1, 0, 0},{28782, 1, 1, 19}, {0, 0, 0, 0}}},{16264, {{28782, 1, 0, 0},{28782, 1, 1, 18}, {0, 0, 0, 0}}},{16264, {{28782, 1, 0, 0},{28782, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16265, {{28783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16266, {{28784, 1, 0, 0},{28784, 1, 1, 19}, {0, 0, 0, 0}}},{16266, {{28784, 1, 0, 0},{28784, 1, 1, 19}, {0, 0, 0, 0}}},{16267, {{28784, 1, 0, 0},{28784, 1, 1, 18}, {0, 0, 0, 0}}},{16267, {{28784, 1, 0, 0},{28784, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16268, {{28785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16269, {{28786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16270, {{28787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16271, {{28788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16272, {{28789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16274, {{28790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16275, {{28791, 1, 0, 0},{28791, 1, 1, 16}, {0, 0, 0, 0}}},{16275, {{28791, 1, 0, 0},{28791, 1, 1, 16}, {0, 0, 0, 0}}},{16276, {{28791, 1, 0, 0},{28791, 1, 1, 17}, {0, 0, 0, 0}}},{16276, {{28791, 1, 0, 0},{28791, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{16277, {{28792, 1, 0, 0},{28792, 1, 1, 18}, {0, 0, 0, 0}}},{16277, {{28792, 1, 0, 0},{28792, 1, 1, 18}, {0, 0, 0, 0}}},{16278, {{28792, 1, 0, 0},{28792, 1, 1, 19}, {0, 0, 0, 0}}},{16278, {{28792, 1, 0, 0},{28792, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{16279, {{28793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16280, {{28794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16281, {{28795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16282, {{28796, 1, 0, 0},{28796, 1, 1, 19}, {0, 0, 0, 0}}},{16282, {{28796, 1, 0, 0},{28796, 1, 1, 19}, {0, 0, 0, 0}}},{16283, {{28796, 1, 0, 0},{28796, 1, 1, 18}, {0, 0, 0, 0}}},{16283, {{28796, 1, 0, 0},{28796, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{16284, {{28797, 1, 0, 0},{28797, 1, 1, 17}, {0, 0, 0, 0}}},{16284, {{28797, 1, 0, 0},{28797, 1, 1, 17}, {0, 0, 0, 0}}},{16285, {{28797, 1, 0, 0},{28797, 1, 1, 18}, {0, 0, 0, 0}}},{16285, {{28797, 1, 0, 0},{28797, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16286, {{28798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{16287, {{28801, 1, 0, 0},{28801, 1, 1, 18}, {0, 0, 0, 0}}},{16287, {{28801, 1, 0, 0},{28801, 1, 1, 18}, {0, 0, 0, 0}}},{16288, {{28801, 1, 0, 0},{28801, 1, 1, 17}, {0, 0, 0, 0}}},{16288, {{28801, 1, 0, 0},{28801, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{16289, {{28802, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16290, {{28803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16291, {{28804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16292, {{28805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16293, {{28806, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16294, {{28807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16295, {{28808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16296, {{28809, 1, 0, 0},{28809, 1, 1, 18}, {0, 0, 0, 0}}},{16296, {{28809, 1, 0, 0},{28809, 1, 1, 18}, {0, 0, 0, 0}}},{16297, {{28809, 1, 0, 0},{28809, 1, 1, 17}, {0, 0, 0, 0}}},{16297, {{28809, 1, 0, 0},{28809, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{16298, {{28810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16299, {{28811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16300, {{28812, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16301, {{28813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16303, {{28814, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16304, {{28815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16305, {{28816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16306, {{28817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16307, {{28818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16308, {{28819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16309, {{28820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16310, {{28821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16311, {{28822, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16312, {{28823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16313, {{28824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16315, {{28825, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16316, {{28826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16317, {{28827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{16318, {{28831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16319, {{28832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16320, {{28833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16321, {{28834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16322, {{28835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16323, {{28836, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16324, {{28837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16325, {{28838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16326, {{28839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16327, {{28840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16328, {{28841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16329, {{28842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16330, {{28843, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16331, {{28844, 1, 0, 0},{28844, 1, 1, 17}, {0, 0, 0, 0}}},{16331, {{28844, 1, 0, 0},{28844, 1, 1, 17}, {0, 0, 0, 0}}},{16332, {{28844, 1, 0, 0},{28844, 1, 1, 18}, {0, 0, 0, 0}}},{16332, {{28844, 1, 0, 0},{28844, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{16333, {{28845, 1, 0, 0},{28845, 1, 1, 18}, {0, 0, 0, 0}}},{16333, {{28845, 1, 0, 0},{28845, 1, 1, 18}, {0, 0, 0, 0}}},{16334, {{28845, 1, 0, 0},{28845, 1, 1, 19}, {0, 0, 0, 0}}},{16334, {{28845, 1, 0, 0},{28845, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{16335, {{28846, 1, 0, 0},{28846, 1, 1, 17}, {0, 0, 0, 0}}},{16335, {{28846, 1, 0, 0},{28846, 1, 1, 17}, {0, 0, 0, 0}}},{16336, {{28846, 1, 0, 0},{28846, 1, 1, 18}, {0, 0, 0, 0}}},{16336, {{28846, 1, 0, 0},{28846, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16337, {{28847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16338, {{28848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16339, {{28849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16340, {{28850, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16341, {{28851, 1, 0, 0},{28851, 1, 1, 17}, {0, 0, 0, 0}}},{16341, {{28851, 1, 0, 0},{28851, 1, 1, 17}, {0, 0, 0, 0}}},{16342, {{28851, 1, 0, 0},{28851, 1, 1, 18}, {0, 0, 0, 0}}},{16342, {{28851, 1, 0, 0},{28851, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16343, {{28852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16344, {{28853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16345, {{28854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16346, {{28855, 1, 0, 0},{28855, 1, 1, 17}, {0, 0, 0, 0}}},{16346, {{28855, 1, 0, 0},{28855, 1, 1, 17}, {0, 0, 0, 0}}},{16347, {{28855, 1, 0, 0},{28855, 1, 1, 18}, {0, 0, 0, 0}}},{16347, {{28855, 1, 0, 0},{28855, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16348, {{28856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16349, {{28857, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16350, {{28858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16351, {{28859, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{16352, {{28868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16353, {{28869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16354, {{28870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16355, {{28871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16356, {{28872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16357, {{28873, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16358, {{28874, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16359, {{28875, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16360, {{28876, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16361, {{28877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16362, {{28878, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16363, {{28879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16364, {{28880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16365, {{28881, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16366, {{28882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16367, {{28883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16368, {{28884, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16369, {{28885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16370, {{28886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16371, {{28887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16372, {{28888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16373, {{28889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16374, {{28890, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16375, {{28891, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16376, {{28892, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16377, {{28893, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16378, {{28894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16379, {{28895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16380, {{28896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16381, {{28897, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16382, {{28898, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16383, {{28899, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16385, {{28900, 1, 0, 0},{28900, 1, 1, 17}, {0, 0, 0, 0}}},{16385, {{28900, 1, 0, 0},{28900, 1, 1, 17}, {0, 0, 0, 0}}},{16386, {{28900, 1, 0, 0},{28900, 1, 1, 18}, {0, 0, 0, 0}}},{16386, {{28900, 1, 0, 0},{28900, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{16387, {{28908, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57989, {{28909, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{16388, {{28911, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16389, {{28912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16390, {{28913, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16391, {{28914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16392, {{28915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16393, {{28916, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16394, {{28917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16395, {{28918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16396, {{28919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16397, {{28920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16398, {{28921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16399, {{28922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16400, {{28923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16401, {{28924, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16402, {{28925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16403, {{28926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16404, {{28927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16405, {{28928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16406, {{28929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16407, {{28930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16408, {{28931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16409, {{28932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16410, {{28933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16411, {{28934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16412, {{28935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16413, {{28936, 1, 0, 0},{28936, 1, 1, 17}, {0, 0, 0, 0}}},{16413, {{28936, 1, 0, 0},{28936, 1, 1, 17}, {0, 0, 0, 0}}},{16414, {{28936, 1, 0, 0},{28936, 1, 1, 18}, {0, 0, 0, 0}}},{16414, {{28936, 1, 0, 0},{28936, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16416, {{28937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16417, {{28938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16418, {{28939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16419, {{28940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16420, {{28941, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16421, {{28942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16422, {{28943, 1, 0, 0},{28943, 1, 1, 17}, {0, 0, 0, 0}}},{16422, {{28943, 1, 0, 0},{28943, 1, 1, 17}, {0, 0, 0, 0}}},{59859, {{28943, 1, 0, 0},{28943, 1, 1, 18}, {0, 0, 0, 0}}},{59859, {{28943, 1, 0, 0},{28943, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{16423, {{28948, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{16424, {{28952, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16426, {{28953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16427, {{28954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16428, {{28955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16429, {{28956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16430, {{28957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16431, {{28958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16432, {{28959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16433, {{28960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16434, {{28961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16435, {{28962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16437, {{28963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16438, {{28964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16439, {{28965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16440, {{28966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16441, {{28967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16442, {{28968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16443, {{28969, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16444, {{28970, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16445, {{28971, 1, 0, 0},{28971, 1, 1, 17}, {0, 0, 0, 0}}},{16445, {{28971, 1, 0, 0},{28971, 1, 1, 17}, {0, 0, 0, 0}}},{16446, {{28971, 1, 0, 0},{28971, 1, 1, 18}, {0, 0, 0, 0}}},{16446, {{28971, 1, 0, 0},{28971, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16447, {{28972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16448, {{28973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16450, {{28974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16451, {{28975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16452, {{28976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16453, {{28977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16454, {{28978, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16455, {{28979, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16456, {{28980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16457, {{28981, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16458, {{28982, 1, 0, 0},{28982, 1, 1, 17}, {0, 0, 0, 0}}},{16458, {{28982, 1, 0, 0},{28982, 1, 1, 17}, {0, 0, 0, 0}}},{57997, {{28982, 1, 0, 0},{28982, 1, 1, 18}, {0, 0, 0, 0}}},{57997, {{28982, 1, 0, 0},{28982, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16459, {{28983, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16460, {{28984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{16461, {{28986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16462, {{28987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16463, {{28988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16464, {{28989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{16465, {{28993, 1, 0, 0},{28993, 1, 1, 17}, {0, 0, 0, 0}}},{16465, {{28993, 1, 0, 0},{28993, 1, 1, 17}, {0, 0, 0, 0}}},{16466, {{28993, 1, 0, 0},{28993, 1, 1, 18}, {0, 0, 0, 0}}},{16466, {{28993, 1, 0, 0},{28993, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16467, {{28994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16468, {{28995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16469, {{28996, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16470, {{28997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16471, {{28998, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16472, {{28999, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16473, {{29000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16474, {{29001, 1, 0, 0},{29001, 1, 1, 18}, {0, 0, 0, 0}}},{16474, {{29001, 1, 0, 0},{29001, 1, 1, 18}, {0, 0, 0, 0}}},{16475, {{29001, 1, 0, 0},{29001, 1, 1, 19}, {0, 0, 0, 0}}},{16475, {{29001, 1, 0, 0},{29001, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{16476, {{29002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16477, {{29003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16478, {{29004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16479, {{29005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16480, {{29006, 1, 0, 0},{29006, 1, 1, 19}, {0, 0, 0, 0}}},{16480, {{29006, 1, 0, 0},{29006, 1, 1, 19}, {0, 0, 0, 0}}},{16481, {{29006, 1, 0, 0},{29006, 1, 1, 18}, {0, 0, 0, 0}}},{16481, {{29006, 1, 0, 0},{29006, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16482, {{29007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16483, {{29008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16485, {{29009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{16486, {{29010, 1, 0, 0},{29010, 1, 1, 19}, {0, 0, 0, 0}}},{16486, {{29010, 1, 0, 0},{29010, 1, 1, 19}, {0, 0, 0, 0}}},{16487, {{29010, 1, 0, 0},{29010, 1, 1, 18}, {0, 0, 0, 0}}},{16487, {{29010, 1, 0, 0},{29010, 1, 1, 18}, {0, 0, 0, 0}}},{16488, {{29010, 1, 0, 0},{29010, 1, 1, 20}, {0, 0, 0, 0}}},{16488, {{29010, 1, 0, 0},{29010, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{16489, {{29011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16490, {{29012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16491, {{29013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16492, {{29014, 1, 0, 0},{29014, 1, 1, 17}, {0, 0, 0, 0}}},{16492, {{29014, 1, 0, 0},{29014, 1, 1, 17}, {0, 0, 0, 0}}},{16493, {{29014, 1, 0, 0},{29014, 1, 1, 18}, {0, 0, 0, 0}}},{16493, {{29014, 1, 0, 0},{29014, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16494, {{29015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16495, {{29016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16496, {{29017, 1, 0, 0},{29017, 1, 1, 19}, {0, 0, 0, 0}}},{16496, {{29017, 1, 0, 0},{29017, 1, 1, 19}, {0, 0, 0, 0}}},{16497, {{29017, 1, 0, 0},{29017, 1, 1, 18}, {0, 0, 0, 0}}},{16497, {{29017, 1, 0, 0},{29017, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16498, {{29018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16499, {{29019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16500, {{29020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16501, {{29021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16502, {{29022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16503, {{29023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16504, {{29024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16505, {{29025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{16506, {{29026, 1, 0, 0},{29026, 1, 1, 20}, {0, 0, 0, 0}}},{16506, {{29026, 1, 0, 0},{29026, 1, 1, 20}, {0, 0, 0, 0}}},{16507, {{29026, 1, 0, 0},{29026, 1, 1, 18}, {0, 0, 0, 0}}},{16507, {{29026, 1, 0, 0},{29026, 1, 1, 18}, {0, 0, 0, 0}}},{16508, {{29026, 1, 0, 0},{29026, 1, 1, 19}, {0, 0, 0, 0}}},{16508, {{29026, 1, 0, 0},{29026, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{16509, {{29027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16510, {{29028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16511, {{29029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16512, {{29030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16513, {{29031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16514, {{29032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16515, {{29033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16516, {{29034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16517, {{29035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16518, {{29036, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16519, {{29037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{16520, {{29038, 1, 0, 0},{29038, 1, 1, 18}, {0, 0, 0, 0}}},{16520, {{29038, 1, 0, 0},{29038, 1, 1, 18}, {0, 0, 0, 0}}},{30269, {{29038, 1, 0, 0},{29038, 1, 1, 0},{29038, 1, 1, 19},{64072, 1, 0, 0}, {0, 0, 0, 0}}},{30269, {{29038, 1, 0, 0},{29038, 1, 1, 0},{29038, 1, 1, 19},{64072, 1, 0, 0}, {0, 0, 0, 0}}},{30269, {{29038, 1, 0, 0},{29038, 1, 1, 0},{29038, 1, 1, 19},{64072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16521, {{29039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16522, {{29040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16523, {{29041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16524, {{29042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16525, {{29043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16526, {{29044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{16527, {{29049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16528, {{29050, 1, 0, 0},{29050, 1, 1, 16}, {0, 0, 0, 0}}},{16528, {{29050, 1, 0, 0},{29050, 1, 1, 16}, {0, 0, 0, 0}}},{16529, {{29050, 1, 0, 0},{29050, 1, 1, 17}, {0, 0, 0, 0}}},{16529, {{29050, 1, 0, 0},{29050, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{16530, {{29051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16531, {{29052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{16532, {{29053, 1, 0, 0},{29053, 1, 1, 20}, {0, 0, 0, 0}}},{16532, {{29053, 1, 0, 0},{29053, 1, 1, 20}, {0, 0, 0, 0}}},{16533, {{29053, 1, 0, 0},{29053, 1, 1, 19}, {0, 0, 0, 0}}},{16533, {{29053, 1, 0, 0},{29053, 1, 1, 19}, {0, 0, 0, 0}}},{16534, {{29053, 1, 0, 0},{29053, 1, 1, 21}, {0, 0, 0, 0}}},{16534, {{29053, 1, 0, 0},{29053, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{16535, {{29054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16536, {{29055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16537, {{29056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16538, {{29057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16539, {{29058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16540, {{29059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16541, {{29060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16542, {{29061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16543, {{29062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16544, {{29063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16545, {{29064, 1, 0, 0},{29064, 1, 1, 17}, {0, 0, 0, 0}}},{16545, {{29064, 1, 0, 0},{29064, 1, 1, 17}, {0, 0, 0, 0}}},{16546, {{29064, 1, 0, 0},{29064, 1, 1, 18}, {0, 0, 0, 0}}},{16546, {{29064, 1, 0, 0},{29064, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16549, {{29065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16550, {{29066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16551, {{29067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16552, {{29068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16553, {{29069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16554, {{29070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16555, {{29071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16557, {{29072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16558, {{29073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16559, {{29074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16560, {{29075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16561, {{29076, 1, 0, 0},{29076, 1, 1, 18}, {0, 0, 0, 0}}},{16561, {{29076, 1, 0, 0},{29076, 1, 1, 18}, {0, 0, 0, 0}}},{16562, {{29076, 1, 0, 0},{29076, 1, 1, 19}, {0, 0, 0, 0}}},{16562, {{29076, 1, 0, 0},{29076, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{16563, {{29077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16564, {{29078, 1, 0, 0},{29078, 1, 1, 17}, {0, 0, 0, 0}}},{16564, {{29078, 1, 0, 0},{29078, 1, 1, 17}, {0, 0, 0, 0}}},{57995, {{29078, 1, 0, 0},{29078, 1, 1, 18}, {0, 0, 0, 0}}},{57995, {{29078, 1, 0, 0},{29078, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{16565, {{29081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16567, {{29082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16568, {{29083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16569, {{29084, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16571, {{29085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16572, {{29086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16573, {{29087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16574, {{29088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16575, {{29089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16576, {{29090, 1, 0, 0},{29090, 1, 1, 18}, {0, 0, 0, 0}}},{16576, {{29090, 1, 0, 0},{29090, 1, 1, 18}, {0, 0, 0, 0}}},{16577, {{29090, 1, 0, 0},{29090, 1, 1, 17}, {0, 0, 0, 0}}},{16577, {{29090, 1, 0, 0},{29090, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{16578, {{29091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16579, {{29092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16580, {{29093, 1, 0, 0},{29093, 1, 1, 16}, {0, 0, 0, 0}}},{16580, {{29093, 1, 0, 0},{29093, 1, 1, 16}, {0, 0, 0, 0}}},{16581, {{29093, 1, 0, 0},{29093, 1, 1, 17}, {0, 0, 0, 0}}},{16581, {{29093, 1, 0, 0},{29093, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{16582, {{29094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16583, {{29095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16584, {{29096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16585, {{29097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16586, {{29098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16587, {{29099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16588, {{29100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16589, {{29101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16590, {{29102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16591, {{29103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16592, {{29104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16593, {{29105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16594, {{29106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16595, {{29107, 1, 0, 0},{29107, 1, 1, 19}, {0, 0, 0, 0}}},{16595, {{29107, 1, 0, 0},{29107, 1, 1, 19}, {0, 0, 0, 0}}},{16596, {{29107, 1, 0, 0},{29107, 1, 1, 18}, {0, 0, 0, 0}}},{16596, {{29107, 1, 0, 0},{29107, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{16597, {{29110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16598, {{29111, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16599, {{29112, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16600, {{29113, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16601, {{29114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16602, {{29115, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16603, {{29116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16604, {{29117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16605, {{29118, 1, 0, 0},{29118, 1, 1, 17}, {0, 0, 0, 0}}},{16605, {{29118, 1, 0, 0},{29118, 1, 1, 17}, {0, 0, 0, 0}}},{16606, {{29118, 1, 0, 0},{29118, 1, 1, 18}, {0, 0, 0, 0}}},{16606, {{29118, 1, 0, 0},{29118, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{16607, {{29119, 1, 0, 0},{29119, 1, 1, 17}, {0, 0, 0, 0}}},{16607, {{29119, 1, 0, 0},{29119, 1, 1, 17}, {0, 0, 0, 0}}},{16608, {{29119, 1, 0, 0},{29119, 1, 1, 18}, {0, 0, 0, 0}}},{16608, {{29119, 1, 0, 0},{29119, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16609, {{29120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{16610, {{29121, 1, 0, 0},{29121, 1, 1, 18}, {0, 0, 0, 0}}},{16610, {{29121, 1, 0, 0},{29121, 1, 1, 18}, {0, 0, 0, 0}}},{16611, {{29121, 1, 0, 0},{29121, 1, 1, 17}, {0, 0, 0, 0}}},{16611, {{29121, 1, 0, 0},{29121, 1, 1, 17}, {0, 0, 0, 0}}},{16612, {{29121, 1, 0, 0},{29121, 1, 1, 20}, {0, 0, 0, 0}}},{16612, {{29121, 1, 0, 0},{29121, 1, 1, 20}, {0, 0, 0, 0}}},{16613, {{29121, 1, 0, 0},{29121, 1, 1, 19}, {0, 0, 0, 0}}},{16613, {{29121, 1, 0, 0},{29121, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{16615, {{29122, 1, 0, 0},{29122, 1, 1, 16}, {0, 0, 0, 0}}},{16615, {{29122, 1, 0, 0},{29122, 1, 1, 16}, {0, 0, 0, 0}}},{16616, {{29122, 1, 0, 0},{29122, 1, 1, 17}, {0, 0, 0, 0}}},{16616, {{29122, 1, 0, 0},{29122, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{16617, {{29123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16618, {{29124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16619, {{29125, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16620, {{29126, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16621, {{29127, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16622, {{29128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16623, {{29129, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16624, {{29130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16625, {{29131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16626, {{29132, 1, 0, 0},{29132, 1, 1, 17}, {0, 0, 0, 0}}},{16626, {{29132, 1, 0, 0},{29132, 1, 1, 17}, {0, 0, 0, 0}}},{58010, {{29132, 1, 0, 0},{29132, 1, 1, 18}, {0, 0, 0, 0}}},{58010, {{29132, 1, 0, 0},{29132, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16627, {{29133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16628, {{29134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16629, {{29135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{16630, {{29136, 1, 0, 0},{29136, 1, 1, 19}, {0, 0, 0, 0}}},{16630, {{29136, 1, 0, 0},{29136, 1, 1, 19}, {0, 0, 0, 0}}},{16631, {{29136, 1, 0, 0},{29136, 1, 1, 18}, {0, 0, 0, 0}}},{16631, {{29136, 1, 0, 0},{29136, 1, 1, 18}, {0, 0, 0, 0}}},{16632, {{29136, 1, 0, 0},{29136, 1, 1, 20}, {0, 0, 0, 0}}},{16632, {{29136, 1, 0, 0},{29136, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{16633, {{29137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16634, {{29138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16635, {{29139, 1, 0, 0},{29139, 1, 1, 17}, {0, 0, 0, 0}}},{16635, {{29139, 1, 0, 0},{29139, 1, 1, 17}, {0, 0, 0, 0}}},{57831, {{29139, 1, 0, 0},{29139, 1, 1, 18}, {0, 0, 0, 0}}},{57831, {{29139, 1, 0, 0},{29139, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16636, {{29140, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16637, {{29141, 1, 0, 0},{29141, 1, 1, 17}, {0, 0, 0, 0}}},{16637, {{29141, 1, 0, 0},{29141, 1, 1, 17}, {0, 0, 0, 0}}},{58005, {{29141, 1, 0, 0},{29141, 1, 1, 18}, {0, 0, 0, 0}}},{58005, {{29141, 1, 0, 0},{29141, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16638, {{29142, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16639, {{29143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16640, {{29144, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16641, {{29145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16642, {{29146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16643, {{29147, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16644, {{29148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{16645, {{29151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16646, {{29152, 1, 0, 0},{29152, 1, 1, 17}, {0, 0, 0, 0}}},{16646, {{29152, 1, 0, 0},{29152, 1, 1, 17}, {0, 0, 0, 0}}},{16647, {{29152, 1, 0, 0},{29152, 1, 1, 18}, {0, 0, 0, 0}}},{16647, {{29152, 1, 0, 0},{29152, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16648, {{29153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16649, {{29154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16650, {{29155, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16651, {{29156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16652, {{29157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16653, {{29158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{16654, {{29159, 1, 0, 0},{29159, 1, 1, 20}, {0, 0, 0, 0}}},{16654, {{29159, 1, 0, 0},{29159, 1, 1, 20}, {0, 0, 0, 0}}},{16655, {{29159, 1, 0, 0},{29159, 1, 1, 17}, {0, 0, 0, 0}}},{16655, {{29159, 1, 0, 0},{29159, 1, 1, 17}, {0, 0, 0, 0}}},{16656, {{29159, 1, 0, 0},{29159, 1, 1, 18}, {0, 0, 0, 0}}},{16656, {{29159, 1, 0, 0},{29159, 1, 1, 18}, {0, 0, 0, 0}}},{16657, {{29159, 1, 0, 0},{29159, 1, 1, 19}, {0, 0, 0, 0}}},{16657, {{29159, 1, 0, 0},{29159, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{16658, {{29160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16659, {{29161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16660, {{29162, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16661, {{29163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16662, {{29164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16663, {{29165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16664, {{29166, 1, 0, 0},{29166, 1, 1, 19}, {0, 0, 0, 0}}},{16664, {{29166, 1, 0, 0},{29166, 1, 1, 19}, {0, 0, 0, 0}}},{16665, {{29166, 1, 0, 0},{29166, 1, 1, 17}, {0, 0, 0, 0}}},{16665, {{29166, 1, 0, 0},{29166, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{16666, {{29167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16667, {{29168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16668, {{29169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16669, {{29170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16670, {{29171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16671, {{29172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16672, {{29173, 1, 0, 0},{29173, 1, 1, 18}, {0, 0, 0, 0}}},{16672, {{29173, 1, 0, 0},{29173, 1, 1, 18}, {0, 0, 0, 0}}},{16673, {{29173, 1, 0, 0},{29173, 1, 1, 17}, {0, 0, 0, 0}}},{16673, {{29173, 1, 0, 0},{29173, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{16674, {{29176, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16675, {{29177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16676, {{29178, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16677, {{29179, 1, 0, 0},{29179, 1, 1, 17}, {0, 0, 0, 0}}},{16677, {{29179, 1, 0, 0},{29179, 1, 1, 17}, {0, 0, 0, 0}}},{16678, {{29179, 1, 0, 0},{29179, 1, 1, 18}, {0, 0, 0, 0}}},{16678, {{29179, 1, 0, 0},{29179, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16679, {{29180, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16680, {{29181, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16681, {{29182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16682, {{29183, 1, 0, 0},{29183, 1, 1, 19}, {0, 0, 0, 0}}},{16682, {{29183, 1, 0, 0},{29183, 1, 1, 19}, {0, 0, 0, 0}}},{16683, {{29183, 1, 0, 0},{29183, 1, 1, 18}, {0, 0, 0, 0}}},{16683, {{29183, 1, 0, 0},{29183, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16684, {{29184, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16685, {{29185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16686, {{29186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16687, {{29187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16688, {{29188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16689, {{29189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16690, {{29190, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16691, {{29191, 1, 0, 0},{29191, 1, 1, 17}, {0, 0, 0, 0}}},{16691, {{29191, 1, 0, 0},{29191, 1, 1, 17}, {0, 0, 0, 0}}},{16692, {{29191, 1, 0, 0},{29191, 1, 1, 18}, {0, 0, 0, 0}}},{16692, {{29191, 1, 0, 0},{29191, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16693, {{29192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16694, {{29193, 1, 0, 0},{29193, 1, 1, 17}, {0, 0, 0, 0}}},{16694, {{29193, 1, 0, 0},{29193, 1, 1, 17}, {0, 0, 0, 0}}},{16695, {{29193, 1, 0, 0},{29193, 1, 1, 18}, {0, 0, 0, 0}}},{16695, {{29193, 1, 0, 0},{29193, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16696, {{29194, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16697, {{29195, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16698, {{29196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16699, {{29197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{16700, {{29199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16701, {{29200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16702, {{29201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16703, {{29202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16704, {{29203, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16705, {{29204, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16706, {{29205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16707, {{29206, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16708, {{29207, 1, 0, 0},{29207, 1, 1, 17}, {0, 0, 0, 0}}},{16708, {{29207, 1, 0, 0},{29207, 1, 1, 17}, {0, 0, 0, 0}}},{16709, {{29207, 1, 0, 0},{29207, 1, 1, 18}, {0, 0, 0, 0}}},{16709, {{29207, 1, 0, 0},{29207, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{16710, {{29209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16711, {{29210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16712, {{29211, 1, 0, 0},{29211, 1, 1, 17}, {0, 0, 0, 0}}},{16712, {{29211, 1, 0, 0},{29211, 1, 1, 17}, {0, 0, 0, 0}}},{16713, {{29211, 1, 0, 0},{29211, 1, 1, 18}, {0, 0, 0, 0}}},{16713, {{29211, 1, 0, 0},{29211, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16714, {{29212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16715, {{29213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16716, {{29214, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16717, {{29215, 1, 0, 0},{29215, 1, 1, 18}, {0, 0, 0, 0}}},{16717, {{29215, 1, 0, 0},{29215, 1, 1, 18}, {0, 0, 0, 0}}},{16718, {{29215, 1, 0, 0},{29215, 1, 1, 17}, {0, 0, 0, 0}}},{16718, {{29215, 1, 0, 0},{29215, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{16719, {{29216, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16720, {{29217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16721, {{29218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16722, {{29219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16723, {{29220, 1, 0, 0},{29220, 1, 1, 17}, {0, 0, 0, 0}}},{16723, {{29220, 1, 0, 0},{29220, 1, 1, 17}, {0, 0, 0, 0}}},{16724, {{29220, 1, 0, 0},{29220, 1, 1, 18}, {0, 0, 0, 0}}},{16724, {{29220, 1, 0, 0},{29220, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16726, {{29221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{16727, {{29223, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{16728, {{29224, 1, 0, 0},{29224, 1, 1, 18}, {0, 0, 0, 0}}},{16728, {{29224, 1, 0, 0},{29224, 1, 1, 18}, {0, 0, 0, 0}}},{16729, {{29224, 1, 0, 0},{29224, 1, 1, 19}, {0, 0, 0, 0}}},{16729, {{29224, 1, 0, 0},{29224, 1, 1, 19}, {0, 0, 0, 0}}},{16730, {{29224, 1, 0, 0},{29224, 1, 1, 21}, {0, 0, 0, 0}}},{16730, {{29224, 1, 0, 0},{29224, 1, 1, 21}, {0, 0, 0, 0}}},{16731, {{29224, 1, 0, 0},{29224, 1, 1, 20}, {0, 0, 0, 0}}},{16731, {{29224, 1, 0, 0},{29224, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{16732, {{29225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16733, {{29226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{16734, {{29227, 1, 0, 0},{29227, 1, 1, 19}, {0, 0, 0, 0}}},{16734, {{29227, 1, 0, 0},{29227, 1, 1, 19}, {0, 0, 0, 0}}},{30270, {{29227, 1, 0, 0},{29227, 1, 1, 0},{29227, 1, 1, 18},{64073, 1, 0, 0}, {0, 0, 0, 0}}},{30270, {{29227, 1, 0, 0},{29227, 1, 1, 0},{29227, 1, 1, 18},{64073, 1, 0, 0}, {0, 0, 0, 0}}},{30270, {{29227, 1, 0, 0},{29227, 1, 1, 0},{29227, 1, 1, 18},{64073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16735, {{29228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16736, {{29229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16737, {{29230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16738, {{29231, 1, 0, 0},{29231, 1, 1, 17}, {0, 0, 0, 0}}},{16738, {{29231, 1, 0, 0},{29231, 1, 1, 17}, {0, 0, 0, 0}}},{58016, {{29231, 1, 0, 0},{29231, 1, 1, 18}, {0, 0, 0, 0}}},{58016, {{29231, 1, 0, 0},{29231, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{16740, {{29232, 1, 0, 0},{29232, 1, 1, 17}, {0, 0, 0, 0}}},{16740, {{29232, 1, 0, 0},{29232, 1, 1, 17}, {0, 0, 0, 0}}},{16741, {{29232, 1, 0, 0},{29232, 1, 1, 18}, {0, 0, 0, 0}}},{16741, {{29232, 1, 0, 0},{29232, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16742, {{29233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16743, {{29234, 1, 0, 0},{29234, 1, 1, 17}, {0, 0, 0, 0}}},{16743, {{29234, 1, 0, 0},{29234, 1, 1, 17}, {0, 0, 0, 0}}},{16744, {{29234, 1, 0, 0},{29234, 1, 1, 18}, {0, 0, 0, 0}}},{16744, {{29234, 1, 0, 0},{29234, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{16746, {{29236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16747, {{29237, 1, 0, 0},{29237, 1, 1, 19}, {0, 0, 0, 0}}},{16747, {{29237, 1, 0, 0},{29237, 1, 1, 19}, {0, 0, 0, 0}}},{16748, {{29237, 1, 0, 0},{29237, 1, 1, 20}, {0, 0, 0, 0}}},{16748, {{29237, 1, 0, 0},{29237, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{16749, {{29238, 1, 0, 0},{29238, 1, 1, 18}, {0, 0, 0, 0}}},{16749, {{29238, 1, 0, 0},{29238, 1, 1, 18}, {0, 0, 0, 0}}},{16750, {{29238, 1, 0, 0},{29238, 1, 1, 19}, {0, 0, 0, 0}}},{16750, {{29238, 1, 0, 0},{29238, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{16751, {{29240, 1, 0, 0},{29240, 1, 1, 17}, {0, 0, 0, 0}}},{16751, {{29240, 1, 0, 0},{29240, 1, 1, 17}, {0, 0, 0, 0}}},{16752, {{29240, 1, 0, 0},{29240, 1, 1, 18}, {0, 0, 0, 0}}},{16752, {{29240, 1, 0, 0},{29240, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{16753, {{29241, 1, 0, 0},{29241, 1, 1, 17}, {0, 0, 0, 0}}},{16753, {{29241, 1, 0, 0},{29241, 1, 1, 17}, {0, 0, 0, 0}}},{16754, {{29241, 1, 0, 0},{29241, 1, 1, 18}, {0, 0, 0, 0}}},{16754, {{29241, 1, 0, 0},{29241, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{16755, {{29242, 1, 0, 0},{29242, 1, 1, 18}, {0, 0, 0, 0}}},{16755, {{29242, 1, 0, 0},{29242, 1, 1, 18}, {0, 0, 0, 0}}},{16756, {{29242, 1, 0, 0},{29242, 1, 1, 19}, {0, 0, 0, 0}}},{16756, {{29242, 1, 0, 0},{29242, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{16757, {{29243, 1, 0, 0},{29243, 1, 1, 18}, {0, 0, 0, 0}}},{16757, {{29243, 1, 0, 0},{29243, 1, 1, 18}, {0, 0, 0, 0}}},{16758, {{29243, 1, 0, 0},{29243, 1, 1, 19}, {0, 0, 0, 0}}},{16758, {{29243, 1, 0, 0},{29243, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{16759, {{29244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16760, {{29245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16762, {{29246, 1, 0, 0},{29246, 1, 1, 18}, {0, 0, 0, 0}}},{16762, {{29246, 1, 0, 0},{29246, 1, 1, 18}, {0, 0, 0, 0}}},{16763, {{29246, 1, 0, 0},{29246, 1, 1, 19}, {0, 0, 0, 0}}},{16763, {{29246, 1, 0, 0},{29246, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{16765, {{29247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16766, {{29248, 1, 0, 0},{29248, 1, 1, 17}, {0, 0, 0, 0}}},{16766, {{29248, 1, 0, 0},{29248, 1, 1, 17}, {0, 0, 0, 0}}},{16767, {{29248, 1, 0, 0},{29248, 1, 1, 18}, {0, 0, 0, 0}}},{16767, {{29248, 1, 0, 0},{29248, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16768, {{29249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16769, {{29250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16770, {{29251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16771, {{29252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16772, {{29253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16773, {{29254, 1, 0, 0},{29254, 1, 1, 17}, {0, 0, 0, 0}}},{16773, {{29254, 1, 0, 0},{29254, 1, 1, 17}, {0, 0, 0, 0}}},{16774, {{29254, 1, 0, 0},{29254, 1, 1, 18}, {0, 0, 0, 0}}},{16774, {{29254, 1, 0, 0},{29254, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{16775, {{29255, 1, 0, 0},{29255, 1, 1, 17}, {0, 0, 0, 0}}},{16775, {{29255, 1, 0, 0},{29255, 1, 1, 17}, {0, 0, 0, 0}}},{16776, {{29255, 1, 0, 0},{29255, 1, 1, 18}, {0, 0, 0, 0}}},{16776, {{29255, 1, 0, 0},{29255, 1, 1, 18}, {0, 0, 0, 0}}},{16777, {{29255, 1, 0, 0},{29255, 1, 1, 19}, {0, 0, 0, 0}}},{16777, {{29255, 1, 0, 0},{29255, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{16779, {{29256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16780, {{29257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16781, {{29258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16782, {{29259, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16783, {{29260, 1, 0, 0},{29260, 1, 1, 19}, {0, 0, 0, 0}}},{16783, {{29260, 1, 0, 0},{29260, 1, 1, 19}, {0, 0, 0, 0}}},{16784, {{29260, 1, 0, 0},{29260, 1, 1, 18}, {0, 0, 0, 0}}},{16784, {{29260, 1, 0, 0},{29260, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{16785, {{29262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16786, {{29263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16787, {{29264, 1, 0, 0},{29264, 1, 1, 18}, {0, 0, 0, 0}}},{16787, {{29264, 1, 0, 0},{29264, 1, 1, 18}, {0, 0, 0, 0}}},{16788, {{29264, 1, 0, 0},{29264, 1, 1, 17}, {0, 0, 0, 0}}},{16788, {{29264, 1, 0, 0},{29264, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{16789, {{29265, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16790, {{29266, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16791, {{29267, 1, 0, 0},{29267, 1, 1, 17}, {0, 0, 0, 0}}},{16791, {{29267, 1, 0, 0},{29267, 1, 1, 17}, {0, 0, 0, 0}}},{16792, {{29267, 1, 0, 0},{29267, 1, 1, 18}, {0, 0, 0, 0}}},{16792, {{29267, 1, 0, 0},{29267, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16793, {{29268, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16794, {{29269, 1, 0, 0},{29269, 1, 1, 17}, {0, 0, 0, 0}}},{16794, {{29269, 1, 0, 0},{29269, 1, 1, 17}, {0, 0, 0, 0}}},{16795, {{29269, 1, 0, 0},{29269, 1, 1, 18}, {0, 0, 0, 0}}},{16795, {{29269, 1, 0, 0},{29269, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16797, {{29270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16798, {{29271, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16800, {{29272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{16801, {{29273, 1, 0, 0},{29273, 1, 1, 19}, {0, 0, 0, 0}}},{16801, {{29273, 1, 0, 0},{29273, 1, 1, 19}, {0, 0, 0, 0}}},{16802, {{29273, 1, 0, 0},{29273, 1, 1, 18}, {0, 0, 0, 0}}},{16802, {{29273, 1, 0, 0},{29273, 1, 1, 18}, {0, 0, 0, 0}}},{16803, {{29273, 1, 0, 0},{29273, 1, 1, 20}, {0, 0, 0, 0}}},{16803, {{29273, 1, 0, 0},{29273, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{16804, {{29274, 1, 0, 0},{29274, 1, 1, 18}, {0, 0, 0, 0}}},{16804, {{29274, 1, 0, 0},{29274, 1, 1, 18}, {0, 0, 0, 0}}},{16805, {{29274, 1, 0, 0},{29274, 1, 1, 17}, {0, 0, 0, 0}}},{16805, {{29274, 1, 0, 0},{29274, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{16806, {{29275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16807, {{29276, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16808, {{29277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16809, {{29278, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16810, {{29279, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16811, {{29280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16812, {{29281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16813, {{29282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16814, {{29283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{16815, {{29285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16816, {{29286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16817, {{29287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16818, {{29288, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16819, {{29289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16820, {{29290, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16821, {{29291, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16822, {{29292, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16823, {{29293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16824, {{29294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16825, {{29295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16826, {{29296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16827, {{29297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16828, {{29298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16829, {{29299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16830, {{29300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{16831, {{29302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16832, {{29303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16833, {{29304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16834, {{29305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{16835, {{29307, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16836, {{29308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16837, {{29309, 1, 0, 0},{29309, 1, 1, 17}, {0, 0, 0, 0}}},{16837, {{29309, 1, 0, 0},{29309, 1, 1, 17}, {0, 0, 0, 0}}},{16838, {{29309, 1, 0, 0},{29309, 1, 1, 18}, {0, 0, 0, 0}}},{16838, {{29309, 1, 0, 0},{29309, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16839, {{29310, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16840, {{29311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16842, {{29312, 1, 0, 0},{29312, 1, 1, 18}, {0, 0, 0, 0}}},{16842, {{29312, 1, 0, 0},{29312, 1, 1, 18}, {0, 0, 0, 0}}},{16843, {{29312, 1, 0, 0},{29312, 1, 1, 17}, {0, 0, 0, 0}}},{16843, {{29312, 1, 0, 0},{29312, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{16844, {{29313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16845, {{29314, 1, 0, 0},{29314, 1, 1, 17}, {0, 0, 0, 0}}},{16845, {{29314, 1, 0, 0},{29314, 1, 1, 17}, {0, 0, 0, 0}}},{59884, {{29314, 1, 0, 0},{29314, 1, 1, 18}, {0, 0, 0, 0}}},{59884, {{29314, 1, 0, 0},{29314, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16846, {{29315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16847, {{29316, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16848, {{29317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16849, {{29318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16850, {{29319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16851, {{29320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16852, {{29321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{16853, {{29324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16854, {{29325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16855, {{29326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16856, {{29327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16857, {{29328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16858, {{29329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16859, {{29330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16860, {{29331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16861, {{29332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16862, {{29333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16863, {{29334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16864, {{29335, 1, 0, 0},{29335, 1, 1, 16}, {0, 0, 0, 0}}},{16864, {{29335, 1, 0, 0},{29335, 1, 1, 16}, {0, 0, 0, 0}}},{58026, {{29335, 1, 0, 0},{29335, 1, 1, 17}, {0, 0, 0, 0}}},{58026, {{29335, 1, 0, 0},{29335, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{16865, {{29336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16866, {{29337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16867, {{29338, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16868, {{29339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16869, {{29340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16870, {{29341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16871, {{29342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{16872, {{29344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16873, {{29345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16874, {{29346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16875, {{29347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16876, {{29348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16877, {{29349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16878, {{29350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16879, {{29351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16880, {{29352, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16881, {{29353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16882, {{29354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16883, {{29355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16884, {{29356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16885, {{29357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16886, {{29358, 1, 0, 0},{29358, 1, 1, 18}, {0, 0, 0, 0}}},{16886, {{29358, 1, 0, 0},{29358, 1, 1, 18}, {0, 0, 0, 0}}},{16887, {{29358, 1, 0, 0},{29358, 1, 1, 17}, {0, 0, 0, 0}}},{16887, {{29358, 1, 0, 0},{29358, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{16889, {{29359, 1, 0, 0},{29359, 1, 1, 17}, {0, 0, 0, 0}}},{16889, {{29359, 1, 0, 0},{29359, 1, 1, 17}, {0, 0, 0, 0}}},{16890, {{29359, 1, 0, 0},{29359, 1, 1, 18}, {0, 0, 0, 0}}},{16890, {{29359, 1, 0, 0},{29359, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16891, {{29360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16892, {{29361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16893, {{29362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16894, {{29363, 1, 0, 0},{29363, 1, 1, 16}, {0, 0, 0, 0}}},{16894, {{29363, 1, 0, 0},{29363, 1, 1, 16}, {0, 0, 0, 0}}},{16895, {{29363, 1, 0, 0},{29363, 1, 1, 17}, {0, 0, 0, 0}}},{16895, {{29363, 1, 0, 0},{29363, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{16897, {{29364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16898, {{29365, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16899, {{29366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{16900, {{29369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16901, {{29370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16902, {{29371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16903, {{29372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16904, {{29373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16905, {{29374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16906, {{29375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16907, {{29376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16908, {{29377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16909, {{29378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16910, {{29379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16911, {{29380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16912, {{29381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16913, {{29382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16914, {{29383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{16915, {{29385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16916, {{29386, 1, 0, 0},{29386, 1, 1, 16}, {0, 0, 0, 0}}},{16916, {{29386, 1, 0, 0},{29386, 1, 1, 16}, {0, 0, 0, 0}}},{16917, {{29386, 1, 0, 0},{29386, 1, 1, 17}, {0, 0, 0, 0}}},{16917, {{29386, 1, 0, 0},{29386, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{16918, {{29387, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16919, {{29388, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16920, {{29389, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16921, {{29390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16922, {{29391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{16923, {{29392, 1, 0, 0},{29392, 1, 1, 18}, {0, 0, 0, 0}}},{16923, {{29392, 1, 0, 0},{29392, 1, 1, 18}, {0, 0, 0, 0}}},{16924, {{29392, 1, 0, 0},{29392, 1, 1, 19}, {0, 0, 0, 0}}},{16924, {{29392, 1, 0, 0},{29392, 1, 1, 19}, {0, 0, 0, 0}}},{16925, {{29392, 1, 0, 0},{29392, 1, 1, 17}, {0, 0, 0, 0}}},{16925, {{29392, 1, 0, 0},{29392, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{16926, {{29393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16927, {{29394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16928, {{29395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16929, {{29396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16930, {{29397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16931, {{29398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16932, {{29399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16933, {{29400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16934, {{29401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16935, {{29402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16936, {{29403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16937, {{29404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{16938, {{29407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16939, {{29408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16940, {{29409, 1, 0, 0},{29409, 1, 1, 18}, {0, 0, 0, 0}}},{16940, {{29409, 1, 0, 0},{29409, 1, 1, 18}, {0, 0, 0, 0}}},{16941, {{29409, 1, 0, 0},{29409, 1, 1, 19}, {0, 0, 0, 0}}},{16941, {{29409, 1, 0, 0},{29409, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{16942, {{29410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16943, {{29411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16944, {{29412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16945, {{29413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16946, {{29414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16947, {{29415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16948, {{29416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16949, {{29417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16950, {{29418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16951, {{29419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16952, {{29420, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16953, {{29421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{16954, {{29427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16955, {{29428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16956, {{29429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16957, {{29430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16958, {{29431, 1, 0, 0},{29431, 1, 1, 17}, {0, 0, 0, 0}}},{16958, {{29431, 1, 0, 0},{29431, 1, 1, 17}, {0, 0, 0, 0}}},{58029, {{29431, 1, 0, 0},{29431, 1, 1, 18}, {0, 0, 0, 0}}},{58029, {{29431, 1, 0, 0},{29431, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{16959, {{29432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16960, {{29433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16961, {{29434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16962, {{29435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16964, {{29436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16965, {{29437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16966, {{29438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16967, {{29439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16968, {{29440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16969, {{29441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16970, {{29442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{16971, {{29444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16972, {{29445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16973, {{29446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16974, {{29447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16975, {{29448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16976, {{29449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16977, {{29450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16978, {{29451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16979, {{29452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16980, {{29453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16981, {{29454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16982, {{29455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16983, {{29456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16984, {{29457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16985, {{29458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16986, {{29459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16987, {{29460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16988, {{29461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16989, {{29462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16990, {{29463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16991, {{29464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16992, {{29465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16993, {{29466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16994, {{29467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{16995, {{29468, 1, 0, 0},{29468, 1, 1, 18}, {0, 0, 0, 0}}},{16995, {{29468, 1, 0, 0},{29468, 1, 1, 18}, {0, 0, 0, 0}}},{16996, {{29468, 1, 0, 0},{29468, 1, 1, 19}, {0, 0, 0, 0}}},{16996, {{29468, 1, 0, 0},{29468, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{16997, {{29469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16998, {{29470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{16999, {{29471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{17000, {{29474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17001, {{29475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17002, {{29476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17003, {{29477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17004, {{29478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17005, {{29479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17006, {{29480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17007, {{29481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{17008, {{29482, 1, 0, 0},{29482, 1, 1, 18}, {0, 0, 0, 0}}},{17008, {{29482, 1, 0, 0},{29482, 1, 1, 18}, {0, 0, 0, 0}}},{30208, {{29482, 1, 0, 0},{29482, 1, 1, 0},{29482, 1, 1, 19},{64022, 1, 0, 0}, {0, 0, 0, 0}}},{30208, {{29482, 1, 0, 0},{29482, 1, 1, 0},{29482, 1, 1, 19},{64022, 1, 0, 0}, {0, 0, 0, 0}}},{30208, {{29482, 1, 0, 0},{29482, 1, 1, 0},{29482, 1, 1, 19},{64022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17009, {{29483, 1, 0, 0},{29483, 1, 1, 17}, {0, 0, 0, 0}}},{17009, {{29483, 1, 0, 0},{29483, 1, 1, 17}, {0, 0, 0, 0}}},{17010, {{29483, 1, 0, 0},{29483, 1, 1, 18}, {0, 0, 0, 0}}},{17010, {{29483, 1, 0, 0},{29483, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17011, {{29484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17012, {{29485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17013, {{29486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17014, {{29487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17015, {{29488, 1, 0, 0},{29488, 1, 1, 16}, {0, 0, 0, 0}}},{17015, {{29488, 1, 0, 0},{29488, 1, 1, 16}, {0, 0, 0, 0}}},{17016, {{29488, 1, 0, 0},{29488, 1, 1, 17}, {0, 0, 0, 0}}},{17016, {{29488, 1, 0, 0},{29488, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{17017, {{29489, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17018, {{29490, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17019, {{29491, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17020, {{29492, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17021, {{29493, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{17022, {{29494, 1, 0, 0},{29494, 1, 1, 21}, {0, 0, 0, 0}}},{17022, {{29494, 1, 0, 0},{29494, 1, 1, 21}, {0, 0, 0, 0}}},{17023, {{29494, 1, 0, 0},{29494, 1, 1, 19}, {0, 0, 0, 0}}},{17023, {{29494, 1, 0, 0},{29494, 1, 1, 19}, {0, 0, 0, 0}}},{17024, {{29494, 1, 0, 0},{29494, 1, 1, 20}, {0, 0, 0, 0}}},{17024, {{29494, 1, 0, 0},{29494, 1, 1, 20}, {0, 0, 0, 0}}},{17025, {{29494, 1, 0, 0},{29494, 1, 1, 22}, {0, 0, 0, 0}}},{17025, {{29494, 1, 0, 0},{29494, 1, 1, 22}, {0, 0, 0, 0}}}}},
    {4, {{17026, {{29495, 1, 0, 0},{29495, 1, 1, 19}, {0, 0, 0, 0}}},{17026, {{29495, 1, 0, 0},{29495, 1, 1, 19}, {0, 0, 0, 0}}},{17027, {{29495, 1, 0, 0},{29495, 1, 1, 20}, {0, 0, 0, 0}}},{17027, {{29495, 1, 0, 0},{29495, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{17028, {{29496, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17029, {{29497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17030, {{29498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17031, {{29499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17032, {{29500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17033, {{29501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17034, {{29502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17035, {{29503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17036, {{29504, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17037, {{29505, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17038, {{29506, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17039, {{29507, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17040, {{29508, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17041, {{29509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17042, {{29510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{17043, {{29512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17045, {{29513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17046, {{29514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17047, {{29515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17048, {{29516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17049, {{29517, 1, 0, 0},{29517, 1, 1, 17}, {0, 0, 0, 0}}},{17049, {{29517, 1, 0, 0},{29517, 1, 1, 17}, {0, 0, 0, 0}}},{17050, {{29517, 1, 0, 0},{29517, 1, 1, 18}, {0, 0, 0, 0}}},{17050, {{29517, 1, 0, 0},{29517, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17051, {{29518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17052, {{29519, 1, 0, 0},{29519, 1, 1, 17}, {0, 0, 0, 0}}},{17052, {{29519, 1, 0, 0},{29519, 1, 1, 17}, {0, 0, 0, 0}}},{17053, {{29519, 1, 0, 0},{29519, 1, 1, 18}, {0, 0, 0, 0}}},{17053, {{29519, 1, 0, 0},{29519, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{17054, {{29520, 1, 0, 0},{29520, 1, 1, 17}, {0, 0, 0, 0}}},{17054, {{29520, 1, 0, 0},{29520, 1, 1, 17}, {0, 0, 0, 0}}},{17055, {{29520, 1, 0, 0},{29520, 1, 1, 18}, {0, 0, 0, 0}}},{17055, {{29520, 1, 0, 0},{29520, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17056, {{29521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17057, {{29522, 1, 0, 0},{29522, 1, 1, 17}, {0, 0, 0, 0}}},{17057, {{29522, 1, 0, 0},{29522, 1, 1, 17}, {0, 0, 0, 0}}},{17058, {{29522, 1, 0, 0},{29522, 1, 1, 18}, {0, 0, 0, 0}}},{17058, {{29522, 1, 0, 0},{29522, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17059, {{29523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17060, {{29524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{17061, {{29526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17062, {{29527, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17063, {{29528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17064, {{29529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17065, {{29530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17066, {{29531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17067, {{29532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17068, {{29533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17069, {{29534, 1, 0, 0},{29534, 1, 1, 17}, {0, 0, 0, 0}}},{17069, {{29534, 1, 0, 0},{29534, 1, 1, 17}, {0, 0, 0, 0}}},{17070, {{29534, 1, 0, 0},{29534, 1, 1, 18}, {0, 0, 0, 0}}},{17070, {{29534, 1, 0, 0},{29534, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17071, {{29535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17072, {{29536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17073, {{29537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17074, {{29538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17075, {{29539, 1, 0, 0},{29539, 1, 1, 18}, {0, 0, 0, 0}}},{17075, {{29539, 1, 0, 0},{29539, 1, 1, 18}, {0, 0, 0, 0}}},{17076, {{29539, 1, 0, 0},{29539, 1, 1, 17}, {0, 0, 0, 0}}},{17076, {{29539, 1, 0, 0},{29539, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{17077, {{29541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17078, {{29542, 1, 0, 0},{29542, 1, 1, 17}, {0, 0, 0, 0}}},{17078, {{29542, 1, 0, 0},{29542, 1, 1, 17}, {0, 0, 0, 0}}},{17079, {{29542, 1, 0, 0},{29542, 1, 1, 18}, {0, 0, 0, 0}}},{17079, {{29542, 1, 0, 0},{29542, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17080, {{29543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17081, {{29544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17082, {{29545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17083, {{29546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17084, {{29547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17085, {{29548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{17086, {{29550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17087, {{29551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17088, {{29552, 1, 0, 0},{29552, 1, 1, 17}, {0, 0, 0, 0}}},{17088, {{29552, 1, 0, 0},{29552, 1, 1, 17}, {0, 0, 0, 0}}},{17089, {{29552, 1, 0, 0},{29552, 1, 1, 18}, {0, 0, 0, 0}}},{17089, {{29552, 1, 0, 0},{29552, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17090, {{29553, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17091, {{29554, 1, 0, 0},{29554, 1, 1, 17}, {0, 0, 0, 0}}},{17091, {{29554, 1, 0, 0},{29554, 1, 1, 17}, {0, 0, 0, 0}}},{17092, {{29554, 1, 0, 0},{29554, 1, 1, 18}, {0, 0, 0, 0}}},{17092, {{29554, 1, 0, 0},{29554, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17093, {{29555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17094, {{29556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17095, {{29557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17096, {{29558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17097, {{29559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17098, {{29560, 1, 0, 0},{29560, 1, 1, 17}, {0, 0, 0, 0}}},{17098, {{29560, 1, 0, 0},{29560, 1, 1, 17}, {0, 0, 0, 0}}},{40909, {{29560, 1, 0, 0},{29560, 1, 1, 18}, {0, 0, 0, 0}}},{40909, {{29560, 1, 0, 0},{29560, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17099, {{29561, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17100, {{29562, 1, 0, 0},{29562, 1, 1, 17}, {0, 0, 0, 0}}},{17100, {{29562, 1, 0, 0},{29562, 1, 1, 17}, {0, 0, 0, 0}}},{17101, {{29562, 1, 0, 0},{29562, 1, 1, 18}, {0, 0, 0, 0}}},{17101, {{29562, 1, 0, 0},{29562, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17102, {{29563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17103, {{29564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17104, {{29565, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17105, {{29566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17106, {{29567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17107, {{29568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17108, {{29569, 1, 0, 0},{29569, 1, 1, 17}, {0, 0, 0, 0}}},{17108, {{29569, 1, 0, 0},{29569, 1, 1, 17}, {0, 0, 0, 0}}},{17109, {{29569, 1, 0, 0},{29569, 1, 1, 18}, {0, 0, 0, 0}}},{17109, {{29569, 1, 0, 0},{29569, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17110, {{29570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17111, {{29571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17112, {{29572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17113, {{29573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17114, {{29574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17115, {{29575, 1, 0, 0},{29575, 1, 1, 19}, {0, 0, 0, 0}}},{17115, {{29575, 1, 0, 0},{29575, 1, 1, 19}, {0, 0, 0, 0}}},{17116, {{29575, 1, 0, 0},{29575, 1, 1, 18}, {0, 0, 0, 0}}},{17116, {{29575, 1, 0, 0},{29575, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{17117, {{29576, 1, 0, 0},{29576, 1, 1, 16}, {0, 0, 0, 0}}},{17117, {{29576, 1, 0, 0},{29576, 1, 1, 16}, {0, 0, 0, 0}}},{59891, {{29576, 1, 0, 0},{29576, 1, 1, 17}, {0, 0, 0, 0}}},{59891, {{29576, 1, 0, 0},{29576, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{17118, {{29577, 1, 0, 0},{29577, 1, 1, 17}, {0, 0, 0, 0}}},{17118, {{29577, 1, 0, 0},{29577, 1, 1, 17}, {0, 0, 0, 0}}},{17121, {{29577, 1, 0, 0},{29577, 1, 1, 18}, {0, 0, 0, 0}}},{17121, {{29577, 1, 0, 0},{29577, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17119, {{29578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17120, {{29579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17122, {{29580, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17123, {{29581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17124, {{29582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17125, {{29583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17126, {{29584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{17127, {{29586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17128, {{29587, 1, 0, 0},{29587, 1, 1, 17}, {0, 0, 0, 0}}},{17128, {{29587, 1, 0, 0},{29587, 1, 1, 17}, {0, 0, 0, 0}}},{17129, {{29587, 1, 0, 0},{29587, 1, 1, 18}, {0, 0, 0, 0}}},{17129, {{29587, 1, 0, 0},{29587, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17130, {{29588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17131, {{29589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17132, {{29590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17133, {{29591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17134, {{29592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{17136, {{29596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17137, {{29597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17138, {{29598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17139, {{29599, 1, 0, 0},{29599, 1, 1, 17}, {0, 0, 0, 0}}},{17139, {{29599, 1, 0, 0},{29599, 1, 1, 17}, {0, 0, 0, 0}}},{17140, {{29599, 1, 0, 0},{29599, 1, 1, 18}, {0, 0, 0, 0}}},{17140, {{29599, 1, 0, 0},{29599, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17141, {{29600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17142, {{29601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17143, {{29602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17144, {{29603, 1, 0, 0},{29603, 1, 1, 16}, {0, 0, 0, 0}}},{17144, {{29603, 1, 0, 0},{29603, 1, 1, 16}, {0, 0, 0, 0}}},{68063, {{29603, 1, 0, 0},{29603, 1, 1, 17}, {0, 0, 0, 0}}},{68063, {{29603, 1, 0, 0},{29603, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{17145, {{29604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17146, {{29605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17148, {{29606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17149, {{29607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17150, {{29608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17151, {{29609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17152, {{29610, 1, 0, 0},{29610, 1, 1, 17}, {0, 0, 0, 0}}},{17152, {{29610, 1, 0, 0},{29610, 1, 1, 17}, {0, 0, 0, 0}}},{59894, {{29610, 1, 0, 0},{29610, 1, 1, 18}, {0, 0, 0, 0}}},{59894, {{29610, 1, 0, 0},{29610, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17154, {{29611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17155, {{29612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17156, {{29613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{17160, {{29618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17162, {{29619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17163, {{29620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17164, {{29621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17165, {{29622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17167, {{29623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17168, {{29624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17169, {{29625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17170, {{29626, 1, 0, 0},{29626, 1, 1, 17}, {0, 0, 0, 0}}},{17170, {{29626, 1, 0, 0},{29626, 1, 1, 17}, {0, 0, 0, 0}}},{17171, {{29626, 1, 0, 0},{29626, 1, 1, 18}, {0, 0, 0, 0}}},{17171, {{29626, 1, 0, 0},{29626, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17172, {{29627, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17173, {{29628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17174, {{29629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17175, {{29630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17176, {{29631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17177, {{29632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17178, {{29633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17179, {{29634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17180, {{29635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17181, {{29636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17182, {{29637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17183, {{29638, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17184, {{29639, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17185, {{29640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17186, {{29641, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{17187, {{29642, 1, 0, 0},{29642, 1, 1, 19}, {0, 0, 0, 0}}},{17187, {{29642, 1, 0, 0},{29642, 1, 1, 19}, {0, 0, 0, 0}}},{17188, {{29642, 1, 0, 0},{29642, 1, 1, 20}, {0, 0, 0, 0}}},{17188, {{29642, 1, 0, 0},{29642, 1, 1, 20}, {0, 0, 0, 0}}},{17189, {{29642, 1, 0, 0},{29642, 1, 1, 21}, {0, 0, 0, 0}}},{17189, {{29642, 1, 0, 0},{29642, 1, 1, 21}, {0, 0, 0, 0}}},{58043, {{29642, 1, 0, 0},{29642, 1, 1, 22}, {0, 0, 0, 0}}},{58043, {{29642, 1, 0, 0},{29642, 1, 1, 22}, {0, 0, 0, 0}}}}},
    {1, {{17190, {{29643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17191, {{29644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17192, {{29645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17193, {{29646, 1, 0, 0},{29646, 1, 1, 18}, {0, 0, 0, 0}}},{17193, {{29646, 1, 0, 0},{29646, 1, 1, 18}, {0, 0, 0, 0}}},{17194, {{29646, 1, 0, 0},{29646, 1, 1, 19}, {0, 0, 0, 0}}},{17194, {{29646, 1, 0, 0},{29646, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{17195, {{29647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17196, {{29648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{17197, {{29650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17198, {{29651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17199, {{29652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17200, {{29653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17201, {{29654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17202, {{29655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17203, {{29656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17204, {{29657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17205, {{29658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17206, {{29659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17207, {{29660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17208, {{29661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17210, {{29662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17211, {{29663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17212, {{29664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17213, {{29665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17214, {{29666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17215, {{29667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17216, {{29668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17217, {{29669, 1, 0, 0},{29669, 1, 1, 18}, {0, 0, 0, 0}}},{17217, {{29669, 1, 0, 0},{29669, 1, 1, 18}, {0, 0, 0, 0}}},{17218, {{29669, 1, 0, 0},{29669, 1, 1, 19}, {0, 0, 0, 0}}},{17218, {{29669, 1, 0, 0},{29669, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{17219, {{29670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17221, {{29671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17222, {{29672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17223, {{29673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17224, {{29674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17225, {{29675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17226, {{29676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17227, {{29677, 1, 0, 0},{29677, 1, 1, 18}, {0, 0, 0, 0}}},{17227, {{29677, 1, 0, 0},{29677, 1, 1, 18}, {0, 0, 0, 0}}},{17228, {{29677, 1, 0, 0},{29677, 1, 1, 19}, {0, 0, 0, 0}}},{17228, {{29677, 1, 0, 0},{29677, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{17229, {{29678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17230, {{29679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{17231, {{29681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{17232, {{29683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17234, {{29684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17235, {{29685, 1, 0, 0},{29685, 1, 1, 17}, {0, 0, 0, 0}}},{17235, {{29685, 1, 0, 0},{29685, 1, 1, 17}, {0, 0, 0, 0}}},{17236, {{29685, 1, 0, 0},{29685, 1, 1, 18}, {0, 0, 0, 0}}},{17236, {{29685, 1, 0, 0},{29685, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17237, {{29686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17238, {{29687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17239, {{29688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17240, {{29689, 1, 0, 0},{29689, 1, 1, 17}, {0, 0, 0, 0}}},{17240, {{29689, 1, 0, 0},{29689, 1, 1, 17}, {0, 0, 0, 0}}},{17241, {{29689, 1, 0, 0},{29689, 1, 1, 18}, {0, 0, 0, 0}}},{17241, {{29689, 1, 0, 0},{29689, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17242, {{29690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17243, {{29691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17245, {{29692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17246, {{29693, 1, 0, 0},{29693, 1, 1, 17}, {0, 0, 0, 0}}},{17246, {{29693, 1, 0, 0},{29693, 1, 1, 17}, {0, 0, 0, 0}}},{17247, {{29693, 1, 0, 0},{29693, 1, 1, 18}, {0, 0, 0, 0}}},{17247, {{29693, 1, 0, 0},{29693, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17248, {{29694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17249, {{29695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17250, {{29696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17251, {{29697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17252, {{29698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17253, {{29699, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17254, {{29700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17255, {{29701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17256, {{29702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17257, {{29703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17258, {{29704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17260, {{29705, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17261, {{29706, 1, 0, 0},{29706, 1, 1, 18}, {0, 0, 0, 0}}},{17261, {{29706, 1, 0, 0},{29706, 1, 1, 18}, {0, 0, 0, 0}}},{17262, {{29706, 1, 0, 0},{29706, 1, 1, 17}, {0, 0, 0, 0}}},{17262, {{29706, 1, 0, 0},{29706, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{17263, {{29707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17264, {{29708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17265, {{29709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{17266, {{29713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{17267, {{29715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17268, {{29716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17269, {{29717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17270, {{29718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17271, {{29719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17272, {{29720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17273, {{29721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17274, {{29722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17275, {{29723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17276, {{29724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17277, {{29725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17278, {{29726, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17279, {{29727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17280, {{29728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17281, {{29729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {9, {{17282, {{29730, 1, 0, 0},{29730, 1, 1, 18}, {0, 0, 0, 0}}},{17282, {{29730, 1, 0, 0},{29730, 1, 1, 18}, {0, 0, 0, 0}}},{17283, {{29730, 1, 0, 0},{29730, 1, 1, 20}, {0, 0, 0, 0}}},{17283, {{29730, 1, 0, 0},{29730, 1, 1, 20}, {0, 0, 0, 0}}},{30271, {{29730, 1, 0, 0},{29730, 1, 1, 0},{29730, 1, 1, 21},{64074, 1, 0, 0}, {0, 0, 0, 0}}},{30271, {{29730, 1, 0, 0},{29730, 1, 1, 0},{29730, 1, 1, 21},{64074, 1, 0, 0}, {0, 0, 0, 0}}},{30271, {{29730, 1, 0, 0},{29730, 1, 1, 0},{29730, 1, 1, 21},{64074, 1, 0, 0}, {0, 0, 0, 0}}},{30273, {{29730, 1, 0, 0},{29730, 1, 1, 19}, {0, 0, 0, 0}}},{30273, {{29730, 1, 0, 0},{29730, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{17284, {{29731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17285, {{29732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17288, {{29733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17290, {{29734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17291, {{29735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17292, {{29736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17293, {{29737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17294, {{29738, 1, 0, 0},{29738, 1, 1, 17}, {0, 0, 0, 0}}},{17294, {{29738, 1, 0, 0},{29738, 1, 1, 17}, {0, 0, 0, 0}}},{59898, {{29738, 1, 0, 0},{29738, 1, 1, 18}, {0, 0, 0, 0}}},{59898, {{29738, 1, 0, 0},{29738, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17295, {{29739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17296, {{29740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17297, {{29741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17298, {{29742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17299, {{29743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17300, {{29744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17301, {{29745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17303, {{29746, 1, 0, 0},{29746, 1, 1, 18}, {0, 0, 0, 0}}},{17303, {{29746, 1, 0, 0},{29746, 1, 1, 18}, {0, 0, 0, 0}}},{17304, {{29746, 1, 0, 0},{29746, 1, 1, 19}, {0, 0, 0, 0}}},{17304, {{29746, 1, 0, 0},{29746, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{17305, {{29747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{17306, {{29748, 1, 0, 0},{29748, 1, 1, 17}, {0, 0, 0, 0}}},{17306, {{29748, 1, 0, 0},{29748, 1, 1, 17}, {0, 0, 0, 0}}},{17307, {{29748, 1, 0, 0},{29748, 1, 1, 18}, {0, 0, 0, 0}}},{17307, {{29748, 1, 0, 0},{29748, 1, 1, 18}, {0, 0, 0, 0}}},{17308, {{29748, 1, 0, 0},{29748, 1, 1, 19}, {0, 0, 0, 0}}},{17308, {{29748, 1, 0, 0},{29748, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{17309, {{29749, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17310, {{29750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17311, {{29751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17312, {{29752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17313, {{29753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17314, {{29754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{17315, {{29756, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17316, {{29757, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17318, {{29758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17319, {{29759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17320, {{29760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17321, {{29761, 1, 0, 0},{29761, 1, 1, 17}, {0, 0, 0, 0}}},{17321, {{29761, 1, 0, 0},{29761, 1, 1, 17}, {0, 0, 0, 0}}},{17322, {{29761, 1, 0, 0},{29761, 1, 1, 18}, {0, 0, 0, 0}}},{17322, {{29761, 1, 0, 0},{29761, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17323, {{29762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17324, {{29763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17325, {{29764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17326, {{29765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17327, {{29766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17328, {{29767, 1, 0, 0},{29767, 1, 1, 18}, {0, 0, 0, 0}}},{17328, {{29767, 1, 0, 0},{29767, 1, 1, 18}, {0, 0, 0, 0}}},{17329, {{29767, 1, 0, 0},{29767, 1, 1, 17}, {0, 0, 0, 0}}},{17329, {{29767, 1, 0, 0},{29767, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{17330, {{29768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17331, {{29769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17332, {{29770, 1, 0, 0},{29770, 1, 1, 16}, {0, 0, 0, 0}}},{17332, {{29770, 1, 0, 0},{29770, 1, 1, 16}, {0, 0, 0, 0}}},{59902, {{29770, 1, 0, 0},{29770, 1, 1, 17}, {0, 0, 0, 0}}},{59902, {{29770, 1, 0, 0},{29770, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {6, {{17333, {{29771, 1, 0, 0},{29771, 1, 1, 18}, {0, 0, 0, 0}}},{17333, {{29771, 1, 0, 0},{29771, 1, 1, 18}, {0, 0, 0, 0}}},{17334, {{29771, 1, 0, 0},{29771, 1, 1, 17}, {0, 0, 0, 0}}},{17334, {{29771, 1, 0, 0},{29771, 1, 1, 17}, {0, 0, 0, 0}}},{17335, {{29771, 1, 0, 0},{29771, 1, 1, 19}, {0, 0, 0, 0}}},{17335, {{29771, 1, 0, 0},{29771, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{17336, {{29772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17337, {{29773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17338, {{29774, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17339, {{29775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17340, {{29776, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17341, {{29777, 1, 0, 0},{29777, 1, 1, 17}, {0, 0, 0, 0}}},{17341, {{29777, 1, 0, 0},{29777, 1, 1, 17}, {0, 0, 0, 0}}},{17342, {{29777, 1, 0, 0},{29777, 1, 1, 18}, {0, 0, 0, 0}}},{17342, {{29777, 1, 0, 0},{29777, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17343, {{29778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17344, {{29779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17345, {{29780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17346, {{29781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17347, {{29782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17348, {{29783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17349, {{29784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17350, {{29785, 1, 0, 0},{29785, 1, 1, 17}, {0, 0, 0, 0}}},{17350, {{29785, 1, 0, 0},{29785, 1, 1, 17}, {0, 0, 0, 0}}},{59905, {{29785, 1, 0, 0},{29785, 1, 1, 18}, {0, 0, 0, 0}}},{59905, {{29785, 1, 0, 0},{29785, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17351, {{29786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17352, {{29787, 1, 0, 0},{29787, 1, 1, 17}, {0, 0, 0, 0}}},{17352, {{29787, 1, 0, 0},{29787, 1, 1, 17}, {0, 0, 0, 0}}},{17353, {{29787, 1, 0, 0},{29787, 1, 1, 18}, {0, 0, 0, 0}}},{17353, {{29787, 1, 0, 0},{29787, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{17354, {{29788, 1, 0, 0},{29788, 1, 1, 17}, {0, 0, 0, 0}}},{17354, {{29788, 1, 0, 0},{29788, 1, 1, 17}, {0, 0, 0, 0}}},{17355, {{29788, 1, 0, 0},{29788, 1, 1, 18}, {0, 0, 0, 0}}},{17355, {{29788, 1, 0, 0},{29788, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17357, {{29789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17358, {{29790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17359, {{29791, 1, 0, 0},{29791, 1, 1, 18}, {0, 0, 0, 0}}},{17359, {{29791, 1, 0, 0},{29791, 1, 1, 18}, {0, 0, 0, 0}}},{17360, {{29791, 1, 0, 0},{29791, 1, 1, 19}, {0, 0, 0, 0}}},{17360, {{29791, 1, 0, 0},{29791, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{17361, {{29792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17362, {{29793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17363, {{29794, 1, 0, 0},{29794, 1, 1, 18}, {0, 0, 0, 0}}},{17363, {{29794, 1, 0, 0},{29794, 1, 1, 18}, {0, 0, 0, 0}}},{17364, {{29794, 1, 0, 0},{29794, 1, 1, 19}, {0, 0, 0, 0}}},{17364, {{29794, 1, 0, 0},{29794, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{17365, {{29795, 1, 0, 0},{29795, 1, 1, 17}, {0, 0, 0, 0}}},{17365, {{29795, 1, 0, 0},{29795, 1, 1, 17}, {0, 0, 0, 0}}},{17366, {{29795, 1, 0, 0},{29795, 1, 1, 18}, {0, 0, 0, 0}}},{17366, {{29795, 1, 0, 0},{29795, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17367, {{29796, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17368, {{29797, 1, 0, 0},{29797, 1, 1, 16}, {0, 0, 0, 0}}},{17368, {{29797, 1, 0, 0},{29797, 1, 1, 16}, {0, 0, 0, 0}}},{58047, {{29797, 1, 0, 0},{29797, 1, 1, 17}, {0, 0, 0, 0}}},{58047, {{29797, 1, 0, 0},{29797, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{17369, {{29798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17370, {{29799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17371, {{29800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17373, {{29801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17374, {{29802, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17375, {{29803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17376, {{29804, 1, 0, 0},{29804, 1, 1, 17}, {0, 0, 0, 0}}},{17376, {{29804, 1, 0, 0},{29804, 1, 1, 17}, {0, 0, 0, 0}}},{17377, {{29804, 1, 0, 0},{29804, 1, 1, 16}, {0, 0, 0, 0}}},{17377, {{29804, 1, 0, 0},{29804, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{17378, {{29805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17379, {{29806, 1, 0, 0},{29806, 1, 1, 17}, {0, 0, 0, 0}}},{17379, {{29806, 1, 0, 0},{29806, 1, 1, 17}, {0, 0, 0, 0}}},{17380, {{29806, 1, 0, 0},{29806, 1, 1, 18}, {0, 0, 0, 0}}},{17380, {{29806, 1, 0, 0},{29806, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{17381, {{29807, 1, 0, 0},{29807, 1, 1, 17}, {0, 0, 0, 0}}},{17381, {{29807, 1, 0, 0},{29807, 1, 1, 17}, {0, 0, 0, 0}}},{17382, {{29807, 1, 0, 0},{29807, 1, 1, 18}, {0, 0, 0, 0}}},{17382, {{29807, 1, 0, 0},{29807, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17383, {{29808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{17384, {{29809, 1, 0, 0},{29809, 1, 1, 17}, {0, 0, 0, 0}}},{17384, {{29809, 1, 0, 0},{29809, 1, 1, 17}, {0, 0, 0, 0}}},{17385, {{29809, 1, 0, 0},{29809, 1, 1, 18}, {0, 0, 0, 0}}},{17385, {{29809, 1, 0, 0},{29809, 1, 1, 18}, {0, 0, 0, 0}}},{58048, {{29809, 1, 0, 0},{29809, 1, 1, 19}, {0, 0, 0, 0}}},{58048, {{29809, 1, 0, 0},{29809, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{17387, {{29810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17388, {{29811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17389, {{29812, 1, 0, 0},{29812, 1, 1, 16}, {0, 0, 0, 0}}},{17389, {{29812, 1, 0, 0},{29812, 1, 1, 16}, {0, 0, 0, 0}}},{58049, {{29812, 1, 0, 0},{29812, 1, 1, 17}, {0, 0, 0, 0}}},{58049, {{29812, 1, 0, 0},{29812, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{17390, {{29813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17391, {{29814, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{17393, {{29817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17395, {{29818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17396, {{29819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17397, {{29820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17399, {{29821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17400, {{29822, 1, 0, 0},{29822, 1, 1, 17}, {0, 0, 0, 0}}},{17400, {{29822, 1, 0, 0},{29822, 1, 1, 17}, {0, 0, 0, 0}}},{17401, {{29822, 1, 0, 0},{29822, 1, 1, 18}, {0, 0, 0, 0}}},{17401, {{29822, 1, 0, 0},{29822, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17402, {{29823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17403, {{29824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17404, {{29825, 1, 0, 0},{29825, 1, 1, 17}, {0, 0, 0, 0}}},{17404, {{29825, 1, 0, 0},{29825, 1, 1, 17}, {0, 0, 0, 0}}},{17405, {{29825, 1, 0, 0},{29825, 1, 1, 18}, {0, 0, 0, 0}}},{17405, {{29825, 1, 0, 0},{29825, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17407, {{29826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17408, {{29827, 1, 0, 0},{29827, 1, 1, 17}, {0, 0, 0, 0}}},{17408, {{29827, 1, 0, 0},{29827, 1, 1, 17}, {0, 0, 0, 0}}},{17409, {{29827, 1, 0, 0},{29827, 1, 1, 18}, {0, 0, 0, 0}}},{17409, {{29827, 1, 0, 0},{29827, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17410, {{29828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17411, {{29829, 1, 0, 0},{29829, 1, 1, 17}, {0, 0, 0, 0}}},{17411, {{29829, 1, 0, 0},{29829, 1, 1, 17}, {0, 0, 0, 0}}},{17412, {{29829, 1, 0, 0},{29829, 1, 1, 18}, {0, 0, 0, 0}}},{17412, {{29829, 1, 0, 0},{29829, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17413, {{29830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17414, {{29831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17415, {{29832, 1, 0, 0},{29832, 1, 1, 17}, {0, 0, 0, 0}}},{17415, {{29832, 1, 0, 0},{29832, 1, 1, 17}, {0, 0, 0, 0}}},{68064, {{29832, 1, 0, 0},{29832, 1, 1, 18}, {0, 0, 0, 0}}},{68064, {{29832, 1, 0, 0},{29832, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{17416, {{29833, 1, 0, 0},{29833, 1, 1, 18}, {0, 0, 0, 0}}},{17416, {{29833, 1, 0, 0},{29833, 1, 1, 18}, {0, 0, 0, 0}}},{17417, {{29833, 1, 0, 0},{29833, 1, 1, 17}, {0, 0, 0, 0}}},{17417, {{29833, 1, 0, 0},{29833, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{17418, {{29834, 1, 0, 0},{29834, 1, 1, 16}, {0, 0, 0, 0}}},{17418, {{29834, 1, 0, 0},{29834, 1, 1, 16}, {0, 0, 0, 0}}},{17419, {{29834, 1, 0, 0},{29834, 1, 1, 17}, {0, 0, 0, 0}}},{17419, {{29834, 1, 0, 0},{29834, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{17420, {{29835, 1, 0, 0},{29835, 1, 1, 17}, {0, 0, 0, 0}}},{17420, {{29835, 1, 0, 0},{29835, 1, 1, 17}, {0, 0, 0, 0}}},{17421, {{29835, 1, 0, 0},{29835, 1, 1, 18}, {0, 0, 0, 0}}},{17421, {{29835, 1, 0, 0},{29835, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17422, {{29836, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{17423, {{29839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17424, {{29840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17425, {{29841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17426, {{29842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17427, {{29843, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17428, {{29844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17429, {{29845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17430, {{29846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17431, {{29847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17432, {{29848, 1, 0, 0},{29848, 1, 1, 18}, {0, 0, 0, 0}}},{17432, {{29848, 1, 0, 0},{29848, 1, 1, 18}, {0, 0, 0, 0}}},{17433, {{29848, 1, 0, 0},{29848, 1, 1, 17}, {0, 0, 0, 0}}},{17433, {{29848, 1, 0, 0},{29848, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{17434, {{29849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17435, {{29850, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17436, {{29851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17437, {{29852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17439, {{29853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17440, {{29854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17441, {{29855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17442, {{29856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17443, {{29857, 1, 0, 0},{29857, 1, 1, 17}, {0, 0, 0, 0}}},{17443, {{29857, 1, 0, 0},{29857, 1, 1, 17}, {0, 0, 0, 0}}},{17444, {{29857, 1, 0, 0},{29857, 1, 1, 18}, {0, 0, 0, 0}}},{17444, {{29857, 1, 0, 0},{29857, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17445, {{29858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{17446, {{29859, 1, 0, 0},{29859, 1, 1, 17}, {0, 0, 0, 0}}},{17446, {{29859, 1, 0, 0},{29859, 1, 1, 17}, {0, 0, 0, 0}}},{17447, {{29859, 1, 0, 0},{29859, 1, 1, 19}, {0, 0, 0, 0}}},{17447, {{29859, 1, 0, 0},{29859, 1, 1, 19}, {0, 0, 0, 0}}},{17448, {{29859, 1, 0, 0},{29859, 1, 1, 18}, {0, 0, 0, 0}}},{17448, {{29859, 1, 0, 0},{29859, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{17449, {{29861, 1, 0, 0},{29861, 1, 1, 17}, {0, 0, 0, 0}}},{17449, {{29861, 1, 0, 0},{29861, 1, 1, 17}, {0, 0, 0, 0}}},{17450, {{29861, 1, 0, 0},{29861, 1, 1, 18}, {0, 0, 0, 0}}},{17450, {{29861, 1, 0, 0},{29861, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{17451, {{29862, 1, 0, 0},{29862, 1, 1, 18}, {0, 0, 0, 0}}},{17451, {{29862, 1, 0, 0},{29862, 1, 1, 18}, {0, 0, 0, 0}}},{17452, {{29862, 1, 0, 0},{29862, 1, 1, 17}, {0, 0, 0, 0}}},{17452, {{29862, 1, 0, 0},{29862, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{17453, {{29863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17454, {{29864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17455, {{29865, 1, 0, 0},{29865, 1, 1, 17}, {0, 0, 0, 0}}},{17455, {{29865, 1, 0, 0},{29865, 1, 1, 17}, {0, 0, 0, 0}}},{17456, {{29865, 1, 0, 0},{29865, 1, 1, 18}, {0, 0, 0, 0}}},{17456, {{29865, 1, 0, 0},{29865, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17457, {{29866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17458, {{29867, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17459, {{29868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17460, {{29869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17462, {{29870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17463, {{29871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17464, {{29872, 1, 0, 0},{29872, 1, 1, 19}, {0, 0, 0, 0}}},{17464, {{29872, 1, 0, 0},{29872, 1, 1, 19}, {0, 0, 0, 0}}},{17465, {{29872, 1, 0, 0},{29872, 1, 1, 18}, {0, 0, 0, 0}}},{17465, {{29872, 1, 0, 0},{29872, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17466, {{29873, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17467, {{29874, 1, 0, 0},{29874, 1, 1, 18}, {0, 0, 0, 0}}},{17467, {{29874, 1, 0, 0},{29874, 1, 1, 18}, {0, 0, 0, 0}}},{17468, {{29874, 1, 0, 0},{29874, 1, 1, 17}, {0, 0, 0, 0}}},{17468, {{29874, 1, 0, 0},{29874, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{17469, {{29875, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{17470, {{29877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17471, {{29878, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17472, {{29879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17473, {{29880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17474, {{29881, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17475, {{29882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17476, {{29883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17477, {{29884, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17478, {{29885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17479, {{29886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17480, {{29887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17482, {{29888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17483, {{29889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17485, {{29890, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17486, {{29891, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17487, {{29892, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17489, {{29893, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17490, {{29894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17491, {{29895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17493, {{29896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17494, {{29897, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{17495, {{29898, 1, 0, 0},{29898, 1, 1, 18}, {0, 0, 0, 0}}},{17495, {{29898, 1, 0, 0},{29898, 1, 1, 18}, {0, 0, 0, 0}}},{17496, {{29898, 1, 0, 0},{29898, 1, 1, 19}, {0, 0, 0, 0}}},{17496, {{29898, 1, 0, 0},{29898, 1, 1, 19}, {0, 0, 0, 0}}},{17497, {{29898, 1, 0, 0},{29898, 1, 1, 20}, {0, 0, 0, 0}}},{17497, {{29898, 1, 0, 0},{29898, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {6, {{17498, {{29899, 1, 0, 0},{29899, 1, 1, 16}, {0, 0, 0, 0}}},{17498, {{29899, 1, 0, 0},{29899, 1, 1, 16}, {0, 0, 0, 0}}},{17499, {{29899, 1, 0, 0},{29899, 1, 1, 17}, {0, 0, 0, 0}}},{17499, {{29899, 1, 0, 0},{29899, 1, 1, 17}, {0, 0, 0, 0}}},{17500, {{29899, 1, 0, 0},{29899, 1, 1, 18}, {0, 0, 0, 0}}},{17500, {{29899, 1, 0, 0},{29899, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17501, {{29900, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17502, {{29901, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17503, {{29902, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17504, {{29903, 1, 0, 0},{29903, 1, 1, 17}, {0, 0, 0, 0}}},{17504, {{29903, 1, 0, 0},{29903, 1, 1, 17}, {0, 0, 0, 0}}},{17505, {{29903, 1, 0, 0},{29903, 1, 1, 18}, {0, 0, 0, 0}}},{17505, {{29903, 1, 0, 0},{29903, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17506, {{29904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17507, {{29905, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{17508, {{29907, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17509, {{29908, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17510, {{29909, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17511, {{29910, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17512, {{29911, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17513, {{29912, 1, 0, 0},{29912, 1, 1, 19}, {0, 0, 0, 0}}},{17513, {{29912, 1, 0, 0},{29912, 1, 1, 19}, {0, 0, 0, 0}}},{17514, {{29912, 1, 0, 0},{29912, 1, 1, 18}, {0, 0, 0, 0}}},{17514, {{29912, 1, 0, 0},{29912, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17515, {{29913, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17516, {{29914, 1, 0, 0},{29914, 1, 1, 17}, {0, 0, 0, 0}}},{17516, {{29914, 1, 0, 0},{29914, 1, 1, 17}, {0, 0, 0, 0}}},{17517, {{29914, 1, 0, 0},{29914, 1, 1, 18}, {0, 0, 0, 0}}},{17517, {{29914, 1, 0, 0},{29914, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17518, {{29915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{17519, {{29916, 1, 0, 0},{29916, 1, 1, 20}, {0, 0, 0, 0}}},{17519, {{29916, 1, 0, 0},{29916, 1, 1, 20}, {0, 0, 0, 0}}},{17520, {{29916, 1, 0, 0},{29916, 1, 1, 19}, {0, 0, 0, 0}}},{17520, {{29916, 1, 0, 0},{29916, 1, 1, 19}, {0, 0, 0, 0}}},{17521, {{29916, 1, 0, 0},{29916, 1, 1, 18}, {0, 0, 0, 0}}},{17521, {{29916, 1, 0, 0},{29916, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17522, {{29917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17523, {{29918, 1, 0, 0},{29918, 1, 1, 18}, {0, 0, 0, 0}}},{17523, {{29918, 1, 0, 0},{29918, 1, 1, 18}, {0, 0, 0, 0}}},{17524, {{29918, 1, 0, 0},{29918, 1, 1, 17}, {0, 0, 0, 0}}},{17524, {{29918, 1, 0, 0},{29918, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {6, {{17525, {{29919, 1, 0, 0},{29919, 1, 1, 18}, {0, 0, 0, 0}}},{17525, {{29919, 1, 0, 0},{29919, 1, 1, 18}, {0, 0, 0, 0}}},{17526, {{29919, 1, 0, 0},{29919, 1, 1, 19}, {0, 0, 0, 0}}},{17526, {{29919, 1, 0, 0},{29919, 1, 1, 19}, {0, 0, 0, 0}}},{17527, {{29919, 1, 0, 0},{29919, 1, 1, 17}, {0, 0, 0, 0}}},{17527, {{29919, 1, 0, 0},{29919, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {6, {{17528, {{29920, 1, 0, 0},{29920, 1, 1, 20}, {0, 0, 0, 0}}},{17528, {{29920, 1, 0, 0},{29920, 1, 1, 20}, {0, 0, 0, 0}}},{17529, {{29920, 1, 0, 0},{29920, 1, 1, 18}, {0, 0, 0, 0}}},{17529, {{29920, 1, 0, 0},{29920, 1, 1, 18}, {0, 0, 0, 0}}},{17530, {{29920, 1, 0, 0},{29920, 1, 1, 19}, {0, 0, 0, 0}}},{17530, {{29920, 1, 0, 0},{29920, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{17531, {{29921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17532, {{29922, 1, 0, 0},{29922, 1, 1, 18}, {0, 0, 0, 0}}},{17532, {{29922, 1, 0, 0},{29922, 1, 1, 18}, {0, 0, 0, 0}}},{17533, {{29922, 1, 0, 0},{29922, 1, 1, 17}, {0, 0, 0, 0}}},{17533, {{29922, 1, 0, 0},{29922, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {6, {{17534, {{29923, 1, 0, 0},{29923, 1, 1, 18}, {0, 0, 0, 0}}},{17534, {{29923, 1, 0, 0},{29923, 1, 1, 18}, {0, 0, 0, 0}}},{17535, {{29923, 1, 0, 0},{29923, 1, 1, 19}, {0, 0, 0, 0}}},{17535, {{29923, 1, 0, 0},{29923, 1, 1, 19}, {0, 0, 0, 0}}},{17536, {{29923, 1, 0, 0},{29923, 1, 1, 17}, {0, 0, 0, 0}}},{17536, {{29923, 1, 0, 0},{29923, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{17537, {{29924, 1, 0, 0},{29924, 1, 1, 18}, {0, 0, 0, 0}}},{17537, {{29924, 1, 0, 0},{29924, 1, 1, 18}, {0, 0, 0, 0}}},{17538, {{29924, 1, 0, 0},{29924, 1, 1, 17}, {0, 0, 0, 0}}},{17538, {{29924, 1, 0, 0},{29924, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{17539, {{29925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17540, {{29926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17541, {{29927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17542, {{29928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17543, {{29929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17544, {{29930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17545, {{29931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17546, {{29932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17547, {{29933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17548, {{29934, 1, 0, 0},{29934, 1, 1, 17}, {0, 0, 0, 0}}},{17548, {{29934, 1, 0, 0},{29934, 1, 1, 17}, {0, 0, 0, 0}}},{17549, {{29934, 1, 0, 0},{29934, 1, 1, 18}, {0, 0, 0, 0}}},{17549, {{29934, 1, 0, 0},{29934, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{17550, {{29935, 1, 0, 0},{29935, 1, 1, 18}, {0, 0, 0, 0}}},{17550, {{29935, 1, 0, 0},{29935, 1, 1, 18}, {0, 0, 0, 0}}},{17551, {{29935, 1, 0, 0},{29935, 1, 1, 19}, {0, 0, 0, 0}}},{17551, {{29935, 1, 0, 0},{29935, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{17552, {{29936, 1, 0, 0},{29936, 1, 1, 17}, {0, 0, 0, 0}}},{17552, {{29936, 1, 0, 0},{29936, 1, 1, 17}, {0, 0, 0, 0}}},{17553, {{29936, 1, 0, 0},{29936, 1, 1, 18}, {0, 0, 0, 0}}},{17553, {{29936, 1, 0, 0},{29936, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17554, {{29937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17555, {{29938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17556, {{29939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17557, {{29940, 1, 0, 0},{29940, 1, 1, 17}, {0, 0, 0, 0}}},{17557, {{29940, 1, 0, 0},{29940, 1, 1, 17}, {0, 0, 0, 0}}},{17558, {{29940, 1, 0, 0},{29940, 1, 1, 18}, {0, 0, 0, 0}}},{17558, {{29940, 1, 0, 0},{29940, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17560, {{29941, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17561, {{29942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17562, {{29943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17563, {{29944, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17564, {{29945, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17565, {{29946, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17566, {{29947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17567, {{29948, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17568, {{29949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17569, {{29950, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17570, {{29951, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17571, {{29952, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17572, {{29953, 1, 0, 0},{29953, 1, 1, 17}, {0, 0, 0, 0}}},{17572, {{29953, 1, 0, 0},{29953, 1, 1, 17}, {0, 0, 0, 0}}},{17573, {{29953, 1, 0, 0},{29953, 1, 1, 18}, {0, 0, 0, 0}}},{17573, {{29953, 1, 0, 0},{29953, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17574, {{29954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17575, {{29955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{17576, {{29956, 1, 0, 0},{29956, 1, 1, 20}, {0, 0, 0, 0}}},{17576, {{29956, 1, 0, 0},{29956, 1, 1, 20}, {0, 0, 0, 0}}},{17577, {{29956, 1, 0, 0},{29956, 1, 1, 18}, {0, 0, 0, 0}}},{17577, {{29956, 1, 0, 0},{29956, 1, 1, 18}, {0, 0, 0, 0}}},{17578, {{29956, 1, 0, 0},{29956, 1, 1, 21}, {0, 0, 0, 0}}},{17578, {{29956, 1, 0, 0},{29956, 1, 1, 21}, {0, 0, 0, 0}}},{17579, {{29956, 1, 0, 0},{29956, 1, 1, 19}, {0, 0, 0, 0}}},{17579, {{29956, 1, 0, 0},{29956, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{17580, {{29957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{17581, {{29958, 1, 0, 0},{29958, 1, 1, 17}, {0, 0, 0, 0}}},{17581, {{29958, 1, 0, 0},{29958, 1, 1, 17}, {0, 0, 0, 0}}},{17582, {{29958, 1, 0, 0},{29958, 1, 1, 18}, {0, 0, 0, 0}}},{17582, {{29958, 1, 0, 0},{29958, 1, 1, 18}, {0, 0, 0, 0}}},{17583, {{29958, 1, 0, 0},{29958, 1, 1, 19}, {0, 0, 0, 0}}},{17583, {{29958, 1, 0, 0},{29958, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{17584, {{29959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17585, {{29960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17586, {{29961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17587, {{29962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17588, {{29963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17589, {{29964, 1, 0, 0},{29964, 1, 1, 18}, {0, 0, 0, 0}}},{17589, {{29964, 1, 0, 0},{29964, 1, 1, 18}, {0, 0, 0, 0}}},{17590, {{29964, 1, 0, 0},{29964, 1, 1, 19}, {0, 0, 0, 0}}},{17590, {{29964, 1, 0, 0},{29964, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {8, {{17591, {{29965, 1, 0, 0},{29965, 1, 1, 18}, {0, 0, 0, 0}}},{17591, {{29965, 1, 0, 0},{29965, 1, 1, 18}, {0, 0, 0, 0}}},{17592, {{29965, 1, 0, 0},{29965, 1, 1, 19}, {0, 0, 0, 0}}},{17592, {{29965, 1, 0, 0},{29965, 1, 1, 19}, {0, 0, 0, 0}}},{17593, {{29965, 1, 0, 0},{29965, 1, 1, 20}, {0, 0, 0, 0}}},{17593, {{29965, 1, 0, 0},{29965, 1, 1, 20}, {0, 0, 0, 0}}},{17594, {{29965, 1, 0, 0},{29965, 1, 1, 21}, {0, 0, 0, 0}}},{17594, {{29965, 1, 0, 0},{29965, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {4, {{17595, {{29966, 1, 0, 0},{29966, 1, 1, 17}, {0, 0, 0, 0}}},{17595, {{29966, 1, 0, 0},{29966, 1, 1, 17}, {0, 0, 0, 0}}},{17596, {{29966, 1, 0, 0},{29966, 1, 1, 18}, {0, 0, 0, 0}}},{17596, {{29966, 1, 0, 0},{29966, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17597, {{29967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17598, {{29968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{17599, {{29969, 1, 0, 0},{29969, 1, 1, 20}, {0, 0, 0, 0}}},{17599, {{29969, 1, 0, 0},{29969, 1, 1, 20}, {0, 0, 0, 0}}},{17600, {{29969, 1, 0, 0},{29969, 1, 1, 22}, {0, 0, 0, 0}}},{17600, {{29969, 1, 0, 0},{29969, 1, 1, 22}, {0, 0, 0, 0}}},{17601, {{29969, 1, 0, 0},{29969, 1, 1, 21}, {0, 0, 0, 0}}},{17601, {{29969, 1, 0, 0},{29969, 1, 1, 21}, {0, 0, 0, 0}}},{17602, {{29969, 1, 0, 0},{29969, 1, 1, 23}, {0, 0, 0, 0}}},{17602, {{29969, 1, 0, 0},{29969, 1, 1, 23}, {0, 0, 0, 0}}}}},
    {1, {{17603, {{29970, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17604, {{29971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17605, {{29972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{17606, {{29973, 1, 0, 0},{29973, 1, 1, 19}, {0, 0, 0, 0}}},{17606, {{29973, 1, 0, 0},{29973, 1, 1, 19}, {0, 0, 0, 0}}},{17607, {{29973, 1, 0, 0},{29973, 1, 1, 21}, {0, 0, 0, 0}}},{17607, {{29973, 1, 0, 0},{29973, 1, 1, 21}, {0, 0, 0, 0}}},{17608, {{29973, 1, 0, 0},{29973, 1, 1, 20}, {0, 0, 0, 0}}},{17608, {{29973, 1, 0, 0},{29973, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{17609, {{29974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17610, {{29975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17611, {{29976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{17612, {{29978, 1, 0, 0},{29978, 1, 1, 17}, {0, 0, 0, 0}}},{17612, {{29978, 1, 0, 0},{29978, 1, 1, 17}, {0, 0, 0, 0}}},{58077, {{29978, 1, 0, 0},{29978, 1, 1, 18}, {0, 0, 0, 0}}},{58077, {{29978, 1, 0, 0},{29978, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17614, {{29979, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17615, {{29980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17616, {{29981, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17617, {{29982, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17618, {{29983, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17619, {{29984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17620, {{29985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17621, {{29986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17622, {{29987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17623, {{29988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17624, {{29989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17625, {{29990, 1, 0, 0},{29990, 1, 1, 18}, {0, 0, 0, 0}}},{17625, {{29990, 1, 0, 0},{29990, 1, 1, 18}, {0, 0, 0, 0}}},{17626, {{29990, 1, 0, 0},{29990, 1, 1, 19}, {0, 0, 0, 0}}},{17626, {{29990, 1, 0, 0},{29990, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{17627, {{29991, 1, 0, 0},{29991, 1, 1, 17}, {0, 0, 0, 0}}},{17627, {{29991, 1, 0, 0},{29991, 1, 1, 17}, {0, 0, 0, 0}}},{17628, {{29991, 1, 0, 0},{29991, 1, 1, 18}, {0, 0, 0, 0}}},{17628, {{29991, 1, 0, 0},{29991, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17629, {{29992, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17630, {{29993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17631, {{29994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17632, {{29995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17633, {{29996, 1, 0, 0},{29996, 1, 1, 17}, {0, 0, 0, 0}}},{17633, {{29996, 1, 0, 0},{29996, 1, 1, 17}, {0, 0, 0, 0}}},{58080, {{29996, 1, 0, 0},{29996, 1, 1, 18}, {0, 0, 0, 0}}},{58080, {{29996, 1, 0, 0},{29996, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17634, {{29997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{17635, {{29999, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17636, {{30000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17637, {{30001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17638, {{30002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17639, {{30003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17640, {{30004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{17641, {{30006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17642, {{30007, 1, 0, 0},{30007, 1, 1, 17}, {0, 0, 0, 0}}},{17642, {{30007, 1, 0, 0},{30007, 1, 1, 17}, {0, 0, 0, 0}}},{58085, {{30007, 1, 0, 0},{30007, 1, 1, 18}, {0, 0, 0, 0}}},{58085, {{30007, 1, 0, 0},{30007, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17643, {{30008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17644, {{30009, 1, 0, 0},{30009, 1, 1, 17}, {0, 0, 0, 0}}},{17644, {{30009, 1, 0, 0},{30009, 1, 1, 17}, {0, 0, 0, 0}}},{56877, {{30009, 1, 0, 0},{30009, 1, 1, 18}, {0, 0, 0, 0}}},{56877, {{30009, 1, 0, 0},{30009, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17645, {{30010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17646, {{30011, 1, 0, 0},{30011, 1, 1, 18}, {0, 0, 0, 0}}},{17646, {{30011, 1, 0, 0},{30011, 1, 1, 18}, {0, 0, 0, 0}}},{17647, {{30011, 1, 0, 0},{30011, 1, 1, 17}, {0, 0, 0, 0}}},{17647, {{30011, 1, 0, 0},{30011, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{17649, {{30012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17650, {{30013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17651, {{30014, 1, 0, 0},{30014, 1, 1, 17}, {0, 0, 0, 0}}},{17651, {{30014, 1, 0, 0},{30014, 1, 1, 17}, {0, 0, 0, 0}}},{17652, {{30014, 1, 0, 0},{30014, 1, 1, 18}, {0, 0, 0, 0}}},{17652, {{30014, 1, 0, 0},{30014, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17653, {{30015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17654, {{30016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17655, {{30017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17656, {{30018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17657, {{30019, 1, 0, 0},{30019, 1, 1, 17}, {0, 0, 0, 0}}},{17657, {{30019, 1, 0, 0},{30019, 1, 1, 17}, {0, 0, 0, 0}}},{17658, {{30019, 1, 0, 0},{30019, 1, 1, 18}, {0, 0, 0, 0}}},{17658, {{30019, 1, 0, 0},{30019, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17659, {{30020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{17660, {{30022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17661, {{30023, 1, 0, 0},{30023, 1, 1, 17}, {0, 0, 0, 0}}},{17661, {{30023, 1, 0, 0},{30023, 1, 1, 17}, {0, 0, 0, 0}}},{17662, {{30023, 1, 0, 0},{30023, 1, 1, 18}, {0, 0, 0, 0}}},{17662, {{30023, 1, 0, 0},{30023, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17663, {{30024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17664, {{30025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17665, {{30026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17666, {{30027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17667, {{30028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17668, {{30029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17669, {{30030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17670, {{30031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17671, {{30032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17672, {{30033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17673, {{30034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17674, {{30035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17675, {{30036, 1, 0, 0},{30036, 1, 1, 19}, {0, 0, 0, 0}}},{17675, {{30036, 1, 0, 0},{30036, 1, 1, 19}, {0, 0, 0, 0}}},{17676, {{30036, 1, 0, 0},{30036, 1, 1, 18}, {0, 0, 0, 0}}},{17676, {{30036, 1, 0, 0},{30036, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17677, {{30037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17678, {{30038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17679, {{30039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17680, {{30040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{17681, {{30041, 1, 0, 0},{30041, 1, 1, 17}, {0, 0, 0, 0}}},{17681, {{30041, 1, 0, 0},{30041, 1, 1, 17}, {0, 0, 0, 0}}},{17682, {{30041, 1, 0, 0},{30041, 1, 1, 18}, {0, 0, 0, 0}}},{17682, {{30041, 1, 0, 0},{30041, 1, 1, 18}, {0, 0, 0, 0}}},{58087, {{30041, 1, 0, 0},{30041, 1, 1, 19}, {0, 0, 0, 0}}},{58087, {{30041, 1, 0, 0},{30041, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{17683, {{30042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17684, {{30043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17685, {{30044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17686, {{30045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17687, {{30046, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17688, {{30047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17690, {{30048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17691, {{30049, 1, 0, 0},{30049, 1, 1, 17}, {0, 0, 0, 0}}},{17691, {{30049, 1, 0, 0},{30049, 1, 1, 17}, {0, 0, 0, 0}}},{17692, {{30049, 1, 0, 0},{30049, 1, 1, 18}, {0, 0, 0, 0}}},{17692, {{30049, 1, 0, 0},{30049, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17693, {{30050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17694, {{30051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17695, {{30052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17696, {{30053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17697, {{30054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17698, {{30055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{17699, {{30057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17700, {{30058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17701, {{30059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17702, {{30060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17703, {{30061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17704, {{30062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17705, {{30063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{17706, {{30064, 1, 0, 0},{30064, 1, 1, 17}, {0, 0, 0, 0}}},{17706, {{30064, 1, 0, 0},{30064, 1, 1, 17}, {0, 0, 0, 0}}},{17707, {{30064, 1, 0, 0},{30064, 1, 1, 18}, {0, 0, 0, 0}}},{17707, {{30064, 1, 0, 0},{30064, 1, 1, 18}, {0, 0, 0, 0}}},{58096, {{30064, 1, 0, 0},{30064, 1, 1, 19}, {0, 0, 0, 0}}},{58096, {{30064, 1, 0, 0},{30064, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{17708, {{30065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17709, {{30066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17710, {{30067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17711, {{30068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17712, {{30069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17713, {{30070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17714, {{30071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17715, {{30072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17716, {{30073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17717, {{30074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17718, {{30075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17719, {{30076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17720, {{30077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17721, {{30078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17722, {{30079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17723, {{30080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17724, {{30081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17725, {{30082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17726, {{30083, 1, 0, 0},{30083, 1, 1, 16}, {0, 0, 0, 0}}},{17726, {{30083, 1, 0, 0},{30083, 1, 1, 16}, {0, 0, 0, 0}}},{17727, {{30083, 1, 0, 0},{30083, 1, 1, 17}, {0, 0, 0, 0}}},{17727, {{30083, 1, 0, 0},{30083, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{17728, {{30084, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17729, {{30085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17730, {{30086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17731, {{30087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17732, {{30088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17733, {{30089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17734, {{30090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17735, {{30091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17736, {{30092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{17737, {{30094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17738, {{30095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17739, {{30096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17740, {{30097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17741, {{30098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17742, {{30099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17743, {{30100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17744, {{30101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{17745, {{30104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17746, {{30105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17748, {{30106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17749, {{30107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17750, {{30108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17751, {{30109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17752, {{30110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{17753, {{30114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17754, {{30115, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17755, {{30116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17756, {{30117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17757, {{30118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17758, {{30119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17759, {{30120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17760, {{30121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17761, {{30122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17762, {{30123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{17764, {{30128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17765, {{30129, 1, 0, 0},{30129, 1, 1, 17}, {0, 0, 0, 0}}},{17765, {{30129, 1, 0, 0},{30129, 1, 1, 17}, {0, 0, 0, 0}}},{17766, {{30129, 1, 0, 0},{30129, 1, 1, 18}, {0, 0, 0, 0}}},{17766, {{30129, 1, 0, 0},{30129, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17767, {{30130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17768, {{30131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17769, {{30132, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17770, {{30133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17771, {{30134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17772, {{30135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17773, {{30136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17774, {{30137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17775, {{30138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17776, {{30139, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17777, {{30140, 1, 0, 0},{30140, 1, 1, 18}, {0, 0, 0, 0}}},{17777, {{30140, 1, 0, 0},{30140, 1, 1, 18}, {0, 0, 0, 0}}},{17778, {{30140, 1, 0, 0},{30140, 1, 1, 19}, {0, 0, 0, 0}}},{17778, {{30140, 1, 0, 0},{30140, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{17779, {{30141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17780, {{30142, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17781, {{30143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17782, {{30144, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17783, {{30145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17784, {{30146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17785, {{30147, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17786, {{30148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17787, {{30149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17788, {{30150, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17789, {{30151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{17790, {{30154, 1, 0, 0},{30154, 1, 1, 17}, {0, 0, 0, 0}}},{17790, {{30154, 1, 0, 0},{30154, 1, 1, 17}, {0, 0, 0, 0}}},{17791, {{30154, 1, 0, 0},{30154, 1, 1, 18}, {0, 0, 0, 0}}},{17791, {{30154, 1, 0, 0},{30154, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17792, {{30155, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17793, {{30156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17794, {{30157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17795, {{30158, 1, 0, 0},{30158, 1, 1, 17}, {0, 0, 0, 0}}},{17795, {{30158, 1, 0, 0},{30158, 1, 1, 17}, {0, 0, 0, 0}}},{17796, {{30158, 1, 0, 0},{30158, 1, 1, 18}, {0, 0, 0, 0}}},{17796, {{30158, 1, 0, 0},{30158, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17797, {{30159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17798, {{30160, 1, 0, 0},{30160, 1, 1, 16}, {0, 0, 0, 0}}},{17798, {{30160, 1, 0, 0},{30160, 1, 1, 16}, {0, 0, 0, 0}}},{17799, {{30160, 1, 0, 0},{30160, 1, 1, 17}, {0, 0, 0, 0}}},{17799, {{30160, 1, 0, 0},{30160, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{17800, {{30161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17801, {{30162, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17802, {{30163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17803, {{30164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17804, {{30165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{17805, {{30167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17806, {{30168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17807, {{30169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17808, {{30170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17809, {{30171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17810, {{30172, 1, 0, 0},{30172, 1, 1, 17}, {0, 0, 0, 0}}},{17810, {{30172, 1, 0, 0},{30172, 1, 1, 17}, {0, 0, 0, 0}}},{17811, {{30172, 1, 0, 0},{30172, 1, 1, 18}, {0, 0, 0, 0}}},{17811, {{30172, 1, 0, 0},{30172, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17812, {{30173, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17813, {{30174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17814, {{30175, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17815, {{30176, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17816, {{30177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17817, {{30178, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17818, {{30179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17819, {{30180, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17820, {{30181, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17821, {{30182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17822, {{30183, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{17823, {{30185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{17824, {{30188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17825, {{30189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17826, {{30190, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17827, {{30191, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17828, {{30192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17829, {{30193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17830, {{30194, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17831, {{30195, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17832, {{30196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17833, {{30197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17834, {{30198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17835, {{30199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17836, {{30200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17837, {{30201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17838, {{30202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17839, {{30203, 1, 0, 0},{30203, 1, 1, 16}, {0, 0, 0, 0}}},{17839, {{30203, 1, 0, 0},{30203, 1, 1, 16}, {0, 0, 0, 0}}},{17840, {{30203, 1, 0, 0},{30203, 1, 1, 17}, {0, 0, 0, 0}}},{17840, {{30203, 1, 0, 0},{30203, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{17841, {{30204, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17842, {{30205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17843, {{30206, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17844, {{30207, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17845, {{30208, 1, 0, 0},{30208, 1, 1, 18}, {0, 0, 0, 0}}},{17845, {{30208, 1, 0, 0},{30208, 1, 1, 18}, {0, 0, 0, 0}}},{17846, {{30208, 1, 0, 0},{30208, 1, 1, 17}, {0, 0, 0, 0}}},{17846, {{30208, 1, 0, 0},{30208, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{17847, {{30209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17848, {{30210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17849, {{30211, 1, 0, 0},{30211, 1, 1, 17}, {0, 0, 0, 0}}},{17849, {{30211, 1, 0, 0},{30211, 1, 1, 17}, {0, 0, 0, 0}}},{17850, {{30211, 1, 0, 0},{30211, 1, 1, 18}, {0, 0, 0, 0}}},{17850, {{30211, 1, 0, 0},{30211, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17851, {{30212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{17852, {{30215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{17853, {{30216, 1, 0, 0},{30216, 1, 1, 18}, {0, 0, 0, 0}}},{17853, {{30216, 1, 0, 0},{30216, 1, 1, 18}, {0, 0, 0, 0}}},{17854, {{30216, 1, 0, 0},{30216, 1, 1, 19}, {0, 0, 0, 0}}},{17854, {{30216, 1, 0, 0},{30216, 1, 1, 19}, {0, 0, 0, 0}}},{58122, {{30216, 1, 0, 0},{30216, 1, 1, 20}, {0, 0, 0, 0}}},{58122, {{30216, 1, 0, 0},{30216, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{17855, {{30217, 1, 0, 0},{30217, 1, 1, 17}, {0, 0, 0, 0}}},{17855, {{30217, 1, 0, 0},{30217, 1, 1, 17}, {0, 0, 0, 0}}},{17856, {{30217, 1, 0, 0},{30217, 1, 1, 18}, {0, 0, 0, 0}}},{17856, {{30217, 1, 0, 0},{30217, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17857, {{30218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17858, {{30219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17859, {{30220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17860, {{30221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17861, {{30222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17862, {{30223, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17863, {{30224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17864, {{30225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17865, {{30226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17866, {{30227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17867, {{30228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17868, {{30229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17869, {{30230, 1, 0, 0},{30230, 1, 1, 17}, {0, 0, 0, 0}}},{17869, {{30230, 1, 0, 0},{30230, 1, 1, 17}, {0, 0, 0, 0}}},{17870, {{30230, 1, 0, 0},{30230, 1, 1, 18}, {0, 0, 0, 0}}},{17870, {{30230, 1, 0, 0},{30230, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{17871, {{30232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17872, {{30233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17873, {{30234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{17874, {{30235, 1, 0, 0},{30235, 1, 1, 17}, {0, 0, 0, 0}}},{17874, {{30235, 1, 0, 0},{30235, 1, 1, 17}, {0, 0, 0, 0}}},{17875, {{30235, 1, 0, 0},{30235, 1, 1, 18}, {0, 0, 0, 0}}},{17875, {{30235, 1, 0, 0},{30235, 1, 1, 18}, {0, 0, 0, 0}}},{58123, {{30235, 1, 0, 0},{30235, 1, 1, 19}, {0, 0, 0, 0}}},{58123, {{30235, 1, 0, 0},{30235, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{17876, {{30236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17877, {{30237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17878, {{30238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17879, {{30239, 1, 0, 0},{30239, 1, 1, 18}, {0, 0, 0, 0}}},{17879, {{30239, 1, 0, 0},{30239, 1, 1, 18}, {0, 0, 0, 0}}},{17880, {{30239, 1, 0, 0},{30239, 1, 1, 17}, {0, 0, 0, 0}}},{17880, {{30239, 1, 0, 0},{30239, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{17881, {{30240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17882, {{30241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17883, {{30242, 1, 0, 0},{30242, 1, 1, 17}, {0, 0, 0, 0}}},{17883, {{30242, 1, 0, 0},{30242, 1, 1, 17}, {0, 0, 0, 0}}},{17884, {{30242, 1, 0, 0},{30242, 1, 1, 18}, {0, 0, 0, 0}}},{17884, {{30242, 1, 0, 0},{30242, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17885, {{30243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17886, {{30244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17887, {{30245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17888, {{30246, 1, 0, 0},{30246, 1, 1, 18}, {0, 0, 0, 0}}},{17888, {{30246, 1, 0, 0},{30246, 1, 1, 18}, {0, 0, 0, 0}}},{17889, {{30246, 1, 0, 0},{30246, 1, 1, 19}, {0, 0, 0, 0}}},{17889, {{30246, 1, 0, 0},{30246, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{17890, {{30247, 1, 0, 0},{30247, 1, 1, 17}, {0, 0, 0, 0}}},{17890, {{30247, 1, 0, 0},{30247, 1, 1, 17}, {0, 0, 0, 0}}},{17891, {{30247, 1, 0, 0},{30247, 1, 1, 18}, {0, 0, 0, 0}}},{17891, {{30247, 1, 0, 0},{30247, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17892, {{30248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17893, {{30249, 1, 0, 0},{30249, 1, 1, 17}, {0, 0, 0, 0}}},{17893, {{30249, 1, 0, 0},{30249, 1, 1, 17}, {0, 0, 0, 0}}},{17894, {{30249, 1, 0, 0},{30249, 1, 1, 18}, {0, 0, 0, 0}}},{17894, {{30249, 1, 0, 0},{30249, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{17895, {{30252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17896, {{30253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17897, {{30254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17898, {{30255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17899, {{30256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17900, {{30257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17901, {{30258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17902, {{30259, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17903, {{30260, 1, 0, 0},{30260, 1, 1, 17}, {0, 0, 0, 0}}},{17903, {{30260, 1, 0, 0},{30260, 1, 1, 17}, {0, 0, 0, 0}}},{17904, {{30260, 1, 0, 0},{30260, 1, 1, 18}, {0, 0, 0, 0}}},{17904, {{30260, 1, 0, 0},{30260, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17905, {{30261, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17906, {{30262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17907, {{30263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17908, {{30264, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17909, {{30265, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17910, {{30266, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17911, {{30267, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17912, {{30268, 1, 0, 0},{30268, 1, 1, 17}, {0, 0, 0, 0}}},{17912, {{30268, 1, 0, 0},{30268, 1, 1, 17}, {0, 0, 0, 0}}},{17913, {{30268, 1, 0, 0},{30268, 1, 1, 18}, {0, 0, 0, 0}}},{17913, {{30268, 1, 0, 0},{30268, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17914, {{30269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{17915, {{30272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17916, {{30273, 1, 0, 0},{30273, 1, 1, 18}, {0, 0, 0, 0}}},{17916, {{30273, 1, 0, 0},{30273, 1, 1, 18}, {0, 0, 0, 0}}},{17917, {{30273, 1, 0, 0},{30273, 1, 1, 17}, {0, 0, 0, 0}}},{17917, {{30273, 1, 0, 0},{30273, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{17918, {{30274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17919, {{30275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17920, {{30276, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17921, {{30277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17922, {{30278, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17923, {{30279, 1, 0, 0},{30279, 1, 1, 17}, {0, 0, 0, 0}}},{17923, {{30279, 1, 0, 0},{30279, 1, 1, 17}, {0, 0, 0, 0}}},{17924, {{30279, 1, 0, 0},{30279, 1, 1, 18}, {0, 0, 0, 0}}},{17924, {{30279, 1, 0, 0},{30279, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17925, {{30280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17926, {{30281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17927, {{30282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17928, {{30283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17929, {{30284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{17930, {{30286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17931, {{30287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17932, {{30288, 1, 0, 0},{30288, 1, 1, 16}, {0, 0, 0, 0}}},{17932, {{30288, 1, 0, 0},{30288, 1, 1, 16}, {0, 0, 0, 0}}},{17933, {{30288, 1, 0, 0},{30288, 1, 1, 17}, {0, 0, 0, 0}}},{17933, {{30288, 1, 0, 0},{30288, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{17934, {{30289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{17935, {{30290, 1, 0, 0},{30290, 1, 1, 18}, {0, 0, 0, 0}}},{17935, {{30290, 1, 0, 0},{30290, 1, 1, 18}, {0, 0, 0, 0}}},{17936, {{30290, 1, 0, 0},{30290, 1, 1, 19}, {0, 0, 0, 0}}},{17936, {{30290, 1, 0, 0},{30290, 1, 1, 19}, {0, 0, 0, 0}}},{17937, {{30290, 1, 0, 0},{30290, 1, 1, 20}, {0, 0, 0, 0}}},{17937, {{30290, 1, 0, 0},{30290, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{17938, {{30291, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17939, {{30292, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17940, {{30293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17941, {{30294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17942, {{30295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17943, {{30296, 1, 0, 0},{30296, 1, 1, 17}, {0, 0, 0, 0}}},{17943, {{30296, 1, 0, 0},{30296, 1, 1, 17}, {0, 0, 0, 0}}},{17944, {{30296, 1, 0, 0},{30296, 1, 1, 18}, {0, 0, 0, 0}}},{17944, {{30296, 1, 0, 0},{30296, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17945, {{30297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17946, {{30298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17947, {{30299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17948, {{30300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{17949, {{30303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17950, {{30304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17951, {{30305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{17952, {{30306, 1, 0, 0},{30306, 1, 1, 17}, {0, 0, 0, 0}}},{17952, {{30306, 1, 0, 0},{30306, 1, 1, 17}, {0, 0, 0, 0}}},{17953, {{30306, 1, 0, 0},{30306, 1, 1, 20}, {0, 0, 0, 0}}},{17953, {{30306, 1, 0, 0},{30306, 1, 1, 20}, {0, 0, 0, 0}}},{17954, {{30306, 1, 0, 0},{30306, 1, 1, 18}, {0, 0, 0, 0}}},{17954, {{30306, 1, 0, 0},{30306, 1, 1, 18}, {0, 0, 0, 0}}},{58126, {{30306, 1, 0, 0},{30306, 1, 1, 19}, {0, 0, 0, 0}}},{58126, {{30306, 1, 0, 0},{30306, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{17955, {{30308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17956, {{30309, 1, 0, 0},{30309, 1, 1, 17}, {0, 0, 0, 0}}},{17956, {{30309, 1, 0, 0},{30309, 1, 1, 17}, {0, 0, 0, 0}}},{17957, {{30309, 1, 0, 0},{30309, 1, 1, 18}, {0, 0, 0, 0}}},{17957, {{30309, 1, 0, 0},{30309, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{17958, {{30311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17960, {{30312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17961, {{30313, 1, 0, 0},{30313, 1, 1, 18}, {0, 0, 0, 0}}},{17961, {{30313, 1, 0, 0},{30313, 1, 1, 18}, {0, 0, 0, 0}}},{17962, {{30313, 1, 0, 0},{30313, 1, 1, 19}, {0, 0, 0, 0}}},{17962, {{30313, 1, 0, 0},{30313, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{17963, {{30314, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{17964, {{30316, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17965, {{30317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17966, {{30318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17967, {{30319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17968, {{30320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17969, {{30321, 1, 0, 0},{30321, 1, 1, 17}, {0, 0, 0, 0}}},{17969, {{30321, 1, 0, 0},{30321, 1, 1, 17}, {0, 0, 0, 0}}},{17970, {{30321, 1, 0, 0},{30321, 1, 1, 18}, {0, 0, 0, 0}}},{17970, {{30321, 1, 0, 0},{30321, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {8, {{17971, {{30322, 1, 0, 0},{30322, 1, 1, 18}, {0, 0, 0, 0}}},{17971, {{30322, 1, 0, 0},{30322, 1, 1, 18}, {0, 0, 0, 0}}},{17972, {{30322, 1, 0, 0},{30322, 1, 1, 19}, {0, 0, 0, 0}}},{17972, {{30322, 1, 0, 0},{30322, 1, 1, 19}, {0, 0, 0, 0}}},{17973, {{30322, 1, 0, 0},{30322, 1, 1, 20}, {0, 0, 0, 0}}},{17973, {{30322, 1, 0, 0},{30322, 1, 1, 20}, {0, 0, 0, 0}}},{17974, {{30322, 1, 0, 0},{30322, 1, 1, 21}, {0, 0, 0, 0}}},{17974, {{30322, 1, 0, 0},{30322, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{17975, {{30324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17976, {{30325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17977, {{30326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{17978, {{30328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17979, {{30329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17980, {{30330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17981, {{30331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17982, {{30332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17983, {{30333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17985, {{30334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17986, {{30335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17987, {{30336, 1, 0, 0},{30336, 1, 1, 17}, {0, 0, 0, 0}}},{17987, {{30336, 1, 0, 0},{30336, 1, 1, 17}, {0, 0, 0, 0}}},{17991, {{30336, 1, 0, 0},{30336, 1, 1, 18}, {0, 0, 0, 0}}},{17991, {{30336, 1, 0, 0},{30336, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17988, {{30337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17989, {{30338, 1, 0, 0},{30338, 1, 1, 17}, {0, 0, 0, 0}}},{17989, {{30338, 1, 0, 0},{30338, 1, 1, 17}, {0, 0, 0, 0}}},{17990, {{30338, 1, 0, 0},{30338, 1, 1, 18}, {0, 0, 0, 0}}},{17990, {{30338, 1, 0, 0},{30338, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{17992, {{30339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{17993, {{30340, 1, 0, 0},{30340, 1, 1, 18}, {0, 0, 0, 0}}},{17993, {{30340, 1, 0, 0},{30340, 1, 1, 18}, {0, 0, 0, 0}}},{17994, {{30340, 1, 0, 0},{30340, 1, 1, 19}, {0, 0, 0, 0}}},{17994, {{30340, 1, 0, 0},{30340, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{17995, {{30341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17996, {{30342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17997, {{30343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17998, {{30344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17999, {{30345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18000, {{30346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18002, {{30347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18004, {{30348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18005, {{30349, 1, 0, 0},{30349, 1, 1, 17}, {0, 0, 0, 0}}},{18005, {{30349, 1, 0, 0},{30349, 1, 1, 17}, {0, 0, 0, 0}}},{18006, {{30349, 1, 0, 0},{30349, 1, 1, 18}, {0, 0, 0, 0}}},{18006, {{30349, 1, 0, 0},{30349, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{18007, {{30350, 1, 0, 0},{30350, 1, 1, 17}, {0, 0, 0, 0}}},{18007, {{30350, 1, 0, 0},{30350, 1, 1, 17}, {0, 0, 0, 0}}},{18008, {{30350, 1, 0, 0},{30350, 1, 1, 18}, {0, 0, 0, 0}}},{18008, {{30350, 1, 0, 0},{30350, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18009, {{30351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18010, {{30352, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{18011, {{30354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18012, {{30355, 1, 0, 0},{30355, 1, 1, 19}, {0, 0, 0, 0}}},{18012, {{30355, 1, 0, 0},{30355, 1, 1, 19}, {0, 0, 0, 0}}},{18013, {{30355, 1, 0, 0},{30355, 1, 1, 18}, {0, 0, 0, 0}}},{18013, {{30355, 1, 0, 0},{30355, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18014, {{30356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18015, {{30357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18016, {{30358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18017, {{30359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18019, {{30360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18020, {{30361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18021, {{30362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18022, {{30363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18023, {{30364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18025, {{30365, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18026, {{30366, 1, 0, 0},{30366, 1, 1, 17}, {0, 0, 0, 0}}},{18026, {{30366, 1, 0, 0},{30366, 1, 1, 17}, {0, 0, 0, 0}}},{18027, {{30366, 1, 0, 0},{30366, 1, 1, 18}, {0, 0, 0, 0}}},{18027, {{30366, 1, 0, 0},{30366, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18028, {{30367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18029, {{30368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18030, {{30369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18031, {{30370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18032, {{30371, 1, 0, 0},{30371, 1, 1, 17}, {0, 0, 0, 0}}},{18032, {{30371, 1, 0, 0},{30371, 1, 1, 17}, {0, 0, 0, 0}}},{18033, {{30371, 1, 0, 0},{30371, 1, 1, 18}, {0, 0, 0, 0}}},{18033, {{30371, 1, 0, 0},{30371, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18035, {{30372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18036, {{30373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18037, {{30374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18038, {{30375, 1, 0, 0},{30375, 1, 1, 18}, {0, 0, 0, 0}}},{18038, {{30375, 1, 0, 0},{30375, 1, 1, 18}, {0, 0, 0, 0}}},{18039, {{30375, 1, 0, 0},{30375, 1, 1, 17}, {0, 0, 0, 0}}},{18039, {{30375, 1, 0, 0},{30375, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{18040, {{30376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18041, {{30377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18042, {{30378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18043, {{30379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18044, {{30380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18045, {{30381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18046, {{30382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18047, {{30383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18048, {{30384, 1, 0, 0},{30384, 1, 1, 17}, {0, 0, 0, 0}}},{18048, {{30384, 1, 0, 0},{30384, 1, 1, 17}, {0, 0, 0, 0}}},{18049, {{30384, 1, 0, 0},{30384, 1, 1, 18}, {0, 0, 0, 0}}},{18049, {{30384, 1, 0, 0},{30384, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{58140, {{30385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{18050, {{30387, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18051, {{30388, 1, 0, 0},{30388, 1, 1, 17}, {0, 0, 0, 0}}},{18051, {{30388, 1, 0, 0},{30388, 1, 1, 17}, {0, 0, 0, 0}}},{18052, {{30388, 1, 0, 0},{30388, 1, 1, 18}, {0, 0, 0, 0}}},{18052, {{30388, 1, 0, 0},{30388, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18054, {{30389, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18055, {{30390, 1, 0, 0},{30390, 1, 1, 17}, {0, 0, 0, 0}}},{18055, {{30390, 1, 0, 0},{30390, 1, 1, 17}, {0, 0, 0, 0}}},{18056, {{30390, 1, 0, 0},{30390, 1, 1, 18}, {0, 0, 0, 0}}},{18056, {{30390, 1, 0, 0},{30390, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{18057, {{30391, 1, 0, 0},{30391, 1, 1, 17}, {0, 0, 0, 0}}},{18057, {{30391, 1, 0, 0},{30391, 1, 1, 17}, {0, 0, 0, 0}}},{18058, {{30391, 1, 0, 0},{30391, 1, 1, 18}, {0, 0, 0, 0}}},{18058, {{30391, 1, 0, 0},{30391, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18059, {{30392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18060, {{30393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18061, {{30394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18062, {{30395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{18063, {{30397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18064, {{30398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18065, {{30399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18066, {{30400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18067, {{30401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18068, {{30402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18069, {{30403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18070, {{30404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18071, {{30405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18072, {{30406, 1, 0, 0},{30406, 1, 1, 18}, {0, 0, 0, 0}}},{18072, {{30406, 1, 0, 0},{30406, 1, 1, 18}, {0, 0, 0, 0}}},{18073, {{30406, 1, 0, 0},{30406, 1, 1, 19}, {0, 0, 0, 0}}},{18073, {{30406, 1, 0, 0},{30406, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{18075, {{30407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18076, {{30408, 1, 0, 0},{30408, 1, 1, 17}, {0, 0, 0, 0}}},{18076, {{30408, 1, 0, 0},{30408, 1, 1, 17}, {0, 0, 0, 0}}},{58144, {{30408, 1, 0, 0},{30408, 1, 1, 18}, {0, 0, 0, 0}}},{58144, {{30408, 1, 0, 0},{30408, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18078, {{30409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {9, {{18079, {{30410, 1, 0, 0},{30410, 1, 1, 18}, {0, 0, 0, 0}}},{18079, {{30410, 1, 0, 0},{30410, 1, 1, 18}, {0, 0, 0, 0}}},{30209, {{30410, 1, 0, 0},{30410, 1, 1, 0},{30410, 1, 1, 19},{64023, 1, 0, 0}, {0, 0, 0, 0}}},{30209, {{30410, 1, 0, 0},{30410, 1, 1, 0},{30410, 1, 1, 19},{64023, 1, 0, 0}, {0, 0, 0, 0}}},{30209, {{30410, 1, 0, 0},{30410, 1, 1, 0},{30410, 1, 1, 19},{64023, 1, 0, 0}, {0, 0, 0, 0}}},{58674, {{30410, 1, 0, 0},{30410, 1, 1, 20}, {0, 0, 0, 0}}},{58674, {{30410, 1, 0, 0},{30410, 1, 1, 20}, {0, 0, 0, 0}}},{59934, {{30410, 1, 0, 0},{30410, 1, 1, 21}, {0, 0, 0, 0}}},{59934, {{30410, 1, 0, 0},{30410, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{18080, {{30411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18081, {{30412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18082, {{30413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18083, {{30414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{18084, {{30418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18085, {{30419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18086, {{30420, 1, 0, 0},{30420, 1, 1, 18}, {0, 0, 0, 0}}},{18086, {{30420, 1, 0, 0},{30420, 1, 1, 18}, {0, 0, 0, 0}}},{18087, {{30420, 1, 0, 0},{30420, 1, 1, 19}, {0, 0, 0, 0}}},{18087, {{30420, 1, 0, 0},{30420, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{18088, {{30421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18089, {{30422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18090, {{30423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{18091, {{30425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18092, {{30426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18093, {{30427, 1, 0, 0},{30427, 1, 1, 18}, {0, 0, 0, 0}}},{18093, {{30427, 1, 0, 0},{30427, 1, 1, 18}, {0, 0, 0, 0}}},{18094, {{30427, 1, 0, 0},{30427, 1, 1, 19}, {0, 0, 0, 0}}},{18094, {{30427, 1, 0, 0},{30427, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{18095, {{30428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18096, {{30429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18097, {{30430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18098, {{30431, 1, 0, 0},{30431, 1, 1, 18}, {0, 0, 0, 0}}},{18098, {{30431, 1, 0, 0},{30431, 1, 1, 18}, {0, 0, 0, 0}}},{18099, {{30431, 1, 0, 0},{30431, 1, 1, 19}, {0, 0, 0, 0}}},{18099, {{30431, 1, 0, 0},{30431, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{18100, {{30432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18101, {{30433, 1, 0, 0},{30433, 1, 1, 17}, {0, 0, 0, 0}}},{18101, {{30433, 1, 0, 0},{30433, 1, 1, 17}, {0, 0, 0, 0}}},{58149, {{30433, 1, 0, 0},{30433, 1, 1, 18}, {0, 0, 0, 0}}},{58149, {{30433, 1, 0, 0},{30433, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18102, {{30434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18103, {{30435, 1, 0, 0},{30435, 1, 1, 17}, {0, 0, 0, 0}}},{18103, {{30435, 1, 0, 0},{30435, 1, 1, 17}, {0, 0, 0, 0}}},{18104, {{30435, 1, 0, 0},{30435, 1, 1, 18}, {0, 0, 0, 0}}},{18104, {{30435, 1, 0, 0},{30435, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18105, {{30436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18106, {{30437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18108, {{30438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18109, {{30439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18110, {{30440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18111, {{30441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18112, {{30442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18113, {{30443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18114, {{30444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18115, {{30445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18116, {{30446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18117, {{30447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18118, {{30448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18119, {{30449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18120, {{30450, 1, 0, 0},{30450, 1, 1, 19}, {0, 0, 0, 0}}},{18120, {{30450, 1, 0, 0},{30450, 1, 1, 19}, {0, 0, 0, 0}}},{18121, {{30450, 1, 0, 0},{30450, 1, 1, 18}, {0, 0, 0, 0}}},{18121, {{30450, 1, 0, 0},{30450, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18122, {{30451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {12, {{18123, {{30452, 1, 0, 0},{30452, 1, 1, 18}, {0, 0, 0, 0}}},{18123, {{30452, 1, 0, 0},{30452, 1, 1, 18}, {0, 0, 0, 0}}},{18124, {{30452, 1, 0, 0},{30452, 1, 1, 19}, {0, 0, 0, 0}}},{18124, {{30452, 1, 0, 0},{30452, 1, 1, 19}, {0, 0, 0, 0}}},{56839, {{30452, 1, 0, 0},{30452, 1, 1, 20}, {0, 0, 0, 0}}},{56839, {{30452, 1, 0, 0},{30452, 1, 1, 20}, {0, 0, 0, 0}}},{56840, {{30452, 1, 0, 0},{30452, 1, 1, 21}, {0, 0, 0, 0}}},{56840, {{30452, 1, 0, 0},{30452, 1, 1, 21}, {0, 0, 0, 0}}},{56841, {{30452, 1, 0, 0},{30452, 1, 1, 24}, {0, 0, 0, 0}}},{56841, {{30452, 1, 0, 0},{30452, 1, 1, 24}, {0, 0, 0, 0}}},{56845, {{30452, 1, 0, 0},{30452, 1, 1, 25}, {0, 0, 0, 0}}},{56845, {{30452, 1, 0, 0},{30452, 1, 1, 25}, {0, 0, 0, 0}}}}},
    {1, {{18126, {{30453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18127, {{30454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18128, {{30455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18129, {{30456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18130, {{30457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18131, {{30458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18132, {{30459, 1, 0, 0},{30459, 1, 1, 17}, {0, 0, 0, 0}}},{18132, {{30459, 1, 0, 0},{30459, 1, 1, 17}, {0, 0, 0, 0}}},{18133, {{30459, 1, 0, 0},{30459, 1, 1, 18}, {0, 0, 0, 0}}},{18133, {{30459, 1, 0, 0},{30459, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18134, {{30460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18135, {{30461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18136, {{30462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18137, {{30463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18138, {{30464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18139, {{30465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18140, {{30466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18141, {{30467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18142, {{30468, 1, 0, 0},{30468, 1, 1, 17}, {0, 0, 0, 0}}},{18142, {{30468, 1, 0, 0},{30468, 1, 1, 17}, {0, 0, 0, 0}}},{18143, {{30468, 1, 0, 0},{30468, 1, 1, 18}, {0, 0, 0, 0}}},{18143, {{30468, 1, 0, 0},{30468, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18144, {{30469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18145, {{30470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18146, {{30471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18147, {{30472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18148, {{30473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18149, {{30474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18150, {{30475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18151, {{30476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{18152, {{30478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18153, {{30479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18154, {{30480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18155, {{30481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18156, {{30482, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18157, {{30483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18159, {{30484, 1, 0, 0},{30484, 1, 1, 17}, {0, 0, 0, 0}}},{18159, {{30484, 1, 0, 0},{30484, 1, 1, 17}, {0, 0, 0, 0}}},{18160, {{30484, 1, 0, 0},{30484, 1, 1, 18}, {0, 0, 0, 0}}},{18160, {{30484, 1, 0, 0},{30484, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18161, {{30485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18162, {{30486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18163, {{30487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18164, {{30488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18165, {{30489, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18166, {{30490, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18167, {{30491, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18168, {{30492, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18169, {{30493, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {10, {{18170, {{30494, 1, 0, 0},{30494, 1, 1, 18}, {0, 0, 0, 0}}},{18170, {{30494, 1, 0, 0},{30494, 1, 1, 18}, {0, 0, 0, 0}}},{18171, {{30494, 1, 0, 0},{30494, 1, 1, 19}, {0, 0, 0, 0}}},{18171, {{30494, 1, 0, 0},{30494, 1, 1, 19}, {0, 0, 0, 0}}},{18172, {{30494, 1, 0, 0},{30494, 1, 1, 22}, {0, 0, 0, 0}}},{18172, {{30494, 1, 0, 0},{30494, 1, 1, 22}, {0, 0, 0, 0}}},{56990, {{30494, 1, 0, 0},{30494, 1, 1, 21}, {0, 0, 0, 0}}},{56990, {{30494, 1, 0, 0},{30494, 1, 1, 21}, {0, 0, 0, 0}}},{58159, {{30494, 1, 0, 0},{30494, 1, 1, 20}, {0, 0, 0, 0}}},{58159, {{30494, 1, 0, 0},{30494, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {8, {{18173, {{30495, 1, 0, 0},{30495, 1, 1, 18}, {0, 0, 0, 0}}},{18173, {{30495, 1, 0, 0},{30495, 1, 1, 18}, {0, 0, 0, 0}}},{18174, {{30495, 1, 0, 0},{30495, 1, 1, 21}, {0, 0, 0, 0}}},{18174, {{30495, 1, 0, 0},{30495, 1, 1, 21}, {0, 0, 0, 0}}},{18175, {{30495, 1, 0, 0},{30495, 1, 1, 23}, {0, 0, 0, 0}}},{18175, {{30495, 1, 0, 0},{30495, 1, 1, 23}, {0, 0, 0, 0}}},{56991, {{30495, 1, 0, 0},{30495, 1, 1, 19}, {0, 0, 0, 0}}},{56991, {{30495, 1, 0, 0},{30495, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{18178, {{30496, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18179, {{30497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18180, {{30498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18181, {{30499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18182, {{30500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18183, {{30501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18184, {{30502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18185, {{30503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18186, {{30504, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18187, {{30505, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18188, {{30506, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18189, {{30507, 1, 0, 0},{30507, 1, 1, 16}, {0, 0, 0, 0}}},{18189, {{30507, 1, 0, 0},{30507, 1, 1, 16}, {0, 0, 0, 0}}},{18190, {{30507, 1, 0, 0},{30507, 1, 1, 17}, {0, 0, 0, 0}}},{18190, {{30507, 1, 0, 0},{30507, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{18191, {{30509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18192, {{30510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18193, {{30511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18194, {{30512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18195, {{30513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18196, {{30514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18197, {{30515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18198, {{30516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18199, {{30517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18200, {{30518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18201, {{30519, 1, 0, 0},{30519, 1, 1, 18}, {0, 0, 0, 0}}},{18201, {{30519, 1, 0, 0},{30519, 1, 1, 18}, {0, 0, 0, 0}}},{18202, {{30519, 1, 0, 0},{30519, 1, 1, 19}, {0, 0, 0, 0}}},{18202, {{30519, 1, 0, 0},{30519, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{18203, {{30520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18204, {{30521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18206, {{30522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18207, {{30523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18208, {{30524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18209, {{30525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{18210, {{30526, 1, 0, 0},{30526, 1, 1, 17}, {0, 0, 0, 0}}},{18210, {{30526, 1, 0, 0},{30526, 1, 1, 17}, {0, 0, 0, 0}}},{58166, {{30526, 1, 0, 0},{30526, 1, 1, 18}, {0, 0, 0, 0}}},{58166, {{30526, 1, 0, 0},{30526, 1, 1, 18}, {0, 0, 0, 0}}},{58167, {{30526, 1, 0, 0},{30526, 1, 1, 19}, {0, 0, 0, 0}}},{58167, {{30526, 1, 0, 0},{30526, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{18211, {{30527, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18212, {{30528, 1, 0, 0},{30528, 1, 1, 18}, {0, 0, 0, 0}}},{18212, {{30528, 1, 0, 0},{30528, 1, 1, 18}, {0, 0, 0, 0}}},{18213, {{30528, 1, 0, 0},{30528, 1, 1, 17}, {0, 0, 0, 0}}},{18213, {{30528, 1, 0, 0},{30528, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{18214, {{30529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18215, {{30530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18216, {{30531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18217, {{30532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18218, {{30533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18219, {{30534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18220, {{30535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18221, {{30536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18222, {{30537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18223, {{30538, 1, 0, 0},{30538, 1, 1, 18}, {0, 0, 0, 0}}},{18223, {{30538, 1, 0, 0},{30538, 1, 1, 18}, {0, 0, 0, 0}}},{18224, {{30538, 1, 0, 0},{30538, 1, 1, 17}, {0, 0, 0, 0}}},{18224, {{30538, 1, 0, 0},{30538, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{18225, {{30539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18226, {{30540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18227, {{30541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18228, {{30542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18229, {{30543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{18230, {{30546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18231, {{30547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18232, {{30548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18233, {{30549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18234, {{30550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18236, {{30551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18237, {{30552, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18238, {{30553, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18239, {{30554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18240, {{30555, 1, 0, 0},{30555, 1, 1, 17}, {0, 0, 0, 0}}},{18240, {{30555, 1, 0, 0},{30555, 1, 1, 17}, {0, 0, 0, 0}}},{18241, {{30555, 1, 0, 0},{30555, 1, 1, 18}, {0, 0, 0, 0}}},{18241, {{30555, 1, 0, 0},{30555, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18242, {{30556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18243, {{30557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18244, {{30558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18245, {{30559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18246, {{30560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18247, {{30561, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18248, {{30562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18249, {{30563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18250, {{30564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18251, {{30565, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18252, {{30566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18253, {{30567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18254, {{30568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18255, {{30569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18256, {{30570, 1, 0, 0},{30570, 1, 1, 17}, {0, 0, 0, 0}}},{18256, {{30570, 1, 0, 0},{30570, 1, 1, 17}, {0, 0, 0, 0}}},{58176, {{30570, 1, 0, 0},{30570, 1, 1, 18}, {0, 0, 0, 0}}},{58176, {{30570, 1, 0, 0},{30570, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18257, {{30571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18258, {{30572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18259, {{30573, 1, 0, 0},{30573, 1, 1, 16}, {0, 0, 0, 0}}},{18259, {{30573, 1, 0, 0},{30573, 1, 1, 16}, {0, 0, 0, 0}}},{18260, {{30573, 1, 0, 0},{30573, 1, 1, 17}, {0, 0, 0, 0}}},{18260, {{30573, 1, 0, 0},{30573, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{18261, {{30574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18262, {{30575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18263, {{30576, 1, 0, 0},{30576, 1, 1, 17}, {0, 0, 0, 0}}},{18263, {{30576, 1, 0, 0},{30576, 1, 1, 17}, {0, 0, 0, 0}}},{18264, {{30576, 1, 0, 0},{30576, 1, 1, 18}, {0, 0, 0, 0}}},{18264, {{30576, 1, 0, 0},{30576, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18265, {{30577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18266, {{30578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18267, {{30579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18268, {{30580, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18269, {{30581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18270, {{30582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18271, {{30583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18272, {{30584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18273, {{30585, 1, 0, 0},{30585, 1, 1, 17}, {0, 0, 0, 0}}},{18273, {{30585, 1, 0, 0},{30585, 1, 1, 17}, {0, 0, 0, 0}}},{18274, {{30585, 1, 0, 0},{30585, 1, 1, 18}, {0, 0, 0, 0}}},{18274, {{30585, 1, 0, 0},{30585, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18275, {{30586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18276, {{30587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18277, {{30588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18278, {{30589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18279, {{30590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18280, {{30591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18281, {{30592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18282, {{30593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18283, {{30594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18284, {{30595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18285, {{30596, 1, 0, 0},{30596, 1, 1, 17}, {0, 0, 0, 0}}},{18285, {{30596, 1, 0, 0},{30596, 1, 1, 17}, {0, 0, 0, 0}}},{18286, {{30596, 1, 0, 0},{30596, 1, 1, 18}, {0, 0, 0, 0}}},{18286, {{30596, 1, 0, 0},{30596, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18287, {{30597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{18288, {{30599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18290, {{30600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18291, {{30601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18292, {{30602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18293, {{30603, 1, 0, 0},{30603, 1, 1, 17}, {0, 0, 0, 0}}},{18293, {{30603, 1, 0, 0},{30603, 1, 1, 17}, {0, 0, 0, 0}}},{18294, {{30603, 1, 0, 0},{30603, 1, 1, 18}, {0, 0, 0, 0}}},{18294, {{30603, 1, 0, 0},{30603, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18295, {{30604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18296, {{30605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{18297, {{30606, 1, 0, 0},{30606, 1, 1, 17}, {0, 0, 0, 0}}},{18297, {{30606, 1, 0, 0},{30606, 1, 1, 17}, {0, 0, 0, 0}}},{18298, {{30606, 1, 0, 0},{30606, 1, 1, 18}, {0, 0, 0, 0}}},{18298, {{30606, 1, 0, 0},{30606, 1, 1, 18}, {0, 0, 0, 0}}},{18299, {{30606, 1, 0, 0},{30606, 1, 1, 19}, {0, 0, 0, 0}}},{18299, {{30606, 1, 0, 0},{30606, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{18300, {{30607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18301, {{30608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18302, {{30609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{18303, {{30612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18304, {{30613, 1, 0, 0},{30613, 1, 1, 17}, {0, 0, 0, 0}}},{18304, {{30613, 1, 0, 0},{30613, 1, 1, 17}, {0, 0, 0, 0}}},{18305, {{30613, 1, 0, 0},{30613, 1, 1, 18}, {0, 0, 0, 0}}},{18305, {{30613, 1, 0, 0},{30613, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18306, {{30614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18307, {{30615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18308, {{30616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18309, {{30617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18310, {{30618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18311, {{30619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18312, {{30620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18313, {{30621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18314, {{30622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18315, {{30623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18316, {{30624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18317, {{30625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18318, {{30626, 1, 0, 0},{30626, 1, 1, 19}, {0, 0, 0, 0}}},{18318, {{30626, 1, 0, 0},{30626, 1, 1, 19}, {0, 0, 0, 0}}},{18319, {{30626, 1, 0, 0},{30626, 1, 1, 18}, {0, 0, 0, 0}}},{18319, {{30626, 1, 0, 0},{30626, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{18320, {{30628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18321, {{30629, 1, 0, 0},{30629, 1, 1, 19}, {0, 0, 0, 0}}},{18321, {{30629, 1, 0, 0},{30629, 1, 1, 19}, {0, 0, 0, 0}}},{18322, {{30629, 1, 0, 0},{30629, 1, 1, 18}, {0, 0, 0, 0}}},{18322, {{30629, 1, 0, 0},{30629, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18323, {{30630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18325, {{30631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18326, {{30632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18327, {{30633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18328, {{30634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18329, {{30635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{18330, {{30636, 1, 0, 0},{30636, 1, 1, 19}, {0, 0, 0, 0}}},{18330, {{30636, 1, 0, 0},{30636, 1, 1, 19}, {0, 0, 0, 0}}},{18331, {{30636, 1, 0, 0},{30636, 1, 1, 18}, {0, 0, 0, 0}}},{18331, {{30636, 1, 0, 0},{30636, 1, 1, 18}, {0, 0, 0, 0}}},{18332, {{30636, 1, 0, 0},{30636, 1, 1, 20}, {0, 0, 0, 0}}},{18332, {{30636, 1, 0, 0},{30636, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{18333, {{30637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18334, {{30638, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18335, {{30639, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18336, {{30640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18337, {{30641, 1, 0, 0},{30641, 1, 1, 17}, {0, 0, 0, 0}}},{18337, {{30641, 1, 0, 0},{30641, 1, 1, 17}, {0, 0, 0, 0}}},{18338, {{30641, 1, 0, 0},{30641, 1, 1, 18}, {0, 0, 0, 0}}},{18338, {{30641, 1, 0, 0},{30641, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18339, {{30642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18340, {{30643, 1, 0, 0},{30643, 1, 1, 17}, {0, 0, 0, 0}}},{18340, {{30643, 1, 0, 0},{30643, 1, 1, 17}, {0, 0, 0, 0}}},{18341, {{30643, 1, 0, 0},{30643, 1, 1, 18}, {0, 0, 0, 0}}},{18341, {{30643, 1, 0, 0},{30643, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18342, {{30644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18343, {{30645, 1, 0, 0},{30645, 1, 1, 18}, {0, 0, 0, 0}}},{18343, {{30645, 1, 0, 0},{30645, 1, 1, 18}, {0, 0, 0, 0}}},{18344, {{30645, 1, 0, 0},{30645, 1, 1, 17}, {0, 0, 0, 0}}},{18344, {{30645, 1, 0, 0},{30645, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{18345, {{30646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{18346, {{30648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18347, {{30649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18348, {{30650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18349, {{30651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18350, {{30652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18351, {{30653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18352, {{30654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18353, {{30655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18354, {{30656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18355, {{30657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{18356, {{30659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18357, {{30660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18358, {{30661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18360, {{30662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{18362, {{30663, 1, 0, 0},{30663, 1, 1, 17}, {0, 0, 0, 0}}},{18362, {{30663, 1, 0, 0},{30663, 1, 1, 17}, {0, 0, 0, 0}}},{18363, {{30663, 1, 0, 0},{30663, 1, 1, 18}, {0, 0, 0, 0}}},{18363, {{30663, 1, 0, 0},{30663, 1, 1, 18}, {0, 0, 0, 0}}},{18364, {{30663, 1, 0, 0},{30663, 1, 1, 19}, {0, 0, 0, 0}}},{18364, {{30663, 1, 0, 0},{30663, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{18365, {{30665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18366, {{30666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{18367, {{30668, 1, 0, 0},{30668, 1, 1, 16}, {0, 0, 0, 0}}},{18367, {{30668, 1, 0, 0},{30668, 1, 1, 16}, {0, 0, 0, 0}}},{18368, {{30668, 1, 0, 0},{30668, 1, 1, 17}, {0, 0, 0, 0}}},{18368, {{30668, 1, 0, 0},{30668, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{18370, {{30669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18371, {{30670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18372, {{30671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18373, {{30672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18374, {{30673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18375, {{30674, 1, 0, 0},{30674, 1, 1, 18}, {0, 0, 0, 0}}},{18375, {{30674, 1, 0, 0},{30674, 1, 1, 18}, {0, 0, 0, 0}}},{18376, {{30674, 1, 0, 0},{30674, 1, 1, 17}, {0, 0, 0, 0}}},{18376, {{30674, 1, 0, 0},{30674, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{18377, {{30675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18378, {{30676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18380, {{30677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18381, {{30678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18382, {{30679, 1, 0, 0},{30679, 1, 1, 17}, {0, 0, 0, 0}}},{18382, {{30679, 1, 0, 0},{30679, 1, 1, 17}, {0, 0, 0, 0}}},{18383, {{30679, 1, 0, 0},{30679, 1, 1, 18}, {0, 0, 0, 0}}},{18383, {{30679, 1, 0, 0},{30679, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18384, {{30680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18385, {{30681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18386, {{30682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18387, {{30683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18388, {{30684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18389, {{30685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18390, {{30686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18391, {{30687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18392, {{30688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18393, {{30689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18394, {{30690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18395, {{30691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18396, {{30692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18397, {{30693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18398, {{30694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18399, {{30695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18400, {{30696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18401, {{30697, 1, 0, 0},{30697, 1, 1, 18}, {0, 0, 0, 0}}},{18401, {{30697, 1, 0, 0},{30697, 1, 1, 18}, {0, 0, 0, 0}}},{18402, {{30697, 1, 0, 0},{30697, 1, 1, 19}, {0, 0, 0, 0}}},{18402, {{30697, 1, 0, 0},{30697, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{18403, {{30698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{18404, {{30700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18405, {{30701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18406, {{30702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18407, {{30703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18408, {{30704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18409, {{30705, 1, 0, 0},{30705, 1, 1, 17}, {0, 0, 0, 0}}},{18409, {{30705, 1, 0, 0},{30705, 1, 1, 17}, {0, 0, 0, 0}}},{18410, {{30705, 1, 0, 0},{30705, 1, 1, 18}, {0, 0, 0, 0}}},{18410, {{30705, 1, 0, 0},{30705, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18411, {{30706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18412, {{30707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18414, {{30708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18415, {{30709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{18416, {{30711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18417, {{30712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18418, {{30713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18419, {{30714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18420, {{30715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18421, {{30716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18422, {{30717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{18423, {{30722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18425, {{30723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18426, {{30724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18427, {{30725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18428, {{30726, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18429, {{30727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18430, {{30728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18431, {{30729, 1, 0, 0},{30729, 1, 1, 17}, {0, 0, 0, 0}}},{18431, {{30729, 1, 0, 0},{30729, 1, 1, 17}, {0, 0, 0, 0}}},{18432, {{30729, 1, 0, 0},{30729, 1, 1, 18}, {0, 0, 0, 0}}},{18432, {{30729, 1, 0, 0},{30729, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18433, {{30730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18434, {{30731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18435, {{30732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18436, {{30733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18437, {{30734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18438, {{30735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18439, {{30736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18441, {{30737, 1, 0, 0},{30737, 1, 1, 18}, {0, 0, 0, 0}}},{18441, {{30737, 1, 0, 0},{30737, 1, 1, 18}, {0, 0, 0, 0}}},{18442, {{30737, 1, 0, 0},{30737, 1, 1, 17}, {0, 0, 0, 0}}},{18442, {{30737, 1, 0, 0},{30737, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{18443, {{30738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18444, {{30739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18445, {{30740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18446, {{30741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{18447, {{30745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{18448, {{30749, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18449, {{30750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18450, {{30751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18451, {{30752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18452, {{30753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18453, {{30754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18454, {{30755, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18455, {{30756, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18456, {{30757, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18457, {{30758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18458, {{30759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18459, {{30760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18460, {{30761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18461, {{30762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18462, {{30763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18463, {{30764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18464, {{30765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18465, {{30766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18466, {{30767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18467, {{30768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18468, {{30769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18470, {{30770, 1, 0, 0},{30770, 1, 1, 19}, {0, 0, 0, 0}}},{18470, {{30770, 1, 0, 0},{30770, 1, 1, 19}, {0, 0, 0, 0}}},{18471, {{30770, 1, 0, 0},{30770, 1, 1, 18}, {0, 0, 0, 0}}},{18471, {{30770, 1, 0, 0},{30770, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18472, {{30771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18473, {{30772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18474, {{30773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18475, {{30774, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18476, {{30775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18477, {{30776, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{18478, {{30778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{18479, {{30783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{18480, {{30786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18482, {{30787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18483, {{30788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18484, {{30789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18485, {{30790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18486, {{30791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18487, {{30792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18488, {{30793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18489, {{30794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18490, {{30795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18491, {{30796, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18492, {{30797, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18493, {{30798, 1, 0, 0},{30798, 1, 1, 17}, {0, 0, 0, 0}}},{18493, {{30798, 1, 0, 0},{30798, 1, 1, 17}, {0, 0, 0, 0}}},{18494, {{30798, 1, 0, 0},{30798, 1, 1, 18}, {0, 0, 0, 0}}},{18494, {{30798, 1, 0, 0},{30798, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{18495, {{30799, 1, 0, 0},{30799, 1, 1, 17}, {0, 0, 0, 0}}},{18495, {{30799, 1, 0, 0},{30799, 1, 1, 17}, {0, 0, 0, 0}}},{18496, {{30799, 1, 0, 0},{30799, 1, 1, 18}, {0, 0, 0, 0}}},{18496, {{30799, 1, 0, 0},{30799, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18497, {{30800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18498, {{30801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18499, {{30802, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{18501, {{30812, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18502, {{30813, 1, 0, 0},{30813, 1, 1, 19}, {0, 0, 0, 0}}},{18502, {{30813, 1, 0, 0},{30813, 1, 1, 19}, {0, 0, 0, 0}}},{18503, {{30813, 1, 0, 0},{30813, 1, 1, 18}, {0, 0, 0, 0}}},{18503, {{30813, 1, 0, 0},{30813, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18504, {{30814, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18506, {{30815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18507, {{30816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18508, {{30817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18509, {{30818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18510, {{30819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18511, {{30820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18512, {{30821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18513, {{30822, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18514, {{30823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18515, {{30824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18516, {{30825, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18517, {{30826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18518, {{30827, 1, 0, 0},{30827, 1, 1, 18}, {0, 0, 0, 0}}},{18518, {{30827, 1, 0, 0},{30827, 1, 1, 18}, {0, 0, 0, 0}}},{18519, {{30827, 1, 0, 0},{30827, 1, 1, 17}, {0, 0, 0, 0}}},{18519, {{30827, 1, 0, 0},{30827, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{18520, {{30828, 1, 0, 0},{30828, 1, 1, 18}, {0, 0, 0, 0}}},{18520, {{30828, 1, 0, 0},{30828, 1, 1, 18}, {0, 0, 0, 0}}},{18521, {{30828, 1, 0, 0},{30828, 1, 1, 19}, {0, 0, 0, 0}}},{18521, {{30828, 1, 0, 0},{30828, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{18522, {{30829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18523, {{30830, 1, 0, 0},{30830, 1, 1, 17}, {0, 0, 0, 0}}},{18523, {{30830, 1, 0, 0},{30830, 1, 1, 17}, {0, 0, 0, 0}}},{18524, {{30830, 1, 0, 0},{30830, 1, 1, 18}, {0, 0, 0, 0}}},{18524, {{30830, 1, 0, 0},{30830, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18525, {{30831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18526, {{30832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18527, {{30833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18528, {{30834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{18529, {{30836, 1, 0, 0},{30836, 1, 1, 17}, {0, 0, 0, 0}}},{18529, {{30836, 1, 0, 0},{30836, 1, 1, 17}, {0, 0, 0, 0}}},{18530, {{30836, 1, 0, 0},{30836, 1, 1, 18}, {0, 0, 0, 0}}},{18530, {{30836, 1, 0, 0},{30836, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{18531, {{30840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18532, {{30841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18533, {{30842, 1, 0, 0},{30842, 1, 1, 17}, {0, 0, 0, 0}}},{18533, {{30842, 1, 0, 0},{30842, 1, 1, 17}, {0, 0, 0, 0}}},{18534, {{30842, 1, 0, 0},{30842, 1, 1, 18}, {0, 0, 0, 0}}},{18534, {{30842, 1, 0, 0},{30842, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{18535, {{30843, 1, 0, 0},{30843, 1, 1, 16}, {0, 0, 0, 0}}},{18535, {{30843, 1, 0, 0},{30843, 1, 1, 16}, {0, 0, 0, 0}}},{18536, {{30843, 1, 0, 0},{30843, 1, 1, 17}, {0, 0, 0, 0}}},{18536, {{30843, 1, 0, 0},{30843, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{18537, {{30844, 1, 0, 0},{30844, 1, 1, 19}, {0, 0, 0, 0}}},{18537, {{30844, 1, 0, 0},{30844, 1, 1, 19}, {0, 0, 0, 0}}},{18538, {{30844, 1, 0, 0},{30844, 1, 1, 18}, {0, 0, 0, 0}}},{18538, {{30844, 1, 0, 0},{30844, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18539, {{30845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18540, {{30846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18541, {{30847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18542, {{30848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18543, {{30849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18544, {{30850, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18545, {{30851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18547, {{30852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18548, {{30853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18549, {{30854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18550, {{30855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18551, {{30856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18552, {{30857, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18553, {{30858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18554, {{30859, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18555, {{30860, 1, 0, 0},{30860, 1, 1, 17}, {0, 0, 0, 0}}},{18555, {{30860, 1, 0, 0},{30860, 1, 1, 17}, {0, 0, 0, 0}}},{18556, {{30860, 1, 0, 0},{30860, 1, 1, 18}, {0, 0, 0, 0}}},{18556, {{30860, 1, 0, 0},{30860, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18557, {{30861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18558, {{30862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18559, {{30863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18560, {{30864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{18561, {{30865, 1, 0, 0},{30865, 1, 1, 18}, {0, 0, 0, 0}}},{18561, {{30865, 1, 0, 0},{30865, 1, 1, 18}, {0, 0, 0, 0}}},{30272, {{30865, 1, 0, 0},{30865, 1, 1, 0},{30865, 1, 1, 19},{64075, 1, 0, 0}, {0, 0, 0, 0}}},{30272, {{30865, 1, 0, 0},{30865, 1, 1, 0},{30865, 1, 1, 19},{64075, 1, 0, 0}, {0, 0, 0, 0}}},{30272, {{30865, 1, 0, 0},{30865, 1, 1, 0},{30865, 1, 1, 19},{64075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18562, {{30866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18563, {{30867, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18564, {{30868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18565, {{30869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18566, {{30870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18567, {{30871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18568, {{30872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18569, {{30873, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18570, {{30874, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{18571, {{30877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18572, {{30878, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18573, {{30879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18574, {{30880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18575, {{30881, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18576, {{30882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18577, {{30883, 1, 0, 0},{30883, 1, 1, 17}, {0, 0, 0, 0}}},{18577, {{30883, 1, 0, 0},{30883, 1, 1, 17}, {0, 0, 0, 0}}},{18578, {{30883, 1, 0, 0},{30883, 1, 1, 18}, {0, 0, 0, 0}}},{18578, {{30883, 1, 0, 0},{30883, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{18579, {{30884, 1, 0, 0},{30884, 1, 1, 17}, {0, 0, 0, 0}}},{18579, {{30884, 1, 0, 0},{30884, 1, 1, 17}, {0, 0, 0, 0}}},{18580, {{30884, 1, 0, 0},{30884, 1, 1, 18}, {0, 0, 0, 0}}},{18580, {{30884, 1, 0, 0},{30884, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18581, {{30885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18582, {{30886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18583, {{30887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18584, {{30888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18585, {{30889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18586, {{30890, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18587, {{30891, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18588, {{30892, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18589, {{30893, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18590, {{30894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18591, {{30895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18592, {{30896, 1, 0, 0},{30896, 1, 1, 18}, {0, 0, 0, 0}}},{18592, {{30896, 1, 0, 0},{30896, 1, 1, 18}, {0, 0, 0, 0}}},{18593, {{30896, 1, 0, 0},{30896, 1, 1, 17}, {0, 0, 0, 0}}},{18593, {{30896, 1, 0, 0},{30896, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{18594, {{30897, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18595, {{30898, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18596, {{30899, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{18597, {{30901, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{18598, {{30906, 1, 0, 0},{30906, 1, 1, 17}, {0, 0, 0, 0}}},{18598, {{30906, 1, 0, 0},{30906, 1, 1, 17}, {0, 0, 0, 0}}},{18599, {{30906, 1, 0, 0},{30906, 1, 1, 18}, {0, 0, 0, 0}}},{18599, {{30906, 1, 0, 0},{30906, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18600, {{30907, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18602, {{30908, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18603, {{30909, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18604, {{30910, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18605, {{30911, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18606, {{30912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18607, {{30913, 1, 0, 0},{30913, 1, 1, 17}, {0, 0, 0, 0}}},{18607, {{30913, 1, 0, 0},{30913, 1, 1, 17}, {0, 0, 0, 0}}},{18608, {{30913, 1, 0, 0},{30913, 1, 1, 18}, {0, 0, 0, 0}}},{18608, {{30913, 1, 0, 0},{30913, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18609, {{30914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18611, {{30915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18612, {{30916, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18614, {{30917, 1, 0, 0},{30917, 1, 1, 17}, {0, 0, 0, 0}}},{18614, {{30917, 1, 0, 0},{30917, 1, 1, 17}, {0, 0, 0, 0}}},{18615, {{30917, 1, 0, 0},{30917, 1, 1, 18}, {0, 0, 0, 0}}},{18615, {{30917, 1, 0, 0},{30917, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18616, {{30918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18617, {{30919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18618, {{30920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18619, {{30921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18620, {{30922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18621, {{30923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18622, {{30924, 1, 0, 0},{30924, 1, 1, 17}, {0, 0, 0, 0}}},{18622, {{30924, 1, 0, 0},{30924, 1, 1, 17}, {0, 0, 0, 0}}},{18623, {{30924, 1, 0, 0},{30924, 1, 1, 18}, {0, 0, 0, 0}}},{18623, {{30924, 1, 0, 0},{30924, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{18624, {{30925, 1, 0, 0},{30925, 1, 1, 16}, {0, 0, 0, 0}}},{18624, {{30925, 1, 0, 0},{30925, 1, 1, 16}, {0, 0, 0, 0}}},{18625, {{30925, 1, 0, 0},{30925, 1, 1, 17}, {0, 0, 0, 0}}},{18625, {{30925, 1, 0, 0},{30925, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{18626, {{30926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18627, {{30927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18629, {{30928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18630, {{30929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18631, {{30930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18632, {{30931, 1, 0, 0},{30931, 1, 1, 18}, {0, 0, 0, 0}}},{18632, {{30931, 1, 0, 0},{30931, 1, 1, 18}, {0, 0, 0, 0}}},{18633, {{30931, 1, 0, 0},{30931, 1, 1, 17}, {0, 0, 0, 0}}},{18633, {{30931, 1, 0, 0},{30931, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {8, {{18634, {{30932, 1, 0, 0},{30932, 1, 1, 19}, {0, 0, 0, 0}}},{18634, {{30932, 1, 0, 0},{30932, 1, 1, 19}, {0, 0, 0, 0}}},{18635, {{30932, 1, 0, 0},{30932, 1, 1, 18}, {0, 0, 0, 0}}},{18635, {{30932, 1, 0, 0},{30932, 1, 1, 18}, {0, 0, 0, 0}}},{18636, {{30932, 1, 0, 0},{30932, 1, 1, 20}, {0, 0, 0, 0}}},{18636, {{30932, 1, 0, 0},{30932, 1, 1, 20}, {0, 0, 0, 0}}},{18637, {{30932, 1, 0, 0},{30932, 1, 1, 21}, {0, 0, 0, 0}}},{18637, {{30932, 1, 0, 0},{30932, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{18638, {{30933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18639, {{30934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{18640, {{30938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18641, {{30939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18642, {{30940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18643, {{30941, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18644, {{30942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18645, {{30943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18646, {{30944, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18647, {{30945, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18648, {{30946, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18649, {{30947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18650, {{30948, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18651, {{30949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18652, {{30950, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18653, {{30951, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18654, {{30952, 1, 0, 0},{30952, 1, 1, 19}, {0, 0, 0, 0}}},{18654, {{30952, 1, 0, 0},{30952, 1, 1, 19}, {0, 0, 0, 0}}},{18655, {{30952, 1, 0, 0},{30952, 1, 1, 18}, {0, 0, 0, 0}}},{18655, {{30952, 1, 0, 0},{30952, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18656, {{30953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18657, {{30954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18658, {{30955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18659, {{30956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18660, {{30957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{18661, {{30959, 1, 0, 0},{30959, 1, 1, 18}, {0, 0, 0, 0}}},{18661, {{30959, 1, 0, 0},{30959, 1, 1, 18}, {0, 0, 0, 0}}},{18662, {{30959, 1, 0, 0},{30959, 1, 1, 19}, {0, 0, 0, 0}}},{18662, {{30959, 1, 0, 0},{30959, 1, 1, 19}, {0, 0, 0, 0}}},{18663, {{30959, 1, 0, 0},{30959, 1, 1, 20}, {0, 0, 0, 0}}},{18663, {{30959, 1, 0, 0},{30959, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{18664, {{30960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18665, {{30961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18666, {{30962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18667, {{30963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18668, {{30964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18669, {{30965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18670, {{30966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18671, {{30967, 1, 0, 0},{30967, 1, 1, 18}, {0, 0, 0, 0}}},{18671, {{30967, 1, 0, 0},{30967, 1, 1, 18}, {0, 0, 0, 0}}},{18672, {{30967, 1, 0, 0},{30967, 1, 1, 17}, {0, 0, 0, 0}}},{18672, {{30967, 1, 0, 0},{30967, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{18673, {{30968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18674, {{30969, 1, 0, 0},{30969, 1, 1, 18}, {0, 0, 0, 0}}},{18674, {{30969, 1, 0, 0},{30969, 1, 1, 18}, {0, 0, 0, 0}}},{18675, {{30969, 1, 0, 0},{30969, 1, 1, 17}, {0, 0, 0, 0}}},{18675, {{30969, 1, 0, 0},{30969, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{18676, {{30970, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18677, {{30971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18678, {{30972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18679, {{30973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18680, {{30974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18681, {{30975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18682, {{30976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18683, {{30977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{18684, {{30982, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18685, {{30983, 1, 0, 0},{30983, 1, 1, 17}, {0, 0, 0, 0}}},{18685, {{30983, 1, 0, 0},{30983, 1, 1, 17}, {0, 0, 0, 0}}},{18686, {{30983, 1, 0, 0},{30983, 1, 1, 18}, {0, 0, 0, 0}}},{18686, {{30983, 1, 0, 0},{30983, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18687, {{30984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18688, {{30985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18689, {{30986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18690, {{30987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18691, {{30988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18692, {{30989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18693, {{30990, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18694, {{30991, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18695, {{30992, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18696, {{30993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18697, {{30994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18698, {{30995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18699, {{30996, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18700, {{30997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{68039, {{30998, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18701, {{30999, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18702, {{31000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18703, {{31001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18704, {{31002, 1, 0, 0},{31002, 1, 1, 17}, {0, 0, 0, 0}}},{18704, {{31002, 1, 0, 0},{31002, 1, 1, 17}, {0, 0, 0, 0}}},{18705, {{31002, 1, 0, 0},{31002, 1, 1, 18}, {0, 0, 0, 0}}},{18705, {{31002, 1, 0, 0},{31002, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18706, {{31003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18707, {{31004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18708, {{31005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18709, {{31006, 1, 0, 0},{31006, 1, 1, 17}, {0, 0, 0, 0}}},{18709, {{31006, 1, 0, 0},{31006, 1, 1, 17}, {0, 0, 0, 0}}},{18710, {{31006, 1, 0, 0},{31006, 1, 1, 18}, {0, 0, 0, 0}}},{18710, {{31006, 1, 0, 0},{31006, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18711, {{31007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18712, {{31008, 1, 0, 0},{31008, 1, 1, 17}, {0, 0, 0, 0}}},{18712, {{31008, 1, 0, 0},{31008, 1, 1, 17}, {0, 0, 0, 0}}},{18713, {{31008, 1, 0, 0},{31008, 1, 1, 18}, {0, 0, 0, 0}}},{18713, {{31008, 1, 0, 0},{31008, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{18714, {{31011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18715, {{31012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18716, {{31013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18717, {{31014, 1, 0, 0},{31014, 1, 1, 17}, {0, 0, 0, 0}}},{18717, {{31014, 1, 0, 0},{31014, 1, 1, 17}, {0, 0, 0, 0}}},{18718, {{31014, 1, 0, 0},{31014, 1, 1, 18}, {0, 0, 0, 0}}},{18718, {{31014, 1, 0, 0},{31014, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18720, {{31015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18721, {{31016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18722, {{31017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18723, {{31018, 1, 0, 0},{31018, 1, 1, 18}, {0, 0, 0, 0}}},{18723, {{31018, 1, 0, 0},{31018, 1, 1, 18}, {0, 0, 0, 0}}},{18724, {{31018, 1, 0, 0},{31018, 1, 1, 19}, {0, 0, 0, 0}}},{18724, {{31018, 1, 0, 0},{31018, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{18725, {{31019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18726, {{31020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18727, {{31021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18728, {{31022, 1, 0, 0},{31022, 1, 1, 17}, {0, 0, 0, 0}}},{18728, {{31022, 1, 0, 0},{31022, 1, 1, 17}, {0, 0, 0, 0}}},{18729, {{31022, 1, 0, 0},{31022, 1, 1, 18}, {0, 0, 0, 0}}},{18729, {{31022, 1, 0, 0},{31022, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18730, {{31023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18731, {{31024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18732, {{31025, 1, 0, 0},{31025, 1, 1, 17}, {0, 0, 0, 0}}},{18732, {{31025, 1, 0, 0},{31025, 1, 1, 17}, {0, 0, 0, 0}}},{18733, {{31025, 1, 0, 0},{31025, 1, 1, 18}, {0, 0, 0, 0}}},{18733, {{31025, 1, 0, 0},{31025, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{18734, {{31026, 1, 0, 0},{31026, 1, 1, 16}, {0, 0, 0, 0}}},{18734, {{31026, 1, 0, 0},{31026, 1, 1, 16}, {0, 0, 0, 0}}},{18735, {{31026, 1, 0, 0},{31026, 1, 1, 17}, {0, 0, 0, 0}}},{18735, {{31026, 1, 0, 0},{31026, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {8, {{18737, {{31028, 1, 0, 0},{31028, 1, 1, 20}, {0, 0, 0, 0}}},{18737, {{31028, 1, 0, 0},{31028, 1, 1, 20}, {0, 0, 0, 0}}},{18738, {{31028, 1, 0, 0},{31028, 1, 1, 18}, {0, 0, 0, 0}}},{18738, {{31028, 1, 0, 0},{31028, 1, 1, 18}, {0, 0, 0, 0}}},{18739, {{31028, 1, 0, 0},{31028, 1, 1, 19}, {0, 0, 0, 0}}},{18739, {{31028, 1, 0, 0},{31028, 1, 1, 19}, {0, 0, 0, 0}}},{18740, {{31028, 1, 0, 0},{31028, 1, 1, 21}, {0, 0, 0, 0}}},{18740, {{31028, 1, 0, 0},{31028, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{18741, {{31029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18742, {{31030, 1, 0, 0},{31030, 1, 1, 16}, {0, 0, 0, 0}}},{18742, {{31030, 1, 0, 0},{31030, 1, 1, 16}, {0, 0, 0, 0}}},{18743, {{31030, 1, 0, 0},{31030, 1, 1, 17}, {0, 0, 0, 0}}},{18743, {{31030, 1, 0, 0},{31030, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{18744, {{31031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18745, {{31032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18746, {{31033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18747, {{31034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18749, {{31035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{18750, {{31036, 1, 0, 0},{31036, 1, 1, 18}, {0, 0, 0, 0}}},{18750, {{31036, 1, 0, 0},{31036, 1, 1, 18}, {0, 0, 0, 0}}},{30210, {{31036, 1, 0, 0},{31036, 1, 1, 0},{31036, 1, 1, 19},{64024, 1, 0, 0}, {0, 0, 0, 0}}},{30210, {{31036, 1, 0, 0},{31036, 1, 1, 0},{31036, 1, 1, 19},{64024, 1, 0, 0}, {0, 0, 0, 0}}},{30210, {{31036, 1, 0, 0},{31036, 1, 1, 0},{31036, 1, 1, 19},{64024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18751, {{31037, 1, 0, 0},{31037, 1, 1, 18}, {0, 0, 0, 0}}},{18751, {{31037, 1, 0, 0},{31037, 1, 1, 18}, {0, 0, 0, 0}}},{18752, {{31037, 1, 0, 0},{31037, 1, 1, 17}, {0, 0, 0, 0}}},{18752, {{31037, 1, 0, 0},{31037, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {7, {{18753, {{31038, 1, 0, 0},{31038, 1, 1, 18}, {0, 0, 0, 0}}},{18753, {{31038, 1, 0, 0},{31038, 1, 1, 18}, {0, 0, 0, 0}}},{30274, {{31038, 1, 0, 0},{31038, 1, 1, 0},{31038, 1, 1, 19},{64076, 1, 0, 0}, {0, 0, 0, 0}}},{30274, {{31038, 1, 0, 0},{31038, 1, 1, 0},{31038, 1, 1, 19},{64076, 1, 0, 0}, {0, 0, 0, 0}}},{30274, {{31038, 1, 0, 0},{31038, 1, 1, 0},{31038, 1, 1, 19},{64076, 1, 0, 0}, {0, 0, 0, 0}}},{58201, {{31038, 1, 0, 0},{31038, 1, 1, 20}, {0, 0, 0, 0}}},{58201, {{31038, 1, 0, 0},{31038, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{18754, {{31039, 1, 0, 0},{31039, 1, 1, 18}, {0, 0, 0, 0}}},{18754, {{31039, 1, 0, 0},{31039, 1, 1, 18}, {0, 0, 0, 0}}},{18755, {{31039, 1, 0, 0},{31039, 1, 1, 17}, {0, 0, 0, 0}}},{18755, {{31039, 1, 0, 0},{31039, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {8, {{18758, {{31040, 1, 0, 0},{31040, 1, 1, 21}, {0, 0, 0, 0}}},{18758, {{31040, 1, 0, 0},{31040, 1, 1, 21}, {0, 0, 0, 0}}},{18759, {{31040, 1, 0, 0},{31040, 1, 1, 18}, {0, 0, 0, 0}}},{18759, {{31040, 1, 0, 0},{31040, 1, 1, 18}, {0, 0, 0, 0}}},{18760, {{31040, 1, 0, 0},{31040, 1, 1, 19}, {0, 0, 0, 0}}},{18760, {{31040, 1, 0, 0},{31040, 1, 1, 19}, {0, 0, 0, 0}}},{18761, {{31040, 1, 0, 0},{31040, 1, 1, 20}, {0, 0, 0, 0}}},{18761, {{31040, 1, 0, 0},{31040, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{18762, {{31041, 1, 0, 0},{31041, 1, 1, 18}, {0, 0, 0, 0}}},{18762, {{31041, 1, 0, 0},{31041, 1, 1, 18}, {0, 0, 0, 0}}},{18763, {{31041, 1, 0, 0},{31041, 1, 1, 19}, {0, 0, 0, 0}}},{18763, {{31041, 1, 0, 0},{31041, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{18764, {{31044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18766, {{31045, 1, 0, 0},{31045, 1, 1, 18}, {0, 0, 0, 0}}},{18766, {{31045, 1, 0, 0},{31045, 1, 1, 18}, {0, 0, 0, 0}}},{18767, {{31045, 1, 0, 0},{31045, 1, 1, 17}, {0, 0, 0, 0}}},{18767, {{31045, 1, 0, 0},{31045, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{18768, {{31046, 1, 0, 0},{31046, 1, 1, 18}, {0, 0, 0, 0}}},{18768, {{31046, 1, 0, 0},{31046, 1, 1, 18}, {0, 0, 0, 0}}},{18769, {{31046, 1, 0, 0},{31046, 1, 1, 17}, {0, 0, 0, 0}}},{18769, {{31046, 1, 0, 0},{31046, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{18770, {{31047, 1, 0, 0},{31047, 1, 1, 18}, {0, 0, 0, 0}}},{18770, {{31047, 1, 0, 0},{31047, 1, 1, 18}, {0, 0, 0, 0}}},{18771, {{31047, 1, 0, 0},{31047, 1, 1, 19}, {0, 0, 0, 0}}},{18771, {{31047, 1, 0, 0},{31047, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {5, {{18772, {{31048, 1, 0, 0},{31048, 1, 1, 18}, {0, 0, 0, 0}}},{18772, {{31048, 1, 0, 0},{31048, 1, 1, 18}, {0, 0, 0, 0}}},{30276, {{31048, 1, 0, 0},{31048, 1, 1, 0},{31048, 1, 1, 19},{64078, 1, 0, 0}, {0, 0, 0, 0}}},{30276, {{31048, 1, 0, 0},{31048, 1, 1, 0},{31048, 1, 1, 19},{64078, 1, 0, 0}, {0, 0, 0, 0}}},{30276, {{31048, 1, 0, 0},{31048, 1, 1, 0},{31048, 1, 1, 19},{64078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{18773, {{31049, 1, 0, 0},{31049, 1, 1, 18}, {0, 0, 0, 0}}},{18773, {{31049, 1, 0, 0},{31049, 1, 1, 18}, {0, 0, 0, 0}}},{30275, {{31049, 1, 0, 0},{31049, 1, 1, 0},{31049, 1, 1, 19},{64077, 1, 0, 0}, {0, 0, 0, 0}}},{30275, {{31049, 1, 0, 0},{31049, 1, 1, 0},{31049, 1, 1, 19},{64077, 1, 0, 0}, {0, 0, 0, 0}}},{30275, {{31049, 1, 0, 0},{31049, 1, 1, 0},{31049, 1, 1, 19},{64077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18774, {{31050, 1, 0, 0},{31050, 1, 1, 18}, {0, 0, 0, 0}}},{18774, {{31050, 1, 0, 0},{31050, 1, 1, 18}, {0, 0, 0, 0}}},{18775, {{31050, 1, 0, 0},{31050, 1, 1, 17}, {0, 0, 0, 0}}},{18775, {{31050, 1, 0, 0},{31050, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{18776, {{31051, 1, 0, 0},{31051, 1, 1, 18}, {0, 0, 0, 0}}},{18776, {{31051, 1, 0, 0},{31051, 1, 1, 18}, {0, 0, 0, 0}}},{18777, {{31051, 1, 0, 0},{31051, 1, 1, 17}, {0, 0, 0, 0}}},{18777, {{31051, 1, 0, 0},{31051, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{18778, {{31052, 1, 0, 0},{31052, 1, 1, 16}, {0, 0, 0, 0}}},{18778, {{31052, 1, 0, 0},{31052, 1, 1, 16}, {0, 0, 0, 0}}},{18779, {{31052, 1, 0, 0},{31052, 1, 1, 17}, {0, 0, 0, 0}}},{18779, {{31052, 1, 0, 0},{31052, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{18780, {{31055, 1, 0, 0},{31055, 1, 1, 18}, {0, 0, 0, 0}}},{18780, {{31055, 1, 0, 0},{31055, 1, 1, 18}, {0, 0, 0, 0}}},{18781, {{31055, 1, 0, 0},{31055, 1, 1, 17}, {0, 0, 0, 0}}},{18781, {{31055, 1, 0, 0},{31055, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {5, {{18782, {{31056, 1, 0, 0},{31056, 1, 1, 18}, {0, 0, 0, 0}}},{18782, {{31056, 1, 0, 0},{31056, 1, 1, 18}, {0, 0, 0, 0}}},{30277, {{31056, 1, 0, 0},{31056, 1, 1, 0},{31056, 1, 1, 19},{64079, 1, 0, 0}, {0, 0, 0, 0}}},{30277, {{31056, 1, 0, 0},{31056, 1, 1, 0},{31056, 1, 1, 19},{64079, 1, 0, 0}, {0, 0, 0, 0}}},{30277, {{31056, 1, 0, 0},{31056, 1, 1, 0},{31056, 1, 1, 19},{64079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18783, {{31057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18784, {{31058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{18785, {{31059, 1, 0, 0},{31059, 1, 1, 22}, {0, 0, 0, 0}}},{18785, {{31059, 1, 0, 0},{31059, 1, 1, 22}, {0, 0, 0, 0}}},{18786, {{31059, 1, 0, 0},{31059, 1, 1, 20}, {0, 0, 0, 0}}},{18786, {{31059, 1, 0, 0},{31059, 1, 1, 20}, {0, 0, 0, 0}}},{18787, {{31059, 1, 0, 0},{31059, 1, 1, 19}, {0, 0, 0, 0}}},{18787, {{31059, 1, 0, 0},{31059, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{18790, {{31060, 1, 0, 0},{31060, 1, 1, 18}, {0, 0, 0, 0}}},{18790, {{31060, 1, 0, 0},{31060, 1, 1, 18}, {0, 0, 0, 0}}},{18791, {{31060, 1, 0, 0},{31060, 1, 1, 17}, {0, 0, 0, 0}}},{18791, {{31060, 1, 0, 0},{31060, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{18792, {{31061, 1, 0, 0},{31061, 1, 1, 17}, {0, 0, 0, 0}}},{18792, {{31061, 1, 0, 0},{31061, 1, 1, 17}, {0, 0, 0, 0}}},{18793, {{31061, 1, 0, 0},{31061, 1, 1, 18}, {0, 0, 0, 0}}},{18793, {{31061, 1, 0, 0},{31061, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {7, {{18794, {{31062, 1, 0, 0},{31062, 1, 1, 18}, {0, 0, 0, 0}}},{18794, {{31062, 1, 0, 0},{31062, 1, 1, 18}, {0, 0, 0, 0}}},{30278, {{31062, 1, 0, 0},{31062, 1, 1, 0},{31062, 1, 1, 19},{64080, 1, 0, 0}, {0, 0, 0, 0}}},{30278, {{31062, 1, 0, 0},{31062, 1, 1, 0},{31062, 1, 1, 19},{64080, 1, 0, 0}, {0, 0, 0, 0}}},{30278, {{31062, 1, 0, 0},{31062, 1, 1, 0},{31062, 1, 1, 19},{64080, 1, 0, 0}, {0, 0, 0, 0}}},{59956, {{31062, 1, 0, 0},{31062, 1, 1, 20}, {0, 0, 0, 0}}},{59956, {{31062, 1, 0, 0},{31062, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{18795, {{31063, 1, 0, 0},{31063, 1, 1, 18}, {0, 0, 0, 0}}},{18795, {{31063, 1, 0, 0},{31063, 1, 1, 18}, {0, 0, 0, 0}}},{18796, {{31063, 1, 0, 0},{31063, 1, 1, 17}, {0, 0, 0, 0}}},{18796, {{31063, 1, 0, 0},{31063, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {8, {{18797, {{31064, 1, 0, 0},{31064, 1, 1, 18}, {0, 0, 0, 0}}},{18797, {{31064, 1, 0, 0},{31064, 1, 1, 18}, {0, 0, 0, 0}}},{18798, {{31064, 1, 0, 0},{31064, 1, 1, 17}, {0, 0, 0, 0}}},{18798, {{31064, 1, 0, 0},{31064, 1, 1, 17}, {0, 0, 0, 0}}},{18799, {{31064, 1, 0, 0},{31064, 1, 1, 19}, {0, 0, 0, 0}}},{18799, {{31064, 1, 0, 0},{31064, 1, 1, 19}, {0, 0, 0, 0}}},{18800, {{31064, 1, 0, 0},{31064, 1, 1, 20}, {0, 0, 0, 0}}},{18800, {{31064, 1, 0, 0},{31064, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{18801, {{31065, 1, 0, 0},{31065, 1, 1, 16}, {0, 0, 0, 0}}},{18801, {{31065, 1, 0, 0},{31065, 1, 1, 16}, {0, 0, 0, 0}}},{18802, {{31065, 1, 0, 0},{31065, 1, 1, 17}, {0, 0, 0, 0}}},{18802, {{31065, 1, 0, 0},{31065, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{18803, {{31066, 1, 0, 0},{31066, 1, 1, 18}, {0, 0, 0, 0}}},{18803, {{31066, 1, 0, 0},{31066, 1, 1, 18}, {0, 0, 0, 0}}},{18804, {{31066, 1, 0, 0},{31066, 1, 1, 17}, {0, 0, 0, 0}}},{18804, {{31066, 1, 0, 0},{31066, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{18806, {{31067, 1, 0, 0},{31067, 1, 1, 18}, {0, 0, 0, 0}}},{18806, {{31067, 1, 0, 0},{31067, 1, 1, 18}, {0, 0, 0, 0}}},{18807, {{31067, 1, 0, 0},{31067, 1, 1, 17}, {0, 0, 0, 0}}},{18807, {{31067, 1, 0, 0},{31067, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{18808, {{31068, 1, 0, 0},{31068, 1, 1, 18}, {0, 0, 0, 0}}},{18808, {{31068, 1, 0, 0},{31068, 1, 1, 18}, {0, 0, 0, 0}}},{18809, {{31068, 1, 0, 0},{31068, 1, 1, 17}, {0, 0, 0, 0}}},{18809, {{31068, 1, 0, 0},{31068, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {5, {{18810, {{31069, 1, 0, 0},{31069, 1, 1, 18}, {0, 0, 0, 0}}},{18810, {{31069, 1, 0, 0},{31069, 1, 1, 18}, {0, 0, 0, 0}}},{30279, {{31069, 1, 0, 0},{31069, 1, 1, 0},{31069, 1, 1, 19},{64081, 1, 0, 0}, {0, 0, 0, 0}}},{30279, {{31069, 1, 0, 0},{31069, 1, 1, 0},{31069, 1, 1, 19},{64081, 1, 0, 0}, {0, 0, 0, 0}}},{30279, {{31069, 1, 0, 0},{31069, 1, 1, 0},{31069, 1, 1, 19},{64081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{18811, {{31070, 1, 0, 0},{31070, 1, 1, 18}, {0, 0, 0, 0}}},{18811, {{31070, 1, 0, 0},{31070, 1, 1, 18}, {0, 0, 0, 0}}},{30211, {{31070, 1, 0, 0},{31070, 1, 1, 0},{31070, 1, 1, 19},{64025, 1, 0, 0}, {0, 0, 0, 0}}},{30211, {{31070, 1, 0, 0},{31070, 1, 1, 0},{31070, 1, 1, 19},{64025, 1, 0, 0}, {0, 0, 0, 0}}},{30211, {{31070, 1, 0, 0},{31070, 1, 1, 0},{31070, 1, 1, 19},{64025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18812, {{31071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18813, {{31072, 1, 0, 0},{31072, 1, 1, 18}, {0, 0, 0, 0}}},{18813, {{31072, 1, 0, 0},{31072, 1, 1, 18}, {0, 0, 0, 0}}},{18814, {{31072, 1, 0, 0},{31072, 1, 1, 17}, {0, 0, 0, 0}}},{18814, {{31072, 1, 0, 0},{31072, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{18815, {{31073, 1, 0, 0},{31073, 1, 1, 16}, {0, 0, 0, 0}}},{18815, {{31073, 1, 0, 0},{31073, 1, 1, 16}, {0, 0, 0, 0}}},{18816, {{31073, 1, 0, 0},{31073, 1, 1, 17}, {0, 0, 0, 0}}},{18816, {{31073, 1, 0, 0},{31073, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {10, {{18817, {{31074, 1, 0, 0},{31074, 1, 1, 18}, {0, 0, 0, 0}}},{18817, {{31074, 1, 0, 0},{31074, 1, 1, 18}, {0, 0, 0, 0}}},{18818, {{31074, 1, 0, 0},{31074, 1, 1, 22}, {0, 0, 0, 0}}},{18818, {{31074, 1, 0, 0},{31074, 1, 1, 22}, {0, 0, 0, 0}}},{18819, {{31074, 1, 0, 0},{31074, 1, 1, 20}, {0, 0, 0, 0}}},{18819, {{31074, 1, 0, 0},{31074, 1, 1, 20}, {0, 0, 0, 0}}},{18820, {{31074, 1, 0, 0},{31074, 1, 1, 21}, {0, 0, 0, 0}}},{18820, {{31074, 1, 0, 0},{31074, 1, 1, 21}, {0, 0, 0, 0}}},{18821, {{31074, 1, 0, 0},{31074, 1, 1, 19}, {0, 0, 0, 0}}},{18821, {{31074, 1, 0, 0},{31074, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{18822, {{31075, 1, 0, 0},{31075, 1, 1, 16}, {0, 0, 0, 0}}},{18822, {{31075, 1, 0, 0},{31075, 1, 1, 16}, {0, 0, 0, 0}}},{18823, {{31075, 1, 0, 0},{31075, 1, 1, 17}, {0, 0, 0, 0}}},{18823, {{31075, 1, 0, 0},{31075, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{18824, {{31076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{18825, {{31077, 1, 0, 0},{31077, 1, 1, 18}, {0, 0, 0, 0}}},{18825, {{31077, 1, 0, 0},{31077, 1, 1, 18}, {0, 0, 0, 0}}},{30212, {{31077, 1, 0, 0},{31077, 1, 1, 0},{31077, 1, 1, 19},{64026, 1, 0, 0}, {0, 0, 0, 0}}},{30212, {{31077, 1, 0, 0},{31077, 1, 1, 0},{31077, 1, 1, 19},{64026, 1, 0, 0}, {0, 0, 0, 0}}},{30212, {{31077, 1, 0, 0},{31077, 1, 1, 0},{31077, 1, 1, 19},{64026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18826, {{31078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18828, {{31079, 1, 0, 0},{31079, 1, 1, 18}, {0, 0, 0, 0}}},{18828, {{31079, 1, 0, 0},{31079, 1, 1, 18}, {0, 0, 0, 0}}},{18829, {{31079, 1, 0, 0},{31079, 1, 1, 17}, {0, 0, 0, 0}}},{18829, {{31079, 1, 0, 0},{31079, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{18830, {{31080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18831, {{31081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18832, {{31082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18833, {{31083, 1, 0, 0},{31083, 1, 1, 18}, {0, 0, 0, 0}}},{18833, {{31083, 1, 0, 0},{31083, 1, 1, 18}, {0, 0, 0, 0}}},{18834, {{31083, 1, 0, 0},{31083, 1, 1, 17}, {0, 0, 0, 0}}},{18834, {{31083, 1, 0, 0},{31083, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{18835, {{31084, 1, 0, 0},{31084, 1, 1, 16}, {0, 0, 0, 0}}},{18835, {{31084, 1, 0, 0},{31084, 1, 1, 16}, {0, 0, 0, 0}}},{18836, {{31084, 1, 0, 0},{31084, 1, 1, 17}, {0, 0, 0, 0}}},{18836, {{31084, 1, 0, 0},{31084, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{18837, {{31085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18838, {{31086, 1, 0, 0},{31086, 1, 1, 16}, {0, 0, 0, 0}}},{18838, {{31086, 1, 0, 0},{31086, 1, 1, 16}, {0, 0, 0, 0}}},{18839, {{31086, 1, 0, 0},{31086, 1, 1, 17}, {0, 0, 0, 0}}},{18839, {{31086, 1, 0, 0},{31086, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{18841, {{31088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{18845, {{31089, 1, 0, 0},{31089, 1, 1, 16}, {0, 0, 0, 0}}},{18845, {{31089, 1, 0, 0},{31089, 1, 1, 16}, {0, 0, 0, 0}}},{18846, {{31089, 1, 0, 0},{31089, 1, 1, 17}, {0, 0, 0, 0}}},{18846, {{31089, 1, 0, 0},{31089, 1, 1, 17}, {0, 0, 0, 0}}},{18847, {{31089, 1, 0, 0},{31089, 1, 1, 18}, {0, 0, 0, 0}}},{18847, {{31089, 1, 0, 0},{31089, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{18848, {{31090, 1, 0, 0},{31090, 1, 1, 18}, {0, 0, 0, 0}}},{18848, {{31090, 1, 0, 0},{31090, 1, 1, 18}, {0, 0, 0, 0}}},{18849, {{31090, 1, 0, 0},{31090, 1, 1, 17}, {0, 0, 0, 0}}},{18849, {{31090, 1, 0, 0},{31090, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{18850, {{31091, 1, 0, 0},{31091, 1, 1, 16}, {0, 0, 0, 0}}},{18850, {{31091, 1, 0, 0},{31091, 1, 1, 16}, {0, 0, 0, 0}}},{18851, {{31091, 1, 0, 0},{31091, 1, 1, 17}, {0, 0, 0, 0}}},{18851, {{31091, 1, 0, 0},{31091, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{18852, {{31092, 1, 0, 0},{31092, 1, 1, 17}, {0, 0, 0, 0}}},{18852, {{31092, 1, 0, 0},{31092, 1, 1, 17}, {0, 0, 0, 0}}},{18853, {{31092, 1, 0, 0},{31092, 1, 1, 16}, {0, 0, 0, 0}}},{18853, {{31092, 1, 0, 0},{31092, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{18854, {{31094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18855, {{31095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{18856, {{31097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18857, {{31098, 1, 0, 0},{31098, 1, 1, 18}, {0, 0, 0, 0}}},{18857, {{31098, 1, 0, 0},{31098, 1, 1, 18}, {0, 0, 0, 0}}},{18858, {{31098, 1, 0, 0},{31098, 1, 1, 17}, {0, 0, 0, 0}}},{18858, {{31098, 1, 0, 0},{31098, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{18859, {{31099, 1, 0, 0},{31099, 1, 1, 18}, {0, 0, 0, 0}}},{18859, {{31099, 1, 0, 0},{31099, 1, 1, 18}, {0, 0, 0, 0}}},{18860, {{31099, 1, 0, 0},{31099, 1, 1, 17}, {0, 0, 0, 0}}},{18860, {{31099, 1, 0, 0},{31099, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{18861, {{31100, 1, 0, 0},{31100, 1, 1, 18}, {0, 0, 0, 0}}},{18861, {{31100, 1, 0, 0},{31100, 1, 1, 18}, {0, 0, 0, 0}}},{18862, {{31100, 1, 0, 0},{31100, 1, 1, 17}, {0, 0, 0, 0}}},{18862, {{31100, 1, 0, 0},{31100, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{18863, {{31101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18864, {{31102, 1, 0, 0},{31102, 1, 1, 18}, {0, 0, 0, 0}}},{18864, {{31102, 1, 0, 0},{31102, 1, 1, 18}, {0, 0, 0, 0}}},{18865, {{31102, 1, 0, 0},{31102, 1, 1, 17}, {0, 0, 0, 0}}},{18865, {{31102, 1, 0, 0},{31102, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{18867, {{31103, 1, 0, 0},{31103, 1, 1, 17}, {0, 0, 0, 0}}},{18867, {{31103, 1, 0, 0},{31103, 1, 1, 17}, {0, 0, 0, 0}}},{18868, {{31103, 1, 0, 0},{31103, 1, 1, 18}, {0, 0, 0, 0}}},{18868, {{31103, 1, 0, 0},{31103, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{18869, {{31104, 1, 0, 0},{31104, 1, 1, 17}, {0, 0, 0, 0}}},{18869, {{31104, 1, 0, 0},{31104, 1, 1, 17}, {0, 0, 0, 0}}},{18870, {{31104, 1, 0, 0},{31104, 1, 1, 18}, {0, 0, 0, 0}}},{18870, {{31104, 1, 0, 0},{31104, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18871, {{31105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18872, {{31106, 1, 0, 0},{31106, 1, 1, 16}, {0, 0, 0, 0}}},{18872, {{31106, 1, 0, 0},{31106, 1, 1, 16}, {0, 0, 0, 0}}},{18873, {{31106, 1, 0, 0},{31106, 1, 1, 17}, {0, 0, 0, 0}}},{18873, {{31106, 1, 0, 0},{31106, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{18875, {{31107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{18877, {{31108, 1, 0, 0},{31108, 1, 1, 18}, {0, 0, 0, 0}}},{18877, {{31108, 1, 0, 0},{31108, 1, 1, 18}, {0, 0, 0, 0}}},{18878, {{31108, 1, 0, 0},{31108, 1, 1, 17}, {0, 0, 0, 0}}},{18878, {{31108, 1, 0, 0},{31108, 1, 1, 17}, {0, 0, 0, 0}}},{18879, {{31108, 1, 0, 0},{31108, 1, 1, 19}, {0, 0, 0, 0}}},{18879, {{31108, 1, 0, 0},{31108, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{18880, {{31109, 1, 0, 0},{31109, 1, 1, 18}, {0, 0, 0, 0}}},{18880, {{31109, 1, 0, 0},{31109, 1, 1, 18}, {0, 0, 0, 0}}},{18881, {{31109, 1, 0, 0},{31109, 1, 1, 17}, {0, 0, 0, 0}}},{18881, {{31109, 1, 0, 0},{31109, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{18882, {{31112, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18884, {{31113, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {10, {{18885, {{31114, 1, 0, 0},{31114, 1, 1, 20}, {0, 0, 0, 0}}},{18885, {{31114, 1, 0, 0},{31114, 1, 1, 20}, {0, 0, 0, 0}}},{18886, {{31114, 1, 0, 0},{31114, 1, 1, 19}, {0, 0, 0, 0}}},{18886, {{31114, 1, 0, 0},{31114, 1, 1, 19}, {0, 0, 0, 0}}},{18887, {{31114, 1, 0, 0},{31114, 1, 1, 17}, {0, 0, 0, 0}}},{18887, {{31114, 1, 0, 0},{31114, 1, 1, 17}, {0, 0, 0, 0}}},{18888, {{31114, 1, 0, 0},{31114, 1, 1, 18}, {0, 0, 0, 0}}},{18888, {{31114, 1, 0, 0},{31114, 1, 1, 18}, {0, 0, 0, 0}}},{18889, {{31114, 1, 0, 0},{31114, 1, 1, 21}, {0, 0, 0, 0}}},{18889, {{31114, 1, 0, 0},{31114, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {4, {{18890, {{31115, 1, 0, 0},{31115, 1, 1, 18}, {0, 0, 0, 0}}},{18890, {{31115, 1, 0, 0},{31115, 1, 1, 18}, {0, 0, 0, 0}}},{18891, {{31115, 1, 0, 0},{31115, 1, 1, 17}, {0, 0, 0, 0}}},{18891, {{31115, 1, 0, 0},{31115, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{18892, {{31116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {7, {{18894, {{31117, 1, 0, 0},{31117, 1, 1, 18}, {0, 0, 0, 0}}},{18894, {{31117, 1, 0, 0},{31117, 1, 1, 18}, {0, 0, 0, 0}}},{30280, {{31117, 1, 0, 0},{31117, 1, 1, 0},{31117, 1, 1, 19},{64082, 1, 0, 0}, {0, 0, 0, 0}}},{30280, {{31117, 1, 0, 0},{31117, 1, 1, 0},{31117, 1, 1, 19},{64082, 1, 0, 0}, {0, 0, 0, 0}}},{30280, {{31117, 1, 0, 0},{31117, 1, 1, 0},{31117, 1, 1, 19},{64082, 1, 0, 0}, {0, 0, 0, 0}}},{58212, {{31117, 1, 0, 0},{31117, 1, 1, 20}, {0, 0, 0, 0}}},{58212, {{31117, 1, 0, 0},{31117, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {5, {{18895, {{31118, 1, 0, 0},{31118, 1, 1, 18}, {0, 0, 0, 0}}},{18895, {{31118, 1, 0, 0},{31118, 1, 1, 18}, {0, 0, 0, 0}}},{30281, {{31118, 1, 0, 0},{31118, 1, 1, 0},{31118, 1, 1, 19},{64083, 1, 0, 0}, {0, 0, 0, 0}}},{30281, {{31118, 1, 0, 0},{31118, 1, 1, 0},{31118, 1, 1, 19},{64083, 1, 0, 0}, {0, 0, 0, 0}}},{30281, {{31118, 1, 0, 0},{31118, 1, 1, 0},{31118, 1, 1, 19},{64083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {7, {{18896, {{31119, 1, 0, 0},{31119, 1, 1, 18}, {0, 0, 0, 0}}},{18896, {{31119, 1, 0, 0},{31119, 1, 1, 18}, {0, 0, 0, 0}}},{18897, {{31119, 1, 0, 0},{31119, 1, 1, 20}, {0, 0, 0, 0}}},{18897, {{31119, 1, 0, 0},{31119, 1, 1, 20}, {0, 0, 0, 0}}},{30213, {{31119, 1, 0, 0},{31119, 1, 1, 0},{31119, 1, 1, 19},{64027, 1, 0, 0}, {0, 0, 0, 0}}},{30213, {{31119, 1, 0, 0},{31119, 1, 1, 0},{31119, 1, 1, 19},{64027, 1, 0, 0}, {0, 0, 0, 0}}},{30213, {{31119, 1, 0, 0},{31119, 1, 1, 0},{31119, 1, 1, 19},{64027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18899, {{31120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18900, {{31121, 1, 0, 0},{31121, 1, 1, 18}, {0, 0, 0, 0}}},{18900, {{31121, 1, 0, 0},{31121, 1, 1, 18}, {0, 0, 0, 0}}},{18901, {{31121, 1, 0, 0},{31121, 1, 1, 17}, {0, 0, 0, 0}}},{18901, {{31121, 1, 0, 0},{31121, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{18902, {{31122, 1, 0, 0},{31122, 1, 1, 16}, {0, 0, 0, 0}}},{18902, {{31122, 1, 0, 0},{31122, 1, 1, 16}, {0, 0, 0, 0}}},{18903, {{31122, 1, 0, 0},{31122, 1, 1, 17}, {0, 0, 0, 0}}},{18903, {{31122, 1, 0, 0},{31122, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{18905, {{31123, 1, 0, 0},{31123, 1, 1, 18}, {0, 0, 0, 0}}},{18905, {{31123, 1, 0, 0},{31123, 1, 1, 18}, {0, 0, 0, 0}}},{18906, {{31123, 1, 0, 0},{31123, 1, 1, 17}, {0, 0, 0, 0}}},{18906, {{31123, 1, 0, 0},{31123, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{18907, {{31124, 1, 0, 0},{31124, 1, 1, 18}, {0, 0, 0, 0}}},{18907, {{31124, 1, 0, 0},{31124, 1, 1, 18}, {0, 0, 0, 0}}},{18908, {{31124, 1, 0, 0},{31124, 1, 1, 17}, {0, 0, 0, 0}}},{18908, {{31124, 1, 0, 0},{31124, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {10, {{18909, {{31125, 1, 0, 0},{31125, 1, 1, 23}, {0, 0, 0, 0}}},{18909, {{31125, 1, 0, 0},{31125, 1, 1, 23}, {0, 0, 0, 0}}},{18910, {{31125, 1, 0, 0},{31125, 1, 1, 18}, {0, 0, 0, 0}}},{18910, {{31125, 1, 0, 0},{31125, 1, 1, 18}, {0, 0, 0, 0}}},{18911, {{31125, 1, 0, 0},{31125, 1, 1, 24}, {0, 0, 0, 0}}},{18911, {{31125, 1, 0, 0},{31125, 1, 1, 24}, {0, 0, 0, 0}}},{18912, {{31125, 1, 0, 0},{31125, 1, 1, 17}, {0, 0, 0, 0}}},{18912, {{31125, 1, 0, 0},{31125, 1, 1, 17}, {0, 0, 0, 0}}},{18913, {{31125, 1, 0, 0},{31125, 1, 1, 19}, {0, 0, 0, 0}}},{18913, {{31125, 1, 0, 0},{31125, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{18914, {{31126, 1, 0, 0},{31126, 1, 1, 18}, {0, 0, 0, 0}}},{18914, {{31126, 1, 0, 0},{31126, 1, 1, 18}, {0, 0, 0, 0}}},{18915, {{31126, 1, 0, 0},{31126, 1, 1, 17}, {0, 0, 0, 0}}},{18915, {{31126, 1, 0, 0},{31126, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{18916, {{31127, 1, 0, 0},{31127, 1, 1, 16}, {0, 0, 0, 0}}},{18916, {{31127, 1, 0, 0},{31127, 1, 1, 16}, {0, 0, 0, 0}}},{18917, {{31127, 1, 0, 0},{31127, 1, 1, 17}, {0, 0, 0, 0}}},{18917, {{31127, 1, 0, 0},{31127, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {6, {{18918, {{31128, 1, 0, 0},{31128, 1, 1, 18}, {0, 0, 0, 0}}},{18918, {{31128, 1, 0, 0},{31128, 1, 1, 18}, {0, 0, 0, 0}}},{18919, {{31128, 1, 0, 0},{31128, 1, 1, 17}, {0, 0, 0, 0}}},{18919, {{31128, 1, 0, 0},{31128, 1, 1, 17}, {0, 0, 0, 0}}},{18920, {{31128, 1, 0, 0},{31128, 1, 1, 19}, {0, 0, 0, 0}}},{18920, {{31128, 1, 0, 0},{31128, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{18921, {{31130, 1, 0, 0},{31130, 1, 1, 17}, {0, 0, 0, 0}}},{18921, {{31130, 1, 0, 0},{31130, 1, 1, 17}, {0, 0, 0, 0}}},{18922, {{31130, 1, 0, 0},{31130, 1, 1, 16}, {0, 0, 0, 0}}},{18922, {{31130, 1, 0, 0},{31130, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{18923, {{31131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18926, {{31132, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{18927, {{31133, 1, 0, 0},{31133, 1, 1, 18}, {0, 0, 0, 0}}},{18927, {{31133, 1, 0, 0},{31133, 1, 1, 18}, {0, 0, 0, 0}}},{18928, {{31133, 1, 0, 0},{31133, 1, 1, 17}, {0, 0, 0, 0}}},{18928, {{31133, 1, 0, 0},{31133, 1, 1, 17}, {0, 0, 0, 0}}},{18929, {{31133, 1, 0, 0},{31133, 1, 1, 19}, {0, 0, 0, 0}}},{18929, {{31133, 1, 0, 0},{31133, 1, 1, 19}, {0, 0, 0, 0}}},{18930, {{31133, 1, 0, 0},{31133, 1, 1, 20}, {0, 0, 0, 0}}},{18930, {{31133, 1, 0, 0},{31133, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{18931, {{31134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18933, {{31135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18935, {{31136, 1, 0, 0},{31136, 1, 1, 17}, {0, 0, 0, 0}}},{18935, {{31136, 1, 0, 0},{31136, 1, 1, 17}, {0, 0, 0, 0}}},{18936, {{31136, 1, 0, 0},{31136, 1, 1, 16}, {0, 0, 0, 0}}},{18936, {{31136, 1, 0, 0},{31136, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {4, {{18937, {{31137, 1, 0, 0},{31137, 1, 1, 18}, {0, 0, 0, 0}}},{18937, {{31137, 1, 0, 0},{31137, 1, 1, 18}, {0, 0, 0, 0}}},{18938, {{31137, 1, 0, 0},{31137, 1, 1, 17}, {0, 0, 0, 0}}},{18938, {{31137, 1, 0, 0},{31137, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{18939, {{31138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{18940, {{31140, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18941, {{31141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18943, {{31142, 1, 0, 0},{31142, 1, 1, 17}, {0, 0, 0, 0}}},{18943, {{31142, 1, 0, 0},{31142, 1, 1, 17}, {0, 0, 0, 0}}},{58214, {{31142, 1, 0, 0},{31142, 1, 1, 18}, {0, 0, 0, 0}}},{58214, {{31142, 1, 0, 0},{31142, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{18944, {{31143, 1, 0, 0},{31143, 1, 1, 19}, {0, 0, 0, 0}}},{18944, {{31143, 1, 0, 0},{31143, 1, 1, 19}, {0, 0, 0, 0}}},{18945, {{31143, 1, 0, 0},{31143, 1, 1, 18}, {0, 0, 0, 0}}},{18945, {{31143, 1, 0, 0},{31143, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{18948, {{31144, 1, 0, 0},{31144, 1, 1, 18}, {0, 0, 0, 0}}},{18948, {{31144, 1, 0, 0},{31144, 1, 1, 18}, {0, 0, 0, 0}}},{18949, {{31144, 1, 0, 0},{31144, 1, 1, 17}, {0, 0, 0, 0}}},{18949, {{31144, 1, 0, 0},{31144, 1, 1, 17}, {0, 0, 0, 0}}},{18950, {{31144, 1, 0, 0},{31144, 1, 1, 19}, {0, 0, 0, 0}}},{18950, {{31144, 1, 0, 0},{31144, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {6, {{18951, {{31145, 1, 0, 0},{31145, 1, 1, 18}, {0, 0, 0, 0}}},{18951, {{31145, 1, 0, 0},{31145, 1, 1, 18}, {0, 0, 0, 0}}},{18952, {{31145, 1, 0, 0},{31145, 1, 1, 17}, {0, 0, 0, 0}}},{18952, {{31145, 1, 0, 0},{31145, 1, 1, 17}, {0, 0, 0, 0}}},{18953, {{31145, 1, 0, 0},{31145, 1, 1, 19}, {0, 0, 0, 0}}},{18953, {{31145, 1, 0, 0},{31145, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {8, {{18954, {{31146, 1, 0, 0},{31146, 1, 1, 17}, {0, 0, 0, 0}}},{18954, {{31146, 1, 0, 0},{31146, 1, 1, 17}, {0, 0, 0, 0}}},{18955, {{31146, 1, 0, 0},{31146, 1, 1, 18}, {0, 0, 0, 0}}},{18955, {{31146, 1, 0, 0},{31146, 1, 1, 18}, {0, 0, 0, 0}}},{58215, {{31146, 1, 0, 0},{31146, 1, 1, 20}, {0, 0, 0, 0}}},{58215, {{31146, 1, 0, 0},{31146, 1, 1, 20}, {0, 0, 0, 0}}},{59973, {{31146, 1, 0, 0},{31146, 1, 1, 19}, {0, 0, 0, 0}}},{59973, {{31146, 1, 0, 0},{31146, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {6, {{18956, {{31147, 1, 0, 0},{31147, 1, 1, 18}, {0, 0, 0, 0}}},{18956, {{31147, 1, 0, 0},{31147, 1, 1, 18}, {0, 0, 0, 0}}},{18957, {{31147, 1, 0, 0},{31147, 1, 1, 17}, {0, 0, 0, 0}}},{18957, {{31147, 1, 0, 0},{31147, 1, 1, 17}, {0, 0, 0, 0}}},{18958, {{31147, 1, 0, 0},{31147, 1, 1, 19}, {0, 0, 0, 0}}},{18958, {{31147, 1, 0, 0},{31147, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{18959, {{31148, 1, 0, 0},{31148, 1, 1, 16}, {0, 0, 0, 0}}},{18959, {{31148, 1, 0, 0},{31148, 1, 1, 16}, {0, 0, 0, 0}}},{18960, {{31148, 1, 0, 0},{31148, 1, 1, 17}, {0, 0, 0, 0}}},{18960, {{31148, 1, 0, 0},{31148, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {6, {{18961, {{31149, 1, 0, 0},{31149, 1, 1, 16}, {0, 0, 0, 0}}},{18961, {{31149, 1, 0, 0},{31149, 1, 1, 16}, {0, 0, 0, 0}}},{18962, {{31149, 1, 0, 0},{31149, 1, 1, 17}, {0, 0, 0, 0}}},{18962, {{31149, 1, 0, 0},{31149, 1, 1, 17}, {0, 0, 0, 0}}},{18963, {{31149, 1, 0, 0},{31149, 1, 1, 18}, {0, 0, 0, 0}}},{18963, {{31149, 1, 0, 0},{31149, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{18964, {{31150, 1, 0, 0},{31150, 1, 1, 19}, {0, 0, 0, 0}}},{18964, {{31150, 1, 0, 0},{31150, 1, 1, 19}, {0, 0, 0, 0}}},{18965, {{31150, 1, 0, 0},{31150, 1, 1, 18}, {0, 0, 0, 0}}},{18965, {{31150, 1, 0, 0},{31150, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18966, {{31151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{18968, {{31152, 1, 0, 0},{31152, 1, 1, 18}, {0, 0, 0, 0}}},{18968, {{31152, 1, 0, 0},{31152, 1, 1, 18}, {0, 0, 0, 0}}},{18969, {{31152, 1, 0, 0},{31152, 1, 1, 19}, {0, 0, 0, 0}}},{18969, {{31152, 1, 0, 0},{31152, 1, 1, 19}, {0, 0, 0, 0}}},{18970, {{31152, 1, 0, 0},{31152, 1, 1, 20}, {0, 0, 0, 0}}},{18970, {{31152, 1, 0, 0},{31152, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{18971, {{31153, 1, 0, 0},{31153, 1, 1, 19}, {0, 0, 0, 0}}},{18971, {{31153, 1, 0, 0},{31153, 1, 1, 19}, {0, 0, 0, 0}}},{18972, {{31153, 1, 0, 0},{31153, 1, 1, 18}, {0, 0, 0, 0}}},{18972, {{31153, 1, 0, 0},{31153, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18973, {{31154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18974, {{31155, 1, 0, 0},{31155, 1, 1, 18}, {0, 0, 0, 0}}},{18974, {{31155, 1, 0, 0},{31155, 1, 1, 18}, {0, 0, 0, 0}}},{18975, {{31155, 1, 0, 0},{31155, 1, 1, 17}, {0, 0, 0, 0}}},{18975, {{31155, 1, 0, 0},{31155, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{18976, {{31156, 1, 0, 0},{31156, 1, 1, 18}, {0, 0, 0, 0}}},{18976, {{31156, 1, 0, 0},{31156, 1, 1, 18}, {0, 0, 0, 0}}},{18977, {{31156, 1, 0, 0},{31156, 1, 1, 17}, {0, 0, 0, 0}}},{18977, {{31156, 1, 0, 0},{31156, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{18978, {{31157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18979, {{31158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18980, {{31159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18981, {{31160, 1, 0, 0},{31160, 1, 1, 17}, {0, 0, 0, 0}}},{18981, {{31160, 1, 0, 0},{31160, 1, 1, 17}, {0, 0, 0, 0}}},{18982, {{31160, 1, 0, 0},{31160, 1, 1, 18}, {0, 0, 0, 0}}},{18982, {{31160, 1, 0, 0},{31160, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18983, {{31161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18984, {{31162, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{18985, {{31163, 1, 0, 0},{31163, 1, 1, 17}, {0, 0, 0, 0}}},{18985, {{31163, 1, 0, 0},{31163, 1, 1, 17}, {0, 0, 0, 0}}},{18986, {{31163, 1, 0, 0},{31163, 1, 1, 18}, {0, 0, 0, 0}}},{18986, {{31163, 1, 0, 0},{31163, 1, 1, 18}, {0, 0, 0, 0}}},{58219, {{31163, 1, 0, 0},{31163, 1, 1, 19}, {0, 0, 0, 0}}},{58219, {{31163, 1, 0, 0},{31163, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{18987, {{31164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{18988, {{31165, 1, 0, 0},{31165, 1, 1, 17}, {0, 0, 0, 0}}},{18988, {{31165, 1, 0, 0},{31165, 1, 1, 17}, {0, 0, 0, 0}}},{18989, {{31165, 1, 0, 0},{31165, 1, 1, 18}, {0, 0, 0, 0}}},{18989, {{31165, 1, 0, 0},{31165, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{18990, {{31166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18991, {{31167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18992, {{31168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18993, {{31169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18994, {{31170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{18996, {{31172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18997, {{31173, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18998, {{31174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18999, {{31175, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19000, {{31176, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19001, {{31177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19004, {{31178, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19006, {{31179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19007, {{31180, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19008, {{31181, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19009, {{31182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19010, {{31183, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19011, {{31184, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19012, {{31185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19013, {{31186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19014, {{31187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19015, {{31188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19016, {{31189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19017, {{31190, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19018, {{31191, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19019, {{31192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19020, {{31193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19021, {{31194, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19022, {{31195, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19023, {{31196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19024, {{31197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19025, {{31198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19026, {{31199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19027, {{31200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19028, {{31201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19030, {{31202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19031, {{31203, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19032, {{31204, 1, 0, 0},{31204, 1, 1, 18}, {0, 0, 0, 0}}},{19032, {{31204, 1, 0, 0},{31204, 1, 1, 18}, {0, 0, 0, 0}}},{19033, {{31204, 1, 0, 0},{31204, 1, 1, 19}, {0, 0, 0, 0}}},{19033, {{31204, 1, 0, 0},{31204, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{19034, {{31205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19035, {{31206, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19036, {{31207, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19037, {{31208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19039, {{31209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19040, {{31210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19041, {{31211, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19042, {{31212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19043, {{31213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19044, {{31214, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19045, {{31215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19046, {{31216, 1, 0, 0},{31216, 1, 1, 18}, {0, 0, 0, 0}}},{19046, {{31216, 1, 0, 0},{31216, 1, 1, 18}, {0, 0, 0, 0}}},{19047, {{31216, 1, 0, 0},{31216, 1, 1, 17}, {0, 0, 0, 0}}},{19047, {{31216, 1, 0, 0},{31216, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{19048, {{31217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19049, {{31218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19050, {{31219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19051, {{31220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19052, {{31221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19053, {{31222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19054, {{31223, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19055, {{31224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19056, {{31225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19057, {{31226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19058, {{31227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19059, {{31228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{19060, {{31231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19061, {{31232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19062, {{31233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19063, {{31234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19064, {{31235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19065, {{31236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19067, {{31237, 1, 0, 0},{31237, 1, 1, 17}, {0, 0, 0, 0}}},{19067, {{31237, 1, 0, 0},{31237, 1, 1, 17}, {0, 0, 0, 0}}},{58222, {{31237, 1, 0, 0},{31237, 1, 1, 18}, {0, 0, 0, 0}}},{58222, {{31237, 1, 0, 0},{31237, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19068, {{31238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19069, {{31239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19070, {{31240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19071, {{31241, 1, 0, 0},{31241, 1, 1, 17}, {0, 0, 0, 0}}},{19071, {{31241, 1, 0, 0},{31241, 1, 1, 17}, {0, 0, 0, 0}}},{19072, {{31241, 1, 0, 0},{31241, 1, 1, 18}, {0, 0, 0, 0}}},{19072, {{31241, 1, 0, 0},{31241, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19073, {{31242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19074, {{31243, 1, 0, 0},{31243, 1, 1, 19}, {0, 0, 0, 0}}},{19074, {{31243, 1, 0, 0},{31243, 1, 1, 19}, {0, 0, 0, 0}}},{19075, {{31243, 1, 0, 0},{31243, 1, 1, 18}, {0, 0, 0, 0}}},{19075, {{31243, 1, 0, 0},{31243, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19076, {{31244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19077, {{31245, 1, 0, 0},{31245, 1, 1, 17}, {0, 0, 0, 0}}},{19077, {{31245, 1, 0, 0},{31245, 1, 1, 17}, {0, 0, 0, 0}}},{19078, {{31245, 1, 0, 0},{31245, 1, 1, 18}, {0, 0, 0, 0}}},{19078, {{31245, 1, 0, 0},{31245, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19079, {{31246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19080, {{31247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19081, {{31248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19082, {{31249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19083, {{31250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19084, {{31251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19085, {{31252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19086, {{31253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19087, {{31254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19088, {{31255, 1, 0, 0},{31255, 1, 1, 19}, {0, 0, 0, 0}}},{19088, {{31255, 1, 0, 0},{31255, 1, 1, 19}, {0, 0, 0, 0}}},{19089, {{31255, 1, 0, 0},{31255, 1, 1, 18}, {0, 0, 0, 0}}},{19089, {{31255, 1, 0, 0},{31255, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19090, {{31256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19091, {{31257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19092, {{31258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19093, {{31259, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19094, {{31260, 1, 0, 0},{31260, 1, 1, 18}, {0, 0, 0, 0}}},{19094, {{31260, 1, 0, 0},{31260, 1, 1, 18}, {0, 0, 0, 0}}},{19095, {{31260, 1, 0, 0},{31260, 1, 1, 17}, {0, 0, 0, 0}}},{19095, {{31260, 1, 0, 0},{31260, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{19096, {{31261, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19097, {{31262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19098, {{31263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19099, {{31264, 1, 0, 0},{31264, 1, 1, 17}, {0, 0, 0, 0}}},{19099, {{31264, 1, 0, 0},{31264, 1, 1, 17}, {0, 0, 0, 0}}},{19100, {{31264, 1, 0, 0},{31264, 1, 1, 18}, {0, 0, 0, 0}}},{19100, {{31264, 1, 0, 0},{31264, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19102, {{31265, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19103, {{31266, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{19104, {{31270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{19105, {{31271, 1, 0, 0},{31271, 1, 1, 17}, {0, 0, 0, 0}}},{19105, {{31271, 1, 0, 0},{31271, 1, 1, 17}, {0, 0, 0, 0}}},{19106, {{31271, 1, 0, 0},{31271, 1, 1, 18}, {0, 0, 0, 0}}},{19106, {{31271, 1, 0, 0},{31271, 1, 1, 18}, {0, 0, 0, 0}}},{19107, {{31271, 1, 0, 0},{31271, 1, 1, 19}, {0, 0, 0, 0}}},{19107, {{31271, 1, 0, 0},{31271, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{19108, {{31272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19109, {{31273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19110, {{31274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19111, {{31275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19112, {{31276, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19113, {{31277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19114, {{31278, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19115, {{31279, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19116, {{31280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{19117, {{31281, 1, 0, 0},{31281, 1, 1, 18}, {0, 0, 0, 0}}},{19117, {{31281, 1, 0, 0},{31281, 1, 1, 18}, {0, 0, 0, 0}}},{19118, {{31281, 1, 0, 0},{31281, 1, 1, 20}, {0, 0, 0, 0}}},{19118, {{31281, 1, 0, 0},{31281, 1, 1, 20}, {0, 0, 0, 0}}},{19119, {{31281, 1, 0, 0},{31281, 1, 1, 19}, {0, 0, 0, 0}}},{19119, {{31281, 1, 0, 0},{31281, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{19120, {{31282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{19121, {{31284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19122, {{31285, 1, 0, 0},{31285, 1, 1, 17}, {0, 0, 0, 0}}},{19122, {{31285, 1, 0, 0},{31285, 1, 1, 17}, {0, 0, 0, 0}}},{19123, {{31285, 1, 0, 0},{31285, 1, 1, 18}, {0, 0, 0, 0}}},{19123, {{31285, 1, 0, 0},{31285, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19125, {{31286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19126, {{31287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19127, {{31288, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19128, {{31289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19129, {{31290, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19130, {{31291, 1, 0, 0},{31291, 1, 1, 17}, {0, 0, 0, 0}}},{19130, {{31291, 1, 0, 0},{31291, 1, 1, 17}, {0, 0, 0, 0}}},{19131, {{31291, 1, 0, 0},{31291, 1, 1, 18}, {0, 0, 0, 0}}},{19131, {{31291, 1, 0, 0},{31291, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{19133, {{31292, 1, 0, 0},{31292, 1, 1, 17}, {0, 0, 0, 0}}},{19133, {{31292, 1, 0, 0},{31292, 1, 1, 17}, {0, 0, 0, 0}}},{19134, {{31292, 1, 0, 0},{31292, 1, 1, 18}, {0, 0, 0, 0}}},{19134, {{31292, 1, 0, 0},{31292, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{19135, {{31293, 1, 0, 0},{31293, 1, 1, 17}, {0, 0, 0, 0}}},{19135, {{31293, 1, 0, 0},{31293, 1, 1, 17}, {0, 0, 0, 0}}},{58230, {{31293, 1, 0, 0},{31293, 1, 1, 18}, {0, 0, 0, 0}}},{58230, {{31293, 1, 0, 0},{31293, 1, 1, 18}, {0, 0, 0, 0}}},{59990, {{31293, 1, 0, 0},{31293, 1, 1, 19}, {0, 0, 0, 0}}},{59990, {{31293, 1, 0, 0},{31293, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{19137, {{31294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19138, {{31295, 1, 0, 0},{31295, 1, 1, 17}, {0, 0, 0, 0}}},{19138, {{31295, 1, 0, 0},{31295, 1, 1, 17}, {0, 0, 0, 0}}},{19139, {{31295, 1, 0, 0},{31295, 1, 1, 18}, {0, 0, 0, 0}}},{19139, {{31295, 1, 0, 0},{31295, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {5, {{19140, {{31296, 1, 0, 0},{31296, 1, 1, 18}, {0, 0, 0, 0}}},{19140, {{31296, 1, 0, 0},{31296, 1, 1, 18}, {0, 0, 0, 0}}},{30282, {{31296, 1, 0, 0},{31296, 1, 1, 0},{31296, 1, 1, 19},{64084, 1, 0, 0}, {0, 0, 0, 0}}},{30282, {{31296, 1, 0, 0},{31296, 1, 1, 0},{31296, 1, 1, 19},{64084, 1, 0, 0}, {0, 0, 0, 0}}},{30282, {{31296, 1, 0, 0},{31296, 1, 1, 0},{31296, 1, 1, 19},{64084, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19141, {{31297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19142, {{31298, 1, 0, 0},{31298, 1, 1, 17}, {0, 0, 0, 0}}},{19142, {{31298, 1, 0, 0},{31298, 1, 1, 17}, {0, 0, 0, 0}}},{59993, {{31298, 1, 0, 0},{31298, 1, 1, 18}, {0, 0, 0, 0}}},{59993, {{31298, 1, 0, 0},{31298, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19143, {{31299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19144, {{31300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19145, {{31301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19146, {{31302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19147, {{31303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19148, {{31304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19150, {{31305, 1, 0, 0},{31305, 1, 1, 18}, {0, 0, 0, 0}}},{19150, {{31305, 1, 0, 0},{31305, 1, 1, 18}, {0, 0, 0, 0}}},{19151, {{31305, 1, 0, 0},{31305, 1, 1, 17}, {0, 0, 0, 0}}},{19151, {{31305, 1, 0, 0},{31305, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{19152, {{31306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19153, {{31307, 1, 0, 0},{31307, 1, 1, 16}, {0, 0, 0, 0}}},{19153, {{31307, 1, 0, 0},{31307, 1, 1, 16}, {0, 0, 0, 0}}},{19154, {{31307, 1, 0, 0},{31307, 1, 1, 17}, {0, 0, 0, 0}}},{19154, {{31307, 1, 0, 0},{31307, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{19155, {{31308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19157, {{31309, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19158, {{31310, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19159, {{31311, 1, 0, 0},{31311, 1, 1, 18}, {0, 0, 0, 0}}},{19159, {{31311, 1, 0, 0},{31311, 1, 1, 18}, {0, 0, 0, 0}}},{19160, {{31311, 1, 0, 0},{31311, 1, 1, 17}, {0, 0, 0, 0}}},{19160, {{31311, 1, 0, 0},{31311, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{19161, {{31312, 1, 0, 0},{31312, 1, 1, 17}, {0, 0, 0, 0}}},{19161, {{31312, 1, 0, 0},{31312, 1, 1, 17}, {0, 0, 0, 0}}},{59994, {{31312, 1, 0, 0},{31312, 1, 1, 18}, {0, 0, 0, 0}}},{59994, {{31312, 1, 0, 0},{31312, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{19162, {{31315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19164, {{31316, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19165, {{31317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19166, {{31318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{19167, {{31319, 1, 0, 0},{31319, 1, 1, 17}, {0, 0, 0, 0}}},{19167, {{31319, 1, 0, 0},{31319, 1, 1, 17}, {0, 0, 0, 0}}},{19168, {{31319, 1, 0, 0},{31319, 1, 1, 18}, {0, 0, 0, 0}}},{19168, {{31319, 1, 0, 0},{31319, 1, 1, 18}, {0, 0, 0, 0}}},{19169, {{31319, 1, 0, 0},{31319, 1, 1, 19}, {0, 0, 0, 0}}},{19169, {{31319, 1, 0, 0},{31319, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{19170, {{31320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19171, {{31321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19172, {{31322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19174, {{31323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19175, {{31324, 1, 0, 0},{31324, 1, 1, 17}, {0, 0, 0, 0}}},{19175, {{31324, 1, 0, 0},{31324, 1, 1, 17}, {0, 0, 0, 0}}},{19176, {{31324, 1, 0, 0},{31324, 1, 1, 18}, {0, 0, 0, 0}}},{19176, {{31324, 1, 0, 0},{31324, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19177, {{31325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{19178, {{31327, 1, 0, 0},{31327, 1, 1, 18}, {0, 0, 0, 0}}},{19178, {{31327, 1, 0, 0},{31327, 1, 1, 18}, {0, 0, 0, 0}}},{19179, {{31327, 1, 0, 0},{31327, 1, 1, 17}, {0, 0, 0, 0}}},{19179, {{31327, 1, 0, 0},{31327, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{19180, {{31328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19181, {{31329, 1, 0, 0},{31329, 1, 1, 17}, {0, 0, 0, 0}}},{19181, {{31329, 1, 0, 0},{31329, 1, 1, 17}, {0, 0, 0, 0}}},{58239, {{31329, 1, 0, 0},{31329, 1, 1, 18}, {0, 0, 0, 0}}},{58239, {{31329, 1, 0, 0},{31329, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{19182, {{31330, 1, 0, 0},{31330, 1, 1, 17}, {0, 0, 0, 0}}},{19182, {{31330, 1, 0, 0},{31330, 1, 1, 17}, {0, 0, 0, 0}}},{19183, {{31330, 1, 0, 0},{31330, 1, 1, 19}, {0, 0, 0, 0}}},{19183, {{31330, 1, 0, 0},{31330, 1, 1, 19}, {0, 0, 0, 0}}},{19184, {{31330, 1, 0, 0},{31330, 1, 1, 18}, {0, 0, 0, 0}}},{19184, {{31330, 1, 0, 0},{31330, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19185, {{31331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19186, {{31332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19187, {{31333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19188, {{31334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19189, {{31335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19190, {{31336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19191, {{31337, 1, 0, 0},{31337, 1, 1, 17}, {0, 0, 0, 0}}},{19191, {{31337, 1, 0, 0},{31337, 1, 1, 17}, {0, 0, 0, 0}}},{19192, {{31337, 1, 0, 0},{31337, 1, 1, 18}, {0, 0, 0, 0}}},{19192, {{31337, 1, 0, 0},{31337, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19193, {{31338, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19194, {{31339, 1, 0, 0},{31339, 1, 1, 17}, {0, 0, 0, 0}}},{19194, {{31339, 1, 0, 0},{31339, 1, 1, 17}, {0, 0, 0, 0}}},{19195, {{31339, 1, 0, 0},{31339, 1, 1, 18}, {0, 0, 0, 0}}},{19195, {{31339, 1, 0, 0},{31339, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19196, {{31340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19197, {{31341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19198, {{31342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{19199, {{31344, 1, 0, 0},{31344, 1, 1, 17}, {0, 0, 0, 0}}},{19199, {{31344, 1, 0, 0},{31344, 1, 1, 17}, {0, 0, 0, 0}}},{60001, {{31344, 1, 0, 0},{31344, 1, 1, 18}, {0, 0, 0, 0}}},{60001, {{31344, 1, 0, 0},{31344, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19200, {{31345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19201, {{31346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19202, {{31347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19203, {{31348, 1, 0, 0},{31348, 1, 1, 19}, {0, 0, 0, 0}}},{19203, {{31348, 1, 0, 0},{31348, 1, 1, 19}, {0, 0, 0, 0}}},{19204, {{31348, 1, 0, 0},{31348, 1, 1, 20}, {0, 0, 0, 0}}},{19204, {{31348, 1, 0, 0},{31348, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{19205, {{31349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19206, {{31350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19207, {{31351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19208, {{31352, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19209, {{31353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19210, {{31354, 1, 0, 0},{31354, 1, 1, 18}, {0, 0, 0, 0}}},{19210, {{31354, 1, 0, 0},{31354, 1, 1, 18}, {0, 0, 0, 0}}},{19211, {{31354, 1, 0, 0},{31354, 1, 1, 19}, {0, 0, 0, 0}}},{19211, {{31354, 1, 0, 0},{31354, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{19212, {{31355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19213, {{31356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19214, {{31357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19215, {{31358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{19216, {{31359, 1, 0, 0},{31359, 1, 1, 19}, {0, 0, 0, 0}}},{19216, {{31359, 1, 0, 0},{31359, 1, 1, 19}, {0, 0, 0, 0}}},{19217, {{31359, 1, 0, 0},{31359, 1, 1, 18}, {0, 0, 0, 0}}},{19217, {{31359, 1, 0, 0},{31359, 1, 1, 18}, {0, 0, 0, 0}}},{19218, {{31359, 1, 0, 0},{31359, 1, 1, 20}, {0, 0, 0, 0}}},{19218, {{31359, 1, 0, 0},{31359, 1, 1, 20}, {0, 0, 0, 0}}},{19219, {{31359, 1, 0, 0},{31359, 1, 1, 21}, {0, 0, 0, 0}}},{19219, {{31359, 1, 0, 0},{31359, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{19220, {{31360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{19221, {{31361, 1, 0, 0},{31361, 1, 1, 18}, {0, 0, 0, 0}}},{19221, {{31361, 1, 0, 0},{31361, 1, 1, 18}, {0, 0, 0, 0}}},{30283, {{31361, 1, 0, 0},{31361, 1, 1, 0},{31361, 1, 1, 19},{64085, 1, 0, 0}, {0, 0, 0, 0}}},{30283, {{31361, 1, 0, 0},{31361, 1, 1, 0},{31361, 1, 1, 19},{64085, 1, 0, 0}, {0, 0, 0, 0}}},{30283, {{31361, 1, 0, 0},{31361, 1, 1, 0},{31361, 1, 1, 19},{64085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19222, {{31362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19223, {{31363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19224, {{31364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19225, {{31365, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19226, {{31366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19227, {{31367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19228, {{31368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19229, {{31369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19230, {{31370, 1, 0, 0},{31370, 1, 1, 18}, {0, 0, 0, 0}}},{19230, {{31370, 1, 0, 0},{31370, 1, 1, 18}, {0, 0, 0, 0}}},{19231, {{31370, 1, 0, 0},{31370, 1, 1, 17}, {0, 0, 0, 0}}},{19231, {{31370, 1, 0, 0},{31370, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{19232, {{31371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19233, {{31372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{19234, {{31375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19235, {{31376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19236, {{31377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19237, {{31378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19238, {{31379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19239, {{31380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19240, {{31381, 1, 0, 0},{31381, 1, 1, 18}, {0, 0, 0, 0}}},{19240, {{31381, 1, 0, 0},{31381, 1, 1, 18}, {0, 0, 0, 0}}},{19241, {{31381, 1, 0, 0},{31381, 1, 1, 19}, {0, 0, 0, 0}}},{19241, {{31381, 1, 0, 0},{31381, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{19242, {{31382, 1, 0, 0},{31382, 1, 1, 17}, {0, 0, 0, 0}}},{19242, {{31382, 1, 0, 0},{31382, 1, 1, 17}, {0, 0, 0, 0}}},{19243, {{31382, 1, 0, 0},{31382, 1, 1, 18}, {0, 0, 0, 0}}},{19243, {{31382, 1, 0, 0},{31382, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{19244, {{31383, 1, 0, 0},{31383, 1, 1, 18}, {0, 0, 0, 0}}},{19244, {{31383, 1, 0, 0},{31383, 1, 1, 18}, {0, 0, 0, 0}}},{19245, {{31383, 1, 0, 0},{31383, 1, 1, 19}, {0, 0, 0, 0}}},{19245, {{31383, 1, 0, 0},{31383, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{19247, {{31384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19248, {{31385, 1, 0, 0},{31385, 1, 1, 16}, {0, 0, 0, 0}}},{19248, {{31385, 1, 0, 0},{31385, 1, 1, 16}, {0, 0, 0, 0}}},{19249, {{31385, 1, 0, 0},{31385, 1, 1, 17}, {0, 0, 0, 0}}},{19249, {{31385, 1, 0, 0},{31385, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{19250, {{31386, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19253, {{31387, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{19254, {{31390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19255, {{31391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19256, {{31392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19257, {{31393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19258, {{31394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19259, {{31395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19260, {{31396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{19262, {{31400, 1, 0, 0},{31400, 1, 1, 16}, {0, 0, 0, 0}}},{19262, {{31400, 1, 0, 0},{31400, 1, 1, 16}, {0, 0, 0, 0}}},{19263, {{31400, 1, 0, 0},{31400, 1, 1, 17}, {0, 0, 0, 0}}},{19263, {{31400, 1, 0, 0},{31400, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{19264, {{31401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19265, {{31402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19266, {{31403, 1, 0, 0},{31403, 1, 1, 16}, {0, 0, 0, 0}}},{19266, {{31403, 1, 0, 0},{31403, 1, 1, 16}, {0, 0, 0, 0}}},{19267, {{31403, 1, 0, 0},{31403, 1, 1, 17}, {0, 0, 0, 0}}},{19267, {{31403, 1, 0, 0},{31403, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{19268, {{31404, 1, 0, 0},{31404, 1, 1, 17}, {0, 0, 0, 0}}},{19268, {{31404, 1, 0, 0},{31404, 1, 1, 17}, {0, 0, 0, 0}}},{19269, {{31404, 1, 0, 0},{31404, 1, 1, 18}, {0, 0, 0, 0}}},{19269, {{31404, 1, 0, 0},{31404, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{19271, {{31406, 1, 0, 0},{31406, 1, 1, 18}, {0, 0, 0, 0}}},{19271, {{31406, 1, 0, 0},{31406, 1, 1, 18}, {0, 0, 0, 0}}},{19272, {{31406, 1, 0, 0},{31406, 1, 1, 19}, {0, 0, 0, 0}}},{19272, {{31406, 1, 0, 0},{31406, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{19273, {{31407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19274, {{31408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19275, {{31409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19277, {{31410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19278, {{31411, 1, 0, 0},{31411, 1, 1, 18}, {0, 0, 0, 0}}},{19278, {{31411, 1, 0, 0},{31411, 1, 1, 18}, {0, 0, 0, 0}}},{19279, {{31411, 1, 0, 0},{31411, 1, 1, 17}, {0, 0, 0, 0}}},{19279, {{31411, 1, 0, 0},{31411, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{19280, {{31412, 1, 0, 0},{31412, 1, 1, 16}, {0, 0, 0, 0}}},{19280, {{31412, 1, 0, 0},{31412, 1, 1, 16}, {0, 0, 0, 0}}},{19281, {{31412, 1, 0, 0},{31412, 1, 1, 17}, {0, 0, 0, 0}}},{19281, {{31412, 1, 0, 0},{31412, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{19282, {{31413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19283, {{31414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19284, {{31415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19285, {{31416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19286, {{31417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19287, {{31418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{19288, {{31419, 1, 0, 0},{31419, 1, 1, 17}, {0, 0, 0, 0}}},{19288, {{31419, 1, 0, 0},{31419, 1, 1, 17}, {0, 0, 0, 0}}},{19289, {{31419, 1, 0, 0},{31419, 1, 1, 18}, {0, 0, 0, 0}}},{19289, {{31419, 1, 0, 0},{31419, 1, 1, 18}, {0, 0, 0, 0}}},{58250, {{31419, 1, 0, 0},{31419, 1, 1, 19}, {0, 0, 0, 0}}},{58250, {{31419, 1, 0, 0},{31419, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{19291, {{31420, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{19292, {{31422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19293, {{31423, 1, 0, 0},{31423, 1, 1, 17}, {0, 0, 0, 0}}},{19293, {{31423, 1, 0, 0},{31423, 1, 1, 17}, {0, 0, 0, 0}}},{19294, {{31423, 1, 0, 0},{31423, 1, 1, 18}, {0, 0, 0, 0}}},{19294, {{31423, 1, 0, 0},{31423, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19295, {{31424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19296, {{31425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19297, {{31426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19298, {{31427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19299, {{31428, 1, 0, 0},{31428, 1, 1, 17}, {0, 0, 0, 0}}},{19299, {{31428, 1, 0, 0},{31428, 1, 1, 17}, {0, 0, 0, 0}}},{19300, {{31428, 1, 0, 0},{31428, 1, 1, 18}, {0, 0, 0, 0}}},{19300, {{31428, 1, 0, 0},{31428, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19301, {{31429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19302, {{31430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19303, {{31431, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19304, {{31432, 1, 0, 0},{31432, 1, 1, 18}, {0, 0, 0, 0}}},{19304, {{31432, 1, 0, 0},{31432, 1, 1, 18}, {0, 0, 0, 0}}},{19305, {{31432, 1, 0, 0},{31432, 1, 1, 19}, {0, 0, 0, 0}}},{19305, {{31432, 1, 0, 0},{31432, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{19306, {{31433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19309, {{31434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19310, {{31435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19311, {{31436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19312, {{31437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19313, {{31438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19314, {{31439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19315, {{31440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19316, {{31441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19317, {{31442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19318, {{31443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19319, {{31444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19320, {{31445, 1, 0, 0},{31445, 1, 1, 17}, {0, 0, 0, 0}}},{19320, {{31445, 1, 0, 0},{31445, 1, 1, 17}, {0, 0, 0, 0}}},{19321, {{31445, 1, 0, 0},{31445, 1, 1, 18}, {0, 0, 0, 0}}},{19321, {{31445, 1, 0, 0},{31445, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{68040, {{31447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19323, {{31448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19324, {{31449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19325, {{31450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19326, {{31451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19327, {{31452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19328, {{31453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{19330, {{31455, 1, 0, 0},{31455, 1, 1, 17}, {0, 0, 0, 0}}},{19330, {{31455, 1, 0, 0},{31455, 1, 1, 17}, {0, 0, 0, 0}}},{19331, {{31455, 1, 0, 0},{31455, 1, 1, 18}, {0, 0, 0, 0}}},{19331, {{31455, 1, 0, 0},{31455, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{19332, {{31456, 1, 0, 0},{31456, 1, 1, 17}, {0, 0, 0, 0}}},{19332, {{31456, 1, 0, 0},{31456, 1, 1, 17}, {0, 0, 0, 0}}},{19333, {{31456, 1, 0, 0},{31456, 1, 1, 18}, {0, 0, 0, 0}}},{19333, {{31456, 1, 0, 0},{31456, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19334, {{31457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19335, {{31458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19336, {{31459, 1, 0, 0},{31459, 1, 1, 18}, {0, 0, 0, 0}}},{19336, {{31459, 1, 0, 0},{31459, 1, 1, 18}, {0, 0, 0, 0}}},{19337, {{31459, 1, 0, 0},{31459, 1, 1, 17}, {0, 0, 0, 0}}},{19337, {{31459, 1, 0, 0},{31459, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{19338, {{31460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19339, {{31461, 1, 0, 0},{31461, 1, 1, 17}, {0, 0, 0, 0}}},{19339, {{31461, 1, 0, 0},{31461, 1, 1, 17}, {0, 0, 0, 0}}},{19340, {{31461, 1, 0, 0},{31461, 1, 1, 18}, {0, 0, 0, 0}}},{19340, {{31461, 1, 0, 0},{31461, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19341, {{31462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19342, {{31463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19343, {{31464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19344, {{31465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19345, {{31466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19346, {{31467, 1, 0, 0},{31467, 1, 1, 17}, {0, 0, 0, 0}}},{19346, {{31467, 1, 0, 0},{31467, 1, 1, 17}, {0, 0, 0, 0}}},{19347, {{31467, 1, 0, 0},{31467, 1, 1, 18}, {0, 0, 0, 0}}},{19347, {{31467, 1, 0, 0},{31467, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19348, {{31468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19349, {{31469, 1, 0, 0},{31469, 1, 1, 17}, {0, 0, 0, 0}}},{19349, {{31469, 1, 0, 0},{31469, 1, 1, 17}, {0, 0, 0, 0}}},{19350, {{31469, 1, 0, 0},{31469, 1, 1, 18}, {0, 0, 0, 0}}},{19350, {{31469, 1, 0, 0},{31469, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{19351, {{31470, 1, 0, 0},{31470, 1, 1, 16}, {0, 0, 0, 0}}},{19351, {{31470, 1, 0, 0},{31470, 1, 1, 16}, {0, 0, 0, 0}}},{19352, {{31470, 1, 0, 0},{31470, 1, 1, 17}, {0, 0, 0, 0}}},{19352, {{31470, 1, 0, 0},{31470, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{19353, {{31471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19354, {{31472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19355, {{31473, 1, 0, 0},{31473, 1, 1, 17}, {0, 0, 0, 0}}},{19355, {{31473, 1, 0, 0},{31473, 1, 1, 17}, {0, 0, 0, 0}}},{19356, {{31473, 1, 0, 0},{31473, 1, 1, 18}, {0, 0, 0, 0}}},{19356, {{31473, 1, 0, 0},{31473, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19357, {{31474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19359, {{31475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19360, {{31476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19362, {{31477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19363, {{31478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19364, {{31479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19365, {{31480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19366, {{31481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19367, {{31482, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19368, {{31483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19369, {{31484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19370, {{31485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19371, {{31486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19372, {{31487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19373, {{31488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19374, {{31489, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19375, {{31490, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{19376, {{31492, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19377, {{31493, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19378, {{31494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19379, {{31495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{19380, {{31496, 1, 0, 0},{31496, 1, 1, 20}, {0, 0, 0, 0}}},{19380, {{31496, 1, 0, 0},{31496, 1, 1, 20}, {0, 0, 0, 0}}},{19381, {{31496, 1, 0, 0},{31496, 1, 1, 19}, {0, 0, 0, 0}}},{19381, {{31496, 1, 0, 0},{31496, 1, 1, 19}, {0, 0, 0, 0}}},{19382, {{31496, 1, 0, 0},{31496, 1, 1, 21}, {0, 0, 0, 0}}},{19382, {{31496, 1, 0, 0},{31496, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{19383, {{31497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19384, {{31498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19385, {{31499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19386, {{31500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19387, {{31501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19388, {{31502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19389, {{31503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19390, {{31504, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19391, {{31505, 1, 0, 0},{31505, 1, 1, 17}, {0, 0, 0, 0}}},{19391, {{31505, 1, 0, 0},{31505, 1, 1, 17}, {0, 0, 0, 0}}},{60016, {{31505, 1, 0, 0},{31505, 1, 1, 18}, {0, 0, 0, 0}}},{60016, {{31505, 1, 0, 0},{31505, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19392, {{31506, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19393, {{31507, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19394, {{31508, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{19395, {{31510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19396, {{31511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19398, {{31512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19399, {{31513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19400, {{31514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19401, {{31515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19402, {{31516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19403, {{31517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19404, {{31518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19405, {{31519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19406, {{31520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19407, {{31521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19408, {{31522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19409, {{31523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19410, {{31524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19411, {{31525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19412, {{31526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19413, {{31527, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19414, {{31528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19415, {{31529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19416, {{31530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19417, {{31531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19418, {{31532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19419, {{31533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19421, {{31534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19422, {{31535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19423, {{31536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19425, {{31537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19426, {{31538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19427, {{31539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19428, {{31540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19429, {{31541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19431, {{31542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19432, {{31543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19433, {{31544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19434, {{31545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{19435, {{31547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{19436, {{31549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{19437, {{31551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19438, {{31552, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19439, {{31553, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19440, {{31554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19441, {{31555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19442, {{31556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19443, {{31557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19444, {{31558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19445, {{31559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19446, {{31560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19447, {{31561, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19448, {{31562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19449, {{31563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19450, {{31564, 1, 0, 0},{31564, 1, 1, 17}, {0, 0, 0, 0}}},{19450, {{31564, 1, 0, 0},{31564, 1, 1, 17}, {0, 0, 0, 0}}},{19451, {{31564, 1, 0, 0},{31564, 1, 1, 18}, {0, 0, 0, 0}}},{19451, {{31564, 1, 0, 0},{31564, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19452, {{31565, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19453, {{31566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19454, {{31567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19455, {{31568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{19456, {{31569, 1, 0, 0},{31569, 1, 1, 20}, {0, 0, 0, 0}}},{19456, {{31569, 1, 0, 0},{31569, 1, 1, 20}, {0, 0, 0, 0}}},{19457, {{31569, 1, 0, 0},{31569, 1, 1, 18}, {0, 0, 0, 0}}},{19457, {{31569, 1, 0, 0},{31569, 1, 1, 18}, {0, 0, 0, 0}}},{19458, {{31569, 1, 0, 0},{31569, 1, 1, 19}, {0, 0, 0, 0}}},{19458, {{31569, 1, 0, 0},{31569, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{19459, {{31570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19460, {{31571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19462, {{31572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19463, {{31573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19464, {{31574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19465, {{31575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19466, {{31576, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{19467, {{31581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19468, {{31582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19469, {{31583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19470, {{31584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19471, {{31585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19472, {{31586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19473, {{31587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19474, {{31588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19475, {{31589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19476, {{31590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19477, {{31591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19478, {{31592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19479, {{31593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19480, {{31594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19481, {{31595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19482, {{31596, 1, 0, 0},{31596, 1, 1, 17}, {0, 0, 0, 0}}},{19482, {{31596, 1, 0, 0},{31596, 1, 1, 17}, {0, 0, 0, 0}}},{19483, {{31596, 1, 0, 0},{31596, 1, 1, 18}, {0, 0, 0, 0}}},{19483, {{31596, 1, 0, 0},{31596, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{19484, {{31597, 1, 0, 0},{31597, 1, 1, 17}, {0, 0, 0, 0}}},{19484, {{31597, 1, 0, 0},{31597, 1, 1, 17}, {0, 0, 0, 0}}},{19485, {{31597, 1, 0, 0},{31597, 1, 1, 18}, {0, 0, 0, 0}}},{19485, {{31597, 1, 0, 0},{31597, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19486, {{31598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19487, {{31599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19488, {{31600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19489, {{31601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19490, {{31602, 1, 0, 0},{31602, 1, 1, 17}, {0, 0, 0, 0}}},{19490, {{31602, 1, 0, 0},{31602, 1, 1, 17}, {0, 0, 0, 0}}},{19491, {{31602, 1, 0, 0},{31602, 1, 1, 18}, {0, 0, 0, 0}}},{19491, {{31602, 1, 0, 0},{31602, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19492, {{31603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19493, {{31604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{19494, {{31605, 1, 0, 0},{31605, 1, 1, 20}, {0, 0, 0, 0}}},{19494, {{31605, 1, 0, 0},{31605, 1, 1, 20}, {0, 0, 0, 0}}},{19495, {{31605, 1, 0, 0},{31605, 1, 1, 21}, {0, 0, 0, 0}}},{19495, {{31605, 1, 0, 0},{31605, 1, 1, 21}, {0, 0, 0, 0}}},{19496, {{31605, 1, 0, 0},{31605, 1, 1, 18}, {0, 0, 0, 0}}},{19496, {{31605, 1, 0, 0},{31605, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19497, {{31606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19498, {{31607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19499, {{31608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19500, {{31609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19501, {{31610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{68041, {{31613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{19502, {{31615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{19503, {{31617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19504, {{31618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19505, {{31619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19506, {{31620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19507, {{31621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19508, {{31622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19509, {{31623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19510, {{31624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19511, {{31625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19512, {{31626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19513, {{31627, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19514, {{31628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19515, {{31629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19516, {{31630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19518, {{31631, 1, 0, 0},{31631, 1, 1, 17}, {0, 0, 0, 0}}},{19518, {{31631, 1, 0, 0},{31631, 1, 1, 17}, {0, 0, 0, 0}}},{19519, {{31631, 1, 0, 0},{31631, 1, 1, 18}, {0, 0, 0, 0}}},{19519, {{31631, 1, 0, 0},{31631, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{19520, {{31632, 1, 0, 0},{31632, 1, 1, 17}, {0, 0, 0, 0}}},{19520, {{31632, 1, 0, 0},{31632, 1, 1, 17}, {0, 0, 0, 0}}},{19521, {{31632, 1, 0, 0},{31632, 1, 1, 18}, {0, 0, 0, 0}}},{19521, {{31632, 1, 0, 0},{31632, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19522, {{31633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19523, {{31634, 1, 0, 0},{31634, 1, 1, 17}, {0, 0, 0, 0}}},{19523, {{31634, 1, 0, 0},{31634, 1, 1, 17}, {0, 0, 0, 0}}},{19524, {{31634, 1, 0, 0},{31634, 1, 1, 18}, {0, 0, 0, 0}}},{19524, {{31634, 1, 0, 0},{31634, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19525, {{31635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19526, {{31636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19527, {{31637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19528, {{31638, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19529, {{31639, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19530, {{31640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19531, {{31641, 1, 0, 0},{31641, 1, 1, 19}, {0, 0, 0, 0}}},{19531, {{31641, 1, 0, 0},{31641, 1, 1, 19}, {0, 0, 0, 0}}},{19532, {{31641, 1, 0, 0},{31641, 1, 1, 18}, {0, 0, 0, 0}}},{19532, {{31641, 1, 0, 0},{31641, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19533, {{31642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19534, {{31643, 1, 0, 0},{31643, 1, 1, 18}, {0, 0, 0, 0}}},{19534, {{31643, 1, 0, 0},{31643, 1, 1, 18}, {0, 0, 0, 0}}},{19535, {{31643, 1, 0, 0},{31643, 1, 1, 17}, {0, 0, 0, 0}}},{19535, {{31643, 1, 0, 0},{31643, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{19536, {{31644, 1, 0, 0},{31644, 1, 1, 17}, {0, 0, 0, 0}}},{19536, {{31644, 1, 0, 0},{31644, 1, 1, 17}, {0, 0, 0, 0}}},{19537, {{31644, 1, 0, 0},{31644, 1, 1, 18}, {0, 0, 0, 0}}},{19537, {{31644, 1, 0, 0},{31644, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19538, {{31645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{19539, {{31646, 1, 0, 0},{31646, 1, 1, 17}, {0, 0, 0, 0}}},{19539, {{31646, 1, 0, 0},{31646, 1, 1, 17}, {0, 0, 0, 0}}},{19540, {{31646, 1, 0, 0},{31646, 1, 1, 18}, {0, 0, 0, 0}}},{19540, {{31646, 1, 0, 0},{31646, 1, 1, 18}, {0, 0, 0, 0}}},{19542, {{31646, 1, 0, 0},{31646, 1, 1, 19}, {0, 0, 0, 0}}},{19542, {{31646, 1, 0, 0},{31646, 1, 1, 19}, {0, 0, 0, 0}}},{19543, {{31646, 1, 0, 0},{31646, 1, 1, 20}, {0, 0, 0, 0}}},{19543, {{31646, 1, 0, 0},{31646, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{19544, {{31647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19545, {{31648, 1, 0, 0},{31648, 1, 1, 17}, {0, 0, 0, 0}}},{19545, {{31648, 1, 0, 0},{31648, 1, 1, 17}, {0, 0, 0, 0}}},{19546, {{31648, 1, 0, 0},{31648, 1, 1, 18}, {0, 0, 0, 0}}},{19546, {{31648, 1, 0, 0},{31648, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19547, {{31649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19548, {{31650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19549, {{31651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19550, {{31652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19551, {{31653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{19552, {{31658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{19553, {{31660, 1, 0, 0},{31660, 1, 1, 17}, {0, 0, 0, 0}}},{19553, {{31660, 1, 0, 0},{31660, 1, 1, 17}, {0, 0, 0, 0}}},{19554, {{31660, 1, 0, 0},{31660, 1, 1, 18}, {0, 0, 0, 0}}},{19554, {{31660, 1, 0, 0},{31660, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{19555, {{31661, 1, 0, 0},{31661, 1, 1, 19}, {0, 0, 0, 0}}},{19555, {{31661, 1, 0, 0},{31661, 1, 1, 19}, {0, 0, 0, 0}}},{19556, {{31661, 1, 0, 0},{31661, 1, 1, 18}, {0, 0, 0, 0}}},{19556, {{31661, 1, 0, 0},{31661, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19557, {{31662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19558, {{31663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19559, {{31664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19560, {{31665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19561, {{31666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19562, {{31667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19563, {{31668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19564, {{31669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19565, {{31670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19566, {{31671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19567, {{31672, 1, 0, 0},{31672, 1, 1, 18}, {0, 0, 0, 0}}},{19567, {{31672, 1, 0, 0},{31672, 1, 1, 18}, {0, 0, 0, 0}}},{19568, {{31672, 1, 0, 0},{31672, 1, 1, 19}, {0, 0, 0, 0}}},{19568, {{31672, 1, 0, 0},{31672, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{19569, {{31673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19570, {{31674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19571, {{31675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19572, {{31676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19573, {{31677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19574, {{31678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19575, {{31679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {7, {{19576, {{31680, 1, 0, 0},{31680, 1, 1, 19}, {0, 0, 0, 0}}},{19576, {{31680, 1, 0, 0},{31680, 1, 1, 19}, {0, 0, 0, 0}}},{19577, {{31680, 1, 0, 0},{31680, 1, 1, 20}, {0, 0, 0, 0}}},{19577, {{31680, 1, 0, 0},{31680, 1, 1, 20}, {0, 0, 0, 0}}},{30285, {{31680, 1, 0, 0},{31680, 1, 1, 0},{31680, 1, 1, 21},{64086, 1, 0, 0}, {0, 0, 0, 0}}},{30285, {{31680, 1, 0, 0},{31680, 1, 1, 0},{31680, 1, 1, 21},{64086, 1, 0, 0}, {0, 0, 0, 0}}},{30285, {{31680, 1, 0, 0},{31680, 1, 1, 0},{31680, 1, 1, 21},{64086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19579, {{31681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19580, {{31682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19581, {{31683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{19582, {{31684, 1, 0, 0},{31684, 1, 1, 17}, {0, 0, 0, 0}}},{19582, {{31684, 1, 0, 0},{31684, 1, 1, 17}, {0, 0, 0, 0}}},{19583, {{31684, 1, 0, 0},{31684, 1, 1, 18}, {0, 0, 0, 0}}},{19583, {{31684, 1, 0, 0},{31684, 1, 1, 18}, {0, 0, 0, 0}}},{19584, {{31684, 1, 0, 0},{31684, 1, 1, 19}, {0, 0, 0, 0}}},{19584, {{31684, 1, 0, 0},{31684, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{19585, {{31685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19586, {{31686, 1, 0, 0},{31686, 1, 1, 17}, {0, 0, 0, 0}}},{19586, {{31686, 1, 0, 0},{31686, 1, 1, 17}, {0, 0, 0, 0}}},{19587, {{31686, 1, 0, 0},{31686, 1, 1, 18}, {0, 0, 0, 0}}},{19587, {{31686, 1, 0, 0},{31686, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{19589, {{31687, 1, 0, 0},{31687, 1, 1, 19}, {0, 0, 0, 0}}},{19589, {{31687, 1, 0, 0},{31687, 1, 1, 19}, {0, 0, 0, 0}}},{19590, {{31687, 1, 0, 0},{31687, 1, 1, 18}, {0, 0, 0, 0}}},{19590, {{31687, 1, 0, 0},{31687, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19591, {{31688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19592, {{31689, 1, 0, 0},{31689, 1, 1, 18}, {0, 0, 0, 0}}},{19592, {{31689, 1, 0, 0},{31689, 1, 1, 18}, {0, 0, 0, 0}}},{19593, {{31689, 1, 0, 0},{31689, 1, 1, 19}, {0, 0, 0, 0}}},{19593, {{31689, 1, 0, 0},{31689, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{19594, {{31690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19595, {{31691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19596, {{31692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19597, {{31693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19598, {{31694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19599, {{31695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{68042, {{31698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{19600, {{31700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19601, {{31701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19602, {{31702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19603, {{31703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19604, {{31704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19605, {{31705, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19607, {{31706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19608, {{31707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19609, {{31708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19610, {{31709, 1, 0, 0},{31709, 1, 1, 19}, {0, 0, 0, 0}}},{19610, {{31709, 1, 0, 0},{31709, 1, 1, 19}, {0, 0, 0, 0}}},{19611, {{31709, 1, 0, 0},{31709, 1, 1, 18}, {0, 0, 0, 0}}},{19611, {{31709, 1, 0, 0},{31709, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19612, {{31710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19613, {{31711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19614, {{31712, 1, 0, 0},{31712, 1, 1, 19}, {0, 0, 0, 0}}},{19614, {{31712, 1, 0, 0},{31712, 1, 1, 19}, {0, 0, 0, 0}}},{19615, {{31712, 1, 0, 0},{31712, 1, 1, 18}, {0, 0, 0, 0}}},{19615, {{31712, 1, 0, 0},{31712, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19616, {{31713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19617, {{31714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19618, {{31715, 1, 0, 0},{31715, 1, 1, 16}, {0, 0, 0, 0}}},{19618, {{31715, 1, 0, 0},{31715, 1, 1, 16}, {0, 0, 0, 0}}},{19619, {{31715, 1, 0, 0},{31715, 1, 1, 17}, {0, 0, 0, 0}}},{19619, {{31715, 1, 0, 0},{31715, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{19620, {{31716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19621, {{31717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19622, {{31718, 1, 0, 0},{31718, 1, 1, 17}, {0, 0, 0, 0}}},{19622, {{31718, 1, 0, 0},{31718, 1, 1, 17}, {0, 0, 0, 0}}},{60027, {{31718, 1, 0, 0},{31718, 1, 1, 18}, {0, 0, 0, 0}}},{60027, {{31718, 1, 0, 0},{31718, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19623, {{31719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19624, {{31720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19625, {{31721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19626, {{31722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19627, {{31723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19629, {{31724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19630, {{31725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{19631, {{31728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19632, {{31729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19633, {{31730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19634, {{31731, 1, 0, 0},{31731, 1, 1, 18}, {0, 0, 0, 0}}},{19634, {{31731, 1, 0, 0},{31731, 1, 1, 18}, {0, 0, 0, 0}}},{19635, {{31731, 1, 0, 0},{31731, 1, 1, 17}, {0, 0, 0, 0}}},{19635, {{31731, 1, 0, 0},{31731, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{19636, {{31732, 1, 0, 0},{31732, 1, 1, 18}, {0, 0, 0, 0}}},{19636, {{31732, 1, 0, 0},{31732, 1, 1, 18}, {0, 0, 0, 0}}},{19637, {{31732, 1, 0, 0},{31732, 1, 1, 17}, {0, 0, 0, 0}}},{19637, {{31732, 1, 0, 0},{31732, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{19638, {{31733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19639, {{31734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{19640, {{31735, 1, 0, 0},{31735, 1, 1, 19}, {0, 0, 0, 0}}},{19640, {{31735, 1, 0, 0},{31735, 1, 1, 19}, {0, 0, 0, 0}}},{19641, {{31735, 1, 0, 0},{31735, 1, 1, 17}, {0, 0, 0, 0}}},{19641, {{31735, 1, 0, 0},{31735, 1, 1, 17}, {0, 0, 0, 0}}},{19642, {{31735, 1, 0, 0},{31735, 1, 1, 18}, {0, 0, 0, 0}}},{19642, {{31735, 1, 0, 0},{31735, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19643, {{31736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19644, {{31737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19645, {{31738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19646, {{31739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19647, {{31740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19648, {{31741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19649, {{31742, 1, 0, 0},{31742, 1, 1, 17}, {0, 0, 0, 0}}},{19649, {{31742, 1, 0, 0},{31742, 1, 1, 17}, {0, 0, 0, 0}}},{19650, {{31742, 1, 0, 0},{31742, 1, 1, 18}, {0, 0, 0, 0}}},{19650, {{31742, 1, 0, 0},{31742, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19651, {{31743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19652, {{31744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19653, {{31745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19654, {{31746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19655, {{31747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19656, {{31748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19657, {{31749, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19658, {{31750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19659, {{31751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19660, {{31752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{19661, {{31753, 1, 0, 0},{31753, 1, 1, 18}, {0, 0, 0, 0}}},{19661, {{31753, 1, 0, 0},{31753, 1, 1, 18}, {0, 0, 0, 0}}},{19662, {{31753, 1, 0, 0},{31753, 1, 1, 17}, {0, 0, 0, 0}}},{19662, {{31753, 1, 0, 0},{31753, 1, 1, 17}, {0, 0, 0, 0}}},{19663, {{31753, 1, 0, 0},{31753, 1, 1, 19}, {0, 0, 0, 0}}},{19663, {{31753, 1, 0, 0},{31753, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{19664, {{31754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19665, {{31755, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19666, {{31756, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19667, {{31757, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19668, {{31758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19669, {{31759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19670, {{31760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19671, {{31761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19672, {{31762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{19673, {{31763, 1, 0, 0},{31763, 1, 1, 17}, {0, 0, 0, 0}}},{19673, {{31763, 1, 0, 0},{31763, 1, 1, 17}, {0, 0, 0, 0}}},{19674, {{31763, 1, 0, 0},{31763, 1, 1, 19}, {0, 0, 0, 0}}},{19674, {{31763, 1, 0, 0},{31763, 1, 1, 19}, {0, 0, 0, 0}}},{19675, {{31763, 1, 0, 0},{31763, 1, 1, 18}, {0, 0, 0, 0}}},{19675, {{31763, 1, 0, 0},{31763, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19676, {{31764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{19677, {{31767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19678, {{31768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19679, {{31769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19680, {{31770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19681, {{31771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19682, {{31772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19683, {{31773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19684, {{31774, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19685, {{31775, 1, 0, 0},{31775, 1, 1, 18}, {0, 0, 0, 0}}},{19685, {{31775, 1, 0, 0},{31775, 1, 1, 18}, {0, 0, 0, 0}}},{19686, {{31775, 1, 0, 0},{31775, 1, 1, 17}, {0, 0, 0, 0}}},{19686, {{31775, 1, 0, 0},{31775, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{19687, {{31776, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19688, {{31777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19689, {{31778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19690, {{31779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19691, {{31780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19692, {{31781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19693, {{31782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19694, {{31783, 1, 0, 0},{31783, 1, 1, 17}, {0, 0, 0, 0}}},{19694, {{31783, 1, 0, 0},{31783, 1, 1, 17}, {0, 0, 0, 0}}},{19695, {{31783, 1, 0, 0},{31783, 1, 1, 18}, {0, 0, 0, 0}}},{19695, {{31783, 1, 0, 0},{31783, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19696, {{31784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19697, {{31785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{19698, {{31786, 1, 0, 0},{31786, 1, 1, 19}, {0, 0, 0, 0}}},{19698, {{31786, 1, 0, 0},{31786, 1, 1, 19}, {0, 0, 0, 0}}},{19699, {{31786, 1, 0, 0},{31786, 1, 1, 18}, {0, 0, 0, 0}}},{19699, {{31786, 1, 0, 0},{31786, 1, 1, 18}, {0, 0, 0, 0}}},{19700, {{31786, 1, 0, 0},{31786, 1, 1, 17}, {0, 0, 0, 0}}},{19700, {{31786, 1, 0, 0},{31786, 1, 1, 17}, {0, 0, 0, 0}}},{19701, {{31786, 1, 0, 0},{31786, 1, 1, 20}, {0, 0, 0, 0}}},{19701, {{31786, 1, 0, 0},{31786, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{19702, {{31787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19703, {{31788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19704, {{31789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{19705, {{31793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{19706, {{31795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19707, {{31796, 1, 0, 0},{31796, 1, 1, 17}, {0, 0, 0, 0}}},{19707, {{31796, 1, 0, 0},{31796, 1, 1, 17}, {0, 0, 0, 0}}},{19708, {{31796, 1, 0, 0},{31796, 1, 1, 18}, {0, 0, 0, 0}}},{19708, {{31796, 1, 0, 0},{31796, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19709, {{31797, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19710, {{31798, 1, 0, 0},{31798, 1, 1, 18}, {0, 0, 0, 0}}},{19710, {{31798, 1, 0, 0},{31798, 1, 1, 18}, {0, 0, 0, 0}}},{19711, {{31798, 1, 0, 0},{31798, 1, 1, 17}, {0, 0, 0, 0}}},{19711, {{31798, 1, 0, 0},{31798, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{19713, {{31799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19714, {{31800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19715, {{31801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19716, {{31802, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19717, {{31803, 1, 0, 0},{31803, 1, 1, 16}, {0, 0, 0, 0}}},{19717, {{31803, 1, 0, 0},{31803, 1, 1, 16}, {0, 0, 0, 0}}},{19718, {{31803, 1, 0, 0},{31803, 1, 1, 17}, {0, 0, 0, 0}}},{19718, {{31803, 1, 0, 0},{31803, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{19719, {{31804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19720, {{31805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{19721, {{31806, 1, 0, 0},{31806, 1, 1, 20}, {0, 0, 0, 0}}},{19721, {{31806, 1, 0, 0},{31806, 1, 1, 20}, {0, 0, 0, 0}}},{19722, {{31806, 1, 0, 0},{31806, 1, 1, 19}, {0, 0, 0, 0}}},{19722, {{31806, 1, 0, 0},{31806, 1, 1, 19}, {0, 0, 0, 0}}},{19723, {{31806, 1, 0, 0},{31806, 1, 1, 21}, {0, 0, 0, 0}}},{19723, {{31806, 1, 0, 0},{31806, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {4, {{19724, {{31807, 1, 0, 0},{31807, 1, 1, 19}, {0, 0, 0, 0}}},{19724, {{31807, 1, 0, 0},{31807, 1, 1, 19}, {0, 0, 0, 0}}},{19725, {{31807, 1, 0, 0},{31807, 1, 1, 18}, {0, 0, 0, 0}}},{19725, {{31807, 1, 0, 0},{31807, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19726, {{31808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{19727, {{31811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19728, {{31812, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19729, {{31813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19730, {{31814, 1, 0, 0},{31814, 1, 1, 17}, {0, 0, 0, 0}}},{19730, {{31814, 1, 0, 0},{31814, 1, 1, 17}, {0, 0, 0, 0}}},{19731, {{31814, 1, 0, 0},{31814, 1, 1, 18}, {0, 0, 0, 0}}},{19731, {{31814, 1, 0, 0},{31814, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19732, {{31815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19733, {{31816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19734, {{31817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19735, {{31818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19736, {{31819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19737, {{31820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19738, {{31821, 1, 0, 0},{31821, 1, 1, 19}, {0, 0, 0, 0}}},{19738, {{31821, 1, 0, 0},{31821, 1, 1, 19}, {0, 0, 0, 0}}},{19739, {{31821, 1, 0, 0},{31821, 1, 1, 18}, {0, 0, 0, 0}}},{19739, {{31821, 1, 0, 0},{31821, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{19740, {{31823, 1, 0, 0},{31823, 1, 1, 17}, {0, 0, 0, 0}}},{19740, {{31823, 1, 0, 0},{31823, 1, 1, 17}, {0, 0, 0, 0}}},{60030, {{31823, 1, 0, 0},{31823, 1, 1, 18}, {0, 0, 0, 0}}},{60030, {{31823, 1, 0, 0},{31823, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{19741, {{31824, 1, 0, 0},{31824, 1, 1, 20}, {0, 0, 0, 0}}},{19741, {{31824, 1, 0, 0},{31824, 1, 1, 20}, {0, 0, 0, 0}}},{19742, {{31824, 1, 0, 0},{31824, 1, 1, 19}, {0, 0, 0, 0}}},{19742, {{31824, 1, 0, 0},{31824, 1, 1, 19}, {0, 0, 0, 0}}},{19743, {{31824, 1, 0, 0},{31824, 1, 1, 21}, {0, 0, 0, 0}}},{19743, {{31824, 1, 0, 0},{31824, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {4, {{19744, {{31825, 1, 0, 0},{31825, 1, 1, 17}, {0, 0, 0, 0}}},{19744, {{31825, 1, 0, 0},{31825, 1, 1, 17}, {0, 0, 0, 0}}},{19745, {{31825, 1, 0, 0},{31825, 1, 1, 18}, {0, 0, 0, 0}}},{19745, {{31825, 1, 0, 0},{31825, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19746, {{31826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19747, {{31827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19748, {{31828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19749, {{31829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19750, {{31830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19751, {{31831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{19752, {{31832, 1, 0, 0},{31832, 1, 1, 19}, {0, 0, 0, 0}}},{19752, {{31832, 1, 0, 0},{31832, 1, 1, 19}, {0, 0, 0, 0}}},{19753, {{31832, 1, 0, 0},{31832, 1, 1, 18}, {0, 0, 0, 0}}},{19753, {{31832, 1, 0, 0},{31832, 1, 1, 18}, {0, 0, 0, 0}}},{19754, {{31832, 1, 0, 0},{31832, 1, 1, 20}, {0, 0, 0, 0}}},{19754, {{31832, 1, 0, 0},{31832, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{19755, {{31833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19757, {{31834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19758, {{31835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19759, {{31836, 1, 0, 0},{31836, 1, 1, 17}, {0, 0, 0, 0}}},{19759, {{31836, 1, 0, 0},{31836, 1, 1, 17}, {0, 0, 0, 0}}},{19760, {{31836, 1, 0, 0},{31836, 1, 1, 18}, {0, 0, 0, 0}}},{19760, {{31836, 1, 0, 0},{31836, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19761, {{31837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19762, {{31838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19763, {{31839, 1, 0, 0},{31839, 1, 1, 17}, {0, 0, 0, 0}}},{19763, {{31839, 1, 0, 0},{31839, 1, 1, 17}, {0, 0, 0, 0}}},{19764, {{31839, 1, 0, 0},{31839, 1, 1, 18}, {0, 0, 0, 0}}},{19764, {{31839, 1, 0, 0},{31839, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {8, {{19765, {{31840, 1, 0, 0},{31840, 1, 1, 17}, {0, 0, 0, 0}}},{19765, {{31840, 1, 0, 0},{31840, 1, 1, 17}, {0, 0, 0, 0}}},{19766, {{31840, 1, 0, 0},{31840, 1, 1, 20}, {0, 0, 0, 0}}},{19766, {{31840, 1, 0, 0},{31840, 1, 1, 20}, {0, 0, 0, 0}}},{19767, {{31840, 1, 0, 0},{31840, 1, 1, 18}, {0, 0, 0, 0}}},{19767, {{31840, 1, 0, 0},{31840, 1, 1, 18}, {0, 0, 0, 0}}},{19768, {{31840, 1, 0, 0},{31840, 1, 1, 19}, {0, 0, 0, 0}}},{19768, {{31840, 1, 0, 0},{31840, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{19769, {{31841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19770, {{31842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19771, {{31843, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19772, {{31844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19773, {{31845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19774, {{31846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{19775, {{31847, 1, 0, 0},{31847, 1, 1, 18}, {0, 0, 0, 0}}},{19775, {{31847, 1, 0, 0},{31847, 1, 1, 18}, {0, 0, 0, 0}}},{19776, {{31847, 1, 0, 0},{31847, 1, 1, 17}, {0, 0, 0, 0}}},{19776, {{31847, 1, 0, 0},{31847, 1, 1, 17}, {0, 0, 0, 0}}},{19777, {{31847, 1, 0, 0},{31847, 1, 1, 19}, {0, 0, 0, 0}}},{19777, {{31847, 1, 0, 0},{31847, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{19778, {{31849, 1, 0, 0},{31849, 1, 1, 18}, {0, 0, 0, 0}}},{19778, {{31849, 1, 0, 0},{31849, 1, 1, 18}, {0, 0, 0, 0}}},{19779, {{31849, 1, 0, 0},{31849, 1, 1, 19}, {0, 0, 0, 0}}},{19779, {{31849, 1, 0, 0},{31849, 1, 1, 19}, {0, 0, 0, 0}}},{19780, {{31849, 1, 0, 0},{31849, 1, 1, 20}, {0, 0, 0, 0}}},{19780, {{31849, 1, 0, 0},{31849, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{19781, {{31850, 1, 0, 0},{31850, 1, 1, 16}, {0, 0, 0, 0}}},{19781, {{31850, 1, 0, 0},{31850, 1, 1, 16}, {0, 0, 0, 0}}},{19782, {{31850, 1, 0, 0},{31850, 1, 1, 17}, {0, 0, 0, 0}}},{19782, {{31850, 1, 0, 0},{31850, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{19783, {{31851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19784, {{31852, 1, 0, 0},{31852, 1, 1, 17}, {0, 0, 0, 0}}},{19784, {{31852, 1, 0, 0},{31852, 1, 1, 17}, {0, 0, 0, 0}}},{19785, {{31852, 1, 0, 0},{31852, 1, 1, 18}, {0, 0, 0, 0}}},{19785, {{31852, 1, 0, 0},{31852, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19786, {{31853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19787, {{31854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19788, {{31855, 1, 0, 0},{31855, 1, 1, 17}, {0, 0, 0, 0}}},{19788, {{31855, 1, 0, 0},{31855, 1, 1, 17}, {0, 0, 0, 0}}},{19789, {{31855, 1, 0, 0},{31855, 1, 1, 18}, {0, 0, 0, 0}}},{19789, {{31855, 1, 0, 0},{31855, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19790, {{31856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19791, {{31857, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19792, {{31858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19793, {{31859, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19795, {{31860, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19796, {{31861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{19797, {{31863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19798, {{31864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19799, {{31865, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19800, {{31866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19801, {{31867, 1, 0, 0},{31867, 1, 1, 17}, {0, 0, 0, 0}}},{19801, {{31867, 1, 0, 0},{31867, 1, 1, 17}, {0, 0, 0, 0}}},{19802, {{31867, 1, 0, 0},{31867, 1, 1, 18}, {0, 0, 0, 0}}},{19802, {{31867, 1, 0, 0},{31867, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19803, {{31868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19804, {{31869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{19805, {{31870, 1, 0, 0},{31870, 1, 1, 21}, {0, 0, 0, 0}}},{19805, {{31870, 1, 0, 0},{31870, 1, 1, 21}, {0, 0, 0, 0}}},{19806, {{31870, 1, 0, 0},{31870, 1, 1, 20}, {0, 0, 0, 0}}},{19806, {{31870, 1, 0, 0},{31870, 1, 1, 20}, {0, 0, 0, 0}}},{19807, {{31870, 1, 0, 0},{31870, 1, 1, 19}, {0, 0, 0, 0}}},{19807, {{31870, 1, 0, 0},{31870, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{19808, {{31871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19809, {{31872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19810, {{31873, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19811, {{31874, 1, 0, 0},{31874, 1, 1, 18}, {0, 0, 0, 0}}},{19811, {{31874, 1, 0, 0},{31874, 1, 1, 18}, {0, 0, 0, 0}}},{19812, {{31874, 1, 0, 0},{31874, 1, 1, 19}, {0, 0, 0, 0}}},{19812, {{31874, 1, 0, 0},{31874, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{19813, {{31875, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19814, {{31876, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19815, {{31877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19816, {{31878, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19817, {{31879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19818, {{31880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19819, {{31881, 1, 0, 0},{31881, 1, 1, 18}, {0, 0, 0, 0}}},{19819, {{31881, 1, 0, 0},{31881, 1, 1, 18}, {0, 0, 0, 0}}},{19820, {{31881, 1, 0, 0},{31881, 1, 1, 19}, {0, 0, 0, 0}}},{19820, {{31881, 1, 0, 0},{31881, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{19821, {{31882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19822, {{31883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19823, {{31884, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19824, {{31885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19825, {{31886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19826, {{31887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19827, {{31888, 1, 0, 0},{31888, 1, 1, 18}, {0, 0, 0, 0}}},{19827, {{31888, 1, 0, 0},{31888, 1, 1, 18}, {0, 0, 0, 0}}},{19828, {{31888, 1, 0, 0},{31888, 1, 1, 19}, {0, 0, 0, 0}}},{19828, {{31888, 1, 0, 0},{31888, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{19830, {{31890, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19831, {{31891, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19832, {{31892, 1, 0, 0},{31892, 1, 1, 17}, {0, 0, 0, 0}}},{19832, {{31892, 1, 0, 0},{31892, 1, 1, 17}, {0, 0, 0, 0}}},{19833, {{31892, 1, 0, 0},{31892, 1, 1, 18}, {0, 0, 0, 0}}},{19833, {{31892, 1, 0, 0},{31892, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19834, {{31893, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19835, {{31894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19836, {{31895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19837, {{31896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19838, {{31897, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19839, {{31898, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19840, {{31899, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19841, {{31900, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19842, {{31901, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19843, {{31902, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19844, {{31903, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19845, {{31904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19846, {{31905, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19847, {{31906, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19848, {{31907, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19849, {{31908, 1, 0, 0},{31908, 1, 1, 17}, {0, 0, 0, 0}}},{19849, {{31908, 1, 0, 0},{31908, 1, 1, 17}, {0, 0, 0, 0}}},{58289, {{31908, 1, 0, 0},{31908, 1, 1, 18}, {0, 0, 0, 0}}},{58289, {{31908, 1, 0, 0},{31908, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19850, {{31909, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19851, {{31910, 1, 0, 0},{31910, 1, 1, 18}, {0, 0, 0, 0}}},{19851, {{31910, 1, 0, 0},{31910, 1, 1, 18}, {0, 0, 0, 0}}},{19852, {{31910, 1, 0, 0},{31910, 1, 1, 17}, {0, 0, 0, 0}}},{19852, {{31910, 1, 0, 0},{31910, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{19853, {{31911, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19854, {{31912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{19855, {{31915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{19856, {{31917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19857, {{31918, 1, 0, 0},{31918, 1, 1, 18}, {0, 0, 0, 0}}},{19857, {{31918, 1, 0, 0},{31918, 1, 1, 18}, {0, 0, 0, 0}}},{19858, {{31918, 1, 0, 0},{31918, 1, 1, 19}, {0, 0, 0, 0}}},{19858, {{31918, 1, 0, 0},{31918, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{19859, {{31919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19860, {{31920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19862, {{31921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19863, {{31922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19864, {{31923, 1, 0, 0},{31923, 1, 1, 17}, {0, 0, 0, 0}}},{19864, {{31923, 1, 0, 0},{31923, 1, 1, 17}, {0, 0, 0, 0}}},{19865, {{31923, 1, 0, 0},{31923, 1, 1, 18}, {0, 0, 0, 0}}},{19865, {{31923, 1, 0, 0},{31923, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19866, {{31924, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{19867, {{31926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19868, {{31927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19869, {{31928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19870, {{31929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19871, {{31930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19872, {{31931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19873, {{31932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19874, {{31933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{19875, {{31934, 1, 0, 0},{31934, 1, 1, 18}, {0, 0, 0, 0}}},{19875, {{31934, 1, 0, 0},{31934, 1, 1, 18}, {0, 0, 0, 0}}},{30215, {{31934, 1, 0, 0},{31934, 1, 1, 0},{31934, 1, 1, 19},{64029, 1, 0, 0}, {0, 0, 0, 0}}},{30215, {{31934, 1, 0, 0},{31934, 1, 1, 0},{31934, 1, 1, 19},{64029, 1, 0, 0}, {0, 0, 0, 0}}},{30215, {{31934, 1, 0, 0},{31934, 1, 1, 0},{31934, 1, 1, 19},{64029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19876, {{31935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{19877, {{31936, 1, 0, 0},{31936, 1, 1, 17}, {0, 0, 0, 0}}},{19877, {{31936, 1, 0, 0},{31936, 1, 1, 17}, {0, 0, 0, 0}}},{19878, {{31936, 1, 0, 0},{31936, 1, 1, 18}, {0, 0, 0, 0}}},{19878, {{31936, 1, 0, 0},{31936, 1, 1, 18}, {0, 0, 0, 0}}},{19879, {{31936, 1, 0, 0},{31936, 1, 1, 19}, {0, 0, 0, 0}}},{19879, {{31936, 1, 0, 0},{31936, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{19880, {{31938, 1, 0, 0},{31938, 1, 1, 17}, {0, 0, 0, 0}}},{19880, {{31938, 1, 0, 0},{31938, 1, 1, 17}, {0, 0, 0, 0}}},{19881, {{31938, 1, 0, 0},{31938, 1, 1, 18}, {0, 0, 0, 0}}},{19881, {{31938, 1, 0, 0},{31938, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19882, {{31939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19883, {{31940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19884, {{31941, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19885, {{31942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19886, {{31943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19887, {{31944, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19888, {{31945, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19889, {{31946, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19890, {{31947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19891, {{31948, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19892, {{31949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19893, {{31950, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19894, {{31951, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19895, {{31952, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19896, {{31953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19897, {{31954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19898, {{31955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19901, {{31956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19902, {{31957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19903, {{31958, 1, 0, 0},{31958, 1, 1, 19}, {0, 0, 0, 0}}},{19903, {{31958, 1, 0, 0},{31958, 1, 1, 19}, {0, 0, 0, 0}}},{19904, {{31958, 1, 0, 0},{31958, 1, 1, 18}, {0, 0, 0, 0}}},{19904, {{31958, 1, 0, 0},{31958, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19905, {{31959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19906, {{31960, 1, 0, 0},{31960, 1, 1, 17}, {0, 0, 0, 0}}},{19906, {{31960, 1, 0, 0},{31960, 1, 1, 17}, {0, 0, 0, 0}}},{19907, {{31960, 1, 0, 0},{31960, 1, 1, 18}, {0, 0, 0, 0}}},{19907, {{31960, 1, 0, 0},{31960, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{19908, {{31961, 1, 0, 0},{31961, 1, 1, 18}, {0, 0, 0, 0}}},{19908, {{31961, 1, 0, 0},{31961, 1, 1, 18}, {0, 0, 0, 0}}},{19909, {{31961, 1, 0, 0},{31961, 1, 1, 17}, {0, 0, 0, 0}}},{19909, {{31961, 1, 0, 0},{31961, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{19910, {{31962, 1, 0, 0},{31962, 1, 1, 17}, {0, 0, 0, 0}}},{19910, {{31962, 1, 0, 0},{31962, 1, 1, 17}, {0, 0, 0, 0}}},{19911, {{31962, 1, 0, 0},{31962, 1, 1, 18}, {0, 0, 0, 0}}},{19911, {{31962, 1, 0, 0},{31962, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19913, {{31963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19914, {{31964, 1, 0, 0},{31964, 1, 1, 17}, {0, 0, 0, 0}}},{19914, {{31964, 1, 0, 0},{31964, 1, 1, 17}, {0, 0, 0, 0}}},{19915, {{31964, 1, 0, 0},{31964, 1, 1, 18}, {0, 0, 0, 0}}},{19915, {{31964, 1, 0, 0},{31964, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19916, {{31965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19917, {{31966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19918, {{31967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19919, {{31968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19920, {{31969, 1, 0, 0},{31969, 1, 1, 16}, {0, 0, 0, 0}}},{19920, {{31969, 1, 0, 0},{31969, 1, 1, 16}, {0, 0, 0, 0}}},{19921, {{31969, 1, 0, 0},{31969, 1, 1, 17}, {0, 0, 0, 0}}},{19921, {{31969, 1, 0, 0},{31969, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{19922, {{31970, 1, 0, 0},{31970, 1, 1, 17}, {0, 0, 0, 0}}},{19922, {{31970, 1, 0, 0},{31970, 1, 1, 17}, {0, 0, 0, 0}}},{19923, {{31970, 1, 0, 0},{31970, 1, 1, 18}, {0, 0, 0, 0}}},{19923, {{31970, 1, 0, 0},{31970, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19924, {{31971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19925, {{31972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19926, {{31973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19927, {{31974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19928, {{31975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19929, {{31976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19930, {{31977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19931, {{31978, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19932, {{31979, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{19933, {{31982, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19934, {{31983, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19935, {{31984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19936, {{31985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19937, {{31986, 1, 0, 0},{31986, 1, 1, 18}, {0, 0, 0, 0}}},{19937, {{31986, 1, 0, 0},{31986, 1, 1, 18}, {0, 0, 0, 0}}},{19938, {{31986, 1, 0, 0},{31986, 1, 1, 19}, {0, 0, 0, 0}}},{19938, {{31986, 1, 0, 0},{31986, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{19939, {{31987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19940, {{31988, 1, 0, 0},{31988, 1, 1, 17}, {0, 0, 0, 0}}},{19940, {{31988, 1, 0, 0},{31988, 1, 1, 17}, {0, 0, 0, 0}}},{19941, {{31988, 1, 0, 0},{31988, 1, 1, 18}, {0, 0, 0, 0}}},{19941, {{31988, 1, 0, 0},{31988, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{19942, {{31989, 1, 0, 0},{31989, 1, 1, 17}, {0, 0, 0, 0}}},{19942, {{31989, 1, 0, 0},{31989, 1, 1, 17}, {0, 0, 0, 0}}},{19943, {{31989, 1, 0, 0},{31989, 1, 1, 18}, {0, 0, 0, 0}}},{19943, {{31989, 1, 0, 0},{31989, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{19944, {{31990, 1, 0, 0},{31990, 1, 1, 17}, {0, 0, 0, 0}}},{19944, {{31990, 1, 0, 0},{31990, 1, 1, 17}, {0, 0, 0, 0}}},{19945, {{31990, 1, 0, 0},{31990, 1, 1, 18}, {0, 0, 0, 0}}},{19945, {{31990, 1, 0, 0},{31990, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19946, {{31991, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19947, {{31992, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{19948, {{31994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19949, {{31995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19950, {{31996, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19951, {{31997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19952, {{31998, 1, 0, 0},{31998, 1, 1, 18}, {0, 0, 0, 0}}},{19952, {{31998, 1, 0, 0},{31998, 1, 1, 18}, {0, 0, 0, 0}}},{19953, {{31998, 1, 0, 0},{31998, 1, 1, 17}, {0, 0, 0, 0}}},{19953, {{31998, 1, 0, 0},{31998, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {8, {{19955, {{32000, 1, 0, 0},{32000, 1, 1, 17}, {0, 0, 0, 0}}},{19955, {{32000, 1, 0, 0},{32000, 1, 1, 17}, {0, 0, 0, 0}}},{19956, {{32000, 1, 0, 0},{32000, 1, 1, 18}, {0, 0, 0, 0}}},{19956, {{32000, 1, 0, 0},{32000, 1, 1, 18}, {0, 0, 0, 0}}},{19957, {{32000, 1, 0, 0},{32000, 1, 1, 19}, {0, 0, 0, 0}}},{19957, {{32000, 1, 0, 0},{32000, 1, 1, 19}, {0, 0, 0, 0}}},{19958, {{32000, 1, 0, 0},{32000, 1, 1, 20}, {0, 0, 0, 0}}},{19958, {{32000, 1, 0, 0},{32000, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{19962, {{32001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19963, {{32002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19964, {{32003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19965, {{32004, 1, 0, 0},{32004, 1, 1, 18}, {0, 0, 0, 0}}},{19965, {{32004, 1, 0, 0},{32004, 1, 1, 18}, {0, 0, 0, 0}}},{19966, {{32004, 1, 0, 0},{32004, 1, 1, 19}, {0, 0, 0, 0}}},{19966, {{32004, 1, 0, 0},{32004, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{19967, {{32005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19969, {{32006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19970, {{32007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19971, {{32008, 1, 0, 0},{32008, 1, 1, 17}, {0, 0, 0, 0}}},{19971, {{32008, 1, 0, 0},{32008, 1, 1, 17}, {0, 0, 0, 0}}},{19972, {{32008, 1, 0, 0},{32008, 1, 1, 18}, {0, 0, 0, 0}}},{19972, {{32008, 1, 0, 0},{32008, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19973, {{32009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19975, {{32010, 1, 0, 0},{32010, 1, 1, 17}, {0, 0, 0, 0}}},{19975, {{32010, 1, 0, 0},{32010, 1, 1, 17}, {0, 0, 0, 0}}},{19976, {{32010, 1, 0, 0},{32010, 1, 1, 18}, {0, 0, 0, 0}}},{19976, {{32010, 1, 0, 0},{32010, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{19977, {{32011, 1, 0, 0},{32011, 1, 1, 18}, {0, 0, 0, 0}}},{19977, {{32011, 1, 0, 0},{32011, 1, 1, 18}, {0, 0, 0, 0}}},{19978, {{32011, 1, 0, 0},{32011, 1, 1, 19}, {0, 0, 0, 0}}},{19978, {{32011, 1, 0, 0},{32011, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{19979, {{32012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{19980, {{32013, 1, 0, 0},{32013, 1, 1, 19}, {0, 0, 0, 0}}},{19980, {{32013, 1, 0, 0},{32013, 1, 1, 19}, {0, 0, 0, 0}}},{19981, {{32013, 1, 0, 0},{32013, 1, 1, 18}, {0, 0, 0, 0}}},{19981, {{32013, 1, 0, 0},{32013, 1, 1, 18}, {0, 0, 0, 0}}},{19982, {{32013, 1, 0, 0},{32013, 1, 1, 20}, {0, 0, 0, 0}}},{19982, {{32013, 1, 0, 0},{32013, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{19984, {{32014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19985, {{32015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19986, {{32016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19987, {{32017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19988, {{32018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19989, {{32019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{19990, {{32020, 1, 0, 0},{32020, 1, 1, 17}, {0, 0, 0, 0}}},{19990, {{32020, 1, 0, 0},{32020, 1, 1, 17}, {0, 0, 0, 0}}},{19991, {{32020, 1, 0, 0},{32020, 1, 1, 18}, {0, 0, 0, 0}}},{19991, {{32020, 1, 0, 0},{32020, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{19994, {{32021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19995, {{32022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19996, {{32023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19997, {{32024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19998, {{32025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20000, {{32026, 1, 0, 0},{32026, 1, 1, 18}, {0, 0, 0, 0}}},{20000, {{32026, 1, 0, 0},{32026, 1, 1, 18}, {0, 0, 0, 0}}},{20001, {{32026, 1, 0, 0},{32026, 1, 1, 19}, {0, 0, 0, 0}}},{20001, {{32026, 1, 0, 0},{32026, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{20002, {{32027, 1, 0, 0},{32027, 1, 1, 18}, {0, 0, 0, 0}}},{20002, {{32027, 1, 0, 0},{32027, 1, 1, 18}, {0, 0, 0, 0}}},{20003, {{32027, 1, 0, 0},{32027, 1, 1, 19}, {0, 0, 0, 0}}},{20003, {{32027, 1, 0, 0},{32027, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{20004, {{32028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20005, {{32029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20006, {{32030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20007, {{32031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20008, {{32032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20009, {{32033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20010, {{32034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20011, {{32035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20012, {{32036, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20013, {{32037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20014, {{32038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{20015, {{32040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20016, {{32041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20017, {{32042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20018, {{32043, 1, 0, 0},{32043, 1, 1, 17}, {0, 0, 0, 0}}},{20018, {{32043, 1, 0, 0},{32043, 1, 1, 17}, {0, 0, 0, 0}}},{20019, {{32043, 1, 0, 0},{32043, 1, 1, 18}, {0, 0, 0, 0}}},{20019, {{32043, 1, 0, 0},{32043, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20020, {{32044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20021, {{32045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20022, {{32046, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20023, {{32047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20024, {{32048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20026, {{32049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20029, {{32050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20030, {{32051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20031, {{32052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20032, {{32053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20033, {{32054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20034, {{32055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20035, {{32056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20036, {{32057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20038, {{32058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20039, {{32059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20040, {{32060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20041, {{32061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20042, {{32062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20043, {{32063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20044, {{32064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20045, {{32065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20046, {{32066, 1, 0, 0},{32066, 1, 1, 19}, {0, 0, 0, 0}}},{20046, {{32066, 1, 0, 0},{32066, 1, 1, 19}, {0, 0, 0, 0}}},{20047, {{32066, 1, 0, 0},{32066, 1, 1, 18}, {0, 0, 0, 0}}},{20047, {{32066, 1, 0, 0},{32066, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20048, {{32067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20049, {{32068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20051, {{32069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20052, {{32070, 1, 0, 0},{32070, 1, 1, 18}, {0, 0, 0, 0}}},{20052, {{32070, 1, 0, 0},{32070, 1, 1, 18}, {0, 0, 0, 0}}},{20053, {{32070, 1, 0, 0},{32070, 1, 1, 19}, {0, 0, 0, 0}}},{20053, {{32070, 1, 0, 0},{32070, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{20054, {{32071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20055, {{32072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20056, {{32073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20057, {{32074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20059, {{32075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20061, {{32076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20063, {{32077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20064, {{32078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20065, {{32079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20066, {{32080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20068, {{32081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20069, {{32082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20070, {{32083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20071, {{32084, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20072, {{32085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20073, {{32086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20074, {{32087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20075, {{32088, 1, 0, 0},{32088, 1, 1, 16}, {0, 0, 0, 0}}},{20075, {{32088, 1, 0, 0},{32088, 1, 1, 16}, {0, 0, 0, 0}}},{20076, {{32088, 1, 0, 0},{32088, 1, 1, 17}, {0, 0, 0, 0}}},{20076, {{32088, 1, 0, 0},{32088, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{20077, {{32089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20078, {{32090, 1, 0, 0},{32090, 1, 1, 17}, {0, 0, 0, 0}}},{20078, {{32090, 1, 0, 0},{32090, 1, 1, 17}, {0, 0, 0, 0}}},{60044, {{32090, 1, 0, 0},{32090, 1, 1, 18}, {0, 0, 0, 0}}},{60044, {{32090, 1, 0, 0},{32090, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20079, {{32091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{20080, {{32092, 1, 0, 0},{32092, 1, 1, 18}, {0, 0, 0, 0}}},{20080, {{32092, 1, 0, 0},{32092, 1, 1, 18}, {0, 0, 0, 0}}},{20081, {{32092, 1, 0, 0},{32092, 1, 1, 20}, {0, 0, 0, 0}}},{20081, {{32092, 1, 0, 0},{32092, 1, 1, 20}, {0, 0, 0, 0}}},{20082, {{32092, 1, 0, 0},{32092, 1, 1, 19}, {0, 0, 0, 0}}},{20082, {{32092, 1, 0, 0},{32092, 1, 1, 19}, {0, 0, 0, 0}}},{20083, {{32092, 1, 0, 0},{32092, 1, 1, 21}, {0, 0, 0, 0}}},{20083, {{32092, 1, 0, 0},{32092, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {4, {{20084, {{32093, 1, 0, 0},{32093, 1, 1, 17}, {0, 0, 0, 0}}},{20084, {{32093, 1, 0, 0},{32093, 1, 1, 17}, {0, 0, 0, 0}}},{58304, {{32093, 1, 0, 0},{32093, 1, 1, 18}, {0, 0, 0, 0}}},{58304, {{32093, 1, 0, 0},{32093, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{20085, {{32094, 1, 0, 0},{32094, 1, 1, 18}, {0, 0, 0, 0}}},{20085, {{32094, 1, 0, 0},{32094, 1, 1, 18}, {0, 0, 0, 0}}},{20086, {{32094, 1, 0, 0},{32094, 1, 1, 19}, {0, 0, 0, 0}}},{20086, {{32094, 1, 0, 0},{32094, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{20087, {{32095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20088, {{32096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20089, {{32097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20090, {{32098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20091, {{32099, 1, 0, 0},{32099, 1, 1, 18}, {0, 0, 0, 0}}},{20091, {{32099, 1, 0, 0},{32099, 1, 1, 18}, {0, 0, 0, 0}}},{20092, {{32099, 1, 0, 0},{32099, 1, 1, 19}, {0, 0, 0, 0}}},{20092, {{32099, 1, 0, 0},{32099, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{20093, {{32100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20094, {{32101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20095, {{32102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20096, {{32103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20097, {{32104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20098, {{32105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20099, {{32106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20100, {{32107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20101, {{32108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20102, {{32109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20103, {{32110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20104, {{32111, 1, 0, 0},{32111, 1, 1, 16}, {0, 0, 0, 0}}},{20104, {{32111, 1, 0, 0},{32111, 1, 1, 16}, {0, 0, 0, 0}}},{20105, {{32111, 1, 0, 0},{32111, 1, 1, 17}, {0, 0, 0, 0}}},{20105, {{32111, 1, 0, 0},{32111, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{20106, {{32112, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20107, {{32113, 1, 0, 0},{32113, 1, 1, 18}, {0, 0, 0, 0}}},{20107, {{32113, 1, 0, 0},{32113, 1, 1, 18}, {0, 0, 0, 0}}},{20108, {{32113, 1, 0, 0},{32113, 1, 1, 17}, {0, 0, 0, 0}}},{20108, {{32113, 1, 0, 0},{32113, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{20109, {{32114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{20110, {{32115, 1, 0, 0},{32115, 1, 1, 18}, {0, 0, 0, 0}}},{20110, {{32115, 1, 0, 0},{32115, 1, 1, 18}, {0, 0, 0, 0}}},{20111, {{32115, 1, 0, 0},{32115, 1, 1, 19}, {0, 0, 0, 0}}},{20111, {{32115, 1, 0, 0},{32115, 1, 1, 19}, {0, 0, 0, 0}}},{20112, {{32115, 1, 0, 0},{32115, 1, 1, 20}, {0, 0, 0, 0}}},{20112, {{32115, 1, 0, 0},{32115, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{20113, {{32116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20114, {{32117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20115, {{32118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{20116, {{32120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20117, {{32121, 1, 0, 0},{32121, 1, 1, 18}, {0, 0, 0, 0}}},{20117, {{32121, 1, 0, 0},{32121, 1, 1, 18}, {0, 0, 0, 0}}},{20118, {{32121, 1, 0, 0},{32121, 1, 1, 17}, {0, 0, 0, 0}}},{20118, {{32121, 1, 0, 0},{32121, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{20120, {{32122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20121, {{32123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20122, {{32124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20123, {{32125, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20124, {{32126, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20125, {{32127, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20126, {{32128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20127, {{32129, 1, 0, 0},{32129, 1, 1, 17}, {0, 0, 0, 0}}},{20127, {{32129, 1, 0, 0},{32129, 1, 1, 17}, {0, 0, 0, 0}}},{20128, {{32129, 1, 0, 0},{32129, 1, 1, 18}, {0, 0, 0, 0}}},{20128, {{32129, 1, 0, 0},{32129, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20129, {{32130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20130, {{32131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20131, {{32132, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20132, {{32133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{20133, {{32134, 1, 0, 0},{32134, 1, 1, 17}, {0, 0, 0, 0}}},{20133, {{32134, 1, 0, 0},{32134, 1, 1, 17}, {0, 0, 0, 0}}},{20134, {{32134, 1, 0, 0},{32134, 1, 1, 19}, {0, 0, 0, 0}}},{20134, {{32134, 1, 0, 0},{32134, 1, 1, 19}, {0, 0, 0, 0}}},{20135, {{32134, 1, 0, 0},{32134, 1, 1, 18}, {0, 0, 0, 0}}},{20135, {{32134, 1, 0, 0},{32134, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20136, {{32135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20137, {{32136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20138, {{32137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20139, {{32138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20140, {{32139, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20141, {{32140, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20142, {{32141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20143, {{32142, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20144, {{32143, 1, 0, 0},{32143, 1, 1, 17}, {0, 0, 0, 0}}},{20144, {{32143, 1, 0, 0},{32143, 1, 1, 17}, {0, 0, 0, 0}}},{20145, {{32143, 1, 0, 0},{32143, 1, 1, 18}, {0, 0, 0, 0}}},{20145, {{32143, 1, 0, 0},{32143, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20146, {{32144, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20147, {{32145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20148, {{32146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20150, {{32147, 1, 0, 0},{32147, 1, 1, 17}, {0, 0, 0, 0}}},{20150, {{32147, 1, 0, 0},{32147, 1, 1, 17}, {0, 0, 0, 0}}},{20151, {{32147, 1, 0, 0},{32147, 1, 1, 18}, {0, 0, 0, 0}}},{20151, {{32147, 1, 0, 0},{32147, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20153, {{32148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20154, {{32149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20155, {{32150, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20156, {{32151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20159, {{32152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20160, {{32153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20161, {{32154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{20162, {{32155, 1, 0, 0},{32155, 1, 1, 20}, {0, 0, 0, 0}}},{20162, {{32155, 1, 0, 0},{32155, 1, 1, 20}, {0, 0, 0, 0}}},{20163, {{32155, 1, 0, 0},{32155, 1, 1, 18}, {0, 0, 0, 0}}},{20163, {{32155, 1, 0, 0},{32155, 1, 1, 18}, {0, 0, 0, 0}}},{20164, {{32155, 1, 0, 0},{32155, 1, 1, 19}, {0, 0, 0, 0}}},{20164, {{32155, 1, 0, 0},{32155, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{20165, {{32156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20166, {{32157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20167, {{32158, 1, 0, 0},{32158, 1, 1, 17}, {0, 0, 0, 0}}},{20167, {{32158, 1, 0, 0},{32158, 1, 1, 17}, {0, 0, 0, 0}}},{60054, {{32158, 1, 0, 0},{32158, 1, 1, 18}, {0, 0, 0, 0}}},{60054, {{32158, 1, 0, 0},{32158, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{20168, {{32159, 1, 0, 0},{32159, 1, 1, 18}, {0, 0, 0, 0}}},{20168, {{32159, 1, 0, 0},{32159, 1, 1, 18}, {0, 0, 0, 0}}},{20169, {{32159, 1, 0, 0},{32159, 1, 1, 19}, {0, 0, 0, 0}}},{20169, {{32159, 1, 0, 0},{32159, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{20170, {{32160, 1, 0, 0},{32160, 1, 1, 17}, {0, 0, 0, 0}}},{20170, {{32160, 1, 0, 0},{32160, 1, 1, 17}, {0, 0, 0, 0}}},{20171, {{32160, 1, 0, 0},{32160, 1, 1, 18}, {0, 0, 0, 0}}},{20171, {{32160, 1, 0, 0},{32160, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20172, {{32161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20173, {{32162, 1, 0, 0},{32162, 1, 1, 17}, {0, 0, 0, 0}}},{20173, {{32162, 1, 0, 0},{32162, 1, 1, 17}, {0, 0, 0, 0}}},{20174, {{32162, 1, 0, 0},{32162, 1, 1, 18}, {0, 0, 0, 0}}},{20174, {{32162, 1, 0, 0},{32162, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{20175, {{32163, 1, 0, 0},{32163, 1, 1, 17}, {0, 0, 0, 0}}},{20175, {{32163, 1, 0, 0},{32163, 1, 1, 17}, {0, 0, 0, 0}}},{20176, {{32163, 1, 0, 0},{32163, 1, 1, 18}, {0, 0, 0, 0}}},{20176, {{32163, 1, 0, 0},{32163, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20177, {{32164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20178, {{32165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20179, {{32166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20180, {{32167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20181, {{32168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20182, {{32169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20183, {{32170, 1, 0, 0},{32170, 1, 1, 17}, {0, 0, 0, 0}}},{20183, {{32170, 1, 0, 0},{32170, 1, 1, 17}, {0, 0, 0, 0}}},{20184, {{32170, 1, 0, 0},{32170, 1, 1, 18}, {0, 0, 0, 0}}},{20184, {{32170, 1, 0, 0},{32170, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20185, {{32171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20186, {{32172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20187, {{32173, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20189, {{32174, 1, 0, 0},{32174, 1, 1, 19}, {0, 0, 0, 0}}},{20189, {{32174, 1, 0, 0},{32174, 1, 1, 19}, {0, 0, 0, 0}}},{20190, {{32174, 1, 0, 0},{32174, 1, 1, 18}, {0, 0, 0, 0}}},{20190, {{32174, 1, 0, 0},{32174, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{20191, {{32175, 1, 0, 0},{32175, 1, 1, 17}, {0, 0, 0, 0}}},{20191, {{32175, 1, 0, 0},{32175, 1, 1, 17}, {0, 0, 0, 0}}},{58312, {{32175, 1, 0, 0},{32175, 1, 1, 18}, {0, 0, 0, 0}}},{58312, {{32175, 1, 0, 0},{32175, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20192, {{32176, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20193, {{32177, 1, 0, 0},{32177, 1, 1, 17}, {0, 0, 0, 0}}},{20193, {{32177, 1, 0, 0},{32177, 1, 1, 17}, {0, 0, 0, 0}}},{58319, {{32177, 1, 0, 0},{32177, 1, 1, 18}, {0, 0, 0, 0}}},{58319, {{32177, 1, 0, 0},{32177, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {8, {{20194, {{32178, 1, 0, 0},{32178, 1, 1, 20}, {0, 0, 0, 0}}},{20194, {{32178, 1, 0, 0},{32178, 1, 1, 20}, {0, 0, 0, 0}}},{20195, {{32178, 1, 0, 0},{32178, 1, 1, 19}, {0, 0, 0, 0}}},{20195, {{32178, 1, 0, 0},{32178, 1, 1, 19}, {0, 0, 0, 0}}},{58313, {{32178, 1, 0, 0},{32178, 1, 1, 22}, {0, 0, 0, 0}}},{58313, {{32178, 1, 0, 0},{32178, 1, 1, 22}, {0, 0, 0, 0}}},{60050, {{32178, 1, 0, 0},{32178, 1, 1, 21}, {0, 0, 0, 0}}},{60050, {{32178, 1, 0, 0},{32178, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{20196, {{32179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20198, {{32180, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20199, {{32181, 1, 0, 0},{32181, 1, 1, 17}, {0, 0, 0, 0}}},{20199, {{32181, 1, 0, 0},{32181, 1, 1, 17}, {0, 0, 0, 0}}},{20200, {{32181, 1, 0, 0},{32181, 1, 1, 18}, {0, 0, 0, 0}}},{20200, {{32181, 1, 0, 0},{32181, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20201, {{32182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20202, {{32183, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20203, {{32184, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20204, {{32185, 1, 0, 0},{32185, 1, 1, 17}, {0, 0, 0, 0}}},{20204, {{32185, 1, 0, 0},{32185, 1, 1, 17}, {0, 0, 0, 0}}},{20205, {{32185, 1, 0, 0},{32185, 1, 1, 18}, {0, 0, 0, 0}}},{20205, {{32185, 1, 0, 0},{32185, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20206, {{32186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20207, {{32187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20208, {{32188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20210, {{32189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20211, {{32190, 1, 0, 0},{32190, 1, 1, 18}, {0, 0, 0, 0}}},{20211, {{32190, 1, 0, 0},{32190, 1, 1, 18}, {0, 0, 0, 0}}},{20212, {{32190, 1, 0, 0},{32190, 1, 1, 17}, {0, 0, 0, 0}}},{20212, {{32190, 1, 0, 0},{32190, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{20214, {{32191, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20216, {{32192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20217, {{32193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20218, {{32194, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20219, {{32195, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20220, {{32196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20221, {{32197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20222, {{32198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20223, {{32199, 1, 0, 0},{32199, 1, 1, 17}, {0, 0, 0, 0}}},{20223, {{32199, 1, 0, 0},{32199, 1, 1, 17}, {0, 0, 0, 0}}},{20224, {{32199, 1, 0, 0},{32199, 1, 1, 18}, {0, 0, 0, 0}}},{20224, {{32199, 1, 0, 0},{32199, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20225, {{32200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20226, {{32201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20227, {{32202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20228, {{32203, 1, 0, 0},{32203, 1, 1, 18}, {0, 0, 0, 0}}},{20228, {{32203, 1, 0, 0},{32203, 1, 1, 18}, {0, 0, 0, 0}}},{20229, {{32203, 1, 0, 0},{32203, 1, 1, 19}, {0, 0, 0, 0}}},{20229, {{32203, 1, 0, 0},{32203, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{20230, {{32204, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20231, {{32205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20232, {{32206, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20233, {{32207, 1, 0, 0},{32207, 1, 1, 18}, {0, 0, 0, 0}}},{20233, {{32207, 1, 0, 0},{32207, 1, 1, 18}, {0, 0, 0, 0}}},{20234, {{32207, 1, 0, 0},{32207, 1, 1, 19}, {0, 0, 0, 0}}},{20234, {{32207, 1, 0, 0},{32207, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{20235, {{32209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20237, {{32210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{20238, {{32213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20239, {{32214, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20240, {{32215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20241, {{32216, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20242, {{32217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20243, {{32218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20244, {{32219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20245, {{32220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20246, {{32221, 1, 0, 0},{32221, 1, 1, 18}, {0, 0, 0, 0}}},{20246, {{32221, 1, 0, 0},{32221, 1, 1, 18}, {0, 0, 0, 0}}},{20247, {{32221, 1, 0, 0},{32221, 1, 1, 19}, {0, 0, 0, 0}}},{20247, {{32221, 1, 0, 0},{32221, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{20248, {{32222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20249, {{32223, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20250, {{32224, 1, 0, 0},{32224, 1, 1, 17}, {0, 0, 0, 0}}},{20250, {{32224, 1, 0, 0},{32224, 1, 1, 17}, {0, 0, 0, 0}}},{20251, {{32224, 1, 0, 0},{32224, 1, 1, 18}, {0, 0, 0, 0}}},{20251, {{32224, 1, 0, 0},{32224, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20252, {{32225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20253, {{32226, 1, 0, 0},{32226, 1, 1, 17}, {0, 0, 0, 0}}},{20253, {{32226, 1, 0, 0},{32226, 1, 1, 17}, {0, 0, 0, 0}}},{20254, {{32226, 1, 0, 0},{32226, 1, 1, 18}, {0, 0, 0, 0}}},{20254, {{32226, 1, 0, 0},{32226, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{20255, {{32227, 1, 0, 0},{32227, 1, 1, 17}, {0, 0, 0, 0}}},{20255, {{32227, 1, 0, 0},{32227, 1, 1, 17}, {0, 0, 0, 0}}},{20256, {{32227, 1, 0, 0},{32227, 1, 1, 18}, {0, 0, 0, 0}}},{20256, {{32227, 1, 0, 0},{32227, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20257, {{32228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20258, {{32229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20259, {{32230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20260, {{32231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20261, {{32232, 1, 0, 0},{32232, 1, 1, 19}, {0, 0, 0, 0}}},{20261, {{32232, 1, 0, 0},{32232, 1, 1, 19}, {0, 0, 0, 0}}},{20262, {{32232, 1, 0, 0},{32232, 1, 1, 18}, {0, 0, 0, 0}}},{20262, {{32232, 1, 0, 0},{32232, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{20263, {{32233, 1, 0, 0},{32233, 1, 1, 19}, {0, 0, 0, 0}}},{20263, {{32233, 1, 0, 0},{32233, 1, 1, 19}, {0, 0, 0, 0}}},{20264, {{32233, 1, 0, 0},{32233, 1, 1, 18}, {0, 0, 0, 0}}},{20264, {{32233, 1, 0, 0},{32233, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20265, {{32234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20266, {{32235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20267, {{32236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20268, {{32237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{20269, {{32239, 1, 0, 0},{32239, 1, 1, 19}, {0, 0, 0, 0}}},{20269, {{32239, 1, 0, 0},{32239, 1, 1, 19}, {0, 0, 0, 0}}},{20270, {{32239, 1, 0, 0},{32239, 1, 1, 18}, {0, 0, 0, 0}}},{20270, {{32239, 1, 0, 0},{32239, 1, 1, 18}, {0, 0, 0, 0}}},{20271, {{32239, 1, 0, 0},{32239, 1, 1, 20}, {0, 0, 0, 0}}},{20271, {{32239, 1, 0, 0},{32239, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{20272, {{32240, 1, 0, 0},{32240, 1, 1, 16}, {0, 0, 0, 0}}},{20272, {{32240, 1, 0, 0},{32240, 1, 1, 16}, {0, 0, 0, 0}}},{20273, {{32240, 1, 0, 0},{32240, 1, 1, 17}, {0, 0, 0, 0}}},{20273, {{32240, 1, 0, 0},{32240, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{20274, {{32241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20275, {{32242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20276, {{32243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{20277, {{32244, 1, 0, 0},{32244, 1, 1, 18}, {0, 0, 0, 0}}},{20277, {{32244, 1, 0, 0},{32244, 1, 1, 18}, {0, 0, 0, 0}}},{30286, {{32244, 1, 0, 0},{32244, 1, 1, 1},{32244, 1, 1, 19},{64087, 1, 0, 0}, {0, 0, 0, 0}}},{30286, {{32244, 1, 0, 0},{32244, 1, 1, 1},{32244, 1, 1, 19},{64087, 1, 0, 0}, {0, 0, 0, 0}}},{30286, {{32244, 1, 0, 0},{32244, 1, 1, 1},{32244, 1, 1, 19},{64087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20278, {{32245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20279, {{32246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20280, {{32247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20281, {{32248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20282, {{32249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20283, {{32250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20284, {{32251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20285, {{32252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20286, {{32253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{20287, {{32255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20288, {{32256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20289, {{32257, 1, 0, 0},{32257, 1, 1, 17}, {0, 0, 0, 0}}},{20289, {{32257, 1, 0, 0},{32257, 1, 1, 17}, {0, 0, 0, 0}}},{20290, {{32257, 1, 0, 0},{32257, 1, 1, 18}, {0, 0, 0, 0}}},{20290, {{32257, 1, 0, 0},{32257, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{20291, {{32258, 1, 0, 0},{32258, 1, 1, 16}, {0, 0, 0, 0}}},{20291, {{32258, 1, 0, 0},{32258, 1, 1, 16}, {0, 0, 0, 0}}},{20292, {{32258, 1, 0, 0},{32258, 1, 1, 17}, {0, 0, 0, 0}}},{20292, {{32258, 1, 0, 0},{32258, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{20293, {{32259, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20294, {{32260, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20295, {{32261, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{20296, {{32264, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{20297, {{32265, 1, 0, 0},{32265, 1, 1, 18}, {0, 0, 0, 0}}},{20297, {{32265, 1, 0, 0},{32265, 1, 1, 18}, {0, 0, 0, 0}}},{30287, {{32265, 1, 0, 0},{32265, 1, 1, 0},{32265, 1, 1, 19},{64088, 1, 0, 0}, {0, 0, 0, 0}}},{30287, {{32265, 1, 0, 0},{32265, 1, 1, 0},{32265, 1, 1, 19},{64088, 1, 0, 0}, {0, 0, 0, 0}}},{30287, {{32265, 1, 0, 0},{32265, 1, 1, 0},{32265, 1, 1, 19},{64088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{20298, {{32266, 1, 0, 0},{32266, 1, 1, 17}, {0, 0, 0, 0}}},{20298, {{32266, 1, 0, 0},{32266, 1, 1, 17}, {0, 0, 0, 0}}},{20299, {{32266, 1, 0, 0},{32266, 1, 1, 18}, {0, 0, 0, 0}}},{20299, {{32266, 1, 0, 0},{32266, 1, 1, 18}, {0, 0, 0, 0}}},{20300, {{32266, 1, 0, 0},{32266, 1, 1, 19}, {0, 0, 0, 0}}},{20300, {{32266, 1, 0, 0},{32266, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{20301, {{32267, 1, 0, 0},{32267, 1, 1, 18}, {0, 0, 0, 0}}},{20301, {{32267, 1, 0, 0},{32267, 1, 1, 18}, {0, 0, 0, 0}}},{20302, {{32267, 1, 0, 0},{32267, 1, 1, 17}, {0, 0, 0, 0}}},{20302, {{32267, 1, 0, 0},{32267, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{20303, {{32268, 1, 0, 0},{32268, 1, 1, 16}, {0, 0, 0, 0}}},{20303, {{32268, 1, 0, 0},{32268, 1, 1, 16}, {0, 0, 0, 0}}},{20304, {{32268, 1, 0, 0},{32268, 1, 1, 17}, {0, 0, 0, 0}}},{20304, {{32268, 1, 0, 0},{32268, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{20305, {{32269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20306, {{32270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20307, {{32271, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20308, {{32272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20309, {{32273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20310, {{32274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20311, {{32275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20312, {{32276, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20313, {{32277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20314, {{32278, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20315, {{32279, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20316, {{32280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20317, {{32281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20318, {{32282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20319, {{32283, 1, 0, 0},{32283, 1, 1, 19}, {0, 0, 0, 0}}},{20319, {{32283, 1, 0, 0},{32283, 1, 1, 19}, {0, 0, 0, 0}}},{20320, {{32283, 1, 0, 0},{32283, 1, 1, 18}, {0, 0, 0, 0}}},{20320, {{32283, 1, 0, 0},{32283, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20321, {{32284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20322, {{32285, 1, 0, 0},{32285, 1, 1, 17}, {0, 0, 0, 0}}},{20322, {{32285, 1, 0, 0},{32285, 1, 1, 17}, {0, 0, 0, 0}}},{20323, {{32285, 1, 0, 0},{32285, 1, 1, 18}, {0, 0, 0, 0}}},{20323, {{32285, 1, 0, 0},{32285, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20324, {{32286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20326, {{32287, 1, 0, 0},{32287, 1, 1, 17}, {0, 0, 0, 0}}},{20326, {{32287, 1, 0, 0},{32287, 1, 1, 17}, {0, 0, 0, 0}}},{20327, {{32287, 1, 0, 0},{32287, 1, 1, 18}, {0, 0, 0, 0}}},{20327, {{32287, 1, 0, 0},{32287, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20328, {{32288, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20329, {{32289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{20330, {{32290, 1, 0, 0},{32290, 1, 1, 20}, {0, 0, 0, 0}}},{20330, {{32290, 1, 0, 0},{32290, 1, 1, 20}, {0, 0, 0, 0}}},{20331, {{32290, 1, 0, 0},{32290, 1, 1, 19}, {0, 0, 0, 0}}},{20331, {{32290, 1, 0, 0},{32290, 1, 1, 19}, {0, 0, 0, 0}}},{20332, {{32290, 1, 0, 0},{32290, 1, 1, 21}, {0, 0, 0, 0}}},{20332, {{32290, 1, 0, 0},{32290, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{20333, {{32291, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20335, {{32292, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20336, {{32293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20337, {{32294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20338, {{32295, 1, 0, 0},{32295, 1, 1, 18}, {0, 0, 0, 0}}},{20338, {{32295, 1, 0, 0},{32295, 1, 1, 18}, {0, 0, 0, 0}}},{20339, {{32295, 1, 0, 0},{32295, 1, 1, 17}, {0, 0, 0, 0}}},{20339, {{32295, 1, 0, 0},{32295, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{20340, {{32296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20342, {{32297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20343, {{32298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20344, {{32299, 1, 0, 0},{32299, 1, 1, 18}, {0, 0, 0, 0}}},{20344, {{32299, 1, 0, 0},{32299, 1, 1, 18}, {0, 0, 0, 0}}},{20345, {{32299, 1, 0, 0},{32299, 1, 1, 19}, {0, 0, 0, 0}}},{20345, {{32299, 1, 0, 0},{32299, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{20346, {{32300, 1, 0, 0},{32300, 1, 1, 17}, {0, 0, 0, 0}}},{20346, {{32300, 1, 0, 0},{32300, 1, 1, 17}, {0, 0, 0, 0}}},{20347, {{32300, 1, 0, 0},{32300, 1, 1, 18}, {0, 0, 0, 0}}},{20347, {{32300, 1, 0, 0},{32300, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{20348, {{32301, 1, 0, 0},{32301, 1, 1, 17}, {0, 0, 0, 0}}},{20348, {{32301, 1, 0, 0},{32301, 1, 1, 17}, {0, 0, 0, 0}}},{20349, {{32301, 1, 0, 0},{32301, 1, 1, 18}, {0, 0, 0, 0}}},{20349, {{32301, 1, 0, 0},{32301, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20350, {{32302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20351, {{32303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20352, {{32304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20353, {{32305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20354, {{32306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20355, {{32307, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20356, {{32308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{20357, {{32309, 1, 0, 0},{32309, 1, 1, 20}, {0, 0, 0, 0}}},{20357, {{32309, 1, 0, 0},{32309, 1, 1, 20}, {0, 0, 0, 0}}},{20358, {{32309, 1, 0, 0},{32309, 1, 1, 19}, {0, 0, 0, 0}}},{20358, {{32309, 1, 0, 0},{32309, 1, 1, 19}, {0, 0, 0, 0}}},{20359, {{32309, 1, 0, 0},{32309, 1, 1, 21}, {0, 0, 0, 0}}},{20359, {{32309, 1, 0, 0},{32309, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {4, {{20360, {{32310, 1, 0, 0},{32310, 1, 1, 17}, {0, 0, 0, 0}}},{20360, {{32310, 1, 0, 0},{32310, 1, 1, 17}, {0, 0, 0, 0}}},{20361, {{32310, 1, 0, 0},{32310, 1, 1, 18}, {0, 0, 0, 0}}},{20361, {{32310, 1, 0, 0},{32310, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{20362, {{32311, 1, 0, 0},{32311, 1, 1, 17}, {0, 0, 0, 0}}},{20362, {{32311, 1, 0, 0},{32311, 1, 1, 17}, {0, 0, 0, 0}}},{58326, {{32311, 1, 0, 0},{32311, 1, 1, 18}, {0, 0, 0, 0}}},{58326, {{32311, 1, 0, 0},{32311, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20363, {{32312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20364, {{32313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20365, {{32314, 1, 0, 0},{32314, 1, 1, 18}, {0, 0, 0, 0}}},{20365, {{32314, 1, 0, 0},{32314, 1, 1, 18}, {0, 0, 0, 0}}},{20366, {{32314, 1, 0, 0},{32314, 1, 1, 17}, {0, 0, 0, 0}}},{20366, {{32314, 1, 0, 0},{32314, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{20367, {{32315, 1, 0, 0},{32315, 1, 1, 17}, {0, 0, 0, 0}}},{20367, {{32315, 1, 0, 0},{32315, 1, 1, 17}, {0, 0, 0, 0}}},{20368, {{32315, 1, 0, 0},{32315, 1, 1, 18}, {0, 0, 0, 0}}},{20368, {{32315, 1, 0, 0},{32315, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20369, {{32316, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20370, {{32317, 1, 0, 0},{32317, 1, 1, 17}, {0, 0, 0, 0}}},{20370, {{32317, 1, 0, 0},{32317, 1, 1, 17}, {0, 0, 0, 0}}},{20371, {{32317, 1, 0, 0},{32317, 1, 1, 18}, {0, 0, 0, 0}}},{20371, {{32317, 1, 0, 0},{32317, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20373, {{32318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20374, {{32319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20375, {{32320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{20376, {{32321, 1, 0, 0},{32321, 1, 1, 18}, {0, 0, 0, 0}}},{20376, {{32321, 1, 0, 0},{32321, 1, 1, 18}, {0, 0, 0, 0}}},{30288, {{32321, 1, 0, 0},{32321, 1, 1, 0},{32321, 1, 1, 19},{64089, 1, 0, 0}, {0, 0, 0, 0}}},{30288, {{32321, 1, 0, 0},{32321, 1, 1, 0},{32321, 1, 1, 19},{64089, 1, 0, 0}, {0, 0, 0, 0}}},{30288, {{32321, 1, 0, 0},{32321, 1, 1, 0},{32321, 1, 1, 19},{64089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20377, {{32322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20379, {{32323, 1, 0, 0},{32323, 1, 1, 19}, {0, 0, 0, 0}}},{20379, {{32323, 1, 0, 0},{32323, 1, 1, 19}, {0, 0, 0, 0}}},{20380, {{32323, 1, 0, 0},{32323, 1, 1, 18}, {0, 0, 0, 0}}},{20380, {{32323, 1, 0, 0},{32323, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{20381, {{32324, 1, 0, 0},{32324, 1, 1, 17}, {0, 0, 0, 0}}},{20381, {{32324, 1, 0, 0},{32324, 1, 1, 17}, {0, 0, 0, 0}}},{20382, {{32324, 1, 0, 0},{32324, 1, 1, 18}, {0, 0, 0, 0}}},{20382, {{32324, 1, 0, 0},{32324, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{20383, {{32325, 1, 0, 0},{32325, 1, 1, 17}, {0, 0, 0, 0}}},{20383, {{32325, 1, 0, 0},{32325, 1, 1, 17}, {0, 0, 0, 0}}},{20384, {{32325, 1, 0, 0},{32325, 1, 1, 18}, {0, 0, 0, 0}}},{20384, {{32325, 1, 0, 0},{32325, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{20385, {{32326, 1, 0, 0},{32326, 1, 1, 17}, {0, 0, 0, 0}}},{20385, {{32326, 1, 0, 0},{32326, 1, 1, 17}, {0, 0, 0, 0}}},{20386, {{32326, 1, 0, 0},{32326, 1, 1, 18}, {0, 0, 0, 0}}},{20386, {{32326, 1, 0, 0},{32326, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20387, {{32327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20388, {{32328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20389, {{32329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20390, {{32330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20392, {{32331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{20393, {{32333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20394, {{32334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20395, {{32335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20396, {{32336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20397, {{32337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20398, {{32338, 1, 0, 0},{32338, 1, 1, 17}, {0, 0, 0, 0}}},{20398, {{32338, 1, 0, 0},{32338, 1, 1, 17}, {0, 0, 0, 0}}},{20399, {{32338, 1, 0, 0},{32338, 1, 1, 18}, {0, 0, 0, 0}}},{20399, {{32338, 1, 0, 0},{32338, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20400, {{32339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20401, {{32340, 1, 0, 0},{32340, 1, 1, 17}, {0, 0, 0, 0}}},{20401, {{32340, 1, 0, 0},{32340, 1, 1, 17}, {0, 0, 0, 0}}},{20402, {{32340, 1, 0, 0},{32340, 1, 1, 18}, {0, 0, 0, 0}}},{20402, {{32340, 1, 0, 0},{32340, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20403, {{32341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20404, {{32342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20405, {{32343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20406, {{32344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20407, {{32345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20408, {{32346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20409, {{32347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20410, {{32348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20411, {{32349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20412, {{32350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20413, {{32351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20414, {{32352, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20415, {{32353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20416, {{32354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20417, {{32355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20418, {{32356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20419, {{32357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20421, {{32358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20422, {{32359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20423, {{32360, 1, 0, 0},{32360, 1, 1, 16}, {0, 0, 0, 0}}},{20423, {{32360, 1, 0, 0},{32360, 1, 1, 16}, {0, 0, 0, 0}}},{20424, {{32360, 1, 0, 0},{32360, 1, 1, 17}, {0, 0, 0, 0}}},{20424, {{32360, 1, 0, 0},{32360, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{20425, {{32361, 1, 0, 0},{32361, 1, 1, 18}, {0, 0, 0, 0}}},{20425, {{32361, 1, 0, 0},{32361, 1, 1, 18}, {0, 0, 0, 0}}},{20426, {{32361, 1, 0, 0},{32361, 1, 1, 17}, {0, 0, 0, 0}}},{20426, {{32361, 1, 0, 0},{32361, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{20427, {{32362, 1, 0, 0},{32362, 1, 1, 17}, {0, 0, 0, 0}}},{20427, {{32362, 1, 0, 0},{32362, 1, 1, 17}, {0, 0, 0, 0}}},{20428, {{32362, 1, 0, 0},{32362, 1, 1, 18}, {0, 0, 0, 0}}},{20428, {{32362, 1, 0, 0},{32362, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{20429, {{32363, 1, 0, 0},{32363, 1, 1, 19}, {0, 0, 0, 0}}},{20429, {{32363, 1, 0, 0},{32363, 1, 1, 19}, {0, 0, 0, 0}}},{58327, {{32363, 1, 0, 0},{32363, 1, 1, 18}, {0, 0, 0, 0}}},{58327, {{32363, 1, 0, 0},{32363, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20430, {{32364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{20431, {{32365, 1, 0, 0},{32365, 1, 1, 18}, {0, 0, 0, 0}}},{20431, {{32365, 1, 0, 0},{32365, 1, 1, 18}, {0, 0, 0, 0}}},{20432, {{32365, 1, 0, 0},{32365, 1, 1, 19}, {0, 0, 0, 0}}},{20432, {{32365, 1, 0, 0},{32365, 1, 1, 19}, {0, 0, 0, 0}}},{20433, {{32365, 1, 0, 0},{32365, 1, 1, 20}, {0, 0, 0, 0}}},{20433, {{32365, 1, 0, 0},{32365, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{20434, {{32366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20435, {{32367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20436, {{32368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20437, {{32369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20438, {{32370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20439, {{32371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20440, {{32372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20441, {{32373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20442, {{32374, 1, 0, 0},{32374, 1, 1, 16}, {0, 0, 0, 0}}},{20442, {{32374, 1, 0, 0},{32374, 1, 1, 16}, {0, 0, 0, 0}}},{20443, {{32374, 1, 0, 0},{32374, 1, 1, 17}, {0, 0, 0, 0}}},{20443, {{32374, 1, 0, 0},{32374, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{20444, {{32375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20445, {{32376, 1, 0, 0},{32376, 1, 1, 18}, {0, 0, 0, 0}}},{20445, {{32376, 1, 0, 0},{32376, 1, 1, 18}, {0, 0, 0, 0}}},{20446, {{32376, 1, 0, 0},{32376, 1, 1, 17}, {0, 0, 0, 0}}},{20446, {{32376, 1, 0, 0},{32376, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{20447, {{32377, 1, 0, 0},{32377, 1, 1, 17}, {0, 0, 0, 0}}},{20447, {{32377, 1, 0, 0},{32377, 1, 1, 17}, {0, 0, 0, 0}}},{20448, {{32377, 1, 0, 0},{32377, 1, 1, 18}, {0, 0, 0, 0}}},{20448, {{32377, 1, 0, 0},{32377, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20449, {{32378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20450, {{32379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20451, {{32380, 1, 0, 0},{32380, 1, 1, 17}, {0, 0, 0, 0}}},{20451, {{32380, 1, 0, 0},{32380, 1, 1, 17}, {0, 0, 0, 0}}},{20452, {{32380, 1, 0, 0},{32380, 1, 1, 18}, {0, 0, 0, 0}}},{20452, {{32380, 1, 0, 0},{32380, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{20453, {{32381, 1, 0, 0},{32381, 1, 1, 17}, {0, 0, 0, 0}}},{20453, {{32381, 1, 0, 0},{32381, 1, 1, 17}, {0, 0, 0, 0}}},{20454, {{32381, 1, 0, 0},{32381, 1, 1, 19}, {0, 0, 0, 0}}},{20454, {{32381, 1, 0, 0},{32381, 1, 1, 19}, {0, 0, 0, 0}}},{20455, {{32381, 1, 0, 0},{32381, 1, 1, 18}, {0, 0, 0, 0}}},{20455, {{32381, 1, 0, 0},{32381, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{20456, {{32382, 1, 0, 0},{32382, 1, 1, 18}, {0, 0, 0, 0}}},{20456, {{32382, 1, 0, 0},{32382, 1, 1, 18}, {0, 0, 0, 0}}},{20457, {{32382, 1, 0, 0},{32382, 1, 1, 17}, {0, 0, 0, 0}}},{20457, {{32382, 1, 0, 0},{32382, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{20458, {{32383, 1, 0, 0},{32383, 1, 1, 17}, {0, 0, 0, 0}}},{20458, {{32383, 1, 0, 0},{32383, 1, 1, 17}, {0, 0, 0, 0}}},{60057, {{32383, 1, 0, 0},{32383, 1, 1, 18}, {0, 0, 0, 0}}},{60057, {{32383, 1, 0, 0},{32383, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20459, {{32384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20460, {{32385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20461, {{32386, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20462, {{32387, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{20463, {{32389, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20464, {{32390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20466, {{32391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20467, {{32392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20468, {{32393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20469, {{32394, 1, 0, 0},{32394, 1, 1, 17}, {0, 0, 0, 0}}},{20469, {{32394, 1, 0, 0},{32394, 1, 1, 17}, {0, 0, 0, 0}}},{20470, {{32394, 1, 0, 0},{32394, 1, 1, 18}, {0, 0, 0, 0}}},{20470, {{32394, 1, 0, 0},{32394, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20471, {{32395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20472, {{32396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20473, {{32397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20474, {{32398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20475, {{32399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20476, {{32400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20477, {{32401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{20478, {{32402, 1, 0, 0},{32402, 1, 1, 17}, {0, 0, 0, 0}}},{20478, {{32402, 1, 0, 0},{32402, 1, 1, 17}, {0, 0, 0, 0}}},{20479, {{32402, 1, 0, 0},{32402, 1, 1, 19}, {0, 0, 0, 0}}},{20479, {{32402, 1, 0, 0},{32402, 1, 1, 19}, {0, 0, 0, 0}}},{20480, {{32402, 1, 0, 0},{32402, 1, 1, 18}, {0, 0, 0, 0}}},{20480, {{32402, 1, 0, 0},{32402, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20481, {{32403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20482, {{32404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20483, {{32405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20484, {{32406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20485, {{32407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20486, {{32408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20487, {{32409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20488, {{32410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20489, {{32411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20490, {{32412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20491, {{32413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20492, {{32414, 1, 0, 0}, {0, 0, 0, 0}}}}}
};

static const MJMappingSet chunked_mj_mappings_007f36_008b9f[3178] = {
    {1, {{20498, {{32566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20499, {{32567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20500, {{32568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20501, {{32569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20502, {{32570, 1, 0, 0},{32570, 1, 1, 17}, {0, 0, 0, 0}}},{20502, {{32570, 1, 0, 0},{32570, 1, 1, 17}, {0, 0, 0, 0}}},{58335, {{32570, 1, 0, 0},{32570, 1, 1, 18}, {0, 0, 0, 0}}},{58335, {{32570, 1, 0, 0},{32570, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20503, {{32571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20504, {{32572, 1, 0, 0},{32572, 1, 1, 17}, {0, 0, 0, 0}}},{20504, {{32572, 1, 0, 0},{32572, 1, 1, 17}, {0, 0, 0, 0}}},{20505, {{32572, 1, 0, 0},{32572, 1, 1, 18}, {0, 0, 0, 0}}},{20505, {{32572, 1, 0, 0},{32572, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20506, {{32573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20507, {{32574, 1, 0, 0},{32574, 1, 1, 18}, {0, 0, 0, 0}}},{20507, {{32574, 1, 0, 0},{32574, 1, 1, 18}, {0, 0, 0, 0}}},{20508, {{32574, 1, 0, 0},{32574, 1, 1, 19}, {0, 0, 0, 0}}},{20508, {{32574, 1, 0, 0},{32574, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{20509, {{32575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20510, {{32576, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20511, {{32577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{20512, {{32579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20513, {{32580, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20514, {{32581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20515, {{32582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20516, {{32583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20517, {{32584, 1, 0, 0},{32584, 1, 1, 16}, {0, 0, 0, 0}}},{20517, {{32584, 1, 0, 0},{32584, 1, 1, 16}, {0, 0, 0, 0}}},{20518, {{32584, 1, 0, 0},{32584, 1, 1, 17}, {0, 0, 0, 0}}},{20518, {{32584, 1, 0, 0},{32584, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{20519, {{32586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20520, {{32587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20521, {{32588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20522, {{32589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20523, {{32590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20524, {{32591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20525, {{32592, 1, 0, 0},{32592, 1, 1, 18}, {0, 0, 0, 0}}},{20525, {{32592, 1, 0, 0},{32592, 1, 1, 18}, {0, 0, 0, 0}}},{20526, {{32592, 1, 0, 0},{32592, 1, 1, 19}, {0, 0, 0, 0}}},{20526, {{32592, 1, 0, 0},{32592, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{20527, {{32593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20528, {{32594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20529, {{32595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{20530, {{32596, 1, 0, 0},{32596, 1, 1, 17}, {0, 0, 0, 0}}},{20530, {{32596, 1, 0, 0},{32596, 1, 1, 17}, {0, 0, 0, 0}}},{20531, {{32596, 1, 0, 0},{32596, 1, 1, 18}, {0, 0, 0, 0}}},{20531, {{32596, 1, 0, 0},{32596, 1, 1, 18}, {0, 0, 0, 0}}},{20532, {{32596, 1, 0, 0},{32596, 1, 1, 19}, {0, 0, 0, 0}}},{20532, {{32596, 1, 0, 0},{32596, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{20533, {{32597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{20534, {{32599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20535, {{32600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20536, {{32601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{20539, {{32603, 1, 0, 0},{32603, 1, 1, 18}, {0, 0, 0, 0}}},{20539, {{32603, 1, 0, 0},{32603, 1, 1, 18}, {0, 0, 0, 0}}},{20540, {{32603, 1, 0, 0},{32603, 1, 1, 17}, {0, 0, 0, 0}}},{20540, {{32603, 1, 0, 0},{32603, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{20541, {{32604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20542, {{32605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20543, {{32606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20544, {{32607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20545, {{32608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20546, {{32609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20547, {{32610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20548, {{32611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20549, {{32612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20550, {{32613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20551, {{32614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20552, {{32615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20553, {{32616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20554, {{32617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20555, {{32618, 1, 0, 0},{32618, 1, 1, 18}, {0, 0, 0, 0}}},{20555, {{32618, 1, 0, 0},{32618, 1, 1, 18}, {0, 0, 0, 0}}},{20556, {{32618, 1, 0, 0},{32618, 1, 1, 19}, {0, 0, 0, 0}}},{20556, {{32618, 1, 0, 0},{32618, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{20557, {{32619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20558, {{32620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20559, {{32621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20560, {{32622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20561, {{32623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20562, {{32624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20563, {{32625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{20564, {{32626, 1, 0, 0},{32626, 1, 1, 18}, {0, 0, 0, 0}}},{20564, {{32626, 1, 0, 0},{32626, 1, 1, 18}, {0, 0, 0, 0}}},{30289, {{32626, 1, 0, 0},{32626, 1, 1, 0},{32626, 1, 1, 19},{64090, 1, 0, 0}, {0, 0, 0, 0}}},{30289, {{32626, 1, 0, 0},{32626, 1, 1, 0},{32626, 1, 1, 19},{64090, 1, 0, 0}, {0, 0, 0, 0}}},{30289, {{32626, 1, 0, 0},{32626, 1, 1, 0},{32626, 1, 1, 19},{64090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20565, {{32627, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{20566, {{32629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20567, {{32630, 1, 0, 0},{32630, 1, 1, 16}, {0, 0, 0, 0}}},{20567, {{32630, 1, 0, 0},{32630, 1, 1, 16}, {0, 0, 0, 0}}},{20568, {{32630, 1, 0, 0},{32630, 1, 1, 17}, {0, 0, 0, 0}}},{20568, {{32630, 1, 0, 0},{32630, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{20569, {{32631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20570, {{32632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20571, {{32633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20572, {{32634, 1, 0, 0},{32634, 1, 1, 16}, {0, 0, 0, 0}}},{20572, {{32634, 1, 0, 0},{32634, 1, 1, 16}, {0, 0, 0, 0}}},{20573, {{32634, 1, 0, 0},{32634, 1, 1, 17}, {0, 0, 0, 0}}},{20573, {{32634, 1, 0, 0},{32634, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{20574, {{32635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20575, {{32636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20576, {{32637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20577, {{32638, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20578, {{32639, 1, 0, 0},{32639, 1, 1, 17}, {0, 0, 0, 0}}},{20578, {{32639, 1, 0, 0},{32639, 1, 1, 17}, {0, 0, 0, 0}}},{20579, {{32639, 1, 0, 0},{32639, 1, 1, 18}, {0, 0, 0, 0}}},{20579, {{32639, 1, 0, 0},{32639, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20580, {{32640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{20581, {{32642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20582, {{32643, 1, 0, 0},{32643, 1, 1, 17}, {0, 0, 0, 0}}},{20582, {{32643, 1, 0, 0},{32643, 1, 1, 17}, {0, 0, 0, 0}}},{20583, {{32643, 1, 0, 0},{32643, 1, 1, 18}, {0, 0, 0, 0}}},{20583, {{32643, 1, 0, 0},{32643, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20584, {{32644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20585, {{32645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20586, {{32646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20587, {{32647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20588, {{32648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20589, {{32649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20590, {{32650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20591, {{32651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20592, {{32652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20593, {{32653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20594, {{32654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20595, {{32655, 1, 0, 0},{32655, 1, 1, 17}, {0, 0, 0, 0}}},{20595, {{32655, 1, 0, 0},{32655, 1, 1, 17}, {0, 0, 0, 0}}},{20596, {{32655, 1, 0, 0},{32655, 1, 1, 18}, {0, 0, 0, 0}}},{20596, {{32655, 1, 0, 0},{32655, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20597, {{32656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20599, {{32657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20600, {{32658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20601, {{32659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20602, {{32660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{20604, {{32661, 1, 0, 0},{32661, 1, 1, 16}, {0, 0, 0, 0}}},{20604, {{32661, 1, 0, 0},{32661, 1, 1, 16}, {0, 0, 0, 0}}},{20605, {{32661, 1, 0, 0},{32661, 1, 1, 17}, {0, 0, 0, 0}}},{20605, {{32661, 1, 0, 0},{32661, 1, 1, 17}, {0, 0, 0, 0}}},{20606, {{32661, 1, 0, 0},{32661, 1, 1, 18}, {0, 0, 0, 0}}},{20606, {{32661, 1, 0, 0},{32661, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20607, {{32662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20608, {{32663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20609, {{32664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20610, {{32665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20611, {{32666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20612, {{32667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20613, {{32668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20614, {{32669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{20615, {{32670, 1, 0, 0},{32670, 1, 1, 18}, {0, 0, 0, 0}}},{20615, {{32670, 1, 0, 0},{32670, 1, 1, 18}, {0, 0, 0, 0}}},{20616, {{32670, 1, 0, 0},{32670, 1, 1, 17}, {0, 0, 0, 0}}},{20616, {{32670, 1, 0, 0},{32670, 1, 1, 17}, {0, 0, 0, 0}}},{58348, {{32670, 1, 0, 0},{32670, 1, 1, 19}, {0, 0, 0, 0}}},{58348, {{32670, 1, 0, 0},{32670, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{20617, {{32672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{20618, {{32673, 1, 0, 0},{32673, 1, 1, 18}, {0, 0, 0, 0}}},{20618, {{32673, 1, 0, 0},{32673, 1, 1, 18}, {0, 0, 0, 0}}},{20619, {{32673, 1, 0, 0},{32673, 1, 1, 19}, {0, 0, 0, 0}}},{20619, {{32673, 1, 0, 0},{32673, 1, 1, 19}, {0, 0, 0, 0}}},{20620, {{32673, 1, 0, 0},{32673, 1, 1, 20}, {0, 0, 0, 0}}},{20620, {{32673, 1, 0, 0},{32673, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{20621, {{32674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20622, {{32675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20623, {{32676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20624, {{32677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20625, {{32678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20626, {{32679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20627, {{32680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20628, {{32681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20629, {{32682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20630, {{32683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20631, {{32684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20632, {{32685, 1, 0, 0},{32685, 1, 1, 17}, {0, 0, 0, 0}}},{20632, {{32685, 1, 0, 0},{32685, 1, 1, 17}, {0, 0, 0, 0}}},{20633, {{32685, 1, 0, 0},{32685, 1, 1, 18}, {0, 0, 0, 0}}},{20633, {{32685, 1, 0, 0},{32685, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{20634, {{32686, 1, 0, 0},{32686, 1, 1, 18}, {0, 0, 0, 0}}},{20634, {{32686, 1, 0, 0},{32686, 1, 1, 18}, {0, 0, 0, 0}}},{20635, {{32686, 1, 0, 0},{32686, 1, 1, 19}, {0, 0, 0, 0}}},{20635, {{32686, 1, 0, 0},{32686, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {6, {{20636, {{32687, 1, 0, 0},{32687, 1, 1, 17}, {0, 0, 0, 0}}},{20636, {{32687, 1, 0, 0},{32687, 1, 1, 17}, {0, 0, 0, 0}}},{20637, {{32687, 1, 0, 0},{32687, 1, 1, 19}, {0, 0, 0, 0}}},{20637, {{32687, 1, 0, 0},{32687, 1, 1, 19}, {0, 0, 0, 0}}},{20638, {{32687, 1, 0, 0},{32687, 1, 1, 18}, {0, 0, 0, 0}}},{20638, {{32687, 1, 0, 0},{32687, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{20639, {{32689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20640, {{32690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20641, {{32691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20642, {{32692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20643, {{32693, 1, 0, 0},{32693, 1, 1, 16}, {0, 0, 0, 0}}},{20643, {{32693, 1, 0, 0},{32693, 1, 1, 16}, {0, 0, 0, 0}}},{20644, {{32693, 1, 0, 0},{32693, 1, 1, 17}, {0, 0, 0, 0}}},{20644, {{32693, 1, 0, 0},{32693, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{20645, {{32694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20646, {{32695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20647, {{32696, 1, 0, 0},{32696, 1, 1, 17}, {0, 0, 0, 0}}},{20647, {{32696, 1, 0, 0},{32696, 1, 1, 17}, {0, 0, 0, 0}}},{20648, {{32696, 1, 0, 0},{32696, 1, 1, 18}, {0, 0, 0, 0}}},{20648, {{32696, 1, 0, 0},{32696, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20649, {{32697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20650, {{32698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20651, {{32699, 1, 0, 0},{32699, 1, 1, 17}, {0, 0, 0, 0}}},{20651, {{32699, 1, 0, 0},{32699, 1, 1, 17}, {0, 0, 0, 0}}},{20652, {{32699, 1, 0, 0},{32699, 1, 1, 16}, {0, 0, 0, 0}}},{20652, {{32699, 1, 0, 0},{32699, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{20653, {{32700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {7, {{20654, {{32701, 1, 0, 0},{32701, 1, 1, 18}, {0, 0, 0, 0}}},{20654, {{32701, 1, 0, 0},{32701, 1, 1, 18}, {0, 0, 0, 0}}},{30216, {{32701, 1, 0, 0},{32701, 1, 1, 0},{32701, 1, 1, 19},{64030, 1, 0, 0}, {0, 0, 0, 0}}},{30216, {{32701, 1, 0, 0},{32701, 1, 1, 0},{32701, 1, 1, 19},{64030, 1, 0, 0}, {0, 0, 0, 0}}},{30216, {{32701, 1, 0, 0},{32701, 1, 1, 0},{32701, 1, 1, 19},{64030, 1, 0, 0}, {0, 0, 0, 0}}},{58352, {{32701, 1, 0, 0},{32701, 1, 1, 20}, {0, 0, 0, 0}}},{58352, {{32701, 1, 0, 0},{32701, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{20655, {{32702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20656, {{32703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20657, {{32704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{20659, {{32705, 1, 0, 0},{32705, 1, 1, 20}, {0, 0, 0, 0}}},{20659, {{32705, 1, 0, 0},{32705, 1, 1, 20}, {0, 0, 0, 0}}},{20660, {{32705, 1, 0, 0},{32705, 1, 1, 19}, {0, 0, 0, 0}}},{20660, {{32705, 1, 0, 0},{32705, 1, 1, 19}, {0, 0, 0, 0}}},{20661, {{32705, 1, 0, 0},{32705, 1, 1, 21}, {0, 0, 0, 0}}},{20661, {{32705, 1, 0, 0},{32705, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{20662, {{32706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20663, {{32707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20664, {{32708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20665, {{32709, 1, 0, 0},{32709, 1, 1, 18}, {0, 0, 0, 0}}},{20665, {{32709, 1, 0, 0},{32709, 1, 1, 18}, {0, 0, 0, 0}}},{20666, {{32709, 1, 0, 0},{32709, 1, 1, 19}, {0, 0, 0, 0}}},{20666, {{32709, 1, 0, 0},{32709, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{20667, {{32710, 1, 0, 0},{32710, 1, 1, 18}, {0, 0, 0, 0}}},{20667, {{32710, 1, 0, 0},{32710, 1, 1, 18}, {0, 0, 0, 0}}},{20668, {{32710, 1, 0, 0},{32710, 1, 1, 19}, {0, 0, 0, 0}}},{20668, {{32710, 1, 0, 0},{32710, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{20669, {{32711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20670, {{32712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20671, {{32713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20672, {{32714, 1, 0, 0},{32714, 1, 1, 17}, {0, 0, 0, 0}}},{20672, {{32714, 1, 0, 0},{32714, 1, 1, 17}, {0, 0, 0, 0}}},{20673, {{32714, 1, 0, 0},{32714, 1, 1, 18}, {0, 0, 0, 0}}},{20673, {{32714, 1, 0, 0},{32714, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20674, {{32715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20675, {{32716, 1, 0, 0},{32716, 1, 1, 19}, {0, 0, 0, 0}}},{20675, {{32716, 1, 0, 0},{32716, 1, 1, 19}, {0, 0, 0, 0}}},{20676, {{32716, 1, 0, 0},{32716, 1, 1, 18}, {0, 0, 0, 0}}},{20676, {{32716, 1, 0, 0},{32716, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20677, {{32717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20678, {{32718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20679, {{32719, 1, 0, 0},{32719, 1, 1, 17}, {0, 0, 0, 0}}},{20679, {{32719, 1, 0, 0},{32719, 1, 1, 17}, {0, 0, 0, 0}}},{58355, {{32719, 1, 0, 0},{32719, 1, 1, 18}, {0, 0, 0, 0}}},{58355, {{32719, 1, 0, 0},{32719, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20680, {{32720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20681, {{32721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20682, {{32722, 1, 0, 0},{32722, 1, 1, 19}, {0, 0, 0, 0}}},{20682, {{32722, 1, 0, 0},{32722, 1, 1, 19}, {0, 0, 0, 0}}},{20683, {{32722, 1, 0, 0},{32722, 1, 1, 18}, {0, 0, 0, 0}}},{20683, {{32722, 1, 0, 0},{32722, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20684, {{32723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20685, {{32724, 1, 0, 0},{32724, 1, 1, 19}, {0, 0, 0, 0}}},{20685, {{32724, 1, 0, 0},{32724, 1, 1, 19}, {0, 0, 0, 0}}},{20686, {{32724, 1, 0, 0},{32724, 1, 1, 18}, {0, 0, 0, 0}}},{20686, {{32724, 1, 0, 0},{32724, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{20687, {{32725, 1, 0, 0},{32725, 1, 1, 17}, {0, 0, 0, 0}}},{20687, {{32725, 1, 0, 0},{32725, 1, 1, 17}, {0, 0, 0, 0}}},{20688, {{32725, 1, 0, 0},{32725, 1, 1, 18}, {0, 0, 0, 0}}},{20688, {{32725, 1, 0, 0},{32725, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20689, {{32726, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20690, {{32727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{20691, {{32731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20692, {{32732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20693, {{32733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20694, {{32734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20695, {{32735, 1, 0, 0},{32735, 1, 1, 17}, {0, 0, 0, 0}}},{20695, {{32735, 1, 0, 0},{32735, 1, 1, 17}, {0, 0, 0, 0}}},{58358, {{32735, 1, 0, 0},{32735, 1, 1, 18}, {0, 0, 0, 0}}},{58358, {{32735, 1, 0, 0},{32735, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{20696, {{32736, 1, 0, 0},{32736, 1, 1, 20}, {0, 0, 0, 0}}},{20696, {{32736, 1, 0, 0},{32736, 1, 1, 20}, {0, 0, 0, 0}}},{20697, {{32736, 1, 0, 0},{32736, 1, 1, 18}, {0, 0, 0, 0}}},{20697, {{32736, 1, 0, 0},{32736, 1, 1, 18}, {0, 0, 0, 0}}},{20698, {{32736, 1, 0, 0},{32736, 1, 1, 19}, {0, 0, 0, 0}}},{20698, {{32736, 1, 0, 0},{32736, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {6, {{20699, {{32737, 1, 0, 0},{32737, 1, 1, 21}, {0, 0, 0, 0}}},{20699, {{32737, 1, 0, 0},{32737, 1, 1, 21}, {0, 0, 0, 0}}},{20700, {{32737, 1, 0, 0},{32737, 1, 1, 22}, {0, 0, 0, 0}}},{20700, {{32737, 1, 0, 0},{32737, 1, 1, 22}, {0, 0, 0, 0}}},{20701, {{32737, 1, 0, 0},{32737, 1, 1, 20}, {0, 0, 0, 0}}},{20701, {{32737, 1, 0, 0},{32737, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{20702, {{32738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20703, {{32739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20704, {{32740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20705, {{32741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{20707, {{32742, 1, 0, 0},{32742, 1, 1, 17}, {0, 0, 0, 0}}},{20707, {{32742, 1, 0, 0},{32742, 1, 1, 17}, {0, 0, 0, 0}}},{20708, {{32742, 1, 0, 0},{32742, 1, 1, 19}, {0, 0, 0, 0}}},{20708, {{32742, 1, 0, 0},{32742, 1, 1, 19}, {0, 0, 0, 0}}},{20709, {{32742, 1, 0, 0},{32742, 1, 1, 18}, {0, 0, 0, 0}}},{20709, {{32742, 1, 0, 0},{32742, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20710, {{32743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20711, {{32744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{20712, {{32745, 1, 0, 0},{32745, 1, 1, 21}, {0, 0, 0, 0}}},{20712, {{32745, 1, 0, 0},{32745, 1, 1, 21}, {0, 0, 0, 0}}},{20713, {{32745, 1, 0, 0},{32745, 1, 1, 19}, {0, 0, 0, 0}}},{20713, {{32745, 1, 0, 0},{32745, 1, 1, 19}, {0, 0, 0, 0}}},{20714, {{32745, 1, 0, 0},{32745, 1, 1, 20}, {0, 0, 0, 0}}},{20714, {{32745, 1, 0, 0},{32745, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{20715, {{32746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{20716, {{32747, 1, 0, 0},{32747, 1, 1, 19}, {0, 0, 0, 0}}},{20716, {{32747, 1, 0, 0},{32747, 1, 1, 19}, {0, 0, 0, 0}}},{20717, {{32747, 1, 0, 0},{32747, 1, 1, 18}, {0, 0, 0, 0}}},{20717, {{32747, 1, 0, 0},{32747, 1, 1, 18}, {0, 0, 0, 0}}},{20718, {{32747, 1, 0, 0},{32747, 1, 1, 20}, {0, 0, 0, 0}}},{20718, {{32747, 1, 0, 0},{32747, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{20719, {{32748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20720, {{32749, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20721, {{32750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20722, {{32751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{20723, {{32752, 1, 0, 0},{32752, 1, 1, 19}, {0, 0, 0, 0}}},{20723, {{32752, 1, 0, 0},{32752, 1, 1, 19}, {0, 0, 0, 0}}},{20724, {{32752, 1, 0, 0},{32752, 1, 1, 18}, {0, 0, 0, 0}}},{20724, {{32752, 1, 0, 0},{32752, 1, 1, 18}, {0, 0, 0, 0}}},{20725, {{32752, 1, 0, 0},{32752, 1, 1, 20}, {0, 0, 0, 0}}},{20725, {{32752, 1, 0, 0},{32752, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{20728, {{32753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20729, {{32754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20730, {{32755, 1, 0, 0},{32755, 1, 1, 17}, {0, 0, 0, 0}}},{20730, {{32755, 1, 0, 0},{32755, 1, 1, 17}, {0, 0, 0, 0}}},{20731, {{32755, 1, 0, 0},{32755, 1, 1, 18}, {0, 0, 0, 0}}},{20731, {{32755, 1, 0, 0},{32755, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20732, {{32756, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20733, {{32757, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{20734, {{32759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20735, {{32760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20736, {{32761, 1, 0, 0},{32761, 1, 1, 17}, {0, 0, 0, 0}}},{20736, {{32761, 1, 0, 0},{32761, 1, 1, 17}, {0, 0, 0, 0}}},{20737, {{32761, 1, 0, 0},{32761, 1, 1, 18}, {0, 0, 0, 0}}},{20737, {{32761, 1, 0, 0},{32761, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20738, {{32762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20739, {{32763, 1, 0, 0},{32763, 1, 1, 19}, {0, 0, 0, 0}}},{20739, {{32763, 1, 0, 0},{32763, 1, 1, 19}, {0, 0, 0, 0}}},{20740, {{32763, 1, 0, 0},{32763, 1, 1, 18}, {0, 0, 0, 0}}},{20740, {{32763, 1, 0, 0},{32763, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {8, {{20741, {{32764, 1, 0, 0},{32764, 1, 1, 19}, {0, 0, 0, 0}}},{20741, {{32764, 1, 0, 0},{32764, 1, 1, 19}, {0, 0, 0, 0}}},{20742, {{32764, 1, 0, 0},{32764, 1, 1, 21}, {0, 0, 0, 0}}},{20742, {{32764, 1, 0, 0},{32764, 1, 1, 21}, {0, 0, 0, 0}}},{20743, {{32764, 1, 0, 0},{32764, 1, 1, 20}, {0, 0, 0, 0}}},{20743, {{32764, 1, 0, 0},{32764, 1, 1, 20}, {0, 0, 0, 0}}},{20744, {{32764, 1, 0, 0},{32764, 1, 1, 22}, {0, 0, 0, 0}}},{20744, {{32764, 1, 0, 0},{32764, 1, 1, 22}, {0, 0, 0, 0}}}}},
    {1, {{20745, {{32765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20746, {{32766, 1, 0, 0},{32766, 1, 1, 17}, {0, 0, 0, 0}}},{20746, {{32766, 1, 0, 0},{32766, 1, 1, 17}, {0, 0, 0, 0}}},{20747, {{32766, 1, 0, 0},{32766, 1, 1, 18}, {0, 0, 0, 0}}},{20747, {{32766, 1, 0, 0},{32766, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20748, {{32767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20749, {{32768, 1, 0, 0},{32768, 1, 1, 19}, {0, 0, 0, 0}}},{20749, {{32768, 1, 0, 0},{32768, 1, 1, 19}, {0, 0, 0, 0}}},{20750, {{32768, 1, 0, 0},{32768, 1, 1, 18}, {0, 0, 0, 0}}},{20750, {{32768, 1, 0, 0},{32768, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20751, {{32769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20752, {{32770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20753, {{32771, 1, 0, 0},{32771, 1, 1, 18}, {0, 0, 0, 0}}},{20753, {{32771, 1, 0, 0},{32771, 1, 1, 18}, {0, 0, 0, 0}}},{20754, {{32771, 1, 0, 0},{32771, 1, 1, 19}, {0, 0, 0, 0}}},{20754, {{32771, 1, 0, 0},{32771, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{20755, {{32772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{20756, {{32773, 1, 0, 0},{32773, 1, 1, 18}, {0, 0, 0, 0}}},{20756, {{32773, 1, 0, 0},{32773, 1, 1, 18}, {0, 0, 0, 0}}},{30290, {{32773, 1, 0, 0},{32773, 1, 1, 0},{32773, 1, 1, 19},{64091, 1, 0, 0}, {0, 0, 0, 0}}},{30290, {{32773, 1, 0, 0},{32773, 1, 1, 0},{32773, 1, 1, 19},{64091, 1, 0, 0}, {0, 0, 0, 0}}},{30290, {{32773, 1, 0, 0},{32773, 1, 1, 0},{32773, 1, 1, 19},{64091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20757, {{32774, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20758, {{32775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20759, {{32776, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{68043, {{32777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20760, {{32778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20761, {{32779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20762, {{32780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20763, {{32781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20764, {{32782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20765, {{32783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20766, {{32784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20767, {{32785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20768, {{32786, 1, 0, 0},{32786, 1, 1, 19}, {0, 0, 0, 0}}},{20768, {{32786, 1, 0, 0},{32786, 1, 1, 19}, {0, 0, 0, 0}}},{20769, {{32786, 1, 0, 0},{32786, 1, 1, 18}, {0, 0, 0, 0}}},{20769, {{32786, 1, 0, 0},{32786, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20770, {{32787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20771, {{32788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20772, {{32789, 1, 0, 0},{32789, 1, 1, 19}, {0, 0, 0, 0}}},{20772, {{32789, 1, 0, 0},{32789, 1, 1, 19}, {0, 0, 0, 0}}},{20773, {{32789, 1, 0, 0},{32789, 1, 1, 18}, {0, 0, 0, 0}}},{20773, {{32789, 1, 0, 0},{32789, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20774, {{32790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20776, {{32791, 1, 0, 0},{32791, 1, 1, 19}, {0, 0, 0, 0}}},{20776, {{32791, 1, 0, 0},{32791, 1, 1, 19}, {0, 0, 0, 0}}},{20777, {{32791, 1, 0, 0},{32791, 1, 1, 18}, {0, 0, 0, 0}}},{20777, {{32791, 1, 0, 0},{32791, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{20778, {{32792, 1, 0, 0},{32792, 1, 1, 17}, {0, 0, 0, 0}}},{20778, {{32792, 1, 0, 0},{32792, 1, 1, 17}, {0, 0, 0, 0}}},{20779, {{32792, 1, 0, 0},{32792, 1, 1, 18}, {0, 0, 0, 0}}},{20779, {{32792, 1, 0, 0},{32792, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{20780, {{32793, 1, 0, 0},{32793, 1, 1, 17}, {0, 0, 0, 0}}},{20780, {{32793, 1, 0, 0},{32793, 1, 1, 17}, {0, 0, 0, 0}}},{20781, {{32793, 1, 0, 0},{32793, 1, 1, 18}, {0, 0, 0, 0}}},{20781, {{32793, 1, 0, 0},{32793, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20782, {{32794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20783, {{32795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{20784, {{32796, 1, 0, 0},{32796, 1, 1, 17}, {0, 0, 0, 0}}},{20784, {{32796, 1, 0, 0},{32796, 1, 1, 17}, {0, 0, 0, 0}}},{20785, {{32796, 1, 0, 0},{32796, 1, 1, 18}, {0, 0, 0, 0}}},{20785, {{32796, 1, 0, 0},{32796, 1, 1, 18}, {0, 0, 0, 0}}},{20786, {{32796, 1, 0, 0},{32796, 1, 1, 19}, {0, 0, 0, 0}}},{20786, {{32796, 1, 0, 0},{32796, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{20787, {{32797, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20788, {{32798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20789, {{32799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20790, {{32800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20791, {{32801, 1, 0, 0},{32801, 1, 1, 17}, {0, 0, 0, 0}}},{20791, {{32801, 1, 0, 0},{32801, 1, 1, 17}, {0, 0, 0, 0}}},{20792, {{32801, 1, 0, 0},{32801, 1, 1, 18}, {0, 0, 0, 0}}},{20792, {{32801, 1, 0, 0},{32801, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{20793, {{32803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20794, {{32804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{20795, {{32806, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{20796, {{32808, 1, 0, 0},{32808, 1, 1, 17}, {0, 0, 0, 0}}},{20796, {{32808, 1, 0, 0},{32808, 1, 1, 17}, {0, 0, 0, 0}}},{20797, {{32808, 1, 0, 0},{32808, 1, 1, 19}, {0, 0, 0, 0}}},{20797, {{32808, 1, 0, 0},{32808, 1, 1, 19}, {0, 0, 0, 0}}},{20798, {{32808, 1, 0, 0},{32808, 1, 1, 18}, {0, 0, 0, 0}}},{20798, {{32808, 1, 0, 0},{32808, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20799, {{32809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20804, {{32810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20805, {{32811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20806, {{32812, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20807, {{32813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20808, {{32814, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20809, {{32815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20810, {{32816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20812, {{32817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{20813, {{32819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20814, {{32820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20815, {{32821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20816, {{32822, 1, 0, 0},{32822, 1, 1, 18}, {0, 0, 0, 0}}},{20816, {{32822, 1, 0, 0},{32822, 1, 1, 18}, {0, 0, 0, 0}}},{20817, {{32822, 1, 0, 0},{32822, 1, 1, 19}, {0, 0, 0, 0}}},{20817, {{32822, 1, 0, 0},{32822, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{20818, {{32823, 1, 0, 0},{32823, 1, 1, 17}, {0, 0, 0, 0}}},{20818, {{32823, 1, 0, 0},{32823, 1, 1, 17}, {0, 0, 0, 0}}},{20819, {{32823, 1, 0, 0},{32823, 1, 1, 18}, {0, 0, 0, 0}}},{20819, {{32823, 1, 0, 0},{32823, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20820, {{32824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20821, {{32825, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20822, {{32826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20823, {{32827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20824, {{32828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20825, {{32829, 1, 0, 0},{32829, 1, 1, 17}, {0, 0, 0, 0}}},{20825, {{32829, 1, 0, 0},{32829, 1, 1, 17}, {0, 0, 0, 0}}},{58365, {{32829, 1, 0, 0},{32829, 1, 1, 18}, {0, 0, 0, 0}}},{58365, {{32829, 1, 0, 0},{32829, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20826, {{32830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20827, {{32831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20828, {{32832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{20829, {{32835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20830, {{32836, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20831, {{32837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20832, {{32838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20833, {{32839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20834, {{32840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20835, {{32841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20836, {{32842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{20837, {{32846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20838, {{32847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20839, {{32848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20840, {{32849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20841, {{32850, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{20842, {{32852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20843, {{32853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{20845, {{32854, 1, 0, 0},{32854, 1, 1, 19}, {0, 0, 0, 0}}},{20845, {{32854, 1, 0, 0},{32854, 1, 1, 19}, {0, 0, 0, 0}}},{20846, {{32854, 1, 0, 0},{32854, 1, 1, 18}, {0, 0, 0, 0}}},{20846, {{32854, 1, 0, 0},{32854, 1, 1, 18}, {0, 0, 0, 0}}},{20847, {{32854, 1, 0, 0},{32854, 1, 1, 20}, {0, 0, 0, 0}}},{20847, {{32854, 1, 0, 0},{32854, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{20848, {{32855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20849, {{32856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20850, {{32857, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20851, {{32858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20852, {{32859, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20853, {{32860, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20854, {{32861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20855, {{32862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20856, {{32863, 1, 0, 0},{32863, 1, 1, 18}, {0, 0, 0, 0}}},{20856, {{32863, 1, 0, 0},{32863, 1, 1, 18}, {0, 0, 0, 0}}},{20857, {{32863, 1, 0, 0},{32863, 1, 1, 19}, {0, 0, 0, 0}}},{20857, {{32863, 1, 0, 0},{32863, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{20858, {{32864, 1, 0, 0},{32864, 1, 1, 17}, {0, 0, 0, 0}}},{20858, {{32864, 1, 0, 0},{32864, 1, 1, 17}, {0, 0, 0, 0}}},{20859, {{32864, 1, 0, 0},{32864, 1, 1, 18}, {0, 0, 0, 0}}},{20859, {{32864, 1, 0, 0},{32864, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20860, {{32865, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20861, {{32866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{20862, {{32868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20863, {{32869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20864, {{32870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20865, {{32871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20866, {{32872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{20867, {{32876, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20868, {{32877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20869, {{32878, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20870, {{32879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{20871, {{32880, 1, 0, 0},{32880, 1, 1, 18}, {0, 0, 0, 0}}},{20871, {{32880, 1, 0, 0},{32880, 1, 1, 18}, {0, 0, 0, 0}}},{20872, {{32880, 1, 0, 0},{32880, 1, 1, 19}, {0, 0, 0, 0}}},{20872, {{32880, 1, 0, 0},{32880, 1, 1, 19}, {0, 0, 0, 0}}},{20873, {{32880, 1, 0, 0},{32880, 1, 1, 20}, {0, 0, 0, 0}}},{20873, {{32880, 1, 0, 0},{32880, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{20875, {{32881, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20876, {{32882, 1, 0, 0},{32882, 1, 1, 17}, {0, 0, 0, 0}}},{20876, {{32882, 1, 0, 0},{32882, 1, 1, 17}, {0, 0, 0, 0}}},{20877, {{32882, 1, 0, 0},{32882, 1, 1, 18}, {0, 0, 0, 0}}},{20877, {{32882, 1, 0, 0},{32882, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{20878, {{32883, 1, 0, 0},{32883, 1, 1, 18}, {0, 0, 0, 0}}},{20878, {{32883, 1, 0, 0},{32883, 1, 1, 18}, {0, 0, 0, 0}}},{20879, {{32883, 1, 0, 0},{32883, 1, 1, 19}, {0, 0, 0, 0}}},{20879, {{32883, 1, 0, 0},{32883, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{20880, {{32884, 1, 0, 0},{32884, 1, 1, 18}, {0, 0, 0, 0}}},{20880, {{32884, 1, 0, 0},{32884, 1, 1, 18}, {0, 0, 0, 0}}},{20881, {{32884, 1, 0, 0},{32884, 1, 1, 19}, {0, 0, 0, 0}}},{20881, {{32884, 1, 0, 0},{32884, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{20882, {{32885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20883, {{32886, 1, 0, 0},{32886, 1, 1, 18}, {0, 0, 0, 0}}},{20883, {{32886, 1, 0, 0},{32886, 1, 1, 18}, {0, 0, 0, 0}}},{20884, {{32886, 1, 0, 0},{32886, 1, 1, 19}, {0, 0, 0, 0}}},{20884, {{32886, 1, 0, 0},{32886, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {6, {{20885, {{32887, 1, 0, 0},{32887, 1, 1, 18}, {0, 0, 0, 0}}},{20885, {{32887, 1, 0, 0},{32887, 1, 1, 18}, {0, 0, 0, 0}}},{20886, {{32887, 1, 0, 0},{32887, 1, 1, 19}, {0, 0, 0, 0}}},{20886, {{32887, 1, 0, 0},{32887, 1, 1, 19}, {0, 0, 0, 0}}},{20887, {{32887, 1, 0, 0},{32887, 1, 1, 20}, {0, 0, 0, 0}}},{20887, {{32887, 1, 0, 0},{32887, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{20888, {{32888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20889, {{32889, 1, 0, 0},{32889, 1, 1, 17}, {0, 0, 0, 0}}},{20889, {{32889, 1, 0, 0},{32889, 1, 1, 17}, {0, 0, 0, 0}}},{20890, {{32889, 1, 0, 0},{32889, 1, 1, 18}, {0, 0, 0, 0}}},{20890, {{32889, 1, 0, 0},{32889, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20891, {{32890, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20892, {{32891, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20893, {{32892, 1, 0, 0},{32892, 1, 1, 17}, {0, 0, 0, 0}}},{20893, {{32892, 1, 0, 0},{32892, 1, 1, 17}, {0, 0, 0, 0}}},{20894, {{32892, 1, 0, 0},{32892, 1, 1, 16}, {0, 0, 0, 0}}},{20894, {{32892, 1, 0, 0},{32892, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {4, {{20895, {{32893, 1, 0, 0},{32893, 1, 1, 17}, {0, 0, 0, 0}}},{20895, {{32893, 1, 0, 0},{32893, 1, 1, 17}, {0, 0, 0, 0}}},{20896, {{32893, 1, 0, 0},{32893, 1, 1, 18}, {0, 0, 0, 0}}},{20896, {{32893, 1, 0, 0},{32893, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{20897, {{32894, 1, 0, 0},{32894, 1, 1, 18}, {0, 0, 0, 0}}},{20897, {{32894, 1, 0, 0},{32894, 1, 1, 18}, {0, 0, 0, 0}}},{20898, {{32894, 1, 0, 0},{32894, 1, 1, 19}, {0, 0, 0, 0}}},{20898, {{32894, 1, 0, 0},{32894, 1, 1, 19}, {0, 0, 0, 0}}},{20899, {{32894, 1, 0, 0},{32894, 1, 1, 20}, {0, 0, 0, 0}}},{20899, {{32894, 1, 0, 0},{32894, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{20900, {{32895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{20901, {{32897, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20903, {{32898, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{20904, {{32900, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20905, {{32901, 1, 0, 0},{32901, 1, 1, 17}, {0, 0, 0, 0}}},{20905, {{32901, 1, 0, 0},{32901, 1, 1, 17}, {0, 0, 0, 0}}},{58373, {{32901, 1, 0, 0},{32901, 1, 1, 18}, {0, 0, 0, 0}}},{58373, {{32901, 1, 0, 0},{32901, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20906, {{32902, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20907, {{32903, 1, 0, 0},{32903, 1, 1, 19}, {0, 0, 0, 0}}},{20907, {{32903, 1, 0, 0},{32903, 1, 1, 19}, {0, 0, 0, 0}}},{20908, {{32903, 1, 0, 0},{32903, 1, 1, 18}, {0, 0, 0, 0}}},{20908, {{32903, 1, 0, 0},{32903, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20910, {{32904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20912, {{32905, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20913, {{32906, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20914, {{32907, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20915, {{32908, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20916, {{32909, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20917, {{32910, 1, 0, 0},{32910, 1, 1, 17}, {0, 0, 0, 0}}},{20917, {{32910, 1, 0, 0},{32910, 1, 1, 17}, {0, 0, 0, 0}}},{20918, {{32910, 1, 0, 0},{32910, 1, 1, 18}, {0, 0, 0, 0}}},{20918, {{32910, 1, 0, 0},{32910, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20919, {{32911, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20920, {{32912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20921, {{32913, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20922, {{32914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20923, {{32915, 1, 0, 0},{32915, 1, 1, 17}, {0, 0, 0, 0}}},{20923, {{32915, 1, 0, 0},{32915, 1, 1, 17}, {0, 0, 0, 0}}},{20924, {{32915, 1, 0, 0},{32915, 1, 1, 18}, {0, 0, 0, 0}}},{20924, {{32915, 1, 0, 0},{32915, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20925, {{32916, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20926, {{32917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20927, {{32918, 1, 0, 0},{32918, 1, 1, 19}, {0, 0, 0, 0}}},{20927, {{32918, 1, 0, 0},{32918, 1, 1, 19}, {0, 0, 0, 0}}},{20928, {{32918, 1, 0, 0},{32918, 1, 1, 18}, {0, 0, 0, 0}}},{20928, {{32918, 1, 0, 0},{32918, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20929, {{32919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20930, {{32920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20931, {{32921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20932, {{32922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20933, {{32923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20934, {{32924, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20935, {{32925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20936, {{32926, 1, 0, 0},{32926, 1, 1, 18}, {0, 0, 0, 0}}},{20936, {{32926, 1, 0, 0},{32926, 1, 1, 18}, {0, 0, 0, 0}}},{20937, {{32926, 1, 0, 0},{32926, 1, 1, 19}, {0, 0, 0, 0}}},{20937, {{32926, 1, 0, 0},{32926, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{20938, {{32929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20939, {{32930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20940, {{32931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20941, {{32932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20942, {{32933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20943, {{32934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20944, {{32935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20945, {{32936, 1, 0, 0},{32936, 1, 1, 16}, {0, 0, 0, 0}}},{20945, {{32936, 1, 0, 0},{32936, 1, 1, 16}, {0, 0, 0, 0}}},{20946, {{32936, 1, 0, 0},{32936, 1, 1, 17}, {0, 0, 0, 0}}},{20946, {{32936, 1, 0, 0},{32936, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{20947, {{32937, 1, 0, 0},{32937, 1, 1, 18}, {0, 0, 0, 0}}},{20947, {{32937, 1, 0, 0},{32937, 1, 1, 18}, {0, 0, 0, 0}}},{20948, {{32937, 1, 0, 0},{32937, 1, 1, 19}, {0, 0, 0, 0}}},{20948, {{32937, 1, 0, 0},{32937, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{20949, {{32938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20950, {{32939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20951, {{32940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{20952, {{32941, 1, 0, 0},{32941, 1, 1, 19}, {0, 0, 0, 0}}},{20952, {{32941, 1, 0, 0},{32941, 1, 1, 19}, {0, 0, 0, 0}}},{20953, {{32941, 1, 0, 0},{32941, 1, 1, 17}, {0, 0, 0, 0}}},{20953, {{32941, 1, 0, 0},{32941, 1, 1, 17}, {0, 0, 0, 0}}},{20954, {{32941, 1, 0, 0},{32941, 1, 1, 18}, {0, 0, 0, 0}}},{20954, {{32941, 1, 0, 0},{32941, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20955, {{32942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20956, {{32943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20957, {{32944, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20958, {{32945, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20959, {{32946, 1, 0, 0},{32946, 1, 1, 18}, {0, 0, 0, 0}}},{20959, {{32946, 1, 0, 0},{32946, 1, 1, 18}, {0, 0, 0, 0}}},{20960, {{32946, 1, 0, 0},{32946, 1, 1, 17}, {0, 0, 0, 0}}},{20960, {{32946, 1, 0, 0},{32946, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{20961, {{32947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20962, {{32948, 1, 0, 0},{32948, 1, 1, 17}, {0, 0, 0, 0}}},{20962, {{32948, 1, 0, 0},{32948, 1, 1, 17}, {0, 0, 0, 0}}},{58380, {{32948, 1, 0, 0},{32948, 1, 1, 18}, {0, 0, 0, 0}}},{58380, {{32948, 1, 0, 0},{32948, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20963, {{32949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20964, {{32950, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20965, {{32951, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20966, {{32952, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20967, {{32953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20968, {{32954, 1, 0, 0},{32954, 1, 1, 19}, {0, 0, 0, 0}}},{20968, {{32954, 1, 0, 0},{32954, 1, 1, 19}, {0, 0, 0, 0}}},{20969, {{32954, 1, 0, 0},{32954, 1, 1, 18}, {0, 0, 0, 0}}},{20969, {{32954, 1, 0, 0},{32954, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{20970, {{32955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{20971, {{32962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20972, {{32963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20973, {{32964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20974, {{32965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20975, {{32966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20976, {{32967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20977, {{32968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20978, {{32969, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20979, {{32970, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20980, {{32971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20981, {{32972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20982, {{32973, 1, 0, 0},{32973, 1, 1, 18}, {0, 0, 0, 0}}},{20982, {{32973, 1, 0, 0},{32973, 1, 1, 18}, {0, 0, 0, 0}}},{20983, {{32973, 1, 0, 0},{32973, 1, 1, 17}, {0, 0, 0, 0}}},{20983, {{32973, 1, 0, 0},{32973, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{20984, {{32974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20985, {{32975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20986, {{32976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20987, {{32977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20988, {{32978, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20989, {{32979, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20991, {{32980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20992, {{32981, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{20993, {{32982, 1, 0, 0},{32982, 1, 1, 18}, {0, 0, 0, 0}}},{20993, {{32982, 1, 0, 0},{32982, 1, 1, 18}, {0, 0, 0, 0}}},{20994, {{32982, 1, 0, 0},{32982, 1, 1, 19}, {0, 0, 0, 0}}},{20994, {{32982, 1, 0, 0},{32982, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{20995, {{32983, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20996, {{32984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20997, {{32985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20998, {{32986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{20999, {{32987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21000, {{32988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21001, {{32989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21002, {{32990, 1, 0, 0},{32990, 1, 1, 19}, {0, 0, 0, 0}}},{21002, {{32990, 1, 0, 0},{32990, 1, 1, 19}, {0, 0, 0, 0}}},{21003, {{32990, 1, 0, 0},{32990, 1, 1, 18}, {0, 0, 0, 0}}},{21003, {{32990, 1, 0, 0},{32990, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21004, {{32991, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21005, {{32992, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21006, {{32993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21007, {{32994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21008, {{32995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21009, {{32996, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21010, {{32997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21011, {{32998, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{21012, {{33005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21013, {{33006, 1, 0, 0},{33006, 1, 1, 17}, {0, 0, 0, 0}}},{21013, {{33006, 1, 0, 0},{33006, 1, 1, 17}, {0, 0, 0, 0}}},{21014, {{33006, 1, 0, 0},{33006, 1, 1, 18}, {0, 0, 0, 0}}},{21014, {{33006, 1, 0, 0},{33006, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21015, {{33007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21016, {{33008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21017, {{33009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21018, {{33010, 1, 0, 0},{33010, 1, 1, 17}, {0, 0, 0, 0}}},{21018, {{33010, 1, 0, 0},{33010, 1, 1, 17}, {0, 0, 0, 0}}},{21019, {{33010, 1, 0, 0},{33010, 1, 1, 18}, {0, 0, 0, 0}}},{21019, {{33010, 1, 0, 0},{33010, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21020, {{33011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21021, {{33012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21022, {{33013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21023, {{33014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21024, {{33015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21026, {{33016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21027, {{33017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21028, {{33018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21029, {{33019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21030, {{33020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21031, {{33021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21032, {{33022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21033, {{33023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21034, {{33024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{21035, {{33026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21036, {{33027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21037, {{33028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21038, {{33029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21039, {{33030, 1, 0, 0},{33030, 1, 1, 18}, {0, 0, 0, 0}}},{21039, {{33030, 1, 0, 0},{33030, 1, 1, 18}, {0, 0, 0, 0}}},{21040, {{33030, 1, 0, 0},{33030, 1, 1, 19}, {0, 0, 0, 0}}},{21040, {{33030, 1, 0, 0},{33030, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{21041, {{33031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{21042, {{33032, 1, 0, 0},{33032, 1, 1, 20}, {0, 0, 0, 0}}},{21042, {{33032, 1, 0, 0},{33032, 1, 1, 20}, {0, 0, 0, 0}}},{21043, {{33032, 1, 0, 0},{33032, 1, 1, 19}, {0, 0, 0, 0}}},{21043, {{33032, 1, 0, 0},{33032, 1, 1, 19}, {0, 0, 0, 0}}},{21044, {{33032, 1, 0, 0},{33032, 1, 1, 21}, {0, 0, 0, 0}}},{21044, {{33032, 1, 0, 0},{33032, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{21045, {{33033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21046, {{33034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21047, {{33035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{21048, {{33037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{21049, {{33044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21050, {{33045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21051, {{33046, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21052, {{33047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21053, {{33048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21054, {{33049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21055, {{33050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21056, {{33051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21057, {{33052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21058, {{33053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21059, {{33054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21060, {{33055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21062, {{33056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21063, {{33057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21064, {{33058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21065, {{33059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21066, {{33060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21067, {{33061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21068, {{33062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21069, {{33063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21071, {{33064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21072, {{33065, 1, 0, 0},{33065, 1, 1, 19}, {0, 0, 0, 0}}},{21072, {{33065, 1, 0, 0},{33065, 1, 1, 19}, {0, 0, 0, 0}}},{21073, {{33065, 1, 0, 0},{33065, 1, 1, 18}, {0, 0, 0, 0}}},{21073, {{33065, 1, 0, 0},{33065, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21074, {{33066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21075, {{33067, 1, 0, 0},{33067, 1, 1, 17}, {0, 0, 0, 0}}},{21075, {{33067, 1, 0, 0},{33067, 1, 1, 17}, {0, 0, 0, 0}}},{58383, {{33067, 1, 0, 0},{33067, 1, 1, 18}, {0, 0, 0, 0}}},{58383, {{33067, 1, 0, 0},{33067, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21076, {{33068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21077, {{33069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21078, {{33070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21079, {{33071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21080, {{33072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21081, {{33073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21082, {{33074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21083, {{33075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21084, {{33076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21085, {{33077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{21089, {{33081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21090, {{33082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21091, {{33083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21092, {{33084, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21093, {{33085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21094, {{33086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21096, {{33087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21097, {{33088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21098, {{33089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21099, {{33090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21100, {{33091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21104, {{33092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21105, {{33093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21106, {{33094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21107, {{33095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21108, {{33096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21109, {{33097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21110, {{33098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21111, {{33099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21112, {{33100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21113, {{33101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21114, {{33102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21115, {{33103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21116, {{33104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21117, {{33105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21118, {{33106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21119, {{33107, 1, 0, 0},{33107, 1, 1, 18}, {0, 0, 0, 0}}},{21119, {{33107, 1, 0, 0},{33107, 1, 1, 18}, {0, 0, 0, 0}}},{21120, {{33107, 1, 0, 0},{33107, 1, 1, 19}, {0, 0, 0, 0}}},{21120, {{33107, 1, 0, 0},{33107, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{21121, {{33108, 1, 0, 0},{33108, 1, 1, 18}, {0, 0, 0, 0}}},{21121, {{33108, 1, 0, 0},{33108, 1, 1, 18}, {0, 0, 0, 0}}},{21122, {{33108, 1, 0, 0},{33108, 1, 1, 19}, {0, 0, 0, 0}}},{21122, {{33108, 1, 0, 0},{33108, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{21123, {{33109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21124, {{33110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21125, {{33111, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21126, {{33112, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{21127, {{33115, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21128, {{33116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21129, {{33117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21130, {{33118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21131, {{33119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21132, {{33120, 1, 0, 0},{33120, 1, 1, 17}, {0, 0, 0, 0}}},{21132, {{33120, 1, 0, 0},{33120, 1, 1, 17}, {0, 0, 0, 0}}},{21133, {{33120, 1, 0, 0},{33120, 1, 1, 18}, {0, 0, 0, 0}}},{21133, {{33120, 1, 0, 0},{33120, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21134, {{33121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21135, {{33122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21137, {{33123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21138, {{33124, 1, 0, 0},{33124, 1, 1, 16}, {0, 0, 0, 0}}},{21138, {{33124, 1, 0, 0},{33124, 1, 1, 16}, {0, 0, 0, 0}}},{21139, {{33124, 1, 0, 0},{33124, 1, 1, 17}, {0, 0, 0, 0}}},{21139, {{33124, 1, 0, 0},{33124, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{21140, {{33125, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21141, {{33126, 1, 0, 0},{33126, 1, 1, 17}, {0, 0, 0, 0}}},{21141, {{33126, 1, 0, 0},{33126, 1, 1, 17}, {0, 0, 0, 0}}},{58390, {{33126, 1, 0, 0},{33126, 1, 1, 18}, {0, 0, 0, 0}}},{58390, {{33126, 1, 0, 0},{33126, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21142, {{33127, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21143, {{33128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21144, {{33129, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21145, {{33130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21146, {{33131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21147, {{33132, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21148, {{33133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21149, {{33134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21150, {{33135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21151, {{33136, 1, 0, 0},{33136, 1, 1, 18}, {0, 0, 0, 0}}},{21151, {{33136, 1, 0, 0},{33136, 1, 1, 18}, {0, 0, 0, 0}}},{21152, {{33136, 1, 0, 0},{33136, 1, 1, 19}, {0, 0, 0, 0}}},{21152, {{33136, 1, 0, 0},{33136, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{21153, {{33137, 1, 0, 0},{33137, 1, 1, 18}, {0, 0, 0, 0}}},{21153, {{33137, 1, 0, 0},{33137, 1, 1, 18}, {0, 0, 0, 0}}},{21154, {{33137, 1, 0, 0},{33137, 1, 1, 19}, {0, 0, 0, 0}}},{21154, {{33137, 1, 0, 0},{33137, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{21155, {{33138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21156, {{33139, 1, 0, 0},{33139, 1, 1, 17}, {0, 0, 0, 0}}},{21156, {{33139, 1, 0, 0},{33139, 1, 1, 17}, {0, 0, 0, 0}}},{21157, {{33139, 1, 0, 0},{33139, 1, 1, 18}, {0, 0, 0, 0}}},{21157, {{33139, 1, 0, 0},{33139, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21158, {{33140, 1, 0, 0},{33140, 1, 1, 17}, {0, 0, 0, 0}}},{21158, {{33140, 1, 0, 0},{33140, 1, 1, 17}, {0, 0, 0, 0}}},{21159, {{33140, 1, 0, 0},{33140, 1, 1, 18}, {0, 0, 0, 0}}},{21159, {{33140, 1, 0, 0},{33140, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21160, {{33141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21161, {{33142, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21162, {{33143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21163, {{33144, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21164, {{33145, 1, 0, 0},{33145, 1, 1, 17}, {0, 0, 0, 0}}},{21164, {{33145, 1, 0, 0},{33145, 1, 1, 17}, {0, 0, 0, 0}}},{21165, {{33145, 1, 0, 0},{33145, 1, 1, 18}, {0, 0, 0, 0}}},{21165, {{33145, 1, 0, 0},{33145, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21166, {{33146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{21167, {{33149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{21168, {{33151, 1, 0, 0},{33151, 1, 1, 18}, {0, 0, 0, 0}}},{21168, {{33151, 1, 0, 0},{33151, 1, 1, 18}, {0, 0, 0, 0}}},{21169, {{33151, 1, 0, 0},{33151, 1, 1, 19}, {0, 0, 0, 0}}},{21169, {{33151, 1, 0, 0},{33151, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{21170, {{33152, 1, 0, 0},{33152, 1, 1, 17}, {0, 0, 0, 0}}},{21170, {{33152, 1, 0, 0},{33152, 1, 1, 17}, {0, 0, 0, 0}}},{21171, {{33152, 1, 0, 0},{33152, 1, 1, 18}, {0, 0, 0, 0}}},{21171, {{33152, 1, 0, 0},{33152, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21172, {{33153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21173, {{33154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21174, {{33155, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{21175, {{33156, 1, 0, 0},{33156, 1, 1, 17}, {0, 0, 0, 0}}},{21175, {{33156, 1, 0, 0},{33156, 1, 1, 17}, {0, 0, 0, 0}}},{21176, {{33156, 1, 0, 0},{33156, 1, 1, 19}, {0, 0, 0, 0}}},{21176, {{33156, 1, 0, 0},{33156, 1, 1, 19}, {0, 0, 0, 0}}},{21177, {{33156, 1, 0, 0},{33156, 1, 1, 18}, {0, 0, 0, 0}}},{21177, {{33156, 1, 0, 0},{33156, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21178, {{33157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21179, {{33158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21180, {{33159, 1, 0, 0},{33159, 1, 1, 16}, {0, 0, 0, 0}}},{21180, {{33159, 1, 0, 0},{33159, 1, 1, 16}, {0, 0, 0, 0}}},{21181, {{33159, 1, 0, 0},{33159, 1, 1, 17}, {0, 0, 0, 0}}},{21181, {{33159, 1, 0, 0},{33159, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{21182, {{33160, 1, 0, 0},{33160, 1, 1, 17}, {0, 0, 0, 0}}},{21182, {{33160, 1, 0, 0},{33160, 1, 1, 17}, {0, 0, 0, 0}}},{58398, {{33160, 1, 0, 0},{33160, 1, 1, 18}, {0, 0, 0, 0}}},{58398, {{33160, 1, 0, 0},{33160, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21183, {{33161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21184, {{33162, 1, 0, 0},{33162, 1, 1, 19}, {0, 0, 0, 0}}},{21184, {{33162, 1, 0, 0},{33162, 1, 1, 19}, {0, 0, 0, 0}}},{21185, {{33162, 1, 0, 0},{33162, 1, 1, 18}, {0, 0, 0, 0}}},{21185, {{33162, 1, 0, 0},{33162, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21186, {{33163, 1, 0, 0},{33163, 1, 1, 17}, {0, 0, 0, 0}}},{21186, {{33163, 1, 0, 0},{33163, 1, 1, 17}, {0, 0, 0, 0}}},{21187, {{33163, 1, 0, 0},{33163, 1, 1, 18}, {0, 0, 0, 0}}},{21187, {{33163, 1, 0, 0},{33163, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21188, {{33164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21189, {{33165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21190, {{33166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21191, {{33167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21192, {{33168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{21193, {{33170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21194, {{33171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21195, {{33172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21196, {{33173, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21197, {{33174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21198, {{33175, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21199, {{33176, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21200, {{33177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21201, {{33178, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21202, {{33179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21203, {{33180, 1, 0, 0},{33180, 1, 1, 17}, {0, 0, 0, 0}}},{21203, {{33180, 1, 0, 0},{33180, 1, 1, 17}, {0, 0, 0, 0}}},{21204, {{33180, 1, 0, 0},{33180, 1, 1, 18}, {0, 0, 0, 0}}},{21204, {{33180, 1, 0, 0},{33180, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21205, {{33181, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21206, {{33182, 1, 0, 0},{33182, 1, 1, 17}, {0, 0, 0, 0}}},{21206, {{33182, 1, 0, 0},{33182, 1, 1, 17}, {0, 0, 0, 0}}},{21207, {{33182, 1, 0, 0},{33182, 1, 1, 18}, {0, 0, 0, 0}}},{21207, {{33182, 1, 0, 0},{33182, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21208, {{33183, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21209, {{33184, 1, 0, 0},{33184, 1, 1, 17}, {0, 0, 0, 0}}},{21209, {{33184, 1, 0, 0},{33184, 1, 1, 17}, {0, 0, 0, 0}}},{21210, {{33184, 1, 0, 0},{33184, 1, 1, 18}, {0, 0, 0, 0}}},{21210, {{33184, 1, 0, 0},{33184, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21211, {{33185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21212, {{33186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21213, {{33187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21214, {{33188, 1, 0, 0},{33188, 1, 1, 17}, {0, 0, 0, 0}}},{21214, {{33188, 1, 0, 0},{33188, 1, 1, 17}, {0, 0, 0, 0}}},{21215, {{33188, 1, 0, 0},{33188, 1, 1, 18}, {0, 0, 0, 0}}},{21215, {{33188, 1, 0, 0},{33188, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{21216, {{33190, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21217, {{33191, 1, 0, 0},{33191, 1, 1, 16}, {0, 0, 0, 0}}},{21217, {{33191, 1, 0, 0},{33191, 1, 1, 16}, {0, 0, 0, 0}}},{58400, {{33191, 1, 0, 0},{33191, 1, 1, 17}, {0, 0, 0, 0}}},{58400, {{33191, 1, 0, 0},{33191, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{21218, {{33192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21219, {{33193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21220, {{33194, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21221, {{33195, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21222, {{33196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21223, {{33197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21224, {{33198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21225, {{33199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21226, {{33200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21227, {{33201, 1, 0, 0},{33201, 1, 1, 16}, {0, 0, 0, 0}}},{21227, {{33201, 1, 0, 0},{33201, 1, 1, 16}, {0, 0, 0, 0}}},{21228, {{33201, 1, 0, 0},{33201, 1, 1, 17}, {0, 0, 0, 0}}},{21228, {{33201, 1, 0, 0},{33201, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{21229, {{33202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21230, {{33203, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21231, {{33204, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{21232, {{33205, 1, 0, 0},{33205, 1, 1, 18}, {0, 0, 0, 0}}},{21232, {{33205, 1, 0, 0},{33205, 1, 1, 18}, {0, 0, 0, 0}}},{21233, {{33205, 1, 0, 0},{33205, 1, 1, 20}, {0, 0, 0, 0}}},{21233, {{33205, 1, 0, 0},{33205, 1, 1, 20}, {0, 0, 0, 0}}},{21234, {{33205, 1, 0, 0},{33205, 1, 1, 19}, {0, 0, 0, 0}}},{21234, {{33205, 1, 0, 0},{33205, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{21235, {{33207, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21236, {{33208, 1, 0, 0},{33208, 1, 1, 18}, {0, 0, 0, 0}}},{21236, {{33208, 1, 0, 0},{33208, 1, 1, 18}, {0, 0, 0, 0}}},{21237, {{33208, 1, 0, 0},{33208, 1, 1, 17}, {0, 0, 0, 0}}},{21237, {{33208, 1, 0, 0},{33208, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{21238, {{33209, 1, 0, 0},{33209, 1, 1, 16}, {0, 0, 0, 0}}},{21238, {{33209, 1, 0, 0},{33209, 1, 1, 16}, {0, 0, 0, 0}}},{21239, {{33209, 1, 0, 0},{33209, 1, 1, 17}, {0, 0, 0, 0}}},{21239, {{33209, 1, 0, 0},{33209, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{21240, {{33210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21241, {{33211, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21242, {{33212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21243, {{33213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21245, {{33214, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21246, {{33215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21247, {{33216, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21248, {{33217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21249, {{33218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21250, {{33219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21251, {{33220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21252, {{33221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21253, {{33222, 1, 0, 0},{33222, 1, 1, 17}, {0, 0, 0, 0}}},{21253, {{33222, 1, 0, 0},{33222, 1, 1, 17}, {0, 0, 0, 0}}},{21254, {{33222, 1, 0, 0},{33222, 1, 1, 18}, {0, 0, 0, 0}}},{21254, {{33222, 1, 0, 0},{33222, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21255, {{33223, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{21256, {{33224, 1, 0, 0},{33224, 1, 1, 17}, {0, 0, 0, 0}}},{21256, {{33224, 1, 0, 0},{33224, 1, 1, 17}, {0, 0, 0, 0}}},{21257, {{33224, 1, 0, 0},{33224, 1, 1, 18}, {0, 0, 0, 0}}},{21257, {{33224, 1, 0, 0},{33224, 1, 1, 18}, {0, 0, 0, 0}}},{21258, {{33224, 1, 0, 0},{33224, 1, 1, 19}, {0, 0, 0, 0}}},{21258, {{33224, 1, 0, 0},{33224, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{21259, {{33225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21260, {{33226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21261, {{33227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21262, {{33228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21263, {{33229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21264, {{33230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21265, {{33231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21266, {{33232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21267, {{33233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21268, {{33234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21269, {{33235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21270, {{33236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21271, {{33237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21272, {{33238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21273, {{33239, 1, 0, 0},{33239, 1, 1, 17}, {0, 0, 0, 0}}},{21273, {{33239, 1, 0, 0},{33239, 1, 1, 17}, {0, 0, 0, 0}}},{21274, {{33239, 1, 0, 0},{33239, 1, 1, 18}, {0, 0, 0, 0}}},{21274, {{33239, 1, 0, 0},{33239, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21275, {{33240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21276, {{33241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21277, {{33242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21278, {{33243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{21279, {{33245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21281, {{33246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21282, {{33247, 1, 0, 0},{33247, 1, 1, 17}, {0, 0, 0, 0}}},{21282, {{33247, 1, 0, 0},{33247, 1, 1, 17}, {0, 0, 0, 0}}},{21283, {{33247, 1, 0, 0},{33247, 1, 1, 18}, {0, 0, 0, 0}}},{21283, {{33247, 1, 0, 0},{33247, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21284, {{33248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21285, {{33249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21286, {{33250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21287, {{33251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21288, {{33252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21289, {{33253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21290, {{33254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21291, {{33255, 1, 0, 0},{33255, 1, 1, 17}, {0, 0, 0, 0}}},{21291, {{33255, 1, 0, 0},{33255, 1, 1, 17}, {0, 0, 0, 0}}},{21292, {{33255, 1, 0, 0},{33255, 1, 1, 18}, {0, 0, 0, 0}}},{21292, {{33255, 1, 0, 0},{33255, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21294, {{33256, 1, 0, 0},{33256, 1, 1, 17}, {0, 0, 0, 0}}},{21294, {{33256, 1, 0, 0},{33256, 1, 1, 17}, {0, 0, 0, 0}}},{58409, {{33256, 1, 0, 0},{33256, 1, 1, 18}, {0, 0, 0, 0}}},{58409, {{33256, 1, 0, 0},{33256, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21295, {{33257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21296, {{33258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21297, {{33259, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21298, {{33260, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{21299, {{33261, 1, 0, 0},{33261, 1, 1, 18}, {0, 0, 0, 0}}},{21299, {{33261, 1, 0, 0},{33261, 1, 1, 18}, {0, 0, 0, 0}}},{30291, {{33261, 1, 0, 0},{33261, 1, 1, 0},{33261, 1, 1, 19},{64092, 1, 0, 0}, {0, 0, 0, 0}}},{30291, {{33261, 1, 0, 0},{33261, 1, 1, 0},{33261, 1, 1, 19},{64092, 1, 0, 0}, {0, 0, 0, 0}}},{30291, {{33261, 1, 0, 0},{33261, 1, 1, 0},{33261, 1, 1, 19},{64092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21300, {{33262, 1, 0, 0},{33262, 1, 1, 16}, {0, 0, 0, 0}}},{21300, {{33262, 1, 0, 0},{33262, 1, 1, 16}, {0, 0, 0, 0}}},{45921, {{33262, 1, 0, 0},{33262, 1, 1, 17}, {0, 0, 0, 0}}},{45921, {{33262, 1, 0, 0},{33262, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{21301, {{33263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21302, {{33264, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21303, {{33265, 1, 0, 0},{33265, 1, 1, 17}, {0, 0, 0, 0}}},{21303, {{33265, 1, 0, 0},{33265, 1, 1, 17}, {0, 0, 0, 0}}},{58139, {{33265, 1, 0, 0},{33265, 1, 1, 18}, {0, 0, 0, 0}}},{58139, {{33265, 1, 0, 0},{33265, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21304, {{33266, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21305, {{33267, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21306, {{33268, 1, 0, 0},{33268, 1, 1, 18}, {0, 0, 0, 0}}},{21306, {{33268, 1, 0, 0},{33268, 1, 1, 18}, {0, 0, 0, 0}}},{21307, {{33268, 1, 0, 0},{33268, 1, 1, 17}, {0, 0, 0, 0}}},{21307, {{33268, 1, 0, 0},{33268, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{21308, {{33269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21309, {{33270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21310, {{33271, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21311, {{33272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21312, {{33273, 1, 0, 0},{33273, 1, 1, 17}, {0, 0, 0, 0}}},{21312, {{33273, 1, 0, 0},{33273, 1, 1, 17}, {0, 0, 0, 0}}},{21313, {{33273, 1, 0, 0},{33273, 1, 1, 18}, {0, 0, 0, 0}}},{21313, {{33273, 1, 0, 0},{33273, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21314, {{33274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21315, {{33275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21316, {{33276, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21317, {{33277, 1, 0, 0},{33277, 1, 1, 17}, {0, 0, 0, 0}}},{21317, {{33277, 1, 0, 0},{33277, 1, 1, 17}, {0, 0, 0, 0}}},{21318, {{33277, 1, 0, 0},{33277, 1, 1, 18}, {0, 0, 0, 0}}},{21318, {{33277, 1, 0, 0},{33277, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21319, {{33278, 1, 0, 0},{33278, 1, 1, 18}, {0, 0, 0, 0}}},{21319, {{33278, 1, 0, 0},{33278, 1, 1, 18}, {0, 0, 0, 0}}},{21320, {{33278, 1, 0, 0},{33278, 1, 1, 17}, {0, 0, 0, 0}}},{21320, {{33278, 1, 0, 0},{33278, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{21321, {{33279, 1, 0, 0},{33279, 1, 1, 17}, {0, 0, 0, 0}}},{21321, {{33279, 1, 0, 0},{33279, 1, 1, 17}, {0, 0, 0, 0}}},{21322, {{33279, 1, 0, 0},{33279, 1, 1, 18}, {0, 0, 0, 0}}},{21322, {{33279, 1, 0, 0},{33279, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {8, {{21323, {{33280, 1, 0, 0},{33280, 1, 1, 18}, {0, 0, 0, 0}}},{21323, {{33280, 1, 0, 0},{33280, 1, 1, 18}, {0, 0, 0, 0}}},{21324, {{33280, 1, 0, 0},{33280, 1, 1, 19}, {0, 0, 0, 0}}},{21324, {{33280, 1, 0, 0},{33280, 1, 1, 19}, {0, 0, 0, 0}}},{21325, {{33280, 1, 0, 0},{33280, 1, 1, 20}, {0, 0, 0, 0}}},{21325, {{33280, 1, 0, 0},{33280, 1, 1, 20}, {0, 0, 0, 0}}},{58416, {{33280, 1, 0, 0},{33280, 1, 1, 21}, {0, 0, 0, 0}}},{58416, {{33280, 1, 0, 0},{33280, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {6, {{21326, {{33281, 1, 0, 0},{33281, 1, 1, 17}, {0, 0, 0, 0}}},{21326, {{33281, 1, 0, 0},{33281, 1, 1, 17}, {0, 0, 0, 0}}},{21327, {{33281, 1, 0, 0},{33281, 1, 1, 19}, {0, 0, 0, 0}}},{21327, {{33281, 1, 0, 0},{33281, 1, 1, 19}, {0, 0, 0, 0}}},{21328, {{33281, 1, 0, 0},{33281, 1, 1, 18}, {0, 0, 0, 0}}},{21328, {{33281, 1, 0, 0},{33281, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21329, {{33282, 1, 0, 0},{33282, 1, 1, 17}, {0, 0, 0, 0}}},{21329, {{33282, 1, 0, 0},{33282, 1, 1, 17}, {0, 0, 0, 0}}},{21330, {{33282, 1, 0, 0},{33282, 1, 1, 18}, {0, 0, 0, 0}}},{21330, {{33282, 1, 0, 0},{33282, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21331, {{33283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21332, {{33284, 1, 0, 0},{33284, 1, 1, 17}, {0, 0, 0, 0}}},{21332, {{33284, 1, 0, 0},{33284, 1, 1, 17}, {0, 0, 0, 0}}},{21333, {{33284, 1, 0, 0},{33284, 1, 1, 18}, {0, 0, 0, 0}}},{21333, {{33284, 1, 0, 0},{33284, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21334, {{33285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{21336, {{33287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21338, {{33288, 1, 0, 0},{33288, 1, 1, 17}, {0, 0, 0, 0}}},{21338, {{33288, 1, 0, 0},{33288, 1, 1, 17}, {0, 0, 0, 0}}},{58419, {{33288, 1, 0, 0},{33288, 1, 1, 18}, {0, 0, 0, 0}}},{58419, {{33288, 1, 0, 0},{33288, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21339, {{33289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21340, {{33290, 1, 0, 0},{33290, 1, 1, 17}, {0, 0, 0, 0}}},{21340, {{33290, 1, 0, 0},{33290, 1, 1, 17}, {0, 0, 0, 0}}},{21341, {{33290, 1, 0, 0},{33290, 1, 1, 18}, {0, 0, 0, 0}}},{21341, {{33290, 1, 0, 0},{33290, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21342, {{33291, 1, 0, 0},{33291, 1, 1, 17}, {0, 0, 0, 0}}},{21342, {{33291, 1, 0, 0},{33291, 1, 1, 17}, {0, 0, 0, 0}}},{58420, {{33291, 1, 0, 0},{33291, 1, 1, 18}, {0, 0, 0, 0}}},{58420, {{33291, 1, 0, 0},{33291, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21343, {{33292, 1, 0, 0},{33292, 1, 1, 18}, {0, 0, 0, 0}}},{21343, {{33292, 1, 0, 0},{33292, 1, 1, 18}, {0, 0, 0, 0}}},{21344, {{33292, 1, 0, 0},{33292, 1, 1, 19}, {0, 0, 0, 0}}},{21344, {{33292, 1, 0, 0},{33292, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{21345, {{33293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21346, {{33294, 1, 0, 0},{33294, 1, 1, 17}, {0, 0, 0, 0}}},{21346, {{33294, 1, 0, 0},{33294, 1, 1, 17}, {0, 0, 0, 0}}},{60095, {{33294, 1, 0, 0},{33294, 1, 1, 20}, {0, 0, 0, 0}}},{60095, {{33294, 1, 0, 0},{33294, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{21349, {{33295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21350, {{33296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21351, {{33297, 1, 0, 0},{33297, 1, 1, 16}, {0, 0, 0, 0}}},{21351, {{33297, 1, 0, 0},{33297, 1, 1, 16}, {0, 0, 0, 0}}},{21352, {{33297, 1, 0, 0},{33297, 1, 1, 17}, {0, 0, 0, 0}}},{21352, {{33297, 1, 0, 0},{33297, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{21353, {{33298, 1, 0, 0},{33298, 1, 1, 18}, {0, 0, 0, 0}}},{21353, {{33298, 1, 0, 0},{33298, 1, 1, 18}, {0, 0, 0, 0}}},{21354, {{33298, 1, 0, 0},{33298, 1, 1, 17}, {0, 0, 0, 0}}},{21354, {{33298, 1, 0, 0},{33298, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{21357, {{33299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21358, {{33300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21359, {{33301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21360, {{33302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21361, {{33303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {7, {{21355, {{33304, 1, 0, 0},{33304, 1, 1, 20}, {0, 0, 0, 0}}},{21355, {{33304, 1, 0, 0},{33304, 1, 1, 20}, {0, 0, 0, 0}}},{21362, {{33304, 1, 0, 0},{33304, 1, 1, 18}, {0, 0, 0, 0}}},{21362, {{33304, 1, 0, 0},{33304, 1, 1, 18}, {0, 0, 0, 0}}},{30310, {{33304, 1, 0, 0},{33304, 1, 1, 0},{33304, 1, 1, 19},{64109, 1, 0, 0}, {0, 0, 0, 0}}},{30310, {{33304, 1, 0, 0},{33304, 1, 1, 0},{33304, 1, 1, 19},{64109, 1, 0, 0}, {0, 0, 0, 0}}},{30310, {{33304, 1, 0, 0},{33304, 1, 1, 0},{33304, 1, 1, 19},{64109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21363, {{33305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21364, {{33306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{21365, {{33307, 1, 0, 0},{33307, 1, 1, 18}, {0, 0, 0, 0}}},{21365, {{33307, 1, 0, 0},{33307, 1, 1, 18}, {0, 0, 0, 0}}},{21366, {{33307, 1, 0, 0},{33307, 1, 1, 22}, {0, 0, 0, 0}}},{21366, {{33307, 1, 0, 0},{33307, 1, 1, 22}, {0, 0, 0, 0}}},{21367, {{33307, 1, 0, 0},{33307, 1, 1, 19}, {0, 0, 0, 0}}},{21367, {{33307, 1, 0, 0},{33307, 1, 1, 19}, {0, 0, 0, 0}}},{21368, {{33307, 1, 0, 0},{33307, 1, 1, 23}, {0, 0, 0, 0}}},{21368, {{33307, 1, 0, 0},{33307, 1, 1, 23}, {0, 0, 0, 0}}}}},
    {6, {{21369, {{33308, 1, 0, 0},{33308, 1, 1, 20}, {0, 0, 0, 0}}},{21369, {{33308, 1, 0, 0},{33308, 1, 1, 20}, {0, 0, 0, 0}}},{21370, {{33308, 1, 0, 0},{33308, 1, 1, 19}, {0, 0, 0, 0}}},{21370, {{33308, 1, 0, 0},{33308, 1, 1, 19}, {0, 0, 0, 0}}},{21371, {{33308, 1, 0, 0},{33308, 1, 1, 18}, {0, 0, 0, 0}}},{21371, {{33308, 1, 0, 0},{33308, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21372, {{33309, 1, 0, 0},{33309, 1, 1, 18}, {0, 0, 0, 0}}},{21372, {{33309, 1, 0, 0},{33309, 1, 1, 18}, {0, 0, 0, 0}}},{21373, {{33309, 1, 0, 0},{33309, 1, 1, 17}, {0, 0, 0, 0}}},{21373, {{33309, 1, 0, 0},{33309, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{21374, {{33310, 1, 0, 0},{33310, 1, 1, 18}, {0, 0, 0, 0}}},{21374, {{33310, 1, 0, 0},{33310, 1, 1, 18}, {0, 0, 0, 0}}},{21375, {{33310, 1, 0, 0},{33310, 1, 1, 17}, {0, 0, 0, 0}}},{21375, {{33310, 1, 0, 0},{33310, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{21376, {{33311, 1, 0, 0},{33311, 1, 1, 18}, {0, 0, 0, 0}}},{21376, {{33311, 1, 0, 0},{33311, 1, 1, 18}, {0, 0, 0, 0}}},{58427, {{33311, 1, 0, 0},{33311, 1, 1, 19}, {0, 0, 0, 0}}},{58427, {{33311, 1, 0, 0},{33311, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{21377, {{33312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21378, {{33313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21379, {{33314, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{21380, {{33317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21381, {{33318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21382, {{33319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21383, {{33320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21384, {{33321, 1, 0, 0},{33321, 1, 1, 17}, {0, 0, 0, 0}}},{21384, {{33321, 1, 0, 0},{33321, 1, 1, 17}, {0, 0, 0, 0}}},{21385, {{33321, 1, 0, 0},{33321, 1, 1, 18}, {0, 0, 0, 0}}},{21385, {{33321, 1, 0, 0},{33321, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21386, {{33322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21387, {{33323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21388, {{33324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21389, {{33325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21390, {{33326, 1, 0, 0},{33326, 1, 1, 18}, {0, 0, 0, 0}}},{21390, {{33326, 1, 0, 0},{33326, 1, 1, 18}, {0, 0, 0, 0}}},{21391, {{33326, 1, 0, 0},{33326, 1, 1, 19}, {0, 0, 0, 0}}},{21391, {{33326, 1, 0, 0},{33326, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{21392, {{33330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21393, {{33331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21394, {{33332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21395, {{33333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21396, {{33334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21397, {{33335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21398, {{33336, 1, 0, 0},{33336, 1, 1, 17}, {0, 0, 0, 0}}},{21398, {{33336, 1, 0, 0},{33336, 1, 1, 17}, {0, 0, 0, 0}}},{21399, {{33336, 1, 0, 0},{33336, 1, 1, 18}, {0, 0, 0, 0}}},{21399, {{33336, 1, 0, 0},{33336, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{21400, {{33337, 1, 0, 0},{33337, 1, 1, 19}, {0, 0, 0, 0}}},{21400, {{33337, 1, 0, 0},{33337, 1, 1, 19}, {0, 0, 0, 0}}},{21401, {{33337, 1, 0, 0},{33337, 1, 1, 20}, {0, 0, 0, 0}}},{21401, {{33337, 1, 0, 0},{33337, 1, 1, 20}, {0, 0, 0, 0}}},{21402, {{33337, 1, 0, 0},{33337, 1, 1, 21}, {0, 0, 0, 0}}},{21402, {{33337, 1, 0, 0},{33337, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{21404, {{33338, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{21405, {{33340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21406, {{33341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{21407, {{33344, 1, 0, 0},{33344, 1, 1, 18}, {0, 0, 0, 0}}},{21407, {{33344, 1, 0, 0},{33344, 1, 1, 18}, {0, 0, 0, 0}}},{21408, {{33344, 1, 0, 0},{33344, 1, 1, 19}, {0, 0, 0, 0}}},{21408, {{33344, 1, 0, 0},{33344, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{21409, {{33345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21410, {{33346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21411, {{33347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21412, {{33348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21413, {{33349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21414, {{33350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21415, {{33351, 1, 0, 0},{33351, 1, 1, 18}, {0, 0, 0, 0}}},{21415, {{33351, 1, 0, 0},{33351, 1, 1, 18}, {0, 0, 0, 0}}},{21416, {{33351, 1, 0, 0},{33351, 1, 1, 19}, {0, 0, 0, 0}}},{21416, {{33351, 1, 0, 0},{33351, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{21417, {{33353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21418, {{33354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21419, {{33355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21420, {{33356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{21421, {{33358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21422, {{33359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21423, {{33360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21424, {{33361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21425, {{33362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21426, {{33363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{21427, {{33365, 1, 0, 0},{33365, 1, 1, 16}, {0, 0, 0, 0}}},{21427, {{33365, 1, 0, 0},{33365, 1, 1, 16}, {0, 0, 0, 0}}},{21428, {{33365, 1, 0, 0},{33365, 1, 1, 17}, {0, 0, 0, 0}}},{21428, {{33365, 1, 0, 0},{33365, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{21429, {{33366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21430, {{33367, 1, 0, 0},{33367, 1, 1, 17}, {0, 0, 0, 0}}},{21430, {{33367, 1, 0, 0},{33367, 1, 1, 17}, {0, 0, 0, 0}}},{21431, {{33367, 1, 0, 0},{33367, 1, 1, 18}, {0, 0, 0, 0}}},{21431, {{33367, 1, 0, 0},{33367, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{21433, {{33368, 1, 0, 0},{33368, 1, 1, 20}, {0, 0, 0, 0}}},{21433, {{33368, 1, 0, 0},{33368, 1, 1, 20}, {0, 0, 0, 0}}},{21434, {{33368, 1, 0, 0},{33368, 1, 1, 18}, {0, 0, 0, 0}}},{21434, {{33368, 1, 0, 0},{33368, 1, 1, 18}, {0, 0, 0, 0}}},{21435, {{33368, 1, 0, 0},{33368, 1, 1, 19}, {0, 0, 0, 0}}},{21435, {{33368, 1, 0, 0},{33368, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{21436, {{33369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21437, {{33370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21438, {{33371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21439, {{33372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21440, {{33373, 1, 0, 0},{33373, 1, 1, 17}, {0, 0, 0, 0}}},{21440, {{33373, 1, 0, 0},{33373, 1, 1, 17}, {0, 0, 0, 0}}},{21441, {{33373, 1, 0, 0},{33373, 1, 1, 18}, {0, 0, 0, 0}}},{21441, {{33373, 1, 0, 0},{33373, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21442, {{33374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21443, {{33375, 1, 0, 0},{33375, 1, 1, 17}, {0, 0, 0, 0}}},{21443, {{33375, 1, 0, 0},{33375, 1, 1, 17}, {0, 0, 0, 0}}},{21444, {{33375, 1, 0, 0},{33375, 1, 1, 18}, {0, 0, 0, 0}}},{21444, {{33375, 1, 0, 0},{33375, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21445, {{33376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21446, {{33377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21447, {{33378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21448, {{33379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21449, {{33380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21450, {{33381, 1, 0, 0},{33381, 1, 1, 16}, {0, 0, 0, 0}}},{21450, {{33381, 1, 0, 0},{33381, 1, 1, 16}, {0, 0, 0, 0}}},{21451, {{33381, 1, 0, 0},{33381, 1, 1, 17}, {0, 0, 0, 0}}},{21451, {{33381, 1, 0, 0},{33381, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{21452, {{33382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21454, {{33383, 1, 0, 0},{33383, 1, 1, 17}, {0, 0, 0, 0}}},{21454, {{33383, 1, 0, 0},{33383, 1, 1, 17}, {0, 0, 0, 0}}},{21455, {{33383, 1, 0, 0},{33383, 1, 1, 18}, {0, 0, 0, 0}}},{21455, {{33383, 1, 0, 0},{33383, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{21456, {{33384, 1, 0, 0},{33384, 1, 1, 17}, {0, 0, 0, 0}}},{21456, {{33384, 1, 0, 0},{33384, 1, 1, 17}, {0, 0, 0, 0}}},{21457, {{33384, 1, 0, 0},{33384, 1, 1, 18}, {0, 0, 0, 0}}},{21457, {{33384, 1, 0, 0},{33384, 1, 1, 18}, {0, 0, 0, 0}}},{21458, {{33384, 1, 0, 0},{33384, 1, 1, 19}, {0, 0, 0, 0}}},{21458, {{33384, 1, 0, 0},{33384, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{21459, {{33385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21460, {{33386, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21461, {{33387, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21462, {{33388, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21463, {{33389, 1, 0, 0},{33389, 1, 1, 17}, {0, 0, 0, 0}}},{21463, {{33389, 1, 0, 0},{33389, 1, 1, 17}, {0, 0, 0, 0}}},{21464, {{33389, 1, 0, 0},{33389, 1, 1, 18}, {0, 0, 0, 0}}},{21464, {{33389, 1, 0, 0},{33389, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21465, {{33390, 1, 0, 0},{33390, 1, 1, 17}, {0, 0, 0, 0}}},{21465, {{33390, 1, 0, 0},{33390, 1, 1, 17}, {0, 0, 0, 0}}},{21466, {{33390, 1, 0, 0},{33390, 1, 1, 18}, {0, 0, 0, 0}}},{21466, {{33390, 1, 0, 0},{33390, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21467, {{33391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{21468, {{33393, 1, 0, 0},{33393, 1, 1, 17}, {0, 0, 0, 0}}},{21468, {{33393, 1, 0, 0},{33393, 1, 1, 17}, {0, 0, 0, 0}}},{21469, {{33393, 1, 0, 0},{33393, 1, 1, 18}, {0, 0, 0, 0}}},{21469, {{33393, 1, 0, 0},{33393, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21470, {{33394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{21475, {{33396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21476, {{33397, 1, 0, 0},{33397, 1, 1, 17}, {0, 0, 0, 0}}},{21476, {{33397, 1, 0, 0},{33397, 1, 1, 17}, {0, 0, 0, 0}}},{21477, {{33397, 1, 0, 0},{33397, 1, 1, 16}, {0, 0, 0, 0}}},{21477, {{33397, 1, 0, 0},{33397, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{21478, {{33398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21479, {{33399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21480, {{33400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{21481, {{33401, 1, 0, 0},{33401, 1, 1, 21}, {0, 0, 0, 0}}},{21481, {{33401, 1, 0, 0},{33401, 1, 1, 21}, {0, 0, 0, 0}}},{30292, {{33401, 1, 0, 0},{33401, 1, 1, 0},{33401, 1, 1, 19},{64093, 1, 0, 0}, {0, 0, 0, 0}}},{30292, {{33401, 1, 0, 0},{33401, 1, 1, 0},{33401, 1, 1, 19},{64093, 1, 0, 0}, {0, 0, 0, 0}}},{30292, {{33401, 1, 0, 0},{33401, 1, 1, 0},{33401, 1, 1, 19},{64093, 1, 0, 0}, {0, 0, 0, 0}}},{30293, {{33401, 1, 0, 0},{33401, 1, 1, 1},{33401, 1, 1, 20},{64094, 1, 0, 0}, {0, 0, 0, 0}}},{30293, {{33401, 1, 0, 0},{33401, 1, 1, 1},{33401, 1, 1, 20},{64094, 1, 0, 0}, {0, 0, 0, 0}}},{30293, {{33401, 1, 0, 0},{33401, 1, 1, 1},{33401, 1, 1, 20},{64094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{21483, {{33403, 1, 0, 0},{33403, 1, 1, 17}, {0, 0, 0, 0}}},{21483, {{33403, 1, 0, 0},{33403, 1, 1, 17}, {0, 0, 0, 0}}},{21484, {{33403, 1, 0, 0},{33403, 1, 1, 18}, {0, 0, 0, 0}}},{21484, {{33403, 1, 0, 0},{33403, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21485, {{33404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21487, {{33405, 1, 0, 0},{33405, 1, 1, 17}, {0, 0, 0, 0}}},{21487, {{33405, 1, 0, 0},{33405, 1, 1, 17}, {0, 0, 0, 0}}},{21488, {{33405, 1, 0, 0},{33405, 1, 1, 18}, {0, 0, 0, 0}}},{21488, {{33405, 1, 0, 0},{33405, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{21489, {{33406, 1, 0, 0},{33406, 1, 1, 19}, {0, 0, 0, 0}}},{21489, {{33406, 1, 0, 0},{33406, 1, 1, 19}, {0, 0, 0, 0}}},{21490, {{33406, 1, 0, 0},{33406, 1, 1, 18}, {0, 0, 0, 0}}},{21490, {{33406, 1, 0, 0},{33406, 1, 1, 18}, {0, 0, 0, 0}}},{21491, {{33406, 1, 0, 0},{33406, 1, 1, 20}, {0, 0, 0, 0}}},{21491, {{33406, 1, 0, 0},{33406, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{21492, {{33407, 1, 0, 0},{33407, 1, 1, 17}, {0, 0, 0, 0}}},{21492, {{33407, 1, 0, 0},{33407, 1, 1, 17}, {0, 0, 0, 0}}},{21493, {{33407, 1, 0, 0},{33407, 1, 1, 18}, {0, 0, 0, 0}}},{21493, {{33407, 1, 0, 0},{33407, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21494, {{33408, 1, 0, 0},{33408, 1, 1, 17}, {0, 0, 0, 0}}},{21494, {{33408, 1, 0, 0},{33408, 1, 1, 17}, {0, 0, 0, 0}}},{21495, {{33408, 1, 0, 0},{33408, 1, 1, 18}, {0, 0, 0, 0}}},{21495, {{33408, 1, 0, 0},{33408, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21496, {{33409, 1, 0, 0},{33409, 1, 1, 17}, {0, 0, 0, 0}}},{21496, {{33409, 1, 0, 0},{33409, 1, 1, 17}, {0, 0, 0, 0}}},{21497, {{33409, 1, 0, 0},{33409, 1, 1, 18}, {0, 0, 0, 0}}},{21497, {{33409, 1, 0, 0},{33409, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{21498, {{33411, 1, 0, 0},{33411, 1, 1, 17}, {0, 0, 0, 0}}},{21498, {{33411, 1, 0, 0},{33411, 1, 1, 17}, {0, 0, 0, 0}}},{21499, {{33411, 1, 0, 0},{33411, 1, 1, 18}, {0, 0, 0, 0}}},{21499, {{33411, 1, 0, 0},{33411, 1, 1, 18}, {0, 0, 0, 0}}},{21500, {{33411, 1, 0, 0},{33411, 1, 1, 19}, {0, 0, 0, 0}}},{21500, {{33411, 1, 0, 0},{33411, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{21501, {{33412, 1, 0, 0},{33412, 1, 1, 17}, {0, 0, 0, 0}}},{21501, {{33412, 1, 0, 0},{33412, 1, 1, 17}, {0, 0, 0, 0}}},{21502, {{33412, 1, 0, 0},{33412, 1, 1, 18}, {0, 0, 0, 0}}},{21502, {{33412, 1, 0, 0},{33412, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21503, {{33413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21504, {{33414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21505, {{33415, 1, 0, 0},{33415, 1, 1, 17}, {0, 0, 0, 0}}},{21505, {{33415, 1, 0, 0},{33415, 1, 1, 17}, {0, 0, 0, 0}}},{21506, {{33415, 1, 0, 0},{33415, 1, 1, 18}, {0, 0, 0, 0}}},{21506, {{33415, 1, 0, 0},{33415, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{21507, {{33417, 1, 0, 0},{33417, 1, 1, 17}, {0, 0, 0, 0}}},{21507, {{33417, 1, 0, 0},{33417, 1, 1, 17}, {0, 0, 0, 0}}},{21508, {{33417, 1, 0, 0},{33417, 1, 1, 18}, {0, 0, 0, 0}}},{21508, {{33417, 1, 0, 0},{33417, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21509, {{33418, 1, 0, 0},{33418, 1, 1, 17}, {0, 0, 0, 0}}},{21509, {{33418, 1, 0, 0},{33418, 1, 1, 17}, {0, 0, 0, 0}}},{21510, {{33418, 1, 0, 0},{33418, 1, 1, 18}, {0, 0, 0, 0}}},{21510, {{33418, 1, 0, 0},{33418, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21511, {{33419, 1, 0, 0},{33419, 1, 1, 17}, {0, 0, 0, 0}}},{21511, {{33419, 1, 0, 0},{33419, 1, 1, 17}, {0, 0, 0, 0}}},{21512, {{33419, 1, 0, 0},{33419, 1, 1, 18}, {0, 0, 0, 0}}},{21512, {{33419, 1, 0, 0},{33419, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21513, {{33420, 1, 0, 0},{33420, 1, 1, 16}, {0, 0, 0, 0}}},{21513, {{33420, 1, 0, 0},{33420, 1, 1, 16}, {0, 0, 0, 0}}},{21514, {{33420, 1, 0, 0},{33420, 1, 1, 17}, {0, 0, 0, 0}}},{21514, {{33420, 1, 0, 0},{33420, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {6, {{21515, {{33421, 1, 0, 0},{33421, 1, 1, 18}, {0, 0, 0, 0}}},{21515, {{33421, 1, 0, 0},{33421, 1, 1, 18}, {0, 0, 0, 0}}},{21516, {{33421, 1, 0, 0},{33421, 1, 1, 19}, {0, 0, 0, 0}}},{21516, {{33421, 1, 0, 0},{33421, 1, 1, 19}, {0, 0, 0, 0}}},{21517, {{33421, 1, 0, 0},{33421, 1, 1, 20}, {0, 0, 0, 0}}},{21517, {{33421, 1, 0, 0},{33421, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{21518, {{33422, 1, 0, 0},{33422, 1, 1, 17}, {0, 0, 0, 0}}},{21518, {{33422, 1, 0, 0},{33422, 1, 1, 17}, {0, 0, 0, 0}}},{21519, {{33422, 1, 0, 0},{33422, 1, 1, 18}, {0, 0, 0, 0}}},{21519, {{33422, 1, 0, 0},{33422, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21520, {{33423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21521, {{33424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21522, {{33425, 1, 0, 0},{33425, 1, 1, 17}, {0, 0, 0, 0}}},{21522, {{33425, 1, 0, 0},{33425, 1, 1, 17}, {0, 0, 0, 0}}},{21524, {{33425, 1, 0, 0},{33425, 1, 1, 18}, {0, 0, 0, 0}}},{21524, {{33425, 1, 0, 0},{33425, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{21525, {{33426, 1, 0, 0},{33426, 1, 1, 18}, {0, 0, 0, 0}}},{21525, {{33426, 1, 0, 0},{33426, 1, 1, 18}, {0, 0, 0, 0}}},{21526, {{33426, 1, 0, 0},{33426, 1, 1, 20}, {0, 0, 0, 0}}},{21526, {{33426, 1, 0, 0},{33426, 1, 1, 20}, {0, 0, 0, 0}}},{21527, {{33426, 1, 0, 0},{33426, 1, 1, 19}, {0, 0, 0, 0}}},{21527, {{33426, 1, 0, 0},{33426, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{21528, {{33427, 1, 0, 0},{33427, 1, 1, 17}, {0, 0, 0, 0}}},{21528, {{33427, 1, 0, 0},{33427, 1, 1, 17}, {0, 0, 0, 0}}},{21529, {{33427, 1, 0, 0},{33427, 1, 1, 18}, {0, 0, 0, 0}}},{21529, {{33427, 1, 0, 0},{33427, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21530, {{33428, 1, 0, 0},{33428, 1, 1, 17}, {0, 0, 0, 0}}},{21530, {{33428, 1, 0, 0},{33428, 1, 1, 17}, {0, 0, 0, 0}}},{21531, {{33428, 1, 0, 0},{33428, 1, 1, 18}, {0, 0, 0, 0}}},{21531, {{33428, 1, 0, 0},{33428, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21532, {{33429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21533, {{33430, 1, 0, 0},{33430, 1, 1, 17}, {0, 0, 0, 0}}},{21533, {{33430, 1, 0, 0},{33430, 1, 1, 17}, {0, 0, 0, 0}}},{21534, {{33430, 1, 0, 0},{33430, 1, 1, 18}, {0, 0, 0, 0}}},{21534, {{33430, 1, 0, 0},{33430, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{21535, {{33432, 1, 0, 0},{33432, 1, 1, 17}, {0, 0, 0, 0}}},{21535, {{33432, 1, 0, 0},{33432, 1, 1, 17}, {0, 0, 0, 0}}},{21536, {{33432, 1, 0, 0},{33432, 1, 1, 18}, {0, 0, 0, 0}}},{21536, {{33432, 1, 0, 0},{33432, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21537, {{33433, 1, 0, 0},{33433, 1, 1, 17}, {0, 0, 0, 0}}},{21537, {{33433, 1, 0, 0},{33433, 1, 1, 17}, {0, 0, 0, 0}}},{21538, {{33433, 1, 0, 0},{33433, 1, 1, 18}, {0, 0, 0, 0}}},{21538, {{33433, 1, 0, 0},{33433, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21539, {{33434, 1, 0, 0},{33434, 1, 1, 18}, {0, 0, 0, 0}}},{21539, {{33434, 1, 0, 0},{33434, 1, 1, 18}, {0, 0, 0, 0}}},{21540, {{33434, 1, 0, 0},{33434, 1, 1, 17}, {0, 0, 0, 0}}},{21540, {{33434, 1, 0, 0},{33434, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{21541, {{33435, 1, 0, 0},{33435, 1, 1, 17}, {0, 0, 0, 0}}},{21541, {{33435, 1, 0, 0},{33435, 1, 1, 17}, {0, 0, 0, 0}}},{21542, {{33435, 1, 0, 0},{33435, 1, 1, 18}, {0, 0, 0, 0}}},{21542, {{33435, 1, 0, 0},{33435, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21543, {{33436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21544, {{33437, 1, 0, 0},{33437, 1, 1, 17}, {0, 0, 0, 0}}},{21544, {{33437, 1, 0, 0},{33437, 1, 1, 17}, {0, 0, 0, 0}}},{21545, {{33437, 1, 0, 0},{33437, 1, 1, 18}, {0, 0, 0, 0}}},{21545, {{33437, 1, 0, 0},{33437, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21546, {{33438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21547, {{33439, 1, 0, 0},{33439, 1, 1, 17}, {0, 0, 0, 0}}},{21547, {{33439, 1, 0, 0},{33439, 1, 1, 17}, {0, 0, 0, 0}}},{21548, {{33439, 1, 0, 0},{33439, 1, 1, 18}, {0, 0, 0, 0}}},{21548, {{33439, 1, 0, 0},{33439, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21549, {{33440, 1, 0, 0},{33440, 1, 1, 18}, {0, 0, 0, 0}}},{21549, {{33440, 1, 0, 0},{33440, 1, 1, 18}, {0, 0, 0, 0}}},{21550, {{33440, 1, 0, 0},{33440, 1, 1, 17}, {0, 0, 0, 0}}},{21550, {{33440, 1, 0, 0},{33440, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{21551, {{33441, 1, 0, 0},{33441, 1, 1, 17}, {0, 0, 0, 0}}},{21551, {{33441, 1, 0, 0},{33441, 1, 1, 17}, {0, 0, 0, 0}}},{21552, {{33441, 1, 0, 0},{33441, 1, 1, 18}, {0, 0, 0, 0}}},{21552, {{33441, 1, 0, 0},{33441, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21553, {{33442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21555, {{33443, 1, 0, 0},{33443, 1, 1, 17}, {0, 0, 0, 0}}},{21555, {{33443, 1, 0, 0},{33443, 1, 1, 17}, {0, 0, 0, 0}}},{21556, {{33443, 1, 0, 0},{33443, 1, 1, 18}, {0, 0, 0, 0}}},{21556, {{33443, 1, 0, 0},{33443, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21557, {{33444, 1, 0, 0},{33444, 1, 1, 17}, {0, 0, 0, 0}}},{21557, {{33444, 1, 0, 0},{33444, 1, 1, 17}, {0, 0, 0, 0}}},{21558, {{33444, 1, 0, 0},{33444, 1, 1, 18}, {0, 0, 0, 0}}},{21558, {{33444, 1, 0, 0},{33444, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21559, {{33445, 1, 0, 0},{33445, 1, 1, 17}, {0, 0, 0, 0}}},{21559, {{33445, 1, 0, 0},{33445, 1, 1, 17}, {0, 0, 0, 0}}},{21560, {{33445, 1, 0, 0},{33445, 1, 1, 18}, {0, 0, 0, 0}}},{21560, {{33445, 1, 0, 0},{33445, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {16, {{21561, {{33446, 1, 0, 0},{33446, 1, 1, 18}, {0, 0, 0, 0}}},{21561, {{33446, 1, 0, 0},{33446, 1, 1, 18}, {0, 0, 0, 0}}},{21562, {{33446, 1, 0, 0},{33446, 1, 1, 19}, {0, 0, 0, 0}}},{21562, {{33446, 1, 0, 0},{33446, 1, 1, 19}, {0, 0, 0, 0}}},{21563, {{33446, 1, 0, 0},{33446, 1, 1, 25}, {0, 0, 0, 0}}},{21563, {{33446, 1, 0, 0},{33446, 1, 1, 25}, {0, 0, 0, 0}}},{21564, {{33446, 1, 0, 0},{33446, 1, 1, 24}, {0, 0, 0, 0}}},{21564, {{33446, 1, 0, 0},{33446, 1, 1, 24}, {0, 0, 0, 0}}},{21565, {{33446, 1, 0, 0},{33446, 1, 1, 20}, {0, 0, 0, 0}}},{21565, {{33446, 1, 0, 0},{33446, 1, 1, 20}, {0, 0, 0, 0}}},{21566, {{33446, 1, 0, 0},{33446, 1, 1, 21}, {0, 0, 0, 0}}},{21566, {{33446, 1, 0, 0},{33446, 1, 1, 21}, {0, 0, 0, 0}}},{21567, {{33446, 1, 0, 0},{33446, 1, 1, 23}, {0, 0, 0, 0}}},{21567, {{33446, 1, 0, 0},{33446, 1, 1, 23}, {0, 0, 0, 0}}},{21568, {{33446, 1, 0, 0},{33446, 1, 1, 22}, {0, 0, 0, 0}}},{21568, {{33446, 1, 0, 0},{33446, 1, 1, 22}, {0, 0, 0, 0}}}}},
    {4, {{21569, {{33447, 1, 0, 0},{33447, 1, 1, 17}, {0, 0, 0, 0}}},{21569, {{33447, 1, 0, 0},{33447, 1, 1, 17}, {0, 0, 0, 0}}},{21570, {{33447, 1, 0, 0},{33447, 1, 1, 18}, {0, 0, 0, 0}}},{21570, {{33447, 1, 0, 0},{33447, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21571, {{33448, 1, 0, 0},{33448, 1, 1, 18}, {0, 0, 0, 0}}},{21571, {{33448, 1, 0, 0},{33448, 1, 1, 18}, {0, 0, 0, 0}}},{21572, {{33448, 1, 0, 0},{33448, 1, 1, 17}, {0, 0, 0, 0}}},{21572, {{33448, 1, 0, 0},{33448, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{21573, {{33449, 1, 0, 0},{33449, 1, 1, 17}, {0, 0, 0, 0}}},{21573, {{33449, 1, 0, 0},{33449, 1, 1, 17}, {0, 0, 0, 0}}},{21574, {{33449, 1, 0, 0},{33449, 1, 1, 18}, {0, 0, 0, 0}}},{21574, {{33449, 1, 0, 0},{33449, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21575, {{33450, 1, 0, 0},{33450, 1, 1, 17}, {0, 0, 0, 0}}},{21575, {{33450, 1, 0, 0},{33450, 1, 1, 17}, {0, 0, 0, 0}}},{21576, {{33450, 1, 0, 0},{33450, 1, 1, 18}, {0, 0, 0, 0}}},{21576, {{33450, 1, 0, 0},{33450, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21577, {{33451, 1, 0, 0},{33451, 1, 1, 17}, {0, 0, 0, 0}}},{21577, {{33451, 1, 0, 0},{33451, 1, 1, 17}, {0, 0, 0, 0}}},{21578, {{33451, 1, 0, 0},{33451, 1, 1, 18}, {0, 0, 0, 0}}},{21578, {{33451, 1, 0, 0},{33451, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{21579, {{33452, 1, 0, 0},{33452, 1, 1, 17}, {0, 0, 0, 0}}},{21579, {{33452, 1, 0, 0},{33452, 1, 1, 17}, {0, 0, 0, 0}}},{21580, {{33452, 1, 0, 0},{33452, 1, 1, 18}, {0, 0, 0, 0}}},{21580, {{33452, 1, 0, 0},{33452, 1, 1, 18}, {0, 0, 0, 0}}},{21581, {{33452, 1, 0, 0},{33452, 1, 1, 19}, {0, 0, 0, 0}}},{21581, {{33452, 1, 0, 0},{33452, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{21582, {{33453, 1, 0, 0},{33453, 1, 1, 17}, {0, 0, 0, 0}}},{21582, {{33453, 1, 0, 0},{33453, 1, 1, 17}, {0, 0, 0, 0}}},{21583, {{33453, 1, 0, 0},{33453, 1, 1, 18}, {0, 0, 0, 0}}},{21583, {{33453, 1, 0, 0},{33453, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{21584, {{33454, 1, 0, 0},{33454, 1, 1, 19}, {0, 0, 0, 0}}},{21584, {{33454, 1, 0, 0},{33454, 1, 1, 19}, {0, 0, 0, 0}}},{21585, {{33454, 1, 0, 0},{33454, 1, 1, 17}, {0, 0, 0, 0}}},{21585, {{33454, 1, 0, 0},{33454, 1, 1, 17}, {0, 0, 0, 0}}},{21586, {{33454, 1, 0, 0},{33454, 1, 1, 18}, {0, 0, 0, 0}}},{21586, {{33454, 1, 0, 0},{33454, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21587, {{33455, 1, 0, 0},{33455, 1, 1, 17}, {0, 0, 0, 0}}},{21587, {{33455, 1, 0, 0},{33455, 1, 1, 17}, {0, 0, 0, 0}}},{21588, {{33455, 1, 0, 0},{33455, 1, 1, 18}, {0, 0, 0, 0}}},{21588, {{33455, 1, 0, 0},{33455, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21589, {{33456, 1, 0, 0},{33456, 1, 1, 18}, {0, 0, 0, 0}}},{21589, {{33456, 1, 0, 0},{33456, 1, 1, 18}, {0, 0, 0, 0}}},{21590, {{33456, 1, 0, 0},{33456, 1, 1, 17}, {0, 0, 0, 0}}},{21590, {{33456, 1, 0, 0},{33456, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {8, {{21591, {{33457, 1, 0, 0},{33457, 1, 1, 18}, {0, 0, 0, 0}}},{21591, {{33457, 1, 0, 0},{33457, 1, 1, 18}, {0, 0, 0, 0}}},{21592, {{33457, 1, 0, 0},{33457, 1, 1, 19}, {0, 0, 0, 0}}},{21592, {{33457, 1, 0, 0},{33457, 1, 1, 19}, {0, 0, 0, 0}}},{21593, {{33457, 1, 0, 0},{33457, 1, 1, 20}, {0, 0, 0, 0}}},{21593, {{33457, 1, 0, 0},{33457, 1, 1, 20}, {0, 0, 0, 0}}},{21594, {{33457, 1, 0, 0},{33457, 1, 1, 22}, {0, 0, 0, 0}}},{21594, {{33457, 1, 0, 0},{33457, 1, 1, 22}, {0, 0, 0, 0}}}}},
    {4, {{21596, {{33458, 1, 0, 0},{33458, 1, 1, 18}, {0, 0, 0, 0}}},{21596, {{33458, 1, 0, 0},{33458, 1, 1, 18}, {0, 0, 0, 0}}},{21597, {{33458, 1, 0, 0},{33458, 1, 1, 17}, {0, 0, 0, 0}}},{21597, {{33458, 1, 0, 0},{33458, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{21598, {{33459, 1, 0, 0},{33459, 1, 1, 17}, {0, 0, 0, 0}}},{21598, {{33459, 1, 0, 0},{33459, 1, 1, 17}, {0, 0, 0, 0}}},{21599, {{33459, 1, 0, 0},{33459, 1, 1, 18}, {0, 0, 0, 0}}},{21599, {{33459, 1, 0, 0},{33459, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21600, {{33460, 1, 0, 0},{33460, 1, 1, 17}, {0, 0, 0, 0}}},{21600, {{33460, 1, 0, 0},{33460, 1, 1, 17}, {0, 0, 0, 0}}},{21601, {{33460, 1, 0, 0},{33460, 1, 1, 18}, {0, 0, 0, 0}}},{21601, {{33460, 1, 0, 0},{33460, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21602, {{33461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21603, {{33462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21604, {{33463, 1, 0, 0},{33463, 1, 1, 17}, {0, 0, 0, 0}}},{21604, {{33463, 1, 0, 0},{33463, 1, 1, 17}, {0, 0, 0, 0}}},{21605, {{33463, 1, 0, 0},{33463, 1, 1, 18}, {0, 0, 0, 0}}},{21605, {{33463, 1, 0, 0},{33463, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21606, {{33464, 1, 0, 0},{33464, 1, 1, 17}, {0, 0, 0, 0}}},{21606, {{33464, 1, 0, 0},{33464, 1, 1, 17}, {0, 0, 0, 0}}},{21607, {{33464, 1, 0, 0},{33464, 1, 1, 18}, {0, 0, 0, 0}}},{21607, {{33464, 1, 0, 0},{33464, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21608, {{33465, 1, 0, 0},{33465, 1, 1, 17}, {0, 0, 0, 0}}},{21608, {{33465, 1, 0, 0},{33465, 1, 1, 17}, {0, 0, 0, 0}}},{21609, {{33465, 1, 0, 0},{33465, 1, 1, 18}, {0, 0, 0, 0}}},{21609, {{33465, 1, 0, 0},{33465, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21610, {{33466, 1, 0, 0},{33466, 1, 1, 17}, {0, 0, 0, 0}}},{21610, {{33466, 1, 0, 0},{33466, 1, 1, 17}, {0, 0, 0, 0}}},{21611, {{33466, 1, 0, 0},{33466, 1, 1, 18}, {0, 0, 0, 0}}},{21611, {{33466, 1, 0, 0},{33466, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21612, {{33467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21613, {{33468, 1, 0, 0},{33468, 1, 1, 17}, {0, 0, 0, 0}}},{21613, {{33468, 1, 0, 0},{33468, 1, 1, 17}, {0, 0, 0, 0}}},{21614, {{33468, 1, 0, 0},{33468, 1, 1, 18}, {0, 0, 0, 0}}},{21614, {{33468, 1, 0, 0},{33468, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{21615, {{33469, 1, 0, 0},{33469, 1, 1, 20}, {0, 0, 0, 0}}},{21615, {{33469, 1, 0, 0},{33469, 1, 1, 20}, {0, 0, 0, 0}}},{21616, {{33469, 1, 0, 0},{33469, 1, 1, 19}, {0, 0, 0, 0}}},{21616, {{33469, 1, 0, 0},{33469, 1, 1, 19}, {0, 0, 0, 0}}},{21617, {{33469, 1, 0, 0},{33469, 1, 1, 22}, {0, 0, 0, 0}}},{21617, {{33469, 1, 0, 0},{33469, 1, 1, 22}, {0, 0, 0, 0}}}}},
    {4, {{21618, {{33470, 1, 0, 0},{33470, 1, 1, 17}, {0, 0, 0, 0}}},{21618, {{33470, 1, 0, 0},{33470, 1, 1, 17}, {0, 0, 0, 0}}},{21619, {{33470, 1, 0, 0},{33470, 1, 1, 18}, {0, 0, 0, 0}}},{21619, {{33470, 1, 0, 0},{33470, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21620, {{33471, 1, 0, 0},{33471, 1, 1, 17}, {0, 0, 0, 0}}},{21620, {{33471, 1, 0, 0},{33471, 1, 1, 17}, {0, 0, 0, 0}}},{21621, {{33471, 1, 0, 0},{33471, 1, 1, 18}, {0, 0, 0, 0}}},{21621, {{33471, 1, 0, 0},{33471, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21622, {{33472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21623, {{33473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21624, {{33474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21625, {{33475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21626, {{33476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21627, {{33477, 1, 0, 0},{33477, 1, 1, 18}, {0, 0, 0, 0}}},{21627, {{33477, 1, 0, 0},{33477, 1, 1, 18}, {0, 0, 0, 0}}},{21628, {{33477, 1, 0, 0},{33477, 1, 1, 19}, {0, 0, 0, 0}}},{21628, {{33477, 1, 0, 0},{33477, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{21629, {{33478, 1, 0, 0},{33478, 1, 1, 17}, {0, 0, 0, 0}}},{21629, {{33478, 1, 0, 0},{33478, 1, 1, 17}, {0, 0, 0, 0}}},{21630, {{33478, 1, 0, 0},{33478, 1, 1, 18}, {0, 0, 0, 0}}},{21630, {{33478, 1, 0, 0},{33478, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{21632, {{33488, 1, 0, 0},{33488, 1, 1, 17}, {0, 0, 0, 0}}},{21632, {{33488, 1, 0, 0},{33488, 1, 1, 17}, {0, 0, 0, 0}}},{21633, {{33488, 1, 0, 0},{33488, 1, 1, 18}, {0, 0, 0, 0}}},{21633, {{33488, 1, 0, 0},{33488, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{21634, {{33489, 1, 0, 0},{33489, 1, 1, 17}, {0, 0, 0, 0}}},{21634, {{33489, 1, 0, 0},{33489, 1, 1, 17}, {0, 0, 0, 0}}},{21635, {{33489, 1, 0, 0},{33489, 1, 1, 18}, {0, 0, 0, 0}}},{21635, {{33489, 1, 0, 0},{33489, 1, 1, 18}, {0, 0, 0, 0}}},{21636, {{33489, 1, 0, 0},{33489, 1, 1, 19}, {0, 0, 0, 0}}},{21636, {{33489, 1, 0, 0},{33489, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {6, {{21637, {{33490, 1, 0, 0},{33490, 1, 1, 18}, {0, 0, 0, 0}}},{21637, {{33490, 1, 0, 0},{33490, 1, 1, 18}, {0, 0, 0, 0}}},{21638, {{33490, 1, 0, 0},{33490, 1, 1, 20}, {0, 0, 0, 0}}},{21638, {{33490, 1, 0, 0},{33490, 1, 1, 20}, {0, 0, 0, 0}}},{21639, {{33490, 1, 0, 0},{33490, 1, 1, 19}, {0, 0, 0, 0}}},{21639, {{33490, 1, 0, 0},{33490, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{21640, {{33491, 1, 0, 0},{33491, 1, 1, 17}, {0, 0, 0, 0}}},{21640, {{33491, 1, 0, 0},{33491, 1, 1, 17}, {0, 0, 0, 0}}},{21641, {{33491, 1, 0, 0},{33491, 1, 1, 18}, {0, 0, 0, 0}}},{21641, {{33491, 1, 0, 0},{33491, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21642, {{33492, 1, 0, 0},{33492, 1, 1, 17}, {0, 0, 0, 0}}},{21642, {{33492, 1, 0, 0},{33492, 1, 1, 17}, {0, 0, 0, 0}}},{21643, {{33492, 1, 0, 0},{33492, 1, 1, 18}, {0, 0, 0, 0}}},{21643, {{33492, 1, 0, 0},{33492, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21644, {{33493, 1, 0, 0},{33493, 1, 1, 17}, {0, 0, 0, 0}}},{21644, {{33493, 1, 0, 0},{33493, 1, 1, 17}, {0, 0, 0, 0}}},{21645, {{33493, 1, 0, 0},{33493, 1, 1, 18}, {0, 0, 0, 0}}},{21645, {{33493, 1, 0, 0},{33493, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21646, {{33494, 1, 0, 0},{33494, 1, 1, 16}, {0, 0, 0, 0}}},{21646, {{33494, 1, 0, 0},{33494, 1, 1, 16}, {0, 0, 0, 0}}},{21647, {{33494, 1, 0, 0},{33494, 1, 1, 17}, {0, 0, 0, 0}}},{21647, {{33494, 1, 0, 0},{33494, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{21648, {{33495, 1, 0, 0},{33495, 1, 1, 17}, {0, 0, 0, 0}}},{21648, {{33495, 1, 0, 0},{33495, 1, 1, 17}, {0, 0, 0, 0}}},{21649, {{33495, 1, 0, 0},{33495, 1, 1, 18}, {0, 0, 0, 0}}},{21649, {{33495, 1, 0, 0},{33495, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21650, {{33496, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21651, {{33497, 1, 0, 0},{33497, 1, 1, 17}, {0, 0, 0, 0}}},{21651, {{33497, 1, 0, 0},{33497, 1, 1, 17}, {0, 0, 0, 0}}},{21652, {{33497, 1, 0, 0},{33497, 1, 1, 18}, {0, 0, 0, 0}}},{21652, {{33497, 1, 0, 0},{33497, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21653, {{33498, 1, 0, 0},{33498, 1, 1, 17}, {0, 0, 0, 0}}},{21653, {{33498, 1, 0, 0},{33498, 1, 1, 17}, {0, 0, 0, 0}}},{21654, {{33498, 1, 0, 0},{33498, 1, 1, 18}, {0, 0, 0, 0}}},{21654, {{33498, 1, 0, 0},{33498, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21655, {{33499, 1, 0, 0},{33499, 1, 1, 17}, {0, 0, 0, 0}}},{21655, {{33499, 1, 0, 0},{33499, 1, 1, 17}, {0, 0, 0, 0}}},{21656, {{33499, 1, 0, 0},{33499, 1, 1, 18}, {0, 0, 0, 0}}},{21656, {{33499, 1, 0, 0},{33499, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21657, {{33500, 1, 0, 0},{33500, 1, 1, 17}, {0, 0, 0, 0}}},{21657, {{33500, 1, 0, 0},{33500, 1, 1, 17}, {0, 0, 0, 0}}},{21658, {{33500, 1, 0, 0},{33500, 1, 1, 18}, {0, 0, 0, 0}}},{21658, {{33500, 1, 0, 0},{33500, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21659, {{33501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{21660, {{33502, 1, 0, 0},{33502, 1, 1, 17}, {0, 0, 0, 0}}},{21660, {{33502, 1, 0, 0},{33502, 1, 1, 17}, {0, 0, 0, 0}}},{21661, {{33502, 1, 0, 0},{33502, 1, 1, 18}, {0, 0, 0, 0}}},{21661, {{33502, 1, 0, 0},{33502, 1, 1, 18}, {0, 0, 0, 0}}},{21662, {{33502, 1, 0, 0},{33502, 1, 1, 19}, {0, 0, 0, 0}}},{21662, {{33502, 1, 0, 0},{33502, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {6, {{21663, {{33503, 1, 0, 0},{33503, 1, 1, 17}, {0, 0, 0, 0}}},{21663, {{33503, 1, 0, 0},{33503, 1, 1, 17}, {0, 0, 0, 0}}},{21664, {{33503, 1, 0, 0},{33503, 1, 1, 18}, {0, 0, 0, 0}}},{21664, {{33503, 1, 0, 0},{33503, 1, 1, 18}, {0, 0, 0, 0}}},{21665, {{33503, 1, 0, 0},{33503, 1, 1, 19}, {0, 0, 0, 0}}},{21665, {{33503, 1, 0, 0},{33503, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{21666, {{33504, 1, 0, 0},{33504, 1, 1, 17}, {0, 0, 0, 0}}},{21666, {{33504, 1, 0, 0},{33504, 1, 1, 17}, {0, 0, 0, 0}}},{21667, {{33504, 1, 0, 0},{33504, 1, 1, 18}, {0, 0, 0, 0}}},{21667, {{33504, 1, 0, 0},{33504, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21668, {{33505, 1, 0, 0},{33505, 1, 1, 17}, {0, 0, 0, 0}}},{21668, {{33505, 1, 0, 0},{33505, 1, 1, 17}, {0, 0, 0, 0}}},{21669, {{33505, 1, 0, 0},{33505, 1, 1, 18}, {0, 0, 0, 0}}},{21669, {{33505, 1, 0, 0},{33505, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{21670, {{33506, 1, 0, 0},{33506, 1, 1, 17}, {0, 0, 0, 0}}},{21670, {{33506, 1, 0, 0},{33506, 1, 1, 17}, {0, 0, 0, 0}}},{21671, {{33506, 1, 0, 0},{33506, 1, 1, 18}, {0, 0, 0, 0}}},{21671, {{33506, 1, 0, 0},{33506, 1, 1, 18}, {0, 0, 0, 0}}},{58449, {{33506, 1, 0, 0},{33506, 1, 1, 19}, {0, 0, 0, 0}}},{58449, {{33506, 1, 0, 0},{33506, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {6, {{21672, {{33507, 1, 0, 0},{33507, 1, 1, 18}, {0, 0, 0, 0}}},{21672, {{33507, 1, 0, 0},{33507, 1, 1, 18}, {0, 0, 0, 0}}},{21673, {{33507, 1, 0, 0},{33507, 1, 1, 19}, {0, 0, 0, 0}}},{21673, {{33507, 1, 0, 0},{33507, 1, 1, 19}, {0, 0, 0, 0}}},{21674, {{33507, 1, 0, 0},{33507, 1, 1, 20}, {0, 0, 0, 0}}},{21674, {{33507, 1, 0, 0},{33507, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{21675, {{33508, 1, 0, 0},{33508, 1, 1, 17}, {0, 0, 0, 0}}},{21675, {{33508, 1, 0, 0},{33508, 1, 1, 17}, {0, 0, 0, 0}}},{21676, {{33508, 1, 0, 0},{33508, 1, 1, 18}, {0, 0, 0, 0}}},{21676, {{33508, 1, 0, 0},{33508, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{21677, {{33509, 1, 0, 0},{33509, 1, 1, 17}, {0, 0, 0, 0}}},{21677, {{33509, 1, 0, 0},{33509, 1, 1, 17}, {0, 0, 0, 0}}},{21679, {{33509, 1, 0, 0},{33509, 1, 1, 19}, {0, 0, 0, 0}}},{21679, {{33509, 1, 0, 0},{33509, 1, 1, 19}, {0, 0, 0, 0}}},{21680, {{33509, 1, 0, 0},{33509, 1, 1, 18}, {0, 0, 0, 0}}},{21680, {{33509, 1, 0, 0},{33509, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21681, {{33510, 1, 0, 0},{33510, 1, 1, 17}, {0, 0, 0, 0}}},{21681, {{33510, 1, 0, 0},{33510, 1, 1, 17}, {0, 0, 0, 0}}},{21682, {{33510, 1, 0, 0},{33510, 1, 1, 18}, {0, 0, 0, 0}}},{21682, {{33510, 1, 0, 0},{33510, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21683, {{33511, 1, 0, 0},{33511, 1, 1, 17}, {0, 0, 0, 0}}},{21683, {{33511, 1, 0, 0},{33511, 1, 1, 17}, {0, 0, 0, 0}}},{21684, {{33511, 1, 0, 0},{33511, 1, 1, 18}, {0, 0, 0, 0}}},{21684, {{33511, 1, 0, 0},{33511, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21685, {{33512, 1, 0, 0},{33512, 1, 1, 17}, {0, 0, 0, 0}}},{21685, {{33512, 1, 0, 0},{33512, 1, 1, 17}, {0, 0, 0, 0}}},{21686, {{33512, 1, 0, 0},{33512, 1, 1, 18}, {0, 0, 0, 0}}},{21686, {{33512, 1, 0, 0},{33512, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21687, {{33513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21688, {{33514, 1, 0, 0},{33514, 1, 1, 17}, {0, 0, 0, 0}}},{21688, {{33514, 1, 0, 0},{33514, 1, 1, 17}, {0, 0, 0, 0}}},{21689, {{33514, 1, 0, 0},{33514, 1, 1, 18}, {0, 0, 0, 0}}},{21689, {{33514, 1, 0, 0},{33514, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21690, {{33515, 1, 0, 0},{33515, 1, 1, 17}, {0, 0, 0, 0}}},{21690, {{33515, 1, 0, 0},{33515, 1, 1, 17}, {0, 0, 0, 0}}},{21691, {{33515, 1, 0, 0},{33515, 1, 1, 18}, {0, 0, 0, 0}}},{21691, {{33515, 1, 0, 0},{33515, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21692, {{33516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21693, {{33517, 1, 0, 0},{33517, 1, 1, 17}, {0, 0, 0, 0}}},{21693, {{33517, 1, 0, 0},{33517, 1, 1, 17}, {0, 0, 0, 0}}},{21694, {{33517, 1, 0, 0},{33517, 1, 1, 18}, {0, 0, 0, 0}}},{21694, {{33517, 1, 0, 0},{33517, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21695, {{33518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21696, {{33519, 1, 0, 0},{33519, 1, 1, 17}, {0, 0, 0, 0}}},{21696, {{33519, 1, 0, 0},{33519, 1, 1, 17}, {0, 0, 0, 0}}},{21697, {{33519, 1, 0, 0},{33519, 1, 1, 18}, {0, 0, 0, 0}}},{21697, {{33519, 1, 0, 0},{33519, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21699, {{33520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{21700, {{33521, 1, 0, 0},{33521, 1, 1, 17}, {0, 0, 0, 0}}},{21700, {{33521, 1, 0, 0},{33521, 1, 1, 17}, {0, 0, 0, 0}}},{21701, {{33521, 1, 0, 0},{33521, 1, 1, 20}, {0, 0, 0, 0}}},{21701, {{33521, 1, 0, 0},{33521, 1, 1, 20}, {0, 0, 0, 0}}},{21702, {{33521, 1, 0, 0},{33521, 1, 1, 18}, {0, 0, 0, 0}}},{21702, {{33521, 1, 0, 0},{33521, 1, 1, 18}, {0, 0, 0, 0}}},{21703, {{33521, 1, 0, 0},{33521, 1, 1, 19}, {0, 0, 0, 0}}},{21703, {{33521, 1, 0, 0},{33521, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{21705, {{33522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{21706, {{33523, 1, 0, 0},{33523, 1, 1, 17}, {0, 0, 0, 0}}},{21706, {{33523, 1, 0, 0},{33523, 1, 1, 17}, {0, 0, 0, 0}}},{21707, {{33523, 1, 0, 0},{33523, 1, 1, 18}, {0, 0, 0, 0}}},{21707, {{33523, 1, 0, 0},{33523, 1, 1, 18}, {0, 0, 0, 0}}},{21708, {{33523, 1, 0, 0},{33523, 1, 1, 19}, {0, 0, 0, 0}}},{21708, {{33523, 1, 0, 0},{33523, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{21709, {{33524, 1, 0, 0},{33524, 1, 1, 17}, {0, 0, 0, 0}}},{21709, {{33524, 1, 0, 0},{33524, 1, 1, 17}, {0, 0, 0, 0}}},{21710, {{33524, 1, 0, 0},{33524, 1, 1, 18}, {0, 0, 0, 0}}},{21710, {{33524, 1, 0, 0},{33524, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21711, {{33525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21712, {{33526, 1, 0, 0},{33526, 1, 1, 17}, {0, 0, 0, 0}}},{21712, {{33526, 1, 0, 0},{33526, 1, 1, 17}, {0, 0, 0, 0}}},{21713, {{33526, 1, 0, 0},{33526, 1, 1, 18}, {0, 0, 0, 0}}},{21713, {{33526, 1, 0, 0},{33526, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21714, {{33527, 1, 0, 0},{33527, 1, 1, 17}, {0, 0, 0, 0}}},{21714, {{33527, 1, 0, 0},{33527, 1, 1, 17}, {0, 0, 0, 0}}},{21715, {{33527, 1, 0, 0},{33527, 1, 1, 18}, {0, 0, 0, 0}}},{21715, {{33527, 1, 0, 0},{33527, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21716, {{33528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{21717, {{33529, 1, 0, 0},{33529, 1, 1, 17}, {0, 0, 0, 0}}},{21717, {{33529, 1, 0, 0},{33529, 1, 1, 17}, {0, 0, 0, 0}}},{21718, {{33529, 1, 0, 0},{33529, 1, 1, 19}, {0, 0, 0, 0}}},{21718, {{33529, 1, 0, 0},{33529, 1, 1, 19}, {0, 0, 0, 0}}},{21719, {{33529, 1, 0, 0},{33529, 1, 1, 18}, {0, 0, 0, 0}}},{21719, {{33529, 1, 0, 0},{33529, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21720, {{33530, 1, 0, 0},{33530, 1, 1, 17}, {0, 0, 0, 0}}},{21720, {{33530, 1, 0, 0},{33530, 1, 1, 17}, {0, 0, 0, 0}}},{21721, {{33530, 1, 0, 0},{33530, 1, 1, 18}, {0, 0, 0, 0}}},{21721, {{33530, 1, 0, 0},{33530, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21722, {{33531, 1, 0, 0},{33531, 1, 1, 17}, {0, 0, 0, 0}}},{21722, {{33531, 1, 0, 0},{33531, 1, 1, 17}, {0, 0, 0, 0}}},{21723, {{33531, 1, 0, 0},{33531, 1, 1, 18}, {0, 0, 0, 0}}},{21723, {{33531, 1, 0, 0},{33531, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21724, {{33532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{21726, {{33533, 1, 0, 0},{33533, 1, 1, 19}, {0, 0, 0, 0}}},{21726, {{33533, 1, 0, 0},{33533, 1, 1, 19}, {0, 0, 0, 0}}},{21727, {{33533, 1, 0, 0},{33533, 1, 1, 17}, {0, 0, 0, 0}}},{21727, {{33533, 1, 0, 0},{33533, 1, 1, 17}, {0, 0, 0, 0}}},{21728, {{33533, 1, 0, 0},{33533, 1, 1, 18}, {0, 0, 0, 0}}},{21728, {{33533, 1, 0, 0},{33533, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21729, {{33534, 1, 0, 0},{33534, 1, 1, 17}, {0, 0, 0, 0}}},{21729, {{33534, 1, 0, 0},{33534, 1, 1, 17}, {0, 0, 0, 0}}},{21730, {{33534, 1, 0, 0},{33534, 1, 1, 18}, {0, 0, 0, 0}}},{21730, {{33534, 1, 0, 0},{33534, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21731, {{33535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21732, {{33536, 1, 0, 0},{33536, 1, 1, 17}, {0, 0, 0, 0}}},{21732, {{33536, 1, 0, 0},{33536, 1, 1, 17}, {0, 0, 0, 0}}},{21733, {{33536, 1, 0, 0},{33536, 1, 1, 18}, {0, 0, 0, 0}}},{21733, {{33536, 1, 0, 0},{33536, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21734, {{33537, 1, 0, 0},{33537, 1, 1, 17}, {0, 0, 0, 0}}},{21734, {{33537, 1, 0, 0},{33537, 1, 1, 17}, {0, 0, 0, 0}}},{21735, {{33537, 1, 0, 0},{33537, 1, 1, 18}, {0, 0, 0, 0}}},{21735, {{33537, 1, 0, 0},{33537, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{21736, {{33538, 1, 0, 0},{33538, 1, 1, 17}, {0, 0, 0, 0}}},{21736, {{33538, 1, 0, 0},{33538, 1, 1, 17}, {0, 0, 0, 0}}},{21737, {{33538, 1, 0, 0},{33538, 1, 1, 19}, {0, 0, 0, 0}}},{21737, {{33538, 1, 0, 0},{33538, 1, 1, 19}, {0, 0, 0, 0}}},{21738, {{33538, 1, 0, 0},{33538, 1, 1, 18}, {0, 0, 0, 0}}},{21738, {{33538, 1, 0, 0},{33538, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{21739, {{33539, 1, 0, 0},{33539, 1, 1, 17}, {0, 0, 0, 0}}},{21739, {{33539, 1, 0, 0},{33539, 1, 1, 17}, {0, 0, 0, 0}}},{21740, {{33539, 1, 0, 0},{33539, 1, 1, 19}, {0, 0, 0, 0}}},{21740, {{33539, 1, 0, 0},{33539, 1, 1, 19}, {0, 0, 0, 0}}},{21741, {{33539, 1, 0, 0},{33539, 1, 1, 18}, {0, 0, 0, 0}}},{21741, {{33539, 1, 0, 0},{33539, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21743, {{33540, 1, 0, 0},{33540, 1, 1, 17}, {0, 0, 0, 0}}},{21743, {{33540, 1, 0, 0},{33540, 1, 1, 17}, {0, 0, 0, 0}}},{21744, {{33540, 1, 0, 0},{33540, 1, 1, 18}, {0, 0, 0, 0}}},{21744, {{33540, 1, 0, 0},{33540, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21745, {{33541, 1, 0, 0},{33541, 1, 1, 17}, {0, 0, 0, 0}}},{21745, {{33541, 1, 0, 0},{33541, 1, 1, 17}, {0, 0, 0, 0}}},{21746, {{33541, 1, 0, 0},{33541, 1, 1, 18}, {0, 0, 0, 0}}},{21746, {{33541, 1, 0, 0},{33541, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21747, {{33542, 1, 0, 0},{33542, 1, 1, 17}, {0, 0, 0, 0}}},{21747, {{33542, 1, 0, 0},{33542, 1, 1, 17}, {0, 0, 0, 0}}},{21748, {{33542, 1, 0, 0},{33542, 1, 1, 18}, {0, 0, 0, 0}}},{21748, {{33542, 1, 0, 0},{33542, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21749, {{33543, 1, 0, 0},{33543, 1, 1, 18}, {0, 0, 0, 0}}},{21749, {{33543, 1, 0, 0},{33543, 1, 1, 18}, {0, 0, 0, 0}}},{21750, {{33543, 1, 0, 0},{33543, 1, 1, 17}, {0, 0, 0, 0}}},{21750, {{33543, 1, 0, 0},{33543, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{21751, {{33544, 1, 0, 0},{33544, 1, 1, 18}, {0, 0, 0, 0}}},{21751, {{33544, 1, 0, 0},{33544, 1, 1, 18}, {0, 0, 0, 0}}},{21752, {{33544, 1, 0, 0},{33544, 1, 1, 17}, {0, 0, 0, 0}}},{21752, {{33544, 1, 0, 0},{33544, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{21753, {{33545, 1, 0, 0},{33545, 1, 1, 17}, {0, 0, 0, 0}}},{21753, {{33545, 1, 0, 0},{33545, 1, 1, 17}, {0, 0, 0, 0}}},{21754, {{33545, 1, 0, 0},{33545, 1, 1, 18}, {0, 0, 0, 0}}},{21754, {{33545, 1, 0, 0},{33545, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21755, {{33546, 1, 0, 0},{33546, 1, 1, 18}, {0, 0, 0, 0}}},{21755, {{33546, 1, 0, 0},{33546, 1, 1, 18}, {0, 0, 0, 0}}},{21756, {{33546, 1, 0, 0},{33546, 1, 1, 17}, {0, 0, 0, 0}}},{21756, {{33546, 1, 0, 0},{33546, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{21757, {{33547, 1, 0, 0},{33547, 1, 1, 17}, {0, 0, 0, 0}}},{21757, {{33547, 1, 0, 0},{33547, 1, 1, 17}, {0, 0, 0, 0}}},{21758, {{33547, 1, 0, 0},{33547, 1, 1, 18}, {0, 0, 0, 0}}},{21758, {{33547, 1, 0, 0},{33547, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21759, {{33548, 1, 0, 0},{33548, 1, 1, 17}, {0, 0, 0, 0}}},{21759, {{33548, 1, 0, 0},{33548, 1, 1, 17}, {0, 0, 0, 0}}},{21760, {{33548, 1, 0, 0},{33548, 1, 1, 18}, {0, 0, 0, 0}}},{21760, {{33548, 1, 0, 0},{33548, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21762, {{33549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21763, {{33550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{21764, {{33558, 1, 0, 0},{33558, 1, 1, 17}, {0, 0, 0, 0}}},{21764, {{33558, 1, 0, 0},{33558, 1, 1, 17}, {0, 0, 0, 0}}},{21765, {{33558, 1, 0, 0},{33558, 1, 1, 18}, {0, 0, 0, 0}}},{21765, {{33558, 1, 0, 0},{33558, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21766, {{33559, 1, 0, 0},{33559, 1, 1, 17}, {0, 0, 0, 0}}},{21766, {{33559, 1, 0, 0},{33559, 1, 1, 17}, {0, 0, 0, 0}}},{21767, {{33559, 1, 0, 0},{33559, 1, 1, 18}, {0, 0, 0, 0}}},{21767, {{33559, 1, 0, 0},{33559, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21768, {{33560, 1, 0, 0},{33560, 1, 1, 17}, {0, 0, 0, 0}}},{21768, {{33560, 1, 0, 0},{33560, 1, 1, 17}, {0, 0, 0, 0}}},{21769, {{33560, 1, 0, 0},{33560, 1, 1, 18}, {0, 0, 0, 0}}},{21769, {{33560, 1, 0, 0},{33560, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21770, {{33561, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21771, {{33562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21772, {{33563, 1, 0, 0},{33563, 1, 1, 17}, {0, 0, 0, 0}}},{21772, {{33563, 1, 0, 0},{33563, 1, 1, 17}, {0, 0, 0, 0}}},{21773, {{33563, 1, 0, 0},{33563, 1, 1, 18}, {0, 0, 0, 0}}},{21773, {{33563, 1, 0, 0},{33563, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21774, {{33564, 1, 0, 0},{33564, 1, 1, 17}, {0, 0, 0, 0}}},{21774, {{33564, 1, 0, 0},{33564, 1, 1, 17}, {0, 0, 0, 0}}},{21775, {{33564, 1, 0, 0},{33564, 1, 1, 18}, {0, 0, 0, 0}}},{21775, {{33564, 1, 0, 0},{33564, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21776, {{33565, 1, 0, 0},{33565, 1, 1, 17}, {0, 0, 0, 0}}},{21776, {{33565, 1, 0, 0},{33565, 1, 1, 17}, {0, 0, 0, 0}}},{21777, {{33565, 1, 0, 0},{33565, 1, 1, 18}, {0, 0, 0, 0}}},{21777, {{33565, 1, 0, 0},{33565, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21778, {{33566, 1, 0, 0},{33566, 1, 1, 17}, {0, 0, 0, 0}}},{21778, {{33566, 1, 0, 0},{33566, 1, 1, 17}, {0, 0, 0, 0}}},{21779, {{33566, 1, 0, 0},{33566, 1, 1, 18}, {0, 0, 0, 0}}},{21779, {{33566, 1, 0, 0},{33566, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21780, {{33567, 1, 0, 0},{33567, 1, 1, 17}, {0, 0, 0, 0}}},{21780, {{33567, 1, 0, 0},{33567, 1, 1, 17}, {0, 0, 0, 0}}},{21781, {{33567, 1, 0, 0},{33567, 1, 1, 18}, {0, 0, 0, 0}}},{21781, {{33567, 1, 0, 0},{33567, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21782, {{33568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21783, {{33569, 1, 0, 0},{33569, 1, 1, 17}, {0, 0, 0, 0}}},{21783, {{33569, 1, 0, 0},{33569, 1, 1, 17}, {0, 0, 0, 0}}},{21784, {{33569, 1, 0, 0},{33569, 1, 1, 18}, {0, 0, 0, 0}}},{21784, {{33569, 1, 0, 0},{33569, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21785, {{33570, 1, 0, 0},{33570, 1, 1, 17}, {0, 0, 0, 0}}},{21785, {{33570, 1, 0, 0},{33570, 1, 1, 17}, {0, 0, 0, 0}}},{21786, {{33570, 1, 0, 0},{33570, 1, 1, 18}, {0, 0, 0, 0}}},{21786, {{33570, 1, 0, 0},{33570, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {10, {{21787, {{33571, 1, 0, 0},{33571, 1, 1, 21}, {0, 0, 0, 0}}},{21787, {{33571, 1, 0, 0},{33571, 1, 1, 21}, {0, 0, 0, 0}}},{21788, {{33571, 1, 0, 0},{33571, 1, 1, 18}, {0, 0, 0, 0}}},{21788, {{33571, 1, 0, 0},{33571, 1, 1, 18}, {0, 0, 0, 0}}},{21789, {{33571, 1, 0, 0},{33571, 1, 1, 20}, {0, 0, 0, 0}}},{21789, {{33571, 1, 0, 0},{33571, 1, 1, 20}, {0, 0, 0, 0}}},{21790, {{33571, 1, 0, 0},{33571, 1, 1, 19}, {0, 0, 0, 0}}},{21790, {{33571, 1, 0, 0},{33571, 1, 1, 19}, {0, 0, 0, 0}}},{21791, {{33571, 1, 0, 0},{33571, 1, 1, 22}, {0, 0, 0, 0}}},{21791, {{33571, 1, 0, 0},{33571, 1, 1, 22}, {0, 0, 0, 0}}}}},
    {1, {{21792, {{33572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21794, {{33573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21796, {{33574, 1, 0, 0},{33574, 1, 1, 16}, {0, 0, 0, 0}}},{21796, {{33574, 1, 0, 0},{33574, 1, 1, 16}, {0, 0, 0, 0}}},{21797, {{33574, 1, 0, 0},{33574, 1, 1, 17}, {0, 0, 0, 0}}},{21797, {{33574, 1, 0, 0},{33574, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{21798, {{33575, 1, 0, 0},{33575, 1, 1, 16}, {0, 0, 0, 0}}},{21798, {{33575, 1, 0, 0},{33575, 1, 1, 16}, {0, 0, 0, 0}}},{21799, {{33575, 1, 0, 0},{33575, 1, 1, 17}, {0, 0, 0, 0}}},{21799, {{33575, 1, 0, 0},{33575, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {6, {{21800, {{33576, 1, 0, 0},{33576, 1, 1, 18}, {0, 0, 0, 0}}},{21800, {{33576, 1, 0, 0},{33576, 1, 1, 18}, {0, 0, 0, 0}}},{21801, {{33576, 1, 0, 0},{33576, 1, 1, 20}, {0, 0, 0, 0}}},{21801, {{33576, 1, 0, 0},{33576, 1, 1, 20}, {0, 0, 0, 0}}},{21802, {{33576, 1, 0, 0},{33576, 1, 1, 19}, {0, 0, 0, 0}}},{21802, {{33576, 1, 0, 0},{33576, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{21805, {{33577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21806, {{33578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{21808, {{33579, 1, 0, 0},{33579, 1, 1, 17}, {0, 0, 0, 0}}},{21808, {{33579, 1, 0, 0},{33579, 1, 1, 17}, {0, 0, 0, 0}}},{21809, {{33579, 1, 0, 0},{33579, 1, 1, 18}, {0, 0, 0, 0}}},{21809, {{33579, 1, 0, 0},{33579, 1, 1, 18}, {0, 0, 0, 0}}},{21810, {{33579, 1, 0, 0},{33579, 1, 1, 19}, {0, 0, 0, 0}}},{21810, {{33579, 1, 0, 0},{33579, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{21811, {{33580, 1, 0, 0},{33580, 1, 1, 17}, {0, 0, 0, 0}}},{21811, {{33580, 1, 0, 0},{33580, 1, 1, 17}, {0, 0, 0, 0}}},{21812, {{33580, 1, 0, 0},{33580, 1, 1, 18}, {0, 0, 0, 0}}},{21812, {{33580, 1, 0, 0},{33580, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21813, {{33581, 1, 0, 0},{33581, 1, 1, 18}, {0, 0, 0, 0}}},{21813, {{33581, 1, 0, 0},{33581, 1, 1, 18}, {0, 0, 0, 0}}},{21814, {{33581, 1, 0, 0},{33581, 1, 1, 17}, {0, 0, 0, 0}}},{21814, {{33581, 1, 0, 0},{33581, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{21815, {{33582, 1, 0, 0},{33582, 1, 1, 19}, {0, 0, 0, 0}}},{21815, {{33582, 1, 0, 0},{33582, 1, 1, 19}, {0, 0, 0, 0}}},{21816, {{33582, 1, 0, 0},{33582, 1, 1, 18}, {0, 0, 0, 0}}},{21816, {{33582, 1, 0, 0},{33582, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21817, {{33583, 1, 0, 0},{33583, 1, 1, 17}, {0, 0, 0, 0}}},{21817, {{33583, 1, 0, 0},{33583, 1, 1, 17}, {0, 0, 0, 0}}},{21818, {{33583, 1, 0, 0},{33583, 1, 1, 18}, {0, 0, 0, 0}}},{21818, {{33583, 1, 0, 0},{33583, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {10, {{21819, {{33584, 1, 0, 0},{33584, 1, 1, 19}, {0, 0, 0, 0}}},{21819, {{33584, 1, 0, 0},{33584, 1, 1, 19}, {0, 0, 0, 0}}},{21820, {{33584, 1, 0, 0},{33584, 1, 1, 21}, {0, 0, 0, 0}}},{21820, {{33584, 1, 0, 0},{33584, 1, 1, 21}, {0, 0, 0, 0}}},{21821, {{33584, 1, 0, 0},{33584, 1, 1, 20}, {0, 0, 0, 0}}},{21821, {{33584, 1, 0, 0},{33584, 1, 1, 20}, {0, 0, 0, 0}}},{21822, {{33584, 1, 0, 0},{33584, 1, 1, 17}, {0, 0, 0, 0}}},{21822, {{33584, 1, 0, 0},{33584, 1, 1, 17}, {0, 0, 0, 0}}},{21823, {{33584, 1, 0, 0},{33584, 1, 1, 18}, {0, 0, 0, 0}}},{21823, {{33584, 1, 0, 0},{33584, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21824, {{33585, 1, 0, 0},{33585, 1, 1, 17}, {0, 0, 0, 0}}},{21824, {{33585, 1, 0, 0},{33585, 1, 1, 17}, {0, 0, 0, 0}}},{21825, {{33585, 1, 0, 0},{33585, 1, 1, 18}, {0, 0, 0, 0}}},{21825, {{33585, 1, 0, 0},{33585, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21826, {{33586, 1, 0, 0},{33586, 1, 1, 17}, {0, 0, 0, 0}}},{21826, {{33586, 1, 0, 0},{33586, 1, 1, 17}, {0, 0, 0, 0}}},{21827, {{33586, 1, 0, 0},{33586, 1, 1, 18}, {0, 0, 0, 0}}},{21827, {{33586, 1, 0, 0},{33586, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21828, {{33587, 1, 0, 0},{33587, 1, 1, 17}, {0, 0, 0, 0}}},{21828, {{33587, 1, 0, 0},{33587, 1, 1, 17}, {0, 0, 0, 0}}},{21829, {{33587, 1, 0, 0},{33587, 1, 1, 18}, {0, 0, 0, 0}}},{21829, {{33587, 1, 0, 0},{33587, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21830, {{33588, 1, 0, 0},{33588, 1, 1, 17}, {0, 0, 0, 0}}},{21830, {{33588, 1, 0, 0},{33588, 1, 1, 17}, {0, 0, 0, 0}}},{21831, {{33588, 1, 0, 0},{33588, 1, 1, 18}, {0, 0, 0, 0}}},{21831, {{33588, 1, 0, 0},{33588, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21832, {{33589, 1, 0, 0},{33589, 1, 1, 17}, {0, 0, 0, 0}}},{21832, {{33589, 1, 0, 0},{33589, 1, 1, 17}, {0, 0, 0, 0}}},{21833, {{33589, 1, 0, 0},{33589, 1, 1, 18}, {0, 0, 0, 0}}},{21833, {{33589, 1, 0, 0},{33589, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21834, {{33590, 1, 0, 0},{33590, 1, 1, 17}, {0, 0, 0, 0}}},{21834, {{33590, 1, 0, 0},{33590, 1, 1, 17}, {0, 0, 0, 0}}},{21835, {{33590, 1, 0, 0},{33590, 1, 1, 18}, {0, 0, 0, 0}}},{21835, {{33590, 1, 0, 0},{33590, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21836, {{33591, 1, 0, 0},{33591, 1, 1, 17}, {0, 0, 0, 0}}},{21836, {{33591, 1, 0, 0},{33591, 1, 1, 17}, {0, 0, 0, 0}}},{21837, {{33591, 1, 0, 0},{33591, 1, 1, 18}, {0, 0, 0, 0}}},{21837, {{33591, 1, 0, 0},{33591, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{21838, {{33592, 1, 0, 0},{33592, 1, 1, 17}, {0, 0, 0, 0}}},{21838, {{33592, 1, 0, 0},{33592, 1, 1, 17}, {0, 0, 0, 0}}},{21839, {{33592, 1, 0, 0},{33592, 1, 1, 18}, {0, 0, 0, 0}}},{21839, {{33592, 1, 0, 0},{33592, 1, 1, 18}, {0, 0, 0, 0}}},{21840, {{33592, 1, 0, 0},{33592, 1, 1, 19}, {0, 0, 0, 0}}},{21840, {{33592, 1, 0, 0},{33592, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{21841, {{33593, 1, 0, 0},{33593, 1, 1, 17}, {0, 0, 0, 0}}},{21841, {{33593, 1, 0, 0},{33593, 1, 1, 17}, {0, 0, 0, 0}}},{21842, {{33593, 1, 0, 0},{33593, 1, 1, 18}, {0, 0, 0, 0}}},{21842, {{33593, 1, 0, 0},{33593, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21843, {{33594, 1, 0, 0},{33594, 1, 1, 17}, {0, 0, 0, 0}}},{21843, {{33594, 1, 0, 0},{33594, 1, 1, 17}, {0, 0, 0, 0}}},{21844, {{33594, 1, 0, 0},{33594, 1, 1, 18}, {0, 0, 0, 0}}},{21844, {{33594, 1, 0, 0},{33594, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21845, {{33595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21846, {{33596, 1, 0, 0},{33596, 1, 1, 17}, {0, 0, 0, 0}}},{21846, {{33596, 1, 0, 0},{33596, 1, 1, 17}, {0, 0, 0, 0}}},{21847, {{33596, 1, 0, 0},{33596, 1, 1, 18}, {0, 0, 0, 0}}},{21847, {{33596, 1, 0, 0},{33596, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21848, {{33597, 1, 0, 0},{33597, 1, 1, 17}, {0, 0, 0, 0}}},{21848, {{33597, 1, 0, 0},{33597, 1, 1, 17}, {0, 0, 0, 0}}},{21849, {{33597, 1, 0, 0},{33597, 1, 1, 18}, {0, 0, 0, 0}}},{21849, {{33597, 1, 0, 0},{33597, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21850, {{33598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21851, {{33599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21852, {{33600, 1, 0, 0},{33600, 1, 1, 17}, {0, 0, 0, 0}}},{21852, {{33600, 1, 0, 0},{33600, 1, 1, 17}, {0, 0, 0, 0}}},{21853, {{33600, 1, 0, 0},{33600, 1, 1, 18}, {0, 0, 0, 0}}},{21853, {{33600, 1, 0, 0},{33600, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21854, {{33601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{21855, {{33602, 1, 0, 0},{33602, 1, 1, 17}, {0, 0, 0, 0}}},{21855, {{33602, 1, 0, 0},{33602, 1, 1, 17}, {0, 0, 0, 0}}},{21856, {{33602, 1, 0, 0},{33602, 1, 1, 18}, {0, 0, 0, 0}}},{21856, {{33602, 1, 0, 0},{33602, 1, 1, 18}, {0, 0, 0, 0}}},{21857, {{33602, 1, 0, 0},{33602, 1, 1, 19}, {0, 0, 0, 0}}},{21857, {{33602, 1, 0, 0},{33602, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {8, {{21858, {{33603, 1, 0, 0},{33603, 1, 1, 18}, {0, 0, 0, 0}}},{21858, {{33603, 1, 0, 0},{33603, 1, 1, 18}, {0, 0, 0, 0}}},{21859, {{33603, 1, 0, 0},{33603, 1, 1, 17}, {0, 0, 0, 0}}},{21859, {{33603, 1, 0, 0},{33603, 1, 1, 17}, {0, 0, 0, 0}}},{21860, {{33603, 1, 0, 0},{33603, 1, 1, 19}, {0, 0, 0, 0}}},{21860, {{33603, 1, 0, 0},{33603, 1, 1, 19}, {0, 0, 0, 0}}},{21861, {{33603, 1, 0, 0},{33603, 1, 1, 20}, {0, 0, 0, 0}}},{21861, {{33603, 1, 0, 0},{33603, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {6, {{21862, {{33604, 1, 0, 0},{33604, 1, 1, 17}, {0, 0, 0, 0}}},{21862, {{33604, 1, 0, 0},{33604, 1, 1, 17}, {0, 0, 0, 0}}},{21863, {{33604, 1, 0, 0},{33604, 1, 1, 19}, {0, 0, 0, 0}}},{21863, {{33604, 1, 0, 0},{33604, 1, 1, 19}, {0, 0, 0, 0}}},{21864, {{33604, 1, 0, 0},{33604, 1, 1, 18}, {0, 0, 0, 0}}},{21864, {{33604, 1, 0, 0},{33604, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{21865, {{33605, 1, 0, 0},{33605, 1, 1, 17}, {0, 0, 0, 0}}},{21865, {{33605, 1, 0, 0},{33605, 1, 1, 17}, {0, 0, 0, 0}}},{21866, {{33605, 1, 0, 0},{33605, 1, 1, 18}, {0, 0, 0, 0}}},{21866, {{33605, 1, 0, 0},{33605, 1, 1, 18}, {0, 0, 0, 0}}},{21867, {{33605, 1, 0, 0},{33605, 1, 1, 19}, {0, 0, 0, 0}}},{21867, {{33605, 1, 0, 0},{33605, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {8, {{21868, {{33606, 1, 0, 0},{33606, 1, 1, 17}, {0, 0, 0, 0}}},{21868, {{33606, 1, 0, 0},{33606, 1, 1, 17}, {0, 0, 0, 0}}},{21869, {{33606, 1, 0, 0},{33606, 1, 1, 19}, {0, 0, 0, 0}}},{21869, {{33606, 1, 0, 0},{33606, 1, 1, 19}, {0, 0, 0, 0}}},{21870, {{33606, 1, 0, 0},{33606, 1, 1, 18}, {0, 0, 0, 0}}},{21870, {{33606, 1, 0, 0},{33606, 1, 1, 18}, {0, 0, 0, 0}}},{21871, {{33606, 1, 0, 0},{33606, 1, 1, 20}, {0, 0, 0, 0}}},{21871, {{33606, 1, 0, 0},{33606, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{21872, {{33607, 1, 0, 0},{33607, 1, 1, 17}, {0, 0, 0, 0}}},{21872, {{33607, 1, 0, 0},{33607, 1, 1, 17}, {0, 0, 0, 0}}},{21873, {{33607, 1, 0, 0},{33607, 1, 1, 18}, {0, 0, 0, 0}}},{21873, {{33607, 1, 0, 0},{33607, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21874, {{33608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21875, {{33609, 1, 0, 0},{33609, 1, 1, 17}, {0, 0, 0, 0}}},{21875, {{33609, 1, 0, 0},{33609, 1, 1, 17}, {0, 0, 0, 0}}},{21876, {{33609, 1, 0, 0},{33609, 1, 1, 18}, {0, 0, 0, 0}}},{21876, {{33609, 1, 0, 0},{33609, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {8, {{21877, {{33610, 1, 0, 0},{33610, 1, 1, 17}, {0, 0, 0, 0}}},{21877, {{33610, 1, 0, 0},{33610, 1, 1, 17}, {0, 0, 0, 0}}},{21878, {{33610, 1, 0, 0},{33610, 1, 1, 19}, {0, 0, 0, 0}}},{21878, {{33610, 1, 0, 0},{33610, 1, 1, 19}, {0, 0, 0, 0}}},{21879, {{33610, 1, 0, 0},{33610, 1, 1, 20}, {0, 0, 0, 0}}},{21879, {{33610, 1, 0, 0},{33610, 1, 1, 20}, {0, 0, 0, 0}}},{21880, {{33610, 1, 0, 0},{33610, 1, 1, 18}, {0, 0, 0, 0}}},{21880, {{33610, 1, 0, 0},{33610, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21881, {{33611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21884, {{33612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21885, {{33613, 1, 0, 0},{33613, 1, 1, 17}, {0, 0, 0, 0}}},{21885, {{33613, 1, 0, 0},{33613, 1, 1, 17}, {0, 0, 0, 0}}},{21886, {{33613, 1, 0, 0},{33613, 1, 1, 18}, {0, 0, 0, 0}}},{21886, {{33613, 1, 0, 0},{33613, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21887, {{33614, 1, 0, 0},{33614, 1, 1, 17}, {0, 0, 0, 0}}},{21887, {{33614, 1, 0, 0},{33614, 1, 1, 17}, {0, 0, 0, 0}}},{21888, {{33614, 1, 0, 0},{33614, 1, 1, 18}, {0, 0, 0, 0}}},{21888, {{33614, 1, 0, 0},{33614, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{21889, {{33615, 1, 0, 0},{33615, 1, 1, 17}, {0, 0, 0, 0}}},{21889, {{33615, 1, 0, 0},{33615, 1, 1, 17}, {0, 0, 0, 0}}},{21890, {{33615, 1, 0, 0},{33615, 1, 1, 18}, {0, 0, 0, 0}}},{21890, {{33615, 1, 0, 0},{33615, 1, 1, 18}, {0, 0, 0, 0}}},{21891, {{33615, 1, 0, 0},{33615, 1, 1, 19}, {0, 0, 0, 0}}},{21891, {{33615, 1, 0, 0},{33615, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{21892, {{33616, 1, 0, 0},{33616, 1, 1, 17}, {0, 0, 0, 0}}},{21892, {{33616, 1, 0, 0},{33616, 1, 1, 17}, {0, 0, 0, 0}}},{21893, {{33616, 1, 0, 0},{33616, 1, 1, 18}, {0, 0, 0, 0}}},{21893, {{33616, 1, 0, 0},{33616, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{21894, {{33617, 1, 0, 0},{33617, 1, 1, 17}, {0, 0, 0, 0}}},{21894, {{33617, 1, 0, 0},{33617, 1, 1, 17}, {0, 0, 0, 0}}},{21895, {{33617, 1, 0, 0},{33617, 1, 1, 18}, {0, 0, 0, 0}}},{21895, {{33617, 1, 0, 0},{33617, 1, 1, 18}, {0, 0, 0, 0}}},{21896, {{33617, 1, 0, 0},{33617, 1, 1, 19}, {0, 0, 0, 0}}},{21896, {{33617, 1, 0, 0},{33617, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {14, {{21897, {{33618, 1, 0, 0},{33618, 1, 1, 18}, {0, 0, 0, 0}}},{21897, {{33618, 1, 0, 0},{33618, 1, 1, 18}, {0, 0, 0, 0}}},{21898, {{33618, 1, 0, 0},{33618, 1, 1, 19}, {0, 0, 0, 0}}},{21898, {{33618, 1, 0, 0},{33618, 1, 1, 19}, {0, 0, 0, 0}}},{21899, {{33618, 1, 0, 0},{33618, 1, 1, 20}, {0, 0, 0, 0}}},{21899, {{33618, 1, 0, 0},{33618, 1, 1, 20}, {0, 0, 0, 0}}},{58459, {{33618, 1, 0, 0},{33618, 1, 1, 22}, {0, 0, 0, 0}}},{58459, {{33618, 1, 0, 0},{33618, 1, 1, 22}, {0, 0, 0, 0}}},{60100, {{33618, 1, 0, 0},{33618, 1, 1, 21}, {0, 0, 0, 0}}},{60100, {{33618, 1, 0, 0},{33618, 1, 1, 21}, {0, 0, 0, 0}}},{60106, {{33618, 1, 0, 0},{33618, 1, 1, 23}, {0, 0, 0, 0}}},{60106, {{33618, 1, 0, 0},{33618, 1, 1, 23}, {0, 0, 0, 0}}},{60384, {{33618, 1, 0, 0},{33618, 1, 1, 24}, {0, 0, 0, 0}}},{60384, {{33618, 1, 0, 0},{33618, 1, 1, 24}, {0, 0, 0, 0}}}}},
    {8, {{21900, {{33619, 1, 0, 0},{33619, 1, 1, 18}, {0, 0, 0, 0}}},{21900, {{33619, 1, 0, 0},{33619, 1, 1, 18}, {0, 0, 0, 0}}},{21901, {{33619, 1, 0, 0},{33619, 1, 1, 19}, {0, 0, 0, 0}}},{21901, {{33619, 1, 0, 0},{33619, 1, 1, 19}, {0, 0, 0, 0}}},{21902, {{33619, 1, 0, 0},{33619, 1, 1, 17}, {0, 0, 0, 0}}},{21902, {{33619, 1, 0, 0},{33619, 1, 1, 17}, {0, 0, 0, 0}}},{21903, {{33619, 1, 0, 0},{33619, 1, 1, 20}, {0, 0, 0, 0}}},{21903, {{33619, 1, 0, 0},{33619, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{21904, {{33620, 1, 0, 0},{33620, 1, 1, 17}, {0, 0, 0, 0}}},{21904, {{33620, 1, 0, 0},{33620, 1, 1, 17}, {0, 0, 0, 0}}},{21905, {{33620, 1, 0, 0},{33620, 1, 1, 18}, {0, 0, 0, 0}}},{21905, {{33620, 1, 0, 0},{33620, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21906, {{33621, 1, 0, 0},{33621, 1, 1, 17}, {0, 0, 0, 0}}},{21906, {{33621, 1, 0, 0},{33621, 1, 1, 17}, {0, 0, 0, 0}}},{21907, {{33621, 1, 0, 0},{33621, 1, 1, 18}, {0, 0, 0, 0}}},{21907, {{33621, 1, 0, 0},{33621, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21908, {{33622, 1, 0, 0},{33622, 1, 1, 17}, {0, 0, 0, 0}}},{21908, {{33622, 1, 0, 0},{33622, 1, 1, 17}, {0, 0, 0, 0}}},{21909, {{33622, 1, 0, 0},{33622, 1, 1, 18}, {0, 0, 0, 0}}},{21909, {{33622, 1, 0, 0},{33622, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21910, {{33623, 1, 0, 0},{33623, 1, 1, 17}, {0, 0, 0, 0}}},{21910, {{33623, 1, 0, 0},{33623, 1, 1, 17}, {0, 0, 0, 0}}},{21911, {{33623, 1, 0, 0},{33623, 1, 1, 18}, {0, 0, 0, 0}}},{21911, {{33623, 1, 0, 0},{33623, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21913, {{33624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{21915, {{33634, 1, 0, 0},{33634, 1, 1, 18}, {0, 0, 0, 0}}},{21915, {{33634, 1, 0, 0},{33634, 1, 1, 18}, {0, 0, 0, 0}}},{21916, {{33634, 1, 0, 0},{33634, 1, 1, 19}, {0, 0, 0, 0}}},{21916, {{33634, 1, 0, 0},{33634, 1, 1, 19}, {0, 0, 0, 0}}},{21917, {{33634, 1, 0, 0},{33634, 1, 1, 17}, {0, 0, 0, 0}}},{21917, {{33634, 1, 0, 0},{33634, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{21918, {{33635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{21920, {{33648, 1, 0, 0},{33648, 1, 1, 17}, {0, 0, 0, 0}}},{21920, {{33648, 1, 0, 0},{33648, 1, 1, 17}, {0, 0, 0, 0}}},{21921, {{33648, 1, 0, 0},{33648, 1, 1, 18}, {0, 0, 0, 0}}},{21921, {{33648, 1, 0, 0},{33648, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21923, {{33649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21924, {{33650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21926, {{33651, 1, 0, 0},{33651, 1, 1, 17}, {0, 0, 0, 0}}},{21926, {{33651, 1, 0, 0},{33651, 1, 1, 17}, {0, 0, 0, 0}}},{21927, {{33651, 1, 0, 0},{33651, 1, 1, 18}, {0, 0, 0, 0}}},{21927, {{33651, 1, 0, 0},{33651, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21928, {{33652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{21929, {{33653, 1, 0, 0},{33653, 1, 1, 18}, {0, 0, 0, 0}}},{21929, {{33653, 1, 0, 0},{33653, 1, 1, 18}, {0, 0, 0, 0}}},{21930, {{33653, 1, 0, 0},{33653, 1, 1, 21}, {0, 0, 0, 0}}},{21930, {{33653, 1, 0, 0},{33653, 1, 1, 21}, {0, 0, 0, 0}}},{21931, {{33653, 1, 0, 0},{33653, 1, 1, 19}, {0, 0, 0, 0}}},{21931, {{33653, 1, 0, 0},{33653, 1, 1, 19}, {0, 0, 0, 0}}},{21932, {{33653, 1, 0, 0},{33653, 1, 1, 20}, {0, 0, 0, 0}}},{21932, {{33653, 1, 0, 0},{33653, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{21933, {{33654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21934, {{33655, 1, 0, 0},{33655, 1, 1, 17}, {0, 0, 0, 0}}},{21934, {{33655, 1, 0, 0},{33655, 1, 1, 17}, {0, 0, 0, 0}}},{21935, {{33655, 1, 0, 0},{33655, 1, 1, 18}, {0, 0, 0, 0}}},{21935, {{33655, 1, 0, 0},{33655, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21936, {{33656, 1, 0, 0},{33656, 1, 1, 17}, {0, 0, 0, 0}}},{21936, {{33656, 1, 0, 0},{33656, 1, 1, 17}, {0, 0, 0, 0}}},{21937, {{33656, 1, 0, 0},{33656, 1, 1, 18}, {0, 0, 0, 0}}},{21937, {{33656, 1, 0, 0},{33656, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21938, {{33657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21939, {{33658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{21940, {{33659, 1, 0, 0},{33659, 1, 1, 17}, {0, 0, 0, 0}}},{21940, {{33659, 1, 0, 0},{33659, 1, 1, 17}, {0, 0, 0, 0}}},{21941, {{33659, 1, 0, 0},{33659, 1, 1, 19}, {0, 0, 0, 0}}},{21941, {{33659, 1, 0, 0},{33659, 1, 1, 19}, {0, 0, 0, 0}}},{21942, {{33659, 1, 0, 0},{33659, 1, 1, 18}, {0, 0, 0, 0}}},{21942, {{33659, 1, 0, 0},{33659, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21943, {{33660, 1, 0, 0},{33660, 1, 1, 17}, {0, 0, 0, 0}}},{21943, {{33660, 1, 0, 0},{33660, 1, 1, 17}, {0, 0, 0, 0}}},{21944, {{33660, 1, 0, 0},{33660, 1, 1, 18}, {0, 0, 0, 0}}},{21944, {{33660, 1, 0, 0},{33660, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21945, {{33661, 1, 0, 0},{33661, 1, 1, 17}, {0, 0, 0, 0}}},{21945, {{33661, 1, 0, 0},{33661, 1, 1, 17}, {0, 0, 0, 0}}},{21946, {{33661, 1, 0, 0},{33661, 1, 1, 18}, {0, 0, 0, 0}}},{21946, {{33661, 1, 0, 0},{33661, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21947, {{33662, 1, 0, 0},{33662, 1, 1, 16}, {0, 0, 0, 0}}},{21947, {{33662, 1, 0, 0},{33662, 1, 1, 16}, {0, 0, 0, 0}}},{21948, {{33662, 1, 0, 0},{33662, 1, 1, 17}, {0, 0, 0, 0}}},{21948, {{33662, 1, 0, 0},{33662, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {10, {{21949, {{33663, 1, 0, 0},{33663, 1, 1, 21}, {0, 0, 0, 0}}},{21949, {{33663, 1, 0, 0},{33663, 1, 1, 21}, {0, 0, 0, 0}}},{21950, {{33663, 1, 0, 0},{33663, 1, 1, 19}, {0, 0, 0, 0}}},{21950, {{33663, 1, 0, 0},{33663, 1, 1, 19}, {0, 0, 0, 0}}},{21951, {{33663, 1, 0, 0},{33663, 1, 1, 18}, {0, 0, 0, 0}}},{21951, {{33663, 1, 0, 0},{33663, 1, 1, 18}, {0, 0, 0, 0}}},{21952, {{33663, 1, 0, 0},{33663, 1, 1, 17}, {0, 0, 0, 0}}},{21952, {{33663, 1, 0, 0},{33663, 1, 1, 17}, {0, 0, 0, 0}}},{21953, {{33663, 1, 0, 0},{33663, 1, 1, 20}, {0, 0, 0, 0}}},{21953, {{33663, 1, 0, 0},{33663, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{21954, {{33664, 1, 0, 0},{33664, 1, 1, 17}, {0, 0, 0, 0}}},{21954, {{33664, 1, 0, 0},{33664, 1, 1, 17}, {0, 0, 0, 0}}},{21955, {{33664, 1, 0, 0},{33664, 1, 1, 18}, {0, 0, 0, 0}}},{21955, {{33664, 1, 0, 0},{33664, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21956, {{33665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21957, {{33666, 1, 0, 0},{33666, 1, 1, 17}, {0, 0, 0, 0}}},{21957, {{33666, 1, 0, 0},{33666, 1, 1, 17}, {0, 0, 0, 0}}},{21958, {{33666, 1, 0, 0},{33666, 1, 1, 18}, {0, 0, 0, 0}}},{21958, {{33666, 1, 0, 0},{33666, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21959, {{33667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21960, {{33668, 1, 0, 0},{33668, 1, 1, 17}, {0, 0, 0, 0}}},{21960, {{33668, 1, 0, 0},{33668, 1, 1, 17}, {0, 0, 0, 0}}},{21961, {{33668, 1, 0, 0},{33668, 1, 1, 18}, {0, 0, 0, 0}}},{21961, {{33668, 1, 0, 0},{33668, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21962, {{33669, 1, 0, 0},{33669, 1, 1, 17}, {0, 0, 0, 0}}},{21962, {{33669, 1, 0, 0},{33669, 1, 1, 17}, {0, 0, 0, 0}}},{21963, {{33669, 1, 0, 0},{33669, 1, 1, 18}, {0, 0, 0, 0}}},{21963, {{33669, 1, 0, 0},{33669, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21964, {{33670, 1, 0, 0},{33670, 1, 1, 17}, {0, 0, 0, 0}}},{21964, {{33670, 1, 0, 0},{33670, 1, 1, 17}, {0, 0, 0, 0}}},{21965, {{33670, 1, 0, 0},{33670, 1, 1, 18}, {0, 0, 0, 0}}},{21965, {{33670, 1, 0, 0},{33670, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21966, {{33671, 1, 0, 0},{33671, 1, 1, 17}, {0, 0, 0, 0}}},{21966, {{33671, 1, 0, 0},{33671, 1, 1, 17}, {0, 0, 0, 0}}},{21967, {{33671, 1, 0, 0},{33671, 1, 1, 18}, {0, 0, 0, 0}}},{21967, {{33671, 1, 0, 0},{33671, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21968, {{33672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21969, {{33673, 1, 0, 0},{33673, 1, 1, 17}, {0, 0, 0, 0}}},{21969, {{33673, 1, 0, 0},{33673, 1, 1, 17}, {0, 0, 0, 0}}},{21970, {{33673, 1, 0, 0},{33673, 1, 1, 18}, {0, 0, 0, 0}}},{21970, {{33673, 1, 0, 0},{33673, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21971, {{33674, 1, 0, 0},{33674, 1, 1, 18}, {0, 0, 0, 0}}},{21971, {{33674, 1, 0, 0},{33674, 1, 1, 18}, {0, 0, 0, 0}}},{21972, {{33674, 1, 0, 0},{33674, 1, 1, 17}, {0, 0, 0, 0}}},{21972, {{33674, 1, 0, 0},{33674, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{21973, {{33675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21975, {{33676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21976, {{33677, 1, 0, 0},{33677, 1, 1, 17}, {0, 0, 0, 0}}},{21976, {{33677, 1, 0, 0},{33677, 1, 1, 17}, {0, 0, 0, 0}}},{21977, {{33677, 1, 0, 0},{33677, 1, 1, 18}, {0, 0, 0, 0}}},{21977, {{33677, 1, 0, 0},{33677, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21978, {{33678, 1, 0, 0},{33678, 1, 1, 17}, {0, 0, 0, 0}}},{21978, {{33678, 1, 0, 0},{33678, 1, 1, 17}, {0, 0, 0, 0}}},{21979, {{33678, 1, 0, 0},{33678, 1, 1, 18}, {0, 0, 0, 0}}},{21979, {{33678, 1, 0, 0},{33678, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21980, {{33679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21981, {{33680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21982, {{33681, 1, 0, 0},{33681, 1, 1, 16}, {0, 0, 0, 0}}},{21982, {{33681, 1, 0, 0},{33681, 1, 1, 16}, {0, 0, 0, 0}}},{21983, {{33681, 1, 0, 0},{33681, 1, 1, 17}, {0, 0, 0, 0}}},{21983, {{33681, 1, 0, 0},{33681, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{21984, {{33682, 1, 0, 0},{33682, 1, 1, 17}, {0, 0, 0, 0}}},{21984, {{33682, 1, 0, 0},{33682, 1, 1, 17}, {0, 0, 0, 0}}},{21985, {{33682, 1, 0, 0},{33682, 1, 1, 18}, {0, 0, 0, 0}}},{21985, {{33682, 1, 0, 0},{33682, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{21986, {{33683, 1, 0, 0},{33683, 1, 1, 17}, {0, 0, 0, 0}}},{21986, {{33683, 1, 0, 0},{33683, 1, 1, 17}, {0, 0, 0, 0}}},{21987, {{33683, 1, 0, 0},{33683, 1, 1, 18}, {0, 0, 0, 0}}},{21987, {{33683, 1, 0, 0},{33683, 1, 1, 18}, {0, 0, 0, 0}}},{21988, {{33683, 1, 0, 0},{33683, 1, 1, 19}, {0, 0, 0, 0}}},{21988, {{33683, 1, 0, 0},{33683, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{21989, {{33684, 1, 0, 0},{33684, 1, 1, 17}, {0, 0, 0, 0}}},{21989, {{33684, 1, 0, 0},{33684, 1, 1, 17}, {0, 0, 0, 0}}},{21990, {{33684, 1, 0, 0},{33684, 1, 1, 18}, {0, 0, 0, 0}}},{21990, {{33684, 1, 0, 0},{33684, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21991, {{33685, 1, 0, 0},{33685, 1, 1, 17}, {0, 0, 0, 0}}},{21991, {{33685, 1, 0, 0},{33685, 1, 1, 17}, {0, 0, 0, 0}}},{21992, {{33685, 1, 0, 0},{33685, 1, 1, 18}, {0, 0, 0, 0}}},{21992, {{33685, 1, 0, 0},{33685, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{21993, {{33686, 1, 0, 0},{33686, 1, 1, 17}, {0, 0, 0, 0}}},{21993, {{33686, 1, 0, 0},{33686, 1, 1, 17}, {0, 0, 0, 0}}},{21994, {{33686, 1, 0, 0},{33686, 1, 1, 19}, {0, 0, 0, 0}}},{21994, {{33686, 1, 0, 0},{33686, 1, 1, 19}, {0, 0, 0, 0}}},{21995, {{33686, 1, 0, 0},{33686, 1, 1, 18}, {0, 0, 0, 0}}},{21995, {{33686, 1, 0, 0},{33686, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{21996, {{33687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21997, {{33688, 1, 0, 0},{33688, 1, 1, 17}, {0, 0, 0, 0}}},{21997, {{33688, 1, 0, 0},{33688, 1, 1, 17}, {0, 0, 0, 0}}},{21998, {{33688, 1, 0, 0},{33688, 1, 1, 18}, {0, 0, 0, 0}}},{21998, {{33688, 1, 0, 0},{33688, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{21999, {{33689, 1, 0, 0},{33689, 1, 1, 17}, {0, 0, 0, 0}}},{21999, {{33689, 1, 0, 0},{33689, 1, 1, 17}, {0, 0, 0, 0}}},{22000, {{33689, 1, 0, 0},{33689, 1, 1, 18}, {0, 0, 0, 0}}},{22000, {{33689, 1, 0, 0},{33689, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{22001, {{33690, 1, 0, 0},{33690, 1, 1, 17}, {0, 0, 0, 0}}},{22001, {{33690, 1, 0, 0},{33690, 1, 1, 17}, {0, 0, 0, 0}}},{22002, {{33690, 1, 0, 0},{33690, 1, 1, 19}, {0, 0, 0, 0}}},{22002, {{33690, 1, 0, 0},{33690, 1, 1, 19}, {0, 0, 0, 0}}},{22003, {{33690, 1, 0, 0},{33690, 1, 1, 18}, {0, 0, 0, 0}}},{22003, {{33690, 1, 0, 0},{33690, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{22004, {{33691, 1, 0, 0},{33691, 1, 1, 19}, {0, 0, 0, 0}}},{22004, {{33691, 1, 0, 0},{33691, 1, 1, 19}, {0, 0, 0, 0}}},{22005, {{33691, 1, 0, 0},{33691, 1, 1, 17}, {0, 0, 0, 0}}},{22005, {{33691, 1, 0, 0},{33691, 1, 1, 17}, {0, 0, 0, 0}}},{22006, {{33691, 1, 0, 0},{33691, 1, 1, 18}, {0, 0, 0, 0}}},{22006, {{33691, 1, 0, 0},{33691, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22007, {{33692, 1, 0, 0},{33692, 1, 1, 17}, {0, 0, 0, 0}}},{22007, {{33692, 1, 0, 0},{33692, 1, 1, 17}, {0, 0, 0, 0}}},{22008, {{33692, 1, 0, 0},{33692, 1, 1, 18}, {0, 0, 0, 0}}},{22008, {{33692, 1, 0, 0},{33692, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22009, {{33693, 1, 0, 0},{33693, 1, 1, 17}, {0, 0, 0, 0}}},{22009, {{33693, 1, 0, 0},{33693, 1, 1, 17}, {0, 0, 0, 0}}},{22010, {{33693, 1, 0, 0},{33693, 1, 1, 18}, {0, 0, 0, 0}}},{22010, {{33693, 1, 0, 0},{33693, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22011, {{33694, 1, 0, 0},{33694, 1, 1, 17}, {0, 0, 0, 0}}},{22011, {{33694, 1, 0, 0},{33694, 1, 1, 17}, {0, 0, 0, 0}}},{22012, {{33694, 1, 0, 0},{33694, 1, 1, 18}, {0, 0, 0, 0}}},{22012, {{33694, 1, 0, 0},{33694, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22013, {{33695, 1, 0, 0},{33695, 1, 1, 17}, {0, 0, 0, 0}}},{22013, {{33695, 1, 0, 0},{33695, 1, 1, 17}, {0, 0, 0, 0}}},{22014, {{33695, 1, 0, 0},{33695, 1, 1, 18}, {0, 0, 0, 0}}},{22014, {{33695, 1, 0, 0},{33695, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22015, {{33696, 1, 0, 0},{33696, 1, 1, 17}, {0, 0, 0, 0}}},{22015, {{33696, 1, 0, 0},{33696, 1, 1, 17}, {0, 0, 0, 0}}},{22016, {{33696, 1, 0, 0},{33696, 1, 1, 18}, {0, 0, 0, 0}}},{22016, {{33696, 1, 0, 0},{33696, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22017, {{33697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22018, {{33698, 1, 0, 0},{33698, 1, 1, 17}, {0, 0, 0, 0}}},{22018, {{33698, 1, 0, 0},{33698, 1, 1, 17}, {0, 0, 0, 0}}},{22019, {{33698, 1, 0, 0},{33698, 1, 1, 18}, {0, 0, 0, 0}}},{22019, {{33698, 1, 0, 0},{33698, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{22021, {{33700, 1, 0, 0},{33700, 1, 1, 16}, {0, 0, 0, 0}}},{22021, {{33700, 1, 0, 0},{33700, 1, 1, 16}, {0, 0, 0, 0}}},{22022, {{33700, 1, 0, 0},{33700, 1, 1, 17}, {0, 0, 0, 0}}},{22022, {{33700, 1, 0, 0},{33700, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{22023, {{33701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22024, {{33702, 1, 0, 0},{33702, 1, 1, 17}, {0, 0, 0, 0}}},{22024, {{33702, 1, 0, 0},{33702, 1, 1, 17}, {0, 0, 0, 0}}},{22025, {{33702, 1, 0, 0},{33702, 1, 1, 18}, {0, 0, 0, 0}}},{22025, {{33702, 1, 0, 0},{33702, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22026, {{33703, 1, 0, 0},{33703, 1, 1, 17}, {0, 0, 0, 0}}},{22026, {{33703, 1, 0, 0},{33703, 1, 1, 17}, {0, 0, 0, 0}}},{22027, {{33703, 1, 0, 0},{33703, 1, 1, 18}, {0, 0, 0, 0}}},{22027, {{33703, 1, 0, 0},{33703, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22028, {{33704, 1, 0, 0},{33704, 1, 1, 17}, {0, 0, 0, 0}}},{22028, {{33704, 1, 0, 0},{33704, 1, 1, 17}, {0, 0, 0, 0}}},{22029, {{33704, 1, 0, 0},{33704, 1, 1, 18}, {0, 0, 0, 0}}},{22029, {{33704, 1, 0, 0},{33704, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22030, {{33705, 1, 0, 0},{33705, 1, 1, 17}, {0, 0, 0, 0}}},{22030, {{33705, 1, 0, 0},{33705, 1, 1, 17}, {0, 0, 0, 0}}},{22031, {{33705, 1, 0, 0},{33705, 1, 1, 18}, {0, 0, 0, 0}}},{22031, {{33705, 1, 0, 0},{33705, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22032, {{33706, 1, 0, 0},{33706, 1, 1, 17}, {0, 0, 0, 0}}},{22032, {{33706, 1, 0, 0},{33706, 1, 1, 17}, {0, 0, 0, 0}}},{22033, {{33706, 1, 0, 0},{33706, 1, 1, 18}, {0, 0, 0, 0}}},{22033, {{33706, 1, 0, 0},{33706, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22034, {{33707, 1, 0, 0},{33707, 1, 1, 17}, {0, 0, 0, 0}}},{22034, {{33707, 1, 0, 0},{33707, 1, 1, 17}, {0, 0, 0, 0}}},{22035, {{33707, 1, 0, 0},{33707, 1, 1, 18}, {0, 0, 0, 0}}},{22035, {{33707, 1, 0, 0},{33707, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22036, {{33708, 1, 0, 0},{33708, 1, 1, 17}, {0, 0, 0, 0}}},{22036, {{33708, 1, 0, 0},{33708, 1, 1, 17}, {0, 0, 0, 0}}},{22037, {{33708, 1, 0, 0},{33708, 1, 1, 18}, {0, 0, 0, 0}}},{22037, {{33708, 1, 0, 0},{33708, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{22038, {{33709, 1, 0, 0},{33709, 1, 1, 17}, {0, 0, 0, 0}}},{22038, {{33709, 1, 0, 0},{33709, 1, 1, 17}, {0, 0, 0, 0}}},{22039, {{33709, 1, 0, 0},{33709, 1, 1, 19}, {0, 0, 0, 0}}},{22039, {{33709, 1, 0, 0},{33709, 1, 1, 19}, {0, 0, 0, 0}}},{22040, {{33709, 1, 0, 0},{33709, 1, 1, 18}, {0, 0, 0, 0}}},{22040, {{33709, 1, 0, 0},{33709, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22042, {{33710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22043, {{33711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{22044, {{33713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {8, {{22046, {{33717, 1, 0, 0},{33717, 1, 1, 17}, {0, 0, 0, 0}}},{22046, {{33717, 1, 0, 0},{33717, 1, 1, 17}, {0, 0, 0, 0}}},{22047, {{33717, 1, 0, 0},{33717, 1, 1, 19}, {0, 0, 0, 0}}},{22047, {{33717, 1, 0, 0},{33717, 1, 1, 19}, {0, 0, 0, 0}}},{22048, {{33717, 1, 0, 0},{33717, 1, 1, 18}, {0, 0, 0, 0}}},{22048, {{33717, 1, 0, 0},{33717, 1, 1, 18}, {0, 0, 0, 0}}},{68065, {{33717, 1, 0, 0},{33717, 1, 1, 20}, {0, 0, 0, 0}}},{68065, {{33717, 1, 0, 0},{33717, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{68044, {{33723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {12, {{22050, {{33725, 1, 0, 0},{33725, 1, 1, 18}, {0, 0, 0, 0}}},{22050, {{33725, 1, 0, 0},{33725, 1, 1, 18}, {0, 0, 0, 0}}},{22051, {{33725, 1, 0, 0},{33725, 1, 1, 20}, {0, 0, 0, 0}}},{22051, {{33725, 1, 0, 0},{33725, 1, 1, 20}, {0, 0, 0, 0}}},{22052, {{33725, 1, 0, 0},{33725, 1, 1, 19}, {0, 0, 0, 0}}},{22052, {{33725, 1, 0, 0},{33725, 1, 1, 19}, {0, 0, 0, 0}}},{22053, {{33725, 1, 0, 0},{33725, 1, 1, 22}, {0, 0, 0, 0}}},{22053, {{33725, 1, 0, 0},{33725, 1, 1, 22}, {0, 0, 0, 0}}},{22054, {{33725, 1, 0, 0},{33725, 1, 1, 21}, {0, 0, 0, 0}}},{22054, {{33725, 1, 0, 0},{33725, 1, 1, 21}, {0, 0, 0, 0}}},{58462, {{33725, 1, 0, 0},{33725, 1, 1, 23}, {0, 0, 0, 0}}},{58462, {{33725, 1, 0, 0},{33725, 1, 1, 23}, {0, 0, 0, 0}}}}},
    {4, {{22055, {{33726, 1, 0, 0},{33726, 1, 1, 17}, {0, 0, 0, 0}}},{22055, {{33726, 1, 0, 0},{33726, 1, 1, 17}, {0, 0, 0, 0}}},{22056, {{33726, 1, 0, 0},{33726, 1, 1, 18}, {0, 0, 0, 0}}},{22056, {{33726, 1, 0, 0},{33726, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22057, {{33727, 1, 0, 0},{33727, 1, 1, 17}, {0, 0, 0, 0}}},{22057, {{33727, 1, 0, 0},{33727, 1, 1, 17}, {0, 0, 0, 0}}},{22058, {{33727, 1, 0, 0},{33727, 1, 1, 18}, {0, 0, 0, 0}}},{22058, {{33727, 1, 0, 0},{33727, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22059, {{33728, 1, 0, 0},{33728, 1, 1, 17}, {0, 0, 0, 0}}},{22059, {{33728, 1, 0, 0},{33728, 1, 1, 17}, {0, 0, 0, 0}}},{22060, {{33728, 1, 0, 0},{33728, 1, 1, 18}, {0, 0, 0, 0}}},{22060, {{33728, 1, 0, 0},{33728, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {8, {{22061, {{33729, 1, 0, 0},{33729, 1, 1, 17}, {0, 0, 0, 0}}},{22061, {{33729, 1, 0, 0},{33729, 1, 1, 17}, {0, 0, 0, 0}}},{22062, {{33729, 1, 0, 0},{33729, 1, 1, 20}, {0, 0, 0, 0}}},{22062, {{33729, 1, 0, 0},{33729, 1, 1, 20}, {0, 0, 0, 0}}},{22063, {{33729, 1, 0, 0},{33729, 1, 1, 19}, {0, 0, 0, 0}}},{22063, {{33729, 1, 0, 0},{33729, 1, 1, 19}, {0, 0, 0, 0}}},{22064, {{33729, 1, 0, 0},{33729, 1, 1, 18}, {0, 0, 0, 0}}},{22064, {{33729, 1, 0, 0},{33729, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22066, {{33730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22068, {{33731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22069, {{33732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22070, {{33733, 1, 0, 0},{33733, 1, 1, 17}, {0, 0, 0, 0}}},{22070, {{33733, 1, 0, 0},{33733, 1, 1, 17}, {0, 0, 0, 0}}},{22071, {{33733, 1, 0, 0},{33733, 1, 1, 18}, {0, 0, 0, 0}}},{22071, {{33733, 1, 0, 0},{33733, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{22072, {{33734, 1, 0, 0},{33734, 1, 1, 16}, {0, 0, 0, 0}}},{22072, {{33734, 1, 0, 0},{33734, 1, 1, 16}, {0, 0, 0, 0}}},{22073, {{33734, 1, 0, 0},{33734, 1, 1, 18}, {0, 0, 0, 0}}},{22073, {{33734, 1, 0, 0},{33734, 1, 1, 18}, {0, 0, 0, 0}}},{22074, {{33734, 1, 0, 0},{33734, 1, 1, 17}, {0, 0, 0, 0}}},{22074, {{33734, 1, 0, 0},{33734, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{22075, {{33735, 1, 0, 0},{33735, 1, 1, 17}, {0, 0, 0, 0}}},{22075, {{33735, 1, 0, 0},{33735, 1, 1, 17}, {0, 0, 0, 0}}},{22076, {{33735, 1, 0, 0},{33735, 1, 1, 18}, {0, 0, 0, 0}}},{22076, {{33735, 1, 0, 0},{33735, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22077, {{33736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22078, {{33737, 1, 0, 0},{33737, 1, 1, 17}, {0, 0, 0, 0}}},{22078, {{33737, 1, 0, 0},{33737, 1, 1, 17}, {0, 0, 0, 0}}},{22079, {{33737, 1, 0, 0},{33737, 1, 1, 18}, {0, 0, 0, 0}}},{22079, {{33737, 1, 0, 0},{33737, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22080, {{33738, 1, 0, 0},{33738, 1, 1, 17}, {0, 0, 0, 0}}},{22080, {{33738, 1, 0, 0},{33738, 1, 1, 17}, {0, 0, 0, 0}}},{22081, {{33738, 1, 0, 0},{33738, 1, 1, 18}, {0, 0, 0, 0}}},{22081, {{33738, 1, 0, 0},{33738, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22082, {{33739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22083, {{33740, 1, 0, 0},{33740, 1, 1, 17}, {0, 0, 0, 0}}},{22083, {{33740, 1, 0, 0},{33740, 1, 1, 17}, {0, 0, 0, 0}}},{22084, {{33740, 1, 0, 0},{33740, 1, 1, 18}, {0, 0, 0, 0}}},{22084, {{33740, 1, 0, 0},{33740, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22085, {{33741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22086, {{33742, 1, 0, 0},{33742, 1, 1, 17}, {0, 0, 0, 0}}},{22086, {{33742, 1, 0, 0},{33742, 1, 1, 17}, {0, 0, 0, 0}}},{22087, {{33742, 1, 0, 0},{33742, 1, 1, 18}, {0, 0, 0, 0}}},{22087, {{33742, 1, 0, 0},{33742, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22088, {{33743, 1, 0, 0},{33743, 1, 1, 17}, {0, 0, 0, 0}}},{22088, {{33743, 1, 0, 0},{33743, 1, 1, 17}, {0, 0, 0, 0}}},{22089, {{33743, 1, 0, 0},{33743, 1, 1, 18}, {0, 0, 0, 0}}},{22089, {{33743, 1, 0, 0},{33743, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22090, {{33744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22091, {{33745, 1, 0, 0},{33745, 1, 1, 17}, {0, 0, 0, 0}}},{22091, {{33745, 1, 0, 0},{33745, 1, 1, 17}, {0, 0, 0, 0}}},{22092, {{33745, 1, 0, 0},{33745, 1, 1, 18}, {0, 0, 0, 0}}},{22092, {{33745, 1, 0, 0},{33745, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22093, {{33746, 1, 0, 0},{33746, 1, 1, 16}, {0, 0, 0, 0}}},{22093, {{33746, 1, 0, 0},{33746, 1, 1, 16}, {0, 0, 0, 0}}},{22094, {{33746, 1, 0, 0},{33746, 1, 1, 17}, {0, 0, 0, 0}}},{22094, {{33746, 1, 0, 0},{33746, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{22095, {{33747, 1, 0, 0},{33747, 1, 1, 18}, {0, 0, 0, 0}}},{22095, {{33747, 1, 0, 0},{33747, 1, 1, 18}, {0, 0, 0, 0}}},{22096, {{33747, 1, 0, 0},{33747, 1, 1, 19}, {0, 0, 0, 0}}},{22096, {{33747, 1, 0, 0},{33747, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{22097, {{33748, 1, 0, 0},{33748, 1, 1, 19}, {0, 0, 0, 0}}},{22097, {{33748, 1, 0, 0},{33748, 1, 1, 19}, {0, 0, 0, 0}}},{22098, {{33748, 1, 0, 0},{33748, 1, 1, 18}, {0, 0, 0, 0}}},{22098, {{33748, 1, 0, 0},{33748, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22099, {{33749, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22100, {{33750, 1, 0, 0},{33750, 1, 1, 17}, {0, 0, 0, 0}}},{22100, {{33750, 1, 0, 0},{33750, 1, 1, 17}, {0, 0, 0, 0}}},{22101, {{33750, 1, 0, 0},{33750, 1, 1, 18}, {0, 0, 0, 0}}},{22101, {{33750, 1, 0, 0},{33750, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22102, {{33751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{22103, {{33752, 1, 0, 0},{33752, 1, 1, 17}, {0, 0, 0, 0}}},{22103, {{33752, 1, 0, 0},{33752, 1, 1, 17}, {0, 0, 0, 0}}},{22104, {{33752, 1, 0, 0},{33752, 1, 1, 18}, {0, 0, 0, 0}}},{22104, {{33752, 1, 0, 0},{33752, 1, 1, 18}, {0, 0, 0, 0}}},{22105, {{33752, 1, 0, 0},{33752, 1, 1, 19}, {0, 0, 0, 0}}},{22105, {{33752, 1, 0, 0},{33752, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{22106, {{33753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22107, {{33754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22108, {{33755, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{22109, {{33756, 1, 0, 0},{33756, 1, 1, 18}, {0, 0, 0, 0}}},{22109, {{33756, 1, 0, 0},{33756, 1, 1, 18}, {0, 0, 0, 0}}},{22110, {{33756, 1, 0, 0},{33756, 1, 1, 19}, {0, 0, 0, 0}}},{22110, {{33756, 1, 0, 0},{33756, 1, 1, 19}, {0, 0, 0, 0}}},{22111, {{33756, 1, 0, 0},{33756, 1, 1, 20}, {0, 0, 0, 0}}},{22111, {{33756, 1, 0, 0},{33756, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{22113, {{33757, 1, 0, 0},{33757, 1, 1, 17}, {0, 0, 0, 0}}},{22113, {{33757, 1, 0, 0},{33757, 1, 1, 17}, {0, 0, 0, 0}}},{22114, {{33757, 1, 0, 0},{33757, 1, 1, 18}, {0, 0, 0, 0}}},{22114, {{33757, 1, 0, 0},{33757, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22115, {{33758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {18, {{22116, {{33759, 1, 0, 0},{33759, 1, 1, 20}, {0, 0, 0, 0}}},{22116, {{33759, 1, 0, 0},{33759, 1, 1, 20}, {0, 0, 0, 0}}},{22117, {{33759, 1, 0, 0},{33759, 1, 1, 25}, {0, 0, 0, 0}}},{22117, {{33759, 1, 0, 0},{33759, 1, 1, 25}, {0, 0, 0, 0}}},{22118, {{33759, 1, 0, 0},{33759, 1, 1, 22}, {0, 0, 0, 0}}},{22118, {{33759, 1, 0, 0},{33759, 1, 1, 22}, {0, 0, 0, 0}}},{22119, {{33759, 1, 0, 0},{33759, 1, 1, 24}, {0, 0, 0, 0}}},{22119, {{33759, 1, 0, 0},{33759, 1, 1, 24}, {0, 0, 0, 0}}},{22120, {{33759, 1, 0, 0},{33759, 1, 1, 21}, {0, 0, 0, 0}}},{22120, {{33759, 1, 0, 0},{33759, 1, 1, 21}, {0, 0, 0, 0}}},{22121, {{33759, 1, 0, 0},{33759, 1, 1, 23}, {0, 0, 0, 0}}},{22121, {{33759, 1, 0, 0},{33759, 1, 1, 23}, {0, 0, 0, 0}}},{22122, {{33759, 1, 0, 0},{33759, 1, 1, 26}, {0, 0, 0, 0}}},{22122, {{33759, 1, 0, 0},{33759, 1, 1, 26}, {0, 0, 0, 0}}},{22124, {{33759, 1, 0, 0},{33759, 1, 1, 27}, {0, 0, 0, 0}}},{22124, {{33759, 1, 0, 0},{33759, 1, 1, 27}, {0, 0, 0, 0}}},{22125, {{33759, 1, 0, 0},{33759, 1, 1, 28}, {0, 0, 0, 0}}},{22125, {{33759, 1, 0, 0},{33759, 1, 1, 28}, {0, 0, 0, 0}}}}},
    {4, {{22126, {{33760, 1, 0, 0},{33760, 1, 1, 17}, {0, 0, 0, 0}}},{22126, {{33760, 1, 0, 0},{33760, 1, 1, 17}, {0, 0, 0, 0}}},{22127, {{33760, 1, 0, 0},{33760, 1, 1, 18}, {0, 0, 0, 0}}},{22127, {{33760, 1, 0, 0},{33760, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{22128, {{33761, 1, 0, 0},{33761, 1, 1, 17}, {0, 0, 0, 0}}},{22128, {{33761, 1, 0, 0},{33761, 1, 1, 17}, {0, 0, 0, 0}}},{22129, {{33761, 1, 0, 0},{33761, 1, 1, 18}, {0, 0, 0, 0}}},{22129, {{33761, 1, 0, 0},{33761, 1, 1, 18}, {0, 0, 0, 0}}},{58493, {{33761, 1, 0, 0},{33761, 1, 1, 19}, {0, 0, 0, 0}}},{58493, {{33761, 1, 0, 0},{33761, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{22130, {{33762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22131, {{33763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22132, {{33764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22133, {{33765, 1, 0, 0},{33765, 1, 1, 17}, {0, 0, 0, 0}}},{22133, {{33765, 1, 0, 0},{33765, 1, 1, 17}, {0, 0, 0, 0}}},{22134, {{33765, 1, 0, 0},{33765, 1, 1, 18}, {0, 0, 0, 0}}},{22134, {{33765, 1, 0, 0},{33765, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22135, {{33766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22136, {{33767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22137, {{33768, 1, 0, 0},{33768, 1, 1, 17}, {0, 0, 0, 0}}},{22137, {{33768, 1, 0, 0},{33768, 1, 1, 17}, {0, 0, 0, 0}}},{22138, {{33768, 1, 0, 0},{33768, 1, 1, 18}, {0, 0, 0, 0}}},{22138, {{33768, 1, 0, 0},{33768, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22139, {{33769, 1, 0, 0},{33769, 1, 1, 17}, {0, 0, 0, 0}}},{22139, {{33769, 1, 0, 0},{33769, 1, 1, 17}, {0, 0, 0, 0}}},{22140, {{33769, 1, 0, 0},{33769, 1, 1, 18}, {0, 0, 0, 0}}},{22140, {{33769, 1, 0, 0},{33769, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22141, {{33770, 1, 0, 0},{33770, 1, 1, 17}, {0, 0, 0, 0}}},{22141, {{33770, 1, 0, 0},{33770, 1, 1, 17}, {0, 0, 0, 0}}},{22142, {{33770, 1, 0, 0},{33770, 1, 1, 18}, {0, 0, 0, 0}}},{22142, {{33770, 1, 0, 0},{33770, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22143, {{33771, 1, 0, 0},{33771, 1, 1, 17}, {0, 0, 0, 0}}},{22143, {{33771, 1, 0, 0},{33771, 1, 1, 17}, {0, 0, 0, 0}}},{22144, {{33771, 1, 0, 0},{33771, 1, 1, 18}, {0, 0, 0, 0}}},{22144, {{33771, 1, 0, 0},{33771, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22145, {{33772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22146, {{33773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22148, {{33774, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{22150, {{33775, 1, 0, 0},{33775, 1, 1, 17}, {0, 0, 0, 0}}},{22150, {{33775, 1, 0, 0},{33775, 1, 1, 17}, {0, 0, 0, 0}}},{22151, {{33775, 1, 0, 0},{33775, 1, 1, 19}, {0, 0, 0, 0}}},{22151, {{33775, 1, 0, 0},{33775, 1, 1, 19}, {0, 0, 0, 0}}},{22152, {{33775, 1, 0, 0},{33775, 1, 1, 18}, {0, 0, 0, 0}}},{22152, {{33775, 1, 0, 0},{33775, 1, 1, 18}, {0, 0, 0, 0}}},{68053, {{33775, 1, 0, 0},{33775, 1, 1, 20}, {0, 0, 0, 0}}},{68053, {{33775, 1, 0, 0},{33775, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {6, {{22154, {{33776, 1, 0, 0},{33776, 1, 1, 17}, {0, 0, 0, 0}}},{22154, {{33776, 1, 0, 0},{33776, 1, 1, 17}, {0, 0, 0, 0}}},{22155, {{33776, 1, 0, 0},{33776, 1, 1, 19}, {0, 0, 0, 0}}},{22155, {{33776, 1, 0, 0},{33776, 1, 1, 19}, {0, 0, 0, 0}}},{22156, {{33776, 1, 0, 0},{33776, 1, 1, 18}, {0, 0, 0, 0}}},{22156, {{33776, 1, 0, 0},{33776, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {8, {{22157, {{33777, 1, 0, 0},{33777, 1, 1, 17}, {0, 0, 0, 0}}},{22157, {{33777, 1, 0, 0},{33777, 1, 1, 17}, {0, 0, 0, 0}}},{22158, {{33777, 1, 0, 0},{33777, 1, 1, 20}, {0, 0, 0, 0}}},{22158, {{33777, 1, 0, 0},{33777, 1, 1, 20}, {0, 0, 0, 0}}},{22159, {{33777, 1, 0, 0},{33777, 1, 1, 19}, {0, 0, 0, 0}}},{22159, {{33777, 1, 0, 0},{33777, 1, 1, 19}, {0, 0, 0, 0}}},{22160, {{33777, 1, 0, 0},{33777, 1, 1, 18}, {0, 0, 0, 0}}},{22160, {{33777, 1, 0, 0},{33777, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{22161, {{33778, 1, 0, 0},{33778, 1, 1, 20}, {0, 0, 0, 0}}},{22161, {{33778, 1, 0, 0},{33778, 1, 1, 20}, {0, 0, 0, 0}}},{22162, {{33778, 1, 0, 0},{33778, 1, 1, 19}, {0, 0, 0, 0}}},{22162, {{33778, 1, 0, 0},{33778, 1, 1, 19}, {0, 0, 0, 0}}},{22163, {{33778, 1, 0, 0},{33778, 1, 1, 21}, {0, 0, 0, 0}}},{22163, {{33778, 1, 0, 0},{33778, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {4, {{22164, {{33779, 1, 0, 0},{33779, 1, 1, 16}, {0, 0, 0, 0}}},{22164, {{33779, 1, 0, 0},{33779, 1, 1, 16}, {0, 0, 0, 0}}},{22165, {{33779, 1, 0, 0},{33779, 1, 1, 17}, {0, 0, 0, 0}}},{22165, {{33779, 1, 0, 0},{33779, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{22166, {{33780, 1, 0, 0},{33780, 1, 1, 17}, {0, 0, 0, 0}}},{22166, {{33780, 1, 0, 0},{33780, 1, 1, 17}, {0, 0, 0, 0}}},{22167, {{33780, 1, 0, 0},{33780, 1, 1, 18}, {0, 0, 0, 0}}},{22167, {{33780, 1, 0, 0},{33780, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22168, {{33781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22169, {{33782, 1, 0, 0},{33782, 1, 1, 17}, {0, 0, 0, 0}}},{22169, {{33782, 1, 0, 0},{33782, 1, 1, 17}, {0, 0, 0, 0}}},{22170, {{33782, 1, 0, 0},{33782, 1, 1, 18}, {0, 0, 0, 0}}},{22170, {{33782, 1, 0, 0},{33782, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{22171, {{33783, 1, 0, 0},{33783, 1, 1, 17}, {0, 0, 0, 0}}},{22171, {{33783, 1, 0, 0},{33783, 1, 1, 17}, {0, 0, 0, 0}}},{22172, {{33783, 1, 0, 0},{33783, 1, 1, 18}, {0, 0, 0, 0}}},{22172, {{33783, 1, 0, 0},{33783, 1, 1, 18}, {0, 0, 0, 0}}},{22173, {{33783, 1, 0, 0},{33783, 1, 1, 19}, {0, 0, 0, 0}}},{22173, {{33783, 1, 0, 0},{33783, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{22174, {{33784, 1, 0, 0},{33784, 1, 1, 18}, {0, 0, 0, 0}}},{22174, {{33784, 1, 0, 0},{33784, 1, 1, 18}, {0, 0, 0, 0}}},{22175, {{33784, 1, 0, 0},{33784, 1, 1, 17}, {0, 0, 0, 0}}},{22175, {{33784, 1, 0, 0},{33784, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{22177, {{33785, 1, 0, 0},{33785, 1, 1, 17}, {0, 0, 0, 0}}},{22177, {{33785, 1, 0, 0},{33785, 1, 1, 17}, {0, 0, 0, 0}}},{22178, {{33785, 1, 0, 0},{33785, 1, 1, 18}, {0, 0, 0, 0}}},{22178, {{33785, 1, 0, 0},{33785, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22179, {{33786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22180, {{33787, 1, 0, 0},{33787, 1, 1, 17}, {0, 0, 0, 0}}},{22180, {{33787, 1, 0, 0},{33787, 1, 1, 17}, {0, 0, 0, 0}}},{22181, {{33787, 1, 0, 0},{33787, 1, 1, 18}, {0, 0, 0, 0}}},{22181, {{33787, 1, 0, 0},{33787, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22182, {{33788, 1, 0, 0},{33788, 1, 1, 17}, {0, 0, 0, 0}}},{22182, {{33788, 1, 0, 0},{33788, 1, 1, 17}, {0, 0, 0, 0}}},{22183, {{33788, 1, 0, 0},{33788, 1, 1, 18}, {0, 0, 0, 0}}},{22183, {{33788, 1, 0, 0},{33788, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22184, {{33789, 1, 0, 0},{33789, 1, 1, 17}, {0, 0, 0, 0}}},{22184, {{33789, 1, 0, 0},{33789, 1, 1, 17}, {0, 0, 0, 0}}},{22185, {{33789, 1, 0, 0},{33789, 1, 1, 18}, {0, 0, 0, 0}}},{22185, {{33789, 1, 0, 0},{33789, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22186, {{33790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22189, {{33791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22190, {{33792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22191, {{33793, 1, 0, 0},{33793, 1, 1, 17}, {0, 0, 0, 0}}},{22191, {{33793, 1, 0, 0},{33793, 1, 1, 17}, {0, 0, 0, 0}}},{22192, {{33793, 1, 0, 0},{33793, 1, 1, 18}, {0, 0, 0, 0}}},{22192, {{33793, 1, 0, 0},{33793, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22193, {{33794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{22194, {{33795, 1, 0, 0},{33795, 1, 1, 17}, {0, 0, 0, 0}}},{22194, {{33795, 1, 0, 0},{33795, 1, 1, 17}, {0, 0, 0, 0}}},{22195, {{33795, 1, 0, 0},{33795, 1, 1, 19}, {0, 0, 0, 0}}},{22195, {{33795, 1, 0, 0},{33795, 1, 1, 19}, {0, 0, 0, 0}}},{22196, {{33795, 1, 0, 0},{33795, 1, 1, 18}, {0, 0, 0, 0}}},{22196, {{33795, 1, 0, 0},{33795, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{22198, {{33796, 1, 0, 0},{33796, 1, 1, 17}, {0, 0, 0, 0}}},{22198, {{33796, 1, 0, 0},{33796, 1, 1, 17}, {0, 0, 0, 0}}},{22199, {{33796, 1, 0, 0},{33796, 1, 1, 18}, {0, 0, 0, 0}}},{22199, {{33796, 1, 0, 0},{33796, 1, 1, 18}, {0, 0, 0, 0}}},{58494, {{33796, 1, 0, 0},{33796, 1, 1, 19}, {0, 0, 0, 0}}},{58494, {{33796, 1, 0, 0},{33796, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{22200, {{33797, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22201, {{33798, 1, 0, 0},{33798, 1, 1, 17}, {0, 0, 0, 0}}},{22201, {{33798, 1, 0, 0},{33798, 1, 1, 17}, {0, 0, 0, 0}}},{22202, {{33798, 1, 0, 0},{33798, 1, 1, 18}, {0, 0, 0, 0}}},{22202, {{33798, 1, 0, 0},{33798, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22203, {{33799, 1, 0, 0},{33799, 1, 1, 17}, {0, 0, 0, 0}}},{22203, {{33799, 1, 0, 0},{33799, 1, 1, 17}, {0, 0, 0, 0}}},{22204, {{33799, 1, 0, 0},{33799, 1, 1, 18}, {0, 0, 0, 0}}},{22204, {{33799, 1, 0, 0},{33799, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22205, {{33800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22206, {{33801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22207, {{33802, 1, 0, 0},{33802, 1, 1, 17}, {0, 0, 0, 0}}},{22207, {{33802, 1, 0, 0},{33802, 1, 1, 17}, {0, 0, 0, 0}}},{22208, {{33802, 1, 0, 0},{33802, 1, 1, 18}, {0, 0, 0, 0}}},{22208, {{33802, 1, 0, 0},{33802, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{22209, {{33803, 1, 0, 0},{33803, 1, 1, 17}, {0, 0, 0, 0}}},{22209, {{33803, 1, 0, 0},{33803, 1, 1, 17}, {0, 0, 0, 0}}},{22210, {{33803, 1, 0, 0},{33803, 1, 1, 18}, {0, 0, 0, 0}}},{22210, {{33803, 1, 0, 0},{33803, 1, 1, 18}, {0, 0, 0, 0}}},{22211, {{33803, 1, 0, 0},{33803, 1, 1, 19}, {0, 0, 0, 0}}},{22211, {{33803, 1, 0, 0},{33803, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{22212, {{33804, 1, 0, 0},{33804, 1, 1, 18}, {0, 0, 0, 0}}},{22212, {{33804, 1, 0, 0},{33804, 1, 1, 18}, {0, 0, 0, 0}}},{22213, {{33804, 1, 0, 0},{33804, 1, 1, 19}, {0, 0, 0, 0}}},{22213, {{33804, 1, 0, 0},{33804, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {6, {{22214, {{33805, 1, 0, 0},{33805, 1, 1, 17}, {0, 0, 0, 0}}},{22214, {{33805, 1, 0, 0},{33805, 1, 1, 17}, {0, 0, 0, 0}}},{22215, {{33805, 1, 0, 0},{33805, 1, 1, 18}, {0, 0, 0, 0}}},{22215, {{33805, 1, 0, 0},{33805, 1, 1, 18}, {0, 0, 0, 0}}},{22216, {{33805, 1, 0, 0},{33805, 1, 1, 19}, {0, 0, 0, 0}}},{22216, {{33805, 1, 0, 0},{33805, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{22217, {{33806, 1, 0, 0},{33806, 1, 1, 17}, {0, 0, 0, 0}}},{22217, {{33806, 1, 0, 0},{33806, 1, 1, 17}, {0, 0, 0, 0}}},{22218, {{33806, 1, 0, 0},{33806, 1, 1, 18}, {0, 0, 0, 0}}},{22218, {{33806, 1, 0, 0},{33806, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22219, {{33807, 1, 0, 0},{33807, 1, 1, 18}, {0, 0, 0, 0}}},{22219, {{33807, 1, 0, 0},{33807, 1, 1, 18}, {0, 0, 0, 0}}},{22220, {{33807, 1, 0, 0},{33807, 1, 1, 19}, {0, 0, 0, 0}}},{22220, {{33807, 1, 0, 0},{33807, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{22221, {{33808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22222, {{33809, 1, 0, 0},{33809, 1, 1, 17}, {0, 0, 0, 0}}},{22222, {{33809, 1, 0, 0},{33809, 1, 1, 17}, {0, 0, 0, 0}}},{22223, {{33809, 1, 0, 0},{33809, 1, 1, 18}, {0, 0, 0, 0}}},{22223, {{33809, 1, 0, 0},{33809, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22224, {{33810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{22225, {{33811, 1, 0, 0},{33811, 1, 1, 17}, {0, 0, 0, 0}}},{22225, {{33811, 1, 0, 0},{33811, 1, 1, 17}, {0, 0, 0, 0}}},{22226, {{33811, 1, 0, 0},{33811, 1, 1, 19}, {0, 0, 0, 0}}},{22226, {{33811, 1, 0, 0},{33811, 1, 1, 19}, {0, 0, 0, 0}}},{22227, {{33811, 1, 0, 0},{33811, 1, 1, 18}, {0, 0, 0, 0}}},{22227, {{33811, 1, 0, 0},{33811, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22228, {{33812, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{22229, {{33813, 1, 0, 0},{33813, 1, 1, 17}, {0, 0, 0, 0}}},{22229, {{33813, 1, 0, 0},{33813, 1, 1, 17}, {0, 0, 0, 0}}},{22230, {{33813, 1, 0, 0},{33813, 1, 1, 19}, {0, 0, 0, 0}}},{22230, {{33813, 1, 0, 0},{33813, 1, 1, 19}, {0, 0, 0, 0}}},{22231, {{33813, 1, 0, 0},{33813, 1, 1, 18}, {0, 0, 0, 0}}},{22231, {{33813, 1, 0, 0},{33813, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{22232, {{33815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22233, {{33816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22234, {{33817, 1, 0, 0},{33817, 1, 1, 17}, {0, 0, 0, 0}}},{22234, {{33817, 1, 0, 0},{33817, 1, 1, 17}, {0, 0, 0, 0}}},{22235, {{33817, 1, 0, 0},{33817, 1, 1, 18}, {0, 0, 0, 0}}},{22235, {{33817, 1, 0, 0},{33817, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{22236, {{33822, 1, 0, 0},{33822, 1, 1, 16}, {0, 0, 0, 0}}},{22236, {{33822, 1, 0, 0},{33822, 1, 1, 16}, {0, 0, 0, 0}}},{22237, {{33822, 1, 0, 0},{33822, 1, 1, 17}, {0, 0, 0, 0}}},{22237, {{33822, 1, 0, 0},{33822, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{22238, {{33823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{22239, {{33824, 1, 0, 0},{33824, 1, 1, 18}, {0, 0, 0, 0}}},{22239, {{33824, 1, 0, 0},{33824, 1, 1, 18}, {0, 0, 0, 0}}},{22240, {{33824, 1, 0, 0},{33824, 1, 1, 19}, {0, 0, 0, 0}}},{22240, {{33824, 1, 0, 0},{33824, 1, 1, 19}, {0, 0, 0, 0}}},{22241, {{33824, 1, 0, 0},{33824, 1, 1, 20}, {0, 0, 0, 0}}},{22241, {{33824, 1, 0, 0},{33824, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{22243, {{33826, 1, 0, 0},{33826, 1, 1, 18}, {0, 0, 0, 0}}},{22243, {{33826, 1, 0, 0},{33826, 1, 1, 18}, {0, 0, 0, 0}}},{22244, {{33826, 1, 0, 0},{33826, 1, 1, 20}, {0, 0, 0, 0}}},{22244, {{33826, 1, 0, 0},{33826, 1, 1, 20}, {0, 0, 0, 0}}},{22245, {{33826, 1, 0, 0},{33826, 1, 1, 19}, {0, 0, 0, 0}}},{22245, {{33826, 1, 0, 0},{33826, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{22246, {{33828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22247, {{33829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{22248, {{33833, 1, 0, 0},{33833, 1, 1, 17}, {0, 0, 0, 0}}},{22248, {{33833, 1, 0, 0},{33833, 1, 1, 17}, {0, 0, 0, 0}}},{22249, {{33833, 1, 0, 0},{33833, 1, 1, 18}, {0, 0, 0, 0}}},{22249, {{33833, 1, 0, 0},{33833, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22250, {{33834, 1, 0, 0},{33834, 1, 1, 17}, {0, 0, 0, 0}}},{22250, {{33834, 1, 0, 0},{33834, 1, 1, 17}, {0, 0, 0, 0}}},{22251, {{33834, 1, 0, 0},{33834, 1, 1, 18}, {0, 0, 0, 0}}},{22251, {{33834, 1, 0, 0},{33834, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22252, {{33835, 1, 0, 0},{33835, 1, 1, 16}, {0, 0, 0, 0}}},{22252, {{33835, 1, 0, 0},{33835, 1, 1, 16}, {0, 0, 0, 0}}},{22253, {{33835, 1, 0, 0},{33835, 1, 1, 17}, {0, 0, 0, 0}}},{22253, {{33835, 1, 0, 0},{33835, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {8, {{22254, {{33836, 1, 0, 0},{33836, 1, 1, 17}, {0, 0, 0, 0}}},{22254, {{33836, 1, 0, 0},{33836, 1, 1, 17}, {0, 0, 0, 0}}},{22255, {{33836, 1, 0, 0},{33836, 1, 1, 20}, {0, 0, 0, 0}}},{22255, {{33836, 1, 0, 0},{33836, 1, 1, 20}, {0, 0, 0, 0}}},{22256, {{33836, 1, 0, 0},{33836, 1, 1, 19}, {0, 0, 0, 0}}},{22256, {{33836, 1, 0, 0},{33836, 1, 1, 19}, {0, 0, 0, 0}}},{22257, {{33836, 1, 0, 0},{33836, 1, 1, 18}, {0, 0, 0, 0}}},{22257, {{33836, 1, 0, 0},{33836, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22258, {{33837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22259, {{33838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22260, {{33839, 1, 0, 0},{33839, 1, 1, 17}, {0, 0, 0, 0}}},{22260, {{33839, 1, 0, 0},{33839, 1, 1, 17}, {0, 0, 0, 0}}},{22261, {{33839, 1, 0, 0},{33839, 1, 1, 18}, {0, 0, 0, 0}}},{22261, {{33839, 1, 0, 0},{33839, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22262, {{33840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22263, {{33841, 1, 0, 0},{33841, 1, 1, 17}, {0, 0, 0, 0}}},{22263, {{33841, 1, 0, 0},{33841, 1, 1, 17}, {0, 0, 0, 0}}},{22264, {{33841, 1, 0, 0},{33841, 1, 1, 18}, {0, 0, 0, 0}}},{22264, {{33841, 1, 0, 0},{33841, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22265, {{33842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22266, {{33843, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22267, {{33844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22268, {{33845, 1, 0, 0},{33845, 1, 1, 17}, {0, 0, 0, 0}}},{22268, {{33845, 1, 0, 0},{33845, 1, 1, 17}, {0, 0, 0, 0}}},{22269, {{33845, 1, 0, 0},{33845, 1, 1, 18}, {0, 0, 0, 0}}},{22269, {{33845, 1, 0, 0},{33845, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22270, {{33846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22272, {{33847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22273, {{33848, 1, 0, 0},{33848, 1, 1, 17}, {0, 0, 0, 0}}},{22273, {{33848, 1, 0, 0},{33848, 1, 1, 17}, {0, 0, 0, 0}}},{22274, {{33848, 1, 0, 0},{33848, 1, 1, 18}, {0, 0, 0, 0}}},{22274, {{33848, 1, 0, 0},{33848, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22275, {{33849, 1, 0, 0},{33849, 1, 1, 17}, {0, 0, 0, 0}}},{22275, {{33849, 1, 0, 0},{33849, 1, 1, 17}, {0, 0, 0, 0}}},{22276, {{33849, 1, 0, 0},{33849, 1, 1, 18}, {0, 0, 0, 0}}},{22276, {{33849, 1, 0, 0},{33849, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22277, {{33850, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22278, {{33851, 1, 0, 0},{33851, 1, 1, 16}, {0, 0, 0, 0}}},{22278, {{33851, 1, 0, 0},{33851, 1, 1, 16}, {0, 0, 0, 0}}},{22279, {{33851, 1, 0, 0},{33851, 1, 1, 17}, {0, 0, 0, 0}}},{22279, {{33851, 1, 0, 0},{33851, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{22280, {{33852, 1, 0, 0},{33852, 1, 1, 17}, {0, 0, 0, 0}}},{22280, {{33852, 1, 0, 0},{33852, 1, 1, 17}, {0, 0, 0, 0}}},{22281, {{33852, 1, 0, 0},{33852, 1, 1, 18}, {0, 0, 0, 0}}},{22281, {{33852, 1, 0, 0},{33852, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22282, {{33853, 1, 0, 0},{33853, 1, 1, 17}, {0, 0, 0, 0}}},{22282, {{33853, 1, 0, 0},{33853, 1, 1, 17}, {0, 0, 0, 0}}},{22283, {{33853, 1, 0, 0},{33853, 1, 1, 18}, {0, 0, 0, 0}}},{22283, {{33853, 1, 0, 0},{33853, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22284, {{33854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22285, {{33855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22287, {{33856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22288, {{33857, 1, 0, 0},{33857, 1, 1, 16}, {0, 0, 0, 0}}},{22288, {{33857, 1, 0, 0},{33857, 1, 1, 16}, {0, 0, 0, 0}}},{22289, {{33857, 1, 0, 0},{33857, 1, 1, 17}, {0, 0, 0, 0}}},{22289, {{33857, 1, 0, 0},{33857, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{22290, {{33858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22291, {{33859, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22292, {{33860, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22293, {{33861, 1, 0, 0},{33861, 1, 1, 17}, {0, 0, 0, 0}}},{22293, {{33861, 1, 0, 0},{33861, 1, 1, 17}, {0, 0, 0, 0}}},{22294, {{33861, 1, 0, 0},{33861, 1, 1, 18}, {0, 0, 0, 0}}},{22294, {{33861, 1, 0, 0},{33861, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22295, {{33862, 1, 0, 0},{33862, 1, 1, 17}, {0, 0, 0, 0}}},{22295, {{33862, 1, 0, 0},{33862, 1, 1, 17}, {0, 0, 0, 0}}},{22296, {{33862, 1, 0, 0},{33862, 1, 1, 18}, {0, 0, 0, 0}}},{22296, {{33862, 1, 0, 0},{33862, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22297, {{33863, 1, 0, 0},{33863, 1, 1, 17}, {0, 0, 0, 0}}},{22297, {{33863, 1, 0, 0},{33863, 1, 1, 17}, {0, 0, 0, 0}}},{22298, {{33863, 1, 0, 0},{33863, 1, 1, 18}, {0, 0, 0, 0}}},{22298, {{33863, 1, 0, 0},{33863, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22299, {{33864, 1, 0, 0},{33864, 1, 1, 17}, {0, 0, 0, 0}}},{22299, {{33864, 1, 0, 0},{33864, 1, 1, 17}, {0, 0, 0, 0}}},{22300, {{33864, 1, 0, 0},{33864, 1, 1, 18}, {0, 0, 0, 0}}},{22300, {{33864, 1, 0, 0},{33864, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{22301, {{33865, 1, 0, 0},{33865, 1, 1, 17}, {0, 0, 0, 0}}},{22301, {{33865, 1, 0, 0},{33865, 1, 1, 17}, {0, 0, 0, 0}}},{22302, {{33865, 1, 0, 0},{33865, 1, 1, 19}, {0, 0, 0, 0}}},{22302, {{33865, 1, 0, 0},{33865, 1, 1, 19}, {0, 0, 0, 0}}},{22303, {{33865, 1, 0, 0},{33865, 1, 1, 18}, {0, 0, 0, 0}}},{22303, {{33865, 1, 0, 0},{33865, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22305, {{33866, 1, 0, 0},{33866, 1, 1, 17}, {0, 0, 0, 0}}},{22305, {{33866, 1, 0, 0},{33866, 1, 1, 17}, {0, 0, 0, 0}}},{22306, {{33866, 1, 0, 0},{33866, 1, 1, 18}, {0, 0, 0, 0}}},{22306, {{33866, 1, 0, 0},{33866, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22307, {{33867, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22308, {{33868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22309, {{33869, 1, 0, 0},{33869, 1, 1, 17}, {0, 0, 0, 0}}},{22309, {{33869, 1, 0, 0},{33869, 1, 1, 17}, {0, 0, 0, 0}}},{22310, {{33869, 1, 0, 0},{33869, 1, 1, 18}, {0, 0, 0, 0}}},{22310, {{33869, 1, 0, 0},{33869, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22311, {{33870, 1, 0, 0},{33870, 1, 1, 17}, {0, 0, 0, 0}}},{22311, {{33870, 1, 0, 0},{33870, 1, 1, 17}, {0, 0, 0, 0}}},{22312, {{33870, 1, 0, 0},{33870, 1, 1, 18}, {0, 0, 0, 0}}},{22312, {{33870, 1, 0, 0},{33870, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22313, {{33871, 1, 0, 0},{33871, 1, 1, 17}, {0, 0, 0, 0}}},{22313, {{33871, 1, 0, 0},{33871, 1, 1, 17}, {0, 0, 0, 0}}},{22314, {{33871, 1, 0, 0},{33871, 1, 1, 18}, {0, 0, 0, 0}}},{22314, {{33871, 1, 0, 0},{33871, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22315, {{33872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22316, {{33873, 1, 0, 0},{33873, 1, 1, 17}, {0, 0, 0, 0}}},{22316, {{33873, 1, 0, 0},{33873, 1, 1, 17}, {0, 0, 0, 0}}},{22317, {{33873, 1, 0, 0},{33873, 1, 1, 18}, {0, 0, 0, 0}}},{22317, {{33873, 1, 0, 0},{33873, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22318, {{33874, 1, 0, 0},{33874, 1, 1, 17}, {0, 0, 0, 0}}},{22318, {{33874, 1, 0, 0},{33874, 1, 1, 17}, {0, 0, 0, 0}}},{22319, {{33874, 1, 0, 0},{33874, 1, 1, 18}, {0, 0, 0, 0}}},{22319, {{33874, 1, 0, 0},{33874, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22320, {{33875, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22321, {{33876, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22322, {{33877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22324, {{33878, 1, 0, 0},{33878, 1, 1, 17}, {0, 0, 0, 0}}},{22324, {{33878, 1, 0, 0},{33878, 1, 1, 17}, {0, 0, 0, 0}}},{22325, {{33878, 1, 0, 0},{33878, 1, 1, 18}, {0, 0, 0, 0}}},{22325, {{33878, 1, 0, 0},{33878, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {7, {{22326, {{33879, 1, 0, 0},{33879, 1, 1, 18}, {0, 0, 0, 0}}},{22326, {{33879, 1, 0, 0},{33879, 1, 1, 18}, {0, 0, 0, 0}}},{22327, {{33879, 1, 0, 0},{33879, 1, 1, 19}, {0, 0, 0, 0}}},{22327, {{33879, 1, 0, 0},{33879, 1, 1, 19}, {0, 0, 0, 0}}},{30294, {{33879, 1, 0, 0},{33879, 1, 1, 0},{33879, 1, 1, 20},{64095, 1, 0, 0}, {0, 0, 0, 0}}},{30294, {{33879, 1, 0, 0},{33879, 1, 1, 0},{33879, 1, 1, 20},{64095, 1, 0, 0}, {0, 0, 0, 0}}},{30294, {{33879, 1, 0, 0},{33879, 1, 1, 0},{33879, 1, 1, 20},{64095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22328, {{33880, 1, 0, 0},{33880, 1, 1, 17}, {0, 0, 0, 0}}},{22328, {{33880, 1, 0, 0},{33880, 1, 1, 17}, {0, 0, 0, 0}}},{22329, {{33880, 1, 0, 0},{33880, 1, 1, 18}, {0, 0, 0, 0}}},{22329, {{33880, 1, 0, 0},{33880, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22330, {{33881, 1, 0, 0},{33881, 1, 1, 17}, {0, 0, 0, 0}}},{22330, {{33881, 1, 0, 0},{33881, 1, 1, 17}, {0, 0, 0, 0}}},{22331, {{33881, 1, 0, 0},{33881, 1, 1, 18}, {0, 0, 0, 0}}},{22331, {{33881, 1, 0, 0},{33881, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{22332, {{33882, 1, 0, 0},{33882, 1, 1, 17}, {0, 0, 0, 0}}},{22332, {{33882, 1, 0, 0},{33882, 1, 1, 17}, {0, 0, 0, 0}}},{22333, {{33882, 1, 0, 0},{33882, 1, 1, 18}, {0, 0, 0, 0}}},{22333, {{33882, 1, 0, 0},{33882, 1, 1, 18}, {0, 0, 0, 0}}},{22334, {{33882, 1, 0, 0},{33882, 1, 1, 19}, {0, 0, 0, 0}}},{22334, {{33882, 1, 0, 0},{33882, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {14, {{22335, {{33883, 1, 0, 0},{33883, 1, 1, 18}, {0, 0, 0, 0}}},{22335, {{33883, 1, 0, 0},{33883, 1, 1, 18}, {0, 0, 0, 0}}},{22336, {{33883, 1, 0, 0},{33883, 1, 1, 19}, {0, 0, 0, 0}}},{22336, {{33883, 1, 0, 0},{33883, 1, 1, 19}, {0, 0, 0, 0}}},{22337, {{33883, 1, 0, 0},{33883, 1, 1, 23}, {0, 0, 0, 0}}},{22337, {{33883, 1, 0, 0},{33883, 1, 1, 23}, {0, 0, 0, 0}}},{22338, {{33883, 1, 0, 0},{33883, 1, 1, 22}, {0, 0, 0, 0}}},{22338, {{33883, 1, 0, 0},{33883, 1, 1, 22}, {0, 0, 0, 0}}},{22339, {{33883, 1, 0, 0},{33883, 1, 1, 24}, {0, 0, 0, 0}}},{22339, {{33883, 1, 0, 0},{33883, 1, 1, 24}, {0, 0, 0, 0}}},{22340, {{33883, 1, 0, 0},{33883, 1, 1, 20}, {0, 0, 0, 0}}},{22340, {{33883, 1, 0, 0},{33883, 1, 1, 20}, {0, 0, 0, 0}}},{22341, {{33883, 1, 0, 0},{33883, 1, 1, 21}, {0, 0, 0, 0}}},{22341, {{33883, 1, 0, 0},{33883, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {6, {{22342, {{33884, 1, 0, 0},{33884, 1, 1, 18}, {0, 0, 0, 0}}},{22342, {{33884, 1, 0, 0},{33884, 1, 1, 18}, {0, 0, 0, 0}}},{22343, {{33884, 1, 0, 0},{33884, 1, 1, 19}, {0, 0, 0, 0}}},{22343, {{33884, 1, 0, 0},{33884, 1, 1, 19}, {0, 0, 0, 0}}},{22344, {{33884, 1, 0, 0},{33884, 1, 1, 20}, {0, 0, 0, 0}}},{22344, {{33884, 1, 0, 0},{33884, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{22345, {{33885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22346, {{33886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22347, {{33887, 1, 0, 0},{33887, 1, 1, 17}, {0, 0, 0, 0}}},{22347, {{33887, 1, 0, 0},{33887, 1, 1, 17}, {0, 0, 0, 0}}},{22348, {{33887, 1, 0, 0},{33887, 1, 1, 18}, {0, 0, 0, 0}}},{22348, {{33887, 1, 0, 0},{33887, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22349, {{33888, 1, 0, 0},{33888, 1, 1, 17}, {0, 0, 0, 0}}},{22349, {{33888, 1, 0, 0},{33888, 1, 1, 17}, {0, 0, 0, 0}}},{22350, {{33888, 1, 0, 0},{33888, 1, 1, 18}, {0, 0, 0, 0}}},{22350, {{33888, 1, 0, 0},{33888, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22351, {{33889, 1, 0, 0},{33889, 1, 1, 17}, {0, 0, 0, 0}}},{22351, {{33889, 1, 0, 0},{33889, 1, 1, 17}, {0, 0, 0, 0}}},{22352, {{33889, 1, 0, 0},{33889, 1, 1, 18}, {0, 0, 0, 0}}},{22352, {{33889, 1, 0, 0},{33889, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{22353, {{33890, 1, 0, 0},{33890, 1, 1, 17}, {0, 0, 0, 0}}},{22353, {{33890, 1, 0, 0},{33890, 1, 1, 17}, {0, 0, 0, 0}}},{22354, {{33890, 1, 0, 0},{33890, 1, 1, 19}, {0, 0, 0, 0}}},{22354, {{33890, 1, 0, 0},{33890, 1, 1, 19}, {0, 0, 0, 0}}},{22355, {{33890, 1, 0, 0},{33890, 1, 1, 18}, {0, 0, 0, 0}}},{22355, {{33890, 1, 0, 0},{33890, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22356, {{33891, 1, 0, 0},{33891, 1, 1, 17}, {0, 0, 0, 0}}},{22356, {{33891, 1, 0, 0},{33891, 1, 1, 17}, {0, 0, 0, 0}}},{22357, {{33891, 1, 0, 0},{33891, 1, 1, 18}, {0, 0, 0, 0}}},{22357, {{33891, 1, 0, 0},{33891, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22359, {{33892, 1, 0, 0},{33892, 1, 1, 17}, {0, 0, 0, 0}}},{22359, {{33892, 1, 0, 0},{33892, 1, 1, 17}, {0, 0, 0, 0}}},{22360, {{33892, 1, 0, 0},{33892, 1, 1, 18}, {0, 0, 0, 0}}},{22360, {{33892, 1, 0, 0},{33892, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22361, {{33893, 1, 0, 0},{33893, 1, 1, 17}, {0, 0, 0, 0}}},{22361, {{33893, 1, 0, 0},{33893, 1, 1, 17}, {0, 0, 0, 0}}},{22362, {{33893, 1, 0, 0},{33893, 1, 1, 18}, {0, 0, 0, 0}}},{22362, {{33893, 1, 0, 0},{33893, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {12, {{22363, {{33894, 1, 0, 0},{33894, 1, 1, 21}, {0, 0, 0, 0}}},{22363, {{33894, 1, 0, 0},{33894, 1, 1, 21}, {0, 0, 0, 0}}},{22364, {{33894, 1, 0, 0},{33894, 1, 1, 17}, {0, 0, 0, 0}}},{22364, {{33894, 1, 0, 0},{33894, 1, 1, 17}, {0, 0, 0, 0}}},{22365, {{33894, 1, 0, 0},{33894, 1, 1, 18}, {0, 0, 0, 0}}},{22365, {{33894, 1, 0, 0},{33894, 1, 1, 18}, {0, 0, 0, 0}}},{22366, {{33894, 1, 0, 0},{33894, 1, 1, 20}, {0, 0, 0, 0}}},{22366, {{33894, 1, 0, 0},{33894, 1, 1, 20}, {0, 0, 0, 0}}},{22367, {{33894, 1, 0, 0},{33894, 1, 1, 22}, {0, 0, 0, 0}}},{22367, {{33894, 1, 0, 0},{33894, 1, 1, 22}, {0, 0, 0, 0}}},{22368, {{33894, 1, 0, 0},{33894, 1, 1, 19}, {0, 0, 0, 0}}},{22368, {{33894, 1, 0, 0},{33894, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{22369, {{33895, 1, 0, 0},{33895, 1, 1, 17}, {0, 0, 0, 0}}},{22369, {{33895, 1, 0, 0},{33895, 1, 1, 17}, {0, 0, 0, 0}}},{22370, {{33895, 1, 0, 0},{33895, 1, 1, 18}, {0, 0, 0, 0}}},{22370, {{33895, 1, 0, 0},{33895, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22371, {{33896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22372, {{33897, 1, 0, 0},{33897, 1, 1, 17}, {0, 0, 0, 0}}},{22372, {{33897, 1, 0, 0},{33897, 1, 1, 17}, {0, 0, 0, 0}}},{22373, {{33897, 1, 0, 0},{33897, 1, 1, 18}, {0, 0, 0, 0}}},{22373, {{33897, 1, 0, 0},{33897, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22374, {{33898, 1, 0, 0},{33898, 1, 1, 17}, {0, 0, 0, 0}}},{22374, {{33898, 1, 0, 0},{33898, 1, 1, 17}, {0, 0, 0, 0}}},{22375, {{33898, 1, 0, 0},{33898, 1, 1, 18}, {0, 0, 0, 0}}},{22375, {{33898, 1, 0, 0},{33898, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22376, {{33899, 1, 0, 0},{33899, 1, 1, 17}, {0, 0, 0, 0}}},{22376, {{33899, 1, 0, 0},{33899, 1, 1, 17}, {0, 0, 0, 0}}},{22377, {{33899, 1, 0, 0},{33899, 1, 1, 18}, {0, 0, 0, 0}}},{22377, {{33899, 1, 0, 0},{33899, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22378, {{33900, 1, 0, 0},{33900, 1, 1, 17}, {0, 0, 0, 0}}},{22378, {{33900, 1, 0, 0},{33900, 1, 1, 17}, {0, 0, 0, 0}}},{22379, {{33900, 1, 0, 0},{33900, 1, 1, 18}, {0, 0, 0, 0}}},{22379, {{33900, 1, 0, 0},{33900, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22380, {{33901, 1, 0, 0},{33901, 1, 1, 17}, {0, 0, 0, 0}}},{22380, {{33901, 1, 0, 0},{33901, 1, 1, 17}, {0, 0, 0, 0}}},{22381, {{33901, 1, 0, 0},{33901, 1, 1, 18}, {0, 0, 0, 0}}},{22381, {{33901, 1, 0, 0},{33901, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22382, {{33902, 1, 0, 0},{33902, 1, 1, 17}, {0, 0, 0, 0}}},{22382, {{33902, 1, 0, 0},{33902, 1, 1, 17}, {0, 0, 0, 0}}},{22383, {{33902, 1, 0, 0},{33902, 1, 1, 18}, {0, 0, 0, 0}}},{22383, {{33902, 1, 0, 0},{33902, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22384, {{33903, 1, 0, 0},{33903, 1, 1, 17}, {0, 0, 0, 0}}},{22384, {{33903, 1, 0, 0},{33903, 1, 1, 17}, {0, 0, 0, 0}}},{22385, {{33903, 1, 0, 0},{33903, 1, 1, 18}, {0, 0, 0, 0}}},{22385, {{33903, 1, 0, 0},{33903, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22386, {{33904, 1, 0, 0},{33904, 1, 1, 17}, {0, 0, 0, 0}}},{22386, {{33904, 1, 0, 0},{33904, 1, 1, 17}, {0, 0, 0, 0}}},{22387, {{33904, 1, 0, 0},{33904, 1, 1, 18}, {0, 0, 0, 0}}},{22387, {{33904, 1, 0, 0},{33904, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22388, {{33905, 1, 0, 0},{33905, 1, 1, 17}, {0, 0, 0, 0}}},{22388, {{33905, 1, 0, 0},{33905, 1, 1, 17}, {0, 0, 0, 0}}},{22389, {{33905, 1, 0, 0},{33905, 1, 1, 18}, {0, 0, 0, 0}}},{22389, {{33905, 1, 0, 0},{33905, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22390, {{33906, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22391, {{33907, 1, 0, 0},{33907, 1, 1, 17}, {0, 0, 0, 0}}},{22391, {{33907, 1, 0, 0},{33907, 1, 1, 17}, {0, 0, 0, 0}}},{22392, {{33907, 1, 0, 0},{33907, 1, 1, 18}, {0, 0, 0, 0}}},{22392, {{33907, 1, 0, 0},{33907, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22393, {{33908, 1, 0, 0},{33908, 1, 1, 17}, {0, 0, 0, 0}}},{22393, {{33908, 1, 0, 0},{33908, 1, 1, 17}, {0, 0, 0, 0}}},{22394, {{33908, 1, 0, 0},{33908, 1, 1, 18}, {0, 0, 0, 0}}},{22394, {{33908, 1, 0, 0},{33908, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22395, {{33909, 1, 0, 0},{33909, 1, 1, 17}, {0, 0, 0, 0}}},{22395, {{33909, 1, 0, 0},{33909, 1, 1, 17}, {0, 0, 0, 0}}},{22396, {{33909, 1, 0, 0},{33909, 1, 1, 18}, {0, 0, 0, 0}}},{22396, {{33909, 1, 0, 0},{33909, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22397, {{33910, 1, 0, 0},{33910, 1, 1, 17}, {0, 0, 0, 0}}},{22397, {{33910, 1, 0, 0},{33910, 1, 1, 17}, {0, 0, 0, 0}}},{22398, {{33910, 1, 0, 0},{33910, 1, 1, 18}, {0, 0, 0, 0}}},{22398, {{33910, 1, 0, 0},{33910, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22399, {{33911, 1, 0, 0},{33911, 1, 1, 17}, {0, 0, 0, 0}}},{22399, {{33911, 1, 0, 0},{33911, 1, 1, 17}, {0, 0, 0, 0}}},{22400, {{33911, 1, 0, 0},{33911, 1, 1, 18}, {0, 0, 0, 0}}},{22400, {{33911, 1, 0, 0},{33911, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22401, {{33912, 1, 0, 0},{33912, 1, 1, 17}, {0, 0, 0, 0}}},{22401, {{33912, 1, 0, 0},{33912, 1, 1, 17}, {0, 0, 0, 0}}},{22402, {{33912, 1, 0, 0},{33912, 1, 1, 18}, {0, 0, 0, 0}}},{22402, {{33912, 1, 0, 0},{33912, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22403, {{33913, 1, 0, 0},{33913, 1, 1, 17}, {0, 0, 0, 0}}},{22403, {{33913, 1, 0, 0},{33913, 1, 1, 17}, {0, 0, 0, 0}}},{22404, {{33913, 1, 0, 0},{33913, 1, 1, 18}, {0, 0, 0, 0}}},{22404, {{33913, 1, 0, 0},{33913, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{22405, {{33914, 1, 0, 0},{33914, 1, 1, 18}, {0, 0, 0, 0}}},{22405, {{33914, 1, 0, 0},{33914, 1, 1, 18}, {0, 0, 0, 0}}},{22406, {{33914, 1, 0, 0},{33914, 1, 1, 19}, {0, 0, 0, 0}}},{22406, {{33914, 1, 0, 0},{33914, 1, 1, 19}, {0, 0, 0, 0}}},{22407, {{33914, 1, 0, 0},{33914, 1, 1, 20}, {0, 0, 0, 0}}},{22407, {{33914, 1, 0, 0},{33914, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{22408, {{33915, 1, 0, 0},{33915, 1, 1, 16}, {0, 0, 0, 0}}},{22408, {{33915, 1, 0, 0},{33915, 1, 1, 16}, {0, 0, 0, 0}}},{22409, {{33915, 1, 0, 0},{33915, 1, 1, 17}, {0, 0, 0, 0}}},{22409, {{33915, 1, 0, 0},{33915, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{22410, {{33916, 1, 0, 0},{33916, 1, 1, 17}, {0, 0, 0, 0}}},{22410, {{33916, 1, 0, 0},{33916, 1, 1, 17}, {0, 0, 0, 0}}},{22411, {{33916, 1, 0, 0},{33916, 1, 1, 18}, {0, 0, 0, 0}}},{22411, {{33916, 1, 0, 0},{33916, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {8, {{22412, {{33917, 1, 0, 0},{33917, 1, 1, 17}, {0, 0, 0, 0}}},{22412, {{33917, 1, 0, 0},{33917, 1, 1, 17}, {0, 0, 0, 0}}},{22413, {{33917, 1, 0, 0},{33917, 1, 1, 19}, {0, 0, 0, 0}}},{22413, {{33917, 1, 0, 0},{33917, 1, 1, 19}, {0, 0, 0, 0}}},{22414, {{33917, 1, 0, 0},{33917, 1, 1, 20}, {0, 0, 0, 0}}},{22414, {{33917, 1, 0, 0},{33917, 1, 1, 20}, {0, 0, 0, 0}}},{22415, {{33917, 1, 0, 0},{33917, 1, 1, 18}, {0, 0, 0, 0}}},{22415, {{33917, 1, 0, 0},{33917, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22416, {{33918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22417, {{33919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22418, {{33920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{22420, {{33921, 1, 0, 0},{33921, 1, 1, 19}, {0, 0, 0, 0}}},{22420, {{33921, 1, 0, 0},{33921, 1, 1, 19}, {0, 0, 0, 0}}},{22421, {{33921, 1, 0, 0},{33921, 1, 1, 17}, {0, 0, 0, 0}}},{22421, {{33921, 1, 0, 0},{33921, 1, 1, 17}, {0, 0, 0, 0}}},{22422, {{33921, 1, 0, 0},{33921, 1, 1, 20}, {0, 0, 0, 0}}},{22422, {{33921, 1, 0, 0},{33921, 1, 1, 20}, {0, 0, 0, 0}}},{22423, {{33921, 1, 0, 0},{33921, 1, 1, 18}, {0, 0, 0, 0}}},{22423, {{33921, 1, 0, 0},{33921, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{22424, {{33922, 1, 0, 0},{33922, 1, 1, 17}, {0, 0, 0, 0}}},{22424, {{33922, 1, 0, 0},{33922, 1, 1, 17}, {0, 0, 0, 0}}},{22425, {{33922, 1, 0, 0},{33922, 1, 1, 19}, {0, 0, 0, 0}}},{22425, {{33922, 1, 0, 0},{33922, 1, 1, 19}, {0, 0, 0, 0}}},{22426, {{33922, 1, 0, 0},{33922, 1, 1, 18}, {0, 0, 0, 0}}},{22426, {{33922, 1, 0, 0},{33922, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22427, {{33923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22428, {{33924, 1, 0, 0},{33924, 1, 1, 17}, {0, 0, 0, 0}}},{22428, {{33924, 1, 0, 0},{33924, 1, 1, 17}, {0, 0, 0, 0}}},{22429, {{33924, 1, 0, 0},{33924, 1, 1, 18}, {0, 0, 0, 0}}},{22429, {{33924, 1, 0, 0},{33924, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{22430, {{33925, 1, 0, 0},{33925, 1, 1, 17}, {0, 0, 0, 0}}},{22430, {{33925, 1, 0, 0},{33925, 1, 1, 17}, {0, 0, 0, 0}}},{22431, {{33925, 1, 0, 0},{33925, 1, 1, 18}, {0, 0, 0, 0}}},{22431, {{33925, 1, 0, 0},{33925, 1, 1, 18}, {0, 0, 0, 0}}},{22432, {{33925, 1, 0, 0},{33925, 1, 1, 19}, {0, 0, 0, 0}}},{22432, {{33925, 1, 0, 0},{33925, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{68045, {{33930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22433, {{33931, 1, 0, 0},{33931, 1, 1, 17}, {0, 0, 0, 0}}},{22433, {{33931, 1, 0, 0},{33931, 1, 1, 17}, {0, 0, 0, 0}}},{22434, {{33931, 1, 0, 0},{33931, 1, 1, 18}, {0, 0, 0, 0}}},{22434, {{33931, 1, 0, 0},{33931, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{22437, {{33936, 1, 0, 0},{33936, 1, 1, 17}, {0, 0, 0, 0}}},{22437, {{33936, 1, 0, 0},{33936, 1, 1, 17}, {0, 0, 0, 0}}},{22438, {{33936, 1, 0, 0},{33936, 1, 1, 18}, {0, 0, 0, 0}}},{22438, {{33936, 1, 0, 0},{33936, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22439, {{33937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22440, {{33938, 1, 0, 0},{33938, 1, 1, 17}, {0, 0, 0, 0}}},{22440, {{33938, 1, 0, 0},{33938, 1, 1, 17}, {0, 0, 0, 0}}},{22441, {{33938, 1, 0, 0},{33938, 1, 1, 18}, {0, 0, 0, 0}}},{22441, {{33938, 1, 0, 0},{33938, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22442, {{33939, 1, 0, 0},{33939, 1, 1, 17}, {0, 0, 0, 0}}},{22442, {{33939, 1, 0, 0},{33939, 1, 1, 17}, {0, 0, 0, 0}}},{22443, {{33939, 1, 0, 0},{33939, 1, 1, 18}, {0, 0, 0, 0}}},{22443, {{33939, 1, 0, 0},{33939, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22444, {{33940, 1, 0, 0},{33940, 1, 1, 17}, {0, 0, 0, 0}}},{22444, {{33940, 1, 0, 0},{33940, 1, 1, 17}, {0, 0, 0, 0}}},{22445, {{33940, 1, 0, 0},{33940, 1, 1, 18}, {0, 0, 0, 0}}},{22445, {{33940, 1, 0, 0},{33940, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22446, {{33941, 1, 0, 0},{33941, 1, 1, 17}, {0, 0, 0, 0}}},{22446, {{33941, 1, 0, 0},{33941, 1, 1, 17}, {0, 0, 0, 0}}},{22447, {{33941, 1, 0, 0},{33941, 1, 1, 18}, {0, 0, 0, 0}}},{22447, {{33941, 1, 0, 0},{33941, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22448, {{33942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22449, {{33943, 1, 0, 0},{33943, 1, 1, 17}, {0, 0, 0, 0}}},{22449, {{33943, 1, 0, 0},{33943, 1, 1, 17}, {0, 0, 0, 0}}},{22450, {{33943, 1, 0, 0},{33943, 1, 1, 18}, {0, 0, 0, 0}}},{22450, {{33943, 1, 0, 0},{33943, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22451, {{33944, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{22452, {{33945, 1, 0, 0},{33945, 1, 1, 17}, {0, 0, 0, 0}}},{22452, {{33945, 1, 0, 0},{33945, 1, 1, 17}, {0, 0, 0, 0}}},{22453, {{33945, 1, 0, 0},{33945, 1, 1, 18}, {0, 0, 0, 0}}},{22453, {{33945, 1, 0, 0},{33945, 1, 1, 18}, {0, 0, 0, 0}}},{22454, {{33945, 1, 0, 0},{33945, 1, 1, 19}, {0, 0, 0, 0}}},{22454, {{33945, 1, 0, 0},{33945, 1, 1, 19}, {0, 0, 0, 0}}},{22455, {{33945, 1, 0, 0},{33945, 1, 1, 20}, {0, 0, 0, 0}}},{22455, {{33945, 1, 0, 0},{33945, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{22456, {{33946, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22457, {{33947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{22458, {{33948, 1, 0, 0},{33948, 1, 1, 19}, {0, 0, 0, 0}}},{22458, {{33948, 1, 0, 0},{33948, 1, 1, 19}, {0, 0, 0, 0}}},{22459, {{33948, 1, 0, 0},{33948, 1, 1, 17}, {0, 0, 0, 0}}},{22459, {{33948, 1, 0, 0},{33948, 1, 1, 17}, {0, 0, 0, 0}}},{22460, {{33948, 1, 0, 0},{33948, 1, 1, 18}, {0, 0, 0, 0}}},{22460, {{33948, 1, 0, 0},{33948, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22461, {{33949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22463, {{33950, 1, 0, 0},{33950, 1, 1, 17}, {0, 0, 0, 0}}},{22463, {{33950, 1, 0, 0},{33950, 1, 1, 17}, {0, 0, 0, 0}}},{22464, {{33950, 1, 0, 0},{33950, 1, 1, 18}, {0, 0, 0, 0}}},{22464, {{33950, 1, 0, 0},{33950, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22465, {{33951, 1, 0, 0},{33951, 1, 1, 17}, {0, 0, 0, 0}}},{22465, {{33951, 1, 0, 0},{33951, 1, 1, 17}, {0, 0, 0, 0}}},{22466, {{33951, 1, 0, 0},{33951, 1, 1, 18}, {0, 0, 0, 0}}},{22466, {{33951, 1, 0, 0},{33951, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22467, {{33952, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{22468, {{33953, 1, 0, 0},{33953, 1, 1, 17}, {0, 0, 0, 0}}},{22468, {{33953, 1, 0, 0},{33953, 1, 1, 17}, {0, 0, 0, 0}}},{22469, {{33953, 1, 0, 0},{33953, 1, 1, 19}, {0, 0, 0, 0}}},{22469, {{33953, 1, 0, 0},{33953, 1, 1, 19}, {0, 0, 0, 0}}},{22470, {{33953, 1, 0, 0},{33953, 1, 1, 18}, {0, 0, 0, 0}}},{22470, {{33953, 1, 0, 0},{33953, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22471, {{33954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22472, {{33955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22473, {{33956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22474, {{33957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{22475, {{33958, 1, 0, 0},{33958, 1, 1, 17}, {0, 0, 0, 0}}},{22475, {{33958, 1, 0, 0},{33958, 1, 1, 17}, {0, 0, 0, 0}}},{22476, {{33958, 1, 0, 0},{33958, 1, 1, 19}, {0, 0, 0, 0}}},{22476, {{33958, 1, 0, 0},{33958, 1, 1, 19}, {0, 0, 0, 0}}},{22477, {{33958, 1, 0, 0},{33958, 1, 1, 18}, {0, 0, 0, 0}}},{22477, {{33958, 1, 0, 0},{33958, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22478, {{33959, 1, 0, 0},{33959, 1, 1, 16}, {0, 0, 0, 0}}},{22478, {{33959, 1, 0, 0},{33959, 1, 1, 16}, {0, 0, 0, 0}}},{22479, {{33959, 1, 0, 0},{33959, 1, 1, 17}, {0, 0, 0, 0}}},{22479, {{33959, 1, 0, 0},{33959, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {6, {{22480, {{33960, 1, 0, 0},{33960, 1, 1, 17}, {0, 0, 0, 0}}},{22480, {{33960, 1, 0, 0},{33960, 1, 1, 17}, {0, 0, 0, 0}}},{22481, {{33960, 1, 0, 0},{33960, 1, 1, 19}, {0, 0, 0, 0}}},{22481, {{33960, 1, 0, 0},{33960, 1, 1, 19}, {0, 0, 0, 0}}},{22482, {{33960, 1, 0, 0},{33960, 1, 1, 18}, {0, 0, 0, 0}}},{22482, {{33960, 1, 0, 0},{33960, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22484, {{33961, 1, 0, 0},{33961, 1, 1, 17}, {0, 0, 0, 0}}},{22484, {{33961, 1, 0, 0},{33961, 1, 1, 17}, {0, 0, 0, 0}}},{22485, {{33961, 1, 0, 0},{33961, 1, 1, 18}, {0, 0, 0, 0}}},{22485, {{33961, 1, 0, 0},{33961, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22486, {{33962, 1, 0, 0},{33962, 1, 1, 17}, {0, 0, 0, 0}}},{22486, {{33962, 1, 0, 0},{33962, 1, 1, 17}, {0, 0, 0, 0}}},{22487, {{33962, 1, 0, 0},{33962, 1, 1, 18}, {0, 0, 0, 0}}},{22487, {{33962, 1, 0, 0},{33962, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22488, {{33963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22489, {{33964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22490, {{33965, 1, 0, 0},{33965, 1, 1, 17}, {0, 0, 0, 0}}},{22490, {{33965, 1, 0, 0},{33965, 1, 1, 17}, {0, 0, 0, 0}}},{22491, {{33965, 1, 0, 0},{33965, 1, 1, 18}, {0, 0, 0, 0}}},{22491, {{33965, 1, 0, 0},{33965, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22492, {{33966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22493, {{33967, 1, 0, 0},{33967, 1, 1, 17}, {0, 0, 0, 0}}},{22493, {{33967, 1, 0, 0},{33967, 1, 1, 17}, {0, 0, 0, 0}}},{22494, {{33967, 1, 0, 0},{33967, 1, 1, 18}, {0, 0, 0, 0}}},{22494, {{33967, 1, 0, 0},{33967, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22495, {{33968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22496, {{33969, 1, 0, 0},{33969, 1, 1, 17}, {0, 0, 0, 0}}},{22496, {{33969, 1, 0, 0},{33969, 1, 1, 17}, {0, 0, 0, 0}}},{22497, {{33969, 1, 0, 0},{33969, 1, 1, 18}, {0, 0, 0, 0}}},{22497, {{33969, 1, 0, 0},{33969, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22498, {{33970, 1, 0, 0},{33970, 1, 1, 17}, {0, 0, 0, 0}}},{22498, {{33970, 1, 0, 0},{33970, 1, 1, 17}, {0, 0, 0, 0}}},{22499, {{33970, 1, 0, 0},{33970, 1, 1, 18}, {0, 0, 0, 0}}},{22499, {{33970, 1, 0, 0},{33970, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22500, {{33971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22501, {{33972, 1, 0, 0},{33972, 1, 1, 17}, {0, 0, 0, 0}}},{22501, {{33972, 1, 0, 0},{33972, 1, 1, 17}, {0, 0, 0, 0}}},{22502, {{33972, 1, 0, 0},{33972, 1, 1, 18}, {0, 0, 0, 0}}},{22502, {{33972, 1, 0, 0},{33972, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22503, {{33973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22504, {{33974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22505, {{33975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22506, {{33976, 1, 0, 0},{33976, 1, 1, 17}, {0, 0, 0, 0}}},{22506, {{33976, 1, 0, 0},{33976, 1, 1, 17}, {0, 0, 0, 0}}},{22507, {{33976, 1, 0, 0},{33976, 1, 1, 18}, {0, 0, 0, 0}}},{22507, {{33976, 1, 0, 0},{33976, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{22508, {{33977, 1, 0, 0},{33977, 1, 1, 17}, {0, 0, 0, 0}}},{22508, {{33977, 1, 0, 0},{33977, 1, 1, 17}, {0, 0, 0, 0}}},{22509, {{33977, 1, 0, 0},{33977, 1, 1, 19}, {0, 0, 0, 0}}},{22509, {{33977, 1, 0, 0},{33977, 1, 1, 19}, {0, 0, 0, 0}}},{22510, {{33977, 1, 0, 0},{33977, 1, 1, 18}, {0, 0, 0, 0}}},{22510, {{33977, 1, 0, 0},{33977, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22511, {{33978, 1, 0, 0},{33978, 1, 1, 17}, {0, 0, 0, 0}}},{22511, {{33978, 1, 0, 0},{33978, 1, 1, 17}, {0, 0, 0, 0}}},{22512, {{33978, 1, 0, 0},{33978, 1, 1, 18}, {0, 0, 0, 0}}},{22512, {{33978, 1, 0, 0},{33978, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{22513, {{33979, 1, 0, 0},{33979, 1, 1, 17}, {0, 0, 0, 0}}},{22513, {{33979, 1, 0, 0},{33979, 1, 1, 17}, {0, 0, 0, 0}}},{22514, {{33979, 1, 0, 0},{33979, 1, 1, 18}, {0, 0, 0, 0}}},{22514, {{33979, 1, 0, 0},{33979, 1, 1, 18}, {0, 0, 0, 0}}},{22515, {{33979, 1, 0, 0},{33979, 1, 1, 19}, {0, 0, 0, 0}}},{22515, {{33979, 1, 0, 0},{33979, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{22516, {{33980, 1, 0, 0},{33980, 1, 1, 17}, {0, 0, 0, 0}}},{22516, {{33980, 1, 0, 0},{33980, 1, 1, 17}, {0, 0, 0, 0}}},{22517, {{33980, 1, 0, 0},{33980, 1, 1, 18}, {0, 0, 0, 0}}},{22517, {{33980, 1, 0, 0},{33980, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22518, {{33981, 1, 0, 0},{33981, 1, 1, 17}, {0, 0, 0, 0}}},{22518, {{33981, 1, 0, 0},{33981, 1, 1, 17}, {0, 0, 0, 0}}},{22519, {{33981, 1, 0, 0},{33981, 1, 1, 18}, {0, 0, 0, 0}}},{22519, {{33981, 1, 0, 0},{33981, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22520, {{33982, 1, 0, 0},{33982, 1, 1, 17}, {0, 0, 0, 0}}},{22520, {{33982, 1, 0, 0},{33982, 1, 1, 17}, {0, 0, 0, 0}}},{22521, {{33982, 1, 0, 0},{33982, 1, 1, 18}, {0, 0, 0, 0}}},{22521, {{33982, 1, 0, 0},{33982, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22522, {{33983, 1, 0, 0},{33983, 1, 1, 17}, {0, 0, 0, 0}}},{22522, {{33983, 1, 0, 0},{33983, 1, 1, 17}, {0, 0, 0, 0}}},{22523, {{33983, 1, 0, 0},{33983, 1, 1, 18}, {0, 0, 0, 0}}},{22523, {{33983, 1, 0, 0},{33983, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22525, {{33984, 1, 0, 0},{33984, 1, 1, 17}, {0, 0, 0, 0}}},{22525, {{33984, 1, 0, 0},{33984, 1, 1, 17}, {0, 0, 0, 0}}},{22526, {{33984, 1, 0, 0},{33984, 1, 1, 18}, {0, 0, 0, 0}}},{22526, {{33984, 1, 0, 0},{33984, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22527, {{33985, 1, 0, 0},{33985, 1, 1, 17}, {0, 0, 0, 0}}},{22527, {{33985, 1, 0, 0},{33985, 1, 1, 17}, {0, 0, 0, 0}}},{22528, {{33985, 1, 0, 0},{33985, 1, 1, 18}, {0, 0, 0, 0}}},{22528, {{33985, 1, 0, 0},{33985, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22529, {{33986, 1, 0, 0},{33986, 1, 1, 17}, {0, 0, 0, 0}}},{22529, {{33986, 1, 0, 0},{33986, 1, 1, 17}, {0, 0, 0, 0}}},{22530, {{33986, 1, 0, 0},{33986, 1, 1, 18}, {0, 0, 0, 0}}},{22530, {{33986, 1, 0, 0},{33986, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22531, {{33987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22532, {{33988, 1, 0, 0},{33988, 1, 1, 17}, {0, 0, 0, 0}}},{22532, {{33988, 1, 0, 0},{33988, 1, 1, 17}, {0, 0, 0, 0}}},{22533, {{33988, 1, 0, 0},{33988, 1, 1, 18}, {0, 0, 0, 0}}},{22533, {{33988, 1, 0, 0},{33988, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22534, {{33989, 1, 0, 0},{33989, 1, 1, 16}, {0, 0, 0, 0}}},{22534, {{33989, 1, 0, 0},{33989, 1, 1, 16}, {0, 0, 0, 0}}},{22535, {{33989, 1, 0, 0},{33989, 1, 1, 17}, {0, 0, 0, 0}}},{22535, {{33989, 1, 0, 0},{33989, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{22536, {{33990, 1, 0, 0},{33990, 1, 1, 17}, {0, 0, 0, 0}}},{22536, {{33990, 1, 0, 0},{33990, 1, 1, 17}, {0, 0, 0, 0}}},{22537, {{33990, 1, 0, 0},{33990, 1, 1, 18}, {0, 0, 0, 0}}},{22537, {{33990, 1, 0, 0},{33990, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22538, {{33991, 1, 0, 0},{33991, 1, 1, 17}, {0, 0, 0, 0}}},{22538, {{33991, 1, 0, 0},{33991, 1, 1, 17}, {0, 0, 0, 0}}},{22539, {{33991, 1, 0, 0},{33991, 1, 1, 18}, {0, 0, 0, 0}}},{22539, {{33991, 1, 0, 0},{33991, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{22540, {{33992, 1, 0, 0},{33992, 1, 1, 17}, {0, 0, 0, 0}}},{22540, {{33992, 1, 0, 0},{33992, 1, 1, 17}, {0, 0, 0, 0}}},{22541, {{33992, 1, 0, 0},{33992, 1, 1, 18}, {0, 0, 0, 0}}},{22541, {{33992, 1, 0, 0},{33992, 1, 1, 18}, {0, 0, 0, 0}}},{60115, {{33992, 1, 0, 0},{33992, 1, 1, 20}, {0, 0, 0, 0}}},{60115, {{33992, 1, 0, 0},{33992, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{22543, {{33993, 1, 0, 0},{33993, 1, 1, 17}, {0, 0, 0, 0}}},{22543, {{33993, 1, 0, 0},{33993, 1, 1, 17}, {0, 0, 0, 0}}},{22544, {{33993, 1, 0, 0},{33993, 1, 1, 18}, {0, 0, 0, 0}}},{22544, {{33993, 1, 0, 0},{33993, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {8, {{22545, {{33994, 1, 0, 0},{33994, 1, 1, 17}, {0, 0, 0, 0}}},{22545, {{33994, 1, 0, 0},{33994, 1, 1, 17}, {0, 0, 0, 0}}},{22546, {{33994, 1, 0, 0},{33994, 1, 1, 18}, {0, 0, 0, 0}}},{22546, {{33994, 1, 0, 0},{33994, 1, 1, 18}, {0, 0, 0, 0}}},{22547, {{33994, 1, 0, 0},{33994, 1, 1, 20}, {0, 0, 0, 0}}},{22547, {{33994, 1, 0, 0},{33994, 1, 1, 20}, {0, 0, 0, 0}}},{22548, {{33994, 1, 0, 0},{33994, 1, 1, 19}, {0, 0, 0, 0}}},{22548, {{33994, 1, 0, 0},{33994, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{22549, {{33995, 1, 0, 0},{33995, 1, 1, 17}, {0, 0, 0, 0}}},{22549, {{33995, 1, 0, 0},{33995, 1, 1, 17}, {0, 0, 0, 0}}},{22550, {{33995, 1, 0, 0},{33995, 1, 1, 18}, {0, 0, 0, 0}}},{22550, {{33995, 1, 0, 0},{33995, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22551, {{33996, 1, 0, 0},{33996, 1, 1, 17}, {0, 0, 0, 0}}},{22551, {{33996, 1, 0, 0},{33996, 1, 1, 17}, {0, 0, 0, 0}}},{22552, {{33996, 1, 0, 0},{33996, 1, 1, 18}, {0, 0, 0, 0}}},{22552, {{33996, 1, 0, 0},{33996, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22553, {{33997, 1, 0, 0},{33997, 1, 1, 17}, {0, 0, 0, 0}}},{22553, {{33997, 1, 0, 0},{33997, 1, 1, 17}, {0, 0, 0, 0}}},{22554, {{33997, 1, 0, 0},{33997, 1, 1, 18}, {0, 0, 0, 0}}},{22554, {{33997, 1, 0, 0},{33997, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22555, {{33998, 1, 0, 0},{33998, 1, 1, 17}, {0, 0, 0, 0}}},{22555, {{33998, 1, 0, 0},{33998, 1, 1, 17}, {0, 0, 0, 0}}},{22556, {{33998, 1, 0, 0},{33998, 1, 1, 18}, {0, 0, 0, 0}}},{22556, {{33998, 1, 0, 0},{33998, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{22558, {{33999, 1, 0, 0},{33999, 1, 1, 17}, {0, 0, 0, 0}}},{22558, {{33999, 1, 0, 0},{33999, 1, 1, 17}, {0, 0, 0, 0}}},{22559, {{33999, 1, 0, 0},{33999, 1, 1, 19}, {0, 0, 0, 0}}},{22559, {{33999, 1, 0, 0},{33999, 1, 1, 19}, {0, 0, 0, 0}}},{22560, {{33999, 1, 0, 0},{33999, 1, 1, 18}, {0, 0, 0, 0}}},{22560, {{33999, 1, 0, 0},{33999, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{22561, {{34000, 1, 0, 0},{34000, 1, 1, 17}, {0, 0, 0, 0}}},{22561, {{34000, 1, 0, 0},{34000, 1, 1, 17}, {0, 0, 0, 0}}},{22562, {{34000, 1, 0, 0},{34000, 1, 1, 19}, {0, 0, 0, 0}}},{22562, {{34000, 1, 0, 0},{34000, 1, 1, 19}, {0, 0, 0, 0}}},{22563, {{34000, 1, 0, 0},{34000, 1, 1, 18}, {0, 0, 0, 0}}},{22563, {{34000, 1, 0, 0},{34000, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22564, {{34001, 1, 0, 0},{34001, 1, 1, 17}, {0, 0, 0, 0}}},{22564, {{34001, 1, 0, 0},{34001, 1, 1, 17}, {0, 0, 0, 0}}},{22565, {{34001, 1, 0, 0},{34001, 1, 1, 18}, {0, 0, 0, 0}}},{22565, {{34001, 1, 0, 0},{34001, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22567, {{34002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22568, {{34003, 1, 0, 0},{34003, 1, 1, 17}, {0, 0, 0, 0}}},{22568, {{34003, 1, 0, 0},{34003, 1, 1, 17}, {0, 0, 0, 0}}},{22569, {{34003, 1, 0, 0},{34003, 1, 1, 18}, {0, 0, 0, 0}}},{22569, {{34003, 1, 0, 0},{34003, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22570, {{34004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22571, {{34005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22572, {{34006, 1, 0, 0},{34006, 1, 1, 17}, {0, 0, 0, 0}}},{22572, {{34006, 1, 0, 0},{34006, 1, 1, 17}, {0, 0, 0, 0}}},{22573, {{34006, 1, 0, 0},{34006, 1, 1, 18}, {0, 0, 0, 0}}},{22573, {{34006, 1, 0, 0},{34006, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{22575, {{34008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{22576, {{34009, 1, 0, 0},{34009, 1, 1, 17}, {0, 0, 0, 0}}},{22576, {{34009, 1, 0, 0},{34009, 1, 1, 17}, {0, 0, 0, 0}}},{22577, {{34009, 1, 0, 0},{34009, 1, 1, 18}, {0, 0, 0, 0}}},{22577, {{34009, 1, 0, 0},{34009, 1, 1, 18}, {0, 0, 0, 0}}},{22578, {{34009, 1, 0, 0},{34009, 1, 1, 19}, {0, 0, 0, 0}}},{22578, {{34009, 1, 0, 0},{34009, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{22579, {{34010, 1, 0, 0},{34010, 1, 1, 17}, {0, 0, 0, 0}}},{22579, {{34010, 1, 0, 0},{34010, 1, 1, 17}, {0, 0, 0, 0}}},{22580, {{34010, 1, 0, 0},{34010, 1, 1, 18}, {0, 0, 0, 0}}},{22580, {{34010, 1, 0, 0},{34010, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{22581, {{34012, 1, 0, 0},{34012, 1, 1, 18}, {0, 0, 0, 0}}},{22581, {{34012, 1, 0, 0},{34012, 1, 1, 18}, {0, 0, 0, 0}}},{22582, {{34012, 1, 0, 0},{34012, 1, 1, 17}, {0, 0, 0, 0}}},{22582, {{34012, 1, 0, 0},{34012, 1, 1, 17}, {0, 0, 0, 0}}},{22583, {{34012, 1, 0, 0},{34012, 1, 1, 19}, {0, 0, 0, 0}}},{22583, {{34012, 1, 0, 0},{34012, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{22587, {{34014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{22589, {{34023, 1, 0, 0},{34023, 1, 1, 18}, {0, 0, 0, 0}}},{22589, {{34023, 1, 0, 0},{34023, 1, 1, 18}, {0, 0, 0, 0}}},{22590, {{34023, 1, 0, 0},{34023, 1, 1, 17}, {0, 0, 0, 0}}},{22590, {{34023, 1, 0, 0},{34023, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{22591, {{34024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22592, {{34025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22593, {{34026, 1, 0, 0},{34026, 1, 1, 18}, {0, 0, 0, 0}}},{22593, {{34026, 1, 0, 0},{34026, 1, 1, 18}, {0, 0, 0, 0}}},{22594, {{34026, 1, 0, 0},{34026, 1, 1, 19}, {0, 0, 0, 0}}},{22594, {{34026, 1, 0, 0},{34026, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{22595, {{34027, 1, 0, 0},{34027, 1, 1, 17}, {0, 0, 0, 0}}},{22595, {{34027, 1, 0, 0},{34027, 1, 1, 17}, {0, 0, 0, 0}}},{22596, {{34027, 1, 0, 0},{34027, 1, 1, 16}, {0, 0, 0, 0}}},{22596, {{34027, 1, 0, 0},{34027, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {10, {{22597, {{34028, 1, 0, 0},{34028, 1, 1, 18}, {0, 0, 0, 0}}},{22597, {{34028, 1, 0, 0},{34028, 1, 1, 18}, {0, 0, 0, 0}}},{22598, {{34028, 1, 0, 0},{34028, 1, 1, 21}, {0, 0, 0, 0}}},{22598, {{34028, 1, 0, 0},{34028, 1, 1, 21}, {0, 0, 0, 0}}},{22599, {{34028, 1, 0, 0},{34028, 1, 1, 19}, {0, 0, 0, 0}}},{22599, {{34028, 1, 0, 0},{34028, 1, 1, 19}, {0, 0, 0, 0}}},{22600, {{34028, 1, 0, 0},{34028, 1, 1, 20}, {0, 0, 0, 0}}},{22600, {{34028, 1, 0, 0},{34028, 1, 1, 20}, {0, 0, 0, 0}}},{22601, {{34028, 1, 0, 0},{34028, 1, 1, 22}, {0, 0, 0, 0}}},{22601, {{34028, 1, 0, 0},{34028, 1, 1, 22}, {0, 0, 0, 0}}}}},
    {1, {{22602, {{34029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {10, {{22603, {{34030, 1, 0, 0},{34030, 1, 1, 18}, {0, 0, 0, 0}}},{22603, {{34030, 1, 0, 0},{34030, 1, 1, 18}, {0, 0, 0, 0}}},{22604, {{34030, 1, 0, 0},{34030, 1, 1, 21}, {0, 0, 0, 0}}},{22604, {{34030, 1, 0, 0},{34030, 1, 1, 21}, {0, 0, 0, 0}}},{22605, {{34030, 1, 0, 0},{34030, 1, 1, 20}, {0, 0, 0, 0}}},{22605, {{34030, 1, 0, 0},{34030, 1, 1, 20}, {0, 0, 0, 0}}},{22606, {{34030, 1, 0, 0},{34030, 1, 1, 19}, {0, 0, 0, 0}}},{22606, {{34030, 1, 0, 0},{34030, 1, 1, 19}, {0, 0, 0, 0}}},{22607, {{34030, 1, 0, 0},{34030, 1, 1, 22}, {0, 0, 0, 0}}},{22607, {{34030, 1, 0, 0},{34030, 1, 1, 22}, {0, 0, 0, 0}}}}},
    {4, {{22609, {{34031, 1, 0, 0},{34031, 1, 1, 17}, {0, 0, 0, 0}}},{22609, {{34031, 1, 0, 0},{34031, 1, 1, 17}, {0, 0, 0, 0}}},{22610, {{34031, 1, 0, 0},{34031, 1, 1, 18}, {0, 0, 0, 0}}},{22610, {{34031, 1, 0, 0},{34031, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22611, {{34032, 1, 0, 0},{34032, 1, 1, 18}, {0, 0, 0, 0}}},{22611, {{34032, 1, 0, 0},{34032, 1, 1, 18}, {0, 0, 0, 0}}},{22612, {{34032, 1, 0, 0},{34032, 1, 1, 17}, {0, 0, 0, 0}}},{22612, {{34032, 1, 0, 0},{34032, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {6, {{22613, {{34033, 1, 0, 0},{34033, 1, 1, 19}, {0, 0, 0, 0}}},{22613, {{34033, 1, 0, 0},{34033, 1, 1, 19}, {0, 0, 0, 0}}},{22614, {{34033, 1, 0, 0},{34033, 1, 1, 17}, {0, 0, 0, 0}}},{22614, {{34033, 1, 0, 0},{34033, 1, 1, 17}, {0, 0, 0, 0}}},{22615, {{34033, 1, 0, 0},{34033, 1, 1, 18}, {0, 0, 0, 0}}},{22615, {{34033, 1, 0, 0},{34033, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22616, {{34034, 1, 0, 0},{34034, 1, 1, 17}, {0, 0, 0, 0}}},{22616, {{34034, 1, 0, 0},{34034, 1, 1, 17}, {0, 0, 0, 0}}},{22617, {{34034, 1, 0, 0},{34034, 1, 1, 18}, {0, 0, 0, 0}}},{22617, {{34034, 1, 0, 0},{34034, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22618, {{34035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{22619, {{34036, 1, 0, 0},{34036, 1, 1, 18}, {0, 0, 0, 0}}},{22619, {{34036, 1, 0, 0},{34036, 1, 1, 18}, {0, 0, 0, 0}}},{22620, {{34036, 1, 0, 0},{34036, 1, 1, 19}, {0, 0, 0, 0}}},{22620, {{34036, 1, 0, 0},{34036, 1, 1, 19}, {0, 0, 0, 0}}},{22621, {{34036, 1, 0, 0},{34036, 1, 1, 20}, {0, 0, 0, 0}}},{22621, {{34036, 1, 0, 0},{34036, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{22622, {{34037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22623, {{34038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22624, {{34039, 1, 0, 0},{34039, 1, 1, 17}, {0, 0, 0, 0}}},{22624, {{34039, 1, 0, 0},{34039, 1, 1, 17}, {0, 0, 0, 0}}},{22625, {{34039, 1, 0, 0},{34039, 1, 1, 18}, {0, 0, 0, 0}}},{22625, {{34039, 1, 0, 0},{34039, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22626, {{34040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22627, {{34041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22628, {{34042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22629, {{34043, 1, 0, 0},{34043, 1, 1, 17}, {0, 0, 0, 0}}},{22629, {{34043, 1, 0, 0},{34043, 1, 1, 17}, {0, 0, 0, 0}}},{22630, {{34043, 1, 0, 0},{34043, 1, 1, 18}, {0, 0, 0, 0}}},{22630, {{34043, 1, 0, 0},{34043, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{22631, {{34044, 1, 0, 0},{34044, 1, 1, 17}, {0, 0, 0, 0}}},{22631, {{34044, 1, 0, 0},{34044, 1, 1, 17}, {0, 0, 0, 0}}},{22632, {{34044, 1, 0, 0},{34044, 1, 1, 18}, {0, 0, 0, 0}}},{22632, {{34044, 1, 0, 0},{34044, 1, 1, 18}, {0, 0, 0, 0}}},{22633, {{34044, 1, 0, 0},{34044, 1, 1, 19}, {0, 0, 0, 0}}},{22633, {{34044, 1, 0, 0},{34044, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {6, {{22634, {{34045, 1, 0, 0},{34045, 1, 1, 17}, {0, 0, 0, 0}}},{22634, {{34045, 1, 0, 0},{34045, 1, 1, 17}, {0, 0, 0, 0}}},{22635, {{34045, 1, 0, 0},{34045, 1, 1, 19}, {0, 0, 0, 0}}},{22635, {{34045, 1, 0, 0},{34045, 1, 1, 19}, {0, 0, 0, 0}}},{22636, {{34045, 1, 0, 0},{34045, 1, 1, 18}, {0, 0, 0, 0}}},{22636, {{34045, 1, 0, 0},{34045, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22637, {{34046, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22638, {{34047, 1, 0, 0},{34047, 1, 1, 17}, {0, 0, 0, 0}}},{22638, {{34047, 1, 0, 0},{34047, 1, 1, 17}, {0, 0, 0, 0}}},{22639, {{34047, 1, 0, 0},{34047, 1, 1, 18}, {0, 0, 0, 0}}},{22639, {{34047, 1, 0, 0},{34047, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22640, {{34048, 1, 0, 0},{34048, 1, 1, 17}, {0, 0, 0, 0}}},{22640, {{34048, 1, 0, 0},{34048, 1, 1, 17}, {0, 0, 0, 0}}},{22641, {{34048, 1, 0, 0},{34048, 1, 1, 18}, {0, 0, 0, 0}}},{22641, {{34048, 1, 0, 0},{34048, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22642, {{34049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22644, {{34050, 1, 0, 0},{34050, 1, 1, 17}, {0, 0, 0, 0}}},{22644, {{34050, 1, 0, 0},{34050, 1, 1, 17}, {0, 0, 0, 0}}},{22645, {{34050, 1, 0, 0},{34050, 1, 1, 18}, {0, 0, 0, 0}}},{22645, {{34050, 1, 0, 0},{34050, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22646, {{34051, 1, 0, 0},{34051, 1, 1, 17}, {0, 0, 0, 0}}},{22646, {{34051, 1, 0, 0},{34051, 1, 1, 17}, {0, 0, 0, 0}}},{22647, {{34051, 1, 0, 0},{34051, 1, 1, 18}, {0, 0, 0, 0}}},{22647, {{34051, 1, 0, 0},{34051, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22648, {{34052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22649, {{34053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22650, {{34054, 1, 0, 0},{34054, 1, 1, 17}, {0, 0, 0, 0}}},{22650, {{34054, 1, 0, 0},{34054, 1, 1, 17}, {0, 0, 0, 0}}},{22651, {{34054, 1, 0, 0},{34054, 1, 1, 18}, {0, 0, 0, 0}}},{22651, {{34054, 1, 0, 0},{34054, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22652, {{34055, 1, 0, 0},{34055, 1, 1, 17}, {0, 0, 0, 0}}},{22652, {{34055, 1, 0, 0},{34055, 1, 1, 17}, {0, 0, 0, 0}}},{22653, {{34055, 1, 0, 0},{34055, 1, 1, 18}, {0, 0, 0, 0}}},{22653, {{34055, 1, 0, 0},{34055, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22654, {{34056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22656, {{34057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22657, {{34058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22658, {{34059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22659, {{34060, 1, 0, 0},{34060, 1, 1, 17}, {0, 0, 0, 0}}},{22659, {{34060, 1, 0, 0},{34060, 1, 1, 17}, {0, 0, 0, 0}}},{22660, {{34060, 1, 0, 0},{34060, 1, 1, 18}, {0, 0, 0, 0}}},{22660, {{34060, 1, 0, 0},{34060, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22661, {{34061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22662, {{34062, 1, 0, 0},{34062, 1, 1, 17}, {0, 0, 0, 0}}},{22662, {{34062, 1, 0, 0},{34062, 1, 1, 17}, {0, 0, 0, 0}}},{22663, {{34062, 1, 0, 0},{34062, 1, 1, 18}, {0, 0, 0, 0}}},{22663, {{34062, 1, 0, 0},{34062, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22664, {{34063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22665, {{34064, 1, 0, 0},{34064, 1, 1, 17}, {0, 0, 0, 0}}},{22665, {{34064, 1, 0, 0},{34064, 1, 1, 17}, {0, 0, 0, 0}}},{22666, {{34064, 1, 0, 0},{34064, 1, 1, 18}, {0, 0, 0, 0}}},{22666, {{34064, 1, 0, 0},{34064, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {10, {{22667, {{34065, 1, 0, 0},{34065, 1, 1, 18}, {0, 0, 0, 0}}},{22667, {{34065, 1, 0, 0},{34065, 1, 1, 18}, {0, 0, 0, 0}}},{22668, {{34065, 1, 0, 0},{34065, 1, 1, 20}, {0, 0, 0, 0}}},{22668, {{34065, 1, 0, 0},{34065, 1, 1, 20}, {0, 0, 0, 0}}},{22669, {{34065, 1, 0, 0},{34065, 1, 1, 19}, {0, 0, 0, 0}}},{22669, {{34065, 1, 0, 0},{34065, 1, 1, 19}, {0, 0, 0, 0}}},{22670, {{34065, 1, 0, 0},{34065, 1, 1, 21}, {0, 0, 0, 0}}},{22670, {{34065, 1, 0, 0},{34065, 1, 1, 21}, {0, 0, 0, 0}}},{58507, {{34065, 1, 0, 0},{34065, 1, 1, 22}, {0, 0, 0, 0}}},{58507, {{34065, 1, 0, 0},{34065, 1, 1, 22}, {0, 0, 0, 0}}}}},
    {4, {{22671, {{34066, 1, 0, 0},{34066, 1, 1, 16}, {0, 0, 0, 0}}},{22671, {{34066, 1, 0, 0},{34066, 1, 1, 16}, {0, 0, 0, 0}}},{22672, {{34066, 1, 0, 0},{34066, 1, 1, 17}, {0, 0, 0, 0}}},{22672, {{34066, 1, 0, 0},{34066, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {8, {{22673, {{34067, 1, 0, 0},{34067, 1, 1, 17}, {0, 0, 0, 0}}},{22673, {{34067, 1, 0, 0},{34067, 1, 1, 17}, {0, 0, 0, 0}}},{22674, {{34067, 1, 0, 0},{34067, 1, 1, 18}, {0, 0, 0, 0}}},{22674, {{34067, 1, 0, 0},{34067, 1, 1, 18}, {0, 0, 0, 0}}},{22675, {{34067, 1, 0, 0},{34067, 1, 1, 19}, {0, 0, 0, 0}}},{22675, {{34067, 1, 0, 0},{34067, 1, 1, 19}, {0, 0, 0, 0}}},{22676, {{34067, 1, 0, 0},{34067, 1, 1, 20}, {0, 0, 0, 0}}},{22676, {{34067, 1, 0, 0},{34067, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{22677, {{34068, 1, 0, 0},{34068, 1, 1, 17}, {0, 0, 0, 0}}},{22677, {{34068, 1, 0, 0},{34068, 1, 1, 17}, {0, 0, 0, 0}}},{22678, {{34068, 1, 0, 0},{34068, 1, 1, 18}, {0, 0, 0, 0}}},{22678, {{34068, 1, 0, 0},{34068, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22679, {{34069, 1, 0, 0},{34069, 1, 1, 17}, {0, 0, 0, 0}}},{22679, {{34069, 1, 0, 0},{34069, 1, 1, 17}, {0, 0, 0, 0}}},{22680, {{34069, 1, 0, 0},{34069, 1, 1, 18}, {0, 0, 0, 0}}},{22680, {{34069, 1, 0, 0},{34069, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22681, {{34070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{22682, {{34071, 1, 0, 0},{34071, 1, 1, 18}, {0, 0, 0, 0}}},{22682, {{34071, 1, 0, 0},{34071, 1, 1, 18}, {0, 0, 0, 0}}},{22683, {{34071, 1, 0, 0},{34071, 1, 1, 20}, {0, 0, 0, 0}}},{22683, {{34071, 1, 0, 0},{34071, 1, 1, 20}, {0, 0, 0, 0}}},{22684, {{34071, 1, 0, 0},{34071, 1, 1, 19}, {0, 0, 0, 0}}},{22684, {{34071, 1, 0, 0},{34071, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{22686, {{34072, 1, 0, 0},{34072, 1, 1, 17}, {0, 0, 0, 0}}},{22686, {{34072, 1, 0, 0},{34072, 1, 1, 17}, {0, 0, 0, 0}}},{22687, {{34072, 1, 0, 0},{34072, 1, 1, 18}, {0, 0, 0, 0}}},{22687, {{34072, 1, 0, 0},{34072, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22688, {{34073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22689, {{34074, 1, 0, 0},{34074, 1, 1, 17}, {0, 0, 0, 0}}},{22689, {{34074, 1, 0, 0},{34074, 1, 1, 17}, {0, 0, 0, 0}}},{22690, {{34074, 1, 0, 0},{34074, 1, 1, 18}, {0, 0, 0, 0}}},{22690, {{34074, 1, 0, 0},{34074, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22691, {{34075, 1, 0, 0},{34075, 1, 1, 17}, {0, 0, 0, 0}}},{22691, {{34075, 1, 0, 0},{34075, 1, 1, 17}, {0, 0, 0, 0}}},{22692, {{34075, 1, 0, 0},{34075, 1, 1, 18}, {0, 0, 0, 0}}},{22692, {{34075, 1, 0, 0},{34075, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22693, {{34076, 1, 0, 0},{34076, 1, 1, 17}, {0, 0, 0, 0}}},{22693, {{34076, 1, 0, 0},{34076, 1, 1, 17}, {0, 0, 0, 0}}},{22694, {{34076, 1, 0, 0},{34076, 1, 1, 18}, {0, 0, 0, 0}}},{22694, {{34076, 1, 0, 0},{34076, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22695, {{34077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22696, {{34078, 1, 0, 0},{34078, 1, 1, 17}, {0, 0, 0, 0}}},{22696, {{34078, 1, 0, 0},{34078, 1, 1, 17}, {0, 0, 0, 0}}},{22697, {{34078, 1, 0, 0},{34078, 1, 1, 18}, {0, 0, 0, 0}}},{22697, {{34078, 1, 0, 0},{34078, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22698, {{34079, 1, 0, 0},{34079, 1, 1, 17}, {0, 0, 0, 0}}},{22698, {{34079, 1, 0, 0},{34079, 1, 1, 17}, {0, 0, 0, 0}}},{22699, {{34079, 1, 0, 0},{34079, 1, 1, 18}, {0, 0, 0, 0}}},{22699, {{34079, 1, 0, 0},{34079, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22700, {{34080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{22701, {{34081, 1, 0, 0},{34081, 1, 1, 17}, {0, 0, 0, 0}}},{22701, {{34081, 1, 0, 0},{34081, 1, 1, 17}, {0, 0, 0, 0}}},{22702, {{34081, 1, 0, 0},{34081, 1, 1, 18}, {0, 0, 0, 0}}},{22702, {{34081, 1, 0, 0},{34081, 1, 1, 18}, {0, 0, 0, 0}}},{60130, {{34081, 1, 0, 0},{34081, 1, 1, 19}, {0, 0, 0, 0}}},{60130, {{34081, 1, 0, 0},{34081, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{22703, {{34082, 1, 0, 0},{34082, 1, 1, 17}, {0, 0, 0, 0}}},{22703, {{34082, 1, 0, 0},{34082, 1, 1, 17}, {0, 0, 0, 0}}},{22704, {{34082, 1, 0, 0},{34082, 1, 1, 18}, {0, 0, 0, 0}}},{22704, {{34082, 1, 0, 0},{34082, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{22705, {{34083, 1, 0, 0},{34083, 1, 1, 19}, {0, 0, 0, 0}}},{22705, {{34083, 1, 0, 0},{34083, 1, 1, 19}, {0, 0, 0, 0}}},{22706, {{34083, 1, 0, 0},{34083, 1, 1, 17}, {0, 0, 0, 0}}},{22706, {{34083, 1, 0, 0},{34083, 1, 1, 17}, {0, 0, 0, 0}}},{22707, {{34083, 1, 0, 0},{34083, 1, 1, 18}, {0, 0, 0, 0}}},{22707, {{34083, 1, 0, 0},{34083, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22708, {{34084, 1, 0, 0},{34084, 1, 1, 17}, {0, 0, 0, 0}}},{22708, {{34084, 1, 0, 0},{34084, 1, 1, 17}, {0, 0, 0, 0}}},{22709, {{34084, 1, 0, 0},{34084, 1, 1, 18}, {0, 0, 0, 0}}},{22709, {{34084, 1, 0, 0},{34084, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {8, {{22710, {{34085, 1, 0, 0},{34085, 1, 1, 18}, {0, 0, 0, 0}}},{22710, {{34085, 1, 0, 0},{34085, 1, 1, 18}, {0, 0, 0, 0}}},{22711, {{34085, 1, 0, 0},{34085, 1, 1, 17}, {0, 0, 0, 0}}},{22711, {{34085, 1, 0, 0},{34085, 1, 1, 17}, {0, 0, 0, 0}}},{22712, {{34085, 1, 0, 0},{34085, 1, 1, 19}, {0, 0, 0, 0}}},{22712, {{34085, 1, 0, 0},{34085, 1, 1, 19}, {0, 0, 0, 0}}},{58531, {{34085, 1, 0, 0},{34085, 1, 1, 20}, {0, 0, 0, 0}}},{58531, {{34085, 1, 0, 0},{34085, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{22713, {{34086, 1, 0, 0},{34086, 1, 1, 17}, {0, 0, 0, 0}}},{22713, {{34086, 1, 0, 0},{34086, 1, 1, 17}, {0, 0, 0, 0}}},{22714, {{34086, 1, 0, 0},{34086, 1, 1, 18}, {0, 0, 0, 0}}},{22714, {{34086, 1, 0, 0},{34086, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22715, {{34087, 1, 0, 0},{34087, 1, 1, 17}, {0, 0, 0, 0}}},{22715, {{34087, 1, 0, 0},{34087, 1, 1, 17}, {0, 0, 0, 0}}},{22716, {{34087, 1, 0, 0},{34087, 1, 1, 18}, {0, 0, 0, 0}}},{22716, {{34087, 1, 0, 0},{34087, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22717, {{34088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22718, {{34089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22719, {{34090, 1, 0, 0},{34090, 1, 1, 17}, {0, 0, 0, 0}}},{22719, {{34090, 1, 0, 0},{34090, 1, 1, 17}, {0, 0, 0, 0}}},{22720, {{34090, 1, 0, 0},{34090, 1, 1, 18}, {0, 0, 0, 0}}},{22720, {{34090, 1, 0, 0},{34090, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22721, {{34091, 1, 0, 0},{34091, 1, 1, 17}, {0, 0, 0, 0}}},{22721, {{34091, 1, 0, 0},{34091, 1, 1, 17}, {0, 0, 0, 0}}},{22722, {{34091, 1, 0, 0},{34091, 1, 1, 18}, {0, 0, 0, 0}}},{22722, {{34091, 1, 0, 0},{34091, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22723, {{34092, 1, 0, 0},{34092, 1, 1, 17}, {0, 0, 0, 0}}},{22723, {{34092, 1, 0, 0},{34092, 1, 1, 17}, {0, 0, 0, 0}}},{22724, {{34092, 1, 0, 0},{34092, 1, 1, 18}, {0, 0, 0, 0}}},{22724, {{34092, 1, 0, 0},{34092, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22725, {{34093, 1, 0, 0},{34093, 1, 1, 17}, {0, 0, 0, 0}}},{22725, {{34093, 1, 0, 0},{34093, 1, 1, 17}, {0, 0, 0, 0}}},{22726, {{34093, 1, 0, 0},{34093, 1, 1, 18}, {0, 0, 0, 0}}},{22726, {{34093, 1, 0, 0},{34093, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22727, {{34094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22728, {{34095, 1, 0, 0},{34095, 1, 1, 17}, {0, 0, 0, 0}}},{22728, {{34095, 1, 0, 0},{34095, 1, 1, 17}, {0, 0, 0, 0}}},{22729, {{34095, 1, 0, 0},{34095, 1, 1, 18}, {0, 0, 0, 0}}},{22729, {{34095, 1, 0, 0},{34095, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22730, {{34096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22731, {{34097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{22732, {{34098, 1, 0, 0},{34098, 1, 1, 17}, {0, 0, 0, 0}}},{22732, {{34098, 1, 0, 0},{34098, 1, 1, 17}, {0, 0, 0, 0}}},{22733, {{34098, 1, 0, 0},{34098, 1, 1, 18}, {0, 0, 0, 0}}},{22733, {{34098, 1, 0, 0},{34098, 1, 1, 18}, {0, 0, 0, 0}}},{60129, {{34098, 1, 0, 0},{34098, 1, 1, 19}, {0, 0, 0, 0}}},{60129, {{34098, 1, 0, 0},{34098, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{22734, {{34099, 1, 0, 0},{34099, 1, 1, 17}, {0, 0, 0, 0}}},{22734, {{34099, 1, 0, 0},{34099, 1, 1, 17}, {0, 0, 0, 0}}},{22735, {{34099, 1, 0, 0},{34099, 1, 1, 18}, {0, 0, 0, 0}}},{22735, {{34099, 1, 0, 0},{34099, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22738, {{34100, 1, 0, 0},{34100, 1, 1, 17}, {0, 0, 0, 0}}},{22738, {{34100, 1, 0, 0},{34100, 1, 1, 17}, {0, 0, 0, 0}}},{22739, {{34100, 1, 0, 0},{34100, 1, 1, 18}, {0, 0, 0, 0}}},{22739, {{34100, 1, 0, 0},{34100, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22740, {{34101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22742, {{34102, 1, 0, 0},{34102, 1, 1, 17}, {0, 0, 0, 0}}},{22742, {{34102, 1, 0, 0},{34102, 1, 1, 17}, {0, 0, 0, 0}}},{22743, {{34102, 1, 0, 0},{34102, 1, 1, 18}, {0, 0, 0, 0}}},{22743, {{34102, 1, 0, 0},{34102, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58504, {{34107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{22744, {{34109, 1, 0, 0},{34109, 1, 1, 19}, {0, 0, 0, 0}}},{22744, {{34109, 1, 0, 0},{34109, 1, 1, 19}, {0, 0, 0, 0}}},{22745, {{34109, 1, 0, 0},{34109, 1, 1, 21}, {0, 0, 0, 0}}},{22745, {{34109, 1, 0, 0},{34109, 1, 1, 21}, {0, 0, 0, 0}}},{22746, {{34109, 1, 0, 0},{34109, 1, 1, 20}, {0, 0, 0, 0}}},{22746, {{34109, 1, 0, 0},{34109, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {6, {{22747, {{34110, 1, 0, 0},{34110, 1, 1, 18}, {0, 0, 0, 0}}},{22747, {{34110, 1, 0, 0},{34110, 1, 1, 18}, {0, 0, 0, 0}}},{22748, {{34110, 1, 0, 0},{34110, 1, 1, 17}, {0, 0, 0, 0}}},{22748, {{34110, 1, 0, 0},{34110, 1, 1, 17}, {0, 0, 0, 0}}},{22749, {{34110, 1, 0, 0},{34110, 1, 1, 19}, {0, 0, 0, 0}}},{22749, {{34110, 1, 0, 0},{34110, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{22750, {{34111, 1, 0, 0},{34111, 1, 1, 17}, {0, 0, 0, 0}}},{22750, {{34111, 1, 0, 0},{34111, 1, 1, 17}, {0, 0, 0, 0}}},{22751, {{34111, 1, 0, 0},{34111, 1, 1, 18}, {0, 0, 0, 0}}},{22751, {{34111, 1, 0, 0},{34111, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22752, {{34112, 1, 0, 0},{34112, 1, 1, 17}, {0, 0, 0, 0}}},{22752, {{34112, 1, 0, 0},{34112, 1, 1, 17}, {0, 0, 0, 0}}},{22753, {{34112, 1, 0, 0},{34112, 1, 1, 18}, {0, 0, 0, 0}}},{22753, {{34112, 1, 0, 0},{34112, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{22754, {{34113, 1, 0, 0},{34113, 1, 1, 17}, {0, 0, 0, 0}}},{22754, {{34113, 1, 0, 0},{34113, 1, 1, 17}, {0, 0, 0, 0}}},{22755, {{34113, 1, 0, 0},{34113, 1, 1, 18}, {0, 0, 0, 0}}},{22755, {{34113, 1, 0, 0},{34113, 1, 1, 18}, {0, 0, 0, 0}}},{22756, {{34113, 1, 0, 0},{34113, 1, 1, 19}, {0, 0, 0, 0}}},{22756, {{34113, 1, 0, 0},{34113, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{22757, {{34114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22760, {{34115, 1, 0, 0},{34115, 1, 1, 18}, {0, 0, 0, 0}}},{22760, {{34115, 1, 0, 0},{34115, 1, 1, 18}, {0, 0, 0, 0}}},{22761, {{34115, 1, 0, 0},{34115, 1, 1, 19}, {0, 0, 0, 0}}},{22761, {{34115, 1, 0, 0},{34115, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{22762, {{34116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22763, {{34117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22764, {{34118, 1, 0, 0},{34118, 1, 1, 17}, {0, 0, 0, 0}}},{22764, {{34118, 1, 0, 0},{34118, 1, 1, 17}, {0, 0, 0, 0}}},{22765, {{34118, 1, 0, 0},{34118, 1, 1, 18}, {0, 0, 0, 0}}},{22765, {{34118, 1, 0, 0},{34118, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22766, {{34119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{22767, {{34120, 1, 0, 0},{34120, 1, 1, 17}, {0, 0, 0, 0}}},{22767, {{34120, 1, 0, 0},{34120, 1, 1, 17}, {0, 0, 0, 0}}},{22768, {{34120, 1, 0, 0},{34120, 1, 1, 18}, {0, 0, 0, 0}}},{22768, {{34120, 1, 0, 0},{34120, 1, 1, 18}, {0, 0, 0, 0}}},{22769, {{34120, 1, 0, 0},{34120, 1, 1, 19}, {0, 0, 0, 0}}},{22769, {{34120, 1, 0, 0},{34120, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{22770, {{34121, 1, 0, 0},{34121, 1, 1, 17}, {0, 0, 0, 0}}},{22770, {{34121, 1, 0, 0},{34121, 1, 1, 17}, {0, 0, 0, 0}}},{22771, {{34121, 1, 0, 0},{34121, 1, 1, 18}, {0, 0, 0, 0}}},{22771, {{34121, 1, 0, 0},{34121, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22772, {{34122, 1, 0, 0},{34122, 1, 1, 17}, {0, 0, 0, 0}}},{22772, {{34122, 1, 0, 0},{34122, 1, 1, 17}, {0, 0, 0, 0}}},{22773, {{34122, 1, 0, 0},{34122, 1, 1, 18}, {0, 0, 0, 0}}},{22773, {{34122, 1, 0, 0},{34122, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22774, {{34123, 1, 0, 0},{34123, 1, 1, 17}, {0, 0, 0, 0}}},{22774, {{34123, 1, 0, 0},{34123, 1, 1, 17}, {0, 0, 0, 0}}},{22775, {{34123, 1, 0, 0},{34123, 1, 1, 18}, {0, 0, 0, 0}}},{22775, {{34123, 1, 0, 0},{34123, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22776, {{34124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22777, {{34125, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22778, {{34126, 1, 0, 0},{34126, 1, 1, 17}, {0, 0, 0, 0}}},{22778, {{34126, 1, 0, 0},{34126, 1, 1, 17}, {0, 0, 0, 0}}},{22779, {{34126, 1, 0, 0},{34126, 1, 1, 18}, {0, 0, 0, 0}}},{22779, {{34126, 1, 0, 0},{34126, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22780, {{34127, 1, 0, 0},{34127, 1, 1, 17}, {0, 0, 0, 0}}},{22780, {{34127, 1, 0, 0},{34127, 1, 1, 17}, {0, 0, 0, 0}}},{22781, {{34127, 1, 0, 0},{34127, 1, 1, 18}, {0, 0, 0, 0}}},{22781, {{34127, 1, 0, 0},{34127, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22782, {{34128, 1, 0, 0},{34128, 1, 1, 17}, {0, 0, 0, 0}}},{22782, {{34128, 1, 0, 0},{34128, 1, 1, 17}, {0, 0, 0, 0}}},{22783, {{34128, 1, 0, 0},{34128, 1, 1, 18}, {0, 0, 0, 0}}},{22783, {{34128, 1, 0, 0},{34128, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22784, {{34129, 1, 0, 0},{34129, 1, 1, 18}, {0, 0, 0, 0}}},{22784, {{34129, 1, 0, 0},{34129, 1, 1, 18}, {0, 0, 0, 0}}},{22785, {{34129, 1, 0, 0},{34129, 1, 1, 19}, {0, 0, 0, 0}}},{22785, {{34129, 1, 0, 0},{34129, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{22786, {{34130, 1, 0, 0},{34130, 1, 1, 17}, {0, 0, 0, 0}}},{22786, {{34130, 1, 0, 0},{34130, 1, 1, 17}, {0, 0, 0, 0}}},{22787, {{34130, 1, 0, 0},{34130, 1, 1, 18}, {0, 0, 0, 0}}},{22787, {{34130, 1, 0, 0},{34130, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22788, {{34131, 1, 0, 0},{34131, 1, 1, 17}, {0, 0, 0, 0}}},{22788, {{34131, 1, 0, 0},{34131, 1, 1, 17}, {0, 0, 0, 0}}},{22789, {{34131, 1, 0, 0},{34131, 1, 1, 18}, {0, 0, 0, 0}}},{22789, {{34131, 1, 0, 0},{34131, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22790, {{34132, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{22791, {{34133, 1, 0, 0},{34133, 1, 1, 18}, {0, 0, 0, 0}}},{22791, {{34133, 1, 0, 0},{34133, 1, 1, 18}, {0, 0, 0, 0}}},{22792, {{34133, 1, 0, 0},{34133, 1, 1, 19}, {0, 0, 0, 0}}},{22792, {{34133, 1, 0, 0},{34133, 1, 1, 19}, {0, 0, 0, 0}}},{22793, {{34133, 1, 0, 0},{34133, 1, 1, 20}, {0, 0, 0, 0}}},{22793, {{34133, 1, 0, 0},{34133, 1, 1, 20}, {0, 0, 0, 0}}},{22794, {{34133, 1, 0, 0},{34133, 1, 1, 21}, {0, 0, 0, 0}}},{22794, {{34133, 1, 0, 0},{34133, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {8, {{22795, {{34134, 1, 0, 0},{34134, 1, 1, 20}, {0, 0, 0, 0}}},{22795, {{34134, 1, 0, 0},{34134, 1, 1, 20}, {0, 0, 0, 0}}},{22796, {{34134, 1, 0, 0},{34134, 1, 1, 17}, {0, 0, 0, 0}}},{22796, {{34134, 1, 0, 0},{34134, 1, 1, 17}, {0, 0, 0, 0}}},{22797, {{34134, 1, 0, 0},{34134, 1, 1, 18}, {0, 0, 0, 0}}},{22797, {{34134, 1, 0, 0},{34134, 1, 1, 18}, {0, 0, 0, 0}}},{22798, {{34134, 1, 0, 0},{34134, 1, 1, 19}, {0, 0, 0, 0}}},{22798, {{34134, 1, 0, 0},{34134, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{22799, {{34135, 1, 0, 0},{34135, 1, 1, 18}, {0, 0, 0, 0}}},{22799, {{34135, 1, 0, 0},{34135, 1, 1, 18}, {0, 0, 0, 0}}},{22800, {{34135, 1, 0, 0},{34135, 1, 1, 17}, {0, 0, 0, 0}}},{22800, {{34135, 1, 0, 0},{34135, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{22801, {{34136, 1, 0, 0},{34136, 1, 1, 17}, {0, 0, 0, 0}}},{22801, {{34136, 1, 0, 0},{34136, 1, 1, 17}, {0, 0, 0, 0}}},{22802, {{34136, 1, 0, 0},{34136, 1, 1, 18}, {0, 0, 0, 0}}},{22802, {{34136, 1, 0, 0},{34136, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {8, {{22803, {{34137, 1, 0, 0},{34137, 1, 1, 17}, {0, 0, 0, 0}}},{22803, {{34137, 1, 0, 0},{34137, 1, 1, 17}, {0, 0, 0, 0}}},{22804, {{34137, 1, 0, 0},{34137, 1, 1, 19}, {0, 0, 0, 0}}},{22804, {{34137, 1, 0, 0},{34137, 1, 1, 19}, {0, 0, 0, 0}}},{22805, {{34137, 1, 0, 0},{34137, 1, 1, 18}, {0, 0, 0, 0}}},{22805, {{34137, 1, 0, 0},{34137, 1, 1, 18}, {0, 0, 0, 0}}},{22806, {{34137, 1, 0, 0},{34137, 1, 1, 20}, {0, 0, 0, 0}}},{22806, {{34137, 1, 0, 0},{34137, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{22807, {{34138, 1, 0, 0},{34138, 1, 1, 17}, {0, 0, 0, 0}}},{22807, {{34138, 1, 0, 0},{34138, 1, 1, 17}, {0, 0, 0, 0}}},{22808, {{34138, 1, 0, 0},{34138, 1, 1, 18}, {0, 0, 0, 0}}},{22808, {{34138, 1, 0, 0},{34138, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22809, {{34139, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22810, {{34140, 1, 0, 0},{34140, 1, 1, 17}, {0, 0, 0, 0}}},{22810, {{34140, 1, 0, 0},{34140, 1, 1, 17}, {0, 0, 0, 0}}},{22811, {{34140, 1, 0, 0},{34140, 1, 1, 18}, {0, 0, 0, 0}}},{22811, {{34140, 1, 0, 0},{34140, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{22812, {{34141, 1, 0, 0},{34141, 1, 1, 19}, {0, 0, 0, 0}}},{22812, {{34141, 1, 0, 0},{34141, 1, 1, 19}, {0, 0, 0, 0}}},{22813, {{34141, 1, 0, 0},{34141, 1, 1, 20}, {0, 0, 0, 0}}},{22813, {{34141, 1, 0, 0},{34141, 1, 1, 20}, {0, 0, 0, 0}}},{22814, {{34141, 1, 0, 0},{34141, 1, 1, 18}, {0, 0, 0, 0}}},{22814, {{34141, 1, 0, 0},{34141, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22815, {{34142, 1, 0, 0},{34142, 1, 1, 17}, {0, 0, 0, 0}}},{22815, {{34142, 1, 0, 0},{34142, 1, 1, 17}, {0, 0, 0, 0}}},{22816, {{34142, 1, 0, 0},{34142, 1, 1, 18}, {0, 0, 0, 0}}},{22816, {{34142, 1, 0, 0},{34142, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22817, {{34143, 1, 0, 0},{34143, 1, 1, 17}, {0, 0, 0, 0}}},{22817, {{34143, 1, 0, 0},{34143, 1, 1, 17}, {0, 0, 0, 0}}},{22818, {{34143, 1, 0, 0},{34143, 1, 1, 18}, {0, 0, 0, 0}}},{22818, {{34143, 1, 0, 0},{34143, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22819, {{34144, 1, 0, 0},{34144, 1, 1, 17}, {0, 0, 0, 0}}},{22819, {{34144, 1, 0, 0},{34144, 1, 1, 17}, {0, 0, 0, 0}}},{22820, {{34144, 1, 0, 0},{34144, 1, 1, 18}, {0, 0, 0, 0}}},{22820, {{34144, 1, 0, 0},{34144, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{22821, {{34145, 1, 0, 0},{34145, 1, 1, 17}, {0, 0, 0, 0}}},{22821, {{34145, 1, 0, 0},{34145, 1, 1, 17}, {0, 0, 0, 0}}},{22822, {{34145, 1, 0, 0},{34145, 1, 1, 18}, {0, 0, 0, 0}}},{22822, {{34145, 1, 0, 0},{34145, 1, 1, 18}, {0, 0, 0, 0}}},{58545, {{34145, 1, 0, 0},{34145, 1, 1, 19}, {0, 0, 0, 0}}},{58545, {{34145, 1, 0, 0},{34145, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{22823, {{34146, 1, 0, 0},{34146, 1, 1, 17}, {0, 0, 0, 0}}},{22823, {{34146, 1, 0, 0},{34146, 1, 1, 17}, {0, 0, 0, 0}}},{22824, {{34146, 1, 0, 0},{34146, 1, 1, 18}, {0, 0, 0, 0}}},{22824, {{34146, 1, 0, 0},{34146, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {10, {{22825, {{34147, 1, 0, 0},{34147, 1, 1, 18}, {0, 0, 0, 0}}},{22825, {{34147, 1, 0, 0},{34147, 1, 1, 18}, {0, 0, 0, 0}}},{22826, {{34147, 1, 0, 0},{34147, 1, 1, 19}, {0, 0, 0, 0}}},{22826, {{34147, 1, 0, 0},{34147, 1, 1, 19}, {0, 0, 0, 0}}},{22827, {{34147, 1, 0, 0},{34147, 1, 1, 20}, {0, 0, 0, 0}}},{22827, {{34147, 1, 0, 0},{34147, 1, 1, 20}, {0, 0, 0, 0}}},{22828, {{34147, 1, 0, 0},{34147, 1, 1, 21}, {0, 0, 0, 0}}},{22828, {{34147, 1, 0, 0},{34147, 1, 1, 21}, {0, 0, 0, 0}}},{22829, {{34147, 1, 0, 0},{34147, 1, 1, 22}, {0, 0, 0, 0}}},{22829, {{34147, 1, 0, 0},{34147, 1, 1, 22}, {0, 0, 0, 0}}}}},
    {4, {{22830, {{34148, 1, 0, 0},{34148, 1, 1, 17}, {0, 0, 0, 0}}},{22830, {{34148, 1, 0, 0},{34148, 1, 1, 17}, {0, 0, 0, 0}}},{22831, {{34148, 1, 0, 0},{34148, 1, 1, 18}, {0, 0, 0, 0}}},{22831, {{34148, 1, 0, 0},{34148, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22832, {{34149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22834, {{34150, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22835, {{34151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{22836, {{34152, 1, 0, 0},{34152, 1, 1, 17}, {0, 0, 0, 0}}},{22836, {{34152, 1, 0, 0},{34152, 1, 1, 17}, {0, 0, 0, 0}}},{22837, {{34152, 1, 0, 0},{34152, 1, 1, 18}, {0, 0, 0, 0}}},{22837, {{34152, 1, 0, 0},{34152, 1, 1, 18}, {0, 0, 0, 0}}},{58518, {{34152, 1, 0, 0},{34152, 1, 1, 19}, {0, 0, 0, 0}}},{58518, {{34152, 1, 0, 0},{34152, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{22838, {{34153, 1, 0, 0},{34153, 1, 1, 17}, {0, 0, 0, 0}}},{22838, {{34153, 1, 0, 0},{34153, 1, 1, 17}, {0, 0, 0, 0}}},{22839, {{34153, 1, 0, 0},{34153, 1, 1, 18}, {0, 0, 0, 0}}},{22839, {{34153, 1, 0, 0},{34153, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22840, {{34154, 1, 0, 0},{34154, 1, 1, 17}, {0, 0, 0, 0}}},{22840, {{34154, 1, 0, 0},{34154, 1, 1, 17}, {0, 0, 0, 0}}},{22841, {{34154, 1, 0, 0},{34154, 1, 1, 18}, {0, 0, 0, 0}}},{22841, {{34154, 1, 0, 0},{34154, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{22842, {{34155, 1, 0, 0},{34155, 1, 1, 17}, {0, 0, 0, 0}}},{22842, {{34155, 1, 0, 0},{34155, 1, 1, 17}, {0, 0, 0, 0}}},{22843, {{34155, 1, 0, 0},{34155, 1, 1, 18}, {0, 0, 0, 0}}},{22843, {{34155, 1, 0, 0},{34155, 1, 1, 18}, {0, 0, 0, 0}}},{22844, {{34155, 1, 0, 0},{34155, 1, 1, 19}, {0, 0, 0, 0}}},{22844, {{34155, 1, 0, 0},{34155, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{22845, {{34156, 1, 0, 0},{34156, 1, 1, 16}, {0, 0, 0, 0}}},{22845, {{34156, 1, 0, 0},{34156, 1, 1, 16}, {0, 0, 0, 0}}},{22846, {{34156, 1, 0, 0},{34156, 1, 1, 17}, {0, 0, 0, 0}}},{22846, {{34156, 1, 0, 0},{34156, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{22847, {{34157, 1, 0, 0},{34157, 1, 1, 17}, {0, 0, 0, 0}}},{22847, {{34157, 1, 0, 0},{34157, 1, 1, 17}, {0, 0, 0, 0}}},{22848, {{34157, 1, 0, 0},{34157, 1, 1, 18}, {0, 0, 0, 0}}},{22848, {{34157, 1, 0, 0},{34157, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22849, {{34158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22850, {{34159, 1, 0, 0},{34159, 1, 1, 17}, {0, 0, 0, 0}}},{22850, {{34159, 1, 0, 0},{34159, 1, 1, 17}, {0, 0, 0, 0}}},{22851, {{34159, 1, 0, 0},{34159, 1, 1, 18}, {0, 0, 0, 0}}},{22851, {{34159, 1, 0, 0},{34159, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22854, {{34160, 1, 0, 0},{34160, 1, 1, 16}, {0, 0, 0, 0}}},{22854, {{34160, 1, 0, 0},{34160, 1, 1, 16}, {0, 0, 0, 0}}},{22855, {{34160, 1, 0, 0},{34160, 1, 1, 17}, {0, 0, 0, 0}}},{22855, {{34160, 1, 0, 0},{34160, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{22856, {{34163, 1, 0, 0},{34163, 1, 1, 16}, {0, 0, 0, 0}}},{22856, {{34163, 1, 0, 0},{34163, 1, 1, 16}, {0, 0, 0, 0}}},{22857, {{34163, 1, 0, 0},{34163, 1, 1, 17}, {0, 0, 0, 0}}},{22857, {{34163, 1, 0, 0},{34163, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{22858, {{34166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22859, {{34167, 1, 0, 0},{34167, 1, 1, 17}, {0, 0, 0, 0}}},{22859, {{34167, 1, 0, 0},{34167, 1, 1, 17}, {0, 0, 0, 0}}},{22860, {{34167, 1, 0, 0},{34167, 1, 1, 18}, {0, 0, 0, 0}}},{22860, {{34167, 1, 0, 0},{34167, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{22861, {{34168, 1, 0, 0},{34168, 1, 1, 17}, {0, 0, 0, 0}}},{22861, {{34168, 1, 0, 0},{34168, 1, 1, 17}, {0, 0, 0, 0}}},{22862, {{34168, 1, 0, 0},{34168, 1, 1, 16}, {0, 0, 0, 0}}},{22862, {{34168, 1, 0, 0},{34168, 1, 1, 16}, {0, 0, 0, 0}}},{22863, {{34168, 1, 0, 0},{34168, 1, 1, 18}, {0, 0, 0, 0}}},{22863, {{34168, 1, 0, 0},{34168, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22864, {{34169, 1, 0, 0},{34169, 1, 1, 17}, {0, 0, 0, 0}}},{22864, {{34169, 1, 0, 0},{34169, 1, 1, 17}, {0, 0, 0, 0}}},{22865, {{34169, 1, 0, 0},{34169, 1, 1, 18}, {0, 0, 0, 0}}},{22865, {{34169, 1, 0, 0},{34169, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22866, {{34170, 1, 0, 0},{34170, 1, 1, 17}, {0, 0, 0, 0}}},{22866, {{34170, 1, 0, 0},{34170, 1, 1, 17}, {0, 0, 0, 0}}},{22867, {{34170, 1, 0, 0},{34170, 1, 1, 18}, {0, 0, 0, 0}}},{22867, {{34170, 1, 0, 0},{34170, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22868, {{34171, 1, 0, 0},{34171, 1, 1, 17}, {0, 0, 0, 0}}},{22868, {{34171, 1, 0, 0},{34171, 1, 1, 17}, {0, 0, 0, 0}}},{22869, {{34171, 1, 0, 0},{34171, 1, 1, 18}, {0, 0, 0, 0}}},{22869, {{34171, 1, 0, 0},{34171, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22870, {{34172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22871, {{34173, 1, 0, 0},{34173, 1, 1, 17}, {0, 0, 0, 0}}},{22871, {{34173, 1, 0, 0},{34173, 1, 1, 17}, {0, 0, 0, 0}}},{22872, {{34173, 1, 0, 0},{34173, 1, 1, 18}, {0, 0, 0, 0}}},{22872, {{34173, 1, 0, 0},{34173, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22873, {{34174, 1, 0, 0},{34174, 1, 1, 17}, {0, 0, 0, 0}}},{22873, {{34174, 1, 0, 0},{34174, 1, 1, 17}, {0, 0, 0, 0}}},{22874, {{34174, 1, 0, 0},{34174, 1, 1, 18}, {0, 0, 0, 0}}},{22874, {{34174, 1, 0, 0},{34174, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22875, {{34175, 1, 0, 0},{34175, 1, 1, 17}, {0, 0, 0, 0}}},{22875, {{34175, 1, 0, 0},{34175, 1, 1, 17}, {0, 0, 0, 0}}},{22876, {{34175, 1, 0, 0},{34175, 1, 1, 18}, {0, 0, 0, 0}}},{22876, {{34175, 1, 0, 0},{34175, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22877, {{34176, 1, 0, 0},{34176, 1, 1, 17}, {0, 0, 0, 0}}},{22877, {{34176, 1, 0, 0},{34176, 1, 1, 17}, {0, 0, 0, 0}}},{22878, {{34176, 1, 0, 0},{34176, 1, 1, 18}, {0, 0, 0, 0}}},{22878, {{34176, 1, 0, 0},{34176, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{22879, {{34177, 1, 0, 0},{34177, 1, 1, 18}, {0, 0, 0, 0}}},{22879, {{34177, 1, 0, 0},{34177, 1, 1, 18}, {0, 0, 0, 0}}},{22880, {{34177, 1, 0, 0},{34177, 1, 1, 17}, {0, 0, 0, 0}}},{22880, {{34177, 1, 0, 0},{34177, 1, 1, 17}, {0, 0, 0, 0}}},{22881, {{34177, 1, 0, 0},{34177, 1, 1, 21}, {0, 0, 0, 0}}},{22881, {{34177, 1, 0, 0},{34177, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{22884, {{34178, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22885, {{34179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{22886, {{34180, 1, 0, 0},{34180, 1, 1, 18}, {0, 0, 0, 0}}},{22886, {{34180, 1, 0, 0},{34180, 1, 1, 18}, {0, 0, 0, 0}}},{22887, {{34180, 1, 0, 0},{34180, 1, 1, 20}, {0, 0, 0, 0}}},{22887, {{34180, 1, 0, 0},{34180, 1, 1, 20}, {0, 0, 0, 0}}},{22888, {{34180, 1, 0, 0},{34180, 1, 1, 19}, {0, 0, 0, 0}}},{22888, {{34180, 1, 0, 0},{34180, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{22890, {{34181, 1, 0, 0},{34181, 1, 1, 18}, {0, 0, 0, 0}}},{22890, {{34181, 1, 0, 0},{34181, 1, 1, 18}, {0, 0, 0, 0}}},{22891, {{34181, 1, 0, 0},{34181, 1, 1, 17}, {0, 0, 0, 0}}},{22891, {{34181, 1, 0, 0},{34181, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{22893, {{34182, 1, 0, 0},{34182, 1, 1, 17}, {0, 0, 0, 0}}},{22893, {{34182, 1, 0, 0},{34182, 1, 1, 17}, {0, 0, 0, 0}}},{22894, {{34182, 1, 0, 0},{34182, 1, 1, 18}, {0, 0, 0, 0}}},{22894, {{34182, 1, 0, 0},{34182, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {10, {{22895, {{34183, 1, 0, 0},{34183, 1, 1, 22}, {0, 0, 0, 0}}},{22895, {{34183, 1, 0, 0},{34183, 1, 1, 22}, {0, 0, 0, 0}}},{22896, {{34183, 1, 0, 0},{34183, 1, 1, 18}, {0, 0, 0, 0}}},{22896, {{34183, 1, 0, 0},{34183, 1, 1, 18}, {0, 0, 0, 0}}},{22897, {{34183, 1, 0, 0},{34183, 1, 1, 19}, {0, 0, 0, 0}}},{22897, {{34183, 1, 0, 0},{34183, 1, 1, 19}, {0, 0, 0, 0}}},{22898, {{34183, 1, 0, 0},{34183, 1, 1, 21}, {0, 0, 0, 0}}},{22898, {{34183, 1, 0, 0},{34183, 1, 1, 21}, {0, 0, 0, 0}}},{22899, {{34183, 1, 0, 0},{34183, 1, 1, 20}, {0, 0, 0, 0}}},{22899, {{34183, 1, 0, 0},{34183, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{22900, {{34184, 1, 0, 0},{34184, 1, 1, 17}, {0, 0, 0, 0}}},{22900, {{34184, 1, 0, 0},{34184, 1, 1, 17}, {0, 0, 0, 0}}},{22901, {{34184, 1, 0, 0},{34184, 1, 1, 18}, {0, 0, 0, 0}}},{22901, {{34184, 1, 0, 0},{34184, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22902, {{34185, 1, 0, 0},{34185, 1, 1, 17}, {0, 0, 0, 0}}},{22902, {{34185, 1, 0, 0},{34185, 1, 1, 17}, {0, 0, 0, 0}}},{22903, {{34185, 1, 0, 0},{34185, 1, 1, 18}, {0, 0, 0, 0}}},{22903, {{34185, 1, 0, 0},{34185, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{22904, {{34186, 1, 0, 0},{34186, 1, 1, 18}, {0, 0, 0, 0}}},{22904, {{34186, 1, 0, 0},{34186, 1, 1, 18}, {0, 0, 0, 0}}},{22905, {{34186, 1, 0, 0},{34186, 1, 1, 17}, {0, 0, 0, 0}}},{22905, {{34186, 1, 0, 0},{34186, 1, 1, 17}, {0, 0, 0, 0}}},{22906, {{34186, 1, 0, 0},{34186, 1, 1, 19}, {0, 0, 0, 0}}},{22906, {{34186, 1, 0, 0},{34186, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{22907, {{34187, 1, 0, 0},{34187, 1, 1, 17}, {0, 0, 0, 0}}},{22907, {{34187, 1, 0, 0},{34187, 1, 1, 17}, {0, 0, 0, 0}}},{22908, {{34187, 1, 0, 0},{34187, 1, 1, 18}, {0, 0, 0, 0}}},{22908, {{34187, 1, 0, 0},{34187, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {8, {{22909, {{34188, 1, 0, 0},{34188, 1, 1, 17}, {0, 0, 0, 0}}},{22909, {{34188, 1, 0, 0},{34188, 1, 1, 17}, {0, 0, 0, 0}}},{22910, {{34188, 1, 0, 0},{34188, 1, 1, 20}, {0, 0, 0, 0}}},{22910, {{34188, 1, 0, 0},{34188, 1, 1, 20}, {0, 0, 0, 0}}},{22911, {{34188, 1, 0, 0},{34188, 1, 1, 18}, {0, 0, 0, 0}}},{22911, {{34188, 1, 0, 0},{34188, 1, 1, 18}, {0, 0, 0, 0}}},{22912, {{34188, 1, 0, 0},{34188, 1, 1, 19}, {0, 0, 0, 0}}},{22912, {{34188, 1, 0, 0},{34188, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{22914, {{34189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22915, {{34190, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{22916, {{34191, 1, 0, 0},{34191, 1, 1, 17}, {0, 0, 0, 0}}},{22916, {{34191, 1, 0, 0},{34191, 1, 1, 17}, {0, 0, 0, 0}}},{22917, {{34191, 1, 0, 0},{34191, 1, 1, 19}, {0, 0, 0, 0}}},{22917, {{34191, 1, 0, 0},{34191, 1, 1, 19}, {0, 0, 0, 0}}},{22918, {{34191, 1, 0, 0},{34191, 1, 1, 18}, {0, 0, 0, 0}}},{22918, {{34191, 1, 0, 0},{34191, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22919, {{34192, 1, 0, 0},{34192, 1, 1, 17}, {0, 0, 0, 0}}},{22919, {{34192, 1, 0, 0},{34192, 1, 1, 17}, {0, 0, 0, 0}}},{22920, {{34192, 1, 0, 0},{34192, 1, 1, 18}, {0, 0, 0, 0}}},{22920, {{34192, 1, 0, 0},{34192, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22921, {{34193, 1, 0, 0},{34193, 1, 1, 17}, {0, 0, 0, 0}}},{22921, {{34193, 1, 0, 0},{34193, 1, 1, 17}, {0, 0, 0, 0}}},{22922, {{34193, 1, 0, 0},{34193, 1, 1, 18}, {0, 0, 0, 0}}},{22922, {{34193, 1, 0, 0},{34193, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22923, {{34194, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22924, {{34195, 1, 0, 0},{34195, 1, 1, 17}, {0, 0, 0, 0}}},{22924, {{34195, 1, 0, 0},{34195, 1, 1, 17}, {0, 0, 0, 0}}},{22925, {{34195, 1, 0, 0},{34195, 1, 1, 18}, {0, 0, 0, 0}}},{22925, {{34195, 1, 0, 0},{34195, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22926, {{34196, 1, 0, 0},{34196, 1, 1, 17}, {0, 0, 0, 0}}},{22926, {{34196, 1, 0, 0},{34196, 1, 1, 17}, {0, 0, 0, 0}}},{22927, {{34196, 1, 0, 0},{34196, 1, 1, 18}, {0, 0, 0, 0}}},{22927, {{34196, 1, 0, 0},{34196, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22928, {{34197, 1, 0, 0},{34197, 1, 1, 16}, {0, 0, 0, 0}}},{22928, {{34197, 1, 0, 0},{34197, 1, 1, 16}, {0, 0, 0, 0}}},{22929, {{34197, 1, 0, 0},{34197, 1, 1, 17}, {0, 0, 0, 0}}},{22929, {{34197, 1, 0, 0},{34197, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{22930, {{34198, 1, 0, 0},{34198, 1, 1, 17}, {0, 0, 0, 0}}},{22930, {{34198, 1, 0, 0},{34198, 1, 1, 17}, {0, 0, 0, 0}}},{22931, {{34198, 1, 0, 0},{34198, 1, 1, 16}, {0, 0, 0, 0}}},{22931, {{34198, 1, 0, 0},{34198, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {4, {{22932, {{34199, 1, 0, 0},{34199, 1, 1, 17}, {0, 0, 0, 0}}},{22932, {{34199, 1, 0, 0},{34199, 1, 1, 17}, {0, 0, 0, 0}}},{22933, {{34199, 1, 0, 0},{34199, 1, 1, 18}, {0, 0, 0, 0}}},{22933, {{34199, 1, 0, 0},{34199, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{22934, {{34200, 1, 0, 0},{34200, 1, 1, 19}, {0, 0, 0, 0}}},{22934, {{34200, 1, 0, 0},{34200, 1, 1, 19}, {0, 0, 0, 0}}},{22935, {{34200, 1, 0, 0},{34200, 1, 1, 17}, {0, 0, 0, 0}}},{22935, {{34200, 1, 0, 0},{34200, 1, 1, 17}, {0, 0, 0, 0}}},{22936, {{34200, 1, 0, 0},{34200, 1, 1, 18}, {0, 0, 0, 0}}},{22936, {{34200, 1, 0, 0},{34200, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22937, {{34201, 1, 0, 0},{34201, 1, 1, 17}, {0, 0, 0, 0}}},{22937, {{34201, 1, 0, 0},{34201, 1, 1, 17}, {0, 0, 0, 0}}},{22938, {{34201, 1, 0, 0},{34201, 1, 1, 18}, {0, 0, 0, 0}}},{22938, {{34201, 1, 0, 0},{34201, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22939, {{34202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22940, {{34203, 1, 0, 0},{34203, 1, 1, 17}, {0, 0, 0, 0}}},{22940, {{34203, 1, 0, 0},{34203, 1, 1, 17}, {0, 0, 0, 0}}},{22941, {{34203, 1, 0, 0},{34203, 1, 1, 18}, {0, 0, 0, 0}}},{22941, {{34203, 1, 0, 0},{34203, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22942, {{34204, 1, 0, 0},{34204, 1, 1, 17}, {0, 0, 0, 0}}},{22942, {{34204, 1, 0, 0},{34204, 1, 1, 17}, {0, 0, 0, 0}}},{22943, {{34204, 1, 0, 0},{34204, 1, 1, 18}, {0, 0, 0, 0}}},{22943, {{34204, 1, 0, 0},{34204, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22944, {{34205, 1, 0, 0},{34205, 1, 1, 17}, {0, 0, 0, 0}}},{22944, {{34205, 1, 0, 0},{34205, 1, 1, 17}, {0, 0, 0, 0}}},{22945, {{34205, 1, 0, 0},{34205, 1, 1, 18}, {0, 0, 0, 0}}},{22945, {{34205, 1, 0, 0},{34205, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22946, {{34206, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22947, {{34207, 1, 0, 0},{34207, 1, 1, 17}, {0, 0, 0, 0}}},{22947, {{34207, 1, 0, 0},{34207, 1, 1, 17}, {0, 0, 0, 0}}},{22948, {{34207, 1, 0, 0},{34207, 1, 1, 18}, {0, 0, 0, 0}}},{22948, {{34207, 1, 0, 0},{34207, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22949, {{34208, 1, 0, 0},{34208, 1, 1, 17}, {0, 0, 0, 0}}},{22949, {{34208, 1, 0, 0},{34208, 1, 1, 17}, {0, 0, 0, 0}}},{22950, {{34208, 1, 0, 0},{34208, 1, 1, 18}, {0, 0, 0, 0}}},{22950, {{34208, 1, 0, 0},{34208, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22951, {{34209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22952, {{34210, 1, 0, 0},{34210, 1, 1, 17}, {0, 0, 0, 0}}},{22952, {{34210, 1, 0, 0},{34210, 1, 1, 17}, {0, 0, 0, 0}}},{22953, {{34210, 1, 0, 0},{34210, 1, 1, 18}, {0, 0, 0, 0}}},{22953, {{34210, 1, 0, 0},{34210, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22954, {{34211, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22955, {{34212, 1, 0, 0},{34212, 1, 1, 17}, {0, 0, 0, 0}}},{22955, {{34212, 1, 0, 0},{34212, 1, 1, 17}, {0, 0, 0, 0}}},{22956, {{34212, 1, 0, 0},{34212, 1, 1, 18}, {0, 0, 0, 0}}},{22956, {{34212, 1, 0, 0},{34212, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22957, {{34213, 1, 0, 0},{34213, 1, 1, 17}, {0, 0, 0, 0}}},{22957, {{34213, 1, 0, 0},{34213, 1, 1, 17}, {0, 0, 0, 0}}},{22958, {{34213, 1, 0, 0},{34213, 1, 1, 18}, {0, 0, 0, 0}}},{22958, {{34213, 1, 0, 0},{34213, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22959, {{34214, 1, 0, 0},{34214, 1, 1, 17}, {0, 0, 0, 0}}},{22959, {{34214, 1, 0, 0},{34214, 1, 1, 17}, {0, 0, 0, 0}}},{22960, {{34214, 1, 0, 0},{34214, 1, 1, 18}, {0, 0, 0, 0}}},{22960, {{34214, 1, 0, 0},{34214, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22961, {{34215, 1, 0, 0},{34215, 1, 1, 17}, {0, 0, 0, 0}}},{22961, {{34215, 1, 0, 0},{34215, 1, 1, 17}, {0, 0, 0, 0}}},{22962, {{34215, 1, 0, 0},{34215, 1, 1, 18}, {0, 0, 0, 0}}},{22962, {{34215, 1, 0, 0},{34215, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22963, {{34216, 1, 0, 0},{34216, 1, 1, 17}, {0, 0, 0, 0}}},{22963, {{34216, 1, 0, 0},{34216, 1, 1, 17}, {0, 0, 0, 0}}},{22964, {{34216, 1, 0, 0},{34216, 1, 1, 18}, {0, 0, 0, 0}}},{22964, {{34216, 1, 0, 0},{34216, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {8, {{22965, {{34217, 1, 0, 0},{34217, 1, 1, 18}, {0, 0, 0, 0}}},{22965, {{34217, 1, 0, 0},{34217, 1, 1, 18}, {0, 0, 0, 0}}},{22966, {{34217, 1, 0, 0},{34217, 1, 1, 19}, {0, 0, 0, 0}}},{22966, {{34217, 1, 0, 0},{34217, 1, 1, 19}, {0, 0, 0, 0}}},{22967, {{34217, 1, 0, 0},{34217, 1, 1, 20}, {0, 0, 0, 0}}},{22967, {{34217, 1, 0, 0},{34217, 1, 1, 20}, {0, 0, 0, 0}}},{22968, {{34217, 1, 0, 0},{34217, 1, 1, 21}, {0, 0, 0, 0}}},{22968, {{34217, 1, 0, 0},{34217, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {4, {{22969, {{34218, 1, 0, 0},{34218, 1, 1, 17}, {0, 0, 0, 0}}},{22969, {{34218, 1, 0, 0},{34218, 1, 1, 17}, {0, 0, 0, 0}}},{22970, {{34218, 1, 0, 0},{34218, 1, 1, 18}, {0, 0, 0, 0}}},{22970, {{34218, 1, 0, 0},{34218, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{22971, {{34219, 1, 0, 0},{34219, 1, 1, 17}, {0, 0, 0, 0}}},{22971, {{34219, 1, 0, 0},{34219, 1, 1, 17}, {0, 0, 0, 0}}},{22972, {{34219, 1, 0, 0},{34219, 1, 1, 19}, {0, 0, 0, 0}}},{22972, {{34219, 1, 0, 0},{34219, 1, 1, 19}, {0, 0, 0, 0}}},{22973, {{34219, 1, 0, 0},{34219, 1, 1, 18}, {0, 0, 0, 0}}},{22973, {{34219, 1, 0, 0},{34219, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22975, {{34220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22977, {{34221, 1, 0, 0},{34221, 1, 1, 17}, {0, 0, 0, 0}}},{22977, {{34221, 1, 0, 0},{34221, 1, 1, 17}, {0, 0, 0, 0}}},{22978, {{34221, 1, 0, 0},{34221, 1, 1, 18}, {0, 0, 0, 0}}},{22978, {{34221, 1, 0, 0},{34221, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22981, {{34222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{22983, {{34223, 1, 0, 0},{34223, 1, 1, 18}, {0, 0, 0, 0}}},{22983, {{34223, 1, 0, 0},{34223, 1, 1, 18}, {0, 0, 0, 0}}},{22984, {{34223, 1, 0, 0},{34223, 1, 1, 19}, {0, 0, 0, 0}}},{22984, {{34223, 1, 0, 0},{34223, 1, 1, 19}, {0, 0, 0, 0}}},{22985, {{34223, 1, 0, 0},{34223, 1, 1, 20}, {0, 0, 0, 0}}},{22985, {{34223, 1, 0, 0},{34223, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{22986, {{34224, 1, 0, 0},{34224, 1, 1, 17}, {0, 0, 0, 0}}},{22986, {{34224, 1, 0, 0},{34224, 1, 1, 17}, {0, 0, 0, 0}}},{22987, {{34224, 1, 0, 0},{34224, 1, 1, 18}, {0, 0, 0, 0}}},{22987, {{34224, 1, 0, 0},{34224, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{22988, {{34225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22991, {{34226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22992, {{34227, 1, 0, 0},{34227, 1, 1, 17}, {0, 0, 0, 0}}},{22992, {{34227, 1, 0, 0},{34227, 1, 1, 17}, {0, 0, 0, 0}}},{22993, {{34227, 1, 0, 0},{34227, 1, 1, 16}, {0, 0, 0, 0}}},{22993, {{34227, 1, 0, 0},{34227, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {4, {{22994, {{34228, 1, 0, 0},{34228, 1, 1, 17}, {0, 0, 0, 0}}},{22994, {{34228, 1, 0, 0},{34228, 1, 1, 17}, {0, 0, 0, 0}}},{22995, {{34228, 1, 0, 0},{34228, 1, 1, 18}, {0, 0, 0, 0}}},{22995, {{34228, 1, 0, 0},{34228, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{22996, {{34229, 1, 0, 0},{34229, 1, 1, 16}, {0, 0, 0, 0}}},{22996, {{34229, 1, 0, 0},{34229, 1, 1, 16}, {0, 0, 0, 0}}},{22997, {{34229, 1, 0, 0},{34229, 1, 1, 17}, {0, 0, 0, 0}}},{22997, {{34229, 1, 0, 0},{34229, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{22998, {{34230, 1, 0, 0},{34230, 1, 1, 17}, {0, 0, 0, 0}}},{22998, {{34230, 1, 0, 0},{34230, 1, 1, 17}, {0, 0, 0, 0}}},{22999, {{34230, 1, 0, 0},{34230, 1, 1, 18}, {0, 0, 0, 0}}},{22999, {{34230, 1, 0, 0},{34230, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{23000, {{34231, 1, 0, 0},{34231, 1, 1, 17}, {0, 0, 0, 0}}},{23000, {{34231, 1, 0, 0},{34231, 1, 1, 17}, {0, 0, 0, 0}}},{23001, {{34231, 1, 0, 0},{34231, 1, 1, 18}, {0, 0, 0, 0}}},{23001, {{34231, 1, 0, 0},{34231, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{23002, {{34232, 1, 0, 0},{34232, 1, 1, 17}, {0, 0, 0, 0}}},{23002, {{34232, 1, 0, 0},{34232, 1, 1, 17}, {0, 0, 0, 0}}},{23003, {{34232, 1, 0, 0},{34232, 1, 1, 18}, {0, 0, 0, 0}}},{23003, {{34232, 1, 0, 0},{34232, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{23004, {{34233, 1, 0, 0},{34233, 1, 1, 17}, {0, 0, 0, 0}}},{23004, {{34233, 1, 0, 0},{34233, 1, 1, 17}, {0, 0, 0, 0}}},{23005, {{34233, 1, 0, 0},{34233, 1, 1, 18}, {0, 0, 0, 0}}},{23005, {{34233, 1, 0, 0},{34233, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{23006, {{34234, 1, 0, 0},{34234, 1, 1, 17}, {0, 0, 0, 0}}},{23006, {{34234, 1, 0, 0},{34234, 1, 1, 17}, {0, 0, 0, 0}}},{23007, {{34234, 1, 0, 0},{34234, 1, 1, 18}, {0, 0, 0, 0}}},{23007, {{34234, 1, 0, 0},{34234, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23008, {{34235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23009, {{34236, 1, 0, 0},{34236, 1, 1, 17}, {0, 0, 0, 0}}},{23009, {{34236, 1, 0, 0},{34236, 1, 1, 17}, {0, 0, 0, 0}}},{23010, {{34236, 1, 0, 0},{34236, 1, 1, 18}, {0, 0, 0, 0}}},{23010, {{34236, 1, 0, 0},{34236, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{23011, {{34237, 1, 0, 0},{34237, 1, 1, 17}, {0, 0, 0, 0}}},{23011, {{34237, 1, 0, 0},{34237, 1, 1, 17}, {0, 0, 0, 0}}},{23012, {{34237, 1, 0, 0},{34237, 1, 1, 18}, {0, 0, 0, 0}}},{23012, {{34237, 1, 0, 0},{34237, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{23013, {{34238, 1, 0, 0},{34238, 1, 1, 17}, {0, 0, 0, 0}}},{23013, {{34238, 1, 0, 0},{34238, 1, 1, 17}, {0, 0, 0, 0}}},{23014, {{34238, 1, 0, 0},{34238, 1, 1, 18}, {0, 0, 0, 0}}},{23014, {{34238, 1, 0, 0},{34238, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{23015, {{34239, 1, 0, 0},{34239, 1, 1, 17}, {0, 0, 0, 0}}},{23015, {{34239, 1, 0, 0},{34239, 1, 1, 17}, {0, 0, 0, 0}}},{23016, {{34239, 1, 0, 0},{34239, 1, 1, 18}, {0, 0, 0, 0}}},{23016, {{34239, 1, 0, 0},{34239, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23017, {{34240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23018, {{34241, 1, 0, 0},{34241, 1, 1, 17}, {0, 0, 0, 0}}},{23018, {{34241, 1, 0, 0},{34241, 1, 1, 17}, {0, 0, 0, 0}}},{23019, {{34241, 1, 0, 0},{34241, 1, 1, 18}, {0, 0, 0, 0}}},{23019, {{34241, 1, 0, 0},{34241, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{23022, {{34242, 1, 0, 0},{34242, 1, 1, 17}, {0, 0, 0, 0}}},{23022, {{34242, 1, 0, 0},{34242, 1, 1, 17}, {0, 0, 0, 0}}},{23023, {{34242, 1, 0, 0},{34242, 1, 1, 18}, {0, 0, 0, 0}}},{23023, {{34242, 1, 0, 0},{34242, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23024, {{34243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23025, {{34244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23026, {{34245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23027, {{34246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23028, {{34247, 1, 0, 0},{34247, 1, 1, 17}, {0, 0, 0, 0}}},{23028, {{34247, 1, 0, 0},{34247, 1, 1, 17}, {0, 0, 0, 0}}},{23029, {{34247, 1, 0, 0},{34247, 1, 1, 18}, {0, 0, 0, 0}}},{23029, {{34247, 1, 0, 0},{34247, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23030, {{34248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{23031, {{34249, 1, 0, 0},{34249, 1, 1, 17}, {0, 0, 0, 0}}},{23031, {{34249, 1, 0, 0},{34249, 1, 1, 17}, {0, 0, 0, 0}}},{23032, {{34249, 1, 0, 0},{34249, 1, 1, 18}, {0, 0, 0, 0}}},{23032, {{34249, 1, 0, 0},{34249, 1, 1, 18}, {0, 0, 0, 0}}},{23033, {{34249, 1, 0, 0},{34249, 1, 1, 19}, {0, 0, 0, 0}}},{23033, {{34249, 1, 0, 0},{34249, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{23034, {{34250, 1, 0, 0},{34250, 1, 1, 17}, {0, 0, 0, 0}}},{23034, {{34250, 1, 0, 0},{34250, 1, 1, 17}, {0, 0, 0, 0}}},{23035, {{34250, 1, 0, 0},{34250, 1, 1, 18}, {0, 0, 0, 0}}},{23035, {{34250, 1, 0, 0},{34250, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{23036, {{34251, 1, 0, 0},{34251, 1, 1, 17}, {0, 0, 0, 0}}},{23036, {{34251, 1, 0, 0},{34251, 1, 1, 17}, {0, 0, 0, 0}}},{23037, {{34251, 1, 0, 0},{34251, 1, 1, 18}, {0, 0, 0, 0}}},{23037, {{34251, 1, 0, 0},{34251, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23038, {{34252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{23039, {{34253, 1, 0, 0},{34253, 1, 1, 19}, {0, 0, 0, 0}}},{23039, {{34253, 1, 0, 0},{34253, 1, 1, 19}, {0, 0, 0, 0}}},{23040, {{34253, 1, 0, 0},{34253, 1, 1, 17}, {0, 0, 0, 0}}},{23040, {{34253, 1, 0, 0},{34253, 1, 1, 17}, {0, 0, 0, 0}}},{23041, {{34253, 1, 0, 0},{34253, 1, 1, 18}, {0, 0, 0, 0}}},{23041, {{34253, 1, 0, 0},{34253, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{23042, {{34254, 1, 0, 0},{34254, 1, 1, 17}, {0, 0, 0, 0}}},{23042, {{34254, 1, 0, 0},{34254, 1, 1, 17}, {0, 0, 0, 0}}},{23043, {{34254, 1, 0, 0},{34254, 1, 1, 18}, {0, 0, 0, 0}}},{23043, {{34254, 1, 0, 0},{34254, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {8, {{23044, {{34255, 1, 0, 0},{34255, 1, 1, 18}, {0, 0, 0, 0}}},{23044, {{34255, 1, 0, 0},{34255, 1, 1, 18}, {0, 0, 0, 0}}},{23045, {{34255, 1, 0, 0},{34255, 1, 1, 21}, {0, 0, 0, 0}}},{23045, {{34255, 1, 0, 0},{34255, 1, 1, 21}, {0, 0, 0, 0}}},{23046, {{34255, 1, 0, 0},{34255, 1, 1, 19}, {0, 0, 0, 0}}},{23046, {{34255, 1, 0, 0},{34255, 1, 1, 19}, {0, 0, 0, 0}}},{23047, {{34255, 1, 0, 0},{34255, 1, 1, 20}, {0, 0, 0, 0}}},{23047, {{34255, 1, 0, 0},{34255, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{23049, {{34256, 1, 0, 0},{34256, 1, 1, 17}, {0, 0, 0, 0}}},{23049, {{34256, 1, 0, 0},{34256, 1, 1, 17}, {0, 0, 0, 0}}},{23050, {{34256, 1, 0, 0},{34256, 1, 1, 18}, {0, 0, 0, 0}}},{23050, {{34256, 1, 0, 0},{34256, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23051, {{34257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23052, {{34258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{23053, {{34261, 1, 0, 0},{34261, 1, 1, 18}, {0, 0, 0, 0}}},{23053, {{34261, 1, 0, 0},{34261, 1, 1, 18}, {0, 0, 0, 0}}},{23054, {{34261, 1, 0, 0},{34261, 1, 1, 19}, {0, 0, 0, 0}}},{23054, {{34261, 1, 0, 0},{34261, 1, 1, 19}, {0, 0, 0, 0}}},{23055, {{34261, 1, 0, 0},{34261, 1, 1, 20}, {0, 0, 0, 0}}},{23055, {{34261, 1, 0, 0},{34261, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{23056, {{34262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23057, {{34263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23058, {{34264, 1, 0, 0},{34264, 1, 1, 17}, {0, 0, 0, 0}}},{23058, {{34264, 1, 0, 0},{34264, 1, 1, 17}, {0, 0, 0, 0}}},{23059, {{34264, 1, 0, 0},{34264, 1, 1, 18}, {0, 0, 0, 0}}},{23059, {{34264, 1, 0, 0},{34264, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{23060, {{34265, 1, 0, 0},{34265, 1, 1, 17}, {0, 0, 0, 0}}},{23060, {{34265, 1, 0, 0},{34265, 1, 1, 17}, {0, 0, 0, 0}}},{23061, {{34265, 1, 0, 0},{34265, 1, 1, 18}, {0, 0, 0, 0}}},{23061, {{34265, 1, 0, 0},{34265, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{23062, {{34266, 1, 0, 0},{34266, 1, 1, 19}, {0, 0, 0, 0}}},{23062, {{34266, 1, 0, 0},{34266, 1, 1, 19}, {0, 0, 0, 0}}},{23063, {{34266, 1, 0, 0},{34266, 1, 1, 17}, {0, 0, 0, 0}}},{23063, {{34266, 1, 0, 0},{34266, 1, 1, 17}, {0, 0, 0, 0}}},{23064, {{34266, 1, 0, 0},{34266, 1, 1, 18}, {0, 0, 0, 0}}},{23064, {{34266, 1, 0, 0},{34266, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23065, {{34267, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23066, {{34268, 1, 0, 0},{34268, 1, 1, 17}, {0, 0, 0, 0}}},{23066, {{34268, 1, 0, 0},{34268, 1, 1, 17}, {0, 0, 0, 0}}},{23067, {{34268, 1, 0, 0},{34268, 1, 1, 18}, {0, 0, 0, 0}}},{23067, {{34268, 1, 0, 0},{34268, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{23068, {{34269, 1, 0, 0},{34269, 1, 1, 17}, {0, 0, 0, 0}}},{23068, {{34269, 1, 0, 0},{34269, 1, 1, 17}, {0, 0, 0, 0}}},{23069, {{34269, 1, 0, 0},{34269, 1, 1, 18}, {0, 0, 0, 0}}},{23069, {{34269, 1, 0, 0},{34269, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{23070, {{34271, 1, 0, 0},{34271, 1, 1, 17}, {0, 0, 0, 0}}},{23070, {{34271, 1, 0, 0},{34271, 1, 1, 17}, {0, 0, 0, 0}}},{23071, {{34271, 1, 0, 0},{34271, 1, 1, 18}, {0, 0, 0, 0}}},{23071, {{34271, 1, 0, 0},{34271, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{23072, {{34272, 1, 0, 0},{34272, 1, 1, 17}, {0, 0, 0, 0}}},{23072, {{34272, 1, 0, 0},{34272, 1, 1, 17}, {0, 0, 0, 0}}},{23073, {{34272, 1, 0, 0},{34272, 1, 1, 18}, {0, 0, 0, 0}}},{23073, {{34272, 1, 0, 0},{34272, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{23074, {{34273, 1, 0, 0},{34273, 1, 1, 19}, {0, 0, 0, 0}}},{23074, {{34273, 1, 0, 0},{34273, 1, 1, 19}, {0, 0, 0, 0}}},{23075, {{34273, 1, 0, 0},{34273, 1, 1, 17}, {0, 0, 0, 0}}},{23075, {{34273, 1, 0, 0},{34273, 1, 1, 17}, {0, 0, 0, 0}}},{23076, {{34273, 1, 0, 0},{34273, 1, 1, 18}, {0, 0, 0, 0}}},{23076, {{34273, 1, 0, 0},{34273, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23077, {{34274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23078, {{34275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {10, {{23079, {{34276, 1, 0, 0},{34276, 1, 1, 18}, {0, 0, 0, 0}}},{23079, {{34276, 1, 0, 0},{34276, 1, 1, 18}, {0, 0, 0, 0}}},{23080, {{34276, 1, 0, 0},{34276, 1, 1, 19}, {0, 0, 0, 0}}},{23080, {{34276, 1, 0, 0},{34276, 1, 1, 19}, {0, 0, 0, 0}}},{23081, {{34276, 1, 0, 0},{34276, 1, 1, 20}, {0, 0, 0, 0}}},{23081, {{34276, 1, 0, 0},{34276, 1, 1, 20}, {0, 0, 0, 0}}},{23082, {{34276, 1, 0, 0},{34276, 1, 1, 21}, {0, 0, 0, 0}}},{23082, {{34276, 1, 0, 0},{34276, 1, 1, 21}, {0, 0, 0, 0}}},{60144, {{34276, 1, 0, 0},{34276, 1, 1, 22}, {0, 0, 0, 0}}},{60144, {{34276, 1, 0, 0},{34276, 1, 1, 22}, {0, 0, 0, 0}}}}},
    {6, {{23086, {{34277, 1, 0, 0},{34277, 1, 1, 17}, {0, 0, 0, 0}}},{23086, {{34277, 1, 0, 0},{34277, 1, 1, 17}, {0, 0, 0, 0}}},{23087, {{34277, 1, 0, 0},{34277, 1, 1, 18}, {0, 0, 0, 0}}},{23087, {{34277, 1, 0, 0},{34277, 1, 1, 18}, {0, 0, 0, 0}}},{23088, {{34277, 1, 0, 0},{34277, 1, 1, 19}, {0, 0, 0, 0}}},{23088, {{34277, 1, 0, 0},{34277, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{23089, {{34278, 1, 0, 0},{34278, 1, 1, 17}, {0, 0, 0, 0}}},{23089, {{34278, 1, 0, 0},{34278, 1, 1, 17}, {0, 0, 0, 0}}},{23090, {{34278, 1, 0, 0},{34278, 1, 1, 18}, {0, 0, 0, 0}}},{23090, {{34278, 1, 0, 0},{34278, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23092, {{34279, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23094, {{34280, 1, 0, 0},{34280, 1, 1, 17}, {0, 0, 0, 0}}},{23094, {{34280, 1, 0, 0},{34280, 1, 1, 17}, {0, 0, 0, 0}}},{23095, {{34280, 1, 0, 0},{34280, 1, 1, 18}, {0, 0, 0, 0}}},{23095, {{34280, 1, 0, 0},{34280, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{23096, {{34281, 1, 0, 0},{34281, 1, 1, 17}, {0, 0, 0, 0}}},{23096, {{34281, 1, 0, 0},{34281, 1, 1, 17}, {0, 0, 0, 0}}},{23097, {{34281, 1, 0, 0},{34281, 1, 1, 18}, {0, 0, 0, 0}}},{23097, {{34281, 1, 0, 0},{34281, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{23098, {{34282, 1, 0, 0},{34282, 1, 1, 17}, {0, 0, 0, 0}}},{23098, {{34282, 1, 0, 0},{34282, 1, 1, 17}, {0, 0, 0, 0}}},{23099, {{34282, 1, 0, 0},{34282, 1, 1, 18}, {0, 0, 0, 0}}},{23099, {{34282, 1, 0, 0},{34282, 1, 1, 18}, {0, 0, 0, 0}}},{23100, {{34282, 1, 0, 0},{34282, 1, 1, 19}, {0, 0, 0, 0}}},{23100, {{34282, 1, 0, 0},{34282, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{23102, {{34283, 1, 0, 0},{34283, 1, 1, 16}, {0, 0, 0, 0}}},{23102, {{34283, 1, 0, 0},{34283, 1, 1, 16}, {0, 0, 0, 0}}},{58562, {{34283, 1, 0, 0},{34283, 1, 1, 17}, {0, 0, 0, 0}}},{58562, {{34283, 1, 0, 0},{34283, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{23103, {{34284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23104, {{34285, 1, 0, 0},{34285, 1, 1, 17}, {0, 0, 0, 0}}},{23104, {{34285, 1, 0, 0},{34285, 1, 1, 17}, {0, 0, 0, 0}}},{23105, {{34285, 1, 0, 0},{34285, 1, 1, 18}, {0, 0, 0, 0}}},{23105, {{34285, 1, 0, 0},{34285, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23106, {{34286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23107, {{34287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23108, {{34288, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23110, {{34289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{23111, {{34291, 1, 0, 0},{34291, 1, 1, 17}, {0, 0, 0, 0}}},{23111, {{34291, 1, 0, 0},{34291, 1, 1, 17}, {0, 0, 0, 0}}},{23112, {{34291, 1, 0, 0},{34291, 1, 1, 18}, {0, 0, 0, 0}}},{23112, {{34291, 1, 0, 0},{34291, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{23113, {{34292, 1, 0, 0},{34292, 1, 1, 17}, {0, 0, 0, 0}}},{23113, {{34292, 1, 0, 0},{34292, 1, 1, 17}, {0, 0, 0, 0}}},{58559, {{34292, 1, 0, 0},{34292, 1, 1, 18}, {0, 0, 0, 0}}},{58559, {{34292, 1, 0, 0},{34292, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{23114, {{34294, 1, 0, 0},{34294, 1, 1, 17}, {0, 0, 0, 0}}},{23114, {{34294, 1, 0, 0},{34294, 1, 1, 17}, {0, 0, 0, 0}}},{23115, {{34294, 1, 0, 0},{34294, 1, 1, 18}, {0, 0, 0, 0}}},{23115, {{34294, 1, 0, 0},{34294, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{23116, {{34295, 1, 0, 0},{34295, 1, 1, 18}, {0, 0, 0, 0}}},{23116, {{34295, 1, 0, 0},{34295, 1, 1, 18}, {0, 0, 0, 0}}},{23117, {{34295, 1, 0, 0},{34295, 1, 1, 20}, {0, 0, 0, 0}}},{23117, {{34295, 1, 0, 0},{34295, 1, 1, 20}, {0, 0, 0, 0}}},{23118, {{34295, 1, 0, 0},{34295, 1, 1, 19}, {0, 0, 0, 0}}},{23118, {{34295, 1, 0, 0},{34295, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{23119, {{34296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{23120, {{34297, 1, 0, 0},{34297, 1, 1, 17}, {0, 0, 0, 0}}},{23120, {{34297, 1, 0, 0},{34297, 1, 1, 17}, {0, 0, 0, 0}}},{23121, {{34297, 1, 0, 0},{34297, 1, 1, 18}, {0, 0, 0, 0}}},{23121, {{34297, 1, 0, 0},{34297, 1, 1, 18}, {0, 0, 0, 0}}},{23122, {{34297, 1, 0, 0},{34297, 1, 1, 19}, {0, 0, 0, 0}}},{23122, {{34297, 1, 0, 0},{34297, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{23123, {{34298, 1, 0, 0},{34298, 1, 1, 17}, {0, 0, 0, 0}}},{23123, {{34298, 1, 0, 0},{34298, 1, 1, 17}, {0, 0, 0, 0}}},{23124, {{34298, 1, 0, 0},{34298, 1, 1, 18}, {0, 0, 0, 0}}},{23124, {{34298, 1, 0, 0},{34298, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{23125, {{34299, 1, 0, 0},{34299, 1, 1, 17}, {0, 0, 0, 0}}},{23125, {{34299, 1, 0, 0},{34299, 1, 1, 17}, {0, 0, 0, 0}}},{23126, {{34299, 1, 0, 0},{34299, 1, 1, 18}, {0, 0, 0, 0}}},{23126, {{34299, 1, 0, 0},{34299, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{23127, {{34300, 1, 0, 0},{34300, 1, 1, 17}, {0, 0, 0, 0}}},{23127, {{34300, 1, 0, 0},{34300, 1, 1, 17}, {0, 0, 0, 0}}},{23128, {{34300, 1, 0, 0},{34300, 1, 1, 18}, {0, 0, 0, 0}}},{23128, {{34300, 1, 0, 0},{34300, 1, 1, 18}, {0, 0, 0, 0}}},{23129, {{34300, 1, 0, 0},{34300, 1, 1, 19}, {0, 0, 0, 0}}},{23129, {{34300, 1, 0, 0},{34300, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{23131, {{34301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{23132, {{34302, 1, 0, 0},{34302, 1, 1, 17}, {0, 0, 0, 0}}},{23132, {{34302, 1, 0, 0},{34302, 1, 1, 17}, {0, 0, 0, 0}}},{23133, {{34302, 1, 0, 0},{34302, 1, 1, 18}, {0, 0, 0, 0}}},{23133, {{34302, 1, 0, 0},{34302, 1, 1, 18}, {0, 0, 0, 0}}},{23134, {{34302, 1, 0, 0},{34302, 1, 1, 19}, {0, 0, 0, 0}}},{23134, {{34302, 1, 0, 0},{34302, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{23135, {{34303, 1, 0, 0},{34303, 1, 1, 17}, {0, 0, 0, 0}}},{23135, {{34303, 1, 0, 0},{34303, 1, 1, 17}, {0, 0, 0, 0}}},{23136, {{34303, 1, 0, 0},{34303, 1, 1, 18}, {0, 0, 0, 0}}},{23136, {{34303, 1, 0, 0},{34303, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{23137, {{34304, 1, 0, 0},{34304, 1, 1, 17}, {0, 0, 0, 0}}},{23137, {{34304, 1, 0, 0},{34304, 1, 1, 17}, {0, 0, 0, 0}}},{23138, {{34304, 1, 0, 0},{34304, 1, 1, 18}, {0, 0, 0, 0}}},{23138, {{34304, 1, 0, 0},{34304, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23139, {{34305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23140, {{34306, 1, 0, 0},{34306, 1, 1, 17}, {0, 0, 0, 0}}},{23140, {{34306, 1, 0, 0},{34306, 1, 1, 17}, {0, 0, 0, 0}}},{23141, {{34306, 1, 0, 0},{34306, 1, 1, 18}, {0, 0, 0, 0}}},{23141, {{34306, 1, 0, 0},{34306, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23142, {{34307, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23143, {{34308, 1, 0, 0},{34308, 1, 1, 17}, {0, 0, 0, 0}}},{23143, {{34308, 1, 0, 0},{34308, 1, 1, 17}, {0, 0, 0, 0}}},{23144, {{34308, 1, 0, 0},{34308, 1, 1, 18}, {0, 0, 0, 0}}},{23144, {{34308, 1, 0, 0},{34308, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{23145, {{34309, 1, 0, 0},{34309, 1, 1, 17}, {0, 0, 0, 0}}},{23145, {{34309, 1, 0, 0},{34309, 1, 1, 17}, {0, 0, 0, 0}}},{23146, {{34309, 1, 0, 0},{34309, 1, 1, 18}, {0, 0, 0, 0}}},{23146, {{34309, 1, 0, 0},{34309, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{23147, {{34310, 1, 0, 0},{34310, 1, 1, 17}, {0, 0, 0, 0}}},{23147, {{34310, 1, 0, 0},{34310, 1, 1, 17}, {0, 0, 0, 0}}},{23148, {{34310, 1, 0, 0},{34310, 1, 1, 18}, {0, 0, 0, 0}}},{23148, {{34310, 1, 0, 0},{34310, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {8, {{23149, {{34311, 1, 0, 0},{34311, 1, 1, 19}, {0, 0, 0, 0}}},{23149, {{34311, 1, 0, 0},{34311, 1, 1, 19}, {0, 0, 0, 0}}},{23150, {{34311, 1, 0, 0},{34311, 1, 1, 17}, {0, 0, 0, 0}}},{23150, {{34311, 1, 0, 0},{34311, 1, 1, 17}, {0, 0, 0, 0}}},{23151, {{34311, 1, 0, 0},{34311, 1, 1, 18}, {0, 0, 0, 0}}},{23151, {{34311, 1, 0, 0},{34311, 1, 1, 18}, {0, 0, 0, 0}}},{23152, {{34311, 1, 0, 0},{34311, 1, 1, 20}, {0, 0, 0, 0}}},{23152, {{34311, 1, 0, 0},{34311, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{23153, {{34312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23154, {{34313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23155, {{34314, 1, 0, 0},{34314, 1, 1, 17}, {0, 0, 0, 0}}},{23155, {{34314, 1, 0, 0},{34314, 1, 1, 17}, {0, 0, 0, 0}}},{23156, {{34314, 1, 0, 0},{34314, 1, 1, 18}, {0, 0, 0, 0}}},{23156, {{34314, 1, 0, 0},{34314, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{23157, {{34315, 1, 0, 0},{34315, 1, 1, 17}, {0, 0, 0, 0}}},{23157, {{34315, 1, 0, 0},{34315, 1, 1, 17}, {0, 0, 0, 0}}},{23158, {{34315, 1, 0, 0},{34315, 1, 1, 18}, {0, 0, 0, 0}}},{23158, {{34315, 1, 0, 0},{34315, 1, 1, 18}, {0, 0, 0, 0}}},{23159, {{34315, 1, 0, 0},{34315, 1, 1, 19}, {0, 0, 0, 0}}},{23159, {{34315, 1, 0, 0},{34315, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{23160, {{34316, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23161, {{34317, 1, 0, 0},{34317, 1, 1, 18}, {0, 0, 0, 0}}},{23161, {{34317, 1, 0, 0},{34317, 1, 1, 18}, {0, 0, 0, 0}}},{23162, {{34317, 1, 0, 0},{34317, 1, 1, 17}, {0, 0, 0, 0}}},{23162, {{34317, 1, 0, 0},{34317, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{23165, {{34318, 1, 0, 0},{34318, 1, 1, 17}, {0, 0, 0, 0}}},{23165, {{34318, 1, 0, 0},{34318, 1, 1, 17}, {0, 0, 0, 0}}},{23166, {{34318, 1, 0, 0},{34318, 1, 1, 18}, {0, 0, 0, 0}}},{23166, {{34318, 1, 0, 0},{34318, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23167, {{34319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23168, {{34320, 1, 0, 0},{34320, 1, 1, 17}, {0, 0, 0, 0}}},{23168, {{34320, 1, 0, 0},{34320, 1, 1, 17}, {0, 0, 0, 0}}},{23169, {{34320, 1, 0, 0},{34320, 1, 1, 18}, {0, 0, 0, 0}}},{23169, {{34320, 1, 0, 0},{34320, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{23170, {{34321, 1, 0, 0},{34321, 1, 1, 17}, {0, 0, 0, 0}}},{23170, {{34321, 1, 0, 0},{34321, 1, 1, 17}, {0, 0, 0, 0}}},{23171, {{34321, 1, 0, 0},{34321, 1, 1, 18}, {0, 0, 0, 0}}},{23171, {{34321, 1, 0, 0},{34321, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {17, {{23172, {{34322, 1, 0, 0},{34322, 1, 1, 19}, {0, 0, 0, 0}}},{23172, {{34322, 1, 0, 0},{34322, 1, 1, 19}, {0, 0, 0, 0}}},{23173, {{34322, 1, 0, 0},{34322, 1, 1, 20}, {0, 0, 0, 0}}},{23173, {{34322, 1, 0, 0},{34322, 1, 1, 20}, {0, 0, 0, 0}}},{30220, {{34322, 1, 0, 0},{34322, 1, 1, 25}, {0, 0, 0, 0}}},{30220, {{34322, 1, 0, 0},{34322, 1, 1, 25}, {0, 0, 0, 0}}},{30221, {{34322, 1, 0, 0},{34322, 1, 1, 23}, {0, 0, 0, 0}}},{30221, {{34322, 1, 0, 0},{34322, 1, 1, 23}, {0, 0, 0, 0}}},{30222, {{34322, 1, 0, 0},{34322, 1, 1, 24}, {0, 0, 0, 0}}},{30222, {{34322, 1, 0, 0},{34322, 1, 1, 24}, {0, 0, 0, 0}}},{30223, {{34322, 1, 0, 0},{34322, 1, 1, 26}, {0, 0, 0, 0}}},{30223, {{34322, 1, 0, 0},{34322, 1, 1, 26}, {0, 0, 0, 0}}},{30224, {{34322, 1, 0, 0},{34322, 1, 1, 0},{34322, 1, 1, 21},{64032, 1, 0, 0}, {0, 0, 0, 0}}},{30224, {{34322, 1, 0, 0},{34322, 1, 1, 0},{34322, 1, 1, 21},{64032, 1, 0, 0}, {0, 0, 0, 0}}},{30224, {{34322, 1, 0, 0},{34322, 1, 1, 0},{34322, 1, 1, 21},{64032, 1, 0, 0}, {0, 0, 0, 0}}},{60150, {{34322, 1, 0, 0},{34322, 1, 1, 22}, {0, 0, 0, 0}}},{60150, {{34322, 1, 0, 0},{34322, 1, 1, 22}, {0, 0, 0, 0}}}}},
    {4, {{23174, {{34323, 1, 0, 0},{34323, 1, 1, 17}, {0, 0, 0, 0}}},{23174, {{34323, 1, 0, 0},{34323, 1, 1, 17}, {0, 0, 0, 0}}},{23175, {{34323, 1, 0, 0},{34323, 1, 1, 18}, {0, 0, 0, 0}}},{23175, {{34323, 1, 0, 0},{34323, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{23177, {{34325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{23178, {{34326, 1, 0, 0},{34326, 1, 1, 17}, {0, 0, 0, 0}}},{23178, {{34326, 1, 0, 0},{34326, 1, 1, 17}, {0, 0, 0, 0}}},{23179, {{34326, 1, 0, 0},{34326, 1, 1, 18}, {0, 0, 0, 0}}},{23179, {{34326, 1, 0, 0},{34326, 1, 1, 18}, {0, 0, 0, 0}}},{60149, {{34326, 1, 0, 0},{34326, 1, 1, 19}, {0, 0, 0, 0}}},{60149, {{34326, 1, 0, 0},{34326, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{23180, {{34327, 1, 0, 0},{34327, 1, 1, 17}, {0, 0, 0, 0}}},{23180, {{34327, 1, 0, 0},{34327, 1, 1, 17}, {0, 0, 0, 0}}},{23181, {{34327, 1, 0, 0},{34327, 1, 1, 18}, {0, 0, 0, 0}}},{23181, {{34327, 1, 0, 0},{34327, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{23182, {{34328, 1, 0, 0},{34328, 1, 1, 18}, {0, 0, 0, 0}}},{23182, {{34328, 1, 0, 0},{34328, 1, 1, 18}, {0, 0, 0, 0}}},{23183, {{34328, 1, 0, 0},{34328, 1, 1, 17}, {0, 0, 0, 0}}},{23183, {{34328, 1, 0, 0},{34328, 1, 1, 17}, {0, 0, 0, 0}}},{23184, {{34328, 1, 0, 0},{34328, 1, 1, 19}, {0, 0, 0, 0}}},{23184, {{34328, 1, 0, 0},{34328, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {6, {{23185, {{34329, 1, 0, 0},{34329, 1, 1, 17}, {0, 0, 0, 0}}},{23185, {{34329, 1, 0, 0},{34329, 1, 1, 17}, {0, 0, 0, 0}}},{23186, {{34329, 1, 0, 0},{34329, 1, 1, 18}, {0, 0, 0, 0}}},{23186, {{34329, 1, 0, 0},{34329, 1, 1, 18}, {0, 0, 0, 0}}},{23187, {{34329, 1, 0, 0},{34329, 1, 1, 19}, {0, 0, 0, 0}}},{23187, {{34329, 1, 0, 0},{34329, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {6, {{23189, {{34330, 1, 0, 0},{34330, 1, 1, 17}, {0, 0, 0, 0}}},{23189, {{34330, 1, 0, 0},{34330, 1, 1, 17}, {0, 0, 0, 0}}},{23190, {{34330, 1, 0, 0},{34330, 1, 1, 18}, {0, 0, 0, 0}}},{23190, {{34330, 1, 0, 0},{34330, 1, 1, 18}, {0, 0, 0, 0}}},{58561, {{34330, 1, 0, 0},{34330, 1, 1, 19}, {0, 0, 0, 0}}},{58561, {{34330, 1, 0, 0},{34330, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{23191, {{34331, 1, 0, 0},{34331, 1, 1, 17}, {0, 0, 0, 0}}},{23191, {{34331, 1, 0, 0},{34331, 1, 1, 17}, {0, 0, 0, 0}}},{23192, {{34331, 1, 0, 0},{34331, 1, 1, 18}, {0, 0, 0, 0}}},{23192, {{34331, 1, 0, 0},{34331, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23193, {{34332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{23194, {{34334, 1, 0, 0},{34334, 1, 1, 17}, {0, 0, 0, 0}}},{23194, {{34334, 1, 0, 0},{34334, 1, 1, 17}, {0, 0, 0, 0}}},{23195, {{34334, 1, 0, 0},{34334, 1, 1, 18}, {0, 0, 0, 0}}},{23195, {{34334, 1, 0, 0},{34334, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23196, {{34335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23197, {{34336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23198, {{34337, 1, 0, 0},{34337, 1, 1, 17}, {0, 0, 0, 0}}},{23198, {{34337, 1, 0, 0},{34337, 1, 1, 17}, {0, 0, 0, 0}}},{23199, {{34337, 1, 0, 0},{34337, 1, 1, 18}, {0, 0, 0, 0}}},{23199, {{34337, 1, 0, 0},{34337, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{23200, {{34338, 1, 0, 0},{34338, 1, 1, 17}, {0, 0, 0, 0}}},{23200, {{34338, 1, 0, 0},{34338, 1, 1, 17}, {0, 0, 0, 0}}},{23201, {{34338, 1, 0, 0},{34338, 1, 1, 19}, {0, 0, 0, 0}}},{23201, {{34338, 1, 0, 0},{34338, 1, 1, 19}, {0, 0, 0, 0}}},{23202, {{34338, 1, 0, 0},{34338, 1, 1, 18}, {0, 0, 0, 0}}},{23202, {{34338, 1, 0, 0},{34338, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23203, {{34339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{23204, {{34340, 1, 0, 0},{34340, 1, 1, 19}, {0, 0, 0, 0}}},{23204, {{34340, 1, 0, 0},{34340, 1, 1, 19}, {0, 0, 0, 0}}},{23205, {{34340, 1, 0, 0},{34340, 1, 1, 17}, {0, 0, 0, 0}}},{23205, {{34340, 1, 0, 0},{34340, 1, 1, 17}, {0, 0, 0, 0}}},{23206, {{34340, 1, 0, 0},{34340, 1, 1, 18}, {0, 0, 0, 0}}},{23206, {{34340, 1, 0, 0},{34340, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23207, {{34341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23208, {{34342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{23209, {{34343, 1, 0, 0},{34343, 1, 1, 19}, {0, 0, 0, 0}}},{23209, {{34343, 1, 0, 0},{34343, 1, 1, 19}, {0, 0, 0, 0}}},{23210, {{34343, 1, 0, 0},{34343, 1, 1, 17}, {0, 0, 0, 0}}},{23210, {{34343, 1, 0, 0},{34343, 1, 1, 17}, {0, 0, 0, 0}}},{23211, {{34343, 1, 0, 0},{34343, 1, 1, 18}, {0, 0, 0, 0}}},{23211, {{34343, 1, 0, 0},{34343, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23212, {{34344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{23214, {{34345, 1, 0, 0},{34345, 1, 1, 17}, {0, 0, 0, 0}}},{23214, {{34345, 1, 0, 0},{34345, 1, 1, 17}, {0, 0, 0, 0}}},{23215, {{34345, 1, 0, 0},{34345, 1, 1, 18}, {0, 0, 0, 0}}},{23215, {{34345, 1, 0, 0},{34345, 1, 1, 18}, {0, 0, 0, 0}}},{23216, {{34345, 1, 0, 0},{34345, 1, 1, 19}, {0, 0, 0, 0}}},{23216, {{34345, 1, 0, 0},{34345, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{23217, {{34346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23218, {{34347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23220, {{34348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{23221, {{34349, 1, 0, 0},{34349, 1, 1, 18}, {0, 0, 0, 0}}},{23221, {{34349, 1, 0, 0},{34349, 1, 1, 18}, {0, 0, 0, 0}}},{23222, {{34349, 1, 0, 0},{34349, 1, 1, 20}, {0, 0, 0, 0}}},{23222, {{34349, 1, 0, 0},{34349, 1, 1, 20}, {0, 0, 0, 0}}},{23223, {{34349, 1, 0, 0},{34349, 1, 1, 19}, {0, 0, 0, 0}}},{23223, {{34349, 1, 0, 0},{34349, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{23225, {{34350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23226, {{34351, 1, 0, 0},{34351, 1, 1, 17}, {0, 0, 0, 0}}},{23226, {{34351, 1, 0, 0},{34351, 1, 1, 17}, {0, 0, 0, 0}}},{23227, {{34351, 1, 0, 0},{34351, 1, 1, 18}, {0, 0, 0, 0}}},{23227, {{34351, 1, 0, 0},{34351, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{23228, {{34352, 1, 0, 0},{34352, 1, 1, 17}, {0, 0, 0, 0}}},{23228, {{34352, 1, 0, 0},{34352, 1, 1, 17}, {0, 0, 0, 0}}},{23229, {{34352, 1, 0, 0},{34352, 1, 1, 19}, {0, 0, 0, 0}}},{23229, {{34352, 1, 0, 0},{34352, 1, 1, 19}, {0, 0, 0, 0}}},{23230, {{34352, 1, 0, 0},{34352, 1, 1, 18}, {0, 0, 0, 0}}},{23230, {{34352, 1, 0, 0},{34352, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23231, {{34353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23232, {{34354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23233, {{34355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23234, {{34356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23235, {{34357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{23236, {{34358, 1, 0, 0},{34358, 1, 1, 19}, {0, 0, 0, 0}}},{23236, {{34358, 1, 0, 0},{34358, 1, 1, 19}, {0, 0, 0, 0}}},{23237, {{34358, 1, 0, 0},{34358, 1, 1, 17}, {0, 0, 0, 0}}},{23237, {{34358, 1, 0, 0},{34358, 1, 1, 17}, {0, 0, 0, 0}}},{23238, {{34358, 1, 0, 0},{34358, 1, 1, 18}, {0, 0, 0, 0}}},{23238, {{34358, 1, 0, 0},{34358, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{23239, {{34360, 1, 0, 0},{34360, 1, 1, 17}, {0, 0, 0, 0}}},{23239, {{34360, 1, 0, 0},{34360, 1, 1, 17}, {0, 0, 0, 0}}},{23240, {{34360, 1, 0, 0},{34360, 1, 1, 18}, {0, 0, 0, 0}}},{23240, {{34360, 1, 0, 0},{34360, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{23241, {{34361, 1, 0, 0},{34361, 1, 1, 17}, {0, 0, 0, 0}}},{23241, {{34361, 1, 0, 0},{34361, 1, 1, 17}, {0, 0, 0, 0}}},{23242, {{34361, 1, 0, 0},{34361, 1, 1, 18}, {0, 0, 0, 0}}},{23242, {{34361, 1, 0, 0},{34361, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{23243, {{34362, 1, 0, 0},{34362, 1, 1, 17}, {0, 0, 0, 0}}},{23243, {{34362, 1, 0, 0},{34362, 1, 1, 17}, {0, 0, 0, 0}}},{23244, {{34362, 1, 0, 0},{34362, 1, 1, 18}, {0, 0, 0, 0}}},{23244, {{34362, 1, 0, 0},{34362, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23245, {{34363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23246, {{34364, 1, 0, 0},{34364, 1, 1, 17}, {0, 0, 0, 0}}},{23246, {{34364, 1, 0, 0},{34364, 1, 1, 17}, {0, 0, 0, 0}}},{23247, {{34364, 1, 0, 0},{34364, 1, 1, 18}, {0, 0, 0, 0}}},{23247, {{34364, 1, 0, 0},{34364, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{23248, {{34365, 1, 0, 0},{34365, 1, 1, 17}, {0, 0, 0, 0}}},{23248, {{34365, 1, 0, 0},{34365, 1, 1, 17}, {0, 0, 0, 0}}},{23249, {{34365, 1, 0, 0},{34365, 1, 1, 18}, {0, 0, 0, 0}}},{23249, {{34365, 1, 0, 0},{34365, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23250, {{34366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23251, {{34367, 1, 0, 0},{34367, 1, 1, 17}, {0, 0, 0, 0}}},{23251, {{34367, 1, 0, 0},{34367, 1, 1, 17}, {0, 0, 0, 0}}},{23252, {{34367, 1, 0, 0},{34367, 1, 1, 18}, {0, 0, 0, 0}}},{23252, {{34367, 1, 0, 0},{34367, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{23253, {{34368, 1, 0, 0},{34368, 1, 1, 17}, {0, 0, 0, 0}}},{23253, {{34368, 1, 0, 0},{34368, 1, 1, 17}, {0, 0, 0, 0}}},{23254, {{34368, 1, 0, 0},{34368, 1, 1, 18}, {0, 0, 0, 0}}},{23254, {{34368, 1, 0, 0},{34368, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {8, {{23255, {{34369, 1, 0, 0},{34369, 1, 1, 19}, {0, 0, 0, 0}}},{23255, {{34369, 1, 0, 0},{34369, 1, 1, 19}, {0, 0, 0, 0}}},{23256, {{34369, 1, 0, 0},{34369, 1, 1, 20}, {0, 0, 0, 0}}},{23256, {{34369, 1, 0, 0},{34369, 1, 1, 20}, {0, 0, 0, 0}}},{23257, {{34369, 1, 0, 0},{34369, 1, 1, 18}, {0, 0, 0, 0}}},{23257, {{34369, 1, 0, 0},{34369, 1, 1, 18}, {0, 0, 0, 0}}},{23258, {{34369, 1, 0, 0},{34369, 1, 1, 17}, {0, 0, 0, 0}}},{23258, {{34369, 1, 0, 0},{34369, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{23259, {{34370, 1, 0, 0},{34370, 1, 1, 17}, {0, 0, 0, 0}}},{23259, {{34370, 1, 0, 0},{34370, 1, 1, 17}, {0, 0, 0, 0}}},{23260, {{34370, 1, 0, 0},{34370, 1, 1, 18}, {0, 0, 0, 0}}},{23260, {{34370, 1, 0, 0},{34370, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23261, {{34371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{23263, {{34374, 1, 0, 0},{34374, 1, 1, 17}, {0, 0, 0, 0}}},{23263, {{34374, 1, 0, 0},{34374, 1, 1, 17}, {0, 0, 0, 0}}},{23264, {{34374, 1, 0, 0},{34374, 1, 1, 18}, {0, 0, 0, 0}}},{23264, {{34374, 1, 0, 0},{34374, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23265, {{34375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23266, {{34376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23267, {{34377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23268, {{34378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23269, {{34379, 1, 0, 0},{34379, 1, 1, 16}, {0, 0, 0, 0}}},{23269, {{34379, 1, 0, 0},{34379, 1, 1, 16}, {0, 0, 0, 0}}},{58575, {{34379, 1, 0, 0},{34379, 1, 1, 17}, {0, 0, 0, 0}}},{58575, {{34379, 1, 0, 0},{34379, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{23270, {{34380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23271, {{34381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{23272, {{34382, 1, 0, 0},{34382, 1, 1, 18}, {0, 0, 0, 0}}},{23272, {{34382, 1, 0, 0},{34382, 1, 1, 18}, {0, 0, 0, 0}}},{23273, {{34382, 1, 0, 0},{34382, 1, 1, 19}, {0, 0, 0, 0}}},{23273, {{34382, 1, 0, 0},{34382, 1, 1, 19}, {0, 0, 0, 0}}},{58583, {{34382, 1, 0, 0},{34382, 1, 1, 24}, {0, 0, 0, 0}}},{58583, {{34382, 1, 0, 0},{34382, 1, 1, 24}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{23278, {{34384, 1, 0, 0},{34384, 1, 1, 19}, {0, 0, 0, 0}}},{23278, {{34384, 1, 0, 0},{34384, 1, 1, 19}, {0, 0, 0, 0}}},{23279, {{34384, 1, 0, 0},{34384, 1, 1, 18}, {0, 0, 0, 0}}},{23279, {{34384, 1, 0, 0},{34384, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23280, {{34385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23281, {{34386, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23283, {{34387, 1, 0, 0},{34387, 1, 1, 17}, {0, 0, 0, 0}}},{23283, {{34387, 1, 0, 0},{34387, 1, 1, 17}, {0, 0, 0, 0}}},{23284, {{34387, 1, 0, 0},{34387, 1, 1, 18}, {0, 0, 0, 0}}},{23284, {{34387, 1, 0, 0},{34387, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{23285, {{34388, 1, 0, 0},{34388, 1, 1, 18}, {0, 0, 0, 0}}},{23285, {{34388, 1, 0, 0},{34388, 1, 1, 18}, {0, 0, 0, 0}}},{23286, {{34388, 1, 0, 0},{34388, 1, 1, 19}, {0, 0, 0, 0}}},{23286, {{34388, 1, 0, 0},{34388, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{23287, {{34389, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23288, {{34390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23289, {{34391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23290, {{34392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23291, {{34393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23292, {{34394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23293, {{34395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{23294, {{34396, 1, 0, 0},{34396, 1, 1, 18}, {0, 0, 0, 0}}},{23294, {{34396, 1, 0, 0},{34396, 1, 1, 18}, {0, 0, 0, 0}}},{30186, {{34396, 1, 0, 0},{34396, 1, 1, 0},{34396, 1, 1, 19},{63798, 1, 0, 0}, {0, 0, 0, 0}}},{30186, {{34396, 1, 0, 0},{34396, 1, 1, 0},{34396, 1, 1, 19},{63798, 1, 0, 0}, {0, 0, 0, 0}}},{30186, {{34396, 1, 0, 0},{34396, 1, 1, 0},{34396, 1, 1, 19},{63798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23295, {{34397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23296, {{34398, 1, 0, 0},{34398, 1, 1, 19}, {0, 0, 0, 0}}},{23296, {{34398, 1, 0, 0},{34398, 1, 1, 19}, {0, 0, 0, 0}}},{23297, {{34398, 1, 0, 0},{34398, 1, 1, 18}, {0, 0, 0, 0}}},{23297, {{34398, 1, 0, 0},{34398, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23299, {{34399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23301, {{34400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23302, {{34401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23303, {{34402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23304, {{34403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23305, {{34404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23306, {{34405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23307, {{34406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23308, {{34407, 1, 0, 0},{34407, 1, 1, 17}, {0, 0, 0, 0}}},{23308, {{34407, 1, 0, 0},{34407, 1, 1, 17}, {0, 0, 0, 0}}},{23309, {{34407, 1, 0, 0},{34407, 1, 1, 18}, {0, 0, 0, 0}}},{23309, {{34407, 1, 0, 0},{34407, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23310, {{34408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23311, {{34409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23312, {{34410, 1, 0, 0},{34410, 1, 1, 16}, {0, 0, 0, 0}}},{23312, {{34410, 1, 0, 0},{34410, 1, 1, 16}, {0, 0, 0, 0}}},{23313, {{34410, 1, 0, 0},{34410, 1, 1, 17}, {0, 0, 0, 0}}},{23313, {{34410, 1, 0, 0},{34410, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{23314, {{34411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23315, {{34412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23316, {{34413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23317, {{34414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23318, {{34415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23319, {{34416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23320, {{34417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{23321, {{34419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23322, {{34420, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23323, {{34421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23324, {{34422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23325, {{34423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23326, {{34424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23327, {{34425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23328, {{34426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23329, {{34427, 1, 0, 0},{34427, 1, 1, 17}, {0, 0, 0, 0}}},{23329, {{34427, 1, 0, 0},{34427, 1, 1, 17}, {0, 0, 0, 0}}},{23330, {{34427, 1, 0, 0},{34427, 1, 1, 18}, {0, 0, 0, 0}}},{23330, {{34427, 1, 0, 0},{34427, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23331, {{34428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23332, {{34429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{23333, {{34436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23334, {{34437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23335, {{34438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23336, {{34439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23337, {{34440, 1, 0, 0},{34440, 1, 1, 18}, {0, 0, 0, 0}}},{23337, {{34440, 1, 0, 0},{34440, 1, 1, 18}, {0, 0, 0, 0}}},{23338, {{34440, 1, 0, 0},{34440, 1, 1, 17}, {0, 0, 0, 0}}},{23338, {{34440, 1, 0, 0},{34440, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{23339, {{34441, 1, 0, 0},{34441, 1, 1, 17}, {0, 0, 0, 0}}},{23339, {{34441, 1, 0, 0},{34441, 1, 1, 17}, {0, 0, 0, 0}}},{23340, {{34441, 1, 0, 0},{34441, 1, 1, 18}, {0, 0, 0, 0}}},{23340, {{34441, 1, 0, 0},{34441, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{23341, {{34442, 1, 0, 0},{34442, 1, 1, 18}, {0, 0, 0, 0}}},{23341, {{34442, 1, 0, 0},{34442, 1, 1, 18}, {0, 0, 0, 0}}},{23342, {{34442, 1, 0, 0},{34442, 1, 1, 19}, {0, 0, 0, 0}}},{23342, {{34442, 1, 0, 0},{34442, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{23343, {{34443, 1, 0, 0},{34443, 1, 1, 17}, {0, 0, 0, 0}}},{23343, {{34443, 1, 0, 0},{34443, 1, 1, 17}, {0, 0, 0, 0}}},{23344, {{34443, 1, 0, 0},{34443, 1, 1, 18}, {0, 0, 0, 0}}},{23344, {{34443, 1, 0, 0},{34443, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{23345, {{34444, 1, 0, 0},{34444, 1, 1, 17}, {0, 0, 0, 0}}},{23345, {{34444, 1, 0, 0},{34444, 1, 1, 17}, {0, 0, 0, 0}}},{23346, {{34444, 1, 0, 0},{34444, 1, 1, 18}, {0, 0, 0, 0}}},{23346, {{34444, 1, 0, 0},{34444, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23347, {{34445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23348, {{34446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23349, {{34447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23350, {{34448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23351, {{34449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23352, {{34450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23353, {{34451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23354, {{34452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23355, {{34453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23356, {{34454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23357, {{34455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23358, {{34456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23359, {{34457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23360, {{34458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23361, {{34459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23362, {{34460, 1, 0, 0},{34460, 1, 1, 18}, {0, 0, 0, 0}}},{23362, {{34460, 1, 0, 0},{34460, 1, 1, 18}, {0, 0, 0, 0}}},{23363, {{34460, 1, 0, 0},{34460, 1, 1, 17}, {0, 0, 0, 0}}},{23363, {{34460, 1, 0, 0},{34460, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{23364, {{34461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23365, {{34462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23366, {{34463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23367, {{34464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23368, {{34465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23369, {{34466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23370, {{34467, 1, 0, 0},{34467, 1, 1, 17}, {0, 0, 0, 0}}},{23370, {{34467, 1, 0, 0},{34467, 1, 1, 17}, {0, 0, 0, 0}}},{23371, {{34467, 1, 0, 0},{34467, 1, 1, 18}, {0, 0, 0, 0}}},{23371, {{34467, 1, 0, 0},{34467, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{23372, {{34468, 1, 0, 0},{34468, 1, 1, 17}, {0, 0, 0, 0}}},{23372, {{34468, 1, 0, 0},{34468, 1, 1, 17}, {0, 0, 0, 0}}},{58609, {{34468, 1, 0, 0},{34468, 1, 1, 18}, {0, 0, 0, 0}}},{58609, {{34468, 1, 0, 0},{34468, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23373, {{34469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23374, {{34470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23375, {{34471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23376, {{34472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{23377, {{34473, 1, 0, 0},{34473, 1, 1, 17}, {0, 0, 0, 0}}},{23377, {{34473, 1, 0, 0},{34473, 1, 1, 17}, {0, 0, 0, 0}}},{23378, {{34473, 1, 0, 0},{34473, 1, 1, 19}, {0, 0, 0, 0}}},{23378, {{34473, 1, 0, 0},{34473, 1, 1, 19}, {0, 0, 0, 0}}},{58606, {{34473, 1, 0, 0},{34473, 1, 1, 18}, {0, 0, 0, 0}}},{58606, {{34473, 1, 0, 0},{34473, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23379, {{34474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{23380, {{34475, 1, 0, 0},{34475, 1, 1, 17}, {0, 0, 0, 0}}},{23380, {{34475, 1, 0, 0},{34475, 1, 1, 17}, {0, 0, 0, 0}}},{23381, {{34475, 1, 0, 0},{34475, 1, 1, 18}, {0, 0, 0, 0}}},{23381, {{34475, 1, 0, 0},{34475, 1, 1, 18}, {0, 0, 0, 0}}},{58601, {{34475, 1, 0, 0},{34475, 1, 1, 19}, {0, 0, 0, 0}}},{58601, {{34475, 1, 0, 0},{34475, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{23382, {{34477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23383, {{34478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23384, {{34479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23385, {{34480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23386, {{34481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23387, {{34482, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23388, {{34483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23389, {{34484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23390, {{34485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23391, {{34486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23392, {{34487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23393, {{34488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23394, {{34489, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23395, {{34490, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23396, {{34491, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23397, {{34492, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23398, {{34493, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23399, {{34494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23400, {{34495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23401, {{34496, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23402, {{34497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23403, {{34498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23404, {{34499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23405, {{34500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23406, {{34501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23407, {{34502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23408, {{34503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23409, {{34504, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23410, {{34505, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23411, {{34506, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23412, {{34507, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23413, {{34508, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23414, {{34509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23415, {{34510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{23416, {{34512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23417, {{34513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23418, {{34514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23419, {{34515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23420, {{34516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23421, {{34517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23422, {{34518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23423, {{34519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23424, {{34520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23425, {{34521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23426, {{34522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23427, {{34523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23428, {{34524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23429, {{34525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23430, {{34526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23431, {{34527, 1, 0, 0},{34527, 1, 1, 17}, {0, 0, 0, 0}}},{23431, {{34527, 1, 0, 0},{34527, 1, 1, 17}, {0, 0, 0, 0}}},{23432, {{34527, 1, 0, 0},{34527, 1, 1, 18}, {0, 0, 0, 0}}},{23432, {{34527, 1, 0, 0},{34527, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23433, {{34528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23434, {{34529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23435, {{34530, 1, 0, 0},{34530, 1, 1, 16}, {0, 0, 0, 0}}},{23435, {{34530, 1, 0, 0},{34530, 1, 1, 16}, {0, 0, 0, 0}}},{23436, {{34530, 1, 0, 0},{34530, 1, 1, 17}, {0, 0, 0, 0}}},{23436, {{34530, 1, 0, 0},{34530, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{23437, {{34531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23438, {{34532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23439, {{34533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23440, {{34534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23441, {{34535, 1, 0, 0},{34535, 1, 1, 17}, {0, 0, 0, 0}}},{23441, {{34535, 1, 0, 0},{34535, 1, 1, 17}, {0, 0, 0, 0}}},{23442, {{34535, 1, 0, 0},{34535, 1, 1, 18}, {0, 0, 0, 0}}},{23442, {{34535, 1, 0, 0},{34535, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23443, {{34536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{23444, {{34537, 1, 0, 0},{34537, 1, 1, 17}, {0, 0, 0, 0}}},{23444, {{34537, 1, 0, 0},{34537, 1, 1, 17}, {0, 0, 0, 0}}},{23445, {{34537, 1, 0, 0},{34537, 1, 1, 18}, {0, 0, 0, 0}}},{23445, {{34537, 1, 0, 0},{34537, 1, 1, 18}, {0, 0, 0, 0}}},{23446, {{34537, 1, 0, 0},{34537, 1, 1, 19}, {0, 0, 0, 0}}},{23446, {{34537, 1, 0, 0},{34537, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{23447, {{34538, 1, 0, 0},{34538, 1, 1, 16}, {0, 0, 0, 0}}},{23447, {{34538, 1, 0, 0},{34538, 1, 1, 16}, {0, 0, 0, 0}}},{23448, {{34538, 1, 0, 0},{34538, 1, 1, 17}, {0, 0, 0, 0}}},{23448, {{34538, 1, 0, 0},{34538, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{23449, {{34539, 1, 0, 0},{34539, 1, 1, 16}, {0, 0, 0, 0}}},{23449, {{34539, 1, 0, 0},{34539, 1, 1, 16}, {0, 0, 0, 0}}},{23450, {{34539, 1, 0, 0},{34539, 1, 1, 17}, {0, 0, 0, 0}}},{23450, {{34539, 1, 0, 0},{34539, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{23451, {{34540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23452, {{34541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23453, {{34542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23454, {{34543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{23455, {{34549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23456, {{34550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23457, {{34551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23458, {{34552, 1, 0, 0},{34552, 1, 1, 19}, {0, 0, 0, 0}}},{23458, {{34552, 1, 0, 0},{34552, 1, 1, 19}, {0, 0, 0, 0}}},{23459, {{34552, 1, 0, 0},{34552, 1, 1, 18}, {0, 0, 0, 0}}},{23459, {{34552, 1, 0, 0},{34552, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23460, {{34553, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23461, {{34554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23462, {{34555, 1, 0, 0},{34555, 1, 1, 17}, {0, 0, 0, 0}}},{23462, {{34555, 1, 0, 0},{34555, 1, 1, 17}, {0, 0, 0, 0}}},{23463, {{34555, 1, 0, 0},{34555, 1, 1, 18}, {0, 0, 0, 0}}},{23463, {{34555, 1, 0, 0},{34555, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23464, {{34556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23465, {{34557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23466, {{34558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23467, {{34559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23468, {{34560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23469, {{34561, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23470, {{34562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23471, {{34563, 1, 0, 0},{34563, 1, 1, 18}, {0, 0, 0, 0}}},{23471, {{34563, 1, 0, 0},{34563, 1, 1, 18}, {0, 0, 0, 0}}},{23472, {{34563, 1, 0, 0},{34563, 1, 1, 19}, {0, 0, 0, 0}}},{23472, {{34563, 1, 0, 0},{34563, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{23473, {{34564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23474, {{34565, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23475, {{34566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23476, {{34567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23477, {{34568, 1, 0, 0},{34568, 1, 1, 17}, {0, 0, 0, 0}}},{23477, {{34568, 1, 0, 0},{34568, 1, 1, 17}, {0, 0, 0, 0}}},{23478, {{34568, 1, 0, 0},{34568, 1, 1, 18}, {0, 0, 0, 0}}},{23478, {{34568, 1, 0, 0},{34568, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{23479, {{34569, 1, 0, 0},{34569, 1, 1, 17}, {0, 0, 0, 0}}},{23479, {{34569, 1, 0, 0},{34569, 1, 1, 17}, {0, 0, 0, 0}}},{23480, {{34569, 1, 0, 0},{34569, 1, 1, 18}, {0, 0, 0, 0}}},{23480, {{34569, 1, 0, 0},{34569, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23481, {{34570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23482, {{34571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23483, {{34572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23484, {{34573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23485, {{34574, 1, 0, 0},{34574, 1, 1, 17}, {0, 0, 0, 0}}},{23485, {{34574, 1, 0, 0},{34574, 1, 1, 17}, {0, 0, 0, 0}}},{58615, {{34574, 1, 0, 0},{34574, 1, 1, 18}, {0, 0, 0, 0}}},{58615, {{34574, 1, 0, 0},{34574, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23486, {{34575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23487, {{34576, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23488, {{34577, 1, 0, 0},{34577, 1, 1, 17}, {0, 0, 0, 0}}},{23488, {{34577, 1, 0, 0},{34577, 1, 1, 17}, {0, 0, 0, 0}}},{23489, {{34577, 1, 0, 0},{34577, 1, 1, 18}, {0, 0, 0, 0}}},{23489, {{34577, 1, 0, 0},{34577, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{23490, {{34578, 1, 0, 0},{34578, 1, 1, 17}, {0, 0, 0, 0}}},{23490, {{34578, 1, 0, 0},{34578, 1, 1, 17}, {0, 0, 0, 0}}},{23491, {{34578, 1, 0, 0},{34578, 1, 1, 18}, {0, 0, 0, 0}}},{23491, {{34578, 1, 0, 0},{34578, 1, 1, 18}, {0, 0, 0, 0}}},{23492, {{34578, 1, 0, 0},{34578, 1, 1, 19}, {0, 0, 0, 0}}},{23492, {{34578, 1, 0, 0},{34578, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{23493, {{34579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23494, {{34580, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23495, {{34581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{23496, {{34584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23497, {{34585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23498, {{34586, 1, 0, 0},{34586, 1, 1, 18}, {0, 0, 0, 0}}},{23498, {{34586, 1, 0, 0},{34586, 1, 1, 18}, {0, 0, 0, 0}}},{23499, {{34586, 1, 0, 0},{34586, 1, 1, 19}, {0, 0, 0, 0}}},{23499, {{34586, 1, 0, 0},{34586, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{23500, {{34587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23501, {{34588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23502, {{34589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23503, {{34590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23504, {{34591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23505, {{34592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23506, {{34593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23507, {{34594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23508, {{34595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23509, {{34596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23510, {{34597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23511, {{34598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23512, {{34599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23513, {{34600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23514, {{34601, 1, 0, 0},{34601, 1, 1, 17}, {0, 0, 0, 0}}},{23514, {{34601, 1, 0, 0},{34601, 1, 1, 17}, {0, 0, 0, 0}}},{23515, {{34601, 1, 0, 0},{34601, 1, 1, 18}, {0, 0, 0, 0}}},{23515, {{34601, 1, 0, 0},{34601, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23516, {{34602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23517, {{34603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23518, {{34604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23519, {{34605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23520, {{34606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23521, {{34607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23522, {{34608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23523, {{34609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23524, {{34610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23525, {{34611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23526, {{34612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23527, {{34613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23528, {{34614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23529, {{34615, 1, 0, 0},{34615, 1, 1, 18}, {0, 0, 0, 0}}},{23529, {{34615, 1, 0, 0},{34615, 1, 1, 18}, {0, 0, 0, 0}}},{23530, {{34615, 1, 0, 0},{34615, 1, 1, 19}, {0, 0, 0, 0}}},{23530, {{34615, 1, 0, 0},{34615, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{23532, {{34616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23533, {{34617, 1, 0, 0},{34617, 1, 1, 17}, {0, 0, 0, 0}}},{23533, {{34617, 1, 0, 0},{34617, 1, 1, 17}, {0, 0, 0, 0}}},{23534, {{34617, 1, 0, 0},{34617, 1, 1, 18}, {0, 0, 0, 0}}},{23534, {{34617, 1, 0, 0},{34617, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23535, {{34618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23536, {{34619, 1, 0, 0},{34619, 1, 1, 18}, {0, 0, 0, 0}}},{23536, {{34619, 1, 0, 0},{34619, 1, 1, 18}, {0, 0, 0, 0}}},{23537, {{34619, 1, 0, 0},{34619, 1, 1, 19}, {0, 0, 0, 0}}},{23537, {{34619, 1, 0, 0},{34619, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{23538, {{34620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23539, {{34621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23540, {{34622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23541, {{34623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23542, {{34624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23543, {{34625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23544, {{34626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23545, {{34627, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23546, {{34628, 1, 0, 0},{34628, 1, 1, 16}, {0, 0, 0, 0}}},{23546, {{34628, 1, 0, 0},{34628, 1, 1, 16}, {0, 0, 0, 0}}},{58634, {{34628, 1, 0, 0},{34628, 1, 1, 17}, {0, 0, 0, 0}}},{58634, {{34628, 1, 0, 0},{34628, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{23547, {{34629, 1, 0, 0},{34629, 1, 1, 17}, {0, 0, 0, 0}}},{23547, {{34629, 1, 0, 0},{34629, 1, 1, 17}, {0, 0, 0, 0}}},{23548, {{34629, 1, 0, 0},{34629, 1, 1, 18}, {0, 0, 0, 0}}},{23548, {{34629, 1, 0, 0},{34629, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23549, {{34630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23550, {{34631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{23551, {{34633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{23552, {{34635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23553, {{34636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23554, {{34637, 1, 0, 0},{34637, 1, 1, 17}, {0, 0, 0, 0}}},{23554, {{34637, 1, 0, 0},{34637, 1, 1, 17}, {0, 0, 0, 0}}},{23555, {{34637, 1, 0, 0},{34637, 1, 1, 18}, {0, 0, 0, 0}}},{23555, {{34637, 1, 0, 0},{34637, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{23556, {{34638, 1, 0, 0},{34638, 1, 1, 17}, {0, 0, 0, 0}}},{23556, {{34638, 1, 0, 0},{34638, 1, 1, 17}, {0, 0, 0, 0}}},{23557, {{34638, 1, 0, 0},{34638, 1, 1, 18}, {0, 0, 0, 0}}},{23557, {{34638, 1, 0, 0},{34638, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23558, {{34639, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23559, {{34640, 1, 0, 0},{34640, 1, 1, 16}, {0, 0, 0, 0}}},{23559, {{34640, 1, 0, 0},{34640, 1, 1, 16}, {0, 0, 0, 0}}},{23560, {{34640, 1, 0, 0},{34640, 1, 1, 17}, {0, 0, 0, 0}}},{23560, {{34640, 1, 0, 0},{34640, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{23561, {{34641, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23562, {{34642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23563, {{34643, 1, 0, 0},{34643, 1, 1, 17}, {0, 0, 0, 0}}},{23563, {{34643, 1, 0, 0},{34643, 1, 1, 17}, {0, 0, 0, 0}}},{23564, {{34643, 1, 0, 0},{34643, 1, 1, 18}, {0, 0, 0, 0}}},{23564, {{34643, 1, 0, 0},{34643, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23565, {{34644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23566, {{34645, 1, 0, 0},{34645, 1, 1, 18}, {0, 0, 0, 0}}},{23566, {{34645, 1, 0, 0},{34645, 1, 1, 18}, {0, 0, 0, 0}}},{23567, {{34645, 1, 0, 0},{34645, 1, 1, 19}, {0, 0, 0, 0}}},{23567, {{34645, 1, 0, 0},{34645, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{23568, {{34646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23569, {{34647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23570, {{34648, 1, 0, 0},{34648, 1, 1, 17}, {0, 0, 0, 0}}},{23570, {{34648, 1, 0, 0},{34648, 1, 1, 17}, {0, 0, 0, 0}}},{23571, {{34648, 1, 0, 0},{34648, 1, 1, 18}, {0, 0, 0, 0}}},{23571, {{34648, 1, 0, 0},{34648, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{23572, {{34649, 1, 0, 0},{34649, 1, 1, 19}, {0, 0, 0, 0}}},{23572, {{34649, 1, 0, 0},{34649, 1, 1, 19}, {0, 0, 0, 0}}},{23573, {{34649, 1, 0, 0},{34649, 1, 1, 18}, {0, 0, 0, 0}}},{23573, {{34649, 1, 0, 0},{34649, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23574, {{34650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23575, {{34651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23576, {{34652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23577, {{34653, 1, 0, 0},{34653, 1, 1, 17}, {0, 0, 0, 0}}},{23577, {{34653, 1, 0, 0},{34653, 1, 1, 17}, {0, 0, 0, 0}}},{23578, {{34653, 1, 0, 0},{34653, 1, 1, 18}, {0, 0, 0, 0}}},{23578, {{34653, 1, 0, 0},{34653, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23579, {{34654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23580, {{34655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23581, {{34656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23582, {{34657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23583, {{34658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23584, {{34659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23585, {{34660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23586, {{34661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23587, {{34662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23588, {{34663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23589, {{34664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23590, {{34665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23591, {{34666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23592, {{34667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23594, {{34668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23595, {{34669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23596, {{34670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23597, {{34671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23598, {{34672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23599, {{34673, 1, 0, 0},{34673, 1, 1, 17}, {0, 0, 0, 0}}},{23599, {{34673, 1, 0, 0},{34673, 1, 1, 17}, {0, 0, 0, 0}}},{23600, {{34673, 1, 0, 0},{34673, 1, 1, 18}, {0, 0, 0, 0}}},{23600, {{34673, 1, 0, 0},{34673, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23601, {{34674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{23602, {{34675, 1, 0, 0},{34675, 1, 1, 16}, {0, 0, 0, 0}}},{23602, {{34675, 1, 0, 0},{34675, 1, 1, 16}, {0, 0, 0, 0}}},{23603, {{34675, 1, 0, 0},{34675, 1, 1, 17}, {0, 0, 0, 0}}},{23603, {{34675, 1, 0, 0},{34675, 1, 1, 17}, {0, 0, 0, 0}}},{23604, {{34675, 1, 0, 0},{34675, 1, 1, 18}, {0, 0, 0, 0}}},{23604, {{34675, 1, 0, 0},{34675, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23605, {{34676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23606, {{34677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23607, {{34678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23608, {{34679, 1, 0, 0},{34679, 1, 1, 16}, {0, 0, 0, 0}}},{23608, {{34679, 1, 0, 0},{34679, 1, 1, 16}, {0, 0, 0, 0}}},{23609, {{34679, 1, 0, 0},{34679, 1, 1, 17}, {0, 0, 0, 0}}},{23609, {{34679, 1, 0, 0},{34679, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{23610, {{34680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23611, {{34681, 1, 0, 0},{34681, 1, 1, 17}, {0, 0, 0, 0}}},{23611, {{34681, 1, 0, 0},{34681, 1, 1, 17}, {0, 0, 0, 0}}},{23612, {{34681, 1, 0, 0},{34681, 1, 1, 16}, {0, 0, 0, 0}}},{23612, {{34681, 1, 0, 0},{34681, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{23613, {{34682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23614, {{34683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23615, {{34684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23616, {{34685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{23617, {{34687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {8, {{23618, {{34690, 1, 0, 0},{34690, 1, 1, 18}, {0, 0, 0, 0}}},{23618, {{34690, 1, 0, 0},{34690, 1, 1, 18}, {0, 0, 0, 0}}},{23619, {{34690, 1, 0, 0},{34690, 1, 1, 21}, {0, 0, 0, 0}}},{23619, {{34690, 1, 0, 0},{34690, 1, 1, 21}, {0, 0, 0, 0}}},{23620, {{34690, 1, 0, 0},{34690, 1, 1, 20}, {0, 0, 0, 0}}},{23620, {{34690, 1, 0, 0},{34690, 1, 1, 20}, {0, 0, 0, 0}}},{23621, {{34690, 1, 0, 0},{34690, 1, 1, 19}, {0, 0, 0, 0}}},{23621, {{34690, 1, 0, 0},{34690, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{23622, {{34691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23623, {{34692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23624, {{34693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23625, {{34694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23626, {{34695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23627, {{34696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23628, {{34697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23629, {{34698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23630, {{34699, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23631, {{34700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23632, {{34701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23633, {{34702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23634, {{34703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23635, {{34704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23636, {{34705, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23637, {{34706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23638, {{34707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23639, {{34708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23640, {{34709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23641, {{34710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23642, {{34711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23643, {{34712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23644, {{34713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23645, {{34714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23646, {{34715, 1, 0, 0},{34715, 1, 1, 16}, {0, 0, 0, 0}}},{23646, {{34715, 1, 0, 0},{34715, 1, 1, 16}, {0, 0, 0, 0}}},{23647, {{34715, 1, 0, 0},{34715, 1, 1, 17}, {0, 0, 0, 0}}},{23647, {{34715, 1, 0, 0},{34715, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{23648, {{34716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23649, {{34717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23650, {{34718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23651, {{34719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23652, {{34720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23653, {{34721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23654, {{34722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23655, {{34723, 1, 0, 0},{34723, 1, 1, 19}, {0, 0, 0, 0}}},{23655, {{34723, 1, 0, 0},{34723, 1, 1, 19}, {0, 0, 0, 0}}},{23656, {{34723, 1, 0, 0},{34723, 1, 1, 18}, {0, 0, 0, 0}}},{23656, {{34723, 1, 0, 0},{34723, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23657, {{34724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23658, {{34725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{23659, {{34727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{23660, {{34729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23661, {{34730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23662, {{34731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23663, {{34732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23664, {{34733, 1, 0, 0},{34733, 1, 1, 17}, {0, 0, 0, 0}}},{23664, {{34733, 1, 0, 0},{34733, 1, 1, 17}, {0, 0, 0, 0}}},{58654, {{34733, 1, 0, 0},{34733, 1, 1, 18}, {0, 0, 0, 0}}},{58654, {{34733, 1, 0, 0},{34733, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23665, {{34734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23666, {{34735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23667, {{34736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23668, {{34737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23669, {{34738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23670, {{34739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23671, {{34740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23672, {{34741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23673, {{34742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23674, {{34743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23675, {{34744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23676, {{34745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23677, {{34746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23678, {{34747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23679, {{34748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23680, {{34749, 1, 0, 0},{34749, 1, 1, 18}, {0, 0, 0, 0}}},{23680, {{34749, 1, 0, 0},{34749, 1, 1, 18}, {0, 0, 0, 0}}},{23681, {{34749, 1, 0, 0},{34749, 1, 1, 19}, {0, 0, 0, 0}}},{23681, {{34749, 1, 0, 0},{34749, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{23682, {{34750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23683, {{34751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23684, {{34752, 1, 0, 0},{34752, 1, 1, 17}, {0, 0, 0, 0}}},{23684, {{34752, 1, 0, 0},{34752, 1, 1, 17}, {0, 0, 0, 0}}},{23685, {{34752, 1, 0, 0},{34752, 1, 1, 18}, {0, 0, 0, 0}}},{23685, {{34752, 1, 0, 0},{34752, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23686, {{34753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23687, {{34754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23688, {{34755, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23689, {{34756, 1, 0, 0},{34756, 1, 1, 17}, {0, 0, 0, 0}}},{23689, {{34756, 1, 0, 0},{34756, 1, 1, 17}, {0, 0, 0, 0}}},{23690, {{34756, 1, 0, 0},{34756, 1, 1, 18}, {0, 0, 0, 0}}},{23690, {{34756, 1, 0, 0},{34756, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23691, {{34757, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23692, {{34758, 1, 0, 0},{34758, 1, 1, 17}, {0, 0, 0, 0}}},{23692, {{34758, 1, 0, 0},{34758, 1, 1, 17}, {0, 0, 0, 0}}},{23693, {{34758, 1, 0, 0},{34758, 1, 1, 18}, {0, 0, 0, 0}}},{23693, {{34758, 1, 0, 0},{34758, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{23694, {{34759, 1, 0, 0},{34759, 1, 1, 17}, {0, 0, 0, 0}}},{23694, {{34759, 1, 0, 0},{34759, 1, 1, 17}, {0, 0, 0, 0}}},{23695, {{34759, 1, 0, 0},{34759, 1, 1, 18}, {0, 0, 0, 0}}},{23695, {{34759, 1, 0, 0},{34759, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23696, {{34760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23697, {{34761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23698, {{34762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23699, {{34763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23700, {{34764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23701, {{34765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23702, {{34766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{23703, {{34768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {8, {{23704, {{34770, 1, 0, 0},{34770, 1, 1, 19}, {0, 0, 0, 0}}},{23704, {{34770, 1, 0, 0},{34770, 1, 1, 19}, {0, 0, 0, 0}}},{23705, {{34770, 1, 0, 0},{34770, 1, 1, 21}, {0, 0, 0, 0}}},{23705, {{34770, 1, 0, 0},{34770, 1, 1, 21}, {0, 0, 0, 0}}},{23706, {{34770, 1, 0, 0},{34770, 1, 1, 22}, {0, 0, 0, 0}}},{23706, {{34770, 1, 0, 0},{34770, 1, 1, 22}, {0, 0, 0, 0}}},{23707, {{34770, 1, 0, 0},{34770, 1, 1, 20}, {0, 0, 0, 0}}},{23707, {{34770, 1, 0, 0},{34770, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{23708, {{34771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23709, {{34772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23710, {{34773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23711, {{34774, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23712, {{34775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23713, {{34776, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23714, {{34777, 1, 0, 0},{34777, 1, 1, 17}, {0, 0, 0, 0}}},{23714, {{34777, 1, 0, 0},{34777, 1, 1, 17}, {0, 0, 0, 0}}},{23715, {{34777, 1, 0, 0},{34777, 1, 1, 18}, {0, 0, 0, 0}}},{23715, {{34777, 1, 0, 0},{34777, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23716, {{34778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23717, {{34779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23718, {{34780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23719, {{34781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23720, {{34782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23721, {{34783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23722, {{34784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23723, {{34785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23724, {{34786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23725, {{34787, 1, 0, 0},{34787, 1, 1, 17}, {0, 0, 0, 0}}},{23725, {{34787, 1, 0, 0},{34787, 1, 1, 17}, {0, 0, 0, 0}}},{23726, {{34787, 1, 0, 0},{34787, 1, 1, 18}, {0, 0, 0, 0}}},{23726, {{34787, 1, 0, 0},{34787, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23727, {{34788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23728, {{34789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23729, {{34790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23730, {{34791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23731, {{34792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23732, {{34793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23733, {{34794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23734, {{34795, 1, 0, 0},{34795, 1, 1, 17}, {0, 0, 0, 0}}},{23734, {{34795, 1, 0, 0},{34795, 1, 1, 17}, {0, 0, 0, 0}}},{23735, {{34795, 1, 0, 0},{34795, 1, 1, 18}, {0, 0, 0, 0}}},{23735, {{34795, 1, 0, 0},{34795, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23736, {{34796, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23737, {{34797, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23738, {{34798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23739, {{34799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23740, {{34800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23741, {{34801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23742, {{34802, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23743, {{34803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23744, {{34804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23745, {{34805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23746, {{34806, 1, 0, 0},{34806, 1, 1, 17}, {0, 0, 0, 0}}},{23746, {{34806, 1, 0, 0},{34806, 1, 1, 17}, {0, 0, 0, 0}}},{23747, {{34806, 1, 0, 0},{34806, 1, 1, 18}, {0, 0, 0, 0}}},{23747, {{34806, 1, 0, 0},{34806, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23748, {{34807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23749, {{34808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23750, {{34809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23751, {{34810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23752, {{34811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23753, {{34812, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23754, {{34813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23755, {{34814, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23756, {{34815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23757, {{34816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23758, {{34817, 1, 0, 0},{34817, 1, 1, 17}, {0, 0, 0, 0}}},{23758, {{34817, 1, 0, 0},{34817, 1, 1, 17}, {0, 0, 0, 0}}},{23759, {{34817, 1, 0, 0},{34817, 1, 1, 18}, {0, 0, 0, 0}}},{23759, {{34817, 1, 0, 0},{34817, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{23760, {{34818, 1, 0, 0},{34818, 1, 1, 16}, {0, 0, 0, 0}}},{23760, {{34818, 1, 0, 0},{34818, 1, 1, 16}, {0, 0, 0, 0}}},{23761, {{34818, 1, 0, 0},{34818, 1, 1, 17}, {0, 0, 0, 0}}},{23761, {{34818, 1, 0, 0},{34818, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{23762, {{34819, 1, 0, 0},{34819, 1, 1, 18}, {0, 0, 0, 0}}},{23762, {{34819, 1, 0, 0},{34819, 1, 1, 18}, {0, 0, 0, 0}}},{23763, {{34819, 1, 0, 0},{34819, 1, 1, 19}, {0, 0, 0, 0}}},{23763, {{34819, 1, 0, 0},{34819, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{23765, {{34820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23766, {{34821, 1, 0, 0},{34821, 1, 1, 19}, {0, 0, 0, 0}}},{23766, {{34821, 1, 0, 0},{34821, 1, 1, 19}, {0, 0, 0, 0}}},{23767, {{34821, 1, 0, 0},{34821, 1, 1, 20}, {0, 0, 0, 0}}},{23767, {{34821, 1, 0, 0},{34821, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{23768, {{34822, 1, 0, 0},{34822, 1, 1, 17}, {0, 0, 0, 0}}},{23768, {{34822, 1, 0, 0},{34822, 1, 1, 17}, {0, 0, 0, 0}}},{23769, {{34822, 1, 0, 0},{34822, 1, 1, 18}, {0, 0, 0, 0}}},{23769, {{34822, 1, 0, 0},{34822, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{23770, {{34823, 1, 0, 0},{34823, 1, 1, 17}, {0, 0, 0, 0}}},{23770, {{34823, 1, 0, 0},{34823, 1, 1, 17}, {0, 0, 0, 0}}},{23771, {{34823, 1, 0, 0},{34823, 1, 1, 18}, {0, 0, 0, 0}}},{23771, {{34823, 1, 0, 0},{34823, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23772, {{34824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23773, {{34825, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23774, {{34826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23775, {{34827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23776, {{34828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{23777, {{34829, 1, 0, 0},{34829, 1, 1, 17}, {0, 0, 0, 0}}},{23777, {{34829, 1, 0, 0},{34829, 1, 1, 17}, {0, 0, 0, 0}}},{23778, {{34829, 1, 0, 0},{34829, 1, 1, 19}, {0, 0, 0, 0}}},{23778, {{34829, 1, 0, 0},{34829, 1, 1, 19}, {0, 0, 0, 0}}},{23779, {{34829, 1, 0, 0},{34829, 1, 1, 18}, {0, 0, 0, 0}}},{23779, {{34829, 1, 0, 0},{34829, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {8, {{23780, {{34830, 1, 0, 0},{34830, 1, 1, 18}, {0, 0, 0, 0}}},{23780, {{34830, 1, 0, 0},{34830, 1, 1, 18}, {0, 0, 0, 0}}},{23781, {{34830, 1, 0, 0},{34830, 1, 1, 19}, {0, 0, 0, 0}}},{23781, {{34830, 1, 0, 0},{34830, 1, 1, 19}, {0, 0, 0, 0}}},{23782, {{34830, 1, 0, 0},{34830, 1, 1, 20}, {0, 0, 0, 0}}},{23782, {{34830, 1, 0, 0},{34830, 1, 1, 20}, {0, 0, 0, 0}}},{23783, {{34830, 1, 0, 0},{34830, 1, 1, 21}, {0, 0, 0, 0}}},{23783, {{34830, 1, 0, 0},{34830, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{23784, {{34831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23785, {{34832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23786, {{34833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23787, {{34834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23788, {{34835, 1, 0, 0},{34835, 1, 1, 17}, {0, 0, 0, 0}}},{23788, {{34835, 1, 0, 0},{34835, 1, 1, 17}, {0, 0, 0, 0}}},{23789, {{34835, 1, 0, 0},{34835, 1, 1, 18}, {0, 0, 0, 0}}},{23789, {{34835, 1, 0, 0},{34835, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23790, {{34836, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23791, {{34837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23792, {{34838, 1, 0, 0},{34838, 1, 1, 17}, {0, 0, 0, 0}}},{23792, {{34838, 1, 0, 0},{34838, 1, 1, 17}, {0, 0, 0, 0}}},{23793, {{34838, 1, 0, 0},{34838, 1, 1, 18}, {0, 0, 0, 0}}},{23793, {{34838, 1, 0, 0},{34838, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23794, {{34839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23795, {{34840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23796, {{34841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23797, {{34842, 1, 0, 0},{34842, 1, 1, 17}, {0, 0, 0, 0}}},{23797, {{34842, 1, 0, 0},{34842, 1, 1, 17}, {0, 0, 0, 0}}},{23798, {{34842, 1, 0, 0},{34842, 1, 1, 18}, {0, 0, 0, 0}}},{23798, {{34842, 1, 0, 0},{34842, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{23799, {{34843, 1, 0, 0},{34843, 1, 1, 18}, {0, 0, 0, 0}}},{23799, {{34843, 1, 0, 0},{34843, 1, 1, 18}, {0, 0, 0, 0}}},{23800, {{34843, 1, 0, 0},{34843, 1, 1, 17}, {0, 0, 0, 0}}},{23800, {{34843, 1, 0, 0},{34843, 1, 1, 17}, {0, 0, 0, 0}}},{23801, {{34843, 1, 0, 0},{34843, 1, 1, 19}, {0, 0, 0, 0}}},{23801, {{34843, 1, 0, 0},{34843, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{23802, {{34844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23803, {{34845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23804, {{34846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23805, {{34847, 1, 0, 0},{34847, 1, 1, 17}, {0, 0, 0, 0}}},{23805, {{34847, 1, 0, 0},{34847, 1, 1, 17}, {0, 0, 0, 0}}},{58666, {{34847, 1, 0, 0},{34847, 1, 1, 18}, {0, 0, 0, 0}}},{58666, {{34847, 1, 0, 0},{34847, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23806, {{34848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{23807, {{34849, 1, 0, 0},{34849, 1, 1, 17}, {0, 0, 0, 0}}},{23807, {{34849, 1, 0, 0},{34849, 1, 1, 17}, {0, 0, 0, 0}}},{23808, {{34849, 1, 0, 0},{34849, 1, 1, 18}, {0, 0, 0, 0}}},{23808, {{34849, 1, 0, 0},{34849, 1, 1, 18}, {0, 0, 0, 0}}},{23809, {{34849, 1, 0, 0},{34849, 1, 1, 19}, {0, 0, 0, 0}}},{23809, {{34849, 1, 0, 0},{34849, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{23810, {{34850, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23811, {{34851, 1, 0, 0},{34851, 1, 1, 17}, {0, 0, 0, 0}}},{23811, {{34851, 1, 0, 0},{34851, 1, 1, 17}, {0, 0, 0, 0}}},{23812, {{34851, 1, 0, 0},{34851, 1, 1, 18}, {0, 0, 0, 0}}},{23812, {{34851, 1, 0, 0},{34851, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{23813, {{34853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23814, {{34854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23815, {{34855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23816, {{34856, 1, 0, 0},{34856, 1, 1, 17}, {0, 0, 0, 0}}},{23816, {{34856, 1, 0, 0},{34856, 1, 1, 17}, {0, 0, 0, 0}}},{23817, {{34856, 1, 0, 0},{34856, 1, 1, 18}, {0, 0, 0, 0}}},{23817, {{34856, 1, 0, 0},{34856, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23818, {{34857, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23819, {{34858, 1, 0, 0},{34858, 1, 1, 16}, {0, 0, 0, 0}}},{23819, {{34858, 1, 0, 0},{34858, 1, 1, 16}, {0, 0, 0, 0}}},{23820, {{34858, 1, 0, 0},{34858, 1, 1, 17}, {0, 0, 0, 0}}},{23820, {{34858, 1, 0, 0},{34858, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{23821, {{34859, 1, 0, 0},{34859, 1, 1, 16}, {0, 0, 0, 0}}},{23821, {{34859, 1, 0, 0},{34859, 1, 1, 16}, {0, 0, 0, 0}}},{23822, {{34859, 1, 0, 0},{34859, 1, 1, 17}, {0, 0, 0, 0}}},{23822, {{34859, 1, 0, 0},{34859, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{23823, {{34860, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23824, {{34861, 1, 0, 0},{34861, 1, 1, 18}, {0, 0, 0, 0}}},{23824, {{34861, 1, 0, 0},{34861, 1, 1, 18}, {0, 0, 0, 0}}},{23825, {{34861, 1, 0, 0},{34861, 1, 1, 17}, {0, 0, 0, 0}}},{23825, {{34861, 1, 0, 0},{34861, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{23826, {{34862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23827, {{34863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23828, {{34864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23829, {{34865, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23830, {{34866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23831, {{34867, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{23832, {{34869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{23833, {{34870, 1, 0, 0},{34870, 1, 1, 20}, {0, 0, 0, 0}}},{23833, {{34870, 1, 0, 0},{34870, 1, 1, 20}, {0, 0, 0, 0}}},{23834, {{34870, 1, 0, 0},{34870, 1, 1, 19}, {0, 0, 0, 0}}},{23834, {{34870, 1, 0, 0},{34870, 1, 1, 19}, {0, 0, 0, 0}}},{23835, {{34870, 1, 0, 0},{34870, 1, 1, 21}, {0, 0, 0, 0}}},{23835, {{34870, 1, 0, 0},{34870, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{23836, {{34871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23837, {{34872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23838, {{34873, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23839, {{34874, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23840, {{34875, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23841, {{34876, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23842, {{34877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23843, {{34878, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23844, {{34879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23845, {{34880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23846, {{34881, 1, 0, 0},{34881, 1, 1, 17}, {0, 0, 0, 0}}},{23846, {{34881, 1, 0, 0},{34881, 1, 1, 17}, {0, 0, 0, 0}}},{23847, {{34881, 1, 0, 0},{34881, 1, 1, 18}, {0, 0, 0, 0}}},{23847, {{34881, 1, 0, 0},{34881, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{23848, {{34882, 1, 0, 0},{34882, 1, 1, 19}, {0, 0, 0, 0}}},{23848, {{34882, 1, 0, 0},{34882, 1, 1, 19}, {0, 0, 0, 0}}},{23849, {{34882, 1, 0, 0},{34882, 1, 1, 20}, {0, 0, 0, 0}}},{23849, {{34882, 1, 0, 0},{34882, 1, 1, 20}, {0, 0, 0, 0}}},{23850, {{34882, 1, 0, 0},{34882, 1, 1, 21}, {0, 0, 0, 0}}},{23850, {{34882, 1, 0, 0},{34882, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{23851, {{34883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23852, {{34884, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23853, {{34885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{23854, {{34886, 1, 0, 0},{34886, 1, 1, 18}, {0, 0, 0, 0}}},{23854, {{34886, 1, 0, 0},{34886, 1, 1, 18}, {0, 0, 0, 0}}},{58673, {{34886, 1, 0, 0},{34886, 1, 1, 19}, {0, 0, 0, 0}}},{58673, {{34886, 1, 0, 0},{34886, 1, 1, 19}, {0, 0, 0, 0}}},{58675, {{34886, 1, 0, 0},{34886, 1, 1, 20}, {0, 0, 0, 0}}},{58675, {{34886, 1, 0, 0},{34886, 1, 1, 20}, {0, 0, 0, 0}}},{58677, {{34886, 1, 0, 0},{34886, 1, 1, 21}, {0, 0, 0, 0}}},{58677, {{34886, 1, 0, 0},{34886, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{23855, {{34887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23856, {{34888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23857, {{34889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{23858, {{34890, 1, 0, 0},{34890, 1, 1, 19}, {0, 0, 0, 0}}},{23858, {{34890, 1, 0, 0},{34890, 1, 1, 19}, {0, 0, 0, 0}}},{23859, {{34890, 1, 0, 0},{34890, 1, 1, 18}, {0, 0, 0, 0}}},{23859, {{34890, 1, 0, 0},{34890, 1, 1, 18}, {0, 0, 0, 0}}},{23860, {{34890, 1, 0, 0},{34890, 1, 1, 17}, {0, 0, 0, 0}}},{23860, {{34890, 1, 0, 0},{34890, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{23861, {{34891, 1, 0, 0},{34891, 1, 1, 18}, {0, 0, 0, 0}}},{23861, {{34891, 1, 0, 0},{34891, 1, 1, 18}, {0, 0, 0, 0}}},{23862, {{34891, 1, 0, 0},{34891, 1, 1, 19}, {0, 0, 0, 0}}},{23862, {{34891, 1, 0, 0},{34891, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{23863, {{34892, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23864, {{34893, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23865, {{34894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23866, {{34895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{23867, {{34897, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23869, {{34898, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{23870, {{34899, 1, 0, 0},{34899, 1, 1, 19}, {0, 0, 0, 0}}},{23870, {{34899, 1, 0, 0},{34899, 1, 1, 19}, {0, 0, 0, 0}}},{23871, {{34899, 1, 0, 0},{34899, 1, 1, 18}, {0, 0, 0, 0}}},{23871, {{34899, 1, 0, 0},{34899, 1, 1, 18}, {0, 0, 0, 0}}},{58680, {{34899, 1, 0, 0},{34899, 1, 1, 20}, {0, 0, 0, 0}}},{58680, {{34899, 1, 0, 0},{34899, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{23872, {{34901, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23873, {{34902, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23874, {{34903, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23875, {{34904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23876, {{34905, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23877, {{34906, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{23878, {{34907, 1, 0, 0},{34907, 1, 1, 19}, {0, 0, 0, 0}}},{23878, {{34907, 1, 0, 0},{34907, 1, 1, 19}, {0, 0, 0, 0}}},{23879, {{34907, 1, 0, 0},{34907, 1, 1, 18}, {0, 0, 0, 0}}},{23879, {{34907, 1, 0, 0},{34907, 1, 1, 18}, {0, 0, 0, 0}}},{23880, {{34907, 1, 0, 0},{34907, 1, 1, 21}, {0, 0, 0, 0}}},{23880, {{34907, 1, 0, 0},{34907, 1, 1, 21}, {0, 0, 0, 0}}},{58685, {{34907, 1, 0, 0},{34907, 1, 1, 20}, {0, 0, 0, 0}}},{58685, {{34907, 1, 0, 0},{34907, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{23881, {{34908, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23882, {{34909, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23883, {{34910, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23884, {{34911, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23885, {{34912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23886, {{34913, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23888, {{34914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23889, {{34915, 1, 0, 0},{34915, 1, 1, 18}, {0, 0, 0, 0}}},{23889, {{34915, 1, 0, 0},{34915, 1, 1, 18}, {0, 0, 0, 0}}},{60165, {{34915, 1, 0, 0},{34915, 1, 1, 19}, {0, 0, 0, 0}}},{60165, {{34915, 1, 0, 0},{34915, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{23890, {{34916, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{23891, {{34918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23892, {{34919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23893, {{34920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23894, {{34921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23895, {{34922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23896, {{34923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{23897, {{34925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23898, {{34926, 1, 0, 0},{34926, 1, 1, 17}, {0, 0, 0, 0}}},{23898, {{34926, 1, 0, 0},{34926, 1, 1, 17}, {0, 0, 0, 0}}},{23899, {{34926, 1, 0, 0},{34926, 1, 1, 18}, {0, 0, 0, 0}}},{23899, {{34926, 1, 0, 0},{34926, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23900, {{34927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23901, {{34928, 1, 0, 0},{34928, 1, 1, 18}, {0, 0, 0, 0}}},{23901, {{34928, 1, 0, 0},{34928, 1, 1, 18}, {0, 0, 0, 0}}},{58692, {{34928, 1, 0, 0},{34928, 1, 1, 20}, {0, 0, 0, 0}}},{58692, {{34928, 1, 0, 0},{34928, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{23902, {{34929, 1, 0, 0},{34929, 1, 1, 18}, {0, 0, 0, 0}}},{23902, {{34929, 1, 0, 0},{34929, 1, 1, 18}, {0, 0, 0, 0}}},{23903, {{34929, 1, 0, 0},{34929, 1, 1, 17}, {0, 0, 0, 0}}},{23903, {{34929, 1, 0, 0},{34929, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{23904, {{34930, 1, 0, 0},{34930, 1, 1, 17}, {0, 0, 0, 0}}},{23904, {{34930, 1, 0, 0},{34930, 1, 1, 17}, {0, 0, 0, 0}}},{23905, {{34930, 1, 0, 0},{34930, 1, 1, 18}, {0, 0, 0, 0}}},{23905, {{34930, 1, 0, 0},{34930, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23906, {{34931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23907, {{34932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23908, {{34933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23909, {{34934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23910, {{34935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23911, {{34936, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23912, {{34937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23913, {{34938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23914, {{34939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23915, {{34940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23916, {{34941, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23917, {{34942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23918, {{34943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23919, {{34944, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23920, {{34945, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23921, {{34946, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23922, {{34947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23923, {{34948, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{23924, {{34952, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23925, {{34953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23926, {{34954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23927, {{34955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23928, {{34956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23929, {{34957, 1, 0, 0},{34957, 1, 1, 17}, {0, 0, 0, 0}}},{23929, {{34957, 1, 0, 0},{34957, 1, 1, 17}, {0, 0, 0, 0}}},{23930, {{34957, 1, 0, 0},{34957, 1, 1, 18}, {0, 0, 0, 0}}},{23930, {{34957, 1, 0, 0},{34957, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23931, {{34958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23932, {{34959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23933, {{34960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23934, {{34961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23935, {{34962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23936, {{34963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23937, {{34964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23938, {{34965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23939, {{34966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23940, {{34967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23941, {{34968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23942, {{34969, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23943, {{34970, 1, 0, 0},{34970, 1, 1, 17}, {0, 0, 0, 0}}},{23943, {{34970, 1, 0, 0},{34970, 1, 1, 17}, {0, 0, 0, 0}}},{23944, {{34970, 1, 0, 0},{34970, 1, 1, 18}, {0, 0, 0, 0}}},{23944, {{34970, 1, 0, 0},{34970, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23945, {{34971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23946, {{34972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23947, {{34973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23948, {{34974, 1, 0, 0},{34974, 1, 1, 18}, {0, 0, 0, 0}}},{23948, {{34974, 1, 0, 0},{34974, 1, 1, 18}, {0, 0, 0, 0}}},{23949, {{34974, 1, 0, 0},{34974, 1, 1, 19}, {0, 0, 0, 0}}},{23949, {{34974, 1, 0, 0},{34974, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{23950, {{34975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23951, {{34976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23952, {{34977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23953, {{34978, 1, 0, 0},{34978, 1, 1, 17}, {0, 0, 0, 0}}},{23953, {{34978, 1, 0, 0},{34978, 1, 1, 17}, {0, 0, 0, 0}}},{23954, {{34978, 1, 0, 0},{34978, 1, 1, 18}, {0, 0, 0, 0}}},{23954, {{34978, 1, 0, 0},{34978, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{23955, {{34979, 1, 0, 0},{34979, 1, 1, 16}, {0, 0, 0, 0}}},{23955, {{34979, 1, 0, 0},{34979, 1, 1, 16}, {0, 0, 0, 0}}},{23956, {{34979, 1, 0, 0},{34979, 1, 1, 17}, {0, 0, 0, 0}}},{23956, {{34979, 1, 0, 0},{34979, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{23957, {{34980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23958, {{34981, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23959, {{34982, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23960, {{34983, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23961, {{34984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23962, {{34985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23963, {{34986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23964, {{34987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23965, {{34988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{23966, {{34990, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{23967, {{34992, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23968, {{34993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23969, {{34994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23970, {{34995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23971, {{34996, 1, 0, 0},{34996, 1, 1, 17}, {0, 0, 0, 0}}},{23971, {{34996, 1, 0, 0},{34996, 1, 1, 17}, {0, 0, 0, 0}}},{58694, {{34996, 1, 0, 0},{34996, 1, 1, 18}, {0, 0, 0, 0}}},{58694, {{34996, 1, 0, 0},{34996, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23972, {{34997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23973, {{34998, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23974, {{34999, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23975, {{35000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23976, {{35001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23977, {{35002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23978, {{35003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23979, {{35004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23980, {{35005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23981, {{35006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23982, {{35007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23983, {{35008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23984, {{35009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23985, {{35010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23986, {{35011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23987, {{35012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23988, {{35013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{23989, {{35018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23990, {{35019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23991, {{35020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23992, {{35021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23993, {{35022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23994, {{35023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23995, {{35024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{23996, {{35025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23997, {{35026, 1, 0, 0},{35026, 1, 1, 17}, {0, 0, 0, 0}}},{23997, {{35026, 1, 0, 0},{35026, 1, 1, 17}, {0, 0, 0, 0}}},{23998, {{35026, 1, 0, 0},{35026, 1, 1, 18}, {0, 0, 0, 0}}},{23998, {{35026, 1, 0, 0},{35026, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{23999, {{35027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24000, {{35028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24001, {{35029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24003, {{35030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24004, {{35031, 1, 0, 0},{35031, 1, 1, 17}, {0, 0, 0, 0}}},{24004, {{35031, 1, 0, 0},{35031, 1, 1, 17}, {0, 0, 0, 0}}},{24005, {{35031, 1, 0, 0},{35031, 1, 1, 16}, {0, 0, 0, 0}}},{24005, {{35031, 1, 0, 0},{35031, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{24006, {{35032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24007, {{35033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24008, {{35034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24009, {{35035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24010, {{35036, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24011, {{35037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24012, {{35038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24013, {{35039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24014, {{35040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24015, {{35041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{24016, {{35047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24017, {{35048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24018, {{35049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24019, {{35050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24020, {{35051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24021, {{35052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24022, {{35053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24023, {{35054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24024, {{35055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24025, {{35056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24026, {{35057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24027, {{35058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24028, {{35059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24029, {{35060, 1, 0, 0},{35060, 1, 1, 18}, {0, 0, 0, 0}}},{24029, {{35060, 1, 0, 0},{35060, 1, 1, 18}, {0, 0, 0, 0}}},{24030, {{35060, 1, 0, 0},{35060, 1, 1, 19}, {0, 0, 0, 0}}},{24030, {{35060, 1, 0, 0},{35060, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{24031, {{35061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24033, {{35062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24034, {{35063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24035, {{35064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24036, {{35065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24037, {{35066, 1, 0, 0},{35066, 1, 1, 16}, {0, 0, 0, 0}}},{24037, {{35066, 1, 0, 0},{35066, 1, 1, 16}, {0, 0, 0, 0}}},{24038, {{35066, 1, 0, 0},{35066, 1, 1, 17}, {0, 0, 0, 0}}},{24038, {{35066, 1, 0, 0},{35066, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{24039, {{35067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24040, {{35068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24041, {{35069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24042, {{35070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24043, {{35071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24044, {{35072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24045, {{35073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24046, {{35074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24047, {{35075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24048, {{35076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24049, {{35077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24050, {{35078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24051, {{35079, 1, 0, 0},{35079, 1, 1, 18}, {0, 0, 0, 0}}},{24051, {{35079, 1, 0, 0},{35079, 1, 1, 18}, {0, 0, 0, 0}}},{24052, {{35079, 1, 0, 0},{35079, 1, 1, 17}, {0, 0, 0, 0}}},{24052, {{35079, 1, 0, 0},{35079, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{24053, {{35080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24054, {{35081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24055, {{35082, 1, 0, 0},{35082, 1, 1, 19}, {0, 0, 0, 0}}},{24055, {{35082, 1, 0, 0},{35082, 1, 1, 19}, {0, 0, 0, 0}}},{24056, {{35082, 1, 0, 0},{35082, 1, 1, 18}, {0, 0, 0, 0}}},{24056, {{35082, 1, 0, 0},{35082, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24057, {{35083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24058, {{35084, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24059, {{35085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24060, {{35086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24061, {{35087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{24062, {{35088, 1, 0, 0},{35088, 1, 1, 18}, {0, 0, 0, 0}}},{24062, {{35088, 1, 0, 0},{35088, 1, 1, 18}, {0, 0, 0, 0}}},{30295, {{35088, 1, 0, 0},{35088, 1, 1, 0},{35088, 1, 1, 19},{64096, 1, 0, 0}, {0, 0, 0, 0}}},{30295, {{35088, 1, 0, 0},{35088, 1, 1, 0},{35088, 1, 1, 19},{64096, 1, 0, 0}, {0, 0, 0, 0}}},{30295, {{35088, 1, 0, 0},{35088, 1, 1, 0},{35088, 1, 1, 19},{64096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24063, {{35089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24064, {{35090, 1, 0, 0},{35090, 1, 1, 17}, {0, 0, 0, 0}}},{24064, {{35090, 1, 0, 0},{35090, 1, 1, 17}, {0, 0, 0, 0}}},{24065, {{35090, 1, 0, 0},{35090, 1, 1, 18}, {0, 0, 0, 0}}},{24065, {{35090, 1, 0, 0},{35090, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24066, {{35091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24067, {{35092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24068, {{35093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24069, {{35094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24070, {{35095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24071, {{35096, 1, 0, 0},{35096, 1, 1, 18}, {0, 0, 0, 0}}},{24071, {{35096, 1, 0, 0},{35096, 1, 1, 18}, {0, 0, 0, 0}}},{24072, {{35096, 1, 0, 0},{35096, 1, 1, 17}, {0, 0, 0, 0}}},{24072, {{35096, 1, 0, 0},{35096, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{24073, {{35097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24074, {{35098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24075, {{35099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24076, {{35100, 1, 0, 0},{35100, 1, 1, 18}, {0, 0, 0, 0}}},{24076, {{35100, 1, 0, 0},{35100, 1, 1, 18}, {0, 0, 0, 0}}},{24077, {{35100, 1, 0, 0},{35100, 1, 1, 19}, {0, 0, 0, 0}}},{24077, {{35100, 1, 0, 0},{35100, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{24078, {{35101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24079, {{35102, 1, 0, 0},{35102, 1, 1, 18}, {0, 0, 0, 0}}},{24079, {{35102, 1, 0, 0},{35102, 1, 1, 18}, {0, 0, 0, 0}}},{24080, {{35102, 1, 0, 0},{35102, 1, 1, 17}, {0, 0, 0, 0}}},{24080, {{35102, 1, 0, 0},{35102, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{24081, {{35103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24082, {{35104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24083, {{35105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24084, {{35106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24085, {{35107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24086, {{35108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24087, {{35109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24088, {{35110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24089, {{35111, 1, 0, 0},{35111, 1, 1, 17}, {0, 0, 0, 0}}},{24089, {{35111, 1, 0, 0},{35111, 1, 1, 17}, {0, 0, 0, 0}}},{24090, {{35111, 1, 0, 0},{35111, 1, 1, 18}, {0, 0, 0, 0}}},{24090, {{35111, 1, 0, 0},{35111, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24091, {{35112, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24092, {{35113, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24093, {{35114, 1, 0, 0},{35114, 1, 1, 18}, {0, 0, 0, 0}}},{24093, {{35114, 1, 0, 0},{35114, 1, 1, 18}, {0, 0, 0, 0}}},{24094, {{35114, 1, 0, 0},{35114, 1, 1, 17}, {0, 0, 0, 0}}},{24094, {{35114, 1, 0, 0},{35114, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{24095, {{35115, 1, 0, 0},{35115, 1, 1, 18}, {0, 0, 0, 0}}},{24095, {{35115, 1, 0, 0},{35115, 1, 1, 18}, {0, 0, 0, 0}}},{24096, {{35115, 1, 0, 0},{35115, 1, 1, 19}, {0, 0, 0, 0}}},{24096, {{35115, 1, 0, 0},{35115, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{24097, {{35116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24098, {{35117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24099, {{35118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24100, {{35119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24101, {{35120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24102, {{35121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24103, {{35122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24104, {{35123, 1, 0, 0},{35123, 1, 1, 16}, {0, 0, 0, 0}}},{24104, {{35123, 1, 0, 0},{35123, 1, 1, 16}, {0, 0, 0, 0}}},{24105, {{35123, 1, 0, 0},{35123, 1, 1, 17}, {0, 0, 0, 0}}},{24105, {{35123, 1, 0, 0},{35123, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{24106, {{35125, 1, 0, 0},{35125, 1, 1, 18}, {0, 0, 0, 0}}},{24106, {{35125, 1, 0, 0},{35125, 1, 1, 18}, {0, 0, 0, 0}}},{24107, {{35125, 1, 0, 0},{35125, 1, 1, 19}, {0, 0, 0, 0}}},{24107, {{35125, 1, 0, 0},{35125, 1, 1, 19}, {0, 0, 0, 0}}},{24108, {{35125, 1, 0, 0},{35125, 1, 1, 17}, {0, 0, 0, 0}}},{24108, {{35125, 1, 0, 0},{35125, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{24109, {{35126, 1, 0, 0},{35126, 1, 1, 17}, {0, 0, 0, 0}}},{24109, {{35126, 1, 0, 0},{35126, 1, 1, 17}, {0, 0, 0, 0}}},{24110, {{35126, 1, 0, 0},{35126, 1, 1, 18}, {0, 0, 0, 0}}},{24110, {{35126, 1, 0, 0},{35126, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24111, {{35127, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24112, {{35128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24113, {{35129, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24114, {{35130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24115, {{35131, 1, 0, 0},{35131, 1, 1, 18}, {0, 0, 0, 0}}},{24115, {{35131, 1, 0, 0},{35131, 1, 1, 18}, {0, 0, 0, 0}}},{24116, {{35131, 1, 0, 0},{35131, 1, 1, 19}, {0, 0, 0, 0}}},{24116, {{35131, 1, 0, 0},{35131, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{24117, {{35132, 1, 0, 0},{35132, 1, 1, 16}, {0, 0, 0, 0}}},{24117, {{35132, 1, 0, 0},{35132, 1, 1, 16}, {0, 0, 0, 0}}},{24118, {{35132, 1, 0, 0},{35132, 1, 1, 17}, {0, 0, 0, 0}}},{24118, {{35132, 1, 0, 0},{35132, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{24119, {{35133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24120, {{35134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24121, {{35135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24122, {{35136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24123, {{35137, 1, 0, 0},{35137, 1, 1, 18}, {0, 0, 0, 0}}},{24123, {{35137, 1, 0, 0},{35137, 1, 1, 18}, {0, 0, 0, 0}}},{24124, {{35137, 1, 0, 0},{35137, 1, 1, 19}, {0, 0, 0, 0}}},{24124, {{35137, 1, 0, 0},{35137, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{24125, {{35138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24126, {{35139, 1, 0, 0},{35139, 1, 1, 17}, {0, 0, 0, 0}}},{24126, {{35139, 1, 0, 0},{35139, 1, 1, 17}, {0, 0, 0, 0}}},{24127, {{35139, 1, 0, 0},{35139, 1, 1, 18}, {0, 0, 0, 0}}},{24127, {{35139, 1, 0, 0},{35139, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{24128, {{35140, 1, 0, 0},{35140, 1, 1, 17}, {0, 0, 0, 0}}},{24128, {{35140, 1, 0, 0},{35140, 1, 1, 17}, {0, 0, 0, 0}}},{59302, {{35140, 1, 0, 0},{35140, 1, 1, 18}, {0, 0, 0, 0}}},{59302, {{35140, 1, 0, 0},{35140, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24129, {{35141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24130, {{35142, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24131, {{35143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24132, {{35144, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24133, {{35145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24134, {{35146, 1, 0, 0},{35146, 1, 1, 16}, {0, 0, 0, 0}}},{24134, {{35146, 1, 0, 0},{35146, 1, 1, 16}, {0, 0, 0, 0}}},{24135, {{35146, 1, 0, 0},{35146, 1, 1, 17}, {0, 0, 0, 0}}},{24135, {{35146, 1, 0, 0},{35146, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{24136, {{35147, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24137, {{35148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24138, {{35149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24139, {{35150, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24140, {{35151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24141, {{35152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24142, {{35153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24143, {{35154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24144, {{35155, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{24145, {{35158, 1, 0, 0},{35158, 1, 1, 18}, {0, 0, 0, 0}}},{24145, {{35158, 1, 0, 0},{35158, 1, 1, 18}, {0, 0, 0, 0}}},{24146, {{35158, 1, 0, 0},{35158, 1, 1, 19}, {0, 0, 0, 0}}},{24146, {{35158, 1, 0, 0},{35158, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{24147, {{35159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24148, {{35160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24149, {{35161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24150, {{35162, 1, 0, 0},{35162, 1, 1, 18}, {0, 0, 0, 0}}},{24150, {{35162, 1, 0, 0},{35162, 1, 1, 18}, {0, 0, 0, 0}}},{24151, {{35162, 1, 0, 0},{35162, 1, 1, 17}, {0, 0, 0, 0}}},{24151, {{35162, 1, 0, 0},{35162, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{24152, {{35163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24153, {{35164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24154, {{35165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24155, {{35166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24156, {{35167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24157, {{35168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24158, {{35169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24159, {{35170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24160, {{35171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24161, {{35172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24162, {{35173, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24163, {{35174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{24164, {{35177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{24165, {{35178, 1, 0, 0},{35178, 1, 1, 18}, {0, 0, 0, 0}}},{24165, {{35178, 1, 0, 0},{35178, 1, 1, 18}, {0, 0, 0, 0}}},{24166, {{35178, 1, 0, 0},{35178, 1, 1, 20}, {0, 0, 0, 0}}},{24166, {{35178, 1, 0, 0},{35178, 1, 1, 20}, {0, 0, 0, 0}}},{24167, {{35178, 1, 0, 0},{35178, 1, 1, 19}, {0, 0, 0, 0}}},{24167, {{35178, 1, 0, 0},{35178, 1, 1, 19}, {0, 0, 0, 0}}},{24168, {{35178, 1, 0, 0},{35178, 1, 1, 21}, {0, 0, 0, 0}}},{24168, {{35178, 1, 0, 0},{35178, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{24169, {{35179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24170, {{35180, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24171, {{35181, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24172, {{35182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24173, {{35183, 1, 0, 0},{35183, 1, 1, 18}, {0, 0, 0, 0}}},{24173, {{35183, 1, 0, 0},{35183, 1, 1, 18}, {0, 0, 0, 0}}},{24174, {{35183, 1, 0, 0},{35183, 1, 1, 19}, {0, 0, 0, 0}}},{24174, {{35183, 1, 0, 0},{35183, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{24175, {{35184, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24176, {{35185, 1, 0, 0},{35185, 1, 1, 16}, {0, 0, 0, 0}}},{24176, {{35185, 1, 0, 0},{35185, 1, 1, 16}, {0, 0, 0, 0}}},{24177, {{35185, 1, 0, 0},{35185, 1, 1, 17}, {0, 0, 0, 0}}},{24177, {{35185, 1, 0, 0},{35185, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{24178, {{35186, 1, 0, 0},{35186, 1, 1, 17}, {0, 0, 0, 0}}},{24178, {{35186, 1, 0, 0},{35186, 1, 1, 17}, {0, 0, 0, 0}}},{24179, {{35186, 1, 0, 0},{35186, 1, 1, 18}, {0, 0, 0, 0}}},{24179, {{35186, 1, 0, 0},{35186, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24180, {{35187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24181, {{35188, 1, 0, 0},{35188, 1, 1, 17}, {0, 0, 0, 0}}},{24181, {{35188, 1, 0, 0},{35188, 1, 1, 17}, {0, 0, 0, 0}}},{24182, {{35188, 1, 0, 0},{35188, 1, 1, 18}, {0, 0, 0, 0}}},{24182, {{35188, 1, 0, 0},{35188, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24183, {{35189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24184, {{35190, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24185, {{35191, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24186, {{35192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24187, {{35193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24188, {{35194, 1, 0, 0},{35194, 1, 1, 17}, {0, 0, 0, 0}}},{24188, {{35194, 1, 0, 0},{35194, 1, 1, 17}, {0, 0, 0, 0}}},{24189, {{35194, 1, 0, 0},{35194, 1, 1, 18}, {0, 0, 0, 0}}},{24189, {{35194, 1, 0, 0},{35194, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24190, {{35195, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24191, {{35196, 1, 0, 0},{35196, 1, 1, 17}, {0, 0, 0, 0}}},{24191, {{35196, 1, 0, 0},{35196, 1, 1, 17}, {0, 0, 0, 0}}},{24192, {{35196, 1, 0, 0},{35196, 1, 1, 18}, {0, 0, 0, 0}}},{24192, {{35196, 1, 0, 0},{35196, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{24193, {{35197, 1, 0, 0},{35197, 1, 1, 17}, {0, 0, 0, 0}}},{24193, {{35197, 1, 0, 0},{35197, 1, 1, 17}, {0, 0, 0, 0}}},{24194, {{35197, 1, 0, 0},{35197, 1, 1, 18}, {0, 0, 0, 0}}},{24194, {{35197, 1, 0, 0},{35197, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24195, {{35198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24196, {{35199, 1, 0, 0},{35199, 1, 1, 17}, {0, 0, 0, 0}}},{24196, {{35199, 1, 0, 0},{35199, 1, 1, 17}, {0, 0, 0, 0}}},{24197, {{35199, 1, 0, 0},{35199, 1, 1, 18}, {0, 0, 0, 0}}},{24197, {{35199, 1, 0, 0},{35199, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24199, {{35200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24200, {{35201, 1, 0, 0},{35201, 1, 1, 19}, {0, 0, 0, 0}}},{24200, {{35201, 1, 0, 0},{35201, 1, 1, 19}, {0, 0, 0, 0}}},{24201, {{35201, 1, 0, 0},{35201, 1, 1, 18}, {0, 0, 0, 0}}},{24201, {{35201, 1, 0, 0},{35201, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24202, {{35202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{24203, {{35203, 1, 0, 0},{35203, 1, 1, 17}, {0, 0, 0, 0}}},{24203, {{35203, 1, 0, 0},{35203, 1, 1, 17}, {0, 0, 0, 0}}},{24204, {{35203, 1, 0, 0},{35203, 1, 1, 18}, {0, 0, 0, 0}}},{24204, {{35203, 1, 0, 0},{35203, 1, 1, 18}, {0, 0, 0, 0}}},{58715, {{35203, 1, 0, 0},{35203, 1, 1, 19}, {0, 0, 0, 0}}},{58715, {{35203, 1, 0, 0},{35203, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{24205, {{35205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{24206, {{35206, 1, 0, 0},{35206, 1, 1, 19}, {0, 0, 0, 0}}},{24206, {{35206, 1, 0, 0},{35206, 1, 1, 19}, {0, 0, 0, 0}}},{24207, {{35206, 1, 0, 0},{35206, 1, 1, 18}, {0, 0, 0, 0}}},{24207, {{35206, 1, 0, 0},{35206, 1, 1, 18}, {0, 0, 0, 0}}},{24208, {{35206, 1, 0, 0},{35206, 1, 1, 20}, {0, 0, 0, 0}}},{24208, {{35206, 1, 0, 0},{35206, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{24209, {{35207, 1, 0, 0},{35207, 1, 1, 19}, {0, 0, 0, 0}}},{24209, {{35207, 1, 0, 0},{35207, 1, 1, 19}, {0, 0, 0, 0}}},{24210, {{35207, 1, 0, 0},{35207, 1, 1, 18}, {0, 0, 0, 0}}},{24210, {{35207, 1, 0, 0},{35207, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{24211, {{35208, 1, 0, 0},{35208, 1, 1, 17}, {0, 0, 0, 0}}},{24211, {{35208, 1, 0, 0},{35208, 1, 1, 17}, {0, 0, 0, 0}}},{24212, {{35208, 1, 0, 0},{35208, 1, 1, 18}, {0, 0, 0, 0}}},{24212, {{35208, 1, 0, 0},{35208, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{24213, {{35209, 1, 0, 0},{35209, 1, 1, 17}, {0, 0, 0, 0}}},{24213, {{35209, 1, 0, 0},{35209, 1, 1, 17}, {0, 0, 0, 0}}},{24214, {{35209, 1, 0, 0},{35209, 1, 1, 18}, {0, 0, 0, 0}}},{24214, {{35209, 1, 0, 0},{35209, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{24215, {{35210, 1, 0, 0},{35210, 1, 1, 17}, {0, 0, 0, 0}}},{24215, {{35210, 1, 0, 0},{35210, 1, 1, 17}, {0, 0, 0, 0}}},{24216, {{35210, 1, 0, 0},{35210, 1, 1, 18}, {0, 0, 0, 0}}},{24216, {{35210, 1, 0, 0},{35210, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24217, {{35211, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24218, {{35212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24219, {{35213, 1, 0, 0},{35213, 1, 1, 17}, {0, 0, 0, 0}}},{24219, {{35213, 1, 0, 0},{35213, 1, 1, 17}, {0, 0, 0, 0}}},{24220, {{35213, 1, 0, 0},{35213, 1, 1, 18}, {0, 0, 0, 0}}},{24220, {{35213, 1, 0, 0},{35213, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24221, {{35214, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24222, {{35215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{24223, {{35216, 1, 0, 0},{35216, 1, 1, 17}, {0, 0, 0, 0}}},{24223, {{35216, 1, 0, 0},{35216, 1, 1, 17}, {0, 0, 0, 0}}},{24224, {{35216, 1, 0, 0},{35216, 1, 1, 18}, {0, 0, 0, 0}}},{24224, {{35216, 1, 0, 0},{35216, 1, 1, 18}, {0, 0, 0, 0}}},{68066, {{35216, 1, 0, 0},{35216, 1, 1, 19}, {0, 0, 0, 0}}},{68066, {{35216, 1, 0, 0},{35216, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{24225, {{35217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24226, {{35218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24227, {{35219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24228, {{35220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24229, {{35221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{24230, {{35222, 1, 0, 0},{35222, 1, 1, 18}, {0, 0, 0, 0}}},{24230, {{35222, 1, 0, 0},{35222, 1, 1, 18}, {0, 0, 0, 0}}},{30296, {{35222, 1, 0, 0},{35222, 1, 1, 0},{35222, 1, 1, 19},{64097, 1, 0, 0}, {0, 0, 0, 0}}},{30296, {{35222, 1, 0, 0},{35222, 1, 1, 0},{35222, 1, 1, 19},{64097, 1, 0, 0}, {0, 0, 0, 0}}},{30296, {{35222, 1, 0, 0},{35222, 1, 1, 0},{35222, 1, 1, 19},{64097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24231, {{35223, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24232, {{35224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24233, {{35225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24234, {{35226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24235, {{35227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24236, {{35228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24237, {{35229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24238, {{35230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24239, {{35231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24240, {{35232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24241, {{35233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24242, {{35234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24243, {{35235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24244, {{35236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24245, {{35237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24246, {{35238, 1, 0, 0},{35238, 1, 1, 17}, {0, 0, 0, 0}}},{24246, {{35238, 1, 0, 0},{35238, 1, 1, 17}, {0, 0, 0, 0}}},{24247, {{35238, 1, 0, 0},{35238, 1, 1, 18}, {0, 0, 0, 0}}},{24247, {{35238, 1, 0, 0},{35238, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24248, {{35239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24249, {{35240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24250, {{35241, 1, 0, 0},{35241, 1, 1, 17}, {0, 0, 0, 0}}},{24250, {{35241, 1, 0, 0},{35241, 1, 1, 17}, {0, 0, 0, 0}}},{24251, {{35241, 1, 0, 0},{35241, 1, 1, 18}, {0, 0, 0, 0}}},{24251, {{35241, 1, 0, 0},{35241, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24252, {{35242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24253, {{35243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24254, {{35244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24255, {{35245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24256, {{35246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24257, {{35247, 1, 0, 0},{35247, 1, 1, 19}, {0, 0, 0, 0}}},{24257, {{35247, 1, 0, 0},{35247, 1, 1, 19}, {0, 0, 0, 0}}},{24258, {{35247, 1, 0, 0},{35247, 1, 1, 18}, {0, 0, 0, 0}}},{24258, {{35247, 1, 0, 0},{35247, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24259, {{35248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24260, {{35249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24261, {{35250, 1, 0, 0},{35250, 1, 1, 17}, {0, 0, 0, 0}}},{24261, {{35250, 1, 0, 0},{35250, 1, 1, 17}, {0, 0, 0, 0}}},{24262, {{35250, 1, 0, 0},{35250, 1, 1, 18}, {0, 0, 0, 0}}},{24262, {{35250, 1, 0, 0},{35250, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24263, {{35251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24264, {{35252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24265, {{35253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24266, {{35254, 1, 0, 0},{35254, 1, 1, 17}, {0, 0, 0, 0}}},{24266, {{35254, 1, 0, 0},{35254, 1, 1, 17}, {0, 0, 0, 0}}},{24267, {{35254, 1, 0, 0},{35254, 1, 1, 18}, {0, 0, 0, 0}}},{24267, {{35254, 1, 0, 0},{35254, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24269, {{35255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24270, {{35256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24271, {{35257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24272, {{35258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24273, {{35259, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24274, {{35260, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24275, {{35261, 1, 0, 0},{35261, 1, 1, 18}, {0, 0, 0, 0}}},{24275, {{35261, 1, 0, 0},{35261, 1, 1, 18}, {0, 0, 0, 0}}},{60178, {{35261, 1, 0, 0},{35261, 1, 1, 19}, {0, 0, 0, 0}}},{60178, {{35261, 1, 0, 0},{35261, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{24276, {{35262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24277, {{35263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{24278, {{35264, 1, 0, 0},{35264, 1, 1, 17}, {0, 0, 0, 0}}},{24278, {{35264, 1, 0, 0},{35264, 1, 1, 17}, {0, 0, 0, 0}}},{24279, {{35264, 1, 0, 0},{35264, 1, 1, 18}, {0, 0, 0, 0}}},{24279, {{35264, 1, 0, 0},{35264, 1, 1, 18}, {0, 0, 0, 0}}},{24280, {{35264, 1, 0, 0},{35264, 1, 1, 19}, {0, 0, 0, 0}}},{24280, {{35264, 1, 0, 0},{35264, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{24281, {{35282, 1, 0, 0},{35282, 1, 1, 18}, {0, 0, 0, 0}}},{24281, {{35282, 1, 0, 0},{35282, 1, 1, 18}, {0, 0, 0, 0}}},{24282, {{35282, 1, 0, 0},{35282, 1, 1, 20}, {0, 0, 0, 0}}},{24282, {{35282, 1, 0, 0},{35282, 1, 1, 20}, {0, 0, 0, 0}}},{24283, {{35282, 1, 0, 0},{35282, 1, 1, 19}, {0, 0, 0, 0}}},{24283, {{35282, 1, 0, 0},{35282, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{24284, {{35283, 1, 0, 0},{35283, 1, 1, 16}, {0, 0, 0, 0}}},{24284, {{35283, 1, 0, 0},{35283, 1, 1, 16}, {0, 0, 0, 0}}},{24285, {{35283, 1, 0, 0},{35283, 1, 1, 17}, {0, 0, 0, 0}}},{24285, {{35283, 1, 0, 0},{35283, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{24286, {{35284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24287, {{35285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24288, {{35286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24289, {{35287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24290, {{35288, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24291, {{35289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24292, {{35290, 1, 0, 0},{35290, 1, 1, 18}, {0, 0, 0, 0}}},{24292, {{35290, 1, 0, 0},{35290, 1, 1, 18}, {0, 0, 0, 0}}},{24293, {{35290, 1, 0, 0},{35290, 1, 1, 17}, {0, 0, 0, 0}}},{24293, {{35290, 1, 0, 0},{35290, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{24294, {{35291, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24295, {{35292, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24296, {{35293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{24297, {{35295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24298, {{35296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24299, {{35297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24300, {{35298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24302, {{35299, 1, 0, 0},{35299, 1, 1, 17}, {0, 0, 0, 0}}},{24302, {{35299, 1, 0, 0},{35299, 1, 1, 17}, {0, 0, 0, 0}}},{58737, {{35299, 1, 0, 0},{35299, 1, 1, 19}, {0, 0, 0, 0}}},{58737, {{35299, 1, 0, 0},{35299, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{24304, {{35300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24305, {{35301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24306, {{35302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24307, {{35303, 1, 0, 0},{35303, 1, 1, 17}, {0, 0, 0, 0}}},{24307, {{35303, 1, 0, 0},{35303, 1, 1, 17}, {0, 0, 0, 0}}},{58736, {{35303, 1, 0, 0},{35303, 1, 1, 18}, {0, 0, 0, 0}}},{58736, {{35303, 1, 0, 0},{35303, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24308, {{35304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24309, {{35305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24310, {{35306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24311, {{35307, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24312, {{35308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24313, {{35309, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24314, {{35310, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{24315, {{35312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24316, {{35313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24317, {{35314, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24318, {{35315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24319, {{35316, 1, 0, 0},{35316, 1, 1, 17}, {0, 0, 0, 0}}},{24319, {{35316, 1, 0, 0},{35316, 1, 1, 17}, {0, 0, 0, 0}}},{24320, {{35316, 1, 0, 0},{35316, 1, 1, 18}, {0, 0, 0, 0}}},{24320, {{35316, 1, 0, 0},{35316, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24321, {{35317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24322, {{35318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24323, {{35319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24324, {{35320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24325, {{35321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24326, {{35322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24327, {{35323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{24328, {{35325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24329, {{35326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24330, {{35327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24331, {{35328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{24332, {{35330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24333, {{35331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24334, {{35332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24335, {{35333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24336, {{35334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24337, {{35335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24338, {{35336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24339, {{35337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24340, {{35338, 1, 0, 0},{35338, 1, 1, 19}, {0, 0, 0, 0}}},{24340, {{35338, 1, 0, 0},{35338, 1, 1, 19}, {0, 0, 0, 0}}},{24341, {{35338, 1, 0, 0},{35338, 1, 1, 20}, {0, 0, 0, 0}}},{24341, {{35338, 1, 0, 0},{35338, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{24342, {{35339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24343, {{35340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24344, {{35341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24345, {{35342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24346, {{35343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24347, {{35344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24348, {{35345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24349, {{35346, 1, 0, 0},{35346, 1, 1, 18}, {0, 0, 0, 0}}},{24349, {{35346, 1, 0, 0},{35346, 1, 1, 18}, {0, 0, 0, 0}}},{24350, {{35346, 1, 0, 0},{35346, 1, 1, 19}, {0, 0, 0, 0}}},{24350, {{35346, 1, 0, 0},{35346, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{24352, {{35347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24353, {{35348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24354, {{35349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24355, {{35350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24356, {{35351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24357, {{35352, 1, 0, 0},{35352, 1, 1, 17}, {0, 0, 0, 0}}},{24357, {{35352, 1, 0, 0},{35352, 1, 1, 17}, {0, 0, 0, 0}}},{24358, {{35352, 1, 0, 0},{35352, 1, 1, 18}, {0, 0, 0, 0}}},{24358, {{35352, 1, 0, 0},{35352, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24359, {{35353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{24360, {{35355, 1, 0, 0},{35355, 1, 1, 17}, {0, 0, 0, 0}}},{24360, {{35355, 1, 0, 0},{35355, 1, 1, 17}, {0, 0, 0, 0}}},{24361, {{35355, 1, 0, 0},{35355, 1, 1, 18}, {0, 0, 0, 0}}},{24361, {{35355, 1, 0, 0},{35355, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24362, {{35356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{24363, {{35357, 1, 0, 0},{35357, 1, 1, 20}, {0, 0, 0, 0}}},{24363, {{35357, 1, 0, 0},{35357, 1, 1, 20}, {0, 0, 0, 0}}},{24364, {{35357, 1, 0, 0},{35357, 1, 1, 19}, {0, 0, 0, 0}}},{24364, {{35357, 1, 0, 0},{35357, 1, 1, 19}, {0, 0, 0, 0}}},{24365, {{35357, 1, 0, 0},{35357, 1, 1, 21}, {0, 0, 0, 0}}},{24365, {{35357, 1, 0, 0},{35357, 1, 1, 21}, {0, 0, 0, 0}}},{24366, {{35357, 1, 0, 0},{35357, 1, 1, 22}, {0, 0, 0, 0}}},{24366, {{35357, 1, 0, 0},{35357, 1, 1, 22}, {0, 0, 0, 0}}}}},
    {1, {{24367, {{35358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24368, {{35359, 1, 0, 0},{35359, 1, 1, 18}, {0, 0, 0, 0}}},{24368, {{35359, 1, 0, 0},{35359, 1, 1, 18}, {0, 0, 0, 0}}},{24369, {{35359, 1, 0, 0},{35359, 1, 1, 19}, {0, 0, 0, 0}}},{24369, {{35359, 1, 0, 0},{35359, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{24370, {{35360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24371, {{35361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24372, {{35362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24373, {{35363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24374, {{35364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24375, {{35365, 1, 0, 0},{35365, 1, 1, 17}, {0, 0, 0, 0}}},{24375, {{35365, 1, 0, 0},{35365, 1, 1, 17}, {0, 0, 0, 0}}},{24376, {{35365, 1, 0, 0},{35365, 1, 1, 18}, {0, 0, 0, 0}}},{24376, {{35365, 1, 0, 0},{35365, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24377, {{35366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24378, {{35367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24379, {{35368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24380, {{35369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24381, {{35370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24382, {{35371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24383, {{35372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24384, {{35373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24385, {{35374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24386, {{35375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24387, {{35376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24389, {{35377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24390, {{35378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24391, {{35379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24392, {{35380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24393, {{35381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24394, {{35382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24395, {{35383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24396, {{35384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24397, {{35385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24398, {{35386, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{24399, {{35387, 1, 0, 0},{35387, 1, 1, 18}, {0, 0, 0, 0}}},{24399, {{35387, 1, 0, 0},{35387, 1, 1, 18}, {0, 0, 0, 0}}},{24400, {{35387, 1, 0, 0},{35387, 1, 1, 19}, {0, 0, 0, 0}}},{24400, {{35387, 1, 0, 0},{35387, 1, 1, 19}, {0, 0, 0, 0}}},{24401, {{35387, 1, 0, 0},{35387, 1, 1, 20}, {0, 0, 0, 0}}},{24401, {{35387, 1, 0, 0},{35387, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{24402, {{35388, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24403, {{35389, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24404, {{35390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24405, {{35391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24406, {{35392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24407, {{35393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24408, {{35394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24409, {{35395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24410, {{35396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24411, {{35397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24412, {{35398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24413, {{35399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24414, {{35400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24415, {{35401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24416, {{35402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24418, {{35403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24419, {{35404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24420, {{35405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24421, {{35406, 1, 0, 0},{35406, 1, 1, 17}, {0, 0, 0, 0}}},{24421, {{35406, 1, 0, 0},{35406, 1, 1, 17}, {0, 0, 0, 0}}},{24422, {{35406, 1, 0, 0},{35406, 1, 1, 18}, {0, 0, 0, 0}}},{24422, {{35406, 1, 0, 0},{35406, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24423, {{35407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24424, {{35408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24425, {{35409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24426, {{35410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24427, {{35411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24428, {{35412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24429, {{35413, 1, 0, 0},{35413, 1, 1, 19}, {0, 0, 0, 0}}},{24429, {{35413, 1, 0, 0},{35413, 1, 1, 19}, {0, 0, 0, 0}}},{24430, {{35413, 1, 0, 0},{35413, 1, 1, 18}, {0, 0, 0, 0}}},{24430, {{35413, 1, 0, 0},{35413, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24431, {{35414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24432, {{35415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24433, {{35416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24434, {{35417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24435, {{35418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24436, {{35419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24437, {{35420, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24438, {{35421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24439, {{35422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{24440, {{35424, 1, 0, 0},{35424, 1, 1, 17}, {0, 0, 0, 0}}},{24440, {{35424, 1, 0, 0},{35424, 1, 1, 17}, {0, 0, 0, 0}}},{24441, {{35424, 1, 0, 0},{35424, 1, 1, 18}, {0, 0, 0, 0}}},{24441, {{35424, 1, 0, 0},{35424, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24442, {{35425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24443, {{35426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24444, {{35427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{24445, {{35428, 1, 0, 0},{35428, 1, 1, 16}, {0, 0, 0, 0}}},{24445, {{35428, 1, 0, 0},{35428, 1, 1, 16}, {0, 0, 0, 0}}},{24446, {{35428, 1, 0, 0},{35428, 1, 1, 17}, {0, 0, 0, 0}}},{24446, {{35428, 1, 0, 0},{35428, 1, 1, 17}, {0, 0, 0, 0}}},{24447, {{35428, 1, 0, 0},{35428, 1, 1, 18}, {0, 0, 0, 0}}},{24447, {{35428, 1, 0, 0},{35428, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24448, {{35429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24449, {{35430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24450, {{35431, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24451, {{35432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24452, {{35433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24453, {{35434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24454, {{35435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24455, {{35436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24456, {{35437, 1, 0, 0},{35437, 1, 1, 17}, {0, 0, 0, 0}}},{24456, {{35437, 1, 0, 0},{35437, 1, 1, 17}, {0, 0, 0, 0}}},{24457, {{35437, 1, 0, 0},{35437, 1, 1, 18}, {0, 0, 0, 0}}},{24457, {{35437, 1, 0, 0},{35437, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{24458, {{35438, 1, 0, 0},{35438, 1, 1, 18}, {0, 0, 0, 0}}},{24458, {{35438, 1, 0, 0},{35438, 1, 1, 18}, {0, 0, 0, 0}}},{24459, {{35438, 1, 0, 0},{35438, 1, 1, 19}, {0, 0, 0, 0}}},{24459, {{35438, 1, 0, 0},{35438, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{24460, {{35439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24461, {{35440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24462, {{35441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24463, {{35442, 1, 0, 0},{35442, 1, 1, 17}, {0, 0, 0, 0}}},{24463, {{35442, 1, 0, 0},{35442, 1, 1, 17}, {0, 0, 0, 0}}},{24464, {{35442, 1, 0, 0},{35442, 1, 1, 18}, {0, 0, 0, 0}}},{24464, {{35442, 1, 0, 0},{35442, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24465, {{35443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24466, {{35444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24467, {{35445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24468, {{35446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24469, {{35447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24470, {{35448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24471, {{35449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24472, {{35450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24473, {{35451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24474, {{35452, 1, 0, 0},{35452, 1, 1, 17}, {0, 0, 0, 0}}},{24474, {{35452, 1, 0, 0},{35452, 1, 1, 17}, {0, 0, 0, 0}}},{24475, {{35452, 1, 0, 0},{35452, 1, 1, 18}, {0, 0, 0, 0}}},{24475, {{35452, 1, 0, 0},{35452, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24476, {{35453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24477, {{35454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24478, {{35455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24479, {{35456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24480, {{35457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24481, {{35458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24482, {{35459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24483, {{35460, 1, 0, 0},{35460, 1, 1, 17}, {0, 0, 0, 0}}},{24483, {{35460, 1, 0, 0},{35460, 1, 1, 17}, {0, 0, 0, 0}}},{24484, {{35460, 1, 0, 0},{35460, 1, 1, 18}, {0, 0, 0, 0}}},{24484, {{35460, 1, 0, 0},{35460, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24485, {{35461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24486, {{35462, 1, 0, 0},{35462, 1, 1, 17}, {0, 0, 0, 0}}},{24486, {{35462, 1, 0, 0},{35462, 1, 1, 17}, {0, 0, 0, 0}}},{24487, {{35462, 1, 0, 0},{35462, 1, 1, 18}, {0, 0, 0, 0}}},{24487, {{35462, 1, 0, 0},{35462, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24488, {{35463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24489, {{35464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24490, {{35465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{24491, {{35467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24492, {{35468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24493, {{35469, 1, 0, 0},{35469, 1, 1, 19}, {0, 0, 0, 0}}},{24493, {{35469, 1, 0, 0},{35469, 1, 1, 19}, {0, 0, 0, 0}}},{24494, {{35469, 1, 0, 0},{35469, 1, 1, 18}, {0, 0, 0, 0}}},{24494, {{35469, 1, 0, 0},{35469, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24495, {{35470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24496, {{35471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24497, {{35472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24498, {{35473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24499, {{35474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24500, {{35475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24501, {{35476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{24502, {{35477, 1, 0, 0},{35477, 1, 1, 20}, {0, 0, 0, 0}}},{24502, {{35477, 1, 0, 0},{35477, 1, 1, 20}, {0, 0, 0, 0}}},{24503, {{35477, 1, 0, 0},{35477, 1, 1, 19}, {0, 0, 0, 0}}},{24503, {{35477, 1, 0, 0},{35477, 1, 1, 19}, {0, 0, 0, 0}}},{24504, {{35477, 1, 0, 0},{35477, 1, 1, 21}, {0, 0, 0, 0}}},{24504, {{35477, 1, 0, 0},{35477, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{24505, {{35478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24506, {{35479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24507, {{35480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24508, {{35481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24510, {{35482, 1, 0, 0},{35482, 1, 1, 17}, {0, 0, 0, 0}}},{24510, {{35482, 1, 0, 0},{35482, 1, 1, 17}, {0, 0, 0, 0}}},{24511, {{35482, 1, 0, 0},{35482, 1, 1, 18}, {0, 0, 0, 0}}},{24511, {{35482, 1, 0, 0},{35482, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24512, {{35483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24513, {{35484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24514, {{35485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24515, {{35486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24516, {{35487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24517, {{35488, 1, 0, 0},{35488, 1, 1, 18}, {0, 0, 0, 0}}},{24517, {{35488, 1, 0, 0},{35488, 1, 1, 18}, {0, 0, 0, 0}}},{24518, {{35488, 1, 0, 0},{35488, 1, 1, 19}, {0, 0, 0, 0}}},{24518, {{35488, 1, 0, 0},{35488, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{24519, {{35489, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24520, {{35490, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24521, {{35491, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{24522, {{35492, 1, 0, 0},{35492, 1, 1, 20}, {0, 0, 0, 0}}},{24522, {{35492, 1, 0, 0},{35492, 1, 1, 20}, {0, 0, 0, 0}}},{24523, {{35492, 1, 0, 0},{35492, 1, 1, 18}, {0, 0, 0, 0}}},{24523, {{35492, 1, 0, 0},{35492, 1, 1, 18}, {0, 0, 0, 0}}},{24524, {{35492, 1, 0, 0},{35492, 1, 1, 19}, {0, 0, 0, 0}}},{24524, {{35492, 1, 0, 0},{35492, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {6, {{24525, {{35493, 1, 0, 0},{35493, 1, 1, 17}, {0, 0, 0, 0}}},{24525, {{35493, 1, 0, 0},{35493, 1, 1, 17}, {0, 0, 0, 0}}},{24526, {{35493, 1, 0, 0},{35493, 1, 1, 19}, {0, 0, 0, 0}}},{24526, {{35493, 1, 0, 0},{35493, 1, 1, 19}, {0, 0, 0, 0}}},{24527, {{35493, 1, 0, 0},{35493, 1, 1, 18}, {0, 0, 0, 0}}},{24527, {{35493, 1, 0, 0},{35493, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24528, {{35494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24529, {{35495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24530, {{35496, 1, 0, 0},{35496, 1, 1, 17}, {0, 0, 0, 0}}},{24530, {{35496, 1, 0, 0},{35496, 1, 1, 17}, {0, 0, 0, 0}}},{24531, {{35496, 1, 0, 0},{35496, 1, 1, 18}, {0, 0, 0, 0}}},{24531, {{35496, 1, 0, 0},{35496, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24532, {{35497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24533, {{35498, 1, 0, 0},{35498, 1, 1, 17}, {0, 0, 0, 0}}},{24533, {{35498, 1, 0, 0},{35498, 1, 1, 17}, {0, 0, 0, 0}}},{58743, {{35498, 1, 0, 0},{35498, 1, 1, 18}, {0, 0, 0, 0}}},{58743, {{35498, 1, 0, 0},{35498, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24534, {{35499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24535, {{35500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24536, {{35501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24537, {{35502, 1, 0, 0},{35502, 1, 1, 17}, {0, 0, 0, 0}}},{24537, {{35502, 1, 0, 0},{35502, 1, 1, 17}, {0, 0, 0, 0}}},{24538, {{35502, 1, 0, 0},{35502, 1, 1, 18}, {0, 0, 0, 0}}},{24538, {{35502, 1, 0, 0},{35502, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24539, {{35503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24540, {{35504, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24541, {{35505, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24542, {{35506, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24543, {{35507, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24544, {{35508, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24545, {{35509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24546, {{35510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24547, {{35511, 1, 0, 0},{35511, 1, 1, 17}, {0, 0, 0, 0}}},{24547, {{35511, 1, 0, 0},{35511, 1, 1, 17}, {0, 0, 0, 0}}},{24548, {{35511, 1, 0, 0},{35511, 1, 1, 18}, {0, 0, 0, 0}}},{24548, {{35511, 1, 0, 0},{35511, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24549, {{35512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24550, {{35513, 1, 0, 0},{35513, 1, 1, 18}, {0, 0, 0, 0}}},{24550, {{35513, 1, 0, 0},{35513, 1, 1, 18}, {0, 0, 0, 0}}},{24551, {{35513, 1, 0, 0},{35513, 1, 1, 19}, {0, 0, 0, 0}}},{24551, {{35513, 1, 0, 0},{35513, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{24552, {{35514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24553, {{35515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24554, {{35516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24555, {{35517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24556, {{35518, 1, 0, 0},{35518, 1, 1, 17}, {0, 0, 0, 0}}},{24556, {{35518, 1, 0, 0},{35518, 1, 1, 17}, {0, 0, 0, 0}}},{24557, {{35518, 1, 0, 0},{35518, 1, 1, 18}, {0, 0, 0, 0}}},{24557, {{35518, 1, 0, 0},{35518, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{24558, {{35519, 1, 0, 0},{35519, 1, 1, 19}, {0, 0, 0, 0}}},{24558, {{35519, 1, 0, 0},{35519, 1, 1, 19}, {0, 0, 0, 0}}},{24559, {{35519, 1, 0, 0},{35519, 1, 1, 18}, {0, 0, 0, 0}}},{24559, {{35519, 1, 0, 0},{35519, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24560, {{35520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24561, {{35521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24562, {{35522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24563, {{35523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24564, {{35524, 1, 0, 0},{35524, 1, 1, 17}, {0, 0, 0, 0}}},{24564, {{35524, 1, 0, 0},{35524, 1, 1, 17}, {0, 0, 0, 0}}},{68089, {{35524, 1, 0, 0},{35524, 1, 1, 18}, {0, 0, 0, 0}}},{68089, {{35524, 1, 0, 0},{35524, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24565, {{35525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24566, {{35526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24567, {{35527, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24568, {{35528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24569, {{35529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24570, {{35530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24571, {{35531, 1, 0, 0},{35531, 1, 1, 19}, {0, 0, 0, 0}}},{24571, {{35531, 1, 0, 0},{35531, 1, 1, 19}, {0, 0, 0, 0}}},{24572, {{35531, 1, 0, 0},{35531, 1, 1, 18}, {0, 0, 0, 0}}},{24572, {{35531, 1, 0, 0},{35531, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24573, {{35532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24574, {{35533, 1, 0, 0},{35533, 1, 1, 17}, {0, 0, 0, 0}}},{24574, {{35533, 1, 0, 0},{35533, 1, 1, 17}, {0, 0, 0, 0}}},{24575, {{35533, 1, 0, 0},{35533, 1, 1, 18}, {0, 0, 0, 0}}},{24575, {{35533, 1, 0, 0},{35533, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24576, {{35534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24577, {{35535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24578, {{35536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24579, {{35537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24580, {{35538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24582, {{35539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24583, {{35540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24584, {{35541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24586, {{35542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24588, {{35543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24590, {{35544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24591, {{35545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24592, {{35546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{24593, {{35547, 1, 0, 0},{35547, 1, 1, 18}, {0, 0, 0, 0}}},{24593, {{35547, 1, 0, 0},{35547, 1, 1, 18}, {0, 0, 0, 0}}},{24594, {{35547, 1, 0, 0},{35547, 1, 1, 20}, {0, 0, 0, 0}}},{24594, {{35547, 1, 0, 0},{35547, 1, 1, 20}, {0, 0, 0, 0}}},{24595, {{35547, 1, 0, 0},{35547, 1, 1, 19}, {0, 0, 0, 0}}},{24595, {{35547, 1, 0, 0},{35547, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{24596, {{35548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24597, {{35549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24598, {{35550, 1, 0, 0},{35550, 1, 1, 19}, {0, 0, 0, 0}}},{24598, {{35550, 1, 0, 0},{35550, 1, 1, 19}, {0, 0, 0, 0}}},{24599, {{35550, 1, 0, 0},{35550, 1, 1, 18}, {0, 0, 0, 0}}},{24599, {{35550, 1, 0, 0},{35550, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24600, {{35551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24601, {{35552, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24603, {{35553, 1, 0, 0},{35553, 1, 1, 17}, {0, 0, 0, 0}}},{24603, {{35553, 1, 0, 0},{35553, 1, 1, 17}, {0, 0, 0, 0}}},{24604, {{35553, 1, 0, 0},{35553, 1, 1, 18}, {0, 0, 0, 0}}},{24604, {{35553, 1, 0, 0},{35553, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24605, {{35554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24606, {{35555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24607, {{35556, 1, 0, 0},{35556, 1, 1, 17}, {0, 0, 0, 0}}},{24607, {{35556, 1, 0, 0},{35556, 1, 1, 17}, {0, 0, 0, 0}}},{24608, {{35556, 1, 0, 0},{35556, 1, 1, 18}, {0, 0, 0, 0}}},{24608, {{35556, 1, 0, 0},{35556, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24609, {{35557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24610, {{35558, 1, 0, 0},{35558, 1, 1, 17}, {0, 0, 0, 0}}},{24610, {{35558, 1, 0, 0},{35558, 1, 1, 17}, {0, 0, 0, 0}}},{24611, {{35558, 1, 0, 0},{35558, 1, 1, 18}, {0, 0, 0, 0}}},{24611, {{35558, 1, 0, 0},{35558, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24612, {{35559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24613, {{35560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24614, {{35561, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24615, {{35562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24617, {{35563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24618, {{35564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{24619, {{35565, 1, 0, 0},{35565, 1, 1, 20}, {0, 0, 0, 0}}},{24619, {{35565, 1, 0, 0},{35565, 1, 1, 20}, {0, 0, 0, 0}}},{24620, {{35565, 1, 0, 0},{35565, 1, 1, 18}, {0, 0, 0, 0}}},{24620, {{35565, 1, 0, 0},{35565, 1, 1, 18}, {0, 0, 0, 0}}},{24621, {{35565, 1, 0, 0},{35565, 1, 1, 19}, {0, 0, 0, 0}}},{24621, {{35565, 1, 0, 0},{35565, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{24622, {{35566, 1, 0, 0},{35566, 1, 1, 20}, {0, 0, 0, 0}}},{24622, {{35566, 1, 0, 0},{35566, 1, 1, 20}, {0, 0, 0, 0}}},{24623, {{35566, 1, 0, 0},{35566, 1, 1, 19}, {0, 0, 0, 0}}},{24623, {{35566, 1, 0, 0},{35566, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{24624, {{35567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24625, {{35568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{24626, {{35569, 1, 0, 0},{35569, 1, 1, 18}, {0, 0, 0, 0}}},{24626, {{35569, 1, 0, 0},{35569, 1, 1, 18}, {0, 0, 0, 0}}},{24627, {{35569, 1, 0, 0},{35569, 1, 1, 17}, {0, 0, 0, 0}}},{24627, {{35569, 1, 0, 0},{35569, 1, 1, 17}, {0, 0, 0, 0}}},{24628, {{35569, 1, 0, 0},{35569, 1, 1, 19}, {0, 0, 0, 0}}},{24628, {{35569, 1, 0, 0},{35569, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{24629, {{35570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24631, {{35571, 1, 0, 0},{35571, 1, 1, 17}, {0, 0, 0, 0}}},{24631, {{35571, 1, 0, 0},{35571, 1, 1, 17}, {0, 0, 0, 0}}},{24632, {{35571, 1, 0, 0},{35571, 1, 1, 18}, {0, 0, 0, 0}}},{24632, {{35571, 1, 0, 0},{35571, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24633, {{35572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24634, {{35573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24635, {{35574, 1, 0, 0},{35574, 1, 1, 17}, {0, 0, 0, 0}}},{24635, {{35574, 1, 0, 0},{35574, 1, 1, 17}, {0, 0, 0, 0}}},{24636, {{35574, 1, 0, 0},{35574, 1, 1, 18}, {0, 0, 0, 0}}},{24636, {{35574, 1, 0, 0},{35574, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24637, {{35575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{24638, {{35576, 1, 0, 0},{35576, 1, 1, 18}, {0, 0, 0, 0}}},{24638, {{35576, 1, 0, 0},{35576, 1, 1, 18}, {0, 0, 0, 0}}},{30226, {{35576, 1, 0, 0},{35576, 1, 1, 0},{35576, 1, 1, 19},{64034, 1, 0, 0}, {0, 0, 0, 0}}},{30226, {{35576, 1, 0, 0},{35576, 1, 1, 0},{35576, 1, 1, 19},{64034, 1, 0, 0}, {0, 0, 0, 0}}},{30226, {{35576, 1, 0, 0},{35576, 1, 1, 0},{35576, 1, 1, 19},{64034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24639, {{35577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24640, {{35578, 1, 0, 0},{35578, 1, 1, 19}, {0, 0, 0, 0}}},{24640, {{35578, 1, 0, 0},{35578, 1, 1, 19}, {0, 0, 0, 0}}},{24641, {{35578, 1, 0, 0},{35578, 1, 1, 18}, {0, 0, 0, 0}}},{24641, {{35578, 1, 0, 0},{35578, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24642, {{35579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24643, {{35580, 1, 0, 0},{35580, 1, 1, 17}, {0, 0, 0, 0}}},{24643, {{35580, 1, 0, 0},{35580, 1, 1, 17}, {0, 0, 0, 0}}},{24644, {{35580, 1, 0, 0},{35580, 1, 1, 18}, {0, 0, 0, 0}}},{24644, {{35580, 1, 0, 0},{35580, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24645, {{35581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24646, {{35582, 1, 0, 0},{35582, 1, 1, 17}, {0, 0, 0, 0}}},{24646, {{35582, 1, 0, 0},{35582, 1, 1, 17}, {0, 0, 0, 0}}},{24647, {{35582, 1, 0, 0},{35582, 1, 1, 18}, {0, 0, 0, 0}}},{24647, {{35582, 1, 0, 0},{35582, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{24648, {{35583, 1, 0, 0},{35583, 1, 1, 17}, {0, 0, 0, 0}}},{24648, {{35583, 1, 0, 0},{35583, 1, 1, 17}, {0, 0, 0, 0}}},{24649, {{35583, 1, 0, 0},{35583, 1, 1, 18}, {0, 0, 0, 0}}},{24649, {{35583, 1, 0, 0},{35583, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24650, {{35584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{24651, {{35585, 1, 0, 0},{35585, 1, 1, 18}, {0, 0, 0, 0}}},{24651, {{35585, 1, 0, 0},{35585, 1, 1, 18}, {0, 0, 0, 0}}},{30297, {{35585, 1, 0, 0},{35585, 1, 1, 0},{35585, 1, 1, 19},{64098, 1, 0, 0}, {0, 0, 0, 0}}},{30297, {{35585, 1, 0, 0},{35585, 1, 1, 0},{35585, 1, 1, 19},{64098, 1, 0, 0}, {0, 0, 0, 0}}},{30297, {{35585, 1, 0, 0},{35585, 1, 1, 0},{35585, 1, 1, 19},{64098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24652, {{35586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{24653, {{35588, 1, 0, 0},{35588, 1, 1, 19}, {0, 0, 0, 0}}},{24653, {{35588, 1, 0, 0},{35588, 1, 1, 19}, {0, 0, 0, 0}}},{24654, {{35588, 1, 0, 0},{35588, 1, 1, 18}, {0, 0, 0, 0}}},{24654, {{35588, 1, 0, 0},{35588, 1, 1, 18}, {0, 0, 0, 0}}},{24655, {{35588, 1, 0, 0},{35588, 1, 1, 20}, {0, 0, 0, 0}}},{24655, {{35588, 1, 0, 0},{35588, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{24656, {{35589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24657, {{35590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24658, {{35591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24659, {{35592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24660, {{35593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{24661, {{35594, 1, 0, 0},{35594, 1, 1, 17}, {0, 0, 0, 0}}},{24661, {{35594, 1, 0, 0},{35594, 1, 1, 17}, {0, 0, 0, 0}}},{24662, {{35594, 1, 0, 0},{35594, 1, 1, 18}, {0, 0, 0, 0}}},{24662, {{35594, 1, 0, 0},{35594, 1, 1, 18}, {0, 0, 0, 0}}},{24663, {{35594, 1, 0, 0},{35594, 1, 1, 20}, {0, 0, 0, 0}}},{24663, {{35594, 1, 0, 0},{35594, 1, 1, 20}, {0, 0, 0, 0}}},{24664, {{35594, 1, 0, 0},{35594, 1, 1, 19}, {0, 0, 0, 0}}},{24664, {{35594, 1, 0, 0},{35594, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{24665, {{35595, 1, 0, 0},{35595, 1, 1, 18}, {0, 0, 0, 0}}},{24665, {{35595, 1, 0, 0},{35595, 1, 1, 18}, {0, 0, 0, 0}}},{24666, {{35595, 1, 0, 0},{35595, 1, 1, 17}, {0, 0, 0, 0}}},{24666, {{35595, 1, 0, 0},{35595, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{24667, {{35596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24668, {{35597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24669, {{35598, 1, 0, 0},{35598, 1, 1, 18}, {0, 0, 0, 0}}},{24669, {{35598, 1, 0, 0},{35598, 1, 1, 18}, {0, 0, 0, 0}}},{24670, {{35598, 1, 0, 0},{35598, 1, 1, 19}, {0, 0, 0, 0}}},{24670, {{35598, 1, 0, 0},{35598, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{24671, {{35599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24673, {{35600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24674, {{35601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24675, {{35602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24676, {{35603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24679, {{35604, 1, 0, 0},{35604, 1, 1, 17}, {0, 0, 0, 0}}},{24679, {{35604, 1, 0, 0},{35604, 1, 1, 17}, {0, 0, 0, 0}}},{24680, {{35604, 1, 0, 0},{35604, 1, 1, 18}, {0, 0, 0, 0}}},{24680, {{35604, 1, 0, 0},{35604, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24681, {{35605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24682, {{35606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24683, {{35607, 1, 0, 0},{35607, 1, 1, 17}, {0, 0, 0, 0}}},{24683, {{35607, 1, 0, 0},{35607, 1, 1, 17}, {0, 0, 0, 0}}},{24684, {{35607, 1, 0, 0},{35607, 1, 1, 18}, {0, 0, 0, 0}}},{24684, {{35607, 1, 0, 0},{35607, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24685, {{35608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24686, {{35609, 1, 0, 0},{35609, 1, 1, 19}, {0, 0, 0, 0}}},{24686, {{35609, 1, 0, 0},{35609, 1, 1, 19}, {0, 0, 0, 0}}},{24687, {{35609, 1, 0, 0},{35609, 1, 1, 18}, {0, 0, 0, 0}}},{24687, {{35609, 1, 0, 0},{35609, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{24690, {{35610, 1, 0, 0},{35610, 1, 1, 17}, {0, 0, 0, 0}}},{24690, {{35610, 1, 0, 0},{35610, 1, 1, 17}, {0, 0, 0, 0}}},{24691, {{35610, 1, 0, 0},{35610, 1, 1, 19}, {0, 0, 0, 0}}},{24691, {{35610, 1, 0, 0},{35610, 1, 1, 19}, {0, 0, 0, 0}}},{24692, {{35610, 1, 0, 0},{35610, 1, 1, 18}, {0, 0, 0, 0}}},{24692, {{35610, 1, 0, 0},{35610, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{24693, {{35611, 1, 0, 0},{35611, 1, 1, 19}, {0, 0, 0, 0}}},{24693, {{35611, 1, 0, 0},{35611, 1, 1, 19}, {0, 0, 0, 0}}},{24694, {{35611, 1, 0, 0},{35611, 1, 1, 18}, {0, 0, 0, 0}}},{24694, {{35611, 1, 0, 0},{35611, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24695, {{35612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24696, {{35613, 1, 0, 0},{35613, 1, 1, 18}, {0, 0, 0, 0}}},{24696, {{35613, 1, 0, 0},{35613, 1, 1, 18}, {0, 0, 0, 0}}},{24697, {{35613, 1, 0, 0},{35613, 1, 1, 19}, {0, 0, 0, 0}}},{24697, {{35613, 1, 0, 0},{35613, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{24698, {{35614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24699, {{35615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24700, {{35616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24701, {{35617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{24703, {{35619, 1, 0, 0},{35619, 1, 1, 16}, {0, 0, 0, 0}}},{24703, {{35619, 1, 0, 0},{35619, 1, 1, 16}, {0, 0, 0, 0}}},{58760, {{35619, 1, 0, 0},{35619, 1, 1, 17}, {0, 0, 0, 0}}},{58760, {{35619, 1, 0, 0},{35619, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{24704, {{35620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24705, {{35621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24706, {{35622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24707, {{35623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24708, {{35624, 1, 0, 0},{35624, 1, 1, 17}, {0, 0, 0, 0}}},{24708, {{35624, 1, 0, 0},{35624, 1, 1, 17}, {0, 0, 0, 0}}},{24709, {{35624, 1, 0, 0},{35624, 1, 1, 18}, {0, 0, 0, 0}}},{24709, {{35624, 1, 0, 0},{35624, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24710, {{35625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24711, {{35626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24712, {{35627, 1, 0, 0},{35627, 1, 1, 17}, {0, 0, 0, 0}}},{24712, {{35627, 1, 0, 0},{35627, 1, 1, 17}, {0, 0, 0, 0}}},{24713, {{35627, 1, 0, 0},{35627, 1, 1, 18}, {0, 0, 0, 0}}},{24713, {{35627, 1, 0, 0},{35627, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{24714, {{35628, 1, 0, 0},{35628, 1, 1, 19}, {0, 0, 0, 0}}},{24714, {{35628, 1, 0, 0},{35628, 1, 1, 19}, {0, 0, 0, 0}}},{24715, {{35628, 1, 0, 0},{35628, 1, 1, 18}, {0, 0, 0, 0}}},{24715, {{35628, 1, 0, 0},{35628, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24716, {{35629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24717, {{35630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24718, {{35631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24719, {{35632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24720, {{35633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24721, {{35634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{24722, {{35635, 1, 0, 0},{35635, 1, 1, 17}, {0, 0, 0, 0}}},{24722, {{35635, 1, 0, 0},{35635, 1, 1, 17}, {0, 0, 0, 0}}},{24723, {{35635, 1, 0, 0},{35635, 1, 1, 18}, {0, 0, 0, 0}}},{24723, {{35635, 1, 0, 0},{35635, 1, 1, 18}, {0, 0, 0, 0}}},{58752, {{35635, 1, 0, 0},{35635, 1, 1, 19}, {0, 0, 0, 0}}},{58752, {{35635, 1, 0, 0},{35635, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{24724, {{35636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24725, {{35637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24726, {{35638, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24727, {{35639, 1, 0, 0},{35639, 1, 1, 17}, {0, 0, 0, 0}}},{24727, {{35639, 1, 0, 0},{35639, 1, 1, 17}, {0, 0, 0, 0}}},{24728, {{35639, 1, 0, 0},{35639, 1, 1, 18}, {0, 0, 0, 0}}},{24728, {{35639, 1, 0, 0},{35639, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24729, {{35640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {9, {{24730, {{35641, 1, 0, 0},{35641, 1, 1, 18}, {0, 0, 0, 0}}},{24730, {{35641, 1, 0, 0},{35641, 1, 1, 18}, {0, 0, 0, 0}}},{24731, {{35641, 1, 0, 0},{35641, 1, 1, 19}, {0, 0, 0, 0}}},{24731, {{35641, 1, 0, 0},{35641, 1, 1, 19}, {0, 0, 0, 0}}},{30298, {{35641, 1, 0, 0},{35641, 1, 1, 0},{35641, 1, 1, 20},{64099, 1, 0, 0}, {0, 0, 0, 0}}},{30298, {{35641, 1, 0, 0},{35641, 1, 1, 0},{35641, 1, 1, 20},{64099, 1, 0, 0}, {0, 0, 0, 0}}},{30298, {{35641, 1, 0, 0},{35641, 1, 1, 0},{35641, 1, 1, 20},{64099, 1, 0, 0}, {0, 0, 0, 0}}},{60186, {{35641, 1, 0, 0},{35641, 1, 1, 21}, {0, 0, 0, 0}}},{60186, {{35641, 1, 0, 0},{35641, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{24732, {{35642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24733, {{35643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24734, {{35644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24735, {{35645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24736, {{35646, 1, 0, 0},{35646, 1, 1, 19}, {0, 0, 0, 0}}},{24736, {{35646, 1, 0, 0},{35646, 1, 1, 19}, {0, 0, 0, 0}}},{24737, {{35646, 1, 0, 0},{35646, 1, 1, 18}, {0, 0, 0, 0}}},{24737, {{35646, 1, 0, 0},{35646, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24738, {{35647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24739, {{35648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{24740, {{35649, 1, 0, 0},{35649, 1, 1, 18}, {0, 0, 0, 0}}},{24740, {{35649, 1, 0, 0},{35649, 1, 1, 18}, {0, 0, 0, 0}}},{24741, {{35649, 1, 0, 0},{35649, 1, 1, 20}, {0, 0, 0, 0}}},{24741, {{35649, 1, 0, 0},{35649, 1, 1, 20}, {0, 0, 0, 0}}},{24742, {{35649, 1, 0, 0},{35649, 1, 1, 19}, {0, 0, 0, 0}}},{24742, {{35649, 1, 0, 0},{35649, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{24743, {{35650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24744, {{35651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{24745, {{35652, 1, 0, 0},{35652, 1, 1, 17}, {0, 0, 0, 0}}},{24745, {{35652, 1, 0, 0},{35652, 1, 1, 17}, {0, 0, 0, 0}}},{24746, {{35652, 1, 0, 0},{35652, 1, 1, 18}, {0, 0, 0, 0}}},{24746, {{35652, 1, 0, 0},{35652, 1, 1, 18}, {0, 0, 0, 0}}},{24747, {{35652, 1, 0, 0},{35652, 1, 1, 19}, {0, 0, 0, 0}}},{24747, {{35652, 1, 0, 0},{35652, 1, 1, 19}, {0, 0, 0, 0}}},{68067, {{35652, 1, 0, 0},{35652, 1, 1, 22}, {0, 0, 0, 0}}},{68067, {{35652, 1, 0, 0},{35652, 1, 1, 22}, {0, 0, 0, 0}}}}},
    {4, {{24750, {{35653, 1, 0, 0},{35653, 1, 1, 17}, {0, 0, 0, 0}}},{24750, {{35653, 1, 0, 0},{35653, 1, 1, 17}, {0, 0, 0, 0}}},{24751, {{35653, 1, 0, 0},{35653, 1, 1, 18}, {0, 0, 0, 0}}},{24751, {{35653, 1, 0, 0},{35653, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24752, {{35654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24753, {{35655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24754, {{35656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24755, {{35657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24756, {{35658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24757, {{35659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{24758, {{35660, 1, 0, 0},{35660, 1, 1, 17}, {0, 0, 0, 0}}},{24758, {{35660, 1, 0, 0},{35660, 1, 1, 17}, {0, 0, 0, 0}}},{24759, {{35660, 1, 0, 0},{35660, 1, 1, 19}, {0, 0, 0, 0}}},{24759, {{35660, 1, 0, 0},{35660, 1, 1, 19}, {0, 0, 0, 0}}},{24760, {{35660, 1, 0, 0},{35660, 1, 1, 18}, {0, 0, 0, 0}}},{24760, {{35660, 1, 0, 0},{35660, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24761, {{35661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24762, {{35662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{24763, {{35663, 1, 0, 0},{35663, 1, 1, 18}, {0, 0, 0, 0}}},{24763, {{35663, 1, 0, 0},{35663, 1, 1, 18}, {0, 0, 0, 0}}},{24764, {{35663, 1, 0, 0},{35663, 1, 1, 19}, {0, 0, 0, 0}}},{24764, {{35663, 1, 0, 0},{35663, 1, 1, 19}, {0, 0, 0, 0}}},{24765, {{35663, 1, 0, 0},{35663, 1, 1, 17}, {0, 0, 0, 0}}},{24765, {{35663, 1, 0, 0},{35663, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{24766, {{35664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24767, {{35665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24768, {{35666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{24769, {{35667, 1, 0, 0},{35667, 1, 1, 17}, {0, 0, 0, 0}}},{24769, {{35667, 1, 0, 0},{35667, 1, 1, 17}, {0, 0, 0, 0}}},{24770, {{35667, 1, 0, 0},{35667, 1, 1, 18}, {0, 0, 0, 0}}},{24770, {{35667, 1, 0, 0},{35667, 1, 1, 18}, {0, 0, 0, 0}}},{60190, {{35667, 1, 0, 0},{35667, 1, 1, 19}, {0, 0, 0, 0}}},{60190, {{35667, 1, 0, 0},{35667, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{24771, {{35668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24772, {{35669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24773, {{35670, 1, 0, 0},{35670, 1, 1, 19}, {0, 0, 0, 0}}},{24773, {{35670, 1, 0, 0},{35670, 1, 1, 19}, {0, 0, 0, 0}}},{24774, {{35670, 1, 0, 0},{35670, 1, 1, 20}, {0, 0, 0, 0}}},{24774, {{35670, 1, 0, 0},{35670, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{24775, {{35671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24776, {{35672, 1, 0, 0},{35672, 1, 1, 17}, {0, 0, 0, 0}}},{24776, {{35672, 1, 0, 0},{35672, 1, 1, 17}, {0, 0, 0, 0}}},{24777, {{35672, 1, 0, 0},{35672, 1, 1, 18}, {0, 0, 0, 0}}},{24777, {{35672, 1, 0, 0},{35672, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24778, {{35673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24779, {{35674, 1, 0, 0},{35674, 1, 1, 19}, {0, 0, 0, 0}}},{24779, {{35674, 1, 0, 0},{35674, 1, 1, 19}, {0, 0, 0, 0}}},{24780, {{35674, 1, 0, 0},{35674, 1, 1, 18}, {0, 0, 0, 0}}},{24780, {{35674, 1, 0, 0},{35674, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24781, {{35675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24782, {{35676, 1, 0, 0},{35676, 1, 1, 18}, {0, 0, 0, 0}}},{24782, {{35676, 1, 0, 0},{35676, 1, 1, 18}, {0, 0, 0, 0}}},{24783, {{35676, 1, 0, 0},{35676, 1, 1, 19}, {0, 0, 0, 0}}},{24783, {{35676, 1, 0, 0},{35676, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{24784, {{35677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24785, {{35678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24786, {{35679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24787, {{35680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24788, {{35681, 1, 0, 0},{35681, 1, 1, 16}, {0, 0, 0, 0}}},{24788, {{35681, 1, 0, 0},{35681, 1, 1, 16}, {0, 0, 0, 0}}},{24789, {{35681, 1, 0, 0},{35681, 1, 1, 17}, {0, 0, 0, 0}}},{24789, {{35681, 1, 0, 0},{35681, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{24790, {{35682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24792, {{35683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24793, {{35684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24794, {{35685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24795, {{35686, 1, 0, 0},{35686, 1, 1, 17}, {0, 0, 0, 0}}},{24795, {{35686, 1, 0, 0},{35686, 1, 1, 17}, {0, 0, 0, 0}}},{24796, {{35686, 1, 0, 0},{35686, 1, 1, 18}, {0, 0, 0, 0}}},{24796, {{35686, 1, 0, 0},{35686, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24797, {{35687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24798, {{35688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24799, {{35689, 1, 0, 0},{35689, 1, 1, 17}, {0, 0, 0, 0}}},{24799, {{35689, 1, 0, 0},{35689, 1, 1, 17}, {0, 0, 0, 0}}},{24800, {{35689, 1, 0, 0},{35689, 1, 1, 18}, {0, 0, 0, 0}}},{24800, {{35689, 1, 0, 0},{35689, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{24801, {{35690, 1, 0, 0},{35690, 1, 1, 16}, {0, 0, 0, 0}}},{24801, {{35690, 1, 0, 0},{35690, 1, 1, 16}, {0, 0, 0, 0}}},{24802, {{35690, 1, 0, 0},{35690, 1, 1, 17}, {0, 0, 0, 0}}},{24802, {{35690, 1, 0, 0},{35690, 1, 1, 17}, {0, 0, 0, 0}}},{24803, {{35690, 1, 0, 0},{35690, 1, 1, 18}, {0, 0, 0, 0}}},{24803, {{35690, 1, 0, 0},{35690, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24804, {{35691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24805, {{35692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24806, {{35693, 1, 0, 0},{35693, 1, 1, 17}, {0, 0, 0, 0}}},{24806, {{35693, 1, 0, 0},{35693, 1, 1, 17}, {0, 0, 0, 0}}},{58766, {{35693, 1, 0, 0},{35693, 1, 1, 18}, {0, 0, 0, 0}}},{58766, {{35693, 1, 0, 0},{35693, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24807, {{35694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24809, {{35695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24810, {{35696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24811, {{35697, 1, 0, 0},{35697, 1, 1, 17}, {0, 0, 0, 0}}},{24811, {{35697, 1, 0, 0},{35697, 1, 1, 17}, {0, 0, 0, 0}}},{24812, {{35697, 1, 0, 0},{35697, 1, 1, 18}, {0, 0, 0, 0}}},{24812, {{35697, 1, 0, 0},{35697, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24813, {{35698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24814, {{35699, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24815, {{35700, 1, 0, 0},{35700, 1, 1, 18}, {0, 0, 0, 0}}},{24815, {{35700, 1, 0, 0},{35700, 1, 1, 18}, {0, 0, 0, 0}}},{24816, {{35700, 1, 0, 0},{35700, 1, 1, 17}, {0, 0, 0, 0}}},{24816, {{35700, 1, 0, 0},{35700, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{24817, {{35701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24818, {{35702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24819, {{35703, 1, 0, 0},{35703, 1, 1, 17}, {0, 0, 0, 0}}},{24819, {{35703, 1, 0, 0},{35703, 1, 1, 17}, {0, 0, 0, 0}}},{24820, {{35703, 1, 0, 0},{35703, 1, 1, 18}, {0, 0, 0, 0}}},{24820, {{35703, 1, 0, 0},{35703, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24821, {{35704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24822, {{35705, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24823, {{35706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24824, {{35707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24825, {{35708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24826, {{35709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24827, {{35710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24828, {{35711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24829, {{35712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{24830, {{35713, 1, 0, 0},{35713, 1, 1, 19}, {0, 0, 0, 0}}},{24830, {{35713, 1, 0, 0},{35713, 1, 1, 19}, {0, 0, 0, 0}}},{24831, {{35713, 1, 0, 0},{35713, 1, 1, 17}, {0, 0, 0, 0}}},{24831, {{35713, 1, 0, 0},{35713, 1, 1, 17}, {0, 0, 0, 0}}},{24832, {{35713, 1, 0, 0},{35713, 1, 1, 18}, {0, 0, 0, 0}}},{24832, {{35713, 1, 0, 0},{35713, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{24833, {{35715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24834, {{35716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24835, {{35717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{24836, {{35719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24837, {{35720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24838, {{35721, 1, 0, 0},{35721, 1, 1, 16}, {0, 0, 0, 0}}},{24838, {{35721, 1, 0, 0},{35721, 1, 1, 16}, {0, 0, 0, 0}}},{24839, {{35721, 1, 0, 0},{35721, 1, 1, 17}, {0, 0, 0, 0}}},{24839, {{35721, 1, 0, 0},{35721, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{24840, {{35722, 1, 0, 0},{35722, 1, 1, 18}, {0, 0, 0, 0}}},{24840, {{35722, 1, 0, 0},{35722, 1, 1, 18}, {0, 0, 0, 0}}},{24841, {{35722, 1, 0, 0},{35722, 1, 1, 17}, {0, 0, 0, 0}}},{24841, {{35722, 1, 0, 0},{35722, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{24842, {{35723, 1, 0, 0},{35723, 1, 1, 17}, {0, 0, 0, 0}}},{24842, {{35723, 1, 0, 0},{35723, 1, 1, 17}, {0, 0, 0, 0}}},{24843, {{35723, 1, 0, 0},{35723, 1, 1, 18}, {0, 0, 0, 0}}},{24843, {{35723, 1, 0, 0},{35723, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{24844, {{35724, 1, 0, 0},{35724, 1, 1, 17}, {0, 0, 0, 0}}},{24844, {{35724, 1, 0, 0},{35724, 1, 1, 17}, {0, 0, 0, 0}}},{24845, {{35724, 1, 0, 0},{35724, 1, 1, 18}, {0, 0, 0, 0}}},{24845, {{35724, 1, 0, 0},{35724, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24846, {{35725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24847, {{35726, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24848, {{35727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24849, {{35728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24850, {{35729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24851, {{35730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24852, {{35731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24853, {{35732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24854, {{35733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24855, {{35734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24856, {{35735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24857, {{35736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24858, {{35737, 1, 0, 0},{35737, 1, 1, 17}, {0, 0, 0, 0}}},{24858, {{35737, 1, 0, 0},{35737, 1, 1, 17}, {0, 0, 0, 0}}},{24859, {{35737, 1, 0, 0},{35737, 1, 1, 18}, {0, 0, 0, 0}}},{24859, {{35737, 1, 0, 0},{35737, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24860, {{35738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24861, {{35739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24862, {{35740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24863, {{35741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24864, {{35742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24865, {{35743, 1, 0, 0}, {0, 0, 0, 0}}}}}
};

static const MJMappingSet chunked_mj_mappings_008c37_008d1c[230] = {
    {1, {{24866, {{35895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24867, {{35896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24868, {{35897, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{24869, {{35898, 1, 0, 0},{35898, 1, 1, 18}, {0, 0, 0, 0}}},{24869, {{35898, 1, 0, 0},{35898, 1, 1, 18}, {0, 0, 0, 0}}},{24870, {{35898, 1, 0, 0},{35898, 1, 1, 17}, {0, 0, 0, 0}}},{24870, {{35898, 1, 0, 0},{35898, 1, 1, 17}, {0, 0, 0, 0}}},{24871, {{35898, 1, 0, 0},{35898, 1, 1, 19}, {0, 0, 0, 0}}},{24871, {{35898, 1, 0, 0},{35898, 1, 1, 19}, {0, 0, 0, 0}}},{24872, {{35898, 1, 0, 0},{35898, 1, 1, 20}, {0, 0, 0, 0}}},{24872, {{35898, 1, 0, 0},{35898, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{24873, {{35899, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24874, {{35900, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24875, {{35901, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24876, {{35902, 1, 0, 0},{35902, 1, 1, 17}, {0, 0, 0, 0}}},{24876, {{35902, 1, 0, 0},{35902, 1, 1, 17}, {0, 0, 0, 0}}},{24877, {{35902, 1, 0, 0},{35902, 1, 1, 18}, {0, 0, 0, 0}}},{24877, {{35902, 1, 0, 0},{35902, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{24878, {{35903, 1, 0, 0},{35903, 1, 1, 17}, {0, 0, 0, 0}}},{24878, {{35903, 1, 0, 0},{35903, 1, 1, 17}, {0, 0, 0, 0}}},{24879, {{35903, 1, 0, 0},{35903, 1, 1, 18}, {0, 0, 0, 0}}},{24879, {{35903, 1, 0, 0},{35903, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24880, {{35904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{24881, {{35905, 1, 0, 0},{35905, 1, 1, 18}, {0, 0, 0, 0}}},{24881, {{35905, 1, 0, 0},{35905, 1, 1, 18}, {0, 0, 0, 0}}},{24882, {{35905, 1, 0, 0},{35905, 1, 1, 17}, {0, 0, 0, 0}}},{24882, {{35905, 1, 0, 0},{35905, 1, 1, 17}, {0, 0, 0, 0}}},{24883, {{35905, 1, 0, 0},{35905, 1, 1, 19}, {0, 0, 0, 0}}},{24883, {{35905, 1, 0, 0},{35905, 1, 1, 19}, {0, 0, 0, 0}}},{24884, {{35905, 1, 0, 0},{35905, 1, 1, 20}, {0, 0, 0, 0}}},{24884, {{35905, 1, 0, 0},{35905, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{24885, {{35906, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24886, {{35907, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24887, {{35908, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24888, {{35909, 1, 0, 0},{35909, 1, 1, 17}, {0, 0, 0, 0}}},{24888, {{35909, 1, 0, 0},{35909, 1, 1, 17}, {0, 0, 0, 0}}},{24889, {{35909, 1, 0, 0},{35909, 1, 1, 18}, {0, 0, 0, 0}}},{24889, {{35909, 1, 0, 0},{35909, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24890, {{35910, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24891, {{35911, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24892, {{35912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24893, {{35913, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24894, {{35914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24895, {{35915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24896, {{35916, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24897, {{35917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24898, {{35918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24899, {{35919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24900, {{35920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24901, {{35921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{24902, {{35923, 1, 0, 0},{35923, 1, 1, 17}, {0, 0, 0, 0}}},{24902, {{35923, 1, 0, 0},{35923, 1, 1, 17}, {0, 0, 0, 0}}},{24903, {{35923, 1, 0, 0},{35923, 1, 1, 18}, {0, 0, 0, 0}}},{24903, {{35923, 1, 0, 0},{35923, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24904, {{35924, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{24905, {{35925, 1, 0, 0},{35925, 1, 1, 18}, {0, 0, 0, 0}}},{24905, {{35925, 1, 0, 0},{35925, 1, 1, 18}, {0, 0, 0, 0}}},{24906, {{35925, 1, 0, 0},{35925, 1, 1, 17}, {0, 0, 0, 0}}},{24906, {{35925, 1, 0, 0},{35925, 1, 1, 17}, {0, 0, 0, 0}}},{24907, {{35925, 1, 0, 0},{35925, 1, 1, 19}, {0, 0, 0, 0}}},{24907, {{35925, 1, 0, 0},{35925, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{24908, {{35926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24909, {{35927, 1, 0, 0},{35927, 1, 1, 17}, {0, 0, 0, 0}}},{24909, {{35927, 1, 0, 0},{35927, 1, 1, 17}, {0, 0, 0, 0}}},{24910, {{35927, 1, 0, 0},{35927, 1, 1, 18}, {0, 0, 0, 0}}},{24910, {{35927, 1, 0, 0},{35927, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24911, {{35928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24912, {{35929, 1, 0, 0},{35929, 1, 1, 17}, {0, 0, 0, 0}}},{24912, {{35929, 1, 0, 0},{35929, 1, 1, 17}, {0, 0, 0, 0}}},{24913, {{35929, 1, 0, 0},{35929, 1, 1, 18}, {0, 0, 0, 0}}},{24913, {{35929, 1, 0, 0},{35929, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{24914, {{35930, 1, 0, 0},{35930, 1, 1, 17}, {0, 0, 0, 0}}},{24914, {{35930, 1, 0, 0},{35930, 1, 1, 17}, {0, 0, 0, 0}}},{24915, {{35930, 1, 0, 0},{35930, 1, 1, 18}, {0, 0, 0, 0}}},{24915, {{35930, 1, 0, 0},{35930, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24916, {{35931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24917, {{35932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24918, {{35933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24919, {{35934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24920, {{35935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24921, {{35936, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{24922, {{35937, 1, 0, 0},{35937, 1, 1, 19}, {0, 0, 0, 0}}},{24922, {{35937, 1, 0, 0},{35937, 1, 1, 19}, {0, 0, 0, 0}}},{24923, {{35937, 1, 0, 0},{35937, 1, 1, 18}, {0, 0, 0, 0}}},{24923, {{35937, 1, 0, 0},{35937, 1, 1, 18}, {0, 0, 0, 0}}},{24924, {{35937, 1, 0, 0},{35937, 1, 1, 17}, {0, 0, 0, 0}}},{24924, {{35937, 1, 0, 0},{35937, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{24926, {{35938, 1, 0, 0},{35938, 1, 1, 17}, {0, 0, 0, 0}}},{24926, {{35938, 1, 0, 0},{35938, 1, 1, 17}, {0, 0, 0, 0}}},{24927, {{35938, 1, 0, 0},{35938, 1, 1, 18}, {0, 0, 0, 0}}},{24927, {{35938, 1, 0, 0},{35938, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24928, {{35939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24929, {{35940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24930, {{35941, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24931, {{35942, 1, 0, 0},{35942, 1, 1, 17}, {0, 0, 0, 0}}},{24931, {{35942, 1, 0, 0},{35942, 1, 1, 17}, {0, 0, 0, 0}}},{24932, {{35942, 1, 0, 0},{35942, 1, 1, 18}, {0, 0, 0, 0}}},{24932, {{35942, 1, 0, 0},{35942, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24933, {{35943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24934, {{35944, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24935, {{35945, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{24936, {{35946, 1, 0, 0},{35946, 1, 1, 18}, {0, 0, 0, 0}}},{24936, {{35946, 1, 0, 0},{35946, 1, 1, 18}, {0, 0, 0, 0}}},{24937, {{35946, 1, 0, 0},{35946, 1, 1, 19}, {0, 0, 0, 0}}},{24937, {{35946, 1, 0, 0},{35946, 1, 1, 19}, {0, 0, 0, 0}}},{24938, {{35946, 1, 0, 0},{35946, 1, 1, 20}, {0, 0, 0, 0}}},{24938, {{35946, 1, 0, 0},{35946, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{24939, {{35947, 1, 0, 0},{35947, 1, 1, 17}, {0, 0, 0, 0}}},{24939, {{35947, 1, 0, 0},{35947, 1, 1, 17}, {0, 0, 0, 0}}},{24940, {{35947, 1, 0, 0},{35947, 1, 1, 18}, {0, 0, 0, 0}}},{24940, {{35947, 1, 0, 0},{35947, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{24941, {{35948, 1, 0, 0},{35948, 1, 1, 17}, {0, 0, 0, 0}}},{24941, {{35948, 1, 0, 0},{35948, 1, 1, 17}, {0, 0, 0, 0}}},{24942, {{35948, 1, 0, 0},{35948, 1, 1, 18}, {0, 0, 0, 0}}},{24942, {{35948, 1, 0, 0},{35948, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24943, {{35949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{24944, {{35951, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24945, {{35952, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24946, {{35953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24947, {{35954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24948, {{35955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24949, {{35956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24950, {{35957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24951, {{35958, 1, 0, 0},{35958, 1, 1, 17}, {0, 0, 0, 0}}},{24951, {{35958, 1, 0, 0},{35958, 1, 1, 17}, {0, 0, 0, 0}}},{24952, {{35958, 1, 0, 0},{35958, 1, 1, 18}, {0, 0, 0, 0}}},{24952, {{35958, 1, 0, 0},{35958, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24953, {{35959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24954, {{35960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24955, {{35961, 1, 0, 0},{35961, 1, 1, 19}, {0, 0, 0, 0}}},{24955, {{35961, 1, 0, 0},{35961, 1, 1, 19}, {0, 0, 0, 0}}},{24956, {{35961, 1, 0, 0},{35961, 1, 1, 18}, {0, 0, 0, 0}}},{24956, {{35961, 1, 0, 0},{35961, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24957, {{35962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24958, {{35963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24959, {{35964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24960, {{35965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24961, {{35966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24962, {{35967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24963, {{35968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24964, {{35969, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24965, {{35970, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24966, {{35971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24967, {{35972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24968, {{35973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24969, {{35974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24970, {{35975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24971, {{35976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24972, {{35977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24973, {{35978, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24974, {{35979, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24975, {{35980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24976, {{35981, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24977, {{35982, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24978, {{35983, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24979, {{35984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24980, {{35985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24981, {{35986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24982, {{35987, 1, 0, 0},{35987, 1, 1, 17}, {0, 0, 0, 0}}},{24982, {{35987, 1, 0, 0},{35987, 1, 1, 17}, {0, 0, 0, 0}}},{24983, {{35987, 1, 0, 0},{35987, 1, 1, 18}, {0, 0, 0, 0}}},{24983, {{35987, 1, 0, 0},{35987, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24984, {{35988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24985, {{35989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24986, {{35990, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24987, {{35991, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24988, {{35992, 1, 0, 0},{35992, 1, 1, 17}, {0, 0, 0, 0}}},{24988, {{35992, 1, 0, 0},{35992, 1, 1, 17}, {0, 0, 0, 0}}},{24989, {{35992, 1, 0, 0},{35992, 1, 1, 18}, {0, 0, 0, 0}}},{24989, {{35992, 1, 0, 0},{35992, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{24990, {{35993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24991, {{35994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{24992, {{35995, 1, 0, 0},{35995, 1, 1, 18}, {0, 0, 0, 0}}},{24992, {{35995, 1, 0, 0},{35995, 1, 1, 18}, {0, 0, 0, 0}}},{24993, {{35995, 1, 0, 0},{35995, 1, 1, 19}, {0, 0, 0, 0}}},{24993, {{35995, 1, 0, 0},{35995, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{24994, {{35996, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24995, {{35997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24996, {{35998, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24997, {{35999, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{24998, {{36000, 1, 0, 0},{36000, 1, 1, 20}, {0, 0, 0, 0}}},{24998, {{36000, 1, 0, 0},{36000, 1, 1, 20}, {0, 0, 0, 0}}},{24999, {{36000, 1, 0, 0},{36000, 1, 1, 19}, {0, 0, 0, 0}}},{24999, {{36000, 1, 0, 0},{36000, 1, 1, 19}, {0, 0, 0, 0}}},{25000, {{36000, 1, 0, 0},{36000, 1, 1, 21}, {0, 0, 0, 0}}},{25000, {{36000, 1, 0, 0},{36000, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{25001, {{36001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25002, {{36002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25003, {{36003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25004, {{36004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25005, {{36005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25006, {{36006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25007, {{36007, 1, 0, 0},{36007, 1, 1, 18}, {0, 0, 0, 0}}},{25007, {{36007, 1, 0, 0},{36007, 1, 1, 18}, {0, 0, 0, 0}}},{25008, {{36007, 1, 0, 0},{36007, 1, 1, 19}, {0, 0, 0, 0}}},{25008, {{36007, 1, 0, 0},{36007, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{25009, {{36008, 1, 0, 0},{36008, 1, 1, 18}, {0, 0, 0, 0}}},{25009, {{36008, 1, 0, 0},{36008, 1, 1, 18}, {0, 0, 0, 0}}},{25010, {{36008, 1, 0, 0},{36008, 1, 1, 19}, {0, 0, 0, 0}}},{25010, {{36008, 1, 0, 0},{36008, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{25011, {{36009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25012, {{36010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{25013, {{36011, 1, 0, 0},{36011, 1, 1, 18}, {0, 0, 0, 0}}},{25013, {{36011, 1, 0, 0},{36011, 1, 1, 18}, {0, 0, 0, 0}}},{25014, {{36011, 1, 0, 0},{36011, 1, 1, 19}, {0, 0, 0, 0}}},{25014, {{36011, 1, 0, 0},{36011, 1, 1, 19}, {0, 0, 0, 0}}},{25015, {{36011, 1, 0, 0},{36011, 1, 1, 20}, {0, 0, 0, 0}}},{25015, {{36011, 1, 0, 0},{36011, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{25016, {{36012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25017, {{36013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25018, {{36014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25019, {{36015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25020, {{36016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25021, {{36017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25022, {{36018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25023, {{36019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25024, {{36020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25025, {{36021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25026, {{36022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25027, {{36023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25028, {{36024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25029, {{36025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25030, {{36026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25031, {{36027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25032, {{36028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25033, {{36029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25034, {{36030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25035, {{36031, 1, 0, 0},{36031, 1, 1, 17}, {0, 0, 0, 0}}},{25035, {{36031, 1, 0, 0},{36031, 1, 1, 17}, {0, 0, 0, 0}}},{58784, {{36031, 1, 0, 0},{36031, 1, 1, 18}, {0, 0, 0, 0}}},{58784, {{36031, 1, 0, 0},{36031, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25036, {{36032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25037, {{36033, 1, 0, 0},{36033, 1, 1, 17}, {0, 0, 0, 0}}},{25037, {{36033, 1, 0, 0},{36033, 1, 1, 17}, {0, 0, 0, 0}}},{25038, {{36033, 1, 0, 0},{36033, 1, 1, 18}, {0, 0, 0, 0}}},{25038, {{36033, 1, 0, 0},{36033, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25039, {{36034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25040, {{36035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25041, {{36036, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25042, {{36037, 1, 0, 0},{36037, 1, 1, 17}, {0, 0, 0, 0}}},{25042, {{36037, 1, 0, 0},{36037, 1, 1, 17}, {0, 0, 0, 0}}},{25043, {{36037, 1, 0, 0},{36037, 1, 1, 18}, {0, 0, 0, 0}}},{25043, {{36037, 1, 0, 0},{36037, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25044, {{36038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25046, {{36039, 1, 0, 0},{36039, 1, 1, 20}, {0, 0, 0, 0}}},{25046, {{36039, 1, 0, 0},{36039, 1, 1, 20}, {0, 0, 0, 0}}},{25047, {{36039, 1, 0, 0},{36039, 1, 1, 19}, {0, 0, 0, 0}}},{25047, {{36039, 1, 0, 0},{36039, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{25048, {{36040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25049, {{36041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25050, {{36042, 1, 0, 0},{36042, 1, 1, 20}, {0, 0, 0, 0}}},{25050, {{36042, 1, 0, 0},{36042, 1, 1, 20}, {0, 0, 0, 0}}},{25051, {{36042, 1, 0, 0},{36042, 1, 1, 19}, {0, 0, 0, 0}}},{25051, {{36042, 1, 0, 0},{36042, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{25052, {{36043, 1, 0, 0},{36043, 1, 1, 17}, {0, 0, 0, 0}}},{25052, {{36043, 1, 0, 0},{36043, 1, 1, 17}, {0, 0, 0, 0}}},{25053, {{36043, 1, 0, 0},{36043, 1, 1, 18}, {0, 0, 0, 0}}},{25053, {{36043, 1, 0, 0},{36043, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25054, {{36044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25055, {{36045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25056, {{36046, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25057, {{36047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25058, {{36048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25060, {{36049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25061, {{36050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {7, {{25062, {{36051, 1, 0, 0},{36051, 1, 1, 18}, {0, 0, 0, 0}}},{25062, {{36051, 1, 0, 0},{36051, 1, 1, 18}, {0, 0, 0, 0}}},{30299, {{36051, 1, 0, 0},{36051, 1, 1, 0},{36051, 1, 1, 19},{64100, 1, 0, 0}, {0, 0, 0, 0}}},{30299, {{36051, 1, 0, 0},{36051, 1, 1, 0},{36051, 1, 1, 19},{64100, 1, 0, 0}, {0, 0, 0, 0}}},{30299, {{36051, 1, 0, 0},{36051, 1, 1, 0},{36051, 1, 1, 19},{64100, 1, 0, 0}, {0, 0, 0, 0}}},{30300, {{36051, 1, 0, 0},{36051, 1, 1, 20}, {0, 0, 0, 0}}},{30300, {{36051, 1, 0, 0},{36051, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{25063, {{36052, 1, 0, 0},{36052, 1, 1, 16}, {0, 0, 0, 0}}},{25063, {{36052, 1, 0, 0},{36052, 1, 1, 16}, {0, 0, 0, 0}}},{58790, {{36052, 1, 0, 0},{36052, 1, 1, 17}, {0, 0, 0, 0}}},{58790, {{36052, 1, 0, 0},{36052, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{25064, {{36053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25065, {{36054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25066, {{36055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{25067, {{36057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25068, {{36058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25069, {{36059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25070, {{36060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25071, {{36061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25072, {{36062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25073, {{36063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25074, {{36064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25075, {{36065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25076, {{36066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25077, {{36067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25078, {{36068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25079, {{36069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25080, {{36070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25081, {{36071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25082, {{36072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25083, {{36073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25084, {{36074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25085, {{36075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25086, {{36076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25087, {{36077, 1, 0, 0},{36077, 1, 1, 18}, {0, 0, 0, 0}}},{25087, {{36077, 1, 0, 0},{36077, 1, 1, 18}, {0, 0, 0, 0}}},{25088, {{36077, 1, 0, 0},{36077, 1, 1, 19}, {0, 0, 0, 0}}},{25088, {{36077, 1, 0, 0},{36077, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{25089, {{36078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25090, {{36079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25091, {{36080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25092, {{36081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25093, {{36082, 1, 0, 0},{36082, 1, 1, 17}, {0, 0, 0, 0}}},{25093, {{36082, 1, 0, 0},{36082, 1, 1, 17}, {0, 0, 0, 0}}},{25094, {{36082, 1, 0, 0},{36082, 1, 1, 18}, {0, 0, 0, 0}}},{25094, {{36082, 1, 0, 0},{36082, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25095, {{36083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25096, {{36084, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25097, {{36085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25098, {{36086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25099, {{36087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25100, {{36088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25101, {{36089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25102, {{36090, 1, 0, 0},{36090, 1, 1, 17}, {0, 0, 0, 0}}},{25102, {{36090, 1, 0, 0},{36090, 1, 1, 17}, {0, 0, 0, 0}}},{25103, {{36090, 1, 0, 0},{36090, 1, 1, 18}, {0, 0, 0, 0}}},{25103, {{36090, 1, 0, 0},{36090, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{25104, {{36091, 1, 0, 0},{36091, 1, 1, 17}, {0, 0, 0, 0}}},{25104, {{36091, 1, 0, 0},{36091, 1, 1, 17}, {0, 0, 0, 0}}},{25105, {{36091, 1, 0, 0},{36091, 1, 1, 18}, {0, 0, 0, 0}}},{25105, {{36091, 1, 0, 0},{36091, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{25106, {{36092, 1, 0, 0},{36092, 1, 1, 20}, {0, 0, 0, 0}}},{25106, {{36092, 1, 0, 0},{36092, 1, 1, 20}, {0, 0, 0, 0}}},{25107, {{36092, 1, 0, 0},{36092, 1, 1, 19}, {0, 0, 0, 0}}},{25107, {{36092, 1, 0, 0},{36092, 1, 1, 19}, {0, 0, 0, 0}}},{25108, {{36092, 1, 0, 0},{36092, 1, 1, 21}, {0, 0, 0, 0}}},{25108, {{36092, 1, 0, 0},{36092, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{25109, {{36093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25110, {{36094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25112, {{36095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25113, {{36096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25114, {{36097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25115, {{36098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25116, {{36099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25117, {{36100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25118, {{36101, 1, 0, 0},{36101, 1, 1, 18}, {0, 0, 0, 0}}},{25118, {{36101, 1, 0, 0},{36101, 1, 1, 18}, {0, 0, 0, 0}}},{25119, {{36101, 1, 0, 0},{36101, 1, 1, 19}, {0, 0, 0, 0}}},{25119, {{36101, 1, 0, 0},{36101, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{25120, {{36102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25121, {{36103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{25122, {{36104, 1, 0, 0},{36104, 1, 1, 18}, {0, 0, 0, 0}}},{25122, {{36104, 1, 0, 0},{36104, 1, 1, 18}, {0, 0, 0, 0}}},{30301, {{36104, 1, 0, 0},{36104, 1, 1, 0},{36104, 1, 1, 19},{64101, 1, 0, 0}, {0, 0, 0, 0}}},{30301, {{36104, 1, 0, 0},{36104, 1, 1, 0},{36104, 1, 1, 19},{64101, 1, 0, 0}, {0, 0, 0, 0}}},{30301, {{36104, 1, 0, 0},{36104, 1, 1, 0},{36104, 1, 1, 19},{64101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25123, {{36105, 1, 0, 0},{36105, 1, 1, 17}, {0, 0, 0, 0}}},{25123, {{36105, 1, 0, 0},{36105, 1, 1, 17}, {0, 0, 0, 0}}},{25124, {{36105, 1, 0, 0},{36105, 1, 1, 18}, {0, 0, 0, 0}}},{25124, {{36105, 1, 0, 0},{36105, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{25125, {{36106, 1, 0, 0},{36106, 1, 1, 17}, {0, 0, 0, 0}}},{25125, {{36106, 1, 0, 0},{36106, 1, 1, 17}, {0, 0, 0, 0}}},{25126, {{36106, 1, 0, 0},{36106, 1, 1, 18}, {0, 0, 0, 0}}},{25126, {{36106, 1, 0, 0},{36106, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25127, {{36107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{25128, {{36109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25129, {{36110, 1, 0, 0},{36110, 1, 1, 17}, {0, 0, 0, 0}}},{25129, {{36110, 1, 0, 0},{36110, 1, 1, 17}, {0, 0, 0, 0}}},{25130, {{36110, 1, 0, 0},{36110, 1, 1, 18}, {0, 0, 0, 0}}},{25130, {{36110, 1, 0, 0},{36110, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {8, {{25131, {{36111, 1, 0, 0},{36111, 1, 1, 19}, {0, 0, 0, 0}}},{25131, {{36111, 1, 0, 0},{36111, 1, 1, 19}, {0, 0, 0, 0}}},{25132, {{36111, 1, 0, 0},{36111, 1, 1, 18}, {0, 0, 0, 0}}},{25132, {{36111, 1, 0, 0},{36111, 1, 1, 18}, {0, 0, 0, 0}}},{25133, {{36111, 1, 0, 0},{36111, 1, 1, 21}, {0, 0, 0, 0}}},{25133, {{36111, 1, 0, 0},{36111, 1, 1, 21}, {0, 0, 0, 0}}},{25134, {{36111, 1, 0, 0},{36111, 1, 1, 20}, {0, 0, 0, 0}}},{25134, {{36111, 1, 0, 0},{36111, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{25135, {{36112, 1, 0, 0},{36112, 1, 1, 17}, {0, 0, 0, 0}}},{25135, {{36112, 1, 0, 0},{36112, 1, 1, 17}, {0, 0, 0, 0}}},{25136, {{36112, 1, 0, 0},{36112, 1, 1, 18}, {0, 0, 0, 0}}},{25136, {{36112, 1, 0, 0},{36112, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{25137, {{36113, 1, 0, 0},{36113, 1, 1, 16}, {0, 0, 0, 0}}},{25137, {{36113, 1, 0, 0},{36113, 1, 1, 16}, {0, 0, 0, 0}}},{25138, {{36113, 1, 0, 0},{36113, 1, 1, 17}, {0, 0, 0, 0}}},{25138, {{36113, 1, 0, 0},{36113, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{25139, {{36114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25140, {{36115, 1, 0, 0},{36115, 1, 1, 17}, {0, 0, 0, 0}}},{25140, {{36115, 1, 0, 0},{36115, 1, 1, 17}, {0, 0, 0, 0}}},{25141, {{36115, 1, 0, 0},{36115, 1, 1, 18}, {0, 0, 0, 0}}},{25141, {{36115, 1, 0, 0},{36115, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25142, {{36116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25143, {{36117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25144, {{36118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25145, {{36119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25146, {{36120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25147, {{36121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25148, {{36122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25149, {{36123, 1, 0, 0},{36123, 1, 1, 17}, {0, 0, 0, 0}}},{25149, {{36123, 1, 0, 0},{36123, 1, 1, 17}, {0, 0, 0, 0}}},{25150, {{36123, 1, 0, 0},{36123, 1, 1, 18}, {0, 0, 0, 0}}},{25150, {{36123, 1, 0, 0},{36123, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{25152, {{36124, 1, 0, 0},{36124, 1, 1, 17}, {0, 0, 0, 0}}},{25152, {{36124, 1, 0, 0},{36124, 1, 1, 17}, {0, 0, 0, 0}}},{25153, {{36124, 1, 0, 0},{36124, 1, 1, 18}, {0, 0, 0, 0}}},{25153, {{36124, 1, 0, 0},{36124, 1, 1, 18}, {0, 0, 0, 0}}},{25154, {{36124, 1, 0, 0},{36124, 1, 1, 19}, {0, 0, 0, 0}}},{25154, {{36124, 1, 0, 0},{36124, 1, 1, 19}, {0, 0, 0, 0}}}}}
};

static const MJMappingSet chunked_mj_mappings_008d64_009484[1825] = {
    {1, {{25157, {{36196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25158, {{36197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25159, {{36198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25160, {{36199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25161, {{36200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25162, {{36201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{25163, {{36203, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25164, {{36204, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25165, {{36205, 1, 0, 0},{36205, 1, 1, 17}, {0, 0, 0, 0}}},{25165, {{36205, 1, 0, 0},{36205, 1, 1, 17}, {0, 0, 0, 0}}},{25166, {{36205, 1, 0, 0},{36205, 1, 1, 18}, {0, 0, 0, 0}}},{25166, {{36205, 1, 0, 0},{36205, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25167, {{36206, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25168, {{36207, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25169, {{36208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25170, {{36209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25171, {{36210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{25172, {{36211, 1, 0, 0},{36211, 1, 1, 19}, {0, 0, 0, 0}}},{25172, {{36211, 1, 0, 0},{36211, 1, 1, 19}, {0, 0, 0, 0}}},{25173, {{36211, 1, 0, 0},{36211, 1, 1, 20}, {0, 0, 0, 0}}},{25173, {{36211, 1, 0, 0},{36211, 1, 1, 20}, {0, 0, 0, 0}}},{25174, {{36211, 1, 0, 0},{36211, 1, 1, 18}, {0, 0, 0, 0}}},{25174, {{36211, 1, 0, 0},{36211, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25175, {{36212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{25177, {{36214, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25178, {{36215, 1, 0, 0},{36215, 1, 1, 18}, {0, 0, 0, 0}}},{25178, {{36215, 1, 0, 0},{36215, 1, 1, 18}, {0, 0, 0, 0}}},{25179, {{36215, 1, 0, 0},{36215, 1, 1, 19}, {0, 0, 0, 0}}},{25179, {{36215, 1, 0, 0},{36215, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{25180, {{36216, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25181, {{36217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25182, {{36218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25183, {{36219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25184, {{36220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25185, {{36221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25186, {{36222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25187, {{36223, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25188, {{36224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25189, {{36225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25190, {{36226, 1, 0, 0},{36226, 1, 1, 17}, {0, 0, 0, 0}}},{25190, {{36226, 1, 0, 0},{36226, 1, 1, 17}, {0, 0, 0, 0}}},{25191, {{36226, 1, 0, 0},{36226, 1, 1, 18}, {0, 0, 0, 0}}},{25191, {{36226, 1, 0, 0},{36226, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25192, {{36227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25193, {{36228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25194, {{36229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25195, {{36230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25196, {{36231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25197, {{36232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25198, {{36233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25199, {{36234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25200, {{36235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25201, {{36236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25202, {{36237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25203, {{36238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25204, {{36239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25205, {{36240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25206, {{36241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25207, {{36242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{25208, {{36244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25209, {{36245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25210, {{36246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25211, {{36247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25212, {{36248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25213, {{36249, 1, 0, 0},{36249, 1, 1, 18}, {0, 0, 0, 0}}},{25213, {{36249, 1, 0, 0},{36249, 1, 1, 18}, {0, 0, 0, 0}}},{25214, {{36249, 1, 0, 0},{36249, 1, 1, 19}, {0, 0, 0, 0}}},{25214, {{36249, 1, 0, 0},{36249, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{25215, {{36250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25216, {{36251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25217, {{36252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25218, {{36253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25219, {{36254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25220, {{36255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25221, {{36256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25222, {{36257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25223, {{36258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25224, {{36259, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25225, {{36260, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25226, {{36261, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25227, {{36262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25228, {{36263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25229, {{36264, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25230, {{36265, 1, 0, 0},{36265, 1, 1, 16}, {0, 0, 0, 0}}},{25230, {{36265, 1, 0, 0},{36265, 1, 1, 16}, {0, 0, 0, 0}}},{25231, {{36265, 1, 0, 0},{36265, 1, 1, 17}, {0, 0, 0, 0}}},{25231, {{36265, 1, 0, 0},{36265, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{25232, {{36266, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25233, {{36267, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25234, {{36268, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25235, {{36269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25236, {{36270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25237, {{36271, 1, 0, 0},{36271, 1, 1, 17}, {0, 0, 0, 0}}},{25237, {{36271, 1, 0, 0},{36271, 1, 1, 17}, {0, 0, 0, 0}}},{25238, {{36271, 1, 0, 0},{36271, 1, 1, 18}, {0, 0, 0, 0}}},{25238, {{36271, 1, 0, 0},{36271, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{25239, {{36274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25240, {{36275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25241, {{36276, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25242, {{36277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25243, {{36278, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25244, {{36279, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{25245, {{36281, 1, 0, 0},{36281, 1, 1, 17}, {0, 0, 0, 0}}},{25245, {{36281, 1, 0, 0},{36281, 1, 1, 17}, {0, 0, 0, 0}}},{25246, {{36281, 1, 0, 0},{36281, 1, 1, 18}, {0, 0, 0, 0}}},{25246, {{36281, 1, 0, 0},{36281, 1, 1, 18}, {0, 0, 0, 0}}},{68068, {{36281, 1, 0, 0},{36281, 1, 1, 19}, {0, 0, 0, 0}}},{68068, {{36281, 1, 0, 0},{36281, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{25247, {{36282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25248, {{36283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25249, {{36284, 1, 0, 0},{36284, 1, 1, 18}, {0, 0, 0, 0}}},{25249, {{36284, 1, 0, 0},{36284, 1, 1, 18}, {0, 0, 0, 0}}},{25250, {{36284, 1, 0, 0},{36284, 1, 1, 17}, {0, 0, 0, 0}}},{25250, {{36284, 1, 0, 0},{36284, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{25251, {{36285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25252, {{36286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25253, {{36287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25254, {{36288, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25255, {{36289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25256, {{36290, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{25258, {{36293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25259, {{36294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25260, {{36295, 1, 0, 0},{36295, 1, 1, 17}, {0, 0, 0, 0}}},{25260, {{36295, 1, 0, 0},{36295, 1, 1, 17}, {0, 0, 0, 0}}},{25261, {{36295, 1, 0, 0},{36295, 1, 1, 18}, {0, 0, 0, 0}}},{25261, {{36295, 1, 0, 0},{36295, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25262, {{36296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25263, {{36297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25264, {{36298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25265, {{36299, 1, 0, 0},{36299, 1, 1, 17}, {0, 0, 0, 0}}},{25265, {{36299, 1, 0, 0},{36299, 1, 1, 17}, {0, 0, 0, 0}}},{25266, {{36299, 1, 0, 0},{36299, 1, 1, 18}, {0, 0, 0, 0}}},{25266, {{36299, 1, 0, 0},{36299, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25267, {{36300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25268, {{36301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25269, {{36302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25270, {{36303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25271, {{36304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25272, {{36305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25274, {{36306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25275, {{36307, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25276, {{36308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25277, {{36309, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25278, {{36310, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25279, {{36311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25280, {{36312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25281, {{36313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{25282, {{36314, 1, 0, 0},{36314, 1, 1, 19}, {0, 0, 0, 0}}},{25282, {{36314, 1, 0, 0},{36314, 1, 1, 19}, {0, 0, 0, 0}}},{25283, {{36314, 1, 0, 0},{36314, 1, 1, 18}, {0, 0, 0, 0}}},{25283, {{36314, 1, 0, 0},{36314, 1, 1, 18}, {0, 0, 0, 0}}},{25284, {{36314, 1, 0, 0},{36314, 1, 1, 20}, {0, 0, 0, 0}}},{25284, {{36314, 1, 0, 0},{36314, 1, 1, 20}, {0, 0, 0, 0}}},{25285, {{36314, 1, 0, 0},{36314, 1, 1, 21}, {0, 0, 0, 0}}},{25285, {{36314, 1, 0, 0},{36314, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{25286, {{36315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25287, {{36316, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25288, {{36317, 1, 0, 0},{36317, 1, 1, 19}, {0, 0, 0, 0}}},{25288, {{36317, 1, 0, 0},{36317, 1, 1, 19}, {0, 0, 0, 0}}},{25289, {{36317, 1, 0, 0},{36317, 1, 1, 18}, {0, 0, 0, 0}}},{25289, {{36317, 1, 0, 0},{36317, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{25290, {{36319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25291, {{36320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25292, {{36321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25293, {{36322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25294, {{36323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25295, {{36324, 1, 0, 0},{36324, 1, 1, 17}, {0, 0, 0, 0}}},{25295, {{36324, 1, 0, 0},{36324, 1, 1, 17}, {0, 0, 0, 0}}},{25296, {{36324, 1, 0, 0},{36324, 1, 1, 18}, {0, 0, 0, 0}}},{25296, {{36324, 1, 0, 0},{36324, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25297, {{36325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25298, {{36326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25299, {{36327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25300, {{36328, 1, 0, 0},{36328, 1, 1, 17}, {0, 0, 0, 0}}},{25300, {{36328, 1, 0, 0},{36328, 1, 1, 17}, {0, 0, 0, 0}}},{25301, {{36328, 1, 0, 0},{36328, 1, 1, 18}, {0, 0, 0, 0}}},{25301, {{36328, 1, 0, 0},{36328, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25302, {{36329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25303, {{36330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25304, {{36331, 1, 0, 0},{36331, 1, 1, 17}, {0, 0, 0, 0}}},{25304, {{36331, 1, 0, 0},{36331, 1, 1, 17}, {0, 0, 0, 0}}},{25305, {{36331, 1, 0, 0},{36331, 1, 1, 18}, {0, 0, 0, 0}}},{25305, {{36331, 1, 0, 0},{36331, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25306, {{36332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25307, {{36333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25308, {{36334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25309, {{36335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25310, {{36336, 1, 0, 0},{36336, 1, 1, 17}, {0, 0, 0, 0}}},{25310, {{36336, 1, 0, 0},{36336, 1, 1, 17}, {0, 0, 0, 0}}},{25311, {{36336, 1, 0, 0},{36336, 1, 1, 18}, {0, 0, 0, 0}}},{25311, {{36336, 1, 0, 0},{36336, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25312, {{36337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25313, {{36338, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25314, {{36339, 1, 0, 0},{36339, 1, 1, 18}, {0, 0, 0, 0}}},{25314, {{36339, 1, 0, 0},{36339, 1, 1, 18}, {0, 0, 0, 0}}},{25315, {{36339, 1, 0, 0},{36339, 1, 1, 19}, {0, 0, 0, 0}}},{25315, {{36339, 1, 0, 0},{36339, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{25316, {{36340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25317, {{36341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{25318, {{36348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25319, {{36349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25320, {{36350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25321, {{36351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25322, {{36352, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25323, {{36353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25324, {{36354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25325, {{36355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25326, {{36356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25327, {{36357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25328, {{36358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25329, {{36359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25330, {{36360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25331, {{36361, 1, 0, 0},{36361, 1, 1, 18}, {0, 0, 0, 0}}},{25331, {{36361, 1, 0, 0},{36361, 1, 1, 18}, {0, 0, 0, 0}}},{25332, {{36361, 1, 0, 0},{36361, 1, 1, 19}, {0, 0, 0, 0}}},{25332, {{36361, 1, 0, 0},{36361, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{25333, {{36362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25334, {{36363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25335, {{36364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{25336, {{36367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25337, {{36368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25338, {{36369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25339, {{36370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25340, {{36371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25341, {{36372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25342, {{36373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25343, {{36374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25344, {{36375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25345, {{36376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{25346, {{36378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25347, {{36379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25348, {{36380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25349, {{36381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25350, {{36382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25351, {{36383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25352, {{36384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25353, {{36385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25354, {{36386, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25355, {{36387, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25356, {{36388, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25357, {{36389, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25358, {{36390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25359, {{36391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25360, {{36392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25361, {{36393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25362, {{36394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25363, {{36395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{25364, {{36400, 1, 0, 0},{36400, 1, 1, 17}, {0, 0, 0, 0}}},{25364, {{36400, 1, 0, 0},{36400, 1, 1, 17}, {0, 0, 0, 0}}},{25365, {{36400, 1, 0, 0},{36400, 1, 1, 18}, {0, 0, 0, 0}}},{25365, {{36400, 1, 0, 0},{36400, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25366, {{36401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25367, {{36402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25368, {{36403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25369, {{36404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25370, {{36405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25371, {{36406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25372, {{36407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25373, {{36408, 1, 0, 0},{36408, 1, 1, 17}, {0, 0, 0, 0}}},{25373, {{36408, 1, 0, 0},{36408, 1, 1, 17}, {0, 0, 0, 0}}},{25374, {{36408, 1, 0, 0},{36408, 1, 1, 18}, {0, 0, 0, 0}}},{25374, {{36408, 1, 0, 0},{36408, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25375, {{36409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25376, {{36410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25377, {{36411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25378, {{36412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25379, {{36413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25380, {{36414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25381, {{36415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25382, {{36416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25383, {{36417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25384, {{36418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25385, {{36419, 1, 0, 0},{36419, 1, 1, 16}, {0, 0, 0, 0}}},{25385, {{36419, 1, 0, 0},{36419, 1, 1, 16}, {0, 0, 0, 0}}},{25386, {{36419, 1, 0, 0},{36419, 1, 1, 17}, {0, 0, 0, 0}}},{25386, {{36419, 1, 0, 0},{36419, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{25387, {{36420, 1, 0, 0},{36420, 1, 1, 17}, {0, 0, 0, 0}}},{25387, {{36420, 1, 0, 0},{36420, 1, 1, 17}, {0, 0, 0, 0}}},{25388, {{36420, 1, 0, 0},{36420, 1, 1, 18}, {0, 0, 0, 0}}},{25388, {{36420, 1, 0, 0},{36420, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25389, {{36421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25390, {{36422, 1, 0, 0},{36422, 1, 1, 16}, {0, 0, 0, 0}}},{25390, {{36422, 1, 0, 0},{36422, 1, 1, 16}, {0, 0, 0, 0}}},{25391, {{36422, 1, 0, 0},{36422, 1, 1, 17}, {0, 0, 0, 0}}},{25391, {{36422, 1, 0, 0},{36422, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{25392, {{36423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25393, {{36424, 1, 0, 0},{36424, 1, 1, 17}, {0, 0, 0, 0}}},{25393, {{36424, 1, 0, 0},{36424, 1, 1, 17}, {0, 0, 0, 0}}},{25394, {{36424, 1, 0, 0},{36424, 1, 1, 18}, {0, 0, 0, 0}}},{25394, {{36424, 1, 0, 0},{36424, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25395, {{36425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25396, {{36426, 1, 0, 0},{36426, 1, 1, 18}, {0, 0, 0, 0}}},{25396, {{36426, 1, 0, 0},{36426, 1, 1, 18}, {0, 0, 0, 0}}},{25397, {{36426, 1, 0, 0},{36426, 1, 1, 19}, {0, 0, 0, 0}}},{25397, {{36426, 1, 0, 0},{36426, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{25398, {{36427, 1, 0, 0},{36427, 1, 1, 17}, {0, 0, 0, 0}}},{25398, {{36427, 1, 0, 0},{36427, 1, 1, 17}, {0, 0, 0, 0}}},{25399, {{36427, 1, 0, 0},{36427, 1, 1, 18}, {0, 0, 0, 0}}},{25399, {{36427, 1, 0, 0},{36427, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25400, {{36428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25401, {{36429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25402, {{36430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25403, {{36431, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25404, {{36432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25405, {{36433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{25406, {{36436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25407, {{36437, 1, 0, 0},{36437, 1, 1, 17}, {0, 0, 0, 0}}},{25407, {{36437, 1, 0, 0},{36437, 1, 1, 17}, {0, 0, 0, 0}}},{25408, {{36437, 1, 0, 0},{36437, 1, 1, 18}, {0, 0, 0, 0}}},{25408, {{36437, 1, 0, 0},{36437, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25409, {{36438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25410, {{36439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25411, {{36440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25412, {{36441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25413, {{36442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25414, {{36443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25415, {{36444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25416, {{36445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25417, {{36446, 1, 0, 0},{36446, 1, 1, 17}, {0, 0, 0, 0}}},{25417, {{36446, 1, 0, 0},{36446, 1, 1, 17}, {0, 0, 0, 0}}},{25418, {{36446, 1, 0, 0},{36446, 1, 1, 18}, {0, 0, 0, 0}}},{25418, {{36446, 1, 0, 0},{36446, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25419, {{36447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25420, {{36448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25421, {{36449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25422, {{36450, 1, 0, 0},{36450, 1, 1, 17}, {0, 0, 0, 0}}},{25422, {{36450, 1, 0, 0},{36450, 1, 1, 17}, {0, 0, 0, 0}}},{25423, {{36450, 1, 0, 0},{36450, 1, 1, 18}, {0, 0, 0, 0}}},{25423, {{36450, 1, 0, 0},{36450, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25424, {{36451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25425, {{36452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25426, {{36453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25427, {{36454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25428, {{36455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25429, {{36456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25430, {{36457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25431, {{36458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25432, {{36459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25433, {{36460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25434, {{36461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{25435, {{36462, 1, 0, 0},{36462, 1, 1, 18}, {0, 0, 0, 0}}},{25435, {{36462, 1, 0, 0},{36462, 1, 1, 18}, {0, 0, 0, 0}}},{25436, {{36462, 1, 0, 0},{36462, 1, 1, 16}, {0, 0, 0, 0}}},{25436, {{36462, 1, 0, 0},{36462, 1, 1, 16}, {0, 0, 0, 0}}},{58816, {{36462, 1, 0, 0},{36462, 1, 1, 17}, {0, 0, 0, 0}}},{58816, {{36462, 1, 0, 0},{36462, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{25437, {{36463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25438, {{36464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25439, {{36465, 1, 0, 0},{36465, 1, 1, 17}, {0, 0, 0, 0}}},{25439, {{36465, 1, 0, 0},{36465, 1, 1, 17}, {0, 0, 0, 0}}},{25440, {{36465, 1, 0, 0},{36465, 1, 1, 18}, {0, 0, 0, 0}}},{25440, {{36465, 1, 0, 0},{36465, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{25441, {{36466, 1, 0, 0},{36466, 1, 1, 17}, {0, 0, 0, 0}}},{25441, {{36466, 1, 0, 0},{36466, 1, 1, 17}, {0, 0, 0, 0}}},{25442, {{36466, 1, 0, 0},{36466, 1, 1, 18}, {0, 0, 0, 0}}},{25442, {{36466, 1, 0, 0},{36466, 1, 1, 18}, {0, 0, 0, 0}}},{25443, {{36466, 1, 0, 0},{36466, 1, 1, 19}, {0, 0, 0, 0}}},{25443, {{36466, 1, 0, 0},{36466, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{25444, {{36467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25445, {{36468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25446, {{36469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25447, {{36470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25448, {{36471, 1, 0, 0},{36471, 1, 1, 17}, {0, 0, 0, 0}}},{25448, {{36471, 1, 0, 0},{36471, 1, 1, 17}, {0, 0, 0, 0}}},{25449, {{36471, 1, 0, 0},{36471, 1, 1, 18}, {0, 0, 0, 0}}},{25449, {{36471, 1, 0, 0},{36471, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25450, {{36472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25451, {{36473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25452, {{36474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25453, {{36475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25454, {{36476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{25455, {{36481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25456, {{36482, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25457, {{36483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25458, {{36484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25459, {{36485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25460, {{36486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{25461, {{36487, 1, 0, 0},{36487, 1, 1, 17}, {0, 0, 0, 0}}},{25461, {{36487, 1, 0, 0},{36487, 1, 1, 17}, {0, 0, 0, 0}}},{25462, {{36487, 1, 0, 0},{36487, 1, 1, 18}, {0, 0, 0, 0}}},{25462, {{36487, 1, 0, 0},{36487, 1, 1, 18}, {0, 0, 0, 0}}},{25463, {{36487, 1, 0, 0},{36487, 1, 1, 19}, {0, 0, 0, 0}}},{25463, {{36487, 1, 0, 0},{36487, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{25464, {{36488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25465, {{36489, 1, 0, 0},{36489, 1, 1, 17}, {0, 0, 0, 0}}},{25465, {{36489, 1, 0, 0},{36489, 1, 1, 17}, {0, 0, 0, 0}}},{25466, {{36489, 1, 0, 0},{36489, 1, 1, 18}, {0, 0, 0, 0}}},{25466, {{36489, 1, 0, 0},{36489, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25467, {{36490, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25468, {{36491, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25469, {{36492, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25470, {{36493, 1, 0, 0},{36493, 1, 1, 19}, {0, 0, 0, 0}}},{25470, {{36493, 1, 0, 0},{36493, 1, 1, 19}, {0, 0, 0, 0}}},{25471, {{36493, 1, 0, 0},{36493, 1, 1, 18}, {0, 0, 0, 0}}},{25471, {{36493, 1, 0, 0},{36493, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25472, {{36494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{25473, {{36496, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{25474, {{36497, 1, 0, 0},{36497, 1, 1, 18}, {0, 0, 0, 0}}},{25474, {{36497, 1, 0, 0},{36497, 1, 1, 18}, {0, 0, 0, 0}}},{25475, {{36497, 1, 0, 0},{36497, 1, 1, 19}, {0, 0, 0, 0}}},{25475, {{36497, 1, 0, 0},{36497, 1, 1, 19}, {0, 0, 0, 0}}},{25476, {{36497, 1, 0, 0},{36497, 1, 1, 20}, {0, 0, 0, 0}}},{25476, {{36497, 1, 0, 0},{36497, 1, 1, 20}, {0, 0, 0, 0}}},{25477, {{36497, 1, 0, 0},{36497, 1, 1, 21}, {0, 0, 0, 0}}},{25477, {{36497, 1, 0, 0},{36497, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{25478, {{36498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25479, {{36499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25480, {{36500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25481, {{36501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25482, {{36502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{25483, {{36504, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25484, {{36505, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{25485, {{36506, 1, 0, 0},{36506, 1, 1, 17}, {0, 0, 0, 0}}},{25485, {{36506, 1, 0, 0},{36506, 1, 1, 17}, {0, 0, 0, 0}}},{25486, {{36506, 1, 0, 0},{36506, 1, 1, 19}, {0, 0, 0, 0}}},{25486, {{36506, 1, 0, 0},{36506, 1, 1, 19}, {0, 0, 0, 0}}},{58819, {{36506, 1, 0, 0},{36506, 1, 1, 18}, {0, 0, 0, 0}}},{58819, {{36506, 1, 0, 0},{36506, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25487, {{36507, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{25488, {{36509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25489, {{36510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25490, {{36511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25491, {{36512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{25492, {{36513, 1, 0, 0},{36513, 1, 1, 18}, {0, 0, 0, 0}}},{25492, {{36513, 1, 0, 0},{36513, 1, 1, 18}, {0, 0, 0, 0}}},{25493, {{36513, 1, 0, 0},{36513, 1, 1, 19}, {0, 0, 0, 0}}},{25493, {{36513, 1, 0, 0},{36513, 1, 1, 19}, {0, 0, 0, 0}}},{25494, {{36513, 1, 0, 0},{36513, 1, 1, 20}, {0, 0, 0, 0}}},{25494, {{36513, 1, 0, 0},{36513, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{25495, {{36514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25496, {{36515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25497, {{36516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25498, {{36517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25499, {{36518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25500, {{36519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25501, {{36520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25502, {{36521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25503, {{36522, 1, 0, 0},{36522, 1, 1, 17}, {0, 0, 0, 0}}},{25503, {{36522, 1, 0, 0},{36522, 1, 1, 17}, {0, 0, 0, 0}}},{25504, {{36522, 1, 0, 0},{36522, 1, 1, 18}, {0, 0, 0, 0}}},{25504, {{36522, 1, 0, 0},{36522, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25505, {{36523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25506, {{36524, 1, 0, 0},{36524, 1, 1, 17}, {0, 0, 0, 0}}},{25506, {{36524, 1, 0, 0},{36524, 1, 1, 17}, {0, 0, 0, 0}}},{25507, {{36524, 1, 0, 0},{36524, 1, 1, 18}, {0, 0, 0, 0}}},{25507, {{36524, 1, 0, 0},{36524, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25508, {{36525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25509, {{36526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25510, {{36527, 1, 0, 0},{36527, 1, 1, 17}, {0, 0, 0, 0}}},{25510, {{36527, 1, 0, 0},{36527, 1, 1, 17}, {0, 0, 0, 0}}},{25511, {{36527, 1, 0, 0},{36527, 1, 1, 18}, {0, 0, 0, 0}}},{25511, {{36527, 1, 0, 0},{36527, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{25512, {{36528, 1, 0, 0},{36528, 1, 1, 17}, {0, 0, 0, 0}}},{25512, {{36528, 1, 0, 0},{36528, 1, 1, 17}, {0, 0, 0, 0}}},{25513, {{36528, 1, 0, 0},{36528, 1, 1, 18}, {0, 0, 0, 0}}},{25513, {{36528, 1, 0, 0},{36528, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25514, {{36529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25515, {{36530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25516, {{36531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25517, {{36532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25518, {{36533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25520, {{36534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25521, {{36535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25522, {{36536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25523, {{36537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25524, {{36538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25525, {{36539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{25526, {{36541, 1, 0, 0},{36541, 1, 1, 16}, {0, 0, 0, 0}}},{25526, {{36541, 1, 0, 0},{36541, 1, 1, 16}, {0, 0, 0, 0}}},{25527, {{36541, 1, 0, 0},{36541, 1, 1, 17}, {0, 0, 0, 0}}},{25527, {{36541, 1, 0, 0},{36541, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{25528, {{36542, 1, 0, 0},{36542, 1, 1, 17}, {0, 0, 0, 0}}},{25528, {{36542, 1, 0, 0},{36542, 1, 1, 17}, {0, 0, 0, 0}}},{25529, {{36542, 1, 0, 0},{36542, 1, 1, 18}, {0, 0, 0, 0}}},{25529, {{36542, 1, 0, 0},{36542, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25530, {{36543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25531, {{36544, 1, 0, 0},{36544, 1, 1, 17}, {0, 0, 0, 0}}},{25531, {{36544, 1, 0, 0},{36544, 1, 1, 17}, {0, 0, 0, 0}}},{25532, {{36544, 1, 0, 0},{36544, 1, 1, 18}, {0, 0, 0, 0}}},{25532, {{36544, 1, 0, 0},{36544, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25533, {{36545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25534, {{36546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25535, {{36547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25536, {{36548, 1, 0, 0},{36548, 1, 1, 17}, {0, 0, 0, 0}}},{25536, {{36548, 1, 0, 0},{36548, 1, 1, 17}, {0, 0, 0, 0}}},{25537, {{36548, 1, 0, 0},{36548, 1, 1, 18}, {0, 0, 0, 0}}},{25537, {{36548, 1, 0, 0},{36548, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{25538, {{36549, 1, 0, 0},{36549, 1, 1, 17}, {0, 0, 0, 0}}},{25538, {{36549, 1, 0, 0},{36549, 1, 1, 17}, {0, 0, 0, 0}}},{25539, {{36549, 1, 0, 0},{36549, 1, 1, 18}, {0, 0, 0, 0}}},{25539, {{36549, 1, 0, 0},{36549, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{25540, {{36550, 1, 0, 0},{36550, 1, 1, 17}, {0, 0, 0, 0}}},{25540, {{36550, 1, 0, 0},{36550, 1, 1, 17}, {0, 0, 0, 0}}},{25541, {{36550, 1, 0, 0},{36550, 1, 1, 18}, {0, 0, 0, 0}}},{25541, {{36550, 1, 0, 0},{36550, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25542, {{36551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25543, {{36552, 1, 0, 0},{36552, 1, 1, 17}, {0, 0, 0, 0}}},{25543, {{36552, 1, 0, 0},{36552, 1, 1, 17}, {0, 0, 0, 0}}},{25544, {{36552, 1, 0, 0},{36552, 1, 1, 18}, {0, 0, 0, 0}}},{25544, {{36552, 1, 0, 0},{36552, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25545, {{36553, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25546, {{36554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25547, {{36555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25548, {{36556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25550, {{36557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25551, {{36558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25552, {{36559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25553, {{36560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25554, {{36561, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25555, {{36562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25556, {{36563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25557, {{36564, 1, 0, 0},{36564, 1, 1, 18}, {0, 0, 0, 0}}},{25557, {{36564, 1, 0, 0},{36564, 1, 1, 18}, {0, 0, 0, 0}}},{25558, {{36564, 1, 0, 0},{36564, 1, 1, 19}, {0, 0, 0, 0}}},{25558, {{36564, 1, 0, 0},{36564, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{25559, {{36566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25560, {{36567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25561, {{36568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25562, {{36569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{25563, {{36571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25564, {{36572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25566, {{36573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25567, {{36574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25568, {{36575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25569, {{36576, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25570, {{36577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25571, {{36578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25572, {{36579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25573, {{36580, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25574, {{36581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25575, {{36582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25576, {{36583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25577, {{36584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25578, {{36585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25579, {{36586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25580, {{36587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25581, {{36588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25582, {{36589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25583, {{36590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25584, {{36591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25585, {{36592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25586, {{36593, 1, 0, 0},{36593, 1, 1, 18}, {0, 0, 0, 0}}},{25586, {{36593, 1, 0, 0},{36593, 1, 1, 18}, {0, 0, 0, 0}}},{25587, {{36593, 1, 0, 0},{36593, 1, 1, 17}, {0, 0, 0, 0}}},{25587, {{36593, 1, 0, 0},{36593, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{25588, {{36594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25589, {{36595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25590, {{36596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25591, {{36597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25592, {{36598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25593, {{36599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25594, {{36600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25595, {{36601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25596, {{36602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25597, {{36603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25598, {{36604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25599, {{36605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25600, {{36606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25601, {{36607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25602, {{36608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25603, {{36609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25604, {{36610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25605, {{36611, 1, 0, 0},{36611, 1, 1, 18}, {0, 0, 0, 0}}},{25605, {{36611, 1, 0, 0},{36611, 1, 1, 18}, {0, 0, 0, 0}}},{25606, {{36611, 1, 0, 0},{36611, 1, 1, 19}, {0, 0, 0, 0}}},{25606, {{36611, 1, 0, 0},{36611, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{25607, {{36612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25608, {{36613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25609, {{36614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25610, {{36615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25611, {{36616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25612, {{36617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25613, {{36618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25614, {{36619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25615, {{36620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25616, {{36621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25617, {{36622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25619, {{36623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25620, {{36624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25621, {{36625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25622, {{36626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25623, {{36627, 1, 0, 0},{36627, 1, 1, 18}, {0, 0, 0, 0}}},{25623, {{36627, 1, 0, 0},{36627, 1, 1, 18}, {0, 0, 0, 0}}},{25624, {{36627, 1, 0, 0},{36627, 1, 1, 19}, {0, 0, 0, 0}}},{25624, {{36627, 1, 0, 0},{36627, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{25625, {{36628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25626, {{36629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25627, {{36630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25629, {{36631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25630, {{36632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25631, {{36633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25632, {{36634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25633, {{36635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25634, {{36636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25635, {{36637, 1, 0, 0},{36637, 1, 1, 17}, {0, 0, 0, 0}}},{25635, {{36637, 1, 0, 0},{36637, 1, 1, 17}, {0, 0, 0, 0}}},{25636, {{36637, 1, 0, 0},{36637, 1, 1, 18}, {0, 0, 0, 0}}},{25636, {{36637, 1, 0, 0},{36637, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{25637, {{36638, 1, 0, 0},{36638, 1, 1, 17}, {0, 0, 0, 0}}},{25637, {{36638, 1, 0, 0},{36638, 1, 1, 17}, {0, 0, 0, 0}}},{25638, {{36638, 1, 0, 0},{36638, 1, 1, 18}, {0, 0, 0, 0}}},{25638, {{36638, 1, 0, 0},{36638, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25639, {{36639, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25640, {{36640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25641, {{36641, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25642, {{36642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25643, {{36643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25644, {{36644, 1, 0, 0},{36644, 1, 1, 17}, {0, 0, 0, 0}}},{25644, {{36644, 1, 0, 0},{36644, 1, 1, 17}, {0, 0, 0, 0}}},{25645, {{36644, 1, 0, 0},{36644, 1, 1, 16}, {0, 0, 0, 0}}},{25645, {{36644, 1, 0, 0},{36644, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{25646, {{36645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25647, {{36646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25648, {{36647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25649, {{36648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25650, {{36649, 1, 0, 0},{36649, 1, 1, 18}, {0, 0, 0, 0}}},{25650, {{36649, 1, 0, 0},{36649, 1, 1, 18}, {0, 0, 0, 0}}},{25651, {{36649, 1, 0, 0},{36649, 1, 1, 19}, {0, 0, 0, 0}}},{25651, {{36649, 1, 0, 0},{36649, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{25652, {{36650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25654, {{36651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25655, {{36652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25656, {{36653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25657, {{36654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25658, {{36655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25659, {{36656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25660, {{36657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25661, {{36658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25662, {{36659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25663, {{36660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25664, {{36661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25666, {{36662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25667, {{36663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25668, {{36664, 1, 0, 0},{36664, 1, 1, 19}, {0, 0, 0, 0}}},{25668, {{36664, 1, 0, 0},{36664, 1, 1, 19}, {0, 0, 0, 0}}},{25669, {{36664, 1, 0, 0},{36664, 1, 1, 18}, {0, 0, 0, 0}}},{25669, {{36664, 1, 0, 0},{36664, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25671, {{36665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25672, {{36666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25673, {{36667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25674, {{36668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25675, {{36669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25676, {{36670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25677, {{36671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25678, {{36672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25679, {{36673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25680, {{36674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25681, {{36675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{25682, {{36676, 1, 0, 0},{36676, 1, 1, 20}, {0, 0, 0, 0}}},{25682, {{36676, 1, 0, 0},{36676, 1, 1, 20}, {0, 0, 0, 0}}},{25683, {{36676, 1, 0, 0},{36676, 1, 1, 19}, {0, 0, 0, 0}}},{25683, {{36676, 1, 0, 0},{36676, 1, 1, 19}, {0, 0, 0, 0}}},{25684, {{36676, 1, 0, 0},{36676, 1, 1, 21}, {0, 0, 0, 0}}},{25684, {{36676, 1, 0, 0},{36676, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{25685, {{36677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25686, {{36678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25687, {{36679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25688, {{36680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25689, {{36681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25690, {{36682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25691, {{36683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25692, {{36684, 1, 0, 0},{36684, 1, 1, 17}, {0, 0, 0, 0}}},{25692, {{36684, 1, 0, 0},{36684, 1, 1, 17}, {0, 0, 0, 0}}},{25693, {{36684, 1, 0, 0},{36684, 1, 1, 18}, {0, 0, 0, 0}}},{25693, {{36684, 1, 0, 0},{36684, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25694, {{36685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25695, {{36686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25696, {{36687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25697, {{36688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25698, {{36689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25699, {{36690, 1, 0, 0},{36690, 1, 1, 17}, {0, 0, 0, 0}}},{25699, {{36690, 1, 0, 0},{36690, 1, 1, 17}, {0, 0, 0, 0}}},{25700, {{36690, 1, 0, 0},{36690, 1, 1, 18}, {0, 0, 0, 0}}},{25700, {{36690, 1, 0, 0},{36690, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25701, {{36691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25702, {{36692, 1, 0, 0},{36692, 1, 1, 18}, {0, 0, 0, 0}}},{25702, {{36692, 1, 0, 0},{36692, 1, 1, 18}, {0, 0, 0, 0}}},{25703, {{36692, 1, 0, 0},{36692, 1, 1, 17}, {0, 0, 0, 0}}},{25703, {{36692, 1, 0, 0},{36692, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{25704, {{36693, 1, 0, 0},{36693, 1, 1, 17}, {0, 0, 0, 0}}},{25704, {{36693, 1, 0, 0},{36693, 1, 1, 17}, {0, 0, 0, 0}}},{25705, {{36693, 1, 0, 0},{36693, 1, 1, 18}, {0, 0, 0, 0}}},{25705, {{36693, 1, 0, 0},{36693, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25706, {{36694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25707, {{36695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25708, {{36696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25709, {{36697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25710, {{36698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25711, {{36699, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25712, {{36700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25713, {{36701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25714, {{36702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25715, {{36703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25716, {{36704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25717, {{36705, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25718, {{36706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25719, {{36707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25721, {{36708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25722, {{36709, 1, 0, 0},{36709, 1, 1, 17}, {0, 0, 0, 0}}},{25722, {{36709, 1, 0, 0},{36709, 1, 1, 17}, {0, 0, 0, 0}}},{25723, {{36709, 1, 0, 0},{36709, 1, 1, 18}, {0, 0, 0, 0}}},{25723, {{36709, 1, 0, 0},{36709, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{25725, {{36763, 1, 0, 0},{36763, 1, 1, 17}, {0, 0, 0, 0}}},{25725, {{36763, 1, 0, 0},{36763, 1, 1, 17}, {0, 0, 0, 0}}},{25726, {{36763, 1, 0, 0},{36763, 1, 1, 18}, {0, 0, 0, 0}}},{25726, {{36763, 1, 0, 0},{36763, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25727, {{36764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25728, {{36765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25729, {{36766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25730, {{36767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25731, {{36768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25732, {{36769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25733, {{36770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25734, {{36771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25735, {{36772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25736, {{36773, 1, 0, 0},{36773, 1, 1, 17}, {0, 0, 0, 0}}},{25736, {{36773, 1, 0, 0},{36773, 1, 1, 17}, {0, 0, 0, 0}}},{58832, {{36773, 1, 0, 0},{36773, 1, 1, 18}, {0, 0, 0, 0}}},{58832, {{36773, 1, 0, 0},{36773, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25737, {{36774, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25738, {{36775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{25739, {{36776, 1, 0, 0},{36776, 1, 1, 17}, {0, 0, 0, 0}}},{25739, {{36776, 1, 0, 0},{36776, 1, 1, 17}, {0, 0, 0, 0}}},{58833, {{36776, 1, 0, 0},{36776, 1, 1, 19}, {0, 0, 0, 0}}},{58833, {{36776, 1, 0, 0},{36776, 1, 1, 19}, {0, 0, 0, 0}}},{59380, {{36776, 1, 0, 0},{36776, 1, 1, 18}, {0, 0, 0, 0}}},{59380, {{36776, 1, 0, 0},{36776, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{25740, {{36780, 1, 0, 0},{36780, 1, 1, 16}, {0, 0, 0, 0}}},{25740, {{36780, 1, 0, 0},{36780, 1, 1, 16}, {0, 0, 0, 0}}},{25741, {{36780, 1, 0, 0},{36780, 1, 1, 17}, {0, 0, 0, 0}}},{25741, {{36780, 1, 0, 0},{36780, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{25742, {{36781, 1, 0, 0},{36781, 1, 1, 17}, {0, 0, 0, 0}}},{25742, {{36781, 1, 0, 0},{36781, 1, 1, 17}, {0, 0, 0, 0}}},{25743, {{36781, 1, 0, 0},{36781, 1, 1, 18}, {0, 0, 0, 0}}},{25743, {{36781, 1, 0, 0},{36781, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25744, {{36782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25745, {{36783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25746, {{36784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{25747, {{36785, 1, 0, 0},{36785, 1, 1, 17}, {0, 0, 0, 0}}},{25747, {{36785, 1, 0, 0},{36785, 1, 1, 17}, {0, 0, 0, 0}}},{25748, {{36785, 1, 0, 0},{36785, 1, 1, 19}, {0, 0, 0, 0}}},{25748, {{36785, 1, 0, 0},{36785, 1, 1, 19}, {0, 0, 0, 0}}},{57352, {{36785, 1, 0, 0},{36785, 1, 1, 18}, {0, 0, 0, 0}}},{57352, {{36785, 1, 0, 0},{36785, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25749, {{36786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25750, {{36787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25751, {{36788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25752, {{36789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {7, {{25753, {{36790, 1, 0, 0},{36790, 1, 1, 19}, {0, 0, 0, 0}}},{25753, {{36790, 1, 0, 0},{36790, 1, 1, 19}, {0, 0, 0, 0}}},{30302, {{36790, 1, 0, 0},{36790, 1, 1, 0},{36790, 1, 1, 18},{64102, 1, 0, 0}, {0, 0, 0, 0}}},{30302, {{36790, 1, 0, 0},{36790, 1, 1, 0},{36790, 1, 1, 18},{64102, 1, 0, 0}, {0, 0, 0, 0}}},{30302, {{36790, 1, 0, 0},{36790, 1, 1, 0},{36790, 1, 1, 18},{64102, 1, 0, 0}, {0, 0, 0, 0}}},{58837, {{36790, 1, 0, 0},{36790, 1, 1, 20}, {0, 0, 0, 0}}},{58837, {{36790, 1, 0, 0},{36790, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{25754, {{36791, 1, 0, 0},{36791, 1, 1, 18}, {0, 0, 0, 0}}},{25754, {{36791, 1, 0, 0},{36791, 1, 1, 18}, {0, 0, 0, 0}}},{25755, {{36791, 1, 0, 0},{36791, 1, 1, 17}, {0, 0, 0, 0}}},{25755, {{36791, 1, 0, 0},{36791, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{25756, {{36792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25757, {{36793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25758, {{36794, 1, 0, 0},{36794, 1, 1, 17}, {0, 0, 0, 0}}},{25758, {{36794, 1, 0, 0},{36794, 1, 1, 17}, {0, 0, 0, 0}}},{25759, {{36794, 1, 0, 0},{36794, 1, 1, 18}, {0, 0, 0, 0}}},{25759, {{36794, 1, 0, 0},{36794, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{25760, {{36795, 1, 0, 0},{36795, 1, 1, 18}, {0, 0, 0, 0}}},{25760, {{36795, 1, 0, 0},{36795, 1, 1, 18}, {0, 0, 0, 0}}},{25761, {{36795, 1, 0, 0},{36795, 1, 1, 19}, {0, 0, 0, 0}}},{25761, {{36795, 1, 0, 0},{36795, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{25762, {{36796, 1, 0, 0},{36796, 1, 1, 18}, {0, 0, 0, 0}}},{25762, {{36796, 1, 0, 0},{36796, 1, 1, 18}, {0, 0, 0, 0}}},{25763, {{36796, 1, 0, 0},{36796, 1, 1, 19}, {0, 0, 0, 0}}},{25763, {{36796, 1, 0, 0},{36796, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{25765, {{36798, 1, 0, 0},{36798, 1, 1, 18}, {0, 0, 0, 0}}},{25765, {{36798, 1, 0, 0},{36798, 1, 1, 18}, {0, 0, 0, 0}}},{25766, {{36798, 1, 0, 0},{36798, 1, 1, 17}, {0, 0, 0, 0}}},{25766, {{36798, 1, 0, 0},{36798, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{25767, {{36799, 1, 0, 0},{36799, 1, 1, 18}, {0, 0, 0, 0}}},{25767, {{36799, 1, 0, 0},{36799, 1, 1, 18}, {0, 0, 0, 0}}},{25768, {{36799, 1, 0, 0},{36799, 1, 1, 19}, {0, 0, 0, 0}}},{25768, {{36799, 1, 0, 0},{36799, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{25769, {{36800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25770, {{36801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25772, {{36802, 1, 0, 0},{36802, 1, 1, 18}, {0, 0, 0, 0}}},{25772, {{36802, 1, 0, 0},{36802, 1, 1, 18}, {0, 0, 0, 0}}},{25773, {{36802, 1, 0, 0},{36802, 1, 1, 19}, {0, 0, 0, 0}}},{25773, {{36802, 1, 0, 0},{36802, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {6, {{25774, {{36803, 1, 0, 0},{36803, 1, 1, 16}, {0, 0, 0, 0}}},{25774, {{36803, 1, 0, 0},{36803, 1, 1, 16}, {0, 0, 0, 0}}},{25775, {{36803, 1, 0, 0},{36803, 1, 1, 17}, {0, 0, 0, 0}}},{25775, {{36803, 1, 0, 0},{36803, 1, 1, 17}, {0, 0, 0, 0}}},{25776, {{36803, 1, 0, 0},{36803, 1, 1, 18}, {0, 0, 0, 0}}},{25776, {{36803, 1, 0, 0},{36803, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{25777, {{36804, 1, 0, 0},{36804, 1, 1, 18}, {0, 0, 0, 0}}},{25777, {{36804, 1, 0, 0},{36804, 1, 1, 18}, {0, 0, 0, 0}}},{25778, {{36804, 1, 0, 0},{36804, 1, 1, 19}, {0, 0, 0, 0}}},{25778, {{36804, 1, 0, 0},{36804, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{25779, {{36805, 1, 0, 0},{36805, 1, 1, 19}, {0, 0, 0, 0}}},{25779, {{36805, 1, 0, 0},{36805, 1, 1, 19}, {0, 0, 0, 0}}},{25780, {{36805, 1, 0, 0},{36805, 1, 1, 20}, {0, 0, 0, 0}}},{25780, {{36805, 1, 0, 0},{36805, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{25781, {{36806, 1, 0, 0},{36806, 1, 1, 19}, {0, 0, 0, 0}}},{25781, {{36806, 1, 0, 0},{36806, 1, 1, 19}, {0, 0, 0, 0}}},{25782, {{36806, 1, 0, 0},{36806, 1, 1, 18}, {0, 0, 0, 0}}},{25782, {{36806, 1, 0, 0},{36806, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{25783, {{36807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{25786, {{36810, 1, 0, 0},{36810, 1, 1, 18}, {0, 0, 0, 0}}},{25786, {{36810, 1, 0, 0},{36810, 1, 1, 18}, {0, 0, 0, 0}}},{25787, {{36810, 1, 0, 0},{36810, 1, 1, 17}, {0, 0, 0, 0}}},{25787, {{36810, 1, 0, 0},{36810, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{25788, {{36811, 1, 0, 0},{36811, 1, 1, 18}, {0, 0, 0, 0}}},{25788, {{36811, 1, 0, 0},{36811, 1, 1, 18}, {0, 0, 0, 0}}},{25789, {{36811, 1, 0, 0},{36811, 1, 1, 17}, {0, 0, 0, 0}}},{25789, {{36811, 1, 0, 0},{36811, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{25790, {{36812, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25791, {{36813, 1, 0, 0},{36813, 1, 1, 18}, {0, 0, 0, 0}}},{25791, {{36813, 1, 0, 0},{36813, 1, 1, 18}, {0, 0, 0, 0}}},{25792, {{36813, 1, 0, 0},{36813, 1, 1, 17}, {0, 0, 0, 0}}},{25792, {{36813, 1, 0, 0},{36813, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{25793, {{36814, 1, 0, 0},{36814, 1, 1, 18}, {0, 0, 0, 0}}},{25793, {{36814, 1, 0, 0},{36814, 1, 1, 18}, {0, 0, 0, 0}}},{25794, {{36814, 1, 0, 0},{36814, 1, 1, 19}, {0, 0, 0, 0}}},{25794, {{36814, 1, 0, 0},{36814, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{25795, {{36815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25796, {{36816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25798, {{36817, 1, 0, 0},{36817, 1, 1, 18}, {0, 0, 0, 0}}},{25798, {{36817, 1, 0, 0},{36817, 1, 1, 18}, {0, 0, 0, 0}}},{25799, {{36817, 1, 0, 0},{36817, 1, 1, 19}, {0, 0, 0, 0}}},{25799, {{36817, 1, 0, 0},{36817, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{25800, {{36818, 1, 0, 0},{36818, 1, 1, 18}, {0, 0, 0, 0}}},{25800, {{36818, 1, 0, 0},{36818, 1, 1, 18}, {0, 0, 0, 0}}},{25801, {{36818, 1, 0, 0},{36818, 1, 1, 17}, {0, 0, 0, 0}}},{25801, {{36818, 1, 0, 0},{36818, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {6, {{25802, {{36819, 1, 0, 0},{36819, 1, 1, 19}, {0, 0, 0, 0}}},{25802, {{36819, 1, 0, 0},{36819, 1, 1, 19}, {0, 0, 0, 0}}},{25803, {{36819, 1, 0, 0},{36819, 1, 1, 18}, {0, 0, 0, 0}}},{25803, {{36819, 1, 0, 0},{36819, 1, 1, 18}, {0, 0, 0, 0}}},{25804, {{36819, 1, 0, 0},{36819, 1, 1, 17}, {0, 0, 0, 0}}},{25804, {{36819, 1, 0, 0},{36819, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{25805, {{36820, 1, 0, 0},{36820, 1, 1, 19}, {0, 0, 0, 0}}},{25805, {{36820, 1, 0, 0},{36820, 1, 1, 19}, {0, 0, 0, 0}}},{25806, {{36820, 1, 0, 0},{36820, 1, 1, 20}, {0, 0, 0, 0}}},{25806, {{36820, 1, 0, 0},{36820, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{25807, {{36821, 1, 0, 0},{36821, 1, 1, 18}, {0, 0, 0, 0}}},{25807, {{36821, 1, 0, 0},{36821, 1, 1, 18}, {0, 0, 0, 0}}},{25808, {{36821, 1, 0, 0},{36821, 1, 1, 17}, {0, 0, 0, 0}}},{25808, {{36821, 1, 0, 0},{36821, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{25809, {{36823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25810, {{36824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{25811, {{36825, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25812, {{36826, 1, 0, 0},{36826, 1, 1, 18}, {0, 0, 0, 0}}},{25812, {{36826, 1, 0, 0},{36826, 1, 1, 18}, {0, 0, 0, 0}}},{25813, {{36826, 1, 0, 0},{36826, 1, 1, 17}, {0, 0, 0, 0}}},{25813, {{36826, 1, 0, 0},{36826, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{25817, {{36832, 1, 0, 0},{36832, 1, 1, 18}, {0, 0, 0, 0}}},{25817, {{36832, 1, 0, 0},{36832, 1, 1, 18}, {0, 0, 0, 0}}},{25818, {{36832, 1, 0, 0},{36832, 1, 1, 17}, {0, 0, 0, 0}}},{25818, {{36832, 1, 0, 0},{36832, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{25819, {{36833, 1, 0, 0},{36833, 1, 1, 16}, {0, 0, 0, 0}}},{25819, {{36833, 1, 0, 0},{36833, 1, 1, 16}, {0, 0, 0, 0}}},{25820, {{36833, 1, 0, 0},{36833, 1, 1, 17}, {0, 0, 0, 0}}},{25820, {{36833, 1, 0, 0},{36833, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{25821, {{36834, 1, 0, 0},{36834, 1, 1, 18}, {0, 0, 0, 0}}},{25821, {{36834, 1, 0, 0},{36834, 1, 1, 18}, {0, 0, 0, 0}}},{25822, {{36834, 1, 0, 0},{36834, 1, 1, 17}, {0, 0, 0, 0}}},{25822, {{36834, 1, 0, 0},{36834, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{25823, {{36835, 1, 0, 0},{36835, 1, 1, 18}, {0, 0, 0, 0}}},{25823, {{36835, 1, 0, 0},{36835, 1, 1, 18}, {0, 0, 0, 0}}},{25824, {{36835, 1, 0, 0},{36835, 1, 1, 17}, {0, 0, 0, 0}}},{25824, {{36835, 1, 0, 0},{36835, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{25825, {{36836, 1, 0, 0},{36836, 1, 1, 18}, {0, 0, 0, 0}}},{25825, {{36836, 1, 0, 0},{36836, 1, 1, 18}, {0, 0, 0, 0}}},{25826, {{36836, 1, 0, 0},{36836, 1, 1, 17}, {0, 0, 0, 0}}},{25826, {{36836, 1, 0, 0},{36836, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{25827, {{36837, 1, 0, 0},{36837, 1, 1, 18}, {0, 0, 0, 0}}},{25827, {{36837, 1, 0, 0},{36837, 1, 1, 18}, {0, 0, 0, 0}}},{25828, {{36837, 1, 0, 0},{36837, 1, 1, 17}, {0, 0, 0, 0}}},{25828, {{36837, 1, 0, 0},{36837, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{25829, {{36838, 1, 0, 0},{36838, 1, 1, 18}, {0, 0, 0, 0}}},{25829, {{36838, 1, 0, 0},{36838, 1, 1, 18}, {0, 0, 0, 0}}},{25830, {{36838, 1, 0, 0},{36838, 1, 1, 19}, {0, 0, 0, 0}}},{25830, {{36838, 1, 0, 0},{36838, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{25831, {{36839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25833, {{36840, 1, 0, 0},{36840, 1, 1, 18}, {0, 0, 0, 0}}},{25833, {{36840, 1, 0, 0},{36840, 1, 1, 18}, {0, 0, 0, 0}}},{25834, {{36840, 1, 0, 0},{36840, 1, 1, 17}, {0, 0, 0, 0}}},{25834, {{36840, 1, 0, 0},{36840, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {6, {{25835, {{36841, 1, 0, 0},{36841, 1, 1, 18}, {0, 0, 0, 0}}},{25835, {{36841, 1, 0, 0},{36841, 1, 1, 18}, {0, 0, 0, 0}}},{25836, {{36841, 1, 0, 0},{36841, 1, 1, 19}, {0, 0, 0, 0}}},{25836, {{36841, 1, 0, 0},{36841, 1, 1, 19}, {0, 0, 0, 0}}},{25837, {{36841, 1, 0, 0},{36841, 1, 1, 20}, {0, 0, 0, 0}}},{25837, {{36841, 1, 0, 0},{36841, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{25838, {{36842, 1, 0, 0},{36842, 1, 1, 18}, {0, 0, 0, 0}}},{25838, {{36842, 1, 0, 0},{36842, 1, 1, 18}, {0, 0, 0, 0}}},{25839, {{36842, 1, 0, 0},{36842, 1, 1, 19}, {0, 0, 0, 0}}},{25839, {{36842, 1, 0, 0},{36842, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{25840, {{36843, 1, 0, 0},{36843, 1, 1, 18}, {0, 0, 0, 0}}},{25840, {{36843, 1, 0, 0},{36843, 1, 1, 18}, {0, 0, 0, 0}}},{25841, {{36843, 1, 0, 0},{36843, 1, 1, 19}, {0, 0, 0, 0}}},{25841, {{36843, 1, 0, 0},{36843, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{25842, {{36844, 1, 0, 0},{36844, 1, 1, 17}, {0, 0, 0, 0}}},{25842, {{36844, 1, 0, 0},{36844, 1, 1, 17}, {0, 0, 0, 0}}},{25843, {{36844, 1, 0, 0},{36844, 1, 1, 16}, {0, 0, 0, 0}}},{25843, {{36844, 1, 0, 0},{36844, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {4, {{25844, {{36845, 1, 0, 0},{36845, 1, 1, 18}, {0, 0, 0, 0}}},{25844, {{36845, 1, 0, 0},{36845, 1, 1, 18}, {0, 0, 0, 0}}},{25845, {{36845, 1, 0, 0},{36845, 1, 1, 19}, {0, 0, 0, 0}}},{25845, {{36845, 1, 0, 0},{36845, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{25846, {{36846, 1, 0, 0},{36846, 1, 1, 18}, {0, 0, 0, 0}}},{25846, {{36846, 1, 0, 0},{36846, 1, 1, 18}, {0, 0, 0, 0}}},{25847, {{36846, 1, 0, 0},{36846, 1, 1, 17}, {0, 0, 0, 0}}},{25847, {{36846, 1, 0, 0},{36846, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {6, {{25848, {{36847, 1, 0, 0},{36847, 1, 1, 18}, {0, 0, 0, 0}}},{25848, {{36847, 1, 0, 0},{36847, 1, 1, 18}, {0, 0, 0, 0}}},{25849, {{36847, 1, 0, 0},{36847, 1, 1, 20}, {0, 0, 0, 0}}},{25849, {{36847, 1, 0, 0},{36847, 1, 1, 20}, {0, 0, 0, 0}}},{25850, {{36847, 1, 0, 0},{36847, 1, 1, 19}, {0, 0, 0, 0}}},{25850, {{36847, 1, 0, 0},{36847, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {6, {{25851, {{36848, 1, 0, 0},{36848, 1, 1, 18}, {0, 0, 0, 0}}},{25851, {{36848, 1, 0, 0},{36848, 1, 1, 18}, {0, 0, 0, 0}}},{25852, {{36848, 1, 0, 0},{36848, 1, 1, 19}, {0, 0, 0, 0}}},{25852, {{36848, 1, 0, 0},{36848, 1, 1, 19}, {0, 0, 0, 0}}},{58845, {{36848, 1, 0, 0},{36848, 1, 1, 20}, {0, 0, 0, 0}}},{58845, {{36848, 1, 0, 0},{36848, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{25853, {{36849, 1, 0, 0},{36849, 1, 1, 18}, {0, 0, 0, 0}}},{25853, {{36849, 1, 0, 0},{36849, 1, 1, 18}, {0, 0, 0, 0}}},{25854, {{36849, 1, 0, 0},{36849, 1, 1, 17}, {0, 0, 0, 0}}},{25854, {{36849, 1, 0, 0},{36849, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{25855, {{36850, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58848, {{36851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25856, {{36852, 1, 0, 0},{36852, 1, 1, 18}, {0, 0, 0, 0}}},{25856, {{36852, 1, 0, 0},{36852, 1, 1, 18}, {0, 0, 0, 0}}},{25857, {{36852, 1, 0, 0},{36852, 1, 1, 17}, {0, 0, 0, 0}}},{25857, {{36852, 1, 0, 0},{36852, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{25858, {{36853, 1, 0, 0},{36853, 1, 1, 18}, {0, 0, 0, 0}}},{25858, {{36853, 1, 0, 0},{36853, 1, 1, 18}, {0, 0, 0, 0}}},{25859, {{36853, 1, 0, 0},{36853, 1, 1, 17}, {0, 0, 0, 0}}},{25859, {{36853, 1, 0, 0},{36853, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{25860, {{36854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25861, {{36855, 1, 0, 0},{36855, 1, 1, 18}, {0, 0, 0, 0}}},{25861, {{36855, 1, 0, 0},{36855, 1, 1, 18}, {0, 0, 0, 0}}},{25862, {{36855, 1, 0, 0},{36855, 1, 1, 19}, {0, 0, 0, 0}}},{25862, {{36855, 1, 0, 0},{36855, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{25863, {{36856, 1, 0, 0},{36856, 1, 1, 17}, {0, 0, 0, 0}}},{25863, {{36856, 1, 0, 0},{36856, 1, 1, 17}, {0, 0, 0, 0}}},{25864, {{36856, 1, 0, 0},{36856, 1, 1, 18}, {0, 0, 0, 0}}},{25864, {{36856, 1, 0, 0},{36856, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{25865, {{36857, 1, 0, 0},{36857, 1, 1, 18}, {0, 0, 0, 0}}},{25865, {{36857, 1, 0, 0},{36857, 1, 1, 18}, {0, 0, 0, 0}}},{25866, {{36857, 1, 0, 0},{36857, 1, 1, 17}, {0, 0, 0, 0}}},{25866, {{36857, 1, 0, 0},{36857, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{25867, {{36858, 1, 0, 0},{36858, 1, 1, 19}, {0, 0, 0, 0}}},{25867, {{36858, 1, 0, 0},{36858, 1, 1, 19}, {0, 0, 0, 0}}},{25868, {{36858, 1, 0, 0},{36858, 1, 1, 18}, {0, 0, 0, 0}}},{25868, {{36858, 1, 0, 0},{36858, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{25869, {{36859, 1, 0, 0},{36859, 1, 1, 18}, {0, 0, 0, 0}}},{25869, {{36859, 1, 0, 0},{36859, 1, 1, 18}, {0, 0, 0, 0}}},{25870, {{36859, 1, 0, 0},{36859, 1, 1, 17}, {0, 0, 0, 0}}},{25870, {{36859, 1, 0, 0},{36859, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{25871, {{36861, 1, 0, 0},{36861, 1, 1, 18}, {0, 0, 0, 0}}},{25871, {{36861, 1, 0, 0},{36861, 1, 1, 18}, {0, 0, 0, 0}}},{25872, {{36861, 1, 0, 0},{36861, 1, 1, 19}, {0, 0, 0, 0}}},{25872, {{36861, 1, 0, 0},{36861, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{25874, {{36862, 1, 0, 0},{36862, 1, 1, 18}, {0, 0, 0, 0}}},{25874, {{36862, 1, 0, 0},{36862, 1, 1, 18}, {0, 0, 0, 0}}},{25875, {{36862, 1, 0, 0},{36862, 1, 1, 17}, {0, 0, 0, 0}}},{25875, {{36862, 1, 0, 0},{36862, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{25876, {{36863, 1, 0, 0},{36863, 1, 1, 16}, {0, 0, 0, 0}}},{25876, {{36863, 1, 0, 0},{36863, 1, 1, 16}, {0, 0, 0, 0}}},{25877, {{36863, 1, 0, 0},{36863, 1, 1, 17}, {0, 0, 0, 0}}},{25877, {{36863, 1, 0, 0},{36863, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{25878, {{36864, 1, 0, 0},{36864, 1, 1, 18}, {0, 0, 0, 0}}},{25878, {{36864, 1, 0, 0},{36864, 1, 1, 18}, {0, 0, 0, 0}}},{25879, {{36864, 1, 0, 0},{36864, 1, 1, 19}, {0, 0, 0, 0}}},{25879, {{36864, 1, 0, 0},{36864, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{25880, {{36865, 1, 0, 0},{36865, 1, 1, 18}, {0, 0, 0, 0}}},{25880, {{36865, 1, 0, 0},{36865, 1, 1, 18}, {0, 0, 0, 0}}},{25881, {{36865, 1, 0, 0},{36865, 1, 1, 19}, {0, 0, 0, 0}}},{25881, {{36865, 1, 0, 0},{36865, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{25882, {{36866, 1, 0, 0},{36866, 1, 1, 18}, {0, 0, 0, 0}}},{25882, {{36866, 1, 0, 0},{36866, 1, 1, 18}, {0, 0, 0, 0}}},{25883, {{36866, 1, 0, 0},{36866, 1, 1, 17}, {0, 0, 0, 0}}},{25883, {{36866, 1, 0, 0},{36866, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{25884, {{36867, 1, 0, 0},{36867, 1, 1, 19}, {0, 0, 0, 0}}},{25884, {{36867, 1, 0, 0},{36867, 1, 1, 19}, {0, 0, 0, 0}}},{25885, {{36867, 1, 0, 0},{36867, 1, 1, 20}, {0, 0, 0, 0}}},{25885, {{36867, 1, 0, 0},{36867, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {6, {{25886, {{36868, 1, 0, 0},{36868, 1, 1, 18}, {0, 0, 0, 0}}},{25886, {{36868, 1, 0, 0},{36868, 1, 1, 18}, {0, 0, 0, 0}}},{25887, {{36868, 1, 0, 0},{36868, 1, 1, 19}, {0, 0, 0, 0}}},{25887, {{36868, 1, 0, 0},{36868, 1, 1, 19}, {0, 0, 0, 0}}},{25888, {{36868, 1, 0, 0},{36868, 1, 1, 17}, {0, 0, 0, 0}}},{25888, {{36868, 1, 0, 0},{36868, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{25889, {{36869, 1, 0, 0},{36869, 1, 1, 18}, {0, 0, 0, 0}}},{25889, {{36869, 1, 0, 0},{36869, 1, 1, 18}, {0, 0, 0, 0}}},{25890, {{36869, 1, 0, 0},{36869, 1, 1, 17}, {0, 0, 0, 0}}},{25890, {{36869, 1, 0, 0},{36869, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{25891, {{36870, 1, 0, 0},{36870, 1, 1, 18}, {0, 0, 0, 0}}},{25891, {{36870, 1, 0, 0},{36870, 1, 1, 18}, {0, 0, 0, 0}}},{25892, {{36870, 1, 0, 0},{36870, 1, 1, 19}, {0, 0, 0, 0}}},{25892, {{36870, 1, 0, 0},{36870, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{25893, {{36871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25894, {{36872, 1, 0, 0},{36872, 1, 1, 17}, {0, 0, 0, 0}}},{25894, {{36872, 1, 0, 0},{36872, 1, 1, 17}, {0, 0, 0, 0}}},{25895, {{36872, 1, 0, 0},{36872, 1, 1, 18}, {0, 0, 0, 0}}},{25895, {{36872, 1, 0, 0},{36872, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{25897, {{36875, 1, 0, 0},{36875, 1, 1, 18}, {0, 0, 0, 0}}},{25897, {{36875, 1, 0, 0},{36875, 1, 1, 18}, {0, 0, 0, 0}}},{25898, {{36875, 1, 0, 0},{36875, 1, 1, 17}, {0, 0, 0, 0}}},{25898, {{36875, 1, 0, 0},{36875, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{25899, {{36876, 1, 0, 0},{36876, 1, 1, 18}, {0, 0, 0, 0}}},{25899, {{36876, 1, 0, 0},{36876, 1, 1, 18}, {0, 0, 0, 0}}},{25900, {{36876, 1, 0, 0},{36876, 1, 1, 17}, {0, 0, 0, 0}}},{25900, {{36876, 1, 0, 0},{36876, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {6, {{25901, {{36877, 1, 0, 0},{36877, 1, 1, 18}, {0, 0, 0, 0}}},{25901, {{36877, 1, 0, 0},{36877, 1, 1, 18}, {0, 0, 0, 0}}},{25902, {{36877, 1, 0, 0},{36877, 1, 1, 17}, {0, 0, 0, 0}}},{25902, {{36877, 1, 0, 0},{36877, 1, 1, 17}, {0, 0, 0, 0}}},{25903, {{36877, 1, 0, 0},{36877, 1, 1, 19}, {0, 0, 0, 0}}},{25903, {{36877, 1, 0, 0},{36877, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {8, {{25904, {{36878, 1, 0, 0},{36878, 1, 1, 19}, {0, 0, 0, 0}}},{25904, {{36878, 1, 0, 0},{36878, 1, 1, 19}, {0, 0, 0, 0}}},{25905, {{36878, 1, 0, 0},{36878, 1, 1, 18}, {0, 0, 0, 0}}},{25905, {{36878, 1, 0, 0},{36878, 1, 1, 18}, {0, 0, 0, 0}}},{25906, {{36878, 1, 0, 0},{36878, 1, 1, 21}, {0, 0, 0, 0}}},{25906, {{36878, 1, 0, 0},{36878, 1, 1, 21}, {0, 0, 0, 0}}},{25907, {{36878, 1, 0, 0},{36878, 1, 1, 20}, {0, 0, 0, 0}}},{25907, {{36878, 1, 0, 0},{36878, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{25908, {{36879, 1, 0, 0},{36879, 1, 1, 18}, {0, 0, 0, 0}}},{25908, {{36879, 1, 0, 0},{36879, 1, 1, 18}, {0, 0, 0, 0}}},{25909, {{36879, 1, 0, 0},{36879, 1, 1, 19}, {0, 0, 0, 0}}},{25909, {{36879, 1, 0, 0},{36879, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {6, {{25910, {{36880, 1, 0, 0},{36880, 1, 1, 18}, {0, 0, 0, 0}}},{25910, {{36880, 1, 0, 0},{36880, 1, 1, 18}, {0, 0, 0, 0}}},{25911, {{36880, 1, 0, 0},{36880, 1, 1, 19}, {0, 0, 0, 0}}},{25911, {{36880, 1, 0, 0},{36880, 1, 1, 19}, {0, 0, 0, 0}}},{25912, {{36880, 1, 0, 0},{36880, 1, 1, 20}, {0, 0, 0, 0}}},{25912, {{36880, 1, 0, 0},{36880, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {6, {{25913, {{36881, 1, 0, 0},{36881, 1, 1, 18}, {0, 0, 0, 0}}},{25913, {{36881, 1, 0, 0},{36881, 1, 1, 18}, {0, 0, 0, 0}}},{25914, {{36881, 1, 0, 0},{36881, 1, 1, 17}, {0, 0, 0, 0}}},{25914, {{36881, 1, 0, 0},{36881, 1, 1, 17}, {0, 0, 0, 0}}},{25915, {{36881, 1, 0, 0},{36881, 1, 1, 19}, {0, 0, 0, 0}}},{25915, {{36881, 1, 0, 0},{36881, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{25916, {{36882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25917, {{36883, 1, 0, 0},{36883, 1, 1, 17}, {0, 0, 0, 0}}},{25917, {{36883, 1, 0, 0},{36883, 1, 1, 17}, {0, 0, 0, 0}}},{25918, {{36883, 1, 0, 0},{36883, 1, 1, 18}, {0, 0, 0, 0}}},{25918, {{36883, 1, 0, 0},{36883, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{25919, {{36884, 1, 0, 0},{36884, 1, 1, 18}, {0, 0, 0, 0}}},{25919, {{36884, 1, 0, 0},{36884, 1, 1, 18}, {0, 0, 0, 0}}},{25920, {{36884, 1, 0, 0},{36884, 1, 1, 19}, {0, 0, 0, 0}}},{25920, {{36884, 1, 0, 0},{36884, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{25921, {{36885, 1, 0, 0},{36885, 1, 1, 18}, {0, 0, 0, 0}}},{25921, {{36885, 1, 0, 0},{36885, 1, 1, 18}, {0, 0, 0, 0}}},{25922, {{36885, 1, 0, 0},{36885, 1, 1, 17}, {0, 0, 0, 0}}},{25922, {{36885, 1, 0, 0},{36885, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {6, {{25923, {{36886, 1, 0, 0},{36886, 1, 1, 18}, {0, 0, 0, 0}}},{25923, {{36886, 1, 0, 0},{36886, 1, 1, 18}, {0, 0, 0, 0}}},{25924, {{36886, 1, 0, 0},{36886, 1, 1, 17}, {0, 0, 0, 0}}},{25924, {{36886, 1, 0, 0},{36886, 1, 1, 17}, {0, 0, 0, 0}}},{25925, {{36886, 1, 0, 0},{36886, 1, 1, 19}, {0, 0, 0, 0}}},{25925, {{36886, 1, 0, 0},{36886, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{25926, {{36887, 1, 0, 0},{36887, 1, 1, 18}, {0, 0, 0, 0}}},{25926, {{36887, 1, 0, 0},{36887, 1, 1, 18}, {0, 0, 0, 0}}},{25927, {{36887, 1, 0, 0},{36887, 1, 1, 19}, {0, 0, 0, 0}}},{25927, {{36887, 1, 0, 0},{36887, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{25928, {{36888, 1, 0, 0},{36888, 1, 1, 17}, {0, 0, 0, 0}}},{25928, {{36888, 1, 0, 0},{36888, 1, 1, 17}, {0, 0, 0, 0}}},{25929, {{36888, 1, 0, 0},{36888, 1, 1, 18}, {0, 0, 0, 0}}},{25929, {{36888, 1, 0, 0},{36888, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{25930, {{36889, 1, 0, 0},{36889, 1, 1, 18}, {0, 0, 0, 0}}},{25930, {{36889, 1, 0, 0},{36889, 1, 1, 18}, {0, 0, 0, 0}}},{25931, {{36889, 1, 0, 0},{36889, 1, 1, 19}, {0, 0, 0, 0}}},{25931, {{36889, 1, 0, 0},{36889, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{25932, {{36890, 1, 0, 0},{36890, 1, 1, 18}, {0, 0, 0, 0}}},{25932, {{36890, 1, 0, 0},{36890, 1, 1, 18}, {0, 0, 0, 0}}},{25933, {{36890, 1, 0, 0},{36890, 1, 1, 19}, {0, 0, 0, 0}}},{25933, {{36890, 1, 0, 0},{36890, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{25934, {{36891, 1, 0, 0},{36891, 1, 1, 18}, {0, 0, 0, 0}}},{25934, {{36891, 1, 0, 0},{36891, 1, 1, 18}, {0, 0, 0, 0}}},{25935, {{36891, 1, 0, 0},{36891, 1, 1, 17}, {0, 0, 0, 0}}},{25935, {{36891, 1, 0, 0},{36891, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{25936, {{36892, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25937, {{36893, 1, 0, 0},{36893, 1, 1, 18}, {0, 0, 0, 0}}},{25937, {{36893, 1, 0, 0},{36893, 1, 1, 18}, {0, 0, 0, 0}}},{25938, {{36893, 1, 0, 0},{36893, 1, 1, 19}, {0, 0, 0, 0}}},{25938, {{36893, 1, 0, 0},{36893, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {8, {{25939, {{36894, 1, 0, 0},{36894, 1, 1, 19}, {0, 0, 0, 0}}},{25939, {{36894, 1, 0, 0},{36894, 1, 1, 19}, {0, 0, 0, 0}}},{25940, {{36894, 1, 0, 0},{36894, 1, 1, 20}, {0, 0, 0, 0}}},{25940, {{36894, 1, 0, 0},{36894, 1, 1, 20}, {0, 0, 0, 0}}},{25941, {{36894, 1, 0, 0},{36894, 1, 1, 18}, {0, 0, 0, 0}}},{25941, {{36894, 1, 0, 0},{36894, 1, 1, 18}, {0, 0, 0, 0}}},{25942, {{36894, 1, 0, 0},{36894, 1, 1, 21}, {0, 0, 0, 0}}},{25942, {{36894, 1, 0, 0},{36894, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {4, {{25944, {{36895, 1, 0, 0},{36895, 1, 1, 18}, {0, 0, 0, 0}}},{25944, {{36895, 1, 0, 0},{36895, 1, 1, 18}, {0, 0, 0, 0}}},{25945, {{36895, 1, 0, 0},{36895, 1, 1, 19}, {0, 0, 0, 0}}},{25945, {{36895, 1, 0, 0},{36895, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{25946, {{36896, 1, 0, 0},{36896, 1, 1, 18}, {0, 0, 0, 0}}},{25946, {{36896, 1, 0, 0},{36896, 1, 1, 18}, {0, 0, 0, 0}}},{25947, {{36896, 1, 0, 0},{36896, 1, 1, 19}, {0, 0, 0, 0}}},{25947, {{36896, 1, 0, 0},{36896, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{25949, {{36897, 1, 0, 0},{36897, 1, 1, 18}, {0, 0, 0, 0}}},{25949, {{36897, 1, 0, 0},{36897, 1, 1, 18}, {0, 0, 0, 0}}},{25950, {{36897, 1, 0, 0},{36897, 1, 1, 17}, {0, 0, 0, 0}}},{25950, {{36897, 1, 0, 0},{36897, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{25951, {{36898, 1, 0, 0},{36898, 1, 1, 19}, {0, 0, 0, 0}}},{25951, {{36898, 1, 0, 0},{36898, 1, 1, 19}, {0, 0, 0, 0}}},{25952, {{36898, 1, 0, 0},{36898, 1, 1, 20}, {0, 0, 0, 0}}},{25952, {{36898, 1, 0, 0},{36898, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{25953, {{36899, 1, 0, 0},{36899, 1, 1, 18}, {0, 0, 0, 0}}},{25953, {{36899, 1, 0, 0},{36899, 1, 1, 18}, {0, 0, 0, 0}}},{25954, {{36899, 1, 0, 0},{36899, 1, 1, 19}, {0, 0, 0, 0}}},{25954, {{36899, 1, 0, 0},{36899, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{25955, {{36900, 1, 0, 0},{36900, 1, 1, 16}, {0, 0, 0, 0}}},{25955, {{36900, 1, 0, 0},{36900, 1, 1, 16}, {0, 0, 0, 0}}},{25956, {{36900, 1, 0, 0},{36900, 1, 1, 17}, {0, 0, 0, 0}}},{25956, {{36900, 1, 0, 0},{36900, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{25957, {{36901, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{25958, {{36903, 1, 0, 0},{36903, 1, 1, 17}, {0, 0, 0, 0}}},{25958, {{36903, 1, 0, 0},{36903, 1, 1, 17}, {0, 0, 0, 0}}},{25959, {{36903, 1, 0, 0},{36903, 1, 1, 18}, {0, 0, 0, 0}}},{25959, {{36903, 1, 0, 0},{36903, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{25960, {{36904, 1, 0, 0},{36904, 1, 1, 17}, {0, 0, 0, 0}}},{25960, {{36904, 1, 0, 0},{36904, 1, 1, 17}, {0, 0, 0, 0}}},{25961, {{36904, 1, 0, 0},{36904, 1, 1, 18}, {0, 0, 0, 0}}},{25961, {{36904, 1, 0, 0},{36904, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{25962, {{36905, 1, 0, 0},{36905, 1, 1, 18}, {0, 0, 0, 0}}},{25962, {{36905, 1, 0, 0},{36905, 1, 1, 18}, {0, 0, 0, 0}}},{25963, {{36905, 1, 0, 0},{36905, 1, 1, 17}, {0, 0, 0, 0}}},{25963, {{36905, 1, 0, 0},{36905, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{25964, {{36906, 1, 0, 0},{36906, 1, 1, 18}, {0, 0, 0, 0}}},{25964, {{36906, 1, 0, 0},{36906, 1, 1, 18}, {0, 0, 0, 0}}},{25965, {{36906, 1, 0, 0},{36906, 1, 1, 17}, {0, 0, 0, 0}}},{25965, {{36906, 1, 0, 0},{36906, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{25966, {{36907, 1, 0, 0},{36907, 1, 1, 16}, {0, 0, 0, 0}}},{25966, {{36907, 1, 0, 0},{36907, 1, 1, 16}, {0, 0, 0, 0}}},{25967, {{36907, 1, 0, 0},{36907, 1, 1, 17}, {0, 0, 0, 0}}},{25967, {{36907, 1, 0, 0},{36907, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{25968, {{36908, 1, 0, 0},{36908, 1, 1, 18}, {0, 0, 0, 0}}},{25968, {{36908, 1, 0, 0},{36908, 1, 1, 18}, {0, 0, 0, 0}}},{25969, {{36908, 1, 0, 0},{36908, 1, 1, 17}, {0, 0, 0, 0}}},{25969, {{36908, 1, 0, 0},{36908, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{25970, {{36909, 1, 0, 0},{36909, 1, 1, 18}, {0, 0, 0, 0}}},{25970, {{36909, 1, 0, 0},{36909, 1, 1, 18}, {0, 0, 0, 0}}},{25971, {{36909, 1, 0, 0},{36909, 1, 1, 17}, {0, 0, 0, 0}}},{25971, {{36909, 1, 0, 0},{36909, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{25972, {{36910, 1, 0, 0},{36910, 1, 1, 18}, {0, 0, 0, 0}}},{25972, {{36910, 1, 0, 0},{36910, 1, 1, 18}, {0, 0, 0, 0}}},{25973, {{36910, 1, 0, 0},{36910, 1, 1, 19}, {0, 0, 0, 0}}},{25973, {{36910, 1, 0, 0},{36910, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{25974, {{36911, 1, 0, 0},{36911, 1, 1, 18}, {0, 0, 0, 0}}},{25974, {{36911, 1, 0, 0},{36911, 1, 1, 18}, {0, 0, 0, 0}}},{25975, {{36911, 1, 0, 0},{36911, 1, 1, 17}, {0, 0, 0, 0}}},{25975, {{36911, 1, 0, 0},{36911, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{25976, {{36912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25977, {{36913, 1, 0, 0},{36913, 1, 1, 18}, {0, 0, 0, 0}}},{25977, {{36913, 1, 0, 0},{36913, 1, 1, 18}, {0, 0, 0, 0}}},{25978, {{36913, 1, 0, 0},{36913, 1, 1, 19}, {0, 0, 0, 0}}},{25978, {{36913, 1, 0, 0},{36913, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{25979, {{36914, 1, 0, 0},{36914, 1, 1, 18}, {0, 0, 0, 0}}},{25979, {{36914, 1, 0, 0},{36914, 1, 1, 18}, {0, 0, 0, 0}}},{25980, {{36914, 1, 0, 0},{36914, 1, 1, 19}, {0, 0, 0, 0}}},{25980, {{36914, 1, 0, 0},{36914, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{25981, {{36915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{25982, {{36916, 1, 0, 0},{36916, 1, 1, 18}, {0, 0, 0, 0}}},{25982, {{36916, 1, 0, 0},{36916, 1, 1, 18}, {0, 0, 0, 0}}},{25983, {{36916, 1, 0, 0},{36916, 1, 1, 17}, {0, 0, 0, 0}}},{25983, {{36916, 1, 0, 0},{36916, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {6, {{25984, {{36917, 1, 0, 0},{36917, 1, 1, 19}, {0, 0, 0, 0}}},{25984, {{36917, 1, 0, 0},{36917, 1, 1, 19}, {0, 0, 0, 0}}},{25985, {{36917, 1, 0, 0},{36917, 1, 1, 18}, {0, 0, 0, 0}}},{25985, {{36917, 1, 0, 0},{36917, 1, 1, 18}, {0, 0, 0, 0}}},{25986, {{36917, 1, 0, 0},{36917, 1, 1, 20}, {0, 0, 0, 0}}},{25986, {{36917, 1, 0, 0},{36917, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{25989, {{36918, 1, 0, 0},{36918, 1, 1, 18}, {0, 0, 0, 0}}},{25989, {{36918, 1, 0, 0},{36918, 1, 1, 18}, {0, 0, 0, 0}}},{25990, {{36918, 1, 0, 0},{36918, 1, 1, 17}, {0, 0, 0, 0}}},{25990, {{36918, 1, 0, 0},{36918, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{25991, {{36919, 1, 0, 0},{36919, 1, 1, 18}, {0, 0, 0, 0}}},{25991, {{36919, 1, 0, 0},{36919, 1, 1, 18}, {0, 0, 0, 0}}},{25992, {{36919, 1, 0, 0},{36919, 1, 1, 17}, {0, 0, 0, 0}}},{25992, {{36919, 1, 0, 0},{36919, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {14, {{25993, {{36920, 1, 0, 0},{36920, 1, 1, 19}, {0, 0, 0, 0}}},{25993, {{36920, 1, 0, 0},{36920, 1, 1, 19}, {0, 0, 0, 0}}},{25994, {{36920, 1, 0, 0},{36920, 1, 1, 22}, {0, 0, 0, 0}}},{25994, {{36920, 1, 0, 0},{36920, 1, 1, 22}, {0, 0, 0, 0}}},{25995, {{36920, 1, 0, 0},{36920, 1, 1, 21}, {0, 0, 0, 0}}},{25995, {{36920, 1, 0, 0},{36920, 1, 1, 21}, {0, 0, 0, 0}}},{25996, {{36920, 1, 0, 0},{36920, 1, 1, 20}, {0, 0, 0, 0}}},{25996, {{36920, 1, 0, 0},{36920, 1, 1, 20}, {0, 0, 0, 0}}},{30234, {{36920, 1, 0, 0},{36920, 1, 1, 0},{36920, 1, 1, 24},{64037, 1, 0, 0}, {0, 0, 0, 0}}},{30234, {{36920, 1, 0, 0},{36920, 1, 1, 0},{36920, 1, 1, 24},{64037, 1, 0, 0}, {0, 0, 0, 0}}},{30234, {{36920, 1, 0, 0},{36920, 1, 1, 0},{36920, 1, 1, 24},{64037, 1, 0, 0}, {0, 0, 0, 0}}},{30303, {{36920, 1, 0, 0},{36920, 1, 1, 1},{36920, 1, 1, 23},{64103, 1, 0, 0}, {0, 0, 0, 0}}},{30303, {{36920, 1, 0, 0},{36920, 1, 1, 1},{36920, 1, 1, 23},{64103, 1, 0, 0}, {0, 0, 0, 0}}},{30303, {{36920, 1, 0, 0},{36920, 1, 1, 1},{36920, 1, 1, 23},{64103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{25998, {{36921, 1, 0, 0},{36921, 1, 1, 19}, {0, 0, 0, 0}}},{25998, {{36921, 1, 0, 0},{36921, 1, 1, 19}, {0, 0, 0, 0}}},{25999, {{36921, 1, 0, 0},{36921, 1, 1, 18}, {0, 0, 0, 0}}},{25999, {{36921, 1, 0, 0},{36921, 1, 1, 18}, {0, 0, 0, 0}}},{26000, {{36921, 1, 0, 0},{36921, 1, 1, 20}, {0, 0, 0, 0}}},{26000, {{36921, 1, 0, 0},{36921, 1, 1, 20}, {0, 0, 0, 0}}},{26001, {{36921, 1, 0, 0},{36921, 1, 1, 22}, {0, 0, 0, 0}}},{26001, {{36921, 1, 0, 0},{36921, 1, 1, 22}, {0, 0, 0, 0}}}}},
    {4, {{58857, {{36922, 1, 0, 0},{36922, 1, 1, 16}, {0, 0, 0, 0}}},{58857, {{36922, 1, 0, 0},{36922, 1, 1, 16}, {0, 0, 0, 0}}},{60230, {{36922, 1, 0, 0},{36922, 1, 1, 19}, {0, 0, 0, 0}}},{60230, {{36922, 1, 0, 0},{36922, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{26002, {{36924, 1, 0, 0},{36924, 1, 1, 18}, {0, 0, 0, 0}}},{26002, {{36924, 1, 0, 0},{36924, 1, 1, 18}, {0, 0, 0, 0}}},{26003, {{36924, 1, 0, 0},{36924, 1, 1, 19}, {0, 0, 0, 0}}},{26003, {{36924, 1, 0, 0},{36924, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{26004, {{36925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{26005, {{36926, 1, 0, 0},{36926, 1, 1, 18}, {0, 0, 0, 0}}},{26005, {{36926, 1, 0, 0},{36926, 1, 1, 18}, {0, 0, 0, 0}}},{26006, {{36926, 1, 0, 0},{36926, 1, 1, 17}, {0, 0, 0, 0}}},{26006, {{36926, 1, 0, 0},{36926, 1, 1, 17}, {0, 0, 0, 0}}},{26007, {{36926, 1, 0, 0},{36926, 1, 1, 19}, {0, 0, 0, 0}}},{26007, {{36926, 1, 0, 0},{36926, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{26008, {{36927, 1, 0, 0},{36927, 1, 1, 18}, {0, 0, 0, 0}}},{26008, {{36927, 1, 0, 0},{36927, 1, 1, 18}, {0, 0, 0, 0}}},{26009, {{36927, 1, 0, 0},{36927, 1, 1, 17}, {0, 0, 0, 0}}},{26009, {{36927, 1, 0, 0},{36927, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{26010, {{36928, 1, 0, 0},{36928, 1, 1, 16}, {0, 0, 0, 0}}},{26010, {{36928, 1, 0, 0},{36928, 1, 1, 16}, {0, 0, 0, 0}}},{26011, {{36928, 1, 0, 0},{36928, 1, 1, 17}, {0, 0, 0, 0}}},{26011, {{36928, 1, 0, 0},{36928, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {6, {{26012, {{36929, 1, 0, 0},{36929, 1, 1, 19}, {0, 0, 0, 0}}},{26012, {{36929, 1, 0, 0},{36929, 1, 1, 19}, {0, 0, 0, 0}}},{26013, {{36929, 1, 0, 0},{36929, 1, 1, 20}, {0, 0, 0, 0}}},{26013, {{36929, 1, 0, 0},{36929, 1, 1, 20}, {0, 0, 0, 0}}},{26014, {{36929, 1, 0, 0},{36929, 1, 1, 21}, {0, 0, 0, 0}}},{26014, {{36929, 1, 0, 0},{36929, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {8, {{26015, {{36930, 1, 0, 0},{36930, 1, 1, 19}, {0, 0, 0, 0}}},{26015, {{36930, 1, 0, 0},{36930, 1, 1, 19}, {0, 0, 0, 0}}},{26016, {{36930, 1, 0, 0},{36930, 1, 1, 20}, {0, 0, 0, 0}}},{26016, {{36930, 1, 0, 0},{36930, 1, 1, 20}, {0, 0, 0, 0}}},{26017, {{36930, 1, 0, 0},{36930, 1, 1, 21}, {0, 0, 0, 0}}},{26017, {{36930, 1, 0, 0},{36930, 1, 1, 21}, {0, 0, 0, 0}}},{26018, {{36930, 1, 0, 0},{36930, 1, 1, 22}, {0, 0, 0, 0}}},{26018, {{36930, 1, 0, 0},{36930, 1, 1, 22}, {0, 0, 0, 0}}}}},
    {1, {{26019, {{36931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26021, {{36932, 1, 0, 0},{36932, 1, 1, 18}, {0, 0, 0, 0}}},{26021, {{36932, 1, 0, 0},{36932, 1, 1, 18}, {0, 0, 0, 0}}},{26022, {{36932, 1, 0, 0},{36932, 1, 1, 17}, {0, 0, 0, 0}}},{26022, {{36932, 1, 0, 0},{36932, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{26023, {{36933, 1, 0, 0},{36933, 1, 1, 17}, {0, 0, 0, 0}}},{26023, {{36933, 1, 0, 0},{36933, 1, 1, 17}, {0, 0, 0, 0}}},{26024, {{36933, 1, 0, 0},{36933, 1, 1, 18}, {0, 0, 0, 0}}},{26024, {{36933, 1, 0, 0},{36933, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26025, {{36934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{26026, {{36935, 1, 0, 0},{36935, 1, 1, 18}, {0, 0, 0, 0}}},{26026, {{36935, 1, 0, 0},{36935, 1, 1, 18}, {0, 0, 0, 0}}},{26027, {{36935, 1, 0, 0},{36935, 1, 1, 20}, {0, 0, 0, 0}}},{26027, {{36935, 1, 0, 0},{36935, 1, 1, 20}, {0, 0, 0, 0}}},{26028, {{36935, 1, 0, 0},{36935, 1, 1, 19}, {0, 0, 0, 0}}},{26028, {{36935, 1, 0, 0},{36935, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{26029, {{36936, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26030, {{36937, 1, 0, 0},{36937, 1, 1, 18}, {0, 0, 0, 0}}},{26030, {{36937, 1, 0, 0},{36937, 1, 1, 18}, {0, 0, 0, 0}}},{26031, {{36937, 1, 0, 0},{36937, 1, 1, 17}, {0, 0, 0, 0}}},{26031, {{36937, 1, 0, 0},{36937, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{26032, {{36938, 1, 0, 0},{36938, 1, 1, 18}, {0, 0, 0, 0}}},{26032, {{36938, 1, 0, 0},{36938, 1, 1, 18}, {0, 0, 0, 0}}},{26033, {{36938, 1, 0, 0},{36938, 1, 1, 19}, {0, 0, 0, 0}}},{26033, {{36938, 1, 0, 0},{36938, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{26034, {{36939, 1, 0, 0},{36939, 1, 1, 18}, {0, 0, 0, 0}}},{26034, {{36939, 1, 0, 0},{36939, 1, 1, 18}, {0, 0, 0, 0}}},{26035, {{36939, 1, 0, 0},{36939, 1, 1, 19}, {0, 0, 0, 0}}},{26035, {{36939, 1, 0, 0},{36939, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {6, {{26036, {{36940, 1, 0, 0},{36940, 1, 1, 18}, {0, 0, 0, 0}}},{26036, {{36940, 1, 0, 0},{36940, 1, 1, 18}, {0, 0, 0, 0}}},{26037, {{36940, 1, 0, 0},{36940, 1, 1, 17}, {0, 0, 0, 0}}},{26037, {{36940, 1, 0, 0},{36940, 1, 1, 17}, {0, 0, 0, 0}}},{26038, {{36940, 1, 0, 0},{36940, 1, 1, 19}, {0, 0, 0, 0}}},{26038, {{36940, 1, 0, 0},{36940, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{26039, {{36941, 1, 0, 0},{36941, 1, 1, 18}, {0, 0, 0, 0}}},{26039, {{36941, 1, 0, 0},{36941, 1, 1, 18}, {0, 0, 0, 0}}},{26040, {{36941, 1, 0, 0},{36941, 1, 1, 19}, {0, 0, 0, 0}}},{26040, {{36941, 1, 0, 0},{36941, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {6, {{26041, {{36942, 1, 0, 0},{36942, 1, 1, 18}, {0, 0, 0, 0}}},{26041, {{36942, 1, 0, 0},{36942, 1, 1, 18}, {0, 0, 0, 0}}},{26042, {{36942, 1, 0, 0},{36942, 1, 1, 19}, {0, 0, 0, 0}}},{26042, {{36942, 1, 0, 0},{36942, 1, 1, 19}, {0, 0, 0, 0}}},{58856, {{36942, 1, 0, 0},{36942, 1, 1, 20}, {0, 0, 0, 0}}},{58856, {{36942, 1, 0, 0},{36942, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {8, {{26043, {{36943, 1, 0, 0},{36943, 1, 1, 20}, {0, 0, 0, 0}}},{26043, {{36943, 1, 0, 0},{36943, 1, 1, 20}, {0, 0, 0, 0}}},{26044, {{36943, 1, 0, 0},{36943, 1, 1, 18}, {0, 0, 0, 0}}},{26044, {{36943, 1, 0, 0},{36943, 1, 1, 18}, {0, 0, 0, 0}}},{26045, {{36943, 1, 0, 0},{36943, 1, 1, 17}, {0, 0, 0, 0}}},{26045, {{36943, 1, 0, 0},{36943, 1, 1, 17}, {0, 0, 0, 0}}},{26046, {{36943, 1, 0, 0},{36943, 1, 1, 19}, {0, 0, 0, 0}}},{26046, {{36943, 1, 0, 0},{36943, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {6, {{26047, {{36944, 1, 0, 0},{36944, 1, 1, 20}, {0, 0, 0, 0}}},{26047, {{36944, 1, 0, 0},{36944, 1, 1, 20}, {0, 0, 0, 0}}},{26048, {{36944, 1, 0, 0},{36944, 1, 1, 18}, {0, 0, 0, 0}}},{26048, {{36944, 1, 0, 0},{36944, 1, 1, 18}, {0, 0, 0, 0}}},{26049, {{36944, 1, 0, 0},{36944, 1, 1, 19}, {0, 0, 0, 0}}},{26049, {{36944, 1, 0, 0},{36944, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{26050, {{36945, 1, 0, 0},{36945, 1, 1, 18}, {0, 0, 0, 0}}},{26050, {{36945, 1, 0, 0},{36945, 1, 1, 18}, {0, 0, 0, 0}}},{26051, {{36945, 1, 0, 0},{36945, 1, 1, 17}, {0, 0, 0, 0}}},{26051, {{36945, 1, 0, 0},{36945, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {12, {{26052, {{36946, 1, 0, 0},{36946, 1, 1, 20}, {0, 0, 0, 0}}},{26052, {{36946, 1, 0, 0},{36946, 1, 1, 20}, {0, 0, 0, 0}}},{26053, {{36946, 1, 0, 0},{36946, 1, 1, 18}, {0, 0, 0, 0}}},{26053, {{36946, 1, 0, 0},{36946, 1, 1, 18}, {0, 0, 0, 0}}},{26054, {{36946, 1, 0, 0},{36946, 1, 1, 23}, {0, 0, 0, 0}}},{26054, {{36946, 1, 0, 0},{36946, 1, 1, 23}, {0, 0, 0, 0}}},{26055, {{36946, 1, 0, 0},{36946, 1, 1, 19}, {0, 0, 0, 0}}},{26055, {{36946, 1, 0, 0},{36946, 1, 1, 19}, {0, 0, 0, 0}}},{26056, {{36946, 1, 0, 0},{36946, 1, 1, 21}, {0, 0, 0, 0}}},{26056, {{36946, 1, 0, 0},{36946, 1, 1, 21}, {0, 0, 0, 0}}},{26057, {{36946, 1, 0, 0},{36946, 1, 1, 22}, {0, 0, 0, 0}}},{26057, {{36946, 1, 0, 0},{36946, 1, 1, 22}, {0, 0, 0, 0}}}}},
    {4, {{26058, {{36947, 1, 0, 0},{36947, 1, 1, 18}, {0, 0, 0, 0}}},{26058, {{36947, 1, 0, 0},{36947, 1, 1, 18}, {0, 0, 0, 0}}},{26059, {{36947, 1, 0, 0},{36947, 1, 1, 19}, {0, 0, 0, 0}}},{26059, {{36947, 1, 0, 0},{36947, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{26060, {{36948, 1, 0, 0},{36948, 1, 1, 18}, {0, 0, 0, 0}}},{26060, {{36948, 1, 0, 0},{36948, 1, 1, 18}, {0, 0, 0, 0}}},{26061, {{36948, 1, 0, 0},{36948, 1, 1, 19}, {0, 0, 0, 0}}},{26061, {{36948, 1, 0, 0},{36948, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {10, {{26062, {{36949, 1, 0, 0},{36949, 1, 1, 21}, {0, 0, 0, 0}}},{26062, {{36949, 1, 0, 0},{36949, 1, 1, 21}, {0, 0, 0, 0}}},{26063, {{36949, 1, 0, 0},{36949, 1, 1, 22}, {0, 0, 0, 0}}},{26063, {{36949, 1, 0, 0},{36949, 1, 1, 22}, {0, 0, 0, 0}}},{26064, {{36949, 1, 0, 0},{36949, 1, 1, 19}, {0, 0, 0, 0}}},{26064, {{36949, 1, 0, 0},{36949, 1, 1, 19}, {0, 0, 0, 0}}},{26065, {{36949, 1, 0, 0},{36949, 1, 1, 23}, {0, 0, 0, 0}}},{26065, {{36949, 1, 0, 0},{36949, 1, 1, 23}, {0, 0, 0, 0}}},{26066, {{36949, 1, 0, 0},{36949, 1, 1, 20}, {0, 0, 0, 0}}},{26066, {{36949, 1, 0, 0},{36949, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{26067, {{36950, 1, 0, 0},{36950, 1, 1, 18}, {0, 0, 0, 0}}},{26067, {{36950, 1, 0, 0},{36950, 1, 1, 18}, {0, 0, 0, 0}}},{26068, {{36950, 1, 0, 0},{36950, 1, 1, 17}, {0, 0, 0, 0}}},{26068, {{36950, 1, 0, 0},{36950, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {8, {{26069, {{36952, 1, 0, 0},{36952, 1, 1, 21}, {0, 0, 0, 0}}},{26069, {{36952, 1, 0, 0},{36952, 1, 1, 21}, {0, 0, 0, 0}}},{26070, {{36952, 1, 0, 0},{36952, 1, 1, 19}, {0, 0, 0, 0}}},{26070, {{36952, 1, 0, 0},{36952, 1, 1, 19}, {0, 0, 0, 0}}},{26071, {{36952, 1, 0, 0},{36952, 1, 1, 18}, {0, 0, 0, 0}}},{26071, {{36952, 1, 0, 0},{36952, 1, 1, 18}, {0, 0, 0, 0}}},{26072, {{36952, 1, 0, 0},{36952, 1, 1, 20}, {0, 0, 0, 0}}},{26072, {{36952, 1, 0, 0},{36952, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{26073, {{36953, 1, 0, 0},{36953, 1, 1, 18}, {0, 0, 0, 0}}},{26073, {{36953, 1, 0, 0},{36953, 1, 1, 18}, {0, 0, 0, 0}}},{26074, {{36953, 1, 0, 0},{36953, 1, 1, 17}, {0, 0, 0, 0}}},{26074, {{36953, 1, 0, 0},{36953, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{26075, {{36954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26076, {{36955, 1, 0, 0},{36955, 1, 1, 18}, {0, 0, 0, 0}}},{26076, {{36955, 1, 0, 0},{36955, 1, 1, 18}, {0, 0, 0, 0}}},{26077, {{36955, 1, 0, 0},{36955, 1, 1, 17}, {0, 0, 0, 0}}},{26077, {{36955, 1, 0, 0},{36955, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{26078, {{36956, 1, 0, 0},{36956, 1, 1, 18}, {0, 0, 0, 0}}},{26078, {{36956, 1, 0, 0},{36956, 1, 1, 18}, {0, 0, 0, 0}}},{26079, {{36956, 1, 0, 0},{36956, 1, 1, 19}, {0, 0, 0, 0}}},{26079, {{36956, 1, 0, 0},{36956, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{26080, {{36957, 1, 0, 0},{36957, 1, 1, 18}, {0, 0, 0, 0}}},{26080, {{36957, 1, 0, 0},{36957, 1, 1, 18}, {0, 0, 0, 0}}},{26081, {{36957, 1, 0, 0},{36957, 1, 1, 17}, {0, 0, 0, 0}}},{26081, {{36957, 1, 0, 0},{36957, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{26082, {{36958, 1, 0, 0},{36958, 1, 1, 18}, {0, 0, 0, 0}}},{26082, {{36958, 1, 0, 0},{36958, 1, 1, 18}, {0, 0, 0, 0}}},{26083, {{36958, 1, 0, 0},{36958, 1, 1, 17}, {0, 0, 0, 0}}},{26083, {{36958, 1, 0, 0},{36958, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{26084, {{36959, 1, 0, 0},{36959, 1, 1, 16}, {0, 0, 0, 0}}},{26084, {{36959, 1, 0, 0},{36959, 1, 1, 16}, {0, 0, 0, 0}}},{26085, {{36959, 1, 0, 0},{36959, 1, 1, 17}, {0, 0, 0, 0}}},{26085, {{36959, 1, 0, 0},{36959, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {6, {{26086, {{36960, 1, 0, 0},{36960, 1, 1, 18}, {0, 0, 0, 0}}},{26086, {{36960, 1, 0, 0},{36960, 1, 1, 18}, {0, 0, 0, 0}}},{26087, {{36960, 1, 0, 0},{36960, 1, 1, 20}, {0, 0, 0, 0}}},{26087, {{36960, 1, 0, 0},{36960, 1, 1, 20}, {0, 0, 0, 0}}},{26088, {{36960, 1, 0, 0},{36960, 1, 1, 19}, {0, 0, 0, 0}}},{26088, {{36960, 1, 0, 0},{36960, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{26089, {{36961, 1, 0, 0},{36961, 1, 1, 18}, {0, 0, 0, 0}}},{26089, {{36961, 1, 0, 0},{36961, 1, 1, 18}, {0, 0, 0, 0}}},{26090, {{36961, 1, 0, 0},{36961, 1, 1, 19}, {0, 0, 0, 0}}},{26090, {{36961, 1, 0, 0},{36961, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {6, {{26091, {{36962, 1, 0, 0},{36962, 1, 1, 19}, {0, 0, 0, 0}}},{26091, {{36962, 1, 0, 0},{36962, 1, 1, 19}, {0, 0, 0, 0}}},{26092, {{36962, 1, 0, 0},{36962, 1, 1, 17}, {0, 0, 0, 0}}},{26092, {{36962, 1, 0, 0},{36962, 1, 1, 17}, {0, 0, 0, 0}}},{26093, {{36962, 1, 0, 0},{36962, 1, 1, 18}, {0, 0, 0, 0}}},{26093, {{36962, 1, 0, 0},{36962, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{26094, {{36963, 1, 0, 0},{36963, 1, 1, 18}, {0, 0, 0, 0}}},{26094, {{36963, 1, 0, 0},{36963, 1, 1, 18}, {0, 0, 0, 0}}},{26095, {{36963, 1, 0, 0},{36963, 1, 1, 19}, {0, 0, 0, 0}}},{26095, {{36963, 1, 0, 0},{36963, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{26096, {{36965, 1, 0, 0},{36965, 1, 1, 17}, {0, 0, 0, 0}}},{26096, {{36965, 1, 0, 0},{36965, 1, 1, 17}, {0, 0, 0, 0}}},{26097, {{36965, 1, 0, 0},{36965, 1, 1, 18}, {0, 0, 0, 0}}},{26097, {{36965, 1, 0, 0},{36965, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{26098, {{36966, 1, 0, 0},{36966, 1, 1, 17}, {0, 0, 0, 0}}},{26098, {{36966, 1, 0, 0},{36966, 1, 1, 17}, {0, 0, 0, 0}}},{26099, {{36966, 1, 0, 0},{36966, 1, 1, 18}, {0, 0, 0, 0}}},{26099, {{36966, 1, 0, 0},{36966, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{26100, {{36967, 1, 0, 0},{36967, 1, 1, 18}, {0, 0, 0, 0}}},{26100, {{36967, 1, 0, 0},{36967, 1, 1, 18}, {0, 0, 0, 0}}},{26101, {{36967, 1, 0, 0},{36967, 1, 1, 17}, {0, 0, 0, 0}}},{26101, {{36967, 1, 0, 0},{36967, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {6, {{26102, {{36968, 1, 0, 0},{36968, 1, 1, 18}, {0, 0, 0, 0}}},{26102, {{36968, 1, 0, 0},{36968, 1, 1, 18}, {0, 0, 0, 0}}},{26103, {{36968, 1, 0, 0},{36968, 1, 1, 17}, {0, 0, 0, 0}}},{26103, {{36968, 1, 0, 0},{36968, 1, 1, 17}, {0, 0, 0, 0}}},{26104, {{36968, 1, 0, 0},{36968, 1, 1, 19}, {0, 0, 0, 0}}},{26104, {{36968, 1, 0, 0},{36968, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {6, {{26105, {{36969, 1, 0, 0},{36969, 1, 1, 18}, {0, 0, 0, 0}}},{26105, {{36969, 1, 0, 0},{36969, 1, 1, 18}, {0, 0, 0, 0}}},{26106, {{36969, 1, 0, 0},{36969, 1, 1, 19}, {0, 0, 0, 0}}},{26106, {{36969, 1, 0, 0},{36969, 1, 1, 19}, {0, 0, 0, 0}}},{26107, {{36969, 1, 0, 0},{36969, 1, 1, 20}, {0, 0, 0, 0}}},{26107, {{36969, 1, 0, 0},{36969, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{26108, {{36970, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26109, {{36971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26110, {{36972, 1, 0, 0},{36972, 1, 1, 18}, {0, 0, 0, 0}}},{26110, {{36972, 1, 0, 0},{36972, 1, 1, 18}, {0, 0, 0, 0}}},{26111, {{36972, 1, 0, 0},{36972, 1, 1, 17}, {0, 0, 0, 0}}},{26111, {{36972, 1, 0, 0},{36972, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{26112, {{36973, 1, 0, 0},{36973, 1, 1, 18}, {0, 0, 0, 0}}},{26112, {{36973, 1, 0, 0},{36973, 1, 1, 18}, {0, 0, 0, 0}}},{26113, {{36973, 1, 0, 0},{36973, 1, 1, 19}, {0, 0, 0, 0}}},{26113, {{36973, 1, 0, 0},{36973, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {8, {{26114, {{36974, 1, 0, 0},{36974, 1, 1, 18}, {0, 0, 0, 0}}},{26114, {{36974, 1, 0, 0},{36974, 1, 1, 18}, {0, 0, 0, 0}}},{26115, {{36974, 1, 0, 0},{36974, 1, 1, 21}, {0, 0, 0, 0}}},{26115, {{36974, 1, 0, 0},{36974, 1, 1, 21}, {0, 0, 0, 0}}},{26116, {{36974, 1, 0, 0},{36974, 1, 1, 20}, {0, 0, 0, 0}}},{26116, {{36974, 1, 0, 0},{36974, 1, 1, 20}, {0, 0, 0, 0}}},{26117, {{36974, 1, 0, 0},{36974, 1, 1, 19}, {0, 0, 0, 0}}},{26117, {{36974, 1, 0, 0},{36974, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {10, {{26118, {{36975, 1, 0, 0},{36975, 1, 1, 18}, {0, 0, 0, 0}}},{26118, {{36975, 1, 0, 0},{36975, 1, 1, 18}, {0, 0, 0, 0}}},{26119, {{36975, 1, 0, 0},{36975, 1, 1, 17}, {0, 0, 0, 0}}},{26119, {{36975, 1, 0, 0},{36975, 1, 1, 17}, {0, 0, 0, 0}}},{26120, {{36975, 1, 0, 0},{36975, 1, 1, 20}, {0, 0, 0, 0}}},{26120, {{36975, 1, 0, 0},{36975, 1, 1, 20}, {0, 0, 0, 0}}},{26121, {{36975, 1, 0, 0},{36975, 1, 1, 19}, {0, 0, 0, 0}}},{26121, {{36975, 1, 0, 0},{36975, 1, 1, 19}, {0, 0, 0, 0}}},{26122, {{36975, 1, 0, 0},{36975, 1, 1, 21}, {0, 0, 0, 0}}},{26122, {{36975, 1, 0, 0},{36975, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {4, {{26123, {{36976, 1, 0, 0},{36976, 1, 1, 18}, {0, 0, 0, 0}}},{26123, {{36976, 1, 0, 0},{36976, 1, 1, 18}, {0, 0, 0, 0}}},{26124, {{36976, 1, 0, 0},{36976, 1, 1, 17}, {0, 0, 0, 0}}},{26124, {{36976, 1, 0, 0},{36976, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{26125, {{36977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{26126, {{36978, 1, 0, 0},{36978, 1, 1, 18}, {0, 0, 0, 0}}},{26126, {{36978, 1, 0, 0},{36978, 1, 1, 18}, {0, 0, 0, 0}}},{26127, {{36978, 1, 0, 0},{36978, 1, 1, 17}, {0, 0, 0, 0}}},{26127, {{36978, 1, 0, 0},{36978, 1, 1, 17}, {0, 0, 0, 0}}},{58867, {{36978, 1, 0, 0},{36978, 1, 1, 19}, {0, 0, 0, 0}}},{58867, {{36978, 1, 0, 0},{36978, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {6, {{26128, {{36979, 1, 0, 0},{36979, 1, 1, 16}, {0, 0, 0, 0}}},{26128, {{36979, 1, 0, 0},{36979, 1, 1, 16}, {0, 0, 0, 0}}},{26129, {{36979, 1, 0, 0},{36979, 1, 1, 17}, {0, 0, 0, 0}}},{26129, {{36979, 1, 0, 0},{36979, 1, 1, 17}, {0, 0, 0, 0}}},{26130, {{36979, 1, 0, 0},{36979, 1, 1, 18}, {0, 0, 0, 0}}},{26130, {{36979, 1, 0, 0},{36979, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {8, {{26131, {{36980, 1, 0, 0},{36980, 1, 1, 19}, {0, 0, 0, 0}}},{26131, {{36980, 1, 0, 0},{36980, 1, 1, 19}, {0, 0, 0, 0}}},{26132, {{36980, 1, 0, 0},{36980, 1, 1, 20}, {0, 0, 0, 0}}},{26132, {{36980, 1, 0, 0},{36980, 1, 1, 20}, {0, 0, 0, 0}}},{26133, {{36980, 1, 0, 0},{36980, 1, 1, 17}, {0, 0, 0, 0}}},{26133, {{36980, 1, 0, 0},{36980, 1, 1, 17}, {0, 0, 0, 0}}},{26134, {{36980, 1, 0, 0},{36980, 1, 1, 18}, {0, 0, 0, 0}}},{26134, {{36980, 1, 0, 0},{36980, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{26135, {{36981, 1, 0, 0},{36981, 1, 1, 19}, {0, 0, 0, 0}}},{26135, {{36981, 1, 0, 0},{36981, 1, 1, 19}, {0, 0, 0, 0}}},{26136, {{36981, 1, 0, 0},{36981, 1, 1, 20}, {0, 0, 0, 0}}},{26136, {{36981, 1, 0, 0},{36981, 1, 1, 20}, {0, 0, 0, 0}}},{26137, {{36981, 1, 0, 0},{36981, 1, 1, 21}, {0, 0, 0, 0}}},{26137, {{36981, 1, 0, 0},{36981, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {4, {{26138, {{36982, 1, 0, 0},{36982, 1, 1, 18}, {0, 0, 0, 0}}},{26138, {{36982, 1, 0, 0},{36982, 1, 1, 18}, {0, 0, 0, 0}}},{26139, {{36982, 1, 0, 0},{36982, 1, 1, 17}, {0, 0, 0, 0}}},{26139, {{36982, 1, 0, 0},{36982, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {14, {{26140, {{36983, 1, 0, 0},{36983, 1, 1, 20}, {0, 0, 0, 0}}},{26140, {{36983, 1, 0, 0},{36983, 1, 1, 20}, {0, 0, 0, 0}}},{26141, {{36983, 1, 0, 0},{36983, 1, 1, 24}, {0, 0, 0, 0}}},{26141, {{36983, 1, 0, 0},{36983, 1, 1, 24}, {0, 0, 0, 0}}},{26142, {{36983, 1, 0, 0},{36983, 1, 1, 21}, {0, 0, 0, 0}}},{26142, {{36983, 1, 0, 0},{36983, 1, 1, 21}, {0, 0, 0, 0}}},{26143, {{36983, 1, 0, 0},{36983, 1, 1, 23}, {0, 0, 0, 0}}},{26143, {{36983, 1, 0, 0},{36983, 1, 1, 23}, {0, 0, 0, 0}}},{26144, {{36983, 1, 0, 0},{36983, 1, 1, 22}, {0, 0, 0, 0}}},{26144, {{36983, 1, 0, 0},{36983, 1, 1, 22}, {0, 0, 0, 0}}},{26145, {{36983, 1, 0, 0},{36983, 1, 1, 26}, {0, 0, 0, 0}}},{26145, {{36983, 1, 0, 0},{36983, 1, 1, 26}, {0, 0, 0, 0}}},{58865, {{36983, 1, 0, 0},{36983, 1, 1, 25}, {0, 0, 0, 0}}},{58865, {{36983, 1, 0, 0},{36983, 1, 1, 25}, {0, 0, 0, 0}}}}},
    {6, {{26146, {{36984, 1, 0, 0},{36984, 1, 1, 18}, {0, 0, 0, 0}}},{26146, {{36984, 1, 0, 0},{36984, 1, 1, 18}, {0, 0, 0, 0}}},{26147, {{36984, 1, 0, 0},{36984, 1, 1, 19}, {0, 0, 0, 0}}},{26147, {{36984, 1, 0, 0},{36984, 1, 1, 19}, {0, 0, 0, 0}}},{26148, {{36984, 1, 0, 0},{36984, 1, 1, 20}, {0, 0, 0, 0}}},{26148, {{36984, 1, 0, 0},{36984, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{26149, {{36985, 1, 0, 0},{36985, 1, 1, 18}, {0, 0, 0, 0}}},{26149, {{36985, 1, 0, 0},{36985, 1, 1, 18}, {0, 0, 0, 0}}},{26150, {{36985, 1, 0, 0},{36985, 1, 1, 17}, {0, 0, 0, 0}}},{26150, {{36985, 1, 0, 0},{36985, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{26151, {{36986, 1, 0, 0},{36986, 1, 1, 18}, {0, 0, 0, 0}}},{26151, {{36986, 1, 0, 0},{36986, 1, 1, 18}, {0, 0, 0, 0}}},{26152, {{36986, 1, 0, 0},{36986, 1, 1, 19}, {0, 0, 0, 0}}},{26152, {{36986, 1, 0, 0},{36986, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{26153, {{36987, 1, 0, 0},{36987, 1, 1, 16}, {0, 0, 0, 0}}},{26153, {{36987, 1, 0, 0},{36987, 1, 1, 16}, {0, 0, 0, 0}}},{26154, {{36987, 1, 0, 0},{36987, 1, 1, 17}, {0, 0, 0, 0}}},{26154, {{36987, 1, 0, 0},{36987, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{26155, {{36988, 1, 0, 0},{36988, 1, 1, 18}, {0, 0, 0, 0}}},{26155, {{36988, 1, 0, 0},{36988, 1, 1, 18}, {0, 0, 0, 0}}},{26156, {{36988, 1, 0, 0},{36988, 1, 1, 19}, {0, 0, 0, 0}}},{26156, {{36988, 1, 0, 0},{36988, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {10, {{26157, {{36989, 1, 0, 0},{36989, 1, 1, 18}, {0, 0, 0, 0}}},{26157, {{36989, 1, 0, 0},{36989, 1, 1, 18}, {0, 0, 0, 0}}},{26158, {{36989, 1, 0, 0},{36989, 1, 1, 17}, {0, 0, 0, 0}}},{26158, {{36989, 1, 0, 0},{36989, 1, 1, 17}, {0, 0, 0, 0}}},{26159, {{36989, 1, 0, 0},{36989, 1, 1, 19}, {0, 0, 0, 0}}},{26159, {{36989, 1, 0, 0},{36989, 1, 1, 19}, {0, 0, 0, 0}}},{26160, {{36989, 1, 0, 0},{36989, 1, 1, 20}, {0, 0, 0, 0}}},{26160, {{36989, 1, 0, 0},{36989, 1, 1, 20}, {0, 0, 0, 0}}},{26161, {{36989, 1, 0, 0},{36989, 1, 1, 21}, {0, 0, 0, 0}}},{26161, {{36989, 1, 0, 0},{36989, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{26162, {{36990, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26163, {{36991, 1, 0, 0},{36991, 1, 1, 18}, {0, 0, 0, 0}}},{26163, {{36991, 1, 0, 0},{36991, 1, 1, 18}, {0, 0, 0, 0}}},{26164, {{36991, 1, 0, 0},{36991, 1, 1, 19}, {0, 0, 0, 0}}},{26164, {{36991, 1, 0, 0},{36991, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {6, {{26165, {{36992, 1, 0, 0},{36992, 1, 1, 18}, {0, 0, 0, 0}}},{26165, {{36992, 1, 0, 0},{36992, 1, 1, 18}, {0, 0, 0, 0}}},{26166, {{36992, 1, 0, 0},{36992, 1, 1, 17}, {0, 0, 0, 0}}},{26166, {{36992, 1, 0, 0},{36992, 1, 1, 17}, {0, 0, 0, 0}}},{26167, {{36992, 1, 0, 0},{36992, 1, 1, 19}, {0, 0, 0, 0}}},{26167, {{36992, 1, 0, 0},{36992, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {8, {{26168, {{36993, 1, 0, 0},{36993, 1, 1, 21}, {0, 0, 0, 0}}},{26168, {{36993, 1, 0, 0},{36993, 1, 1, 21}, {0, 0, 0, 0}}},{26169, {{36993, 1, 0, 0},{36993, 1, 1, 18}, {0, 0, 0, 0}}},{26169, {{36993, 1, 0, 0},{36993, 1, 1, 18}, {0, 0, 0, 0}}},{26170, {{36993, 1, 0, 0},{36993, 1, 1, 20}, {0, 0, 0, 0}}},{26170, {{36993, 1, 0, 0},{36993, 1, 1, 20}, {0, 0, 0, 0}}},{26171, {{36993, 1, 0, 0},{36993, 1, 1, 19}, {0, 0, 0, 0}}},{26171, {{36993, 1, 0, 0},{36993, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{26172, {{36994, 1, 0, 0},{36994, 1, 1, 18}, {0, 0, 0, 0}}},{26172, {{36994, 1, 0, 0},{36994, 1, 1, 18}, {0, 0, 0, 0}}},{26173, {{36994, 1, 0, 0},{36994, 1, 1, 17}, {0, 0, 0, 0}}},{26173, {{36994, 1, 0, 0},{36994, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {12, {{26174, {{36995, 1, 0, 0},{36995, 1, 1, 19}, {0, 0, 0, 0}}},{26174, {{36995, 1, 0, 0},{36995, 1, 1, 19}, {0, 0, 0, 0}}},{26175, {{36995, 1, 0, 0},{36995, 1, 1, 23}, {0, 0, 0, 0}}},{26175, {{36995, 1, 0, 0},{36995, 1, 1, 23}, {0, 0, 0, 0}}},{26176, {{36995, 1, 0, 0},{36995, 1, 1, 18}, {0, 0, 0, 0}}},{26176, {{36995, 1, 0, 0},{36995, 1, 1, 18}, {0, 0, 0, 0}}},{26177, {{36995, 1, 0, 0},{36995, 1, 1, 24}, {0, 0, 0, 0}}},{26177, {{36995, 1, 0, 0},{36995, 1, 1, 24}, {0, 0, 0, 0}}},{26178, {{36995, 1, 0, 0},{36995, 1, 1, 21}, {0, 0, 0, 0}}},{26178, {{36995, 1, 0, 0},{36995, 1, 1, 21}, {0, 0, 0, 0}}},{26179, {{36995, 1, 0, 0},{36995, 1, 1, 20}, {0, 0, 0, 0}}},{26179, {{36995, 1, 0, 0},{36995, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {6, {{26180, {{36996, 1, 0, 0},{36996, 1, 1, 18}, {0, 0, 0, 0}}},{26180, {{36996, 1, 0, 0},{36996, 1, 1, 18}, {0, 0, 0, 0}}},{26181, {{36996, 1, 0, 0},{36996, 1, 1, 19}, {0, 0, 0, 0}}},{26181, {{36996, 1, 0, 0},{36996, 1, 1, 19}, {0, 0, 0, 0}}},{58868, {{36996, 1, 0, 0},{36996, 1, 1, 20}, {0, 0, 0, 0}}},{58868, {{36996, 1, 0, 0},{36996, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{26182, {{36997, 1, 0, 0},{36997, 1, 1, 18}, {0, 0, 0, 0}}},{26182, {{36997, 1, 0, 0},{36997, 1, 1, 18}, {0, 0, 0, 0}}},{26183, {{36997, 1, 0, 0},{36997, 1, 1, 17}, {0, 0, 0, 0}}},{26183, {{36997, 1, 0, 0},{36997, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{26184, {{36998, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{26185, {{36999, 1, 0, 0},{36999, 1, 1, 20}, {0, 0, 0, 0}}},{26185, {{36999, 1, 0, 0},{36999, 1, 1, 20}, {0, 0, 0, 0}}},{26186, {{36999, 1, 0, 0},{36999, 1, 1, 18}, {0, 0, 0, 0}}},{26186, {{36999, 1, 0, 0},{36999, 1, 1, 18}, {0, 0, 0, 0}}},{26187, {{36999, 1, 0, 0},{36999, 1, 1, 19}, {0, 0, 0, 0}}},{26187, {{36999, 1, 0, 0},{36999, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{26188, {{37000, 1, 0, 0},{37000, 1, 1, 18}, {0, 0, 0, 0}}},{26188, {{37000, 1, 0, 0},{37000, 1, 1, 18}, {0, 0, 0, 0}}},{26189, {{37000, 1, 0, 0},{37000, 1, 1, 17}, {0, 0, 0, 0}}},{26189, {{37000, 1, 0, 0},{37000, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {32, {{26190, {{37001, 1, 0, 0},{37001, 1, 1, 31}, {0, 0, 0, 0}}},{26190, {{37001, 1, 0, 0},{37001, 1, 1, 31}, {0, 0, 0, 0}}},{26191, {{37001, 1, 0, 0},{37001, 1, 1, 41}, {0, 0, 0, 0}}},{26191, {{37001, 1, 0, 0},{37001, 1, 1, 41}, {0, 0, 0, 0}}},{26192, {{37001, 1, 0, 0},{37001, 1, 1, 43}, {0, 0, 0, 0}}},{26192, {{37001, 1, 0, 0},{37001, 1, 1, 43}, {0, 0, 0, 0}}},{26193, {{37001, 1, 0, 0},{37001, 1, 1, 47}, {0, 0, 0, 0}}},{26193, {{37001, 1, 0, 0},{37001, 1, 1, 47}, {0, 0, 0, 0}}},{26194, {{37001, 1, 0, 0},{37001, 1, 1, 42}, {0, 0, 0, 0}}},{26194, {{37001, 1, 0, 0},{37001, 1, 1, 42}, {0, 0, 0, 0}}},{26195, {{37001, 1, 0, 0},{37001, 1, 1, 44}, {0, 0, 0, 0}}},{26195, {{37001, 1, 0, 0},{37001, 1, 1, 44}, {0, 0, 0, 0}}},{26196, {{37001, 1, 0, 0},{37001, 1, 1, 45}, {0, 0, 0, 0}}},{26196, {{37001, 1, 0, 0},{37001, 1, 1, 45}, {0, 0, 0, 0}}},{26197, {{37001, 1, 0, 0},{37001, 1, 1, 39}, {0, 0, 0, 0}}},{26197, {{37001, 1, 0, 0},{37001, 1, 1, 39}, {0, 0, 0, 0}}},{58866, {{37001, 1, 0, 0},{37001, 1, 1, 38}, {0, 0, 0, 0}}},{58866, {{37001, 1, 0, 0},{37001, 1, 1, 38}, {0, 0, 0, 0}}},{60234, {{37001, 1, 0, 0},{37001, 1, 1, 37}, {0, 0, 0, 0}}},{60234, {{37001, 1, 0, 0},{37001, 1, 1, 37}, {0, 0, 0, 0}}},{60235, {{37001, 1, 0, 0},{37001, 1, 1, 36}, {0, 0, 0, 0}}},{60235, {{37001, 1, 0, 0},{37001, 1, 1, 36}, {0, 0, 0, 0}}},{60236, {{37001, 1, 0, 0},{37001, 1, 1, 40}, {0, 0, 0, 0}}},{60236, {{37001, 1, 0, 0},{37001, 1, 1, 40}, {0, 0, 0, 0}}},{60237, {{37001, 1, 0, 0},{37001, 1, 1, 35}, {0, 0, 0, 0}}},{60237, {{37001, 1, 0, 0},{37001, 1, 1, 35}, {0, 0, 0, 0}}},{60238, {{37001, 1, 0, 0},{37001, 1, 1, 34}, {0, 0, 0, 0}}},{60238, {{37001, 1, 0, 0},{37001, 1, 1, 34}, {0, 0, 0, 0}}},{60239, {{37001, 1, 0, 0},{37001, 1, 1, 33}, {0, 0, 0, 0}}},{60239, {{37001, 1, 0, 0},{37001, 1, 1, 33}, {0, 0, 0, 0}}},{60248, {{37001, 1, 0, 0},{37001, 1, 1, 32}, {0, 0, 0, 0}}},{60248, {{37001, 1, 0, 0},{37001, 1, 1, 32}, {0, 0, 0, 0}}}}},
    {22, {{26199, {{37002, 1, 0, 0},{37002, 1, 1, 31}, {0, 0, 0, 0}}},{26199, {{37002, 1, 0, 0},{37002, 1, 1, 31}, {0, 0, 0, 0}}},{26200, {{37002, 1, 0, 0},{37002, 1, 1, 24}, {0, 0, 0, 0}}},{26200, {{37002, 1, 0, 0},{37002, 1, 1, 24}, {0, 0, 0, 0}}},{26201, {{37002, 1, 0, 0},{37002, 1, 1, 30}, {0, 0, 0, 0}}},{26201, {{37002, 1, 0, 0},{37002, 1, 1, 30}, {0, 0, 0, 0}}},{26202, {{37002, 1, 0, 0},{37002, 1, 1, 29}, {0, 0, 0, 0}}},{26202, {{37002, 1, 0, 0},{37002, 1, 1, 29}, {0, 0, 0, 0}}},{26203, {{37002, 1, 0, 0},{37002, 1, 1, 28}, {0, 0, 0, 0}}},{26203, {{37002, 1, 0, 0},{37002, 1, 1, 28}, {0, 0, 0, 0}}},{26204, {{37002, 1, 0, 0},{37002, 1, 1, 27}, {0, 0, 0, 0}}},{26204, {{37002, 1, 0, 0},{37002, 1, 1, 27}, {0, 0, 0, 0}}},{26205, {{37002, 1, 0, 0},{37002, 1, 1, 26}, {0, 0, 0, 0}}},{26205, {{37002, 1, 0, 0},{37002, 1, 1, 26}, {0, 0, 0, 0}}},{26206, {{37002, 1, 0, 0},{37002, 1, 1, 32}, {0, 0, 0, 0}}},{26206, {{37002, 1, 0, 0},{37002, 1, 1, 32}, {0, 0, 0, 0}}},{26207, {{37002, 1, 0, 0},{37002, 1, 1, 34}, {0, 0, 0, 0}}},{26207, {{37002, 1, 0, 0},{37002, 1, 1, 34}, {0, 0, 0, 0}}},{58870, {{37002, 1, 0, 0},{37002, 1, 1, 33}, {0, 0, 0, 0}}},{58870, {{37002, 1, 0, 0},{37002, 1, 1, 33}, {0, 0, 0, 0}}},{60240, {{37002, 1, 0, 0},{37002, 1, 1, 25}, {0, 0, 0, 0}}},{60240, {{37002, 1, 0, 0},{37002, 1, 1, 25}, {0, 0, 0, 0}}}}},
    {4, {{26209, {{37003, 1, 0, 0},{37003, 1, 1, 18}, {0, 0, 0, 0}}},{26209, {{37003, 1, 0, 0},{37003, 1, 1, 18}, {0, 0, 0, 0}}},{26210, {{37003, 1, 0, 0},{37003, 1, 1, 17}, {0, 0, 0, 0}}},{26210, {{37003, 1, 0, 0},{37003, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{26211, {{37004, 1, 0, 0},{37004, 1, 1, 18}, {0, 0, 0, 0}}},{26211, {{37004, 1, 0, 0},{37004, 1, 1, 18}, {0, 0, 0, 0}}},{26212, {{37004, 1, 0, 0},{37004, 1, 1, 17}, {0, 0, 0, 0}}},{26212, {{37004, 1, 0, 0},{37004, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {6, {{26213, {{37005, 1, 0, 0},{37005, 1, 1, 16}, {0, 0, 0, 0}}},{26213, {{37005, 1, 0, 0},{37005, 1, 1, 16}, {0, 0, 0, 0}}},{26214, {{37005, 1, 0, 0},{37005, 1, 1, 17}, {0, 0, 0, 0}}},{26214, {{37005, 1, 0, 0},{37005, 1, 1, 17}, {0, 0, 0, 0}}},{58874, {{37005, 1, 0, 0},{37005, 1, 1, 18}, {0, 0, 0, 0}}},{58874, {{37005, 1, 0, 0},{37005, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{26215, {{37006, 1, 0, 0},{37006, 1, 1, 18}, {0, 0, 0, 0}}},{26215, {{37006, 1, 0, 0},{37006, 1, 1, 18}, {0, 0, 0, 0}}},{26216, {{37006, 1, 0, 0},{37006, 1, 1, 17}, {0, 0, 0, 0}}},{26216, {{37006, 1, 0, 0},{37006, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{26217, {{37007, 1, 0, 0},{37007, 1, 1, 18}, {0, 0, 0, 0}}},{26217, {{37007, 1, 0, 0},{37007, 1, 1, 18}, {0, 0, 0, 0}}},{26218, {{37007, 1, 0, 0},{37007, 1, 1, 17}, {0, 0, 0, 0}}},{26218, {{37007, 1, 0, 0},{37007, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{26219, {{37008, 1, 0, 0},{37008, 1, 1, 18}, {0, 0, 0, 0}}},{26219, {{37008, 1, 0, 0},{37008, 1, 1, 18}, {0, 0, 0, 0}}},{26220, {{37008, 1, 0, 0},{37008, 1, 1, 17}, {0, 0, 0, 0}}},{26220, {{37008, 1, 0, 0},{37008, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{26221, {{37009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26222, {{37010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26223, {{37011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26224, {{37012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26225, {{37013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26226, {{37014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26227, {{37015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26228, {{37016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26229, {{37017, 1, 0, 0},{37017, 1, 1, 17}, {0, 0, 0, 0}}},{26229, {{37017, 1, 0, 0},{37017, 1, 1, 17}, {0, 0, 0, 0}}},{26230, {{37017, 1, 0, 0},{37017, 1, 1, 18}, {0, 0, 0, 0}}},{26230, {{37017, 1, 0, 0},{37017, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26231, {{37018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26232, {{37019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{26233, {{37022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26234, {{37023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26235, {{37024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26236, {{37025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26237, {{37026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26238, {{37027, 1, 0, 0},{37027, 1, 1, 19}, {0, 0, 0, 0}}},{26238, {{37027, 1, 0, 0},{37027, 1, 1, 19}, {0, 0, 0, 0}}},{26239, {{37027, 1, 0, 0},{37027, 1, 1, 18}, {0, 0, 0, 0}}},{26239, {{37027, 1, 0, 0},{37027, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26240, {{37028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26241, {{37029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {10, {{26242, {{37030, 1, 0, 0},{37030, 1, 1, 19}, {0, 0, 0, 0}}},{26242, {{37030, 1, 0, 0},{37030, 1, 1, 19}, {0, 0, 0, 0}}},{26243, {{37030, 1, 0, 0},{37030, 1, 1, 22}, {0, 0, 0, 0}}},{26243, {{37030, 1, 0, 0},{37030, 1, 1, 22}, {0, 0, 0, 0}}},{26244, {{37030, 1, 0, 0},{37030, 1, 1, 21}, {0, 0, 0, 0}}},{26244, {{37030, 1, 0, 0},{37030, 1, 1, 21}, {0, 0, 0, 0}}},{26245, {{37030, 1, 0, 0},{37030, 1, 1, 24}, {0, 0, 0, 0}}},{26245, {{37030, 1, 0, 0},{37030, 1, 1, 24}, {0, 0, 0, 0}}},{26246, {{37030, 1, 0, 0},{37030, 1, 1, 20}, {0, 0, 0, 0}}},{26246, {{37030, 1, 0, 0},{37030, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{26248, {{37031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26249, {{37032, 1, 0, 0},{37032, 1, 1, 18}, {0, 0, 0, 0}}},{26249, {{37032, 1, 0, 0},{37032, 1, 1, 18}, {0, 0, 0, 0}}},{26250, {{37032, 1, 0, 0},{37032, 1, 1, 19}, {0, 0, 0, 0}}},{26250, {{37032, 1, 0, 0},{37032, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{26253, {{37033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{26254, {{37034, 1, 0, 0},{37034, 1, 1, 20}, {0, 0, 0, 0}}},{26254, {{37034, 1, 0, 0},{37034, 1, 1, 20}, {0, 0, 0, 0}}},{26255, {{37034, 1, 0, 0},{37034, 1, 1, 19}, {0, 0, 0, 0}}},{26255, {{37034, 1, 0, 0},{37034, 1, 1, 19}, {0, 0, 0, 0}}},{26256, {{37034, 1, 0, 0},{37034, 1, 1, 21}, {0, 0, 0, 0}}},{26256, {{37034, 1, 0, 0},{37034, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {4, {{26257, {{37035, 1, 0, 0},{37035, 1, 1, 16}, {0, 0, 0, 0}}},{26257, {{37035, 1, 0, 0},{37035, 1, 1, 16}, {0, 0, 0, 0}}},{26258, {{37035, 1, 0, 0},{37035, 1, 1, 17}, {0, 0, 0, 0}}},{26258, {{37035, 1, 0, 0},{37035, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{26261, {{37037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26262, {{37038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26263, {{37039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26264, {{37040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26265, {{37041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26266, {{37042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26267, {{37043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26268, {{37044, 1, 0, 0},{37044, 1, 1, 17}, {0, 0, 0, 0}}},{26268, {{37044, 1, 0, 0},{37044, 1, 1, 17}, {0, 0, 0, 0}}},{26269, {{37044, 1, 0, 0},{37044, 1, 1, 18}, {0, 0, 0, 0}}},{26269, {{37044, 1, 0, 0},{37044, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26270, {{37045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26271, {{37046, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26272, {{37047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26273, {{37048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26275, {{37049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{26276, {{37052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26277, {{37053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26278, {{37054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26279, {{37055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26280, {{37056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26281, {{37057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26282, {{37058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26283, {{37059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26284, {{37060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26285, {{37061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26286, {{37062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26287, {{37063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26288, {{37064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26289, {{37065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26290, {{37066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26291, {{37067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26292, {{37068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26293, {{37069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26294, {{37070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{26295, {{37073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26296, {{37074, 1, 0, 0},{37074, 1, 1, 17}, {0, 0, 0, 0}}},{26296, {{37074, 1, 0, 0},{37074, 1, 1, 17}, {0, 0, 0, 0}}},{26297, {{37074, 1, 0, 0},{37074, 1, 1, 18}, {0, 0, 0, 0}}},{26297, {{37074, 1, 0, 0},{37074, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{26298, {{37076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26299, {{37077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26300, {{37078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26301, {{37079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26302, {{37080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26303, {{37081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26304, {{37082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26305, {{37083, 1, 0, 0},{37083, 1, 1, 17}, {0, 0, 0, 0}}},{26305, {{37083, 1, 0, 0},{37083, 1, 1, 17}, {0, 0, 0, 0}}},{26306, {{37083, 1, 0, 0},{37083, 1, 1, 18}, {0, 0, 0, 0}}},{26306, {{37083, 1, 0, 0},{37083, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26307, {{37084, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26308, {{37085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26309, {{37086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26310, {{37087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26311, {{37088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26312, {{37089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{26313, {{37090, 1, 0, 0},{37090, 1, 1, 17}, {0, 0, 0, 0}}},{26313, {{37090, 1, 0, 0},{37090, 1, 1, 17}, {0, 0, 0, 0}}},{26314, {{37090, 1, 0, 0},{37090, 1, 1, 18}, {0, 0, 0, 0}}},{26314, {{37090, 1, 0, 0},{37090, 1, 1, 18}, {0, 0, 0, 0}}},{26315, {{37090, 1, 0, 0},{37090, 1, 1, 19}, {0, 0, 0, 0}}},{26315, {{37090, 1, 0, 0},{37090, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{26316, {{37091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26317, {{37092, 1, 0, 0},{37092, 1, 1, 17}, {0, 0, 0, 0}}},{26317, {{37092, 1, 0, 0},{37092, 1, 1, 17}, {0, 0, 0, 0}}},{26318, {{37092, 1, 0, 0},{37092, 1, 1, 18}, {0, 0, 0, 0}}},{26318, {{37092, 1, 0, 0},{37092, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26319, {{37093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{26320, {{37096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26321, {{37097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26322, {{37098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26323, {{37099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26324, {{37100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26325, {{37101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26326, {{37102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26327, {{37103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26328, {{37104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26329, {{37105, 1, 0, 0},{37105, 1, 1, 17}, {0, 0, 0, 0}}},{26329, {{37105, 1, 0, 0},{37105, 1, 1, 17}, {0, 0, 0, 0}}},{26330, {{37105, 1, 0, 0},{37105, 1, 1, 16}, {0, 0, 0, 0}}},{26330, {{37105, 1, 0, 0},{37105, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{26331, {{37106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26332, {{37107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26333, {{37108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26334, {{37109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26335, {{37110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26336, {{37111, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{26337, {{37113, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26338, {{37114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26339, {{37115, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26340, {{37116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{26341, {{37117, 1, 0, 0},{37117, 1, 1, 18}, {0, 0, 0, 0}}},{26341, {{37117, 1, 0, 0},{37117, 1, 1, 18}, {0, 0, 0, 0}}},{30235, {{37117, 1, 0, 0},{37117, 1, 1, 0},{37117, 1, 1, 19},{64038, 1, 0, 0}, {0, 0, 0, 0}}},{30235, {{37117, 1, 0, 0},{37117, 1, 1, 0},{37117, 1, 1, 19},{64038, 1, 0, 0}, {0, 0, 0, 0}}},{30235, {{37117, 1, 0, 0},{37117, 1, 1, 0},{37117, 1, 1, 19},{64038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26342, {{37118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26343, {{37119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26344, {{37120, 1, 0, 0},{37120, 1, 1, 17}, {0, 0, 0, 0}}},{26344, {{37120, 1, 0, 0},{37120, 1, 1, 17}, {0, 0, 0, 0}}},{26345, {{37120, 1, 0, 0},{37120, 1, 1, 18}, {0, 0, 0, 0}}},{26345, {{37120, 1, 0, 0},{37120, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26346, {{37121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26347, {{37122, 1, 0, 0},{37122, 1, 1, 17}, {0, 0, 0, 0}}},{26347, {{37122, 1, 0, 0},{37122, 1, 1, 17}, {0, 0, 0, 0}}},{26348, {{37122, 1, 0, 0},{37122, 1, 1, 18}, {0, 0, 0, 0}}},{26348, {{37122, 1, 0, 0},{37122, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26349, {{37123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26350, {{37124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26351, {{37125, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26352, {{37126, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26353, {{37127, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26354, {{37128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26355, {{37129, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26356, {{37130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26357, {{37131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26358, {{37132, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26359, {{37133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26360, {{37134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26361, {{37135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26362, {{37136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26363, {{37137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26364, {{37138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{26365, {{37140, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{26366, {{37141, 1, 0, 0},{37141, 1, 1, 18}, {0, 0, 0, 0}}},{26366, {{37141, 1, 0, 0},{37141, 1, 1, 18}, {0, 0, 0, 0}}},{26367, {{37141, 1, 0, 0},{37141, 1, 1, 17}, {0, 0, 0, 0}}},{26367, {{37141, 1, 0, 0},{37141, 1, 1, 17}, {0, 0, 0, 0}}},{26368, {{37141, 1, 0, 0},{37141, 1, 1, 21}, {0, 0, 0, 0}}},{26368, {{37141, 1, 0, 0},{37141, 1, 1, 21}, {0, 0, 0, 0}}},{26369, {{37141, 1, 0, 0},{37141, 1, 1, 19}, {0, 0, 0, 0}}},{26369, {{37141, 1, 0, 0},{37141, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{26371, {{37142, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26372, {{37143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26373, {{37144, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26374, {{37145, 1, 0, 0},{37145, 1, 1, 17}, {0, 0, 0, 0}}},{26374, {{37145, 1, 0, 0},{37145, 1, 1, 17}, {0, 0, 0, 0}}},{58878, {{37145, 1, 0, 0},{37145, 1, 1, 18}, {0, 0, 0, 0}}},{58878, {{37145, 1, 0, 0},{37145, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{26375, {{37146, 1, 0, 0},{37146, 1, 1, 17}, {0, 0, 0, 0}}},{26375, {{37146, 1, 0, 0},{37146, 1, 1, 17}, {0, 0, 0, 0}}},{26376, {{37146, 1, 0, 0},{37146, 1, 1, 18}, {0, 0, 0, 0}}},{26376, {{37146, 1, 0, 0},{37146, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{26377, {{37147, 1, 0, 0},{37147, 1, 1, 16}, {0, 0, 0, 0}}},{26377, {{37147, 1, 0, 0},{37147, 1, 1, 16}, {0, 0, 0, 0}}},{26378, {{37147, 1, 0, 0},{37147, 1, 1, 17}, {0, 0, 0, 0}}},{26378, {{37147, 1, 0, 0},{37147, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{26379, {{37148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26380, {{37149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26381, {{37150, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26383, {{37151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26384, {{37152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26385, {{37153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26386, {{37154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26387, {{37155, 1, 0, 0},{37155, 1, 1, 17}, {0, 0, 0, 0}}},{26387, {{37155, 1, 0, 0},{37155, 1, 1, 17}, {0, 0, 0, 0}}},{26388, {{37155, 1, 0, 0},{37155, 1, 1, 18}, {0, 0, 0, 0}}},{26388, {{37155, 1, 0, 0},{37155, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26389, {{37156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26390, {{37157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26391, {{37158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26392, {{37159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26393, {{37160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26394, {{37161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26395, {{37162, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26396, {{37163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26397, {{37164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26398, {{37165, 1, 0, 0},{37165, 1, 1, 19}, {0, 0, 0, 0}}},{26398, {{37165, 1, 0, 0},{37165, 1, 1, 19}, {0, 0, 0, 0}}},{26399, {{37165, 1, 0, 0},{37165, 1, 1, 18}, {0, 0, 0, 0}}},{26399, {{37165, 1, 0, 0},{37165, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{26400, {{37166, 1, 0, 0},{37166, 1, 1, 17}, {0, 0, 0, 0}}},{26400, {{37166, 1, 0, 0},{37166, 1, 1, 17}, {0, 0, 0, 0}}},{26401, {{37166, 1, 0, 0},{37166, 1, 1, 18}, {0, 0, 0, 0}}},{26401, {{37166, 1, 0, 0},{37166, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26402, {{37167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{26403, {{37168, 1, 0, 0},{37168, 1, 1, 18}, {0, 0, 0, 0}}},{26403, {{37168, 1, 0, 0},{37168, 1, 1, 18}, {0, 0, 0, 0}}},{26404, {{37168, 1, 0, 0},{37168, 1, 1, 19}, {0, 0, 0, 0}}},{26404, {{37168, 1, 0, 0},{37168, 1, 1, 19}, {0, 0, 0, 0}}},{26405, {{37168, 1, 0, 0},{37168, 1, 1, 20}, {0, 0, 0, 0}}},{26405, {{37168, 1, 0, 0},{37168, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{26406, {{37169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26407, {{37170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26408, {{37171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26409, {{37172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26410, {{37173, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26411, {{37174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26412, {{37175, 1, 0, 0},{37175, 1, 1, 17}, {0, 0, 0, 0}}},{26412, {{37175, 1, 0, 0},{37175, 1, 1, 17}, {0, 0, 0, 0}}},{26413, {{37175, 1, 0, 0},{37175, 1, 1, 18}, {0, 0, 0, 0}}},{26413, {{37175, 1, 0, 0},{37175, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26414, {{37176, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26415, {{37177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26416, {{37178, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26417, {{37179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26418, {{37180, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26419, {{37181, 1, 0, 0},{37181, 1, 1, 17}, {0, 0, 0, 0}}},{26419, {{37181, 1, 0, 0},{37181, 1, 1, 17}, {0, 0, 0, 0}}},{26420, {{37181, 1, 0, 0},{37181, 1, 1, 18}, {0, 0, 0, 0}}},{26420, {{37181, 1, 0, 0},{37181, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26421, {{37182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26422, {{37183, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26423, {{37184, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26424, {{37185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{26425, {{37187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26426, {{37188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26427, {{37189, 1, 0, 0},{37189, 1, 1, 16}, {0, 0, 0, 0}}},{26427, {{37189, 1, 0, 0},{37189, 1, 1, 16}, {0, 0, 0, 0}}},{26428, {{37189, 1, 0, 0},{37189, 1, 1, 17}, {0, 0, 0, 0}}},{26428, {{37189, 1, 0, 0},{37189, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{26429, {{37190, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26430, {{37191, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26431, {{37192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26432, {{37193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26433, {{37194, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{26434, {{37195, 1, 0, 0},{37195, 1, 1, 19}, {0, 0, 0, 0}}},{26434, {{37195, 1, 0, 0},{37195, 1, 1, 19}, {0, 0, 0, 0}}},{26435, {{37195, 1, 0, 0},{37195, 1, 1, 18}, {0, 0, 0, 0}}},{26435, {{37195, 1, 0, 0},{37195, 1, 1, 18}, {0, 0, 0, 0}}},{59344, {{37195, 1, 0, 0},{37195, 1, 1, 21}, {0, 0, 0, 0}}},{59344, {{37195, 1, 0, 0},{37195, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {4, {{26437, {{37196, 1, 0, 0},{37196, 1, 1, 19}, {0, 0, 0, 0}}},{26437, {{37196, 1, 0, 0},{37196, 1, 1, 19}, {0, 0, 0, 0}}},{26438, {{37196, 1, 0, 0},{37196, 1, 1, 18}, {0, 0, 0, 0}}},{26438, {{37196, 1, 0, 0},{37196, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {8, {{26439, {{37197, 1, 0, 0},{37197, 1, 1, 21}, {0, 0, 0, 0}}},{26439, {{37197, 1, 0, 0},{37197, 1, 1, 21}, {0, 0, 0, 0}}},{26440, {{37197, 1, 0, 0},{37197, 1, 1, 19}, {0, 0, 0, 0}}},{26440, {{37197, 1, 0, 0},{37197, 1, 1, 19}, {0, 0, 0, 0}}},{26441, {{37197, 1, 0, 0},{37197, 1, 1, 22}, {0, 0, 0, 0}}},{26441, {{37197, 1, 0, 0},{37197, 1, 1, 22}, {0, 0, 0, 0}}},{26442, {{37197, 1, 0, 0},{37197, 1, 1, 20}, {0, 0, 0, 0}}},{26442, {{37197, 1, 0, 0},{37197, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{26443, {{37198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26444, {{37199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26445, {{37200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26446, {{37201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26447, {{37202, 1, 0, 0},{37202, 1, 1, 17}, {0, 0, 0, 0}}},{26447, {{37202, 1, 0, 0},{37202, 1, 1, 17}, {0, 0, 0, 0}}},{26448, {{37202, 1, 0, 0},{37202, 1, 1, 18}, {0, 0, 0, 0}}},{26448, {{37202, 1, 0, 0},{37202, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26450, {{37203, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26451, {{37204, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26452, {{37205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26453, {{37206, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26454, {{37207, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26455, {{37208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26456, {{37209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26457, {{37210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26458, {{37211, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{26459, {{37215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26460, {{37216, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26461, {{37217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26462, {{37218, 1, 0, 0},{37218, 1, 1, 17}, {0, 0, 0, 0}}},{26462, {{37218, 1, 0, 0},{37218, 1, 1, 17}, {0, 0, 0, 0}}},{26463, {{37218, 1, 0, 0},{37218, 1, 1, 18}, {0, 0, 0, 0}}},{26463, {{37218, 1, 0, 0},{37218, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26464, {{37219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26465, {{37220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26466, {{37221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{26467, {{37223, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26468, {{37224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26469, {{37225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26470, {{37226, 1, 0, 0},{37226, 1, 1, 17}, {0, 0, 0, 0}}},{26470, {{37226, 1, 0, 0},{37226, 1, 1, 17}, {0, 0, 0, 0}}},{26471, {{37226, 1, 0, 0},{37226, 1, 1, 18}, {0, 0, 0, 0}}},{26471, {{37226, 1, 0, 0},{37226, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26472, {{37227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26473, {{37228, 1, 0, 0},{37228, 1, 1, 17}, {0, 0, 0, 0}}},{26473, {{37228, 1, 0, 0},{37228, 1, 1, 17}, {0, 0, 0, 0}}},{26474, {{37228, 1, 0, 0},{37228, 1, 1, 18}, {0, 0, 0, 0}}},{26474, {{37228, 1, 0, 0},{37228, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26475, {{37229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26476, {{37230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26477, {{37231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{26478, {{37234, 1, 0, 0},{37234, 1, 1, 18}, {0, 0, 0, 0}}},{26478, {{37234, 1, 0, 0},{37234, 1, 1, 18}, {0, 0, 0, 0}}},{26479, {{37234, 1, 0, 0},{37234, 1, 1, 19}, {0, 0, 0, 0}}},{26479, {{37234, 1, 0, 0},{37234, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{26480, {{37235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26481, {{37236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26482, {{37237, 1, 0, 0},{37237, 1, 1, 17}, {0, 0, 0, 0}}},{26482, {{37237, 1, 0, 0},{37237, 1, 1, 17}, {0, 0, 0, 0}}},{26483, {{37237, 1, 0, 0},{37237, 1, 1, 18}, {0, 0, 0, 0}}},{26483, {{37237, 1, 0, 0},{37237, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26484, {{37238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26485, {{37239, 1, 0, 0},{37239, 1, 1, 19}, {0, 0, 0, 0}}},{26485, {{37239, 1, 0, 0},{37239, 1, 1, 19}, {0, 0, 0, 0}}},{26486, {{37239, 1, 0, 0},{37239, 1, 1, 18}, {0, 0, 0, 0}}},{26486, {{37239, 1, 0, 0},{37239, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{26487, {{37240, 1, 0, 0},{37240, 1, 1, 18}, {0, 0, 0, 0}}},{26487, {{37240, 1, 0, 0},{37240, 1, 1, 18}, {0, 0, 0, 0}}},{26488, {{37240, 1, 0, 0},{37240, 1, 1, 17}, {0, 0, 0, 0}}},{26488, {{37240, 1, 0, 0},{37240, 1, 1, 17}, {0, 0, 0, 0}}},{26489, {{37240, 1, 0, 0},{37240, 1, 1, 19}, {0, 0, 0, 0}}},{26489, {{37240, 1, 0, 0},{37240, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{26490, {{37241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26491, {{37242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26492, {{37243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{26494, {{37248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26495, {{37249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26498, {{37250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26499, {{37251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26500, {{37252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26501, {{37253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26502, {{37254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26503, {{37255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26504, {{37256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26505, {{37257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26506, {{37258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26507, {{37259, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26508, {{37260, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26509, {{37261, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26510, {{37262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26511, {{37263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26512, {{37264, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26513, {{37265, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26514, {{37266, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26515, {{37267, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26516, {{37268, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26517, {{37269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26519, {{37270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26520, {{37271, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26521, {{37272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26522, {{37273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26523, {{37274, 1, 0, 0},{37274, 1, 1, 16}, {0, 0, 0, 0}}},{26523, {{37274, 1, 0, 0},{37274, 1, 1, 16}, {0, 0, 0, 0}}},{26524, {{37274, 1, 0, 0},{37274, 1, 1, 17}, {0, 0, 0, 0}}},{26524, {{37274, 1, 0, 0},{37274, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{26525, {{37275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26526, {{37276, 1, 0, 0},{37276, 1, 1, 18}, {0, 0, 0, 0}}},{26526, {{37276, 1, 0, 0},{37276, 1, 1, 18}, {0, 0, 0, 0}}},{26527, {{37276, 1, 0, 0},{37276, 1, 1, 17}, {0, 0, 0, 0}}},{26527, {{37276, 1, 0, 0},{37276, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{26528, {{37277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26529, {{37278, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26530, {{37279, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26531, {{37280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26532, {{37281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26533, {{37282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26534, {{37283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26535, {{37284, 1, 0, 0},{37284, 1, 1, 17}, {0, 0, 0, 0}}},{26535, {{37284, 1, 0, 0},{37284, 1, 1, 17}, {0, 0, 0, 0}}},{26536, {{37284, 1, 0, 0},{37284, 1, 1, 18}, {0, 0, 0, 0}}},{26536, {{37284, 1, 0, 0},{37284, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26538, {{37285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26539, {{37286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26540, {{37287, 1, 0, 0},{37287, 1, 1, 17}, {0, 0, 0, 0}}},{26540, {{37287, 1, 0, 0},{37287, 1, 1, 17}, {0, 0, 0, 0}}},{26541, {{37287, 1, 0, 0},{37287, 1, 1, 18}, {0, 0, 0, 0}}},{26541, {{37287, 1, 0, 0},{37287, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{26542, {{37288, 1, 0, 0},{37288, 1, 1, 17}, {0, 0, 0, 0}}},{26542, {{37288, 1, 0, 0},{37288, 1, 1, 17}, {0, 0, 0, 0}}},{26543, {{37288, 1, 0, 0},{37288, 1, 1, 18}, {0, 0, 0, 0}}},{26543, {{37288, 1, 0, 0},{37288, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26544, {{37289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26545, {{37290, 1, 0, 0},{37290, 1, 1, 18}, {0, 0, 0, 0}}},{26545, {{37290, 1, 0, 0},{37290, 1, 1, 18}, {0, 0, 0, 0}}},{26546, {{37290, 1, 0, 0},{37290, 1, 1, 19}, {0, 0, 0, 0}}},{26546, {{37290, 1, 0, 0},{37290, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{26547, {{37291, 1, 0, 0},{37291, 1, 1, 17}, {0, 0, 0, 0}}},{26547, {{37291, 1, 0, 0},{37291, 1, 1, 17}, {0, 0, 0, 0}}},{26548, {{37291, 1, 0, 0},{37291, 1, 1, 18}, {0, 0, 0, 0}}},{26548, {{37291, 1, 0, 0},{37291, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{26549, {{37292, 1, 0, 0},{37292, 1, 1, 17}, {0, 0, 0, 0}}},{26549, {{37292, 1, 0, 0},{37292, 1, 1, 17}, {0, 0, 0, 0}}},{26550, {{37292, 1, 0, 0},{37292, 1, 1, 18}, {0, 0, 0, 0}}},{26550, {{37292, 1, 0, 0},{37292, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26551, {{37293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26552, {{37294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26553, {{37295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26554, {{37296, 1, 0, 0},{37296, 1, 1, 17}, {0, 0, 0, 0}}},{26554, {{37296, 1, 0, 0},{37296, 1, 1, 17}, {0, 0, 0, 0}}},{26555, {{37296, 1, 0, 0},{37296, 1, 1, 18}, {0, 0, 0, 0}}},{26555, {{37296, 1, 0, 0},{37296, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26556, {{37297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26557, {{37298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26558, {{37299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26559, {{37300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{26560, {{37301, 1, 0, 0},{37301, 1, 1, 17}, {0, 0, 0, 0}}},{26560, {{37301, 1, 0, 0},{37301, 1, 1, 17}, {0, 0, 0, 0}}},{26561, {{37301, 1, 0, 0},{37301, 1, 1, 18}, {0, 0, 0, 0}}},{26561, {{37301, 1, 0, 0},{37301, 1, 1, 18}, {0, 0, 0, 0}}},{26562, {{37301, 1, 0, 0},{37301, 1, 1, 19}, {0, 0, 0, 0}}},{26562, {{37301, 1, 0, 0},{37301, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{26563, {{37302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26564, {{37303, 1, 0, 0},{37303, 1, 1, 16}, {0, 0, 0, 0}}},{26564, {{37303, 1, 0, 0},{37303, 1, 1, 16}, {0, 0, 0, 0}}},{26565, {{37303, 1, 0, 0},{37303, 1, 1, 17}, {0, 0, 0, 0}}},{26565, {{37303, 1, 0, 0},{37303, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{26566, {{37304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26567, {{37305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26568, {{37306, 1, 0, 0},{37306, 1, 1, 18}, {0, 0, 0, 0}}},{26568, {{37306, 1, 0, 0},{37306, 1, 1, 18}, {0, 0, 0, 0}}},{26569, {{37306, 1, 0, 0},{37306, 1, 1, 19}, {0, 0, 0, 0}}},{26569, {{37306, 1, 0, 0},{37306, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{26570, {{37307, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26571, {{37308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26572, {{37309, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26573, {{37310, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26574, {{37311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26575, {{37312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {10, {{26576, {{37313, 1, 0, 0},{37313, 1, 1, 21}, {0, 0, 0, 0}}},{26576, {{37313, 1, 0, 0},{37313, 1, 1, 21}, {0, 0, 0, 0}}},{26577, {{37313, 1, 0, 0},{37313, 1, 1, 18}, {0, 0, 0, 0}}},{26577, {{37313, 1, 0, 0},{37313, 1, 1, 18}, {0, 0, 0, 0}}},{26578, {{37313, 1, 0, 0},{37313, 1, 1, 22}, {0, 0, 0, 0}}},{26578, {{37313, 1, 0, 0},{37313, 1, 1, 22}, {0, 0, 0, 0}}},{26579, {{37313, 1, 0, 0},{37313, 1, 1, 19}, {0, 0, 0, 0}}},{26579, {{37313, 1, 0, 0},{37313, 1, 1, 19}, {0, 0, 0, 0}}},{26580, {{37313, 1, 0, 0},{37313, 1, 1, 20}, {0, 0, 0, 0}}},{26580, {{37313, 1, 0, 0},{37313, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{26581, {{37314, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26582, {{37315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26583, {{37316, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26584, {{37317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26585, {{37318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26586, {{37319, 1, 0, 0},{37319, 1, 1, 18}, {0, 0, 0, 0}}},{26586, {{37319, 1, 0, 0},{37319, 1, 1, 18}, {0, 0, 0, 0}}},{26587, {{37319, 1, 0, 0},{37319, 1, 1, 19}, {0, 0, 0, 0}}},{26587, {{37319, 1, 0, 0},{37319, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{26588, {{37320, 1, 0, 0},{37320, 1, 1, 17}, {0, 0, 0, 0}}},{26588, {{37320, 1, 0, 0},{37320, 1, 1, 17}, {0, 0, 0, 0}}},{58893, {{37320, 1, 0, 0},{37320, 1, 1, 18}, {0, 0, 0, 0}}},{58893, {{37320, 1, 0, 0},{37320, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26589, {{37321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{26590, {{37323, 1, 0, 0},{37323, 1, 1, 17}, {0, 0, 0, 0}}},{26590, {{37323, 1, 0, 0},{37323, 1, 1, 17}, {0, 0, 0, 0}}},{26591, {{37323, 1, 0, 0},{37323, 1, 1, 18}, {0, 0, 0, 0}}},{26591, {{37323, 1, 0, 0},{37323, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26592, {{37324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26593, {{37325, 1, 0, 0},{37325, 1, 1, 17}, {0, 0, 0, 0}}},{26593, {{37325, 1, 0, 0},{37325, 1, 1, 17}, {0, 0, 0, 0}}},{60258, {{37325, 1, 0, 0},{37325, 1, 1, 18}, {0, 0, 0, 0}}},{60258, {{37325, 1, 0, 0},{37325, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26594, {{37326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26595, {{37327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26596, {{37328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26597, {{37329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{26598, {{37331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26599, {{37332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26600, {{37333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26601, {{37334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26602, {{37335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26603, {{37336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26604, {{37337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26605, {{37338, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26606, {{37339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26607, {{37340, 1, 0, 0},{37340, 1, 1, 18}, {0, 0, 0, 0}}},{26607, {{37340, 1, 0, 0},{37340, 1, 1, 18}, {0, 0, 0, 0}}},{26608, {{37340, 1, 0, 0},{37340, 1, 1, 19}, {0, 0, 0, 0}}},{26608, {{37340, 1, 0, 0},{37340, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{26609, {{37341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26610, {{37342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26611, {{37343, 1, 0, 0},{37343, 1, 1, 17}, {0, 0, 0, 0}}},{26611, {{37343, 1, 0, 0},{37343, 1, 1, 17}, {0, 0, 0, 0}}},{26612, {{37343, 1, 0, 0},{37343, 1, 1, 18}, {0, 0, 0, 0}}},{26612, {{37343, 1, 0, 0},{37343, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26613, {{37344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26614, {{37345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26615, {{37346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26616, {{37347, 1, 0, 0},{37347, 1, 1, 19}, {0, 0, 0, 0}}},{26616, {{37347, 1, 0, 0},{37347, 1, 1, 19}, {0, 0, 0, 0}}},{26617, {{37347, 1, 0, 0},{37347, 1, 1, 18}, {0, 0, 0, 0}}},{26617, {{37347, 1, 0, 0},{37347, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26618, {{37348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26619, {{37349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26620, {{37350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26621, {{37351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26622, {{37352, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26623, {{37353, 1, 0, 0},{37353, 1, 1, 17}, {0, 0, 0, 0}}},{26623, {{37353, 1, 0, 0},{37353, 1, 1, 17}, {0, 0, 0, 0}}},{26624, {{37353, 1, 0, 0},{37353, 1, 1, 18}, {0, 0, 0, 0}}},{26624, {{37353, 1, 0, 0},{37353, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26625, {{37354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26626, {{37355, 1, 0, 0},{37355, 1, 1, 16}, {0, 0, 0, 0}}},{26626, {{37355, 1, 0, 0},{37355, 1, 1, 16}, {0, 0, 0, 0}}},{26627, {{37355, 1, 0, 0},{37355, 1, 1, 17}, {0, 0, 0, 0}}},{26627, {{37355, 1, 0, 0},{37355, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{26628, {{37356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26629, {{37357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26630, {{37358, 1, 0, 0},{37358, 1, 1, 17}, {0, 0, 0, 0}}},{26630, {{37358, 1, 0, 0},{37358, 1, 1, 17}, {0, 0, 0, 0}}},{26631, {{37358, 1, 0, 0},{37358, 1, 1, 18}, {0, 0, 0, 0}}},{26631, {{37358, 1, 0, 0},{37358, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{26632, {{37359, 1, 0, 0},{37359, 1, 1, 17}, {0, 0, 0, 0}}},{26632, {{37359, 1, 0, 0},{37359, 1, 1, 17}, {0, 0, 0, 0}}},{26633, {{37359, 1, 0, 0},{37359, 1, 1, 18}, {0, 0, 0, 0}}},{26633, {{37359, 1, 0, 0},{37359, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26634, {{37360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26635, {{37361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26636, {{37362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26637, {{37363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26638, {{37364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26639, {{37365, 1, 0, 0},{37365, 1, 1, 17}, {0, 0, 0, 0}}},{26639, {{37365, 1, 0, 0},{37365, 1, 1, 17}, {0, 0, 0, 0}}},{26640, {{37365, 1, 0, 0},{37365, 1, 1, 18}, {0, 0, 0, 0}}},{26640, {{37365, 1, 0, 0},{37365, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26641, {{37366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26642, {{37367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26643, {{37368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26644, {{37369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{26646, {{37371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{26647, {{37372, 1, 0, 0},{37372, 1, 1, 20}, {0, 0, 0, 0}}},{26647, {{37372, 1, 0, 0},{37372, 1, 1, 20}, {0, 0, 0, 0}}},{26648, {{37372, 1, 0, 0},{37372, 1, 1, 19}, {0, 0, 0, 0}}},{26648, {{37372, 1, 0, 0},{37372, 1, 1, 19}, {0, 0, 0, 0}}},{26649, {{37372, 1, 0, 0},{37372, 1, 1, 21}, {0, 0, 0, 0}}},{26649, {{37372, 1, 0, 0},{37372, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{26650, {{37373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26651, {{37374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26653, {{37375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26654, {{37376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26655, {{37377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26656, {{37378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26657, {{37379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26658, {{37380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26659, {{37381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26660, {{37382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26661, {{37383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26662, {{37384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26663, {{37385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26665, {{37386, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26666, {{37387, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26667, {{37388, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26668, {{37389, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26670, {{37390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26671, {{37391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26672, {{37392, 1, 0, 0},{37392, 1, 1, 17}, {0, 0, 0, 0}}},{26672, {{37392, 1, 0, 0},{37392, 1, 1, 17}, {0, 0, 0, 0}}},{26673, {{37392, 1, 0, 0},{37392, 1, 1, 18}, {0, 0, 0, 0}}},{26673, {{37392, 1, 0, 0},{37392, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26674, {{37393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26675, {{37394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26677, {{37395, 1, 0, 0},{37395, 1, 1, 17}, {0, 0, 0, 0}}},{26677, {{37395, 1, 0, 0},{37395, 1, 1, 17}, {0, 0, 0, 0}}},{26678, {{37395, 1, 0, 0},{37395, 1, 1, 18}, {0, 0, 0, 0}}},{26678, {{37395, 1, 0, 0},{37395, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26680, {{37396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26681, {{37397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26682, {{37398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26683, {{37399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26684, {{37400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26685, {{37401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26686, {{37402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26687, {{37403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26688, {{37404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26689, {{37405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26690, {{37406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26691, {{37407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26692, {{37408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26693, {{37409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{26694, {{37411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26695, {{37412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26696, {{37413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26697, {{37414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26698, {{37415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26700, {{37416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26701, {{37417, 1, 0, 0},{37417, 1, 1, 17}, {0, 0, 0, 0}}},{26701, {{37417, 1, 0, 0},{37417, 1, 1, 17}, {0, 0, 0, 0}}},{26702, {{37417, 1, 0, 0},{37417, 1, 1, 18}, {0, 0, 0, 0}}},{26702, {{37417, 1, 0, 0},{37417, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{26704, {{37420, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26705, {{37421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26706, {{37422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26707, {{37423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26708, {{37424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26709, {{37425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26710, {{37426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26711, {{37427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26712, {{37428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26714, {{37429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26715, {{37430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26716, {{37431, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26717, {{37432, 1, 0, 0},{37432, 1, 1, 17}, {0, 0, 0, 0}}},{26717, {{37432, 1, 0, 0},{37432, 1, 1, 17}, {0, 0, 0, 0}}},{26718, {{37432, 1, 0, 0},{37432, 1, 1, 18}, {0, 0, 0, 0}}},{26718, {{37432, 1, 0, 0},{37432, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26719, {{37433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26720, {{37434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26721, {{37435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26722, {{37436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26723, {{37437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26724, {{37438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26725, {{37439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26726, {{37440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26727, {{37441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26728, {{37442, 1, 0, 0},{37442, 1, 1, 17}, {0, 0, 0, 0}}},{26728, {{37442, 1, 0, 0},{37442, 1, 1, 17}, {0, 0, 0, 0}}},{26729, {{37442, 1, 0, 0},{37442, 1, 1, 18}, {0, 0, 0, 0}}},{26729, {{37442, 1, 0, 0},{37442, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26730, {{37443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26731, {{37444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26732, {{37445, 1, 0, 0},{37445, 1, 1, 17}, {0, 0, 0, 0}}},{26732, {{37445, 1, 0, 0},{37445, 1, 1, 17}, {0, 0, 0, 0}}},{26733, {{37445, 1, 0, 0},{37445, 1, 1, 18}, {0, 0, 0, 0}}},{26733, {{37445, 1, 0, 0},{37445, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26734, {{37446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26735, {{37447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26736, {{37448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26737, {{37449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26738, {{37450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26739, {{37451, 1, 0, 0},{37451, 1, 1, 17}, {0, 0, 0, 0}}},{26739, {{37451, 1, 0, 0},{37451, 1, 1, 17}, {0, 0, 0, 0}}},{26740, {{37451, 1, 0, 0},{37451, 1, 1, 18}, {0, 0, 0, 0}}},{26740, {{37451, 1, 0, 0},{37451, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26741, {{37452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26742, {{37453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26743, {{37454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26744, {{37455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26745, {{37456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26746, {{37457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26747, {{37458, 1, 0, 0},{37458, 1, 1, 16}, {0, 0, 0, 0}}},{26747, {{37458, 1, 0, 0},{37458, 1, 1, 16}, {0, 0, 0, 0}}},{26748, {{37458, 1, 0, 0},{37458, 1, 1, 17}, {0, 0, 0, 0}}},{26748, {{37458, 1, 0, 0},{37458, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{26749, {{37459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26750, {{37460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26751, {{37461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26752, {{37462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26753, {{37463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26754, {{37464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26755, {{37465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26756, {{37466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{26757, {{37467, 1, 0, 0},{37467, 1, 1, 19}, {0, 0, 0, 0}}},{26757, {{37467, 1, 0, 0},{37467, 1, 1, 19}, {0, 0, 0, 0}}},{26758, {{37467, 1, 0, 0},{37467, 1, 1, 20}, {0, 0, 0, 0}}},{26758, {{37467, 1, 0, 0},{37467, 1, 1, 20}, {0, 0, 0, 0}}},{26759, {{37467, 1, 0, 0},{37467, 1, 1, 21}, {0, 0, 0, 0}}},{26759, {{37467, 1, 0, 0},{37467, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{26760, {{37468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26761, {{37469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26762, {{37470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26763, {{37471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26764, {{37472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26765, {{37473, 1, 0, 0},{37473, 1, 1, 17}, {0, 0, 0, 0}}},{26765, {{37473, 1, 0, 0},{37473, 1, 1, 17}, {0, 0, 0, 0}}},{26766, {{37473, 1, 0, 0},{37473, 1, 1, 18}, {0, 0, 0, 0}}},{26766, {{37473, 1, 0, 0},{37473, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26767, {{37474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26768, {{37475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26769, {{37476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26770, {{37477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26771, {{37478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26772, {{37479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26773, {{37480, 1, 0, 0},{37480, 1, 1, 17}, {0, 0, 0, 0}}},{26773, {{37480, 1, 0, 0},{37480, 1, 1, 17}, {0, 0, 0, 0}}},{26774, {{37480, 1, 0, 0},{37480, 1, 1, 18}, {0, 0, 0, 0}}},{26774, {{37480, 1, 0, 0},{37480, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26775, {{37481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26776, {{37482, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26777, {{37483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26778, {{37484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26779, {{37485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26780, {{37486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26781, {{37487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26782, {{37488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26783, {{37489, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26784, {{37490, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{26785, {{37493, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26786, {{37494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26787, {{37495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26788, {{37496, 1, 0, 0},{37496, 1, 1, 17}, {0, 0, 0, 0}}},{26788, {{37496, 1, 0, 0},{37496, 1, 1, 17}, {0, 0, 0, 0}}},{26789, {{37496, 1, 0, 0},{37496, 1, 1, 18}, {0, 0, 0, 0}}},{26789, {{37496, 1, 0, 0},{37496, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26790, {{37497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26791, {{37498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26792, {{37499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26793, {{37500, 1, 0, 0},{37500, 1, 1, 17}, {0, 0, 0, 0}}},{26793, {{37500, 1, 0, 0},{37500, 1, 1, 17}, {0, 0, 0, 0}}},{26794, {{37500, 1, 0, 0},{37500, 1, 1, 18}, {0, 0, 0, 0}}},{26794, {{37500, 1, 0, 0},{37500, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26795, {{37501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26796, {{37502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26797, {{37503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26798, {{37504, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26799, {{37505, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26800, {{37506, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26801, {{37507, 1, 0, 0},{37507, 1, 1, 18}, {0, 0, 0, 0}}},{26801, {{37507, 1, 0, 0},{37507, 1, 1, 18}, {0, 0, 0, 0}}},{26802, {{37507, 1, 0, 0},{37507, 1, 1, 17}, {0, 0, 0, 0}}},{26802, {{37507, 1, 0, 0},{37507, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{26804, {{37509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26805, {{37510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26806, {{37511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26808, {{37512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26809, {{37513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26810, {{37514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26812, {{37515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26813, {{37516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26814, {{37517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26815, {{37518, 1, 0, 0},{37518, 1, 1, 17}, {0, 0, 0, 0}}},{26815, {{37518, 1, 0, 0},{37518, 1, 1, 17}, {0, 0, 0, 0}}},{26816, {{37518, 1, 0, 0},{37518, 1, 1, 18}, {0, 0, 0, 0}}},{26816, {{37518, 1, 0, 0},{37518, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26817, {{37519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26818, {{37520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26819, {{37521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26820, {{37522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26821, {{37523, 1, 0, 0},{37523, 1, 1, 17}, {0, 0, 0, 0}}},{26821, {{37523, 1, 0, 0},{37523, 1, 1, 17}, {0, 0, 0, 0}}},{26822, {{37523, 1, 0, 0},{37523, 1, 1, 18}, {0, 0, 0, 0}}},{26822, {{37523, 1, 0, 0},{37523, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26823, {{37524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26824, {{37525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26825, {{37526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26826, {{37527, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26827, {{37528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26828, {{37529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26829, {{37530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26830, {{37531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26831, {{37532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26832, {{37533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26833, {{37534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26834, {{37535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26835, {{37536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26836, {{37537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{26837, {{37538, 1, 0, 0},{37538, 1, 1, 16}, {0, 0, 0, 0}}},{26837, {{37538, 1, 0, 0},{37538, 1, 1, 16}, {0, 0, 0, 0}}},{26838, {{37538, 1, 0, 0},{37538, 1, 1, 17}, {0, 0, 0, 0}}},{26838, {{37538, 1, 0, 0},{37538, 1, 1, 17}, {0, 0, 0, 0}}},{68069, {{37538, 1, 0, 0},{37538, 1, 1, 18}, {0, 0, 0, 0}}},{68069, {{37538, 1, 0, 0},{37538, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26839, {{37539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26840, {{37540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26841, {{37541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26842, {{37542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26843, {{37543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26844, {{37544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26845, {{37545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26846, {{37546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26847, {{37547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26848, {{37548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26849, {{37549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26850, {{37550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26851, {{37551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{26852, {{37554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26853, {{37555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26854, {{37556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26855, {{37557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26856, {{37558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26857, {{37559, 1, 0, 0},{37559, 1, 1, 17}, {0, 0, 0, 0}}},{26857, {{37559, 1, 0, 0},{37559, 1, 1, 17}, {0, 0, 0, 0}}},{26858, {{37559, 1, 0, 0},{37559, 1, 1, 18}, {0, 0, 0, 0}}},{26858, {{37559, 1, 0, 0},{37559, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26859, {{37560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26860, {{37561, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26861, {{37562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26862, {{37563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26863, {{37564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26864, {{37565, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26865, {{37566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26866, {{37567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26867, {{37568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26868, {{37569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26869, {{37570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26870, {{37571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26871, {{37572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26872, {{37573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26873, {{37574, 1, 0, 0},{37574, 1, 1, 17}, {0, 0, 0, 0}}},{26873, {{37574, 1, 0, 0},{37574, 1, 1, 17}, {0, 0, 0, 0}}},{26874, {{37574, 1, 0, 0},{37574, 1, 1, 18}, {0, 0, 0, 0}}},{26874, {{37574, 1, 0, 0},{37574, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26875, {{37575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26876, {{37576, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26877, {{37577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26878, {{37578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26879, {{37579, 1, 0, 0},{37579, 1, 1, 17}, {0, 0, 0, 0}}},{26879, {{37579, 1, 0, 0},{37579, 1, 1, 17}, {0, 0, 0, 0}}},{26880, {{37579, 1, 0, 0},{37579, 1, 1, 18}, {0, 0, 0, 0}}},{26880, {{37579, 1, 0, 0},{37579, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{26881, {{37580, 1, 0, 0},{37580, 1, 1, 17}, {0, 0, 0, 0}}},{26881, {{37580, 1, 0, 0},{37580, 1, 1, 17}, {0, 0, 0, 0}}},{26882, {{37580, 1, 0, 0},{37580, 1, 1, 19}, {0, 0, 0, 0}}},{26882, {{37580, 1, 0, 0},{37580, 1, 1, 19}, {0, 0, 0, 0}}},{26883, {{37580, 1, 0, 0},{37580, 1, 1, 18}, {0, 0, 0, 0}}},{26883, {{37580, 1, 0, 0},{37580, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26884, {{37581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26885, {{37582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26886, {{37583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26887, {{37584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26888, {{37585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26889, {{37586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26890, {{37587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26891, {{37588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26892, {{37589, 1, 0, 0},{37589, 1, 1, 17}, {0, 0, 0, 0}}},{26892, {{37589, 1, 0, 0},{37589, 1, 1, 17}, {0, 0, 0, 0}}},{26893, {{37589, 1, 0, 0},{37589, 1, 1, 18}, {0, 0, 0, 0}}},{26893, {{37589, 1, 0, 0},{37589, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26894, {{37590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26895, {{37591, 1, 0, 0},{37591, 1, 1, 18}, {0, 0, 0, 0}}},{26895, {{37591, 1, 0, 0},{37591, 1, 1, 18}, {0, 0, 0, 0}}},{26896, {{37591, 1, 0, 0},{37591, 1, 1, 17}, {0, 0, 0, 0}}},{26896, {{37591, 1, 0, 0},{37591, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{26897, {{37592, 1, 0, 0},{37592, 1, 1, 18}, {0, 0, 0, 0}}},{26897, {{37592, 1, 0, 0},{37592, 1, 1, 18}, {0, 0, 0, 0}}},{26898, {{37592, 1, 0, 0},{37592, 1, 1, 17}, {0, 0, 0, 0}}},{26898, {{37592, 1, 0, 0},{37592, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{26899, {{37593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26900, {{37594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26901, {{37595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26902, {{37596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26903, {{37597, 1, 0, 0},{37597, 1, 1, 17}, {0, 0, 0, 0}}},{26903, {{37597, 1, 0, 0},{37597, 1, 1, 17}, {0, 0, 0, 0}}},{26904, {{37597, 1, 0, 0},{37597, 1, 1, 18}, {0, 0, 0, 0}}},{26904, {{37597, 1, 0, 0},{37597, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26905, {{37598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26906, {{37599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26907, {{37600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26908, {{37601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{26909, {{37603, 1, 0, 0},{37603, 1, 1, 18}, {0, 0, 0, 0}}},{26909, {{37603, 1, 0, 0},{37603, 1, 1, 18}, {0, 0, 0, 0}}},{26910, {{37603, 1, 0, 0},{37603, 1, 1, 17}, {0, 0, 0, 0}}},{26910, {{37603, 1, 0, 0},{37603, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{26911, {{37604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26912, {{37605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26913, {{37606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26914, {{37607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26915, {{37608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{26916, {{37609, 1, 0, 0},{37609, 1, 1, 18}, {0, 0, 0, 0}}},{26916, {{37609, 1, 0, 0},{37609, 1, 1, 18}, {0, 0, 0, 0}}},{26917, {{37609, 1, 0, 0},{37609, 1, 1, 20}, {0, 0, 0, 0}}},{26917, {{37609, 1, 0, 0},{37609, 1, 1, 20}, {0, 0, 0, 0}}},{26918, {{37609, 1, 0, 0},{37609, 1, 1, 19}, {0, 0, 0, 0}}},{26918, {{37609, 1, 0, 0},{37609, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{26919, {{37610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26920, {{37611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26921, {{37612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26922, {{37613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26923, {{37614, 1, 0, 0},{37614, 1, 1, 17}, {0, 0, 0, 0}}},{26923, {{37614, 1, 0, 0},{37614, 1, 1, 17}, {0, 0, 0, 0}}},{26924, {{37614, 1, 0, 0},{37614, 1, 1, 18}, {0, 0, 0, 0}}},{26924, {{37614, 1, 0, 0},{37614, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26925, {{37615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26927, {{37616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26928, {{37617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26929, {{37618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26930, {{37619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26931, {{37620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{26932, {{37623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26933, {{37624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26934, {{37625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26935, {{37626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26936, {{37627, 1, 0, 0},{37627, 1, 1, 17}, {0, 0, 0, 0}}},{26936, {{37627, 1, 0, 0},{37627, 1, 1, 17}, {0, 0, 0, 0}}},{26937, {{37627, 1, 0, 0},{37627, 1, 1, 18}, {0, 0, 0, 0}}},{26937, {{37627, 1, 0, 0},{37627, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26938, {{37628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26939, {{37629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26940, {{37630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26941, {{37631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26942, {{37632, 1, 0, 0},{37632, 1, 1, 17}, {0, 0, 0, 0}}},{26942, {{37632, 1, 0, 0},{37632, 1, 1, 17}, {0, 0, 0, 0}}},{26943, {{37632, 1, 0, 0},{37632, 1, 1, 18}, {0, 0, 0, 0}}},{26943, {{37632, 1, 0, 0},{37632, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26944, {{37633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26945, {{37634, 1, 0, 0},{37634, 1, 1, 18}, {0, 0, 0, 0}}},{26945, {{37634, 1, 0, 0},{37634, 1, 1, 18}, {0, 0, 0, 0}}},{26946, {{37634, 1, 0, 0},{37634, 1, 1, 17}, {0, 0, 0, 0}}},{26946, {{37634, 1, 0, 0},{37634, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{26947, {{37635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26948, {{37636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26949, {{37637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26950, {{37638, 1, 0, 0},{37638, 1, 1, 19}, {0, 0, 0, 0}}},{26950, {{37638, 1, 0, 0},{37638, 1, 1, 19}, {0, 0, 0, 0}}},{26951, {{37638, 1, 0, 0},{37638, 1, 1, 18}, {0, 0, 0, 0}}},{26951, {{37638, 1, 0, 0},{37638, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26952, {{37639, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26953, {{37640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26955, {{37641, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26956, {{37642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26957, {{37643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26959, {{37644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26960, {{37645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26961, {{37646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26962, {{37647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26963, {{37648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26964, {{37649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26965, {{37650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26966, {{37651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26967, {{37652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26968, {{37653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26969, {{37654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26970, {{37655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26971, {{37656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26972, {{37657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{26973, {{37658, 1, 0, 0},{37658, 1, 1, 17}, {0, 0, 0, 0}}},{26973, {{37658, 1, 0, 0},{37658, 1, 1, 17}, {0, 0, 0, 0}}},{26974, {{37658, 1, 0, 0},{37658, 1, 1, 18}, {0, 0, 0, 0}}},{26974, {{37658, 1, 0, 0},{37658, 1, 1, 18}, {0, 0, 0, 0}}},{26975, {{37658, 1, 0, 0},{37658, 1, 1, 19}, {0, 0, 0, 0}}},{26975, {{37658, 1, 0, 0},{37658, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{26976, {{37659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26977, {{37660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26978, {{37661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26979, {{37662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26980, {{37663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26981, {{37664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26982, {{37665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26983, {{37666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26984, {{37667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26985, {{37668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26986, {{37669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26987, {{37670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26988, {{37671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{26989, {{37672, 1, 0, 0},{37672, 1, 1, 17}, {0, 0, 0, 0}}},{26989, {{37672, 1, 0, 0},{37672, 1, 1, 17}, {0, 0, 0, 0}}},{26990, {{37672, 1, 0, 0},{37672, 1, 1, 18}, {0, 0, 0, 0}}},{26990, {{37672, 1, 0, 0},{37672, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{26991, {{37673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26992, {{37674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26993, {{37675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26994, {{37676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26995, {{37677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26996, {{37678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26997, {{37679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{26998, {{37681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{26999, {{37682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27000, {{37683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27001, {{37684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27002, {{37685, 1, 0, 0},{37685, 1, 1, 18}, {0, 0, 0, 0}}},{27002, {{37685, 1, 0, 0},{37685, 1, 1, 18}, {0, 0, 0, 0}}},{27003, {{37685, 1, 0, 0},{37685, 1, 1, 19}, {0, 0, 0, 0}}},{27003, {{37685, 1, 0, 0},{37685, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{27004, {{37686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27005, {{37687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27006, {{37688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27007, {{37689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27008, {{37690, 1, 0, 0},{37690, 1, 1, 17}, {0, 0, 0, 0}}},{27008, {{37690, 1, 0, 0},{37690, 1, 1, 17}, {0, 0, 0, 0}}},{27009, {{37690, 1, 0, 0},{37690, 1, 1, 18}, {0, 0, 0, 0}}},{27009, {{37690, 1, 0, 0},{37690, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27010, {{37691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{27014, {{37700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{27015, {{37703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27016, {{37704, 1, 0, 0},{37704, 1, 1, 17}, {0, 0, 0, 0}}},{27016, {{37704, 1, 0, 0},{37704, 1, 1, 17}, {0, 0, 0, 0}}},{27017, {{37704, 1, 0, 0},{37704, 1, 1, 18}, {0, 0, 0, 0}}},{27017, {{37704, 1, 0, 0},{37704, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27018, {{37705, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27019, {{37706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27020, {{37707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27021, {{37708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27022, {{37709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27023, {{37710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27024, {{37711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27026, {{37712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27027, {{37713, 1, 0, 0},{37713, 1, 1, 18}, {0, 0, 0, 0}}},{27027, {{37713, 1, 0, 0},{37713, 1, 1, 18}, {0, 0, 0, 0}}},{27028, {{37713, 1, 0, 0},{37713, 1, 1, 17}, {0, 0, 0, 0}}},{27028, {{37713, 1, 0, 0},{37713, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{27029, {{37714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27030, {{37715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27031, {{37716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27033, {{37717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27034, {{37718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27035, {{37719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27036, {{37720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27037, {{37721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27038, {{37722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27039, {{37723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27040, {{37724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27041, {{37725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27042, {{37726, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27043, {{37727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27044, {{37728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27045, {{37729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27046, {{37730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27047, {{37731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27048, {{37732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{27049, {{37733, 1, 0, 0},{37733, 1, 1, 18}, {0, 0, 0, 0}}},{27049, {{37733, 1, 0, 0},{37733, 1, 1, 18}, {0, 0, 0, 0}}},{27050, {{37733, 1, 0, 0},{37733, 1, 1, 19}, {0, 0, 0, 0}}},{27050, {{37733, 1, 0, 0},{37733, 1, 1, 19}, {0, 0, 0, 0}}},{27051, {{37733, 1, 0, 0},{37733, 1, 1, 20}, {0, 0, 0, 0}}},{27051, {{37733, 1, 0, 0},{37733, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{27052, {{37734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27053, {{37735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27054, {{37736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27055, {{37737, 1, 0, 0},{37737, 1, 1, 17}, {0, 0, 0, 0}}},{27055, {{37737, 1, 0, 0},{37737, 1, 1, 17}, {0, 0, 0, 0}}},{27056, {{37737, 1, 0, 0},{37737, 1, 1, 18}, {0, 0, 0, 0}}},{27056, {{37737, 1, 0, 0},{37737, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27057, {{37738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27058, {{37739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27059, {{37740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27060, {{37741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27061, {{37742, 1, 0, 0},{37742, 1, 1, 17}, {0, 0, 0, 0}}},{27061, {{37742, 1, 0, 0},{37742, 1, 1, 17}, {0, 0, 0, 0}}},{27062, {{37742, 1, 0, 0},{37742, 1, 1, 18}, {0, 0, 0, 0}}},{27062, {{37742, 1, 0, 0},{37742, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27063, {{37743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27064, {{37744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27065, {{37745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27066, {{37746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27067, {{37747, 1, 0, 0},{37747, 1, 1, 17}, {0, 0, 0, 0}}},{27067, {{37747, 1, 0, 0},{37747, 1, 1, 17}, {0, 0, 0, 0}}},{27068, {{37747, 1, 0, 0},{37747, 1, 1, 18}, {0, 0, 0, 0}}},{27068, {{37747, 1, 0, 0},{37747, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27069, {{37748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27070, {{37749, 1, 0, 0},{37749, 1, 1, 18}, {0, 0, 0, 0}}},{27070, {{37749, 1, 0, 0},{37749, 1, 1, 18}, {0, 0, 0, 0}}},{27071, {{37749, 1, 0, 0},{37749, 1, 1, 19}, {0, 0, 0, 0}}},{27071, {{37749, 1, 0, 0},{37749, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{27072, {{37750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27073, {{37751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27074, {{37752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27075, {{37753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27076, {{37754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27077, {{37755, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27078, {{37756, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27079, {{37757, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27080, {{37758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27081, {{37759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27082, {{37760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27083, {{37761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27084, {{37762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27085, {{37763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27086, {{37764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{27087, {{37768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27088, {{37769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27089, {{37770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27090, {{37771, 1, 0, 0},{37771, 1, 1, 18}, {0, 0, 0, 0}}},{27090, {{37771, 1, 0, 0},{37771, 1, 1, 18}, {0, 0, 0, 0}}},{27091, {{37771, 1, 0, 0},{37771, 1, 1, 19}, {0, 0, 0, 0}}},{27091, {{37771, 1, 0, 0},{37771, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{27092, {{37772, 1, 0, 0},{37772, 1, 1, 19}, {0, 0, 0, 0}}},{27092, {{37772, 1, 0, 0},{37772, 1, 1, 19}, {0, 0, 0, 0}}},{27093, {{37772, 1, 0, 0},{37772, 1, 1, 18}, {0, 0, 0, 0}}},{27093, {{37772, 1, 0, 0},{37772, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27094, {{37773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27095, {{37774, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27096, {{37775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27097, {{37776, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27098, {{37777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27099, {{37778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27100, {{37779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27101, {{37780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27102, {{37781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27103, {{37782, 1, 0, 0},{37782, 1, 1, 19}, {0, 0, 0, 0}}},{27103, {{37782, 1, 0, 0},{37782, 1, 1, 19}, {0, 0, 0, 0}}},{27104, {{37782, 1, 0, 0},{37782, 1, 1, 18}, {0, 0, 0, 0}}},{27104, {{37782, 1, 0, 0},{37782, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27105, {{37783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27106, {{37784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27107, {{37785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27108, {{37786, 1, 0, 0},{37786, 1, 1, 18}, {0, 0, 0, 0}}},{27108, {{37786, 1, 0, 0},{37786, 1, 1, 18}, {0, 0, 0, 0}}},{27109, {{37786, 1, 0, 0},{37786, 1, 1, 19}, {0, 0, 0, 0}}},{27109, {{37786, 1, 0, 0},{37786, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{27110, {{37787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27111, {{37788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27112, {{37789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27113, {{37790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27114, {{37791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27115, {{37792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{27116, {{37793, 1, 0, 0},{37793, 1, 1, 18}, {0, 0, 0, 0}}},{27116, {{37793, 1, 0, 0},{37793, 1, 1, 18}, {0, 0, 0, 0}}},{27117, {{37793, 1, 0, 0},{37793, 1, 1, 20}, {0, 0, 0, 0}}},{27117, {{37793, 1, 0, 0},{37793, 1, 1, 20}, {0, 0, 0, 0}}},{27118, {{37793, 1, 0, 0},{37793, 1, 1, 19}, {0, 0, 0, 0}}},{27118, {{37793, 1, 0, 0},{37793, 1, 1, 19}, {0, 0, 0, 0}}},{58932, {{37793, 1, 0, 0},{37793, 1, 1, 21}, {0, 0, 0, 0}}},{58932, {{37793, 1, 0, 0},{37793, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{27119, {{37794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27120, {{37795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27121, {{37796, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27122, {{37797, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27123, {{37798, 1, 0, 0},{37798, 1, 1, 17}, {0, 0, 0, 0}}},{27123, {{37798, 1, 0, 0},{37798, 1, 1, 17}, {0, 0, 0, 0}}},{27124, {{37798, 1, 0, 0},{37798, 1, 1, 18}, {0, 0, 0, 0}}},{27124, {{37798, 1, 0, 0},{37798, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27125, {{37799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27126, {{37800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27127, {{37801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27128, {{37802, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27129, {{37803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27130, {{37804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27132, {{37805, 1, 0, 0},{37805, 1, 1, 17}, {0, 0, 0, 0}}},{27132, {{37805, 1, 0, 0},{37805, 1, 1, 17}, {0, 0, 0, 0}}},{60292, {{37805, 1, 0, 0},{37805, 1, 1, 18}, {0, 0, 0, 0}}},{60292, {{37805, 1, 0, 0},{37805, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27133, {{37806, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27134, {{37807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{27136, {{37808, 1, 0, 0},{37808, 1, 1, 17}, {0, 0, 0, 0}}},{27136, {{37808, 1, 0, 0},{37808, 1, 1, 17}, {0, 0, 0, 0}}},{27137, {{37808, 1, 0, 0},{37808, 1, 1, 18}, {0, 0, 0, 0}}},{27137, {{37808, 1, 0, 0},{37808, 1, 1, 18}, {0, 0, 0, 0}}},{27138, {{37808, 1, 0, 0},{37808, 1, 1, 19}, {0, 0, 0, 0}}},{27138, {{37808, 1, 0, 0},{37808, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{27139, {{37809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27140, {{37810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27141, {{37811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27142, {{37812, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27143, {{37813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27144, {{37814, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27145, {{37815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27146, {{37816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {10, {{27147, {{37817, 1, 0, 0},{37817, 1, 1, 21}, {0, 0, 0, 0}}},{27147, {{37817, 1, 0, 0},{37817, 1, 1, 21}, {0, 0, 0, 0}}},{27148, {{37817, 1, 0, 0},{37817, 1, 1, 17}, {0, 0, 0, 0}}},{27148, {{37817, 1, 0, 0},{37817, 1, 1, 17}, {0, 0, 0, 0}}},{27149, {{37817, 1, 0, 0},{37817, 1, 1, 18}, {0, 0, 0, 0}}},{27149, {{37817, 1, 0, 0},{37817, 1, 1, 18}, {0, 0, 0, 0}}},{27150, {{37817, 1, 0, 0},{37817, 1, 1, 19}, {0, 0, 0, 0}}},{27150, {{37817, 1, 0, 0},{37817, 1, 1, 19}, {0, 0, 0, 0}}},{27151, {{37817, 1, 0, 0},{37817, 1, 1, 20}, {0, 0, 0, 0}}},{27151, {{37817, 1, 0, 0},{37817, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{27152, {{37818, 1, 0, 0},{37818, 1, 1, 18}, {0, 0, 0, 0}}},{27152, {{37818, 1, 0, 0},{37818, 1, 1, 18}, {0, 0, 0, 0}}},{27153, {{37818, 1, 0, 0},{37818, 1, 1, 17}, {0, 0, 0, 0}}},{27153, {{37818, 1, 0, 0},{37818, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{27154, {{37819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{27156, {{37824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27157, {{37825, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27158, {{37826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27159, {{37827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27160, {{37828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27161, {{37829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27163, {{37830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27164, {{37831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27165, {{37832, 1, 0, 0},{37832, 1, 1, 18}, {0, 0, 0, 0}}},{27165, {{37832, 1, 0, 0},{37832, 1, 1, 18}, {0, 0, 0, 0}}},{27166, {{37832, 1, 0, 0},{37832, 1, 1, 17}, {0, 0, 0, 0}}},{27166, {{37832, 1, 0, 0},{37832, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{27167, {{37833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27168, {{37834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27169, {{37835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27170, {{37836, 1, 0, 0},{37836, 1, 1, 17}, {0, 0, 0, 0}}},{27170, {{37836, 1, 0, 0},{37836, 1, 1, 17}, {0, 0, 0, 0}}},{27171, {{37836, 1, 0, 0},{37836, 1, 1, 18}, {0, 0, 0, 0}}},{27171, {{37836, 1, 0, 0},{37836, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27172, {{37837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27173, {{37838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27174, {{37839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27175, {{37840, 1, 0, 0},{37840, 1, 1, 17}, {0, 0, 0, 0}}},{27175, {{37840, 1, 0, 0},{37840, 1, 1, 17}, {0, 0, 0, 0}}},{27176, {{37840, 1, 0, 0},{37840, 1, 1, 18}, {0, 0, 0, 0}}},{27176, {{37840, 1, 0, 0},{37840, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{27177, {{37841, 1, 0, 0},{37841, 1, 1, 17}, {0, 0, 0, 0}}},{27177, {{37841, 1, 0, 0},{37841, 1, 1, 17}, {0, 0, 0, 0}}},{27178, {{37841, 1, 0, 0},{37841, 1, 1, 18}, {0, 0, 0, 0}}},{27178, {{37841, 1, 0, 0},{37841, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27179, {{37842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27180, {{37843, 1, 0, 0},{37843, 1, 1, 17}, {0, 0, 0, 0}}},{27180, {{37843, 1, 0, 0},{37843, 1, 1, 17}, {0, 0, 0, 0}}},{27181, {{37843, 1, 0, 0},{37843, 1, 1, 18}, {0, 0, 0, 0}}},{27181, {{37843, 1, 0, 0},{37843, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27183, {{37844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27184, {{37845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27185, {{37846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27186, {{37847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27187, {{37848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27188, {{37849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27189, {{37850, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27190, {{37851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27191, {{37852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27192, {{37853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27193, {{37854, 1, 0, 0},{37854, 1, 1, 17}, {0, 0, 0, 0}}},{27193, {{37854, 1, 0, 0},{37854, 1, 1, 17}, {0, 0, 0, 0}}},{27194, {{37854, 1, 0, 0},{37854, 1, 1, 18}, {0, 0, 0, 0}}},{27194, {{37854, 1, 0, 0},{37854, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27195, {{37855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27196, {{37856, 1, 0, 0},{37856, 1, 1, 16}, {0, 0, 0, 0}}},{27196, {{37856, 1, 0, 0},{37856, 1, 1, 16}, {0, 0, 0, 0}}},{27197, {{37856, 1, 0, 0},{37856, 1, 1, 17}, {0, 0, 0, 0}}},{27197, {{37856, 1, 0, 0},{37856, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{27198, {{37857, 1, 0, 0},{37857, 1, 1, 17}, {0, 0, 0, 0}}},{27198, {{37857, 1, 0, 0},{37857, 1, 1, 17}, {0, 0, 0, 0}}},{27199, {{37857, 1, 0, 0},{37857, 1, 1, 18}, {0, 0, 0, 0}}},{27199, {{37857, 1, 0, 0},{37857, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27200, {{37858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27201, {{37859, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27202, {{37860, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27203, {{37861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27204, {{37862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27205, {{37863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27206, {{37864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27207, {{37865, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27208, {{37866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27209, {{37867, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27210, {{37868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27211, {{37869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27212, {{37870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{27213, {{37873, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{27214, {{37875, 1, 0, 0},{37875, 1, 1, 16}, {0, 0, 0, 0}}},{27214, {{37875, 1, 0, 0},{37875, 1, 1, 16}, {0, 0, 0, 0}}},{27215, {{37875, 1, 0, 0},{37875, 1, 1, 17}, {0, 0, 0, 0}}},{27215, {{37875, 1, 0, 0},{37875, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{27217, {{37876, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27218, {{37877, 1, 0, 0},{37877, 1, 1, 17}, {0, 0, 0, 0}}},{27218, {{37877, 1, 0, 0},{37877, 1, 1, 17}, {0, 0, 0, 0}}},{27219, {{37877, 1, 0, 0},{37877, 1, 1, 18}, {0, 0, 0, 0}}},{27219, {{37877, 1, 0, 0},{37877, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27220, {{37878, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27221, {{37879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27222, {{37880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27223, {{37881, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27225, {{37882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27226, {{37883, 1, 0, 0},{37883, 1, 1, 18}, {0, 0, 0, 0}}},{27226, {{37883, 1, 0, 0},{37883, 1, 1, 18}, {0, 0, 0, 0}}},{27227, {{37883, 1, 0, 0},{37883, 1, 1, 17}, {0, 0, 0, 0}}},{27227, {{37883, 1, 0, 0},{37883, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{27228, {{37884, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27229, {{37885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27230, {{37886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27231, {{37887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27232, {{37888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27233, {{37889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27234, {{37890, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27235, {{37891, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27236, {{37892, 1, 0, 0},{37892, 1, 1, 17}, {0, 0, 0, 0}}},{27236, {{37892, 1, 0, 0},{37892, 1, 1, 17}, {0, 0, 0, 0}}},{27237, {{37892, 1, 0, 0},{37892, 1, 1, 18}, {0, 0, 0, 0}}},{27237, {{37892, 1, 0, 0},{37892, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27238, {{37893, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27239, {{37894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27240, {{37895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27241, {{37896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27242, {{37897, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27243, {{37898, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27244, {{37899, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27245, {{37900, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27246, {{37901, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27247, {{37902, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27248, {{37903, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27249, {{37904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27250, {{37905, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27251, {{37906, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27252, {{37907, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27253, {{37908, 1, 0, 0},{37908, 1, 1, 18}, {0, 0, 0, 0}}},{27253, {{37908, 1, 0, 0},{37908, 1, 1, 18}, {0, 0, 0, 0}}},{27254, {{37908, 1, 0, 0},{37908, 1, 1, 17}, {0, 0, 0, 0}}},{27254, {{37908, 1, 0, 0},{37908, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{27255, {{37909, 1, 0, 0},{37909, 1, 1, 18}, {0, 0, 0, 0}}},{27255, {{37909, 1, 0, 0},{37909, 1, 1, 18}, {0, 0, 0, 0}}},{27256, {{37909, 1, 0, 0},{37909, 1, 1, 17}, {0, 0, 0, 0}}},{27256, {{37909, 1, 0, 0},{37909, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{27258, {{37910, 1, 0, 0},{37910, 1, 1, 17}, {0, 0, 0, 0}}},{27258, {{37910, 1, 0, 0},{37910, 1, 1, 17}, {0, 0, 0, 0}}},{27259, {{37910, 1, 0, 0},{37910, 1, 1, 18}, {0, 0, 0, 0}}},{27259, {{37910, 1, 0, 0},{37910, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27261, {{37911, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27262, {{37912, 1, 0, 0},{37912, 1, 1, 17}, {0, 0, 0, 0}}},{27262, {{37912, 1, 0, 0},{37912, 1, 1, 17}, {0, 0, 0, 0}}},{27263, {{37912, 1, 0, 0},{37912, 1, 1, 18}, {0, 0, 0, 0}}},{27263, {{37912, 1, 0, 0},{37912, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27264, {{37913, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27265, {{37914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{27267, {{37916, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27268, {{37917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27269, {{37918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27270, {{37919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27271, {{37920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27272, {{37921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{68046, {{37925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{27274, {{37929, 1, 0, 0},{37929, 1, 1, 16}, {0, 0, 0, 0}}},{27274, {{37929, 1, 0, 0},{37929, 1, 1, 16}, {0, 0, 0, 0}}},{27275, {{37929, 1, 0, 0},{37929, 1, 1, 17}, {0, 0, 0, 0}}},{27275, {{37929, 1, 0, 0},{37929, 1, 1, 17}, {0, 0, 0, 0}}},{27276, {{37929, 1, 0, 0},{37929, 1, 1, 18}, {0, 0, 0, 0}}},{27276, {{37929, 1, 0, 0},{37929, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{27277, {{37930, 1, 0, 0},{37930, 1, 1, 16}, {0, 0, 0, 0}}},{27277, {{37930, 1, 0, 0},{37930, 1, 1, 16}, {0, 0, 0, 0}}},{27278, {{37930, 1, 0, 0},{37930, 1, 1, 17}, {0, 0, 0, 0}}},{27278, {{37930, 1, 0, 0},{37930, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{27279, {{37931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27280, {{37932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27281, {{37933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27282, {{37934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27283, {{37935, 1, 0, 0},{37935, 1, 1, 17}, {0, 0, 0, 0}}},{27283, {{37935, 1, 0, 0},{37935, 1, 1, 17}, {0, 0, 0, 0}}},{27284, {{37935, 1, 0, 0},{37935, 1, 1, 18}, {0, 0, 0, 0}}},{27284, {{37935, 1, 0, 0},{37935, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27285, {{37936, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27286, {{37937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27287, {{37938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27288, {{37939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27289, {{37940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27290, {{37941, 1, 0, 0},{37941, 1, 1, 17}, {0, 0, 0, 0}}},{27290, {{37941, 1, 0, 0},{37941, 1, 1, 17}, {0, 0, 0, 0}}},{27291, {{37941, 1, 0, 0},{37941, 1, 1, 18}, {0, 0, 0, 0}}},{27291, {{37941, 1, 0, 0},{37941, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27292, {{37942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27293, {{37943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27294, {{37944, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27295, {{37945, 1, 0, 0},{37945, 1, 1, 16}, {0, 0, 0, 0}}},{27295, {{37945, 1, 0, 0},{37945, 1, 1, 16}, {0, 0, 0, 0}}},{27296, {{37945, 1, 0, 0},{37945, 1, 1, 17}, {0, 0, 0, 0}}},{27296, {{37945, 1, 0, 0},{37945, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{27297, {{37946, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27298, {{37947, 1, 0, 0},{37947, 1, 1, 17}, {0, 0, 0, 0}}},{27298, {{37947, 1, 0, 0},{37947, 1, 1, 17}, {0, 0, 0, 0}}},{27299, {{37947, 1, 0, 0},{37947, 1, 1, 18}, {0, 0, 0, 0}}},{27299, {{37947, 1, 0, 0},{37947, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27300, {{37948, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27301, {{37949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27302, {{37950, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27303, {{37951, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{27305, {{37953, 1, 0, 0},{37953, 1, 1, 17}, {0, 0, 0, 0}}},{27305, {{37953, 1, 0, 0},{37953, 1, 1, 17}, {0, 0, 0, 0}}},{58945, {{37953, 1, 0, 0},{37953, 1, 1, 18}, {0, 0, 0, 0}}},{58945, {{37953, 1, 0, 0},{37953, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27306, {{37954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27307, {{37955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27308, {{37956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27309, {{37957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27310, {{37958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27311, {{37959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27312, {{37960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27313, {{37961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27314, {{37962, 1, 0, 0},{37962, 1, 1, 17}, {0, 0, 0, 0}}},{27314, {{37962, 1, 0, 0},{37962, 1, 1, 17}, {0, 0, 0, 0}}},{27315, {{37962, 1, 0, 0},{37962, 1, 1, 18}, {0, 0, 0, 0}}},{27315, {{37962, 1, 0, 0},{37962, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27316, {{37963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27317, {{37964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27319, {{37965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27320, {{37966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27321, {{37967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27322, {{37968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27323, {{37969, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27324, {{37970, 1, 0, 0},{37970, 1, 1, 17}, {0, 0, 0, 0}}},{27324, {{37970, 1, 0, 0},{37970, 1, 1, 17}, {0, 0, 0, 0}}},{27325, {{37970, 1, 0, 0},{37970, 1, 1, 18}, {0, 0, 0, 0}}},{27325, {{37970, 1, 0, 0},{37970, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{27326, {{37971, 1, 0, 0},{37971, 1, 1, 18}, {0, 0, 0, 0}}},{27326, {{37971, 1, 0, 0},{37971, 1, 1, 18}, {0, 0, 0, 0}}},{27327, {{37971, 1, 0, 0},{37971, 1, 1, 19}, {0, 0, 0, 0}}},{27327, {{37971, 1, 0, 0},{37971, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{27328, {{37973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27329, {{37974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27330, {{37975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27331, {{37976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27332, {{37977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27333, {{37978, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27334, {{37979, 1, 0, 0},{37979, 1, 1, 17}, {0, 0, 0, 0}}},{27334, {{37979, 1, 0, 0},{37979, 1, 1, 17}, {0, 0, 0, 0}}},{27335, {{37979, 1, 0, 0},{37979, 1, 1, 18}, {0, 0, 0, 0}}},{27335, {{37979, 1, 0, 0},{37979, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27339, {{37980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27340, {{37981, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27341, {{37982, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27342, {{37983, 1, 0, 0},{37983, 1, 1, 17}, {0, 0, 0, 0}}},{27342, {{37983, 1, 0, 0},{37983, 1, 1, 17}, {0, 0, 0, 0}}},{27343, {{37983, 1, 0, 0},{37983, 1, 1, 18}, {0, 0, 0, 0}}},{27343, {{37983, 1, 0, 0},{37983, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27344, {{37984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27345, {{37985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27346, {{37986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27347, {{37987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27348, {{37988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27349, {{37989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{27352, {{37992, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27353, {{37993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27354, {{37994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27355, {{37995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27356, {{37996, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27357, {{37997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27358, {{37998, 1, 0, 0},{37998, 1, 1, 17}, {0, 0, 0, 0}}},{27358, {{37998, 1, 0, 0},{37998, 1, 1, 17}, {0, 0, 0, 0}}},{27359, {{37998, 1, 0, 0},{37998, 1, 1, 18}, {0, 0, 0, 0}}},{27359, {{37998, 1, 0, 0},{37998, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27361, {{37999, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27362, {{38000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27363, {{38001, 1, 0, 0},{38001, 1, 1, 17}, {0, 0, 0, 0}}},{27363, {{38001, 1, 0, 0},{38001, 1, 1, 17}, {0, 0, 0, 0}}},{27364, {{38001, 1, 0, 0},{38001, 1, 1, 18}, {0, 0, 0, 0}}},{27364, {{38001, 1, 0, 0},{38001, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27365, {{38002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27366, {{38003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27367, {{38004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27369, {{38005, 1, 0, 0},{38005, 1, 1, 17}, {0, 0, 0, 0}}},{27369, {{38005, 1, 0, 0},{38005, 1, 1, 17}, {0, 0, 0, 0}}},{27370, {{38005, 1, 0, 0},{38005, 1, 1, 18}, {0, 0, 0, 0}}},{27370, {{38005, 1, 0, 0},{38005, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{27371, {{38006, 1, 0, 0},{38006, 1, 1, 18}, {0, 0, 0, 0}}},{27371, {{38006, 1, 0, 0},{38006, 1, 1, 18}, {0, 0, 0, 0}}},{27372, {{38006, 1, 0, 0},{38006, 1, 1, 16}, {0, 0, 0, 0}}},{27372, {{38006, 1, 0, 0},{38006, 1, 1, 16}, {0, 0, 0, 0}}},{27373, {{38006, 1, 0, 0},{38006, 1, 1, 17}, {0, 0, 0, 0}}},{27373, {{38006, 1, 0, 0},{38006, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {6, {{27374, {{38007, 1, 0, 0},{38007, 1, 1, 18}, {0, 0, 0, 0}}},{27374, {{38007, 1, 0, 0},{38007, 1, 1, 18}, {0, 0, 0, 0}}},{27375, {{38007, 1, 0, 0},{38007, 1, 1, 19}, {0, 0, 0, 0}}},{27375, {{38007, 1, 0, 0},{38007, 1, 1, 19}, {0, 0, 0, 0}}},{27376, {{38007, 1, 0, 0},{38007, 1, 1, 20}, {0, 0, 0, 0}}},{27376, {{38007, 1, 0, 0},{38007, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{27377, {{38008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27378, {{38009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27379, {{38010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27380, {{38011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27381, {{38012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27382, {{38013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27383, {{38014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27384, {{38015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27385, {{38016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27386, {{38017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27387, {{38018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27388, {{38019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27389, {{38020, 1, 0, 0},{38020, 1, 1, 17}, {0, 0, 0, 0}}},{27389, {{38020, 1, 0, 0},{38020, 1, 1, 17}, {0, 0, 0, 0}}},{27390, {{38020, 1, 0, 0},{38020, 1, 1, 18}, {0, 0, 0, 0}}},{27390, {{38020, 1, 0, 0},{38020, 1, 1, 18}, {0, 0, 0, 0}}}}}
};

static const MJMappingSet chunked_mj_mappings_009577_009c7b[1797] = {
    {1, {{27400, {{38263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27401, {{38264, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27402, {{38265, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27403, {{38266, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27404, {{38267, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27405, {{38268, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27406, {{38269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27407, {{38270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{27409, {{38272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27410, {{38273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27411, {{38274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27412, {{38275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27413, {{38276, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27414, {{38277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27415, {{38278, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27416, {{38279, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27417, {{38280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27418, {{38281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27419, {{38282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27420, {{38283, 1, 0, 0},{38283, 1, 1, 17}, {0, 0, 0, 0}}},{27420, {{38283, 1, 0, 0},{38283, 1, 1, 17}, {0, 0, 0, 0}}},{27421, {{38283, 1, 0, 0},{38283, 1, 1, 18}, {0, 0, 0, 0}}},{27421, {{38283, 1, 0, 0},{38283, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27423, {{38284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27424, {{38285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27425, {{38286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27426, {{38287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27428, {{38288, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27429, {{38289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27430, {{38290, 1, 0, 0},{38290, 1, 1, 18}, {0, 0, 0, 0}}},{27430, {{38290, 1, 0, 0},{38290, 1, 1, 18}, {0, 0, 0, 0}}},{27431, {{38290, 1, 0, 0},{38290, 1, 1, 19}, {0, 0, 0, 0}}},{27431, {{38290, 1, 0, 0},{38290, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{27432, {{38291, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27433, {{38292, 1, 0, 0},{38292, 1, 1, 17}, {0, 0, 0, 0}}},{27433, {{38292, 1, 0, 0},{38292, 1, 1, 17}, {0, 0, 0, 0}}},{27434, {{38292, 1, 0, 0},{38292, 1, 1, 18}, {0, 0, 0, 0}}},{27434, {{38292, 1, 0, 0},{38292, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27435, {{38293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27436, {{38294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{27437, {{38296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27438, {{38297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27439, {{38298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27440, {{38299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27441, {{38300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27442, {{38301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27443, {{38302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27444, {{38303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27445, {{38304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27446, {{38305, 1, 0, 0},{38305, 1, 1, 17}, {0, 0, 0, 0}}},{27446, {{38305, 1, 0, 0},{38305, 1, 1, 17}, {0, 0, 0, 0}}},{27447, {{38305, 1, 0, 0},{38305, 1, 1, 18}, {0, 0, 0, 0}}},{27447, {{38305, 1, 0, 0},{38305, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27448, {{38306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27450, {{38307, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27451, {{38308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27452, {{38309, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27453, {{38310, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27454, {{38311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27455, {{38312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27456, {{38313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{27457, {{38315, 1, 0, 0},{38315, 1, 1, 18}, {0, 0, 0, 0}}},{27457, {{38315, 1, 0, 0},{38315, 1, 1, 18}, {0, 0, 0, 0}}},{27458, {{38315, 1, 0, 0},{38315, 1, 1, 19}, {0, 0, 0, 0}}},{27458, {{38315, 1, 0, 0},{38315, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{27459, {{38316, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27460, {{38317, 1, 0, 0},{38317, 1, 1, 17}, {0, 0, 0, 0}}},{27460, {{38317, 1, 0, 0},{38317, 1, 1, 17}, {0, 0, 0, 0}}},{27461, {{38317, 1, 0, 0},{38317, 1, 1, 18}, {0, 0, 0, 0}}},{27461, {{38317, 1, 0, 0},{38317, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27462, {{38318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27463, {{38319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27464, {{38320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27465, {{38321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27466, {{38322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{27468, {{38324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27469, {{38325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27470, {{38326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27471, {{38327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27472, {{38328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27473, {{38329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27474, {{38330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27475, {{38331, 1, 0, 0},{38331, 1, 1, 18}, {0, 0, 0, 0}}},{27475, {{38331, 1, 0, 0},{38331, 1, 1, 18}, {0, 0, 0, 0}}},{27476, {{38331, 1, 0, 0},{38331, 1, 1, 19}, {0, 0, 0, 0}}},{27476, {{38331, 1, 0, 0},{38331, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{27477, {{38332, 1, 0, 0},{38332, 1, 1, 18}, {0, 0, 0, 0}}},{27477, {{38332, 1, 0, 0},{38332, 1, 1, 18}, {0, 0, 0, 0}}},{27478, {{38332, 1, 0, 0},{38332, 1, 1, 19}, {0, 0, 0, 0}}},{27478, {{38332, 1, 0, 0},{38332, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{27479, {{38333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27480, {{38334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27481, {{38335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27482, {{38336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27483, {{38337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{27484, {{38339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27486, {{38340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27487, {{38341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27488, {{38342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27489, {{38343, 1, 0, 0},{38343, 1, 1, 17}, {0, 0, 0, 0}}},{27489, {{38343, 1, 0, 0},{38343, 1, 1, 17}, {0, 0, 0, 0}}},{27490, {{38343, 1, 0, 0},{38343, 1, 1, 18}, {0, 0, 0, 0}}},{27490, {{38343, 1, 0, 0},{38343, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{27491, {{38344, 1, 0, 0},{38344, 1, 1, 18}, {0, 0, 0, 0}}},{27491, {{38344, 1, 0, 0},{38344, 1, 1, 18}, {0, 0, 0, 0}}},{27492, {{38344, 1, 0, 0},{38344, 1, 1, 17}, {0, 0, 0, 0}}},{27492, {{38344, 1, 0, 0},{38344, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{27493, {{38345, 1, 0, 0},{38345, 1, 1, 17}, {0, 0, 0, 0}}},{27493, {{38345, 1, 0, 0},{38345, 1, 1, 17}, {0, 0, 0, 0}}},{27494, {{38345, 1, 0, 0},{38345, 1, 1, 18}, {0, 0, 0, 0}}},{27494, {{38345, 1, 0, 0},{38345, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27495, {{38346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27496, {{38347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27497, {{38348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27498, {{38349, 1, 0, 0},{38349, 1, 1, 18}, {0, 0, 0, 0}}},{27498, {{38349, 1, 0, 0},{38349, 1, 1, 18}, {0, 0, 0, 0}}},{27499, {{38349, 1, 0, 0},{38349, 1, 1, 19}, {0, 0, 0, 0}}},{27499, {{38349, 1, 0, 0},{38349, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{27500, {{38352, 1, 0, 0},{38352, 1, 1, 17}, {0, 0, 0, 0}}},{27500, {{38352, 1, 0, 0},{38352, 1, 1, 17}, {0, 0, 0, 0}}},{27501, {{38352, 1, 0, 0},{38352, 1, 1, 19}, {0, 0, 0, 0}}},{27501, {{38352, 1, 0, 0},{38352, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{27503, {{38353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27504, {{38354, 1, 0, 0},{38354, 1, 1, 17}, {0, 0, 0, 0}}},{27504, {{38354, 1, 0, 0},{38354, 1, 1, 17}, {0, 0, 0, 0}}},{27505, {{38354, 1, 0, 0},{38354, 1, 1, 18}, {0, 0, 0, 0}}},{27505, {{38354, 1, 0, 0},{38354, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27506, {{38355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27507, {{38356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27508, {{38357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27509, {{38358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27510, {{38359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27511, {{38360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27512, {{38361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27513, {{38362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27514, {{38363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27515, {{38364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27516, {{38365, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27517, {{38366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27518, {{38367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27519, {{38368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27520, {{38369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27521, {{38370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27522, {{38371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27523, {{38372, 1, 0, 0},{38372, 1, 1, 17}, {0, 0, 0, 0}}},{27523, {{38372, 1, 0, 0},{38372, 1, 1, 17}, {0, 0, 0, 0}}},{27524, {{38372, 1, 0, 0},{38372, 1, 1, 18}, {0, 0, 0, 0}}},{27524, {{38372, 1, 0, 0},{38372, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{27525, {{38373, 1, 0, 0},{38373, 1, 1, 18}, {0, 0, 0, 0}}},{27525, {{38373, 1, 0, 0},{38373, 1, 1, 18}, {0, 0, 0, 0}}},{27526, {{38373, 1, 0, 0},{38373, 1, 1, 17}, {0, 0, 0, 0}}},{27526, {{38373, 1, 0, 0},{38373, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{27527, {{38374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{27530, {{38428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27531, {{38429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27532, {{38430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{27533, {{38432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27534, {{38433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27535, {{38434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27536, {{38435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27537, {{38436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27538, {{38437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27539, {{38438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27540, {{38439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27541, {{38440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27542, {{38441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27543, {{38442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27544, {{38443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27545, {{38444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27546, {{38445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27547, {{38446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27548, {{38447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27549, {{38448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27550, {{38449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27551, {{38450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27552, {{38451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27553, {{38452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{27554, {{38455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27555, {{38456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27556, {{38457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27557, {{38458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27558, {{38459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27559, {{38460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27560, {{38461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27561, {{38462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27562, {{38463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27563, {{38464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27564, {{38465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27565, {{38466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27566, {{38467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27567, {{38468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{27571, {{38474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27572, {{38475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27573, {{38476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{27574, {{38477, 1, 0, 0},{38477, 1, 1, 19}, {0, 0, 0, 0}}},{27574, {{38477, 1, 0, 0},{38477, 1, 1, 19}, {0, 0, 0, 0}}},{27575, {{38477, 1, 0, 0},{38477, 1, 1, 18}, {0, 0, 0, 0}}},{27575, {{38477, 1, 0, 0},{38477, 1, 1, 18}, {0, 0, 0, 0}}},{27576, {{38477, 1, 0, 0},{38477, 1, 1, 20}, {0, 0, 0, 0}}},{27576, {{38477, 1, 0, 0},{38477, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{27577, {{38478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27578, {{38479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27579, {{38480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27580, {{38481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27581, {{38482, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27582, {{38483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27583, {{38484, 1, 0, 0},{38484, 1, 1, 17}, {0, 0, 0, 0}}},{27583, {{38484, 1, 0, 0},{38484, 1, 1, 17}, {0, 0, 0, 0}}},{27584, {{38484, 1, 0, 0},{38484, 1, 1, 18}, {0, 0, 0, 0}}},{27584, {{38484, 1, 0, 0},{38484, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27585, {{38485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27586, {{38486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27587, {{38487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27589, {{38488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27590, {{38489, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27591, {{38490, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27592, {{38491, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27593, {{38492, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27594, {{38493, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27595, {{38494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27596, {{38495, 1, 0, 0},{38495, 1, 1, 17}, {0, 0, 0, 0}}},{27596, {{38495, 1, 0, 0},{38495, 1, 1, 17}, {0, 0, 0, 0}}},{27597, {{38495, 1, 0, 0},{38495, 1, 1, 18}, {0, 0, 0, 0}}},{27597, {{38495, 1, 0, 0},{38495, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27598, {{38496, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27599, {{38497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27600, {{38498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27601, {{38499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27602, {{38500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27603, {{38501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27604, {{38502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{27605, {{38506, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27606, {{38507, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27607, {{38508, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27608, {{38509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27609, {{38510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27610, {{38511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27611, {{38512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27612, {{38513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27613, {{38514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27614, {{38515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27615, {{38516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27616, {{38517, 1, 0, 0},{38517, 1, 1, 18}, {0, 0, 0, 0}}},{27616, {{38517, 1, 0, 0},{38517, 1, 1, 18}, {0, 0, 0, 0}}},{27617, {{38517, 1, 0, 0},{38517, 1, 1, 17}, {0, 0, 0, 0}}},{27617, {{38517, 1, 0, 0},{38517, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{27618, {{38518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27619, {{38519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27620, {{38520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27621, {{38521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27622, {{38522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27623, {{38523, 1, 0, 0},{38523, 1, 1, 17}, {0, 0, 0, 0}}},{27623, {{38523, 1, 0, 0},{38523, 1, 1, 17}, {0, 0, 0, 0}}},{27624, {{38523, 1, 0, 0},{38523, 1, 1, 18}, {0, 0, 0, 0}}},{27624, {{38523, 1, 0, 0},{38523, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27625, {{38524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27626, {{38525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27627, {{38526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27628, {{38527, 1, 0, 0},{38527, 1, 1, 17}, {0, 0, 0, 0}}},{27628, {{38527, 1, 0, 0},{38527, 1, 1, 17}, {0, 0, 0, 0}}},{27629, {{38527, 1, 0, 0},{38527, 1, 1, 18}, {0, 0, 0, 0}}},{27629, {{38527, 1, 0, 0},{38527, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27630, {{38528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27631, {{38529, 1, 0, 0},{38529, 1, 1, 17}, {0, 0, 0, 0}}},{27631, {{38529, 1, 0, 0},{38529, 1, 1, 17}, {0, 0, 0, 0}}},{27632, {{38529, 1, 0, 0},{38529, 1, 1, 18}, {0, 0, 0, 0}}},{27632, {{38529, 1, 0, 0},{38529, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{27633, {{38530, 1, 0, 0},{38530, 1, 1, 17}, {0, 0, 0, 0}}},{27633, {{38530, 1, 0, 0},{38530, 1, 1, 17}, {0, 0, 0, 0}}},{27634, {{38530, 1, 0, 0},{38530, 1, 1, 18}, {0, 0, 0, 0}}},{27634, {{38530, 1, 0, 0},{38530, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27635, {{38531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27637, {{38532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27638, {{38533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {9, {{27639, {{38534, 1, 0, 0},{38534, 1, 1, 19}, {0, 0, 0, 0}}},{27639, {{38534, 1, 0, 0},{38534, 1, 1, 19}, {0, 0, 0, 0}}},{27640, {{38534, 1, 0, 0},{38534, 1, 1, 20}, {0, 0, 0, 0}}},{27640, {{38534, 1, 0, 0},{38534, 1, 1, 20}, {0, 0, 0, 0}}},{30189, {{38534, 1, 0, 0},{38534, 1, 1, 0},{38534, 1, 1, 22},{63964, 1, 0, 0}, {0, 0, 0, 0}}},{30189, {{38534, 1, 0, 0},{38534, 1, 1, 0},{38534, 1, 1, 22},{63964, 1, 0, 0}, {0, 0, 0, 0}}},{30189, {{38534, 1, 0, 0},{38534, 1, 1, 0},{38534, 1, 1, 22},{63964, 1, 0, 0}, {0, 0, 0, 0}}},{30190, {{38534, 1, 0, 0},{38534, 1, 1, 21}, {0, 0, 0, 0}}},{30190, {{38534, 1, 0, 0},{38534, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{27641, {{38535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27642, {{38536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27643, {{38537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{27644, {{38538, 1, 0, 0},{38538, 1, 1, 20}, {0, 0, 0, 0}}},{27644, {{38538, 1, 0, 0},{38538, 1, 1, 20}, {0, 0, 0, 0}}},{27645, {{38538, 1, 0, 0},{38538, 1, 1, 18}, {0, 0, 0, 0}}},{27645, {{38538, 1, 0, 0},{38538, 1, 1, 18}, {0, 0, 0, 0}}},{27646, {{38538, 1, 0, 0},{38538, 1, 1, 19}, {0, 0, 0, 0}}},{27646, {{38538, 1, 0, 0},{38538, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{27647, {{38539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27648, {{38540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27649, {{38541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27650, {{38542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27651, {{38543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58972, {{38544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27652, {{38545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27653, {{38546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27655, {{38547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27656, {{38548, 1, 0, 0},{38548, 1, 1, 19}, {0, 0, 0, 0}}},{27656, {{38548, 1, 0, 0},{38548, 1, 1, 19}, {0, 0, 0, 0}}},{27657, {{38548, 1, 0, 0},{38548, 1, 1, 18}, {0, 0, 0, 0}}},{27657, {{38548, 1, 0, 0},{38548, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27658, {{38549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27659, {{38550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27660, {{38551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{27661, {{38552, 1, 0, 0},{38552, 1, 1, 19}, {0, 0, 0, 0}}},{27661, {{38552, 1, 0, 0},{38552, 1, 1, 19}, {0, 0, 0, 0}}},{27662, {{38552, 1, 0, 0},{38552, 1, 1, 20}, {0, 0, 0, 0}}},{27662, {{38552, 1, 0, 0},{38552, 1, 1, 20}, {0, 0, 0, 0}}},{27663, {{38552, 1, 0, 0},{38552, 1, 1, 22}, {0, 0, 0, 0}}},{27663, {{38552, 1, 0, 0},{38552, 1, 1, 22}, {0, 0, 0, 0}}},{27664, {{38552, 1, 0, 0},{38552, 1, 1, 21}, {0, 0, 0, 0}}},{27664, {{38552, 1, 0, 0},{38552, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {4, {{27665, {{38553, 1, 0, 0},{38553, 1, 1, 18}, {0, 0, 0, 0}}},{27665, {{38553, 1, 0, 0},{38553, 1, 1, 18}, {0, 0, 0, 0}}},{27666, {{38553, 1, 0, 0},{38553, 1, 1, 19}, {0, 0, 0, 0}}},{27666, {{38553, 1, 0, 0},{38553, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{27667, {{38554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27668, {{38555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27669, {{38556, 1, 0, 0},{38556, 1, 1, 17}, {0, 0, 0, 0}}},{27669, {{38556, 1, 0, 0},{38556, 1, 1, 17}, {0, 0, 0, 0}}},{27670, {{38556, 1, 0, 0},{38556, 1, 1, 18}, {0, 0, 0, 0}}},{27670, {{38556, 1, 0, 0},{38556, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27671, {{38557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27672, {{38558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27673, {{38559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27674, {{38560, 1, 0, 0},{38560, 1, 1, 18}, {0, 0, 0, 0}}},{27674, {{38560, 1, 0, 0},{38560, 1, 1, 18}, {0, 0, 0, 0}}},{27675, {{38560, 1, 0, 0},{38560, 1, 1, 17}, {0, 0, 0, 0}}},{27675, {{38560, 1, 0, 0},{38560, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{27677, {{38562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{27678, {{38563, 1, 0, 0},{38563, 1, 1, 20}, {0, 0, 0, 0}}},{27678, {{38563, 1, 0, 0},{38563, 1, 1, 20}, {0, 0, 0, 0}}},{27679, {{38563, 1, 0, 0},{38563, 1, 1, 19}, {0, 0, 0, 0}}},{27679, {{38563, 1, 0, 0},{38563, 1, 1, 19}, {0, 0, 0, 0}}},{27680, {{38563, 1, 0, 0},{38563, 1, 1, 21}, {0, 0, 0, 0}}},{27680, {{38563, 1, 0, 0},{38563, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{27681, {{38564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27682, {{38565, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27683, {{38566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{27684, {{38567, 1, 0, 0},{38567, 1, 1, 20}, {0, 0, 0, 0}}},{27684, {{38567, 1, 0, 0},{38567, 1, 1, 20}, {0, 0, 0, 0}}},{27685, {{38567, 1, 0, 0},{38567, 1, 1, 18}, {0, 0, 0, 0}}},{27685, {{38567, 1, 0, 0},{38567, 1, 1, 18}, {0, 0, 0, 0}}},{27686, {{38567, 1, 0, 0},{38567, 1, 1, 19}, {0, 0, 0, 0}}},{27686, {{38567, 1, 0, 0},{38567, 1, 1, 19}, {0, 0, 0, 0}}},{27687, {{38567, 1, 0, 0},{38567, 1, 1, 21}, {0, 0, 0, 0}}},{27687, {{38567, 1, 0, 0},{38567, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {4, {{27688, {{38568, 1, 0, 0},{38568, 1, 1, 18}, {0, 0, 0, 0}}},{27688, {{38568, 1, 0, 0},{38568, 1, 1, 18}, {0, 0, 0, 0}}},{27689, {{38568, 1, 0, 0},{38568, 1, 1, 17}, {0, 0, 0, 0}}},{27689, {{38568, 1, 0, 0},{38568, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{27690, {{38569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27691, {{38570, 1, 0, 0},{38570, 1, 1, 17}, {0, 0, 0, 0}}},{27691, {{38570, 1, 0, 0},{38570, 1, 1, 17}, {0, 0, 0, 0}}},{58977, {{38570, 1, 0, 0},{38570, 1, 1, 18}, {0, 0, 0, 0}}},{58977, {{38570, 1, 0, 0},{38570, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27692, {{38571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27693, {{38572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27694, {{38573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27695, {{38574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27696, {{38575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27697, {{38576, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27698, {{38577, 1, 0, 0},{38577, 1, 1, 17}, {0, 0, 0, 0}}},{27698, {{38577, 1, 0, 0},{38577, 1, 1, 17}, {0, 0, 0, 0}}},{27699, {{38577, 1, 0, 0},{38577, 1, 1, 18}, {0, 0, 0, 0}}},{27699, {{38577, 1, 0, 0},{38577, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27700, {{38578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27701, {{38579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27702, {{38580, 1, 0, 0},{38580, 1, 1, 17}, {0, 0, 0, 0}}},{27702, {{38580, 1, 0, 0},{38580, 1, 1, 17}, {0, 0, 0, 0}}},{27703, {{38580, 1, 0, 0},{38580, 1, 1, 18}, {0, 0, 0, 0}}},{27703, {{38580, 1, 0, 0},{38580, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27704, {{38581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27705, {{38582, 1, 0, 0},{38582, 1, 1, 18}, {0, 0, 0, 0}}},{27705, {{38582, 1, 0, 0},{38582, 1, 1, 18}, {0, 0, 0, 0}}},{27706, {{38582, 1, 0, 0},{38582, 1, 1, 17}, {0, 0, 0, 0}}},{27706, {{38582, 1, 0, 0},{38582, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{27707, {{38583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27708, {{38584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27709, {{38585, 1, 0, 0},{38585, 1, 1, 17}, {0, 0, 0, 0}}},{27709, {{38585, 1, 0, 0},{38585, 1, 1, 17}, {0, 0, 0, 0}}},{56921, {{38585, 1, 0, 0},{38585, 1, 1, 18}, {0, 0, 0, 0}}},{56921, {{38585, 1, 0, 0},{38585, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27710, {{38586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27711, {{38587, 1, 0, 0},{38587, 1, 1, 18}, {0, 0, 0, 0}}},{27711, {{38587, 1, 0, 0},{38587, 1, 1, 18}, {0, 0, 0, 0}}},{27712, {{38587, 1, 0, 0},{38587, 1, 1, 19}, {0, 0, 0, 0}}},{27712, {{38587, 1, 0, 0},{38587, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{27713, {{38588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27714, {{38589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{27715, {{38591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27716, {{38592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27717, {{38593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27718, {{38594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27719, {{38595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27720, {{38596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{27721, {{38597, 1, 0, 0},{38597, 1, 1, 20}, {0, 0, 0, 0}}},{27721, {{38597, 1, 0, 0},{38597, 1, 1, 20}, {0, 0, 0, 0}}},{27722, {{38597, 1, 0, 0},{38597, 1, 1, 19}, {0, 0, 0, 0}}},{27722, {{38597, 1, 0, 0},{38597, 1, 1, 19}, {0, 0, 0, 0}}},{27723, {{38597, 1, 0, 0},{38597, 1, 1, 21}, {0, 0, 0, 0}}},{27723, {{38597, 1, 0, 0},{38597, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{27724, {{38598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27725, {{38599, 1, 0, 0},{38599, 1, 1, 19}, {0, 0, 0, 0}}},{27725, {{38599, 1, 0, 0},{38599, 1, 1, 19}, {0, 0, 0, 0}}},{27726, {{38599, 1, 0, 0},{38599, 1, 1, 18}, {0, 0, 0, 0}}},{27726, {{38599, 1, 0, 0},{38599, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27727, {{38600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27728, {{38601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27729, {{38602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27730, {{38603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27731, {{38604, 1, 0, 0},{38604, 1, 1, 17}, {0, 0, 0, 0}}},{27731, {{38604, 1, 0, 0},{38604, 1, 1, 17}, {0, 0, 0, 0}}},{27732, {{38604, 1, 0, 0},{38604, 1, 1, 18}, {0, 0, 0, 0}}},{27732, {{38604, 1, 0, 0},{38604, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27733, {{38605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27734, {{38606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27735, {{38607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27736, {{38608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27737, {{38609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27738, {{38610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27739, {{38611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27740, {{38612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27741, {{38613, 1, 0, 0},{38613, 1, 1, 17}, {0, 0, 0, 0}}},{27741, {{38613, 1, 0, 0},{38613, 1, 1, 17}, {0, 0, 0, 0}}},{27742, {{38613, 1, 0, 0},{38613, 1, 1, 18}, {0, 0, 0, 0}}},{27742, {{38613, 1, 0, 0},{38613, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27743, {{38614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27744, {{38615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{27745, {{38616, 1, 0, 0},{38616, 1, 1, 19}, {0, 0, 0, 0}}},{27745, {{38616, 1, 0, 0},{38616, 1, 1, 19}, {0, 0, 0, 0}}},{27746, {{38616, 1, 0, 0},{38616, 1, 1, 18}, {0, 0, 0, 0}}},{27746, {{38616, 1, 0, 0},{38616, 1, 1, 18}, {0, 0, 0, 0}}},{27747, {{38616, 1, 0, 0},{38616, 1, 1, 17}, {0, 0, 0, 0}}},{27747, {{38616, 1, 0, 0},{38616, 1, 1, 17}, {0, 0, 0, 0}}},{27748, {{38616, 1, 0, 0},{38616, 1, 1, 20}, {0, 0, 0, 0}}},{27748, {{38616, 1, 0, 0},{38616, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{27749, {{38617, 1, 0, 0},{38617, 1, 1, 18}, {0, 0, 0, 0}}},{27749, {{38617, 1, 0, 0},{38617, 1, 1, 18}, {0, 0, 0, 0}}},{27750, {{38617, 1, 0, 0},{38617, 1, 1, 19}, {0, 0, 0, 0}}},{27750, {{38617, 1, 0, 0},{38617, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {6, {{27751, {{38618, 1, 0, 0},{38618, 1, 1, 20}, {0, 0, 0, 0}}},{27751, {{38618, 1, 0, 0},{38618, 1, 1, 20}, {0, 0, 0, 0}}},{27752, {{38618, 1, 0, 0},{38618, 1, 1, 18}, {0, 0, 0, 0}}},{27752, {{38618, 1, 0, 0},{38618, 1, 1, 18}, {0, 0, 0, 0}}},{27753, {{38618, 1, 0, 0},{38618, 1, 1, 19}, {0, 0, 0, 0}}},{27753, {{38618, 1, 0, 0},{38618, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{27754, {{38619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27755, {{38620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27756, {{38621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27757, {{38622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27758, {{38623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{27759, {{38625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27760, {{38626, 1, 0, 0},{38626, 1, 1, 17}, {0, 0, 0, 0}}},{27760, {{38626, 1, 0, 0},{38626, 1, 1, 17}, {0, 0, 0, 0}}},{27761, {{38626, 1, 0, 0},{38626, 1, 1, 18}, {0, 0, 0, 0}}},{27761, {{38626, 1, 0, 0},{38626, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {9, {{27762, {{38627, 1, 0, 0},{38627, 1, 1, 18}, {0, 0, 0, 0}}},{27762, {{38627, 1, 0, 0},{38627, 1, 1, 18}, {0, 0, 0, 0}}},{30304, {{38627, 1, 0, 0},{38627, 1, 1, 0},{38627, 1, 1, 20},{64104, 1, 0, 0}, {0, 0, 0, 0}}},{30304, {{38627, 1, 0, 0},{38627, 1, 1, 0},{38627, 1, 1, 20},{64104, 1, 0, 0}, {0, 0, 0, 0}}},{30304, {{38627, 1, 0, 0},{38627, 1, 1, 0},{38627, 1, 1, 20},{64104, 1, 0, 0}, {0, 0, 0, 0}}},{58983, {{38627, 1, 0, 0},{38627, 1, 1, 24}, {0, 0, 0, 0}}},{58983, {{38627, 1, 0, 0},{38627, 1, 1, 24}, {0, 0, 0, 0}}},{58984, {{38627, 1, 0, 0},{38627, 1, 1, 19}, {0, 0, 0, 0}}},{58984, {{38627, 1, 0, 0},{38627, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{27766, {{38628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27767, {{38629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27768, {{38630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27769, {{38631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27770, {{38632, 1, 0, 0},{38632, 1, 1, 18}, {0, 0, 0, 0}}},{27770, {{38632, 1, 0, 0},{38632, 1, 1, 18}, {0, 0, 0, 0}}},{27771, {{38632, 1, 0, 0},{38632, 1, 1, 19}, {0, 0, 0, 0}}},{27771, {{38632, 1, 0, 0},{38632, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{27772, {{38633, 1, 0, 0},{38633, 1, 1, 17}, {0, 0, 0, 0}}},{27772, {{38633, 1, 0, 0},{38633, 1, 1, 17}, {0, 0, 0, 0}}},{27773, {{38633, 1, 0, 0},{38633, 1, 1, 18}, {0, 0, 0, 0}}},{27773, {{38633, 1, 0, 0},{38633, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{27774, {{38634, 1, 0, 0},{38634, 1, 1, 19}, {0, 0, 0, 0}}},{27774, {{38634, 1, 0, 0},{38634, 1, 1, 19}, {0, 0, 0, 0}}},{27775, {{38634, 1, 0, 0},{38634, 1, 1, 18}, {0, 0, 0, 0}}},{27775, {{38634, 1, 0, 0},{38634, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27778, {{38635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{27779, {{38637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27780, {{38638, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27781, {{38639, 1, 0, 0},{38639, 1, 1, 17}, {0, 0, 0, 0}}},{27781, {{38639, 1, 0, 0},{38639, 1, 1, 17}, {0, 0, 0, 0}}},{27782, {{38639, 1, 0, 0},{38639, 1, 1, 18}, {0, 0, 0, 0}}},{27782, {{38639, 1, 0, 0},{38639, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{27783, {{38640, 1, 0, 0},{38640, 1, 1, 18}, {0, 0, 0, 0}}},{27783, {{38640, 1, 0, 0},{38640, 1, 1, 18}, {0, 0, 0, 0}}},{27784, {{38640, 1, 0, 0},{38640, 1, 1, 19}, {0, 0, 0, 0}}},{27784, {{38640, 1, 0, 0},{38640, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{27785, {{38641, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27786, {{38642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{27787, {{38644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27788, {{38645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27789, {{38646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27790, {{38647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27792, {{38648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27793, {{38649, 1, 0, 0},{38649, 1, 1, 17}, {0, 0, 0, 0}}},{27793, {{38649, 1, 0, 0},{38649, 1, 1, 17}, {0, 0, 0, 0}}},{27794, {{38649, 1, 0, 0},{38649, 1, 1, 18}, {0, 0, 0, 0}}},{27794, {{38649, 1, 0, 0},{38649, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27795, {{38650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27796, {{38651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27797, {{38652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27798, {{38653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{27799, {{38655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27800, {{38656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{27801, {{38658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27802, {{38659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27803, {{38660, 1, 0, 0},{38660, 1, 1, 17}, {0, 0, 0, 0}}},{27803, {{38660, 1, 0, 0},{38660, 1, 1, 17}, {0, 0, 0, 0}}},{27804, {{38660, 1, 0, 0},{38660, 1, 1, 18}, {0, 0, 0, 0}}},{27804, {{38660, 1, 0, 0},{38660, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27805, {{38661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{27806, {{38662, 1, 0, 0},{38662, 1, 1, 19}, {0, 0, 0, 0}}},{27806, {{38662, 1, 0, 0},{38662, 1, 1, 19}, {0, 0, 0, 0}}},{27807, {{38662, 1, 0, 0},{38662, 1, 1, 18}, {0, 0, 0, 0}}},{27807, {{38662, 1, 0, 0},{38662, 1, 1, 18}, {0, 0, 0, 0}}},{27808, {{38662, 1, 0, 0},{38662, 1, 1, 20}, {0, 0, 0, 0}}},{27808, {{38662, 1, 0, 0},{38662, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{27809, {{38663, 1, 0, 0},{38663, 1, 1, 17}, {0, 0, 0, 0}}},{27809, {{38663, 1, 0, 0},{38663, 1, 1, 17}, {0, 0, 0, 0}}},{58994, {{38663, 1, 0, 0},{38663, 1, 1, 18}, {0, 0, 0, 0}}},{58994, {{38663, 1, 0, 0},{38663, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27810, {{38664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27811, {{38665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27812, {{38666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27813, {{38667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27814, {{38668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27816, {{38669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27817, {{38670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27818, {{38671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27819, {{38672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27820, {{38673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27821, {{38674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27822, {{38675, 1, 0, 0},{38675, 1, 1, 17}, {0, 0, 0, 0}}},{27822, {{38675, 1, 0, 0},{38675, 1, 1, 17}, {0, 0, 0, 0}}},{27823, {{38675, 1, 0, 0},{38675, 1, 1, 18}, {0, 0, 0, 0}}},{27823, {{38675, 1, 0, 0},{38675, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27824, {{38676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27825, {{38677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27826, {{38678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27827, {{38679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27828, {{38680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27829, {{38681, 1, 0, 0},{38681, 1, 1, 17}, {0, 0, 0, 0}}},{27829, {{38681, 1, 0, 0},{38681, 1, 1, 17}, {0, 0, 0, 0}}},{27830, {{38681, 1, 0, 0},{38681, 1, 1, 18}, {0, 0, 0, 0}}},{27830, {{38681, 1, 0, 0},{38681, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27831, {{38682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27832, {{38683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27833, {{38684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27834, {{38685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27835, {{38686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27836, {{38687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27837, {{38688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{27838, {{38689, 1, 0, 0},{38689, 1, 1, 19}, {0, 0, 0, 0}}},{27838, {{38689, 1, 0, 0},{38689, 1, 1, 19}, {0, 0, 0, 0}}},{27839, {{38689, 1, 0, 0},{38689, 1, 1, 18}, {0, 0, 0, 0}}},{27839, {{38689, 1, 0, 0},{38689, 1, 1, 18}, {0, 0, 0, 0}}},{27840, {{38689, 1, 0, 0},{38689, 1, 1, 20}, {0, 0, 0, 0}}},{27840, {{38689, 1, 0, 0},{38689, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{27841, {{38690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27842, {{38691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27843, {{38692, 1, 0, 0},{38692, 1, 1, 19}, {0, 0, 0, 0}}},{27843, {{38692, 1, 0, 0},{38692, 1, 1, 19}, {0, 0, 0, 0}}},{27844, {{38692, 1, 0, 0},{38692, 1, 1, 18}, {0, 0, 0, 0}}},{27844, {{38692, 1, 0, 0},{38692, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27845, {{38693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27846, {{38694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27847, {{38695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27848, {{38696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27849, {{38697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27850, {{38698, 1, 0, 0},{38698, 1, 1, 17}, {0, 0, 0, 0}}},{27850, {{38698, 1, 0, 0},{38698, 1, 1, 17}, {0, 0, 0, 0}}},{27851, {{38698, 1, 0, 0},{38698, 1, 1, 18}, {0, 0, 0, 0}}},{27851, {{38698, 1, 0, 0},{38698, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27852, {{38699, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27853, {{38700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{27854, {{38702, 1, 0, 0},{38702, 1, 1, 16}, {0, 0, 0, 0}}},{27854, {{38702, 1, 0, 0},{38702, 1, 1, 16}, {0, 0, 0, 0}}},{27855, {{38702, 1, 0, 0},{38702, 1, 1, 17}, {0, 0, 0, 0}}},{27855, {{38702, 1, 0, 0},{38702, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{27856, {{38703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27857, {{38704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27858, {{38705, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27859, {{38706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27860, {{38707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27861, {{38708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27862, {{38709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27863, {{38710, 1, 0, 0},{38710, 1, 1, 17}, {0, 0, 0, 0}}},{27863, {{38710, 1, 0, 0},{38710, 1, 1, 17}, {0, 0, 0, 0}}},{27864, {{38710, 1, 0, 0},{38710, 1, 1, 18}, {0, 0, 0, 0}}},{27864, {{38710, 1, 0, 0},{38710, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27865, {{38711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27866, {{38712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27867, {{38713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27868, {{38714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27869, {{38715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27870, {{38716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27871, {{38717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27872, {{38718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27873, {{38719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27874, {{38720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27875, {{38721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27876, {{38722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27878, {{38723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27879, {{38724, 1, 0, 0},{38724, 1, 1, 19}, {0, 0, 0, 0}}},{27879, {{38724, 1, 0, 0},{38724, 1, 1, 19}, {0, 0, 0, 0}}},{27880, {{38724, 1, 0, 0},{38724, 1, 1, 17}, {0, 0, 0, 0}}},{27880, {{38724, 1, 0, 0},{38724, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{27881, {{38725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27882, {{38726, 1, 0, 0},{38726, 1, 1, 18}, {0, 0, 0, 0}}},{27882, {{38726, 1, 0, 0},{38726, 1, 1, 18}, {0, 0, 0, 0}}},{27883, {{38726, 1, 0, 0},{38726, 1, 1, 17}, {0, 0, 0, 0}}},{27883, {{38726, 1, 0, 0},{38726, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{27884, {{38727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27886, {{38728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27887, {{38729, 1, 0, 0},{38729, 1, 1, 17}, {0, 0, 0, 0}}},{27887, {{38729, 1, 0, 0},{38729, 1, 1, 17}, {0, 0, 0, 0}}},{27888, {{38729, 1, 0, 0},{38729, 1, 1, 18}, {0, 0, 0, 0}}},{27888, {{38729, 1, 0, 0},{38729, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27889, {{38730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27890, {{38731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27891, {{38732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27892, {{38733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27893, {{38734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27894, {{38735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27895, {{38736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27896, {{38737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27897, {{38738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{27898, {{38741, 1, 0, 0},{38741, 1, 1, 19}, {0, 0, 0, 0}}},{27898, {{38741, 1, 0, 0},{38741, 1, 1, 19}, {0, 0, 0, 0}}},{27899, {{38741, 1, 0, 0},{38741, 1, 1, 20}, {0, 0, 0, 0}}},{27899, {{38741, 1, 0, 0},{38741, 1, 1, 20}, {0, 0, 0, 0}}},{27900, {{38741, 1, 0, 0},{38741, 1, 1, 18}, {0, 0, 0, 0}}},{27900, {{38741, 1, 0, 0},{38741, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {5, {{27901, {{38742, 1, 0, 0},{38742, 1, 1, 18}, {0, 0, 0, 0}}},{27901, {{38742, 1, 0, 0},{38742, 1, 1, 18}, {0, 0, 0, 0}}},{30214, {{38742, 1, 0, 0},{38742, 1, 1, 0},{38742, 1, 1, 19},{64028, 1, 0, 0}, {0, 0, 0, 0}}},{30214, {{38742, 1, 0, 0},{38742, 1, 1, 0},{38742, 1, 1, 19},{64028, 1, 0, 0}, {0, 0, 0, 0}}},{30214, {{38742, 1, 0, 0},{38742, 1, 1, 0},{38742, 1, 1, 19},{64028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27902, {{38743, 1, 0, 0},{38743, 1, 1, 17}, {0, 0, 0, 0}}},{27902, {{38743, 1, 0, 0},{38743, 1, 1, 17}, {0, 0, 0, 0}}},{27903, {{38743, 1, 0, 0},{38743, 1, 1, 18}, {0, 0, 0, 0}}},{27903, {{38743, 1, 0, 0},{38743, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{27904, {{38744, 1, 0, 0},{38744, 1, 1, 17}, {0, 0, 0, 0}}},{27904, {{38744, 1, 0, 0},{38744, 1, 1, 17}, {0, 0, 0, 0}}},{27905, {{38744, 1, 0, 0},{38744, 1, 1, 18}, {0, 0, 0, 0}}},{27905, {{38744, 1, 0, 0},{38744, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27906, {{38745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27909, {{38746, 1, 0, 0},{38746, 1, 1, 17}, {0, 0, 0, 0}}},{27909, {{38746, 1, 0, 0},{38746, 1, 1, 17}, {0, 0, 0, 0}}},{27910, {{38746, 1, 0, 0},{38746, 1, 1, 18}, {0, 0, 0, 0}}},{27910, {{38746, 1, 0, 0},{38746, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{27911, {{38747, 1, 0, 0},{38747, 1, 1, 17}, {0, 0, 0, 0}}},{27911, {{38747, 1, 0, 0},{38747, 1, 1, 17}, {0, 0, 0, 0}}},{27912, {{38747, 1, 0, 0},{38747, 1, 1, 18}, {0, 0, 0, 0}}},{27912, {{38747, 1, 0, 0},{38747, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{27913, {{38748, 1, 0, 0},{38748, 1, 1, 19}, {0, 0, 0, 0}}},{27913, {{38748, 1, 0, 0},{38748, 1, 1, 19}, {0, 0, 0, 0}}},{27914, {{38748, 1, 0, 0},{38748, 1, 1, 20}, {0, 0, 0, 0}}},{27914, {{38748, 1, 0, 0},{38748, 1, 1, 20}, {0, 0, 0, 0}}},{27915, {{38748, 1, 0, 0},{38748, 1, 1, 21}, {0, 0, 0, 0}}},{27915, {{38748, 1, 0, 0},{38748, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {4, {{27918, {{38749, 1, 0, 0},{38749, 1, 1, 17}, {0, 0, 0, 0}}},{27918, {{38749, 1, 0, 0},{38749, 1, 1, 17}, {0, 0, 0, 0}}},{27919, {{38749, 1, 0, 0},{38749, 1, 1, 16}, {0, 0, 0, 0}}},{27919, {{38749, 1, 0, 0},{38749, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {4, {{27920, {{38750, 1, 0, 0},{38750, 1, 1, 17}, {0, 0, 0, 0}}},{27920, {{38750, 1, 0, 0},{38750, 1, 1, 17}, {0, 0, 0, 0}}},{27921, {{38750, 1, 0, 0},{38750, 1, 1, 18}, {0, 0, 0, 0}}},{27921, {{38750, 1, 0, 0},{38750, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27922, {{38751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{27923, {{38752, 1, 0, 0},{38752, 1, 1, 21}, {0, 0, 0, 0}}},{27923, {{38752, 1, 0, 0},{38752, 1, 1, 21}, {0, 0, 0, 0}}},{27924, {{38752, 1, 0, 0},{38752, 1, 1, 22}, {0, 0, 0, 0}}},{27924, {{38752, 1, 0, 0},{38752, 1, 1, 22}, {0, 0, 0, 0}}},{27925, {{38752, 1, 0, 0},{38752, 1, 1, 20}, {0, 0, 0, 0}}},{27925, {{38752, 1, 0, 0},{38752, 1, 1, 20}, {0, 0, 0, 0}}},{27926, {{38752, 1, 0, 0},{38752, 1, 1, 23}, {0, 0, 0, 0}}},{27926, {{38752, 1, 0, 0},{38752, 1, 1, 23}, {0, 0, 0, 0}}}}},
    {4, {{27927, {{38753, 1, 0, 0},{38753, 1, 1, 17}, {0, 0, 0, 0}}},{27927, {{38753, 1, 0, 0},{38753, 1, 1, 17}, {0, 0, 0, 0}}},{27928, {{38753, 1, 0, 0},{38753, 1, 1, 18}, {0, 0, 0, 0}}},{27928, {{38753, 1, 0, 0},{38753, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27929, {{38754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27930, {{38755, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27931, {{38756, 1, 0, 0},{38756, 1, 1, 17}, {0, 0, 0, 0}}},{27931, {{38756, 1, 0, 0},{38756, 1, 1, 17}, {0, 0, 0, 0}}},{27932, {{38756, 1, 0, 0},{38756, 1, 1, 18}, {0, 0, 0, 0}}},{27932, {{38756, 1, 0, 0},{38756, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{27933, {{38758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27934, {{38759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27935, {{38760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27936, {{38761, 1, 0, 0},{38761, 1, 1, 17}, {0, 0, 0, 0}}},{27936, {{38761, 1, 0, 0},{38761, 1, 1, 17}, {0, 0, 0, 0}}},{59006, {{38761, 1, 0, 0},{38761, 1, 1, 18}, {0, 0, 0, 0}}},{59006, {{38761, 1, 0, 0},{38761, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27937, {{38762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27938, {{38763, 1, 0, 0},{38763, 1, 1, 17}, {0, 0, 0, 0}}},{27938, {{38763, 1, 0, 0},{38763, 1, 1, 17}, {0, 0, 0, 0}}},{27939, {{38763, 1, 0, 0},{38763, 1, 1, 18}, {0, 0, 0, 0}}},{27939, {{38763, 1, 0, 0},{38763, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27940, {{38764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27941, {{38765, 1, 0, 0},{38765, 1, 1, 20}, {0, 0, 0, 0}}},{27941, {{38765, 1, 0, 0},{38765, 1, 1, 20}, {0, 0, 0, 0}}},{27942, {{38765, 1, 0, 0},{38765, 1, 1, 19}, {0, 0, 0, 0}}},{27942, {{38765, 1, 0, 0},{38765, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{27943, {{38766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27944, {{38767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{27945, {{38769, 1, 0, 0},{38769, 1, 1, 21}, {0, 0, 0, 0}}},{27945, {{38769, 1, 0, 0},{38769, 1, 1, 21}, {0, 0, 0, 0}}},{27946, {{38769, 1, 0, 0},{38769, 1, 1, 22}, {0, 0, 0, 0}}},{27946, {{38769, 1, 0, 0},{38769, 1, 1, 22}, {0, 0, 0, 0}}},{27947, {{38769, 1, 0, 0},{38769, 1, 1, 20}, {0, 0, 0, 0}}},{27947, {{38769, 1, 0, 0},{38769, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{27948, {{38770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27949, {{38771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27950, {{38772, 1, 0, 0},{38772, 1, 1, 18}, {0, 0, 0, 0}}},{27950, {{38772, 1, 0, 0},{38772, 1, 1, 18}, {0, 0, 0, 0}}},{27951, {{38772, 1, 0, 0},{38772, 1, 1, 19}, {0, 0, 0, 0}}},{27951, {{38772, 1, 0, 0},{38772, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{27952, {{38773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27953, {{38774, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27954, {{38775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27955, {{38776, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27957, {{38777, 1, 0, 0},{38777, 1, 1, 17}, {0, 0, 0, 0}}},{27957, {{38777, 1, 0, 0},{38777, 1, 1, 17}, {0, 0, 0, 0}}},{27958, {{38777, 1, 0, 0},{38777, 1, 1, 18}, {0, 0, 0, 0}}},{27958, {{38777, 1, 0, 0},{38777, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27959, {{38778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27960, {{38779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27961, {{38780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27962, {{38781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27963, {{38782, 1, 0, 0},{38782, 1, 1, 16}, {0, 0, 0, 0}}},{27963, {{38782, 1, 0, 0},{38782, 1, 1, 16}, {0, 0, 0, 0}}},{27964, {{38782, 1, 0, 0},{38782, 1, 1, 17}, {0, 0, 0, 0}}},{27964, {{38782, 1, 0, 0},{38782, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{27965, {{38783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27966, {{38784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27967, {{38785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27968, {{38786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27969, {{38787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27970, {{38788, 1, 0, 0},{38788, 1, 1, 19}, {0, 0, 0, 0}}},{27970, {{38788, 1, 0, 0},{38788, 1, 1, 19}, {0, 0, 0, 0}}},{27971, {{38788, 1, 0, 0},{38788, 1, 1, 18}, {0, 0, 0, 0}}},{27971, {{38788, 1, 0, 0},{38788, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27972, {{38789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27973, {{38790, 1, 0, 0},{38790, 1, 1, 17}, {0, 0, 0, 0}}},{27973, {{38790, 1, 0, 0},{38790, 1, 1, 17}, {0, 0, 0, 0}}},{27974, {{38790, 1, 0, 0},{38790, 1, 1, 18}, {0, 0, 0, 0}}},{27974, {{38790, 1, 0, 0},{38790, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27975, {{38791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27976, {{38792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27977, {{38793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27978, {{38794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27979, {{38795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27980, {{38796, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27981, {{38797, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27982, {{38798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27983, {{38799, 1, 0, 0},{38799, 1, 1, 17}, {0, 0, 0, 0}}},{27983, {{38799, 1, 0, 0},{38799, 1, 1, 17}, {0, 0, 0, 0}}},{27984, {{38799, 1, 0, 0},{38799, 1, 1, 18}, {0, 0, 0, 0}}},{27984, {{38799, 1, 0, 0},{38799, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27985, {{38800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{27986, {{38803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27987, {{38804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27988, {{38805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27989, {{38806, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27990, {{38807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27991, {{38808, 1, 0, 0},{38808, 1, 1, 19}, {0, 0, 0, 0}}},{27991, {{38808, 1, 0, 0},{38808, 1, 1, 19}, {0, 0, 0, 0}}},{27992, {{38808, 1, 0, 0},{38808, 1, 1, 18}, {0, 0, 0, 0}}},{27992, {{38808, 1, 0, 0},{38808, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27993, {{38809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27994, {{38810, 1, 0, 0},{38810, 1, 1, 17}, {0, 0, 0, 0}}},{27994, {{38810, 1, 0, 0},{38810, 1, 1, 17}, {0, 0, 0, 0}}},{27995, {{38810, 1, 0, 0},{38810, 1, 1, 18}, {0, 0, 0, 0}}},{27995, {{38810, 1, 0, 0},{38810, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{27996, {{38811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27997, {{38812, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27998, {{38813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{27999, {{38814, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28000, {{38815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28001, {{38816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{28002, {{38818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28003, {{38819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28004, {{38820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28005, {{38821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28006, {{38822, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28007, {{38823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{28008, {{38824, 1, 0, 0},{38824, 1, 1, 17}, {0, 0, 0, 0}}},{28008, {{38824, 1, 0, 0},{38824, 1, 1, 17}, {0, 0, 0, 0}}},{28009, {{38824, 1, 0, 0},{38824, 1, 1, 18}, {0, 0, 0, 0}}},{28009, {{38824, 1, 0, 0},{38824, 1, 1, 18}, {0, 0, 0, 0}}},{28010, {{38824, 1, 0, 0},{38824, 1, 1, 19}, {0, 0, 0, 0}}},{28010, {{38824, 1, 0, 0},{38824, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{28011, {{38825, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28012, {{38826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28013, {{38827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28014, {{38828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28015, {{38829, 1, 0, 0},{38829, 1, 1, 18}, {0, 0, 0, 0}}},{28015, {{38829, 1, 0, 0},{38829, 1, 1, 18}, {0, 0, 0, 0}}},{28016, {{38829, 1, 0, 0},{38829, 1, 1, 19}, {0, 0, 0, 0}}},{28016, {{38829, 1, 0, 0},{38829, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{28017, {{38830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{28018, {{38832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28019, {{38833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28020, {{38834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28021, {{38835, 1, 0, 0},{38835, 1, 1, 17}, {0, 0, 0, 0}}},{28021, {{38835, 1, 0, 0},{38835, 1, 1, 17}, {0, 0, 0, 0}}},{28022, {{38835, 1, 0, 0},{38835, 1, 1, 18}, {0, 0, 0, 0}}},{28022, {{38835, 1, 0, 0},{38835, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28023, {{38836, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28024, {{38837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28025, {{38838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{28026, {{38840, 1, 0, 0},{38840, 1, 1, 18}, {0, 0, 0, 0}}},{28026, {{38840, 1, 0, 0},{38840, 1, 1, 18}, {0, 0, 0, 0}}},{28027, {{38840, 1, 0, 0},{38840, 1, 1, 17}, {0, 0, 0, 0}}},{28027, {{38840, 1, 0, 0},{38840, 1, 1, 17}, {0, 0, 0, 0}}},{68070, {{38840, 1, 0, 0},{38840, 1, 1, 19}, {0, 0, 0, 0}}},{68070, {{38840, 1, 0, 0},{38840, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{28028, {{38841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28029, {{38842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28030, {{38843, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28031, {{38844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28032, {{38845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28033, {{38846, 1, 0, 0},{38846, 1, 1, 17}, {0, 0, 0, 0}}},{28033, {{38846, 1, 0, 0},{38846, 1, 1, 17}, {0, 0, 0, 0}}},{28034, {{38846, 1, 0, 0},{38846, 1, 1, 18}, {0, 0, 0, 0}}},{28034, {{38846, 1, 0, 0},{38846, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{28035, {{38847, 1, 0, 0},{38847, 1, 1, 17}, {0, 0, 0, 0}}},{28035, {{38847, 1, 0, 0},{38847, 1, 1, 17}, {0, 0, 0, 0}}},{28036, {{38847, 1, 0, 0},{38847, 1, 1, 18}, {0, 0, 0, 0}}},{28036, {{38847, 1, 0, 0},{38847, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28037, {{38848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28038, {{38849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28039, {{38850, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28040, {{38851, 1, 0, 0},{38851, 1, 1, 18}, {0, 0, 0, 0}}},{28040, {{38851, 1, 0, 0},{38851, 1, 1, 18}, {0, 0, 0, 0}}},{28041, {{38851, 1, 0, 0},{38851, 1, 1, 17}, {0, 0, 0, 0}}},{28041, {{38851, 1, 0, 0},{38851, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{28042, {{38852, 1, 0, 0},{38852, 1, 1, 17}, {0, 0, 0, 0}}},{28042, {{38852, 1, 0, 0},{38852, 1, 1, 17}, {0, 0, 0, 0}}},{28043, {{38852, 1, 0, 0},{38852, 1, 1, 18}, {0, 0, 0, 0}}},{28043, {{38852, 1, 0, 0},{38852, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28044, {{38853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {10, {{28045, {{38854, 1, 0, 0},{38854, 1, 1, 19}, {0, 0, 0, 0}}},{28045, {{38854, 1, 0, 0},{38854, 1, 1, 19}, {0, 0, 0, 0}}},{28046, {{38854, 1, 0, 0},{38854, 1, 1, 21}, {0, 0, 0, 0}}},{28046, {{38854, 1, 0, 0},{38854, 1, 1, 21}, {0, 0, 0, 0}}},{28047, {{38854, 1, 0, 0},{38854, 1, 1, 17}, {0, 0, 0, 0}}},{28047, {{38854, 1, 0, 0},{38854, 1, 1, 17}, {0, 0, 0, 0}}},{28048, {{38854, 1, 0, 0},{38854, 1, 1, 20}, {0, 0, 0, 0}}},{28048, {{38854, 1, 0, 0},{38854, 1, 1, 20}, {0, 0, 0, 0}}},{28049, {{38854, 1, 0, 0},{38854, 1, 1, 18}, {0, 0, 0, 0}}},{28049, {{38854, 1, 0, 0},{38854, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28050, {{38855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{28051, {{38856, 1, 0, 0},{38856, 1, 1, 17}, {0, 0, 0, 0}}},{28051, {{38856, 1, 0, 0},{38856, 1, 1, 17}, {0, 0, 0, 0}}},{28052, {{38856, 1, 0, 0},{38856, 1, 1, 19}, {0, 0, 0, 0}}},{28052, {{38856, 1, 0, 0},{38856, 1, 1, 19}, {0, 0, 0, 0}}},{28053, {{38856, 1, 0, 0},{38856, 1, 1, 18}, {0, 0, 0, 0}}},{28053, {{38856, 1, 0, 0},{38856, 1, 1, 18}, {0, 0, 0, 0}}},{28054, {{38856, 1, 0, 0},{38856, 1, 1, 20}, {0, 0, 0, 0}}},{28054, {{38856, 1, 0, 0},{38856, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{28055, {{38857, 1, 0, 0},{38857, 1, 1, 17}, {0, 0, 0, 0}}},{28055, {{38857, 1, 0, 0},{38857, 1, 1, 17}, {0, 0, 0, 0}}},{28056, {{38857, 1, 0, 0},{38857, 1, 1, 18}, {0, 0, 0, 0}}},{28056, {{38857, 1, 0, 0},{38857, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{28057, {{38858, 1, 0, 0},{38858, 1, 1, 17}, {0, 0, 0, 0}}},{28057, {{38858, 1, 0, 0},{38858, 1, 1, 17}, {0, 0, 0, 0}}},{28058, {{38858, 1, 0, 0},{38858, 1, 1, 18}, {0, 0, 0, 0}}},{28058, {{38858, 1, 0, 0},{38858, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{28059, {{38859, 1, 0, 0},{38859, 1, 1, 18}, {0, 0, 0, 0}}},{28059, {{38859, 1, 0, 0},{38859, 1, 1, 18}, {0, 0, 0, 0}}},{28060, {{38859, 1, 0, 0},{38859, 1, 1, 17}, {0, 0, 0, 0}}},{28060, {{38859, 1, 0, 0},{38859, 1, 1, 17}, {0, 0, 0, 0}}},{28061, {{38859, 1, 0, 0},{38859, 1, 1, 19}, {0, 0, 0, 0}}},{28061, {{38859, 1, 0, 0},{38859, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {6, {{28062, {{38860, 1, 0, 0},{38860, 1, 1, 18}, {0, 0, 0, 0}}},{28062, {{38860, 1, 0, 0},{38860, 1, 1, 18}, {0, 0, 0, 0}}},{28063, {{38860, 1, 0, 0},{38860, 1, 1, 19}, {0, 0, 0, 0}}},{28063, {{38860, 1, 0, 0},{38860, 1, 1, 19}, {0, 0, 0, 0}}},{28064, {{38860, 1, 0, 0},{38860, 1, 1, 17}, {0, 0, 0, 0}}},{28064, {{38860, 1, 0, 0},{38860, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{28066, {{38861, 1, 0, 0},{38861, 1, 1, 18}, {0, 0, 0, 0}}},{28066, {{38861, 1, 0, 0},{38861, 1, 1, 18}, {0, 0, 0, 0}}},{28067, {{38861, 1, 0, 0},{38861, 1, 1, 17}, {0, 0, 0, 0}}},{28067, {{38861, 1, 0, 0},{38861, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{28068, {{38862, 1, 0, 0},{38862, 1, 1, 18}, {0, 0, 0, 0}}},{28068, {{38862, 1, 0, 0},{38862, 1, 1, 18}, {0, 0, 0, 0}}},{28069, {{38862, 1, 0, 0},{38862, 1, 1, 17}, {0, 0, 0, 0}}},{28069, {{38862, 1, 0, 0},{38862, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{28070, {{38863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28071, {{38864, 1, 0, 0},{38864, 1, 1, 18}, {0, 0, 0, 0}}},{28071, {{38864, 1, 0, 0},{38864, 1, 1, 18}, {0, 0, 0, 0}}},{28072, {{38864, 1, 0, 0},{38864, 1, 1, 17}, {0, 0, 0, 0}}},{28072, {{38864, 1, 0, 0},{38864, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {6, {{28073, {{38865, 1, 0, 0},{38865, 1, 1, 18}, {0, 0, 0, 0}}},{28073, {{38865, 1, 0, 0},{38865, 1, 1, 18}, {0, 0, 0, 0}}},{28074, {{38865, 1, 0, 0},{38865, 1, 1, 17}, {0, 0, 0, 0}}},{28074, {{38865, 1, 0, 0},{38865, 1, 1, 17}, {0, 0, 0, 0}}},{28075, {{38865, 1, 0, 0},{38865, 1, 1, 19}, {0, 0, 0, 0}}},{28075, {{38865, 1, 0, 0},{38865, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{28076, {{38866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{28077, {{38867, 1, 0, 0},{38867, 1, 1, 19}, {0, 0, 0, 0}}},{28077, {{38867, 1, 0, 0},{38867, 1, 1, 19}, {0, 0, 0, 0}}},{28078, {{38867, 1, 0, 0},{38867, 1, 1, 18}, {0, 0, 0, 0}}},{28078, {{38867, 1, 0, 0},{38867, 1, 1, 18}, {0, 0, 0, 0}}},{28079, {{38867, 1, 0, 0},{38867, 1, 1, 21}, {0, 0, 0, 0}}},{28079, {{38867, 1, 0, 0},{38867, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {4, {{28080, {{38868, 1, 0, 0},{38868, 1, 1, 18}, {0, 0, 0, 0}}},{28080, {{38868, 1, 0, 0},{38868, 1, 1, 18}, {0, 0, 0, 0}}},{28081, {{38868, 1, 0, 0},{38868, 1, 1, 17}, {0, 0, 0, 0}}},{28081, {{38868, 1, 0, 0},{38868, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{28082, {{38869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28083, {{38870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28084, {{38871, 1, 0, 0},{38871, 1, 1, 18}, {0, 0, 0, 0}}},{28084, {{38871, 1, 0, 0},{38871, 1, 1, 18}, {0, 0, 0, 0}}},{28085, {{38871, 1, 0, 0},{38871, 1, 1, 17}, {0, 0, 0, 0}}},{28085, {{38871, 1, 0, 0},{38871, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{28086, {{38872, 1, 0, 0},{38872, 1, 1, 18}, {0, 0, 0, 0}}},{28086, {{38872, 1, 0, 0},{38872, 1, 1, 18}, {0, 0, 0, 0}}},{28087, {{38872, 1, 0, 0},{38872, 1, 1, 17}, {0, 0, 0, 0}}},{28087, {{38872, 1, 0, 0},{38872, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{28088, {{38873, 1, 0, 0},{38873, 1, 1, 18}, {0, 0, 0, 0}}},{28088, {{38873, 1, 0, 0},{38873, 1, 1, 18}, {0, 0, 0, 0}}},{28089, {{38873, 1, 0, 0},{38873, 1, 1, 17}, {0, 0, 0, 0}}},{28089, {{38873, 1, 0, 0},{38873, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{28090, {{38874, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28091, {{38875, 1, 0, 0},{38875, 1, 1, 18}, {0, 0, 0, 0}}},{28091, {{38875, 1, 0, 0},{38875, 1, 1, 18}, {0, 0, 0, 0}}},{28092, {{38875, 1, 0, 0},{38875, 1, 1, 17}, {0, 0, 0, 0}}},{28092, {{38875, 1, 0, 0},{38875, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {8, {{28093, {{38876, 1, 0, 0},{38876, 1, 1, 18}, {0, 0, 0, 0}}},{28093, {{38876, 1, 0, 0},{38876, 1, 1, 18}, {0, 0, 0, 0}}},{28094, {{38876, 1, 0, 0},{38876, 1, 1, 17}, {0, 0, 0, 0}}},{28094, {{38876, 1, 0, 0},{38876, 1, 1, 17}, {0, 0, 0, 0}}},{28095, {{38876, 1, 0, 0},{38876, 1, 1, 20}, {0, 0, 0, 0}}},{28095, {{38876, 1, 0, 0},{38876, 1, 1, 20}, {0, 0, 0, 0}}},{28096, {{38876, 1, 0, 0},{38876, 1, 1, 19}, {0, 0, 0, 0}}},{28096, {{38876, 1, 0, 0},{38876, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{28097, {{38877, 1, 0, 0},{38877, 1, 1, 18}, {0, 0, 0, 0}}},{28097, {{38877, 1, 0, 0},{38877, 1, 1, 18}, {0, 0, 0, 0}}},{28098, {{38877, 1, 0, 0},{38877, 1, 1, 17}, {0, 0, 0, 0}}},{28098, {{38877, 1, 0, 0},{38877, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {8, {{28099, {{38878, 1, 0, 0},{38878, 1, 1, 20}, {0, 0, 0, 0}}},{28099, {{38878, 1, 0, 0},{38878, 1, 1, 20}, {0, 0, 0, 0}}},{28100, {{38878, 1, 0, 0},{38878, 1, 1, 19}, {0, 0, 0, 0}}},{28100, {{38878, 1, 0, 0},{38878, 1, 1, 19}, {0, 0, 0, 0}}},{28101, {{38878, 1, 0, 0},{38878, 1, 1, 18}, {0, 0, 0, 0}}},{28101, {{38878, 1, 0, 0},{38878, 1, 1, 18}, {0, 0, 0, 0}}},{28102, {{38878, 1, 0, 0},{38878, 1, 1, 21}, {0, 0, 0, 0}}},{28102, {{38878, 1, 0, 0},{38878, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{28103, {{38880, 1, 0, 0},{38880, 1, 1, 18}, {0, 0, 0, 0}}},{28103, {{38880, 1, 0, 0},{38880, 1, 1, 18}, {0, 0, 0, 0}}},{28104, {{38880, 1, 0, 0},{38880, 1, 1, 17}, {0, 0, 0, 0}}},{28104, {{38880, 1, 0, 0},{38880, 1, 1, 17}, {0, 0, 0, 0}}},{68071, {{38880, 1, 0, 0},{38880, 1, 1, 19}, {0, 0, 0, 0}}},{68071, {{38880, 1, 0, 0},{38880, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{28105, {{38881, 1, 0, 0},{38881, 1, 1, 17}, {0, 0, 0, 0}}},{28105, {{38881, 1, 0, 0},{38881, 1, 1, 17}, {0, 0, 0, 0}}},{28106, {{38881, 1, 0, 0},{38881, 1, 1, 18}, {0, 0, 0, 0}}},{28106, {{38881, 1, 0, 0},{38881, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28108, {{38882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28109, {{38883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{28110, {{38884, 1, 0, 0},{38884, 1, 1, 18}, {0, 0, 0, 0}}},{28110, {{38884, 1, 0, 0},{38884, 1, 1, 18}, {0, 0, 0, 0}}},{28111, {{38884, 1, 0, 0},{38884, 1, 1, 17}, {0, 0, 0, 0}}},{28111, {{38884, 1, 0, 0},{38884, 1, 1, 17}, {0, 0, 0, 0}}},{28112, {{38884, 1, 0, 0},{38884, 1, 1, 19}, {0, 0, 0, 0}}},{28112, {{38884, 1, 0, 0},{38884, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{28113, {{38885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{28115, {{38893, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28116, {{38894, 1, 0, 0},{38894, 1, 1, 17}, {0, 0, 0, 0}}},{28116, {{38894, 1, 0, 0},{38894, 1, 1, 17}, {0, 0, 0, 0}}},{28117, {{38894, 1, 0, 0},{38894, 1, 1, 18}, {0, 0, 0, 0}}},{28117, {{38894, 1, 0, 0},{38894, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28118, {{38895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28119, {{38896, 1, 0, 0},{38896, 1, 1, 16}, {0, 0, 0, 0}}},{28119, {{38896, 1, 0, 0},{38896, 1, 1, 16}, {0, 0, 0, 0}}},{59020, {{38896, 1, 0, 0},{38896, 1, 1, 17}, {0, 0, 0, 0}}},{59020, {{38896, 1, 0, 0},{38896, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{28120, {{38897, 1, 0, 0},{38897, 1, 1, 17}, {0, 0, 0, 0}}},{28120, {{38897, 1, 0, 0},{38897, 1, 1, 17}, {0, 0, 0, 0}}},{59021, {{38897, 1, 0, 0},{38897, 1, 1, 18}, {0, 0, 0, 0}}},{59021, {{38897, 1, 0, 0},{38897, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28121, {{38898, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28122, {{38899, 1, 0, 0},{38899, 1, 1, 18}, {0, 0, 0, 0}}},{28122, {{38899, 1, 0, 0},{38899, 1, 1, 18}, {0, 0, 0, 0}}},{28123, {{38899, 1, 0, 0},{38899, 1, 1, 19}, {0, 0, 0, 0}}},{28123, {{38899, 1, 0, 0},{38899, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{28124, {{38900, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28125, {{38901, 1, 0, 0},{38901, 1, 1, 17}, {0, 0, 0, 0}}},{28125, {{38901, 1, 0, 0},{38901, 1, 1, 17}, {0, 0, 0, 0}}},{28126, {{38901, 1, 0, 0},{38901, 1, 1, 18}, {0, 0, 0, 0}}},{28126, {{38901, 1, 0, 0},{38901, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{28127, {{38902, 1, 0, 0},{38902, 1, 1, 17}, {0, 0, 0, 0}}},{28127, {{38902, 1, 0, 0},{38902, 1, 1, 17}, {0, 0, 0, 0}}},{28128, {{38902, 1, 0, 0},{38902, 1, 1, 18}, {0, 0, 0, 0}}},{28128, {{38902, 1, 0, 0},{38902, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28129, {{38903, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28130, {{38904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28131, {{38905, 1, 0, 0},{38905, 1, 1, 16}, {0, 0, 0, 0}}},{28131, {{38905, 1, 0, 0},{38905, 1, 1, 16}, {0, 0, 0, 0}}},{28132, {{38905, 1, 0, 0},{38905, 1, 1, 17}, {0, 0, 0, 0}}},{28132, {{38905, 1, 0, 0},{38905, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {6, {{28133, {{38906, 1, 0, 0},{38906, 1, 1, 17}, {0, 0, 0, 0}}},{28133, {{38906, 1, 0, 0},{38906, 1, 1, 17}, {0, 0, 0, 0}}},{28134, {{38906, 1, 0, 0},{38906, 1, 1, 18}, {0, 0, 0, 0}}},{28134, {{38906, 1, 0, 0},{38906, 1, 1, 18}, {0, 0, 0, 0}}},{28135, {{38906, 1, 0, 0},{38906, 1, 1, 19}, {0, 0, 0, 0}}},{28135, {{38906, 1, 0, 0},{38906, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {6, {{28137, {{38907, 1, 0, 0},{38907, 1, 1, 18}, {0, 0, 0, 0}}},{28137, {{38907, 1, 0, 0},{38907, 1, 1, 18}, {0, 0, 0, 0}}},{28138, {{38907, 1, 0, 0},{38907, 1, 1, 17}, {0, 0, 0, 0}}},{28138, {{38907, 1, 0, 0},{38907, 1, 1, 17}, {0, 0, 0, 0}}},{28139, {{38907, 1, 0, 0},{38907, 1, 1, 19}, {0, 0, 0, 0}}},{28139, {{38907, 1, 0, 0},{38907, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{28140, {{38908, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28141, {{38909, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28142, {{38910, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {13, {{28143, {{38911, 1, 0, 0},{38911, 1, 1, 21}, {0, 0, 0, 0}}},{28143, {{38911, 1, 0, 0},{38911, 1, 1, 21}, {0, 0, 0, 0}}},{28144, {{38911, 1, 0, 0},{38911, 1, 1, 23}, {0, 0, 0, 0}}},{28144, {{38911, 1, 0, 0},{38911, 1, 1, 23}, {0, 0, 0, 0}}},{28145, {{38911, 1, 0, 0},{38911, 1, 1, 22}, {0, 0, 0, 0}}},{28145, {{38911, 1, 0, 0},{38911, 1, 1, 22}, {0, 0, 0, 0}}},{30305, {{38911, 1, 0, 0},{38911, 1, 1, 0},{38911, 1, 1, 24},{64105, 1, 0, 0}, {0, 0, 0, 0}}},{30305, {{38911, 1, 0, 0},{38911, 1, 1, 0},{38911, 1, 1, 24},{64105, 1, 0, 0}, {0, 0, 0, 0}}},{30305, {{38911, 1, 0, 0},{38911, 1, 1, 0},{38911, 1, 1, 24},{64105, 1, 0, 0}, {0, 0, 0, 0}}},{30306, {{38911, 1, 0, 0},{38911, 1, 1, 25}, {0, 0, 0, 0}}},{30306, {{38911, 1, 0, 0},{38911, 1, 1, 25}, {0, 0, 0, 0}}},{68072, {{38911, 1, 0, 0},{38911, 1, 1, 26}, {0, 0, 0, 0}}},{68072, {{38911, 1, 0, 0},{38911, 1, 1, 26}, {0, 0, 0, 0}}}}},
    {6, {{28146, {{38912, 1, 0, 0},{38912, 1, 1, 17}, {0, 0, 0, 0}}},{28146, {{38912, 1, 0, 0},{38912, 1, 1, 17}, {0, 0, 0, 0}}},{28147, {{38912, 1, 0, 0},{38912, 1, 1, 18}, {0, 0, 0, 0}}},{28147, {{38912, 1, 0, 0},{38912, 1, 1, 18}, {0, 0, 0, 0}}},{68047, {{38912, 1, 0, 0},{38912, 1, 1, 19}, {0, 0, 0, 0}}},{68047, {{38912, 1, 0, 0},{38912, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{28149, {{38913, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28150, {{38914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28151, {{38915, 1, 0, 0},{38915, 1, 1, 17}, {0, 0, 0, 0}}},{28151, {{38915, 1, 0, 0},{38915, 1, 1, 17}, {0, 0, 0, 0}}},{28152, {{38915, 1, 0, 0},{38915, 1, 1, 18}, {0, 0, 0, 0}}},{28152, {{38915, 1, 0, 0},{38915, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{28153, {{38916, 1, 0, 0},{38916, 1, 1, 17}, {0, 0, 0, 0}}},{28153, {{38916, 1, 0, 0},{38916, 1, 1, 17}, {0, 0, 0, 0}}},{28154, {{38916, 1, 0, 0},{38916, 1, 1, 18}, {0, 0, 0, 0}}},{28154, {{38916, 1, 0, 0},{38916, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28155, {{38917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28156, {{38918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28157, {{38919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28158, {{38920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28159, {{38921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28160, {{38922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28161, {{38923, 1, 0, 0},{38923, 1, 1, 16}, {0, 0, 0, 0}}},{28161, {{38923, 1, 0, 0},{38923, 1, 1, 16}, {0, 0, 0, 0}}},{28162, {{38923, 1, 0, 0},{38923, 1, 1, 17}, {0, 0, 0, 0}}},{28162, {{38923, 1, 0, 0},{38923, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{28163, {{38924, 1, 0, 0},{38924, 1, 1, 18}, {0, 0, 0, 0}}},{28163, {{38924, 1, 0, 0},{38924, 1, 1, 18}, {0, 0, 0, 0}}},{28164, {{38924, 1, 0, 0},{38924, 1, 1, 19}, {0, 0, 0, 0}}},{28164, {{38924, 1, 0, 0},{38924, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{28165, {{38925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28166, {{38926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28167, {{38927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28168, {{38928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28169, {{38929, 1, 0, 0},{38929, 1, 1, 18}, {0, 0, 0, 0}}},{28169, {{38929, 1, 0, 0},{38929, 1, 1, 18}, {0, 0, 0, 0}}},{28170, {{38929, 1, 0, 0},{38929, 1, 1, 19}, {0, 0, 0, 0}}},{28170, {{38929, 1, 0, 0},{38929, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{28171, {{38930, 1, 0, 0},{38930, 1, 1, 18}, {0, 0, 0, 0}}},{28171, {{38930, 1, 0, 0},{38930, 1, 1, 18}, {0, 0, 0, 0}}},{28172, {{38930, 1, 0, 0},{38930, 1, 1, 19}, {0, 0, 0, 0}}},{28172, {{38930, 1, 0, 0},{38930, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{28173, {{38931, 1, 0, 0},{38931, 1, 1, 19}, {0, 0, 0, 0}}},{28173, {{38931, 1, 0, 0},{38931, 1, 1, 19}, {0, 0, 0, 0}}},{28174, {{38931, 1, 0, 0},{38931, 1, 1, 18}, {0, 0, 0, 0}}},{28174, {{38931, 1, 0, 0},{38931, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28176, {{38932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28177, {{38933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28178, {{38934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28180, {{38935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28181, {{38936, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28182, {{38937, 1, 0, 0},{38937, 1, 1, 17}, {0, 0, 0, 0}}},{28182, {{38937, 1, 0, 0},{38937, 1, 1, 17}, {0, 0, 0, 0}}},{28183, {{38937, 1, 0, 0},{38937, 1, 1, 18}, {0, 0, 0, 0}}},{28183, {{38937, 1, 0, 0},{38937, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28184, {{38938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28185, {{38939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28186, {{38940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28187, {{38941, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28188, {{38942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28189, {{38943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28190, {{38944, 1, 0, 0},{38944, 1, 1, 17}, {0, 0, 0, 0}}},{28190, {{38944, 1, 0, 0},{38944, 1, 1, 17}, {0, 0, 0, 0}}},{28191, {{38944, 1, 0, 0},{38944, 1, 1, 18}, {0, 0, 0, 0}}},{28191, {{38944, 1, 0, 0},{38944, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28192, {{38945, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28193, {{38946, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28194, {{38947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28195, {{38948, 1, 0, 0},{38948, 1, 1, 18}, {0, 0, 0, 0}}},{28195, {{38948, 1, 0, 0},{38948, 1, 1, 18}, {0, 0, 0, 0}}},{28196, {{38948, 1, 0, 0},{38948, 1, 1, 19}, {0, 0, 0, 0}}},{28196, {{38948, 1, 0, 0},{38948, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{28197, {{38949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28198, {{38950, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28199, {{38951, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28200, {{38952, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28201, {{38953, 1, 0, 0},{38953, 1, 1, 16}, {0, 0, 0, 0}}},{28201, {{38953, 1, 0, 0},{38953, 1, 1, 16}, {0, 0, 0, 0}}},{28202, {{38953, 1, 0, 0},{38953, 1, 1, 17}, {0, 0, 0, 0}}},{28202, {{38953, 1, 0, 0},{38953, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{28203, {{38954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28204, {{38955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28205, {{38956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28206, {{38957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28207, {{38958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28208, {{38959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28209, {{38960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28210, {{38961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{28211, {{38962, 1, 0, 0},{38962, 1, 1, 17}, {0, 0, 0, 0}}},{28211, {{38962, 1, 0, 0},{38962, 1, 1, 17}, {0, 0, 0, 0}}},{28212, {{38962, 1, 0, 0},{38962, 1, 1, 18}, {0, 0, 0, 0}}},{28212, {{38962, 1, 0, 0},{38962, 1, 1, 18}, {0, 0, 0, 0}}},{28213, {{38962, 1, 0, 0},{38962, 1, 1, 19}, {0, 0, 0, 0}}},{28213, {{38962, 1, 0, 0},{38962, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{28214, {{38963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28215, {{38964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28216, {{38965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28217, {{38966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28218, {{38967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28219, {{38968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28220, {{38969, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28221, {{38970, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{28222, {{38971, 1, 0, 0},{38971, 1, 1, 18}, {0, 0, 0, 0}}},{28222, {{38971, 1, 0, 0},{38971, 1, 1, 18}, {0, 0, 0, 0}}},{30307, {{38971, 1, 0, 0},{38971, 1, 1, 0},{38971, 1, 1, 19},{64106, 1, 0, 0}, {0, 0, 0, 0}}},{30307, {{38971, 1, 0, 0},{38971, 1, 1, 0},{38971, 1, 1, 19},{64106, 1, 0, 0}, {0, 0, 0, 0}}},{30307, {{38971, 1, 0, 0},{38971, 1, 1, 0},{38971, 1, 1, 19},{64106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28223, {{38972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28224, {{38973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28225, {{38974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28226, {{38975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28227, {{38976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28228, {{38977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28230, {{38978, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28231, {{38979, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28232, {{38980, 1, 0, 0},{38980, 1, 1, 17}, {0, 0, 0, 0}}},{28232, {{38980, 1, 0, 0},{38980, 1, 1, 17}, {0, 0, 0, 0}}},{28233, {{38980, 1, 0, 0},{38980, 1, 1, 18}, {0, 0, 0, 0}}},{28233, {{38980, 1, 0, 0},{38980, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28234, {{38981, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28235, {{38982, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28236, {{38983, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28237, {{38984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28239, {{38985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28240, {{38986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28241, {{38987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28242, {{38988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28243, {{38989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28244, {{38990, 1, 0, 0},{38990, 1, 1, 17}, {0, 0, 0, 0}}},{28244, {{38990, 1, 0, 0},{38990, 1, 1, 17}, {0, 0, 0, 0}}},{28245, {{38990, 1, 0, 0},{38990, 1, 1, 18}, {0, 0, 0, 0}}},{28245, {{38990, 1, 0, 0},{38990, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28246, {{38991, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28247, {{38992, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28248, {{38993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28249, {{38994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28250, {{38995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28251, {{38996, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28252, {{38997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28253, {{38998, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28254, {{38999, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28255, {{39000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28257, {{39001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28258, {{39002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28259, {{39003, 1, 0, 0},{39003, 1, 1, 18}, {0, 0, 0, 0}}},{28259, {{39003, 1, 0, 0},{39003, 1, 1, 18}, {0, 0, 0, 0}}},{28260, {{39003, 1, 0, 0},{39003, 1, 1, 17}, {0, 0, 0, 0}}},{28260, {{39003, 1, 0, 0},{39003, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{28261, {{39004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28262, {{39005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {7, {{28263, {{39006, 1, 0, 0},{39006, 1, 1, 18}, {0, 0, 0, 0}}},{28263, {{39006, 1, 0, 0},{39006, 1, 1, 18}, {0, 0, 0, 0}}},{30188, {{39006, 1, 0, 0},{39006, 1, 1, 0},{39006, 1, 1, 19},{63952, 1, 0, 0}, {0, 0, 0, 0}}},{30188, {{39006, 1, 0, 0},{39006, 1, 1, 0},{39006, 1, 1, 19},{63952, 1, 0, 0}, {0, 0, 0, 0}}},{30188, {{39006, 1, 0, 0},{39006, 1, 1, 0},{39006, 1, 1, 19},{63952, 1, 0, 0}, {0, 0, 0, 0}}},{59027, {{39006, 1, 0, 0},{39006, 1, 1, 20}, {0, 0, 0, 0}}},{59027, {{39006, 1, 0, 0},{39006, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{28264, {{39007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28265, {{39008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28266, {{39009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28267, {{39010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28268, {{39011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28269, {{39012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28270, {{39013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28271, {{39014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28272, {{39015, 1, 0, 0},{39015, 1, 1, 19}, {0, 0, 0, 0}}},{28272, {{39015, 1, 0, 0},{39015, 1, 1, 19}, {0, 0, 0, 0}}},{28273, {{39015, 1, 0, 0},{39015, 1, 1, 18}, {0, 0, 0, 0}}},{28273, {{39015, 1, 0, 0},{39015, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28274, {{39016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28275, {{39017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28276, {{39018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28277, {{39019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28278, {{39020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28279, {{39021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28280, {{39022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28281, {{39023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28282, {{39024, 1, 0, 0},{39024, 1, 1, 17}, {0, 0, 0, 0}}},{28282, {{39024, 1, 0, 0},{39024, 1, 1, 17}, {0, 0, 0, 0}}},{28283, {{39024, 1, 0, 0},{39024, 1, 1, 18}, {0, 0, 0, 0}}},{28283, {{39024, 1, 0, 0},{39024, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28284, {{39025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28285, {{39026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{28286, {{39027, 1, 0, 0},{39027, 1, 1, 18}, {0, 0, 0, 0}}},{28286, {{39027, 1, 0, 0},{39027, 1, 1, 18}, {0, 0, 0, 0}}},{28287, {{39027, 1, 0, 0},{39027, 1, 1, 19}, {0, 0, 0, 0}}},{28287, {{39027, 1, 0, 0},{39027, 1, 1, 19}, {0, 0, 0, 0}}},{28288, {{39027, 1, 0, 0},{39027, 1, 1, 20}, {0, 0, 0, 0}}},{28288, {{39027, 1, 0, 0},{39027, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{28289, {{39028, 1, 0, 0},{39028, 1, 1, 17}, {0, 0, 0, 0}}},{28289, {{39028, 1, 0, 0},{39028, 1, 1, 17}, {0, 0, 0, 0}}},{28290, {{39028, 1, 0, 0},{39028, 1, 1, 18}, {0, 0, 0, 0}}},{28290, {{39028, 1, 0, 0},{39028, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{28294, {{39080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28295, {{39081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28296, {{39082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28297, {{39083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28298, {{39084, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28299, {{39085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28300, {{39086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28301, {{39087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28302, {{39088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28303, {{39089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28304, {{39090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28305, {{39091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28306, {{39092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28307, {{39093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28308, {{39094, 1, 0, 0},{39094, 1, 1, 17}, {0, 0, 0, 0}}},{28308, {{39094, 1, 0, 0},{39094, 1, 1, 17}, {0, 0, 0, 0}}},{28309, {{39094, 1, 0, 0},{39094, 1, 1, 18}, {0, 0, 0, 0}}},{28309, {{39094, 1, 0, 0},{39094, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28310, {{39095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28311, {{39096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28312, {{39097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28313, {{39098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28314, {{39099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28315, {{39100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28316, {{39101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28317, {{39102, 1, 0, 0},{39102, 1, 1, 16}, {0, 0, 0, 0}}},{28317, {{39102, 1, 0, 0},{39102, 1, 1, 16}, {0, 0, 0, 0}}},{28318, {{39102, 1, 0, 0},{39102, 1, 1, 17}, {0, 0, 0, 0}}},{28318, {{39102, 1, 0, 0},{39102, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{28319, {{39103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28320, {{39104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28321, {{39105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28322, {{39106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28323, {{39107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28324, {{39108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28325, {{39109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28326, {{39110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28327, {{39111, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28328, {{39112, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28329, {{39113, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{28330, {{39115, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28331, {{39116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28332, {{39117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{28334, {{39131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28335, {{39132, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28336, {{39133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{28338, {{39135, 1, 0, 0},{39135, 1, 1, 18}, {0, 0, 0, 0}}},{28338, {{39135, 1, 0, 0},{39135, 1, 1, 18}, {0, 0, 0, 0}}},{28339, {{39135, 1, 0, 0},{39135, 1, 1, 19}, {0, 0, 0, 0}}},{28339, {{39135, 1, 0, 0},{39135, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{28340, {{39136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28341, {{39137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{28342, {{39138, 1, 0, 0},{39138, 1, 1, 20}, {0, 0, 0, 0}}},{28342, {{39138, 1, 0, 0},{39138, 1, 1, 20}, {0, 0, 0, 0}}},{28343, {{39138, 1, 0, 0},{39138, 1, 1, 18}, {0, 0, 0, 0}}},{28343, {{39138, 1, 0, 0},{39138, 1, 1, 18}, {0, 0, 0, 0}}},{28344, {{39138, 1, 0, 0},{39138, 1, 1, 19}, {0, 0, 0, 0}}},{28344, {{39138, 1, 0, 0},{39138, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{28345, {{39139, 1, 0, 0},{39139, 1, 1, 18}, {0, 0, 0, 0}}},{28345, {{39139, 1, 0, 0},{39139, 1, 1, 18}, {0, 0, 0, 0}}},{28346, {{39139, 1, 0, 0},{39139, 1, 1, 17}, {0, 0, 0, 0}}},{28346, {{39139, 1, 0, 0},{39139, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{28347, {{39140, 1, 0, 0},{39140, 1, 1, 16}, {0, 0, 0, 0}}},{28347, {{39140, 1, 0, 0},{39140, 1, 1, 16}, {0, 0, 0, 0}}},{28348, {{39140, 1, 0, 0},{39140, 1, 1, 17}, {0, 0, 0, 0}}},{28348, {{39140, 1, 0, 0},{39140, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{28349, {{39141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{28350, {{39142, 1, 0, 0},{39142, 1, 1, 18}, {0, 0, 0, 0}}},{28350, {{39142, 1, 0, 0},{39142, 1, 1, 18}, {0, 0, 0, 0}}},{28351, {{39142, 1, 0, 0},{39142, 1, 1, 17}, {0, 0, 0, 0}}},{28351, {{39142, 1, 0, 0},{39142, 1, 1, 17}, {0, 0, 0, 0}}},{28352, {{39142, 1, 0, 0},{39142, 1, 1, 19}, {0, 0, 0, 0}}},{28352, {{39142, 1, 0, 0},{39142, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{28353, {{39143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{28354, {{39145, 1, 0, 0},{39145, 1, 1, 18}, {0, 0, 0, 0}}},{28354, {{39145, 1, 0, 0},{39145, 1, 1, 18}, {0, 0, 0, 0}}},{28355, {{39145, 1, 0, 0},{39145, 1, 1, 17}, {0, 0, 0, 0}}},{28355, {{39145, 1, 0, 0},{39145, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{28356, {{39146, 1, 0, 0},{39146, 1, 1, 18}, {0, 0, 0, 0}}},{28356, {{39146, 1, 0, 0},{39146, 1, 1, 18}, {0, 0, 0, 0}}},{28357, {{39146, 1, 0, 0},{39146, 1, 1, 17}, {0, 0, 0, 0}}},{28357, {{39146, 1, 0, 0},{39146, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{28358, {{39147, 1, 0, 0},{39147, 1, 1, 19}, {0, 0, 0, 0}}},{28358, {{39147, 1, 0, 0},{39147, 1, 1, 19}, {0, 0, 0, 0}}},{28359, {{39147, 1, 0, 0},{39147, 1, 1, 18}, {0, 0, 0, 0}}},{28359, {{39147, 1, 0, 0},{39147, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28360, {{39148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{28361, {{39149, 1, 0, 0},{39149, 1, 1, 18}, {0, 0, 0, 0}}},{28361, {{39149, 1, 0, 0},{39149, 1, 1, 18}, {0, 0, 0, 0}}},{28362, {{39149, 1, 0, 0},{39149, 1, 1, 17}, {0, 0, 0, 0}}},{28362, {{39149, 1, 0, 0},{39149, 1, 1, 17}, {0, 0, 0, 0}}},{59041, {{39149, 1, 0, 0},{39149, 1, 1, 19}, {0, 0, 0, 0}}},{59041, {{39149, 1, 0, 0},{39149, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{28363, {{39150, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{28364, {{39151, 1, 0, 0},{39151, 1, 1, 19}, {0, 0, 0, 0}}},{28364, {{39151, 1, 0, 0},{39151, 1, 1, 19}, {0, 0, 0, 0}}},{30240, {{39151, 1, 0, 0},{39151, 1, 1, 0},{39151, 1, 1, 20},{64042, 1, 0, 0}, {0, 0, 0, 0}}},{30240, {{39151, 1, 0, 0},{39151, 1, 1, 0},{39151, 1, 1, 20},{64042, 1, 0, 0}, {0, 0, 0, 0}}},{30240, {{39151, 1, 0, 0},{39151, 1, 1, 0},{39151, 1, 1, 20},{64042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28365, {{39152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28366, {{39153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28367, {{39154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28368, {{39155, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28369, {{39156, 1, 0, 0},{39156, 1, 1, 18}, {0, 0, 0, 0}}},{28369, {{39156, 1, 0, 0},{39156, 1, 1, 18}, {0, 0, 0, 0}}},{28370, {{39156, 1, 0, 0},{39156, 1, 1, 19}, {0, 0, 0, 0}}},{28370, {{39156, 1, 0, 0},{39156, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{28371, {{39157, 1, 0, 0},{39157, 1, 1, 16}, {0, 0, 0, 0}}},{28371, {{39157, 1, 0, 0},{39157, 1, 1, 16}, {0, 0, 0, 0}}},{28372, {{39157, 1, 0, 0},{39157, 1, 1, 17}, {0, 0, 0, 0}}},{28372, {{39157, 1, 0, 0},{39157, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{28373, {{39158, 1, 0, 0},{39158, 1, 1, 18}, {0, 0, 0, 0}}},{28373, {{39158, 1, 0, 0},{39158, 1, 1, 18}, {0, 0, 0, 0}}},{28374, {{39158, 1, 0, 0},{39158, 1, 1, 17}, {0, 0, 0, 0}}},{28374, {{39158, 1, 0, 0},{39158, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{28375, {{39159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28376, {{39160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28377, {{39161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28378, {{39162, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28379, {{39163, 1, 0, 0},{39163, 1, 1, 16}, {0, 0, 0, 0}}},{28379, {{39163, 1, 0, 0},{39163, 1, 1, 16}, {0, 0, 0, 0}}},{28380, {{39163, 1, 0, 0},{39163, 1, 1, 17}, {0, 0, 0, 0}}},{28380, {{39163, 1, 0, 0},{39163, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {5, {{28381, {{39164, 1, 0, 0},{39164, 1, 1, 18}, {0, 0, 0, 0}}},{28381, {{39164, 1, 0, 0},{39164, 1, 1, 18}, {0, 0, 0, 0}}},{30241, {{39164, 1, 0, 0},{39164, 1, 1, 0},{39164, 1, 1, 19},{64043, 1, 0, 0}, {0, 0, 0, 0}}},{30241, {{39164, 1, 0, 0},{39164, 1, 1, 0},{39164, 1, 1, 19},{64043, 1, 0, 0}, {0, 0, 0, 0}}},{30241, {{39164, 1, 0, 0},{39164, 1, 1, 0},{39164, 1, 1, 19},{64043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{28382, {{39165, 1, 0, 0},{39165, 1, 1, 18}, {0, 0, 0, 0}}},{28382, {{39165, 1, 0, 0},{39165, 1, 1, 18}, {0, 0, 0, 0}}},{28383, {{39165, 1, 0, 0},{39165, 1, 1, 20}, {0, 0, 0, 0}}},{28383, {{39165, 1, 0, 0},{39165, 1, 1, 20}, {0, 0, 0, 0}}},{28384, {{39165, 1, 0, 0},{39165, 1, 1, 19}, {0, 0, 0, 0}}},{28384, {{39165, 1, 0, 0},{39165, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{28385, {{39166, 1, 0, 0},{39166, 1, 1, 18}, {0, 0, 0, 0}}},{28385, {{39166, 1, 0, 0},{39166, 1, 1, 18}, {0, 0, 0, 0}}},{28386, {{39166, 1, 0, 0},{39166, 1, 1, 19}, {0, 0, 0, 0}}},{28386, {{39166, 1, 0, 0},{39166, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{28387, {{39167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28388, {{39168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28389, {{39169, 1, 0, 0},{39169, 1, 1, 17}, {0, 0, 0, 0}}},{28389, {{39169, 1, 0, 0},{39169, 1, 1, 17}, {0, 0, 0, 0}}},{28390, {{39169, 1, 0, 0},{39169, 1, 1, 16}, {0, 0, 0, 0}}},{28390, {{39169, 1, 0, 0},{39169, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {4, {{28391, {{39170, 1, 0, 0},{39170, 1, 1, 18}, {0, 0, 0, 0}}},{28391, {{39170, 1, 0, 0},{39170, 1, 1, 18}, {0, 0, 0, 0}}},{28392, {{39170, 1, 0, 0},{39170, 1, 1, 17}, {0, 0, 0, 0}}},{28392, {{39170, 1, 0, 0},{39170, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {6, {{28393, {{39171, 1, 0, 0},{39171, 1, 1, 19}, {0, 0, 0, 0}}},{28393, {{39171, 1, 0, 0},{39171, 1, 1, 19}, {0, 0, 0, 0}}},{28394, {{39171, 1, 0, 0},{39171, 1, 1, 18}, {0, 0, 0, 0}}},{28394, {{39171, 1, 0, 0},{39171, 1, 1, 18}, {0, 0, 0, 0}}},{28395, {{39171, 1, 0, 0},{39171, 1, 1, 20}, {0, 0, 0, 0}}},{28395, {{39171, 1, 0, 0},{39171, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{28396, {{39172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28397, {{39173, 1, 0, 0},{39173, 1, 1, 18}, {0, 0, 0, 0}}},{28397, {{39173, 1, 0, 0},{39173, 1, 1, 18}, {0, 0, 0, 0}}},{28398, {{39173, 1, 0, 0},{39173, 1, 1, 19}, {0, 0, 0, 0}}},{28398, {{39173, 1, 0, 0},{39173, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{28399, {{39174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28400, {{39175, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28401, {{39176, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28403, {{39177, 1, 0, 0},{39177, 1, 1, 19}, {0, 0, 0, 0}}},{28403, {{39177, 1, 0, 0},{39177, 1, 1, 19}, {0, 0, 0, 0}}},{28404, {{39177, 1, 0, 0},{39177, 1, 1, 18}, {0, 0, 0, 0}}},{28404, {{39177, 1, 0, 0},{39177, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {9, {{28405, {{39178, 1, 0, 0},{39178, 1, 1, 18}, {0, 0, 0, 0}}},{28405, {{39178, 1, 0, 0},{39178, 1, 1, 18}, {0, 0, 0, 0}}},{28406, {{39178, 1, 0, 0},{39178, 1, 1, 20}, {0, 0, 0, 0}}},{28406, {{39178, 1, 0, 0},{39178, 1, 1, 20}, {0, 0, 0, 0}}},{59042, {{39178, 1, 0, 0},{39178, 1, 1, 19}, {0, 0, 0, 0}}},{59042, {{39178, 1, 0, 0},{39178, 1, 1, 19}, {0, 0, 0, 0}}},{68101, {{39178, 1, 0, 0},{39178, 1, 1, 21}, {0, 0, 0, 0}}},{68101, {{39178, 1, 0, 0},{39178, 1, 1, 21}, {0, 0, 0, 0}}},{59043, {{39178, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28407, {{39179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {12, {{28408, {{39180, 1, 0, 0},{39180, 1, 1, 24}, {0, 0, 0, 0}}},{28408, {{39180, 1, 0, 0},{39180, 1, 1, 24}, {0, 0, 0, 0}}},{28409, {{39180, 1, 0, 0},{39180, 1, 1, 20}, {0, 0, 0, 0}}},{28409, {{39180, 1, 0, 0},{39180, 1, 1, 20}, {0, 0, 0, 0}}},{28410, {{39180, 1, 0, 0},{39180, 1, 1, 22}, {0, 0, 0, 0}}},{28410, {{39180, 1, 0, 0},{39180, 1, 1, 22}, {0, 0, 0, 0}}},{28411, {{39180, 1, 0, 0},{39180, 1, 1, 21}, {0, 0, 0, 0}}},{28411, {{39180, 1, 0, 0},{39180, 1, 1, 21}, {0, 0, 0, 0}}},{28412, {{39180, 1, 0, 0},{39180, 1, 1, 25}, {0, 0, 0, 0}}},{28412, {{39180, 1, 0, 0},{39180, 1, 1, 25}, {0, 0, 0, 0}}},{28413, {{39180, 1, 0, 0},{39180, 1, 1, 23}, {0, 0, 0, 0}}},{28413, {{39180, 1, 0, 0},{39180, 1, 1, 23}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{28414, {{39184, 1, 0, 0},{39184, 1, 1, 18}, {0, 0, 0, 0}}},{28414, {{39184, 1, 0, 0},{39184, 1, 1, 18}, {0, 0, 0, 0}}},{28415, {{39184, 1, 0, 0},{39184, 1, 1, 19}, {0, 0, 0, 0}}},{28415, {{39184, 1, 0, 0},{39184, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{28416, {{39185, 1, 0, 0},{39185, 1, 1, 18}, {0, 0, 0, 0}}},{28416, {{39185, 1, 0, 0},{39185, 1, 1, 18}, {0, 0, 0, 0}}},{28417, {{39185, 1, 0, 0},{39185, 1, 1, 17}, {0, 0, 0, 0}}},{28417, {{39185, 1, 0, 0},{39185, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {6, {{28418, {{39186, 1, 0, 0},{39186, 1, 1, 19}, {0, 0, 0, 0}}},{28418, {{39186, 1, 0, 0},{39186, 1, 1, 19}, {0, 0, 0, 0}}},{28419, {{39186, 1, 0, 0},{39186, 1, 1, 17}, {0, 0, 0, 0}}},{28419, {{39186, 1, 0, 0},{39186, 1, 1, 17}, {0, 0, 0, 0}}},{28420, {{39186, 1, 0, 0},{39186, 1, 1, 18}, {0, 0, 0, 0}}},{28420, {{39186, 1, 0, 0},{39186, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{28421, {{39187, 1, 0, 0},{39187, 1, 1, 18}, {0, 0, 0, 0}}},{28421, {{39187, 1, 0, 0},{39187, 1, 1, 18}, {0, 0, 0, 0}}},{28422, {{39187, 1, 0, 0},{39187, 1, 1, 19}, {0, 0, 0, 0}}},{28422, {{39187, 1, 0, 0},{39187, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{28423, {{39188, 1, 0, 0},{39188, 1, 1, 18}, {0, 0, 0, 0}}},{28423, {{39188, 1, 0, 0},{39188, 1, 1, 18}, {0, 0, 0, 0}}},{28424, {{39188, 1, 0, 0},{39188, 1, 1, 17}, {0, 0, 0, 0}}},{28424, {{39188, 1, 0, 0},{39188, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {6, {{28425, {{39189, 1, 0, 0},{39189, 1, 1, 19}, {0, 0, 0, 0}}},{28425, {{39189, 1, 0, 0},{39189, 1, 1, 19}, {0, 0, 0, 0}}},{28426, {{39189, 1, 0, 0},{39189, 1, 1, 17}, {0, 0, 0, 0}}},{28426, {{39189, 1, 0, 0},{39189, 1, 1, 17}, {0, 0, 0, 0}}},{28427, {{39189, 1, 0, 0},{39189, 1, 1, 18}, {0, 0, 0, 0}}},{28427, {{39189, 1, 0, 0},{39189, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{28428, {{39190, 1, 0, 0},{39190, 1, 1, 18}, {0, 0, 0, 0}}},{28428, {{39190, 1, 0, 0},{39190, 1, 1, 18}, {0, 0, 0, 0}}},{28429, {{39190, 1, 0, 0},{39190, 1, 1, 17}, {0, 0, 0, 0}}},{28429, {{39190, 1, 0, 0},{39190, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{28430, {{39191, 1, 0, 0},{39191, 1, 1, 18}, {0, 0, 0, 0}}},{28430, {{39191, 1, 0, 0},{39191, 1, 1, 18}, {0, 0, 0, 0}}},{28431, {{39191, 1, 0, 0},{39191, 1, 1, 17}, {0, 0, 0, 0}}},{28431, {{39191, 1, 0, 0},{39191, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{28432, {{39192, 1, 0, 0},{39192, 1, 1, 18}, {0, 0, 0, 0}}},{28432, {{39192, 1, 0, 0},{39192, 1, 1, 18}, {0, 0, 0, 0}}},{28433, {{39192, 1, 0, 0},{39192, 1, 1, 17}, {0, 0, 0, 0}}},{28433, {{39192, 1, 0, 0},{39192, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{28434, {{39194, 1, 0, 0},{39194, 1, 1, 18}, {0, 0, 0, 0}}},{28434, {{39194, 1, 0, 0},{39194, 1, 1, 18}, {0, 0, 0, 0}}},{28435, {{39194, 1, 0, 0},{39194, 1, 1, 17}, {0, 0, 0, 0}}},{28435, {{39194, 1, 0, 0},{39194, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{28436, {{39195, 1, 0, 0},{39195, 1, 1, 18}, {0, 0, 0, 0}}},{28436, {{39195, 1, 0, 0},{39195, 1, 1, 18}, {0, 0, 0, 0}}},{28437, {{39195, 1, 0, 0},{39195, 1, 1, 17}, {0, 0, 0, 0}}},{28437, {{39195, 1, 0, 0},{39195, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{28438, {{39196, 1, 0, 0},{39196, 1, 1, 18}, {0, 0, 0, 0}}},{28438, {{39196, 1, 0, 0},{39196, 1, 1, 18}, {0, 0, 0, 0}}},{28439, {{39196, 1, 0, 0},{39196, 1, 1, 17}, {0, 0, 0, 0}}},{28439, {{39196, 1, 0, 0},{39196, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {6, {{28440, {{39197, 1, 0, 0},{39197, 1, 1, 18}, {0, 0, 0, 0}}},{28440, {{39197, 1, 0, 0},{39197, 1, 1, 18}, {0, 0, 0, 0}}},{28441, {{39197, 1, 0, 0},{39197, 1, 1, 17}, {0, 0, 0, 0}}},{28441, {{39197, 1, 0, 0},{39197, 1, 1, 17}, {0, 0, 0, 0}}},{28442, {{39197, 1, 0, 0},{39197, 1, 1, 19}, {0, 0, 0, 0}}},{28442, {{39197, 1, 0, 0},{39197, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{28443, {{39198, 1, 0, 0},{39198, 1, 1, 18}, {0, 0, 0, 0}}},{28443, {{39198, 1, 0, 0},{39198, 1, 1, 18}, {0, 0, 0, 0}}},{28444, {{39198, 1, 0, 0},{39198, 1, 1, 17}, {0, 0, 0, 0}}},{28444, {{39198, 1, 0, 0},{39198, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{28445, {{39199, 1, 0, 0},{39199, 1, 1, 18}, {0, 0, 0, 0}}},{28445, {{39199, 1, 0, 0},{39199, 1, 1, 18}, {0, 0, 0, 0}}},{28446, {{39199, 1, 0, 0},{39199, 1, 1, 17}, {0, 0, 0, 0}}},{28446, {{39199, 1, 0, 0},{39199, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{28447, {{39200, 1, 0, 0},{39200, 1, 1, 18}, {0, 0, 0, 0}}},{28447, {{39200, 1, 0, 0},{39200, 1, 1, 18}, {0, 0, 0, 0}}},{28448, {{39200, 1, 0, 0},{39200, 1, 1, 17}, {0, 0, 0, 0}}},{28448, {{39200, 1, 0, 0},{39200, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{28449, {{39201, 1, 0, 0},{39201, 1, 1, 19}, {0, 0, 0, 0}}},{28449, {{39201, 1, 0, 0},{39201, 1, 1, 19}, {0, 0, 0, 0}}},{28450, {{39201, 1, 0, 0},{39201, 1, 1, 18}, {0, 0, 0, 0}}},{28450, {{39201, 1, 0, 0},{39201, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{28452, {{39202, 1, 0, 0},{39202, 1, 1, 18}, {0, 0, 0, 0}}},{28452, {{39202, 1, 0, 0},{39202, 1, 1, 18}, {0, 0, 0, 0}}},{28453, {{39202, 1, 0, 0},{39202, 1, 1, 17}, {0, 0, 0, 0}}},{28453, {{39202, 1, 0, 0},{39202, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{28454, {{39203, 1, 0, 0},{39203, 1, 1, 16}, {0, 0, 0, 0}}},{28454, {{39203, 1, 0, 0},{39203, 1, 1, 16}, {0, 0, 0, 0}}},{28455, {{39203, 1, 0, 0},{39203, 1, 1, 17}, {0, 0, 0, 0}}},{28455, {{39203, 1, 0, 0},{39203, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{28456, {{39204, 1, 0, 0},{39204, 1, 1, 18}, {0, 0, 0, 0}}},{28456, {{39204, 1, 0, 0},{39204, 1, 1, 18}, {0, 0, 0, 0}}},{28457, {{39204, 1, 0, 0},{39204, 1, 1, 17}, {0, 0, 0, 0}}},{28457, {{39204, 1, 0, 0},{39204, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{28458, {{39205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28459, {{39206, 1, 0, 0},{39206, 1, 1, 18}, {0, 0, 0, 0}}},{28459, {{39206, 1, 0, 0},{39206, 1, 1, 18}, {0, 0, 0, 0}}},{28460, {{39206, 1, 0, 0},{39206, 1, 1, 17}, {0, 0, 0, 0}}},{28460, {{39206, 1, 0, 0},{39206, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {6, {{28461, {{39207, 1, 0, 0},{39207, 1, 1, 18}, {0, 0, 0, 0}}},{28461, {{39207, 1, 0, 0},{39207, 1, 1, 18}, {0, 0, 0, 0}}},{28462, {{39207, 1, 0, 0},{39207, 1, 1, 17}, {0, 0, 0, 0}}},{28462, {{39207, 1, 0, 0},{39207, 1, 1, 17}, {0, 0, 0, 0}}},{28463, {{39207, 1, 0, 0},{39207, 1, 1, 19}, {0, 0, 0, 0}}},{28463, {{39207, 1, 0, 0},{39207, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {5, {{28464, {{39208, 1, 0, 0},{39208, 1, 1, 18}, {0, 0, 0, 0}}},{28464, {{39208, 1, 0, 0},{39208, 1, 1, 18}, {0, 0, 0, 0}}},{30242, {{39208, 1, 0, 0},{39208, 1, 1, 0},{39208, 1, 1, 19},{64044, 1, 0, 0}, {0, 0, 0, 0}}},{30242, {{39208, 1, 0, 0},{39208, 1, 1, 0},{39208, 1, 1, 19},{64044, 1, 0, 0}, {0, 0, 0, 0}}},{30242, {{39208, 1, 0, 0},{39208, 1, 1, 0},{39208, 1, 1, 19},{64044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{28465, {{39209, 1, 0, 0},{39209, 1, 1, 16}, {0, 0, 0, 0}}},{28465, {{39209, 1, 0, 0},{39209, 1, 1, 16}, {0, 0, 0, 0}}},{28466, {{39209, 1, 0, 0},{39209, 1, 1, 17}, {0, 0, 0, 0}}},{28466, {{39209, 1, 0, 0},{39209, 1, 1, 17}, {0, 0, 0, 0}}},{28467, {{39209, 1, 0, 0},{39209, 1, 1, 18}, {0, 0, 0, 0}}},{28467, {{39209, 1, 0, 0},{39209, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{28468, {{39210, 1, 0, 0},{39210, 1, 1, 16}, {0, 0, 0, 0}}},{28468, {{39210, 1, 0, 0},{39210, 1, 1, 16}, {0, 0, 0, 0}}},{28469, {{39210, 1, 0, 0},{39210, 1, 1, 17}, {0, 0, 0, 0}}},{28469, {{39210, 1, 0, 0},{39210, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{28470, {{39211, 1, 0, 0},{39211, 1, 1, 18}, {0, 0, 0, 0}}},{28470, {{39211, 1, 0, 0},{39211, 1, 1, 18}, {0, 0, 0, 0}}},{28471, {{39211, 1, 0, 0},{39211, 1, 1, 17}, {0, 0, 0, 0}}},{28471, {{39211, 1, 0, 0},{39211, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{28472, {{39212, 1, 0, 0},{39212, 1, 1, 18}, {0, 0, 0, 0}}},{28472, {{39212, 1, 0, 0},{39212, 1, 1, 18}, {0, 0, 0, 0}}},{28473, {{39212, 1, 0, 0},{39212, 1, 1, 17}, {0, 0, 0, 0}}},{28473, {{39212, 1, 0, 0},{39212, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{28474, {{39213, 1, 0, 0},{39213, 1, 1, 16}, {0, 0, 0, 0}}},{28474, {{39213, 1, 0, 0},{39213, 1, 1, 16}, {0, 0, 0, 0}}},{28475, {{39213, 1, 0, 0},{39213, 1, 1, 17}, {0, 0, 0, 0}}},{28475, {{39213, 1, 0, 0},{39213, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{28476, {{39214, 1, 0, 0},{39214, 1, 1, 17}, {0, 0, 0, 0}}},{28476, {{39214, 1, 0, 0},{39214, 1, 1, 17}, {0, 0, 0, 0}}},{28477, {{39214, 1, 0, 0},{39214, 1, 1, 18}, {0, 0, 0, 0}}},{28477, {{39214, 1, 0, 0},{39214, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28478, {{39215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28479, {{39216, 1, 0, 0},{39216, 1, 1, 16}, {0, 0, 0, 0}}},{28479, {{39216, 1, 0, 0},{39216, 1, 1, 16}, {0, 0, 0, 0}}},{28480, {{39216, 1, 0, 0},{39216, 1, 1, 17}, {0, 0, 0, 0}}},{28480, {{39216, 1, 0, 0},{39216, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{28481, {{39217, 1, 0, 0},{39217, 1, 1, 18}, {0, 0, 0, 0}}},{28481, {{39217, 1, 0, 0},{39217, 1, 1, 18}, {0, 0, 0, 0}}},{28482, {{39217, 1, 0, 0},{39217, 1, 1, 17}, {0, 0, 0, 0}}},{28482, {{39217, 1, 0, 0},{39217, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{28483, {{39218, 1, 0, 0},{39218, 1, 1, 18}, {0, 0, 0, 0}}},{28483, {{39218, 1, 0, 0},{39218, 1, 1, 18}, {0, 0, 0, 0}}},{28484, {{39218, 1, 0, 0},{39218, 1, 1, 17}, {0, 0, 0, 0}}},{28484, {{39218, 1, 0, 0},{39218, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{28485, {{39219, 1, 0, 0},{39219, 1, 1, 18}, {0, 0, 0, 0}}},{28485, {{39219, 1, 0, 0},{39219, 1, 1, 18}, {0, 0, 0, 0}}},{28486, {{39219, 1, 0, 0},{39219, 1, 1, 17}, {0, 0, 0, 0}}},{28486, {{39219, 1, 0, 0},{39219, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {6, {{28487, {{39220, 1, 0, 0},{39220, 1, 1, 17}, {0, 0, 0, 0}}},{28487, {{39220, 1, 0, 0},{39220, 1, 1, 17}, {0, 0, 0, 0}}},{28488, {{39220, 1, 0, 0},{39220, 1, 1, 18}, {0, 0, 0, 0}}},{28488, {{39220, 1, 0, 0},{39220, 1, 1, 18}, {0, 0, 0, 0}}},{28489, {{39220, 1, 0, 0},{39220, 1, 1, 19}, {0, 0, 0, 0}}},{28489, {{39220, 1, 0, 0},{39220, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{28490, {{39221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28491, {{39222, 1, 0, 0},{39222, 1, 1, 16}, {0, 0, 0, 0}}},{28491, {{39222, 1, 0, 0},{39222, 1, 1, 16}, {0, 0, 0, 0}}},{28492, {{39222, 1, 0, 0},{39222, 1, 1, 17}, {0, 0, 0, 0}}},{28492, {{39222, 1, 0, 0},{39222, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{28493, {{39225, 1, 0, 0},{39225, 1, 1, 19}, {0, 0, 0, 0}}},{28493, {{39225, 1, 0, 0},{39225, 1, 1, 19}, {0, 0, 0, 0}}},{28494, {{39225, 1, 0, 0},{39225, 1, 1, 17}, {0, 0, 0, 0}}},{28494, {{39225, 1, 0, 0},{39225, 1, 1, 17}, {0, 0, 0, 0}}},{28495, {{39225, 1, 0, 0},{39225, 1, 1, 18}, {0, 0, 0, 0}}},{28495, {{39225, 1, 0, 0},{39225, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{28496, {{39226, 1, 0, 0},{39226, 1, 1, 18}, {0, 0, 0, 0}}},{28496, {{39226, 1, 0, 0},{39226, 1, 1, 18}, {0, 0, 0, 0}}},{28497, {{39226, 1, 0, 0},{39226, 1, 1, 17}, {0, 0, 0, 0}}},{28497, {{39226, 1, 0, 0},{39226, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{28498, {{39227, 1, 0, 0},{39227, 1, 1, 18}, {0, 0, 0, 0}}},{28498, {{39227, 1, 0, 0},{39227, 1, 1, 18}, {0, 0, 0, 0}}},{28499, {{39227, 1, 0, 0},{39227, 1, 1, 17}, {0, 0, 0, 0}}},{28499, {{39227, 1, 0, 0},{39227, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{28500, {{39228, 1, 0, 0},{39228, 1, 1, 18}, {0, 0, 0, 0}}},{28500, {{39228, 1, 0, 0},{39228, 1, 1, 18}, {0, 0, 0, 0}}},{28501, {{39228, 1, 0, 0},{39228, 1, 1, 17}, {0, 0, 0, 0}}},{28501, {{39228, 1, 0, 0},{39228, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{28502, {{39229, 1, 0, 0},{39229, 1, 1, 18}, {0, 0, 0, 0}}},{28502, {{39229, 1, 0, 0},{39229, 1, 1, 18}, {0, 0, 0, 0}}},{28503, {{39229, 1, 0, 0},{39229, 1, 1, 17}, {0, 0, 0, 0}}},{28503, {{39229, 1, 0, 0},{39229, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{28504, {{39230, 1, 0, 0},{39230, 1, 1, 18}, {0, 0, 0, 0}}},{28504, {{39230, 1, 0, 0},{39230, 1, 1, 18}, {0, 0, 0, 0}}},{28505, {{39230, 1, 0, 0},{39230, 1, 1, 17}, {0, 0, 0, 0}}},{28505, {{39230, 1, 0, 0},{39230, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{28506, {{39231, 1, 0, 0},{39231, 1, 1, 17}, {0, 0, 0, 0}}},{28506, {{39231, 1, 0, 0},{39231, 1, 1, 17}, {0, 0, 0, 0}}},{28507, {{39231, 1, 0, 0},{39231, 1, 1, 16}, {0, 0, 0, 0}}},{28507, {{39231, 1, 0, 0},{39231, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {4, {{28508, {{39232, 1, 0, 0},{39232, 1, 1, 18}, {0, 0, 0, 0}}},{28508, {{39232, 1, 0, 0},{39232, 1, 1, 18}, {0, 0, 0, 0}}},{28509, {{39232, 1, 0, 0},{39232, 1, 1, 17}, {0, 0, 0, 0}}},{28509, {{39232, 1, 0, 0},{39232, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{28514, {{39233, 1, 0, 0},{39233, 1, 1, 18}, {0, 0, 0, 0}}},{28514, {{39233, 1, 0, 0},{39233, 1, 1, 18}, {0, 0, 0, 0}}},{28515, {{39233, 1, 0, 0},{39233, 1, 1, 17}, {0, 0, 0, 0}}},{28515, {{39233, 1, 0, 0},{39233, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {8, {{28516, {{39234, 1, 0, 0},{39234, 1, 1, 18}, {0, 0, 0, 0}}},{28516, {{39234, 1, 0, 0},{39234, 1, 1, 18}, {0, 0, 0, 0}}},{28517, {{39234, 1, 0, 0},{39234, 1, 1, 19}, {0, 0, 0, 0}}},{28517, {{39234, 1, 0, 0},{39234, 1, 1, 19}, {0, 0, 0, 0}}},{28518, {{39234, 1, 0, 0},{39234, 1, 1, 20}, {0, 0, 0, 0}}},{28518, {{39234, 1, 0, 0},{39234, 1, 1, 20}, {0, 0, 0, 0}}},{28519, {{39234, 1, 0, 0},{39234, 1, 1, 17}, {0, 0, 0, 0}}},{28519, {{39234, 1, 0, 0},{39234, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{28520, {{39236, 1, 0, 0},{39236, 1, 1, 16}, {0, 0, 0, 0}}},{28520, {{39236, 1, 0, 0},{39236, 1, 1, 16}, {0, 0, 0, 0}}},{28521, {{39236, 1, 0, 0},{39236, 1, 1, 17}, {0, 0, 0, 0}}},{28521, {{39236, 1, 0, 0},{39236, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {6, {{28522, {{39237, 1, 0, 0},{39237, 1, 1, 21}, {0, 0, 0, 0}}},{28522, {{39237, 1, 0, 0},{39237, 1, 1, 21}, {0, 0, 0, 0}}},{28523, {{39237, 1, 0, 0},{39237, 1, 1, 20}, {0, 0, 0, 0}}},{28523, {{39237, 1, 0, 0},{39237, 1, 1, 20}, {0, 0, 0, 0}}},{28524, {{39237, 1, 0, 0},{39237, 1, 1, 19}, {0, 0, 0, 0}}},{28524, {{39237, 1, 0, 0},{39237, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {6, {{28526, {{39238, 1, 0, 0},{39238, 1, 1, 18}, {0, 0, 0, 0}}},{28526, {{39238, 1, 0, 0},{39238, 1, 1, 18}, {0, 0, 0, 0}}},{28527, {{39238, 1, 0, 0},{39238, 1, 1, 17}, {0, 0, 0, 0}}},{28527, {{39238, 1, 0, 0},{39238, 1, 1, 17}, {0, 0, 0, 0}}},{28528, {{39238, 1, 0, 0},{39238, 1, 1, 19}, {0, 0, 0, 0}}},{28528, {{39238, 1, 0, 0},{39238, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {6, {{28529, {{39239, 1, 0, 0},{39239, 1, 1, 18}, {0, 0, 0, 0}}},{28529, {{39239, 1, 0, 0},{39239, 1, 1, 18}, {0, 0, 0, 0}}},{28530, {{39239, 1, 0, 0},{39239, 1, 1, 17}, {0, 0, 0, 0}}},{28530, {{39239, 1, 0, 0},{39239, 1, 1, 17}, {0, 0, 0, 0}}},{28531, {{39239, 1, 0, 0},{39239, 1, 1, 19}, {0, 0, 0, 0}}},{28531, {{39239, 1, 0, 0},{39239, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{28532, {{39240, 1, 0, 0},{39240, 1, 1, 18}, {0, 0, 0, 0}}},{28532, {{39240, 1, 0, 0},{39240, 1, 1, 18}, {0, 0, 0, 0}}},{28533, {{39240, 1, 0, 0},{39240, 1, 1, 17}, {0, 0, 0, 0}}},{28533, {{39240, 1, 0, 0},{39240, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {6, {{28534, {{39241, 1, 0, 0},{39241, 1, 1, 18}, {0, 0, 0, 0}}},{28534, {{39241, 1, 0, 0},{39241, 1, 1, 18}, {0, 0, 0, 0}}},{28535, {{39241, 1, 0, 0},{39241, 1, 1, 17}, {0, 0, 0, 0}}},{28535, {{39241, 1, 0, 0},{39241, 1, 1, 17}, {0, 0, 0, 0}}},{28536, {{39241, 1, 0, 0},{39241, 1, 1, 19}, {0, 0, 0, 0}}},{28536, {{39241, 1, 0, 0},{39241, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{28537, {{39242, 1, 0, 0},{39242, 1, 1, 16}, {0, 0, 0, 0}}},{28537, {{39242, 1, 0, 0},{39242, 1, 1, 16}, {0, 0, 0, 0}}},{28538, {{39242, 1, 0, 0},{39242, 1, 1, 17}, {0, 0, 0, 0}}},{28538, {{39242, 1, 0, 0},{39242, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{28539, {{39243, 1, 0, 0},{39243, 1, 1, 19}, {0, 0, 0, 0}}},{28539, {{39243, 1, 0, 0},{39243, 1, 1, 19}, {0, 0, 0, 0}}},{28540, {{39243, 1, 0, 0},{39243, 1, 1, 18}, {0, 0, 0, 0}}},{28540, {{39243, 1, 0, 0},{39243, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{28541, {{39244, 1, 0, 0},{39244, 1, 1, 18}, {0, 0, 0, 0}}},{28541, {{39244, 1, 0, 0},{39244, 1, 1, 18}, {0, 0, 0, 0}}},{28542, {{39244, 1, 0, 0},{39244, 1, 1, 17}, {0, 0, 0, 0}}},{28542, {{39244, 1, 0, 0},{39244, 1, 1, 17}, {0, 0, 0, 0}}},{28543, {{39244, 1, 0, 0},{39244, 1, 1, 19}, {0, 0, 0, 0}}},{28543, {{39244, 1, 0, 0},{39244, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{28544, {{39245, 1, 0, 0},{39245, 1, 1, 18}, {0, 0, 0, 0}}},{28544, {{39245, 1, 0, 0},{39245, 1, 1, 18}, {0, 0, 0, 0}}},{28545, {{39245, 1, 0, 0},{39245, 1, 1, 17}, {0, 0, 0, 0}}},{28545, {{39245, 1, 0, 0},{39245, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{28546, {{39246, 1, 0, 0},{39246, 1, 1, 18}, {0, 0, 0, 0}}},{28546, {{39246, 1, 0, 0},{39246, 1, 1, 18}, {0, 0, 0, 0}}},{28547, {{39246, 1, 0, 0},{39246, 1, 1, 17}, {0, 0, 0, 0}}},{28547, {{39246, 1, 0, 0},{39246, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{28548, {{39247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28549, {{39248, 1, 0, 0},{39248, 1, 1, 18}, {0, 0, 0, 0}}},{28549, {{39248, 1, 0, 0},{39248, 1, 1, 18}, {0, 0, 0, 0}}},{28550, {{39248, 1, 0, 0},{39248, 1, 1, 17}, {0, 0, 0, 0}}},{28550, {{39248, 1, 0, 0},{39248, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {8, {{28551, {{39249, 1, 0, 0},{39249, 1, 1, 18}, {0, 0, 0, 0}}},{28551, {{39249, 1, 0, 0},{39249, 1, 1, 18}, {0, 0, 0, 0}}},{28552, {{39249, 1, 0, 0},{39249, 1, 1, 17}, {0, 0, 0, 0}}},{28552, {{39249, 1, 0, 0},{39249, 1, 1, 17}, {0, 0, 0, 0}}},{28553, {{39249, 1, 0, 0},{39249, 1, 1, 19}, {0, 0, 0, 0}}},{28553, {{39249, 1, 0, 0},{39249, 1, 1, 19}, {0, 0, 0, 0}}},{28554, {{39249, 1, 0, 0},{39249, 1, 1, 20}, {0, 0, 0, 0}}},{28554, {{39249, 1, 0, 0},{39249, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{28555, {{39250, 1, 0, 0},{39250, 1, 1, 18}, {0, 0, 0, 0}}},{28555, {{39250, 1, 0, 0},{39250, 1, 1, 18}, {0, 0, 0, 0}}},{28556, {{39250, 1, 0, 0},{39250, 1, 1, 17}, {0, 0, 0, 0}}},{28556, {{39250, 1, 0, 0},{39250, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{28558, {{39251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28559, {{39252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28560, {{39253, 1, 0, 0},{39253, 1, 1, 17}, {0, 0, 0, 0}}},{28560, {{39253, 1, 0, 0},{39253, 1, 1, 17}, {0, 0, 0, 0}}},{28561, {{39253, 1, 0, 0},{39253, 1, 1, 18}, {0, 0, 0, 0}}},{28561, {{39253, 1, 0, 0},{39253, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{28562, {{39254, 1, 0, 0},{39254, 1, 1, 16}, {0, 0, 0, 0}}},{28562, {{39254, 1, 0, 0},{39254, 1, 1, 16}, {0, 0, 0, 0}}},{28563, {{39254, 1, 0, 0},{39254, 1, 1, 17}, {0, 0, 0, 0}}},{28563, {{39254, 1, 0, 0},{39254, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {14, {{28564, {{39255, 1, 0, 0},{39255, 1, 1, 20}, {0, 0, 0, 0}}},{28564, {{39255, 1, 0, 0},{39255, 1, 1, 20}, {0, 0, 0, 0}}},{28565, {{39255, 1, 0, 0},{39255, 1, 1, 25}, {0, 0, 0, 0}}},{28565, {{39255, 1, 0, 0},{39255, 1, 1, 25}, {0, 0, 0, 0}}},{28566, {{39255, 1, 0, 0},{39255, 1, 1, 22}, {0, 0, 0, 0}}},{28566, {{39255, 1, 0, 0},{39255, 1, 1, 22}, {0, 0, 0, 0}}},{28567, {{39255, 1, 0, 0},{39255, 1, 1, 21}, {0, 0, 0, 0}}},{28567, {{39255, 1, 0, 0},{39255, 1, 1, 21}, {0, 0, 0, 0}}},{28568, {{39255, 1, 0, 0},{39255, 1, 1, 23}, {0, 0, 0, 0}}},{28568, {{39255, 1, 0, 0},{39255, 1, 1, 23}, {0, 0, 0, 0}}},{28569, {{39255, 1, 0, 0},{39255, 1, 1, 24}, {0, 0, 0, 0}}},{28569, {{39255, 1, 0, 0},{39255, 1, 1, 24}, {0, 0, 0, 0}}},{28570, {{39255, 1, 0, 0},{39255, 1, 1, 26}, {0, 0, 0, 0}}},{28570, {{39255, 1, 0, 0},{39255, 1, 1, 26}, {0, 0, 0, 0}}}}},
    {4, {{28571, {{39256, 1, 0, 0},{39256, 1, 1, 18}, {0, 0, 0, 0}}},{28571, {{39256, 1, 0, 0},{39256, 1, 1, 18}, {0, 0, 0, 0}}},{28572, {{39256, 1, 0, 0},{39256, 1, 1, 17}, {0, 0, 0, 0}}},{28572, {{39256, 1, 0, 0},{39256, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {6, {{28573, {{39257, 1, 0, 0},{39257, 1, 1, 18}, {0, 0, 0, 0}}},{28573, {{39257, 1, 0, 0},{39257, 1, 1, 18}, {0, 0, 0, 0}}},{28574, {{39257, 1, 0, 0},{39257, 1, 1, 17}, {0, 0, 0, 0}}},{28574, {{39257, 1, 0, 0},{39257, 1, 1, 17}, {0, 0, 0, 0}}},{28575, {{39257, 1, 0, 0},{39257, 1, 1, 19}, {0, 0, 0, 0}}},{28575, {{39257, 1, 0, 0},{39257, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{28576, {{39258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{28577, {{39259, 1, 0, 0},{39259, 1, 1, 17}, {0, 0, 0, 0}}},{28577, {{39259, 1, 0, 0},{39259, 1, 1, 17}, {0, 0, 0, 0}}},{28578, {{39259, 1, 0, 0},{39259, 1, 1, 19}, {0, 0, 0, 0}}},{28578, {{39259, 1, 0, 0},{39259, 1, 1, 19}, {0, 0, 0, 0}}},{28579, {{39259, 1, 0, 0},{39259, 1, 1, 20}, {0, 0, 0, 0}}},{28579, {{39259, 1, 0, 0},{39259, 1, 1, 20}, {0, 0, 0, 0}}},{28580, {{39259, 1, 0, 0},{39259, 1, 1, 18}, {0, 0, 0, 0}}},{28580, {{39259, 1, 0, 0},{39259, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{28581, {{39260, 1, 0, 0},{39260, 1, 1, 19}, {0, 0, 0, 0}}},{28581, {{39260, 1, 0, 0},{39260, 1, 1, 19}, {0, 0, 0, 0}}},{28582, {{39260, 1, 0, 0},{39260, 1, 1, 17}, {0, 0, 0, 0}}},{28582, {{39260, 1, 0, 0},{39260, 1, 1, 17}, {0, 0, 0, 0}}},{28583, {{39260, 1, 0, 0},{39260, 1, 1, 18}, {0, 0, 0, 0}}},{28583, {{39260, 1, 0, 0},{39260, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{28584, {{39261, 1, 0, 0},{39261, 1, 1, 16}, {0, 0, 0, 0}}},{28584, {{39261, 1, 0, 0},{39261, 1, 1, 16}, {0, 0, 0, 0}}},{28585, {{39261, 1, 0, 0},{39261, 1, 1, 17}, {0, 0, 0, 0}}},{28585, {{39261, 1, 0, 0},{39261, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{28586, {{39262, 1, 0, 0},{39262, 1, 1, 18}, {0, 0, 0, 0}}},{28586, {{39262, 1, 0, 0},{39262, 1, 1, 18}, {0, 0, 0, 0}}},{28587, {{39262, 1, 0, 0},{39262, 1, 1, 17}, {0, 0, 0, 0}}},{28587, {{39262, 1, 0, 0},{39262, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{28588, {{39263, 1, 0, 0},{39263, 1, 1, 17}, {0, 0, 0, 0}}},{28588, {{39263, 1, 0, 0},{39263, 1, 1, 17}, {0, 0, 0, 0}}},{28589, {{39263, 1, 0, 0},{39263, 1, 1, 18}, {0, 0, 0, 0}}},{28589, {{39263, 1, 0, 0},{39263, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{28590, {{39264, 1, 0, 0},{39264, 1, 1, 18}, {0, 0, 0, 0}}},{28590, {{39264, 1, 0, 0},{39264, 1, 1, 18}, {0, 0, 0, 0}}},{28591, {{39264, 1, 0, 0},{39264, 1, 1, 17}, {0, 0, 0, 0}}},{28591, {{39264, 1, 0, 0},{39264, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{28592, {{39265, 1, 0, 0},{39265, 1, 1, 16}, {0, 0, 0, 0}}},{28592, {{39265, 1, 0, 0},{39265, 1, 1, 16}, {0, 0, 0, 0}}},{28593, {{39265, 1, 0, 0},{39265, 1, 1, 17}, {0, 0, 0, 0}}},{28593, {{39265, 1, 0, 0},{39265, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{28594, {{39318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28595, {{39319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28597, {{39320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28598, {{39321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28599, {{39322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28600, {{39323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28601, {{39324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28602, {{39325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28603, {{39326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28604, {{39327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28605, {{39328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28606, {{39329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28607, {{39330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28608, {{39331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28609, {{39332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28611, {{39333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28612, {{39334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28613, {{39335, 1, 0, 0},{39335, 1, 1, 17}, {0, 0, 0, 0}}},{28613, {{39335, 1, 0, 0},{39335, 1, 1, 17}, {0, 0, 0, 0}}},{28614, {{39335, 1, 0, 0},{39335, 1, 1, 16}, {0, 0, 0, 0}}},{28614, {{39335, 1, 0, 0},{39335, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{28615, {{39336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28616, {{39337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28617, {{39338, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28618, {{39339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28619, {{39340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28620, {{39341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28621, {{39342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28622, {{39343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28623, {{39344, 1, 0, 0},{39344, 1, 1, 17}, {0, 0, 0, 0}}},{28623, {{39344, 1, 0, 0},{39344, 1, 1, 17}, {0, 0, 0, 0}}},{28624, {{39344, 1, 0, 0},{39344, 1, 1, 18}, {0, 0, 0, 0}}},{28624, {{39344, 1, 0, 0},{39344, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28625, {{39345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28626, {{39346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28627, {{39347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28628, {{39348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28629, {{39349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28630, {{39350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28631, {{39351, 1, 0, 0},{39351, 1, 1, 16}, {0, 0, 0, 0}}},{28631, {{39351, 1, 0, 0},{39351, 1, 1, 16}, {0, 0, 0, 0}}},{28632, {{39351, 1, 0, 0},{39351, 1, 1, 17}, {0, 0, 0, 0}}},{28632, {{39351, 1, 0, 0},{39351, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{28633, {{39352, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28634, {{39353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28635, {{39354, 1, 0, 0},{39354, 1, 1, 17}, {0, 0, 0, 0}}},{28635, {{39354, 1, 0, 0},{39354, 1, 1, 17}, {0, 0, 0, 0}}},{28636, {{39354, 1, 0, 0},{39354, 1, 1, 18}, {0, 0, 0, 0}}},{28636, {{39354, 1, 0, 0},{39354, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28637, {{39355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28638, {{39356, 1, 0, 0},{39356, 1, 1, 17}, {0, 0, 0, 0}}},{28638, {{39356, 1, 0, 0},{39356, 1, 1, 17}, {0, 0, 0, 0}}},{28639, {{39356, 1, 0, 0},{39356, 1, 1, 18}, {0, 0, 0, 0}}},{28639, {{39356, 1, 0, 0},{39356, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28640, {{39357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28641, {{39358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28642, {{39359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28643, {{39360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28644, {{39361, 1, 0, 0},{39361, 1, 1, 18}, {0, 0, 0, 0}}},{28644, {{39361, 1, 0, 0},{39361, 1, 1, 18}, {0, 0, 0, 0}}},{28645, {{39361, 1, 0, 0},{39361, 1, 1, 19}, {0, 0, 0, 0}}},{28645, {{39361, 1, 0, 0},{39361, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{28646, {{39362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28647, {{39363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28648, {{39364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28649, {{39365, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28650, {{39366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{28651, {{39368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28652, {{39369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28653, {{39370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28654, {{39371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28655, {{39372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28656, {{39373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28657, {{39374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28658, {{39375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28659, {{39376, 1, 0, 0},{39376, 1, 1, 18}, {0, 0, 0, 0}}},{28659, {{39376, 1, 0, 0},{39376, 1, 1, 18}, {0, 0, 0, 0}}},{28660, {{39376, 1, 0, 0},{39376, 1, 1, 19}, {0, 0, 0, 0}}},{28660, {{39376, 1, 0, 0},{39376, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{28661, {{39377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28662, {{39378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28663, {{39379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28664, {{39380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28665, {{39381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28666, {{39382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28667, {{39383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28668, {{39384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28669, {{39385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28670, {{39386, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28671, {{39387, 1, 0, 0},{39387, 1, 1, 17}, {0, 0, 0, 0}}},{28671, {{39387, 1, 0, 0},{39387, 1, 1, 17}, {0, 0, 0, 0}}},{28672, {{39387, 1, 0, 0},{39387, 1, 1, 18}, {0, 0, 0, 0}}},{28672, {{39387, 1, 0, 0},{39387, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28673, {{39388, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28674, {{39389, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28675, {{39390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28676, {{39391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28677, {{39392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28678, {{39393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28679, {{39394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28680, {{39395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28681, {{39396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28682, {{39397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28683, {{39398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28685, {{39399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28686, {{39400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28687, {{39401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28689, {{39402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28690, {{39403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28691, {{39404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28692, {{39405, 1, 0, 0},{39405, 1, 1, 17}, {0, 0, 0, 0}}},{28692, {{39405, 1, 0, 0},{39405, 1, 1, 17}, {0, 0, 0, 0}}},{28693, {{39405, 1, 0, 0},{39405, 1, 1, 18}, {0, 0, 0, 0}}},{28693, {{39405, 1, 0, 0},{39405, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{28694, {{39406, 1, 0, 0},{39406, 1, 1, 17}, {0, 0, 0, 0}}},{28694, {{39406, 1, 0, 0},{39406, 1, 1, 17}, {0, 0, 0, 0}}},{28695, {{39406, 1, 0, 0},{39406, 1, 1, 18}, {0, 0, 0, 0}}},{28695, {{39406, 1, 0, 0},{39406, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28696, {{39407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28697, {{39408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28698, {{39409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28699, {{39410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28700, {{39411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28701, {{39412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28702, {{39413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28703, {{39414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28704, {{39415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28705, {{39416, 1, 0, 0},{39416, 1, 1, 17}, {0, 0, 0, 0}}},{28705, {{39416, 1, 0, 0},{39416, 1, 1, 17}, {0, 0, 0, 0}}},{28706, {{39416, 1, 0, 0},{39416, 1, 1, 18}, {0, 0, 0, 0}}},{28706, {{39416, 1, 0, 0},{39416, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28707, {{39417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28708, {{39418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28709, {{39419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28710, {{39420, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28711, {{39421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28712, {{39422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28713, {{39423, 1, 0, 0},{39423, 1, 1, 18}, {0, 0, 0, 0}}},{28713, {{39423, 1, 0, 0},{39423, 1, 1, 18}, {0, 0, 0, 0}}},{28714, {{39423, 1, 0, 0},{39423, 1, 1, 17}, {0, 0, 0, 0}}},{28714, {{39423, 1, 0, 0},{39423, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{28715, {{39424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28716, {{39425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28717, {{39426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28718, {{39427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28719, {{39428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28720, {{39429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28721, {{39430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28722, {{39431, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28723, {{39432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28724, {{39433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28725, {{39434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28726, {{39435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28727, {{39436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28728, {{39437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28729, {{39438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28730, {{39439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28731, {{39440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28732, {{39441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28733, {{39442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28734, {{39443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28735, {{39444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28736, {{39445, 1, 0, 0},{39445, 1, 1, 16}, {0, 0, 0, 0}}},{28736, {{39445, 1, 0, 0},{39445, 1, 1, 16}, {0, 0, 0, 0}}},{28737, {{39445, 1, 0, 0},{39445, 1, 1, 17}, {0, 0, 0, 0}}},{28737, {{39445, 1, 0, 0},{39445, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{28738, {{39446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28739, {{39447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28740, {{39448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28741, {{39449, 1, 0, 0},{39449, 1, 1, 19}, {0, 0, 0, 0}}},{28741, {{39449, 1, 0, 0},{39449, 1, 1, 19}, {0, 0, 0, 0}}},{28742, {{39449, 1, 0, 0},{39449, 1, 1, 18}, {0, 0, 0, 0}}},{28742, {{39449, 1, 0, 0},{39449, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28743, {{39450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28745, {{39451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28746, {{39452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28747, {{39453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28748, {{39454, 1, 0, 0},{39454, 1, 1, 17}, {0, 0, 0, 0}}},{28748, {{39454, 1, 0, 0},{39454, 1, 1, 17}, {0, 0, 0, 0}}},{28749, {{39454, 1, 0, 0},{39454, 1, 1, 18}, {0, 0, 0, 0}}},{28749, {{39454, 1, 0, 0},{39454, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28750, {{39455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28751, {{39456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28752, {{39457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28753, {{39458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28754, {{39459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28755, {{39460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28756, {{39461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28757, {{39462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28758, {{39463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28759, {{39464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28760, {{39465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28761, {{39466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28762, {{39467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28763, {{39468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28764, {{39469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28766, {{39470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28767, {{39471, 1, 0, 0},{39471, 1, 1, 16}, {0, 0, 0, 0}}},{28767, {{39471, 1, 0, 0},{39471, 1, 1, 16}, {0, 0, 0, 0}}},{28768, {{39471, 1, 0, 0},{39471, 1, 1, 17}, {0, 0, 0, 0}}},{28768, {{39471, 1, 0, 0},{39471, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{28769, {{39472, 1, 0, 0},{39472, 1, 1, 19}, {0, 0, 0, 0}}},{28769, {{39472, 1, 0, 0},{39472, 1, 1, 19}, {0, 0, 0, 0}}},{28770, {{39472, 1, 0, 0},{39472, 1, 1, 18}, {0, 0, 0, 0}}},{28770, {{39472, 1, 0, 0},{39472, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28771, {{39473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28772, {{39474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28773, {{39475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28774, {{39476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28775, {{39477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28776, {{39478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28777, {{39479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28778, {{39480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28779, {{39481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28780, {{39482, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28781, {{39483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28782, {{39484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28783, {{39485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28784, {{39486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28785, {{39487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28786, {{39488, 1, 0, 0},{39488, 1, 1, 17}, {0, 0, 0, 0}}},{28786, {{39488, 1, 0, 0},{39488, 1, 1, 17}, {0, 0, 0, 0}}},{28787, {{39488, 1, 0, 0},{39488, 1, 1, 18}, {0, 0, 0, 0}}},{28787, {{39488, 1, 0, 0},{39488, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{28788, {{39489, 1, 0, 0},{39489, 1, 1, 17}, {0, 0, 0, 0}}},{28788, {{39489, 1, 0, 0},{39489, 1, 1, 17}, {0, 0, 0, 0}}},{28789, {{39489, 1, 0, 0},{39489, 1, 1, 18}, {0, 0, 0, 0}}},{28789, {{39489, 1, 0, 0},{39489, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28790, {{39490, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28791, {{39491, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28792, {{39492, 1, 0, 0},{39492, 1, 1, 17}, {0, 0, 0, 0}}},{28792, {{39492, 1, 0, 0},{39492, 1, 1, 17}, {0, 0, 0, 0}}},{28793, {{39492, 1, 0, 0},{39492, 1, 1, 18}, {0, 0, 0, 0}}},{28793, {{39492, 1, 0, 0},{39492, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28795, {{39493, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28796, {{39494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28797, {{39495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28798, {{39496, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28799, {{39497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{28800, {{39498, 1, 0, 0},{39498, 1, 1, 18}, {0, 0, 0, 0}}},{28800, {{39498, 1, 0, 0},{39498, 1, 1, 18}, {0, 0, 0, 0}}},{28801, {{39498, 1, 0, 0},{39498, 1, 1, 19}, {0, 0, 0, 0}}},{28801, {{39498, 1, 0, 0},{39498, 1, 1, 19}, {0, 0, 0, 0}}},{68073, {{39498, 1, 0, 0},{39498, 1, 1, 20}, {0, 0, 0, 0}}},{68073, {{39498, 1, 0, 0},{39498, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{28802, {{39499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28803, {{39500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28804, {{39501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28805, {{39502, 1, 0, 0},{39502, 1, 1, 17}, {0, 0, 0, 0}}},{28805, {{39502, 1, 0, 0},{39502, 1, 1, 17}, {0, 0, 0, 0}}},{28806, {{39502, 1, 0, 0},{39502, 1, 1, 18}, {0, 0, 0, 0}}},{28806, {{39502, 1, 0, 0},{39502, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28807, {{39503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28808, {{39504, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28809, {{39505, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28810, {{39506, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28811, {{39507, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28812, {{39508, 1, 0, 0},{39508, 1, 1, 17}, {0, 0, 0, 0}}},{28812, {{39508, 1, 0, 0},{39508, 1, 1, 17}, {0, 0, 0, 0}}},{28813, {{39508, 1, 0, 0},{39508, 1, 1, 18}, {0, 0, 0, 0}}},{28813, {{39508, 1, 0, 0},{39508, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28814, {{39509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{28815, {{39510, 1, 0, 0},{39510, 1, 1, 17}, {0, 0, 0, 0}}},{28815, {{39510, 1, 0, 0},{39510, 1, 1, 17}, {0, 0, 0, 0}}},{28816, {{39510, 1, 0, 0},{39510, 1, 1, 19}, {0, 0, 0, 0}}},{28816, {{39510, 1, 0, 0},{39510, 1, 1, 19}, {0, 0, 0, 0}}},{28817, {{39510, 1, 0, 0},{39510, 1, 1, 18}, {0, 0, 0, 0}}},{28817, {{39510, 1, 0, 0},{39510, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28818, {{39511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28819, {{39512, 1, 0, 0},{39512, 1, 1, 17}, {0, 0, 0, 0}}},{28819, {{39512, 1, 0, 0},{39512, 1, 1, 17}, {0, 0, 0, 0}}},{28820, {{39512, 1, 0, 0},{39512, 1, 1, 18}, {0, 0, 0, 0}}},{28820, {{39512, 1, 0, 0},{39512, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28822, {{39513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28823, {{39514, 1, 0, 0},{39514, 1, 1, 17}, {0, 0, 0, 0}}},{28823, {{39514, 1, 0, 0},{39514, 1, 1, 17}, {0, 0, 0, 0}}},{28824, {{39514, 1, 0, 0},{39514, 1, 1, 18}, {0, 0, 0, 0}}},{28824, {{39514, 1, 0, 0},{39514, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{28825, {{39515, 1, 0, 0},{39515, 1, 1, 17}, {0, 0, 0, 0}}},{28825, {{39515, 1, 0, 0},{39515, 1, 1, 17}, {0, 0, 0, 0}}},{28826, {{39515, 1, 0, 0},{39515, 1, 1, 18}, {0, 0, 0, 0}}},{28826, {{39515, 1, 0, 0},{39515, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28827, {{39516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28828, {{39517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28829, {{39518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28830, {{39519, 1, 0, 0},{39519, 1, 1, 18}, {0, 0, 0, 0}}},{28830, {{39519, 1, 0, 0},{39519, 1, 1, 18}, {0, 0, 0, 0}}},{28831, {{39519, 1, 0, 0},{39519, 1, 1, 19}, {0, 0, 0, 0}}},{28831, {{39519, 1, 0, 0},{39519, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{28832, {{39520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28833, {{39521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28834, {{39522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28835, {{39523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28836, {{39524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28837, {{39525, 1, 0, 0},{39525, 1, 1, 18}, {0, 0, 0, 0}}},{28837, {{39525, 1, 0, 0},{39525, 1, 1, 18}, {0, 0, 0, 0}}},{28838, {{39525, 1, 0, 0},{39525, 1, 1, 19}, {0, 0, 0, 0}}},{28838, {{39525, 1, 0, 0},{39525, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{28839, {{39526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28840, {{39527, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28841, {{39528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28842, {{39529, 1, 0, 0},{39529, 1, 1, 17}, {0, 0, 0, 0}}},{28842, {{39529, 1, 0, 0},{39529, 1, 1, 17}, {0, 0, 0, 0}}},{28843, {{39529, 1, 0, 0},{39529, 1, 1, 18}, {0, 0, 0, 0}}},{28843, {{39529, 1, 0, 0},{39529, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28844, {{39530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28845, {{39531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{28848, {{39592, 1, 0, 0},{39592, 1, 1, 17}, {0, 0, 0, 0}}},{28848, {{39592, 1, 0, 0},{39592, 1, 1, 17}, {0, 0, 0, 0}}},{59058, {{39592, 1, 0, 0},{39592, 1, 1, 18}, {0, 0, 0, 0}}},{59058, {{39592, 1, 0, 0},{39592, 1, 1, 18}, {0, 0, 0, 0}}},{59059, {{39592, 1, 0, 0},{39592, 1, 1, 19}, {0, 0, 0, 0}}},{59059, {{39592, 1, 0, 0},{39592, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{28849, {{39593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28850, {{39594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28851, {{39595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28852, {{39596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28853, {{39597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28854, {{39598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28855, {{39599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28856, {{39600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28857, {{39601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28858, {{39602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28859, {{39603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28860, {{39604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28861, {{39605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28862, {{39606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28863, {{39607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28864, {{39608, 1, 0, 0},{39608, 1, 1, 17}, {0, 0, 0, 0}}},{28864, {{39608, 1, 0, 0},{39608, 1, 1, 17}, {0, 0, 0, 0}}},{28865, {{39608, 1, 0, 0},{39608, 1, 1, 18}, {0, 0, 0, 0}}},{28865, {{39608, 1, 0, 0},{39608, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28866, {{39609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28867, {{39610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28868, {{39611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28869, {{39612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28870, {{39613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28871, {{39614, 1, 0, 0},{39614, 1, 1, 17}, {0, 0, 0, 0}}},{28871, {{39614, 1, 0, 0},{39614, 1, 1, 17}, {0, 0, 0, 0}}},{28872, {{39614, 1, 0, 0},{39614, 1, 1, 18}, {0, 0, 0, 0}}},{28872, {{39614, 1, 0, 0},{39614, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{28873, {{39615, 1, 0, 0},{39615, 1, 1, 18}, {0, 0, 0, 0}}},{28873, {{39615, 1, 0, 0},{39615, 1, 1, 18}, {0, 0, 0, 0}}},{28874, {{39615, 1, 0, 0},{39615, 1, 1, 17}, {0, 0, 0, 0}}},{28874, {{39615, 1, 0, 0},{39615, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{28875, {{39616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28876, {{39617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28877, {{39618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28878, {{39619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28879, {{39620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{28880, {{39622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28881, {{39623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28882, {{39624, 1, 0, 0},{39624, 1, 1, 17}, {0, 0, 0, 0}}},{28882, {{39624, 1, 0, 0},{39624, 1, 1, 17}, {0, 0, 0, 0}}},{28883, {{39624, 1, 0, 0},{39624, 1, 1, 18}, {0, 0, 0, 0}}},{28883, {{39624, 1, 0, 0},{39624, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28884, {{39625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28885, {{39626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{28886, {{39629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28887, {{39630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28888, {{39631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28889, {{39632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28890, {{39633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28891, {{39634, 1, 0, 0},{39634, 1, 1, 17}, {0, 0, 0, 0}}},{28891, {{39634, 1, 0, 0},{39634, 1, 1, 17}, {0, 0, 0, 0}}},{28892, {{39634, 1, 0, 0},{39634, 1, 1, 18}, {0, 0, 0, 0}}},{28892, {{39634, 1, 0, 0},{39634, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{28893, {{39635, 1, 0, 0},{39635, 1, 1, 17}, {0, 0, 0, 0}}},{28893, {{39635, 1, 0, 0},{39635, 1, 1, 17}, {0, 0, 0, 0}}},{28894, {{39635, 1, 0, 0},{39635, 1, 1, 18}, {0, 0, 0, 0}}},{28894, {{39635, 1, 0, 0},{39635, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28895, {{39636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28896, {{39637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{28897, {{39638, 1, 0, 0},{39638, 1, 1, 19}, {0, 0, 0, 0}}},{28897, {{39638, 1, 0, 0},{39638, 1, 1, 19}, {0, 0, 0, 0}}},{28898, {{39638, 1, 0, 0},{39638, 1, 1, 17}, {0, 0, 0, 0}}},{28898, {{39638, 1, 0, 0},{39638, 1, 1, 17}, {0, 0, 0, 0}}},{28899, {{39638, 1, 0, 0},{39638, 1, 1, 18}, {0, 0, 0, 0}}},{28899, {{39638, 1, 0, 0},{39638, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28900, {{39639, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28901, {{39640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28902, {{39641, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{28904, {{39643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28905, {{39644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28907, {{39645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28908, {{39646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28910, {{39647, 1, 0, 0},{39647, 1, 1, 17}, {0, 0, 0, 0}}},{28910, {{39647, 1, 0, 0},{39647, 1, 1, 17}, {0, 0, 0, 0}}},{60349, {{39647, 1, 0, 0},{39647, 1, 1, 18}, {0, 0, 0, 0}}},{60349, {{39647, 1, 0, 0},{39647, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28911, {{39648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28912, {{39649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28913, {{39650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28914, {{39651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28915, {{39652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28916, {{39653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28917, {{39654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28918, {{39655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28919, {{39656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28920, {{39657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28921, {{39658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28922, {{39659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28923, {{39660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28924, {{39661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28925, {{39662, 1, 0, 0},{39662, 1, 1, 17}, {0, 0, 0, 0}}},{28925, {{39662, 1, 0, 0},{39662, 1, 1, 17}, {0, 0, 0, 0}}},{59067, {{39662, 1, 0, 0},{39662, 1, 1, 18}, {0, 0, 0, 0}}},{59067, {{39662, 1, 0, 0},{39662, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{28926, {{39663, 1, 0, 0},{39663, 1, 1, 18}, {0, 0, 0, 0}}},{28926, {{39663, 1, 0, 0},{39663, 1, 1, 18}, {0, 0, 0, 0}}},{28927, {{39663, 1, 0, 0},{39663, 1, 1, 19}, {0, 0, 0, 0}}},{28927, {{39663, 1, 0, 0},{39663, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{28928, {{39664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28929, {{39665, 1, 0, 0},{39665, 1, 1, 17}, {0, 0, 0, 0}}},{28929, {{39665, 1, 0, 0},{39665, 1, 1, 17}, {0, 0, 0, 0}}},{28930, {{39665, 1, 0, 0},{39665, 1, 1, 18}, {0, 0, 0, 0}}},{28930, {{39665, 1, 0, 0},{39665, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28931, {{39666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28932, {{39667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28933, {{39668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28934, {{39669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28935, {{39670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28936, {{39671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28937, {{39672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28938, {{39673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28939, {{39674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28940, {{39675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28941, {{39676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28942, {{39677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28943, {{39678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28944, {{39679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28945, {{39680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28946, {{39681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28947, {{39682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28948, {{39683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28949, {{39684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28950, {{39685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28951, {{39686, 1, 0, 0},{39686, 1, 1, 17}, {0, 0, 0, 0}}},{28951, {{39686, 1, 0, 0},{39686, 1, 1, 17}, {0, 0, 0, 0}}},{28952, {{39686, 1, 0, 0},{39686, 1, 1, 18}, {0, 0, 0, 0}}},{28952, {{39686, 1, 0, 0},{39686, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28953, {{39687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28954, {{39688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28955, {{39689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28956, {{39690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28957, {{39691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28958, {{39692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28959, {{39693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28960, {{39694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{28961, {{39696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28962, {{39697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28963, {{39698, 1, 0, 0},{39698, 1, 1, 17}, {0, 0, 0, 0}}},{28963, {{39698, 1, 0, 0},{39698, 1, 1, 17}, {0, 0, 0, 0}}},{28964, {{39698, 1, 0, 0},{39698, 1, 1, 18}, {0, 0, 0, 0}}},{28964, {{39698, 1, 0, 0},{39698, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{28965, {{39700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28966, {{39701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28967, {{39702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28968, {{39703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28969, {{39704, 1, 0, 0},{39704, 1, 1, 19}, {0, 0, 0, 0}}},{28969, {{39704, 1, 0, 0},{39704, 1, 1, 19}, {0, 0, 0, 0}}},{28970, {{39704, 1, 0, 0},{39704, 1, 1, 18}, {0, 0, 0, 0}}},{28970, {{39704, 1, 0, 0},{39704, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28971, {{39705, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28972, {{39706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28973, {{39707, 1, 0, 0},{39707, 1, 1, 17}, {0, 0, 0, 0}}},{28973, {{39707, 1, 0, 0},{39707, 1, 1, 17}, {0, 0, 0, 0}}},{28974, {{39707, 1, 0, 0},{39707, 1, 1, 18}, {0, 0, 0, 0}}},{28974, {{39707, 1, 0, 0},{39707, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{28975, {{39708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{28976, {{39710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28977, {{39711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28978, {{39712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28979, {{39713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28980, {{39714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{28981, {{39715, 1, 0, 0},{39715, 1, 1, 18}, {0, 0, 0, 0}}},{28981, {{39715, 1, 0, 0},{39715, 1, 1, 18}, {0, 0, 0, 0}}},{28982, {{39715, 1, 0, 0},{39715, 1, 1, 17}, {0, 0, 0, 0}}},{28982, {{39715, 1, 0, 0},{39715, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{28983, {{39716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28984, {{39717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28985, {{39718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28986, {{39719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28987, {{39720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28988, {{39721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28989, {{39722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28990, {{39723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{28991, {{39724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{28993, {{39725, 1, 0, 0},{39725, 1, 1, 18}, {0, 0, 0, 0}}},{28993, {{39725, 1, 0, 0},{39725, 1, 1, 18}, {0, 0, 0, 0}}},{28994, {{39725, 1, 0, 0},{39725, 1, 1, 19}, {0, 0, 0, 0}}},{28994, {{39725, 1, 0, 0},{39725, 1, 1, 19}, {0, 0, 0, 0}}},{28995, {{39725, 1, 0, 0},{39725, 1, 1, 20}, {0, 0, 0, 0}}},{28995, {{39725, 1, 0, 0},{39725, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {10, {{28996, {{39726, 1, 0, 0},{39726, 1, 1, 18}, {0, 0, 0, 0}}},{28996, {{39726, 1, 0, 0},{39726, 1, 1, 18}, {0, 0, 0, 0}}},{28997, {{39726, 1, 0, 0},{39726, 1, 1, 21}, {0, 0, 0, 0}}},{28997, {{39726, 1, 0, 0},{39726, 1, 1, 21}, {0, 0, 0, 0}}},{28998, {{39726, 1, 0, 0},{39726, 1, 1, 20}, {0, 0, 0, 0}}},{28998, {{39726, 1, 0, 0},{39726, 1, 1, 20}, {0, 0, 0, 0}}},{28999, {{39726, 1, 0, 0},{39726, 1, 1, 19}, {0, 0, 0, 0}}},{28999, {{39726, 1, 0, 0},{39726, 1, 1, 19}, {0, 0, 0, 0}}},{29000, {{39726, 1, 0, 0},{39726, 1, 1, 22}, {0, 0, 0, 0}}},{29000, {{39726, 1, 0, 0},{39726, 1, 1, 22}, {0, 0, 0, 0}}}}},
    {1, {{29001, {{39727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29002, {{39728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29003, {{39729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{29004, {{39730, 1, 0, 0},{39730, 1, 1, 17}, {0, 0, 0, 0}}},{29004, {{39730, 1, 0, 0},{39730, 1, 1, 17}, {0, 0, 0, 0}}},{59076, {{39730, 1, 0, 0},{39730, 1, 1, 18}, {0, 0, 0, 0}}},{59076, {{39730, 1, 0, 0},{39730, 1, 1, 18}, {0, 0, 0, 0}}},{59077, {{39730, 1, 0, 0},{39730, 1, 1, 19}, {0, 0, 0, 0}}},{59077, {{39730, 1, 0, 0},{39730, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{29005, {{39731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29006, {{39732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29007, {{39733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{29008, {{39735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29009, {{39736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29010, {{39737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29011, {{39738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29012, {{39739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29013, {{39740, 1, 0, 0},{39740, 1, 1, 17}, {0, 0, 0, 0}}},{29013, {{39740, 1, 0, 0},{39740, 1, 1, 17}, {0, 0, 0, 0}}},{59079, {{39740, 1, 0, 0},{39740, 1, 1, 18}, {0, 0, 0, 0}}},{59079, {{39740, 1, 0, 0},{39740, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29014, {{39741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29015, {{39742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29016, {{39743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29017, {{39744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29018, {{39745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29019, {{39746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29020, {{39747, 1, 0, 0},{39747, 1, 1, 17}, {0, 0, 0, 0}}},{29020, {{39747, 1, 0, 0},{39747, 1, 1, 17}, {0, 0, 0, 0}}},{59083, {{39747, 1, 0, 0},{39747, 1, 1, 18}, {0, 0, 0, 0}}},{59083, {{39747, 1, 0, 0},{39747, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29021, {{39748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29022, {{39749, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29023, {{39750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{29024, {{39752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{29025, {{39754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29026, {{39755, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29028, {{39756, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29029, {{39757, 1, 0, 0},{39757, 1, 1, 18}, {0, 0, 0, 0}}},{29029, {{39757, 1, 0, 0},{39757, 1, 1, 18}, {0, 0, 0, 0}}},{29030, {{39757, 1, 0, 0},{39757, 1, 1, 19}, {0, 0, 0, 0}}},{29030, {{39757, 1, 0, 0},{39757, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{29031, {{39758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29032, {{39759, 1, 0, 0},{39759, 1, 1, 17}, {0, 0, 0, 0}}},{29032, {{39759, 1, 0, 0},{39759, 1, 1, 17}, {0, 0, 0, 0}}},{29033, {{39759, 1, 0, 0},{39759, 1, 1, 18}, {0, 0, 0, 0}}},{29033, {{39759, 1, 0, 0},{39759, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29034, {{39760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29035, {{39761, 1, 0, 0},{39761, 1, 1, 17}, {0, 0, 0, 0}}},{29035, {{39761, 1, 0, 0},{39761, 1, 1, 17}, {0, 0, 0, 0}}},{29036, {{39761, 1, 0, 0},{39761, 1, 1, 18}, {0, 0, 0, 0}}},{29036, {{39761, 1, 0, 0},{39761, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29037, {{39762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29038, {{39763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29039, {{39764, 1, 0, 0},{39764, 1, 1, 19}, {0, 0, 0, 0}}},{29039, {{39764, 1, 0, 0},{39764, 1, 1, 19}, {0, 0, 0, 0}}},{29040, {{39764, 1, 0, 0},{39764, 1, 1, 18}, {0, 0, 0, 0}}},{29040, {{39764, 1, 0, 0},{39764, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29041, {{39765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29042, {{39766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29043, {{39767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29044, {{39768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29045, {{39769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29046, {{39770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29047, {{39771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29048, {{39772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29049, {{39773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29050, {{39774, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29051, {{39775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29052, {{39776, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29053, {{39777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{29054, {{39779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29055, {{39780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29056, {{39781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29058, {{39782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29059, {{39783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29060, {{39784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29061, {{39785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29062, {{39786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29063, {{39787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29064, {{39788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29065, {{39789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29066, {{39790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29067, {{39791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29068, {{39792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29069, {{39793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29070, {{39794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29071, {{39795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29072, {{39796, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29073, {{39797, 1, 0, 0},{39797, 1, 1, 17}, {0, 0, 0, 0}}},{29073, {{39797, 1, 0, 0},{39797, 1, 1, 17}, {0, 0, 0, 0}}},{29074, {{39797, 1, 0, 0},{39797, 1, 1, 18}, {0, 0, 0, 0}}},{29074, {{39797, 1, 0, 0},{39797, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29076, {{39798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29077, {{39799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29078, {{39800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29079, {{39801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29081, {{39802, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29082, {{39803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29083, {{39804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29084, {{39805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29085, {{39806, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29086, {{39807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29087, {{39808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29088, {{39809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29089, {{39810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29090, {{39811, 1, 0, 0},{39811, 1, 1, 17}, {0, 0, 0, 0}}},{29090, {{39811, 1, 0, 0},{39811, 1, 1, 17}, {0, 0, 0, 0}}},{29091, {{39811, 1, 0, 0},{39811, 1, 1, 18}, {0, 0, 0, 0}}},{29091, {{39811, 1, 0, 0},{39811, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29092, {{39812, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29094, {{39813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29095, {{39814, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29096, {{39815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29097, {{39816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29098, {{39817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29099, {{39818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29100, {{39819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29101, {{39820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29102, {{39821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{29103, {{39822, 1, 0, 0},{39822, 1, 1, 18}, {0, 0, 0, 0}}},{29103, {{39822, 1, 0, 0},{39822, 1, 1, 18}, {0, 0, 0, 0}}},{29104, {{39822, 1, 0, 0},{39822, 1, 1, 17}, {0, 0, 0, 0}}},{29104, {{39822, 1, 0, 0},{39822, 1, 1, 17}, {0, 0, 0, 0}}},{59095, {{39822, 1, 0, 0},{39822, 1, 1, 19}, {0, 0, 0, 0}}},{59095, {{39822, 1, 0, 0},{39822, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{29105, {{39823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29107, {{39824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29108, {{39825, 1, 0, 0},{39825, 1, 1, 17}, {0, 0, 0, 0}}},{29108, {{39825, 1, 0, 0},{39825, 1, 1, 17}, {0, 0, 0, 0}}},{29109, {{39825, 1, 0, 0},{39825, 1, 1, 18}, {0, 0, 0, 0}}},{29109, {{39825, 1, 0, 0},{39825, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{29111, {{39826, 1, 0, 0},{39826, 1, 1, 17}, {0, 0, 0, 0}}},{29111, {{39826, 1, 0, 0},{39826, 1, 1, 17}, {0, 0, 0, 0}}},{29112, {{39826, 1, 0, 0},{39826, 1, 1, 18}, {0, 0, 0, 0}}},{29112, {{39826, 1, 0, 0},{39826, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29113, {{39827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29116, {{39828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29117, {{39829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29118, {{39830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29119, {{39831, 1, 0, 0},{39831, 1, 1, 18}, {0, 0, 0, 0}}},{29119, {{39831, 1, 0, 0},{39831, 1, 1, 18}, {0, 0, 0, 0}}},{29120, {{39831, 1, 0, 0},{39831, 1, 1, 19}, {0, 0, 0, 0}}},{29120, {{39831, 1, 0, 0},{39831, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{29121, {{39833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29122, {{39834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29123, {{39835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29124, {{39836, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29125, {{39837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29126, {{39838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29127, {{39839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29128, {{39840, 1, 0, 0},{39840, 1, 1, 17}, {0, 0, 0, 0}}},{29128, {{39840, 1, 0, 0},{39840, 1, 1, 17}, {0, 0, 0, 0}}},{29129, {{39840, 1, 0, 0},{39840, 1, 1, 18}, {0, 0, 0, 0}}},{29129, {{39840, 1, 0, 0},{39840, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29130, {{39841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29131, {{39842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29132, {{39843, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29133, {{39844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29134, {{39845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29135, {{39846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29136, {{39847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29137, {{39848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29139, {{39849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29140, {{39850, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{29141, {{39851, 1, 0, 0},{39851, 1, 1, 20}, {0, 0, 0, 0}}},{29141, {{39851, 1, 0, 0},{39851, 1, 1, 20}, {0, 0, 0, 0}}},{29142, {{39851, 1, 0, 0},{39851, 1, 1, 21}, {0, 0, 0, 0}}},{29142, {{39851, 1, 0, 0},{39851, 1, 1, 21}, {0, 0, 0, 0}}},{29143, {{39851, 1, 0, 0},{39851, 1, 1, 19}, {0, 0, 0, 0}}},{29143, {{39851, 1, 0, 0},{39851, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{29144, {{39852, 1, 0, 0},{39852, 1, 1, 17}, {0, 0, 0, 0}}},{29144, {{39852, 1, 0, 0},{39852, 1, 1, 17}, {0, 0, 0, 0}}},{29145, {{39852, 1, 0, 0},{39852, 1, 1, 18}, {0, 0, 0, 0}}},{29145, {{39852, 1, 0, 0},{39852, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29146, {{39853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29148, {{39854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29150, {{39855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29151, {{39856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{29152, {{39857, 1, 0, 0},{39857, 1, 1, 17}, {0, 0, 0, 0}}},{29152, {{39857, 1, 0, 0},{39857, 1, 1, 17}, {0, 0, 0, 0}}},{29153, {{39857, 1, 0, 0},{39857, 1, 1, 18}, {0, 0, 0, 0}}},{29153, {{39857, 1, 0, 0},{39857, 1, 1, 18}, {0, 0, 0, 0}}},{29154, {{39857, 1, 0, 0},{39857, 1, 1, 19}, {0, 0, 0, 0}}},{29154, {{39857, 1, 0, 0},{39857, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{29155, {{39858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{29156, {{39860, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29157, {{39861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29158, {{39862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29159, {{39863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29160, {{39864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29161, {{39865, 1, 0, 0},{39865, 1, 1, 18}, {0, 0, 0, 0}}},{29161, {{39865, 1, 0, 0},{39865, 1, 1, 18}, {0, 0, 0, 0}}},{29162, {{39865, 1, 0, 0},{39865, 1, 1, 19}, {0, 0, 0, 0}}},{29162, {{39865, 1, 0, 0},{39865, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{29163, {{39866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29164, {{39867, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29165, {{39868, 1, 0, 0},{39868, 1, 1, 17}, {0, 0, 0, 0}}},{29165, {{39868, 1, 0, 0},{39868, 1, 1, 17}, {0, 0, 0, 0}}},{29166, {{39868, 1, 0, 0},{39868, 1, 1, 18}, {0, 0, 0, 0}}},{29166, {{39868, 1, 0, 0},{39868, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29167, {{39869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29168, {{39870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29169, {{39871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29170, {{39872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29171, {{39873, 1, 0, 0},{39873, 1, 1, 17}, {0, 0, 0, 0}}},{29171, {{39873, 1, 0, 0},{39873, 1, 1, 17}, {0, 0, 0, 0}}},{29172, {{39873, 1, 0, 0},{39873, 1, 1, 18}, {0, 0, 0, 0}}},{29172, {{39873, 1, 0, 0},{39873, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29173, {{39874, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29174, {{39875, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29175, {{39876, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29176, {{39877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29177, {{39878, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29178, {{39879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29179, {{39880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29180, {{39881, 1, 0, 0},{39881, 1, 1, 17}, {0, 0, 0, 0}}},{29180, {{39881, 1, 0, 0},{39881, 1, 1, 17}, {0, 0, 0, 0}}},{29181, {{39881, 1, 0, 0},{39881, 1, 1, 18}, {0, 0, 0, 0}}},{29181, {{39881, 1, 0, 0},{39881, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29182, {{39882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29183, {{39883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29184, {{39884, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{29185, {{39886, 1, 0, 0},{39886, 1, 1, 17}, {0, 0, 0, 0}}},{29185, {{39886, 1, 0, 0},{39886, 1, 1, 17}, {0, 0, 0, 0}}},{29186, {{39886, 1, 0, 0},{39886, 1, 1, 18}, {0, 0, 0, 0}}},{29186, {{39886, 1, 0, 0},{39886, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29187, {{39887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29188, {{39888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29189, {{39889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29190, {{39890, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{29191, {{39892, 1, 0, 0},{39892, 1, 1, 17}, {0, 0, 0, 0}}},{29191, {{39892, 1, 0, 0},{39892, 1, 1, 17}, {0, 0, 0, 0}}},{29192, {{39892, 1, 0, 0},{39892, 1, 1, 18}, {0, 0, 0, 0}}},{29192, {{39892, 1, 0, 0},{39892, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29193, {{39893, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29194, {{39894, 1, 0, 0},{39894, 1, 1, 19}, {0, 0, 0, 0}}},{29194, {{39894, 1, 0, 0},{39894, 1, 1, 19}, {0, 0, 0, 0}}},{29195, {{39894, 1, 0, 0},{39894, 1, 1, 18}, {0, 0, 0, 0}}},{29195, {{39894, 1, 0, 0},{39894, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29197, {{39895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29198, {{39896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29199, {{39897, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29200, {{39898, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{29201, {{39899, 1, 0, 0},{39899, 1, 1, 20}, {0, 0, 0, 0}}},{29201, {{39899, 1, 0, 0},{39899, 1, 1, 20}, {0, 0, 0, 0}}},{29202, {{39899, 1, 0, 0},{39899, 1, 1, 18}, {0, 0, 0, 0}}},{29202, {{39899, 1, 0, 0},{39899, 1, 1, 18}, {0, 0, 0, 0}}},{29203, {{39899, 1, 0, 0},{39899, 1, 1, 19}, {0, 0, 0, 0}}},{29203, {{39899, 1, 0, 0},{39899, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{29204, {{39900, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29205, {{39901, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29206, {{39902, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29207, {{39903, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29208, {{39904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29209, {{39905, 1, 0, 0},{39905, 1, 1, 18}, {0, 0, 0, 0}}},{29209, {{39905, 1, 0, 0},{39905, 1, 1, 18}, {0, 0, 0, 0}}},{29210, {{39905, 1, 0, 0},{39905, 1, 1, 19}, {0, 0, 0, 0}}},{29210, {{39905, 1, 0, 0},{39905, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{29211, {{39906, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29212, {{39907, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29213, {{39908, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29214, {{39909, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29215, {{39910, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29216, {{39911, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29217, {{39912, 1, 0, 0},{39912, 1, 1, 17}, {0, 0, 0, 0}}},{29217, {{39912, 1, 0, 0},{39912, 1, 1, 17}, {0, 0, 0, 0}}},{29218, {{39912, 1, 0, 0},{39912, 1, 1, 18}, {0, 0, 0, 0}}},{29218, {{39912, 1, 0, 0},{39912, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29219, {{39913, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29220, {{39914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29221, {{39915, 1, 0, 0},{39915, 1, 1, 17}, {0, 0, 0, 0}}},{29221, {{39915, 1, 0, 0},{39915, 1, 1, 17}, {0, 0, 0, 0}}},{29222, {{39915, 1, 0, 0},{39915, 1, 1, 18}, {0, 0, 0, 0}}},{29222, {{39915, 1, 0, 0},{39915, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29223, {{39916, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29224, {{39917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29225, {{39918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29226, {{39919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29227, {{39920, 1, 0, 0},{39920, 1, 1, 18}, {0, 0, 0, 0}}},{29227, {{39920, 1, 0, 0},{39920, 1, 1, 18}, {0, 0, 0, 0}}},{29228, {{39920, 1, 0, 0},{39920, 1, 1, 17}, {0, 0, 0, 0}}},{29228, {{39920, 1, 0, 0},{39920, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{29229, {{39921, 1, 0, 0},{39921, 1, 1, 18}, {0, 0, 0, 0}}},{29229, {{39921, 1, 0, 0},{39921, 1, 1, 18}, {0, 0, 0, 0}}},{29230, {{39921, 1, 0, 0},{39921, 1, 1, 19}, {0, 0, 0, 0}}},{29230, {{39921, 1, 0, 0},{39921, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{29233, {{39922, 1, 0, 0},{39922, 1, 1, 19}, {0, 0, 0, 0}}},{29233, {{39922, 1, 0, 0},{39922, 1, 1, 19}, {0, 0, 0, 0}}},{29234, {{39922, 1, 0, 0},{39922, 1, 1, 18}, {0, 0, 0, 0}}},{29234, {{39922, 1, 0, 0},{39922, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29235, {{39923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{29236, {{39925, 1, 0, 0},{39925, 1, 1, 17}, {0, 0, 0, 0}}},{29236, {{39925, 1, 0, 0},{39925, 1, 1, 17}, {0, 0, 0, 0}}},{29237, {{39925, 1, 0, 0},{39925, 1, 1, 18}, {0, 0, 0, 0}}},{29237, {{39925, 1, 0, 0},{39925, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29238, {{39926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29239, {{39927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29240, {{39928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29241, {{39929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29242, {{39930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29243, {{39931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29244, {{39932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29245, {{39933, 1, 0, 0},{39933, 1, 1, 17}, {0, 0, 0, 0}}},{29245, {{39933, 1, 0, 0},{39933, 1, 1, 17}, {0, 0, 0, 0}}},{29246, {{39933, 1, 0, 0},{39933, 1, 1, 18}, {0, 0, 0, 0}}},{29246, {{39933, 1, 0, 0},{39933, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29247, {{39934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29248, {{39935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29249, {{39936, 1, 0, 0},{39936, 1, 1, 17}, {0, 0, 0, 0}}},{29249, {{39936, 1, 0, 0},{39936, 1, 1, 17}, {0, 0, 0, 0}}},{29250, {{39936, 1, 0, 0},{39936, 1, 1, 18}, {0, 0, 0, 0}}},{29250, {{39936, 1, 0, 0},{39936, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29252, {{39937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29253, {{39938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29254, {{39939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29255, {{39940, 1, 0, 0},{39940, 1, 1, 17}, {0, 0, 0, 0}}},{29255, {{39940, 1, 0, 0},{39940, 1, 1, 17}, {0, 0, 0, 0}}},{29256, {{39940, 1, 0, 0},{39940, 1, 1, 18}, {0, 0, 0, 0}}},{29256, {{39940, 1, 0, 0},{39940, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29257, {{39941, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29258, {{39942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29259, {{39943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29260, {{39944, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29261, {{39945, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29262, {{39946, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29263, {{39947, 1, 0, 0},{39947, 1, 1, 17}, {0, 0, 0, 0}}},{29263, {{39947, 1, 0, 0},{39947, 1, 1, 17}, {0, 0, 0, 0}}},{29264, {{39947, 1, 0, 0},{39947, 1, 1, 18}, {0, 0, 0, 0}}},{29264, {{39947, 1, 0, 0},{39947, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{29265, {{39948, 1, 0, 0},{39948, 1, 1, 17}, {0, 0, 0, 0}}},{29265, {{39948, 1, 0, 0},{39948, 1, 1, 17}, {0, 0, 0, 0}}},{29266, {{39948, 1, 0, 0},{39948, 1, 1, 19}, {0, 0, 0, 0}}},{29266, {{39948, 1, 0, 0},{39948, 1, 1, 19}, {0, 0, 0, 0}}},{29267, {{39948, 1, 0, 0},{39948, 1, 1, 18}, {0, 0, 0, 0}}},{29267, {{39948, 1, 0, 0},{39948, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29268, {{39949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29269, {{39950, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29270, {{39951, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{29271, {{39952, 1, 0, 0},{39952, 1, 1, 19}, {0, 0, 0, 0}}},{29271, {{39952, 1, 0, 0},{39952, 1, 1, 19}, {0, 0, 0, 0}}},{29272, {{39952, 1, 0, 0},{39952, 1, 1, 17}, {0, 0, 0, 0}}},{29272, {{39952, 1, 0, 0},{39952, 1, 1, 17}, {0, 0, 0, 0}}},{29273, {{39952, 1, 0, 0},{39952, 1, 1, 18}, {0, 0, 0, 0}}},{29273, {{39952, 1, 0, 0},{39952, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29274, {{39953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29275, {{39954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29276, {{39955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29277, {{39956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29278, {{39957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29279, {{39958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29280, {{39959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29281, {{39960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29282, {{39961, 1, 0, 0},{39961, 1, 1, 17}, {0, 0, 0, 0}}},{29282, {{39961, 1, 0, 0},{39961, 1, 1, 17}, {0, 0, 0, 0}}},{29283, {{39961, 1, 0, 0},{39961, 1, 1, 18}, {0, 0, 0, 0}}},{29283, {{39961, 1, 0, 0},{39961, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29284, {{39962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29285, {{39963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29286, {{39964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29287, {{39965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29289, {{39966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29290, {{39967, 1, 0, 0},{39967, 1, 1, 16}, {0, 0, 0, 0}}},{29290, {{39967, 1, 0, 0},{39967, 1, 1, 16}, {0, 0, 0, 0}}},{29291, {{39967, 1, 0, 0},{39967, 1, 1, 17}, {0, 0, 0, 0}}},{29291, {{39967, 1, 0, 0},{39967, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{29292, {{39968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29293, {{39969, 1, 0, 0},{39969, 1, 1, 17}, {0, 0, 0, 0}}},{29293, {{39969, 1, 0, 0},{39969, 1, 1, 17}, {0, 0, 0, 0}}},{29294, {{39969, 1, 0, 0},{39969, 1, 1, 18}, {0, 0, 0, 0}}},{29294, {{39969, 1, 0, 0},{39969, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{29295, {{39970, 1, 0, 0},{39970, 1, 1, 18}, {0, 0, 0, 0}}},{29295, {{39970, 1, 0, 0},{39970, 1, 1, 18}, {0, 0, 0, 0}}},{29296, {{39970, 1, 0, 0},{39970, 1, 1, 17}, {0, 0, 0, 0}}},{29296, {{39970, 1, 0, 0},{39970, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{29297, {{39971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29298, {{39972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29299, {{39973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29300, {{39974, 1, 0, 0},{39974, 1, 1, 17}, {0, 0, 0, 0}}},{29300, {{39974, 1, 0, 0},{39974, 1, 1, 17}, {0, 0, 0, 0}}},{29301, {{39974, 1, 0, 0},{39974, 1, 1, 18}, {0, 0, 0, 0}}},{29301, {{39974, 1, 0, 0},{39974, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{29302, {{39975, 1, 0, 0},{39975, 1, 1, 17}, {0, 0, 0, 0}}},{29302, {{39975, 1, 0, 0},{39975, 1, 1, 17}, {0, 0, 0, 0}}},{29303, {{39975, 1, 0, 0},{39975, 1, 1, 18}, {0, 0, 0, 0}}},{29303, {{39975, 1, 0, 0},{39975, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29304, {{39976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29305, {{39977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29306, {{39978, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29307, {{39979, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29308, {{39980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29309, {{39981, 1, 0, 0},{39981, 1, 1, 17}, {0, 0, 0, 0}}},{29309, {{39981, 1, 0, 0},{39981, 1, 1, 17}, {0, 0, 0, 0}}},{29310, {{39981, 1, 0, 0},{39981, 1, 1, 18}, {0, 0, 0, 0}}},{29310, {{39981, 1, 0, 0},{39981, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29311, {{39982, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29312, {{39983, 1, 0, 0},{39983, 1, 1, 19}, {0, 0, 0, 0}}},{29312, {{39983, 1, 0, 0},{39983, 1, 1, 19}, {0, 0, 0, 0}}},{29313, {{39983, 1, 0, 0},{39983, 1, 1, 18}, {0, 0, 0, 0}}},{29313, {{39983, 1, 0, 0},{39983, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{29315, {{39984, 1, 0, 0},{39984, 1, 1, 18}, {0, 0, 0, 0}}},{29315, {{39984, 1, 0, 0},{39984, 1, 1, 18}, {0, 0, 0, 0}}},{29316, {{39984, 1, 0, 0},{39984, 1, 1, 17}, {0, 0, 0, 0}}},{29316, {{39984, 1, 0, 0},{39984, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{29317, {{39985, 1, 0, 0},{39985, 1, 1, 18}, {0, 0, 0, 0}}},{29317, {{39985, 1, 0, 0},{39985, 1, 1, 18}, {0, 0, 0, 0}}},{29318, {{39985, 1, 0, 0},{39985, 1, 1, 17}, {0, 0, 0, 0}}},{29318, {{39985, 1, 0, 0},{39985, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{29319, {{39986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29320, {{39987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29321, {{39988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29322, {{39989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29323, {{39990, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29324, {{39991, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29325, {{39992, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29326, {{39993, 1, 0, 0},{39993, 1, 1, 18}, {0, 0, 0, 0}}},{29326, {{39993, 1, 0, 0},{39993, 1, 1, 18}, {0, 0, 0, 0}}},{29327, {{39993, 1, 0, 0},{39993, 1, 1, 17}, {0, 0, 0, 0}}},{29327, {{39993, 1, 0, 0},{39993, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{29328, {{39994, 1, 0, 0},{39994, 1, 1, 18}, {0, 0, 0, 0}}},{29328, {{39994, 1, 0, 0},{39994, 1, 1, 18}, {0, 0, 0, 0}}},{29329, {{39994, 1, 0, 0},{39994, 1, 1, 17}, {0, 0, 0, 0}}},{29329, {{39994, 1, 0, 0},{39994, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{29330, {{39995, 1, 0, 0},{39995, 1, 1, 19}, {0, 0, 0, 0}}},{29330, {{39995, 1, 0, 0},{39995, 1, 1, 19}, {0, 0, 0, 0}}},{29331, {{39995, 1, 0, 0},{39995, 1, 1, 18}, {0, 0, 0, 0}}},{29331, {{39995, 1, 0, 0},{39995, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29333, {{39996, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29334, {{39997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29335, {{39998, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29336, {{39999, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29337, {{40000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{29338, {{40001, 1, 0, 0},{40001, 1, 1, 19}, {0, 0, 0, 0}}},{29338, {{40001, 1, 0, 0},{40001, 1, 1, 19}, {0, 0, 0, 0}}},{29339, {{40001, 1, 0, 0},{40001, 1, 1, 17}, {0, 0, 0, 0}}},{29339, {{40001, 1, 0, 0},{40001, 1, 1, 17}, {0, 0, 0, 0}}},{29340, {{40001, 1, 0, 0},{40001, 1, 1, 18}, {0, 0, 0, 0}}},{29340, {{40001, 1, 0, 0},{40001, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29341, {{40002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29342, {{40003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29343, {{40004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29344, {{40005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29345, {{40006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29346, {{40007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29347, {{40008, 1, 0, 0},{40008, 1, 1, 19}, {0, 0, 0, 0}}},{29347, {{40008, 1, 0, 0},{40008, 1, 1, 19}, {0, 0, 0, 0}}},{29348, {{40008, 1, 0, 0},{40008, 1, 1, 18}, {0, 0, 0, 0}}},{29348, {{40008, 1, 0, 0},{40008, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29349, {{40009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29350, {{40010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29351, {{40011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29352, {{40012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29353, {{40013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29354, {{40014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29355, {{40015, 1, 0, 0},{40015, 1, 1, 17}, {0, 0, 0, 0}}},{29355, {{40015, 1, 0, 0},{40015, 1, 1, 17}, {0, 0, 0, 0}}},{29356, {{40015, 1, 0, 0},{40015, 1, 1, 18}, {0, 0, 0, 0}}},{29356, {{40015, 1, 0, 0},{40015, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29357, {{40016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29358, {{40017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{29359, {{40018, 1, 0, 0},{40018, 1, 1, 20}, {0, 0, 0, 0}}},{29359, {{40018, 1, 0, 0},{40018, 1, 1, 20}, {0, 0, 0, 0}}},{29360, {{40018, 1, 0, 0},{40018, 1, 1, 19}, {0, 0, 0, 0}}},{29360, {{40018, 1, 0, 0},{40018, 1, 1, 19}, {0, 0, 0, 0}}},{29361, {{40018, 1, 0, 0},{40018, 1, 1, 18}, {0, 0, 0, 0}}},{29361, {{40018, 1, 0, 0},{40018, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29362, {{40019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29363, {{40020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29364, {{40021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29365, {{40022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{29366, {{40023, 1, 0, 0},{40023, 1, 1, 19}, {0, 0, 0, 0}}},{29366, {{40023, 1, 0, 0},{40023, 1, 1, 19}, {0, 0, 0, 0}}},{29367, {{40023, 1, 0, 0},{40023, 1, 1, 18}, {0, 0, 0, 0}}},{29367, {{40023, 1, 0, 0},{40023, 1, 1, 18}, {0, 0, 0, 0}}},{29368, {{40023, 1, 0, 0},{40023, 1, 1, 21}, {0, 0, 0, 0}}},{29368, {{40023, 1, 0, 0},{40023, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{29370, {{40024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29371, {{40025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29372, {{40026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29373, {{40027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29374, {{40028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29375, {{40029, 1, 0, 0},{40029, 1, 1, 17}, {0, 0, 0, 0}}},{29375, {{40029, 1, 0, 0},{40029, 1, 1, 17}, {0, 0, 0, 0}}},{29376, {{40029, 1, 0, 0},{40029, 1, 1, 18}, {0, 0, 0, 0}}},{29376, {{40029, 1, 0, 0},{40029, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29377, {{40030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29378, {{40031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29379, {{40032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29381, {{40033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29382, {{40034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29383, {{40035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29384, {{40036, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29385, {{40037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29386, {{40038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29387, {{40039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29388, {{40040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29389, {{40041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29390, {{40042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29391, {{40043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29392, {{40044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29393, {{40045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29394, {{40046, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29395, {{40047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29396, {{40048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29397, {{40049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29398, {{40050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29399, {{40051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29400, {{40052, 1, 0, 0},{40052, 1, 1, 16}, {0, 0, 0, 0}}},{29400, {{40052, 1, 0, 0},{40052, 1, 1, 16}, {0, 0, 0, 0}}},{29401, {{40052, 1, 0, 0},{40052, 1, 1, 17}, {0, 0, 0, 0}}},{29401, {{40052, 1, 0, 0},{40052, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{29402, {{40053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{29403, {{40054, 1, 0, 0},{40054, 1, 1, 17}, {0, 0, 0, 0}}},{29403, {{40054, 1, 0, 0},{40054, 1, 1, 17}, {0, 0, 0, 0}}},{29404, {{40054, 1, 0, 0},{40054, 1, 1, 18}, {0, 0, 0, 0}}},{29404, {{40054, 1, 0, 0},{40054, 1, 1, 18}, {0, 0, 0, 0}}},{29405, {{40054, 1, 0, 0},{40054, 1, 1, 19}, {0, 0, 0, 0}}},{29405, {{40054, 1, 0, 0},{40054, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{29406, {{40055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29407, {{40056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29408, {{40057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29409, {{40058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29410, {{40059, 1, 0, 0}, {0, 0, 0, 0}}}}}
};

static const MJMappingSet chunked_mj_mappings_009ce5_009e1e[314] = {
    {1, {{29411, {{40165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29412, {{40166, 1, 0, 0},{40166, 1, 1, 18}, {0, 0, 0, 0}}},{29412, {{40166, 1, 0, 0},{40166, 1, 1, 18}, {0, 0, 0, 0}}},{29413, {{40166, 1, 0, 0},{40166, 1, 1, 19}, {0, 0, 0, 0}}},{29413, {{40166, 1, 0, 0},{40166, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{29414, {{40167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29415, {{40168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29416, {{40169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29417, {{40170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29418, {{40171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29419, {{40172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29420, {{40173, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29421, {{40174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29422, {{40175, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29423, {{40176, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29424, {{40177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29425, {{40178, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29426, {{40179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29427, {{40180, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29428, {{40181, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29429, {{40182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29430, {{40183, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29431, {{40184, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29432, {{40185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29433, {{40186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29434, {{40187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29435, {{40188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29436, {{40189, 1, 0, 0},{40189, 1, 1, 16}, {0, 0, 0, 0}}},{29436, {{40189, 1, 0, 0},{40189, 1, 1, 16}, {0, 0, 0, 0}}},{29437, {{40189, 1, 0, 0},{40189, 1, 1, 17}, {0, 0, 0, 0}}},{29437, {{40189, 1, 0, 0},{40189, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{29438, {{40190, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29439, {{40191, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29440, {{40192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29441, {{40193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29442, {{40194, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29443, {{40195, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29444, {{40196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29445, {{40197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29446, {{40198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29447, {{40199, 1, 0, 0},{40199, 1, 1, 18}, {0, 0, 0, 0}}},{29447, {{40199, 1, 0, 0},{40199, 1, 1, 18}, {0, 0, 0, 0}}},{29448, {{40199, 1, 0, 0},{40199, 1, 1, 19}, {0, 0, 0, 0}}},{29448, {{40199, 1, 0, 0},{40199, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{29449, {{40200, 1, 0, 0},{40200, 1, 1, 18}, {0, 0, 0, 0}}},{29449, {{40200, 1, 0, 0},{40200, 1, 1, 18}, {0, 0, 0, 0}}},{29450, {{40200, 1, 0, 0},{40200, 1, 1, 19}, {0, 0, 0, 0}}},{29450, {{40200, 1, 0, 0},{40200, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {6, {{29451, {{40201, 1, 0, 0},{40201, 1, 1, 19}, {0, 0, 0, 0}}},{29451, {{40201, 1, 0, 0},{40201, 1, 1, 19}, {0, 0, 0, 0}}},{29452, {{40201, 1, 0, 0},{40201, 1, 1, 18}, {0, 0, 0, 0}}},{29452, {{40201, 1, 0, 0},{40201, 1, 1, 18}, {0, 0, 0, 0}}},{29453, {{40201, 1, 0, 0},{40201, 1, 1, 20}, {0, 0, 0, 0}}},{29453, {{40201, 1, 0, 0},{40201, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{29454, {{40202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29455, {{40203, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29456, {{40204, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29457, {{40205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29458, {{40206, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29459, {{40207, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29460, {{40208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29461, {{40209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29462, {{40210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29463, {{40211, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29464, {{40212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29465, {{40213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29466, {{40214, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29467, {{40215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29468, {{40216, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29469, {{40217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29470, {{40218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29471, {{40219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29472, {{40220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29473, {{40221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29474, {{40222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29475, {{40223, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29476, {{40224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29477, {{40225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29478, {{40226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29479, {{40227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29480, {{40228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29481, {{40229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29482, {{40230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29483, {{40231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29484, {{40232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29485, {{40233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29486, {{40234, 1, 0, 0},{40234, 1, 1, 17}, {0, 0, 0, 0}}},{29486, {{40234, 1, 0, 0},{40234, 1, 1, 17}, {0, 0, 0, 0}}},{29487, {{40234, 1, 0, 0},{40234, 1, 1, 18}, {0, 0, 0, 0}}},{29487, {{40234, 1, 0, 0},{40234, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29488, {{40235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29489, {{40236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29490, {{40237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29491, {{40238, 1, 0, 0},{40238, 1, 1, 16}, {0, 0, 0, 0}}},{29491, {{40238, 1, 0, 0},{40238, 1, 1, 16}, {0, 0, 0, 0}}},{29492, {{40238, 1, 0, 0},{40238, 1, 1, 17}, {0, 0, 0, 0}}},{29492, {{40238, 1, 0, 0},{40238, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{29493, {{40239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29494, {{40240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29495, {{40241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29496, {{40242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29497, {{40243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29498, {{40244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29499, {{40245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29500, {{40246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29501, {{40247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29502, {{40248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29503, {{40249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29504, {{40250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29505, {{40251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29506, {{40252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29507, {{40253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29508, {{40254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29509, {{40255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29510, {{40256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29511, {{40257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29512, {{40258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29513, {{40259, 1, 0, 0},{40259, 1, 1, 17}, {0, 0, 0, 0}}},{29513, {{40259, 1, 0, 0},{40259, 1, 1, 17}, {0, 0, 0, 0}}},{29514, {{40259, 1, 0, 0},{40259, 1, 1, 18}, {0, 0, 0, 0}}},{29514, {{40259, 1, 0, 0},{40259, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29515, {{40260, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29516, {{40261, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29517, {{40262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29518, {{40263, 1, 0, 0},{40263, 1, 1, 17}, {0, 0, 0, 0}}},{29518, {{40263, 1, 0, 0},{40263, 1, 1, 17}, {0, 0, 0, 0}}},{29519, {{40263, 1, 0, 0},{40263, 1, 1, 18}, {0, 0, 0, 0}}},{29519, {{40263, 1, 0, 0},{40263, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{29520, {{40264, 1, 0, 0},{40264, 1, 1, 18}, {0, 0, 0, 0}}},{29520, {{40264, 1, 0, 0},{40264, 1, 1, 18}, {0, 0, 0, 0}}},{29521, {{40264, 1, 0, 0},{40264, 1, 1, 19}, {0, 0, 0, 0}}},{29521, {{40264, 1, 0, 0},{40264, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{29522, {{40265, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29523, {{40266, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29524, {{40267, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29525, {{40268, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29526, {{40269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29527, {{40270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29528, {{40271, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29529, {{40272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29530, {{40273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29531, {{40274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29532, {{40275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29533, {{40276, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29534, {{40277, 1, 0, 0},{40277, 1, 1, 16}, {0, 0, 0, 0}}},{29534, {{40277, 1, 0, 0},{40277, 1, 1, 16}, {0, 0, 0, 0}}},{29535, {{40277, 1, 0, 0},{40277, 1, 1, 17}, {0, 0, 0, 0}}},{29535, {{40277, 1, 0, 0},{40277, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{29536, {{40278, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29537, {{40279, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29538, {{40280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29539, {{40281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29540, {{40282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29541, {{40283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29542, {{40284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29544, {{40285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29545, {{40286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29546, {{40287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29547, {{40288, 1, 0, 0},{40288, 1, 1, 19}, {0, 0, 0, 0}}},{29547, {{40288, 1, 0, 0},{40288, 1, 1, 19}, {0, 0, 0, 0}}},{29548, {{40288, 1, 0, 0},{40288, 1, 1, 18}, {0, 0, 0, 0}}},{29548, {{40288, 1, 0, 0},{40288, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29549, {{40289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29550, {{40290, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29551, {{40291, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29552, {{40292, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29553, {{40293, 1, 0, 0},{40293, 1, 1, 17}, {0, 0, 0, 0}}},{29553, {{40293, 1, 0, 0},{40293, 1, 1, 17}, {0, 0, 0, 0}}},{29554, {{40293, 1, 0, 0},{40293, 1, 1, 18}, {0, 0, 0, 0}}},{29554, {{40293, 1, 0, 0},{40293, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29555, {{40294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29556, {{40295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29557, {{40296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29558, {{40297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29560, {{40298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29561, {{40299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29562, {{40300, 1, 0, 0},{40300, 1, 1, 19}, {0, 0, 0, 0}}},{29562, {{40300, 1, 0, 0},{40300, 1, 1, 19}, {0, 0, 0, 0}}},{29563, {{40300, 1, 0, 0},{40300, 1, 1, 18}, {0, 0, 0, 0}}},{29563, {{40300, 1, 0, 0},{40300, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29564, {{40301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29565, {{40302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29566, {{40303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29567, {{40304, 1, 0, 0},{40304, 1, 1, 17}, {0, 0, 0, 0}}},{29567, {{40304, 1, 0, 0},{40304, 1, 1, 17}, {0, 0, 0, 0}}},{29568, {{40304, 1, 0, 0},{40304, 1, 1, 18}, {0, 0, 0, 0}}},{29568, {{40304, 1, 0, 0},{40304, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29569, {{40305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29570, {{40306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29571, {{40307, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29572, {{40308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29573, {{40309, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29574, {{40310, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29575, {{40311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29576, {{40312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29577, {{40313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29578, {{40314, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29579, {{40315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29580, {{40316, 1, 0, 0},{40316, 1, 1, 17}, {0, 0, 0, 0}}},{29580, {{40316, 1, 0, 0},{40316, 1, 1, 17}, {0, 0, 0, 0}}},{29581, {{40316, 1, 0, 0},{40316, 1, 1, 18}, {0, 0, 0, 0}}},{29581, {{40316, 1, 0, 0},{40316, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29582, {{40317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29583, {{40318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29584, {{40319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29585, {{40320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29586, {{40321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29587, {{40322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29588, {{40323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29589, {{40324, 1, 0, 0},{40324, 1, 1, 17}, {0, 0, 0, 0}}},{29589, {{40324, 1, 0, 0},{40324, 1, 1, 17}, {0, 0, 0, 0}}},{29590, {{40324, 1, 0, 0},{40324, 1, 1, 18}, {0, 0, 0, 0}}},{29590, {{40324, 1, 0, 0},{40324, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29591, {{40325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29592, {{40326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29593, {{40327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29594, {{40328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29595, {{40329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29596, {{40330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29597, {{40331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29598, {{40332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29599, {{40333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29600, {{40334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29601, {{40335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29602, {{40336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{29603, {{40338, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29604, {{40339, 1, 0, 0},{40339, 1, 1, 17}, {0, 0, 0, 0}}},{29604, {{40339, 1, 0, 0},{40339, 1, 1, 17}, {0, 0, 0, 0}}},{29605, {{40339, 1, 0, 0},{40339, 1, 1, 18}, {0, 0, 0, 0}}},{29605, {{40339, 1, 0, 0},{40339, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29606, {{40340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29607, {{40341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29608, {{40342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29609, {{40343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29610, {{40344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29611, {{40345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29612, {{40346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29613, {{40347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29614, {{40348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29615, {{40349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29616, {{40350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29617, {{40351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29618, {{40352, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29619, {{40353, 1, 0, 0},{40353, 1, 1, 17}, {0, 0, 0, 0}}},{29619, {{40353, 1, 0, 0},{40353, 1, 1, 17}, {0, 0, 0, 0}}},{60360, {{40353, 1, 0, 0},{40353, 1, 1, 18}, {0, 0, 0, 0}}},{60360, {{40353, 1, 0, 0},{40353, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29620, {{40354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29621, {{40355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29622, {{40356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29623, {{40357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29624, {{40358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29625, {{40359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29626, {{40360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29627, {{40361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29628, {{40362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29629, {{40363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29630, {{40364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29631, {{40365, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29632, {{40366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29633, {{40367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29634, {{40368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29635, {{40369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{29636, {{40370, 1, 0, 0},{40370, 1, 1, 17}, {0, 0, 0, 0}}},{29636, {{40370, 1, 0, 0},{40370, 1, 1, 17}, {0, 0, 0, 0}}},{29637, {{40370, 1, 0, 0},{40370, 1, 1, 18}, {0, 0, 0, 0}}},{29637, {{40370, 1, 0, 0},{40370, 1, 1, 18}, {0, 0, 0, 0}}},{29638, {{40370, 1, 0, 0},{40370, 1, 1, 19}, {0, 0, 0, 0}}},{29638, {{40370, 1, 0, 0},{40370, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{29639, {{40371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{29640, {{40372, 1, 0, 0},{40372, 1, 1, 18}, {0, 0, 0, 0}}},{29640, {{40372, 1, 0, 0},{40372, 1, 1, 18}, {0, 0, 0, 0}}},{30243, {{40372, 1, 0, 0},{40372, 1, 1, 0},{40372, 1, 1, 19},{64045, 1, 0, 0}, {0, 0, 0, 0}}},{30243, {{40372, 1, 0, 0},{40372, 1, 1, 0},{40372, 1, 1, 19},{64045, 1, 0, 0}, {0, 0, 0, 0}}},{30243, {{40372, 1, 0, 0},{40372, 1, 1, 0},{40372, 1, 1, 19},{64045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29641, {{40373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29642, {{40374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29643, {{40375, 1, 0, 0},{40375, 1, 1, 16}, {0, 0, 0, 0}}},{29643, {{40375, 1, 0, 0},{40375, 1, 1, 16}, {0, 0, 0, 0}}},{29644, {{40375, 1, 0, 0},{40375, 1, 1, 17}, {0, 0, 0, 0}}},{29644, {{40375, 1, 0, 0},{40375, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{29645, {{40376, 1, 0, 0},{40376, 1, 1, 17}, {0, 0, 0, 0}}},{29645, {{40376, 1, 0, 0},{40376, 1, 1, 17}, {0, 0, 0, 0}}},{29646, {{40376, 1, 0, 0},{40376, 1, 1, 18}, {0, 0, 0, 0}}},{29646, {{40376, 1, 0, 0},{40376, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{29647, {{40377, 1, 0, 0},{40377, 1, 1, 17}, {0, 0, 0, 0}}},{29647, {{40377, 1, 0, 0},{40377, 1, 1, 17}, {0, 0, 0, 0}}},{29648, {{40377, 1, 0, 0},{40377, 1, 1, 18}, {0, 0, 0, 0}}},{29648, {{40377, 1, 0, 0},{40377, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29649, {{40378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29650, {{40379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29651, {{40380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29652, {{40381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29653, {{40382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{29654, {{40383, 1, 0, 0},{40383, 1, 1, 18}, {0, 0, 0, 0}}},{29654, {{40383, 1, 0, 0},{40383, 1, 1, 18}, {0, 0, 0, 0}}},{29655, {{40383, 1, 0, 0},{40383, 1, 1, 20}, {0, 0, 0, 0}}},{29655, {{40383, 1, 0, 0},{40383, 1, 1, 20}, {0, 0, 0, 0}}},{29656, {{40383, 1, 0, 0},{40383, 1, 1, 19}, {0, 0, 0, 0}}},{29656, {{40383, 1, 0, 0},{40383, 1, 1, 19}, {0, 0, 0, 0}}},{29657, {{40383, 1, 0, 0},{40383, 1, 1, 21}, {0, 0, 0, 0}}},{29657, {{40383, 1, 0, 0},{40383, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {8, {{29658, {{40384, 1, 0, 0},{40384, 1, 1, 21}, {0, 0, 0, 0}}},{29658, {{40384, 1, 0, 0},{40384, 1, 1, 21}, {0, 0, 0, 0}}},{29659, {{40384, 1, 0, 0},{40384, 1, 1, 20}, {0, 0, 0, 0}}},{29659, {{40384, 1, 0, 0},{40384, 1, 1, 20}, {0, 0, 0, 0}}},{29660, {{40384, 1, 0, 0},{40384, 1, 1, 19}, {0, 0, 0, 0}}},{29660, {{40384, 1, 0, 0},{40384, 1, 1, 19}, {0, 0, 0, 0}}},{29661, {{40384, 1, 0, 0},{40384, 1, 1, 18}, {0, 0, 0, 0}}},{29661, {{40384, 1, 0, 0},{40384, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{29663, {{40385, 1, 0, 0},{40385, 1, 1, 17}, {0, 0, 0, 0}}},{29663, {{40385, 1, 0, 0},{40385, 1, 1, 17}, {0, 0, 0, 0}}},{29664, {{40385, 1, 0, 0},{40385, 1, 1, 18}, {0, 0, 0, 0}}},{29664, {{40385, 1, 0, 0},{40385, 1, 1, 18}, {0, 0, 0, 0}}},{29665, {{40385, 1, 0, 0},{40385, 1, 1, 19}, {0, 0, 0, 0}}},{29665, {{40385, 1, 0, 0},{40385, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{29666, {{40386, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29667, {{40387, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29668, {{40388, 1, 0, 0},{40388, 1, 1, 17}, {0, 0, 0, 0}}},{29668, {{40388, 1, 0, 0},{40388, 1, 1, 17}, {0, 0, 0, 0}}},{29669, {{40388, 1, 0, 0},{40388, 1, 1, 18}, {0, 0, 0, 0}}},{29669, {{40388, 1, 0, 0},{40388, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29670, {{40389, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29671, {{40390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29672, {{40391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29673, {{40392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29674, {{40393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29675, {{40394, 1, 0, 0},{40394, 1, 1, 17}, {0, 0, 0, 0}}},{29675, {{40394, 1, 0, 0},{40394, 1, 1, 17}, {0, 0, 0, 0}}},{29676, {{40394, 1, 0, 0},{40394, 1, 1, 18}, {0, 0, 0, 0}}},{29676, {{40394, 1, 0, 0},{40394, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29677, {{40395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29678, {{40396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29679, {{40397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29680, {{40398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29681, {{40399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29682, {{40400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29683, {{40401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29684, {{40402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29685, {{40403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29686, {{40404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29687, {{40405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29688, {{40406, 1, 0, 0},{40406, 1, 1, 17}, {0, 0, 0, 0}}},{29688, {{40406, 1, 0, 0},{40406, 1, 1, 17}, {0, 0, 0, 0}}},{29689, {{40406, 1, 0, 0},{40406, 1, 1, 18}, {0, 0, 0, 0}}},{29689, {{40406, 1, 0, 0},{40406, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{29690, {{40407, 1, 0, 0},{40407, 1, 1, 17}, {0, 0, 0, 0}}},{29690, {{40407, 1, 0, 0},{40407, 1, 1, 17}, {0, 0, 0, 0}}},{29691, {{40407, 1, 0, 0},{40407, 1, 1, 18}, {0, 0, 0, 0}}},{29691, {{40407, 1, 0, 0},{40407, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29692, {{40408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29693, {{40409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29694, {{40410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29695, {{40411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29696, {{40412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29697, {{40413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29698, {{40414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29699, {{40415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29700, {{40416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29702, {{40417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29703, {{40418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29704, {{40419, 1, 0, 0},{40419, 1, 1, 18}, {0, 0, 0, 0}}},{29704, {{40419, 1, 0, 0},{40419, 1, 1, 18}, {0, 0, 0, 0}}},{29705, {{40419, 1, 0, 0},{40419, 1, 1, 17}, {0, 0, 0, 0}}},{29705, {{40419, 1, 0, 0},{40419, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{29706, {{40420, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29707, {{40421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29708, {{40422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29709, {{40423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29710, {{40424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29711, {{40425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29712, {{40426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29713, {{40427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29714, {{40428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29715, {{40429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29716, {{40430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29717, {{40431, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29718, {{40432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29719, {{40433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29720, {{40434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29721, {{40435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29722, {{40436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29723, {{40437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29724, {{40438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29725, {{40439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29726, {{40440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29727, {{40441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29728, {{40442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29729, {{40443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29730, {{40444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29731, {{40445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29732, {{40446, 1, 0, 0},{40446, 1, 1, 17}, {0, 0, 0, 0}}},{29732, {{40446, 1, 0, 0},{40446, 1, 1, 17}, {0, 0, 0, 0}}},{29733, {{40446, 1, 0, 0},{40446, 1, 1, 18}, {0, 0, 0, 0}}},{29733, {{40446, 1, 0, 0},{40446, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29734, {{40447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29735, {{40448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29736, {{40449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29737, {{40450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29738, {{40451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29739, {{40452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29740, {{40453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29741, {{40454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29742, {{40455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29743, {{40456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29744, {{40457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29745, {{40458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29746, {{40459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29747, {{40460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29748, {{40461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29749, {{40462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29750, {{40463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29751, {{40464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29752, {{40465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29753, {{40466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29754, {{40467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29755, {{40468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29756, {{40469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29757, {{40470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29759, {{40471, 1, 0, 0},{40471, 1, 1, 16}, {0, 0, 0, 0}}},{29759, {{40471, 1, 0, 0},{40471, 1, 1, 16}, {0, 0, 0, 0}}},{29760, {{40471, 1, 0, 0},{40471, 1, 1, 17}, {0, 0, 0, 0}}},{29760, {{40471, 1, 0, 0},{40471, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{29761, {{40472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29762, {{40473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29763, {{40474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29764, {{40475, 1, 0, 0},{40475, 1, 1, 17}, {0, 0, 0, 0}}},{29764, {{40475, 1, 0, 0},{40475, 1, 1, 17}, {0, 0, 0, 0}}},{29765, {{40475, 1, 0, 0},{40475, 1, 1, 18}, {0, 0, 0, 0}}},{29765, {{40475, 1, 0, 0},{40475, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29766, {{40476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29767, {{40477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29768, {{40478, 1, 0, 0}, {0, 0, 0, 0}}}}}
};

static const MJMappingSet chunked_mj_mappings_009e75_009fef[379] = {
    {1, {{29769, {{40565, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29770, {{40566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29771, {{40567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29772, {{40568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29773, {{40569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29774, {{40570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29775, {{40571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29776, {{40572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29777, {{40573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{29778, {{40575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29779, {{40576, 1, 0, 0},{40576, 1, 1, 17}, {0, 0, 0, 0}}},{29779, {{40576, 1, 0, 0},{40576, 1, 1, 17}, {0, 0, 0, 0}}},{29780, {{40576, 1, 0, 0},{40576, 1, 1, 18}, {0, 0, 0, 0}}},{29780, {{40576, 1, 0, 0},{40576, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{29781, {{40577, 1, 0, 0},{40577, 1, 1, 17}, {0, 0, 0, 0}}},{29781, {{40577, 1, 0, 0},{40577, 1, 1, 17}, {0, 0, 0, 0}}},{29782, {{40577, 1, 0, 0},{40577, 1, 1, 18}, {0, 0, 0, 0}}},{29782, {{40577, 1, 0, 0},{40577, 1, 1, 18}, {0, 0, 0, 0}}},{60364, {{40577, 1, 0, 0},{40577, 1, 1, 19}, {0, 0, 0, 0}}},{60364, {{40577, 1, 0, 0},{40577, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{29783, {{40578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29784, {{40579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29785, {{40580, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29786, {{40581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29787, {{40582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29788, {{40583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29789, {{40584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29790, {{40585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29791, {{40586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29792, {{40587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29793, {{40588, 1, 0, 0},{40588, 1, 1, 17}, {0, 0, 0, 0}}},{29793, {{40588, 1, 0, 0},{40588, 1, 1, 17}, {0, 0, 0, 0}}},{29794, {{40588, 1, 0, 0},{40588, 1, 1, 18}, {0, 0, 0, 0}}},{29794, {{40588, 1, 0, 0},{40588, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29795, {{40589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29796, {{40590, 1, 0, 0},{40590, 1, 1, 18}, {0, 0, 0, 0}}},{29796, {{40590, 1, 0, 0},{40590, 1, 1, 18}, {0, 0, 0, 0}}},{29797, {{40590, 1, 0, 0},{40590, 1, 1, 19}, {0, 0, 0, 0}}},{29797, {{40590, 1, 0, 0},{40590, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{29798, {{40591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29799, {{40592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29800, {{40593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29801, {{40594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29802, {{40595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29803, {{40596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29804, {{40597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29805, {{40598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29806, {{40599, 1, 0, 0},{40599, 1, 1, 18}, {0, 0, 0, 0}}},{29806, {{40599, 1, 0, 0},{40599, 1, 1, 18}, {0, 0, 0, 0}}},{29807, {{40599, 1, 0, 0},{40599, 1, 1, 19}, {0, 0, 0, 0}}},{29807, {{40599, 1, 0, 0},{40599, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{29808, {{40600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29809, {{40601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29810, {{40602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29811, {{40603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29812, {{40604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29813, {{40605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29814, {{40606, 1, 0, 0},{40606, 1, 1, 17}, {0, 0, 0, 0}}},{29814, {{40606, 1, 0, 0},{40606, 1, 1, 17}, {0, 0, 0, 0}}},{29815, {{40606, 1, 0, 0},{40606, 1, 1, 18}, {0, 0, 0, 0}}},{29815, {{40606, 1, 0, 0},{40606, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{29816, {{40607, 1, 0, 0},{40607, 1, 1, 20}, {0, 0, 0, 0}}},{29816, {{40607, 1, 0, 0},{40607, 1, 1, 20}, {0, 0, 0, 0}}},{29817, {{40607, 1, 0, 0},{40607, 1, 1, 19}, {0, 0, 0, 0}}},{29817, {{40607, 1, 0, 0},{40607, 1, 1, 19}, {0, 0, 0, 0}}},{29818, {{40607, 1, 0, 0},{40607, 1, 1, 21}, {0, 0, 0, 0}}},{29818, {{40607, 1, 0, 0},{40607, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{29819, {{40608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29820, {{40609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29821, {{40610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29822, {{40611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29823, {{40612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29824, {{40613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {2, {{29826, {{40614, 1, 0, 0}, {0, 0, 0, 0}}},{29825, {{40614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29827, {{40615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29828, {{40616, 1, 0, 0},{40616, 1, 1, 17}, {0, 0, 0, 0}}},{29828, {{40616, 1, 0, 0},{40616, 1, 1, 17}, {0, 0, 0, 0}}},{29829, {{40616, 1, 0, 0},{40616, 1, 1, 18}, {0, 0, 0, 0}}},{29829, {{40616, 1, 0, 0},{40616, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{29830, {{40617, 1, 0, 0},{40617, 1, 1, 17}, {0, 0, 0, 0}}},{29830, {{40617, 1, 0, 0},{40617, 1, 1, 17}, {0, 0, 0, 0}}},{29831, {{40617, 1, 0, 0},{40617, 1, 1, 18}, {0, 0, 0, 0}}},{29831, {{40617, 1, 0, 0},{40617, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{29832, {{40618, 1, 0, 0},{40618, 1, 1, 19}, {0, 0, 0, 0}}},{29832, {{40618, 1, 0, 0},{40618, 1, 1, 19}, {0, 0, 0, 0}}},{29833, {{40618, 1, 0, 0},{40618, 1, 1, 18}, {0, 0, 0, 0}}},{29833, {{40618, 1, 0, 0},{40618, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29834, {{40619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29835, {{40620, 1, 0, 0},{40620, 1, 1, 17}, {0, 0, 0, 0}}},{29835, {{40620, 1, 0, 0},{40620, 1, 1, 17}, {0, 0, 0, 0}}},{29836, {{40620, 1, 0, 0},{40620, 1, 1, 18}, {0, 0, 0, 0}}},{29836, {{40620, 1, 0, 0},{40620, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{29837, {{40621, 1, 0, 0},{40621, 1, 1, 18}, {0, 0, 0, 0}}},{29837, {{40621, 1, 0, 0},{40621, 1, 1, 18}, {0, 0, 0, 0}}},{29838, {{40621, 1, 0, 0},{40621, 1, 1, 19}, {0, 0, 0, 0}}},{29838, {{40621, 1, 0, 0},{40621, 1, 1, 19}, {0, 0, 0, 0}}},{29839, {{40621, 1, 0, 0},{40621, 1, 1, 20}, {0, 0, 0, 0}}},{29839, {{40621, 1, 0, 0},{40621, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{29840, {{40622, 1, 0, 0},{40622, 1, 1, 17}, {0, 0, 0, 0}}},{29840, {{40622, 1, 0, 0},{40622, 1, 1, 17}, {0, 0, 0, 0}}},{59258, {{40622, 1, 0, 0},{40622, 1, 1, 18}, {0, 0, 0, 0}}},{59258, {{40622, 1, 0, 0},{40622, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29841, {{40623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29843, {{40624, 1, 0, 0},{40624, 1, 1, 17}, {0, 0, 0, 0}}},{29843, {{40624, 1, 0, 0},{40624, 1, 1, 17}, {0, 0, 0, 0}}},{29844, {{40624, 1, 0, 0},{40624, 1, 1, 18}, {0, 0, 0, 0}}},{29844, {{40624, 1, 0, 0},{40624, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29845, {{40625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29846, {{40626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29847, {{40627, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29848, {{40628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29849, {{40629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{29850, {{40631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29851, {{40632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29852, {{40633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29853, {{40634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29854, {{40635, 1, 0, 0},{40635, 1, 1, 19}, {0, 0, 0, 0}}},{29854, {{40635, 1, 0, 0},{40635, 1, 1, 19}, {0, 0, 0, 0}}},{29855, {{40635, 1, 0, 0},{40635, 1, 1, 18}, {0, 0, 0, 0}}},{29855, {{40635, 1, 0, 0},{40635, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{29856, {{40636, 1, 0, 0},{40636, 1, 1, 17}, {0, 0, 0, 0}}},{29856, {{40636, 1, 0, 0},{40636, 1, 1, 17}, {0, 0, 0, 0}}},{29857, {{40636, 1, 0, 0},{40636, 1, 1, 18}, {0, 0, 0, 0}}},{29857, {{40636, 1, 0, 0},{40636, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{29858, {{40637, 1, 0, 0},{40637, 1, 1, 17}, {0, 0, 0, 0}}},{29858, {{40637, 1, 0, 0},{40637, 1, 1, 17}, {0, 0, 0, 0}}},{29859, {{40637, 1, 0, 0},{40637, 1, 1, 18}, {0, 0, 0, 0}}},{29859, {{40637, 1, 0, 0},{40637, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{29860, {{40638, 1, 0, 0},{40638, 1, 1, 17}, {0, 0, 0, 0}}},{29860, {{40638, 1, 0, 0},{40638, 1, 1, 17}, {0, 0, 0, 0}}},{29861, {{40638, 1, 0, 0},{40638, 1, 1, 18}, {0, 0, 0, 0}}},{29861, {{40638, 1, 0, 0},{40638, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{29862, {{40639, 1, 0, 0},{40639, 1, 1, 19}, {0, 0, 0, 0}}},{29862, {{40639, 1, 0, 0},{40639, 1, 1, 19}, {0, 0, 0, 0}}},{29863, {{40639, 1, 0, 0},{40639, 1, 1, 18}, {0, 0, 0, 0}}},{29863, {{40639, 1, 0, 0},{40639, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29864, {{40640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29865, {{40641, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29866, {{40642, 1, 0, 0},{40642, 1, 1, 16}, {0, 0, 0, 0}}},{29866, {{40642, 1, 0, 0},{40642, 1, 1, 16}, {0, 0, 0, 0}}},{29867, {{40642, 1, 0, 0},{40642, 1, 1, 17}, {0, 0, 0, 0}}},{29867, {{40642, 1, 0, 0},{40642, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{29868, {{40643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29870, {{40644, 1, 0, 0},{40644, 1, 1, 17}, {0, 0, 0, 0}}},{29870, {{40644, 1, 0, 0},{40644, 1, 1, 17}, {0, 0, 0, 0}}},{59263, {{40644, 1, 0, 0},{40644, 1, 1, 18}, {0, 0, 0, 0}}},{59263, {{40644, 1, 0, 0},{40644, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29872, {{40645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29873, {{40646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29877, {{40647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29878, {{40648, 1, 0, 0},{40648, 1, 1, 17}, {0, 0, 0, 0}}},{29878, {{40648, 1, 0, 0},{40648, 1, 1, 17}, {0, 0, 0, 0}}},{29879, {{40648, 1, 0, 0},{40648, 1, 1, 18}, {0, 0, 0, 0}}},{29879, {{40648, 1, 0, 0},{40648, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{29881, {{40650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29882, {{40651, 1, 0, 0},{40651, 1, 1, 17}, {0, 0, 0, 0}}},{29882, {{40651, 1, 0, 0},{40651, 1, 1, 17}, {0, 0, 0, 0}}},{29883, {{40651, 1, 0, 0},{40651, 1, 1, 18}, {0, 0, 0, 0}}},{29883, {{40651, 1, 0, 0},{40651, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{29885, {{40652, 1, 0, 0},{40652, 1, 1, 18}, {0, 0, 0, 0}}},{29885, {{40652, 1, 0, 0},{40652, 1, 1, 18}, {0, 0, 0, 0}}},{29886, {{40652, 1, 0, 0},{40652, 1, 1, 19}, {0, 0, 0, 0}}},{29886, {{40652, 1, 0, 0},{40652, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{29887, {{40653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29888, {{40654, 1, 0, 0},{40654, 1, 1, 17}, {0, 0, 0, 0}}},{29888, {{40654, 1, 0, 0},{40654, 1, 1, 17}, {0, 0, 0, 0}}},{29889, {{40654, 1, 0, 0},{40654, 1, 1, 18}, {0, 0, 0, 0}}},{29889, {{40654, 1, 0, 0},{40654, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29890, {{40655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29891, {{40656, 1, 0, 0},{40656, 1, 1, 17}, {0, 0, 0, 0}}},{29891, {{40656, 1, 0, 0},{40656, 1, 1, 17}, {0, 0, 0, 0}}},{59266, {{40656, 1, 0, 0},{40656, 1, 1, 18}, {0, 0, 0, 0}}},{59266, {{40656, 1, 0, 0},{40656, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29892, {{40657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {2, {{29894, {{40658, 1, 0, 0}, {0, 0, 0, 0}}},{29893, {{40658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29895, {{40659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29896, {{40660, 1, 0, 0},{40660, 1, 1, 17}, {0, 0, 0, 0}}},{29896, {{40660, 1, 0, 0},{40660, 1, 1, 17}, {0, 0, 0, 0}}},{29897, {{40660, 1, 0, 0},{40660, 1, 1, 18}, {0, 0, 0, 0}}},{29897, {{40660, 1, 0, 0},{40660, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29898, {{40661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29899, {{40662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29900, {{40663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29901, {{40664, 1, 0, 0},{40664, 1, 1, 17}, {0, 0, 0, 0}}},{29901, {{40664, 1, 0, 0},{40664, 1, 1, 17}, {0, 0, 0, 0}}},{29902, {{40664, 1, 0, 0},{40664, 1, 1, 18}, {0, 0, 0, 0}}},{29902, {{40664, 1, 0, 0},{40664, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29903, {{40665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29904, {{40666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29905, {{40667, 1, 0, 0},{40667, 1, 1, 18}, {0, 0, 0, 0}}},{29905, {{40667, 1, 0, 0},{40667, 1, 1, 18}, {0, 0, 0, 0}}},{29906, {{40667, 1, 0, 0},{40667, 1, 1, 19}, {0, 0, 0, 0}}},{29906, {{40667, 1, 0, 0},{40667, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{29907, {{40668, 1, 0, 0},{40668, 1, 1, 17}, {0, 0, 0, 0}}},{29907, {{40668, 1, 0, 0},{40668, 1, 1, 17}, {0, 0, 0, 0}}},{29908, {{40668, 1, 0, 0},{40668, 1, 1, 18}, {0, 0, 0, 0}}},{29908, {{40668, 1, 0, 0},{40668, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{29909, {{40669, 1, 0, 0},{40669, 1, 1, 17}, {0, 0, 0, 0}}},{29909, {{40669, 1, 0, 0},{40669, 1, 1, 17}, {0, 0, 0, 0}}},{29910, {{40669, 1, 0, 0},{40669, 1, 1, 18}, {0, 0, 0, 0}}},{29910, {{40669, 1, 0, 0},{40669, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{29911, {{40670, 1, 0, 0},{40670, 1, 1, 17}, {0, 0, 0, 0}}},{29911, {{40670, 1, 0, 0},{40670, 1, 1, 17}, {0, 0, 0, 0}}},{29912, {{40670, 1, 0, 0},{40670, 1, 1, 18}, {0, 0, 0, 0}}},{29912, {{40670, 1, 0, 0},{40670, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29913, {{40671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29914, {{40672, 1, 0, 0},{40672, 1, 1, 17}, {0, 0, 0, 0}}},{29914, {{40672, 1, 0, 0},{40672, 1, 1, 17}, {0, 0, 0, 0}}},{29915, {{40672, 1, 0, 0},{40672, 1, 1, 18}, {0, 0, 0, 0}}},{29915, {{40672, 1, 0, 0},{40672, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{29916, {{40674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29917, {{40675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29918, {{40676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29919, {{40677, 1, 0, 0},{40677, 1, 1, 17}, {0, 0, 0, 0}}},{29919, {{40677, 1, 0, 0},{40677, 1, 1, 17}, {0, 0, 0, 0}}},{29920, {{40677, 1, 0, 0},{40677, 1, 1, 18}, {0, 0, 0, 0}}},{29920, {{40677, 1, 0, 0},{40677, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29921, {{40678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29922, {{40679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29923, {{40680, 1, 0, 0},{40680, 1, 1, 17}, {0, 0, 0, 0}}},{29923, {{40680, 1, 0, 0},{40680, 1, 1, 17}, {0, 0, 0, 0}}},{29924, {{40680, 1, 0, 0},{40680, 1, 1, 18}, {0, 0, 0, 0}}},{29924, {{40680, 1, 0, 0},{40680, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{29925, {{40683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29926, {{40684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29927, {{40685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29928, {{40686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{29929, {{40687, 1, 0, 0},{40687, 1, 1, 17}, {0, 0, 0, 0}}},{29929, {{40687, 1, 0, 0},{40687, 1, 1, 17}, {0, 0, 0, 0}}},{29930, {{40687, 1, 0, 0},{40687, 1, 1, 18}, {0, 0, 0, 0}}},{29930, {{40687, 1, 0, 0},{40687, 1, 1, 18}, {0, 0, 0, 0}}},{29931, {{40687, 1, 0, 0},{40687, 1, 1, 19}, {0, 0, 0, 0}}},{29931, {{40687, 1, 0, 0},{40687, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{29932, {{40688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29933, {{40689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29934, {{40690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29935, {{40691, 1, 0, 0},{40691, 1, 1, 16}, {0, 0, 0, 0}}},{29935, {{40691, 1, 0, 0},{40691, 1, 1, 16}, {0, 0, 0, 0}}},{29936, {{40691, 1, 0, 0},{40691, 1, 1, 17}, {0, 0, 0, 0}}},{29936, {{40691, 1, 0, 0},{40691, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{29937, {{40692, 1, 0, 0},{40692, 1, 1, 17}, {0, 0, 0, 0}}},{29937, {{40692, 1, 0, 0},{40692, 1, 1, 17}, {0, 0, 0, 0}}},{29938, {{40692, 1, 0, 0},{40692, 1, 1, 18}, {0, 0, 0, 0}}},{29938, {{40692, 1, 0, 0},{40692, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29939, {{40693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29940, {{40694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29941, {{40695, 1, 0, 0},{40695, 1, 1, 17}, {0, 0, 0, 0}}},{29941, {{40695, 1, 0, 0},{40695, 1, 1, 17}, {0, 0, 0, 0}}},{29942, {{40695, 1, 0, 0},{40695, 1, 1, 18}, {0, 0, 0, 0}}},{29942, {{40695, 1, 0, 0},{40695, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29943, {{40696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{29944, {{40697, 1, 0, 0},{40697, 1, 1, 17}, {0, 0, 0, 0}}},{29944, {{40697, 1, 0, 0},{40697, 1, 1, 17}, {0, 0, 0, 0}}},{29945, {{40697, 1, 0, 0},{40697, 1, 1, 18}, {0, 0, 0, 0}}},{29945, {{40697, 1, 0, 0},{40697, 1, 1, 18}, {0, 0, 0, 0}}},{58678, {{40697, 1, 0, 0},{40697, 1, 1, 19}, {0, 0, 0, 0}}},{58678, {{40697, 1, 0, 0},{40697, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{29946, {{40698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29947, {{40699, 1, 0, 0},{40699, 1, 1, 17}, {0, 0, 0, 0}}},{29947, {{40699, 1, 0, 0},{40699, 1, 1, 17}, {0, 0, 0, 0}}},{29948, {{40699, 1, 0, 0},{40699, 1, 1, 18}, {0, 0, 0, 0}}},{29948, {{40699, 1, 0, 0},{40699, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{29949, {{40700, 1, 0, 0},{40700, 1, 1, 17}, {0, 0, 0, 0}}},{29949, {{40700, 1, 0, 0},{40700, 1, 1, 17}, {0, 0, 0, 0}}},{29950, {{40700, 1, 0, 0},{40700, 1, 1, 18}, {0, 0, 0, 0}}},{29950, {{40700, 1, 0, 0},{40700, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{29951, {{40701, 1, 0, 0},{40701, 1, 1, 17}, {0, 0, 0, 0}}},{29951, {{40701, 1, 0, 0},{40701, 1, 1, 17}, {0, 0, 0, 0}}},{29952, {{40701, 1, 0, 0},{40701, 1, 1, 18}, {0, 0, 0, 0}}},{29952, {{40701, 1, 0, 0},{40701, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29953, {{40702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29954, {{40703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29955, {{40704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29956, {{40705, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29957, {{40706, 1, 0, 0},{40706, 1, 1, 17}, {0, 0, 0, 0}}},{29957, {{40706, 1, 0, 0},{40706, 1, 1, 17}, {0, 0, 0, 0}}},{29958, {{40706, 1, 0, 0},{40706, 1, 1, 18}, {0, 0, 0, 0}}},{29958, {{40706, 1, 0, 0},{40706, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{29959, {{40707, 1, 0, 0},{40707, 1, 1, 17}, {0, 0, 0, 0}}},{29959, {{40707, 1, 0, 0},{40707, 1, 1, 17}, {0, 0, 0, 0}}},{29960, {{40707, 1, 0, 0},{40707, 1, 1, 18}, {0, 0, 0, 0}}},{29960, {{40707, 1, 0, 0},{40707, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29961, {{40708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29962, {{40709, 1, 0, 0},{40709, 1, 1, 16}, {0, 0, 0, 0}}},{29962, {{40709, 1, 0, 0},{40709, 1, 1, 16}, {0, 0, 0, 0}}},{29963, {{40709, 1, 0, 0},{40709, 1, 1, 17}, {0, 0, 0, 0}}},{29963, {{40709, 1, 0, 0},{40709, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{29964, {{40710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{29965, {{40711, 1, 0, 0},{40711, 1, 1, 17}, {0, 0, 0, 0}}},{29965, {{40711, 1, 0, 0},{40711, 1, 1, 17}, {0, 0, 0, 0}}},{29966, {{40711, 1, 0, 0},{40711, 1, 1, 19}, {0, 0, 0, 0}}},{29966, {{40711, 1, 0, 0},{40711, 1, 1, 19}, {0, 0, 0, 0}}},{29967, {{40711, 1, 0, 0},{40711, 1, 1, 18}, {0, 0, 0, 0}}},{29967, {{40711, 1, 0, 0},{40711, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {10, {{29968, {{40712, 1, 0, 0},{40712, 1, 1, 19}, {0, 0, 0, 0}}},{29968, {{40712, 1, 0, 0},{40712, 1, 1, 19}, {0, 0, 0, 0}}},{29969, {{40712, 1, 0, 0},{40712, 1, 1, 21}, {0, 0, 0, 0}}},{29969, {{40712, 1, 0, 0},{40712, 1, 1, 21}, {0, 0, 0, 0}}},{29970, {{40712, 1, 0, 0},{40712, 1, 1, 23}, {0, 0, 0, 0}}},{29970, {{40712, 1, 0, 0},{40712, 1, 1, 23}, {0, 0, 0, 0}}},{29971, {{40712, 1, 0, 0},{40712, 1, 1, 22}, {0, 0, 0, 0}}},{29971, {{40712, 1, 0, 0},{40712, 1, 1, 22}, {0, 0, 0, 0}}},{29972, {{40712, 1, 0, 0},{40712, 1, 1, 20}, {0, 0, 0, 0}}},{29972, {{40712, 1, 0, 0},{40712, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{29973, {{40713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29974, {{40714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{29975, {{40718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29976, {{40719, 1, 0, 0},{40719, 1, 1, 17}, {0, 0, 0, 0}}},{29976, {{40719, 1, 0, 0},{40719, 1, 1, 17}, {0, 0, 0, 0}}},{29978, {{40719, 1, 0, 0},{40719, 1, 1, 18}, {0, 0, 0, 0}}},{29978, {{40719, 1, 0, 0},{40719, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{29979, {{40720, 1, 0, 0},{40720, 1, 1, 17}, {0, 0, 0, 0}}},{29979, {{40720, 1, 0, 0},{40720, 1, 1, 17}, {0, 0, 0, 0}}},{29980, {{40720, 1, 0, 0},{40720, 1, 1, 18}, {0, 0, 0, 0}}},{29980, {{40720, 1, 0, 0},{40720, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29981, {{40721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29982, {{40722, 1, 0, 0},{40722, 1, 1, 17}, {0, 0, 0, 0}}},{29982, {{40722, 1, 0, 0},{40722, 1, 1, 17}, {0, 0, 0, 0}}},{29983, {{40722, 1, 0, 0},{40722, 1, 1, 18}, {0, 0, 0, 0}}},{29983, {{40722, 1, 0, 0},{40722, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29984, {{40723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29985, {{40724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{29986, {{40725, 1, 0, 0},{40725, 1, 1, 17}, {0, 0, 0, 0}}},{29986, {{40725, 1, 0, 0},{40725, 1, 1, 17}, {0, 0, 0, 0}}},{29987, {{40725, 1, 0, 0},{40725, 1, 1, 18}, {0, 0, 0, 0}}},{29987, {{40725, 1, 0, 0},{40725, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{29988, {{40726, 1, 0, 0},{40726, 1, 1, 17}, {0, 0, 0, 0}}},{29988, {{40726, 1, 0, 0},{40726, 1, 1, 17}, {0, 0, 0, 0}}},{29989, {{40726, 1, 0, 0},{40726, 1, 1, 18}, {0, 0, 0, 0}}},{29989, {{40726, 1, 0, 0},{40726, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29990, {{40727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29991, {{40728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29992, {{40729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29993, {{40730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{29994, {{40731, 1, 0, 0},{40731, 1, 1, 17}, {0, 0, 0, 0}}},{29994, {{40731, 1, 0, 0},{40731, 1, 1, 17}, {0, 0, 0, 0}}},{29995, {{40731, 1, 0, 0},{40731, 1, 1, 19}, {0, 0, 0, 0}}},{29995, {{40731, 1, 0, 0},{40731, 1, 1, 19}, {0, 0, 0, 0}}},{29996, {{40731, 1, 0, 0},{40731, 1, 1, 18}, {0, 0, 0, 0}}},{29996, {{40731, 1, 0, 0},{40731, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29997, {{40732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29998, {{40733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{29999, {{40734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30000, {{40735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30001, {{40736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30002, {{40737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{30004, {{40738, 1, 0, 0},{40738, 1, 1, 17}, {0, 0, 0, 0}}},{30004, {{40738, 1, 0, 0},{40738, 1, 1, 17}, {0, 0, 0, 0}}},{30005, {{40738, 1, 0, 0},{40738, 1, 1, 18}, {0, 0, 0, 0}}},{30005, {{40738, 1, 0, 0},{40738, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{30006, {{40739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30007, {{40740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{30008, {{40741, 1, 0, 0},{40741, 1, 1, 16}, {0, 0, 0, 0}}},{30008, {{40741, 1, 0, 0},{40741, 1, 1, 16}, {0, 0, 0, 0}}},{30009, {{40741, 1, 0, 0},{40741, 1, 1, 17}, {0, 0, 0, 0}}},{30009, {{40741, 1, 0, 0},{40741, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{30010, {{40742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30011, {{40743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{30012, {{40744, 1, 0, 0},{40744, 1, 1, 16}, {0, 0, 0, 0}}},{30012, {{40744, 1, 0, 0},{40744, 1, 1, 16}, {0, 0, 0, 0}}},{30013, {{40744, 1, 0, 0},{40744, 1, 1, 17}, {0, 0, 0, 0}}},{30013, {{40744, 1, 0, 0},{40744, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{30014, {{40745, 1, 0, 0},{40745, 1, 1, 16}, {0, 0, 0, 0}}},{30014, {{40745, 1, 0, 0},{40745, 1, 1, 16}, {0, 0, 0, 0}}},{30015, {{40745, 1, 0, 0},{40745, 1, 1, 17}, {0, 0, 0, 0}}},{30015, {{40745, 1, 0, 0},{40745, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{30016, {{40746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30017, {{40747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{30018, {{40748, 1, 0, 0},{40748, 1, 1, 17}, {0, 0, 0, 0}}},{30018, {{40748, 1, 0, 0},{40748, 1, 1, 17}, {0, 0, 0, 0}}},{30019, {{40748, 1, 0, 0},{40748, 1, 1, 18}, {0, 0, 0, 0}}},{30019, {{40748, 1, 0, 0},{40748, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{30020, {{40749, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{30021, {{40750, 1, 0, 0},{40750, 1, 1, 16}, {0, 0, 0, 0}}},{30021, {{40750, 1, 0, 0},{40750, 1, 1, 16}, {0, 0, 0, 0}}},{30022, {{40750, 1, 0, 0},{40750, 1, 1, 17}, {0, 0, 0, 0}}},{30022, {{40750, 1, 0, 0},{40750, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{30023, {{40751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30024, {{40752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{30025, {{40753, 1, 0, 0},{40753, 1, 1, 17}, {0, 0, 0, 0}}},{30025, {{40753, 1, 0, 0},{40753, 1, 1, 17}, {0, 0, 0, 0}}},{30026, {{40753, 1, 0, 0},{40753, 1, 1, 18}, {0, 0, 0, 0}}},{30026, {{40753, 1, 0, 0},{40753, 1, 1, 18}, {0, 0, 0, 0}}},{30027, {{40753, 1, 0, 0},{40753, 1, 1, 19}, {0, 0, 0, 0}}},{30027, {{40753, 1, 0, 0},{40753, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{30028, {{40754, 1, 0, 0},{40754, 1, 1, 17}, {0, 0, 0, 0}}},{30028, {{40754, 1, 0, 0},{40754, 1, 1, 17}, {0, 0, 0, 0}}},{30029, {{40754, 1, 0, 0},{40754, 1, 1, 18}, {0, 0, 0, 0}}},{30029, {{40754, 1, 0, 0},{40754, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{30030, {{40755, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{30031, {{40756, 1, 0, 0},{40756, 1, 1, 17}, {0, 0, 0, 0}}},{30031, {{40756, 1, 0, 0},{40756, 1, 1, 17}, {0, 0, 0, 0}}},{30032, {{40756, 1, 0, 0},{40756, 1, 1, 18}, {0, 0, 0, 0}}},{30032, {{40756, 1, 0, 0},{40756, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{30033, {{40757, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30034, {{40758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30035, {{40759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30036, {{40760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{30037, {{40761, 1, 0, 0},{40761, 1, 1, 17}, {0, 0, 0, 0}}},{30037, {{40761, 1, 0, 0},{40761, 1, 1, 17}, {0, 0, 0, 0}}},{30038, {{40761, 1, 0, 0},{40761, 1, 1, 18}, {0, 0, 0, 0}}},{30038, {{40761, 1, 0, 0},{40761, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{30039, {{40762, 1, 0, 0},{40762, 1, 1, 17}, {0, 0, 0, 0}}},{30039, {{40762, 1, 0, 0},{40762, 1, 1, 17}, {0, 0, 0, 0}}},{30040, {{40762, 1, 0, 0},{40762, 1, 1, 18}, {0, 0, 0, 0}}},{30040, {{40762, 1, 0, 0},{40762, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{30041, {{40763, 1, 0, 0},{40763, 1, 1, 19}, {0, 0, 0, 0}}},{30041, {{40763, 1, 0, 0},{40763, 1, 1, 19}, {0, 0, 0, 0}}},{30042, {{40763, 1, 0, 0},{40763, 1, 1, 18}, {0, 0, 0, 0}}},{30042, {{40763, 1, 0, 0},{40763, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{30043, {{40764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30044, {{40765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{30045, {{40766, 1, 0, 0},{40766, 1, 1, 17}, {0, 0, 0, 0}}},{30045, {{40766, 1, 0, 0},{40766, 1, 1, 17}, {0, 0, 0, 0}}},{30046, {{40766, 1, 0, 0},{40766, 1, 1, 18}, {0, 0, 0, 0}}},{30046, {{40766, 1, 0, 0},{40766, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{30047, {{40767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30048, {{40768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30049, {{40769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30050, {{40770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30051, {{40771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30052, {{40772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30053, {{40773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30054, {{40774, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30055, {{40775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30056, {{40776, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30057, {{40777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{30058, {{40778, 1, 0, 0},{40778, 1, 1, 19}, {0, 0, 0, 0}}},{30058, {{40778, 1, 0, 0},{40778, 1, 1, 19}, {0, 0, 0, 0}}},{30059, {{40778, 1, 0, 0},{40778, 1, 1, 18}, {0, 0, 0, 0}}},{30059, {{40778, 1, 0, 0},{40778, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {10, {{30060, {{40779, 1, 0, 0},{40779, 1, 1, 18}, {0, 0, 0, 0}}},{30060, {{40779, 1, 0, 0},{40779, 1, 1, 18}, {0, 0, 0, 0}}},{30061, {{40779, 1, 0, 0},{40779, 1, 1, 20}, {0, 0, 0, 0}}},{30061, {{40779, 1, 0, 0},{40779, 1, 1, 20}, {0, 0, 0, 0}}},{30062, {{40779, 1, 0, 0},{40779, 1, 1, 19}, {0, 0, 0, 0}}},{30062, {{40779, 1, 0, 0},{40779, 1, 1, 19}, {0, 0, 0, 0}}},{30063, {{40779, 1, 0, 0},{40779, 1, 1, 22}, {0, 0, 0, 0}}},{30063, {{40779, 1, 0, 0},{40779, 1, 1, 22}, {0, 0, 0, 0}}},{59306, {{40779, 1, 0, 0},{40779, 1, 1, 21}, {0, 0, 0, 0}}},{59306, {{40779, 1, 0, 0},{40779, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{30064, {{40780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{30065, {{40781, 1, 0, 0},{40781, 1, 1, 16}, {0, 0, 0, 0}}},{30065, {{40781, 1, 0, 0},{40781, 1, 1, 16}, {0, 0, 0, 0}}},{30066, {{40781, 1, 0, 0},{40781, 1, 1, 17}, {0, 0, 0, 0}}},{30066, {{40781, 1, 0, 0},{40781, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{30067, {{40782, 1, 0, 0},{40782, 1, 1, 18}, {0, 0, 0, 0}}},{30067, {{40782, 1, 0, 0},{40782, 1, 1, 18}, {0, 0, 0, 0}}},{30068, {{40782, 1, 0, 0},{40782, 1, 1, 19}, {0, 0, 0, 0}}},{30068, {{40782, 1, 0, 0},{40782, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{30069, {{40783, 1, 0, 0},{40783, 1, 1, 17}, {0, 0, 0, 0}}},{30069, {{40783, 1, 0, 0},{40783, 1, 1, 17}, {0, 0, 0, 0}}},{30070, {{40783, 1, 0, 0},{40783, 1, 1, 18}, {0, 0, 0, 0}}},{30070, {{40783, 1, 0, 0},{40783, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{30071, {{40784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30072, {{40786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30073, {{40787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30074, {{40788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30075, {{40789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{30076, {{40790, 1, 0, 0},{40790, 1, 1, 18}, {0, 0, 0, 0}}},{30076, {{40790, 1, 0, 0},{40790, 1, 1, 18}, {0, 0, 0, 0}}},{30077, {{40790, 1, 0, 0},{40790, 1, 1, 17}, {0, 0, 0, 0}}},{30077, {{40790, 1, 0, 0},{40790, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{30078, {{40791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30079, {{40792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30080, {{40793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30081, {{40794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30082, {{40795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30083, {{40796, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30084, {{40797, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30085, {{40798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30086, {{40799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{30087, {{40800, 1, 0, 0},{40800, 1, 1, 18}, {0, 0, 0, 0}}},{30087, {{40800, 1, 0, 0},{40800, 1, 1, 18}, {0, 0, 0, 0}}},{30088, {{40800, 1, 0, 0},{40800, 1, 1, 17}, {0, 0, 0, 0}}},{30088, {{40800, 1, 0, 0},{40800, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{30089, {{40801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30090, {{40802, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30091, {{40803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30092, {{40804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30094, {{40805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30095, {{40806, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{30096, {{40807, 1, 0, 0},{40807, 1, 1, 18}, {0, 0, 0, 0}}},{30096, {{40807, 1, 0, 0},{40807, 1, 1, 18}, {0, 0, 0, 0}}},{30097, {{40807, 1, 0, 0},{40807, 1, 1, 20}, {0, 0, 0, 0}}},{30097, {{40807, 1, 0, 0},{40807, 1, 1, 20}, {0, 0, 0, 0}}},{30098, {{40807, 1, 0, 0},{40807, 1, 1, 19}, {0, 0, 0, 0}}},{30098, {{40807, 1, 0, 0},{40807, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{30099, {{40808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30100, {{40809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30101, {{40810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30102, {{40811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30103, {{40812, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30104, {{40813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30105, {{40814, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30106, {{40815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30107, {{40816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30108, {{40817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30109, {{40818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30110, {{40819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30111, {{40820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30112, {{40821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{30113, {{40822, 1, 0, 0},{40822, 1, 1, 17}, {0, 0, 0, 0}}},{30113, {{40822, 1, 0, 0},{40822, 1, 1, 17}, {0, 0, 0, 0}}},{30114, {{40822, 1, 0, 0},{40822, 1, 1, 18}, {0, 0, 0, 0}}},{30114, {{40822, 1, 0, 0},{40822, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{30115, {{40823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30116, {{40824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30117, {{40825, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30118, {{40826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30119, {{40827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30120, {{40828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30121, {{40829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30122, {{40830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {12, {{30123, {{40845, 1, 0, 0},{40845, 1, 1, 19}, {0, 0, 0, 0}}},{30123, {{40845, 1, 0, 0},{40845, 1, 1, 19}, {0, 0, 0, 0}}},{30124, {{40845, 1, 0, 0},{40845, 1, 1, 23}, {0, 0, 0, 0}}},{30124, {{40845, 1, 0, 0},{40845, 1, 1, 23}, {0, 0, 0, 0}}},{30125, {{40845, 1, 0, 0},{40845, 1, 1, 20}, {0, 0, 0, 0}}},{30125, {{40845, 1, 0, 0},{40845, 1, 1, 20}, {0, 0, 0, 0}}},{30126, {{40845, 1, 0, 0},{40845, 1, 1, 22}, {0, 0, 0, 0}}},{30126, {{40845, 1, 0, 0},{40845, 1, 1, 22}, {0, 0, 0, 0}}},{30127, {{40845, 1, 0, 0},{40845, 1, 1, 21}, {0, 0, 0, 0}}},{30127, {{40845, 1, 0, 0},{40845, 1, 1, 21}, {0, 0, 0, 0}}},{59285, {{40845, 1, 0, 0},{40845, 1, 1, 30}, {0, 0, 0, 0}}},{59285, {{40845, 1, 0, 0},{40845, 1, 1, 30}, {0, 0, 0, 0}}}}},
    {1, {{30131, {{40846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{30132, {{40847, 1, 0, 0},{40847, 1, 1, 17}, {0, 0, 0, 0}}},{30132, {{40847, 1, 0, 0},{40847, 1, 1, 17}, {0, 0, 0, 0}}},{60378, {{40847, 1, 0, 0},{40847, 1, 1, 20}, {0, 0, 0, 0}}},{60378, {{40847, 1, 0, 0},{40847, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {4, {{30135, {{40848, 1, 0, 0},{40848, 1, 1, 17}, {0, 0, 0, 0}}},{30135, {{40848, 1, 0, 0},{40848, 1, 1, 17}, {0, 0, 0, 0}}},{30136, {{40848, 1, 0, 0},{40848, 1, 1, 18}, {0, 0, 0, 0}}},{30136, {{40848, 1, 0, 0},{40848, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{30138, {{40849, 1, 0, 0},{40849, 1, 1, 17}, {0, 0, 0, 0}}},{30138, {{40849, 1, 0, 0},{40849, 1, 1, 17}, {0, 0, 0, 0}}},{30139, {{40849, 1, 0, 0},{40849, 1, 1, 18}, {0, 0, 0, 0}}},{30139, {{40849, 1, 0, 0},{40849, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{30140, {{40850, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{30141, {{40851, 1, 0, 0},{40851, 1, 1, 16}, {0, 0, 0, 0}}},{30141, {{40851, 1, 0, 0},{40851, 1, 1, 16}, {0, 0, 0, 0}}},{30142, {{40851, 1, 0, 0},{40851, 1, 1, 17}, {0, 0, 0, 0}}},{30142, {{40851, 1, 0, 0},{40851, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{30143, {{40852, 1, 0, 0},{40852, 1, 1, 17}, {0, 0, 0, 0}}},{30143, {{40852, 1, 0, 0},{40852, 1, 1, 17}, {0, 0, 0, 0}}},{30144, {{40852, 1, 0, 0},{40852, 1, 1, 18}, {0, 0, 0, 0}}},{30144, {{40852, 1, 0, 0},{40852, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{30145, {{40853, 1, 0, 0},{40853, 1, 1, 17}, {0, 0, 0, 0}}},{30145, {{40853, 1, 0, 0},{40853, 1, 1, 17}, {0, 0, 0, 0}}},{30146, {{40853, 1, 0, 0},{40853, 1, 1, 18}, {0, 0, 0, 0}}},{30146, {{40853, 1, 0, 0},{40853, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{30147, {{40854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30148, {{40855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{30150, {{40856, 1, 0, 0},{40856, 1, 1, 16}, {0, 0, 0, 0}}},{30150, {{40856, 1, 0, 0},{40856, 1, 1, 16}, {0, 0, 0, 0}}},{30151, {{40856, 1, 0, 0},{40856, 1, 1, 17}, {0, 0, 0, 0}}},{30151, {{40856, 1, 0, 0},{40856, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {12, {{30154, {{40860, 1, 0, 0},{40860, 1, 1, 20}, {0, 0, 0, 0}}},{30154, {{40860, 1, 0, 0},{40860, 1, 1, 20}, {0, 0, 0, 0}}},{30155, {{40860, 1, 0, 0},{40860, 1, 1, 18}, {0, 0, 0, 0}}},{30155, {{40860, 1, 0, 0},{40860, 1, 1, 18}, {0, 0, 0, 0}}},{30156, {{40860, 1, 0, 0},{40860, 1, 1, 22}, {0, 0, 0, 0}}},{30156, {{40860, 1, 0, 0},{40860, 1, 1, 22}, {0, 0, 0, 0}}},{30157, {{40860, 1, 0, 0},{40860, 1, 1, 19}, {0, 0, 0, 0}}},{30157, {{40860, 1, 0, 0},{40860, 1, 1, 19}, {0, 0, 0, 0}}},{30158, {{40860, 1, 0, 0},{40860, 1, 1, 21}, {0, 0, 0, 0}}},{30158, {{40860, 1, 0, 0},{40860, 1, 1, 21}, {0, 0, 0, 0}}},{59288, {{40860, 1, 0, 0},{40860, 1, 1, 25}, {0, 0, 0, 0}}},{59288, {{40860, 1, 0, 0},{40860, 1, 1, 25}, {0, 0, 0, 0}}}}},
    {8, {{30159, {{40861, 1, 0, 0},{40861, 1, 1, 19}, {0, 0, 0, 0}}},{30159, {{40861, 1, 0, 0},{40861, 1, 1, 19}, {0, 0, 0, 0}}},{30160, {{40861, 1, 0, 0},{40861, 1, 1, 18}, {0, 0, 0, 0}}},{30160, {{40861, 1, 0, 0},{40861, 1, 1, 18}, {0, 0, 0, 0}}},{30161, {{40861, 1, 0, 0},{40861, 1, 1, 20}, {0, 0, 0, 0}}},{30161, {{40861, 1, 0, 0},{40861, 1, 1, 20}, {0, 0, 0, 0}}},{60002, {{40861, 1, 0, 0},{40861, 1, 1, 21}, {0, 0, 0, 0}}},{60002, {{40861, 1, 0, 0},{40861, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{30163, {{40862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30164, {{40864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30165, {{40865, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30166, {{40866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30168, {{40867, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30169, {{40868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30170, {{40869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{30171, {{40876, 1, 0, 0},{40876, 1, 1, 16}, {0, 0, 0, 0}}},{30171, {{40876, 1, 0, 0},{40876, 1, 1, 16}, {0, 0, 0, 0}}},{30172, {{40876, 1, 0, 0},{40876, 1, 1, 17}, {0, 0, 0, 0}}},{30172, {{40876, 1, 0, 0},{40876, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{30174, {{40877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57203, {{40881, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30177, {{40884, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57273, {{40889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59459, {{40892, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30179, {{40893, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30180, {{40894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60118, {{40895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59215, {{40898, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2838, {{40899, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30181, {{40900, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{18788, {{40902, 1, 0, 0},{40902, 1, 1, 17}, {0, 0, 0, 0}}},{18788, {{40902, 1, 0, 0},{40902, 1, 1, 17}, {0, 0, 0, 0}}},{30182, {{40902, 1, 0, 0},{40902, 1, 1, 18}, {0, 0, 0, 0}}},{30182, {{40902, 1, 0, 0},{40902, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60171, {{40907, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57449, {{40942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60040, {{40943, 1, 0, 0}, {0, 0, 0, 0}}}}}
};

static const MJMappingSet chunked_mj_mappings_00f91d_00f970[84] = {
    {1, {{30183, {{27396, 1, 0, 0},{27396, 1, 1, 0},{27396, 1, 1, 19},{63773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30184, {{24266, 1, 0, 0},{24266, 1, 1, 0},{24266, 1, 1, 20},{63784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30185, {{26391, 1, 0, 0},{26391, 1, 1, 0},{26391, 1, 1, 22},{63785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30186, {{34396, 1, 0, 0},{34396, 1, 1, 0},{34396, 1, 1, 19},{63798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30187, {{27578, 1, 0, 0},{27578, 1, 1, 0},{27578, 1, 1, 19},{63856, 1, 0, 0}, {0, 0, 0, 0}}}}}
};

static const MJMappingSet chunked_mj_mappings_00f9d0_00fa6d[158] = {
    {1, {{30188, {{39006, 1, 0, 0},{39006, 1, 1, 0},{39006, 1, 1, 19},{63952, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30189, {{38534, 1, 0, 0},{38534, 1, 1, 0},{38534, 1, 1, 22},{63964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{30191, {{64014, 1, 0, 0},{64014, 1, 1, 17}, {0, 0, 0, 0}}},{30191, {{64014, 1, 0, 0},{64014, 1, 1, 17}, {0, 0, 0, 0}}},{30192, {{64014, 1, 0, 0},{64014, 1, 1, 18}, {0, 0, 0, 0}}},{30192, {{64014, 1, 0, 0},{64014, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{30193, {{64015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30194, {{22618, 1, 0, 0},{22618, 1, 1, 0},{22618, 1, 1, 21},{64016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{30196, {{64017, 1, 0, 0},{64017, 1, 1, 18}, {0, 0, 0, 0}}},{30196, {{64017, 1, 0, 0},{64017, 1, 1, 18}, {0, 0, 0, 0}}},{30197, {{64017, 1, 0, 0},{64017, 1, 1, 19}, {0, 0, 0, 0}}},{30197, {{64017, 1, 0, 0},{64017, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{30199, {{26228, 1, 0, 0},{26228, 1, 1, 0},{26228, 1, 1, 19},{64018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{30200, {{64019, 1, 0, 0},{64019, 1, 1, 17}, {0, 0, 0, 0}}},{30200, {{64019, 1, 0, 0},{64019, 1, 1, 17}, {0, 0, 0, 0}}},{30201, {{64019, 1, 0, 0},{64019, 1, 1, 18}, {0, 0, 0, 0}}},{30201, {{64019, 1, 0, 0},{64019, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{30202, {{64020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30205, {{20958, 1, 0, 0},{20958, 1, 1, 0},{20958, 1, 1, 26},{64021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30208, {{29482, 1, 0, 0},{29482, 1, 1, 0},{29482, 1, 1, 19},{64022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30209, {{30410, 1, 0, 0},{30410, 1, 1, 0},{30410, 1, 1, 19},{64023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30210, {{31036, 1, 0, 0},{31036, 1, 1, 0},{31036, 1, 1, 19},{64024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30211, {{31070, 1, 0, 0},{31070, 1, 1, 0},{31070, 1, 1, 19},{64025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30212, {{31077, 1, 0, 0},{31077, 1, 1, 0},{31077, 1, 1, 19},{64026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30213, {{31119, 1, 0, 0},{31119, 1, 1, 0},{31119, 1, 1, 19},{64027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30214, {{38742, 1, 0, 0},{38742, 1, 1, 0},{38742, 1, 1, 19},{64028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30215, {{31934, 1, 0, 0},{31934, 1, 1, 0},{31934, 1, 1, 19},{64029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30216, {{32701, 1, 0, 0},{32701, 1, 1, 0},{32701, 1, 1, 19},{64030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{30217, {{64031, 1, 0, 0},{64031, 1, 1, 17}, {0, 0, 0, 0}}},{30217, {{64031, 1, 0, 0},{64031, 1, 1, 17}, {0, 0, 0, 0}}},{30218, {{64031, 1, 0, 0},{64031, 1, 1, 19}, {0, 0, 0, 0}}},{30218, {{64031, 1, 0, 0},{64031, 1, 1, 19}, {0, 0, 0, 0}}},{30219, {{64031, 1, 0, 0},{64031, 1, 1, 18}, {0, 0, 0, 0}}},{30219, {{64031, 1, 0, 0},{64031, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{30224, {{34322, 1, 0, 0},{34322, 1, 1, 0},{34322, 1, 1, 21},{64032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30225, {{64033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30226, {{35576, 1, 0, 0},{35576, 1, 1, 0},{35576, 1, 1, 19},{64034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30227, {{64035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {12, {{30228, {{64036, 1, 0, 0},{64036, 1, 1, 23}, {0, 0, 0, 0}}},{30228, {{64036, 1, 0, 0},{64036, 1, 1, 23}, {0, 0, 0, 0}}},{30229, {{64036, 1, 0, 0},{64036, 1, 1, 19}, {0, 0, 0, 0}}},{30229, {{64036, 1, 0, 0},{64036, 1, 1, 19}, {0, 0, 0, 0}}},{30230, {{64036, 1, 0, 0},{64036, 1, 1, 20}, {0, 0, 0, 0}}},{30230, {{64036, 1, 0, 0},{64036, 1, 1, 20}, {0, 0, 0, 0}}},{30231, {{64036, 1, 0, 0},{64036, 1, 1, 21}, {0, 0, 0, 0}}},{30231, {{64036, 1, 0, 0},{64036, 1, 1, 21}, {0, 0, 0, 0}}},{30232, {{64036, 1, 0, 0},{64036, 1, 1, 22}, {0, 0, 0, 0}}},{30232, {{64036, 1, 0, 0},{64036, 1, 1, 22}, {0, 0, 0, 0}}},{30233, {{64036, 1, 0, 0},{64036, 1, 1, 18}, {0, 0, 0, 0}}},{30233, {{64036, 1, 0, 0},{64036, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{30234, {{36920, 1, 0, 0},{36920, 1, 1, 0},{36920, 1, 1, 24},{64037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30235, {{37117, 1, 0, 0},{37117, 1, 1, 0},{37117, 1, 1, 19},{64038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{30236, {{64039, 1, 0, 0},{64039, 1, 1, 17}, {0, 0, 0, 0}}},{30236, {{64039, 1, 0, 0},{64039, 1, 1, 17}, {0, 0, 0, 0}}},{30237, {{64039, 1, 0, 0},{64039, 1, 1, 18}, {0, 0, 0, 0}}},{30237, {{64039, 1, 0, 0},{64039, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{30238, {{64040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30239, {{64041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30240, {{39151, 1, 0, 0},{39151, 1, 1, 0},{39151, 1, 1, 20},{64042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30241, {{39164, 1, 0, 0},{39164, 1, 1, 0},{39164, 1, 1, 19},{64043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30242, {{39208, 1, 0, 0},{39208, 1, 1, 0},{39208, 1, 1, 19},{64044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30243, {{40372, 1, 0, 0},{40372, 1, 1, 0},{40372, 1, 1, 19},{64045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30244, {{20398, 1, 0, 0},{20398, 1, 1, 0},{20398, 1, 1, 19},{64048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30245, {{20711, 1, 0, 0},{20711, 1, 1, 0},{20711, 1, 1, 21},{64049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30246, {{20813, 1, 0, 0},{20813, 1, 1, 0},{20813, 1, 1, 20},{64050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30247, {{21193, 1, 0, 0},{21193, 1, 1, 0},{21193, 1, 1, 19},{64051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30248, {{21220, 1, 0, 0},{21220, 1, 1, 0},{21220, 1, 1, 19},{64052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30249, {{21329, 1, 0, 0},{21329, 1, 1, 0},{21329, 1, 1, 19},{64053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30250, {{21917, 1, 0, 0},{21917, 1, 1, 0},{21917, 1, 1, 19},{64054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30251, {{22022, 1, 0, 0},{22022, 1, 1, 0},{22022, 1, 1, 19},{64055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30252, {{22120, 1, 0, 0},{22120, 1, 1, 0},{22120, 1, 1, 19},{64056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30253, {{22592, 1, 0, 0},{22592, 1, 1, 0},{22592, 1, 1, 19},{64057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30254, {{22696, 1, 0, 0},{22696, 1, 1, 0},{22696, 1, 1, 19},{64058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30255, {{23652, 1, 0, 0},{23652, 1, 1, 0},{23652, 1, 1, 19},{64059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30256, {{23662, 1, 0, 0},{23662, 1, 1, 0},{23662, 1, 1, 19},{64060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30257, {{24724, 1, 0, 0},{24724, 1, 1, 0},{24724, 1, 1, 19},{64061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30258, {{24936, 1, 0, 0},{24936, 1, 1, 0},{24936, 1, 1, 24},{64062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30259, {{24974, 1, 0, 0},{24974, 1, 1, 0},{24974, 1, 1, 19},{64063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30260, {{25074, 1, 0, 0},{25074, 1, 1, 0},{25074, 1, 1, 21},{64064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30262, {{25935, 1, 0, 0},{25935, 1, 1, 0},{25935, 1, 1, 19},{64065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30263, {{26082, 1, 0, 0},{26082, 1, 1, 0},{26082, 1, 1, 20},{64066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30264, {{26257, 1, 0, 0},{26257, 1, 1, 0},{26257, 1, 1, 19},{64067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30265, {{26757, 1, 0, 0},{26757, 1, 1, 0},{26757, 1, 1, 19},{64068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30266, {{28023, 1, 0, 0},{28023, 1, 1, 0},{28023, 1, 1, 19},{64069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30267, {{28186, 1, 0, 0},{28186, 1, 1, 0},{28186, 1, 1, 19},{64070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30268, {{28450, 1, 0, 0},{28450, 1, 1, 0},{28450, 1, 1, 19},{64071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30269, {{29038, 1, 0, 0},{29038, 1, 1, 0},{29038, 1, 1, 19},{64072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30270, {{29227, 1, 0, 0},{29227, 1, 1, 0},{29227, 1, 1, 18},{64073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30271, {{29730, 1, 0, 0},{29730, 1, 1, 0},{29730, 1, 1, 21},{64074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30272, {{30865, 1, 0, 0},{30865, 1, 1, 0},{30865, 1, 1, 19},{64075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30274, {{31038, 1, 0, 0},{31038, 1, 1, 0},{31038, 1, 1, 19},{64076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30275, {{31049, 1, 0, 0},{31049, 1, 1, 0},{31049, 1, 1, 19},{64077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30276, {{31048, 1, 0, 0},{31048, 1, 1, 0},{31048, 1, 1, 19},{64078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30277, {{31056, 1, 0, 0},{31056, 1, 1, 0},{31056, 1, 1, 19},{64079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30278, {{31062, 1, 0, 0},{31062, 1, 1, 0},{31062, 1, 1, 19},{64080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30279, {{31069, 1, 0, 0},{31069, 1, 1, 0},{31069, 1, 1, 19},{64081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30280, {{31117, 1, 0, 0},{31117, 1, 1, 0},{31117, 1, 1, 19},{64082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30281, {{31118, 1, 0, 0},{31118, 1, 1, 0},{31118, 1, 1, 19},{64083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30282, {{31296, 1, 0, 0},{31296, 1, 1, 0},{31296, 1, 1, 19},{64084, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30283, {{31361, 1, 0, 0},{31361, 1, 1, 0},{31361, 1, 1, 19},{64085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30285, {{31680, 1, 0, 0},{31680, 1, 1, 0},{31680, 1, 1, 21},{64086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30286, {{32244, 1, 0, 0},{32244, 1, 1, 1},{32244, 1, 1, 19},{64087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30287, {{32265, 1, 0, 0},{32265, 1, 1, 0},{32265, 1, 1, 19},{64088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30288, {{32321, 1, 0, 0},{32321, 1, 1, 0},{32321, 1, 1, 19},{64089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30289, {{32626, 1, 0, 0},{32626, 1, 1, 0},{32626, 1, 1, 19},{64090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30290, {{32773, 1, 0, 0},{32773, 1, 1, 0},{32773, 1, 1, 19},{64091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30291, {{33261, 1, 0, 0},{33261, 1, 1, 0},{33261, 1, 1, 19},{64092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30292, {{33401, 1, 0, 0},{33401, 1, 1, 0},{33401, 1, 1, 19},{64093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30293, {{33401, 1, 0, 0},{33401, 1, 1, 1},{33401, 1, 1, 20},{64094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30294, {{33879, 1, 0, 0},{33879, 1, 1, 0},{33879, 1, 1, 20},{64095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30295, {{35088, 1, 0, 0},{35088, 1, 1, 0},{35088, 1, 1, 19},{64096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30296, {{35222, 1, 0, 0},{35222, 1, 1, 0},{35222, 1, 1, 19},{64097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30297, {{35585, 1, 0, 0},{35585, 1, 1, 0},{35585, 1, 1, 19},{64098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30298, {{35641, 1, 0, 0},{35641, 1, 1, 0},{35641, 1, 1, 20},{64099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30299, {{36051, 1, 0, 0},{36051, 1, 1, 0},{36051, 1, 1, 19},{64100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30301, {{36104, 1, 0, 0},{36104, 1, 1, 0},{36104, 1, 1, 19},{64101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30302, {{36790, 1, 0, 0},{36790, 1, 1, 0},{36790, 1, 1, 18},{64102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30303, {{36920, 1, 0, 0},{36920, 1, 1, 1},{36920, 1, 1, 23},{64103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30304, {{38627, 1, 0, 0},{38627, 1, 1, 0},{38627, 1, 1, 20},{64104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30305, {{38911, 1, 0, 0},{38911, 1, 1, 0},{38911, 1, 1, 24},{64105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30307, {{38971, 1, 0, 0},{38971, 1, 1, 0},{38971, 1, 1, 19},{64106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30308, {{24693, 1, 0, 0},{24693, 1, 1, 0},{24693, 1, 1, 19},{64107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30309, {{64108, 1, 0, 0},{148206, 1, 0, 0},{148206, 1, 1, 0},{148206, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{30310, {{33304, 1, 0, 0},{33304, 1, 1, 0},{33304, 1, 1, 19},{64109, 1, 0, 0}, {0, 0, 0, 0}}}}}
};

static const MJMappingSet chunked_mj_mappings_020000_02a6d6[42711] = {
    {6, {{30312, {{131072, 1, 0, 0},{131072, 1, 1, 16}, {0, 0, 0, 0}}},{30312, {{131072, 1, 0, 0},{131072, 1, 1, 16}, {0, 0, 0, 0}}},{30313, {{131072, 1, 0, 0},{131072, 1, 1, 17}, {0, 0, 0, 0}}},{30313, {{131072, 1, 0, 0},{131072, 1, 1, 17}, {0, 0, 0, 0}}},{56848, {{131072, 1, 0, 0},{131072, 1, 1, 18}, {0, 0, 0, 0}}},{56848, {{131072, 1, 0, 0},{131072, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{30314, {{131073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30315, {{131075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30316, {{131077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30317, {{131079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30318, {{131080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{30319, {{131083, 1, 0, 0},{131083, 1, 1, 19}, {0, 0, 0, 0}}},{30319, {{131083, 1, 0, 0},{131083, 1, 1, 19}, {0, 0, 0, 0}}},{30320, {{131083, 1, 0, 0},{131083, 1, 1, 17}, {0, 0, 0, 0}}},{30320, {{131083, 1, 0, 0},{131083, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30322, {{131085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30323, {{131088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30324, {{131090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30325, {{131091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30326, {{131092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30328, {{131096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30329, {{131097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30330, {{131098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30331, {{131099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30333, {{131104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30334, {{131105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30335, {{131106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30336, {{131110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30337, {{131113, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30338, {{131116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30339, {{131117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30340, {{131118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30341, {{131127, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30342, {{131128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30343, {{131129, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30344, {{131131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{30345, {{131137, 1, 0, 0},{131137, 1, 1, 16}, {0, 0, 0, 0}}},{30345, {{131137, 1, 0, 0},{131137, 1, 1, 16}, {0, 0, 0, 0}}},{30346, {{131137, 1, 0, 0},{131137, 1, 1, 18}, {0, 0, 0, 0}}},{30346, {{131137, 1, 0, 0},{131137, 1, 1, 18}, {0, 0, 0, 0}}},{56847, {{131137, 1, 0, 0},{131137, 1, 1, 19}, {0, 0, 0, 0}}},{56847, {{131137, 1, 0, 0},{131137, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30347, {{131139, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{58590, {{131141, 1, 0, 0},{131141, 1, 1, 16}, {0, 0, 0, 0}}},{58590, {{131141, 1, 0, 0},{131141, 1, 1, 16}, {0, 0, 0, 0}}},{58592, {{131141, 1, 0, 0},{131141, 1, 1, 17}, {0, 0, 0, 0}}},{58592, {{131141, 1, 0, 0},{131141, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30348, {{131144, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30349, {{131145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{56998, {{131146, 1, 0, 0},{131146, 1, 1, 16}, {0, 0, 0, 0}}},{56998, {{131146, 1, 0, 0},{131146, 1, 1, 16}, {0, 0, 0, 0}}},{57285, {{131146, 1, 0, 0},{131146, 1, 1, 17}, {0, 0, 0, 0}}},{57285, {{131146, 1, 0, 0},{131146, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30350, {{131157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30351, {{131159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30352, {{131168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30353, {{131169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30354, {{131172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30355, {{131173, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30356, {{131174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30357, {{131175, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30358, {{131177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30359, {{131181, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30360, {{131182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30361, {{131184, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30362, {{131185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30364, {{131188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30365, {{131191, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30366, {{131196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30367, {{131197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30368, {{131198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30369, {{131199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30370, {{131202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30371, {{131204, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30372, {{131206, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30373, {{131207, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30374, {{131208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30375, {{131209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30376, {{131213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30377, {{131215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30378, {{131216, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30379, {{131220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30380, {{131221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30381, {{131222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30382, {{131228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30383, {{131230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30384, {{131231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30385, {{131232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{30386, {{131234, 1, 0, 0},{131234, 1, 1, 18}, {0, 0, 0, 0}}},{30386, {{131234, 1, 0, 0},{131234, 1, 1, 18}, {0, 0, 0, 0}}},{30387, {{131234, 1, 0, 0},{131234, 1, 1, 17}, {0, 0, 0, 0}}},{30387, {{131234, 1, 0, 0},{131234, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{30388, {{131235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30389, {{131236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30390, {{131237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30391, {{131239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30392, {{131240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30393, {{131241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30394, {{131244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30395, {{131246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30396, {{131247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30397, {{131249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30398, {{131250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30399, {{131251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30400, {{131253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30401, {{131254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30402, {{131255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30403, {{131256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30404, {{131257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30405, {{131258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30406, {{131260, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30407, {{131261, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30408, {{131262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30409, {{131263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30410, {{131266, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30411, {{131267, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30412, {{131268, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30413, {{131273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30414, {{131275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30416, {{131279, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30417, {{131280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30418, {{131281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30419, {{131282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30420, {{131285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30421, {{131286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30422, {{131294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30424, {{131295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30425, {{131296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30426, {{131297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{30427, {{131300, 1, 0, 0},{131300, 1, 1, 16}, {0, 0, 0, 0}}},{30427, {{131300, 1, 0, 0},{131300, 1, 1, 16}, {0, 0, 0, 0}}},{57398, {{131300, 1, 0, 0},{131300, 1, 1, 17}, {0, 0, 0, 0}}},{57398, {{131300, 1, 0, 0},{131300, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{30428, {{131301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30429, {{131303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58083, {{131307, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30430, {{131309, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30431, {{131312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30432, {{131314, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30433, {{131315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30434, {{131318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30435, {{131322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30436, {{131324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30437, {{131325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30438, {{131326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30439, {{131332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30440, {{131333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30441, {{131335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30442, {{131336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30443, {{131337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30444, {{131338, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30445, {{131339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{30446, {{131340, 1, 0, 0},{131340, 1, 1, 16}, {0, 0, 0, 0}}},{30446, {{131340, 1, 0, 0},{131340, 1, 1, 16}, {0, 0, 0, 0}}},{30447, {{131340, 1, 0, 0},{131340, 1, 1, 17}, {0, 0, 0, 0}}},{30447, {{131340, 1, 0, 0},{131340, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{30448, {{131341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30449, {{131342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30450, {{131343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30451, {{131345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30452, {{131346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30453, {{131347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30454, {{131348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30455, {{131349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30456, {{131350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30457, {{131351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30458, {{131353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30459, {{131354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30460, {{131355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30461, {{131357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30462, {{131358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30463, {{131359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30464, {{131360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30465, {{131361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{30466, {{131362, 1, 0, 0},{131362, 1, 1, 16}, {0, 0, 0, 0}}},{30466, {{131362, 1, 0, 0},{131362, 1, 1, 16}, {0, 0, 0, 0}}},{30467, {{131362, 1, 0, 0},{131362, 1, 1, 17}, {0, 0, 0, 0}}},{30467, {{131362, 1, 0, 0},{131362, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{30468, {{131363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30469, {{131364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30470, {{131368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30471, {{131372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30474, {{131373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30475, {{131374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30476, {{131376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30477, {{131377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30478, {{131380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30479, {{131381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30480, {{131383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30481, {{131385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30482, {{131391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30483, {{131394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30484, {{131395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30485, {{131397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30486, {{131400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30487, {{131404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30488, {{131411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30489, {{131412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30490, {{131414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30491, {{131415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30492, {{131416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30493, {{131417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30494, {{131418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{30496, {{131422, 1, 0, 0},{131422, 1, 1, 16}, {0, 0, 0, 0}}},{30496, {{131422, 1, 0, 0},{131422, 1, 1, 16}, {0, 0, 0, 0}}},{30497, {{131422, 1, 0, 0},{131422, 1, 1, 17}, {0, 0, 0, 0}}},{30497, {{131422, 1, 0, 0},{131422, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{30498, {{131423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30499, {{131424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30500, {{131425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30501, {{131426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30502, {{131430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30503, {{131433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30504, {{131434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30505, {{131435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30506, {{131436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30507, {{131440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30508, {{131441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30510, {{131447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30509, {{131449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30511, {{131457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30512, {{131458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30513, {{131462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30514, {{131463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30515, {{131465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30516, {{131466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30517, {{131467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30518, {{131468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30519, {{131470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30520, {{131471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30521, {{131472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30522, {{131473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30523, {{131475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30525, {{131477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30526, {{131481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30527, {{131482, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30528, {{131483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30529, {{131484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30530, {{131485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30531, {{131487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30532, {{131490, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30533, {{131491, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30534, {{131492, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30535, {{131494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30536, {{131497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30537, {{131498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30538, {{131501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30539, {{131502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30540, {{131503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30541, {{131504, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30542, {{131505, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30543, {{131506, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30544, {{131508, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30545, {{131509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30546, {{131510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30547, {{131512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30548, {{131513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30549, {{131514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{30550, {{131515, 1, 0, 0},{131515, 1, 1, 17}, {0, 0, 0, 0}}},{30550, {{131515, 1, 0, 0},{131515, 1, 1, 17}, {0, 0, 0, 0}}},{30551, {{131515, 1, 0, 0},{131515, 1, 1, 16}, {0, 0, 0, 0}}},{30551, {{131515, 1, 0, 0},{131515, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{30552, {{131516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30554, {{131522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56912, {{131530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30556, {{131531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59315, {{131533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30557, {{131538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30558, {{131543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30559, {{131544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30560, {{131545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30561, {{131546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30562, {{131547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30563, {{131548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30564, {{131549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30565, {{131550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30566, {{131566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30567, {{131567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30568, {{131569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30569, {{131570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30570, {{131571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30571, {{131572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30572, {{131573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{30573, {{131582, 1, 0, 0},{131582, 1, 1, 16}, {0, 0, 0, 0}}},{30573, {{131582, 1, 0, 0},{131582, 1, 1, 16}, {0, 0, 0, 0}}},{56920, {{131582, 1, 0, 0},{131582, 1, 1, 17}, {0, 0, 0, 0}}},{56920, {{131582, 1, 0, 0},{131582, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{30574, {{131583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30575, {{131584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30576, {{131585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30577, {{131587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30578, {{131588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30579, {{131589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30580, {{131590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30581, {{131591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30582, {{131592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30583, {{131593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30584, {{131594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30585, {{131595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30586, {{131596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30587, {{131597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30588, {{131598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30589, {{131599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30590, {{131600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30591, {{131601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30592, {{131603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30600, {{131606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30594, {{131618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30595, {{131620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30596, {{131622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30597, {{131623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30598, {{131624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30599, {{131626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30602, {{131628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30603, {{131629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30604, {{131630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30605, {{131631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30606, {{131632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30607, {{131633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30608, {{131634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30609, {{131635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30610, {{131636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30611, {{131637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30612, {{131638, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{30613, {{131639, 1, 0, 0},{131639, 1, 1, 16}, {0, 0, 0, 0}}},{30613, {{131639, 1, 0, 0},{131639, 1, 1, 16}, {0, 0, 0, 0}}},{30614, {{131639, 1, 0, 0},{131639, 1, 1, 17}, {0, 0, 0, 0}}},{30614, {{131639, 1, 0, 0},{131639, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{30615, {{131640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30616, {{131641, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30617, {{131642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30618, {{131643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30619, {{131644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30620, {{131645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30621, {{131646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30622, {{131647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59321, {{131660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30625, {{131663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30627, {{131666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30628, {{131668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{30629, {{131669, 1, 0, 0},{131669, 1, 1, 16}, {0, 0, 0, 0}}},{30629, {{131669, 1, 0, 0},{131669, 1, 1, 16}, {0, 0, 0, 0}}},{30630, {{131669, 1, 0, 0},{131669, 1, 1, 17}, {0, 0, 0, 0}}},{30630, {{131669, 1, 0, 0},{131669, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{30631, {{131670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30633, {{131676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30634, {{131681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30635, {{131682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30636, {{131683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30637, {{131684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30638, {{131685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30639, {{131686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30640, {{131687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30641, {{131688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30642, {{131689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30643, {{131690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30644, {{131691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30645, {{131692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30646, {{131693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30647, {{131694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30649, {{131699, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30651, {{131722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30652, {{131723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30653, {{131724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30654, {{131727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30655, {{131729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30656, {{131734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30657, {{131746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30659, {{131747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30660, {{131748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30661, {{131749, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30662, {{131750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30663, {{131751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30665, {{131752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30666, {{131753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30667, {{131754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30668, {{131755, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30669, {{131756, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30670, {{131757, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30671, {{131758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30672, {{131760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30673, {{131761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30674, {{131762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30675, {{131763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30676, {{131764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30677, {{131765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30678, {{131766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30679, {{131767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30680, {{131768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30681, {{131769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30682, {{131770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30683, {{131771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30684, {{131774, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30686, {{131785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30688, {{131793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30689, {{131794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30690, {{131795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30691, {{131797, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30692, {{131799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30694, {{131803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30695, {{131804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30696, {{131805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30697, {{131806, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30698, {{131809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30699, {{131810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30700, {{131811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30701, {{131815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30702, {{131816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30703, {{131817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30704, {{131818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30705, {{131819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{30706, {{131820, 1, 0, 0},{131820, 1, 1, 17}, {0, 0, 0, 0}}},{30706, {{131820, 1, 0, 0},{131820, 1, 1, 17}, {0, 0, 0, 0}}},{30707, {{131820, 1, 0, 0},{131820, 1, 1, 16}, {0, 0, 0, 0}}},{30707, {{131820, 1, 0, 0},{131820, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{30708, {{131821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30709, {{131822, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30710, {{131823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30711, {{131824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30712, {{131825, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30713, {{131826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30714, {{131827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30715, {{131828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30716, {{131829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30717, {{131830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30718, {{131831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30719, {{131832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30720, {{131833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30721, {{131834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30722, {{131835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30723, {{131836, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30724, {{131838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30725, {{131839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30762, {{131840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30727, {{131848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30728, {{131849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30729, {{131861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30730, {{131863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{30731, {{131864, 1, 0, 0},{131864, 1, 1, 17}, {0, 0, 0, 0}}},{30731, {{131864, 1, 0, 0},{131864, 1, 1, 17}, {0, 0, 0, 0}}},{30732, {{131864, 1, 0, 0},{131864, 1, 1, 16}, {0, 0, 0, 0}}},{30732, {{131864, 1, 0, 0},{131864, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30733, {{131866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30734, {{131868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30735, {{131870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30736, {{131871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30737, {{131873, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30738, {{131874, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30739, {{131881, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30740, {{131882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{30741, {{131883, 1, 0, 0},{131883, 1, 1, 18}, {0, 0, 0, 0}}},{30741, {{131883, 1, 0, 0},{131883, 1, 1, 18}, {0, 0, 0, 0}}},{30742, {{131883, 1, 0, 0},{131883, 1, 1, 17}, {0, 0, 0, 0}}},{30742, {{131883, 1, 0, 0},{131883, 1, 1, 17}, {0, 0, 0, 0}}},{68074, {{131883, 1, 0, 0},{131883, 1, 1, 19}, {0, 0, 0, 0}}},{68074, {{131883, 1, 0, 0},{131883, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{30743, {{131884, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30744, {{131885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30745, {{131886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30746, {{131887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30747, {{131888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30748, {{131889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30749, {{131890, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30750, {{131891, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30751, {{131892, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30752, {{131893, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30753, {{131894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30754, {{131895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30755, {{131896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30756, {{131897, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30758, {{131898, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30759, {{131902, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30760, {{131903, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30761, {{131904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30763, {{131907, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30764, {{131908, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30765, {{131927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30766, {{131928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30767, {{131930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30768, {{131931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30769, {{131933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30770, {{131934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30771, {{131935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30772, {{131938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30773, {{131943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30774, {{131944, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30775, {{131945, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30776, {{131946, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30777, {{131947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30778, {{131948, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30779, {{131950, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30780, {{131951, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30781, {{131952, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30782, {{131953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30783, {{131954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30784, {{131955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30785, {{131956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30786, {{131957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30787, {{131958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30788, {{131959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30789, {{131960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30790, {{131961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30791, {{131962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30792, {{131963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30793, {{131969, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30794, {{131971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30795, {{131992, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30796, {{131994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30797, {{131995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30798, {{132001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30799, {{132002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30800, {{132007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30801, {{132013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30802, {{132014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30803, {{132015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30804, {{132016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30805, {{132017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30806, {{132018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30807, {{132019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30808, {{132022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30809, {{132023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30810, {{132024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{30811, {{132025, 1, 0, 0},{132025, 1, 1, 16}, {0, 0, 0, 0}}},{30811, {{132025, 1, 0, 0},{132025, 1, 1, 16}, {0, 0, 0, 0}}},{30812, {{132025, 1, 0, 0},{132025, 1, 1, 17}, {0, 0, 0, 0}}},{30812, {{132025, 1, 0, 0},{132025, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{30814, {{132026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30815, {{132027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30816, {{132028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30817, {{132029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30818, {{132030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30819, {{132031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30820, {{132032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30821, {{132033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30822, {{132034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30823, {{132035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30824, {{132036, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30825, {{132038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30826, {{132039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30827, {{132040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59335, {{132041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30828, {{132051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30829, {{132052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30831, {{132055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30832, {{132059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30833, {{132065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30834, {{132066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30835, {{132067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30836, {{132068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30837, {{132071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30838, {{132072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30840, {{132073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30841, {{132074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30842, {{132075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30843, {{132076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30844, {{132077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30845, {{132078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30846, {{132079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30847, {{132080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{30848, {{132089, 1, 0, 0},{132089, 1, 1, 17}, {0, 0, 0, 0}}},{30848, {{132089, 1, 0, 0},{132089, 1, 1, 17}, {0, 0, 0, 0}}},{56954, {{132089, 1, 0, 0},{132089, 1, 1, 18}, {0, 0, 0, 0}}},{56954, {{132089, 1, 0, 0},{132089, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30849, {{132093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30850, {{132102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30851, {{132104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30852, {{132105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30854, {{132106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30855, {{132107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30856, {{132108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30857, {{132109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30858, {{132110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30859, {{132122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30860, {{132123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30862, {{132128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30863, {{132129, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30864, {{132130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30865, {{132131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30866, {{132132, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30867, {{132133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30868, {{132134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30869, {{132135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30870, {{132136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30871, {{132145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30872, {{132149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30873, {{132153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30874, {{132154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30875, {{132155, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30876, {{132156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30877, {{132157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30878, {{132158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30879, {{132159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30880, {{132162, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30881, {{132164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30882, {{132165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30883, {{132170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30890, {{132173, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30885, {{132174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30886, {{132176, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{30887, {{132183, 1, 0, 0},{132183, 1, 1, 16}, {0, 0, 0, 0}}},{30887, {{132183, 1, 0, 0},{132183, 1, 1, 16}, {0, 0, 0, 0}}},{56959, {{132183, 1, 0, 0},{132183, 1, 1, 17}, {0, 0, 0, 0}}},{56959, {{132183, 1, 0, 0},{132183, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{30888, {{132184, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30889, {{132185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30884, {{132186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30891, {{132187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30892, {{132190, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30893, {{132191, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30894, {{132193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30895, {{132196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30896, {{132201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30897, {{132202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30898, {{132203, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30899, {{132207, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30900, {{132208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30901, {{132209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30902, {{132210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30903, {{132212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30904, {{132214, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30905, {{132215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30906, {{132217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30907, {{132218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30908, {{132219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30909, {{132221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30910, {{132222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30911, {{132223, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30912, {{132224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30913, {{132225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30914, {{132227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30915, {{132228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30916, {{132229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30917, {{132230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30920, {{132234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30918, {{132235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30921, {{132236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30923, {{132237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30924, {{132239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30925, {{132240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{30926, {{132246, 1, 0, 0},{132246, 1, 1, 16}, {0, 0, 0, 0}}},{30926, {{132246, 1, 0, 0},{132246, 1, 1, 16}, {0, 0, 0, 0}}},{30927, {{132246, 1, 0, 0},{132246, 1, 1, 17}, {0, 0, 0, 0}}},{30927, {{132246, 1, 0, 0},{132246, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30928, {{132248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30929, {{132249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30930, {{132253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30931, {{132254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30932, {{132258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30940, {{132263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30933, {{132264, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30934, {{132269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30935, {{132270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30936, {{132272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30937, {{132273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30938, {{132274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30939, {{132280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30941, {{132294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30942, {{132296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30943, {{132298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30945, {{132299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30946, {{132300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30947, {{132302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30948, {{132303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30949, {{132311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30950, {{132313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30951, {{132316, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30952, {{132317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30953, {{132318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30954, {{132319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30955, {{132320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30956, {{132321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30957, {{132323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30958, {{132324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30959, {{132325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30960, {{132326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30961, {{132327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30962, {{132330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30963, {{132331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30964, {{132333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30966, {{132334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30967, {{132337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30968, {{132339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30969, {{132340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30971, {{132341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30970, {{132342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30972, {{132344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30973, {{132349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30974, {{132351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30975, {{132352, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30976, {{132353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30977, {{132354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30978, {{132355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30979, {{132356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30980, {{132358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30981, {{132359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30982, {{132360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{30983, {{132361, 1, 0, 0},{132361, 1, 1, 18}, {0, 0, 0, 0}}},{30983, {{132361, 1, 0, 0},{132361, 1, 1, 18}, {0, 0, 0, 0}}},{30984, {{132361, 1, 0, 0},{132361, 1, 1, 17}, {0, 0, 0, 0}}},{30984, {{132361, 1, 0, 0},{132361, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{30985, {{132362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30986, {{132363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30987, {{132364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30988, {{132365, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30989, {{132366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30990, {{132367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30991, {{132368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30992, {{132369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30993, {{132370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30994, {{132371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30995, {{132372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30996, {{132373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30997, {{132374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30998, {{132375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30999, {{132376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31000, {{132377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31001, {{132378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31002, {{132379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31003, {{132380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31004, {{132381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31005, {{132382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31006, {{132383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31007, {{132384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57410, {{132387, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {14, {{31009, {{132389, 1, 0, 0},{132389, 1, 1, 19}, {0, 0, 0, 0}}},{31009, {{132389, 1, 0, 0},{132389, 1, 1, 19}, {0, 0, 0, 0}}},{31010, {{132389, 1, 0, 0},{132389, 1, 1, 16}, {0, 0, 0, 0}}},{31010, {{132389, 1, 0, 0},{132389, 1, 1, 16}, {0, 0, 0, 0}}},{31011, {{132389, 1, 0, 0},{132389, 1, 1, 20}, {0, 0, 0, 0}}},{31011, {{132389, 1, 0, 0},{132389, 1, 1, 20}, {0, 0, 0, 0}}},{31012, {{132389, 1, 0, 0},{132389, 1, 1, 21}, {0, 0, 0, 0}}},{31012, {{132389, 1, 0, 0},{132389, 1, 1, 21}, {0, 0, 0, 0}}},{56992, {{132389, 1, 0, 0},{132389, 1, 1, 17}, {0, 0, 0, 0}}},{56992, {{132389, 1, 0, 0},{132389, 1, 1, 17}, {0, 0, 0, 0}}},{56997, {{132389, 1, 0, 0},{132389, 1, 1, 18}, {0, 0, 0, 0}}},{56997, {{132389, 1, 0, 0},{132389, 1, 1, 18}, {0, 0, 0, 0}}},{59329, {{132389, 1, 0, 0},{132389, 1, 1, 24}, {0, 0, 0, 0}}},{59329, {{132389, 1, 0, 0},{132389, 1, 1, 24}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31013, {{132391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31014, {{132392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31015, {{132396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31016, {{132399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31017, {{132400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31018, {{132401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31019, {{132402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31020, {{132406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31021, {{132408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31022, {{132409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31023, {{132411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31024, {{132412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31025, {{132413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31026, {{132414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31027, {{132415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{31028, {{132416, 1, 0, 0},{132416, 1, 1, 16}, {0, 0, 0, 0}}},{31028, {{132416, 1, 0, 0},{132416, 1, 1, 16}, {0, 0, 0, 0}}},{31029, {{132416, 1, 0, 0},{132416, 1, 1, 17}, {0, 0, 0, 0}}},{31029, {{132416, 1, 0, 0},{132416, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{31030, {{132417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31031, {{132418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31032, {{132419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31033, {{132420, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31034, {{132424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{31035, {{132427, 1, 0, 0},{132427, 1, 1, 16}, {0, 0, 0, 0}}},{31035, {{132427, 1, 0, 0},{132427, 1, 1, 16}, {0, 0, 0, 0}}},{57002, {{132427, 1, 0, 0},{132427, 1, 1, 17}, {0, 0, 0, 0}}},{57002, {{132427, 1, 0, 0},{132427, 1, 1, 17}, {0, 0, 0, 0}}},{57003, {{132427, 1, 0, 0},{132427, 1, 1, 18}, {0, 0, 0, 0}}},{57003, {{132427, 1, 0, 0},{132427, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{31036, {{132428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31037, {{132429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31038, {{132432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31039, {{132433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31040, {{132434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31041, {{132436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31042, {{132437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31043, {{132438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31044, {{132440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31045, {{132442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31046, {{132443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31047, {{132445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31048, {{132446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31049, {{132448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31050, {{132449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31051, {{132450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31052, {{132451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31053, {{132452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31054, {{132453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31055, {{132454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31056, {{132455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31057, {{132456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31058, {{132457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31059, {{132459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31060, {{132460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31061, {{132461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31062, {{132462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31063, {{132464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31064, {{132466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31065, {{132467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31066, {{132469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31067, {{132473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31068, {{132474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31069, {{132475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31070, {{132479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31071, {{132482, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31072, {{132484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31073, {{132485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31074, {{132486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31075, {{132487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31076, {{132490, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31077, {{132491, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31078, {{132492, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31079, {{132499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31080, {{132500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31081, {{132508, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31082, {{132511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31083, {{132512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31084, {{132515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31085, {{132517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31086, {{132518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31087, {{132519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31088, {{132521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31089, {{132524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31090, {{132529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31091, {{132530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31092, {{132533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31093, {{132534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31094, {{132535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31095, {{132536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31096, {{132537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31097, {{132546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31098, {{132547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31099, {{132548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31100, {{132550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31101, {{132552, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31102, {{132553, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31103, {{132554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31104, {{132555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31105, {{132556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31106, {{132557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31107, {{132558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31108, {{132559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31110, {{132566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31111, {{132567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31112, {{132568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31113, {{132569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31114, {{132570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31115, {{132571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31116, {{132572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31117, {{132573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31118, {{132578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31119, {{132581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31120, {{132582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31121, {{132584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31122, {{132585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31123, {{132594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31124, {{132595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31125, {{132596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31126, {{132597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31127, {{132598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31128, {{132602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31129, {{132603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31130, {{132604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31131, {{132605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31132, {{132606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31133, {{132607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31134, {{132608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31135, {{132609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31136, {{132612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31137, {{132613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31138, {{132614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31139, {{132615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31140, {{132618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31141, {{132619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31142, {{132620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31143, {{132621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31144, {{132626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31145, {{132630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31146, {{132631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31147, {{132632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31148, {{132637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31149, {{132638, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31150, {{132639, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31151, {{132641, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31152, {{132643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31153, {{132644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31154, {{132645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31155, {{132646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31156, {{132648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31157, {{132649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31158, {{132650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31159, {{132653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{31160, {{132655, 1, 0, 0},{132655, 1, 1, 16}, {0, 0, 0, 0}}},{31160, {{132655, 1, 0, 0},{132655, 1, 1, 16}, {0, 0, 0, 0}}},{31965, {{132655, 1, 0, 0},{132655, 1, 1, 17}, {0, 0, 0, 0}}},{31965, {{132655, 1, 0, 0},{132655, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31161, {{132657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31162, {{132658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31163, {{132659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31164, {{132661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31165, {{132662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31166, {{132663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31167, {{132664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31168, {{132666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31169, {{132667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31170, {{132668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31171, {{132669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31172, {{132670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31173, {{132673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31174, {{132674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31175, {{132678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31176, {{132679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31177, {{132680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31178, {{132681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31179, {{132682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31180, {{132683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31181, {{132686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31182, {{132687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31183, {{132688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31184, {{132689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31185, {{132690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31186, {{132692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31187, {{132693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31188, {{132695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31189, {{132696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31190, {{132697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31191, {{132698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31192, {{132700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31193, {{132702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31194, {{132703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31195, {{132704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31196, {{132706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31197, {{132708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57035, {{132709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31198, {{132716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31199, {{132717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{7437, {{132724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31200, {{132725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31201, {{132726, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31202, {{132727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31203, {{132728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31204, {{132729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31205, {{132730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31206, {{132731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31207, {{132732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31208, {{132734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31209, {{132735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31210, {{132736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31211, {{132737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31212, {{132739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31213, {{132740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31214, {{132741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31215, {{132742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31216, {{132745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31217, {{132746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31218, {{132747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31219, {{132748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31220, {{132749, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31221, {{132750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31222, {{132751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31223, {{132753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31224, {{132754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31225, {{132758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31226, {{132759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31227, {{132760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31228, {{132761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31229, {{132763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31230, {{132765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31231, {{132766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31232, {{132767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31233, {{132768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31234, {{132769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{31235, {{132771, 1, 0, 0},{132771, 1, 1, 16}, {0, 0, 0, 0}}},{31235, {{132771, 1, 0, 0},{132771, 1, 1, 16}, {0, 0, 0, 0}}},{31236, {{132771, 1, 0, 0},{132771, 1, 1, 17}, {0, 0, 0, 0}}},{31236, {{132771, 1, 0, 0},{132771, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31237, {{132773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31238, {{132774, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31239, {{132775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31240, {{132776, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31241, {{132777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31242, {{132778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31243, {{132779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31244, {{132780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31245, {{132781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31246, {{132782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31247, {{132783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31248, {{132784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31249, {{132785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31250, {{132791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31251, {{132792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31252, {{132793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31253, {{132794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31254, {{132795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31255, {{132796, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31256, {{132797, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31257, {{132798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31258, {{132799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31259, {{132800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31260, {{132801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31261, {{132802, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31262, {{132803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31263, {{132804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{31264, {{132809, 1, 0, 0},{132809, 1, 1, 16}, {0, 0, 0, 0}}},{31264, {{132809, 1, 0, 0},{132809, 1, 1, 16}, {0, 0, 0, 0}}},{31265, {{132809, 1, 0, 0},{132809, 1, 1, 17}, {0, 0, 0, 0}}},{31265, {{132809, 1, 0, 0},{132809, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{31266, {{132810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31267, {{132813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31268, {{132814, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31269, {{132815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31270, {{132816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31271, {{132817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31272, {{132818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31273, {{132819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31274, {{132820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31275, {{132821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31276, {{132822, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31277, {{132823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31278, {{132824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31279, {{132825, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31280, {{132826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31281, {{132827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31282, {{132832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31283, {{132837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31284, {{132838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31285, {{132839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31286, {{132840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31287, {{132842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31288, {{132843, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31289, {{132844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31290, {{132845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{31291, {{132846, 1, 0, 0},{132846, 1, 1, 16}, {0, 0, 0, 0}}},{31291, {{132846, 1, 0, 0},{132846, 1, 1, 16}, {0, 0, 0, 0}}},{31295, {{132846, 1, 0, 0},{132846, 1, 1, 17}, {0, 0, 0, 0}}},{31295, {{132846, 1, 0, 0},{132846, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{31292, {{132847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31293, {{132848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31294, {{132849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31296, {{132851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31297, {{132852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31298, {{132853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31299, {{132854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31300, {{132855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31301, {{132856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{31302, {{132857, 1, 0, 0},{132857, 1, 1, 16}, {0, 0, 0, 0}}},{31302, {{132857, 1, 0, 0},{132857, 1, 1, 16}, {0, 0, 0, 0}}},{31303, {{132857, 1, 0, 0},{132857, 1, 1, 17}, {0, 0, 0, 0}}},{31303, {{132857, 1, 0, 0},{132857, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31304, {{132859, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31305, {{132863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31306, {{132866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31307, {{132867, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31308, {{132868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31309, {{132870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31310, {{132871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31311, {{132872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31312, {{132873, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31313, {{132876, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31314, {{132880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31315, {{132881, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31316, {{132882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31317, {{132883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31318, {{132884, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31319, {{132885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31320, {{132886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31321, {{132887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31322, {{132888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31323, {{132889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31324, {{132890, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31325, {{132891, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31326, {{132892, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31327, {{132893, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31328, {{132894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31329, {{132895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31330, {{132896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31331, {{132901, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31332, {{132904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31333, {{132913, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31334, {{132914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31335, {{132915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31336, {{132916, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31337, {{132917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31338, {{132918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31339, {{132919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31340, {{132920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31341, {{132921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31342, {{132922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31343, {{132923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31344, {{132924, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31345, {{132926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31346, {{132928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31347, {{132929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31348, {{132930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31349, {{132931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31350, {{132932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31351, {{132933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31352, {{132943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31353, {{132945, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31354, {{132946, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31355, {{132952, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31356, {{132953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31357, {{132954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31358, {{132955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31359, {{132956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31360, {{132957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31361, {{132958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31362, {{132959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31364, {{132960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31365, {{132962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31366, {{132963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31367, {{132964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31368, {{132965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31369, {{132966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31370, {{132967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31371, {{132968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31372, {{132969, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31373, {{132970, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31374, {{132978, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31375, {{132979, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31376, {{132984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31377, {{132989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31378, {{132990, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31379, {{132991, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31380, {{132992, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31381, {{132993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31382, {{132994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31383, {{132995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31384, {{132996, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31385, {{132997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31386, {{132998, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31387, {{132999, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31388, {{133000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31389, {{133001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31390, {{133002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31391, {{133003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31392, {{133004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31393, {{133005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31394, {{133006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31395, {{133007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31396, {{133008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31397, {{133015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31398, {{133020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31399, {{133021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31400, {{133027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31401, {{133028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31402, {{133029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31403, {{133030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31404, {{133031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31405, {{133032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31406, {{133033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31407, {{133034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31408, {{133035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31409, {{133036, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31410, {{133037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31411, {{133038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31412, {{133039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31413, {{133040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31414, {{133041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31415, {{133042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31416, {{133047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31417, {{133051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31418, {{133052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31419, {{133053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31420, {{133054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31421, {{133055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31422, {{133058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31423, {{133059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31424, {{133060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31425, {{133061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31426, {{133062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31427, {{133063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31428, {{133064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31429, {{133065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31430, {{133066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31431, {{133067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31432, {{133068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31433, {{133069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31434, {{133070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31435, {{133071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31436, {{133072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31437, {{133073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31438, {{133074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31439, {{133075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31440, {{133076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31441, {{133077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31442, {{133078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31443, {{133079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31444, {{133080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31445, {{133087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31446, {{133094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31447, {{133095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31448, {{133096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31449, {{133097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31450, {{133098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31451, {{133099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31452, {{133100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31453, {{133101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31454, {{133102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31455, {{133103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31456, {{133104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31457, {{133105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31458, {{133107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31459, {{133109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31460, {{133114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31461, {{133115, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31462, {{133116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31463, {{133117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31464, {{133118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31465, {{133119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31466, {{133120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31467, {{133121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31468, {{133122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31469, {{133123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31470, {{133124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31471, {{133125, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31472, {{133126, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{31473, {{133127, 1, 0, 0},{133127, 1, 1, 17}, {0, 0, 0, 0}}},{31473, {{133127, 1, 0, 0},{133127, 1, 1, 17}, {0, 0, 0, 0}}},{31474, {{133127, 1, 0, 0},{133127, 1, 1, 18}, {0, 0, 0, 0}}},{31474, {{133127, 1, 0, 0},{133127, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31475, {{133131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31480, {{133132, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31476, {{133133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31477, {{133134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31478, {{133135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31479, {{133136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31481, {{133147, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31482, {{133148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31483, {{133149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31484, {{133150, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31485, {{133153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31486, {{133154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31487, {{133155, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31488, {{133158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31489, {{133159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31490, {{133162, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31491, {{133163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31492, {{133166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31493, {{133167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31494, {{133168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31495, {{133169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31496, {{133170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31497, {{133171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31498, {{133174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31499, {{133175, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31500, {{133177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31501, {{133178, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31503, {{133179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31505, {{133186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31506, {{133187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31507, {{133188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31504, {{133189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31508, {{133195, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31509, {{133197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31510, {{133201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31511, {{133202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31512, {{133203, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31513, {{133204, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31514, {{133205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31515, {{133214, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31517, {{133215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31518, {{133216, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31519, {{133217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31520, {{133219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31521, {{133220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31522, {{133221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31523, {{133222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31524, {{133227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31525, {{133228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31526, {{133229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31527, {{133230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31528, {{133231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31529, {{133232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31530, {{133233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31531, {{133238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31533, {{133239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31534, {{133240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31535, {{133241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31536, {{133243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31537, {{133244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31538, {{133245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31539, {{133246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31540, {{133247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31541, {{133251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31542, {{133253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31543, {{133254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31545, {{133255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31546, {{133256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31547, {{133257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31548, {{133260, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31549, {{133262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31550, {{133263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31551, {{133266, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31552, {{133267, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31553, {{133268, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31554, {{133269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31556, {{133270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31557, {{133271, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31558, {{133273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31559, {{133274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31560, {{133275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31561, {{133280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31562, {{133281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31563, {{133282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31564, {{133283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31565, {{133284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31566, {{133285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31567, {{133286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31568, {{133287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31570, {{133289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31571, {{133290, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31572, {{133293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31573, {{133294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31574, {{133295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31575, {{133296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31576, {{133297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31577, {{133298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31578, {{133299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31579, {{133300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31580, {{133301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31581, {{133302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31582, {{133303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31583, {{133305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31585, {{133306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31586, {{133307, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31587, {{133308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31588, {{133310, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31589, {{133311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31590, {{133312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31591, {{133313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31592, {{133315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31593, {{133316, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31594, {{133318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31595, {{133319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31596, {{133320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31597, {{133321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31598, {{133322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31599, {{133323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31600, {{133325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31601, {{133326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31602, {{133327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31603, {{133328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31604, {{133329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31605, {{133330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31606, {{133332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31607, {{133333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31608, {{133334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31609, {{133336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31610, {{133337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31611, {{133338, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31612, {{133342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31613, {{133343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31614, {{133344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31615, {{133345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31616, {{133347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31617, {{133348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{31618, {{133349, 1, 0, 0},{133349, 1, 1, 16}, {0, 0, 0, 0}}},{31618, {{133349, 1, 0, 0},{133349, 1, 1, 16}, {0, 0, 0, 0}}},{31619, {{133349, 1, 0, 0},{133349, 1, 1, 17}, {0, 0, 0, 0}}},{31619, {{133349, 1, 0, 0},{133349, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{31620, {{133350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31621, {{133351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31622, {{133352, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31623, {{133353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31624, {{133354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31625, {{133355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31626, {{133359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31627, {{133360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31628, {{133361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31629, {{133364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31630, {{133365, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31631, {{133368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31632, {{133370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31633, {{133373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31634, {{133374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31635, {{133375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31636, {{133376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31637, {{133380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31638, {{133381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31639, {{133383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31640, {{133384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31641, {{133385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31642, {{133386, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31643, {{133387, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31644, {{133390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31645, {{133391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31646, {{133392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31647, {{133393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31648, {{133394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31649, {{133395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31650, {{133396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31651, {{133397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31652, {{133399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31653, {{133400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31654, {{133401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31655, {{133402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31656, {{133403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31657, {{133404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31658, {{133405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31659, {{133406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31660, {{133407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31661, {{133410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31662, {{133411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{59937, {{133412, 1, 0, 0},{133412, 1, 1, 16}, {0, 0, 0, 0}}},{59937, {{133412, 1, 0, 0},{133412, 1, 1, 16}, {0, 0, 0, 0}}},{59938, {{133412, 1, 0, 0},{133412, 1, 1, 17}, {0, 0, 0, 0}}},{59938, {{133412, 1, 0, 0},{133412, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{31663, {{133413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31664, {{133418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31665, {{133419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31666, {{133420, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31667, {{133422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31668, {{133423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31669, {{133424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31670, {{133425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31671, {{133426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31672, {{133427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31673, {{133428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31674, {{133429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31675, {{133432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31676, {{133433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31677, {{133434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31678, {{133436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31679, {{133438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31680, {{133439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31681, {{133440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31682, {{133441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31683, {{133442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31684, {{133443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31685, {{133444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31686, {{133449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31687, {{133450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31688, {{133451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31689, {{133452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31690, {{133454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31691, {{133456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31692, {{133457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31693, {{133458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31694, {{133459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31695, {{133462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31696, {{133464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31697, {{133465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31698, {{133466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31699, {{133467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31700, {{133468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31701, {{133469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31702, {{133470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31703, {{133474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31704, {{133475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31705, {{133478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31706, {{133480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31707, {{133481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31708, {{133482, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31709, {{133483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31710, {{133484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31711, {{133486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31712, {{133487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31713, {{133488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31714, {{133489, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31715, {{133492, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31716, {{133493, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31717, {{133494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31718, {{133496, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31719, {{133498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31720, {{133499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31721, {{133500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31722, {{133501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31723, {{133502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31724, {{133503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31725, {{133504, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31726, {{133506, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31727, {{133507, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{31728, {{133508, 1, 0, 0},{133508, 1, 1, 17}, {0, 0, 0, 0}}},{31728, {{133508, 1, 0, 0},{133508, 1, 1, 17}, {0, 0, 0, 0}}},{31729, {{133508, 1, 0, 0},{133508, 1, 1, 18}, {0, 0, 0, 0}}},{31729, {{133508, 1, 0, 0},{133508, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{31730, {{133509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31731, {{133510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31732, {{133511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31733, {{133512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31734, {{133514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31735, {{133515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31736, {{133516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31737, {{133517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31738, {{133518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31739, {{133519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31740, {{133520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31741, {{133521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31742, {{133522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31743, {{133523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31744, {{133524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31745, {{133525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31746, {{133527, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31747, {{133528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31748, {{133529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31749, {{133530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31750, {{133531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31751, {{133532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31752, {{133533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31753, {{133537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31754, {{133538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31755, {{133539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31756, {{133540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31757, {{133543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31758, {{133544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31759, {{133546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31760, {{133547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31761, {{133548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31762, {{133556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31763, {{133557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31764, {{133558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57100, {{133560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31765, {{133568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31766, {{133576, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31767, {{133582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31768, {{133583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31769, {{133585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31770, {{133586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31771, {{133587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31772, {{133589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31773, {{133590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31774, {{133591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31775, {{133592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31776, {{133593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31777, {{133594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31778, {{133595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31779, {{133596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31780, {{133597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31781, {{133598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31782, {{133599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31783, {{133600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31784, {{133601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31785, {{133602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31786, {{133603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31787, {{133604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31788, {{133605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31789, {{133609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31791, {{133610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31792, {{133611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31793, {{133616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31794, {{133618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31795, {{133619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31796, {{133620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31797, {{133621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31798, {{133622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31799, {{133623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31800, {{133624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31801, {{133625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31802, {{133626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31803, {{133627, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31804, {{133628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31805, {{133629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31806, {{133631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31807, {{133632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31808, {{133633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31809, {{133634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31810, {{133635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31811, {{133637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31812, {{133638, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31813, {{133639, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31814, {{133640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31815, {{133642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31816, {{133643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31817, {{133644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31818, {{133645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31819, {{133646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31820, {{133648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31822, {{133650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31823, {{133651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31824, {{133652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31825, {{133653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31826, {{133654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31827, {{133656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31828, {{133657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31829, {{133658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31830, {{133659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31831, {{133660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31832, {{133661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31833, {{133662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31834, {{133663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31835, {{133664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31836, {{133666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31837, {{133667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31838, {{133668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31839, {{133669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31840, {{133670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{31841, {{133671, 1, 0, 0},{133671, 1, 1, 16}, {0, 0, 0, 0}}},{31841, {{133671, 1, 0, 0},{133671, 1, 1, 16}, {0, 0, 0, 0}}},{57416, {{133671, 1, 0, 0},{133671, 1, 1, 17}, {0, 0, 0, 0}}},{57416, {{133671, 1, 0, 0},{133671, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57112, {{133673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31842, {{133676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31843, {{133677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31844, {{133678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31845, {{133679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31846, {{133680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31847, {{133681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31848, {{133682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31849, {{133684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31850, {{133685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31851, {{133686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31852, {{133687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31853, {{133688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31854, {{133689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31855, {{133690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31856, {{133691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31857, {{133694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31858, {{133696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31859, {{133697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31860, {{133698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31861, {{133700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31862, {{133702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31863, {{133704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31864, {{133705, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31865, {{133706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31867, {{133707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31868, {{133708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31869, {{133709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31870, {{133710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31871, {{133711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31872, {{133716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31873, {{133718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31874, {{133719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31875, {{133720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31876, {{133721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31877, {{133723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31878, {{133724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31879, {{133725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31880, {{133726, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31881, {{133727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31882, {{133728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31883, {{133732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31885, {{133733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31886, {{133734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31887, {{133735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31888, {{133736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31889, {{133737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31890, {{133738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31891, {{133739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31892, {{133740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31893, {{133741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31894, {{133742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31895, {{133743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31896, {{133744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31897, {{133745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31898, {{133751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31899, {{133752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31900, {{133753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31901, {{133754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31902, {{133755, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31903, {{133756, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31904, {{133757, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31905, {{133759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31906, {{133762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31907, {{133763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31908, {{133764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31909, {{133766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31910, {{133767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31911, {{133768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31912, {{133769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31913, {{133770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31914, {{133771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31915, {{133772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31916, {{133773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31917, {{133775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31918, {{133776, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31919, {{133777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31920, {{133778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31921, {{133779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31922, {{133780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31923, {{133783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31924, {{133784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31925, {{133785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31926, {{133786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31927, {{133787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31928, {{133788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31929, {{133789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31930, {{133790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31931, {{133791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31932, {{133795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31933, {{133797, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31934, {{133798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31935, {{133799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31936, {{133800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31937, {{133801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31938, {{133802, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31939, {{133803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31940, {{133805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31941, {{133806, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31942, {{133807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31943, {{133808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31944, {{133811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31945, {{133814, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31946, {{133815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31947, {{133817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31948, {{133818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31949, {{133819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31950, {{133820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31951, {{133823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31952, {{133827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31953, {{133828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31954, {{133832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31955, {{133833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31956, {{133836, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31957, {{133838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31958, {{133840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{31959, {{133843, 1, 0, 0},{133843, 1, 1, 18}, {0, 0, 0, 0}}},{31959, {{133843, 1, 0, 0},{133843, 1, 1, 18}, {0, 0, 0, 0}}},{31960, {{133843, 1, 0, 0},{133843, 1, 1, 17}, {0, 0, 0, 0}}},{31960, {{133843, 1, 0, 0},{133843, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{31962, {{133844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31963, {{133845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31964, {{133846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31966, {{133848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31967, {{133849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58023, {{133850, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31968, {{133851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31969, {{133852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31970, {{133854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31971, {{133855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31972, {{133856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31973, {{133859, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{31974, {{133860, 1, 0, 0},{133860, 1, 1, 16}, {0, 0, 0, 0}}},{31974, {{133860, 1, 0, 0},{133860, 1, 1, 16}, {0, 0, 0, 0}}},{57125, {{133860, 1, 0, 0},{133860, 1, 1, 17}, {0, 0, 0, 0}}},{57125, {{133860, 1, 0, 0},{133860, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{31975, {{133861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31976, {{133862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31977, {{133864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31978, {{133865, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31979, {{133866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31980, {{133867, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31981, {{133868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31982, {{133869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31983, {{133870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31984, {{133871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31985, {{133872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31986, {{133873, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31987, {{133875, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31988, {{133877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31989, {{133878, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31990, {{133881, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31991, {{133882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31992, {{133883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31993, {{133884, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31994, {{133887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31995, {{133888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31996, {{133889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31997, {{133890, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{31998, {{133894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{31999, {{133895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32000, {{133896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32001, {{133899, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32002, {{133902, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32003, {{133904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32004, {{133905, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32005, {{133907, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32006, {{133912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32007, {{133914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32008, {{133915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32009, {{133916, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32010, {{133917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32011, {{133918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32012, {{133919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32013, {{133921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32014, {{133922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32015, {{133923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32016, {{133925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32017, {{133926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32018, {{133927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32019, {{133928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32020, {{133929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32021, {{133930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32022, {{133931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32023, {{133932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32024, {{133933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32025, {{133934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32026, {{133935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32027, {{133937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32028, {{133938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32029, {{133939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32030, {{133944, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32031, {{133945, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32032, {{133946, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32033, {{133947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32034, {{133948, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32035, {{133950, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32036, {{133951, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32037, {{133952, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32038, {{133953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32039, {{133954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32040, {{133959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32041, {{133960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32042, {{133961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32043, {{133962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32044, {{133963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32045, {{133964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32046, {{133965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32048, {{133966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32049, {{133967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32050, {{133972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32051, {{133973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32052, {{133974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32053, {{133975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32054, {{133976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32055, {{133977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32060, {{133978, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32057, {{133979, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32058, {{133980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32059, {{133981, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32056, {{133982, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32061, {{133984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32062, {{133985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{32063, {{133987, 1, 0, 0},{133987, 1, 1, 16}, {0, 0, 0, 0}}},{32063, {{133987, 1, 0, 0},{133987, 1, 1, 16}, {0, 0, 0, 0}}},{32064, {{133987, 1, 0, 0},{133987, 1, 1, 17}, {0, 0, 0, 0}}},{32064, {{133987, 1, 0, 0},{133987, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32065, {{133989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32066, {{133990, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32067, {{133991, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32068, {{133992, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{32069, {{133999, 1, 0, 0},{133999, 1, 1, 16}, {0, 0, 0, 0}}},{32069, {{133999, 1, 0, 0},{133999, 1, 1, 16}, {0, 0, 0, 0}}},{32070, {{133999, 1, 0, 0},{133999, 1, 1, 17}, {0, 0, 0, 0}}},{32070, {{133999, 1, 0, 0},{133999, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{32071, {{134000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32072, {{134001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32073, {{134002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32074, {{134003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32075, {{134004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32076, {{134005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32077, {{134006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32078, {{134008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32079, {{134009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32080, {{134011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32081, {{134012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32082, {{134014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32083, {{134015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32084, {{134016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32085, {{134017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32086, {{134019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32087, {{134020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32088, {{134022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32089, {{134023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32090, {{134025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32091, {{134027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32092, {{134029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32093, {{134030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32094, {{134032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32095, {{134033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32096, {{134034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{32097, {{134035, 1, 0, 0},{134035, 1, 1, 16}, {0, 0, 0, 0}}},{32097, {{134035, 1, 0, 0},{134035, 1, 1, 16}, {0, 0, 0, 0}}},{32098, {{134035, 1, 0, 0},{134035, 1, 1, 17}, {0, 0, 0, 0}}},{32098, {{134035, 1, 0, 0},{134035, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32099, {{134037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32100, {{134038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32101, {{134039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32102, {{134040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32103, {{134041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32104, {{134042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32105, {{134043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32106, {{134044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32107, {{134046, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32108, {{134047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32109, {{134048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32110, {{134049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32111, {{134050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32112, {{134051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32113, {{134052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32114, {{134053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32115, {{134054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32116, {{134058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32117, {{134059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32118, {{134060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32119, {{134061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32120, {{134063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32121, {{134064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{32123, {{134065, 1, 0, 0},{134065, 1, 1, 16}, {0, 0, 0, 0}}},{32123, {{134065, 1, 0, 0},{134065, 1, 1, 16}, {0, 0, 0, 0}}},{32124, {{134065, 1, 0, 0},{134065, 1, 1, 17}, {0, 0, 0, 0}}},{32124, {{134065, 1, 0, 0},{134065, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{32125, {{134066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32126, {{134067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32127, {{134068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32128, {{134069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32129, {{134071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32130, {{134076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32131, {{134077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32132, {{134080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32133, {{134081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32134, {{134082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32135, {{134083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32136, {{134084, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32137, {{134088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32138, {{134089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32139, {{134090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32140, {{134091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{32141, {{134092, 1, 0, 0},{134092, 1, 1, 16}, {0, 0, 0, 0}}},{32141, {{134092, 1, 0, 0},{134092, 1, 1, 16}, {0, 0, 0, 0}}},{57146, {{134092, 1, 0, 0},{134092, 1, 1, 17}, {0, 0, 0, 0}}},{57146, {{134092, 1, 0, 0},{134092, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{32142, {{134093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32143, {{134095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32144, {{134096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32145, {{134097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32146, {{134098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32147, {{134099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32148, {{134100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32149, {{134101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32150, {{134102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32151, {{134103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32152, {{134104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32153, {{134105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32154, {{134106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32155, {{134107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32156, {{134108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32157, {{134109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32158, {{134110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32159, {{134118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32160, {{134122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32161, {{134123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32162, {{134124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32163, {{134125, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32164, {{134126, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32165, {{134128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32166, {{134129, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32167, {{134130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32168, {{134131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32169, {{134132, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32170, {{134133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32171, {{134134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32172, {{134135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32173, {{134148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32174, {{134149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32175, {{134150, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32176, {{134152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32177, {{134153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32178, {{134154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32179, {{134155, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32180, {{134156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32181, {{134157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32182, {{134158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32183, {{134159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32184, {{134160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32185, {{134161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32186, {{134162, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32187, {{134163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32188, {{134164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32189, {{134165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32190, {{134166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32191, {{134167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32192, {{134168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32193, {{134169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32194, {{134170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32195, {{134171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32196, {{134172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32197, {{134173, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32198, {{134174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32199, {{134177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{21678, {{134181, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32200, {{134194, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32201, {{134198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32202, {{134200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32203, {{134202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32204, {{134204, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32205, {{134205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32206, {{134206, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32207, {{134208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32209, {{134222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{32210, {{134224, 1, 0, 0},{134224, 1, 1, 16}, {0, 0, 0, 0}}},{32210, {{134224, 1, 0, 0},{134224, 1, 1, 16}, {0, 0, 0, 0}}},{57161, {{134224, 1, 0, 0},{134224, 1, 1, 17}, {0, 0, 0, 0}}},{57161, {{134224, 1, 0, 0},{134224, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{32211, {{134225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32212, {{134226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32213, {{134227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32214, {{134228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32215, {{134229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32216, {{134230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32217, {{134231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32218, {{134232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32219, {{134233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32220, {{134234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32221, {{134235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32222, {{134236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32223, {{134237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32224, {{134238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32225, {{134239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32226, {{134240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32227, {{134241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32228, {{134242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32229, {{134243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32230, {{134244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32231, {{134245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32232, {{134247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32233, {{134248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32234, {{134249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32235, {{134251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32236, {{134252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32237, {{134253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32238, {{134254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32239, {{134255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32240, {{134256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32241, {{134257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32242, {{134258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32243, {{134259, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32244, {{134260, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32246, {{134283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32247, {{134284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32248, {{134286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32249, {{134289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32250, {{134291, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32251, {{134295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32252, {{134297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32253, {{134298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32254, {{134313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32255, {{134314, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32256, {{134317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32257, {{134318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32258, {{134319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32259, {{134320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32260, {{134321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32261, {{134322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32262, {{134323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32263, {{134324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32264, {{134325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32265, {{134326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32266, {{134327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32267, {{134328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32268, {{134329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32269, {{134330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32270, {{134331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32271, {{134332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32272, {{134333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32273, {{134334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32274, {{134335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32275, {{134336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32276, {{134338, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32277, {{134339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32278, {{134340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32279, {{134341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32280, {{134342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32281, {{134343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32282, {{134344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32283, {{134345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32284, {{134346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32285, {{134347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32286, {{134348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32287, {{134349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32288, {{134350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32289, {{134355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32290, {{134379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32291, {{134380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32292, {{134383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32293, {{134384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32294, {{134385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32295, {{134391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32296, {{134392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32297, {{134393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32298, {{134394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32299, {{134395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32300, {{134400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32301, {{134403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32302, {{134404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32303, {{134433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32304, {{134434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32305, {{134435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32306, {{134439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32307, {{134440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32308, {{134441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32309, {{134442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32310, {{134443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32311, {{134444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32312, {{134445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32313, {{134446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32314, {{134447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32315, {{134448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32316, {{134449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32317, {{134450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32318, {{134451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32319, {{134452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32320, {{134453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32321, {{134454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32322, {{134455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32323, {{134456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32324, {{134457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32325, {{134458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32326, {{134459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32327, {{134460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32328, {{134461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32329, {{134462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32330, {{134463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32331, {{134464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32332, {{134465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32333, {{134466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32334, {{134467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{32335, {{134469, 1, 0, 0},{134469, 1, 1, 18}, {0, 0, 0, 0}}},{32335, {{134469, 1, 0, 0},{134469, 1, 1, 18}, {0, 0, 0, 0}}},{32336, {{134469, 1, 0, 0},{134469, 1, 1, 17}, {0, 0, 0, 0}}},{32336, {{134469, 1, 0, 0},{134469, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{32337, {{134474, 1, 0, 0},{134474, 1, 1, 17}, {0, 0, 0, 0}}},{32337, {{134474, 1, 0, 0},{134474, 1, 1, 17}, {0, 0, 0, 0}}},{32338, {{134474, 1, 0, 0},{134474, 1, 1, 16}, {0, 0, 0, 0}}},{32338, {{134474, 1, 0, 0},{134474, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32339, {{134481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32340, {{134482, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32341, {{134483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32342, {{134488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32343, {{134496, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32344, {{134497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32345, {{134498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32346, {{134501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32347, {{134506, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32348, {{134508, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32349, {{134510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32350, {{134512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32351, {{134513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32352, {{134514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32353, {{134515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32354, {{134516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32355, {{134517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32356, {{134519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32357, {{134520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32358, {{134521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32359, {{134525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32360, {{134542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32361, {{134567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32362, {{134568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32363, {{134569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32364, {{134570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32365, {{134571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32366, {{134572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32367, {{134573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{32368, {{134574, 1, 0, 0},{134574, 1, 1, 16}, {0, 0, 0, 0}}},{32368, {{134574, 1, 0, 0},{134574, 1, 1, 16}, {0, 0, 0, 0}}},{32369, {{134574, 1, 0, 0},{134574, 1, 1, 17}, {0, 0, 0, 0}}},{32369, {{134574, 1, 0, 0},{134574, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32370, {{134576, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32371, {{134577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32372, {{134578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32373, {{134579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32374, {{134580, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32375, {{134581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{32376, {{134583, 1, 0, 0},{134583, 1, 1, 16}, {0, 0, 0, 0}}},{32376, {{134583, 1, 0, 0},{134583, 1, 1, 16}, {0, 0, 0, 0}}},{32377, {{134583, 1, 0, 0},{134583, 1, 1, 17}, {0, 0, 0, 0}}},{32377, {{134583, 1, 0, 0},{134583, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{32378, {{134584, 1, 0, 0},{134584, 1, 1, 16}, {0, 0, 0, 0}}},{32378, {{134584, 1, 0, 0},{134584, 1, 1, 16}, {0, 0, 0, 0}}},{32379, {{134584, 1, 0, 0},{134584, 1, 1, 17}, {0, 0, 0, 0}}},{32379, {{134584, 1, 0, 0},{134584, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{32380, {{134585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32381, {{134586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32382, {{134587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32383, {{134588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32384, {{134589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32385, {{134590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32386, {{134591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32387, {{134592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32388, {{134593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32389, {{134594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32390, {{134595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32391, {{134596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32392, {{134597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32393, {{134598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32394, {{134599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32395, {{134600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32396, {{134601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32397, {{134602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32398, {{134603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32399, {{134604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32400, {{134605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32401, {{134606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32402, {{134607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32403, {{134608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32404, {{134609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32405, {{134610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32406, {{134611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{32407, {{134612, 1, 0, 0},{134612, 1, 1, 16}, {0, 0, 0, 0}}},{32407, {{134612, 1, 0, 0},{134612, 1, 1, 16}, {0, 0, 0, 0}}},{57185, {{134612, 1, 0, 0},{134612, 1, 1, 17}, {0, 0, 0, 0}}},{57185, {{134612, 1, 0, 0},{134612, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{32408, {{134613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32409, {{134614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32410, {{134615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32411, {{134616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32412, {{134617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32413, {{134618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32414, {{134619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32415, {{134620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32416, {{134621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32417, {{134622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32418, {{134623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32419, {{134625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{57130, {{134640, 1, 0, 0},{134640, 1, 1, 16}, {0, 0, 0, 0}}},{57130, {{134640, 1, 0, 0},{134640, 1, 1, 16}, {0, 0, 0, 0}}},{57181, {{134640, 1, 0, 0},{134640, 1, 1, 17}, {0, 0, 0, 0}}},{57181, {{134640, 1, 0, 0},{134640, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32421, {{134648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32422, {{134651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32423, {{134652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32424, {{134653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32425, {{134654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32426, {{134656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32427, {{134701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32428, {{134702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32429, {{134703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32430, {{134704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32431, {{134705, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32432, {{134706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32433, {{134707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32434, {{134708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32435, {{134709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32436, {{134710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32437, {{134711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32438, {{134712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32439, {{134713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32440, {{134714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32441, {{134715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32442, {{134719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32443, {{134720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32444, {{134721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32445, {{134722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32446, {{134723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32447, {{134724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32448, {{134725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32449, {{134726, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32450, {{134727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32451, {{134728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32452, {{134729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32453, {{134730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32454, {{134731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32455, {{134732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32456, {{134733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32457, {{134734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32458, {{134735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32459, {{134736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32460, {{134737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32461, {{134738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32462, {{134739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32463, {{134740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32464, {{134741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32465, {{134742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32466, {{134743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32467, {{134744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32468, {{134745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32469, {{134746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32470, {{134747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32471, {{134751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32472, {{134752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32473, {{134753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32474, {{134754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32475, {{134755, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32476, {{134756, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32477, {{134757, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32478, {{134758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32479, {{134759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32480, {{134760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32481, {{134761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32482, {{134762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32483, {{134765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32484, {{134798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32485, {{134799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32486, {{134800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32487, {{134803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32488, {{134804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32489, {{134805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32490, {{134808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32491, {{134820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32492, {{134860, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32493, {{134861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32494, {{134862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32495, {{134863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32496, {{134864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32497, {{134869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32498, {{134870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32499, {{134871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32500, {{134872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32501, {{134873, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32502, {{134874, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{32503, {{134875, 1, 0, 0},{134875, 1, 1, 17}, {0, 0, 0, 0}}},{32503, {{134875, 1, 0, 0},{134875, 1, 1, 17}, {0, 0, 0, 0}}},{32504, {{134875, 1, 0, 0},{134875, 1, 1, 16}, {0, 0, 0, 0}}},{32504, {{134875, 1, 0, 0},{134875, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{32505, {{134876, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32506, {{134878, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32507, {{134879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32508, {{134880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32509, {{134881, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32510, {{134882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32511, {{134883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32512, {{134884, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32513, {{134885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32514, {{134886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32515, {{134887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32516, {{134888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32517, {{134889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32518, {{134890, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32519, {{134891, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32520, {{134892, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32521, {{134893, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32522, {{134894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32523, {{134895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32524, {{134896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32525, {{134897, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57189, {{134899, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32526, {{134902, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{8548, {{134904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32527, {{134907, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32529, {{134933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32530, {{134936, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32531, {{134938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32532, {{134940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32533, {{134941, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32534, {{134942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32535, {{134944, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32536, {{134946, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32537, {{134947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32538, {{134950, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32539, {{135004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32540, {{135005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32541, {{135007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32542, {{135009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32543, {{135010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32544, {{135011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32545, {{135012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32546, {{135013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32547, {{135014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32548, {{135015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32549, {{135017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32550, {{135018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32551, {{135019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32552, {{135020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32553, {{135021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32554, {{135022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32555, {{135023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32556, {{135024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32557, {{135025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32558, {{135026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32559, {{135027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32560, {{135028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32561, {{135029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32562, {{135030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32563, {{135031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32564, {{135032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32565, {{135033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32566, {{135034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32567, {{135035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32568, {{135036, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32569, {{135038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32570, {{135039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32571, {{135040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32572, {{135041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32573, {{135042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32574, {{135043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32575, {{135044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32576, {{135045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32577, {{135048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32578, {{135049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32579, {{135050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32580, {{135051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32581, {{135052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59436, {{135055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32583, {{135074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32584, {{135075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32585, {{135076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32586, {{135077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32587, {{135078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32588, {{135079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32589, {{135081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32590, {{135082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32591, {{135083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32592, {{135084, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32593, {{135086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32594, {{135087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32595, {{135088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32596, {{135098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32597, {{135109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32598, {{135110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32599, {{135111, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32600, {{135112, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32601, {{135113, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32602, {{135114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{32603, {{135115, 1, 0, 0},{135115, 1, 1, 16}, {0, 0, 0, 0}}},{32603, {{135115, 1, 0, 0},{135115, 1, 1, 16}, {0, 0, 0, 0}}},{32604, {{135115, 1, 0, 0},{135115, 1, 1, 17}, {0, 0, 0, 0}}},{32604, {{135115, 1, 0, 0},{135115, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{32605, {{135116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32606, {{135117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32607, {{135118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32608, {{135119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32609, {{135120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32610, {{135121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32611, {{135122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32612, {{135123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32613, {{135124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{32671, {{135125, 1, 0, 0},{135125, 1, 1, 16}, {0, 0, 0, 0}}},{32671, {{135125, 1, 0, 0},{135125, 1, 1, 16}, {0, 0, 0, 0}}},{32672, {{135125, 1, 0, 0},{135125, 1, 1, 17}, {0, 0, 0, 0}}},{32672, {{135125, 1, 0, 0},{135125, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{32615, {{135126, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32616, {{135127, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32617, {{135128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32618, {{135129, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32619, {{135130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32620, {{135131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32621, {{135132, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32622, {{135133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32623, {{135134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32624, {{135135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32625, {{135136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32627, {{135137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32628, {{135138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32629, {{135139, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32630, {{135140, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32631, {{135141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32632, {{135142, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32633, {{135143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32634, {{135144, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32635, {{135167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32636, {{135184, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32637, {{135185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32638, {{135186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32639, {{135187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32640, {{135188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32641, {{135190, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32642, {{135191, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32643, {{135218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32644, {{135224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32645, {{135225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32646, {{135226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32648, {{135229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32649, {{135230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32650, {{135231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32582, {{135232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32651, {{135233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32652, {{135234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32653, {{135235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32654, {{135236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32655, {{135237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32656, {{135238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32657, {{135239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32658, {{135240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32659, {{135241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32647, {{135242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32660, {{135243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32661, {{135244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32662, {{135246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32663, {{135247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32664, {{135248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32665, {{135249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32666, {{135250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32667, {{135251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32668, {{135252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32669, {{135253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32670, {{135254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32614, {{135255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32673, {{135271, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32674, {{135272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32675, {{135273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32676, {{135274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32677, {{135275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32678, {{135276, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32679, {{135277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32680, {{135280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32681, {{135306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32682, {{135307, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32683, {{135311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32684, {{135312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32685, {{135313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32686, {{135314, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32687, {{135315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32688, {{135316, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32689, {{135317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32690, {{135318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32691, {{135319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32692, {{135320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32693, {{135321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32694, {{135322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32695, {{135323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32696, {{135324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32697, {{135325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32698, {{135326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32699, {{135327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32700, {{135328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32701, {{135329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32702, {{135330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32703, {{135331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32704, {{135332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32705, {{135333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32706, {{135334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32707, {{135335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32708, {{135336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32709, {{135337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32710, {{135350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32711, {{135362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32712, {{135363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32713, {{135389, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32714, {{135390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32715, {{135392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32716, {{135393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32717, {{135394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32718, {{135395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32719, {{135396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32720, {{135397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32721, {{135398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32722, {{135399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32723, {{135400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32724, {{135401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32725, {{135402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32726, {{135403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32727, {{135404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32728, {{135405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32729, {{135406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32730, {{135407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32731, {{135411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32732, {{135415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32734, {{135431, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32735, {{135432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32736, {{135433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32737, {{135434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32738, {{135441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32739, {{135442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32740, {{135443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32741, {{135444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32742, {{135445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32743, {{135446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32744, {{135447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32745, {{135448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32746, {{135449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32747, {{135450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32748, {{135451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59466, {{135452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32749, {{135456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32750, {{135458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32751, {{135459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32752, {{135473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32753, {{135474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32754, {{135475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32755, {{135476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32756, {{135477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32757, {{135478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32758, {{135479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32759, {{135480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32760, {{135481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32761, {{135482, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{32762, {{135483, 1, 0, 0},{135483, 1, 1, 17}, {0, 0, 0, 0}}},{32762, {{135483, 1, 0, 0},{135483, 1, 1, 17}, {0, 0, 0, 0}}},{32763, {{135483, 1, 0, 0},{135483, 1, 1, 16}, {0, 0, 0, 0}}},{32763, {{135483, 1, 0, 0},{135483, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32764, {{135492, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32765, {{135507, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32766, {{135508, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32768, {{135511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32769, {{135512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32770, {{135513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32771, {{135514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32772, {{135515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32773, {{135516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32774, {{135517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32775, {{135518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32776, {{135519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32777, {{135521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32778, {{135533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32779, {{135534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32780, {{135536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32781, {{135537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32794, {{135540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32782, {{135541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32783, {{135542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32784, {{135544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32785, {{135545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32786, {{135546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32787, {{135547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32788, {{135548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32789, {{135549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32790, {{135550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32791, {{135551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32792, {{135557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32793, {{135558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32795, {{135566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32796, {{135567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32797, {{135568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32798, {{135569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32799, {{135570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32800, {{135571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32801, {{135572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32802, {{135575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32803, {{135584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32804, {{135585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32805, {{135587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32806, {{135588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32807, {{135589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32808, {{135590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32809, {{135591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32810, {{135592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32811, {{135593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{465, {{135594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32812, {{135598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32813, {{135599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32814, {{135600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32815, {{135601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32816, {{135602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32817, {{135603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32818, {{135604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32819, {{135605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32820, {{135606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32821, {{135607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32822, {{135608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32823, {{135611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32824, {{135612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32825, {{135613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32826, {{135614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32827, {{135615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32828, {{135616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32829, {{135617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32830, {{135618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32831, {{135619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32832, {{135620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32833, {{135622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32834, {{135623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32835, {{135624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32836, {{135625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32837, {{135626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32838, {{135627, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32839, {{135628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32840, {{135629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32841, {{135630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32843, {{135633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32844, {{135634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32845, {{135635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32846, {{135636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32847, {{135637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32848, {{135638, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32849, {{135639, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32850, {{135640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32851, {{135642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32852, {{135643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32853, {{135644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32854, {{135645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32855, {{135646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32856, {{135647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32857, {{135648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32859, {{135649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32860, {{135650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32861, {{135651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32862, {{135652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32863, {{135654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32864, {{135655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32865, {{135657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32866, {{135658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32867, {{135659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32868, {{135660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32869, {{135661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32870, {{135662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32871, {{135664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32872, {{135665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32873, {{135666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32874, {{135667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32875, {{135668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32876, {{135669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32877, {{135670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32878, {{135671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32879, {{135672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32880, {{135673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32881, {{135674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32882, {{135676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32883, {{135677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32884, {{135678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32885, {{135679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32886, {{135680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32887, {{135681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32888, {{135685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32889, {{135686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32890, {{135687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32891, {{135689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32892, {{135690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32893, {{135691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32894, {{135692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32895, {{135693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32896, {{135694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32897, {{135695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32904, {{135696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32898, {{135697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32899, {{135698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32900, {{135699, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32901, {{135701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32903, {{135702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32905, {{135704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32906, {{135705, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32907, {{135706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32908, {{135710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32909, {{135711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32910, {{135712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32911, {{135714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32912, {{135715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32913, {{135716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32914, {{135718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32915, {{135719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32916, {{135720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32917, {{135722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32918, {{135725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32919, {{135726, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32920, {{135727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32921, {{135728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32922, {{135729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32923, {{135730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32924, {{135731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32925, {{135732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32926, {{135733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32927, {{135734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32928, {{135735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32929, {{135736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32930, {{135737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32931, {{135740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32932, {{135741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32933, {{135742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32934, {{135743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32935, {{135748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32936, {{135752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32937, {{135753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32938, {{135754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32939, {{135758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{8910, {{135759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32941, {{135760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32942, {{135762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32943, {{135765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32944, {{135767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32945, {{135768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32946, {{135769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32947, {{135770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32948, {{135771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32949, {{135773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32950, {{135775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32951, {{135785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32952, {{135786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32953, {{135788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32954, {{135789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32955, {{135790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32956, {{135793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32957, {{135794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32958, {{135795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{32959, {{135796, 1, 0, 0},{135796, 1, 1, 17}, {0, 0, 0, 0}}},{32959, {{135796, 1, 0, 0},{135796, 1, 1, 17}, {0, 0, 0, 0}}},{32960, {{135796, 1, 0, 0},{135796, 1, 1, 18}, {0, 0, 0, 0}}},{32960, {{135796, 1, 0, 0},{135796, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {4, {{32961, {{135797, 1, 0, 0},{135797, 1, 1, 17}, {0, 0, 0, 0}}},{32961, {{135797, 1, 0, 0},{135797, 1, 1, 17}, {0, 0, 0, 0}}},{57209, {{135797, 1, 0, 0},{135797, 1, 1, 16}, {0, 0, 0, 0}}},{57209, {{135797, 1, 0, 0},{135797, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{32962, {{135798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32963, {{135799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32964, {{135800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32965, {{135803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32966, {{135808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32967, {{135809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32968, {{135810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32969, {{135811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32970, {{135812, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32971, {{135813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32972, {{135814, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32973, {{135815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32974, {{135816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32975, {{135817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32940, {{135818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32976, {{135819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32977, {{135820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32978, {{135821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32979, {{135822, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{32980, {{135823, 1, 0, 0},{135823, 1, 1, 16}, {0, 0, 0, 0}}},{32980, {{135823, 1, 0, 0},{135823, 1, 1, 16}, {0, 0, 0, 0}}},{32981, {{135823, 1, 0, 0},{135823, 1, 1, 17}, {0, 0, 0, 0}}},{32981, {{135823, 1, 0, 0},{135823, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32982, {{135830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{57214, {{135833, 1, 0, 0},{135833, 1, 1, 16}, {0, 0, 0, 0}}},{57214, {{135833, 1, 0, 0},{135833, 1, 1, 16}, {0, 0, 0, 0}}},{57215, {{135833, 1, 0, 0},{135833, 1, 1, 17}, {0, 0, 0, 0}}},{57215, {{135833, 1, 0, 0},{135833, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32983, {{135835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32984, {{135837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32985, {{135838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32986, {{135839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32988, {{135842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32989, {{135843, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32990, {{135844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{32991, {{135845, 1, 0, 0},{135845, 1, 1, 17}, {0, 0, 0, 0}}},{32991, {{135845, 1, 0, 0},{135845, 1, 1, 17}, {0, 0, 0, 0}}},{57212, {{135845, 1, 0, 0},{135845, 1, 1, 16}, {0, 0, 0, 0}}},{57212, {{135845, 1, 0, 0},{135845, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{32992, {{135846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32993, {{135852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57216, {{135854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32994, {{135855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32995, {{135860, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32996, {{135861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32997, {{135862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32998, {{135863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32999, {{135864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33000, {{135865, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33001, {{135866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33002, {{135867, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33003, {{135868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33004, {{135869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33005, {{135870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33006, {{135895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33007, {{135896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33008, {{135899, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33009, {{135900, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33010, {{135901, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33011, {{135902, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33012, {{135904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33013, {{135908, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33014, {{135911, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33015, {{135912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33016, {{135913, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33017, {{135914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33018, {{135915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33019, {{135916, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33020, {{135917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33021, {{135918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33022, {{135919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33023, {{135920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33024, {{135921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33025, {{135922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{33026, {{135923, 1, 0, 0},{135923, 1, 1, 16}, {0, 0, 0, 0}}},{33026, {{135923, 1, 0, 0},{135923, 1, 1, 16}, {0, 0, 0, 0}}},{33027, {{135923, 1, 0, 0},{135923, 1, 1, 17}, {0, 0, 0, 0}}},{33027, {{135923, 1, 0, 0},{135923, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{33028, {{135924, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33029, {{135925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33030, {{135926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33031, {{135927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33032, {{135928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33033, {{135933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33034, {{135937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33035, {{135942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33036, {{135948, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33037, {{135957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33038, {{135958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33039, {{135959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33040, {{135960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33041, {{135961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{33042, {{135963, 1, 0, 0},{135963, 1, 1, 17}, {0, 0, 0, 0}}},{33042, {{135963, 1, 0, 0},{135963, 1, 1, 17}, {0, 0, 0, 0}}},{33043, {{135963, 1, 0, 0},{135963, 1, 1, 18}, {0, 0, 0, 0}}},{33043, {{135963, 1, 0, 0},{135963, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{33044, {{135964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33045, {{135965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33046, {{135966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33047, {{135967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33048, {{135972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33049, {{135973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33050, {{135974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33051, {{135975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{33052, {{135976, 1, 0, 0},{135976, 1, 1, 17}, {0, 0, 0, 0}}},{33052, {{135976, 1, 0, 0},{135976, 1, 1, 17}, {0, 0, 0, 0}}},{57222, {{135976, 1, 0, 0},{135976, 1, 1, 16}, {0, 0, 0, 0}}},{57222, {{135976, 1, 0, 0},{135976, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{33053, {{135977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33054, {{135978, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33055, {{135979, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33056, {{135980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33057, {{135981, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33058, {{135982, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33059, {{135983, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33060, {{135984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33061, {{135985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33062, {{135986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33063, {{135987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33064, {{135988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33065, {{135990, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33066, {{135992, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33068, {{136004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33069, {{136010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33070, {{136013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33071, {{136015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33072, {{136016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33073, {{136017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33074, {{136019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33075, {{136020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33076, {{136032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33077, {{136033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33078, {{136036, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33079, {{136037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33080, {{136038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33081, {{136039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33082, {{136040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{33083, {{136041, 1, 0, 0},{136041, 1, 1, 16}, {0, 0, 0, 0}}},{33083, {{136041, 1, 0, 0},{136041, 1, 1, 16}, {0, 0, 0, 0}}},{57235, {{136041, 1, 0, 0},{136041, 1, 1, 17}, {0, 0, 0, 0}}},{57235, {{136041, 1, 0, 0},{136041, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{33084, {{136042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33085, {{136043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33086, {{136044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33087, {{136045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{33088, {{136046, 1, 0, 0},{136046, 1, 1, 16}, {0, 0, 0, 0}}},{33088, {{136046, 1, 0, 0},{136046, 1, 1, 16}, {0, 0, 0, 0}}},{33089, {{136046, 1, 0, 0},{136046, 1, 1, 17}, {0, 0, 0, 0}}},{33089, {{136046, 1, 0, 0},{136046, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{33090, {{136047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33091, {{136048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33092, {{136049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33093, {{136050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33094, {{136051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33095, {{136052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33096, {{136053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33097, {{136054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33099, {{136058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33100, {{136059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59454, {{136060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33101, {{136072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33102, {{136077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33103, {{136078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33104, {{136080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33105, {{136082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33106, {{136084, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33107, {{136085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33108, {{136087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33109, {{136103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33110, {{136104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33111, {{136105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33112, {{136106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33113, {{136107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33114, {{136108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33115, {{136110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33116, {{136111, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33117, {{136112, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33118, {{136113, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33119, {{136114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33120, {{136115, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33121, {{136116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33122, {{136117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33123, {{136118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33124, {{136119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33125, {{136120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33126, {{136121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33127, {{136122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33128, {{136123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33129, {{136124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33130, {{136125, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33131, {{136126, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33132, {{136127, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33133, {{136128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33134, {{136129, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33135, {{136130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33136, {{136131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33137, {{136132, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33139, {{136135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33140, {{136136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33141, {{136151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33142, {{136155, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33143, {{136169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33144, {{136170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33145, {{136171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33146, {{136173, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33147, {{136174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33148, {{136175, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33149, {{136176, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33150, {{136177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33151, {{136178, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33152, {{136179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33153, {{136180, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33154, {{136181, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33155, {{136182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33156, {{136183, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33157, {{136184, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33158, {{136201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33159, {{136206, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33160, {{136208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33161, {{136209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33162, {{136213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33163, {{136219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33164, {{136220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33165, {{136221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33166, {{136222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33167, {{136223, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33168, {{136224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33169, {{136225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33170, {{136226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33171, {{136227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33172, {{136228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33174, {{136229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33175, {{136230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33176, {{136233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33177, {{136234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33178, {{136235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33179, {{136242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33180, {{136243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33181, {{136249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33182, {{136254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33183, {{136258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33184, {{136263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33186, {{136265, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33187, {{136266, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33188, {{136271, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33189, {{136280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33190, {{136281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33191, {{136282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33192, {{136283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33193, {{136284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{33194, {{136286, 1, 0, 0},{136286, 1, 1, 17}, {0, 0, 0, 0}}},{33194, {{136286, 1, 0, 0},{136286, 1, 1, 17}, {0, 0, 0, 0}}},{33195, {{136286, 1, 0, 0},{136286, 1, 1, 16}, {0, 0, 0, 0}}},{33195, {{136286, 1, 0, 0},{136286, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {4, {{33196, {{136287, 1, 0, 0},{136287, 1, 1, 17}, {0, 0, 0, 0}}},{33196, {{136287, 1, 0, 0},{136287, 1, 1, 17}, {0, 0, 0, 0}}},{33197, {{136287, 1, 0, 0},{136287, 1, 1, 16}, {0, 0, 0, 0}}},{33197, {{136287, 1, 0, 0},{136287, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{33198, {{136288, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33199, {{136289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33200, {{136290, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33201, {{136291, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33202, {{136292, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33203, {{136294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{33204, {{136301, 1, 0, 0},{136301, 1, 1, 18}, {0, 0, 0, 0}}},{33204, {{136301, 1, 0, 0},{136301, 1, 1, 18}, {0, 0, 0, 0}}},{33205, {{136301, 1, 0, 0},{136301, 1, 1, 17}, {0, 0, 0, 0}}},{33205, {{136301, 1, 0, 0},{136301, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{33206, {{136302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33207, {{136310, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33208, {{136311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33209, {{136312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33210, {{136313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33211, {{136314, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33212, {{136322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59457, {{136323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33213, {{136326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33214, {{136328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {2, {{33215, {{136329, 1, 0, 0}, {0, 0, 0, 0}}},{33216, {{136329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33217, {{136330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33218, {{136331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33219, {{136332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33220, {{136333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33221, {{136334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33222, {{136335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33223, {{136336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33224, {{136337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33225, {{136338, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33226, {{136339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33227, {{136342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33228, {{136353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33229, {{136355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33230, {{136356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33231, {{136360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33232, {{136367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33233, {{136368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33234, {{136369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33235, {{136370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33236, {{136371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33238, {{136380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33239, {{136381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33237, {{136383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33240, {{136393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33241, {{136394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33242, {{136397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33243, {{136400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33244, {{136401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33245, {{136402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33246, {{136403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33247, {{136404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33248, {{136406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33249, {{136408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33250, {{136409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33251, {{136410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33252, {{136411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33253, {{136419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{33254, {{136420, 1, 0, 0},{136420, 1, 1, 16}, {0, 0, 0, 0}}},{33254, {{136420, 1, 0, 0},{136420, 1, 1, 16}, {0, 0, 0, 0}}},{33255, {{136420, 1, 0, 0},{136420, 1, 1, 17}, {0, 0, 0, 0}}},{33255, {{136420, 1, 0, 0},{136420, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33256, {{136422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33257, {{136428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33258, {{136432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33259, {{136433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33260, {{136434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33261, {{136435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33262, {{136436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33263, {{136437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33264, {{136438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59045, {{136439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33265, {{136444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33266, {{136447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33267, {{136448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33268, {{136449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33269, {{136450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33270, {{136455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33271, {{136456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33272, {{136457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33273, {{136458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33274, {{136459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33275, {{136463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33276, {{136469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33277, {{136471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33278, {{136472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33279, {{136473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33280, {{136474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33281, {{136475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33282, {{136476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33283, {{136477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33284, {{136478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33285, {{136479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33286, {{136483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33287, {{136484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33288, {{136485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33289, {{136486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33290, {{136488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33291, {{136489, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33292, {{136490, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33293, {{136491, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33294, {{136492, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33295, {{136495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33296, {{136496, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33297, {{136497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33298, {{136499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33299, {{136500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33300, {{136501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33301, {{136502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33302, {{136503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33303, {{136505, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33304, {{136507, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33305, {{136508, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33306, {{136510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33307, {{136513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33308, {{136518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33309, {{136520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33310, {{136521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33311, {{136523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33312, {{136524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33313, {{136525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33314, {{136526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33315, {{136528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33316, {{136529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{33317, {{136530, 1, 0, 0},{136530, 1, 1, 16}, {0, 0, 0, 0}}},{33317, {{136530, 1, 0, 0},{136530, 1, 1, 16}, {0, 0, 0, 0}}},{33318, {{136530, 1, 0, 0},{136530, 1, 1, 17}, {0, 0, 0, 0}}},{33318, {{136530, 1, 0, 0},{136530, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{33319, {{136531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33320, {{136533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{33321, {{136534, 1, 0, 0},{136534, 1, 1, 16}, {0, 0, 0, 0}}},{33321, {{136534, 1, 0, 0},{136534, 1, 1, 16}, {0, 0, 0, 0}}},{57252, {{136534, 1, 0, 0},{136534, 1, 1, 17}, {0, 0, 0, 0}}},{57252, {{136534, 1, 0, 0},{136534, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{33322, {{136535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33323, {{136536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33324, {{136538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33325, {{136539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33326, {{136540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33327, {{136541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33328, {{136542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33329, {{136543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33330, {{136544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33331, {{136545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33332, {{136546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33333, {{136547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33334, {{136551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33335, {{136552, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{33336, {{136553, 1, 0, 0},{136553, 1, 1, 16}, {0, 0, 0, 0}}},{33336, {{136553, 1, 0, 0},{136553, 1, 1, 16}, {0, 0, 0, 0}}},{33337, {{136553, 1, 0, 0},{136553, 1, 1, 17}, {0, 0, 0, 0}}},{33337, {{136553, 1, 0, 0},{136553, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{33338, {{136554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33339, {{136555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33340, {{136556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33341, {{136557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33342, {{136558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33343, {{136559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57258, {{136560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33344, {{136562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33345, {{136564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33346, {{136566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33347, {{136569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33348, {{136570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33349, {{136571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33351, {{136572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33350, {{136573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33352, {{136574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33353, {{136575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33354, {{136576, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33355, {{136577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33356, {{136579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33357, {{136580, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33358, {{136581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33359, {{136582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33360, {{136583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33361, {{136584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33362, {{136586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33363, {{136587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33364, {{136588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33365, {{136589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33366, {{136592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33367, {{136593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33368, {{136594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33369, {{136595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33370, {{136596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33371, {{136597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33372, {{136598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33373, {{136599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33374, {{136604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33375, {{136605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33376, {{136606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33377, {{136607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33378, {{136610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33379, {{136611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33380, {{136612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33381, {{136615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33382, {{136616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33383, {{136617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33384, {{136618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33385, {{136619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33386, {{136620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33387, {{136621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33388, {{136623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33389, {{136624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33390, {{136626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33391, {{136627, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33395, {{136628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33393, {{136630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33392, {{136632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33396, {{136633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33397, {{136634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33398, {{136637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33399, {{136638, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33400, {{136639, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33401, {{136641, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33402, {{136643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33403, {{136646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33404, {{136648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33405, {{136652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33406, {{136653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33407, {{136655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33408, {{136657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{33409, {{136658, 1, 0, 0},{136658, 1, 1, 16}, {0, 0, 0, 0}}},{33409, {{136658, 1, 0, 0},{136658, 1, 1, 16}, {0, 0, 0, 0}}},{57270, {{136658, 1, 0, 0},{136658, 1, 1, 17}, {0, 0, 0, 0}}},{57270, {{136658, 1, 0, 0},{136658, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{33410, {{136659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33411, {{136661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33412, {{136663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33413, {{136666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33414, {{136668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33415, {{136669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33416, {{136670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33417, {{136671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33418, {{136672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33419, {{136673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33420, {{136674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33421, {{136677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33422, {{136678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33423, {{136679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33424, {{136680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33425, {{136691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33426, {{136692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33427, {{136693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33428, {{136695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33429, {{136696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33430, {{136697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33431, {{136698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33432, {{136699, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33433, {{136700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33434, {{136701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33435, {{136702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33436, {{136703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33437, {{136706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33438, {{136707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33439, {{136708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{33440, {{136710, 1, 0, 0},{136710, 1, 1, 16}, {0, 0, 0, 0}}},{33440, {{136710, 1, 0, 0},{136710, 1, 1, 16}, {0, 0, 0, 0}}},{33441, {{136710, 1, 0, 0},{136710, 1, 1, 17}, {0, 0, 0, 0}}},{33441, {{136710, 1, 0, 0},{136710, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{33442, {{136711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33443, {{136713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33444, {{136715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33445, {{136716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33446, {{136717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33447, {{136718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33448, {{136719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33449, {{136720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33450, {{136729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33451, {{136731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33452, {{136733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33453, {{136735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33454, {{136736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33455, {{136737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33456, {{136738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33457, {{136739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33458, {{136740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33459, {{136745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33460, {{136747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33461, {{136752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33462, {{136754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33463, {{136756, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33464, {{136757, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33465, {{136758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33466, {{136759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33467, {{136760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33468, {{136761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33469, {{136762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33470, {{136763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33471, {{136764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33472, {{136769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33473, {{136771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33474, {{136772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33475, {{136773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33476, {{136775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33477, {{136776, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33478, {{136777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33479, {{136778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33480, {{136780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33481, {{136782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33482, {{136783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33483, {{136784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33484, {{136785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33485, {{136786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33486, {{136787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33487, {{136789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33488, {{136790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33489, {{136791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33490, {{136792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33491, {{136793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33492, {{136794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33493, {{136796, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33494, {{136797, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33495, {{136800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33496, {{136801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33497, {{136804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33498, {{136808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33499, {{136809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33500, {{136810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33501, {{136811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33502, {{136812, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33503, {{136813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33504, {{136814, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33505, {{136822, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33506, {{136824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33507, {{136825, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33508, {{136826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33509, {{136827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33510, {{136828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33511, {{136833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33512, {{136837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33513, {{136838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33514, {{136839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33515, {{136840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33516, {{136841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33517, {{136842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33518, {{136844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33519, {{136847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33520, {{136848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33521, {{136849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33523, {{136852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33525, {{136855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33526, {{136857, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33527, {{136859, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33528, {{136860, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33529, {{136863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33530, {{136864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33531, {{136868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33532, {{136871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33533, {{136872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33534, {{136877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33536, {{136878, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33535, {{136879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33537, {{136884, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33538, {{136887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33539, {{136888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33540, {{136889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33541, {{136890, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33542, {{136891, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33543, {{136892, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33544, {{136893, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33545, {{136894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33546, {{136908, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33548, {{136909, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33549, {{136910, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33551, {{136912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33552, {{136914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33553, {{136919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33554, {{136920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33555, {{136921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33556, {{136922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33557, {{136923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33558, {{136924, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33559, {{136925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33560, {{136926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33561, {{136927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33562, {{136928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33563, {{136929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33564, {{136930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33565, {{136949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33566, {{136950, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33567, {{136951, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33568, {{136953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33569, {{136957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33570, {{136958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33571, {{136959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33572, {{136960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33573, {{136961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33574, {{136962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33575, {{136963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33576, {{136964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33577, {{136965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{33578, {{136966, 1, 0, 0},{136966, 1, 1, 18}, {0, 0, 0, 0}}},{33578, {{136966, 1, 0, 0},{136966, 1, 1, 18}, {0, 0, 0, 0}}},{33579, {{136966, 1, 0, 0},{136966, 1, 1, 17}, {0, 0, 0, 0}}},{33579, {{136966, 1, 0, 0},{136966, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{33580, {{136967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33581, {{136968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33582, {{136969, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33583, {{136970, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33584, {{136971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33585, {{136972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33586, {{136973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33587, {{136974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33588, {{136987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33589, {{136988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33590, {{136989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33591, {{136991, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33592, {{136993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33593, {{136994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33594, {{136995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33595, {{136997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33596, {{136998, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33597, {{136999, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33598, {{137000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33599, {{137001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33600, {{137002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33601, {{137003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33602, {{137004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33603, {{137005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33604, {{137006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33605, {{137007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33606, {{137008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33607, {{137009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33608, {{137010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33609, {{137011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33610, {{137012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33611, {{137013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33612, {{137014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57301, {{137015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33613, {{137016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33614, {{137021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33615, {{137026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33616, {{137032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33617, {{137033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33618, {{137034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33619, {{137035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33620, {{137036, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33621, {{137037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33622, {{137038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33623, {{137039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33624, {{137041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33625, {{137042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33626, {{137043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33627, {{137047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33628, {{137049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33629, {{137050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33630, {{137051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33631, {{137052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33632, {{137053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33633, {{137054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33634, {{137055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33635, {{137056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33636, {{137057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33637, {{137058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{33638, {{137060, 1, 0, 0},{137060, 1, 1, 16}, {0, 0, 0, 0}}},{33638, {{137060, 1, 0, 0},{137060, 1, 1, 16}, {0, 0, 0, 0}}},{57303, {{137060, 1, 0, 0},{137060, 1, 1, 17}, {0, 0, 0, 0}}},{57303, {{137060, 1, 0, 0},{137060, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{33639, {{137061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33640, {{137062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33641, {{137063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33642, {{137064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33643, {{137065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33644, {{137066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33645, {{137067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33646, {{137074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33647, {{137104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33648, {{137105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33649, {{137107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33650, {{137108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33651, {{137112, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33652, {{137113, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33653, {{137114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33654, {{137115, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33655, {{137116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33656, {{137118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33657, {{137119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33658, {{137121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33659, {{137122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33660, {{137123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33661, {{137124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33662, {{137126, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33663, {{137127, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33664, {{137128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33665, {{137129, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33666, {{137130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33667, {{137131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33668, {{137132, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33669, {{137133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33670, {{137134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33671, {{137135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33672, {{137164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33673, {{137168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33674, {{137169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33675, {{137170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33676, {{137172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33677, {{137174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33678, {{137175, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33679, {{137176, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33680, {{137181, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33681, {{137182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33682, {{137183, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33683, {{137184, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33684, {{137185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33685, {{137186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33686, {{137187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33687, {{137188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33688, {{137189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33689, {{137190, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33690, {{137192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33691, {{137193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33692, {{137194, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33693, {{137195, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33694, {{137196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33695, {{137197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33696, {{137198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33697, {{137199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33698, {{137200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33699, {{137202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33700, {{137203, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33701, {{137204, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33703, {{137213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33704, {{137223, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33705, {{137224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33706, {{137227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33707, {{137232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33708, {{137233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33709, {{137234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33710, {{137235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33702, {{137238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33713, {{137241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33714, {{137242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33715, {{137243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33716, {{137244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33717, {{137246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33718, {{137247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33719, {{137248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33711, {{137249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33720, {{137250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33721, {{137253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33712, {{137254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33722, {{137271, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33723, {{137272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33724, {{137277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33725, {{137281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33726, {{137282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33727, {{137283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33728, {{137287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33729, {{137288, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33730, {{137289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33731, {{137290, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33732, {{137291, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33733, {{137292, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33734, {{137293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33735, {{137295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33736, {{137296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33737, {{137297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33738, {{137298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33739, {{137299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33740, {{137300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33741, {{137301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33742, {{137302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33743, {{137303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33758, {{137304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33744, {{137305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33745, {{137306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33746, {{137307, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33747, {{137308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33748, {{137311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33750, {{137317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33751, {{137320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33752, {{137337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33753, {{137338, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33754, {{137340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33755, {{137341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33756, {{137342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33757, {{137343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33759, {{137347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33760, {{137350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33761, {{137351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33762, {{137352, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33763, {{137353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33764, {{137354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33765, {{137355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33766, {{137361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33767, {{137362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33768, {{137363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33769, {{137364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33770, {{137365, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33771, {{137366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33772, {{137367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{33773, {{137368, 1, 0, 0},{137368, 1, 1, 16}, {0, 0, 0, 0}}},{33773, {{137368, 1, 0, 0},{137368, 1, 1, 16}, {0, 0, 0, 0}}},{33774, {{137368, 1, 0, 0},{137368, 1, 1, 17}, {0, 0, 0, 0}}},{33774, {{137368, 1, 0, 0},{137368, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{33775, {{137369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33776, {{137370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33777, {{137371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33778, {{137372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33779, {{137373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33780, {{137379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33781, {{137391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33782, {{137394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33783, {{137395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33784, {{137396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33785, {{137397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33786, {{137399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33787, {{137400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33788, {{137401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33789, {{137405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33790, {{137408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33791, {{137409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33792, {{137410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33793, {{137411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33794, {{137414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33795, {{137415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33796, {{137416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33797, {{137417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33798, {{137419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33799, {{137420, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33800, {{137421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33801, {{137422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33802, {{137423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33803, {{137424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33804, {{137425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33805, {{137426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33806, {{137428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33807, {{137429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33809, {{137433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33810, {{137440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33811, {{137441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33812, {{137442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33813, {{137443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33814, {{137444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33815, {{137446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33816, {{137449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{33817, {{137450, 1, 0, 0},{137450, 1, 1, 17}, {0, 0, 0, 0}}},{33817, {{137450, 1, 0, 0},{137450, 1, 1, 17}, {0, 0, 0, 0}}},{33818, {{137450, 1, 0, 0},{137450, 1, 1, 16}, {0, 0, 0, 0}}},{33818, {{137450, 1, 0, 0},{137450, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{33819, {{137451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33820, {{137452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33821, {{137453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33822, {{137454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33823, {{137455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33824, {{137456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33825, {{137457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33826, {{137459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33827, {{137463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33828, {{137464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33829, {{137465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33830, {{137468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33831, {{137469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33832, {{137470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33833, {{137471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33834, {{137472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33835, {{137473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33836, {{137480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33837, {{137483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33838, {{137484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33839, {{137487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33840, {{137494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33841, {{137495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33842, {{137496, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33843, {{137497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33844, {{137498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33845, {{137499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33847, {{137505, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33848, {{137506, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33849, {{137507, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33850, {{137508, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33851, {{137509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33852, {{137512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33854, {{137515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33855, {{137516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33856, {{137518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33857, {{137519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33858, {{137521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33859, {{137522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33860, {{137526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33853, {{137528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33861, {{137529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33862, {{137530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33863, {{137532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33864, {{137533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33865, {{137534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33866, {{137535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33867, {{137536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33868, {{137537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33869, {{137539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33870, {{137541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33871, {{137542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33872, {{137543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33873, {{137544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33874, {{137545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33875, {{137547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33876, {{137549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33877, {{137550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33878, {{137551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33879, {{137552, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33880, {{137553, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33881, {{137559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33882, {{137563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33883, {{137564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33884, {{137565, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33885, {{137566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33886, {{137574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33887, {{137575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33888, {{137576, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33889, {{137577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33892, {{137578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33893, {{137581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33894, {{137587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33895, {{137590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33896, {{137591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33897, {{137593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33898, {{137594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33899, {{137597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33900, {{137599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33901, {{137600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33902, {{137601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33903, {{137603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33904, {{137604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33905, {{137605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33906, {{137606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33907, {{137610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33908, {{137611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33909, {{137612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33910, {{137613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33911, {{137614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33912, {{137615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33913, {{137617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33914, {{137620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33915, {{137621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33916, {{137623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33917, {{137624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33918, {{137625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33919, {{137627, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33920, {{137628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33921, {{137629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33922, {{137631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33923, {{137632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33924, {{137633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33925, {{137634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33926, {{137635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33927, {{137640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33928, {{137641, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33929, {{137642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33930, {{137645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33931, {{137649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33932, {{137655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33933, {{137657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33934, {{137658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33935, {{137659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33936, {{137660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33937, {{137661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33938, {{137662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33939, {{137663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33940, {{137667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33941, {{137671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{33942, {{137672, 1, 0, 0},{137672, 1, 1, 16}, {0, 0, 0, 0}}},{33942, {{137672, 1, 0, 0},{137672, 1, 1, 16}, {0, 0, 0, 0}}},{33943, {{137672, 1, 0, 0},{137672, 1, 1, 17}, {0, 0, 0, 0}}},{33943, {{137672, 1, 0, 0},{137672, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33944, {{137675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33945, {{137676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33946, {{137677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33947, {{137678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33948, {{137679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33949, {{137682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33950, {{137684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33951, {{137685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33952, {{137686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33953, {{137687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33954, {{137688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33955, {{137689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33956, {{137690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33957, {{137692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33958, {{137697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33959, {{137698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33960, {{137699, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33961, {{137700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33962, {{137701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33963, {{137702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33964, {{137704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33965, {{137705, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33966, {{137706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33967, {{137707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33968, {{137708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33969, {{137709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33970, {{137710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33971, {{137711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33972, {{137712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{33973, {{137713, 1, 0, 0},{137713, 1, 1, 16}, {0, 0, 0, 0}}},{33973, {{137713, 1, 0, 0},{137713, 1, 1, 16}, {0, 0, 0, 0}}},{57330, {{137713, 1, 0, 0},{137713, 1, 1, 17}, {0, 0, 0, 0}}},{57330, {{137713, 1, 0, 0},{137713, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33974, {{137721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33975, {{137722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33976, {{137723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33977, {{137724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59504, {{137725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33979, {{137728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33981, {{137730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33982, {{137731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33983, {{137732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33984, {{137733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33985, {{137734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33986, {{137735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33988, {{137736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33989, {{137737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33990, {{137738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{33991, {{137739, 1, 0, 0},{137739, 1, 1, 16}, {0, 0, 0, 0}}},{33991, {{137739, 1, 0, 0},{137739, 1, 1, 16}, {0, 0, 0, 0}}},{57336, {{137739, 1, 0, 0},{137739, 1, 1, 17}, {0, 0, 0, 0}}},{57336, {{137739, 1, 0, 0},{137739, 1, 1, 17}, {0, 0, 0, 0}}},{57337, {{137739, 1, 0, 0},{137739, 1, 1, 18}, {0, 0, 0, 0}}},{57337, {{137739, 1, 0, 0},{137739, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33992, {{137742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33993, {{137743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33994, {{137747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33995, {{137748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33996, {{137750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33987, {{137751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{33997, {{137754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33998, {{137755, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33999, {{137756, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34000, {{137758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34001, {{137759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34002, {{137760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34003, {{137762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34004, {{137763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34006, {{137765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34007, {{137766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34008, {{137767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34009, {{137768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34010, {{137769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34011, {{137778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34012, {{137780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34013, {{137781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34015, {{137782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34016, {{137783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34017, {{137787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34018, {{137788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34019, {{137789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34020, {{137790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34021, {{137791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34022, {{137792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34023, {{137793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34025, {{137794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34026, {{137795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34027, {{137796, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34028, {{137797, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34029, {{137798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34030, {{137799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34031, {{137800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34032, {{137801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34033, {{137807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34034, {{137817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34035, {{137819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34036, {{137823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34037, {{137824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34038, {{137825, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34039, {{137826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34041, {{137827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34042, {{137828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34043, {{137829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34044, {{137830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34045, {{137831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34046, {{137832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34047, {{137833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34048, {{137834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34049, {{137837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34050, {{137846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34051, {{137847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34052, {{137848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34053, {{137851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34054, {{137853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34055, {{137854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34056, {{137855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34057, {{137856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34058, {{137857, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34059, {{137858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34060, {{137859, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34061, {{137860, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34062, {{137861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34063, {{137862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34064, {{137868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34065, {{137870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34066, {{137871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34067, {{137872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34068, {{137873, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34069, {{137874, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34070, {{137875, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34071, {{137878, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34072, {{137880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34073, {{137881, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34074, {{137883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34075, {{137886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34076, {{137887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34077, {{137888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34078, {{137889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{34079, {{137890, 1, 0, 0},{137890, 1, 1, 16}, {0, 0, 0, 0}}},{34079, {{137890, 1, 0, 0},{137890, 1, 1, 16}, {0, 0, 0, 0}}},{34080, {{137890, 1, 0, 0},{137890, 1, 1, 17}, {0, 0, 0, 0}}},{34080, {{137890, 1, 0, 0},{137890, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34081, {{137892, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34082, {{137893, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34083, {{137904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34084, {{137905, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34085, {{137906, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34086, {{137907, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34087, {{137908, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34088, {{137909, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34089, {{137910, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34090, {{137912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34091, {{137916, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34092, {{137917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34093, {{137918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34094, {{137919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34095, {{137920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34096, {{137921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34097, {{137922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34098, {{137923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34099, {{137924, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34100, {{137925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34101, {{137926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34102, {{137938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34103, {{137939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34104, {{137946, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34105, {{137947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34106, {{137948, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34107, {{137949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34108, {{137950, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34109, {{137951, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34110, {{137956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34111, {{137957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34112, {{137961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34113, {{137964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34114, {{137965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34115, {{137966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34116, {{137967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34117, {{137968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34118, {{137969, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34119, {{137971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34120, {{137972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34121, {{137973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34122, {{137975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34123, {{137976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34124, {{137977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34125, {{137979, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34126, {{137980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34127, {{137981, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34128, {{137982, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34129, {{137985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34130, {{137986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34131, {{137987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34132, {{137988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34133, {{137989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34134, {{137990, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34135, {{137991, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34136, {{137993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34137, {{137994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34138, {{137995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34139, {{137996, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34140, {{137997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34141, {{137998, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34142, {{138000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34143, {{138001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34144, {{138002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34145, {{138003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34146, {{138005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34147, {{138006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34148, {{138007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34149, {{138008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34150, {{138009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34151, {{138010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34152, {{138011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34153, {{138012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34154, {{138013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34155, {{138014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34156, {{138015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34157, {{138016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34158, {{138020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34159, {{138021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34160, {{138023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34161, {{138024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34162, {{138026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34163, {{138027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34164, {{138028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34165, {{138030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34166, {{138031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34167, {{138032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34168, {{138033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34169, {{138034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34170, {{138035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34172, {{138036, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34173, {{138037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{57349, {{138038, 1, 0, 0},{138038, 1, 1, 16}, {0, 0, 0, 0}}},{57349, {{138038, 1, 0, 0},{138038, 1, 1, 16}, {0, 0, 0, 0}}},{59514, {{138038, 1, 0, 0},{138038, 1, 1, 17}, {0, 0, 0, 0}}},{59514, {{138038, 1, 0, 0},{138038, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34174, {{138040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34175, {{138041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34176, {{138043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34177, {{138045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34178, {{138046, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34179, {{138049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34180, {{138050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34181, {{138054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34182, {{138055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34183, {{138056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34184, {{138058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34185, {{138060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{34186, {{138062, 1, 0, 0},{138062, 1, 1, 17}, {0, 0, 0, 0}}},{34186, {{138062, 1, 0, 0},{138062, 1, 1, 17}, {0, 0, 0, 0}}},{34187, {{138062, 1, 0, 0},{138062, 1, 1, 16}, {0, 0, 0, 0}}},{34187, {{138062, 1, 0, 0},{138062, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34188, {{138064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34189, {{138065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34190, {{138066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34191, {{138067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34192, {{138069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10288, {{138071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34193, {{138073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34195, {{138074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34196, {{138075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34197, {{138077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34198, {{138078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34199, {{138079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34200, {{138080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34201, {{138082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34202, {{138083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34203, {{138085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34204, {{138089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34206, {{138091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34207, {{138093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34208, {{138094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34209, {{138095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34210, {{138099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34211, {{138100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34212, {{138103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34213, {{138104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34214, {{138105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34215, {{138108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34216, {{138109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34217, {{138110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34218, {{138111, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34219, {{138112, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34220, {{138113, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34221, {{138114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34222, {{138126, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34224, {{138128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34225, {{138133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34226, {{138134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34227, {{138138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34228, {{138142, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34229, {{138145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34230, {{138150, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34231, {{138151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34232, {{138160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34233, {{138161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34234, {{138171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34235, {{138175, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10302, {{138177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34236, {{138178, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34237, {{138180, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34239, {{138183, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34240, {{138184, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34241, {{138185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34242, {{138186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34244, {{138188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34245, {{138189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34246, {{138190, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34248, {{138191, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34249, {{138192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{782, {{138193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34250, {{138194, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34251, {{138196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34252, {{138198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34253, {{138200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34254, {{138201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34259, {{138202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34255, {{138203, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34256, {{138204, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34257, {{138205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34258, {{138208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34260, {{138210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34261, {{138213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34262, {{138215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34263, {{138216, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34264, {{138217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34266, {{138219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{34265, {{138221, 1, 0, 0},{138221, 1, 1, 16}, {0, 0, 0, 0}}},{34265, {{138221, 1, 0, 0},{138221, 1, 1, 16}, {0, 0, 0, 0}}},{34267, {{138221, 1, 0, 0},{138221, 1, 1, 17}, {0, 0, 0, 0}}},{34267, {{138221, 1, 0, 0},{138221, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{34268, {{138222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34269, {{138226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34270, {{138227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34271, {{138228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34272, {{138229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34273, {{138230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34274, {{138231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34275, {{138232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34276, {{138234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34277, {{138235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34278, {{138238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34279, {{138239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34280, {{138240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34281, {{138242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34282, {{138243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34283, {{138246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34284, {{138247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34285, {{138250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34286, {{138251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34287, {{138252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34289, {{138254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{34292, {{138258, 1, 0, 0},{138258, 1, 1, 16}, {0, 0, 0, 0}}},{34292, {{138258, 1, 0, 0},{138258, 1, 1, 16}, {0, 0, 0, 0}}},{34293, {{138258, 1, 0, 0},{138258, 1, 1, 17}, {0, 0, 0, 0}}},{34293, {{138258, 1, 0, 0},{138258, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{34288, {{138259, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34291, {{138260, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34290, {{138261, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34294, {{138262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34297, {{138263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34296, {{138266, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34295, {{138268, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34298, {{138269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34299, {{138271, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34300, {{138273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34301, {{138275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34302, {{138276, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34303, {{138277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34304, {{138278, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34305, {{138280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34306, {{138281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34307, {{138282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34308, {{138283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34309, {{138284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34310, {{138285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34311, {{138286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34312, {{138287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{34313, {{138289, 1, 0, 0},{138289, 1, 1, 17}, {0, 0, 0, 0}}},{34313, {{138289, 1, 0, 0},{138289, 1, 1, 17}, {0, 0, 0, 0}}},{57366, {{138289, 1, 0, 0},{138289, 1, 1, 16}, {0, 0, 0, 0}}},{57366, {{138289, 1, 0, 0},{138289, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{34314, {{138290, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34315, {{138294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34316, {{138296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34317, {{138298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34318, {{138299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34319, {{138300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34320, {{138301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34321, {{138302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34322, {{138303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34323, {{138304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34324, {{138305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34325, {{138306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34326, {{138307, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34327, {{138309, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34328, {{138311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34329, {{138312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34330, {{138313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34331, {{138314, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34332, {{138315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34333, {{138316, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34334, {{138317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34335, {{138318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34336, {{138319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34337, {{138320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34338, {{138321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34339, {{138322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34340, {{138323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34341, {{138324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34342, {{138325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34343, {{138326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34344, {{138332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34345, {{138333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34346, {{138335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34347, {{138336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34348, {{138337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34349, {{138338, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34350, {{138339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34351, {{138340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34352, {{138341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34353, {{138342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34354, {{138343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34355, {{138344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34356, {{138353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34357, {{138354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34358, {{138355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34359, {{138356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34360, {{138359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34361, {{138361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34362, {{138362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34363, {{138363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34364, {{138364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34365, {{138365, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34366, {{138366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34367, {{138367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34368, {{138368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34369, {{138369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34370, {{138370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34371, {{138371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34372, {{138380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34373, {{138381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34374, {{138382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34375, {{138384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34376, {{138385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34377, {{138386, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34378, {{138388, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34379, {{138389, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34380, {{138390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34381, {{138391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34382, {{138392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34383, {{138393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34384, {{138394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34385, {{138395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34386, {{138396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34387, {{138397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34388, {{138398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34389, {{138399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34390, {{138400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34391, {{138401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34392, {{138403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34393, {{138409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34394, {{138410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34395, {{138412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34396, {{138413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34397, {{138414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34398, {{138415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34399, {{138416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34400, {{138417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34401, {{138418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34402, {{138419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34403, {{138420, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34404, {{138421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34405, {{138422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34406, {{138427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34407, {{138428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34408, {{138430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34409, {{138431, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34410, {{138434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34411, {{138436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34412, {{138437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34413, {{138438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34414, {{138439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34415, {{138440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34416, {{138441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34417, {{138445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34418, {{138447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34419, {{138448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34420, {{138449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34421, {{138456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34422, {{138458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34423, {{138462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34424, {{138463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34425, {{138464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34426, {{138465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34427, {{138471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34429, {{138475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34430, {{138476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34428, {{138477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34431, {{138478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34432, {{138484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34433, {{138488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34434, {{138491, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34435, {{138494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34436, {{138495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34437, {{138496, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34438, {{138497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34439, {{138498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34440, {{138500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34441, {{138501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34442, {{138502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34443, {{138504, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34444, {{138507, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34445, {{138508, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34446, {{138509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34447, {{138510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34448, {{138511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34449, {{138512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34450, {{138513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34451, {{138514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34452, {{138515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34453, {{138517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34455, {{138518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34456, {{138520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34457, {{138523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34458, {{138524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34459, {{138525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34460, {{138526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34461, {{138527, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34462, {{138528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34463, {{138529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34464, {{138532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34465, {{138533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34466, {{138534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34467, {{138535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34468, {{138536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34469, {{138537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34470, {{138539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34471, {{138540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34472, {{138541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34473, {{138542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34474, {{138543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34475, {{138544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34476, {{138545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34477, {{138546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34478, {{138547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34479, {{138552, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34480, {{138555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34481, {{138556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34482, {{138558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34483, {{138559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34484, {{138560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34485, {{138561, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34487, {{138565, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34488, {{138570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34489, {{138571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34490, {{138574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34491, {{138575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34492, {{138577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34493, {{138579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34494, {{138580, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34495, {{138581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34496, {{138582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34497, {{138583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34498, {{138584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34500, {{138585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34501, {{138586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34502, {{138587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34503, {{138588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34504, {{138589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34505, {{138590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34506, {{138594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34508, {{138604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34509, {{138606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34510, {{138607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34511, {{138608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34512, {{138609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59521, {{138611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34513, {{138612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34514, {{138613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34515, {{138614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34516, {{138615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34517, {{138616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34518, {{138617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34519, {{138618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34520, {{138620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34521, {{138622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34522, {{138626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34523, {{138627, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34524, {{138628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34525, {{138629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34526, {{138630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34527, {{138631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34528, {{138632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34529, {{138633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34530, {{138634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34531, {{138635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34533, {{138637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34534, {{138638, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34535, {{138639, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34536, {{138640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34537, {{138641, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34538, {{138642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34539, {{138643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34540, {{138650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34541, {{138652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{34542, {{138657, 1, 0, 0},{138657, 1, 1, 18}, {0, 0, 0, 0}}},{34542, {{138657, 1, 0, 0},{138657, 1, 1, 18}, {0, 0, 0, 0}}},{34543, {{138657, 1, 0, 0},{138657, 1, 1, 17}, {0, 0, 0, 0}}},{34543, {{138657, 1, 0, 0},{138657, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{34544, {{138658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34545, {{138659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34546, {{138660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34547, {{138665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34548, {{138666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34549, {{138667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34550, {{138668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34551, {{138669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34552, {{138671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34553, {{138672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34554, {{138673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34569, {{138674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34555, {{138675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34556, {{138678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34557, {{138679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34558, {{138680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34559, {{138693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34560, {{138694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34561, {{138695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34562, {{138696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34563, {{138697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34564, {{138700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34565, {{138701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34566, {{138702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34567, {{138706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34568, {{138707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34570, {{138709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34571, {{138710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34572, {{138711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34573, {{138712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34574, {{138713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34575, {{138714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34576, {{138715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34577, {{138717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34578, {{138718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34579, {{138719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34580, {{138720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34581, {{138721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34582, {{138722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34583, {{138723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{34584, {{138724, 1, 0, 0},{138724, 1, 1, 16}, {0, 0, 0, 0}}},{34584, {{138724, 1, 0, 0},{138724, 1, 1, 16}, {0, 0, 0, 0}}},{34585, {{138724, 1, 0, 0},{138724, 1, 1, 17}, {0, 0, 0, 0}}},{34585, {{138724, 1, 0, 0},{138724, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{34586, {{138725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{34588, {{138726, 1, 0, 0},{138726, 1, 1, 16}, {0, 0, 0, 0}}},{34588, {{138726, 1, 0, 0},{138726, 1, 1, 16}, {0, 0, 0, 0}}},{34589, {{138726, 1, 0, 0},{138726, 1, 1, 17}, {0, 0, 0, 0}}},{34589, {{138726, 1, 0, 0},{138726, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{34590, {{138727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34591, {{138728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34592, {{138729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34593, {{138730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34594, {{138731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34596, {{138732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34597, {{138733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34598, {{138734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34599, {{138744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34600, {{138746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34601, {{138747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34602, {{138748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34603, {{138751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57383, {{138753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34604, {{138754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34605, {{138760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34606, {{138761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34607, {{138762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34608, {{138763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34609, {{138764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34610, {{138765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34611, {{138766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34612, {{138769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34613, {{138770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34614, {{138771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34615, {{138772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34616, {{138773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34617, {{138774, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34618, {{138775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34619, {{138776, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34620, {{138777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34621, {{138778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34622, {{138779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34623, {{138780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34624, {{138782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34625, {{138783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34626, {{138784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34627, {{138785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34628, {{138786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34629, {{138787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34630, {{138788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34631, {{138789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34632, {{138790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34633, {{138791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34634, {{138792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34635, {{138793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34636, {{138794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34637, {{138795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34638, {{138796, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34639, {{138803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34640, {{138804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34641, {{138805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34643, {{138819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34644, {{138822, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34645, {{138826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34646, {{138827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34647, {{138830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34648, {{138831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34649, {{138835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34650, {{138836, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34651, {{138838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34652, {{138840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34653, {{138850, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34654, {{138851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34655, {{138852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34656, {{138853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34657, {{138854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34658, {{138855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34659, {{138856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34660, {{138857, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34661, {{138858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34662, {{138859, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34663, {{138860, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34664, {{138861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34665, {{138862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34666, {{138863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34667, {{138864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34668, {{138866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34669, {{138867, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34670, {{138869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34671, {{138870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34672, {{138871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34673, {{138872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34674, {{138873, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34675, {{138874, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34676, {{138875, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34677, {{138876, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34678, {{138877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34679, {{138878, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34680, {{138886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34681, {{138887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34682, {{138896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34683, {{138898, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34684, {{138899, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34685, {{138900, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34686, {{138901, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34687, {{138902, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34688, {{138903, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34689, {{138905, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34690, {{138906, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34691, {{138909, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34693, {{138911, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34694, {{138913, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34695, {{138919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34696, {{138923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34697, {{138924, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34699, {{138925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34700, {{138926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34701, {{138927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34702, {{138928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34703, {{138929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34704, {{138930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34705, {{138932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34706, {{138933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34707, {{138934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34708, {{138935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34709, {{138936, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34710, {{138937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34711, {{138938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34712, {{138940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34713, {{138941, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34714, {{138942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34715, {{138943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34717, {{138949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34718, {{138955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34719, {{138956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34720, {{138957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34721, {{138958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34722, {{138961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34723, {{138965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34724, {{138967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34725, {{138968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34726, {{138969, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34727, {{138970, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34728, {{138972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34729, {{138973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34730, {{138974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34731, {{138975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34733, {{138976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34734, {{138977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34735, {{138978, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34736, {{138979, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34737, {{138980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34739, {{138981, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34740, {{138983, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34741, {{138984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34742, {{138985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34743, {{138986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34744, {{138988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34745, {{138989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34746, {{138990, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34747, {{138991, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34748, {{138992, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34749, {{138993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34750, {{138994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34751, {{138996, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34752, {{138997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34753, {{139002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34754, {{139006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34755, {{139007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34756, {{139011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34757, {{139017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34758, {{139018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34759, {{139019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34760, {{139020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34761, {{139021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34762, {{139022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34764, {{139024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34765, {{139025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34766, {{139026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34767, {{139027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34768, {{139028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34769, {{139029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34770, {{139030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34771, {{139031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34772, {{139032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{34773, {{139033, 1, 0, 0},{139033, 1, 1, 16}, {0, 0, 0, 0}}},{34773, {{139033, 1, 0, 0},{139033, 1, 1, 16}, {0, 0, 0, 0}}},{34774, {{139033, 1, 0, 0},{139033, 1, 1, 17}, {0, 0, 0, 0}}},{34774, {{139033, 1, 0, 0},{139033, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{34775, {{139034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34776, {{139035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34777, {{139036, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34778, {{139037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34779, {{139038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34780, {{139045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34781, {{139055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34782, {{139056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34783, {{139062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34784, {{139066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{33185, {{139067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34785, {{139068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34786, {{139069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34787, {{139070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34788, {{139071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34789, {{139072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34790, {{139073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34791, {{139074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34792, {{139075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34793, {{139076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34795, {{139077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34796, {{139078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34797, {{139079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34798, {{139080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34799, {{139081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34800, {{139082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34801, {{139088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34802, {{139090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34803, {{139091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34804, {{139092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34805, {{139093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34806, {{139096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34807, {{139097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34808, {{139098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34809, {{139099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34810, {{139100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34811, {{139101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34812, {{139102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34813, {{139103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34814, {{139104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34815, {{139105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34816, {{139106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34817, {{139107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34818, {{139108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34819, {{139109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34820, {{139110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34821, {{139111, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34822, {{139116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34823, {{139117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34824, {{139122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34825, {{139125, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{34826, {{139126, 1, 0, 0},{139126, 1, 1, 18}, {0, 0, 0, 0}}},{34826, {{139126, 1, 0, 0},{139126, 1, 1, 18}, {0, 0, 0, 0}}},{34827, {{139126, 1, 0, 0},{139126, 1, 1, 17}, {0, 0, 0, 0}}},{34827, {{139126, 1, 0, 0},{139126, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{34828, {{139127, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34829, {{139128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34830, {{139130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34831, {{139132, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34832, {{139133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34833, {{139135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34834, {{139136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57393, {{139139, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34835, {{139140, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34836, {{139141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34837, {{139142, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34838, {{139143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34839, {{139144, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34840, {{139145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34841, {{139146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34842, {{139147, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34843, {{139148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34844, {{139149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34845, {{139152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34846, {{139153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34847, {{139157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34848, {{139159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34849, {{139160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34850, {{139161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34851, {{139162, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34852, {{139163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34853, {{139164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34854, {{139165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34855, {{139166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34856, {{139167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34858, {{139173, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34859, {{139174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34860, {{139175, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34861, {{139177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34862, {{139182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34863, {{139183, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34864, {{139184, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34865, {{139185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34866, {{139186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34867, {{139191, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34868, {{139195, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34869, {{139196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34870, {{139197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34871, {{139198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34872, {{139201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34873, {{139202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34874, {{139206, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34875, {{139207, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34876, {{139208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34877, {{139209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34878, {{139210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34879, {{139211, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34880, {{139212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34881, {{139214, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34882, {{139215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34883, {{139217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34884, {{139220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34885, {{139221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{34886, {{139222, 1, 0, 0},{139222, 1, 1, 16}, {0, 0, 0, 0}}},{34886, {{139222, 1, 0, 0},{139222, 1, 1, 16}, {0, 0, 0, 0}}},{34887, {{139222, 1, 0, 0},{139222, 1, 1, 17}, {0, 0, 0, 0}}},{34887, {{139222, 1, 0, 0},{139222, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34888, {{139224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34889, {{139225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34890, {{139227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34891, {{139228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34892, {{139229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34893, {{139232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34894, {{139233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34895, {{139234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34896, {{139235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34897, {{139236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34898, {{139237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34899, {{139238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{34900, {{139239, 1, 0, 0},{139239, 1, 1, 16}, {0, 0, 0, 0}}},{34900, {{139239, 1, 0, 0},{139239, 1, 1, 16}, {0, 0, 0, 0}}},{34901, {{139239, 1, 0, 0},{139239, 1, 1, 17}, {0, 0, 0, 0}}},{34901, {{139239, 1, 0, 0},{139239, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{34902, {{139240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{34903, {{139241, 1, 0, 0},{139241, 1, 1, 16}, {0, 0, 0, 0}}},{34903, {{139241, 1, 0, 0},{139241, 1, 1, 16}, {0, 0, 0, 0}}},{34904, {{139241, 1, 0, 0},{139241, 1, 1, 17}, {0, 0, 0, 0}}},{34904, {{139241, 1, 0, 0},{139241, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{34905, {{139242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10765, {{139243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34908, {{139244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34909, {{139245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{34910, {{139246, 1, 0, 0},{139246, 1, 1, 16}, {0, 0, 0, 0}}},{34910, {{139246, 1, 0, 0},{139246, 1, 1, 16}, {0, 0, 0, 0}}},{34911, {{139246, 1, 0, 0},{139246, 1, 1, 18}, {0, 0, 0, 0}}},{34911, {{139246, 1, 0, 0},{139246, 1, 1, 18}, {0, 0, 0, 0}}},{34912, {{139246, 1, 0, 0},{139246, 1, 1, 17}, {0, 0, 0, 0}}},{34912, {{139246, 1, 0, 0},{139246, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{34913, {{139247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34914, {{139248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34915, {{139249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30003, {{139251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34916, {{139253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34917, {{139255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34918, {{139256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34919, {{139257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34920, {{139258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34921, {{139259, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34922, {{139260, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34923, {{139261, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34924, {{139262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34925, {{139263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34926, {{139264, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34927, {{139265, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34928, {{139266, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34929, {{139267, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34930, {{139268, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34931, {{139271, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34932, {{139272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34933, {{139273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34934, {{139274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34935, {{139275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34936, {{139276, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34937, {{139277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34938, {{139278, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34939, {{139280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34940, {{139281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34941, {{139282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34942, {{139283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34943, {{139285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34944, {{139287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34945, {{139288, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34946, {{139289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34947, {{139292, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34948, {{139293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34949, {{139294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34950, {{139296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34951, {{139297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34952, {{139299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{34953, {{139305, 1, 0, 0},{139305, 1, 1, 16}, {0, 0, 0, 0}}},{34953, {{139305, 1, 0, 0},{139305, 1, 1, 16}, {0, 0, 0, 0}}},{34954, {{139305, 1, 0, 0},{139305, 1, 1, 17}, {0, 0, 0, 0}}},{34954, {{139305, 1, 0, 0},{139305, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{34955, {{139306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34956, {{139311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{7920, {{139316, 1, 0, 0},{139316, 1, 1, 16}, {0, 0, 0, 0}}},{7920, {{139316, 1, 0, 0},{139316, 1, 1, 16}, {0, 0, 0, 0}}},{34957, {{139316, 1, 0, 0},{139316, 1, 1, 17}, {0, 0, 0, 0}}},{34957, {{139316, 1, 0, 0},{139316, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{34958, {{139317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34959, {{139318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{34960, {{139319, 1, 0, 0},{139319, 1, 1, 17}, {0, 0, 0, 0}}},{34960, {{139319, 1, 0, 0},{139319, 1, 1, 17}, {0, 0, 0, 0}}},{34961, {{139319, 1, 0, 0},{139319, 1, 1, 16}, {0, 0, 0, 0}}},{34961, {{139319, 1, 0, 0},{139319, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34962, {{139321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34963, {{139322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34964, {{139327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34965, {{139328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34966, {{139329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34967, {{139330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34968, {{139331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34969, {{139332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34970, {{139333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34971, {{139340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34972, {{139342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34973, {{139343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34974, {{139344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34975, {{139346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34976, {{139347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34977, {{139348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34978, {{139349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34979, {{139350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34980, {{139351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34981, {{139352, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34982, {{139353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34983, {{139354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34984, {{139355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34985, {{139357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34986, {{139358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34987, {{139359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34988, {{139360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34989, {{139362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34990, {{139363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34991, {{139364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34992, {{139365, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34993, {{139366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34994, {{139367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34995, {{139368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34996, {{139369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34997, {{139370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34998, {{139371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{34999, {{139372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35000, {{139373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35001, {{139374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35002, {{139375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35003, {{139376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35004, {{139377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35005, {{139379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35006, {{139380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35007, {{139381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35008, {{139382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35009, {{139383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35010, {{139385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35011, {{139386, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35012, {{139387, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35013, {{139388, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35014, {{139389, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35015, {{139390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35016, {{139391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35017, {{139392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35018, {{139393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35019, {{139394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35020, {{139395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35021, {{139396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35022, {{139398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35023, {{139399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35024, {{139401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35025, {{139402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35026, {{139403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35027, {{139405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35028, {{139407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35029, {{139408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35030, {{139409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35031, {{139410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35032, {{139411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35033, {{139412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35034, {{139413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35035, {{139414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35036, {{139415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35037, {{139416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35041, {{139417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35038, {{139418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35039, {{139421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35040, {{139423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35042, {{139429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35043, {{139430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35044, {{139432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35045, {{139433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35046, {{139434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35047, {{139435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35048, {{139436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35049, {{139438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35050, {{139439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35051, {{139441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35052, {{139442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35053, {{139443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35054, {{139444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35055, {{139445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35056, {{139446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35057, {{139447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35058, {{139448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35059, {{139449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35060, {{139450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35061, {{139451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35062, {{139452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35063, {{139453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35064, {{139454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35065, {{139455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35066, {{139456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35067, {{139457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{10851, {{139458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35068, {{139462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35069, {{139467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35070, {{139468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35071, {{139469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35072, {{139470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35073, {{139471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35074, {{139472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35075, {{139473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35076, {{139474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35077, {{139475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35078, {{139476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35079, {{139477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35080, {{139478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35081, {{139479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35082, {{139480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35083, {{139481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35084, {{139482, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35085, {{139483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35086, {{139484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35087, {{139485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35088, {{139490, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35089, {{139492, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35090, {{139493, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35091, {{139494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35092, {{139495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35093, {{139496, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35094, {{139498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35095, {{139499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35096, {{139500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35097, {{139501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35098, {{139502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35099, {{139503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35100, {{139504, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35101, {{139505, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35102, {{139506, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35103, {{139507, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35104, {{139508, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35105, {{139509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35106, {{139510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35107, {{139511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35108, {{139512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35109, {{139513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35110, {{139514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35111, {{139520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35112, {{139524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35113, {{139525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35114, {{139526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35115, {{139527, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35116, {{139529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35117, {{139530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35118, {{139531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35119, {{139532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35120, {{139533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35121, {{139534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35122, {{139535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35123, {{139536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35124, {{139538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35125, {{139539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35126, {{139540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35127, {{139541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35128, {{139542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35129, {{139543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35130, {{139544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35131, {{139546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35132, {{139547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35133, {{139550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35134, {{139551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35135, {{139552, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35136, {{139553, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35137, {{139554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35138, {{139555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35139, {{139556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35140, {{139557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35141, {{139558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35142, {{139559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35143, {{139564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35144, {{139568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35145, {{139569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35146, {{139570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35147, {{139571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35148, {{139572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35149, {{139574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35150, {{139575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35151, {{139576, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35152, {{139577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35153, {{139578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35154, {{139579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35155, {{139580, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35156, {{139581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35157, {{139582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35158, {{139583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35159, {{139584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35160, {{139585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35161, {{139586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35162, {{139591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35163, {{139592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35164, {{139593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35165, {{139594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35166, {{139595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35167, {{139597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35168, {{139598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35169, {{139599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35170, {{139600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35171, {{139601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35172, {{139602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35173, {{139603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35174, {{139604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35175, {{139605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35176, {{139607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35177, {{139609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35178, {{139610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35179, {{139611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35180, {{139612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35181, {{139613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35182, {{139614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35183, {{139615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35184, {{139616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35185, {{139617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35186, {{139618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35187, {{139620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35188, {{139622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35189, {{139624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35190, {{139625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35191, {{139626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35192, {{139627, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35193, {{139629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35194, {{139630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35195, {{139631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35196, {{139632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35197, {{139633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35198, {{139637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35199, {{139640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35200, {{139641, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35201, {{139642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35202, {{139643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35203, {{139645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35204, {{139646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35205, {{139647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35206, {{139648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35207, {{139650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35208, {{139651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35209, {{139652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35210, {{139653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35211, {{139655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35212, {{139656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35213, {{139657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35214, {{139658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35215, {{139659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35216, {{139660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35217, {{139661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35218, {{139662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35219, {{139665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35220, {{139666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35221, {{139670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35222, {{139673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35224, {{139675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35225, {{139676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35223, {{139678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{35226, {{139679, 1, 0, 0},{139679, 1, 1, 16}, {0, 0, 0, 0}}},{35226, {{139679, 1, 0, 0},{139679, 1, 1, 16}, {0, 0, 0, 0}}},{35227, {{139679, 1, 0, 0},{139679, 1, 1, 17}, {0, 0, 0, 0}}},{35227, {{139679, 1, 0, 0},{139679, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35228, {{139682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30524, {{139683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35229, {{139686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35230, {{139692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35231, {{139695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{35232, {{139696, 1, 0, 0},{139696, 1, 1, 16}, {0, 0, 0, 0}}},{35232, {{139696, 1, 0, 0},{139696, 1, 1, 16}, {0, 0, 0, 0}}},{56823, {{139696, 1, 0, 0},{139696, 1, 1, 17}, {0, 0, 0, 0}}},{56823, {{139696, 1, 0, 0},{139696, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35233, {{139698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35234, {{139699, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35235, {{139700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35236, {{139702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35237, {{139703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35238, {{139704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35239, {{139706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35240, {{139707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35241, {{139708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35242, {{139709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35243, {{139710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35244, {{139711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35245, {{139712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35247, {{139714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35246, {{139715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35248, {{139718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35249, {{139719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35250, {{139720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35251, {{139722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35252, {{139723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35253, {{139724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35254, {{139725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35255, {{139727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35256, {{139729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35257, {{139730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35258, {{139731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35259, {{139732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35261, {{139733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35262, {{139734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35263, {{139735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35264, {{139736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35265, {{139737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35266, {{139738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35267, {{139740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35268, {{139741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35269, {{139742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35270, {{139743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35271, {{139744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35272, {{139747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35273, {{139748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35274, {{139749, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35275, {{139750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35276, {{139751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35277, {{139752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35278, {{139753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35279, {{139754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35280, {{139755, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35281, {{139757, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35282, {{139759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35283, {{139760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35284, {{139762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35285, {{139763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35286, {{139764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35287, {{139765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35288, {{139766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35289, {{139767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35290, {{139768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35291, {{139769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35292, {{139770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35293, {{139771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35294, {{139772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35295, {{139778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35296, {{139779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35297, {{139780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35298, {{139781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35299, {{139782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35300, {{139783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35301, {{139784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35302, {{139785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35303, {{139786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35304, {{139787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35305, {{139788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35306, {{139794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35307, {{139795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35308, {{139799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35309, {{139800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35312, {{139801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35313, {{139802, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35314, {{139803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35315, {{139804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35316, {{139805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35317, {{139806, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35318, {{139807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35319, {{139808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35320, {{139809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35321, {{139810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35322, {{139811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35323, {{139812, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35324, {{139813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35325, {{139814, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35326, {{139823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35327, {{139826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35328, {{139827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35329, {{139828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35330, {{139829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35331, {{139830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35332, {{139831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35333, {{139832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35334, {{139833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35335, {{139834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{35336, {{139835, 1, 0, 0},{139835, 1, 1, 16}, {0, 0, 0, 0}}},{35336, {{139835, 1, 0, 0},{139835, 1, 1, 16}, {0, 0, 0, 0}}},{57443, {{139835, 1, 0, 0},{139835, 1, 1, 17}, {0, 0, 0, 0}}},{57443, {{139835, 1, 0, 0},{139835, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{35337, {{139836, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35338, {{139837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35339, {{139838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35340, {{139839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35341, {{139840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35342, {{139841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35343, {{139842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35344, {{139843, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35345, {{139844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35346, {{139845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35347, {{139846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35348, {{139847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35349, {{139848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35350, {{139857, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35351, {{139859, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35352, {{139861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{35353, {{139862, 1, 0, 0},{139862, 1, 1, 16}, {0, 0, 0, 0}}},{35353, {{139862, 1, 0, 0},{139862, 1, 1, 16}, {0, 0, 0, 0}}},{59250, {{139862, 1, 0, 0},{139862, 1, 1, 17}, {0, 0, 0, 0}}},{59250, {{139862, 1, 0, 0},{139862, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{35354, {{139863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35355, {{139864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35356, {{139865, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35357, {{139866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35358, {{139867, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35359, {{139868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35360, {{139869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35361, {{139870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35362, {{139871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35363, {{139872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35364, {{139873, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35365, {{139874, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35366, {{139875, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35367, {{139876, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35368, {{139877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35369, {{139878, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35370, {{139879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35371, {{139880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35372, {{139881, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35373, {{139882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35374, {{139883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35375, {{139884, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35376, {{139885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35377, {{139892, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35378, {{139893, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35379, {{139894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35380, {{139898, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35381, {{139903, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35382, {{139904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35383, {{139905, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35384, {{139906, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35385, {{139907, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35386, {{139908, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35387, {{139909, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35388, {{139910, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35389, {{139911, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35390, {{139914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35392, {{139918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35393, {{139920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35394, {{139922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35395, {{139923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35396, {{139925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35397, {{139926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35398, {{139927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35399, {{139928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35400, {{139929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35401, {{139930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35402, {{139931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35403, {{139932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35404, {{139933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35406, {{139943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35407, {{139946, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35408, {{139949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35409, {{139951, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35410, {{139952, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35411, {{139953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35412, {{139954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35413, {{139955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35414, {{139956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35415, {{139957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35416, {{139962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35417, {{139963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35418, {{139968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35419, {{139969, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35420, {{139970, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35421, {{139971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35422, {{139972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35423, {{139973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35424, {{139974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35425, {{139975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35426, {{139980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35427, {{139985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35428, {{139986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35429, {{139987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35430, {{139988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35431, {{139994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35432, {{139996, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35433, {{139997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35434, {{140001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35435, {{140002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35436, {{140003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35437, {{140006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35438, {{140008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35440, {{140010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35441, {{140013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35442, {{140014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35444, {{140015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35445, {{140016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{35446, {{140017, 1, 0, 0},{140017, 1, 1, 16}, {0, 0, 0, 0}}},{35446, {{140017, 1, 0, 0},{140017, 1, 1, 16}, {0, 0, 0, 0}}},{58011, {{140017, 1, 0, 0},{140017, 1, 1, 17}, {0, 0, 0, 0}}},{58011, {{140017, 1, 0, 0},{140017, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{35447, {{140018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35448, {{140025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35457, {{140027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35450, {{140028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35451, {{140029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35452, {{140030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{35453, {{140031, 1, 0, 0},{140031, 1, 1, 16}, {0, 0, 0, 0}}},{35453, {{140031, 1, 0, 0},{140031, 1, 1, 16}, {0, 0, 0, 0}}},{35454, {{140031, 1, 0, 0},{140031, 1, 1, 17}, {0, 0, 0, 0}}},{35454, {{140031, 1, 0, 0},{140031, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{35455, {{140032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35456, {{140034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35449, {{140035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35458, {{140036, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35459, {{140037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35460, {{140039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35461, {{140040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35462, {{140041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35463, {{140043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35464, {{140045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35465, {{140049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35466, {{140050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35467, {{140051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35468, {{140052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35469, {{140053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{35470, {{140059, 1, 0, 0},{140059, 1, 1, 16}, {0, 0, 0, 0}}},{35470, {{140059, 1, 0, 0},{140059, 1, 1, 16}, {0, 0, 0, 0}}},{35471, {{140059, 1, 0, 0},{140059, 1, 1, 17}, {0, 0, 0, 0}}},{35471, {{140059, 1, 0, 0},{140059, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35472, {{140061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35473, {{140062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35476, {{140070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35477, {{140072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35478, {{140075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35479, {{140076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35480, {{140077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35481, {{140078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35482, {{140079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{35483, {{140081, 1, 0, 0},{140081, 1, 1, 16}, {0, 0, 0, 0}}},{35483, {{140081, 1, 0, 0},{140081, 1, 1, 16}, {0, 0, 0, 0}}},{35484, {{140081, 1, 0, 0},{140081, 1, 1, 17}, {0, 0, 0, 0}}},{35484, {{140081, 1, 0, 0},{140081, 1, 1, 17}, {0, 0, 0, 0}}},{35485, {{140081, 1, 0, 0},{140081, 1, 1, 18}, {0, 0, 0, 0}}},{35485, {{140081, 1, 0, 0},{140081, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35486, {{140083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35487, {{140084, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35488, {{140085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35489, {{140088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35490, {{140089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35491, {{140090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35492, {{140091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35493, {{140092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35494, {{140093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35495, {{140094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35496, {{140095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{35497, {{140097, 1, 0, 0},{140097, 1, 1, 16}, {0, 0, 0, 0}}},{35497, {{140097, 1, 0, 0},{140097, 1, 1, 16}, {0, 0, 0, 0}}},{35498, {{140097, 1, 0, 0},{140097, 1, 1, 17}, {0, 0, 0, 0}}},{35498, {{140097, 1, 0, 0},{140097, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{35499, {{140098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35500, {{140099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35501, {{140100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35502, {{140101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35503, {{140102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35504, {{140103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35505, {{140104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35506, {{140106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35507, {{140109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35508, {{140110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35509, {{140111, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35510, {{140112, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35511, {{140113, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35512, {{140114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35513, {{140115, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35514, {{140116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35515, {{140118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35516, {{140119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35517, {{140120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35518, {{140121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35519, {{140122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35520, {{140123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35521, {{140124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35522, {{140125, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35523, {{140126, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35524, {{140128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35525, {{140130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35526, {{140131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35527, {{140135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35528, {{140136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35529, {{140139, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35530, {{140140, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35531, {{140141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35532, {{140142, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35533, {{140143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35534, {{140144, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35535, {{140145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35536, {{140146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35537, {{140147, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35538, {{140150, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35539, {{140151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35540, {{140152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35541, {{140153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35542, {{140154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35543, {{140155, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35544, {{140156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35545, {{140157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35546, {{140159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35547, {{140160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35548, {{140161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35549, {{140162, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35550, {{140163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35551, {{140164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35552, {{140165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35554, {{140167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35555, {{140168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35556, {{140169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35557, {{140170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35558, {{140171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35559, {{140172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35560, {{140173, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35561, {{140174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35562, {{140176, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35563, {{140177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35564, {{140178, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35565, {{140179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35566, {{140180, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35567, {{140181, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35568, {{140182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35569, {{140183, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35570, {{140184, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35571, {{140185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35572, {{140186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35573, {{140188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35574, {{140189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35575, {{140191, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35576, {{140192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35577, {{140193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35578, {{140194, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35579, {{140195, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35580, {{140196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35581, {{140197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35582, {{140198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35583, {{140199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35584, {{140200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35585, {{140202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35586, {{140203, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35587, {{140204, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35588, {{140205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35593, {{140206, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57271, {{140207, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35589, {{140208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35590, {{140209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35591, {{140212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35592, {{140213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35594, {{140215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35595, {{140216, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35596, {{140217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35597, {{140218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35598, {{140219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35600, {{140225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35601, {{140226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35602, {{140227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35603, {{140228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35604, {{140229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35605, {{140230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35606, {{140231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35607, {{140233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35608, {{140237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35609, {{140238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35610, {{140242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35611, {{140243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35612, {{140244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35613, {{140245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35614, {{140246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35615, {{140247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35616, {{140248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35617, {{140249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35618, {{140250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35619, {{140252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35620, {{140253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35621, {{140254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35622, {{140255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35623, {{140260, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35625, {{140261, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35626, {{140262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35627, {{140263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35628, {{140264, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35629, {{140265, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35630, {{140266, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35631, {{140269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35632, {{140270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35633, {{140271, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35634, {{140272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35635, {{140273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35636, {{140274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35637, {{140275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35638, {{140276, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35639, {{140277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35640, {{140278, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35641, {{140279, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35642, {{140280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35643, {{140284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35644, {{140285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35645, {{140286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35646, {{140288, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35647, {{140289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35648, {{140290, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35649, {{140291, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35650, {{140292, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35651, {{140293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35652, {{140294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35653, {{140295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35654, {{140296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35655, {{140298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35656, {{140299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35657, {{140300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35658, {{140302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35659, {{140303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35660, {{140304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35661, {{140306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35662, {{140307, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35663, {{140308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35664, {{140309, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35665, {{140310, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35666, {{140314, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35667, {{140317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35668, {{140318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35669, {{140319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35670, {{140320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35671, {{140321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35672, {{140324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35679, {{140325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35674, {{140326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35675, {{140327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35677, {{140328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35678, {{140331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35673, {{140332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35680, {{140333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35681, {{140335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35682, {{140338, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35683, {{140339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35684, {{140340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35685, {{140342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35686, {{140343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35687, {{140344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35688, {{140345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35689, {{140346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35690, {{140347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35691, {{140348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35692, {{140350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35693, {{140351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35694, {{140352, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35696, {{140353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35697, {{140354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35698, {{140355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35699, {{140357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35700, {{140358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35701, {{140359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35702, {{140360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35703, {{140361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35704, {{140362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35705, {{140363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35706, {{140364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35707, {{140365, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35708, {{140368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35709, {{140369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35710, {{140370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35711, {{140371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35712, {{140372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35713, {{140373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35714, {{140374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35715, {{140375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35716, {{140376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35717, {{140377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35718, {{140378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35719, {{140379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35720, {{140380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35721, {{140381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35722, {{140382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35723, {{140383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35724, {{140384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35725, {{140385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35726, {{140386, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35727, {{140387, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35728, {{140388, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35729, {{140390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35730, {{140391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35731, {{140392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35732, {{140393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35733, {{140394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35734, {{140395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35735, {{140396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35736, {{140397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35737, {{140398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35738, {{140399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35739, {{140400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57481, {{140401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35740, {{140402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35741, {{140403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35742, {{140405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35743, {{140406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35744, {{140407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35745, {{140408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35746, {{140409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35747, {{140413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35748, {{140415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35749, {{140416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35750, {{140417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35751, {{140419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35752, {{140420, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35753, {{140422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35754, {{140424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35755, {{140425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35756, {{140426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35757, {{140429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35758, {{140431, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35759, {{140432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35760, {{140433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35761, {{140434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35762, {{140435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35763, {{140436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35764, {{140437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35765, {{140438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35766, {{140439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35767, {{140443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35768, {{140444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35769, {{140445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35770, {{140446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35771, {{140447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35772, {{140448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35773, {{140449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35774, {{140451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35775, {{140452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35776, {{140453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35777, {{140454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35778, {{140455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35779, {{140457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35780, {{140458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35781, {{140459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35782, {{140460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35783, {{140461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35784, {{140462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35785, {{140463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35786, {{140464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35787, {{140465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35788, {{140466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35789, {{140467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35790, {{140468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35791, {{140469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35792, {{140471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35793, {{140472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35794, {{140473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35795, {{140477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35796, {{140478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35797, {{140479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35798, {{140480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35799, {{140482, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35800, {{140483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35801, {{140484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35802, {{140485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35803, {{140486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35804, {{140487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35805, {{140488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35806, {{140489, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35807, {{140490, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35808, {{140491, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35809, {{140494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35810, {{140495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35811, {{140498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35812, {{140499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35813, {{140500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35814, {{140501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35815, {{140502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35816, {{140503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35817, {{140505, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35818, {{140508, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35819, {{140509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35820, {{140510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35821, {{140511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35822, {{140512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35823, {{140513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35824, {{140514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35825, {{140521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35826, {{140522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35827, {{140523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{35828, {{140525, 1, 0, 0},{140525, 1, 1, 17}, {0, 0, 0, 0}}},{35828, {{140525, 1, 0, 0},{140525, 1, 1, 17}, {0, 0, 0, 0}}},{35829, {{140525, 1, 0, 0},{140525, 1, 1, 16}, {0, 0, 0, 0}}},{35829, {{140525, 1, 0, 0},{140525, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{35830, {{140526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35831, {{140527, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35832, {{140528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35833, {{140529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35834, {{140530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35835, {{140531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35836, {{140532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35837, {{140533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35838, {{140534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35839, {{140535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35840, {{140536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35841, {{140539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35842, {{140540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35843, {{140541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35844, {{140542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35845, {{140543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35846, {{140544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35847, {{140545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35849, {{140546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35850, {{140547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35851, {{140548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35852, {{140549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35853, {{140550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35854, {{140551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35855, {{140552, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35856, {{140553, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35857, {{140554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35858, {{140555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35859, {{140556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35860, {{140557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35861, {{140558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35862, {{140559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35863, {{140560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35864, {{140561, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35865, {{140564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35866, {{140568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35867, {{140570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35868, {{140571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35869, {{140572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35870, {{140575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35871, {{140576, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35872, {{140577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35873, {{140578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35874, {{140579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35875, {{140580, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35876, {{140581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35877, {{140582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35878, {{140583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35879, {{140584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35880, {{140585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35881, {{140586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35882, {{140588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35883, {{140589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35884, {{140593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35885, {{140595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {2, {{35886, {{140596, 1, 0, 0}, {0, 0, 0, 0}}},{35887, {{140596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35888, {{140597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35889, {{140598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{35890, {{140599, 1, 0, 0},{140599, 1, 1, 16}, {0, 0, 0, 0}}},{35890, {{140599, 1, 0, 0},{140599, 1, 1, 16}, {0, 0, 0, 0}}},{68075, {{140599, 1, 0, 0},{140599, 1, 1, 17}, {0, 0, 0, 0}}},{68075, {{140599, 1, 0, 0},{140599, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{35891, {{140600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35892, {{140601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35893, {{140611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35894, {{140613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35895, {{140615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35896, {{140621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35897, {{140622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35898, {{140623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35899, {{140624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35900, {{140625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{35901, {{140626, 1, 0, 0},{140626, 1, 1, 16}, {0, 0, 0, 0}}},{35901, {{140626, 1, 0, 0},{140626, 1, 1, 16}, {0, 0, 0, 0}}},{35902, {{140626, 1, 0, 0},{140626, 1, 1, 17}, {0, 0, 0, 0}}},{35902, {{140626, 1, 0, 0},{140626, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{35903, {{140627, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35904, {{140628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35905, {{140629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35906, {{140631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35907, {{140632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35908, {{140633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35909, {{140634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35910, {{140635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35911, {{140636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35912, {{140637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35913, {{140638, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35914, {{140639, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35915, {{140646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35916, {{140647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35917, {{140648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35918, {{140651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35919, {{140652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35920, {{140653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35921, {{140655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35922, {{140656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35923, {{140657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35924, {{140658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35925, {{140662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35926, {{140663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35927, {{140665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35928, {{140666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35929, {{140667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35930, {{140668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35931, {{140669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35932, {{140670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35933, {{140675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35934, {{140676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35935, {{140677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35936, {{140680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35937, {{140681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35938, {{140682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35939, {{140683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{23887, {{140685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35940, {{140686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35941, {{140687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35942, {{140688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35943, {{140689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35944, {{140690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35945, {{140695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35946, {{140696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35947, {{140697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35948, {{140700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35949, {{140701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35950, {{140702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35951, {{140703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35952, {{140705, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35953, {{140706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35954, {{140707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35955, {{140708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35956, {{140709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35957, {{140710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35958, {{140711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35959, {{140713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35960, {{140714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35961, {{140715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35962, {{140716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35963, {{140717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35964, {{140722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35965, {{140723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35966, {{140724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35967, {{140725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35968, {{140726, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35969, {{140727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35970, {{140728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35971, {{140729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35972, {{140730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35973, {{140731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35974, {{140735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35975, {{140736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35976, {{140737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35977, {{140738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35978, {{140739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35979, {{140740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35980, {{140743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35981, {{140744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35982, {{140745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35983, {{140746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35984, {{140747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35985, {{140748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35986, {{140749, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35987, {{140750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35988, {{140751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35989, {{140752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35990, {{140753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35991, {{140754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35992, {{140755, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35993, {{140756, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35994, {{140757, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35995, {{140758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35996, {{140759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{35997, {{140760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{35999, {{140779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36000, {{140780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36001, {{140781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36002, {{140782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36003, {{140783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36004, {{140784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36005, {{140785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36006, {{140786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36007, {{140788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36008, {{140793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36009, {{140794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36010, {{140799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36011, {{140800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36012, {{140801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36013, {{140802, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36014, {{140803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36015, {{140804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36016, {{140805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36017, {{140806, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36018, {{140807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36019, {{140808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36020, {{140809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36021, {{140810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36022, {{140811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36023, {{140812, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36024, {{140813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36025, {{140814, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36026, {{140815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36027, {{140816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36028, {{140817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36029, {{140818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36030, {{140819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36031, {{140820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36032, {{140821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36033, {{140822, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36034, {{140831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36035, {{140832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36036, {{140834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36037, {{140835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36038, {{140838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36039, {{140839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36040, {{140840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36041, {{140841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36042, {{140842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36044, {{140852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36045, {{140853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36046, {{140854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36047, {{140855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36048, {{140856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36049, {{140857, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36050, {{140858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36051, {{140859, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36052, {{140860, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36053, {{140861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36054, {{140863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36055, {{140864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36056, {{140865, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36057, {{140866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36058, {{140867, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36059, {{140868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36060, {{140869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36061, {{140870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36062, {{140871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36063, {{140872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36064, {{140873, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36065, {{140874, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36067, {{140876, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36068, {{140877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36069, {{140878, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36071, {{140886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36072, {{140898, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36074, {{140910, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36075, {{140911, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36076, {{140912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36077, {{140913, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36078, {{140914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36080, {{140915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36081, {{140916, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36082, {{140917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36083, {{140918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36084, {{140919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36085, {{140920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36086, {{140921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36087, {{140922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36088, {{140923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36089, {{140924, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36090, {{140925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36091, {{140926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36092, {{140927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36093, {{140928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36094, {{140929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36095, {{140930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36096, {{140931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36098, {{140933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36099, {{140934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36100, {{140935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36101, {{140936, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36102, {{140937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36103, {{140938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36104, {{140939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36105, {{140940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36106, {{140941, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36107, {{140942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36108, {{140943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36109, {{140944, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36110, {{140945, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36111, {{140947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59611, {{140949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36114, {{140966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36115, {{140968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36116, {{140970, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36117, {{140971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36118, {{140973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36119, {{140974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36120, {{140975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36121, {{140976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36122, {{140986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36123, {{140987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36124, {{140993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36125, {{140994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36126, {{140995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36127, {{140996, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36128, {{140997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36129, {{140998, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36130, {{140999, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36131, {{141000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36132, {{141001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36133, {{141002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36134, {{141003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36135, {{141004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36136, {{141005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36137, {{141006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36138, {{141008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36139, {{141009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36140, {{141010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36141, {{141011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{36142, {{141012, 1, 0, 0},{141012, 1, 1, 16}, {0, 0, 0, 0}}},{36142, {{141012, 1, 0, 0},{141012, 1, 1, 16}, {0, 0, 0, 0}}},{36143, {{141012, 1, 0, 0},{141012, 1, 1, 17}, {0, 0, 0, 0}}},{36143, {{141012, 1, 0, 0},{141012, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{36144, {{141013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36145, {{141014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36146, {{141015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36147, {{141016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36148, {{141018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36149, {{141019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36150, {{141020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36151, {{141022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36152, {{141023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36153, {{141024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36154, {{141025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36155, {{141026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36156, {{141027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36157, {{141028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36158, {{141029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36159, {{141030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36160, {{141031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36161, {{141032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36162, {{141033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36163, {{141034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36164, {{141035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36165, {{141036, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36166, {{141037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36167, {{141038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36168, {{141043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36169, {{141070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36170, {{141078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36171, {{141079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36172, {{141082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36173, {{141091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36174, {{141092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36175, {{141093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36176, {{141094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36177, {{141096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36178, {{141097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36179, {{141098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36180, {{141099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36181, {{141100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36182, {{141101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36183, {{141102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36184, {{141103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36185, {{141104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36186, {{141105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36187, {{141106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36188, {{141107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36189, {{141108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36190, {{141109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36191, {{141110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36192, {{141111, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36193, {{141112, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36194, {{141113, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36195, {{141114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36196, {{141115, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36197, {{141116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36198, {{141117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36199, {{141118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36200, {{141119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36201, {{141120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36202, {{141121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36203, {{141122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36204, {{141124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36205, {{141125, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36206, {{141126, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36207, {{141127, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36208, {{141128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36209, {{141129, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36210, {{141130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36211, {{141131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36212, {{141132, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36213, {{141133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36214, {{141134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36215, {{141135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36216, {{141136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36217, {{141137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36218, {{141138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36219, {{141139, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36220, {{141140, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36221, {{141141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36222, {{141143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36223, {{141163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36225, {{141164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36227, {{141165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36228, {{141166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36229, {{141167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36230, {{141170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36231, {{141171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36232, {{141172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36233, {{141176, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59617, {{141180, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36234, {{141181, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36235, {{141191, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36236, {{141193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36237, {{141194, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36238, {{141195, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36239, {{141198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36240, {{141199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36241, {{141200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36243, {{141201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36244, {{141202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36245, {{141203, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36246, {{141204, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36247, {{141205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36248, {{141206, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36249, {{141207, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36250, {{141208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36251, {{141209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36252, {{141210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36253, {{141211, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36254, {{141212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36255, {{141213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36256, {{141214, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36257, {{141215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36258, {{141216, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36259, {{141217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36261, {{141218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36262, {{141219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36263, {{141220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36264, {{141221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36265, {{141222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36266, {{141223, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36267, {{141224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36268, {{141225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36269, {{141226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36270, {{141227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36271, {{141228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36272, {{141229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36273, {{141230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36274, {{141231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36275, {{141232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36276, {{141233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36277, {{141234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36278, {{141235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36279, {{141238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36280, {{141251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36281, {{141252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36282, {{141253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36283, {{141258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36284, {{141259, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36285, {{141276, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36286, {{141277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36287, {{141280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36288, {{141281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36289, {{141282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36290, {{141283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36291, {{141284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36292, {{141286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36293, {{141287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36294, {{141288, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36295, {{141289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36296, {{141290, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36297, {{141291, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36298, {{141292, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36324, {{141293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36299, {{141294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36300, {{141295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36301, {{141296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36302, {{141297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36303, {{141298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36304, {{141299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36305, {{141300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36306, {{141301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36307, {{141302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36308, {{141303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36309, {{141304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36310, {{141305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{36311, {{141306, 1, 0, 0},{141306, 1, 1, 16}, {0, 0, 0, 0}}},{36311, {{141306, 1, 0, 0},{141306, 1, 1, 16}, {0, 0, 0, 0}}},{57520, {{141306, 1, 0, 0},{141306, 1, 1, 17}, {0, 0, 0, 0}}},{57520, {{141306, 1, 0, 0},{141306, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{36312, {{141307, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36313, {{141308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36314, {{141309, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36315, {{141310, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36316, {{141311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36317, {{141312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36319, {{141339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36320, {{141340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36321, {{141341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36322, {{141342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36323, {{141343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36325, {{141345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36326, {{141355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36327, {{141356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36328, {{141357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36329, {{141358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36330, {{141361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36331, {{141362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36332, {{141363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36333, {{141364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{36334, {{141365, 1, 0, 0},{141365, 1, 1, 17}, {0, 0, 0, 0}}},{36334, {{141365, 1, 0, 0},{141365, 1, 1, 17}, {0, 0, 0, 0}}},{36335, {{141365, 1, 0, 0},{141365, 1, 1, 16}, {0, 0, 0, 0}}},{36335, {{141365, 1, 0, 0},{141365, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{36336, {{141366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36337, {{141367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36338, {{141368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36339, {{141369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36340, {{141370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36341, {{141371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36342, {{141372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36343, {{141373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36344, {{141374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36345, {{141375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36346, {{141376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36347, {{141377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36349, {{141378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{36350, {{141379, 1, 0, 0},{141379, 1, 1, 16}, {0, 0, 0, 0}}},{36350, {{141379, 1, 0, 0},{141379, 1, 1, 16}, {0, 0, 0, 0}}},{36351, {{141379, 1, 0, 0},{141379, 1, 1, 17}, {0, 0, 0, 0}}},{36351, {{141379, 1, 0, 0},{141379, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{36352, {{141380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36353, {{141381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36354, {{141382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36355, {{141383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36356, {{141384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36357, {{141385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36358, {{141386, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36359, {{141387, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36360, {{141390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36361, {{141391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36362, {{141392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36363, {{141393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36364, {{141394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36365, {{141395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36366, {{141396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36367, {{141397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36368, {{141398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36369, {{141399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36370, {{141400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36371, {{141401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{36372, {{141402, 1, 0, 0},{141402, 1, 1, 16}, {0, 0, 0, 0}}},{36372, {{141402, 1, 0, 0},{141402, 1, 1, 16}, {0, 0, 0, 0}}},{36373, {{141402, 1, 0, 0},{141402, 1, 1, 17}, {0, 0, 0, 0}}},{36373, {{141402, 1, 0, 0},{141402, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{36374, {{141403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36375, {{141410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36376, {{141412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36380, {{141419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36381, {{141424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36382, {{141427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36383, {{141428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36384, {{141429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36385, {{141432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36386, {{141433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36379, {{141435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36387, {{141437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36388, {{141438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36389, {{141439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36390, {{141442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36391, {{141452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36392, {{141453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36393, {{141458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36394, {{141459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{36395, {{141460, 1, 0, 0},{141460, 1, 1, 16}, {0, 0, 0, 0}}},{36395, {{141460, 1, 0, 0},{141460, 1, 1, 16}, {0, 0, 0, 0}}},{36396, {{141460, 1, 0, 0},{141460, 1, 1, 17}, {0, 0, 0, 0}}},{36396, {{141460, 1, 0, 0},{141460, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{36397, {{141461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36398, {{141462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36399, {{141463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36400, {{141464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36401, {{141465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36402, {{141466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36403, {{141467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36404, {{141468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36405, {{141469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36406, {{141470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36407, {{141471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36408, {{141472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36409, {{141473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36410, {{141474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36411, {{141475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36412, {{141476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36413, {{141477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36414, {{141478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36415, {{141479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36416, {{141483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36417, {{141489, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36419, {{141497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36420, {{141513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36421, {{141514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36422, {{141517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36423, {{141518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36424, {{141519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36425, {{141520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36426, {{141521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36427, {{141522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36428, {{141523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36429, {{141524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36430, {{141525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36431, {{141526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36432, {{141528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36433, {{141529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36434, {{141530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36435, {{141531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36436, {{141532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36437, {{141533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36438, {{141534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36439, {{141535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36440, {{141536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36441, {{141537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36442, {{141538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36443, {{141555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36444, {{141558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36446, {{141562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36447, {{141563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36448, {{141564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36449, {{141566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36450, {{141567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36451, {{141568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36452, {{141570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36453, {{141571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36454, {{141572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36457, {{141573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36458, {{141574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36459, {{141575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36460, {{141576, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36461, {{141577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36462, {{141578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36463, {{141579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36464, {{141580, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36465, {{141581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36466, {{141582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36467, {{141583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36468, {{141584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36469, {{141585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36470, {{141586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36471, {{141587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36472, {{141592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36473, {{141593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36474, {{141594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36475, {{141604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{36476, {{141606, 1, 0, 0},{141606, 1, 1, 16}, {0, 0, 0, 0}}},{36476, {{141606, 1, 0, 0},{141606, 1, 1, 16}, {0, 0, 0, 0}}},{36477, {{141606, 1, 0, 0},{141606, 1, 1, 17}, {0, 0, 0, 0}}},{36477, {{141606, 1, 0, 0},{141606, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{36478, {{141607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36479, {{141608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36480, {{141609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36481, {{141610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36482, {{141611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36483, {{141612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36484, {{141613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36485, {{141614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36486, {{141615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36487, {{141616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36488, {{141617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36489, {{141618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36490, {{141619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36491, {{141620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36492, {{141621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36493, {{141622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36494, {{141623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36495, {{141632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36496, {{141633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36497, {{141635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36498, {{141644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36499, {{141645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36501, {{141646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36502, {{141647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36503, {{141648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36504, {{141649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36505, {{141650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36506, {{141651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36507, {{141655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36508, {{141656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36509, {{141658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36512, {{141659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36513, {{141660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36514, {{141661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36515, {{141662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36516, {{141663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36517, {{141664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36518, {{141665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36519, {{141669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36520, {{141670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36521, {{141671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36522, {{141672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36523, {{141673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36524, {{141679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36525, {{141680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36526, {{141681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36527, {{141682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36528, {{141683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36529, {{141691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36530, {{141692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36531, {{141693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{36532, {{141701, 1, 0, 0},{141701, 1, 1, 16}, {0, 0, 0, 0}}},{36532, {{141701, 1, 0, 0},{141701, 1, 1, 16}, {0, 0, 0, 0}}},{36533, {{141701, 1, 0, 0},{141701, 1, 1, 17}, {0, 0, 0, 0}}},{36533, {{141701, 1, 0, 0},{141701, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36534, {{141704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36535, {{141705, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36536, {{141706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36537, {{141708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36538, {{141709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36539, {{141710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36540, {{141711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36541, {{141712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36542, {{141713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36543, {{141714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36544, {{141715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36545, {{141717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36546, {{141718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36547, {{141720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36548, {{141721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36549, {{141722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36550, {{141723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36551, {{141724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36552, {{141726, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36553, {{141727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36554, {{141728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36555, {{141729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36556, {{141730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57653, {{141731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36557, {{141734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36558, {{141735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36559, {{141736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36560, {{141737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36561, {{141739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36562, {{141740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36563, {{141744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36564, {{141745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36565, {{141746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36566, {{141747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36567, {{141748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36568, {{141750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36569, {{141751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36570, {{141752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36571, {{141753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36572, {{141754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36573, {{141755, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36574, {{141756, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36575, {{141757, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36576, {{141758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36577, {{141759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36578, {{141760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36579, {{141761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36580, {{141765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36581, {{141766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36582, {{141767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36583, {{141768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36584, {{141769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36585, {{141770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36586, {{141771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36587, {{141772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36590, {{141775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36591, {{141777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36592, {{141780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36593, {{141781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36594, {{141782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36595, {{141783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36596, {{141784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36597, {{141788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36598, {{141789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36599, {{141790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36600, {{141791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36601, {{141792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36589, {{141793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36602, {{141794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36603, {{141796, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36604, {{141797, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36605, {{141798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36606, {{141799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36607, {{141800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36608, {{141804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36609, {{141806, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36610, {{141808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36611, {{141809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36612, {{141811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36613, {{141812, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36614, {{141813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36615, {{141814, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36616, {{141815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36617, {{141816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36618, {{141817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36619, {{141818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36620, {{141820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36621, {{141821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36622, {{141822, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36623, {{141823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36624, {{141825, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36625, {{141826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36626, {{141831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36627, {{141832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36628, {{141833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36629, {{141834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36630, {{141835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36631, {{141836, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36632, {{141837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36633, {{141838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36634, {{141839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36635, {{141840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36636, {{141841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58897, {{141842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36637, {{141843, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36638, {{141844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36639, {{141846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36640, {{141847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36641, {{141849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36642, {{141850, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36643, {{141851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36644, {{141852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36645, {{141853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36646, {{141854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36647, {{141855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36648, {{141857, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36649, {{141859, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36650, {{141861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36652, {{141862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36653, {{141863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36654, {{141864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36655, {{141865, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36656, {{141866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36657, {{141870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36658, {{141871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36659, {{141872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36660, {{141873, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36661, {{141874, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36662, {{141875, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36663, {{141876, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36664, {{141879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36665, {{141880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36667, {{141881, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36668, {{141882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36669, {{141886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36670, {{141887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36671, {{141888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36672, {{141889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36673, {{141890, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36674, {{141892, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36675, {{141893, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36676, {{141894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36678, {{141896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36679, {{141899, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36680, {{141901, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36677, {{141903, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36682, {{141905, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36683, {{141906, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36684, {{141909, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36685, {{141910, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36686, {{141911, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36687, {{141912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36688, {{141913, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36689, {{141914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36690, {{141917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36691, {{141918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36692, {{141919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36693, {{141920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36694, {{141921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36695, {{141922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36696, {{141925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36697, {{141926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36698, {{141928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36699, {{141931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36700, {{141932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36701, {{141933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36702, {{141936, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36703, {{141937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36704, {{141938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36706, {{141943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36707, {{141944, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36708, {{141945, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36709, {{141946, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36710, {{141947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36711, {{141948, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36712, {{141954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36713, {{141955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36714, {{141956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36715, {{141959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36716, {{141960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36717, {{141961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36718, {{141962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36719, {{141963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36720, {{141965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36721, {{141966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36722, {{141967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36723, {{141968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36724, {{141969, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36725, {{141970, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36726, {{141971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36727, {{141972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36728, {{141973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36729, {{141974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36730, {{141975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36732, {{141989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36733, {{141990, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36734, {{141991, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36735, {{142004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36736, {{142005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36737, {{142007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36738, {{142008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36739, {{142010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36740, {{142011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36741, {{142012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36742, {{142013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36743, {{142014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36744, {{142015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36745, {{142016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36746, {{142018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36747, {{142032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36748, {{142033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36749, {{142034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36750, {{142050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36751, {{142053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36752, {{142054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36753, {{142055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36754, {{142056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36755, {{142057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36756, {{142059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36757, {{142060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36758, {{142061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36759, {{142062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36760, {{142063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36761, {{142064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36762, {{142066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36763, {{142068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36764, {{142069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36765, {{142070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36766, {{142071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36767, {{142072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36768, {{142073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36769, {{142074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36770, {{142075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36771, {{142076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36772, {{142077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36773, {{142078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36774, {{142079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36775, {{142080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36776, {{142083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36777, {{142086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36779, {{142110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36780, {{142122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36781, {{142123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36782, {{142124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36783, {{142129, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36784, {{142130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36785, {{142131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36786, {{142132, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36787, {{142133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36788, {{142134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36789, {{142135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36790, {{142136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36791, {{142137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36792, {{142139, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36793, {{142140, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36794, {{142141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36795, {{142142, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36796, {{142143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36797, {{142144, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36798, {{142145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36799, {{142146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36800, {{142147, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36801, {{142148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36802, {{142149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36803, {{142150, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36804, {{142151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36805, {{142152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36806, {{142153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36807, {{142154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36808, {{142155, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36809, {{142156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36810, {{142157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36811, {{142158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36812, {{142159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36813, {{142160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36814, {{142162, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36815, {{142163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36816, {{142180, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36817, {{142181, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36818, {{142182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36819, {{142183, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36820, {{142184, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36821, {{142209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36822, {{142210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36823, {{142211, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36824, {{142212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36825, {{142213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36826, {{142214, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36827, {{142215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36828, {{142216, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36829, {{142217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36830, {{142218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36831, {{142219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36832, {{142220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36833, {{142221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36834, {{142222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36835, {{142223, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36836, {{142224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36837, {{142225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36838, {{142226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36839, {{142227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36840, {{142228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36841, {{142229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36842, {{142230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36843, {{142231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36844, {{142232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36845, {{142233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36846, {{142234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36847, {{142236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36848, {{142237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36849, {{142238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36850, {{142239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36851, {{142240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36852, {{142241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36853, {{142242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36854, {{142243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36855, {{142244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36856, {{142246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36857, {{142247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36858, {{142248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36859, {{142254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36860, {{142277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36861, {{142279, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36862, {{142281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36863, {{142296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36864, {{142313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36865, {{142314, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36866, {{142315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36867, {{142316, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36868, {{142317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36869, {{142318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36870, {{142319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36871, {{142320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{36872, {{142321, 1, 0, 0},{142321, 1, 1, 16}, {0, 0, 0, 0}}},{36872, {{142321, 1, 0, 0},{142321, 1, 1, 16}, {0, 0, 0, 0}}},{57562, {{142321, 1, 0, 0},{142321, 1, 1, 17}, {0, 0, 0, 0}}},{57562, {{142321, 1, 0, 0},{142321, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{36873, {{142322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36874, {{142323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36875, {{142324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36876, {{142325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36877, {{142326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36878, {{142327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36879, {{142328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36880, {{142329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36881, {{142330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36882, {{142331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36883, {{142332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36884, {{142333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36885, {{142334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36886, {{142335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36887, {{142336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36888, {{142337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36889, {{142338, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36890, {{142339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36891, {{142340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36892, {{142341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36893, {{142342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36894, {{142343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36895, {{142344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36896, {{142345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36897, {{142346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36898, {{142347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36899, {{142349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36900, {{142350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36901, {{142356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36902, {{142357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36903, {{142358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36904, {{142359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36905, {{142361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36906, {{142362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36907, {{142363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36908, {{142364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{36909, {{142365, 1, 0, 0},{142365, 1, 1, 17}, {0, 0, 0, 0}}},{36909, {{142365, 1, 0, 0},{142365, 1, 1, 17}, {0, 0, 0, 0}}},{36910, {{142365, 1, 0, 0},{142365, 1, 1, 18}, {0, 0, 0, 0}}},{36910, {{142365, 1, 0, 0},{142365, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36911, {{142372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36912, {{142376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36913, {{142399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36914, {{142400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36915, {{142401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36916, {{142402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36917, {{142405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36918, {{142406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36919, {{142407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36920, {{142418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36921, {{142431, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36922, {{142432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36923, {{142433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36924, {{142434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36925, {{142435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36926, {{142436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36927, {{142437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36928, {{142438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36929, {{142439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36930, {{142440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36931, {{142441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36932, {{142443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36933, {{142444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36934, {{142446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36935, {{142447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36936, {{142448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36937, {{142449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{36938, {{142450, 1, 0, 0},{142450, 1, 1, 16}, {0, 0, 0, 0}}},{36938, {{142450, 1, 0, 0},{142450, 1, 1, 16}, {0, 0, 0, 0}}},{57566, {{142450, 1, 0, 0},{142450, 1, 1, 17}, {0, 0, 0, 0}}},{57566, {{142450, 1, 0, 0},{142450, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{36939, {{142451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36940, {{142452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38856, {{142453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36941, {{142454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36942, {{142455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36943, {{142457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36944, {{142459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36945, {{142460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36946, {{142463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36947, {{142464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36948, {{142485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36949, {{142486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36950, {{142488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36951, {{142512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36952, {{142515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36953, {{142516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36954, {{142517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36955, {{142518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36956, {{142519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36957, {{142520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36958, {{142521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36959, {{142522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36960, {{142523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36961, {{142524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36962, {{142525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36963, {{142526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36964, {{142528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36965, {{142529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36966, {{142530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36967, {{142531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36968, {{142532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36969, {{142533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36970, {{142534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36971, {{142535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36972, {{142536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36973, {{142537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36974, {{142538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36975, {{142539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36976, {{142541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36977, {{142542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36978, {{142543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36979, {{142544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36980, {{142545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36981, {{142546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36982, {{142547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36983, {{142548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36984, {{142549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36985, {{142550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36986, {{142552, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36987, {{142553, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36988, {{142554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36989, {{142559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36990, {{142589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36991, {{142590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36992, {{142592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36993, {{142624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36994, {{142625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36995, {{142626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36996, {{142627, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36998, {{142628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{36999, {{142629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37000, {{142630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37001, {{142631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37002, {{142632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37003, {{142633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37004, {{142634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37005, {{142635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37006, {{142636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37007, {{142637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37008, {{142638, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37009, {{142640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37010, {{142641, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37011, {{142642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37012, {{142643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37013, {{142644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37014, {{142645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37015, {{142646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37016, {{142647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37017, {{142648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37018, {{142649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37019, {{142650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37020, {{142651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{37021, {{142652, 1, 0, 0},{142652, 1, 1, 16}, {0, 0, 0, 0}}},{37021, {{142652, 1, 0, 0},{142652, 1, 1, 16}, {0, 0, 0, 0}}},{37022, {{142652, 1, 0, 0},{142652, 1, 1, 17}, {0, 0, 0, 0}}},{37022, {{142652, 1, 0, 0},{142652, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{37023, {{142653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37024, {{142654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37025, {{142655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37026, {{142656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37027, {{142657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37028, {{142658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37029, {{142659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37030, {{142660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37031, {{142664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37032, {{142669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37033, {{142670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37034, {{142671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37035, {{142679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37037, {{142695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37038, {{142719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37039, {{142720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37040, {{142721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37041, {{142722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37042, {{142723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37043, {{142725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37044, {{142726, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37045, {{142728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37046, {{142729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{37047, {{142730, 1, 0, 0},{142730, 1, 1, 17}, {0, 0, 0, 0}}},{37047, {{142730, 1, 0, 0},{142730, 1, 1, 17}, {0, 0, 0, 0}}},{57573, {{142730, 1, 0, 0},{142730, 1, 1, 16}, {0, 0, 0, 0}}},{57573, {{142730, 1, 0, 0},{142730, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{37048, {{142731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37049, {{142732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37050, {{142733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37051, {{142734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37052, {{142735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37053, {{142736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37054, {{142737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37055, {{142738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37056, {{142739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37057, {{142740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37058, {{142741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37059, {{142742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37060, {{142743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37061, {{142744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37062, {{142745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37063, {{142746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37064, {{142747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37065, {{142748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37066, {{142749, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37067, {{142753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37068, {{142758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37069, {{142772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37070, {{142773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37071, {{142791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37072, {{142792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37073, {{142795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37074, {{142796, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37075, {{142797, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37076, {{142798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37077, {{142799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{37078, {{142800, 1, 0, 0},{142800, 1, 1, 16}, {0, 0, 0, 0}}},{37078, {{142800, 1, 0, 0},{142800, 1, 1, 16}, {0, 0, 0, 0}}},{68076, {{142800, 1, 0, 0},{142800, 1, 1, 17}, {0, 0, 0, 0}}},{68076, {{142800, 1, 0, 0},{142800, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{37079, {{142801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37080, {{142802, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37081, {{142803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37082, {{142804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37083, {{142805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37084, {{142806, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37085, {{142807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37086, {{142808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37087, {{142809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37088, {{142810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37089, {{142811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37090, {{142812, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37091, {{142813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37092, {{142817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37093, {{142818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37094, {{142820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37095, {{142821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37096, {{142829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37097, {{142838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37098, {{142839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37099, {{142840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37100, {{142841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37101, {{142842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37102, {{142843, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37103, {{142844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37104, {{142845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37105, {{142846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37106, {{142847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37107, {{142848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37108, {{142849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37109, {{142850, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37110, {{142851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37111, {{142852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37112, {{142853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37113, {{142854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37114, {{142855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37115, {{142856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{37116, {{142857, 1, 0, 0},{142857, 1, 1, 16}, {0, 0, 0, 0}}},{37116, {{142857, 1, 0, 0},{142857, 1, 1, 16}, {0, 0, 0, 0}}},{57578, {{142857, 1, 0, 0},{142857, 1, 1, 17}, {0, 0, 0, 0}}},{57578, {{142857, 1, 0, 0},{142857, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37117, {{142883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37118, {{142884, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37119, {{142885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37120, {{142886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37121, {{142887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37122, {{142888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37123, {{142889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37124, {{142890, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37125, {{142891, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37126, {{142892, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{37127, {{142893, 1, 0, 0},{142893, 1, 1, 16}, {0, 0, 0, 0}}},{37127, {{142893, 1, 0, 0},{142893, 1, 1, 16}, {0, 0, 0, 0}}},{37128, {{142893, 1, 0, 0},{142893, 1, 1, 17}, {0, 0, 0, 0}}},{37128, {{142893, 1, 0, 0},{142893, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{37129, {{142894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37130, {{142895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37131, {{142896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37132, {{142897, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37133, {{142899, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37134, {{142900, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37135, {{142925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57579, {{142926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37136, {{142927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37137, {{142928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37138, {{142929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37175, {{142930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37139, {{142931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37140, {{142932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37141, {{142933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37142, {{142934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37143, {{142935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37144, {{142936, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37145, {{142937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37146, {{142938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37147, {{142956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37148, {{142957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37149, {{142958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37150, {{142959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37151, {{142960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37152, {{142961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37153, {{142962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37154, {{142963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37155, {{142964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37156, {{142965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37157, {{142978, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37158, {{142979, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37159, {{142980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37160, {{142981, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37161, {{142982, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37162, {{142984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37163, {{142985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37164, {{142986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37165, {{142987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37166, {{142988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37167, {{142993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37168, {{142997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37169, {{142998, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37170, {{143006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37171, {{143007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37172, {{143008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37173, {{143009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37174, {{143010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37176, {{143014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37177, {{143019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37178, {{143020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37179, {{143024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37180, {{143025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37181, {{143027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37182, {{143028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37183, {{143029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37184, {{143030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37185, {{143031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37186, {{143036, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37187, {{143039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37188, {{143040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37189, {{143042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37190, {{143043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37191, {{143044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37192, {{143045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37193, {{143047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37194, {{143048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37195, {{143049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37196, {{143051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37197, {{143052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37198, {{143053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37199, {{143055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37200, {{143059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37201, {{143060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37202, {{143063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37203, {{143064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37204, {{143065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37205, {{143066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37206, {{143067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37207, {{143068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37208, {{143069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37209, {{143070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37210, {{143071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37211, {{143072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37213, {{143075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37214, {{143076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37215, {{143078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37216, {{143079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37217, {{143081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37218, {{143082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37224, {{143083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{37219, {{143084, 1, 0, 0},{143084, 1, 1, 16}, {0, 0, 0, 0}}},{37219, {{143084, 1, 0, 0},{143084, 1, 1, 16}, {0, 0, 0, 0}}},{37220, {{143084, 1, 0, 0},{143084, 1, 1, 17}, {0, 0, 0, 0}}},{37220, {{143084, 1, 0, 0},{143084, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{37221, {{143085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37222, {{143086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37223, {{143087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37225, {{143090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37226, {{143091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37227, {{143092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37240, {{143093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37228, {{143094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{68077, {{143095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37241, {{143096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37230, {{143097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37231, {{143099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37242, {{143100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37232, {{143101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37233, {{143102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37234, {{143103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37235, {{143104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37236, {{143105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37237, {{143107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37238, {{143108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37239, {{143109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37243, {{143113, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37244, {{143114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37245, {{143115, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37246, {{143116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37247, {{143117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37248, {{143118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37249, {{143119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37250, {{143120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37251, {{143121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37252, {{143126, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37253, {{143128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37254, {{143129, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37255, {{143131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37256, {{143132, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37257, {{143133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37258, {{143134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37259, {{143135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37260, {{143136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37261, {{143137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{37262, {{143138, 1, 0, 0},{143138, 1, 1, 16}, {0, 0, 0, 0}}},{37262, {{143138, 1, 0, 0},{143138, 1, 1, 16}, {0, 0, 0, 0}}},{57595, {{143138, 1, 0, 0},{143138, 1, 1, 17}, {0, 0, 0, 0}}},{57595, {{143138, 1, 0, 0},{143138, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{37263, {{143139, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37264, {{143140, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37265, {{143141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37270, {{143142, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37266, {{143143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37267, {{143144, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37275, {{143145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37268, {{143146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37269, {{143148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37271, {{143151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37272, {{143152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37273, {{143153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37274, {{143157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37445, {{143158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37276, {{143159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37291, {{143161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37277, {{143162, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37278, {{143163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37293, {{143164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37280, {{143165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37292, {{143166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37281, {{143167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37282, {{143168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37283, {{143169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37284, {{143170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37285, {{143171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37286, {{143172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37287, {{143173, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37288, {{143174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37289, {{143175, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37290, {{143177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37297, {{143189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37295, {{143191, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37296, {{143192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37317, {{143194, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37298, {{143195, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37299, {{143196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37300, {{143197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37301, {{143198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37302, {{143199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37303, {{143200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37304, {{143201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37318, {{143203, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37305, {{143204, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37306, {{143205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37307, {{143206, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37308, {{143207, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37309, {{143208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37310, {{143209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37311, {{143210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37343, {{143211, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37312, {{143212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37313, {{143213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37314, {{143214, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37315, {{143215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37316, {{143216, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37319, {{143224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37727, {{143225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37320, {{143226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37321, {{143230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37322, {{143231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37323, {{143232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37341, {{143233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37324, {{143234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37325, {{143235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37326, {{143236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37327, {{143237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37328, {{143238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37329, {{143239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37330, {{143240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37339, {{143242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37331, {{143244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37332, {{143245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37340, {{143246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37333, {{143247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37334, {{143248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37338, {{143249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37335, {{143250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37336, {{143251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37337, {{143252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37342, {{143264, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37344, {{143269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37364, {{143270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37345, {{143271, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37346, {{143272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37347, {{143273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37348, {{143275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37349, {{143276, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37350, {{143277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37351, {{143278, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37352, {{143279, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37353, {{143280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37354, {{143281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37363, {{143282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37355, {{143283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37356, {{143284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37357, {{143285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37358, {{143286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37359, {{143287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37360, {{143288, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37361, {{143289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37362, {{143290, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57611, {{143294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37365, {{143302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37366, {{143304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37367, {{143305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37368, {{143306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{37369, {{143308, 1, 0, 0},{143308, 1, 1, 16}, {0, 0, 0, 0}}},{37369, {{143308, 1, 0, 0},{143308, 1, 1, 16}, {0, 0, 0, 0}}},{57615, {{143308, 1, 0, 0},{143308, 1, 1, 17}, {0, 0, 0, 0}}},{57615, {{143308, 1, 0, 0},{143308, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37370, {{143310, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37376, {{143311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37371, {{143312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37372, {{143313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37373, {{143314, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37374, {{143315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37375, {{143316, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{37377, {{143320, 1, 0, 0},{143320, 1, 1, 17}, {0, 0, 0, 0}}},{37377, {{143320, 1, 0, 0},{143320, 1, 1, 17}, {0, 0, 0, 0}}},{57612, {{143320, 1, 0, 0},{143320, 1, 1, 16}, {0, 0, 0, 0}}},{57612, {{143320, 1, 0, 0},{143320, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {4, {{37378, {{143321, 1, 0, 0},{143321, 1, 1, 16}, {0, 0, 0, 0}}},{37378, {{143321, 1, 0, 0},{143321, 1, 1, 16}, {0, 0, 0, 0}}},{37379, {{143321, 1, 0, 0},{143321, 1, 1, 17}, {0, 0, 0, 0}}},{37379, {{143321, 1, 0, 0},{143321, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{37380, {{143322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37381, {{143323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37382, {{143326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37383, {{143327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{37384, {{143328, 1, 0, 0},{143328, 1, 1, 16}, {0, 0, 0, 0}}},{37384, {{143328, 1, 0, 0},{143328, 1, 1, 16}, {0, 0, 0, 0}}},{57618, {{143328, 1, 0, 0},{143328, 1, 1, 17}, {0, 0, 0, 0}}},{57618, {{143328, 1, 0, 0},{143328, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{37385, {{143329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37397, {{143330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37395, {{143331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37386, {{143332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37387, {{143333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37403, {{143334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37398, {{143335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37388, {{143336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37389, {{143337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37390, {{143338, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37391, {{143339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37392, {{143340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37393, {{143341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37394, {{143342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37396, {{143347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37399, {{143351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37400, {{143352, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37401, {{143353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37402, {{143355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37404, {{143358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37405, {{143360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37406, {{143361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37407, {{143362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37410, {{143363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37408, {{143364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57619, {{143367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37409, {{143369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37411, {{143373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37412, {{143374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37413, {{143375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37414, {{143376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37415, {{143378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37419, {{143379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37416, {{143380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37417, {{143381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37418, {{143382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37421, {{143386, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37424, {{143387, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37425, {{143388, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37422, {{143389, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37423, {{143390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37426, {{143395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37427, {{143396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37428, {{143397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37429, {{143398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37430, {{143399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37431, {{143400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37432, {{143402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37433, {{143404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37434, {{143405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37435, {{143406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37436, {{143407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37437, {{143408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37438, {{143411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37439, {{143413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37440, {{143414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37441, {{143415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37442, {{143419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37443, {{143427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37444, {{143433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57623, {{143434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37446, {{143435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37447, {{143438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37449, {{143447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37450, {{143449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37451, {{143452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{37452, {{143453, 1, 0, 0},{143453, 1, 1, 16}, {0, 0, 0, 0}}},{37452, {{143453, 1, 0, 0},{143453, 1, 1, 16}, {0, 0, 0, 0}}},{37453, {{143453, 1, 0, 0},{143453, 1, 1, 17}, {0, 0, 0, 0}}},{37453, {{143453, 1, 0, 0},{143453, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37455, {{143455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37456, {{143458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37457, {{143459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37458, {{143462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37459, {{143463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37460, {{143464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37461, {{143467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37462, {{143468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37463, {{143469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37464, {{143471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37465, {{143472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37466, {{143473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37467, {{143475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37468, {{143476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37469, {{143477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37470, {{143478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37471, {{143479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37472, {{143481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37473, {{143482, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37474, {{143483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37475, {{143484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59669, {{143485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37476, {{143487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37477, {{143489, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37478, {{143490, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37479, {{143492, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37480, {{143494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37481, {{143495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37482, {{143497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37483, {{143498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37484, {{143500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37485, {{143505, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37486, {{143506, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37487, {{143508, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37488, {{143509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37489, {{143510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37490, {{143511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37492, {{143516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37493, {{143517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37494, {{143518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37495, {{143519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37496, {{143520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37497, {{143521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37498, {{143522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37499, {{143524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37500, {{143525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37501, {{143526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37502, {{143527, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37503, {{143528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37504, {{143530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37505, {{143531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37506, {{143532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37507, {{143533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37508, {{143534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37509, {{143535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{37510, {{143536, 1, 0, 0},{143536, 1, 1, 16}, {0, 0, 0, 0}}},{37510, {{143536, 1, 0, 0},{143536, 1, 1, 16}, {0, 0, 0, 0}}},{57626, {{143536, 1, 0, 0},{143536, 1, 1, 17}, {0, 0, 0, 0}}},{57626, {{143536, 1, 0, 0},{143536, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{37511, {{143537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37512, {{143538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37513, {{143539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37514, {{143540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37515, {{143541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37516, {{143542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37517, {{143545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37518, {{143546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37519, {{143547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37520, {{143549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37521, {{143551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37522, {{143552, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37523, {{143553, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37524, {{143554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37525, {{143555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37526, {{143556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37527, {{143557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37528, {{143558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37529, {{143560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37530, {{143561, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37531, {{143562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37532, {{143563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37533, {{143564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37534, {{143565, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37535, {{143566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37536, {{143567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37537, {{143568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37538, {{143569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37539, {{143570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{37540, {{143572, 1, 0, 0},{143572, 1, 1, 16}, {0, 0, 0, 0}}},{37540, {{143572, 1, 0, 0},{143572, 1, 1, 16}, {0, 0, 0, 0}}},{37541, {{143572, 1, 0, 0},{143572, 1, 1, 17}, {0, 0, 0, 0}}},{37541, {{143572, 1, 0, 0},{143572, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37542, {{143574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37543, {{143575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37544, {{143576, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37545, {{143577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13101, {{143578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37546, {{143581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37547, {{143582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37548, {{143583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37549, {{143584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37551, {{143586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37553, {{143590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37554, {{143591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37555, {{143592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37556, {{143594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37557, {{143599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37558, {{143603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37560, {{143604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37559, {{143605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37561, {{143611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37562, {{143612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37563, {{143613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37564, {{143614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37565, {{143615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37566, {{143616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37567, {{143617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37569, {{143621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{37570, {{143622, 1, 0, 0},{143622, 1, 1, 16}, {0, 0, 0, 0}}},{37570, {{143622, 1, 0, 0},{143622, 1, 1, 16}, {0, 0, 0, 0}}},{57642, {{143622, 1, 0, 0},{143622, 1, 1, 17}, {0, 0, 0, 0}}},{57642, {{143622, 1, 0, 0},{143622, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37571, {{143624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37572, {{143625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37573, {{143626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37574, {{143629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37575, {{143630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37576, {{143631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37577, {{143632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37578, {{143633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37579, {{143635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37580, {{143636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37581, {{143640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37582, {{143641, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37583, {{143642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37584, {{143644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37585, {{143645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{37586, {{143646, 1, 0, 0},{143646, 1, 1, 16}, {0, 0, 0, 0}}},{37586, {{143646, 1, 0, 0},{143646, 1, 1, 16}, {0, 0, 0, 0}}},{37587, {{143646, 1, 0, 0},{143646, 1, 1, 17}, {0, 0, 0, 0}}},{37587, {{143646, 1, 0, 0},{143646, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{37588, {{143647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37589, {{143648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37590, {{143649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{37591, {{143650, 1, 0, 0},{143650, 1, 1, 16}, {0, 0, 0, 0}}},{37591, {{143650, 1, 0, 0},{143650, 1, 1, 16}, {0, 0, 0, 0}}},{37592, {{143650, 1, 0, 0},{143650, 1, 1, 17}, {0, 0, 0, 0}}},{37592, {{143650, 1, 0, 0},{143650, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37593, {{143652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37594, {{143654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37595, {{143655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37596, {{143656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37597, {{143658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37598, {{143659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37599, {{143662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37600, {{143663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37601, {{143664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37602, {{143665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37603, {{143666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37604, {{143668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37605, {{143669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37606, {{143670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37607, {{143671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37608, {{143672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37609, {{143674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37610, {{143675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37611, {{143676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37612, {{143678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37613, {{143679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37614, {{143680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37615, {{143681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37616, {{143684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37617, {{143686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37618, {{143687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37619, {{143688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37620, {{143689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37621, {{143690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37622, {{143691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37623, {{143692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{37624, {{143693, 1, 0, 0},{143693, 1, 1, 16}, {0, 0, 0, 0}}},{37624, {{143693, 1, 0, 0},{143693, 1, 1, 16}, {0, 0, 0, 0}}},{37625, {{143693, 1, 0, 0},{143693, 1, 1, 17}, {0, 0, 0, 0}}},{37625, {{143693, 1, 0, 0},{143693, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{37626, {{143694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37627, {{143696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37628, {{143702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37629, {{143706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37630, {{143707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37631, {{143708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37632, {{143709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37633, {{143710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37634, {{143711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37635, {{143712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37636, {{143713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37637, {{143714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37638, {{143715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37639, {{143716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37640, {{143717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37641, {{143718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37642, {{143719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37643, {{143727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37644, {{143729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37645, {{143730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37646, {{143731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37647, {{143732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37648, {{143735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37649, {{143736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37650, {{143737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37651, {{143738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37652, {{143739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37653, {{143740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37654, {{143741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57662, {{143745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37656, {{143748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37657, {{143751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37658, {{143752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37659, {{143753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37660, {{143754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37661, {{143755, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37662, {{143756, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37663, {{143757, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37664, {{143761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37665, {{143762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37666, {{143763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37667, {{143764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37668, {{143765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37669, {{143766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37671, {{143767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37672, {{143768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37673, {{143770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37674, {{143771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37675, {{143772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37676, {{143773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37677, {{143774, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37678, {{143775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37679, {{143776, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37680, {{143781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37681, {{143783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57663, {{143784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37682, {{143785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37683, {{143787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37684, {{143794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37685, {{143795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37687, {{143796, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37688, {{143798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37690, {{143799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37691, {{143800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37692, {{143801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37693, {{143802, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37694, {{143803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37695, {{143804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37696, {{143805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37697, {{143806, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37698, {{143807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37699, {{143808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37700, {{143811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37702, {{143812, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37704, {{143830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37705, {{143833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37706, {{143834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37707, {{143842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37708, {{143844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37709, {{143845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37710, {{143846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37711, {{143847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37712, {{143848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37713, {{143849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37714, {{143850, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37715, {{143851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37716, {{143852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37717, {{143854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37718, {{143855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37719, {{143856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37720, {{143857, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37721, {{143858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37722, {{143859, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37723, {{143860, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37724, {{143861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37725, {{143862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37726, {{143866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37728, {{143877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37729, {{143878, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37730, {{143881, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37731, {{143882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37732, {{143885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37733, {{143886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37734, {{143887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37735, {{143889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37738, {{143903, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37739, {{143904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37740, {{143905, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37741, {{143908, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37742, {{143909, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37743, {{143910, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37744, {{143911, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37745, {{143912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37746, {{143913, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37747, {{143914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37748, {{143921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37749, {{143926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37750, {{143942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37751, {{143943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37752, {{143944, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37753, {{143946, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37754, {{143953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37755, {{143954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37756, {{143955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37757, {{143956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37758, {{143957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37759, {{143958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37760, {{143959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37761, {{143960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37762, {{143961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37763, {{143962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37764, {{143963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37765, {{143964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37767, {{143969, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37768, {{143977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37770, {{143979, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37771, {{143980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37772, {{143981, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37773, {{143984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37774, {{143994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37775, {{143995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37776, {{143996, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37777, {{143998, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37778, {{143999, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37769, {{144000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37779, {{144001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37780, {{144002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37781, {{144003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37782, {{144004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37783, {{144005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37784, {{144018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37785, {{144021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37786, {{144022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37787, {{144025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37788, {{144026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37789, {{144027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37790, {{144029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37791, {{144030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37792, {{144031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37793, {{144032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37794, {{144033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37795, {{144035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37796, {{144036, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37797, {{144037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37798, {{144038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37799, {{144039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37800, {{144040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37802, {{144052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37803, {{144053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37804, {{144055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{37805, {{144056, 1, 0, 0},{144056, 1, 1, 16}, {0, 0, 0, 0}}},{37805, {{144056, 1, 0, 0},{144056, 1, 1, 16}, {0, 0, 0, 0}}},{37806, {{144056, 1, 0, 0},{144056, 1, 1, 17}, {0, 0, 0, 0}}},{37806, {{144056, 1, 0, 0},{144056, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37807, {{144061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37808, {{144068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37809, {{144069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37810, {{144073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37811, {{144074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37812, {{144075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37813, {{144076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37814, {{144077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37815, {{144078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37816, {{144079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37817, {{144080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37818, {{144083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37819, {{144087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37820, {{144088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37821, {{144094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37823, {{144095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57681, {{144096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37824, {{144098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37825, {{144099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37826, {{144101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37827, {{144108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37828, {{144111, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37829, {{144115, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37830, {{144116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37831, {{144117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37832, {{144118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37833, {{144119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37834, {{144120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37835, {{144121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37836, {{144127, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37837, {{144135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37838, {{144136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37839, {{144141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37840, {{144147, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37841, {{144150, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37842, {{144151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37843, {{144152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37844, {{144154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37845, {{144156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37846, {{144158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37847, {{144159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37848, {{144160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37849, {{144163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37850, {{144166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37851, {{144167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37852, {{144168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37853, {{144169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37854, {{144172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37855, {{144173, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37856, {{144188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37857, {{144189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37858, {{144190, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37859, {{144195, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37860, {{144196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37861, {{144198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37862, {{144199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37863, {{144200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37864, {{144207, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37865, {{144209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37866, {{144211, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37867, {{144212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37868, {{144214, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37869, {{144216, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37870, {{144217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37871, {{144218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37872, {{144219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37873, {{144220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37874, {{144221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37875, {{144222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{37876, {{144223, 1, 0, 0},{144223, 1, 1, 16}, {0, 0, 0, 0}}},{37876, {{144223, 1, 0, 0},{144223, 1, 1, 16}, {0, 0, 0, 0}}},{37877, {{144223, 1, 0, 0},{144223, 1, 1, 17}, {0, 0, 0, 0}}},{37877, {{144223, 1, 0, 0},{144223, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{37878, {{144224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37879, {{144225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37880, {{144226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37881, {{144227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37882, {{144228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37883, {{144229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37884, {{144230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37885, {{144231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37887, {{144236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{37888, {{144238, 1, 0, 0},{144238, 1, 1, 16}, {0, 0, 0, 0}}},{37888, {{144238, 1, 0, 0},{144238, 1, 1, 16}, {0, 0, 0, 0}}},{37889, {{144238, 1, 0, 0},{144238, 1, 1, 17}, {0, 0, 0, 0}}},{37889, {{144238, 1, 0, 0},{144238, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37890, {{144242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37891, {{144244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37892, {{144245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37893, {{144247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37894, {{144248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37895, {{144249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37896, {{144250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37897, {{144260, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37898, {{144261, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37899, {{144268, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37900, {{144269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37901, {{144270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37902, {{144274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {2, {{37903, {{144275, 1, 0, 0}, {0, 0, 0, 0}}},{37904, {{144275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37905, {{144276, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37906, {{144277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37907, {{144282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37908, {{144283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {2, {{37909, {{144284, 1, 0, 0}, {0, 0, 0, 0}}},{37910, {{144284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37911, {{144287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37912, {{144291, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37913, {{144292, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37914, {{144297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37915, {{144301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37916, {{144305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37917, {{144306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37919, {{144307, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37920, {{144308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{37921, {{144309, 1, 0, 0},{144309, 1, 1, 16}, {0, 0, 0, 0}}},{37921, {{144309, 1, 0, 0},{144309, 1, 1, 16}, {0, 0, 0, 0}}},{68078, {{144309, 1, 0, 0},{144309, 1, 1, 17}, {0, 0, 0, 0}}},{68078, {{144309, 1, 0, 0},{144309, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37923, {{144312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37924, {{144313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{30332, {{144314, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37927, {{144316, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37925, {{144318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37928, {{144319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37929, {{144320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37930, {{144321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37931, {{144322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37932, {{144323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37934, {{144324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37935, {{144331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{13633, {{144332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37936, {{144336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37937, {{144337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{37938, {{144338, 1, 0, 0},{144338, 1, 1, 18}, {0, 0, 0, 0}}},{37938, {{144338, 1, 0, 0},{144338, 1, 1, 18}, {0, 0, 0, 0}}},{37939, {{144338, 1, 0, 0},{144338, 1, 1, 17}, {0, 0, 0, 0}}},{37939, {{144338, 1, 0, 0},{144338, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{37940, {{144339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37941, {{144340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37942, {{144341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37943, {{144342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37944, {{144343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37945, {{144344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37946, {{144345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37947, {{144346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37948, {{144347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37949, {{144348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37950, {{144349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37951, {{144350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37952, {{144351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{37953, {{144352, 1, 0, 0},{144352, 1, 1, 16}, {0, 0, 0, 0}}},{37953, {{144352, 1, 0, 0},{144352, 1, 1, 16}, {0, 0, 0, 0}}},{37954, {{144352, 1, 0, 0},{144352, 1, 1, 17}, {0, 0, 0, 0}}},{37954, {{144352, 1, 0, 0},{144352, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{37955, {{144353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37956, {{144356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57707, {{144359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37957, {{144370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37958, {{144371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37959, {{144374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37960, {{144377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37961, {{144378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37962, {{144379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37963, {{144381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37964, {{144382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37965, {{144385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37966, {{144387, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37967, {{144391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37968, {{144392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37970, {{144393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37971, {{144394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37972, {{144395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37973, {{144396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37974, {{144397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37975, {{144398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37976, {{144399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37977, {{144400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37978, {{144401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37979, {{144402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37980, {{144403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57713, {{144425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37981, {{144427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37982, {{144428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37983, {{144429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37984, {{144430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37985, {{144431, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37986, {{144432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37987, {{144437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37988, {{144438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37989, {{144439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37990, {{144440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37991, {{144441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37992, {{144442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37993, {{144443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37994, {{144445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37995, {{144446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37996, {{144447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37997, {{144448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{37999, {{144449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38000, {{144450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38001, {{144451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38002, {{144452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38003, {{144458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38004, {{144459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38005, {{144465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38006, {{144485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38007, {{144487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38008, {{144488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38009, {{144490, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{38010, {{144493, 1, 0, 0},{144493, 1, 1, 16}, {0, 0, 0, 0}}},{38010, {{144493, 1, 0, 0},{144493, 1, 1, 16}, {0, 0, 0, 0}}},{38011, {{144493, 1, 0, 0},{144493, 1, 1, 17}, {0, 0, 0, 0}}},{38011, {{144493, 1, 0, 0},{144493, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57727, {{144495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57729, {{144496, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38012, {{144497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57721, {{144501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38013, {{144504, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38014, {{144505, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57720, {{144506, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38015, {{144510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38016, {{144511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38017, {{144512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38018, {{144513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38019, {{144514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38020, {{144515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38021, {{144516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38022, {{144517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38023, {{144518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38024, {{144519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38025, {{144520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38026, {{144521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38027, {{144552, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38028, {{144553, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38029, {{144554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38030, {{144555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38032, {{144556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38033, {{144557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38034, {{144559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38035, {{144560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38037, {{144566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38038, {{144567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38039, {{144568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38040, {{144569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38041, {{144570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38042, {{144571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38043, {{144572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38044, {{144576, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38045, {{144577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38046, {{144578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38047, {{144579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38048, {{144581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38049, {{144582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38051, {{144584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{38052, {{144585, 1, 0, 0},{144585, 1, 1, 17}, {0, 0, 0, 0}}},{38052, {{144585, 1, 0, 0},{144585, 1, 1, 17}, {0, 0, 0, 0}}},{38053, {{144585, 1, 0, 0},{144585, 1, 1, 16}, {0, 0, 0, 0}}},{38053, {{144585, 1, 0, 0},{144585, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{38054, {{144586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38055, {{144587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38056, {{144588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38057, {{144589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38058, {{144590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38061, {{144591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38062, {{144592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38063, {{144593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38064, {{144594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38065, {{144595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38066, {{144596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38067, {{144597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38068, {{144598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38071, {{144599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38072, {{144600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38073, {{144601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38074, {{144603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38075, {{144604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38076, {{144605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38077, {{144606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38078, {{144607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38079, {{144608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38080, {{144609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38081, {{144612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38082, {{144615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38083, {{144617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57752, {{144618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57746, {{144651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38086, {{144654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38087, {{144655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38088, {{144657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38089, {{144658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38090, {{144659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38091, {{144660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38092, {{144661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38093, {{144662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38084, {{144682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38094, {{144683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38095, {{144684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38096, {{144685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38097, {{144686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38098, {{144687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{38099, {{144689, 1, 0, 0},{144689, 1, 1, 16}, {0, 0, 0, 0}}},{38099, {{144689, 1, 0, 0},{144689, 1, 1, 16}, {0, 0, 0, 0}}},{57773, {{144689, 1, 0, 0},{144689, 1, 1, 17}, {0, 0, 0, 0}}},{57773, {{144689, 1, 0, 0},{144689, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{38100, {{144690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38101, {{144691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38102, {{144693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38103, {{144694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38104, {{144695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38105, {{144696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38106, {{144697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38107, {{144698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38108, {{144699, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38109, {{144700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38110, {{144701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38111, {{144702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38112, {{144703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38113, {{144704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38114, {{144705, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38115, {{144706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38116, {{144707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38117, {{144708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38118, {{144709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38119, {{144710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38120, {{144711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38121, {{144712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38122, {{144713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38123, {{144714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38124, {{144715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38125, {{144716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38126, {{144717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38127, {{144718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38128, {{144720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38129, {{144721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38130, {{144722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38131, {{144723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{38132, {{144730, 1, 0, 0},{144730, 1, 1, 17}, {0, 0, 0, 0}}},{38132, {{144730, 1, 0, 0},{144730, 1, 1, 17}, {0, 0, 0, 0}}},{38133, {{144730, 1, 0, 0},{144730, 1, 1, 18}, {0, 0, 0, 0}}},{38133, {{144730, 1, 0, 0},{144730, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57768, {{144754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38134, {{144765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38135, {{144766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38136, {{144767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38137, {{144768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38138, {{144769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38139, {{144770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38140, {{144771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38141, {{144775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38142, {{144777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38143, {{144778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38144, {{144788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38145, {{144803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38146, {{144804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38147, {{144805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38148, {{144808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38149, {{144810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38150, {{144811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38151, {{144812, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38152, {{144813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38154, {{144814, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38155, {{144816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38156, {{144817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38157, {{144818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38158, {{144820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38159, {{144821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38160, {{144822, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38161, {{144823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38162, {{144824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38163, {{144825, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38164, {{144826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38165, {{144827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38166, {{144828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38167, {{144829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38168, {{144830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38169, {{144831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38170, {{144833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38171, {{144834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38172, {{144835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{38173, {{144836, 1, 0, 0},{144836, 1, 1, 18}, {0, 0, 0, 0}}},{38173, {{144836, 1, 0, 0},{144836, 1, 1, 18}, {0, 0, 0, 0}}},{38174, {{144836, 1, 0, 0},{144836, 1, 1, 19}, {0, 0, 0, 0}}},{38174, {{144836, 1, 0, 0},{144836, 1, 1, 19}, {0, 0, 0, 0}}},{38175, {{144836, 1, 0, 0},{144836, 1, 1, 20}, {0, 0, 0, 0}}},{38175, {{144836, 1, 0, 0},{144836, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38176, {{144848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38177, {{144854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57790, {{144869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38178, {{144883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38179, {{144884, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38180, {{144885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38181, {{144887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38182, {{144888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38183, {{144889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38184, {{144901, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38185, {{144903, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38186, {{144912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38187, {{144913, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38188, {{144914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38189, {{144917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38190, {{144918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38191, {{144919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38193, {{144920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38194, {{144921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38195, {{144922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38196, {{144923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38197, {{144924, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38198, {{144925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38199, {{144926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38200, {{144927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38201, {{144928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38202, {{144929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38203, {{144930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38204, {{144931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38205, {{144932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38206, {{144933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{38207, {{144934, 1, 0, 0},{144934, 1, 1, 16}, {0, 0, 0, 0}}},{38207, {{144934, 1, 0, 0},{144934, 1, 1, 16}, {0, 0, 0, 0}}},{38208, {{144934, 1, 0, 0},{144934, 1, 1, 17}, {0, 0, 0, 0}}},{38208, {{144934, 1, 0, 0},{144934, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{38209, {{144935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38210, {{144936, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38211, {{144937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38212, {{144938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38213, {{144939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38214, {{144940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38215, {{144941, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38216, {{144942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38217, {{144943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38218, {{144944, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38219, {{144945, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38220, {{144946, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38221, {{144947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{38222, {{144952, 1, 0, 0},{144952, 1, 1, 18}, {0, 0, 0, 0}}},{38222, {{144952, 1, 0, 0},{144952, 1, 1, 18}, {0, 0, 0, 0}}},{38223, {{144952, 1, 0, 0},{144952, 1, 1, 17}, {0, 0, 0, 0}}},{38223, {{144952, 1, 0, 0},{144952, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{38224, {{144953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38225, {{144954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38227, {{144955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38228, {{144962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38229, {{144967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57801, {{144980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38231, {{144986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38232, {{144987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38233, {{144988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57799, {{144991, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38367, {{144993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38234, {{144996, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57807, {{145002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38235, {{145008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38236, {{145009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38237, {{145011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38238, {{145012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38239, {{145013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38240, {{145014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38241, {{145019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38242, {{145020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38243, {{145021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38244, {{145022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38245, {{145023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38246, {{145024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38247, {{145025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38248, {{145026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38249, {{145027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38250, {{145028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38251, {{145029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38252, {{145030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38253, {{145031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38254, {{145032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38255, {{145033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38256, {{145034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38257, {{145035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38258, {{145036, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38259, {{145037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38260, {{145038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38261, {{145039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38262, {{145040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38263, {{145041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38264, {{145042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{1829, {{145043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38265, {{145044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38266, {{145045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38267, {{145046, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38268, {{145047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38269, {{145048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38270, {{145049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38271, {{145050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38272, {{145051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38273, {{145052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38274, {{145053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38275, {{145055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38276, {{145056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38277, {{145057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38278, {{145058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{38279, {{145059, 1, 0, 0},{145059, 1, 1, 17}, {0, 0, 0, 0}}},{38279, {{145059, 1, 0, 0},{145059, 1, 1, 17}, {0, 0, 0, 0}}},{57817, {{145059, 1, 0, 0},{145059, 1, 1, 16}, {0, 0, 0, 0}}},{57817, {{145059, 1, 0, 0},{145059, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{38280, {{145060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38281, {{145061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38282, {{145062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38283, {{145063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38284, {{145064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38285, {{145065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38286, {{145066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38287, {{145069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57829, {{145088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38289, {{145110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38290, {{145112, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{22974, {{145113, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38291, {{145114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57836, {{145115, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57819, {{145116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57824, {{145123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38292, {{145131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38294, {{145137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38295, {{145139, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38296, {{145140, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38297, {{145141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38298, {{145142, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38299, {{145143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38300, {{145144, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38301, {{145145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38302, {{145146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38303, {{145147, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38304, {{145148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38305, {{145149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38306, {{145151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38307, {{145152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38308, {{145153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38309, {{145154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38310, {{145155, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38311, {{145156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38312, {{145157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38313, {{145158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38314, {{145159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38315, {{145160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38316, {{145164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38317, {{145166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{38318, {{145180, 1, 0, 0},{145180, 1, 1, 17}, {0, 0, 0, 0}}},{38318, {{145180, 1, 0, 0},{145180, 1, 1, 17}, {0, 0, 0, 0}}},{38319, {{145180, 1, 0, 0},{145180, 1, 1, 18}, {0, 0, 0, 0}}},{38319, {{145180, 1, 0, 0},{145180, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38321, {{145186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38322, {{145187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38323, {{145188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38324, {{145189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38325, {{145190, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38327, {{145191, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38328, {{145192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38329, {{145209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38330, {{145210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38331, {{145213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{38332, {{145215, 1, 0, 0},{145215, 1, 1, 17}, {0, 0, 0, 0}}},{38332, {{145215, 1, 0, 0},{145215, 1, 1, 17}, {0, 0, 0, 0}}},{38333, {{145215, 1, 0, 0},{145215, 1, 1, 18}, {0, 0, 0, 0}}},{38333, {{145215, 1, 0, 0},{145215, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38334, {{145217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38335, {{145218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38336, {{145219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38337, {{145220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38338, {{145221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38339, {{145222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38340, {{145223, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38341, {{145224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38342, {{145225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38343, {{145226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{38344, {{145227, 1, 0, 0},{145227, 1, 1, 17}, {0, 0, 0, 0}}},{38344, {{145227, 1, 0, 0},{145227, 1, 1, 17}, {0, 0, 0, 0}}},{38345, {{145227, 1, 0, 0},{145227, 1, 1, 16}, {0, 0, 0, 0}}},{38345, {{145227, 1, 0, 0},{145227, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{38346, {{145228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38347, {{145229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38348, {{145230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38349, {{145231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38350, {{145233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38351, {{145234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38352, {{145235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38353, {{145236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38354, {{145237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38355, {{145238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38356, {{145239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38357, {{145240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38358, {{145241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38359, {{145242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38360, {{145243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38361, {{145244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38362, {{145245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38363, {{145246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38364, {{145251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{38365, {{145252, 1, 0, 0},{145252, 1, 1, 17}, {0, 0, 0, 0}}},{38365, {{145252, 1, 0, 0},{145252, 1, 1, 17}, {0, 0, 0, 0}}},{57856, {{145252, 1, 0, 0},{145252, 1, 1, 18}, {0, 0, 0, 0}}},{57856, {{145252, 1, 0, 0},{145252, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{38366, {{145253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38368, {{145263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38369, {{145268, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38370, {{145271, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38371, {{145272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38372, {{145273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38373, {{145274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38374, {{145275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38375, {{145276, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38376, {{145278, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{38377, {{145280, 1, 0, 0},{145280, 1, 1, 17}, {0, 0, 0, 0}}},{38377, {{145280, 1, 0, 0},{145280, 1, 1, 17}, {0, 0, 0, 0}}},{57852, {{145280, 1, 0, 0},{145280, 1, 1, 16}, {0, 0, 0, 0}}},{57852, {{145280, 1, 0, 0},{145280, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57845, {{145282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38378, {{145286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38379, {{145287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38380, {{145288, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38381, {{145291, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38382, {{145295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38383, {{145296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38384, {{145297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38385, {{145298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38386, {{145299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38387, {{145300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38388, {{145301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38389, {{145302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38390, {{145303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38391, {{145304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38392, {{145305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38393, {{145306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38394, {{145307, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38395, {{145308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38396, {{145309, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38397, {{145310, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38398, {{145311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38399, {{145315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38400, {{145316, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38401, {{145330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38402, {{145332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38403, {{145333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38404, {{145335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38405, {{145348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38406, {{145349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38407, {{145350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38408, {{145351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38409, {{145353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38410, {{145354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38411, {{145355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38412, {{145356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38413, {{145357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38414, {{145358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38415, {{145359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38416, {{145360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38417, {{145361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38418, {{145362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57864, {{145381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{38419, {{145383, 1, 0, 0},{145383, 1, 1, 18}, {0, 0, 0, 0}}},{38419, {{145383, 1, 0, 0},{145383, 1, 1, 18}, {0, 0, 0, 0}}},{38420, {{145383, 1, 0, 0},{145383, 1, 1, 17}, {0, 0, 0, 0}}},{38420, {{145383, 1, 0, 0},{145383, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{38421, {{145384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38422, {{145391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38423, {{145392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38424, {{145394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{38425, {{145395, 1, 0, 0},{145395, 1, 1, 17}, {0, 0, 0, 0}}},{38425, {{145395, 1, 0, 0},{145395, 1, 1, 17}, {0, 0, 0, 0}}},{38426, {{145395, 1, 0, 0},{145395, 1, 1, 16}, {0, 0, 0, 0}}},{38426, {{145395, 1, 0, 0},{145395, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{38427, {{145396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38428, {{145397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38429, {{145398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38430, {{145400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38431, {{145401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38432, {{145402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38433, {{145403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38434, {{145404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38435, {{145405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38436, {{145406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38437, {{145407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38438, {{145412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38439, {{145424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38440, {{145428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38441, {{145433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38442, {{145434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38443, {{145435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38444, {{145436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38445, {{145437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38446, {{145438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38447, {{145439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38448, {{145440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38449, {{145441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38450, {{145444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38451, {{145453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38452, {{145454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38453, {{145460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38454, {{145463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38455, {{145464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38456, {{145465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38457, {{145466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38458, {{145467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38459, {{145468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38460, {{145469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38461, {{145470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38462, {{145472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38463, {{145475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38464, {{145478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38465, {{145480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38466, {{145481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38467, {{145482, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38468, {{145484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38469, {{145485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38470, {{145486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38472, {{145493, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38473, {{145495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38474, {{145496, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38475, {{145499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38476, {{145502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57873, {{145503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38477, {{145504, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38478, {{145507, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38479, {{145509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38471, {{145510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38480, {{145511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38481, {{145513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38482, {{145514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38483, {{145515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38484, {{145517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38485, {{145520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38486, {{145523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38487, {{145524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38488, {{145526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38489, {{145527, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38490, {{145528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38491, {{145530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38492, {{145532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38493, {{145533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38494, {{145534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38495, {{145536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38496, {{145537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38497, {{145538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38498, {{145539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38499, {{145540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38500, {{145541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38501, {{145542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38502, {{145543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38503, {{145545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38504, {{145546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38505, {{145547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38506, {{145548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38507, {{145549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38508, {{145550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38509, {{145551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38510, {{145552, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38511, {{145553, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38512, {{145554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38513, {{145555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38514, {{145557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38515, {{145558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38516, {{145559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38517, {{145560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38518, {{145561, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38519, {{145562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38520, {{145563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38521, {{145564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38522, {{145565, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38523, {{145566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38524, {{145567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38525, {{145568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38526, {{145569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38527, {{145570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38528, {{145571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38529, {{145572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38530, {{145573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38531, {{145574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{38532, {{145575, 1, 0, 0},{145575, 1, 1, 16}, {0, 0, 0, 0}}},{38532, {{145575, 1, 0, 0},{145575, 1, 1, 16}, {0, 0, 0, 0}}},{38533, {{145575, 1, 0, 0},{145575, 1, 1, 17}, {0, 0, 0, 0}}},{38533, {{145575, 1, 0, 0},{145575, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{38534, {{145576, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38535, {{145577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38536, {{145578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38537, {{145579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38538, {{145580, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38539, {{145581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38540, {{145582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38542, {{145583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38543, {{145584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38544, {{145585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38545, {{145586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38546, {{145587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38547, {{145588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38548, {{145590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57880, {{145592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38550, {{145594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38551, {{145595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38552, {{145597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38554, {{145608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38555, {{145609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38556, {{145610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38557, {{145611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38558, {{145612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38559, {{145614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38560, {{145615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38561, {{145616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38562, {{145617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38563, {{145618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38564, {{145619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38565, {{145621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38566, {{145622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38567, {{145625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38568, {{145626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38569, {{145627, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38570, {{145629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38571, {{145630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38572, {{145631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38573, {{145632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38574, {{145633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38575, {{145634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38576, {{145635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38577, {{145636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38578, {{145637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38579, {{145638, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38580, {{145639, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38581, {{145640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38582, {{145641, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38583, {{145644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38584, {{145645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38585, {{145646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38586, {{145647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38587, {{145648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38588, {{145650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38589, {{145651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38590, {{145652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38591, {{145653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38592, {{145654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38593, {{145655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38594, {{145656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38595, {{145657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38596, {{145658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38597, {{145659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38598, {{145660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38599, {{145661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38600, {{145662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38601, {{145663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38602, {{145664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38603, {{145665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38604, {{145666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38605, {{145667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38606, {{145672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38607, {{145673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38608, {{145674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38609, {{145675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38610, {{145676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38611, {{145677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38612, {{145678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38613, {{145679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38614, {{145680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38615, {{145681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38616, {{145682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38617, {{145687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38618, {{145688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38619, {{145689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38620, {{145690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38621, {{145692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38622, {{145695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38623, {{145696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38624, {{145697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38625, {{145698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38626, {{145699, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38627, {{145704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38628, {{145705, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38629, {{145706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38630, {{145708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38631, {{145711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38632, {{145712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38633, {{145713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38634, {{145714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38635, {{145715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38636, {{145716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38637, {{145717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38638, {{145718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38639, {{145719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38640, {{145721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38641, {{145722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38642, {{145723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38643, {{145724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38644, {{145727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38645, {{145728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38646, {{145729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38647, {{145730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38648, {{145731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38649, {{145732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38650, {{145733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38651, {{145735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38652, {{145736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38653, {{145737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38654, {{145738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38655, {{145739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38656, {{145740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38657, {{145741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38658, {{145742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38659, {{145743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38660, {{145744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38661, {{145745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38662, {{145746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38663, {{145747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38664, {{145748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38665, {{145750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38666, {{145751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38667, {{145752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38668, {{145753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38670, {{145760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38671, {{145763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38672, {{145764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38673, {{145765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38674, {{145766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38675, {{145767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38676, {{145768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{38677, {{145769, 1, 0, 0},{145769, 1, 1, 16}, {0, 0, 0, 0}}},{38677, {{145769, 1, 0, 0},{145769, 1, 1, 16}, {0, 0, 0, 0}}},{57892, {{145769, 1, 0, 0},{145769, 1, 1, 17}, {0, 0, 0, 0}}},{57892, {{145769, 1, 0, 0},{145769, 1, 1, 17}, {0, 0, 0, 0}}},{57894, {{145769, 1, 0, 0},{145769, 1, 1, 18}, {0, 0, 0, 0}}},{57894, {{145769, 1, 0, 0},{145769, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{38678, {{145770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38679, {{145771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38680, {{145772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38681, {{145773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38682, {{145774, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38683, {{145776, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38684, {{145778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38685, {{145779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38686, {{145780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38687, {{145783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38688, {{145785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38689, {{145786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38690, {{145787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38691, {{145788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38692, {{145789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38693, {{145790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38694, {{145791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38695, {{145792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38696, {{145793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38697, {{145794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38698, {{145795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38699, {{145796, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38700, {{145797, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38701, {{145799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38702, {{145800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38703, {{145801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38704, {{145803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38705, {{145804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38706, {{145807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38707, {{145808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{68079, {{145809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38708, {{145812, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38709, {{145813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38710, {{145814, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38711, {{145815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38712, {{145816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38713, {{145817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38714, {{145820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38715, {{145821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38716, {{145822, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38717, {{145823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38718, {{145825, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38719, {{145826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38720, {{145827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38721, {{145830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38722, {{145831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38723, {{145832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38724, {{145833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38725, {{145834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38726, {{145835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38727, {{145837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38728, {{145839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38729, {{145841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38730, {{145842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38731, {{145845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38732, {{145846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38733, {{145847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38734, {{145848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38735, {{145849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38736, {{145850, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38737, {{145851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38738, {{145852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38739, {{145853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38740, {{145856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38741, {{145857, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38742, {{145858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38743, {{145859, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38744, {{145860, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38745, {{145862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38746, {{145863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38747, {{145864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38748, {{145865, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38749, {{145866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38750, {{145867, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38751, {{145868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38752, {{145869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38753, {{145870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38754, {{145871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38755, {{145873, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38756, {{145874, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38757, {{145875, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38758, {{145876, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38759, {{145877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38760, {{145878, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38761, {{145881, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38762, {{145882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38763, {{145883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38764, {{145884, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38765, {{145885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38766, {{145886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38767, {{145887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38768, {{145888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38769, {{145889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38770, {{145890, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38771, {{145891, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38772, {{145892, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38773, {{145893, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38774, {{145894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38775, {{145895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38776, {{145896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38777, {{145897, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38778, {{145898, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38779, {{145899, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38780, {{145901, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38781, {{145902, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38782, {{145906, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38783, {{145907, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38784, {{145908, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38785, {{145910, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38787, {{145912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38788, {{145913, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38789, {{145914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38790, {{145915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38791, {{145916, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38792, {{145917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38793, {{145918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38794, {{145919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38795, {{145920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38796, {{145921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38797, {{145922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38798, {{145923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38799, {{145925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38800, {{145926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38801, {{145927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38802, {{145928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38803, {{145929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38804, {{145930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38805, {{145931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38806, {{145932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38807, {{145933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38808, {{145934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38809, {{145935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38810, {{145936, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38811, {{145937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38812, {{145939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38813, {{145940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38814, {{145942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38815, {{145943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38816, {{145944, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38817, {{145945, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38818, {{145946, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38819, {{145947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38820, {{145948, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38821, {{145949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38822, {{145950, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38823, {{145951, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38824, {{145952, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38825, {{145953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38826, {{145954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38827, {{145955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38828, {{145956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38829, {{145957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38830, {{145958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38831, {{145962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38832, {{145963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38833, {{145964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38834, {{145969, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38835, {{145970, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38836, {{145971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38837, {{145972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38838, {{145973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38839, {{145974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38840, {{145975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38841, {{145976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38842, {{145977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38843, {{145978, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38844, {{145979, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38845, {{145982, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38846, {{145983, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38847, {{145984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38848, {{145987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38849, {{145988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38850, {{145989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38851, {{145990, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38852, {{145991, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38853, {{145992, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38854, {{145993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38855, {{145995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38857, {{145997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38858, {{145998, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38859, {{145999, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38860, {{146000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38861, {{146001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38862, {{146002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38863, {{146003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38864, {{146004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38865, {{146005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38866, {{146006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38867, {{146009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38868, {{146010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38869, {{146013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38870, {{146014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38871, {{146015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38872, {{146016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38873, {{146017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38874, {{146018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{38875, {{146019, 1, 0, 0},{146019, 1, 1, 16}, {0, 0, 0, 0}}},{38875, {{146019, 1, 0, 0},{146019, 1, 1, 16}, {0, 0, 0, 0}}},{38876, {{146019, 1, 0, 0},{146019, 1, 1, 17}, {0, 0, 0, 0}}},{38876, {{146019, 1, 0, 0},{146019, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38877, {{146023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38878, {{146024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38879, {{146025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38880, {{146026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38881, {{146027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38882, {{146028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38883, {{146029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38884, {{146031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38885, {{146032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38886, {{146033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38887, {{146034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38888, {{146035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{38889, {{146036, 1, 0, 0},{146036, 1, 1, 17}, {0, 0, 0, 0}}},{38889, {{146036, 1, 0, 0},{146036, 1, 1, 17}, {0, 0, 0, 0}}},{57912, {{146036, 1, 0, 0},{146036, 1, 1, 16}, {0, 0, 0, 0}}},{57912, {{146036, 1, 0, 0},{146036, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{38890, {{146037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38891, {{146038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38892, {{146039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57913, {{146041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38893, {{146042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38894, {{146043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38895, {{146044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38896, {{146046, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38897, {{146047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38898, {{146048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38899, {{146050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38900, {{146051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38901, {{146052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38902, {{146053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38903, {{146054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38904, {{146055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38905, {{146057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57917, {{146058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38906, {{146059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38907, {{146060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{38908, {{146061, 1, 0, 0},{146061, 1, 1, 16}, {0, 0, 0, 0}}},{38908, {{146061, 1, 0, 0},{146061, 1, 1, 16}, {0, 0, 0, 0}}},{38909, {{146061, 1, 0, 0},{146061, 1, 1, 17}, {0, 0, 0, 0}}},{38909, {{146061, 1, 0, 0},{146061, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38910, {{146063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38911, {{146064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38912, {{146070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38913, {{146071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38914, {{146072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38915, {{146073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57923, {{146074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38916, {{146075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38917, {{146081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38918, {{146083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38919, {{146089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38920, {{146090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38921, {{146093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38922, {{146094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38923, {{146095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38924, {{146096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38925, {{146098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38926, {{146099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57925, {{146100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38927, {{146102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38928, {{146103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38929, {{146104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38930, {{146105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38931, {{146108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38932, {{146109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38933, {{146110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38934, {{146112, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38935, {{146113, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38936, {{146121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38937, {{146122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38938, {{146124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38939, {{146128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38940, {{146130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38941, {{146131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38942, {{146134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38943, {{146136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38944, {{146137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38945, {{146138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38946, {{146140, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38947, {{146141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38948, {{146142, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38949, {{146143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38950, {{146145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38951, {{146147, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38952, {{146148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38953, {{146149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38954, {{146150, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38955, {{146151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38956, {{146152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38957, {{146153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38958, {{146154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38959, {{146155, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38960, {{146156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38961, {{146157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38962, {{146159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38963, {{146161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38964, {{146163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38965, {{146164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38966, {{146165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38967, {{146166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38968, {{146169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38969, {{146170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38971, {{146171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38972, {{146178, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38973, {{146179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38974, {{146181, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38975, {{146182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38976, {{146184, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38977, {{146185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38978, {{146186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38979, {{146187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38980, {{146188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38981, {{146189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38982, {{146190, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38983, {{146191, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38984, {{146193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38986, {{146194, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38985, {{146195, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38987, {{146197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38988, {{146198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38989, {{146199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38990, {{146201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38991, {{146202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38992, {{146203, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38993, {{146204, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38994, {{146209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38995, {{146210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38996, {{146211, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38997, {{146217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38998, {{146218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{38999, {{146219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39000, {{146220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39001, {{146221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39002, {{146222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39003, {{146228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39004, {{146232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39005, {{146233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39006, {{146234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39007, {{146235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39008, {{146236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39009, {{146237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39010, {{146238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39011, {{146239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39012, {{146240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39013, {{146241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39014, {{146245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39015, {{146246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39016, {{146247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39017, {{146256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39018, {{146262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39019, {{146263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39020, {{146264, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39021, {{146268, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39022, {{146269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39023, {{146270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39024, {{146271, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39025, {{146272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39026, {{146273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39027, {{146274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39028, {{146275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39029, {{146285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39030, {{146286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39031, {{146289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39032, {{146290, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39033, {{146291, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39034, {{146292, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39035, {{146293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39036, {{146294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39037, {{146295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39038, {{146296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39039, {{146297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39040, {{146298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39041, {{146299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39042, {{146300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39043, {{146301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39044, {{146307, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39045, {{146308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39046, {{146312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39047, {{146313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39048, {{146314, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39049, {{146315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39050, {{146316, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39051, {{146317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39052, {{146318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39053, {{146319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39054, {{146320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39055, {{146321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39056, {{146322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39057, {{146324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39058, {{146325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39059, {{146336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39060, {{146337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39061, {{146338, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39062, {{146339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39063, {{146343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39064, {{146345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39065, {{146346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39066, {{146347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{39067, {{146348, 1, 0, 0},{146348, 1, 1, 16}, {0, 0, 0, 0}}},{39067, {{146348, 1, 0, 0},{146348, 1, 1, 16}, {0, 0, 0, 0}}},{57931, {{146348, 1, 0, 0},{146348, 1, 1, 17}, {0, 0, 0, 0}}},{57931, {{146348, 1, 0, 0},{146348, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39068, {{146350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39069, {{146351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39070, {{146353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39071, {{146354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39072, {{146355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39073, {{146356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57932, {{146360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39075, {{146366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39076, {{146367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39077, {{146370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39078, {{146371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39079, {{146373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39080, {{146374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39081, {{146375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39082, {{146376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39083, {{146377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39084, {{146378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39085, {{146379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39086, {{146380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39087, {{146381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39088, {{146382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39089, {{146383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39074, {{146384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39090, {{146386, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39091, {{146387, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39092, {{146389, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39093, {{146396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39094, {{146397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39095, {{146399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39096, {{146402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39097, {{146404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39098, {{146405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39099, {{146406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39100, {{146407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39101, {{146408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39102, {{146409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39103, {{146410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39104, {{146411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39105, {{146412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39106, {{146413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39107, {{146414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39108, {{146415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39109, {{146416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39110, {{146417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39111, {{146418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39112, {{146421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39113, {{146424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39114, {{146425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39115, {{146426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39116, {{146427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39117, {{146428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39118, {{146429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39119, {{146430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39120, {{146435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39121, {{146438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39122, {{146441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39123, {{146442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39124, {{146443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39125, {{146444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39126, {{146445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39127, {{146446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{39128, {{146448, 1, 0, 0},{146448, 1, 1, 16}, {0, 0, 0, 0}}},{39128, {{146448, 1, 0, 0},{146448, 1, 1, 16}, {0, 0, 0, 0}}},{68080, {{146448, 1, 0, 0},{146448, 1, 1, 17}, {0, 0, 0, 0}}},{68080, {{146448, 1, 0, 0},{146448, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39129, {{146452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39130, {{146454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39131, {{146455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39132, {{146456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39133, {{146458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39134, {{146459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39135, {{146460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39136, {{146462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39137, {{146463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39138, {{146464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39139, {{146465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39140, {{146466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39141, {{146471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39142, {{146472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39143, {{146473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39144, {{146475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39145, {{146476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39146, {{146477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39147, {{146478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39148, {{146482, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39149, {{146486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39150, {{146487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39151, {{146488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39152, {{146490, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39153, {{146492, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39154, {{146495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39155, {{146496, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39156, {{146498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39157, {{146500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39158, {{146501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39159, {{146503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39160, {{146504, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39161, {{146505, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39162, {{146506, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39163, {{146509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39164, {{146510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39165, {{146511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39166, {{146512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39167, {{146513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39168, {{146514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39169, {{146515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39170, {{146517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39171, {{146518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39172, {{146521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39173, {{146522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39174, {{146523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39175, {{146524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39176, {{146526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39177, {{146528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39178, {{146529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39179, {{146532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39180, {{146534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39181, {{146537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39182, {{146538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39183, {{146539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39184, {{146542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39185, {{146543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39186, {{146545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39187, {{146547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39188, {{146548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{39189, {{146549, 1, 0, 0},{146549, 1, 1, 16}, {0, 0, 0, 0}}},{39189, {{146549, 1, 0, 0},{146549, 1, 1, 16}, {0, 0, 0, 0}}},{39190, {{146549, 1, 0, 0},{146549, 1, 1, 18}, {0, 0, 0, 0}}},{39190, {{146549, 1, 0, 0},{146549, 1, 1, 18}, {0, 0, 0, 0}}},{57939, {{146549, 1, 0, 0},{146549, 1, 1, 17}, {0, 0, 0, 0}}},{57939, {{146549, 1, 0, 0},{146549, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{39191, {{146550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39192, {{146551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39193, {{146554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39194, {{146555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39195, {{146557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39196, {{146559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39197, {{146560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39198, {{146561, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39199, {{146562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39200, {{146563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39201, {{146564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39202, {{146570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39203, {{146571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39204, {{146574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39205, {{146575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39206, {{146578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39207, {{146579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39208, {{146580, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39209, {{146581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39210, {{146585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39211, {{146588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39212, {{146589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39213, {{146591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39214, {{146592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39215, {{146593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39216, {{146597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{39217, {{146600, 1, 0, 0},{146600, 1, 1, 16}, {0, 0, 0, 0}}},{39217, {{146600, 1, 0, 0},{146600, 1, 1, 16}, {0, 0, 0, 0}}},{39218, {{146600, 1, 0, 0},{146600, 1, 1, 17}, {0, 0, 0, 0}}},{39218, {{146600, 1, 0, 0},{146600, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39219, {{146609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39220, {{146610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39221, {{146611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39222, {{146612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39223, {{146613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39224, {{146614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39225, {{146615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39226, {{146616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39227, {{146617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39228, {{146618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39229, {{146619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39230, {{146620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39231, {{146621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39232, {{146622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39233, {{146623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39234, {{146624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39235, {{146625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39236, {{146626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39237, {{146628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39238, {{146629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39239, {{146634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39240, {{146635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39241, {{146641, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39242, {{146648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39243, {{146649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39244, {{146650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39245, {{146651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39246, {{146653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39247, {{146654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39248, {{146655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39249, {{146656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39250, {{146659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39251, {{146660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39252, {{146661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39253, {{146662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39254, {{146663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39255, {{146664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39256, {{146665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39257, {{146666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39258, {{146667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39259, {{146668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39260, {{146669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39261, {{146670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39263, {{146671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39264, {{146672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39265, {{146673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39267, {{146674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39268, {{146675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39269, {{146676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39270, {{146678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39271, {{146679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{39272, {{146686, 1, 0, 0},{146686, 1, 1, 17}, {0, 0, 0, 0}}},{39272, {{146686, 1, 0, 0},{146686, 1, 1, 17}, {0, 0, 0, 0}}},{59775, {{146686, 1, 0, 0},{146686, 1, 1, 18}, {0, 0, 0, 0}}},{59775, {{146686, 1, 0, 0},{146686, 1, 1, 18}, {0, 0, 0, 0}}},{59776, {{146686, 1, 0, 0},{146686, 1, 1, 19}, {0, 0, 0, 0}}},{59776, {{146686, 1, 0, 0},{146686, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39273, {{146688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39274, {{146689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39275, {{146702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39278, {{146705, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39279, {{146706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39280, {{146708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39281, {{146709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39282, {{146710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39283, {{146713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39284, {{146719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39285, {{146720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39286, {{146721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39287, {{146722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39288, {{146723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39289, {{146724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39290, {{146725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39291, {{146726, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39292, {{146727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39293, {{146729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39294, {{146730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39295, {{146731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39296, {{146732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39297, {{146733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39298, {{146734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39299, {{146735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39300, {{146736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39301, {{146737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39302, {{146738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39303, {{146739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39304, {{146740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39307, {{146741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39308, {{146742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39309, {{146743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39310, {{146744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39311, {{146745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39312, {{146746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39313, {{146747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39314, {{146748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39315, {{146749, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39316, {{146750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39317, {{146751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39318, {{146752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39320, {{146779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39322, {{146783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{15437, {{146784, 1, 0, 0},{146784, 1, 1, 16}, {0, 0, 0, 0}}},{15437, {{146784, 1, 0, 0},{146784, 1, 1, 16}, {0, 0, 0, 0}}},{68061, {{146784, 1, 0, 0},{146784, 1, 1, 17}, {0, 0, 0, 0}}},{68061, {{146784, 1, 0, 0},{146784, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39323, {{146786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39324, {{146788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39325, {{146790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39326, {{146791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39327, {{146792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39328, {{146793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39329, {{146796, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39330, {{146801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39331, {{146807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39332, {{146808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39333, {{146809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39334, {{146810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39335, {{146811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{39336, {{146813, 1, 0, 0},{146813, 1, 1, 16}, {0, 0, 0, 0}}},{39336, {{146813, 1, 0, 0},{146813, 1, 1, 16}, {0, 0, 0, 0}}},{57954, {{146813, 1, 0, 0},{146813, 1, 1, 17}, {0, 0, 0, 0}}},{57954, {{146813, 1, 0, 0},{146813, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{39337, {{146814, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39338, {{146815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39339, {{146816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39340, {{146817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39341, {{146818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39342, {{146819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39343, {{146820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39344, {{146821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39345, {{146822, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39346, {{146823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39347, {{146824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39348, {{146825, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39349, {{146826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39350, {{146827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39351, {{146828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39352, {{146829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39353, {{146830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39354, {{146831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39355, {{146832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39356, {{146833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39357, {{146834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39358, {{146835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39359, {{146836, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39360, {{146837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39361, {{146838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39319, {{146839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39362, {{146840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39363, {{146841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39364, {{146842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39365, {{146843, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39366, {{146844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39367, {{146845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39368, {{146846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39369, {{146847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39378, {{146848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39370, {{146849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39371, {{146850, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39372, {{146851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39373, {{146852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39374, {{146853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39375, {{146854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39379, {{146879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39376, {{146889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39381, {{146899, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39383, {{146908, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39384, {{146909, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39385, {{146911, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39386, {{146913, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39387, {{146914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39389, {{146916, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39390, {{146917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39391, {{146918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39392, {{146919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39393, {{146921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39395, {{146933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39396, {{146934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{39398, {{146937, 1, 0, 0},{146937, 1, 1, 17}, {0, 0, 0, 0}}},{39398, {{146937, 1, 0, 0},{146937, 1, 1, 17}, {0, 0, 0, 0}}},{39399, {{146937, 1, 0, 0},{146937, 1, 1, 18}, {0, 0, 0, 0}}},{39399, {{146937, 1, 0, 0},{146937, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{39400, {{146938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39402, {{146940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39403, {{146941, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39404, {{146942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39405, {{146943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39406, {{146944, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39407, {{146945, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39408, {{146946, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39409, {{146947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39410, {{146948, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39411, {{146949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39412, {{146950, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39414, {{146951, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39415, {{146952, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39416, {{146953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39417, {{146954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39418, {{146956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39419, {{146957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39420, {{146958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39421, {{146959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39422, {{146960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39423, {{146961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39424, {{146962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39425, {{146963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39426, {{146964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39427, {{146965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39428, {{146966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39429, {{146967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39430, {{146968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39431, {{146969, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39432, {{146970, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39433, {{146971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39434, {{146972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39435, {{146973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39436, {{146974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39437, {{146975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39438, {{146976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39439, {{146977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39440, {{146978, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{39441, {{146994, 1, 0, 0},{146994, 1, 1, 16}, {0, 0, 0, 0}}},{39441, {{146994, 1, 0, 0},{146994, 1, 1, 16}, {0, 0, 0, 0}}},{39442, {{146994, 1, 0, 0},{146994, 1, 1, 17}, {0, 0, 0, 0}}},{39442, {{146994, 1, 0, 0},{146994, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39397, {{147000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39444, {{147013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39445, {{147014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39446, {{147015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39447, {{147017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39448, {{147018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39449, {{147020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39450, {{147021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39451, {{147022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39452, {{147034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39453, {{147037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39454, {{147038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39455, {{147039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39456, {{147040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39457, {{147041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39458, {{147042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39459, {{147043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39460, {{147044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39461, {{147045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39463, {{147046, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39464, {{147047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39465, {{147048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39466, {{147049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39467, {{147050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39468, {{147052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39469, {{147053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39470, {{147054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39471, {{147055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39490, {{147056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39472, {{147057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39473, {{147058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39474, {{147059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39475, {{147060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39476, {{147061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39477, {{147062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39478, {{147063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39479, {{147064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39480, {{147065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39481, {{147066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39482, {{147067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39483, {{147068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39484, {{147069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39485, {{147070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39486, {{147071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39487, {{147072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39488, {{147073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39489, {{147085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59801, {{147108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39493, {{147116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39495, {{147119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39496, {{147120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39497, {{147121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39498, {{147122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39499, {{147123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39500, {{147124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39501, {{147148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39503, {{147149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39504, {{147150, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39505, {{147151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39506, {{147152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39507, {{147153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39508, {{147154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39509, {{147155, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39510, {{147156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39511, {{147157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39512, {{147158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39513, {{147159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39514, {{147160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39515, {{147161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39516, {{147162, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39517, {{147163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39518, {{147164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39519, {{147165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39520, {{147166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39521, {{147168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39522, {{147169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39523, {{147170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39524, {{147171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39525, {{147173, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39526, {{147174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39527, {{147175, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39528, {{147176, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39529, {{147177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39530, {{147178, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39531, {{147179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39532, {{147180, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39533, {{147181, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39534, {{147182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39535, {{147183, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39536, {{147185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39537, {{147186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39538, {{147225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39539, {{147226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {12, {{39540, {{147227, 1, 0, 0},{147227, 1, 1, 21}, {0, 0, 0, 0}}},{39540, {{147227, 1, 0, 0},{147227, 1, 1, 21}, {0, 0, 0, 0}}},{59815, {{147227, 1, 0, 0},{147227, 1, 1, 16}, {0, 0, 0, 0}}},{59815, {{147227, 1, 0, 0},{147227, 1, 1, 16}, {0, 0, 0, 0}}},{59816, {{147227, 1, 0, 0},{147227, 1, 1, 17}, {0, 0, 0, 0}}},{59816, {{147227, 1, 0, 0},{147227, 1, 1, 17}, {0, 0, 0, 0}}},{59817, {{147227, 1, 0, 0},{147227, 1, 1, 18}, {0, 0, 0, 0}}},{59817, {{147227, 1, 0, 0},{147227, 1, 1, 18}, {0, 0, 0, 0}}},{59818, {{147227, 1, 0, 0},{147227, 1, 1, 19}, {0, 0, 0, 0}}},{59818, {{147227, 1, 0, 0},{147227, 1, 1, 19}, {0, 0, 0, 0}}},{59819, {{147227, 1, 0, 0},{147227, 1, 1, 20}, {0, 0, 0, 0}}},{59819, {{147227, 1, 0, 0},{147227, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{39541, {{147228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39542, {{147229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39543, {{147230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39544, {{147231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39545, {{147232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39546, {{147233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39547, {{147234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39548, {{147235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39549, {{147259, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39550, {{147267, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39551, {{147268, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39552, {{147269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39553, {{147270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39554, {{147271, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39555, {{147272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39556, {{147273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39557, {{147274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39558, {{147276, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39601, {{147277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39559, {{147278, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39560, {{147279, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39561, {{147280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39562, {{147281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39563, {{147282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39564, {{147283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39565, {{147284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39566, {{147285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39567, {{147286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39568, {{147288, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39569, {{147290, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39570, {{147291, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39571, {{147292, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39572, {{147293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39573, {{147294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39574, {{147295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39575, {{147296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39576, {{147297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39578, {{147298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39579, {{147299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39580, {{147300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39581, {{147301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39582, {{147302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39583, {{147303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39584, {{147304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39585, {{147305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39586, {{147306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39587, {{147307, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39588, {{147308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39589, {{147309, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39590, {{147310, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39591, {{147311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39592, {{147312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39593, {{147313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39594, {{147314, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39595, {{147315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39596, {{147316, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39597, {{147317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39598, {{147318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39599, {{147325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39600, {{147326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57970, {{147328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39602, {{147333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39603, {{147363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39604, {{147364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39605, {{147366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39606, {{147368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39607, {{147369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39608, {{147370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39609, {{147371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39610, {{147372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39611, {{147373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39612, {{147374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57971, {{147376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39613, {{147397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39614, {{147398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39615, {{147399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39616, {{147400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39617, {{147401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39618, {{147402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39619, {{147403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39620, {{147404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39621, {{147405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39622, {{147407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39623, {{147408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39624, {{147409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39625, {{147410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39626, {{147411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39627, {{147412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39628, {{147413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39629, {{147414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39630, {{147415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39631, {{147416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39632, {{147417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39633, {{147418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39634, {{147419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39635, {{147420, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39636, {{147421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39637, {{147422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39638, {{147423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39639, {{147424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39640, {{147425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39641, {{147426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39642, {{147427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39643, {{147428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39644, {{147429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39645, {{147442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39646, {{147443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39647, {{147444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39648, {{147460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39649, {{147461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39650, {{147462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39651, {{147463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39652, {{147464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39653, {{147467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39654, {{147470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39655, {{147485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39656, {{147490, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39657, {{147491, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39658, {{147492, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39659, {{147494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39660, {{147495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39661, {{147496, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39662, {{147497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39663, {{147498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39664, {{147499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39665, {{147500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39666, {{147502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39667, {{147503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39668, {{147504, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39669, {{147505, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39670, {{147506, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39671, {{147507, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39672, {{147508, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39673, {{147509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39674, {{147510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39675, {{147511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39676, {{147529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39677, {{147530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39678, {{147531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39680, {{147532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39681, {{147533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39683, {{147551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39684, {{147553, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39685, {{147554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39686, {{147555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39687, {{147556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39688, {{147557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39689, {{147558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39690, {{147559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39691, {{147560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39692, {{147562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39693, {{147563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39694, {{147564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39695, {{147566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39696, {{147567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39697, {{147568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39698, {{147569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39699, {{147570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39700, {{147571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39701, {{147572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39702, {{147573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39703, {{147574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39704, {{147575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39705, {{147576, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39706, {{147577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39707, {{147578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39708, {{147579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39709, {{147580, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39710, {{147581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39711, {{147582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39712, {{147583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39713, {{147584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39715, {{147585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39716, {{147586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39717, {{147587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39718, {{147588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39719, {{147589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39720, {{147590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39721, {{147591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39723, {{147606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39724, {{147616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39725, {{147617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39726, {{147618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{39727, {{147619, 1, 0, 0},{147619, 1, 1, 17}, {0, 0, 0, 0}}},{39727, {{147619, 1, 0, 0},{147619, 1, 1, 17}, {0, 0, 0, 0}}},{39728, {{147619, 1, 0, 0},{147619, 1, 1, 16}, {0, 0, 0, 0}}},{39728, {{147619, 1, 0, 0},{147619, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{39729, {{147620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39730, {{147621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39731, {{147622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39732, {{147634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39733, {{147636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39734, {{147637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39735, {{147638, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39736, {{147639, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39737, {{147640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39738, {{147641, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39739, {{147642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39740, {{147643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39741, {{147644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39742, {{147645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39743, {{147646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39744, {{147647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39745, {{147648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39746, {{147649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39747, {{147650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39748, {{147651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39749, {{147652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39750, {{147658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39751, {{147669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39752, {{147671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39753, {{147672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39754, {{147673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39756, {{147689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39757, {{147690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39758, {{147691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39759, {{147692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39760, {{147693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39722, {{147694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39761, {{147695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39762, {{147696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39763, {{147697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39764, {{147698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39798, {{147699, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39765, {{147700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39766, {{147701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39767, {{147702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39768, {{147703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39769, {{147704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39770, {{147705, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39771, {{147706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39772, {{147707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39773, {{147708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39774, {{147709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39775, {{147715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39777, {{147721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39778, {{147726, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39779, {{147727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39780, {{147728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39781, {{147729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2233, {{147730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39782, {{147731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39783, {{147732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39784, {{147733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39785, {{147734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39786, {{147735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39787, {{147736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39788, {{147737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39789, {{147738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39790, {{147739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39791, {{147740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39792, {{147741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39793, {{147750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39794, {{147751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39795, {{147752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39796, {{147753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39797, {{147754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57979, {{147755, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39799, {{147764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39800, {{147765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39801, {{147766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39802, {{147767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{39803, {{147768, 1, 0, 0},{147768, 1, 1, 16}, {0, 0, 0, 0}}},{39803, {{147768, 1, 0, 0},{147768, 1, 1, 16}, {0, 0, 0, 0}}},{68081, {{147768, 1, 0, 0},{147768, 1, 1, 17}, {0, 0, 0, 0}}},{68081, {{147768, 1, 0, 0},{147768, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39804, {{147770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39805, {{147771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39806, {{147773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39807, {{147781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39808, {{147782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39809, {{147783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39810, {{147786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39811, {{147787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39812, {{147788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39813, {{147789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39814, {{147790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39815, {{147791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39816, {{147792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39817, {{147793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39818, {{147794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39819, {{147795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39820, {{147803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39821, {{147805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39822, {{147809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39823, {{147810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39824, {{147811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39825, {{147812, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39826, {{147814, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39827, {{147816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39828, {{147817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39829, {{147818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39830, {{147819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39831, {{147821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39832, {{147824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39833, {{147825, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39834, {{147826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39835, {{147827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39836, {{147828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39837, {{147831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39838, {{147832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39839, {{147833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39840, {{147835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39841, {{147836, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39842, {{147837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39843, {{147839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39844, {{147840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39845, {{147841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39846, {{147842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39847, {{147843, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39848, {{147844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39849, {{147845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39850, {{147846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39851, {{147851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39852, {{147853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39853, {{147854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39854, {{147855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39855, {{147856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39856, {{147857, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39857, {{147858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39858, {{147865, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39859, {{147866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39860, {{147868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39861, {{147869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39862, {{147870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39863, {{147871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39864, {{147887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39865, {{147888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39866, {{147889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39867, {{147890, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39868, {{147891, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39869, {{147892, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39870, {{147900, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39871, {{147901, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39872, {{147902, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39873, {{147903, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39874, {{147904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39875, {{147905, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39877, {{147910, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39878, {{147911, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39880, {{147930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39881, {{147931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39882, {{147932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39883, {{147933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39884, {{147934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39885, {{147935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39886, {{147936, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39887, {{147937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39888, {{147947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39889, {{147951, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39890, {{147952, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39891, {{147953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39892, {{147954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39897, {{147955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39893, {{147956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39894, {{147957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39895, {{147958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39896, {{147959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39898, {{147966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39899, {{147985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39900, {{147986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39901, {{147987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39902, {{147988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39903, {{147989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39904, {{147990, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39905, {{148002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39906, {{148003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39907, {{148006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39908, {{148007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39909, {{148008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39910, {{148009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39911, {{148010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39912, {{148011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39913, {{148012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39914, {{148013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39915, {{148014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39916, {{148015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39917, {{148016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39918, {{148017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39919, {{148018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57992, {{148039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39921, {{148042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39922, {{148044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39923, {{148045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39924, {{148046, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39925, {{148047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39926, {{148049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39927, {{148061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39920, {{148062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39928, {{148063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39929, {{148065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39930, {{148066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{39931, {{148067, 1, 0, 0},{148067, 1, 1, 16}, {0, 0, 0, 0}}},{39931, {{148067, 1, 0, 0},{148067, 1, 1, 16}, {0, 0, 0, 0}}},{39932, {{148067, 1, 0, 0},{148067, 1, 1, 17}, {0, 0, 0, 0}}},{39932, {{148067, 1, 0, 0},{148067, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{39933, {{148068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39934, {{148069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39935, {{148070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39936, {{148071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39937, {{148073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39938, {{148074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39939, {{148075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39940, {{148076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39941, {{148077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39942, {{148078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39943, {{148079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39944, {{148089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39945, {{148090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39946, {{148091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{39947, {{148101, 1, 0, 0},{148101, 1, 1, 17}, {0, 0, 0, 0}}},{39947, {{148101, 1, 0, 0},{148101, 1, 1, 17}, {0, 0, 0, 0}}},{59860, {{148101, 1, 0, 0},{148101, 1, 1, 18}, {0, 0, 0, 0}}},{59860, {{148101, 1, 0, 0},{148101, 1, 1, 18}, {0, 0, 0, 0}}},{59861, {{148101, 1, 0, 0},{148101, 1, 1, 16}, {0, 0, 0, 0}}},{59861, {{148101, 1, 0, 0},{148101, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39948, {{148119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39949, {{148121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39950, {{148122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39951, {{148123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39952, {{148124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39954, {{148143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39956, {{148146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39957, {{148148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39958, {{148149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39959, {{148150, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39960, {{148151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39961, {{148153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39962, {{148154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39963, {{148155, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39964, {{148156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39965, {{148157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39966, {{148158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39967, {{148159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39968, {{148160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39969, {{148161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39970, {{148162, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39971, {{148163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39972, {{148164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2290, {{148165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39973, {{148166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39974, {{148167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59865, {{148188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39976, {{148200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39977, {{148201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39978, {{148202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39979, {{148203, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39980, {{148204, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39981, {{148205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {5, {{30309, {{64108, 1, 0, 0},{148206, 1, 0, 0},{148206, 1, 1, 0},{148206, 1, 1, 19}, {0, 0, 0, 0}}},{30309, {{64108, 1, 0, 0},{148206, 1, 0, 0},{148206, 1, 1, 0},{148206, 1, 1, 19}, {0, 0, 0, 0}}},{30309, {{64108, 1, 0, 0},{148206, 1, 0, 0},{148206, 1, 1, 0},{148206, 1, 1, 19}, {0, 0, 0, 0}}},{39982, {{148206, 1, 0, 0},{148206, 1, 1, 18}, {0, 0, 0, 0}}},{39982, {{148206, 1, 0, 0},{148206, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{39983, {{148207, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39984, {{148208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39985, {{148209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39986, {{148211, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39987, {{148222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39988, {{148225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39989, {{148227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39990, {{148232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39991, {{148233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39992, {{148234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39993, {{148235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39994, {{148236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39995, {{148237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39996, {{148238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39997, {{148239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39998, {{148240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39999, {{148241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40000, {{148242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40001, {{148243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40002, {{148244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40003, {{148245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40004, {{148247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40005, {{148251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40006, {{148252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40008, {{148279, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40009, {{148280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40010, {{148282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40011, {{148283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40012, {{148284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40013, {{148285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40014, {{148287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40015, {{148288, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40016, {{148289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40017, {{148301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40018, {{148304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40019, {{148306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40020, {{148307, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40021, {{148308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40022, {{148309, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40023, {{148310, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40024, {{148311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40025, {{148312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40026, {{148313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40027, {{148314, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40028, {{148315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40029, {{148316, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40030, {{148317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40031, {{148327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40032, {{148349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40033, {{148350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40068, {{148353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40034, {{148354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40035, {{148375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40036, {{148376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40037, {{148377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40038, {{148378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40039, {{148381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40040, {{148382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40041, {{148383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40042, {{148384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40043, {{148385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40044, {{148386, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40045, {{148387, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40046, {{148388, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40047, {{148389, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40048, {{148390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40049, {{148391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40050, {{148392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40051, {{148393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40052, {{148395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40053, {{148396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40054, {{148397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40055, {{148398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40056, {{148399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40057, {{148400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40058, {{148401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40059, {{148402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40060, {{148403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40061, {{148404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40062, {{148405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40063, {{148406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40064, {{148407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40065, {{148408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40067, {{148412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40069, {{148416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{40070, {{148417, 1, 0, 0},{148417, 1, 1, 16}, {0, 0, 0, 0}}},{40070, {{148417, 1, 0, 0},{148417, 1, 1, 16}, {0, 0, 0, 0}}},{59875, {{148417, 1, 0, 0},{148417, 1, 1, 18}, {0, 0, 0, 0}}},{59875, {{148417, 1, 0, 0},{148417, 1, 1, 18}, {0, 0, 0, 0}}},{59876, {{148417, 1, 0, 0},{148417, 1, 1, 17}, {0, 0, 0, 0}}},{59876, {{148417, 1, 0, 0},{148417, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40072, {{148432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40073, {{148439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40074, {{148440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40075, {{148441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40076, {{148442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40077, {{148443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40078, {{148444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40079, {{148445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40080, {{148446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40081, {{148447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40082, {{148448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40083, {{148461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40084, {{148462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40085, {{148463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40086, {{148464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40087, {{148468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40088, {{148469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40089, {{148470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40090, {{148471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40091, {{148472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40092, {{148473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40093, {{148474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40094, {{148475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40095, {{148476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40096, {{148477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40097, {{148478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40098, {{148479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40099, {{148480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40100, {{148498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40101, {{148500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40102, {{148501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40103, {{148503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40104, {{148510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40105, {{148511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40106, {{148518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40107, {{148519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40108, {{148520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40109, {{148521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40110, {{148522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40111, {{148523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40112, {{148524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40113, {{148525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40114, {{148526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40115, {{148527, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40116, {{148528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40117, {{148529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40118, {{148530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40119, {{148531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40120, {{148535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40121, {{148546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40122, {{148547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40123, {{148548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40124, {{148549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40125, {{148550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40126, {{148554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40127, {{148555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40128, {{148556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40129, {{148562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40130, {{148563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40131, {{148564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40132, {{148566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40133, {{148567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40135, {{148577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40136, {{148578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40137, {{148579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40138, {{148580, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40139, {{148581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40140, {{148582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40141, {{148583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40134, {{148584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40142, {{148585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40143, {{148586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40144, {{148593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40145, {{148595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40146, {{148596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40147, {{148597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40148, {{148598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40149, {{148599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40150, {{148600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40151, {{148601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40152, {{148602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40153, {{148604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40154, {{148605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40155, {{148606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40156, {{148607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40157, {{148608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40158, {{148609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40159, {{148610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40160, {{148614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40162, {{148617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40161, {{148627, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40163, {{148628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40164, {{148630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40165, {{148631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40166, {{148636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40167, {{148637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40168, {{148638, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40169, {{148644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40170, {{148645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40171, {{148646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40172, {{148647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40173, {{148651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40174, {{148652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40175, {{148654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40176, {{148666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40177, {{148667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40178, {{148668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40179, {{148669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40180, {{148670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40181, {{148676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40182, {{148677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40183, {{148680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40184, {{148681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40185, {{148682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40186, {{148687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40187, {{148688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40188, {{148689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40189, {{148690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40190, {{148692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40191, {{148693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40192, {{148695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40193, {{148698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40194, {{148700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40195, {{148701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40196, {{148702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40197, {{148703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40198, {{148704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40199, {{148705, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40200, {{148708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40201, {{148709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40202, {{148710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40203, {{148714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40204, {{148715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40205, {{148716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40206, {{148717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40207, {{148718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40208, {{148719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40209, {{148721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40210, {{148722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40211, {{148723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40212, {{148724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40213, {{148725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40214, {{148726, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40215, {{148727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40216, {{148728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40217, {{148733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40218, {{148734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40219, {{148736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40220, {{148737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40221, {{148738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59270, {{148742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40223, {{148743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40224, {{148744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40225, {{148745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40226, {{148746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40227, {{148747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40228, {{148748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40222, {{148749, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{40229, {{148752, 1, 0, 0},{148752, 1, 1, 16}, {0, 0, 0, 0}}},{40229, {{148752, 1, 0, 0},{148752, 1, 1, 16}, {0, 0, 0, 0}}},{40230, {{148752, 1, 0, 0},{148752, 1, 1, 17}, {0, 0, 0, 0}}},{40230, {{148752, 1, 0, 0},{148752, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{40231, {{148753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40232, {{148754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{40233, {{148756, 1, 0, 0},{148756, 1, 1, 16}, {0, 0, 0, 0}}},{40233, {{148756, 1, 0, 0},{148756, 1, 1, 16}, {0, 0, 0, 0}}},{40234, {{148756, 1, 0, 0},{148756, 1, 1, 17}, {0, 0, 0, 0}}},{40234, {{148756, 1, 0, 0},{148756, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{40235, {{148757, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40236, {{148759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40237, {{148760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40238, {{148761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40239, {{148762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40240, {{148764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40241, {{148765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40242, {{148767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40243, {{148768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40244, {{148770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40245, {{148771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40246, {{148772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40247, {{148773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40248, {{148774, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40249, {{148775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40251, {{148777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40252, {{148778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40253, {{148780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40254, {{148782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40255, {{148784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40256, {{148785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40257, {{148786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40258, {{148790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40259, {{148793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40260, {{148800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40261, {{148801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40262, {{148804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40263, {{148805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40264, {{148806, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40265, {{148807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40266, {{148808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40267, {{148810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40268, {{148811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40269, {{148813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40270, {{148816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40271, {{148818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40272, {{148819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40273, {{148821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40274, {{148824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40275, {{148826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40276, {{148827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40277, {{148829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40278, {{148830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40279, {{148831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40280, {{148832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {2, {{40281, {{148834, 1, 0, 0}, {0, 0, 0, 0}}},{40282, {{148834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40283, {{148835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{40284, {{148836, 1, 0, 0},{148836, 1, 1, 17}, {0, 0, 0, 0}}},{40284, {{148836, 1, 0, 0},{148836, 1, 1, 17}, {0, 0, 0, 0}}},{58019, {{148836, 1, 0, 0},{148836, 1, 1, 16}, {0, 0, 0, 0}}},{58019, {{148836, 1, 0, 0},{148836, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{40285, {{148837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40286, {{148838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40287, {{148839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{40288, {{148840, 1, 0, 0},{148840, 1, 1, 16}, {0, 0, 0, 0}}},{40288, {{148840, 1, 0, 0},{148840, 1, 1, 16}, {0, 0, 0, 0}}},{58021, {{148840, 1, 0, 0},{148840, 1, 1, 17}, {0, 0, 0, 0}}},{58021, {{148840, 1, 0, 0},{148840, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{40289, {{148841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40290, {{148842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40291, {{148843, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40292, {{148845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40293, {{148846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40294, {{148847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40295, {{148848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40296, {{148849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40297, {{148850, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40298, {{148852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40299, {{148853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40300, {{148854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40301, {{148855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40302, {{148858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40303, {{148860, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40304, {{148861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40305, {{148862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40306, {{148863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40307, {{148864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40308, {{148865, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40314, {{148868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40309, {{148869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40310, {{148870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40311, {{148871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40312, {{148872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40313, {{148873, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40316, {{148876, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40317, {{148877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40318, {{148879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40319, {{148880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40320, {{148882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40321, {{148883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40322, {{148885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40323, {{148886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40324, {{148887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40325, {{148888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40326, {{148889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40327, {{148890, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40328, {{148891, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40332, {{148892, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40333, {{148893, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40334, {{148894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40335, {{148896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40336, {{148898, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40337, {{148899, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40338, {{148900, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40339, {{148902, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40340, {{148903, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40341, {{148905, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40342, {{148907, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40343, {{148908, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40344, {{148909, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40345, {{148910, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40346, {{148911, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40347, {{148912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40348, {{148913, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40349, {{148914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40350, {{148915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40351, {{148916, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40352, {{148917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40353, {{148919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40354, {{148922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40355, {{148923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40356, {{148924, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40357, {{148925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40358, {{148926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40359, {{148927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40360, {{148928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40361, {{148929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40362, {{148930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40363, {{148931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40364, {{148932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40365, {{148933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40366, {{148935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40367, {{148936, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40368, {{148937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40369, {{148938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40370, {{148939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40371, {{148940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40372, {{148941, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40373, {{148942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40374, {{148943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40376, {{148947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40375, {{148949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40377, {{148951, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40378, {{148952, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40379, {{148953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40380, {{148954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40381, {{148955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40382, {{148956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40383, {{148958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40384, {{148960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40385, {{148962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40386, {{148963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40387, {{148965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40388, {{148966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40389, {{148968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40390, {{148969, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40391, {{148970, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40392, {{148971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40393, {{148972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40394, {{148973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40395, {{148974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40396, {{148975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40397, {{148977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40398, {{148980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40399, {{148981, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40400, {{148982, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40401, {{148983, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40402, {{148984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40403, {{148986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40404, {{148987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40405, {{148990, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40406, {{148991, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40407, {{148992, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40408, {{148995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40409, {{148996, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40410, {{148997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40411, {{148998, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40412, {{148999, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40413, {{149000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40414, {{149001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40415, {{149002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40416, {{149004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40417, {{149008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40418, {{149009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40419, {{149012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40420, {{149014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40421, {{149016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40422, {{149018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40423, {{149019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40424, {{149020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40425, {{149021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40426, {{149022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40427, {{149024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40428, {{149025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40429, {{149026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40430, {{149027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40431, {{149028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40432, {{149029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40433, {{149030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40434, {{149031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40435, {{149033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40436, {{149035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40437, {{149041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40438, {{149042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40439, {{149043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40440, {{149044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40441, {{149045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40442, {{149046, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40443, {{149047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40444, {{149048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40445, {{149049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40446, {{149050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40447, {{149051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40448, {{149053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40449, {{149054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40450, {{149055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40451, {{149056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40452, {{149061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40453, {{149063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40454, {{149064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40455, {{149066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40456, {{149067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40457, {{149068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40458, {{149069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40459, {{149070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40460, {{149071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40461, {{149073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40462, {{149074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40463, {{149075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40464, {{149076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40465, {{149077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40466, {{149078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40467, {{149079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40468, {{149080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40469, {{149083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40470, {{149084, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40471, {{149086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40472, {{149087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40473, {{149088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40474, {{149089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40475, {{149090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40476, {{149091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40477, {{149092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40478, {{149096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40479, {{149097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40480, {{149098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40481, {{149101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40482, {{149104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40483, {{149105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40484, {{149106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40485, {{149107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40486, {{149108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40487, {{149109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40488, {{149112, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40489, {{149117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40490, {{149118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40491, {{149119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40492, {{149120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40493, {{149121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40494, {{149122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40495, {{149123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40496, {{149124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40497, {{149125, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40498, {{149128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40499, {{149129, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40500, {{149130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40501, {{149132, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40502, {{149133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40503, {{149134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40504, {{149135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40505, {{149137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40506, {{149138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40507, {{149140, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40508, {{149141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40509, {{149142, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40510, {{149146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40511, {{149147, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40512, {{149148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40513, {{149149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40514, {{149150, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40515, {{149151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40516, {{149153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40517, {{149154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40518, {{149155, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40519, {{149156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40520, {{149157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40521, {{149160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40522, {{149161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40523, {{149162, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40524, {{149163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40525, {{149164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40526, {{149165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40527, {{149166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40528, {{149167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40529, {{149169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40530, {{149170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40531, {{149171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40532, {{149172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40533, {{149173, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40534, {{149174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40535, {{149180, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40536, {{149182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40537, {{149183, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40538, {{149184, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40539, {{149185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40540, {{149188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40541, {{149189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40542, {{149194, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40543, {{149195, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40544, {{149196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40545, {{149197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40546, {{149198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40547, {{149199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40548, {{149200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40549, {{149201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40573, {{149202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40550, {{149206, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56818, {{149207, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40551, {{149209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40552, {{149210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40553, {{149211, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40554, {{149212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40555, {{149213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40556, {{149215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40557, {{149217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40558, {{149218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40559, {{149219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40560, {{149220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40561, {{149221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40562, {{149222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40563, {{149223, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40564, {{149225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40565, {{149226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40566, {{149227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40567, {{149229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40568, {{149231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40569, {{149232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40570, {{149233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40571, {{149234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40572, {{149235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40574, {{149236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40575, {{149238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40576, {{149239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40577, {{149240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40578, {{149241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{68082, {{149242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40581, {{149243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40580, {{149244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40582, {{149246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40583, {{149247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40584, {{149250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40585, {{149251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40586, {{149252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40587, {{149253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40588, {{149255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40589, {{149256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40590, {{149257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40591, {{149258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40592, {{149259, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40593, {{149260, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40594, {{149261, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40595, {{149263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40596, {{149264, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40597, {{149265, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40598, {{149266, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40599, {{149267, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40600, {{149268, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40601, {{149269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40602, {{149270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40603, {{149272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40604, {{149273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40605, {{149275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40606, {{149276, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40607, {{149277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40608, {{149281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40609, {{149282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40610, {{149283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40611, {{149284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40612, {{149287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40613, {{149288, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40614, {{149289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40615, {{149291, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40616, {{149294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40617, {{149295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40618, {{149296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40619, {{149297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40620, {{149298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40621, {{149299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40622, {{149300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{40623, {{149301, 1, 0, 0},{149301, 1, 1, 16}, {0, 0, 0, 0}}},{40623, {{149301, 1, 0, 0},{149301, 1, 1, 16}, {0, 0, 0, 0}}},{40624, {{149301, 1, 0, 0},{149301, 1, 1, 17}, {0, 0, 0, 0}}},{40624, {{149301, 1, 0, 0},{149301, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{40625, {{149302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40627, {{149303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40628, {{149304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40629, {{149305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40630, {{149306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40631, {{149307, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40632, {{149308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40633, {{149309, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40634, {{149310, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40635, {{149311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40636, {{149315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40637, {{149316, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40638, {{149317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40639, {{149318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40640, {{149319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40641, {{149320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40642, {{149321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40643, {{149324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40644, {{149325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40645, {{149326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40646, {{149327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40647, {{149328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40648, {{149329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40649, {{149330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40650, {{149331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40651, {{149332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40652, {{149333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40653, {{149334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40654, {{149335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40655, {{149336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40656, {{149337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40657, {{149338, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40658, {{149339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40659, {{149340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40660, {{149341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40661, {{149342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40662, {{149343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40663, {{149344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40664, {{149345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40665, {{149347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40666, {{149348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40667, {{149350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40668, {{149351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40669, {{149352, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40670, {{149353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40671, {{149356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40672, {{149357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40673, {{149358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40674, {{149359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40675, {{149360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40676, {{149361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40677, {{149362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40679, {{149363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40680, {{149365, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40681, {{149366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40682, {{149367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40683, {{149368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40684, {{149369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40685, {{149370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40686, {{149371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40687, {{149372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40688, {{149373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40689, {{149374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40690, {{149375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40691, {{149376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40692, {{149377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40693, {{149378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40694, {{149384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40695, {{149385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40696, {{149386, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40697, {{149387, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40698, {{149388, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40699, {{149389, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40700, {{149390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40701, {{149399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40702, {{149401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40703, {{149402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40704, {{149403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40705, {{149404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40706, {{149405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40707, {{149406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40708, {{149407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40709, {{149408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40710, {{149409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40711, {{149410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40712, {{149411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40713, {{149413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40714, {{149422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40715, {{149423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40716, {{149424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40717, {{149426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40718, {{149427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40719, {{149428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40720, {{149429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40721, {{149432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40722, {{149433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40723, {{149438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40724, {{149439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40725, {{149440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40726, {{149444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40727, {{149445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40728, {{149447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40729, {{149448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40730, {{149449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40731, {{149451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40732, {{149452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40733, {{149453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40734, {{149454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40735, {{149455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40736, {{149456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40737, {{149457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40738, {{149458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40739, {{149459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40740, {{149460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40741, {{149461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40742, {{149462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40743, {{149463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40744, {{149464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40745, {{149465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40746, {{149466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40747, {{149468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40748, {{149469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40749, {{149470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40750, {{149471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40751, {{149472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40752, {{149473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40753, {{149474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40754, {{149475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40755, {{149476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40756, {{149477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40757, {{149478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40758, {{149479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40759, {{149480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40760, {{149481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40761, {{149482, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40762, {{149483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40763, {{149484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40764, {{149485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40765, {{149489, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40766, {{149492, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40767, {{149493, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40768, {{149494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40769, {{149496, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40770, {{149499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40771, {{149500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40772, {{149501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40773, {{149502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40774, {{149503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40775, {{149504, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40776, {{149511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40777, {{149512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40778, {{149516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40779, {{149517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40780, {{149518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40781, {{149519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40782, {{149520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40783, {{149521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40785, {{149522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40786, {{149523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40789, {{149524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40790, {{149525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40791, {{149526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40793, {{149527, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40794, {{149528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40795, {{149529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40796, {{149530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40803, {{149531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40797, {{149532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40798, {{149533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40799, {{149534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40800, {{149535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40801, {{149536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40802, {{149537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40805, {{149541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40806, {{149546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40807, {{149547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40808, {{149548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40809, {{149549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40810, {{149557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40811, {{149558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40812, {{149560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40813, {{149561, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40814, {{149562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40815, {{149563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40816, {{149565, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40817, {{149566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40818, {{149567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40819, {{149568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40820, {{149569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40821, {{149570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40822, {{149571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40823, {{149572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40824, {{149573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40825, {{149574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40826, {{149575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40827, {{149582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40828, {{149583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40829, {{149584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40830, {{149585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40831, {{149586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40832, {{149587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40833, {{149589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40834, {{149590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40835, {{149591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40836, {{149592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40837, {{149593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40838, {{149596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40839, {{149597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40840, {{149598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40841, {{149599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40842, {{149600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40843, {{149601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40844, {{149602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40845, {{149603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40846, {{149604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40847, {{149605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40848, {{149606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40849, {{149607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40850, {{149609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40851, {{149610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40852, {{149611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40853, {{149612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40854, {{149613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40855, {{149614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40856, {{149616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40857, {{149617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40858, {{149618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40862, {{149619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40859, {{149620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40860, {{149621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40861, {{149622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40863, {{149630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40864, {{149631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40865, {{149632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40866, {{149633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40867, {{149635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40868, {{149636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40869, {{149637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40870, {{149642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40871, {{149643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40872, {{149644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40873, {{149645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40874, {{149646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40875, {{149647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40876, {{149648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40877, {{149649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40878, {{149650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40879, {{149651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40880, {{149652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40881, {{149654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40882, {{149655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40883, {{149656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40884, {{149657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40885, {{149658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40886, {{149660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40887, {{149661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40888, {{149662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40889, {{149664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40890, {{149666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40891, {{149667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40892, {{149670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40893, {{149671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40894, {{149672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40895, {{149673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40896, {{149674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40897, {{149678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40898, {{149679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40899, {{149684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40900, {{149685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40901, {{149686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40902, {{149687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40903, {{149688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40904, {{149689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40905, {{149690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40906, {{149691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40907, {{149692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40908, {{149699, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40911, {{149701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40910, {{149702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40912, {{149706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40913, {{149707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40914, {{149708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40915, {{149709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40916, {{149710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40917, {{149713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40918, {{149714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40919, {{149715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40920, {{149716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40921, {{149720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40922, {{149721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40923, {{149723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40924, {{149724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40925, {{149725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40926, {{149727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40927, {{149728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40928, {{149731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40929, {{149732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40930, {{149735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40931, {{149737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40932, {{149738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40933, {{149739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40934, {{149740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40938, {{149741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40936, {{149743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40937, {{149744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40939, {{149748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40940, {{149749, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40941, {{149750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40942, {{149752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40943, {{149753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40944, {{149754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40945, {{149755, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40946, {{149756, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{17147, {{149757, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40947, {{149768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40948, {{149769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40949, {{149770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40950, {{149771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40951, {{149773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40952, {{149775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40953, {{149776, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40954, {{149777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40955, {{149778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40956, {{149779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40957, {{149784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58040, {{149787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40959, {{149798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40960, {{149799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40961, {{149800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40962, {{149801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40963, {{149802, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40964, {{149803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40965, {{149804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40966, {{149805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40967, {{149807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40969, {{149809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40970, {{149810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40971, {{149814, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40972, {{149815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40973, {{149816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40974, {{149817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40975, {{149818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40976, {{149819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40977, {{149820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58042, {{149821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40979, {{149838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40980, {{149842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40981, {{149843, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40982, {{149844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40983, {{149845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40984, {{149846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40985, {{149847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40986, {{149849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40987, {{149850, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40988, {{149851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40989, {{149852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40990, {{149853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40991, {{149854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40992, {{149870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40993, {{149871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40994, {{149872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40995, {{149879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40996, {{149880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40999, {{149884, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41000, {{149885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41001, {{149886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41002, {{149887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41003, {{149888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41004, {{149891, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41005, {{149892, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41006, {{149906, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59900, {{149915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41007, {{149920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41008, {{149921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41009, {{149924, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41010, {{149925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41011, {{149926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41012, {{149928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41014, {{149934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41015, {{149935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41016, {{149936, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41017, {{149937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41018, {{149938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41019, {{149939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41020, {{149940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58046, {{149945, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41022, {{149948, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41024, {{149974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41025, {{149975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41026, {{149976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41027, {{149977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41029, {{149984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41030, {{149986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41031, {{149991, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41032, {{149994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41033, {{149995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41035, {{149997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41036, {{149998, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41037, {{149999, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41038, {{150000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41039, {{150001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41040, {{150002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41041, {{150010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41044, {{150029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41045, {{150030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41046, {{150031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41047, {{150032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41048, {{150036, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41049, {{150037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41050, {{150038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41051, {{150040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41052, {{150041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41053, {{150043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41054, {{150044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41055, {{150045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41056, {{150061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41057, {{150072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41058, {{150073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41059, {{150074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41060, {{150075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41061, {{150080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41062, {{150081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41063, {{150082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41064, {{150083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41065, {{150084, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41066, {{150085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41067, {{150086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41068, {{150087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41069, {{150088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41070, {{150089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41071, {{150090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41072, {{150091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41073, {{150093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41074, {{150098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41076, {{150111, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41077, {{150113, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41078, {{150115, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41079, {{150116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41080, {{150120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41081, {{150121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41082, {{150123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41083, {{150124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41084, {{150125, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41085, {{150126, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41086, {{150127, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41087, {{150128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41088, {{150129, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41089, {{150130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41090, {{150131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41091, {{150132, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41092, {{150133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41093, {{150134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41094, {{150147, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41095, {{150148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41096, {{150149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41097, {{150150, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41098, {{150151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41099, {{150156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41100, {{150157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41101, {{150158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41102, {{150159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41103, {{150160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41104, {{150177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41105, {{150178, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41106, {{150179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58051, {{150182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41107, {{150185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41108, {{150186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41109, {{150187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41110, {{150188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41111, {{150190, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41113, {{150196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41114, {{150201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41115, {{150207, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41116, {{150208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41117, {{150213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41118, {{150214, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41119, {{150217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41120, {{150218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41121, {{150219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41122, {{150220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41123, {{150221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41124, {{150228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41125, {{150229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41126, {{150232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41127, {{150237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41128, {{150238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41129, {{150243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41130, {{150244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41131, {{150245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41132, {{150247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41133, {{150248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41134, {{150249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41135, {{150250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41136, {{150251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41137, {{150252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41138, {{150253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41139, {{150255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41140, {{150256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41141, {{150257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41142, {{150258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41143, {{150259, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41144, {{150260, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41145, {{150261, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41146, {{150262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41147, {{150263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41148, {{150265, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41149, {{150266, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41150, {{150267, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41151, {{150268, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41152, {{150270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41153, {{150273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41154, {{150274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41155, {{150275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41156, {{150279, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41157, {{150280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41158, {{150281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41159, {{150284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41160, {{150285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41161, {{150286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41162, {{150287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41163, {{150288, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41164, {{150290, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41165, {{150293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41166, {{150294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41167, {{150295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41168, {{150296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41169, {{150298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41170, {{150299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41171, {{150300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41172, {{150301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41173, {{150304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41174, {{150305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41175, {{150306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41176, {{150307, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41177, {{150308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41178, {{150309, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41179, {{150310, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41180, {{150311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41181, {{150312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41182, {{150313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41183, {{150314, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41184, {{150315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41185, {{150316, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41186, {{150317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41187, {{150318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41188, {{150319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41189, {{150320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41190, {{150321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41191, {{150323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41192, {{150325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41193, {{150326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41194, {{150327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41195, {{150330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41196, {{150331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41197, {{150332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41199, {{150335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41200, {{150336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41201, {{150337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41202, {{150338, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41203, {{150339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41204, {{150340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41205, {{150348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41206, {{150349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41207, {{150350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41208, {{150351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41209, {{150352, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41210, {{150353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41211, {{150354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41212, {{150355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41213, {{150356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41214, {{150357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41215, {{150358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41216, {{150359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41217, {{150360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41218, {{150363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41219, {{150364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41220, {{150365, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41221, {{150366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41222, {{150367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41223, {{150368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41224, {{150369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41225, {{150370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41226, {{150371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41227, {{150374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41228, {{150375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41229, {{150376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41230, {{150377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41231, {{150378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41232, {{150379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41233, {{150380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41234, {{150383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41235, {{150384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41236, {{150385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41237, {{150387, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41238, {{150388, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41239, {{150393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41240, {{150394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41241, {{150395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41242, {{150396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41243, {{150397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41244, {{150400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41245, {{150405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41246, {{150406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41247, {{150407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41248, {{150408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41249, {{150409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41250, {{150410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41251, {{150411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41252, {{150412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41253, {{150413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41254, {{150414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41255, {{150415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41256, {{150416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41257, {{150419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41258, {{150420, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41259, {{150421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41260, {{150422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41261, {{150424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41262, {{150425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41263, {{150426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41264, {{150427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41265, {{150428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41266, {{150433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41267, {{150434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41268, {{150438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41269, {{150439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41270, {{150440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41271, {{150441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41272, {{150442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41273, {{150443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41274, {{150445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41275, {{150446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41276, {{150447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41277, {{150448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41278, {{150449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41279, {{150450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41280, {{150451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41281, {{150453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41282, {{150455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41283, {{150456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41284, {{150457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41285, {{150458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41286, {{150459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41287, {{150460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41288, {{150461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41289, {{150465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41290, {{150466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41291, {{150468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41292, {{150469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41293, {{150471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41294, {{150472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41295, {{150473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41296, {{150475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41297, {{150476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41298, {{150477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41299, {{150478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41300, {{150479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41301, {{150480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41302, {{150483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41303, {{150484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41304, {{150485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41305, {{150486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41306, {{150487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41307, {{150490, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41308, {{150491, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41309, {{150492, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41310, {{150493, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41311, {{150495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41312, {{150496, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41313, {{150497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41314, {{150498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41315, {{150499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41316, {{150502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41317, {{150503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41318, {{150508, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41319, {{150511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41320, {{150514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41321, {{150515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41322, {{150519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41323, {{150523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41324, {{150525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {2, {{41325, {{150528, 1, 0, 0}, {0, 0, 0, 0}}},{41326, {{150528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41327, {{150530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41328, {{150531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41329, {{150533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41330, {{150534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41331, {{150535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41332, {{150536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41333, {{150539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41334, {{150540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41335, {{150542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41336, {{150543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41337, {{150545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41338, {{150547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41339, {{150549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41340, {{150550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41341, {{150555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41342, {{150557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{41343, {{150558, 1, 0, 0},{150558, 1, 1, 18}, {0, 0, 0, 0}}},{41343, {{150558, 1, 0, 0},{150558, 1, 1, 18}, {0, 0, 0, 0}}},{41344, {{150558, 1, 0, 0},{150558, 1, 1, 19}, {0, 0, 0, 0}}},{41344, {{150558, 1, 0, 0},{150558, 1, 1, 19}, {0, 0, 0, 0}}},{57093, {{150558, 1, 0, 0},{150558, 1, 1, 16}, {0, 0, 0, 0}}},{57093, {{150558, 1, 0, 0},{150558, 1, 1, 16}, {0, 0, 0, 0}}},{57095, {{150558, 1, 0, 0},{150558, 1, 1, 17}, {0, 0, 0, 0}}},{57095, {{150558, 1, 0, 0},{150558, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{41345, {{150559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41346, {{150560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41347, {{150561, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41348, {{150562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41349, {{150563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41350, {{150566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41351, {{150567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41352, {{150568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41353, {{150569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41354, {{150570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41355, {{150571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41356, {{150572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41357, {{150573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41358, {{150574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41359, {{150576, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41360, {{150577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41361, {{150578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41362, {{150579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41363, {{150580, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41364, {{150581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41365, {{150582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41367, {{150595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41368, {{150596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41369, {{150597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41370, {{150598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41371, {{150599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58088, {{150602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41372, {{150603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41373, {{150604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41374, {{150605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41375, {{150606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41376, {{150607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41377, {{150610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41366, {{150611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41378, {{150614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41379, {{150619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41380, {{150620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41381, {{150622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41382, {{150623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41383, {{150624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41384, {{150625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41385, {{150627, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41386, {{150628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41387, {{150629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41388, {{150630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41390, {{150632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41391, {{150633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41392, {{150634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{41393, {{150635, 1, 0, 0},{150635, 1, 1, 16}, {0, 0, 0, 0}}},{41393, {{150635, 1, 0, 0},{150635, 1, 1, 16}, {0, 0, 0, 0}}},{68083, {{150635, 1, 0, 0},{150635, 1, 1, 17}, {0, 0, 0, 0}}},{68083, {{150635, 1, 0, 0},{150635, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{41394, {{150636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41395, {{150637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41396, {{150638, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41397, {{150641, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41398, {{150642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41399, {{150643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41400, {{150644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41401, {{150645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41402, {{150647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41403, {{150648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41404, {{150650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41405, {{150652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41406, {{150653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41407, {{150654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{41408, {{150659, 1, 0, 0},{150659, 1, 1, 16}, {0, 0, 0, 0}}},{41408, {{150659, 1, 0, 0},{150659, 1, 1, 16}, {0, 0, 0, 0}}},{41409, {{150659, 1, 0, 0},{150659, 1, 1, 17}, {0, 0, 0, 0}}},{41409, {{150659, 1, 0, 0},{150659, 1, 1, 17}, {0, 0, 0, 0}}},{58099, {{150659, 1, 0, 0},{150659, 1, 1, 18}, {0, 0, 0, 0}}},{58099, {{150659, 1, 0, 0},{150659, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{41410, {{150660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41412, {{150665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41413, {{150666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41414, {{150667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41415, {{150669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41416, {{150673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41417, {{150674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41418, {{150675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41419, {{150676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41420, {{150677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41421, {{150678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41422, {{150679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41423, {{150680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41424, {{150681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41425, {{150682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41426, {{150683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41427, {{150684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41428, {{150685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41429, {{150690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41430, {{150691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41431, {{150692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41432, {{150694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41433, {{150695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41434, {{150696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41435, {{150697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41436, {{150699, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41437, {{150700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41438, {{150701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41439, {{150702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41440, {{150703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41441, {{150704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41442, {{150706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41443, {{150709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41444, {{150710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41445, {{150711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41446, {{150712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41447, {{150713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41448, {{150714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41449, {{150715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41450, {{150717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41451, {{150718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41452, {{150719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41453, {{150721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41454, {{150722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41455, {{150725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41456, {{150728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41457, {{150729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41458, {{150730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41459, {{150731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41460, {{150732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41461, {{150734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41462, {{150735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41463, {{150738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41464, {{150739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41465, {{150740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41466, {{150741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41467, {{150742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41468, {{150743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {2, {{41469, {{150744, 1, 0, 0}, {0, 0, 0, 0}}},{41470, {{150744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41471, {{150747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41472, {{150749, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41473, {{150751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41474, {{150752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41475, {{150756, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41476, {{150757, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41477, {{150761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41478, {{150762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41479, {{150763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41480, {{150764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41481, {{150765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41482, {{150766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41483, {{150767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41484, {{150771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41485, {{150772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41486, {{150773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41487, {{150774, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41488, {{150777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41489, {{150779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41490, {{150780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{41491, {{150783, 1, 0, 0},{150783, 1, 1, 16}, {0, 0, 0, 0}}},{41491, {{150783, 1, 0, 0},{150783, 1, 1, 16}, {0, 0, 0, 0}}},{68084, {{150783, 1, 0, 0},{150783, 1, 1, 17}, {0, 0, 0, 0}}},{68084, {{150783, 1, 0, 0},{150783, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{41492, {{150784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41493, {{150786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41494, {{150788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41495, {{150789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41496, {{150791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41497, {{150792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41498, {{150793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41499, {{150795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41500, {{150796, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41501, {{150798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41502, {{150800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41503, {{150801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41504, {{150802, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41505, {{150803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41506, {{150804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41507, {{150805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41508, {{150807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41509, {{150808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41510, {{150809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41511, {{150812, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41512, {{150813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41513, {{150814, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41514, {{150815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{41515, {{150817, 1, 0, 0},{150817, 1, 1, 16}, {0, 0, 0, 0}}},{41515, {{150817, 1, 0, 0},{150817, 1, 1, 16}, {0, 0, 0, 0}}},{58117, {{150817, 1, 0, 0},{150817, 1, 1, 17}, {0, 0, 0, 0}}},{58117, {{150817, 1, 0, 0},{150817, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{41516, {{150818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41517, {{150819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41518, {{150820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41519, {{150821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41520, {{150822, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41521, {{150823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41522, {{150824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41523, {{150825, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41524, {{150826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41525, {{150827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41526, {{150830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41527, {{150831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41528, {{150832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41529, {{150833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41530, {{150836, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{17747, {{150840, 1, 0, 0},{150840, 1, 1, 17}, {0, 0, 0, 0}}},{17747, {{150840, 1, 0, 0},{150840, 1, 1, 17}, {0, 0, 0, 0}}},{41531, {{150840, 1, 0, 0},{150840, 1, 1, 16}, {0, 0, 0, 0}}},{41531, {{150840, 1, 0, 0},{150840, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41532, {{150843, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41533, {{150844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41534, {{150845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41535, {{150846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41536, {{150847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41537, {{150848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41538, {{150849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41539, {{150850, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41540, {{150851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41541, {{150855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41542, {{150858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41543, {{150859, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41544, {{150861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41545, {{150862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41546, {{150865, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41547, {{150866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41548, {{150868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41549, {{150871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41550, {{150872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41551, {{150873, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41552, {{150874, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41553, {{150875, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41554, {{150876, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41555, {{150877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41556, {{150878, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41557, {{150879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41558, {{150880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41559, {{150881, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41560, {{150882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41561, {{150883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41562, {{150884, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41563, {{150885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41564, {{150886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41565, {{150887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41567, {{150891, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41568, {{150892, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41569, {{150894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41570, {{150895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41571, {{150896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41566, {{150901, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41572, {{150903, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41573, {{150904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41574, {{150905, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41575, {{150906, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41576, {{150907, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41577, {{150908, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41578, {{150909, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41579, {{150910, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41580, {{150911, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41581, {{150912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41582, {{150913, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41583, {{150917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41584, {{150918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41585, {{150920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41586, {{150921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41587, {{150923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41588, {{150924, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41589, {{150925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41590, {{150931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41591, {{150932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41592, {{150933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41593, {{150934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41594, {{150935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41595, {{150936, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41596, {{150937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41597, {{150938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41598, {{150939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41599, {{150940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41600, {{150941, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41601, {{150944, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41602, {{150945, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41603, {{150946, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41604, {{150947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41605, {{150948, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41606, {{150949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41607, {{150952, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41608, {{150953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41609, {{150954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41610, {{150955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41611, {{150958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41612, {{150959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41613, {{150960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41614, {{150962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41615, {{150963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41616, {{150964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41617, {{150965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41618, {{150966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41619, {{150967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41620, {{150976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41621, {{150977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41622, {{150978, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41623, {{150979, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41624, {{150980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41625, {{150981, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41626, {{150982, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41627, {{150983, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41628, {{150984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41629, {{150985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41630, {{150986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41631, {{150988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41632, {{150989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41633, {{150990, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41634, {{150991, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41635, {{150992, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41636, {{150993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41637, {{150995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41638, {{150996, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41639, {{150997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41640, {{150998, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41641, {{150999, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41642, {{151000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41643, {{151001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41644, {{151002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41645, {{151003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41646, {{151004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41647, {{151010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41648, {{151012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41649, {{151013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41650, {{151014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41651, {{151015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41652, {{151016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41653, {{151031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41654, {{151032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41655, {{151033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41656, {{151034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41657, {{151035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41658, {{151036, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41659, {{151037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41660, {{151038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41661, {{151039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41662, {{151040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41663, {{151041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41664, {{151042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41665, {{151043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41666, {{151044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41667, {{151045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41668, {{151046, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41669, {{151047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41670, {{151048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41671, {{151049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41672, {{151050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41673, {{151051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41674, {{151052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41675, {{151054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41676, {{151055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41677, {{151057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41678, {{151058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41679, {{151059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41680, {{151060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41681, {{151061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41682, {{151062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41683, {{151063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41684, {{151065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41685, {{151070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41686, {{151071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41687, {{151072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41688, {{151073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41689, {{151074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41690, {{151076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41691, {{151078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41692, {{151081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41693, {{151082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41694, {{151083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41695, {{151084, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41696, {{151085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41697, {{151086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41698, {{151087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41699, {{151088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41700, {{151089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41701, {{151090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41702, {{151091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41703, {{151092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41704, {{151093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41705, {{151094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41706, {{151095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41707, {{151096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41708, {{151097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41709, {{151098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41710, {{151099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41711, {{151100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41712, {{151101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41713, {{151102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41714, {{151103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41715, {{151111, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41716, {{151112, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41717, {{151113, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41718, {{151114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41719, {{151115, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41720, {{151116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41721, {{151117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41722, {{151118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41723, {{151119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41724, {{151131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41725, {{151132, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41726, {{151133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41727, {{151134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41728, {{151135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41729, {{151136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41730, {{151137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41731, {{151138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41732, {{151139, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41733, {{151140, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41734, {{151141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41735, {{151142, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41736, {{151143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41737, {{151144, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41738, {{151145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41739, {{151146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41740, {{151148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41741, {{151153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41742, {{151154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41743, {{151155, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41744, {{151156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41745, {{151157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41746, {{151158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41747, {{151159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41748, {{151160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41749, {{151161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41750, {{151163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41751, {{151167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41752, {{151168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41753, {{151169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41754, {{151171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{41755, {{151172, 1, 0, 0},{151172, 1, 1, 16}, {0, 0, 0, 0}}},{41755, {{151172, 1, 0, 0},{151172, 1, 1, 16}, {0, 0, 0, 0}}},{41756, {{151172, 1, 0, 0},{151172, 1, 1, 17}, {0, 0, 0, 0}}},{41756, {{151172, 1, 0, 0},{151172, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{41757, {{151173, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41758, {{151174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41759, {{151176, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41760, {{151177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41761, {{151178, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41762, {{151179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41763, {{151180, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41764, {{151181, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41765, {{151182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41766, {{151183, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41767, {{151184, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41768, {{151185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41769, {{151186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41770, {{151187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41771, {{151188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41772, {{151189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41773, {{151190, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41774, {{151191, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41775, {{151192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41776, {{151193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41777, {{151194, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41778, {{151195, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41779, {{151196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41780, {{151197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41781, {{151206, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41782, {{151207, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41783, {{151208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41784, {{151209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41785, {{151211, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41786, {{151223, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41787, {{151224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41788, {{151226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41789, {{151227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41790, {{151228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41791, {{151229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41792, {{151230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41793, {{151231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41794, {{151232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41795, {{151233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41796, {{151234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41797, {{151235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41798, {{151236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41799, {{151237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41800, {{151238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41801, {{151239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41802, {{151240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41803, {{151241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41804, {{151244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41805, {{151245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41806, {{151253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41807, {{151254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41808, {{151255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41809, {{151256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41810, {{151257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41811, {{151258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41812, {{151259, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41813, {{151260, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41814, {{151261, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41815, {{151262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41816, {{151266, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41817, {{151267, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41818, {{151269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41819, {{151270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41820, {{151272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41821, {{151273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41822, {{151278, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41823, {{151279, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41824, {{151281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41825, {{151282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41826, {{151283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41827, {{151284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41828, {{151285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41829, {{151286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41830, {{151287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41831, {{151288, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41832, {{151289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41833, {{151290, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41834, {{151296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41835, {{151297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41836, {{151298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41837, {{151299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41838, {{151300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41839, {{151301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41840, {{151302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41841, {{151307, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41842, {{151308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41843, {{151309, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41844, {{151312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41845, {{151313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41846, {{151314, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41847, {{151318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41848, {{151319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41849, {{151320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41850, {{151321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41851, {{151322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41852, {{151323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41853, {{151324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41854, {{151325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41855, {{151327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41856, {{151328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41857, {{151329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41858, {{151330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41859, {{151331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41860, {{151332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41861, {{151333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41862, {{151334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41863, {{151335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41864, {{151337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41865, {{151338, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41867, {{151340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41868, {{151342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41869, {{151345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41871, {{151346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41872, {{151347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41873, {{151348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41874, {{151350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41875, {{151351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41876, {{151353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41877, {{151354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41878, {{151356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41879, {{151357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41880, {{151358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41881, {{151359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41882, {{151361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41883, {{151362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41884, {{151363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41885, {{151365, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41886, {{151366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41887, {{151367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41888, {{151369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41889, {{151370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41890, {{151371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41891, {{151372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41893, {{151373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41894, {{151374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41895, {{151376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41896, {{151379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41897, {{151381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41898, {{151382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41899, {{151384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41900, {{151385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41901, {{151386, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41902, {{151387, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41903, {{151394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41904, {{151395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41905, {{151396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41906, {{151397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41907, {{151398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41908, {{151399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41909, {{151403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41910, {{151404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41911, {{151408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41912, {{151409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41913, {{151410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41914, {{151411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41915, {{151417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41916, {{151418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41917, {{151419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41918, {{151420, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41919, {{151421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41920, {{151422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41921, {{151423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41922, {{151428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41923, {{151429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41924, {{151430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41925, {{151431, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41926, {{151432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41927, {{151433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41928, {{151434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41929, {{151440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41930, {{151441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41931, {{151442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41932, {{151445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41934, {{151449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41935, {{151451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41936, {{151453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41937, {{151454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41938, {{151456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41939, {{151457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41940, {{151458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41941, {{151459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41942, {{151460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41943, {{151461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41944, {{151462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41945, {{151463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41946, {{151464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41947, {{151466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41948, {{151467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41949, {{151468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41950, {{151469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41951, {{151470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41952, {{151474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41953, {{151476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41955, {{151477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41956, {{151478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41957, {{151480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41958, {{151482, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41959, {{151483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41960, {{151484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41961, {{151486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41962, {{151487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41963, {{151488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41964, {{151491, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41965, {{151492, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41966, {{151493, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41967, {{151494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41968, {{151495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41969, {{151496, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41970, {{151497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41971, {{151498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41972, {{151499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41973, {{151500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41974, {{151502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41975, {{151503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41976, {{151504, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41977, {{151505, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41978, {{151506, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41979, {{151507, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41980, {{151508, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41981, {{151509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41982, {{151510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41983, {{151511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41984, {{151512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41985, {{151513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41986, {{151514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41987, {{151515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41988, {{151516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41990, {{151519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41991, {{151520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41992, {{151521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41993, {{151523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41994, {{151524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41995, {{151525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41996, {{151527, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41997, {{151528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41998, {{151529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{41999, {{151530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42000, {{151531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41989, {{151533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42001, {{151534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42002, {{151535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42003, {{151539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42004, {{151540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42005, {{151541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42006, {{151542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42007, {{151543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42008, {{151547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42009, {{151548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42010, {{151549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42011, {{151550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42012, {{151551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42013, {{151553, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42014, {{151554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42015, {{151555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42016, {{151556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42017, {{151560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42018, {{151561, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42019, {{151562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42020, {{151565, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42021, {{151566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42022, {{151567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42023, {{151568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42024, {{151571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42025, {{151572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42026, {{151573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42027, {{151574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42028, {{151575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42029, {{151576, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42030, {{151577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42031, {{151579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42032, {{151581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42033, {{151582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42034, {{151583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42035, {{151584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42036, {{151586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42037, {{151587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42038, {{151588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42039, {{151589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42040, {{151590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42041, {{151591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42042, {{151592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42043, {{151593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42044, {{151595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42045, {{151596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42046, {{151597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42047, {{151598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42048, {{151599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42049, {{151600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42050, {{151601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42051, {{151602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42052, {{151604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42053, {{151605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42054, {{151608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42055, {{151609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42056, {{151610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42057, {{151612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42058, {{151613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42059, {{151614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42060, {{151617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42061, {{151618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42062, {{151619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{42063, {{151620, 1, 0, 0},{151620, 1, 1, 16}, {0, 0, 0, 0}}},{42063, {{151620, 1, 0, 0},{151620, 1, 1, 16}, {0, 0, 0, 0}}},{42064, {{151620, 1, 0, 0},{151620, 1, 1, 17}, {0, 0, 0, 0}}},{42064, {{151620, 1, 0, 0},{151620, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{42065, {{151621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42066, {{151622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42067, {{151623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42068, {{151624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42069, {{151625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42070, {{151626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18074, {{151627, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42071, {{151630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42072, {{151631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42073, {{151632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42074, {{151633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42075, {{151634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42076, {{151635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{68085, {{151636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42078, {{151637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42079, {{151638, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42080, {{151642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42081, {{151643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42082, {{151644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42083, {{151646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42084, {{151647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42085, {{151648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42086, {{151649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42087, {{151657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42088, {{151658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42089, {{151659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42090, {{151660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42091, {{151661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42092, {{151662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42093, {{151663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42094, {{151664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42095, {{151665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42096, {{151666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42097, {{151670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42098, {{151671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42099, {{151672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42100, {{151673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42101, {{151674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42102, {{151675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42103, {{151676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42104, {{151680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42105, {{151682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42106, {{151684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42108, {{151687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42109, {{151688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42110, {{151690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42111, {{151691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42112, {{151698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42113, {{151699, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42114, {{151700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42115, {{151702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42116, {{151703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42117, {{151704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42118, {{151705, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42119, {{151712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42120, {{151713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42121, {{151715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42122, {{151716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42123, {{151717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42124, {{151718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42125, {{151719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42126, {{151720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42127, {{151721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42128, {{151722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42129, {{151724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42130, {{151727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42131, {{151729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42132, {{151730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42133, {{151731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42134, {{151733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42135, {{151734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42136, {{151736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42137, {{151737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42138, {{151738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42139, {{151740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42140, {{151742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42141, {{151744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42142, {{151746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42143, {{151747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42144, {{151751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42145, {{151752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42146, {{151753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42147, {{151756, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42148, {{151757, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42149, {{151758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42150, {{151763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42151, {{151764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42152, {{151765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42153, {{151766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42154, {{151768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42155, {{151769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42156, {{151772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42157, {{151775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42158, {{151776, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42159, {{151777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42160, {{151779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42161, {{151780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42162, {{151782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42163, {{151783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42164, {{151784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42165, {{151785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42166, {{151786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42167, {{151787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42168, {{151790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42169, {{151792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{42170, {{151794, 1, 0, 0},{151794, 1, 1, 16}, {0, 0, 0, 0}}},{42170, {{151794, 1, 0, 0},{151794, 1, 1, 16}, {0, 0, 0, 0}}},{42171, {{151794, 1, 0, 0},{151794, 1, 1, 17}, {0, 0, 0, 0}}},{42171, {{151794, 1, 0, 0},{151794, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{42172, {{151795, 1, 0, 0},{151795, 1, 1, 16}, {0, 0, 0, 0}}},{42172, {{151795, 1, 0, 0},{151795, 1, 1, 16}, {0, 0, 0, 0}}},{42173, {{151795, 1, 0, 0},{151795, 1, 1, 17}, {0, 0, 0, 0}}},{42173, {{151795, 1, 0, 0},{151795, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{42174, {{151796, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42175, {{151797, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42176, {{151798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42178, {{151801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42179, {{151802, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42180, {{151803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42181, {{151804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42182, {{151805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42183, {{151806, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42184, {{151807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42185, {{151808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42186, {{151809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{42187, {{151810, 1, 0, 0},{151810, 1, 1, 17}, {0, 0, 0, 0}}},{42187, {{151810, 1, 0, 0},{151810, 1, 1, 17}, {0, 0, 0, 0}}},{42188, {{151810, 1, 0, 0},{151810, 1, 1, 18}, {0, 0, 0, 0}}},{42188, {{151810, 1, 0, 0},{151810, 1, 1, 18}, {0, 0, 0, 0}}},{42189, {{151810, 1, 0, 0},{151810, 1, 1, 16}, {0, 0, 0, 0}}},{42189, {{151810, 1, 0, 0},{151810, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42190, {{151814, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42191, {{151815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42192, {{151816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42193, {{151817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42194, {{151818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42195, {{151819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42196, {{151820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42197, {{151821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42198, {{151822, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42199, {{151823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42200, {{151824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42201, {{151825, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42202, {{151826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42203, {{151827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42204, {{151828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42205, {{151829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42206, {{151830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42207, {{151831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42208, {{151832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42209, {{151833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42210, {{151834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42211, {{151835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42212, {{151836, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42213, {{151837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42214, {{151838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42215, {{151842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42216, {{151845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42217, {{151846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42218, {{151847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42219, {{151849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42220, {{151855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42221, {{151857, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42222, {{151858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18158, {{151859, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42223, {{151860, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42224, {{151861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42225, {{151862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42226, {{151863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42227, {{151864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42228, {{151865, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42229, {{151866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42230, {{151867, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42231, {{151868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42232, {{151869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42233, {{151870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42234, {{151871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42235, {{151872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42236, {{151873, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42237, {{151874, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42238, {{151876, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42240, {{151887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42241, {{151889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42242, {{151890, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42243, {{151891, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42244, {{151892, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42239, {{151893, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42245, {{151894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42246, {{151895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42247, {{151896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42248, {{151897, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42249, {{151898, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42250, {{151899, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42251, {{151900, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42252, {{151901, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42253, {{151902, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42254, {{151903, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42255, {{151904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42256, {{151907, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42257, {{151910, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42258, {{151911, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42259, {{151912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42260, {{151914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42261, {{151915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42262, {{151916, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42263, {{151917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42264, {{151918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42265, {{151920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42266, {{151921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42267, {{151923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42268, {{151925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42269, {{151926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42270, {{151927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42271, {{151928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42272, {{151929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42273, {{151930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42274, {{151932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42275, {{151941, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42277, {{151942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42278, {{151943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42279, {{151944, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42280, {{151945, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42281, {{151947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42282, {{151948, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42283, {{151949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42284, {{151950, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42285, {{151951, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42286, {{151952, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42287, {{151953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{42288, {{151954, 1, 0, 0},{151954, 1, 1, 16}, {0, 0, 0, 0}}},{42288, {{151954, 1, 0, 0},{151954, 1, 1, 16}, {0, 0, 0, 0}}},{42289, {{151954, 1, 0, 0},{151954, 1, 1, 17}, {0, 0, 0, 0}}},{42289, {{151954, 1, 0, 0},{151954, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{42290, {{151955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42291, {{151956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42292, {{151958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42293, {{151959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42294, {{151960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42295, {{151961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42296, {{151962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42297, {{151963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42298, {{151964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42299, {{151965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42300, {{151966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42301, {{151967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42302, {{151968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42303, {{151969, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42304, {{151970, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42305, {{151971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42306, {{151972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42307, {{151973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42308, {{151974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42309, {{151977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42310, {{151983, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42311, {{151984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42312, {{151986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42313, {{151987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42314, {{151989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42315, {{151990, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42316, {{151991, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42317, {{152000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42318, {{152004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42319, {{152005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42320, {{152006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42321, {{152007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42322, {{152009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42323, {{152010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42324, {{152011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42325, {{152012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42326, {{152013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42327, {{152014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42328, {{152015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42329, {{152016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42330, {{152017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42331, {{152018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42332, {{152019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42333, {{152020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42334, {{152021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42335, {{152022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42346, {{152023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42336, {{152024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42337, {{152025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42338, {{152026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42339, {{152027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42340, {{152028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42341, {{152029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42342, {{152030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42343, {{152031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42344, {{152032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42345, {{152033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42347, {{152037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42348, {{152038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42349, {{152043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42350, {{152045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42351, {{152046, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42352, {{152047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42353, {{152050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42354, {{152051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42355, {{152052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42356, {{152053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42357, {{152062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42358, {{152063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42359, {{152064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42360, {{152066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42361, {{152067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42107, {{152068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42362, {{152069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42363, {{152070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42364, {{152071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42365, {{152072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42366, {{152073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42367, {{152074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42368, {{152075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42369, {{152078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42370, {{152080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42371, {{152081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42372, {{152082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42373, {{152083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42374, {{152084, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42375, {{152085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42376, {{152086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42377, {{152087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42378, {{152088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42379, {{152089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42380, {{152090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42381, {{152091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42382, {{152092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42383, {{152094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58175, {{152095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42384, {{152096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42386, {{152105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42387, {{152106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42388, {{152107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42389, {{152119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42390, {{152120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42391, {{152123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42392, {{152124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42393, {{152125, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42394, {{152126, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42395, {{152127, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42396, {{152128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42397, {{152129, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42398, {{152130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42399, {{152131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42400, {{152132, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42401, {{152133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42402, {{152134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42403, {{152135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42404, {{152136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42405, {{152137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42406, {{152138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42407, {{152139, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{42408, {{152140, 1, 0, 0},{152140, 1, 1, 17}, {0, 0, 0, 0}}},{42408, {{152140, 1, 0, 0},{152140, 1, 1, 17}, {0, 0, 0, 0}}},{42409, {{152140, 1, 0, 0},{152140, 1, 1, 18}, {0, 0, 0, 0}}},{42409, {{152140, 1, 0, 0},{152140, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{42410, {{152141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{42411, {{152143, 1, 0, 0},{152143, 1, 1, 16}, {0, 0, 0, 0}}},{42411, {{152143, 1, 0, 0},{152143, 1, 1, 16}, {0, 0, 0, 0}}},{58180, {{152143, 1, 0, 0},{152143, 1, 1, 17}, {0, 0, 0, 0}}},{58180, {{152143, 1, 0, 0},{152143, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{42412, {{152144, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42413, {{152145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42414, {{152146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42415, {{152147, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42416, {{152148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42417, {{152149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42418, {{152150, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42419, {{152151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42420, {{152152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42421, {{152153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42422, {{152154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42423, {{152155, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42424, {{152156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42425, {{152157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42426, {{152158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42427, {{152159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42428, {{152162, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42429, {{152166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42430, {{152167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42431, {{152176, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{2853, {{152177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42432, {{152180, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42433, {{152181, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42434, {{152182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42435, {{152183, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42436, {{152184, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42437, {{152185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42438, {{152186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42439, {{152187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42440, {{152188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42441, {{152189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42442, {{152190, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42443, {{152191, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42444, {{152192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42445, {{152193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42446, {{152194, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42447, {{152196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42448, {{152197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42449, {{152198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42450, {{152199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42451, {{152200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42452, {{152201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42453, {{152202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42454, {{152208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42455, {{152211, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42456, {{152213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42457, {{152231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42458, {{152232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42459, {{152234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42460, {{152235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42461, {{152236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42462, {{152237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42463, {{152238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42464, {{152239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42465, {{152240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42466, {{152241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42467, {{152242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42468, {{152243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42469, {{152244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42470, {{152246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42471, {{152248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42472, {{152249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42473, {{152250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42474, {{152251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42475, {{152252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42476, {{152253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42477, {{152254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42478, {{152255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42479, {{152256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42480, {{152257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42481, {{152258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42482, {{152259, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42483, {{152260, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42484, {{152268, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42485, {{152269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42486, {{152272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42487, {{152281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42488, {{152282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42489, {{152284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42490, {{152285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42491, {{152286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42492, {{152287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42493, {{152288, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42494, {{152289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42495, {{152290, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42496, {{152291, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42497, {{152292, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42498, {{152293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42499, {{152294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42500, {{152299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42501, {{152300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42502, {{152301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42503, {{152303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42504, {{152304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42505, {{152315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42506, {{152319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42507, {{152320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42508, {{152321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42509, {{152322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42510, {{152323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42511, {{152324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42512, {{152325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42513, {{152326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42514, {{152328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42515, {{152329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42516, {{152330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42517, {{152331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42518, {{152332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42519, {{152333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42520, {{152334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42521, {{152335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42522, {{152336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42523, {{152341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42524, {{152342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42525, {{152343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42526, {{152346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42527, {{152347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42528, {{152348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42529, {{152349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42530, {{152350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42531, {{152351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42532, {{152353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42534, {{152355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42535, {{152360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42536, {{152361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42537, {{152362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42538, {{152365, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42539, {{152366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42540, {{152367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42541, {{152368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42542, {{152369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42543, {{152370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42544, {{152373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42545, {{152374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42546, {{152378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42547, {{152379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42548, {{152380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42549, {{152381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42550, {{152382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42551, {{152383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42552, {{152384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42553, {{152385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42554, {{152386, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42555, {{152387, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42556, {{152388, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42557, {{152393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42558, {{152394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42559, {{152395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42560, {{152396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42561, {{152399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42562, {{152403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42563, {{152404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42564, {{152405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42565, {{152407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42566, {{152410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42567, {{152413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42568, {{152414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42569, {{152415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42570, {{152416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42571, {{152417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42572, {{152418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42573, {{152419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42574, {{152420, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42575, {{152421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42576, {{152423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42577, {{152424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42578, {{152426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42579, {{152427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42580, {{152428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42581, {{152429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42582, {{152430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42583, {{152431, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42584, {{152432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42585, {{152433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42586, {{152434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42587, {{152435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42588, {{152436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42589, {{152437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42590, {{152438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42591, {{152439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42592, {{152440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42593, {{152441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42594, {{152442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42595, {{152443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42596, {{152444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42597, {{152445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42598, {{152446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42599, {{152447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42600, {{152448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42601, {{152450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42602, {{152451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42603, {{152452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42604, {{152453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42605, {{152454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42606, {{152455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42607, {{152456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42608, {{152457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42609, {{152458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42610, {{152459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42611, {{152460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42612, {{152461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42613, {{152462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42614, {{152463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42615, {{152464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42616, {{152465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42617, {{152466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42618, {{152467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42619, {{152468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42620, {{152470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42621, {{152471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42622, {{152472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42623, {{152473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42624, {{152474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42625, {{152475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42626, {{152476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42627, {{152478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42628, {{152479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42629, {{152480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42630, {{152481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42631, {{152482, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42632, {{152483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42633, {{152484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42634, {{152485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42635, {{152486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42636, {{152487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42637, {{152488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42638, {{152489, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42639, {{152490, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42640, {{152491, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42641, {{152493, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42642, {{152494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42643, {{152495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42644, {{152496, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42645, {{152497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42646, {{152498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42647, {{152501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42648, {{152502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42649, {{152504, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42650, {{152505, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42651, {{152506, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42652, {{152507, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42653, {{152508, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42654, {{152509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42655, {{152510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42656, {{152511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42657, {{152512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42658, {{152513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42659, {{152514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42660, {{152516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42661, {{152517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42662, {{152518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42663, {{152519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42664, {{152520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42665, {{152521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42666, {{152526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57922, {{152527, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42667, {{152530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42668, {{152531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42669, {{152532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42670, {{152534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42671, {{152536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42672, {{152537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42673, {{152538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42674, {{152540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42675, {{152541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42676, {{152542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42677, {{152543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42678, {{152544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42679, {{152546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42680, {{152549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42681, {{152553, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42682, {{152554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42683, {{152555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42684, {{152556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42685, {{152559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42686, {{152560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42687, {{152562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42688, {{152563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42689, {{152564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42690, {{152565, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42691, {{152567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42692, {{152568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42694, {{152571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42695, {{152572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42696, {{152574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42697, {{152576, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42698, {{152577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42699, {{152578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42700, {{152579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42701, {{152580, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42702, {{152593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42703, {{152595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42704, {{152597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42705, {{152598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42706, {{152599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42707, {{152600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42708, {{152601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42709, {{152602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42710, {{152603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42711, {{152605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42712, {{152606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42713, {{152612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42714, {{152617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42715, {{152618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42716, {{152619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42717, {{152620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42718, {{152622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42719, {{152634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42720, {{152636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42721, {{152637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42722, {{152639, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42723, {{152641, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42724, {{152644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42725, {{152645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42726, {{152646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42727, {{152647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42728, {{152648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42729, {{152649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42730, {{152650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42731, {{152651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42732, {{152652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42733, {{152653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42734, {{152654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42735, {{152655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42736, {{152656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42737, {{152657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42738, {{152658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42739, {{152659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42740, {{152660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42742, {{152666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42743, {{152671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42744, {{152673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42745, {{152674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42746, {{152675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42747, {{152677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42748, {{152678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42749, {{152681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42750, {{152685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42751, {{152687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42752, {{152688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42741, {{152689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42753, {{152690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42754, {{152691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42755, {{152692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42756, {{152693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42757, {{152694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42770, {{152695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42758, {{152696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42759, {{152697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42760, {{152698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42761, {{152699, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42762, {{152700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42763, {{152701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42764, {{152702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42765, {{152703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42766, {{152704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42767, {{152705, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42768, {{152706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42769, {{152718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42771, {{152720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42772, {{152721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42773, {{152722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42774, {{152724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42775, {{152727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42776, {{152729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42777, {{152731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42778, {{152732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42779, {{152734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42780, {{152736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42781, {{152737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42782, {{152738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42783, {{152739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42784, {{152740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42785, {{152743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42786, {{152744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42787, {{152745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42788, {{152746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42789, {{152747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42790, {{152748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42791, {{152749, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42792, {{152750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42793, {{152751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42794, {{152752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42795, {{152753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42796, {{152754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42797, {{152755, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42798, {{152756, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42799, {{152757, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42800, {{152758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42801, {{152764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42802, {{152765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42803, {{152768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42804, {{152769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42805, {{152771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42806, {{152774, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42807, {{152775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42808, {{152777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42810, {{152778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42811, {{152779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42812, {{152780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42813, {{152781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42814, {{152782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42815, {{152783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42816, {{152784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42817, {{152785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42818, {{152786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42819, {{152787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42820, {{152788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42821, {{152789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42822, {{152790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42823, {{152791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42824, {{152793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42825, {{152808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42826, {{152809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42827, {{152811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42828, {{152813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42829, {{152816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42830, {{152817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42831, {{152826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42832, {{152827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42833, {{152828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42834, {{152829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42835, {{152830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42836, {{152831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42837, {{152832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42838, {{152833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42839, {{152834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42840, {{152835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42841, {{152836, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42842, {{152837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42843, {{152838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42844, {{152839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42845, {{152840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42846, {{152841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42847, {{152842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42848, {{152844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42849, {{152846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42850, {{152848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42851, {{152865, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42852, {{152866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42853, {{152872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42854, {{152873, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42855, {{152874, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42856, {{152875, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42857, {{152877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42859, {{152879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42860, {{152880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42861, {{152881, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42862, {{152882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42863, {{152883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42865, {{152885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42866, {{152891, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42867, {{152897, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42868, {{152899, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42869, {{152901, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42870, {{152902, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42871, {{152903, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42872, {{152904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42873, {{152905, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42874, {{152908, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42875, {{152909, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42876, {{152910, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42877, {{152911, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42878, {{152912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42879, {{152913, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42880, {{152915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42881, {{152916, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42882, {{152917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42883, {{152918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42884, {{152919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42885, {{152920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42864, {{152921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42886, {{152926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42887, {{152929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42888, {{152931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42889, {{152944, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42890, {{152945, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42891, {{152946, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42892, {{152947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42893, {{152948, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42894, {{152949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42895, {{152950, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42896, {{152951, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42897, {{152952, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42898, {{152953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42899, {{152954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42900, {{152955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42901, {{152956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42902, {{152957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42903, {{152958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42904, {{152959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42905, {{152971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42906, {{152972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42907, {{152984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42908, {{152985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42909, {{152986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42910, {{152987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42911, {{152988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42912, {{152989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42913, {{152990, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42914, {{152991, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42915, {{152992, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42916, {{152993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42917, {{152999, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42918, {{153000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42919, {{153002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42920, {{153003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42921, {{153004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42922, {{153011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42923, {{153012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42924, {{153015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42925, {{153019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42926, {{153020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42927, {{153021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42928, {{153022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42929, {{153023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42930, {{153024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42931, {{153025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42932, {{153026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42933, {{153033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42934, {{153034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42935, {{153042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42936, {{153043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42937, {{153048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42938, {{153049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42939, {{153053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42940, {{153054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42941, {{153055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42942, {{153059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42943, {{153062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42944, {{153063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42945, {{153066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42946, {{153067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42947, {{153068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42948, {{153069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42949, {{153070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42950, {{153073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42951, {{153074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42952, {{153076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42953, {{153077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42954, {{153078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42955, {{153079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42956, {{153080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42957, {{153082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42958, {{153086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42959, {{153087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42960, {{153091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42961, {{153092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42962, {{153093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42963, {{153094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{42964, {{153095, 1, 0, 0},{153095, 1, 1, 17}, {0, 0, 0, 0}}},{42964, {{153095, 1, 0, 0},{153095, 1, 1, 17}, {0, 0, 0, 0}}},{42965, {{153095, 1, 0, 0},{153095, 1, 1, 16}, {0, 0, 0, 0}}},{42965, {{153095, 1, 0, 0},{153095, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{42966, {{153096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7482, {{153097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42967, {{153104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42968, {{153109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42969, {{153110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{42970, {{153126, 1, 0, 0},{153126, 1, 1, 17}, {0, 0, 0, 0}}},{42970, {{153126, 1, 0, 0},{153126, 1, 1, 17}, {0, 0, 0, 0}}},{42971, {{153126, 1, 0, 0},{153126, 1, 1, 16}, {0, 0, 0, 0}}},{42971, {{153126, 1, 0, 0},{153126, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42973, {{153131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{42975, {{153132, 1, 0, 0},{153132, 1, 1, 17}, {0, 0, 0, 0}}},{42975, {{153132, 1, 0, 0},{153132, 1, 1, 17}, {0, 0, 0, 0}}},{42976, {{153132, 1, 0, 0},{153132, 1, 1, 16}, {0, 0, 0, 0}}},{42976, {{153132, 1, 0, 0},{153132, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{42977, {{153133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42978, {{153134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{42980, {{153136, 1, 0, 0},{153136, 1, 1, 17}, {0, 0, 0, 0}}},{42980, {{153136, 1, 0, 0},{153136, 1, 1, 17}, {0, 0, 0, 0}}},{42981, {{153136, 1, 0, 0},{153136, 1, 1, 16}, {0, 0, 0, 0}}},{42981, {{153136, 1, 0, 0},{153136, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{42982, {{153137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42983, {{153138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42985, {{153150, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42986, {{153151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42987, {{153152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42988, {{153154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42989, {{153156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42990, {{153161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42992, {{153162, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42994, {{153163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42995, {{153165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42996, {{153166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{42997, {{153167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42998, {{153174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43000, {{153188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43001, {{153190, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43003, {{153195, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43005, {{153196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43006, {{153197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{43007, {{153198, 1, 0, 0},{153198, 1, 1, 17}, {0, 0, 0, 0}}},{43007, {{153198, 1, 0, 0},{153198, 1, 1, 17}, {0, 0, 0, 0}}},{43008, {{153198, 1, 0, 0},{153198, 1, 1, 16}, {0, 0, 0, 0}}},{43008, {{153198, 1, 0, 0},{153198, 1, 1, 16}, {0, 0, 0, 0}}},{43009, {{153198, 1, 0, 0},{153198, 1, 1, 18}, {0, 0, 0, 0}}},{43009, {{153198, 1, 0, 0},{153198, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{43010, {{153199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43011, {{153200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43012, {{153201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43013, {{153202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43014, {{153203, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43017, {{153208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{24002, {{153215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43018, {{153219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43019, {{153225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43020, {{153226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43021, {{153227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43022, {{153228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43023, {{153229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43025, {{153244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43026, {{153246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43027, {{153254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43028, {{153255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43030, {{153256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43031, {{153257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43032, {{153258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43033, {{153259, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59972, {{153274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43042, {{153280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43038, {{153281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43039, {{153282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43040, {{153283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43041, {{153285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{43035, {{153286, 1, 0, 0},{153286, 1, 1, 17}, {0, 0, 0, 0}}},{43035, {{153286, 1, 0, 0},{153286, 1, 1, 17}, {0, 0, 0, 0}}},{43036, {{153286, 1, 0, 0},{153286, 1, 1, 16}, {0, 0, 0, 0}}},{43036, {{153286, 1, 0, 0},{153286, 1, 1, 16}, {0, 0, 0, 0}}},{43037, {{153286, 1, 0, 0},{153286, 1, 1, 18}, {0, 0, 0, 0}}},{43037, {{153286, 1, 0, 0},{153286, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{43043, {{153287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43044, {{153288, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43045, {{153300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43046, {{153304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{43047, {{153305, 1, 0, 0},{153305, 1, 1, 17}, {0, 0, 0, 0}}},{43047, {{153305, 1, 0, 0},{153305, 1, 1, 17}, {0, 0, 0, 0}}},{43048, {{153305, 1, 0, 0},{153305, 1, 1, 16}, {0, 0, 0, 0}}},{43048, {{153305, 1, 0, 0},{153305, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{43049, {{153306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43050, {{153307, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{43051, {{153308, 1, 0, 0},{153308, 1, 1, 17}, {0, 0, 0, 0}}},{43051, {{153308, 1, 0, 0},{153308, 1, 1, 17}, {0, 0, 0, 0}}},{43052, {{153308, 1, 0, 0},{153308, 1, 1, 16}, {0, 0, 0, 0}}},{43052, {{153308, 1, 0, 0},{153308, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{43053, {{153309, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43054, {{153310, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43055, {{153311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43057, {{153312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43058, {{153313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43059, {{153326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43060, {{153327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43061, {{153328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43062, {{153329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{43063, {{153330, 1, 0, 0},{153330, 1, 1, 17}, {0, 0, 0, 0}}},{43063, {{153330, 1, 0, 0},{153330, 1, 1, 17}, {0, 0, 0, 0}}},{43064, {{153330, 1, 0, 0},{153330, 1, 1, 16}, {0, 0, 0, 0}}},{43064, {{153330, 1, 0, 0},{153330, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{43065, {{153331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43066, {{153332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43067, {{153334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43069, {{153339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43070, {{153340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43071, {{153341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43072, {{153343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43073, {{153347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43074, {{153348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{43075, {{153349, 1, 0, 0},{153349, 1, 1, 17}, {0, 0, 0, 0}}},{43075, {{153349, 1, 0, 0},{153349, 1, 1, 17}, {0, 0, 0, 0}}},{43076, {{153349, 1, 0, 0},{153349, 1, 1, 16}, {0, 0, 0, 0}}},{43076, {{153349, 1, 0, 0},{153349, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43077, {{153359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43078, {{153361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{43079, {{153362, 1, 0, 0},{153362, 1, 1, 17}, {0, 0, 0, 0}}},{43079, {{153362, 1, 0, 0},{153362, 1, 1, 17}, {0, 0, 0, 0}}},{43080, {{153362, 1, 0, 0},{153362, 1, 1, 16}, {0, 0, 0, 0}}},{43080, {{153362, 1, 0, 0},{153362, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{43081, {{153363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43082, {{153364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43083, {{153366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43084, {{153369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43085, {{153370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43088, {{153371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43089, {{153376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43090, {{153378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43091, {{153379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{43092, {{153382, 1, 0, 0},{153382, 1, 1, 17}, {0, 0, 0, 0}}},{43092, {{153382, 1, 0, 0},{153382, 1, 1, 17}, {0, 0, 0, 0}}},{43093, {{153382, 1, 0, 0},{153382, 1, 1, 16}, {0, 0, 0, 0}}},{43093, {{153382, 1, 0, 0},{153382, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{43094, {{153383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43098, {{153384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43097, {{153385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43099, {{153388, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43100, {{153391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43101, {{153392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43102, {{153393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43103, {{153394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43104, {{153396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43105, {{153399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43106, {{153400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43107, {{153401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43108, {{153404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43109, {{153405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43110, {{153407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43111, {{153408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43112, {{153409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43113, {{153410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43114, {{153411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43115, {{153412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43116, {{153413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43118, {{153416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43119, {{153418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43120, {{153419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43121, {{153420, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43122, {{153421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43123, {{153424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43124, {{153425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43125, {{153426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43126, {{153427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43127, {{153428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43128, {{153429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43130, {{153430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43131, {{153432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43132, {{153439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43133, {{153442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43134, {{153443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43135, {{153444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43136, {{153445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43137, {{153446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43138, {{153447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43139, {{153448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43140, {{153449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43141, {{153450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43142, {{153451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43143, {{153452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43144, {{153454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43145, {{153455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43146, {{153457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43148, {{153464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43150, {{153465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43151, {{153466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43152, {{153467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43153, {{153470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43154, {{153471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43155, {{153472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43156, {{153473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43157, {{153474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43158, {{153475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43159, {{153476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43160, {{153477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43161, {{153478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43162, {{153485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43163, {{153486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43164, {{153487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43165, {{153495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43166, {{153496, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43167, {{153497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43168, {{153498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43170, {{153500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43171, {{153501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43172, {{153502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43173, {{153511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{43174, {{153513, 1, 0, 0},{153513, 1, 1, 18}, {0, 0, 0, 0}}},{43174, {{153513, 1, 0, 0},{153513, 1, 1, 18}, {0, 0, 0, 0}}},{43175, {{153513, 1, 0, 0},{153513, 1, 1, 17}, {0, 0, 0, 0}}},{43175, {{153513, 1, 0, 0},{153513, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{43176, {{153514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43177, {{153515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43178, {{153519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43179, {{153520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43180, {{153521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43181, {{153522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43182, {{153523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{43183, {{153524, 1, 0, 0},{153524, 1, 1, 18}, {0, 0, 0, 0}}},{43183, {{153524, 1, 0, 0},{153524, 1, 1, 18}, {0, 0, 0, 0}}},{43184, {{153524, 1, 0, 0},{153524, 1, 1, 17}, {0, 0, 0, 0}}},{43184, {{153524, 1, 0, 0},{153524, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{43185, {{153525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43186, {{153526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43187, {{153527, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43188, {{153528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43189, {{153529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43190, {{153530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43191, {{153531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43192, {{153532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43193, {{153539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43194, {{153549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43195, {{153550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43196, {{153551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43197, {{153552, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43198, {{153553, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43199, {{153554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43200, {{153555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43201, {{153556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43202, {{153557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43203, {{153558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43204, {{153559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43205, {{153560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43206, {{153561, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43207, {{153562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43208, {{153563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43209, {{153564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43210, {{153565, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43211, {{153577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43212, {{153583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43213, {{153591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43214, {{153592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43215, {{153594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43216, {{153595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43217, {{153597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {2, {{43218, {{153599, 1, 0, 0}, {0, 0, 0, 0}}},{43219, {{153599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43220, {{153600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43221, {{153601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43222, {{153602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43223, {{153603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43224, {{153604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43225, {{153605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43226, {{153606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43227, {{153607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43228, {{153608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43230, {{153609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43231, {{153610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43232, {{153611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43233, {{153612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43234, {{153613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43235, {{153614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57517, {{153622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43237, {{153633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43239, {{153649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43240, {{153650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43241, {{153651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43242, {{153652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{43243, {{153653, 1, 0, 0},{153653, 1, 1, 17}, {0, 0, 0, 0}}},{43243, {{153653, 1, 0, 0},{153653, 1, 1, 17}, {0, 0, 0, 0}}},{43244, {{153653, 1, 0, 0},{153653, 1, 1, 16}, {0, 0, 0, 0}}},{43244, {{153653, 1, 0, 0},{153653, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{43246, {{153654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43247, {{153655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43257, {{153656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43248, {{153657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43249, {{153658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43250, {{153660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43251, {{153661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43252, {{153662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43253, {{153663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43254, {{153664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43255, {{153665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43256, {{153666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43258, {{153670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43260, {{153679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43261, {{153684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43262, {{153689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43263, {{153690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43264, {{153692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43265, {{153693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58229, {{153694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43266, {{153695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43267, {{153696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43268, {{153697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43269, {{153699, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43270, {{153700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43271, {{153701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43273, {{153702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43274, {{153703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43275, {{153704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43276, {{153705, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43277, {{153706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43279, {{153707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43280, {{153708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43281, {{153709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43282, {{153710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43283, {{153711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43284, {{153712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43285, {{153715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{19136, {{153716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43287, {{153721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43288, {{153728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43289, {{153729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43290, {{153735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43291, {{153736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43292, {{153738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43293, {{153739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43294, {{153740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43295, {{153741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43296, {{153742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43297, {{153743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43298, {{153744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43299, {{153745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43300, {{153746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43301, {{153747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43302, {{153748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43303, {{153749, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43304, {{153750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43305, {{153751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43307, {{153758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43308, {{153773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43309, {{153774, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43310, {{153775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43311, {{153776, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43312, {{153778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43313, {{153780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43314, {{153781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43315, {{153782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43316, {{153783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43317, {{153784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43320, {{153785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43321, {{153786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43322, {{153787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43323, {{153788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43324, {{153789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58238, {{153802, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43325, {{153803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43326, {{153804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43327, {{153811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43328, {{153814, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43329, {{153815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43330, {{153816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43331, {{153817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43332, {{153818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43333, {{153819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43334, {{153820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43335, {{153821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43336, {{153822, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43337, {{153823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43338, {{153824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43339, {{153828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43340, {{153829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43341, {{153830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43342, {{153832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43343, {{153833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43344, {{153834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43345, {{153835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43347, {{153836, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43348, {{153837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43349, {{153838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43350, {{153839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43351, {{153840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43352, {{153841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43353, {{153842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43354, {{153847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43355, {{153848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43356, {{153852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43358, {{153856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43359, {{153857, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43360, {{153858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43361, {{153859, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43362, {{153861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43363, {{153862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43364, {{153863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43357, {{153864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43365, {{153866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43366, {{153867, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43367, {{153868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43368, {{153870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43369, {{153873, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43370, {{153874, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43371, {{153875, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43372, {{153876, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43373, {{153879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43374, {{153880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43375, {{153883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43376, {{153884, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43377, {{153886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43378, {{153887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43379, {{153890, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43380, {{153891, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43381, {{153892, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43382, {{153894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43383, {{153895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43384, {{153896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43385, {{153897, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43386, {{153898, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43387, {{153899, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43388, {{153900, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43389, {{153905, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43390, {{153906, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43391, {{153907, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43392, {{153908, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43393, {{153909, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43394, {{153910, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43395, {{153911, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43396, {{153912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43397, {{153913, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43398, {{153914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43399, {{153915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43400, {{153916, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43401, {{153917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43402, {{153918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43403, {{153919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43405, {{153924, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43406, {{153927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43407, {{153928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43408, {{153929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43409, {{153930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43410, {{153931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43411, {{153932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43412, {{153933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43413, {{153934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43414, {{153935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43415, {{153936, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43416, {{153937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43417, {{153938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43418, {{153939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43419, {{153940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43420, {{153949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43421, {{153950, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43422, {{153952, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43423, {{153953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43424, {{153954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43425, {{153955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43426, {{153957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43427, {{153958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43428, {{153959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43429, {{153960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43430, {{153961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43431, {{153969, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43432, {{153971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43433, {{153972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43434, {{153974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43435, {{153975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43436, {{153976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43437, {{153977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43438, {{153978, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43439, {{153979, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43440, {{153980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43442, {{153981, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43443, {{153982, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43444, {{153983, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43445, {{153984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43446, {{153985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43447, {{153986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43448, {{153987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43449, {{153988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43450, {{153989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43451, {{153990, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43452, {{153991, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43453, {{153993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43454, {{153994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43455, {{153996, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43456, {{153997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43457, {{153998, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43458, {{153999, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43459, {{154000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43460, {{154001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43461, {{154011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43462, {{154012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43463, {{154013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43464, {{154014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43465, {{154015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43466, {{154016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43467, {{154017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43468, {{154018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43469, {{154019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43470, {{154020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43471, {{154021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43472, {{154022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43473, {{154023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43474, {{154024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43475, {{154025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43476, {{154029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43478, {{154037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43479, {{154038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43480, {{154040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43481, {{154041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43482, {{154042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43483, {{154043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43484, {{154044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43485, {{154045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43486, {{154046, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43487, {{154047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43441, {{154048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43488, {{154052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43490, {{154055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43491, {{154058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43492, {{154059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58244, {{154060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43493, {{154066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43494, {{154067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43495, {{154068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43496, {{154069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43497, {{154071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43498, {{154072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43499, {{154073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43500, {{154074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43501, {{154075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43502, {{154076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43503, {{154077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43504, {{154078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43505, {{154079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43506, {{154080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43507, {{154082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43508, {{154092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43509, {{154093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43510, {{154094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43511, {{154095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43512, {{154096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43513, {{154097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43514, {{154098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43515, {{154099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43516, {{154100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43517, {{154101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43518, {{154102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43519, {{154103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43520, {{154104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43521, {{154105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43522, {{154106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43523, {{154107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43524, {{154108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43525, {{154109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43526, {{154117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43527, {{154119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43528, {{154124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43529, {{154125, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43530, {{154126, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43531, {{154127, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43532, {{154128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43533, {{154129, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43534, {{154130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43535, {{154131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43536, {{154132, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43537, {{154133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43538, {{154134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43539, {{154135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43540, {{154136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43541, {{154140, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43542, {{154144, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43543, {{154145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43544, {{154146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43545, {{154154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{43546, {{154155, 1, 0, 0},{154155, 1, 1, 16}, {0, 0, 0, 0}}},{43546, {{154155, 1, 0, 0},{154155, 1, 1, 16}, {0, 0, 0, 0}}},{43547, {{154155, 1, 0, 0},{154155, 1, 1, 17}, {0, 0, 0, 0}}},{43547, {{154155, 1, 0, 0},{154155, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{43548, {{154156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43549, {{154157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43550, {{154158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43551, {{154161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43552, {{154162, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43553, {{154163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43554, {{154164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43555, {{154166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43556, {{154167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43557, {{154171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43558, {{154172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43559, {{154173, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43560, {{154174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43561, {{154175, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43562, {{154176, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43563, {{154177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43564, {{154178, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58252, {{154184, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43565, {{154186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43566, {{154187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43567, {{154188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43568, {{154190, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43569, {{154192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43570, {{154194, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43571, {{154195, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43572, {{154196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43573, {{154197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43574, {{154198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43575, {{154199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43576, {{154200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43577, {{154201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43578, {{154208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43579, {{154209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43580, {{154210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43581, {{154211, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43582, {{154212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43583, {{154221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43584, {{154222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43585, {{154225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43586, {{154226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43587, {{154227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43588, {{154228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43589, {{154229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43590, {{154230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43591, {{154235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43592, {{154238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43593, {{154239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43594, {{154240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43595, {{154241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43596, {{154242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43597, {{154243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43598, {{154250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43599, {{154251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43600, {{154252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43601, {{154253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43602, {{154254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43603, {{154255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43604, {{154256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43605, {{154257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43606, {{154261, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43607, {{154264, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43608, {{154265, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43609, {{154266, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43610, {{154267, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43611, {{154277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43612, {{154278, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{43613, {{154279, 1, 0, 0},{154279, 1, 1, 16}, {0, 0, 0, 0}}},{43613, {{154279, 1, 0, 0},{154279, 1, 1, 16}, {0, 0, 0, 0}}},{43614, {{154279, 1, 0, 0},{154279, 1, 1, 17}, {0, 0, 0, 0}}},{43614, {{154279, 1, 0, 0},{154279, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43615, {{154283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43616, {{154284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43617, {{154286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43618, {{154287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43619, {{154288, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43620, {{154289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43621, {{154291, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43622, {{154293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43623, {{154294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43624, {{154298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43625, {{154299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43626, {{154300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43627, {{154303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43628, {{154306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43629, {{154307, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43631, {{154309, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43632, {{154312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43633, {{154313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43634, {{154315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43635, {{154317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43636, {{154318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43637, {{154320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{43638, {{154324, 1, 0, 0},{154324, 1, 1, 16}, {0, 0, 0, 0}}},{43638, {{154324, 1, 0, 0},{154324, 1, 1, 16}, {0, 0, 0, 0}}},{43639, {{154324, 1, 0, 0},{154324, 1, 1, 17}, {0, 0, 0, 0}}},{43639, {{154324, 1, 0, 0},{154324, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43640, {{154326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43641, {{154329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43642, {{154330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43643, {{154331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43644, {{154332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43645, {{154333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43646, {{154334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43647, {{154335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43648, {{154336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{43650, {{154337, 1, 0, 0},{154337, 1, 1, 16}, {0, 0, 0, 0}}},{43650, {{154337, 1, 0, 0},{154337, 1, 1, 16}, {0, 0, 0, 0}}},{58258, {{154337, 1, 0, 0},{154337, 1, 1, 17}, {0, 0, 0, 0}}},{58258, {{154337, 1, 0, 0},{154337, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{43651, {{154338, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43652, {{154339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43653, {{154340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43655, {{154348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43656, {{154351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43657, {{154352, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43658, {{154353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43649, {{154354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43659, {{154355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43660, {{154356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43661, {{154357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43662, {{154358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43663, {{154359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43664, {{154360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43665, {{154361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43666, {{154363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43667, {{154364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43668, {{154365, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60015, {{154369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58259, {{154371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43669, {{154374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43670, {{154375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43671, {{154376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43672, {{154377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43673, {{154380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43674, {{154381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43675, {{154382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43676, {{154383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43677, {{154384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43678, {{154386, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43679, {{154387, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43680, {{154388, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43681, {{154389, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43683, {{154390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43684, {{154391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43690, {{154392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43685, {{154393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43686, {{154394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43687, {{154395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43688, {{154397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43689, {{154398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43691, {{154410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43692, {{154411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43693, {{154412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43694, {{154413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43695, {{154414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43696, {{154415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43697, {{154416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43698, {{154417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43699, {{154418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43700, {{154419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43701, {{154420, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43702, {{154421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43703, {{154422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43704, {{154423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43705, {{154424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43706, {{154425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43707, {{154426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43708, {{154427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43709, {{154428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43710, {{154429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43711, {{154430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43712, {{154431, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43714, {{154443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43715, {{154444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43716, {{154448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43717, {{154449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43718, {{154450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43719, {{154451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43720, {{154452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43721, {{154453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43722, {{154454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43723, {{154455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43724, {{154456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43725, {{154457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43726, {{154458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43727, {{154459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43728, {{154460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43729, {{154461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43730, {{154462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{43731, {{154463, 1, 0, 0},{154463, 1, 1, 16}, {0, 0, 0, 0}}},{43731, {{154463, 1, 0, 0},{154463, 1, 1, 16}, {0, 0, 0, 0}}},{43732, {{154463, 1, 0, 0},{154463, 1, 1, 17}, {0, 0, 0, 0}}},{43732, {{154463, 1, 0, 0},{154463, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{43733, {{154464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43734, {{154465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43735, {{154466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43736, {{154467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43737, {{154468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43738, {{154469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43739, {{154470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43740, {{154471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43741, {{154472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43742, {{154473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43743, {{154474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43744, {{154475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43745, {{154476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43746, {{154477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43747, {{154478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43748, {{154480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43749, {{154481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43750, {{154487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43751, {{154488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43752, {{154501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43753, {{154509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43754, {{154510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43755, {{154511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43756, {{154512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43757, {{154513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43758, {{154514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43759, {{154515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43760, {{154516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43761, {{154517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43762, {{154518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43763, {{154519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43764, {{154520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43765, {{154521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43766, {{154522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43767, {{154524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43768, {{154525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43769, {{154528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43770, {{154529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43771, {{154530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43772, {{154531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43773, {{154532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43774, {{154533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43775, {{154534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43776, {{154535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43777, {{154536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43778, {{154538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{43779, {{154539, 1, 0, 0},{154539, 1, 1, 16}, {0, 0, 0, 0}}},{43779, {{154539, 1, 0, 0},{154539, 1, 1, 16}, {0, 0, 0, 0}}},{43780, {{154539, 1, 0, 0},{154539, 1, 1, 17}, {0, 0, 0, 0}}},{43780, {{154539, 1, 0, 0},{154539, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{43781, {{154540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43782, {{154541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43783, {{154542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43784, {{154543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43785, {{154546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43786, {{154563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43787, {{154577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43788, {{154578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43789, {{154579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43790, {{154580, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43791, {{154581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43792, {{154582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43793, {{154583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43794, {{154584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43795, {{154585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43796, {{154586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43797, {{154587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43798, {{154588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43799, {{154589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43800, {{154590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43801, {{154591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43802, {{154592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43803, {{154593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43804, {{154594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43805, {{154595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43806, {{154596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43807, {{154597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43808, {{154598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43809, {{154599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43810, {{154600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43811, {{154601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43812, {{154602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43813, {{154603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43814, {{154604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43815, {{154605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43816, {{154606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43817, {{154607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43818, {{154608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43819, {{154609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43820, {{154611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43821, {{154612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43822, {{154613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43823, {{154614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43824, {{154615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43825, {{154616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43826, {{154617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43827, {{154618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43828, {{154619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43829, {{154620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43830, {{154621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43831, {{154622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{43832, {{154623, 1, 0, 0},{154623, 1, 1, 16}, {0, 0, 0, 0}}},{43832, {{154623, 1, 0, 0},{154623, 1, 1, 16}, {0, 0, 0, 0}}},{68086, {{154623, 1, 0, 0},{154623, 1, 1, 17}, {0, 0, 0, 0}}},{68086, {{154623, 1, 0, 0},{154623, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43833, {{154626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43835, {{154635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43836, {{154648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43837, {{154649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43838, {{154650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43839, {{154651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43840, {{154652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43841, {{154653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43842, {{154654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43843, {{154655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43844, {{154656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43845, {{154657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43846, {{154658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43847, {{154659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43848, {{154660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43849, {{154661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43850, {{154662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43851, {{154663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43852, {{154664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43853, {{154665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43854, {{154666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43855, {{154667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43856, {{154668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43857, {{154669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43858, {{154670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43859, {{154671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43860, {{154672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43861, {{154673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43862, {{154674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43863, {{154675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43864, {{154676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43865, {{154677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43866, {{154678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43867, {{154679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43868, {{154680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43869, {{154681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43870, {{154682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43871, {{154683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43872, {{154684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43873, {{154685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43874, {{154686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43875, {{154687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43876, {{154688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43877, {{154689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43878, {{154690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43879, {{154692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43880, {{154693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43881, {{154695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{43882, {{154699, 1, 0, 0},{154699, 1, 1, 17}, {0, 0, 0, 0}}},{43882, {{154699, 1, 0, 0},{154699, 1, 1, 17}, {0, 0, 0, 0}}},{60022, {{154699, 1, 0, 0},{154699, 1, 1, 18}, {0, 0, 0, 0}}},{60022, {{154699, 1, 0, 0},{154699, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43883, {{154703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43884, {{154724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43885, {{154728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43886, {{154729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43887, {{154741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43888, {{154742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43889, {{154743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43890, {{154744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43891, {{154745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43892, {{154746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43894, {{154747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43895, {{154748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43896, {{154749, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43897, {{154750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43898, {{154751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{43899, {{154752, 1, 0, 0},{154752, 1, 1, 16}, {0, 0, 0, 0}}},{43899, {{154752, 1, 0, 0},{154752, 1, 1, 16}, {0, 0, 0, 0}}},{43900, {{154752, 1, 0, 0},{154752, 1, 1, 17}, {0, 0, 0, 0}}},{43900, {{154752, 1, 0, 0},{154752, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{43901, {{154753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43834, {{154754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43902, {{154756, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43903, {{154758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43904, {{154759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43905, {{154760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43906, {{154761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43907, {{154762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43908, {{154763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43909, {{154764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43910, {{154765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43911, {{154766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43912, {{154767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43913, {{154768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43914, {{154769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43915, {{154770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43916, {{154773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43917, {{154774, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43918, {{154775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43919, {{154776, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43920, {{154777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43921, {{154778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43922, {{154779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43923, {{154780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43924, {{154781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43925, {{154782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43926, {{154783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43927, {{154784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43928, {{154785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43929, {{154786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43930, {{154787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43931, {{154788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43932, {{154789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43933, {{154790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43934, {{154791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43935, {{154792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43936, {{154793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43937, {{154794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43938, {{154795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43939, {{154798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43940, {{154799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43941, {{154811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43942, {{154812, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43943, {{154813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43944, {{154822, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43945, {{154823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43946, {{154824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43947, {{154825, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43949, {{154826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43950, {{154827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43951, {{154828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43953, {{154829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43955, {{154830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43956, {{154831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43958, {{154832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43959, {{154833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43960, {{154834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43961, {{154835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43962, {{154836, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43963, {{154837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43964, {{154838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43965, {{154839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43966, {{154840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43967, {{154841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43968, {{154842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43969, {{154843, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43970, {{154844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43971, {{154845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43972, {{154846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43973, {{154847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43974, {{154848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43975, {{154849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43976, {{154850, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43977, {{154851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43978, {{154852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43979, {{154853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43980, {{154854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43981, {{154855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43982, {{154856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43983, {{154857, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43984, {{154858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43985, {{154859, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43986, {{154860, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43987, {{154861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43988, {{154862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43989, {{154863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43990, {{154864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43991, {{154865, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43992, {{154866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43993, {{154867, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43994, {{154868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43995, {{154869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43996, {{154870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43997, {{154871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43998, {{154879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43999, {{154880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44000, {{154882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44001, {{154896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44002, {{154897, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44003, {{154898, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44004, {{154899, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44005, {{154900, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44006, {{154913, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44007, {{154914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44008, {{154915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44009, {{154916, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44010, {{154917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44011, {{154918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44012, {{154919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44013, {{154920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44014, {{154921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44015, {{154922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44016, {{154923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44017, {{154924, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44018, {{154925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44019, {{154926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44020, {{154927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44021, {{154928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44022, {{154929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44023, {{154930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44024, {{154932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44025, {{154933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44026, {{154934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44027, {{154935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44028, {{154936, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44029, {{154937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44030, {{154938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44031, {{154939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44032, {{154940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44033, {{154943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44034, {{154949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44035, {{154959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44036, {{154960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44037, {{154962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44038, {{154963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44039, {{154972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44040, {{154973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44041, {{154974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44042, {{154975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44043, {{154976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44044, {{154977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44045, {{154978, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44046, {{154979, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44047, {{154980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44048, {{154981, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44049, {{154982, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44050, {{154983, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44051, {{154984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44052, {{154985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44053, {{154986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44054, {{154987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44055, {{154988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44056, {{154989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44057, {{154990, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44058, {{154991, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44059, {{154992, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44060, {{154993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44061, {{154994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44062, {{154995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44063, {{154996, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44064, {{154997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44065, {{154998, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44066, {{155000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44067, {{155001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44068, {{155005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44069, {{155013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44070, {{155014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44071, {{155015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44072, {{155016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44073, {{155018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44074, {{155019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44075, {{155020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44076, {{155021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44077, {{155022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44078, {{155023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44079, {{155024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44080, {{155025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44081, {{155026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44082, {{155027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44083, {{155028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44084, {{155029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44085, {{155030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44086, {{155031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44087, {{155032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44088, {{155033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44089, {{155034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44090, {{155035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44091, {{155036, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44092, {{155037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44093, {{155038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44094, {{155039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44095, {{155041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44097, {{155042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44098, {{155050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44099, {{155051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44100, {{155060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44101, {{155061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44102, {{155062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44103, {{155063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44104, {{155064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44105, {{155065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44106, {{155066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44107, {{155067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44108, {{155068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44109, {{155069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44110, {{155070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44111, {{155071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44112, {{155072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44113, {{155073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44114, {{155074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44115, {{155075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44116, {{155076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44117, {{155077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44118, {{155078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44119, {{155079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44120, {{155080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44121, {{155081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44122, {{155088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44123, {{155092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44124, {{155093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44125, {{155094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44126, {{155095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44127, {{155096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44128, {{155097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44129, {{155098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44130, {{155099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44131, {{155100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44132, {{155101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44133, {{155102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44134, {{155103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44135, {{155104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44136, {{155106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44137, {{155107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44138, {{155108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44139, {{155112, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44140, {{155113, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44141, {{155115, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44142, {{155116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44143, {{155117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44144, {{155118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44145, {{155119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44146, {{155120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44147, {{155121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44148, {{155122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44149, {{155123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44150, {{155128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44151, {{155129, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44152, {{155132, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44153, {{155133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44154, {{155135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44155, {{155136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44158, {{155138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44162, {{155139, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44157, {{155141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44159, {{155144, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44160, {{155145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44161, {{155146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44156, {{155147, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44163, {{155152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44164, {{155153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44165, {{155158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44166, {{155159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44167, {{155160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44168, {{155162, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44169, {{155164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44170, {{155169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44171, {{155170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44172, {{155171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44173, {{155172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44174, {{155173, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44175, {{155175, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44176, {{155176, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44177, {{155178, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44178, {{155179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44179, {{155180, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44180, {{155181, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44181, {{155182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44182, {{155183, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44183, {{155186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44184, {{155187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44185, {{155188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44186, {{155189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44187, {{155190, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44188, {{155191, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44189, {{155192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44190, {{155197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44191, {{155198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44192, {{155201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44193, {{155202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44194, {{155203, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44195, {{155204, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44196, {{155205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44197, {{155206, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44198, {{155207, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44199, {{155208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44200, {{155209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44201, {{155210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44202, {{155211, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44203, {{155212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44204, {{155213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44205, {{155214, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{44206, {{155215, 1, 0, 0},{155215, 1, 1, 16}, {0, 0, 0, 0}}},{44206, {{155215, 1, 0, 0},{155215, 1, 1, 16}, {0, 0, 0, 0}}},{58287, {{155215, 1, 0, 0},{155215, 1, 1, 17}, {0, 0, 0, 0}}},{58287, {{155215, 1, 0, 0},{155215, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44207, {{155219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44208, {{155220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44209, {{155221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44210, {{155222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44211, {{155226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44212, {{155227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44213, {{155228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44214, {{155229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44215, {{155230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44216, {{155231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44217, {{155232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44218, {{155233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44219, {{155234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44220, {{155236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44221, {{155237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44222, {{155241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44223, {{155243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44225, {{155249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44226, {{155250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44227, {{155251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44228, {{155252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44229, {{155253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44230, {{155254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44231, {{155255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44232, {{155256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44233, {{155257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44234, {{155258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44235, {{155259, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44236, {{155260, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44237, {{155261, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44238, {{155262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44239, {{155263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44240, {{155272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44241, {{155273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44242, {{155274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44243, {{155275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44244, {{155288, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44245, {{155289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44246, {{155290, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{44247, {{155291, 1, 0, 0},{155291, 1, 1, 17}, {0, 0, 0, 0}}},{44247, {{155291, 1, 0, 0},{155291, 1, 1, 17}, {0, 0, 0, 0}}},{58292, {{155291, 1, 0, 0},{155291, 1, 1, 16}, {0, 0, 0, 0}}},{58292, {{155291, 1, 0, 0},{155291, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{44248, {{155292, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44249, {{155293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44251, {{155294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44252, {{155295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44253, {{155296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44254, {{155297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44255, {{155298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44256, {{155299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44257, {{155300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44258, {{155315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44259, {{155317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44260, {{155318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44261, {{155319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44262, {{155320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44263, {{155329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44264, {{155330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44266, {{155331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44267, {{155332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44268, {{155333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44269, {{155334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44270, {{155335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44271, {{155336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44272, {{155337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44273, {{155338, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44274, {{155339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44275, {{155340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44276, {{155341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44277, {{155342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44278, {{155343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44279, {{155344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44280, {{155345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44281, {{155346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44282, {{155347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44283, {{155348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44284, {{155349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44285, {{155350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44286, {{155352, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44287, {{155357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44288, {{155358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44289, {{155359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44290, {{155364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44291, {{155365, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44292, {{155366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44293, {{155367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44295, {{155368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44296, {{155369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44297, {{155370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44298, {{155371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44299, {{155372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44300, {{155373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44301, {{155374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44302, {{155375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44303, {{155376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44304, {{155379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44305, {{155380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44306, {{155381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44307, {{155390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44308, {{155391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44309, {{155392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44310, {{155393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44311, {{155394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44312, {{155395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44313, {{155396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44314, {{155397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44315, {{155398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44316, {{155399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44317, {{155400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44318, {{155401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44319, {{155403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44320, {{155411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44321, {{155416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44322, {{155417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44323, {{155418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44324, {{155419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44325, {{155420, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44326, {{155421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44327, {{155422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44328, {{155423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44329, {{155424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44330, {{155425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44332, {{155427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44333, {{155428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44334, {{155433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44335, {{155434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44336, {{155435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44337, {{155436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44338, {{155437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44339, {{155438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44341, {{155446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44342, {{155447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44343, {{155448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44344, {{155449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44345, {{155450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44346, {{155451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44347, {{155452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44348, {{155453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44349, {{155454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44350, {{155455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44351, {{155456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44352, {{155457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44353, {{155458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44354, {{155459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44355, {{155461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44356, {{155463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44357, {{155464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44358, {{155465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44359, {{155466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44360, {{155471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44361, {{155472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44362, {{155473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44331, {{155474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44363, {{155479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44364, {{155480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44365, {{155481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44366, {{155482, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44367, {{155483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44368, {{155484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44369, {{155487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44370, {{155491, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44371, {{155493, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44372, {{155494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44373, {{155495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44374, {{155500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44375, {{155501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44376, {{155502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44377, {{155504, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44378, {{155505, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44379, {{155506, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44380, {{155507, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44381, {{155508, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44382, {{155510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44383, {{155511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44384, {{155512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44385, {{155513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44386, {{155515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44387, {{155516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44388, {{155517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44389, {{155518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44390, {{155519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44391, {{155520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44392, {{155521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44393, {{155522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44394, {{155524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44395, {{155525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{44396, {{155526, 1, 0, 0},{155526, 1, 1, 16}, {0, 0, 0, 0}}},{44396, {{155526, 1, 0, 0},{155526, 1, 1, 16}, {0, 0, 0, 0}}},{44397, {{155526, 1, 0, 0},{155526, 1, 1, 17}, {0, 0, 0, 0}}},{44397, {{155526, 1, 0, 0},{155526, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{44398, {{155527, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44399, {{155530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44400, {{155531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44401, {{155532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44402, {{155533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44403, {{155534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44404, {{155535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44405, {{155536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44406, {{155538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44407, {{155539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44408, {{155541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44409, {{155542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44410, {{155545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44411, {{155547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44412, {{155548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44413, {{155549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{44414, {{155550, 1, 0, 0},{155550, 1, 1, 17}, {0, 0, 0, 0}}},{44414, {{155550, 1, 0, 0},{155550, 1, 1, 17}, {0, 0, 0, 0}}},{58300, {{155550, 1, 0, 0},{155550, 1, 1, 16}, {0, 0, 0, 0}}},{58300, {{155550, 1, 0, 0},{155550, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44415, {{155553, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44416, {{155554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44417, {{155555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44418, {{155556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44419, {{155557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44420, {{155558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44421, {{155559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44422, {{155560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44423, {{155561, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44424, {{155562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44425, {{155563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44426, {{155564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44427, {{155566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44428, {{155567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44429, {{155577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44430, {{155578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44431, {{155579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44432, {{155580, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44433, {{155587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44434, {{155588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44435, {{155589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44436, {{155590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44437, {{155591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44438, {{155592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44439, {{155593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44440, {{155594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44441, {{155595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44442, {{155596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44443, {{155597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44444, {{155598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44445, {{155599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44446, {{155600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44447, {{155601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44448, {{155603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44449, {{155604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44450, {{155612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44452, {{155615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44453, {{155616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44454, {{155622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44455, {{155623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44456, {{155624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44457, {{155625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44459, {{155626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44460, {{155627, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44461, {{155628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44462, {{155629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44463, {{155630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44464, {{155631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44465, {{155632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44466, {{155633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44467, {{155634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44468, {{155635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44469, {{155636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44470, {{155637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44471, {{155640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44472, {{155641, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44473, {{155642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44474, {{155643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44475, {{155645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44476, {{155656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44478, {{155657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44479, {{155660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44480, {{155666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44481, {{155667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44482, {{155668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44483, {{155669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44484, {{155670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44485, {{155671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44486, {{155672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44487, {{155673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44488, {{155674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44489, {{155675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44490, {{155676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44491, {{155677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44492, {{155678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44493, {{155679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44494, {{155680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44495, {{155681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44496, {{155682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44497, {{155683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44498, {{155684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44499, {{155685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44500, {{155686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44501, {{155687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44502, {{155688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44503, {{155704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44504, {{155715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44505, {{155718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44506, {{155720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44507, {{155721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44508, {{155722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44509, {{155723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44510, {{155725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44511, {{155726, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44512, {{155727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44513, {{155728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44514, {{155729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44515, {{155730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44516, {{155731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44517, {{155732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44518, {{155733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44519, {{155734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44520, {{155735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44521, {{155736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44522, {{155737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44523, {{155738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44524, {{155739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44525, {{155740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44526, {{155742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44527, {{155743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44528, {{155744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44529, {{155748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44530, {{155749, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58310, {{155760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44532, {{155761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44533, {{155762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44534, {{155763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44535, {{155764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44536, {{155766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44537, {{155767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44538, {{155776, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44539, {{155777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44540, {{155778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44541, {{155779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44542, {{155780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44543, {{155781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44544, {{155782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44545, {{155783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44546, {{155784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44547, {{155785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44548, {{155786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44549, {{155787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44550, {{155788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44551, {{155789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44552, {{155790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44553, {{155791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44554, {{155792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44555, {{155793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44556, {{155794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44557, {{155795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44558, {{155796, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44559, {{155797, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44531, {{155798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44560, {{155799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44561, {{155800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44562, {{155801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44563, {{155802, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44564, {{155803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44565, {{155804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44566, {{155805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44567, {{155808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44568, {{155809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44570, {{155828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44571, {{155829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44573, {{155831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44574, {{155836, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44575, {{155841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44576, {{155843, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44577, {{155844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44578, {{155845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44579, {{155846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44580, {{155847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{44581, {{155848, 1, 0, 0},{155848, 1, 1, 16}, {0, 0, 0, 0}}},{44581, {{155848, 1, 0, 0},{155848, 1, 1, 16}, {0, 0, 0, 0}}},{44582, {{155848, 1, 0, 0},{155848, 1, 1, 17}, {0, 0, 0, 0}}},{44582, {{155848, 1, 0, 0},{155848, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{44583, {{155849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44584, {{155850, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58323, {{155851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44585, {{155852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44586, {{155853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44587, {{155854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44588, {{155855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44589, {{155856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44590, {{155857, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44591, {{155858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44592, {{155859, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44593, {{155860, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44594, {{155861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44595, {{155862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44596, {{155863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44597, {{155864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44598, {{155865, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44599, {{155866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44600, {{155867, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44601, {{155868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44602, {{155869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44603, {{155870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44604, {{155871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44605, {{155872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44606, {{155873, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44607, {{155874, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44608, {{155875, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44609, {{155877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44610, {{155879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44611, {{155885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44612, {{155901, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44613, {{155906, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44614, {{155913, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44615, {{155914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44616, {{155915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44617, {{155916, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44618, {{155917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44619, {{155918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44620, {{155919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44621, {{155920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44622, {{155921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44623, {{155922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44624, {{155923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44625, {{155924, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44626, {{155925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44627, {{155926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44628, {{155927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44629, {{155928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44630, {{155929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44631, {{155930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44632, {{155931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44633, {{155932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44634, {{155951, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44635, {{155952, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44636, {{155953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44637, {{155954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44638, {{155963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44639, {{155964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44640, {{155965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44641, {{155966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44642, {{155967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44643, {{155968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44645, {{155969, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44646, {{155971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44647, {{155972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44648, {{155974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44649, {{155975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44650, {{155976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44651, {{155977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44652, {{155978, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44653, {{155979, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44654, {{155980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44655, {{155981, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44656, {{155982, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44657, {{155983, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44658, {{155984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44659, {{155985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44660, {{155986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44661, {{155987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44662, {{155988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44663, {{155989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44664, {{155992, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44665, {{156005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44666, {{156006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44667, {{156007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44668, {{156008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44669, {{156009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44670, {{156012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44671, {{156021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44672, {{156022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44673, {{156023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44674, {{156024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44675, {{156026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44676, {{156027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44678, {{156028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44679, {{156029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44680, {{156030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44681, {{156031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44682, {{156032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44683, {{156033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44684, {{156034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44685, {{156035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44686, {{156036, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44687, {{156037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44688, {{156038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44689, {{156040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44690, {{156041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44691, {{156042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44692, {{156043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44693, {{156045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44694, {{156046, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44695, {{156047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44696, {{156053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44697, {{156054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44698, {{156055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44699, {{156056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44700, {{156060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44701, {{156061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44702, {{156062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44703, {{156063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44704, {{156064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44705, {{156065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{44706, {{156066, 1, 0, 0},{156066, 1, 1, 16}, {0, 0, 0, 0}}},{44706, {{156066, 1, 0, 0},{156066, 1, 1, 16}, {0, 0, 0, 0}}},{58330, {{156066, 1, 0, 0},{156066, 1, 1, 17}, {0, 0, 0, 0}}},{58330, {{156066, 1, 0, 0},{156066, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{44707, {{156067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44708, {{156069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44709, {{156070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44710, {{156071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44711, {{156072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44712, {{156073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44713, {{156074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44714, {{156075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44715, {{156076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44716, {{156079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44717, {{156080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44718, {{156081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44719, {{156088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44720, {{156091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44721, {{156093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44722, {{156094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44723, {{156095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44724, {{156096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44725, {{156097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44726, {{156098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44727, {{156099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44728, {{156100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44729, {{156101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44730, {{156104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44731, {{156110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44732, {{156111, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44733, {{156113, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44734, {{156115, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44735, {{156116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44736, {{156117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44737, {{156118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{44738, {{156119, 1, 0, 0},{156119, 1, 1, 17}, {0, 0, 0, 0}}},{44738, {{156119, 1, 0, 0},{156119, 1, 1, 17}, {0, 0, 0, 0}}},{44739, {{156119, 1, 0, 0},{156119, 1, 1, 16}, {0, 0, 0, 0}}},{44739, {{156119, 1, 0, 0},{156119, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{44740, {{156120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44741, {{156121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{44742, {{156122, 1, 0, 0},{156122, 1, 1, 16}, {0, 0, 0, 0}}},{44742, {{156122, 1, 0, 0},{156122, 1, 1, 16}, {0, 0, 0, 0}}},{44743, {{156122, 1, 0, 0},{156122, 1, 1, 17}, {0, 0, 0, 0}}},{44743, {{156122, 1, 0, 0},{156122, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{44744, {{156123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44745, {{156126, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44746, {{156130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44747, {{156131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44748, {{156133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44749, {{156135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44750, {{156136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44751, {{156137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44752, {{156139, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44753, {{156140, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44754, {{156141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44755, {{156144, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44756, {{156145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44757, {{156147, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44758, {{156149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44759, {{156150, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44760, {{156151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44761, {{156152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44762, {{156153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44763, {{156154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44764, {{156155, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44765, {{156156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44766, {{156158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44767, {{156160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44768, {{156161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44769, {{156163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44770, {{156165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44771, {{156167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{44772, {{156194, 1, 0, 0},{156194, 1, 1, 18}, {0, 0, 0, 0}}},{44772, {{156194, 1, 0, 0},{156194, 1, 1, 18}, {0, 0, 0, 0}}},{57891, {{156194, 1, 0, 0},{156194, 1, 1, 17}, {0, 0, 0, 0}}},{57891, {{156194, 1, 0, 0},{156194, 1, 1, 17}, {0, 0, 0, 0}}},{58334, {{156194, 1, 0, 0},{156194, 1, 1, 19}, {0, 0, 0, 0}}},{58334, {{156194, 1, 0, 0},{156194, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{44774, {{156195, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44775, {{156196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44776, {{156198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44777, {{156199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{44778, {{156200, 1, 0, 0},{156200, 1, 1, 16}, {0, 0, 0, 0}}},{44778, {{156200, 1, 0, 0},{156200, 1, 1, 16}, {0, 0, 0, 0}}},{44779, {{156200, 1, 0, 0},{156200, 1, 1, 17}, {0, 0, 0, 0}}},{44779, {{156200, 1, 0, 0},{156200, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{44780, {{156201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44781, {{156203, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44782, {{156204, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44783, {{156205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44784, {{156206, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44785, {{156207, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44786, {{156208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44787, {{156209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44788, {{156210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44789, {{156212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44790, {{156213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44791, {{156214, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44792, {{156215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44793, {{156216, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44794, {{156217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44795, {{156218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44796, {{156219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44797, {{156220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44798, {{156225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44799, {{156226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44800, {{156230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{44801, {{156231, 1, 0, 0},{156231, 1, 1, 16}, {0, 0, 0, 0}}},{44801, {{156231, 1, 0, 0},{156231, 1, 1, 16}, {0, 0, 0, 0}}},{44802, {{156231, 1, 0, 0},{156231, 1, 1, 17}, {0, 0, 0, 0}}},{44802, {{156231, 1, 0, 0},{156231, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{44803, {{156232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44804, {{156233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44805, {{156238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44806, {{156239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44807, {{156240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44808, {{156245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44809, {{156246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44810, {{156249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44811, {{156250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44812, {{156253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44813, {{156254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44814, {{156255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44815, {{156256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44816, {{156258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44817, {{156259, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44818, {{156261, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44819, {{156262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44820, {{156263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44821, {{156265, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44822, {{156266, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44823, {{156267, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44824, {{156268, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44825, {{156269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44826, {{156270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44827, {{156271, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44828, {{156272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{44829, {{156275, 1, 0, 0},{156275, 1, 1, 16}, {0, 0, 0, 0}}},{44829, {{156275, 1, 0, 0},{156275, 1, 1, 16}, {0, 0, 0, 0}}},{44830, {{156275, 1, 0, 0},{156275, 1, 1, 17}, {0, 0, 0, 0}}},{44830, {{156275, 1, 0, 0},{156275, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44831, {{156279, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44832, {{156280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44833, {{156281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44834, {{156283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44835, {{156284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44838, {{156285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44836, {{156286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44837, {{156287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44839, {{156289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44840, {{156290, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44841, {{156291, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44842, {{156294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{44844, {{156299, 1, 0, 0},{156299, 1, 1, 16}, {0, 0, 0, 0}}},{44844, {{156299, 1, 0, 0},{156299, 1, 1, 16}, {0, 0, 0, 0}}},{57703, {{156299, 1, 0, 0},{156299, 1, 1, 17}, {0, 0, 0, 0}}},{57703, {{156299, 1, 0, 0},{156299, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{44843, {{156300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44845, {{156301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44846, {{156303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44847, {{156304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44848, {{156305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44849, {{156306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44850, {{156307, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44851, {{156308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44852, {{156309, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44853, {{156310, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44854, {{156312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44855, {{156317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44856, {{156318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44857, {{156319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44858, {{156320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44859, {{156321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44860, {{156324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44867, {{156326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44861, {{156327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44862, {{156329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44863, {{156330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44864, {{156331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44865, {{156333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44866, {{156334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44868, {{156347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44869, {{156348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44870, {{156349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44871, {{156350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44872, {{156351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44873, {{156356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44874, {{156357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44875, {{156358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44876, {{156359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44877, {{156360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44878, {{156361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44879, {{156362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44880, {{156363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44881, {{156364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44882, {{156365, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44883, {{156369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44884, {{156374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44885, {{156375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44886, {{156376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{44887, {{156377, 1, 0, 0},{156377, 1, 1, 16}, {0, 0, 0, 0}}},{44887, {{156377, 1, 0, 0},{156377, 1, 1, 16}, {0, 0, 0, 0}}},{44888, {{156377, 1, 0, 0},{156377, 1, 1, 17}, {0, 0, 0, 0}}},{44888, {{156377, 1, 0, 0},{156377, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{44889, {{156378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44890, {{156379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44891, {{156380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44892, {{156381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44893, {{156382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44894, {{156383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44895, {{156384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44896, {{156385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44897, {{156387, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44898, {{156391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44899, {{156394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44900, {{156395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44901, {{156396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44902, {{156397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44903, {{156398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44904, {{156399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44905, {{156400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44906, {{156402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44907, {{156403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44908, {{156404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44909, {{156409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44910, {{156411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44911, {{156412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44912, {{156413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44913, {{156414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44914, {{156415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44915, {{156416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44916, {{156417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44917, {{156418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44918, {{156419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44919, {{156421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44920, {{156422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44921, {{156426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44923, {{156431, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44924, {{156432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44925, {{156433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44926, {{156434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44927, {{156435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44928, {{156436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44929, {{156437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44930, {{156438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44931, {{156442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44932, {{156444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44933, {{156445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44934, {{156446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44935, {{156447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44936, {{156448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44937, {{156449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44938, {{156450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44939, {{156451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44940, {{156452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44941, {{156454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44943, {{156457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44944, {{156458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44945, {{156459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44946, {{156460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44947, {{156461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44948, {{156462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44949, {{156465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44950, {{156466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44951, {{156467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44952, {{156468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44953, {{156469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44954, {{156470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44955, {{156471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44956, {{156472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44957, {{156474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44958, {{156475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44959, {{156476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44960, {{156478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44961, {{156479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44962, {{156481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44964, {{156483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44963, {{156485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44965, {{156486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44966, {{156487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44967, {{156488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44968, {{156490, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44969, {{156491, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44970, {{156492, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44971, {{156494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44972, {{156495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44973, {{156496, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44974, {{156497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44975, {{156498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44976, {{156500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44977, {{156501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44978, {{156503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44979, {{156504, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44980, {{156506, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44981, {{156507, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44982, {{156508, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44983, {{156509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44984, {{156510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44985, {{156511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59667, {{156513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57686, {{156516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44986, {{156517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44987, {{156518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44988, {{156519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44989, {{156520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44990, {{156521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44991, {{156522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44992, {{156523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44993, {{156524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44994, {{156525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44995, {{156526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44996, {{156529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44997, {{156530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44998, {{156532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{44999, {{156533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45000, {{156535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45001, {{156537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45002, {{156538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45003, {{156539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45004, {{156540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45005, {{156541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45006, {{156542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45007, {{156543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45008, {{156544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45009, {{156547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45010, {{156548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45011, {{156550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45012, {{156551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45013, {{156552, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45014, {{156553, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45015, {{156554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45016, {{156555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45017, {{156556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45018, {{156557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45019, {{156558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45020, {{156562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45021, {{156567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45022, {{156568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45023, {{156569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45024, {{156570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45025, {{156573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45026, {{156574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45027, {{156575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45028, {{156578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45029, {{156579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45030, {{156580, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45031, {{156581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45032, {{156582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45033, {{156583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45034, {{156584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45035, {{156585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45036, {{156589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45037, {{156591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45038, {{156592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{45039, {{156593, 1, 0, 0},{156593, 1, 1, 16}, {0, 0, 0, 0}}},{45039, {{156593, 1, 0, 0},{156593, 1, 1, 16}, {0, 0, 0, 0}}},{45040, {{156593, 1, 0, 0},{156593, 1, 1, 17}, {0, 0, 0, 0}}},{45040, {{156593, 1, 0, 0},{156593, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{45041, {{156594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45042, {{156596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45043, {{156597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45044, {{156598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45045, {{156599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45046, {{156600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45047, {{156601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45048, {{156603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45049, {{156604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{45050, {{156609, 1, 0, 0},{156609, 1, 1, 18}, {0, 0, 0, 0}}},{45050, {{156609, 1, 0, 0},{156609, 1, 1, 18}, {0, 0, 0, 0}}},{45051, {{156609, 1, 0, 0},{156609, 1, 1, 17}, {0, 0, 0, 0}}},{45051, {{156609, 1, 0, 0},{156609, 1, 1, 17}, {0, 0, 0, 0}}},{45052, {{156609, 1, 0, 0},{156609, 1, 1, 16}, {0, 0, 0, 0}}},{45052, {{156609, 1, 0, 0},{156609, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{45053, {{156610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45054, {{156611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45055, {{156613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45056, {{156614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45057, {{156615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45058, {{156616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45059, {{156618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45060, {{156619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45061, {{156620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45062, {{156621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45063, {{156628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45064, {{156629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45065, {{156630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45066, {{156631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45067, {{156632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45068, {{156633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45069, {{156634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45070, {{156636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45071, {{156637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45072, {{156638, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45073, {{156639, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45074, {{156640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45075, {{156641, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45076, {{156642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45077, {{156643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45078, {{156644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45079, {{156645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45080, {{156646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45081, {{156647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45082, {{156648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45083, {{156649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45084, {{156650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45085, {{156651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45086, {{156652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45087, {{156654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45088, {{156655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45089, {{156656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45090, {{156657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45091, {{156658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45092, {{156659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45093, {{156660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45094, {{156661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45095, {{156662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45096, {{156663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45097, {{156664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45098, {{156665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45099, {{156666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45100, {{156667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45101, {{156668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45102, {{156672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45103, {{156674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45104, {{156675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45105, {{156676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45106, {{156677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45108, {{156678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{45109, {{156679, 1, 0, 0},{156679, 1, 1, 16}, {0, 0, 0, 0}}},{45109, {{156679, 1, 0, 0},{156679, 1, 1, 16}, {0, 0, 0, 0}}},{45110, {{156679, 1, 0, 0},{156679, 1, 1, 17}, {0, 0, 0, 0}}},{45110, {{156679, 1, 0, 0},{156679, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{45111, {{156680, 1, 0, 0},{156680, 1, 1, 17}, {0, 0, 0, 0}}},{45111, {{156680, 1, 0, 0},{156680, 1, 1, 17}, {0, 0, 0, 0}}},{45112, {{156680, 1, 0, 0},{156680, 1, 1, 16}, {0, 0, 0, 0}}},{45112, {{156680, 1, 0, 0},{156680, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{45114, {{156681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45115, {{156682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45116, {{156683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45117, {{156684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45118, {{156685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45120, {{156693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45121, {{156694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45122, {{156695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45123, {{156696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45124, {{156697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45125, {{156698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45126, {{156699, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45127, {{156702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45128, {{156705, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45129, {{156706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45130, {{156707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45131, {{156708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45132, {{156709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45133, {{156710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45134, {{156711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45135, {{156712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45136, {{156713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45137, {{156714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45138, {{156715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45139, {{156716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45140, {{156717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45141, {{156718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45142, {{156719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45143, {{156720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45144, {{156721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45145, {{156722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45146, {{156723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45147, {{156727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45148, {{156728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45149, {{156729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45150, {{156730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45151, {{156732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45152, {{156733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45153, {{156734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45154, {{156735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45155, {{156736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45156, {{156743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45157, {{156744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45158, {{156746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45159, {{156747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45160, {{156748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45161, {{156749, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45162, {{156750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45163, {{156751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45164, {{156752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45165, {{156753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45166, {{156760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45167, {{156761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45168, {{156762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45169, {{156763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45170, {{156764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45171, {{156765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45172, {{156766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45173, {{156767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45174, {{156768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45175, {{156769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{45176, {{156770, 1, 0, 0},{156770, 1, 1, 16}, {0, 0, 0, 0}}},{45176, {{156770, 1, 0, 0},{156770, 1, 1, 16}, {0, 0, 0, 0}}},{45177, {{156770, 1, 0, 0},{156770, 1, 1, 17}, {0, 0, 0, 0}}},{45177, {{156770, 1, 0, 0},{156770, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{45178, {{156771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45179, {{156772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45180, {{156773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45181, {{156774, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45182, {{156775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45183, {{156776, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45184, {{156777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45185, {{156778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45186, {{156784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45187, {{156785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45188, {{156786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45189, {{156787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45190, {{156788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45191, {{156789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45192, {{156790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45193, {{156792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45194, {{156793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45195, {{156794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45196, {{156795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45197, {{156796, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45198, {{156797, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45199, {{156798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45200, {{156803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45201, {{156804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45202, {{156805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45203, {{156806, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45204, {{156807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45205, {{156808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45206, {{156810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45207, {{156813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45208, {{156814, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45209, {{156815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45210, {{156816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45211, {{156817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45212, {{156818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45213, {{156819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45214, {{156820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45215, {{156821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45216, {{156822, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45218, {{156828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45219, {{156829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45217, {{156830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45220, {{156831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45221, {{156834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45222, {{156835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45223, {{156837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45224, {{156838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45225, {{156839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45226, {{156840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45227, {{156841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45228, {{156842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45231, {{156843, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45229, {{156844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45230, {{156846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45232, {{156848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45233, {{156849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45234, {{156850, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{45235, {{156851, 1, 0, 0},{156851, 1, 1, 17}, {0, 0, 0, 0}}},{45235, {{156851, 1, 0, 0},{156851, 1, 1, 17}, {0, 0, 0, 0}}},{58361, {{156851, 1, 0, 0},{156851, 1, 1, 16}, {0, 0, 0, 0}}},{58361, {{156851, 1, 0, 0},{156851, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{45236, {{156852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45237, {{156854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45238, {{156855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45239, {{156856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45240, {{156858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45241, {{156859, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45242, {{156862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45243, {{156863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45244, {{156864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45245, {{156866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45247, {{156868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45246, {{156870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45248, {{156872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45249, {{156875, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45250, {{156877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45251, {{156878, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45252, {{156880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45253, {{156882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45254, {{156883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45255, {{156884, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45256, {{156885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45257, {{156886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45258, {{156887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45259, {{156888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45260, {{156890, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45261, {{156893, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45262, {{156894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45263, {{156895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45264, {{156896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45265, {{156898, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45266, {{156900, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45267, {{156904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45268, {{156906, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45269, {{156908, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45270, {{156910, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45271, {{156911, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45273, {{156912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45274, {{156913, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45275, {{156915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45276, {{156916, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45277, {{156917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45278, {{156919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45279, {{156921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45280, {{156922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45281, {{156923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45282, {{156924, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45283, {{156925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45284, {{156926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45285, {{156929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45286, {{156930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45287, {{156931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45289, {{156932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45290, {{156933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45291, {{156934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45292, {{156936, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45293, {{156937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45294, {{156939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45295, {{156940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45296, {{156941, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45297, {{156942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45298, {{156943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45299, {{156944, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45300, {{156947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45301, {{156948, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45302, {{156951, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{45303, {{156952, 1, 0, 0},{156952, 1, 1, 16}, {0, 0, 0, 0}}},{45303, {{156952, 1, 0, 0},{156952, 1, 1, 16}, {0, 0, 0, 0}}},{45304, {{156952, 1, 0, 0},{156952, 1, 1, 17}, {0, 0, 0, 0}}},{45304, {{156952, 1, 0, 0},{156952, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45305, {{156955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45306, {{156956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45307, {{156957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45308, {{156958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45309, {{156959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45310, {{156960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45311, {{156961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45312, {{156962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45313, {{156965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45314, {{156966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45315, {{156968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45316, {{156969, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45317, {{156970, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45318, {{156971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45319, {{156972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45320, {{156973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58364, {{156974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45321, {{156975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45322, {{156976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45324, {{156982, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45325, {{156983, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45326, {{156984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45327, {{156985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45328, {{156988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45329, {{156989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45330, {{156990, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45331, {{156991, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45332, {{156992, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45333, {{156993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45335, {{156998, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45336, {{156999, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45337, {{157001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45338, {{157002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45339, {{157003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45340, {{157004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45341, {{157005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45342, {{157007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45343, {{157008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45344, {{157009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45345, {{157010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45346, {{157011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45347, {{157012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45348, {{157013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45349, {{157015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45350, {{157016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45351, {{157017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45352, {{157024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45353, {{157025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45354, {{157026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45356, {{157041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45355, {{157044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45357, {{157045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45358, {{157046, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45359, {{157047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45360, {{157048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45361, {{157049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45362, {{157050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45363, {{157051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45364, {{157052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45365, {{157053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45366, {{157054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45367, {{157055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45368, {{157056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45369, {{157061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45370, {{157062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45371, {{157064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45372, {{157065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45373, {{157066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45374, {{157067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45375, {{157069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45376, {{157070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45377, {{157071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45378, {{157072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45379, {{157084, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45387, {{157085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{45380, {{157090, 1, 0, 0},{157090, 1, 1, 16}, {0, 0, 0, 0}}},{45380, {{157090, 1, 0, 0},{157090, 1, 1, 16}, {0, 0, 0, 0}}},{45381, {{157090, 1, 0, 0},{157090, 1, 1, 17}, {0, 0, 0, 0}}},{45381, {{157090, 1, 0, 0},{157090, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{45382, {{157091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45383, {{157092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45384, {{157093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45385, {{157094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45386, {{157095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45388, {{157096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45389, {{157097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45390, {{157098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45391, {{157099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45392, {{157100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45393, {{157112, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45394, {{157113, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45395, {{157114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45396, {{157116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45397, {{157117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45398, {{157118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45399, {{157119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45400, {{157120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45401, {{157121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45402, {{157122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45403, {{157123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45404, {{157130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45405, {{157133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45406, {{157134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45407, {{157136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45408, {{157137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45409, {{157138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45410, {{157139, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45411, {{157140, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45412, {{157141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45413, {{157142, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45414, {{157143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45415, {{157145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45416, {{157148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45417, {{157151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45418, {{157152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45419, {{157153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45420, {{157154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45421, {{157155, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45422, {{157156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45423, {{157157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45424, {{157158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45425, {{157159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45426, {{157160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45427, {{157166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45428, {{157167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45429, {{157171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45430, {{157172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45431, {{157173, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45432, {{157174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45433, {{157175, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45434, {{157179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45435, {{157180, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45436, {{157182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45437, {{157183, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45438, {{157187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45439, {{157188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45440, {{157189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45441, {{157192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45442, {{157193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45443, {{157194, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45444, {{157195, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45445, {{157197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45446, {{157202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45447, {{157203, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45448, {{157204, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45449, {{157205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45450, {{157208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45451, {{157211, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45452, {{157213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45453, {{157215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45454, {{157216, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45455, {{157219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45456, {{157220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45457, {{157224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45458, {{157225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45459, {{157226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45460, {{157227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45461, {{157228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45462, {{157229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45463, {{157234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45464, {{157235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45465, {{157236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45466, {{157237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45467, {{157238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45468, {{157239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45469, {{157240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45470, {{157241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45471, {{157242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45472, {{157244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45473, {{157245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45474, {{157246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45475, {{157247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45476, {{157248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45477, {{157249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45478, {{157250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45479, {{157251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45480, {{157252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45481, {{157253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45482, {{157254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45483, {{157255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45484, {{157256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45485, {{157257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45486, {{157258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45487, {{157259, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45488, {{157261, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45489, {{157264, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45490, {{157271, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45491, {{157272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45492, {{157273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57684, {{157274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45493, {{157275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45494, {{157279, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45495, {{157280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45496, {{157281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45497, {{157282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45498, {{157284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45499, {{157287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45500, {{157288, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45501, {{157289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45502, {{157290, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45503, {{157291, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45504, {{157292, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45505, {{157293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45506, {{157294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45507, {{157295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45508, {{157296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45509, {{157297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45510, {{157298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45511, {{157299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45512, {{157300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45513, {{157301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45514, {{157302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45515, {{157303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45516, {{157304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45517, {{157305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45518, {{157306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45519, {{157307, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45520, {{157308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45521, {{157310, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45522, {{157312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45523, {{157313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45524, {{157315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45525, {{157320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45526, {{157321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45527, {{157324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45528, {{157325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45529, {{157326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45530, {{157327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45531, {{157329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45532, {{157330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45533, {{157331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45534, {{157336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45535, {{157337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45536, {{157338, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45537, {{157339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45538, {{157341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45539, {{157342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45540, {{157343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45541, {{157344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45542, {{157345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45543, {{157346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45544, {{157347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45545, {{157348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45546, {{157349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45547, {{157350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45548, {{157351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{45549, {{157352, 1, 0, 0},{157352, 1, 1, 16}, {0, 0, 0, 0}}},{45549, {{157352, 1, 0, 0},{157352, 1, 1, 16}, {0, 0, 0, 0}}},{45550, {{157352, 1, 0, 0},{157352, 1, 1, 17}, {0, 0, 0, 0}}},{45550, {{157352, 1, 0, 0},{157352, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45552, {{157354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45553, {{157355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45554, {{157356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45555, {{157357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45556, {{157358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{45557, {{157359, 1, 0, 0},{157359, 1, 1, 16}, {0, 0, 0, 0}}},{45557, {{157359, 1, 0, 0},{157359, 1, 1, 16}, {0, 0, 0, 0}}},{45558, {{157359, 1, 0, 0},{157359, 1, 1, 17}, {0, 0, 0, 0}}},{45558, {{157359, 1, 0, 0},{157359, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{45559, {{157360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45561, {{157369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45562, {{157370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45563, {{157371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45564, {{157377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45565, {{157378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45566, {{157379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45567, {{157380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45568, {{157381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45569, {{157390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45570, {{157391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45571, {{157392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45560, {{157393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45572, {{157394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45573, {{157395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45574, {{157396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45575, {{157397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45576, {{157398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45577, {{157399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45578, {{157400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45579, {{157401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45580, {{157402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45581, {{157403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45582, {{157404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45583, {{157405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45584, {{157406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45585, {{157407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45586, {{157408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45587, {{157409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45588, {{157411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45589, {{157412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45590, {{157413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45591, {{157414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45592, {{157415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45593, {{157417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45594, {{157419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45595, {{157420, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45596, {{157424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45597, {{157426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45598, {{157427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45599, {{157428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45600, {{157430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45601, {{157431, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45602, {{157433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45603, {{157441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45604, {{157442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45605, {{157443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45606, {{157444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45607, {{157446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45608, {{157447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45609, {{157448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{21095, {{157449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45610, {{157450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45611, {{157451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45612, {{157452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45613, {{157453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45614, {{157454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45615, {{157455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45616, {{157456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45617, {{157457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45618, {{157458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45619, {{157459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45620, {{157460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45621, {{157461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45622, {{157462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45623, {{157463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45624, {{157464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45625, {{157465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45626, {{157466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45627, {{157467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45628, {{157468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45629, {{157469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45630, {{157471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45631, {{157473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45632, {{157482, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45633, {{157484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45634, {{157488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45635, {{157491, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45636, {{157493, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45637, {{157494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45638, {{157496, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45639, {{157497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45640, {{157498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45641, {{157499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45642, {{157501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45643, {{157502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45644, {{157503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45645, {{157519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45646, {{157520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45647, {{157521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45648, {{157522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45649, {{157523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45650, {{157524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45651, {{157525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45652, {{157526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45653, {{157527, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45654, {{157528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45655, {{157530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45656, {{157531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45657, {{157532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45658, {{157533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45659, {{157534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45660, {{157535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45661, {{157536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45662, {{157537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45663, {{157538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45664, {{157539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45665, {{157540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45666, {{157541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45667, {{157542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45668, {{157543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45669, {{157544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45670, {{157545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45671, {{157546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{45672, {{157547, 1, 0, 0},{157547, 1, 1, 16}, {0, 0, 0, 0}}},{45672, {{157547, 1, 0, 0},{157547, 1, 1, 16}, {0, 0, 0, 0}}},{45673, {{157547, 1, 0, 0},{157547, 1, 1, 17}, {0, 0, 0, 0}}},{45673, {{157547, 1, 0, 0},{157547, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{45674, {{157548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45675, {{157549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45676, {{157550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45677, {{157551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45678, {{157552, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45679, {{157553, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45680, {{157554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45681, {{157555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45682, {{157556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45683, {{157557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45684, {{157558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45685, {{157559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45686, {{157560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45687, {{157569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45688, {{157572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45689, {{157573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45690, {{157577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45691, {{157579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45692, {{157589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45693, {{157590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45694, {{157591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45695, {{157593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45696, {{157594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45697, {{157595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45698, {{157596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45699, {{157597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45700, {{157598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45701, {{157599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45702, {{157600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45703, {{157601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45704, {{157602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45705, {{157603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45706, {{157604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45707, {{157605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45708, {{157606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45709, {{157607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45711, {{157608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45712, {{157609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45713, {{157610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45714, {{157611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45715, {{157612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45716, {{157613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45717, {{157614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45718, {{157615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45719, {{157616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45720, {{157617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45721, {{157618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45722, {{157622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45723, {{157632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45724, {{157633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45725, {{157635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45726, {{157652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45727, {{157653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45728, {{157656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45729, {{157657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45730, {{157659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45731, {{157660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45732, {{157661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45733, {{157662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45734, {{157663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45735, {{157664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45736, {{157665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45737, {{157666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45738, {{157667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45739, {{157668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45740, {{157669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45741, {{157670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45742, {{157671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45743, {{157673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45744, {{157679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45745, {{157680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45746, {{157683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45747, {{157696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45748, {{157697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45749, {{157699, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45750, {{157700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45751, {{157701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45752, {{157702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45753, {{157703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45754, {{157704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45755, {{157705, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45756, {{157706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45757, {{157707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45758, {{157708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45759, {{157709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45760, {{157710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45761, {{157711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45762, {{157712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45763, {{157713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45764, {{157714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45765, {{157715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45766, {{157716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45767, {{157717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45768, {{157718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45769, {{157719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45770, {{157724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45771, {{157726, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45772, {{157727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45773, {{157728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45774, {{157732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45775, {{157733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45776, {{157734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45777, {{157736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45778, {{157737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45779, {{157738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45780, {{157739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45781, {{157740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45782, {{157751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45783, {{157752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45784, {{157753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45785, {{157754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45786, {{157755, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58399, {{157756, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45788, {{157758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45789, {{157759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45790, {{157760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45791, {{157761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45792, {{157762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45793, {{157763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45794, {{157764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45795, {{157765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45796, {{157766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45797, {{157767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45798, {{157768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45799, {{157772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45800, {{157781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45801, {{157782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45802, {{157783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45803, {{157784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45804, {{157787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45805, {{157788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45806, {{157799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45807, {{157800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45808, {{157801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45809, {{157802, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45810, {{157803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45811, {{157804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45812, {{157805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45813, {{157808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45814, {{157809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45815, {{157810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{45787, {{157811, 1, 0, 0},{157811, 1, 1, 16}, {0, 0, 0, 0}}},{45787, {{157811, 1, 0, 0},{157811, 1, 1, 16}, {0, 0, 0, 0}}},{58403, {{157811, 1, 0, 0},{157811, 1, 1, 17}, {0, 0, 0, 0}}},{58403, {{157811, 1, 0, 0},{157811, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{45816, {{157812, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45817, {{157813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45818, {{157814, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45819, {{157815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45820, {{157816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45821, {{157817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45822, {{157818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45823, {{157819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45824, {{157820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45825, {{157821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45826, {{157827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45827, {{157829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45828, {{157830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45829, {{157832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45830, {{157834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45859, {{157835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45831, {{157836, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45832, {{157837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45833, {{157838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45834, {{157839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45835, {{157840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45836, {{157841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45837, {{157842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45838, {{157845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45839, {{157846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45840, {{157849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45841, {{157850, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45842, {{157851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45843, {{157852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45844, {{157853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45845, {{157861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45846, {{157862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45847, {{157863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45848, {{157864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45849, {{157865, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{45850, {{157866, 1, 0, 0},{157866, 1, 1, 16}, {0, 0, 0, 0}}},{45850, {{157866, 1, 0, 0},{157866, 1, 1, 16}, {0, 0, 0, 0}}},{45851, {{157866, 1, 0, 0},{157866, 1, 1, 17}, {0, 0, 0, 0}}},{45851, {{157866, 1, 0, 0},{157866, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{45853, {{157867, 1, 0, 0},{157867, 1, 1, 16}, {0, 0, 0, 0}}},{45853, {{157867, 1, 0, 0},{157867, 1, 1, 16}, {0, 0, 0, 0}}},{45854, {{157867, 1, 0, 0},{157867, 1, 1, 17}, {0, 0, 0, 0}}},{45854, {{157867, 1, 0, 0},{157867, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45855, {{157869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45856, {{157870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45857, {{157871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45858, {{157880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45860, {{157882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45861, {{157883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{45862, {{157884, 1, 0, 0},{157884, 1, 1, 16}, {0, 0, 0, 0}}},{45862, {{157884, 1, 0, 0},{157884, 1, 1, 16}, {0, 0, 0, 0}}},{45863, {{157884, 1, 0, 0},{157884, 1, 1, 17}, {0, 0, 0, 0}}},{45863, {{157884, 1, 0, 0},{157884, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{45864, {{157885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45865, {{157887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45866, {{157888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45867, {{157889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45868, {{157890, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45869, {{157894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45870, {{157895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45871, {{157896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45872, {{157897, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45873, {{157898, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45874, {{157899, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45875, {{157900, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45876, {{157901, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45877, {{157903, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45878, {{157904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45879, {{157905, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45880, {{157906, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45881, {{157907, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45882, {{157910, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45883, {{157914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45884, {{157915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45885, {{157916, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45886, {{157917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45887, {{157919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45888, {{157920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45889, {{157922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45890, {{157924, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45891, {{157925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45892, {{157926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45893, {{157927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45894, {{157928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45895, {{157929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45896, {{157930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45897, {{157934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45898, {{157935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45900, {{157938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45901, {{157940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45902, {{157943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45903, {{157944, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45904, {{157945, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45905, {{157946, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45906, {{157947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45907, {{157948, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45908, {{157949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45909, {{157950, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45910, {{157951, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45911, {{157953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45912, {{157954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45913, {{157955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45914, {{157956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45915, {{157958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45916, {{157959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45917, {{157960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45918, {{157961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45919, {{157962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45920, {{157964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45922, {{157965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45923, {{157968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45924, {{157970, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45925, {{157972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45926, {{157973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45927, {{157975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45928, {{157976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45929, {{157977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45930, {{157978, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45931, {{157980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{45932, {{157981, 1, 0, 0},{157981, 1, 1, 16}, {0, 0, 0, 0}}},{45932, {{157981, 1, 0, 0},{157981, 1, 1, 16}, {0, 0, 0, 0}}},{45933, {{157981, 1, 0, 0},{157981, 1, 1, 17}, {0, 0, 0, 0}}},{45933, {{157981, 1, 0, 0},{157981, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{45934, {{157982, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45935, {{157983, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45936, {{157985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45937, {{157986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45938, {{157987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45939, {{157990, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45940, {{157991, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45941, {{157992, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45942, {{157995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45943, {{157996, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45944, {{158002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45945, {{158003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45946, {{158004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45947, {{158005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45948, {{158008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45949, {{158009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45950, {{158010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45951, {{158011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{45952, {{158012, 1, 0, 0},{158012, 1, 1, 16}, {0, 0, 0, 0}}},{45952, {{158012, 1, 0, 0},{158012, 1, 1, 16}, {0, 0, 0, 0}}},{60086, {{158012, 1, 0, 0},{158012, 1, 1, 17}, {0, 0, 0, 0}}},{60086, {{158012, 1, 0, 0},{158012, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{45953, {{158013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45954, {{158015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45955, {{158017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45956, {{158018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45957, {{158020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45958, {{158021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45959, {{158022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45960, {{158024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45961, {{158025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45962, {{158026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45963, {{158027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45964, {{158029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45965, {{158030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45966, {{158031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45967, {{158032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45968, {{158033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45969, {{158034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45970, {{158035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45971, {{158036, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45972, {{158041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{45973, {{158043, 1, 0, 0},{158043, 1, 1, 17}, {0, 0, 0, 0}}},{45973, {{158043, 1, 0, 0},{158043, 1, 1, 17}, {0, 0, 0, 0}}},{58415, {{158043, 1, 0, 0},{158043, 1, 1, 16}, {0, 0, 0, 0}}},{58415, {{158043, 1, 0, 0},{158043, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45974, {{158047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45975, {{158048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45976, {{158049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45977, {{158050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45978, {{158051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45979, {{158053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45980, {{158054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45981, {{158055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45982, {{158056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45983, {{158057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45984, {{158059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45985, {{158060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45986, {{158061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45987, {{158062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45988, {{158063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45989, {{158064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45990, {{158066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{45991, {{158067, 1, 0, 0},{158067, 1, 1, 16}, {0, 0, 0, 0}}},{45991, {{158067, 1, 0, 0},{158067, 1, 1, 16}, {0, 0, 0, 0}}},{58418, {{158067, 1, 0, 0},{158067, 1, 1, 17}, {0, 0, 0, 0}}},{58418, {{158067, 1, 0, 0},{158067, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{45992, {{158068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45993, {{158069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45994, {{158070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{45995, {{158071, 1, 0, 0},{158071, 1, 1, 16}, {0, 0, 0, 0}}},{45995, {{158071, 1, 0, 0},{158071, 1, 1, 16}, {0, 0, 0, 0}}},{45996, {{158071, 1, 0, 0},{158071, 1, 1, 17}, {0, 0, 0, 0}}},{45996, {{158071, 1, 0, 0},{158071, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{45997, {{158072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45998, {{158073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{45999, {{158074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46000, {{158075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46001, {{158076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46002, {{158077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46003, {{158078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46004, {{158079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46005, {{158080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46006, {{158081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46007, {{158082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46008, {{158083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46009, {{158084, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46010, {{158085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46011, {{158086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46012, {{158087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46013, {{158088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46014, {{158089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46015, {{158090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46016, {{158092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46017, {{158095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46018, {{158096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46019, {{158097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46020, {{158099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43117, {{158100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46021, {{158101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46022, {{158103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46023, {{158104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46024, {{158105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46025, {{158106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46026, {{158107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46027, {{158108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46028, {{158109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46029, {{158110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46030, {{158114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46031, {{158115, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46032, {{158116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46033, {{158117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46034, {{158121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46035, {{158122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46036, {{158123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46037, {{158124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46038, {{158125, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46039, {{158127, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46040, {{158128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46041, {{158129, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46042, {{158130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46043, {{158134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46044, {{158137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46045, {{158138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46046, {{158140, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46047, {{158141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46048, {{158142, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46049, {{158143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46050, {{158145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46051, {{158146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46052, {{158148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46053, {{158149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46054, {{158151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46055, {{158152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46056, {{158153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46057, {{158154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46058, {{158155, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46059, {{158157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46060, {{158158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46061, {{158159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46062, {{158160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46063, {{158162, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46064, {{158163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46065, {{158164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46066, {{158165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46067, {{158166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46068, {{158168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46069, {{158169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46070, {{158170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46071, {{158171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46072, {{158173, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46073, {{158175, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{46074, {{158176, 1, 0, 0},{158176, 1, 1, 16}, {0, 0, 0, 0}}},{46074, {{158176, 1, 0, 0},{158176, 1, 1, 16}, {0, 0, 0, 0}}},{46075, {{158176, 1, 0, 0},{158176, 1, 1, 17}, {0, 0, 0, 0}}},{46075, {{158176, 1, 0, 0},{158176, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{46076, {{158177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46077, {{158178, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46078, {{158180, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46079, {{158181, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46080, {{158182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46081, {{158183, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46082, {{158185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46083, {{158186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46084, {{158187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46085, {{158189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46086, {{158190, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46087, {{158192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46088, {{158193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46089, {{158196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46090, {{158197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46091, {{158198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46092, {{158200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46093, {{158201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46094, {{158203, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46095, {{158204, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46096, {{158209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46098, {{158211, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46099, {{158212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46100, {{158213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46101, {{158214, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46102, {{158215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46103, {{158216, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46104, {{158217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46105, {{158218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46106, {{158219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46107, {{158221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46108, {{158222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46109, {{158223, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46110, {{158224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46111, {{158228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46112, {{158229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46113, {{158230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46114, {{158231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46115, {{158232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46116, {{158233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46118, {{158235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46119, {{158236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46120, {{158237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46121, {{158238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46122, {{158239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46123, {{158241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46124, {{158242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46125, {{158243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46126, {{158244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46127, {{158245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46128, {{158246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46129, {{158247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46130, {{158248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46131, {{158252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46132, {{158255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46133, {{158256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46134, {{158257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46135, {{158258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46136, {{158260, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46137, {{158261, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46138, {{158262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46139, {{158263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46140, {{158264, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46141, {{158266, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46142, {{158268, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46143, {{158269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46144, {{158270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46145, {{158271, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46147, {{158276, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46148, {{158279, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46149, {{158280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46150, {{158281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46151, {{158282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46152, {{158283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46153, {{158284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46154, {{158285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46155, {{158286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46156, {{158287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46157, {{158294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46158, {{158296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46159, {{158297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46160, {{158300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46161, {{158301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46162, {{158302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46163, {{158303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46164, {{158304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46165, {{158305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46166, {{158306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46167, {{158307, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46168, {{158308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46169, {{158309, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46170, {{158310, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46171, {{158311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46172, {{158312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46173, {{158313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46174, {{158314, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46175, {{158315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46176, {{158316, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46177, {{158317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46178, {{158319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46179, {{158321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46180, {{158322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46181, {{158324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46182, {{158325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46183, {{158326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46184, {{158327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46185, {{158328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46186, {{158329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46187, {{158330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46188, {{158331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46189, {{158332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46190, {{158333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46191, {{158340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46192, {{158341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46193, {{158342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46194, {{158343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46195, {{158344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46196, {{158345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46197, {{158346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46198, {{158347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46199, {{158348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46200, {{158349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46201, {{158350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46202, {{158353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46203, {{158354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46204, {{158355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46205, {{158356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46206, {{158357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46207, {{158358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46208, {{158359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46209, {{158360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46210, {{158361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46211, {{158362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46212, {{158363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46213, {{158364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46214, {{158365, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46215, {{158369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46216, {{158371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46217, {{158372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46218, {{158373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46219, {{158375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46220, {{158377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46221, {{158378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46222, {{158379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46223, {{158380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{46224, {{158381, 1, 0, 0},{158381, 1, 1, 17}, {0, 0, 0, 0}}},{46224, {{158381, 1, 0, 0},{158381, 1, 1, 17}, {0, 0, 0, 0}}},{46225, {{158381, 1, 0, 0},{158381, 1, 1, 16}, {0, 0, 0, 0}}},{46225, {{158381, 1, 0, 0},{158381, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{46226, {{158382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46227, {{158384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46228, {{158385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46229, {{158386, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46230, {{158390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46231, {{158391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46232, {{158392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46233, {{158393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46234, {{158397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46235, {{158398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46236, {{158399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46237, {{158400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46238, {{158401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46239, {{158402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46240, {{158403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46241, {{158404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46242, {{158405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46243, {{158406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46244, {{158407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46245, {{158408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46246, {{158410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46247, {{158413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46248, {{158414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46249, {{158418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46250, {{158419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46251, {{158420, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46252, {{158421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46253, {{158422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46254, {{158423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46255, {{158425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46256, {{158427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46257, {{158428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46258, {{158429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46259, {{158430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46260, {{158433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46261, {{158434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46262, {{158435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46263, {{158436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46264, {{158438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46265, {{158440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46266, {{158442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46267, {{158443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46268, {{158444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46269, {{158446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46270, {{158447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46271, {{158448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46272, {{158449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46273, {{158451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46274, {{158454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46275, {{158455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46276, {{158456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46278, {{158457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46279, {{158458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46280, {{158459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46281, {{158460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46282, {{158463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46283, {{158464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46284, {{158465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46285, {{158466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46286, {{158467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46287, {{158468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46288, {{158469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46289, {{158470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46290, {{158472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46291, {{158473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{21523, {{158474, 1, 0, 0},{158474, 1, 1, 16}, {0, 0, 0, 0}}},{21523, {{158474, 1, 0, 0},{158474, 1, 1, 16}, {0, 0, 0, 0}}},{46292, {{158474, 1, 0, 0},{158474, 1, 1, 17}, {0, 0, 0, 0}}},{46292, {{158474, 1, 0, 0},{158474, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46293, {{158481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46294, {{158483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46295, {{158484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46298, {{158485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46299, {{158486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46300, {{158487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46301, {{158488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46302, {{158489, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46304, {{158490, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46305, {{158491, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46306, {{158492, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46307, {{158493, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{46308, {{158494, 1, 0, 0},{158494, 1, 1, 16}, {0, 0, 0, 0}}},{46308, {{158494, 1, 0, 0},{158494, 1, 1, 16}, {0, 0, 0, 0}}},{46309, {{158494, 1, 0, 0},{158494, 1, 1, 17}, {0, 0, 0, 0}}},{46309, {{158494, 1, 0, 0},{158494, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{46310, {{158495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{46312, {{158496, 1, 0, 0},{158496, 1, 1, 16}, {0, 0, 0, 0}}},{46312, {{158496, 1, 0, 0},{158496, 1, 1, 16}, {0, 0, 0, 0}}},{46313, {{158496, 1, 0, 0},{158496, 1, 1, 17}, {0, 0, 0, 0}}},{46313, {{158496, 1, 0, 0},{158496, 1, 1, 17}, {0, 0, 0, 0}}},{46314, {{158496, 1, 0, 0},{158496, 1, 1, 18}, {0, 0, 0, 0}}},{46314, {{158496, 1, 0, 0},{158496, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{46315, {{158497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46317, {{158498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46319, {{158499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46320, {{158500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46321, {{158501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46322, {{158502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46323, {{158503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58446, {{158515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46324, {{158517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46325, {{158519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46326, {{158520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46327, {{158521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46328, {{158522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46330, {{158523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46331, {{158524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46332, {{158525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46333, {{158526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46334, {{158527, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46335, {{158528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46336, {{158529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46337, {{158530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46338, {{158531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46339, {{158532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46340, {{158533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46341, {{158534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46342, {{158535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46343, {{158536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46344, {{158537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46345, {{158538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46346, {{158539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46348, {{158557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46349, {{158562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46350, {{158568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46351, {{158569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46352, {{158571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46353, {{158573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46354, {{158574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46355, {{158575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46356, {{158576, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46357, {{158577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46358, {{158578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46359, {{158579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46361, {{158580, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46363, {{158581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46365, {{158582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46366, {{158583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46367, {{158584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46368, {{158585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46369, {{158586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46370, {{158587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46371, {{158588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46372, {{158591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46373, {{158592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46374, {{158593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46375, {{158594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46376, {{158595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46377, {{158596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46378, {{158597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46379, {{158598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46380, {{158599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46381, {{158600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46382, {{158601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46383, {{158602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46384, {{158603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46385, {{158604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46386, {{158605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46387, {{158606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46389, {{158607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46390, {{158608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46391, {{158609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46392, {{158610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46393, {{158611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46394, {{158616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46395, {{158617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{21912, {{158624, 1, 0, 0},{158624, 1, 1, 16}, {0, 0, 0, 0}}},{21912, {{158624, 1, 0, 0},{158624, 1, 1, 16}, {0, 0, 0, 0}}},{58465, {{158624, 1, 0, 0},{158624, 1, 1, 17}, {0, 0, 0, 0}}},{58465, {{158624, 1, 0, 0},{158624, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46396, {{158638, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{379, {{158646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46397, {{158647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46398, {{158648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46399, {{158649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46400, {{158650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46401, {{158651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46402, {{158652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46403, {{158653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46404, {{158654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46405, {{158655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46406, {{158656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46407, {{158657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46408, {{158658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46409, {{158659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46410, {{158660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46411, {{158661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46412, {{158662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46413, {{158663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46414, {{158664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46415, {{158665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46416, {{158666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46417, {{158667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{46418, {{158668, 1, 0, 0},{158668, 1, 1, 16}, {0, 0, 0, 0}}},{46418, {{158668, 1, 0, 0},{158668, 1, 1, 16}, {0, 0, 0, 0}}},{60109, {{158668, 1, 0, 0},{158668, 1, 1, 17}, {0, 0, 0, 0}}},{60109, {{158668, 1, 0, 0},{158668, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{46419, {{158669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46420, {{158670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46421, {{158671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46422, {{158672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46423, {{158673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46424, {{158674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46425, {{158675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46426, {{158676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46427, {{158677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46429, {{158678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46430, {{158680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46431, {{158681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46432, {{158682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46433, {{158683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46434, {{158684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46436, {{158685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46437, {{158686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46438, {{158687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46439, {{158688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46440, {{158689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46441, {{158690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46442, {{158691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46443, {{158692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46444, {{158693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46445, {{158694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46446, {{158695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46447, {{158696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46448, {{158697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46449, {{158698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46450, {{158699, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46451, {{158700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46452, {{158702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46453, {{158703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46454, {{158704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46455, {{158708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46456, {{158712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46457, {{158742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46458, {{158743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46459, {{158744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46460, {{158748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46461, {{158750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46462, {{158757, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46463, {{158759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{46464, {{158761, 1, 0, 0},{158761, 1, 1, 17}, {0, 0, 0, 0}}},{46464, {{158761, 1, 0, 0},{158761, 1, 1, 17}, {0, 0, 0, 0}}},{57282, {{158761, 1, 0, 0},{158761, 1, 1, 18}, {0, 0, 0, 0}}},{57282, {{158761, 1, 0, 0},{158761, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{46465, {{158762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46466, {{158763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46467, {{158764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46468, {{158765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46469, {{158766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46470, {{158767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46471, {{158768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46472, {{158769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46473, {{158770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46474, {{158771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46475, {{158772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46476, {{158773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46477, {{158774, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46478, {{158775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46479, {{158776, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46480, {{158777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46481, {{158778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46482, {{158780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46483, {{158781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46484, {{158782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46485, {{158783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46486, {{158784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46487, {{158785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46488, {{158786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46489, {{158787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46490, {{158788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46491, {{158789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46493, {{158790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46494, {{158791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46496, {{158793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46497, {{158794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46498, {{158795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46499, {{158796, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46500, {{158797, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46501, {{158798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46502, {{158800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46503, {{158801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46504, {{158802, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46505, {{158803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46506, {{158804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46507, {{158805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46508, {{158806, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46509, {{158807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46510, {{158808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46511, {{158809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46512, {{158810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46514, {{158811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46515, {{158812, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46516, {{158814, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46517, {{158815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46518, {{158816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46519, {{158817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46520, {{158818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46521, {{158819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{46522, {{158820, 1, 0, 0},{158820, 1, 1, 16}, {0, 0, 0, 0}}},{46522, {{158820, 1, 0, 0},{158820, 1, 1, 16}, {0, 0, 0, 0}}},{46523, {{158820, 1, 0, 0},{158820, 1, 1, 17}, {0, 0, 0, 0}}},{46523, {{158820, 1, 0, 0},{158820, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{46524, {{158821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46525, {{158822, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46526, {{158823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46527, {{158824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46528, {{158825, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46529, {{158826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46530, {{158827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46531, {{158828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46532, {{158830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46533, {{158832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46534, {{158833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46535, {{158834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{46536, {{158835, 1, 0, 0},{158835, 1, 1, 18}, {0, 0, 0, 0}}},{46536, {{158835, 1, 0, 0},{158835, 1, 1, 18}, {0, 0, 0, 0}}},{46537, {{158835, 1, 0, 0},{158835, 1, 1, 17}, {0, 0, 0, 0}}},{46537, {{158835, 1, 0, 0},{158835, 1, 1, 17}, {0, 0, 0, 0}}},{46538, {{158835, 1, 0, 0},{158835, 1, 1, 19}, {0, 0, 0, 0}}},{46538, {{158835, 1, 0, 0},{158835, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{46539, {{158836, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46540, {{158837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46541, {{158838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46542, {{158839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46543, {{158840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46544, {{158841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46545, {{158842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46546, {{158843, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46547, {{158844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46549, {{158851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46550, {{158854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46551, {{158855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46552, {{158856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60117, {{158878, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46553, {{158884, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46554, {{158890, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46555, {{158892, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46556, {{158895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46557, {{158896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46558, {{158904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46559, {{158913, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46560, {{158914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46561, {{158915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46562, {{158916, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46563, {{158917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46564, {{158919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46565, {{158920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46566, {{158921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46567, {{158922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46568, {{158923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46569, {{158924, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46570, {{158926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46571, {{158927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46572, {{158928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46573, {{158929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46574, {{158930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46575, {{158931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46576, {{158932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46577, {{158933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46578, {{158934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46579, {{158935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46580, {{158936, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46581, {{158937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46582, {{158938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46583, {{158939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46584, {{158940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{46585, {{158941, 1, 0, 0},{158941, 1, 1, 17}, {0, 0, 0, 0}}},{46585, {{158941, 1, 0, 0},{158941, 1, 1, 17}, {0, 0, 0, 0}}},{46586, {{158941, 1, 0, 0},{158941, 1, 1, 18}, {0, 0, 0, 0}}},{46586, {{158941, 1, 0, 0},{158941, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{46587, {{158942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46588, {{158943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46589, {{158944, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46590, {{158945, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46591, {{158946, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46592, {{158947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46593, {{158948, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46594, {{158949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46595, {{158950, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46596, {{158951, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46597, {{158953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46598, {{158954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46599, {{158955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46600, {{158956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46601, {{158957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46602, {{158958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46603, {{158960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46604, {{158961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46605, {{158962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46606, {{158963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46607, {{158964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46608, {{158965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46609, {{158966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46610, {{158967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46611, {{158968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46612, {{158969, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46613, {{158970, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46614, {{158971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46615, {{158972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46616, {{158973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46617, {{158974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46618, {{158975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46619, {{158976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46620, {{158977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46621, {{158981, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46622, {{158982, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46623, {{158983, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46624, {{158984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46625, {{158985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46627, {{158986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46628, {{158987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46629, {{158988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46630, {{158989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46631, {{158990, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46632, {{158991, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46633, {{158992, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46634, {{158993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46635, {{158994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46636, {{158995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46637, {{158996, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46638, {{158997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46639, {{158999, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46640, {{159000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46641, {{159001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46642, {{159002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46643, {{159003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46644, {{159006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46645, {{159007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46646, {{159008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46648, {{159022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46649, {{159024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46650, {{159025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46651, {{159026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46652, {{159027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46653, {{159035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46654, {{159036, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46655, {{159056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46656, {{159058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46657, {{159069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46658, {{159071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58499, {{159072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46659, {{159073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46661, {{159074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46662, {{159075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46663, {{159077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46665, {{159078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46666, {{159079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46667, {{159080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46668, {{159081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46669, {{159082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46670, {{159083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46671, {{159084, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46672, {{159085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46673, {{159086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46674, {{159087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46675, {{159089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46676, {{159090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46677, {{159091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46678, {{159092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46679, {{159093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46680, {{159094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46681, {{159095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46682, {{159096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46683, {{159097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46684, {{159098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46685, {{159099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46686, {{159100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46687, {{159101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46688, {{159102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46689, {{159103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46690, {{159104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46692, {{159105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46693, {{159106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46694, {{159108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46695, {{159109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46697, {{159110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46698, {{159111, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46699, {{159112, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46700, {{159113, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46701, {{159114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46702, {{159115, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46703, {{159116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46704, {{159117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46705, {{159118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46706, {{159119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46707, {{159120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46708, {{159121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46709, {{159122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46710, {{159123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46711, {{159124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46712, {{159125, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46713, {{159126, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46714, {{159127, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46715, {{159128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46716, {{159130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46748, {{159131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46718, {{159145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46719, {{159146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46720, {{159147, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46721, {{159148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46722, {{159149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46723, {{159150, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46724, {{159151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46725, {{159152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46726, {{159179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46727, {{159187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46728, {{159188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46729, {{159189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46730, {{159191, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46731, {{159192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46732, {{159193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46733, {{159194, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46734, {{159195, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46735, {{159196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46736, {{159197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46737, {{159198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46738, {{159199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46739, {{159200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46740, {{159201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46741, {{159203, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46743, {{159217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46744, {{159218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46745, {{159219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46746, {{159220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46747, {{159221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46717, {{159222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46749, {{159223, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46750, {{159225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46751, {{159226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46752, {{159227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46753, {{159229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46754, {{159230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46755, {{159231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46756, {{159232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46757, {{159233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46758, {{159234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46759, {{159235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46760, {{159236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46761, {{159238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46762, {{159239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46763, {{159240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46764, {{159241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46765, {{159242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46766, {{159243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46767, {{159244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46768, {{159245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46769, {{159246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46770, {{159247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46771, {{159248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{46772, {{159249, 1, 0, 0},{159249, 1, 1, 16}, {0, 0, 0, 0}}},{46772, {{159249, 1, 0, 0},{159249, 1, 1, 16}, {0, 0, 0, 0}}},{46773, {{159249, 1, 0, 0},{159249, 1, 1, 17}, {0, 0, 0, 0}}},{46773, {{159249, 1, 0, 0},{159249, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{46774, {{159250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46775, {{159251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46776, {{159252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46777, {{159253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46778, {{159254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46779, {{159255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46780, {{159256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46781, {{159257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46782, {{159258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46783, {{159259, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46784, {{159260, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46785, {{159261, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46786, {{159262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46787, {{159263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46788, {{159264, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46789, {{159265, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46790, {{159266, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46791, {{159267, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46792, {{159268, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46793, {{159270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46794, {{159271, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46795, {{159272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46796, {{159273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46797, {{159274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46798, {{159275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46799, {{159276, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46800, {{159277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46801, {{159278, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46802, {{159279, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46803, {{159280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46804, {{159281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46805, {{159282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46806, {{159283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46807, {{159284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46808, {{159285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46809, {{159286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46810, {{159287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46811, {{159288, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46812, {{159289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46813, {{159290, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{46814, {{159296, 1, 0, 0},{159296, 1, 1, 17}, {0, 0, 0, 0}}},{46814, {{159296, 1, 0, 0},{159296, 1, 1, 17}, {0, 0, 0, 0}}},{46815, {{159296, 1, 0, 0},{159296, 1, 1, 18}, {0, 0, 0, 0}}},{46815, {{159296, 1, 0, 0},{159296, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46817, {{159302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{46818, {{159303, 1, 0, 0},{159303, 1, 1, 16}, {0, 0, 0, 0}}},{46818, {{159303, 1, 0, 0},{159303, 1, 1, 16}, {0, 0, 0, 0}}},{46819, {{159303, 1, 0, 0},{159303, 1, 1, 17}, {0, 0, 0, 0}}},{46819, {{159303, 1, 0, 0},{159303, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{46820, {{159304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46821, {{159305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46822, {{159306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46823, {{159308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46824, {{159309, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46825, {{159325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46826, {{159328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46827, {{159329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46828, {{159330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46829, {{159331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46830, {{159332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46831, {{159333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46833, {{159334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46834, {{159336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46835, {{159337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46836, {{159338, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46837, {{159357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46838, {{159358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46840, {{159359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46841, {{159360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46842, {{159361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46843, {{159362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46844, {{159366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46845, {{159367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46846, {{159368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46847, {{159369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46848, {{159370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46849, {{159371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46850, {{159372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46851, {{159373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46852, {{159374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46853, {{159375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46854, {{159376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46855, {{159377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46856, {{159378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46857, {{159379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46858, {{159380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46859, {{159381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46860, {{159382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46861, {{159383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46862, {{159385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46863, {{159386, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46864, {{159387, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46865, {{159388, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46866, {{159389, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46867, {{159390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46868, {{159391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46869, {{159392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46870, {{159393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46871, {{159394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46872, {{159395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46873, {{159396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46874, {{159397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46875, {{159398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46876, {{159399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46877, {{159400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46878, {{159401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46879, {{159402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46880, {{159403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46881, {{159404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46882, {{159405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46883, {{159406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46884, {{159407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46885, {{159408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46886, {{159409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46887, {{159410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46888, {{159411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46889, {{159412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46890, {{159413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46891, {{159414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46892, {{159415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46893, {{159416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46894, {{159417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46895, {{159418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46896, {{159419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46897, {{159420, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46898, {{159421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46899, {{159422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46900, {{159423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46901, {{159424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46902, {{159425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46903, {{159426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46904, {{159427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46905, {{159428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46906, {{159429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46907, {{159430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46908, {{159431, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46909, {{159432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46910, {{159433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46911, {{159434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46913, {{159435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46914, {{159436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46915, {{159437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46916, {{159447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46917, {{159453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46919, {{159455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46920, {{159456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46921, {{159457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46922, {{159458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46923, {{159459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46924, {{159460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46925, {{159461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46926, {{159462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46927, {{159463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46928, {{159464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46929, {{159465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46930, {{159466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46931, {{159494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46932, {{159495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46933, {{159496, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46934, {{159498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46935, {{159499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46936, {{159500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46937, {{159502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46938, {{159503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46939, {{159504, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46940, {{159505, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46941, {{159506, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46942, {{159513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46943, {{159520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46944, {{159521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46945, {{159522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46918, {{159523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46946, {{159524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46947, {{159529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46948, {{159530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46949, {{159531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{46950, {{159532, 1, 0, 0},{159532, 1, 1, 16}, {0, 0, 0, 0}}},{46950, {{159532, 1, 0, 0},{159532, 1, 1, 16}, {0, 0, 0, 0}}},{46951, {{159532, 1, 0, 0},{159532, 1, 1, 17}, {0, 0, 0, 0}}},{46951, {{159532, 1, 0, 0},{159532, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{46952, {{159533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46953, {{159534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{46954, {{159535, 1, 0, 0},{159535, 1, 1, 17}, {0, 0, 0, 0}}},{46954, {{159535, 1, 0, 0},{159535, 1, 1, 17}, {0, 0, 0, 0}}},{46955, {{159535, 1, 0, 0},{159535, 1, 1, 16}, {0, 0, 0, 0}}},{46955, {{159535, 1, 0, 0},{159535, 1, 1, 16}, {0, 0, 0, 0}}},{46956, {{159535, 1, 0, 0},{159535, 1, 1, 18}, {0, 0, 0, 0}}},{46956, {{159535, 1, 0, 0},{159535, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{46957, {{159536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46958, {{159537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46959, {{159538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46960, {{159539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46961, {{159540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46962, {{159541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46963, {{159542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46964, {{159543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46965, {{159544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46966, {{159545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46967, {{159546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46968, {{159547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46969, {{159548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46970, {{159549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46971, {{159550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46972, {{159551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46973, {{159552, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46974, {{159553, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46975, {{159554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46976, {{159555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46977, {{159556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46978, {{159557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46979, {{159558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46980, {{159559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46981, {{159560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46982, {{159561, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46983, {{159562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46984, {{159563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47014, {{159564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46985, {{159566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46986, {{159567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46987, {{159568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46988, {{159569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46989, {{159570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46990, {{159571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46991, {{159572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46992, {{159573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46993, {{159574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46994, {{159575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46995, {{159576, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46996, {{159577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46997, {{159578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46998, {{159579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{46999, {{159580, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47000, {{159581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47001, {{159582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47002, {{159583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47003, {{159584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47004, {{159585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47006, {{159586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47007, {{159587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47008, {{159588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47009, {{159589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47010, {{159590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47011, {{159591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47012, {{159592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47013, {{159593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47015, {{159595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47016, {{159596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47017, {{159597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47018, {{159598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47019, {{159599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47020, {{159601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47021, {{159603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47022, {{159605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47023, {{159607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47024, {{159608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47025, {{159609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47026, {{159611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{47027, {{159631, 1, 0, 0},{159631, 1, 1, 16}, {0, 0, 0, 0}}},{47027, {{159631, 1, 0, 0},{159631, 1, 1, 16}, {0, 0, 0, 0}}},{47028, {{159631, 1, 0, 0},{159631, 1, 1, 17}, {0, 0, 0, 0}}},{47028, {{159631, 1, 0, 0},{159631, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{47029, {{159632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47030, {{159634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58542, {{159635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{47031, {{159636, 1, 0, 0},{159636, 1, 1, 18}, {0, 0, 0, 0}}},{47031, {{159636, 1, 0, 0},{159636, 1, 1, 18}, {0, 0, 0, 0}}},{47032, {{159636, 1, 0, 0},{159636, 1, 1, 17}, {0, 0, 0, 0}}},{47032, {{159636, 1, 0, 0},{159636, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47033, {{159639, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47034, {{159654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47035, {{159659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47036, {{159660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47037, {{159661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47038, {{159662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47039, {{159663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47040, {{159664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{47041, {{159665, 1, 0, 0},{159665, 1, 1, 16}, {0, 0, 0, 0}}},{47041, {{159665, 1, 0, 0},{159665, 1, 1, 16}, {0, 0, 0, 0}}},{47042, {{159665, 1, 0, 0},{159665, 1, 1, 17}, {0, 0, 0, 0}}},{47042, {{159665, 1, 0, 0},{159665, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{47043, {{159666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47044, {{159667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47045, {{159668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47046, {{159669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47047, {{159670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47048, {{159671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47049, {{159672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47050, {{159673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47051, {{159674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47052, {{159675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47053, {{159676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47054, {{159677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47055, {{159678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47056, {{159679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47057, {{159680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47058, {{159681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47060, {{159682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47061, {{159683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47062, {{159684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47064, {{159685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47066, {{159686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47067, {{159688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47068, {{159689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47069, {{159690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47070, {{159691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47071, {{159692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47072, {{159693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47073, {{159694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47074, {{159695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47075, {{159696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47076, {{159697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47077, {{159698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47078, {{159699, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{47079, {{159700, 1, 0, 0},{159700, 1, 1, 16}, {0, 0, 0, 0}}},{47079, {{159700, 1, 0, 0},{159700, 1, 1, 16}, {0, 0, 0, 0}}},{47080, {{159700, 1, 0, 0},{159700, 1, 1, 17}, {0, 0, 0, 0}}},{47080, {{159700, 1, 0, 0},{159700, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{47081, {{159701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47082, {{159702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47083, {{159703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47084, {{159704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47085, {{159705, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47086, {{159712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47087, {{159714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47088, {{159715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47089, {{159716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47090, {{159717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47091, {{159720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47092, {{159721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47093, {{159722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47094, {{159734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47096, {{159735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{47097, {{159736, 1, 0, 0},{159736, 1, 1, 17}, {0, 0, 0, 0}}},{47097, {{159736, 1, 0, 0},{159736, 1, 1, 17}, {0, 0, 0, 0}}},{68087, {{159736, 1, 0, 0},{159736, 1, 1, 18}, {0, 0, 0, 0}}},{68087, {{159736, 1, 0, 0},{159736, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47098, {{159747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47099, {{159748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47100, {{159749, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47101, {{159750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47102, {{159751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47103, {{159754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47104, {{159757, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47106, {{159764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47107, {{159765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47108, {{159766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47109, {{159767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47110, {{159768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47111, {{159769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47112, {{159770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47113, {{159771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47114, {{159773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47115, {{159776, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47116, {{159777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47117, {{159778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47118, {{159779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47119, {{159780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47120, {{159781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47121, {{159782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47122, {{159783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47123, {{159784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47124, {{159785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47125, {{159786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47126, {{159787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47127, {{159788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47128, {{159789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47129, {{159790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47130, {{159791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47132, {{159792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47133, {{159793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47134, {{159794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47135, {{159795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47136, {{159796, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47137, {{159797, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47138, {{159799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47140, {{159800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{47141, {{159801, 1, 0, 0},{159801, 1, 1, 17}, {0, 0, 0, 0}}},{47141, {{159801, 1, 0, 0},{159801, 1, 1, 17}, {0, 0, 0, 0}}},{47142, {{159801, 1, 0, 0},{159801, 1, 1, 16}, {0, 0, 0, 0}}},{47142, {{159801, 1, 0, 0},{159801, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{47143, {{159802, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47144, {{159803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47145, {{159804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47146, {{159805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47147, {{159806, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47148, {{159807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47149, {{159808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47150, {{159809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47151, {{159810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47152, {{159811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47153, {{159812, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47154, {{159813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47155, {{159814, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47156, {{159815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47157, {{159816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47158, {{159817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47159, {{159818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47160, {{159819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47161, {{159820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47162, {{159821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47163, {{159822, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47164, {{159823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47165, {{159825, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47166, {{159828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47167, {{159829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47168, {{159830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47169, {{159831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47170, {{159832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{23176, {{159848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47171, {{159856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47172, {{159867, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47173, {{159868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47174, {{159869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47175, {{159870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47176, {{159871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47177, {{159872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47178, {{159873, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47179, {{159874, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47180, {{159875, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47181, {{159876, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47182, {{159877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47183, {{159878, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47184, {{159879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47185, {{159880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47187, {{159881, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47188, {{159882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47189, {{159883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47190, {{159884, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47191, {{159885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47192, {{159886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47193, {{159887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47194, {{159888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47195, {{159889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47196, {{159890, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47197, {{159891, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47198, {{159892, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47199, {{159893, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47200, {{159894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47201, {{159895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47202, {{159896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47203, {{159897, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47204, {{159898, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47205, {{159899, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47206, {{159900, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47207, {{159902, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47208, {{159903, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47209, {{159904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47210, {{159905, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47211, {{159906, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47212, {{159907, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47213, {{159909, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47214, {{159911, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47215, {{159912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47216, {{159913, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47217, {{159914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47218, {{159915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47219, {{159920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47220, {{159921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47221, {{159922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47222, {{159923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47223, {{159924, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47224, {{159925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47225, {{159945, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47226, {{159950, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47227, {{159951, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47228, {{159952, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47229, {{159953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47230, {{159954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47231, {{159955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47232, {{159956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47233, {{159957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47234, {{159958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47235, {{159959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47236, {{159960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47237, {{159961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47238, {{159962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47239, {{159963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47240, {{159965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47241, {{159966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47242, {{159967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47243, {{159968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47244, {{159969, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47245, {{159970, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47246, {{159971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47247, {{159972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47248, {{159973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47249, {{159974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47250, {{159975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47251, {{159976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47252, {{159977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47253, {{159978, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47254, {{159979, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47255, {{159980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47256, {{159982, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47257, {{159984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{47258, {{159988, 1, 0, 0},{159988, 1, 1, 18}, {0, 0, 0, 0}}},{47258, {{159988, 1, 0, 0},{159988, 1, 1, 18}, {0, 0, 0, 0}}},{47259, {{159988, 1, 0, 0},{159988, 1, 1, 17}, {0, 0, 0, 0}}},{47259, {{159988, 1, 0, 0},{159988, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47260, {{159994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47261, {{159995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47262, {{159996, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47263, {{160011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{47264, {{160013, 1, 0, 0},{160013, 1, 1, 17}, {0, 0, 0, 0}}},{47264, {{160013, 1, 0, 0},{160013, 1, 1, 17}, {0, 0, 0, 0}}},{47265, {{160013, 1, 0, 0},{160013, 1, 1, 18}, {0, 0, 0, 0}}},{47265, {{160013, 1, 0, 0},{160013, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{47266, {{160014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47267, {{160015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47268, {{160016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47269, {{160017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47270, {{160018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47272, {{160019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47273, {{160020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47274, {{160021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47275, {{160022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47276, {{160023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47277, {{160024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47278, {{160025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47279, {{160026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47280, {{160027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47281, {{160028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47282, {{160029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47283, {{160030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47284, {{160031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47285, {{160032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47286, {{160033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47287, {{160035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47288, {{160036, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47289, {{160041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47290, {{160042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47291, {{160043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47292, {{160054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47293, {{160055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47294, {{160056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{47295, {{160057, 1, 0, 0},{160057, 1, 1, 17}, {0, 0, 0, 0}}},{47295, {{160057, 1, 0, 0},{160057, 1, 1, 17}, {0, 0, 0, 0}}},{47296, {{160057, 1, 0, 0},{160057, 1, 1, 18}, {0, 0, 0, 0}}},{47296, {{160057, 1, 0, 0},{160057, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{47297, {{160058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47298, {{160059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47299, {{160060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47300, {{160061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47301, {{160062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47302, {{160063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47303, {{160064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47304, {{160065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47305, {{160066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47306, {{160067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47307, {{160068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47308, {{160069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47309, {{160070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47310, {{160071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47311, {{160072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47312, {{160073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47313, {{160074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47314, {{160075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47315, {{160077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47316, {{160078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47317, {{160079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47318, {{160083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47319, {{160086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47320, {{160087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47321, {{160088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47322, {{160089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47323, {{160090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47324, {{160091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47325, {{160092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47326, {{160093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47327, {{160094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47328, {{160095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47329, {{160096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47330, {{160097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47331, {{160098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47332, {{160099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47333, {{160105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47334, {{160106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47335, {{160110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47336, {{160111, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47337, {{160112, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{47340, {{160113, 1, 0, 0},{160113, 1, 1, 16}, {0, 0, 0, 0}}},{47340, {{160113, 1, 0, 0},{160113, 1, 1, 16}, {0, 0, 0, 0}}},{58573, {{160113, 1, 0, 0},{160113, 1, 1, 17}, {0, 0, 0, 0}}},{58573, {{160113, 1, 0, 0},{160113, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{47341, {{160114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47342, {{160115, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47343, {{160116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47344, {{160120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47345, {{160122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47346, {{160123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47347, {{160124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47348, {{160125, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47349, {{160129, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47350, {{160130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47351, {{160131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47352, {{160134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47353, {{160135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47354, {{160136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47355, {{160137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47356, {{160138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47357, {{160139, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47358, {{160143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47359, {{160144, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47360, {{160145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47361, {{160146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47362, {{160147, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47363, {{160149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47364, {{160150, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47365, {{160151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47366, {{160152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47367, {{160153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47369, {{160154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47371, {{160156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47372, {{160157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47373, {{160158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47374, {{160161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47375, {{160162, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47376, {{160163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47377, {{160164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47378, {{160165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47379, {{160166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47380, {{160167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47381, {{160168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47382, {{160169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47383, {{160170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47384, {{160171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47385, {{160172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47386, {{160175, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47387, {{160177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47388, {{160179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47389, {{160184, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47390, {{160185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47391, {{160186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47392, {{160187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47393, {{160188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47394, {{160189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47395, {{160190, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47396, {{160193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47397, {{160194, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47398, {{160195, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47399, {{160196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47400, {{160197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47401, {{160198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47402, {{160199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47403, {{160200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47404, {{160205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47412, {{160207, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47405, {{160208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47406, {{160209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47407, {{160210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47408, {{160211, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47411, {{160212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47413, {{160216, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47414, {{160219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47415, {{160220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47416, {{160221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47417, {{160222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47418, {{160223, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47419, {{160224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47420, {{160225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47421, {{160226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47422, {{160227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47423, {{160228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47424, {{160229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47425, {{160230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47426, {{160231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47427, {{160237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47428, {{160240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47429, {{160242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47430, {{160247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47431, {{160251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47432, {{160252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{47433, {{160253, 1, 0, 0},{160253, 1, 1, 16}, {0, 0, 0, 0}}},{47433, {{160253, 1, 0, 0},{160253, 1, 1, 16}, {0, 0, 0, 0}}},{47434, {{160253, 1, 0, 0},{160253, 1, 1, 17}, {0, 0, 0, 0}}},{47434, {{160253, 1, 0, 0},{160253, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{47435, {{160254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47436, {{160255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47437, {{160256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47438, {{160257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47439, {{160262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47440, {{160263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47441, {{160264, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47442, {{160265, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47443, {{160267, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47444, {{160268, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47445, {{160269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47446, {{160270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47447, {{160272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47448, {{160273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47449, {{160274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47450, {{160275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47451, {{160277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47452, {{160278, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47453, {{160279, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47454, {{160280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47455, {{160284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47456, {{160285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47457, {{160287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47458, {{160288, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47459, {{160294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47460, {{160295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47461, {{160296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47462, {{160297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{47463, {{160298, 1, 0, 0},{160298, 1, 1, 16}, {0, 0, 0, 0}}},{47463, {{160298, 1, 0, 0},{160298, 1, 1, 16}, {0, 0, 0, 0}}},{47464, {{160298, 1, 0, 0},{160298, 1, 1, 17}, {0, 0, 0, 0}}},{47464, {{160298, 1, 0, 0},{160298, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{47465, {{160299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47466, {{160300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47467, {{160301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47468, {{160303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47469, {{160304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58598, {{160312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47470, {{160313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47471, {{160314, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47472, {{160315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47473, {{160317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47474, {{160318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47475, {{160319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47476, {{160320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47477, {{160321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47478, {{160322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47479, {{160323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47480, {{160324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47481, {{160325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47482, {{160326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47483, {{160327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47484, {{160328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47485, {{160329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47486, {{160330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47487, {{160331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47488, {{160332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47489, {{160333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47490, {{160334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47491, {{160335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47492, {{160336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47493, {{160337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47494, {{160340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47495, {{160343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47496, {{160344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47497, {{160347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47498, {{160348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47499, {{160350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47500, {{160351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47501, {{160352, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47502, {{160353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47503, {{160354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47504, {{160355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47505, {{160356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47506, {{160357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47507, {{160359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47508, {{160360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47509, {{160361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47510, {{160362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47511, {{160363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47512, {{160364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47513, {{160365, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47514, {{160366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47515, {{160367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47516, {{160368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47517, {{160369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47518, {{160371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47519, {{160372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47520, {{160373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47521, {{160374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47522, {{160385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47523, {{160388, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47524, {{160389, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47525, {{160390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47526, {{160391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47527, {{160397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47528, {{160398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47529, {{160399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47530, {{160400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47531, {{160402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47532, {{160403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47533, {{160404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47534, {{160405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47535, {{160406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47536, {{160407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47537, {{160408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47538, {{160409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47539, {{160410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47540, {{160411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{47541, {{160412, 1, 0, 0},{160412, 1, 1, 16}, {0, 0, 0, 0}}},{47541, {{160412, 1, 0, 0},{160412, 1, 1, 16}, {0, 0, 0, 0}}},{47542, {{160412, 1, 0, 0},{160412, 1, 1, 17}, {0, 0, 0, 0}}},{47542, {{160412, 1, 0, 0},{160412, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{47543, {{160413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47544, {{160414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47545, {{160415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47546, {{160416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47547, {{160417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47548, {{160419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47549, {{160420, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47550, {{160421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47551, {{160422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47552, {{160423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47553, {{160424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47554, {{160425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47555, {{160426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47556, {{160427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47557, {{160428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47558, {{160429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47559, {{160430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47560, {{160431, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47561, {{160432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47562, {{160433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47563, {{160434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{47564, {{160439, 1, 0, 0},{160439, 1, 1, 16}, {0, 0, 0, 0}}},{47564, {{160439, 1, 0, 0},{160439, 1, 1, 16}, {0, 0, 0, 0}}},{47565, {{160439, 1, 0, 0},{160439, 1, 1, 17}, {0, 0, 0, 0}}},{47565, {{160439, 1, 0, 0},{160439, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{47566, {{160440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47567, {{160445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47568, {{160446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47569, {{160447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47570, {{160456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47571, {{160457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47572, {{160458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47573, {{160459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47574, {{160460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47575, {{160461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47576, {{160462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47577, {{160463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47578, {{160464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47579, {{160465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47580, {{160466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47581, {{160467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47582, {{160468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47583, {{160469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47584, {{160470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47585, {{160471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47586, {{160472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47587, {{160473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47588, {{160474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47589, {{160475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47590, {{160476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{47591, {{160477, 1, 0, 0},{160477, 1, 1, 16}, {0, 0, 0, 0}}},{47591, {{160477, 1, 0, 0},{160477, 1, 1, 16}, {0, 0, 0, 0}}},{58636, {{160477, 1, 0, 0},{160477, 1, 1, 17}, {0, 0, 0, 0}}},{58636, {{160477, 1, 0, 0},{160477, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{47592, {{160478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47593, {{160479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47594, {{160480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47595, {{160481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47596, {{160485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47597, {{160488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47598, {{160492, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47599, {{160495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47601, {{160497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47602, {{160498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47603, {{160499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47604, {{160513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47605, {{160514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47606, {{160515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47607, {{160516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47608, {{160517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47609, {{160518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47610, {{160519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47611, {{160520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47612, {{160521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{47613, {{160522, 1, 0, 0},{160522, 1, 1, 16}, {0, 0, 0, 0}}},{47613, {{160522, 1, 0, 0},{160522, 1, 1, 16}, {0, 0, 0, 0}}},{47614, {{160522, 1, 0, 0},{160522, 1, 1, 17}, {0, 0, 0, 0}}},{47614, {{160522, 1, 0, 0},{160522, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{47615, {{160523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47616, {{160524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47617, {{160525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47618, {{160526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47619, {{160527, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47620, {{160528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47621, {{160529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47622, {{160530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47623, {{160531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47624, {{160532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47625, {{160533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47626, {{160534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47627, {{160535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47628, {{160536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47629, {{160537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47630, {{160538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47631, {{160539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47632, {{160540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47633, {{160541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47634, {{160543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47635, {{160544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47636, {{160545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47637, {{160546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47638, {{160547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47639, {{160548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47640, {{160549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47641, {{160550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47642, {{160551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47643, {{160552, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47644, {{160553, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47645, {{160554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47646, {{160555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47647, {{160556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47648, {{160557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47649, {{160564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47650, {{160565, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47651, {{160571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47652, {{160572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47653, {{160573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47654, {{160574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47655, {{160575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47656, {{160576, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47657, {{160577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47658, {{160578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47659, {{160593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47660, {{160594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47661, {{160595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47662, {{160596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47664, {{160597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47665, {{160598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47666, {{160599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47668, {{160600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47669, {{160601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47670, {{160602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47671, {{160603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47672, {{160604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47673, {{160605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47674, {{160606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47675, {{160607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47676, {{160608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47677, {{160609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47678, {{160610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47679, {{160611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47680, {{160612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47681, {{160613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47682, {{160614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47683, {{160615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{47684, {{160617, 1, 0, 0},{160617, 1, 1, 16}, {0, 0, 0, 0}}},{47684, {{160617, 1, 0, 0},{160617, 1, 1, 16}, {0, 0, 0, 0}}},{47772, {{160617, 1, 0, 0},{160617, 1, 1, 17}, {0, 0, 0, 0}}},{47772, {{160617, 1, 0, 0},{160617, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{47685, {{160618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47686, {{160619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47687, {{160620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47688, {{160621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47689, {{160622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47690, {{160623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47691, {{160624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47692, {{160625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47693, {{160626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47694, {{160627, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47695, {{160628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47696, {{160629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47697, {{160630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47698, {{160631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47699, {{160632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47700, {{160633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47701, {{160634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47702, {{160635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47703, {{160636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47704, {{160637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47705, {{160638, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47706, {{160639, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47707, {{160640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47708, {{160641, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47709, {{160642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47710, {{160643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47711, {{160644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47712, {{160661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47713, {{160662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47714, {{160663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47715, {{160664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47716, {{160665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47718, {{160670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47719, {{160673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47720, {{160674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47721, {{160675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47722, {{160676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47723, {{160677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47724, {{160678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47725, {{160680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47726, {{160681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47727, {{160683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47728, {{160684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47729, {{160685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47730, {{160686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47731, {{160687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47732, {{160688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47733, {{160689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47734, {{160690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47735, {{160691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47736, {{160692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47737, {{160693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47738, {{160694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47739, {{160695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47740, {{160696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47741, {{160697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47742, {{160698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47743, {{160699, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47744, {{160700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47745, {{160701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47746, {{160702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47747, {{160703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47748, {{160704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47749, {{160705, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47750, {{160706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47751, {{160707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47752, {{160708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47753, {{160709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47754, {{160710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47755, {{160711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47756, {{160712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{47757, {{160714, 1, 0, 0},{160714, 1, 1, 16}, {0, 0, 0, 0}}},{47757, {{160714, 1, 0, 0},{160714, 1, 1, 16}, {0, 0, 0, 0}}},{47758, {{160714, 1, 0, 0},{160714, 1, 1, 17}, {0, 0, 0, 0}}},{47758, {{160714, 1, 0, 0},{160714, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{47759, {{160715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47760, {{160716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47761, {{160717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47762, {{160718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47763, {{160719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47764, {{160720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47765, {{160721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47766, {{160722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47767, {{160723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47768, {{160724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47769, {{160725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47770, {{160730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47771, {{160731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47773, {{160733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47775, {{160740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47776, {{160741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47777, {{160742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47778, {{160743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47779, {{160744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47780, {{160745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47781, {{160746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47782, {{160747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47783, {{160748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47784, {{160760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47785, {{160761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47786, {{160762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47787, {{160763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47788, {{160764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47789, {{160765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{47790, {{160766, 1, 0, 0},{160766, 1, 1, 17}, {0, 0, 0, 0}}},{47790, {{160766, 1, 0, 0},{160766, 1, 1, 17}, {0, 0, 0, 0}}},{47791, {{160766, 1, 0, 0},{160766, 1, 1, 18}, {0, 0, 0, 0}}},{47791, {{160766, 1, 0, 0},{160766, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{47792, {{160767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47793, {{160768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47794, {{160769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{47795, {{160770, 1, 0, 0},{160770, 1, 1, 16}, {0, 0, 0, 0}}},{47795, {{160770, 1, 0, 0},{160770, 1, 1, 16}, {0, 0, 0, 0}}},{58653, {{160770, 1, 0, 0},{160770, 1, 1, 17}, {0, 0, 0, 0}}},{58653, {{160770, 1, 0, 0},{160770, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{47796, {{160771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47797, {{160772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47798, {{160773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47799, {{160774, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47800, {{160775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47801, {{160776, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47802, {{160777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47803, {{160778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47804, {{160779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47805, {{160780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47806, {{160781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47807, {{160782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47808, {{160783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47809, {{160784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47810, {{160785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47811, {{160786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47812, {{160787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47813, {{160788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47814, {{160789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47815, {{160790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47816, {{160791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47817, {{160792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47818, {{160793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47819, {{160794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47820, {{160795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47821, {{160796, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47822, {{160797, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47823, {{160799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47824, {{160803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47825, {{160812, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47826, {{160813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47827, {{160814, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47828, {{160815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47829, {{160816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47830, {{160817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47831, {{160818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47832, {{160819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47833, {{160820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47834, {{160822, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47835, {{160832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47836, {{160836, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47837, {{160837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47838, {{160838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47839, {{160839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47840, {{160840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47841, {{160841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47842, {{160842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47843, {{160843, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47844, {{160844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47845, {{160845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47846, {{160846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47847, {{160847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47848, {{160848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47849, {{160849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47850, {{160850, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47851, {{160851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47852, {{160852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47853, {{160853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47854, {{160854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47855, {{160855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47856, {{160856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47857, {{160857, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47858, {{160858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47859, {{160860, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47860, {{160861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47861, {{160862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47862, {{160863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47863, {{160864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47864, {{160865, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47865, {{160866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47866, {{160867, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47867, {{160868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47868, {{160869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47869, {{160870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47870, {{160871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47871, {{160872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47872, {{160873, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47873, {{160874, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47874, {{160875, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47875, {{160876, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47876, {{160877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47877, {{160878, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47878, {{160879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47879, {{160880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47880, {{160881, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47881, {{160882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47882, {{160883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47883, {{160884, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47884, {{160885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47885, {{160888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47886, {{160897, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47887, {{160898, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47888, {{160899, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47889, {{160904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47890, {{160906, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47891, {{160910, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47892, {{160911, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47893, {{160912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47894, {{160913, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47895, {{160914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47896, {{160915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47897, {{160916, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47898, {{160918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47899, {{160919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47900, {{160920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47600, {{160921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47901, {{160922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47902, {{160924, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47903, {{160926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47904, {{160927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47905, {{160929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47906, {{160930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47907, {{160931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47908, {{160932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47909, {{160933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47910, {{160934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47911, {{160935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47912, {{160939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47913, {{160956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47914, {{160957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47915, {{160958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47916, {{160959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47917, {{160960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47918, {{160961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47919, {{160962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47920, {{160969, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47921, {{160970, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47922, {{160971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47923, {{160972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47924, {{160974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47925, {{160975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47926, {{160976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47927, {{160977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47928, {{160978, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47929, {{160979, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47930, {{160980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47931, {{160981, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47932, {{160982, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47933, {{160983, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47934, {{160984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47935, {{160985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47936, {{160986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47937, {{160987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47938, {{160988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47939, {{160989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47940, {{160990, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47941, {{160991, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47942, {{160994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47943, {{161000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47944, {{161001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47945, {{161002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47946, {{161003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47947, {{161004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47948, {{161009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47949, {{161010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47950, {{161011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47951, {{161012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47952, {{161013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47953, {{161014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47954, {{161015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47955, {{161016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47956, {{161017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47957, {{161018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47958, {{161019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47959, {{161020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47960, {{161021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47961, {{161022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47962, {{161023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47963, {{161024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47964, {{161025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47965, {{161026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47966, {{161027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47967, {{161028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47968, {{161029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47969, {{161030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47970, {{161031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47971, {{161032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47972, {{161038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47973, {{161039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47974, {{161040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47975, {{161046, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47976, {{161047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47977, {{161048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47978, {{161053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47979, {{161054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47980, {{161055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47981, {{161056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47982, {{161057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47983, {{161058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47984, {{161059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47985, {{161060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{47986, {{161061, 1, 0, 0},{161061, 1, 1, 17}, {0, 0, 0, 0}}},{47986, {{161061, 1, 0, 0},{161061, 1, 1, 17}, {0, 0, 0, 0}}},{47987, {{161061, 1, 0, 0},{161061, 1, 1, 16}, {0, 0, 0, 0}}},{47987, {{161061, 1, 0, 0},{161061, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{47988, {{161062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47989, {{161064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{47990, {{161066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47991, {{161067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47992, {{161068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47993, {{161069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47994, {{161070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47995, {{161071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47996, {{161072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47997, {{161073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47998, {{161074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{47999, {{161075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48000, {{161076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48001, {{161077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48002, {{161078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48003, {{161079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48004, {{161082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48005, {{161084, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48006, {{161091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48007, {{161092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48008, {{161093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48009, {{161094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48010, {{161095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48011, {{161096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48012, {{161097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48013, {{161098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48014, {{161099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48015, {{161100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48016, {{161101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48017, {{161102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48018, {{161103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48019, {{161104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48020, {{161105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48021, {{161106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48022, {{161107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48023, {{161108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48024, {{161109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48025, {{161113, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48026, {{161115, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48027, {{161116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48028, {{161117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48029, {{161118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48030, {{161119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48031, {{161120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48032, {{161121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48033, {{161122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48034, {{161123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48035, {{161124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48036, {{161125, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48037, {{161126, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48038, {{161127, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48039, {{161128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48040, {{161129, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48041, {{161134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48042, {{161135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48043, {{161136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48044, {{161137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48045, {{161138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48046, {{161139, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48047, {{161140, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48048, {{161141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48049, {{161142, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48050, {{161143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48051, {{161144, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48052, {{161145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48053, {{161146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48054, {{161148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48055, {{161149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48056, {{161150, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48057, {{161151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48058, {{161153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48059, {{161154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48060, {{161156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48061, {{161157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48062, {{161158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48063, {{161159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48064, {{161160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48069, {{161161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48065, {{161162, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48066, {{161163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48067, {{161164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48068, {{161165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48070, {{161169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48071, {{161170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48072, {{161171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48073, {{161172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48074, {{161173, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48075, {{161174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48076, {{161175, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48077, {{161176, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48078, {{161177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48079, {{161178, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48080, {{161179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48081, {{161180, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48082, {{161181, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48083, {{161182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48084, {{161184, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48085, {{161186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48086, {{161187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48087, {{161188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48088, {{161191, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48089, {{161192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48090, {{161193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48091, {{161194, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48092, {{161195, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48093, {{161196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48094, {{161197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48095, {{161198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48096, {{161199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48097, {{161203, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48098, {{161204, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48099, {{161205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48100, {{161206, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48101, {{161207, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48102, {{161208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48103, {{161210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48104, {{161211, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48105, {{161212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48106, {{161213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48107, {{161214, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48108, {{161215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48109, {{161216, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48110, {{161217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48111, {{161218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48112, {{161219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48113, {{161220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48114, {{161222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48115, {{161223, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48116, {{161224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48117, {{161225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48118, {{161226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48119, {{161227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48120, {{161228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48121, {{161230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48122, {{161232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48123, {{161233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48124, {{161234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48125, {{161236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48126, {{161237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48127, {{161238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48128, {{161239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48129, {{161240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48130, {{161241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48131, {{161242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48132, {{161243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48133, {{161244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48134, {{161245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48135, {{161246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48136, {{161247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48137, {{161248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48139, {{161251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48140, {{161254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48141, {{161255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48142, {{161256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48143, {{161257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48144, {{161258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48145, {{161259, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48146, {{161260, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48147, {{161261, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60163, {{161262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48148, {{161264, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48149, {{161265, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48150, {{161267, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48151, {{161268, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48152, {{161269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48153, {{161270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48154, {{161271, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48155, {{161272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48156, {{161273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48157, {{161274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48158, {{161275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48159, {{161276, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48160, {{161279, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{48161, {{161282, 1, 0, 0},{161282, 1, 1, 16}, {0, 0, 0, 0}}},{48161, {{161282, 1, 0, 0},{161282, 1, 1, 16}, {0, 0, 0, 0}}},{48162, {{161282, 1, 0, 0},{161282, 1, 1, 17}, {0, 0, 0, 0}}},{48162, {{161282, 1, 0, 0},{161282, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{48163, {{161283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48164, {{161284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48165, {{161285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48166, {{161286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48167, {{161288, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48168, {{161289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48169, {{161290, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48170, {{161291, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48171, {{161293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{48172, {{161294, 1, 0, 0},{161294, 1, 1, 16}, {0, 0, 0, 0}}},{48172, {{161294, 1, 0, 0},{161294, 1, 1, 16}, {0, 0, 0, 0}}},{58686, {{161294, 1, 0, 0},{161294, 1, 1, 17}, {0, 0, 0, 0}}},{58686, {{161294, 1, 0, 0},{161294, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{48173, {{161295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48174, {{161296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48175, {{161297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48177, {{161298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48178, {{161299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48179, {{161300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48180, {{161301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{48181, {{161305, 1, 0, 0},{161305, 1, 1, 16}, {0, 0, 0, 0}}},{48181, {{161305, 1, 0, 0},{161305, 1, 1, 16}, {0, 0, 0, 0}}},{48182, {{161305, 1, 0, 0},{161305, 1, 1, 17}, {0, 0, 0, 0}}},{48182, {{161305, 1, 0, 0},{161305, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{48183, {{161306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48184, {{161308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48185, {{161309, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48186, {{161310, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48187, {{161311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48188, {{161312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48189, {{161313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48190, {{161314, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48191, {{161315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48192, {{161316, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48193, {{161317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48194, {{161318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48195, {{161319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48196, {{161320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48197, {{161321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48198, {{161322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48199, {{161323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48200, {{161324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48201, {{161325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48202, {{161326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48203, {{161327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48204, {{161328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48205, {{161329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48206, {{161333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48207, {{161334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48208, {{161335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48209, {{161341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48210, {{161342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48211, {{161343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48212, {{161344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48213, {{161345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48214, {{161346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48215, {{161347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48216, {{161348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48217, {{161349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48218, {{161350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48219, {{161351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48220, {{161352, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48221, {{161353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48222, {{161354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48223, {{161355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48224, {{161356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48225, {{161357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48226, {{161358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48227, {{161359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48228, {{161360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48229, {{161361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48230, {{161362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48232, {{161367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48233, {{161368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48234, {{161371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48235, {{161372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48236, {{161373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48237, {{161374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48238, {{161375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48239, {{161376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48240, {{161377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48241, {{161378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48242, {{161379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48243, {{161380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48244, {{161381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{48245, {{161383, 1, 0, 0},{161383, 1, 1, 16}, {0, 0, 0, 0}}},{48245, {{161383, 1, 0, 0},{161383, 1, 1, 16}, {0, 0, 0, 0}}},{48246, {{161383, 1, 0, 0},{161383, 1, 1, 17}, {0, 0, 0, 0}}},{48246, {{161383, 1, 0, 0},{161383, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{48247, {{161384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48248, {{161385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48249, {{161386, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48250, {{161387, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48252, {{161389, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48253, {{161390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48254, {{161391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48255, {{161396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48256, {{161397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48257, {{161398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48258, {{161399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48259, {{161400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48260, {{161401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48261, {{161402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48262, {{161406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48263, {{161408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48264, {{161409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48265, {{161410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48266, {{161411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48267, {{161412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48268, {{161413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48269, {{161414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48270, {{161415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48271, {{161416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48272, {{161417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48274, {{161418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48275, {{161419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48276, {{161420, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48277, {{161421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48278, {{161422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48279, {{161423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48280, {{161424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48281, {{161425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48282, {{161427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48283, {{161429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48284, {{161435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48285, {{161436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48286, {{161437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48287, {{161441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48288, {{161444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48289, {{161445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48290, {{161446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48291, {{161447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48292, {{161448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48293, {{161449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48294, {{161450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48295, {{161451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48296, {{161452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48297, {{161453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48298, {{161454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48299, {{161455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48300, {{161456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48301, {{161457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48302, {{161458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48303, {{161459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48304, {{161460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48305, {{161461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48306, {{161463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48307, {{161464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48308, {{161465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48309, {{161466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48311, {{161473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48312, {{161474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48313, {{161475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48314, {{161476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48316, {{161487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48317, {{161488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48318, {{161489, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48319, {{161490, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48320, {{161491, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{48321, {{161492, 1, 0, 0},{161492, 1, 1, 16}, {0, 0, 0, 0}}},{48321, {{161492, 1, 0, 0},{161492, 1, 1, 16}, {0, 0, 0, 0}}},{48322, {{161492, 1, 0, 0},{161492, 1, 1, 17}, {0, 0, 0, 0}}},{48322, {{161492, 1, 0, 0},{161492, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{48323, {{161493, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48324, {{161494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48325, {{161495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48326, {{161496, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48327, {{161497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48328, {{161498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48329, {{161499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48330, {{161500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48331, {{161501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48332, {{161502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48333, {{161503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48334, {{161504, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48335, {{161505, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48336, {{161506, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48337, {{161507, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48338, {{161508, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48339, {{161509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48340, {{161510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48341, {{161511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48342, {{161512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48343, {{161513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48344, {{161514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48345, {{161515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48346, {{161516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48347, {{161522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48348, {{161523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48349, {{161528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48350, {{161529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48351, {{161530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48352, {{161531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48353, {{161532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48354, {{161533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48355, {{161534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48356, {{161535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48357, {{161536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{48358, {{161537, 1, 0, 0},{161537, 1, 1, 16}, {0, 0, 0, 0}}},{48358, {{161537, 1, 0, 0},{161537, 1, 1, 16}, {0, 0, 0, 0}}},{48359, {{161537, 1, 0, 0},{161537, 1, 1, 17}, {0, 0, 0, 0}}},{48359, {{161537, 1, 0, 0},{161537, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{48360, {{161538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48361, {{161539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48362, {{161540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{48363, {{161541, 1, 0, 0},{161541, 1, 1, 16}, {0, 0, 0, 0}}},{48363, {{161541, 1, 0, 0},{161541, 1, 1, 16}, {0, 0, 0, 0}}},{58705, {{161541, 1, 0, 0},{161541, 1, 1, 17}, {0, 0, 0, 0}}},{58705, {{161541, 1, 0, 0},{161541, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{48364, {{161542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48365, {{161543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48366, {{161544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48367, {{161545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48368, {{161546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48369, {{161547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48370, {{161548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48371, {{161549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48372, {{161550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{48373, {{161551, 1, 0, 0},{161551, 1, 1, 16}, {0, 0, 0, 0}}},{48373, {{161551, 1, 0, 0},{161551, 1, 1, 16}, {0, 0, 0, 0}}},{48374, {{161551, 1, 0, 0},{161551, 1, 1, 17}, {0, 0, 0, 0}}},{48374, {{161551, 1, 0, 0},{161551, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48375, {{161555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48376, {{161556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48377, {{161557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48378, {{161562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48379, {{161565, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48380, {{161566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48381, {{161567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48382, {{161568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48383, {{161569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48384, {{161570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48385, {{161571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48386, {{161572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48387, {{161573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48388, {{161574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48389, {{161575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48390, {{161577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48392, {{161581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48393, {{161582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48394, {{161583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48395, {{161584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48396, {{161586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48397, {{161587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48398, {{161591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48399, {{161592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48400, {{161597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48401, {{161602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48402, {{161603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48403, {{161604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48404, {{161605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48405, {{161606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48406, {{161607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48407, {{161608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48408, {{161609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48409, {{161610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48410, {{161611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48411, {{161612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48412, {{161613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48413, {{161614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48414, {{161615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48415, {{161616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48416, {{161617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48417, {{161618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{48418, {{161619, 1, 0, 0},{161619, 1, 1, 16}, {0, 0, 0, 0}}},{48418, {{161619, 1, 0, 0},{161619, 1, 1, 16}, {0, 0, 0, 0}}},{58708, {{161619, 1, 0, 0},{161619, 1, 1, 17}, {0, 0, 0, 0}}},{58708, {{161619, 1, 0, 0},{161619, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{48419, {{161620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48420, {{161621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48421, {{161622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48422, {{161623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48423, {{161624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48424, {{161625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48425, {{161626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48426, {{161627, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48427, {{161628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48428, {{161642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48429, {{161643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48430, {{161644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48431, {{161645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48432, {{161646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48433, {{161647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{48434, {{161649, 1, 0, 0},{161649, 1, 1, 16}, {0, 0, 0, 0}}},{48434, {{161649, 1, 0, 0},{161649, 1, 1, 16}, {0, 0, 0, 0}}},{58710, {{161649, 1, 0, 0},{161649, 1, 1, 17}, {0, 0, 0, 0}}},{58710, {{161649, 1, 0, 0},{161649, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{48435, {{161650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48436, {{161651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48437, {{161652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48438, {{161653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48439, {{161654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48440, {{161655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48441, {{161656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48442, {{161657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48443, {{161658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48444, {{161659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48445, {{161660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48446, {{161661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48447, {{161662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48448, {{161663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48449, {{161664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48450, {{161665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48451, {{161666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48452, {{161676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48453, {{161677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48454, {{161678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48455, {{161679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48456, {{161680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4473, {{161681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48457, {{161682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48458, {{161683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48459, {{161684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48460, {{161685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48461, {{161686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48462, {{161687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48463, {{161689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48464, {{161691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48465, {{161692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48466, {{161693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48467, {{161694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48468, {{161695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48469, {{161698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48470, {{161699, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48471, {{161702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48472, {{161703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48473, {{161704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48474, {{161705, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{48475, {{161706, 1, 0, 0},{161706, 1, 1, 16}, {0, 0, 0, 0}}},{48475, {{161706, 1, 0, 0},{161706, 1, 1, 16}, {0, 0, 0, 0}}},{68088, {{161706, 1, 0, 0},{161706, 1, 1, 17}, {0, 0, 0, 0}}},{68088, {{161706, 1, 0, 0},{161706, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{48476, {{161707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48477, {{161708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48478, {{161709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48479, {{161710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48480, {{161711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48481, {{161712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48482, {{161713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48483, {{161714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48484, {{161718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48485, {{161719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{48486, {{161720, 1, 0, 0},{161720, 1, 1, 17}, {0, 0, 0, 0}}},{48486, {{161720, 1, 0, 0},{161720, 1, 1, 17}, {0, 0, 0, 0}}},{48487, {{161720, 1, 0, 0},{161720, 1, 1, 16}, {0, 0, 0, 0}}},{48487, {{161720, 1, 0, 0},{161720, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{48488, {{161721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48489, {{161722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48490, {{161723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48491, {{161724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48492, {{161725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48493, {{161730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48494, {{161731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48495, {{161732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48496, {{161736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48497, {{161737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48498, {{161738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48499, {{161740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48500, {{161741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48501, {{161742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48502, {{161743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48503, {{161744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48504, {{161745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48505, {{161746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48506, {{161748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48507, {{161749, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48508, {{161750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48509, {{161751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48510, {{161752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48511, {{161753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48512, {{161755, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48513, {{161756, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48514, {{161758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48515, {{161759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48516, {{161760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48517, {{161762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48518, {{161763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48520, {{161764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48519, {{161765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48521, {{161767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{48522, {{161768, 1, 0, 0},{161768, 1, 1, 17}, {0, 0, 0, 0}}},{48522, {{161768, 1, 0, 0},{161768, 1, 1, 17}, {0, 0, 0, 0}}},{48523, {{161768, 1, 0, 0},{161768, 1, 1, 16}, {0, 0, 0, 0}}},{48523, {{161768, 1, 0, 0},{161768, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{48524, {{161769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48525, {{161770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48526, {{161772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48527, {{161775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48528, {{161777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48529, {{161778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48530, {{161779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48531, {{161780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48532, {{161785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48533, {{161788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48534, {{161789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48535, {{161791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48536, {{161794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48537, {{161795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48538, {{161796, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48539, {{161798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48540, {{161799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48543, {{161800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48544, {{161801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48545, {{161802, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48546, {{161803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48547, {{161805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48548, {{161806, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48549, {{161807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48550, {{161809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48551, {{161810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48552, {{161811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48553, {{161816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48554, {{161817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48555, {{161818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48556, {{161819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48557, {{161820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48558, {{161821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48559, {{161822, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48560, {{161823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48561, {{161824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48562, {{161825, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48563, {{161826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48564, {{161830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48565, {{161832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48566, {{161833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48567, {{161834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48568, {{161835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48572, {{161836, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48569, {{161837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48570, {{161838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48571, {{161839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48573, {{161844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48574, {{161845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48575, {{161846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48576, {{161848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48577, {{161851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48578, {{161852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48580, {{161853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48581, {{161854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48582, {{161855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48583, {{161861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48584, {{161863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48585, {{161864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48586, {{161865, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48587, {{161867, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48588, {{161868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48589, {{161869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48590, {{161870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48591, {{161871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48592, {{161872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48593, {{161873, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48594, {{161874, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48595, {{161875, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48596, {{161880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48597, {{161882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48598, {{161883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48599, {{161884, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48600, {{161885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48601, {{161886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48602, {{161887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48603, {{161888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48604, {{161889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48605, {{161890, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48606, {{161891, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48607, {{161892, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48608, {{161893, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48609, {{161894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48610, {{161896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48611, {{161897, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48612, {{161898, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48613, {{161899, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48614, {{161900, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48615, {{161901, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48616, {{161902, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48617, {{161903, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48618, {{161909, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48619, {{161910, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48620, {{161911, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48621, {{161912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48622, {{161913, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48623, {{161914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48624, {{161915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48625, {{161916, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48626, {{161917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48627, {{161918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48628, {{161919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48629, {{161922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48630, {{161923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48631, {{161924, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48632, {{161925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48633, {{161926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48634, {{161931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48635, {{161932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48636, {{161933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48637, {{161934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48638, {{161935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48639, {{161936, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48640, {{161938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48641, {{161939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48642, {{161940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48643, {{161942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48644, {{161946, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58726, {{161947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48645, {{161948, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48646, {{161949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48648, {{161950, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48649, {{161954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48650, {{161955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48651, {{161957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48652, {{161958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48653, {{161959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48654, {{161960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48655, {{161962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48656, {{161963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48657, {{161964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48658, {{161965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48659, {{161966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48660, {{161967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48661, {{161969, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48662, {{161970, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48663, {{161971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48664, {{161972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48665, {{161973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48666, {{161974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48667, {{161975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48668, {{161976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48669, {{161977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48670, {{161979, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48671, {{161980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48672, {{161981, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48674, {{161982, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48675, {{161983, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48676, {{161984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48677, {{161985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48678, {{161987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48679, {{161988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48680, {{161992, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48681, {{161993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48682, {{161994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48683, {{161995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48684, {{161996, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48685, {{161997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48686, {{161998, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48687, {{161999, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48688, {{162000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48689, {{162001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48690, {{162002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48691, {{162003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48692, {{162004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48693, {{162005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48694, {{162006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48695, {{162007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48696, {{162008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48697, {{162009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48698, {{162010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48699, {{162011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48700, {{162012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48701, {{162015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48702, {{162017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48703, {{162018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48704, {{162019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48705, {{162020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48706, {{162022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48707, {{162024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48708, {{162025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48709, {{162026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48710, {{162027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48711, {{162028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48712, {{162029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48713, {{162031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48714, {{162032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48715, {{162033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48716, {{162036, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48717, {{162037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48718, {{162038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48719, {{162039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48720, {{162040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48721, {{162041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48722, {{162042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48723, {{162043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48724, {{162044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48725, {{162045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48726, {{162046, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48727, {{162048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48728, {{162051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48729, {{162052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48730, {{162053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48731, {{162054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48732, {{162055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48733, {{162056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48734, {{162057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48735, {{162059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48736, {{162060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48737, {{162061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48738, {{162062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48739, {{162063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48740, {{162064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48741, {{162065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48742, {{162066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48743, {{162067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48744, {{162068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48745, {{162069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48746, {{162070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48747, {{162071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48748, {{162072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48749, {{162073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48750, {{162074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48751, {{162075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48752, {{162076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48753, {{162077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48754, {{162078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48755, {{162079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48756, {{162080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48757, {{162081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48758, {{162082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48759, {{162083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48760, {{162086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48761, {{162087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48762, {{162088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48763, {{162089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48764, {{162091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48765, {{162092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48766, {{162094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48767, {{162095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48768, {{162096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48769, {{162097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48770, {{162098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48771, {{162099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48772, {{162101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48773, {{162102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48774, {{162106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48775, {{162107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48776, {{162108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48777, {{162109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48778, {{162110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48779, {{162111, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48780, {{162112, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48781, {{162113, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48782, {{162114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48783, {{162116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48784, {{162120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48785, {{162121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48786, {{162122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48787, {{162123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48788, {{162124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48789, {{162125, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48790, {{162126, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48791, {{162129, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48792, {{162131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48793, {{162133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48794, {{162134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48795, {{162135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48796, {{162136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48797, {{162137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48798, {{162138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48799, {{162139, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48800, {{162140, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48801, {{162141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48802, {{162142, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48803, {{162143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48804, {{162145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48805, {{162146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48806, {{162147, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48807, {{162148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48808, {{162149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{48809, {{162150, 1, 0, 0},{162150, 1, 1, 16}, {0, 0, 0, 0}}},{48809, {{162150, 1, 0, 0},{162150, 1, 1, 16}, {0, 0, 0, 0}}},{48810, {{162150, 1, 0, 0},{162150, 1, 1, 17}, {0, 0, 0, 0}}},{48810, {{162150, 1, 0, 0},{162150, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{48811, {{162151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48812, {{162152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48813, {{162153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48814, {{162154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48815, {{162155, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48816, {{162156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48817, {{162157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48818, {{162158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48819, {{162159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48820, {{162160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48821, {{162161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48822, {{162162, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48823, {{162163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48824, {{162164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48825, {{162165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{4559, {{162166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48826, {{162167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48827, {{162168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48828, {{162169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48829, {{162170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48830, {{162171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48831, {{162172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48832, {{162173, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48833, {{162174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48834, {{162175, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48835, {{162176, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48836, {{162177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48837, {{162181, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48838, {{162183, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48839, {{162190, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48840, {{162191, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48841, {{162192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48842, {{162193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48843, {{162194, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48844, {{162199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48845, {{162200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48846, {{162201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48847, {{162202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48848, {{162203, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48849, {{162204, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48850, {{162205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48851, {{162206, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48852, {{162207, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48853, {{162208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48854, {{162209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48855, {{162210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48856, {{162211, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48857, {{162212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48858, {{162213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48859, {{162214, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48860, {{162215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48861, {{162216, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48862, {{162217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48863, {{162218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48864, {{162219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48865, {{162221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48866, {{162222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48868, {{162232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48869, {{162233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48870, {{162234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48871, {{162235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48872, {{162238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48873, {{162241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48874, {{162242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48875, {{162243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48876, {{162244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48877, {{162245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48878, {{162246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48905, {{162247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48879, {{162249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48880, {{162250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48881, {{162251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48882, {{162252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48883, {{162253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48884, {{162254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48885, {{162255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48886, {{162256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48887, {{162257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48888, {{162258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48889, {{162259, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48890, {{162260, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48891, {{162261, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48892, {{162262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48893, {{162263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48894, {{162265, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{48895, {{162266, 1, 0, 0},{162266, 1, 1, 17}, {0, 0, 0, 0}}},{48895, {{162266, 1, 0, 0},{162266, 1, 1, 17}, {0, 0, 0, 0}}},{58741, {{162266, 1, 0, 0},{162266, 1, 1, 16}, {0, 0, 0, 0}}},{58741, {{162266, 1, 0, 0},{162266, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{48896, {{162267, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48897, {{162268, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48898, {{162269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48899, {{162270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48900, {{162271, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48901, {{162272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48902, {{162273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48903, {{162274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48904, {{162279, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48906, {{162281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48907, {{162282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48908, {{162283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48909, {{162288, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48910, {{162289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48911, {{162290, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48912, {{162291, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48913, {{162292, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48914, {{162293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48915, {{162294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48916, {{162295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48917, {{162296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48918, {{162297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48919, {{162298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48920, {{162299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48921, {{162300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48923, {{162301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48924, {{162302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48925, {{162303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48926, {{162304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48927, {{162305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48928, {{162306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48929, {{162307, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48930, {{162308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48931, {{162309, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48932, {{162310, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48933, {{162311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48934, {{162312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48935, {{162313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48936, {{162314, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48937, {{162315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48938, {{162319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48940, {{162327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48941, {{162328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48942, {{162329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48943, {{162330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48944, {{162331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48945, {{162333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48946, {{162334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48947, {{162335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48948, {{162336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48949, {{162337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48950, {{162339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48951, {{162341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48952, {{162342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48953, {{162343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48954, {{162345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48955, {{162346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48956, {{162347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48957, {{162348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48958, {{162349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48959, {{162350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48960, {{162351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48961, {{162352, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48962, {{162353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48963, {{162354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48964, {{162355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48965, {{162356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48966, {{162357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48967, {{162358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48968, {{162359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48969, {{162360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48970, {{162361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48971, {{162362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48972, {{162364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48973, {{162365, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48974, {{162366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48975, {{162367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48976, {{162368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48977, {{162369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48978, {{162370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48979, {{162371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48980, {{162372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48981, {{162373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48982, {{162374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48983, {{162375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48984, {{162376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48985, {{162378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48986, {{162384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48987, {{162385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48988, {{162386, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48989, {{162387, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48990, {{162388, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48991, {{162389, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48992, {{162390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48993, {{162392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48994, {{162394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48995, {{162395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48996, {{162396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48997, {{162397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48998, {{162398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{48999, {{162399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49000, {{162400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49001, {{162401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49002, {{162402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49003, {{162403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49004, {{162404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49005, {{162405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49006, {{162406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49007, {{162407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49008, {{162408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49009, {{162409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49010, {{162410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49011, {{162411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49012, {{162412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49013, {{162413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{49014, {{162414, 1, 0, 0},{162414, 1, 1, 16}, {0, 0, 0, 0}}},{49014, {{162414, 1, 0, 0},{162414, 1, 1, 16}, {0, 0, 0, 0}}},{58750, {{162414, 1, 0, 0},{162414, 1, 1, 17}, {0, 0, 0, 0}}},{58750, {{162414, 1, 0, 0},{162414, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{49015, {{162415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49016, {{162416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49017, {{162417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49018, {{162418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49019, {{162419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49020, {{162420, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49021, {{162421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49022, {{162422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49024, {{162423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49025, {{162424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49026, {{162425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49027, {{162426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{49028, {{162427, 1, 0, 0},{162427, 1, 1, 16}, {0, 0, 0, 0}}},{49028, {{162427, 1, 0, 0},{162427, 1, 1, 16}, {0, 0, 0, 0}}},{49029, {{162427, 1, 0, 0},{162427, 1, 1, 17}, {0, 0, 0, 0}}},{49029, {{162427, 1, 0, 0},{162427, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{49030, {{162428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49031, {{162429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49032, {{162430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49034, {{162431, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49035, {{162432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49036, {{162433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49038, {{162434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49039, {{162435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49040, {{162436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49041, {{162437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49042, {{162438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49043, {{162439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49044, {{162451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49045, {{162452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49046, {{162453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49047, {{162454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49048, {{162455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49049, {{162456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49050, {{162457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49051, {{162458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49052, {{162459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49053, {{162460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49054, {{162461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49055, {{162462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49056, {{162463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49057, {{162464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49058, {{162465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49059, {{162466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49060, {{162467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49061, {{162468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49062, {{162470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49063, {{162471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49064, {{162472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49065, {{162473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49066, {{162474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49067, {{162475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49068, {{162476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49069, {{162477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{49070, {{162478, 1, 0, 0},{162478, 1, 1, 16}, {0, 0, 0, 0}}},{49070, {{162478, 1, 0, 0},{162478, 1, 1, 16}, {0, 0, 0, 0}}},{58758, {{162478, 1, 0, 0},{162478, 1, 1, 17}, {0, 0, 0, 0}}},{58758, {{162478, 1, 0, 0},{162478, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{49071, {{162479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49072, {{162480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49073, {{162481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49074, {{162482, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49075, {{162483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49076, {{162484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49077, {{162485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49078, {{162486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49079, {{162487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49080, {{162489, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49081, {{162490, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49082, {{162495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49083, {{162506, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49084, {{162507, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49085, {{162511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49086, {{162512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49087, {{162513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49088, {{162514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49089, {{162515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49090, {{162516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49091, {{162517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49092, {{162518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49093, {{162519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49094, {{162520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49095, {{162521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49096, {{162522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49097, {{162523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49098, {{162524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49099, {{162525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49100, {{162526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49101, {{162527, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49102, {{162528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49103, {{162529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{49104, {{162530, 1, 0, 0},{162530, 1, 1, 16}, {0, 0, 0, 0}}},{49104, {{162530, 1, 0, 0},{162530, 1, 1, 16}, {0, 0, 0, 0}}},{49105, {{162530, 1, 0, 0},{162530, 1, 1, 17}, {0, 0, 0, 0}}},{49105, {{162530, 1, 0, 0},{162530, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{49106, {{162531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49107, {{162532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49108, {{162533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49109, {{162534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49110, {{162535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49111, {{162536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49112, {{162537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49113, {{162538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49114, {{162539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49115, {{162540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49116, {{162541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49117, {{162542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49118, {{162543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49119, {{162544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49120, {{162545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49121, {{162546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58762, {{162551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49122, {{162552, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49123, {{162553, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49124, {{162557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49125, {{162558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49126, {{162559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49127, {{162560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49128, {{162561, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49129, {{162562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49130, {{162563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49131, {{162564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49132, {{162565, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49134, {{162567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49135, {{162568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49136, {{162569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49137, {{162570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49138, {{162571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49139, {{162572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49140, {{162573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49141, {{162574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49142, {{162575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49143, {{162576, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49144, {{162577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49145, {{162578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49146, {{162579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49147, {{162581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49148, {{162582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49149, {{162583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49150, {{162588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49151, {{162589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49152, {{162590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49153, {{162591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49154, {{162592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49155, {{162593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49156, {{162594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49157, {{162595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49158, {{162596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49159, {{162598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49160, {{162599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49162, {{162600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49163, {{162601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49164, {{162602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49165, {{162603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49166, {{162604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49167, {{162605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49168, {{162606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{49169, {{162607, 1, 0, 0},{162607, 1, 1, 17}, {0, 0, 0, 0}}},{49169, {{162607, 1, 0, 0},{162607, 1, 1, 17}, {0, 0, 0, 0}}},{58763, {{162607, 1, 0, 0},{162607, 1, 1, 16}, {0, 0, 0, 0}}},{58763, {{162607, 1, 0, 0},{162607, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{49170, {{162608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49171, {{162609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49172, {{162610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49173, {{162611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49174, {{162612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49175, {{162613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49176, {{162614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49177, {{162623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49178, {{162624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49179, {{162625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49180, {{162626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49181, {{162627, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49182, {{162628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49183, {{162629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49184, {{162630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49185, {{162632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49186, {{162633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49187, {{162634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49188, {{162635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49189, {{162636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49190, {{162637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49191, {{162638, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49192, {{162639, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49193, {{162640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49194, {{162641, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49195, {{162642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49196, {{162643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49197, {{162644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49198, {{162645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49199, {{162646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49200, {{162648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49201, {{162650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49202, {{162651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49203, {{162652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49204, {{162653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49205, {{162654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49206, {{162655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49207, {{162656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49208, {{162657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49209, {{162658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49210, {{162659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49211, {{162660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49212, {{162661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49213, {{162662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49214, {{162663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49215, {{162664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49216, {{162665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49217, {{162666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49218, {{162667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49219, {{162668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49220, {{162669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49221, {{162670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49223, {{162675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49224, {{162676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49225, {{162677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49226, {{162679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49227, {{162680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49228, {{162681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49229, {{162682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49230, {{162683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49231, {{162684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49232, {{162685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49233, {{162686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49234, {{162687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49235, {{162688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60194, {{162690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49236, {{162691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49237, {{162692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49238, {{162694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{49239, {{162695, 1, 0, 0},{162695, 1, 1, 17}, {0, 0, 0, 0}}},{49239, {{162695, 1, 0, 0},{162695, 1, 1, 17}, {0, 0, 0, 0}}},{49240, {{162695, 1, 0, 0},{162695, 1, 1, 16}, {0, 0, 0, 0}}},{49240, {{162695, 1, 0, 0},{162695, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{49241, {{162696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49242, {{162697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49243, {{162698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49244, {{162701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49245, {{162702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49246, {{162703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49247, {{162704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49248, {{162705, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49249, {{162706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49250, {{162707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49251, {{162708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49252, {{162711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49253, {{162712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49254, {{162713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49255, {{162714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49256, {{162715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49257, {{162717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49258, {{162718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49259, {{162719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49260, {{162720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49261, {{162721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49262, {{162722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49263, {{162723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49264, {{162724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49265, {{162725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49266, {{162726, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49267, {{162728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49268, {{162729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49269, {{162731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49270, {{162732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49271, {{162733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49272, {{162734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49273, {{162735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49274, {{162736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49275, {{162737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49276, {{162738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49277, {{162739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49278, {{162740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49279, {{162741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49280, {{162742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49281, {{162743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49282, {{162744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49283, {{162746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49284, {{162747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49285, {{162748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49286, {{162749, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49287, {{162750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49289, {{162751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49290, {{162752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49291, {{162753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49292, {{162754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49293, {{162755, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{49294, {{162758, 1, 0, 0},{162758, 1, 1, 17}, {0, 0, 0, 0}}},{49294, {{162758, 1, 0, 0},{162758, 1, 1, 17}, {0, 0, 0, 0}}},{49295, {{162758, 1, 0, 0},{162758, 1, 1, 16}, {0, 0, 0, 0}}},{49295, {{162758, 1, 0, 0},{162758, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{49296, {{162759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49297, {{162761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49298, {{162762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49299, {{162763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{49300, {{162764, 1, 0, 0},{162764, 1, 1, 16}, {0, 0, 0, 0}}},{49300, {{162764, 1, 0, 0},{162764, 1, 1, 16}, {0, 0, 0, 0}}},{49301, {{162764, 1, 0, 0},{162764, 1, 1, 17}, {0, 0, 0, 0}}},{49301, {{162764, 1, 0, 0},{162764, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{49302, {{162765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49303, {{162766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49304, {{162767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49305, {{162768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49306, {{162769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49307, {{162771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49308, {{162772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49309, {{162773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49310, {{162775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49311, {{162776, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49312, {{162777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49313, {{162778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49314, {{162779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49315, {{162780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49316, {{162781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49317, {{162782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49318, {{162784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49319, {{162785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49320, {{162786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49321, {{162788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49322, {{162789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49323, {{162790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49324, {{162791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49325, {{162792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49326, {{162793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49327, {{162794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49328, {{162795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49329, {{162797, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49330, {{162798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49332, {{162799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49333, {{162800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49334, {{162801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49335, {{162802, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49336, {{162803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49337, {{162807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49338, {{162808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49339, {{162809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49340, {{162810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49341, {{162811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49342, {{162812, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49343, {{162813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{49344, {{162814, 1, 0, 0},{162814, 1, 1, 16}, {0, 0, 0, 0}}},{49344, {{162814, 1, 0, 0},{162814, 1, 1, 16}, {0, 0, 0, 0}}},{49345, {{162814, 1, 0, 0},{162814, 1, 1, 17}, {0, 0, 0, 0}}},{49345, {{162814, 1, 0, 0},{162814, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{49346, {{162815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49347, {{162816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49348, {{162819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49349, {{162820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49350, {{162821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49351, {{162822, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49352, {{162823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49353, {{162824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49354, {{162825, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49355, {{162826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49356, {{162827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49357, {{162828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49358, {{162829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49359, {{162832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49360, {{162833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49361, {{162834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49362, {{162835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49363, {{162837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49364, {{162840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49365, {{162841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49367, {{162848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49368, {{162849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49369, {{162850, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49370, {{162851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49371, {{162853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49372, {{162855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49373, {{162856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49374, {{162857, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49375, {{162858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49376, {{162859, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49377, {{162860, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49378, {{162861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49379, {{162862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49380, {{162863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49381, {{162864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49382, {{162865, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49383, {{162866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49384, {{162868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49385, {{162869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49386, {{162870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49387, {{162871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49388, {{162872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49389, {{162873, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49390, {{162875, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49391, {{162876, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49392, {{162877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49393, {{162879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49394, {{162880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49395, {{162881, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49396, {{162882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49397, {{162883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49398, {{162884, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49399, {{162885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49400, {{162886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49401, {{162887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49402, {{162888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49403, {{162889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49404, {{162890, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49405, {{162894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49406, {{162895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49407, {{162896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49408, {{162897, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49409, {{162898, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49410, {{162899, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49411, {{162900, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49412, {{162901, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49413, {{162902, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49414, {{162903, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49415, {{162904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49416, {{162905, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49417, {{162906, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49418, {{162907, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49419, {{162908, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49420, {{162909, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49421, {{162910, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49422, {{162911, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49423, {{162912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49424, {{162913, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49425, {{162914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49426, {{162915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49427, {{162916, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49428, {{162920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49429, {{162921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49430, {{162922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49431, {{162923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49432, {{162924, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49433, {{162925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49434, {{162926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49435, {{162927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49436, {{162928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49437, {{162929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49438, {{162930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49439, {{162931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49440, {{162932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49441, {{162933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49442, {{162934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49444, {{162939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49445, {{162940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49446, {{162941, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49447, {{162942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49448, {{162944, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49449, {{162945, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49450, {{162946, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49451, {{162947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49452, {{162948, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49453, {{162949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49454, {{162950, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49455, {{162951, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49456, {{162952, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49457, {{162953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49458, {{162955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49459, {{162956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49460, {{162957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49461, {{162958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49462, {{162959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49463, {{162960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49464, {{162961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49465, {{162962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49466, {{162963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49467, {{162964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49468, {{162965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49469, {{162967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49470, {{162968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49471, {{162969, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49472, {{162970, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49473, {{162971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49474, {{162972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49475, {{162973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49476, {{162974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49477, {{162975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49478, {{162977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49479, {{162978, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49481, {{162980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49482, {{162981, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49483, {{162982, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49484, {{162983, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{49485, {{162984, 1, 0, 0},{162984, 1, 1, 16}, {0, 0, 0, 0}}},{49485, {{162984, 1, 0, 0},{162984, 1, 1, 16}, {0, 0, 0, 0}}},{49486, {{162984, 1, 0, 0},{162984, 1, 1, 17}, {0, 0, 0, 0}}},{49486, {{162984, 1, 0, 0},{162984, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{49487, {{162985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49488, {{162986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49489, {{162988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49490, {{162989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49491, {{162990, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49492, {{162991, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49493, {{162992, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49494, {{162993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49495, {{162994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49496, {{162995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49497, {{162996, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49498, {{162997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49499, {{162998, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49500, {{162999, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49501, {{163000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49502, {{163002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49503, {{163003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49504, {{163006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49505, {{163007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49506, {{163008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49507, {{163009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49508, {{163010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49509, {{163011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49510, {{163012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49511, {{163013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49512, {{163014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49513, {{163017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49514, {{163018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49515, {{163019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49516, {{163020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49517, {{163021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49518, {{163022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49519, {{163023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49520, {{163024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49521, {{163025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49522, {{163026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49523, {{163028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49524, {{163030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49525, {{163033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49526, {{163034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49527, {{163035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49528, {{163036, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49529, {{163037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49530, {{163038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49531, {{163039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49532, {{163040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49533, {{163041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49534, {{163042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49535, {{163043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49536, {{163046, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49537, {{163047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49538, {{163048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49539, {{163049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49540, {{163050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49541, {{163051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49542, {{163052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49543, {{163053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49544, {{163054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49545, {{163055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49546, {{163058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49547, {{163059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49548, {{163060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49549, {{163061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49550, {{163062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49551, {{163063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49552, {{163064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49553, {{163065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49554, {{163066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49555, {{163067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49556, {{163068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49557, {{163071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49558, {{163073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49559, {{163074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49560, {{163075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49561, {{163076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49562, {{163077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49563, {{163078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49564, {{163079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49565, {{163082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49566, {{163084, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49567, {{163085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49568, {{163086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49569, {{163087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49570, {{163091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49571, {{163092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49572, {{163093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49573, {{163094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49574, {{163095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49575, {{163096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49576, {{163097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49577, {{163098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49578, {{163101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49579, {{163102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49580, {{163103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49581, {{163104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49582, {{163105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49583, {{163106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49584, {{163107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49585, {{163108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49587, {{163109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49588, {{163110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{49589, {{163114, 1, 0, 0},{163114, 1, 1, 16}, {0, 0, 0, 0}}},{49589, {{163114, 1, 0, 0},{163114, 1, 1, 16}, {0, 0, 0, 0}}},{49590, {{163114, 1, 0, 0},{163114, 1, 1, 17}, {0, 0, 0, 0}}},{49590, {{163114, 1, 0, 0},{163114, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{49592, {{163115, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49593, {{163116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49594, {{163117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49595, {{163122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49596, {{163123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49597, {{163124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49598, {{163125, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49599, {{163126, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49600, {{163127, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49601, {{163128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49602, {{163129, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49603, {{163130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49604, {{163139, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49605, {{163140, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49606, {{163141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49607, {{163143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49608, {{163144, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49609, {{163145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49610, {{163146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49611, {{163147, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49612, {{163148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{49613, {{163149, 1, 0, 0},{163149, 1, 1, 16}, {0, 0, 0, 0}}},{49613, {{163149, 1, 0, 0},{163149, 1, 1, 16}, {0, 0, 0, 0}}},{49614, {{163149, 1, 0, 0},{163149, 1, 1, 17}, {0, 0, 0, 0}}},{49614, {{163149, 1, 0, 0},{163149, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{49615, {{163150, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49616, {{163151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49617, {{163152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49618, {{163153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49619, {{163163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49620, {{163164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49621, {{163165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49622, {{163166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49623, {{163168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49624, {{163170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49625, {{163171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49626, {{163172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49627, {{163173, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49628, {{163174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49629, {{163175, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49630, {{163176, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49631, {{163177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49632, {{163178, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49633, {{163179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49634, {{163180, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49635, {{163181, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49636, {{163182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49637, {{163183, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49638, {{163184, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49639, {{163185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49640, {{163186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49641, {{163195, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49642, {{163196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49643, {{163197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49644, {{163198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49645, {{163199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49646, {{163202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49647, {{163205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49648, {{163206, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49649, {{163207, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49650, {{163208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49651, {{163209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49652, {{163210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49653, {{163211, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49654, {{163212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49655, {{163213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49656, {{163214, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49657, {{163221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49658, {{163222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49659, {{163223, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49660, {{163225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49661, {{163226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49662, {{163227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49663, {{163228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49664, {{163229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49665, {{163230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49666, {{163231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49667, {{163232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49668, {{163233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49669, {{163234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49670, {{163235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49671, {{163236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49672, {{163241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49673, {{163242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49674, {{163249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49675, {{163250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49676, {{163251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49677, {{163252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49678, {{163253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49679, {{163254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49680, {{163255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49681, {{163256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49682, {{163257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49683, {{163258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49684, {{163259, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49685, {{163261, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49686, {{163268, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49687, {{163269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49688, {{163271, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49689, {{163273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49690, {{163274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49691, {{163275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49692, {{163276, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49693, {{163277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49694, {{163278, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49695, {{163279, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49696, {{163280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49697, {{163281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49698, {{163282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49699, {{163283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49700, {{163284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49701, {{163285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49702, {{163286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49703, {{163287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49704, {{163294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49705, {{163295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49706, {{163296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49707, {{163297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49708, {{163298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49709, {{163299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49710, {{163300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49711, {{163301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49712, {{163302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49713, {{163303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49714, {{163304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49715, {{163305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49716, {{163313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49717, {{163314, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49718, {{163315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49719, {{163317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49720, {{163323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49721, {{163325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49722, {{163326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49723, {{163327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49724, {{163328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49725, {{163329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49726, {{163330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49727, {{163331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49728, {{163332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49729, {{163333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49730, {{163334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49731, {{163335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49732, {{163336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49733, {{163338, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49734, {{163339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49735, {{163344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49736, {{163345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49737, {{163348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49738, {{163349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49739, {{163350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49740, {{163351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49741, {{163352, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{49742, {{163353, 1, 0, 0},{163353, 1, 1, 17}, {0, 0, 0, 0}}},{49742, {{163353, 1, 0, 0},{163353, 1, 1, 17}, {0, 0, 0, 0}}},{49743, {{163353, 1, 0, 0},{163353, 1, 1, 16}, {0, 0, 0, 0}}},{49743, {{163353, 1, 0, 0},{163353, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49744, {{163358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49745, {{163365, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49746, {{163366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49747, {{163367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49748, {{163368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49749, {{163369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49754, {{163370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49750, {{163371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49751, {{163375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49752, {{163377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49753, {{163378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49755, {{163383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49756, {{163384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49757, {{163385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{49758, {{163389, 1, 0, 0},{163389, 1, 1, 17}, {0, 0, 0, 0}}},{49758, {{163389, 1, 0, 0},{163389, 1, 1, 17}, {0, 0, 0, 0}}},{49759, {{163389, 1, 0, 0},{163389, 1, 1, 16}, {0, 0, 0, 0}}},{49759, {{163389, 1, 0, 0},{163389, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{49760, {{163390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49761, {{163391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49762, {{163392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49763, {{163398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49764, {{163399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49765, {{163400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49766, {{163401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49767, {{163402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49768, {{163404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49769, {{163406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49770, {{163408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49771, {{163416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49772, {{163417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49773, {{163418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49774, {{163419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49775, {{163421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49776, {{163422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49777, {{163424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49778, {{163427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49779, {{163431, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49780, {{163432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49781, {{163433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49782, {{163436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49783, {{163437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49784, {{163439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49785, {{163442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49786, {{163443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49787, {{163444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49788, {{163446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49789, {{163447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49790, {{163448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{49791, {{163449, 1, 0, 0},{163449, 1, 1, 16}, {0, 0, 0, 0}}},{49791, {{163449, 1, 0, 0},{163449, 1, 1, 16}, {0, 0, 0, 0}}},{49792, {{163449, 1, 0, 0},{163449, 1, 1, 17}, {0, 0, 0, 0}}},{49792, {{163449, 1, 0, 0},{163449, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{49793, {{163450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49794, {{163453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49795, {{163454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49796, {{163460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49797, {{163461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49798, {{163462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49799, {{163463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49802, {{163470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49803, {{163471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49804, {{163474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49805, {{163475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49806, {{163476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49807, {{163481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49808, {{163483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49809, {{163484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49810, {{163485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49811, {{163486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49812, {{163487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49813, {{163488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49814, {{163489, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49815, {{163490, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49816, {{163491, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49817, {{163492, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49818, {{163493, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49819, {{163494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49820, {{163495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49821, {{163496, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49822, {{163504, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49823, {{163506, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49824, {{163507, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49825, {{163510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49826, {{163511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49827, {{163512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49828, {{163513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49829, {{163514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49830, {{163515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49831, {{163516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49832, {{163517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49833, {{163518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49834, {{163519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49835, {{163520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49836, {{163521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49837, {{163522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49838, {{163523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49839, {{163533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49840, {{163537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49841, {{163538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49842, {{163539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49843, {{163540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49844, {{163541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49845, {{163542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49846, {{163543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49847, {{163544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49848, {{163545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49849, {{163546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49850, {{163547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49851, {{163548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49852, {{163549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49853, {{163550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49854, {{163559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49855, {{163563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49856, {{163564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49857, {{163567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49858, {{163568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49859, {{163569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49860, {{163570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49861, {{163571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49862, {{163572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49863, {{163573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49864, {{163574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49865, {{163575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49866, {{163576, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49867, {{163577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49868, {{163578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49869, {{163579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49870, {{163580, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49871, {{163581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49872, {{163583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49873, {{163593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49874, {{163594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49875, {{163598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49876, {{163599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49877, {{163600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49878, {{163601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49879, {{163602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49880, {{163603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49881, {{163604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49882, {{163605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49883, {{163606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49884, {{163607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49885, {{163608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49886, {{163609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49887, {{163610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49888, {{163611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49889, {{163612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49890, {{163613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49891, {{163614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49892, {{163615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49893, {{163616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49894, {{163617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49895, {{163620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49896, {{163621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49897, {{163622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49898, {{163624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49899, {{163625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49900, {{163626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49901, {{163627, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49902, {{163628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49903, {{163629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49904, {{163630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49905, {{163631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49906, {{163632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49907, {{163633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49908, {{163634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49909, {{163635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49910, {{163636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49911, {{163637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49912, {{163638, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49913, {{163640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49914, {{163649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49915, {{163650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49916, {{163652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49917, {{163653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49918, {{163654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49919, {{163657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49920, {{163658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49921, {{163659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49922, {{163660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49923, {{163661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49924, {{163662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49925, {{163663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49926, {{163664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49927, {{163665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49928, {{163666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49929, {{163667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49930, {{163668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49931, {{163669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49932, {{163670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49933, {{163677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49934, {{163678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49935, {{163679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49936, {{163680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49937, {{163682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49938, {{163683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49939, {{163684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49940, {{163685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49941, {{163686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49942, {{163687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49943, {{163688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49944, {{163689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49945, {{163690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49946, {{163691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49947, {{163693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49948, {{163695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49949, {{163698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49950, {{163699, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49951, {{163700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49952, {{163701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49953, {{163702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49954, {{163703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49955, {{163704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49956, {{163705, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49957, {{163706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49958, {{163707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49959, {{163708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49960, {{163709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49961, {{163710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49962, {{163711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49963, {{163712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49964, {{163713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49965, {{163714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49966, {{163719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49967, {{163720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49968, {{163722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49969, {{163723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49970, {{163725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49971, {{163726, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49972, {{163727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49973, {{163728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49974, {{163729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49975, {{163730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49976, {{163732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49977, {{163733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49978, {{163737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49979, {{163738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49980, {{163739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49981, {{163740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49982, {{163741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49983, {{163742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49984, {{163743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49985, {{163744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49986, {{163745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49987, {{163746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49988, {{163747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49989, {{163748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49990, {{163749, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49991, {{163751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{49992, {{163752, 1, 0, 0},{163752, 1, 1, 16}, {0, 0, 0, 0}}},{49992, {{163752, 1, 0, 0},{163752, 1, 1, 16}, {0, 0, 0, 0}}},{49993, {{163752, 1, 0, 0},{163752, 1, 1, 17}, {0, 0, 0, 0}}},{49993, {{163752, 1, 0, 0},{163752, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{49994, {{163753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49995, {{163755, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49996, {{163756, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49997, {{163757, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49998, {{163758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{49999, {{163759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50000, {{163760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50001, {{163761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50002, {{163762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50003, {{163763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50004, {{163764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50005, {{163765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50006, {{163766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50007, {{163767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50008, {{163768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50009, {{163769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50010, {{163771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50011, {{163773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50012, {{163774, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{50013, {{163776, 1, 0, 0},{163776, 1, 1, 16}, {0, 0, 0, 0}}},{50013, {{163776, 1, 0, 0},{163776, 1, 1, 16}, {0, 0, 0, 0}}},{68090, {{163776, 1, 0, 0},{163776, 1, 1, 17}, {0, 0, 0, 0}}},{68090, {{163776, 1, 0, 0},{163776, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{50014, {{163777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50015, {{163778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50016, {{163779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50017, {{163780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50018, {{163781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50019, {{163782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50020, {{163785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50021, {{163788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50022, {{163789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50023, {{163794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50024, {{163795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50025, {{163796, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50026, {{163797, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50027, {{163798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50028, {{163799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50029, {{163800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50030, {{163801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50031, {{163802, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50032, {{163803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50033, {{163804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50034, {{163805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50035, {{163806, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50036, {{163807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50037, {{163811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50038, {{163812, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50039, {{163813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50040, {{163814, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50041, {{163817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50042, {{163819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50043, {{163822, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50044, {{163823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50045, {{163826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50046, {{163827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50047, {{163828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50048, {{163829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50049, {{163830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50050, {{163831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50051, {{163832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50052, {{163833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50053, {{163834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50054, {{163835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50055, {{163836, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50056, {{163837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50057, {{163838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50058, {{163839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50059, {{163840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50060, {{163844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50061, {{163846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50062, {{163847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50063, {{163859, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50064, {{163860, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50065, {{163861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50066, {{163862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50067, {{163863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50068, {{163864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50069, {{163865, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50070, {{163866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50071, {{163867, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50072, {{163868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50073, {{163869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50074, {{163870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50075, {{163871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50076, {{163872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50077, {{163873, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50078, {{163874, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50079, {{163875, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50080, {{163876, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50081, {{163877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50082, {{163878, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50083, {{163879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50084, {{163880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50085, {{163881, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50086, {{163882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50087, {{163883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50088, {{163884, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50089, {{163885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50090, {{163886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50092, {{163887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50093, {{163891, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50094, {{163892, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50095, {{163894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50096, {{163895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50097, {{163896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50098, {{163904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50099, {{163905, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50100, {{163906, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50101, {{163907, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50102, {{163908, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50103, {{163909, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50104, {{163910, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50105, {{163911, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50106, {{163912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50107, {{163914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50108, {{163915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50109, {{163916, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50110, {{163917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50111, {{163918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50112, {{163919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50113, {{163920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50114, {{163921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50115, {{163922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50116, {{163923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50117, {{163924, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50118, {{163925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50119, {{163932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50120, {{163933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50121, {{163934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50122, {{163935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50123, {{163936, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50124, {{163937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50125, {{163938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50126, {{163939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50127, {{163955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50128, {{163957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50129, {{163958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50130, {{163959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50131, {{163960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50132, {{163961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50133, {{163962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50134, {{163963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50135, {{163964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50136, {{163965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50137, {{163966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50138, {{163967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50139, {{163968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50140, {{163969, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50141, {{163970, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50142, {{163971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50143, {{163972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50144, {{163973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50145, {{163974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50146, {{163975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50147, {{163976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50148, {{163978, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50149, {{163979, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50150, {{163983, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50151, {{163985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50152, {{163986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50153, {{163987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50154, {{163995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50155, {{163996, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50156, {{163997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50157, {{163998, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50158, {{163999, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50159, {{164000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50160, {{164001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50161, {{164002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50162, {{164003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50163, {{164004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50164, {{164005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50165, {{164006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50166, {{164007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50167, {{164008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50168, {{164009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50169, {{164010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50170, {{164011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50171, {{164012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50172, {{164013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50173, {{164014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50174, {{164015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50175, {{164016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50176, {{164017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50177, {{164018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50178, {{164019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50179, {{164020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50180, {{164021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50181, {{164022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50182, {{164023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50183, {{164024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50184, {{164027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50185, {{164050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50186, {{164051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50187, {{164052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50188, {{164053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50189, {{164054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50190, {{164055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50191, {{164056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50192, {{164057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50193, {{164058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50194, {{164059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50195, {{164060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50196, {{164061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50197, {{164062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50198, {{164063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50199, {{164064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50200, {{164066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50201, {{164067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50202, {{164068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50203, {{164069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50204, {{164070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50205, {{164071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50206, {{164072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50207, {{164077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50208, {{164078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50209, {{164079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50210, {{164082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50211, {{164095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50212, {{164096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50213, {{164099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50214, {{164101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50215, {{164102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50216, {{164103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50217, {{164104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50218, {{164105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50219, {{164106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50220, {{164107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50221, {{164108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50222, {{164109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50223, {{164110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50224, {{164111, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50225, {{164112, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50226, {{164113, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50227, {{164114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50228, {{164115, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50229, {{164116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50230, {{164117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50231, {{164118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50232, {{164119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50233, {{164120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{50234, {{164121, 1, 0, 0},{164121, 1, 1, 16}, {0, 0, 0, 0}}},{50234, {{164121, 1, 0, 0},{164121, 1, 1, 16}, {0, 0, 0, 0}}},{50235, {{164121, 1, 0, 0},{164121, 1, 1, 17}, {0, 0, 0, 0}}},{50235, {{164121, 1, 0, 0},{164121, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{50236, {{164122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50237, {{164123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50238, {{164125, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50239, {{164135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50240, {{164136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50241, {{164137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50242, {{164138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50243, {{164139, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50244, {{164140, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50245, {{164141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50246, {{164143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50247, {{164144, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50248, {{164165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50249, {{164166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50250, {{164168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50251, {{164170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50252, {{164171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50253, {{164172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50254, {{164173, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50255, {{164174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50256, {{164175, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50257, {{164176, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50258, {{164177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50259, {{164178, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50260, {{164179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50261, {{164180, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50262, {{164181, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50263, {{164182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50264, {{164183, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50265, {{164184, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50266, {{164185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50267, {{164186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50268, {{164187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50269, {{164188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50270, {{164189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50271, {{164190, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50272, {{164191, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50273, {{164192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50274, {{164193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50275, {{164194, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50276, {{164195, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50277, {{164196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50278, {{164197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50279, {{164198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50280, {{164199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50281, {{164200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50282, {{164201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50283, {{164202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50345, {{164203, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50284, {{164209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50285, {{164210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50286, {{164212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50287, {{164213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50288, {{164214, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50289, {{164225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50290, {{164226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50291, {{164227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50292, {{164228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50293, {{164229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50294, {{164230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50295, {{164231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50296, {{164232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50297, {{164233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50298, {{164234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50299, {{164235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50300, {{164236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50301, {{164237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50302, {{164238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50303, {{164239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50304, {{164240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50305, {{164241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50306, {{164249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50307, {{164250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50308, {{164251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50309, {{164252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50310, {{164262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50311, {{164263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50312, {{164264, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50313, {{164266, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50314, {{164267, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50315, {{164268, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50316, {{164269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50317, {{164270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50318, {{164271, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50319, {{164272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50320, {{164273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50321, {{164274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50322, {{164275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50323, {{164276, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50324, {{164277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50325, {{164289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50326, {{164290, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50327, {{164291, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50328, {{164292, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50329, {{164293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50330, {{164294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50331, {{164295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50332, {{164296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50333, {{164298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50334, {{164299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50335, {{164302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50336, {{164303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50337, {{164304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50338, {{164308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50339, {{164310, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50340, {{164311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50341, {{164312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50342, {{164313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50343, {{164314, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50344, {{164317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50346, {{164320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50347, {{164324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50348, {{164325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50349, {{164326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50350, {{164327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50351, {{164328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50352, {{164329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50353, {{164330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50354, {{164331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50355, {{164333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50356, {{164334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50357, {{164335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50358, {{164336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50359, {{164337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50360, {{164338, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50361, {{164339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50362, {{164340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50363, {{164347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50364, {{164348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50365, {{164349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50366, {{164355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50367, {{164356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50368, {{164357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50369, {{164360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50370, {{164361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50371, {{164363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50372, {{164367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50373, {{164368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50374, {{164369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50375, {{164371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50376, {{164372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50377, {{164373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50378, {{164374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50379, {{164375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50380, {{164376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50381, {{164377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50382, {{164378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50383, {{164379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50384, {{164380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50385, {{164381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50386, {{164382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50387, {{164383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50388, {{164384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50389, {{164385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50390, {{164386, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50391, {{164389, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50392, {{164391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50393, {{164393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50394, {{164394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50395, {{164395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50396, {{164396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50397, {{164397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50398, {{164398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50399, {{164404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50400, {{164405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50401, {{164406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50402, {{164408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50403, {{164409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50404, {{164410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50405, {{164411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50406, {{164412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50407, {{164416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50408, {{164418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50409, {{164419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50410, {{164423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50411, {{164424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50412, {{164425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50413, {{164426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50414, {{164427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50415, {{164428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50416, {{164429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50417, {{164434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50418, {{164436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50419, {{164437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50420, {{164438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50421, {{164439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50422, {{164443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50423, {{164444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50424, {{164445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50425, {{164450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50426, {{164451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50427, {{164452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50428, {{164453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50429, {{164454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50430, {{164457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50431, {{164458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50432, {{164460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50433, {{164461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50434, {{164462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50435, {{164463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50436, {{164464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50437, {{164465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50438, {{164466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50439, {{164468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50440, {{164469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50441, {{164471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50442, {{164472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50443, {{164473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50444, {{164474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50445, {{164477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50446, {{164478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50447, {{164479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50448, {{164480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50449, {{164481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50450, {{164482, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50451, {{164485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50452, {{164487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50453, {{164489, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50454, {{164490, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50455, {{164493, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50456, {{164494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50457, {{164498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50458, {{164499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50459, {{164500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50460, {{164501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50461, {{164502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50462, {{164503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50463, {{164504, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50464, {{164507, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50465, {{164509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50466, {{164510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50467, {{164511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50468, {{164512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50469, {{164513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50470, {{164514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50471, {{164519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50473, {{164527, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50474, {{164528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50475, {{164529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50476, {{164531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50477, {{164532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50478, {{164533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50479, {{164534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50480, {{164536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50481, {{164537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50482, {{164538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50483, {{164539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50484, {{164540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50485, {{164541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50486, {{164542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50487, {{164543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50488, {{164544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50489, {{164545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50490, {{164546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50491, {{164550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50492, {{164551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50493, {{164552, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50494, {{164553, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50495, {{164556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50496, {{164558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50497, {{164559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50498, {{164560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50499, {{164561, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50500, {{164562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50501, {{164563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50502, {{164564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50503, {{164565, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50504, {{164566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50505, {{164567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50506, {{164568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50507, {{164569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50508, {{164570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50509, {{164571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50510, {{164573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50511, {{164574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50512, {{164575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50513, {{164577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50514, {{164581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50515, {{164582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50516, {{164583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50517, {{164584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50518, {{164585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50519, {{164586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50520, {{164587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50521, {{164588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50522, {{164589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50523, {{164590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50524, {{164591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50525, {{164592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50526, {{164593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50527, {{164594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50528, {{164595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50529, {{164599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50530, {{164601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50531, {{164602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50532, {{164603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50533, {{164604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50534, {{164605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50535, {{164606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50536, {{164609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50537, {{164610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50538, {{164612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50539, {{164613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50540, {{164614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50541, {{164615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50542, {{164616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50543, {{164617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50544, {{164618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50545, {{164619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50546, {{164621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50547, {{164622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50548, {{164623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50549, {{164627, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50550, {{164630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50551, {{164632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50552, {{164637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50553, {{164638, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50554, {{164639, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50555, {{164640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50556, {{164641, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50557, {{164642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50558, {{164643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50559, {{164644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50560, {{164645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50561, {{164646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50562, {{164647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50563, {{164649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50564, {{164650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50565, {{164651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50566, {{164652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50567, {{164653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50568, {{164654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50569, {{164655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50570, {{164656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50571, {{164657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50572, {{164658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50573, {{164659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50574, {{164660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50575, {{164661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50576, {{164662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50577, {{164663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50578, {{164664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60220, {{164666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50579, {{164672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50580, {{164673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50581, {{164674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50582, {{164675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50583, {{164677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50584, {{164678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50585, {{164679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50586, {{164680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50587, {{164681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50588, {{164682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50589, {{164683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50590, {{164684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50591, {{164685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50592, {{164686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50593, {{164687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50594, {{164688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50595, {{164689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50596, {{164690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50597, {{164691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50598, {{164692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50599, {{164699, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50600, {{164701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50601, {{164702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50602, {{164703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50603, {{164706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50604, {{164707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50605, {{164711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50606, {{164712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50607, {{164713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50608, {{164714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50609, {{164715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50610, {{164716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50611, {{164717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50612, {{164718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50613, {{164719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50614, {{164720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50615, {{164721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50616, {{164722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50617, {{164723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50618, {{164724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50619, {{164726, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50620, {{164727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50621, {{164728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50622, {{164729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50623, {{164731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50624, {{164735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50625, {{164737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50626, {{164740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50627, {{164741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50628, {{164743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50629, {{164744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50630, {{164745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{50631, {{164746, 1, 0, 0},{164746, 1, 1, 16}, {0, 0, 0, 0}}},{50631, {{164746, 1, 0, 0},{164746, 1, 1, 16}, {0, 0, 0, 0}}},{50632, {{164746, 1, 0, 0},{164746, 1, 1, 17}, {0, 0, 0, 0}}},{50632, {{164746, 1, 0, 0},{164746, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{50633, {{164747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50634, {{164748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50635, {{164749, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50636, {{164752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50637, {{164756, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50638, {{164757, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50639, {{164758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50640, {{164759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50641, {{164760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50642, {{164761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50643, {{164764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50644, {{164769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50645, {{164770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50646, {{164771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50647, {{164772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50648, {{164773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50649, {{164774, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50650, {{164775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50651, {{164776, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50652, {{164777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50653, {{164778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50654, {{164779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50655, {{164780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50656, {{164781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50657, {{164782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50658, {{164783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50659, {{164785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50660, {{164786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50661, {{164787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50662, {{164788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50663, {{164789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50664, {{164790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50665, {{164791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50666, {{164793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50667, {{164798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50668, {{164800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50669, {{164801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50670, {{164805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50671, {{164806, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50672, {{164807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50673, {{164808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50674, {{164809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50675, {{164810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50677, {{164811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50678, {{164813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50679, {{164814, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50680, {{164815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50681, {{164817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50682, {{164818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50683, {{164819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50684, {{164820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50685, {{164821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50686, {{164822, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50687, {{164824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50688, {{164825, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50689, {{164826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50690, {{164827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50691, {{164829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50692, {{164831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50693, {{164832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50694, {{164833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50695, {{164834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50696, {{164835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50697, {{164836, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50698, {{164837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50699, {{164838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50700, {{164840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50701, {{164841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50702, {{164842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50703, {{164843, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50704, {{164844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50705, {{164847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50706, {{164848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50707, {{164849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50708, {{164850, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50709, {{164851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50711, {{164852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50712, {{164853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50714, {{164854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50715, {{164855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50716, {{164856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50717, {{164857, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50718, {{164858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50719, {{164861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50720, {{164862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50721, {{164863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50722, {{164865, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50713, {{164866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50723, {{164867, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50724, {{164876, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50725, {{164877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50726, {{164879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50727, {{164880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50728, {{164881, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50729, {{164884, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50730, {{164885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50731, {{164886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50732, {{164891, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50733, {{164892, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50734, {{164893, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50735, {{164894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50736, {{164895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50737, {{164896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50738, {{164897, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50739, {{164898, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50740, {{164899, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50741, {{164901, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50742, {{164902, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50743, {{164904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50744, {{164905, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50745, {{164911, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50746, {{164912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50747, {{164913, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{50748, {{164914, 1, 0, 0},{164914, 1, 1, 16}, {0, 0, 0, 0}}},{50748, {{164914, 1, 0, 0},{164914, 1, 1, 16}, {0, 0, 0, 0}}},{50749, {{164914, 1, 0, 0},{164914, 1, 1, 17}, {0, 0, 0, 0}}},{50749, {{164914, 1, 0, 0},{164914, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{50750, {{164915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50751, {{164916, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50752, {{164917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50753, {{164918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50754, {{164921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50755, {{164924, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50756, {{164926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50757, {{164928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50758, {{164929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50759, {{164930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50760, {{164931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50761, {{164932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50762, {{164934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50763, {{164935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50764, {{164937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50765, {{164938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50766, {{164939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50767, {{164940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{50768, {{164941, 1, 0, 0},{164941, 1, 1, 17}, {0, 0, 0, 0}}},{50768, {{164941, 1, 0, 0},{164941, 1, 1, 17}, {0, 0, 0, 0}}},{50769, {{164941, 1, 0, 0},{164941, 1, 1, 16}, {0, 0, 0, 0}}},{50769, {{164941, 1, 0, 0},{164941, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{50770, {{164942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50771, {{164943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50772, {{164944, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50773, {{164945, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{50774, {{164946, 1, 0, 0},{164946, 1, 1, 17}, {0, 0, 0, 0}}},{50774, {{164946, 1, 0, 0},{164946, 1, 1, 17}, {0, 0, 0, 0}}},{50775, {{164946, 1, 0, 0},{164946, 1, 1, 16}, {0, 0, 0, 0}}},{50775, {{164946, 1, 0, 0},{164946, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{50776, {{164947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50778, {{164948, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{50779, {{164949, 1, 0, 0},{164949, 1, 1, 18}, {0, 0, 0, 0}}},{50779, {{164949, 1, 0, 0},{164949, 1, 1, 18}, {0, 0, 0, 0}}},{50780, {{164949, 1, 0, 0},{164949, 1, 1, 17}, {0, 0, 0, 0}}},{50780, {{164949, 1, 0, 0},{164949, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50781, {{164951, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50782, {{164955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50783, {{164956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50784, {{164957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50785, {{164958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50786, {{164960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50787, {{164961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50788, {{164962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50789, {{164963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50790, {{164964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50792, {{164965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50793, {{164966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50794, {{164967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50795, {{164968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50797, {{164969, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50798, {{164970, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50799, {{164971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{50800, {{164973, 1, 0, 0},{164973, 1, 1, 17}, {0, 0, 0, 0}}},{50800, {{164973, 1, 0, 0},{164973, 1, 1, 17}, {0, 0, 0, 0}}},{50801, {{164973, 1, 0, 0},{164973, 1, 1, 16}, {0, 0, 0, 0}}},{50801, {{164973, 1, 0, 0},{164973, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58842, {{164978, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50803, {{164981, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50804, {{164983, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50805, {{164985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50806, {{164988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50807, {{164989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50808, {{164991, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50809, {{164992, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50810, {{164993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{57455, {{164994, 1, 0, 0},{164994, 1, 1, 16}, {0, 0, 0, 0}}},{57455, {{164994, 1, 0, 0},{164994, 1, 1, 16}, {0, 0, 0, 0}}},{58843, {{164994, 1, 0, 0},{164994, 1, 1, 17}, {0, 0, 0, 0}}},{58843, {{164994, 1, 0, 0},{164994, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{50811, {{164995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50812, {{164996, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50814, {{164997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50815, {{164998, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50816, {{164999, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50817, {{165000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{50818, {{165001, 1, 0, 0},{165001, 1, 1, 17}, {0, 0, 0, 0}}},{50818, {{165001, 1, 0, 0},{165001, 1, 1, 17}, {0, 0, 0, 0}}},{50819, {{165001, 1, 0, 0},{165001, 1, 1, 16}, {0, 0, 0, 0}}},{50819, {{165001, 1, 0, 0},{165001, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{50820, {{165002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50821, {{165003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{50822, {{165004, 1, 0, 0},{165004, 1, 1, 17}, {0, 0, 0, 0}}},{50822, {{165004, 1, 0, 0},{165004, 1, 1, 17}, {0, 0, 0, 0}}},{50823, {{165004, 1, 0, 0},{165004, 1, 1, 16}, {0, 0, 0, 0}}},{50823, {{165004, 1, 0, 0},{165004, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50824, {{165006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50825, {{165007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50826, {{165009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50828, {{165018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50829, {{165019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50830, {{165020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50831, {{165021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50832, {{165022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50833, {{165024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50834, {{165025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50835, {{165026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50836, {{165027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50837, {{165028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50838, {{165029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50839, {{165030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50841, {{165031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50842, {{165032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50843, {{165033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50844, {{165034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50845, {{165035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50847, {{165036, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{50848, {{165037, 1, 0, 0},{165037, 1, 1, 17}, {0, 0, 0, 0}}},{50848, {{165037, 1, 0, 0},{165037, 1, 1, 17}, {0, 0, 0, 0}}},{50849, {{165037, 1, 0, 0},{165037, 1, 1, 16}, {0, 0, 0, 0}}},{50849, {{165037, 1, 0, 0},{165037, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{50850, {{165038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50851, {{165039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{50852, {{165040, 1, 0, 0},{165040, 1, 1, 17}, {0, 0, 0, 0}}},{50852, {{165040, 1, 0, 0},{165040, 1, 1, 17}, {0, 0, 0, 0}}},{50853, {{165040, 1, 0, 0},{165040, 1, 1, 16}, {0, 0, 0, 0}}},{50853, {{165040, 1, 0, 0},{165040, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{50854, {{165041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50855, {{165042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50856, {{165043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50857, {{165044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50858, {{165048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50859, {{165052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50860, {{165053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50861, {{165054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50862, {{165055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50863, {{165056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50864, {{165057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50865, {{165058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50866, {{165059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50867, {{165060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{50868, {{165061, 1, 0, 0},{165061, 1, 1, 17}, {0, 0, 0, 0}}},{50868, {{165061, 1, 0, 0},{165061, 1, 1, 17}, {0, 0, 0, 0}}},{50869, {{165061, 1, 0, 0},{165061, 1, 1, 16}, {0, 0, 0, 0}}},{50869, {{165061, 1, 0, 0},{165061, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50870, {{165063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50871, {{165064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50872, {{165065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50874, {{165066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50875, {{165067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50876, {{165068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{50877, {{165069, 1, 0, 0},{165069, 1, 1, 17}, {0, 0, 0, 0}}},{50877, {{165069, 1, 0, 0},{165069, 1, 1, 17}, {0, 0, 0, 0}}},{58849, {{165069, 1, 0, 0},{165069, 1, 1, 16}, {0, 0, 0, 0}}},{58849, {{165069, 1, 0, 0},{165069, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{50878, {{165070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50879, {{165071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58850, {{165077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50880, {{165078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50882, {{165080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50883, {{165081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50884, {{165083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50885, {{165085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50886, {{165086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50887, {{165091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{50888, {{165092, 1, 0, 0},{165092, 1, 1, 17}, {0, 0, 0, 0}}},{50888, {{165092, 1, 0, 0},{165092, 1, 1, 17}, {0, 0, 0, 0}}},{50889, {{165092, 1, 0, 0},{165092, 1, 1, 16}, {0, 0, 0, 0}}},{50889, {{165092, 1, 0, 0},{165092, 1, 1, 16}, {0, 0, 0, 0}}},{50890, {{165092, 1, 0, 0},{165092, 1, 1, 18}, {0, 0, 0, 0}}},{50890, {{165092, 1, 0, 0},{165092, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{50891, {{165093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50892, {{165094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50893, {{165095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50894, {{165096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50895, {{165097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50896, {{165099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50897, {{165100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50898, {{165101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50899, {{165102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50900, {{165103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50901, {{165104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{50902, {{165105, 1, 0, 0},{165105, 1, 1, 16}, {0, 0, 0, 0}}},{50902, {{165105, 1, 0, 0},{165105, 1, 1, 16}, {0, 0, 0, 0}}},{50903, {{165105, 1, 0, 0},{165105, 1, 1, 17}, {0, 0, 0, 0}}},{50903, {{165105, 1, 0, 0},{165105, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{50904, {{165106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50905, {{165107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50906, {{165108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{50907, {{165109, 1, 0, 0},{165109, 1, 1, 17}, {0, 0, 0, 0}}},{50907, {{165109, 1, 0, 0},{165109, 1, 1, 17}, {0, 0, 0, 0}}},{50908, {{165109, 1, 0, 0},{165109, 1, 1, 16}, {0, 0, 0, 0}}},{50908, {{165109, 1, 0, 0},{165109, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{50909, {{165110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50910, {{165111, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50911, {{165112, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50912, {{165113, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50913, {{165114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50914, {{165115, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50915, {{165116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50916, {{165117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50918, {{165118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50919, {{165119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50920, {{165120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50921, {{165122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50922, {{165123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50923, {{165124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50924, {{165125, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50926, {{165131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50927, {{165142, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50928, {{165143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50929, {{165145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50930, {{165146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50931, {{165147, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50932, {{165148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{50933, {{165149, 1, 0, 0},{165149, 1, 1, 17}, {0, 0, 0, 0}}},{50933, {{165149, 1, 0, 0},{165149, 1, 1, 17}, {0, 0, 0, 0}}},{50934, {{165149, 1, 0, 0},{165149, 1, 1, 16}, {0, 0, 0, 0}}},{50934, {{165149, 1, 0, 0},{165149, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{50935, {{165150, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{50936, {{165151, 1, 0, 0},{165151, 1, 1, 17}, {0, 0, 0, 0}}},{50936, {{165151, 1, 0, 0},{165151, 1, 1, 17}, {0, 0, 0, 0}}},{50937, {{165151, 1, 0, 0},{165151, 1, 1, 16}, {0, 0, 0, 0}}},{50937, {{165151, 1, 0, 0},{165151, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50938, {{165153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50940, {{165154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50941, {{165155, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50942, {{165156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50971, {{165157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50943, {{165158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{50944, {{165159, 1, 0, 0},{165159, 1, 1, 17}, {0, 0, 0, 0}}},{50944, {{165159, 1, 0, 0},{165159, 1, 1, 17}, {0, 0, 0, 0}}},{50945, {{165159, 1, 0, 0},{165159, 1, 1, 16}, {0, 0, 0, 0}}},{50945, {{165159, 1, 0, 0},{165159, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{50946, {{165160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50947, {{165161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50948, {{165162, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{50949, {{165163, 1, 0, 0},{165163, 1, 1, 17}, {0, 0, 0, 0}}},{50949, {{165163, 1, 0, 0},{165163, 1, 1, 17}, {0, 0, 0, 0}}},{50950, {{165163, 1, 0, 0},{165163, 1, 1, 16}, {0, 0, 0, 0}}},{50950, {{165163, 1, 0, 0},{165163, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{50951, {{165164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50954, {{165165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50955, {{165166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{50956, {{165167, 1, 0, 0},{165167, 1, 1, 17}, {0, 0, 0, 0}}},{50956, {{165167, 1, 0, 0},{165167, 1, 1, 17}, {0, 0, 0, 0}}},{60228, {{165167, 1, 0, 0},{165167, 1, 1, 16}, {0, 0, 0, 0}}},{60228, {{165167, 1, 0, 0},{165167, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {4, {{50957, {{165168, 1, 0, 0},{165168, 1, 1, 16}, {0, 0, 0, 0}}},{50957, {{165168, 1, 0, 0},{165168, 1, 1, 16}, {0, 0, 0, 0}}},{50958, {{165168, 1, 0, 0},{165168, 1, 1, 17}, {0, 0, 0, 0}}},{50958, {{165168, 1, 0, 0},{165168, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{50959, {{165169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50960, {{165170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50939, {{165173, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50952, {{165174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58854, {{165179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50962, {{165197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50963, {{165199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50964, {{165200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50965, {{165201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50966, {{165204, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50967, {{165205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50968, {{165206, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50969, {{165207, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58859, {{165209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50970, {{165210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50972, {{165212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50973, {{165213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50974, {{165214, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50975, {{165215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{50976, {{165216, 1, 0, 0},{165216, 1, 1, 17}, {0, 0, 0, 0}}},{50976, {{165216, 1, 0, 0},{165216, 1, 1, 17}, {0, 0, 0, 0}}},{50977, {{165216, 1, 0, 0},{165216, 1, 1, 16}, {0, 0, 0, 0}}},{50977, {{165216, 1, 0, 0},{165216, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{50978, {{165217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50979, {{165218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{50980, {{165219, 1, 0, 0},{165219, 1, 1, 17}, {0, 0, 0, 0}}},{50980, {{165219, 1, 0, 0},{165219, 1, 1, 17}, {0, 0, 0, 0}}},{50981, {{165219, 1, 0, 0},{165219, 1, 1, 16}, {0, 0, 0, 0}}},{50981, {{165219, 1, 0, 0},{165219, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{50982, {{165220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{50983, {{165221, 1, 0, 0},{165221, 1, 1, 17}, {0, 0, 0, 0}}},{50983, {{165221, 1, 0, 0},{165221, 1, 1, 17}, {0, 0, 0, 0}}},{50984, {{165221, 1, 0, 0},{165221, 1, 1, 16}, {0, 0, 0, 0}}},{50984, {{165221, 1, 0, 0},{165221, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{50985, {{165222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50986, {{165223, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50987, {{165224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50988, {{165225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50989, {{165226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{50990, {{165227, 1, 0, 0},{165227, 1, 1, 18}, {0, 0, 0, 0}}},{50990, {{165227, 1, 0, 0},{165227, 1, 1, 18}, {0, 0, 0, 0}}},{50991, {{165227, 1, 0, 0},{165227, 1, 1, 17}, {0, 0, 0, 0}}},{50991, {{165227, 1, 0, 0},{165227, 1, 1, 17}, {0, 0, 0, 0}}},{50992, {{165227, 1, 0, 0},{165227, 1, 1, 19}, {0, 0, 0, 0}}},{50992, {{165227, 1, 0, 0},{165227, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50994, {{165239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50995, {{165240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50996, {{165241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50997, {{165242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{50998, {{165244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{50999, {{165245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51000, {{165247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51001, {{165248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51002, {{165249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51003, {{165251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51004, {{165252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51005, {{165254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51006, {{165255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{51008, {{165256, 1, 0, 0},{165256, 1, 1, 16}, {0, 0, 0, 0}}},{51008, {{165256, 1, 0, 0},{165256, 1, 1, 16}, {0, 0, 0, 0}}},{51009, {{165256, 1, 0, 0},{165256, 1, 1, 17}, {0, 0, 0, 0}}},{51009, {{165256, 1, 0, 0},{165256, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{51010, {{165257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{51011, {{165258, 1, 0, 0},{165258, 1, 1, 17}, {0, 0, 0, 0}}},{51011, {{165258, 1, 0, 0},{165258, 1, 1, 17}, {0, 0, 0, 0}}},{51012, {{165258, 1, 0, 0},{165258, 1, 1, 16}, {0, 0, 0, 0}}},{51012, {{165258, 1, 0, 0},{165258, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{51013, {{165259, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51014, {{165260, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51015, {{165261, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51016, {{165262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51017, {{165263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51018, {{165264, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51019, {{165275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51020, {{165278, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51021, {{165279, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51022, {{165280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51023, {{165281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51024, {{165282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51025, {{165283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51027, {{165284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51028, {{165285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51029, {{165290, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51030, {{165291, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51031, {{165292, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51032, {{165293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51033, {{165294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51034, {{165295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51035, {{165296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51036, {{165297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51037, {{165299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51038, {{165300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51039, {{165301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51040, {{165302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51042, {{165303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51043, {{165304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{51044, {{165305, 1, 0, 0},{165305, 1, 1, 17}, {0, 0, 0, 0}}},{51044, {{165305, 1, 0, 0},{165305, 1, 1, 17}, {0, 0, 0, 0}}},{51045, {{165305, 1, 0, 0},{165305, 1, 1, 16}, {0, 0, 0, 0}}},{51045, {{165305, 1, 0, 0},{165305, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{51046, {{165307, 1, 0, 0},{165307, 1, 1, 17}, {0, 0, 0, 0}}},{51046, {{165307, 1, 0, 0},{165307, 1, 1, 17}, {0, 0, 0, 0}}},{51047, {{165307, 1, 0, 0},{165307, 1, 1, 16}, {0, 0, 0, 0}}},{51047, {{165307, 1, 0, 0},{165307, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{51048, {{165308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51049, {{165310, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{51051, {{165311, 1, 0, 0},{165311, 1, 1, 17}, {0, 0, 0, 0}}},{51051, {{165311, 1, 0, 0},{165311, 1, 1, 17}, {0, 0, 0, 0}}},{51052, {{165311, 1, 0, 0},{165311, 1, 1, 16}, {0, 0, 0, 0}}},{51052, {{165311, 1, 0, 0},{165311, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{51053, {{165312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51054, {{165313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51055, {{165314, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51056, {{165315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51057, {{165316, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51058, {{165320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{51060, {{165321, 1, 0, 0},{165321, 1, 1, 18}, {0, 0, 0, 0}}},{51060, {{165321, 1, 0, 0},{165321, 1, 1, 18}, {0, 0, 0, 0}}},{51061, {{165321, 1, 0, 0},{165321, 1, 1, 17}, {0, 0, 0, 0}}},{51061, {{165321, 1, 0, 0},{165321, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51062, {{165323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51063, {{165324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51064, {{165328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51065, {{165330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51066, {{165331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51067, {{165332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51068, {{165333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51069, {{165335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51070, {{165336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51071, {{165337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51072, {{165341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51073, {{165342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51074, {{165343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51075, {{165344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51076, {{165345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51078, {{165346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51079, {{165347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51080, {{165349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51081, {{165350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51082, {{165351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51083, {{165352, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51084, {{165354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51086, {{165355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51087, {{165356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{51088, {{165357, 1, 0, 0},{165357, 1, 1, 16}, {0, 0, 0, 0}}},{51088, {{165357, 1, 0, 0},{165357, 1, 1, 16}, {0, 0, 0, 0}}},{51089, {{165357, 1, 0, 0},{165357, 1, 1, 17}, {0, 0, 0, 0}}},{51089, {{165357, 1, 0, 0},{165357, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{51090, {{165358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51091, {{165359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51092, {{165360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{51093, {{165361, 1, 0, 0},{165361, 1, 1, 17}, {0, 0, 0, 0}}},{51093, {{165361, 1, 0, 0},{165361, 1, 1, 17}, {0, 0, 0, 0}}},{51094, {{165361, 1, 0, 0},{165361, 1, 1, 16}, {0, 0, 0, 0}}},{51094, {{165361, 1, 0, 0},{165361, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51097, {{165370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51098, {{165371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51099, {{165372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51100, {{165373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51101, {{165377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51102, {{165378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51103, {{165379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51104, {{165380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51105, {{165383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51106, {{165384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51107, {{165385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51108, {{165386, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51109, {{165388, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51111, {{165393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51112, {{165395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51113, {{165396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51114, {{165401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51115, {{165402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51116, {{165403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51117, {{165404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51118, {{165405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51119, {{165406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51120, {{165407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51121, {{165408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51122, {{165409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{51123, {{165410, 1, 0, 0},{165410, 1, 1, 17}, {0, 0, 0, 0}}},{51123, {{165410, 1, 0, 0},{165410, 1, 1, 17}, {0, 0, 0, 0}}},{51124, {{165410, 1, 0, 0},{165410, 1, 1, 16}, {0, 0, 0, 0}}},{51124, {{165410, 1, 0, 0},{165410, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{51125, {{165411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51126, {{165418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51127, {{165419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51128, {{165420, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51129, {{165421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51130, {{165422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51131, {{165423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51132, {{165424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51133, {{165426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51134, {{165427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51135, {{165428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51136, {{165429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{51137, {{165431, 1, 0, 0},{165431, 1, 1, 17}, {0, 0, 0, 0}}},{51137, {{165431, 1, 0, 0},{165431, 1, 1, 17}, {0, 0, 0, 0}}},{51138, {{165431, 1, 0, 0},{165431, 1, 1, 16}, {0, 0, 0, 0}}},{51138, {{165431, 1, 0, 0},{165431, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{51139, {{165432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51140, {{165433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51141, {{165436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51142, {{165438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51143, {{165439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51144, {{165440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51145, {{165441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{51146, {{165442, 1, 0, 0},{165442, 1, 1, 17}, {0, 0, 0, 0}}},{51146, {{165442, 1, 0, 0},{165442, 1, 1, 17}, {0, 0, 0, 0}}},{51147, {{165442, 1, 0, 0},{165442, 1, 1, 16}, {0, 0, 0, 0}}},{51147, {{165442, 1, 0, 0},{165442, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{51148, {{165443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51149, {{165444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51151, {{165445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51152, {{165446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51153, {{165454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51155, {{165455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51156, {{165456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51157, {{165457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51158, {{165458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51159, {{165459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51160, {{165460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{51161, {{165461, 1, 0, 0},{165461, 1, 1, 17}, {0, 0, 0, 0}}},{51161, {{165461, 1, 0, 0},{165461, 1, 1, 17}, {0, 0, 0, 0}}},{51162, {{165461, 1, 0, 0},{165461, 1, 1, 16}, {0, 0, 0, 0}}},{51162, {{165461, 1, 0, 0},{165461, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51163, {{165463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51164, {{165464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{51165, {{165465, 1, 0, 0},{165465, 1, 1, 17}, {0, 0, 0, 0}}},{51165, {{165465, 1, 0, 0},{165465, 1, 1, 17}, {0, 0, 0, 0}}},{51166, {{165465, 1, 0, 0},{165465, 1, 1, 16}, {0, 0, 0, 0}}},{51166, {{165465, 1, 0, 0},{165465, 1, 1, 16}, {0, 0, 0, 0}}},{58876, {{165465, 1, 0, 0},{165465, 1, 1, 18}, {0, 0, 0, 0}}},{58876, {{165465, 1, 0, 0},{165465, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{51167, {{165466, 1, 0, 0},{165466, 1, 1, 17}, {0, 0, 0, 0}}},{51167, {{165466, 1, 0, 0},{165466, 1, 1, 17}, {0, 0, 0, 0}}},{51168, {{165466, 1, 0, 0},{165466, 1, 1, 16}, {0, 0, 0, 0}}},{51168, {{165466, 1, 0, 0},{165466, 1, 1, 16}, {0, 0, 0, 0}}},{51169, {{165466, 1, 0, 0},{165466, 1, 1, 18}, {0, 0, 0, 0}}},{51169, {{165466, 1, 0, 0},{165466, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51170, {{165468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51171, {{165470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{51172, {{165471, 1, 0, 0},{165471, 1, 1, 17}, {0, 0, 0, 0}}},{51172, {{165471, 1, 0, 0},{165471, 1, 1, 17}, {0, 0, 0, 0}}},{51173, {{165471, 1, 0, 0},{165471, 1, 1, 16}, {0, 0, 0, 0}}},{51173, {{165471, 1, 0, 0},{165471, 1, 1, 16}, {0, 0, 0, 0}}},{51174, {{165471, 1, 0, 0},{165471, 1, 1, 18}, {0, 0, 0, 0}}},{51174, {{165471, 1, 0, 0},{165471, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{51175, {{165472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51176, {{165473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51177, {{165475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51178, {{165476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51179, {{165478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51180, {{165479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51181, {{165480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51182, {{165481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51183, {{165483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51184, {{165485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51185, {{165486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51186, {{165487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51187, {{165488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51188, {{165489, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51189, {{165490, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51190, {{165491, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51191, {{165493, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51192, {{165494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51193, {{165495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51194, {{165496, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51195, {{165497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51196, {{165498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51197, {{165499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51198, {{165500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51199, {{165501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51200, {{165502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51201, {{165506, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51202, {{165514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51203, {{165515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51204, {{165516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51205, {{165519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51206, {{165521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51207, {{165522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51208, {{165523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51209, {{165524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51210, {{165525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51211, {{165526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51212, {{165527, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51213, {{165528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51214, {{165529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51215, {{165530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51216, {{165531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51217, {{165536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51218, {{165538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51219, {{165539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51220, {{165541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51221, {{165543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51222, {{165544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51223, {{165546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51224, {{165548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51225, {{165549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51226, {{165550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51227, {{165551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51228, {{165552, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51229, {{165553, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51230, {{165554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51231, {{165555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51232, {{165556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51233, {{165557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51234, {{165558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51235, {{165562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51236, {{165563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51237, {{165564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51238, {{165565, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51239, {{165566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51240, {{165567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51241, {{165569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51242, {{165571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51243, {{165572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51244, {{165574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51245, {{165575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51246, {{165576, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51247, {{165577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51248, {{165578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51249, {{165579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51250, {{165580, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51251, {{165581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51252, {{165582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51253, {{165583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51254, {{165584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51255, {{165585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51256, {{165586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51257, {{165587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51258, {{165588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51259, {{165589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51260, {{165590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51261, {{165591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51262, {{165593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51263, {{165594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51264, {{165595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51265, {{165596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51266, {{165598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51267, {{165599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51268, {{165600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51269, {{165601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51270, {{165602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51271, {{165603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51272, {{165604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51273, {{165605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51274, {{165606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51275, {{165607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51276, {{165611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51277, {{165612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51278, {{165613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51279, {{165614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51280, {{165615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51281, {{165616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51282, {{165617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51283, {{165619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51284, {{165620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51285, {{165621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51286, {{165622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51287, {{165623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51288, {{165624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51289, {{165626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51290, {{165627, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51293, {{165631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51295, {{165632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51297, {{165634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51298, {{165635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51299, {{165637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51300, {{165638, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51301, {{165640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51302, {{165643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51303, {{165644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51304, {{165645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51305, {{165646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51306, {{165647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51307, {{165648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51308, {{165649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51309, {{165650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51310, {{165651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51311, {{165652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51312, {{165653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51313, {{165654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51314, {{165655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51315, {{165656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51316, {{165657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51317, {{165658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51318, {{165659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51319, {{165660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51320, {{165663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51321, {{165667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51322, {{165668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51323, {{165669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51324, {{165670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51325, {{165671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51326, {{165672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51327, {{165674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51328, {{165678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51329, {{165679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51330, {{165680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51331, {{165681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51332, {{165682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51333, {{165683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51334, {{165684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51335, {{165685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51336, {{165686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51337, {{165687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51338, {{165688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51339, {{165689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51340, {{165690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51341, {{165691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51342, {{165692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51343, {{165693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51344, {{165694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51345, {{165695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51346, {{165696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51347, {{165697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51348, {{165702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51349, {{165703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51350, {{165704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51351, {{165705, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51352, {{165706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51354, {{165707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51355, {{165708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51356, {{165709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51357, {{165710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51358, {{165711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51359, {{165712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51360, {{165713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51361, {{165715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51362, {{165716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51363, {{165717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51364, {{165718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51365, {{165719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51366, {{165720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51367, {{165723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51368, {{165724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51369, {{165725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51370, {{165726, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51371, {{165727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51372, {{165728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51373, {{165730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{51374, {{165731, 1, 0, 0},{165731, 1, 1, 16}, {0, 0, 0, 0}}},{51374, {{165731, 1, 0, 0},{165731, 1, 1, 16}, {0, 0, 0, 0}}},{58877, {{165731, 1, 0, 0},{165731, 1, 1, 17}, {0, 0, 0, 0}}},{58877, {{165731, 1, 0, 0},{165731, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{51375, {{165732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51376, {{165733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51377, {{165734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51378, {{165735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51379, {{165736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51380, {{165737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51381, {{165738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51382, {{165739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51383, {{165740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51384, {{165741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51385, {{165742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51386, {{165743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51387, {{165744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51388, {{165745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51389, {{165746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51390, {{165747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51391, {{165748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51392, {{165749, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51393, {{165750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51394, {{165751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51395, {{165752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51396, {{165760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51397, {{165761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51398, {{165762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51399, {{165763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51400, {{165764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51401, {{165765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51402, {{165769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51403, {{165770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51404, {{165771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51405, {{165772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51406, {{165774, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51407, {{165776, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51408, {{165777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51409, {{165779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51410, {{165780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51411, {{165781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51412, {{165782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51413, {{165783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51414, {{165784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51415, {{165785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51416, {{165786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51417, {{165787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51418, {{165792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51419, {{165793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51420, {{165795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51421, {{165798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51422, {{165799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51423, {{165800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51424, {{165801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51425, {{165802, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51426, {{165803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51427, {{165806, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51428, {{165807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51429, {{165808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51430, {{165809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51431, {{165810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51432, {{165811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51433, {{165813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51434, {{165814, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51435, {{165815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51436, {{165817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51437, {{165818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51438, {{165819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51439, {{165820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51440, {{165821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51441, {{165822, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51442, {{165823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51443, {{165824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51444, {{165826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51445, {{165827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51446, {{165828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51447, {{165829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51448, {{165836, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51449, {{165837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51450, {{165838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51451, {{165839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51452, {{165840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51453, {{165841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51454, {{165844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51455, {{165845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51456, {{165846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51457, {{165847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51458, {{165848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51459, {{165849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51460, {{165850, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51461, {{165851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51462, {{165852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51463, {{165854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51464, {{165855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51465, {{165856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51466, {{165858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51467, {{165859, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51468, {{165860, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51469, {{165861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51470, {{165862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51471, {{165864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51472, {{165866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51473, {{165867, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51474, {{165868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51475, {{165870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51476, {{165871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51477, {{165872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51478, {{165874, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51479, {{165876, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51480, {{165877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51481, {{165878, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51482, {{165879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51483, {{165880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51484, {{165881, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51485, {{165882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51486, {{165883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51487, {{165884, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51488, {{165885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51489, {{165886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51490, {{165887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51491, {{165888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51492, {{165889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51493, {{165890, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51494, {{165894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51495, {{165895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51496, {{165896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51497, {{165898, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{51498, {{165899, 1, 0, 0},{165899, 1, 1, 16}, {0, 0, 0, 0}}},{51498, {{165899, 1, 0, 0},{165899, 1, 1, 16}, {0, 0, 0, 0}}},{51499, {{165899, 1, 0, 0},{165899, 1, 1, 17}, {0, 0, 0, 0}}},{51499, {{165899, 1, 0, 0},{165899, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{51500, {{165900, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51501, {{165901, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51502, {{165902, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51503, {{165903, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51504, {{165904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51505, {{165905, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51506, {{165906, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51507, {{165908, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51508, {{165909, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51509, {{165910, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51510, {{165911, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51511, {{165912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51512, {{165919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51513, {{165921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51514, {{165922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51515, {{165924, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51516, {{165925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51517, {{165926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51518, {{165927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51519, {{165932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51520, {{165934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51521, {{165942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51522, {{165943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51523, {{165944, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51524, {{165949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51525, {{165950, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51526, {{165951, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51527, {{165954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51528, {{165955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51529, {{165956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51530, {{165963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51531, {{165964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51532, {{165965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51533, {{165966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51534, {{165967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51535, {{165968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51536, {{165969, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51537, {{165970, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51538, {{165971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51539, {{165972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51540, {{165973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51541, {{165974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51542, {{165976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51543, {{165984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51544, {{165985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51545, {{165986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51546, {{165987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51547, {{165988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51548, {{165993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51549, {{165994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51550, {{165995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51551, {{165996, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51552, {{165997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51553, {{165998, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51554, {{165999, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51555, {{166000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51556, {{166001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51557, {{166002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51558, {{166003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51559, {{166004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51560, {{166005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51561, {{166006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51562, {{166007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51563, {{166008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58889, {{166011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51564, {{166016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51565, {{166017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51566, {{166019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51567, {{166021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51568, {{166022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51569, {{166023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51570, {{166024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51571, {{166025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51572, {{166026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51573, {{166027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51574, {{166028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51575, {{166029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51576, {{166030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51577, {{166031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51578, {{166032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51579, {{166034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51580, {{166041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51581, {{166042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51582, {{166043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51583, {{166048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51584, {{166049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51585, {{166050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51586, {{166051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51587, {{166053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51588, {{166054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51589, {{166056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51590, {{166057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51591, {{166058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51592, {{166059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51593, {{166060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51594, {{166061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51597, {{166062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51595, {{166063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51596, {{166064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51598, {{166066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51599, {{166069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51600, {{166070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51601, {{166073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51602, {{166074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51603, {{166077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51604, {{166078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51605, {{166079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51606, {{166080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51607, {{166081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51608, {{166082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51609, {{166083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51610, {{166084, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51611, {{166085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51612, {{166086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51613, {{166087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51614, {{166088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51615, {{166091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51616, {{166093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51617, {{166094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51618, {{166096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51619, {{166097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51620, {{166099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51621, {{166100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51622, {{166101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51623, {{166102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51624, {{166104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51625, {{166105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51626, {{166106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51627, {{166107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51628, {{166108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51629, {{166109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51630, {{166112, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51631, {{166113, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51632, {{166114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51633, {{166115, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51634, {{166118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51635, {{166119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51636, {{166120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51637, {{166121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51638, {{166122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51639, {{166124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51640, {{166125, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51641, {{166126, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51642, {{166127, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51643, {{166128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51644, {{166129, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51645, {{166130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51646, {{166131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51647, {{166133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51648, {{166134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51649, {{166135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51650, {{166136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51651, {{166137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51652, {{166138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51653, {{166139, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51654, {{166140, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51655, {{166142, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51656, {{166143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51657, {{166145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51658, {{166146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51659, {{166148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51660, {{166149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51661, {{166151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51662, {{166152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51663, {{166153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51664, {{166154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51665, {{166156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51666, {{166157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51667, {{166158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51668, {{166159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51669, {{166161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51670, {{166162, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51671, {{166163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51672, {{166165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51673, {{166166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51674, {{166167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51675, {{166168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51676, {{166169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51677, {{166170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51678, {{166173, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51679, {{166177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51680, {{166178, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51681, {{166179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51682, {{166181, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51683, {{166182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51684, {{166184, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51685, {{166185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51686, {{166187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51687, {{166188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51688, {{166189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51689, {{166193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51690, {{166196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51691, {{166199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51692, {{166201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51693, {{166202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51694, {{166205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51695, {{166206, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51696, {{166207, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51697, {{166208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51699, {{166210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51700, {{166211, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51701, {{166212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{51702, {{166213, 1, 0, 0},{166213, 1, 1, 16}, {0, 0, 0, 0}}},{51702, {{166213, 1, 0, 0},{166213, 1, 1, 16}, {0, 0, 0, 0}}},{60262, {{166213, 1, 0, 0},{166213, 1, 1, 17}, {0, 0, 0, 0}}},{60262, {{166213, 1, 0, 0},{166213, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{51703, {{166214, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51704, {{166215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51705, {{166217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51706, {{166218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51707, {{166219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51708, {{166220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51709, {{166221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51710, {{166222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51711, {{166223, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51712, {{166224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51713, {{166225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51714, {{166226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51715, {{166233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51716, {{166238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51717, {{166242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51718, {{166243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51720, {{166245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51721, {{166246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51722, {{166247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51723, {{166248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51724, {{166249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51725, {{166250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51726, {{166251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51727, {{166259, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51728, {{166261, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51729, {{166265, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51731, {{166267, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51732, {{166274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51733, {{166275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51734, {{166276, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51735, {{166277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51736, {{166278, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51737, {{166279, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51738, {{166280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51741, {{166294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51742, {{166297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51743, {{166300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51744, {{166301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51745, {{166302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51746, {{166303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51747, {{166305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51748, {{166307, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58909, {{166315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51750, {{166316, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51751, {{166317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51752, {{166318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51753, {{166319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51754, {{166320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51755, {{166321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51756, {{166322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51757, {{166323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51758, {{166324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51759, {{166325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51760, {{166326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51761, {{166327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51762, {{166328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51763, {{166330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51764, {{166331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51765, {{166335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51766, {{166350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51767, {{166351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51768, {{166355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51769, {{166357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51770, {{166358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51771, {{166359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51772, {{166361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51774, {{166373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51775, {{166374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51776, {{166375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51777, {{166376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51778, {{166377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51779, {{166378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51780, {{166379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51781, {{166380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51782, {{166381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51783, {{166382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51785, {{166383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51786, {{166384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51787, {{166385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51788, {{166386, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51789, {{166387, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51790, {{166389, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51791, {{166390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51793, {{166399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51794, {{166425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51795, {{166428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51796, {{166430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51797, {{166431, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51798, {{166432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51799, {{166441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51800, {{166444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51801, {{166445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51802, {{166446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51803, {{166447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51804, {{166448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51805, {{166450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51806, {{166451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51807, {{166452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51808, {{166453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51809, {{166454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51810, {{166455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51811, {{166456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51812, {{166457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51813, {{166458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51814, {{166459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51815, {{166460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51816, {{166461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51817, {{166462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51818, {{166467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51819, {{166480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51820, {{166487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51821, {{166496, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51822, {{166501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51823, {{166503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51824, {{166505, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51825, {{166508, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51826, {{166511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{51827, {{166513, 1, 0, 0},{166513, 1, 1, 17}, {0, 0, 0, 0}}},{51827, {{166513, 1, 0, 0},{166513, 1, 1, 17}, {0, 0, 0, 0}}},{51828, {{166513, 1, 0, 0},{166513, 1, 1, 18}, {0, 0, 0, 0}}},{51828, {{166513, 1, 0, 0},{166513, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51830, {{166517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60283, {{166519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51831, {{166521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51832, {{166522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51835, {{166532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51836, {{166533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51838, {{166535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51839, {{166536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51840, {{166537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51841, {{166538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51842, {{166539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51844, {{166540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51845, {{166541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51846, {{166542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51847, {{166543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51848, {{166544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51849, {{166545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51850, {{166546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51851, {{166548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51852, {{166549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51853, {{166550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51854, {{166551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51855, {{166553, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51856, {{166559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51857, {{166582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51858, {{166583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51859, {{166585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51860, {{166588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51861, {{166589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51863, {{166605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51864, {{166607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51865, {{166608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51866, {{166609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51867, {{166610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51868, {{166611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51869, {{166612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51870, {{166613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51871, {{166615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51872, {{166616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51873, {{166617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51874, {{166621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51876, {{166628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58935, {{166642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51877, {{166649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51878, {{166651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51879, {{166652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51880, {{166653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51882, {{166657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51883, {{166659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51884, {{166660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51885, {{166669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51886, {{166671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51887, {{166672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51888, {{166673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51890, {{166674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51891, {{166675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51892, {{166676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51893, {{166678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51894, {{166679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51895, {{166680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51896, {{166681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51897, {{166682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51898, {{166683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51899, {{166684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51900, {{166685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51901, {{166687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58940, {{166713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51902, {{166719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51903, {{166721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51904, {{166722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51905, {{166724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51906, {{166725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51907, {{166727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51908, {{166734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51910, {{166739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51911, {{166741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51912, {{166742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51913, {{166743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51914, {{166744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51915, {{166745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51916, {{166746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51917, {{166747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51918, {{166748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51919, {{166749, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51920, {{166750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58952, {{166760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51921, {{166762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51923, {{166764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51924, {{166780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51925, {{166782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51926, {{166784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51927, {{166787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51928, {{166788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51929, {{166802, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51931, {{166803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51932, {{166804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51933, {{166805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51934, {{166806, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51935, {{166807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51936, {{166810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51937, {{166827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51938, {{166831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51939, {{166839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51940, {{166840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51941, {{166841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51942, {{166843, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51943, {{166844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51944, {{166845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51945, {{166846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51946, {{166847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51947, {{166849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51949, {{166866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51950, {{166867, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51951, {{166868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51952, {{166877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51953, {{166878, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51954, {{166879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51955, {{166880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51956, {{166881, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51957, {{166882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51958, {{166883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{51959, {{166895, 1, 0, 0},{166895, 1, 1, 17}, {0, 0, 0, 0}}},{51959, {{166895, 1, 0, 0},{166895, 1, 1, 17}, {0, 0, 0, 0}}},{51960, {{166895, 1, 0, 0},{166895, 1, 1, 18}, {0, 0, 0, 0}}},{51960, {{166895, 1, 0, 0},{166895, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51961, {{166898, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51962, {{166908, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58953, {{166915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51963, {{166917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51964, {{166918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51965, {{166921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51966, {{166923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51967, {{166924, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51968, {{166925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51969, {{166926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51970, {{166927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51972, {{166928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51973, {{166929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51974, {{166930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51976, {{166931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51975, {{166940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51977, {{166942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51978, {{166943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51979, {{166944, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51980, {{166948, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51981, {{166950, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51982, {{166953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51983, {{166957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51984, {{166958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51985, {{166959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51986, {{166962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51987, {{166963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51988, {{166965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51989, {{166968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51990, {{166971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51991, {{166999, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51992, {{167000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51993, {{167001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51994, {{167002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51995, {{167003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51996, {{167004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51997, {{167005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{51998, {{167006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51999, {{167008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52000, {{167009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52001, {{167010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52002, {{167011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52003, {{167012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52004, {{167013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52005, {{167014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52006, {{167015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52008, {{167016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52009, {{167019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52010, {{167020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52011, {{167021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52012, {{167022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52013, {{167023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52014, {{167024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52015, {{167025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52016, {{167027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52017, {{167028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52018, {{167029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52019, {{167030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52020, {{167031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52021, {{167032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52022, {{167034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52023, {{167035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52024, {{167036, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52025, {{167039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52026, {{167040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52027, {{167041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52028, {{167042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52029, {{167043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52030, {{167044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52031, {{167046, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52032, {{167047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52033, {{167048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52034, {{167049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52035, {{167050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52036, {{167051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52037, {{167052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52038, {{167054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52039, {{167056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52040, {{167057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52041, {{167058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52042, {{167059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52043, {{167060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52044, {{167061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52045, {{167063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52046, {{167064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52047, {{167065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52048, {{167066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52049, {{167067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52050, {{167068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52051, {{167069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52052, {{167070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52053, {{167071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52054, {{167072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52055, {{167073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52056, {{167074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52057, {{167077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52058, {{167078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52059, {{167079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52060, {{167080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52061, {{167081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52062, {{167082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52063, {{167083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52064, {{167084, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52065, {{167085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52066, {{167086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52067, {{167087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52068, {{167088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52069, {{167089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52070, {{167090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52071, {{167091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52072, {{167092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52073, {{167093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52074, {{167094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52075, {{167096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52076, {{167098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52077, {{167099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52078, {{167100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52079, {{167102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52080, {{167103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52081, {{167104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52082, {{167105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52084, {{167106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52085, {{167107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52086, {{167108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52087, {{167109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52088, {{167113, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52089, {{167114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52090, {{167115, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52091, {{167116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52092, {{167120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52093, {{167121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52094, {{167125, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52095, {{167126, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52096, {{167127, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52097, {{167128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52098, {{167129, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52099, {{167130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52100, {{167131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52102, {{167132, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{27427, {{167133, 1, 0, 0},{167133, 1, 1, 17}, {0, 0, 0, 0}}},{27427, {{167133, 1, 0, 0},{167133, 1, 1, 17}, {0, 0, 0, 0}}},{52103, {{167133, 1, 0, 0},{167133, 1, 1, 18}, {0, 0, 0, 0}}},{52103, {{167133, 1, 0, 0},{167133, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{52104, {{167134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52105, {{167135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52106, {{167136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52107, {{167137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52108, {{167138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52109, {{167139, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52110, {{167141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52111, {{167151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52112, {{167154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52113, {{167155, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52114, {{167157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52115, {{167158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52116, {{167159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52117, {{167160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52118, {{167161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52119, {{167162, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52120, {{167163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52121, {{167164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52122, {{167165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52123, {{167166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52124, {{167167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52125, {{167168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52126, {{167169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52127, {{167170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52129, {{167174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52130, {{167181, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52131, {{167182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52132, {{167183, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52133, {{167184, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52134, {{167185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52135, {{167186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52136, {{167187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52137, {{167188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52138, {{167189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52139, {{167190, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52140, {{167191, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52141, {{167192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52142, {{167193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52143, {{167194, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52144, {{167195, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52145, {{167197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52146, {{167198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52147, {{167199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52148, {{167200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52149, {{167201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52150, {{167203, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52151, {{167205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52152, {{167208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52153, {{167209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52154, {{167210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52155, {{167212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52156, {{167213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52157, {{167214, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52158, {{167215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52159, {{167216, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52160, {{167217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52161, {{167219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52163, {{167224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52164, {{167226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52165, {{167227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52166, {{167228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52167, {{167230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52168, {{167231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52169, {{167232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52170, {{167233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52171, {{167234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52172, {{167235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52173, {{167237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52174, {{167238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52175, {{167240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52176, {{167241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52177, {{167242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52178, {{167243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52179, {{167244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52180, {{167245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52181, {{167247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52182, {{167248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52183, {{167249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52184, {{167250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52185, {{167251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52186, {{167255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52187, {{167256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52188, {{167258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52189, {{167259, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52190, {{167260, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52191, {{167262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52192, {{167263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52193, {{167264, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52194, {{167265, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52195, {{167268, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52196, {{167269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52197, {{167270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52198, {{167271, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52199, {{167272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52200, {{167273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52201, {{167274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52202, {{167275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52203, {{167276, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52204, {{167277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52205, {{167278, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52206, {{167280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52207, {{167281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52208, {{167284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52209, {{167285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52210, {{167286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52211, {{167287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52212, {{167289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52213, {{167292, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60311, {{167295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52214, {{167296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52215, {{167297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52216, {{167298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52217, {{167299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{52218, {{167300, 1, 0, 0},{167300, 1, 1, 16}, {0, 0, 0, 0}}},{52218, {{167300, 1, 0, 0},{167300, 1, 1, 16}, {0, 0, 0, 0}}},{52219, {{167300, 1, 0, 0},{167300, 1, 1, 17}, {0, 0, 0, 0}}},{52219, {{167300, 1, 0, 0},{167300, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{52220, {{167301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52221, {{167302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52222, {{167303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52223, {{167304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52224, {{167305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52225, {{167306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52226, {{167307, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{27485, {{167313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52227, {{167318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52248, {{167320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52228, {{167321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52229, {{167322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52230, {{167324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52231, {{167325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52232, {{167326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52233, {{167327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52234, {{167328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52235, {{167329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52236, {{167330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52237, {{167336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52238, {{167340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52239, {{167341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52240, {{167342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52242, {{167343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52243, {{167344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52244, {{167345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52245, {{167346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52246, {{167347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52247, {{167350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52249, {{167353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52250, {{167354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52251, {{167356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52252, {{167357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52253, {{167358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52254, {{167361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52255, {{167362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52256, {{167363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52257, {{167364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52258, {{167365, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52259, {{167366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52260, {{167367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52261, {{167374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52262, {{167376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52263, {{167378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52264, {{167379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52265, {{167380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52266, {{167381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52267, {{167384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52268, {{167385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52269, {{167396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52270, {{167397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52271, {{167398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52272, {{167401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52273, {{167402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52274, {{167403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52275, {{167404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52276, {{167406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52277, {{167408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52278, {{167409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52279, {{167410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52280, {{167411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52281, {{167415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52282, {{167416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52283, {{167419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52284, {{167421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52285, {{167422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52286, {{167439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52287, {{167440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52288, {{167441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52289, {{167442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52290, {{167443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52291, {{167444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52292, {{167446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52293, {{167447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52294, {{167448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52295, {{167449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52296, {{167450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52297, {{167451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52298, {{167452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52299, {{167453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52300, {{167454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52301, {{167455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{30363, {{167461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52302, {{167465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52303, {{167466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52304, {{167467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52305, {{167468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52306, {{167469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52307, {{167470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52308, {{167471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52309, {{167472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52310, {{167473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52311, {{167474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52312, {{167475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52313, {{167478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52314, {{167482, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52315, {{167483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52316, {{167484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52317, {{167488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52318, {{167489, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52319, {{167490, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52320, {{167491, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52321, {{167492, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52322, {{167493, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52323, {{167494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52324, {{167495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52325, {{167496, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52326, {{167497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52327, {{167499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52328, {{167500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52329, {{167503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52330, {{167507, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52331, {{167509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52332, {{167512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52333, {{167513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52334, {{167514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52335, {{167515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52336, {{167516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52337, {{167517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52338, {{167518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52339, {{167519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52340, {{167520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52341, {{167521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52342, {{167522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52343, {{167527, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52344, {{167528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52345, {{167541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52346, {{167542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52347, {{167543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52348, {{167544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52349, {{167545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52350, {{167546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52351, {{167547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52352, {{167548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52353, {{167549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52354, {{167550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52355, {{167551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52356, {{167552, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52357, {{167553, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52358, {{167554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52359, {{167555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52360, {{167556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52361, {{167557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52362, {{167558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52363, {{167561, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52365, {{167574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60317, {{167575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52366, {{167576, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52367, {{167580, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52368, {{167581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52369, {{167583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52370, {{167584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52371, {{167585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52372, {{167586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52373, {{167587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52374, {{167588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52375, {{167589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52376, {{167590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52377, {{167591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52378, {{167592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52379, {{167593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52380, {{167594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52381, {{167595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52382, {{167596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52383, {{167597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52384, {{167598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52385, {{167599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52386, {{167605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52387, {{167608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52388, {{167611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52389, {{167615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52390, {{167616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52391, {{167617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52392, {{167618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52393, {{167619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52394, {{167620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52395, {{167621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52396, {{167622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52397, {{167623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52398, {{167624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52399, {{167625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52400, {{167626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52401, {{167627, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52402, {{167628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52403, {{167629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52404, {{167631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52405, {{167632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52406, {{167633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52407, {{167634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52408, {{167635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52409, {{167636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52410, {{167637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52411, {{167638, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52412, {{167639, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52413, {{167640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52414, {{167654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52415, {{167656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52416, {{167658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52417, {{167659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52418, {{167663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52419, {{167664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52420, {{167665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52421, {{167666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52422, {{167667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52423, {{167668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52424, {{167669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52425, {{167670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52426, {{167671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52427, {{167672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52428, {{167673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52429, {{167674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52430, {{167675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52431, {{167676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52432, {{167677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52433, {{167678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52434, {{167679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52435, {{167686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52436, {{167688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52437, {{167690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52438, {{167691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52439, {{167692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52440, {{167693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52441, {{167694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52442, {{167695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52443, {{167696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52444, {{167697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52445, {{167698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52446, {{167699, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52447, {{167700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52448, {{167701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52449, {{167702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52450, {{167703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52451, {{167704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52452, {{167705, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52453, {{167717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52454, {{167718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52455, {{167722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52456, {{167723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52457, {{167724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52458, {{167725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52459, {{167726, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52460, {{167727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52461, {{167728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52462, {{167729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52463, {{167730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52464, {{167741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52465, {{167742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52466, {{167743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52467, {{167744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{52468, {{167745, 1, 0, 0},{167745, 1, 1, 16}, {0, 0, 0, 0}}},{52468, {{167745, 1, 0, 0},{167745, 1, 1, 16}, {0, 0, 0, 0}}},{52469, {{167745, 1, 0, 0},{167745, 1, 1, 17}, {0, 0, 0, 0}}},{52469, {{167745, 1, 0, 0},{167745, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{52470, {{167746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52471, {{167747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52472, {{167748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52473, {{167757, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52474, {{167758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52475, {{167759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52476, {{167760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52477, {{167761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52478, {{167766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52479, {{167767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52480, {{167768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52481, {{167769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52482, {{167773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52483, {{167775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52484, {{167776, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52485, {{167777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52486, {{167781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52487, {{167782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52488, {{167783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52489, {{167784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52490, {{167786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52491, {{167787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52492, {{167788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52493, {{167790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52494, {{167791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52495, {{167792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52496, {{167794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52497, {{167796, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52498, {{167797, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52499, {{167798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52500, {{167799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52501, {{167800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52502, {{167801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52503, {{167804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52504, {{167805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52505, {{167807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52506, {{167808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52507, {{167809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52508, {{167810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52509, {{167811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52510, {{167812, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52511, {{167813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52512, {{167814, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52513, {{167815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52514, {{167816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52515, {{167817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52516, {{167818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52517, {{167819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52518, {{167820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52519, {{167821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52520, {{167823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52521, {{167825, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52522, {{167826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52523, {{167827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52524, {{167828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52525, {{167829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52526, {{167830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52527, {{167831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52528, {{167832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52529, {{167833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52530, {{167834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52531, {{167835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52532, {{167836, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52533, {{167837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52534, {{167841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52535, {{167842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52536, {{167843, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52537, {{167844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52538, {{167845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52539, {{167846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52540, {{167847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52541, {{167848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52542, {{167849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52543, {{167850, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52544, {{167851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52545, {{167852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52546, {{167853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52547, {{167854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52548, {{167855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52549, {{167856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52550, {{167857, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52551, {{167858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52552, {{167859, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52553, {{167860, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52554, {{167861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52555, {{167862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52556, {{167863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52557, {{167864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52558, {{167865, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52559, {{167866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52561, {{167867, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52562, {{167868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52563, {{167869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52564, {{167870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52565, {{167871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52566, {{167872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52567, {{167873, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52568, {{167874, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52569, {{167875, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52570, {{167878, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52571, {{167879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52572, {{167880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52573, {{167881, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52574, {{167883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52575, {{167884, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52576, {{167885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52577, {{167886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52578, {{167887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52579, {{167888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52580, {{167889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52581, {{167890, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52582, {{167891, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52583, {{167892, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52584, {{167893, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52585, {{167894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52586, {{167895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52587, {{167896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52588, {{167897, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52589, {{167898, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52590, {{167899, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52591, {{167902, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52592, {{167904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52593, {{167905, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52594, {{167906, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52595, {{167907, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{52596, {{167908, 1, 0, 0},{167908, 1, 1, 16}, {0, 0, 0, 0}}},{52596, {{167908, 1, 0, 0},{167908, 1, 1, 16}, {0, 0, 0, 0}}},{52597, {{167908, 1, 0, 0},{167908, 1, 1, 17}, {0, 0, 0, 0}}},{52597, {{167908, 1, 0, 0},{167908, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{52598, {{167909, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52599, {{167910, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52600, {{167911, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52601, {{167912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52602, {{167913, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52603, {{167914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52604, {{167915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52605, {{167916, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52606, {{167917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52607, {{167918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52608, {{167919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52609, {{167921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52610, {{167922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52611, {{167923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52612, {{167924, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52613, {{167925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52614, {{167926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52615, {{167928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52616, {{167929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52617, {{167931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52619, {{167933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52620, {{167935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52621, {{167936, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52622, {{167937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52623, {{167938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52624, {{167939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52625, {{167940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52626, {{167941, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52627, {{167942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52628, {{167943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52629, {{167944, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52630, {{167945, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52631, {{167946, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52632, {{167947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52633, {{167948, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52634, {{167949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52635, {{167950, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52636, {{167951, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52637, {{167952, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52638, {{167953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52639, {{167956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52640, {{167957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52641, {{167959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52642, {{167960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52643, {{167961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52644, {{167962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52645, {{167963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52646, {{167964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52647, {{167965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52648, {{167966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52649, {{167967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52650, {{167968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52651, {{167969, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52652, {{167971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52653, {{167972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52654, {{167973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52655, {{167974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52656, {{167975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52657, {{167976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52658, {{167977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52659, {{167978, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52660, {{167979, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52661, {{167980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52662, {{167981, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52663, {{167982, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52664, {{167983, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52665, {{167984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{52666, {{167985, 1, 0, 0},{167985, 1, 1, 17}, {0, 0, 0, 0}}},{52666, {{167985, 1, 0, 0},{167985, 1, 1, 17}, {0, 0, 0, 0}}},{58985, {{167985, 1, 0, 0},{167985, 1, 1, 16}, {0, 0, 0, 0}}},{58985, {{167985, 1, 0, 0},{167985, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{52667, {{167986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52668, {{167988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52669, {{167989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52670, {{167990, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52671, {{167991, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52672, {{167992, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52673, {{167993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52674, {{167994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52675, {{167995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52676, {{167996, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52677, {{167997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52678, {{167998, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52679, {{167999, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52680, {{168000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52681, {{168002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52682, {{168003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52683, {{168004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52684, {{168005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52685, {{168006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52686, {{168007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52687, {{168008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52688, {{168009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52689, {{168010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52690, {{168011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52691, {{168012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52692, {{168013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52693, {{168014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52694, {{168015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52695, {{168016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52696, {{168018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52697, {{168019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52698, {{168020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52699, {{168021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52700, {{168022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52701, {{168023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52702, {{168024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52703, {{168027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52704, {{168028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52705, {{168029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52706, {{168030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52707, {{168031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52708, {{168032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52709, {{168033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52710, {{168034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52711, {{168035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52712, {{168036, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52713, {{168039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52714, {{168040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52715, {{168041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52716, {{168042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52717, {{168043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52718, {{168044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52719, {{168047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52720, {{168048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52721, {{168050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52722, {{168051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52723, {{168052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52724, {{168053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52725, {{168055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52726, {{168057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52727, {{168058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52728, {{168059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52729, {{168060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52730, {{168066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52731, {{168067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52732, {{168068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52733, {{168069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52734, {{168070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52735, {{168071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52736, {{168072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52737, {{168073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52738, {{168074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52739, {{168077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52740, {{168078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52741, {{168082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52742, {{168083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52744, {{168084, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52745, {{168085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52746, {{168086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52747, {{168087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52748, {{168088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52749, {{168089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52750, {{168094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52751, {{168097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52752, {{168098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52753, {{168099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52754, {{168100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52755, {{168101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52756, {{168102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52757, {{168103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52758, {{168104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52759, {{168105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52760, {{168106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52762, {{168107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52761, {{168108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52764, {{168115, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52765, {{168117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52766, {{168120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52767, {{168121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52768, {{168122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52769, {{168123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52770, {{168124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52771, {{168125, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52772, {{168126, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52773, {{168127, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52774, {{168128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52775, {{168129, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52776, {{168130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52777, {{168131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52778, {{168139, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52779, {{168140, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52780, {{168144, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52781, {{168145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52782, {{168146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52783, {{168147, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52784, {{168148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52785, {{168149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52786, {{168150, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52787, {{168151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52788, {{168152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52789, {{168153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52790, {{168154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52791, {{168155, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52792, {{168156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52793, {{168157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52794, {{168158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52795, {{168159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52796, {{168160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52797, {{168161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52799, {{168172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52800, {{168175, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52801, {{168180, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52802, {{168181, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52803, {{168182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52804, {{168183, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52805, {{168184, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52806, {{168185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52807, {{168186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52808, {{168187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52809, {{168188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52810, {{168191, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52811, {{168192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52812, {{168193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52813, {{168196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52814, {{168198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52815, {{168201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52816, {{168202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52817, {{168213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52818, {{168214, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52819, {{168215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52820, {{168216, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52821, {{168217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52822, {{168218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52823, {{168219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52824, {{168220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52825, {{168221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52826, {{168222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52827, {{168223, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52828, {{168224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52829, {{168225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52830, {{168226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52831, {{168227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52832, {{168228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52833, {{168229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52834, {{168230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52835, {{168231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52836, {{168234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52837, {{168235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52838, {{168237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{52839, {{168238, 1, 0, 0},{168238, 1, 1, 16}, {0, 0, 0, 0}}},{52839, {{168238, 1, 0, 0},{168238, 1, 1, 16}, {0, 0, 0, 0}}},{58872, {{168238, 1, 0, 0},{168238, 1, 1, 17}, {0, 0, 0, 0}}},{58872, {{168238, 1, 0, 0},{168238, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52840, {{168247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52841, {{168248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52842, {{168250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52843, {{168251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52844, {{168252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52845, {{168253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52846, {{168254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52847, {{168255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52848, {{168256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52849, {{168257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52850, {{168258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52851, {{168259, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52852, {{168260, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52853, {{168261, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52854, {{168262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52855, {{168263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52856, {{168264, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52857, {{168265, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52858, {{168268, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52859, {{168270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52860, {{168271, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52861, {{168272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52862, {{168274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52863, {{168275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52864, {{168278, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52865, {{168285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{52866, {{168286, 1, 0, 0},{168286, 1, 1, 16}, {0, 0, 0, 0}}},{52866, {{168286, 1, 0, 0},{168286, 1, 1, 16}, {0, 0, 0, 0}}},{52867, {{168286, 1, 0, 0},{168286, 1, 1, 17}, {0, 0, 0, 0}}},{52867, {{168286, 1, 0, 0},{168286, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{52868, {{168287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52869, {{168288, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52870, {{168289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52871, {{168290, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52872, {{168291, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52873, {{168292, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52874, {{168293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52875, {{168294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52876, {{168296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52877, {{168297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52878, {{168298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52879, {{168299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52880, {{168303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52881, {{168305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52882, {{168306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52883, {{168308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52884, {{168310, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52885, {{168311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58997, {{168312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52886, {{168316, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52887, {{168317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{52888, {{168318, 1, 0, 0},{168318, 1, 1, 16}, {0, 0, 0, 0}}},{52888, {{168318, 1, 0, 0},{168318, 1, 1, 16}, {0, 0, 0, 0}}},{52889, {{168318, 1, 0, 0},{168318, 1, 1, 17}, {0, 0, 0, 0}}},{52889, {{168318, 1, 0, 0},{168318, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{52890, {{168319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52891, {{168320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52892, {{168321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52893, {{168322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52894, {{168323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52895, {{168324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52896, {{168325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52897, {{168326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52898, {{168327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52899, {{168328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52900, {{168329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52901, {{168337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52902, {{168338, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52903, {{168339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52904, {{168340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52905, {{168341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52906, {{168342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52907, {{168343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52908, {{168344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52909, {{168346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58054, {{168348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52910, {{168351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52911, {{168355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52912, {{168356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52913, {{168357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52914, {{168363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52915, {{168365, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52916, {{168366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52917, {{168367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52918, {{168368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52919, {{168369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52920, {{168370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52921, {{168373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52922, {{168374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52923, {{168375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52924, {{168376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52930, {{168377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52925, {{168378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52926, {{168381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52927, {{168383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52928, {{168384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52929, {{168385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52931, {{168386, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52932, {{168388, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52933, {{168391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52934, {{168392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52935, {{168394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52936, {{168395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52937, {{168396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52938, {{168398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52939, {{168401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52940, {{168402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52941, {{168403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52942, {{168404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{52943, {{168405, 1, 0, 0},{168405, 1, 1, 16}, {0, 0, 0, 0}}},{52943, {{168405, 1, 0, 0},{168405, 1, 1, 16}, {0, 0, 0, 0}}},{52944, {{168405, 1, 0, 0},{168405, 1, 1, 17}, {0, 0, 0, 0}}},{52944, {{168405, 1, 0, 0},{168405, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{52945, {{168406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52946, {{168407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52947, {{168411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52948, {{168412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52950, {{168413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52951, {{168414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52952, {{168415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52953, {{168416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52954, {{168419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52956, {{168422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52957, {{168423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52958, {{168424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52959, {{168425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52960, {{168426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52961, {{168427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52962, {{168429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52963, {{168430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52964, {{168431, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52965, {{168432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52966, {{168433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52967, {{168436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52968, {{168437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52969, {{168438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52970, {{168440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52971, {{168442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52972, {{168443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52973, {{168445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52974, {{168447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52975, {{168448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52976, {{168449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52977, {{168451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52978, {{168452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52979, {{168453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52980, {{168454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52981, {{168455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52982, {{168456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52983, {{168457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52984, {{168458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52985, {{168459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52986, {{168460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52987, {{168461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52988, {{168462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52989, {{168463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52990, {{168464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52991, {{168465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52992, {{168468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52993, {{168470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52994, {{168471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{52995, {{168473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{52996, {{168474, 1, 0, 0},{168474, 1, 1, 16}, {0, 0, 0, 0}}},{52996, {{168474, 1, 0, 0},{168474, 1, 1, 16}, {0, 0, 0, 0}}},{52997, {{168474, 1, 0, 0},{168474, 1, 1, 17}, {0, 0, 0, 0}}},{52997, {{168474, 1, 0, 0},{168474, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{52998, {{168475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{52999, {{168476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53000, {{168477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53001, {{168478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53002, {{168479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53003, {{168481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53004, {{168482, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53005, {{168483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53006, {{168484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53007, {{168485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53008, {{168486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53009, {{168488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53010, {{168492, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53011, {{168493, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53012, {{168494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53013, {{168495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53014, {{168497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53015, {{168498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53016, {{168500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53017, {{168501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53018, {{168502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53019, {{168503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53020, {{168504, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53021, {{168505, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53022, {{168506, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53023, {{168507, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53024, {{168508, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {2, {{53025, {{168509, 1, 0, 0}, {0, 0, 0, 0}}},{53026, {{168509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53027, {{168511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53028, {{168512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53029, {{168513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53030, {{168514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53031, {{168515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53032, {{168516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53033, {{168518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53034, {{168519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53035, {{168520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53036, {{168521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53037, {{168522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53038, {{168523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53039, {{168524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53040, {{168525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53041, {{168527, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53042, {{168528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53043, {{168529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53044, {{168530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53045, {{168532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53046, {{168533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53047, {{168534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53048, {{168535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59004, {{168536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53049, {{168537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53050, {{168538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53051, {{168540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53052, {{168541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{53053, {{168542, 1, 0, 0},{168542, 1, 1, 16}, {0, 0, 0, 0}}},{53053, {{168542, 1, 0, 0},{168542, 1, 1, 16}, {0, 0, 0, 0}}},{53054, {{168542, 1, 0, 0},{168542, 1, 1, 17}, {0, 0, 0, 0}}},{53054, {{168542, 1, 0, 0},{168542, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{53055, {{168543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53056, {{168545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53057, {{168546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53058, {{168547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53059, {{168548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53060, {{168549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53061, {{168550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53062, {{168551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53063, {{168552, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53064, {{168553, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53065, {{168554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53066, {{168555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53067, {{168556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53068, {{168559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53069, {{168560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53070, {{168562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53071, {{168563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53072, {{168565, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53073, {{168566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53074, {{168568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53075, {{168569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53076, {{168570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53077, {{168571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53078, {{168572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53079, {{168573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53080, {{168574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53081, {{168575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53082, {{168576, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53083, {{168577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53084, {{168578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53085, {{168579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53086, {{168580, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53087, {{168581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53088, {{168582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53089, {{168583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53090, {{168584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53091, {{168585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53092, {{168587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53093, {{168589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53094, {{168591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53095, {{168592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53096, {{168593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53097, {{168594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{53098, {{168595, 1, 0, 0},{168595, 1, 1, 16}, {0, 0, 0, 0}}},{53098, {{168595, 1, 0, 0},{168595, 1, 1, 16}, {0, 0, 0, 0}}},{59008, {{168595, 1, 0, 0},{168595, 1, 1, 17}, {0, 0, 0, 0}}},{59008, {{168595, 1, 0, 0},{168595, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{53099, {{168596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53100, {{168598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53101, {{168599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53102, {{168601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53103, {{168602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53104, {{168603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53105, {{168604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53106, {{168605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53107, {{168606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53108, {{168607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53109, {{168608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53110, {{168609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53111, {{168610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53112, {{168611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53113, {{168612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53114, {{168613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53115, {{168614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53116, {{168615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53117, {{168616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53118, {{168617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53119, {{168618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53120, {{168619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53121, {{168620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53122, {{168621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53123, {{168622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53124, {{168623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53125, {{168624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53126, {{168625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53127, {{168626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53128, {{168627, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53129, {{168628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53130, {{168629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53131, {{168630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53132, {{168631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53133, {{168633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53134, {{168634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53135, {{168635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53136, {{168639, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53137, {{168640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53138, {{168641, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53139, {{168642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53140, {{168643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53141, {{168644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53142, {{168645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53143, {{168646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53144, {{168647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53145, {{168648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53146, {{168649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53147, {{168650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53148, {{168651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53149, {{168652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53150, {{168653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53151, {{168654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53152, {{168655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53153, {{168656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53154, {{168657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53155, {{168658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53156, {{168659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53157, {{168660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53158, {{168663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53159, {{168665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53160, {{168667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53161, {{168668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53162, {{168669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53163, {{168670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53164, {{168671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53165, {{168672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53166, {{168673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53167, {{168674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53168, {{168675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53169, {{168676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53170, {{168677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53171, {{168678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53172, {{168679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53173, {{168680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53174, {{168681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53175, {{168682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53176, {{168683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53177, {{168684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53178, {{168685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53179, {{168686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53180, {{168687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53181, {{168688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53182, {{168689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53183, {{168690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53184, {{168693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53185, {{168696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53186, {{168697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53187, {{168698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53188, {{168700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53189, {{168701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53190, {{168702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53191, {{168703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53192, {{168704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53193, {{168707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53194, {{168708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53195, {{168709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53196, {{168710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53197, {{168711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53198, {{168712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53199, {{168713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53200, {{168714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53201, {{168715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53202, {{168716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53203, {{168717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53204, {{168718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53205, {{168719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53206, {{168720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53207, {{168721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53208, {{168722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53209, {{168723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53210, {{168730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53211, {{168731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53212, {{168733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53213, {{168736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53214, {{168737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53215, {{168738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53216, {{168742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53217, {{168743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53218, {{168744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53219, {{168745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53220, {{168746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53221, {{168747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53222, {{168748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53223, {{168749, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53224, {{168750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53225, {{168751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53226, {{168752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53227, {{168753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53228, {{168754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53229, {{168755, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53230, {{168758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53231, {{168761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53232, {{168762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53233, {{168763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53234, {{168766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53235, {{168767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53236, {{168768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53237, {{168769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53238, {{168770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53239, {{168771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53240, {{168772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53241, {{168773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53242, {{168774, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53243, {{168775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53244, {{168777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53245, {{168778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53246, {{168779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53247, {{168782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53248, {{168783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53249, {{168784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53250, {{168785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53251, {{168787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53252, {{168788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53253, {{168789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{53254, {{168790, 1, 0, 0},{168790, 1, 1, 16}, {0, 0, 0, 0}}},{53254, {{168790, 1, 0, 0},{168790, 1, 1, 16}, {0, 0, 0, 0}}},{53255, {{168790, 1, 0, 0},{168790, 1, 1, 17}, {0, 0, 0, 0}}},{53255, {{168790, 1, 0, 0},{168790, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{53256, {{168791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53257, {{168792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53258, {{168793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53259, {{168794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53260, {{168795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53261, {{168796, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53262, {{168797, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53263, {{168798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53264, {{168799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53265, {{168804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53266, {{168805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53267, {{168806, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53268, {{168807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53269, {{168808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53270, {{168809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53271, {{168810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53272, {{168811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53273, {{168814, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53274, {{168817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53275, {{168818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53276, {{168819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53277, {{168820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53278, {{168821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53279, {{168823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53280, {{168824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{53281, {{168825, 1, 0, 0},{168825, 1, 1, 16}, {0, 0, 0, 0}}},{53281, {{168825, 1, 0, 0},{168825, 1, 1, 16}, {0, 0, 0, 0}}},{53282, {{168825, 1, 0, 0},{168825, 1, 1, 17}, {0, 0, 0, 0}}},{53282, {{168825, 1, 0, 0},{168825, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53283, {{168827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53284, {{168830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53285, {{168831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53286, {{168832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53287, {{168833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53288, {{168836, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53289, {{168837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53290, {{168839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53291, {{168840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53292, {{168841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53293, {{168842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53294, {{168843, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53295, {{168844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53296, {{168845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53297, {{168846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53298, {{168847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53299, {{168849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53300, {{168850, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53301, {{168851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53302, {{168852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53303, {{168853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53304, {{168855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53305, {{168856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53306, {{168857, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53307, {{168858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53308, {{168859, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53309, {{168860, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53310, {{168861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53311, {{168863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53312, {{168865, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53313, {{168868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53314, {{168869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53315, {{168870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53316, {{168871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53317, {{168872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53318, {{168873, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53319, {{168874, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53320, {{168875, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53321, {{168876, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53322, {{168877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53323, {{168878, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53324, {{168879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53325, {{168880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53326, {{168881, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53327, {{168882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53328, {{168883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53329, {{168885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53330, {{168887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53331, {{168888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53332, {{168889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53333, {{168890, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53334, {{168891, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53335, {{168892, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53336, {{168893, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53337, {{168894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53338, {{168897, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53339, {{168899, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53340, {{168901, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53341, {{168902, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53342, {{168903, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53343, {{168904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53344, {{168906, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53345, {{168907, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53346, {{168910, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53347, {{168912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53348, {{168913, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53349, {{168914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53350, {{168915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53351, {{168916, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53352, {{168917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53353, {{168918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53354, {{168919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53355, {{168920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53356, {{168921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53357, {{168922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53358, {{168926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53359, {{168927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53360, {{168928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53361, {{168929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53362, {{168930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53363, {{168931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53364, {{168932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53365, {{168933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53366, {{168934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53367, {{168935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53368, {{168936, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53369, {{168937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53370, {{168938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53371, {{168939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53372, {{168940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53373, {{168941, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53374, {{168942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53375, {{168943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53376, {{168944, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53377, {{168945, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53378, {{168946, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53379, {{168948, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53380, {{168949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53381, {{168950, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53382, {{168952, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53383, {{168953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53384, {{168954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53385, {{168955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53386, {{168961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53387, {{168963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53388, {{168964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53389, {{168965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53390, {{168967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53391, {{168970, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53392, {{168971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53393, {{168972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53394, {{168973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53395, {{168975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53396, {{168979, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53397, {{168980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53398, {{168981, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53399, {{168982, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53400, {{168984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53401, {{168985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{53402, {{168986, 1, 0, 0},{168986, 1, 1, 16}, {0, 0, 0, 0}}},{53402, {{168986, 1, 0, 0},{168986, 1, 1, 16}, {0, 0, 0, 0}}},{53403, {{168986, 1, 0, 0},{168986, 1, 1, 17}, {0, 0, 0, 0}}},{53403, {{168986, 1, 0, 0},{168986, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{53404, {{168987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53405, {{168988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53406, {{168990, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{53407, {{168991, 1, 0, 0},{168991, 1, 1, 16}, {0, 0, 0, 0}}},{53407, {{168991, 1, 0, 0},{168991, 1, 1, 16}, {0, 0, 0, 0}}},{53408, {{168991, 1, 0, 0},{168991, 1, 1, 17}, {0, 0, 0, 0}}},{53408, {{168991, 1, 0, 0},{168991, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{53409, {{168992, 1, 0, 0},{168992, 1, 1, 16}, {0, 0, 0, 0}}},{53409, {{168992, 1, 0, 0},{168992, 1, 1, 16}, {0, 0, 0, 0}}},{53410, {{168992, 1, 0, 0},{168992, 1, 1, 17}, {0, 0, 0, 0}}},{53410, {{168992, 1, 0, 0},{168992, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{53411, {{168993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{53412, {{168994, 1, 0, 0},{168994, 1, 1, 16}, {0, 0, 0, 0}}},{53412, {{168994, 1, 0, 0},{168994, 1, 1, 16}, {0, 0, 0, 0}}},{53413, {{168994, 1, 0, 0},{168994, 1, 1, 17}, {0, 0, 0, 0}}},{53413, {{168994, 1, 0, 0},{168994, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{53414, {{168995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53415, {{168996, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53416, {{168997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53417, {{168998, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{53418, {{168999, 1, 0, 0},{168999, 1, 1, 16}, {0, 0, 0, 0}}},{53418, {{168999, 1, 0, 0},{168999, 1, 1, 16}, {0, 0, 0, 0}}},{53419, {{168999, 1, 0, 0},{168999, 1, 1, 17}, {0, 0, 0, 0}}},{53419, {{168999, 1, 0, 0},{168999, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{53420, {{169000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53421, {{169003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53422, {{169005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53423, {{169006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53424, {{169008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53425, {{169009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53426, {{169010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53427, {{169012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53428, {{169013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53429, {{169014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53430, {{169015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53431, {{169016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53432, {{169017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53433, {{169018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53434, {{169019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53435, {{169020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53436, {{169022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{53437, {{169023, 1, 0, 0},{169023, 1, 1, 16}, {0, 0, 0, 0}}},{53437, {{169023, 1, 0, 0},{169023, 1, 1, 16}, {0, 0, 0, 0}}},{53438, {{169023, 1, 0, 0},{169023, 1, 1, 17}, {0, 0, 0, 0}}},{53438, {{169023, 1, 0, 0},{169023, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{53439, {{169024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53440, {{169025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53441, {{169026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53442, {{169027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53443, {{169031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53444, {{169032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53445, {{169033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53446, {{169034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53447, {{169035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53448, {{169036, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53449, {{169037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53450, {{169040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53451, {{169041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53452, {{169042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53453, {{169043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53454, {{169044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53455, {{169045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53456, {{169046, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53457, {{169048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53458, {{169049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53459, {{169050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53460, {{169051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53461, {{169053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53462, {{169054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53463, {{169055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53464, {{169057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53465, {{169058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53466, {{169059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53467, {{169060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53468, {{169061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53469, {{169062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53470, {{169063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53471, {{169064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53472, {{169065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53473, {{169066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53474, {{169068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53475, {{169069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53476, {{169070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53477, {{169071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53478, {{169072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53479, {{169073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53480, {{169074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53481, {{169075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53482, {{169076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53483, {{169077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53484, {{169078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53485, {{169079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53486, {{169080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53487, {{169081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53488, {{169082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53489, {{169083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53490, {{169084, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53491, {{169085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53492, {{169086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53493, {{169090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53494, {{169091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53495, {{169092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53496, {{169094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53497, {{169095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53498, {{169096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53499, {{169097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53500, {{169098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53501, {{169100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53502, {{169103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53503, {{169104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53504, {{169105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53505, {{169106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53506, {{169107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53507, {{169108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53508, {{169109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53509, {{169110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53510, {{169111, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53511, {{169112, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53512, {{169113, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53513, {{169114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53514, {{169115, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53515, {{169116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53516, {{169117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53517, {{169118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53518, {{169122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53519, {{169124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53520, {{169126, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53521, {{169127, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53522, {{169130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53523, {{169131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53524, {{169133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53525, {{169134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53526, {{169135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53527, {{169136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53528, {{169137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53529, {{169138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53530, {{169139, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53531, {{169140, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53532, {{169141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53533, {{169143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53534, {{169144, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53535, {{169145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53536, {{169146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53537, {{169147, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53538, {{169148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53539, {{169149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53540, {{169150, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53541, {{169151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53542, {{169152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53543, {{169153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53544, {{169154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53545, {{169155, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53546, {{169156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53547, {{169157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53548, {{169158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53549, {{169169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53550, {{169171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53551, {{169172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53552, {{169173, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53553, {{169174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53554, {{169176, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60336, {{169177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53555, {{169181, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53556, {{169182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53557, {{169183, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53558, {{169184, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53559, {{169185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53560, {{169186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53561, {{169187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53562, {{169188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53563, {{169189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53564, {{169190, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53565, {{169191, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53566, {{169192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53568, {{169194, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53569, {{169195, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53570, {{169196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53571, {{169197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53572, {{169198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53573, {{169199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53574, {{169208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53575, {{169210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53576, {{169211, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53577, {{169212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53578, {{169213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53579, {{169214, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53580, {{169216, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53581, {{169217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53582, {{169218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53583, {{169219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53584, {{169220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53585, {{169221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53586, {{169222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53587, {{169223, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53588, {{169224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53589, {{169225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53590, {{169226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53591, {{169227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53592, {{169228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53593, {{169234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53594, {{169236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53595, {{169237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53596, {{169240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53597, {{169241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53598, {{169242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53599, {{169243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53600, {{169244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53601, {{169245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53602, {{169246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53603, {{169247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53604, {{169248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53605, {{169249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53606, {{169250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53607, {{169251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{53608, {{169252, 1, 0, 0},{169252, 1, 1, 16}, {0, 0, 0, 0}}},{53608, {{169252, 1, 0, 0},{169252, 1, 1, 16}, {0, 0, 0, 0}}},{59031, {{169252, 1, 0, 0},{169252, 1, 1, 17}, {0, 0, 0, 0}}},{59031, {{169252, 1, 0, 0},{169252, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{53609, {{169253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53610, {{169254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53611, {{169262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53612, {{169263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53613, {{169264, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53614, {{169267, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53615, {{169268, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53616, {{169269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53617, {{169270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53618, {{169271, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53619, {{169272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53620, {{169273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{53621, {{169274, 1, 0, 0},{169274, 1, 1, 16}, {0, 0, 0, 0}}},{53621, {{169274, 1, 0, 0},{169274, 1, 1, 16}, {0, 0, 0, 0}}},{59032, {{169274, 1, 0, 0},{169274, 1, 1, 17}, {0, 0, 0, 0}}},{59032, {{169274, 1, 0, 0},{169274, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{53622, {{169275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53623, {{169276, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53624, {{169277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53625, {{169278, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53626, {{169280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53627, {{169282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53628, {{169285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53629, {{169286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{53630, {{169288, 1, 0, 0},{169288, 1, 1, 16}, {0, 0, 0, 0}}},{53630, {{169288, 1, 0, 0},{169288, 1, 1, 16}, {0, 0, 0, 0}}},{68091, {{169288, 1, 0, 0},{169288, 1, 1, 17}, {0, 0, 0, 0}}},{68091, {{169288, 1, 0, 0},{169288, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{53631, {{169289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53632, {{169290, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53633, {{169291, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53634, {{169292, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53635, {{169293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53636, {{169294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53637, {{169295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53638, {{169296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53639, {{169298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53640, {{169299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53641, {{169300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53642, {{169301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53643, {{169302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53644, {{169304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53645, {{169308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53646, {{169309, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53647, {{169310, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53648, {{169311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53649, {{169312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53650, {{169313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53651, {{169314, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53652, {{169315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53653, {{169316, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53654, {{169318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53655, {{169321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53656, {{169322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53657, {{169323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53658, {{169325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53659, {{169326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53660, {{169327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53661, {{169328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53662, {{169329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53663, {{169330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53664, {{169331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53665, {{169332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53666, {{169333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53667, {{169334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53668, {{169336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53669, {{169337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53670, {{169338, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53671, {{169340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53672, {{169342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53673, {{169344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53674, {{169345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53675, {{169350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53676, {{169351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53677, {{169354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53678, {{169356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53679, {{169357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53680, {{169358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53681, {{169359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53682, {{169360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53683, {{169362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53684, {{169363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53685, {{169364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53686, {{169369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53687, {{169370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53688, {{169371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53689, {{169372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53690, {{169373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53691, {{169376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53692, {{169377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53693, {{169378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53694, {{169379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53695, {{169380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53696, {{169381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53697, {{169382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53698, {{169384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53699, {{169385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53700, {{169388, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53701, {{169389, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53702, {{169390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53703, {{169391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53704, {{169392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53705, {{169394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53706, {{169395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53707, {{169396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53708, {{169397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{53709, {{169398, 1, 0, 0},{169398, 1, 1, 16}, {0, 0, 0, 0}}},{53709, {{169398, 1, 0, 0},{169398, 1, 1, 16}, {0, 0, 0, 0}}},{53710, {{169398, 1, 0, 0},{169398, 1, 1, 17}, {0, 0, 0, 0}}},{53710, {{169398, 1, 0, 0},{169398, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{53711, {{169399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53712, {{169400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53713, {{169401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53714, {{169402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53715, {{169404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53716, {{169405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53717, {{169406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53718, {{169407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53719, {{169409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53720, {{169412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53721, {{169413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53722, {{169414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53723, {{169415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53724, {{169416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53725, {{169417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53726, {{169418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53727, {{169419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53728, {{169422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53729, {{169423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53730, {{169424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53731, {{169427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53732, {{169428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53733, {{169429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53734, {{169430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53735, {{169432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53736, {{169433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53737, {{169434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53738, {{169435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53739, {{169437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53740, {{169438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53741, {{169440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53742, {{169442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53743, {{169443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53744, {{169444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53745, {{169445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53746, {{169447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53747, {{169448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53748, {{169449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53749, {{169450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53750, {{169451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53751, {{169452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53752, {{169453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53753, {{169454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53754, {{169455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53755, {{169456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53756, {{169457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53757, {{169458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53758, {{169459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53759, {{169460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53760, {{169461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53761, {{169462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53762, {{169463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53763, {{169464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53764, {{169468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53765, {{169469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53766, {{169470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53767, {{169471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53768, {{169473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53769, {{169475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53770, {{169477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53771, {{169478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53772, {{169479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53773, {{169480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53774, {{169481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53775, {{169482, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53776, {{169483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53777, {{169484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53778, {{169485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53779, {{169486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53780, {{169487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53781, {{169488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53782, {{169489, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53783, {{169490, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53784, {{169491, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53785, {{169493, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53786, {{169494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53787, {{169495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53788, {{169496, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53789, {{169497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53790, {{169498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53791, {{169500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53792, {{169501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53793, {{169502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53794, {{169503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53795, {{169504, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53796, {{169505, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53797, {{169509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53798, {{169510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53799, {{169511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53800, {{169517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53801, {{169518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53802, {{169519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53803, {{169520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53804, {{169521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53805, {{169522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53806, {{169523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53807, {{169525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53808, {{169527, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{53809, {{169528, 1, 0, 0},{169528, 1, 1, 17}, {0, 0, 0, 0}}},{53809, {{169528, 1, 0, 0},{169528, 1, 1, 17}, {0, 0, 0, 0}}},{59038, {{169528, 1, 0, 0},{169528, 1, 1, 16}, {0, 0, 0, 0}}},{59038, {{169528, 1, 0, 0},{169528, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{53810, {{169529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53811, {{169530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53812, {{169531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53813, {{169532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53814, {{169533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53815, {{169534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53816, {{169535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53817, {{169537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53818, {{169538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53819, {{169539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53820, {{169540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53821, {{169541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53822, {{169542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53823, {{169543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53824, {{169544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53825, {{169546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53826, {{169550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53827, {{169551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53828, {{169552, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53829, {{169553, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53830, {{169554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53831, {{169555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53832, {{169556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53833, {{169558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53834, {{169560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53835, {{169561, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53836, {{169562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53837, {{169563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53838, {{169564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53839, {{169565, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{53840, {{169566, 1, 0, 0},{169566, 1, 1, 16}, {0, 0, 0, 0}}},{53840, {{169566, 1, 0, 0},{169566, 1, 1, 16}, {0, 0, 0, 0}}},{53841, {{169566, 1, 0, 0},{169566, 1, 1, 17}, {0, 0, 0, 0}}},{53841, {{169566, 1, 0, 0},{169566, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53842, {{169568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53843, {{169569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53844, {{169570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53845, {{169571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53846, {{169572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53847, {{169585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53848, {{169586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53849, {{169588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53850, {{169590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53851, {{169591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53852, {{169592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{53853, {{169594, 1, 0, 0},{169594, 1, 1, 16}, {0, 0, 0, 0}}},{53853, {{169594, 1, 0, 0},{169594, 1, 1, 16}, {0, 0, 0, 0}}},{53854, {{169594, 1, 0, 0},{169594, 1, 1, 17}, {0, 0, 0, 0}}},{53854, {{169594, 1, 0, 0},{169594, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53855, {{169596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53856, {{169597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53857, {{169598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53858, {{169599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53859, {{169600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53860, {{169601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{53861, {{169602, 1, 0, 0},{169602, 1, 1, 17}, {0, 0, 0, 0}}},{53861, {{169602, 1, 0, 0},{169602, 1, 1, 17}, {0, 0, 0, 0}}},{53862, {{169602, 1, 0, 0},{169602, 1, 1, 16}, {0, 0, 0, 0}}},{53862, {{169602, 1, 0, 0},{169602, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{53863, {{169603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53864, {{169604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{53865, {{169605, 1, 0, 0},{169605, 1, 1, 17}, {0, 0, 0, 0}}},{53865, {{169605, 1, 0, 0},{169605, 1, 1, 17}, {0, 0, 0, 0}}},{53866, {{169605, 1, 0, 0},{169605, 1, 1, 16}, {0, 0, 0, 0}}},{53866, {{169605, 1, 0, 0},{169605, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53867, {{169607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{53868, {{169608, 1, 0, 0},{169608, 1, 1, 17}, {0, 0, 0, 0}}},{53868, {{169608, 1, 0, 0},{169608, 1, 1, 17}, {0, 0, 0, 0}}},{53869, {{169608, 1, 0, 0},{169608, 1, 1, 16}, {0, 0, 0, 0}}},{53869, {{169608, 1, 0, 0},{169608, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{53870, {{169609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53871, {{169610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53872, {{169611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53873, {{169612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53874, {{169613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53875, {{169614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53876, {{169615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53877, {{169618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53878, {{169619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{53879, {{169621, 1, 0, 0},{169621, 1, 1, 17}, {0, 0, 0, 0}}},{53879, {{169621, 1, 0, 0},{169621, 1, 1, 17}, {0, 0, 0, 0}}},{53880, {{169621, 1, 0, 0},{169621, 1, 1, 16}, {0, 0, 0, 0}}},{53880, {{169621, 1, 0, 0},{169621, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {4, {{53881, {{169622, 1, 0, 0},{169622, 1, 1, 17}, {0, 0, 0, 0}}},{53881, {{169622, 1, 0, 0},{169622, 1, 1, 17}, {0, 0, 0, 0}}},{53882, {{169622, 1, 0, 0},{169622, 1, 1, 16}, {0, 0, 0, 0}}},{53882, {{169622, 1, 0, 0},{169622, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{53883, {{169623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53884, {{169624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53885, {{169625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53886, {{169626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53887, {{169627, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53888, {{169628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53889, {{169629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53890, {{169630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53891, {{169635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53892, {{169636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53893, {{169640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{53894, {{169641, 1, 0, 0},{169641, 1, 1, 17}, {0, 0, 0, 0}}},{53894, {{169641, 1, 0, 0},{169641, 1, 1, 17}, {0, 0, 0, 0}}},{53895, {{169641, 1, 0, 0},{169641, 1, 1, 16}, {0, 0, 0, 0}}},{53895, {{169641, 1, 0, 0},{169641, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{53896, {{169642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53897, {{169643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53898, {{169644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53899, {{169645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53900, {{169646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53901, {{169647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53902, {{169648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53903, {{169649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53904, {{169650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53905, {{169651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53906, {{169652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53907, {{169653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53908, {{169654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53909, {{169655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53910, {{169656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{53911, {{169657, 1, 0, 0},{169657, 1, 1, 17}, {0, 0, 0, 0}}},{53911, {{169657, 1, 0, 0},{169657, 1, 1, 17}, {0, 0, 0, 0}}},{53912, {{169657, 1, 0, 0},{169657, 1, 1, 16}, {0, 0, 0, 0}}},{53912, {{169657, 1, 0, 0},{169657, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{53913, {{169658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53914, {{169659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53915, {{169662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53916, {{169663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53917, {{169664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53918, {{169665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53919, {{169669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{53920, {{169670, 1, 0, 0},{169670, 1, 1, 17}, {0, 0, 0, 0}}},{53920, {{169670, 1, 0, 0},{169670, 1, 1, 17}, {0, 0, 0, 0}}},{53921, {{169670, 1, 0, 0},{169670, 1, 1, 16}, {0, 0, 0, 0}}},{53921, {{169670, 1, 0, 0},{169670, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{53922, {{169671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53923, {{169673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53924, {{169674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53925, {{169675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53926, {{169676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53927, {{169677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53928, {{169678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53929, {{169679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53930, {{169680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53931, {{169681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53932, {{169682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53933, {{169683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53934, {{169688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53935, {{169690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53936, {{169691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53937, {{169693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{53938, {{169694, 1, 0, 0},{169694, 1, 1, 17}, {0, 0, 0, 0}}},{53938, {{169694, 1, 0, 0},{169694, 1, 1, 17}, {0, 0, 0, 0}}},{53939, {{169694, 1, 0, 0},{169694, 1, 1, 16}, {0, 0, 0, 0}}},{53939, {{169694, 1, 0, 0},{169694, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{53940, {{169695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53941, {{169696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53942, {{169697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53943, {{169698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53944, {{169699, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53945, {{169700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{53946, {{169701, 1, 0, 0},{169701, 1, 1, 17}, {0, 0, 0, 0}}},{53946, {{169701, 1, 0, 0},{169701, 1, 1, 17}, {0, 0, 0, 0}}},{53947, {{169701, 1, 0, 0},{169701, 1, 1, 16}, {0, 0, 0, 0}}},{53947, {{169701, 1, 0, 0},{169701, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{53949, {{169702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53950, {{169703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53951, {{169704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53952, {{169705, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53953, {{169707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53954, {{169708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53955, {{169709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53956, {{169710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53957, {{169711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53958, {{169712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53959, {{169715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53960, {{169717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53961, {{169718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{53962, {{169721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{53963, {{169722, 1, 0, 0},{169722, 1, 1, 17}, {0, 0, 0, 0}}},{53963, {{169722, 1, 0, 0},{169722, 1, 1, 17}, {0, 0, 0, 0}}},{53964, {{169722, 1, 0, 0},{169722, 1, 1, 16}, {0, 0, 0, 0}}},{53964, {{169722, 1, 0, 0},{169722, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{53965, {{169723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53966, {{169724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53967, {{169725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53968, {{169726, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53969, {{169727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53970, {{169728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53971, {{169729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53973, {{169730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53974, {{169731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53975, {{169732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53976, {{169733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{53977, {{169734, 1, 0, 0},{169734, 1, 1, 17}, {0, 0, 0, 0}}},{53977, {{169734, 1, 0, 0},{169734, 1, 1, 17}, {0, 0, 0, 0}}},{53978, {{169734, 1, 0, 0},{169734, 1, 1, 16}, {0, 0, 0, 0}}},{53978, {{169734, 1, 0, 0},{169734, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{53979, {{169735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{53980, {{169736, 1, 0, 0},{169736, 1, 1, 17}, {0, 0, 0, 0}}},{53980, {{169736, 1, 0, 0},{169736, 1, 1, 17}, {0, 0, 0, 0}}},{53981, {{169736, 1, 0, 0},{169736, 1, 1, 16}, {0, 0, 0, 0}}},{53981, {{169736, 1, 0, 0},{169736, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{53982, {{169737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53983, {{169738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{53984, {{169739, 1, 0, 0},{169739, 1, 1, 17}, {0, 0, 0, 0}}},{53984, {{169739, 1, 0, 0},{169739, 1, 1, 17}, {0, 0, 0, 0}}},{53985, {{169739, 1, 0, 0},{169739, 1, 1, 16}, {0, 0, 0, 0}}},{53985, {{169739, 1, 0, 0},{169739, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{53986, {{169740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53987, {{169741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53988, {{169742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{53989, {{169743, 1, 0, 0},{169743, 1, 1, 17}, {0, 0, 0, 0}}},{53989, {{169743, 1, 0, 0},{169743, 1, 1, 17}, {0, 0, 0, 0}}},{53990, {{169743, 1, 0, 0},{169743, 1, 1, 18}, {0, 0, 0, 0}}},{53990, {{169743, 1, 0, 0},{169743, 1, 1, 18}, {0, 0, 0, 0}}},{53991, {{169743, 1, 0, 0},{169743, 1, 1, 16}, {0, 0, 0, 0}}},{53991, {{169743, 1, 0, 0},{169743, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{53992, {{169744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53993, {{169745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53994, {{169746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{53995, {{169747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{53996, {{169749, 1, 0, 0},{169749, 1, 1, 16}, {0, 0, 0, 0}}},{53996, {{169749, 1, 0, 0},{169749, 1, 1, 16}, {0, 0, 0, 0}}},{53997, {{169749, 1, 0, 0},{169749, 1, 1, 17}, {0, 0, 0, 0}}},{53997, {{169749, 1, 0, 0},{169749, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{53998, {{169750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{53999, {{169751, 1, 0, 0},{169751, 1, 1, 17}, {0, 0, 0, 0}}},{53999, {{169751, 1, 0, 0},{169751, 1, 1, 17}, {0, 0, 0, 0}}},{54000, {{169751, 1, 0, 0},{169751, 1, 1, 16}, {0, 0, 0, 0}}},{54000, {{169751, 1, 0, 0},{169751, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{54001, {{169753, 1, 0, 0},{169753, 1, 1, 17}, {0, 0, 0, 0}}},{54001, {{169753, 1, 0, 0},{169753, 1, 1, 17}, {0, 0, 0, 0}}},{54002, {{169753, 1, 0, 0},{169753, 1, 1, 18}, {0, 0, 0, 0}}},{54002, {{169753, 1, 0, 0},{169753, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{54003, {{169754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54004, {{169761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54005, {{169762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54006, {{169763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54007, {{169772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54008, {{169773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{54009, {{169775, 1, 0, 0},{169775, 1, 1, 17}, {0, 0, 0, 0}}},{54009, {{169775, 1, 0, 0},{169775, 1, 1, 17}, {0, 0, 0, 0}}},{54010, {{169775, 1, 0, 0},{169775, 1, 1, 16}, {0, 0, 0, 0}}},{54010, {{169775, 1, 0, 0},{169775, 1, 1, 16}, {0, 0, 0, 0}}},{59044, {{169775, 1, 0, 0},{169775, 1, 1, 18}, {0, 0, 0, 0}}},{59044, {{169775, 1, 0, 0},{169775, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{54011, {{169776, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54012, {{169777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54013, {{169778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54014, {{169779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{54015, {{169780, 1, 0, 0},{169780, 1, 1, 17}, {0, 0, 0, 0}}},{54015, {{169780, 1, 0, 0},{169780, 1, 1, 17}, {0, 0, 0, 0}}},{54016, {{169780, 1, 0, 0},{169780, 1, 1, 18}, {0, 0, 0, 0}}},{54016, {{169780, 1, 0, 0},{169780, 1, 1, 18}, {0, 0, 0, 0}}},{54017, {{169780, 1, 0, 0},{169780, 1, 1, 16}, {0, 0, 0, 0}}},{54017, {{169780, 1, 0, 0},{169780, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{54018, {{169781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54019, {{169782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54020, {{169783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54021, {{169784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{54022, {{169785, 1, 0, 0},{169785, 1, 1, 17}, {0, 0, 0, 0}}},{54022, {{169785, 1, 0, 0},{169785, 1, 1, 17}, {0, 0, 0, 0}}},{54023, {{169785, 1, 0, 0},{169785, 1, 1, 16}, {0, 0, 0, 0}}},{54023, {{169785, 1, 0, 0},{169785, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54024, {{169787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54025, {{169788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54026, {{169789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54027, {{169790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{54028, {{169791, 1, 0, 0},{169791, 1, 1, 16}, {0, 0, 0, 0}}},{54028, {{169791, 1, 0, 0},{169791, 1, 1, 16}, {0, 0, 0, 0}}},{54062, {{169791, 1, 0, 0},{169791, 1, 1, 17}, {0, 0, 0, 0}}},{54062, {{169791, 1, 0, 0},{169791, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{54029, {{169792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54030, {{169793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54031, {{169794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54032, {{169800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54033, {{169801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54034, {{169804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54035, {{169808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54036, {{169813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54037, {{169814, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54038, {{169815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54039, {{169816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{54040, {{169817, 1, 0, 0},{169817, 1, 1, 17}, {0, 0, 0, 0}}},{54040, {{169817, 1, 0, 0},{169817, 1, 1, 17}, {0, 0, 0, 0}}},{54041, {{169817, 1, 0, 0},{169817, 1, 1, 16}, {0, 0, 0, 0}}},{54041, {{169817, 1, 0, 0},{169817, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{54042, {{169818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54043, {{169819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54068, {{169820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{54044, {{169821, 1, 0, 0},{169821, 1, 1, 17}, {0, 0, 0, 0}}},{54044, {{169821, 1, 0, 0},{169821, 1, 1, 17}, {0, 0, 0, 0}}},{54045, {{169821, 1, 0, 0},{169821, 1, 1, 16}, {0, 0, 0, 0}}},{54045, {{169821, 1, 0, 0},{169821, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{54046, {{169822, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54047, {{169823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54048, {{169824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54049, {{169825, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54050, {{169826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54051, {{169827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54052, {{169828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54053, {{169829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54054, {{169830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54055, {{169831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54056, {{169832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54057, {{169833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54058, {{169834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54059, {{169835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54060, {{169836, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54061, {{169838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54063, {{169840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54064, {{169844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54065, {{169845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54066, {{169846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54067, {{169847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54069, {{169849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54070, {{169852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54071, {{169853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54072, {{169854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54073, {{169855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{54074, {{169856, 1, 0, 0},{169856, 1, 1, 17}, {0, 0, 0, 0}}},{54074, {{169856, 1, 0, 0},{169856, 1, 1, 17}, {0, 0, 0, 0}}},{54075, {{169856, 1, 0, 0},{169856, 1, 1, 16}, {0, 0, 0, 0}}},{54075, {{169856, 1, 0, 0},{169856, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{54076, {{169857, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54077, {{169858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{54078, {{169859, 1, 0, 0},{169859, 1, 1, 17}, {0, 0, 0, 0}}},{54078, {{169859, 1, 0, 0},{169859, 1, 1, 17}, {0, 0, 0, 0}}},{54079, {{169859, 1, 0, 0},{169859, 1, 1, 16}, {0, 0, 0, 0}}},{54079, {{169859, 1, 0, 0},{169859, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{54080, {{169860, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54081, {{169861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54082, {{169862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54083, {{169863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54084, {{169864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54085, {{169865, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54086, {{169866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54087, {{169867, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54088, {{169868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{54089, {{169869, 1, 0, 0},{169869, 1, 1, 17}, {0, 0, 0, 0}}},{54089, {{169869, 1, 0, 0},{169869, 1, 1, 17}, {0, 0, 0, 0}}},{54090, {{169869, 1, 0, 0},{169869, 1, 1, 16}, {0, 0, 0, 0}}},{54090, {{169869, 1, 0, 0},{169869, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{54091, {{169870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{54092, {{169871, 1, 0, 0},{169871, 1, 1, 17}, {0, 0, 0, 0}}},{54092, {{169871, 1, 0, 0},{169871, 1, 1, 17}, {0, 0, 0, 0}}},{54093, {{169871, 1, 0, 0},{169871, 1, 1, 16}, {0, 0, 0, 0}}},{54093, {{169871, 1, 0, 0},{169871, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{54094, {{169872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{54095, {{169873, 1, 0, 0},{169873, 1, 1, 17}, {0, 0, 0, 0}}},{54095, {{169873, 1, 0, 0},{169873, 1, 1, 17}, {0, 0, 0, 0}}},{54096, {{169873, 1, 0, 0},{169873, 1, 1, 16}, {0, 0, 0, 0}}},{54096, {{169873, 1, 0, 0},{169873, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54097, {{169882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54098, {{169886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{54099, {{169889, 1, 0, 0},{169889, 1, 1, 17}, {0, 0, 0, 0}}},{54099, {{169889, 1, 0, 0},{169889, 1, 1, 17}, {0, 0, 0, 0}}},{54100, {{169889, 1, 0, 0},{169889, 1, 1, 16}, {0, 0, 0, 0}}},{54100, {{169889, 1, 0, 0},{169889, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{54101, {{169890, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54102, {{169891, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54103, {{169892, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{54104, {{169893, 1, 0, 0},{169893, 1, 1, 17}, {0, 0, 0, 0}}},{54104, {{169893, 1, 0, 0},{169893, 1, 1, 17}, {0, 0, 0, 0}}},{54105, {{169893, 1, 0, 0},{169893, 1, 1, 16}, {0, 0, 0, 0}}},{54105, {{169893, 1, 0, 0},{169893, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{54106, {{169894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{54107, {{169895, 1, 0, 0},{169895, 1, 1, 17}, {0, 0, 0, 0}}},{54107, {{169895, 1, 0, 0},{169895, 1, 1, 17}, {0, 0, 0, 0}}},{54108, {{169895, 1, 0, 0},{169895, 1, 1, 16}, {0, 0, 0, 0}}},{54108, {{169895, 1, 0, 0},{169895, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{54109, {{169896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54110, {{169897, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54111, {{169898, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{54112, {{169899, 1, 0, 0},{169899, 1, 1, 17}, {0, 0, 0, 0}}},{54112, {{169899, 1, 0, 0},{169899, 1, 1, 17}, {0, 0, 0, 0}}},{54113, {{169899, 1, 0, 0},{169899, 1, 1, 16}, {0, 0, 0, 0}}},{54113, {{169899, 1, 0, 0},{169899, 1, 1, 16}, {0, 0, 0, 0}}},{54114, {{169899, 1, 0, 0},{169899, 1, 1, 18}, {0, 0, 0, 0}}},{54114, {{169899, 1, 0, 0},{169899, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{54115, {{169900, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{54116, {{169901, 1, 0, 0},{169901, 1, 1, 17}, {0, 0, 0, 0}}},{54116, {{169901, 1, 0, 0},{169901, 1, 1, 17}, {0, 0, 0, 0}}},{54117, {{169901, 1, 0, 0},{169901, 1, 1, 16}, {0, 0, 0, 0}}},{54117, {{169901, 1, 0, 0},{169901, 1, 1, 16}, {0, 0, 0, 0}}},{54118, {{169901, 1, 0, 0},{169901, 1, 1, 18}, {0, 0, 0, 0}}},{54118, {{169901, 1, 0, 0},{169901, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{54119, {{169902, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54120, {{169903, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54121, {{169906, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54122, {{169909, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54123, {{169910, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{54124, {{169911, 1, 0, 0},{169911, 1, 1, 17}, {0, 0, 0, 0}}},{54124, {{169911, 1, 0, 0},{169911, 1, 1, 17}, {0, 0, 0, 0}}},{54125, {{169911, 1, 0, 0},{169911, 1, 1, 16}, {0, 0, 0, 0}}},{54125, {{169911, 1, 0, 0},{169911, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54126, {{169914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54127, {{169915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54128, {{169916, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54129, {{169917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54130, {{169918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54131, {{169919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54132, {{169920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54133, {{169921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54134, {{169922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54135, {{169923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{54136, {{169924, 1, 0, 0},{169924, 1, 1, 17}, {0, 0, 0, 0}}},{54136, {{169924, 1, 0, 0},{169924, 1, 1, 17}, {0, 0, 0, 0}}},{54137, {{169924, 1, 0, 0},{169924, 1, 1, 16}, {0, 0, 0, 0}}},{54137, {{169924, 1, 0, 0},{169924, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{54138, {{169925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54139, {{169926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54140, {{169927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54141, {{169928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54142, {{169930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{54143, {{169931, 1, 0, 0},{169931, 1, 1, 17}, {0, 0, 0, 0}}},{54143, {{169931, 1, 0, 0},{169931, 1, 1, 17}, {0, 0, 0, 0}}},{54144, {{169931, 1, 0, 0},{169931, 1, 1, 16}, {0, 0, 0, 0}}},{54144, {{169931, 1, 0, 0},{169931, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{54145, {{169932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54146, {{169933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54147, {{169937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54148, {{169938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54149, {{169939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54150, {{169940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54151, {{169941, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54152, {{169942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54153, {{169943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54154, {{169944, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54155, {{169945, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54156, {{169946, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54157, {{169947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54158, {{169948, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54159, {{169949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54160, {{169950, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54161, {{169955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54162, {{169956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54163, {{169957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54164, {{169958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54165, {{169959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54166, {{169960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54167, {{169961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54168, {{169962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54169, {{169965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54170, {{169966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54171, {{169967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54172, {{169968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{54173, {{169969, 1, 0, 0},{169969, 1, 1, 17}, {0, 0, 0, 0}}},{54173, {{169969, 1, 0, 0},{169969, 1, 1, 17}, {0, 0, 0, 0}}},{54174, {{169969, 1, 0, 0},{169969, 1, 1, 16}, {0, 0, 0, 0}}},{54174, {{169969, 1, 0, 0},{169969, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54175, {{169971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54176, {{169972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54177, {{169974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54178, {{169975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54179, {{169976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54180, {{169978, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{54181, {{169981, 1, 0, 0},{169981, 1, 1, 16}, {0, 0, 0, 0}}},{54181, {{169981, 1, 0, 0},{169981, 1, 1, 16}, {0, 0, 0, 0}}},{54182, {{169981, 1, 0, 0},{169981, 1, 1, 17}, {0, 0, 0, 0}}},{54182, {{169981, 1, 0, 0},{169981, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54183, {{170000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54184, {{170001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54185, {{170005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54186, {{170006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54187, {{170007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54188, {{170009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54189, {{170010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54190, {{170011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54191, {{170012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54192, {{170015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54193, {{170017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54194, {{170018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54195, {{170019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54196, {{170020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54197, {{170021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54198, {{170022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54199, {{170026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54200, {{170027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54201, {{170028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54202, {{170031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54203, {{170034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54204, {{170038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54205, {{170039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54206, {{170041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54208, {{170043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54209, {{170044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54210, {{170045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54211, {{170046, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54212, {{170047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54213, {{170048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54214, {{170051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54215, {{170052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54216, {{170053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54217, {{170056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54218, {{170058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54219, {{170060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54220, {{170061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54221, {{170062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54222, {{170064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54223, {{170065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54224, {{170066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54225, {{170067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54226, {{170068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54228, {{170069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54229, {{170072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54231, {{170074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54232, {{170077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54234, {{170080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54235, {{170081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54236, {{170082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54237, {{170084, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54238, {{170085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54239, {{170087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54240, {{170088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54241, {{170089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54242, {{170090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54243, {{170091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54244, {{170092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54245, {{170093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54246, {{170095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54247, {{170096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54248, {{170097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54249, {{170098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54250, {{170099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54251, {{170101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54252, {{170103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54253, {{170104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54254, {{170105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54256, {{170106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54257, {{170107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54258, {{170108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54259, {{170109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54260, {{170110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54261, {{170111, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54262, {{170116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54263, {{170117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54264, {{170120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54265, {{170121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54266, {{170122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54267, {{170124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54268, {{170125, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54269, {{170126, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54270, {{170128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54271, {{170129, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54272, {{170130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54273, {{170131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{54274, {{170132, 1, 0, 0},{170132, 1, 1, 17}, {0, 0, 0, 0}}},{54274, {{170132, 1, 0, 0},{170132, 1, 1, 17}, {0, 0, 0, 0}}},{54275, {{170132, 1, 0, 0},{170132, 1, 1, 16}, {0, 0, 0, 0}}},{54275, {{170132, 1, 0, 0},{170132, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{54276, {{170133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54277, {{170134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54278, {{170135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54279, {{170136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54280, {{170137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54281, {{170138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54282, {{170139, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54283, {{170140, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{54284, {{170141, 1, 0, 0},{170141, 1, 1, 16}, {0, 0, 0, 0}}},{54284, {{170141, 1, 0, 0},{170141, 1, 1, 16}, {0, 0, 0, 0}}},{54285, {{170141, 1, 0, 0},{170141, 1, 1, 17}, {0, 0, 0, 0}}},{54285, {{170141, 1, 0, 0},{170141, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{54286, {{170142, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54287, {{170143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54288, {{170144, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54289, {{170145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54290, {{170151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54291, {{170152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54292, {{170154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54293, {{170155, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54294, {{170157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54295, {{170158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54296, {{170159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54297, {{170160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54298, {{170161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54299, {{170162, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54300, {{170163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54301, {{170164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54302, {{170165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54303, {{170166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54304, {{170167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54305, {{170168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54306, {{170169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54307, {{170170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54308, {{170171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54309, {{170172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54310, {{170173, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54311, {{170174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54312, {{170175, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54313, {{170176, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54314, {{170177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54315, {{170178, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54316, {{170179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54317, {{170180, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54318, {{170181, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54319, {{170182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54320, {{170186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54321, {{170188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54322, {{170189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54323, {{170192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54324, {{170194, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54325, {{170197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54326, {{170198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54327, {{170199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54328, {{170200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54329, {{170201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54330, {{170202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54331, {{170203, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54332, {{170205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54333, {{170206, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54334, {{170207, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54335, {{170208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54336, {{170209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54337, {{170210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54338, {{170211, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54339, {{170212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54340, {{170217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54341, {{170219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54342, {{170221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54343, {{170222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54344, {{170223, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54345, {{170224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54346, {{170225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54347, {{170226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54348, {{170227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54349, {{170228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54350, {{170229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54351, {{170230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54352, {{170231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54353, {{170232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54354, {{170233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54355, {{170234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54356, {{170235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54357, {{170236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54358, {{170237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54359, {{170238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54360, {{170239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{54361, {{170240, 1, 0, 0},{170240, 1, 1, 16}, {0, 0, 0, 0}}},{54361, {{170240, 1, 0, 0},{170240, 1, 1, 16}, {0, 0, 0, 0}}},{59053, {{170240, 1, 0, 0},{170240, 1, 1, 17}, {0, 0, 0, 0}}},{59053, {{170240, 1, 0, 0},{170240, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{54362, {{170241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54363, {{170248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54364, {{170249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54365, {{170251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54366, {{170256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54367, {{170259, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54368, {{170262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54369, {{170263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{54370, {{170264, 1, 0, 0},{170264, 1, 1, 16}, {0, 0, 0, 0}}},{54370, {{170264, 1, 0, 0},{170264, 1, 1, 16}, {0, 0, 0, 0}}},{68092, {{170264, 1, 0, 0},{170264, 1, 1, 17}, {0, 0, 0, 0}}},{68092, {{170264, 1, 0, 0},{170264, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{54371, {{170265, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54372, {{170266, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54373, {{170267, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54374, {{170268, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54375, {{170269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54376, {{170270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54377, {{170271, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54378, {{170272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54379, {{170273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54380, {{170274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54381, {{170275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54382, {{170276, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54383, {{170277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54384, {{170278, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54385, {{170280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54386, {{170281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54387, {{170282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54388, {{170283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54389, {{170284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54390, {{170285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54391, {{170291, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54392, {{170297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54393, {{170298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54394, {{170299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54395, {{170301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54396, {{170302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54397, {{170303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54398, {{170304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54399, {{170305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54400, {{170306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54401, {{170307, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54402, {{170308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54403, {{170309, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54404, {{170311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54405, {{170312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54406, {{170313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54407, {{170314, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54409, {{170315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54410, {{170316, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54411, {{170317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54412, {{170318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54413, {{170319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54414, {{170320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54415, {{170321, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54416, {{170322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54417, {{170323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54418, {{170324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54419, {{170326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54420, {{170327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54421, {{170328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54422, {{170332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54423, {{170334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54424, {{170338, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54425, {{170341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54426, {{170347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54427, {{170348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54428, {{170349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54429, {{170350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54430, {{170351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54431, {{170352, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54432, {{170353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54433, {{170354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54434, {{170355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54435, {{170356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54436, {{170365, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54437, {{170368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54438, {{170369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54439, {{170370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54440, {{170371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54441, {{170372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54442, {{170373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54443, {{170374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54444, {{170375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54446, {{170376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54447, {{170377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54448, {{170378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54449, {{170379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54451, {{170380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54452, {{170381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54453, {{170382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54454, {{170383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54455, {{170385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54456, {{170388, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54458, {{170395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54459, {{170396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54460, {{170398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54461, {{170399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54462, {{170400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54463, {{170401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54464, {{170402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54465, {{170403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54466, {{170404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54467, {{170405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54468, {{170406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54469, {{170407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54470, {{170408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54471, {{170410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54472, {{170419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54473, {{170421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54474, {{170422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54475, {{170423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54476, {{170424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54477, {{170425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54478, {{170426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54479, {{170427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54480, {{170428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54481, {{170429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54482, {{170430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54483, {{170432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54484, {{170434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54485, {{170435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54486, {{170436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54487, {{170437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54488, {{170438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54489, {{170439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54490, {{170440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54491, {{170441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54492, {{170442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54493, {{170443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54494, {{170444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54495, {{170446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54496, {{170447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54497, {{170448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54498, {{170450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54499, {{170452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54500, {{170453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54501, {{170454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54502, {{170456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54503, {{170457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54504, {{170458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54505, {{170459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54506, {{170460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54507, {{170464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54508, {{170465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54509, {{170466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54510, {{170467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54511, {{170468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54512, {{170469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54513, {{170513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54514, {{170514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54515, {{170517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54516, {{170519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54517, {{170520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54518, {{170521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54519, {{170522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54520, {{170524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54521, {{170525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54522, {{170526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54523, {{170527, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54524, {{170528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54525, {{170529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54526, {{170531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54527, {{170534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54528, {{170535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54529, {{170537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54530, {{170539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54531, {{170540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54532, {{170541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54533, {{170542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54534, {{170543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54535, {{170544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54536, {{170545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54537, {{170546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54538, {{170547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54539, {{170548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54540, {{170549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54541, {{170550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54542, {{170551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54543, {{170552, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54544, {{170553, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54545, {{170554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54546, {{170555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54547, {{170557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54548, {{170558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54549, {{170559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54550, {{170560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54551, {{170562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54552, {{170564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54553, {{170566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54554, {{170569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54555, {{170570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54556, {{170571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54557, {{170572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54558, {{170573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54559, {{170574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59061, {{170576, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54560, {{170578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54561, {{170580, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54562, {{170583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54563, {{170584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{54564, {{170585, 1, 0, 0},{170585, 1, 1, 17}, {0, 0, 0, 0}}},{54564, {{170585, 1, 0, 0},{170585, 1, 1, 17}, {0, 0, 0, 0}}},{59062, {{170585, 1, 0, 0},{170585, 1, 1, 16}, {0, 0, 0, 0}}},{59062, {{170585, 1, 0, 0},{170585, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{54565, {{170586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54566, {{170587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54567, {{170588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54568, {{170589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54569, {{170590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54570, {{170591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54571, {{170592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54572, {{170593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54573, {{170594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54574, {{170595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54575, {{170596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54576, {{170598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54577, {{170599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54578, {{170605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54579, {{170606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54580, {{170607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54581, {{170608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54582, {{170609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54583, {{170610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54584, {{170611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54585, {{170612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54586, {{170613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54587, {{170614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54588, {{170616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54589, {{170617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54590, {{170620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54591, {{170621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54592, {{170622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54593, {{170624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54594, {{170625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54595, {{170626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54596, {{170627, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54597, {{170628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54598, {{170629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54599, {{170630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54600, {{170632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54601, {{170633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54602, {{170634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54603, {{170635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54604, {{170636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54605, {{170637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54606, {{170638, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54607, {{170639, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54608, {{170640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54609, {{170641, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54610, {{170642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54611, {{170644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54612, {{170645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54613, {{170647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54614, {{170648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54615, {{170649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54616, {{170650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54617, {{170651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54618, {{170652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54619, {{170656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54620, {{170657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54621, {{170658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54622, {{170659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54623, {{170660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54624, {{170661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54625, {{170662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54626, {{170663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54627, {{170664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54628, {{170665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54629, {{170666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54630, {{170669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54631, {{170670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59064, {{170671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54632, {{170673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54633, {{170674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54634, {{170675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54635, {{170676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{54636, {{170679, 1, 0, 0},{170679, 1, 1, 16}, {0, 0, 0, 0}}},{54636, {{170679, 1, 0, 0},{170679, 1, 1, 16}, {0, 0, 0, 0}}},{54637, {{170679, 1, 0, 0},{170679, 1, 1, 17}, {0, 0, 0, 0}}},{54637, {{170679, 1, 0, 0},{170679, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{54638, {{170680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54639, {{170681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54640, {{170682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54641, {{170684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54642, {{170686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54643, {{170687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54644, {{170688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54645, {{170689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54646, {{170690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54647, {{170691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54648, {{170694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54649, {{170695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54650, {{170698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54651, {{170701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54652, {{170702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54653, {{170703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54654, {{170708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54655, {{170709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54656, {{170710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54657, {{170714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54658, {{170718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54659, {{170720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54660, {{170725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54661, {{170726, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54662, {{170727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54663, {{170728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54664, {{170729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54665, {{170730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54666, {{170731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54667, {{170733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54668, {{170734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54669, {{170735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54670, {{170736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54671, {{170737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54672, {{170738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54673, {{170739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54674, {{170740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54675, {{170743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54676, {{170746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54677, {{170751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54678, {{170752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54679, {{170754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54680, {{170755, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54681, {{170756, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54683, {{170761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54684, {{170763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54685, {{170765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54686, {{170767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54687, {{170768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54688, {{170769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54689, {{170772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54690, {{170774, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54691, {{170775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54692, {{170776, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54693, {{170777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54694, {{170778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54695, {{170779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54696, {{170780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54697, {{170781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54698, {{170782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54699, {{170783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54700, {{170784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54701, {{170795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54702, {{170797, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54703, {{170798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54704, {{170800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54705, {{170801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54706, {{170802, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54707, {{170803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54708, {{170804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54709, {{170805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54710, {{170806, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54711, {{170807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54712, {{170808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54713, {{170809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54714, {{170810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54716, {{170812, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54717, {{170813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54718, {{170819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54719, {{170823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54720, {{170824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54721, {{170826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54722, {{170827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54723, {{170828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54724, {{170829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54725, {{170830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54726, {{170831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54727, {{170832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54728, {{170833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54729, {{170834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54731, {{170840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54732, {{170842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54733, {{170845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54734, {{170846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54735, {{170847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54736, {{170848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54737, {{170849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54738, {{170850, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54739, {{170851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54740, {{170852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54741, {{170853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54742, {{170854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54743, {{170855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54744, {{170856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54745, {{170857, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54746, {{170858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54747, {{170859, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54748, {{170871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54749, {{170872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54750, {{170874, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54751, {{170875, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54752, {{170876, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54753, {{170877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54754, {{170878, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54755, {{170879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54756, {{170880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54757, {{170881, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54758, {{170882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54759, {{170883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54760, {{170884, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54761, {{170885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54762, {{170886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54763, {{170888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54764, {{170895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54765, {{170896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54766, {{170902, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54767, {{170903, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54768, {{170904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54769, {{170905, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54770, {{170906, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54771, {{170907, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54772, {{170908, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54773, {{170909, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54774, {{170910, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54775, {{170911, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54776, {{170912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54777, {{170913, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54778, {{170914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54779, {{170915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54780, {{170923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54781, {{170926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54782, {{170927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54783, {{170928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54784, {{170929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54785, {{170930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54786, {{170931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54787, {{170932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54788, {{170933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54789, {{170934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54790, {{170935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54791, {{170937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{54792, {{170938, 1, 0, 0},{170938, 1, 1, 16}, {0, 0, 0, 0}}},{54792, {{170938, 1, 0, 0},{170938, 1, 1, 16}, {0, 0, 0, 0}}},{54793, {{170938, 1, 0, 0},{170938, 1, 1, 17}, {0, 0, 0, 0}}},{54793, {{170938, 1, 0, 0},{170938, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54794, {{170941, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54795, {{170947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54796, {{170948, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54797, {{170949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54798, {{170950, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54799, {{170951, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54800, {{170952, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54801, {{170953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54802, {{170954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54803, {{170955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54804, {{170956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54805, {{170957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54806, {{170958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54807, {{170959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54808, {{170960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54809, {{170961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54810, {{170971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54811, {{170972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54812, {{170973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54813, {{170974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54814, {{170975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54815, {{170976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54816, {{170980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54817, {{170982, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54818, {{170983, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54819, {{170984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54820, {{170985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54821, {{170989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54822, {{170992, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54823, {{170993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54824, {{170994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54825, {{170995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54826, {{171000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54827, {{171002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54828, {{171003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54829, {{171006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54830, {{171007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54831, {{171008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54832, {{171009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54833, {{171010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54834, {{171011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54835, {{171013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54836, {{171015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54837, {{171018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54838, {{171019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54839, {{171021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54840, {{171022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54841, {{171023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54842, {{171024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54843, {{171025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54844, {{171026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{54845, {{171027, 1, 0, 0},{171027, 1, 1, 16}, {0, 0, 0, 0}}},{54845, {{171027, 1, 0, 0},{171027, 1, 1, 16}, {0, 0, 0, 0}}},{54846, {{171027, 1, 0, 0},{171027, 1, 1, 17}, {0, 0, 0, 0}}},{54846, {{171027, 1, 0, 0},{171027, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{54847, {{171028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54848, {{171029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54849, {{171030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54850, {{171031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54851, {{171033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54852, {{171037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54853, {{171038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54854, {{171039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54855, {{171040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54856, {{171042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54857, {{171043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54858, {{171044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54859, {{171045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54861, {{171046, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54862, {{171047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54863, {{171048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54864, {{171050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54865, {{171051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54866, {{171052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54867, {{171053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54868, {{171055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54869, {{171057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54870, {{171058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54871, {{171059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54872, {{171060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54873, {{171062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54874, {{171065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54875, {{171067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54876, {{171068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54877, {{171070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54878, {{171071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54879, {{171073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54880, {{171074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54881, {{171075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54882, {{171077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54883, {{171078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54884, {{171079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54885, {{171080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54886, {{171081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54887, {{171083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54888, {{171085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54889, {{171086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54890, {{171087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54891, {{171088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54892, {{171089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54893, {{171090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54894, {{171091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54895, {{171092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54896, {{171094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54897, {{171095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54898, {{171096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54899, {{171097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54900, {{171098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54901, {{171099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54902, {{171100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54903, {{171101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54904, {{171102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54905, {{171103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54906, {{171104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54907, {{171105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54908, {{171106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54909, {{171107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54910, {{171108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54911, {{171109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54912, {{171110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54913, {{171111, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54914, {{171112, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54915, {{171114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54916, {{171115, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54917, {{171116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54918, {{171117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54919, {{171118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54920, {{171120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54921, {{171121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54922, {{171122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54923, {{171123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54924, {{171124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54925, {{171125, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54926, {{171126, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54927, {{171127, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54928, {{171128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54929, {{171129, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54930, {{171131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54931, {{171132, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54932, {{171134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{54933, {{171135, 1, 0, 0},{171135, 1, 1, 16}, {0, 0, 0, 0}}},{54933, {{171135, 1, 0, 0},{171135, 1, 1, 16}, {0, 0, 0, 0}}},{54934, {{171135, 1, 0, 0},{171135, 1, 1, 17}, {0, 0, 0, 0}}},{54934, {{171135, 1, 0, 0},{171135, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{54935, {{171136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54936, {{171137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54937, {{171138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54938, {{171139, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54939, {{171140, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54941, {{171144, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54942, {{171145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54943, {{171146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54944, {{171147, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54945, {{171148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54946, {{171149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54947, {{171150, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54948, {{171151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54949, {{171152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54950, {{171153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54951, {{171156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54952, {{171160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54953, {{171161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54954, {{171162, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54955, {{171163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54956, {{171164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54957, {{171165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54958, {{171166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54959, {{171168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54960, {{171169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54961, {{171170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54962, {{171171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54963, {{171172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54964, {{171173, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54965, {{171174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54966, {{171175, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54967, {{171176, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54968, {{171177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54969, {{171178, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54970, {{171179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54971, {{171180, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54972, {{171185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54973, {{171187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54974, {{171188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54975, {{171189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54976, {{171190, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54977, {{171191, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54978, {{171193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54979, {{171195, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54980, {{171196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54981, {{171197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54982, {{171198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54983, {{171199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54984, {{171200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54985, {{171201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54986, {{171202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54987, {{171205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54988, {{171206, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54989, {{171207, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54990, {{171208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54991, {{171209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54992, {{171211, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54993, {{171212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54994, {{171213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54995, {{171214, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54996, {{171215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54997, {{171216, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54998, {{171217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54999, {{171218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55000, {{171219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55001, {{171220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55002, {{171221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55003, {{171222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55004, {{171224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55005, {{171225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55006, {{171229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55007, {{171231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55008, {{171232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55009, {{171233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55010, {{171234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55011, {{171235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55012, {{171236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55013, {{171237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55014, {{171239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55015, {{171240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55016, {{171241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55017, {{171242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55018, {{171245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55019, {{171246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55020, {{171247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55021, {{171248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55022, {{171249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55023, {{171250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55024, {{171253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55025, {{171254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55026, {{171255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55027, {{171256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55028, {{171257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55029, {{171258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55030, {{171264, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55031, {{171267, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55032, {{171270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55033, {{171271, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55034, {{171273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55035, {{171274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55036, {{171275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55037, {{171277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55038, {{171278, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55039, {{171280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55040, {{171281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55041, {{171282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55042, {{171283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55043, {{171288, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55044, {{171289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55045, {{171291, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55046, {{171292, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55047, {{171293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55048, {{171294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55049, {{171295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55050, {{171296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55051, {{171297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55052, {{171298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55053, {{171299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55054, {{171300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55055, {{171304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55056, {{171305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55057, {{171306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55058, {{171309, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55059, {{171310, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55060, {{171311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55061, {{171313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55062, {{171314, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55063, {{171315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{55064, {{171316, 1, 0, 0},{171316, 1, 1, 16}, {0, 0, 0, 0}}},{55064, {{171316, 1, 0, 0},{171316, 1, 1, 16}, {0, 0, 0, 0}}},{55065, {{171316, 1, 0, 0},{171316, 1, 1, 17}, {0, 0, 0, 0}}},{55065, {{171316, 1, 0, 0},{171316, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55066, {{171318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55067, {{171319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55068, {{171320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55069, {{171322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55070, {{171323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55071, {{171327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55072, {{171328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55073, {{171329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55074, {{171330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55075, {{171332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55076, {{171333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55077, {{171335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55078, {{171336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{55079, {{171337, 1, 0, 0},{171337, 1, 1, 17}, {0, 0, 0, 0}}},{55079, {{171337, 1, 0, 0},{171337, 1, 1, 17}, {0, 0, 0, 0}}},{59085, {{171337, 1, 0, 0},{171337, 1, 1, 16}, {0, 0, 0, 0}}},{59085, {{171337, 1, 0, 0},{171337, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55080, {{171339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55081, {{171340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55082, {{171341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55083, {{171342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55084, {{171343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55085, {{171344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55086, {{171345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55087, {{171347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55088, {{171348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55089, {{171349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55090, {{171350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55091, {{171351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55092, {{171355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55093, {{171360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55094, {{171361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55095, {{171362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55096, {{171363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55097, {{171364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55098, {{171365, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55099, {{171366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55100, {{171367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55101, {{171368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55102, {{171369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55103, {{171370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55104, {{171371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55105, {{171372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55106, {{171373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55107, {{171374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55108, {{171375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55109, {{171376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55110, {{171377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55111, {{171382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55112, {{171383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55113, {{171385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55114, {{171390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55115, {{171392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55116, {{171393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55117, {{171394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55118, {{171395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55119, {{171396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55120, {{171397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55121, {{171398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55122, {{171399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55124, {{171400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55125, {{171401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55126, {{171402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55127, {{171403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55128, {{171406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55129, {{171409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55130, {{171414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55131, {{171415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55132, {{171416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55133, {{171418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55134, {{171422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55135, {{171424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55151, {{171425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55136, {{171426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55137, {{171427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55138, {{171428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55139, {{171429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55140, {{171430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55141, {{171431, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55142, {{171432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55143, {{171433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55144, {{171434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55145, {{171436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55146, {{171437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55147, {{171438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55148, {{171439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55149, {{171440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55150, {{171441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55152, {{171453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55153, {{171454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55154, {{171455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55155, {{171458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55156, {{171459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55157, {{171462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55158, {{171467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55159, {{171468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55160, {{171469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55161, {{171470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55162, {{171471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55163, {{171472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55164, {{171473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55165, {{171474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55166, {{171475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55167, {{171476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55168, {{171477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55169, {{171478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55170, {{171479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55171, {{171480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55172, {{171481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55173, {{171482, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55174, {{171483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59117, {{171494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55175, {{171495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59124, {{171498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55176, {{171500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55177, {{171508, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55178, {{171509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55179, {{171511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{55180, {{171512, 1, 0, 0},{171512, 1, 1, 16}, {0, 0, 0, 0}}},{55180, {{171512, 1, 0, 0},{171512, 1, 1, 16}, {0, 0, 0, 0}}},{59133, {{171512, 1, 0, 0},{171512, 1, 1, 17}, {0, 0, 0, 0}}},{59133, {{171512, 1, 0, 0},{171512, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{55181, {{171513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55182, {{171514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55183, {{171515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55184, {{171516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55185, {{171518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55186, {{171519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55187, {{171520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55188, {{171521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55189, {{171522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55190, {{171524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55191, {{171525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55192, {{171526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55193, {{171527, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55194, {{171529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55195, {{171530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55196, {{171531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55197, {{171532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55198, {{171533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55199, {{171534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55200, {{171536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55201, {{171541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55202, {{171542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55203, {{171551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55204, {{171552, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55205, {{171557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55206, {{171559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55207, {{171560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55208, {{171561, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55209, {{171563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55210, {{171565, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55211, {{171566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55212, {{171574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55213, {{171577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55214, {{171579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55215, {{171580, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{55216, {{171581, 1, 0, 0},{171581, 1, 1, 19}, {0, 0, 0, 0}}},{55216, {{171581, 1, 0, 0},{171581, 1, 1, 19}, {0, 0, 0, 0}}},{55217, {{171581, 1, 0, 0},{171581, 1, 1, 18}, {0, 0, 0, 0}}},{55217, {{171581, 1, 0, 0},{171581, 1, 1, 18}, {0, 0, 0, 0}}},{55218, {{171581, 1, 0, 0},{171581, 1, 1, 20}, {0, 0, 0, 0}}},{55218, {{171581, 1, 0, 0},{171581, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {1, {{55219, {{171582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55220, {{171583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55221, {{171584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55222, {{171585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55223, {{171586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55224, {{171587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55225, {{171588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55226, {{171589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55227, {{171590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55228, {{171591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55229, {{171592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55230, {{171593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55231, {{171594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55232, {{171595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55233, {{171597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55234, {{171598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55235, {{171599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55236, {{171600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55237, {{171601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55238, {{171602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55239, {{171613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55240, {{171618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55241, {{171621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60354, {{171624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55242, {{171630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55243, {{171633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55244, {{171634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55245, {{171635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55246, {{171636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55247, {{171638, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{55248, {{171639, 1, 0, 0},{171639, 1, 1, 16}, {0, 0, 0, 0}}},{55248, {{171639, 1, 0, 0},{171639, 1, 1, 16}, {0, 0, 0, 0}}},{68093, {{171639, 1, 0, 0},{171639, 1, 1, 17}, {0, 0, 0, 0}}},{68093, {{171639, 1, 0, 0},{171639, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{55249, {{171640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{55250, {{171642, 1, 0, 0},{171642, 1, 1, 16}, {0, 0, 0, 0}}},{55250, {{171642, 1, 0, 0},{171642, 1, 1, 16}, {0, 0, 0, 0}}},{55251, {{171642, 1, 0, 0},{171642, 1, 1, 17}, {0, 0, 0, 0}}},{55251, {{171642, 1, 0, 0},{171642, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{55252, {{171643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55253, {{171644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55254, {{171645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55255, {{171646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55256, {{171647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55257, {{171648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55258, {{171649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55259, {{171650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55260, {{171651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55261, {{171652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55262, {{171653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55263, {{171654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{55264, {{171658, 1, 0, 0},{171658, 1, 1, 18}, {0, 0, 0, 0}}},{55264, {{171658, 1, 0, 0},{171658, 1, 1, 18}, {0, 0, 0, 0}}},{55265, {{171658, 1, 0, 0},{171658, 1, 1, 17}, {0, 0, 0, 0}}},{55265, {{171658, 1, 0, 0},{171658, 1, 1, 17}, {0, 0, 0, 0}}},{55266, {{171658, 1, 0, 0},{171658, 1, 1, 19}, {0, 0, 0, 0}}},{55266, {{171658, 1, 0, 0},{171658, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55267, {{171664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55268, {{171673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55269, {{171674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55270, {{171675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55271, {{171676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55272, {{171681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55273, {{171691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59154, {{171692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55274, {{171693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55275, {{171694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55276, {{171695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55277, {{171696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55278, {{171697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55279, {{171698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55280, {{171699, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55281, {{171700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55282, {{171701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55283, {{171702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55284, {{171703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55285, {{171704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55286, {{171705, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55287, {{171706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55288, {{171707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55289, {{171708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55290, {{171709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59157, {{171711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55291, {{171716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55292, {{171721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55293, {{171723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55294, {{171725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55295, {{171730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55296, {{171734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55297, {{171736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55298, {{171737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55299, {{171738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{55300, {{171739, 1, 0, 0},{171739, 1, 1, 18}, {0, 0, 0, 0}}},{55300, {{171739, 1, 0, 0},{171739, 1, 1, 18}, {0, 0, 0, 0}}},{55301, {{171739, 1, 0, 0},{171739, 1, 1, 17}, {0, 0, 0, 0}}},{55301, {{171739, 1, 0, 0},{171739, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{55302, {{171740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55303, {{171741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55304, {{171742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55305, {{171743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{55306, {{171744, 1, 0, 0},{171744, 1, 1, 16}, {0, 0, 0, 0}}},{55306, {{171744, 1, 0, 0},{171744, 1, 1, 16}, {0, 0, 0, 0}}},{55307, {{171744, 1, 0, 0},{171744, 1, 1, 17}, {0, 0, 0, 0}}},{55307, {{171744, 1, 0, 0},{171744, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{55308, {{171745, 1, 0, 0},{171745, 1, 1, 16}, {0, 0, 0, 0}}},{55308, {{171745, 1, 0, 0},{171745, 1, 1, 16}, {0, 0, 0, 0}}},{55309, {{171745, 1, 0, 0},{171745, 1, 1, 17}, {0, 0, 0, 0}}},{55309, {{171745, 1, 0, 0},{171745, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{55310, {{171746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55311, {{171747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55312, {{171748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55314, {{171749, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55315, {{171751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55316, {{171752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55317, {{171753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55318, {{171762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55319, {{171764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55320, {{171765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55321, {{171766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55322, {{171771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55323, {{171777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55324, {{171778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55325, {{171779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55326, {{171780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55327, {{171781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55328, {{171782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55329, {{171783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55330, {{171784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55331, {{171785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55332, {{171786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55333, {{171787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55334, {{171788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55335, {{171789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55336, {{171790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55337, {{171791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55338, {{171792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55339, {{171794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55340, {{171800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55341, {{171804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55342, {{171805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55343, {{171806, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55344, {{171807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55345, {{171808, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55346, {{171809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55347, {{171813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55348, {{171814, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55349, {{171815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55350, {{171816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55351, {{171817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {2, {{55352, {{171818, 1, 0, 0}, {0, 0, 0, 0}}},{55353, {{171818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55354, {{171819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55355, {{171820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55356, {{171821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55357, {{171822, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55358, {{171828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55360, {{171835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55361, {{171836, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55362, {{171837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55363, {{171839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55364, {{171840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55365, {{171841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55366, {{171846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55367, {{171852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55368, {{171853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55369, {{171854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55370, {{171855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55371, {{171856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55372, {{171859, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55373, {{171861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55374, {{171864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55375, {{171866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55376, {{171867, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55377, {{171868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55378, {{171869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55379, {{171870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55380, {{171871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55381, {{171873, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55382, {{171875, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55383, {{171879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55384, {{171880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55385, {{171881, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55386, {{171882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55387, {{171883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55388, {{171884, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59198, {{171887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55389, {{171888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55390, {{171889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55391, {{171891, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55392, {{171893, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55393, {{171894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55394, {{171895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55395, {{171919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55396, {{171921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55397, {{171922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55398, {{171923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55399, {{171924, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55400, {{171925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55401, {{171926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55402, {{171927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55403, {{171928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55404, {{171929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55405, {{171931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55406, {{171932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55407, {{171933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55408, {{171934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55409, {{171935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55410, {{171936, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55411, {{171937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55412, {{171938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55413, {{171939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55414, {{171940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55415, {{171941, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55416, {{171942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55417, {{171943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55418, {{171952, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55419, {{171953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55420, {{171954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55421, {{171956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55422, {{171957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55423, {{171958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55424, {{171959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55425, {{171960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55426, {{171961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55427, {{171962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55428, {{171963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55429, {{171964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55430, {{171965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55431, {{171966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55432, {{171967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55433, {{171968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55434, {{171970, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55435, {{171971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55436, {{171972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55437, {{171973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55438, {{171974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55439, {{171975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55441, {{171976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55442, {{171977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55443, {{171978, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55444, {{171979, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55445, {{171982, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55446, {{171984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55447, {{171985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55448, {{171986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55449, {{171987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55450, {{171988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55451, {{171989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55452, {{171990, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55453, {{171995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55454, {{171996, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55455, {{171997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55456, {{171998, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55457, {{171999, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55458, {{172000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55459, {{172001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55460, {{172002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55461, {{172003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55462, {{172004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55463, {{172005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55464, {{172006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55465, {{172007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55466, {{172008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55467, {{172009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55468, {{172010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55469, {{172011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55470, {{172012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55471, {{172013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55472, {{172014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55473, {{172023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55474, {{172024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55475, {{172028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59209, {{172031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55476, {{172032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55477, {{172033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55478, {{172039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55479, {{172040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55480, {{172041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55481, {{172042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55482, {{172043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55483, {{172044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55484, {{172045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55485, {{172047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55486, {{172050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55487, {{172051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55488, {{172052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55489, {{172053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55490, {{172054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55491, {{172055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55492, {{172056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55493, {{172057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55494, {{172058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55495, {{172059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55496, {{172061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55497, {{172062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55498, {{172063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55499, {{172064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55500, {{172065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55501, {{172066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55502, {{172067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55503, {{172068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55504, {{172069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55505, {{172070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55506, {{172071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55507, {{172072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55508, {{172073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55509, {{172074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55510, {{172075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55511, {{172076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55512, {{172077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{55513, {{172079, 1, 0, 0},{172079, 1, 1, 17}, {0, 0, 0, 0}}},{55513, {{172079, 1, 0, 0},{172079, 1, 1, 17}, {0, 0, 0, 0}}},{55514, {{172079, 1, 0, 0},{172079, 1, 1, 18}, {0, 0, 0, 0}}},{55514, {{172079, 1, 0, 0},{172079, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55515, {{172088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55516, {{172090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55517, {{172092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55518, {{172093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55519, {{172095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55520, {{172096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55521, {{172104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55522, {{172105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55523, {{172106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{55524, {{172107, 1, 0, 0},{172107, 1, 1, 16}, {0, 0, 0, 0}}},{55524, {{172107, 1, 0, 0},{172107, 1, 1, 16}, {0, 0, 0, 0}}},{68094, {{172107, 1, 0, 0},{172107, 1, 1, 17}, {0, 0, 0, 0}}},{68094, {{172107, 1, 0, 0},{172107, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{55525, {{172108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55526, {{172109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55527, {{172110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55528, {{172111, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55529, {{172112, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55530, {{172113, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55531, {{172114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55532, {{172115, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55533, {{172116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55534, {{172117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55535, {{172118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55536, {{172119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55537, {{172120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55538, {{172121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55539, {{172122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55540, {{172123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55541, {{172124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55542, {{172125, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55543, {{172126, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55544, {{172127, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55545, {{172128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{55546, {{172129, 1, 0, 0},{172129, 1, 1, 16}, {0, 0, 0, 0}}},{55546, {{172129, 1, 0, 0},{172129, 1, 1, 16}, {0, 0, 0, 0}}},{55547, {{172129, 1, 0, 0},{172129, 1, 1, 17}, {0, 0, 0, 0}}},{55547, {{172129, 1, 0, 0},{172129, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{55548, {{172130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55549, {{172131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55550, {{172133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55551, {{172134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55552, {{172135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55553, {{172136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55554, {{172137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55555, {{172138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55556, {{172141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55557, {{172142, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55558, {{172144, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55559, {{172146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55560, {{172151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55561, {{172152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55562, {{172153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55563, {{172154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55564, {{172155, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55565, {{172156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55566, {{172157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55567, {{172158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55568, {{172159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55569, {{172160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55570, {{172161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55571, {{172162, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55572, {{172163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55573, {{172164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55574, {{172165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55575, {{172166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55576, {{172167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55577, {{172168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55578, {{172169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55579, {{172170, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55580, {{172171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55581, {{172172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55582, {{172173, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55583, {{172174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55584, {{172175, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55585, {{172176, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55586, {{172177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55587, {{172178, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55588, {{172179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55589, {{172180, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55590, {{172181, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55591, {{172182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55592, {{172186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55593, {{172187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55594, {{172188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55595, {{172189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55596, {{172190, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55597, {{172191, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55598, {{172192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55599, {{172193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55600, {{172196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55601, {{172197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55602, {{172202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55603, {{172205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55604, {{172212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55605, {{172213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55606, {{172214, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55607, {{172215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55608, {{172216, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55609, {{172217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55610, {{172218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55611, {{172219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55612, {{172220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55613, {{172221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55614, {{172222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55615, {{172223, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55616, {{172224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55617, {{172225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55618, {{172226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55619, {{172227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55620, {{172228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55621, {{172229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55622, {{172230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55623, {{172231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{55624, {{172232, 1, 0, 0},{172232, 1, 1, 16}, {0, 0, 0, 0}}},{55624, {{172232, 1, 0, 0},{172232, 1, 1, 16}, {0, 0, 0, 0}}},{55625, {{172232, 1, 0, 0},{172232, 1, 1, 17}, {0, 0, 0, 0}}},{55625, {{172232, 1, 0, 0},{172232, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{55627, {{172233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55628, {{172234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55629, {{172235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55630, {{172236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55631, {{172237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55632, {{172238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55633, {{172239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55634, {{172240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55635, {{172241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55636, {{172242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55637, {{172243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55638, {{172244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55639, {{172245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55640, {{172247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55641, {{172248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55642, {{172249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55643, {{172250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55644, {{172251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55645, {{172252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55646, {{172253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55647, {{172259, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55648, {{172260, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55649, {{172262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55650, {{172263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55651, {{172266, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55652, {{172267, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55653, {{172269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55654, {{172271, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55655, {{172272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55656, {{172278, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55657, {{172279, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{55658, {{172281, 1, 0, 0},{172281, 1, 1, 17}, {0, 0, 0, 0}}},{55658, {{172281, 1, 0, 0},{172281, 1, 1, 17}, {0, 0, 0, 0}}},{55659, {{172281, 1, 0, 0},{172281, 1, 1, 18}, {0, 0, 0, 0}}},{55659, {{172281, 1, 0, 0},{172281, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{55660, {{172282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55661, {{172283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55662, {{172284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55663, {{172285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55664, {{172286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55665, {{172287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55666, {{172288, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55667, {{172289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55668, {{172290, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55669, {{172291, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55670, {{172292, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55671, {{172293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55672, {{172294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55673, {{172295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55674, {{172296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55675, {{172297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55676, {{172298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55677, {{172299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55678, {{172300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55679, {{172301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55680, {{172302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55681, {{172303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55682, {{172304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55683, {{172305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55684, {{172311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55685, {{172312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55686, {{172315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55687, {{172316, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55688, {{172317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55689, {{172318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55690, {{172329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55691, {{172330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55692, {{172331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55693, {{172333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55694, {{172334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55695, {{172335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55696, {{172336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55697, {{172337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55698, {{172338, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55699, {{172339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55700, {{172340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55701, {{172341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55702, {{172342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55703, {{172343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55704, {{172344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55705, {{172345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59232, {{172346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55706, {{172347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55707, {{172348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55708, {{172349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55709, {{172351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55710, {{172354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55711, {{172355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55712, {{172356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55713, {{172357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55714, {{172358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55715, {{172359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55716, {{172360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55717, {{172361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55718, {{172362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55719, {{172363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55720, {{172364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55721, {{172365, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59233, {{172368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55722, {{172374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55723, {{172378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55724, {{172380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55725, {{172381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55726, {{172383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55727, {{172385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55728, {{172388, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55729, {{172398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55730, {{172399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55731, {{172400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55732, {{172401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55733, {{172402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55734, {{172403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55735, {{172404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55736, {{172405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55737, {{172406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55738, {{172408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55739, {{172409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55740, {{172410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55741, {{172411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55742, {{172412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55743, {{172413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55744, {{172414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55745, {{172415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55746, {{172416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55747, {{172417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55748, {{172418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55749, {{172419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55750, {{172420, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55751, {{172421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55752, {{172422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55753, {{172423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55754, {{172424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55755, {{172425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55756, {{172426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55757, {{172427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55758, {{172428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55759, {{172429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55760, {{172432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55761, {{172438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55762, {{172439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55763, {{172442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55764, {{172443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55765, {{172445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55766, {{172448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55767, {{172449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55768, {{172452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55769, {{172456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55770, {{172457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55771, {{172458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55772, {{172459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55773, {{172461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55774, {{172463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55775, {{172464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55776, {{172465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55777, {{172466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55778, {{172467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55779, {{172468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55780, {{172469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55781, {{172470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55782, {{172471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55783, {{172472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55784, {{172473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55785, {{172474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55786, {{172475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55787, {{172476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55788, {{172477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55789, {{172478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55790, {{172479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55791, {{172480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55792, {{172481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55793, {{172482, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55794, {{172484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55795, {{172486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55796, {{172487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55797, {{172488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55798, {{172489, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55799, {{172490, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55800, {{172497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55801, {{172498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55802, {{172499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55803, {{172500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55804, {{172501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55805, {{172502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55806, {{172503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55807, {{172504, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55808, {{172505, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55809, {{172506, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55810, {{172507, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55811, {{172508, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55812, {{172509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55813, {{172510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55814, {{172512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55815, {{172513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59243, {{172514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55816, {{172524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55817, {{172525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55818, {{172526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55819, {{172527, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55820, {{172528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55821, {{172529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55822, {{172530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55823, {{172531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{55824, {{172532, 1, 0, 0},{172532, 1, 1, 16}, {0, 0, 0, 0}}},{55824, {{172532, 1, 0, 0},{172532, 1, 1, 16}, {0, 0, 0, 0}}},{55825, {{172532, 1, 0, 0},{172532, 1, 1, 17}, {0, 0, 0, 0}}},{55825, {{172532, 1, 0, 0},{172532, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{55826, {{172533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55827, {{172535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55828, {{172536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55829, {{172537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55830, {{172538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55831, {{172540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55832, {{172543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55833, {{172544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55834, {{172545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55835, {{172547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55836, {{172548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55837, {{172549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55838, {{172550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55839, {{172551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55840, {{172553, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55841, {{172554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55842, {{172555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55843, {{172556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55844, {{172557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55845, {{172560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55846, {{172561, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55847, {{172562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55848, {{172563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55849, {{172565, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55850, {{172567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55851, {{172568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55852, {{172570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55853, {{172571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55854, {{172572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55855, {{172573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55856, {{172574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55857, {{172578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55858, {{172580, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55859, {{172581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55860, {{172582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55861, {{172583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55862, {{172584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55863, {{172585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55864, {{172587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55865, {{172590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55866, {{172591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55867, {{172592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55868, {{172593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55869, {{172596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55870, {{172597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55871, {{172599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55872, {{172600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55873, {{172601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55874, {{172602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55875, {{172603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55876, {{172604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55877, {{172607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55878, {{172608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55879, {{172630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55880, {{172631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55881, {{172632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55882, {{172633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55883, {{172634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55884, {{172635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55885, {{172636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55886, {{172637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55887, {{172638, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55888, {{172639, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55889, {{172641, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55890, {{172642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55891, {{172643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55892, {{172644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55893, {{172646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55894, {{172647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55895, {{172648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55896, {{172649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55898, {{172650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55899, {{172653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55900, {{172654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55901, {{172655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55902, {{172656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55903, {{172657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55904, {{172658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55905, {{172659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55906, {{172661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55907, {{172662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55908, {{172663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55909, {{172664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55910, {{172666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55911, {{172667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55912, {{172671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55913, {{172672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55914, {{172673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55915, {{172674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55916, {{172675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55917, {{172676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55918, {{172678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55919, {{172679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55920, {{172680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55921, {{172681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55922, {{172682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55923, {{172683, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55924, {{172685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55925, {{172687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55926, {{172688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{55927, {{172689, 1, 0, 0},{172689, 1, 1, 16}, {0, 0, 0, 0}}},{55927, {{172689, 1, 0, 0},{172689, 1, 1, 16}, {0, 0, 0, 0}}},{55928, {{172689, 1, 0, 0},{172689, 1, 1, 17}, {0, 0, 0, 0}}},{55928, {{172689, 1, 0, 0},{172689, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{55929, {{172690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55930, {{172691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55931, {{172692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55932, {{172693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{55933, {{172694, 1, 0, 0},{172694, 1, 1, 16}, {0, 0, 0, 0}}},{55933, {{172694, 1, 0, 0},{172694, 1, 1, 16}, {0, 0, 0, 0}}},{55934, {{172694, 1, 0, 0},{172694, 1, 1, 17}, {0, 0, 0, 0}}},{55934, {{172694, 1, 0, 0},{172694, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{55935, {{172695, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55936, {{172696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55937, {{172697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55938, {{172699, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55939, {{172700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55940, {{172706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55941, {{172709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55942, {{172710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55943, {{172711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{55944, {{172712, 1, 0, 0},{172712, 1, 1, 16}, {0, 0, 0, 0}}},{55944, {{172712, 1, 0, 0},{172712, 1, 1, 16}, {0, 0, 0, 0}}},{55945, {{172712, 1, 0, 0},{172712, 1, 1, 17}, {0, 0, 0, 0}}},{55945, {{172712, 1, 0, 0},{172712, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{55946, {{172713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55947, {{172714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55948, {{172715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55949, {{172716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55950, {{172717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55951, {{172719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55952, {{172721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55953, {{172723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55954, {{172724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55955, {{172725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55956, {{172726, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55957, {{172727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55958, {{172728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55959, {{172729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55960, {{172730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55961, {{172731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55962, {{172737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55963, {{172738, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55964, {{172739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55965, {{172740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55966, {{172741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55967, {{172742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55968, {{172743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55969, {{172744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55970, {{172745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55971, {{172746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55972, {{172747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55973, {{172748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55974, {{172750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55975, {{172751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55976, {{172752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55977, {{172753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55978, {{172755, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55979, {{172756, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55980, {{172757, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55981, {{172758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55982, {{172759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55983, {{172760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55984, {{172761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55985, {{172762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55986, {{172764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55987, {{172765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55988, {{172766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55989, {{172769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55990, {{172770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55991, {{172771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55992, {{172773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55993, {{172775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55994, {{172778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55995, {{172779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55996, {{172780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55997, {{172782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55998, {{172783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{55999, {{172784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56000, {{172785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56001, {{172786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56002, {{172787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56003, {{172789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56004, {{172791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56005, {{172792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56006, {{172793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56007, {{172794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56008, {{172795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56009, {{172796, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56010, {{172798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56011, {{172800, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{56012, {{172801, 1, 0, 0},{172801, 1, 1, 16}, {0, 0, 0, 0}}},{56012, {{172801, 1, 0, 0},{172801, 1, 1, 16}, {0, 0, 0, 0}}},{56013, {{172801, 1, 0, 0},{172801, 1, 1, 17}, {0, 0, 0, 0}}},{56013, {{172801, 1, 0, 0},{172801, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{56014, {{172802, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56015, {{172803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56016, {{172804, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56017, {{172805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56018, {{172806, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56019, {{172807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{56020, {{172808, 1, 0, 0},{172808, 1, 1, 16}, {0, 0, 0, 0}}},{56020, {{172808, 1, 0, 0},{172808, 1, 1, 16}, {0, 0, 0, 0}}},{59257, {{172808, 1, 0, 0},{172808, 1, 1, 17}, {0, 0, 0, 0}}},{59257, {{172808, 1, 0, 0},{172808, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{56021, {{172809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56022, {{172810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56023, {{172811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56024, {{172812, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56025, {{172813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56026, {{172815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56027, {{172817, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56028, {{172818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56029, {{172819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56030, {{172821, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56031, {{172822, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56032, {{172823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56033, {{172824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56034, {{172827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56035, {{172829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56036, {{172830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56037, {{172831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56038, {{172834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56039, {{172835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56040, {{172836, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56041, {{172839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56042, {{172841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56043, {{172842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56044, {{172843, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56045, {{172844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56046, {{172845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56047, {{172846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56048, {{172847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56049, {{172848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56050, {{172849, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56051, {{172850, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{56052, {{172851, 1, 0, 0},{172851, 1, 1, 16}, {0, 0, 0, 0}}},{56052, {{172851, 1, 0, 0},{172851, 1, 1, 16}, {0, 0, 0, 0}}},{59259, {{172851, 1, 0, 0},{172851, 1, 1, 17}, {0, 0, 0, 0}}},{59259, {{172851, 1, 0, 0},{172851, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56053, {{172853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56054, {{172854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56055, {{172855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56056, {{172856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56057, {{172857, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56058, {{172858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56059, {{172860, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56060, {{172861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56061, {{172862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56062, {{172863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56063, {{172864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56064, {{172865, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56065, {{172866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56066, {{172867, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56067, {{172868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56068, {{172869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56069, {{172870, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{56070, {{172871, 1, 0, 0},{172871, 1, 1, 16}, {0, 0, 0, 0}}},{56070, {{172871, 1, 0, 0},{172871, 1, 1, 16}, {0, 0, 0, 0}}},{56071, {{172871, 1, 0, 0},{172871, 1, 1, 17}, {0, 0, 0, 0}}},{56071, {{172871, 1, 0, 0},{172871, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{56072, {{172872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56073, {{172873, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56074, {{172875, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56075, {{172876, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56076, {{172877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56077, {{172878, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56078, {{172879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56079, {{172880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56080, {{172881, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{56081, {{172882, 1, 0, 0},{172882, 1, 1, 16}, {0, 0, 0, 0}}},{56081, {{172882, 1, 0, 0},{172882, 1, 1, 16}, {0, 0, 0, 0}}},{59260, {{172882, 1, 0, 0},{172882, 1, 1, 17}, {0, 0, 0, 0}}},{59260, {{172882, 1, 0, 0},{172882, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{56082, {{172883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56083, {{172884, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56084, {{172885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56085, {{172887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56086, {{172888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56087, {{172890, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56088, {{172891, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56089, {{172892, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56090, {{172893, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56091, {{172894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56092, {{172895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56093, {{172896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56094, {{172897, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56095, {{172899, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56096, {{172900, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56097, {{172901, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56098, {{172902, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56099, {{172903, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56100, {{172904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56101, {{172905, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{56102, {{172906, 1, 0, 0},{172906, 1, 1, 16}, {0, 0, 0, 0}}},{56102, {{172906, 1, 0, 0},{172906, 1, 1, 16}, {0, 0, 0, 0}}},{59262, {{172906, 1, 0, 0},{172906, 1, 1, 17}, {0, 0, 0, 0}}},{59262, {{172906, 1, 0, 0},{172906, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{56103, {{172907, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56104, {{172908, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56105, {{172909, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56106, {{172912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56107, {{172914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56108, {{172915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56109, {{172916, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56110, {{172917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56111, {{172918, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56112, {{172920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56113, {{172921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56114, {{172922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56115, {{172923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56116, {{172924, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56117, {{172925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56118, {{172926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56119, {{172927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56120, {{172928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56121, {{172929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56122, {{172930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56123, {{172931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56124, {{172932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56125, {{172933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56126, {{172934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56127, {{172935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56128, {{172940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{56129, {{172946, 1, 0, 0},{172946, 1, 1, 16}, {0, 0, 0, 0}}},{56129, {{172946, 1, 0, 0},{172946, 1, 1, 16}, {0, 0, 0, 0}}},{56130, {{172946, 1, 0, 0},{172946, 1, 1, 17}, {0, 0, 0, 0}}},{56130, {{172946, 1, 0, 0},{172946, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{56131, {{172947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56132, {{172948, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56133, {{172949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56134, {{172950, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56135, {{172951, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56136, {{172954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56137, {{172955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56138, {{172956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56139, {{172957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56140, {{172958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56141, {{172960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56142, {{172961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56143, {{172963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56144, {{172964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56145, {{172965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56146, {{172966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56147, {{172967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56148, {{172968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56149, {{172970, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56150, {{172971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56151, {{172972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56152, {{172973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56153, {{172974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56154, {{172975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{56155, {{172976, 1, 0, 0},{172976, 1, 1, 16}, {0, 0, 0, 0}}},{56155, {{172976, 1, 0, 0},{172976, 1, 1, 16}, {0, 0, 0, 0}}},{56156, {{172976, 1, 0, 0},{172976, 1, 1, 17}, {0, 0, 0, 0}}},{56156, {{172976, 1, 0, 0},{172976, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56157, {{172978, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56158, {{172980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56159, {{172981, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56160, {{172982, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56161, {{172984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56162, {{172985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56163, {{172986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56164, {{172987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56165, {{172988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56166, {{172989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56167, {{172990, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56168, {{172991, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56169, {{172992, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56170, {{172993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56171, {{172994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56172, {{172997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56173, {{172998, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56174, {{172999, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56175, {{173000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56176, {{173001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56177, {{173002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56178, {{173003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56179, {{173004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56180, {{173005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56181, {{173007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56182, {{173008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56183, {{173009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56184, {{173010, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56185, {{173011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56186, {{173012, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56187, {{173013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56188, {{173015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56189, {{173016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56190, {{173017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56191, {{173018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56192, {{173019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56193, {{173020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56194, {{173021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56195, {{173022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56196, {{173023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56197, {{173024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56198, {{173025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56199, {{173026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56200, {{173028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56201, {{173029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56202, {{173030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56203, {{173031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56204, {{173032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56205, {{173033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56206, {{173034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56207, {{173035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56208, {{173036, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56209, {{173038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56210, {{173039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56211, {{173040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56212, {{173042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56213, {{173043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56214, {{173044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56215, {{173046, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56216, {{173047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56217, {{173048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56218, {{173049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56219, {{173050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56220, {{173051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56221, {{173052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56222, {{173053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56223, {{173055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56224, {{173056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56225, {{173058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56226, {{173059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56227, {{173060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56228, {{173063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56229, {{173064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56230, {{173065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56231, {{173067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56232, {{173068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56233, {{173069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56234, {{173070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56235, {{173071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56236, {{173073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56237, {{173074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56238, {{173075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56239, {{173076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56240, {{173077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56241, {{173078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56242, {{173079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56243, {{173080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56244, {{173081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56245, {{173082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56246, {{173083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56247, {{173084, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56248, {{173085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56249, {{173086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56250, {{173087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56251, {{173088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56252, {{173091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56253, {{173092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56254, {{173093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56255, {{173094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56256, {{173095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56257, {{173096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56258, {{173097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56259, {{173098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56260, {{173099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56261, {{173100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56262, {{173101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56263, {{173102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56264, {{173103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56265, {{173104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56266, {{173106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56267, {{173107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56268, {{173108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56269, {{173111, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56270, {{173113, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56271, {{173114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56272, {{173115, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56273, {{173119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56274, {{173120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56275, {{173121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56276, {{173122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56277, {{173123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56278, {{173124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56279, {{173126, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56280, {{173127, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56281, {{173128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56282, {{173129, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56283, {{173130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56284, {{173131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56285, {{173132, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56286, {{173135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56287, {{173136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56288, {{173138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56289, {{173139, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56290, {{173140, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56291, {{173141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56292, {{173142, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56293, {{173143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56294, {{173144, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56295, {{173145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56296, {{173146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56297, {{173147, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56298, {{173148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56299, {{173149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56300, {{173150, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56301, {{173151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56302, {{173154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56303, {{173155, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56304, {{173158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56305, {{173159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56306, {{173160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56307, {{173161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56308, {{173162, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56309, {{173163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56310, {{173164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56311, {{173165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56312, {{173166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56313, {{173167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56314, {{173168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56315, {{173169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{56316, {{173170, 1, 0, 0},{173170, 1, 1, 16}, {0, 0, 0, 0}}},{56316, {{173170, 1, 0, 0},{173170, 1, 1, 16}, {0, 0, 0, 0}}},{56317, {{173170, 1, 0, 0},{173170, 1, 1, 17}, {0, 0, 0, 0}}},{56317, {{173170, 1, 0, 0},{173170, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{56318, {{173171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56319, {{173173, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56320, {{173174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56321, {{173175, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56322, {{173176, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56323, {{173177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56324, {{173179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56325, {{173180, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56326, {{173181, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56327, {{173182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56328, {{173183, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56329, {{173184, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56330, {{173185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56331, {{173186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56332, {{173187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56333, {{173188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56334, {{173190, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56335, {{173191, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56336, {{173192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56337, {{173194, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56338, {{173195, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56339, {{173197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56340, {{173199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56341, {{173200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56342, {{173201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56343, {{173202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56344, {{173203, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56345, {{173204, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56346, {{173205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56347, {{173206, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56348, {{173208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56349, {{173209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56350, {{173210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56351, {{173211, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56352, {{173212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56353, {{173213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56354, {{173214, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56355, {{173215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56356, {{173216, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56357, {{173217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56358, {{173218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56359, {{173219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56360, {{173223, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56361, {{173225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56362, {{173226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56363, {{173227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56364, {{173228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56365, {{173229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56366, {{173230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56367, {{173231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56368, {{173232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56369, {{173234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56370, {{173236, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56371, {{173240, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56372, {{173241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56373, {{173242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56374, {{173243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56375, {{173244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56376, {{173246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56377, {{173247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56378, {{173248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56379, {{173249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56380, {{173251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56381, {{173252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56382, {{173253, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56383, {{173254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56384, {{173257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56385, {{173258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56386, {{173259, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56387, {{173260, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56388, {{173262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56389, {{173263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56390, {{173264, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56391, {{173265, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56392, {{173266, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56393, {{173267, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56394, {{173268, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56395, {{173269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56396, {{173270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56397, {{173272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56398, {{173275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56399, {{173276, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56400, {{173277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56401, {{173278, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56402, {{173279, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56403, {{173282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56404, {{173284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56405, {{173285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56406, {{173287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56407, {{173288, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56408, {{173289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56409, {{173292, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56410, {{173293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56411, {{173294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56412, {{173295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56413, {{173296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56414, {{173297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56415, {{173298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56416, {{173300, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56417, {{173301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56418, {{173302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56419, {{173303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56420, {{173304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56421, {{173305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56422, {{173307, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56423, {{173308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56424, {{173309, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56425, {{173310, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56426, {{173311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56427, {{173312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56428, {{173313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{56429, {{173314, 1, 0, 0},{173314, 1, 1, 16}, {0, 0, 0, 0}}},{56429, {{173314, 1, 0, 0},{173314, 1, 1, 16}, {0, 0, 0, 0}}},{56430, {{173314, 1, 0, 0},{173314, 1, 1, 17}, {0, 0, 0, 0}}},{56430, {{173314, 1, 0, 0},{173314, 1, 1, 17}, {0, 0, 0, 0}}},{56431, {{173314, 1, 0, 0},{173314, 1, 1, 18}, {0, 0, 0, 0}}},{56431, {{173314, 1, 0, 0},{173314, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{29977, {{173315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{56432, {{173316, 1, 0, 0},{173316, 1, 1, 16}, {0, 0, 0, 0}}},{56432, {{173316, 1, 0, 0},{173316, 1, 1, 16}, {0, 0, 0, 0}}},{56433, {{173316, 1, 0, 0},{173316, 1, 1, 17}, {0, 0, 0, 0}}},{56433, {{173316, 1, 0, 0},{173316, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{56434, {{173317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56435, {{173318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{56436, {{173320, 1, 0, 0},{173320, 1, 1, 16}, {0, 0, 0, 0}}},{56436, {{173320, 1, 0, 0},{173320, 1, 1, 16}, {0, 0, 0, 0}}},{56437, {{173320, 1, 0, 0},{173320, 1, 1, 17}, {0, 0, 0, 0}}},{56437, {{173320, 1, 0, 0},{173320, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56438, {{173322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56439, {{173323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{56440, {{173325, 1, 0, 0},{173325, 1, 1, 17}, {0, 0, 0, 0}}},{56440, {{173325, 1, 0, 0},{173325, 1, 1, 17}, {0, 0, 0, 0}}},{57575, {{173325, 1, 0, 0},{173325, 1, 1, 16}, {0, 0, 0, 0}}},{57575, {{173325, 1, 0, 0},{173325, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{56441, {{173326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56442, {{173327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56443, {{173328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56444, {{173329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56445, {{173332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56446, {{173333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56447, {{173334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56448, {{173335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56449, {{173336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56450, {{173337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56451, {{173339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56452, {{173340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56453, {{173341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56454, {{173343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56463, {{173346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56456, {{173347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56457, {{173348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56458, {{173349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56459, {{173350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56460, {{173351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56461, {{173352, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56462, {{173353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56455, {{173354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56464, {{173356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56465, {{173358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56466, {{173359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56467, {{173360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56468, {{173361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56469, {{173363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56470, {{173365, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56471, {{173366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56472, {{173367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56473, {{173368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56474, {{173370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56475, {{173371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56476, {{173372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56477, {{173373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56478, {{173374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56479, {{173375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56480, {{173376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56481, {{173377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56482, {{173378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56483, {{173379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56484, {{173380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56485, {{173382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56486, {{173383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56487, {{173384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56488, {{173385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56489, {{173386, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56490, {{173387, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56491, {{173388, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{56492, {{173389, 1, 0, 0},{173389, 1, 1, 16}, {0, 0, 0, 0}}},{56492, {{173389, 1, 0, 0},{173389, 1, 1, 16}, {0, 0, 0, 0}}},{56493, {{173389, 1, 0, 0},{173389, 1, 1, 17}, {0, 0, 0, 0}}},{56493, {{173389, 1, 0, 0},{173389, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{56494, {{173390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56495, {{173391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56496, {{173392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56497, {{173393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56498, {{173394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56499, {{173395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56500, {{173396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56501, {{173399, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56502, {{173400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56503, {{173402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56504, {{173403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56505, {{173404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56506, {{173405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56507, {{173406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56508, {{173407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56509, {{173408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{56510, {{173412, 1, 0, 0},{173412, 1, 1, 16}, {0, 0, 0, 0}}},{56510, {{173412, 1, 0, 0},{173412, 1, 1, 16}, {0, 0, 0, 0}}},{56511, {{173412, 1, 0, 0},{173412, 1, 1, 17}, {0, 0, 0, 0}}},{56511, {{173412, 1, 0, 0},{173412, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56512, {{173414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56513, {{173415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56514, {{173416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56515, {{173417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56516, {{173418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56517, {{173419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56518, {{173420, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56519, {{173421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56520, {{173422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{56521, {{173423, 1, 0, 0},{173423, 1, 1, 16}, {0, 0, 0, 0}}},{56521, {{173423, 1, 0, 0},{173423, 1, 1, 16}, {0, 0, 0, 0}}},{68095, {{173423, 1, 0, 0},{173423, 1, 1, 17}, {0, 0, 0, 0}}},{68095, {{173423, 1, 0, 0},{173423, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{56522, {{173424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56523, {{173425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56524, {{173428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56525, {{173429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56526, {{173430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56527, {{173431, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56528, {{173432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56529, {{173433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56530, {{173434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56531, {{173437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56532, {{173438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56533, {{173439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56534, {{173440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56535, {{173441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56536, {{173442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56537, {{173443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56538, {{173444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{56539, {{173445, 1, 0, 0},{173445, 1, 1, 16}, {0, 0, 0, 0}}},{56539, {{173445, 1, 0, 0},{173445, 1, 1, 16}, {0, 0, 0, 0}}},{56540, {{173445, 1, 0, 0},{173445, 1, 1, 17}, {0, 0, 0, 0}}},{56540, {{173445, 1, 0, 0},{173445, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{56541, {{173446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56542, {{173447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56543, {{173448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56544, {{173449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56545, {{173451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56546, {{173452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56547, {{173453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56548, {{173454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56549, {{173455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56550, {{173456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56551, {{173457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56552, {{173458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56553, {{173459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56554, {{173461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56555, {{173463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56556, {{173466, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56557, {{173467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56558, {{173468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56559, {{173470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56560, {{173473, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56561, {{173474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56562, {{173475, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56563, {{173476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56564, {{173477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56565, {{173478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56566, {{173479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56567, {{173480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56568, {{173481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56569, {{173482, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56570, {{173486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56571, {{173487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56572, {{173488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56573, {{173491, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56574, {{173493, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56575, {{173494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56576, {{173495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56577, {{173499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56578, {{173500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56579, {{173501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56580, {{173502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56581, {{173505, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56582, {{173506, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56583, {{173509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56584, {{173510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{56585, {{173511, 1, 0, 0},{173511, 1, 1, 16}, {0, 0, 0, 0}}},{56585, {{173511, 1, 0, 0},{173511, 1, 1, 16}, {0, 0, 0, 0}}},{56586, {{173511, 1, 0, 0},{173511, 1, 1, 17}, {0, 0, 0, 0}}},{56586, {{173511, 1, 0, 0},{173511, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56587, {{173513, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56588, {{173514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56589, {{173515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56590, {{173516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56591, {{173517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56592, {{173519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56593, {{173520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56594, {{173521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56595, {{173522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56596, {{173524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56597, {{173526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56598, {{173528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56599, {{173529, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56600, {{173530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56601, {{173531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56602, {{173532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56603, {{173533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56604, {{173534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56605, {{173535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56606, {{173536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56607, {{173537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56608, {{173538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56609, {{173539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56610, {{173540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56611, {{173541, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56612, {{173542, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56613, {{173543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56614, {{173544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56615, {{173545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56616, {{173546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56617, {{173547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56618, {{173548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56619, {{173549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56620, {{173550, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56621, {{173551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56622, {{173553, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56623, {{173554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56624, {{173556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56625, {{173557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56626, {{173558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56627, {{173559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56628, {{173560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56629, {{173561, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56630, {{173562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56631, {{173563, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56632, {{173565, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56633, {{173566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56634, {{173567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{56635, {{173568, 1, 0, 0},{173568, 1, 1, 16}, {0, 0, 0, 0}}},{56635, {{173568, 1, 0, 0},{173568, 1, 1, 16}, {0, 0, 0, 0}}},{56636, {{173568, 1, 0, 0},{173568, 1, 1, 17}, {0, 0, 0, 0}}},{56636, {{173568, 1, 0, 0},{173568, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{56637, {{173569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56638, {{173570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56639, {{173571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56640, {{173572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56641, {{173573, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56642, {{173574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56643, {{173575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56644, {{173578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56645, {{173579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56646, {{173581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56647, {{173582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56648, {{173583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56649, {{173584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56650, {{173585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56651, {{173586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56652, {{173587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56653, {{173588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56654, {{173589, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56655, {{173590, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56656, {{173591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56657, {{173592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56658, {{173594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56659, {{173595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56660, {{173596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56661, {{173597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56662, {{173599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56663, {{173600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56664, {{173601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56665, {{173602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56666, {{173603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56667, {{173604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56668, {{173605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56669, {{173606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56670, {{173607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56671, {{173608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56672, {{173609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56673, {{173610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56674, {{173612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56675, {{173613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56676, {{173614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56677, {{173616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56678, {{173617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56679, {{173618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56680, {{173619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56681, {{173620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56682, {{173622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56683, {{173623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56684, {{173624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56685, {{173625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56686, {{173626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56687, {{173628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56688, {{173629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56689, {{173630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56690, {{173631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56691, {{173632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56692, {{173633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56693, {{173634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56694, {{173635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56695, {{173637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56696, {{173638, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56697, {{173639, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56698, {{173640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56699, {{173641, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56700, {{173642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56701, {{173643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56702, {{173644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56703, {{173645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56704, {{173646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56705, {{173647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56706, {{173648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56707, {{173649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56708, {{173650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56709, {{173651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56710, {{173652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56711, {{173654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56712, {{173655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56713, {{173656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56714, {{173657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56715, {{173658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56716, {{173659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56717, {{173660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56718, {{173661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56719, {{173662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56720, {{173663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56721, {{173666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56722, {{173668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56723, {{173669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56724, {{173670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56725, {{173671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56726, {{173672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56727, {{173673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56728, {{173674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56729, {{173675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56730, {{173676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56731, {{173677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56732, {{173678, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56733, {{173679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56734, {{173680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56735, {{173682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56736, {{173684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56737, {{173685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56738, {{173686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56739, {{173687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56740, {{173688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56741, {{173689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56742, {{173690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56743, {{173691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56744, {{173692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56745, {{173693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56746, {{173694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56747, {{173696, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56748, {{173697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56749, {{173698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56750, {{173699, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56751, {{173700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56752, {{173703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56753, {{173704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56754, {{173705, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56755, {{173707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56756, {{173709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56757, {{173710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56758, {{173713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56760, {{173714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56761, {{173715, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{56762, {{173717, 1, 0, 0},{173717, 1, 1, 16}, {0, 0, 0, 0}}},{56762, {{173717, 1, 0, 0},{173717, 1, 1, 16}, {0, 0, 0, 0}}},{56763, {{173717, 1, 0, 0},{173717, 1, 1, 17}, {0, 0, 0, 0}}},{56763, {{173717, 1, 0, 0},{173717, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{56764, {{173718, 1, 0, 0},{173718, 1, 1, 16}, {0, 0, 0, 0}}},{56764, {{173718, 1, 0, 0},{173718, 1, 1, 16}, {0, 0, 0, 0}}},{56765, {{173718, 1, 0, 0},{173718, 1, 1, 17}, {0, 0, 0, 0}}},{56765, {{173718, 1, 0, 0},{173718, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{56766, {{173719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56767, {{173720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{56768, {{173721, 1, 0, 0},{173721, 1, 1, 16}, {0, 0, 0, 0}}},{56768, {{173721, 1, 0, 0},{173721, 1, 1, 16}, {0, 0, 0, 0}}},{56769, {{173721, 1, 0, 0},{173721, 1, 1, 17}, {0, 0, 0, 0}}},{56769, {{173721, 1, 0, 0},{173721, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56770, {{173724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56771, {{173725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56772, {{173726, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56773, {{173727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56774, {{173728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56775, {{173730, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56776, {{173733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56778, {{173734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56779, {{173735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56780, {{173736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56781, {{173739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56782, {{173740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56784, {{173741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56785, {{173742, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56786, {{173743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56787, {{173744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56788, {{173745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56789, {{173746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56790, {{173747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56791, {{173749, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56792, {{173750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56793, {{173752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56794, {{173753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56795, {{173754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56796, {{173755, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56797, {{173756, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56798, {{173757, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56799, {{173758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56800, {{173759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56801, {{173760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56802, {{173761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56803, {{173762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56804, {{173763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56805, {{173768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56806, {{173770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56807, {{173771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56808, {{173773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56809, {{173774, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56810, {{173775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56811, {{173776, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56812, {{173777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56813, {{173778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56814, {{173779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56815, {{173780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56816, {{173781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56817, {{173782, 1, 0, 0}, {0, 0, 0, 0}}}}}
};

static const MJMappingSet chunked_mj_mappings_02a716_02a88d[376] = {
    {1, {{57098, {{173846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56907, {{173865, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56908, {{173866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56915, {{173868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56913, {{173880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56922, {{173885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{56923, {{173894, 1, 0, 0},{173894, 1, 1, 16}, {0, 0, 0, 0}}},{56923, {{173894, 1, 0, 0},{173894, 1, 1, 16}, {0, 0, 0, 0}}},{56928, {{173894, 1, 0, 0},{173894, 1, 1, 17}, {0, 0, 0, 0}}},{56928, {{173894, 1, 0, 0},{173894, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59327, {{173906, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56940, {{173912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56939, {{173919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59358, {{173974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59361, {{173975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59360, {{173977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59369, {{173998, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57034, {{174005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59376, {{174020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57051, {{174023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57063, {{174039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58110, {{174055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57101, {{174071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59409, {{174088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57122, {{174104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57914, {{174130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57158, {{174149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57170, {{174176, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57173, {{174177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57172, {{174184, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57188, {{174201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57196, {{174210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57193, {{174218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59435, {{174221, 1, 0, 0}, {0, 0, 0, 0}}}}}
};

static const MJMappingSet chunked_mj_mappings_02a8e5_02a8e8[4] = {
    {1, {{57229, {{174309, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57227, {{174312, 1, 0, 0}, {0, 0, 0, 0}}}}}
};

static const MJMappingSet chunked_mj_mappings_02a944_02aa76[307] = {
    {1, {{59476, {{174404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58558, {{174435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57293, {{174446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57296, {{174450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57302, {{174461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57308, {{174474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57306, {{174479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57316, {{174502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57315, {{174504, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57317, {{174510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57319, {{174522, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59915, {{174559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{59509, {{174568, 1, 0, 0},{174568, 1, 1, 16}, {0, 0, 0, 0}}},{59509, {{174568, 1, 0, 0},{174568, 1, 1, 16}, {0, 0, 0, 0}}},{68099, {{174568, 1, 0, 0},{174568, 1, 1, 17}, {0, 0, 0, 0}}},{68099, {{174568, 1, 0, 0},{174568, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57067, {{174606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57373, {{174625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57379, {{174626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57381, {{174636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57413, {{174697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60004, {{174710, 1, 0, 0}, {0, 0, 0, 0}}}}}
};

static const MJMappingSet chunked_mj_mappings_02aaba_02ad21[616] = {
    {1, {{57461, {{174778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57465, {{174783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57473, {{174788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57494, {{174823, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57495, {{174828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57503, {{174840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59616, {{174851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57508, {{174853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57529, {{174911, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57533, {{174921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57555, {{174961, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57558, {{174967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{12395, {{174970, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57559, {{174972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57560, {{174983, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57572, {{175026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57574, {{175031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57584, {{175046, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57617, {{175065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57655, {{175122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59703, {{175126, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59685, {{175143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57671, {{175146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57677, {{175168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57692, {{175204, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58393, {{175208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58392, {{175209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57702, {{175217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57701, {{175218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57700, {{175219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57709, {{175222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57708, {{175225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57712, {{175227, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57722, {{175238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57723, {{175239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57735, {{175250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57736, {{175251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57733, {{175254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57748, {{175266, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57742, {{175270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59722, {{175271, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57744, {{175272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59725, {{175274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57757, {{175283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57764, {{175292, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57761, {{175298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57762, {{175299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57758, {{175302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57779, {{175305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57780, {{175306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59734, {{175311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57776, {{175314, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57808, {{175315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57781, {{175318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57793, {{175329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57797, {{175330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57805, {{175332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57802, {{175335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57825, {{175345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57822, {{175347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57821, {{175354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57834, {{175360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{59743, {{175363, 1, 0, 0},{175363, 1, 1, 16}, {0, 0, 0, 0}}},{59743, {{175363, 1, 0, 0},{175363, 1, 1, 16}, {0, 0, 0, 0}}},{59747, {{175363, 1, 0, 0},{175363, 1, 1, 17}, {0, 0, 0, 0}}},{59747, {{175363, 1, 0, 0},{175363, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57847, {{175366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57833, {{175367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57849, {{175374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57857, {{175376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57858, {{175379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57855, {{175381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57868, {{175392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57869, {{175393, 1, 0, 0}, {0, 0, 0, 0}}}}}
};

static const MJMappingSet chunked_mj_mappings_02ad89_02b117[911] = {
    {4, {{57949, {{175497, 1, 0, 0},{175497, 1, 1, 17}, {0, 0, 0, 0}}},{57949, {{175497, 1, 0, 0},{175497, 1, 1, 17}, {0, 0, 0, 0}}},{59779, {{175497, 1, 0, 0},{175497, 1, 1, 16}, {0, 0, 0, 0}}},{59779, {{175497, 1, 0, 0},{175497, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{59781, {{175498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{57522, {{175554, 1, 0, 0},{175554, 1, 1, 16}, {0, 0, 0, 0}}},{57522, {{175554, 1, 0, 0},{175554, 1, 1, 16}, {0, 0, 0, 0}}},{57969, {{175554, 1, 0, 0},{175554, 1, 1, 17}, {0, 0, 0, 0}}},{57969, {{175554, 1, 0, 0},{175554, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57967, {{175564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59835, {{175588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57986, {{175650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57993, {{175659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58004, {{175717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58008, {{175734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58013, {{175754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58018, {{175760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58025, {{175794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40788, {{175813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58035, {{175820, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58039, {{175827, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59896, {{175832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59907, {{175875, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58050, {{175889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58053, {{175914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58060, {{175926, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58063, {{175927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58062, {{175928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58064, {{175929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58067, {{175931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58069, {{175932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58086, {{175947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58100, {{175960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58111, {{175973, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58125, {{175994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58153, {{176023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58157, {{176039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58164, {{176040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58170, {{176044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{58172, {{176050, 1, 0, 0},{176050, 1, 1, 17}, {0, 0, 0, 0}}},{58172, {{176050, 1, 0, 0},{176050, 1, 1, 17}, {0, 0, 0, 0}}},{59939, {{176050, 1, 0, 0},{176050, 1, 1, 16}, {0, 0, 0, 0}}},{59939, {{176050, 1, 0, 0},{176050, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58178, {{176053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58182, {{176059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58186, {{176060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58185, {{176063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58194, {{176072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58830, {{176098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{58200, {{176131, 1, 0, 0},{176131, 1, 1, 16}, {0, 0, 0, 0}}},{58200, {{176131, 1, 0, 0},{176131, 1, 1, 16}, {0, 0, 0, 0}}},{58203, {{176131, 1, 0, 0},{176131, 1, 1, 17}, {0, 0, 0, 0}}},{58203, {{176131, 1, 0, 0},{176131, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59967, {{176143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58220, {{176167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59987, {{176183, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58224, {{176185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58228, {{176193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58235, {{176198, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{58236, {{176200, 1, 0, 0},{176200, 1, 1, 17}, {0, 0, 0, 0}}},{58236, {{176200, 1, 0, 0},{176200, 1, 1, 17}, {0, 0, 0, 0}}},{59999, {{176200, 1, 0, 0},{176200, 1, 1, 16}, {0, 0, 0, 0}}},{59999, {{176200, 1, 0, 0},{176200, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58243, {{176209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58246, {{176214, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58248, {{176215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58268, {{176261, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60019, {{176265, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58274, {{176274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58277, {{176291, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58279, {{176297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58285, {{176306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58293, {{176324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58295, {{176327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58299, {{176348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60052, {{176364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58311, {{176370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58321, {{176373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58328, {{176394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58332, {{176407, 1, 0, 0}, {0, 0, 0, 0}}}}}
};

static const MJMappingSet chunked_mj_mappings_02b168_02b5c6[1119] = {
    {1, {{60068, {{176488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58363, {{176511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58371, {{176531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58381, {{176539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58389, {{176552, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58397, {{176567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58404, {{176570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58435, {{176604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58436, {{176608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58440, {{176616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58442, {{176617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58457, {{176635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58471, {{176639, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58473, {{176642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58475, {{176644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58469, {{176646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58470, {{176647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58485, {{176659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58487, {{176670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58502, {{176676, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60127, {{176692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58516, {{176697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58510, {{176701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58517, {{176704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58514, {{176711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58530, {{176721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58527, {{176722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58529, {{176729, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58525, {{176732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58536, {{176736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58546, {{176748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58543, {{176752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58549, {{176763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58547, {{176768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58548, {{176772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58571, {{176792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58574, {{176796, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58585, {{176798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60153, {{176799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58597, {{176806, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58602, {{176807, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58600, {{176811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58607, {{176816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58611, {{176819, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58612, {{176826, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58618, {{176829, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58619, {{176830, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58628, {{176834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58624, {{176835, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58629, {{176837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58633, {{176838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58627, {{176841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58639, {{176843, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58640, {{176846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58642, {{176847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58648, {{176848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58643, {{176853, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58649, {{176854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58651, {{176856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58646, {{176858, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58656, {{176861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58660, {{176866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58662, {{176868, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58682, {{176879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58687, {{176885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58691, {{176892, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58699, {{176894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58698, {{176901, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58701, {{176904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58709, {{176915, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57484, {{176927, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58725, {{176933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58729, {{176935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58739, {{176951, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58747, {{176964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58754, {{176966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58759, {{176971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58768, {{176978, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58775, {{177025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60198, {{177031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58786, {{177045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58798, {{177070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58802, {{177088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58804, {{177090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58805, {{177092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58807, {{177097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58810, {{177103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58813, {{177108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60213, {{177110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58818, {{177121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58820, {{177129, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58823, {{177134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58838, {{177182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58847, {{177188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58860, {{177199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58875, {{177238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{60252, {{177256, 1, 0, 0},{177256, 1, 1, 16}, {0, 0, 0, 0}}},{60252, {{177256, 1, 0, 0},{177256, 1, 1, 16}, {0, 0, 0, 0}}},{60253, {{177256, 1, 0, 0},{177256, 1, 1, 17}, {0, 0, 0, 0}}},{60253, {{177256, 1, 0, 0},{177256, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58885, {{177269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58288, {{177280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58896, {{177281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58899, {{177286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58898, {{177289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58901, {{177294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58903, {{177298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58902, {{177299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58906, {{177302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58907, {{177306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58904, {{177308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58919, {{177318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58916, {{177322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58920, {{177327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58923, {{177330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58924, {{177332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{58928, {{177340, 1, 0, 0},{177340, 1, 1, 17}, {0, 0, 0, 0}}},{58928, {{177340, 1, 0, 0},{177340, 1, 1, 17}, {0, 0, 0, 0}}},{58929, {{177340, 1, 0, 0},{177340, 1, 1, 16}, {0, 0, 0, 0}}},{58929, {{177340, 1, 0, 0},{177340, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58934, {{177344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58936, {{177347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60300, {{177350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58939, {{177354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58941, {{177357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58946, {{177360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58944, {{177364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58950, {{177368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58949, {{177370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57411, {{177433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58960, {{177434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58961, {{177437, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58962, {{177438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60309, {{177439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58964, {{177450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60316, {{177477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58991, {{177503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58993, {{177507, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58998, {{177531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59014, {{177547, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59023, {{177565, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59026, {{177567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60338, {{177570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59034, {{177576, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59039, {{177606, 1, 0, 0}, {0, 0, 0, 0}}}}}
};

static const MJMappingSet chunked_mj_mappings_02b607_02b6d9[211] = {
    {1, {{59048, {{177671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59051, {{177675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59054, {{177685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59065, {{177721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59080, {{177736, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59084, {{177739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59091, {{177743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59098, {{177745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59097, {{177746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59101, {{177749, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59107, {{177750, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59114, {{177754, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59110, {{177756, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59109, {{177758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59112, {{177759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59118, {{177760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59121, {{177761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59131, {{177765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59132, {{177767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59137, {{177768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59139, {{177769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59141, {{177770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59145, {{177771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59148, {{177772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59144, {{177774, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59151, {{177775, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59166, {{177776, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59163, {{177777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59143, {{177778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59160, {{177779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59164, {{177781, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59159, {{177782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59169, {{177784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59176, {{177785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59173, {{177787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59175, {{177788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59177, {{177790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59182, {{177792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59187, {{177794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59189, {{177796, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59195, {{177797, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59200, {{177798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59202, {{177799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59998, {{177844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59207, {{177845, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59213, {{177851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59217, {{177854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59219, {{177856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59216, {{177860, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59222, {{177861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59229, {{177869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59236, {{177873, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59238, {{177874, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59235, {{177876, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59242, {{177878, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59244, {{177879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59246, {{177880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58255, {{177881, 1, 0, 0}, {0, 0, 0, 0}}}}}
};

static const MJMappingSet chunked_mj_mappings_02b71b_02b9e4[714] = {
    {1, {{59267, {{177947, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59278, {{177956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56880, {{177984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{56830, {{177985, 1, 0, 0},{177985, 1, 1, 17}, {0, 0, 0, 0}}},{56830, {{177985, 1, 0, 0},{177985, 1, 1, 17}, {0, 0, 0, 0}}},{59299, {{177985, 1, 0, 0},{177985, 1, 1, 16}, {0, 0, 0, 0}}},{59299, {{177985, 1, 0, 0},{177985, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {6, {{56835, {{177986, 1, 0, 0},{177986, 1, 1, 17}, {0, 0, 0, 0}}},{56835, {{177986, 1, 0, 0},{177986, 1, 1, 17}, {0, 0, 0, 0}}},{56836, {{177986, 1, 0, 0},{177986, 1, 1, 18}, {0, 0, 0, 0}}},{56836, {{177986, 1, 0, 0},{177986, 1, 1, 18}, {0, 0, 0, 0}}},{57546, {{177986, 1, 0, 0},{177986, 1, 1, 16}, {0, 0, 0, 0}}},{57546, {{177986, 1, 0, 0},{177986, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{58020, {{177987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56855, {{177988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56909, {{177989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{56904, {{177990, 1, 0, 0},{177990, 1, 1, 19}, {0, 0, 0, 0}}},{56904, {{177990, 1, 0, 0},{177990, 1, 1, 19}, {0, 0, 0, 0}}},{59309, {{177990, 1, 0, 0},{177990, 1, 1, 17}, {0, 0, 0, 0}}},{59309, {{177990, 1, 0, 0},{177990, 1, 1, 17}, {0, 0, 0, 0}}},{59340, {{177990, 1, 0, 0},{177990, 1, 1, 18}, {0, 0, 0, 0}}},{59340, {{177990, 1, 0, 0},{177990, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{57148, {{177991, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56929, {{177993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{7022, {{177994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{56984, {{177996, 1, 0, 0},{177996, 1, 1, 16}, {0, 0, 0, 0}}},{56984, {{177996, 1, 0, 0},{177996, 1, 1, 16}, {0, 0, 0, 0}}},{59345, {{177996, 1, 0, 0},{177996, 1, 1, 17}, {0, 0, 0, 0}}},{59345, {{177996, 1, 0, 0},{177996, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{57382, {{177997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57052, {{178000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{57073, {{178001, 1, 0, 0},{178001, 1, 1, 17}, {0, 0, 0, 0}}},{57073, {{178001, 1, 0, 0},{178001, 1, 1, 17}, {0, 0, 0, 0}}},{60385, {{178001, 1, 0, 0},{178001, 1, 1, 18}, {0, 0, 0, 0}}},{60385, {{178001, 1, 0, 0},{178001, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{57094, {{178002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59397, {{178003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57218, {{178004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{58834, {{178005, 1, 0, 0},{178005, 1, 1, 16}, {0, 0, 0, 0}}},{58834, {{178005, 1, 0, 0},{178005, 1, 1, 16}, {0, 0, 0, 0}}},{59406, {{178005, 1, 0, 0},{178005, 1, 1, 17}, {0, 0, 0, 0}}},{59406, {{178005, 1, 0, 0},{178005, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{57113, {{178006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57114, {{178007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{57200, {{178013, 1, 0, 0},{178013, 1, 1, 17}, {0, 0, 0, 0}}},{57200, {{178013, 1, 0, 0},{178013, 1, 1, 17}, {0, 0, 0, 0}}},{57394, {{178013, 1, 0, 0},{178013, 1, 1, 16}, {0, 0, 0, 0}}},{57394, {{178013, 1, 0, 0},{178013, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57233, {{178015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57240, {{178016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57242, {{178017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{57091, {{178018, 1, 0, 0},{178018, 1, 1, 16}, {0, 0, 0, 0}}},{57091, {{178018, 1, 0, 0},{178018, 1, 1, 16}, {0, 0, 0, 0}}},{57262, {{178018, 1, 0, 0},{178018, 1, 1, 18}, {0, 0, 0, 0}}},{57262, {{178018, 1, 0, 0},{178018, 1, 1, 18}, {0, 0, 0, 0}}},{58424, {{178018, 1, 0, 0},{178018, 1, 1, 17}, {0, 0, 0, 0}}},{58424, {{178018, 1, 0, 0},{178018, 1, 1, 17}, {0, 0, 0, 0}}},{59474, {{178018, 1, 0, 0},{178018, 1, 1, 20}, {0, 0, 0, 0}}},{59474, {{178018, 1, 0, 0},{178018, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {6, {{57264, {{178019, 1, 0, 0},{178019, 1, 1, 17}, {0, 0, 0, 0}}},{57264, {{178019, 1, 0, 0},{178019, 1, 1, 17}, {0, 0, 0, 0}}},{57266, {{178019, 1, 0, 0},{178019, 1, 1, 18}, {0, 0, 0, 0}}},{57266, {{178019, 1, 0, 0},{178019, 1, 1, 18}, {0, 0, 0, 0}}},{58425, {{178019, 1, 0, 0},{178019, 1, 1, 16}, {0, 0, 0, 0}}},{58425, {{178019, 1, 0, 0},{178019, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{57278, {{178020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{57326, {{178031, 1, 0, 0},{178031, 1, 1, 16}, {0, 0, 0, 0}}},{57326, {{178031, 1, 0, 0},{178031, 1, 1, 16}, {0, 0, 0, 0}}},{59501, {{178031, 1, 0, 0},{178031, 1, 1, 17}, {0, 0, 0, 0}}},{59501, {{178031, 1, 0, 0},{178031, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{57343, {{178032, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57345, {{178033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57367, {{178034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57380, {{178035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{57446, {{178038, 1, 0, 0},{178038, 1, 1, 17}, {0, 0, 0, 0}}},{57446, {{178038, 1, 0, 0},{178038, 1, 1, 17}, {0, 0, 0, 0}}},{59555, {{178038, 1, 0, 0},{178038, 1, 1, 18}, {0, 0, 0, 0}}},{59555, {{178038, 1, 0, 0},{178038, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {6, {{57445, {{178039, 1, 0, 0},{178039, 1, 1, 18}, {0, 0, 0, 0}}},{57445, {{178039, 1, 0, 0},{178039, 1, 1, 18}, {0, 0, 0, 0}}},{57447, {{178039, 1, 0, 0},{178039, 1, 1, 17}, {0, 0, 0, 0}}},{57447, {{178039, 1, 0, 0},{178039, 1, 1, 17}, {0, 0, 0, 0}}},{59556, {{178039, 1, 0, 0},{178039, 1, 1, 19}, {0, 0, 0, 0}}},{59556, {{178039, 1, 0, 0},{178039, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {4, {{57482, {{178040, 1, 0, 0},{178040, 1, 1, 17}, {0, 0, 0, 0}}},{57482, {{178040, 1, 0, 0},{178040, 1, 1, 17}, {0, 0, 0, 0}}},{59587, {{178040, 1, 0, 0},{178040, 1, 1, 16}, {0, 0, 0, 0}}},{59587, {{178040, 1, 0, 0},{178040, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{57496, {{178041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59646, {{178044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57577, {{178047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{57640, {{178048, 1, 0, 0},{178048, 1, 1, 16}, {0, 0, 0, 0}}},{57640, {{178048, 1, 0, 0},{178048, 1, 1, 16}, {0, 0, 0, 0}}},{57641, {{178048, 1, 0, 0},{178048, 1, 1, 17}, {0, 0, 0, 0}}},{57641, {{178048, 1, 0, 0},{178048, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{57644, {{178049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59679, {{178050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57657, {{178051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57664, {{178052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57688, {{178054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57689, {{178055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57710, {{178056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{57724, {{178057, 1, 0, 0},{178057, 1, 1, 17}, {0, 0, 0, 0}}},{57724, {{178057, 1, 0, 0},{178057, 1, 1, 17}, {0, 0, 0, 0}}},{59715, {{178057, 1, 0, 0},{178057, 1, 1, 18}, {0, 0, 0, 0}}},{59715, {{178057, 1, 0, 0},{178057, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{57725, {{178058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14410, {{178059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57783, {{178060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57813, {{178061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{57812, {{178062, 1, 0, 0},{178062, 1, 1, 17}, {0, 0, 0, 0}}},{57812, {{178062, 1, 0, 0},{178062, 1, 1, 17}, {0, 0, 0, 0}}},{59735, {{178062, 1, 0, 0},{178062, 1, 1, 18}, {0, 0, 0, 0}}},{59735, {{178062, 1, 0, 0},{178062, 1, 1, 18}, {0, 0, 0, 0}}},{59736, {{178062, 1, 0, 0},{178062, 1, 1, 21}, {0, 0, 0, 0}}},{59736, {{178062, 1, 0, 0},{178062, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {1, {{57811, {{178063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57830, {{178064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57839, {{178065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57851, {{178066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{57896, {{178067, 1, 0, 0},{178067, 1, 1, 16}, {0, 0, 0, 0}}},{57896, {{178067, 1, 0, 0},{178067, 1, 1, 16}, {0, 0, 0, 0}}},{59758, {{178067, 1, 0, 0},{178067, 1, 1, 17}, {0, 0, 0, 0}}},{59758, {{178067, 1, 0, 0},{178067, 1, 1, 17}, {0, 0, 0, 0}}},{59759, {{178067, 1, 0, 0},{178067, 1, 1, 18}, {0, 0, 0, 0}}},{59759, {{178067, 1, 0, 0},{178067, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57933, {{178069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57934, {{178070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57953, {{178072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57961, {{178073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{57968, {{178076, 1, 0, 0},{178076, 1, 1, 16}, {0, 0, 0, 0}}},{57968, {{178076, 1, 0, 0},{178076, 1, 1, 16}, {0, 0, 0, 0}}},{59829, {{178076, 1, 0, 0},{178076, 1, 1, 17}, {0, 0, 0, 0}}},{59829, {{178076, 1, 0, 0},{178076, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57977, {{178078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57990, {{178079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58031, {{178084, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58076, {{178090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58095, {{178091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59920, {{178092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58138, {{178094, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58142, {{178095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{18077, {{178096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57237, {{178097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58160, {{178098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58207, {{178100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58208, {{178101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58217, {{178102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58225, {{178104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{58240, {{178105, 1, 0, 0},{178105, 1, 1, 17}, {0, 0, 0, 0}}},{58240, {{178105, 1, 0, 0},{178105, 1, 1, 17}, {0, 0, 0, 0}}},{58241, {{178105, 1, 0, 0},{178105, 1, 1, 18}, {0, 0, 0, 0}}},{58241, {{178105, 1, 0, 0},{178105, 1, 1, 18}, {0, 0, 0, 0}}},{60003, {{178105, 1, 0, 0},{178105, 1, 1, 16}, {0, 0, 0, 0}}},{60003, {{178105, 1, 0, 0},{178105, 1, 1, 16}, {0, 0, 0, 0}}}}},
    {1, {{58254, {{178106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60026, {{178109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58271, {{178110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58272, {{178111, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58297, {{178112, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58318, {{178113, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{58345, {{178120, 1, 0, 0},{178120, 1, 1, 16}, {0, 0, 0, 0}}},{58345, {{178120, 1, 0, 0},{178120, 1, 1, 16}, {0, 0, 0, 0}}},{60062, {{178120, 1, 0, 0},{178120, 1, 1, 17}, {0, 0, 0, 0}}},{60062, {{178120, 1, 0, 0},{178120, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{60071, {{178121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{21335, {{178123, 1, 0, 0},{178123, 1, 1, 18}, {0, 0, 0, 0}}},{21335, {{178123, 1, 0, 0},{178123, 1, 1, 18}, {0, 0, 0, 0}}},{60089, {{178123, 1, 0, 0},{178123, 1, 1, 16}, {0, 0, 0, 0}}},{60089, {{178123, 1, 0, 0},{178123, 1, 1, 16}, {0, 0, 0, 0}}},{60090, {{178123, 1, 0, 0},{178123, 1, 1, 17}, {0, 0, 0, 0}}},{60090, {{178123, 1, 0, 0},{178123, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{58444, {{178124, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{58443, {{178125, 1, 0, 0},{178125, 1, 1, 16}, {0, 0, 0, 0}}},{58443, {{178125, 1, 0, 0},{178125, 1, 1, 16}, {0, 0, 0, 0}}},{60099, {{178125, 1, 0, 0},{178125, 1, 1, 17}, {0, 0, 0, 0}}},{60099, {{178125, 1, 0, 0},{178125, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{58448, {{178126, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{22123, {{178127, 1, 0, 0},{178127, 1, 1, 18}, {0, 0, 0, 0}}},{22123, {{178127, 1, 0, 0},{178127, 1, 1, 18}, {0, 0, 0, 0}}},{58460, {{178127, 1, 0, 0},{178127, 1, 1, 17}, {0, 0, 0, 0}}},{58460, {{178127, 1, 0, 0},{178127, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{58458, {{178128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{58491, {{178130, 1, 0, 0},{178130, 1, 1, 17}, {0, 0, 0, 0}}},{58491, {{178130, 1, 0, 0},{178130, 1, 1, 17}, {0, 0, 0, 0}}},{58495, {{178130, 1, 0, 0},{178130, 1, 1, 18}, {0, 0, 0, 0}}},{58495, {{178130, 1, 0, 0},{178130, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{58492, {{178131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58505, {{178132, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58532, {{178134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{58683, {{178136, 1, 0, 0},{178136, 1, 1, 17}, {0, 0, 0, 0}}},{58683, {{178136, 1, 0, 0},{178136, 1, 1, 17}, {0, 0, 0, 0}}},{60164, {{178136, 1, 0, 0},{178136, 1, 1, 18}, {0, 0, 0, 0}}},{60164, {{178136, 1, 0, 0},{178136, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{58684, {{178137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58697, {{178138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58714, {{178139, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58727, {{178140, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58771, {{178141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58853, {{178152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58855, {{178153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{58869, {{178154, 1, 0, 0},{178154, 1, 1, 18}, {0, 0, 0, 0}}},{58869, {{178154, 1, 0, 0},{178154, 1, 1, 18}, {0, 0, 0, 0}}},{60241, {{178154, 1, 0, 0},{178154, 1, 1, 16}, {0, 0, 0, 0}}},{60241, {{178154, 1, 0, 0},{178154, 1, 1, 16}, {0, 0, 0, 0}}},{60242, {{178154, 1, 0, 0},{178154, 1, 1, 17}, {0, 0, 0, 0}}},{60242, {{178154, 1, 0, 0},{178154, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58895, {{178159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58910, {{178160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58937, {{178161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58971, {{178179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60343, {{178185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59149, {{178189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59203, {{178195, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59210, {{178196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59228, {{178197, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59271, {{178200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60369, {{178202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59292, {{178205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56870, {{178213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56869, {{178276, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56884, {{178295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56910, {{178319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56935, {{178365, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56934, {{178369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56941, {{178372, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56946, {{178386, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56947, {{178389, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56957, {{178421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59337, {{178456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56976, {{178460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56975, {{178462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{59346, {{178480, 1, 0, 0},{178480, 1, 1, 16}, {0, 0, 0, 0}}},{59346, {{178480, 1, 0, 0},{178480, 1, 1, 16}, {0, 0, 0, 0}}},{59348, {{178480, 1, 0, 0},{178480, 1, 1, 17}, {0, 0, 0, 0}}},{59348, {{178480, 1, 0, 0},{178480, 1, 1, 17}, {0, 0, 0, 0}}},{59349, {{178480, 1, 0, 0},{178480, 1, 1, 18}, {0, 0, 0, 0}}},{59349, {{178480, 1, 0, 0},{178480, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57033, {{178528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57042, {{178546, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{57044, {{178555, 1, 0, 0},{178555, 1, 1, 16}, {0, 0, 0, 0}}},{57044, {{178555, 1, 0, 0},{178555, 1, 1, 16}, {0, 0, 0, 0}}},{59375, {{178555, 1, 0, 0},{178555, 1, 1, 17}, {0, 0, 0, 0}}},{59375, {{178555, 1, 0, 0},{178555, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57099, {{178568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57059, {{178596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57062, {{178597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57066, {{178603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59384, {{178609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58101, {{178614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57077, {{178617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57078, {{178623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59394, {{178637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {8, {{59399, {{21338, 1, 1, 26},{178660, 1, 0, 0},{178660, 1, 1, 16}, {0, 0, 0, 0}}},{59399, {{21338, 1, 1, 26},{178660, 1, 0, 0},{178660, 1, 1, 16}, {0, 0, 0, 0}}},{59400, {{21338, 1, 1, 27},{178660, 1, 0, 0},{178660, 1, 1, 17}, {0, 0, 0, 0}}},{59400, {{21338, 1, 1, 27},{178660, 1, 0, 0},{178660, 1, 1, 17}, {0, 0, 0, 0}}},{59401, {{178660, 1, 0, 0},{178660, 1, 1, 18}, {0, 0, 0, 0}}},{59401, {{178660, 1, 0, 0},{178660, 1, 1, 18}, {0, 0, 0, 0}}},{59402, {{178660, 1, 0, 0},{178660, 1, 1, 19}, {0, 0, 0, 0}}},{59402, {{178660, 1, 0, 0},{178660, 1, 1, 19}, {0, 0, 0, 0}}}}}
};

static const MJMappingSet chunked_mj_mappings_02ba2d_02bb07[219] = {
    {1, {{57131, {{178733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57915, {{178747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57150, {{178791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57152, {{178798, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59420, {{178834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57163, {{178843, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57162, {{178852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59421, {{178861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59424, {{178892, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57184, {{178893, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57187, {{178912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57194, {{178925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57195, {{178929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59438, {{178951, 1, 0, 0}, {0, 0, 0, 0}}}}}
};

static const MJMappingSet chunked_mj_mappings_02bb4f_02bcd0[386] = {
    {1, {{57206, {{179023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57208, {{179031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57221, {{179057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57230, {{179077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59453, {{179085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57231, {{179086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57241, {{179108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57243, {{179117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57247, {{179130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57248, {{179134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{59482, {{179185, 1, 0, 0},{179185, 1, 1, 16}, {0, 0, 0, 0}}},{59482, {{179185, 1, 0, 0},{179185, 1, 1, 16}, {0, 0, 0, 0}}},{59483, {{179185, 1, 0, 0},{179185, 1, 1, 17}, {0, 0, 0, 0}}},{59483, {{179185, 1, 0, 0},{179185, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57521, {{179192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57814, {{179193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58829, {{179200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57289, {{179201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57292, {{179219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57294, {{179226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57297, {{179231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57305, {{179272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57304, {{179273, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57307, {{179275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57311, {{179299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57312, {{179303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57314, {{179323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57321, {{179367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58489, {{179369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57324, {{179374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57327, {{179381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57333, {{179405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59506, {{179408, 1, 0, 0}, {0, 0, 0, 0}}}}}
};

static const MJMappingSet chunked_mj_mappings_02bd34_02bdd2[159] = {
    {1, {{57966, {{179508, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57368, {{179543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57372, {{179561, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57376, {{179572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57385, {{179581, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57392, {{179609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59534, {{179613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58999, {{179616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57396, {{179623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59549, {{179658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57428, {{179665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57429, {{179666, 1, 0, 0}, {0, 0, 0, 0}}}}}
};

static const MJMappingSet chunked_mj_mappings_02be28_02c03b[532] = {
    {1, {{57462, {{179752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57471, {{179774, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57486, {{179795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59029, {{179815, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59030, {{179816, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57502, {{179855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59613, {{179860, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57509, {{179863, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57510, {{179885, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{9310, {{179907, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{57527, {{179931, 1, 0, 0},{179931, 1, 1, 16}, {0, 0, 0, 0}}},{57527, {{179931, 1, 0, 0},{179931, 1, 1, 16}, {0, 0, 0, 0}}},{57530, {{179931, 1, 0, 0},{179931, 1, 1, 17}, {0, 0, 0, 0}}},{57530, {{179931, 1, 0, 0},{179931, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57532, {{179941, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57534, {{179943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57549, {{179979, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57551, {{179987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57557, {{180019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57565, {{180074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57569, {{180076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57571, {{180101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57576, {{180114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57587, {{180141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59655, {{180153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57597, {{180155, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57622, {{180205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57628, {{180222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57645, {{180247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57651, {{180263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57656, {{180268, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59688, {{180283, 1, 0, 0}, {0, 0, 0, 0}}}}}
};

static const MJMappingSet chunked_mj_mappings_02c083_02c27a[504] = {
    {1, {{57687, {{180355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57694, {{180375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56819, {{180379, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57696, {{180380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57699, {{180383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57705, {{180384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57706, {{180385, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57719, {{180396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57731, {{180403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57734, {{180405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57732, {{180406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57745, {{180415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57743, {{180425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57750, {{180431, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57749, {{180435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57760, {{180439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57759, {{180445, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57763, {{180450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57765, {{180451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57772, {{180458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57774, {{180464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57777, {{180469, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57775, {{180470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57787, {{180471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57778, {{180474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{57767, {{180478, 1, 0, 0},{180478, 1, 1, 16}, {0, 0, 0, 0}}},{57767, {{180478, 1, 0, 0},{180478, 1, 1, 16}, {0, 0, 0, 0}}},{57782, {{180478, 1, 0, 0},{180478, 1, 1, 17}, {0, 0, 0, 0}}},{57782, {{180478, 1, 0, 0},{180478, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57795, {{180489, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57800, {{180491, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57806, {{180492, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57796, {{180494, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57798, {{180498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{15903, {{180501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57803, {{180502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57804, {{180503, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57820, {{180506, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57823, {{180511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57826, {{180512, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59745, {{180515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57827, {{180517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57818, {{180519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57828, {{180523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57835, {{180530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57837, {{180534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57850, {{180536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57843, {{180538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57848, {{180539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57844, {{180540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57860, {{180549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57859, {{180551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57862, {{180554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57866, {{180555, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57865, {{180558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57867, {{180562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{57871, {{180565, 1, 0, 0},{180565, 1, 1, 16}, {0, 0, 0, 0}}},{57871, {{180565, 1, 0, 0},{180565, 1, 1, 16}, {0, 0, 0, 0}}},{57872, {{180565, 1, 0, 0},{180565, 1, 1, 17}, {0, 0, 0, 0}}},{57872, {{180565, 1, 0, 0},{180565, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57874, {{180567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58369, {{180621, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38786, {{180628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57906, {{180629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57910, {{180635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59768, {{180658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56871, {{180688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59772, {{180692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59771, {{180693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57945, {{180700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57958, {{180743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57959, {{180753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59803, {{180760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59808, {{180780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57965, {{180789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57973, {{180824, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57984, {{180857, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57983, {{180858, 1, 0, 0}, {0, 0, 0, 0}}}}}
};

static const MJMappingSet chunked_mj_mappings_02c2c8_02c608[833] = {
    {1, {{14469, {{180936, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58009, {{180942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58014, {{180960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58017, {{180972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59883, {{181023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56857, {{181033, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59886, {{181042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58032, {{181052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58038, {{181076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58037, {{181078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58044, {{181103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59903, {{181107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58059, {{181156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58061, {{181158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41198, {{181160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58065, {{181163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58068, {{181166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58070, {{181167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58073, {{181168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58079, {{181172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59914, {{181182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58098, {{181192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58102, {{181200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58118, {{181219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58121, {{181238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58124, {{181244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58146, {{181284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58156, {{181320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58158, {{181326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58165, {{181328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58171, {{181334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58174, {{181340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58173, {{181341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58177, {{181346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58184, {{181352, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58189, {{181356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58195, {{181373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58196, {{181381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59950, {{181393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58199, {{181416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58202, {{181421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58204, {{181426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59957, {{181427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58211, {{181441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58210, {{181444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59969, {{181446, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59970, {{181447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58216, {{181462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58221, {{181478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59988, {{181484, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59989, {{181490, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58226, {{181491, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58227, {{181497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58234, {{181504, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58245, {{181518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58247, {{181519, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58256, {{181544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58257, {{181548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58260, {{181552, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58262, {{181560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58263, {{181569, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58265, {{181579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58269, {{181582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58273, {{181607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58276, {{181614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58278, {{181622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58282, {{181639, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58286, {{181650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58294, {{181667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60038, {{181680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58298, {{181691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58306, {{181718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60048, {{181721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60047, {{181722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58317, {{181740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58329, {{181768, 1, 0, 0}, {0, 0, 0, 0}}}}}
};

static const MJMappingSet chunked_mj_mappings_02c670_02c8d7[616] = {
    {1, {{58347, {{181872, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58366, {{181917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57395, {{181929, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{45551, {{181945, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60080, {{181949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58386, {{181954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58396, {{181964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58395, {{181971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58401, {{181980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58423, {{182027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58428, {{182034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58432, {{182038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58434, {{182039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58441, {{182051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58453, {{182057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58452, {{182061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58456, {{182062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58468, {{182064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58466, {{182065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58477, {{182066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58474, {{182068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{58461, {{182071, 1, 0, 0},{182071, 1, 1, 16}, {0, 0, 0, 0}}},{58461, {{182071, 1, 0, 0},{182071, 1, 1, 16}, {0, 0, 0, 0}}},{58478, {{182071, 1, 0, 0},{182071, 1, 1, 17}, {0, 0, 0, 0}}},{58478, {{182071, 1, 0, 0},{182071, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {8, {{58476, {{182075, 1, 0, 0},{182075, 1, 1, 16}, {0, 0, 0, 0}}},{58476, {{182075, 1, 0, 0},{182075, 1, 1, 16}, {0, 0, 0, 0}}},{58479, {{182075, 1, 0, 0},{182075, 1, 1, 17}, {0, 0, 0, 0}}},{58479, {{182075, 1, 0, 0},{182075, 1, 1, 17}, {0, 0, 0, 0}}},{60112, {{182075, 1, 0, 0},{182075, 1, 1, 18}, {0, 0, 0, 0}}},{60112, {{182075, 1, 0, 0},{182075, 1, 1, 18}, {0, 0, 0, 0}}},{60113, {{182075, 1, 0, 0},{182075, 1, 1, 19}, {0, 0, 0, 0}}},{60113, {{182075, 1, 0, 0},{182075, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{58484, {{182076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58488, {{182082, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58503, {{182093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58500, {{182100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58501, {{182102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60125, {{182104, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58511, {{182109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58512, {{182110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58509, {{182117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58515, {{182118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58513, {{182121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58508, {{182125, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58522, {{182139, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58524, {{182143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58520, {{182153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58521, {{182157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58535, {{182160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58528, {{182164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58537, {{182169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58544, {{182177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60136, {{182182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58556, {{182200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58551, {{182201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60139, {{182204, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58555, {{182210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58553, {{182219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58565, {{182224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58564, {{182226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{23163, {{182227, 1, 0, 0},{182227, 1, 1, 16}, {0, 0, 0, 0}}},{23163, {{182227, 1, 0, 0},{182227, 1, 1, 16}, {0, 0, 0, 0}}},{23164, {{182227, 1, 0, 0},{182227, 1, 1, 17}, {0, 0, 0, 0}}},{23164, {{182227, 1, 0, 0},{182227, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60147, {{182232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58570, {{182242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58572, {{182246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58603, {{182262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58605, {{182263, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58604, {{182264, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58608, {{182272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58610, {{182274, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58620, {{182281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58617, {{182282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58616, {{182283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58621, {{182285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58631, {{182286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58626, {{182290, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58625, {{182296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58632, {{182297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58637, {{182301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58638, {{182302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58645, {{182303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58644, {{182308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58650, {{182311, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58652, {{182313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58647, {{182316, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58655, {{182317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58657, {{182318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58659, {{182319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58664, {{182323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58663, {{182324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58669, {{182331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58681, {{182336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58696, {{182363, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60170, {{182373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58702, {{182377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58720, {{182397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58721, {{182398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58723, {{182404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58728, {{182406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58734, {{182422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60189, {{182453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58748, {{182455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58755, {{182461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58756, {{182468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58767, {{182480, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58772, {{182485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58773, {{182486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58774, {{182487, 1, 0, 0}, {0, 0, 0, 0}}}}}
};

static const MJMappingSet chunked_mj_mappings_02c932_02c941[16] = {
    {1, {{58776, {{182578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58367, {{182593, 1, 0, 0}, {0, 0, 0, 0}}}}}
};

static const MJMappingSet chunked_mj_mappings_02c981_02ca56[214] = {
    {1, {{58797, {{182657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60209, {{182668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58801, {{182684, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58803, {{182694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58808, {{182701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58812, {{182711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58809, {{182714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58814, {{182725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58817, {{182739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60217, {{182752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58822, {{182759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58824, {{182764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58825, {{182765, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58826, {{182771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58839, {{182812, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58852, {{182831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58871, {{182870, 1, 0, 0}, {0, 0, 0, 0}}}}}
};

static const MJMappingSet chunked_mj_mappings_02caa6_02cb23[126] = {
    {1, {{58884, {{182950, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58886, {{182956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58894, {{182971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56885, {{182975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58900, {{182980, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60269, {{182982, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58908, {{182993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58912, {{182998, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58913, {{182999, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60271, {{183005, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58914, {{183008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58917, {{183013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58918, {{183015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58922, {{183021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58921, {{183024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58925, {{183025, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60294, {{183034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58930, {{183035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58931, {{183038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58933, {{183042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58943, {{183048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58942, {{183049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58938, {{183050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60303, {{183054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58951, {{183063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60308, {{183073, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58954, {{183074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58955, {{183075, 1, 0, 0}, {0, 0, 0, 0}}}}}
};

static const MJMappingSet chunked_mj_mappings_02cb85_02cc9d[281] = {
    {1, {{58957, {{183173, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58959, {{183174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58963, {{183185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58966, {{183190, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60310, {{183206, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58968, {{183233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58986, {{183282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58987, {{183283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60325, {{183285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58990, {{183288, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60326, {{183297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58996, {{183304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60329, {{183317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59010, {{183334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59012, {{183336, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59019, {{183340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59028, {{183371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59033, {{183375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59036, {{183412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59046, {{183453, 1, 0, 0}, {0, 0, 0, 0}}}}}
};

static const MJMappingSet chunked_mj_mappings_02ccef_02cd7e[144] = {
    {1, {{60347, {{183535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59055, {{183536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59066, {{183591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59069, {{183596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59081, {{183613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59082, {{183614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59090, {{183617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59092, {{183618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59094, {{183620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59096, {{183622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59103, {{183623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59105, {{183624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59102, {{183626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59104, {{183627, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59100, {{183628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59116, {{183629, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59113, {{183631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59111, {{183632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59115, {{183633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59123, {{183634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59119, {{183636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59122, {{183637, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59125, {{183638, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59130, {{183639, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59129, {{183642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59138, {{183644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59134, {{183645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59136, {{183646, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59140, {{183647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59146, {{183648, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59150, {{183649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59147, {{183650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59155, {{183651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59165, {{183654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59158, {{183655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59162, {{183656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59161, {{183657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59167, {{183658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{59156, {{183659, 1, 0, 0},{183659, 1, 1, 16}, {0, 0, 0, 0}}},{59156, {{183659, 1, 0, 0},{183659, 1, 1, 16}, {0, 0, 0, 0}}},{59171, {{183659, 1, 0, 0},{183659, 1, 1, 17}, {0, 0, 0, 0}}},{59171, {{183659, 1, 0, 0},{183659, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{59172, {{183660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59174, {{183661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59181, {{183663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59180, {{183664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59178, {{183665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59185, {{183666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59184, {{183667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59188, {{183668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59186, {{183669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59190, {{183670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59191, {{183671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59193, {{183673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59196, {{183674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59197, {{183675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59199, {{183677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59201, {{183678, 1, 0, 0}, {0, 0, 0, 0}}}}}
};

static const MJMappingSet chunked_mj_mappings_02cdbf_02cdfa[60] = {
    {1, {{59204, {{183743, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59205, {{183748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59208, {{183753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59212, {{183762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59218, {{183764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59214, {{183767, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59220, {{183769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59223, {{183773, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59221, {{183777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59225, {{183778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59227, {{183782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59226, {{183783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59230, {{183787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59231, {{183788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59234, {{183792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59239, {{183793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59237, {{183795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59241, {{183797, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59245, {{183799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59249, {{183802, 1, 0, 0}, {0, 0, 0, 0}}}}}
};

static const MJMappingSet chunked_mj_mappings_02ce3d_02ebe0[7588] = {
    {1, {{58368, {{183869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59283, {{183925, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59293, {{183984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56820, {{183985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56821, {{183987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56874, {{183989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56825, {{183990, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56875, {{183992, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56876, {{183994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56829, {{183995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59296, {{183997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{46311, {{184000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57104, {{184001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56828, {{184002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56832, {{184006, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59297, {{184007, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57474, {{184008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58576, {{184009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56833, {{184011, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{59298, {{184014, 1, 0, 0},{184014, 1, 1, 16}, {0, 0, 0, 0}}},{59298, {{184014, 1, 0, 0},{184014, 1, 1, 16}, {0, 0, 0, 0}}},{59381, {{184014, 1, 0, 0},{184014, 1, 1, 17}, {0, 0, 0, 0}}},{59381, {{184014, 1, 0, 0},{184014, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{56838, {{184016, 1, 0, 0},{184016, 1, 1, 16}, {0, 0, 0, 0}}},{56838, {{184016, 1, 0, 0},{184016, 1, 1, 16}, {0, 0, 0, 0}}},{57545, {{184016, 1, 0, 0},{184016, 1, 1, 17}, {0, 0, 0, 0}}},{57545, {{184016, 1, 0, 0},{184016, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{57476, {{184017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56834, {{184018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58881, {{184021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57477, {{184024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59300, {{184027, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{56843, {{184028, 1, 0, 0},{184028, 1, 1, 16}, {0, 0, 0, 0}}},{56843, {{184028, 1, 0, 0},{184028, 1, 1, 16}, {0, 0, 0, 0}}},{56844, {{184028, 1, 0, 0},{184028, 1, 1, 17}, {0, 0, 0, 0}}},{56844, {{184028, 1, 0, 0},{184028, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{56842, {{184029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56846, {{184030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57479, {{184037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57638, {{184053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57639, {{184055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59301, {{184056, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56849, {{184059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56853, {{184063, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56854, {{184064, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56851, {{184065, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56850, {{184066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56852, {{184068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56856, {{184070, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56860, {{184072, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56863, {{184076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59073, {{184077, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56865, {{184086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56864, {{184087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{57888, {{184088, 1, 0, 0},{184088, 1, 1, 16}, {0, 0, 0, 0}}},{57888, {{184088, 1, 0, 0},{184088, 1, 1, 16}, {0, 0, 0, 0}}},{57889, {{184088, 1, 0, 0},{184088, 1, 1, 17}, {0, 0, 0, 0}}},{57889, {{184088, 1, 0, 0},{184088, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{56867, {{184089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56868, {{184090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{12, {{184092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{68051, {{184095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57893, {{184107, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56872, {{184120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56878, {{184122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56881, {{184125, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56882, {{184126, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56883, {{184127, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56887, {{184131, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56888, {{184134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56889, {{184135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56892, {{184136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56891, {{184137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56896, {{184138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56897, {{184139, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {8, {{56862, {{184140, 1, 0, 0},{184140, 1, 1, 16}, {0, 0, 0, 0}}},{56862, {{184140, 1, 0, 0},{184140, 1, 1, 16}, {0, 0, 0, 0}}},{56893, {{184140, 1, 0, 0},{184140, 1, 1, 17}, {0, 0, 0, 0}}},{56893, {{184140, 1, 0, 0},{184140, 1, 1, 17}, {0, 0, 0, 0}}},{56894, {{184140, 1, 0, 0},{184140, 1, 1, 18}, {0, 0, 0, 0}}},{56894, {{184140, 1, 0, 0},{184140, 1, 1, 18}, {0, 0, 0, 0}}},{56898, {{184140, 1, 0, 0},{184140, 1, 1, 19}, {0, 0, 0, 0}}},{56898, {{184140, 1, 0, 0},{184140, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {1, {{57386, {{184141, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56900, {{184149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56899, {{184150, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59305, {{184153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59304, {{184154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58981, {{184156, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59307, {{184157, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59308, {{184160, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56902, {{184161, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56906, {{184163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59311, {{184164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{6505, {{184166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59310, {{184167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59313, {{184169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59312, {{184171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59319, {{184174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59317, {{184182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59318, {{184183, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56916, {{184185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59320, {{184193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59322, {{184196, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56925, {{184202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59323, {{184206, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56924, {{184215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56927, {{184216, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59325, {{184218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56930, {{184232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56937, {{184238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59328, {{184249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59326, {{184251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56943, {{184255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56938, {{184258, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59331, {{184283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56950, {{184294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59334, {{184301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56952, {{184306, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59333, {{184310, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56956, {{184315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56955, {{184320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59336, {{184329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56961, {{184343, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56962, {{184344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56964, {{184345, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56963, {{184346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56965, {{184348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58058, {{184349, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56968, {{184351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{56969, {{184352, 1, 0, 0},{184352, 1, 1, 16}, {0, 0, 0, 0}}},{56969, {{184352, 1, 0, 0},{184352, 1, 1, 16}, {0, 0, 0, 0}}},{56970, {{184352, 1, 0, 0},{184352, 1, 1, 17}, {0, 0, 0, 0}}},{56970, {{184352, 1, 0, 0},{184352, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{56972, {{184353, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56971, {{184355, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56974, {{184356, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56973, {{184357, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{59338, {{184360, 1, 0, 0},{184360, 1, 1, 16}, {0, 0, 0, 0}}},{59338, {{184360, 1, 0, 0},{184360, 1, 1, 16}, {0, 0, 0, 0}}},{59339, {{184360, 1, 0, 0},{184360, 1, 1, 17}, {0, 0, 0, 0}}},{59339, {{184360, 1, 0, 0},{184360, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56977, {{184368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56978, {{184369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56980, {{184388, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56982, {{184390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58412, {{184391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{58413, {{184392, 1, 0, 0},{184392, 1, 1, 16}, {0, 0, 0, 0}}},{58413, {{184392, 1, 0, 0},{184392, 1, 1, 16}, {0, 0, 0, 0}}},{59341, {{184392, 1, 0, 0},{184392, 1, 1, 17}, {0, 0, 0, 0}}},{59341, {{184392, 1, 0, 0},{184392, 1, 1, 17}, {0, 0, 0, 0}}},{59342, {{184392, 1, 0, 0},{184392, 1, 1, 18}, {0, 0, 0, 0}}},{59342, {{184392, 1, 0, 0},{184392, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{59343, {{184393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56987, {{184400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56996, {{184404, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59347, {{184405, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59350, {{184407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57001, {{184416, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57005, {{184417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57008, {{184418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57012, {{184419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59353, {{184422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59352, {{184424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57011, {{184425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57013, {{184426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{57014, {{184427, 1, 0, 0},{184427, 1, 1, 16}, {0, 0, 0, 0}}},{57014, {{184427, 1, 0, 0},{184427, 1, 1, 16}, {0, 0, 0, 0}}},{57015, {{184427, 1, 0, 0},{184427, 1, 1, 17}, {0, 0, 0, 0}}},{57015, {{184427, 1, 0, 0},{184427, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57020, {{184429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{57017, {{184431, 1, 0, 0},{184431, 1, 1, 16}, {0, 0, 0, 0}}},{57017, {{184431, 1, 0, 0},{184431, 1, 1, 16}, {0, 0, 0, 0}}},{57018, {{184431, 1, 0, 0},{184431, 1, 1, 17}, {0, 0, 0, 0}}},{57018, {{184431, 1, 0, 0},{184431, 1, 1, 17}, {0, 0, 0, 0}}},{57019, {{184431, 1, 0, 0},{184431, 1, 1, 18}, {0, 0, 0, 0}}},{57019, {{184431, 1, 0, 0},{184431, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59354, {{184434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57023, {{184435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59355, {{184436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59356, {{184438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{57026, {{184439, 1, 0, 0},{184439, 1, 1, 16}, {0, 0, 0, 0}}},{57026, {{184439, 1, 0, 0},{184439, 1, 1, 16}, {0, 0, 0, 0}}},{57028, {{184439, 1, 0, 0},{184439, 1, 1, 17}, {0, 0, 0, 0}}},{57028, {{184439, 1, 0, 0},{184439, 1, 1, 17}, {0, 0, 0, 0}}},{57029, {{184439, 1, 0, 0},{184439, 1, 1, 18}, {0, 0, 0, 0}}},{57029, {{184439, 1, 0, 0},{184439, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57025, {{184442, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57030, {{184443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59357, {{184444, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57031, {{184449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57032, {{184453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59359, {{184456, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59362, {{184462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59363, {{184465, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59364, {{184467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59365, {{184468, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59366, {{184471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59367, {{184474, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59368, {{184476, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59371, {{184491, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58055, {{184492, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{56981, {{184495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{59372, {{184498, 1, 0, 0},{184498, 1, 1, 16}, {0, 0, 0, 0}}},{59372, {{184498, 1, 0, 0},{184498, 1, 1, 16}, {0, 0, 0, 0}}},{68097, {{184498, 1, 0, 0},{184498, 1, 1, 17}, {0, 0, 0, 0}}},{68097, {{184498, 1, 0, 0},{184498, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{59374, {{184499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59373, {{184500, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57039, {{184510, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{7458, {{184516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57041, {{184517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57040, {{184518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57043, {{184524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57047, {{184537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{57048, {{184538, 1, 0, 0},{184538, 1, 1, 16}, {0, 0, 0, 0}}},{57048, {{184538, 1, 0, 0},{184538, 1, 1, 16}, {0, 0, 0, 0}}},{57049, {{184538, 1, 0, 0},{184538, 1, 1, 17}, {0, 0, 0, 0}}},{57049, {{184538, 1, 0, 0},{184538, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57046, {{184545, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59377, {{184548, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57053, {{184549, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57050, {{184552, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57055, {{184557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59378, {{184560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58346, {{184561, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57054, {{184562, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59379, {{184575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57056, {{184591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57058, {{184600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57060, {{184606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57065, {{184611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59382, {{184613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57069, {{184615, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57068, {{184619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59385, {{184627, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57071, {{184628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57072, {{184633, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57076, {{184639, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59391, {{184640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59390, {{184641, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57074, {{184642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57075, {{184643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57079, {{184645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57084, {{184650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59392, {{184653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57085, {{184656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57086, {{184657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57087, {{184658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59393, {{184660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57088, {{184664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57089, {{184669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59395, {{184671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59396, {{184672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57096, {{184673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{57102, {{184681, 1, 0, 0},{184681, 1, 1, 16}, {0, 0, 0, 0}}},{57102, {{184681, 1, 0, 0},{184681, 1, 1, 16}, {0, 0, 0, 0}}},{57103, {{184681, 1, 0, 0},{184681, 1, 1, 17}, {0, 0, 0, 0}}},{57103, {{184681, 1, 0, 0},{184681, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58577, {{184687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58579, {{184688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57105, {{184690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57107, {{184691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57106, {{184692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57109, {{184693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59403, {{184697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57110, {{184698, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59404, {{184700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57111, {{184701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59405, {{184705, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{68050, {{184706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59407, {{184707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59408, {{184708, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{57115, {{184710, 1, 0, 0},{184710, 1, 1, 16}, {0, 0, 0, 0}}},{57115, {{184710, 1, 0, 0},{184710, 1, 1, 16}, {0, 0, 0, 0}}},{57116, {{184710, 1, 0, 0},{184710, 1, 1, 17}, {0, 0, 0, 0}}},{57116, {{184710, 1, 0, 0},{184710, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{57117, {{184711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57118, {{184713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{57120, {{184719, 1, 0, 0},{184719, 1, 1, 16}, {0, 0, 0, 0}}},{57120, {{184719, 1, 0, 0},{184719, 1, 1, 16}, {0, 0, 0, 0}}},{57121, {{184719, 1, 0, 0},{184719, 1, 1, 17}, {0, 0, 0, 0}}},{57121, {{184719, 1, 0, 0},{184719, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{59410, {{184720, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58836, {{184722, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59411, {{184724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{57126, {{184729, 1, 0, 0},{184729, 1, 1, 16}, {0, 0, 0, 0}}},{57126, {{184729, 1, 0, 0},{184729, 1, 1, 16}, {0, 0, 0, 0}}},{57127, {{184729, 1, 0, 0},{184729, 1, 1, 17}, {0, 0, 0, 0}}},{57127, {{184729, 1, 0, 0},{184729, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57129, {{184734, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57128, {{184737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{56986, {{184738, 1, 0, 0},{184738, 1, 1, 16}, {0, 0, 0, 0}}},{56986, {{184738, 1, 0, 0},{184738, 1, 1, 16}, {0, 0, 0, 0}}},{56988, {{184738, 1, 0, 0},{184738, 1, 1, 17}, {0, 0, 0, 0}}},{56988, {{184738, 1, 0, 0},{184738, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57132, {{184740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57133, {{184741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57136, {{184746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{57137, {{184747, 1, 0, 0},{184747, 1, 1, 16}, {0, 0, 0, 0}}},{57137, {{184747, 1, 0, 0},{184747, 1, 1, 16}, {0, 0, 0, 0}}},{58030, {{184747, 1, 0, 0},{184747, 1, 1, 17}, {0, 0, 0, 0}}},{58030, {{184747, 1, 0, 0},{184747, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{57138, {{184748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57141, {{184751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57142, {{184756, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57143, {{184758, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59412, {{184759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57145, {{184769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59414, {{184774, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59413, {{184778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59415, {{184785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59418, {{184790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59416, {{184791, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57156, {{184802, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57155, {{184805, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57157, {{184810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57159, {{184838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59419, {{184861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57168, {{184876, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{57174, {{184889, 1, 0, 0},{184889, 1, 1, 16}, {0, 0, 0, 0}}},{57174, {{184889, 1, 0, 0},{184889, 1, 1, 16}, {0, 0, 0, 0}}},{57183, {{184889, 1, 0, 0},{184889, 1, 1, 17}, {0, 0, 0, 0}}},{57183, {{184889, 1, 0, 0},{184889, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57177, {{184892, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59422, {{184903, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57182, {{184904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57179, {{184920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57171, {{184921, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {8, {{59425, {{184928, 1, 0, 0},{184928, 1, 1, 16}, {0, 0, 0, 0}}},{59425, {{184928, 1, 0, 0},{184928, 1, 1, 16}, {0, 0, 0, 0}}},{59426, {{184928, 1, 0, 0},{184928, 1, 1, 17}, {0, 0, 0, 0}}},{59426, {{184928, 1, 0, 0},{184928, 1, 1, 17}, {0, 0, 0, 0}}},{59430, {{184928, 1, 0, 0},{184928, 1, 1, 18}, {0, 0, 0, 0}}},{59430, {{184928, 1, 0, 0},{184928, 1, 1, 18}, {0, 0, 0, 0}}},{59464, {{184928, 1, 0, 0},{184928, 1, 1, 19}, {0, 0, 0, 0}}},{59464, {{184928, 1, 0, 0},{184928, 1, 1, 19}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59429, {{184930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59427, {{184937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57186, {{184949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57192, {{184959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59432, {{184969, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59428, {{184970, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57190, {{184972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59431, {{184975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59433, {{185030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57199, {{185048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{57197, {{185051, 1, 0, 0},{185051, 1, 1, 16}, {0, 0, 0, 0}}},{57197, {{185051, 1, 0, 0},{185051, 1, 1, 16}, {0, 0, 0, 0}}},{57198, {{185051, 1, 0, 0},{185051, 1, 1, 17}, {0, 0, 0, 0}}},{57198, {{185051, 1, 0, 0},{185051, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57202, {{185100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57201, {{185108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32733, {{185126, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59440, {{185136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59442, {{185167, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{32858, {{185182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57160, {{185184, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57205, {{185189, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59443, {{185190, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59439, {{185206, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59446, {{185211, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57210, {{185212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{59445, {{185216, 1, 0, 0},{185216, 1, 1, 16}, {0, 0, 0, 0}}},{59445, {{185216, 1, 0, 0},{185216, 1, 1, 16}, {0, 0, 0, 0}}},{60381, {{185216, 1, 0, 0},{185216, 1, 1, 17}, {0, 0, 0, 0}}},{60381, {{185216, 1, 0, 0},{185216, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59447, {{185221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59351, {{185224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10952, {{185230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57220, {{185237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57219, {{185241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59448, {{185243, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59451, {{185244, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59449, {{185250, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59450, {{185252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59452, {{185262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57232, {{185280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59455, {{185293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59456, {{185296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57236, {{185301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{57238, {{185307, 1, 0, 0},{185307, 1, 1, 16}, {0, 0, 0, 0}}},{57238, {{185307, 1, 0, 0},{185307, 1, 1, 16}, {0, 0, 0, 0}}},{57239, {{185307, 1, 0, 0},{185307, 1, 1, 17}, {0, 0, 0, 0}}},{57239, {{185307, 1, 0, 0},{185307, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59458, {{185342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57246, {{185344, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59461, {{185354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59462, {{185362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57483, {{185365, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57251, {{185366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59465, {{185369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58129, {{185373, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59470, {{185376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59471, {{185378, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{57254, {{185379, 1, 0, 0},{185379, 1, 1, 16}, {0, 0, 0, 0}}},{57254, {{185379, 1, 0, 0},{185379, 1, 1, 16}, {0, 0, 0, 0}}},{57257, {{185379, 1, 0, 0},{185379, 1, 1, 17}, {0, 0, 0, 0}}},{57257, {{185379, 1, 0, 0},{185379, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59472, {{185381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59467, {{185382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59473, {{185383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59628, {{185387, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57259, {{185388, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59468, {{185390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59469, {{185392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57260, {{185393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{57265, {{185395, 1, 0, 0},{185395, 1, 1, 16}, {0, 0, 0, 0}}},{57265, {{185395, 1, 0, 0},{185395, 1, 1, 16}, {0, 0, 0, 0}}},{57267, {{185395, 1, 0, 0},{185395, 1, 1, 17}, {0, 0, 0, 0}}},{57267, {{185395, 1, 0, 0},{185395, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{57263, {{185396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57268, {{185400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59475, {{185410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57272, {{185411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{57274, {{185413, 1, 0, 0},{185413, 1, 1, 16}, {0, 0, 0, 0}}},{57274, {{185413, 1, 0, 0},{185413, 1, 1, 16}, {0, 0, 0, 0}}},{57276, {{185413, 1, 0, 0},{185413, 1, 1, 17}, {0, 0, 0, 0}}},{57276, {{185413, 1, 0, 0},{185413, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57275, {{185418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59477, {{185419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59478, {{185421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57281, {{185423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57279, {{185425, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57280, {{185426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{57283, {{185429, 1, 0, 0},{185429, 1, 1, 16}, {0, 0, 0, 0}}},{57283, {{185429, 1, 0, 0},{185429, 1, 1, 16}, {0, 0, 0, 0}}},{57284, {{185429, 1, 0, 0},{185429, 1, 1, 17}, {0, 0, 0, 0}}},{57284, {{185429, 1, 0, 0},{185429, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{59479, {{185430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57286, {{185432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59480, {{185434, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59481, {{185436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59484, {{185438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57287, {{185439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57288, {{185443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57290, {{185448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59486, {{185450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59487, {{185454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59488, {{185455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57295, {{185461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57298, {{185470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57299, {{185471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59489, {{185472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59490, {{185489, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59491, {{185492, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59492, {{185495, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59493, {{185502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57313, {{185505, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57318, {{185517, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59494, {{185528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59495, {{185532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59496, {{185533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57320, {{185538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59497, {{185543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59498, {{185544, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57323, {{185556, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59500, {{185557, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57325, {{185559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59499, {{185560, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59502, {{185574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{57329, {{185577, 1, 0, 0},{185577, 1, 1, 16}, {0, 0, 0, 0}}},{57329, {{185577, 1, 0, 0},{185577, 1, 1, 16}, {0, 0, 0, 0}}},{57331, {{185577, 1, 0, 0},{185577, 1, 1, 17}, {0, 0, 0, 0}}},{57331, {{185577, 1, 0, 0},{185577, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57332, {{185583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{59505, {{185585, 1, 0, 0},{185585, 1, 1, 16}, {0, 0, 0, 0}}},{59505, {{185585, 1, 0, 0},{185585, 1, 1, 16}, {0, 0, 0, 0}}},{68098, {{185585, 1, 0, 0},{185585, 1, 1, 17}, {0, 0, 0, 0}}},{68098, {{185585, 1, 0, 0},{185585, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59503, {{185587, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59507, {{185595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{67952, {{185597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57335, {{185598, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57338, {{185599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57342, {{185600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57340, {{185603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57341, {{185604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57339, {{185605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59268, {{185612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57344, {{185622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59510, {{185623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59511, {{185631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57346, {{185644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57348, {{185649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57347, {{185650, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57351, {{185653, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59515, {{185654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59516, {{185655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57354, {{185656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57353, {{185657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57360, {{185662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57359, {{185663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59517, {{185668, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57361, {{185671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57362, {{185673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57363, {{185675, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{34243, {{185677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57364, {{185680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57365, {{185682, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57369, {{185688, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59518, {{185689, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57370, {{185693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59519, {{185701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57371, {{185709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59520, {{185711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57374, {{185726, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57375, {{185727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57377, {{185728, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59522, {{185731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57378, {{185732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59524, {{185740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57211, {{185746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59525, {{185747, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59528, {{185769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59526, {{185771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59527, {{185772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57387, {{185777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59529, {{185795, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59530, {{185799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57389, {{185810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57390, {{185811, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59533, {{185812, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{59531, {{185813, 1, 0, 0},{185813, 1, 1, 16}, {0, 0, 0, 0}}},{59531, {{185813, 1, 0, 0},{185813, 1, 1, 16}, {0, 0, 0, 0}}},{59532, {{185813, 1, 0, 0},{185813, 1, 1, 17}, {0, 0, 0, 0}}},{59532, {{185813, 1, 0, 0},{185813, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57391, {{185822, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59535, {{185834, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59536, {{185837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59538, {{185846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57399, {{185851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57400, {{185852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57401, {{185854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57402, {{185855, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57403, {{185856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57407, {{185862, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57412, {{185865, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57414, {{185866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{10793, {{185869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59543, {{185873, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59542, {{185874, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59546, {{185880, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59545, {{185882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57419, {{185886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57425, {{185891, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57422, {{185895, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59547, {{185897, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59548, {{185899, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57427, {{185908, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{57430, {{185918, 1, 0, 0},{185918, 1, 1, 16}, {0, 0, 0, 0}}},{57430, {{185918, 1, 0, 0},{185918, 1, 1, 16}, {0, 0, 0, 0}}},{57431, {{185918, 1, 0, 0},{185918, 1, 1, 17}, {0, 0, 0, 0}}},{57431, {{185918, 1, 0, 0},{185918, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{57433, {{185932, 1, 0, 0},{185932, 1, 1, 16}, {0, 0, 0, 0}}},{57433, {{185932, 1, 0, 0},{185932, 1, 1, 16}, {0, 0, 0, 0}}},{57434, {{185932, 1, 0, 0},{185932, 1, 1, 17}, {0, 0, 0, 0}}},{57434, {{185932, 1, 0, 0},{185932, 1, 1, 17}, {0, 0, 0, 0}}},{57435, {{185932, 1, 0, 0},{185932, 1, 1, 18}, {0, 0, 0, 0}}},{57435, {{185932, 1, 0, 0},{185932, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{59551, {{185933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59552, {{185938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59553, {{185940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57438, {{185943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59554, {{185953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57444, {{185955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59557, {{185959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59559, {{185967, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59560, {{185969, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59562, {{185974, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59561, {{185976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57451, {{185978, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57450, {{185979, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59565, {{185981, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59563, {{185984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59564, {{185986, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57452, {{185987, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59566, {{186001, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57454, {{186002, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57453, {{186003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{57456, {{186008, 1, 0, 0},{186008, 1, 1, 16}, {0, 0, 0, 0}}},{57456, {{186008, 1, 0, 0},{186008, 1, 1, 16}, {0, 0, 0, 0}}},{57457, {{186008, 1, 0, 0},{186008, 1, 1, 17}, {0, 0, 0, 0}}},{57457, {{186008, 1, 0, 0},{186008, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59567, {{186015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57458, {{186017, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59568, {{186019, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57460, {{186021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57463, {{186024, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59570, {{186026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59571, {{186028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59572, {{186030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59573, {{186036, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57464, {{186037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {12, {{57466, {{186038, 1, 0, 0},{186038, 1, 1, 16}, {0, 0, 0, 0}}},{57466, {{186038, 1, 0, 0},{186038, 1, 1, 16}, {0, 0, 0, 0}}},{59574, {{186038, 1, 0, 0},{186038, 1, 1, 17}, {0, 0, 0, 0}}},{59574, {{186038, 1, 0, 0},{186038, 1, 1, 17}, {0, 0, 0, 0}}},{59575, {{186038, 1, 0, 0},{186038, 1, 1, 18}, {0, 0, 0, 0}}},{59575, {{186038, 1, 0, 0},{186038, 1, 1, 18}, {0, 0, 0, 0}}},{59576, {{186038, 1, 0, 0},{186038, 1, 1, 19}, {0, 0, 0, 0}}},{59576, {{186038, 1, 0, 0},{186038, 1, 1, 19}, {0, 0, 0, 0}}},{59577, {{186038, 1, 0, 0},{186038, 1, 1, 20}, {0, 0, 0, 0}}},{59577, {{186038, 1, 0, 0},{186038, 1, 1, 20}, {0, 0, 0, 0}}},{59578, {{186038, 1, 0, 0},{186038, 1, 1, 21}, {0, 0, 0, 0}}},{59578, {{186038, 1, 0, 0},{186038, 1, 1, 21}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57467, {{186040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57468, {{186041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59579, {{186042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57469, {{186044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59580, {{186047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57470, {{186049, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59581, {{186050, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59582, {{186053, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57472, {{186059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59584, {{186060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59585, {{186061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{59588, {{186077, 1, 0, 0},{186077, 1, 1, 16}, {0, 0, 0, 0}}},{59588, {{186077, 1, 0, 0},{186077, 1, 1, 16}, {0, 0, 0, 0}}},{59589, {{186077, 1, 0, 0},{186077, 1, 1, 17}, {0, 0, 0, 0}}},{59589, {{186077, 1, 0, 0},{186077, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{59590, {{186078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57485, {{186080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59591, {{186083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59592, {{186087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57487, {{186089, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59593, {{186090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59594, {{186099, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59597, {{186114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59598, {{186120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59599, {{186122, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57492, {{186125, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57491, {{186126, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{59600, {{186128, 1, 0, 0},{186128, 1, 1, 16}, {0, 0, 0, 0}}},{59600, {{186128, 1, 0, 0},{186128, 1, 1, 16}, {0, 0, 0, 0}}},{59601, {{186128, 1, 0, 0},{186128, 1, 1, 17}, {0, 0, 0, 0}}},{59601, {{186128, 1, 0, 0},{186128, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{59602, {{186129, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59603, {{186136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59606, {{186137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59605, {{186139, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59604, {{186145, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{11474, {{186147, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57498, {{186148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57499, {{186149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57500, {{186150, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57501, {{186151, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{59608, {{186158, 1, 0, 0},{186158, 1, 1, 16}, {0, 0, 0, 0}}},{59608, {{186158, 1, 0, 0},{186158, 1, 1, 16}, {0, 0, 0, 0}}},{59609, {{186158, 1, 0, 0},{186158, 1, 1, 17}, {0, 0, 0, 0}}},{59609, {{186158, 1, 0, 0},{186158, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{57506, {{186163, 1, 0, 0},{186163, 1, 1, 16}, {0, 0, 0, 0}}},{57506, {{186163, 1, 0, 0},{186163, 1, 1, 16}, {0, 0, 0, 0}}},{57512, {{186163, 1, 0, 0},{186163, 1, 1, 17}, {0, 0, 0, 0}}},{57512, {{186163, 1, 0, 0},{186163, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{57497, {{186164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57505, {{186165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59614, {{186166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59610, {{186174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59612, {{186181, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59615, {{186182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59620, {{186188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59618, {{186193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59621, {{186201, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59619, {{186205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57755, {{186208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59622, {{186209, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59624, {{186210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59627, {{186216, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59626, {{186231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59625, {{186241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57513, {{186245, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57514, {{186251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57516, {{186252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59633, {{186255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57523, {{186256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59631, {{186264, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59629, {{186270, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57519, {{186277, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59632, {{186293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57526, {{186297, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{59635, {{186302, 1, 0, 0},{186302, 1, 1, 16}, {0, 0, 0, 0}}},{59635, {{186302, 1, 0, 0},{186302, 1, 1, 16}, {0, 0, 0, 0}}},{59636, {{186302, 1, 0, 0},{186302, 1, 1, 17}, {0, 0, 0, 0}}},{59636, {{186302, 1, 0, 0},{186302, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59637, {{186305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59638, {{186308, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57528, {{186333, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57531, {{186334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59639, {{186350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57536, {{186351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{59640, {{186353, 1, 0, 0},{186353, 1, 1, 16}, {0, 0, 0, 0}}},{59640, {{186353, 1, 0, 0},{186353, 1, 1, 16}, {0, 0, 0, 0}}},{59641, {{186353, 1, 0, 0},{186353, 1, 1, 17}, {0, 0, 0, 0}}},{59641, {{186353, 1, 0, 0},{186353, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{57538, {{186354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57541, {{186358, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59642, {{186361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57542, {{186364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57543, {{186366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57547, {{186370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{57548, {{186373, 1, 0, 0},{186373, 1, 1, 16}, {0, 0, 0, 0}}},{57548, {{186373, 1, 0, 0},{186373, 1, 1, 16}, {0, 0, 0, 0}}},{57550, {{186373, 1, 0, 0},{186373, 1, 1, 17}, {0, 0, 0, 0}}},{57550, {{186373, 1, 0, 0},{186373, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{36681, {{186377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59643, {{186381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57554, {{186392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57556, {{186400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57552, {{186401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59644, {{186440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57563, {{186464, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59645, {{186487, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59647, {{186488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59650, {{186501, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{57564, {{186502, 1, 0, 0},{186502, 1, 1, 16}, {0, 0, 0, 0}}},{57564, {{186502, 1, 0, 0},{186502, 1, 1, 16}, {0, 0, 0, 0}}},{57567, {{186502, 1, 0, 0},{186502, 1, 1, 17}, {0, 0, 0, 0}}},{57567, {{186502, 1, 0, 0},{186502, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59649, {{186511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59652, {{186520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59651, {{186535, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{12884, {{186539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59653, {{186543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59654, {{186564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57580, {{186580, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57603, {{186586, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57582, {{186594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57583, {{186595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57585, {{186601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57591, {{186604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57588, {{186607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57589, {{186608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57590, {{186609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57593, {{186612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57596, {{186613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57600, {{186616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57598, {{186617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57602, {{186618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57601, {{186619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57609, {{186622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57607, {{186624, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57604, {{186625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57605, {{186626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59658, {{186627, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59657, {{186628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57613, {{186634, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57616, {{186635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57620, {{186639, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57621, {{186643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59660, {{186645, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{59663, {{186646, 1, 0, 0},{186646, 1, 1, 16}, {0, 0, 0, 0}}},{59663, {{186646, 1, 0, 0},{186646, 1, 1, 16}, {0, 0, 0, 0}}},{59664, {{186646, 1, 0, 0},{186646, 1, 1, 17}, {0, 0, 0, 0}}},{59664, {{186646, 1, 0, 0},{186646, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{59665, {{186647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{59661, {{186648, 1, 0, 0},{186648, 1, 1, 16}, {0, 0, 0, 0}}},{59661, {{186648, 1, 0, 0},{186648, 1, 1, 16}, {0, 0, 0, 0}}},{59662, {{186648, 1, 0, 0},{186648, 1, 1, 17}, {0, 0, 0, 0}}},{59662, {{186648, 1, 0, 0},{186648, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{59666, {{186649, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59668, {{186651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57625, {{186655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59753, {{186656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59670, {{186659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57627, {{186661, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57630, {{186663, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57634, {{186665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57632, {{186666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57636, {{186670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59672, {{186671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57643, {{186685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57646, {{186694, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57648, {{186701, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57649, {{186702, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57654, {{186711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59675, {{186712, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59674, {{186714, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59673, {{186716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57652, {{186719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59676, {{186726, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{59677, {{186727, 1, 0, 0},{186727, 1, 1, 16}, {0, 0, 0, 0}}},{59677, {{186727, 1, 0, 0},{186727, 1, 1, 16}, {0, 0, 0, 0}}},{59681, {{186727, 1, 0, 0},{186727, 1, 1, 17}, {0, 0, 0, 0}}},{59681, {{186727, 1, 0, 0},{186727, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57658, {{186731, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57660, {{186732, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57661, {{186733, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59680, {{186744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59682, {{186752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59683, {{186753, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57667, {{186755, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57670, {{186764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59684, {{186769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{57668, {{186775, 1, 0, 0},{186775, 1, 1, 16}, {0, 0, 0, 0}}},{57668, {{186775, 1, 0, 0},{186775, 1, 1, 16}, {0, 0, 0, 0}}},{57669, {{186775, 1, 0, 0},{186775, 1, 1, 17}, {0, 0, 0, 0}}},{57669, {{186775, 1, 0, 0},{186775, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59686, {{186777, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59687, {{186778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59690, {{186779, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59689, {{186784, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59705, {{186792, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59692, {{186793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59693, {{186813, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57675, {{186818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59695, {{186828, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57676, {{186833, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59697, {{186843, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57678, {{186847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59696, {{186848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59699, {{186857, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{37801, {{186867, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59700, {{186869, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59698, {{186877, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59701, {{186893, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59702, {{186900, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57683, {{186903, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57650, {{186904, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59706, {{186928, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59707, {{186937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57691, {{186941, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59708, {{186972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59709, {{186975, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57715, {{186976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57704, {{186977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57217, {{186981, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59713, {{186983, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59712, {{186985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57716, {{186993, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57717, {{186994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59714, {{186995, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57730, {{187008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57726, {{187009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59717, {{187016, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59716, {{187026, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57739, {{187029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57738, {{187030, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57737, {{187031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{57756, {{187033, 1, 0, 0},{187033, 1, 1, 16}, {0, 0, 0, 0}}},{57756, {{187033, 1, 0, 0},{187033, 1, 1, 16}, {0, 0, 0, 0}}},{57769, {{187033, 1, 0, 0},{187033, 1, 1, 17}, {0, 0, 0, 0}}},{57769, {{187033, 1, 0, 0},{187033, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{59719, {{187034, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59721, {{187035, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59724, {{187036, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59718, {{187054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59720, {{187055, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59723, {{187058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57754, {{187060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57747, {{187061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59726, {{187062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59732, {{187068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57771, {{187080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59727, {{187085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57751, {{187086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57786, {{187090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59733, {{187091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59728, {{187093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57789, {{187096, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59730, {{187098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59731, {{187100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57788, {{187101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57792, {{187102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57791, {{187103, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59738, {{187106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57815, {{187116, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59741, {{187119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59742, {{187120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57810, {{187123, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59740, {{187128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57794, {{187130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59739, {{187135, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59744, {{187137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59746, {{187138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59748, {{187146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57838, {{187152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57840, {{187153, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{14692, {{187154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57846, {{187164, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59749, {{187172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57854, {{187176, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59751, {{187183, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60324, {{187186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57863, {{187187, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59752, {{187188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57870, {{187194, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57875, {{187203, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{38549, {{187210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{57878, {{187211, 1, 0, 0},{187211, 1, 1, 16}, {0, 0, 0, 0}}},{57878, {{187211, 1, 0, 0},{187211, 1, 1, 16}, {0, 0, 0, 0}}},{57879, {{187211, 1, 0, 0},{187211, 1, 1, 17}, {0, 0, 0, 0}}},{57879, {{187211, 1, 0, 0},{187211, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{57882, {{187214, 1, 0, 0},{187214, 1, 1, 16}, {0, 0, 0, 0}}},{57882, {{187214, 1, 0, 0},{187214, 1, 1, 16}, {0, 0, 0, 0}}},{57883, {{187214, 1, 0, 0},{187214, 1, 1, 17}, {0, 0, 0, 0}}},{57883, {{187214, 1, 0, 0},{187214, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{57881, {{187215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59755, {{187218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57885, {{187219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57886, {{187220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57887, {{187221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59756, {{187224, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59757, {{187238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57895, {{187241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59761, {{187248, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57901, {{187261, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57900, {{187262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57903, {{187264, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57902, {{187265, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59762, {{187267, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57904, {{187269, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57905, {{187272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59763, {{187278, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57907, {{187281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57908, {{187286, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59764, {{187287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57911, {{187299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59765, {{187307, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57916, {{187309, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57919, {{187312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57920, {{187313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57924, {{187315, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59767, {{187316, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59766, {{187317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57921, {{187318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57926, {{187322, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57927, {{187323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57928, {{187324, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59769, {{187331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57929, {{187335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57936, {{187376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57938, {{187380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59773, {{187382, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57941, {{187393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{15203, {{187401, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57944, {{187402, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57942, {{187403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57947, {{187407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57946, {{187408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59774, {{187409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57948, {{187414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59777, {{187427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57951, {{187430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59785, {{187433, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59780, {{187439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59784, {{187440, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59793, {{187441, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57955, {{187449, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59786, {{187452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57770, {{187457, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57952, {{187458, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59787, {{187460, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59795, {{187461, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59796, {{187462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59794, {{187470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59790, {{187471, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59788, {{187482, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57960, {{187489, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59804, {{187490, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59799, {{187497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59800, {{187498, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59805, {{187502, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59797, {{187504, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59798, {{187505, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59802, {{187508, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59824, {{187511, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59810, {{187516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{59820, {{187517, 1, 0, 0},{187517, 1, 1, 16}, {0, 0, 0, 0}}},{59820, {{187517, 1, 0, 0},{187517, 1, 1, 16}, {0, 0, 0, 0}}},{59821, {{187517, 1, 0, 0},{187517, 1, 1, 17}, {0, 0, 0, 0}}},{59821, {{187517, 1, 0, 0},{187517, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59806, {{187523, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59807, {{187524, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59823, {{187525, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59809, {{187531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59811, {{187532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59812, {{187533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57963, {{187537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59822, {{187538, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59826, {{187539, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59832, {{187540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{59830, {{187550, 1, 0, 0},{187550, 1, 1, 16}, {0, 0, 0, 0}}},{59830, {{187550, 1, 0, 0},{187550, 1, 1, 16}, {0, 0, 0, 0}}},{59831, {{187550, 1, 0, 0},{187550, 1, 1, 17}, {0, 0, 0, 0}}},{59831, {{187550, 1, 0, 0},{187550, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59828, {{187552, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59827, {{187559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59836, {{187566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59837, {{187571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59840, {{187582, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59838, {{187585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59834, {{187592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59841, {{187594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57974, {{187597, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57972, {{187599, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59847, {{187600, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {10, {{59843, {{187603, 1, 0, 0},{187603, 1, 1, 16}, {0, 0, 0, 0}}},{59843, {{187603, 1, 0, 0},{187603, 1, 1, 16}, {0, 0, 0, 0}}},{59844, {{187603, 1, 0, 0},{187603, 1, 1, 17}, {0, 0, 0, 0}}},{59844, {{187603, 1, 0, 0},{187603, 1, 1, 17}, {0, 0, 0, 0}}},{59845, {{187603, 1, 0, 0},{187603, 1, 1, 18}, {0, 0, 0, 0}}},{59845, {{187603, 1, 0, 0},{187603, 1, 1, 18}, {0, 0, 0, 0}}},{59846, {{187603, 1, 0, 0},{187603, 1, 1, 19}, {0, 0, 0, 0}}},{59846, {{187603, 1, 0, 0},{187603, 1, 1, 19}, {0, 0, 0, 0}}},{59850, {{187603, 1, 0, 0},{187603, 1, 1, 20}, {0, 0, 0, 0}}},{59850, {{187603, 1, 0, 0},{187603, 1, 1, 20}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59848, {{187606, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59842, {{187607, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59839, {{187610, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{39755, {{187611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59851, {{187616, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59852, {{187619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59853, {{187620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59849, {{187625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{59854, {{187631, 1, 0, 0},{187631, 1, 1, 16}, {0, 0, 0, 0}}},{59854, {{187631, 1, 0, 0},{187631, 1, 1, 16}, {0, 0, 0, 0}}},{59855, {{187631, 1, 0, 0},{187631, 1, 1, 17}, {0, 0, 0, 0}}},{59855, {{187631, 1, 0, 0},{187631, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57981, {{187642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57982, {{187647, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57985, {{187665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59856, {{187670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57988, {{187673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57987, {{187674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59857, {{187691, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59858, {{187692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57994, {{187703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57998, {{187706, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59863, {{187707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59864, {{187709, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57996, {{187710, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59862, {{187711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{56993, {{187716, 1, 0, 0},{187716, 1, 1, 16}, {0, 0, 0, 0}}},{56993, {{187716, 1, 0, 0},{187716, 1, 1, 16}, {0, 0, 0, 0}}},{56994, {{187716, 1, 0, 0},{187716, 1, 1, 17}, {0, 0, 0, 0}}},{56994, {{187716, 1, 0, 0},{187716, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{39953, {{187719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{56944, {{187720, 1, 0, 0},{187720, 1, 1, 16}, {0, 0, 0, 0}}},{56944, {{187720, 1, 0, 0},{187720, 1, 1, 16}, {0, 0, 0, 0}}},{57999, {{187720, 1, 0, 0},{187720, 1, 1, 17}, {0, 0, 0, 0}}},{57999, {{187720, 1, 0, 0},{187720, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59866, {{187723, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59867, {{187737, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58002, {{187740, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59870, {{187741, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58001, {{187744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58000, {{187745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59871, {{187752, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59869, {{187763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59874, {{187764, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59873, {{187770, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58006, {{187772, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59872, {{187778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59868, {{187785, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58007, {{187786, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59878, {{187788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59877, {{187794, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58197, {{187799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58012, {{187809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58015, {{187818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{40250, {{187832, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59880, {{187844, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59882, {{187864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59881, {{187865, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58024, {{187875, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58027, {{187888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58028, {{187898, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59885, {{187912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59887, {{187917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59888, {{187922, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59889, {{187936, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59890, {{187937, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58034, {{187939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58036, {{187950, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59892, {{187957, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59893, {{187958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59895, {{187964, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58041, {{187965, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59897, {{187972, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58045, {{187989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59899, {{187990, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59901, {{187991, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59904, {{188003, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59906, {{188004, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59909, {{188018, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59912, {{188023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59911, {{188028, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59910, {{188029, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58052, {{188038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58056, {{188059, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58057, {{188060, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58066, {{188067, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58071, {{188069, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58072, {{188071, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58075, {{188074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58074, {{188075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57191, {{188078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59913, {{188081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58078, {{188083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58081, {{188088, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58082, {{188090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{58090, {{188094, 1, 0, 0},{188094, 1, 1, 16}, {0, 0, 0, 0}}},{58090, {{188094, 1, 0, 0},{188094, 1, 1, 16}, {0, 0, 0, 0}}},{58093, {{188094, 1, 0, 0},{188094, 1, 1, 17}, {0, 0, 0, 0}}},{58093, {{188094, 1, 0, 0},{188094, 1, 1, 17}, {0, 0, 0, 0}}},{58094, {{188094, 1, 0, 0},{188094, 1, 1, 18}, {0, 0, 0, 0}}},{58094, {{188094, 1, 0, 0},{188094, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {1, {{58089, {{188095, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58092, {{188105, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58091, {{188106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59916, {{188108, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{41389, {{188112, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59917, {{188114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58097, {{188119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59918, {{188121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58103, {{188133, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58105, {{188136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58104, {{188137, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58107, {{188143, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58106, {{188144, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58109, {{188148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58108, {{188149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59919, {{188152, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58112, {{188154, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58113, {{188159, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58114, {{188162, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58115, {{188163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58116, {{188165, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58119, {{188171, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59921, {{188174, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58120, {{188182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{17959, {{188208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58127, {{188213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59923, {{188216, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59924, {{188220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{58128, {{188221, 1, 0, 0},{188221, 1, 1, 16}, {0, 0, 0, 0}}},{58128, {{188221, 1, 0, 0},{188221, 1, 1, 16}, {0, 0, 0, 0}}},{58130, {{188221, 1, 0, 0},{188221, 1, 1, 17}, {0, 0, 0, 0}}},{58130, {{188221, 1, 0, 0},{188221, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{58131, {{188222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{58132, {{188225, 1, 0, 0},{188225, 1, 1, 16}, {0, 0, 0, 0}}},{58132, {{188225, 1, 0, 0},{188225, 1, 1, 16}, {0, 0, 0, 0}}},{58133, {{188225, 1, 0, 0},{188225, 1, 1, 17}, {0, 0, 0, 0}}},{58133, {{188225, 1, 0, 0},{188225, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{58134, {{188226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59925, {{188229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58135, {{188230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58136, {{188231, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59926, {{188235, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59927, {{188237, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59928, {{188239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58137, {{188242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59929, {{188249, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59930, {{188252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59931, {{188254, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58141, {{188266, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59932, {{188280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58143, {{188284, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59933, {{188287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58145, {{188290, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58147, {{188294, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59935, {{188295, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58150, {{188298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58148, {{188299, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58151, {{188303, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58152, {{188304, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58154, {{188316, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58155, {{188319, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59936, {{188326, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58162, {{188329, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58163, {{188330, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58169, {{188337, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58179, {{188371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58181, {{188375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58183, {{188380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58187, {{188389, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59942, {{188390, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59943, {{188393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58188, {{188394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{58190, {{188405, 1, 0, 0},{188405, 1, 1, 16}, {0, 0, 0, 0}}},{58190, {{188405, 1, 0, 0},{188405, 1, 1, 16}, {0, 0, 0, 0}}},{58191, {{188405, 1, 0, 0},{188405, 1, 1, 17}, {0, 0, 0, 0}}},{58191, {{188405, 1, 0, 0},{188405, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{58192, {{188406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58193, {{188407, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59944, {{188415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59945, {{188417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59946, {{188419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59947, {{188423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59948, {{188432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59949, {{188448, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59951, {{188463, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59952, {{188467, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59953, {{188470, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{42858, {{188472, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59954, {{188479, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59955, {{188483, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58198, {{188492, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59959, {{188514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59958, {{188515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59961, {{188516, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{58205, {{188520, 1, 0, 0},{188520, 1, 1, 16}, {0, 0, 0, 0}}},{58205, {{188520, 1, 0, 0},{188520, 1, 1, 16}, {0, 0, 0, 0}}},{58209, {{188520, 1, 0, 0},{188520, 1, 1, 17}, {0, 0, 0, 0}}},{58209, {{188520, 1, 0, 0},{188520, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{59960, {{188521, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58206, {{188526, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59962, {{188527, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59965, {{188528, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59964, {{188530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{43002, {{188531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59963, {{188532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59966, {{188534, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59968, {{188537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59971, {{188551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58213, {{188554, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59974, {{188559, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59975, {{188564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59977, {{188567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59976, {{188570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59978, {{188572, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58218, {{188574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59979, {{188578, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59980, {{188583, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59981, {{188591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59982, {{188594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{59983, {{188602, 1, 0, 0},{188602, 1, 1, 16}, {0, 0, 0, 0}}},{59983, {{188602, 1, 0, 0},{188602, 1, 1, 16}, {0, 0, 0, 0}}},{59984, {{188602, 1, 0, 0},{188602, 1, 1, 17}, {0, 0, 0, 0}}},{59984, {{188602, 1, 0, 0},{188602, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{68048, {{188603, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59986, {{188604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59985, {{188609, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43169, {{188612, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58223, {{188622, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{43238, {{188630, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59991, {{188631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58231, {{188635, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59992, {{188636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58232, {{188642, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58233, {{188643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{59995, {{188652, 1, 0, 0},{188652, 1, 1, 16}, {0, 0, 0, 0}}},{59995, {{188652, 1, 0, 0},{188652, 1, 1, 16}, {0, 0, 0, 0}}},{59996, {{188652, 1, 0, 0},{188652, 1, 1, 17}, {0, 0, 0, 0}}},{59996, {{188652, 1, 0, 0},{188652, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59997, {{188656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60000, {{188660, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58242, {{188667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60006, {{188673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60005, {{188674, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58249, {{188690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58251, {{188700, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58253, {{188707, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60007, {{188711, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60009, {{188717, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60008, {{188718, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60010, {{188719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60011, {{188721, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60012, {{188724, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60013, {{188725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60014, {{188735, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58261, {{188739, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60017, {{188745, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58264, {{188751, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{19541, {{188754, 1, 0, 0},{188754, 1, 1, 16}, {0, 0, 0, 0}}},{19541, {{188754, 1, 0, 0},{188754, 1, 1, 16}, {0, 0, 0, 0}}},{58266, {{188754, 1, 0, 0},{188754, 1, 1, 17}, {0, 0, 0, 0}}},{58266, {{188754, 1, 0, 0},{188754, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60018, {{188756, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58270, {{188757, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58267, {{188761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60020, {{188766, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60021, {{188769, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60023, {{188783, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{60024, {{188784, 1, 0, 0},{188784, 1, 1, 16}, {0, 0, 0, 0}}},{60024, {{188784, 1, 0, 0},{188784, 1, 1, 16}, {0, 0, 0, 0}}},{60025, {{188784, 1, 0, 0},{188784, 1, 1, 17}, {0, 0, 0, 0}}},{60025, {{188784, 1, 0, 0},{188784, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60029, {{188790, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58275, {{188803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60028, {{188806, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{2237, {{188840, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60033, {{188842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60031, {{188847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58280, {{188852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60032, {{188854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58281, {{188861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58283, {{188864, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58284, {{188866, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60034, {{188875, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{68049, {{188878, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58290, {{188883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60035, {{188892, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60036, {{188894, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58291, {{188899, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60037, {{188905, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58296, {{188908, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60039, {{188917, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44340, {{188923, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60041, {{188931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58302, {{188932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58301, {{188933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60042, {{188934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60043, {{188935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58303, {{188939, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60045, {{188940, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60046, {{188943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58308, {{188945, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58307, {{188948, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60049, {{188949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58314, {{188955, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60051, {{188956, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58320, {{188963, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60053, {{188966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60055, {{188976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58324, {{188977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58322, {{188981, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58325, {{188983, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60056, {{189000, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58331, {{189015, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58333, {{189023, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57384, {{189040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60058, {{189041, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58336, {{189045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60059, {{189048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{57009, {{189052, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58337, {{189057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58340, {{189058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{58338, {{189060, 1, 0, 0},{189060, 1, 1, 16}, {0, 0, 0, 0}}},{58338, {{189060, 1, 0, 0},{189060, 1, 1, 16}, {0, 0, 0, 0}}},{58339, {{189060, 1, 0, 0},{189060, 1, 1, 17}, {0, 0, 0, 0}}},{58339, {{189060, 1, 0, 0},{189060, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{60060, {{189061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{56995, {{189062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58343, {{189068, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58344, {{189074, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58341, {{189075, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58342, {{189076, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44942, {{189078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{44922, {{189080, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60061, {{189090, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60063, {{189093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60065, {{189098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{58349, {{189099, 1, 0, 0},{189099, 1, 1, 16}, {0, 0, 0, 0}}},{58349, {{189099, 1, 0, 0},{189099, 1, 1, 16}, {0, 0, 0, 0}}},{60064, {{189099, 1, 0, 0},{189099, 1, 1, 17}, {0, 0, 0, 0}}},{60064, {{189099, 1, 0, 0},{189099, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58350, {{189109, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{32767, {{189110, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58351, {{189111, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58354, {{189112, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58353, {{189114, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60066, {{189118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58356, {{189119, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58357, {{189120, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60067, {{189121, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58359, {{189130, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58360, {{189138, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60069, {{189146, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58362, {{189149, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60070, {{189158, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60072, {{189166, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60073, {{189168, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60074, {{189176, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60075, {{189177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58370, {{189179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58372, {{189185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60076, {{189186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58374, {{189188, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{58375, {{189190, 1, 0, 0},{189190, 1, 1, 16}, {0, 0, 0, 0}}},{58375, {{189190, 1, 0, 0},{189190, 1, 1, 16}, {0, 0, 0, 0}}},{58376, {{189190, 1, 0, 0},{189190, 1, 1, 17}, {0, 0, 0, 0}}},{58376, {{189190, 1, 0, 0},{189190, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {4, {{58377, {{189191, 1, 0, 0},{189191, 1, 1, 16}, {0, 0, 0, 0}}},{58377, {{189191, 1, 0, 0},{189191, 1, 1, 16}, {0, 0, 0, 0}}},{58378, {{189191, 1, 0, 0},{189191, 1, 1, 17}, {0, 0, 0, 0}}},{58378, {{189191, 1, 0, 0},{189191, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58379, {{189195, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60077, {{189199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60078, {{189200, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58382, {{189202, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58385, {{189204, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58388, {{189205, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58384, {{189206, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58387, {{189210, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60081, {{189211, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58391, {{189212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60082, {{189222, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58394, {{189225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60083, {{189239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60084, {{189242, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58402, {{189252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60085, {{189256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58405, {{189257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58406, {{189260, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58407, {{189261, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{40315, {{189262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58408, {{189264, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59922, {{189266, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58410, {{189275, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58411, {{189280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60087, {{189283, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58414, {{189285, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60088, {{189287, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58417, {{189288, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60091, {{189291, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60092, {{189292, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60093, {{189293, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60094, {{189296, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58421, {{189298, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58422, {{189305, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58426, {{189313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58429, {{189317, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58430, {{189318, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58431, {{189323, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60096, {{189325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58433, {{189331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58437, {{189335, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58438, {{189339, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58439, {{189340, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60097, {{189346, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60098, {{189347, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58445, {{189350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58451, {{189351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60103, {{189354, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58454, {{189359, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {6, {{58447, {{189360, 1, 0, 0},{189360, 1, 1, 16}, {0, 0, 0, 0}}},{58447, {{189360, 1, 0, 0},{189360, 1, 1, 16}, {0, 0, 0, 0}}},{60101, {{189360, 1, 0, 0},{189360, 1, 1, 17}, {0, 0, 0, 0}}},{60101, {{189360, 1, 0, 0},{189360, 1, 1, 17}, {0, 0, 0, 0}}},{60102, {{189360, 1, 0, 0},{189360, 1, 1, 18}, {0, 0, 0, 0}}},{60102, {{189360, 1, 0, 0},{189360, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58455, {{189365, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58450, {{189367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60104, {{189368, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57225, {{189369, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{57224, {{189370, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60105, {{189371, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58464, {{189374, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58483, {{189375, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{58472, {{189376, 1, 0, 0},{189376, 1, 1, 16}, {0, 0, 0, 0}}},{58472, {{189376, 1, 0, 0},{189376, 1, 1, 16}, {0, 0, 0, 0}}},{60108, {{189376, 1, 0, 0},{189376, 1, 1, 17}, {0, 0, 0, 0}}},{60108, {{189376, 1, 0, 0},{189376, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{60119, {{189377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58463, {{189380, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60107, {{189383, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58467, {{189387, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60110, {{189388, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60111, {{189389, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60114, {{189392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58480, {{189393, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60120, {{189394, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58482, {{189395, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58490, {{189396, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58481, {{189398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60116, {{189406, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60124, {{189411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60126, {{189413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60123, {{189414, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60121, {{189415, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58496, {{189417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60128, {{189427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58498, {{189435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58497, {{189436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60122, {{189439, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58506, {{189443, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60131, {{189447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{22574, {{189459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58526, {{189477, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58519, {{189478, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60132, {{189481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58523, {{189486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60133, {{189497, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60134, {{189499, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{58533, {{189500, 1, 0, 0},{189500, 1, 1, 16}, {0, 0, 0, 0}}},{58533, {{189500, 1, 0, 0},{189500, 1, 1, 16}, {0, 0, 0, 0}}},{58541, {{189500, 1, 0, 0},{189500, 1, 1, 17}, {0, 0, 0, 0}}},{58541, {{189500, 1, 0, 0},{189500, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58534, {{189514, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58540, {{189530, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60137, {{189531, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60138, {{189532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60135, {{189533, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60140, {{189536, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58539, {{189558, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60142, {{189564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60143, {{189579, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58550, {{189584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60141, {{189588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58552, {{189591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60146, {{189592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58560, {{189593, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58554, {{189594, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60145, {{189595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60148, {{189596, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58557, {{189608, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58563, {{189611, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58568, {{189625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58567, {{189631, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60151, {{189632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58566, {{189640, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58569, {{189643, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55359, {{189654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58578, {{189655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58582, {{189656, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{58580, {{189657, 1, 0, 0},{189657, 1, 1, 16}, {0, 0, 0, 0}}},{58580, {{189657, 1, 0, 0},{189657, 1, 1, 16}, {0, 0, 0, 0}}},{58581, {{189657, 1, 0, 0},{189657, 1, 1, 17}, {0, 0, 0, 0}}},{58581, {{189657, 1, 0, 0},{189657, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {1, {{58584, {{189658, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60152, {{189659, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{58586, {{189660, 1, 0, 0},{189660, 1, 1, 16}, {0, 0, 0, 0}}},{58586, {{189660, 1, 0, 0},{189660, 1, 1, 16}, {0, 0, 0, 0}}},{58587, {{189660, 1, 0, 0},{189660, 1, 1, 17}, {0, 0, 0, 0}}},{58587, {{189660, 1, 0, 0},{189660, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58588, {{189662, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58589, {{189664, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58591, {{189665, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60079, {{189666, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58593, {{189667, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58594, {{189669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60154, {{189670, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60155, {{189679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58595, {{189680, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58596, {{189681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58599, {{189685, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60157, {{189690, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60156, {{189692, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60158, {{189693, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58613, {{189697, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58614, {{189703, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60159, {{189704, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{58623, {{189705, 1, 0, 0},{189705, 1, 1, 16}, {0, 0, 0, 0}}},{58623, {{189705, 1, 0, 0},{189705, 1, 1, 16}, {0, 0, 0, 0}}},{58635, {{189705, 1, 0, 0},{189705, 1, 1, 17}, {0, 0, 0, 0}}},{58635, {{189705, 1, 0, 0},{189705, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58622, {{189716, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60160, {{189719, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58630, {{189725, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60161, {{189744, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58658, {{189762, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58661, {{189763, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58665, {{189768, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60162, {{189771, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58667, {{189778, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58668, {{189780, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58670, {{189787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58671, {{189788, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58672, {{189789, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58679, {{189793, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60166, {{189796, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60167, {{189797, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58688, {{189801, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58690, {{189802, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58689, {{189803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58693, {{189809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58695, {{189818, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60168, {{189831, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58703, {{189850, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60169, {{189851, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58704, {{189856, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58707, {{189861, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60172, {{189871, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60173, {{189883, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58711, {{189886, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58712, {{189887, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58713, {{189888, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58718, {{189889, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58716, {{189890, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58719, {{189891, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60174, {{189896, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60175, {{189907, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60176, {{189910, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58722, {{189912, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58724, {{189914, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60177, {{189916, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58731, {{189919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58730, {{189920, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{58732, {{189928, 1, 0, 0},{189928, 1, 1, 16}, {0, 0, 0, 0}}},{58732, {{189928, 1, 0, 0},{189928, 1, 1, 16}, {0, 0, 0, 0}}},{58733, {{189928, 1, 0, 0},{189928, 1, 1, 17}, {0, 0, 0, 0}}},{58733, {{189928, 1, 0, 0},{189928, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60179, {{189931, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60181, {{189932, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60180, {{189933, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58735, {{189934, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60182, {{189938, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58738, {{189943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60183, {{189949, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58740, {{189953, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58742, {{189954, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58744, {{189960, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60184, {{189962, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58745, {{189966, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58746, {{189968, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60185, {{189977, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60187, {{189985, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58753, {{189988, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60188, {{189989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58751, {{189990, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58749, {{189994, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58757, {{189997, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58761, {{190008, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60191, {{190009, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60192, {{190013, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58765, {{190014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60193, {{190020, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58770, {{190021, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58769, {{190022, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60195, {{190039, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58777, {{190040, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60196, {{190042, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58778, {{190043, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60197, {{190044, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{24925, {{190045, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58779, {{190046, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58780, {{190047, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58781, {{190048, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{49443, {{190051, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60200, {{190054, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58782, {{190058, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60201, {{190061, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58783, {{190062, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58785, {{190066, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58788, {{190078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58787, {{190079, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60202, {{190081, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58789, {{190083, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58309, {{190084, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58792, {{190085, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58791, {{190086, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58315, {{190087, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58793, {{190091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58794, {{190092, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60203, {{190093, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58795, {{190097, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60205, {{190101, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58796, {{190102, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60204, {{190106, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60206, {{190118, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60207, {{190128, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60208, {{190132, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58799, {{190134, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58800, {{190136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60210, {{190169, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60212, {{190177, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58811, {{190185, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60211, {{190186, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60214, {{190208, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60215, {{190217, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58815, {{190226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60216, {{190229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58821, {{190252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60218, {{190262, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{25549, {{190271, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60219, {{190272, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58827, {{190279, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58828, {{190301, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60221, {{190302, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60222, {{190313, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{25720, {{190316, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58831, {{190331, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60223, {{190332, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58835, {{190334, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58840, {{190338, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58844, {{190341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58846, {{190342, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{60224, {{190345, 1, 0, 0},{190345, 1, 1, 16}, {0, 0, 0, 0}}},{60224, {{190345, 1, 0, 0},{190345, 1, 1, 16}, {0, 0, 0, 0}}},{60225, {{190345, 1, 0, 0},{190345, 1, 1, 17}, {0, 0, 0, 0}}},{60225, {{190345, 1, 0, 0},{190345, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58851, {{190360, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60227, {{190361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60229, {{190362, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60226, {{190367, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58858, {{190381, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60231, {{190384, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58861, {{190391, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58864, {{190397, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60232, {{190398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58862, {{190400, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58863, {{190403, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60233, {{190408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{60244, {{190416, 1, 0, 0},{190416, 1, 1, 16}, {0, 0, 0, 0}}},{60244, {{190416, 1, 0, 0},{190416, 1, 1, 16}, {0, 0, 0, 0}}},{60245, {{190416, 1, 0, 0},{190416, 1, 1, 17}, {0, 0, 0, 0}}},{60245, {{190416, 1, 0, 0},{190416, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60243, {{190423, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60247, {{190424, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60246, {{190426, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{1865, {{190429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58873, {{190430, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60249, {{190435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60250, {{190450, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60251, {{190459, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{51296, {{190462, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58879, {{190481, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58880, {{190485, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58882, {{190486, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60254, {{190488, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58883, {{190490, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60255, {{190509, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60256, {{190515, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58887, {{190518, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58888, {{190520, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58890, {{190532, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58891, {{190537, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58892, {{190540, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60257, {{190543, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60259, {{190551, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60261, {{190564, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60260, {{190566, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60263, {{190567, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60265, {{190568, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60264, {{190570, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60266, {{190571, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60270, {{190574, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60267, {{190575, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60268, {{190576, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58905, {{190577, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60272, {{190580, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60273, {{190584, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60275, {{190585, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60274, {{190588, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58911, {{190591, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60276, {{190592, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60278, {{190595, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58915, {{190601, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60282, {{190602, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60277, {{190604, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60279, {{190605, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60280, {{190613, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60281, {{190614, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60288, {{190617, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60285, {{190618, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60286, {{190619, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60284, {{190620, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58926, {{190623, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60287, {{190625, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60289, {{190626, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58927, {{190628, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60290, {{190632, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60293, {{190636, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60291, {{190644, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60295, {{190651, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60296, {{190652, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60297, {{190654, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60298, {{190655, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60305, {{190657, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60299, {{190669, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60302, {{190671, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58947, {{190672, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58948, {{190673, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60304, {{190677, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60301, {{190679, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{5236, {{190681, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60306, {{190686, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60307, {{190687, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58956, {{190713, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58958, {{190727, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60312, {{190746, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60313, {{190748, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60314, {{190759, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58965, {{190760, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58967, {{190761, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60315, {{190782, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58976, {{190787, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58969, {{190796, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58970, {{190799, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58973, {{190803, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60318, {{190809, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58974, {{190810, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60319, {{190836, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60320, {{190837, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58978, {{190838, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60321, {{190839, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58980, {{190841, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58979, {{190842, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60322, {{190846, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{58982, {{190847, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60323, {{190848, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58988, {{190850, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58989, {{190852, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58992, {{190854, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60328, {{190876, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58995, {{190879, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60327, {{190882, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{27877, {{190903, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{60330, {{190912, 1, 0, 0},{190912, 1, 1, 16}, {0, 0, 0, 0}}},{60330, {{190912, 1, 0, 0},{190912, 1, 1, 16}, {0, 0, 0, 0}}},{60331, {{190912, 1, 0, 0},{190912, 1, 1, 17}, {0, 0, 0, 0}}},{60331, {{190912, 1, 0, 0},{190912, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59000, {{190919, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{59001, {{190922, 1, 0, 0},{190922, 1, 1, 16}, {0, 0, 0, 0}}},{59001, {{190922, 1, 0, 0},{190922, 1, 1, 16}, {0, 0, 0, 0}}},{59002, {{190922, 1, 0, 0},{190922, 1, 1, 17}, {0, 0, 0, 0}}},{59002, {{190922, 1, 0, 0},{190922, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59003, {{190924, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60332, {{190930, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{59005, {{190931, 1, 0, 0},{190931, 1, 1, 16}, {0, 0, 0, 0}}},{59005, {{190931, 1, 0, 0},{190931, 1, 1, 16}, {0, 0, 0, 0}}},{59007, {{190931, 1, 0, 0},{190931, 1, 1, 17}, {0, 0, 0, 0}}},{59007, {{190931, 1, 0, 0},{190931, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60333, {{190935, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59009, {{190942, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59011, {{190943, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59018, {{190958, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59017, {{190959, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {4, {{59015, {{190960, 1, 0, 0},{190960, 1, 1, 16}, {0, 0, 0, 0}}},{59015, {{190960, 1, 0, 0},{190960, 1, 1, 16}, {0, 0, 0, 0}}},{59016, {{190960, 1, 0, 0},{190960, 1, 1, 17}, {0, 0, 0, 0}}},{59016, {{190960, 1, 0, 0},{190960, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{60334, {{190962, 1, 0, 0},{190962, 1, 1, 16}, {0, 0, 0, 0}}},{60334, {{190962, 1, 0, 0},{190962, 1, 1, 16}, {0, 0, 0, 0}}},{60335, {{190962, 1, 0, 0},{190962, 1, 1, 17}, {0, 0, 0, 0}}},{60335, {{190962, 1, 0, 0},{190962, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59286, {{190971, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59022, {{190976, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59024, {{190983, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59025, {{190984, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60337, {{190989, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59035, {{191014, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59040, {{191031, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60339, {{191037, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60340, {{191038, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {6, {{60341, {{191041, 1, 0, 0},{191041, 1, 1, 16}, {0, 0, 0, 0}}},{60341, {{191041, 1, 0, 0},{191041, 1, 1, 16}, {0, 0, 0, 0}}},{60342, {{191041, 1, 0, 0},{191041, 1, 1, 17}, {0, 0, 0, 0}}},{60342, {{191041, 1, 0, 0},{191041, 1, 1, 17}, {0, 0, 0, 0}}},{68100, {{191041, 1, 0, 0},{191041, 1, 1, 18}, {0, 0, 0, 0}}},{68100, {{191041, 1, 0, 0},{191041, 1, 1, 18}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59047, {{191057, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58237, {{191078, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60344, {{191091, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60345, {{191098, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59049, {{191100, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59050, {{191111, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60346, {{191112, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59052, {{191117, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59056, {{191132, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59057, {{191136, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59060, {{191142, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59063, {{191148, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60348, {{191163, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{54715, {{191172, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59068, {{191175, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60350, {{191179, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59071, {{191182, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59072, {{191183, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59074, {{191190, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{54860, {{191191, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59075, {{191192, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59078, {{191193, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58706, {{191199, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59088, {{191212, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59087, {{191213, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59086, {{191214, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59089, {{191215, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59093, {{191216, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59099, {{191218, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59108, {{191219, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59106, {{191220, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60351, {{191221, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60352, {{191225, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59127, {{191226, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59126, {{191228, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59120, {{191229, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59128, {{191230, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60353, {{191232, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59135, {{191233, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59142, {{191234, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59153, {{191238, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60355, {{191239, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59152, {{191241, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59170, {{191246, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59168, {{191247, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59183, {{191251, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59179, {{191252, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59194, {{191255, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59192, {{191256, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60356, {{191257, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59206, {{191276, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60357, {{191280, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60358, {{191281, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59211, {{191282, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60359, {{191289, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59224, {{191292, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60361, {{191307, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60362, {{191312, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59240, {{191320, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59247, {{191325, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59248, {{191327, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60363, {{191328, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{55897, {{191341, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{59251, {{191345, 1, 0, 0},{191345, 1, 1, 16}, {0, 0, 0, 0}}},{59251, {{191345, 1, 0, 0},{191345, 1, 1, 16}, {0, 0, 0, 0}}},{59252, {{191345, 1, 0, 0},{191345, 1, 1, 17}, {0, 0, 0, 0}}},{59252, {{191345, 1, 0, 0},{191345, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{48391, {{191348, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59253, {{191350, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59254, {{191351, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {4, {{59255, {{191353, 1, 0, 0},{191353, 1, 1, 16}, {0, 0, 0, 0}}},{59255, {{191353, 1, 0, 0},{191353, 1, 1, 16}, {0, 0, 0, 0}}},{59256, {{191353, 1, 0, 0},{191353, 1, 1, 17}, {0, 0, 0, 0}}},{59256, {{191353, 1, 0, 0},{191353, 1, 1, 17}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59261, {{191361, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60365, {{191364, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60366, {{191366, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59265, {{191376, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59264, {{191377, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{58676, {{191392, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59269, {{191398, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59275, {{191408, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59277, {{191409, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59273, {{191410, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59272, {{191411, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59276, {{191412, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59274, {{191413, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60367, {{191417, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59279, {{191418, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59303, {{191419, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59280, {{191420, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59282, {{191421, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60368, {{191422, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60370, {{191427, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60372, {{191428, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60371, {{191429, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60373, {{191431, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60374, {{191432, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60377, {{191435, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60376, {{191436, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60375, {{191438, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{59284, {{191447, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {0, {{-1, {{0, 0, 0, 0}}}}},
    {1, {{60379, {{191451, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59291, {{191452, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59290, {{191453, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59289, {{191454, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{59287, {{191455, 1, 0, 0}, {0, 0, 0, 0}}}}},
    {1, {{60380, {{191456, 1, 0, 0}, {0, 0, 0, 0}}}}}
};

static const URangeToMJMappings urange_to_mj_mappings[34] = {
    {12293, 12347, chunked_mj_mappings_003005_00303b},
    {13312, 19893, chunked_mj_mappings_003400_004db5},
    {19968, 32414, chunked_mj_mappings_004e00_007e9e},
    {32566, 35743, chunked_mj_mappings_007f36_008b9f},
    {35895, 36124, chunked_mj_mappings_008c37_008d1c},
    {36196, 38020, chunked_mj_mappings_008d64_009484},
    {38263, 40059, chunked_mj_mappings_009577_009c7b},
    {40165, 40478, chunked_mj_mappings_009ce5_009e1e},
    {40565, 40943, chunked_mj_mappings_009e75_009fef},
    {63773, 63856, chunked_mj_mappings_00f91d_00f970},
    {63952, 64109, chunked_mj_mappings_00f9d0_00fa6d},
    {131072, 173782, chunked_mj_mappings_020000_02a6d6},
    {173846, 174221, chunked_mj_mappings_02a716_02a88d},
    {174309, 174312, chunked_mj_mappings_02a8e5_02a8e8},
    {174404, 174710, chunked_mj_mappings_02a944_02aa76},
    {174778, 175393, chunked_mj_mappings_02aaba_02ad21},
    {175497, 176407, chunked_mj_mappings_02ad89_02b117},
    {176488, 177606, chunked_mj_mappings_02b168_02b5c6},
    {177671, 177881, chunked_mj_mappings_02b607_02b6d9},
    {177947, 178660, chunked_mj_mappings_02b71b_02b9e4},
    {178733, 178951, chunked_mj_mappings_02ba2d_02bb07},
    {179023, 179408, chunked_mj_mappings_02bb4f_02bcd0},
    {179508, 179666, chunked_mj_mappings_02bd34_02bdd2},
    {179752, 180283, chunked_mj_mappings_02be28_02c03b},
    {180355, 180858, chunked_mj_mappings_02c083_02c27a},
    {180936, 181768, chunked_mj_mappings_02c2c8_02c608},
    {181872, 182487, chunked_mj_mappings_02c670_02c8d7},
    {182578, 182593, chunked_mj_mappings_02c932_02c941},
    {182657, 182870, chunked_mj_mappings_02c981_02ca56},
    {182950, 183075, chunked_mj_mappings_02caa6_02cb23},
    {183173, 183453, chunked_mj_mappings_02cb85_02cc9d},
    {183535, 183678, chunked_mj_mappings_02ccef_02cd7e},
    {183743, 183802, chunked_mj_mappings_02cdbf_02cdfa},
    {183869, 191456, chunked_mj_mappings_02ce3d_02ebe0}
};

static const MJShrinkMappingUnicodeSet mj_shrink_mappings[68103] = {
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {12293,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {12294,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {12347,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {19992,-1},
        {19992,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {13314,-1},
        {-1},
        {-1,-1},
        {21916,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29275,20117},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29275,20117},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20116,-1},
        {20116,-1,-1}
    },
    {
        {13318,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33394,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33394,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20081,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20167,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25062,-1},
        {25062,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31532,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20104,25234},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20117,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23451,20133},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23451,20133},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {13356,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24297,-1},
        {24297,-1,-1}
    },
    {
        {13358,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {13358,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {13358,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24248,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20449,-1},
        {20449,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23571,29246},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20239,-1},
        {20239,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20800,-1},
        {20800,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25856,-1},
        {25856,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21269,21270},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21269,21270},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27424,-1},
        {27424,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20193,-1},
        {20193,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20049,-1},
        {20049,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25327,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25305,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25305,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20223,-1},
        {20223,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20223,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25303,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {64048,20398},
        {64048,20398,-1}
    },
    {
        {-1,-1},
        {-1},
        {20316,-1},
        {20316,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24627,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22809,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20134,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20279,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23593,-1},
        {23593,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32786,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24465,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20681,-1},
        {20681,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20426,-1},
        {20426,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36523,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20215,-1},
        {20215,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37027,-1},
        {37027,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24730,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25240,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23221,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20853,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33495,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33495,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25147,25150},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24760,-1},
        {-1,-1,-1}
    },
    {
        {13416,-1},
        {-1},
        {-1,-1},
        {20451,-1,-1}
    },
    {
        {13418,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24818,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38914,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25545,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20405,-1},
        {20405,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23241,-1},
        {23241,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24894,-1},
        {24894,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24907,-1},
        {24907,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20506,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20250,26371},
        {20250,26371,-1}
    },
    {
        {-1,-1},
        {-1},
        {27154,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30053,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20636,-1},
        {20636,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24034,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25688,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33879,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24336,36019},
        {24336,36019,-1}
    },
    {
        {-1,-1},
        {-1},
        {25754,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32725,-1},
        {32725,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25006,-1},
        {25006,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27204,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24797,-1},
        {24797,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28544,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28544,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21422,-1},
        {24285,21422,-1}
    },
    {
        {-1,-1},
        {-1},
        {29026,-1},
        {15808,29026,-1}
    },
    {
        {-1,-1},
        {-1},
        {20650,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20689,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21295,-1},
        {-1,-1,-1}
    },
    {
        {13458,-1},
        {-1},
        {-1,-1},
        {20693,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36948,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25793,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34261,-1},
        {34261,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31337,31311},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25826,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33287,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24173,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20737,-1},
        {20737,20426,-1}
    },
    {
        {-1,-1},
        {-1},
        {35018,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28707,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32327,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24948,25086},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28739,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20769,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20791,-1},
        {20791,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20791,-1},
        {20791,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20185,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35064,26804},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37675,-1},
        {37675,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26081,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26118,-1},
        {26118,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29121,-1},
        {29121,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20840,-1},
        {20840,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36001,-1},
        {36001,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20001,20841},
        {20841,20001,-1}
    },
    {
        {-1,-1},
        {-1},
        {20186,-1},
        {-1,-1,-1}
    },
    {
        {13493,-1},
        {-1},
        {-1,-1},
        {30339,35980,-1}
    },
    {
        {-1,-1},
        {-1},
        {20844,-1},
        {20844,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36930,27507},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36930,27507},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39002,-1},
        {39002,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39002,-1},
        {39002,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32596,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20882,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {13500,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31646,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33913,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31852,-1},
        {-1,-1,-1}
    },
    {
        {13505,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23495,-1},
        {23495,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35435,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35435,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27647,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27712,-1},
        {-1,-1,-1}
    },
    {
        {13511,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27963,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27934,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27946,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27940,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28024,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27934,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28117,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28076,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35613,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28357,-1},
        {28357,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28500,-1},
        {28500,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28500,-1},
        {28500,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29578,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20982,-1},
        {20982,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21151,-1},
        {-1,-1,-1}
    },
    {
        {13531,-1},
        {-1},
        {-1,-1},
        {20999,21151,-1}
    },
    {
        {-1,-1},
        {-1},
        {24188,-1},
        {24188,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21009,-1},
        {21009,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22865,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22865,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22865,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30383,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29623,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24355,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21038,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21106,-1},
        {21106,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21180,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21547,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21050,-1},
        {21050,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35284,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26391,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21043,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38824,-1},
        {38824,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40304,38613},
        {24427,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21119,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20842,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25153,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21273,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32782,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37716,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37716,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37733,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37733,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21139,-1},
        {21139,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {133127,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37613,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21127,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21083,-1},
        {21083,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21049,-1},
        {21049,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23815,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29312,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24433,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37159,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31932,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32118,-1},
        {32118,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21117,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32901,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26369,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20150,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30435,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37177,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31639,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27377,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30922,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21049,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21169,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24343,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40668,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21071,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21187,-1},
        {21187,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21187,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20841,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21097,20054},
        {-1,-1,-1}
    },
    {
        {13599,-1},
        {-1},
        {-1,-1},
        {21218,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27463,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27608,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30078,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23529,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30130,-1},
        {30130,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39178,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32972,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30413,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40565,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23586,-1},
        {24038,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21331,-1},
        {21331,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21273,-1},
        {21273,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38627,33021},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24658,24646},
        {24658,24646,-1}
    },
    {
        {-1,-1},
        {-1},
        {24658,24646},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24254,-1},
        {24254,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24254,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31755,-1},
        {31755,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30064,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30064,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24061,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36123,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38475,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20162,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {19990,-1},
        {19990,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {19990,-1},
        {19990,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {19990,-1},
        {19990,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {19989,-1},
        {19989,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21347,-1},
        {21347,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21353,-1},
        {21353,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22863,-1},
        {22863,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21371,21364},
        {21371,21364,-1}
    },
    {
        {-1,-1},
        {-1},
        {21535,-1},
        {21535,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30757,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21400,21283},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23621,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20301,22311},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25158,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37329,30154},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21402,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38593,22534},
        {38593,22534,-1}
    },
    {
        {-1,-1},
        {-1},
        {30413,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20905,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21416,24282},
        {21416,24282,-1}
    },
    {
        {-1,-1},
        {-1},
        {20736,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24283,-1},
        {24283,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24283,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24187,-1},
        {24187,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32032,-1},
        {32032,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35480,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30388,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29226,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29243,21451},
        {21451,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29243,21451},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21490,-1},
        {21490,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21490,-1},
        {21490,-1,-1}
    },
    {
        {13661,-1},
        {-1},
        {-1,-1},
        {20107,-1,-1}
    },
    {
        {13662,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21919,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38617,21452},
        {38617,21452,-1}
    },
    {
        {-1,-1},
        {-1},
        {22737,-1},
        {22737,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22737,-1},
        {22737,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21473,-1},
        {-1,-1,-1}
    },
    {
        {13667,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21492,-1},
        {21492,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21498,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26098,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21485,24070},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21673,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30472,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26041,20223},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35359,-1},
        {35359,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21676,-1},
        {-1,-1,-1}
    },
    {
        {13678,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21558,-1},
        {21558,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21558,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22043,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21536,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21642,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21589,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27785,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26106,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20237,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30764,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22210,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33016,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21648,20197},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24459,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21533,-1},
        {21533,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35436,21564},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27954,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26354,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21665,-1},
        {21665,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25423,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21666,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27448,-1},
        {27448,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23792,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21585,21780},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21840,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36763,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23839,-1},
        {23839,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21767,-1},
        {21767,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25480,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33099,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24199,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28085,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21693,-1},
        {-1,-1,-1}
    },
    {
        {13734,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21563,-1},
        {-1,-1,-1}
    },
    {
        {13736,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22014,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22495,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21766,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21693,-1},
        {21693,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21693,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21884,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22129,-1},
        {22129,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22132,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26265,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25545,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35584,-1},
        {35584,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21666,-1},
        {21666,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26928,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21606,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40822,-1},
        {40822,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30465,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21563,-1},
        {21563,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35646,-1},
        {35646,-1,-1}
    },
    {
        {13765,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30340,21981},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26247,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32092,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38647,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21897,-1},
        {21897,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33119,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27007,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21969,-1},
        {21969,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27084,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34388,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34388,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40221,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21847,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21847,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33162,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35589,-1},
        {-1,-1,-1}
    },
    {
        {13786,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31505,-1},
        {31505,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40773,-1},
        {21957,40773,-1}
    },
    {
        {-1,-1},
        {-1},
        {30772,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24927,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22164,-1},
        {22164,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36441,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32842,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33187,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32043,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {14428,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29241,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34763,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {131953,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22612,-1},
        {22612,-1,-1}
    },
    {
        {13812,-1},
        {-1},
        {-1,-1},
        {40791,-1,-1}
    },
    {
        {13812,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38911,-1},
        {38911,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38772,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {13812,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {13829,-1},
        {-1},
        {-1,-1},
        {37532,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23992,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31001,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31680,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31013,-1},
        {31013,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22129,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25130,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31022,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38583,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35733,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21898,23146},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38996,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23884,-1},
        {23884,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35728,35726},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22234,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22234,-1},
        {22234,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24517,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26585,-1},
        {26585,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20809,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22291,20870},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28149,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22254,-1},
        {22254,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21151,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24207,-1},
        {24207,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32791,-1},
        {32791,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31357,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20315,20175},
        {20315,22343,20175}
    },
    {
        {-1,-1},
        {-1},
        {22387,-1},
        {22387,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26015,-1},
        {26015,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22349,-1},
        {22349,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23611,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {19992,22389},
        {19992,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22352,-1},
        {22352,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22352,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26688,-1},
        {26688,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33406,-1},
        {33406,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22352,-1},
        {22352,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22578,-1},
        {22578,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28037,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31350,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29287,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20812,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26809,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38491,-1},
        {38491,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38491,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20445,-1},
        {20445,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {64074,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29733,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33094,-1},
        {33094,-1,-1}
    },
    {
        {13898,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26994,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22428,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23431,-1},
        {23431,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37326,-1},
        {37326,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22575,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38867,-1},
        {38867,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22602,-1},
        {22602,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23592,22640},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26771,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25615,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25615,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33260,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27141,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28466,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32581,-1},
        {32581,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28414,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29858,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30079,-1},
        {30079,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25769,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22094,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28503,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22432,-1},
        {22432,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22518,-1},
        {22518,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34809,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34399,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22725,-1},
        {22725,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25788,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25812,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21488,33274},
        {21488,33274,-1}
    },
    {
        {-1,-1},
        {-1},
        {22530,-1},
        {22530,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28629,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22818,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36917,39319},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24283,-1},
        {24283,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22744,22593},
        {22744,22593,-1}
    },
    {
        {-1,-1},
        {-1},
        {30830,22470},
        {30992,30830,-1}
    },
    {
        {-1,-1},
        {-1},
        {34407,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32761,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22777,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {63964,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22797,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21319,-1},
        {33307,21319,-1}
    },
    {
        {-1,-1},
        {-1},
        {23348,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23348,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22877,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22821,-1},
        {22821,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22871,-1},
        {22871,-1,-1}
    },
    {
        {13969,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {13969,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22846,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22856,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24184,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24380,-1},
        {-1,-1,-1}
    },
    {
        {13974,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26619,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20276,25206},
        {-1,-1,-1}
    },
    {
        {13977,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21517,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22495,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22862,-1},
        {22862,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26862,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31407,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22855,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22900,-1},
        {64048,20398,-1}
    },
    {
        {-1,-1},
        {-1},
        {22900,-1},
        {22900,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23011,-1},
        {23011,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22908,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22904,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23019,-1},
        {23019,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22901,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22992,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31443,22908},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20235,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32940,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20887,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32986,-1},
        {32986,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22389,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22993,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21654,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22985,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26570,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22994,-1},
        {22994,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23341,-1},
        {23341,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26611,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23381,-1},
        {23381,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22943,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23079,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24214,-1},
        {21253,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24614,-1},
        {24614,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23012,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22865,23094},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20381,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21665,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22428,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24860,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26784,-1},
        {-1,-1,-1}
    },
    {
        {14031,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23234,-1},
        {23234,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31922,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23022,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21775,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28925,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23030,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30700,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21745,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22974,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22471,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23072,-1},
        {23072,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24736,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23196,21764},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25397,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36367,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23721,22956},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34595,23172},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21813,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28117,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23234,-1},
        {23234,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23234,-1},
        {23234,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23081,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23130,-1},
        {23130,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29032,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21883,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24745,24817},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31281,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23878,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23305,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26985,23236},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23043,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25554,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25554,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22989,-1},
        {22989,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29326,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20449,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39087,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25616,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23063,36855},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23304,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23162,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {136966,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30910,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21964,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27580,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22616,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {18613,37142},
        {18613,37142,-1}
    },
    {
        {-1,-1},
        {-1},
        {30929,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23110,-1},
        {23110,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33281,23138},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22612,23030},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22024,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22667,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27165,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25693,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23280,23273},
        {23273,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20658,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27141,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29123,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30969,23254},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23202,36023},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {136966,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26032,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23172,-1},
        {23172,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27323,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23085,-1},
        {23305,23085,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31019,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23348,-1},
        {23348,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23348,-1},
        {23348,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23363,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23243,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24396,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23420,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23333,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23372,-1},
        {23372,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23018,-1},
        {23018,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23376,-1},
        {23376,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23401,-1},
        {23401,21683,-1}
    },
    {
        {-1,-1},
        {-1},
        {22909,-1},
        {22909,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23081,-1},
        {23081,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32020,-1},
        {32020,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27030,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40329,-1},
        {40329,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23423,-1},
        {23423,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23432,-1},
        {23432,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23425,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30106,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20889,23531},
        {20889,23531,-1}
    },
    {
        {-1,-1},
        {-1},
        {23450,-1},
        {23450,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20171,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23481,-1},
        {23481,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26049,-1},
        {26049,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23621,-1},
        {23621,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20985,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23527,-1},
        {23527,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23452,-1},
        {23452,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33704,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35348,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23493,-1},
        {23493,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33690,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23487,-1},
        {23487,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26132,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33787,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23462,20126},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20882,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20901,-1},
        {20901,-1,-1}
    },
    {
        {14177,-1},
        {-1},
        {-1,-1},
        {26368,-1,-1}
    },
    {
        {14178,-1},
        {-1},
        {-1,-1},
        {23431,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33884,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20883,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23504,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31337,31311},
        {31337,31311,-1}
    },
    {
        {-1,-1},
        {-1},
        {33998,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {14187,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {14188,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24826,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31422,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34126,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22818,-1},
        {22818,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23522,23517},
        {23522,23517,-1}
    },
    {
        {-1,-1},
        {-1},
        {21011,-1},
        {21011,-1,-1}
    },
    {
        {14197,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24471,-1},
        {24471,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27465,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20677,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36802,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23592,24038},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36315,-1},
        {36315,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28784,35927},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36864,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22949,36918},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35897,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23600,25513},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26524,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36864,30031},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36941,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25619,39319},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21453,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23381,-1},
        {23381,-1,-1}
    },
    {
        {14221,-1},
        {-1},
        {-1,-1},
        {23611,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23621,-1},
        {23621,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23614,-1},
        {23614,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23635,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26088,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23633,-1},
        {23633,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {138326,26431},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23630,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35930,-1},
        {35930,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23615,-1},
        {23615,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23626,33279},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23653,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23656,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22880,22666},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23637,-1},
        {23637,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20043,-1},
        {20043,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33452,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23697,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23574,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23676,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20250,26371},
        {20250,26371,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23888,-1},
        {23888,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36208,-1},
        {36208,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20186,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20986,-1},
        {20986,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23674,-1},
        {23674,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26105,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24545,23724},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {138657,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23735,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23735,-1},
        {23735,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21576,23718},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21576,23718},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23714,-1},
        {23714,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26063,-1},
        {26063,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21495,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {19998,-1},
        {19998,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25097,23735},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23612,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23696,-1},
        {23696,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30409,23731},
        {-1,-1,-1}
    },
    {
        {14273,-1},
        {-1},
        {-1,-1},
        {23736,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21496,-1},
        {21496,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {138657,21673},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20807,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {138657,25387},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29699,23821},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31362,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23842,26740},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23592,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22476,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25404,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35901,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21745,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23844,21775},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35901,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21825,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23696,-1},
        {23696,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31645,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24357,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22521,-1},
        {22521,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30858,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24055,-1},
        {24055,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24055,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23713,-1},
        {23713,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31646,-1},
        {-1,-1,-1}
    },
    {
        {14306,-1},
        {-1},
        {-1,-1},
        {23822,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26248,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {14312,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23735,-1},
        {23735,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21890,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27849,-1},
        {27849,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21471,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23735,-1},
        {23735,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24142,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31957,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23236,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {13786,-1},
        {-1,-1,-1}
    },
    {
        {14324,-1},
        {-1},
        {-1,-1},
        {23916,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23713,-1},
        {23713,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21407,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23959,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23937,-1},
        {23937,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22649,-1},
        {22649,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {14333,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22665,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {14336,-1},
        {-1},
        {-1,-1},
        {23798,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22007,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24489,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22077,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23601,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30628,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24162,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24162,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23943,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23884,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38598,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28503,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22057,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23968,-1},
        {23968,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32618,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {14333,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34204,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23916,-1},
        {23916,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23916,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23980,18919},
        {23980,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23988,-1},
        {23988,22885,22887}
    },
    {
        {-1,-1},
        {-1},
        {23988,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23996,-1},
        {23996,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23844,-1},
        {23844,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23565,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24017,-1},
        {24017,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23980,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23916,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22744,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31150,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38753,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27421,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33618,33426},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20116,-1},
        {24040,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23637,-1},
        {23637,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32923,-1},
        {-1,-1,-1}
    },
    {
        {14383,-1},
        {-1},
        {-1,-1},
        {20197,-1,-1}
    },
    {
        {14383,-1},
        {-1},
        {-1,-1},
        {24050,-1,-1}
    },
    {
        {14383,-1},
        {-1},
        {-1,-1},
        {20197,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31637,26462},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21489,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33470,-1},
        {-1,-1,-1}
    },
    {
        {14390,-1},
        {-1},
        {-1,-1},
        {24070,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26497,24084},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35122,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21540,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24076,-1},
        {24076,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24571,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26570,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24081,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21616,-1},
        {-1,-1,-1}
    },
    {
        {14400,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26803,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26803,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24642,-1},
        {24642,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36626,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24727,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24358,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23846,-1},
        {-1,-1,-1}
    },
    {
        {24125,-1},
        {-1},
        {24125,-1},
        {24125,-1,-1}
    },
    {
        {24125,-1},
        {-1},
        {24125,-1},
        {24125,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32379,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25539,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25579,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26991,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35084,-1},
        {35084,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23162,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22830,-1},
        {22830,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24162,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32273,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27083,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24908,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24908,-1},
        {-1,-1,-1}
    },
    {
        {14428,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23937,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22024,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24128,-1},
        {24128,-1,-1}
    },
    {
        {14433,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24148,-1},
        {24148,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24148,-1},
        {24148,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40845,-1},
        {31452,40845,-1}
    },
    {
        {-1,-1},
        {-1},
        {38577,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27315,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24128,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25082,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24018,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26373,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24190,-1},
        {24190,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32153,32380},
        {32153,32380,-1}
    },
    {
        {-1,-1},
        {-1},
        {32153,32380},
        {32153,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24190,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23429,-1},
        {23429,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23431,-1},
        {23431,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28792,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33538,25100},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24213,-1},
        {24213,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31435,24196},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20196,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21482,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24202,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30141,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30136,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24278,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26355,24254},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24235,32895},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30165,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26021,-1},
        {26021,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24202,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24215,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30165,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30168,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25354,24324},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24283,-1},
        {24283,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {131236,22534},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21412,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20506,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24249,33521},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24254,33279},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36926,24254},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {131236,22534},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24300,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38556,-1},
        {38556,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24305,38605},
        {24305,38605,-1}
    },
    {
        {-1,-1},
        {-1},
        {23529,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20689,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24300,63798},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20745,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34311,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24281,32764},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24310,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22888,22892},
        {22888,22892,-1}
    },
    {
        {-1,-1},
        {-1},
        {24330,-1},
        {24330,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24330,-1},
        {24330,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25806,-1},
        {25806,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25806,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21688,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24059,21367},
        {24059,21367,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30340,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24344,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24347,-1},
        {24347,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24372,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38466,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24061,-1},
        {24061,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26053,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35199,-1},
        {35199,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35199,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24381,-1},
        {24381,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24380,-1},
        {24380,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24380,-1},
        {24380,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24358,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39180,-1},
        {39180,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30924,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20633,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38548,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39739,-1},
        {39739,31909,-1}
    },
    {
        {-1,-1},
        {-1},
        {30962,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24396,24357},
        {24396,24357,-1}
    },
    {
        {-1,-1},
        {-1},
        {24076,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24076,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32902,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24037,-1},
        {24037,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20025,-1},
        {20025,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20073,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21488,33274},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30347,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38745,38748},
        {38745,38748,-1}
    },
    {
        {-1,-1},
        {-1},
        {22793,35722},
        {22793,35722,-1}
    },
    {
        {-1,-1},
        {-1},
        {23481,-1},
        {23481,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26366,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31922,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37194,20163},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23449,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20375,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20250,26371},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20342,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20319,-1},
        {20319,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20452,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20448,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36368,36341},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24453,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36918,-1},
        {36918,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21205,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26987,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26934,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24713,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20661,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24494,-1},
        {24494,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36948,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20739,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {14586,-1},
        {-1},
        {-1,-1},
        {24515,-1,-1}
    },
    {
        {14586,-1},
        {-1},
        {-1,-1},
        {24515,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24542,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24525,-1},
        {24525,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24525,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24594,24661},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25184,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24541,-1},
        {24541,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24613,-1},
        {24613,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22952,-1},
        {22952,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24859,-1},
        {24859,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24859,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20206,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25034,-1},
        {25034,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26514,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32027,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29976,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24039,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24661,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24682,20319},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24599,-1},
        {24599,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20305,24601},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24560,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20307,24629},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24612,-1},
        {24612,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25305,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35421,-1},
        {-1,-1,-1}
    },
    {
        {14615,-1},
        {-1},
        {-1,-1},
        {27888,-1,-1}
    },
    {
        {14615,-1},
        {-1},
        {-1,-1},
        {27888,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24681,-1},
        {24681,-1,-1}
    },
    {
        {14618,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24641,-1},
        {24641,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20376,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24656,-1},
        {24656,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28872,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20363,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20369,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24651,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27945,20415},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20357,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20351,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22434,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20379,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24682,-1},
        {24682,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24616,-1},
        {24616,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24666,-1},
        {24666,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29378,29791},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28913,24683},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24589,-1},
        {24589,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26794,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24713,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24684,26734},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24904,-1},
        {24904,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24688,27995},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28931,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25937,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24908,-1},
        {24908,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24908,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24616,21365},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25506,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20474,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24499,24503},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24499,24755},
        {24499,24755,24503}
    },
    {
        {-1,-1},
        {-1},
        {24471,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24758,-1},
        {24758,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25480,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26851,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26855,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24821,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20493,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24736,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25516,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24803,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20494,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24616,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24962,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29468,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20506,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24976,-1},
        {24976,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32865,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26997,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28205,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24365,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38538,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25534,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24617,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29494,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24800,24693},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30684,-1},
        {30684,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20598,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24860,-1},
        {24860,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38918,-1},
        {38918,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {134756,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20566,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20621,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {14703,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {14703,-1},
        {-1},
        {-1,-1},
        {14703,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24753,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24753,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24892,24910},
        {24892,24910,-1}
    },
    {
        {-1,-1},
        {-1},
        {24895,-1},
        {24895,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35115,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24897,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22622,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38292,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31881,24575},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27068,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23267,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36012,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26289,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24928,-1},
        {24928,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29090,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36899,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24120,-1},
        {24120,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28411,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40637,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28409,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25693,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25087,-1},
        {25087,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29121,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27635,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24853,-1},
        {24853,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30059,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20686,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27186,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25000,-1},
        {25000,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35400,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29123,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25045,-1},
        {25045,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {63964,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25059,-1},
        {25059,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35167,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36776,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31004,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39514,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35980,25031},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29182,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35981,25031},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27138,27005},
        {27138,27005,-1}
    },
    {
        {-1,-1},
        {-1},
        {15877,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {15877,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34910,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25087,-1},
        {25087,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24808,24920},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23358,-1},
        {23358,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {14703,-1},
        {14703,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29166,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25886,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24604,-1},
        {24604,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25108,37549},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25108,37549},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25110,-1},
        {25110,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {141711,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27494,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21705,-1},
        {21705,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26685,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21191,-1},
        {21191,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25119,-1},
        {25119,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26140,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36785,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35186,-1},
        {35186,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33282,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25144,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25158,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38356,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23626,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25910,21454},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25187,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {13531,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25187,26623},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25285,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24341,-1},
        {24341,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27773,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26623,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29366,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27611,-1},
        {27611,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20223,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26514,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25235,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25214,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26611,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38440,25212},
        {25212,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25209,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35390,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25300,25244},
        {25244,25300,-1}
    },
    {
        {-1,-1},
        {-1},
        {25300,25244},
        {25244,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26680,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28895,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26666,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20805,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27927,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25516,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26803,-1},
        {26803,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22434,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29678,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22429,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25360,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25360,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25293,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26675,25335},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25273,23435},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25198,25295},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35025,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21585,25324},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26831,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27795,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20469,-1},
        {20469,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25556,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24478,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25254,-1},
        {26564,25254,-1}
    },
    {
        {-1,-1},
        {-1},
        {26893,-1},
        {-1,-1,-1}
    },
    {
        {25536,-1},
        {-1},
        {26847,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25361,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29747,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25516,-1},
        {25516,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26965,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26990,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29326,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28845,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35091,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25662,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25788,-1},
        {25788,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27001,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32251,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25463,-1},
        {25463,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25463,-1},
        {25463,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24495,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33600,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25818,34388},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27071,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27040,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28343,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25793,-1},
        {25793,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28377,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32013,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {14963,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30042,25773},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25277,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27045,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27124,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20828,-1},
        {20828,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25313,25844},
        {25313,25844,-1}
    },
    {
        {-1,-1},
        {-1},
        {25313,25844},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28414,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25594,25622},
        {25594,25622,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24314,-1},
        {24314,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25539,-1},
        {25539,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25827,25623},
        {25827,25623,-1}
    },
    {
        {-1,-1},
        {-1},
        {33393,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27154,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25451,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25731,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25504,-1},
        {25504,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21103,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27171,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25312,25818},
        {25312,25818,-1}
    },
    {
        {-1,-1},
        {-1},
        {25312,25818},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25312,25818},
        {25312,25818,-1}
    },
    {
        {-1,-1},
        {-1},
        {25652,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25277,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20710,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30059,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25411,-1},
        {25411,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25289,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28506,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36027,25861},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31574,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25810,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25802,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28593,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27263,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25658,25884},
        {25658,25884,-1}
    },
    {
        {-1,-1},
        {-1},
        {22622,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27308,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25718,-1},
        {25718,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25900,25829},
        {25829,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22741,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30922,25299},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35178,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25582,-1},
        {25582,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25975,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25658,25884},
        {25658,25884,-1}
    },
    {
        {-1,-1},
        {-1},
        {25169,25778},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24935,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28481,25847},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25847,-1},
        {25847,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25135,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25785,25898},
        {-1,-1,-1}
    },
    {
        {14958,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {14958,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25516,25644},
        {25516,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34299,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27402,-1},
        {25331,-1,-1}
    },
    {
        {14963,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27403,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29550,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30089,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27495,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25906,-1},
        {25906,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22810,-1},
        {22810,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21201,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35913,136775},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20062,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25297,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20062,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25959,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24343,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26356,-1},
        {26356,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27664,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23553,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37059,-1},
        {37059,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32676,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25421,25182},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21192,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21187,-1},
        {21187,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25933,25150},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25906,22852},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27584,20818},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21097,20054},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25955,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25119,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25933,21093},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25986,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25121,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25571,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26604,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25958,-1},
        {25958,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27465,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {14188,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25422,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25970,-1},
        {25970,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25985,30050},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25985,30050},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35248,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23493,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27481,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32341,-1},
        {32341,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32901,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25139,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25138,30439},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25369,25831},
        {25369,25831,-1}
    },
    {
        {-1,-1},
        {-1},
        {40779,25998},
        {40779,25998,-1}
    },
    {
        {-1,-1},
        {-1},
        {29773,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21011,25238},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26009,21028},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26011,-1},
        {26011,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26007,-1},
        {26007,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26015,-1},
        {26015,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25976,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22893,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23423,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26027,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25152,-1},
        {25152,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35362,21508},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25240,-1},
        {25240,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40718,-1},
        {40718,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26029,26039},
        {26029,26039,-1}
    },
    {
        {-1,-1},
        {-1},
        {24285,24246},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20547,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26049,-1},
        {26049,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26071,-1},
        {26071,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26051,-1},
        {26051,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24509,25582},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26075,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30336,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26088,-1},
        {26088,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24318,-1},
        {24318,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27683,27671},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33390,-1},
        {33390,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20809,-1},
        {20809,-1,-1}
    },
    {
        {15062,-1},
        {-1},
        {-1,-1},
        {26088,-1,-1}
    },
    {
        {15063,-1},
        {-1},
        {-1,-1},
        {21402,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21542,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32932,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21610,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26157,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30036,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30036,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20908,-1},
        {20908,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21631,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26166,-1},
        {26166,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20901,-1},
        {20901,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33065,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33065,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26292,-1},
        {26292,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26149,-1},
        {26149,-1,-1}
    },
    {
        {15082,-1},
        {-1},
        {-1,-1},
        {33282,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38997,39023},
        {38997,39023,-1}
    },
    {
        {-1,-1},
        {-1},
        {30516,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20880,20882},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26390,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36196,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21745,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26216,-1},
        {26216,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21742,-1},
        {21742,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21742,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33107,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26221,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26399,-1},
        {26399,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26399,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {143812,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26132,-1},
        {26132,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23832,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26210,26211},
        {26211,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {13736,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27965,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21889,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28849,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26364,-1},
        {26364,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28259,-1},
        {28259,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21914,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21986,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21986,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26262,-1},
        {26262,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24962,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24962,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21897,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26223,-1},
        {26223,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26333,-1},
        {26333,-1,-1}
    },
    {
        {15118,-1},
        {-1},
        {-1,-1},
        {38997,39023,-1}
    },
    {
        {-1,-1},
        {-1},
        {21981,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26185,26187},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34770,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30623,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22032,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26203,-1},
        {26203,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22063,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26256,-1},
        {26256,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {15130,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {15132,-1},
        {-1},
        {-1,-1},
        {26185,26187,-1}
    },
    {
        {-1,-1},
        {-1},
        {22092,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22092,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26253,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29022,-1},
        {63856,27578,29022}
    },
    {
        {-1,-1},
        {-1},
        {26148,-1},
        {26148,-1,-1}
    },
    {
        {15138,-1},
        {-1},
        {-1,-1},
        {26342,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23565,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38576,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26292,-1},
        {26292,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32397,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32397,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32397,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35711,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30067,-1},
        {30090,30067,-1}
    },
    {
        {-1,-1},
        {-1},
        {29141,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26332,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {64104,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28699,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33278,-1},
        {33278,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26367,-1},
        {26367,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34720,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32925,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26376,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20132,-1},
        {20132,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28009,-1},
        {28009,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {134469,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30556,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24908,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36665,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36665,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26474,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32177,-1},
        {32177,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {144338,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27414,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25225,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33545,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26512,-1},
        {26512,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26834,-1},
        {26834,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26713,-1},
        {26713,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32796,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26544,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25342,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25345,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26053,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27966,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26800,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26805,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25368,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24685,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26692,27927},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26530,30791},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21009,15213},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26928,-1},
        {26928,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25429,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25387,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {15213,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25334,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26928,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22068,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21283,-1},
        {21283,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26497,-1},
        {26497,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27179,-1},
        {27179,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27179,-1},
        {27179,-1,-1}
    },
    {
        {15223,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26778,-1},
        {-1},
        {26778,-1},
        {-1,-1,-1}
    },
    {
        {26778,-1},
        {-1},
        {26778,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24358,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25513,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26848,-1},
        {26848,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27007,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26594,-1},
        {26594,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25496,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26114,26829},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20989,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26741,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27036,-1},
        {27036,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26894,-1},
        {26894,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27377,-1},
        {27377,-1,-1}
    },
    {
        {15239,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {15240,-1},
        {-1},
        {-1,-1},
        {22856,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26969,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26825,-1},
        {26825,-1,-1}
    },
    {
        {15245,-1},
        {-1},
        {-1,-1},
        {31627,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26479,-1},
        {26479,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26608,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20567,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25554,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20120,20121},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27238,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26669,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32217,27135},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30684,-1},
        {30684,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26969,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24853,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26647,-1},
        {21828,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21273,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27047,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21371,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35566,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26853,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {15268,-1},
        {-1},
        {-1,-1},
        {26494,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26943,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32034,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27170,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31287,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29060,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27153,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27029,30606},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26976,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26708,-1},
        {26708,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26829,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26862,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27251,26389},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33368,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26676,-1},
        {26676,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26676,-1},
        {26676,-1,-1}
    },
    {
        {15286,-1},
        {-1},
        {-1,-1},
        {27082,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31293,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36865,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36865,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32251,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26412,-1},
        {26412,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25620,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26996,-1},
        {-1,-1,-1}
    },
    {
        {15299,-1},
        {-1},
        {-1,-1},
        {26748,28422,-1}
    },
    {
        {-1,-1},
        {-1},
        {25160,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28411,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26768,-1},
        {26768,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27157,-1},
        {27157,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27170,23267},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31310,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26941,-1},
        {26941,-1,-1}
    },
    {
        {15309,-1},
        {-1},
        {-1,-1},
        {27123,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26853,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39336,-1},
        {39336,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27234,-1},
        {27234,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27067,-1},
        {27067,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27635,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25749,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26831,-1},
        {26831,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28566,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25764,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22778,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26890,-1},
        {26890,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28510,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40595,-1},
        {40595,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29309,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21263,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25776,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26875,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26840,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26361,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26826,-1},
        {26826,26827,-1}
    },
    {
        {-1,-1},
        {-1},
        {27386,-1},
        {27386,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25818,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24314,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24314,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27347,-1},
        {27347,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27315,-1},
        {-1,-1,-1}
    },
    {
        {15344,-1},
        {-1},
        {-1,-1},
        {26893,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27216,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35167,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27296,-1},
        {27296,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27296,-1},
        {27296,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27192,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20760,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25829,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27216,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27133,-1},
        {27133,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27133,-1},
        {27133,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29896,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26584,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28683,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28666,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38875,-1},
        {38875,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27035,-1},
        {27035,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26809,27315},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27063,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27379,-1},
        {27379,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27379,-1},
        {27379,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {15375,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27234,-1},
        {27234,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34306,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27397,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27414,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36510,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29351,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27414,-1},
        {27414,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27117,-1},
        {27117,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24413,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25892,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31852,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27359,-1},
        {27359,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21988,-1},
        {21988,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20864,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27439,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28810,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27427,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27431,-1},
        {-1,-1,-1}
    },
    {
        {15398,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21645,-1},
        {21645,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30733,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34439,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23553,15398},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {15398,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29920,27449},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21542,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27448,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27459,-1},
        {27459,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27468,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24350,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27462,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21330,27436},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22036,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27515,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27483,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21688,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21466,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40723,27439},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27472,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37118,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25963,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27448,38622},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27454,-1},
        {27454,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24247,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39154,39150},
        {39154,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25087,-1},
        {25087,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27476,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27466,21959},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27454,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37172,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21956,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22617,-1},
        {22617,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32886,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {18188,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27488,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27488,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27491,-1},
        {27491,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35382,-1},
        {35382,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24112,27512},
        {27512,24112,-1}
    },
    {
        {-1,-1},
        {-1},
        {22499,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21957,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39768,-1},
        {39768,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21692,26429},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27544,27531},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24067,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21169,27515},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30764,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30772,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36764,-1},
        {36764,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37521,27530},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34597,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30869,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30871,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27566,-1},
        {27566,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28259,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29001,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27592,-1},
        {27592,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30929,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28371,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30920,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34720,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21215,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30951,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33192,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31024,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33242,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32696,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27572,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27521,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27436,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27579,27580},
        {27579,27580,-1}
    },
    {
        {-1,-1},
        {-1},
        {27583,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27584,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25950,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {146072,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27580,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30956,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40725,-1},
        {40725,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40727,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {160730,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21473,-1},
        {21473,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21473,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34006,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31443,32709},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30474,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31443,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32104,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31443,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31885,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27622,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37083,31885},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35039,-1},
        {35039,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25360,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27631,-1},
        {27631,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32791,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33521,34461},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27650,-1},
        {27650,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26608,31885},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32745,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21471,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27656,34928},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21776,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24185,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27622,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27647,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28966,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27631,-1},
        {27631,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30435,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23527,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39715,-1},
        {39715,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29918,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38660,-1},
        {38660,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27684,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27710,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27726,-1},
        {27726,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27825,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28346,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38632,-1},
        {38632,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27973,25214},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27763,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27839,-1},
        {27839,-1,-1}
    },
    {
        {15555,-1},
        {-1},
        {-1,-1},
        {28140,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28046,-1},
        {28046,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27969,-1},
        {27969,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25238,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27803,-1},
        {27803,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25912,-1},
        {25912,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22389,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23724,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24598,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27889,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28101,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28322,-1},
        {28322,-1,-1}
    },
    {
        {15570,-1},
        {-1},
        {-1,-1},
        {27861,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20381,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27969,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28051,-1},
        {28051,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27945,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28107,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27798,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24685,-1},
        {24685,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24685,-1},
        {24685,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24685,-1},
        {24685,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28113,-1},
        {28113,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28113,-1},
        {28113,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25397,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27953,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22349,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27963,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28046,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23615,-1},
        {23615,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26771,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28117,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24760,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20083,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26907,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27754,-1},
        {27754,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28216,-1},
        {28216,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24796,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27947,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28720,-1},
        {28720,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32652,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28203,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28155,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28057,28122},
        {28057,28122,-1}
    },
    {
        {-1,-1},
        {-1},
        {28088,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28088,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28040,-1},
        {28040,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28024,-1},
        {28024,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28169,28041},
        {28041,28169,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28168,28639},
        {28168,28639,-1}
    },
    {
        {-1,-1},
        {-1},
        {28149,-1},
        {28149,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28453,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30468,-1},
        {-1,-1,-1}
    },
    {
        {15633,-1},
        {-1},
        {-1,-1},
        {36961,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28544,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28503,-1},
        {28503,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28668,-1},
        {28668,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26950,28327},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29761,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28856,29024},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28095,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33268,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33268,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24487,-1},
        {24487,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40654,-1},
        {-1,-1,-1}
    },
    {
        {15646,-1},
        {-1},
        {-1,-1},
        {28372,-1,-1}
    },
    {
        {15646,-1},
        {-1},
        {-1,-1},
        {28372,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28085,-1},
        {28085,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23789,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28075,30437},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29060,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26389,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28340,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37089,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27053,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28210,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33026,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33021,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28145,-1},
        {28145,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28145,-1},
        {28145,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30173,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28251,-1},
        {28251,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28243,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28476,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32829,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23487,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24713,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28196,-1},
        {28196,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24908,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34503,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28450,22495},
        {28450,64071,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28564,-1},
        {28009,28564,-1}
    },
    {
        {-1,-1},
        {-1},
        {25774,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27752,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24490,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29148,-1},
        {-1,-1,-1}
    },
    {
        {28567,-1},
        {-1},
        {28567,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28476,-1},
        {28476,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36032,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27832,28227},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28252,-1},
        {28252,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28402,-1},
        {28402,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28500,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24977,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28156,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35028,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35028,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24833,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28546,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31391,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38649,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25794,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28212,28167},
        {28212,28167,-1}
    },
    {
        {15716,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28632,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38568,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29542,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29542,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27287,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {145215,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28046,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28682,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28186,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28110,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40179,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28469,-1},
        {28469,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28418,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20763,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36115,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28520,32858},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25130,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36776,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28729,-1},
        {28729,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28448,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33156,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28681,-1},
        {28511,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20769,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30363,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33184,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33184,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34402,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23416,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28034,27700},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27243,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28366,-1},
        {28366,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32879,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30316,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36996,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30318,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38358,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29243,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29243,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28111,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20789,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31775,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {15770,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28149,-1},
        {28149,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28732,-1},
        {28732,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32397,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39515,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28779,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30402,28789},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28818,-1},
        {28818,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24531,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28851,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {19989,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31180,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25303,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28851,-1},
        {28851,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26611,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25297,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31180,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29180,-1},
        {29180,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39146,-1},
        {39146,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28889,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28925,-1},
        {28925,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25384,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31080,-1},
        {31080,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24726,-1},
        {-1,-1,-1}
    },
    {
        {15808,-1},
        {-1},
        {-1,-1},
        {29026,-1,-1}
    },
    {
        {15808,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31466,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22871,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24776,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28818,-1},
        {28818,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22496,28975},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24674,-1},
        {24674,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25513,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23561,-1},
        {23561,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28954,-1},
        {28954,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24788,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26253,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28954,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29164,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29017,25860},
        {29017,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29032,20598},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24865,-1},
        {-1,-1,-1}
    },
    {
        {15828,-1},
        {-1},
        {-1,-1},
        {28976,15828,-1}
    },
    {
        {15828,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29495,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29031,-1},
        {29031,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20879,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26546,29157},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29006,-1},
        {29006,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25637,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25637,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31910,-1},
        {29136,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29009,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28874,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31296,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28893,-1},
        {28893,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32831,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30956,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25720,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28925,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28925,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27133,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29739,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25691,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29066,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36107,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27511,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {131969,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24970,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25776,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29121,-1},
        {29121,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28975,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28961,23341},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29141,-1},
        {29141,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29200,29865},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28849,-1},
        {28849,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29103,-1},
        {29103,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29159,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29159,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28857,40670},
        {28857,40670,-1}
    },
    {
        {-1,-1},
        {-1},
        {25824,-1},
        {-1,-1,-1}
    },
    {
        {15877,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28966,-1},
        {28966,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29166,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25078,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33200,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25086,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31964,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31964,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28982,-1},
        {28982,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29224,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29229,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25899,-1},
        {25899,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20430,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29240,33596},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33836,-1},
        {33836,19975,-1}
    },
    {
        {-1,-1},
        {-1},
        {34223,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21461,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33108,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26484,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29255,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33026,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26387,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27530,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20633,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26989,-1},
        {26989,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36776,-1},
        {36776,24321,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20114,-1},
        {20114,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40806,-1},
        {40806,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29277,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29281,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29588,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29325,-1},
        {29325,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26623,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25226,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25198,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25238,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26590,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25326,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28895,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25352,29314},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25352,29314},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22434,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29305,23558},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25394,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25384,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29326,-1},
        {29326,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29437,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29283,-1},
        {29283,-1,-1}
    },
    {
        {15935,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29304,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25504,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25492,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26832,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31287,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25556,-1},
        {25556,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25582,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29325,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25609,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28395,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28411,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25688,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27151,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27135,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32883,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27186,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25778,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25747,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27263,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38929,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35938,-1},
        {35938,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25827,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25861,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29374,-1},
        {29374,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29434,40791},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29546,-1},
        {-1,-1,-1}
    },
    {
        {15968,-1},
        {-1},
        {-1,-1},
        {35930,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29431,-1},
        {29431,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26551,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26570,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24623,-1},
        {24623,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29378,-1},
        {-1,-1,-1}
    },
    {
        {15974,-1},
        {-1},
        {-1,-1},
        {35970,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31207,-1},
        {-1,-1,-1}
    },
    {
        {15976,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22839,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29369,20351},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25341,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35992,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35973,-1},
        {35973,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25417,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20414,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25402,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35944,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35980,-1},
        {35980,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24218,29436},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29497,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29479,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25516,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26855,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21646,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25481,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29489,-1},
        {29489,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29785,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {16003,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {16003,-1},
        {-1},
        {-1,-1},
        {16003,29399,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29492,-1},
        {29492,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29508,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31296,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24665,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27087,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23105,-1},
        {23105,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29392,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25678,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36447,-1},
        {36447,-1,-1}
    },
    {
        {16020,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25773,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27204,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29562,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29562,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31328,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27323,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29877,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25826,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27610,-1},
        {27610,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29600,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29854,-1},
        {29854,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21002,29589},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29657,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29589,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20173,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37397,-1},
        {37397,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29611,-1},
        {29611,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21536,29599},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31211,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20973,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29761,-1},
        {29761,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25341,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29808,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35914,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23453,23542},
        {23453,23542,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22487,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25405,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29779,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29688,-1},
        {29688,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29791,-1},
        {29791,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24760,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31259,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25515,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22517,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22528,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22509,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29746,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22539,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28290,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24428,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26965,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25580,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28203,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25641,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22533,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29864,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28409,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29857,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31558,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26840,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28507,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29887,29831},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22725,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29859,-1},
        {29859,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29898,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29918,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37032,26599},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29392,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29392,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33999,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32793,26599},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29920,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32745,35290},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40363,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36919,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23920,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29922,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29922,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34849,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20324,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29955,-1},
        {29955,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32568,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29942,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29942,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30414,-1},
        {30414,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30412,30871},
        {30412,30871,-1}
    },
    {
        {-1,-1},
        {-1},
        {29942,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30745,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29942,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29935,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30745,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29955,27951},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34537,29943},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29946,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29953,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22519,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22666,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29966,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36571,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29966,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29965,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29956,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29969,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29966,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38988,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29969,35238},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29966,-1},
        {29966,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29951,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32571,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29969,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29955,32883},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29922,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29245,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40594,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29951,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37165,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39730,-1},
        {39730,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27481,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32763,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29969,30070},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30992,29965},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30097,29971},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37970,29971},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29166,29975},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35722,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22377,-1},
        {22377,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25121,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39334,-1},
        {39334,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29989,-1},
        {29989,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30007,-1},
        {30007,-1,-1}
    },
    {
        {16215,-1},
        {-1},
        {-1,-1},
        {30045,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30029,21549},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20801,-1},
        {20801,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21329,31362},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26155,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30073,-1},
        {30073,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30041,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30041,-1},
        {30041,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22482,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30025,23805},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21860,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26235,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22495,-1},
        {22495,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32579,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30245,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30081,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22092,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22725,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21103,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22567,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26143,-1},
        {26143,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30054,-1},
        {30054,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {16242,-1},
        {-1,-1,-1}
    },
    {
        {16242,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30100,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30142,-1},
        {-1,-1,-1}
    },
    {
        {16245,-1},
        {-1},
        {-1,-1},
        {33624,33674,24196}
    },
    {
        {16247,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30115,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30130,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30130,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30130,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30204,-1},
        {30204,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23624,30109},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31370,30117},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24623,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24629,-1},
        {24629,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31518,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30157,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30141,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24220,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30137,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30240,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30195,24229},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30316,-1},
        {30316,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30149,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20985,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30163,21395},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21169,21237},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30174,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30146,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24705,-1},
        {24705,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30168,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23472,30209},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21400,16302},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30165,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26775,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24725,30159},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30133,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23517,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30192,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28681,-1},
        {28681,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30149,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28041,30183},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29378,30158},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21364,30196},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30313,-1},
        {30313,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30154,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21836,-1},
        {21836,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30229,20470},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30230,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24776,30201},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20506,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20054,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30175,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30158,-1},
        {30158,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24832,30195},
        {-1,-1,-1}
    },
    {
        {16302,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30201,30031},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28246,24287},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30280,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30240,28369},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30247,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26612,30133},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39342,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28353,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36864,30165},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36861,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28610,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {63784,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30309,-1},
        {30309,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30309,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34069,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30242,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30316,-1},
        {30316,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30207,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30259,24948},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30320,-1},
        {30320,-1,-1}
    },
    {
        {16329,-1},
        {-1},
        {-1,-1},
        {16343,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28507,30319},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28507,30319},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30233,-1},
        {30233,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30290,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30117,-1},
        {30117,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {16329,24996},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {151095,-1},
        {151095,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32735,30319},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28644,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30238,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30279,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24300,-1},
        {-1,-1,-1}
    },
    {
        {16343,-1},
        {-1},
        {-1,-1},
        {16329,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30284,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30293,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30311,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39321,-1},
        {39321,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32772,-1},
        {32772,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30333,-1},
        {30333,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30333,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20828,-1},
        {20828,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32768,-1},
        {32768,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26253,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26265,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30368,30906},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30956,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30087,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30378,-1},
        {30378,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25897,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22519,30383},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30382,30342},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24052,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34987,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34449,24084},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25343,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25928,30388},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27449,-1},
        {27449,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30390,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30877,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24171,-1},
        {24171,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32309,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38851,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26292,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30402,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30402,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30402,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30406,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30402,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30409,32997},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30412,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30420,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30435,64076},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30433,23613},
        {30433,23613,-1}
    },
    {
        {-1,-1},
        {-1},
        {30444,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22669,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21932,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30442,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22633,40573},
        {22633,40573,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30468,-1},
        {30468,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30516,-1},
        {30516,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32924,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30472,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32907,29283},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30449,31046},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26144,35363},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30449,26501},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30449,26501},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30732,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26114,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24573,31365},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30471,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21640,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30655,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35224,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30636,30516},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31061,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26190,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32850,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33011,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33019,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21704,30459},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35438,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29705,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26209,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30489,23786},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23792,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26214,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30491,31462},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33051,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30636,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30524,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34576,33050},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26210,143812},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {16441,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30592,33065},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32856,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30603,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30563,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26211,-1},
        {26211,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40718,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33102,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20469,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30559,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29467,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26235,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31365,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21899,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21919,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21898,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30624,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33136,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33136,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25562,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25558,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {16472,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26191,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30552,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24140,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29074,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30465,-1},
        {30465,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24046,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33153,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29831,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35627,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30500,26165},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35248,-1},
        {35248,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26293,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27112,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39770,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22079,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36104,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26241,26313},
        {26241,26313,-1}
    },
    {
        {-1,-1},
        {-1},
        {36118,36023},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33211,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35678,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30374,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35258,35226},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33215,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30565,-1},
        {30565,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30570,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30378,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {158391,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33240,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24173,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24173,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40664,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26342,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24009,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25899,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34875,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30684,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29379,25924},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29420,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29492,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35947,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27409,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28836,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25305,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29628,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {16531,-1},
        {-1},
        {-1,-1},
        {35215,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30707,-1},
        {30707,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30707,-1},
        {30707,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30707,-1},
        {30707,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21489,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {152622,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37064,30752},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22948,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23849,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30932,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26584,-1},
        {26584,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21759,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30692,30726},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30757,-1},
        {30757,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26570,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30738,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22399,-1},
        {22399,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29641,-1},
        {29641,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27524,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30733,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26027,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30831,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37530,30973},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26673,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21676,30733},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30887,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30887,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30753,24196},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30798,21774},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21065,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30740,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20054,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28825,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38982,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37605,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21816,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30824,21829},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30820,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36367,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21854,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23824,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21896,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29641,-1},
        {29641,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30752,33524},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21929,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30950,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22684,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22684,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30869,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26258,30861},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30928,26001},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21981,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23230,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29743,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23445,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22626,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30860,21085},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27162,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37827,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37832,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27593,30956},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22654,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23956,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27584,-1},
        {27584,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22063,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22087,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22698,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {63964,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22727,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30920,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25788,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22112,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31019,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27307,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27569,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30928,-1},
        {30928,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33240,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27367,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28699,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27570,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30952,-1},
        {30952,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22761,-1},
        {22761,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31059,31124},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31059,31124},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {161329,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {161329,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31117,-1},
        {64082,31117,-1}
    },
    {
        {-1,-1},
        {-1},
        {32050,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34966,-1},
        {34966,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34966,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27523,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27523,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {16645,-1},
        {-1},
        {-1,-1},
        {31206,-1,-1}
    },
    {
        {16645,-1},
        {-1},
        {-1,-1},
        {31206,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26653,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35022,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35022,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35022,-1},
        {31243,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23077,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34593,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34593,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26885,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31045,22934},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31045,22934},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31040,31071},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26836,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31252,-1},
        {31252,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31252,-1},
        {31252,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24488,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31038,-1},
        {64076,31038,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31128,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27040,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27133,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39761,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27635,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31109,-1},
        {31109,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31109,-1},
        {31109,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25993,40778},
        {25993,40778,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27422,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26457,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26451,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26451,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31227,-1},
        {31227,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24180,-1},
        {24180,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31235,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {153513,26501},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {153513,26501},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33469,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26507,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26520,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31049,31189},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25282,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30707,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26575,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31258,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26691,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26646,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26800,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20363,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26696,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26684,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26753,24041},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21332,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26690,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31903,-1},
        {31903,-1,-1}
    },
    {
        {16712,-1},
        {-1},
        {-1,-1},
        {31206,22863,-1}
    },
    {
        {16712,-1},
        {-1},
        {-1,-1},
        {31206,22863,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20354,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26778,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25424,31245},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26755,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {142150,-1},
        {-1,-1,-1}
    },
    {
        {16719,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26776,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26741,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33426,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20742,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33745,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39331,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25468,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26829,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26860,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26905,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26864,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26892,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26831,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31242,-1},
        {31242,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32964,-1},
        {-1,-1,-1}
    },
    {
        {16739,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25554,31278},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26988,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26988,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25551,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34093,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39321,-1},
        {39321,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27495,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27036,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27085,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31958,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24681,31298},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27071,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27056,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {155368,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31957,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32808,-1},
        {32808,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34295,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33688,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32812,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31186,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38626,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27134,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39336,-1},
        {39336,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20717,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27194,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40692,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27297,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23566,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23566,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27264,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31309,36039},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31258,-1},
        {31258,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25850,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20864,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27379,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31966,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22283,22269},
        {22283,22269,-1}
    },
    {
        {-1,-1},
        {-1},
        {23425,31349},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31350,-1},
        {31350,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31350,-1},
        {31350,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20810,31379},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24088,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23439,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31368,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25334,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31408,31407},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27934,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23461,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31549,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31365,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20985,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23462,-1},
        {23462,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31365,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31368,25225},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23480,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31359,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36051,64100},
        {36051,64100,-1}
    },
    {
        {-1,-1},
        {-1},
        {28145,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31401,-1},
        {31401,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20901,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20901,-1},
        {20901,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33108,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31379,31383},
        {31379,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38691,-1},
        {38691,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36562,31795},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31414,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22889,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28129,31402},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24247,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {165321,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {165321,-1},
        {-1,-1,-1}
    },
    {
        {16820,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32638,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28966,31407},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30686,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31782,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28504,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31429,-1},
        {31429,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31392,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31811,-1},
        {-1,-1,-1}
    },
    {
        {16831,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31418,32647},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34330,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24838,-1},
        {24838,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31458,-1},
        {31458,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25238,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27667,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24452,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22956,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31441,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21702,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30813,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40810,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33950,25647},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23138,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23626,23622},
        {23626,23622,-1}
    },
    {
        {-1,-1},
        {-1},
        {23626,23622},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38920,38656},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23348,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23340,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26537,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27664,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24341,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25778,25908},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33463,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31505,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33527,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33137,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26492,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33510,32607},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31534,-1},
        {31534,-1,-1}
    },
    {
        {16870,-1},
        {-1},
        {-1,-1},
        {33521,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31637,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {16883,26355},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31718,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33585,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26675,-1},
        {26675,-1,-1}
    },
    {
        {16878,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32823,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31598,19998},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31640,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31640,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31604,-1},
        {31604,-1,-1}
    },
    {
        {16883,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27778,31541},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31593,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31681,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31640,-1},
        {31640,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35910,-1},
        {35910,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37013,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33765,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35328,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31574,-1},
        {31574,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31527,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31845,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31492,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27836,31541},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33761,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31703,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25150,-1},
        {-1,-1,-1}
    },
    {
        {16903,-1},
        {-1},
        {-1,-1},
        {32617,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31494,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31633,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22796,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27795,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33785,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30450,33500},
        {-1,-1,-1}
    },
    {
        {16910,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27126,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33845,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36432,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33883,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31539,-1},
        {33901,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21273,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33834,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33904,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33586,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33492,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31527,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27777,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32997,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31482,31716},
        {31716,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23857,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29232,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33905,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33864,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31601,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37463,-1},
        {37463,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31563,-1},
        {31563,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33916,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31705,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31705,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32013,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33970,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33936,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33951,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23512,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26786,31602},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33324,33902},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32741,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33977,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33994,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33993,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26684,-1},
        {26684,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34075,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31828,-1},
        {31828,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32066,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33071,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33771,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26786,31245},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33986,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34071,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32838,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34083,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38517,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34086,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25955,16996},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35290,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31774,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23194,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38538,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31777,24758},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26848,34241},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34210,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34154,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27156,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27544,31627},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32638,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31533,-1},
        {31533,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25558,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26837,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31700,-1},
        {31700,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29965,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34177,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34201,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32681,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31668,24863},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34183,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20150,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34196,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23531,-1},
        {-1,-1,-1}
    },
    {
        {16996,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25696,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36983,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21129,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34265,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27166,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34271,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34268,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24290,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34298,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36066,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39609,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27166,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24396,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34330,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34328,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39244,-1},
        {31825,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39244,-1},
        {31825,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31703,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24022,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36783,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38728,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31904,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31190,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26009,-1},
        {26009,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34978,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31875,31189},
        {31875,31189,-1}
    },
    {
        {-1,-1},
        {-1},
        {31946,-1},
        {31946,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31946,-1},
        {31946,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37521,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31933,30094},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29678,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37528,31917},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31959,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39188,-1},
        {39188,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26757,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31933,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {16719,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22766,22767},
        {22766,22767,-1}
    },
    {
        {-1,-1},
        {-1},
        {40654,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31885,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31938,31965},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39195,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39177,-1},
        {39177,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31934,-1},
        {31934,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31124,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29008,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39256,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26976,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {16739,-1},
        {16739,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31922,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27084,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27115,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34763,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27683,27671},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39237,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28479,-1},
        {28479,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27243,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32353,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31974,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23617,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32696,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27214,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31986,-1},
        {31986,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32618,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31262,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27310,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27330,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25844,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27367,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31903,37855},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38753,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31841,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32127,-1},
        {32127,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32199,-1},
        {32199,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32006,-1},
        {32006,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32006,-1},
        {32020,32006,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32090,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27448,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31049,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34946,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32011,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25158,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32134,-1},
        {32134,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32094,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32070,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32178,-1},
        {32178,-1,-1}
    },
    {
        {17094,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {17094,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37320,17094},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32110,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26620,32053},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32021,26622},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32608,-1},
        {32608,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22533,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32069,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32187,-1},
        {32187,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31066,32006},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32104,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32013,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32272,-1},
        {32272,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32029,-1},
        {32029,-1,-1}
    },
    {
        {17110,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22533,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32267,-1},
        {32267,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32221,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32764,-1},
        {32764,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32153,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32174,-1},
        {32174,-1,-1}
    },
    {
        {17117,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32155,36349},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35036,-1},
        {35036,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32019,-1},
        {32019,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32046,34567},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32321,-1},
        {32321,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32321,-1},
        {32321,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32032,32004},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32182,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32071,-1},
        {32071,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32080,36387},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34618,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23104,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37658,32051},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32230,15239},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32121,37669},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37449,32173},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32174,-1},
        {32174,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32386,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25988,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35097,-1},
        {35097,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31941,32181},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {64096,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38971,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37738,32174},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24490,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24133,-1},
        {24133,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38822,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32029,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27001,32266},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28247,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32306,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34636,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32313,26950},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32313,26950},
        {-1,-1,-1}
    },
    {
        {17154,-1},
        {-1},
        {-1,-1},
        {25569,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32131,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32134,-1},
        {32134,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27088,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32034,-1},
        {32034,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25640,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32177,-1},
        {32177,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35126,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27135,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27154,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22533,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27162,32394},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32321,-1},
        {64089,32321,-1}
    },
    {
        {-1,-1},
        {-1},
        {32321,-1},
        {32321,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28442,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32315,-1},
        {32315,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35149,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27205,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24163,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {161618,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32313,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32327,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {13458,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32058,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27298,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28604,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32321,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31150,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {18021,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32176,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32065,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31337,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27330,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32044,-1},
        {32044,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34081,27307},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35178,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32349,-1},
        {-1,-1,-1}
    },
    {
        {17195,-1},
        {-1},
        {-1,-1},
        {32412,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27379,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27189,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23360,32379},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27402,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28699,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32065,-1},
        {32065,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32696,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32696,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32568,-1},
        {-1,-1,-1}
    },
    {
        {17219,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {17219,-1},
        {-1},
        {-1,-1},
        {17219,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31450,32572},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32572,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29942,32574},
        {32574,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26547,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34527,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37656,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29951,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {64084,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38857,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32593,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32597,-1},
        {32597,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32597,-1},
        {32597,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31359,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23439,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32608,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31903,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32608,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32607,31623},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32633,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32626,27795},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32621,-1},
        {32621,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40599,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34154,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34174,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38717,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38722,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29250,-1},
        {29250,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32666,27803},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39178,-1},
        {39178,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27819,32668},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27825,32668},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27865,32668},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29250,-1},
        {29250,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27950,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32669,33030},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32666,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29450,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35056,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29465,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28235,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28325,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32694,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28470,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28512,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28507,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28701,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32724,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20976,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {156674,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35425,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34957,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35425,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32714,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32716,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32714,158238},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25341,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30686,-1},
        {30686,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35425,38610},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35425,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25417,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35495,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35425,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32762,26397},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31926,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27544,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32283,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33994,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29969,32750},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32750,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32750,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32762,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35711,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32769,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32741,-1},
        {32741,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31915,22989},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31915,22989},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25062,24862},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32782,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32793,20094},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32796,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25513,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33745,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33745,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26831,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29103,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32801,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32793,-1},
        {26355,25341,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32828,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32854,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24677,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21678,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26212,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32862,-1},
        {32862,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29464,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36057,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {13736,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32862,-1},
        {32862,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21886,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21895,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32289,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30609,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33187,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38555,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33370,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30623,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25891,32879},
        {25891,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38576,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32894,-1},
        {32894,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32894,-1},
        {32894,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31899,32901},
        {31899,32901,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20350,-1},
        {-1,-1,-1}
    },
    {
        {32921,-1},
        {-1},
        {32921,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32908,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21505,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32998,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28825,-1},
        {28825,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32982,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33048,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32920,34884},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39146,-1},
        {39146,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39146,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27773,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32933,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30471,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26112,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26381,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22972,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27713,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26381,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34502,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38998,22239},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32835,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32850,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30517,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32966,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33029,-1},
        {33029,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21665,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30520,-1},
        {30520,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21678,-1},
        {-1,-1,-1}
    },
    {
        {17390,-1},
        {-1},
        {-1,-1},
        {33031,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30179,-1},
        {30179,-1,-1}
    },
    {
        {17392,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {17392,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32965,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26211,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21736,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39616,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26212,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30541,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33021,-1},
        {33021,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36068,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36064,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33120,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25486,-1},
        {-1,-1,-1}
    },
    {
        {35930,-1},
        {-1},
        {35930,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {13736,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38519,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21563,21767},
        {-1,-1,-1}
    },
    {
        {17416,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33075,33126},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39195,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32865,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33094,-1},
        {33094,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33094,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {17442,-1},
        {17442,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30588,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26265,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21886,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21695,-1},
        {-1,-1,-1}
    },
    {
        {17431,-1},
        {-1},
        {-1,-1},
        {29482,35948,-1}
    },
    {
        {-1,-1},
        {-1},
        {26262,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33160,-1},
        {33160,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33050,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33007,-1},
        {-1,-1,-1}
    },
    {
        {17436,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21993,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23243,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36092,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35594,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33229,-1},
        {33229,-1,-1}
    },
    {
        {17442,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27550,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28355,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29330,33243},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35315,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36452,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33370,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33372,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37855,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27151,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32670,-1},
        {32670,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33129,36450},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33184,25722},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39770,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26305,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27563,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33173,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22063,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36764,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32290,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30622,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30294,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30992,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29877,-1},
        {-1,-1,-1}
    },
    {
        {26406,-1},
        {-1},
        {26406,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30236,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32907,-1},
        {32907,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30378,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28699,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33208,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29211,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30317,-1},
        {30317,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27570,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33252,29646},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20693,-1},
        {20693,-1,-1}
    },
    {
        {17491,-1},
        {-1},
        {-1,-1},
        {21488,33274,-1}
    },
    {
        {-1,-1},
        {-1},
        {32954,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40736,-1},
        {40736,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26264,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29980,-1},
        {29980,-1,-1}
    },
    {
        {17499,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30879,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26013,-1},
        {26013,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33308,-1},
        {33308,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33887,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33313,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {158238,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24420,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33330,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33336,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33332,32981},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33333,-1},
        {33333,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33337,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33340,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33340,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30522,33340},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33313,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30535,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {158296,38495},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33333,33109},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27330,-1},
        {27330,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23832,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33340,-1},
        {-1,-1,-1}
    },
    {
        {17526,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33368,-1},
        {-1,-1,-1}
    },
    {
        {17530,-1},
        {-1},
        {-1,-1},
        {24070,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33344,34659},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {17442,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22604,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32981,31526},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25722,33334},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27123,-1},
        {27123,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36896,-1},
        {36896,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33378,25803},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {17526,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22132,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20763,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33396,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26049,33398},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33396,33725},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33398,26366},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30626,33398},
        {-1,-1,-1}
    },
    {
        {17553,-1},
        {-1},
        {-1,-1},
        {33400,64094,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33437,20062},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33904,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33678,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29282,33511},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33406,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20888,33717},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20025,27500},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20016,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24341,33422},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33542,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33542,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33490,-1},
        {33490,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33500,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33470,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33437,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33464,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31542,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33707,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33707,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24323,33725},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33707,26286},
        {33707,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31561,-1},
        {31561,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31568,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33976,-1},
        {33976,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33976,-1},
        {33976,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22908,33735},
        {-1,-1,-1}
    },
    {
        {17587,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {17587,-1},
        {-1},
        {-1,-1},
        {17587,26422,-1}
    },
    {
        {-1,-1},
        {-1},
        {32948,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22909,33735},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22825,-1},
        {22825,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22825,-1},
        {22825,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31560,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31549,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33545,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33545,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33545,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33545,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32034,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33539,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31553,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33454,-1},
        {-1,-1,-1}
    },
    {
        {17598,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {17598,-1},
        {-1},
        {-1,-1},
        {17598,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31561,-1},
        {31561,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31561,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35400,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33453,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33659,33765},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33465,-1},
        {33465,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33765,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36203,22151},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24441,33902},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26928,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33571,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33760,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24665,24535},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24665,24535},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33609,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28796,33903},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33685,-1},
        {33443,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33780,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33780,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33805,-1},
        {33805,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33967,-1},
        {33967,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {17620,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {17620,-1},
        {-1},
        {-1,-1},
        {17620,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26851,34085},
        {34085,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33558,26231},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20054,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24450,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23851,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27868,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23391,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33446,25150},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33570,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27795,33879},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33593,30693},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31623,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {13500,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33905,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33609,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20856,-1},
        {20856,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20856,-1},
        {20856,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20127,-1},
        {20127,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21049,33477},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33542,-1},
        {33542,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27602,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33589,-1},
        {33589,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33544,26612},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24456,33563},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32782,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23891,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23891,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34302,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33685,33607},
        {33685,33607,-1}
    },
    {
        {-1,-1},
        {-1},
        {23015,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39248,-1},
        {39248,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34364,-1},
        {34364,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34364,-1},
        {34364,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34098,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21398,33735},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35352,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23057,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33761,-1},
        {33761,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33904,33661},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26647,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33688,23472},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {64080,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31408,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33499,21733},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35912,33651},
        {-1,-1,-1}
    },
    {
        {17672,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {17672,-1},
        {-1},
        {-1,-1},
        {17672,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33948,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34033,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34033,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22675,30347},
        {-1,-1,-1}
    },
    {
        {17677,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33982,36895},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31627,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33834,26012},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34003,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35377,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31455,34191},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33071,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33777,34054},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31077,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37347,33903},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34000,33059},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32835,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32043,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29575,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31747,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26786,30687},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39770,34186},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33775,-1},
        {33775,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29313,34110},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33897,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33897,-1},
        {-1,-1,-1}
    },
    {
        {17701,-1},
        {-1},
        {-1,-1},
        {34157,-1,-1}
    },
    {
        {17701,-1},
        {-1},
        {-1,-1},
        {34157,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34110,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33698,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29999,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37220,34012},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33914,25558},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27225,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28982,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33789,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34129,-1},
        {34129,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34129,-1},
        {34129,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25552,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29748,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38538,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38539,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25958,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25958,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34123,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34123,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28212,64031},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25581,64031},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23562,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31422,27454},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26832,-1},
        {-1,-1,-1}
    },
    {
        {17731,-1},
        {-1},
        {-1,-1},
        {17731,34093,-1}
    },
    {
        {17731,-1},
        {-1},
        {-1,-1},
        {34093,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33976,33890},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23234,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26977,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33936,24262},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23655,23647},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30570,34045},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20150,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22616,33998},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31305,31258},
        {31305,31258,-1}
    },
    {
        {-1,-1},
        {-1},
        {31305,31258},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31305,31258},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35028,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33706,-1},
        {33706,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33904,-1},
        {33904,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22025,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34081,23519},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37532,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33984,-1},
        {33984,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40763,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33491,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25130,31830},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34146,-1},
        {34146,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31281,34250},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34298,38316},
        {33704,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34298,38316},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34298,38316},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33288,34247},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34129,38317},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30923,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {64096,34297},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34308,-1},
        {34308,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33512,-1},
        {33512,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33581,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34276,-1},
        {34276,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34304,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22804,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34193,-1},
        {34193,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34042,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33738,-1},
        {33738,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34281,-1},
        {34281,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34315,38957},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33787,-1},
        {33787,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33787,-1},
        {33787,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24396,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34323,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40288,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35703,-1},
        {35703,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35703,-1},
        {35703,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32315,34278},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31277,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38620,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21474,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21474,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31853,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38627,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36485,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24022,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24022,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40475,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22210,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35689,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25891,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30922,33509},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38728,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37312,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39729,27421},
        {39729,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39729,27421},
        {39729,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34384,-1},
        {34384,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34390,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34384,-1},
        {34384,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24426,21127},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34388,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34388,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34389,20966},
        {34389,20966,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24426,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29733,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29733,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29390,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31668,34389},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40393,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32034,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {15132,-1},
        {-1,-1,-1}
    },
    {
        {17821,-1},
        {-1},
        {-1,-1},
        {34411,-1,-1}
    },
    {
        {17821,-1},
        {-1},
        {-1,-1},
        {34411,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34673,34427},
        {34673,34427,-1}
    },
    {
        {-1,-1},
        {-1},
        {34673,34427},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34502,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34442,29573},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34468,-1},
        {34468,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34623,-1},
        {34623,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34661,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34480,32064},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34567,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {17898,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34461,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34586,30758},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34444,-1},
        {34444,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34453,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34817,-1},
        {34817,22190,-1}
    },
    {
        {-1,-1},
        {-1},
        {34521,-1},
        {34521,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34503,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29671,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34802,34411},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31232,34517},
        {-1,-1,-1}
    },
    {
        {17848,-1},
        {-1},
        {-1,-1},
        {34558,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34513,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26781,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {18429,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34576,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34570,-1},
        {34570,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34570,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34615,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34595,29733},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34633,34796},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {166430,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34503,29743},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31927,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34673,34427},
        {34673,34427,-1}
    },
    {
        {-1,-1},
        {-1},
        {34673,34427},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37673,34574},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34554,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34513,-1},
        {34513,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34619,-1},
        {34619,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34849,-1},
        {34849,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22553,34647},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22553,34647},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34699,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34731,26045},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29020,34647},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34486,135990},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32870,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22560,34662},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34623,-1},
        {34623,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34720,27068},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34540,36423},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27025,34565},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29792,34784},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34453,34870},
        {34453,34870,-1}
    },
    {
        {-1,-1},
        {-1},
        {34453,34870},
        {34453,34870,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26059,34735},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23267,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34468,-1},
        {34468,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34753,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22665,-1},
        {-1,-1,-1}
    },
    {
        {17893,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34562,-1},
        {34562,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34562,-1},
        {34562,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31968,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28417,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32300,-1},
        {-1,-1,-1}
    },
    {
        {17898,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34758,-1},
        {34758,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34562,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34554,29245},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29857,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34790,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36476,34758},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29848,34684},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36460,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34656,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34847,-1},
        {34847,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34794,27277},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34817,29973},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34461,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34577,-1},
        {34577,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34797,22734},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34875,22725},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34566,-1},
        {34566,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27320,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34519,-1},
        {34519,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {17848,25130},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32390,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34850,32741},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33021,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34794,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34442,-1},
        {34442,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34453,34870},
        {34453,34870,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {160766,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34822,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34822,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28771,-1},
        {-1,-1,-1}
    },
    {
        {17935,-1},
        {-1},
        {-1,-1},
        {24676,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34884,-1},
        {34884,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33033,-1},
        {33032,33033,-1}
    },
    {
        {-1,-1},
        {-1},
        {30063,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34884,21847},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21888,-1},
        {21888,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34898,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34904,37532},
        {34904,37532,-1}
    },
    {
        {-1,-1},
        {-1},
        {24481,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34907,34910},
        {34907,34910,-1}
    },
    {
        {-1,-1},
        {-1},
        {21330,-1},
        {21330,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24062,-1},
        {24062,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26457,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26617,34978},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34955,27606},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32053,34943},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {161329,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31211,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31070,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34957,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26657,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24642,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32113,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35005,-1},
        {35005,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35027,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26702,35007},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21930,-1},
        {21930,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35097,32131},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35114,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35029,-1},
        {35029,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34943,35167},
        {35167,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34927,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31108,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28085,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35058,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26861,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26861,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26890,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32187,-1},
        {32187,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35097,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26880,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20522,35222},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32176,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20500,35070},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32323,-1,-1}
    },
    {
        {17985,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34927,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35097,26965},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25554,34966},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35076,26950},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25542,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24862,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32241,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35070,26991},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24495,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34180,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31119,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32279,-1},
        {32279,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25616,35168},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27166,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35127,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32323,-1},
        {32323,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35027,27117},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35018,-1},
        {35018,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34974,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36898,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36898,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35131,-1},
        {35131,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35131,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27225,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35096,22285},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27227,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32362,32117},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32203,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {18021,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20762,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35065,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27387,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35120,-1},
        {35120,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27402,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36441,27117},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36983,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29083,-1},
        {29083,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36983,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36983,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36983,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24471,-1},
        {24471,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28098,24471},
        {24471,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {64097,37026},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24265,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37045,30831},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35220,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20282,26045},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26776,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30541,158238},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35223,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {64097,36204},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35224,21347},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21085,37607},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35233,37106},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33521,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35241,30465},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35244,29956},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22865,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40385,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37142,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35263,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35241,33282},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26234,35251},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35263,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35250,31461},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35203,35247},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36051,35263},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37192,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38728,35233},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25179,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34913,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32089,35293},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37521,35301},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35493,35302},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39616,35318},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37257,35303},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27513,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39908,-1},
        {-1,-1,-1}
    },
    {
        {18081,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39944,35307},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39927,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39983,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39973,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25963,36101},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40023,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40023,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39529,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {18095,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35355,-1},
        {35355,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35355,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20449,-1},
        {20449,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35349,-1},
        {35349,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32993,35419},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35336,25238},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35463,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35362,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35616,35617},
        {35616,35617,-1}
    },
    {
        {-1,-1},
        {-1},
        {35616,35617},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35419,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32791,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35473,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35387,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35363,27439},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21622,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21891,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35445,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33109,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21969,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21610,21650},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35373,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35475,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30496,35398},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37465,35379},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35488,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30770,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27877,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35422,-1},
        {35422,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35600,-1},
        {35600,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24321,36783},
        {24321,36783,-1}
    },
    {
        {-1,-1},
        {-1},
        {35542,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35547,-1},
        {35547,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35405,-1},
        {35405,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21761,38948},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35473,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35338,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35617,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {16441,35491},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35657,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35441,-1},
        {35441,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35480,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32856,35556},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35551,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30542,35558},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35359,-1},
        {35359,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35524,36766},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24801,24746},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31353,35463},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33081,35398},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35426,32175},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28076,35695},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21646,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35551,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {166441,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35646,-1},
        {35646,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35435,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35596,30869},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25480,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35460,-1},
        {35460,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23113,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35489,-1},
        {35489,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35480,-1},
        {35480,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35617,37732},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35584,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33160,35627},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22534,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31305,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31305,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35504,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40639,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35698,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21222,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35584,33181},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22024,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35715,30646},
        {-1,-1,-1}
    },
    {
        {18188,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35612,30637},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35607,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27234,35531},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22702,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24801,24746},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30983,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35548,-1},
        {35548,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36110,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35572,36695},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35686,-1},
        {35686,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28629,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36957,35678},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36957,35678},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35730,-1},
        {35730,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35691,28652},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32380,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33229,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35734,-1},
        {35734,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35734,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35715,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32816,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27442,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21371,30716},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23723,-1},
        {23723,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28652,30591},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29282,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30366,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35646,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35913,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35913,-1},
        {35913,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35913,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35913,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33353,38957},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37221,-1},
        {37221,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36374,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35913,-1},
        {35913,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34661,35918},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36400,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26263,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22777,29151},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35918,-1},
        {31466,35918,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29988,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35927,-1},
        {35927,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33945,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25463,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24793,35949},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25582,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35986,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32812,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39519,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39715,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35995,-1},
        {35995,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25885,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25235,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35973,-1},
        {35973,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27589,-1},
        {27589,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35970,25259},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29435,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29463,-1},
        {29463,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34399,21495},
        {21495,34399,-1}
    },
    {
        {-1,-1},
        {-1},
        {35992,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35992,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {18276,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35993,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30962,-1},
        {-1,-1,-1}
    },
    {
        {18276,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22071,-1},
        {22071,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30487,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36066,-1},
        {36066,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36066,-1},
        {36066,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36049,-1},
        {36049,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30588,36022},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32985,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30487,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34898,-1},
        {34898,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36004,-1},
        {36004,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26178,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33029,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36066,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40782,-1},
        {40782,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33071,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26234,-1},
        {26234,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33109,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36011,-1},
        {36011,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36064,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30555,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30555,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30554,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36070,-1},
        {36070,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36020,-1},
        {36020,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30586,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32251,25943},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20469,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36029,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21993,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36100,-1},
        {36100,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36008,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37357,36198},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37509,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36204,-1},
        {36204,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33353,36199},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28189,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36234,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36212,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36804,36215},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40618,36229},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36387,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36228,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36228,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36228,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36846,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22396,36212},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36229,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36249,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36262,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36869,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36332,-1},
        {36332,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36229,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21280,36228},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36262,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36879,-1},
        {36879,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36897,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36884,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36245,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {155671,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36903,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36387,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36913,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34627,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36245,36919},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22528,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36945,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36929,36249},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36404,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31428,-1},
        {31428,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36958,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36423,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36425,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29829,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37853,36245},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36225,31965},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32318,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36436,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36437,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36234,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36234,37941},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36485,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36989,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36249,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25130,36271},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37002,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36245,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22754,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36493,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36271,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37194,20163},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36387,36212},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34921,36290},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36286,-1},
        {36286,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33296,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36321,26477},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36305,30732},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36420,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34443,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36290,27573},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36234,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32989,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36296,-1},
        {36296,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37477,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39654,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36299,-1},
        {36299,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30033,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30496,36319},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34986,36282},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36022,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26848,23508},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34479,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38460,-1},
        {38460,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21264,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34526,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29673,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36857,-1},
        {36857,36321,-1}
    },
    {
        {-1,-1},
        {-1},
        {36319,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34597,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29437,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36394,34573},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36341,36070},
        {-1,-1,-1}
    },
    {
        {18429,-1},
        {-1},
        {-1,-1},
        {30095,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34627,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37631,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36491,-1},
        {36491,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26970,-1},
        {26970,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36451,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36362,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36361,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25496,36382},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22509,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36909,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21077,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39616,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34662,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40823,-1},
        {40823,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36460,30589},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36406,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36332,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34676,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34660,36461},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38538,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {18454,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21957,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36420,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36420,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29796,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36432,33592},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22628,36447},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36441,-1},
        {36441,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34756,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28382,28399},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27165,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30626,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34796,36437},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36499,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22712,36487},
        {36487,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34814,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31330,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32380,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39519,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36497,37159},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23358,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39472,-1},
        {39472,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38000,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36385,-1},
        {36385,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34827,36485},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37034,23556},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23556,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23556,32981},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23556,-1},
        {23556,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30522,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {164471,33007},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39608,-1},
        {39608,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31406,-1},
        {31406,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36664,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36550,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32885,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33211,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26407,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36555,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36564,-1},
        {-1,-1,-1}
    },
    {
        {18500,-1},
        {-1},
        {-1,-1},
        {36556,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36578,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36611,-1},
        {36611,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33322,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30024,26028},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34449,36611},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36578,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36604,-1},
        {-1,-1,-1}
    },
    {
        {18510,-1},
        {-1},
        {-1,-1},
        {36708,-1,-1}
    },
    {
        {18510,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36578,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37217,36571},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30496,36670},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36600,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36562,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36587,22821},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25327,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36618,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38799,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36631,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36554,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39909,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30073,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38814,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24128,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37703,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {18510,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36681,36081},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27367,-1},
        {27367,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31299,36676},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40356,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24046,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37806,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32752,39480},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37832,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37853,36653},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32670,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36452,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36672,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36674,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39493,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39493,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22194,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36633,34774},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24162,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36676,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36682,-1},
        {36682,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29914,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36708,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21128,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26546,-1},
        {26546,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26216,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39712,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36786,-1},
        {36786,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36786,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36938,-1},
        {36938,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21705,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36810,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36810,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36856,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24450,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24450,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36817,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36857,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36802,23431},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36840,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36840,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36234,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24314,-1},
        {24314,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36842,26354},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25943,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36837,-1},
        {36837,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24112,27512},
        {27512,24112,-1}
    },
    {
        {-1,-1},
        {-1},
        {20124,20126},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28107,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30001,-1},
        {30001,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36865,-1},
        {36865,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22558,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20630,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36953,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36848,36881},
        {36848,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29575,-1},
        {29575,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29575,-1},
        {29575,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36931,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36452,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24481,31142},
        {24481,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27207,-1},
        {27207,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38543,-1},
        {38543,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38543,-1},
        {38543,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36983,-1},
        {36983,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32680,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36924,38647},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36271,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36271,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {163232,36986},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26297,30669},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {18613,-1},
        {-1},
        {-1,-1},
        {37142,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37030,-1},
        {37030,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32822,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33252,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21351,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {165626,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24055,-1},
        {24055,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37054,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37066,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37026,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37106,32796},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21368,-1},
        {21368,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21465,37059},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37119,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21060,37061},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37106,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24055,-1},
        {24055,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40680,20826},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37083,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37048,-1},
        {37048,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22852,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20539,20379},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33521,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33521,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24283,-1},
        {24283,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34701,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37142,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39359,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {166330,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22530,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23849,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25976,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33181,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37191,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23648,37119},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37122,-1},
        {37122,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26369,37191},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32681,37167},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32804,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40680,20826},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30669,37191},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37197,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37194,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37194,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32020,37255},
        {37255,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37255,-1},
        {37255,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37204,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37208,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37208,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38758,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37301,-1},
        {37301,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37207,-1},
        {37207,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37211,23749},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37196,35422},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37207,-1},
        {37207,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37218,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29980,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34523,37221},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20973,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37219,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28092,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37261,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35516,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37206,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35584,37219},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37253,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28238,-1},
        {28238,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35565,37302},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37271,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37295,-1},
        {37295,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35600,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37235,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39320,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36451,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35662,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35670,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34789,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35663,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30663,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32891,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36706,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26333,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35734,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34849,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26446,-1},
        {26446,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37338,-1},
        {37338,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37389,-1},
        {37389,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37466,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37399,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32835,37559},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37339,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37449,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36330,37547},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37509,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37448,35435},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38001,-1},
        {38001,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37396,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37719,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37396,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37574,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35475,-1},
        {35475,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22812,37861},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20989,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25161,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34648,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37663,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26020,37375},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37669,39958},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37682,39973},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36424,37732},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37733,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37733,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37733,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21988,37799},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20989,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37648,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37679,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29748,-1},
        {29748,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29748,-1},
        {29748,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33725,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38001,23491},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37638,-1},
        {37638,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28921,37662},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35657,35388},
        {35657,35388,-1}
    },
    {
        {-1,-1},
        {-1},
        {34756,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33059,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33059,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37593,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37738,-1},
        {37738,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37858,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37439,-1},
        {37439,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27454,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37679,-1},
        {37679,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38538,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21021,-1},
        {21021,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37733,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37846,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32371,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26988,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37756,-1},
        {37756,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37745,-1},
        {37745,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37863,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38015,-1},
        {38015,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38649,37939},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29887,-1},
        {29887,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37986,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24494,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {15118,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37444,37941},
        {37444,37941,-1}
    },
    {
        {-1,-1},
        {-1},
        {37787,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37787,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35657,35388},
        {35657,35388,-1}
    },
    {
        {-1,-1},
        {-1},
        {37908,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38012,32631},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34673,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20889,23531},
        {20889,23531,-1}
    },
    {
        {-1,-1},
        {-1},
        {27472,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32816,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26313,37891},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39730,-1},
        {39730,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24396,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39675,-1},
        {39675,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21983,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32902,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22007,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38287,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38286,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38284,-1},
        {38284,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38310,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38283,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38306,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21839,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35224,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38291,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38283,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38283,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38307,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38480,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23546,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38346,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38307,21566},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38283,36196},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38287,-1},
        {38287,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31337,31311},
        {31337,31311,-1}
    },
    {
        {-1,-1},
        {-1},
        {35925,38303},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23448,38317},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38343,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21461,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38289,26519},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29748,-1},
        {29748,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39721,-1},
        {39721,-1,-1}
    },
    {
        {38348,-1},
        {-1},
        {38348,-1},
        {38348,-1,-1}
    },
    {
        {38348,-1},
        {-1},
        {38348,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26359,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33521,38306},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22865,38347},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39321,38349},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26149,38349},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33268,38366},
        {-1,-1,-1}
    },
    {
        {18864,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28858,38358},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35449,35518},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38364,37111},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30570,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38433,21494},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38433,21486},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38433,20180},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25225,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24207,-1},
        {24207,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23803,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21620,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26157,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27834,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21679,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21670,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26190,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20381,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21737,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26215,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26203,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38553,-1},
        {38553,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38534,-1},
        {38477,38534,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38532,-1},
        {38532,22564,-1}
    },
    {
        {-1,-1},
        {-1},
        {38492,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26221,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21816,38512},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22495,-1},
        {22495,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27544,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38532,24846},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21913,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38517,-1},
        {38517,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20621,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21776,22616},
        {22616,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21988,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21966,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23269,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22665,-1},
        {22665,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23943,167455},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23943,167455},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28654,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38560,24979},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23296,-1},
        {-1,-1,-1}
    },
    {
        {18919,-1},
        {-1},
        {-1,-1},
        {23980,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {13458,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30238,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28640,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28657,27996},
        {28693,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38997,39023},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39872,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40169,-1},
        {40169,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25512,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38604,-1},
        {38604,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40198,-1},
        {40198,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38598,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38610,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40227,-1},
        {40227,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40441,-1},
        {40441,-1,-1}
    },
    {
        {18938,-1},
        {-1},
        {-1,-1},
        {38596,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40255,-1},
        {40255,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40288,-1},
        {40288,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38597,-1},
        {38597,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35504,33302},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40306,-1},
        {40306,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31988,-1},
        {31988,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40361,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40342,-1},
        {40342,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40387,-1},
        {40387,-1,-1}
    },
    {
        {18948,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40431,-1},
        {40431,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38617,21452},
        {38617,21452,-1}
    },
    {
        {-1,-1},
        {-1},
        {36051,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40474,-1},
        {40474,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35728,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34384,-1},
        {34384,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38649,-1},
        {38649,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38633,31357},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23446,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27965,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38633,22839},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32701,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38721,-1},
        {38721,38647,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38717,-1},
        {38717,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38633,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38704,-1},
        {38704,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38649,24573},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26512,38678},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23849,38686},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38738,38660},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {14178,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31402,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38724,-1},
        {38724,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33761,38647},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33882,38660},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33882,38660},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38712,38799},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36842,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36842,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38669,30669},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28354,-1},
        {28354,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38948,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38948,-1},
        {-1,-1,-1}
    },
    {
        {18985,-1},
        {-1},
        {-1,-1},
        {38728,38666,-1}
    },
    {
        {-1,-1},
        {-1},
        {38718,-1},
        {38718,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28107,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38660,-1},
        {38660,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38634,-1},
        {38634,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38707,38515},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28668,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35206,-1},
        {35206,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39020,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38675,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23565,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28525,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38704,-1},
        {38704,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38577,38707},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20789,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38642,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38721,38647},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30193,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32618,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36649,36059},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27635,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38758,27607},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38758,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37218,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36028,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {168625,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38754,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23113,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37294,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34890,37262},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38788,20094},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38762,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38766,37390},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30828,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23584,21202},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38712,35207},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38762,38830},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38795,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38762,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36705,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20881,-1},
        {20881,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21253,21263},
        {38788,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36028,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34534,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34579,38830},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34554,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38780,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34573,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35250,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34578,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38827,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38868,-1},
        {38868,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34623,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36635,38790},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29739,-1},
        {29739,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32648,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31798,-1},
        {31798,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38761,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36664,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38836,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38797,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34676,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21988,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24171,-1},
        {24171,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34697,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38712,35207},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38851,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36544,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36448,38857},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38761,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36476,38789},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38761,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29004,-1},
        {29004,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36485,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36550,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39226,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38851,32382},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34847,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36504,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38777,38860},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37448,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39155,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38875,36628},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38816,-1},
        {38816,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36665,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38873,-1},
        {38873,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39622,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36427,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36463,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38894,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38898,40783},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40783,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38898,40783},
        {40783,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29856,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34212,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38900,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35345,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35345,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38901,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21535,-1},
        {21535,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39326,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38899,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35589,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35597,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22123,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23401,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39648,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38916,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38916,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38928,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38928,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30889,38606},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27665,38926},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37043,38925},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {169104,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {19132,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38990,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38998,22239},
        {38998,22239,-1}
    },
    {
        {-1,-1},
        {-1},
        {21363,38997},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33258,38932},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33059,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38929,20812},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23435,38983},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39653,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38967,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30339,35980},
        {35980,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38945,21542},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38564,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38982,40303},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39001,21461},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27091,37108},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26127,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38945,38907},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22865,39006},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24093,38983},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21208,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21466,38990},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40353,38997},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39001,-1},
        {39001,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38994,21692},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20670,20399},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24189,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39333,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39333,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38930,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39003,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38968,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40380,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39025,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38622,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39013,20901},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {169104,30413},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32400,25942},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21443,38991},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38960,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37860,39631},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26369,39028},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39003,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39020,-1},
        {39020,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39633,-1},
        {39633,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31105,38964},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27146,39006},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38990,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39083,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34425,39083},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39087,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39085,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39088,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39083,34461},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39115,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39089,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34503,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39088,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39087,-1},
        {39087,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25100,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39112,21158},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34570,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20812,39087},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34555,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36196,39095},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39083,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39100,22796},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34382,39103},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39096,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39089,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39095,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39096,34655},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39100,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39089,38899},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39089,38899},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34643,39094},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39080,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39098,34647},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39098,-1},
        {39098,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30907,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39099,-1},
        {39099,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39099,-1},
        {39099,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39103,-1},
        {24070,39103,-1}
    },
    {
        {-1,-1},
        {-1},
        {39107,36437},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39761,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39107,34789},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36490,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22176,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22176,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38671,-1},
        {38671,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32741,-1},
        {32741,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39253,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39253,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39228,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39256,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39248,-1},
        {39248,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {146072,39151},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39138,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23428,39165},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40779,25998},
        {40779,25998,-1}
    },
    {
        {-1,-1},
        {-1},
        {39226,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39226,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39141,35293},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37472,39147},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39253,-1},
        {39253,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39156,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39156,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37450,39156},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37474,39191},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39147,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37473,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37473,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31203,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31203,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39218,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37504,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39204,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39184,28236},
        {39184,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37517,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37517,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35443,39227},
        {-1,-1,-1}
    },
    {
        {19259,-1},
        {-1},
        {-1,-1},
        {39198,-1,-1}
    },
    {
        {19259,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39194,37591},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39194,37591},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37613,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {16719,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36805,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36805,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39176,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39146,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39146,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39212,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39212,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37627,33253},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22094,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22094,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39256,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39194,39958},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39194,39958},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37704,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39165,-1},
        {39165,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39165,-1},
        {39165,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39212,-1},
        {39212,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39212,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37745,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39227,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39227,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27465,39201},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27465,39201},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39190,37799},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37786,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37786,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37864,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36448,39227},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37857,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22087,-1},
        {22087,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30640,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33214,-1},
        {33214,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32377,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37745,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27465,39201},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31337,39186},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31337,39186},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32889,39226},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39176,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39176,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37971,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40054,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39226,-1},
        {39226,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31964,-1},
        {31964,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35327,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35327,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38753,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38753,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31293,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31293,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39003,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39714,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39326,-1},
        {39326,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29739,39326},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37665,39331},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39336,39759},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39357,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24070,39103},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39439,-1},
        {39439,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39361,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25233,39348},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39356,39599},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39797,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39439,-1},
        {39439,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32675,32676},
        {32675,32676,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39378,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39394,-1},
        {39394,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39493,39366},
        {39493,39366,-1}
    },
    {
        {-1,-1},
        {-1},
        {39405,173111},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39498,32902},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39347,-1},
        {39347,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39390,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39847,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39410,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39385,39860},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35018,-1},
        {35018,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39889,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39384,173594},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39369,-1},
        {39369,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37241,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39878,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39419,36372},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37656,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36639,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39419,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39423,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39385,-1},
        {39385,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39944,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39506,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39942,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25436,25628},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39361,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39519,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39378,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33160,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39509,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {172162,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39426,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39502,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39522,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39488,-1},
        {39488,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35219,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36450,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40001,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37892,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34115,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37913,39460},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39531,-1},
        {39531,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35320,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40445,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40037,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40045,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39502,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37987,-1},
        {37987,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31452,40845},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39467,-1},
        {39467,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36513,-1},
        {-1,-1,-1}
    },
    {
        {19394,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39599,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39150,39599},
        {-1,-1,-1}
    },
    {
        {19402,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39609,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23611,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39597,-1},
        {39597,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39607,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39608,-1},
        {39608,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35301,-1},
        {-1,-1,-1}
    },
    {
        {19410,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38795,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35435,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33351,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26390,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39616,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22456,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39622,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37235,39620},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33109,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {170610,36386},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39635,39620},
        {39635,39620,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39608,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33155,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33153,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25705,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39243,39637},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33214,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38856,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {19432,-1},
        {-1},
        {-1,-1},
        {25970,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38957,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39013,-1},
        {39013,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22194,-1},
        {22194,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22194,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39658,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39648,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39648,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39665,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39686,-1},
        {39686,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39658,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39662,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39704,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39675,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21608,39653},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32736,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22996,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24404,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24605,39671},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38920,39706},
        {38920,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26580,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26305,-1},
        {26305,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24046,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33324,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23555,39663},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39658,39662},
        {39704,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39686,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39673,-1},
        {39673,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39698,31779},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25130,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34253,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38920,-1},
        {38920,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33945,39683},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39714,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30436,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39714,36106},
        {-1,-1,-1}
    },
    {
        {19479,-1},
        {-1},
        {-1,-1},
        {39726,-1,-1}
    },
    {
        {19479,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {19479,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32719,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34701,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39730,-1},
        {39730,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39730,-1},
        {39730,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29969,-1},
        {29969,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29038,-1},
        {64072,29038,-1}
    },
    {
        {-1,-1},
        {-1},
        {39746,-1},
        {39746,-1,-1}
    },
    {
        {19488,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39749,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {19410,39740},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27088,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35313,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {19488,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39749,-1},
        {39749,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35241,37276},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39761,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38656,37276},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31001,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39020,39759},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40845,-1},
        {31452,40845,-1}
    },
    {
        {-1,-1},
        {-1},
        {39774,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40789,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39779,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34444,-1},
        {34444,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39977,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28417,-1},
        {28417,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39795,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39864,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40669,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39823,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39852,23381},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40019,-1},
        {40019,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39848,-1},
        {39848,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40019,-1},
        {39995,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31308,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33340,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32090,39788},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32089,39962},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39190,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33344,39779},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33344,39779},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21747,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {166330,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33351,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {168625,39839},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39969,-1},
        {39969,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40058,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39915,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39918,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39956,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37259,39911},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40019,38814},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39918,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37624,39961},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36068,39940},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38816,39825},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38748,39854},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37257,39851},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39990,39919},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39890,39429},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40020,-1},
        {40019,40020,-1}
    },
    {
        {-1,-1},
        {-1},
        {39990,39919},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39899,36638},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40009,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39958,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39850,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39928,37733},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39986,39446},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39846,-1},
        {39846,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39889,37719},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {171716,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40035,-1},
        {40035,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34697,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39933,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28417,-1},
        {28417,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33368,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39827,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37799,40039},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30918,39958},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {171541,36670},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39911,-1},
        {39911,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39972,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40409,39986},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23561,39826},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39922,37831},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39944,37825},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27131,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39770,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22043,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34774,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39958,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {171739,-1},
        {171739,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39949,27454},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39975,-1},
        {39975,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39797,23959},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40056,33379},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33380,40054},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34831,-1},
        {34809,34831,-1}
    },
    {
        {-1,-1},
        {-1},
        {39920,39511},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39912,-1},
        {39912,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35676,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35676,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40024,30087},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {13829,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39933,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23358,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40169,-1},
        {40169,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40195,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22626,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40251,-1},
        {40251,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40178,-1},
        {40178,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40176,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40195,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38929,40198},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39746,40215},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39796,40213},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38924,40255},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40203,38927},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40198,-1},
        {40198,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38592,-1},
        {38592,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26783,-1},
        {26783,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40234,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40198,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40306,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40182,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40227,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40384,-1},
        {40383,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40195,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {172058,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40288,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40167,-1},
        {-1,-1,-1}
    },
    {
        {19652,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36601,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40180,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40289,-1},
        {40289,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40199,-1},
        {40199,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40198,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40273,-1},
        {40273,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40306,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40180,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40180,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40195,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40209,36039},
        {-1,-1,-1}
    },
    {
        {19665,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40182,-1},
        {40182,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40199,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40335,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40285,-1},
        {40285,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40285,-1},
        {40285,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40346,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40254,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40387,-1},
        {40387,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40378,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40300,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {136775,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40281,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40182,-1},
        {40182,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40272,25949},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39217,40195},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20976,-1},
        {20976,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40387,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38995,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40335,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40385,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {172281,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40434,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40356,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40199,-1},
        {40199,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40361,-1},
        {40361,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40361,-1},
        {40361,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40287,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40363,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40450,-1},
        {40450,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32883,40287},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40441,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36436,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40407,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40372,-1},
        {40372,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40423,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40285,28982},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31461,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23507,40356},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35937,40292},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22575,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33308,40311},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36033,40399},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33744,40388},
        {-1,-1,-1}
    },
    {
        {19719,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22887,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30552,40423},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32681,40285},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40300,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30089,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29141,-1},
        {29141,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40199,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30090,40455},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40210,-1},
        {40210,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40792,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30813,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37276,40565},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40571,31988},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40599,-1},
        {40599,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37148,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40584,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40607,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31108,40575},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31108,40575},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32666,-1},
        {32666,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40617,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40616,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40616,33314},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40628,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40617,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36849,40621},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40617,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40655,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40655,-1},
        {31896,40655,-1}
    },
    {
        {-1,-1},
        {-1},
        {39200,39173},
        {39173,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37586,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40617,37582},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40617,-1},
        {40617,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31262,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31262,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30874,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40613,40654},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27544,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31255,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27086,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40620,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40613,37782},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37848,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37841,40629},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37892,40623},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37892,40623},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37300,40623},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29100,-1},
        {29100,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33945,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40635,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40635,24460},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38753,31404},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40643,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40643,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40643,39630},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40643,35318},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40655,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29314,-1},
        {29314,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37676,40655},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37624,40655},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32202,40653},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40656,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26348,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31282,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40656,37847},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31328,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40677,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30383,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40660,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40667,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40664,30491},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40672,32966},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40679,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33398,40671},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27737,27738},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25140,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39417,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40679,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39908,40686},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36367,40687},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39930,40687},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30077,-1},
        {30077,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40686,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40026,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36481,40677},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40032,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {19831,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34224,32386},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34821,-1},
        {-1,-1,-1}
    },
    {
        {19831,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40712,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34584,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30010,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36682,40718},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25290,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40660,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39811,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39409,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39622,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40748,39470},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34884,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30384,-1},
        {30384,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22159,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40769,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28053,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40032,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22159,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40779,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40788,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40801,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40803,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40809,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40818,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35010,40807},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40812,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40809,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40816,40594},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40800,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35313,40807},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40816,-1},
        {40816,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40569,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40816,30929},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40816,30918},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40799,-1},
        {40799,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40816,36094},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40809,-1},
        {40809,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40572,40822},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40045,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25130,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31840,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40861,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40860,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40860,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40861,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21561,-1},
        {21561,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31722,-1},
        {31722,-1,-1}
    },
    {
        {19968,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {19969,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {19970,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {19971,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {19978,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {19979,-1,-1}
    },
    {
        {19975,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {19975,-1},
        {-1},
        {-1,-1},
        {19975,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {19976,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {19976,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {19977,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {19978,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {19979,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20854,-1,-1}
    },
    {
        {19981,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {19982,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {19982,-1},
        {-1},
        {-1,-1},
        {19982,33287,-1}
    },
    {
        {19982,-1},
        {-1},
        {-1,-1},
        {19982,33287,-1}
    },
    {
        {19983,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {19984,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {19985,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {19986,-1},
        {-1},
        {-1,-1},
        {19985,-1,-1}
    },
    {
        {19986,-1},
        {-1},
        {-1,-1},
        {19985,-1,-1}
    },
    {
        {19986,-1},
        {-1},
        {-1,-1},
        {19985,-1,-1}
    },
    {
        {19986,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {19986,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {19988,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {19989,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {19990,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {19991,-1},
        {-1},
        {-1,-1},
        {21317,19990,-1}
    },
    {
        {19992,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {19993,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {19993,-1},
        {-1},
        {-1,-1},
        {19993,-1,-1}
    },
    {
        {19993,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26989,21271},
        {27604,21271,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {19998,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {19992,-1},
        {19992,-1,-1}
    },
    {
        {20001,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37193,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20841,20001,-1}
    },
    {
        {20006,-1},
        {-1},
        {-1,-1},
        {31453,20006,-1}
    },
    {
        {20006,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20008,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20009,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20010,-1},
        {-1},
        {-1,-1},
        {20171,-1,-1}
    },
    {
        {20011,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20012,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20013,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20014,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20015,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20016,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20016,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20017,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20018,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20018,24343},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20022,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20024,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20024,-1},
        {-1},
        {-1,-1},
        {20024,-1,-1}
    },
    {
        {20025,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20027,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20027,-1},
        {-1},
        {-1,-1},
        {20027,-1,-1}
    },
    {
        {20028,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20791,40599},
        {40599,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20791,40599},
        {40599,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20031,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20032,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21450,-1,-1}
    },
    {
        {20034,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20034,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20034,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20035,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {12294,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {12294,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20037,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20037,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20037,-1},
        {20037,-1,-1}
    },
    {
        {20039,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20040,-1},
        {-1},
        {-1,-1},
        {24186,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20736,32681},
        {32681,-1,-1}
    },
    {
        {20043,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20045,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20046,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20047,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20049,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {19992,20853},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {19992,20853},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23968,21932},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20053,-1},
        {-1},
        {-1,-1},
        {34382,-1,-1}
    },
    {
        {20053,-1},
        {-1},
        {-1,-1},
        {34382,-1,-1}
    },
    {
        {20053,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20054,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20055,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20056,-1},
        {-1},
        {-1,-1},
        {20055,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20057,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20058,-1},
        {-1},
        {-1,-1},
        {38560,38577,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20061,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20062,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20063,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20066,-1},
        {-1},
        {-1,-1},
        {33995,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22987,-1,-1}
    },
    {
        {20073,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20291,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20081,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20083,-1},
        {-1},
        {-1,-1},
        {20083,-1,-1}
    },
    {
        {20083,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20094,-1,-1}
    },
    {
        {20094,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20096,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24185,20094},
        {20094,-1,-1}
    },
    {
        {20098,-1},
        {-1},
        {-1,-1},
        {20081,-1,-1}
    },
    {
        {20098,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31932,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22777,-1},
        {-1,-1,-1}
    },
    {
        {20101,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20102,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20104,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20105,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20106,-1},
        {-1},
        {-1,-1},
        {20107,-1,-1}
    },
    {
        {20107,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20108,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20109,-1},
        {-1},
        {-1,-1},
        {24178,-1,-1}
    },
    {
        {20110,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20110,34407},
        {20110,-1,-1}
    },
    {
        {20113,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20114,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20854,-1},
        {20854,-1,-1}
    },
    {
        {20116,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20117,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22235,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27507,-1,-1}
    },
    {
        {20120,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20121,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20123,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20124,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20125,-1},
        {-1},
        {-1,-1},
        {25993,40778,-1}
    },
    {
        {20126,-1},
        {-1},
        {-1,-1},
        {20124,-1,-1}
    },
    {
        {20127,-1},
        {-1},
        {-1,-1},
        {19998,-1,-1}
    },
    {
        {20127,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20127,-1},
        {-1},
        {-1,-1},
        {19998,-1,-1}
    },
    {
        {20128,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20128,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20129,-1},
        {-1},
        {-1,-1},
        {20129,-1,-1}
    },
    {
        {20129,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20130,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20130,-1},
        {-1},
        {-1,-1},
        {20130,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22823,-1},
        {22823,-1,-1}
    },
    {
        {20132,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20132,-1},
        {-1},
        {-1,-1},
        {20132,-1,-1}
    },
    {
        {20132,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20133,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20133,-1},
        {-1},
        {-1,-1},
        {20133,-1,-1}
    },
    {
        {20134,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20136,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20139,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20140,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20141,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20142,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20142,-1},
        {-1},
        {-1,-1},
        {20142,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20139,-1,-1}
    },
    {
        {20144,-1},
        {-1},
        {-1,-1},
        {20140,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22812,25494},
        {22812,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35242,-1},
        {35242,-1,-1}
    },
    {
        {20147,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {19432,-1},
        {-1,-1,-1}
    },
    {
        {20150,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20150,-1},
        {-1},
        {-1,-1},
        {20150,-1,-1}
    },
    {
        {20150,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24265,27087},
        {24265,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24265,27087},
        {24265,-1,-1}
    },
    {
        {20153,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20153,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20153,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20154,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20155,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20156,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20160,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20161,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20162,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20163,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20164,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20184,20677},
        {20184,-1,-1}
    },
    {
        {20166,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20167,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20167,-1,-1}
    },
    {
        {20168,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25484,-1},
        {-1,-1,-1}
    },
    {
        {20170,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20171,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20912,-1},
        {20912,-1,-1}
    },
    {
        {20173,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20174,-1},
        {-1},
        {-1,-1},
        {24467,24478,-1}
    },
    {
        {20175,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20176,-1},
        {-1},
        {-1,-1},
        {20632,-1,-1}
    },
    {
        {20180,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20181,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20182,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20183,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20183,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20184,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20185,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20186,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20181,-1},
        {-1,-1,-1}
    },
    {
        {20189,-1},
        {-1},
        {-1,-1},
        {21516,-1,-1}
    },
    {
        {20190,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20190,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20191,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20193,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20195,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20196,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20197,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38263,-1},
        {38263,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20203,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20205,-1},
        {-1},
        {-1,-1},
        {20190,-1,-1}
    },
    {
        {20205,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20206,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20208,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20209,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20210,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20350,-1},
        {-1,-1,-1}
    },
    {
        {20213,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20214,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20215,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20315,20175},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20016,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20489,-1},
        {20489,-1,-1}
    },
    {
        {20219,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20219,-1},
        {-1,-1,-1}
    },
    {
        {20221,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20302,-1,-1}
    },
    {
        {20223,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20224,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20224,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20225,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20227,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20315,24340},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24563,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27797,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24441,-1},
        {24441,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20233,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20234,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20235,-1},
        {-1},
        {-1,-1},
        {20235,-1,-1}
    },
    {
        {20235,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20237,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20238,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20239,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20240,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20241,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20286,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27785,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20246,-1},
        {-1},
        {-1,-1},
        {24081,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28672,20778},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28779,-1,-1}
    },
    {
        {20250,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {151146,20660},
        {20660,-1,-1}
    },
    {
        {20252,-1},
        {-1},
        {-1,-1},
        {20485,-1,-1}
    },
    {
        {20253,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24455,-1},
        {24455,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29572,20357},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21162,-1,-1}
    },
    {
        {20271,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20272,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20320,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20320,-1},
        {20320,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20276,-1},
        {-1},
        {-1,-1},
        {20276,-1,-1}
    },
    {
        {20276,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22235,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20278,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20278,-1},
        {-1},
        {-1,-1},
        {20278,-1,-1}
    },
    {
        {20279,-1},
        {-1},
        {-1,-1},
        {32964,-1,-1}
    },
    {
        {20280,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20294,-1,-1}
    },
    {
        {20282,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20284,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20285,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20286,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26547,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20284,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20284,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22987,20284},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24449,-1,-1}
    },
    {
        {20291,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37219,-1},
        {37219,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26151,-1},
        {-1,-1,-1}
    },
    {
        {20294,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20295,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20296,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20297,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24444,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20301,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20302,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20303,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20303,-1},
        {-1},
        {-1,-1},
        {20303,-1,-1}
    },
    {
        {20304,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20305,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20307,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20307,-1},
        {-1},
        {-1,-1},
        {20307,-1,-1}
    },
    {
        {20308,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20309,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20310,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20311,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20312,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20313,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20314,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20315,-1},
        {-1},
        {-1,-1},
        {20175,-1,-1}
    },
    {
        {20316,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20317,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20318,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20319,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20320,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20320,-1},
        {-1},
        {-1,-1},
        {20320,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20185,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20324,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28846,-1},
        {-1,-1,-1}
    },
    {
        {20329,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20330,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20330,-1},
        {-1},
        {-1,-1},
        {20330,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20332,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38477,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38477,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20335,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20336,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27861,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37225,-1},
        {37225,-1,-1}
    },
    {
        {20339,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32827,-1},
        {-1,-1,-1}
    },
    {
        {20341,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20342,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20343,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20344,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20346,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20346,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20347,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20348,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20348,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20349,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20349,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20350,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20350,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20351,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20351,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24418,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20354,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20355,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20357,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20358,-1},
        {-1},
        {-1,-1},
        {26469,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22839,-1},
        {-1,-1,-1}
    },
    {
        {20360,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20362,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20363,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38512,-1,-1}
    },
    {
        {20365,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20367,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20369,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20370,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24459,-1,-1}
    },
    {
        {20372,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20385,-1},
        {-1,-1,-1}
    },
    {
        {20374,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20375,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20376,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20378,-1},
        {-1},
        {-1,-1},
        {24455,-1,-1}
    },
    {
        {20379,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20381,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34214,-1,-1}
    },
    {
        {20384,-1},
        {-1},
        {-1,-1},
        {20448,-1,-1}
    },
    {
        {20385,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28900,-1},
        {-1,-1,-1}
    },
    {
        {20395,-1},
        {-1},
        {-1,-1},
        {20318,-1,-1}
    },
    {
        {20395,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20397,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20398,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20399,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31384,-1},
        {31384,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20402,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25387,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20842,-1},
        {-1,-1,-1}
    },
    {
        {20405,-1},
        {-1},
        {-1,-1},
        {20405,-1,-1}
    },
    {
        {20405,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20406,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27193,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28052,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30007,-1,-1}
    },
    {
        {20414,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20415,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20415,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20451,-1},
        {-1},
        {-1,-1},
        {20451,-1,-1}
    },
    {
        {20418,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20419,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20420,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20421,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29378,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22195,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20425,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20426,-1},
        {-1},
        {-1,-1},
        {20426,-1,-1}
    },
    {
        {20426,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20427,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36628,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20430,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20431,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20432,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20433,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20434,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20436,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23435,20519},
        {-1,-1,-1}
    },
    {
        {20439,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20440,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20440,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24725,-1},
        {-1,-1,-1}
    },
    {
        {20442,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20443,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20445,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20842,-1},
        {-1},
        {20842,-1},
        {20842,-1,-1}
    },
    {
        {20842,-1},
        {-1},
        {20842,-1},
        {20842,-1,-1}
    },
    {
        {20842,-1},
        {-1},
        {20842,-1},
        {-1,-1,-1}
    },
    {
        {20447,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20448,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20449,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20462,-1,-1}
    },
    {
        {20451,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20452,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20453,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20454,-1},
        {-1},
        {-1,-1},
        {20756,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20462,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20463,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20465,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20466,-1},
        {-1},
        {-1,-1},
        {20634,-1,-1}
    },
    {
        {20467,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28154,-1},
        {-1,-1,-1}
    },
    {
        {20469,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20470,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28125,-1},
        {-1,-1,-1}
    },
    {
        {20472,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26895,-1},
        {-1,-1,-1}
    },
    {
        {20474,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20633,33889},
        {20633,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39739,31909,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27427,-1,-1}
    },
    {
        {20478,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20480,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20481,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20482,-1},
        {-1},
        {-1,-1},
        {20341,-1,-1}
    },
    {
        {20482,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20485,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20486,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26469,20358},
        {24480,-1,-1}
    },
    {
        {20489,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20588,-1,-1}
    },
    {
        {20491,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20493,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20494,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20495,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20495,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20496,-1},
        {-1},
        {-1,-1},
        {20495,-1,-1}
    },
    {
        {20496,-1},
        {-1},
        {-1,-1},
        {20495,-1,-1}
    },
    {
        {20497,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20498,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20499,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20500,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20502,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20502,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20504,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20505,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20506,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20508,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20510,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20511,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24799,-1},
        {-1,-1,-1}
    },
    {
        {20513,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20514,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20515,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20516,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20517,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20517,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20518,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20518,-1},
        {-1},
        {-1,-1},
        {20518,-1,-1}
    },
    {
        {20519,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20520,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20521,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20521,-1},
        {-1},
        {-1,-1},
        {20521,-1,-1}
    },
    {
        {20522,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20523,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20524,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20525,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20526,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28126,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20426,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20426,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28956,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22868,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20534,-1},
        {-1},
        {-1,-1},
        {20465,-1,-1}
    },
    {
        {20534,-1},
        {-1},
        {-1,-1},
        {20465,20534,-1}
    },
    {
        {-1,-1},
        {-1},
        {20320,-1},
        {20320,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25505,-1},
        {-1,-1,-1}
    },
    {
        {20537,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20539,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25509,-1},
        {20525,-1,-1}
    },
    {
        {20544,-1},
        {-1},
        {-1,-1},
        {20544,33521,-1}
    },
    {
        {20544,-1},
        {-1},
        {-1,-1},
        {20544,33521,-1}
    },
    {
        {20545,-1},
        {-1},
        {-1,-1},
        {31216,31281,-1}
    },
    {
        {20546,-1},
        {-1},
        {-1,-1},
        {27500,21069,-1}
    },
    {
        {20546,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20547,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20547,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20754,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20550,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20551,-1},
        {-1},
        {-1,-1},
        {20206,-1,-1}
    },
    {
        {20552,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20552,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20553,-1},
        {-1},
        {-1,-1},
        {20553,-1,-1}
    },
    {
        {20553,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20553,-1},
        {-1},
        {-1,-1},
        {20553,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23631,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23198,-1},
        {-1,-1,-1}
    },
    {
        {20558,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20559,-1},
        {-1},
        {-1,-1},
        {20559,-1,-1}
    },
    {
        {20559,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20560,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20560,-1},
        {-1},
        {-1,-1},
        {20560,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20563,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20565,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20566,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20566,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20567,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20355,-1},
        {20355,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20570,-1},
        {-1},
        {-1,-1},
        {20316,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25554,-1},
        {-1,-1,-1}
    },
    {
        {20572,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32972,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24821,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20579,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29494,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20581,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20581,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20582,-1},
        {-1},
        {-1,-1},
        {32997,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26612,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36000,-1},
        {-1,-1,-1}
    },
    {
        {20586,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24453,-1},
        {-1,-1,-1}
    },
    {
        {20588,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25558,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21696,-1},
        {-1,-1,-1}
    },
    {
        {20592,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20592,-1},
        {-1},
        {-1,-1},
        {20592,-1,-1}
    },
    {
        {20592,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20594,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24820,-1},
        {-1,-1,-1}
    },
    {
        {20596,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20597,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20598,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20600,-1},
        {-1},
        {20600,-1},
        {20600,-1,-1}
    },
    {
        {20600,-1},
        {-1},
        {20600,-1},
        {-1,-1,-1}
    },
    {
        {20600,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20633,-1},
        {20633,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23105,20674},
        {20674,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20514,-1},
        {20514,-1,-1}
    },
    {
        {20605,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20608,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20613,-1},
        {-1},
        {-1,-1},
        {20659,20253,-1}
    },
    {
        {20613,-1},
        {-1},
        {-1,-1},
        {20659,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33592,25558},
        {-1,-1,-1}
    },
    {
        {20616,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36785,28349},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32629,21966},
        {32629,-1,-1}
    },
    {
        {20621,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20621,-1},
        {-1},
        {-1,-1},
        {20621,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20625,-1},
        {-1},
        {-1,-1},
        {20625,-1,-1}
    },
    {
        {20625,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20625,-1},
        {-1},
        {-1,-1},
        {20625,-1,-1}
    },
    {
        {20626,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20627,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20628,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20628,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20629,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20630,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20632,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20633,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20633,-1},
        {-1},
        {-1,-1},
        {20633,-1,-1}
    },
    {
        {20634,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20636,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25640,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20006,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20643,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23578,20795},
        {20795,-1,-1}
    },
    {
        {20650,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28463,-1},
        {-1,-1,-1}
    },
    {
        {20652,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20653,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26361,36973},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20588,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20588,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20657,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20658,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20659,20253,-1}
    },
    {
        {20659,-1},
        {-1},
        {-1,-1},
        {20253,-1,-1}
    },
    {
        {20659,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20660,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20660,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20661,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27117,-1},
        {-1,-1,-1}
    },
    {
        {20663,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29245,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20666,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20667,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20667,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29103,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20670,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37154,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24713,-1},
        {-1,-1,-1}
    },
    {
        {20674,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20633,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20633,-1,-1}
    },
    {
        {20676,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20677,-1},
        {-1},
        {-1,-1},
        {20677,-1,-1}
    },
    {
        {20677,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28451,-1},
        {-1,-1,-1}
    },
    {
        {20679,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24930,28459},
        {-1,-1,-1}
    },
    {
        {20681,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20682,-1},
        {-1},
        {-1,-1},
        {20185,20682,-1}
    },
    {
        {20682,-1},
        {-1},
        {-1,-1},
        {20185,20682,-1}
    },
    {
        {20682,-1},
        {-1},
        {-1,-1},
        {20185,-1,-1}
    },
    {
        {20682,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36010,-1},
        {-1,-1,-1}
    },
    {
        {20684,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20685,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20686,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20687,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20687,-1},
        {-1},
        {-1,-1},
        {20687,-1,-1}
    },
    {
        {20687,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20688,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20689,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31384,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20692,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20693,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20694,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21172,21214},
        {21172,21214,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20697,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20698,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27450,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27450,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20702,-1},
        {-1},
        {-1,-1},
        {20605,-1,-1}
    },
    {
        {20702,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30058,-1},
        {30058,-1,-1}
    },
    {
        {20705,-1},
        {-1},
        {-1,-1},
        {24800,24693,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20707,-1},
        {-1},
        {-1,-1},
        {20717,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20709,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20710,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20711,-1},
        {-1},
        {-1,-1},
        {64049,20711,-1}
    },
    {
        {20711,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20711,-1},
        {-1},
        {-1,-1},
        {20711,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20713,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35662,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24801,24746},
        {24801,24746,-1}
    },
    {
        {-1,-1},
        {-1},
        {24980,-1},
        {-1,-1,-1}
    },
    {
        {20717,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20717,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20718,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20718,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26840,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38599,-1,-1}
    },
    {
        {20722,-1},
        {-1},
        {-1,-1},
        {20185,20682,-1}
    },
    {
        {20722,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20722,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20723,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20713,-1},
        {-1},
        {20713,-1},
        {20713,-1,-1}
    },
    {
        {20725,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {145215,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35167,-1},
        {-1,-1,-1}
    },
    {
        {20729,-1},
        {-1},
        {-1,-1},
        {20385,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20731,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28583,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20736,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20737,-1},
        {-1},
        {-1,-1},
        {20426,-1,-1}
    },
    {
        {20738,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20739,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20740,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20740,-1},
        {-1},
        {-1,-1},
        {20740,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24403,30070},
        {-1,-1,-1}
    },
    {
        {20742,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20742,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20744,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20745,-1},
        {-1},
        {-1,-1},
        {20537,-1,-1}
    },
    {
        {20747,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20667,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36617,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38576,-1},
        {-1,-1,-1}
    },
    {
        {20754,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25825,-1},
        {-1,-1,-1}
    },
    {
        {20756,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20756,-1},
        {-1},
        {-1,-1},
        {20756,-1,-1}
    },
    {
        {20757,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20758,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20759,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20760,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20762,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20762,-1},
        {-1},
        {-1,-1},
        {20762,-1,-1}
    },
    {
        {20763,-1},
        {-1},
        {-1,-1},
        {20763,33310,-1}
    },
    {
        {20763,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20766,-1},
        {-1},
        {-1,-1},
        {29246,-1,-1}
    },
    {
        {20767,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20769,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22745,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29190,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35263,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35263,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28425,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28556,-1},
        {28556,-1,-1}
    },
    {
        {20778,-1},
        {-1},
        {-1,-1},
        {20778,-1,-1}
    },
    {
        {20778,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20778,-1},
        {-1},
        {-1,-1},
        {20778,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35946,-1},
        {35946,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20786,-1},
        {-1},
        {-1,-1},
        {20786,-1,-1}
    },
    {
        {20786,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20789,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25884,-1},
        {-1,-1,-1}
    },
    {
        {20791,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20794,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20794,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20795,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20795,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20796,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35064,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20799,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20800,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20800,-1},
        {-1},
        {-1,-1},
        {23586,-1,-1}
    },
    {
        {20801,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31786,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31786,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31786,-1,-1}
    },
    {
        {20803,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20804,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20805,-1},
        {-1},
        {-1,-1},
        {20805,-1,-1}
    },
    {
        {20805,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20806,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20806,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20807,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20808,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20809,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20810,-1},
        {-1},
        {-1,-1},
        {20812,-1,-1}
    },
    {
        {20810,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20810,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20811,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20812,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20812,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20813,-1},
        {-1},
        {-1,-1},
        {20813,-1,-1}
    },
    {
        {20813,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20814,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20814,-1},
        {-1},
        {-1,-1},
        {20814,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38263,-1,-1}
    },
    {
        {20816,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20812,-1},
        {-1},
        {20810,20812},
        {20812,-1,-1}
    },
    {
        {20818,-1},
        {-1},
        {-1,-1},
        {20816,-1,-1}
    },
    {
        {20818,-1},
        {-1},
        {-1,-1},
        {20816,20818,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20820,-1},
        {-1},
        {-1,-1},
        {20814,-1,-1}
    },
    {
        {20820,-1},
        {-1},
        {-1,-1},
        {20820,20814,-1}
    },
    {
        {20820,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20821,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20823,-1},
        {-1},
        {20823,-1},
        {20823,-1,-1}
    },
    {
        {20823,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22987,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23557,-1},
        {-1,-1,-1}
    },
    {
        {20826,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23557,-1},
        {-1,-1,-1}
    },
    {
        {20828,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21194,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31443,23557},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20832,-1},
        {-1},
        {-1,-1},
        {20828,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31457,21194},
        {-1,-1,-1}
    },
    {
        {20834,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29957,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20837,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20129,-1,-1}
    },
    {
        {20869,-1},
        {-1},
        {-1,-1},
        {20869,-1,-1}
    },
    {
        {20840,-1},
        {-1},
        {-1,-1},
        {20840,-1,-1}
    },
    {
        {20840,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20841,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20841,-1},
        {-1},
        {-1,-1},
        {20001,-1,-1}
    },
    {
        {20842,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20842,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20843,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20843,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20844,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20844,-1},
        {-1},
        {-1,-1},
        {20844,-1,-1}
    },
    {
        {20845,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20846,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20846,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20849,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22825,-1},
        {22825,-1,-1}
    },
    {
        {20851,-1},
        {-1},
        {-1,-1},
        {31505,-1,-1}
    },
    {
        {20851,-1},
        {-1},
        {-1,-1},
        {31505,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33288,-1},
        {-1,-1,-1}
    },
    {
        {20853,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20854,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20855,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20855,-1},
        {-1},
        {-1,-1},
        {20855,-1,-1}
    },
    {
        {20855,-1},
        {-1},
        {-1,-1},
        {20855,-1,-1}
    },
    {
        {20856,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21672,33586},
        {33586,-1,-1}
    },
    {
        {20859,-1},
        {-1},
        {-1,-1},
        {32654,-1,-1}
    },
    {
        {20859,-1},
        {-1},
        {-1,-1},
        {32654,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20860,-1},
        {-1},
        {-1,-1},
        {20860,-1,-1}
    },
    {
        {20860,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20860,-1},
        {-1},
        {-1,-1},
        {20860,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20864,-1,-1}
    },
    {
        {20864,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20866,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20867,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20873,-1,-1}
    },
    {
        {20869,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20870,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23713,23831},
        {-1,-1,-1}
    },
    {
        {20873,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20873,-1},
        {-1},
        {-1,-1},
        {20873,-1,-1}
    },
    {
        {20874,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20875,-1},
        {-1},
        {-1,-1},
        {20866,-1,-1}
    },
    {
        {20876,-1},
        {-1},
        {-1,-1},
        {20874,-1,-1}
    },
    {
        {20877,-1},
        {-1},
        {-1,-1},
        {20877,-1,-1}
    },
    {
        {20877,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20879,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20880,-1},
        {-1},
        {-1,-1},
        {20882,-1,-1}
    },
    {
        {20881,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20882,-1},
        {-1},
        {-1,-1},
        {20882,-1,-1}
    },
    {
        {20882,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20883,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20883,-1},
        {-1},
        {-1,-1},
        {20883,-1,-1}
    },
    {
        {20883,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23563,-1},
        {-1,-1,-1}
    },
    {
        {20885,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20885,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20885,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20885,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20886,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20887,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20887,-1},
        {-1},
        {-1,-1},
        {20887,-1,-1}
    },
    {
        {20888,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20889,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20893,-1},
        {-1},
        {-1,-1},
        {23452,-1,-1}
    },
    {
        {20893,-1},
        {-1},
        {-1,-1},
        {23459,23452,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30336,-1},
        {-1,-1,-1}
    },
    {
        {20896,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20898,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20898,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20898,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20898,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20899,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20900,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20900,-1},
        {-1},
        {-1,-1},
        {20900,-1,-1}
    },
    {
        {20900,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20900,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20901,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20902,-1},
        {-1},
        {-1,-1},
        {23495,-1,-1}
    },
    {
        {20902,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20904,-1},
        {-1},
        {-1,-1},
        {23500,-1,-1}
    },
    {
        {20905,-1},
        {-1},
        {-1,-1},
        {20889,-1,-1}
    },
    {
        {20906,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20906,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20907,-1},
        {-1},
        {-1,-1},
        {20912,-1,-1}
    },
    {
        {20908,-1},
        {-1},
        {-1,-1},
        {20908,-1,-1}
    },
    {
        {20908,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20909,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27743,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20912,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20913,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20914,-1},
        {-1},
        {-1,-1},
        {27798,-1,-1}
    },
    {
        {20915,-1},
        {-1},
        {-1,-1},
        {27770,-1,-1}
    },
    {
        {20916,-1},
        {-1},
        {-1,-1},
        {20916,-1,-1}
    },
    {
        {20916,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20916,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20916,-1},
        {-1},
        {-1,-1},
        {20916,-1,-1}
    },
    {
        {20917,-1},
        {-1},
        {-1,-1},
        {27841,-1,-1}
    },
    {
        {20917,-1},
        {-1},
        {-1,-1},
        {27841,-1,-1}
    },
    {
        {20918,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20919,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27887,-1,-1}
    },
    {
        {20924,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20925,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27941,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28049,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20931,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20932,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20933,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20934,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28126,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20937,-1},
        {-1},
        {-1,-1},
        {28092,-1,-1}
    },
    {
        {20938,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20938,-1},
        {-1},
        {-1,-1},
        {20938,-1,-1}
    },
    {
        {20939,-1},
        {-1},
        {-1,-1},
        {20939,-1,-1}
    },
    {
        {20939,-1},
        {-1},
        {-1,-1},
        {20939,-1,-1}
    },
    {
        {20940,-1},
        {-1},
        {-1,-1},
        {20940,-1,-1}
    },
    {
        {20940,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20941,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28187,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28234,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28356,24884},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20950,-1},
        {-1},
        {-1,-1},
        {28310,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28476,-1},
        {28476,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28580,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31105,-1},
        {-1,-1,-1}
    },
    {
        {20955,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20956,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20957,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20958,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20958,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28678,-1},
        {-1,-1,-1}
    },
    {
        {20960,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20961,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20962,-1},
        {-1},
        {-1,-1},
        {20961,-1,-1}
    },
    {
        {20962,-1},
        {-1},
        {-1,-1},
        {20961,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40179,40300},
        {40179,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23621,-1,-1}
    },
    {
        {20966,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20967,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20969,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20970,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39080,20974},
        {39080,-1,-1}
    },
    {
        {20973,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20974,-1},
        {-1},
        {-1,-1},
        {39080,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20976,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20977,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20979,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20980,-1},
        {-1},
        {-1,-1},
        {20973,-1,-1}
    },
    {
        {20981,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20981,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20982,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20984,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20985,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20986,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20989,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20989,-1},
        {-1},
        {-1,-1},
        {20989,-1,-1}
    },
    {
        {20990,-1},
        {-1},
        {-1,-1},
        {20989,-1,-1}
    },
    {
        {20990,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20992,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20993,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20994,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20995,-1},
        {-1},
        {-1,-1},
        {20995,-1,-1}
    },
    {
        {20995,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20996,-1},
        {-1},
        {-1,-1},
        {20995,-1,-1}
    },
    {
        {20996,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21109,-1,-1}
    },
    {
        {20998,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20998,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20999,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20999,-1,-1}
    },
    {
        {20999,-1},
        {-1},
        {-1,-1},
        {20999,21151,-1}
    },
    {
        {-1,-1},
        {-1},
        {21151,-1},
        {20999,-1,-1}
    },
    {
        {21000,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24190,-1},
        {-1,-1,-1}
    },
    {
        {21002,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21003,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21003,-1},
        {-1},
        {-1,-1},
        {21003,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24534,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33467,33965},
        {33467,-1,-1}
    },
    {
        {21006,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24190,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24420,-1},
        {-1,-1,-1}
    },
    {
        {21009,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21010,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21011,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21011,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21012,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21013,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21014,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21015,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21129,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25282,25173},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21021,-1},
        {21021,-1,-1}
    },
    {
        {21021,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37455,-1},
        {37455,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21165,-1},
        {-1,-1,-1}
    },
    {
        {21034,-1},
        {-1},
        {21034,-1},
        {21034,-1,-1}
    },
    {
        {21034,-1},
        {-1},
        {21034,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37040,21038},
        {-1,-1,-1}
    },
    {
        {21028,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21028,-1},
        {-1},
        {-1,-1},
        {21028,-1,-1}
    },
    {
        {21028,-1},
        {-1},
        {-1,-1},
        {21028,-1,-1}
    },
    {
        {21029,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21163,21031},
        {21163,-1,-1}
    },
    {
        {21031,-1},
        {-1},
        {-1,-1},
        {21163,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21033,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21034,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21029,-1},
        {-1},
        {21029,24487},
        {21029,-1,-1}
    },
    {
        {21038,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21040,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21043,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21049,26422},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21046,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21047,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21048,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21048,-1},
        {-1},
        {-1,-1},
        {21048,-1,-1}
    },
    {
        {21049,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21050,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21051,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21051,-1},
        {-1},
        {-1,-1},
        {21051,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21163,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21050,-1},
        {21050,-1,-1}
    },
    {
        {21059,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21060,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21063,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32921,21066},
        {-1,-1,-1}
    },
    {
        {21065,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21066,-1},
        {-1},
        {-1,-1},
        {21066,-1,-1}
    },
    {
        {21066,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21067,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21068,-1},
        {-1},
        {-1,-1},
        {21050,-1,-1}
    },
    {
        {21069,-1},
        {-1},
        {-1,-1},
        {21069,-1,-1}
    },
    {
        {21069,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21049,-1},
        {-1},
        {21049,-1},
        {21049,-1,-1}
    },
    {
        {21071,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21071,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37679,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26819,-1},
        {-1,-1,-1}
    },
    {
        {21076,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21077,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21078,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21079,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21071,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21083,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21084,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21085,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21086,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21089,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26027,-1},
        {-1,-1,-1}
    },
    {
        {21091,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21092,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21093,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38329,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30729,-1},
        {30729,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30729,-1},
        {30729,-1,-1}
    },
    {
        {21097,-1},
        {-1},
        {-1,-1},
        {21104,-1,-1}
    },
    {
        {21098,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21098,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24230,-1},
        {-1,-1,-1}
    },
    {
        {21100,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23627,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21103,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21104,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21105,-1},
        {-1},
        {-1,-1},
        {21091,21133,-1}
    },
    {
        {21105,-1},
        {-1},
        {-1,-1},
        {21138,21091,-1}
    },
    {
        {21105,-1},
        {-1},
        {-1,-1},
        {21091,21133,-1}
    },
    {
        {21106,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21106,-1},
        {-1},
        {-1,-1},
        {21106,-1,-1}
    },
    {
        {21106,-1},
        {-1},
        {-1,-1},
        {21106,-1,-1}
    },
    {
        {21107,-1},
        {-1},
        {-1,-1},
        {21124,-1,-1}
    },
    {
        {21107,-1},
        {-1},
        {-1,-1},
        {21107,21124,-1}
    },
    {
        {21108,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21109,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21109,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21085,-1},
        {-1,-1,-1}
    },
    {
        {21111,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25134,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26819,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23849,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21119,-1},
        {-1,-1,-1}
    },
    {
        {21117,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27472,-1},
        {-1,-1,-1}
    },
    {
        {21119,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21119,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28966,-1},
        {-1,-1,-1}
    },
    {
        {21122,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21123,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21123,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21124,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37772,-1,-1}
    },
    {
        {21127,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21128,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21129,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21129,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25220,21119},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21133,-1},
        {-1},
        {-1,-1},
        {21091,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21091,21133},
        {21091,21133,-1}
    },
    {
        {-1,-1},
        {-1},
        {21123,31339},
        {21123,-1,-1}
    },
    {
        {21137,-1},
        {-1},
        {-1,-1},
        {21092,-1,-1}
    },
    {
        {21137,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21138,-1},
        {-1},
        {-1,-1},
        {21091,-1,-1}
    },
    {
        {21138,-1},
        {-1},
        {-1,-1},
        {21138,21091,-1}
    },
    {
        {21139,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21140,-1},
        {-1},
        {-1,-1},
        {21091,-1,-1}
    },
    {
        {21140,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21144,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21144,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34849,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26040,-1,-1}
    },
    {
        {21147,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21240,21223},
        {21240,21223,-1}
    },
    {
        {-1,-1},
        {-1},
        {36776,36774},
        {36774,-1,-1}
    },
    {
        {21151,-1},
        {-1},
        {-1,-1},
        {20999,-1,-1}
    },
    {
        {21152,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21155,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21156,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21158,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21161,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21162,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21163,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21164,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21165,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36920,-1},
        {-1,-1,-1}
    },
    {
        {21167,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21169,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21172,21214},
        {21172,21214,-1}
    },
    {
        {21172,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21173,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21173,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21177,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21178,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21179,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21180,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21182,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21182,-1},
        {-1},
        {-1,-1},
        {21182,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21218,13599},
        {21218,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21185,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21180,-1},
        {21578,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21187,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64065,25935,-1}
    },
    {
        {21189,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21191,-1},
        {-1},
        {-1,-1},
        {21191,-1,-1}
    },
    {
        {21191,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21192,-1},
        {-1},
        {-1,-1},
        {21191,-1,-1}
    },
    {
        {21193,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21194,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21196,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21196,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21196,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21197,-1},
        {-1},
        {-1,-1},
        {20510,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21216,-1},
        {21216,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21200,-1},
        {-1},
        {-1,-1},
        {29467,-1,-1}
    },
    {
        {21201,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21202,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21202,-1},
        {-1},
        {-1,-1},
        {21202,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21182,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21205,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21206,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21206,-1},
        {-1},
        {-1,-1},
        {21206,-1,-1}
    },
    {
        {21207,-1},
        {-1},
        {-1,-1},
        {21206,21207,-1}
    },
    {
        {21207,-1},
        {-1},
        {-1,-1},
        {21206,-1,-1}
    },
    {
        {21208,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21209,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21211,-1},
        {-1},
        {-1,-1},
        {21234,21235,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21213,-1},
        {-1},
        {-1,-1},
        {21213,-1,-1}
    },
    {
        {21213,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21214,-1},
        {-1},
        {-1,-1},
        {21172,-1,-1}
    },
    {
        {21215,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21215,-1},
        {-1},
        {-1,-1},
        {21215,-1,-1}
    },
    {
        {21216,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21216,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21218,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21219,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21220,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21222,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21222,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21223,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37169,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21205,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21127,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20150,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21232,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21234,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21234,-1},
        {-1},
        {-1,-1},
        {21235,-1,-1}
    },
    {
        {21235,-1},
        {-1},
        {-1,-1},
        {21234,-1,-1}
    },
    {
        {21235,-1},
        {-1},
        {-1,-1},
        {21235,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21237,-1},
        {-1},
        {-1,-1},
        {21169,-1,-1}
    },
    {
        {21237,-1},
        {-1},
        {-1,-1},
        {21169,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21239,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21240,-1},
        {-1},
        {-1,-1},
        {21223,-1,-1}
    },
    {
        {21240,-1},
        {-1},
        {-1,-1},
        {21223,-1,-1}
    },
    {
        {21241,-1},
        {-1},
        {-1,-1},
        {21253,-1,-1}
    },
    {
        {21241,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21242,-1},
        {-1},
        {-1,-1},
        {21242,-1,-1}
    },
    {
        {21242,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21243,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25937,40169},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25265,-1},
        {-1,-1,-1}
    },
    {
        {21246,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21247,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21248,-1},
        {-1},
        {-1,-1},
        {21243,-1,-1}
    },
    {
        {21249,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21250,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {19984,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {19984,-1},
        {-1,-1,-1}
    },
    {
        {21253,-1},
        {-1},
        {-1,-1},
        {21253,-1,-1}
    },
    {
        {21253,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21254,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21254,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21255,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21255,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21256,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21258,24179},
        {-1,-1,-1}
    },
    {
        {21258,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21259,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21261,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21264,22852},
        {-1,-1,-1}
    },
    {
        {21263,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21263,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21264,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33605,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21269,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21270,-1},
        {-1},
        {-1,-1},
        {21270,-1,-1}
    },
    {
        {21270,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21271,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21271,-1},
        {-1},
        {-1,-1},
        {21271,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21273,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21274,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26601,-1,-1}
    },
    {
        {21276,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21277,-1},
        {-1},
        {-1,-1},
        {24064,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21280,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21280,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21281,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21283,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21284,-1},
        {-1},
        {-1,-1},
        {21281,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34101,34255,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21281,-1,-1}
    },
    {
        {21290,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31755,-1,-1}
    },
    {
        {21295,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21297,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21299,-1,-1}
    },
    {
        {21299,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31639,-1},
        {-1,-1,-1}
    },
    {
        {21301,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21301,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26601,-1},
        {26601,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26601,-1},
        {26601,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30655,-1},
        {-1,-1,-1}
    },
    {
        {21304,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21305,-1},
        {-1},
        {-1,-1},
        {21305,-1,-1}
    },
    {
        {21305,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21306,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21307,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21307,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21310,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21310,-1},
        {-1},
        {-1,-1},
        {21310,-1,-1}
    },
    {
        {21310,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21311,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21311,-1},
        {-1},
        {-1,-1},
        {21311,-1,-1}
    },
    {
        {21312,-1},
        {-1},
        {-1,-1},
        {21306,-1,-1}
    },
    {
        {21312,-1},
        {-1},
        {-1,-1},
        {21312,21306,-1}
    },
    {
        {21313,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21314,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21315,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24319,-1,-1}
    },
    {
        {21317,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21318,-1},
        {-1},
        {-1,-1},
        {21330,-1,-1}
    },
    {
        {21319,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21320,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21321,-1},
        {-1},
        {-1,-1},
        {21321,-1,-1}
    },
    {
        {21322,-1},
        {-1},
        {-1,-1},
        {21322,-1,-1}
    },
    {
        {21322,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {19990,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21325,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21325,-1},
        {21325,-1,-1}
    },
    {
        {21329,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21330,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21330,-1},
        {-1},
        {-1,-1},
        {21330,-1,-1}
    },
    {
        {21331,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21332,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21335,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21336,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21338,-1},
        {-1},
        {-1,-1},
        {21338,-1,-1}
    },
    {
        {21338,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21338,-1},
        {-1},
        {-1,-1},
        {21338,-1,-1}
    },
    {
        {21338,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29575,-1,-1}
    },
    {
        {21340,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31014,-1},
        {20017,-1,-1}
    },
    {
        {21342,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21342,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20073,31293},
        {20073,-1,-1}
    },
    {
        {21344,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21345,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21347,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40565,28407},
        {35199,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35199,-1,-1}
    },
    {
        {21350,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21351,-1},
        {-1},
        {-1,-1},
        {33253,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21353,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31680,21353},
        {-1,-1,-1}
    },
    {
        {21356,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37019,-1,-1}
    },
    {
        {21358,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21359,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21360,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21361,-1},
        {-1},
        {-1,-1},
        {21361,-1,-1}
    },
    {
        {21361,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21361,-1},
        {-1},
        {-1,-1},
        {21361,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21363,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21363,-1},
        {-1},
        {-1,-1},
        {21363,21373,-1}
    },
    {
        {21363,-1},
        {-1},
        {-1,-1},
        {21363,21373,-1}
    },
    {
        {21364,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21365,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21367,-1},
        {-1},
        {-1,-1},
        {24059,-1,-1}
    },
    {
        {21367,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21367,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21367,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21368,-1},
        {-1},
        {-1,-1},
        {21368,-1,-1}
    },
    {
        {21368,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24676,-1,-1}
    },
    {
        {21370,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21371,-1},
        {-1},
        {-1,-1},
        {21364,-1,-1}
    },
    {
        {21371,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26444,-1},
        {26444,-1,-1}
    },
    {
        {21373,-1},
        {-1},
        {-1,-1},
        {21363,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21375,-1},
        {-1},
        {-1,-1},
        {21375,-1,-1}
    },
    {
        {21375,-1},
        {-1},
        {-1,-1},
        {21375,-1,-1}
    },
    {
        {21375,-1},
        {-1},
        {-1,-1},
        {21375,-1,-1}
    },
    {
        {21375,-1},
        {-1},
        {-1,-1},
        {21375,-1,-1}
    },
    {
        {21375,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33181,-1},
        {33181,-1,-1}
    },
    {
        {21378,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21380,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21380,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21378,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30296,21426},
        {21426,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29273,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30757,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31364,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32566,-1},
        {-1,-1,-1}
    },
    {
        {21395,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21398,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21398,-1},
        {-1},
        {-1,-1},
        {21398,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21400,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21402,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22402,-1},
        {-1,-1,-1}
    },
    {
        {21405,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21290,-1},
        {-1,-1,-1}
    },
    {
        {21407,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21408,-1},
        {-1},
        {-1,-1},
        {24257,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21407,-1,-1}
    },
    {
        {21412,-1},
        {-1},
        {-1,-1},
        {26278,26310,-1}
    },
    {
        {21413,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21414,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39003,-1},
        {-1,-1,-1}
    },
    {
        {21416,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21417,-1},
        {-1},
        {-1,-1},
        {21417,-1,-1}
    },
    {
        {21417,-1},
        {-1},
        {-1,-1},
        {21417,-1,-1}
    },
    {
        {21417,-1},
        {-1},
        {-1,-1},
        {21417,24271,-1}
    },
    {
        {21417,-1},
        {-1},
        {-1,-1},
        {21417,-1,-1}
    },
    {
        {21417,-1},
        {-1},
        {-1,-1},
        {21417,24271,-1}
    },
    {
        {21417,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21417,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21417,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24274,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27711,-1},
        {-1,-1,-1}
    },
    {
        {21421,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21422,-1},
        {-1},
        {-1,-1},
        {24285,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26278,26310,-1}
    },
    {
        {21424,-1},
        {-1},
        {-1,-1},
        {24288,-1,-1}
    },
    {
        {21424,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21299,-1},
        {-1,-1,-1}
    },
    {
        {21426,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21426,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21427,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21428,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28304,-1,-1}
    },
    {
        {21430,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21431,-1},
        {-1},
        {-1,-1},
        {32945,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37168,38563,-1}
    },
    {
        {-1,-1},
        {-1},
        {36418,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21435,-1,-1}
    },
    {
        {21435,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21440,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21442,-1},
        {-1},
        {21442,21443},
        {21442,21443,-1}
    },
    {
        {21442,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21443,-1},
        {-1},
        {-1,-1},
        {21442,-1,-1}
    },
    {
        {21443,-1},
        {-1},
        {21442,21443},
        {21442,21443,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21442,21443,-1}
    },
    {
        {21448,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21448,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21449,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21449,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21450,-1},
        {-1},
        {-1,-1},
        {21450,-1,-1}
    },
    {
        {21450,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21451,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21452,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21453,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21454,-1},
        {-1},
        {-1,-1},
        {21454,-1,-1}
    },
    {
        {21454,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22828,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25300,25244},
        {29358,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20107,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21460,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21461,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21462,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21463,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21465,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21466,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21467,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21467,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21467,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21467,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21471,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23561,-1,-1}
    },
    {
        {21471,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21471,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21471,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30090,30067,-1}
    },
    {
        {21473,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21474,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21475,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21476,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21477,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21480,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21481,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21482,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21483,-1},
        {-1},
        {-1,-1},
        {21483,-1,-1}
    },
    {
        {21483,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21484,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21485,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21485,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21486,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21487,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21488,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21489,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21489,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21490,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21490,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21491,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21492,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21493,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21494,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21495,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21496,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21448,22022},
        {21448,-1,-1}
    },
    {
        {21498,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21480,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20993,20939},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21700,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21505,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21507,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21508,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21927,35359},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21599,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21599,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21505,22909},
        {-1,-1,-1}
    },
    {
        {21512,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21513,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21514,-1},
        {-1},
        {-1,-1},
        {24340,-1,-1}
    },
    {
        {21515,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21516,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21517,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21518,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21519,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21519,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21520,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21521,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21522,-1},
        {-1},
        {-1,-1},
        {21668,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22151,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21565,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21531,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21531,-1},
        {-1},
        {-1,-1},
        {21531,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25197,-1},
        {-1,-1,-1}
    },
    {
        {21533,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21533,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21534,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21535,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21536,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24515,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22972,-1},
        {-1,-1,-1}
    },
    {
        {21540,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21542,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21543,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21544,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21545,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21545,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21547,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21548,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21549,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21550,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23439,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21698,24350},
        {24350,-1,-1}
    },
    {
        {21577,-1},
        {-1},
        {-1,-1},
        {21577,-1,-1}
    },
    {
        {21577,-1},
        {-1},
        {-1,-1},
        {21577,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21558,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27770,21852},
        {-1,-1,-1}
    },
    {
        {21560,-1},
        {-1},
        {-1,-1},
        {21560,-1,-1}
    },
    {
        {21560,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21561,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25237,-1},
        {-1,-1,-1}
    },
    {
        {21563,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21564,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21565,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21566,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21578,-1},
        {-1},
        {21578,-1},
        {21578,-1,-1}
    },
    {
        {21568,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21568,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26112,-1},
        {-1,-1,-1}
    },
    {
        {21570,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21571,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21563,-1},
        {-1,-1,-1}
    },
    {
        {21574,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21576,-1},
        {-1},
        {-1,-1},
        {21576,-1,-1}
    },
    {
        {21576,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21576,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21577,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21578,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36282,-1},
        {36282,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21483,22125},
        {21483,-1,-1}
    },
    {
        {21581,-1},
        {-1},
        {-1,-1},
        {21565,-1,-1}
    },
    {
        {21582,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21558,-1},
        {-1},
        {21558,-1},
        {-1,-1,-1}
    },
    {
        {21585,-1},
        {-1},
        {-1,-1},
        {21534,-1,-1}
    },
    {
        {21589,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21589,-1},
        {-1},
        {-1,-1},
        {22036,21589,-1}
    },
    {
        {-1,-1},
        {-1},
        {30027,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21571,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40797,-1,-1}
    },
    {
        {21599,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21563,-1},
        {21563,-1,-1}
    },
    {
        {21602,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35584,-1},
        {35584,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20873,-1},
        {-1,-1,-1}
    },
    {
        {21606,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35398,-1,-1}
    },
    {
        {21608,-1},
        {-1},
        {-1,-1},
        {21608,-1,-1}
    },
    {
        {21608,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21980,-1,-1}
    },
    {
        {21610,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21611,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21612,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21628,-1},
        {-1,-1,-1}
    },
    {
        {21616,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21617,-1},
        {-1},
        {-1,-1},
        {21617,-1,-1}
    },
    {
        {21617,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21619,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21620,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21621,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21622,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21623,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21857,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20314,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21495,-1},
        {-1,-1,-1}
    },
    {
        {21627,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21628,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21629,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21632,-1},
        {-1,-1,-1}
    },
    {
        {21631,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21632,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21277,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21636,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21638,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21638,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21640,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21642,-1},
        {-1},
        {-1,-1},
        {21644,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21643,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21644,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21645,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21646,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21647,-1},
        {-1},
        {-1,-1},
        {35424,-1,-1}
    },
    {
        {21648,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21650,-1},
        {-1},
        {-1,-1},
        {21610,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30745,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21730,-1},
        {-1,-1,-1}
    },
    {
        {21653,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21654,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21660,-1},
        {-1},
        {-1,-1},
        {21668,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34566,21743},
        {21743,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34566,21743},
        {21743,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21664,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21665,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21666,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21666,-1},
        {-1},
        {-1,-1},
        {21666,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26180,-1},
        {-1,-1,-1}
    },
    {
        {21668,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21669,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21670,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21672,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21673,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21673,-1,-1}
    },
    {
        {21675,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21676,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21676,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21677,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21678,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21679,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35426,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21682,-1},
        {-1},
        {-1,-1},
        {21682,-1,-1}
    },
    {
        {21682,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21683,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24674,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35441,32850},
        {-1,-1,-1}
    },
    {
        {21687,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21688,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21690,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21692,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21693,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21694,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21695,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21696,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21697,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21698,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21699,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21700,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24671,-1},
        {-1,-1,-1}
    },
    {
        {21702,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21703,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21704,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21705,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21705,-1},
        {-1},
        {-1,-1},
        {21705,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22320,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38911,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24456,-1},
        {-1,-1,-1}
    },
    {
        {21720,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22130,21890},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21729,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21730,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21733,-1},
        {-1},
        {-1,-1},
        {27468,-1,-1}
    },
    {
        {21734,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22151,-1},
        {-1,-1,-1}
    },
    {
        {21736,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21736,-1},
        {-1},
        {-1,-1},
        {21736,-1,-1}
    },
    {
        {21737,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25417,-1},
        {-1,-1,-1}
    },
    {
        {21740,-1},
        {-1},
        {-1,-1},
        {35382,-1,-1}
    },
    {
        {21741,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21742,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21743,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29282,21565},
        {-1,-1,-1}
    },
    {
        {21745,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21746,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21746,-1},
        {-1},
        {-1,-1},
        {21746,-1,-1}
    },
    {
        {21747,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21928,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25420,15213},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22482,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34569,-1},
        {-1,-1,-1}
    },
    {
        {21754,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40766,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21757,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21757,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21759,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21760,-1},
        {-1},
        {-1,-1},
        {35480,-1,-1}
    },
    {
        {21761,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {64015,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22470,-1},
        {-1,-1,-1}
    },
    {
        {21764,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21766,-1},
        {-1},
        {-1,-1},
        {21766,-1,-1}
    },
    {
        {21766,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21767,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21769,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30820,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38500,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33694,-1,-1}
    },
    {
        {21774,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21775,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21776,-1},
        {-1},
        {-1,-1},
        {21776,-1,-1}
    },
    {
        {21776,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22124,-1},
        {22124,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21698,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30824,-1},
        {-1,-1,-1}
    },
    {
        {21780,-1},
        {-1},
        {-1,-1},
        {21566,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21782,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38497,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21843,-1},
        {21843,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29380,21886},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28024,-1},
        {-1,-1,-1}
    },
    {
        {21918,-1},
        {-1},
        {21918,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21803,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34382,21495},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21764,20854},
        {-1,-1,-1}
    },
    {
        {21806,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21807,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28078,-1},
        {-1,-1,-1}
    },
    {
        {21809,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21811,-1},
        {-1},
        {-1,-1},
        {21811,-1,-1}
    },
    {
        {21811,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21811,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34595,-1},
        {-1,-1,-1}
    },
    {
        {21813,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22537,-1},
        {-1,-1,-1}
    },
    {
        {21816,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21817,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21817,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26218,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21820,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21822,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21628,-1},
        {-1,-1,-1}
    },
    {
        {21824,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21825,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20083,21564},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40806,-1},
        {-1,-1,-1}
    },
    {
        {21828,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21828,-1},
        {-1},
        {-1,-1},
        {21828,-1,-1}
    },
    {
        {21828,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21828,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21828,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21829,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21830,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21830,-1},
        {-1},
        {-1,-1},
        {21830,-1,-1}
    },
    {
        {21831,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21834,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24425,37319},
        {-1,-1,-1}
    },
    {
        {21836,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21836,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29310,-1},
        {-1,-1,-1}
    },
    {
        {21839,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21840,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21820,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23835,-1},
        {-1,-1,-1}
    },
    {
        {21843,-1},
        {-1},
        {-1,-1},
        {21843,-1,-1}
    },
    {
        {21843,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21843,-1},
        {21843,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38518,-1},
        {-1,-1,-1}
    },
    {
        {21846,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21847,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30133,35390},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21852,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21853,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21853,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21854,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21843,-1},
        {21843,-1,-1}
    },
    {
        {21856,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21857,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21859,-1},
        {-1},
        {-1,-1},
        {37532,-1,-1}
    },
    {
        {21860,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33293,-1},
        {33293,33294,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24773,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24773,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27874,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21776,-1},
        {21776,-1,-1}
    },
    {
        {21883,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21884,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21884,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21885,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21886,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21888,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21889,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21890,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21891,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21892,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21892,-1},
        {-1},
        {-1,-1},
        {21892,-1,-1}
    },
    {
        {21892,-1},
        {-1},
        {-1,-1},
        {21892,-1,-1}
    },
    {
        {21892,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21892,-1},
        {-1},
        {-1,-1},
        {21892,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29020,-1},
        {-1,-1,-1}
    },
    {
        {21894,-1},
        {-1},
        {-1,-1},
        {21746,-1,-1}
    },
    {
        {21894,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21895,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21896,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21897,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21898,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21899,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21610,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26612,-1},
        {-1,-1,-1}
    },
    {
        {21902,-1},
        {-1},
        {-1,-1},
        {21692,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37057,-1},
        {-1,-1,-1}
    },
    {
        {21905,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21905,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24848,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21653,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21912,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21913,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21913,-1},
        {-1},
        {-1,-1},
        {21913,-1,-1}
    },
    {
        {21913,-1},
        {-1},
        {-1,-1},
        {21913,-1,-1}
    },
    {
        {21913,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21914,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21690,-1},
        {-1,-1,-1}
    },
    {
        {21916,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21916,-1},
        {-1},
        {-1,-1},
        {21916,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21917,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21918,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21918,-1},
        {-1},
        {-1,-1},
        {21918,-1,-1}
    },
    {
        {21919,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22087,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26256,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27459,21611},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21522,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23891,24022},
        {-1,-1,-1}
    },
    {
        {21927,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21928,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21928,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21929,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21929,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21930,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21931,-1},
        {-1},
        {-1,-1},
        {21931,-1,-1}
    },
    {
        {21931,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21931,-1},
        {-1},
        {-1,-1},
        {21931,-1,-1}
    },
    {
        {21931,-1},
        {-1},
        {-1,-1},
        {21931,-1,-1}
    },
    {
        {21932,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21933,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21934,-1},
        {-1},
        {-1,-1},
        {21336,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22132,-1},
        {22132,-1,-1}
    },
    {
        {21936,-1},
        {-1},
        {-1,-1},
        {21936,-1,-1}
    },
    {
        {21936,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31472,21737},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26619,-1},
        {26619,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21942,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21929,-1},
        {-1},
        {21929,-1},
        {21929,-1,-1}
    },
    {
        {21929,-1},
        {-1},
        {21929,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21951,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27580,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21956,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21957,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21957,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21959,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37013,-1},
        {-1,-1,-1}
    },
    {
        {21961,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21964,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22013,-1},
        {-1,-1,-1}
    },
    {
        {21966,-1},
        {-1},
        {-1,-1},
        {32629,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30606,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30606,-1},
        {-1,-1,-1}
    },
    {
        {21969,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21970,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21970,-1},
        {-1},
        {-1,-1},
        {21970,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21972,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28349,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23234,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30917,-1},
        {-1,-1,-1}
    },
    {
        {21978,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21980,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21981,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21981,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21983,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37226,-1},
        {37226,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32705,-1},
        {-1,-1,-1}
    },
    {
        {21986,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21986,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21987,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21987,-1,-1}
    },
    {
        {21988,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21988,-1},
        {-1},
        {-1,-1},
        {21988,-1,-1}
    },
    {
        {21988,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30347,22071},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21961,32034},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23833,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {64069,26214},
        {-1,-1,-1}
    },
    {
        {21993,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21885,25343},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22007,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22007,-1},
        {22007,-1,-1}
    },
    {
        {22009,-1},
        {-1},
        {-1,-1},
        {22009,-1,-1}
    },
    {
        {22009,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30368,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33180,-1},
        {-1,-1,-1}
    },
    {
        {22013,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22014,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35627,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27117,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21483,-1},
        {21483,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22665,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21483,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24936,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22022,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22023,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22024,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22025,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22025,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22025,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21824,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21742,-1},
        {21742,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23254,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22030,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22032,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33173,-1},
        {-1,-1,-1}
    },
    {
        {22036,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22036,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23267,-1},
        {-1,-1,-1}
    },
    {
        {22038,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22039,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22040,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24471,-1},
        {-1,-1,-1}
    },
    {
        {22043,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22043,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40613,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27122,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21202,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37117,-1},
        {-1,-1,-1}
    },
    {
        {22056,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22057,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22057,-1},
        {-1},
        {-1,-1},
        {35649,22057,-1}
    },
    {
        {-1,-1},
        {-1},
        {36023,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28982,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33192,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23959,-1},
        {-1,-1,-1}
    },
    {
        {22063,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22064,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22064,-1},
        {-1},
        {-1,-1},
        {22064,-1,-1}
    },
    {
        {22065,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22066,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22066,-1},
        {-1},
        {-1,-1},
        {22066,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22068,-1},
        {-1},
        {-1,-1},
        {35292,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22070,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22071,-1},
        {-1,-1,-1}
    },
    {
        {22071,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22072,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23308,-1},
        {-1,-1,-1}
    },
    {
        {22075,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30044,-1,-1}
    },
    {
        {22077,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30969,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30969,-1},
        {-1,-1,-1}
    },
    {
        {22079,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22080,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22082,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22082,-1},
        {-1},
        {-1,-1},
        {22082,-1,-1}
    },
    {
        {22082,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21931,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23296,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22087,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22087,-1},
        {-1},
        {-1,-1},
        {22087,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36441,-1},
        {-1,-1,-1}
    },
    {
        {22089,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35662,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22092,-1},
        {-1},
        {-1,-1},
        {22092,-1,-1}
    },
    {
        {22092,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22094,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22096,-1},
        {-1},
        {-1,-1},
        {22120,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30967,-1},
        {-1,-1,-1}
    },
    {
        {22099,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29234,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21535,-1},
        {21535,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24158,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25496,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25810,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22107,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22110,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22112,-1},
        {-1},
        {-1,-1},
        {22112,-1,-1}
    },
    {
        {22112,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22116,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28611,-1},
        {-1,-1,-1}
    },
    {
        {22118,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {145180,-1},
        {-1,-1,-1}
    },
    {
        {22120,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22122,-1},
        {-1},
        {-1,-1},
        {35679,-1,-1}
    },
    {
        {22123,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22123,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22124,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22125,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22127,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33219,-1},
        {-1,-1,-1}
    },
    {
        {22129,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22129,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22130,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34398,-1},
        {-1,-1,-1}
    },
    {
        {22132,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22132,-1},
        {-1},
        {-1,-1},
        {22132,-1,-1}
    },
    {
        {22132,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36947,-1,-1}
    },
    {
        {22134,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22134,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27462,-1},
        {-1,-1,-1}
    },
    {
        {22136,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22136,-1},
        {-1},
        {-1,-1},
        {22136,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24403,30070},
        {-1,-1,-1}
    },
    {
        {22138,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22189,-1},
        {22189,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21905,-1},
        {21905,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22144,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22144,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26332,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30435,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36957,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36957,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22150,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22150,-1},
        {-1},
        {-1,-1},
        {22150,-1,-1}
    },
    {
        {22151,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22152,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21636,-1},
        {21636,-1,-1}
    },
    {
        {22154,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22154,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22156,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22743,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35946,21495},
        {-1,-1,-1}
    },
    {
        {22159,-1},
        {-1},
        {-1,-1},
        {22164,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29982,22039},
        {22039,-1,-1}
    },
    {
        {22164,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22164,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22164,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22165,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26333,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25854,-1},
        {-1,-1,-1}
    },
    {
        {22169,-1},
        {-1},
        {-1,-1},
        {40809,-1,-1}
    },
    {
        {22170,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31019,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40664,40665},
        {-1,-1,-1}
    },
    {
        {22173,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22173,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22174,-1},
        {-1},
        {-1,-1},
        {21746,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22176,-1},
        {-1},
        {-1,-1},
        {28687,-1,-1}
    },
    {
        {22176,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22178,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22181,-1},
        {-1},
        {-1,-1},
        {21693,-1,-1}
    },
    {
        {22181,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22184,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22184,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22185,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21846,-1},
        {21847,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22188,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22189,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22190,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22190,-1},
        {-1},
        {-1,-1},
        {22190,-1,-1}
    },
    {
        {22190,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22190,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35604,-1},
        {35604,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22194,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22195,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22196,-1},
        {-1},
        {-1,-1},
        {21427,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22198,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22210,-1},
        {22210,-1,-1}
    },
    {
        {22204,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22204,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21852,-1},
        {21852,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33290,-1},
        {-1,-1,-1}
    },
    {
        {22208,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22208,-1},
        {-1},
        {-1,-1},
        {22208,-1,-1}
    },
    {
        {22209,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22209,-1},
        {-1},
        {-1,-1},
        {22209,-1,-1}
    },
    {
        {22209,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22210,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22211,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22213,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31018,34822},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26348,-1},
        {-1,-1,-1}
    },
    {
        {22216,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22216,-1},
        {-1},
        {-1,-1},
        {22216,-1,-1}
    },
    {
        {22217,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22218,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22221,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22222,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22222,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22222,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33393,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22225,-1},
        {-1},
        {-1,-1},
        {22065,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22227,-1},
        {-1},
        {-1,-1},
        {40807,-1,-1}
    },
    {
        {22227,-1},
        {-1},
        {-1,-1},
        {40807,-1,-1}
    },
    {
        {22227,-1},
        {-1},
        {-1,-1},
        {22227,40807,-1}
    },
    {
        {-1,-1},
        {-1},
        {22218,-1},
        {-1,-1,-1}
    },
    {
        {22231,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22232,-1},
        {-1},
        {-1,-1},
        {22238,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22240,-1,-1}
    },
    {
        {22234,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22235,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22235,-1},
        {-1},
        {-1,-1},
        {22235,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22238,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22239,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22240,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22296,22243},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22243,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28149,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20879,-1,-1}
    },
    {
        {22248,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26085,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22238,-1},
        {22238,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22254,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22254,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22269,-1},
        {-1},
        {22283,22269},
        {22269,-1,-1}
    },
    {
        {22256,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22258,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22259,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22262,-1},
        {-1},
        {-1,-1},
        {22283,22269,-1}
    },
    {
        {22262,-1},
        {-1},
        {-1,-1},
        {22269,-1,-1}
    },
    {
        {22263,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26085,23455},
        {26085,-1,-1}
    },
    {
        {22265,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22266,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22283,22269},
        {22283,22269,-1}
    },
    {
        {-1,-1},
        {-1},
        {32974,-1},
        {32974,-1,-1}
    },
    {
        {22269,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22271,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22272,-1},
        {-1},
        {-1,-1},
        {22269,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22275,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22276,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20989,20990},
        {20989,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22280,-1},
        {-1},
        {-1,-1},
        {22287,-1,-1}
    },
    {
        {22280,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22281,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22282,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22282,-1},
        {-1},
        {-1,-1},
        {22282,-1,-1}
    },
    {
        {22283,-1},
        {-1},
        {-1,-1},
        {22269,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22285,-1},
        {-1},
        {-1,-1},
        {22258,-1,-1}
    },
    {
        {22285,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22285,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22285,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22291,-1},
        {-1},
        {22291,20870},
        {22291,20870,-1}
    },
    {
        {22287,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22275,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22290,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22291,-1},
        {-1},
        {-1,-1},
        {20870,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38356,-1},
        {-1,-1,-1}
    },
    {
        {22293,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22294,-1},
        {-1},
        {-1,-1},
        {22259,-1,-1}
    },
    {
        {22294,-1},
        {-1},
        {-1,-1},
        {22259,22294,-1}
    },
    {
        {22294,-1},
        {-1},
        {22259,22294},
        {22259,22294,-1}
    },
    {
        {22294,-1},
        {-1},
        {22259,22294},
        {22259,22294,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22296,-1},
        {-1},
        {-1,-1},
        {22243,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22300,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27410,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27410,-1,-1}
    },
    {
        {22303,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22305,-1},
        {-1},
        {-1,-1},
        {22303,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22307,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22307,-1},
        {-1},
        {-1,-1},
        {22307,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22310,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22311,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22312,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22313,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22714,-1},
        {22714,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26439,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26439,-1,-1}
    },
    {
        {22317,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22319,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22320,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38433,20439},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22313,-1},
        {-1,-1,-1}
    },
    {
        {22323,-1},
        {-1},
        {-1,-1},
        {30013,30030,-1}
    },
    {
        {22324,-1},
        {-1},
        {-1,-1},
        {22343,-1,-1}
    },
    {
        {22324,-1},
        {-1},
        {-1,-1},
        {22343,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22336,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22327,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22328,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22331,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37324,30343},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22336,-1},
        {-1},
        {-1,-1},
        {38447,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22395,22564},
        {-1,-1,-1}
    },
    {
        {22338,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22734,-1},
        {22734,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25237,-1},
        {-1,-1,-1}
    },
    {
        {22341,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22342,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22343,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22343,-1},
        {-1},
        {-1,-1},
        {22343,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22353,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22353,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27788,-1},
        {-1,-1,-1}
    },
    {
        {22346,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22348,-1},
        {22348,-1,-1}
    },
    {
        {22348,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22349,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22350,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22351,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22352,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22353,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22411,-1},
        {22411,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22320,-1},
        {22320,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32147,32076},
        {32147,32076,-1}
    },
    {
        {-1,-1},
        {-1},
        {24524,-1},
        {24524,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20608,22602},
        {22602,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22395,-1},
        {22395,-1,-1}
    },
    {
        {24032,-1},
        {-1},
        {24032,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33252,22533},
        {22533,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22369,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22530,-1},
        {22530,-1,-1}
    },
    {
        {22372,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22374,-1,-1}
    },
    {
        {22374,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22376,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22377,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22378,-1},
        {-1},
        {-1,-1},
        {22378,-1,-1}
    },
    {
        {22378,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23396,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21488,33274,-1}
    },
    {
        {22383,-1},
        {-1},
        {-1,-1},
        {22351,-1,-1}
    },
    {
        {22384,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25282,-1},
        {-1,-1,-1}
    },
    {
        {22387,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22388,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22389,-1},
        {-1},
        {-1,-1},
        {19992,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25287,-1},
        {-1,-1,-1}
    },
    {
        {22391,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22434,-1},
        {22434,22343,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22395,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22396,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22399,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22378,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22467,-1,-1}
    },
    {
        {22402,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22408,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22409,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22409,-1},
        {-1},
        {-1,-1},
        {22409,-1,-1}
    },
    {
        {22411,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22411,-1},
        {-1},
        {-1,-1},
        {22411,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27918,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26684,25356},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22578,-1},
        {22578,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22744,22593},
        {-1,-1,-1}
    },
    {
        {22419,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22419,-1},
        {-1},
        {-1,-1},
        {22419,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21402,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22534,-1},
        {22534,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22426,-1},
        {-1},
        {-1,-1},
        {21916,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22428,-1},
        {-1,-1,-1}
    },
    {
        {22428,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22429,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22429,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22430,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22432,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22434,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22435,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22436,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25342,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22432,22718},
        {22432,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22384,-1,-1}
    },
    {
        {22440,-1},
        {-1},
        {-1,-1},
        {23432,22483,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32854,-1,-1}
    },
    {
        {22442,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22444,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22448,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22451,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26771,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29694,-1},
        {-1,-1,-1}
    },
    {
        {22456,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37083,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24441,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23736,-1},
        {23736,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24207,-1},
        {-1,-1,-1}
    },
    {
        {22464,-1},
        {-1},
        {-1,-1},
        {22402,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23697,-1},
        {23697,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22467,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38450,22346},
        {38450,-1,-1}
    },
    {
        {22470,-1},
        {-1},
        {-1,-1},
        {30830,-1,-1}
    },
    {
        {22471,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22472,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22472,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23777,23805},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22320,-1},
        {22320,-1,-1}
    },
    {
        {22475,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22476,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25424,-1},
        {-1,-1,-1}
    },
    {
        {22478,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22478,-1},
        {-1},
        {-1,-1},
        {22478,-1,-1}
    },
    {
        {22479,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26795,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21746,-1},
        {21746,-1,-1}
    },
    {
        {22482,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22482,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22483,-1},
        {-1},
        {-1,-1},
        {22482,-1,-1}
    },
    {
        {22483,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22484,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22486,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22487,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22492,-1},
        {-1},
        {-1,-1},
        {37326,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22494,-1},
        {-1},
        {-1,-1},
        {22564,-1,-1}
    },
    {
        {22495,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22496,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25458,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22499,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22500,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22502,-1},
        {-1},
        {-1,-1},
        {26880,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25476,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25511,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25536,-1},
        {-1,-1,-1}
    },
    {
        {22509,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22748,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25513,-1},
        {-1,-1,-1}
    },
    {
        {22512,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20470,36879},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25447,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38501,22350},
        {22350,-1,-1}
    },
    {
        {22516,-1},
        {-1},
        {-1,-1},
        {22516,-1,-1}
    },
    {
        {22516,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22517,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22518,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22518,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22519,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22520,-1},
        {-1},
        {-1,-1},
        {22580,-1,-1}
    },
    {
        {22520,-1},
        {-1},
        {-1,-1},
        {22520,22580,-1}
    },
    {
        {22521,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22522,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22523,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22524,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22525,-1},
        {-1},
        {-1,-1},
        {25475,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22527,-1},
        {-1},
        {-1,-1},
        {27877,-1,-1}
    },
    {
        {22528,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22530,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22372,-1,-1}
    },
    {
        {22532,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22533,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22534,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35641,20677},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22537,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22538,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22539,-1},
        {-1},
        {-1,-1},
        {22539,-1,-1}
    },
    {
        {22539,-1},
        {-1},
        {-1,-1},
        {22539,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22549,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31361,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22553,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22553,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25582,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25536,-1},
        {-1,-1,-1}
    },
    {
        {22557,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22558,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22560,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22561,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22561,-1},
        {22561,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38533,-1},
        {-1,-1,-1}
    },
    {
        {22564,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29279,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38542,-1,-1}
    },
    {
        {22567,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22738,36943},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24646,-1},
        {-1,-1,-1}
    },
    {
        {22570,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26934,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25540,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37122,-1},
        {-1,-1,-1}
    },
    {
        {22575,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22576,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22576,-1},
        {-1},
        {-1,-1},
        {22576,-1,-1}
    },
    {
        {22577,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22577,-1},
        {-1},
        {-1,-1},
        {22577,-1,-1}
    },
    {
        {22578,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22578,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23883,-1},
        {-1,-1,-1}
    },
    {
        {22580,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22580,-1},
        {-1},
        {-1,-1},
        {22580,-1,-1}
    },
    {
        {22581,-1},
        {-1},
        {-1,-1},
        {22581,-1,-1}
    },
    {
        {22581,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22521,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22521,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39087,-1},
        {-1,-1,-1}
    },
    {
        {22585,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22586,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22640,27941},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29326,-1},
        {-1,-1,-1}
    },
    {
        {22589,-1},
        {-1},
        {-1,-1},
        {23713,-1,-1}
    },
    {
        {22589,-1},
        {-1},
        {-1,-1},
        {23713,22589,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22592,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22593,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22601,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22602,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22603,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22604,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22604,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29794,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22609,-1,-1}
    },
    {
        {22609,-1},
        {-1},
        {-1,-1},
        {22609,-1,-1}
    },
    {
        {22609,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22610,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27040,-1},
        {-1,-1,-1}
    },
    {
        {22612,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22612,-1},
        {-1},
        {-1,-1},
        {22612,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22615,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22616,-1},
        {-1},
        {-1,-1},
        {22616,-1,-1}
    },
    {
        {22616,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22617,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22617,-1},
        {-1},
        {-1,-1},
        {22617,-1,-1}
    },
    {
        {22617,-1},
        {-1},
        {-1,-1},
        {22617,-1,-1}
    },
    {
        {22618,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22618,-1},
        {-1},
        {-1,-1},
        {22618,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22618,-1},
        {22618,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22622,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33900,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33900,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22534,-1},
        {-1,-1,-1}
    },
    {
        {22625,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22626,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22628,-1},
        {-1},
        {-1,-1},
        {22734,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38499,-1},
        {38499,-1,-1}
    },
    {
        {22631,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22633,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22635,-1},
        {-1},
        {-1,-1},
        {22625,-1,-1}
    },
    {
        {22640,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22640,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22642,-1},
        {-1},
        {-1,-1},
        {22580,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22539,-1},
        {22539,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22645,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25693,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40565,-1},
        {40565,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22649,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27511,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25720,-1},
        {-1,-1,-1}
    },
    {
        {22652,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29245,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22654,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27161,-1},
        {-1,-1,-1}
    },
    {
        {22659,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22659,-1},
        {-1},
        {-1,-1},
        {22659,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27117,-1},
        {-1,-1,-1}
    },
    {
        {22661,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28382,28399},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22350,-1,-1}
    },
    {
        {22665,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22666,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22667,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22669,-1},
        {-1},
        {-1,-1},
        {22669,-1,-1}
    },
    {
        {22669,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22669,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37101,-1},
        {37101,-1,-1}
    },
    {
        {22671,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22672,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30340,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25688,-1},
        {-1,-1,-1}
    },
    {
        {22675,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22675,-1},
        {-1},
        {-1,-1},
        {22675,-1,-1}
    },
    {
        {22676,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22576,-1},
        {22576,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22612,-1,-1}
    },
    {
        {22679,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22684,-1},
        {-1},
        {-1,-1},
        {22684,-1,-1}
    },
    {
        {22684,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22684,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22685,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22686,-1},
        {-1},
        {-1,-1},
        {22679,-1,-1}
    },
    {
        {22679,-1},
        {-1},
        {-1,-1},
        {22686,22679,-1}
    },
    {
        {22687,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22687,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22602,38564},
        {22602,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25758,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22617,27211},
        {22617,-1,-1}
    },
    {
        {22696,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22697,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22698,-1},
        {-1},
        {-1,-1},
        {22697,-1,-1}
    },
    {
        {22699,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22699,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22699,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22320,-1},
        {22320,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30427,-1},
        {-1,-1,-1}
    },
    {
        {22702,-1},
        {-1},
        {-1,-1},
        {22549,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22549,22702,-1}
    },
    {
        {-1,-1},
        {-1},
        {22549,22702},
        {22549,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22748,-1},
        {22748,-1,-1}
    },
    {
        {22705,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22707,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22748,-1},
        {22748,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27194,-1},
        {-1,-1,-1}
    },
    {
        {22712,-1},
        {-1},
        {-1,-1},
        {36487,-1,-1}
    },
    {
        {22712,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22713,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22714,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22714,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22715,-1},
        {-1},
        {-1,-1},
        {29254,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22617,30973},
        {22685,-1,-1}
    },
    {
        {22718,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39515,39365},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22500,-1},
        {22500,-1,-1}
    },
    {
        {22721,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22723,-1},
        {-1},
        {-1,-1},
        {30086,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37326,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37326,-1,-1}
    },
    {
        {22725,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29882,-1},
        {-1,-1,-1}
    },
    {
        {22727,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22727,-1},
        {-1},
        {-1,-1},
        {22727,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25818,-1},
        {-1,-1,-1}
    },
    {
        {22730,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22732,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22733,-1},
        {-1},
        {-1,-1},
        {22649,-1,-1}
    },
    {
        {22734,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22734,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27323,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29885,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22737,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22738,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22738,-1},
        {-1},
        {-1,-1},
        {22738,-1,-1}
    },
    {
        {22739,-1},
        {-1},
        {-1,-1},
        {22311,-1,-1}
    },
    {
        {22739,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22740,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22741,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22741,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22743,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22744,-1},
        {-1},
        {-1,-1},
        {22593,-1,-1}
    },
    {
        {22745,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22745,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22746,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38331,-1},
        {38331,-1,-1}
    },
    {
        {22748,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22750,-1},
        {-1},
        {-1,-1},
        {22730,-1,-1}
    },
    {
        {22751,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22751,-1},
        {-1},
        {-1,-1},
        {22751,-1,-1}
    },
    {
        {22752,-1},
        {-1},
        {-1,-1},
        {22751,-1,-1}
    },
    {
        {22752,-1},
        {-1},
        {-1,-1},
        {22751,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21473,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22754,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32879,-1},
        {-1,-1,-1}
    },
    {
        {22756,-1},
        {-1},
        {-1,-1},
        {22732,-1,-1}
    },
    {
        {22757,-1},
        {-1},
        {-1,-1},
        {24283,-1,-1}
    },
    {
        {22757,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22734,-1},
        {22734,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24022,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22744,22593},
        {38647,-1,-1}
    },
    {
        {22761,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22763,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22764,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22766,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22767,-1},
        {-1},
        {-1,-1},
        {22766,-1,-1}
    },
    {
        {22768,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22769,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22770,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22770,-1},
        {-1},
        {-1,-1},
        {22770,-1,-1}
    },
    {
        {22771,-1},
        {-1},
        {-1,-1},
        {27579,27580,-1}
    },
    {
        {22772,-1},
        {-1},
        {-1,-1},
        {22772,-1,-1}
    },
    {
        {22772,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22766,22767},
        {21916,22766,22767}
    },
    {
        {22775,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22777,-1},
        {-1},
        {-1,-1},
        {22769,-1,-1}
    },
    {
        {22778,-1},
        {-1},
        {-1,-1},
        {22775,-1,-1}
    },
    {
        {22779,-1},
        {-1},
        {-1,-1},
        {23167,-1,-1}
    },
    {
        {22780,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22781,-1},
        {-1},
        {-1,-1},
        {23551,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36317,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22699,-1},
        {-1,-1,-1}
    },
    {
        {22781,-1},
        {-1},
        {22781,23551},
        {22781,23551,-1}
    },
    {
        {-1,-1},
        {-1},
        {22769,22777},
        {22777,22769,-1}
    },
    {
        {22786,-1},
        {-1},
        {-1,-1},
        {32066,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30408,36111},
        {-1,-1,-1}
    },
    {
        {22789,-1},
        {-1},
        {-1,-1},
        {38477,-1,-1}
    },
    {
        {22789,-1},
        {-1},
        {-1,-1},
        {38477,22789,-1}
    },
    {
        {22790,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40779,25998},
        {31874,-1,-1}
    },
    {
        {22793,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22794,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22794,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22795,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22796,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22797,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22799,-1},
        {-1},
        {-1,-1},
        {22799,-1,-1}
    },
    {
        {22799,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22800,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22799,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22804,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22804,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22804,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22805,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22806,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22806,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22808,40219},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22808,40219},
        {-1,-1,-1}
    },
    {
        {22808,-1},
        {-1},
        {-1,-1},
        {21359,-1,-1}
    },
    {
        {22809,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22810,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22811,-1},
        {-1},
        {-1,-1},
        {22810,-1,-1}
    },
    {
        {22811,-1},
        {-1},
        {-1,-1},
        {22810,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22812,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24384,21246},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24384,21246},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22818,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22818,-1},
        {-1},
        {-1,-1},
        {22818,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22818,-1,-1}
    },
    {
        {22820,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22821,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23584,-1},
        {-1,-1,-1}
    },
    {
        {22823,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22825,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22826,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22827,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22828,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22829,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22830,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22833,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22834,-1},
        {-1},
        {-1,-1},
        {26412,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27888,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27604,-1,-1}
    },
    {
        {22839,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22840,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22840,-1},
        {-1},
        {-1,-1},
        {22840,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20448,22846},
        {22846,-1,-1}
    },
    {
        {22845,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22846,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23574,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22852,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31022,-1},
        {-1,-1,-1}
    },
    {
        {22854,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22854,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22855,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22856,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22857,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36557,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22894,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28857,40670},
        {-1,-1,-1}
    },
    {
        {22862,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22863,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22864,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22865,-1},
        {-1},
        {-1,-1},
        {22865,-1,-1}
    },
    {
        {22865,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22865,-1},
        {-1},
        {-1,-1},
        {22865,-1,-1}
    },
    {
        {22865,-1},
        {-1},
        {-1,-1},
        {22865,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22868,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22868,-1},
        {-1},
        {-1,-1},
        {22868,-1,-1}
    },
    {
        {22869,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22871,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22872,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22872,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22874,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22875,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26000,-1},
        {-1,-1,-1}
    },
    {
        {22877,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22879,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22879,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22880,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22880,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22880,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22880,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20658,-1,-1}
    },
    {
        {22882,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22882,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38754,-1},
        {-1,-1,-1}
    },
    {
        {22885,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22885,-1},
        {-1},
        {-1,-1},
        {22885,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21209,-1},
        {-1,-1,-1}
    },
    {
        {22887,-1},
        {-1},
        {-1,-1},
        {22885,-1,-1}
    },
    {
        {22888,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22889,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22889,-1},
        {-1},
        {-1,-1},
        {22889,-1,-1}
    },
    {
        {22890,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22892,-1},
        {-1},
        {-1,-1},
        {22888,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22893,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22894,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27344,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23372,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22899,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22900,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22901,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22902,-1},
        {-1},
        {-1,-1},
        {23341,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22904,-1},
        {-1,-1,-1}
    },
    {
        {22904,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29590,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22908,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22909,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22328,22914},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26093,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25913,-1},
        {-1,-1,-1}
    },
    {
        {22913,-1},
        {-1},
        {-1,-1},
        {22913,-1,-1}
    },
    {
        {22913,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22914,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22915,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22915,-1},
        {-1},
        {-1,-1},
        {22915,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22916,-1},
        {-1},
        {-1,-1},
        {22916,-1,-1}
    },
    {
        {22916,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22904,34425},
        {-1,-1,-1}
    },
    {
        {22941,-1},
        {-1},
        {22941,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22570,-1},
        {-1,-1,-1}
    },
    {
        {22922,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22923,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22925,-1},
        {-1},
        {-1,-1},
        {23032,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23011,29599},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23030,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {136966,-1},
        {-1,-1,-1}
    },
    {
        {22930,-1},
        {-1},
        {-1,-1},
        {22956,-1,-1}
    },
    {
        {22931,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21549,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22934,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {18613,37142},
        {18613,37142,-1}
    },
    {
        {22937,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32986,37043},
        {-1,-1,-1}
    },
    {
        {22939,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22939,-1},
        {-1},
        {-1,-1},
        {136884,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22923,29606},
        {-1,-1,-1}
    },
    {
        {22941,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22962,-1},
        {-1,-1,-1}
    },
    {
        {22943,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21558,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21558,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26027,22925},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22984,21545},
        {-1,-1,-1}
    },
    {
        {22947,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22948,-1},
        {-1},
        {-1,-1},
        {20227,-1,-1}
    },
    {
        {22949,-1},
        {-1},
        {-1,-1},
        {22949,-1,-1}
    },
    {
        {22949,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22904,34444},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22952,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22956,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22958,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22987,26157},
        {-1,-1,-1}
    },
    {
        {22962,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23341,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {142008,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22995,28855},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21599,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23104,-1},
        {23104,-1,-1}
    },
    {
        {22969,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22971,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22972,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21627,-1},
        {-1,-1,-1}
    },
    {
        {22974,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21487,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21642,22934},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30765,-1},
        {-1,-1,-1}
    },
    {
        {22979,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22982,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24612,-1},
        {-1,-1,-1}
    },
    {
        {22984,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22985,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22985,-1,-1}
    },
    {
        {22987,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32835,23230},
        {-1,-1,-1}
    },
    {
        {22989,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22989,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22992,-1},
        {-1,-1,-1}
    },
    {
        {22992,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22993,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22994,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22995,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22996,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21616,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23019,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23338,23631},
        {23338,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23338,23631},
        {-1,-1,-1}
    },
    {
        {23001,-1},
        {-1},
        {-1,-1},
        {22922,-1,-1}
    },
    {
        {23002,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23002,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21699,-1},
        {-1,-1,-1}
    },
    {
        {23004,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23005,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23006,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22993,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23071,-1},
        {23071,-1,-1}
    },
    {
        {23011,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23011,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23012,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23013,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23014,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23015,-1},
        {-1},
        {-1,-1},
        {23014,-1,-1}
    },
    {
        {23016,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22904,-1},
        {-1,-1,-1}
    },
    {
        {23018,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23019,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23019,-1},
        {-1},
        {-1,-1},
        {23019,-1,-1}
    },
    {
        {23019,-1},
        {-1},
        {-1,-1},
        {23019,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33029,-1},
        {-1,-1,-1}
    },
    {
        {23022,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37528,22956},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29678,-1},
        {-1,-1,-1}
    },
    {
        {23030,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23032,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35435,22908},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23035,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22931,23198},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22429,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23039,-1},
        {-1},
        {-1,-1},
        {23039,-1,-1}
    },
    {
        {23039,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23041,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23041,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21700,-1},
        {-1,-1,-1}
    },
    {
        {23043,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23105,-1,-1}
    },
    {
        {23049,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21743,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21736,-1},
        {-1,-1,-1}
    },
    {
        {23052,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23053,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27533,-1},
        {-1,-1,-1}
    },
    {
        {23057,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22982,-1,-1}
    },
    {
        {23059,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20811,23079},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31462,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23063,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23064,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38488,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23066,-1},
        {-1},
        {-1,-1},
        {23248,23346,-1}
    },
    {
        {23087,-1},
        {-1},
        {-1,-1},
        {23087,-1,-1}
    },
    {
        {23087,-1},
        {-1},
        {23087,-1},
        {23087,-1,-1}
    },
    {
        {23068,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23068,-1},
        {-1},
        {-1,-1},
        {23068,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26734,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23071,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23072,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23006,24535},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23075,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22941,-1,-1}
    },
    {
        {23077,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23338,-1},
        {23338,-1,-1}
    },
    {
        {23079,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23081,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23081,-1},
        {-1},
        {-1,-1},
        {23081,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23063,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23285,-1},
        {23285,-1,-1}
    },
    {
        {23085,-1},
        {-1},
        {-1,-1},
        {23305,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21761,-1},
        {-1,-1,-1}
    },
    {
        {23087,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22994,-1,-1}
    },
    {
        {23093,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23094,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23094,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29738,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24358,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21852,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {13734,-1},
        {-1,-1,-1}
    },
    {
        {23100,-1},
        {-1},
        {-1,-1},
        {20513,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {13736,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21824,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23104,-1},
        {23104,-1,-1}
    },
    {
        {23104,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23105,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26825,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29742,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23110,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23113,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20469,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28113,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28113,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30869,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21807,-1},
        {-1,-1,-1}
    },
    {
        {23081,-1},
        {-1},
        {23081,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21846,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26000,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21857,-1},
        {-1,-1,-1}
    },
    {
        {23125,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20816,20818},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24059,21367},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23143,-1},
        {23143,-1,-1}
    },
    {
        {23130,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31466,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30855,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22266,23035},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29743,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23138,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23035,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21825,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23141,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23142,-1},
        {-1},
        {-1,-1},
        {23142,-1,-1}
    },
    {
        {23142,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23143,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23143,-1},
        {-1},
        {-1,-1},
        {23143,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23832,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23736,-1},
        {-1,-1,-1}
    },
    {
        {23146,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29736,-1},
        {-1,-1,-1}
    },
    {
        {23148,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23148,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23149,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {136004,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40599,-1},
        {40599,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26218,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24358,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {134469,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23159,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26264,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23162,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26247,-1},
        {-1,-1,-1}
    },
    {
        {23166,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23166,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23167,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21889,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26256,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21884,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29489,-1},
        {-1,-1,-1}
    },
    {
        {23172,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22570,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23280,23273},
        {23273,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35242,-1},
        {35242,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25539,-1},
        {-1,-1,-1}
    },
    {
        {23179,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23142,-1},
        {23142,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22992,-1},
        {22992,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31471,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22585,-1},
        {-1,-1,-1}
    },
    {
        {23186,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24848,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21885,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21927,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21896,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23194,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23195,-1},
        {-1},
        {-1,-1},
        {23195,-1,-1}
    },
    {
        {23195,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23196,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24833,-1},
        {-1,-1,-1}
    },
    {
        {23198,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23199,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23200,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29779,-1},
        {-1,-1,-1}
    },
    {
        {23202,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23202,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26579,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21919,-1},
        {-1,-1,-1}
    },
    {
        {23207,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29494,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {13765,-1},
        {-1,-1,-1}
    },
    {
        {23228,-1},
        {-1},
        {23228,-1},
        {23228,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23660,23646},
        {-1,-1,-1}
    },
    {
        {23217,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23219,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34945,-1},
        {-1,-1,-1}
    },
    {
        {23221,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23221,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33592,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {134805,-1},
        {-1,-1,-1}
    },
    {
        {23226,-1},
        {-1},
        {-1,-1},
        {23172,-1,-1}
    },
    {
        {23226,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23226,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23110,-1,-1}
    },
    {
        {23228,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23229,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23230,-1},
        {-1},
        {-1,-1},
        {23230,-1,-1}
    },
    {
        {23230,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23231,-1},
        {-1},
        {-1,-1},
        {24871,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23233,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23233,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23234,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23234,-1},
        {-1},
        {-1,-1},
        {23234,-1,-1}
    },
    {
        {23234,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23196,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23236,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21983,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26269,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23241,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21961,-1},
        {-1,-1,-1}
    },
    {
        {23243,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23243,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23244,-1},
        {-1},
        {-1,-1},
        {23244,-1,-1}
    },
    {
        {23244,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29803,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30917,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23248,-1},
        {-1},
        {-1,-1},
        {23346,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23254,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23255,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23255,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34746,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29831,-1},
        {-1,-1,-1}
    },
    {
        {23258,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23258,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23260,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23260,-1},
        {-1},
        {-1,-1},
        {23260,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24951,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22665,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26289,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26289,-1},
        {-1,-1,-1}
    },
    {
        {23264,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23265,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23265,-1},
        {-1},
        {-1,-1},
        {23265,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23344,35215},
        {-1,-1,-1}
    },
    {
        {23267,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23269,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23270,-1},
        {-1},
        {-1,-1},
        {23022,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22038,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23273,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33180,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23278,-1},
        {-1,-1,-1}
    },
    {
        {23278,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20658,-1},
        {-1,-1,-1}
    },
    {
        {23280,-1},
        {-1},
        {-1,-1},
        {23273,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20272,-1},
        {-1,-1,-1}
    },
    {
        {23285,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24980,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23200,-1},
        {23200,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30970,-1},
        {-1,-1,-1}
    },
    {
        {23290,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23291,-1},
        {-1},
        {-1,-1},
        {23290,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22079,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30967,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30059,-1},
        {-1,-1,-1}
    },
    {
        {23296,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29118,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22094,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23304,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23305,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29141,-1},
        {-1,-1,-1}
    },
    {
        {23307,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23308,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32654,-1},
        {23226,23172,-1}
    },
    {
        {-1,-1},
        {-1},
        {23081,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22110,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22123,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34184,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22125,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23081,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23318,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23319,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23321,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33217,23244},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27462,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23243,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30994,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22122,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23330,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23333,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22743,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29877,-1},
        {-1,-1,-1}
    },
    {
        {23338,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23338,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23992,-1},
        {-1,-1,-1}
    },
    {
        {23340,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23341,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21421,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25825,-1},
        {-1,-1,-1}
    },
    {
        {23344,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22733,-1},
        {-1,-1,-1}
    },
    {
        {23346,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23348,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23348,-1},
        {-1},
        {-1,-1},
        {23348,-1,-1}
    },
    {
        {23348,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23350,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23350,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33288,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24925,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28678,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22176,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23358,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23358,-1},
        {-1},
        {-1,-1},
        {23358,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33241,-1},
        {-1,-1,-1}
    },
    {
        {23360,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23361,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31755,-1},
        {-1,-1,-1}
    },
    {
        {23363,-1},
        {-1},
        {-1,-1},
        {23330,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23358,25078},
        {23358,-1,-1}
    },
    {
        {23365,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23344,-1},
        {23344,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {139643,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25884,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27489,27475},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38753,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23372,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29165,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23358,25078},
        {23358,-1,-1}
    },
    {
        {23376,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23377,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23378,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23380,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23381,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23382,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23383,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23384,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23386,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23386,-1},
        {-1},
        {-1,-1},
        {23386,-1,-1}
    },
    {
        {23387,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23388,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23389,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23391,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21987,-1},
        {21987,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32974,-1},
        {32974,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25265,-1},
        {-1,-1,-1}
    },
    {
        {23395,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23396,-1},
        {-1},
        {-1,-1},
        {23396,-1,-1}
    },
    {
        {23396,-1},
        {-1},
        {-1,-1},
        {23396,-1,-1}
    },
    {
        {23396,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23397,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23398,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23400,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23401,-1},
        {-1},
        {-1,-1},
        {21683,-1,-1}
    },
    {
        {23401,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23403,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27498,22909},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23407,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23408,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23409,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26895,-1},
        {-1,-1,-1}
    },
    {
        {23411,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23411,-1},
        {-1},
        {-1,-1},
        {23411,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23400,-1},
        {-1,-1,-1}
    },
    {
        {23413,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23413,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23411,-1},
        {-1},
        {23411,-1},
        {-1,-1,-1}
    },
    {
        {23411,-1},
        {-1},
        {23411,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29339,-1},
        {-1,-1,-1}
    },
    {
        {23416,-1},
        {-1},
        {-1,-1},
        {23398,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25816,-1},
        {25816,-1,-1}
    },
    {
        {23418,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22743,-1},
        {-1,-1,-1}
    },
    {
        {23420,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23420,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23421,-1},
        {-1},
        {-1,-1},
        {23420,-1,-1}
    },
    {
        {23421,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23423,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23424,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23425,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20887,-1},
        {20887,-1,-1}
    },
    {
        {23427,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23428,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23429,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31353,-1,-1}
    },
    {
        {23431,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23432,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23433,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31361,64085,-1}
    },
    {
        {23435,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23436,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23437,-1},
        {-1},
        {-1,-1},
        {32905,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23439,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23452,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23452,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23429,-1},
        {-1,-1,-1}
    },
    {
        {23443,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23445,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23446,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23447,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23448,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23449,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23450,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23451,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23452,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23453,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23526,23455},
        {23526,23455,-1}
    },
    {
        {23455,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23458,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23459,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23459,-1},
        {-1},
        {-1,-1},
        {23459,-1,-1}
    },
    {
        {23460,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23461,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23462,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23462,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25010,-1},
        {-1,-1,-1}
    },
    {
        {23470,-1},
        {-1},
        {20849,23470},
        {23470,-1,-1}
    },
    {
        {23468,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23470,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23472,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23526,23455,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23475,-1},
        {-1},
        {-1,-1},
        {23475,-1,-1}
    },
    {
        {23475,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23475,-1},
        {-1},
        {-1,-1},
        {23475,-1,-1}
    },
    {
        {23476,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23477,-1},
        {-1},
        {-1,-1},
        {23477,-1,-1}
    },
    {
        {23477,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23478,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23478,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23529,-1,-1}
    },
    {
        {23480,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23481,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23494,-1},
        {23494,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23495,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23487,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23488,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23488,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23490,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23491,-1},
        {-1},
        {-1,-1},
        {20900,-1,-1}
    },
    {
        {23491,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23491,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23492,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23493,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23494,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23495,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23495,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38738,-1},
        {38738,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38738,-1},
        {38738,-1,-1}
    },
    {
        {23497,-1},
        {-1},
        {-1,-1},
        {40372,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35998,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23500,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23527,-1,-1}
    },
    {
        {23502,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23504,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23504,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23522,23517},
        {-1,-1,-1}
    },
    {
        {23506,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23506,-1},
        {-1},
        {-1,-1},
        {23506,-1,-1}
    },
    {
        {23507,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23508,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23510,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23527,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23512,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23453,23542,-1}
    },
    {
        {23515,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23515,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23527,-1},
        {-1},
        {36015,23527},
        {23527,-1,-1}
    },
    {
        {23517,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23518,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23518,-1},
        {-1},
        {-1,-1},
        {23518,-1,-1}
    },
    {
        {23519,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23521,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23522,-1},
        {-1},
        {-1,-1},
        {23517,-1,-1}
    },
    {
        {23522,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23524,-1},
        {-1,-1,-1}
    },
    {
        {23524,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23524,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23525,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23525,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23526,-1},
        {-1},
        {-1,-1},
        {23455,-1,-1}
    },
    {
        {23527,-1},
        {-1},
        {-1,-1},
        {23527,-1,-1}
    },
    {
        {23527,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23528,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23529,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23531,-1},
        {-1},
        {-1,-1},
        {20889,-1,-1}
    },
    {
        {23532,-1},
        {-1},
        {-1,-1},
        {23515,-1,-1}
    },
    {
        {23515,-1},
        {-1},
        {-1,-1},
        {23515,-1,-1}
    },
    {
        {23515,23532},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34137,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23534,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23536,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23537,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23991,-1},
        {-1,-1,-1}
    },
    {
        {23539,-1},
        {-1},
        {-1,-1},
        {23453,23542,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23541,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23541,-1},
        {-1},
        {-1,-1},
        {23541,-1,-1}
    },
    {
        {23541,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23542,-1},
        {-1},
        {-1,-1},
        {23453,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23544,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23565,23550},
        {23565,23550,-1}
    },
    {
        {23546,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23549,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23549,-1},
        {-1},
        {-1,-1},
        {23549,-1,-1}
    },
    {
        {23550,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23551,-1},
        {-1},
        {-1,-1},
        {22781,23551,-1}
    },
    {
        {23551,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21493,-1},
        {21493,-1,-1}
    },
    {
        {23553,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23554,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23555,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23556,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23556,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23557,-1},
        {-1},
        {-1,-1},
        {21067,-1,-1}
    },
    {
        {23558,-1},
        {-1},
        {-1,-1},
        {23558,23559,-1}
    },
    {
        {23558,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23558,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23559,-1},
        {-1},
        {-1,-1},
        {23558,-1,-1}
    },
    {
        {23560,-1},
        {-1},
        {-1,-1},
        {23554,-1,-1}
    },
    {
        {23560,-1},
        {-1},
        {-1,-1},
        {23560,23554,-1}
    },
    {
        {23561,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23562,-1},
        {-1},
        {-1,-1},
        {23562,-1,-1}
    },
    {
        {23562,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23562,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23563,-1},
        {-1},
        {-1,-1},
        {23563,-1,-1}
    },
    {
        {23563,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23565,-1},
        {-1},
        {-1,-1},
        {23550,-1,-1}
    },
    {
        {23566,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23566,-1},
        {-1},
        {-1,-1},
        {23566,-1,-1}
    },
    {
        {23567,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23574,-1},
        {-1,-1,-1}
    },
    {
        {23569,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23570,-1},
        {-1},
        {-1,-1},
        {29246,-1,-1}
    },
    {
        {23570,-1},
        {-1},
        {-1,-1},
        {23570,29246,-1}
    },
    {
        {23571,-1},
        {-1},
        {-1,-1},
        {31034,29246,-1}
    },
    {
        {23571,-1},
        {-1},
        {23571,-1},
        {31034,29246,-1}
    },
    {
        {23574,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33789,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22645,-1},
        {22645,-1,-1}
    },
    {
        {23578,-1},
        {-1},
        {-1,-1},
        {23578,-1,-1}
    },
    {
        {23578,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23578,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20040,40637},
        {40636,40637,-1}
    },
    {
        {-1,-1},
        {-1},
        {23574,-1},
        {-1,-1,-1}
    },
    {
        {23582,-1},
        {-1},
        {-1,-1},
        {29134,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23584,-1,-1}
    },
    {
        {23584,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23586,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23586,-1},
        {-1},
        {-1,-1},
        {23588,-1,-1}
    },
    {
        {23587,-1},
        {-1},
        {-1,-1},
        {23586,-1,-1}
    },
    {
        {23587,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23588,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23592,-1},
        {-1},
        {-1,-1},
        {23592,-1,-1}
    },
    {
        {23592,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23592,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23593,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23595,-1},
        {-1},
        {23595,-1},
        {23593,-1,-1}
    },
    {
        {23595,-1},
        {-1},
        {-1,-1},
        {23593,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23597,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23597,-1},
        {-1},
        {-1,-1},
        {22575,-1,-1}
    },
    {
        {23597,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26693,-1},
        {-1,-1,-1}
    },
    {
        {23600,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23601,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27070,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23608,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23609,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23609,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23610,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23611,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23612,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23613,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23614,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23615,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23616,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23617,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31348,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23621,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23622,-1},
        {-1},
        {-1,-1},
        {23626,-1,-1}
    },
    {
        {23622,-1},
        {-1},
        {-1,-1},
        {23626,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23626,-1},
        {-1,-1,-1}
    },
    {
        {23624,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23647,-1},
        {-1,-1,-1}
    },
    {
        {23626,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23627,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21514,-1},
        {-1,-1,-1}
    },
    {
        {23629,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23630,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23631,-1},
        {-1},
        {-1,-1},
        {23643,-1,-1}
    },
    {
        {23631,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23632,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23633,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23633,-1},
        {-1},
        {-1,-1},
        {23633,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23480,-1},
        {-1,-1,-1}
    },
    {
        {23635,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23635,-1},
        {-1},
        {-1,-1},
        {36000,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22958,-1},
        {-1,-1,-1}
    },
    {
        {23637,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23637,-1},
        {-1},
        {-1,-1},
        {23637,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29312,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28431,-1,-1}
    },
    {
        {23643,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33778,-1},
        {-1,-1,-1}
    },
    {
        {23646,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23647,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23648,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23648,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23649,-1},
        {-1},
        {-1,-1},
        {23650,-1,-1}
    },
    {
        {23650,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23651,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23652,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23653,-1},
        {-1},
        {-1,-1},
        {23653,-1,-1}
    },
    {
        {23653,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23655,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23656,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23657,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20698,23534},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23657,-1},
        {23657,-1,-1}
    },
    {
        {23660,-1},
        {-1},
        {-1,-1},
        {23646,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23662,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23663,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23664,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23665,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20185,-1},
        {20185,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23670,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {138541,-1},
        {-1,-1,-1}
    },
    {
        {23673,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23674,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23676,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23665,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20986,-1},
        {20986,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32993,23789},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22338,38442},
        {22338,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23688,-1},
        {-1},
        {-1,-1},
        {23688,35898,-1}
    },
    {
        {23688,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23690,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23692,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23692,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23692,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38283,24181},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38283,24181},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23700,-1},
        {-1,-1,-1}
    },
    {
        {23695,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23696,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23697,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {138652,-1},
        {-1,-1,-1}
    },
    {
        {23700,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23700,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33445,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23711,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23712,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23713,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23714,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23715,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31348,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22369,-1},
        {-1,-1,-1}
    },
    {
        {23718,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23719,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23720,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23721,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23722,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23723,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23724,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23725,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38464,-1},
        {38464,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37043,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22387,-1},
        {-1,-1,-1}
    },
    {
        {23729,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23731,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23943,-1,-1}
    },
    {
        {23733,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23734,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23735,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23736,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23719,-1},
        {23719,-1,-1}
    },
    {
        {23738,-1},
        {-1},
        {-1,-1},
        {23725,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23739,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23740,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23740,-1},
        {-1},
        {-1,-1},
        {23740,-1,-1}
    },
    {
        {23742,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37030,-1,-1}
    },
    {
        {23749,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23751,-1},
        {-1,-1,-1}
    },
    {
        {23751,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24682,-1},
        {-1,-1,-1}
    },
    {
        {23753,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38428,-1},
        {38428,-1,-1}
    },
    {
        {23755,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22436,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23874,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33563,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {14312,26669},
        {-1,-1,-1}
    },
    {
        {23760,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33603,-1},
        {-1,-1,-1}
    },
    {
        {23762,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21670,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21694,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28892,-1},
        {-1,-1,-1}
    },
    {
        {23769,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22862,23839},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33570,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35336,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23916,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23461,-1},
        {-1,-1,-1}
    },
    {
        {23776,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23777,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24018,-1},
        {-1,-1,-1}
    },
    {
        {23784,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23785,-1},
        {-1},
        {-1,-1},
        {23784,-1,-1}
    },
    {
        {23786,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29706,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36628,-1},
        {36628,-1,-1}
    },
    {
        {23789,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23789,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23791,-1},
        {-1},
        {-1,-1},
        {23792,-1,-1}
    },
    {
        {23792,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23796,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23992,-1},
        {23992,-1,-1}
    },
    {
        {23798,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33688,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22478,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23700,-1},
        {-1,-1,-1}
    },
    {
        {23802,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23802,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23803,-1},
        {-1},
        {-1,-1},
        {23803,-1,-1}
    },
    {
        {23803,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23805,-1},
        {-1},
        {-1,-1},
        {23777,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21780,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31588,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23814,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23814,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23815,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23815,-1},
        {23815,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27795,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23819,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23819,-1},
        {-1},
        {-1,-1},
        {23819,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {136004,-1},
        {-1,-1,-1}
    },
    {
        {23821,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23822,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23735,-1},
        {23735,-1,-1}
    },
    {
        {23824,-1},
        {-1},
        {-1,-1},
        {23825,-1,-1}
    },
    {
        {23825,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23828,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23829,-1},
        {-1},
        {-1,-1},
        {23830,-1,-1}
    },
    {
        {23830,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23831,-1},
        {-1},
        {-1,-1},
        {23713,-1,-1}
    },
    {
        {23832,-1},
        {-1},
        {-1,-1},
        {23833,-1,-1}
    },
    {
        {23833,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23834,-1},
        {-1},
        {-1,-1},
        {23834,-1,-1}
    },
    {
        {23834,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23835,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22517,-1},
        {-1,-1,-1}
    },
    {
        {23837,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23837,-1},
        {-1},
        {-1,-1},
        {23837,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22523,-1},
        {-1,-1,-1}
    },
    {
        {23839,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23840,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28085,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28085,-1},
        {-1,-1,-1}
    },
    {
        {23842,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23842,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23844,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22500,-1},
        {-1,-1,-1}
    },
    {
        {23846,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23847,-1},
        {-1},
        {-1,-1},
        {23913,-1,-1}
    },
    {
        {23847,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23849,-1},
        {-1},
        {-1,-1},
        {23849,-1,-1}
    },
    {
        {23849,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23851,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23857,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24841,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22580,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33873,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26978,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {19132,-1},
        {-1,-1,-1}
    },
    {
        {23865,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33913,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23769,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22564,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26950,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26609,-1},
        {26609,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30890,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30890,-1},
        {-1,-1,-1}
    },
    {
        {23874,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23172,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23878,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23879,-1},
        {-1},
        {-1,-1},
        {23878,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23159,-1},
        {-1,-1,-1}
    },
    {
        {23882,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23883,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23884,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21209,26052},
        {-1,-1,-1}
    },
    {
        {23886,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23888,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23890,-1},
        {-1},
        {-1,-1},
        {23721,-1,-1}
    },
    {
        {23891,-1},
        {-1},
        {-1,-1},
        {23721,23890,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29497,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27507,-1},
        {27507,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23900,-1},
        {-1},
        {-1,-1},
        {23822,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22615,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28179,35903,-1}
    },
    {
        {23905,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22616,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25641,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {134805,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22626,-1},
        {-1,-1,-1}
    },
    {
        {23913,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23913,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23914,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23916,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23917,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22625,22635},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22625,22635},
        {-1,-1,-1}
    },
    {
        {23919,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23920,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33260,-1},
        {-1,-1,-1}
    },
    {
        {23923,-1},
        {-1},
        {-1,-1},
        {23919,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22601,-1},
        {-1,-1,-1}
    },
    {
        {23926,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23926,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27141,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34031,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24951,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28411,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28399,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27159,-1},
        {-1,-1,-1}
    },
    {
        {23937,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23938,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23938,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23940,-1,-1}
    },
    {
        {23940,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29100,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22024,-1},
        {-1,-1,-1}
    },
    {
        {23943,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23943,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23944,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23828,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23947,-1},
        {-1},
        {-1,-1},
        {23798,-1,-1}
    },
    {
        {23948,-1},
        {-1},
        {-1,-1},
        {23798,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24948,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34012,-1},
        {-1,-1,-1}
    },
    {
        {23952,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23952,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27233,-1},
        {-1,-1,-1}
    },
    {
        {23954,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23954,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23956,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23846,-1},
        {23846,-1,-1}
    },
    {
        {23959,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26855,-1},
        {-1,-1,-1}
    },
    {
        {23961,-1},
        {-1},
        {-1,-1},
        {23961,-1,-1}
    },
    {
        {23961,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31786,-1},
        {-1,-1,-1}
    },
    {
        {23965,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22549,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23968,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27227,-1},
        {-1,-1,-1}
    },
    {
        {23970,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23972,-1},
        {-1},
        {-1,-1},
        {23970,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23975,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30992,22470},
        {30992,30830,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23980,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34203,-1},
        {-1,-1,-1}
    },
    {
        {23982,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34170,-1},
        {-1,-1,-1}
    },
    {
        {23984,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22134,-1},
        {-1,-1,-1}
    },
    {
        {23986,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23986,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22320,-1},
        {22320,-1,-1}
    },
    {
        {23988,-1},
        {-1},
        {-1,-1},
        {22885,22887,-1}
    },
    {
        {-1,-1},
        {-1},
        {38585,-1},
        {38585,-1,-1}
    },
    {
        {23991,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23992,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23993,-1},
        {-1},
        {-1,-1},
        {23947,23798,-1}
    },
    {
        {23994,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25130,-1},
        {-1,-1,-1}
    },
    {
        {23996,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23997,-1},
        {-1},
        {-1,-1},
        {23731,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38577,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25130,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27348,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27348,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23986,35215},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23992,35597},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38816,-1},
        {-1,-1,-1}
    },
    {
        {24009,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24009,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22198,-1},
        {-1,-1,-1}
    },
    {
        {24011,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24012,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24013,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24013,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24013,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25085,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24017,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24018,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24019,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24019,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24019,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24022,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24023,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24024,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24024,-1},
        {24024,-1,-1}
    },
    {
        {24027,-1},
        {-1},
        {-1,-1},
        {24029,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24029,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24030,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24032,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24033,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24033,-1},
        {-1},
        {-1,-1},
        {24033,-1,-1}
    },
    {
        {24034,-1},
        {-1},
        {-1,-1},
        {24035,-1,-1}
    },
    {
        {24034,-1},
        {-1},
        {-1,-1},
        {24034,24035,-1}
    },
    {
        {24035,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24036,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24037,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24038,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24039,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24040,-1},
        {-1},
        {-1,-1},
        {24040,-1,-1}
    },
    {
        {24040,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24041,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24041,-1},
        {-1},
        {-1,-1},
        {24041,-1,-1}
    },
    {
        {24043,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24046,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24049,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24050,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24051,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24052,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24053,-1},
        {-1},
        {-1,-1},
        {21358,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26157,-1},
        {-1,-1,-1}
    },
    {
        {24055,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24055,-1},
        {-1},
        {-1,-1},
        {24055,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21370,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24061,-1},
        {24061,-1,-1}
    },
    {
        {24059,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24059,-1},
        {-1},
        {-1,-1},
        {24059,-1,-1}
    },
    {
        {24061,-1},
        {-1},
        {-1,-1},
        {24061,-1,-1}
    },
    {
        {24061,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24061,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24062,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24066,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24064,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21360,24163},
        {21360,-1,-1}
    },
    {
        {24066,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24067,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24070,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24107,29509},
        {24107,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20221,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34969,24085},
        {-1,-1,-1}
    },
    {
        {24075,-1},
        {-1},
        {-1,-1},
        {32025,-1,-1}
    },
    {
        {24076,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24561,-1},
        {-1,-1,-1}
    },
    {
        {24081,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24082,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24084,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24085,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24086,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24088,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24089,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24090,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24090,-1},
        {-1},
        {-1,-1},
        {24090,-1,-1}
    },
    {
        {24091,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24093,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24093,-1},
        {-1},
        {-1,-1},
        {24093,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32072,-1},
        {-1,-1,-1}
    },
    {
        {24095,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24688,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24081,-1},
        {-1,-1,-1}
    },
    {
        {24101,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24742,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24742,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23789,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20402,-1},
        {-1,-1,-1}
    },
    {
        {24107,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35033,-1},
        {35033,-1,-1}
    },
    {
        {24109,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23609,-1,-1}
    },
    {
        {24110,-1},
        {-1},
        {-1,-1},
        {24171,-1,-1}
    },
    {
        {24110,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24111,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24112,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24112,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24112,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24115,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38857,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24118,-1},
        {-1},
        {-1,-1},
        {24111,-1,-1}
    },
    {
        {24119,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24120,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25509,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26827,-1},
        {-1,-1,-1}
    },
    {
        {24125,-1},
        {-1},
        {-1,-1},
        {24125,-1,-1}
    },
    {
        {24125,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24126,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24126,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20399,-1,-1}
    },
    {
        {24128,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24128,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20906,32643},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20906,32643},
        {-1,-1,-1}
    },
    {
        {24131,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24131,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24131,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24132,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24133,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24135,-1},
        {-1},
        {-1,-1},
        {24171,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23643,23631},
        {23643,-1,-1}
    },
    {
        {24137,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27091,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24140,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24902,-1},
        {-1,-1,-1}
    },
    {
        {24142,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35138,-1},
        {-1,-1,-1}
    },
    {
        {24148,-1},
        {-1},
        {-1,-1},
        {24148,-1,-1}
    },
    {
        {24148,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24149,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24149,-1},
        {-1},
        {-1,-1},
        {24149,-1,-1}
    },
    {
        {24150,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24151,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24152,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24148,24149},
        {24149,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24148,24149},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24171,24135},
        {24171,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24986,-1},
        {-1,-1,-1}
    },
    {
        {24158,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24159,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24159,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28961,25771},
        {-1,-1,-1}
    },
    {
        {24161,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24162,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24162,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24163,-1},
        {-1},
        {-1,-1},
        {24163,-1,-1}
    },
    {
        {24163,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24164,-1},
        {-1},
        {-1,-1},
        {24163,-1,-1}
    },
    {
        {24164,-1},
        {-1},
        {-1,-1},
        {24163,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27287,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25030,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24996,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24996,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24171,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24172,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24173,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24173,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24173,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24173,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24174,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27355,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27355,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35188,-1},
        {-1,-1,-1}
    },
    {
        {24178,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24179,-1},
        {-1},
        {-1,-1},
        {24179,-1,-1}
    },
    {
        {24179,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24180,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24180,-1},
        {-1},
        {-1,-1},
        {24180,-1,-1}
    },
    {
        {24181,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24181,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24181,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24182,-1},
        {-1},
        {-1,-1},
        {24183,-1,-1}
    },
    {
        {24183,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24184,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24184,-1,-1}
    },
    {
        {24185,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24186,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24187,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24188,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24189,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24190,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24191,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24193,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20164,-1},
        {-1,-1,-1}
    },
    {
        {24195,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24196,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24199,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20170,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24202,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20916,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24193,24307},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24207,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24300,30439},
        {24300,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24196,-1},
        {24196,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24213,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24214,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24214,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24215,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24235,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24255,24287},
        {24287,-1,-1}
    },
    {
        {24218,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30133,-1},
        {-1,-1,-1}
    },
    {
        {24220,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30140,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24224,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {17598,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24229,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24230,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24231,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33660,-1},
        {-1,-1,-1}
    },
    {
        {24234,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24235,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24236,-1},
        {-1},
        {-1,-1},
        {21398,-1,-1}
    },
    {
        {24236,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24236,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24237,-1},
        {-1},
        {-1,-1},
        {24237,-1,-1}
    },
    {
        {24237,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37193,33564},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23631,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33802,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21405,-1},
        {-1,-1,-1}
    },
    {
        {24245,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24246,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24247,-1},
        {-1},
        {-1,-1},
        {24247,-1,-1}
    },
    {
        {24247,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24248,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24249,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24249,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26494,-1},
        {26494,-1,-1}
    },
    {
        {24246,-1},
        {-1},
        {36974,24246},
        {24246,-1,-1}
    },
    {
        {24246,-1},
        {-1},
        {36974,24246},
        {24246,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24254,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24254,-1},
        {-1},
        {-1,-1},
        {24254,-1,-1}
    },
    {
        {24255,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24255,-1},
        {-1},
        {-1,-1},
        {24255,24287,-1}
    },
    {
        {-1,-1},
        {-1},
        {24267,-1},
        {-1,-1,-1}
    },
    {
        {24257,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24258,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24259,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24271,-1},
        {-1},
        {21417,24271},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33995,-1},
        {-1,-1,-1}
    },
    {
        {24262,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38692,-1},
        {-1,-1,-1}
    },
    {
        {24264,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24265,-1},
        {-1},
        {-1,-1},
        {24265,-1,-1}
    },
    {
        {24265,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24266,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24267,-1},
        {-1},
        {-1,-1},
        {24267,-1,-1}
    },
    {
        {24267,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24271,-1},
        {-1},
        {-1,-1},
        {21417,-1,-1}
    },
    {
        {24271,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24272,-1},
        {-1},
        {-1,-1},
        {21417,24271,-1}
    },
    {
        {24272,-1},
        {-1},
        {-1,-1},
        {21417,24271,-1}
    },
    {
        {24272,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24272,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24272,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24272,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24274,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24275,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24278,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24278,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24109,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40575,-1},
        {-1,-1,-1}
    },
    {
        {24281,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24281,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24282,-1},
        {-1},
        {-1,-1},
        {21416,-1,-1}
    },
    {
        {24283,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24285,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24287,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24287,-1},
        {-1},
        {-1,-1},
        {24287,-1,-1}
    },
    {
        {24288,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24288,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24289,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24290,-1},
        {-1},
        {-1,-1},
        {24259,-1,-1}
    },
    {
        {24291,-1},
        {-1},
        {-1,-1},
        {24291,24195,-1}
    },
    {
        {24291,-1},
        {-1},
        {-1,-1},
        {24291,24195,-1}
    },
    {
        {24291,-1},
        {-1},
        {-1,-1},
        {24195,-1,-1}
    },
    {
        {24291,-1},
        {-1},
        {-1,-1},
        {24291,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22721,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29254,21959},
        {-1,-1,-1}
    },
    {
        {24296,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24297,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24297,-1},
        {24297,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24300,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30317,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30316,-1},
        {-1,-1,-1}
    },
    {
        {24304,-1},
        {-1},
        {-1,-1},
        {24193,24307,-1}
    },
    {
        {24305,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20791,-1},
        {-1,-1,-1}
    },
    {
        {24307,-1},
        {-1},
        {-1,-1},
        {24193,-1,-1}
    },
    {
        {24308,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24308,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24310,-1},
        {-1},
        {-1,-1},
        {24310,-1,-1}
    },
    {
        {24310,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24310,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24311,-1},
        {-1},
        {-1,-1},
        {24311,-1,-1}
    },
    {
        {24311,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24311,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24312,-1},
        {-1},
        {-1,-1},
        {36842,-1,-1}
    },
    {
        {24312,-1},
        {-1},
        {-1,-1},
        {36842,-1,-1}
    },
    {
        {24313,-1},
        {-1},
        {-1,-1},
        {36843,-1,-1}
    },
    {
        {24313,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24314,-1},
        {-1},
        {-1,-1},
        {24314,-1,-1}
    },
    {
        {24314,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24314,-1},
        {-1},
        {-1,-1},
        {24314,-1,-1}
    },
    {
        {24315,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24315,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24316,-1},
        {-1},
        {-1,-1},
        {36858,-1,-1}
    },
    {
        {24316,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24316,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24317,-1},
        {-1},
        {-1,-1},
        {24315,-1,-1}
    },
    {
        {24317,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24318,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24318,-1},
        {-1},
        {-1,-1},
        {24318,-1,-1}
    },
    {
        {24319,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24320,-1},
        {-1},
        {-1,-1},
        {38283,-1,-1}
    },
    {
        {24321,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24322,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24322,-1},
        {-1},
        {-1,-1},
        {24322,-1,-1}
    },
    {
        {24323,-1},
        {-1},
        {-1,-1},
        {26820,-1,-1}
    },
    {
        {24324,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23700,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24327,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24328,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24329,-1},
        {-1},
        {-1,-1},
        {22872,-1,-1}
    },
    {
        {24329,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24330,-1},
        {-1},
        {-1,-1},
        {24330,-1,-1}
    },
    {
        {24330,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24331,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24332,-1},
        {-1},
        {-1,-1},
        {19968,-1,-1}
    },
    {
        {24333,-1},
        {-1},
        {-1,-1},
        {20108,-1,-1}
    },
    {
        {24333,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24334,-1},
        {-1},
        {-1,-1},
        {19977,-1,-1}
    },
    {
        {24334,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24335,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24336,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24337,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24337,-1},
        {-1},
        {24337,-1},
        {24337,-1,-1}
    },
    {
        {24339,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24340,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24341,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24342,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24342,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24343,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24344,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24351,-1},
        {-1,-1,-1}
    },
    {
        {24347,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24348,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24349,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24350,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24351,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24355,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24357,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24358,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24359,-1},
        {-1},
        {-1,-1},
        {24359,-1,-1}
    },
    {
        {24359,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24361,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24361,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24365,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24365,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24367,-1},
        {-1},
        {-1,-1},
        {24398,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24369,-1},
        {-1},
        {-1,-1},
        {24369,-1,-1}
    },
    {
        {24369,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24372,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24373,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24374,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24375,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24376,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24376,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24375,-1},
        {-1},
        {-1,-1},
        {24375,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24380,-1,-1}
    },
    {
        {24380,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24381,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24382,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24384,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24385,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30050,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24389,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24389,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21029,-1},
        {-1,-1,-1}
    },
    {
        {24391,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24392,-1},
        {-1},
        {-1,-1},
        {24382,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24394,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25808,-1},
        {-1,-1,-1}
    },
    {
        {24396,-1},
        {-1},
        {-1,-1},
        {24357,-1,-1}
    },
    {
        {24396,-1},
        {-1},
        {-1,-1},
        {24396,-1,-1}
    },
    {
        {24396,-1},
        {-1},
        {-1,-1},
        {24396,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24398,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25899,-1},
        {-1,-1,-1}
    },
    {
        {24401,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24401,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24401,-1},
        {-1},
        {-1,-1},
        {24401,-1,-1}
    },
    {
        {24401,-1},
        {-1},
        {-1,-1},
        {24401,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24112,27512},
        {27512,24112,-1}
    },
    {
        {24403,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24404,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24404,-1},
        {-1},
        {37682,37636},
        {24404,-1,-1}
    },
    {
        {24406,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24406,-1},
        {-1},
        {-1,-1},
        {24406,-1,-1}
    },
    {
        {24406,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24407,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24407,-1},
        {-1},
        {-1,-1},
        {24407,-1,-1}
    },
    {
        {24408,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24409,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24409,-1},
        {-1},
        {24409,21295},
        {24409,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24412,-1},
        {-1},
        {-1,-1},
        {24413,-1,-1}
    },
    {
        {24412,-1},
        {-1},
        {-1,-1},
        {24413,-1,-1}
    },
    {
        {24412,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24413,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24413,-1},
        {-1},
        {24412,24413},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24417,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24418,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24418,-1},
        {-1},
        {-1,-1},
        {24418,-1,-1}
    },
    {
        {24419,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24420,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24422,-1},
        {-1},
        {24422,-1},
        {24422,-1,-1}
    },
    {
        {24422,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24423,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24425,-1},
        {-1},
        {-1,-1},
        {24425,-1,-1}
    },
    {
        {24425,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24426,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24426,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24427,-1},
        {-1},
        {-1,-1},
        {24427,-1,-1}
    },
    {
        {24427,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24428,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24429,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23481,-1},
        {23481,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24432,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24432,-1},
        {-1},
        {-1,-1},
        {24432,-1,-1}
    },
    {
        {24433,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24434,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24435,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20235,-1},
        {-1,-1,-1}
    },
    {
        {24439,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24441,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24444,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24445,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24446,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24447,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24448,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24448,-1},
        {-1},
        {-1,-1},
        {24448,-1,-1}
    },
    {
        {24449,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24450,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24451,-1},
        {-1},
        {-1,-1},
        {24448,-1,-1}
    },
    {
        {24452,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24453,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20385,-1},
        {-1,-1,-1}
    },
    {
        {24455,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24456,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24457,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24458,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24459,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24460,-1},
        {-1},
        {-1,-1},
        {24460,-1,-1}
    },
    {
        {24460,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20339,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36894,-1},
        {-1,-1,-1}
    },
    {
        {24463,-1},
        {-1},
        {-1,-1},
        {38495,-1,-1}
    },
    {
        {24463,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24464,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24465,-1},
        {-1},
        {-1,-1},
        {24452,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24466,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24467,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24471,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24472,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24472,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24473,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30869,-1},
        {-1,-1,-1}
    },
    {
        {24476,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24478,-1},
        {-1},
        {-1,-1},
        {24467,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20508,-1},
        {-1,-1,-1}
    },
    {
        {24480,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24481,-1},
        {-1},
        {-1,-1},
        {24481,-1,-1}
    },
    {
        {24481,-1},
        {-1},
        {-1,-1},
        {24481,-1,-1}
    },
    {
        {24481,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24482,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20511,-1},
        {20511,-1,-1}
    },
    {
        {24484,-1},
        {-1},
        {-1,-1},
        {20581,-1,-1}
    },
    {
        {24484,-1},
        {-1},
        {-1,-1},
        {20581,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24643,26159},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36944,34662},
        {-1,-1,-1}
    },
    {
        {24487,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24488,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24489,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24490,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24493,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24494,-1},
        {-1},
        {-1,-1},
        {24494,-1,-1}
    },
    {
        {24494,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24494,-1},
        {-1},
        {-1,-1},
        {24494,-1,-1}
    },
    {
        {24494,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24495,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24449,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20676,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24499,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24500,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24500,-1},
        {-1},
        {-1,-1},
        {24500,24501,-1}
    },
    {
        {24501,-1},
        {-1},
        {-1,-1},
        {24500,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25735,-1},
        {-1,-1,-1}
    },
    {
        {24503,-1},
        {-1},
        {-1,-1},
        {24499,-1,-1}
    },
    {
        {24503,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24504,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24505,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20744,-1},
        {-1,-1,-1}
    },
    {
        {24508,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24509,-1},
        {-1},
        {-1,-1},
        {24509,-1,-1}
    },
    {
        {24509,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24500,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28689,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25084,-1},
        {-1,-1,-1}
    },
    {
        {24515,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24516,-1},
        {-1},
        {-1,-1},
        {24515,-1,-1}
    },
    {
        {24517,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24519,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20161,-1,-1}
    },
    {
        {24521,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20163,24531},
        {-1,-1,-1}
    },
    {
        {24523,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24524,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24525,-1},
        {-1},
        {-1,-1},
        {24525,-1,-1}
    },
    {
        {24525,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24525,-1},
        {-1},
        {-1,-1},
        {24525,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20161,-1},
        {20161,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25082,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24530,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24531,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24532,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24534,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24535,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24536,-1},
        {-1},
        {-1,-1},
        {24536,-1,-1}
    },
    {
        {24536,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24537,-1},
        {-1},
        {-1,-1},
        {24537,-1,-1}
    },
    {
        {24537,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20182,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24540,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24541,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24542,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24542,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24542,-1},
        {24542,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24542,-1},
        {-1,-1,-1}
    },
    {
        {24544,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24545,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24546,-1},
        {-1},
        {-1,-1},
        {24735,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24613,-1},
        {24613,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24548,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24962,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24553,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24555,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24558,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24559,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24560,-1},
        {-1},
        {-1,-1},
        {24756,-1,-1}
    },
    {
        {24560,-1},
        {-1},
        {-1,-1},
        {24756,-1,-1}
    },
    {
        {24561,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24563,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20209,24604},
        {-1,-1,-1}
    },
    {
        {24565,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24671,-1},
        {24671,-1,-1}
    },
    {
        {24568,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29378,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24571,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24572,-1},
        {-1},
        {-1,-1},
        {24951,-1,-1}
    },
    {
        {24573,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24532,24894},
        {-1,-1,-1}
    },
    {
        {24575,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24575,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25040,25079},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28846,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24795,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20286,-1},
        {-1,-1,-1}
    },
    {
        {24589,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24590,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24591,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24592,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20276,-1},
        {-1,-1,-1}
    },
    {
        {24594,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24596,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24597,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24598,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24599,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24600,-1},
        {-1},
        {-1,-1},
        {22266,-1,-1}
    },
    {
        {24601,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24603,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24604,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24605,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20279,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24608,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24609,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24610,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23588,-1},
        {23588,-1,-1}
    },
    {
        {24612,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24613,-1},
        {-1},
        {-1,-1},
        {24613,-1,-1}
    },
    {
        {24613,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24613,-1,-1}
    },
    {
        {24614,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24614,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24615,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24616,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24617,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24618,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24619,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21612,-1},
        {21612,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22387,-1},
        {-1,-1,-1}
    },
    {
        {24623,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28843,-1},
        {-1,-1,-1}
    },
    {
        {24625,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28851,-1},
        {-1,-1,-1}
    },
    {
        {24627,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24629,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24444,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24713,26414},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24713,-1},
        {24713,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20182,-1},
        {20182,-1,-1}
    },
    {
        {24634,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24641,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24642,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24643,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20342,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20332,-1},
        {-1,-1,-1}
    },
    {
        {24646,-1},
        {-1},
        {-1,-1},
        {24658,-1,-1}
    },
    {
        {24647,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24650,-1},
        {-1},
        {-1,-1},
        {21332,-1,-1}
    },
    {
        {24651,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24652,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24653,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24594,22909},
        {-1,-1,-1}
    },
    {
        {24656,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24658,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24676,24765},
        {24765,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24661,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24605,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24774,-1},
        {-1,-1,-1}
    },
    {
        {24665,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24666,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20330,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22839,-1},
        {-1,-1,-1}
    },
    {
        {24671,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24672,-1},
        {-1},
        {-1,-1},
        {24618,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24715,-1,-1}
    },
    {
        {24674,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24674,-1},
        {-1},
        {-1,-1},
        {24674,-1,-1}
    },
    {
        {24674,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24675,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24676,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24677,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24677,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24797,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24680,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24681,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24682,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24683,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24684,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24685,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20346,-1},
        {-1,-1,-1}
    },
    {
        {24687,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24688,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24742,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24742,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20341,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25031,-1},
        {25031,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24499,24503},
        {24499,24503,-1}
    },
    {
        {24693,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24695,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24703,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24705,-1},
        {-1},
        {-1,-1},
        {24705,-1,-1}
    },
    {
        {24705,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24707,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24708,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24708,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24742,-1},
        {-1},
        {-1,-1},
        {24742,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24713,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24714,-1},
        {-1},
        {-1,-1},
        {21746,-1,-1}
    },
    {
        {24715,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24715,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24716,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24717,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20448,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24789,-1},
        {24789,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24598,-1,-1}
    },
    {
        {24722,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20436,-1,-1}
    },
    {
        {24724,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24725,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24726,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24727,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24727,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24727,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24775,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24730,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24731,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24733,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24734,-1},
        {-1},
        {-1,-1},
        {35492,-1,-1}
    },
    {
        {24734,-1},
        {-1},
        {35492,24734},
        {35492,-1,-1}
    },
    {
        {24735,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24736,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24833,21033},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24739,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24742,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24743,-1},
        {-1},
        {-1,-1},
        {20432,-1,-1}
    },
    {
        {24744,-1},
        {-1},
        {-1,-1},
        {20320,-1,-1}
    },
    {
        {24744,-1},
        {-1},
        {-1,-1},
        {24744,20320,-1}
    },
    {
        {24745,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24746,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24752,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24753,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24754,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24754,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24755,-1},
        {-1},
        {-1,-1},
        {24499,24503,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24756,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24757,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24758,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24760,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23448,24857},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24764,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24765,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24766,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24766,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20523,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24714,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28956,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24773,-1},
        {-1},
        {-1,-1},
        {24773,-1,-1}
    },
    {
        {24773,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24774,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24774,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24775,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24775,-1},
        {-1},
        {-1,-1},
        {24775,-1,-1}
    },
    {
        {24776,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24779,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24616,23451},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24785,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21644,-1},
        {21644,-1,-1}
    },
    {
        {24787,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24787,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24788,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24789,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24789,-1},
        {24789,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25467,-1},
        {-1,-1,-1}
    },
    {
        {24792,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24792,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24792,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24793,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24794,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24795,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24796,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24797,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24797,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27427,-1,-1}
    },
    {
        {24799,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24800,-1},
        {-1},
        {-1,-1},
        {24693,-1,-1}
    },
    {
        {24800,-1},
        {-1},
        {-1,-1},
        {24693,-1,-1}
    },
    {
        {24801,-1},
        {-1},
        {-1,-1},
        {24746,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24803,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24358,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24898,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24807,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24807,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24808,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {64064,25074},
        {64064,25074,-1}
    },
    {
        {24814,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24816,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24817,-1},
        {-1},
        {-1,-1},
        {24745,-1,-1}
    },
    {
        {24818,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24819,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24820,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24821,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24822,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24823,-1},
        {-1},
        {-1,-1},
        {34850,-1,-1}
    },
    {
        {24824,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24825,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24825,-1},
        {-1},
        {-1,-1},
        {24825,-1,-1}
    },
    {
        {24826,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24827,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24795,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26934,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35551,-1},
        {-1,-1,-1}
    },
    {
        {24832,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24833,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24835,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20558,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38761,-1},
        {-1,-1,-1}
    },
    {
        {24838,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38873,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38873,-1,-1}
    },
    {
        {24840,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24840,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24840,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24841,-1},
        {-1},
        {-1,-1},
        {24841,-1,-1}
    },
    {
        {24841,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25588,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24845,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24846,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24847,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24847,-1},
        {-1},
        {-1,-1},
        {24847,-1,-1}
    },
    {
        {24848,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24850,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24850,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24851,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24853,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24853,-1},
        {-1},
        {-1,-1},
        {24853,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24682,-1},
        {-1,-1,-1}
    },
    {
        {24857,-1},
        {-1},
        {-1,-1},
        {24682,-1,-1}
    },
    {
        {24858,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24859,-1},
        {-1},
        {-1,-1},
        {24859,-1,-1}
    },
    {
        {24859,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24860,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20547,-1},
        {-1,-1,-1}
    },
    {
        {24862,-1},
        {-1},
        {-1,-1},
        {25062,-1,-1}
    },
    {
        {24863,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24909,-1},
        {-1},
        {24616,24909},
        {24909,-1,-1}
    },
    {
        {24865,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26974,-1},
        {26974,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29008,-1},
        {-1,-1,-1}
    },
    {
        {24871,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24872,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29077,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24875,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24876,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27056,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24652,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24730,-1},
        {-1,-1,-1}
    },
    {
        {24880,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25084,-1},
        {25084,-1,-1}
    },
    {
        {24884,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24887,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25972,-1},
        {25972,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24890,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24890,-1},
        {-1},
        {-1,-1},
        {24890,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36956,-1},
        {-1,-1,-1}
    },
    {
        {24892,-1},
        {-1},
        {-1,-1},
        {24910,-1,-1}
    },
    {
        {24892,-1},
        {-1},
        {-1,-1},
        {24910,-1,-1}
    },
    {
        {24892,-1},
        {-1},
        {-1,-1},
        {24892,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24893,-1},
        {-1},
        {-1,-1},
        {21338,-1,-1}
    },
    {
        {24893,-1},
        {-1},
        {-1,-1},
        {21338,-1,-1}
    },
    {
        {24893,-1},
        {-1},
        {-1,-1},
        {21338,-1,-1}
    },
    {
        {24894,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24895,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20626,-1},
        {-1,-1,-1}
    },
    {
        {24897,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24898,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24900,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24902,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24903,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24904,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24904,-1},
        {-1},
        {-1,-1},
        {24904,-1,-1}
    },
    {
        {24904,-1},
        {-1},
        {-1,-1},
        {24904,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24906,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24906,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24907,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24908,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24908,-1},
        {-1},
        {-1,-1},
        {24908,-1,-1}
    },
    {
        {24908,-1},
        {-1},
        {-1,-1},
        {24908,-1,-1}
    },
    {
        {24909,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24910,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24910,-1},
        {-1},
        {-1,-1},
        {24892,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28319,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27133,-1},
        {-1,-1,-1}
    },
    {
        {24915,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27169,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24917,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24917,-1},
        {-1},
        {-1,-1},
        {24917,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24151,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24983,-1},
        {-1,-1,-1}
    },
    {
        {24920,-1},
        {-1},
        {-1,-1},
        {24808,-1,-1}
    },
    {
        {24921,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24922,-1},
        {-1},
        {-1,-1},
        {24921,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32633,27119},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28259,24845},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28259,24845},
        {-1,-1,-1}
    },
    {
        {24925,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24925,-1},
        {-1},
        {-1,-1},
        {24925,-1,-1}
    },
    {
        {24925,-1},
        {-1},
        {-1,-1},
        {24925,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24927,-1},
        {-1},
        {-1,-1},
        {24683,-1,-1}
    },
    {
        {24928,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29245,-1},
        {-1,-1,-1}
    },
    {
        {24930,-1},
        {-1},
        {-1,-1},
        {24930,-1,-1}
    },
    {
        {24930,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24931,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24872,-1,-1}
    },
    {
        {24933,-1},
        {-1},
        {-1,-1},
        {24933,-1,-1}
    },
    {
        {24933,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24933,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24983,-1},
        {-1,-1,-1}
    },
    {
        {24935,-1},
        {-1},
        {-1,-1},
        {24935,-1,-1}
    },
    {
        {24935,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24936,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24936,-1},
        {-1},
        {-1,-1},
        {24936,-1,-1}
    },
    {
        {24936,-1},
        {-1},
        {-1,-1},
        {24936,-1,-1}
    },
    {
        {24936,-1},
        {-1},
        {-1,-1},
        {24936,-1,-1}
    },
    {
        {24936,-1},
        {-1},
        {-1,-1},
        {24936,-1,-1}
    },
    {
        {24936,-1},
        {-1},
        {-1,-1},
        {24936,-1,-1}
    },
    {
        {24936,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24936,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36899,28451},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27572,27590},
        {-1,-1,-1}
    },
    {
        {24939,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24982,-1},
        {24982,-1,-1}
    },
    {
        {24942,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24943,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24944,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24945,-1},
        {-1},
        {-1,-1},
        {21338,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24947,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24948,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24948,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24949,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24950,-1},
        {-1},
        {-1,-1},
        {24950,-1,-1}
    },
    {
        {24950,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24951,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22519,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20674,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20674,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24956,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24956,-1,-1}
    },
    {
        {24958,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24977,-1},
        {-1},
        {-1,-1},
        {24977,-1,-1}
    },
    {
        {24977,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24962,-1},
        {-1},
        {-1,-1},
        {24962,-1,-1}
    },
    {
        {24962,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24927,-1},
        {24683,24927,-1}
    },
    {
        {-1,-1},
        {-1},
        {30624,-1},
        {30624,-1,-1}
    },
    {
        {24967,-1},
        {-1},
        {-1,-1},
        {25001,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22687,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28558,-1},
        {-1,-1,-1}
    },
    {
        {24970,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24970,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37406,-1},
        {-1,-1,-1}
    },
    {
        {24973,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24974,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {14428,-1},
        {-1,-1,-1}
    },
    {
        {24976,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24976,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24976,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24977,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24978,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24979,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24979,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24980,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29128,-1},
        {-1,-1,-1}
    },
    {
        {24982,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24983,-1},
        {-1},
        {-1,-1},
        {24982,-1,-1}
    },
    {
        {24984,-1},
        {-1},
        {-1,-1},
        {21916,-1,-1}
    },
    {
        {24985,-1},
        {-1},
        {-1,-1},
        {21916,-1,-1}
    },
    {
        {24986,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23563,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23563,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24816,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25958,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31574,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20709,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32763,-1},
        {-1,-1,-1}
    },
    {
        {24996,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24996,-1},
        {-1},
        {-1,-1},
        {24996,-1,-1}
    },
    {
        {24996,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24997,-1},
        {-1},
        {-1,-1},
        {21172,21214,-1}
    },
    {
        {-1,-1},
        {-1},
        {24997,-1},
        {21172,24997,21214}
    },
    {
        {24999,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24999,-1},
        {-1},
        {-1,-1},
        {24999,-1,-1}
    },
    {
        {25000,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25001,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25003,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25004,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25005,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25006,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24808,24920},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24808,24920},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27224,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20710,-1},
        {-1,-1,-1}
    },
    {
        {25010,-1},
        {-1},
        {-1,-1},
        {25010,-1,-1}
    },
    {
        {25010,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25010,-1},
        {-1},
        {-1,-1},
        {25010,-1,-1}
    },
    {
        {25010,-1},
        {-1},
        {-1,-1},
        {25010,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24944,23318},
        {-1,-1,-1}
    },
    {
        {25014,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25014,-1},
        {-1},
        {-1,-1},
        {25014,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30990,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25017,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25018,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22374,-1},
        {22374,-1,-1}
    },
    {
        {25020,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25020,-1},
        {-1},
        {-1,-1},
        {25020,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24865,33905},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25022,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20709,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25027,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21220,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25818,-1},
        {-1,-1,-1}
    },
    {
        {25030,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25031,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25032,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25033,-1},
        {-1},
        {-1,-1},
        {24540,-1,-1}
    },
    {
        {25034,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25034,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25035,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25036,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25036,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25037,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25040,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24297,-1},
        {25037,-1,-1}
    },
    {
        {25045,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25045,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25001,37531},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29184,36203},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31337,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25053,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25055,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20757,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25059,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20756,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25062,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25062,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25045,-1},
        {25045,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29190,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24962,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24612,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21129,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24173,-1},
        {-1,-1,-1}
    },
    {
        {25074,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25074,-1},
        {-1},
        {-1,-1},
        {64064,25074,-1}
    },
    {
        {25076,-1},
        {-1},
        {-1,-1},
        {25082,-1,-1}
    },
    {
        {25077,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25077,-1},
        {-1},
        {-1,-1},
        {25077,-1,-1}
    },
    {
        {25078,-1},
        {-1},
        {-1,-1},
        {23358,25078,-1}
    },
    {
        {25078,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25079,-1},
        {-1},
        {-1,-1},
        {25040,-1,-1}
    },
    {
        {25080,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25082,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20864,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20864,-1},
        {-1,-1,-1}
    },
    {
        {25084,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25085,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25085,-1},
        {-1},
        {-1,-1},
        {25085,-1,-1}
    },
    {
        {25086,-1},
        {-1},
        {-1,-1},
        {24948,-1,-1}
    },
    {
        {25086,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25087,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25087,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25088,-1},
        {-1},
        {-1,-1},
        {24651,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38753,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24797,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25096,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25097,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25098,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25100,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25101,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25102,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25102,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25104,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25104,-1},
        {-1},
        {-1,-1},
        {25104,-1,-1}
    },
    {
        {25105,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25106,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25108,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25109,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25110,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25136,25126},
        {25136,25126,-1}
    },
    {
        {25114,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25115,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25115,-1},
        {-1},
        {-1,-1},
        {25115,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25117,-1},
        {-1},
        {-1,-1},
        {36001,-1,-1}
    },
    {
        {25118,-1},
        {-1},
        {-1,-1},
        {25115,-1,-1}
    },
    {
        {25119,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25121,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25122,-1},
        {-1},
        {-1,-1},
        {25122,-1,-1}
    },
    {
        {25122,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25123,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30408,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31561,-1},
        {31561,-1,-1}
    },
    {
        {25126,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21109,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27468,-1},
        {-1,-1,-1}
    },
    {
        {25129,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25130,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25129,-1},
        {-1},
        {25129,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23493,-1},
        {-1,-1,-1}
    },
    {
        {25134,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25134,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25135,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25136,-1},
        {-1},
        {-1,-1},
        {25126,-1,-1}
    },
    {
        {25135,-1},
        {-1},
        {-1,-1},
        {25138,25135,-1}
    },
    {
        {25138,-1},
        {-1},
        {-1,-1},
        {25135,-1,-1}
    },
    {
        {25139,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25139,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25140,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25140,-1},
        {-1},
        {-1,-1},
        {25140,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30655,-1},
        {-1,-1,-1}
    },
    {
        {25144,-1},
        {-1},
        {-1,-1},
        {25144,-1,-1}
    },
    {
        {25144,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25147,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25147,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22808,21359},
        {21359,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25150,-1},
        {-1},
        {-1,-1},
        {25147,-1,-1}
    },
    {
        {25151,-1},
        {-1},
        {-1,-1},
        {25151,-1,-1}
    },
    {
        {25151,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25151,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25152,-1},
        {-1},
        {-1,-1},
        {25152,-1,-1}
    },
    {
        {25152,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25153,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25153,-1},
        {-1},
        {-1,-1},
        {25153,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25155,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22810,25159},
        {-1,-1,-1}
    },
    {
        {25158,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25159,-1},
        {-1},
        {-1,-1},
        {25159,-1,-1}
    },
    {
        {25159,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25160,-1},
        {-1},
        {-1,-1},
        {25160,-1,-1}
    },
    {
        {25160,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25161,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25161,-1},
        {-1},
        {-1,-1},
        {25161,-1,-1}
    },
    {
        {25161,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25161,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28814,25159},
        {-1,-1,-1}
    },
    {
        {25163,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25164,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25165,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25166,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29360,25170},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25169,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25170,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25171,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25173,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25174,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22312,-1},
        {22312,-1,-1}
    },
    {
        {25176,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26454,-1},
        {-1,-1,-1}
    },
    {
        {25178,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25178,-1},
        {-1},
        {-1,-1},
        {25178,-1,-1}
    },
    {
        {25179,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26439,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26439,-1},
        {-1,-1,-1}
    },
    {
        {25182,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24041,-1},
        {-1,-1,-1}
    },
    {
        {25184,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25184,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25185,-1},
        {-1},
        {-1,-1},
        {25301,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25187,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25188,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25182,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31175,22519},
        {22519,-1,-1}
    },
    {
        {25192,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25192,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25192,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25197,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25198,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25198,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25199,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25854,-1},
        {-1,-1,-1}
    },
    {
        {25201,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25201,-1},
        {-1},
        {-1,-1},
        {25201,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25810,-1},
        {-1,-1,-1}
    },
    {
        {25203,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26044,-1},
        {26044,-1,-1}
    },
    {
        {25206,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26512,25240},
        {26512,-1,-1}
    },
    {
        {25209,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25211,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25212,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25214,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25215,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25216,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25216,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29377,-1},
        {-1,-1,-1}
    },
    {
        {25218,-1},
        {-1},
        {-1,-1},
        {29378,-1,-1}
    },
    {
        {25219,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25219,-1},
        {-1},
        {-1,-1},
        {25219,-1,-1}
    },
    {
        {25220,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25221,-1},
        {-1},
        {-1,-1},
        {25304,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25451,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25451,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25188,-1},
        {-1,-1,-1}
    },
    {
        {25225,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25226,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20888,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25327,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25233,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25234,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25235,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25236,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25237,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25238,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25239,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25240,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25241,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25290,25771},
        {25771,-1,-1}
    },
    {
        {25243,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25244,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25303,-1},
        {25303,-1,-1}
    },
    {
        {25246,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22577,-1},
        {22577,-1,-1}
    },
    {
        {25254,-1},
        {-1},
        {-1,-1},
        {26564,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26547,-1},
        {26547,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20873,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24598,-1},
        {-1,-1,-1}
    },
    {
        {25259,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25260,-1},
        {-1},
        {-1,-1},
        {25825,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29645,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25265,-1},
        {-1},
        {-1,-1},
        {25265,-1,-1}
    },
    {
        {25265,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33655,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25586,27844},
        {-1,-1,-1}
    },
    {
        {25269,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25406,25375},
        {25406,25375,-1}
    },
    {
        {25273,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25275,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25276,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25277,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25282,-1},
        {-1},
        {-1,-1},
        {25173,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26590,25334},
        {-1,-1,-1}
    },
    {
        {25284,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25285,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25286,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25287,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25288,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25289,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25290,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25243,-1},
        {-1},
        {25243,-1},
        {-1,-1,-1}
    },
    {
        {25292,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25292,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25293,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25295,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25296,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25296,-1},
        {-1},
        {-1,-1},
        {25296,-1,-1}
    },
    {
        {25296,-1},
        {-1},
        {-1,-1},
        {25296,-1,-1}
    },
    {
        {25297,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25298,-1},
        {-1},
        {-1,-1},
        {25298,-1,-1}
    },
    {
        {25298,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25299,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25300,-1},
        {-1},
        {-1,-1},
        {25244,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25301,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25302,-1},
        {-1},
        {-1,-1},
        {25301,-1,-1}
    },
    {
        {25303,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25304,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25305,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25307,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25308,-1},
        {-1},
        {-1,-1},
        {25309,-1,-1}
    },
    {
        {25309,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25269,-1},
        {25269,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30097,25836},
        {25836,-1,-1}
    },
    {
        {25312,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25313,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25472,33390},
        {-1,-1,-1}
    },
    {
        {25324,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25325,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25326,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25327,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25218,20219},
        {-1,-1,-1}
    },
    {
        {25329,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25331,-1},
        {-1},
        {-1,-1},
        {25331,-1,-1}
    },
    {
        {25331,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25333,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25334,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25335,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25335,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21702,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21332,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35927,-1},
        {35927,-1,-1}
    },
    {
        {25340,-1},
        {-1},
        {-1,-1},
        {25340,-1,-1}
    },
    {
        {25340,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25341,-1},
        {-1},
        {-1,-1},
        {26355,-1,-1}
    },
    {
        {25342,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25343,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27966,-1},
        {-1,-1,-1}
    },
    {
        {25345,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25346,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25347,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25571,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22428,-1},
        {-1,-1,-1}
    },
    {
        {25351,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25352,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25352,-1},
        {-1},
        {-1,-1},
        {25352,-1,-1}
    },
    {
        {25352,-1},
        {-1},
        {-1,-1},
        {25352,-1,-1}
    },
    {
        {25352,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25353,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25354,-1},
        {-1},
        {-1,-1},
        {24324,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25356,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25357,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25715,21043},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24683,-1},
        {-1,-1,-1}
    },
    {
        {25360,-1},
        {-1},
        {-1,-1},
        {25295,-1,-1}
    },
    {
        {25361,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25420,21015},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26053,-1},
        {26053,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25666,25885},
        {25666,25885,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25212,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25225,-1},
        {25225,-1,-1}
    },
    {
        {25368,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25369,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25375,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25383,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25383,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25384,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25387,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26738,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26799,-1},
        {-1,-1,-1}
    },
    {
        {25391,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25394,25220},
        {-1,-1,-1}
    },
    {
        {25394,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {64068,-1},
        {-1,-1,-1}
    },
    {
        {25397,-1},
        {-1},
        {-1,-1},
        {24324,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26776,-1},
        {-1,-1,-1}
    },
    {
        {25401,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25402,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25402,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25403,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25404,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25404,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25405,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25405,-1},
        {-1},
        {-1,-1},
        {25405,-1,-1}
    },
    {
        {25406,-1},
        {-1},
        {-1,-1},
        {25375,-1,-1}
    },
    {
        {25407,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25447,-1},
        {-1,-1,-1}
    },
    {
        {25409,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25411,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25412,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25412,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26742,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21050,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25860,-1},
        {-1,-1,-1}
    },
    {
        {25417,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25420,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25421,-1},
        {-1},
        {-1,-1},
        {25182,-1,-1}
    },
    {
        {25421,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25422,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25423,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25423,-1},
        {-1},
        {-1,-1},
        {25423,-1,-1}
    },
    {
        {25424,-1},
        {-1},
        {-1,-1},
        {25424,-1,-1}
    },
    {
        {25424,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25423,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25590,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25429,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25431,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25431,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25433,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22475,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25436,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26806,-1},
        {26806,-1,-1}
    },
    {
        {25445,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25447,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25448,-1},
        {-1},
        {-1,-1},
        {25448,-1,-1}
    },
    {
        {25448,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25449,-1},
        {-1},
        {-1,-1},
        {25449,-1,-1}
    },
    {
        {25449,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25451,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25452,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25454,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25454,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25827,-1},
        {25827,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25404,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25458,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25458,-1},
        {-1},
        {-1,-1},
        {25458,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25305,23731},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25696,32317,32207}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25462,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25463,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25466,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25467,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25468,-1},
        {-1},
        {-1,-1},
        {25404,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26874,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26866,-1,-1}
    },
    {
        {25472,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {13898,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25475,-1},
        {-1},
        {-1,-1},
        {25475,-1,-1}
    },
    {
        {25475,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25476,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26824,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25179,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25480,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25481,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25269,-1},
        {-1,-1,-1}
    },
    {
        {25484,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28151,-1},
        {28151,-1,-1}
    },
    {
        {25486,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25487,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26827,-1},
        {-1,-1,-1}
    },
    {
        {25490,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25490,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21460,-1},
        {-1,-1,-1}
    },
    {
        {25492,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25494,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26895,-1},
        {-1,-1,-1}
    },
    {
        {25496,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25497,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25499,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25423,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22495,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25503,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25504,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25505,-1},
        {-1},
        {-1,-1},
        {25505,-1,-1}
    },
    {
        {25505,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25506,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25507,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25509,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25511,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25511,-1},
        {-1},
        {-1,-1},
        {25511,-1,-1}
    },
    {
        {25512,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25513,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25513,-1},
        {-1},
        {-1,-1},
        {25513,-1,-1}
    },
    {
        {25514,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25515,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25516,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {15239,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23400,25163},
        {-1,-1,-1}
    },
    {
        {25522,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25524,-1},
        {-1},
        {-1,-1},
        {25681,-1,-1}
    },
    {
        {25525,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25531,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25533,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25534,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25534,-1},
        {-1},
        {-1,-1},
        {25534,-1,-1}
    },
    {
        {25534,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26953,-1},
        {-1,-1,-1}
    },
    {
        {25536,-1},
        {-1},
        {-1,-1},
        {26604,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26978,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25539,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25539,-1},
        {-1},
        {-1,-1},
        {25539,-1,-1}
    },
    {
        {25540,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25540,-1},
        {-1},
        {-1,-1},
        {25540,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25542,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26976,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25516,-1,-1}
    },
    {
        {25545,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25169,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22622,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22863,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25551,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25551,-1},
        {-1},
        {-1,-1},
        {25551,-1,-1}
    },
    {
        {25552,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25423,-1},
        {-1},
        {-1,-1},
        {25423,-1,-1}
    },
    {
        {25554,-1},
        {-1},
        {-1,-1},
        {25407,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25556,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25557,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25558,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25558,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24490,26991},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25562,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25563,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25564,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22779,-1},
        {-1,-1,-1}
    },
    {
        {25568,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25568,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25569,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21888,25356},
        {-1,-1,-1}
    },
    {
        {25571,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21918,-1},
        {-1,-1,-1}
    },
    {
        {25573,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25573,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28290,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25507,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25577,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25579,-1,-1}
    },
    {
        {25579,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25580,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25581,-1},
        {-1},
        {-1,-1},
        {25522,-1,-1}
    },
    {
        {25582,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20121,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22585,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26786,-1},
        {-1,-1,-1}
    },
    {
        {25586,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25588,-1},
        {-1},
        {-1,-1},
        {25588,-1,-1}
    },
    {
        {25588,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25589,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25589,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25590,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25591,-1},
        {-1},
        {-1,-1},
        {25554,25407,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25594,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25606,-1},
        {-1},
        {-1,-1},
        {25606,-1,-1}
    },
    {
        {25606,-1},
        {-1},
        {-1,-1},
        {25606,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27453,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27029,-1},
        {-1,-1,-1}
    },
    {
        {25609,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25184,25302},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23637,-1},
        {-1,-1,-1}
    },
    {
        {25613,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25613,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25615,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25615,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25616,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33592,25558},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25619,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25620,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27068,-1},
        {-1,-1,-1}
    },
    {
        {25622,-1},
        {-1},
        {-1,-1},
        {25594,-1,-1}
    },
    {
        {25623,-1},
        {-1},
        {-1,-1},
        {25827,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25289,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27087,-1},
        {-1,-1,-1}
    },
    {
        {25628,-1},
        {-1},
        {-1,-1},
        {25436,-1,-1}
    },
    {
        {25628,-1},
        {-1},
        {-1,-1},
        {25436,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28340,-1},
        {-1,-1,-1}
    },
    {
        {25630,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25472,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26705,-1},
        {-1,-1,-1}
    },
    {
        {25634,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25634,-1},
        {-1},
        {-1,-1},
        {25634,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28357,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25637,-1},
        {-1},
        {-1,-1},
        {25637,-1,-1}
    },
    {
        {25637,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25638,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25638,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20627,-1},
        {-1,-1,-1}
    },
    {
        {25640,-1},
        {-1},
        {-1,-1},
        {25640,-1,-1}
    },
    {
        {25640,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25640,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25641,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25641,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25644,-1},
        {25516,-1,-1}
    },
    {
        {25644,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25645,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25645,-1},
        {-1},
        {-1,-1},
        {25645,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24900,-1},
        {-1,-1,-1}
    },
    {
        {25647,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38601,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31370,25580},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30606,-1},
        {-1,-1,-1}
    },
    {
        {25652,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25654,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20280,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27035,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39730,25212},
        {-1,-1,-1}
    },
    {
        {25658,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25638,-1},
        {25638,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25331,-1},
        {25331,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25662,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25696,-1},
        {-1,-1,-1}
    },
    {
        {25666,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25666,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28411,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32302,-1},
        {-1,-1,-1}
    },
    {
        {25678,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25678,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25681,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25709,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25709,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20666,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27157,-1},
        {-1,-1,-1}
    },
    {
        {25688,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25688,-1},
        {-1},
        {-1,-1},
        {25688,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29833,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29833,-1},
        {-1,-1,-1}
    },
    {
        {25690,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25691,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25693,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27151,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25696,-1},
        {-1},
        {-1,-1},
        {32317,32207,-1}
    },
    {
        {25696,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25696,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25654,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22642,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29920,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29920,-1},
        {-1,-1,-1}
    },
    {
        {25703,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25705,-1},
        {-1},
        {-1,-1},
        {25705,-1,-1}
    },
    {
        {25705,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23558,23559},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27131,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22659,-1},
        {-1,-1,-1}
    },
    {
        {25709,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25709,-1},
        {-1},
        {-1,-1},
        {25709,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25970,-1},
        {-1,-1,-1}
    },
    {
        {25711,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25507,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24922,-1},
        {-1,-1,-1}
    },
    {
        {25715,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25715,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27117,-1},
        {-1,-1,-1}
    },
    {
        {25718,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37396,21119},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37396,21119},
        {-1,-1,-1}
    },
    {
        {25720,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25720,-1},
        {-1},
        {-1,-1},
        {25720,-1,-1}
    },
    {
        {25721,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25721,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25722,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25722,-1},
        {-1},
        {-1,-1},
        {25722,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29309,-1},
        {-1,-1,-1}
    },
    {
        {25725,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24374,-1},
        {24374,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25448,27292},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25731,30956},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29309,-1},
        {29309,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25504,-1},
        {25504,-1,-1}
    },
    {
        {25731,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25735,-1},
        {25735,-1,-1}
    },
    {
        {25735,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25736,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22698,25711},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20713,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25468,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21421,25558},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25745,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25746,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25747,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25004,-1},
        {-1,-1,-1}
    },
    {
        {25749,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27204,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25313,27178},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25645,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25754,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25327,-1},
        {25327,-1,-1}
    },
    {
        {25757,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25758,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25758,-1},
        {-1},
        {-1,-1},
        {25758,-1,-1}
    },
    {
        {25759,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25119,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25506,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25506,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25764,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25765,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32118,-1},
        {32118,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25769,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27275,-1},
        {-1,-1,-1}
    },
    {
        {25771,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25773,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25774,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30528,25619},
        {-1,-1,-1}
    },
    {
        {25776,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25776,-1},
        {-1},
        {-1,-1},
        {25776,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27234,-1},
        {-1,-1,-1}
    },
    {
        {25778,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22697,-1},
        {-1,-1,-1}
    },
    {
        {25785,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25785,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25787,-1},
        {-1},
        {-1,-1},
        {25787,-1,-1}
    },
    {
        {25787,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25788,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25790,-1},
        {-1},
        {-1,-1},
        {25790,-1,-1}
    },
    {
        {25790,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25791,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25182,-1},
        {-1,-1,-1}
    },
    {
        {25793,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25794,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25017,-1},
        {-1,-1,-1}
    },
    {
        {25796,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25796,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25797,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22712,-1},
        {-1,-1,-1}
    },
    {
        {25799,-1},
        {-1},
        {-1,-1},
        {25246,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25778,-1},
        {25778,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25802,-1},
        {-1},
        {-1,-1},
        {25731,-1,-1}
    },
    {
        {25803,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25804,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25805,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25806,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25806,-1},
        {-1},
        {-1,-1},
        {25806,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25806,20742},
        {25806,-1,-1}
    },
    {
        {25808,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25509,-1,-1}
    },
    {
        {25810,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25810,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25793,21295},
        {-1,-1,-1}
    },
    {
        {25812,-1},
        {-1},
        {-1,-1},
        {25285,-1,-1}
    },
    {
        {25813,-1},
        {-1},
        {-1,-1},
        {25658,25884,-1}
    },
    {
        {-1,-1},
        {-1},
        {33883,25581},
        {-1,-1,-1}
    },
    {
        {25815,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25816,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22714,-1},
        {-1,-1,-1}
    },
    {
        {25818,-1},
        {-1},
        {-1,-1},
        {25312,-1,-1}
    },
    {
        {25818,-1},
        {-1},
        {-1,-1},
        {25312,-1,-1}
    },
    {
        {25818,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32886,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29550,-1},
        {-1,-1,-1}
    },
    {
        {25824,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25825,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25826,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25826,-1},
        {-1},
        {-1,-1},
        {25826,-1,-1}
    },
    {
        {25827,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25828,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25828,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25829,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25830,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25831,-1},
        {-1},
        {-1,-1},
        {25369,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27476,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25836,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25130,-1},
        {-1,-1,-1}
    },
    {
        {25839,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25839,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25841,-1},
        {-1},
        {-1,-1},
        {38307,-1,-1}
    },
    {
        {25842,-1},
        {-1},
        {-1,-1},
        {25855,-1,-1}
    },
    {
        {25842,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25842,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25842,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27355,-1},
        {-1,-1,-1}
    },
    {
        {25844,-1},
        {-1},
        {-1,-1},
        {25313,-1,-1}
    },
    {
        {25844,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25705,-1,-1}
    },
    {
        {25846,-1},
        {-1},
        {-1,-1},
        {25846,-1,-1}
    },
    {
        {25846,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25847,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29557,29471},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25892,-1},
        {25892,-1,-1}
    },
    {
        {25850,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25851,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27347,-1},
        {-1,-1,-1}
    },
    {
        {25853,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {25853},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25854,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25855,-1},
        {-1},
        {-1,-1},
        {25855,-1,-1}
    },
    {
        {25855,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25855,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25856,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25794,-1},
        {25794,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25746,25830},
        {-1,-1,-1}
    },
    {
        {25860,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25861,-1},
        {-1},
        {-1,-1},
        {25890,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30910,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25411,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25609,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25853,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29562,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25681,22300},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28699,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25886,-1},
        {-1,-1,-1}
    },
    {
        {25871,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25871,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25652,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25876,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25876,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25878,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25880,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25884,-1},
        {-1},
        {-1,-1},
        {25658,-1,-1}
    },
    {
        {25885,-1},
        {-1},
        {-1,-1},
        {25666,-1,-1}
    },
    {
        {25885,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25885,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25885,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25886,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25411,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25705,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25691,-1},
        {-1,-1,-1}
    },
    {
        {25890,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25891,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25892,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25892,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25890,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31411,-1},
        {-1,-1,-1}
    },
    {
        {25897,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25898,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25899,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25900,-1},
        {-1},
        {-1,-1},
        {25829,-1,-1}
    },
    {
        {25900,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34849,25534},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25903,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25903,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25906,-1},
        {25906,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25906,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32709,-1},
        {-1,-1,-1}
    },
    {
        {25908,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25909,-1},
        {-1},
        {-1,-1},
        {25908,-1,-1}
    },
    {
        {25909,-1},
        {-1},
        {-1,-1},
        {25991,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25910,-1},
        {-1},
        {-1,-1},
        {21454,-1,-1}
    },
    {
        {25911,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25912,-1},
        {-1},
        {-1,-1},
        {25912,-1,-1}
    },
    {
        {25912,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25913,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25915,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25918,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25919,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25919,-1},
        {-1},
        {-1,-1},
        {25919,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36843,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25925,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21481,21475},
        {25187,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25924,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25925,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21512,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31574,-1,-1}
    },
    {
        {25928,-1},
        {-1},
        {-1,-1},
        {21177,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25910,21454},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25973,-1},
        {-1,-1,-1}
    },
    {
        {25933,-1},
        {-1},
        {-1,-1},
        {21465,-1,-1}
    },
    {
        {25945,-1},
        {-1},
        {-1,-1},
        {25945,-1,-1}
    },
    {
        {25935,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25937,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22890,23550},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25940,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25941,-1},
        {-1},
        {-1,-1},
        {21189,-1,-1}
    },
    {
        {25942,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25943,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25944,-1},
        {-1},
        {-1,-1},
        {21465,25933,-1}
    },
    {
        {25945,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25467,-1},
        {-1,-1,-1}
    },
    {
        {25949,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25949,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25949,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25950,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25950,-1},
        {-1},
        {-1,-1},
        {25949,25950,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25954,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25955,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25933,22821},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25958,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25959,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25989,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25963,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25964,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25964,-1},
        {-1},
        {-1,-1},
        {25964,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25562,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27459,25933},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25968,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25970,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25970,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25972,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25973,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25973,-1},
        {-1},
        {-1,-1},
        {25973,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38515,-1},
        {-1,-1,-1}
    },
    {
        {25975,-1},
        {-1},
        {-1,-1},
        {25975,-1,-1}
    },
    {
        {25975,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25976,-1},
        {-1},
        {-1,-1},
        {25968,-1,-1}
    },
    {
        {25976,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25978,-1},
        {-1},
        {-1,-1},
        {39493,39366,-1}
    },
    {
        {25978,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27478,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25985,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25985,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25986,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25987,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25987,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25987,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25988,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25989,-1},
        {-1},
        {-1,-1},
        {23416,23398,-1}
    },
    {
        {-1,-1},
        {-1},
        {25989,23398},
        {23416,25989,23398}
    },
    {
        {25991,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25991,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25992,-1},
        {-1},
        {-1,-1},
        {23416,23398,-1}
    },
    {
        {25992,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25993,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25994,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40779,25998},
        {-1,-1,-1}
    },
    {
        {25996,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35258,35226},
        {35258,35226,-1}
    },
    {
        {25998,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25998,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26000,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26000,-1},
        {-1},
        {-1,-1},
        {26000,-1,-1}
    },
    {
        {26000,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26001,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24254,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20153,-1},
        {20153,-1,-1}
    },
    {
        {26007,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26009,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26013,-1},
        {26013,-1,-1}
    },
    {
        {26011,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26012,-1},
        {-1},
        {-1,-1},
        {26012,-1,-1}
    },
    {
        {26012,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26013,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24254,-1},
        {-1,-1,-1}
    },
    {
        {26015,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26015,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26017,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26020,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26021,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26023,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26023,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26512,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26027,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26028,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26029,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26031,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26032,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26027,-1},
        {26027,-1,-1}
    },
    {
        {26034,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26034,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33465,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26037,-1},
        {-1},
        {-1,-1},
        {26034,-1,-1}
    },
    {
        {26037,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26039,-1},
        {-1},
        {-1,-1},
        {26029,-1,-1}
    },
    {
        {26039,-1},
        {-1},
        {-1,-1},
        {26029,-1,-1}
    },
    {
        {26040,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26041,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26044,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26044,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26045,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26054,-1},
        {26054,-1,-1}
    },
    {
        {26047,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26049,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26050,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26051,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26052,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26053,-1},
        {-1},
        {-1,-1},
        {26053,-1,-1}
    },
    {
        {26053,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26054,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30772,26063},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26066,-1,-1}
    },
    {
        {26057,-1},
        {-1},
        {-1,-1},
        {25975,-1,-1}
    },
    {
        {26057,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30745,-1},
        {-1,-1,-1}
    },
    {
        {26059,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26060,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26060,-1},
        {26060,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26063,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26066,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26786,-1},
        {-1,-1,-1}
    },
    {
        {26068,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26071,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24159,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24159,-1,-1}
    },
    {
        {26073,-1},
        {-1},
        {-1,-1},
        {26075,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39108,-1},
        {-1,-1,-1}
    },
    {
        {26075,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26051,-1},
        {26051,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20250,26371},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20250,26371},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26068,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26068,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26068,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26080,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26081,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26081,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26081,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26082,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26082,-1},
        {-1},
        {-1,-1},
        {26082,-1,-1}
    },
    {
        {26082,-1},
        {-1},
        {-1,-1},
        {26082,-1,-1}
    },
    {
        {26082,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {64082,31117},
        {64082,31117,-1}
    },
    {
        {26085,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26085,-1},
        {-1},
        {-1,-1},
        {26085,-1,-1}
    },
    {
        {26086,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26087,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26088,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26089,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21332,21494},
        {21332,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21480,-1},
        {-1,-1,-1}
    },
    {
        {26092,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26093,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26093,-1},
        {26093,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26097,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26098,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30340,-1},
        {30340,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30340,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26105,26178},
        {26178,-1,-1}
    },
    {
        {26105,-1},
        {-1},
        {-1,-1},
        {26178,-1,-1}
    },
    {
        {26106,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26107,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26107,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26108,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26302,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26149,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26212,-1,-1}
    },
    {
        {26112,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26112,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32954,-1},
        {-1,-1,-1}
    },
    {
        {26114,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26114,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26115,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26116,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21560,-1},
        {-1,-1,-1}
    },
    {
        {26118,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26119,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26120,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26120,-1},
        {-1},
        {-1,-1},
        {26120,-1,-1}
    },
    {
        {26120,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26121,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26122,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21534,-1},
        {39321,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21534,-1},
        {-1,-1,-1}
    },
    {
        {26124,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26126,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26126,-1},
        {-1},
        {-1,-1},
        {26126,-1,-1}
    },
    {
        {26127,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26131,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26132,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26133,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26115,-1},
        {26115,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26121,35370},
        {35370,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26311,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24892,24910},
        {24892,24910,-1}
    },
    {
        {-1,-1},
        {-1},
        {23712,-1},
        {-1,-1,-1}
    },
    {
        {26140,-1},
        {-1},
        {-1,-1},
        {26131,38525,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26142,-1},
        {-1},
        {-1,-1},
        {26170,-1,-1}
    },
    {
        {26142,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26142,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26143,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26144,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26145,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26146,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26148,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26148,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26149,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26122,-1,-1}
    },
    {
        {26151,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26152,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26154,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26155,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26127,-1,-1}
    },
    {
        {26157,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26159,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26160,-1},
        {-1},
        {-1,-1},
        {26159,-1,-1}
    },
    {
        {26161,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21640,-1},
        {-1,-1,-1}
    },
    {
        {26163,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26164,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26165,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26166,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26166,-1},
        {-1},
        {-1,-1},
        {26166,-1,-1}
    },
    {
        {151637,-1},
        {-1},
        {-1,-1},
        {151637,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20908,-1},
        {20908,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26170,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26114,-1},
        {-1},
        {-1,-1},
        {26114,-1,-1}
    },
    {
        {26172,-1},
        {-1},
        {-1,-1},
        {26172,-1,-1}
    },
    {
        {26172,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26175,-1},
        {-1},
        {-1,-1},
        {26336,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30522,-1,-1}
    },
    {
        {26177,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26178,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26179,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26180,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26181,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30566,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26184,-1},
        {-1},
        {-1,-1},
        {30350,-1,-1}
    },
    {
        {26184,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26184,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26185,-1},
        {-1},
        {-1,-1},
        {26187,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26187,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26188,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26312,-1},
        {26312,-1,-1}
    },
    {
        {26190,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26191,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21683,-1},
        {-1,-1,-1}
    },
    {
        {26193,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26194,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26199,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26097,-1,-1}
    },
    {
        {26201,-1},
        {-1},
        {-1,-1},
        {26201,-1,-1}
    },
    {
        {26201,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26202,-1},
        {-1},
        {-1,-1},
        {26217,-1,-1}
    },
    {
        {26203,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26205,-1},
        {-1},
        {-1,-1},
        {26172,-1,-1}
    },
    {
        {26206,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26207,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26207,-1},
        {-1},
        {-1,-1},
        {26207,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26207,-1,-1}
    },
    {
        {26209,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26210,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26211,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26212,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26213,-1},
        {-1},
        {-1,-1},
        {30358,-1,-1}
    },
    {
        {26214,-1},
        {-1},
        {-1,-1},
        {26214,-1,-1}
    },
    {
        {26214,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26215,-1},
        {-1},
        {-1,-1},
        {30355,-1,-1}
    },
    {
        {26215,-1},
        {-1},
        {-1,-1},
        {30355,26215,-1}
    },
    {
        {26216,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26217,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26218,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26219,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26220,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26221,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26222,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26223,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26224,-1},
        {-1},
        {-1,-1},
        {26227,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30568,-1},
        {-1,-1,-1}
    },
    {
        {26227,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26228,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26230,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26231,-1},
        {-1},
        {-1,-1},
        {26231,-1,-1}
    },
    {
        {26231,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26231,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26232,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26234,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26235,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26230,-1,-1}
    },
    {
        {26240,-1},
        {-1},
        {-1,-1},
        {26240,26106,-1}
    },
    {
        {26240,-1},
        {-1},
        {-1,-1},
        {26106,-1,-1}
    },
    {
        {26241,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26241,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26243,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26244,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26045,-1},
        {-1,-1,-1}
    },
    {
        {26247,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26248,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26249,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30889,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26253,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26254,-1},
        {-1},
        {-1,-1},
        {26144,-1,-1}
    },
    {
        {26254,-1},
        {-1},
        {-1,-1},
        {26144,-1,-1}
    },
    {
        {26254,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26329,-1},
        {26329,-1,-1}
    },
    {
        {26256,-1},
        {-1},
        {-1,-1},
        {26256,-1,-1}
    },
    {
        {26256,-1},
        {-1},
        {-1,-1},
        {26256,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26257,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26258,-1},
        {-1},
        {-1,-1},
        {26228,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30592,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21891,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {13734,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26262,-1},
        {-1},
        {-1,-1},
        {26262,-1,-1}
    },
    {
        {26262,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26263,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26263,-1},
        {-1},
        {-1,-1},
        {26263,-1,-1}
    },
    {
        {26264,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26265,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26367,-1,-1}
    },
    {
        {26269,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30366,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26272,-1},
        {-1},
        {-1,-1},
        {30364,-1,-1}
    },
    {
        {26272,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34697,-1},
        {-1,-1,-1}
    },
    {
        {26274,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27683,27671},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30366,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26253,26191},
        {-1,-1,-1}
    },
    {
        {26278,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26305,-1},
        {-1},
        {26305,23626},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38555,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30622,-1},
        {-1,-1,-1}
    },
    {
        {26283,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35131,-1},
        {-1,-1,-1}
    },
    {
        {26285,-1},
        {-1},
        {-1,-1},
        {30366,-1,-1}
    },
    {
        {26286,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26286,-1},
        {-1},
        {-1,-1},
        {26286,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33180,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36452,-1},
        {-1,-1,-1}
    },
    {
        {26289,-1},
        {-1},
        {-1,-1},
        {26165,-1,-1}
    },
    {
        {26289,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26289,-1},
        {-1},
        {-1,-1},
        {26289,26165,-1}
    },
    {
        {26290,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26292,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26293,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29831,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26296,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26297,-1},
        {-1},
        {-1,-1},
        {26297,-1,-1}
    },
    {
        {26297,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22077,-1},
        {-1,-1,-1}
    },
    {
        {26299,-1},
        {-1},
        {-1,-1},
        {26223,-1,-1}
    },
    {
        {26300,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26300,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26302,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26303,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26305,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26305,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26305,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26305,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26308,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26308,-1},
        {-1},
        {-1,-1},
        {26308,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26308,-1},
        {26308,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26308,-1},
        {-1,-1,-1}
    },
    {
        {26310,-1},
        {-1},
        {-1,-1},
        {26278,-1,-1}
    },
    {
        {26311,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26312,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26313,-1},
        {-1},
        {-1,-1},
        {26241,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36027,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30969,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29031,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26143,-1},
        {26143,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21442,21443},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30374,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26292,-1},
        {26292,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26326,-1},
        {-1},
        {-1,-1},
        {26326,-1,-1}
    },
    {
        {26326,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26308,-1},
        {26308,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26329,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26329,-1},
        {-1},
        {-1,-1},
        {26329,-1,-1}
    },
    {
        {26330,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26330,-1},
        {-1},
        {-1,-1},
        {26330,-1,-1}
    },
    {
        {26330,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26331,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26332,-1},
        {-1},
        {-1,-1},
        {26332,-1,-1}
    },
    {
        {26332,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26333,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31018,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26216,-1,-1}
    },
    {
        {26336,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26336,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23476,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33243,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26342,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34701,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26344,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26344,-1},
        {-1},
        {-1,-1},
        {26344,-1,-1}
    },
    {
        {26345,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22217,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26348,-1},
        {-1},
        {-1,-1},
        {26194,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30682,-1},
        {-1,-1,-1}
    },
    {
        {26352,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30002,-1},
        {-1,-1,-1}
    },
    {
        {26354,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26355,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26356,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26356,-1},
        {-1},
        {-1,-1},
        {26356,-1,-1}
    },
    {
        {26357,-1},
        {-1},
        {-1,-1},
        {26355,-1,-1}
    },
    {
        {26357,-1},
        {-1},
        {-1,-1},
        {26355,-1,-1}
    },
    {
        {26357,-1},
        {-1},
        {-1,-1},
        {26355,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24573,-1},
        {-1,-1,-1}
    },
    {
        {26359,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26359,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26360,-1},
        {-1},
        {-1,-1},
        {26172,26205,-1}
    },
    {
        {26360,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26361,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26362,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26363,-1},
        {-1},
        {-1,-1},
        {26119,-1,-1}
    },
    {
        {26363,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26363,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26364,-1},
        {-1},
        {-1,-1},
        {26364,-1,-1}
    },
    {
        {26364,-1},
        {-1},
        {-1,-1},
        {26364,-1,-1}
    },
    {
        {26365,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26366,-1},
        {-1},
        {-1,-1},
        {26365,-1,-1}
    },
    {
        {26366,-1},
        {-1},
        {-1,-1},
        {26365,26366,-1}
    },
    {
        {26366,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26367,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26368,-1},
        {-1},
        {-1,-1},
        {26368,-1,-1}
    },
    {
        {26368,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26369,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26369,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21206,-1},
        {21206,-1,-1}
    },
    {
        {26371,-1},
        {-1},
        {-1,-1},
        {20250,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20250,26371},
        {26371,-1,-1}
    },
    {
        {26371,-1},
        {-1},
        {-1,-1},
        {26371,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25536,-1},
        {-1,-1,-1}
    },
    {
        {26373,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29969,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35048,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26376,-1},
        {-1},
        {-1,-1},
        {26376,-1,-1}
    },
    {
        {26376,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26377,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26377,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33048,-1},
        {-1,-1,-1}
    },
    {
        {26379,-1},
        {-1},
        {-1,-1},
        {26379,-1,-1}
    },
    {
        {26379,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38930,-1},
        {38930,-1,-1}
    },
    {
        {26381,-1},
        {-1},
        {-1,-1},
        {26381,-1,-1}
    },
    {
        {26381,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26383,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32970,-1},
        {-1},
        {32970,-1},
        {32970,-1,-1}
    },
    {
        {26386,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26387,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26387,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26388,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26388,-1},
        {-1},
        {-1,-1},
        {26388,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26389,-1},
        {-1},
        {-1,-1},
        {26389,-1,-1}
    },
    {
        {26389,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26390,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26391,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26391,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26393,-1},
        {-1},
        {-1,-1},
        {26126,-1,-1}
    },
    {
        {26393,-1},
        {-1},
        {-1,-1},
        {26126,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24537,30450},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24537,30450},
        {-1,-1,-1}
    },
    {
        {26395,-1},
        {-1},
        {-1,-1},
        {26395,-1,-1}
    },
    {
        {26395,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26395,-1},
        {-1},
        {-1,-1},
        {26395,-1,-1}
    },
    {
        {26395,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26395,-1},
        {-1},
        {-1,-1},
        {26395,-1,-1}
    },
    {
        {26395,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26397,-1},
        {-1},
        {-1,-1},
        {26397,-1,-1}
    },
    {
        {26397,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26398,-1},
        {-1},
        {-1,-1},
        {26399,-1,-1}
    },
    {
        {26398,-1},
        {-1},
        {-1,-1},
        {26399,-1,-1}
    },
    {
        {26398,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26399,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26399,-1},
        {-1},
        {-1,-1},
        {26399,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33063,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33063,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26395,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30643,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26391,-1},
        {-1,-1,-1}
    },
    {
        {26406,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26406,-1},
        {-1},
        {-1,-1},
        {26406,-1,-1}
    },
    {
        {26406,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26407,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26408,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26410,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26411,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26412,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26413,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26414,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26494,-1,-1}
    },
    {
        {26417,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26433,-1},
        {-1,-1,-1}
    },
    {
        {26419,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26420,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {17587,26422},
        {26422,-1,-1}
    },
    {
        {26422,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26423,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26424,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {144338,31755},
        {-1,-1,-1}
    },
    {
        {26426,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27163,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26503,-1},
        {26503,-1,-1}
    },
    {
        {26429,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26431,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26433,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27177,27402},
        {27177,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26438,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26438,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26439,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26440,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26440,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26441,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26426,22323},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26963,-1},
        {-1,-1,-1}
    },
    {
        {26444,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26445,-1},
        {-1},
        {-1,-1},
        {26446,-1,-1}
    },
    {
        {26446,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26447,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26448,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26449,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36564,-1},
        {-1,-1,-1}
    },
    {
        {26451,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26451,-1},
        {-1},
        {-1,-1},
        {26451,-1,-1}
    },
    {
        {26452,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26454,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26454,-1},
        {-1},
        {-1,-1},
        {26454,-1,-1}
    },
    {
        {26454,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22916,21574},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22916,21574},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26457,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24532,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24339,26439},
        {-1,-1,-1}
    },
    {
        {26460,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26461,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26462,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26462,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26462,-1},
        {-1},
        {26487,-1},
        {-1,-1,-1}
    },
    {
        {26463,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26464,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26465,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26466,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26467,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26468,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26469,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26470,-1},
        {-1},
        {-1,-1},
        {26441,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26474,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26622,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26477,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26623,-1},
        {-1},
        {-1,-1},
        {144341,-1,-1}
    },
    {
        {26479,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26480,-1},
        {-1},
        {-1,-1},
        {20625,-1,-1}
    },
    {
        {26481,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26482,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26483,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26484,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26485,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26486,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26487,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27571,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26503,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26491,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26492,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26491,37397},
        {26491,-1,-1}
    },
    {
        {26494,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26494,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26495,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26494,-1},
        {26494,-1,-1}
    },
    {
        {26497,-1},
        {-1},
        {-1,-1},
        {26497,-1,-1}
    },
    {
        {26497,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30456,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26451,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27083,-1,-1}
    },
    {
        {26501,-1},
        {-1},
        {-1,-1},
        {26501,-1,-1}
    },
    {
        {26501,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26501,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26969,-1},
        {-1,-1,-1}
    },
    {
        {26503,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26503,-1},
        {-1,-1,-1}
    },
    {
        {26505,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26114,-1},
        {-1,-1,-1}
    },
    {
        {26507,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26508,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26508,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26439,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26512,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26514,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26514,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26515,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26563,-1},
        {-1,-1,-1}
    },
    {
        {26517,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26519,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26520,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26522,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26523,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26524,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26525,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27141,-1},
        {27141,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26528,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26529,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26530,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26534,-1},
        {-1},
        {-1,-1},
        {27368,-1,-1}
    },
    {
        {26534,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26537,-1},
        {-1},
        {-1,-1},
        {26494,-1,-1}
    },
    {
        {26537,-1},
        {-1},
        {-1,-1},
        {26494,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22349,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26937,-1},
        {-1,-1,-1}
    },
    {
        {26543,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26544,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26546,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26547,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26548,-1},
        {-1},
        {-1,-1},
        {26548,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26550,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26551,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26552,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26553,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26553,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31203,-1},
        {-1,-1,-1}
    },
    {
        {26555,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26705,26706,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26617,26623,-1}
    },
    {
        {-1,-1},
        {-1},
        {34326,-1},
        {-1,-1,-1}
    },
    {
        {26560,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26561,-1},
        {-1},
        {-1,-1},
        {33333,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26561,33333,-1}
    },
    {
        {26563,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26564,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26564,-1},
        {-1},
        {-1,-1},
        {26564,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26566,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26568,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26570,-1},
        {-1},
        {-1,-1},
        {26570,-1,-1}
    },
    {
        {26570,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24082,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31207,-1},
        {-1,-1,-1}
    },
    {
        {26574,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26575,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26576,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26577,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26578,-1},
        {-1},
        {-1,-1},
        {28422,-1,-1}
    },
    {
        {26579,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26580,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26580,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26969,-1},
        {26969,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25307,-1},
        {-1,-1,-1}
    },
    {
        {26583,-1},
        {-1},
        {-1,-1},
        {26494,-1,-1}
    },
    {
        {26584,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26585,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26586,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26587,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26589,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26590,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26976,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27320,26990},
        {26990,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26594,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26596,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26619,-1},
        {-1},
        {26619,-1},
        {26619,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26599,-1},
        {-1},
        {-1,-1},
        {26599,-1,-1}
    },
    {
        {26599,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24598,-1},
        {-1,-1,-1}
    },
    {
        {26601,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25303,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25282,25173},
        {-1,-1,-1}
    },
    {
        {26604,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31201,-1},
        {-1,-1,-1}
    },
    {
        {26606,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26607,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26608,-1},
        {-1},
        {-1,-1},
        {22856,-1,-1}
    },
    {
        {26609,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26609,-1},
        {-1},
        {-1,-1},
        {26609,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26611,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26611,-1},
        {-1},
        {-1,-1},
        {26611,-1,-1}
    },
    {
        {26612,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26613,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26615,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26479,-1},
        {-1,-1,-1}
    },
    {
        {26617,-1},
        {-1},
        {-1,-1},
        {26623,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26548,-1},
        {-1},
        {26548,-1},
        {-1,-1,-1}
    },
    {
        {26548,-1},
        {-1},
        {26548,-1},
        {-1,-1,-1}
    },
    {
        {26619,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26620,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27273,-1},
        {-1,-1,-1}
    },
    {
        {26622,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26623,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26772,25903},
        {26772,26693,-1}
    },
    {
        {26625,-1},
        {-1},
        {-1,-1},
        {26611,-1,-1}
    },
    {
        {26625,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26626,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26627,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26627,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26628,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26613,-1},
        {-1},
        {31604,26613},
        {26613,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26640,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26643,-1},
        {-1},
        {-1,-1},
        {26643,-1,-1}
    },
    {
        {26643,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26646,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26646,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26647,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26648,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21682,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21332,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26973,-1},
        {-1,-1,-1}
    },
    {
        {26653,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26653,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26654,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26655,-1},
        {-1},
        {-1,-1},
        {26655,-1,-1}
    },
    {
        {26655,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20973,-1},
        {-1,-1,-1}
    },
    {
        {26657,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26657,-1},
        {-1},
        {-1,-1},
        {26657,-1,-1}
    },
    {
        {26658,-1},
        {-1},
        {-1,-1},
        {26575,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34997,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26640,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31906,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31906,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27954,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26555,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26665,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26665,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26666,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26667,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26668,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26669,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26669,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31567,-1,-1}
    },
    {
        {26673,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26674,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26674,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26675,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26676,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26676,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20363,15213},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26801,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27096,27171},
        {-1,-1,-1}
    },
    {
        {26680,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26680,-1},
        {-1},
        {-1,-1},
        {26680,-1,-1}
    },
    {
        {26681,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25351,-1},
        {-1,-1,-1}
    },
    {
        {26683,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26684,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26685,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26686,-1},
        {-1},
        {-1,-1},
        {27410,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26688,-1},
        {-1},
        {-1,-1},
        {26688,-1,-1}
    },
    {
        {26688,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26688,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26689,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26690,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26691,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26692,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26693,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26693,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26694,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26550,-1},
        {-1,-1,-1}
    },
    {
        {26696,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26697,-1},
        {-1},
        {-1,-1},
        {26696,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27935,-1},
        {-1,-1,-1}
    },
    {
        {26700,-1},
        {-1},
        {-1,-1},
        {21331,-1,-1}
    },
    {
        {26701,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26702,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26704,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26705,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26706,-1},
        {-1},
        {-1,-1},
        {26705,26706,-1}
    },
    {
        {26706,-1},
        {-1},
        {-1,-1},
        {26706,-1,-1}
    },
    {
        {26706,-1},
        {-1},
        {-1,-1},
        {26705,26706,-1}
    },
    {
        {26706,-1},
        {-1},
        {-1,-1},
        {26705,-1,-1}
    },
    {
        {26707,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26708,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26709,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24676,-1},
        {-1,-1,-1}
    },
    {
        {26711,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26894,-1,-1}
    },
    {
        {26713,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26715,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26716,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26717,-1},
        {-1},
        {-1,-1},
        {26529,-1,-1}
    },
    {
        {26717,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26717,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26719,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26723,-1},
        {-1},
        {-1,-1},
        {27284,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26727,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26731,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26731,-1},
        {26731,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26734,-1},
        {-1},
        {-1,-1},
        {26479,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26738,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26713,-1},
        {-1,-1,-1}
    },
    {
        {26740,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26740,-1},
        {-1},
        {-1,-1},
        {26740,-1,-1}
    },
    {
        {26741,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26741,-1},
        {-1},
        {-1,-1},
        {26741,-1,-1}
    },
    {
        {26742,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26743,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29563,29486},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26746,-1},
        {-1},
        {-1,-1},
        {26611,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26748,-1},
        {-1},
        {-1,-1},
        {28422,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33693,-1},
        {-1,-1,-1}
    },
    {
        {26750,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26751,-1},
        {-1},
        {-1,-1},
        {26438,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26753,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26754,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26755,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26755,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26757,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25397,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22767,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37942,37591},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26765,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26765,-1},
        {-1},
        {26765,-1},
        {-1,-1,-1}
    },
    {
        {26767,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26767,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26768,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27000,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25342,-1},
        {-1,-1,-1}
    },
    {
        {26771,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26772,-1},
        {-1},
        {-1,-1},
        {26693,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28042,26980},
        {-1,-1,-1}
    },
    {
        {26774,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26775,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26775,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26776,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26777,-1},
        {-1},
        {-1,-1},
        {27125,-1,-1}
    },
    {
        {26778,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26779,-1},
        {-1},
        {-1,-1},
        {26779,-1,-1}
    },
    {
        {26779,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26780,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26781,-1},
        {-1},
        {-1,-1},
        {26465,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36349,-1},
        {-1,-1,-1}
    },
    {
        {26783,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26783,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26784,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26786,-1},
        {-1},
        {-1,-1},
        {26786,-1,-1}
    },
    {
        {26786,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26787,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26875,-1},
        {-1,-1,-1}
    },
    {
        {26789,-1},
        {-1},
        {-1,-1},
        {26494,-1,-1}
    },
    {
        {26789,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26790,-1},
        {-1},
        {-1,-1},
        {22818,26790,-1}
    },
    {
        {26790,-1},
        {-1},
        {-1,-1},
        {22818,-1,-1}
    },
    {
        {26791,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26792,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26794,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26795,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26797,-1},
        {-1},
        {-1,-1},
        {26797,-1,-1}
    },
    {
        {26797,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26799,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26800,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26801,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26802,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26803,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26804,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26805,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26805,-1},
        {-1},
        {-1,-1},
        {26805,-1,-1}
    },
    {
        {26806,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26856,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26819,-1},
        {26792,26819,-1}
    },
    {
        {26809,-1},
        {-1},
        {-1,-1},
        {27315,-1,-1}
    },
    {
        {26810,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26811,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26812,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26802,-1},
        {-1},
        {26802,-1},
        {-1,-1,-1}
    },
    {
        {26819,-1},
        {-1},
        {-1,-1},
        {26792,-1,-1}
    },
    {
        {26820,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26820,-1},
        {-1},
        {-1,-1},
        {26820,-1,-1}
    },
    {
        {26821,-1},
        {-1},
        {-1,-1},
        {26564,-1,-1}
    },
    {
        {26822,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26980,-1},
        {-1,-1,-1}
    },
    {
        {26824,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26824,-1},
        {-1},
        {-1,-1},
        {26824,-1,-1}
    },
    {
        {26825,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26826,-1},
        {-1},
        {-1,-1},
        {26827,-1,-1}
    },
    {
        {26827,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26828,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26829,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25506,-1},
        {-1,-1,-1}
    },
    {
        {26831,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26831,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26832,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26834,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26836,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26837,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26838,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26839,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26840,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26841,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26842,-1},
        {-1},
        {-1,-1},
        {26842,-1,-1}
    },
    {
        {26842,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26824,32946},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31259,-1},
        {26801,-1,-1}
    },
    {
        {26847,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26848,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26849,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36638,-1},
        {36638,-1,-1}
    },
    {
        {26851,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26852,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26853,-1},
        {-1},
        {-1,-1},
        {27146,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25497,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26855,-1},
        {-1},
        {-1,-1},
        {26719,-1,-1}
    },
    {
        {26856,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28149,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26860,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26861,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26862,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26863,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26864,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26865,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26866,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26381,26933},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26871,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23094,-1},
        {-1,-1,-1}
    },
    {
        {26873,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26874,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26875,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26875,-1},
        {-1},
        {-1,-1},
        {26875,-1,-1}
    },
    {
        {26876,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26805,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36631,-1},
        {-1,-1,-1}
    },
    {
        {26880,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26881,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31103,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26883,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26884,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26885,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26887,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26887,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26888,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20056,20055,-1}
    },
    {
        {26890,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26891,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26892,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26892,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26893,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26894,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26895,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26898,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30014,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30014,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26969,26508},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30855,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26970,-1},
        {26970,-1,-1}
    },
    {
        {26905,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26906,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26907,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26907,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26907,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26908,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26913,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26914,-1},
        {-1},
        {-1,-1},
        {27126,-1,-1}
    },
    {
        {26915,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26917,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26918,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26918,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26920,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26922,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25571,-1},
        {-1,-1,-1}
    },
    {
        {26928,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26928,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27022,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26932,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26933,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26934,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32216,20989},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26937,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26937,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35091,-1},
        {-1,-1,-1}
    },
    {
        {26939,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26689,34893},
        {-1,-1,-1}
    },
    {
        {26941,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26941,-1},
        {-1},
        {-1,-1},
        {26941,-1,-1}
    },
    {
        {26941,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31627,25539},
        {31627,-1,-1}
    },
    {
        {26943,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23458,26684},
        {-1,-1,-1}
    },
    {
        {26946,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24132,-1},
        {24132,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26950,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26950,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32997,-1},
        {-1,-1,-1}
    },
    {
        {26953,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26953,-1},
        {-1},
        {-1,-1},
        {26953,-1,-1}
    },
    {
        {26954,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28290,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20560,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26412,-1},
        {26412,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25542,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24605,-1},
        {-1,-1,-1}
    },
    {
        {26963,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26964,-1},
        {-1},
        {-1,-1},
        {26964,-1,-1}
    },
    {
        {26964,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26964,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26965,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27355,-1},
        {27355,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27355,-1},
        {27355,-1,-1}
    },
    {
        {26967,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26967,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26580,-1},
        {-1,-1,-1}
    },
    {
        {26969,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26970,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26972,-1},
        {-1},
        {-1,-1},
        {32993,-1,-1}
    },
    {
        {26973,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26974,-1},
        {-1},
        {-1,-1},
        {26974,-1,-1}
    },
    {
        {26974,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26881,-1},
        {-1,-1,-1}
    },
    {
        {26976,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26977,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26978,-1},
        {-1},
        {-1,-1},
        {26978,-1,-1}
    },
    {
        {26978,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26978,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26979,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26980,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20544,-1},
        {-1,-1,-1}
    },
    {
        {26984,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26985,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26986,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26987,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26987,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26988,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26989,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26990,-1},
        {-1},
        {-1,-1},
        {26990,-1,-1}
    },
    {
        {26990,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26991,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26991,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26994,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26995,-1},
        {-1},
        {-1,-1},
        {64068,26757,-1}
    },
    {
        {26996,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26997,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31906,15268},
        {-1,-1,-1}
    },
    {
        {26999,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27000,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27001,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27002,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27004,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27005,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27006,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27007,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27008,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27009,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27010,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27010,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27058,-1},
        {-1},
        {27058,27170},
        {27058,-1,-1}
    },
    {
        {26977,-1},
        {-1},
        {26977,-1},
        {26977,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27018,-1},
        {-1},
        {-1,-1},
        {27018,-1,-1}
    },
    {
        {27018,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27022,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33367,-1},
        {33367,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30430,30910},
        {-1,-1,-1}
    },
    {
        {27025,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27025,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27026,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27026,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27329,27154},
        {27154,-1,-1}
    },
    {
        {27028,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27028,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27029,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27030,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25634,-1},
        {-1,-1,-1}
    },
    {
        {27032,-1},
        {-1},
        {-1,-1},
        {30697,-1,-1}
    },
    {
        {27032,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25645,-1},
        {-1,-1,-1}
    },
    {
        {27035,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27036,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27036,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {144788,21407},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26771,-1},
        {26771,-1,-1}
    },
    {
        {27040,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26550,-1},
        {26550,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25622,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25622,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26688,-1},
        {-1,-1,-1}
    },
    {
        {27045,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27046,-1},
        {-1},
        {-1,-1},
        {24185,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27047,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27047,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34388,27165},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31137,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27053,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27054,-1},
        {-1},
        {-1,-1},
        {26628,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27056,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27057,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27058,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26755,24237},
        {-1,-1,-1}
    },
    {
        {27060,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33592,26987},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27063,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26883,-1},
        {-1,-1,-1}
    },
    {
        {27066,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27066,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27066,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27067,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27067,-1},
        {-1},
        {-1,-1},
        {27067,-1,-1}
    },
    {
        {27067,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27068,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24495,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27070,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27071,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27072,-1},
        {-1},
        {-1,-1},
        {27073,-1,-1}
    },
    {
        {27073,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27073,-1},
        {-1},
        {-1,-1},
        {144954,27211,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27075,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31291,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26486,-1},
        {-1,-1,-1}
    },
    {
        {27079,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27079,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32808,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27082,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27083,-1},
        {-1},
        {-1,-1},
        {27083,-1,-1}
    },
    {
        {27083,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27084,-1},
        {-1},
        {-1,-1},
        {27084,-1,-1}
    },
    {
        {27084,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27085,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27086,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27087,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27087,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27087,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27088,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27089,-1},
        {-1},
        {-1,-1},
        {26576,64068,26757}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27091,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27129,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26873,26700},
        {21331,-1,-1}
    },
    {
        {27094,-1},
        {-1},
        {-1,-1},
        {27216,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {144954,27211,-1}
    },
    {
        {27096,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27097,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27079,-1},
        {-1},
        {-1,-1},
        {27097,27079,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27101,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27102,-1},
        {-1},
        {-1,-1},
        {27379,-1,-1}
    },
    {
        {27106,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27106,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27107,-1},
        {-1},
        {-1,-1},
        {25959,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36899,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36899,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24949,-1},
        {-1,-1,-1}
    },
    {
        {27111,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27112,-1},
        {-1},
        {-1,-1},
        {26881,-1,-1}
    },
    {
        {27113,-1},
        {-1},
        {-1,-1},
        {27010,-1,-1}
    },
    {
        {27113,-1},
        {-1},
        {-1,-1},
        {27113,27010,-1}
    },
    {
        {27113,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27010,-1},
        {-1},
        {-1,-1},
        {27010,-1,-1}
    },
    {
        {27010,-1},
        {-1},
        {-1,-1},
        {27010,-1,-1}
    },
    {
        {27114,-1},
        {-1},
        {-1,-1},
        {27010,-1,-1}
    },
    {
        {27115,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27115,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29920,26599},
        {-1,-1,-1}
    },
    {
        {27117,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27118,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27119,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27122,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27123,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27124,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27125,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27126,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26964,33260},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35129,-1},
        {-1,-1,-1}
    },
    {
        {27129,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27131,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {16531,35215},
        {35215,-1,-1}
    },
    {
        {27133,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27134,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27134,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27135,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27135,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27135,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27138,-1},
        {-1},
        {-1,-1},
        {27005,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38515,-1},
        {-1,-1,-1}
    },
    {
        {27141,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38626,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33065,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22659,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26780,29245},
        {-1,-1,-1}
    },
    {
        {27146,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27147,-1},
        {-1},
        {-1,-1},
        {27147,-1,-1}
    },
    {
        {27147,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27148,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31309,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32302,-1},
        {-1,-1,-1}
    },
    {
        {27151,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33386,27347},
        {27347,-1,-1}
    },
    {
        {27153,-1},
        {-1},
        {-1,-1},
        {26753,-1,-1}
    },
    {
        {27154,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27155,-1},
        {-1},
        {-1,-1},
        {27004,-1,-1}
    },
    {
        {27156,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27156,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27157,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32725,26607},
        {-1,-1,-1}
    },
    {
        {27159,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26848,27222},
        {-1,-1,-1}
    },
    {
        {27161,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27162,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27163,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27163,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26584,-1},
        {26584,-1,-1}
    },
    {
        {27165,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27166,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27166,-1},
        {-1},
        {-1,-1},
        {26530,-1,-1}
    },
    {
        {27167,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27167,-1},
        {-1},
        {-1,-1},
        {27167,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27169,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27169,-1},
        {-1},
        {-1,-1},
        {27169,-1,-1}
    },
    {
        {27170,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27171,-1},
        {-1},
        {-1,-1},
        {27096,-1,-1}
    },
    {
        {27171,-1},
        {-1},
        {-1,-1},
        {27096,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23791,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26609,31513},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27177,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27178,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27178,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27179,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25696,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27182,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27184,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27186,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27187,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27188,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27188,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27189,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27189,-1,-1}
    },
    {
        {27190,-1},
        {-1},
        {-1,-1},
        {27190,-1,-1}
    },
    {
        {27190,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21474,-1,-1}
    },
    {
        {27192,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27193,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27193,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27194,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27194,-1},
        {-1},
        {-1,-1},
        {27194,-1,-1}
    },
    {
        {27195,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27035,-1},
        {27035,-1,-1}
    },
    {
        {27197,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27197,-1},
        {-1},
        {-1,-1},
        {27197,-1,-1}
    },
    {
        {27198,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27199,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27068,37295},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26486,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22880,27197},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31567,-1},
        {31567,-1,-1}
    },
    {
        {27204,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27204,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27205,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27206,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27207,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27208,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27209,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27209,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27060,-1,-1}
    },
    {
        {27211,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27258,-1},
        {-1},
        {27258,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27258,38287},
        {-1,-1,-1}
    },
    {
        {27214,-1},
        {-1},
        {-1,-1},
        {25773,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20688,-1},
        {-1,-1,-1}
    },
    {
        {27216,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27218,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34147,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22697,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27222,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26842,33804},
        {-1,-1,-1}
    },
    {
        {27224,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27225,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27225,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27000,26786},
        {-1,-1,-1}
    },
    {
        {27227,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27227,21413},
        {27227,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35203,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35203,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27231,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27231,-1},
        {-1},
        {-1,-1},
        {27231,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27233,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27233,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27234,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26742,29999},
        {-1,-1,-1}
    },
    {
        {27236,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29482,-1},
        {-1,-1,-1}
    },
    {
        {27238,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27238,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24996,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25736,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27243,-1},
        {-1},
        {-1,-1},
        {27178,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26787,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26543,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26611,-1},
        {26611,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25736,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27129,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21416,24282,-1}
    },
    {
        {27250,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27251,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27251,-1},
        {-1},
        {-1,-1},
        {27251,-1,-1}
    },
    {
        {27251,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27256,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27258,-1},
        {-1},
        {-1,-1},
        {27258,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25787,-1},
        {-1,-1,-1}
    },
    {
        {27262,-1},
        {-1},
        {-1,-1},
        {25805,-1,-1}
    },
    {
        {27263,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27264,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27264,-1},
        {-1},
        {-1,-1},
        {27264,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31328,-1},
        {-1,-1,-1}
    },
    {
        {27267,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27268,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26441,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25808,-1},
        {-1,-1,-1}
    },
    {
        {27273,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27273,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26438,-1},
        {26438,-1,-1}
    },
    {
        {27275,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32618,-1},
        {-1,-1,-1}
    },
    {
        {27277,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27277,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27278,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27192,-1},
        {27192,-1,-1}
    },
    {
        {27280,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27280,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27281,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39080,26963},
        {39080,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27584,26932},
        {-1,-1,-1}
    },
    {
        {27284,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32363,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36995,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36995,-1},
        {-1,-1,-1}
    },
    {
        {27287,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26544,27287},
        {27287,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26970,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27292,-1},
        {-1},
        {-1,-1},
        {27292,-1,-1}
    },
    {
        {27292,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27292,-1},
        {-1},
        {-1,-1},
        {27292,-1,-1}
    },
    {
        {27292,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27293,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27294,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27295,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27296,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27296,-1},
        {-1},
        {-1,-1},
        {27296,-1,-1}
    },
    {
        {27297,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27298,-1},
        {-1},
        {-1,-1},
        {26908,-1,-1}
    },
    {
        {27299,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36947,27147},
        {-1,-1,-1}
    },
    {
        {27301,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27161,21117},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26980,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27171,-1},
        {-1,-1,-1}
    },
    {
        {27306,-1},
        {-1},
        {-1,-1},
        {27359,-1,-1}
    },
    {
        {27307,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27308,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27308,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27308,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27308,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37504,-1},
        {-1,-1,-1}
    },
    {
        {27310,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26827,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22296,26801},
        {-1,-1,-1}
    },
    {
        {27315,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27315,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27315,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26462,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26484,-1},
        {26484,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37480,-1},
        {-1,-1,-1}
    },
    {
        {27320,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27320,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27107,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35946,27308},
        {-1,-1,-1}
    },
    {
        {27323,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27267,27389},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27325,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26827,-1},
        {26826,26827,-1}
    },
    {
        {27329,-1},
        {-1},
        {-1,-1},
        {27154,-1,-1}
    },
    {
        {27330,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27330,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27331,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26486,26943},
        {26486,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31335,-1},
        {-1,-1,-1}
    },
    {
        {27334,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38935,25847},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36074,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34913,27178},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34913,27178},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36983,-1},
        {-1,-1,-1}
    },
    {
        {27344,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27345,-1},
        {-1},
        {-1,-1},
        {32589,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27307,-1},
        {-1,-1,-1}
    },
    {
        {27347,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27348,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27348,-1},
        {-1},
        {-1,-1},
        {27348,-1,-1}
    },
    {
        {27348,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27409,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35178,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27354,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27355,-1},
        {-1},
        {-1,-1},
        {27355,-1,-1}
    },
    {
        {27355,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27355,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27355,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27356,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27356,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27357,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27357,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27358,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27358,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27358,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27358,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27359,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25765,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31672,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25851,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34030,-1},
        {-1,-1,-1}
    },
    {
        {27364,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27364,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27367,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27368,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27280,-1},
        {27280,-1,-1}
    },
    {
        {27370,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27370,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35948,-1},
        {-1,-1,-1}
    },
    {
        {27372,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26988,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33852,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31308,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27377,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27378,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27379,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27379,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25078,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28703,-1},
        {-1,-1,-1}
    },
    {
        {27386,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27387,-1},
        {-1},
        {-1,-1},
        {26716,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27389,-1},
        {-1},
        {-1,-1},
        {27267,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26586,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31344,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36786,-1},
        {36786,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27396,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27397,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25666,25885,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25885,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21474,-1,-1}
    },
    {
        {27402,-1},
        {-1},
        {-1,-1},
        {27177,-1,-1}
    },
    {
        {27402,-1},
        {-1},
        {-1,-1},
        {27177,-1,-1}
    },
    {
        {27403,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27407,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27408,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27409,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27410,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25897,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25899,-1},
        {-1,-1,-1}
    },
    {
        {27414,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27415,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27415,-1},
        {-1},
        {-1,-1},
        {27396,63773,27415}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26040,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32306,27151},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40599,-1},
        {-1,-1,-1}
    },
    {
        {27419,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27421,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27421,-1},
        {-1},
        {-1,-1},
        {39729,-1,-1}
    },
    {
        {27422,-1},
        {-1},
        {-1,-1},
        {27386,-1,-1}
    },
    {
        {27423,-1},
        {-1},
        {-1,-1},
        {27423,-1,-1}
    },
    {
        {27423,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27424,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27425,-1},
        {-1},
        {-1,-1},
        {27425,-1,-1}
    },
    {
        {27425,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27427,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27487,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21561,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22350,-1},
        {-1,-1,-1}
    },
    {
        {27431,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21620,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21165,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21669,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27436,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27436,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27431,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25496,36470},
        {-1,-1,-1}
    },
    {
        {27439,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21561,34882},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37059,30733},
        {-1,-1,-1}
    },
    {
        {27442,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21919,-1},
        {21919,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39154,-1},
        {-1,-1,-1}
    },
    {
        {27445,-1},
        {-1},
        {-1,-1},
        {27454,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27447,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27448,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27449,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27450,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27453,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27454,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27454,-1},
        {27454,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21208,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21912,-1},
        {-1,-1,-1}
    },
    {
        {27459,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27466,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28270,-1},
        {-1,-1,-1}
    },
    {
        {27462,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27462,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27463,-1},
        {-1},
        {-1,-1},
        {27463,-1,-1}
    },
    {
        {27463,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24841,-1},
        {-1,-1,-1}
    },
    {
        {27465,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27465,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27466,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27476,-1},
        {-1,-1,-1}
    },
    {
        {27468,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27470,-1},
        {-1},
        {-1,-1},
        {27470,-1,-1}
    },
    {
        {27470,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27489,-1},
        {-1,-1,-1}
    },
    {
        {27472,-1},
        {-1},
        {-1,-1},
        {27431,-1,-1}
    },
    {
        {27472,-1},
        {-1},
        {-1,-1},
        {27431,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27476,34390},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27475,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27476,-1},
        {-1},
        {-1,-1},
        {27476,-1,-1}
    },
    {
        {27476,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22132,-1},
        {22132,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22132,-1},
        {-1,-1,-1}
    },
    {
        {27478,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22063,-1},
        {22063,-1,-1}
    },
    {
        {27480,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27481,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27481,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27483,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25985,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39023,-1},
        {-1,-1,-1}
    },
    {
        {27487,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27488,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27489,-1},
        {-1},
        {-1,-1},
        {27475,-1,-1}
    },
    {
        {27489,-1},
        {-1},
        {-1,-1},
        {27475,-1,-1}
    },
    {
        {27490,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27490,-1},
        {-1},
        {-1,-1},
        {27490,-1,-1}
    },
    {
        {27491,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27492,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27493,-1},
        {-1},
        {-1,-1},
        {27497,-1,-1}
    },
    {
        {27494,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27495,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27497,27493},
        {27497,27493,-1}
    },
    {
        {27497,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27498,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27500,-1},
        {-1},
        {-1,-1},
        {21069,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36383,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28544,-1,-1}
    },
    {
        {27503,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36405,-1},
        {-1,-1,-1}
    },
    {
        {27507,-1},
        {-1},
        {-1,-1},
        {27507,-1,-1}
    },
    {
        {27507,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27507,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27507,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27508,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27509,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33836,-1},
        {-1,-1,-1}
    },
    {
        {27511,-1},
        {-1},
        {-1,-1},
        {27508,-1,-1}
    },
    {
        {27512,-1},
        {-1},
        {-1,-1},
        {24112,-1,-1}
    },
    {
        {27512,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27513,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27514,-1},
        {-1},
        {-1,-1},
        {27513,-1,-1}
    },
    {
        {27515,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27519,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27520,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27521,-1},
        {-1},
        {-1,-1},
        {27519,-1,-1}
    },
    {
        {27522,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27523,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27524,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27526,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27529,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27530,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27531,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27533,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27533,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30820,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {14031,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31462,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33073,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27541,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27542,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27544,-1},
        {-1},
        {-1,-1},
        {27531,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23130,-1},
        {-1,-1,-1}
    },
    {
        {27547,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24381,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28337,-1},
        {-1,-1,-1}
    },
    {
        {27550,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27556,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23493,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40223,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20710,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27561,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27561,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27562,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27563,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27565,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27566,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27567,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27567,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27567,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27569,-1},
        {-1},
        {-1,-1},
        {27570,-1,-1}
    },
    {
        {27570,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27571,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27571,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27572,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27573,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29627,32662},
        {-1,-1,-1}
    },
    {
        {27575,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27578,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27579,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27579,-1},
        {-1},
        {-1,-1},
        {27579,-1,-1}
    },
    {
        {27580,-1},
        {-1},
        {-1,-1},
        {27579,-1,-1}
    },
    {
        {27580,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27581,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27582,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27583,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27584,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27584,-1},
        {-1},
        {-1,-1},
        {27584,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27589,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27590,-1},
        {-1},
        {-1,-1},
        {27590,-1,-1}
    },
    {
        {27590,-1},
        {-1},
        {-1,-1},
        {27572,-1,-1}
    },
    {
        {27591,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27592,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27593,-1},
        {-1},
        {-1,-1},
        {21307,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39336,-1},
        {-1,-1,-1}
    },
    {
        {27595,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27597,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27598,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27599,-1},
        {-1},
        {-1,-1},
        {27598,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27602,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22992,-1},
        {-1,-1,-1}
    },
    {
        {27602,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27603,-1},
        {-1},
        {-1,-1},
        {32946,-1,-1}
    },
    {
        {27603,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27604,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27606,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27607,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27608,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27610,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27611,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32021,-1},
        {32021,-1,-1}
    },
    {
        {27615,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34952,-1},
        {34952,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27656,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36858,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23593,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27633,29667},
        {-1,-1,-1}
    },
    {
        {27622,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27623,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27623,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27633,-1},
        {27628,-1,-1}
    },
    {
        {27627,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27628,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34461,35911},
        {-1,-1,-1}
    },
    {
        {27630,-1},
        {-1},
        {-1,-1},
        {27615,-1,-1}
    },
    {
        {27631,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22521,-1},
        {-1,-1,-1}
    },
    {
        {27633,-1},
        {-1},
        {-1,-1},
        {38816,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27635,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27628,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27633,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40353,26988},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29325,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27647,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29519,-1},
        {-1,-1,-1}
    },
    {
        {27650,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27656,31461},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27656,31461},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27653,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27658,-1},
        {-1,-1,-1}
    },
    {
        {27656,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27656,-1},
        {-1},
        {-1,-1},
        {27656,-1,-1}
    },
    {
        {27656,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27658,-1},
        {-1},
        {-1,-1},
        {27656,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33945,27647},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31341,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27662,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27662,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27663,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27664,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27665,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21413,-1,-1}
    },
    {
        {27667,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27667,-1},
        {-1},
        {-1,-1},
        {27667,-1,-1}
    },
    {
        {27667,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27668,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27675,-1},
        {-1,-1,-1}
    },
    {
        {27671,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27675,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27675,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38525,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27684,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27683,-1},
        {-1},
        {-1,-1},
        {27671,-1,-1}
    },
    {
        {27684,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20811,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24032,27699},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26481,27683},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27683,24404},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38738,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27699,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27700,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27701,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27702,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27703,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27704,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27706,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28346,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23570,-1},
        {-1,-1,-1}
    },
    {
        {27710,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27711,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27712,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27713,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27714,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35955,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {146559,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27737,27796},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27954,-1},
        {-1,-1,-1}
    },
    {
        {27725,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27725,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27726,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27726,-1},
        {-1},
        {-1,-1},
        {27725,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27728,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28216,27845},
        {27845,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27733,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27735,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27735,-1},
        {-1,-1,-1}
    },
    {
        {27737,-1},
        {-1},
        {-1,-1},
        {27738,-1,-1}
    },
    {
        {27738,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27739,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27739,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27740,-1},
        {-1},
        {-1,-1},
        {27710,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27741,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27742,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27743,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27744,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27738,-1},
        {-1},
        {27737,27738},
        {27738,-1,-1}
    },
    {
        {27746,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24558,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27868,-1},
        {-1,-1,-1}
    },
    {
        {27751,-1},
        {-1},
        {-1,-1},
        {27751,-1,-1}
    },
    {
        {27751,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27752,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27752,-1},
        {-1},
        {28104,-1},
        {-1,-1,-1}
    },
    {
        {27754,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31357,-1},
        {31357,-1,-1}
    },
    {
        {27757,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27757,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27757,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22343,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27760,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27762,-1},
        {-1},
        {-1,-1},
        {27762,-1,-1}
    },
    {
        {27762,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27763,-1},
        {-1},
        {-1,-1},
        {27764,-1,-1}
    },
    {
        {27764,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27764,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28085,28134},
        {-1,-1,-1}
    },
    {
        {27766,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27766,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22395,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27958,-1,-1}
    },
    {
        {27770,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27773,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27774,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27774,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25234,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26492,-1},
        {-1,-1,-1}
    },
    {
        {27777,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27778,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27779,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27780,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27781,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27782,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27784,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27785,-1},
        {-1},
        {-1,-1},
        {27784,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36848,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27784,-1},
        {-1,-1,-1}
    },
    {
        {27788,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27789,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27800,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30732,-1},
        {-1,-1,-1}
    },
    {
        {27792,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24568,-1},
        {-1,-1,-1}
    },
    {
        {27794,-1},
        {-1},
        {-1,-1},
        {27809,-1,-1}
    },
    {
        {27795,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27796,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27797,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27798,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27799,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27799,-1},
        {-1},
        {-1,-1},
        {28155,-1,-1}
    },
    {
        {27799,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27800,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27801,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27802,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27803,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27804,-1},
        {-1},
        {-1,-1},
        {27886,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27700,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27803,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27966,-1},
        {27966,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27966,-1},
        {27966,-1,-1}
    },
    {
        {27809,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27809,-1},
        {-1},
        {-1,-1},
        {27809,27794,-1}
    },
    {
        {27810,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27818,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27818,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27819,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23504,26151},
        {-1,-1,-1}
    },
    {
        {27821,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27822,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27825,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27825,-1,-1}
    },
    {
        {27827,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22395,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27832,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27833,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27834,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27835,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27836,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27837,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27838,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27839,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27839,-1},
        {-1},
        {-1,-1},
        {27839,-1,-1}
    },
    {
        {27839,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27934,-1},
        {-1,-1,-1}
    },
    {
        {27841,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27842,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27957,-1},
        {-1,-1,-1}
    },
    {
        {27844,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27845,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27846,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20285,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20319,-1},
        {-1,-1,-1}
    },
    {
        {27849,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27850,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27852,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24589,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27856,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27859,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27860,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27861,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27862,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27863,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28409,-1},
        {28409,-1,-1}
    },
    {
        {27865,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27865,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27866,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27867,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27868,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27869,-1},
        {-1},
        {-1,-1},
        {28335,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28632,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24596,-1},
        {-1,-1,-1}
    },
    {
        {27872,-1},
        {-1},
        {-1,-1},
        {20919,-1,-1}
    },
    {
        {27873,-1},
        {-1},
        {-1,-1},
        {27873,-1,-1}
    },
    {
        {27873,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27874,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27875,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20284,-1},
        {-1,-1,-1}
    },
    {
        {27877,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27880,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27880,-1},
        {-1},
        {-1,-1},
        {27880,-1,-1}
    },
    {
        {27881,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27882,-1},
        {-1},
        {-1,-1},
        {28057,-1,-1}
    },
    {
        {27883,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27884,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27885,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27886,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27887,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27888,-1},
        {-1},
        {-1,-1},
        {27888,-1,-1}
    },
    {
        {27888,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27889,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27891,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27849,-1},
        {-1,-1,-1}
    },
    {
        {27899,-1},
        {-1},
        {-1,-1},
        {28511,-1,-1}
    },
    {
        {27899,-1},
        {-1},
        {-1,-1},
        {28511,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22432,-1},
        {22432,-1,-1}
    },
    {
        {27904,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28500,32092},
        {28009,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28082,-1},
        {28082,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24674,-1},
        {-1,-1,-1}
    },
    {
        {27908,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28006,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25327,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22434,-1},
        {-1,-1,-1}
    },
    {
        {27914,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27915,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27916,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27740,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27918,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {14618,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27922,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28460,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28382,28399},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20350,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24734,-1},
        {-1,-1,-1}
    },
    {
        {27927,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25335,-1},
        {-1,-1,-1}
    },
    {
        {27929,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27931,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27931,-1},
        {27931,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20370,-1},
        {-1,-1,-1}
    },
    {
        {27934,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27935,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27819,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20234,-1},
        {20234,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27940,-1},
        {-1},
        {-1,-1},
        {27849,-1,-1}
    },
    {
        {27941,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27942,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27943,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27945,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27946,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27947,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28194,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24647,-1},
        {-1,-1,-1}
    },
    {
        {27950,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27951,-1},
        {-1},
        {-1,-1},
        {28500,-1,-1}
    },
    {
        {27951,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27953,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27954,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27955,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27956,-1},
        {-1},
        {-1,-1},
        {27956,-1,-1}
    },
    {
        {27956,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27957,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27958,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27960,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27961,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27931,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27963,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27964,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27965,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27966,-1},
        {-1},
        {-1,-1},
        {27966,-1,-1}
    },
    {
        {27966,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27967,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27833,-1},
        {-1,-1,-1}
    },
    {
        {27969,-1},
        {-1},
        {-1,-1},
        {27969,-1,-1}
    },
    {
        {27969,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27846,-1},
        {27846,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28025,-1},
        {28025,-1,-1}
    },
    {
        {27972,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27973,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28609,-1},
        {28609,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22482,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27992,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27993,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27994,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27995,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27996,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30173,-1},
        {-1,-1,-1}
    },
    {
        {27998,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28000,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35910,-1},
        {-1,-1,-1}
    },
    {
        {28003,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28004,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28005,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28006,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28121,23435},
        {-1,-1,-1}
    },
    {
        {28009,-1},
        {-1},
        {-1,-1},
        {28009,-1,-1}
    },
    {
        {28009,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28010,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28145,-1},
        {-1,-1,-1}
    },
    {
        {28012,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20415,-1},
        {-1,-1,-1}
    },
    {
        {28014,-1},
        {-1},
        {-1,-1},
        {28014,-1,-1}
    },
    {
        {28014,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28016,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28440,-1},
        {28440,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28020,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24420,28558},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29282,-1},
        {-1,-1,-1}
    },
    {
        {28023,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28024,-1},
        {-1},
        {-1,-1},
        {28024,-1,-1}
    },
    {
        {28024,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28025,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27798,-1},
        {27798,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28913,27842},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {135933,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28032,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28033,-1},
        {-1},
        {-1,-1},
        {28402,-1,-1}
    },
    {
        {28034,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24707,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28037,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27751,-1},
        {-1,-1,-1}
    },
    {
        {28039,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28040,-1},
        {-1},
        {-1,-1},
        {28040,-1,-1}
    },
    {
        {28040,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28041,-1},
        {-1},
        {-1,-1},
        {28169,-1,-1}
    },
    {
        {28042,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27889,-1},
        {-1,-1,-1}
    },
    {
        {28044,-1},
        {-1},
        {-1,-1},
        {28263,-1,-1}
    },
    {
        {28045,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28046,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28046,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28046,-1},
        {-1,-1,-1}
    },
    {
        {28049,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28051,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28052,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28053,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28056,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28057,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28059,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28060,-1},
        {-1},
        {-1,-1},
        {28678,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28516,-1},
        {-1,-1,-1}
    },
    {
        {28074,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28075,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28076,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28076,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28644,-1},
        {-1,-1,-1}
    },
    {
        {28078,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28079,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21852,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28466,-1},
        {28466,-1,-1}
    },
    {
        {28082,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24766,-1},
        {-1,-1,-1}
    },
    {
        {28084,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28085,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28085,-1},
        {-1},
        {-1,-1},
        {28085,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21822,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28088,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20520,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25448,-1},
        {-1,-1,-1}
    },
    {
        {28092,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36367,-1},
        {-1,-1,-1}
    },
    {
        {28095,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28095,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28095,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28096,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28503,-1},
        {-1,-1,-1}
    },
    {
        {28098,-1},
        {-1},
        {-1,-1},
        {24471,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28207,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28207,-1,-1}
    },
    {
        {28100,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28101,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28102,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28103,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28104,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28107,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28108,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28108,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21608,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28110,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28111,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28111,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28112,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28113,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28114,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28117,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28118,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25516,-1},
        {-1,-1,-1}
    },
    {
        {28120,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28121,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28122,-1},
        {-1},
        {-1,-1},
        {28057,-1,-1}
    },
    {
        {28123,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39342,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28125,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28126,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28126,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28127,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28129,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20522,-1},
        {-1,-1,-1}
    },
    {
        {28132,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28134,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23443,27852},
        {-1,-1,-1}
    },
    {
        {28136,-1},
        {-1},
        {-1,-1},
        {27972,-1,-1}
    },
    {
        {28136,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28137,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28138,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28139,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28139,-1},
        {-1},
        {-1,-1},
        {28139,-1,-1}
    },
    {
        {28140,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28122,-1},
        {-1,-1,-1}
    },
    {
        {28142,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28145,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28408,-1},
        {-1,-1,-1}
    },
    {
        {28147,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28149,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28150,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28151,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28165,-1},
        {-1},
        {-1,-1},
        {28165,-1,-1}
    },
    {
        {28153,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28154,-1},
        {-1},
        {-1,-1},
        {27973,-1,-1}
    },
    {
        {28155,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28155,-1},
        {-1},
        {-1,-1},
        {28155,-1,-1}
    },
    {
        {28156,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27802,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39154,39150},
        {39154,39150,-1}
    },
    {
        {-1,-1},
        {-1},
        {27850,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22868,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28149,-1},
        {28149,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27766,23735},
        {27766,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28165,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28149,-1},
        {28149,-1,-1}
    },
    {
        {28167,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28168,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28169,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28170,-1},
        {-1},
        {-1,-1},
        {28181,28149,-1}
    },
    {
        {28171,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20478,-1},
        {-1,-1,-1}
    },
    {
        {28179,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28181,-1},
        {-1},
        {-1,-1},
        {28149,-1,-1}
    },
    {
        {28183,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26580,-1},
        {-1,-1,-1}
    },
    {
        {28185,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28186,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28187,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28641,-1},
        {-1,-1,-1}
    },
    {
        {28189,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28189,-1},
        {-1},
        {-1,-1},
        {28189,-1,-1}
    },
    {
        {28190,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28191,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28192,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28192,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28193,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28194,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28195,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28195,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28196,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28197,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28198,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28199,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28199,-1},
        {-1},
        {-1,-1},
        {28199,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20558,-1},
        {-1,-1,-1}
    },
    {
        {28201,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28641,-1},
        {28641,-1,-1}
    },
    {
        {28203,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28204,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28205,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28206,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28206,-1},
        {-1},
        {-1,-1},
        {28206,-1,-1}
    },
    {
        {28207,-1},
        {-1},
        {-1,-1},
        {28207,-1,-1}
    },
    {
        {28207,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28207,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28153,26235},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34425,-1},
        {-1,-1,-1}
    },
    {
        {28210,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28466,24365},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28466,24365},
        {-1,-1,-1}
    },
    {
        {28212,-1},
        {-1},
        {-1,-1},
        {28167,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25551,-1},
        {-1,-1,-1}
    },
    {
        {28214,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28214,-1},
        {-1},
        {-1,-1},
        {28214,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28216,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28218,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28220,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28222,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {15240,28422},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25542,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28233,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24853,-1},
        {-1,-1,-1}
    },
    {
        {28227,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28228,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20579,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28232,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28233,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28234,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28235,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28236,-1},
        {-1},
        {-1,-1},
        {39184,-1,-1}
    },
    {
        {28237,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28238,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28239,-1},
        {-1},
        {-1,-1},
        {38920,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27850,27942},
        {27942,-1,-1}
    },
    {
        {28241,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28243,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28244,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24484,-1},
        {-1,-1,-1}
    },
    {
        {28246,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28247,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28248,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22869,28286},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32996,-1},
        {-1,-1,-1}
    },
    {
        {28251,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28251,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28252,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28254,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28255,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20598,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28259,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25580,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24826,-1},
        {-1,-1,-1}
    },
    {
        {28263,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28263,-1},
        {-1},
        {-1,-1},
        {28263,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37325,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28095,-1},
        {-1,-1,-1}
    },
    {
        {28267,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28267,-1},
        {28267,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27845,-1},
        {-1,-1,-1}
    },
    {
        {28270,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28270,-1},
        {-1},
        {-1,-1},
        {28270,-1,-1}
    },
    {
        {28270,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28271,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {63964,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30729,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28274,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28274,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22527,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22527,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32657,-1},
        {-1,-1,-1}
    },
    {
        {28278,-1},
        {-1},
        {-1,-1},
        {27849,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28222,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21928,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28147,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {146752,28037},
        {28037,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {146752,28037},
        {28037,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28100,-1,-1}
    },
    {
        {28286,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28286,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28287,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28288,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28288,-1},
        {-1},
        {-1,-1},
        {28288,-1,-1}
    },
    {
        {28290,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28433,-1},
        {-1},
        {28433,-1},
        {-1,-1,-1}
    },
    {
        {28300,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26690,27964},
        {-1,-1,-1}
    },
    {
        {28303,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28303,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28304,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28254,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34945,28414},
        {-1,-1,-1}
    },
    {
        {28307,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28307,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32660,28220},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28635,-1},
        {28635,-1,-1}
    },
    {
        {28310,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23882,-1},
        {-1,-1,-1}
    },
    {
        {28312,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28234,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22612,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31370,-1},
        {31370,-1,-1}
    },
    {
        {28316,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28317,-1},
        {-1},
        {-1,-1},
        {28317,-1,-1}
    },
    {
        {28317,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28319,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22610,-1},
        {-1,-1,-1}
    },
    {
        {28322,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28322,-1},
        {-1},
        {-1,-1},
        {28322,-1,-1}
    },
    {
        {28322,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28138,20523},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39342,-1},
        {39342,-1,-1}
    },
    {
        {28325,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28325,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28228,-1},
        {-1,-1,-1}
    },
    {
        {28327,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26685,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22626,-1},
        {-1,-1,-1}
    },
    {
        {28330,-1},
        {-1},
        {-1,-1},
        {28179,-1,-1}
    },
    {
        {28330,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28331,-1},
        {-1},
        {-1,-1},
        {28201,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32663,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28335,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30362,-1},
        {30362,-1,-1}
    },
    {
        {28337,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28338,-1},
        {-1},
        {-1,-1},
        {28338,-1,-1}
    },
    {
        {28338,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28338,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28338,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28340,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27575,-1},
        {-1,-1,-1}
    },
    {
        {28342,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28343,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27869,28335},
        {15633,36961,-1}
    },
    {
        {-1,-1},
        {-1},
        {32034,-1},
        {-1,-1,-1}
    },
    {
        {28346,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28346,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28629,28287},
        {28629,28287,-1}
    },
    {
        {28349,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20608,-1},
        {-1,-1,-1}
    },
    {
        {28351,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28351,-1},
        {-1},
        {-1,-1},
        {28351,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28353,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28354,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28355,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28356,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28357,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38548,-1},
        {-1,-1,-1}
    },
    {
        {28359,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28359,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28361,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28362,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28363,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28363,-1},
        {-1},
        {-1,-1},
        {28363,-1,-1}
    },
    {
        {28363,-1},
        {-1},
        {-1,-1},
        {28363,-1,-1}
    },
    {
        {28363,-1},
        {-1},
        {-1,-1},
        {28363,-1,-1}
    },
    {
        {28364,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28366,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20625,-1},
        {-1,-1,-1}
    },
    {
        {28369,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24385,27827},
        {-1,-1,-1}
    },
    {
        {28371,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28372,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28372,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28373,-1},
        {-1},
        {-1,-1},
        {28373,-1,-1}
    },
    {
        {28373,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28373,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27057,28414},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28342,-1},
        {-1,-1,-1}
    },
    {
        {28377,-1},
        {-1},
        {-1,-1},
        {21295,-1,-1}
    },
    {
        {28414,-1},
        {-1},
        {28414,-1},
        {28414,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27784,28139},
        {28139,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28278,-1},
        {-1,-1,-1}
    },
    {
        {28381,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28382,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28440,-1},
        {28440,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28395,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28396,-1},
        {-1},
        {-1,-1},
        {28396,-1,-1}
    },
    {
        {28396,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28399,-1},
        {-1},
        {-1,-1},
        {28382,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {147326,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {147326,-1},
        {-1,-1,-1}
    },
    {
        {28402,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28404,-1},
        {-1,-1,-1}
    },
    {
        {28404,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28404,-1},
        {-1},
        {-1,-1},
        {28404,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27154,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28608,24928},
        {-1,-1,-1}
    },
    {
        {28407,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28408,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28409,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28411,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28536,26805},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28414,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28414,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28415,-1},
        {-1},
        {-1,-1},
        {28288,-1,-1}
    },
    {
        {28415,-1},
        {-1},
        {-1,-1},
        {28415,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30956,-1},
        {-1,-1,-1}
    },
    {
        {28417,-1},
        {-1},
        {-1,-1},
        {28417,-1,-1}
    },
    {
        {28417,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28418,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23490,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28079,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24034,24035},
        {-1,-1,-1}
    },
    {
        {28422,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20657,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28425,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28426,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30275,-1},
        {-1,-1,-1}
    },
    {
        {28428,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28520,-1},
        {-1,-1,-1}
    },
    {
        {28431,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28432,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28433,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28433,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28433,-1},
        {-1},
        {-1,-1},
        {28433,-1,-1}
    },
    {
        {28433,-1},
        {-1},
        {-1,-1},
        {28433,-1,-1}
    },
    {
        {28433,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28433,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28433,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35137,-1},
        {-1,-1,-1}
    },
    {
        {28435,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28436,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28437,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28440,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28237,-1},
        {-1,-1,-1}
    },
    {
        {28442,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28442,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28568,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24948,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27752,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25690,-1},
        {-1,-1,-1}
    },
    {
        {28448,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28448,-1},
        {-1},
        {-1,-1},
        {28448,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29145,28271},
        {28271,-1,-1}
    },
    {
        {28450,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {64071,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28451,-1},
        {-1},
        {-1,-1},
        {28451,-1,-1}
    },
    {
        {28451,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28651,-1},
        {-1,-1,-1}
    },
    {
        {28453,-1},
        {-1},
        {-1,-1},
        {31402,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20094,-1},
        {20094,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28458,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28459,-1},
        {-1},
        {-1,-1},
        {28459,-1,-1}
    },
    {
        {28459,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28460,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24951,-1},
        {-1,-1,-1}
    },
    {
        {28463,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28465,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28466,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28467,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28467,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28121,-1},
        {-1,-1,-1}
    },
    {
        {28469,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28470,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27112,-1},
        {-1,-1,-1}
    },
    {
        {28472,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37154,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28025,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28476,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28478,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28478,-1},
        {-1},
        {-1,-1},
        {28478,-1,-1}
    },
    {
        {28479,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28520,-1},
        {28520,-1,-1}
    },
    {
        {28481,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27946,-1},
        {27946,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28395,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28465,22013},
        {28465,-1,-1}
    },
    {
        {28485,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28418,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24980,-1},
        {-1,-1,-1}
    },
    {
        {28497,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28498,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28500,-1},
        {-1},
        {-1,-1},
        {28500,-1,-1}
    },
    {
        {28500,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28500,-1},
        {-1},
        {-1,-1},
        {28500,-1,-1}
    },
    {
        {28500,-1},
        {-1},
        {-1,-1},
        {28500,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25006,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28005,29750},
        {-1,-1,-1}
    },
    {
        {28503,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28504,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28505,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28506,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28507,-1},
        {-1},
        {-1,-1},
        {28508,-1,-1}
    },
    {
        {28507,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28507,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28507,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28507,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28508,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28510,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28511,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28512,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28513,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28514,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28239,-1},
        {-1,-1,-1}
    },
    {
        {28516,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28516,-1},
        {-1},
        {-1,-1},
        {28516,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28327,-1},
        {-1,-1,-1}
    },
    {
        {28518,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28337,-1},
        {-1,-1,-1}
    },
    {
        {28520,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28512,30064},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26234,28508},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28525,-1},
        {-1},
        {-1,-1},
        {28525,-1,-1}
    },
    {
        {28525,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28526,-1},
        {-1},
        {-1,-1},
        {28526,-1,-1}
    },
    {
        {28526,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28527,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28527,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28528,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31245,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28532,-1},
        {-1},
        {-1,-1},
        {28710,-1,-1}
    },
    {
        {28532,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28552,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {144952,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33871,-1},
        {-1,-1,-1}
    },
    {
        {28536,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28536,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28538,-1},
        {28538,-1,-1}
    },
    {
        {28538,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28422,-1},
        {-1,-1,-1}
    },
    {
        {28540,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28540,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28541,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28542,-1},
        {-1},
        {-1,-1},
        {28542,-1,-1}
    },
    {
        {28542,-1},
        {-1},
        {-1,-1},
        {28542,-1,-1}
    },
    {
        {28542,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22285,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22285,-1},
        {-1,-1,-1}
    },
    {
        {28544,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28544,-1},
        {-1},
        {-1,-1},
        {28171,-1,-1}
    },
    {
        {28544,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28545,-1},
        {-1},
        {-1,-1},
        {28171,-1,-1}
    },
    {
        {28546,-1},
        {-1},
        {-1,-1},
        {28548,-1,-1}
    },
    {
        {28546,-1},
        {-1},
        {-1,-1},
        {28548,-1,-1}
    },
    {
        {28546,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28548,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30059,-1},
        {-1,-1,-1}
    },
    {
        {28550,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28552,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27204,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20692,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28556,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28557,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28558,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27995,-1},
        {27995,-1,-1}
    },
    {
        {28560,-1},
        {-1},
        {-1,-1},
        {27780,-1,-1}
    },
    {
        {28561,-1},
        {-1},
        {-1,-1},
        {28316,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28564,-1},
        {-1},
        {-1,-1},
        {28009,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29121,-1},
        {-1,-1,-1}
    },
    {
        {28566,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28567,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28568,-1},
        {-1},
        {-1,-1},
        {28536,-1,-1}
    },
    {
        {28568,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28595,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34398,28670},
        {-1,-1,-1}
    },
    {
        {28575,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28576,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28576,-1},
        {-1},
        {-1,-1},
        {28576,-1,-1}
    },
    {
        {28577,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25803,-1},
        {-1,-1,-1}
    },
    {
        {28579,-1},
        {-1},
        {-1,-1},
        {28003,-1,-1}
    },
    {
        {28580,-1},
        {-1},
        {-1,-1},
        {27810,-1,-1}
    },
    {
        {28580,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28580,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28581,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28583,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23416,28366},
        {-1,-1,-1}
    },
    {
        {28586,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28459,34261},
        {28288,28459,-1}
    },
    {
        {-1,-1},
        {-1},
        {28459,34261},
        {28288,28459,-1}
    },
    {
        {-1,-1},
        {-1},
        {28459,34261},
        {28288,28459,-1}
    },
    {
        {-1,-1},
        {-1},
        {28459,34261},
        {28288,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36039,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28591,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28591,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28593,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {13458,-1},
        {-1,-1,-1}
    },
    {
        {28595,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28595,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28597,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28598,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28459,-1},
        {28459,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28459,-1},
        {28459,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25022,-1},
        {-1,-1,-1}
    },
    {
        {28601,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25014,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29159,-1},
        {-1,-1,-1}
    },
    {
        {28604,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25818,-1},
        {-1,-1,-1}
    },
    {
        {28606,-1},
        {-1},
        {-1,-1},
        {28606,-1,-1}
    },
    {
        {28606,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35167,-1},
        {-1,-1,-1}
    },
    {
        {28608,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28609,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28610,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28610,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28611,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28198,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28198,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28024,-1,-1}
    },
    {
        {28614,-1},
        {-1},
        {-1,-1},
        {28614,-1,-1}
    },
    {
        {28614,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28615,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28616,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28616,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28617,-1},
        {-1},
        {-1,-1},
        {30562,-1,-1}
    },
    {
        {28618,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30990,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24865,33905},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40718,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28186,-1},
        {-1,-1,-1}
    },
    {
        {28628,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28629,-1},
        {-1},
        {-1,-1},
        {28287,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28710,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28448,24149},
        {-1,-1,-1}
    },
    {
        {28632,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28632,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28366,-1},
        {28366,-1,-1}
    },
    {
        {28634,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28635,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28635,-1},
        {-1},
        {-1,-1},
        {28635,-1,-1}
    },
    {
        {28635,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20760,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28103,-1},
        {28103,-1,-1}
    },
    {
        {28638,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28639,-1},
        {-1},
        {-1,-1},
        {28168,-1,-1}
    },
    {
        {28640,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28640,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28641,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29883,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28579,-1},
        {-1,-1,-1}
    },
    {
        {28644,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23493,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38577,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23565,28557},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28649,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28649,-1},
        {-1},
        {-1,-1},
        {28649,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28651,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28652,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28312,-1},
        {-1,-1,-1}
    },
    {
        {28654,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28655,-1},
        {-1},
        {-1,-1},
        {28655,-1,-1}
    },
    {
        {28655,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28655,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28656,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28657,-1},
        {-1},
        {-1,-1},
        {27996,-1,-1}
    },
    {
        {28657,-1},
        {-1},
        {-1,-1},
        {27996,-1,-1}
    },
    {
        {28657,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28659,-1},
        {-1},
        {-1,-1},
        {28507,28508,-1}
    },
    {
        {-1,-1},
        {-1},
        {28634,-1},
        {-1,-1,-1}
    },
    {
        {28661,-1},
        {-1},
        {-1,-1},
        {27996,-1,-1}
    },
    {
        {28662,-1},
        {-1},
        {-1,-1},
        {38346,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28415,-1},
        {-1,-1,-1}
    },
    {
        {28665,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28665,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28666,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28528,38564},
        {-1,-1,-1}
    },
    {
        {28668,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28670,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21426,30725},
        {30725,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28672,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28673,-1},
        {-1},
        {-1,-1},
        {28478,-1,-1}
    },
    {
        {28673,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40565,28407},
        {28407,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36060,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {15877,-1},
        {-1,-1,-1}
    },
    {
        {28677,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28678,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28679,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28679,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32748,-1},
        {-1,-1,-1}
    },
    {
        {28681,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28682,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28683,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28425,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25273,-1},
        {-1,-1,-1}
    },
    {
        {28687,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25076,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25076,-1},
        {-1,-1,-1}
    },
    {
        {28689,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28544,28615},
        {28615,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28546,28548},
        {28546,28548,-1}
    },
    {
        {-1,-1},
        {-1},
        {64084,-1},
        {-1,-1,-1}
    },
    {
        {28693,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28693,-1},
        {-1},
        {-1,-1},
        {28693,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28696,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28597,-1},
        {-1,-1,-1}
    },
    {
        {28698,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28698,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28699,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28699,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28699,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28699,-1},
        {-1},
        {-1,-1},
        {28699,-1,-1}
    },
    {
        {28699,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28701,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28702,-1},
        {-1},
        {-1,-1},
        {28702,-1,-1}
    },
    {
        {28702,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28702,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28702,-1},
        {-1},
        {-1,-1},
        {28702,-1,-1}
    },
    {
        {28702,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28703,-1},
        {-1},
        {-1,-1},
        {28703,-1,-1}
    },
    {
        {28703,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38568,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38568,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28707,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25079,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28710,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28710,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28711,-1},
        {-1},
        {-1,-1},
        {28381,-1,-1}
    },
    {
        {28711,-1},
        {-1},
        {-1,-1},
        {28381,28711,-1}
    },
    {
        {28711,-1},
        {-1},
        {-1,-1},
        {28711,-1,-1}
    },
    {
        {28711,-1},
        {-1},
        {-1,-1},
        {28711,-1,-1}
    },
    {
        {28711,-1},
        {-1},
        {-1,-1},
        {28711,-1,-1}
    },
    {
        {28711,-1},
        {-1},
        {-1,-1},
        {28711,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28712,-1},
        {-1},
        {-1,-1},
        {28716,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38973,28693},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {64089,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32288,-1},
        {-1,-1,-1}
    },
    {
        {28716,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28720,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32637,-1},
        {-1,-1,-1}
    },
    {
        {28722,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27914,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31966,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28307,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27570,-1},
        {-1,-1,-1}
    },
    {
        {28729,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28730,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25140,-1},
        {-1,-1,-1}
    },
    {
        {28732,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28734,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28734,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {64089,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38684,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28149,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28149,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30474,-1},
        {-1,-1,-1}
    },
    {
        {28739,-1},
        {-1},
        {-1,-1},
        {35920,35914,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28463,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28651,34253},
        {-1,-1,-1}
    },
    {
        {28743,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28743,-1},
        {-1},
        {-1,-1},
        {28520,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28746,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28746,-1},
        {-1},
        {-1,-1},
        {28746,-1,-1}
    },
    {
        {28747,-1},
        {-1},
        {-1,-1},
        {27861,-1,-1}
    },
    {
        {28748,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28748,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {147715,38357},
        {-1,-1,-1}
    },
    {
        {28750,-1},
        {-1},
        {-1,-1},
        {28777,-1,-1}
    },
    {
        {28753,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27973,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20081,28772},
        {-1,-1,-1}
    },
    {
        {28756,-1},
        {-1},
        {-1,-1},
        {28777,28750,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28720,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28760,-1},
        {-1},
        {-1,-1},
        {28760,-1,-1}
    },
    {
        {28760,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25898,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38369,-1},
        {-1,-1,-1}
    },
    {
        {28765,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28766,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23660,23646},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28651,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28734,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28771,-1},
        {-1},
        {-1,-1},
        {28286,-1,-1}
    },
    {
        {28772,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28773,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28765,39023},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28777,-1},
        {28777,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36123,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36123,-1},
        {-1,-1,-1}
    },
    {
        {28777,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39729,-1},
        {-1,-1,-1}
    },
    {
        {28779,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28780,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28782,-1},
        {-1},
        {-1,-1},
        {20809,-1,-1}
    },
    {
        {28782,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28782,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28783,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28784,-1},
        {-1},
        {-1,-1},
        {28784,-1,-1}
    },
    {
        {28784,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29063,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21089,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36637,-1},
        {36637,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28888,-1},
        {28888,-1,-1}
    },
    {
        {28789,-1},
        {-1},
        {-1,-1},
        {38666,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28790,-1},
        {-1},
        {-1,-1},
        {31432,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28782,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28782,-1},
        {-1,-1,-1}
    },
    {
        {28792,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28792,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29152,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22320,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36196,-1},
        {36196,-1,-1}
    },
    {
        {28796,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28796,-1},
        {-1},
        {-1,-1},
        {28796,-1,-1}
    },
    {
        {28797,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28797,-1},
        {-1},
        {-1,-1},
        {28797,-1,-1}
    },
    {
        {28798,-1},
        {-1},
        {-1,-1},
        {28797,-1,-1}
    },
    {
        {28801,-1},
        {-1},
        {-1,-1},
        {28801,27683,27671}
    },
    {
        {28801,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28954,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24568,-1},
        {-1,-1,-1}
    },
    {
        {28805,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28806,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28806,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25237,-1},
        {-1,-1,-1}
    },
    {
        {28809,-1},
        {-1},
        {-1,-1},
        {29200,28809,-1}
    },
    {
        {28809,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28810,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28783,22351},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21467,-1},
        {21467,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28814,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28814,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28925,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31180,-1},
        {-1,-1,-1}
    },
    {
        {28818,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28790,25238},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28821,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28823,-1},
        {-1},
        {-1,-1},
        {20809,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28825,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28831,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25276,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24629,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20309,28847},
        {-1,-1,-1}
    },
    {
        {28836,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24619,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25302,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22376,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24604,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25305,-1},
        {-1,-1,-1}
    },
    {
        {28843,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28844,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28844,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28845,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28845,-1},
        {-1},
        {-1,-1},
        {28845,-1,-1}
    },
    {
        {28846,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28846,-1},
        {-1},
        {-1,-1},
        {28846,-1,-1}
    },
    {
        {28847,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28849,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28849,-1},
        {28849,-1,-1}
    },
    {
        {28851,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28851,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26570,28810},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24599,-1},
        {-1,-1,-1}
    },
    {
        {28855,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28855,-1},
        {-1},
        {-1,-1},
        {28855,-1,-1}
    },
    {
        {28856,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28857,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28858,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28859,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24676,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {14618,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20346,-1},
        {-1,-1,-1}
    },
    {
        {28872,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29029,-1},
        {29029,-1,-1}
    },
    {
        {28874,-1},
        {-1},
        {-1,-1},
        {29036,-1,-1}
    },
    {
        {28875,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28988,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26654,28823},
        {-1,-1,-1}
    },
    {
        {28879,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30698,-1},
        {-1,-1,-1}
    },
    {
        {28881,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28883,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28884,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28797,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22419,-1},
        {-1,-1,-1}
    },
    {
        {28888,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28889,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24688,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30682,29165},
        {-1,-1,-1}
    },
    {
        {28892,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28893,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29190,-1},
        {29190,-1,-1}
    },
    {
        {28895,-1},
        {-1},
        {-1,-1},
        {29017,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24061,28782},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22430,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24674,-1},
        {24674,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28900,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28900,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28908,-1},
        {-1},
        {-1,-1},
        {29180,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26206,-1},
        {26206,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24214,-1},
        {-1,-1,-1}
    },
    {
        {28913,-1},
        {-1},
        {-1,-1},
        {28847,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24714,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24465,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28796,-1},
        {28796,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22479,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29715,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {64061,-1},
        {-1,-1,-1}
    },
    {
        {28921,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22479,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29730,-1},
        {-1,-1,-1}
    },
    {
        {28925,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36196,-1},
        {36196,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28884,24420},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29076,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21561,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20496,-1},
        {-1,-1,-1}
    },
    {
        {28931,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28932,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28934,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28935,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28936,-1},
        {-1},
        {-1,-1},
        {29081,29013,-1}
    },
    {
        {28936,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28937,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37554,37356},
        {29103,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28940,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24716,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24714,-1},
        {-1,-1,-1}
    },
    {
        {28943,-1},
        {-1},
        {-1,-1},
        {19998,20127,-1}
    },
    {
        {28948,-1},
        {-1},
        {-1,-1},
        {28976,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28644,29182},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28953,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28954,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38275,-1},
        {-1,-1,-1}
    },
    {
        {28956,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24795,-1},
        {-1,-1,-1}
    },
    {
        {28958,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24796,-1},
        {-1,-1,-1}
    },
    {
        {28960,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28961,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24766,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32858,28818},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33104,-1},
        {33104,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24779,-1},
        {-1,-1,-1}
    },
    {
        {28966,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28934,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22539,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31353,28286},
        {-1,-1,-1}
    },
    {
        {28971,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28971,-1},
        {-1},
        {-1,-1},
        {28971,29191,-1}
    },
    {
        {-1,-1},
        {-1},
        {24789,-1},
        {-1,-1,-1}
    },
    {
        {29026,-1},
        {-1},
        {-1,-1},
        {29026,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28974,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28975,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28976,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28977,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28977,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29020,-1},
        {29020,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28982,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22500,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29056,26119},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20513,-1},
        {-1,-1,-1}
    },
    {
        {28988,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38761,29028},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28883,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22585,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37723,-1,-1}
    },
    {
        {28998,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28999,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29001,-1},
        {-1},
        {-1,-1},
        {29001,-1,-1}
    },
    {
        {29001,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29002,-1},
        {-1},
        {-1,-1},
        {26244,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29004,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24832,-1},
        {-1,-1,-1}
    },
    {
        {29006,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29006,-1},
        {-1},
        {-1,-1},
        {29006,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29008,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29009,-1},
        {-1},
        {-1,-1},
        {29038,-1,-1}
    },
    {
        {29010,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29010,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29010,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29011,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26441,38275},
        {-1,-1,-1}
    },
    {
        {29013,-1},
        {-1},
        {-1,-1},
        {29081,-1,-1}
    },
    {
        {29014,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29014,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29014,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29017,-1},
        {-1},
        {-1,-1},
        {29017,-1,-1}
    },
    {
        {29017,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29020,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23194,-1},
        {-1,-1,-1}
    },
    {
        {29022,-1},
        {-1},
        {-1,-1},
        {63856,27578,-1}
    },
    {
        {-1,-1},
        {-1},
        {28205,-1},
        {-1,-1,-1}
    },
    {
        {29024,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29180,-1},
        {-1,-1,-1}
    },
    {
        {29026,-1},
        {-1},
        {-1,-1},
        {29026,-1,-1}
    },
    {
        {29026,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29026,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25545,27002},
        {-1,-1,-1}
    },
    {
        {29028,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29029,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29030,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29031,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29032,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29033,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26978,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29159,-1},
        {29159,-1,-1}
    },
    {
        {29036,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28872,-1},
        {28872,-1,-1}
    },
    {
        {29038,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29157,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22557,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22561,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31946,-1},
        {-1,-1,-1}
    },
    {
        {29061,-1},
        {-1},
        {29096,29061},
        {29061,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33151,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33151,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22616,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28818,-1},
        {28818,-1,-1}
    },
    {
        {29053,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29053,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29053,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29190,-1},
        {29190,-1,-1}
    },
    {
        {29056,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24894,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21209,23162},
        {-1,-1,-1}
    },
    {
        {29060,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29061,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28312,-1},
        {-1,-1,-1}
    },
    {
        {29063,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29064,-1},
        {-1},
        {-1,-1},
        {29081,29013,-1}
    },
    {
        {29064,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29066,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29066,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23470,29743},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24493,-1},
        {-1,-1,-1}
    },
    {
        {29071,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28319,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24906,-1},
        {-1,-1,-1}
    },
    {
        {29074,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22626,-1},
        {-1,-1,-1}
    },
    {
        {29076,-1},
        {-1},
        {-1,-1},
        {37780,-1,-1}
    },
    {
        {29076,-1},
        {-1},
        {-1,-1},
        {37780,29076,-1}
    },
    {
        {29077,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29078,-1},
        {-1},
        {-1,-1},
        {28976,-1,-1}
    },
    {
        {29081,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29083,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25693,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24947,-1},
        {-1,-1,-1}
    },
    {
        {29087,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29088,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29090,-1},
        {-1},
        {-1,-1},
        {29090,28925,-1}
    },
    {
        {29090,-1},
        {-1},
        {-1,-1},
        {28925,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22676,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29088,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27147,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27147,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24939,-1},
        {-1,-1,-1}
    },
    {
        {29096,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28396,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31227,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28825,-1},
        {28825,-1,-1}
    },
    {
        {29100,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24935,-1},
        {-1,-1,-1}
    },
    {
        {29102,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29103,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29105,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29107,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29107,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29224,-1},
        {29224,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22679,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20717,-1},
        {-1,-1,-1}
    },
    {
        {29113,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29114,-1},
        {-1},
        {-1,-1},
        {29113,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29032,30064},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28506,-1},
        {-1,-1,-1}
    },
    {
        {29118,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29118,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29121,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29121,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29121,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29121,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29123,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29124,-1},
        {-1},
        {-1,-1},
        {28976,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29063,-1},
        {29063,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28940,-1},
        {-1,-1,-1}
    },
    {
        {29128,-1},
        {-1},
        {-1,-1},
        {28783,-1,-1}
    },
    {
        {29129,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28977,27054},
        {-1,-1,-1}
    },
    {
        {29131,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28954,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25749,-1},
        {-1,-1,-1}
    },
    {
        {29134,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29136,-1},
        {-1},
        {-1,-1},
        {29136,-1,-1}
    },
    {
        {29136,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29136,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28884,-1},
        {28884,-1,-1}
    },
    {
        {29138,-1},
        {-1},
        {-1,-1},
        {28988,-1,-1}
    },
    {
        {29139,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29140,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29141,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29143,-1},
        {-1},
        {-1,-1},
        {29211,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29148,-1},
        {-1,-1,-1}
    },
    {
        {29145,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29148,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29151,-1},
        {-1},
        {-1,-1},
        {21942,-1,-1}
    },
    {
        {29152,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29152,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29139,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25037,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {145180,-1},
        {-1,-1,-1}
    },
    {
        {29157,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29158,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29159,-1},
        {-1},
        {-1,-1},
        {29159,-1,-1}
    },
    {
        {29159,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29159,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29159,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20736,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24865,33905},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28610,-1},
        {-1,-1,-1}
    },
    {
        {29164,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29165,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29166,-1},
        {-1},
        {-1,-1},
        {29166,-1,-1}
    },
    {
        {29166,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28586,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29032,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25014,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29031,-1},
        {29031,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29173,-1},
        {-1},
        {-1,-1},
        {29173,-1,-1}
    },
    {
        {29173,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29177,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29063,-1},
        {29063,-1,-1}
    },
    {
        {29179,-1},
        {-1},
        {-1,-1},
        {29071,-1,-1}
    },
    {
        {29179,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29180,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22740,-1},
        {-1,-1,-1}
    },
    {
        {29182,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29183,-1},
        {-1},
        {-1,-1},
        {29183,-1,-1}
    },
    {
        {29183,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29184,-1},
        {-1},
        {-1,-1},
        {36203,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29224,29074},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27348,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28665,-1},
        {-1,-1,-1}
    },
    {
        {29190,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29191,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29191,-1},
        {-1},
        {-1,-1},
        {29191,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28425,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29071,-1},
        {29071,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26336,26179},
        {-1,-1,-1}
    },
    {
        {29197,-1},
        {-1},
        {-1,-1},
        {37984,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22754,-1},
        {-1,-1,-1}
    },
    {
        {29200,-1},
        {-1},
        {-1,-1},
        {28809,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {18276,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26342,-1},
        {-1,-1,-1}
    },
    {
        {29205,-1},
        {-1},
        {-1,-1},
        {29166,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22752,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29121,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29121,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22756,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29211,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29211,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28743,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28884,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30319,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {145180,34822},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40587,31964},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29211,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29211,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29165,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27422,-1},
        {-1,-1,-1}
    },
    {
        {29224,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29224,-1},
        {-1},
        {-1,-1},
        {29224,-1,-1}
    },
    {
        {29224,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29224,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39729,-1},
        {-1,-1,-1}
    },
    {
        {29226,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29227,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29228,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29229,-1},
        {-1},
        {-1,-1},
        {20105,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29232,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29232,-1},
        {-1},
        {-1,-1},
        {29232,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24859,-1},
        {-1,-1,-1}
    },
    {
        {29234,-1},
        {-1},
        {-1,-1},
        {28858,-1,-1}
    },
    {
        {29234,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29236,-1},
        {-1},
        {-1,-1},
        {25899,-1,-1}
    },
    {
        {29237,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29237,-1},
        {-1},
        {-1,-1},
        {29237,-1,-1}
    },
    {
        {29238,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29238,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29240,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29240,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29241,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29241,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29242,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29242,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29243,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29243,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29244,-1},
        {-1},
        {-1,-1},
        {20430,-1,-1}
    },
    {
        {29245,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29246,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29246,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29247,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29248,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29248,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29250,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22665,-1,-1}
    },
    {
        {29254,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29254,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29255,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29255,-1},
        {-1},
        {-1,-1},
        {29255,-1,-1}
    },
    {
        {29255,-1},
        {-1},
        {-1,-1},
        {29255,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29256,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21028,29255},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29259,-1},
        {-1},
        {-1,-1},
        {31627,-1,-1}
    },
    {
        {29260,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29260,-1},
        {-1},
        {-1,-1},
        {29260,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31383,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29270,-1},
        {-1,-1,-1}
    },
    {
        {29266,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29267,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29267,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29260,24893},
        {-1,-1,-1}
    },
    {
        {29269,-1},
        {-1},
        {-1,-1},
        {31419,31379,-1}
    },
    {
        {29269,-1},
        {-1},
        {-1,-1},
        {31379,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29270,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29271,-1},
        {-1},
        {-1,-1},
        {29270,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29272,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29273,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29273,-1},
        {-1},
        {-1,-1},
        {29273,-1,-1}
    },
    {
        {29273,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29275,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29277,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29279,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29281,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29282,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29283,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22346,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27650,29339},
        {-1,-1,-1}
    },
    {
        {29287,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29289,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29281,29275},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26617,29287},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29294,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29295,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25277,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29298,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25287,-1},
        {-1,-1,-1}
    },
    {
        {29300,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29304,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29305,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29309,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29309,-1},
        {-1},
        {-1,-1},
        {29309,-1,-1}
    },
    {
        {29310,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29311,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29312,-1},
        {-1},
        {-1,-1},
        {29312,-1,-1}
    },
    {
        {29312,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29313,-1},
        {-1},
        {-1,-1},
        {29314,-1,-1}
    },
    {
        {29314,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29310,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21083,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29305,-1},
        {-1,-1,-1}
    },
    {
        {29319,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25458,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29491,-1},
        {-1,-1,-1}
    },
    {
        {29325,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29326,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34636,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31959,-1},
        {-1,-1,-1}
    },
    {
        {29330,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33467,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24894,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26381,-1},
        {-1,-1,-1}
    },
    {
        {29334,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29310,30606},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25705,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29096,29326},
        {-1,-1,-1}
    },
    {
        {29339,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22697,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25758,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25759,-1},
        {-1,-1,-1}
    },
    {
        {29344,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27348,-1},
        {-1,-1,-1}
    },
    {
        {29346,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29557,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25850,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29351,-1},
        {-1},
        {-1,-1},
        {29344,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39759,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22804,-1},
        {34369,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29356,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29357,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29358,-1},
        {-1},
        {-1,-1},
        {21451,-1,-1}
    },
    {
        {29358,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29359,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29359,-1},
        {-1},
        {-1,-1},
        {29359,-1,-1}
    },
    {
        {29360,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29361,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29362,-1},
        {-1},
        {-1,-1},
        {35962,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35961,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35961,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29366,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29369,-1},
        {-1},
        {-1,-1},
        {29494,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21536,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21564,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26514,-1},
        {-1,-1,-1}
    },
    {
        {29374,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35995,-1,-1}
    },
    {
        {29376,-1},
        {-1},
        {-1,-1},
        {29366,-1,-1}
    },
    {
        {29377,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29378,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29379,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29380,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29378,-1},
        {-1,-1,-1}
    },
    {
        {29382,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {64092,-1},
        {20855,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {64092,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29427,-1},
        {-1,-1,-1}
    },
    {
        {29388,-1},
        {-1},
        {-1,-1},
        {29481,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26553,-1},
        {-1,-1,-1}
    },
    {
        {29390,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25302,-1},
        {-1,-1,-1}
    },
    {
        {29392,-1},
        {-1},
        {-1,-1},
        {29392,-1,-1}
    },
    {
        {29392,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29392,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29427,-1},
        {-1,-1,-1}
    },
    {
        {29394,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25259,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29399,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29401,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29403,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {15974,29295},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29408,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29409,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29409,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29410,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25361,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29546,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24455,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29394,26613},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33296,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32104,29489},
        {29489,-1,-1}
    },
    {
        {29417,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20375,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20332,-1},
        {-1,-1,-1}
    },
    {
        {29420,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29421,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29427,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23592,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35944,-1},
        {-1,-1,-1}
    },
    {
        {29431,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29432,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29433,-1},
        {-1},
        {-1,-1},
        {29421,-1,-1}
    },
    {
        {29434,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29435,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29436,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29437,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25403,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26731,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24743,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24717,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26000,-1},
        {-1,-1,-1}
    },
    {
        {29447,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29738,-1},
        {-1,-1,-1}
    },
    {
        {29450,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21201,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29432,35981,-1}
    },
    {
        {-1,-1},
        {-1},
        {29557,29471},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32652,-1},
        {32652,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29736,-1},
        {-1,-1,-1}
    },
    {
        {29458,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29564,29550},
        {29564,-1,-1}
    },
    {
        {29462,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29463,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29464,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29465,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29467,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29468,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29468,-1},
        {-1},
        {-1,-1},
        {29468,-1,-1}
    },
    {
        {29469,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29471,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29435,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29477,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26963,-1},
        {-1,-1,-1}
    },
    {
        {29479,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29480,-1},
        {-1},
        {-1,-1},
        {29503,-1,-1}
    },
    {
        {29481,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29482,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29483,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29483,-1},
        {-1},
        {-1,-1},
        {29483,-1,-1}
    },
    {
        {29484,-1},
        {-1},
        {-1,-1},
        {34655,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25534,-1},
        {-1,-1,-1}
    },
    {
        {29486,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29487,-1},
        {-1},
        {-1,-1},
        {35986,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22865,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22865,-1},
        {-1,-1,-1}
    },
    {
        {29489,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29491,-1},
        {-1},
        {-1,-1},
        {35949,-1,-1}
    },
    {
        {29492,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29494,-1},
        {-1},
        {-1,-1},
        {29494,-1,-1}
    },
    {
        {29494,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29494,-1},
        {-1},
        {-1,-1},
        {29494,-1,-1}
    },
    {
        {29494,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29495,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29495,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26979,-1},
        {-1,-1,-1}
    },
    {
        {29497,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25615,-1},
        {-1,-1,-1}
    },
    {
        {29501,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29502,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29503,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25436,25628},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29802,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28304,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29508,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29509,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22071,-1},
        {22071,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27035,-1},
        {27035,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25654,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30347,22071},
        {22071,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27134,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29518,-1},
        {-1},
        {-1,-1},
        {22888,22892,-1}
    },
    {
        {29519,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29519,-1},
        {-1},
        {-1,-1},
        {35992,29519,-1}
    },
    {
        {29520,-1},
        {-1},
        {-1,-1},
        {40606,-1,-1}
    },
    {
        {29520,-1},
        {-1},
        {-1,-1},
        {40606,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28472,-1},
        {-1,-1,-1}
    },
    {
        {29522,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29522,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29522,-1},
        {29522,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22071,-1},
        {22071,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29527,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29852,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28654,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37883,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28511,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39509,-1},
        {39509,-1,-1}
    },
    {
        {29539,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29539,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29542,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29542,-1},
        {-1},
        {-1,-1},
        {29542,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29544,-1},
        {-1},
        {-1,-1},
        {29420,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29546,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29547,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29548,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29550,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29551,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29552,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29552,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29553,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29554,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29554,-1},
        {-1},
        {-1,-1},
        {29554,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23418,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27308,-1},
        {-1,-1,-1}
    },
    {
        {29557,-1},
        {-1},
        {-1,-1},
        {29471,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29489,-1},
        {-1,-1,-1}
    },
    {
        {29559,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29560,-1},
        {-1},
        {-1,-1},
        {29539,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29562,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29562,-1},
        {-1},
        {-1,-1},
        {29562,-1,-1}
    },
    {
        {29563,-1},
        {-1},
        {-1,-1},
        {29486,-1,-1}
    },
    {
        {29564,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25880,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35995,-1},
        {35995,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29489,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29569,-1},
        {-1},
        {-1,-1},
        {29547,-1,-1}
    },
    {
        {29569,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34308,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29572,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29573,-1},
        {-1},
        {-1,-1},
        {22937,-1,-1}
    },
    {
        {33586,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29575,-1},
        {-1},
        {-1,-1},
        {29575,-1,-1}
    },
    {
        {29575,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26053,-1},
        {-1,-1,-1}
    },
    {
        {29577,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29578,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29579,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29579,-1},
        {-1},
        {-1,-1},
        {29577,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29983,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21151,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29657,29596},
        {29657,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29588,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29589,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29590,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29592,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29596,-1},
        {-1},
        {-1,-1},
        {29657,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29589,-1},
        {-1,-1,-1}
    },
    {
        {29598,-1},
        {-1},
        {-1,-1},
        {30726,-1,-1}
    },
    {
        {29599,-1},
        {-1},
        {-1,-1},
        {29641,-1,-1}
    },
    {
        {29599,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29600,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26514,-1},
        {-1,-1,-1}
    },
    {
        {29602,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29589,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25308,-1},
        {-1,-1,-1}
    },
    {
        {29605,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29761,-1,-1}
    },
    {
        {29606,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29608,-1},
        {-1},
        {-1,-1},
        {29647,-1,-1}
    },
    {
        {29609,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29611,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22349,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29618,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29619,-1},
        {-1},
        {-1,-1},
        {29767,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29621,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22378,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29623,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22409,-1},
        {-1,-1,-1}
    },
    {
        {29625,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29885,-1},
        {29885,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29885,-1},
        {29885,-1,-1}
    },
    {
        {29627,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29628,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29637,-1},
        {-1,-1,-1}
    },
    {
        {29631,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29632,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29943,-1},
        {29943,-1,-1}
    },
    {
        {29634,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29642,32835},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29298,-1},
        {-1,-1,-1}
    },
    {
        {29637,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29640,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29641,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29642,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29642,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29642,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29643,-1},
        {-1},
        {-1,-1},
        {29705,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29645,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29646,-1},
        {-1},
        {-1,-1},
        {29645,-1,-1}
    },
    {
        {29646,-1},
        {-1},
        {-1,-1},
        {29645,-1,-1}
    },
    {
        {29647,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29754,-1},
        {29754,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25333,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29654,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27904,29669},
        {-1,-1,-1}
    },
    {
        {29657,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22240,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29662,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29578,-1},
        {29578,-1,-1}
    },
    {
        {29664,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29665,-1},
        {-1},
        {-1,-1},
        {29748,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29701,-1},
        {-1,-1,-1}
    },
    {
        {29667,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23453,23542,-1}
    },
    {
        {29669,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29669,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29671,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25342,-1},
        {-1,-1,-1}
    },
    {
        {29673,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29674,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29705,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29689,-1},
        {-1,-1,-1}
    },
    {
        {29677,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29677,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29678,-1},
        {-1},
        {-1,-1},
        {20329,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20332,-1},
        {-1,-1,-1}
    },
    {
        {29681,-1},
        {-1},
        {-1,-1},
        {29908,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29599,-1},
        {29641,29599,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24716,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30868,-1,-1}
    },
    {
        {29688,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29689,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29689,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29690,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29437,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29694,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29697,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29434,-1},
        {-1,-1,-1}
    },
    {
        {29699,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29701,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29702,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29703,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23386,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29705,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29706,-1},
        {-1},
        {-1,-1},
        {29706,-1,-1}
    },
    {
        {29706,-1},
        {-1},
        {-1,-1},
        {29706,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24725,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38517,-1},
        {38517,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29896,29827},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29795,-1,-1}
    },
    {
        {29715,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29697,25503},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38814,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25108,30430},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32183,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29641,-1},
        {29641,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22495,-1},
        {-1,-1,-1}
    },
    {
        {29722,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29723,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29641,-1},
        {29641,29599,-1}
    },
    {
        {-1,-1},
        {-1},
        {26395,32854},
        {32854,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22534,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29809,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29881,-1,-1}
    },
    {
        {29730,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29730,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29739,-1},
        {-1,-1,-1}
    },
    {
        {29732,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29733,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29734,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22538,-1},
        {-1,-1,-1}
    },
    {
        {29736,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29738,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29739,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29740,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29742,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29743,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29744,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29745,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29746,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29746,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29747,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29748,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29748,-1},
        {-1},
        {-1,-1},
        {29748,-1,-1}
    },
    {
        {29748,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29749,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29750,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29811,32652},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29753,-1},
        {-1},
        {-1,-1},
        {29748,-1,-1}
    },
    {
        {29754,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29898,29887},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22581,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29600,-1},
        {29600,-1,-1}
    },
    {
        {29759,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29760,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29761,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29761,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26979,-1},
        {-1,-1,-1}
    },
    {
        {29763,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29764,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22564,-1},
        {-1,-1,-1}
    },
    {
        {29766,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29767,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29767,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25545,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29641,-1},
        {29641,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29771,-1},
        {-1},
        {-1,-1},
        {29771,-1,-1}
    },
    {
        {29771,-1},
        {-1},
        {-1,-1},
        {29771,-1,-1}
    },
    {
        {29771,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30877,-1},
        {-1,-1,-1}
    },
    {
        {29773,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31359,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25539,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29778,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29779,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27006,-1},
        {-1,-1,-1}
    },
    {
        {29781,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29781,-1},
        {-1,-1,-1}
    },
    {
        {29783,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29706,-1},
        {29706,-1,-1}
    },
    {
        {29785,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29786,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29787,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29787,-1},
        {-1},
        {-1,-1},
        {29787,-1,-1}
    },
    {
        {29788,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29788,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29789,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29790,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29791,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29791,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29792,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29509,-1},
        {-1,-1,-1}
    },
    {
        {29794,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29794,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29795,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29795,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29796,-1},
        {-1},
        {-1,-1},
        {29814,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29857,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29801,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29802,-1},
        {-1},
        {-1,-1},
        {30908,-1,-1}
    },
    {
        {29803,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26066,37775},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26066,37775},
        {-1,-1,-1}
    },
    {
        {29805,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29807,-1},
        {-1},
        {-1,-1},
        {29701,-1,-1}
    },
    {
        {29807,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29808,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29809,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29809,-1},
        {-1},
        {-1,-1},
        {29809,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29811,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29608,-1},
        {29647,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25620,-1},
        {-1,-1,-1}
    },
    {
        {29814,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34768,29867},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29736,-1},
        {29736,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22652,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20657,-1},
        {-1,-1,-1}
    },
    {
        {29822,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29822,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27593,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29825,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29825,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29827,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29827,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29855,-1},
        {29855,-1,-1}
    },
    {
        {29829,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29829,-1},
        {-1},
        {-1,-1},
        {29795,-1,-1}
    },
    {
        {29830,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29831,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29832,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29833,-1},
        {-1},
        {-1,-1},
        {29833,-1,-1}
    },
    {
        {29833,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29835,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29835,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {144953,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29840,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29842,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29703,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22679,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29865,-1},
        {29865,-1,-1}
    },
    {
        {29847,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29848,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29848,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29849,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29578,-1},
        {-1,-1,-1}
    },
    {
        {29852,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29808,-1},
        {29808,-1,-1}
    },
    {
        {29854,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29855,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29856,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29857,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29857,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29858,-1},
        {-1},
        {-1,-1},
        {29792,-1,-1}
    },
    {
        {29859,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29859,-1},
        {-1},
        {-1,-1},
        {29859,-1,-1}
    },
    {
        {29859,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29861,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29861,-1},
        {-1},
        {-1,-1},
        {29861,-1,-1}
    },
    {
        {29862,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29862,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29863,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29864,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29865,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29865,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29866,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29867,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29743,-1},
        {29743,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29872,-1},
        {-1},
        {-1,-1},
        {29872,-1,-1}
    },
    {
        {29872,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29795,-1},
        {-1,-1,-1}
    },
    {
        {29877,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22743,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25975,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29602,-1},
        {-1,-1,-1}
    },
    {
        {29881,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29882,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29883,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20758,-1},
        {-1,-1,-1}
    },
    {
        {29885,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25824,-1},
        {-1,-1,-1}
    },
    {
        {29887,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22738,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27345,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29346,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25853,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25854,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29896,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29897,-1},
        {-1},
        {-1,-1},
        {29914,-1,-1}
    },
    {
        {29898,-1},
        {-1},
        {-1,-1},
        {29898,-1,-1}
    },
    {
        {29898,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29898,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25855,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25855,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25855,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29808,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38568,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25078,-1},
        {-1,-1,-1}
    },
    {
        {29903,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29903,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22754,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29908,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24396,29885},
        {-1,-1,-1}
    },
    {
        {29910,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29898,-1},
        {-1,-1,-1}
    },
    {
        {29912,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29912,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29881,29182},
        {-1,-1,-1}
    },
    {
        {29914,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29914,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29916,-1},
        {-1},
        {-1,-1},
        {29916,-1,-1}
    },
    {
        {29916,-1},
        {-1},
        {-1,-1},
        {29916,-1,-1}
    },
    {
        {29916,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29918,-1},
        {-1},
        {-1,-1},
        {29918,-1,-1}
    },
    {
        {29918,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29920,-1},
        {-1},
        {-1,-1},
        {29920,-1,-1}
    },
    {
        {29920,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29920,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29392,22519},
        {-1,-1,-1}
    },
    {
        {29922,-1},
        {-1},
        {-1,-1},
        {29922,-1,-1}
    },
    {
        {29922,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29923,-1},
        {-1},
        {-1,-1},
        {24321,-1,-1}
    },
    {
        {29923,-1},
        {-1},
        {-1,-1},
        {24321,-1,-1}
    },
    {
        {29923,-1},
        {-1},
        {-1,-1},
        {24321,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29926,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29927,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32568,-1,-1}
    },
    {
        {29929,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29931,-1},
        {-1},
        {-1,-1},
        {30406,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30745,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29934,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29934,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29935,-1},
        {-1},
        {-1,-1},
        {29964,-1,-1}
    },
    {
        {29935,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29936,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29936,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29937,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29938,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30745,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21488,-1},
        {-1,-1,-1}
    },
    {
        {29942,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29943,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29944,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29953,-1},
        {-1,-1,-1}
    },
    {
        {29946,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29934,29982},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31376,-1},
        {-1,-1,-1}
    },
    {
        {29951,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29956,136775},
        {-1,-1,-1}
    },
    {
        {29953,-1},
        {-1},
        {-1,-1},
        {29942,-1,-1}
    },
    {
        {29953,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32745,29975},
        {-1,-1,-1}
    },
    {
        {29955,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29956,-1},
        {-1},
        {-1,-1},
        {29956,-1,-1}
    },
    {
        {29956,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29956,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29956,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29957,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29943,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29943,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29943,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32588,-1},
        {32579,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29966,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27465,29942},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25976,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25973,-1},
        {-1,-1,-1}
    },
    {
        {29964,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29964,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29965,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29965,-1},
        {-1},
        {-1,-1},
        {29965,-1,-1}
    },
    {
        {29965,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29965,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29966,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29966,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30967,-1},
        {-1,-1,-1}
    },
    {
        {29969,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29969,-1},
        {-1},
        {-1,-1},
        {29969,-1,-1}
    },
    {
        {29969,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29969,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29971,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20747,-1},
        {-1,-1,-1}
    },
    {
        {29973,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29973,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29973,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32588,-1,-1}
    },
    {
        {29975,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29976,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29978,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29980,-1,-1}
    },
    {
        {29980,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24426,-1},
        {-1,-1,-1}
    },
    {
        {29982,-1},
        {-1},
        {-1,-1},
        {22039,-1,-1}
    },
    {
        {29983,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29984,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29987,-1},
        {-1},
        {29987,-1},
        {29987,-1,-1}
    },
    {
        {29987,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29988,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29989,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29990,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29990,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29992,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29994,-1},
        {-1},
        {-1,-1},
        {35282,-1,-1}
    },
    {
        {29995,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29996,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26820,-1},
        {-1,-1,-1}
    },
    {
        {29999,-1},
        {-1},
        {-1,-1},
        {23527,-1,-1}
    },
    {
        {30000,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30001,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30002,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30003,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30001,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30007,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30008,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30010,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30011,-1},
        {-1},
        {-1,-1},
        {30011,30059,-1}
    },
    {
        {30011,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30012,-1},
        {-1},
        {-1,-1},
        {30010,-1,-1}
    },
    {
        {30013,-1},
        {-1},
        {-1,-1},
        {30030,-1,-1}
    },
    {
        {30014,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30014,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30016,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30016,-1},
        {30016,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30045,150550},
        {30045,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30020,-1},
        {-1},
        {-1,-1},
        {30041,-1,-1}
    },
    {
        {30022,-1},
        {-1},
        {-1,-1},
        {30045,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30024,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30025,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30026,-1},
        {-1},
        {-1,-1},
        {32789,-1,-1}
    },
    {
        {30027,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30028,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30029,-1},
        {-1},
        {-1,-1},
        {30028,-1,-1}
    },
    {
        {30030,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30031,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30032,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30033,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30034,-1},
        {-1},
        {-1,-1},
        {30045,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30036,-1},
        {-1},
        {-1,-1},
        {30036,-1,-1}
    },
    {
        {30036,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30086,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21617,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20873,30052},
        {-1,-1,-1}
    },
    {
        {30041,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30041,-1},
        {-1},
        {-1,-1},
        {30041,-1,-1}
    },
    {
        {30042,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30043,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30044,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30045,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30045,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30048,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22419,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22419,-1,-1}
    },
    {
        {30050,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31572,-1},
        {31572,-1,-1}
    },
    {
        {30052,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30053,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30054,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30055,-1},
        {-1},
        {-1,-1},
        {30053,-1,-1}
    },
    {
        {30057,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30058,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30059,-1},
        {-1},
        {-1,-1},
        {30011,-1,-1}
    },
    {
        {30060,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30061,-1},
        {-1},
        {-1,-1},
        {30060,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30045,30022},
        {30045,-1,-1}
    },
    {
        {30063,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30064,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30064,-1},
        {-1},
        {-1,-1},
        {30064,-1,-1}
    },
    {
        {30065,-1},
        {-1},
        {-1,-1},
        {30041,-1,-1}
    },
    {
        {30066,-1},
        {-1},
        {-1,-1},
        {30060,-1,-1}
    },
    {
        {30067,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30068,-1},
        {-1},
        {-1,-1},
        {30087,-1,-1}
    },
    {
        {30059,-1},
        {-1},
        {-1,-1},
        {30011,30059,-1}
    },
    {
        {30070,-1},
        {-1},
        {-1,-1},
        {24403,-1,-1}
    },
    {
        {30071,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30072,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30073,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30074,-1},
        {-1},
        {-1,-1},
        {30086,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26274,-1,-1}
    },
    {
        {30077,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30078,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30079,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30571,30014},
        {-1,-1,-1}
    },
    {
        {30081,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30082,-1},
        {-1},
        {-1,-1},
        {30067,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30077,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30077,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36692,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30074,30086,-1}
    },
    {
        {30086,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30087,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21103,-1},
        {21103,-1,-1}
    },
    {
        {30089,-1},
        {-1},
        {-1,-1},
        {30067,-1,-1}
    },
    {
        {30090,-1},
        {-1},
        {-1,-1},
        {30067,-1,-1}
    },
    {
        {30091,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30092,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30094,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30095,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30097,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30098,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30100,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32923,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30106,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24220,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30109,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {16242,-1},
        {-1,-1,-1}
    },
    {
        {30114,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30115,-1},
        {-1},
        {-1,-1},
        {32940,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30117,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30142,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30229,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30209,-1},
        {30209,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30201,-1},
        {30201,-1,-1}
    },
    {
        {30123,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30128,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30129,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30129,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30130,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30131,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30133,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30135,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30136,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30137,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30168,21482},
        {-1,-1,-1}
    },
    {
        {30140,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30140,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30141,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30142,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30143,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30144,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30146,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30147,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30149,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30150,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30151,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30154,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30154,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30140,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24683,-1,-1}
    },
    {
        {30157,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30158,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30158,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30159,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34456,34516},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34456,34516},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30137,-1},
        {-1,-1,-1}
    },
    {
        {30162,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30163,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30164,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30165,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30168,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30169,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30171,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30173,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30174,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30175,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30176,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30178,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30179,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30180,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30168,20823},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20425,30174},
        {-1,-1,-1}
    },
    {
        {30183,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30185,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30188,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30190,-1},
        {-1},
        {-1,-1},
        {33081,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30192,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30193,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30194,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30195,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30196,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24760,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20494,30143},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30235,-1},
        {-1,-1,-1}
    },
    {
        {30201,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30202,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30135,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30135,-1},
        {-1,-1,-1}
    },
    {
        {30204,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20629,-1},
        {-1,-1,-1}
    },
    {
        {30206,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30207,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30208,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30208,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30209,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30210,-1},
        {-1},
        {-1,-1},
        {21854,-1,-1}
    },
    {
        {30211,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30211,-1},
        {-1},
        {-1,-1},
        {30211,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30215,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30216,-1},
        {-1},
        {-1,-1},
        {30216,-1,-1}
    },
    {
        {30216,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30217,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30217,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30219,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30221,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28251,-1},
        {-1,-1,-1}
    },
    {
        {30223,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24254,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21692,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30227,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33510,-1},
        {33510,-1,-1}
    },
    {
        {30229,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30230,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30230,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30232,-1},
        {-1},
        {-1,-1},
        {30267,-1,-1}
    },
    {
        {30233,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21413,-1},
        {-1,-1,-1}
    },
    {
        {30235,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30235,-1},
        {-1},
        {-1,-1},
        {30235,-1,-1}
    },
    {
        {30236,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30238,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30239,-1},
        {-1},
        {-1,-1},
        {30239,-1,-1}
    },
    {
        {30239,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30240,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30241,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30242,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30242,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30244,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30245,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30185,-1},
        {-1},
        {-1,-1},
        {30185,30246,-1}
    },
    {
        {30246,-1},
        {-1},
        {-1,-1},
        {30185,-1,-1}
    },
    {
        {30247,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30247,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30248,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33081,-1},
        {33081,30190,-1}
    },
    {
        {30253,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20650,30259},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26063,-1},
        {-1,-1,-1}
    },
    {
        {30256,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30238,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30259,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30260,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30260,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30261,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30240,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30267,-1,-1}
    },
    {
        {30267,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30268,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30268,-1},
        {-1},
        {-1,-1},
        {30268,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21220,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {16302,-1,-1}
    },
    {
        {30273,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30273,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30274,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30275,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24980,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30244,-1,-1}
    },
    {
        {30278,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30279,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30279,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30280,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30281,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30283,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30284,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30279,-1},
        {-1},
        {-1,-1},
        {30279,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33215,-1},
        {-1,-1,-1}
    },
    {
        {30290,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30290,-1},
        {-1},
        {-1,-1},
        {30290,-1,-1}
    },
    {
        {30290,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30309,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24847,-1},
        {-1,-1,-1}
    },
    {
        {30293,-1},
        {-1},
        {-1,-1},
        {30320,-1,-1}
    },
    {
        {30294,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30296,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30296,-1},
        {-1},
        {-1,-1},
        {30296,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28601,30241},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20956,-1},
        {-1,-1,-1}
    },
    {
        {30300,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28639,-1},
        {-1,-1,-1}
    },
    {
        {30305,-1},
        {-1},
        {-1,-1},
        {30196,-1,-1}
    },
    {
        {30306,-1},
        {-1},
        {-1,-1},
        {30162,-1,-1}
    },
    {
        {30306,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30306,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30308,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30309,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30309,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30311,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30311,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30312,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30313,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30313,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30314,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30316,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30317,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30318,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30319,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30320,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30321,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30321,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30322,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30322,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30322,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30322,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30326,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30328,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30332,-1},
        {-1,-1,-1}
    },
    {
        {30330,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30331,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30332,-1},
        {-1},
        {-1,-1},
        {30330,-1,-1}
    },
    {
        {30333,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30334,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30336,-1},
        {-1},
        {-1,-1},
        {39321,-1,-1}
    },
    {
        {30337,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30336,-1},
        {-1},
        {30336,30337},
        {30337,-1,-1}
    },
    {
        {30336,-1},
        {-1},
        {30336,30337},
        {-1,-1,-1}
    },
    {
        {30336,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30339,-1},
        {-1},
        {-1,-1},
        {35980,-1,-1}
    },
    {
        {30340,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30340,-1},
        {-1},
        {-1,-1},
        {30340,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30342,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30343,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30344,-1},
        {-1},
        {-1,-1},
        {27512,24112,-1}
    },
    {
        {-1,-1},
        {-1},
        {29628,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26148,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30347,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21363,21373,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30350,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30350,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33020,-1},
        {-1,-1,-1}
    },
    {
        {30352,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21734,-1},
        {-1,-1,-1}
    },
    {
        {30355,-1},
        {-1},
        {-1,-1},
        {30355,-1,-1}
    },
    {
        {30355,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30357,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30358,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26221,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26228,-1},
        {-1,-1,-1}
    },
    {
        {30361,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30362,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30363,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30364,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30365,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30366,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30366,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30368,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30369,-1},
        {-1},
        {-1,-1},
        {30366,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30369,-1},
        {-1},
        {-1,-1},
        {30366,-1,-1}
    },
    {
        {30374,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30375,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30375,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30376,-1},
        {-1},
        {-1,-1},
        {26143,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30365,-1},
        {-1,-1,-1}
    },
    {
        {30378,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33243,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30382,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30383,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30384,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30384,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27628,-1},
        {27628,-1,-1}
    },
    {
        {30388,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30388,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30390,-1},
        {-1,-1,-1}
    },
    {
        {30390,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30390,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30391,-1},
        {-1},
        {-1,-1},
        {40723,-1,-1}
    },
    {
        {30391,-1},
        {-1},
        {-1,-1},
        {40723,-1,-1}
    },
    {
        {30392,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30393,-1},
        {-1},
        {-1,-1},
        {30392,-1,-1}
    },
    {
        {30394,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30390,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30399,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30403,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30408,-1,-1}
    },
    {
        {30402,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30403,-1},
        {-1},
        {-1,-1},
        {26479,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30402,-1},
        {-1,-1,-1}
    },
    {
        {30405,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30406,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30406,-1},
        {-1},
        {-1,-1},
        {30406,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30413,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30413,-1},
        {30413,-1,-1}
    },
    {
        {30408,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30408,-1,-1}
    },
    {
        {30409,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30410,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37474,-1,-1}
    },
    {
        {30412,-1},
        {-1},
        {-1,-1},
        {30871,-1,-1}
    },
    {
        {30413,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30414,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30418,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30402,-1},
        {-1,-1,-1}
    },
    {
        {30420,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30420,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30428,-1},
        {-1,-1,-1}
    },
    {
        {30422,-1},
        {-1},
        {-1,-1},
        {33995,-1,-1}
    },
    {
        {30423,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31776,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35032,-1},
        {-1,-1,-1}
    },
    {
        {30427,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30427,-1},
        {-1},
        {-1,-1},
        {30427,-1,-1}
    },
    {
        {30428,-1},
        {-1},
        {-1,-1},
        {30423,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28425,-1},
        {-1,-1,-1}
    },
    {
        {30430,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30431,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30431,-1},
        {-1},
        {-1,-1},
        {30431,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30433,-1},
        {-1},
        {-1,-1},
        {23613,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30435,-1},
        {-1,-1,-1}
    },
    {
        {30435,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30435,-1},
        {-1},
        {-1,-1},
        {30435,-1,-1}
    },
    {
        {30436,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30437,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30438,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30439,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30440,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22666,-1},
        {-1,-1,-1}
    },
    {
        {30442,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30438,-1},
        {30438,-1,-1}
    },
    {
        {30444,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25147,32159},
        {-1,-1,-1}
    },
    {
        {30446,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32821,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30449,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30450,-1},
        {-1},
        {-1,-1},
        {30450,-1,-1}
    },
    {
        {30450,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30450,-1},
        {30450,-1,-1}
    },
    {
        {30452,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30452,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21507,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26112,-1},
        {-1,-1,-1}
    },
    {
        {30456,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26133,-1},
        {-1,-1,-1}
    },
    {
        {30459,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30459,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30460,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30474,-1},
        {-1,-1,-1}
    },
    {
        {30462,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26108,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26379,26126,-1}
    },
    {
        {30465,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {64097,35222},
        {35222,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30468,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30468,-1},
        {-1},
        {-1,-1},
        {30468,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26116,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30471,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30472,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30473,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30474,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30475,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30476,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64097,35222,-1}
    },
    {
        {-1,-1},
        {-1},
        {32998,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30468,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31365,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30487,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24910,-1},
        {24892,24910,-1}
    },
    {
        {30489,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30490,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30491,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36015,-1},
        {-1,-1,-1}
    },
    {
        {30494,-1},
        {-1},
        {-1,-1},
        {30495,-1,-1}
    },
    {
        {30494,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30494,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30495,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30495,-1},
        {-1},
        {-1,-1},
        {30494,30495,-1}
    },
    {
        {30495,-1},
        {-1},
        {-1,-1},
        {30494,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30496,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {64097,35222},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32965,-1},
        {-1,-1,-1}
    },
    {
        {30500,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30501,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30502,-1},
        {-1},
        {-1,-1},
        {30501,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26157,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30505,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32982,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32982,-1,-1}
    },
    {
        {30509,-1},
        {-1},
        {-1,-1},
        {30566,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33013,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30535,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21665,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26157,-1},
        {-1,-1,-1}
    },
    {
        {30516,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30517,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30518,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30519,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30519,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30520,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30522,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21673,-1},
        {-1,-1,-1}
    },
    {
        {30524,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34886,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30528,-1},
        {-1},
        {-1,-1},
        {33879,-1,-1}
    },
    {
        {30528,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30556,-1},
        {-1},
        {20105,29229},
        {30556,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30473,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33063,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30471,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30534,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30535,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33054,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20420,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26217,-1},
        {-1,-1,-1}
    },
    {
        {30541,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30542,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33094,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23832,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33109,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30552,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21811,-1},
        {-1,-1,-1}
    },
    {
        {30554,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30555,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30555,-1},
        {-1},
        {-1,-1},
        {30555,-1,-1}
    },
    {
        {30556,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30592,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30559,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30560,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30561,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30562,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30563,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30565,-1,-1}
    },
    {
        {30565,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30566,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30568,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {13736,-1},
        {-1,-1,-1}
    },
    {
        {30570,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30571,-1},
        {-1},
        {-1,-1},
        {16441,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26221,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26221,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33127,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26247,-1},
        {-1,-1,-1}
    },
    {
        {30578,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26249,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26265,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33137,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23883,-1},
        {-1,-1,-1}
    },
    {
        {30585,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30585,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30586,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26244,-1},
        {-1,-1,-1}
    },
    {
        {30588,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30589,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30590,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30591,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30592,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35215,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20240,25765},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21886,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24384,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33157,-1},
        {-1,-1,-1}
    },
    {
        {30603,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30603,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30606,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30606,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30606,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30552,36996},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26230,-1},
        {-1,-1,-1}
    },
    {
        {30609,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30612,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40264,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30487,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33180,-1},
        {-1,-1,-1}
    },
    {
        {30618,-1},
        {-1},
        {-1,-1},
        {30636,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27141,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29827,-1},
        {-1,-1,-1}
    },
    {
        {30622,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30623,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30624,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27131,-1},
        {-1,-1,-1}
    },
    {
        {30626,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30626,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30628,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30629,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30629,-1},
        {-1},
        {-1,-1},
        {30629,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26303,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24158,-1},
        {-1,-1,-1}
    },
    {
        {30633,-1},
        {-1},
        {-1,-1},
        {30682,-1,-1}
    },
    {
        {30634,-1},
        {-1},
        {-1,-1},
        {30624,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36105,-1},
        {-1,-1,-1}
    },
    {
        {30636,-1},
        {-1},
        {-1,-1},
        {30636,-1,-1}
    },
    {
        {30636,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30636,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30637,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30640,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30643,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30643,-1},
        {-1},
        {-1,-1},
        {30643,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30646,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21899,-1},
        {-1,-1,-1}
    },
    {
        {30649,-1},
        {-1},
        {-1,-1},
        {26326,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33214,-1},
        {-1,-1,-1}
    },
    {
        {30651,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30652,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30653,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29031,-1,-1}
    },
    {
        {30655,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24494,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24833,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26331,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26332,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30663,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30663,-1},
        {-1},
        {-1,-1},
        {30663,-1,-1}
    },
    {
        {30663,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39024,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26336,-1},
        {26336,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26336,-1},
        {26336,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30669,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22800,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33243,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30626,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30626,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26407,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24931,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26348,-1},
        {-1,-1,-1}
    },
    {
        {30679,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30679,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25897,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30682,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30683,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30684,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30684,-1},
        {-1,-1,-1}
    },
    {
        {30686,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30687,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30688,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25899,-1},
        {-1,-1,-1}
    },
    {
        {30690,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30691,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30692,-1},
        {-1},
        {-1,-1},
        {30695,-1,-1}
    },
    {
        {30693,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30694,-1},
        {-1},
        {-1,-1},
        {20399,-1,-1}
    },
    {
        {30695,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27520,-1},
        {-1,-1,-1}
    },
    {
        {30697,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30697,-1},
        {-1},
        {-1,-1},
        {30697,-1,-1}
    },
    {
        {30698,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30700,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30701,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30702,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30703,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30704,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25850,-1},
        {-1,-1,-1}
    },
    {
        {30707,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30708,-1},
        {-1},
        {-1,-1},
        {30855,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30732,-1},
        {30732,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23676,30740},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30932,-1},
        {-1,-1,-1}
    },
    {
        {30715,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30716,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21167,-1},
        {-1,-1,-1}
    },
    {
        {30722,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22349,30799},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30733,29606},
        {-1,-1,-1}
    },
    {
        {30725,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30726,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29641,29599},
        {29641,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21571,-1},
        {-1,-1,-1}
    },
    {
        {30729,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30729,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21549,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32827,-1},
        {-1,-1,-1}
    },
    {
        {30732,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30733,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21545,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21560,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30738,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27521,-1},
        {-1,-1,-1}
    },
    {
        {30740,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30741,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30745,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30787,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26152,-1},
        {-1,-1,-1}
    },
    {
        {30752,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30753,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30757,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30758,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30759,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38440,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21640,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22982,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26609,-1},
        {-1,-1,-1}
    },
    {
        {30764,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30765,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30770,-1},
        {-1},
        {-1,-1},
        {30770,-1,-1}
    },
    {
        {30770,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30707,30759},
        {-1,-1,-1}
    },
    {
        {30772,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29632,-1},
        {29632,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21277,-1},
        {-1,-1,-1}
    },
    {
        {30778,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30783,-1},
        {-1},
        {-1,-1},
        {37489,37979,31014}
    },
    {
        {-1,-1},
        {-1},
        {37523,-1},
        {37523,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30787,-1},
        {-1},
        {-1,-1},
        {26417,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30789,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21704,30834},
        {-1,-1,-1}
    },
    {
        {30791,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24459,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30920,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31001,-1},
        {31001,30861,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30798,-1},
        {-1},
        {-1,-1},
        {30798,-1,-1}
    },
    {
        {30798,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30799,-1},
        {-1},
        {-1,-1},
        {30740,-1,-1}
    },
    {
        {30740,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27934,-1},
        {-1,-1,-1}
    },
    {
        {30801,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30813,-1},
        {-1},
        {-1,-1},
        {30813,-1,-1}
    },
    {
        {30813,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34573,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30820,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34444,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21730,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34553,-1},
        {-1,-1,-1}
    },
    {
        {30824,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21746,-1},
        {-1,-1,-1}
    },
    {
        {30826,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30827,-1},
        {-1},
        {-1,-1},
        {30827,-1,-1}
    },
    {
        {30827,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30828,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30828,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30830,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30830,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30831,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23057,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30834,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30836,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30836,-1},
        {-1},
        {-1,-1},
        {30836,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23736,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34585,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37847,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37847,-1},
        {-1,-1,-1}
    },
    {
        {30844,-1},
        {-1},
        {-1,-1},
        {30844,-1,-1}
    },
    {
        {30844,-1},
        {-1},
        {-1,-1},
        {30844,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21813,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21836,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29732,-1},
        {-1,-1,-1}
    },
    {
        {30849,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29742,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26228,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31259,30791},
        {-1,-1,-1}
    },
    {
        {30854,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30855,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29641,-1},
        {29641,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38613,-1},
        {-1,-1,-1}
    },
    {
        {30858,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23714,30869},
        {-1,-1,-1}
    },
    {
        {30860,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30860,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30861,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30862,-1},
        {-1},
        {-1,-1},
        {30741,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23834,-1},
        {-1,-1,-1}
    },
    {
        {30865,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23839,-1},
        {23839,-1,-1}
    },
    {
        {30867,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30868,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30869,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23832,-1},
        {-1,-1,-1}
    },
    {
        {30871,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30791,-1},
        {-1,-1,-1}
    },
    {
        {30874,-1},
        {-1},
        {-1,-1},
        {22521,-1,-1}
    },
    {
        {30877,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30878,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30879,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30855,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30883,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30883,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30884,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30884,-1},
        {-1},
        {-1,-1},
        {30884,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21888,-1},
        {-1,-1,-1}
    },
    {
        {30887,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30889,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30890,-1},
        {-1},
        {-1,-1},
        {30759,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27573,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30893,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38532,-1},
        {-1,-1,-1}
    },
    {
        {30895,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30896,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30896,-1},
        {-1},
        {-1,-1},
        {25533,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40572,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30901,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30906,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30906,-1},
        {-1},
        {-1,-1},
        {30906,-1,-1}
    },
    {
        {30907,-1},
        {-1},
        {-1,-1},
        {30906,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30908,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30889,-1,-1}
    },
    {
        {30910,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36870,-1},
        {-1,-1,-1}
    },
    {
        {30913,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30913,-1},
        {-1},
        {-1,-1},
        {30913,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30827,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33157,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30917,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30917,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30918,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30738,-1,-1}
    },
    {
        {30920,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30922,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30923,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30924,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30924,-1},
        {-1},
        {-1,-1},
        {30924,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30606,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30606,-1},
        {-1,-1,-1}
    },
    {
        {30926,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24265,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30928,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30929,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30932,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30932,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30932,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30932,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30933,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30938,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30732,-1},
        {30732,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30973,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30944,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30945,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30820,29245},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22667,-1},
        {-1,-1,-1}
    },
    {
        {30948,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30922,-1},
        {-1,-1,-1}
    },
    {
        {30950,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30951,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30952,-1},
        {-1},
        {-1,-1},
        {30952,-1,-1}
    },
    {
        {30952,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30732,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {17893,-1},
        {-1,-1,-1}
    },
    {
        {30956,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28440,33059},
        {-1,-1,-1}
    },
    {
        {30959,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30959,-1},
        {-1},
        {-1,-1},
        {30959,-1,-1}
    },
    {
        {30959,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23959,-1},
        {-1,-1,-1}
    },
    {
        {30962,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30964,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30976,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30967,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30967,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30855,-1},
        {30708,30855,-1}
    },
    {
        {30969,-1},
        {-1},
        {-1,-1},
        {30969,-1,-1}
    },
    {
        {30969,-1},
        {-1},
        {-1,-1},
        {30969,-1,-1}
    },
    {
        {30970,-1},
        {-1},
        {-1,-1},
        {37489,37979,-1}
    },
    {
        {30971,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30973,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30974,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30976,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30977,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38570,38522,-1}
    },
    {
        {30983,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30983,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22684,-1},
        {22684,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28608,30973},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {152622,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30883,-1},
        {30883,-1,-1}
    },
    {
        {30990,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30992,-1},
        {-1},
        {-1,-1},
        {30830,-1,-1}
    },
    {
        {30993,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30994,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27565,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38713,-1},
        {38713,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21128,38713},
        {38713,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27567,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38576,-1},
        {-1,-1,-1}
    },
    {
        {31001,-1},
        {-1},
        {-1,-1},
        {30861,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30933,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30933,-1,-1}
    },
    {
        {31003,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31004,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30877,-1},
        {30877,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30770,31022,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30913,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30913,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24173,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34081,-1},
        {-1,-1,-1}
    },
    {
        {31013,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31014,-1},
        {-1},
        {-1,-1},
        {37489,37979,-1}
    },
    {
        {31014,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31018,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31018,-1},
        {-1},
        {-1,-1},
        {31018,-1,-1}
    },
    {
        {31019,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31020,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31022,-1},
        {-1},
        {-1,-1},
        {30770,-1,-1}
    },
    {
        {31022,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31024,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31025,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31025,-1},
        {-1},
        {-1,-1},
        {31025,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31025,-1},
        {31025,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31025,-1},
        {31025,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31028,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31028,-1},
        {-1},
        {-1,-1},
        {31028,-1,-1}
    },
    {
        {31028,-1},
        {-1},
        {-1,-1},
        {31028,-1,-1}
    },
    {
        {31028,-1},
        {-1},
        {-1,-1},
        {31028,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32592,32566},
        {32592,32566,-1}
    },
    {
        {-1,-1},
        {-1},
        {32592,32566},
        {32592,32566,-1}
    },
    {
        {-1,-1},
        {-1},
        {31003,34253},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36106,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24022,-1},
        {-1,-1,-1}
    },
    {
        {31034,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31035,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31036,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31038,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31040,-1},
        {-1},
        {-1,-1},
        {31040,-1,-1}
    },
    {
        {31040,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31040,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31040,-1},
        {-1},
        {-1,-1},
        {31040,-1,-1}
    },
    {
        {31041,-1},
        {-1},
        {-1,-1},
        {31041,-1,-1}
    },
    {
        {31041,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31045,-1},
        {-1},
        {-1,-1},
        {31045,-1,-1}
    },
    {
        {31045,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31046,-1},
        {-1},
        {-1,-1},
        {31046,-1,-1}
    },
    {
        {31046,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31047,-1},
        {-1},
        {-1,-1},
        {31047,-1,-1}
    },
    {
        {31047,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31048,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31049,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31050,-1},
        {-1},
        {-1,-1},
        {31050,-1,-1}
    },
    {
        {31050,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27798,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27798,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31056,64079,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31056,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31209,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32057,-1},
        {-1,-1,-1}
    },
    {
        {31059,-1},
        {-1},
        {-1,-1},
        {31059,-1,-1}
    },
    {
        {31059,-1},
        {-1},
        {-1,-1},
        {31059,-1,-1}
    },
    {
        {31059,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31061,-1},
        {-1},
        {-1,-1},
        {31192,-1,-1}
    },
    {
        {31061,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31062,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31063,-1},
        {-1},
        {-1,-1},
        {31063,-1,-1}
    },
    {
        {31063,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31064,-1},
        {-1},
        {-1,-1},
        {31064,-1,-1}
    },
    {
        {31064,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31064,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31064,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31066,-1},
        {-1},
        {-1,-1},
        {31066,-1,-1}
    },
    {
        {31066,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31067,-1},
        {-1},
        {-1,-1},
        {31067,-1,-1}
    },
    {
        {31067,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31068,-1},
        {-1},
        {-1,-1},
        {31068,-1,-1}
    },
    {
        {31068,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31069,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31070,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31071,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31072,-1},
        {-1},
        {-1,-1},
        {31072,-1,-1}
    },
    {
        {31072,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26612,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26612,-1},
        {-1,-1,-1}
    },
    {
        {31074,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31074,-1},
        {-1},
        {-1,-1},
        {31152,31074,-1}
    },
    {
        {31074,-1},
        {-1},
        {-1,-1},
        {31152,31074,-1}
    },
    {
        {31074,-1},
        {-1},
        {-1,-1},
        {31152,-1,-1}
    },
    {
        {31074,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26053,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26053,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26665,-1},
        {-1,-1,-1}
    },
    {
        {31077,-1},
        {-1},
        {64026,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31079,-1},
        {-1},
        {-1,-1},
        {31079,-1,-1}
    },
    {
        {31079,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31080,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26693,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31063,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31063,-1},
        {-1,-1,-1}
    },
    {
        {31085,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64081,31069,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31090,-1},
        {-1},
        {-1,-1},
        {31090,-1,-1}
    },
    {
        {31090,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33060,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33060,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35027,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35027,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31128,-1},
        {31128,-1,-1}
    },
    {
        {31095,-1},
        {-1},
        {-1,-1},
        {31153,-1,-1}
    },
    {
        {31097,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31098,-1},
        {-1},
        {-1,-1},
        {31098,-1,-1}
    },
    {
        {31098,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26890,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31103,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31103,-1},
        {-1},
        {-1,-1},
        {31108,31103,-1}
    },
    {
        {31104,-1},
        {-1},
        {-1,-1},
        {31263,-1,-1}
    },
    {
        {31104,-1},
        {-1},
        {-1,-1},
        {31263,-1,-1}
    },
    {
        {31105,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26893,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31108,-1},
        {-1},
        {-1,-1},
        {31103,-1,-1}
    },
    {
        {31108,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31108,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31109,-1},
        {-1},
        {-1,-1},
        {31146,31109,-1}
    },
    {
        {31109,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35084,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31114,-1},
        {-1},
        {-1,-1},
        {31114,-1,-1}
    },
    {
        {31114,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31114,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31114,-1},
        {-1},
        {-1,-1},
        {31114,-1,-1}
    },
    {
        {31114,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31117,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31118,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31119,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {64027,-1},
        {-1},
        {-1,-1},
        {64027,31119,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32233,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31124,-1},
        {-1},
        {-1,-1},
        {31124,-1,-1}
    },
    {
        {31124,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31125,-1},
        {-1},
        {-1,-1},
        {31125,-1,-1}
    },
    {
        {31125,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31125,-1},
        {-1},
        {-1,-1},
        {31125,-1,-1}
    },
    {
        {31125,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31125,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31126,-1},
        {-1},
        {-1,-1},
        {31126,-1,-1}
    },
    {
        {31126,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32230,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32230,-1},
        {-1,-1,-1}
    },
    {
        {31128,-1},
        {-1},
        {-1,-1},
        {31128,-1,-1}
    },
    {
        {31128,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31128,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31957,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31957,-1},
        {-1,-1,-1}
    },
    {
        {31131,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31133,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31133,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31133,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31133,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32286,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35115,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35115,-1},
        {-1,-1,-1}
    },
    {
        {31137,-1},
        {-1},
        {-1,-1},
        {31137,-1,-1}
    },
    {
        {31137,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27067,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27067,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31098,-1},
        {31098,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31142,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31143,-1},
        {-1},
        {-1,-1},
        {31143,-1,-1}
    },
    {
        {31143,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31145,-1},
        {-1},
        {-1,-1},
        {31145,-1,-1}
    },
    {
        {31145,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31145,-1},
        {-1},
        {-1,-1},
        {31145,-1,-1}
    },
    {
        {31146,-1},
        {-1},
        {-1,-1},
        {31109,-1,-1}
    },
    {
        {31146,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27292,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27292,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31150,-1},
        {-1},
        {-1,-1},
        {31036,31150,-1}
    },
    {
        {31150,-1},
        {-1},
        {-1,-1},
        {31036,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31152,-1},
        {-1},
        {-1,-1},
        {31152,-1,-1}
    },
    {
        {31152,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31152,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31153,-1},
        {-1},
        {-1,-1},
        {31153,-1,-1}
    },
    {
        {31153,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27348,-1},
        {-1,-1,-1}
    },
    {
        {31155,-1},
        {-1},
        {-1,-1},
        {31155,-1,-1}
    },
    {
        {31155,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31156,-1},
        {-1},
        {-1,-1},
        {31156,-1,-1}
    },
    {
        {31156,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31124,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27409,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39006,-1},
        {-1,-1,-1}
    },
    {
        {31160,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31160,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31161,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31162,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31163,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31163,-1},
        {-1},
        {-1,-1},
        {31163,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22865,-1},
        {-1,-1,-1}
    },
    {
        {31165,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31165,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31166,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31167,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31168,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31169,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31170,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32788,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26452,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31240,-1},
        {31240,-1,-1}
    },
    {
        {31175,-1},
        {-1},
        {-1,-1},
        {33464,34269,-1}
    },
    {
        {31176,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31177,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31178,-1},
        {-1},
        {-1,-1},
        {24180,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31179,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31180,-1},
        {-1},
        {-1,-1},
        {31179,-1,-1}
    },
    {
        {31181,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26508,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32791,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32792,-1},
        {32792,-1,-1}
    },
    {
        {31185,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31186,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26525,-1},
        {-1,-1,-1}
    },
    {
        {31188,-1},
        {-1},
        {-1,-1},
        {31923,-1,-1}
    },
    {
        {31189,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31190,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39147,-1},
        {39147,-1,-1}
    },
    {
        {31192,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26543,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26560,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31198,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31199,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31201,-1},
        {-1},
        {-1,-1},
        {31059,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40801,40802,-1}
    },
    {
        {31203,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31204,-1},
        {-1},
        {-1,-1},
        {31204,-1,-1}
    },
    {
        {31204,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31206,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31207,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26590,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31209,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31190,-1,-1}
    },
    {
        {31211,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31212,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31213,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31309,-1},
        {-1,-1,-1}
    },
    {
        {31216,-1},
        {-1},
        {-1,-1},
        {31216,31281,-1}
    },
    {
        {31216,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33940,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26653,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28895,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31906,-1},
        {31906,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25347,-1},
        {25347,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34997,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31227,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35036,-1},
        {-1,-1,-1}
    },
    {
        {31232,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31294,-1},
        {31294,31295,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31235,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26797,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31246,-1},
        {-1},
        {-1,-1},
        {31246,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31341,-1},
        {31341,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31240,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31923,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31242,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31243,-1},
        {-1},
        {-1,-1},
        {31243,-1,-1}
    },
    {
        {31243,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31245,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31245,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31246,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26895,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26822,-1},
        {-1,-1,-1}
    },
    {
        {31249,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26852,-1},
        {-1,-1,-1}
    },
    {
        {31252,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31253,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31255,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31255,-1},
        {-1},
        {-1,-1},
        {31255,-1,-1}
    },
    {
        {31256,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31257,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31258,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31259,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31260,-1},
        {-1},
        {-1,-1},
        {31260,-1,-1}
    },
    {
        {31260,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26842,-1},
        {-1,-1,-1}
    },
    {
        {31262,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31263,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31264,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31264,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31125,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34250,31940},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31983,-1},
        {31983,-1,-1}
    },
    {
        {31277,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31278,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31281,-1},
        {-1},
        {-1,-1},
        {31216,-1,-1}
    },
    {
        {31281,-1},
        {-1},
        {-1,-1},
        {31216,31281,-1}
    },
    {
        {31281,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31282,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31287,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31288,-1},
        {-1},
        {-1,-1},
        {33988,-1,-1}
    },
    {
        {31289,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31258,-1,-1}
    },
    {
        {31291,-1},
        {-1},
        {-1,-1},
        {31282,-1,-1}
    },
    {
        {31291,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31292,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31292,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31293,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31293,-1},
        {-1},
        {-1,-1},
        {31293,-1,-1}
    },
    {
        {31294,-1},
        {-1},
        {-1,-1},
        {31295,-1,-1}
    },
    {
        {31295,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31295,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31296,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33592,25558},
        {-1,-1,-1}
    },
    {
        {31298,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31299,-1},
        {-1},
        {-1,-1},
        {27029,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31302,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31303,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31305,-1},
        {-1},
        {-1,-1},
        {31258,-1,-1}
    },
    {
        {31305,-1},
        {-1},
        {-1,-1},
        {31258,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31318,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31249,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31249,-1},
        {-1,-1,-1}
    },
    {
        {31308,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31309,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31310,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31311,-1},
        {-1},
        {-1,-1},
        {31337,31311,-1}
    },
    {
        {31311,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31312,-1},
        {-1},
        {-1,-1},
        {31179,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31145,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27243,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31318,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31319,-1},
        {-1},
        {-1,-1},
        {31298,-1,-1}
    },
    {
        {31319,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31319,-1},
        {-1},
        {-1,-1},
        {31319,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27208,-1},
        {-1,-1,-1}
    },
    {
        {31321,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {144954,27211},
        {144954,27211,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27189,-1},
        {-1,-1,-1}
    },
    {
        {31324,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31324,-1},
        {-1},
        {-1,-1},
        {31324,-1,-1}
    },
    {
        {31325,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31327,-1},
        {-1},
        {-1,-1},
        {31327,-1,-1}
    },
    {
        {31327,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31328,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31329,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31330,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31330,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31330,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31331,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31983,-1},
        {31983,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27315,-1},
        {-1,-1,-1}
    },
    {
        {31335,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38973,-1},
        {38973,-1,-1}
    },
    {
        {31337,-1},
        {-1},
        {-1,-1},
        {31311,-1,-1}
    },
    {
        {31337,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31338,-1},
        {-1},
        {-1,-1},
        {31216,31281,-1}
    },
    {
        {31339,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31339,-1},
        {-1},
        {-1,-1},
        {31339,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25844,-1},
        {-1,-1,-1}
    },
    {
        {31341,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31344,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27408,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27409,-1},
        {-1,-1,-1}
    },
    {
        {31348,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31348,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31349,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31350,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31406,-1},
        {-1,-1,-1}
    },
    {
        {31352,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31353,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31354,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31354,-1},
        {-1},
        {-1,-1},
        {31354,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23431,-1},
        {23431,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31361,-1},
        {-1,-1,-1}
    },
    {
        {31357,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31359,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31359,-1},
        {-1},
        {-1,-1},
        {31359,-1,-1}
    },
    {
        {31359,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31359,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31361,-1},
        {-1},
        {64085,-1},
        {-1,-1,-1}
    },
    {
        {31362,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31363,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31364,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31365,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31368,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23502,-1},
        {-1,-1,-1}
    },
    {
        {31370,-1},
        {-1},
        {-1,-1},
        {31370,-1,-1}
    },
    {
        {31370,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31382,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27737,31402},
        {-1,-1,-1}
    },
    {
        {31376,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31407,-1,-1}
    },
    {
        {31378,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31379,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31381,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31381,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31382,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31382,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31383,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31383,-1},
        {-1},
        {-1,-1},
        {31379,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31384,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23468,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31391,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31392,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21461,31381},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32621,-1},
        {-1,-1,-1}
    },
    {
        {31395,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31647,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38512,34093},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38512,34093},
        {-1,-1,-1}
    },
    {
        {31401,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31402,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31349,-1},
        {31349,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31349,-1},
        {31349,-1,-1}
    },
    {
        {31404,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31404,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31406,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31406,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31407,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31408,-1},
        {-1},
        {-1,-1},
        {31407,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31712,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26009,-1},
        {-1,-1,-1}
    },
    {
        {31411,-1},
        {-1},
        {-1,-1},
        {31411,-1,-1}
    },
    {
        {31411,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23512,38352},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23512,38352},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31414,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32842,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24713,31379},
        {-1,-1,-1}
    },
    {
        {31417,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31418,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31419,-1},
        {-1},
        {-1,-1},
        {31379,-1,-1}
    },
    {
        {31419,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31420,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31422,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31423,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31423,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36204,38743},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27635,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23534,-1},
        {-1,-1,-1}
    },
    {
        {31427,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31428,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31428,-1},
        {-1},
        {-1,-1},
        {31428,-1,-1}
    },
    {
        {31429,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31406,-1,-1}
    },
    {
        {31431,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31432,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31432,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31434,-1},
        {-1},
        {-1,-1},
        {31363,-1,-1}
    },
    {
        {31435,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31436,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31437,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31438,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31439,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25237,-1},
        {-1,-1,-1}
    },
    {
        {31441,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31442,-1},
        {-1},
        {-1,-1},
        {22855,-1,-1}
    },
    {
        {31443,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31445,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31445,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31449,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31450,-1},
        {-1},
        {-1,-1},
        {20295,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31452,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31453,-1},
        {-1},
        {-1,-1},
        {20006,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31455,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31455,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31456,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31456,-1},
        {-1},
        {-1,-1},
        {31456,-1,-1}
    },
    {
        {31457,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31458,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31459,-1},
        {-1},
        {-1,-1},
        {31459,-1,-1}
    },
    {
        {31459,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31441,-1},
        {-1,-1,-1}
    },
    {
        {31461,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31461,-1},
        {-1},
        {-1,-1},
        {31461,-1,-1}
    },
    {
        {31462,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31463,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31464,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35516,-1,-1}
    },
    {
        {31466,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31467,-1},
        {-1},
        {-1,-1},
        {38742,-1,-1}
    },
    {
        {31467,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31469,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31469,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22592,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22592,-1},
        {-1,-1,-1}
    },
    {
        {31471,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31472,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22705,-1},
        {-1,-1,-1}
    },
    {
        {31476,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27498,-1},
        {-1,-1,-1}
    },
    {
        {31478,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25211,22350},
        {-1,-1,-1}
    },
    {
        {31480,-1},
        {-1},
        {-1,-1},
        {31478,-1,-1}
    },
    {
        {31481,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31482,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31735,-1},
        {-1,-1,-1}
    },
    {
        {31485,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31722,-1,-1}
    },
    {
        {31487,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33426,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31482,31716},
        {31482,31716,-1}
    },
    {
        {31490,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31492,-1},
        {-1},
        {-1,-1},
        {31492,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29243,-1},
        {-1,-1,-1}
    },
    {
        {31494,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31495,-1},
        {-1},
        {-1,-1},
        {31639,-1,-1}
    },
    {
        {31496,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31496,-1},
        {-1},
        {-1,-1},
        {31496,-1,-1}
    },
    {
        {31496,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21248,31503},
        {-1,-1,-1}
    },
    {
        {31498,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31499,-1},
        {-1},
        {-1,-1},
        {31565,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33469,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33454,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33451,-1},
        {-1,-1,-1}
    },
    {
        {31503,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20130,-1},
        {-1,-1,-1}
    },
    {
        {31505,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31506,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31718,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31558,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33505,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33523,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31512,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31513,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31515,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33537,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31623,-1},
        {-1,-1,-1}
    },
    {
        {31518,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31629,-1,-1}
    },
    {
        {31520,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33524,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33493,-1},
        {-1,-1,-1}
    },
    {
        {31525,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31526,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31527,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31528,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31567,-1,-1}
    },
    {
        {31530,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31532,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31533,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31534,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31535,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31537,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31539,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31541,-1},
        {-1},
        {-1,-1},
        {31684,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31542,-1},
        {-1},
        {-1,-1},
        {30690,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33542,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21493,-1},
        {-1,-1,-1}
    },
    {
        {31545,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31547,-1},
        {-1},
        {-1,-1},
        {31559,-1,-1}
    },
    {
        {31549,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31553,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31722,-1},
        {31722,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33589,-1},
        {33589,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31381,-1},
        {-1,-1,-1}
    },
    {
        {31557,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31558,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31559,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31560,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31561,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31563,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31564,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31564,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31565,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31566,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31567,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31568,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31569,-1},
        {-1},
        {-1,-1},
        {31569,-1,-1}
    },
    {
        {31569,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31569,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31570,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31492,-1},
        {-1},
        {31492,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31572,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31573,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31574,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31645,-1},
        {-1,-1,-1}
    },
    {
        {31581,-1},
        {-1},
        {-1,-1},
        {31631,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31604,31574},
        {31574,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33705,-1},
        {-1,-1,-1}
    },
    {
        {31584,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33660,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25226,31494},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31852,-1},
        {-1,-1,-1}
    },
    {
        {31588,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31589,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31590,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31591,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33695,-1},
        {-1,-1,-1}
    },
    {
        {31593,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36074,-1},
        {-1,-1,-1}
    },
    {
        {31596,-1},
        {-1},
        {-1,-1},
        {31596,-1,-1}
    },
    {
        {31596,-1},
        {-1},
        {-1,-1},
        {31596,-1,-1}
    },
    {
        {31597,-1},
        {-1},
        {-1,-1},
        {31639,-1,-1}
    },
    {
        {31597,-1},
        {-1},
        {-1,-1},
        {31597,31639,-1}
    },
    {
        {31598,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31599,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31600,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31601,-1},
        {-1},
        {-1,-1},
        {31712,-1,-1}
    },
    {
        {31602,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31602,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31603,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31604,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31605,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31605,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31605,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31560,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {16903,-1},
        {-1,-1,-1}
    },
    {
        {31609,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31610,-1},
        {-1},
        {-1,-1},
        {31568,-1,-1}
    },
    {
        {31615,-1},
        {-1},
        {-1,-1},
        {31712,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33769,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33802,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31622,-1},
        {-1},
        {-1,-1},
        {31718,-1,-1}
    },
    {
        {31623,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33492,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31627,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32617,-1},
        {32617,16903,-1}
    },
    {
        {31629,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31631,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31631,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31632,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31632,-1},
        {-1},
        {-1,-1},
        {31632,-1,-1}
    },
    {
        {31633,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31634,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31634,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31833,31759},
        {31759,-1,-1}
    },
    {
        {31636,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31637,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31639,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31640,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31641,-1},
        {-1},
        {-1,-1},
        {31641,-1,-1}
    },
    {
        {31641,-1},
        {-1},
        {-1,-1},
        {31641,-1,-1}
    },
    {
        {31642,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31643,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31643,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31644,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31644,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31645,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31646,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31646,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {31646},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31646,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31646,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31647,-1},
        {-1},
        {-1,-1},
        {31640,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31649,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33728,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31574,-1},
        {31574,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33795,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31658,-1},
        {-1},
        {-1,-1},
        {31774,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31661,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31661,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33841,-1},
        {-1,-1,-1}
    },
    {
        {31663,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31565,-1,-1}
    },
    {
        {31665,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23631,-1},
        {-1,-1,-1}
    },
    {
        {31668,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31669,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31672,-1},
        {-1},
        {-1,-1},
        {31672,-1,-1}
    },
    {
        {31672,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33903,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31676,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38808,26786},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33914,-1},
        {-1,-1,-1}
    },
    {
        {31680,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31680,-1},
        {-1},
        {-1,-1},
        {31680,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31681,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30473,-1},
        {-1,-1,-1}
    },
    {
        {31684,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31684,-1},
        {-1},
        {-1,-1},
        {31684,-1,-1}
    },
    {
        {31684,-1},
        {-1},
        {-1,-1},
        {31684,-1,-1}
    },
    {
        {31685,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31686,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31686,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31687,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31687,-1},
        {-1},
        {-1,-1},
        {31687,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31689,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31689,-1},
        {-1},
        {-1,-1},
        {31689,-1,-1}
    },
    {
        {31690,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31691,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31692,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33833,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30471,-1},
        {-1,-1,-1}
    },
    {
        {31695,-1},
        {-1},
        {-1,-1},
        {31645,-1,-1}
    },
    {
        {31700,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33995,-1},
        {-1,-1,-1}
    },
    {
        {31702,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31703,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31705,-1},
        {-1},
        {-1,-1},
        {15375,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31706,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33976,-1},
        {-1,-1,-1}
    },
    {
        {31709,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31709,-1},
        {-1},
        {-1,-1},
        {31709,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28037,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31712,-1},
        {-1},
        {-1,-1},
        {31712,-1,-1}
    },
    {
        {31712,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31762,-1},
        {-1},
        {31762,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {17672,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24429,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24429,-1},
        {-1,-1,-1}
    },
    {
        {31716,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31717,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31718,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38586,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31721,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31722,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31689,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33980,-1},
        {-1,-1,-1}
    },
    {
        {31725,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31728,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31852,26461},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31731,-1},
        {-1},
        {-1,-1},
        {31731,-1,-1}
    },
    {
        {31731,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31734,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31735,-1},
        {-1},
        {-1,-1},
        {31735,-1,-1}
    },
    {
        {31735,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31735,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31717,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31838,-1},
        {31838,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34036,-1},
        {-1,-1,-1}
    },
    {
        {31744,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31745,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31747,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29987,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31751,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23643,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31637,-1},
        {-1,-1,-1}
    },
    {
        {31755,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31757,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31758,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31759,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31684,-1},
        {-1,-1,-1}
    },
    {
        {31761,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31762,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31763,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31763,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31763,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31764,-1},
        {-1},
        {-1,-1},
        {31761,-1,-1}
    },
    {
        {31767,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31787,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34153,30442},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23534,-1},
        {-1,-1,-1}
    },
    {
        {31774,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31775,-1},
        {-1},
        {-1,-1},
        {31775,-1,-1}
    },
    {
        {31775,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31776,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31777,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38292,-1},
        {-1,-1,-1}
    },
    {
        {31779,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31782,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31783,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31783,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21214,-1},
        {-1,-1,-1}
    },
    {
        {31786,-1},
        {-1},
        {-1,-1},
        {31786,-1,-1}
    },
    {
        {31786,-1},
        {-1},
        {-1,-1},
        {31786,-1,-1}
    },
    {
        {31786,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31786,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31787,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22124,31598},
        {31598,-1,-1}
    },
    {
        {31793,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31795,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31798,-1},
        {-1},
        {-1,-1},
        {31798,-1,-1}
    },
    {
        {31798,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31799,-1},
        {-1},
        {-1,-1},
        {27280,-1,-1}
    },
    {
        {31800,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36942,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36942,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31709,-1},
        {31709,-1,-1}
    },
    {
        {31805,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31806,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31806,-1},
        {-1},
        {-1,-1},
        {31806,-1,-1}
    },
    {
        {31806,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31807,-1},
        {-1},
        {-1,-1},
        {31807,-1,-1}
    },
    {
        {31807,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31808,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31811,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31779,-1},
        {31779,-1,-1}
    },
    {
        {31813,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31705,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34233,-1},
        {-1,-1,-1}
    },
    {
        {31818,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36513,38007},
        {-1,-1,-1}
    },
    {
        {31820,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31821,-1},
        {-1},
        {-1,-1},
        {31821,-1,-1}
    },
    {
        {31821,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31823,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31824,-1},
        {-1},
        {-1,-1},
        {31824,-1,-1}
    },
    {
        {31824,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31824,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31825,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31825,-1},
        {-1},
        {-1,-1},
        {25776,31825,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31828,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31830,-1},
        {-1},
        {-1,-1},
        {31844,-1,-1}
    },
    {
        {31831,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31832,-1},
        {-1},
        {-1,-1},
        {31824,31832,-1}
    },
    {
        {31832,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31832,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31833,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31855,-1,-1}
    },
    {
        {31838,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31839,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31839,-1},
        {-1},
        {-1,-1},
        {31839,-1,-1}
    },
    {
        {31840,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31840,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31840,-1},
        {-1},
        {-1,-1},
        {31840,-1,-1}
    },
    {
        {31840,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31841,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22889,21299},
        {21299,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31844,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31845,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37758,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31849,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31849,-1},
        {-1},
        {-1,-1},
        {31849,-1,-1}
    },
    {
        {31849,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26029,26039},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26029,26039},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31825,-1},
        {-1,-1,-1}
    },
    {
        {31852,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31852,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31853,-1},
        {-1},
        {-1,-1},
        {31721,-1,-1}
    },
    {
        {31854,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31855,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31855,-1},
        {-1},
        {-1,-1},
        {31855,-1,-1}
    },
    {
        {31856,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31831,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31859,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38609,31988},
        {31988,-1,-1}
    },
    {
        {31861,-1},
        {-1},
        {-1,-1},
        {36463,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31885,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31873,24041},
        {-1,-1,-1}
    },
    {
        {31865,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35208,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {63952,39006},
        {63952,39006,-1}
    },
    {
        {-1,-1},
        {-1},
        {63952,39006},
        {63952,39006,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31176,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31870,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31870,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31870,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26449,26009},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31873,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31874,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31874,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31875,-1},
        {-1},
        {-1,-1},
        {31189,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26495,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29289,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31941,-1},
        {-1,-1,-1}
    },
    {
        {31881,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31881,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31875,31189},
        {31875,31189,-1}
    },
    {
        {31883,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30695,-1},
        {-1,-1,-1}
    },
    {
        {31885,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37320,-1},
        {-1,-1,-1}
    },
    {
        {31887,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31888,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31888,-1},
        {-1},
        {-1,-1},
        {31888,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31890,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27860,-1},
        {27860,-1,-1}
    },
    {
        {31892,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31892,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31893,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31203,-1},
        {-1,-1,-1}
    },
    {
        {31895,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31896,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31198,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40656,-1},
        {40656,-1,-1}
    },
    {
        {31899,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31990,-1},
        {31990,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31986,-1},
        {31986,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31903,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31904,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31905,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31906,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26613,-1},
        {-1,-1,-1}
    },
    {
        {31908,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31909,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31910,-1},
        {-1},
        {-1,-1},
        {29136,-1,-1}
    },
    {
        {31910,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31911,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31912,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31915,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31917,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31918,-1},
        {-1},
        {-1,-1},
        {31975,-1,-1}
    },
    {
        {31918,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26776,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31921,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31922,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31923,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31923,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31935,-1},
        {-1,-1,-1}
    },
    {
        {31926,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31927,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31256,-1},
        {-1,-1,-1}
    },
    {
        {31929,-1},
        {-1},
        {-1,-1},
        {31883,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31932,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31933,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31934,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31935,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31936,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31936,-1},
        {-1},
        {-1,-1},
        {31936,-1,-1}
    },
    {
        {31936,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31938,-1},
        {-1},
        {-1,-1},
        {31965,-1,-1}
    },
    {
        {31938,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39219,31958},
        {39219,31958,-1}
    },
    {
        {31940,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31941,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40634,40629},
        {40618,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39217,-1,-1}
    },
    {
        {31944,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31946,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25539,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31944,-1},
        {-1,-1,-1}
    },
    {
        {31949,-1},
        {-1},
        {-1,-1},
        {39176,-1,-1}
    },
    {
        {31950,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27025,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27026,-1},
        {-1,-1,-1}
    },
    {
        {31954,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31296,64084,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31957,-1},
        {-1},
        {-1,-1},
        {39227,-1,-1}
    },
    {
        {31958,-1},
        {-1},
        {-1,-1},
        {31958,-1,-1}
    },
    {
        {31958,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31959,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31960,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31960,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31961,-1},
        {-1},
        {-1,-1},
        {31961,-1,-1}
    },
    {
        {31961,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22941,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22941,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31958,-1},
        {31958,-1,-1}
    },
    {
        {31964,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31964,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31965,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31966,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31967,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31968,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37857,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37857,-1},
        {-1,-1,-1}
    },
    {
        {31970,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31970,-1},
        {-1},
        {-1,-1},
        {27169,31970,-1}
    },
    {
        {-1,-1},
        {-1},
        {31965,-1},
        {31938,31965,-1}
    },
    {
        {-1,-1},
        {-1},
        {25746,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31983,-1},
        {31983,-1,-1}
    },
    {
        {31974,-1},
        {-1},
        {-1,-1},
        {39246,-1,-1}
    },
    {
        {31975,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28479,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27287,-1},
        {-1,-1,-1}
    },
    {
        {31979,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27323,-1},
        {-1,-1,-1}
    },
    {
        {31983,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22296,22243},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23421,31989},
        {-1,-1,-1}
    },
    {
        {31986,-1},
        {-1},
        {-1,-1},
        {31986,-1,-1}
    },
    {
        {31986,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38015,-1},
        {-1,-1,-1}
    },
    {
        {31988,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31988,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31989,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31989,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31990,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31990,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {155330,-1},
        {-1,-1,-1}
    },
    {
        {31992,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31994,-1},
        {-1},
        {-1,-1},
        {31998,-1,-1}
    },
    {
        {31995,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21151,-1},
        {21151,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32006,-1},
        {-1,-1,-1}
    },
    {
        {31998,-1},
        {-1},
        {-1,-1},
        {31998,-1,-1}
    },
    {
        {31998,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32000,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32000,-1},
        {-1},
        {-1,-1},
        {32000,-1,-1}
    },
    {
        {32000,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32000,-1},
        {-1},
        {-1,-1},
        {32000,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34921,-1},
        {-1,-1,-1}
    },
    {
        {32002,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32003,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32004,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32004,-1},
        {-1},
        {-1,-1},
        {32004,-1,-1}
    },
    {
        {32005,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32006,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32007,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32008,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32008,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32009,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32010,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32010,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32011,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32011,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32030,-1},
        {-1,-1,-1}
    },
    {
        {32013,-1},
        {-1},
        {-1,-1},
        {32013,-1,-1}
    },
    {
        {32013,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32013,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32049,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32016,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32018,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32019,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32020,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32020,-1},
        {-1},
        {-1,-1},
        {32020,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32021,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32023,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32024,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32025,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32026,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32026,-1},
        {-1},
        {-1,-1},
        {32026,-1,-1}
    },
    {
        {32027,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32027,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32028,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32029,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32030,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31105,-1},
        {-1,-1,-1}
    },
    {
        {32032,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32033,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32034,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32035,-1},
        {-1},
        {-1,-1},
        {32183,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31048,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32046,25166},
        {32046,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32283,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32064,37341},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32192,-1,-1}
    },
    {
        {32043,-1},
        {-1},
        {-1,-1},
        {32043,-1,-1}
    },
    {
        {32043,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32044,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32024,-1,-1}
    },
    {
        {32046,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32047,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32048,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32049,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32050,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32051,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34987,-1},
        {-1,-1,-1}
    },
    {
        {32053,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34986,32028},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26563,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32083,-1},
        {-1,-1,-1}
    },
    {
        {32057,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32058,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {155604,24591},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32061,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32062,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32063,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32064,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32065,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32066,-1},
        {-1},
        {-1,-1},
        {32066,-1,-1}
    },
    {
        {32066,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32067,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32068,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32069,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32070,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32070,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32071,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32072,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24629,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37474,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32075,-1},
        {-1},
        {-1,-1},
        {32024,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32076,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32029,-1,-1}
    },
    {
        {32078,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32079,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32080,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32003,-1},
        {-1,-1,-1}
    },
    {
        {32083,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32072,-1},
        {-1,-1,-1}
    },
    {
        {32118,-1},
        {-1},
        {32118,-1},
        {32118,-1,-1}
    },
    {
        {32086,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32089,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32090,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32091,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32092,-1},
        {-1},
        {-1,-1},
        {32092,-1,-1}
    },
    {
        {32092,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32092,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32092,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34996,-1,-1}
    },
    {
        {32094,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32094,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33603,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32097,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32098,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32099,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32099,-1},
        {-1},
        {-1,-1},
        {32099,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32171,-1},
        {32218,32171,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32102,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32104,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31079,-1},
        {-1,-1,-1}
    },
    {
        {32106,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32047,-1},
        {32047,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32080,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24059,21367},
        {-1,-1,-1}
    },
    {
        {32110,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39405,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39405,-1},
        {-1,-1,-1}
    },
    {
        {32112,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32113,-1},
        {-1},
        {-1,-1},
        {32113,-1,-1}
    },
    {
        {32113,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32114,-1},
        {-1},
        {-1,-1},
        {31992,-1,-1}
    },
    {
        {32115,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32115,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32115,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31077,-1},
        {-1,-1,-1}
    },
    {
        {32117,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32118,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32365,-1,-1}
    },
    {
        {32121,-1},
        {-1},
        {-1,-1},
        {32121,-1,-1}
    },
    {
        {32121,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32122,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38601,-1},
        {-1,-1,-1}
    },
    {
        {32125,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32104,-1},
        {-1,-1,-1}
    },
    {
        {32127,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {64087,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32113,-1,-1}
    },
    {
        {32131,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32176,20436},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32134,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32134,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32134,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32136,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32137,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32139,-1},
        {-1},
        {-1,-1},
        {32024,-1,-1}
    },
    {
        {32140,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32143,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32143,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26740,32181},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32147,-1},
        {-1},
        {-1,-1},
        {32076,-1,-1}
    },
    {
        {32147,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29920,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32340,-1},
        {32340,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32151,-1},
        {-1},
        {-1,-1},
        {32069,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32299,-1},
        {32299,-1,-1}
    },
    {
        {32153,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32154,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32155,-1},
        {-1},
        {-1,-1},
        {32155,-1,-1}
    },
    {
        {32155,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32155,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32156,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32157,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32159,-1},
        {-1},
        {-1,-1},
        {32159,-1,-1}
    },
    {
        {32159,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32160,-1},
        {-1},
        {-1,-1},
        {32209,-1,-1}
    },
    {
        {32160,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26891,-1},
        {-1,-1,-1}
    },
    {
        {32162,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32162,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32163,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32163,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32057,-1},
        {32057,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32166,-1},
        {-1,-1,-1}
    },
    {
        {32166,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32167,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32166,-1},
        {32166,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26880,-1},
        {-1,-1,-1}
    },
    {
        {32170,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32170,-1},
        {-1},
        {-1,-1},
        {32170,-1,-1}
    },
    {
        {32171,-1},
        {-1},
        {-1,-1},
        {32218,-1,-1}
    },
    {
        {32172,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32173,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32174,-1},
        {-1},
        {-1,-1},
        {32174,-1,-1}
    },
    {
        {32174,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32175,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32176,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32177,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32178,-1},
        {-1},
        {-1,-1},
        {32178,-1,-1}
    },
    {
        {32178,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32323,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32180,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32181,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32181,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32182,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32183,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32184,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32186,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32187,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35048,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32189,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32190,-1},
        {-1},
        {-1,-1},
        {32190,-1,-1}
    },
    {
        {32190,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32191,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32192,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24482,-1},
        {-1,-1,-1}
    },
    {
        {32194,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32199,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32199,-1},
        {-1},
        {-1,-1},
        {32199,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32289,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35058,-1},
        {-1,-1,-1}
    },
    {
        {32202,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32203,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32203,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32204,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32225,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32207,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32207,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32209,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32210,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32213,-1},
        {-1},
        {-1,-1},
        {32387,-1,-1}
    },
    {
        {32214,-1},
        {-1},
        {-1,-1},
        {32210,-1,-1}
    },
    {
        {32215,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32216,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32217,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32218,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24862,-1},
        {-1,-1,-1}
    },
    {
        {32220,-1},
        {-1},
        {-1,-1},
        {32191,-1,-1}
    },
    {
        {32221,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32221,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32222,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37325,-1},
        {-1,-1,-1}
    },
    {
        {32224,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32224,-1},
        {-1},
        {-1,-1},
        {32224,-1,-1}
    },
    {
        {32225,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32227,-1},
        {-1},
        {-1,-1},
        {32257,-1,-1}
    },
    {
        {32227,-1},
        {-1},
        {-1,-1},
        {32257,32227,-1}
    },
    {
        {32228,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35091,-1,-1}
    },
    {
        {32230,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31312,31179},
        {-1,-1,-1}
    },
    {
        {32232,-1},
        {-1},
        {-1,-1},
        {32232,-1,-1}
    },
    {
        {32232,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32233,-1},
        {-1},
        {-1,-1},
        {32233,-1,-1}
    },
    {
        {32233,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32317,32207,-1}
    },
    {
        {32236,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32239,-1},
        {-1},
        {-1,-1},
        {32239,-1,-1}
    },
    {
        {32239,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32239,-1},
        {-1},
        {-1,-1},
        {32239,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32379,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32379,-1},
        {-1,-1,-1}
    },
    {
        {32241,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32242,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32315,32092},
        {-1,-1,-1}
    },
    {
        {32244,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35084,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32313,-1},
        {-1,-1,-1}
    },
    {
        {32249,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32251,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32277,-1},
        {-1},
        {32277,-1},
        {32277,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36204,-1},
        {36204,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32063,-1},
        {32063,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32257,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32257,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32317,32207},
        {32317,32207,-1}
    },
    {
        {-1,-1},
        {-1},
        {32317,32207},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32997,-1},
        {32997,-1,-1}
    },
    {
        {32260,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32261,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32264,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32265,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32266,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32266,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32266,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32267,-1},
        {-1},
        {-1,-1},
        {32267,-1,-1}
    },
    {
        {32267,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36870,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36870,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24171,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27070,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32321,-1},
        {-1,-1,-1}
    },
    {
        {32272,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32273,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32274,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32368,-1},
        {-1,-1,-1}
    },
    {
        {32277,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37771,-1},
        {-1,-1,-1}
    },
    {
        {32279,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32363,-1},
        {32363,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33592,32221},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32091,32295},
        {-1,-1,-1}
    },
    {
        {32283,-1},
        {-1},
        {-1,-1},
        {32283,-1,-1}
    },
    {
        {32283,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32121,-1,-1}
    },
    {
        {32285,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32285,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32286,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32287,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32287,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32288,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32289,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32290,-1},
        {-1},
        {-1,-1},
        {32290,-1,-1}
    },
    {
        {32290,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32290,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32291,-1},
        {-1},
        {-1,-1},
        {30476,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32032,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27035,-1},
        {-1,-1,-1}
    },
    {
        {32294,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32295,-1},
        {-1},
        {-1,-1},
        {32091,-1,-1}
    },
    {
        {32295,-1},
        {-1},
        {-1,-1},
        {32091,-1,-1}
    },
    {
        {32296,-1},
        {-1},
        {-1,-1},
        {24140,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32156,14428},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30050,-1},
        {-1,-1,-1}
    },
    {
        {32299,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32299,-1},
        {-1},
        {-1,-1},
        {32299,-1,-1}
    },
    {
        {32300,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32300,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32302,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32410,-1},
        {32410,-1,-1}
    },
    {
        {32305,-1},
        {-1},
        {-1,-1},
        {32294,-1,-1}
    },
    {
        {32306,-1},
        {-1},
        {-1,-1},
        {32047,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32330,29309},
        {-1,-1,-1}
    },
    {
        {32309,-1},
        {-1},
        {-1,-1},
        {32309,-1,-1}
    },
    {
        {32309,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32309,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32311,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27169,-1},
        {-1,-1,-1}
    },
    {
        {32313,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32314,-1},
        {-1},
        {-1,-1},
        {32314,-1,-1}
    },
    {
        {32314,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32315,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32315,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29831,-1},
        {-1,-1,-1}
    },
    {
        {32317,-1},
        {-1},
        {-1,-1},
        {32207,-1,-1}
    },
    {
        {32317,-1},
        {-1},
        {-1,-1},
        {32317,32207,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32318,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32173,27119},
        {-1,-1,-1}
    },
    {
        {32321,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29575,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32323,-1},
        {-1},
        {-1,-1},
        {32323,-1,-1}
    },
    {
        {32323,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32325,-1},
        {-1},
        {-1,-1},
        {32368,-1,-1}
    },
    {
        {32325,-1},
        {-1},
        {-1,-1},
        {32368,32325,-1}
    },
    {
        {32326,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32326,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32327,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32358,-1},
        {-1},
        {35137,32358},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32330,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32331,-1},
        {-1},
        {-1,-1},
        {32363,-1,-1}
    },
    {
        {32333,-1},
        {-1},
        {-1,-1},
        {32353,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37897,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23657,-1},
        {-1,-1,-1}
    },
    {
        {32338,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32338,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27190,-1},
        {-1,-1,-1}
    },
    {
        {32340,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32340,-1},
        {-1},
        {-1,-1},
        {32340,-1,-1}
    },
    {
        {32341,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32342,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27209,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32345,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32346,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32189,-1},
        {32189,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32349,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32350,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34306,27236},
        {-1,-1,-1}
    },
    {
        {32353,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30059,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32386,-1},
        {32386,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32358,-1},
        {-1},
        {-1,-1},
        {32358,-1,-1}
    },
    {
        {32359,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32299,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32299,-1},
        {-1,-1,-1}
    },
    {
        {32361,-1},
        {-1},
        {-1,-1},
        {32260,-1,-1}
    },
    {
        {32361,-1},
        {-1},
        {-1,-1},
        {32260,-1,-1}
    },
    {
        {32362,-1},
        {-1},
        {-1,-1},
        {32117,-1,-1}
    },
    {
        {32362,-1},
        {-1},
        {-1,-1},
        {32362,-1,-1}
    },
    {
        {32363,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27299,-1},
        {-1,-1,-1}
    },
    {
        {32365,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32365,-1},
        {-1},
        {-1,-1},
        {32365,-1,-1}
    },
    {
        {32365,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38849,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32368,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32207,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27294,-1},
        {-1,-1,-1}
    },
    {
        {32371,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32174,35166},
        {-1,-1,-1}
    },
    {
        {32373,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27277,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27277,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32377,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32377,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29022,-1},
        {-1,-1,-1}
    },
    {
        {32379,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32380,-1},
        {-1},
        {-1,-1},
        {32153,-1,-1}
    },
    {
        {32380,-1},
        {-1},
        {-1,-1},
        {32153,-1,-1}
    },
    {
        {32381,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32381,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32381,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32382,-1},
        {-1},
        {-1,-1},
        {32382,-1,-1}
    },
    {
        {32382,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32383,-1},
        {-1},
        {-1,-1},
        {35172,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24158,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32386,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32387,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27359,-1},
        {-1,-1,-1}
    },
    {
        {32390,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32391,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32392,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32393,-1},
        {-1},
        {-1,-1},
        {32408,-1,-1}
    },
    {
        {32394,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32394,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32396,-1},
        {-1},
        {-1,-1},
        {32154,-1,-1}
    },
    {
        {32397,-1},
        {-1},
        {-1,-1},
        {32047,-1,-1}
    },
    {
        {32398,-1},
        {-1},
        {-1,-1},
        {32330,32406,-1}
    },
    {
        {32399,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32400,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32401,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32402,-1},
        {-1},
        {-1,-1},
        {32399,-1,-1}
    },
    {
        {32402,-1},
        {-1},
        {-1,-1},
        {32399,-1,-1}
    },
    {
        {32402,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32403,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32404,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32406,-1},
        {-1},
        {-1,-1},
        {32330,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25884,-1},
        {-1,-1,-1}
    },
    {
        {32408,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27407,-1},
        {-1,-1,-1}
    },
    {
        {32410,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32411,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32412,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32566,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21368,-1},
        {21368,-1,-1}
    },
    {
        {32568,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28966,32566},
        {-1,-1,-1}
    },
    {
        {32570,-1},
        {-1},
        {-1,-1},
        {27424,-1,-1}
    },
    {
        {32571,-1},
        {-1},
        {-1,-1},
        {32566,-1,-1}
    },
    {
        {32572,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32572,-1},
        {-1},
        {-1,-1},
        {27424,32570,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37474,-1,-1}
    },
    {
        {32574,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32574,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37530,32572},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32568,-1},
        {32568,-1,-1}
    },
    {
        {32579,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32580,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32581,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32592,32566},
        {-1,-1,-1}
    },
    {
        {32583,-1},
        {-1},
        {-1,-1},
        {27197,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32590,22727},
        {22748,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32590,22727},
        {22748,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29973,29934},
        {29973,-1,-1}
    },
    {
        {32588,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32589,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32590,-1},
        {-1},
        {-1,-1},
        {22748,-1,-1}
    },
    {
        {32591,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32592,-1},
        {-1},
        {-1,-1},
        {32566,-1,-1}
    },
    {
        {32592,-1},
        {-1},
        {-1,-1},
        {32566,-1,-1}
    },
    {
        {32593,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32594,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32596,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32596,-1},
        {-1},
        {-1,-1},
        {32596,-1,-1}
    },
    {
        {32596,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32597,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32645,37007},
        {32645,-1,-1}
    },
    {
        {32600,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32597,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32600,33541},
        {-1,-1,-1}
    },
    {
        {32607,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32608,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32609,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32646,32631},
        {32631,-1,-1}
    },
    {
        {32611,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32612,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32615,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32616,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32617,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32618,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32618,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32619,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36639,-1},
        {-1,-1,-1}
    },
    {
        {32621,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32622,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32626,38899},
        {-1,-1,-1}
    },
    {
        {32624,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32626,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32629,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30041,38692},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30041,38692},
        {-1,-1,-1}
    },
    {
        {32631,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32632,-1},
        {-1},
        {-1,-1},
        {32624,-1,-1}
    },
    {
        {32633,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27156,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27156,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23561,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30050,-1},
        {30050,-1,-1}
    },
    {
        {32637,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32638,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32642,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32643,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32643,-1},
        {-1},
        {-1,-1},
        {20906,32643,-1}
    },
    {
        {-1,-1},
        {-1},
        {32617,-1},
        {-1,-1,-1}
    },
    {
        {32645,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32646,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32647,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32648,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {18188,-1},
        {-1,-1,-1}
    },
    {
        {32650,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32652,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32654,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32656,-1},
        {-1},
        {-1,-1},
        {32657,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32657,-1},
        {-1},
        {-1,-1},
        {32656,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32666,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32660,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {146702,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {146702,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {146702,-1},
        {-1,-1,-1}
    },
    {
        {32662,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32663,-1},
        {-1},
        {-1,-1},
        {32652,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32982,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32660,32654},
        {32654,-1,-1}
    },
    {
        {32666,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32681,-1},
        {32681,-1,-1}
    },
    {
        {32668,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32669,-1},
        {-1},
        {-1,-1},
        {29300,-1,-1}
    },
    {
        {32670,-1},
        {-1},
        {-1,-1},
        {32670,-1,-1}
    },
    {
        {32670,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27935,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32675,-1},
        {-1},
        {-1,-1},
        {32676,-1,-1}
    },
    {
        {32676,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32662,-1},
        {-1,-1,-1}
    },
    {
        {32680,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32681,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39178,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33108,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28187,-1},
        {-1,-1,-1}
    },
    {
        {32685,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32685,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32686,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32686,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32687,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32687,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32687,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21407,-1},
        {-1,-1,-1}
    },
    {
        {32690,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29856,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32694,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32694,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32666,29547},
        {-1,-1,-1}
    },
    {
        {32696,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32696,-1},
        {-1},
        {-1,-1},
        {32696,-1,-1}
    },
    {
        {32697,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23418,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24950,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24950,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32701,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {156674,-1},
        {-1,-1,-1}
    },
    {
        {32703,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32705,-1},
        {-1},
        {-1,-1},
        {32705,-1,-1}
    },
    {
        {32705,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32705,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {156674,-1},
        {-1,-1,-1}
    },
    {
        {32707,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32709,-1},
        {32709,-1,-1}
    },
    {
        {32709,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32709,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32710,-1},
        {-1},
        {-1,-1},
        {32736,-1,-1}
    },
    {
        {32710,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39662,32703},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26665,-1},
        {-1,-1,-1}
    },
    {
        {32714,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32714,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {156674,-1},
        {-1,-1,-1}
    },
    {
        {32716,-1},
        {-1},
        {-1,-1},
        {32716,-1,-1}
    },
    {
        {32716,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25259,-1},
        {-1,-1,-1}
    },
    {
        {32718,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32719,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36604,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32970,156674},
        {-1,-1,-1}
    },
    {
        {32722,-1},
        {-1},
        {-1,-1},
        {32722,-1,-1}
    },
    {
        {32722,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35425,-1},
        {-1,-1,-1}
    },
    {
        {32724,-1},
        {-1},
        {-1,-1},
        {32724,-1,-1}
    },
    {
        {32724,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32725,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32725,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32725,-1},
        {32725,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32718,-1},
        {-1,-1,-1}
    },
    {
        {32731,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32739,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32707,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24374,-1},
        {-1,-1,-1}
    },
    {
        {32735,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32736,-1},
        {-1},
        {-1,-1},
        {32736,-1,-1}
    },
    {
        {32736,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32736,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32737,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32737,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32737,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32411,-1},
        {-1,-1,-1}
    },
    {
        {32739,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32709,-1},
        {32709,-1,-1}
    },
    {
        {32741,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32742,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32742,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32742,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32763,23459},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32745,-1},
        {-1},
        {-1,-1},
        {32745,-1,-1}
    },
    {
        {32745,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32745,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32747,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32747,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32747,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32748,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32750,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32751,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32752,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32752,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32752,-1},
        {-1},
        {-1,-1},
        {32752,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32762,-1},
        {32762,-1,-1}
    },
    {
        {32754,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32755,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32755,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32314,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40300,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32763,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36033,32762},
        {-1,-1,-1}
    },
    {
        {32761,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32761,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32762,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32763,-1},
        {-1},
        {-1,-1},
        {32763,-1,-1}
    },
    {
        {32763,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32764,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32764,-1},
        {-1},
        {-1,-1},
        {32764,-1,-1}
    },
    {
        {32764,-1},
        {-1},
        {-1,-1},
        {32764,-1,-1}
    },
    {
        {32764,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32768,-1},
        {-1},
        {-1,-1},
        {32768,-1,-1}
    },
    {
        {32768,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32769,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32770,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32771,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32771,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32772,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32773,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32774,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32776,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32778,-1},
        {-1},
        {-1,-1},
        {32779,-1,-1}
    },
    {
        {32779,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32780,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32782,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32784,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32785,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32786,-1},
        {-1},
        {-1,-1},
        {26469,-1,-1}
    },
    {
        {32786,-1},
        {-1},
        {-1,-1},
        {26469,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32788,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32789,-1},
        {-1},
        {-1,-1},
        {32789,-1,-1}
    },
    {
        {32789,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32790,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32791,-1},
        {-1},
        {-1,-1},
        {32791,-1,-1}
    },
    {
        {32791,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32792,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32792,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32793,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32793,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26560,155222},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32796,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32796,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32796,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32801,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32801,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26822,-1},
        {-1,-1,-1}
    },
    {
        {32804,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32806,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32808,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32808,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32808,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27083,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27036,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31309,-1},
        {-1,-1,-1}
    },
    {
        {32812,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31339,-1},
        {31339,-1,-1}
    },
    {
        {32816,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30952,-1},
        {-1,-1,-1}
    },
    {
        {32819,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32821,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32822,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32822,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32823,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32823,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32883,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32827,-1},
        {-1},
        {-1,-1},
        {24677,-1,-1}
    },
    {
        {32828,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32829,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32831,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32887,-1,-1}
    },
    {
        {32835,-1},
        {-1},
        {-1,-1},
        {32828,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32965,32831},
        {-1,-1,-1}
    },
    {
        {32838,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32827,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25303,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26383,-1},
        {-1,-1,-1}
    },
    {
        {32842,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30522,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32850,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32819,-1},
        {-1,-1,-1}
    },
    {
        {32850,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32872,32879},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32850,26215},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32854,-1},
        {-1},
        {-1,-1},
        {32854,-1,-1}
    },
    {
        {32854,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32854,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {16441,-1},
        {-1,-1,-1}
    },
    {
        {32856,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30555,-1},
        {-1,-1,-1}
    },
    {
        {32858,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30565,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36539,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39320,-1},
        {39320,-1,-1}
    },
    {
        {32862,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32863,-1},
        {-1},
        {-1,-1},
        {22779,23167,-1}
    },
    {
        {32863,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32865,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32866,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29760,-1},
        {-1,-1,-1}
    },
    {
        {32870,-1},
        {-1},
        {-1,-1},
        {32880,32865,-1}
    },
    {
        {-1,-1},
        {-1},
        {30589,-1},
        {-1,-1,-1}
    },
    {
        {32872,-1},
        {-1},
        {-1,-1},
        {32879,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32865,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24871,23231,-1}
    },
    {
        {-1,-1},
        {-1},
        {32879,-1},
        {32879,-1,-1}
    },
    {
        {32879,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32880,-1},
        {-1},
        {-1,-1},
        {32865,-1,-1}
    },
    {
        {32880,-1},
        {-1},
        {-1,-1},
        {32865,-1,-1}
    },
    {
        {32880,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32881,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32882,-1},
        {-1},
        {-1,-1},
        {22768,-1,-1}
    },
    {
        {32882,-1},
        {-1},
        {-1,-1},
        {22768,-1,-1}
    },
    {
        {32883,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32883,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32884,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32884,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32885,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32886,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32886,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32887,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32887,-1},
        {-1},
        {-1,-1},
        {32887,-1,-1}
    },
    {
        {32887,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20747,30651},
        {-1,-1,-1}
    },
    {
        {32889,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32889,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25830,-1},
        {-1,-1,-1}
    },
    {
        {32891,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32884,32893},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32884,32893},
        {32884,32893,-1}
    },
    {
        {32893,-1},
        {-1},
        {-1,-1},
        {32884,-1,-1}
    },
    {
        {32893,-1},
        {-1},
        {-1,-1},
        {32884,-1,-1}
    },
    {
        {32894,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32894,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32894,-1},
        {-1},
        {-1,-1},
        {32894,-1,-1}
    },
    {
        {32895,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24459,-1},
        {-1,-1,-1}
    },
    {
        {32900,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32901,-1},
        {-1},
        {-1,-1},
        {31899,-1,-1}
    },
    {
        {32902,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32903,-1},
        {-1},
        {-1,-1},
        {32903,-1,-1}
    },
    {
        {32903,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32903,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32905,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33222,-1},
        {33222,-1,-1}
    },
    {
        {32907,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32908,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32907,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32943,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32943,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32905,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33011,39612},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32970,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {144338,32907},
        {-1,-1,-1}
    },
    {
        {32915,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32915,-1},
        {-1},
        {-1,-1},
        {32915,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36004,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38860,-1},
        {38765,-1,-1}
    },
    {
        {32918,-1},
        {-1},
        {-1,-1},
        {32918,-1,-1}
    },
    {
        {32918,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32920,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32921,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32922,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32923,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32924,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32925,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32929,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32930,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20989,-1},
        {-1,-1,-1}
    },
    {
        {32932,-1},
        {-1},
        {-1,-1},
        {33178,-1,-1}
    },
    {
        {32933,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32935,-1},
        {-1},
        {-1,-1},
        {32986,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32982,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32982,-1},
        {-1,-1,-1}
    },
    {
        {32937,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32937,-1},
        {-1},
        {-1,-1},
        {32937,-1,-1}
    },
    {
        {32938,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32940,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32941,-1},
        {-1},
        {-1,-1},
        {26386,-1,-1}
    },
    {
        {32941,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32941,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20130,39599},
        {-1,-1,-1}
    },
    {
        {32943,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28982,-1},
        {-1,-1,-1}
    },
    {
        {32945,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32946,-1},
        {-1},
        {-1,-1},
        {32946,-1,-1}
    },
    {
        {32946,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21563,-1},
        {-1,-1,-1}
    },
    {
        {32948,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26133,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39616,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27450,32965},
        {-1,-1,-1}
    },
    {
        {32952,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32952,-1,-1}
    },
    {
        {32954,-1},
        {-1},
        {-1,-1},
        {32954,-1,-1}
    },
    {
        {32954,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32943,-1},
        {32943,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32987,-1},
        {-1,-1,-1}
    },
    {
        {32963,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32964,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32965,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32966,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32954,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39622,-1},
        {-1,-1,-1}
    },
    {
        {32970,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36028,-1},
        {-1,-1,-1}
    },
    {
        {32972,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32974,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26383,-1},
        {26383,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32930,26547},
        {32930,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32925,167478},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32981,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32982,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32982,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32983,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32984,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32985,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32986,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32987,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33125,21213},
        {-1,-1,-1}
    },
    {
        {32989,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32990,-1},
        {-1},
        {-1,-1},
        {32990,-1,-1}
    },
    {
        {32990,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25287,-1},
        {25287,-1,-1}
    },
    {
        {32992,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32993,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {19402,-1},
        {19402,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38465,-1},
        {-1,-1,-1}
    },
    {
        {32996,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32997,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32998,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21693,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33007,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33009,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33011,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33012,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33013,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33016,21256},
        {33016,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33016,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33019,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33020,-1},
        {-1},
        {-1,-1},
        {33020,-1,-1}
    },
    {
        {33021,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30509,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33034,-1},
        {-1,-1,-1}
    },
    {
        {33026,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33030,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24674,-1},
        {-1,-1,-1}
    },
    {
        {33029,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33030,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33030,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33031,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33032,-1},
        {-1},
        {-1,-1},
        {33032,-1,-1}
    },
    {
        {33032,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33032,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33033,-1},
        {-1},
        {-1,-1},
        {33032,-1,-1}
    },
    {
        {33034,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33029,-1,-1}
    },
    {
        {33037,-1},
        {-1},
        {-1,-1},
        {33214,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33248,-1},
        {33248,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33046,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21563,-1,-1}
    },
    {
        {33048,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36053,-1},
        {-1,-1,-1}
    },
    {
        {33050,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33051,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33054,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33248,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25402,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33059,-1},
        {-1},
        {-1,-1},
        {21767,-1,-1}
    },
    {
        {33060,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38960,33029},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33193,24530},
        {-1,-1,-1}
    },
    {
        {33063,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33063,-1},
        {-1},
        {20840,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30240,-1},
        {-1,-1,-1}
    },
    {
        {33065,-1},
        {-1},
        {-1,-1},
        {33065,-1,-1}
    },
    {
        {33065,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30542,-1},
        {-1,-1,-1}
    },
    {
        {33073,-1},
        {-1},
        {-1,-1},
        {33073,-1,-1}
    },
    {
        {33068,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33077,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39870,-1},
        {-1,-1,-1}
    },
    {
        {33071,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33073,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23615,-1},
        {23615,-1,-1}
    },
    {
        {33075,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30174,-1},
        {30174,-1,-1}
    },
    {
        {33077,-1},
        {-1},
        {-1,-1},
        {32929,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33081,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30571,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33085,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33086,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33086,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33176,20469},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25476,33127},
        {-1,-1,-1}
    },
    {
        {33020,-1},
        {-1},
        {33020,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36381,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30519,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30561,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33094,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26258,26228},
        {26228,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30568,-1},
        {-1,-1,-1}
    },
    {
        {33098,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33099,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25467,-1},
        {-1,-1,-1}
    },
    {
        {33102,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33104,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33105,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33107,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33107,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33108,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33108,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33109,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38515,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33086,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33173,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28197,-1},
        {28197,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23186,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33233,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35930,-1},
        {-1,-1,-1}
    },
    {
        {33119,-1},
        {-1},
        {-1,-1},
        {33187,-1,-1}
    },
    {
        {33120,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33120,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39619,-1},
        {39619,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33152,25573},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26263,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26263,-1},
        {-1,-1,-1}
    },
    {
        {33125,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33126,-1},
        {-1},
        {-1,-1},
        {33075,-1,-1}
    },
    {
        {33127,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33128,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33129,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36081,-1},
        {26249,-1,-1}
    },
    {
        {33131,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25545,-1},
        {-1,-1,-1}
    },
    {
        {33133,-1},
        {-1},
        {-1,-1},
        {40822,-1,-1}
    },
    {
        {33134,-1},
        {-1},
        {-1,-1},
        {38987,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33136,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33136,-1},
        {-1},
        {-1,-1},
        {33136,-1,-1}
    },
    {
        {33137,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33137,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38536,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33050,21371},
        {33050,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33050,21371},
        {-1,-1,-1}
    },
    {
        {33140,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33140,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26247,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37723,27573},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33144,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33145,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33145,-1},
        {-1},
        {-1,-1},
        {33145,-1,-1}
    },
    {
        {33146,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33155,-1},
        {-1},
        {33155,-1},
        {-1,-1,-1}
    },
    {
        {33151,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33151,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33152,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33152,-1},
        {-1},
        {-1,-1},
        {33152,-1,-1}
    },
    {
        {33153,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33154,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33155,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33156,-1},
        {-1},
        {-1,-1},
        {30185,30246,-1}
    },
    {
        {33156,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33156,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33157,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25637,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25637,-1},
        {-1,-1,-1}
    },
    {
        {33160,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33162,-1},
        {-1},
        {-1,-1},
        {33162,-1,-1}
    },
    {
        {33162,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30240,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33167,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33154,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36544,-1},
        {-1,-1,-1}
    },
    {
        {33171,-1},
        {-1},
        {-1,-1},
        {33144,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33176,-1},
        {-1,-1,-1}
    },
    {
        {33173,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30368,-1},
        {-1,-1,-1}
    },
    {
        {33176,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35137,-1},
        {-1,-1,-1}
    },
    {
        {33178,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33180,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33180,-1},
        {-1},
        {-1,-1},
        {33180,-1,-1}
    },
    {
        {33181,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33184,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33184,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36088,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33187,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33188,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33188,-1},
        {-1},
        {-1,-1},
        {33188,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30643,-1},
        {-1,-1,-1}
    },
    {
        {33192,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33193,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33152,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25005,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33030,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30646,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33200,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32887,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32887,-1},
        {-1,-1,-1}
    },
    {
        {33202,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33203,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33205,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33205,-1},
        {-1},
        {-1,-1},
        {33205,-1,-1}
    },
    {
        {33205,-1},
        {-1},
        {-1,-1},
        {33205,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40028,37141},
        {-1,-1,-1}
    },
    {
        {33208,-1},
        {-1},
        {-1,-1},
        {33208,-1,-1}
    },
    {
        {33208,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24996,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24996,-1},
        {-1,-1,-1}
    },
    {
        {33210,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33211,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25790,33151},
        {-1,-1,-1}
    },
    {
        {33213,-1},
        {-1},
        {-1,-1},
        {32966,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33214,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33215,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33216,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33217,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33218,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33219,-1},
        {-1},
        {-1,-1},
        {30320,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33222,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33222,-1},
        {-1},
        {-1,-1},
        {33222,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25813,-1},
        {-1,-1,-1}
    },
    {
        {33224,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33224,-1},
        {-1},
        {-1,-1},
        {33240,33224,-1}
    },
    {
        {33224,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33225,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33226,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33216,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33192,-1},
        {-1,-1,-1}
    },
    {
        {33229,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33231,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26331,-1},
        {-1,-1,-1}
    },
    {
        {33233,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38616,-1},
        {-1,-1,-1}
    },
    {
        {33235,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31013,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33238,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33239,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33239,-1},
        {-1},
        {-1,-1},
        {33239,-1,-1}
    },
    {
        {33240,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33241,-1},
        {-1},
        {-1,-1},
        {21693,-1,-1}
    },
    {
        {33242,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33243,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33247,-1},
        {-1},
        {-1,-1},
        {33235,-1,-1}
    },
    {
        {33247,-1},
        {-1},
        {-1,-1},
        {33235,-1,-1}
    },
    {
        {33248,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25890,-1},
        {-1,-1,-1}
    },
    {
        {33251,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33252,-1},
        {-1},
        {-1,-1},
        {36066,-1,-1}
    },
    {
        {33253,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20879,-1},
        {-1,-1,-1}
    },
    {
        {33255,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33255,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33256,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20879,-1},
        {-1,-1,-1}
    },
    {
        {33258,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33260,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33261,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33260,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30352,30347},
        {30347,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64092,33261,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33267,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33268,-1},
        {-1},
        {-1,-1},
        {33268,-1,-1}
    },
    {
        {33268,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32779,36845},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33274,-1},
        {-1},
        {-1,-1},
        {21488,-1,-1}
    },
    {
        {33275,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33276,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33277,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33277,-1},
        {-1},
        {-1,-1},
        {33277,-1,-1}
    },
    {
        {33278,-1},
        {-1},
        {-1,-1},
        {33278,-1,-1}
    },
    {
        {33278,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33279,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33279,-1},
        {-1},
        {-1,-1},
        {33279,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33281,-1},
        {-1},
        {-1,-1},
        {33281,-1,-1}
    },
    {
        {33281,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33281,-1},
        {-1},
        {-1,-1},
        {33281,-1,-1}
    },
    {
        {33282,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33282,-1},
        {-1},
        {-1,-1},
        {33282,-1,-1}
    },
    {
        {33284,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33284,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33284,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33285,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33287,-1},
        {-1},
        {-1,-1},
        {19982,33287,-1}
    },
    {
        {33287,-1},
        {-1},
        {-1,-1},
        {19982,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33288,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33289,-1},
        {-1},
        {-1,-1},
        {25369,25831,-1}
    },
    {
        {33290,-1},
        {-1},
        {-1,-1},
        {26087,-1,-1}
    },
    {
        {33290,-1},
        {-1},
        {-1,-1},
        {26087,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37313,-1},
        {-1,-1,-1}
    },
    {
        {33292,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33292,-1},
        {-1},
        {-1,-1},
        {33292,-1,-1}
    },
    {
        {33293,-1},
        {-1},
        {-1,-1},
        {33294,-1,-1}
    },
    {
        {33294,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33296,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33298,-1},
        {-1},
        {-1,-1},
        {33298,-1,-1}
    },
    {
        {33298,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33304,-1},
        {-1},
        {-1,-1},
        {33304,39208,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33296,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33302,-1},
        {-1},
        {-1,-1},
        {37610,33303,-1}
    },
    {
        {33303,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33304,-1},
        {-1},
        {-1,-1},
        {39208,-1,-1}
    },
    {
        {33305,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33307,-1},
        {-1},
        {-1,-1},
        {21319,-1,-1}
    },
    {
        {33307,-1},
        {-1},
        {-1,-1},
        {21319,-1,-1}
    },
    {
        {33307,-1},
        {-1},
        {-1,-1},
        {33307,21319,-1}
    },
    {
        {33307,-1},
        {-1},
        {-1,-1},
        {21319,-1,-1}
    },
    {
        {33308,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33308,-1},
        {-1},
        {-1,-1},
        {33308,-1,-1}
    },
    {
        {33308,-1},
        {-1},
        {-1,-1},
        {33308,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33310,-1},
        {-1},
        {-1,-1},
        {33310,-1,-1}
    },
    {
        {33310,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33311,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33323,-1},
        {-1,-1,-1}
    },
    {
        {33313,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33314,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32933,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33322,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33322,24070},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33321,-1},
        {-1},
        {-1,-1},
        {33337,-1,-1}
    },
    {
        {33321,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33322,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33323,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33324,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33340,-1},
        {-1,-1,-1}
    },
    {
        {33326,-1},
        {-1},
        {-1,-1},
        {33387,-1,-1}
    },
    {
        {33326,-1},
        {-1},
        {-1,-1},
        {33387,-1,-1}
    },
    {
        {33330,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33331,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33332,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33333,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33334,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33335,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33336,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33336,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33337,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33337,-1},
        {-1},
        {-1,-1},
        {33337,-1,-1}
    },
    {
        {33337,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33340,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33344,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33344,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36896,-1},
        {36896,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34562,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33349,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33350,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33351,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33351,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33353,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33334,-1},
        {33334,-1,-1}
    },
    {
        {33355,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33330,31252},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33359,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23626,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {158391,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24137,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33323,-1},
        {33323,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33323,-1},
        {33323,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33367,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33367,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33368,-1},
        {-1},
        {-1,-1},
        {33368,-1,-1}
    },
    {
        {33368,-1},
        {-1},
        {-1,-1},
        {33368,-1,-1}
    },
    {
        {33368,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33369,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33370,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33372,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33373,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33373,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33375,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33375,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33376,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25803,-1},
        {-1,-1,-1}
    },
    {
        {33378,-1},
        {-1},
        {-1,-1},
        {27299,-1,-1}
    },
    {
        {33379,-1},
        {-1},
        {-1,-1},
        {33386,-1,-1}
    },
    {
        {33380,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26987,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26987,-1},
        {-1,-1,-1}
    },
    {
        {33382,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33384,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33384,-1},
        {-1},
        {-1,-1},
        {33384,-1,-1}
    },
    {
        {33384,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33229,-1},
        {-1,-1,-1}
    },
    {
        {33386,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33387,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33390,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33390,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33391,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33393,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33393,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33394,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33396,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33396,29942},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33396,29942},
        {-1,-1,-1}
    },
    {
        {33398,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33399,-1},
        {-1},
        {-1,-1},
        {33398,-1,-1}
    },
    {
        {33400,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33401,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37194,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33405,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33405,-1},
        {-1},
        {-1,-1},
        {33405,-1,-1}
    },
    {
        {33406,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33406,-1},
        {-1},
        {-1,-1},
        {33406,-1,-1}
    },
    {
        {33406,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33407,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33407,-1},
        {-1},
        {-1,-1},
        {33407,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33411,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33411,-1},
        {-1},
        {-1,-1},
        {33411,-1,-1}
    },
    {
        {33411,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24331,33543},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33406,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33418,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33418,-1},
        {-1},
        {-1,-1},
        {33418,-1,-1}
    },
    {
        {33419,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33419,-1},
        {-1},
        {-1,-1},
        {33419,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33419,-1},
        {33419,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33419,-1},
        {33419,-1,-1}
    },
    {
        {33421,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33421,-1},
        {-1},
        {-1,-1},
        {33421,-1,-1}
    },
    {
        {33421,-1},
        {-1},
        {-1,-1},
        {33421,-1,-1}
    },
    {
        {33422,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33422,-1},
        {-1},
        {-1,-1},
        {33422,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33550,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33446,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33426,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33426,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33426,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33427,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33427,-1},
        {-1},
        {-1,-1},
        {33427,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21321,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21321,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33661,-1},
        {33661,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33433,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33433,-1},
        {-1},
        {-1,-1},
        {33433,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28961,34154},
        {-1,-1,-1}
    },
    {
        {33437,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33437,-1},
        {-1},
        {-1,-1},
        {33437,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27668,-1},
        {-1,-1,-1}
    },
    {
        {33439,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33439,-1},
        {-1},
        {-1,-1},
        {33439,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33441,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33441,-1},
        {-1},
        {-1,-1},
        {33441,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33548,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33443,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33443,-1},
        {-1},
        {-1,-1},
        {33443,-1,-1}
    },
    {
        {33444,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33444,-1},
        {-1},
        {-1,-1},
        {33444,-1,-1}
    },
    {
        {33445,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33445,-1},
        {-1},
        {-1,-1},
        {33445,-1,-1}
    },
    {
        {33446,-1},
        {-1},
        {-1,-1},
        {33446,-1,-1}
    },
    {
        {33446,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33446,-1},
        {-1},
        {-1,-1},
        {33446,-1,-1}
    },
    {
        {33446,-1},
        {-1},
        {-1,-1},
        {33446,34310,-1}
    },
    {
        {33446,-1},
        {-1},
        {-1,-1},
        {33446,34310,-1}
    },
    {
        {33446,-1},
        {-1},
        {-1,-1},
        {33446,34310,-1}
    },
    {
        {33446,-1},
        {-1},
        {-1,-1},
        {33446,34310,-1}
    },
    {
        {33446,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33447,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33447,-1},
        {-1},
        {-1,-1},
        {33447,-1,-1}
    },
    {
        {33448,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33448,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33449,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33449,-1},
        {-1},
        {-1,-1},
        {33449,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33451,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33451,-1},
        {-1},
        {-1,-1},
        {33451,-1,-1}
    },
    {
        {33452,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33452,-1},
        {-1},
        {-1,-1},
        {33452,-1,-1}
    },
    {
        {33452,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33453,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33453,-1},
        {-1},
        {-1,-1},
        {33453,-1,-1}
    },
    {
        {33454,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33454,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33454,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33455,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33455,-1},
        {-1},
        {-1,-1},
        {33455,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33457,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33457,-1},
        {-1},
        {-1,-1},
        {33457,-1,-1}
    },
    {
        {33457,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33457,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33458,-1},
        {-1},
        {-1,-1},
        {33457,-1,-1}
    },
    {
        {33458,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33459,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33459,-1},
        {-1},
        {-1,-1},
        {33459,-1,-1}
    },
    {
        {33460,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33460,-1},
        {-1},
        {-1,-1},
        {33460,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33521,22828},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33503,-1},
        {-1,-1,-1}
    },
    {
        {33463,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33463,-1},
        {-1},
        {-1,-1},
        {33463,-1,-1}
    },
    {
        {33464,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33464,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33465,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33465,-1},
        {-1},
        {-1,-1},
        {33465,-1,-1}
    },
    {
        {33466,-1},
        {-1},
        {-1,-1},
        {33433,-1,-1}
    },
    {
        {33466,-1},
        {-1},
        {-1,-1},
        {33433,-1,-1}
    },
    {
        {33467,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33468,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33468,-1},
        {-1},
        {-1,-1},
        {33468,-1,-1}
    },
    {
        {33469,-1},
        {-1},
        {-1,-1},
        {33469,-1,-1}
    },
    {
        {33469,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33469,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33470,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33470,-1},
        {-1},
        {-1,-1},
        {33470,-1,-1}
    },
    {
        {33471,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33471,-1},
        {-1},
        {-1,-1},
        {33471,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20130,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24467,34031},
        {24467,24478,-1}
    },
    {
        {-1,-1},
        {-1},
        {28823,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33543,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31495,-1},
        {-1,-1,-1}
    },
    {
        {33477,-1},
        {-1},
        {-1,-1},
        {21000,-1,-1}
    },
    {
        {33477,-1},
        {-1},
        {-1,-1},
        {21000,33477,-1}
    },
    {
        {33478,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33478,-1},
        {-1},
        {-1,-1},
        {33478,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33617,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33617,-1,-1}
    },
    {
        {33489,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33489,-1},
        {-1},
        {-1,-1},
        {33489,-1,-1}
    },
    {
        {33489,-1},
        {-1},
        {-1,-1},
        {33489,-1,-1}
    },
    {
        {33490,-1},
        {-1},
        {-1,-1},
        {33490,-1,-1}
    },
    {
        {33490,-1},
        {-1},
        {-1,-1},
        {33490,-1,-1}
    },
    {
        {33490,-1},
        {-1},
        {-1,-1},
        {33490,-1,-1}
    },
    {
        {33491,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33491,-1},
        {-1},
        {-1,-1},
        {33491,-1,-1}
    },
    {
        {33492,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33492,-1},
        {-1},
        {-1,-1},
        {33492,-1,-1}
    },
    {
        {33493,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33493,-1},
        {-1},
        {-1,-1},
        {33493,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31515,-1},
        {33495,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31515,-1},
        {33495,-1,-1}
    },
    {
        {33495,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33495,-1},
        {-1},
        {-1,-1},
        {33495,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33684,-1},
        {-1,-1,-1}
    },
    {
        {33497,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33497,-1},
        {-1},
        {-1,-1},
        {33497,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33499,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33499,-1},
        {-1},
        {-1,-1},
        {33499,-1,-1}
    },
    {
        {33500,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33500,-1},
        {-1},
        {-1,-1},
        {33500,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33502,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33502,-1},
        {-1},
        {-1,-1},
        {33502,-1,-1}
    },
    {
        {33502,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33503,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33503,-1},
        {-1},
        {-1,-1},
        {33503,-1,-1}
    },
    {
        {33503,-1},
        {-1},
        {-1,-1},
        {33503,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33505,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33505,-1},
        {-1},
        {-1,-1},
        {33505,-1,-1}
    },
    {
        {33506,-1},
        {-1},
        {-1,-1},
        {33505,-1,-1}
    },
    {
        {33506,-1},
        {-1},
        {-1,-1},
        {33505,33506,-1}
    },
    {
        {33507,-1},
        {-1},
        {-1,-1},
        {33507,-1,-1}
    },
    {
        {33507,-1},
        {-1},
        {-1,-1},
        {33507,-1,-1}
    },
    {
        {33507,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33509,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33509,-1,-1}
    },
    {
        {33509,-1},
        {-1},
        {-1,-1},
        {33509,-1,-1}
    },
    {
        {33509,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33510,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33510,-1},
        {-1},
        {-1,-1},
        {33510,-1,-1}
    },
    {
        {33511,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33511,-1},
        {-1},
        {-1,-1},
        {33511,-1,-1}
    },
    {
        {33512,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33512,-1},
        {-1},
        {-1,-1},
        {33512,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33897,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33515,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33515,-1},
        {-1},
        {-1,-1},
        {33515,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33589,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20185,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23446,-1},
        {-1,-1,-1}
    },
    {
        {33521,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33521,-1},
        {-1},
        {-1,-1},
        {33521,-1,-1}
    },
    {
        {33521,-1},
        {-1},
        {-1,-1},
        {33521,-1,-1}
    },
    {
        {33521,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33523,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33523,-1},
        {-1},
        {-1,-1},
        {33523,-1,-1}
    },
    {
        {33523,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33524,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33524,-1},
        {-1},
        {-1,-1},
        {33524,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33433,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33527,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33527,-1},
        {-1},
        {-1,-1},
        {33527,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33529,-1},
        {-1,-1,-1}
    },
    {
        {33529,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33529,-1},
        {-1},
        {-1,-1},
        {33529,-1,-1}
    },
    {
        {33529,-1},
        {-1},
        {-1,-1},
        {33529,-1,-1}
    },
    {
        {33530,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33530,-1},
        {-1},
        {-1,-1},
        {33530,33683,-1}
    },
    {
        {33531,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33531,-1},
        {-1},
        {-1,-1},
        {33531,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34148,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33533,-1},
        {-1},
        {-1,-1},
        {33776,33533,-1}
    },
    {
        {33533,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33533,-1},
        {-1},
        {-1,-1},
        {33776,-1,-1}
    },
    {
        {33534,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33534,-1},
        {-1},
        {-1,-1},
        {33534,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33545,-1},
        {21619,-1,-1}
    },
    {
        {33536,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33536,-1},
        {-1},
        {-1,-1},
        {33536,-1,-1}
    },
    {
        {33537,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33537,-1},
        {-1},
        {-1,-1},
        {33537,-1,-1}
    },
    {
        {33538,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33538,-1},
        {-1},
        {-1,-1},
        {33538,-1,-1}
    },
    {
        {33538,-1},
        {-1},
        {-1,-1},
        {33538,-1,-1}
    },
    {
        {33539,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33539,-1},
        {-1},
        {-1,-1},
        {33539,-1,-1}
    },
    {
        {33539,-1},
        {-1},
        {-1,-1},
        {33539,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33540,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33540,-1},
        {-1},
        {-1,-1},
        {33540,-1,-1}
    },
    {
        {33541,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33541,-1},
        {-1},
        {-1,-1},
        {33541,-1,-1}
    },
    {
        {33542,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33542,-1},
        {-1},
        {-1,-1},
        {33542,-1,-1}
    },
    {
        {33543,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33543,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33544,-1},
        {-1},
        {-1,-1},
        {33544,-1,-1}
    },
    {
        {33544,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33545,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33545,-1},
        {-1},
        {-1,-1},
        {33545,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33586,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33548,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33548,-1},
        {-1},
        {-1,-1},
        {33548,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33503,-1},
        {-1},
        {33503,-1},
        {-1,-1,-1}
    },
    {
        {33550,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33558,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33558,-1},
        {-1},
        {-1,-1},
        {33558,-1,-1}
    },
    {
        {33559,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33559,-1},
        {-1},
        {-1,-1},
        {33559,-1,-1}
    },
    {
        {33560,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33560,-1},
        {-1},
        {-1,-1},
        {33560,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21360,-1},
        {-1,-1,-1}
    },
    {
        {33563,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33563,-1},
        {-1},
        {-1,-1},
        {33563,-1,-1}
    },
    {
        {33564,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33564,-1},
        {-1},
        {-1,-1},
        {33564,-1,-1}
    },
    {
        {33565,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33565,-1},
        {-1},
        {-1,-1},
        {33565,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20241,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33427,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33427,-1,-1}
    },
    {
        {33570,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33570,-1},
        {-1},
        {-1,-1},
        {33570,-1,-1}
    },
    {
        {33571,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33571,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33571,-1},
        {-1},
        {-1,-1},
        {33571,-1,-1}
    },
    {
        {33571,-1},
        {-1},
        {-1,-1},
        {33571,-1,-1}
    },
    {
        {33571,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31574,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31574,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32365,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32365,-1},
        {-1,-1,-1}
    },
    {
        {33576,-1},
        {-1},
        {-1,-1},
        {33576,-1,-1}
    },
    {
        {33576,-1},
        {-1},
        {-1,-1},
        {33576,-1,-1}
    },
    {
        {33576,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21518,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20809,33451},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33579,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33579,-1},
        {-1},
        {-1,-1},
        {33579,-1,-1}
    },
    {
        {33579,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33581,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33581,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33583,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33583,-1},
        {-1},
        {-1,-1},
        {33583,-1,-1}
    },
    {
        {33584,-1},
        {-1},
        {-1,-1},
        {33848,-1,-1}
    },
    {
        {33584,-1},
        {-1},
        {-1,-1},
        {33584,33848,-1}
    },
    {
        {33584,-1},
        {-1},
        {-1,-1},
        {33848,-1,-1}
    },
    {
        {33584,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33584,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33585,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33585,-1},
        {-1},
        {-1,-1},
        {33585,-1,-1}
    },
    {
        {33586,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33586,-1},
        {-1},
        {-1,-1},
        {33586,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33588,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33588,-1},
        {-1},
        {-1,-1},
        {33588,-1,-1}
    },
    {
        {33589,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33589,-1},
        {-1},
        {-1,-1},
        {33589,-1,-1}
    },
    {
        {33590,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33590,-1},
        {-1},
        {-1,-1},
        {33590,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33592,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33592,-1},
        {-1},
        {-1,-1},
        {33592,-1,-1}
    },
    {
        {33592,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33593,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33593,-1},
        {-1},
        {-1,-1},
        {33593,-1,-1}
    },
    {
        {33594,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33594,-1},
        {-1},
        {-1,-1},
        {33594,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33596,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33596,-1},
        {-1},
        {-1,-1},
        {33596,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31492,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31569,-1},
        {-1,-1,-1}
    },
    {
        {33600,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33600,-1},
        {-1},
        {-1,-1},
        {33600,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33524,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33603,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33603,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33603,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33603,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33604,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33604,-1},
        {-1},
        {-1,-1},
        {33604,-1,-1}
    },
    {
        {33604,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33605,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33605,-1},
        {-1},
        {-1,-1},
        {31572,33605,-1}
    },
    {
        {33605,-1},
        {-1},
        {-1,-1},
        {33605,-1,-1}
    },
    {
        {33610,-1},
        {-1},
        {33610,-1},
        {-1,-1,-1}
    },
    {
        {33610,-1},
        {-1},
        {33610,-1},
        {33610,-1,-1}
    },
    {
        {33610,-1},
        {-1},
        {33610,-1},
        {33610,-1,-1}
    },
    {
        {33610,-1},
        {-1},
        {33610,-1},
        {33610,-1,-1}
    },
    {
        {33607,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33607,-1},
        {-1},
        {-1,-1},
        {33607,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33489,138679},
        {-1,-1,-1}
    },
    {
        {33609,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33609,-1},
        {-1},
        {-1,-1},
        {33609,-1,-1}
    },
    {
        {33610,-1},
        {-1},
        {-1,-1},
        {33610,-1,-1}
    },
    {
        {33610,-1},
        {-1},
        {-1,-1},
        {33610,-1,-1}
    },
    {
        {33610,-1},
        {-1},
        {-1,-1},
        {33610,-1,-1}
    },
    {
        {33610,-1},
        {-1},
        {-1,-1},
        {33610,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32785,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23433,33661},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33615,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33615,-1},
        {-1},
        {-1,-1},
        {33615,-1,-1}
    },
    {
        {33615,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33616,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33616,-1},
        {-1},
        {-1,-1},
        {33616,-1,-1}
    },
    {
        {33617,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33617,-1},
        {-1},
        {-1,-1},
        {33617,-1,-1}
    },
    {
        {33617,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33618,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33618,-1},
        {-1},
        {-1,-1},
        {33618,-1,-1}
    },
    {
        {33618,-1},
        {-1},
        {-1,-1},
        {33618,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33805,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33620,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33620,-1},
        {-1},
        {-1,-1},
        {33560,33620,-1}
    },
    {
        {33621,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33621,-1},
        {-1},
        {-1,-1},
        {33621,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33623,-1},
        {-1},
        {-1,-1},
        {33538,-1,-1}
    },
    {
        {33623,-1},
        {-1},
        {-1,-1},
        {33538,33623,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33538,-1,-1}
    },
    {
        {33624,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33634,-1},
        {-1},
        {-1,-1},
        {33634,33511,-1}
    },
    {
        {33634,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33634,-1},
        {-1},
        {-1,-1},
        {33511,-1,-1}
    },
    {
        {33635,-1},
        {-1},
        {-1,-1},
        {26628,27054,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23614,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33891,37324},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33651,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33651,-1},
        {-1},
        {-1,-1},
        {33651,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25206,-1},
        {-1,-1,-1}
    },
    {
        {33653,-1},
        {-1},
        {-1,-1},
        {33653,-1,-1}
    },
    {
        {33653,-1},
        {-1},
        {-1,-1},
        {33653,-1,-1}
    },
    {
        {33653,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33653,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33449,-1},
        {-1,-1,-1}
    },
    {
        {33655,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33655,-1},
        {-1},
        {-1,-1},
        {33655,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27493,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31584,-1},
        {-1,-1,-1}
    },
    {
        {33659,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33659,-1},
        {-1},
        {-1,-1},
        {33659,-1,-1}
    },
    {
        {33659,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33660,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33660,-1},
        {-1},
        {-1,-1},
        {33660,-1,-1}
    },
    {
        {33661,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33661,-1},
        {-1},
        {-1,-1},
        {33661,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33904,33661},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33904,33661},
        {-1,-1,-1}
    },
    {
        {33663,-1},
        {-1},
        {-1,-1},
        {33538,33663,-1}
    },
    {
        {33663,-1},
        {-1},
        {-1,-1},
        {33663,-1,-1}
    },
    {
        {33663,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33663,-1},
        {-1},
        {-1,-1},
        {33538,-1,-1}
    },
    {
        {33663,-1},
        {-1},
        {-1,-1},
        {33538,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31598,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24076,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33669,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33669,-1},
        {-1},
        {-1,-1},
        {33669,-1,-1}
    },
    {
        {33670,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33670,-1},
        {-1},
        {-1,-1},
        {33670,-1,-1}
    },
    {
        {33671,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33671,-1},
        {-1},
        {-1,-1},
        {33671,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27794,33853},
        {-1,-1,-1}
    },
    {
        {33673,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33673,-1},
        {-1},
        {-1,-1},
        {33673,-1,-1}
    },
    {
        {33674,-1},
        {-1},
        {-1,-1},
        {33624,-1,-1}
    },
    {
        {33674,-1},
        {-1},
        {-1,-1},
        {33624,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31600,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20823,-1},
        {-1,-1,-1}
    },
    {
        {33677,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33677,-1},
        {-1},
        {-1,-1},
        {33677,-1,-1}
    },
    {
        {33678,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33678,-1},
        {-1},
        {-1,-1},
        {33678,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33678,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27784,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34028,-1},
        {34028,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34028,-1},
        {34028,-1,-1}
    },
    {
        {33682,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33682,-1},
        {-1},
        {-1,-1},
        {33682,-1,-1}
    },
    {
        {33683,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33683,-1},
        {-1},
        {-1,-1},
        {33683,-1,-1}
    },
    {
        {33683,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33684,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33684,-1},
        {-1},
        {-1,-1},
        {33684,-1,-1}
    },
    {
        {33685,-1},
        {-1},
        {-1,-1},
        {33607,-1,-1}
    },
    {
        {33685,-1},
        {-1},
        {-1,-1},
        {33685,33607,-1}
    },
    {
        {33686,-1},
        {-1},
        {-1,-1},
        {33550,-1,-1}
    },
    {
        {33686,-1},
        {-1},
        {-1,-1},
        {33550,-1,-1}
    },
    {
        {33686,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33911,-1},
        {-1,-1,-1}
    },
    {
        {33688,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33688,-1},
        {-1},
        {-1,-1},
        {33688,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33690,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33690,-1},
        {-1},
        {-1,-1},
        {33690,-1,-1}
    },
    {
        {33690,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33691,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33691,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33691,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33693,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33693,-1},
        {-1},
        {-1,-1},
        {33693,-1,-1}
    },
    {
        {33694,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33694,-1},
        {-1},
        {-1,-1},
        {33694,-1,-1}
    },
    {
        {33695,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33695,-1},
        {-1},
        {-1,-1},
        {33695,-1,-1}
    },
    {
        {33696,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33696,-1},
        {-1},
        {-1,-1},
        {33696,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36275,-1},
        {-1,-1,-1}
    },
    {
        {33698,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33698,-1},
        {-1},
        {-1,-1},
        {33698,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33564,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33564,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29379,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33703,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33703,-1},
        {-1},
        {-1,-1},
        {33703,-1,-1}
    },
    {
        {33704,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33704,-1},
        {-1},
        {-1,-1},
        {33704,-1,-1}
    },
    {
        {33705,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33705,-1},
        {-1},
        {-1,-1},
        {33705,-1,-1}
    },
    {
        {33706,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33706,-1},
        {-1},
        {-1,-1},
        {33706,-1,-1}
    },
    {
        {33707,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33707,-1},
        {-1},
        {-1,-1},
        {33707,-1,-1}
    },
    {
        {33759,-1},
        {-1},
        {-1,-1},
        {33759,-1,-1}
    },
    {
        {33759,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31680,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30007,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27792,-1},
        {-1,-1,-1}
    },
    {
        {33713,-1},
        {-1},
        {-1,-1},
        {33802,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33717,-1},
        {-1},
        {-1,-1},
        {33759,-1,-1}
    },
    {
        {33717,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33717,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33725,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33725,-1},
        {-1},
        {-1,-1},
        {33725,-1,-1}
    },
    {
        {33725,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33725,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33725,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33727,-1},
        {-1},
        {-1,-1},
        {33727,-1,-1}
    },
    {
        {33727,-1},
        {-1},
        {-1,-1},
        {33727,-1,-1}
    },
    {
        {33728,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33728,-1},
        {-1},
        {-1,-1},
        {33728,-1,-1}
    },
    {
        {33729,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33729,-1},
        {-1},
        {-1,-1},
        {33729,-1,-1}
    },
    {
        {33729,-1},
        {-1},
        {-1,-1},
        {33729,-1,-1}
    },
    {
        {33729,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30340,33421},
        {33421,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33507,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26481,-1},
        {-1,-1,-1}
    },
    {
        {33733,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33733,-1},
        {-1},
        {-1,-1},
        {33733,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21474,26871},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21474,26871},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21474,26871},
        {-1,-1,-1}
    },
    {
        {33735,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33735,-1},
        {-1},
        {-1,-1},
        {33735,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25289,33669},
        {-1,-1,-1}
    },
    {
        {33737,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33737,-1},
        {-1},
        {-1,-1},
        {33737,-1,-1}
    },
    {
        {33738,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33738,-1},
        {-1},
        {-1,-1},
        {33738,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21619,-1},
        {-1,-1,-1}
    },
    {
        {33740,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33740,-1},
        {-1},
        {-1,-1},
        {33740,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26863,-1},
        {-1,-1,-1}
    },
    {
        {33742,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33742,-1},
        {-1},
        {-1,-1},
        {33742,-1,-1}
    },
    {
        {33743,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33743,-1},
        {-1},
        {-1,-1},
        {33743,-1,-1}
    },
    {
        {33744,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33745,-1},
        {-1},
        {-1,-1},
        {33745,-1,-1}
    },
    {
        {33745,-1},
        {-1},
        {-1,-1},
        {33745,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34241,-1},
        {31294,31295,-1}
    },
    {
        {-1,-1},
        {-1},
        {34241,-1},
        {31294,31295,-1}
    },
    {
        {33747,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33747,-1},
        {-1},
        {-1,-1},
        {33747,-1,-1}
    },
    {
        {33748,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33748,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23833,-1},
        {-1,-1,-1}
    },
    {
        {33750,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33750,-1},
        {-1},
        {-1,-1},
        {33750,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25551,-1},
        {-1,-1,-1}
    },
    {
        {33752,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33752,-1},
        {-1},
        {-1,-1},
        {33752,-1,-1}
    },
    {
        {33752,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31627,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38272,34349},
        {-1,-1,-1}
    },
    {
        {33756,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33756,-1},
        {-1},
        {-1,-1},
        {33756,-1,-1}
    },
    {
        {33756,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33757,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33757,-1},
        {-1},
        {-1,-1},
        {33757,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33673,40654},
        {-1,-1,-1}
    },
    {
        {33759,-1},
        {-1},
        {-1,-1},
        {33759,-1,-1}
    },
    {
        {33759,-1},
        {-1},
        {-1,-1},
        {33759,-1,-1}
    },
    {
        {33759,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33759,-1},
        {-1},
        {-1,-1},
        {33759,-1,-1}
    },
    {
        {33759,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33759,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33759,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33759,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33759,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33760,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33760,-1},
        {-1},
        {-1,-1},
        {33760,-1,-1}
    },
    {
        {33761,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33761,-1},
        {-1},
        {-1,-1},
        {33761,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25265,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33252,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31646,-1},
        {-1,-1,-1}
    },
    {
        {33765,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33765,-1},
        {-1},
        {-1,-1},
        {33765,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33465,-1},
        {33465,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33769,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33769,-1},
        {-1},
        {-1,-1},
        {33769,-1,-1}
    },
    {
        {33770,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33770,-1},
        {-1},
        {-1,-1},
        {33770,-1,-1}
    },
    {
        {33771,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33771,-1},
        {-1},
        {-1,-1},
        {33771,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27836,33493},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33492,-1},
        {33492,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24218,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33775,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33775,-1},
        {-1},
        {-1,-1},
        {33775,-1,-1}
    },
    {
        {33775,-1},
        {-1},
        {-1,-1},
        {33775,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33776,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33776,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33776,-1},
        {-1},
        {-1,-1},
        {33776,-1,-1}
    },
    {
        {33777,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33777,-1},
        {-1},
        {-1,-1},
        {33777,-1,-1}
    },
    {
        {33777,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33777,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33778,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33778,-1},
        {-1},
        {-1,-1},
        {33778,-1,-1}
    },
    {
        {33778,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33449,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33449,-1},
        {-1,-1,-1}
    },
    {
        {33780,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33780,-1},
        {-1},
        {-1,-1},
        {33780,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32596,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33783,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33783,-1},
        {-1},
        {-1,-1},
        {24090,33783,-1}
    },
    {
        {33783,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33785,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33785,-1},
        {-1},
        {-1,-1},
        {33785,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32937,33849},
        {-1,-1,-1}
    },
    {
        {33787,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33787,-1},
        {-1},
        {-1,-1},
        {33787,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33789,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33789,-1},
        {-1},
        {-1,-1},
        {33789,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24541,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21040,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33793,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33793,-1},
        {-1},
        {-1,-1},
        {33793,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21644,-1},
        {-1,-1,-1}
    },
    {
        {33795,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33795,-1},
        {-1},
        {-1,-1},
        {33795,-1,-1}
    },
    {
        {33795,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33796,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33796,-1},
        {-1},
        {-1,-1},
        {33796,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26149,-1},
        {26149,-1,-1}
    },
    {
        {33798,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33798,-1},
        {-1},
        {-1,-1},
        {33798,-1,-1}
    },
    {
        {33799,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33799,-1},
        {-1},
        {-1,-1},
        {33799,-1,-1}
    },
    {
        {33703,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33897,32933},
        {-1,-1,-1}
    },
    {
        {33802,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33802,-1},
        {-1},
        {-1,-1},
        {33802,-1,-1}
    },
    {
        {33803,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33803,-1},
        {-1},
        {-1,-1},
        {33803,-1,-1}
    },
    {
        {33803,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33804,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33804,-1},
        {-1},
        {-1,-1},
        {33804,-1,-1}
    },
    {
        {33805,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33805,-1},
        {-1},
        {-1,-1},
        {33805,-1,-1}
    },
    {
        {33805,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33806,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33806,-1},
        {-1},
        {-1,-1},
        {33806,-1,-1}
    },
    {
        {33807,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33807,-1},
        {-1},
        {-1,-1},
        {33807,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31633,-1},
        {-1,-1,-1}
    },
    {
        {33809,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33809,-1},
        {-1},
        {-1,-1},
        {33809,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20823,-1},
        {-1,-1,-1}
    },
    {
        {33811,-1},
        {-1},
        {-1,-1},
        {33841,-1,-1}
    },
    {
        {33811,-1},
        {-1},
        {-1,-1},
        {33841,-1,-1}
    },
    {
        {33811,-1},
        {-1},
        {-1,-1},
        {33841,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33493,-1},
        {33493,-1,-1}
    },
    {
        {33813,-1},
        {-1},
        {-1,-1},
        {34234,33813,-1}
    },
    {
        {33813,-1},
        {-1},
        {-1,-1},
        {34234,33813,-1}
    },
    {
        {33813,-1},
        {-1},
        {-1,-1},
        {34234,-1,-1}
    },
    {
        {33815,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22856,33660},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31622,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31622,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33833,31175},
        {-1,-1,-1}
    },
    {
        {33824,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33824,-1},
        {-1},
        {-1,-1},
        {33804,-1,-1}
    },
    {
        {33824,-1},
        {-1},
        {-1,-1},
        {33824,33804,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33826,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33826,-1},
        {-1},
        {-1,-1},
        {33826,-1,-1}
    },
    {
        {33826,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34722,34509},
        {34509,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21942,29151},
        {21942,29151,-1}
    },
    {
        {33833,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33833,-1},
        {-1},
        {-1,-1},
        {33833,-1,-1}
    },
    {
        {33834,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33834,-1},
        {-1},
        {-1,-1},
        {33834,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39321,34115},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39321,34115},
        {-1,-1,-1}
    },
    {
        {33836,-1},
        {-1},
        {-1,-1},
        {19975,-1,-1}
    },
    {
        {33836,-1},
        {-1},
        {-1,-1},
        {19975,-1,-1}
    },
    {
        {33836,-1},
        {-1},
        {-1,-1},
        {19975,-1,-1}
    },
    {
        {33836,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30697,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20842,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33713,26604},
        {-1,-1,-1}
    },
    {
        {33841,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33841,-1},
        {-1},
        {-1,-1},
        {33841,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33841,33811},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23857,-1},
        {-1,-1,-1}
    },
    {
        {33845,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33845,-1},
        {-1},
        {-1,-1},
        {33845,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26149,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26786,-1},
        {-1,-1,-1}
    },
    {
        {33848,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33848,-1},
        {-1},
        {-1,-1},
        {33848,-1,-1}
    },
    {
        {33849,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33849,-1},
        {-1},
        {-1,-1},
        {33849,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20882,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33769,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33769,-1},
        {-1,-1,-1}
    },
    {
        {33852,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33852,-1},
        {-1},
        {-1,-1},
        {33852,-1,-1}
    },
    {
        {33853,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33853,-1},
        {-1},
        {-1,-1},
        {33853,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27963,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25324,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34193,-1},
        {34193,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34193,-1},
        {34193,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {64051,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26152,33940},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32985,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33785,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33785,-1,-1}
    },
    {
        {33862,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33862,-1},
        {-1},
        {-1,-1},
        {33862,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33864,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33864,-1},
        {-1},
        {-1,-1},
        {33864,-1,-1}
    },
    {
        {33865,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33865,-1},
        {-1},
        {-1,-1},
        {33865,-1,-1}
    },
    {
        {33865,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33866,-1},
        {-1},
        {-1,-1},
        {33780,-1,-1}
    },
    {
        {33866,-1},
        {-1},
        {-1,-1},
        {33866,33780,-1}
    },
    {
        {-1,-1},
        {-1},
        {32970,33621},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34129,33733},
        {34129,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33870,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33870,-1},
        {-1},
        {-1,-1},
        {33870,-1,-1}
    },
    {
        {33871,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33871,-1},
        {-1},
        {-1,-1},
        {33871,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33995,30406},
        {-1,-1,-1}
    },
    {
        {33873,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33873,-1},
        {-1},
        {-1,-1},
        {33873,-1,-1}
    },
    {
        {33874,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33874,-1},
        {-1},
        {-1,-1},
        {33874,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31690,33874},
        {33874,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31692,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33690,-1},
        {33690,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33879,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33879,-1},
        {-1},
        {-1,-1},
        {33879,-1,-1}
    },
    {
        {33745,-1},
        {-1},
        {-1,-1},
        {33745,-1,-1}
    },
    {
        {33745,-1},
        {-1},
        {-1,-1},
        {33745,-1,-1}
    },
    {
        {33881,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33881,-1},
        {-1},
        {-1,-1},
        {34328,33881,-1}
    },
    {
        {33882,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33882,-1},
        {-1},
        {-1,-1},
        {33882,-1,-1}
    },
    {
        {33882,-1},
        {-1},
        {-1,-1},
        {33882,-1,-1}
    },
    {
        {33883,-1},
        {-1},
        {-1,-1},
        {33883,-1,-1}
    },
    {
        {33883,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33883,-1},
        {-1},
        {-1,-1},
        {33883,-1,-1}
    },
    {
        {33883,-1},
        {-1},
        {-1,-1},
        {33883,-1,-1}
    },
    {
        {33883,-1},
        {-1},
        {-1,-1},
        {33883,-1,-1}
    },
    {
        {33883,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33883,-1},
        {-1},
        {-1,-1},
        {33883,-1,-1}
    },
    {
        {33884,-1},
        {-1},
        {-1,-1},
        {33884,-1,-1}
    },
    {
        {33884,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33884,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21185,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24365,-1},
        {-1,-1,-1}
    },
    {
        {33887,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33887,-1},
        {-1},
        {-1,-1},
        {33887,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33889,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33889,-1},
        {-1},
        {-1,-1},
        {33889,-1,-1}
    },
    {
        {33890,-1},
        {-1},
        {-1,-1},
        {33995,-1,-1}
    },
    {
        {33890,-1},
        {-1},
        {-1,-1},
        {33890,33995,-1}
    },
    {
        {33890,-1},
        {-1},
        {-1,-1},
        {33995,-1,-1}
    },
    {
        {33891,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33891,-1},
        {-1},
        {-1,-1},
        {33891,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33893,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33893,-1},
        {-1},
        {-1,-1},
        {33893,-1,-1}
    },
    {
        {33894,-1},
        {-1},
        {-1,-1},
        {33894,-1,-1}
    },
    {
        {33894,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33894,-1},
        {-1},
        {-1,-1},
        {33894,-1,-1}
    },
    {
        {33894,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33894,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33894,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33563,30031},
        {-1,-1,-1}
    },
    {
        {33897,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33897,-1},
        {-1},
        {-1,-1},
        {33897,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34186,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34186,-1,-1}
    },
    {
        {33899,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33899,-1},
        {-1},
        {-1,-1},
        {33899,-1,-1}
    },
    {
        {33900,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33900,-1},
        {-1},
        {-1,-1},
        {33900,-1,-1}
    },
    {
        {33901,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33901,-1},
        {-1},
        {-1,-1},
        {33901,-1,-1}
    },
    {
        {33902,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33902,-1},
        {-1},
        {-1,-1},
        {33902,-1,-1}
    },
    {
        {33903,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33903,-1},
        {-1},
        {-1,-1},
        {33903,-1,-1}
    },
    {
        {33904,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33904,-1},
        {-1},
        {-1,-1},
        {33904,-1,-1}
    },
    {
        {33905,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33905,-1},
        {-1},
        {-1,-1},
        {33905,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33737,27849},
        {-1,-1,-1}
    },
    {
        {33907,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33907,-1},
        {-1},
        {-1,-1},
        {33907,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33909,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33909,-1},
        {-1},
        {-1,-1},
        {33909,-1,-1}
    },
    {
        {33910,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33910,-1},
        {-1},
        {-1,-1},
        {33910,-1,-1}
    },
    {
        {33911,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33911,-1},
        {-1},
        {-1,-1},
        {33911,-1,-1}
    },
    {
        {33912,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33912,-1},
        {-1},
        {-1,-1},
        {33912,-1,-1}
    },
    {
        {33913,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33913,-1},
        {-1},
        {-1,-1},
        {33913,-1,-1}
    },
    {
        {33914,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33914,-1},
        {-1},
        {-1,-1},
        {33914,-1,-1}
    },
    {
        {33914,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23888,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23888,-1},
        {-1,-1,-1}
    },
    {
        {33916,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33916,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24616,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30473,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33921,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33921,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33921,-1},
        {-1},
        {-1,-1},
        {33921,-1,-1}
    },
    {
        {33921,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33922,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33922,-1},
        {-1},
        {-1,-1},
        {33922,34069,-1}
    },
    {
        {33922,-1},
        {-1},
        {-1,-1},
        {34069,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31686,-1},
        {31686,-1,-1}
    },
    {
        {33924,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33924,-1},
        {-1},
        {-1,-1},
        {33924,-1,-1}
    },
    {
        {33925,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33925,-1},
        {-1},
        {-1,-1},
        {33925,-1,-1}
    },
    {
        {33925,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33931,-1},
        {-1},
        {-1,-1},
        {34083,-1,-1}
    },
    {
        {33931,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33936,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33936,-1},
        {-1},
        {-1,-1},
        {33936,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27575,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34036,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34036,-1,-1}
    },
    {
        {33940,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33940,-1},
        {-1},
        {-1,-1},
        {33940,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34145,30495},
        {-1,-1,-1}
    },
    {
        {33943,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33943,-1},
        {-1},
        {-1,-1},
        {33943,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25360,-1},
        {-1,-1,-1}
    },
    {
        {33945,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33945,-1},
        {-1},
        {-1,-1},
        {33945,-1,-1}
    },
    {
        {33945,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33945,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39730,33845},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32570,-1},
        {-1,-1,-1}
    },
    {
        {33948,-1},
        {-1},
        {-1,-1},
        {33948,-1,-1}
    },
    {
        {33948,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33948,-1},
        {-1},
        {-1,-1},
        {33948,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21407,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33950,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33950,-1},
        {-1},
        {-1,-1},
        {33950,-1,-1}
    },
    {
        {33951,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33951,-1},
        {-1},
        {-1,-1},
        {33951,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24687,34191},
        {-1,-1,-1}
    },
    {
        {33953,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33953,-1},
        {-1},
        {-1,-1},
        {33953,-1,-1}
    },
    {
        {33953,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34295,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28034,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30041,-1},
        {-1,-1,-1}
    },
    {
        {33958,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33958,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33958,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23923,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33965,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33965,-1},
        {-1},
        {-1,-1},
        {33467,33965,-1}
    },
    {
        {-1,-1},
        {-1},
        {38586,-1},
        {-1,-1,-1}
    },
    {
        {33967,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33967,-1},
        {-1},
        {-1,-1},
        {33967,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33324,33902},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33970,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33970,-1},
        {-1},
        {-1,-1},
        {33970,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32013,-1},
        {-1,-1,-1}
    },
    {
        {33972,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33972,-1},
        {-1},
        {-1,-1},
        {33972,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33756,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32027,33903},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33464,34131},
        {34131,-1,-1}
    },
    {
        {33976,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33976,-1},
        {-1},
        {-1,-1},
        {33976,-1,-1}
    },
    {
        {33977,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33977,-1},
        {-1},
        {-1,-1},
        {33977,-1,-1}
    },
    {
        {33977,-1},
        {-1},
        {-1,-1},
        {33977,-1,-1}
    },
    {
        {33978,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33978,-1},
        {-1},
        {-1,-1},
        {33978,-1,-1}
    },
    {
        {33979,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33979,-1},
        {-1},
        {-1,-1},
        {33979,-1,-1}
    },
    {
        {33979,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33980,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33980,-1},
        {-1},
        {-1,-1},
        {33980,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33982,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33982,-1},
        {-1},
        {-1,-1},
        {33982,-1,-1}
    },
    {
        {33983,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33983,-1},
        {-1},
        {-1,-1},
        {33983,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33984,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33984,-1},
        {-1},
        {-1,-1},
        {33984,-1,-1}
    },
    {
        {33985,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33985,-1},
        {-1},
        {-1,-1},
        {33985,-1,-1}
    },
    {
        {33986,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33986,-1},
        {-1},
        {-1,-1},
        {33986,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34282,-1},
        {-1,-1,-1}
    },
    {
        {33988,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33988,-1},
        {-1},
        {-1,-1},
        {33988,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27969,34092},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27969,34092},
        {-1,-1,-1}
    },
    {
        {33990,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33990,-1},
        {-1},
        {-1,-1},
        {33990,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33993,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33993,-1},
        {-1},
        {-1,-1},
        {33993,-1,-1}
    },
    {
        {33994,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33994,-1},
        {-1},
        {-1,-1},
        {33994,-1,-1}
    },
    {
        {33994,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33994,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33995,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33995,-1},
        {-1},
        {-1,-1},
        {33995,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33997,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33997,-1},
        {-1},
        {-1,-1},
        {33997,-1,-1}
    },
    {
        {33998,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33998,-1},
        {-1},
        {-1,-1},
        {33998,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33999,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33999,-1},
        {-1},
        {-1,-1},
        {33999,-1,-1}
    },
    {
        {33999,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34000,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34000,-1},
        {-1},
        {-1,-1},
        {34000,-1,-1}
    },
    {
        {34000,-1},
        {-1},
        {-1,-1},
        {34000,-1,-1}
    },
    {
        {34001,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34001,-1},
        {-1},
        {-1,-1},
        {34001,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36562,-1},
        {-1,-1,-1}
    },
    {
        {34003,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34003,-1},
        {-1},
        {-1,-1},
        {34003,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33976,32660},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26690,-1},
        {-1,-1,-1}
    },
    {
        {34006,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34006,-1},
        {-1},
        {-1,-1},
        {34006,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34974,-1},
        {-1,-1,-1}
    },
    {
        {34009,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34009,-1},
        {-1},
        {-1,-1},
        {34009,-1,-1}
    },
    {
        {34009,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34010,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34010,-1},
        {-1},
        {-1,-1},
        {34010,-1,-1}
    },
    {
        {34012,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34012,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34012,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33807,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34023,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34023,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33065,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21209,-1},
        {-1,-1,-1}
    },
    {
        {34026,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34026,-1},
        {-1},
        {-1,-1},
        {34026,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34028,-1},
        {-1},
        {-1,-1},
        {34028,-1,-1}
    },
    {
        {34028,-1},
        {-1},
        {-1,-1},
        {34028,-1,-1}
    },
    {
        {34028,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34028,-1},
        {-1},
        {-1,-1},
        {34028,-1,-1}
    },
    {
        {34028,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24245,-1},
        {-1,-1,-1}
    },
    {
        {34030,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34030,-1},
        {-1},
        {-1,-1},
        {34030,-1,-1}
    },
    {
        {34030,-1},
        {-1},
        {-1,-1},
        {34030,-1,-1}
    },
    {
        {34030,-1},
        {-1},
        {-1,-1},
        {34030,-1,-1}
    },
    {
        {34030,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34031,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34031,-1},
        {-1},
        {-1,-1},
        {34031,-1,-1}
    },
    {
        {34032,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34032,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34033,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34033,-1},
        {-1},
        {-1,-1},
        {34033,33805,-1}
    },
    {
        {34033,-1},
        {-1},
        {-1,-1},
        {34033,33805,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33771,-1},
        {33771,-1,-1}
    },
    {
        {34036,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34036,-1},
        {-1},
        {-1,-1},
        {34036,-1,-1}
    },
    {
        {34036,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25463,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21807,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26361,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24481,-1},
        {-1,-1,-1}
    },
    {
        {34042,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34044,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34044,-1},
        {-1},
        {-1,-1},
        {34044,-1,-1}
    },
    {
        {34044,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34045,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34045,-1},
        {-1},
        {-1,-1},
        {34045,-1,-1}
    },
    {
        {34045,-1},
        {-1},
        {-1,-1},
        {34045,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40565,-1},
        {-1,-1,-1}
    },
    {
        {34047,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34047,-1},
        {-1},
        {-1,-1},
        {34047,-1,-1}
    },
    {
        {34048,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34048,-1},
        {-1},
        {-1,-1},
        {34048,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33609,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21839,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22925,-1},
        {23032,-1,-1}
    },
    {
        {34054,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34054,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31186,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34974,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28113,33789},
        {-1,-1,-1}
    },
    {
        {34060,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34060,-1},
        {-1},
        {-1,-1},
        {34060,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31895,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33659,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33659,-1,-1}
    },
    {
        {34065,-1},
        {-1},
        {-1,-1},
        {34065,-1,-1}
    },
    {
        {34065,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34065,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34065,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33911,-1},
        {33911,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33911,-1},
        {33911,-1,-1}
    },
    {
        {34067,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34067,-1},
        {-1},
        {-1,-1},
        {34067,-1,-1}
    },
    {
        {34067,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34067,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34068,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34068,-1},
        {-1},
        {-1,-1},
        {33748,34068,-1}
    },
    {
        {34069,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34069,-1},
        {-1},
        {-1,-1},
        {34069,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34071,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34071,-1},
        {-1},
        {-1,-1},
        {34071,-1,-1}
    },
    {
        {34071,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34072,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34072,-1},
        {-1},
        {-1,-1},
        {34072,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33913,-1},
        {-1,-1,-1}
    },
    {
        {34074,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34074,-1},
        {-1},
        {-1,-1},
        {34074,-1,-1}
    },
    {
        {34075,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34075,-1},
        {-1},
        {-1,-1},
        {34075,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33982,-1},
        {-1,-1,-1}
    },
    {
        {34078,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34078,-1},
        {-1},
        {-1,-1},
        {34078,-1,-1}
    },
    {
        {34079,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34079,-1},
        {-1},
        {-1,-1},
        {34079,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32066,33903},
        {-1,-1,-1}
    },
    {
        {34081,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34081,-1},
        {-1},
        {-1,-1},
        {34081,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34083,-1},
        {-1},
        {-1,-1},
        {34083,-1,-1}
    },
    {
        {34083,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34083,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34084,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34084,-1},
        {-1},
        {-1,-1},
        {34084,-1,-1}
    },
    {
        {34085,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34085,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34085,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34086,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34086,-1},
        {-1},
        {-1,-1},
        {34086,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33740,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23493,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34091,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34091,-1},
        {-1},
        {-1,-1},
        {34091,-1,-1}
    },
    {
        {34092,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34092,-1},
        {-1},
        {-1,-1},
        {34092,-1,-1}
    },
    {
        {34093,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34093,-1},
        {-1},
        {-1,-1},
        {34093,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34095,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34095,-1},
        {-1},
        {-1,-1},
        {34236,34095,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27578,-1},
        {-1,-1,-1}
    },
    {
        {34098,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34098,-1},
        {-1},
        {-1,-1},
        {34098,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34100,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34100,-1},
        {-1},
        {-1,-1},
        {40635,34100,-1}
    },
    {
        {34101,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31744,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31744,-1,-1}
    },
    {
        {34109,-1},
        {-1},
        {-1,-1},
        {34109,-1,-1}
    },
    {
        {34109,-1},
        {-1},
        {-1,-1},
        {34109,-1,-1}
    },
    {
        {34109,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34110,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34110,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34110,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34112,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34112,-1},
        {-1},
        {-1,-1},
        {34112,-1,-1}
    },
    {
        {34113,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34113,-1},
        {-1},
        {-1,-1},
        {34113,-1,-1}
    },
    {
        {34113,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34276,21213},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34115,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34115,-1},
        {-1},
        {-1,-1},
        {34115,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33824,33804},
        {33804,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34261,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31774,-1},
        {-1,-1,-1}
    },
    {
        {34120,-1},
        {-1},
        {-1,-1},
        {34120,-1,-1}
    },
    {
        {34120,-1},
        {-1},
        {-1,-1},
        {34120,-1,-1}
    },
    {
        {34120,-1},
        {-1},
        {-1,-1},
        {34120,-1,-1}
    },
    {
        {34121,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34121,-1},
        {-1},
        {-1,-1},
        {34121,-1,-1}
    },
    {
        {34122,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34122,-1},
        {-1},
        {-1,-1},
        {34122,-1,-1}
    },
    {
        {34123,-1},
        {-1},
        {-1,-1},
        {34306,34122,-1}
    },
    {
        {34123,-1},
        {-1},
        {-1,-1},
        {34122,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34271,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28189,-1},
        {-1,-1,-1}
    },
    {
        {34126,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34126,-1},
        {-1},
        {-1,-1},
        {34126,-1,-1}
    },
    {
        {34127,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34127,-1},
        {-1},
        {-1,-1},
        {34127,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33775,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33775,-1,-1}
    },
    {
        {34129,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34129,-1},
        {-1},
        {-1,-1},
        {34129,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34131,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34131,-1},
        {-1},
        {-1,-1},
        {34131,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22577,-1},
        {-1,-1,-1}
    },
    {
        {34133,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34133,-1},
        {-1},
        {-1,-1},
        {34133,-1,-1}
    },
    {
        {34133,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34133,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34135,-1},
        {-1},
        {-1,-1},
        {34135,-1,-1}
    },
    {
        {34135,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34136,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34136,-1},
        {-1},
        {-1,-1},
        {34136,-1,-1}
    },
    {
        {34137,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34137,-1},
        {-1},
        {-1,-1},
        {34137,-1,-1}
    },
    {
        {34137,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34137,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34138,-1},
        {-1},
        {-1,-1},
        {33852,-1,-1}
    },
    {
        {34138,-1},
        {-1},
        {-1,-1},
        {33852,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31305,31242},
        {31242,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34142,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34142,-1},
        {-1},
        {-1,-1},
        {34142,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34145,-1},
        {-1},
        {-1,-1},
        {34145,-1,-1}
    },
    {
        {34145,-1},
        {-1},
        {-1,-1},
        {34145,-1,-1}
    },
    {
        {34146,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34146,-1},
        {-1},
        {-1,-1},
        {34146,-1,-1}
    },
    {
        {34147,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34147,-1},
        {-1},
        {-1,-1},
        {34147,-1,-1}
    },
    {
        {34147,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34147,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34147,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34148,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34148,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34201,38597},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38920,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24489,-1},
        {-1,-1,-1}
    },
    {
        {34152,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34152,-1},
        {-1},
        {-1,-1},
        {34152,-1,-1}
    },
    {
        {34153,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34153,-1},
        {-1},
        {-1,-1},
        {34153,-1,-1}
    },
    {
        {34154,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34154,-1},
        {-1},
        {-1,-1},
        {34154,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33984,32114},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33984,32114},
        {-1,-1,-1}
    },
    {
        {34157,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34157,-1},
        {-1},
        {-1,-1},
        {34157,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20905,34086},
        {-1,-1,-1}
    },
    {
        {34159,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34159,-1},
        {-1},
        {-1,-1},
        {34159,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34176,-1},
        {-1},
        {34176,28201},
        {34176,34314,-1}
    },
    {
        {34176,-1},
        {-1},
        {34176,28201},
        {34176,34314,-1}
    },
    {
        {34129,-1},
        {-1},
        {33733,-1},
        {-1,-1,-1}
    },
    {
        {34129,-1},
        {-1},
        {33733,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38646,-1},
        {-1,-1,-1}
    },
    {
        {34167,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34167,-1},
        {-1},
        {-1,-1},
        {34167,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33901,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33901,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33901,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34170,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34170,-1},
        {-1},
        {-1,-1},
        {34170,-1,-1}
    },
    {
        {34171,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34171,-1},
        {-1},
        {-1,-1},
        {34171,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32902,-1},
        {-1,-1,-1}
    },
    {
        {34173,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34173,-1},
        {-1},
        {-1,-1},
        {34173,-1,-1}
    },
    {
        {34174,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34174,-1},
        {-1},
        {-1,-1},
        {34174,-1,-1}
    },
    {
        {34175,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34175,-1},
        {-1},
        {-1,-1},
        {33841,34175,-1}
    },
    {
        {34176,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34176,-1},
        {-1},
        {-1,-1},
        {34176,34314,-1}
    },
    {
        {34177,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34177,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34177,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33421,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33983,-1},
        {-1,-1,-1}
    },
    {
        {34180,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34180,-1},
        {-1},
        {-1,-1},
        {34180,-1,-1}
    },
    {
        {34180,-1},
        {-1},
        {-1,-1},
        {34180,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34182,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34182,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34183,-1},
        {-1},
        {-1,-1},
        {34183,-1,-1}
    },
    {
        {34183,-1},
        {-1},
        {-1,-1},
        {34183,-1,-1}
    },
    {
        {34183,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34183,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34183,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34184,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34184,-1},
        {-1},
        {-1,-1},
        {34184,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34186,-1},
        {-1},
        {-1,-1},
        {34186,-1,-1}
    },
    {
        {34186,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34186,-1},
        {-1},
        {-1,-1},
        {34186,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34188,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34188,-1},
        {-1},
        {-1,-1},
        {34188,-1,-1}
    },
    {
        {34188,-1},
        {-1},
        {-1,-1},
        {34188,-1,-1}
    },
    {
        {34188,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20098,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34065,-1},
        {-1,-1,-1}
    },
    {
        {34191,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34191,-1},
        {-1},
        {-1,-1},
        {34191,-1,-1}
    },
    {
        {34191,-1},
        {-1},
        {-1,-1},
        {34191,-1,-1}
    },
    {
        {34192,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34192,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34193,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34193,-1},
        {-1},
        {-1,-1},
        {34193,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34081,31922},
        {-1,-1,-1}
    },
    {
        {34195,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34195,-1},
        {-1},
        {-1,-1},
        {34195,-1,-1}
    },
    {
        {34196,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34196,-1},
        {-1},
        {-1,-1},
        {34196,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31806,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31806,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34273,36942},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34273,36942},
        {-1,-1,-1}
    },
    {
        {34199,-1},
        {-1},
        {-1,-1},
        {22290,-1,-1}
    },
    {
        {34199,-1},
        {-1},
        {-1,-1},
        {22290,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34201,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34201,-1},
        {-1},
        {-1,-1},
        {34201,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34153,27556},
        {-1,-1,-1}
    },
    {
        {34203,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34203,-1},
        {-1},
        {-1,-1},
        {34203,-1,-1}
    },
    {
        {34204,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34204,-1},
        {-1},
        {-1,-1},
        {34204,-1,-1}
    },
    {
        {34205,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34205,-1},
        {-1},
        {-1,-1},
        {34205,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39153,-1},
        {-1,-1,-1}
    },
    {
        {34207,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34207,-1},
        {-1},
        {-1,-1},
        {34207,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34210,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34210,-1},
        {-1},
        {-1,-1},
        {34210,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33874,-1,-1}
    },
    {
        {34212,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34212,-1},
        {-1},
        {-1,-1},
        {34212,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34214,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34214,-1},
        {-1},
        {-1,-1},
        {34214,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34216,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34216,-1},
        {-1},
        {-1,-1},
        {34216,-1,-1}
    },
    {
        {34217,-1},
        {-1},
        {-1,-1},
        {34217,-1,-1}
    },
    {
        {34217,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34217,-1},
        {-1},
        {-1,-1},
        {34217,-1,-1}
    },
    {
        {34217,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34218,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34218,-1},
        {-1},
        {-1,-1},
        {34218,-1,-1}
    },
    {
        {34219,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34219,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34219,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34220,34277,-1}
    },
    {
        {34220,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34221,-1},
        {-1},
        {-1,-1},
        {34221,-1,-1}
    },
    {
        {34221,-1},
        {-1},
        {-1,-1},
        {34221,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34222,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34223,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34223,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34223,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34224,-1},
        {-1},
        {-1,-1},
        {34219,-1,-1}
    },
    {
        {34219,-1},
        {-1},
        {-1,-1},
        {34224,34219,-1}
    },
    {
        {-1,-1},
        {-1},
        {33873,23565},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34315,-1},
        {34315,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36960,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36960,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22475,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22475,-1,-1}
    },
    {
        {34231,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34231,-1},
        {-1},
        {-1,-1},
        {34231,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34233,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34233,-1},
        {-1},
        {-1,-1},
        {34233,-1,-1}
    },
    {
        {34234,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34234,-1},
        {-1},
        {-1,-1},
        {34234,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34299,-1},
        {34299,-1,-1}
    },
    {
        {34236,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34236,-1},
        {-1},
        {-1,-1},
        {34236,34095,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29074,-1},
        {-1,-1,-1}
    },
    {
        {34241,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34241,-1},
        {-1},
        {-1,-1},
        {34241,31295,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21474,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21474,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27466,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32386,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32624,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25652,-1},
        {-1,-1,-1}
    },
    {
        {34247,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34247,-1},
        {-1},
        {-1,-1},
        {34247,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30589,-1},
        {-1,-1,-1}
    },
    {
        {34249,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34249,-1},
        {-1},
        {-1,-1},
        {34249,-1,-1}
    },
    {
        {34249,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34250,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34250,-1},
        {-1},
        {-1,-1},
        {34250,-1,-1}
    },
    {
        {34251,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34251,-1},
        {-1},
        {-1,-1},
        {34251,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34084,-1},
        {34084,-1,-1}
    },
    {
        {34253,-1},
        {-1},
        {-1,-1},
        {34253,-1,-1}
    },
    {
        {34253,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34253,-1},
        {-1},
        {-1,-1},
        {34253,-1,-1}
    },
    {
        {34254,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34254,-1},
        {-1},
        {-1,-1},
        {34254,-1,-1}
    },
    {
        {34255,-1},
        {-1},
        {34101,34255},
        {34101,-1,-1}
    },
    {
        {34255,-1},
        {-1},
        {34101,34255},
        {34101,34255,-1}
    },
    {
        {34255,-1},
        {-1},
        {34101,34255},
        {34101,-1,-1}
    },
    {
        {34255,-1},
        {-1},
        {34101,34255},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34256,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34256,-1},
        {-1},
        {-1,-1},
        {34256,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22800,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {64031,-1},
        {-1,-1,-1}
    },
    {
        {34261,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34261,-1},
        {-1},
        {-1,-1},
        {34261,-1,-1}
    },
    {
        {34261,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36066,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34264,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34264,-1},
        {-1},
        {-1,-1},
        {34264,-1,-1}
    },
    {
        {34265,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34265,-1},
        {-1},
        {-1,-1},
        {34265,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23531,-1},
        {-1,-1,-1}
    },
    {
        {34268,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34268,-1},
        {-1},
        {-1,-1},
        {34268,-1,-1}
    },
    {
        {34269,-1},
        {-1},
        {-1,-1},
        {33464,-1,-1}
    },
    {
        {34269,-1},
        {-1},
        {-1,-1},
        {33464,-1,-1}
    },
    {
        {34271,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34271,-1},
        {-1},
        {-1,-1},
        {34271,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34273,-1},
        {-1},
        {-1,-1},
        {34273,33659,-1}
    },
    {
        {34273,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34273,-1},
        {-1},
        {-1,-1},
        {34273,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34183,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32631,-1},
        {-1,-1,-1}
    },
    {
        {34276,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34276,-1},
        {-1},
        {-1,-1},
        {34276,-1,-1}
    },
    {
        {34276,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34276,-1},
        {-1},
        {-1,-1},
        {34276,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34277,-1},
        {-1},
        {-1,-1},
        {34220,-1,-1}
    },
    {
        {34277,-1},
        {-1},
        {-1,-1},
        {34220,-1,-1}
    },
    {
        {34277,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34278,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34278,-1},
        {-1},
        {-1,-1},
        {34278,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32233,34175},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34281,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34281,-1},
        {-1},
        {-1,-1},
        {34281,-1,-1}
    },
    {
        {34282,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34282,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34282,-1},
        {-1},
        {-1,-1},
        {34282,34222,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28525,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38564,-1},
        {-1,-1,-1}
    },
    {
        {34285,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34285,-1},
        {-1},
        {-1,-1},
        {34285,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27189,-1},
        {27189,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24944,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21129,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31295,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31295,-1,-1}
    },
    {
        {34314,-1},
        {-1},
        {34176,34314},
        {34314,-1,-1}
    },
    {
        {34294,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34294,-1},
        {-1},
        {-1,-1},
        {34294,-1,-1}
    },
    {
        {34295,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34295,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34295,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34127,-1},
        {-1,-1,-1}
    },
    {
        {34297,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34297,-1},
        {-1},
        {-1,-1},
        {34297,-1,-1}
    },
    {
        {34297,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34298,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34298,-1},
        {-1},
        {-1,-1},
        {34298,-1,-1}
    },
    {
        {34299,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34299,-1},
        {-1},
        {-1,-1},
        {34299,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33841,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33841,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33841,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27372,-1},
        {-1,-1,-1}
    },
    {
        {34302,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34302,-1},
        {-1},
        {-1,-1},
        {34302,-1,-1}
    },
    {
        {34302,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34303,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34303,-1},
        {-1},
        {-1,-1},
        {34303,-1,-1}
    },
    {
        {34304,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34304,-1},
        {-1},
        {-1,-1},
        {34304,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33852,-1},
        {-1,-1,-1}
    },
    {
        {34306,-1},
        {-1},
        {-1,-1},
        {34122,-1,-1}
    },
    {
        {34306,-1},
        {-1},
        {-1,-1},
        {34306,34122,-1}
    },
    {
        {-1,-1},
        {-1},
        {34306,34123},
        {34306,34122,-1}
    },
    {
        {34308,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34308,-1},
        {-1},
        {-1,-1},
        {34308,-1,-1}
    },
    {
        {34309,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34309,-1},
        {-1},
        {-1,-1},
        {34309,-1,-1}
    },
    {
        {34310,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34310,-1},
        {-1},
        {-1,-1},
        {33446,34310,-1}
    },
    {
        {34311,-1},
        {-1},
        {-1,-1},
        {34323,34311,-1}
    },
    {
        {34311,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34311,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34311,-1},
        {-1},
        {-1,-1},
        {34311,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34315,38973},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34314,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34314,-1},
        {-1},
        {-1,-1},
        {34314,-1,-1}
    },
    {
        {34315,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34315,-1},
        {-1},
        {-1,-1},
        {34315,-1,-1}
    },
    {
        {34315,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31838,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34219,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21234,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39525,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39525,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32069,-1},
        {32069,-1,-1}
    },
    {
        {34320,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34320,-1},
        {-1},
        {-1,-1},
        {34320,33841,-1}
    },
    {
        {34321,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34321,-1},
        {-1},
        {-1,-1},
        {34321,-1,-1}
    },
    {
        {34322,-1},
        {-1},
        {-1,-1},
        {33833,-1,-1}
    },
    {
        {34322,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34323,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34323,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34323,-1},
        {-1},
        {-1,-1},
        {34323,34311,-1}
    },
    {
        {-1,-1},
        {-1},
        {34028,-1},
        {34028,-1,-1}
    },
    {
        {34326,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34326,-1},
        {-1},
        {-1,-1},
        {27377,34326,-1}
    },
    {
        {34327,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34327,-1},
        {-1},
        {-1,-1},
        {34327,27287,-1}
    },
    {
        {34328,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34328,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34328,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34330,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34330,-1},
        {-1},
        {-1,-1},
        {34330,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33738,-1,-1}
    },
    {
        {34334,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34334,-1},
        {-1},
        {-1,-1},
        {34334,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38577,34093},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22715,-1},
        {-1,-1,-1}
    },
    {
        {34337,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34337,-1},
        {-1},
        {-1,-1},
        {34337,-1,-1}
    },
    {
        {34338,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34338,-1},
        {-1},
        {-1,-1},
        {34338,-1,-1}
    },
    {
        {34338,-1},
        {-1},
        {-1,-1},
        {34338,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34340,-1},
        {-1},
        {-1,-1},
        {33457,34340,-1}
    },
    {
        {34340,-1},
        {-1},
        {-1,-1},
        {33457,-1,-1}
    },
    {
        {34340,-1},
        {-1},
        {-1,-1},
        {33457,34340,-1}
    },
    {
        {-1,-1},
        {-1},
        {31845,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38646,-1},
        {-1,-1,-1}
    },
    {
        {34343,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34343,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34343,-1},
        {-1},
        {-1,-1},
        {34343,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32327,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34345,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34345,-1},
        {-1},
        {-1,-1},
        {34345,-1,-1}
    },
    {
        {34345,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22744,40587},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28651,34253},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24011,-1},
        {-1,-1,-1}
    },
    {
        {34349,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34349,-1},
        {-1},
        {-1,-1},
        {34349,-1,-1}
    },
    {
        {34349,-1},
        {-1},
        {-1,-1},
        {34349,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32637,-1},
        {-1,-1,-1}
    },
    {
        {34351,-1},
        {-1},
        {-1,-1},
        {34153,-1,-1}
    },
    {
        {34351,-1},
        {-1},
        {-1,-1},
        {34153,34351,-1}
    },
    {
        {34352,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34352,-1},
        {-1},
        {-1,-1},
        {34352,-1,-1}
    },
    {
        {34352,-1},
        {-1},
        {-1,-1},
        {34352,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {63952,34167},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22593,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33752,33873},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {160057,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34360,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34360,-1},
        {-1},
        {-1,-1},
        {34360,-1,-1}
    },
    {
        {34361,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34361,-1},
        {-1},
        {-1,-1},
        {34361,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34345,-1},
        {-1,-1,-1}
    },
    {
        {34364,-1},
        {-1},
        {-1,-1},
        {34364,-1,-1}
    },
    {
        {34364,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34271,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34271,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34361,-1},
        {-1,-1,-1}
    },
    {
        {34367,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34367,-1},
        {-1},
        {-1,-1},
        {34367,-1,-1}
    },
    {
        {34368,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34368,-1},
        {-1},
        {-1,-1},
        {34368,-1,-1}
    },
    {
        {34369,-1},
        {-1},
        {-1,-1},
        {34369,-1,-1}
    },
    {
        {34369,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34369,-1},
        {-1},
        {-1,-1},
        {34369,-1,-1}
    },
    {
        {34369,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31844,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25085,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34247,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27410,34277},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37313,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40712,-1},
        {-1,-1,-1}
    },
    {
        {34381,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34382,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34382,-1},
        {-1},
        {-1,-1},
        {34382,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34384,-1},
        {-1},
        {-1,-1},
        {34384,-1,-1}
    },
    {
        {34384,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24942,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34387,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34387,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34388,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34388,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34389,-1},
        {-1},
        {-1,-1},
        {20966,-1,-1}
    },
    {
        {34390,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34391,-1},
        {-1},
        {-1,-1},
        {34394,34395,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34394,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34395,-1},
        {-1},
        {-1,-1},
        {34394,-1,-1}
    },
    {
        {34396,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34382,-1,-1}
    },
    {
        {34398,-1},
        {-1},
        {-1,-1},
        {34398,-1,-1}
    },
    {
        {34398,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34399,-1},
        {-1},
        {-1,-1},
        {21495,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34402,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25135,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34382,-1},
        {-1,-1,-1}
    },
    {
        {34407,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34407,-1},
        {-1},
        {-1,-1},
        {34407,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24426,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20789,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20789,-1},
        {-1,-1,-1}
    },
    {
        {34411,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34412,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34497,-1},
        {34497,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34787,34811},
        {-1,-1,-1}
    },
    {
        {34415,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34417,-1},
        {-1},
        {-1,-1},
        {34664,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34481,-1},
        {-1,-1,-1}
    },
    {
        {34421,-1},
        {-1},
        {-1,-1},
        {34503,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34423,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34425,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34426,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34427,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34427,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34427,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38614,-1},
        {38614,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34444,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34475,26487},
        {-1,-1,-1}
    },
    {
        {34439,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34441,-1},
        {-1},
        {-1,-1},
        {34442,-1,-1}
    },
    {
        {34441,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34442,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34442,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34443,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34443,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34444,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34444,-1},
        {-1},
        {-1,-1},
        {34444,-1,-1}
    },
    {
        {34445,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34486,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34449,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24420,-1},
        {24420,-1,-1}
    },
    {
        {34451,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34483,-1},
        {-1,-1,-1}
    },
    {
        {34453,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34472,-1},
        {-1,-1,-1}
    },
    {
        {34456,-1},
        {-1},
        {-1,-1},
        {34516,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34505,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34460,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34460,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34461,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34523,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22707,24996},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34467,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34467,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34468,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34527,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34472,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34473,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34473,-1},
        {-1},
        {-1,-1},
        {34473,-1,-1}
    },
    {
        {34474,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34475,-1},
        {-1},
        {-1,-1},
        {39825,-1,-1}
    },
    {
        {34475,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34558,-1},
        {-1,-1,-1}
    },
    {
        {34479,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34480,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34481,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26544,34423},
        {-1,-1,-1}
    },
    {
        {34483,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34486,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34488,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34507,32046},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34541,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36315,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34497,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34456,36299},
        {-1,-1,-1}
    },
    {
        {34499,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34500,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34502,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34503,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34472,-1},
        {-1,-1,-1}
    },
    {
        {34505,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34865,-1},
        {34865,-1,-1}
    },
    {
        {34507,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34500,-1},
        {-1,-1,-1}
    },
    {
        {34509,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34510,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34480,-1},
        {-1,-1,-1}
    },
    {
        {34513,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34461,-1},
        {-1,-1,-1}
    },
    {
        {34516,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34517,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34444,-1},
        {-1,-1,-1}
    },
    {
        {34519,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30306,-1},
        {-1,-1,-1}
    },
    {
        {34521,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34523,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34571,-1},
        {-1,-1,-1}
    },
    {
        {34526,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34527,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34527,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26665,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34531,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34532,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34534,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26658,34526},
        {-1,-1,-1}
    },
    {
        {34537,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34537,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34537,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38675,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38675,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36330,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36330,-1},
        {-1,-1,-1}
    },
    {
        {34540,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34541,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34542,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34543,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34569,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27714,-1},
        {-1,-1,-1}
    },
    {
        {34552,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34552,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34553,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34554,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34555,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34555,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34556,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34557,-1},
        {-1},
        {-1,-1},
        {35997,-1,-1}
    },
    {
        {34558,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34532,-1},
        {-1,-1,-1}
    },
    {
        {34560,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29697,34578},
        {-1,-1,-1}
    },
    {
        {34562,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34563,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34563,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34563,-1,-1}
    },
    {
        {34565,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34566,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34567,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34568,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34568,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34569,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34569,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34570,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34571,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26768,-1},
        {-1,-1,-1}
    },
    {
        {34573,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34574,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34576,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34577,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34577,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34578,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34578,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34578,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34579,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34555,-1},
        {-1},
        {34555,-1},
        {-1,-1,-1}
    },
    {
        {34584,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34585,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34586,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34586,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36382,-1},
        {-1,-1,-1}
    },
    {
        {34588,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34591,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31259,-1},
        {-1,-1,-1}
    },
    {
        {34593,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34720,-1},
        {-1,-1,-1}
    },
    {
        {34595,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34597,-1},
        {-1,-1,-1}
    },
    {
        {34597,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34643,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26841,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34678,-1,-1}
    },
    {
        {34601,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34601,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31097,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26829,34647},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34486,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38519,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34444,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34586,-1},
        {-1,-1,-1}
    },
    {
        {34609,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26881,-1},
        {-1,-1,-1}
    },
    {
        {34612,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34574,-1},
        {34574,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34752,-1},
        {-1,-1,-1}
    },
    {
        {34615,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34615,-1},
        {-1},
        {-1,-1},
        {34615,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22533,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34618,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34619,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34619,-1},
        {-1},
        {-1,-1},
        {34619,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34622,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34623,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34624,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22538,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29256,34662},
        {-1,-1,-1}
    },
    {
        {34627,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39757,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34453,34870,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34444,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34821,34687},
        {34821,-1,-1}
    },
    {
        {34633,-1},
        {-1},
        {-1,-1},
        {34796,-1,-1}
    },
    {
        {34635,-1},
        {-1},
        {-1,-1},
        {34847,-1,-1}
    },
    {
        {34636,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34638,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34638,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23159,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29761,-1},
        {29761,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29761,-1},
        {29761,-1,-1}
    },
    {
        {34641,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32236,-1},
        {-1,-1,-1}
    },
    {
        {34643,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34643,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26999,-1},
        {-1,-1,-1}
    },
    {
        {34645,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34645,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34647,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34648,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34648,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34649,-1},
        {-1},
        {-1,-1},
        {34649,-1,-1}
    },
    {
        {34649,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29489,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26994,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34557,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26979,34574},
        {-1,-1,-1}
    },
    {
        {34655,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34656,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34837,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38972,-1},
        {-1,-1,-1}
    },
    {
        {34659,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34660,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34661,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34662,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29787,-1},
        {-1,-1,-1}
    },
    {
        {34664,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36405,-1},
        {-1,-1,-1}
    },
    {
        {34666,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32210,34543},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26934,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36406,-1},
        {-1,-1,-1}
    },
    {
        {34670,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29480,29503,-1}
    },
    {
        {-1,-1},
        {-1},
        {34521,-1},
        {-1,-1,-1}
    },
    {
        {34673,-1},
        {-1},
        {-1,-1},
        {34427,-1,-1}
    },
    {
        {34673,-1},
        {-1},
        {-1,-1},
        {34673,34427,-1}
    },
    {
        {34674,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29619,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29619,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29619,-1},
        {-1,-1,-1}
    },
    {
        {34676,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24833,34661},
        {-1,-1,-1}
    },
    {
        {34678,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34488,34597},
        {34488,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34488,34597},
        {34488,-1,-1}
    },
    {
        {34680,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21986,34612},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21986,34612},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36413,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34684,-1},
        {-1},
        {-1,-1},
        {34747,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34687,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34690,-1},
        {-1},
        {-1,-1},
        {34571,-1,-1}
    },
    {
        {34690,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34690,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34690,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34695,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34696,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34697,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34826,-1},
        {-1,-1,-1}
    },
    {
        {34699,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34701,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38548,34701},
        {34701,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21957,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27046,-1},
        {-1,-1,-1}
    },
    {
        {34707,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27070,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34873,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33021,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23475,34444},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23475,34444},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {160730,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34426,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34719,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34720,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34441,34442},
        {34442,-1,-1}
    },
    {
        {34722,-1},
        {-1},
        {-1,-1},
        {34509,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37758,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36428,-1},
        {-1,-1,-1}
    },
    {
        {34727,-1},
        {-1},
        {-1,-1},
        {34543,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36450,21830},
        {-1,-1,-1}
    },
    {
        {34731,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34732,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34733,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34735,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27162,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33187,-1},
        {-1,-1,-1}
    },
    {
        {34739,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34624,38515},
        {-1,-1,-1}
    },
    {
        {34741,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34753,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34609,34444},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34756,24958},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34774,-1},
        {-1,-1,-1}
    },
    {
        {34746,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34747,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27135,-1},
        {-1,-1,-1}
    },
    {
        {34749,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34749,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34750,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34752,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34752,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34753,-1},
        {-1},
        {-1,-1},
        {34442,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27134,-1},
        {-1,-1,-1}
    },
    {
        {34756,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34756,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {17898,-1},
        {-1,-1,-1}
    },
    {
        {34758,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34758,-1},
        {-1},
        {-1,-1},
        {34758,-1,-1}
    },
    {
        {34759,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34759,-1},
        {-1},
        {-1,-1},
        {34758,34759,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34661,-1,-1}
    },
    {
        {34763,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29825,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26792,34570},
        {-1,-1,-1}
    },
    {
        {34766,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34768,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34770,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34770,-1},
        {-1},
        {-1,-1},
        {34770,-1,-1}
    },
    {
        {34770,-1},
        {-1},
        {-1,-1},
        {34770,-1,-1}
    },
    {
        {34770,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27233,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {144952,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34774,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34873,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28192,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40009,-1},
        {-1,-1,-1}
    },
    {
        {34783,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34784,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34787,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34787,-1},
        {-1},
        {-1,-1},
        {34787,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34789,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34790,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34783,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36470,-1},
        {-1,-1,-1}
    },
    {
        {34794,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34795,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34795,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34796,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34797,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40020,-1},
        {-1,-1,-1}
    },
    {
        {34799,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34856,-1},
        {34856,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34661,-1,-1}
    },
    {
        {34802,-1},
        {-1},
        {-1,-1},
        {34411,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34774,-1},
        {-1,-1,-1}
    },
    {
        {34805,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34806,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34806,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34807,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34849,-1,-1}
    },
    {
        {34809,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34811,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27296,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34814,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34790,36039},
        {-1,-1,-1}
    },
    {
        {34817,-1},
        {-1},
        {-1,-1},
        {22190,-1,-1}
    },
    {
        {34817,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34678,-1},
        {34678,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34678,-1},
        {34678,-1,-1}
    },
    {
        {34819,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34819,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34733,-1},
        {-1,-1,-1}
    },
    {
        {34821,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34821,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34822,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34822,-1},
        {-1},
        {-1,-1},
        {34822,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34851,-1},
        {34851,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34851,-1},
        {34851,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36042,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34826,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34827,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27297,-1},
        {-1,-1,-1}
    },
    {
        {34829,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34829,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34829,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34830,-1},
        {-1},
        {-1,-1},
        {34770,-1,-1}
    },
    {
        {34830,-1},
        {-1},
        {-1,-1},
        {34770,34830,-1}
    },
    {
        {34830,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34830,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34831,-1},
        {-1},
        {-1,-1},
        {34809,-1,-1}
    },
    {
        {34832,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34833,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32365,-1,-1}
    },
    {
        {34835,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34835,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34836,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34837,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34838,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34838,-1},
        {-1},
        {-1,-1},
        {34838,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27330,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27345,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34847,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38754,-1},
        {-1,-1,-1}
    },
    {
        {34849,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34849,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34849,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34850,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34851,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34851,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27377,23421},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27368,-1},
        {-1,-1,-1}
    },
    {
        {34855,-1},
        {-1},
        {-1,-1},
        {34873,-1,-1}
    },
    {
        {34856,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34856,-1},
        {-1},
        {-1,-1},
        {34856,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39930,34573},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40251,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40251,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34849,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34849,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36504,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34562,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34862,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34873,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34865,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34866,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27387,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34870,-1},
        {-1},
        {-1,-1},
        {34453,34870,-1}
    },
    {
        {34870,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34870,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34876,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27402,-1},
        {-1,-1,-1}
    },
    {
        {34873,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34453,34870,-1}
    },
    {
        {34875,-1},
        {-1},
        {-1,-1},
        {34542,-1,-1}
    },
    {
        {34876,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34827,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34627,-1},
        {34627,-1,-1}
    },
    {
        {34880,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34882,-1},
        {-1},
        {-1,-1},
        {34884,-1,-1}
    },
    {
        {34882,-1},
        {-1},
        {-1,-1},
        {34884,-1,-1}
    },
    {
        {34882,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34884,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34886,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33032,33033},
        {33032,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21888,-1,-1}
    },
    {
        {34890,-1},
        {-1},
        {-1,-1},
        {34890,-1,-1}
    },
    {
        {34890,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34890,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34892,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34893,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34898,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34899,-1},
        {-1},
        {-1,-1},
        {34899,-1,-1}
    },
    {
        {34899,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24055,-1,-1}
    },
    {
        {34903,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34904,-1},
        {-1},
        {-1,-1},
        {37532,-1,-1}
    },
    {
        {34905,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34907,-1},
        {-1},
        {-1,-1},
        {34907,34910,-1}
    },
    {
        {34907,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34907,-1},
        {-1},
        {-1,-1},
        {34907,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36947,-1,-1}
    },
    {
        {34909,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34910,-1},
        {-1},
        {-1,-1},
        {34907,-1,-1}
    },
    {
        {34911,-1},
        {-1},
        {-1,-1},
        {36947,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34913,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34913,-1},
        {-1},
        {-1,-1},
        {34913,-1,-1}
    },
    {
        {34914,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34915,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34916,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {161300,26438},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {161300,-1},
        {-1,-1,-1}
    },
    {
        {34920,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34921,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34968,-1},
        {-1,-1,-1}
    },
    {
        {34923,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34974,-1},
        {-1},
        {24059,34974},
        {-1,-1,-1}
    },
    {
        {34974,-1},
        {-1},
        {24059,34974},
        {-1,-1,-1}
    },
    {
        {34927,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34928,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34930,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34930,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34927,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34946,26517},
        {-1,-1,-1}
    },
    {
        {34933,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31070,-1},
        {-1,-1,-1}
    },
    {
        {34935,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37034,26012},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34987,26525},
        {-1,-1,-1}
    },
    {
        {34941,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34942,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34943,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34945,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34946,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34069,-1},
        {34069,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22934,35158},
        {35158,-1,-1}
    },
    {
        {34952,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34968,25301},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38936,-1},
        {-1,-1,-1}
    },
    {
        {34955,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34957,-1},
        {-1,-1,-1}
    },
    {
        {34957,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34957,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24188,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20304,31056},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25307,-1},
        {-1,-1,-1}
    },
    {
        {34962,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31062,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26607,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26620,-1},
        {-1,-1,-1}
    },
    {
        {34966,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34967,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34968,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34969,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20184,38468},
        {-1,-1,-1}
    },
    {
        {34974,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34974,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34976,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34978,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34978,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32079,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32079,-1},
        {-1,-1,-1}
    },
    {
        {34980,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {64079,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26606,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26552,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25288,-1},
        {-1,-1,-1}
    },
    {
        {34986,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34987,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32946,-1},
        {32946,-1,-1}
    },
    {
        {34990,-1},
        {-1},
        {-1,-1},
        {31152,31074,-1}
    },
    {
        {34992,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34993,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34980,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26648,-1},
        {-1,-1,-1}
    },
    {
        {34996,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34997,-1},
        {-1},
        {-1,-1},
        {34941,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34999,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26667,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34969,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26669,-1},
        {-1,-1,-1}
    },
    {
        {35004,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35005,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35006,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35007,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35008,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35009,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35010,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35011,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35012,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35013,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35018,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35022,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35023,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25424,-1},
        {-1,-1,-1}
    },
    {
        {35025,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35026,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35026,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35027,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35028,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35029,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35029,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34967,-1},
        {34967,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26803,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26803,-1},
        {-1,-1,-1}
    },
    {
        {35032,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35033,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35013,-1},
        {-1,-1,-1}
    },
    {
        {35035,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35036,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35037,-1},
        {-1},
        {-1,-1},
        {35013,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35039,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35033,-1},
        {35033,-1,-1}
    },
    {
        {35041,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35048,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35084,-1},
        {35084,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31097,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35115,-1},
        {35115,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26905,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35056,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35057,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35058,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35059,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35060,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35060,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35061,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35060,35061},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35064,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35065,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25513,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25513,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35068,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35069,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35070,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26885,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31098,-1},
        {-1,-1,-1}
    },
    {
        {35073,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35074,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35097,-1},
        {-1,-1,-1}
    },
    {
        {35076,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31128,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35079,-1},
        {-1},
        {-1,-1},
        {35079,-1,-1}
    },
    {
        {35079,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31278,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31114,-1},
        {-1,-1,-1}
    },
    {
        {35082,-1},
        {-1},
        {-1,-1},
        {35082,-1,-1}
    },
    {
        {35082,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26986,-1},
        {-1,-1,-1}
    },
    {
        {35084,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35088,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35090,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35090,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35091,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31119,-1},
        {64027,31119,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26939,-1},
        {-1,-1,-1}
    },
    {
        {35096,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35096,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35097,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35098,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35128,32311},
        {35128,-1,-1}
    },
    {
        {35100,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35100,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35101,-1},
        {-1},
        {-1,-1},
        {35148,-1,-1}
    },
    {
        {35102,-1},
        {-1},
        {-1,-1},
        {35102,-1,-1}
    },
    {
        {35102,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27067,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25645,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25040,25079},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27029,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29503,-1},
        {-1,-1,-1}
    },
    {
        {35109,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35111,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35111,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35114,-1},
        {-1},
        {-1,-1},
        {35114,-1,-1}
    },
    {
        {35114,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35115,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35115,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35056,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35018,-1},
        {35018,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32264,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35122,24109},
        {-1,-1,-1}
    },
    {
        {35120,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35122,-1},
        {-1},
        {-1,-1},
        {34996,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32314,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32314,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35126,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35126,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35127,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35128,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35129,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35131,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35131,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20682,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20682,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27133,-1},
        {-1,-1,-1}
    },
    {
        {35136,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35137,-1},
        {-1},
        {-1,-1},
        {32358,-1,-1}
    },
    {
        {35137,-1},
        {-1},
        {-1,-1},
        {35137,32358,-1}
    },
    {
        {35138,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35139,-1},
        {-1},
        {-1,-1},
        {35090,-1,-1}
    },
    {
        {35139,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35140,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35141,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24158,-1,-1}
    },
    {
        {35145,-1},
        {-1},
        {35145,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25776,-1},
        {-1,-1,-1}
    },
    {
        {35145,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27190,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27190,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26840,-1},
        {-1,-1,-1}
    },
    {
        {35148,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35149,-1},
        {-1},
        {-1,-1},
        {38609,38620,-1}
    },
    {
        {-1,-1},
        {-1},
        {27214,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27233,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27187,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27208,-1},
        {-1,-1,-1}
    },
    {
        {35158,-1},
        {-1},
        {-1,-1},
        {35158,-1,-1}
    },
    {
        {35158,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27292,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27262,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35166,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35167,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35168,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35170,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35172,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24158,-1,-1}
    },
    {
        {35174,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35178,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35178,-1},
        {-1},
        {-1,-1},
        {35178,-1,-1}
    },
    {
        {35178,-1},
        {-1},
        {-1,-1},
        {35178,-1,-1}
    },
    {
        {35178,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25850,-1},
        {-1,-1,-1}
    },
    {
        {35181,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35183,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35183,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27379,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27379,-1},
        {-1,-1,-1}
    },
    {
        {35186,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35186,-1},
        {-1},
        {-1,-1},
        {35186,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35188,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35188,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25140,-1},
        {-1,-1,-1}
    },
    {
        {35191,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27409,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27408,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35188,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35188,-1,-1}
    },
    {
        {35198,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35199,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35199,-1},
        {-1},
        {-1,-1},
        {35199,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35200,-1},
        {-1},
        {-1,-1},
        {35198,-1,-1}
    },
    {
        {35201,-1},
        {-1},
        {-1,-1},
        {35201,-1,-1}
    },
    {
        {35201,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35201,-1},
        {-1,-1,-1}
    },
    {
        {35203,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35203,-1},
        {-1},
        {-1,-1},
        {35203,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35201,-1},
        {-1,-1,-1}
    },
    {
        {35206,-1},
        {-1},
        {-1,-1},
        {35206,-1,-1}
    },
    {
        {35206,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35206,-1},
        {-1},
        {-1,-1},
        {35206,-1,-1}
    },
    {
        {35207,-1},
        {-1},
        {-1,-1},
        {35207,-1,-1}
    },
    {
        {35207,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35208,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35208,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35209,-1},
        {-1},
        {-1,-1},
        {32647,-1,-1}
    },
    {
        {35209,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35210,-1},
        {-1},
        {-1,-1},
        {32648,-1,-1}
    },
    {
        {35210,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35211,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35264,35251},
        {35264,35251,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35216,20180},
        {-1,-1,-1}
    },
    {
        {35215,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35216,-1},
        {-1},
        {-1,-1},
        {35258,35226,-1}
    },
    {
        {35216,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35222,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28032,30474},
        {-1,-1,-1}
    },
    {
        {35219,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35220,-1},
        {-1},
        {-1,-1},
        {35219,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35222,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35223,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35224,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35386,-1},
        {-1,-1,-1}
    },
    {
        {35226,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {16531,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30541,-1},
        {-1,-1,-1}
    },
    {
        {35231,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35233,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39759,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35255,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35238,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35238,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35239,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35244,38990},
        {-1,-1,-1}
    },
    {
        {35241,-1},
        {-1},
        {-1,-1},
        {30585,-1,-1}
    },
    {
        {35241,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35242,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26049,35242},
        {-1,-1,-1}
    },
    {
        {35244,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35244,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35226,-1},
        {-1,-1,-1}
    },
    {
        {35247,-1},
        {-1},
        {-1,-1},
        {35247,-1,-1}
    },
    {
        {35247,-1},
        {-1},
        {-1,-1},
        {35247,-1,-1}
    },
    {
        {35248,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26283,35261},
        {-1,-1,-1}
    },
    {
        {35250,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35250,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35251,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35255,-1},
        {-1},
        {-1,-1},
        {35248,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24494,-1},
        {-1,-1,-1}
    },
    {
        {35258,-1},
        {-1},
        {-1,-1},
        {35226,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35248,-1},
        {35248,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35261,-1},
        {-1},
        {-1,-1},
        {35239,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23529,-1},
        {-1,-1,-1}
    },
    {
        {35263,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35264,-1},
        {-1},
        {-1,-1},
        {35251,-1,-1}
    },
    {
        {35264,-1},
        {-1},
        {-1,-1},
        {35251,-1,-1}
    },
    {
        {35264,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35282,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35282,-1},
        {-1},
        {-1,-1},
        {35282,-1,-1}
    },
    {
        {35282,-1},
        {-1},
        {-1,-1},
        {35282,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35284,-1},
        {-1},
        {-1,-1},
        {31563,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35286,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35318,-1,-1}
    },
    {
        {35288,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20235,-1},
        {-1,-1,-1}
    },
    {
        {35290,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35290,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32966,35302},
        {-1,-1,-1}
    },
    {
        {35292,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35293,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39853,35303},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35292,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33011,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25352,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35299,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33030,-1},
        {-1,-1,-1}
    },
    {
        {35301,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35302,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35303,-1},
        {-1},
        {-1,-1},
        {35299,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27742,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35303,-1},
        {-1,-1,-1}
    },
    {
        {35307,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39906,35301},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35282,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {17431,-1},
        {-1,-1,-1}
    },
    {
        {35313,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35299,-1},
        {-1,-1,-1}
    },
    {
        {35315,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35316,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35316,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35301,-1},
        {35301,-1,-1}
    },
    {
        {35318,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30992,-1},
        {-1,-1,-1}
    },
    {
        {35320,-1},
        {-1},
        {-1,-1},
        {35302,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35327,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30097,35292},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31019,-1},
        {-1,-1,-1}
    },
    {
        {35325,-1},
        {-1},
        {-1,-1},
        {35327,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29141,-1},
        {29141,-1,-1}
    },
    {
        {35327,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35328,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35330,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35331,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21483,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35336,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35338,-1},
        {-1,-1,-1}
    },
    {
        {35338,-1},
        {-1},
        {-1,-1},
        {35338,-1,-1}
    },
    {
        {35338,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35340,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35342,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35344,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35345,-1},
        {-1},
        {-1,-1},
        {35409,-1,-1}
    },
    {
        {35346,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35346,-1},
        {-1},
        {-1,-1},
        {35346,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35347,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35348,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35349,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35350,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35351,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35352,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35352,-1},
        {-1},
        {-1,-1},
        {35352,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35338,-1},
        {-1,-1,-1}
    },
    {
        {35355,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35355,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35397,-1},
        {-1,-1,-1}
    },
    {
        {35357,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35357,-1},
        {-1},
        {-1,-1},
        {35357,-1,-1}
    },
    {
        {35357,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35357,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35359,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35359,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30695,-1,-1}
    },
    {
        {35361,-1},
        {-1},
        {-1,-1},
        {21535,-1,-1}
    },
    {
        {35362,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35363,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35365,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35365,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37109,29494},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35388,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35370,-1},
        {-1,-1,-1}
    },
    {
        {35370,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20449,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35373,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35344,30740},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35351,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35377,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35383,-1},
        {-1,-1,-1}
    },
    {
        {35379,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35380,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35381,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35382,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35383,-1},
        {-1},
        {-1,-1},
        {20280,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21644,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24629,37477},
        {-1,-1,-1}
    },
    {
        {35386,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35387,-1},
        {-1},
        {-1,-1},
        {35387,-1,-1}
    },
    {
        {35387,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35387,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35388,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35436,-1,-1}
    },
    {
        {35390,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35390,-1},
        {35390,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35393,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35342,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35445,-1},
        {-1,-1,-1}
    },
    {
        {35397,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35398,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35399,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35400,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35413,32982},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21610,-1},
        {21610,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37463,-1},
        {-1,-1,-1}
    },
    {
        {35405,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35406,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35406,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35408,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35409,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35410,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35412,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35413,-1},
        {-1},
        {-1,-1},
        {35413,-1,-1}
    },
    {
        {35413,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35416,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35419,-1},
        {-1,-1,-1}
    },
    {
        {35419,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35421,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35422,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35424,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35424,-1},
        {-1},
        {-1,-1},
        {35424,-1,-1}
    },
    {
        {35425,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35426,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35427,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35594,-1},
        {35594,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35594,-1},
        {35594,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35594,-1},
        {35594,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35430,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35433,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24456,-1},
        {-1,-1,-1}
    },
    {
        {35435,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35436,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35437,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35437,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35438,-1},
        {-1},
        {-1,-1},
        {35438,-1,-1}
    },
    {
        {35438,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35482,-1},
        {-1,-1,-1}
    },
    {
        {35440,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35441,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35442,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35442,-1},
        {-1},
        {-1,-1},
        {35442,-1,-1}
    },
    {
        {35443,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35482,36036},
        {-1,-1,-1}
    },
    {
        {35445,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35598,-1},
        {35598,-1,-1}
    },
    {
        {35449,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35452,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35452,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35344,30799},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30801,-1},
        {-1,-1,-1}
    },
    {
        {35458,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35460,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35460,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35461,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35463,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35491,-1},
        {35491,-1,-1}
    },
    {
        {35465,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35468,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35469,-1},
        {-1},
        {-1,-1},
        {35469,-1,-1}
    },
    {
        {35469,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35563,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35472,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35473,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35475,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35477,-1},
        {-1,-1,-1}
    },
    {
        {35477,-1},
        {-1},
        {-1,-1},
        {35477,-1,-1}
    },
    {
        {35477,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35477,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35480,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35482,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35482,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39416,35373},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21766,-1},
        {21766,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35543,-1},
        {-1,-1,-1}
    },
    {
        {35486,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35488,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35488,-1},
        {-1},
        {-1,-1},
        {35488,-1,-1}
    },
    {
        {35489,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35500,-1},
        {-1,-1,-1}
    },
    {
        {35491,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35492,-1},
        {-1},
        {-1,-1},
        {35492,-1,-1}
    },
    {
        {35492,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35492,-1},
        {-1},
        {-1,-1},
        {35492,-1,-1}
    },
    {
        {35493,-1},
        {-1},
        {-1,-1},
        {35440,-1,-1}
    },
    {
        {35493,-1},
        {-1},
        {-1,-1},
        {35440,35493,-1}
    },
    {
        {35493,-1},
        {-1},
        {-1,-1},
        {35440,-1,-1}
    },
    {
        {35494,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35495,-1},
        {-1},
        {-1,-1},
        {36628,-1,-1}
    },
    {
        {35496,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35496,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35500,-1},
        {-1},
        {-1,-1},
        {35500,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38663,-1},
        {-1,-1,-1}
    },
    {
        {35500,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35501,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35502,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35502,-1},
        {-1},
        {-1,-1},
        {35502,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21809,-1,-1}
    },
    {
        {35504,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35551,-1},
        {-1,-1,-1}
    },
    {
        {35506,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24752,-1},
        {24752,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35511,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35511,-1},
        {-1},
        {-1,-1},
        {35511,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24358,-1},
        {-1,-1,-1}
    },
    {
        {35513,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35513,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35460,20358},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35516,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35547,-1},
        {-1,-1,-1}
    },
    {
        {35518,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35518,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35519,-1},
        {-1},
        {-1,-1},
        {35519,-1,-1}
    },
    {
        {35519,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21852,-1},
        {-1,-1,-1}
    },
    {
        {35522,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35524,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24524,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35527,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38827,-1,-1}
    },
    {
        {35531,-1},
        {-1},
        {-1,-1},
        {35531,-1,-1}
    },
    {
        {35531,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35532,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35533,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35533,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35535,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35536,-1},
        {-1},
        {-1,-1},
        {24838,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35538,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35542,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35543,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35057,35442},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35441,-1},
        {-1,-1,-1}
    },
    {
        {35546,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35547,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35547,-1},
        {-1},
        {-1,-1},
        {35547,-1,-1}
    },
    {
        {35547,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35548,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35550,-1},
        {-1},
        {-1,-1},
        {35550,-1,-1}
    },
    {
        {35550,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35551,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35552,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35553,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35553,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35554,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21902,-1},
        {-1,-1,-1}
    },
    {
        {35556,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35556,-1},
        {-1},
        {-1,-1},
        {35556,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36405,-1},
        {-1,-1,-1}
    },
    {
        {35558,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35558,-1},
        {-1},
        {-1,-1},
        {35558,-1,-1}
    },
    {
        {35559,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36667,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35676,-1},
        {35676,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35563,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21843,-1,-1}
    },
    {
        {35565,-1},
        {-1},
        {-1,-1},
        {35565,-1,-1}
    },
    {
        {35565,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35565,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35566,-1},
        {-1},
        {-1,-1},
        {35566,-1,-1}
    },
    {
        {35566,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36409,35550},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35569,-1},
        {-1},
        {-1,-1},
        {35569,-1,-1}
    },
    {
        {35569,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35569,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21693,22181},
        {21693,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35571,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35571,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35572,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35574,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35574,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35575,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35576,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35585,36060},
        {-1,-1,-1}
    },
    {
        {35578,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35578,-1},
        {-1},
        {-1,-1},
        {35578,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37728,35641},
        {-1,-1,-1}
    },
    {
        {35580,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35580,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38761,-1},
        {-1,-1,-1}
    },
    {
        {35582,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35582,-1},
        {-1},
        {-1,-1},
        {35582,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35584,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35585,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35586,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35588,-1},
        {-1},
        {-1,-1},
        {35588,-1,-1}
    },
    {
        {35588,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35588,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35589,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35591,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26292,35709},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35437,24871},
        {-1,-1,-1}
    },
    {
        {35594,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35594,-1},
        {-1},
        {-1,-1},
        {35594,-1,-1}
    },
    {
        {35594,-1},
        {-1},
        {-1,-1},
        {35594,-1,-1}
    },
    {
        {35594,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35596,-1},
        {-1},
        {-1,-1},
        {27468,-1,-1}
    },
    {
        {35597,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35598,-1},
        {-1},
        {-1,-1},
        {35598,-1,-1}
    },
    {
        {35598,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21471,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35600,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36428,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21972,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35604,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35604,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21884,-1},
        {-1,-1,-1}
    },
    {
        {35606,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35607,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35607,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35674,-1},
        {-1,-1,-1}
    },
    {
        {35609,-1},
        {-1},
        {-1,-1},
        {35609,-1,-1}
    },
    {
        {35609,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35610,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35610,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35610,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35611,-1},
        {-1},
        {-1,-1},
        {35611,-1,-1}
    },
    {
        {35611,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35612,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35613,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35613,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35615,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35616,-1},
        {-1},
        {-1,-1},
        {35617,-1,-1}
    },
    {
        {35617,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35651,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35674,30950},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32317,-1},
        {-1,-1,-1}
    },
    {
        {35622,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35627,29827},
        {-1,-1,-1}
    },
    {
        {35624,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35624,-1},
        {-1},
        {-1,-1},
        {35624,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35624,-1},
        {35624,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21830,-1},
        {21830,-1,-1}
    },
    {
        {35627,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35627,-1},
        {-1},
        {-1,-1},
        {35627,-1,-1}
    },
    {
        {35628,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35628,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35629,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36012,22038},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35419,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35679,-1},
        {-1,-1,-1}
    },
    {
        {35635,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35635,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35554,-1},
        {35554,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35559,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24246,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35641,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {64099,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36100,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35692,-1},
        {-1,-1,-1}
    },
    {
        {35646,-1},
        {-1},
        {-1,-1},
        {35646,-1,-1}
    },
    {
        {35646,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22066,-1},
        {22066,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21427,22196},
        {-1,-1,-1}
    },
    {
        {35649,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35649,-1},
        {-1},
        {-1,-1},
        {35649,-1,-1}
    },
    {
        {35649,-1},
        {-1},
        {-1,-1},
        {35649,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35651,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35654,-1},
        {-1},
        {-1,-1},
        {22075,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {64034,22882},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35657,-1},
        {-1},
        {-1,-1},
        {35388,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35733,-1},
        {35733,-1,-1}
    },
    {
        {35660,-1},
        {-1},
        {-1,-1},
        {35355,-1,-1}
    },
    {
        {35660,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35660,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35662,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35663,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35663,-1},
        {-1},
        {-1,-1},
        {35663,-1,-1}
    },
    {
        {35663,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22082,-1},
        {22082,-1,-1}
    },
    {
        {35665,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35668,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35624,-1},
        {-1,-1,-1}
    },
    {
        {35670,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35670,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31572,-1},
        {31572,-1,-1}
    },
    {
        {35672,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35672,-1},
        {-1},
        {-1,-1},
        {35672,-1,-1}
    },
    {
        {35673,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35674,-1},
        {-1},
        {-1,-1},
        {35674,-1,-1}
    },
    {
        {35674,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35675,-1},
        {-1},
        {-1,-1},
        {35670,-1,-1}
    },
    {
        {35676,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35676,-1},
        {-1},
        {-1,-1},
        {35676,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32361,-1},
        {-1,-1,-1}
    },
    {
        {35678,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35679,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32373,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35740,-1},
        {35740,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35740,-1},
        {35740,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25963,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21483,-1},
        {-1,-1,-1}
    },
    {
        {35686,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35686,-1},
        {-1},
        {-1,-1},
        {35686,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28611,-1},
        {-1,-1,-1}
    },
    {
        {35689,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35689,-1},
        {-1},
        {-1,-1},
        {22123,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34297,-1},
        {34297,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34297,-1},
        {34297,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34297,-1},
        {-1,-1,-1}
    },
    {
        {35691,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35692,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27448,35441},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35695,-1},
        {-1},
        {-1,-1},
        {35379,-1,-1}
    },
    {
        {35696,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35697,-1},
        {-1},
        {-1,-1},
        {21892,-1,-1}
    },
    {
        {35697,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35698,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32379,-1},
        {-1,-1,-1}
    },
    {
        {35700,-1},
        {-1},
        {-1,-1},
        {35700,-1,-1}
    },
    {
        {35700,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25055,-1},
        {23565,23550,-1}
    },
    {
        {35702,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35703,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35703,-1},
        {-1},
        {-1,-1},
        {35703,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25836,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22198,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30435,-1,-1}
    },
    {
        {35709,-1},
        {-1},
        {-1,-1},
        {35465,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35711,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35712,-1},
        {-1},
        {-1,-1},
        {35501,-1,-1}
    },
    {
        {35713,-1},
        {-1},
        {-1,-1},
        {35713,-1,-1}
    },
    {
        {35713,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35713,-1},
        {-1},
        {-1,-1},
        {35627,-1,-1}
    },
    {
        {35715,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35522,-1},
        {35522,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36707,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35700,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35700,-1},
        {-1,-1,-1}
    },
    {
        {35722,-1},
        {-1},
        {-1,-1},
        {22793,35722,-1}
    },
    {
        {35722,-1},
        {-1},
        {-1,-1},
        {22793,-1,-1}
    },
    {
        {35723,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35723,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35724,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35724,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35556,-1,-1}
    },
    {
        {35726,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35728,-1},
        {-1},
        {-1,-1},
        {35726,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40864,35542},
        {-1,-1,-1}
    },
    {
        {35730,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35731,-1},
        {-1},
        {-1,-1},
        {35698,-1,-1}
    },
    {
        {35732,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35733,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35734,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36513,-1},
        {-1,-1,-1}
    },
    {
        {35737,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35737,-1},
        {-1},
        {-1,-1},
        {35737,-1,-1}
    },
    {
        {35738,-1},
        {-1},
        {-1,-1},
        {35715,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22216,-1},
        {22216,-1,-1}
    },
    {
        {35740,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35741,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35742,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35895,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35897,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35898,-1},
        {-1},
        {-1,-1},
        {35898,-1,-1}
    },
    {
        {35898,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35898,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35898,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35338,162739},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27946,-1},
        {-1,-1,-1}
    },
    {
        {35901,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35903,-1},
        {-1},
        {-1,-1},
        {28179,-1,-1}
    },
    {
        {35903,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35903,-1},
        {28179,35903,-1}
    },
    {
        {35905,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35905,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35905,-1},
        {-1},
        {-1,-1},
        {35905,-1,-1}
    },
    {
        {35905,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22089,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35712,-1},
        {-1,-1,-1}
    },
    {
        {35909,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35909,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35910,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35911,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35912,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35913,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35914,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35916,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34250,-1},
        {-1,-1,-1}
    },
    {
        {35918,-1},
        {-1},
        {-1,-1},
        {31466,-1,-1}
    },
    {
        {35919,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35920,-1},
        {-1},
        {-1,-1},
        {35914,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35924,33398,33399}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35924,-1},
        {-1},
        {-1,-1},
        {33398,33399,-1}
    },
    {
        {35925,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35925,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35925,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35925,-1},
        {-1},
        {35925,-1},
        {-1,-1,-1}
    },
    {
        {35927,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35927,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35930,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35930,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35930,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21742,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29401,35978},
        {-1,-1,-1}
    },
    {
        {35937,-1},
        {-1},
        {-1,-1},
        {35937,-1,-1}
    },
    {
        {35937,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35937,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35938,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35938,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23401,35945},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35948,35495},
        {-1,-1,-1}
    },
    {
        {35944,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35945,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35946,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35946,-1},
        {-1},
        {-1,-1},
        {35946,-1,-1}
    },
    {
        {35946,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35947,-1},
        {-1},
        {-1,-1},
        {20104,-1,-1}
    },
    {
        {35947,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35948,-1},
        {-1},
        {-1,-1},
        {29482,-1,-1}
    },
    {
        {35948,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35949,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27589,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35102,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35612,-1},
        {-1,-1,-1}
    },
    {
        {35955,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36450,-1},
        {36450,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27562,-1},
        {-1,-1,-1}
    },
    {
        {35960,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35961,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35961,-1},
        {-1},
        {-1,-1},
        {35961,-1,-1}
    },
    {
        {35962,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35964,-1},
        {-1},
        {-1,-1},
        {35988,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35961,34930},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29399,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35416,35964},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35970,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35978,-1},
        {35978,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32900,35981},
        {-1,-1,-1}
    },
    {
        {35973,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35977,-1},
        {-1,-1,-1}
    },
    {
        {35977,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35978,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35980,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35981,-1},
        {-1},
        {-1,-1},
        {29432,-1,-1}
    },
    {
        {35982,-1},
        {-1},
        {-1,-1},
        {29450,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35048,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35949,-1},
        {-1,-1,-1}
    },
    {
        {35986,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35987,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35987,-1},
        {-1},
        {-1,-1},
        {29483,35987,-1}
    },
    {
        {35988,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22874,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35977,35610},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35992,32812},
        {-1,-1,-1}
    },
    {
        {35992,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35992,-1},
        {-1},
        {-1,-1},
        {35992,-1,-1}
    },
    {
        {35993,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35995,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35995,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35997,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35998,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21729,-1},
        {21729,-1,-1}
    },
    {
        {36000,-1},
        {-1},
        {-1,-1},
        {36000,-1,-1}
    },
    {
        {36000,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36000,-1},
        {-1},
        {-1,-1},
        {36000,-1,-1}
    },
    {
        {36001,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36002,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36024,24530},
        {-1,-1,-1}
    },
    {
        {36004,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33322,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29609,-1},
        {29609,-1,-1}
    },
    {
        {36007,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36007,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36008,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36008,-1},
        {-1},
        {-1,-1},
        {36008,-1,-1}
    },
    {
        {36009,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36010,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36011,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36011,-1},
        {-1},
        {-1,-1},
        {36011,-1,-1}
    },
    {
        {36011,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36012,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36013,-1},
        {-1},
        {-1,-1},
        {36074,-1,-1}
    },
    {
        {36014,-1},
        {-1},
        {-1,-1},
        {24336,36019,-1}
    },
    {
        {36015,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36016,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36009,36315},
        {-1,-1,-1}
    },
    {
        {36018,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36019,-1},
        {-1},
        {-1,-1},
        {24336,-1,-1}
    },
    {
        {36020,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30042,36059},
        {-1,-1,-1}
    },
    {
        {36022,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36023,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36024,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36027,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36028,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36029,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32989,36070},
        {-1,-1,-1}
    },
    {
        {36031,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36032,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36033,-1},
        {-1},
        {-1,-1},
        {36033,-1,-1}
    },
    {
        {36033,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36034,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36035,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36036,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36039,-1},
        {-1},
        {-1,-1},
        {36039,-1,-1}
    },
    {
        {36039,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36040,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24676,-1,-1}
    },
    {
        {36042,-1},
        {-1},
        {-1,-1},
        {36042,-1,-1}
    },
    {
        {36042,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36045,-1},
        {-1},
        {-1,-1},
        {36115,-1,-1}
    },
    {
        {36046,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33063,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36049,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22882,36054},
        {36054,-1,-1}
    },
    {
        {36051,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36051,-1},
        {36051,64100,-1}
    },
    {
        {36053,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36054,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26777,-1},
        {-1,-1,-1}
    },
    {
        {36057,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36058,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36059,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36060,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29723,-1,-1}
    },
    {
        {36062,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33094,-1},
        {-1,-1,-1}
    },
    {
        {36064,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36065,-1},
        {-1},
        {-1,-1},
        {32154,32396,-1}
    },
    {
        {36066,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36067,-1},
        {-1},
        {-1,-1},
        {22770,-1,-1}
    },
    {
        {36068,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31929,31883},
        {-1,-1,-1}
    },
    {
        {36070,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20499,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36074,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40782,-1},
        {40782,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24115,-1,-1}
    },
    {
        {36077,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36077,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36112,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36080,-1},
        {-1},
        {-1,-1},
        {20550,-1,-1}
    },
    {
        {36081,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21705,25117},
        {-1,-1,-1}
    },
    {
        {36084,-1},
        {-1},
        {-1,-1},
        {38972,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40782,-1,-1}
    },
    {
        {36088,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37808,-1},
        {-1,-1,-1}
    },
    {
        {36090,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36090,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36091,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36091,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36092,-1},
        {-1},
        {-1,-1},
        {36092,-1,-1}
    },
    {
        {36092,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36092,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36093,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36094,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25943,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36100,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36101,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36101,-1},
        {-1},
        {-1,-1},
        {36101,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36103,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36104,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36105,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36105,-1},
        {-1},
        {-1,-1},
        {36105,-1,-1}
    },
    {
        {36106,-1},
        {-1},
        {-1,-1},
        {36059,-1,-1}
    },
    {
        {36106,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36107,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36109,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36110,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36110,-1},
        {-1},
        {-1,-1},
        {36110,-1,-1}
    },
    {
        {36111,-1},
        {-1},
        {-1,-1},
        {36111,-1,-1}
    },
    {
        {36111,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36111,-1},
        {-1},
        {-1,-1},
        {36111,-1,-1}
    },
    {
        {36111,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36112,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36112,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36123,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36123,-1},
        {36123,-1,-1}
    },
    {
        {36114,-1},
        {-1},
        {-1,-1},
        {36066,-1,-1}
    },
    {
        {36115,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36115,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36116,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36118,-1},
        {-1,-1,-1}
    },
    {
        {36118,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36114,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26407,-1},
        {-1,-1,-1}
    },
    {
        {36123,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36123,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36115,-1},
        {36115,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36115,-1},
        {36115,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36115,-1},
        {36115,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36196,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36198,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36199,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36203,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36204,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36205,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36205,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38963,31958},
        {-1,-1,-1}
    },
    {
        {36208,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36209,-1},
        {-1},
        {-1,-1},
        {36208,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21163,36212},
        {-1,-1,-1}
    },
    {
        {36211,-1},
        {-1},
        {-1,-1},
        {36211,-1,-1}
    },
    {
        {36211,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36211,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36212,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36245,-1},
        {-1,-1,-1}
    },
    {
        {36215,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36215,-1},
        {-1},
        {-1,-1},
        {36215,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36211,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36225,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39854,-1},
        {39854,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36817,36215},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36978,36933,36234}
    },
    {
        {-1,-1},
        {-1},
        {26617,36211},
        {-1,-1,-1}
    },
    {
        {36225,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36225,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36225,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36300,-1},
        {-1,-1,-1}
    },
    {
        {36228,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36229,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36228,26566},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31449,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36228,26383},
        {-1,-1,-1}
    },
    {
        {36234,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36264,36259},
        {36264,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26666,36259},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36234,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36259,-1},
        {-1,-1,-1}
    },
    {
        {36245,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30700,36245},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20419,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36228,26806},
        {-1,-1,-1}
    },
    {
        {36249,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36249,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36895,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36225,37329},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36264,40633},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36249,36914},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36259,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36229,23445},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36946,-1},
        {-1,-1,-1}
    },
    {
        {36262,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36245,36406},
        {-1,-1,-1}
    },
    {
        {36264,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31145,36229},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31145,36229},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20697,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27189,36271},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36481,-1},
        {36481,-1,-1}
    },
    {
        {36271,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36271,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36275,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29228,-1},
        {29228,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33007,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36282,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32033,36310},
        {-1,-1,-1}
    },
    {
        {36286,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36299,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29228,-1},
        {-1,-1,-1}
    },
    {
        {36290,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36294,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36296,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36299,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36299,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36300,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36302,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36303,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29628,36286},
        {-1,-1,-1}
    },
    {
        {36305,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36332,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36310,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36311,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36451,-1},
        {36451,-1,-1}
    },
    {
        {36313,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36314,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36314,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36314,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36314,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36315,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36302,-1},
        {-1,-1,-1}
    },
    {
        {36317,-1},
        {-1},
        {-1,-1},
        {36317,-1,-1}
    },
    {
        {36317,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36319,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22839,-1},
        {-1,-1,-1}
    },
    {
        {36321,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36264,-1},
        {-1,-1,-1}
    },
    {
        {36323,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39609,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36328,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36328,-1},
        {-1},
        {-1,-1},
        {36328,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36341,32079},
        {-1,-1,-1}
    },
    {
        {36330,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36331,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36331,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36332,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32115,36335},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36332,34541},
        {-1,-1,-1}
    },
    {
        {36335,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36339,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36339,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36341,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36348,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36349,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20496,20495},
        {-1,-1,-1}
    },
    {
        {36351,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40810,-1},
        {40810,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36313,37559},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36358,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37570,-1},
        {-1,-1,-1}
    },
    {
        {36360,-1},
        {-1},
        {-1,-1},
        {30094,-1,-1}
    },
    {
        {36361,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36361,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36362,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33050,-1,-1}
    },
    {
        {36364,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36367,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36368,-1},
        {-1},
        {-1,-1},
        {36341,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36918,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36468,-1},
        {-1,-1,-1}
    },
    {
        {36372,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36406,-1},
        {-1,-1,-1}
    },
    {
        {36374,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35542,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36332,30566},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36358,-1},
        {-1,-1,-1}
    },
    {
        {36381,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36382,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36383,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36384,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36385,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36386,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36387,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33795,29469},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36510,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36391,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36452,-1},
        {36452,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36394,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30896,25533},
        {-1,-1,-1}
    },
    {
        {36400,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36400,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36929,-1},
        {36929,36975,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36404,-1},
        {-1},
        {-1,-1},
        {36362,-1,-1}
    },
    {
        {36405,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36406,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36409,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37749,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34680,36475},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36386,-1},
        {-1,-1,-1}
    },
    {
        {36413,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34656,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36441,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36418,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35582,36487},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35582,36487},
        {-1,-1,-1}
    },
    {
        {36420,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36420,-1},
        {-1},
        {-1,-1},
        {36420,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26946,36367},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33151,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33151,-1},
        {-1,-1,-1}
    },
    {
        {36423,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36424,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36424,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36425,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36426,-1},
        {-1},
        {-1,-1},
        {24495,-1,-1}
    },
    {
        {36426,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36427,-1},
        {-1},
        {-1,-1},
        {36367,-1,-1}
    },
    {
        {36427,-1},
        {-1},
        {-1,-1},
        {36367,-1,-1}
    },
    {
        {36428,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36420,-1,-1}
    },
    {
        {36432,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36513,-1},
        {-1,-1,-1}
    },
    {
        {36436,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36437,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36437,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27162,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36296,-1},
        {-1,-1,-1}
    },
    {
        {36441,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25745,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23651,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36332,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36332,-1,-1}
    },
    {
        {36447,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36448,-1},
        {-1},
        {-1,-1},
        {36310,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36450,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36450,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36451,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36452,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32323,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31967,-1},
        {36973,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36225,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32885,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36475,-1},
        {-1,-1,-1}
    },
    {
        {36460,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36461,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36463,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36464,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36466,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36466,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36466,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36468,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36441,36468},
        {36468,-1,-1}
    },
    {
        {36470,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36470,-1},
        {36470,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36470,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36522,-1},
        {36522,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36427,36367,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36475,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36476,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36481,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36484,-1,-1}
    },
    {
        {36484,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36485,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22129,-1},
        {-1,-1,-1}
    },
    {
        {36487,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36487,-1},
        {-1},
        {-1,-1},
        {36487,-1,-1}
    },
    {
        {36487,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27268,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36490,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36491,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20763,-1},
        {-1,-1,-1}
    },
    {
        {36493,-1},
        {-1},
        {-1,-1},
        {36493,-1,-1}
    },
    {
        {36493,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36497,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36497,-1},
        {-1},
        {-1,-1},
        {36497,-1,-1}
    },
    {
        {36497,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36497,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36499,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36500,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32380,-1},
        {-1,-1,-1}
    },
    {
        {36504,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36505,-1},
        {-1},
        {-1,-1},
        {36522,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36510,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38002,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36484,34326},
        {-1,-1,-1}
    },
    {
        {36513,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36513,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36513,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36427,36367,-1}
    },
    {
        {-1,-1},
        {-1},
        {36493,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36374,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31428,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36466,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36522,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36522,-1},
        {-1},
        {-1,-1},
        {36522,-1,-1}
    },
    {
        {36523,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36524,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36524,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32829,-1,-1}
    },
    {
        {36526,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36527,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36527,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36528,-1},
        {-1},
        {-1,-1},
        {20307,39636,-1}
    },
    {
        {36528,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36529,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36529,-1},
        {36529,-1,-1}
    },
    {
        {36531,-1},
        {-1},
        {-1,-1},
        {36524,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36524,33350},
        {-1,-1,-1}
    },
    {
        {36533,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36534,-1},
        {-1},
        {-1,-1},
        {35064,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30702,-1},
        {30702,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36531,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38816,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36539,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34648,36544},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34648,36544},
        {-1,-1,-1}
    },
    {
        {36542,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36542,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31968,36552},
        {-1,-1,-1}
    },
    {
        {36544,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36544,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36526,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22194,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32887,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32887,-1,-1}
    },
    {
        {36549,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36549,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36550,-1},
        {-1},
        {-1,-1},
        {20307,39636,-1}
    },
    {
        {36550,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36552,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36552,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23338,-1},
        {-1,-1,-1}
    },
    {
        {36554,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36555,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36556,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36556,-1,-1}
    },
    {
        {36557,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36682,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35338,36564},
        {-1,-1,-1}
    },
    {
        {36561,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36562,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33539,-1},
        {-1,-1,-1}
    },
    {
        {36564,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36564,-1},
        {-1},
        {-1,-1},
        {36564,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36674,36611},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38515,-1},
        {38515,-1,-1}
    },
    {
        {36571,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36555,34461},
        {-1,-1,-1}
    },
    {
        {36575,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36562,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40660,-1},
        {-1,-1,-1}
    },
    {
        {36578,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36579,-1},
        {-1},
        {-1,-1},
        {36703,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36562,21628},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32970,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36601,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36556,35293},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36602,25260},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36564,-1},
        {-1,-1,-1}
    },
    {
        {36587,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36562,22378},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36649,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34500,36613},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38788,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36618,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39385,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36571,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36600,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36601,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36602,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36603,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36604,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36605,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36606,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36608,-1},
        {-1},
        {-1,-1},
        {36700,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36611,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36611,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35301,-1},
        {-1,-1,-1}
    },
    {
        {36613,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36611,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36616,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36617,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36618,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30060,-1},
        {-1,-1,-1}
    },
    {
        {36620,-1},
        {-1},
        {-1,-1},
        {36635,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36613,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38808,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36662,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25411,36613},
        {-1,-1,-1}
    },
    {
        {36626,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36627,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36627,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36628,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36629,-1},
        {-1},
        {-1,-1},
        {36605,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36631,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36633,-1},
        {-1},
        {-1,-1},
        {36626,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36635,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36636,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36637,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36637,-1},
        {-1},
        {-1,-1},
        {36637,-1,-1}
    },
    {
        {36638,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36638,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36639,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36695,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20506,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34619,36628},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34619,36628},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36646,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36649,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36649,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36650,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36651,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36653,-1},
        {-1},
        {-1,-1},
        {36575,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36655,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22296,26274},
        {26274,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37262,36695},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22243,-1},
        {-1,-1,-1}
    },
    {
        {36659,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36662,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36703,-1,-1}
    },
    {
        {36664,-1},
        {-1},
        {-1,-1},
        {36664,-1,-1}
    },
    {
        {36664,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36665,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36636,-1},
        {-1},
        {-1,-1},
        {36636,-1,-1}
    },
    {
        {36667,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36672,-1},
        {-1},
        {36672,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30042,-1},
        {-1,-1,-1}
    },
    {
        {36670,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36671,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36672,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38876,-1,-1}
    },
    {
        {36674,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36676,-1},
        {-1},
        {-1,-1},
        {36676,-1,-1}
    },
    {
        {36676,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36676,-1},
        {-1},
        {-1,-1},
        {36676,-1,-1}
    },
    {
        {36677,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36678,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37825,-1},
        {-1,-1,-1}
    },
    {
        {36681,-1},
        {-1},
        {-1,-1},
        {36578,-1,-1}
    },
    {
        {36682,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29833,36628},
        {-1,-1,-1}
    },
    {
        {36684,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36684,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36685,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36686,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36665,20693},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36692,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36692,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33378,36686},
        {-1,-1,-1}
    },
    {
        {36695,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36696,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33380,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36674,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36685,23565},
        {-1,-1,-1}
    },
    {
        {36700,-1},
        {-1},
        {-1,-1},
        {36608,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36671,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36703,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27345,-1},
        {-1,-1,-1}
    },
    {
        {36705,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36706,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36707,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36707,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36708,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36692,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36692,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36763,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36763,-1},
        {-1},
        {-1,-1},
        {36763,-1,-1}
    },
    {
        {36764,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36772,-1,-1}
    },
    {
        {36766,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36767,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32618,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36771,-1},
        {36771,-1,-1}
    },
    {
        {36771,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36772,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36774,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36775,-1},
        {-1},
        {-1,-1},
        {36776,24321,-1}
    },
    {
        {36776,-1},
        {-1},
        {-1,-1},
        {24321,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26001,29677},
        {26001,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26001,29677},
        {-1,-1,-1}
    },
    {
        {36781,-1},
        {-1},
        {-1,-1},
        {36766,-1,-1}
    },
    {
        {36781,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36782,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36783,-1},
        {-1},
        {-1,-1},
        {24321,-1,-1}
    },
    {
        {36784,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36785,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36785,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36786,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36786,-1},
        {-1,-1,-1}
    },
    {
        {36788,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36789,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36790,-1},
        {-1},
        {-1,-1},
        {36789,-1,-1}
    },
    {
        {36791,-1},
        {-1},
        {-1,-1},
        {36791,-1,-1}
    },
    {
        {36791,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36794,37002},
        {36794,37002,-1}
    },
    {
        {36794,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36794,-1},
        {-1},
        {-1,-1},
        {36794,37002,-1}
    },
    {
        {36795,-1},
        {-1},
        {-1,-1},
        {36795,-1,-1}
    },
    {
        {36795,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36796,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36796,-1},
        {-1},
        {-1,-1},
        {36796,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36948,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36948,-1,-1}
    },
    {
        {36799,-1},
        {-1},
        {-1,-1},
        {36799,-1,-1}
    },
    {
        {36799,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36801,-1},
        {-1},
        {-1,-1},
        {36983,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36802,-1},
        {-1},
        {-1,-1},
        {36802,-1,-1}
    },
    {
        {36802,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36802,-1},
        {36802,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36802,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36802,-1},
        {36802,-1,-1}
    },
    {
        {36804,-1},
        {-1},
        {-1,-1},
        {36804,-1,-1}
    },
    {
        {36804,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36805,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36805,-1},
        {-1},
        {-1,-1},
        {36805,-1,-1}
    },
    {
        {36806,-1},
        {-1},
        {-1,-1},
        {36806,-1,-1}
    },
    {
        {36806,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36942,-1},
        {36942,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36810,-1},
        {-1},
        {-1,-1},
        {36810,-1,-1}
    },
    {
        {36810,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36913,36877},
        {-1,-1,-1}
    },
    {
        {36813,-1},
        {-1},
        {-1,-1},
        {36813,-1,-1}
    },
    {
        {36813,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36814,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36814,-1},
        {-1},
        {-1,-1},
        {36814,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36948,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36817,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36817,-1},
        {-1},
        {-1,-1},
        {36817,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36819,-1},
        {-1},
        {-1,-1},
        {36819,-1,-1}
    },
    {
        {36819,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36819,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36820,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36820,-1},
        {-1},
        {-1,-1},
        {36820,-1,-1}
    },
    {
        {36821,-1},
        {-1},
        {-1,-1},
        {36821,-1,-1}
    },
    {
        {36821,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36996,-1},
        {36996,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36889,-1},
        {36889,-1,-1}
    },
    {
        {36826,-1},
        {-1},
        {-1,-1},
        {36826,-1,-1}
    },
    {
        {36826,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36832,-1},
        {-1},
        {-1,-1},
        {36832,-1,-1}
    },
    {
        {36832,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36978,36933},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36978,36933},
        {-1,-1,-1}
    },
    {
        {36834,-1},
        {-1},
        {-1,-1},
        {36834,-1,-1}
    },
    {
        {36834,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36836,-1},
        {-1},
        {-1,-1},
        {36836,-1,-1}
    },
    {
        {36836,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36837,-1},
        {-1},
        {-1,-1},
        {36837,-1,-1}
    },
    {
        {36837,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36838,-1},
        {-1},
        {-1,-1},
        {36838,-1,-1}
    },
    {
        {36838,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38515,-1},
        {38515,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36840,-1},
        {-1},
        {-1,-1},
        {36840,-1,-1}
    },
    {
        {36840,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36841,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36841,-1},
        {-1},
        {-1,-1},
        {36999,-1,-1}
    },
    {
        {36841,-1},
        {-1},
        {-1,-1},
        {36999,-1,-1}
    },
    {
        {36842,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36842,-1},
        {-1},
        {-1,-1},
        {36842,-1,-1}
    },
    {
        {36843,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36843,-1},
        {-1},
        {-1,-1},
        {36843,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24448,21281},
        {21281,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24448,21281},
        {24448,-1,-1}
    },
    {
        {36845,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36845,-1},
        {-1},
        {-1,-1},
        {36845,-1,-1}
    },
    {
        {36846,-1},
        {-1},
        {-1,-1},
        {36846,-1,-1}
    },
    {
        {36846,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36847,-1},
        {-1},
        {-1,-1},
        {36867,-1,-1}
    },
    {
        {36847,-1},
        {-1},
        {-1,-1},
        {36867,-1,-1}
    },
    {
        {36847,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36848,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36848,-1},
        {-1},
        {-1,-1},
        {36848,-1,-1}
    },
    {
        {36849,-1},
        {-1},
        {-1,-1},
        {36849,-1,-1}
    },
    {
        {36849,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36852,-1},
        {-1},
        {-1,-1},
        {36852,22238,-1}
    },
    {
        {36852,-1},
        {-1},
        {-1,-1},
        {22238,-1,-1}
    },
    {
        {36853,-1},
        {-1},
        {-1,-1},
        {36853,-1,-1}
    },
    {
        {36853,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36855,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36855,-1},
        {-1},
        {-1,-1},
        {36855,-1,-1}
    },
    {
        {36856,-1},
        {-1},
        {-1,-1},
        {36856,-1,-1}
    },
    {
        {36856,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36857,-1},
        {-1},
        {-1,-1},
        {36857,36321,-1}
    },
    {
        {36857,-1},
        {-1},
        {-1,-1},
        {36321,-1,-1}
    },
    {
        {36858,-1},
        {-1},
        {-1,-1},
        {36858,-1,-1}
    },
    {
        {36858,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36859,-1},
        {-1},
        {-1,-1},
        {36859,-1,-1}
    },
    {
        {36859,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36861,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36861,-1},
        {-1},
        {-1,-1},
        {36861,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27957,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27957,-1},
        {-1,-1,-1}
    },
    {
        {36864,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36864,-1},
        {-1},
        {-1,-1},
        {36864,-1,-1}
    },
    {
        {36865,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36865,-1},
        {-1},
        {-1,-1},
        {36865,-1,-1}
    },
    {
        {36866,-1},
        {-1},
        {-1,-1},
        {36866,-1,-1}
    },
    {
        {36866,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36867,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36867,-1},
        {-1},
        {-1,-1},
        {36867,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36869,-1},
        {-1},
        {-1,-1},
        {36869,-1,-1}
    },
    {
        {36869,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36870,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36870,-1},
        {-1},
        {-1,-1},
        {36870,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36806,-1},
        {-1,-1,-1}
    },
    {
        {36872,-1},
        {-1},
        {-1,-1},
        {36837,-1,-1}
    },
    {
        {36872,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36875,-1},
        {-1},
        {-1,-1},
        {36875,-1,-1}
    },
    {
        {36875,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36876,-1},
        {-1},
        {-1,-1},
        {36876,-1,-1}
    },
    {
        {36876,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36877,-1},
        {-1},
        {-1,-1},
        {36877,-1,-1}
    },
    {
        {36877,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36877,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36878,-1},
        {-1},
        {-1,-1},
        {36878,-1,-1}
    },
    {
        {36878,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36878,-1},
        {-1},
        {-1,-1},
        {36946,36878,-1}
    },
    {
        {36878,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36879,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36879,-1},
        {-1},
        {-1,-1},
        {36879,-1,-1}
    },
    {
        {36880,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36880,-1},
        {-1},
        {-1,-1},
        {36880,-1,-1}
    },
    {
        {36880,-1},
        {-1},
        {-1,-1},
        {36880,-1,-1}
    },
    {
        {36881,-1},
        {-1},
        {-1,-1},
        {36848,36881,-1}
    },
    {
        {36881,-1},
        {-1},
        {-1,-1},
        {36848,-1,-1}
    },
    {
        {36881,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36883,36958},
        {36883,36958,-1}
    },
    {
        {36883,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36883,-1},
        {-1},
        {-1,-1},
        {36883,36958,-1}
    },
    {
        {36884,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36884,-1},
        {-1},
        {-1,-1},
        {36884,-1,-1}
    },
    {
        {36885,-1},
        {-1},
        {-1,-1},
        {24452,36885,-1}
    },
    {
        {36885,-1},
        {-1},
        {-1,-1},
        {24452,-1,-1}
    },
    {
        {36886,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36886,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36886,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36887,-1},
        {-1},
        {-1,-1},
        {36887,-1,-1}
    },
    {
        {36887,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36889,-1},
        {-1},
        {-1,-1},
        {36889,-1,-1}
    },
    {
        {36889,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36890,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36890,-1},
        {-1},
        {-1,-1},
        {36890,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36821,-1},
        {-1,-1,-1}
    },
    {
        {36893,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36893,-1},
        {-1},
        {-1,-1},
        {36893,-1,-1}
    },
    {
        {36894,-1},
        {-1},
        {-1,-1},
        {36894,-1,-1}
    },
    {
        {36894,-1},
        {-1},
        {-1,-1},
        {36894,-1,-1}
    },
    {
        {36894,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36894,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36895,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36895,-1},
        {-1},
        {-1,-1},
        {36895,-1,-1}
    },
    {
        {36896,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36896,-1},
        {-1},
        {-1,-1},
        {36896,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36897,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36897,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36898,-1},
        {-1},
        {-1,-1},
        {36898,-1,-1}
    },
    {
        {36898,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36899,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36899,-1},
        {-1},
        {-1,-1},
        {36899,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27801,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27801,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36852,24317},
        {-1,-1,-1}
    },
    {
        {36903,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36903,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24793,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24793,-1},
        {-1,-1,-1}
    },
    {
        {36856,-1},
        {-1},
        {-1,-1},
        {36856,-1,-1}
    },
    {
        {36856,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36909,-1},
        {-1},
        {-1,-1},
        {36909,-1,-1}
    },
    {
        {36909,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36910,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36910,-1},
        {-1},
        {-1,-1},
        {36910,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36938,-1},
        {36938,-1,-1}
    },
    {
        {36913,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36913,-1},
        {-1},
        {-1,-1},
        {36913,-1,-1}
    },
    {
        {36914,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36914,-1},
        {-1},
        {-1,-1},
        {36914,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36917,-1},
        {-1},
        {-1,-1},
        {36917,-1,-1}
    },
    {
        {36917,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36917,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36918,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36918,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36919,-1},
        {-1},
        {-1,-1},
        {36919,-1,-1}
    },
    {
        {36919,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36920,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36920,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36920,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36920,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36921,-1},
        {-1},
        {-1,-1},
        {36948,-1,-1}
    },
    {
        {36921,-1},
        {-1},
        {-1,-1},
        {36948,-1,-1}
    },
    {
        {36921,-1},
        {-1},
        {-1,-1},
        {36948,-1,-1}
    },
    {
        {36921,-1},
        {-1},
        {-1,-1},
        {36948,-1,-1}
    },
    {
        {36924,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36924,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36896,33509},
        {-1,-1,-1}
    },
    {
        {36926,-1},
        {-1},
        {-1,-1},
        {36926,-1,-1}
    },
    {
        {36926,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36926,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38568,38543},
        {38543,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38568,38543},
        {38543,-1,-1}
    },
    {
        {36929,-1},
        {-1},
        {-1,-1},
        {36929,-1,-1}
    },
    {
        {36929,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36929,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36930,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36930,-1},
        {-1},
        {-1,-1},
        {36930,-1,-1}
    },
    {
        {36930,-1},
        {-1},
        {-1,-1},
        {36930,-1,-1}
    },
    {
        {36930,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36931,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36932,-1},
        {-1},
        {-1,-1},
        {36932,-1,-1}
    },
    {
        {36932,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36933,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36933,-1},
        {-1},
        {-1,-1},
        {36978,36933,-1}
    },
    {
        {-1,-1},
        {-1},
        {36883,36958},
        {-1,-1,-1}
    },
    {
        {36935,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36935,-1},
        {-1},
        {-1,-1},
        {36935,-1,-1}
    },
    {
        {36935,-1},
        {-1},
        {-1,-1},
        {36935,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28252,-1},
        {-1,-1,-1}
    },
    {
        {36937,-1},
        {-1},
        {-1,-1},
        {36937,-1,-1}
    },
    {
        {36937,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36938,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36938,-1},
        {-1},
        {-1,-1},
        {36938,-1,-1}
    },
    {
        {36939,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36939,-1},
        {-1},
        {-1,-1},
        {36939,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36941,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36941,-1},
        {-1},
        {-1,-1},
        {36941,-1,-1}
    },
    {
        {36942,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36942,-1},
        {-1},
        {-1,-1},
        {36942,-1,-1}
    },
    {
        {36943,-1},
        {-1},
        {-1,-1},
        {36943,-1,-1}
    },
    {
        {36943,-1},
        {-1},
        {-1,-1},
        {36943,-1,-1}
    },
    {
        {36943,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36943,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36944,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36944,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36944,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36945,-1},
        {-1},
        {-1,-1},
        {36945,-1,-1}
    },
    {
        {36945,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36946,-1},
        {-1},
        {-1,-1},
        {36946,-1,-1}
    },
    {
        {36946,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36946,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36946,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36946,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36946,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36947,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36947,-1},
        {-1},
        {-1,-1},
        {36947,-1,-1}
    },
    {
        {36948,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36948,-1},
        {-1},
        {-1,-1},
        {36948,-1,-1}
    },
    {
        {36949,-1},
        {-1},
        {-1,-1},
        {36949,-1,-1}
    },
    {
        {36949,-1},
        {-1},
        {-1,-1},
        {36949,-1,-1}
    },
    {
        {36949,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36949,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36949,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36950,-1},
        {-1},
        {-1,-1},
        {36950,-1,-1}
    },
    {
        {36950,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36952,-1},
        {-1},
        {-1,-1},
        {36952,-1,-1}
    },
    {
        {36952,-1},
        {-1},
        {-1,-1},
        {36952,-1,-1}
    },
    {
        {36952,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36952,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36953,-1},
        {-1},
        {-1,-1},
        {36953,-1,-1}
    },
    {
        {36953,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28338,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36956,-1},
        {-1},
        {-1,-1},
        {36956,-1,-1}
    },
    {
        {36956,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36957,-1},
        {-1},
        {-1,-1},
        {36957,-1,-1}
    },
    {
        {36957,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36958,-1},
        {-1},
        {-1,-1},
        {36883,36958,-1}
    },
    {
        {36958,-1},
        {-1},
        {-1,-1},
        {36883,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36978,36933},
        {36933,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36978,36933},
        {36933,-1,-1}
    },
    {
        {36960,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36960,-1},
        {-1},
        {-1,-1},
        {36960,-1,-1}
    },
    {
        {36960,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36961,-1},
        {-1},
        {-1,-1},
        {36961,-1,-1}
    },
    {
        {36961,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36963,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36963,-1},
        {-1},
        {-1,-1},
        {36963,-1,-1}
    },
    {
        {36965,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36965,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36968,-1},
        {-1},
        {-1,-1},
        {36968,-1,-1}
    },
    {
        {36968,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36968,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36969,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36969,-1},
        {-1},
        {-1,-1},
        {36969,-1,-1}
    },
    {
        {36969,-1},
        {-1},
        {-1,-1},
        {36969,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28402,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23273,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36973,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36973,-1},
        {-1},
        {-1,-1},
        {36973,-1,-1}
    },
    {
        {36974,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36974,-1},
        {-1},
        {-1,-1},
        {36974,-1,-1}
    },
    {
        {36974,-1},
        {-1},
        {-1,-1},
        {36974,-1,-1}
    },
    {
        {36974,-1},
        {-1},
        {-1,-1},
        {36974,-1,-1}
    },
    {
        {36975,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36975,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36975,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36975,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36975,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35128,-1},
        {-1,-1,-1}
    },
    {
        {36978,-1},
        {-1},
        {-1,-1},
        {36978,36933,-1}
    },
    {
        {36978,-1},
        {-1},
        {-1,-1},
        {36933,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30700,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30700,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30700,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36981,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36981,-1},
        {-1},
        {-1,-1},
        {36981,-1,-1}
    },
    {
        {36981,-1},
        {-1},
        {-1,-1},
        {36981,-1,-1}
    },
    {
        {36982,-1},
        {-1},
        {-1,-1},
        {36982,-1,-1}
    },
    {
        {36982,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36983,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36983,-1},
        {-1},
        {-1,-1},
        {36983,-1,-1}
    },
    {
        {36983,-1},
        {-1},
        {-1,-1},
        {36983,-1,-1}
    },
    {
        {36983,-1},
        {-1},
        {-1,-1},
        {36983,-1,-1}
    },
    {
        {36983,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36983,-1},
        {-1},
        {-1,-1},
        {36983,-1,-1}
    },
    {
        {36984,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36984,-1},
        {-1},
        {-1,-1},
        {36984,-1,-1}
    },
    {
        {36984,-1},
        {-1},
        {-1,-1},
        {36984,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36986,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36986,-1},
        {-1},
        {-1,-1},
        {36986,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36988,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36988,-1},
        {-1},
        {-1,-1},
        {36988,-1,-1}
    },
    {
        {36989,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36989,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36989,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36989,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36989,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36991,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36991,-1},
        {-1},
        {-1,-1},
        {36991,-1,-1}
    },
    {
        {36992,-1},
        {-1},
        {-1,-1},
        {36992,-1,-1}
    },
    {
        {36992,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36992,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36993,-1},
        {-1},
        {-1,-1},
        {36993,-1,-1}
    },
    {
        {36993,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36993,-1},
        {-1},
        {-1,-1},
        {36993,-1,-1}
    },
    {
        {36993,-1},
        {-1},
        {-1,-1},
        {36993,-1,-1}
    },
    {
        {36994,-1},
        {-1},
        {-1,-1},
        {36994,-1,-1}
    },
    {
        {36994,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36995,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36995,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36995,-1},
        {-1},
        {-1,-1},
        {36995,-1,-1}
    },
    {
        {36995,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36995,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36995,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36996,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36996,-1},
        {-1},
        {-1,-1},
        {36996,-1,-1}
    },
    {
        {36997,-1},
        {-1},
        {-1,-1},
        {36997,-1,-1}
    },
    {
        {36997,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28548,-1},
        {-1,-1,-1}
    },
    {
        {36999,-1},
        {-1},
        {-1,-1},
        {36999,-1,-1}
    },
    {
        {36999,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36999,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37000,-1},
        {-1},
        {-1,-1},
        {37000,-1,-1}
    },
    {
        {37000,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37001,-1},
        {-1},
        {-1,-1},
        {36794,37002,-1}
    },
    {
        {37001,-1},
        {-1},
        {-1,-1},
        {37001,-1,-1}
    },
    {
        {37001,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37001,-1},
        {-1},
        {-1,-1},
        {37001,-1,-1}
    },
    {
        {37001,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37001,-1},
        {-1},
        {-1,-1},
        {37001,-1,-1}
    },
    {
        {37001,-1},
        {-1},
        {-1,-1},
        {37001,-1,-1}
    },
    {
        {37001,-1},
        {-1},
        {-1,-1},
        {37001,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37002,-1},
        {-1},
        {-1,-1},
        {36794,37002,-1}
    },
    {
        {37002,-1},
        {-1},
        {-1,-1},
        {36794,-1,-1}
    },
    {
        {37002,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37002,-1},
        {-1},
        {-1,-1},
        {37002,-1,-1}
    },
    {
        {37002,-1},
        {-1},
        {-1,-1},
        {37002,-1,-1}
    },
    {
        {37002,-1},
        {-1},
        {-1,-1},
        {37002,-1,-1}
    },
    {
        {37002,-1},
        {-1},
        {-1,-1},
        {37002,-1,-1}
    },
    {
        {37002,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37002,-1},
        {-1},
        {-1,-1},
        {37002,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37004,-1},
        {-1},
        {-1,-1},
        {37004,-1,-1}
    },
    {
        {37004,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21407,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21407,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37007,-1},
        {-1},
        {-1,-1},
        {37007,-1,-1}
    },
    {
        {37007,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37008,-1},
        {-1},
        {-1,-1},
        {37008,-1,-1}
    },
    {
        {37008,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37009,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37159,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37019,-1},
        {-1,-1,-1}
    },
    {
        {37013,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37019,-1},
        {-1,-1,-1}
    },
    {
        {37015,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37017,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37017,-1},
        {-1},
        {-1,-1},
        {37017,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23142,-1},
        {-1,-1,-1}
    },
    {
        {37019,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37087,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37025,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37026,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37027,-1},
        {-1},
        {-1,-1},
        {37027,-1,-1}
    },
    {
        {37027,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37041,21156},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27784,-1,-1}
    },
    {
        {37030,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37030,-1},
        {-1},
        {-1,-1},
        {37030,-1,-1}
    },
    {
        {37030,-1},
        {-1},
        {-1,-1},
        {37030,-1,-1}
    },
    {
        {37030,-1},
        {-1},
        {-1,-1},
        {37030,-1,-1}
    },
    {
        {37030,-1},
        {-1},
        {-1,-1},
        {37030,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21011,18613},
        {-1,-1,-1}
    },
    {
        {37032,-1},
        {-1},
        {-1,-1},
        {26449,-1,-1}
    },
    {
        {37032,-1},
        {-1},
        {-1,-1},
        {37032,26449,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28783,37103},
        {-1,-1,-1}
    },
    {
        {37034,-1},
        {-1},
        {-1,-1},
        {37034,-1,-1}
    },
    {
        {37034,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37034,-1},
        {-1},
        {-1,-1},
        {37034,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37030,-1},
        {37030,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37030,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37063,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37109,-1},
        {-1,-1,-1}
    },
    {
        {37039,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37040,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37041,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37043,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37045,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37046,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29929,37039},
        {-1,-1,-1}
    },
    {
        {37048,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37177,37138},
        {37138,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21179,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37054,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23546,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37057,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37066,21182},
        {-1,-1,-1}
    },
    {
        {37059,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37060,-1},
        {-1},
        {-1,-1},
        {37092,-1,-1}
    },
    {
        {37061,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21180,37084},
        {-1,-1,-1}
    },
    {
        {37063,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37064,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37026,-1},
        {-1},
        {37026,-1},
        {37026,-1,-1}
    },
    {
        {37066,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32822,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37027,-1},
        {-1,-1,-1}
    },
    {
        {37070,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22880,37165},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37070,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24310,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37159,-1},
        {-1,-1,-1}
    },
    {
        {37079,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37096,25940},
        {-1,-1,-1}
    },
    {
        {37083,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37083,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37084,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37085,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37086,-1},
        {-1},
        {-1,-1},
        {37070,-1,-1}
    },
    {
        {37087,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37099,29990},
        {-1,-1,-1}
    },
    {
        {37089,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37090,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37090,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37090,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28196,-1},
        {-1,-1,-1}
    },
    {
        {37092,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37092,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37096,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37057,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37089,22971},
        {-1,-1,-1}
    },
    {
        {37099,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37057,38738},
        {-1,-1,-1}
    },
    {
        {37101,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24427,37119},
        {-1,-1,-1}
    },
    {
        {37103,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37026,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37026,-1},
        {-1,-1,-1}
    },
    {
        {37106,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20816,20818},
        {-1,-1,-1}
    },
    {
        {37108,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37109,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37110,-1},
        {-1},
        {-1,-1},
        {37096,-1,-1}
    },
    {
        {37111,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {64092,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37013,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37155,38860},
        {-1,-1,-1}
    },
    {
        {37117,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37118,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37119,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37120,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37120,-1},
        {-1},
        {-1,-1},
        {37120,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37046,-1},
        {-1,-1,-1}
    },
    {
        {37122,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37122,-1},
        {-1},
        {-1,-1},
        {37122,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37159,-1},
        {-1,-1,-1}
    },
    {
        {37124,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37126,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20539,20399},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37141,37111},
        {-1},
        {37141,37111},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37141,37111},
        {37111,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37099,21471},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37144,21776},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37086,24687},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36785,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26187,64038},
        {-1,-1,-1}
    },
    {
        {37138,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37140,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37111,-1},
        {-1},
        {-1,-1},
        {37111,-1,-1}
    },
    {
        {37141,-1},
        {-1},
        {-1,-1},
        {37111,-1,-1}
    },
    {
        {37111,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37111,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37142,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37144,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37145,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37172,21119},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37172,21119},
        {-1,-1,-1}
    },
    {
        {37148,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21216,-1},
        {-1,-1,-1}
    },
    {
        {37150,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37154,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37155,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37155,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26364,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35377,-1},
        {-1,-1,-1}
    },
    {
        {37159,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26300,37013},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36027,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37154,29234},
        {-1,-1,-1}
    },
    {
        {37165,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37165,-1},
        {-1},
        {-1,-1},
        {37165,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37167,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37168,-1},
        {-1},
        {-1,-1},
        {38563,-1,-1}
    },
    {
        {37168,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37168,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37169,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37170,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37170,40701},
        {-1,-1,-1}
    },
    {
        {37172,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37054,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37175,-1},
        {-1},
        {-1,-1},
        {37190,-1,-1}
    },
    {
        {37175,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22818,-1},
        {-1,-1,-1}
    },
    {
        {37177,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37178,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36646,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37191,-1,-1}
    },
    {
        {37181,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37181,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24962,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34308,-1},
        {34308,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29141,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21240,37150},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37181,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37181,-1},
        {-1,-1,-1}
    },
    {
        {37190,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37191,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37192,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37193,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37194,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37195,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37195,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37196,-1},
        {-1},
        {-1,-1},
        {37196,-1,-1}
    },
    {
        {37196,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37197,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37197,-1},
        {-1},
        {-1,-1},
        {37197,-1,-1}
    },
    {
        {37197,-1},
        {-1},
        {-1,-1},
        {37197,-1,-1}
    },
    {
        {37197,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37198,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37194,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37194,-1},
        {-1,-1,-1}
    },
    {
        {37202,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37202,-1},
        {-1},
        {-1,-1},
        {37202,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37204,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37221,27622},
        {-1,-1,-1}
    },
    {
        {37206,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37207,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37208,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37209,-1},
        {-1},
        {-1,-1},
        {26015,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37211,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27838,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37194,-1},
        {-1,-1,-1}
    },
    {
        {37217,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37218,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37218,-1},
        {-1},
        {-1,-1},
        {37218,-1,-1}
    },
    {
        {37219,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37220,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37221,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37228,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25130,-1},
        {-1,-1,-1}
    },
    {
        {37225,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37226,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37226,-1},
        {-1},
        {-1,-1},
        {37226,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37236,-1},
        {-1,-1,-1}
    },
    {
        {37228,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37228,-1},
        {-1},
        {-1,-1},
        {37228,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37239,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35427,37259},
        {-1,-1,-1}
    },
    {
        {37234,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37234,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37235,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37236,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37237,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37237,-1},
        {-1},
        {-1,-1},
        {37237,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35496,-1},
        {-1,-1,-1}
    },
    {
        {37239,-1},
        {-1},
        {-1,-1},
        {37239,-1,-1}
    },
    {
        {37239,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37240,-1},
        {-1},
        {-1,-1},
        {37240,-1,-1}
    },
    {
        {37240,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37240,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37241,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35504,37294},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37250,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37251,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37196,32175},
        {-1,-1,-1}
    },
    {
        {37253,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30430,-1,-1}
    },
    {
        {37255,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29124,-1},
        {-1,-1,-1}
    },
    {
        {37257,-1},
        {-1},
        {-1,-1},
        {37204,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37259,-1},
        {-1},
        {-1,-1},
        {37218,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37261,-1},
        {-1,-1,-1}
    },
    {
        {37261,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37262,-1},
        {-1},
        {-1,-1},
        {40569,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35576,37259},
        {-1,-1,-1}
    },
    {
        {37264,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37266,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37195,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37255,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37278,-1},
        {-1},
        {28201,37278},
        {37278,-1,-1}
    },
    {
        {37271,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27068,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37208,34699},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35598,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35598,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33603,37236},
        {-1,-1,-1}
    },
    {
        {37276,-1},
        {-1},
        {-1,-1},
        {37276,-1,-1}
    },
    {
        {37276,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37191,-1},
        {-1,-1,-1}
    },
    {
        {37278,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37282,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37284,-1},
        {-1},
        {-1,-1},
        {37292,-1,-1}
    },
    {
        {37284,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37858,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35635,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35635,-1},
        {-1,-1,-1}
    },
    {
        {37288,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37288,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31967,-1},
        {31967,-1,-1}
    },
    {
        {37290,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37290,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37291,-1},
        {-1},
        {-1,-1},
        {21307,-1,-1}
    },
    {
        {37291,-1},
        {-1},
        {-1,-1},
        {21307,-1,-1}
    },
    {
        {37292,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37292,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37294,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37295,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37297,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37299,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37300,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37301,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37301,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37301,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37302,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21307,37291},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21307,37291},
        {-1,-1,-1}
    },
    {
        {37304,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36700,-1},
        {-1,-1,-1}
    },
    {
        {37306,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37306,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37228,-1,-1}
    },
    {
        {37308,-1},
        {-1},
        {-1,-1},
        {23476,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37316,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37312,-1},
        {-1},
        {-1,-1},
        {37304,-1,-1}
    },
    {
        {37313,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37313,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37313,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37313,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37313,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37315,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37316,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37318,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37319,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37319,-1},
        {-1},
        {-1,-1},
        {37319,-1,-1}
    },
    {
        {37320,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37321,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37323,-1},
        {-1},
        {-1,-1},
        {37320,-1,-1}
    },
    {
        {37323,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37324,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37325,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37326,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37327,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37328,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37329,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37338,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37336,-1},
        {-1,-1,-1}
    },
    {
        {37334,-1},
        {-1},
        {-1,-1},
        {20992,-1,-1}
    },
    {
        {37335,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37336,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37338,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37339,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37340,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37340,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37341,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37343,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37343,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20081,20098},
        {20081,20098,-1}
    },
    {
        {37345,-1},
        {-1},
        {-1,-1},
        {37340,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37334,-1},
        {-1,-1,-1}
    },
    {
        {37347,-1},
        {-1},
        {-1,-1},
        {37347,-1,-1}
    },
    {
        {37347,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37348,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37349,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37350,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37351,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37877,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37877,-1},
        {-1,-1,-1}
    },
    {
        {37356,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37357,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37358,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37358,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37609,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37361,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37501,-1},
        {-1,-1,-1}
    },
    {
        {37365,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37365,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37366,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37371,-1},
        {-1},
        {-1,-1},
        {37338,-1,-1}
    },
    {
        {37372,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37372,-1},
        {-1},
        {-1,-1},
        {21138,21091,-1}
    },
    {
        {37372,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37375,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37377,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27453,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37467,-1,-1}
    },
    {
        {37383,-1},
        {-1},
        {-1,-1},
        {37444,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37356,-1},
        {37433,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37386,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35355,37685},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37389,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37390,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37392,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37392,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37393,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37396,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37397,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37398,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37399,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25810,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37645,37790},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37707,-1},
        {37707,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37406,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37347,-1},
        {37347,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27581,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33322,37399},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37818,-1,-1}
    },
    {
        {37417,-1},
        {-1},
        {-1,-1},
        {29200,28809,37994}
    },
    {
        {37417,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37420,-1},
        {-1},
        {-1,-1},
        {37944,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35292,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37504,30496},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37427,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37428,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37431,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37432,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37432,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37433,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37434,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37436,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23449,-1,-1}
    },
    {
        {37439,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37440,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29645,37806},
        {29645,-1,-1}
    },
    {
        {37442,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37442,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37443,-1},
        {-1},
        {-1,-1},
        {37444,-1,-1}
    },
    {
        {37444,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37445,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37445,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37447,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37448,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37449,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37450,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37451,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37451,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21644,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37454,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37455,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37456,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37457,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27880,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27880,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21277,37445},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37445,-1},
        {-1,-1,-1}
    },
    {
        {37462,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37463,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37465,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37466,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37467,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37467,-1},
        {-1},
        {-1,-1},
        {37467,-1,-1}
    },
    {
        {37467,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37470,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37433,-1},
        {37433,-1,-1}
    },
    {
        {37472,-1},
        {-1},
        {-1,-1},
        {37704,-1,-1}
    },
    {
        {37473,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37473,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37474,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37449,34986},
        {-1,-1,-1}
    },
    {
        {37476,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37477,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37478,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37479,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37480,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37480,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37361,-1},
        {37361,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37604,36303},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37455,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37489,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39730,-1},
        {39730,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37496,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37496,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39180,37847},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37500,-1},
        {-1},
        {-1,-1},
        {37500,-1,-1}
    },
    {
        {37500,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37501,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37502,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37503,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37504,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37406,-1},
        {37406,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37351,-1},
        {-1,-1,-1}
    },
    {
        {37507,-1},
        {-1},
        {-1,-1},
        {37507,-1,-1}
    },
    {
        {37507,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37509,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37457,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37649,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37512,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37513,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37517,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27082,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35010,-1},
        {-1,-1,-1}
    },
    {
        {37521,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37523,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37523,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37473,-1},
        {-1,-1,-1}
    },
    {
        {37525,-1},
        {-1},
        {-1,-1},
        {37444,-1,-1}
    },
    {
        {37526,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37527,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37528,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37529,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37530,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37531,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37532,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35973,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37406,-1},
        {37406,-1,-1}
    },
    {
        {37535,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37431,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37682,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37682,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35005,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37678,-1},
        {-1,-1,-1}
    },
    {
        {37543,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37517,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39850,-1},
        {-1,-1,-1}
    },
    {
        {37547,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26674,-1},
        {-1,-1,-1}
    },
    {
        {37549,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27410,38014},
        {38014,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37554,-1},
        {-1},
        {-1,-1},
        {37356,-1,-1}
    },
    {
        {37613,-1},
        {-1},
        {-1,-1},
        {37613,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37847,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37559,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37559,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37561,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27742,-1},
        {-1,-1,-1}
    },
    {
        {37567,-1},
        {-1},
        {-1,-1},
        {37912,-1,-1}
    },
    {
        {37568,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37986,-1,-1}
    },
    {
        {37570,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37953,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37574,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37574,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37706,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37467,35029},
        {-1,-1,-1}
    },
    {
        {37579,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37579,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37580,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37580,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37580,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37582,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37583,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37584,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37931,-1},
        {-1,-1,-1}
    },
    {
        {37586,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37587,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25405,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37591,-1},
        {-1},
        {-1,-1},
        {37591,-1,-1}
    },
    {
        {37591,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37593,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24736,37738},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37489,37979},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37600,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37604,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37605,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37624,36348},
        {-1,-1,-1}
    },
    {
        {37607,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37444,37941,-1}
    },
    {
        {37609,-1},
        {-1},
        {-1,-1},
        {37609,-1,-1}
    },
    {
        {37609,-1},
        {-1},
        {-1,-1},
        {37609,-1,-1}
    },
    {
        {37609,-1},
        {-1},
        {-1,-1},
        {37609,-1,-1}
    },
    {
        {37610,-1},
        {-1},
        {-1,-1},
        {33303,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29313,37739},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37613,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37470,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37470,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37531,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37501,-1},
        {-1,-1,-1}
    },
    {
        {37618,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37619,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37805,37806},
        {-1,-1,-1}
    },
    {
        {37623,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37624,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37625,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37626,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37627,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37627,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37628,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37530,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37476,32175},
        {-1,-1,-1}
    },
    {
        {37631,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37529,-1},
        {-1,-1,-1}
    },
    {
        {37634,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37634,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37636,-1},
        {-1},
        {-1,-1},
        {37682,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37739,40654},
        {-1,-1,-1}
    },
    {
        {37638,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37638,-1},
        {-1},
        {-1,-1},
        {37638,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37593,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23736,37502},
        {-1,-1,-1}
    },
    {
        {37645,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37450,-1},
        {-1,-1,-1}
    },
    {
        {37647,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37648,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37649,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37427,-1},
        {37427,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37653,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37738,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37561,-1},
        {-1,-1,-1}
    },
    {
        {37656,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37657,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37658,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37658,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37658,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37883,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37661,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37662,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37663,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37664,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37665,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37666,-1},
        {-1},
        {-1,-1},
        {37549,-1,-1}
    },
    {
        {37667,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37669,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37670,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37671,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37672,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37672,-1},
        {-1},
        {-1,-1},
        {37672,-1,-1}
    },
    {
        {37673,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37675,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37676,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24427,-1},
        {-1,-1,-1}
    },
    {
        {37678,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37679,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29645,29646},
        {29645,-1,-1}
    },
    {
        {37682,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37685,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37685,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37583,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37559,-1},
        {-1,-1,-1}
    },
    {
        {37690,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37690,-1},
        {-1},
        {-1,-1},
        {37690,-1,-1}
    },
    {
        {37691,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37700,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37703,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37704,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37704,-1},
        {-1},
        {-1,-1},
        {37704,-1,-1}
    },
    {
        {37705,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37706,-1},
        {-1},
        {-1,-1},
        {37676,-1,-1}
    },
    {
        {37707,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37521,-1},
        {-1,-1,-1}
    },
    {
        {37709,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32239,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37713,-1},
        {-1},
        {-1,-1},
        {37713,-1,-1}
    },
    {
        {37713,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37716,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37718,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37719,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37722,-1},
        {-1},
        {-1,-1},
        {37675,-1,-1}
    },
    {
        {37723,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37724,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32806,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37806,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37454,-1},
        {37454,-1,-1}
    },
    {
        {37728,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37713,-1},
        {37713,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37450,-1},
        {-1,-1,-1}
    },
    {
        {37732,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37733,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37733,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37733,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37366,37447},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25123,-1},
        {25123,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37738,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37739,-1},
        {-1},
        {-1,-1},
        {37740,-1,-1}
    },
    {
        {37740,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37742,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37742,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37744,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37745,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32225,37653},
        {-1,-1,-1}
    },
    {
        {37747,-1},
        {-1},
        {-1,-1},
        {37969,-1,-1}
    },
    {
        {37747,-1},
        {-1},
        {-1,-1},
        {37969,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37749,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37749,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29786,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35552,37679},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37705,-1},
        {-1,-1,-1}
    },
    {
        {37756,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37758,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37657,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27966,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37768,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36427,37653},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37771,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37771,-1},
        {-1},
        {-1,-1},
        {36676,-1,-1}
    },
    {
        {37772,-1},
        {-1},
        {-1,-1},
        {37772,-1,-1}
    },
    {
        {37772,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37625,39228},
        {-1,-1,-1}
    },
    {
        {37775,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35616,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39233,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32808,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37808,34697},
        {-1,-1,-1}
    },
    {
        {37780,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37782,-1},
        {-1},
        {-1,-1},
        {37782,-1,-1}
    },
    {
        {37782,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37783,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39730,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27082,-1},
        {30687,-1,-1}
    },
    {
        {37786,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37786,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37787,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30436,-1},
        {30436,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37790,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21083,37628},
        {21083,-1,-1}
    },
    {
        {37793,-1},
        {-1},
        {-1,-1},
        {37793,-1,-1}
    },
    {
        {37793,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37793,-1},
        {-1},
        {-1,-1},
        {37793,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37795,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26781,37775},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37799,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37804,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37805,-1},
        {-1},
        {37805,37806},
        {37806,-1,-1}
    },
    {
        {37806,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34690,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37808,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37808,-1},
        {-1},
        {-1,-1},
        {37808,-1,-1}
    },
    {
        {37808,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36677,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37852,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37745,33260},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37427,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29802,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37931,-1},
        {37931,-1,-1}
    },
    {
        {37817,-1},
        {-1},
        {-1,-1},
        {37817,-1,-1}
    },
    {
        {37817,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37817,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37817,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37817,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37818,-1},
        {-1},
        {-1,-1},
        {37818,-1,-1}
    },
    {
        {37818,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37782,-1},
        {-1},
        {29795,37782},
        {37782,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30944,-1},
        {-1,-1,-1}
    },
    {
        {37825,-1},
        {-1},
        {-1,-1},
        {37782,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37827,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37830,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37831,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37832,-1},
        {-1},
        {-1,-1},
        {37832,-1,-1}
    },
    {
        {37832,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30050,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37840,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37840,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37841,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37841,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37840,39994},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37892,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37846,-1},
        {-1,-1,-1}
    },
    {
        {37846,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37847,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37848,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25114,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37631,-1},
        {-1,-1,-1}
    },
    {
        {37852,-1},
        {-1},
        {-1,-1},
        {37946,-1,-1}
    },
    {
        {37853,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37854,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37854,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37855,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37586,-1},
        {37586,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37586,-1},
        {-1,-1,-1}
    },
    {
        {37857,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37857,-1},
        {-1},
        {-1,-1},
        {37857,-1,-1}
    },
    {
        {37858,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36448,37709},
        {-1,-1,-1}
    },
    {
        {37860,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37861,-1},
        {-1},
        {-1,-1},
        {37561,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37863,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37864,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37864,-1},
        {37864,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33370,37679},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26792,37745},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32581,-1},
        {32581,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34763,166605},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31968,37854},
        {-1,-1,-1}
    },
    {
        {37873,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37658,-1},
        {37658,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37658,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29840,-1},
        {-1,-1,-1}
    },
    {
        {37877,-1},
        {-1},
        {-1,-1},
        {37877,-1,-1}
    },
    {
        {37877,-1},
        {-1},
        {-1,-1},
        {37877,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37883,-1},
        {-1},
        {-1,-1},
        {37883,-1,-1}
    },
    {
        {37883,-1},
        {-1},
        {-1,-1},
        {37883,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37957,31574},
        {-1,-1,-1}
    },
    {
        {37885,-1},
        {-1},
        {-1,-1},
        {37561,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32342,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37783,-1},
        {37783,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21297,-1},
        {21297,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37891,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37892,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37892,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22684,-1},
        {-1,-1,-1}
    },
    {
        {37895,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37897,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37722,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37722,28271},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35947,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37903,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37904,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37733,-1},
        {37733,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37957,-1},
        {-1,-1,-1}
    },
    {
        {37907,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37908,-1},
        {-1},
        {-1,-1},
        {37908,-1,-1}
    },
    {
        {37908,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37910,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37910,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37911,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37912,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37912,-1},
        {-1},
        {-1,-1},
        {37912,-1,-1}
    },
    {
        {37913,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37914,-1},
        {-1},
        {-1,-1},
        {37647,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37907,-1},
        {37907,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36470,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32808,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35676,-1},
        {-1,-1,-1}
    },
    {
        {37921,-1},
        {-1},
        {-1,-1},
        {37444,37941,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29159,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29159,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29159,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37986,33379},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37986,33379},
        {-1,-1,-1}
    },
    {
        {37931,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37470,-1},
        {37470,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30983,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37772,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37740,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37938,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37939,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37940,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37941,-1},
        {-1},
        {-1,-1},
        {37444,-1,-1}
    },
    {
        {37941,-1},
        {-1},
        {-1,-1},
        {37444,-1,-1}
    },
    {
        {37942,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37745,-1},
        {37745,-1,-1}
    },
    {
        {37944,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37707,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37707,-1},
        {-1,-1,-1}
    },
    {
        {37946,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37947,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37947,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37782,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37940,-1},
        {37940,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37857,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37953,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37306,37758},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37530,-1,-1}
    },
    {
        {37956,-1},
        {-1},
        {-1,-1},
        {37619,-1,-1}
    },
    {
        {37957,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23565,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21137,21092},
        {21137,21092,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37440,-1},
        {37440,-1,-1}
    },
    {
        {37962,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37962,-1},
        {-1},
        {-1,-1},
        {37962,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37864,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32588,37863},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21297,-1},
        {21297,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27320,23527},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37969,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37970,-1},
        {-1},
        {-1,-1},
        {37969,-1,-1}
    },
    {
        {37970,-1},
        {-1},
        {-1,-1},
        {37969,-1,-1}
    },
    {
        {37971,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37971,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37910,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29896,37733},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27345,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37978,-1},
        {-1},
        {-1,-1},
        {38013,-1,-1}
    },
    {
        {37979,-1},
        {-1},
        {-1,-1},
        {37489,-1,-1}
    },
    {
        {37979,-1},
        {-1},
        {-1,-1},
        {37979,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34028,-1},
        {-1,-1,-1}
    },
    {
        {37982,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37984,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37986,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37987,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37451,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33386,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40055,-1},
        {-1,-1,-1}
    },
    {
        {37994,-1},
        {-1},
        {-1,-1},
        {29200,28809,-1}
    },
    {
        {37995,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37969,-1},
        {37969,-1,-1}
    },
    {
        {37997,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37999,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38000,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38001,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38001,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38002,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37749,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35327,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38005,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38005,-1},
        {-1},
        {-1,-1},
        {32592,38005,32566}
    },
    {
        {-1,-1},
        {-1},
        {33247,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33247,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33247,-1},
        {-1,-1,-1}
    },
    {
        {38007,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38007,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38007,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22744,37847},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38012,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38013,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38014,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38015,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38017,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25897,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38263,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38264,-1},
        {-1},
        {-1,-1},
        {38263,-1,-1}
    },
    {
        {38265,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38265,39147},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38265,37444},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32902,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37904,32902},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38272,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38274,-1},
        {-1,-1,-1}
    },
    {
        {38274,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38275,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38272,-1},
        {38272,-1,-1}
    },
    {
        {38278,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38279,-1},
        {-1},
        {-1,-1},
        {38281,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38281,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38282,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38283,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38283,-1},
        {-1},
        {-1,-1},
        {38283,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38284,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38285,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38286,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38287,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38287,-1},
        {-1},
        {38287,-1},
        {38287,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28516,-1},
        {-1,-1,-1}
    },
    {
        {38289,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38291,-1},
        {-1},
        {-1,-1},
        {38291,-1,-1}
    },
    {
        {38291,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38291,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38292,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38292,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38281,-1},
        {-1,-1,-1}
    },
    {
        {38294,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38296,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38297,-1},
        {-1},
        {-1,-1},
        {39719,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38285,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38308,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38303,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38304,-1},
        {-1},
        {-1,-1},
        {38287,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38306,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38307,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38308,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38309,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38310,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38311,-1},
        {-1},
        {-1,-1},
        {39720,24055,-1}
    },
    {
        {38312,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38313,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38315,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38315,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38316,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38317,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38317,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24237,24311},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27801,38346},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38291,-1},
        {-1,-1,-1}
    },
    {
        {38322,-1},
        {-1},
        {-1,-1},
        {38322,-1,-1}
    },
    {
        {38322,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38324,-1},
        {-1},
        {-1,-1},
        {38339,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36522,-1},
        {-1,-1,-1}
    },
    {
        {38326,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27578,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38315,22780},
        {-1,-1,-1}
    },
    {
        {38329,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38331,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38331,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38332,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38332,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38333,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38334,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39720,38311},
        {39720,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35090,35139},
        {-1,-1,-1}
    },
    {
        {38339,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35201,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38343,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38343,-1},
        {-1},
        {-1,-1},
        {38343,-1,-1}
    },
    {
        {38344,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38344,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38346,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38347,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38348,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38349,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38349,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38352,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38352,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38355,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38356,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38357,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38358,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38306,-1},
        {38306,38364,-1}
    },
    {
        {38360,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38362,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38326,-1},
        {-1,-1,-1}
    },
    {
        {38364,-1},
        {-1},
        {-1,-1},
        {38306,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38366,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38369,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38370,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38355,30070},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38373,-1},
        {-1},
        {-1,-1},
        {38373,-1,-1}
    },
    {
        {38373,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38428,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38429,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24041,-1},
        {-1,-1,-1}
    },
    {
        {38433,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23673,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38512,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38525,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38497,-1},
        {38497,-1,-1}
    },
    {
        {38440,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21319,-1},
        {38494,-1,-1}
    },
    {
        {38442,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22351,-1},
        {22351,-1,-1}
    },
    {
        {38444,-1},
        {-1},
        {-1,-1},
        {22353,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38444,21550},
        {-1,-1,-1}
    },
    {
        {38446,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38447,-1},
        {-1},
        {-1,-1},
        {22336,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38447,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31357,-1,-1}
    },
    {
        {38450,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38451,-1},
        {-1},
        {-1,-1},
        {38525,-1,-1}
    },
    {
        {38452,-1},
        {-1},
        {-1,-1},
        {38512,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38440,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38459,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38460,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23994,-1},
        {-1,-1,-1}
    },
    {
        {38463,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38464,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38465,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38466,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26142,-1},
        {-1,-1,-1}
    },
    {
        {38468,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31227,22549},
        {-1,-1,-1}
    },
    {
        {38475,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38476,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38477,-1},
        {-1},
        {-1,-1},
        {38477,-1,-1}
    },
    {
        {38477,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38477,-1},
        {-1},
        {-1,-1},
        {38477,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21678,-1},
        {-1,-1,-1}
    },
    {
        {38479,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38480,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38493,-1},
        {38493,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23789,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38488,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23072,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25996,-1},
        {-1,-1,-1}
    },
    {
        {38491,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38492,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38493,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38494,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38495,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38495,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25429,-1},
        {-1,-1,-1}
    },
    {
        {38497,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38498,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38499,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38500,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38501,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38502,-1},
        {-1},
        {-1,-1},
        {14336,23798,38575}
    },
    {
        {38506,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21857,-1},
        {-1,-1,-1}
    },
    {
        {38508,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23822,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28138,-1},
        {-1,-1,-1}
    },
    {
        {38512,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38816,-1},
        {-1,-1,-1}
    },
    {
        {38514,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38515,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38517,-1},
        {-1},
        {-1,-1},
        {38517,-1,-1}
    },
    {
        {38517,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38518,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38519,-1},
        {-1},
        {-1,-1},
        {38501,-1,-1}
    },
    {
        {38520,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26119,-1},
        {26119,-1,-1}
    },
    {
        {38522,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38525,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38492,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38492,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22585,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22576,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22576,-1,-1}
    },
    {
        {38530,-1},
        {-1},
        {-1,-1},
        {38512,-1,-1}
    },
    {
        {38530,-1},
        {-1},
        {-1,-1},
        {38512,38530,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38532,-1},
        {-1},
        {-1,-1},
        {22564,-1,-1}
    },
    {
        {38533,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38534,-1},
        {-1},
        {63964,-1},
        {-1,-1,-1}
    },
    {
        {38534,-1},
        {-1},
        {63964,-1},
        {38534,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38536,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38538,-1},
        {-1},
        {-1,-1},
        {38538,-1,-1}
    },
    {
        {38538,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38538,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38539,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26263,-1},
        {26235,-1,-1}
    },
    {
        {38541,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38542,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38543,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38579,22549},
        {22549,-1,-1}
    },
    {
        {38548,-1},
        {-1},
        {-1,-1},
        {38548,-1,-1}
    },
    {
        {38548,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38549,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38551,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38552,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38552,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38552,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38552,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38553,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38553,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38554,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38555,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38556,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38556,-1},
        {-1},
        {-1,-1},
        {38556,-1,-1}
    },
    {
        {38557,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25942,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38553,-1,-1}
    },
    {
        {38560,-1},
        {-1},
        {-1,-1},
        {38560,38577,-1}
    },
    {
        {38560,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38563,-1},
        {-1},
        {-1,-1},
        {38563,-1,-1}
    },
    {
        {38563,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38563,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38564,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38565,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38567,-1},
        {-1},
        {-1,-1},
        {38567,-1,-1}
    },
    {
        {38567,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38567,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38567,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38568,-1},
        {-1},
        {-1,-1},
        {38568,38543,-1}
    },
    {
        {38568,-1},
        {-1},
        {-1,-1},
        {38543,-1,-1}
    },
    {
        {38569,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38570,-1},
        {-1},
        {-1,-1},
        {38522,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22707,28614},
        {22707,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23341,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38575,-1},
        {-1},
        {-1,-1},
        {14336,23798,-1}
    },
    {
        {38576,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38577,-1},
        {-1},
        {-1,-1},
        {38560,-1,-1}
    },
    {
        {38577,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38578,-1},
        {-1},
        {-1,-1},
        {39469,-1,-1}
    },
    {
        {38579,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38580,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38580,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34407,-1},
        {-1,-1,-1}
    },
    {
        {38582,-1},
        {-1},
        {-1,-1},
        {38582,-1,-1}
    },
    {
        {38582,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38583,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38584,-1},
        {-1},
        {-1,-1},
        {38583,-1,-1}
    },
    {
        {38585,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38586,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38587,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38587,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38588,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38589,-1},
        {-1},
        {-1,-1},
        {38603,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24331,-1},
        {24331,-1,-1}
    },
    {
        {38592,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38593,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38601,-1},
        {-1,-1,-1}
    },
    {
        {38596,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38597,-1},
        {-1},
        {-1,-1},
        {38597,-1,-1}
    },
    {
        {38597,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38597,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38598,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38599,-1},
        {-1},
        {-1,-1},
        {38599,-1,-1}
    },
    {
        {38599,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33809,-1},
        {-1,-1,-1}
    },
    {
        {38601,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38603,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38604,-1},
        {-1},
        {-1,-1},
        {38604,-1,-1}
    },
    {
        {38604,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38605,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38606,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38619,40373},
        {38619,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34384,-1},
        {-1,-1,-1}
    },
    {
        {38609,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38610,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37648,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37648,38585},
        {-1,-1,-1}
    },
    {
        {38613,-1},
        {-1},
        {-1,-1},
        {40304,-1,-1}
    },
    {
        {38613,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38614,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {172162,-1},
        {-1,-1,-1}
    },
    {
        {38616,-1},
        {-1},
        {-1,-1},
        {38616,-1,-1}
    },
    {
        {38616,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38616,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38616,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38617,-1},
        {-1},
        {-1,-1},
        {21452,-1,-1}
    },
    {
        {38617,-1},
        {-1},
        {-1,-1},
        {21452,-1,-1}
    },
    {
        {38618,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38618,-1},
        {-1},
        {-1,-1},
        {38618,-1,-1}
    },
    {
        {38618,-1},
        {-1},
        {-1,-1},
        {38618,-1,-1}
    },
    {
        {38619,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38620,-1},
        {-1},
        {-1,-1},
        {38609,-1,-1}
    },
    {
        {38621,-1},
        {-1},
        {-1,-1},
        {38605,-1,-1}
    },
    {
        {38622,-1},
        {-1},
        {-1,-1},
        {40335,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40410,-1},
        {-1,-1,-1}
    },
    {
        {38626,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38626,-1},
        {-1},
        {-1,-1},
        {38626,-1,-1}
    },
    {
        {38627,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40445,-1},
        {40445,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38609,38620},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38598,-1},
        {38598,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38598,-1},
        {38598,-1,-1}
    },
    {
        {38632,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38632,-1},
        {-1},
        {-1,-1},
        {38632,-1,-1}
    },
    {
        {38633,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38633,-1},
        {-1},
        {-1,-1},
        {38633,-1,-1}
    },
    {
        {38634,-1},
        {-1},
        {-1,-1},
        {38634,-1,-1}
    },
    {
        {38634,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38635,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23692,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38651,-1},
        {-1,-1,-1}
    },
    {
        {38639,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38639,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38640,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38640,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28354,-1,-1}
    },
    {
        {38642,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32716,38639},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38681,-1},
        {-1,-1,-1}
    },
    {
        {38646,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38647,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38634,29982},
        {-1,-1,-1}
    },
    {
        {38649,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38649,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38651,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23445,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38633,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26177,38671},
        {-1,-1,-1}
    },
    {
        {38656,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38660,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38660,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38662,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38662,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38662,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38663,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38664,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38666,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33803,38670},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38660,22877},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38669,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38670,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38671,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27859,-1},
        {-1,-1,-1}
    },
    {
        {38673,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38512,38530},
        {38512,-1,-1}
    },
    {
        {38675,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38675,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38676,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39145,-1},
        {-1,-1,-1}
    },
    {
        {38678,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38646,-1},
        {38646,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38664,-1},
        {-1,-1,-1}
    },
    {
        {38681,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38681,-1},
        {-1},
        {-1,-1},
        {38681,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38695,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38728,38666,-1}
    },
    {
        {38684,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38686,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38676,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38646,21535},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38691,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38692,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38692,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33945,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24428,38678},
        {-1,-1,-1}
    },
    {
        {38695,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24275,-1},
        {-1,-1,-1}
    },
    {
        {38698,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38698,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38692,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30064,38715},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38710,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38710,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38715,31782},
        {-1,-1,-1}
    },
    {
        {38704,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38706,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38707,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38726,-1},
        {38726,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34170,-1},
        {-1,-1,-1}
    },
    {
        {38710,-1},
        {-1},
        {-1,-1},
        {28354,-1,-1}
    },
    {
        {38710,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38525,-1},
        {-1,-1,-1}
    },
    {
        {38712,-1},
        {-1},
        {-1,-1},
        {35207,-1,-1}
    },
    {
        {38713,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34183,-1},
        {-1,-1,-1}
    },
    {
        {38715,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38729,39228},
        {-1,-1,-1}
    },
    {
        {38717,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38718,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38695,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28635,-1},
        {28635,-1,-1}
    },
    {
        {38721,-1},
        {-1},
        {-1,-1},
        {38647,-1,-1}
    },
    {
        {38722,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38722,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38724,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38724,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36027,38726},
        {-1,-1,-1}
    },
    {
        {38726,-1},
        {-1},
        {-1,-1},
        {38726,-1,-1}
    },
    {
        {38726,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38727,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38728,-1},
        {-1},
        {-1,-1},
        {38666,-1,-1}
    },
    {
        {38729,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38729,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38722,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23453,23542},
        {23453,23542,-1}
    },
    {
        {38733,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38734,-1},
        {-1},
        {-1,-1},
        {40372,-1,-1}
    },
    {
        {38735,-1},
        {-1},
        {-1,-1},
        {40372,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38721,38647},
        {-1,-1,-1}
    },
    {
        {38738,-1},
        {-1},
        {-1,-1},
        {38738,-1,-1}
    },
    {
        {38738,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38741,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38741,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38741,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38742,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38743,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38743,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38745,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38746,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38746,-1},
        {-1},
        {-1,-1},
        {38746,-1,-1}
    },
    {
        {38747,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38747,-1},
        {-1},
        {-1,-1},
        {38747,-1,-1}
    },
    {
        {38748,-1},
        {-1},
        {-1,-1},
        {38745,-1,-1}
    },
    {
        {38748,-1},
        {-1},
        {-1,-1},
        {38745,38748,-1}
    },
    {
        {38748,-1},
        {-1},
        {-1,-1},
        {38745,38748,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22825,-1},
        {22825,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22825,-1},
        {22825,-1,-1}
    },
    {
        {38750,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38750,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24754,-1},
        {-1,-1,-1}
    },
    {
        {38752,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38752,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38752,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38752,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38753,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38753,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38754,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38754,-1,-1}
    },
    {
        {38756,-1},
        {-1},
        {-1,-1},
        {30384,-1,-1}
    },
    {
        {38756,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38758,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38760,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38761,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38762,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38763,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38763,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37356,-1},
        {-1,-1,-1}
    },
    {
        {38765,-1},
        {-1},
        {-1,-1},
        {38765,-1,-1}
    },
    {
        {38765,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38766,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38795,-1},
        {-1,-1,-1}
    },
    {
        {38769,-1},
        {-1},
        {-1,-1},
        {38765,-1,-1}
    },
    {
        {38769,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38769,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21202,34943},
        {-1,-1,-1}
    },
    {
        {38771,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38772,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38772,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32016,-1},
        {32016,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38777,-1},
        {-1},
        {-1,-1},
        {38790,-1,-1}
    },
    {
        {38777,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38778,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38780,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40727,-1,-1}
    },
    {
        {38785,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31277,-1},
        {31277,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38788,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38788,-1},
        {-1},
        {-1,-1},
        {38788,-1,-1}
    },
    {
        {38789,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38790,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38790,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33589,-1},
        {33589,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40727,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35440,-1},
        {-1,-1,-1}
    },
    {
        {38795,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38797,-1},
        {38797,-1,-1}
    },
    {
        {38797,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33393,38830},
        {-1,-1,-1}
    },
    {
        {38799,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38799,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38800,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37234,38795},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36627,38789},
        {-1,-1,-1}
    },
    {
        {38805,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38806,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38808,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38808,-1},
        {-1},
        {-1,-1},
        {38808,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38810,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38810,-1},
        {-1},
        {-1,-1},
        {38810,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29739,-1},
        {29739,-1,-1}
    },
    {
        {38812,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38808,37631},
        {-1,-1,-1}
    },
    {
        {38814,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38816,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38818,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38819,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24171,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38835,21885},
        {-1,-1,-1}
    },
    {
        {38822,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38822,-1},
        {-1,-1,-1}
    },
    {
        {38824,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38824,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38824,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38808,-1},
        {38808,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37738,-1},
        {-1,-1,-1}
    },
    {
        {38827,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38829,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38829,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38830,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38878,-1},
        {38878,-1,-1}
    },
    {
        {38833,-1},
        {-1},
        {-1,-1},
        {38876,-1,-1}
    },
    {
        {38834,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38835,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38835,-1},
        {-1},
        {-1,-1},
        {38835,-1,-1}
    },
    {
        {38836,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38795,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38842,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37860,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23657,27211},
        {-1,-1,-1}
    },
    {
        {38846,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38846,-1},
        {-1},
        {-1,-1},
        {38772,38846,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38857,-1},
        {-1,-1,-1}
    },
    {
        {38849,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38851,-1},
        {-1},
        {-1,-1},
        {38851,-1,-1}
    },
    {
        {38851,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38854,-1},
        {-1},
        {-1,-1},
        {38854,-1,-1}
    },
    {
        {38854,-1},
        {-1},
        {-1,-1},
        {38854,-1,-1}
    },
    {
        {38854,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38854,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38854,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38856,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38856,-1},
        {-1},
        {-1,-1},
        {38856,35178,-1}
    },
    {
        {38856,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38856,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38857,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38857,-1},
        {-1},
        {-1,-1},
        {38857,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38859,-1},
        {-1},
        {-1,-1},
        {38859,-1,-1}
    },
    {
        {38859,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38859,-1},
        {-1},
        {-1,-1},
        {38859,-1,-1}
    },
    {
        {38860,-1},
        {-1},
        {-1,-1},
        {38765,-1,-1}
    },
    {
        {38860,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38860,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33894,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38865,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38865,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38865,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38808,-1},
        {38808,-1,-1}
    },
    {
        {38867,-1},
        {-1},
        {-1,-1},
        {38867,-1,-1}
    },
    {
        {38867,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38867,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38868,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38868,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38872,38819},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38872,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38872,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38873,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38873,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38881,-1},
        {-1,-1,-1}
    },
    {
        {38875,-1},
        {-1},
        {-1,-1},
        {38875,-1,-1}
    },
    {
        {38875,-1},
        {-1},
        {-1,-1},
        {38875,-1,-1}
    },
    {
        {38876,-1},
        {-1},
        {-1,-1},
        {38876,-1,-1}
    },
    {
        {38876,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38876,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38876,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38834,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38834,-1,-1}
    },
    {
        {38878,-1},
        {-1},
        {-1,-1},
        {38878,-1,-1}
    },
    {
        {38878,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38878,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38878,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38881,-1},
        {-1},
        {-1,-1},
        {38881,-1,-1}
    },
    {
        {38881,-1},
        {-1},
        {-1,-1},
        {38881,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35178,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35178,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38893,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38894,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38894,-1},
        {-1},
        {-1,-1},
        {38893,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38897,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34212,-1},
        {-1,-1,-1}
    },
    {
        {38897,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38898,-1},
        {-1},
        {-1,-1},
        {40783,-1,-1}
    },
    {
        {38899,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38899,-1},
        {-1},
        {-1,-1},
        {38899,-1,-1}
    },
    {
        {38900,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38901,-1},
        {-1},
        {-1,-1},
        {38907,-1,-1}
    },
    {
        {38901,-1},
        {-1},
        {-1,-1},
        {38907,-1,-1}
    },
    {
        {38902,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38902,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37728,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37728,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38907,-1},
        {-1},
        {-1,-1},
        {38907,-1,-1}
    },
    {
        {38907,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38907,-1},
        {-1},
        {-1,-1},
        {38907,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35571,18948},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35622,-1},
        {-1,-1,-1}
    },
    {
        {38911,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38911,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38911,-1},
        {-1},
        {-1,-1},
        {38911,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38913,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38914,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38915,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38915,-1},
        {-1},
        {-1,-1},
        {38915,-1,-1}
    },
    {
        {38916,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38916,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38917,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38918,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38920,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38948,-1},
        {-1,-1,-1}
    },
    {
        {38922,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39015,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39015,-1},
        {39015,-1,-1}
    },
    {
        {38924,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38924,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38925,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38926,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38927,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38928,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38929,-1},
        {-1},
        {-1,-1},
        {38929,-1,-1}
    },
    {
        {38929,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38930,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38930,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38931,-1},
        {-1},
        {-1,-1},
        {38931,-1,-1}
    },
    {
        {38931,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38932,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30889,-1},
        {-1,-1,-1}
    },
    {
        {38934,-1},
        {-1},
        {-1,-1},
        {27886,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38935,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38936,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38938,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31310,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38925,-1},
        {-1,-1,-1}
    },
    {
        {38942,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38989,38942},
        {38989,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38945,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38945,-1},
        {-1,-1,-1}
    },
    {
        {38947,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38948,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38948,-1},
        {-1},
        {-1,-1},
        {38948,-1,-1}
    },
    {
        {38949,-1},
        {-1},
        {-1,-1},
        {38948,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36861,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38955,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38929,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38929,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39006,-1},
        {-1,-1,-1}
    },
    {
        {38955,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38956,-1},
        {-1},
        {-1,-1},
        {38960,-1,-1}
    },
    {
        {38957,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38960,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38962,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38962,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38962,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38963,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38964,-1},
        {-1},
        {-1,-1},
        {31310,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38945,-1},
        {-1,-1,-1}
    },
    {
        {38967,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38968,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38973,-1},
        {-1},
        {-1,-1},
        {38973,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38973,-1},
        {38973,-1,-1}
    },
    {
        {38971,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38972,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38973,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39661,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39661,-1},
        {39661,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26894,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38914,23450},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36084,38972},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39006,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38982,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38983,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35111,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27453,38967},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38987,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38988,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38989,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38990,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38990,-1},
        {-1},
        {-1,-1},
        {38990,-1,-1}
    },
    {
        {38991,-1},
        {-1},
        {-1,-1},
        {38996,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35554,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38994,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38995,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38996,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38997,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38998,-1},
        {-1},
        {-1,-1},
        {22239,-1,-1}
    },
    {
        {38999,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39000,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39001,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39002,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39003,-1},
        {-1},
        {-1,-1},
        {39002,-1,-1}
    },
    {
        {39003,-1},
        {-1},
        {-1,-1},
        {39002,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35611,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39006,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40410,38991},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32754,38964},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38996,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39001,22575},
        {-1,-1,-1}
    },
    {
        {39013,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39014,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39015,-1},
        {-1},
        {-1,-1},
        {39015,-1,-1}
    },
    {
        {39015,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24061,-1},
        {24061,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21140,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39019,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39020,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20762,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39002,64100},
        {-1,-1,-1}
    },
    {
        {39023,-1},
        {-1},
        {-1,-1},
        {38997,-1,-1}
    },
    {
        {39024,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39024,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39025,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39019,-1},
        {-1,-1,-1}
    },
    {
        {39027,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39027,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39027,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39028,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39028,-1},
        {-1},
        {-1,-1},
        {39028,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39080,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24426,-1},
        {-1,-1,-1}
    },
    {
        {39082,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39083,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39088,-1},
        {-1,-1,-1}
    },
    {
        {39085,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39087,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39088,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39089,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21038,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39094,-1},
        {-1,-1,-1}
    },
    {
        {39094,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39094,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39095,-1},
        {-1},
        {-1,-1},
        {39110,-1,-1}
    },
    {
        {39096,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38873,39107},
        {-1,-1,-1}
    },
    {
        {39098,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39099,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39100,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20977,39108},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39100,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39100,-1},
        {-1,-1,-1}
    },
    {
        {39103,-1},
        {-1},
        {-1,-1},
        {24070,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39106,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32747,-1},
        {-1,-1,-1}
    },
    {
        {39106,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39107,-1},
        {-1},
        {-1,-1},
        {39108,-1,-1}
    },
    {
        {39108,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39110,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39111,-1},
        {-1},
        {-1,-1},
        {39110,-1,-1}
    },
    {
        {39112,-1},
        {-1},
        {-1,-1},
        {39110,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39107,36988},
        {-1,-1,-1}
    },
    {
        {39115,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39080,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39080,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39131,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39132,-1},
        {-1},
        {-1,-1},
        {32763,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39131,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39135,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39135,-1},
        {-1},
        {-1,-1},
        {39135,-1,-1}
    },
    {
        {39136,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39137,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39138,-1},
        {-1},
        {-1,-1},
        {39138,-1,-1}
    },
    {
        {39138,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39138,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39139,-1},
        {-1},
        {-1,-1},
        {39139,-1,-1}
    },
    {
        {39139,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39164,39135},
        {39164,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39164,39135},
        {39164,-1,-1}
    },
    {
        {39141,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39256,-1,-1}
    },
    {
        {39143,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39145,-1},
        {-1},
        {-1,-1},
        {39145,-1,-1}
    },
    {
        {39145,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39146,-1},
        {-1},
        {-1,-1},
        {39146,-1,-1}
    },
    {
        {39146,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39147,-1},
        {-1},
        {-1,-1},
        {39147,-1,-1}
    },
    {
        {39147,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39178,-1},
        {39178,-1,-1}
    },
    {
        {39149,-1},
        {-1},
        {-1,-1},
        {39149,-1,-1}
    },
    {
        {39149,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39150,-1},
        {-1},
        {-1,-1},
        {39154,-1,-1}
    },
    {
        {39151,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39152,-1},
        {-1},
        {-1,-1},
        {39151,-1,-1}
    },
    {
        {39153,-1},
        {-1},
        {-1,-1},
        {39143,-1,-1}
    },
    {
        {39154,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39155,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39156,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39156,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39152,37436},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39152,37436},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39180,37455},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39253,-1},
        {39253,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39165,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39176,36018},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39214,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39214,-1},
        {-1,-1,-1}
    },
    {
        {39164,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39165,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39165,-1},
        {-1},
        {-1,-1},
        {39165,-1,-1}
    },
    {
        {39165,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39166,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39166,-1},
        {-1},
        {-1,-1},
        {39166,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39155,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39171,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39146,31252},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39146,31252},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39171,-1},
        {-1},
        {-1,-1},
        {39171,-1,-1}
    },
    {
        {39171,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39171,-1},
        {-1},
        {-1,-1},
        {39171,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39156,-1},
        {-1,-1,-1}
    },
    {
        {39173,-1},
        {-1},
        {-1,-1},
        {39173,-1,-1}
    },
    {
        {39173,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37530,39204},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39176,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39177,-1},
        {-1},
        {-1,-1},
        {39177,-1,-1}
    },
    {
        {39177,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39178,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39178,-1},
        {-1},
        {-1,-1},
        {39178,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {16645,-1},
        {-1,-1,-1}
    },
    {
        {39180,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39180,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39180,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39180,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39180,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39180,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39184,-1},
        {-1},
        {-1,-1},
        {39184,-1,-1}
    },
    {
        {39184,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39186,-1},
        {-1},
        {-1,-1},
        {39186,-1,-1}
    },
    {
        {39186,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39186,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39187,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39187,-1},
        {-1},
        {-1,-1},
        {39187,-1,-1}
    },
    {
        {39188,-1},
        {-1},
        {-1,-1},
        {39188,-1,-1}
    },
    {
        {39188,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39190,-1},
        {-1},
        {-1,-1},
        {39190,-1,-1}
    },
    {
        {39190,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39191,-1},
        {-1},
        {-1,-1},
        {39191,-1,-1}
    },
    {
        {39191,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39192,-1},
        {-1},
        {-1,-1},
        {39192,20313,-1}
    },
    {
        {39192,-1},
        {-1},
        {-1,-1},
        {20313,-1,-1}
    },
    {
        {39194,-1},
        {-1},
        {-1,-1},
        {39194,32948,-1}
    },
    {
        {39194,-1},
        {-1},
        {-1,-1},
        {32948,-1,-1}
    },
    {
        {39195,-1},
        {-1},
        {-1,-1},
        {39195,-1,-1}
    },
    {
        {39195,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39196,-1},
        {-1},
        {-1,-1},
        {39196,-1,-1}
    },
    {
        {39196,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39197,-1},
        {-1},
        {-1,-1},
        {39197,39166,-1}
    },
    {
        {39197,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39197,-1},
        {-1},
        {-1,-1},
        {39197,39166,-1}
    },
    {
        {39198,-1},
        {-1},
        {-1,-1},
        {39198,-1,-1}
    },
    {
        {39198,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39200,-1},
        {-1},
        {-1,-1},
        {39200,39173,-1}
    },
    {
        {39200,-1},
        {-1},
        {-1,-1},
        {39173,-1,-1}
    },
    {
        {39201,-1},
        {-1},
        {-1,-1},
        {39201,-1,-1}
    },
    {
        {39201,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39145,37251},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39145,37251},
        {-1,-1,-1}
    },
    {
        {39204,-1},
        {-1},
        {-1,-1},
        {39204,-1,-1}
    },
    {
        {39204,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39176,36649},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39208,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39192,39608},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39192,39608},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39192,39608},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39186,36653},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39186,36653},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39212,-1},
        {-1},
        {-1,-1},
        {39212,-1,-1}
    },
    {
        {39212,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37728,39241},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37728,39241},
        {-1,-1,-1}
    },
    {
        {39214,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39214,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39256,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39256,-1},
        {-1,-1,-1}
    },
    {
        {39217,-1},
        {-1},
        {-1,-1},
        {39217,-1,-1}
    },
    {
        {39217,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39218,-1},
        {-1},
        {-1,-1},
        {39218,-1,-1}
    },
    {
        {39218,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39219,-1},
        {-1},
        {-1,-1},
        {39219,31958,-1}
    },
    {
        {39219,-1},
        {-1},
        {-1,-1},
        {31958,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39194,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39194,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31958,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31958,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39226,-1},
        {-1},
        {-1,-1},
        {39226,-1,-1}
    },
    {
        {39226,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39227,-1},
        {-1},
        {-1,-1},
        {39227,-1,-1}
    },
    {
        {39227,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39228,-1},
        {-1},
        {-1,-1},
        {39228,-1,-1}
    },
    {
        {39228,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39229,-1},
        {-1},
        {-1,-1},
        {39229,-1,-1}
    },
    {
        {39229,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39230,-1},
        {-1},
        {-1,-1},
        {39230,-1,-1}
    },
    {
        {39230,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28338,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28338,-1},
        {-1,-1,-1}
    },
    {
        {39232,-1},
        {-1},
        {-1,-1},
        {39232,-1,-1}
    },
    {
        {39232,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39233,-1},
        {-1},
        {-1,-1},
        {39233,-1,-1}
    },
    {
        {39233,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39234,-1},
        {-1},
        {-1,-1},
        {39234,-1,-1}
    },
    {
        {39234,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39234,-1},
        {-1},
        {-1,-1},
        {39234,-1,-1}
    },
    {
        {39234,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39237,-1},
        {-1},
        {-1,-1},
        {39237,-1,-1}
    },
    {
        {39237,-1},
        {-1},
        {-1,-1},
        {39237,-1,-1}
    },
    {
        {39237,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39238,-1},
        {-1},
        {-1,-1},
        {39238,-1,-1}
    },
    {
        {39238,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39238,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39241,-1},
        {-1},
        {-1,-1},
        {39241,-1,-1}
    },
    {
        {39241,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39241,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36685,32342},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36685,32342},
        {-1,-1,-1}
    },
    {
        {39243,-1},
        {-1},
        {-1,-1},
        {39243,-1,-1}
    },
    {
        {39243,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39244,-1},
        {-1},
        {-1,-1},
        {39244,-1,-1}
    },
    {
        {39244,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39244,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39245,-1},
        {-1},
        {-1,-1},
        {33203,39245,-1}
    },
    {
        {39245,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39246,-1},
        {-1},
        {-1,-1},
        {39246,-1,-1}
    },
    {
        {39246,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39184,-1},
        {-1,-1,-1}
    },
    {
        {39248,-1},
        {-1},
        {-1,-1},
        {39248,-1,-1}
    },
    {
        {39248,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39249,-1},
        {-1},
        {-1,-1},
        {39249,-1,-1}
    },
    {
        {39249,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39249,-1},
        {-1},
        {-1,-1},
        {39249,-1,-1}
    },
    {
        {39249,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39250,-1},
        {-1},
        {-1,-1},
        {39250,-1,-1}
    },
    {
        {39250,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27222,39191},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39253,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39253,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39249,31330},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39249,31330},
        {-1,-1,-1}
    },
    {
        {39255,-1},
        {-1},
        {-1,-1},
        {39255,-1,-1}
    },
    {
        {39255,-1},
        {-1},
        {-1,-1},
        {39255,-1,-1}
    },
    {
        {39255,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39255,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39255,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39255,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39255,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39256,-1},
        {-1},
        {-1,-1},
        {39256,-1,-1}
    },
    {
        {39256,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39233,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39260,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39260,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39260,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30952,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30952,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39263,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39263,-1},
        {-1},
        {-1,-1},
        {39263,-1,-1}
    },
    {
        {39264,-1},
        {-1},
        {-1,-1},
        {39264,-1,-1}
    },
    {
        {39264,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38013,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38013,-1},
        {-1,-1,-1}
    },
    {
        {39318,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39319,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39320,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39321,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39331,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33534,-1,-1}
    },
    {
        {39326,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39326,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35901,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39331,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {64098,39333},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39333,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39334,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39333,37278},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39333,37278},
        {-1,-1,-1}
    },
    {
        {39336,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39336,-1},
        {-1,-1,-1}
    },
    {
        {39340,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39341,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39342,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39419,-1},
        {39419,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39378,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39378,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39364,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39347,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39348,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39356,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39356,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39356,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39368,-1},
        {-1,-1,-1}
    },
    {
        {39353,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37399,-1},
        {-1,-1,-1}
    },
    {
        {39356,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39356,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39357,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39359,-1},
        {-1},
        {-1,-1},
        {39522,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39799,-1},
        {-1,-1,-1}
    },
    {
        {39361,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39361,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40199,-1},
        {-1,-1,-1}
    },
    {
        {39363,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39364,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39365,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39366,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39368,-1},
        {-1},
        {-1,-1},
        {39366,-1,-1}
    },
    {
        {39369,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39384,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40219,39381},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39811,170182},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39519,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39380,-1},
        {-1,-1,-1}
    },
    {
        {39376,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39376,-1},
        {-1},
        {-1,-1},
        {39376,-1,-1}
    },
    {
        {39377,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39378,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39380,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39381,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39420,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36587,39405},
        {-1,-1,-1}
    },
    {
        {39384,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39385,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39387,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39387,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39389,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39390,-1},
        {-1},
        {-1,-1},
        {39389,-1,-1}
    },
    {
        {39391,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39470,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32629,21966},
        {-1,-1,-1}
    },
    {
        {39394,-1},
        {-1},
        {-1,-1},
        {39394,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39441,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36618,39427},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39472,-1},
        {39472,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39376,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39405,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39405,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39406,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39406,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34523,-1},
        {-1,-1,-1}
    },
    {
        {39408,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39409,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39410,-1},
        {-1},
        {-1,-1},
        {39348,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36348,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35307,-1},
        {-1,-1,-1}
    },
    {
        {39416,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39416,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39417,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39456,26390},
        {-1,-1,-1}
    },
    {
        {39419,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39420,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39423,-1},
        {-1},
        {-1,-1},
        {39423,-1,-1}
    },
    {
        {39423,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35472,-1},
        {-1,-1,-1}
    },
    {
        {39425,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39426,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39427,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39429,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39420,-1},
        {-1,-1,-1}
    },
    {
        {39394,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39419,39908},
        {-1,-1,-1}
    },
    {
        {39434,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39464,-1},
        {-1,-1,-1}
    },
    {
        {39438,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39439,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39443,39511,-1}
    },
    {
        {39441,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39442,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39443,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39456,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33136,39491},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33136,39491},
        {-1,-1,-1}
    },
    {
        {39446,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39449,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39492,-1},
        {39492,-1,-1}
    },
    {
        {39449,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39449,-1},
        {-1},
        {-1,-1},
        {39449,-1,-1}
    },
    {
        {39450,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {171593,39417},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39945,39419},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33137,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36664,39450},
        {-1,-1,-1}
    },
    {
        {39456,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33146,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39460,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39492,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39464,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25436,25628},
        {-1,-1,-1}
    },
    {
        {39467,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23882,-1},
        {-1,-1,-1}
    },
    {
        {39469,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39470,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39426,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39426,-1},
        {-1,-1,-1}
    },
    {
        {39472,-1},
        {-1},
        {-1,-1},
        {39472,-1,-1}
    },
    {
        {39472,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39473,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39416,40761},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39478,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39479,-1},
        {-1},
        {-1,-1},
        {39442,-1,-1}
    },
    {
        {39480,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39439,-1},
        {39439,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39530,36678},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24948,39470},
        {-1,-1,-1}
    },
    {
        {39486,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39488,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39488,-1},
        {-1},
        {-1,-1},
        {39488,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39490,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39491,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39492,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39492,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39493,-1},
        {-1},
        {-1,-1},
        {39366,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36437,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40010,39509},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39522,-1},
        {-1,-1,-1}
    },
    {
        {39498,-1},
        {-1},
        {-1,-1},
        {39498,-1,-1}
    },
    {
        {39498,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39500,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39501,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39502,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39502,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37907,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39506,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36461,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39509,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39511,-1},
        {-1},
        {-1,-1},
        {39443,-1,-1}
    },
    {
        {39512,-1},
        {-1},
        {-1,-1},
        {39486,-1,-1}
    },
    {
        {39512,-1},
        {-1},
        {-1,-1},
        {39512,39486,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40035,-1},
        {-1,-1,-1}
    },
    {
        {39514,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39514,-1},
        {-1},
        {-1,-1},
        {39514,-1,-1}
    },
    {
        {39515,-1},
        {-1},
        {-1,-1},
        {39365,-1,-1}
    },
    {
        {39515,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39637,-1},
        {-1,-1,-1}
    },
    {
        {39519,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39519,-1},
        {-1},
        {-1,-1},
        {39519,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37308,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32894,-1},
        {-1,-1,-1}
    },
    {
        {39522,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39472,-1},
        {39472,-1,-1}
    },
    {
        {39524,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39525,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39525,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23360,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38816,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39509,35327},
        {-1,-1,-1}
    },
    {
        {39529,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39529,-1},
        {-1},
        {-1,-1},
        {39529,-1,-1}
    },
    {
        {39530,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39531,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39592,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39597,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39599,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39600,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39599,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39600,38785},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20307,39636,-1}
    },
    {
        {39606,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39607,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39608,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39608,-1},
        {-1},
        {-1,-1},
        {39608,-1,-1}
    },
    {
        {39609,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35436,39612},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39612,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33151,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39616,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39617,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {19402,39612},
        {-1,-1,-1}
    },
    {
        {39619,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39620,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39622,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39632,39609},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33152,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33152,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30906,39620},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22043,-1},
        {-1,-1,-1}
    },
    {
        {39630,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39631,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39632,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39633,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39634,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39634,-1},
        {-1},
        {-1,-1},
        {39634,-1,-1}
    },
    {
        {39635,-1},
        {-1},
        {-1,-1},
        {39620,-1,-1}
    },
    {
        {39635,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39636,-1},
        {-1},
        {-1,-1},
        {20307,-1,-1}
    },
    {
        {39637,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39638,-1},
        {-1},
        {-1,-1},
        {39638,-1,-1}
    },
    {
        {39638,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39638,-1},
        {-1},
        {-1,-1},
        {39638,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39025,-1,-1}
    },
    {
        {39640,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39640,-1},
        {-1},
        {-1,-1},
        {39640,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39644,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23959,-1},
        {-1,-1,-1}
    },
    {
        {39646,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39647,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39648,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39648,-1},
        {-1,-1,-1}
    },
    {
        {39650,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39651,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39673,-1,-1}
    },
    {
        {39653,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39654,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39648,-1},
        {-1,-1,-1}
    },
    {
        {39657,-1},
        {-1},
        {-1,-1},
        {39714,-1,-1}
    },
    {
        {39658,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39659,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39661,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39662,-1},
        {-1},
        {-1,-1},
        {39658,-1,-1}
    },
    {
        {39663,-1},
        {-1},
        {-1,-1},
        {39653,-1,-1}
    },
    {
        {39663,-1},
        {-1},
        {-1,-1},
        {39653,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21059,-1},
        {39650,-1,-1}
    },
    {
        {39665,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39665,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39654,-1,-1}
    },
    {
        {39668,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39653,-1},
        {-1,-1,-1}
    },
    {
        {39671,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39653,-1},
        {-1,-1,-1}
    },
    {
        {39673,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39675,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39663,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21059,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39682,-1},
        {-1},
        {-1,-1},
        {39714,-1,-1}
    },
    {
        {39683,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39650,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39686,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39686,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39654,-1},
        {-1,-1,-1}
    },
    {
        {39688,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24027,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39692,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39696,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33977,39668},
        {-1,-1,-1}
    },
    {
        {39698,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39698,-1},
        {-1},
        {-1,-1},
        {39698,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31735,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39704,-1},
        {-1,-1,-1}
    },
    {
        {39702,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {13500,-1},
        {-1,-1,-1}
    },
    {
        {39704,-1},
        {-1},
        {-1,-1},
        {39704,-1,-1}
    },
    {
        {39704,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39706,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39708,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39711,-1},
        {-1,-1,-1}
    },
    {
        {39711,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39712,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39683,-1},
        {-1,-1,-1}
    },
    {
        {39714,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39715,-1},
        {-1},
        {-1,-1},
        {39715,-1,-1}
    },
    {
        {39715,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26345,39711},
        {-1,-1,-1}
    },
    {
        {39717,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38360,39725,-1}
    },
    {
        {39719,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39720,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39721,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39722,-1},
        {-1},
        {-1,-1},
        {38360,39725,-1}
    },
    {
        {39723,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39725,-1},
        {-1},
        {39725,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39725,-1},
        {-1},
        {-1,-1},
        {38360,-1,-1}
    },
    {
        {39725,-1},
        {-1},
        {-1,-1},
        {38360,-1,-1}
    },
    {
        {39725,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39726,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39726,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39726,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39726,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39726,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39727,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39729,27421},
        {39729,-1,-1}
    },
    {
        {39729,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39730,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39731,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39230,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39739,-1},
        {-1},
        {-1,-1},
        {31909,-1,-1}
    },
    {
        {39740,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39741,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39747,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39747,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39745,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39746,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39747,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39748,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39749,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39747,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39320,39747},
        {-1,-1,-1}
    },
    {
        {39755,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39757,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39757,-1},
        {-1},
        {-1,-1},
        {39757,-1,-1}
    },
    {
        {39758,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39759,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39759,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39761,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39761,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39107,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {19488,-1},
        {-1,-1,-1}
    },
    {
        {39764,-1},
        {-1},
        {-1,-1},
        {39764,-1,-1}
    },
    {
        {39764,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37276,-1,-1}
    },
    {
        {39768,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32891,-1},
        {-1,-1,-1}
    },
    {
        {39770,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39774,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39774,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33313,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39801,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39779,-1},
        {-1},
        {-1,-1},
        {40046,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39781,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39782,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33322,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39838,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39787,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39788,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39791,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39851,39356},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39794,-1},
        {-1},
        {-1,-1},
        {40056,-1,-1}
    },
    {
        {39795,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39796,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39797,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39797,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39798,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39799,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39801,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36603,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37440,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39944,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34486,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39811,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39788,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39798,-1},
        {-1,-1,-1}
    },
    {
        {39811,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39811,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39812,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39825,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39899,-1,-1}
    },
    {
        {39818,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39872,-1},
        {39872,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39822,-1},
        {-1},
        {-1,-1},
        {39822,-1,-1}
    },
    {
        {39822,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39823,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39825,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39825,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39826,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39826,-1},
        {-1},
        {-1,-1},
        {39826,-1,-1}
    },
    {
        {39827,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39822,-1},
        {-1,-1,-1}
    },
    {
        {39830,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39831,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39831,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39830,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39838,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39839,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39840,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39840,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39905,37530},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39933,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39409,-1},
        {-1,-1,-1}
    },
    {
        {39846,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39847,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39848,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39850,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39851,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39851,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39851,-1},
        {-1},
        {-1,-1},
        {39851,-1,-1}
    },
    {
        {39852,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39852,-1},
        {-1},
        {-1,-1},
        {39852,-1,-1}
    },
    {
        {39853,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39854,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39857,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39857,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39857,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39858,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39860,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39906,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39961,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39864,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39865,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39865,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39827,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39420,39918},
        {-1,-1,-1}
    },
    {
        {39870,-1},
        {-1},
        {-1,-1},
        {39896,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39872,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39873,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39873,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31308,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39858,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40001,37580},
        {-1,-1,-1}
    },
    {
        {39878,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39879,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39881,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39881,-1},
        {-1},
        {-1,-1},
        {39881,-1,-1}
    },
    {
        {39882,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39882,-1},
        {39882,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39886,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39886,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39887,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39889,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39890,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39892,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39892,-1},
        {-1},
        {-1,-1},
        {39892,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39850,39439},
        {-1,-1,-1}
    },
    {
        {39894,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39894,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39896,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39912,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39896,24760},
        {-1,-1,-1}
    },
    {
        {39899,-1},
        {-1},
        {-1,-1},
        {39899,-1,-1}
    },
    {
        {39899,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39899,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39944,-1},
        {-1,-1,-1}
    },
    {
        {39901,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27512,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39946,-1},
        {-1,-1,-1}
    },
    {
        {39905,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39905,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39906,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39907,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39908,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39909,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39911,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39912,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39912,-1},
        {-1},
        {-1,-1},
        {39912,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36650,-1},
        {-1,-1,-1}
    },
    {
        {39914,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39915,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39915,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40058,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33355,39963},
        {-1,-1,-1}
    },
    {
        {39918,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39919,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39920,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39920,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39921,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39921,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39922,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39922,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39923,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39925,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39925,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39879,-1},
        {39879,-1,-1}
    },
    {
        {39927,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39928,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39823,-1,-1}
    },
    {
        {39930,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34674,39919},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39918,-1},
        {-1,-1,-1}
    },
    {
        {39933,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39933,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39936,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39936,-1},
        {-1},
        {-1,-1},
        {39936,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39945,33146},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {171541,-1},
        {-1,-1,-1}
    },
    {
        {39940,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39940,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39881,39619},
        {-1,-1,-1}
    },
    {
        {39942,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40010,38819},
        {-1,-1,-1}
    },
    {
        {39944,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39945,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39946,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39948,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39948,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39948,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39949,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39985,37749},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39952,-1},
        {-1},
        {-1,-1},
        {39952,-1,-1}
    },
    {
        {39952,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39952,-1},
        {-1},
        {-1,-1},
        {39952,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39954,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39955,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39956,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39957,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39958,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39899,34676},
        {-1,-1,-1}
    },
    {
        {39960,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39961,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39961,-1},
        {-1},
        {-1,-1},
        {39961,-1,-1}
    },
    {
        {39962,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39963,-1},
        {-1},
        {-1,-1},
        {39982,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39965,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39796,-1},
        {39796,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39796,-1},
        {39796,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39479,-1},
        {-1,-1,-1}
    },
    {
        {39969,-1},
        {-1},
        {-1,-1},
        {39969,-1,-1}
    },
    {
        {39969,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39970,-1},
        {-1},
        {-1,-1},
        {39970,-1,-1}
    },
    {
        {39970,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39971,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39972,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39973,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39975,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39975,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39977,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39978,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40005,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39921,-1},
        {-1,-1,-1}
    },
    {
        {39981,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39981,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39982,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39983,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39983,-1},
        {-1},
        {-1,-1},
        {39983,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39984,-1},
        {-1},
        {-1,-1},
        {39984,-1,-1}
    },
    {
        {39984,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39985,-1},
        {-1},
        {-1,-1},
        {39985,-1,-1}
    },
    {
        {39985,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39986,-1},
        {-1},
        {-1,-1},
        {40711,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21202,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40692,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39990,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39991,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39493,-1},
        {-1,-1,-1}
    },
    {
        {39993,-1},
        {-1},
        {-1,-1},
        {39993,-1,-1}
    },
    {
        {39993,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39994,-1},
        {-1},
        {-1,-1},
        {39994,-1,-1}
    },
    {
        {39994,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39995,-1},
        {-1},
        {-1,-1},
        {39995,-1,-1}
    },
    {
        {39995,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39998,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39933,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26305,39986},
        {-1,-1,-1}
    },
    {
        {40001,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40001,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40001,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37848,39971},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40005,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40006,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40007,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40008,-1},
        {-1},
        {-1,-1},
        {40008,-1,-1}
    },
    {
        {40008,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40009,-1},
        {-1},
        {-1,-1},
        {40712,-1,-1}
    },
    {
        {40010,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39921,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {171716,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39957,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40015,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40015,-1},
        {-1},
        {-1,-1},
        {40015,-1,-1}
    },
    {
        {40016,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39945,-1},
        {39945,-1,-1}
    },
    {
        {40018,-1},
        {-1},
        {-1,-1},
        {40018,-1,-1}
    },
    {
        {40018,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40018,-1},
        {-1},
        {-1,-1},
        {40018,-1,-1}
    },
    {
        {40019,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40020,-1},
        {-1},
        {-1,-1},
        {40019,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40041,30971},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40023,-1},
        {-1},
        {-1,-1},
        {40023,-1,-1}
    },
    {
        {40023,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40023,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40024,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33495,-1},
        {33495,-1,-1}
    },
    {
        {40026,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40027,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40028,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40029,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40029,-1},
        {-1},
        {-1,-1},
        {40029,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40031,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40032,-1},
        {-1},
        {-1,-1},
        {33214,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39994,-1},
        {33226,-1,-1}
    },
    {
        {40035,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39956,-1},
        {-1,-1,-1}
    },
    {
        {40037,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34821,-1},
        {-1,-1,-1}
    },
    {
        {40039,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40041,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40042,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40043,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40041,33233},
        {-1,-1,-1}
    },
    {
        {40045,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40046,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40048,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34851,40005},
        {-1,-1,-1}
    },
    {
        {40050,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39944,31019},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38856,40037},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38856,40037},
        {-1,-1,-1}
    },
    {
        {40053,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40054,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40054,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40054,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40055,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40056,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39529,-1},
        {-1,-1,-1}
    },
    {
        {40058,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40165,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40166,-1},
        {-1},
        {-1,-1},
        {40166,-1,-1}
    },
    {
        {40166,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40167,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21235,-1},
        {-1,-1,-1}
    },
    {
        {40169,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40176,-1},
        {-1,-1,-1}
    },
    {
        {40171,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40172,-1},
        {-1},
        {-1,-1},
        {40167,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38557,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40388,40335},
        {40388,40335,-1}
    },
    {
        {40179,-1},
        {-1},
        {40179,40300},
        {40179,-1,-1}
    },
    {
        {40176,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40200,-1},
        {-1,-1,-1}
    },
    {
        {40178,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40179,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40180,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40199,-1},
        {40199,-1,-1}
    },
    {
        {40182,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40171,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40194,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40210,38930},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40455,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40455,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40232,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27170,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40194,-1},
        {-1,-1,-1}
    },
    {
        {40194,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40195,-1},
        {-1},
        {-1,-1},
        {40194,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21305,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40259,-1},
        {-1,-1,-1}
    },
    {
        {40198,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40199,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40199,-1},
        {-1},
        {-1,-1},
        {40199,-1,-1}
    },
    {
        {40200,-1},
        {-1},
        {-1,-1},
        {38593,-1,-1}
    },
    {
        {40200,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40201,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40201,-1},
        {-1},
        {-1,-1},
        {40201,-1,-1}
    },
    {
        {40201,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {19652,-1},
        {-1,-1,-1}
    },
    {
        {40203,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40179,-1},
        {40179,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40182,-1},
        {-1,-1,-1}
    },
    {
        {40206,-1},
        {-1},
        {-1,-1},
        {40407,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {172058,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40209,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40210,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {172079,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40215,-1},
        {-1,-1,-1}
    },
    {
        {40213,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40223,-1},
        {-1,-1,-1}
    },
    {
        {40215,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40216,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38601,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40166,-1},
        {-1,-1,-1}
    },
    {
        {40219,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40221,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40222,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40223,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34507,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38606,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24188,-1},
        {-1,-1,-1}
    },
    {
        {40227,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39831,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40234,-1},
        {-1,-1,-1}
    },
    {
        {40230,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40234,-1},
        {-1,-1,-1}
    },
    {
        {40232,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40194,-1},
        {-1,-1,-1}
    },
    {
        {40234,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40234,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40235,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40236,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40242,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40346,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40346,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40257,-1},
        {-1,-1,-1}
    },
    {
        {40242,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26783,-1},
        {26783,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40288,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27170,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32724,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40251,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40442,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40254,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40255,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40287,-1},
        {-1,-1,-1}
    },
    {
        {40257,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40258,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40259,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40259,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40260,-1},
        {-1},
        {-1,-1},
        {40223,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40262,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40263,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40263,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40264,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40264,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40478,-1},
        {-1,-1,-1}
    },
    {
        {40266,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40199,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40195,39420},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40475,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40335,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40292,-1},
        {-1,-1,-1}
    },
    {
        {40272,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40273,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40274,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40318,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40345,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40281,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40198,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38968,-1},
        {-1,-1,-1}
    },
    {
        {40284,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40285,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40286,-1},
        {-1},
        {-1,-1},
        {40285,-1,-1}
    },
    {
        {40287,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40288,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40288,-1},
        {-1},
        {-1,-1},
        {40288,-1,-1}
    },
    {
        {40289,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40290,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40291,-1},
        {-1},
        {-1,-1},
        {40284,-1,-1}
    },
    {
        {40292,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40216,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40288,-1},
        {-1,-1,-1}
    },
    {
        {40297,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40299,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40300,-1},
        {-1},
        {-1,-1},
        {40300,-1,-1}
    },
    {
        {40300,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40242,29124},
        {-1,-1,-1}
    },
    {
        {40303,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40304,-1},
        {-1},
        {-1,-1},
        {38613,-1,-1}
    },
    {
        {40304,-1},
        {-1},
        {-1,-1},
        {40304,38613,-1}
    },
    {
        {-1,-1},
        {-1},
        {40260,22518},
        {-1,-1,-1}
    },
    {
        {40306,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40307,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40221,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40291,143798},
        {-1,-1,-1}
    },
    {
        {40310,-1},
        {-1},
        {-1,-1},
        {40201,-1,-1}
    },
    {
        {40311,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27449,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40477,-1},
        {-1,-1,-1}
    },
    {
        {40314,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40316,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40316,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21461,-1},
        {-1,-1,-1}
    },
    {
        {40318,-1},
        {-1},
        {-1,-1},
        {40356,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39644,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39439,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40329,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40385,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40324,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40324,-1},
        {-1},
        {-1,-1},
        {40324,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30014,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40327,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40474,-1},
        {-1,-1,-1}
    },
    {
        {40329,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40330,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40307,-1},
        {-1,-1,-1}
    },
    {
        {40333,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40335,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40440,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40440,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {18948,-1,-1}
    },
    {
        {40342,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40345,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40346,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40242,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {171982,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40307,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40316,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37118,40407},
        {-1,-1,-1}
    },
    {
        {40353,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40335,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25153,-1},
        {-1,-1,-1}
    },
    {
        {40356,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40307,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38601,-1},
        {-1,-1,-1}
    },
    {
        {40361,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40363,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40364,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40203,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40367,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40273,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40370,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40370,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40370,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39972,-1},
        {-1,-1,-1}
    },
    {
        {40372,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40373,-1},
        {-1},
        {-1,-1},
        {38619,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40330,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40288,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40288,-1},
        {-1,-1,-1}
    },
    {
        {40376,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40376,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40377,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40377,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40378,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40379,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40380,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40381,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32752,172162},
        {-1,-1,-1}
    },
    {
        {40383,-1},
        {-1},
        {-1,-1},
        {40383,-1,-1}
    },
    {
        {40383,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40383,-1},
        {-1},
        {-1,-1},
        {40383,-1,-1}
    },
    {
        {40383,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40384,-1},
        {-1},
        {-1,-1},
        {40384,40383,-1}
    },
    {
        {40384,-1},
        {-1},
        {-1,-1},
        {40383,-1,-1}
    },
    {
        {40384,-1},
        {-1},
        {-1,-1},
        {40383,-1,-1}
    },
    {
        {40384,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40385,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40385,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40385,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40386,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40387,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40388,-1},
        {-1},
        {-1,-1},
        {40335,-1,-1}
    },
    {
        {40388,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {172432,-1},
        {-1,-1,-1}
    },
    {
        {40390,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40391,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40393,-1},
        {-1,-1,-1}
    },
    {
        {40393,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26783,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40399,-1},
        {-1},
        {-1,-1},
        {40390,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26216,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40215,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29966,40303},
        {-1,-1,-1}
    },
    {
        {40403,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40406,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40406,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40407,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40407,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40409,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40410,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37144,40330},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25976,40474},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40419,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40415,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40416,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40422,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40441,21413},
        {-1,-1,-1}
    },
    {
        {40419,-1},
        {-1},
        {-1,-1},
        {40419,-1,-1}
    },
    {
        {40419,-1},
        {-1},
        {-1,-1},
        {40419,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40346,22213},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40422,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40423,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40356,33775},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40367,-1},
        {40367,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40643,40399},
        {-1,-1,-1}
    },
    {
        {40429,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40431,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29141,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32762,40299},
        {-1,-1,-1}
    },
    {
        {40434,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40436,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23648,40307},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40455,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36466,-1},
        {-1,-1,-1}
    },
    {
        {40440,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40441,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40442,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40329,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40445,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40458,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40353,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39512,-1},
        {-1,-1,-1}
    },
    {
        {40450,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29344,172058},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25985,40419},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40469,34398},
        {-1,-1,-1}
    },
    {
        {40455,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {172432,37172},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26954,-1},
        {-1,-1,-1}
    },
    {
        {40458,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23527,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40461,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40367,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33945,40475},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40235,30078},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40469,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35723,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35723,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40473,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40474,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40475,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40475,-1},
        {-1},
        {-1,-1},
        {40475,-1,-1}
    },
    {
        {40476,-1},
        {-1},
        {-1,-1},
        {40221,-1,-1}
    },
    {
        {40477,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40478,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40565,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40568,40802},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40802,-1},
        {-1,-1,-1}
    },
    {
        {40568,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40569,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40571,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40572,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40573,-1},
        {-1},
        {-1,-1},
        {22633,-1,-1}
    },
    {
        {40575,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40576,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40576,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40577,-1},
        {-1},
        {-1,-1},
        {40612,-1,-1}
    },
    {
        {40577,-1},
        {-1},
        {-1,-1},
        {40612,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40612,-1,-1}
    },
    {
        {40581,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40612,31895},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40584,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31759,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40587,-1},
        {-1,-1,-1}
    },
    {
        {40587,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40588,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40588,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40607,-1},
        {-1,-1,-1}
    },
    {
        {40593,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40594,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40595,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21646,-1},
        {-1,-1,-1}
    },
    {
        {40597,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40599,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40599,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31668,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40603,38686},
        {-1,-1,-1}
    },
    {
        {40603,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40587,-1},
        {-1,-1,-1}
    },
    {
        {40605,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40606,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40606,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40607,-1},
        {-1},
        {-1,-1},
        {40607,-1,-1}
    },
    {
        {40607,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40607,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38717,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32666,-1},
        {32666,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40607,-1},
        {-1,-1,-1}
    },
    {
        {40612,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40613,-1},
        {-1},
        {-1,-1},
        {40614,-1,-1}
    },
    {
        {40614,-1},
        {-1},
        {-1,-1},
        {40613,40614,-1}
    },
    {
        {40614,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20094,-1},
        {-1,-1,-1}
    },
    {
        {40616,-1},
        {-1},
        {-1,-1},
        {40616,-1,-1}
    },
    {
        {40616,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40617,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40617,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40618,-1},
        {-1},
        {-1,-1},
        {40618,-1,-1}
    },
    {
        {40618,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40618,-1},
        {-1},
        {40634,40618},
        {40618,-1,-1}
    },
    {
        {40620,-1},
        {-1},
        {-1,-1},
        {40617,-1,-1}
    },
    {
        {40620,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40621,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40621,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40621,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40623,-1},
        {-1},
        {-1,-1},
        {40628,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40617,-1},
        {40617,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35242,40629},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40628,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40629,-1},
        {-1},
        {-1,-1},
        {40618,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35914,36898},
        {-1,-1,-1}
    },
    {
        {40632,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40633,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40634,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40635,-1},
        {-1},
        {-1,-1},
        {40635,-1,-1}
    },
    {
        {40635,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40636,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40636,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40637,-1},
        {-1},
        {-1,-1},
        {40636,-1,-1}
    },
    {
        {40637,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40638,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40638,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40639,-1},
        {-1},
        {-1,-1},
        {40639,-1,-1}
    },
    {
        {40639,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25515,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34145,31964},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34145,31964},
        {-1,-1,-1}
    },
    {
        {40643,-1},
        {-1},
        {-1,-1},
        {40644,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40644,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40670,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40652,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40652,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40653,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40654,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40654,-1},
        {-1},
        {-1,-1},
        {40654,-1,-1}
    },
    {
        {40655,-1},
        {-1},
        {-1,-1},
        {31896,-1,-1}
    },
    {
        {40656,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40657,-1},
        {-1},
        {-1,-1},
        {40658,-1,-1}
    },
    {
        {40658,-1},
        {-1},
        {-1,-1},
        {40658,-1,-1}
    },
    {
        {40658,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24331,-1},
        {-1,-1,-1}
    },
    {
        {40660,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40660,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {64066,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40668,-1},
        {-1,-1,-1}
    },
    {
        {40664,-1},
        {-1},
        {-1,-1},
        {40665,-1,-1}
    },
    {
        {40664,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40665,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37219,40668},
        {-1,-1,-1}
    },
    {
        {40667,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40667,-1},
        {-1},
        {-1,-1},
        {40667,-1,-1}
    },
    {
        {40668,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40668,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40669,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40669,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40670,-1},
        {-1},
        {-1,-1},
        {28857,-1,-1}
    },
    {
        {40670,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40671,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40672,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40672,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39423,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40692,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40677,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40677,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30073,40671},
        {-1,-1,-1}
    },
    {
        {40679,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40680,-1},
        {-1},
        {-1,-1},
        {20826,-1,-1}
    },
    {
        {40680,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40686,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40687,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40687,-1},
        {-1},
        {-1,-1},
        {40687,-1,-1}
    },
    {
        {40687,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40667,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40406,40679},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40406,40679},
        {-1,-1,-1}
    },
    {
        {40692,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40692,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40694,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40695,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40695,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40697,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40697,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31881,40699},
        {-1,-1,-1}
    },
    {
        {40699,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40699,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40700,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40700,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40701,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40701,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40701,-1},
        {40701,-1,-1}
    },
    {
        {40703,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40703,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40707,-1},
        {-1,-1,-1}
    },
    {
        {40706,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40706,-1},
        {-1},
        {-1,-1},
        {40706,-1,-1}
    },
    {
        {40707,-1},
        {-1},
        {-1,-1},
        {19831,-1,-1}
    },
    {
        {40707,-1},
        {-1},
        {-1,-1},
        {34521,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34523,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34584,-1},
        {34584,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34584,-1},
        {34584,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {19831,-1},
        {-1,-1,-1}
    },
    {
        {40711,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40711,-1},
        {-1},
        {-1,-1},
        {40711,-1,-1}
    },
    {
        {40711,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40712,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40712,-1},
        {-1},
        {-1,-1},
        {40712,-1,-1}
    },
    {
        {40712,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40712,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40712,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36484,-1},
        {-1,-1,-1}
    },
    {
        {40718,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40720,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40720,-1},
        {-1},
        {-1,-1},
        {40720,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40718,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40723,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40725,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40725,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40727,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28149,-1},
        {-1,-1,-1}
    },
    {
        {40729,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28149,40727},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38326,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40736,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40737,-1},
        {-1},
        {-1,-1},
        {40736,-1,-1}
    },
    {
        {40737,-1},
        {-1},
        {-1,-1},
        {40736,40737,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40664,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39356,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39747,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39747,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35970,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39389,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39831,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39831,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40769,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40769,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40748,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40748,-1},
        {-1},
        {-1,-1},
        {40748,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33351,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33351,-1},
        {-1,-1,-1}
    },
    {
        {40751,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40748,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40761,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25580,40759},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40759,-1},
        {-1,-1,-1}
    },
    {
        {40759,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37772,40759},
        {-1,-1,-1}
    },
    {
        {40761,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40761,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40762,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40762,-1},
        {-1},
        {-1,-1},
        {40762,-1,-1}
    },
    {
        {40763,-1},
        {-1},
        {-1,-1},
        {40763,-1,-1}
    },
    {
        {40763,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40765,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40766,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40766,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29966,-1},
        {-1,-1,-1}
    },
    {
        {40769,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21775,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40773,-1},
        {-1},
        {-1,-1},
        {21957,-1,-1}
    },
    {
        {40774,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40774,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40774,-1},
        {-1,-1,-1}
    },
    {
        {40778,-1},
        {-1},
        {-1,-1},
        {25993,40778,-1}
    },
    {
        {40778,-1},
        {-1},
        {-1,-1},
        {25993,-1,-1}
    },
    {
        {40779,-1},
        {-1},
        {-1,-1},
        {25998,-1,-1}
    },
    {
        {40779,-1},
        {-1},
        {-1,-1},
        {40779,-1,-1}
    },
    {
        {40779,-1},
        {-1},
        {-1,-1},
        {40779,-1,-1}
    },
    {
        {40779,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33229,40782},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25993,36039},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25993,36039},
        {-1,-1,-1}
    },
    {
        {40782,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40782,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40783,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40783,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25993,40778},
        {25993,40778,-1}
    },
    {
        {40786,-1},
        {-1},
        {-1,-1},
        {27503,-1,-1}
    },
    {
        {40787,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40788,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40789,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40791,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40792,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40803,38756},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40816,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22124,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21616,-1},
        {-1,-1,-1}
    },
    {
        {40797,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40799,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40800,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40800,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40801,-1},
        {-1},
        {-1,-1},
        {40802,-1,-1}
    },
    {
        {40802,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40803,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40807,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40806,32079},
        {-1,-1,-1}
    },
    {
        {40806,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40807,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40807,-1},
        {-1},
        {-1,-1},
        {40807,-1,-1}
    },
    {
        {40807,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40809,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40810,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40812,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40813,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40816,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40818,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40791,-1},
        {-1,-1,-1}
    },
    {
        {40821,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40822,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40822,-1},
        {-1},
        {-1,-1},
        {40822,-1,-1}
    },
    {
        {40823,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40816,32266},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24046,40807},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40799,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40813,-1},
        {40813,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29563,40807},
        {-1,-1,-1}
    },
    {
        {40845,-1},
        {-1},
        {-1,-1},
        {31452,40845,-1}
    },
    {
        {40845,-1},
        {-1},
        {-1,-1},
        {31452,-1,-1}
    },
    {
        {40845,-1},
        {-1},
        {-1,-1},
        {31452,40845,-1}
    },
    {
        {40845,-1},
        {-1},
        {-1,-1},
        {40845,-1,-1}
    },
    {
        {40845,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40848,-1},
        {40848,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40848,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40848,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31452,40845,-1}
    },
    {
        {-1,-1},
        {-1},
        {26407,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26407,-1},
        {-1,-1,-1}
    },
    {
        {40852,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40852,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40853,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40853,-1},
        {-1},
        {-1,-1},
        {40853,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40855,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35723,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35723,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40860,-1},
        {-1},
        {-1,-1},
        {20096,40860,-1}
    },
    {
        {40860,-1},
        {-1},
        {-1,-1},
        {20096,-1,-1}
    },
    {
        {40860,-1},
        {-1},
        {-1,-1},
        {20096,-1,-1}
    },
    {
        {40860,-1},
        {-1},
        {-1,-1},
        {20096,40860,-1}
    },
    {
        {40860,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40861,-1},
        {-1},
        {-1,-1},
        {31179,-1,-1}
    },
    {
        {40861,-1},
        {-1},
        {-1,-1},
        {31179,-1,-1}
    },
    {
        {40861,-1},
        {-1},
        {-1,-1},
        {31179,40861,-1}
    },
    {
        {40861,-1},
        {-1},
        {40861,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40864,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40866,-1},
        {-1},
        {-1,-1},
        {21644,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35559,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37628,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37628,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26753,-1,-1}
    },
    {
        {31059,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {63773,-1},
        {-1},
        {-1,-1},
        {27396,-1,-1}
    },
    {
        {63784,-1},
        {-1},
        {-1,-1},
        {24266,-1,-1}
    },
    {
        {63785,-1},
        {-1},
        {-1,-1},
        {26391,-1,-1}
    },
    {
        {63798,-1},
        {-1},
        {-1,-1},
        {34396,-1,-1}
    },
    {
        {63856,-1},
        {-1},
        {-1,-1},
        {27578,-1,-1}
    },
    {
        {63952,-1},
        {-1},
        {-1,-1},
        {39006,-1,-1}
    },
    {
        {63964,-1},
        {-1},
        {-1,-1},
        {38534,-1,-1}
    },
    {
        {38534,-1},
        {-1},
        {63964,-1},
        {38534,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38617,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38617,-1},
        {-1,-1,-1}
    },
    {
        {64015,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {64016,-1},
        {-1},
        {-1,-1},
        {22618,-1,-1}
    },
    {
        {64016,-1},
        {-1},
        {-1,-1},
        {64016,22618,-1}
    },
    {
        {64017,-1},
        {-1},
        {-1,-1},
        {23822,-1,-1}
    },
    {
        {64017,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26228,-1},
        {-1},
        {-1,-1},
        {26228,-1,-1}
    },
    {
        {64019,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {64019,-1},
        {-1},
        {-1,-1},
        {64019,-1,-1}
    },
    {
        {64020,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {64021,-1},
        {-1},
        {-1,-1},
        {29081,-1,-1}
    },
    {
        {64021,-1},
        {-1},
        {-1,-1},
        {29081,-1,-1}
    },
    {
        {64021,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {64022,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30410,-1},
        {-1},
        {-1,-1},
        {30410,-1,-1}
    },
    {
        {31036,-1},
        {-1},
        {31036,-1},
        {31036,-1,-1}
    },
    {
        {64025,-1},
        {-1},
        {-1,-1},
        {31070,-1,-1}
    },
    {
        {64026,-1},
        {-1},
        {-1,-1},
        {31077,-1,-1}
    },
    {
        {64027,-1},
        {-1},
        {-1,-1},
        {31119,-1,-1}
    },
    {
        {38742,-1},
        {-1},
        {38742,-1},
        {38742,-1,-1}
    },
    {
        {31934,-1},
        {-1},
        {-1,-1},
        {31934,-1,-1}
    },
    {
        {32701,-1},
        {-1},
        {-1,-1},
        {32701,-1,-1}
    },
    {
        {64031,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {64031,-1},
        {-1},
        {-1,-1},
        {33224,33240,64031}
    },
    {
        {64031,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {64032,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {64032,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {64032,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {64032,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {64032,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {64033,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {64034,-1},
        {-1},
        {-1,-1},
        {35576,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36211,-1},
        {36211,-1,-1}
    },
    {
        {64036,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {64036,-1},
        {-1},
        {-1,-1},
        {36820,-1,-1}
    },
    {
        {64036,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {64036,-1},
        {-1},
        {-1,-1},
        {36820,-1,-1}
    },
    {
        {64036,-1},
        {-1},
        {-1,-1},
        {36820,-1,-1}
    },
    {
        {64036,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36920,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {64038,-1},
        {-1},
        {-1,-1},
        {37117,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38557,25623},
        {-1,-1,-1}
    },
    {
        {39151,-1},
        {-1},
        {-1,-1},
        {39151,-1,-1}
    },
    {
        {39164,-1},
        {-1},
        {-1,-1},
        {39164,-1,-1}
    },
    {
        {39208,-1},
        {-1},
        {-1,-1},
        {39208,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40372,-1,-1}
    },
    {
        {64048,-1},
        {-1},
        {-1,-1},
        {20398,-1,-1}
    },
    {
        {64049,-1},
        {-1},
        {-1,-1},
        {20711,-1,-1}
    },
    {
        {64050,-1},
        {-1},
        {-1,-1},
        {20813,-1,-1}
    },
    {
        {64051,-1},
        {-1},
        {-1,-1},
        {21193,-1,-1}
    },
    {
        {64052,-1},
        {-1},
        {-1,-1},
        {21220,-1,-1}
    },
    {
        {64053,-1},
        {-1},
        {-1,-1},
        {21329,-1,-1}
    },
    {
        {64054,-1},
        {-1},
        {-1,-1},
        {21917,-1,-1}
    },
    {
        {64055,-1},
        {-1},
        {-1,-1},
        {22022,-1,-1}
    },
    {
        {64056,-1},
        {-1},
        {-1,-1},
        {22120,-1,-1}
    },
    {
        {64057,-1},
        {-1},
        {-1,-1},
        {22592,-1,-1}
    },
    {
        {64058,-1},
        {-1},
        {-1,-1},
        {22696,-1,-1}
    },
    {
        {64059,-1},
        {-1},
        {-1,-1},
        {23652,-1,-1}
    },
    {
        {64060,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {64061,-1},
        {-1},
        {-1,-1},
        {24724,-1,-1}
    },
    {
        {64062,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {64063,-1},
        {-1},
        {-1,-1},
        {24974,-1,-1}
    },
    {
        {64064,-1},
        {-1},
        {-1,-1},
        {25074,-1,-1}
    },
    {
        {64064,-1},
        {-1},
        {-1,-1},
        {25074,-1,-1}
    },
    {
        {64065,-1},
        {-1},
        {-1,-1},
        {25935,-1,-1}
    },
    {
        {64066,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {64067,-1},
        {-1},
        {-1,-1},
        {26257,-1,-1}
    },
    {
        {64068,-1},
        {-1},
        {-1,-1},
        {26757,-1,-1}
    },
    {
        {64069,-1},
        {-1},
        {-1,-1},
        {28023,-1,-1}
    },
    {
        {64070,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {64071,-1},
        {-1},
        {-1,-1},
        {28450,-1,-1}
    },
    {
        {64072,-1},
        {-1},
        {-1,-1},
        {29038,-1,-1}
    },
    {
        {64073,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {64074,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {64075,-1},
        {-1},
        {-1,-1},
        {30865,-1,-1}
    },
    {
        {64074,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {64076,-1},
        {-1},
        {-1,-1},
        {31038,-1,-1}
    },
    {
        {64077,-1},
        {-1},
        {-1,-1},
        {31049,-1,-1}
    },
    {
        {64078,-1},
        {-1},
        {-1,-1},
        {31048,-1,-1}
    },
    {
        {64079,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {64080,-1},
        {-1},
        {-1,-1},
        {31062,-1,-1}
    },
    {
        {64081,-1},
        {-1},
        {-1,-1},
        {31069,-1,-1}
    },
    {
        {64082,-1},
        {-1},
        {-1,-1},
        {31117,-1,-1}
    },
    {
        {64083,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {64084,-1},
        {-1},
        {-1,-1},
        {31296,-1,-1}
    },
    {
        {64085,-1},
        {-1},
        {-1,-1},
        {31361,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {64086,-1},
        {-1},
        {-1,-1},
        {31680,-1,-1}
    },
    {
        {64087,-1},
        {-1},
        {-1,-1},
        {32244,-1,-1}
    },
    {
        {64088,-1},
        {-1},
        {-1,-1},
        {32265,-1,-1}
    },
    {
        {64089,-1},
        {-1},
        {-1,-1},
        {32321,-1,-1}
    },
    {
        {64090,-1},
        {-1},
        {-1,-1},
        {32626,-1,-1}
    },
    {
        {64091,-1},
        {-1},
        {-1,-1},
        {32773,-1,-1}
    },
    {
        {64092,-1},
        {-1},
        {-1,-1},
        {33261,-1,-1}
    },
    {
        {64093,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {64094,-1},
        {-1},
        {-1,-1},
        {33400,-1,-1}
    },
    {
        {64095,-1},
        {-1},
        {-1,-1},
        {33879,-1,-1}
    },
    {
        {64096,-1},
        {-1},
        {-1,-1},
        {35088,-1,-1}
    },
    {
        {64097,-1},
        {-1},
        {-1,-1},
        {35222,-1,-1}
    },
    {
        {64098,-1},
        {-1},
        {-1,-1},
        {35585,-1,-1}
    },
    {
        {64099,-1},
        {-1},
        {-1,-1},
        {35641,-1,-1}
    },
    {
        {64100,-1},
        {-1},
        {-1,-1},
        {36051,-1,-1}
    },
    {
        {64100,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {64101,-1},
        {-1},
        {-1,-1},
        {36104,-1,-1}
    },
    {
        {64102,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {64103,-1},
        {-1},
        {-1,-1},
        {36920,-1,-1}
    },
    {
        {64104,-1},
        {-1},
        {-1,-1},
        {38627,-1,-1}
    },
    {
        {64105,-1},
        {-1},
        {-1,-1},
        {38911,-1,-1}
    },
    {
        {64105,-1},
        {-1},
        {-1,-1},
        {38911,-1,-1}
    },
    {
        {64106,-1},
        {-1},
        {-1,-1},
        {38971,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24693,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29081,-1,-1}
    },
    {
        {33304,-1},
        {-1},
        {-1,-1},
        {39208,33304,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21621,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {19971,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {19988,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {19988,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {19992,-1,-1}
    },
    {
        {131083,-1},
        {-1},
        {-1,-1},
        {19976,-1,-1}
    },
    {
        {131083,-1},
        {-1},
        {-1,-1},
        {19976,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {19991,-1},
        {-1},
        {-1,-1},
        {19990,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24179,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23378,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22825,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40599,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {19981,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31637,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22825,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27712,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24179,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31637,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40703,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21531,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22780,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22893,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21340,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22235,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20013,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20013,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20013,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21460,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20054,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20107,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31357,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20096,40860},
        {20096,40860,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20024,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22863,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20025,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32066,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24038,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {131209,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33836,19975,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20083,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30690,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24351,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20047,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20046,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37318,-1,-1}
    },
    {
        {131234,-1},
        {-1},
        {-1,-1},
        {27966,-1,-1}
    },
    {
        {131234,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {131236,-1},
        {-1},
        {-1,-1},
        {22534,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31383,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27663,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21380,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36784,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20054,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26839,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32317,32207,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29992,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28879,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24184,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22402,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24605,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30031,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22351,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25163,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24840,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32945,21431,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20009,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20250,26371,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38560,38577,20058}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20063,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24030,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31036,31150,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20063,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {19985,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33618,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25369,25831,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39131,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22987,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20633,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25130,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39319,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20081,20098,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24605,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20096,40860,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20094,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37168,38563,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20094,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20094,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20035,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22402,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24030,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24187,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20035,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21608,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20105,29229,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20104,35947,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20104,35947,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {19978,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {19979,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20116,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20120,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20120,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20120,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32020,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24658,24646,-1}
    },
    {
        {-1,-1},
        {-1},
        {26839,-1},
        {26839,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26748,26839},
        {26839,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20120,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32769,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32854,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25993,40778,20125}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20127,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20134,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26411,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20811,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28921,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20141,-1},
        {-1},
        {-1,-1},
        {20141,-1,-1}
    },
    {
        {20142,-1},
        {-1},
        {-1,-1},
        {20142,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20056,20055,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20056,20055,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20056,20055,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20139,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20811,-1,-1}
    },
    {
        {20147,-1},
        {-1},
        {-1,-1},
        {20147,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23408,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32850,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21213,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23495,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24408,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34928,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22885,22887,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26820,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22780,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24248,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34928,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35140,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {131490,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23521,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {19969,20171,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20045,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20168,27726},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39640,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38512,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20316,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39719,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24561,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64048,20398,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26045,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23713,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20491,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21083,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30002,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20315,20175,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26045,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21083,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20238,-1},
        {20238,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21629,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20241,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20215,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20241,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30326,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20354,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20445,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23462,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22809,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34398,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20633,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20360,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20745,20537,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20280,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20809,-1},
        {20809,-1,-1}
    },
    {
        {131603,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24616,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24441,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20316,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20225,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20853,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20344,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36767,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30305,30196,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20478,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20478,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25912,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38918,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20234,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23472,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24441,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31452,40845,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20853,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20351,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20351,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20372,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38495,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32996,-1},
        {32996,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22809,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21163,21031,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30428,30423,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20518,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24481,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39135,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31452,40845,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20493,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20363,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20006,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20686,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20378,24455,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22557,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20505,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20415,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20354,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36767,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38512,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20445,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30589,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39135,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21629,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29237,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33065,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32023,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20506,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23462,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20505,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {20534},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20493,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22781,23551,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30589,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20499,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37225,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20378,24455,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23453,23542,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20374,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20778,-1},
        {20778,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20778,-1},
        {20778,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24495,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36865,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20973,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {131883,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {131883,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20791,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35018,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33457,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22519,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38495,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34928,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32681,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20514,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20445,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20632,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32091,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21487,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20524,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {131953,-1},
        {-1},
        {-1,-1},
        {38598,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21964,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24838,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20754,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22799,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20756,-1,-1}
    },
    {
        {131969,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20632,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20374,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29245,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33288,-1,-1}
    },
    {
        {-1,-1},
        {20592},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34153,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24956,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24838,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20550,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20313,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21959,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20608,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20676,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29548,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21270,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22320,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20597,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36441,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20698,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38646,-1},
        {38646,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31838,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {132089,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31680,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35918,-1},
        {31466,35918,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32631,-1},
        {32631,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28288,28415,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20791,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32747,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25000,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20608,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20740,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20608,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24838,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29246,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {132170,-1},
        {-1},
        {-1,-1},
        {20796,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20676,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20185,20682,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30669,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22210,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20688,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20681,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38263,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22825,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38263,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20811,20805,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27515,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38263,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27515,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20821,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26444,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20821,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20816,20818,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20807,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20821,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32652,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33278,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23597,22575},
        {23597,22575,-1}
    },
    {
        {-1,-1},
        {-1},
        {23597,22575},
        {23597,22575,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40703,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20937,28092,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23346,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20731,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20834,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20176,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20047,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30690,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36935,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20840,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20840,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20056,20055,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21029,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20846,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20043,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20013,-1,-1}
    },
    {
        {132361,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {132361,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38263,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31637,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20849,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31637,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40786,27503,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20856,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {19982,33287,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25369,25831,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20013,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39668,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20908,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24323,26820,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31637,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22372,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31637,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22890,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20860,-1},
        {20860,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20860,-1},
        {20860,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20860,-1},
        {20860,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20860,-1},
        {20860,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22372,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20856,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30064,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26399,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26216,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20154,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32066,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30000,32593,-1}
    },
    {
        {-1,-1},
        {-1},
        {21516,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21516,-1},
        {32593,-1,-1}
    },
    {
        {20874,-1},
        {-1},
        {-1,-1},
        {20874,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20877,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20874,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23448,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20874,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20025,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38632,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32943,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26482,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38632,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20867,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23475,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27425,-1,-1}
    },
    {
        {26207,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29237,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20885,-1},
        {-1},
        {-1,-1},
        {20900,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26305,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20882,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38632,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20130,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22825,-1},
        {22825,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23481,-1},
        {23481,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {23461},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20901,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21466,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20896,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23478,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29017,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21476,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29017,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23541,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34588,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20913,-1,-1}
    },
    {
        {-1,-1},
        {27774},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27842,-1},
        {27842,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22846,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27931,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33624,33674,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {132566,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20925,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28165,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20925,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39342,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20940,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29245,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20908,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21360,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20956,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20908,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20956,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20919,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {132648,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28961,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20801,-1},
        {-1},
        {-1,-1},
        {20801,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22809,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22823,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22809,-1},
        {22809,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21542,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27704,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20805,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27665,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21830,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20973,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22812,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39080,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22987,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39080,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20197,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {19981,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22812,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39080,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29237,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20896,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34389,20966,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22519,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21213,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29245,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23376,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21475,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22239,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29976,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27491,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29579,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29976,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38998,22239,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22259,22294,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38754,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40786,27503,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26354,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20989,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31492,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22602,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20035,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20992,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20284,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21085,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24478,24467,20174}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21000,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21085,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21051,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21050,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37476,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37855,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29992,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21109,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21033,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21000,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21033,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21033,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25130,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37476,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24030,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21046,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21043,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21163,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21106,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21084,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21085,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21009,-1},
        {-1},
        {-1,-1},
        {21009,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21046,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21029,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21015,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21109,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21051,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21049,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21106,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21034,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21034,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21034,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21021,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21051,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21076,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21046,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37092,37060,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21009,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21083,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21038,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24037,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21182,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33075,33126,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26029,26039,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21137,21092,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21015,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21123,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21109,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21046,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30519,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28149,-1},
        {28149,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26029,26039,-1}
    },
    {
        {132943,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21078,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38329,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21069,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21046,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21216,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21209,-1},
        {21209,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30087,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21083,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31574,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38824,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21063,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21078,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21563,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21163,21031,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21065,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33570,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31574,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25134,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29522,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21051,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21123,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21049,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32742,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21104,21097,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21109,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32624,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21063,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32624,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29791,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21063,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21119,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21123,-1},
        {21123,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21127,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38656,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21063,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21138,21091,-1}
    },
    {
        {133127,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {133127,-1},
        {-1},
        {-1,-1},
        {26034,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28641,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35010,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21076,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38753,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21103,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21117,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21147,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {133178,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24380,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24038,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21185,-1},
        {21185,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36920,64103,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21185,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36211,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21201,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24180,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27544,27531,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21206,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21197,20510,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20853,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33285,-1},
        {33285,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21173,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29522,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21193,64051,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25973,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21213,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21216,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25032,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21222,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {133305,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21234,21235,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21172,21214,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25891,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32752,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24188,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22402,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26092,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26092,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36557,-1},
        {36557,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26092,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20898,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33467,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23452,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32004,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32970,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33967,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40179,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32118,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34664,34417,-1}
    },
    {
        {21269,-1},
        {-1},
        {-1,-1},
        {21270,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38263,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26045,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26088,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30691,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21417,24271,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21331,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30494,30495,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21331,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21331,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {19992,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27425,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28641,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {19968,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21274,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26479,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31776,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31776,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29983,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26734,26479,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26601,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29978,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31637,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24119,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21274,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24384,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24119,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31637,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26601,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25269,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26376,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28134,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30142,-1,-1}
    },
    {
        {133500,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20445,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22834,26412,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21317,-1,-1}
    },
    {
        {21321,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21321,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24180,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20133,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22834,26412,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21460,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {19989,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {19990,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25903,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24180,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20853,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {133533,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24184,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21332,-1},
        {21332,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21336,21934,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20054,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36554,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24107,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38761,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25958,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20081,20098,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20081,20098,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22732,22756,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21085,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34388,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35199,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20806,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36876,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23560,23554,30452}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21440,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36817,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20035,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40599,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21347,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34913,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20811,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36876,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21067,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36876,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40565,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20811,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34388,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20811,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31903,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34913,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35998,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25105,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20056,20055,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21347,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21375,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21356,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24380,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25233,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24481,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24616,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26444,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24616,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21375,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24059,21367,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20164,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23431,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22899,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36784,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20399,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36784,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25152,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26088,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26021,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24109,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25033,24540,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24218,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23784,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24109,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29467,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25151,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21402,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36784,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21407,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21407,-1},
        {21407,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20164,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25988,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28147,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30462,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23384,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21402,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29881,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39243,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37328,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27573,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30428,30423,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24246,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21426,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23919,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24265,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21402,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24246,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24962,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24246,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24246,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22880,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21405,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21407,-1},
        {21407,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24283,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24274,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21407,-1},
        {21407,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34382,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29965,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21021,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21407,-1,-1}
    },
    {
        {133843,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {133843,-1},
        {-1},
        {-1,-1},
        {133843,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32945,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24040,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31353,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20812,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21435,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24466,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21533,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21435,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24093,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20105,29229,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20845,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33775,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24189,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21442,21443,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21442,21443,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25993,40778,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21442,21443,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24180,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30042,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25993,40778,20125}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30044,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26013,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24323,26820,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40782,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33229,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21435,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23428,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {133917,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21454,-1},
        {-1},
        {-1,-1},
        {24318,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32945,21431,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21453,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25910,21454,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {133917,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24618,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24076,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30382,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24076,-1},
        {24076,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21451,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22857,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22868,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20853,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33509,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24471,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30045,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21466,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26705,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26356,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21843,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20107,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38442,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22863,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25954,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24321,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22863,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {15968,35930,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21463,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26705,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21129,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25562,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24247,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22781,23551,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34214,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24248,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {63856,27578,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21473,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20652,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25628,25436,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24043,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38816,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22793,35722,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22793,35722,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24962,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20081,20098,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21475,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26088,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21505,-1,-1}
    },
    {
        {134047,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21482,-1},
        {21482,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21491,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21566,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21484,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21484,-1},
        {21484,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21483,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21505,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21505,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21505,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21513,-1},
        {-1},
        {21513,-1},
        {21513,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36275,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20127,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21505,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21533,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21507,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33292,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21270,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29978,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33618,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35398,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35398,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21697,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21577,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21620,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21507,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26157,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30382,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21619,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35584,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21628,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37084,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21612,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36848,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29579,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21693,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35398,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34001,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22154,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21531,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21608,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21531,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22352,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21608,-1},
        {21531,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35398,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21606,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35496,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21533,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21627,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21891,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40807,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38728,38666,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21839,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21705,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21533,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22539,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22993,21476},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21495,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21631,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22087,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21533,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21533,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21741,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33694,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22768,32882,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22156,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31637,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21899,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22132,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27472,27431,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32993,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27795,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21533,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24107,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35726,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {134469,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {134469,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21892,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21892,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21620,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21840,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29572,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36203,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24735,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21759,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21813,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21550,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21892,21916},
        {21916,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21892,21916},
        {21916,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22877,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21843,-1},
        {21843,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26462,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21757,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21664,20582},
        {21664,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21483,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35742,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21643,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21918,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21930,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21561,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35578,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21886,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31637,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33503,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21919,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30028,-1,-1}
    },
    {
        {134625,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28418,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31574,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21676,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21606,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21483,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21930,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21767,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21820,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26029,26039,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28512,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21700,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20794,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22039,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28875,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22122,35679,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21776,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21747,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21931,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21683,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35591,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {134756,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21830,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27522,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32675,-1},
        {32675,32676,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {134765,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32629,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21883,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21824,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {134805,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21746,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26143,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22204,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35624,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22014,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35458,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22780,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21959,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21747,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38428,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21764,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40807,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21852,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {135007,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21693,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21669,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21602,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35591,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21912,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22210,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22170,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22132,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22216,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21959,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21830,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21892,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35201,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28216,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21822,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35292,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24754,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30044,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22210,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21693,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21931,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31908,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31908,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21816,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35646,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21917,64054,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21638,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21899,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32680,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21919,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22170,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21830,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20295,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22032,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38728,38666,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21917,64054,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36094,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21964,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21931,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23565,23550,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38907,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22170,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21764,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35712,35501,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22216,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21930,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22022,64055,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21830,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21930,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22170,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30883,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21542,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22184,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25136,25126,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22075,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21427,22196,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21830,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27468,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21776,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39336,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21507,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36094,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21830,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64056,22120,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21912,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25087,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25087,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20796,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22170,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20796,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21830,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35556,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21427,22196,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22024,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39467,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33393,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21427,22196,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21427,22196,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38758,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33393,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33393,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26085,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36215,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33391,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30446,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27169,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26085,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22283,22269,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28149,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20879,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22283,22269,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26085,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37009,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20879,-1},
        {20879,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31503,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22275,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28149,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22238,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30011,30059,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22283,22269,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23105,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22283,22269,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26085,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30452,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30446,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30446,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38754,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21476,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31382,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23341,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22259,22294,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26376,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29017,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20882,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {135681,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22259,22294,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38646,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26085,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22283,22269,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22280,22287,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22259,22294,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20882,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22254,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32618,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30475,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22290,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28288,28415,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29508,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22254,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31259,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22254,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22271,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {135741,-1},
        {-1},
        {-1,-1},
        {22303,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26015,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32854,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22312,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22522,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25176,-1},
        {-1,-1,-1}
    },
    {
        {135765,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23553,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22669,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23553,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22522,-1,-1}
    },
    {
        {22767,-1},
        {-1},
        {-1,-1},
        {22766,22767,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22395,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22349,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22343,-1},
        {22343,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22395,-1,-1}
    },
    {
        {135796,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {135796,-1},
        {-1},
        {-1,-1},
        {135796,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20806,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22395,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {135803,-1},
        {-1},
        {-1,-1},
        {22499,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22395,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22395,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22395,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23553,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22352,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33267,-1},
        {33267,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33267,-1},
        {33267,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22378,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {19992,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22539,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20806,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22434,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22353,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27877,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22432,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27915,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23553,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23429,-1,-1}
    },
    {
        {135895,-1},
        {-1},
        {-1,-1},
        {136301,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21488,33274,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32993,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22561,-1,-1}
    },
    {
        {135908,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22714,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26738,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22575,-1},
        {22575,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22352,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22432,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22432,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22522,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27801,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22714,-1,-1}
    },
    {
        {135933,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22411,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30060,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23713,22589,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22353,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22456,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22534,-1,-1}
    },
    {
        {135963,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {135963,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22419,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28132,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22500,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22612,-1},
        {22612,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21488,33274,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21488,33274,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22521,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27877,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22714,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22495,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22516,-1,-1}
    },
    {
        {135990,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22500,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {136004,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22320,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22395,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22396,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26612,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22396,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30566,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22718,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22402,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22402,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33255,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22669,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37325,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37325,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22779,23167,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22518,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34811,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21488,33274,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30758,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30238,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33771,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37083,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24283,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {136132,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22320,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22810,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22669,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22432,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22707,-1},
        {22707,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21488,33274,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22778,22775,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22539,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30566,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22500,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22649,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22714,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22509,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31966,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22549,22702,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33308,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23376,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22564,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22517,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22740,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38567,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38567,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22516,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22558,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31391,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35064,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26364,-1,-1}
    },
    {
        {136301,-1},
        {-1},
        {-1,-1},
        {136301,-1,-1}
    },
    {
        {136301,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {136302,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22738,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32854,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27584,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22625,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33255,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33255,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22685,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36884,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29087,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22669,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22672,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33255,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27567,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22320,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22645,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22320,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22523,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22718,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22737,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22523,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22602,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22602,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22320,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22725,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23493,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22170,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22737,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22622,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21021,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22654,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22697,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22725,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22645,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40612,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22645,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30956,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22916,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20973,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34001,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22778,22775,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22778,22775,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36617,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21916,-1},
        {21916,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22799,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22778,22775,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22780,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40723,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22777,22769,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34521,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21488,33274,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22777,22769,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25087,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22777,22769,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22777,22769,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22777,22769,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34873,34855,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20756,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25993,40778,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23416,23398,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22790,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29282,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35258,35226,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40643,40644,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23624,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24473,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40614,-1},
        {22829,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38263,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35201,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21213,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35201,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24481,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22799,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23130,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23130,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22799,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23130,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38816,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22806,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22809,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22810,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22810,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22809,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20134,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22812,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22809,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21375,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28879,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22810,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31227,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22820,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24685,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22821,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28879,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22828,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31435,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20134,-1,-1}
    },
    {
        {136663,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22839,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20171,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20869,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38738,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22840,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22840,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36557,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36557,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21577,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21577,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22871,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22840,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22840,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29282,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22852,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22372,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31337,31311},
        {31337,31311,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27169,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22777,22769,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22863,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22871,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22852,-1},
        {22852,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22519,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34913,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29791,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29245,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {136775,-1},
        {-1},
        {-1,-1},
        {27424,32570,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22577,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34913,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30494,30495,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27888,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30350,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34913,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22519,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22889,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36617,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22890,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38859,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20632,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22890,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22863,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36617,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28961,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24330,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27424,32570,136775}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29791,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32080,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21930,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22893,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22947,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {136884,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22971,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22900,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22934,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22909,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23005,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {18613,37142,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22949,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21606,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22985,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31922,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23199,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {136966,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {136966,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22971,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22971,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23241,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22890,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22971,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35349,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22922,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23006,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23035,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23105,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23650,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22971,-1,-1}
    },
    {
        {137026,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23072,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22904,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23338,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22915,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22915,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23105,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23199,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23105,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22934,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23130,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32856,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22941,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23105,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23269,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {18613,37142,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21897,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27442,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23110,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23068,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26285,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23307,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {64019,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25134,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23167,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27171,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23105,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24816,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23194,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30427,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40599,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24817,24745,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23105,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23194,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23265,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23290,-1},
        {-1},
        {-1,-1},
        {23290,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23340,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23207,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23348,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35201,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23372,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23348,-1},
        {23348,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31368,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {137405,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23285,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23348,-1},
        {23348,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40727,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23338,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23338,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23110,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23321,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23365,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35140,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {18613,37142,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23338,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23199,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23341,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23254,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23372,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37984,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20789,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23372,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33398,33399,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23403,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23380,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20445,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22909,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23384,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22909,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23387,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23389,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24726,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20351,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34850,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23381,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27530,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23383,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23411,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23387,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20449,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23413,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29312,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23018,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23407,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27131,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23418,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23418,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23395,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22581,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21402,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23448,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32596,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31361,64085,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32066,-1,-1}
    },
    {
        {137667,-1},
        {-1},
        {-1,-1},
        {23431,-1,-1}
    },
    {
        {-1,-1},
        {23427},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23431,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23431,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36007,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23428,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36051,64100,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32596,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23526,23455,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23490,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23478,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23429,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32905,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23445,-1},
        {23439,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23490,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23429,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23428,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21361,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23490,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23526,23455,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23475,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23448,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33509,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23524,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27714,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32581,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23452,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38911,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23459,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27627,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34388,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20901,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36557,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26647,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23472,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23491,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23490,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23522,23517,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23621,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20902,23495},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23522,23517,-1}
    },
    {
        {-1,-1},
        {-1},
        {23472,-1},
        {23472,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23527,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36015,23527},
        {23527,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20901,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29017,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31686,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35242,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23478,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33495,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22190,-1,-1}
    },
    {
        {-1,-1},
        {36051},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23478,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23493,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23459,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23453,23542,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23529,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23493,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31908,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23476,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28024,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22622,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23452,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21885,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37328,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22885,22887,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36015,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38911,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22818,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23459,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23453,23542,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23494,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36051,64100,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24826,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31422,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23472,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22190,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23487,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23507,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22885,22887,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31350,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36051,64100,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29254,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23522,23517,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23506,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36400,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22622,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38816,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23506,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23525,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23453,23542,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23506,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39002,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23537,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23522,23517,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25824,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23527,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31561,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21460,-1},
        {21460,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29237,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21460,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22865,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32784,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21034,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23432,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36022,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23563,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20140,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23432,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36947,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23563,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21117,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27193,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25134,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23562,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21129,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23565,23550,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36031,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23566,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23566,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23566,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27835,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21117,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31034,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36020,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31859,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20123,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24403,30070,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31558,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21460,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40636,40637,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38553,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22937,29573,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38553,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30465,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21329,64053,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20937,28092,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23567,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34191,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40636,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38553,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23587,-1},
        {-1},
        {-1,-1},
        {23586,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21448,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23588,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23592,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34426,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30175,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23593,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23593,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23593,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23601,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23495,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23601,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23600,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23601,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23601,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23608,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27513,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22839,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33391,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35930,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36523,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22829,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27513,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35925,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23612,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30475,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23632,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30382,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23611,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20811,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23630,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23616,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39608,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32937,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23614,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26088,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23614,-1,-1}
    },
    {
        {138326,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29312,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20811,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31458,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23615,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23651,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23611,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36764,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23614,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23615,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32071,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27522,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23627,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23612,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23630,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23624,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23630,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23615,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21693,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25153,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26355,25341,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36764,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23653,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30772,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24450,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38477,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23617,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23624,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22863,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22863,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28346,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24473,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23614,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23624,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32675,32676,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24850,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35248,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33308,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33308,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32071,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22863,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27155,27004,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24300,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20043,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20016,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21440,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25856,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24066,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27598,27599,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30002,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38738,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38738,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38738,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24605,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27598,27599,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21335,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24184,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27602,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24892,24910,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22863,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21335,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26412,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25163,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31163,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29071,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21930,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22863,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26412,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25308,25309,-1}
    },
    {
        {138541,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25160,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25160,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21361,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23731,23997,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21361,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23690,-1,-1}
    },
    {
        {23674,-1},
        {-1},
        {-1,-1},
        {23674,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21361,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {138565,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21440,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23992,-1},
        {23992,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30029,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23791,-1},
        {23792,23791,-1}
    },
    {
        {138594,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23739,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27497,27493,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27969,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {138616,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23731,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23715,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28845,-1},
        {28845,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35895,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21638,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22369,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23494,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38465,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20489,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {138642,-1},
        {-1},
        {-1,-1},
        {23722,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23731,23997,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {138652,-1},
        {-1},
        {-1,-1},
        {14273,-1,-1}
    },
    {
        {138657,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {138657,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29992,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23713,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23713,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23731,23997,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23792,-1,-1}
    },
    {
        {138679,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23665,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23792,23791,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23713,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38738,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23849,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23731,23997,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {138720,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23690,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23839,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22402,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20957,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20351,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32652,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23916,14324,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22350,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22534,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23751,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23735,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22534,-1,-1}
    },
    {
        {138803,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {138804,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23842,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23834,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27602,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22432,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23735,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23713,23831,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23849,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22384,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23846,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23789,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32652,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24962,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27849,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30408,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22622,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22534,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23834,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23947,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31391,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23784,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27507,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24019,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30566,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23803,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27665,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23822,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23735,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23814,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22863,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23819,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22707,-1,-1}
    },
    {
        {139038,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28567,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28526,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22528,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23919,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40786,27503,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36335,-1},
        {36335,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24009,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23961,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23784,23785,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23996,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {139126,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {139126,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23731,23997,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29071,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23753,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23526,23455,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23919,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30922,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27508,27511,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23954,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24009,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24027,-1},
        {-1},
        {-1,-1},
        {22372,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28797,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {13356,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {13356,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22372,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {133843,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {13356,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {133843,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24032,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37013,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31637,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23376,-1,-1}
    },
    {
        {139258,-1},
        {-1},
        {-1,-1},
        {33075,33126,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34886,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24033,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31452,40845,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24036,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36617,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23376,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23376,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23376,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27827,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23411,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24040,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24049,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24040,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26032,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24046,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38579,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24043,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24046,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24046,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21021,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21380,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21370,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26389,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21370,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36491,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24061,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24061,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23493,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24070,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35925,20133,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24413,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24067,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {63856,27578,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26503,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23609,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24076,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33836,19975,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20133,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31995,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34382,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39719,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24093,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20140,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25325,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24107,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21331,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23609,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21331,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23609,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23609,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24089,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20134,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32140,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24159,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24118,24111,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35033,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23609,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35033,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24107,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24172,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20896,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35084,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20399,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32217,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23609,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32411,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29279,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23609,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31574,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35930,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32264,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {19975,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30064,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32069,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36880,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24172,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32900,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30431,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27216,27094,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24172,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27216,27094,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34962,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24173,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24152,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31452,40845,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32160,32209,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35172,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25628,25436,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27512,24112,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27512,24112,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24172,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21299,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {139643,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24150,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39409,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27712,-1},
        {27712,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33279,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24184,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21460,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25136,25126,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24183,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24181,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31992,29572,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24188,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24190,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24190,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32230,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27877,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31992,32114,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32230,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32057,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32118,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24413,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24189,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26377,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23411,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24281,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24289,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24213,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22889,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30058,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25033,24540},
        {25033,24540,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24220,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25155,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30693,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39740,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24188,-1,-1}
    },
    {
        {139800,-1},
        {-1},
        {-1,-1},
        {40575,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31458,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35242,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24297,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25110,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28431,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29254,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40575,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24246,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23431,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31458,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23526,23455,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20660,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33104,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30238,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27508,27511,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24245,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29791,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37739,37740,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24674,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24247,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31209,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22889,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40763,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36988,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35390,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24289,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21959,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24283,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34382,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21426,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24265,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24289,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24275,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35131,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {37181},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24289,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31964,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24275,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24449,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24449,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36225,-1,-1}
    },
    {
        {140062,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20061,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31561,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30028,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {19982,33287,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {19982,33287,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {19982,33287,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24061,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25106,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24328,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31639,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25215,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24328,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30016,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40763,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22868,-1},
        {22868,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24605,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22864,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24321,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35440,35493,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32066,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24327,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35328,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24321,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24321,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23562,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34849,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26820,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22865,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22885,22887},
        {22885,22887,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32622,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35328,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24043,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33288,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36981,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24061,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26278,26310,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25806,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24331,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27494,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28797,28798,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24337,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24337,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20108,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25140,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30704,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24337,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24339,-1},
        {-1},
        {-1,-1},
        {20035,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24392,24382,-1}
    },
    {
        {24339,-1},
        {-1},
        {-1,-1},
        {21450,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24392,24382,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31680,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24341,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24340,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23556,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31998,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27571,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20035,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24340,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24392,24382,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24392,24382,-1}
    },
    {
        {140205,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24340,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27664,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24358,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24380,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22402,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21698,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32703,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20035,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29447,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24381,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24358,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24381,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28858,29234,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24380,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24358,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24381,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24380,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24347,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39191,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24347,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37848,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24380,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24380,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24380,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27296,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28818,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21916,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32690,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40179,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39256,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28818,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24372,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29009,29038,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24396,24357,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24391,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24396,24357,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39739,31909,-1}
    },
    {
        {24403,-1},
        {-1},
        {-1,-1},
        {24403,30070,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22810,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22909,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30087,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21216,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26089,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24404,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39749,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39749,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30003,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39741,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30087,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24408,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24409,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32902,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32900,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29399,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30394,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24409,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34655,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22810,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24413,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64022,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32763,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24409,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35946,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22778,22775,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37027,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22218,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26045,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24037,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24418,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24422,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35036,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64026,31077,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35576,64034,-1}
    },
    {
        {-1,-1},
        {-1},
        {24423,-1},
        {24423,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39340,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22793,35722,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24426,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22793,35722,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39340,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26748,28422,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28343,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20009,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22793,35722,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39340,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22793,35722,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23563,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24433,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22793,35722,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27193,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31137,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33394,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20009,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29359,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24466,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24467,24478,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24455,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20206,-1},
        {36820,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24473,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20316,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20888,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20219,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31458,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20426,-1},
        {20426,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20426,-1},
        {20426,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21566,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36890,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24481,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24448,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24456,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24495,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20439,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20405,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23653,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24465,24452,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20478,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20350,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24448,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24460,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24455,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20462,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24471,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36952,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24471,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20657,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36437,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24481,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24928,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24494,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20692,-1},
        {36981,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36960,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34503,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24505,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36996,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36991,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20693,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24505,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24962,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31462,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34909,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36786,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20676,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24525,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24616,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24656,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24708,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24859,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24565,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24597,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24535,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24561,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24561,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24848,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24592,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24641,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24573,-1},
        {24573,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24671,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24535,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24594,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26580,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24658,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24603,-1},
        {24603,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24616,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26585,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20280,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24380,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24862,25062,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33251,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24617,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35584,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24616,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24616,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24656,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24597,24687,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24555,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24618,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24545,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24713,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33678,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24680,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24685,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24687,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24817,24745,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24675,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29378,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24713,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24950,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24684,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24656,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24859,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24715,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23014,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24656,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24789,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31899,32901,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24724,64061,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21155,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29378,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24685,-1},
        {24685,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24680,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24713,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23490,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24773,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24736,-1},
        {24736,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24745,-1},
        {-1},
        {-1,-1},
        {24817,24745,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24603,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24764,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24601,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35726,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24705,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24821,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23358,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20518,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36785,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24658,24646,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28357,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24735,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24799,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24758,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24616,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24801,24746,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23494,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24859,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {141043,-1},
        {-1},
        {-1,-1},
        {24755,24499,24503}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24935,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24755,24499,24503}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24859,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24826,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35624,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21696,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24760,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24962,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24647,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24801,24746,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24838,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24950,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24827,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24713,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {24985},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24656,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25010,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24833,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24970,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24816,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24962,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24950,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24642,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25010,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24789,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24800,24693,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24845,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23241,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24656,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24962,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24903,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24996,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24970,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24944,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25059,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36785,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36785,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20762,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24917,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24755,24499,24503}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24800,24693,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25030,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24976,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24797,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24739,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25001,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24851,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25031,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24838,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25084,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24789,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24920,24808,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33804,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24800,24693,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24739,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {141403,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24816,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34850,24823,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30471,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24796,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30471,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24930,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24930,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24900,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24850,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24816,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24928,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24872,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25032,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24827,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36705,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35726,-1,-1}
    },
    {
        {141483,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24970,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25053,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24816,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24683,24927,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21232,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24976,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25705,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24925,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24775,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24755,24499,24503}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22123,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24739,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24735,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24996,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25037,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24978,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24800,24693,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20303,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21172,21214,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25055,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24545,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24915,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39000,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36705,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25017,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24597,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25030,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31354,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {141711,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25105,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25105,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30691,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25105,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25110,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39320,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25129,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22570,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25105,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24863,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24351,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30683,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21191,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25104,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25114,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25106,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25106,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28357,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27507,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22570,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26856,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34850,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22283,22269,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27507,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34850,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37707,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30683,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25119,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34850,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34850,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25130,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25681,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34850,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31330,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29118,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34101,34255,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25140,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30704,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22495,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38902,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34850,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25122,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33255,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24337,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29118,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25140,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35672,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34850,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21380,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25151,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24053,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25152,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25147,25150,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23616,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36764,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38728,38666,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38728,38666,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22833,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20014,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24039,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21033,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24341,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25215,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24318,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33281,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25269,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29289,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {142008,-1},
        {-1},
        {-1,-1},
        {25445,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25244,25300,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25215,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25747,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23558,23559,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36764,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29309,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36983,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25293,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25641,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25308,25309,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25237,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25233,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25351,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25233,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21370,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33655,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25324,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25361,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26059,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {142150,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25696,32317,32207}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25554,25407,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25594,25622,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25352,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25594,25622,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25571,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {142159,-1},
        {-1},
        {-1,-1},
        {25827,-1,-1}
    },
    {
        {142160,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25334,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25198,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25166,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25327,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25326,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25693,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25637,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25507,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25586,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33109,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25445,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25331,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25421,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25308,25309,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20081,20098,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25226,-1,-1}
    },
    {
        {142246,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30016,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25233,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22396,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25436,25628},
        {25628,25436,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33109,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25718,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25303,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25582,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25776,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26053,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25606,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25554,25407,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36983,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25347,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25571,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25308,25309,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {142365,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {142365,-1},
        {-1},
        {-1,-1},
        {142365,-1,-1}
    },
    {
        {142372,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25579,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25463,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25303,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31462,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25855,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25209,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25516,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25829,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33281,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25582,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25628,25436,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25206,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25542,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25387,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25506,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25754,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31518,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25774,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25705,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29831,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25507,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25304,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20682,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25308,25309,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21046,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25594,25622,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29309,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22686,22679,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32118,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25711,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25462,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25746,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25308,25309,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25481,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25556,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25514,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25539,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25586,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25522,25581,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25827,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25652,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25725,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25539,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25540,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25793,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25463,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25496,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25654,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31391,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25299,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25308,25309,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25652,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31644,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25678,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25312,25818,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31462,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25556,-1,-1}
    },
    {
        {142817,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25678,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25619,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27010,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25725,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25808,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25539,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25856,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25850,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25688,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25299,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25771,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25634,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33281,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25816,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25184,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25787,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25793,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25793,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25725,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25411,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25334,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25787,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25892,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25771,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25880,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27419,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25793,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25794,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25903,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36764,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22810,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25945,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39791,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22793,35722,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25955,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25955,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26045,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25906,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23416,23398},
        {23416,23398,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25206,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21843,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25945,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25305,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20316,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36859,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26045,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36198,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39178,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21047,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {63856,27578,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {63856,27578,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24380,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25906,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25958,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {63856,27578,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22949,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25954,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36198,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39178,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {63856,27578,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {63856,27578,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23495,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21919,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {63856,27578,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25787,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21189,25941,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24500,24501,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24500,24501,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25954,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22863,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {63856,27578,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {63856,27578,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23142,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {63856,27578,-1}
    },
    {
        {-1,-1},
        {25989},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25954,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27544,27531,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25577,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {63856,27578,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27584,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25942,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25356,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31302,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34850,24823,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25964,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26057,25975},
        {25975,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24509,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25787,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25940,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21213,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24500,24501,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32903,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31302,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30027,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31434,31363,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36011,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {63856,27578,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25976,25968,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25968,25976,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25976,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25978,39493,39366}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25773,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25955,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25964,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25973,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25972,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31302,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20081,20098,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20081,20098,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25955,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {63856,27578,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25805,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25805,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {63856,27578,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25943,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25964,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22804,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25958,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25282,25173,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27449,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25976,25968,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25880,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22793,35722,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23416,23398,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24494,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35465,35709,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37670,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24425,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24425,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {29166},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20153,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26007,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26015,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29096,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26007,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26512,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26031,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25240,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36817,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26034,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26029,26039,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21086,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21079,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26034,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25240,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34402,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35475,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26029,26039,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25240,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35475,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25240,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26032,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26040,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26034,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26034,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26034,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40718,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26034,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26034,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36024,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26029,26039,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26029,26039,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26049,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25151,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26049,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26060,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21247,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26053,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28354,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26051,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38515,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38515,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26051,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21463,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26053,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24159,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24159,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24162,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24162,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31048,64078,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26127,-1},
        {26127,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26127,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26088,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34955,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26399,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20908,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26124,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35910,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26088,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26212,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26107,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30336,30337,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26115,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30342,-1},
        {30342,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21413,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26220,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35910,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26115,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32828,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26172,26205,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26178,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21413,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26115,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26154,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26089,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26146,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38525,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21402,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26149,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26165,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20360,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26263,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26151,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26159,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39740,-1,-1}
    },
    {
        {-1,-1},
        {26179},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26122,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26124,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {143798,-1},
        {-1},
        {-1,-1},
        {26217,26202,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26177,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25954,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30090,30067,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23477,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26164,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20908,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26127,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26213,30358,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26209,-1,-1}
    },
    {
        {143811,-1},
        {-1},
        {-1,-1},
        {33282,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {143812,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26201,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26181,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26332,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26224,26227,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26164,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27089,32615},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26126,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26399,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30431,-1},
        {-1,-1,-1}
    },
    {
        {143861,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27809,27794,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33094,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26185,26187,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30563,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30090,30067,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25562,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26149,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26240,26106,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32066,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {63952,39006,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26274,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26181,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21063,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26127,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26313,26241,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26181,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29031,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33945,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24409,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21776,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26234,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36101,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32066,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21209,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28212,28167,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29031,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26283,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26308,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26361,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38997,15118,39023}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33284,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36786,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26336,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26132,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26149,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26292,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27766,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29152,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26292,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33879,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26308,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22799,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26118,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26127,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26292,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26292,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26336,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26185,26187,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26333,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26302,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26148,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33391,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20105,29229,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33391,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20250,26371,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26361,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31755,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38859,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35920,35914,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30883,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26361,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40652,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33311,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26381,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32943,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26399,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26399,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26399,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38512,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30472,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26383,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33337,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26388,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26388,-1,-1}
    },
    {
        {144242,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30928,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38712,35207,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26389,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28373,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26397,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20940,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29694,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23381,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30350,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36786,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27377,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26422,-1},
        {26422,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20445,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26422,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26576,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {63856,27578,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25903,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26681,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {144336,-1},
        {-1},
        {-1,-1},
        {26451,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {144338,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {144338,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {144339,-1},
        {-1},
        {-1,-1},
        {26439,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22256,-1,-1}
    },
    {
        {144341,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26574,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {144346,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26825,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26517,-1},
        {26517,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {144351,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {144356,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31177,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26422,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26522,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26448,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21050,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32786,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26617,26623,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26803,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27227,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26653,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26898,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26439,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26874,-1,-1}
    },
    {
        {144458,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {144459,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {144465,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {144485,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28961,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26676,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26613,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26613,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26579,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26704,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34384,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26553,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27063,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26505,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26999,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26681,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64068,26757,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26685,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26707,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26676,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26446,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26465,26781,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26654,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25463,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25463,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26501,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27063,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26898,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26803,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27082,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30452,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26611,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32177,-1,-1}
    },
    {
        {144612,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30042,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26976,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27082,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26398,26399,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20056,20055,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26589,-1},
        {26589,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27165,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26552,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26653,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26705,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26970,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27193,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23376,-1,-1}
    },
    {
        {144730,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {144730,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26839,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27068,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27377,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30452,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {144788,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27227,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27190,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30452,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26795,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26970,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {144836,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {144836,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {144836,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27193,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26647,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21083,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26840,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24185,27046,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26787,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31518,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26731,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26883,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26585,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27122,-1,-1}
    },
    {
        {144952,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {144952,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {144953,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {144954,-1},
        {-1},
        {-1,-1},
        {27211,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {144967,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34503,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27063,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26803,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27198,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26708,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27118,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27163,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26928,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27277,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27325,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27310,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33284,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27066,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32789,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27193,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24374,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28961,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27053,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27236,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26852,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27378,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27198,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26978,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31206,-1,-1}
    },
    {
        {145164,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {145180,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {145180,-1},
        {-1},
        {-1,-1},
        {145180,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27268,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29026,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {145215,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {145215,-1},
        {-1},
        {-1,-1},
        {145215,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31644,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23565,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26589,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26680,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26486,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34394,34395,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {145251,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {145252,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29246,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25856,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27310,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40595,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35316,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27035,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27264,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27299,-1},
        {27299,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29965,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34081,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26484,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27086,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26420,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33882,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27001,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37027,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28961,-1,-1}
    },
    {
        {145383,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {145383,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27195,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27189,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27118,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27277,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {145407,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27386,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26881,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27161,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25666,25885,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26787,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20769,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26586,-1,-1}
    },
    {
        {145444,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28961,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {145469,-1},
        {-1},
        {-1,-1},
        {27419,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27377,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26840,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22218,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27133,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26601,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27280,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22218,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29237,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22826,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26647,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39729,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34271,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21828,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24350,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21627,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27472,27431,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27454,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27447,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27424,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33761,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27454,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27459,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36039,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21988,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21964,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27453,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21854,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27454,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27448,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27468,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27468,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39154,39150,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27466,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26368,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22118,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35556,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33394,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27476,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27470,-1,-1}
    },
    {
        {23569,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20047,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27492,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20250,26371,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21675,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24187,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36817,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25136,25126,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26053,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26053,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21069,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26053,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27491,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23569,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25144,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30326,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36319,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31672,33879,64095}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28041,28169,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36817,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40786,27503,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25136,25126,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24466,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27515,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28041,28169,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35576,64034,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21021,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21069,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34394,34395,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31071,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36487,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40786,27503,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25972,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20094,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40786,27503,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36475,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36466,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21098,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36487,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36484,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27512,24112,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36705,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21473,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27515,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27524,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22809,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24344,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27515,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23692,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20982,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27519,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24030,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27529,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22829,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27524,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27524,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32066,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27522,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36764,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33140,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33140,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27533,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21761,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27519,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27522,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27562,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27522,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27570,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64082,31117,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28232,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64082,31117,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21761,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21375,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20901,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33104,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27562,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27522,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22857,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25134,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32302,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21222,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34928,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34091,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30279,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30279,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34212,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34216,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29677,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27590,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {63856,27578,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25937,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25237,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27583,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25954,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {63856,27578,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {146072,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27589,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {63856,27578,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21084,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38015,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27584,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40391,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25970,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27589,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27589,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38015,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37328,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31995,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38015,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38015,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33945,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27598,27599,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36463,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29767,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32946,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23105,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27604,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27607,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27607,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25308,25309,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27425,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28151,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25308,25309,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27607,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21442,21443,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31167,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40638,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39653,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32021,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26052,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {27628},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27622,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32772,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27650,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27628,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {16441,30571,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23615,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23624,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25988,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27650,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {16441,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {16441,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {16441,30571,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {16441,30571,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38816,27633,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {16441,30571,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34952,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30473,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34952,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32772,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38816,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27662,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32287,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37328,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27635,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {16441,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39675,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {16441,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {16441,30571,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38816,27633,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27647,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27653,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {16441,30571,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27656,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {16441,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32381,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39715,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27662,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27662,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20282,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31063,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27683,27671,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38512,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27683,27671,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27684,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27675,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28040,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27738,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27704,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27704,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36008,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {146559,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27788,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20061,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22395,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39154,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25912,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27801,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27875,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22349,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20912,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27844,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27728,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27801,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20916,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20916,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27792,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27832,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22349,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27770,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27711,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24618,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28356,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20811,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27867,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27844,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27869,28335,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27969,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27886,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28193,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28046,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22534,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28049,20351},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27754,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27835,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28363,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28500,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27922,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27861,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {146686,-1},
        {-1},
        {-1,-1},
        {27888,-1,-1}
    },
    {
        {146688,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {146702,-1},
        {-1},
        {-1,-1},
        {25215,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27969,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29356,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27869,28335,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27960,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27963,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27859,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27950,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28657,27996,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27874,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27737,27738,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27915,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28020,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64076,31038,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64069,28023,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28422,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28149,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28550,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28149,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28544,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28003,28579,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27825,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {146752,-1},
        {-1},
        {-1,-1},
        {28037,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39154,39150,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28046,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28096,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27961,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28644,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27825,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27825,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27779,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27964,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28046,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28101,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28526,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27916,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28753,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27942,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27766,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23734,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27958,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28170,28149},
        {28149,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27922,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28450,64071,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28046,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28046,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28046,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28046,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28356,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28354,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {28085},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {146899,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28550,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28076,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28241,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28422,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28074,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28155,-1,-1}
    },
    {
        {146937,-1},
        {-1},
        {-1,-1},
        {28155,-1,-1}
    },
    {
        {146937,-1},
        {-1},
        {-1,-1},
        {28155,-1,-1}
    },
    {
        {146938,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28024,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27941,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27963,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28338,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28606,-1},
        {28606,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26753,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28168,28639,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27700,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28076,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28335,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32705,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28459,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27916,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28422,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {28500},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {28500},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28312,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30437,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28267,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27922,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28512,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28239,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22395,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28085,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28544,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28114,20932},
        {28114,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28024,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28123,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27969,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27965,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27969,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26753,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28373,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28357,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28046,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27849,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28147,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28354,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28024,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33618,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28629,28287,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28168,28639,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28478,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28498,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28418,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27885,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28154,27973,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28041,28169,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27995,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23615,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30442,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28481,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28100,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39749,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28288,28415,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28153,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28364,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28425,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28024,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28431,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28409,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28601,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27963,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31569,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28520,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28641,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28641,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28652,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27737,27738,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28606,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28270,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28666,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26165,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28207,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28460,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28651,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28373,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26047,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28228,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23506,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38698,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28168,28639,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28234,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28104,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28536,-1,-1}
    },
    {
        {147326,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28244,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28606,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28228,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28404,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28545,28171,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28149,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28526,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28593,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28409,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28207,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28606,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28418,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34153,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28460,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28327,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27963,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28544,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28417,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28104,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28518,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34153,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28356,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28595,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28436,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27810,28580,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28961,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28730,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28363,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28417,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28606,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28550,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28409,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28450,64071,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28185,-1,-1}
    },
    {
        {28701,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28595,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28544,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28644,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28760,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28552,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28628,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28404,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28204,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28136,27972,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28363,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {147606,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28673,28478,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28753,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28703,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28748,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28373,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28147,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28629,28287,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28618,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28512,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28629,28287,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28527,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28654,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28644,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28528,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28654,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28003,28579,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28088,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28228,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28228,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28689,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28668,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28701,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {147715,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28687,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28316,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28689,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24189,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28418,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28504,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28777,28750,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28670,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28678,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28746,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28154,27973,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28009,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35742,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28611,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28437,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28732,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28611,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28729,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20153,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28628,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28525,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28760,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38646,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28732,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28729,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22320,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28760,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28760,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28777,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34146,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28849,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28779,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36198,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28784,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28797,28798,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36196,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28797,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64072,29038,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26262,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28797,28798,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22240,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33995,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {147910,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22396,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28821,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28937,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29572,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31179,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39146,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23475,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28893,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {147966,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28875,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28818,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23561,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33836,19975,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34909,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28797,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29152,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29152,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24246,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29134,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28831,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28892,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20821,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35937,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28797,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28797,28798,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28872,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29024,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20809,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39770,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28888,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29017,28895,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28966,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28818,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28849,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28818,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29134,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29036,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30428,30423,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64097,35222,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29105,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29190,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29013,-1},
        {-1},
        {29081,29013},
        {29081,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29009,29038,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21161,-1,-1}
    },
    {
        {29071,-1},
        {-1},
        {-1,-1},
        {29071,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39770,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28821,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29083,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28883,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28892,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29138,28988,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29006,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21350,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29004,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28961,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36051,64100,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33836,19975,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29166,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29157,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29105,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29224,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37294,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28975,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29134,23582,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29017,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37619,37956,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29159,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20809,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29166,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29105,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29006,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29006,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29118,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26989,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {148412,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28872,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {64021},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29143,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24785,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26262,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31932,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29157,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27169,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29017,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29113,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29224,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29100,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39146,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30430,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36637,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29134,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29063,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29118,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29177,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29136,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31080,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28825,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30428,30423,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20547,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29103,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28925,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20809,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29191,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29224,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29009,29038,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25774,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28999,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28958,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29121,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26989,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29159,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29190,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26332,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29087,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28825,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29121,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29159,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29190,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26180,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29083,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28593,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37984,29197,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28982,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29190,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32831,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29224,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29224,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28966,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29159,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28966,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29096,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25484,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31994,31998,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21365,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37319,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28858,29234,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23386,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24179,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20445,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20107,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32147,32076,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38560,38577,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20081,20098,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20081,20098,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31036,31150,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20081,20098,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20081,20098,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20081,20098,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32147,32076,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25899,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64027,31119,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24859,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24859,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24034,24035,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20081,20098,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36781,36766,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25899,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {18188,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20081,20098,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {63856,27578,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35316,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21496,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20081,20098,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25088,24651,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22781,23551,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20081,20098,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29242,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29240,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25945,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30095,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23011,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23011,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20096,40860,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29245,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39658,39662,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29246,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29247,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30098,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23558,23559,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28479,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21359,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20430,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33624,33674,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30142,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30142,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27299,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28479,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33624,33674,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30149,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33624,33674,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34101,34255,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36920,64103,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30142,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25943,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31744,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37292,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37292,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36920,64103,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27280,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29254,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29254,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20633,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29254,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22941,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24070,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29256,27573},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21477,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27573,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26543,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22396,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21521,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21521,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31419,31379,29269}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32996,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31419,31379,29269}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29266,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29260,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26847,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29266,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26989,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38548,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32581,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31744,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29270,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27280,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27310,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29272,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29273,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29273,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40818,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24193,24307,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29277,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {149033,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29300,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29287,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29300,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20320,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39381,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26041,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39426,-1},
        {39426,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26463,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29305,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38597,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29346,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29300,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29312,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40605,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {149157,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25391,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29325,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29281,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28879,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25391,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25391,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39426,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29346,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29548,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29314,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29314,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29359,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {15968,35930,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29509,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29392,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29480,29503,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29359,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30027,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29544,29420,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29403,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29362,35962,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29378,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29378,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29464,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21179,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29458,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20445,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33108,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29409,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22888,22892,29518}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20495,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29464,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22888,22892,29518}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24757,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29480,29503,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28982,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29563,29486,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22071,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33912,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {149489,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25147,25150,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29458,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29480,29503,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29503,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30044,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29066,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35988,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39509,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29563,29486},
        {29563,29486,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20676,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29458,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25147,25150,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39509,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {149654,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29560,29539,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29544,29420,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29522,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29557,-1},
        {-1},
        {-1,-1},
        {29557,29471,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40762,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29560,-1},
        {-1},
        {-1,-1},
        {29560,29539,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29559,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29550,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20778,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29550,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29577,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26007,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29608,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35511,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29786,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29641,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29606,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {19992,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29808,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29642,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24634,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36861,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29677,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30895,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29744,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29578,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29697,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23453,23542,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29705,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29589,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29848,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29825,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22283,22269,-1}
    },
    {
        {-1,-1},
        {-1},
        {29600,-1},
        {29600,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29748,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29677,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32854,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29898,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29759,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29701,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29872,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21644,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29848,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29723,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36861,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29749,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29750,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29764,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29795,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29885,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29872,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26001,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29701,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29791,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29723,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29642,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29865,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29887,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29773,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29881,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29881,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29881,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22685,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29887,-1,-1}
    },
    {
        {150093,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29748,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29887,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29808,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34399,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29809,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29722,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29849,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29796,29814,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29848,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29877,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29822,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29811,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29722,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29872,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20608,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29887,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30887,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29885,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38728,38666,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29877,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29748,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29618,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29922,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29926,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32568,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30414,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29965,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {150358,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {150383,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29955,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30414,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29943,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29956,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22748,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32591,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31025,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29973,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29976,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26088,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33292,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21577,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34071,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32829,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34071,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29983,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20154,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22995,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31337,31311,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31513,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38738,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23035,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26143,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22995,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23273,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22995,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22995,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23273,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {63964,38534,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29990,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27515,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29992,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20633,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29992,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20633,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20633,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24248,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22665,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30030,-1,-1}
    },
    {
        {150550,-1},
        {-1},
        {-1,-1},
        {30045,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30041,30065,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30030,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21329,64053,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21329,64053,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30012,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30031,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30060,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37030,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30031,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30011,30059,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30031,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23041,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30031,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30003,-1},
        {30003,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30043,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38520,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30031,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30027,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30043,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30011,30059,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24403,30070,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39749,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36617,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30043,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39741,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26274,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30042,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30042,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21959,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30041,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30045,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30087,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30011,30059,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29761,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31966,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23346,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31329,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30071,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36426,24495,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26657,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30011,30059,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22642,22580,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21959,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30078,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30054,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34426,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32047,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30053,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26143,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32047,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38647,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38721,38647,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38647,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27491,-1,-1}
    },
    {
        {150804,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36275,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30092,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26803,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22799,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38728,38666,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20164,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30100,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30117,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32989,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26021,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30114,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30147,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30149,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26413,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32923,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {151054,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30267,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30142,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30185,30246,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30305,30196,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30123,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30142,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30165,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30162,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30106,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30241,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30193,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30115,32940,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32983,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30450,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30211,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30185,30246,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30185,30246,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30207,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30275,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30117,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {151054,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30150,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25000,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30164,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24287,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24817,24745,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34645,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34928,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30223,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40441,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24970,-1,-1}
    },
    {
        {151095,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27436,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24230,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39260,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30240,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {151146,-1},
        {-1},
        {-1,-1},
        {30144,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30229,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40636,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30185,30246,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30227,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24817,24745,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30185,30246,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30223,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {151179,-1},
        {-1},
        {-1,-1},
        {30137,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25000,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35390,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21854,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30235,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30279,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30247,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30238,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33178,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30241,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {151095,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30188,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30221,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40200,38593,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25540,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30320,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30201,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30318,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24970,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30308,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30114,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30142,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30227,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25033,24540,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30185,30246,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30256,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30308,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30245,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30296,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30185,30246,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30253,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30296,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30256,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40441,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36491,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30328,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30331,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29677,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30330,30332,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30331,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30331,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30333,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27849,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35910,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27683,27671,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38512,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36843,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32066,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30343,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30343,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26143,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {30342},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30378,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30343,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32066,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30568,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33369,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26234,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30340,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33457,34340,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27512,24112,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26143,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33457,34340,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30378,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30383,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36315,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35178,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36764,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30394,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38808,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30394,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30394,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38859,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32646,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32646,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32646,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23381,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35178,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27449,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23391,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30402,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30402,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30437,-1,-1}
    },
    {
        {151626,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {151637,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22778,22775,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30433,23613,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23391,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30430,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31755,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30431,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30431,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39188,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30413,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30431,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30409,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30430,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30442,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34849,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30409,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27345,32589,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30439,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27068,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40572,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30446,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33390,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20855,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20855,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30449,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30449,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39318,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34701,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31365,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30452,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30452,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30452,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35222,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30566,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30559,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30517,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32932,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64097,35222,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64097,35222,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {151842,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32774,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35223,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30465,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40718,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30473,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36004,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30449,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64097,35222,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30473,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30465,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32854,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30542,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36031,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26212,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38758,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30524,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30446,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {151977,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30563,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30450,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {152013,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {16441,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30570,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30570,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {152037,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30473,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38758,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30487,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30606,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30501,30502,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30586,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {152094,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30591,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23562,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39973,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {152140,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {152140,-1},
        {-1},
        {-1,-1},
        {152140,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30517,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30475,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30562,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30509,30566,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30474,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64097,35222,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30561,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30496,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30626,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30554,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {16441,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29031,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30624,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30624,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30624,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30628,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30623,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30561,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30471,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26342,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30623,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29031,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30618,30636,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30626,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30628,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30653,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38525,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30623,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30646,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33393,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30646,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35264,35251,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35264,35251,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30623,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35264,35251,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30683,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30683,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30462,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27085,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30684,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30688,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30694,20399,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30695,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30690,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26063,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30692,30695,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37827,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30097,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30693,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30701,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30698,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26063,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26063,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30693,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37827,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26063,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30142,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24408,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32900,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21776,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30142,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30244,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38601,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26234,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26234,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30142,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38859,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30707,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30765,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30707,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37489,30970,37979}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {152622,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30757,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30791,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30765,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30910,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30928,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30791,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30787,22995},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30791,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {152718,-1},
        {-1},
        {-1,-1},
        {30858,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30828,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30789,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30967,-1},
        {30967,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30895,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37489,37979,31014}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29001,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21754,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30791,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30740,30799,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22509,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30865,64075,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30956,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {152793,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30895,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30962,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26612,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24109,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {152846,-1},
        {-1},
        {-1,-1},
        {30869,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30933,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30770,31022,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30830,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30913,-1},
        {-1},
        {-1,-1},
        {30913,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23720,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31024,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21405,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26881,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30844,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31019,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30877,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22072,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30895,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26027,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38911,-1,-1}
    },
    {
        {152999,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25705,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30922,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25705,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30992,30830,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30922,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38713,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30791,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31034,-1},
        {31034,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31036,31150,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31071,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31071,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31098,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38957,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36198,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36198,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64081,31069,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64080,31062,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22856,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31206,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31063,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31152,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31040,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31085,-1},
        {31085,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64025,31070,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27523,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31153,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64076,31038,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64082,31117,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64076,31038,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36628,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31153,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64082,31117,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31153,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64025,31070,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31050,-1},
        {31050,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31944,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40779,25998,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64025,31070,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31090,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31153,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31153,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64082,31117,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64025,31070,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33200,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29134,23582,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31050,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31040,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31050,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36330,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31153,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29550,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29550,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64025,31070,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31036,31150,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31071,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40779,25998,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31153,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33836,19975,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33836,19975,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29577,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31177,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31189,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21033,-1},
        {21033,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21033,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26491,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31298,31319,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29314,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22868,-1},
        {22868,-1,-1}
    },
    {
        {153457,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27799,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31296,64084,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31227,-1},
        {31227,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31235,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40653,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31069,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {153513,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {153513,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37019,26654},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {153524,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {153524,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31921,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31255,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31302,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31287,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31256,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31262,-1},
        {31262,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31327,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31302,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26987,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24297,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25480,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39727,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27058,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31206,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27088,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31302,-1},
        {31302,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39727,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40654,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27151,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20860,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33464,34269,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25347,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31309,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31278,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31293,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40654,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31327,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25480,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31278,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31143,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31206,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31329,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31309,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31321,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31206,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31305,31258,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31179,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31335,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31329,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25812,25285,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31975,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31253,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35480,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34214,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31303,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38973,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31179,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31348,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22622,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23383,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34384,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32905,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23617,154068,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36995,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23490,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31429,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23723,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31368,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23720,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23490,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23617,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22435,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21521,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31431,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29237,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29237,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35219,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31429,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31350,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36051,64100,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31395,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32619,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22539,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26368,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24327,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31404,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25513,-1,-1}
    },
    {
        {-1,-1},
        {23472},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31350,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33277,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31404,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31428,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26584,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23508,-1},
        {23508,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31383,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {154052,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31391,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {154068,-1},
        {-1},
        {-1,-1},
        {23617,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31407,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31406,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31350,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31428,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31350,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31418,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31428,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22216,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31432,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31406,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31432,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31428,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31434,31363},
        {31434,31363,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31350,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31434,31363,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31432,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31431,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31434,31363,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20134,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31449,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24453,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22312,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31445,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26395,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22484,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31452,40845,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31452,40845,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31478,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20736,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31461,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31452,40845,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31461,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21830,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36123,-1},
        {36123,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36123,-1},
        {36123,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31637,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31485,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {154339,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {154340,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {154353,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33459,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31637,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21273,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31505,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31722,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30382,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33507,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31534,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31534,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20874,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31685,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31569,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {154353,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31735,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33467,33965,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33391,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31590,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31689,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33391,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31712,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31639,-1},
        {31639,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20856,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {15375,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31661,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31645,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31589,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20874,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20874,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30854,-1,-1}
    },
    {
        {154546,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26685,-1},
        {26685,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31786,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {158941,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31597,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31534,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31600,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31855,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31525,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31828,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34849,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31689,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38827,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31589,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31813,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31598,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38827,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31565,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31728,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31689,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31722,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31565,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29978,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {154699,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31709,-1,-1}
    },
    {
        {154724,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31589,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36890,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31689,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31535,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31709,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38827,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32742,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27602,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31663,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30700,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31777,-1},
        {-1},
        {31777,-1},
        {31777,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31685,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31932,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31722,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31515,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31565,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31745,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31639,31495,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33094,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31661,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31574,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31689,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31731,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31735,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31709,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31685,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32617,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31689,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25516,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31633,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20906,32643,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33094,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31820,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33785,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33094,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27123,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {155041,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31758,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27591,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31598,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27123,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31838,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31932,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38827,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33785,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31589,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31598,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38827,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35264,35251,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31838,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31821,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35264,35251,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31775,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31786,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36463,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31903,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25516,-1,-1}
    },
    {
        {155182,-1},
        {-1},
        {-1,-1},
        {26706,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39145,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31906,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26009,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31235,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26561,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31922,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {155222,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39177,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31903,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31865,-1,-1}
    },
    {
        {155234,-1},
        {-1},
        {-1,-1},
        {32862,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {155237,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39188,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31968,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31896,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38287,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31990,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31875,31189,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28101,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31909,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31875,31189,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30496,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31954,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31922,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31944,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23630,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31944,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {155330,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31903,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31966,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39176,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31966,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {155352,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31983,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {155368,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30058,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38584,38583,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31966,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31966,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31954,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31954,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22064,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31926,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40656,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31875,31189,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31895,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {155427,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37323,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39256,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27591,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31958,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31990,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31959,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31875,31189,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {155484,-1},
        {-1},
        {-1,-1},
        {27368,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31958,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31967,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27591,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31903,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31998,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31998,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32166,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24358,-1},
        {24358,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32199,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32003,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32317,32207,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32305,32294,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32098,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32166,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32049,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32007,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28879,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32080,-1},
        {32080,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32025,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32305,32294,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32166,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32025,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32049,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32321,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38518,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32184,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32000,-1,-1}
    },
    {
        {155604,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32062,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32034,-1,-1}
    },
    {
        {155616,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32140,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32272,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32048,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39673,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32115,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {155643,-1},
        {-1},
        {-1,-1},
        {32121,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32143,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {155660,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32338,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {155671,-1},
        {-1},
        {-1,-1},
        {32340,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32317,32207,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32023,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32089,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32023,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26066,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32025,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32057,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32199,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32340,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25513,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32178,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32340,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27631,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32032,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32153,32380,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32228,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27236,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {155744,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35048,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32154,32396,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32051,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32053,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32300,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32241,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32290,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32131,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32065,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32177,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32305,32294,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32218,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32140,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32032,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32097,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32097,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32321,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32340,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32368,32325,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32194,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32032,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32140,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32098,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32230,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31995,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32199,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32175,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30887,-1,-1}
    },
    {
        {155885,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32300,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32174,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32066,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32346,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32305,32294,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32199,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32302,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32390,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23563,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32290,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32379,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32379,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32368,32325,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32189,-1,-1}
    },
    {
        {32349,-1},
        {-1},
        {-1,-1},
        {32349,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32362,32117,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32053,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32233,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32371,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32309,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38857,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38849,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32377,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32341,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36705,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32346,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32272,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32313,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23494,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32274,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24413,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35065,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32305,32294,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32381,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32365,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32053,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20769,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32274,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32153,32380,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32342,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24413,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24413,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32265,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22793,35722,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32411,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38857,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32218,32171,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32300,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32313,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32382,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36705,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32233,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32065,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32189,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32566,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26734,26479,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27424,32570,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36234,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64056,22120,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29943,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37758,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32574,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32574,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32581,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39729,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39729,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32588,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32589,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22235,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {156272,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32593,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32596,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32597,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36557,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32596,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31908,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {156294,-1},
        {-1},
        {-1,-1},
        {23713,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31537,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32607,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32608,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32607,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23713,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32617,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31637,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20828,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40701,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32615,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32622,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32616,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32616,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32617,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32618,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32645,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64090,32626,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32624,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32622,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32629,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32631,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32615,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32622,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32633,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32637,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32622,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32638,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32622,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32646,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32645,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32642,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31840,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36794,37002,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32648,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32650,-1,-1}
    },
    {
        {156492,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32652,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32652,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24184,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32660,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32954,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29300,32669,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29300,32669,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32670,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32687,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32662,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36764,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32670,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30162,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29300,32669,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22519,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29250,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29548,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35960,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21892,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35480,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30570,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38819,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32660,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33703,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32690,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32690,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33703,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29548,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23348,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32696,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21892,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32660,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32690,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29087,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32666,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32694,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21892,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32666,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38633,-1},
        {38633,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32764,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33310,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38633,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33310,-1,-1}
    },
    {
        {156674,-1},
        {-1},
        {-1,-1},
        {32764,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32705,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32705,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32709,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21675,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32725,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39131,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21451,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21451,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25134,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27809,27794,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26060,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32411,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32411,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36814,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32742,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32748,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32742,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32750,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32761,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32742,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32764,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32762,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28629,28287,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39131,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32771,23389,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32769,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32774,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32774,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22781,23551,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22781,23551,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32772,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38656,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40699,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40700,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32796,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32789,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32796,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31278,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32808,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32808,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24505,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32804,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32804,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32812,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32884,32893,-1}
    },
    {
        {-1,-1},
        {-1},
        {32932,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29809,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32862,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32856,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32862,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32829,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32828,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22779,23167,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21359,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32884,32893,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32854,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32884,32893,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32879,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32842,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32880,32865,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32842,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32850,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27533,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32858,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22779,23167,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32885,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32862,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32879,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29809,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32842,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32884,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32894,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32850,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32885,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32885,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32879,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32885,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30011,30059,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26172,26205,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31899,32901,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31899,32901,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32903,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26360,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26034,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32902,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32943,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32948,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39139,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33242,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32984,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32933,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32989,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32997,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21256,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32996,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32985,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32989,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25993,33229,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32989,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32943,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32948,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32937,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30146,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32983,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33109,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23611,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {157310,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32992,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32954,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24892,24910,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35930,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33016,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32937,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32992,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32982,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36319,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33140,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30516,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33007,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32996,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32996,-1,-1}
    },
    {
        {157360,-1},
        {-1},
        {-1,-1},
        {33031,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33075,33126,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32963,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33011,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33016,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32992,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28825,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32996,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33075,33126,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {17442,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33050,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32983,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33048,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33181,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36349,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20083,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39146,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32948,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {157469,-1},
        {-1},
        {-1,-1},
        {33229,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28825,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32941,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33152,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36767,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {17442,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30384,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33021,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32981,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39512,39486,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33129,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32963,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21693,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33136,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33136,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21563,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32965,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33075,33126,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33075,33126,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32963,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25134,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21897,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30236,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24894,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20989,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33226,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30244,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32996,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33137,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21980,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33145,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39512,39486,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20653,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33034,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33059,21767,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32670,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36975,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33180,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37301,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33229,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33075,33126,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27835,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40822,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33034,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33218,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33075,33126,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33226,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33098,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23273,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33240,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33181,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30320,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33218,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33181,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40656,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33193,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30317,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33203,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33210,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33178,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30185,30246,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37221,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26407,-1},
        {-1},
        {-1,-1},
        {26407,-1,-1}
    },
    {
        {26407,-1},
        {-1},
        {-1,-1},
        {26407,-1,-1}
    },
    {
        {26407,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33162,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33210,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36111,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33176,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33162,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32696,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33248,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37221,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {157917,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26395,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {157930,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23396,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33256,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33256,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33258,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30333,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33258,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24800,24693,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20140,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30343,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38642,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34882,34884,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33260,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30343,-1},
        {30343,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30343,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32066,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24248,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20315,20175,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33267,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33267,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33267,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33268,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33268,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21488,33274},
        {23627,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23627,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21488,33274,17491}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25569,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36618,-1},
        {36618,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {158033,-1},
        {-1},
        {-1,-1},
        {33276,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40786,27503,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22239,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30003,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26355,-1},
        {26355,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33279,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33281,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33277,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36022,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40736,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33279,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28858,29234,-1}
    },
    {
        {158063,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33282,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33285,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33288,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33288,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {19982,33287,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33281,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35201,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35201,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33282,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28976,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30003,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40786,27503,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33279,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33285,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38898,40783,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33282,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25369,25831,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23493,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36786,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36786,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35698,35731,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21930,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33279,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22732,22756,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38015,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28338,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22645,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33296,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21550,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33296,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35449,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {158173,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36781,36766,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33296,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33296,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33296,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25001,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35441,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33293,33294,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33887,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33310,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33321,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26381,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26381,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26381,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38765,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26381,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33324,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33336,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {158238,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35318,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27941,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26389,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31567,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33351,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33367,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26389,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27941,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {158296,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20842,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26381,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27941,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27941,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23221,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26397,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33359,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27941,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {158348,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31567,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27227,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27208,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33323,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33330,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36948,33351},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36948,33351},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {158391,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39472,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35588,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36088,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33330,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33330,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33330,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33393,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30384,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33396,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30388,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38996,38991,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35924,33398,33399}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38996,38991,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38996,38991,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30384,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33653,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21451,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27515,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {158463,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20039,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33653,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33426,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33445,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33433,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33433,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33793,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25369,25831,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33795,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30097,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21451,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33524,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33862,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33967,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27515,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33733,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33805,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33980,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26045,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {33511},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33437,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36428,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33997,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33684,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33707,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33467,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33618,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33967,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27425,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33610,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33407,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33909,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27779,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33789,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33542,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33594,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33426,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34085,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33609,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33785,-1},
        {33785,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34006,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33694,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20633,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33841,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34256,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33865,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33661,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33421,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20234,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33775,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33865,-1,-1}
    },
    {
        {158761,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33967,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33503,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34084,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31452,40845,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33684,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40635,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33940,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31176,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33500,-1},
        {33500,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33940,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33953,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33464,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30626,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33421,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33509,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33509,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33707,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33609,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33891,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20127,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33467,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33914,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33900,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31177,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {158835,-1},
        {-1},
        {-1,-1},
        {33586,-1,-1}
    },
    {
        {158835,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {158835,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33707,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33509,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31452,40845,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20849,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21930,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33467,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33900,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35480,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33503,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33460,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33445,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40635,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {13500,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27935,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33848,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33967,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33448,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34254,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40167,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33524,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {158941,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {158941,-1},
        {-1},
        {-1,-1},
        {158941,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26685,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33984,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20981,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33609,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33985,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33884,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33684,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33543,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29191,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33503,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33509,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22818,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20139,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33900,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24409,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35299,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40654,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36834,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33564,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22825,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33570,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33500,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33539,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33493,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33900,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33503,-1,-1}
    },
    {
        {-1,-1},
        {33761},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {33985},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33833,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33523,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34074,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34183,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34204,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33738,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33564,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33740,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34145,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33544,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34033,33805,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33743,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34171,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34271,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33738,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33592,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33865,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33904,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33560,33620,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34193,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35657,35388,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33912,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33467,33965,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34328,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33909,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33609,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33893,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33707,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34069,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28145,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33694,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33940,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34001,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34065,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34065,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27602,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33745,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34588,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33900,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34142,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34045,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33742,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26149,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33688,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34000,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33405,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {159296,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {159296,-1},
        {-1},
        {-1,-1},
        {159296,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33678,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33542,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34822,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33660,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27533,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36896,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {159333,-1},
        {-1},
        {-1,-1},
        {17731,34093,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35657,35388,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33804,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32772,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33897,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33984,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33694,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33581,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33776,33533,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34295,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33704,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33848,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34191,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38512,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33807,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33809,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33871,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24800,24693,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34180,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33748,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33624,33674,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31776,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33771,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33802,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33564,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33778,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33796,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33809,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34084,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33894,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34184,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34047,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34072,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33785,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31327,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31838,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34133,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39498,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40167,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31253,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33984,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23420,23421,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34098,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33807,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33999,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33849,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32772,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34069,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27035,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34067,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33696,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33581,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25545,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33900,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33871,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33901,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33862,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34253,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {159636,-1},
        {-1},
        {-1,-1},
        {33290,159636,26087}
    },
    {
        {159636,-1},
        {-1},
        {-1,-1},
        {33290,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34254,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33953,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33737,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33787,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34345,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34044,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33770,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33900,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27035,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34109,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34109,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27189,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32772,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33785,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33945,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33983,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33988,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34028,-1,-1}
    },
    {
        {159734,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {159735,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {159736,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33745,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34360,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23647,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34069,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20693,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34203,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33785,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31242,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34328,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33988,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33893,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35657,35388,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35657,35388,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33509,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22818,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34299,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34299,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31776,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34033,33805,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33988,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34085,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34113,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34256,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33695,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33807,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34212,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34147,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34214,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39347,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32772,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34328,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33785,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34078,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34146,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33900,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34254,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34152,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33738,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33809,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34115,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34203,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38897,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34299,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33742,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34254,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38618,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34249,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29136,31910,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34369,-1,-1}
    },
    {
        {159988,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {159988,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34196,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39498,-1,-1}
    },
    {
        {160013,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {160013,-1},
        {-1},
        {-1,-1},
        {160013,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34343,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33785,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34265,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34306,34122,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34214,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27189,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29138,28988,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34299,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39498,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34345,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {160057,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {160057,-1},
        {-1},
        {-1,-1},
        {160057,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34193,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34299,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33775,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34285,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33777,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34328,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34268,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34285,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34265,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32858,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34368,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31277,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34850,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34368,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34120,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34265,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28982,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27489,27475,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34384,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34382,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34384,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34384,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34390,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34387,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34384,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34394,34395,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24426,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34381,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34382,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20977,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35419,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34389,20966,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28888,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34407,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38728,38666,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38712,35207,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33759,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39253,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {160766,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34468,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34473,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34425,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34425,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34411,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34453,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34426,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34423,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27602,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34473,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34453,34870,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34576,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34537,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34753,34442,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34503,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34444,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34588,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34601,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34537,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34612,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34532,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34562,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34540,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35960,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34695,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34568,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34873,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34483,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34664,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34678,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34678,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34513,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34849,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34540,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34519,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34537,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34641,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34641,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34659,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34641,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34415,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34609,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34664,34417,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34673,34427,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34597,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34453,34870,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34540,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34425,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34467,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34752,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34516,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34553,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34661,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34822,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34453,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34624,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34425,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34811,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34597,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34662,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34636,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34444,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34468,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34597,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34787,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34661,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34483,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34731,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33684,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34442,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34502,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34540,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34483,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {160730,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {160731,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34832,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34837,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21988,-1},
        {21988,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {160766,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {160766,-1},
        {-1},
        {-1,-1},
        {160766,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34480,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34645,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34502,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34584,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34483,-1,-1}
    },
    {
        {160784,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34661,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34474,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34731,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {17893,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34749,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {160841,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24851,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34799,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34784,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34749,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34567,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34837,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34453,34870,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34567,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20096,40860,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34749,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {17848,34558,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34563,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34622,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34661,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34655,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34749,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34664,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34638,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34735,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34756,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34588,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34807,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34832,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34722,34509,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34609,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34750,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34619,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34784,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34829,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34789,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34873,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34622,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34850,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34442,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34588,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34648,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34541,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34822,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34753,34442,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34661,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34837,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34829,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34445,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34866,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34822,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34661,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34445,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34822,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34719,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34822,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34695,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34678,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34732,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34661,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34829,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34569,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34588,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34811,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34850,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34453,34870,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34588,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34445,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34795,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34873,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24375,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34701,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34445,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34875,34542,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34566,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34569,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34797,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40035,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34732,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34884,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27424,32570,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22259,22294,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33032,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30431,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27472,27431,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30431,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30435,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39188,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34884,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36556,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34911,-1},
        {36947,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36947,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34893,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34913,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24448,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27491,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24494,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24838,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34899,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34898,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29575,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21106,-1,-1}
    },
    {
        {-1,-1},
        {34899},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24838,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31142,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29575,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34909,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34909,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23566,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34914,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35028,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35028,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {161300,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {161301,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34920,-1},
        {34920,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34927,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35006,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34920,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34942,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35028,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34920,-1,-1}
    },
    {
        {161329,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25937,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31072,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21482,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34957,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34976,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34966,35023,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34942,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31056,64079,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34957,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35013,35037,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34941,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35010,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35010,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64082,31117,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {64068,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {161412,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35013,35037,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35032,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {161427,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31227,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35115,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35009,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {161571,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35056,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35090,35139,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35010,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35069,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35137,32358,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32323,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35127,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31259,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24132,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35120,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24125,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34968,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35090,35139,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35167,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35027,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34952,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22218,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35028,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35033,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35028,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25993,40778,-1}
    },
    {
        {161550,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35028,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {161571,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34920,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35090,35139,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35079,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34928,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35174,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {161618,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20608,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35120,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64096,35088,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35064,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34980,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64096,35088,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35090,35139,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35137,32358,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35140,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35140,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34920,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34957,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34920,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35140,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35120,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35137,32358,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35145,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27662,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34923,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {18021,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35186,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35186,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35203,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29645,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34407,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35207,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38712,35207,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24218,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35203,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28883,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32631,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32854,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35206,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32647,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35211,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20816,20818,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35386,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35248,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33703,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35258,35226,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21914,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30554,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30568,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35242,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38758,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36084,38972,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35242,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38322,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30629,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30554,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35250,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24615,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35282,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35320,35302,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35318,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35320,35302,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35293,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35293,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35293,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35318,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20821,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35327,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20821,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35313,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24413,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38824,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35327,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25681,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {63856,27578,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35313,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35327,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35313,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35320,35302,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35422,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35477,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35463,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35347,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35405,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22827,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35433,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35350,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35672,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35347,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21612,-1,-1}
    },
    {
        {162181,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35511,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35398,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35386,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35380,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20318,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35584,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35409,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35409,35345,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24321,36783,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35410,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35363,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35386,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35563,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23490,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21698,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35578,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24676,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35440,35493,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35349,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22916,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35440,35493,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35359,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35400,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35349,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35398,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37682,37636,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35463,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35516,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25945,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21983,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21621,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39721,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21892,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38728,38666,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35516,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35458,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35365,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35610,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30435,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35548,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35398,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35672,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35427,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35607,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26395,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35463,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35589,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35426,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35548,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35535,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35575,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35740,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35390,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35400,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35380,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24321,36783,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35571,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35409,35345,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35535,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35535,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35524,-1},
        {-1},
        {-1,-1},
        {35524,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30435,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35668,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35409,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35426,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {162436,-1},
        {-1},
        {-1,-1},
        {35352,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35382,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36776,24321,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35489,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35380,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35591,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24687,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32327,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35426,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35473,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35542,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35611,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35380,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35558,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36895,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35571,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35609,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35646,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35604,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35676,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21983,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35550,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35613,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30330,30332,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35627,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35646,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35390,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35672,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35554,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35591,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35382,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35379,35695,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35586,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35628,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25812,25285,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35386,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35465,35709,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35696,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35676,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35702,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20889,23531,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35740,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35641,64099,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35524,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35678,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35576,64034,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35524,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24321,36783,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24321,36783,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35724,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35591,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35649,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35591,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35732,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22123,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35554,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21892,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21927,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35698,35731,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35723,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35897,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35901,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {13667,-1,-1}
    },
    {
        {162739,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22737,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {162750,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35905,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35905,-1,-1}
    },
    {
        {162759,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38553,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35898,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28567,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28567,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35898,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35910,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22772,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35910,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35916,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21370,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35920,35914,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33793,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35916,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21370,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22189,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35914,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35925,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35925,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35930,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29066,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35937,-1},
        {-1},
        {-1,-1},
        {35937,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20821,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20821,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35927,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35930,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34849,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35930,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35930,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40179,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35930,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32900,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35945,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34214,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26118,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29392,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35973,-1,-1}
    },
    {
        {163000,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29399,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35960,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29548,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35980,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35980,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35986,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24409,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29480,29503,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29482,35948,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35986,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29431,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31163,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29432,35981,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29562,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36008,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29795,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29795,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24471,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36012,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29795,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36054,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36031,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36001,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36022,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32829,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36016,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36049,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36031,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36031,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36040,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36070,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36040,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38914,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36001,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36012,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36042,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36008,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36031,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36023,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23521,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36036,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26211,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36051,64100,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32028,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36012,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36051,64100,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32972,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36016,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36058,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21830,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36031,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {163232,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23494,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36070,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36068,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36092,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36020,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36077,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36009,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36094,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32588,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36031,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36060,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29723,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36101,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {163232,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22770,36067,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40782,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36039,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36058,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {163232,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21830,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36115,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36094,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22770,36067,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34146,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36070,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22770,36067,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36088,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24163,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36986,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36001,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36101,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {163232,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36092,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36020,-1,-1}
    },
    {
        {163344,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36104,64101,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36103,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36090,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36101,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36986,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36109,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36986,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36986,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36123,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36118,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28931,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36204,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36204,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36199,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29001,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29118,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29118,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29128,28783,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36208,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36211,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24466,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {36211},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36259,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36234,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36234,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36259,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36330,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36264,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36264,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36896,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20166,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36362,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36259,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26469,20358,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36264,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36259,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21467,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36953,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33275,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36952,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25718,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36997,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36491,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22868,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36271,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22868,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36271,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {163767,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27491,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23621,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20225,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36331,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38494,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36452,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22868,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36311,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36302,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36328,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36328,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36391,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36466,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36427,36367,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36314,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20225,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25308,25309,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36451,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36339,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36391,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36282,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36319,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38494,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36463,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36328,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36311,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22868,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23651,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36448,36310,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36424,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33051,-1,-1}
    },
    {
        {163978,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {36491},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23653,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22799,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36420,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {164027,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36362,36404,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36328,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36339,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36351,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36451,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22868,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36475,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36485,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21264,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33128,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36348,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21264,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36420,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36400,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36470,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36450,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36330,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36490,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36374,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36463,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36367,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36481,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36313,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24053,21358,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36463,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36499,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36427,36367,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37389,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36463,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36374,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36476,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36522,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36432,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36425,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23653,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36470,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36500,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36423,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36332,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36470,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36470,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36468,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39002,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29273,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21413,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21413,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32930,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32828,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29200,28809,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26381,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25105,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33213,32966,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23556,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32930,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36544,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32828,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36531,36524,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36315,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32850,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33258,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23381,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32856,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23001,22922,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36468,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35978,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23556,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38754,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20307,39636,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32856,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {164471,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31406,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36575,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32856,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32779,-1,-1}
    },
    {
        {164482,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27138,27005,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32889,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29544,29420,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32889,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23338,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36544,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37357,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20094,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38515,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36628,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36703,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36587,-1},
        {36587,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36587,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30824,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25282,25173,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20347,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {164595,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21507,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26017,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31441,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36600,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36631,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36617,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36678,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36631,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36605,36629,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36639,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36618,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25582,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36562,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36674,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36636,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36706,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36602,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36646,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36686,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36695,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36618,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36636,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36608,36700,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36613,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31735,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36685,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {164813,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34913,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36692,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36618,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36678,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36554,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36696,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36618,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36685,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36587,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32696,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36706,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36705,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36616,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36707,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {164876,-1},
        {-1},
        {-1,-1},
        {36763,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33688,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26032,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33688,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36771,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36781,36766,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36781,36766,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31480,31478,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36776,24321,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36764,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33288,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35702,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36784,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36784,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36786,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34483,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36786,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36556,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36556,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36804,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24466,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24466,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36215,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36215,-1,-1}
    },
    {
        {164949,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {164949,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36802,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36813,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24466,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24467,24478,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24464,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36864,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36804,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36805,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24467,24478,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24478,24467,20174}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24467,24478,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36978,36933,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24449,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24449,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36932,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36857,36321,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36999,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24460,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36857,36321,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24674,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36867,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36855,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21330,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36896,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36983,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24237,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35980,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36866,-1},
        {36866,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36794,37002,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26820,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36938,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24448,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36867,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36876,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36794,37002,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36890,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36883,36958},
        {36883,36958,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36864,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36864,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36953,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36957,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36867,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36867,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36865,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36991,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36605,36629,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36930,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24448,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36942,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36910,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27512,24112,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28165,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38564,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35536,24838,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36883,36958,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36842,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21205,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21205,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36947,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36939,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36864,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36864,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36948,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36948,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36865,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36983,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36930,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36930,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36212,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22889,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36883,36958,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36996,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36984,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27941,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36875,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36983,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24257,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36896,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21450,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36947,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36947,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27941,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36865,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36865,-1,-1}
    },
    {
        {165227,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {165227,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {165227,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36949,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36865,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36947,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36932,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36953,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25105,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38015,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36996,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24487,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27579,27580,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36974,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22577,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36948,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36960,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36969,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24190,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36914,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {165320,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {165321,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {165321,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24499,24503,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36973,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38662,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36947,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36986,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36981,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36875,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36983,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24962,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24473,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36986,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36986,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36986,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36968,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36983,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36910,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24118,24111,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36953,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25787,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36981,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36899,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25847,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28382,28399,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36973,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21407,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36794,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36794,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36996,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36646,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36917,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36883,36958,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36975,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22889,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36975,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38564,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37000,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21297,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36983,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36983,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36983,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36914,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36914,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36914,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36973,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40440,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37032,26449,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24055,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37027,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37090,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37048,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37041,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37027,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37030,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37039,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37092,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37064,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37041,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {37057},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32774,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37057,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21034,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37030,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37138,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37138,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37041,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {165591,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24055,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37030,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37089,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37084,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37117,64038,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {165626,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32858,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37117,64038,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25959,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37084,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24055,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37109,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {37165},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37192,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24676,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37092,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21375,-1},
        {21375,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37109,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37064,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37063,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21830,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25160,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40654,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37145,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37070,37086,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37165,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37168,38563,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40654,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37109,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37119,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37124,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24055,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24055,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37168,38563,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32858,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37111,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21139,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37148,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37101,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30952,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37117,64038,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37148,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37122,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37167,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37192,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24676,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37253,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28238,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37292,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37264,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37206,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37264,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37202,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26015,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37202,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35290,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35290,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37261,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31893,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37207,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31967,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37253,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37292,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22039,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39154,39150,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37297,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37292,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39154,39150,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37264,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37237,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39135,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21980,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21307,37291,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37228,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37236,-1},
        {37236,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37221,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31085,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28921,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37218,37259,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22633,40573,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37294,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37237,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37299,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37255,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37255,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31967,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37221,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37306,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31967,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25773,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28858,29234,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30519,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24059,21367,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30519,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31966,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20809,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37325,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37327,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37327,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37328,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37328,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37645,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37329,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37329,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37329,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {166214,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {166217,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26491,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37338,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37489,30970,37979}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37406,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37489,37979,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30683,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37392,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {166251,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37805,37806,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37451,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {166279,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {166280,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37502,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37586,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37613,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37840,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37609,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37805,37806,-1}
    },
    {
        {166330,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {166331,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37504,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37517,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35657,35388,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37448,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37745,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {37795},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {166430,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {166441,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35657,35388,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35657,35388,-1}
    },
    {
        {166467,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37860,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {166513,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {166513,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37738,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37855,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35657,35388,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29748,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {166553,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37656,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37728,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {166605,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37517,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37942,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37444,37941,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37716,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37713,-1,-1}
    },
    {
        {166621,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {166628,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29748,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37656,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37877,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39730,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37470,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37745,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37841,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26657,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37444,37941,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39156,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37656,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37913,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37982,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {145180,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37904,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37786,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37987,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26034,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37619,37956,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37444,37941,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37841,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37756,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37619,37956,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {166849,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37962,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37662,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {166895,-1},
        {-1},
        {-1,-1},
        {37971,-1,-1}
    },
    {
        {166895,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37439,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22612,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37858,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37530,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38015,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37984,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37907,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37907,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37907,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37969,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37908,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38263,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38263,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39686,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39654,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39654,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39654,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39658,39662,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36335,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39683,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39692,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32854,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39704,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39661,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39675,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32854,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39661,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37855,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39704,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39714,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39704,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38310,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38287,-1},
        {-1},
        {38287,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38291,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38291,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38370,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38306,38364,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64079,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25155,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38369,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38283,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {167184,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38370,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26613,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38310,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38369,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38310,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38310,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24739,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38310,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38357,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38346,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38331,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38331,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38283,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38370,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38357,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39726,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30465,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38334,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38480,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {167281,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38356,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38326,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38283,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38346,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38306,38364,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38360,39725,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22654,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38366,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35518,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38283,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38364,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38357,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24283,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38326,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22654,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37931,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38000,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38313,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {167419,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38428,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24070,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38446,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32026,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25237,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27663,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {167455,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25913,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38520,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38515,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24101,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {167478,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30428,30423,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38525,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38512,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38518,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38464,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38477,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38464,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24207,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38450,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29987,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38475,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {63964,38534,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38488,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38517,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38480,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36491,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36491,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38491,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23713,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {63964,38534,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38501,38519,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38518,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38488,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {167561,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20094,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30028,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38536,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22684,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30238,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38576,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30238,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22320,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38506,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38512,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38491,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {63964,38534,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38517,-1,-1}
    },
    {
        {167659,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38515,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23849,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32581,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38553,-1},
        {-1},
        {-1,-1},
        {38553,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38576,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38552,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20670,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38576,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38576,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38560,38577,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37192,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38568,38543,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38512,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38525,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38525,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22549,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {167730,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22684,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38508,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38508,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37168,38563,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22684,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38520,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28657,27996,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38512,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38552,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38564,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38564,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38567,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38552,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38520,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38520,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38552,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29159,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30011,30059,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32900,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32900,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38584,38583,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38584,38583,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40178,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40251,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40178,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24331,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38587,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40199,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40194,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40199,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21806,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38614,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40361,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40289,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40223,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40274,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40260,40223},
        {40223,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21806,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38619,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40285,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40273,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40287,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38601,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21806,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40329,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40314,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40299,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40356,40318,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40324,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40307,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40441,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40303,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40330,-1,-1}
    },
    {
        {167928,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38614,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {18948,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40346,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38626,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40274,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35726,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40393,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40386,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38617,21452,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40364,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40391,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64104,38627,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64104,38627,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40415,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38617,21452},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40429,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40434,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40419,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40306,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40455,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {19719,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40442,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40461,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38610,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64104,38627,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40329,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64104,38627,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64104,38627,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36909,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40469,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64104,38627,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24059,21367,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64104,38627,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64104,38627,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38633,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38632,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38512,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38656,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38492,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38633,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38721,38647,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38651,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34389,20966,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38728,38666},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38651,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28460,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38704,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38651,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38512,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38704,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38646,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38717,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38651,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38642,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38726,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38718,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27675,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38651,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38728,38666,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38717,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38649,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30362,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38721,38647,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38712,35207,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38647,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38704,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38662,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31417,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40599,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34048,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38651,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38651,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38692,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38649,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35208,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38649,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35208,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38704,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38726,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38646,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40855,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40855,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38663,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38646,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38649,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38704,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38728,38666,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38718,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28732,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32647,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38691,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38691,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38649,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38663,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38721,38647,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22825,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22825,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21359,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21359,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21359,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29791,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24107,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33510,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25105,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20849,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32784,-1},
        {32784,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30383,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38758,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30394,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38760,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38760,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38758,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24904,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40723,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40723,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40723,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32051,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20881,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40723,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38788,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33032,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27356,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {168608,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30394,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29417,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {168625,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40727,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38799,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38876,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25787,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20500,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38818,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38795,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32002,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36562,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38788,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38805,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25787,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38836,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38835,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38814,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31833,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35250,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40727,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38836,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38788,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38772,38846,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38780,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38836,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38824,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20500,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38816,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38772,38846,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27656,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38788,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27356,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22240,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38834,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38875,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38872,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38875,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38833,38876,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38834,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38867,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36676,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38875,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38881,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38881,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34212,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23395,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40783,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33258,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22909,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23395,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38911,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38911,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35543,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38911,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38911,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38913,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39658,39662,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38948,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39658,39662,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32993,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38935,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39661,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26045,-1},
        {26045,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38914,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {169104,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30518,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38955,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25114,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36529,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38968,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38998,22239,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38915,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36084,38972,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29229,38990},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64106,38971,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38994,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38924,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64106,38971,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38915,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38960,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38997,39023,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39025,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24962,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23653,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38928,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64106,38971,28693}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38996,38991,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38914,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38990,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38914,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20762,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39000,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39000,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38990,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30317,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36116,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39318,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39024,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39110,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39110,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39083,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22799,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39088,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39087,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21013,39111},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39110,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24070,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {169423,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34426,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39096,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22852,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36938,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39108,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39100,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39106,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38973,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20794,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20794,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39108,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39108,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39131,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32752,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32752,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32764,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {169599,-1},
        {-1},
        {-1,-1},
        {39135,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39135,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39135,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39135,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39164,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39164,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39145,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39166,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39138,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39143,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39253,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39154,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39154,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40616,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39154,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39180,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39151,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35916,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64081,31069,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31890,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39147,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39165,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39246,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39165,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39165,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39156,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39165,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39165,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39180,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39217,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39178,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32946,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39156,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {169712,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39219,31958,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39246,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35916,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39197,39166,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39165,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39165,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {169753,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {169753,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39176,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39253,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39165,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39244,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39233,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {169808,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31959,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40779,25998,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39165,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28479,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39246,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39165,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39165,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28479,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39253,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39263,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39253,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39241,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39260,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39253,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39892,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39212,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39226,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28893,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39226,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39218,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39141,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39318,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38914,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39658,39662,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39658,39662,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38960,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39658,39662,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38948,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38948,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38948,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38960,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38996,38991,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38931,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39318,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39320,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39321,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39326,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39340,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39353,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39357,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39394,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22766,22767,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32629,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39522,39359,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39356,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39394,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39394,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39381,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39384,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39493,39366,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39387,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39425,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39361,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {170182,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39493,39366,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39470,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39493,39366,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39492,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39492,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28084,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39389,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39478,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39368,39366,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39498,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38994,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39446,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39456,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39416,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39394,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39473,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39416,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39525,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39426,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39470,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29987,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39472,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39491,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39472,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34398,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39634,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39704,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {39524},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32666,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39525,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39512,39486,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39409,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39384,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39491,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36703,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39472,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39491,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39425,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39512,39486,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39522,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32908,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26487,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36290,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39609,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32938,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32965,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31391,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32930,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31391,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39599,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33075,33126,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38967,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38967,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33051,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39616,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39616,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39635,39620,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {170610,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39260,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21202,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39635,39620,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39616,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20307,39636,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20676,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36544,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39635,39620,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33173,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37825,37782,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39630,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33210,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39635,39620,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39028,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23408,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22665,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22612,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35946,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37101,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35946,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35946,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35946,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39646,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22435,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22478,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37101,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22581,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21338,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27155,27004,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22665,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22210,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39654,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39658,39662,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39704,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39650,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32902,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39692,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39704,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39673,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32902,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39708,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39692,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39654,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39708,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37855,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39704,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39704,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39702,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39715,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39702,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27656,27658,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39714,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39715,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20014,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38360,39725,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38360,39725,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39720,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39720,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29748,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39721,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38360,39725,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31944,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29237,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31212,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29237,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29237,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29237,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29237,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29237,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37707,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31946,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39730,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27832,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39730,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28921,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27832,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29100,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39730,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39191,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31909,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27832,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40779,25998,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32697,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28107,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29969,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64072,29038,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28729,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40779,25998,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28729,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39741,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39749,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39749,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39741,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39741,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39741,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39745,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32772,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39745,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39740,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39757,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39749,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39740,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39740,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39741,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39740,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39749,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39749,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39745,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39758,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {19488,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39740,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37276,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {19488,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39749,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34394,34395,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39740,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39770,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31838,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39787,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39782,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39973,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39791,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22238,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39899,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39971,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39852,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {40032},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34444,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {171483,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39796,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39826,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39848,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39826,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39818,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39782,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {171541,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {171581,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {171581,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {171581,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {171593,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39873,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39822,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39892,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39827,-1,-1}
    },
    {
        {171658,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {171658,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {171658,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39969,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39975,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39782,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39995,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39826,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39978,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39878,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {171716,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39908,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40055,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {171739,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {171739,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39995,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39995,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39846,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39846,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39892,-1,-1}
    },
    {
        {171753,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39973,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39973,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39912,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39881,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40053,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {160784,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22039,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39881,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34809,34831,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39986,40711,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40023,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28417,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28417,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39827,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37218,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40035,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39908,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39827,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40035,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40039,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40015,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28937,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40165,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26783,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26783,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40167,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40169,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38592,23948,23798}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40199,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26783,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40379,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40379,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40200,38593,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38604,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40379,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40434,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40198,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40379,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40257,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38596,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40379,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40178,-1},
        {38599,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26783,-1,-1}
    },
    {
        {171982,-1},
        {-1},
        {-1,-1},
        {40201,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24973,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40223,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40461,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40223,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38619,40373,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40232,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40199,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40311,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40221,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40232,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40232,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40274,-1,-1}
    },
    {
        {172058,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40201,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40198,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40258,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38619,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {172079,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {172079,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40257,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40221,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40232,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40751,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40194,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40203,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40232,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38601,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38601,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {172058,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {172162,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40303,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40213,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40329,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40272,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40297,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40327,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40324,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40378,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {18948,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40199,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40442,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40182,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40274,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {172281,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {172281,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40455,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40304,38613,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38605,38621,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40455,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38619,40373,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38601,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40477,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40410,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40251,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40377,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27653,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40422,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40422,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25847,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40179,-1,-1}
    },
    {
        {172432,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40423,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39015,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40434,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40379,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40442,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40329,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40383,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40385,30435},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40455,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40442,-1},
        {40442,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64104,38627,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40441,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40440,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40469,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40210,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40391,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40474,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40475,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40372,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40179,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40477,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23660,23646,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40565,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35203,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25509,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35203,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22633,40573,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22633,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35203,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40569,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35203,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40594,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34382,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34382,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34920,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40576,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34920,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29277,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39420,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40576,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34382,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {40599,-1},
        {40599,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40605,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32666,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40607,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40607,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22645,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40628,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33426,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33426,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39141,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27377,34326,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29105,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29100,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40616,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31892,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40628,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37168,38563,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40628,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31954,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40616,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40616,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31946,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40628,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31954,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39226,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39237,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39238,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33678,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29100,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39226,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {172940,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38753,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31964,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25705,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40638,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40638,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26332,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36637,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29066,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29004,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40652,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40654,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39321,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31946,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31946,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39321,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31946,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40656,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39336,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40657,40658,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32940,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40657,40658,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20022,-1,-1}
    },
    {
        {173111,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40692,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28593,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28593,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40679,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40679,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40694,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28593,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40679,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40701,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40703,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34415,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34415,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40701,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40701,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40701,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40707,19831,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34563,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40706,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40706,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40711,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40711,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40711,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31179,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40718,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40718,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40718,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21729,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {18613,37142,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {18613,37142,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40723,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40723,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40727,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40725,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40725,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40723,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35970,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40748,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40751,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40765,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40765,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22164,22159,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21310,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22164,22159,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40774,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21957,40773,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33229,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33229,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36039,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40779,25998,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40807,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40789,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40789,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40788,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40809,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40792,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40792,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40788,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40797,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40791,-1,-1}
    },
    {
        {173553,-1},
        {-1},
        {-1,-1},
        {40799,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40801,40802,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40797,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {173570,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40806,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40807,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40791,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40810,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40807,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {173594,-1},
        {-1},
        {-1,-1},
        {40812,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40810,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40806,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40810,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40822,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40813,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40822,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40813,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40799,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40821,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20999,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40797,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40807,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33296,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40816,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40799,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33296,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40853,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40853,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31840,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31840,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40855,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29903,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20096,40860,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20096,40860,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20096,40860,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {173746,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31179,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35997,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20096,40860,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31179,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21809,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31722,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21561,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21561,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31722,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31787,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31787,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {19975,-1,-1}
    },
    {
        {20116,-1},
        {-1},
        {-1,-1},
        {20116,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27664,-1,-1}
    },
    {
        {19985,-1},
        {-1},
        {-1,-1},
        {19985,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20025,-1},
        {-1},
        {-1,-1},
        {20025,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28858,29234},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21359,-1,-1}
    },
    {
        {19988,-1},
        {-1},
        {-1,-1},
        {19988,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28879,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22839,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25152,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25152,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20124,20126,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25152,-1,-1}
    },
    {
        {30452,-1},
        {-1},
        {-1,-1},
        {30452,-1,-1}
    },
    {
        {30452,-1},
        {-1},
        {-1,-1},
        {30452,-1,-1}
    },
    {
        {30452,-1},
        {-1},
        {-1,-1},
        {30452,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28961,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28961,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28961,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30452,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21281,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20063,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20063,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20037,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20039,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25300,25244},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25300,25244},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20055,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31177,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20129,-1,-1}
    },
    {
        {27668,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20806,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20806,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20102,-1,-1}
    },
    {
        {20104,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26007,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24351,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20110,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20116,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24178,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20883,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20127,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20130,-1},
        {-1},
        {-1,-1},
        {20130,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27704,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20133,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20805,20133,-1}
    },
    {
        {20136,-1},
        {-1},
        {-1,-1},
        {20136,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20136,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30452,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20055,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20139,-1},
        {-1},
        {-1,-1},
        {20139,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20139,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33279,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20056,20055,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20129,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24478,24467,20174}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20170,-1,-1}
    },
    {
        {20196,-1},
        {-1},
        {-1,-1},
        {20196,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29255,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {20133},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20302,-1,-1}
    },
    {
        {20183,-1},
        {-1},
        {-1,-1},
        {20183,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29916,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20329,-1},
        {-1},
        {-1,-1},
        {20329,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24441,-1},
        {24441,-1,-1}
    },
    {
        {20385,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20374,-1},
        {-1},
        {-1,-1},
        {20374,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20241,-1,-1}
    },
    {
        {38585,-1},
        {-1},
        {-1,-1},
        {38585,20339,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20399,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20842,-1,-1}
    },
    {
        {20374,-1},
        {-1},
        {-1,-1},
        {20374,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39131,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20489,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20633,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20842,-1},
        {-1},
        {20842,-1},
        {20842,-1,-1}
    },
    {
        {26781,-1},
        {-1},
        {-1,-1},
        {26465,26781,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20681,-1},
        {-1},
        {-1,-1},
        {20681,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20842,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20572,-1},
        {-1},
        {-1,-1},
        {20572,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20689,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20860,-1,-1}
    },
    {
        {20581,-1},
        {-1},
        {-1,-1},
        {20581,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20681,-1},
        {-1},
        {-1,-1},
        {20681,-1,-1}
    },
    {
        {20744,-1},
        {-1},
        {-1,-1},
        {20744,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20670,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20778,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {132089,-1},
        {-1},
        {-1,-1},
        {132089,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20756,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20796,-1},
        {-1},
        {-1,-1},
        {20796,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20796,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20801,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20812,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20805,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20812,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20812,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31786,-1,-1}
    },
    {
        {20807,-1},
        {-1},
        {-1,-1},
        {20807,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38263,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20820,20814,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20806,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20820,20814,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27610,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27610,-1,-1}
    },
    {
        {20130,-1},
        {-1},
        {-1,-1},
        {20130,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23664,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20130,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22833,-1,-1}
    },
    {
        {20856,-1},
        {-1},
        {-1,-1},
        {20856,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {19982,33287,-1}
    },
    {
        {20860,-1},
        {-1},
        {-1,-1},
        {20860,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20860,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20860,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20860,-1,-1}
    },
    {
        {20860,-1},
        {-1},
        {-1,-1},
        {20860,-1,-1}
    },
    {
        {30494,-1},
        {-1},
        {-1,-1},
        {30494,-1,-1}
    },
    {
        {30495,-1},
        {-1},
        {-1,-1},
        {30495,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20860,-1},
        {20860,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20860,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20860,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22864,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20860,-1},
        {20860,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29245,-1,-1}
    },
    {
        {20864,-1},
        {-1},
        {-1,-1},
        {20864,-1,-1}
    },
    {
        {20869,-1},
        {-1},
        {-1,-1},
        {20869,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20873,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20874,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20874,-1,-1}
    },
    {
        {20877,-1},
        {-1},
        {-1,-1},
        {20877,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20877,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20841,20001,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20873,-1,-1}
    },
    {
        {20877,-1},
        {-1},
        {-1,-1},
        {20877,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20883,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20883,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20883,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20883,-1,-1}
    },
    {
        {20883,-1},
        {-1},
        {-1,-1},
        {20883,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20883,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20883,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20883,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {14400,-1,-1}
    },
    {
        {13500,-1},
        {-1},
        {-1,-1},
        {13500,-1,-1}
    },
    {
        {20888,-1},
        {-1},
        {-1,-1},
        {20888,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21253,-1,-1}
    },
    {
        {20893,-1},
        {-1},
        {-1,-1},
        {23452,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20896,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20901,-1,-1}
    },
    {
        {20901,-1},
        {-1},
        {-1,-1},
        {20901,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20901,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20901,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23500,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20889,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25010,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20973,-1,-1}
    },
    {
        {20828,-1},
        {-1},
        {-1,-1},
        {20828,-1,-1}
    },
    {
        {20982,-1},
        {-1},
        {-1,-1},
        {20982,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20982,-1},
        {20982,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21000,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21000,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21060,-1},
        {21060,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21021,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21021,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21069,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21046,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21046,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21127,-1,-1}
    },
    {
        {21129,-1},
        {-1},
        {-1,-1},
        {21129,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21129,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20998,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20999,21151,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21161,-1},
        {-1},
        {-1,-1},
        {21161,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21218,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21191,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22839,-1,-1}
    },
    {
        {21187,-1},
        {-1},
        {-1,-1},
        {21187,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21205,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21218,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21240,21223,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21240,21223,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21256,-1},
        {-1},
        {-1,-1},
        {21256,-1,-1}
    },
    {
        {21256,-1},
        {-1},
        {-1,-1},
        {21256,-1,-1}
    },
    {
        {21259,-1},
        {-1},
        {-1,-1},
        {21259,-1,-1}
    },
    {
        {21259,-1},
        {-1},
        {-1,-1},
        {21259,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21259,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21361,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24059,21367,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24046,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21281,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24319,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33307,21319,-1}
    },
    {
        {21317,-1},
        {-1},
        {-1,-1},
        {21317,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21329,64053,-1}
    },
    {
        {-1,-1},
        {26097},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21329,64053,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21329,64053,-1}
    },
    {
        {24650,-1},
        {-1},
        {-1,-1},
        {24650,21332,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29575,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29575,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29575,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21353,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21360,-1,-1}
    },
    {
        {21359,-1},
        {-1},
        {-1,-1},
        {21359,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21359,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37111,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40577,40612,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21402,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37328,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37328,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21413,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37328,-1,-1}
    },
    {
        {21413,-1},
        {-1},
        {-1,-1},
        {21413,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21413,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21413,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24296,-1,-1}
    },
    {
        {20197,-1},
        {-1},
        {-1,-1},
        {20197,-1,-1}
    },
    {
        {20197,-1},
        {-1},
        {-1,-1},
        {20197,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21533,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21442,21443,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21442,21443,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21453,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25300,25244},
        {-1,-1,-1}
    },
    {
        {21460,-1},
        {-1},
        {-1,-1},
        {21460,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21466,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21471,-1},
        {-1,-1,-1}
    },
    {
        {21467,-1},
        {-1},
        {-1,-1},
        {21467,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21467,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21495,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21566,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21533,-1,-1}
    },
    {
        {21646,-1},
        {-1},
        {-1,-1},
        {21646,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21629,-1,-1}
    },
    {
        {21646,-1},
        {-1},
        {-1,-1},
        {21646,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21547,-1,-1}
    },
    {
        {21547,-1},
        {-1},
        {-1,-1},
        {21547,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21692,-1},
        {-1},
        {-1,-1},
        {21692,-1,-1}
    },
    {
        {21692,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21547,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21776,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21843,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21696,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21959,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22120,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21831,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21830,-1},
        {-1},
        {-1,-1},
        {21830,-1,-1}
    },
    {
        {21934,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21930,-1},
        {-1},
        {-1,-1},
        {21930,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24613,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21495,34399,-1}
    },
    {
        {21916,-1},
        {-1},
        {-1,-1},
        {21916,-1,-1}
    },
    {
        {21930,-1},
        {-1},
        {-1,-1},
        {21930,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21934,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21312,21306,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21983,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21930,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35449,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22013,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22099,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22039,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22218,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22218,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24369,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22196,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22218,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22234,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {31379,-1},
        {31379,21254,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21959,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22294,-1},
        {-1},
        {-1,-1},
        {22259,22294,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20806,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23553,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20806,-1,-1}
    },
    {
        {22317,-1},
        {-1},
        {-1,-1},
        {22317,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22352,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22352,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34945,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23597,22575,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34945,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22500,-1,-1}
    },
    {
        {22533,-1},
        {-1},
        {-1,-1},
        {22533,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22486,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22442,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22667,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22585,-1,-1}
    },
    {
        {22602,-1},
        {-1},
        {-1,-1},
        {22602,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22396,-1,-1}
    },
    {
        {-1,-1},
        {22604},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22633,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22727,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22549,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22707,-1},
        {-1},
        {-1,-1},
        {22707,-1,-1}
    },
    {
        {22712,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22727,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22640,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22781,-1},
        {-1},
        {-1,-1},
        {22781,23551,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22781,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22796,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22799,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22799,-1},
        {-1},
        {-1,-1},
        {22799,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22799,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21319,-1},
        {33307,21319,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21460,21319,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22810,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21460,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33307,21319,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21460,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39131,-1,-1}
    },
    {
        {22818,-1},
        {-1},
        {-1,-1},
        {22818,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22828,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22839,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22840,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22839,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24059,21367,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22839,-1,-1}
    },
    {
        {27888,-1},
        {-1},
        {-1,-1},
        {27888,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22874,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22864,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22864,-1,-1}
    },
    {
        {158761,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29245,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29245,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23433,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23142,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23022,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23041,-1,-1}
    },
    {
        {23105,-1},
        {-1},
        {-1,-1},
        {23105,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23338,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23105,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23142,-1},
        {23142,-1,-1}
    },
    {
        {23231,-1},
        {-1},
        {-1,-1},
        {24871,23231,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23403,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23408,-1},
        {-1},
        {-1,-1},
        {23408,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23408,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23416,23398,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23416,23398,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23429,-1},
        {-1},
        {-1,-1},
        {23429,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20133,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23475,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20133,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23475,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23494,-1},
        {23494,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20901,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20901,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20901,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23492,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23527,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23494,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31432,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23527,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36785,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23563,-1,-1}
    },
    {
        {23561,-1},
        {-1},
        {-1,-1},
        {23561,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23553,-1,-1}
    },
    {
        {36785,-1},
        {-1},
        {-1,-1},
        {36785,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23565,23550,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23565,23550,-1}
    },
    {
        {22781,-1},
        {-1},
        {22781,23551},
        {22781,23551,-1}
    },
    {
        {22781,-1},
        {-1},
        {22781,23551},
        {-1,-1,-1}
    },
    {
        {22781,-1},
        {-1},
        {22781,23551},
        {22781,-1,-1}
    },
    {
        {22781,-1},
        {-1},
        {22781,23551},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23562,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23565,23550,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24541,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40654,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23601,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23601,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23612,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23611,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30473,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29312,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23660,23646,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23663,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23663,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20197,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25160,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23713,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22402,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22402,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22402,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23878,-1},
        {-1},
        {-1,-1},
        {23878,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24494,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24500,24501,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21427,22196,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35347,-1,-1}
    },
    {
        {13356,-1},
        {-1},
        {-1,-1},
        {13356,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24032,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24030,-1,-1}
    },
    {
        {24034,-1},
        {-1},
        {-1,-1},
        {24035,-1,-1}
    },
    {
        {24036,-1},
        {-1},
        {-1,-1},
        {24036,-1,-1}
    },
    {
        {24038,-1},
        {-1},
        {-1,-1},
        {24038,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24043,-1,-1}
    },
    {
        {24046,-1},
        {-1},
        {-1,-1},
        {24046,-1,-1}
    },
    {
        {24046,-1},
        {-1},
        {-1,-1},
        {24046,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21361,-1,-1}
    },
    {
        {21367,-1},
        {-1},
        {-1,-1},
        {24059,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24101,-1},
        {24101,-1,-1}
    },
    {
        {24076,-1},
        {-1},
        {-1,-1},
        {24076,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24107,-1},
        {-1},
        {-1,-1},
        {24107,24101,-1}
    },
    {
        {24107,-1},
        {-1},
        {-1,-1},
        {24101,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24118,24111,-1}
    },
    {
        {-1,-1},
        {-1},
        {24118,24111},
        {24118,24111,-1}
    },
    {
        {24107,-1},
        {-1},
        {-1,-1},
        {24107,24101,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38859,-1,-1}
    },
    {
        {24111,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24180,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24180,-1,-1}
    },
    {
        {24184,-1},
        {-1},
        {-1,-1},
        {24184,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24190,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24190,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24190,-1,-1}
    },
    {
        {24190,-1},
        {-1},
        {-1,-1},
        {24190,-1,-1}
    },
    {
        {24190,-1},
        {-1},
        {-1,-1},
        {24190,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24213,-1,-1}
    },
    {
        {24230,-1},
        {-1},
        {-1,-1},
        {24230,-1,-1}
    },
    {
        {24213,-1},
        {-1},
        {-1,-1},
        {24213,-1,-1}
    },
    {
        {24237,-1},
        {-1},
        {-1,-1},
        {24237,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30092,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24231,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24231,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24231,-1,-1}
    },
    {
        {24246,-1},
        {-1},
        {-1,-1},
        {24246,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24109,-1},
        {24109,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24109,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25993,40778,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24310,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24310,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24310,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24335,-1,-1}
    },
    {
        {24339,-1},
        {-1},
        {-1,-1},
        {24339,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24339,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24344,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22839,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24396,24357,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24375,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24375,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24375,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30330,30332,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33467,-1,-1}
    },
    {
        {24404,-1},
        {-1},
        {37682,37636},
        {24404,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33467,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33467,-1,-1}
    },
    {
        {24407,-1},
        {-1},
        {-1,-1},
        {24407,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24407,-1,-1}
    },
    {
        {24409,-1},
        {-1},
        {24409,21295},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24413,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24422,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24429,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24441,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24465,24452,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24481,-1},
        {-1},
        {-1,-1},
        {24481,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24536,-1},
        {24536,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24536,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24536,-1},
        {24536,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24565,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24650,21332,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24613,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24616,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24656,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24713,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24656,-1},
        {-1},
        {-1,-1},
        {24656,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {14703,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {14703,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24755,-1},
        {-1},
        {24499,24503},
        {24755,24499,24503}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {14703,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24859,-1},
        {-1},
        {-1,-1},
        {24859,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25010,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40654,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20674,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24917,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24942,-1,-1}
    },
    {
        {24962,-1},
        {-1},
        {-1,-1},
        {24962,-1,-1}
    },
    {
        {25040,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25085,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25014,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25108,-1},
        {-1},
        {-1,-1},
        {25108,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25108,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25108,37549},
        {25108,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25119,-1},
        {25119,-1,-1}
    },
    {
        {25114,-1},
        {-1},
        {-1,-1},
        {25114,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25122,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25138,25135,-1}
    },
    {
        {25144,-1},
        {-1},
        {25144,-1},
        {25144,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25152,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25152,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23616,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25147,25150,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25147,25150,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25307,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25237,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25237,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25308,25309,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25436,25628},
        {25628,25436,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25466,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26053,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33281,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26053,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27096,27171,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25691,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27096,27171,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25652,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {25758},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25903,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25908,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25912,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25910,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25305,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25942,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31574,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {63856,27578,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {63856,27578,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21460,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37087,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21189,-1,-1}
    },
    {
        {25954,-1},
        {-1},
        {-1,-1},
        {25954,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {63856,27578,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30330,30332,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25955,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25955,-1},
        {25955,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25958,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26000,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30330,30332,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25955,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25963,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25963,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25571,-1},
        {25571,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25577,-1,-1}
    },
    {
        {25958,-1},
        {-1},
        {-1,-1},
        {25958,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25958,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25975,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22800,-1},
        {-1},
        {22800,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25968,25976,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25973,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21427,22196,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26021,-1},
        {-1},
        {-1,-1},
        {26021,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26021,29255,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40718,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40718,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26041,-1,-1}
    },
    {
        {24340,-1},
        {-1},
        {-1,-1},
        {24340,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24343,-1},
        {-1},
        {-1,-1},
        {24343,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24344,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26689,33322},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26045,-1,-1}
    },
    {
        {26049,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20860,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21466,27573,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26063,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26063,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26053,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24375,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26051,-1,-1}
    },
    {
        {26082,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26088,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26088,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26355,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32996,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26355,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20882,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26157,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32963,-1,-1}
    },
    {
        {26127,-1},
        {-1},
        {-1,-1},
        {26127,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30475,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26359,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32997,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40736,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26185,26187,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26127,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33034,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26364,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26364,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26365,26366,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26367,-1,-1}
    },
    {
        {26369,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20250,26371},
        {20250,26371,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26292,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26373,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24433,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30366,-1,-1}
    },
    {
        {26312,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30067,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26388,-1,-1}
    },
    {
        {26388,-1},
        {-1},
        {-1,-1},
        {26388,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26388,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32929,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33240,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26397,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30643,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23387,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26414,-1},
        {26414,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {63856,27578},
        {63856,27578,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32177,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {26491},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26460,-1},
        {-1},
        {-1,-1},
        {26460,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32796,-1},
        {32796,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26576,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26580,-1,-1}
    },
    {
        {26517,-1},
        {-1},
        {-1,-1},
        {26517,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26746,26611,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26627,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26691,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26619,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26748,28422,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26580,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26688,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26688,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26688,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26820,-1,-1}
    },
    {
        {26893,-1},
        {-1},
        {-1,-1},
        {26893,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26688,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26748,28422,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26989,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26964,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26989,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26748,28422,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26748,28422,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26885,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {15309,-1},
        {-1},
        {-1,-1},
        {27123,15309,-1}
    },
    {
        {-1,-1},
        {-1},
        {26708,-1},
        {26708,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26688,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26839,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26839,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26839,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26840,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {33378,27299},
        {27299,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {145215,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {144954,27211,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {144954,27211,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27264,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27146,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27146,-1,-1}
    },
    {
        {27155,-1},
        {-1},
        {-1,-1},
        {27155,27004,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32789,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27177,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27146,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27171,-1},
        {27096,27171,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25856,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26934,-1},
        {-1,-1,-1}
    },
    {
        {27278,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {27370},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27264,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27315,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {145252,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24174,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27177,27402,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27277,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39729,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39729,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25496,36470},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27454,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27454,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27454,-1,-1}
    },
    {
        {27470,-1},
        {-1},
        {-1,-1},
        {27470,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27476,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27489,27475,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27492,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27492,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27491,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28041,28169,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28041,28169,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27507,-1,-1}
    },
    {
        {27507,-1},
        {-1},
        {-1,-1},
        {27507,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27507,-1},
        {-1},
        {-1,-1},
        {27507,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27515,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27515,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21365,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27519,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23403,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39131,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27573,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27573,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27579,27580,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27579,27580,-1}
    },
    {
        {27575,-1},
        {-1},
        {-1,-1},
        {27575,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25955,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27584,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27589,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27583,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27602,-1},
        {-1},
        {-1,-1},
        {27602,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27663,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27665,-1,-1}
    },
    {
        {27664,-1},
        {-1},
        {-1,-1},
        {27664,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27699,-1},
        {-1},
        {27699,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27704,-1,-1}
    },
    {
        {27714,-1},
        {-1},
        {-1,-1},
        {27714,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27908,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21442,21443,-1}
    },
    {
        {27784,-1},
        {-1},
        {-1,-1},
        {27784,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27954,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26579,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27945,-1},
        {-1},
        {-1,-1},
        {27945,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28154,27973,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28181,28149,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27779,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28145,-1,-1}
    },
    {
        {28238,-1},
        {-1},
        {-1,-1},
        {28238,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27784,28139},
        {28139,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28319,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28354,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28435,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28288,28415,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24917,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21295,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {15299,28422},
        {28422,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28644,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {25273,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28546,28548},
        {28546,28548,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28748,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28856,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28825,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28977,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28879,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28857,40670,-1}
    },
    {
        {-1,-1},
        {-1},
        {29105,-1},
        {29105,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28858,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29180,-1},
        {29180,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29078,-1},
        {-1},
        {-1,-1},
        {28976,29078,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28961,-1,-1}
    },
    {
        {28982,-1},
        {-1},
        {-1,-1},
        {28982,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20127,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20860,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {19998,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29071,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29066,-1},
        {29066,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29141,-1},
        {-1},
        {-1,-1},
        {29141,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24980,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29246,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29246,-1,-1}
    },
    {
        {23559,-1},
        {-1},
        {-1,-1},
        {23558,23559,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29279,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29309,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29310,30606},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {15968,35930,-1}
    },
    {
        {29431,-1},
        {-1},
        {-1,-1},
        {29431,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29503,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {64056,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29563,29486,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29563,29486,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29645,-1,-1}
    },
    {
        {29642,-1},
        {-1},
        {-1,-1},
        {29642,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29761,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29809,-1},
        {-1},
        {-1,-1},
        {29809,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29608,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29862,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38728,38666,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29926,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29926,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29926,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29926,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30741,30862,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29976,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29976,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29978,-1,-1}
    },
    {
        {29978,-1},
        {-1},
        {-1,-1},
        {29978,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29996,-1},
        {-1},
        {-1,-1},
        {29996,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29996,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20633,-1},
        {30041,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30007,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30041,-1},
        {-1},
        {-1,-1},
        {30041,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30041,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30028,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30032,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30057,-1,-1}
    },
    {
        {30064,-1},
        {-1},
        {-1,-1},
        {30064,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30050,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26292,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30074,30086,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30074,30086,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30090,30067,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30074,30086,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30090,30067,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30074,30086,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30092,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22799,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30097,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30097,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30142,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30216,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30235,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30306,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30328,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30331,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30331,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30330,30332,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35980,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30003,-1,-1}
    },
    {
        {-1,-1},
        {30340},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30394,-1},
        {30394,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30394,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30408,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {151637,-1,-1}
    },
    {
        {30408,-1},
        {-1},
        {-1,-1},
        {30408,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30435,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30435,-1,-1}
    },
    {
        {30433,-1},
        {-1},
        {-1,-1},
        {30433,23613,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30439,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30444,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30439,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30462,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64097,35222,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30494,-1},
        {-1},
        {-1,-1},
        {30494,30495,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30494,30495,-1}
    },
    {
        {-1,-1},
        {-1},
        {30475,-1},
        {30475,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30475,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30552,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34886,-1},
        {-1},
        {-1,-1},
        {34886,-1,-1}
    },
    {
        {34886,-1},
        {-1},
        {-1,-1},
        {34886,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30475,-1},
        {30475,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30552,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30475,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30609,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30669,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30683,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30683,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30683,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30686,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30977,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30933,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31038,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31059,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31085,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31085,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31085,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31117,-1},
        {-1},
        {-1,-1},
        {64082,31117,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64027,31119,-1}
    },
    {
        {31142,-1},
        {-1},
        {-1,-1},
        {31142,-1,-1}
    },
    {
        {31146,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25771,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31153,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31246,-1},
        {-1},
        {-1,-1},
        {31246,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31216,31281,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31293,-1,-1}
    },
    {
        {31293,-1},
        {-1},
        {-1,-1},
        {31293,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31293,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31293,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31308,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31293,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31329,-1},
        {-1},
        {-1,-1},
        {31329,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31354,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31354,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23448,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31428,-1,-1}
    },
    {
        {31419,-1},
        {-1},
        {-1,-1},
        {31379,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {31434},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31434,31363,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31485,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31532,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31639,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31718,31622,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31767,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31922,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31908,-1,-1}
    },
    {
        {31908,-1},
        {-1},
        {-1,-1},
        {31908,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31922,31874,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29136,31910,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38584,38583,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37350,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32305,32294,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32011,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32175,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {32065,64068},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31992,32114,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32291,30476,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32175,-1},
        {-1},
        {-1,-1},
        {32175,-1,-1}
    },
    {
        {32178,-1},
        {-1},
        {-1,-1},
        {32178,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32291,30476,-1}
    },
    {
        {-1,-1},
        {-1},
        {32361,32260},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32177,-1},
        {-1},
        {-1,-1},
        {32177,32178,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32177,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32153,32380,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32311,-1},
        {-1},
        {-1,-1},
        {32311,-1,-1}
    },
    {
        {32363,-1},
        {-1},
        {-1,-1},
        {32363,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32401,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32410,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32570,-1},
        {-1},
        {-1,-1},
        {27424,32570,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32607,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32608,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32608,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32631,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32178,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32631,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32654,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32670,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32670,-1},
        {-1},
        {-1,-1},
        {32670,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32681,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32701,-1},
        {-1},
        {-1,-1},
        {32701,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32722,-1,-1}
    },
    {
        {32719,-1},
        {-1},
        {-1,-1},
        {32719,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32719,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32719,-1,-1}
    },
    {
        {32735,-1},
        {-1},
        {-1,-1},
        {32735,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32763,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32769,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32774,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32829,-1},
        {-1},
        {-1,-1},
        {32829,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22768,32882,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32895,-1,-1}
    },
    {
        {32901,-1},
        {-1},
        {-1,-1},
        {31899,32901,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31899,32901,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31899,32901,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31899,32901,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32905,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32905,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32948,-1},
        {-1},
        {-1,-1},
        {32948,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33181,-1,-1}
    },
    {
        {33073,-1},
        {-1},
        {-1,-1},
        {33073,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33026,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33167,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33016,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33126,-1},
        {-1},
        {-1,-1},
        {33075,33126,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26407,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39512,39486,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33160,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33240,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30643,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33231,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33235,33247,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33240,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33251,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36066,33252,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33255,-1,-1}
    },
    {
        {33256,-1},
        {-1},
        {-1,-1},
        {33256,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33268,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33288,19982,33287}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {19982,33287,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33277,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33279,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33282,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33288,-1},
        {-1},
        {-1,-1},
        {33288,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37313,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37313,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35449,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33307,21319,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33307,21319,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33310,-1,-1}
    },
    {
        {33311,-1},
        {-1},
        {-1,-1},
        {33311,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33324,-1},
        {-1},
        {-1,-1},
        {33324,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33334,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33387,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33394,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33394,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33489,-1,-1}
    },
    {
        {-1,-1},
        {33477},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33521,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33618,-1},
        {-1},
        {-1,-1},
        {33618,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33618,-1,-1}
    },
    {
        {33506,-1},
        {-1},
        {-1,-1},
        {33505,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33579,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33521,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33618,-1},
        {-1},
        {-1,-1},
        {33618,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33759,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33725,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33740,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33744,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33777,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33775,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33728,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33744,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {158941,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34136,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33865,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33695,-1,-1}
    },
    {
        {33761,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33796,-1},
        {-1},
        {-1,-1},
        {33796,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33865,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26989,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34098,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34152,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34065,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34152,-1},
        {-1},
        {-1,-1},
        {34152,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34085,-1},
        {-1},
        {-1,-1},
        {33905,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34101,34255,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34101,34255,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34186,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {17731,-1},
        {-1},
        {-1,-1},
        {17731,34093,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33976,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34276,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33457,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34145,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29554,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31339,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34314,-1},
        {-1},
        {34176,34314},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34196,-1,-1}
    },
    {
        {34330,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {28525,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29191,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37962,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21892,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37313,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34381,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34381,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34381,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34381,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34382,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34382,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34384,-1,-1}
    },
    {
        {34382,-1},
        {-1},
        {-1,-1},
        {34382,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34388,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34384,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34382,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34382,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24248,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34394,34395,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34389,20966,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34394,34395,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34389,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34390,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34398,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34468,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34468,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34468,-1,-1}
    },
    {
        {34475,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34473,-1},
        {-1},
        {-1,-1},
        {34473,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34468,-1},
        {-1},
        {-1,-1},
        {34468,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34468,-1,-1}
    },
    {
        {34574,-1},
        {-1},
        {-1,-1},
        {34574,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34588,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39757,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34588,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34678,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34699,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34645,-1,-1}
    },
    {
        {34733,-1},
        {-1},
        {-1,-1},
        {34733,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34847,-1},
        {-1},
        {-1,-1},
        {34847,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34873,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34880,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34884,-1,-1}
    },
    {
        {34886,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30410,-1},
        {-1},
        {-1,-1},
        {30410,-1,-1}
    },
    {
        {34886,-1},
        {-1},
        {-1,-1},
        {34886,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40697,-1,-1}
    },
    {
        {34886,-1},
        {-1},
        {-1,-1},
        {34886,-1,-1}
    },
    {
        {40697,-1},
        {-1},
        {-1,-1},
        {40697,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34899,-1},
        {-1},
        {-1,-1},
        {34899,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34907,34910,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34907,34910,-1}
    },
    {
        {34907,-1},
        {-1},
        {-1,-1},
        {34907,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34928,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34928,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34928,-1},
        {-1},
        {-1,-1},
        {34928,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34928,-1,-1}
    },
    {
        {34996,-1},
        {-1},
        {-1,-1},
        {34996,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35028,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35023,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34996,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35076,26950},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35140,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35140,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35120,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35203,-1,-1}
    },
    {
        {35203,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36983,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35261,35239,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35264,35251,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35264,35251,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35282,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35282,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35299,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35299,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35299,-1,-1}
    },
    {
        {35303,-1},
        {-1},
        {-1,-1},
        {35299,-1,-1}
    },
    {
        {35299,-1},
        {-1},
        {-1,-1},
        {35299,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35349,-1,-1}
    },
    {
        {35373,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35500,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35442,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {18188,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35449,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35569,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35615,-1,-1}
    },
    {
        {35635,-1},
        {-1},
        {-1,-1},
        {35635,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35465,35709,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35591,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35649,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {35651,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35628,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36895,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35591,-1,-1}
    },
    {
        {18188,-1},
        {-1},
        {-1,-1},
        {18188,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35742,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35542,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35698,35731,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35920,35914,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35980,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {14197,-1,-1}
    },
    {
        {36031,-1},
        {-1},
        {-1,-1},
        {36031,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36020,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36010,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36049,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36010,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36051,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36042,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36058,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36106,36059,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36106,36059,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36062,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36225,-1},
        {36225,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36299,-1},
        {36299,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36481,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36767,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36767,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36776,-1},
        {-1},
        {-1,-1},
        {36776,24321,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36784,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36784,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36786,-1,-1}
    },
    {
        {64102,-1},
        {-1},
        {-1,-1},
        {36789,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36804,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21705,-1},
        {21305,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24311,-1},
        {24311,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24310,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24310,-1,-1}
    },
    {
        {36848,-1},
        {-1},
        {-1,-1},
        {36848,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21280,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {24452,36885},
        {24452,36885,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24237,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36999,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36960,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36949,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36942,-1},
        {-1},
        {-1,-1},
        {36942,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36960,-1},
        {36960,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36963,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36794,37002,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21312,21306,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36996,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36794,37002,-1}
    },
    {
        {36983,-1},
        {-1},
        {-1,-1},
        {36983,-1,-1}
    },
    {
        {37001,-1},
        {-1},
        {-1,-1},
        {37001,36794,37002}
    },
    {
        {36978,-1},
        {-1},
        {-1,-1},
        {36978,36933,-1}
    },
    {
        {36996,-1},
        {-1},
        {-1,-1},
        {36996,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37001,36794,37002}
    },
    {
        {37002,-1},
        {-1},
        {-1,-1},
        {36794,37002,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36999,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21407,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37145,-1,-1}
    },
    {
        {37145,-1},
        {-1},
        {-1,-1},
        {37145,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37193,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37193,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37195,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37195,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37290,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37312,37304,-1}
    },
    {
        {37320,-1},
        {-1},
        {-1,-1},
        {37320,37323,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37325,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22516,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37946,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37793,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37953,-1},
        {-1},
        {-1,-1},
        {37953,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38002,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39658,39662,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38373,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38512,-1,-1}
    },
    {
        {-1,-1},
        {38576},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38560,38577},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {63964,38534,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38560,38577,-1}
    },
    {
        {-1,-1},
        {-1},
        {38501,38519},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38477,-1,-1}
    },
    {
        {38570,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38618,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38587,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38618,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38605,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38618,-1,-1}
    },
    {
        {38627,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {64104,-1},
        {-1},
        {-1,-1},
        {64104,38627,-1}
    },
    {
        {-1,-1},
        {-1},
        {38617,21452},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38632,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23477,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38632,29246,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38663,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38692,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38897,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20150,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20150,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21959,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38761,-1,-1}
    },
    {
        {38761,-1},
        {-1},
        {-1,-1},
        {38761,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38761,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38859,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38859,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38859,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38859,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {34212,-1},
        {-1,-1,-1}
    },
    {
        {38897,-1},
        {-1},
        {-1,-1},
        {38897,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39015,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38948,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39006,-1},
        {-1},
        {-1,-1},
        {63952,39006,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64106,38971,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39025,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {39100,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39106,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35909,-1,-1}
    },
    {
        {39149,-1},
        {-1},
        {-1,-1},
        {39149,-1,-1}
    },
    {
        {39178,-1},
        {-1},
        {-1,-1},
        {39178,-1,-1}
    },
    {
        {39178,-1},
        {-1},
        {-1,-1},
        {39178,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39377,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39438,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39493,39366,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39592,-1},
        {-1},
        {-1,-1},
        {39592,-1,-1}
    },
    {
        {39592,-1},
        {-1},
        {-1,-1},
        {39592,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39616,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39608,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39637,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39662,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {30896,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39717,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39727,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39729,-1,-1}
    },
    {
        {39730,-1},
        {-1},
        {-1,-1},
        {39730,-1,-1}
    },
    {
        {39730,-1},
        {-1},
        {-1,-1},
        {39730,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39730,-1,-1}
    },
    {
        {39740,-1},
        {-1},
        {-1,-1},
        {39740,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39747,-1},
        {-1},
        {-1,-1},
        {39747,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39770,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39770,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39770,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39770,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39822,-1},
        {-1},
        {-1,-1},
        {39822,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39791,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39796,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40056,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39998,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40165,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40199,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40219,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40213,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40441,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40179,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40575,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40575,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40575,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40599,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40599,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40599,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40599,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31946,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39238,-1,-1}
    },
    {
        {40644,-1},
        {-1},
        {-1,-1},
        {40644,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40653,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40653,-1,-1}
    },
    {
        {40656,-1},
        {-1},
        {-1,-1},
        {40656,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40736,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40748,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40763,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40763,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40763,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40763,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40763,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40763,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25993,40778,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25993,40778,-1}
    },
    {
        {27503,-1},
        {-1},
        {-1,-1},
        {40786,27503,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40786,27503,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40845,-1},
        {-1},
        {-1,-1},
        {31452,40845,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31452,40845,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20096,40860,-1}
    },
    {
        {40860,-1},
        {-1},
        {-1,-1},
        {20096,40860,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20096,40860,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20096,40860,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20096,40860,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {20130},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25165,-1},
        {-1},
        {-1,-1},
        {25165,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21916,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22025,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21359,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {26274},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35140,-1},
        {-1},
        {-1,-1},
        {35140,-1,-1}
    },
    {
        {-1,-1},
        {40778},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40779,-1},
        {-1},
        {-1,-1},
        {40779,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20170,-1,-1}
    },
    {
        {-1,-1},
        {20037},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20304,-1,-1}
    },
    {
        {20234,-1},
        {-1},
        {-1,-1},
        {20234,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20320,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {20233},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {22745},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20355,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {23729},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {20551},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {20628},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {20860,-1},
        {20860,-1,-1}
    },
    {
        {26371,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20744,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {34907},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {20789},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20818,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20170,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33287,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33287,-1,-1}
    },
    {
        {-1,-1},
        {33287},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37195,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33287,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33287,-1,-1}
    },
    {
        {-1,-1},
        {33287},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33287,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {22880},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23459,-1,-1}
    },
    {
        {-1,-1},
        {20901},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27726,-1,-1}
    },
    {
        {-1,-1},
        {27733},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {27766},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {27915},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28113,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28147,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {28216},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {28322},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {28422},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {64021,-1},
        {-1},
        {-1,-1},
        {29081,-1,-1}
    },
    {
        {-1,-1},
        {20961},
        {20961,20962},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {39082},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {26388},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {21084},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {36776,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22025,-1,-1}
    },
    {
        {-1,-1},
        {21162},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {13599,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {21206},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {33671},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21208,-1},
        {-1},
        {-1,-1},
        {21208,-1,-1}
    },
    {
        {21208,-1},
        {-1},
        {-1,-1},
        {21208,-1,-1}
    },
    {
        {64052,-1},
        {-1},
        {-1,-1},
        {64052,-1,-1}
    },
    {
        {64052,-1},
        {-1},
        {-1,-1},
        {64052,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {34389},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33618,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {21332},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21338,-1},
        {-1},
        {-1,-1},
        {21338,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21338,24893},
        {21338,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21338,24893},
        {21338,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {21338,24893},
        {21338,-1,-1}
    },
    {
        {-1,-1},
        {21338},
        {21338,24893},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37111,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36784,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36784,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34382,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {22173},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21535,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {21843},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {21828},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {21972,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {38902},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22781,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22781,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {22781},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {21929},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22030,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {22227},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22222,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {22272},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22343,-1,-1}
    },
    {
        {-1,-1},
        {22352},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {26627},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {25383},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {135963},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22435,-1,-1}
    },
    {
        {-1,-1},
        {22740},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22615,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22580,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {22714},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {29887},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {22739,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22025,-1,-1}
    },
    {
        {22775,-1},
        {-1},
        {-1,-1},
        {22775,-1,-1}
    },
    {
        {-1,-1},
        {22781},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22781,-1,-1}
    },
    {
        {-1,-1},
        {22218},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {22793},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {20778},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {25919},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22799,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22799,-1,-1}
    },
    {
        {-1,-1},
        {26790},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21319,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {22872},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22885,-1,-1}
    },
    {
        {-1,-1},
        {28417},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27344,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {23162},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {25173},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {23416},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {23416},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23416,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23416,-1,-1}
    },
    {
        {-1,-1},
        {23435},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {23472,-1},
        {23472,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23515,-1},
        {-1},
        {-1,-1},
        {23515,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {23532},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {23542},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23542,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23559,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {23567},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {23643},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {24264},
        {24264,21414},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {29312},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {33452},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23697,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {23720},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {138657},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {23975},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23948,-1,-1}
    },
    {
        {-1,-1},
        {23993},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {24012},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {24012},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {27119},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24012,-1,-1}
    },
    {
        {-1,-1},
        {24013},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {24022},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24022,-1},
        {-1},
        {-1,-1},
        {24022,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24022,-1},
        {-1},
        {-1,-1},
        {24022,-1,-1}
    },
    {
        {24059,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {14390,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {20053},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20053,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {24609},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {24604},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26274,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {24892},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24164,-1},
        {-1},
        {-1,-1},
        {24163,-1,-1}
    },
    {
        {-1,-1},
        {20130},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {30123},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {30171},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24231,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24231,-1,-1}
    },
    {
        {-1,-1},
        {34389},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24248,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {33210},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {40848},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {40637},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {40441},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {24336},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24339,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24396,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24396,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24396,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24396,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24396,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {24389},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24396,-1,-1}
    },
    {
        {24389,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {30086},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {26164},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24409,-1},
        {-1},
        {24409,21295},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24422,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {20227},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {24441},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {24480},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {24495},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24478,-1},
        {-1},
        {-1,-1},
        {24478,-1,-1}
    },
    {
        {24499,-1},
        {-1},
        {-1,-1},
        {24503,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24505,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {24735},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {24735},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {24685},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {24536},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {24676},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {24685},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24735,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27888,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27888,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {28932},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {24985},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {24594},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {24835},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {24824},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {24776},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24800,-1},
        {-1},
        {-1,-1},
        {24800,-1,-1}
    },
    {
        {-1,-1},
        {24754},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {24939},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {25005},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29081,-1,-1}
    },
    {
        {24859,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {25085},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24935,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25035,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {25785},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {25088},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {25102},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33538,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33538,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34255,-1,-1}
    },
    {
        {-1,-1},
        {25165},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25463,-1,-1}
    },
    {
        {-1,-1},
        {142365},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {25765},
        {25765,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {25463},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25805,-1},
        {-1},
        {-1,-1},
        {25805,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {25831},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {25900},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {25976,-1},
        {-1},
        {-1,-1},
        {25976,-1,-1}
    },
    {
        {-1,-1},
        {31296},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27468,-1,-1}
    },
    {
        {25993,-1},
        {-1},
        {-1,-1},
        {25993,-1,-1}
    },
    {
        {-1,-1},
        {25998},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {25993},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {25993},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {25919},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {25919},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {25919},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {32654},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {35722},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {39745},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {40718},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26049,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {26060},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {26114},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26363,-1,-1}
    },
    {
        {26119,-1},
        {-1},
        {-1,-1},
        {26119,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26363,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26363,-1,-1}
    },
    {
        {-1,-1},
        {33031},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {26184},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {26185},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {134625},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30566,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {26219},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26241,26313,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26313,-1},
        {-1},
        {-1,-1},
        {26241,26313,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {173746},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {26345},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {26308},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {26230,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {38902},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {32929},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {26505},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {26862},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26563,-1},
        {-1,-1,-1}
    },
    {
        {26583,-1},
        {-1},
        {-1,-1},
        {26583,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26617,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {27250},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26746,-1,-1}
    },
    {
        {-1,-1},
        {27141},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {26880},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {26746,-1,-1}
    },
    {
        {-1,-1},
        {26978},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {32092},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {27141},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {27047},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {31263},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27022,-1,-1}
    },
    {
        {-1,-1},
        {26933},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {27097},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27106,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27224,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {26974},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {144954,27211,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {144954,27211,-1}
    },
    {
        {27224,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27194,-1,-1}
    },
    {
        {-1,-1},
        {26997},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {27035},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27035,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27402,-1},
        {-1},
        {-1,-1},
        {27402,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {31806},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27427,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {28937},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {38741},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27507,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27507,-1,-1}
    },
    {
        {27507,-1},
        {-1},
        {-1,-1},
        {27507,-1,-1}
    },
    {
        {-1,-1},
        {28113},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27573,-1,-1}
    },
    {
        {-1,-1},
        {27584},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {27575},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27589,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {27714,-1},
        {-1},
        {-1,-1},
        {27714,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {27743},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {27773},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {146686,-1},
        {-1},
        {-1,-1},
        {27888,-1,-1}
    },
    {
        {146686,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {27888,-1,-1}
    },
    {
        {27941,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {28288},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28149,-1},
        {-1},
        {-1,-1},
        {28181,28149,-1}
    },
    {
        {28149,-1},
        {-1},
        {-1,-1},
        {28181,28149,-1}
    },
    {
        {-1,-1},
        {28052},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {28290},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {28634},
        {28634,-1},
        {-1,-1,-1}
    },
    {
        {28304,-1},
        {-1},
        {-1,-1},
        {28304,-1,-1}
    },
    {
        {-1,-1},
        {28111},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28149,-1},
        {-1},
        {-1,-1},
        {28181,28149,-1}
    },
    {
        {28251,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {28545},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28288,-1,-1}
    },
    {
        {-1,-1},
        {28196},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28147,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28234,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28288,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28415,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {28196},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40654,-1,-1}
    },
    {
        {-1,-1},
        {28322},
        {28322,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {28361},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {28552},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28234,-1,-1}
    },
    {
        {-1,-1},
        {28168},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28415,-1},
        {-1},
        {-1,-1},
        {28415,-1,-1}
    },
    {
        {28415,-1},
        {-1},
        {-1,-1},
        {28415,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28415,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28415,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28415,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28415,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28415,-1,-1}
    },
    {
        {-1,-1},
        {28415},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {28415},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28525,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {28511},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {28722},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28415,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28415,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28415,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28661,-1},
        {-1},
        {-1,-1},
        {28661,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {28575},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {28598},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {28652},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {34180},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {28652},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28760,-1,-1}
    },
    {
        {-1,-1},
        {28629},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28415,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28415,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28415,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {28415,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {28743},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {28670},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {28415},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {40251},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {154699},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {28722},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {28639},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {28639},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {29031},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {40658,-1,-1}
    },
    {
        {28943,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {28961},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29031,-1,-1}
    },
    {
        {-1,-1},
        {29121},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {64021},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {24904},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {29121},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29121,-1,-1}
    },
    {
        {-1,-1},
        {64021},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {29105},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {64021},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {64021},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {29197},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {23559,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29314,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {29494},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {29450},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29467,-1,-1}
    },
    {
        {-1,-1},
        {25589},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26053,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {29801},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {29738,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {26234},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {29761},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {29792},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29730,-1,-1}
    },
    {
        {29785,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {22533,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {29849},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {22699},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {29864},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {29999},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30065,-1,-1}
    },
    {
        {-1,-1},
        {30063},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {30090},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {30169},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {30278},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34255,-1,-1}
    },
    {
        {-1,-1},
        {30306},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30352,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {26243},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {40725},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {30410,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30494,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30494,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26241,26313},
        {26313,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {26241,26313},
        {26313,-1,-1}
    },
    {
        {-1,-1},
        {26329},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {30697,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {30906},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {31013},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31062,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {34987},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {64079},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34996,-1,-1}
    },
    {
        {-1,-1},
        {35012},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {25998,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {31085},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {31277},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31146,-1},
        {-1},
        {-1,-1},
        {31146,31109,-1}
    },
    {
        {-1,-1},
        {35137},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {64027},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {31155},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {31152},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {31178},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {31201},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {31201},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64026,31077,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {161412},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64083,31118,-1}
    },
    {
        {31293,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {31302},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31298,-1},
        {-1},
        {-1,-1},
        {31298,-1,-1}
    },
    {
        {31312,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {40866},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31344,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40861,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31354,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {31402},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {31427},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {24432,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {31499},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31505,-1},
        {-1},
        {-1,-1},
        {31505,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31545,-1,-1}
    },
    {
        {-1,-1},
        {31807},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31680,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {33862},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {154699,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31684,-1,-1}
    },
    {
        {-1,-1},
        {32789},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31761,-1,-1}
    },
    {
        {31718,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {31823,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {31734},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {31918},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {31929},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {40654},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32090,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {32099},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {32098},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {32294},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32178,-1},
        {-1},
        {-1,-1},
        {32178,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {32190},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {32273},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {32383,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {32570},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {32592},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {32654},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32654,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {32681,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {32764},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {32725},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {32748},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {31990},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {32883},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {30519},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33021,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33144,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {33274},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {33275},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {21471},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33287,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33287,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {33287},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {33287},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {33287},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {33294},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {33453},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33489,-1,-1}
    },
    {
        {33618,-1},
        {-1},
        {-1,-1},
        {33618,-1,-1}
    },
    {
        {33618,-1},
        {-1},
        {-1,-1},
        {33618,-1,-1}
    },
    {
        {33618,-1},
        {-1},
        {-1,-1},
        {33618,-1,-1}
    },
    {
        {-1,-1},
        {33433},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {33634},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33618,-1,-1}
    },
    {
        {33618,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33618,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33777,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {33594},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33775,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33775,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31680,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {33618,-1,-1}
    },
    {
        {-1,-1},
        {34072},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {34157},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34001,-1,-1}
    },
    {
        {-1,-1},
        {33586},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {20860,-1,-1}
    },
    {
        {34098,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {34081},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {34193},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {33976},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34255,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {24413},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34219,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {34204},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34276,-1},
        {-1},
        {-1,-1},
        {34276,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34269,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {155041},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34326,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {34322,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {34390},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {40607},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {34510},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {34674},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {34905},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {34910,-1,-1}
    },
    {
        {34915,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {31036,-1,-1}
    },
    {
        {-1,-1},
        {21021},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {64083,31118,-1}
    },
    {
        {-1,-1},
        {31143},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {31153},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {32645},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35258,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35258,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35258,-1,-1}
    },
    {
        {35261,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {35463},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {35552},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {64099,-1},
        {-1},
        {-1,-1},
        {64099,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35609,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {35670},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {20989},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {33945},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {27589},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {35962},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35998,-1,-1}
    },
    {
        {-1,-1},
        {40782},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36066,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36066,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {40632},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36234,-1,-1}
    },
    {
        {-1,-1},
        {36452},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {36426},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {36452},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {36452},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {36452},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {36837},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {36837},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {36999},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {36982},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {36960,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37002,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37001,-1},
        {-1},
        {-1,-1},
        {37001,-1,-1}
    },
    {
        {37001,-1},
        {-1},
        {-1,-1},
        {37001,-1,-1}
    },
    {
        {37001,-1},
        {-1},
        {-1,-1},
        {37001,-1,-1}
    },
    {
        {37001,-1},
        {-1},
        {-1,-1},
        {37001,-1,-1}
    },
    {
        {37001,-1},
        {-1},
        {-1,-1},
        {37001,-1,-1}
    },
    {
        {37001,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37002,-1},
        {-1},
        {-1,-1},
        {37002,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37001,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {37002,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37001,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {37040},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {37239},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37325,-1},
        {-1},
        {-1,-1},
        {37325,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {37361},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {37593},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {37466},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {21137},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {37724},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {37941},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {37957},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {37586},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {37941},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {37780},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {37805,-1},
        {-1},
        {37805,37806},
        {37805,-1,-1}
    },
    {
        {-1,-1},
        {37593},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {37808},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {37941},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {37782},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {166553},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38306,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38364,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38364,-1,-1}
    },
    {
        {-1,-1},
        {38366},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {36491},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {38582},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {30655},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {38609,38620,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {30655},
        {30655,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {38670},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {35210},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {38799},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {38876},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {38876},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {31310},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {31310},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {38997},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {39178},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {169753},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {39232},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {39232},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39340,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {39501},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {39490},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {39644},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39647,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {39715},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {39993},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {39957},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {39940},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {40199},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40353,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {40383},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {40476},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {40577,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {40618},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {40621},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {40788},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {40803},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {40800},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {40806},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {40810},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {40568},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {40823},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {40818},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {40822},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {40860},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {22343,-1,-1}
    },
    {
        {28251,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33618,-1},
        {-1},
        {-1,-1},
        {33618,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {21218,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23494,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {32776},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33775,-1},
        {-1},
        {-1,-1},
        {33775,-1,-1}
    },
    {
        {24409,-1},
        {-1},
        {24409,21295},
        {24409,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29275,20117},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {27488,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20898,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {20900,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {24013,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28095,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {28095,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {29589,-1},
        {-1,-1,-1}
    },
    {
        {29832,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {33717,-1},
        {-1},
        {-1,-1},
        {33717,33759,-1}
    },
    {
        {35216,-1},
        {-1},
        {-1,-1},
        {35216,35258,35226}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {37682,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {38911,-1},
        {-1},
        {-1,-1},
        {38911,-1,-1}
    },
    {
        {39498,-1},
        {-1},
        {-1,-1},
        {39498,-1,-1}
    },
    {
        {131883,-1},
        {-1},
        {-1,-1},
        {36437,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23387,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {39368,39366,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {35032,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {29314,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {23041,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {159736,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {35524,-1},
        {-1},
        {-1,-1},
        {35524,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {36463,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {38501,38519},
        {38501,38519,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {-1},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {-1,-1},
        {39232},
        {-1,-1},
        {-1,-1,-1}
    },
    {
        {39178,-1},
        {-1},
        {-1,-1},
        {39178,-1,-1}
    }
};